#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>

#include <ToolKit/ActionAreaClass.h>

/*
 * Function: AmActionAreaClass::initializeValues()
 *
 * Purpose:
 *   Initialize the member variables to sane values.
 */

void
AmActionAreaClass::initializeValues()
{
    _pwcForm       = NULL;
    _pwcSeparator  = NULL;
    _pwcButtonForm = NULL;
    _nButtons      = 0;
}


/*
 * Function: AmActionAreaClass::createInterface()
 *
 * Purpose:
 *   Creates the widgets used in the AmActionAreaClass.  The
 *   action area consists of a form that contains a separator
 *   and under the seprator a form that contains evenly spaced
 *   action buttons.
 */

void
AmActionAreaClass::createInterface()
{

/* Create the layout widgets used in the action area. */ 

    _pwcForm = new AmFormClass( _wParent, _szName);
    _pwcForm->AttachLeft();
    _pwcForm->AttachRight();
    _pwcForm->AttachBottom();

    _pwcSeparator = new AmSeparatorClass( _pwcForm, "actionAreaSeparator");
    _pwcSeparator->AttachLeft();
    _pwcSeparator->AttachTop();
    _pwcSeparator->AttachRight();

    _pwcButtonForm = new AmFormClass( _pwcForm, "actionAreaButtonForm");
    _pwcButtonForm->AttachLeft();
    _pwcButtonForm->AttachTop( _pwcSeparator);
    _pwcButtonForm->AttachRight();
    _pwcButtonForm->AttachBottom();

/* Assign the top most widget to be the base widget. */

    _wBase = *_pwcForm;
 
/* Hook in a destroy callback. */

    addDestroyCallback();
}


/*
 * Function: AmActionAreaClass::AmActionAreaClass()
 *
 * Purpose:
 *   Default constructor.
 *
 * Comments:
 *   Clears out all the fields specific to AmActionAreaClass also 
 *   calls the AmBaseWidgetClass default constructor.
 */

AmActionAreaClass::AmActionAreaClass()
    : AmBaseWidgetClass()
{
    initializeValues();
}


/*
 * Functions: AmActionAreaClass::AmActionAreaClass( wParent, szName)
 *            AmActionAreaClass::AmActionAreaClass( pwcParent, szName)
 *
 * Purpose:
 *   Constructor for the AmActionAreaClass.  Both functions work
 *   the same way except that one takes a widget and the other
 *   takes a "wrapped" widget.
 *
 * Parameters:
 *   Widget          wParent    : Parent widget of the new widget.
 *   AmBaseWidgetClass *pwcParent : Pointer to "wrapped" parent widget.
 *   const char      *szName    : Name of the widget.
 * 
 * Comments:
 *   Creates a action area widget class which consists of two forms, a
 *   separator, and as many buttons as are added later.  The constructor
 *   uses the passed in parent and passed in name as the widget's parent
 *   and widget's name respectively.  The AmBaseWidgetClass constructor is
 *   called with the widget and name arguments so that everything gets
 *   initialized correctly.
 */

AmActionAreaClass::AmActionAreaClass( Widget     wParent,
                                      const char *szName)
    : AmBaseWidgetClass( wParent, szName)
{
    initializeValues();
    createInterface();
}


/* See comment above. */

AmActionAreaClass::AmActionAreaClass( AmBaseWidgetClass *pwcParent,
                                      const char        *szName)
    : AmBaseWidgetClass( pwcParent, szName)
{
    initializeValues();
    createInterface();
}


/*
 * Function: AmActionAreaClass::~AmActionAreaClass()
 *
 * Purpose:
 *   Destructor for the AmActionAreaClass.
 *
 * Comments:
 *   Frees up all the memory used by the widget.
 *    
 */

AmActionAreaClass::~AmActionAreaClass()
{
}


/*
 * Function: AmActionAreaClass::AddButton( szName, xcp, xp)
 *
 * Purpose:
 *   Add a button to the action area.  AddButton also associated the
 *   passed in callback with the button.
 *
 * Parameters:
 *   const char     *szName : Name of the button.
 *   XtCallbackProc xcp     : Callback function to associated with
 *                                 the button.
 *   XtPointer      xp      : Client data to pass to the callback
 *                            function.
 *
 * Comments:
 *   Adds a button to the action area.  The button gets added to the right
 *   of the existing buttons.  The buttons are evenly spaced along the
 *   widths of the action area.  If a callback function is given then that
 *   function is associated with the button's activate callback.
 */

AmButtonClass *
AmActionAreaClass::AddButton( char           *szName,
                              XtCallbackProc xcp,
                              XtPointer      xp)
{
    AmButtonClass *pwcButton;
    int           iWidth;

/* One more button. */

    _nButtons += 1;   

/* Change the spacing of the existing buttons. */

    _pwcButtonForm->SetFractionBase( _nButtons * 2);

/* Add the new button on the right of the existing buttons. */

    pwcButton = new AmButtonClass( _pwcButtonForm, szName, xcp, xp);
    pwcButton->AttachLeft( 1 + ( _nButtons - 1) * 2);
    pwcButton->AttachTop();
    pwcButton->AttachBottom();

/* Center the button on the fraction base divison. */

    iWidth = pwcButton->GetWidth();
    pwcButton->LeftOffset(( - iWidth) / 2); 

    return( pwcButton);
}
