#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>

#include <ToolKit/BaseWidgetClass.h>

/*
 * Function: AmBaseWidgetClass::initializeValues()
 *
 * Purpose:
 *   Initialize the member variables to sane values.
 */

void
AmBaseWidgetClass::initializeValues()
{
    _wParent          = NULL;
    _szName           = NULL;
    _wBase            = NULL;

    _fDestroyCallback = FALSE;
}

void
AmBaseWidgetClass::setDefaultResources( const char **szResources)
{
    Display     *pDisplay;
    XrmDatabase xDatabase;
    char        sz[ RESLEN], *szAppName, szClassName;
    int         ii;

    if ( _wParent != NULL) {

    /* Initializations. */

        pDisplay  = XtDisplay( _wParent);
        xDatabase = XtDatabase( pDisplay);

        XtGetApplicationNameAndClass( pDisplay,
                                      ( String *) &szAppName,
                                      ( String *) &szClassName);

    /* Go through each of the defaults and add them to the database. */

        for ( ii = 0; szResources[ ii] != NULL; ii++) {

        /*  
         * Generate the resource string according to the AmFirst letter
         *   of the passed in string "-" means that the string is not
         *   to be modified other than removing the "-", "+"  means that
         *   application name is to be prepended in addition to the widget's
         *   name, and everything else means that the widget's name is to
         *   be prepended.
         */

            if ( szResources[ ii][ 0] == '-') { 
                sprintf( sz, "%s", &szResources[ ii][ 1]);
            } else if ( szResources[ ii][ 0] == '+') { 
                if ( _szName != NULL) {
                    sprintf( sz, "%s*%s%s", szAppName, _szName,
                             &szResources[ ii][ 1]);
                } else {
                    sprintf( sz, "%s*%s", szAppName, &szResources[ ii][ 1]);
                }
            } else {
                if ( _szName != NULL) {
                    sprintf( sz, "%s%s", _szName, szResources[ ii]);
                } else {
                    sprintf( sz, "%s", szResources[ ii]);
                }
            }

        /* Add the resource into the database. */

            XrmPutLineResource( &xDatabase, sz);
        }
    }
}


void
AmBaseWidgetClass::getResources( Widget               w,
                                 const XtResourceList xResources,
                                 const int            nResources)
{
    int ii;

    if ( w == NULL) { 
        w = _wBase;
    }

    if ( w != NULL) {
        for ( ii = 0; ii < nResources; ii++) {
            XtGetSubresources( w,
                               ( XtPointer) this,
                               NULL, NULL,
                               &xResources[ ii], 
                               1, NULL, 0);
        }
    }
}

/*
 * Function: AmBaseWidgetClass::AmBaseWidgetClass()
 *
 * Purpose:
 *   Default constructor.
 *
 * Comments:
 *   Clears out all the fields specific to AmBaseWidgetClass.
 */

AmBaseWidgetClass::AmBaseWidgetClass()
{
    initializeValues();
}


/*
 * Functions: AmBaseWidgetClass::AmBaseWidgetClass( wParent, szName)
 *            AmBaseWidgetClass::AmBaseWidgetClass( bwcParent, szName)
 *
 * Purpose:
 *   Constructor for the most basic widget.  Both functions work
 *   the same way except that one takes a widget and the other
 *   takes a "wrapped" widget.
 *
 * Parameters:
 *   Widget     wParent : Parent widget of the new widget.
 *   const char *szName : Name of the widget.
 * 
 * Description:
 *   Creates a widget class using the passed in name as the widget's name.
 */

AmBaseWidgetClass::AmBaseWidgetClass( Widget     wParent,
                                      const char *szName)
{
    initializeValues();

    _wParent = wParent;

    if ( szName != NULL) {
        _szName = strdup( szName);
    } else {
        _szName = NULL;
    }
}


/* See comment above. */

AmBaseWidgetClass::AmBaseWidgetClass( AmBaseWidgetClass *bwcParent,
                                      const char      *szName)
{
    initializeValues();

    if ( bwcParent != NULL) {
        _wParent = bwcParent->GetWidget();
    } else {
        _wParent = NULL;
    }

    if ( szName != NULL) {
        _szName = strdup( szName);
    } else {
        _szName = NULL;
    }
}


/*
 * Function: AmBaseWidgetClass::~AmBaseWidgetClass()
 *
 * Purpose:
 *   Destructor for the most basic widget.
 *
 * Description:
 *   Frees up all the memory used by the widget.
 *    
 */

AmBaseWidgetClass::~AmBaseWidgetClass()
{
    removeDestroyCallback();

    if ( _szName != NULL) {
        free( _szName);
    }

    if ( _wBase != NULL) {
        XtDestroyWidget( _wBase);
        _wBase = NULL;
    }
}

/*
 * Function: AmBaseWidgetClass::show()
 *
 * Purpose:
 *   Display the widget and its' children.
 *
 * Comments:
 *   This is the private (specific to AmBaseWidgetClass) function
 *   to manage and display the widget.
 */

void
AmBaseWidgetClass::show()
{
    if (( _wBase != NULL) &&
	( XtIsRealized( _wBase)) &&
	( XtIsManaged( _wBase) == FALSE)) {
	    
        XtManageChild( _wBase);

    }
}


/*
 * Function: AmBaseWidgetClass::hide()
 *
 * Purpose:
 *   Hide - i.e remove from view - the widget and its' children.
 *
 * Comments:
 *   This is the private (specific to AmBaseWidgetClass) function
 *   to unmanage and hide the widget.
 */

void
AmBaseWidgetClass::hide()
{
    if (( _wBase!= NULL) &&
	( XtIsRealized( _wBase)) &&
	( XtIsManaged( _wBase) == TRUE)) {
	    
	XtUnmanageChild( _wBase);
    }
}

/*
 * Function: AmBaseWidgetClass::Show()
 *
 * Purpose:
 *   Show the widget and its' children.
 *
 * Comments:
 *   This is the public interface function to show the widget.
 */

void
AmBaseWidgetClass::Show()
{
    show();
}


/*
 * Function: AmBaseWidgetClass::Hide()
 *
 * Purpose:
 *   Hide - i.e remove from view - the widget and its' children.
 *
 * Comments:
 *   This is the private interface function to hide the widget.
 */

void
AmBaseWidgetClass::Hide()
{
    hide();
}

/*
 * Function: AmBaseWidgetClass::GetName()
 *
 * Purpose:
 *   Returns the name of the widget.
 *
 */

const char *
AmBaseWidgetClass::GetName()
{
    return( _szName);
}


/*
 * Function: AmBaseWidgetClass::GetParentWidget()
 *
 * Purpose:
 *   Returns the parent widget.
 *
 */

Widget
AmBaseWidgetClass::GetParentWidget()
{
    return( _wParent);
}


/*
 * Function: AmBaseWidgetClass::GetWidget()
 *
 * Purpose:
 *   Returns the base widget.
 *
 * Comments:
 *   The base widget is the upmost widget in the widge tree for this 
 *   WidgetClass.
 */

Widget
AmBaseWidgetClass::GetWidget()
{
    return( _wBase);
}


/*
 * Function: AmBaseWidgetClass::operator Widget()
 *
 * Purpose:
 *   Provides a cast to a Widget.  The value of the 
 *   class cast to a Widget is the base widget.
 *
 */

AmBaseWidgetClass::operator Widget ()
{
    return( _wBase);
}

AmBaseWidgetClass::operator const char * ()
{
    return( _szName);
}


/*
 * Function: AmBaseWidgetClass::addDestroyCallback()
 *
 * Purpose:
 *   Internal function to add a destroy callback to the
 *   widget.
 *
 * Comments:
 *   The addDestroyCallback function checks to make sure the
 *   callback hasn't already been added and the widget actually
 *   exists.
 */

void
AmBaseWidgetClass::addDestroyCallback()
{
    if ( _wBase != NULL && _fDestroyCallback != TRUE) {
        XtAddCallback( _wBase, XmNdestroyCallback,
            ( XtCallbackProc) & AmBaseWidgetClass::AmDestroyCallback,
            ( XtPointer) this);
        _fDestroyCallback = TRUE;
    }
}


/*
 * Function: AmBaseWidgetClass::removeDestroyCallback()
 *
 * Purpose:
 *   Internal function to remove a destroy callback from the
 *   widget.
 *
 * Comments:
 *   The removeDestroyCallback function checks to make sure the
 *   callback was already added and the widget actually exists.
 */

void
AmBaseWidgetClass::removeDestroyCallback()
{
    if ( _wBase != NULL && _fDestroyCallback == TRUE) {
        XtRemoveCallback( _wBase, XmNdestroyCallback,
            ( XtCallbackProc) & AmBaseWidgetClass::AmDestroyCallback,
            ( XtPointer) this);
        _fDestroyCallback = FALSE;
    }
}

/*
 * Function: AmBaseWidgetClass::AmDestroyCallback( w, xpClientData, xpCallData)
 *
 * Purpose:
 *   Internal function to destroy the memory associated with the 
 *   widget.
 *
 * Parameters:
 *   Widget    w            : The widget that the callback was attached to.
 *   XtPointer xpClientData : The AmBaseWidgetClass pointer. 
 *   XtPointer xpCallData   : Calldata associated with a destroy callback.
 *
 * Comments:
 *   The AmDestroyCallback simply calls the destroy function.
 */

void
AmBaseWidgetClass::AmDestroyCallback( Widget    w,
                                      XtPointer xpClientData,
                                      XtPointer xpCallData)
{
    AmBaseWidgetClass *bwc = ( AmBaseWidgetClass *) xpClientData;

/* Call the actual destroy function in the class, so that we
   get access to the member variables without having to use
   the rather tedious "->". */

    bwc->destroy( w, xpCallData);
}


/*
 * Function: AmBaseWidgetClass::destroy( w, xpCallData)
 *
 * Purpose:
 *   Internal function to destroy the memory associated with the 
 *   widget.
 *
 * Parameters:
 *   Widget    w            : The widget that the callback was attached to.
 *   XtPointer xpCallData   : Calldata associated with a destroy callback.
 *
 * Comments:
 *   Removes the destroy callback so that we don't end up calling this
 *   function over and over again (since the destructor destroys the
 *   widgets which in turn would call the destroy callback... etc...).
 */

/* ARGSUSED */
void
AmBaseWidgetClass::destroy( Widget w,
                            XtPointer xpCallData)
{
    removeDestroyCallback();

    _wBase = NULL;
    delete this;
}


/*
 * Functions: AmBaseWidgetClass::AttachLeft()
 *            AmBaseWidgetClass::AttachLeft( w)
 *            AmBaseWidgetClass::AttachLeft( iPos)
 *            AmBaseWidgetClass::AttachTop()
 *            AmBaseWidgetClass::AttachTop( w)
 *            AmBaseWidgetClass::AttachTop( iPos)
 *            AmBaseWidgetClass::AttachRight()
 *            AmBaseWidgetClass::AttachRight( w)
 *            AmBaseWidgetClass::AttachRight( iPos)
 *            AmBaseWidgetClass::AttachBottom()
 *            AmBaseWidgetClass::AttachBottom( w)
 *            AmBaseWidgetClass::AttachBottom( iPos)
 *
 * Purpose:
 *   These functions are used in conjunction with FormClasses.  The version
 *   without an argument attaches the widget to the form.  The version with
 *   the argument attaches the widget to the specified widget.
 *
 * Parameters:
 *   Widget w    : The widget to attach to. 
 *   int    iPos : The position to attach to. 
 *
 */

void
AmBaseWidgetClass::AttachLeft()
{
    XtVaSetValues( _wBase,
                   XmNleftAttachment, XmATTACH_FORM,
                   NULL);
}

void
AmBaseWidgetClass::AttachLeft( Widget w)
{
    XtVaSetValues( _wBase,
                   XmNleftAttachment, XmATTACH_WIDGET,
                   XmNleftWidget, w,
                   NULL);
}

void
AmBaseWidgetClass::AttachLeft( AmBaseWidgetClass *pwc)
{
    XtVaSetValues( _wBase,
                   XmNleftAttachment, XmATTACH_WIDGET,
                   XmNleftWidget, ( Widget) *pwc,
                   NULL);
}

void
AmBaseWidgetClass::AttachLeft( int iPos)
{
    XtVaSetValues( _wBase,
                   XmNleftAttachment, XmATTACH_POSITION,
                   XmNleftPosition, iPos,
                   NULL);
}

void
AmBaseWidgetClass::AttachTop()
{
    XtVaSetValues( _wBase,
                   XmNtopAttachment, XmATTACH_FORM,
                   NULL);
}

void
AmBaseWidgetClass::AttachTop( Widget w)
{
    XtVaSetValues( _wBase,
                   XmNtopAttachment, XmATTACH_WIDGET,
                   XmNtopWidget, w,
                   NULL);
}

void
AmBaseWidgetClass::AttachTop( AmBaseWidgetClass *pwc)
{
    XtVaSetValues( _wBase,
                   XmNtopAttachment, XmATTACH_WIDGET,
                   XmNtopWidget, ( Widget) *pwc,
                   NULL);
}

void
AmBaseWidgetClass::AttachTop( int iPos)
{
    XtVaSetValues( _wBase,
                   XmNtopAttachment, XmATTACH_POSITION,
                   XmNtopPosition, iPos,
                   NULL);
}

void
AmBaseWidgetClass::AttachRight()
{
    XtVaSetValues( _wBase,
                   XmNrightAttachment, XmATTACH_FORM,
                   NULL);
}

void
AmBaseWidgetClass::AttachRight( Widget w)
{
    XtVaSetValues( _wBase,
                   XmNrightAttachment, XmATTACH_WIDGET,
                   XmNrightWidget, w,
                   NULL);
}

void
AmBaseWidgetClass::AttachRight( AmBaseWidgetClass *pwc)
{
    XtVaSetValues( _wBase,
                   XmNrightAttachment, XmATTACH_WIDGET,
                   XmNrightWidget, ( Widget) *pwc,
                   NULL);
}

void
AmBaseWidgetClass::AttachRight( int iPos)
{
    XtVaSetValues( _wBase,
                   XmNrightAttachment, XmATTACH_POSITION,
                   XmNrightPosition, iPos,
                   NULL);
}

void
AmBaseWidgetClass::AttachBottom()
{
    XtVaSetValues( _wBase,
                   XmNbottomAttachment, XmATTACH_FORM,
                   NULL);
}

void
AmBaseWidgetClass::AttachBottom( Widget w)
{
    XtVaSetValues( _wBase,
                   XmNbottomAttachment, XmATTACH_WIDGET,
                   XmNbottomWidget, w,
                   NULL);
}

void
AmBaseWidgetClass::AttachBottom( AmBaseWidgetClass *pwc)
{
    XtVaSetValues( _wBase,
                   XmNbottomAttachment, XmATTACH_WIDGET,
                   XmNbottomWidget, ( Widget) *pwc,
                   NULL);
}

void
AmBaseWidgetClass::AttachBottom( int iPos)
{
    XtVaSetValues( _wBase,
                   XmNbottomAttachment, XmATTACH_POSITION,
                   XmNbottomPosition, iPos,
                   NULL);
}


/*
 * Functions: AmBaseWidgetClass::LeftOffset( iOffset)
 *            AmBaseWidgetClass::TopOffset( iOffset)
 *            AmBaseWidgetClass::RightOffset( iOffset)
 *            AmBaseWidgetClass::BottomOffset( iOffset)
 *
 * Purpose:
 *   These functions are used in conjunction with FormClasses.  They specifiy
 *   the offset of a given attachment.
 *
 * Parameters:
 *   int    iOffset : The offset from the given attachment. 
 *
 */

void
AmBaseWidgetClass::LeftOffset( int iOffset)
{
    XtVaSetValues( _wBase,
                   XmNleftOffset, iOffset,
                   NULL);
}

void
AmBaseWidgetClass::TopOffset( int iOffset)
{
    XtVaSetValues( _wBase,
                   XmNtopOffset, iOffset,
                   NULL);
}

void
AmBaseWidgetClass::RightOffset( int iOffset)
{
    XtVaSetValues( _wBase,
                   XmNrightOffset, iOffset,
                   NULL);
}

void
AmBaseWidgetClass::BottomOffset( int iOffset)
{
    XtVaSetValues( _wBase,
                   XmNbottomOffset, iOffset,
                   NULL);
}


/*
 * Functions: AmBaseWidgetClass::GetWidth()
 *            AmBaseWidgetClass::GetHeight()
 *
 * Purpose:
 *   These functions retrieve the dimensions of the widget.
 *
 * Returns:
 *   int : The widht or height of the widget. 
 *
 */

int
AmBaseWidgetClass::GetWidth()
{
    Dimension dimWidth;

    XtVaGetValues( _wBase,
                   XmNwidth, &dimWidth,
                   NULL);

    return(( int) dimWidth);
}

int
AmBaseWidgetClass::GetHeight()
{
    Dimension dimHeight;

    XtVaGetValues( _wBase,
                   XmNheight, &dimHeight,
                   NULL);

    return(( int) dimHeight);
}

/*
 * Functions: AmBaseWidgetClass::AddCallback( szReason, xcpCallback, xpData)
 *            AmBaseWidgetClass::RemoveCallback( szReason, xcpCallback, xpData)
 *
 * Purpose:
 *   These functions add and remove callback functions from the widget.
 *
 * Parameters:
 *   const char     *szReason   : The reason for the callback function (for 
 *                                example XmNactivateCallback)
 *   XtCallbackProc xcpCallback : The callback function to call.
 *   XtPointer      xpData      : The client data to pass to the callback
 *                                function.
 * Comments:
 *   To remove a previously added callback the same parameters that were
 *   passed to AddCallback need to be bassed to RemoveCallback.
 */


void
AmBaseWidgetClass::AddCallback( const char     *szReason,
                                XtCallbackProc xcpCallback,
                                XtPointer      xpData)
{
    if ( xcpCallback != NULL) {
        XtAddCallback( _wBase, szReason, xcpCallback, xpData);
    }
}


void
AmBaseWidgetClass::RemoveCallback( const char     *szReason,
                                   XtCallbackProc xcpCallback,
                                   XtPointer      xpData)
{
    if ( xcpCallback != NULL) {
        XtRemoveCallback( _wBase, szReason, xcpCallback, xpData);
    }
}
