#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/MessageB.h>

#include <ToolKit/BoxClass.h>


/*
 * Function: initializeValues()
 *
 * Purpose:
 *   Initialize the member variables to sane values.
 */

void
AmBoxClass::initializeValues()
{
    _wShell = NULL;
    _wBox   = NULL;
}


/*
 * Function: createInterface()
 *
 * Purpose:
 *   Creates the widgets used in the AmBoxClass.
 *
 * Comments:
 *   The box class consists of a popup shell with a message 
 *   box inside.
 *
 *   The message box consists of four things.  An icon, a message
 *   next to the icon, and three buttons below the icon (Ok, Cancel,
 *   Help).
 * 
 *   By default, there is no icon and no message.
 */

void
AmBoxClass::createInterface()
{

/* Create the message box. */


    _wShell = XmCreateDialogShell( _wParent, _szName, NULL, 0);
    _wBox = XtVaCreateWidget( _szName,
                              xmMessageBoxWidgetClass, _wShell,
                              NULL);

/* Make sure the top level widget is treated as the base widget. */

    _wBase = _wShell;

/* Hook in a destroy callback. */
 
    addDestroyCallback();
}


/*
 * Function: setIcon( iType) 
 *
 * Purpose:
 *   Changes the icon of box.
 * 
 * Parameters:
 *   int iType : The type of icon to set the icon to.
 */

void
AmBoxClass::setIcon( int iType)
{
    XtVaSetValues( _wBox, XmNdialogType, iType, NULL);
}


/*
 * Function: setMessageString( sz)
 *
 * Purpose:
 *   Changes the message that is displayed in the box. 
 *
 * Parameters:
 *   const char *sz : The new message to display in the box.
 */

void
AmBoxClass::setMessageString( const char *sz)
{
    XmString xsz;

    xsz = XmStringCreateLocalized(( char *) sz);
    XtVaSetValues( _wBox, XmNmessageString, xsz, NULL);
    XmStringFree( xsz);
}


void
AmBoxClass::hideOkButton()
{
    Widget w;

    w = XmMessageBoxGetChild( _wBox, XmDIALOG_OK_BUTTON);
    if ( XtIsManaged( w)) {
        XtUnmanageChild( w);
    }
}


void
AmBoxClass::hideCancelButton()
{
    Widget w;

    w = XmMessageBoxGetChild( _wBox, XmDIALOG_CANCEL_BUTTON);
    if ( XtIsManaged( w)) {
        XtUnmanageChild( w);
    }
}


void
AmBoxClass::hideHelpButton()
{
    Widget w;

    w = XmMessageBoxGetChild( _wBox, XmDIALOG_HELP_BUTTON);
    if ( XtIsManaged( w)) {
        XtUnmanageChild( w);
    }
}


void
AmBoxClass::setOkCallback( XtCallbackProc xcp,
                           XtPointer      xp)
{
    if ( xcp != NULL) {
        XtAddCallback( _wBox, XmNokCallback, xcp, xp);
    }
}


void
AmBoxClass::setCancelCallback( XtCallbackProc xcp,
                               XtPointer      xp)
{
    if ( xcp != NULL) {
        XtAddCallback( _wBox, XmNcancelCallback, xcp, xp);
    }
}


void
AmBoxClass::setHelpCallback( XtCallbackProc xcp,
                             XtPointer      xp)
{
    if ( xcp != NULL) {
        XtAddCallback( _wBox, XmNhelpCallback, xcp, xp);
    }
}


/*
 * Function: AmBoxClass()
 *
 * Purpose:
 *   Default constructor.
 *
 * Comments:
 *   Clears out all the fields specific to AmBoxClass also
 *   calls the AmBaseWidgetClass default constructor.
 */

AmBoxClass::AmBoxClass()
    : AmBaseWidgetClass()
{
    initializeValues();
}


/*
 * Functions: AmBoxClass( wParent, szName)
 *            AmBoxClass( bwcParent, szName)
 *
 * Purpose:
 *   Constructor for the AmBoxClass.  Both functions work the same
 *   way except that one takes a widget and the other takes a "wrapped"
 *   widget.
 *
 * Parameters:
 *   Widget     wParent : Parent widget of the new widget.
 *   const char *szName : Name of the widget.
 *
 */

AmBoxClass::AmBoxClass( Widget     wParent,
                        const char *szName)
    : AmBaseWidgetClass( wParent, szName)
{
    initializeValues();
    createInterface();
}


/* See above comment. */

AmBoxClass::AmBoxClass( AmBaseWidgetClass *bwcParent,
                        const char      *szName)
    : AmBaseWidgetClass( bwcParent, szName)
{
    initializeValues();
    createInterface();
}


/*
 * Function: ~AmBoxClass()
 *
 * Purpose:
 *   Destructor for the AmBoxClass.
 *
 * Comments:
 *   Frees up all the memory used by the widget class.
 *
 */

AmBoxClass::~AmBoxClass()
{
}


/*
 * Function: Show()
 *
 * Purpose:
 *   Show the dialog's widgets.
 */

void
AmBoxClass::Show()
{

/* Make sure the message box exists before attempting to pop it up. */

    if ( _wShell != NULL) {
        XtManageChild( _wBox);
        AmForceDisplayDialog( _wBox);
    }
}


/*
 * Function: Hide()
 *
 * Purpose:
 *   Hide the dialog's widgets.
 */

void
AmBoxClass::Hide()
{

/* Make sure a shell exists before attempting to pop it down. */

    if ( _wShell != NULL) {
        XtUnmanageChild( _wBox);
        AmForceRemoveDialog( _wBox);
    }
}


/*
 * Function: DestroyOnDelete()
 *
 * Purpose:
 *   Destroy the dialog's widgets when the shell is deleted.
 */

void
AmBoxClass::DestroyOnDelete()
{
    if ( _wShell != NULL) {
        XtVaSetValues( _wShell, XmNdeleteResponse, XmDESTROY, NULL);
    }
}


/*
 * Function: HideOnDelete()
 *
 * Purpose:
 *   Hide the dialog's widgets when the shell is deleted.
 */

void
AmBoxClass::HideOnDelete()
{
    if ( _wShell != NULL) {
        XtVaSetValues( _wShell, XmNdeleteResponse, XmUNMAP, NULL);
    }
}
