#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>

#include <anvil/cad/defs.h>
#include <anvil/cad/hash.h>
#include <anvil/cad/segment.h>

#ifdef STATISTICS

long numTouch = 0;
long numInsert = 0;
long numDelete = 0;
long numHashStep = 0;

#endif

AmChTablePtr
AmMakeCHTable( ushort_t cacheSize, ushort_t hashSize)
{
    AmChTablePtr table;
    ushort_t   ii;

/* Sanity checking. */

    if ( cacheSize > hashSize) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

/* Allocate the needed memory. */

    table = ( AmChTablePtr) malloc( sizeof( AmChTableRec));
    if ( table == NULL) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    table->cache = ( AmCachePtr) malloc( sizeof( AmCacheRec) * cacheSize);
    if ( table->cache == NULL) {
        free( table);
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    table->hash = ( AmHashPtr) malloc( sizeof( AmHashRec) * hashSize);
    if ( table->hash == NULL) {
        free( table->cache);
        free( table);
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    table->hashSize = hashSize;
    table->cacheSize = cacheSize;
    table->num = 0;
    table->AmFirst = AmCACHE_INVALID;
    table->AmLast = AmCACHE_INVALID;

    for ( ii = 0; ii < cacheSize; ii++) {
        table->cache[ ii].ID = AmINVALID;
    }

    for ( ii = 0; ii < hashSize; ii++) {
        table->hash[ ii].status = AmHASH_EMPTY;
    }

    return( table);
}


AmErrorCode
AmFreeCHTable( AmChTablePtr table)
{
    if ( table == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }
    if ( table->cache != NULL) {
        free( table->cache);
    }
    if ( table->hash != NULL) {
        free( table->hash);
    }
    free( table);

    return( AmOK);
}


AmErrorCode
AmInsertCHTable( AmChTablePtr table, long ID, long *swapID)
{
    ushort_t ii, index, offset;

#ifdef STATISTICS
    numInsert++;
#endif

    if ( table->num == table->cacheSize) {

        /* Perform replacement when the cache is full. */

        ii = table->AmLast;
        table->AmLast = table->cache[ ii].prev;
        table->cache[ table->AmLast].next = AmCACHE_INVALID;

        /* Delete the hash table entry. */

        *swapID = table->cache[ ii].ID;
        table->hash[ table->cache[ ii].offset].status = AmHASH_DELETED;

    } else {

        /* Find empty slot in the cache. */

        *swapID = AmINVALID;
        for ( ii = 0; ii < table->cacheSize; ii++) {
            if ( table->cache[ ii].ID == AmINVALID) {
                break;
            }
        }

        table->num += 1;
    }

    if ( table->AmFirst != AmCACHE_INVALID) {
        table->cache[ table->AmFirst].prev = ii;
    }

    table->cache[ ii].prev = AmCACHE_INVALID;
    table->cache[ ii].next = table->AmFirst;
    table->cache[ ii].ID = ID;
    table->AmFirst = ii;

    if ( table->AmLast == AmCACHE_INVALID) {
        table->AmLast = ii;
    }

    /* Do hash table insertion. */

    index = ID % table->hashSize;
//    offset = ID % ( table->hashSize - 1);
    offset = 12;

    while ( !( table->hash[ index].status & AmHASH_EMPTY)) {

#ifdef STATISTICS
    numHashStep++;
#endif

        index += offset;
        index %= table->hashSize;
    }

    table->hash[ index].status = AmHASH_FULL;
    table->hash[ index].offset = ii;

    table->cache[ ii].offset = index;

    return( AmOK);
}


AmErrorCode
AmTouchCHTable( AmChTablePtr table, long ID)
{
    ushort_t index, offset, prev, next, temp;

#ifdef STATISTICS
    numTouch++;
#endif

    index = ID % table->hashSize;
//    offset = ID % ( table->hashSize - 1);
    offset = 12;

    temp = AmCACHE_INVALID;
    while ( table->hash[ index].status != AmHASH_EMPTY) {

#ifdef STATISTICS
    numHashStep++;
#endif

        temp = table->hash[ index].offset;
        if ( table->cache[ temp].ID == ID) {
            break;
        }
        index += offset;
        index %= table->hashSize;
    }

    if ( table->hash[ index].status == AmHASH_EMPTY) {
        RETURN_ERROR( AmID_NOT_IN_CACHE);
    }

    if ( table->AmFirst == temp) {
        return( AmOK);
    }

    prev = table->cache[ temp].prev;
    next = table->cache[ temp].next;

    if ( prev != AmCACHE_INVALID) {
        table->cache[ prev].next = next;
    }

    if ( next != AmCACHE_INVALID) {
        table->cache[ next].prev = prev;
    } else {
        if ( prev != AmCACHE_INVALID) {
            table->AmLast = prev;
        } else {
            table->AmLast = temp;
        }
    }

    table->cache[ temp].next = table->AmFirst;
    table->cache[ temp].prev = AmCACHE_INVALID;

    if ( table->AmFirst != AmCACHE_INVALID) {
        table->cache[ table->AmFirst].prev = temp;
    }
    table->AmFirst = temp; 

    return( AmOK);
}


AmErrorCode
AmDeleteCHTable( AmChTablePtr table, long ID)
{
    ushort_t index, offset, prev, next, temp;

#ifdef STATISTICS
    numDelete++;
#endif

    index = ID % table->hashSize;
//    offset = ID % ( table->hashSize - 1);
    offset = 12;

    temp = AmCACHE_INVALID;
    while ( table->hash[ index].status != AmHASH_EMPTY) {

#ifdef STATISTICS
    numHashStep++;
#endif

        temp = table->hash[ index].offset;
        if ( table->cache[ temp].ID == ID) {
            break;
        }
        index += offset;
        index %= table->hashSize;
    }

    if ( table->hash[ index].status == AmHASH_EMPTY) {
        RETURN_ERROR( AmID_NOT_IN_CACHE);
    }

    table->hash[ index].status = AmHASH_DELETED;

    table->cache[ temp].ID = AmINVALID;
    prev = table->cache[ temp].prev;
    next = table->cache[ temp].next;

    if ( prev != AmCACHE_INVALID) {
        table->cache[ prev].next = next;
    } else {
        table->AmFirst = next;
    }

    if ( next != AmCACHE_INVALID) {
        table->cache[ next].prev = prev;
    } else {
        table->AmLast = prev;
    }

    table->num -= 1;

    return( AmOK);
}

AmErrorCode
AmPrintCHTable( AmChTablePtr table)
{
    ushort_t index;

    index = table->AmFirst;
    fprintf( stderr, "Cache:\n");
    while ( index != AmCACHE_INVALID) {
        fprintf( stderr, "  Pos: %d\n    ID: %ld\n", index,
            table->cache[ index].ID);
        index = table->cache[ index].next;
    }

    return( AmOK);
}
