#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_WIN32)
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>

#include <anvil/cad/defs.h>
#include <anvil/cad/mem.h>
#include <anvil/cad/segmentProto.h>
#include <anvil/cad/objectProto.h>

AmMemoryPtr
AmMakeMemory( long numSegments,
            long numObjects,
            ushort_t cacheSize, 
            ushort_t hashSize,
            char     *filename)
{
    AmMemoryPtr mem;

    mem = ( AmMemoryPtr) malloc( sizeof( AmMemoryRec));
    if ( mem == NULL) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    mem->firstID = AmINVALID;
    mem->lastID = AmINVALID;

    mem->segTable = AmMakeSegmentTable( numSegments);
    mem->objTable = AmMakeObjectTable( numObjects);
    mem->chTable = AmMakeCHTable( cacheSize, hashSize);
    mem->filename = strdup( filename);

    return( mem);
}

AmErrorCode
AmFreeMemory( AmMemoryPtr memory)
{

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

    AmFreeCHTable( memory->chTable);
    AmFreeObjectTable( memory->objTable);
    AmFreeSegmentTable( memory->segTable);

    if ( memory->filename != NULL) {
        free( memory->filename);
    }
    free( memory);

    return( AmOK);
}

AmErrorCode
AmDumpMemory( AmMemoryPtr memory)
{
    char name[ PATH_LEN];
    int  file, temp;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    sprintf( name, "%s.mem", memory->filename);
    file = open( name, O_WRONLY | O_CREAT, 0666);
    if ( file == -1) {
        RETURN_ERROR( AmOPEN_FAILED);
    }

    temp = write( file, &memory->firstID, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    temp = write( file, &memory->lastID, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    close( file);

    if ( AmDumpSegments( memory) != AmOK) {
        return( ANVIL_ERROR_CODE);
    }

    if ( AmDumpObjectTable( memory) != AmOK) {
        return( ANVIL_ERROR_CODE);
    }

    if ( AmDumpSegmentTable( memory) != AmOK) {
        return( ANVIL_ERROR_CODE);
    }

    return( AmOK);
}


AmMemoryPtr
AmLoadMemory( char     *filename,
            long     objSize,
            ushort_t cacheSize,
            ushort_t hashSize)
{
    AmMemoryPtr mem;
    char      name[ PATH_LEN];
    long      firstID, lastID;
    int       file, temp;

    if ( filename == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

    sprintf( name, "%s.mem", filename);
    file = open( name, O_RDONLY);
    if ( file == -1) {
        ANVIL_ERROR_CODE = AmOPEN_FAILED;
        return( NULL);
    }

    temp = read( file, &firstID, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

    temp = read( file, &lastID, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

    close( file);

    mem = ( AmMemoryPtr) malloc( sizeof( AmMemoryRec));
    if ( mem == NULL) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    mem->filename = strdup( filename);
    mem->firstID = firstID;
    mem->lastID = lastID;

    mem->chTable = AmMakeCHTable( cacheSize, hashSize);
    if ( mem->chTable == NULL) {
        free( mem->filename);
        free( mem);
        return( NULL);
    }

    mem->segTable = AmLoadSegmentTable( filename);
    if ( mem->segTable == NULL) {
        free( mem->filename);
        AmFreeCHTable( mem->chTable);
        return( NULL);
    }

    mem->objTable = AmLoadObjectTable( filename, objSize);
    if ( mem->objTable == NULL) {
        free( mem->filename);
        AmFreeCHTable( mem->chTable);
        AmFreeSegmentTable( mem->segTable);
        return( NULL);
    }

    return( mem);
}
