#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined( _WIN32)
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#if !defined( _WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif

#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>

#include <anvil/cad/defs.h>
#include <anvil/cad/mem.h>
#include <anvil/cad/segmentProto.h>
#include <anvil/cad/objectProto.h>

#ifdef STATISTICS

void printStatistics();

extern long numCompacts;
extern long numAllocs;
extern long numAllocSeeks;
extern long numReallocs;
extern long numReallocSeeks;
extern long numInsert;
extern long numTouch;
extern long numDelete;
extern long numHashStep;

#endif

#ifdef __cplusplus 
extern "C" {
#endif
// these live in anvil/lib/compat/compat.c
ulong_t pread( int fildes, void *buf, size_t nbyte, long offset);
ulong_t pwrite( int fildes, const void *buf, size_t nbyte, long offset);
#ifdef __cplusplus 
};
#endif


int
AmCmpLong( void *l1, void *l2)
{
    return( ( long) l1 - ( long) l2);
}


/*
 * Function: AmErrorCode AmSetObjectInvalid( AmCadObjectRefPtr object)
 *
 * Parameters:
 *   object:    The object to set invalid.
 *
 * Description:
 *   Sets the object to indicate that it is not valid.
 *   This is needed for unused parts in the object table.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 */

AmErrorCode
AmSetObjectInvalid( AmCadObjectRefPtr object)
{

/* Sanity checking. */

    if ( object == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

/* Set the values to indicate an uninitialzed object. */

    object->segmentID = AmINVALID;
    object->offset = AmINVALID;

    return( AmOK);
}


/*
 * Function: AmObjectTablePtr AmMakeObjectTable( long size)
 *
 * Parameters:
 *   size:      The number of objects that the table can hold.
 *
 * Description:
 *   Creates a new (empty) object table.
 *
 * Returns:
 *   The new segment table if successfull.  NULL if
 *   unsuccessfull.
 *
 * Errors:
 *   AmERROR_MALLOC_FAILED
 */

AmObjectTablePtr
AmMakeObjectTable( long size)
{
    AmObjectTablePtr table;
    long           ii;

    table = ( AmObjectTablePtr) malloc( sizeof( AmObjectTableRec));
    if ( table == NULL) { 
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    table->segSize = size;
    table->AmLast = 0;
    table->free = 0;
    table->loadedSeg = 0;
    table->flag = AmOBJECT_DIRTY | AmOBJECT_INITIALIZED;

    table->objects = ( AmCadObjectRefPtr) malloc(
                                            sizeof( AmCadObjectRefRec) * size); 
    if ( table->objects == NULL) {
        free( table);
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    for ( ii = 0; ii < size; ii++) {
        AmSetObjectInvalid( &( table->objects[ ii]));
    }

    return( table);
}


/*
 * Function: AmErrorCode AmFreeObjectTable( AmObjectTablePtr table)
 *
 * Parameters:
 *   table:     The object table to be removed from memory.
 *
 * Description:
 *   Frees all the memory used by the object table.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 */

AmErrorCode
AmFreeObjectTable( AmObjectTablePtr table)
{

#ifdef SANITY_CHECKING

    if ( table == NULL) { 
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#endif

    free( table->objects);
    free( table);

    return( AmOK);
}



/*
 * Function: uchar_t *AmAllocObject( AmMemoryPtr memory,
 *                                 long      size,
 *                                 long      type,
 *                                 long      *ID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   size:      The size of the object to allocate. 
 *   type:      The type of the object to allocate.
 *   ID:        Where the object ID gets returned.
 *
 * Description:
 *   Sets aside an area in memory of the specified size.
 *   This is done by going through all the segments and 
 *   by finding a segment that has enough free space for
 *   the new object.  If their isn't a large enough
 *   continuous free block, the function tries to see
 *   if their is a segment that when compressed has
 *   enough free memory.  If there is no such segment, a
 *   new one is allocated and initialized.
 *
 * Returns:
 *   A pointer to the memory segment that was allocated, the
 *   object's ID is returned through the reference paramter.
 *   If the object wasn't allocated, NULL is returned and the
 *   reference parameter is set to AmINVALID.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_BAD_PARAM
 *   AmERROR_SEGMENT_INVALID
 *   AmERROR_SEGMENT_FULL
 */

uchar_t *
AmAllocObject( AmMemoryPtr memory,
             long      size,
             long      type,
             long      *ID)
{
    AmCadNodePtr      node;
    AmHeaderPtr       header;
    AmSegmentTablePtr segments;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj=NULL;
    AmMemorySegmentPtr      segment;
    long            tempSize, ii, jj;

#ifdef STATISTICS    
    numAllocs++;
#endif

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        *ID = AmINVALID;
        return( NULL);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        *ID = AmINVALID;
        return( NULL);
    }

    if ( memory->objTable == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        *ID = AmINVALID;
        return( NULL);
    }

    if ( size == AmINVALID) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        *ID = AmINVALID;
        return( NULL);
    }

#endif

#endif

    segments = memory->segTable;
    objTable = memory->objTable;

    tempSize = size + sizeof( AmHeaderRec);
    PARITYSIZE( tempSize);

/*
 * Find a segment with enough free space at the end (without the need
 *   to do a compression run.
 */

    node = segments->free[ type];
    if ( node == NULL) {
        node = segments->map[ type]->head;
    }

    while ( node != NULL) {

#ifdef STATISTICS    
    numAllocSeeks++;
#endif

        ii = ( long) node->data;
        segment = &segments->segments[ ii];
        if ( tempSize <= AmFreeSpace( segment)) {
            AmActivateSegment( memory, ii);
            break;
        } else if ( tempSize <= AmTotalFreeSpace( segment)) {
            AmActivateSegment( memory, ii);
            AmCompactSegment( memory, segment);
            break;
        }
        node = node->next;
    }

/* If no segment was found, create a segment for the new object. */

    if ( node == NULL) {
        ii = segments->AmLast;
        if ( ii == segments->size) {
            ANVIL_ERROR_CODE = AmSEGMENT_TABLE_FULL;
            *ID = AmINVALID;
            return( NULL);
        }

        if (( segment = AmMakeSegment( memory, ii, SEGMENT_SIZE)) == NULL) {
            *ID = AmINVALID;
            return( NULL);
        }

    /* 
     * Keep track of the segment in the list of segments associated with
     *   this particular object type.
     */

        AmInsertTail( segments->map[ type], ( void *) ii);
        node = segments->map[ type]->tail;
        segments->AmLast += 1;
    }

/*
 * Allocate the memory for the object - we know there is enough space
 *   since we made sure of that in the code above.  Also, we know that
 *   the segment data is in memory due to the AmActivateSegment, or
 *   AmMakeSegment statements above.
 */

    header = ( AmHeaderPtr) AmAllocBlock( segment, size);
    if ( header == NULL) {
        *ID = AmINVALID;
        return( NULL);
    }

/* Set the objects ID in the block and update the object table. */

    SET_HTYPE( header, type);

/* Find an empty slot in the object table. */

    for ( jj = objTable->free; jj < objTable->AmLast; jj++) {
        obj = AmGetObject( memory, jj);
        if ( obj->segmentID == AmINVALID) {
            break;
        }
    }

    if ( jj == objTable->AmLast) {
        obj = AmGetObject( memory, jj);
        objTable->AmLast++;
    }

    SET_HID( header, jj);
    objTable->free = jj + 1;

    obj->segmentID = ii;
    obj->offset = ( uchar_t *) header - segments->segments[ ii].data;

    objTable->flag |= AmOBJECT_DIRTY;

/*
 * Set the segment free pointer to point to the segment that we
 *   just inserted into.
 */

    segments->free[ type] = node;

/* Return the new object's ID through the reference parameter. */

    *ID = HID( header);
    return( ( uchar_t *) header);
}


/*
 * Function: uchar_t * AmReallocObject( AmMemoryPtr memory,
 *                                    long      ID,
 *                                    long      size)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   ID:        The object to reallocate.
 *   size:      The size of the new object.
 *
 * Description:
 *   Sets aside a new area in memory of the specified size
 *   and copies the contents of the object to the new area
 *   the amount copied depends on the smaller of the two
 *   sizes.  The memory used by the object is freed and
 *   the object pointers are updated to point to the new
 *   memory segment.
 *
 * Returns:
 *   The pointer to the new memory area, NULL on failure otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_BAD_PARAM
 *   AmERROR_SEGMENT_INVALID
 *   AmERROR_SEGMENT_FULL
 */

uchar_t *
AmReallocObject( AmMemoryPtr memory,
               long      ID,
               long      size)
{
    AmCadNodePtr      node;
    AmSegmentTablePtr segments;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj;
    AmMemorySegmentPtr      currSegment;
    AmHeaderPtr       oldHeader, header;
    uchar_t         *data, *oldData;
    long            segment=0, offset, tempSize, ii, copySize,
                    oldID, oldType, oldSize, oldSeg;

#ifdef STATISTICS    
    numReallocs++;
#endif

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

    if ( memory->objTable == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

    if ( ID == AmINVALID) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }
 
    if ( size == AmINVALID) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

#endif

#endif

    segments = memory->segTable;
    objTable = memory->objTable;

/* Bring the segment with the old object into memory and
   get the data. */

    obj = AmGetObject( memory, ID);

    ii = obj->segmentID;
    offset = obj->offset;

    AmActivateSegment( memory, ii);
    currSegment = &memory->segTable->segments[ ii];

    data = &currSegment->data[ offset];
    oldHeader = ( AmHeaderPtr) data;

    oldType = HTYPE( oldHeader);
    oldSize = HSIZE( oldHeader);
    oldID = HID( oldHeader);
    oldSeg = ii;

/* Make sure the new size of the object is word aligned. */

    tempSize = size + sizeof( AmHeaderRec);
    PARITYSIZE( tempSize);

/* If it's the same size as the old block then nothing needs to be done. */

    if ( tempSize == oldSize) {
        currSegment->status |= AmSEGMENT_DIRTY;
        return( data);
    }

    oldData = ( uchar_t *) alloca( oldSize);
    memcpy( oldData, data, oldSize);

    AmFreeBlock( currSegment, ( void *) data);

/* Find a segment with enough free space. */

    node = segments->free[ oldType];
    if ( node == NULL) {
        node = segments->map[ oldType]->head;
    }

    while ( node != NULL) {

#ifdef STATISTICS    
    numReallocSeeks++;
#endif

        ii = ( long) node->data;
        currSegment = &segments->segments[ ii];
        if ( tempSize <= AmFreeSpace( currSegment)) {
            AmActivateSegment( memory, ii);
            break;
        } else if ( tempSize <= AmTotalFreeSpace( currSegment)) {
            AmActivateSegment( memory, ii);
            AmCompactSegment( memory, currSegment);
            break;
        }
        node = node->next;
    }

/* Create a segment for the new object. */

    if ( node == NULL) {
        ii = segments->AmLast;
        if ( ii == segments->size) {
            ANVIL_ERROR_CODE = AmSEGMENT_TABLE_FULL;
            return( NULL);
        }

        if (( currSegment = AmMakeSegment( memory, ii, SEGMENT_SIZE)) == NULL) {
            ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
            return( NULL);
        }

        AmInsertTail( segments->map[ oldType], ( void *) ii);
        node = segments->map[ oldType]->tail;
        segments->AmLast += 1;
    }

/*
 * Allocate the memory for the object - we know there is enough space
 *   since we made sure of that in the code above.
 */

    header = ( AmHeaderPtr) AmAllocBlock( currSegment, size);
    if ( header == NULL) {
        ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
        return( NULL);
    }

/* Set the objects ID in the block and update the object table. */

    SET_HTYPE( header, oldType);
    SET_HID( header, oldID);

    obj->segmentID = ii;
    obj->offset = ( uchar_t *) header - currSegment->data;

    objTable->flag |= AmOBJECT_DIRTY;

    copySize = (( oldSize <= HSIZE( header)) ?
                   oldSize : HSIZE( header));
    memcpy( ( uchar_t *) header + sizeof( AmHeaderRec),
            oldData + sizeof( AmHeaderRec),
            copySize - sizeof( AmHeaderRec));


/*
 * Set the segment free pointer to point to the AmFirst segment with
 *   empty space in it.
 */

    node = AmGetNode( segments->map[ oldType], ( void *) oldSeg,
               ( AmSearchFn) AmCmpLong); 
    if ( segments->free[ oldType] == NULL) { 
        segments->free[ oldType] = node;
    } else {
        if ( segment < ( long) segments->free[ oldType]->data) {
            segments->free[ oldType] = node;
        }
    }
 
    return( ( uchar_t *) header);
}


/*
 * Function: AmErrorCode AmFreeObject( AmMemoryPtr memory,
 *                                   long      ID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   ID:        The object to be freed.
 *
 * Description:
 *   Frees up the memory used by the object referenced by the
 *   passed in ID.
 *
 * Returns:
 *   AmOK if successfull, the error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_BAD_PARAM
 */

AmErrorCode
AmFreeObject( AmMemoryPtr memory,
            long      ID)
{
    AmCadNodePtr      node;
    AmSegmentTablePtr segTable;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj;
    AmHeaderPtr       header;
    long            seg, offset, type;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING
 
    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segTable = memory->segTable;
    objTable = memory->objTable;

    obj = AmGetObject( memory, ID);
    seg = obj->segmentID;
    offset = obj->offset;

    AmActivateSegment( memory, seg);
    header = ( AmHeaderPtr) &segTable->segments[ seg].data[ offset];
    type = HTYPE( header);

    AmFreeBlock( &segTable->segments[ seg], ( void *) header);

    obj->segmentID = AmINVALID;
    obj->offset = AmINVALID;

    node = AmGetNode( segTable->map[ type], ( void * ) seg,
               ( AmSearchFn) AmCmpLong); 
    if ( segTable->free[ type] == NULL) { 
        segTable->free[ type] = node;
    } else {
        if ( seg < ( long) segTable->free[ type]->data) {
            segTable->free[ type] = node;
        }
    }

    if ( objTable->free > ID) {
        objTable->free = ID;
    }

    return( AmOK);
}


/*
 * Function: void AmPrintObject( AmMemoryPtr memory,
 *                             long      ID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   ID:        The object to be printed.
 *
 * Description:
 *   Prints out some information about the object.  That is, its ID, its
 *   type, and any other pertinent information.
 *
 * Returns:
 *   -nothing-
 */

void
AmPrintObject( AmMemoryPtr memory,
             int       ID)
{
    AmSegmentTablePtr segments;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj;
    AmHeaderPtr       header;
    AmCadPointPtr     point;
    AmCadLinePtr      line;
    AmCadRefListPtr   refList;
    uchar_t         *data;
    long            *refs;
    long            segment, offset, ii;

    segments = memory->segTable;
    objTable = memory->objTable;

    obj = AmGetObject( memory, ID);
    segment = obj->segmentID;
    offset = obj->offset;

    if ( segment == AmINVALID || offset == AmINVALID) {
        fprintf( stderr, "Object: #%d -> INVALID\n", ID);
        return;
    }

    AmActivateSegment( memory, segment);
    data = &segments->segments[ segment].data[ offset];

    header = ( AmHeaderPtr) data;

    fprintf( stderr, "Object: #%d (size = %ld)\n", ID, HSIZE( header)); 

    switch( HTYPE( header)) {
        case AmTYPE_POINT:
            point = ( AmCadPointPtr)( data + sizeof( AmHeaderRec) +
                                 sizeof( AmCadObjectRec));
            fprintf( stderr, "    Type: Point\n"); 
            fprintf( stderr, "    x: %f\n", point->x);
            fprintf( stderr, "    y: %f\n", point->y);
            break;

        case AmTYPE_LINE:
            line = ( AmCadLinePtr)( data + sizeof( AmHeaderRec) +
                               sizeof( AmCadObjectRec));
            fprintf( stderr, "    Type: Line\n"); 
            fprintf( stderr, "    P1: %ld\n", line->p1);
            fprintf( stderr, "    P2: %ld\n", line->p2);
            break;

        case AmTYPE_REFERENCE_LIST:
            refList = ( AmCadRefListPtr)( data + sizeof( AmHeaderRec));
            fprintf( stderr, "    Type: Reference\n"); 
            fprintf( stderr, "    Num: %ld\n", refList->numRefs); 
            refs = ( long *)( data + sizeof( AmHeaderRec) +
                              sizeof( AmCadRefListRec));
            for ( ii = 0; ii < refList->numRefs; ii++) {
                fprintf( stderr, "        Ref #%ld: %ld\n", ii + 1, refs[ ii]);
            }
            break;

        case AmTYPE_FREE:
            fprintf( stderr, "    Type: Free\n"); 
            break;

        default:
            fprintf( stderr, "    Type: %ld not implemented.\n", HTYPE( header));
    }
}


/*
 * Function: AmErrorCode AmAddReference( AmMemoryPtr memory,
 *                                     long      objectID,
 *                                     long      referenceID)
 *
 * Parameters:
 *   memory:      The memory record that contains the segment
 *                table, object table, and the cache.
 *   objectID:    The object that is to get a new reference.
 *   referenceID: The object that is issuing the reference.
 *
 * Description:
 *   Adds a new reference to an object.  If no reference list exists
 *   for that object, one is created.
 *
 * Returns:
 *   AmOk if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 *   AmERROR_MALLOC_FAILED
 */

AmErrorCode
AmAddReference( AmMemoryPtr memory,
              long      objectID,
              long      referenceID)
{
    AmSegmentTablePtr segments;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj;
    uchar_t         *data;
    AmCadObjectPtr    object;
    AmCadRefListPtr   refList;
    long            *refListData;
    long            segment, offset, temp, size, oldID, newID;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( objectID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( referenceID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segments = memory->segTable;
    objTable = memory->objTable;

/* Get the object. */

    obj = AmGetObject( memory, objectID);
 
    segment = obj->segmentID;
    offset = obj->offset;

    AmActivateSegment( memory, segment);
    data = &segments->segments[ segment].data[ offset];

    object = GET_OBJECT( data);

/* Get the reference list for that object. */

    oldID = object->references;

    if ( oldID != AmINVALID) {

    /* 
     * There is an existing reference list.  Get the old reference list
     *   and add the new reference to the end of it.
     */

        obj = AmGetObject( memory, oldID);
        temp = obj->segmentID;
        offset = obj->offset;

        if ( temp != segment) {
            segment = temp;
            AmActivateSegment( memory, segment);
        }

        data = &segments->segments[ segment].data[ offset];
        refList = GET_REFERENCE_LIST( data);

        size = sizeof( AmCadRefListRec) +
               ( refList->numRefs + 1) * sizeof( long);
        data = AmReallocObject( memory, oldID, size);

    /* Re-get the reference list since it might have moved in the realloc. */

        refList = GET_REFERENCE_LIST( data);

    /* Put the new reference into the list. */

        refListData = GET_REFERENCE_LIST_DATA( data);
        refListData[ refList->numRefs] = referenceID;
        refList->numRefs += 1;

    } else {

    /* There wasn't a reference list, so create a new one. */

        size = sizeof( AmCadRefListRec) + sizeof( long);
        data = AmAllocObject( memory, size, AmTYPE_REFERENCE_LIST, &newID);
        if ( data == NULL) {
            return( ANVIL_ERROR_CODE);
        }

        refList = GET_REFERENCE_LIST( data);
        refList->numRefs = 1;

    /* Put the new reference into the list. */

        refListData = GET_REFERENCE_LIST_DATA( data);
        *refListData = referenceID;

    /* Re-get the object since it might have moved in the alloc. */

        obj = AmGetObject( memory, objectID);
        segment = obj->segmentID;
        offset = obj->offset;
        AmActivateSegment( memory, segment);

        data = &segments->segments[ segment].data[ offset];
        object = GET_OBJECT( data);

    /* Insert the new reference list. */

        object->references = newID;
    }

    return( AmOK);
}


/*
 * Function: AmErrorCode AmAddAttribute( AmMemoryPtr memory,
 *                                     long      objectID,
 *                                     long      attributeID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   objectID:    The object that is to get a new attribute.
 *   attributeID: The new attribute.
 *
 * Description:
 *   Adds a new attribute to an object.
 *   If no attribute list exists for that object, one is created.
 *
 * Returns:
 *   AmOk if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 *   AmERROR_MALLOC_FAILED
 */

AmErrorCode
AmAddAttribute( AmMemoryPtr memory,
              long      objectID,
              long      attributeID)
{
    AmSegmentTablePtr  segments;
    AmObjectTablePtr   objTable;
    AmCadObjectRefPtr  obj;
    uchar_t          *data;
    AmCadObjectPtr     object;
    AmCadAttribListPtr attribList;
    long             *attribListData;
    long             segment, offset, temp, size, oldID, newID;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( objectID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( attributeID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segments = memory->segTable;
    objTable = memory->objTable;

/* Get the object. */

    obj = AmGetObject( memory, objectID);
 
    segment = obj->segmentID;
    offset = obj->offset;

    AmActivateSegment( memory, segment);
    data = &segments->segments[ segment].data[ offset];

    object = GET_OBJECT( data);

/* Get the attribute list for that object. */

    oldID = object->attributes;

    if ( oldID != AmINVALID) {

    /* 
     * There is an existing attribute list.  Get the old attribute list
     *   and add the new attribute to the end of it.
     */

        obj = AmGetObject( memory, oldID);
        temp = obj->segmentID;
        offset = obj->offset;

        if ( temp != segment) {
            segment = temp;
            AmActivateSegment( memory, segment);
        }

        data = &segments->segments[ segment].data[ offset];
        attribList = GET_ATTRIBUTE_LIST( data);

        size = sizeof( AmCadAttribListRec) +
               ( attribList->numAttribs + 1) * sizeof( long);
        data = AmReallocObject( memory, oldID, size);

    /* Re-get the attribute list since it might have moved in the realloc. */

        attribList = GET_ATTRIBUTE_LIST( data);

    /* Put the attribute into the list. */

        attribListData = GET_ATTRIBUTE_LIST_DATA( data);
        attribListData[ attribList->numAttribs] = attributeID;
        attribList->numAttribs += 1;

    } else {

    /* There wasn't an attribute list, so create a new one. */

        size = sizeof( AmCadAttribListRec) + sizeof( long);
        data = AmAllocObject( memory, size, AmTYPE_ATTRIBUTE_LIST, &newID);
        if ( data == NULL) {
            return( ANVIL_ERROR_CODE);
        }

        attribList = GET_ATTRIBUTE_LIST( data);
        attribList->numAttribs = 1;

    /* Put the attribute into the list. */

        attribListData = GET_ATTRIBUTE_LIST_DATA( data);
        *attribListData = attributeID;

    /* Re-get the object since it might have moved in the alloc. */

        obj = AmGetObject( memory, objectID);
        segment = obj->segmentID;
        offset = obj->offset;
        AmActivateSegment( memory, segment);

        data = &segments->segments[ segment].data[ offset];
        object = GET_OBJECT( data);

    /* Insert the new reference list. */

        object->attributes = newID;
    }

    return( AmOK);
}


/*
 * Function: AmErrorCode AmRemoveReference( AmMemoryPtr memory,
 *                                        long      objectID,
 *                                        long      refObjectID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   objectID:    The object from which the reference is to be removed. 
 *   refObjectID: The object that is referencing the other object. 
 *
 * Description:
 *   Removes a reference from the object accessed by objectID.  The
 *   reference that is removed is that of the object accessed by
 *   refObjectID.  If the reference list is empty after the removal,
 *   it is removed completely.
 *
 * Returns:
 *   AmOk if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 *   AmERROR_MALLOC_FAILED
 */

AmErrorCode
AmRemoveReference( AmMemoryPtr memory,
                 int       objectID,
                 int       refObjectID)
{
    AmSegmentTablePtr segments;
    AmObjectTablePtr  objTable;
    AmCadObjectRefPtr obj;
    uchar_t         *data;
    AmCadObjectPtr    object;
    AmCadRefListPtr   refList;
    long            *refListData;
    long            segment, offset, temp, numRefs, oldID, ii, jj;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( objectID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( refObjectID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segments = memory->segTable;
    objTable = memory->objTable;

/* Get the object that has the existing reference. */

    obj = AmGetObject( memory, objectID);
 
    segment = obj->segmentID; 
    offset = obj->offset; 

    AmActivateSegment( memory, segment);
    data = &segments->segments[ segment].data[ offset];

    object = GET_OBJECT( data);
    oldID = object->references;

/* Make sure there is actually a reference list. */

    if ( oldID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* Get the reference list. */

    obj = AmGetObject( memory, oldID);
    temp = obj->segmentID;
    offset = obj->offset;

    if ( temp != segment) {
        segment = temp;
        AmActivateSegment( memory, segment);
    }

    data = &segments->segments[ segment].data[ offset];

    refList = GET_REFERENCE_LIST( data);
    refListData = GET_REFERENCE_LIST_DATA( data);

    if ( refList->numRefs == 1) {

    /* The AmLast reference was removed, so remove the reference object. */

        if ( refListData[ 0] != refObjectID) {
            RETURN_ERROR( AmBAD_ARGUMENT);
        }

        object->references = AmINVALID;
        AmFreeObject( memory, oldID);
    } else {

    /* Find the reference object in the list and remove it. */

        numRefs = refList->numRefs;
        for ( ii = 0, jj = 0; ii < numRefs; ii++) {
            if ( refListData[ ii] == refObjectID) {
                continue;
            }   
            refList[ jj++] = refList[ ii];
        }

    /* If the reference object was not in the list something is hokey. */

        if ( jj >= numRefs) {
            RETURN_ERROR( AmBAD_ARGUMENT);
        }

    /* Re-utilize the freed memory. */

        refList->numRefs -= 1;
        AmReallocObject( memory, oldID,
            sizeof( AmCadRefListRec) + ( numRefs - 1) * sizeof( long));
    }

    return( AmOK);
}


AmErrorCode
AmDumpObjectTable( AmMemoryPtr memory)
{
    AmObjectTablePtr  objTable;
    char            filename[ PATH_LEN];
    int             temp, file;
    long            pos, size;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    sprintf( filename, "%s.obj", memory->filename);
    file = open( filename, O_WRONLY | O_CREAT, 0666);
    if ( file == -1) {
        RETURN_ERROR( AmOPEN_FAILED);
    }

/*
    temp = write( file, &objTable->segSize, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }
*/

    temp = write( file, &objTable->AmLast, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    if ( objTable->flag & AmOBJECT_DIRTY) {

        size = objTable->segSize * sizeof( AmCadObjectRefRec);
        pos = sizeof( long) + objTable->loadedSeg * size;

        temp = pwrite( file, objTable->objects, size, pos);
        if ( temp != size) {
            close( file);
            RETURN_ERROR( AmWRITE_FAILED);
        }

        objTable->flag &= AmOBJECT_MASK_DIRTY;
    }
    
    close( file);

    return( AmOK);
}


AmObjectTablePtr
AmLoadObjectTable( char *filename, long segSize)
{
    AmObjectTablePtr objTable;
    char           name[ PATH_LEN];
    int            file;
    long           size, temp;

#ifdef SANITY_CHECKING

    if ( filename == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#endif

    sprintf( name, "%s.obj", filename);
    file = open( name, O_RDONLY);
    if ( file == -1) {
        ANVIL_ERROR_CODE = AmOPEN_FAILED;
        return( NULL);
    }

    objTable = AmMakeObjectTable( segSize);
    if ( objTable == NULL) {
        close( file);
        return( NULL);
    }

    temp = read( file, &objTable->AmLast, sizeof( long));
    if ( temp != sizeof( long)) {
        AmFreeObjectTable( objTable);
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

/* Read in the AmFirst block of the object table. */

    size = objTable->segSize * sizeof( AmCadObjectRefRec);
    temp = read( file, objTable->objects, size);
    if ( temp != size) {
        AmFreeObjectTable( objTable);
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }
    
    close( file);
    return( objTable);
}

AmErrorCode
AmWriteObjSegment( AmMemoryPtr memory)
{
    AmObjectTablePtr objTable;
    char           filename[ PATH_LEN];
    long           pos, size;
    int            file, temp;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    sprintf( filename, "%s.obj", memory->filename);
    file = open( filename, O_WRONLY | O_CREAT, 0666);

    if ( file == -1) {
        RETURN_ERROR( AmOPEN_FAILED);
    }

    size = objTable->segSize * sizeof( AmCadObjectRefRec);
    pos = sizeof( long) + objTable->loadedSeg * size;

    temp = pwrite( file, objTable->objects, size, pos);
    if ( temp != size) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    close( file);
    return( AmOK);
}


AmErrorCode
AmCreateObjSegment( AmMemoryPtr memory, long ID)
{
    AmObjectTablePtr objTable;
    long           ii;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    if ( objTable->flag & AmOBJECT_DIRTY) {
        if ( AmWriteObjSegment( memory) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }
    }

    for ( ii = 0; ii < objTable->segSize; ii++) {
        AmSetObjectInvalid( &( objTable->objects[ ii]));
    }

    objTable->loadedSeg = ID;
    objTable->flag = AmOBJECT_DIRTY | AmOBJECT_INITIALIZED;

    return( AmOK);
}


AmErrorCode
AmReadObjSegment( AmMemoryPtr memory, long ID)
{
    AmObjectTablePtr objTable;
    char           filename[ PATH_LEN];
    long           pos, size;
    int            file, temp;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    if ( objTable->flag & AmOBJECT_DIRTY) {
        if ( AmWriteObjSegment( memory) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }
    }

    sprintf( filename, "%s.obj", memory->filename);
    file = open( filename, O_RDONLY);

    if ( file == -1) {
        RETURN_ERROR( AmOPEN_FAILED);
    }

    size = objTable->segSize * sizeof( AmCadObjectRefRec);
    pos = sizeof( long) + ID * size;

    temp = pread( file, objTable->objects, size, pos);
    if ( temp != size) {
        close( file);
        RETURN_ERROR( AmREAD_FAILED);
    }

    close( file);

    objTable->loadedSeg = ID;
    objTable->flag &= AmOBJECT_MASK_DIRTY;

    return( AmOK);
}


AmCadObjectRefPtr
AmGetObject( AmMemoryPtr memory, long ID)
{
    AmObjectTablePtr objTable;
    long           objSeg, objOff;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( ID == AmINVALID) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

#endif

#endif

    objTable = memory->objTable;

    objSeg = ID / objTable->segSize;
    objOff = ID % objTable->segSize;

    if ( objSeg != objTable->loadedSeg) {

        if ( objSeg > (( objTable->AmLast - 1) / objTable->segSize)) {
            if ( AmCreateObjSegment( memory, objSeg) != AmOK) {
                return( NULL);
            }
        } else if ( AmReadObjSegment( memory, objSeg) != AmOK) {
            return( NULL);
        }
    }

    return( &objTable->objects[ objOff]);
}


AmErrorCode
AmSetObject( AmMemoryPtr memory, long ID, long segID, long segOff)
{
    AmObjectTablePtr objTable;
    long           objSeg, objOff;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( segID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( segOff == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    objSeg = ID / objTable->segSize;
    objOff = ID % objTable->segSize;

    if ( objSeg != objTable->loadedSeg) {
        if ( AmReadObjSegment( memory, objSeg) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }
    }

    objTable->objects[ objOff].segmentID = segID;
    objTable->objects[ objOff].offset = segOff;

    objTable->flag |= AmOBJECT_DIRTY;

    return( AmOK);
}


uchar_t *
AmGetData( AmMemoryPtr memory, long ID)
{
    AmCadObjectRefPtr obj;
    long            segment, offset;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( ID == AmINVALID) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

#endif

#endif

    obj = AmGetObject( memory, ID);
    if ( obj == NULL) {
        return( NULL);
    }

    segment = obj->segmentID;
    offset = obj->offset;

    if ( segment == AmINVALID || offset == AmINVALID) {
        return( NULL);
    }

    if ( AmActivateSegment( memory, segment) != AmOK) {
        return( NULL);
    }

    return( &memory->segTable->segments[ segment].data[ offset]);
}


/*
 * Function: AmErrorCode AmUpdateObjectOffset( AmObjectTablePtr object,
 *                                           long           ID,
 *                                           long           offset)
 *
 * Parameters:
 *   objectTable: The table of objects that contains the object to
 *                be updated.
 *   ID:          The ID of the object to be updated.
 *   offset:      The new offset.
 *
 * Description:
 *   Updates the object, referenced by the object ID, with a new
 *   offset.  This is needed when the object gets moved around
 *   within a segment.
 *   
 * Returns:
 *   AmOK if successfull.  The error coded otherwise.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 */

AmErrorCode
AmUpdateObjectOffset( AmMemoryPtr memory,
                    long      ID, 
                    long      newOffset)
{
    AmObjectTablePtr objTable;
    long           objSeg, objOff;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    objSeg = ID / objTable->segSize;
    objOff = ID % objTable->segSize;

    if ( objSeg != objTable->loadedSeg) {
        if ( AmReadObjSegment( memory, objSeg) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }
    }

    objTable->objects[ objOff].offset = newOffset;

    return( AmOK);
}

/*
 * Function: AmErrorCode AmChangedObject( AmObjectTablePtr object,
 *                                      long           ID)
 *
 * Parameters:
 *   memory:    The memory record that contains the segment
 *              table, object table, and the cache.
 *   ID:        The ID of the object to be updated.
 *
 * Description:
 *   Marks the segment that the object is in as dirty.  This
 *   function should get called whenever memory within a segment
 *   gets changed.
 *   
 * Returns:
 *   AmOK if successfull.  The error coded otherwise.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 */

AmErrorCode
AmChangedObject( AmMemoryPtr memory,
               long      ID)
{
    AmObjectTablePtr  objTable;
    AmSegmentTablePtr segTable;
    AmMemorySegmentPtr      segment;
    long            objSeg, objOff, segID;

#ifdef SANITY_CHECKING

/* Sanity check. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objTable = memory->objTable;

    objSeg = ID / objTable->segSize;
    objOff = ID % objTable->segSize;

    if ( objSeg != objTable->loadedSeg) {
        if ( AmReadObjSegment( memory, objSeg) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }
    }

    segID = objTable->objects[ objOff].segmentID;

    segTable = memory->segTable;
    segment = &segTable->segments[ segID];

    segment->status |= AmSEGMENT_DIRTY;

    return( AmOK);
}
