#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_WIN32)
#include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>

#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>

#include <anvil/cad/defs.h>
#include <anvil/cad/mem.h>
#include <anvil/cad/list.h>
#include <anvil/cad/objectProto.h>

/* Global variables for statistics collection: */

#ifdef STATISTICS

long numCompacts = 0;
long numAllocs = 0;
long numAllocSeeks = 0;
long numReallocs = 0;
long numReallocSeeks = 0;

#endif

long mallocs = 0;

#ifdef __cplusplus 
extern "C" {
#endif
// these live in anvil/lib/compat/compat.c
ulong_t pread( int fildes, void *buf, size_t nbyte, long offset);
ulong_t pwrite( int fildes, const void *buf, size_t nbyte, long offset);
#ifdef __cplusplus 
};
#endif

/*
 * Function: AmErrorCode AmSetSegmentInvalid( AmMemorySegmentPtr segment)
 *
 * Parameters:
 *   segment:   The segment to set invalid.
 *
 * Description:
 *   Sets the segment to indicate that it is not a valid.
 *   This is needed for unused parts in the segment table.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 */

AmErrorCode
AmSetSegmentInvalid( AmMemorySegmentPtr segment)
{

/* Sanity checking. */

    if ( segment == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

/* Set the values to indicate an uninitialzed segment. */

    segment->status = AmSEGMENT_UNINITIALIZED;
    segment->size = AmINVALID;
    segment->AmLast = AmINVALID;
    segment->freed = AmINVALID;
    segment->data = NULL;
    segment->swap = NULL;

    return( AmOK);
}


/*
 * Function: AmErrorCode AmInitializeSegment( AmMemorySegmentPtr segment,
 *                                          long size)
 *
 * Parameters:
 *   segment:   The segment to initialize.
 *   size:      The amount of memory to associate with the segment.
 *
 * Description:
 *   Initializes the passed in segment.  A memory block is
 *   allocated for the data portion of the segment.
 *
 * Note:
 *   The data portion is rounded up if needed to make sure it ends on 
 *   even memory boundaries.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_MALLOC_FAILED
 */

AmErrorCode
AmInitializeSegment( AmMemorySegmentPtr segment, long size)
{

/* Sanity checking. */

    if ( segment == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

/* Make sure the size of the segment ends on even boundary */

    PARITYSIZE( size);

    segment->status = AmSEGMENT_INITIALIZED | AmSEGMENT_DIRTY;
    segment->size = size;
    segment->AmLast = 0;
    segment->freed = 0;

    segment->data = ( uchar_t *) calloc( sizeof( uchar_t), size);
    mallocs++;

    if ( segment->data == NULL) { 
        RETURN_ERROR( AmMALLOC_FAILED);
    }

    return( AmOK);
}


/*
 * Function: AmSegmentTablePtr AmMakeSegmentTable( long size)
 *
 * Parameters:
 *   size:      The number of segments that the table can hold.
 *
 * Description:
 *   Creates a new (empty) segment table.
 *
 * Returns:
 *   The new segment table if successfull.  NULL if
 *   unsuccessfull.
 *
 * Errors:
 *   AmERROR_MALLOC_FAILED
 */

AmSegmentTablePtr
AmMakeSegmentTable( long size)
{
    AmSegmentTablePtr table;
    long            ii;

    table = ( AmSegmentTablePtr) malloc( sizeof( AmSegmentTableRec));
    if ( table == NULL) { 
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

    table->size = size;
    table->AmLast = 0;

    table->segments = ( AmMemorySegmentPtr) malloc( sizeof( AmMemorySegmentRec) * size); 
    if ( table->segments == NULL) {
        free( table);
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    for ( ii = 0; ii < NUM_TYPES; ii++) {
        table->map[ ii] = AmMakeList();
        table->free[ ii] = NULL;
    }

    for ( ii = 0; ii < size; ii++) {
        AmSetSegmentInvalid( &( table->segments[ ii]));
    }

    return( table);
}


/*
 * Function: AmErrorCode AmFreeSegmentTable( AmSegmentTablePtr table)
 *
 * Parameters:
 *   table:     The segment table to be removed from memory.
 *
 * Description:
 *   Frees the memory used by the passed-in segment table.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 */

AmErrorCode
AmFreeSegmentTable( AmSegmentTablePtr table)
{
    AmMemorySegmentPtr segment;
    long       ii;

/* Sanity checking. */

    if ( table == NULL) { 
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

    for ( ii = 0; ii < NUM_TYPES; ii++) {
        AmFreeList( table->map[ ii]);
    }

    for ( ii = 0; ii < table->AmLast; ii++) {
        segment = &table->segments[ ii];
        if ( segment->swap != NULL) {
            free( segment->swap);
        }
    }

    free( table->segments);
    free( table);

    return( AmOK);
}


/*
 * Function: long AmFreeSpace( AmMemorySegmentPtr segment)
 *
 * Parameters:
 *   segment:   The segment to query. 
 *
 * Description:
 *   Figures out the amount of continuous free space
 *   at the end of the segment.
 *
 * Returns:
 *   The amount of free space.  AmINVALID if there was an
 *   error.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 */

long
AmFreeSpace( AmMemorySegmentPtr segment)
{

#ifdef STRICT_SANITY_CHECKING

/* Sanity checking. */

    if ( segment == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( AmINVALID);
    } 

#endif

    return( segment->size - segment->AmLast);
}


/*
 * Function: long AmTotalFreeSpace( AmMemorySegmentPtr segment)
 *
 * Parameters:
 *   segment:   The segment to query. 
 *
 * Description:
 *   Figures out the amount of free space in the segment. 
 *   This includes those blocks in the segment that have
 *   been freed previously.
 *
 * Returns:
 *   The amount of free space.  AmINVALID if there was an error.
 *
 * Errors:
 *   AmERROR_BAD_PARAM
 */

long
AmTotalFreeSpace( AmMemorySegmentPtr segment)
{

#ifdef STRICT_SANITY_CHECKING

/* Sanity checking. */

    if ( segment == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    } 

#endif

    return( segment->size - segment->AmLast + segment->freed);
}


/*
 * Function: AmErrorCode AmCompactSegment( AmMemoryPtr memory, AmMemorySegmentPtr segment)
 *
 * Parameters:
 *   segment:   The segment to allocate memory in. 
 *
 * Description:
 *   Performs garbage collection on the segment.  That is, it
 *   goes through the segment and moves the free blocks to
 *   the end - thereby allowing larger blocks of memory to
 *   be allocated.
 *
 * Returns:
 *   AmOK if successfull.  The error coded otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_SEGMENT_INVALID
 */

AmErrorCode
AmCompactSegment( AmMemoryPtr mem, AmMemorySegmentPtr segment)
{
    AmObjectTablePtr objectTable;
    AmHeaderPtr      header=NULL;
    long           index, moveTo, moveFrom, moveFromSize;

#ifdef STATISTICS    
    numCompacts++;
#endif

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( mem == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

    if ( segment == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( segment->status & AmSEGMENT_UNINITIALIZED) {
        RETURN_ERROR( AmSEGMENT_INVALID);
    }

    if ( mem->objTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    objectTable = mem->objTable;

/* Go through the blocks and move them around. */

    index = 0;

/* Find the AmFirst free block. */

    if ( segment->AmLast != 0) {
        header = ( AmHeaderPtr)( &segment->data[ index]);

        while ( ( index < segment->AmLast) &&
                ( HTYPE( header) != AmTYPE_FREE)) {
            index += HSIZE( header);
            header = ( AmHeaderPtr)( &segment->data[ index]);
        }
    }
    moveTo = index;

/*
 * Keep moving the allocated areas forward over the previously freed
 *   areas until the whole segment has been examined.
 */

    while ( index < segment->AmLast) {

    /* 
     * Find the end of the free space (there may be several freed
     *   regions in a row.
     */

        while ( ( index < segment->AmLast) &&
                ( HTYPE( header) == AmTYPE_FREE)) {
            index += HSIZE( header); 
            header = ( AmHeaderPtr)( &segment->data[ index]);
        }

    /*
     * If the end of the allocted region has been reached, nothing  
     *   has to be done.
     */

        if ( index == segment->AmLast) {  
            break;
        }
        moveFrom = index;

    /*
     * Find the end of the allocated region (again there may be more than
     *   one allocated region.  It is more efficient to move it all in one
     *   chunk).
     */

        while ( ( index < segment->AmLast) &&
                ( HTYPE( header) != AmTYPE_FREE)) {
            AmUpdateObjectOffset( mem, HID( header),
                index - moveFrom + moveTo);
            index += HSIZE( header);
            header = ( AmHeaderPtr)( &segment->data[ index]);
        }

    /* Actually move the allocated region over the freed region. */

        moveFromSize = index - moveFrom;
        memcpy( &segment->data[ moveTo], &segment->data[ moveFrom],
            moveFromSize);
        header = ( AmHeaderPtr) &segment->data[ moveTo + moveFromSize];
        SET_H( header, AmTYPE_FREE, moveFrom - moveTo);

        moveTo = moveTo + moveFromSize;
        header = ( AmHeaderPtr) &segment->data[ index];
    }

/* Update the segement information. */

    segment->freed -= ( segment->AmLast - moveTo);
    segment->AmLast = moveTo;

    return( AmOK);
}


/*
 * Function: void *AmAllocBlock( AmMemorySegmentPtr segment, long size)
 *
 * Parameters:
 *   segment:   The segment to allocate memory in. 
 *   size:      The amount of memory to allocate. 
 *
 * Description:
 *   Sets aside an area in memory of the specified size.
 *
 * Returns:
 *   A pointer to the memory segment allocated.  NULL if
 *   allocation is unsuccessfull.
 *
 *   Note: The segment data needs to be in memory before
 *         this function gets called.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_SEGMENT_INACTIVE
 *   AmERROR_BAD_PARAM
 *   AmERROR_SEGMENT_INVALID
 *   AmERROR_SEGMENT_FULL
 */

void *
AmAllocBlock( AmMemorySegmentPtr segment, long size)
{
    AmHeaderPtr  header;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( segment == NULL) {
        ANVIL_ERROR_CODE =  AmNULL_ARGUMENT;
        return( NULL);
    } 

    if ( segment->data == NULL) {
        ANVIL_ERROR_CODE =  AmBAD_ARGUMENT;
        return( NULL);
    } 

    if ( size == AmINVALID) {
        ANVIL_ERROR_CODE =  AmBAD_ARGUMENT;
        return( NULL);
    } 

#endif

/* The allocated block has the passed-in size plus the size of
   the block header indicator. */

    size += sizeof( AmHeaderRec);

/* Make sure the size of the allocated block ends on an even boundary. */

    PARITYSIZE( size);

    if ( AmFreeSpace( segment) < size) {
        ANVIL_ERROR_CODE =  AmSEGMENT_FULL;
        return( NULL);
    }

/* Save the size in the segment. */

    header = ( AmHeaderPtr)( segment->data + segment->AmLast);

    SET_H( header, AmTYPE_OBJECT, size);
    SET_HID( header, AmINVALID);
    segment->AmLast += size;
   
    segment->status |= AmSEGMENT_DIRTY;

    return( ( void *) header);
}


/*
 * Function: AmErrorCode AmFreeBlock( AmMemorySegmentPtr segment, void *mem)
 *
 * Parameters:
 *   segment:   The segment to free memory from.
 *   mem:       The memory block to be freed.
 *
 * Description:
 *   Makes a previously allocated memory block available for use
 *   again.
 *
 * Returns:
 *   AmOK if successfull.  The error code otherwise.
 *
 * Errors:
 *   AmERROR_NULL_PARAM
 *   AmERROR_SEGMENT_INVALID
 */

AmErrorCode
AmFreeBlock( AmMemorySegmentPtr segment, void *mem)
{
    long      size;
    AmHeaderPtr header;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( segment == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

    if ( mem == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    } 

#ifdef STRICT_SANITY_CHECKING

    if ( segment->status & AmSEGMENT_UNINITIALIZED) {
        RETURN_ERROR( AmSEGMENT_INVALID);
    }

#endif

#endif

    header = ( AmHeaderPtr) mem;
    size = HSIZE( header);

    if ( HTYPE( header) == AmTYPE_FREE) {
        RETURN_ERROR( AmFREE_MEMORY);
    }

    segment->status |= AmSEGMENT_DIRTY;

/* Empty the memory block. */

    SET_H( header, AmTYPE_FREE, size);
    SET_HID( header, AmINVALID);
    segment->freed += size;

    return( AmOK);
}


AmErrorCode
AmReadSegment( AmMemoryPtr memory, long ID)
{
    AmMemorySegmentPtr segment;
    char filename[ PATH_LEN];
    long pos;
    int  file, temp;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segment = &memory->segTable->segments[ ID];

    sprintf( filename, "%s.%03lu", memory->filename, ID / NUM_SEGS_PER_FILE);
    file = open( filename, O_RDONLY);
    if ( file == -1) { 
        RETURN_ERROR( AmOPEN_FAILED);
    }

    pos = ( ID % NUM_SEGS_PER_FILE) * SEGMENT_SIZE; 
    temp = pread( file, segment->data, segment->size, pos);
    if ( temp != segment->size) {
        close( file);
        RETURN_ERROR( AmREAD_FAILED);
    }

    close( file);
  
    return( AmOK);
}

AmErrorCode
AmWriteSegment( AmMemoryPtr memory, long ID)
{
    AmMemorySegmentPtr segment;
    char filename[ PATH_LEN];
    long pos;
    int  file, temp;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( ID == AmINVALID) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segment = &memory->segTable->segments[ ID];

    sprintf( filename, "%s.%03lu", memory->filename, ID / NUM_SEGS_PER_FILE);
    file = open( filename, O_WRONLY | O_CREAT, 0666);
    if ( file == -1) { 
        RETURN_ERROR( AmOPEN_FAILED);
    }

    pos = ( ID % NUM_SEGS_PER_FILE) * SEGMENT_SIZE; 
    if ( lseek( file, pos, SEEK_SET) != pos) {
        RETURN_ERROR( AmSEEK_FAILED);
    }

   
    temp = write( file, segment->data, segment->size);
    if ( temp != segment->size) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    close( file);

    free( segment->data);
    mallocs--;

    segment->data = NULL;

/* Mark the segment not dirty. */

    segment->status &= AmSEGMENT_MASK_DIRTY;
  
    return( AmOK);
}


AmErrorCode
AmSwapSegment( AmMemorySegmentPtr segment, long ID)
{
    char buffer[ TEXTLEN];
    int  file, temp;

#ifdef SANITY_CHECKING

    if ( segment == NULL) {
        DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( ID == AmINVALID) {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    if ( segment->swap != NULL) {
        free( segment->swap);
    }

    sprintf( buffer, "/tmp/CAD.swap.%ld.XXXXXX", ID);
    segment->swap = strdup( buffer);

    mktemp( segment->swap);
    file = open( segment->swap, O_CREAT | O_RDWR, PERMS);
    if ( file == -1) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

    if ( (temp = write( file, segment->data, segment->size)) !=
         segment->size) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    close( file);

    segment->data = NULL;

/* Mark the segment as swapped. */

    segment->status |= AmSEGMENT_SWAPPED;
  
    return( AmOK);
}


AmErrorCode
AmUnswapSegment( AmMemorySegmentPtr segment, long ID)
{
    int  file, temp;

#ifdef SANITY_CHECKING

    if ( segment == NULL) {
        DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }

#endif

    if ( !( segment->status & AmSEGMENT_SWAPPED)) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( segment->swap == NULL) {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    
    file = open( segment->swap, O_RDONLY);
    if ( file == -1) { 
        DO_RETURN_ERROR( AmOPEN_FAILED);
    }

    if ( (temp = read( file, segment->data, segment->size) != segment->size) != 0) {
        close( file);
        DO_RETURN_ERROR( AmREAD_FAILED);
    }

    close( file);
    unlink( segment->swap);

    free( segment->swap);
    segment->swap = NULL;
    

/* Mark the segment not swapped. */

    segment->status &= AmSEGMENT_MASK_SWAPPED;
  
    return( AmOK);
}


AmErrorCode
AmActivateSegment( AmMemoryPtr memory, long ID) 
{
    AmMemorySegmentPtr segment, sSegment;
    long       swapID;

    segment = &memory->segTable->segments[ ID];
    if ( segment->data != NULL) {
        return( AmTouchCHTable( memory->chTable, ID)); 
    } else {
        if ( AmInsertCHTable( memory->chTable, ID, &swapID) != AmOK) { 
            return( ANVIL_ERROR_CODE);
        }
        if ( swapID != AmINVALID) {
            sSegment = &memory->segTable->segments[ swapID];
            segment->data = sSegment->data;
            if ( sSegment->status & AmSEGMENT_DIRTY) {
                if ( AmSwapSegment( sSegment, swapID) != AmOK) {
                    return( ANVIL_ERROR_CODE);
                }
            } else {
                sSegment->data = NULL;
            }
        } else {
            segment->data = ( uchar_t *) malloc( sizeof( uchar_t) *
                segment->size);
            mallocs++;

            if ( segment->data == NULL) {
                RETURN_ERROR( AmMALLOC_FAILED);
            }
        }

        if ( SWAPPED( segment)) {
            return( AmUnswapSegment( segment, ID));
        } else {
            return( AmReadSegment( memory, ID));
        }
    }
}


AmMemorySegmentPtr
AmMakeSegment( AmMemoryPtr memory, long ID, long size)
{
    AmMemorySegmentPtr      segment, sSegment;
    long            swapID;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

#endif

#endif

    segment = &memory->segTable->segments[ ID];
    if ( VALID( segment)) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

    if ( AmInsertCHTable( memory->chTable, ID, &swapID) != AmOK) { 
        return( NULL);
    }

    if ( swapID == AmINVALID) {
        if ( AmInitializeSegment( segment, size) != AmOK) {
            return( NULL);
        }
    } else {
        sSegment = &memory->segTable->segments[ swapID];

        segment->status = AmSEGMENT_INITIALIZED | AmSEGMENT_DIRTY;
        segment->size = size;
        segment->AmLast = 0;
        segment->freed = 0;

    /* Re-use the memory about to be freed by the segment that
       is getting swapped. */

        segment->data = sSegment->data;
        AmSwapSegment( sSegment, swapID);
    }

    return( segment);
}

AmErrorCode
AmDiscardSegments( AmMemoryPtr memory)
{
    AmSegmentTablePtr segTable;
    char            filename[ PATH_LEN];
    long            ii;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

   if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segTable = memory->segTable;
    for ( ii = 0; ii < segTable->AmLast; ii++) {

        if ( segTable->segments[ ii].data != NULL) {
            AmDeleteCHTable( memory->chTable, ii);
            free( segTable->segments[ ii].data);
            segTable->segments[ ii].data = NULL;
            segTable->segments[ ii].status &= AmSEGMENT_MASK_DIRTY;
            mallocs--;
        } else if ( SWAPPED(( &segTable->segments[ ii]))) {
            sprintf( filename, ".swap/swap.%lu", ii);
            unlink( filename);
            segTable->segments[ ii].status &= ( AmSEGMENT_MASK_DIRTY &
                                                AmSEGMENT_MASK_SWAPPED);
        }
    }

    return( AmOK);
}


AmErrorCode
AmDumpSegments( AmMemoryPtr memory)
{
    AmSegmentTablePtr segTable;
    long            ii;

#ifdef SANITY_CHECKING

/* Sanity checking. */

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

   if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segTable = memory->segTable;
    for ( ii = 0; ii < segTable->AmLast; ii++) {
        if ( AmActivateSegment( memory, ii) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }

        if ( AmWriteSegment( memory, ii) != AmOK) {
            return( ANVIL_ERROR_CODE);
        }

        AmDeleteCHTable( memory->chTable, ii);
    }

    return( AmOK);
}

AmErrorCode
AmDumpSegmentTable( AmMemoryPtr memory)
{
    AmSegmentTablePtr segTable;
    AmCadNodePtr      node;
    char            filename[ PATH_LEN];
    int             temp, size, file, ii, endOfList;

#ifdef SANITY_CHECKING

    if ( memory == NULL) {
        RETURN_ERROR( AmNULL_ARGUMENT);
    }

#ifdef STRICT_SANITY_CHECKING

    if ( memory->segTable == NULL) {
        RETURN_ERROR( AmBAD_ARGUMENT);
    }

#endif

#endif

    segTable = memory->segTable;

    sprintf( filename, "%s.seg", memory->filename);
    file = open( filename, O_WRONLY | O_CREAT, 0666);
    if ( file == -1) {
        RETURN_ERROR( AmOPEN_FAILED);
    }

    temp = write( file, &segTable->size, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    temp = write( file, &segTable->AmLast, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }

    size = sizeof( AmMemorySegmentRec) * segTable->AmLast;
    temp = write( file, segTable->segments, size);
    if ( temp != size) {
        close( file);
        RETURN_ERROR( AmWRITE_FAILED);
    }
   
    endOfList = AmINVALID; 

    for ( ii = 0; ii < NUM_TYPES; ii++) {
        if ( segTable->map[ ii] != NULL) {
            temp = write( file, &ii, sizeof( long));
            if ( temp != sizeof( long)) {
                close( file);
                RETURN_ERROR( AmWRITE_FAILED);
            }
            node = segTable->map[ ii]->head;
            while ( node != NULL) {
                temp = write( file, &node->data, sizeof( long));
                if ( temp != sizeof( long)) {
                    close( file);
                    RETURN_ERROR( AmWRITE_FAILED);
                }
                node = node->next;
            }

            temp = write( file, &endOfList, sizeof( long));
            if ( temp != sizeof( long)) {
                close( file);
                RETURN_ERROR( AmWRITE_FAILED);
            }
        }
    }

    close( file);

    return( AmOK);
}

AmSegmentTablePtr
AmLoadSegmentTable( char *filename)
{
    AmSegmentTablePtr segTable;
    char            name[ PATH_LEN];
    int             file;
    long            size, temp, ID, type;

#ifdef SANITY_CHECKING

    if ( filename == NULL) {
        ANVIL_ERROR_CODE = AmNULL_ARGUMENT;
        return( NULL);
    }

#endif

    sprintf( name, "%s.seg", filename);
    file = open( name, O_RDONLY);
    if ( file == -1) {
        ANVIL_ERROR_CODE = AmOPEN_FAILED;
        return( NULL);
    }

    temp = read( file, &size, sizeof( long));
    if ( temp != sizeof( long)) {
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

    segTable = AmMakeSegmentTable( size);
    if ( segTable == NULL) {
        close( file);
        return( NULL);
    }

    temp = read( file, &segTable->AmLast, sizeof( long));
    if ( temp != sizeof( long)) {
        AmFreeSegmentTable( segTable);
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

    size = segTable->AmLast * sizeof( AmMemorySegmentRec);
    temp = read( file, segTable->segments, size);
    if ( temp != size) {
        AmFreeSegmentTable( segTable);
        close( file);
        ANVIL_ERROR_CODE = AmREAD_FAILED;
        return( NULL);
    }

    while (( temp = read( file, &type, sizeof( long))) == sizeof( long)) {
        while ( TRUE) {
            temp = read( file, &ID, sizeof( long));
            if ( temp != sizeof( long)) {
                AmFreeSegmentTable( segTable);
                close( file);
                ANVIL_ERROR_CODE = AmREAD_FAILED;
                return( NULL);
            }
            if ( ID == AmINVALID) {
                break;
            }
            AmInsertTail( segTable->map[ type], ( void *) ID);
        }
    }

    close( file);

    return( segTable);
}
