#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/display.h>
#include <anvil/imageP.h>

/**************************************************************************
 * 
 * Description:
 *  	project a subrect of a source matrix to a subrect of destination 
 *	using a pixel function and client data for each element of the
 *	source matrix.
 * Arguments:
 *	srcData: source matrix. Actually can be of arbitrary type.
 *	srcSize: size of each element in the source matrix.
 *	rows, cols: rows, cols of source matrix.
 *	flipRows, flipCols: controls transposition during projection.
 *	srcRect: subrectangle of source matrix to project.  Note the
 *	source rectangle is in float coordinates for subpixel projection
 *	dstRect: subrectangle in destination image.
 *	pixelFn: pixel function to apply to source matrix.
 *	clientData: passed as second argument to the pixel function.
 *	transparency: transparency value 0.0 - 1.0
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED if the differencing or 
 *	interpolation remainder vectors couldn't be allocated.
 * Recommended enhancements:
 *	Using the previous row, and precalcing some of the remainder
 *	stuff and allocaing the vectors.
 *
 *************************************************************************/


/* ARGSUSED */

AmErrorCode 
AmImageClass::bilinMatrixImage( uchar_t			*pcSrcData,
			      int			cbSrcSize,
			      int			nRows,
			      int			nCols,
			      AnvilBoolean			fFlipRows,
			      AnvilBoolean			fFlipCols,
			      AmRealRectRec const 	*pSrcRealRect,
			      AmRectangleRec const	*pDstRect,
			      AmColorFn			fnPixel,
			      void			*pPixelClientData,
			      float			dTransparency)
{
    int			nHorizPixels, nVertPixels, *hIndex, *vIndex, 
			*pHIndex, *pVIndex, ii, *remRow, *remCol,
			*pRemRow, *pRemCol;
    uchar_t		*pcSrc, *pcDst, *pcTemp;
    ulong_t		*plData, *plRow, *plEndRow, *plCol, *plEndCol,
    			pl1, pl2, pl3, pl4;
    long		lPixel;
    static ulong_t      vSrcRed[ 256], vSrcGreen[ 256], vSrcBlue[ 256],   
			vDstRed[ 256], vDstGreen[ 256], vDstBlue[ 256];
    static float        dLastTransparency = 1.0;
    AnvilBoolean		fTransparency = ( dTransparency != 1.0);
    AmRectangleRec	rctScreen;

/* shortcut if we are completely transparent */

    if ( dTransparency == 0.0) {
	return( AmOK);
    }

/* translate the destination rectangle from virtual to screen coordinates */

    GetDisplay()->ScreenRect( pDstRect, &rctScreen);

/* get the information out of the destination image */

    plData = ( ulong_t *) AmGetData();
    nHorizPixels = _cxPixels;
    nVertPixels = _cyPixels;

/* assign up front for efficiency */

    pcSrc = ( uchar_t *) &lPixel;

/* create the forward differencing vectors, then scale the horizontal vector
   by the source pixel size, and the vertical vector by the source row length
   The source rectangle is calculated with -1, because neighboring pixels
   are sampled for bilinear interpolation. */

    if (( hIndex = AmMakeRealFwdDiffVec( pSrcRealRect->x, 
				       pSrcRealRect->x + 
					   pSrcRealRect->width - 1,
				       rctScreen.width)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ii = 0; ii < rctScreen.width; ii++) {
	hIndex[ ii] *= cbSrcSize;
    }
    if (( vIndex = AmMakeRealFwdDiffVec( pSrcRealRect->y, 
				       pSrcRealRect->y + 
					   pSrcRealRect->height - 1,
				       rctScreen.height)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    pVIndex = vIndex;
    for ( ii = 0; ii < rctScreen.height; ii++) {
	vIndex[ ii] *= cbSrcSize*nCols;
    }

/* create the remainder vectors for interpolation */

    if (( remRow = AmMakeRealRemainVec( pSrcRealRect->y, 
				      pSrcRealRect->y + 
					  pSrcRealRect->height - 1,
				      rctScreen.height)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( remCol = AmMakeRealRemainVec( pSrcRealRect->x, 
				      pSrcRealRect->x + 
					pSrcRealRect->width - 1,
				      rctScreen.width)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* Create lookup tables to translate the rgb values from the original
   value into the transparent value. */

    if ( fTransparency) {
        if ( dTransparency != dLastTransparency) {
            AmGenRGBLookups( dTransparency,
                           vSrcRed, vSrcGreen, vSrcBlue,
                           vDstRed, vDstGreen, vDstBlue);
        }
        dLastTransparency = dTransparency;
    }

/* offset to the beginning of the source subimage, set the pointer to
   the start of the remainder row. */

    pcSrcData += (( int) pSrcRealRect->x + 
		  ( int) pSrcRealRect->y*nCols)*cbSrcSize;
    pRemRow = remRow;

/* loop until the end of the destination subimage, setting pointers at the
   start and end of each row.  Use the vertical and horizontal forward
   differencing vectors to increment through the source image.  
   dstDataLong destLong point to the current pixel and end of the 
   destination image. The boolean flipRows and flipCols parameters control 
   the increment and decrement of the destination loops */

    if ( fFlipRows) {

	plEndRow = plData + rctScreen.x + rctScreen.y*nHorizPixels;
	plRow = plData + rctScreen.x + 
		     ( rctScreen.y + rctScreen.height - 1)*nHorizPixels;

/* rows decrease, columns decrease */

	while ( plRow >= plEndRow) {

/* set pointers to the horizontal forward differencing vector, the source
   data and the column remainder vector. */

	    pHIndex = hIndex;
	    pcTemp = pcSrcData;
	    pRemCol = remCol;	
	    if ( fFlipCols) {
		plCol = plRow + rctScreen.width - 1;
		plEndCol = plRow;
		while ( plCol >= plEndCol) {

/* get the four neighboring pixels to interpolate. */

		    pl1 = ( fnPixel)( pcTemp, 
				      pPixelClientData);
		    pl2 = ( fnPixel)( pcTemp + cbSrcSize, 
				      pPixelClientData);
		    pl3 = ( fnPixel)( pcTemp + cbSrcSize*nCols, 
				      pPixelClientData);
		    pl4 = ( fnPixel)( pcTemp + cbSrcSize*( nCols + 1), 
				      pPixelClientData);
		    lPixel = AmInterpPixels( pl1, pl2, pl3, pl4, 
					   *pRemRow, *pRemCol);

/* take the resulting interpolated pixel and blend it with the destination
   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {
			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
				       vDstBlue[ pcDst[ BLUE]]) |
				     ( vSrcGreen[ pcSrc[ GREEN]] +
				       vDstGreen[ pcDst[ GREEN]]) |
				     ( vSrcRed[ pcSrc[ RED]] +
				       vDstRed[ pcDst[ RED]]);
			} else {
			    *plCol = lPixel;
			}
		    }

/* decrement the row pointer, increment the row interpolation remainder 
   pointer, and increment the pointer to the source data 
   (by forward differencing) */

		    plCol--;
		    pRemCol++;
		    pcTemp += *pHIndex++;
		}
	    } else {

/* Iterate through the cols in increasing order (start at left and go right) */

		plCol = plRow;
		plEndCol = plRow + rctScreen.width;
		while ( plCol < plEndCol) {

/* get the four neighboring pixels to interpolate. */

		    pl1 = ( fnPixel)( pcTemp, 
				      pPixelClientData);
		    pl2 = ( fnPixel)( pcTemp + cbSrcSize, 
				      pPixelClientData);
		    pl3 = ( fnPixel)( pcTemp + cbSrcSize*nCols, 
				      pPixelClientData);
		    pl4 = ( fnPixel)( pcTemp + cbSrcSize*( nCols + 1), 
				      pPixelClientData);
		    lPixel = AmInterpPixels( pl1, pl2, pl3, pl4, 
					   *pRemRow, *pRemCol);

/* take the resulting interpolated pixel and blend it with the destination
   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {
			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
				       vDstBlue[ pcDst[ BLUE]]) |
				     ( vSrcGreen[ pcSrc[ GREEN]] +
				       vDstGreen[ pcDst[ GREEN]]) |
				     ( vSrcRed[ pcSrc[ RED]] +
				       vDstRed[ pcDst[ RED]]);
			} else {
			    *plCol = lPixel;
			}
		    }

/* increment the row pointer, the row interpolation remainder pointer, and
   the pointer to the source data (by forward differencing) */

		    plCol++;
		    pRemCol++;
		    pcTemp += *pHIndex++;
		}
	    }

/* step to the previous row in the destination image, the next row in
   the source image ( by forward difference), and the next interpolation
   remainder */

	    plRow -= nHorizPixels;
	    pcSrcData += *pVIndex++;
	    pRemRow++;
	}
    } else {
	plRow = plData + rctScreen.x + rctScreen.y*nHorizPixels;
	plEndRow = plData + rctScreen.x + 
		     ( rctScreen.y + rctScreen.height)*nHorizPixels;

/* rows increase */

	while ( plRow < plEndRow) {
	    pHIndex = hIndex;
	    pcTemp = pcSrcData;
	    pRemCol = remCol;	

/* columns decrease.  Set the start pointer at the end of the destination
   row, and terminate at the start of the destination row */

	    if ( fFlipCols) {
		plCol = plRow + rctScreen.width - 1;
		plEndCol = plRow;
		while ( plCol >= plEndCol) {

/* get the four neighboring pixels to interpolate. */

		    pl1 = ( fnPixel)( pcTemp, 
				      pPixelClientData);
		    pl2 = ( fnPixel)( pcTemp + cbSrcSize, 
				      pPixelClientData);
		    pl3 = ( fnPixel)( pcTemp + cbSrcSize*nCols, 
				      pPixelClientData);
		    pl4 = ( fnPixel)( pcTemp + cbSrcSize*( nCols + 1), 
				      pPixelClientData);
		    lPixel = AmInterpPixels( pl1, pl2, pl3, pl4, 
					   *pRemRow, *pRemCol);

/* take the resulting interpolated pixel and blend it with the destination
   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {
			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
				       vDstBlue[ pcDst[ BLUE]]) |
				     ( vSrcGreen[ pcSrc[ GREEN]] +
				       vDstGreen[ pcDst[ GREEN]]) |
				     ( vSrcRed[ pcSrc[ RED]] +
				       vDstRed[ pcDst[ RED]]);
			} else {
			    *plCol = lPixel;
			}
		    }

/* decrement the row pointer, increment the row interpolation remainder 
   pointer, and the pointer to the source data (by forward differencing) */

		    plCol--;
		    pRemCol++;
		    pcTemp += *pHIndex++;
		}
	    } else {

/* columns increase.  Set the start pointer at the start of the destination
   rows and terminate at the end of the destatination row */

		plCol = plRow;
		plEndCol = plRow + rctScreen.width;
		while ( plCol < plEndCol) {

/* get the four neighboring pixels to interpolate. */

		    pl1 = ( fnPixel)( pcTemp, 
				      pPixelClientData);
		    pl2 = ( fnPixel)( pcTemp + cbSrcSize, 
				      pPixelClientData);
		    pl3 = ( fnPixel)( pcTemp + cbSrcSize*nCols, 
				      pPixelClientData);
		    pl4 = ( fnPixel)( pcTemp + cbSrcSize*( nCols + 1), 
				      pPixelClientData);
		    lPixel = AmInterpPixels( pl1, pl2, pl3, pl4, 
					   *pRemRow, *pRemCol);

/* take the resulting interpolated pixel and blend it with the destination
   if there's trasparency, otherwise, just set it directly */

		    if ( fTransparency) {
			pcDst = ( uchar_t *) plCol;
			*plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
				   vDstBlue[ pcDst[ BLUE]]) |
				 ( vSrcGreen[ pcSrc[ GREEN]] +
				   vDstGreen[ pcDst[ GREEN]]) |
				 ( vSrcRed[ pcSrc[ RED]] +
				   vDstRed[ pcDst[ RED]]);
		    } else {
			*plCol = lPixel;
		    }

/* increment the row pointer, the row interpolation remainder pointer, and
   the pointer to the source data (by forward differencing) */

		    plCol++;
		    pRemCol++;
		    pcTemp += *pHIndex++;
		}
	    }

/* step to the next row in the destination and source images, and the
   next element in the interpolation remainder array. */

	    plRow += nHorizPixels;
	    pcSrcData += *pVIndex++;
	    pRemRow++;
	}
    } 
    free( hIndex);
    free( vIndex);
    free( remCol);
    free( remRow);
    return( AmOK);
}
