#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/**********************************************************************
 *
 * AmDisplayClass and AmWindowClass base definitions and member functions.
 * The machine specific graphics are derived from these classes.
 * The display class contains an enumeration for the display type,
 * the display real and virtual resolution, and a list of windows
 * created on the display.  The window class contains a backpointer to
 * the display, and a list of the GC's created on that window.
 *
 **********************************************************************/

#include <stdio.h>
#include <string.h>
#if defined(_WIN32) || defined(__BSD44__)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <math.h>
#include <anvil/display.h>
#include <anvil/macros.h>

/****************************************************************************
 *
 * Description:
 *   AmDisplayClass constructor.  Sets the resolution of the display
 *   and the virtual display (used for DPI scaling in printing and
 *   the map editor).
 *
 ****************************************************************************/

AmDisplayClass::AmDisplayClass()
{
    _iDPIBase  = 72;   /* The resolution of the "virtual display". */
    _iDPI      = 72;   /* The resolution of the display. */
    _lstWindow = NULL;
}

/****************************************************************************
 *
 * Description:
 *   AmDisplayClass destructor: clean up the window list.
 *
 ****************************************************************************/

AmDisplayClass::~AmDisplayClass()
{
    AmListPtr     lstTemp;
    AmWindowClass *pWindow;

/* If there are any windows that still exist and that are attached to
   the display, make sure that their display gets set to NULL so that
   they know that it no longer exists. */

    for ( lstTemp = _lstWindow; lstTemp != NULL; ) {
        pWindow = ( AmWindowClass *) lstTemp->data;
        pWindow->RemoveDisplay( this);

        _lstWindow = lstTemp->next;
        free( lstTemp);
        lstTemp = _lstWindow;
    }

    _lstWindow = NULL;
}

/****************************************************************************
 *
 * Description:
 *   Retrieve the screen resolution in DPI.
 * Returns:
 *   int - Screen resolution.
 *
 ****************************************************************************/

int AmDisplayClass::GetDPI()
{
    return( _iDPI);
}

/****************************************************************************
 *
 * Description:
 *   Retrieve the virtual resolution in DPI.
 * Returns:
 *   int - Screen resolution.
 *
 ****************************************************************************/

int AmDisplayClass::GetDPIBase()
{
    return( _iDPIBase);
}

/****************************************************************************
 *
 * Description:
 * 	Convert a "virtual" coordinate value to a screen coordinate value.
 * Arguments:
 *	iVirtual - virtual coordinate value
 * Returns:
 *	screen value.
 * Modifications:
 *	hmm...maybe some roundoff problems here?
 *
 ****************************************************************************/

int
AmDisplayClass::ScreenCoord( int iVirtual)
{
    return( iVirtual * _iDPI / _iDPIBase);
}

/****************************************************************************
 *
 * Description:
 * 	Convert a screen coordinate value to a "virtual" coordinate value.
 * Arguments:
 *	iScreen - screen coordinate value
 * Returns:
 *	virtual coordinate value.
 * Modifications:
 *	hmm...maybe some roundoff problems here?
 *
 ****************************************************************************/

int
AmDisplayClass::VirtualCoord( int iScreen)
{
    return( iScreen * _iDPIBase / _iDPI);
}

/****************************************************************************
 *
 * Description:
 * 	Convert a virtual coordinate value to a screen coordinate value.
 *   	This function uses floating point math to avoid round-off errors.
 * Arguments:
 *	dVirtual - virtual coordinate value
 * Returns:
 *	screen coordinate value.
 *
 ****************************************************************************/

float
AmDisplayClass::ScreenFloat( float dVirtual)
{
    return( dVirtual * _iDPI / ( float) _iDPIBase);
}

/****************************************************************************
 *
 * Description:
 * 	Convert a screen coordinate value to a "virtual" coordinate value.
 *   	This function uses floating point math to avoid round-off errors.
 * Arguments:
 *	dScreen - virtual coordinate value
 * Returns:
 *	virtual coordinate value.
 *
 ****************************************************************************/

float
AmDisplayClass::VirtualFloat( float dScreen)
{
    return( dScreen * _iDPIBase / ( float) _iDPI);
}

/****************************************************************************
 *
 * Description:
 * 	Convert a "virtual" rectangle to a screen rectangle. 
 * Arguments:
 *	pVirtualRect - virtual coordinate rectangle.
 *	pScreenRect - returned screen rectangle.
 *
 ****************************************************************************/

void
AmDisplayClass::ScreenRect( AmRectangleRec const *pVirtualRect,
                          AmRectanglePtr       pScreenRect)
{
#if defined( CONVERT_BY_COORDINATE)
    int iRight, iBottom;
#endif

    pScreenRect->x      = ScreenCoord( pVirtualRect->x);
    pScreenRect->y      = ScreenCoord( pVirtualRect->y);

#if defined( CONVERT_BY_COORDINATE)

    iRight  = ScreenCoord( pVirtualRect->x + pVirtualRect->width);
    iBottom = ScreenCoord( pVirtualRect->y + pVirtualRect->height);

    pScreenRect->width  = iRight - pScreenRect->x; 
    pScreenRect->height = iBottom - pScreenRect->y;

#else

    pScreenRect->width  = ScreenCoord( pVirtualRect->width);
    pScreenRect->height = ScreenCoord( pVirtualRect->height); 

#endif

}

/****************************************************************************
 *
 * Description:
 * 	Convert a screen rectangle to a "virtual" rectangle. 
 * Arguments:
 *	pScreenRect - screen rectangle.
 *	pVirtualRect - returned virtual coordinate rectangle.
 *
 ****************************************************************************/

void
AmDisplayClass::VirtualRect( AmRectangleRec const *pScreenRect,
                           AmRectanglePtr       pVirtualRect)
{
#if defined( CONVERT_BY_COORDINATE)
    int iRight, iBottom;
#endif

    pVirtualRect->x      = VirtualCoord( pScreenRect->x);
    pVirtualRect->y      = VirtualCoord( pScreenRect->y);

#if defined( CONVERT_BY_COORDINATE)

    iRight  = VirtualCoord( pScreenRect->x + pScreenRect->width);
    iBottom = VirtualCoord( pScreenRect->y + pScreenRect->height);

    pVirtualRect->width  = iRight - pScreenRect->x; 
    pVirtualRect->height = iBottom - pScreenRect->y;

#else

    pVirtualRect->width  = VirtualCoord( pScreenRect->width);
    pVirtualRect->height = VirtualCoord( pScreenRect->height);

#endif
}

/****************************************************************************
 *
 * Description:
 * 	Convert a "virtual" rectangle to a screen rectangle.
 *   	This function uses floating point math to avoid round-off errors.
 * Arguments:
 *	pVirtualRealRect - virtual coordinate rectangle.
 *	pScreenRealRect - returned screen coordinate rectangle.
 *
 ****************************************************************************/

void
AmDisplayClass::ScreenRealRect( AmRealRectRec const *pVirtualRealRect,
                              AmRealRectPtr       pScreenRealRect)
{
#if defined( CONVERT_BY_COORDINATE)
    float dRight, dBottom;
#endif

    pScreenRealRect->x      = ScreenFloat( pVirtualRealRect->x);
    pScreenRealRect->y      = ScreenFloat( pVirtualRealRect->y);

#if defined( CONVERT_BY_COORDINATE)

    dRight  = ScreenFloat( pVirtualRealRect->x + pVirtualRealRect->width);
    dBottom = ScreenFloat( pVirtualRealRect->y + pVirtualRealRect->height);

    pScreenRealRect->width  = dRight - pScreenRealRect->x;
    pScreenRealRect->height = dBottom - pScreenRealRect->y; 

#else

    pScreenRealRect->width  = ScreenFloat( pVirtualRealRect->width);
    pScreenRealRect->height = ScreenFloat( pVirtualRealRect->height);

#endif

}

/****************************************************************************
 *
 * Description:
 *	Convert a screen rectangle to a "virtual" rectangle.
 *   	This function uses floating point math to avoid round-off errors.
 * Arguments:
 *	pScreenRealRect - screen coordinate rectangle.
 *	pVirtualRealRect - returned virtual coordinate rectangle.
 *
 ****************************************************************************/

void
AmDisplayClass::VirtualRealRect( AmRealRectRec const *pScreenRealRect,
                               AmRealRectPtr       pVirtualRealRect)
{
#if defined( CONVERT_BY_COORDINATE)
    float dRight, dBottom;
#endif

    pVirtualRealRect->x      = ScreenFloat( pScreenRealRect->x);
    pVirtualRealRect->y      = ScreenFloat( pScreenRealRect->y);

#if defined( CONVERT_BY_COORDINATE)

    dRight  = VirtualFloat( pScreenRealRect->x + pScreenRealRect->width);
    dBottom = VirtualFloat( pScreenRealRect->y + pScreenRealRect->height);

    pVirtualRealRect->width  = dRight - pVirtualRealRect->x;
    pVirtualRealRect->height = dBottom - pVirtualRealRect->y; 

#else

    pVirtualRealRect->width  = VirtualFloat( pScreenRealRect->width);
    pVirtualRealRect->height = VirtualFloat( pScreenRealRect->height);

#endif

}


/****************************************************************************
 *
 * Function: int AmDisplayClass::DimInPixels( iDimension)
 *
 * Description:
 *   Convert a screen dimension which is either width or height into
 *   number of pixels represented by the dimension.
 *
 * Parameters:
 *   int iDimension - screen dimension.
 *
 * Returns:
 *   int - the number of pixels represented by the dimension.
 *
 ****************************************************************************/

int
AmDisplayClass::DimInPixels( int iDimension)
{
    return( iDimension + 1);
}

/****************************************************************************
 *
 * Description:
 * 	Set the resolution of the display in DPI, and update any windows
 *	created on the display.
 * Arguments:
 *	iDPI - dpi to set.
 *
 ****************************************************************************/

void
AmDisplayClass::SetDPI( int iDPI)
{
    AmListPtr     lstTemp;
    AmWindowClass *pWindow;

    _iDPI = iDPI;

    /* Update the windows so that they recognize the new DPI. */

    for ( lstTemp = _lstWindow; lstTemp != NULL; lstTemp = lstTemp->next) {
        pWindow = ( AmWindowClass *) lstTemp->data; 
        pWindow->Update();
    }
}

/****************************************************************************
 *
 * Description:
 * 	Add a window to the list of windows that have been created
 *   	with this display as a parent.  The display tracks windows 
 *	that have been created on it so they can be updated when
 *	the display's attributes have been changed.
 * Arguments:
 *	pWindow - window to add to the display's window list.
 *
 ****************************************************************************/

void
AmDisplayClass::AddWindow( AmWindowClass *pWindow)
{
    _lstWindow = AmInsertNodeLst( pWindow, _lstWindow); 
}

/****************************************************************************
 *
 * Description:
 * 	Remove a window from the display's list of windows that have 
 *	been created with this display as a parent.
 * Arguments:
 *	pWindow - window to remove from the display's window list.
 *
 ****************************************************************************/

void
AmDisplayClass::RemoveWindow( AmWindowClass *pWindow)
{
    _lstWindow = AmDelPtrLst( pWindow, _lstWindow, ( AmSearchFn) AmEquals); 
}

/****************************************************************************
 *
 * Description:
 * 	Wrapper function for drawing a single line. calls the DrawLines
 *	virtual function with the line's endpoints.
 * Arguments:
 *	pWindow - window to draw on.
 *	pGC - graphics context to use for drawing.
 *	x0, x0, x1, y1 - line endpoints.
 *
 ****************************************************************************/

void 
AmDisplayClass::DrawLine( AmWindowClass *pWindow,
			AmGCClass     *pGC,
			int         x0,
			int         y0,
			int         x1,
			int         y1)
{
    AmPointRec pPoint[ 2];

    pPoint[ 0].x = x0;
    pPoint[ 0].y = y0;
    pPoint[ 1].x = x1;
    pPoint[ 1].y = y1;
    DrawLines( pWindow, pGC, pPoint, 2);
}

/****************************************************************************
 *
 * Description:
 * 	Wrapper function for drawing multiline text with the appropriate
 *  	alignments. calls the ExtentsText and DrawText virtual functions.
 * Arguments:
 *	pWindow - window to draw on.
 *	pGC - graphics context to use for drawing.
 *	x, y - text starting position (ascent below, descent above the
 *	y point).
 * 	szText - test string to draw.
 *	eAlign - enumeration describing how the text should be drawn.
 *
 ****************************************************************************/

void
AmDisplayClass::DrawMultilineText( AmWindowClass    *pWindow,
				 AmGCClass        *pGC,
				 int            x,
				 int            y,
				 char           *szText,
				 AmTextAlignEnum  eAlign)
{
    char  *szLine, *szLineTemp, *szTemp;
    int   iWidth, iAscent, iDescent,
          iWidthMax, iHeightMax, iAscentMax, iDescentMax,
          xStart=0, yStart=0, iLine;
    float dAngle = 0.0;

    /* Create a buffer on the stack for a single line of text.  The
       size of the buffer is big enough for the worst case scenario. */

    szLine = ( char *) alloca( strlen( szText) + 1);
    if ( szLine == NULL) {
        DO_ERROR( AmALLOCA_FAILED);
	return;
    }

    /* Figure out the largest line extents. */

    szTemp = szText;
    iWidthMax = 0;
    iAscentMax = 0;
    iDescentMax = 0;

    while (( szTemp != NULL) && ( *szTemp != '\0')) {

        /* Copy one line of the szText. */

	for ( szLineTemp = szLine;
	      ( *szTemp != '\0') && ( *szTemp != '\n');
	      szTemp++, szLineTemp++) {
	    *szLineTemp = *szTemp;
	}
	*szLineTemp = '\0';

        /* Skip over the AmLast character if it's a '\n'. */

	if ( *szTemp == '\n') {
	    szTemp++;
	}

	ExtentsText( pGC, szLine, &iWidth, &iAscent, &iDescent);

	if ( iWidth > iWidthMax) {
	    iWidthMax = iWidth;
	}
	if ( iAscent > iAscentMax) {
	    iAscentMax = iAscent;
	}
	if ( iDescent > iDescentMax) {
	    iDescentMax = iDescent;
	}
    }

    iHeightMax = iAscentMax + iDescentMax;

    /* Now that we know the spacing, draw the szText. */

    iLine = 0;
    szTemp = szText;

    while (( szTemp != NULL) && ( *szTemp != '\0')) {

        /* Copy one line of the szText. */

	for ( szLineTemp = szLine;
	      ( *szTemp != '\0') && ( *szTemp != '\n');
	      szTemp++, szLineTemp++) {
	    *szLineTemp = *szTemp;
	}
        *szLineTemp = '\0';

        /* Skip over the AmLast character if it's a '\n'. */

	if ( *szTemp == '\n') {
	    szTemp++;
	}

        /* Get this line's extents. */

	ExtentsText( pGC, szLine, &iWidth, &iAscent, &iDescent);

        /* Set the starting x or y position depending on the alignment
           enumeration. */

	switch ( eAlign) {

	case AmCENTER_HORIZONTAL:
	    xStart = x - iWidth / 2;
	    yStart = y + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    break;

	case AmLEFT_HORIZONTAL:
	    xStart = x;
	    yStart = y + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    break;

	case AmRIGHT_HORIZONTAL:	
	    xStart = x + iWidthMax - iWidth;
	    yStart = y + iLine * iHeightMax + ( iAscent + iDescent ) / 2 -
                     iDescent;
	    break;

	case AmCENTER_UP_VERTICAL:
	    dAngle = M_PI / 2;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    yStart = y + iWidth / 2;
	    break;

	case AmLEFT_UP_VERTICAL:
	    dAngle = M_PI / 2;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    yStart = y;
	    break;

	case AmRIGHT_UP_VERTICAL:
	    dAngle = M_PI / 2;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    yStart = y + iWidthMax - iWidth;
	    break;

	case AmCENTER_DOWN_VERTICAL:
	    dAngle = 3 * M_PI / 2;
	    yStart = y + iWidth / 2;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    break;

	case AmLEFT_DOWN_VERTICAL:
	    dAngle = 3 * M_PI / 2;
	    yStart = y;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    break;

	case AmRIGHT_DOWN_VERTICAL:
	    dAngle = 3 * M_PI / 2;
	    yStart = y + iWidthMax - iWidth;
	    xStart = x + iLine * iHeightMax + ( iAscent + iDescent) / 2 -
                     iDescent;
	    break;
	default:
	    break;

	}

        /* If the dAngle is 0, then horizontal, else draw at an angle. */

	if ( dAngle == 0.0) {
	    DrawText( pWindow, pGC, xStart, yStart, szLine);
	} else {
	    DrawRotatedText( pWindow, pGC, xStart, yStart, dAngle, szLine);
	}

	iLine++;
    }
}
		
/****************************************************************************
 *
 * Description:
 * 	wrapper for drawing szText with specified alignments 
 * Arguments:
 *	pWindow - window to draw on.
 *	pGC - graphics context to use for drawing.
 *	x, y - text starting position (ascent below, descent above the
 *	y point).
 * 	szText - test string to draw.
 *	eAlign - enumeration describing how the text should be drawn.
 *
 ****************************************************************************/

void
AmDisplayClass::DrawAlignedText( AmWindowClass   *pWindow,
                               AmGCClass       *pGC,
                               int           x, 
                               int           y,
                               char          *szText,
                               AmTextAlignEnum eAlign)
{
    int   iWidth, iAscent, iDescent;
    float dAngle = 0.0;

    ExtentsText( pGC, szText, &iWidth, &iAscent, &iDescent);

/* Set the starting x or y position depending on the alignment enumeration. */

    switch ( eAlign) {

    case AmCENTER_HORIZONTAL:
	x -= iWidth / 2;
	y += ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmLEFT_HORIZONTAL:
	y += ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmRIGHT_HORIZONTAL:
	x -= iWidth;
	y += ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmCENTER_UP_VERTICAL:
	dAngle = M_PI / 2;
	y += iWidth / 2;
	x += ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmLEFT_UP_VERTICAL:
	dAngle = M_PI / 2;
	x += ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmRIGHT_UP_VERTICAL:
	dAngle = M_PI/2;
	x += ( iAscent + iDescent) / 2 - iDescent;
	y += iWidth;
	break;

    case AmCENTER_DOWN_VERTICAL:
	dAngle = 3 * M_PI / 2;
	y -= iWidth / 2;
	x -= ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmLEFT_DOWN_VERTICAL:
	dAngle = 3*M_PI/2;
	x -= ( iAscent + iDescent) / 2 - iDescent;
	break;

    case AmRIGHT_DOWN_VERTICAL:
	y -= iWidth;
	x -= ( iAscent + iDescent) / 2 - iDescent;
	break;

    default:
	break;

    }

/* if the dAngle is 0, then horizontal, else draw at an dAngle */

    if ( dAngle == 0.0) {
	DrawText( pWindow, pGC, x, y, szText);
    } else {
	DrawRotatedText( pWindow, pGC, x, y, dAngle, szText);
    }
}

/****************************************************************************
 *
 * Description:
 * 	retrieve the enumerated type of the display: windows, x, 
 *	postscript...
 * Returns: 
 * 	the display type enumeration.
 *
 ****************************************************************************/

AmDisplayTypeEnum
AmDisplayClass::GetType()
{
    return( _eDisplayType);
}

/****************************************************************************
 *
 * Description:
 * 	wrapper member function to allocate a vector of colors, calls
 *	the virtual AllocColor member function.
 * Arguments:
 *	pCol - pointer to the vector of colors.
 *	nCol - number of colors in the vector.
 *
 ****************************************************************************/

void 
AmDisplayClass::AllocateColors( AmColorCellPtr	pCol,
			      int		nCol)
{
    int	ii;

    for ( ii = 0; ii < nCol; ii++) {
	if ( AllocColor( &pCol[ ii]) == FALSE) {
	    pCol[ ii].pixel = ( ulong_t) -1;
	}
    }
}

/****************************************************************************
 *
 * Description: 
 * 	compare a pair of color cells by their RGB values.
 * Arguments:
 *	pColSrc, pColDst - colors to compare.
 * Modifications:
 *	what is this function doing here anyway?  does anyting use it?
 *
 ****************************************************************************/

AnvilBoolean
AmCompareColor( AmColorCellPtr pColSrc,
              AmColorCellPtr pColDst)
{
    return(( pColSrc->red == pColDst->red) &&
           ( pColSrc->blue == pColDst->blue) && 
    	   ( pColSrc->green == pColDst->green));
}   

/****************************************************************************
 *
 * Description: 
 * 	Wrapper function for drawing a rotated ellipse, bounded by 
 *	x0, y0, x1, y1, with opposing axis axisB, with a resolution of 
 *	nseg.  Used for drawing fresnel zones and such (OK, it's like
 *	a really application specific function, like I don't know it)
 * Arguments:
 *	pWindow - window to draw on.
 *	pGC - graphics context to use for drawing.
 *	x0, y0, x1, y1 - endpoints of the ellipse (NOT THE FOCI)
 * 	axisB - width of the other axis
 *	nseg - number of segments to draw with.
 * Modifications:
 *	should return an error code if the alloca() fails.
 *
 ****************************************************************************/

void
AmDisplayClass::DrawRotatedEllipse( AmWindowClass *pWindow,
                                  AmGCClass     *pGC,
                                  int         x0,
                                  int         y0,
                                  int         x1,
                                  int         y1,
                                  int         axisB,
                                  int         nseg)
{
    float	cosTheta, sinTheta, axisA;
    int		npts, cx, cy, ii, x, y;
    AmPointPtr	pts;

    cx = ( x1 + x0)/2;
    cy = ( y1 + y0)/2;
    axisA = sqrt(( double) ( SQR( x1 - x0) + SQR( y1 - y0)))/2.0;
    cosTheta = ( x1 - x0)/( axisA*2.0);
    sinTheta = ( y1 - y0)/( axisA*2.0);
    pts = ( AmPointPtr) alloca( nseg*sizeof( AmPointRec));
    if ( pts == NULL) {
	DO_ERROR( AmALLOCA_FAILED);
	return;
    }
    npts = nseg/4;
    for ( ii = 0; ii < npts; ii++) {

        /* x is from 0 to axisA/2 */

	x = (int)(( axisA*ii)/( npts - 1));
	y = (int)(axisB*sqrt( 1 - SQR( x)/SQR( axisA)));

        /* do that symmetry thang... */

	pts[ ii].x = x;
	pts[ ii].y = y;
	pts[ npts*2 - ii - 1].x = x;
	pts[ npts*2 - ii - 1].y = -y;
	pts[ npts*2 + ii].x = -x;
	pts[ npts*2 + ii].y = -y;
	pts[ npts*4 - ii - 1].x = -x;
	pts[ npts*4 - ii - 1].y = y;
    }

    /* apply the rotation and translation to the ellipse */

    for ( ii = 0; ii < npts*4; ii++) {
	pts[ ii].x = (short int)(cosTheta*pts[ ii].x - sinTheta*pts[ ii].y);
	pts[ ii].y = (short int)(sinTheta*pts[ ii].x + cosTheta*pts[ ii].y);
	pts[ ii].x += cx;
	pts[ ii].y += cy;
    }

    DrawLines( pWindow, pGC, pts, npts*4);
}

/**********************************************************************
 *
 * AmWindowClass member functions.
 *
 **********************************************************************/
  
/****************************************************************************
 *
 * Description: 
 * 	Window class constructor. Initialize the display backpointer
 *	and the list of graphics contexts.
 *
 ****************************************************************************/

AmWindowClass::AmWindowClass()
{
    _pDisplay = NULL;
    _lstGC = NULL;
}

/****************************************************************************
 *
 * Description: 
 * 	Window class constructor. Initialize the display backpointer
 *	to the passed display and NULL out the list of graphics contexts.
 * Arguments:
 *	display - display the window was constructed on.
 *
 ****************************************************************************/

AmWindowClass::AmWindowClass( AmDisplayClass *pDisplay)
{
    _pDisplay = pDisplay;
    _lstGC = NULL;

    _pDisplay->AddWindow( this);
}

/****************************************************************************
 *
 * Description: 
 * 	AmWindowClass destructor. Clean the window from any GC's that were
 *	created on it, and remove the window from the display's window
 *	list.
 *
 ****************************************************************************/

AmWindowClass::~AmWindowClass()
{
    AmListPtr lstTemp;
    AmGCClass *pGC;

    /* If there are any GCs that still exist and that are attached to
       the display, make sure that their window gets set to NULL so that
       they know that it no longer exists. */

    for ( lstTemp = _lstGC; lstTemp != NULL; ) {
        pGC = ( AmGCClass *) lstTemp->data;
        pGC->RemoveWindow( this);

        _lstGC = lstTemp->next;
        free( lstTemp);
        lstTemp = _lstGC;
    }

    if ( _pDisplay != NULL) {
        _pDisplay->RemoveWindow( this);
    }
}

/****************************************************************************
 *
 * Description: 
 *	update a window.  Call the virtual Update() function for all the
 *	GC's created on this window.
 *
 ****************************************************************************/

void
AmWindowClass::Update()
{
    AmListPtr lstTemp;
	AmGCClass *pGC;

    /* Update the GCs so that they recognize the changes. */

    for ( lstTemp = _lstGC; lstTemp != NULL; lstTemp = lstTemp->next) {
        pGC = ( AmGCClass *) lstTemp->data; 
        pGC->Update();
    }
}

/****************************************************************************
 *
 * Description: 
 * 	Add a GC to the list of GCs that have been created
 *   	with this window as a parent.
 * Arguments:
 *	pGC - gc to add to the window's GC list
 *
 ****************************************************************************/

void
AmWindowClass::AddGC( AmGCClass *pGC)
{
    _lstGC = AmInsertNodeLst( pGC, _lstGC);
}

/****************************************************************************
 *
 * Description: 
 * 	Remove a GC from the list of GCs that have been created
 *   	with this window as a parent.
 * Arguments:
 *	pGC - gc to remove from the window's GC list.
 *
 ****************************************************************************/

void
AmWindowClass::RemoveGC( AmGCClass *pGC)
{
    _lstGC = AmDelPtrLst( pGC, _lstGC, ( AmSearchFn) AmEquals); 
}

/****************************************************************************
 *
 * Description: 
 * 	return the display that the window was created on
 * Returns:
 * 	the display, of course.
 *
 ****************************************************************************/

AmDisplayClass *
AmWindowClass::GetDisplay()
{
    return( _pDisplay);
}

/****************************************************************************
 *
 * Description: 
 * 	DO NOT CALL THIS FUNCTION, it gets called from the AmDisplayClass 
 *  	when it gets destroyed. 
 * Arguments:
 *	pDisplay - display the window was created on 
 *
 ****************************************************************************/

void
AmWindowClass::RemoveDisplay( AmDisplayClass *pDisplay)
{

/* This is included to make sure that it is actually this window's 
   display that is getting destroyed. */

    if ( _pDisplay == pDisplay) {
        _pDisplay = NULL;
    }
}
