#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


/* an extents display is a "fake" display which is used to track the 
   extents of a series of drawing commands. Typically they're used to
   figure out the extents of an object on the screen so it can be 
   cleared out properly */


#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

/* because windows defines bessel functions with the names y0, y1 */

#undef y0
#undef y1
#include <anvil/extents.h>
#include <anvil/macros.h>

/****************************************************************************
 *
 * Description:
 * 	AmExtentsClass destructor
 *
 ****************************************************************************/

AmExtentsClass::~AmExtentsClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	AmExtentsClass constructor: initialize the range values and set
 *	the display.  
 * Arguments:
 *	disp - display the AmExtentsClass is created with.
 *
 ****************************************************************************/

AmExtentsClass::AmExtentsClass( AmDisplayClass	*disp)
{
    sourceDisplay = disp;
    StartExtents();
}
    
/****************************************************************************
 *
 * Description:
 * 	initialize the range values for the extents display.
 *
 ****************************************************************************/

void
AmExtentsClass::StartExtents()
{
    xmin = MY_HUGEINT;
    ymin = MY_HUGEINT;
    xmax = -MY_HUGEINT;
    ymax = -MY_HUGEINT;
}

/****************************************************************************
 *
 * Description:
 * 	add the passed extents rectangle to the current extents 
 *
 ****************************************************************************/

void
AmExtentsClass::addExtents( int xminArg,
			  int yminArg,
			  int xmaxArg,
			  int ymaxArg)
{
    if ( xminArg < xmin) {
	xmin = xminArg;
    }
    if ( yminArg < ymin) {
	ymin = yminArg;
    }
    if ( xmaxArg > xmax) {
	xmax = xmaxArg;
    }
    if ( ymaxArg > ymax) {
	ymax = ymaxArg;
    }
}

/****************************************************************************
 *
 * Description:
 * 	add the passed point to the current extents 
 * Arguments:
 *	x, y - point to add to the extents.
 *
 ****************************************************************************/

void
AmExtentsClass::addExtents( int x, 
			  int y)
{
    if ( x < xmin) {
	xmin = x;
    }
    if ( y < ymin) {
	ymin = y;
    }
    if ( x > xmax) {
	xmax = x;
    }
    if ( y > ymax) {
	ymax = y;
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the extents rectangle, which bounds the extents of the 
 *	list of graphics calls.
 * Returns:
 *	extents rectangle.
 *
 ****************************************************************************/

AmRectangleRec const *
AmExtentsClass::GetExtents()
{
    static AmRectangleRec rect;

    rect.x = xmin;
    rect.y = ymin;
    rect.width = xmax - xmin;
    rect.height = ymax - ymin;
    return( &rect);
}

/****************************************************************************
 *
 * Description:
 * 	call the DrawLines function from the extents wrapper.
 * Arguments:
 *      w - window to draw on.
 *      gc - graphics context to use for drawing.
 *      pts - points to draw.
 *	numpts - length of the points vector.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawLines( AmWindowClass *w,
			   AmGCClass     *gc,
			   AmPointPtr    pts,
			   int         	 numpts)
{
    AmExtentsClass::DrawPoints( w, gc, pts, numpts);
}

/****************************************************************************
 *
 * Description:
 * 	call the XDrawSegments function from the wrapper 
 * Arguments:
 *      w - window to draw on.
 *      gc - graphics context to use for drawing.
 *      segments - segments to draw.
 *	nseg - length of the segments vector.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawSegments( AmWindowClass    *,
                              AmGCClass        *,
                              AmSegmentRec     *segments,
                              int              nseg) 
{
    int ii;

    for ( ii = 0; ii < nseg; ii++) {
	addExtents( segments[ ii].x0, segments[ ii].y0);
	addExtents( segments[ ii].x1, segments[ ii].y1);
    }
}

/****************************************************************************
 *
 * Description:
 * 	call the DrawLines function from the extents wrapper.
 * Arguments:
 *      w - window to draw on.
 *      gc - graphics context to use for drawing.
 *      pts - points to draw.
 *	numpts - length of the points vector.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawPoints( AmWindowClass	*,
			    AmGCClass      	*,
			    AmPointPtr        	pts,
			    int          	numpts)
{
    int	ii;

    for ( ii = 0; ii < numpts; ii++) {
	addExtents( pts[ ii].x, pts[ ii].y);
    }
}

/****************************************************************************
 *
 * Description:
 * 	draw an arc starting at angle1, of angle2 width 
 * Arguments:
 *      w - window to draw on.
 *      gc - graphics context to use for drawing.
 *	x,y - start of arc.
 *	width, height - width, height of arc.
 *	x, y - arc extents.
 * Modifications:
 *	over-extents for partial arcs.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawArc( AmWindowClass	*,
			AmGCClass		*,
			int           	x,
			int           	y,
			int           	width,
			int           	height,
			float         	,
			float         	)
{
    addExtents( x, y, x + width, y + height);
}

/****************************************************************************
 *
 * Description:
 * 	fill an arc starting at angle1, of angle2 width 
 * Arguments:
 *      w - window to draw on.
 *      gc - graphics context to use for drawing.
 *	x,y - start of arc.
 *	width, height - width, height of arc.
 *	x, y - arc extents.
 * Modifications:
 *	over-extents for partial arcs.
 *
 ****************************************************************************/

void
AmExtentsClass::FillArc( AmWindowClass	*,
			AmGCClass		*,
			int           	x,
			int           	y,
			int           	width,
			int           	height,
			float         	,
			float         	)
{
    addExtents( x, y, x + width, y + height);
}

/****************************************************************************
 *
 * Description:
 * 	copy a rectangle in a window 
 * Arguments:
 *	dstX, dstY, srcWidth, srcHeight - dimensions of copied rectangle.
 *
 ****************************************************************************/

void
AmExtentsClass::CopyArea( AmWindowClass *,
			AmWindowClass *,
			AmGCClass     *,
			int	     ,
			int	     ,
			int	     dstX,
			int	     dstY,
			int          srcWidth,
			int          srcHeight)
{
    addExtents( dstX, dstY, dstX + srcWidth, dstY + srcHeight);
}

/****************************************************************************
 *
 * Description:
 * 	draw an image.
 * Arguments:
 *	dstX, dstY, srcWidth, srcHeight - dimensions of copied rectangle.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawImage( AmWindowClass *,
			  AmGCClass     *,
			  int         ,
			  int         ,
			  int         dstX,
			  int         dstY,
			  int         srcWidth,
			  int         srcHeight,
			  AmImageClass  *)
{
    addExtents( dstX, dstY, dstX + srcWidth, dstY + srcHeight);
}

/****************************************************************************
 *
 * Description:
 * 	draw a glyph
 * Arguments:
 *	w - window to draw to.
 *	glyph - glyph to draw
 *	x, y - position of glyph
 *
 ****************************************************************************/

void 
AmExtentsClass::DrawGlyph( AmWindowClass    *w,
			 AmGlyphClass     *glyph,
			 int        	x,
			 int        	y)
{
    addExtents( x, y, x + glyph->GetWidth(), y + glyph->GetHeight());
}

/****************************************************************************
 *
 * Description:
 * 	draw text on the screen.  This is only for 8-bit text.  uses
 *	ExtentsText() virtual function to find the text extents.
 * Arguments:
 *	pWindow - window to draw to 
 *	pGC - gc to draw with.
 *	x, y - position of text.
 *	szText - text string.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawText( AmWindowClass *pWindow,
			 AmGCClass     *pGC,
			 int         x,
			 int         y,
			 char        *szText)
{
    int	iWidth, iAscent, iDescent;

    sourceDisplay->ExtentsText( pGC, szText, &iWidth, &iAscent, &iDescent);
    addExtents( x, y - iAscent, x + iWidth, y + iAscent + iDescent);
}

/****************************************************************************
 *
 * Description:
 * 	get the extents of a text string.  This is only for 8-bit text.  
 *	uses *	ExtentsText() virtual function to find the text extents.
 * Arguments:
 *	gc - gc to draw with.
 *	text - text string.
 *	width - returned width.
 *	ascentReturn, descentReturn - returned ascent and descent.
 *
 ****************************************************************************/

void
AmExtentsClass::ExtentsText( AmGCClass    *gc,
			   char       *text,
			   int        *width,
			   int        *ascentReturn,
			   int        *descentReturn)
{
    sourceDisplay->ExtentsText( gc, text, width, 
				ascentReturn, descentReturn);
}

/****************************************************************************
 *
 * Description:
 * 	return the width of the text string in the source display 
 * Arguments:
 *	gc - graphics context (for font and font size information).
 *	text - text string to extents.
 *	length - length of the text string
 * Returns:
 *	the text width in pixels
 * Modifications:
 *	the length isn't really neccessary, since the string is 
 *	NULL-terminated.
 *
 ****************************************************************************/

int
AmExtentsClass::TextWidth( AmGCClass    *gc,
			 char       *text,
			 int        length)
{
    return( sourceDisplay->TextWidth( gc, text, length));
}

/****************************************************************************
 *
 * Description:
 * 	return the height of the font in the source display 
 * Arguments:
 *	gc - graphics context (for font and font size information).
 *	ascentReturn, descentReturn - returned ascent and descent.
 * Returns:
 *	the total font height in pixels.
 *
 ****************************************************************************/

int
AmExtentsClass::FontHeight( AmGCClass             *gc,
			  int                 *ascentReturn,
			  int                 *descentReturn)
{
    return( sourceDisplay->FontHeight( gc, ascentReturn, descentReturn));
}

/****************************************************************************
 *
 * Description:
 * 	allocate a color cell in the source display, not really used for
 *	extents, but for linking with pure virtual functions. 
 * Arguments:
 *	pCol - color to allocate.
 * Returns:
 *	TRUE if color was allocated successfully, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmExtentsClass::AllocColor( AmColorCellPtr pCol) 
{
    return( sourceDisplay->AllocColor( pCol));
}

/****************************************************************************
 *
 * Description:
 * 	allocate a named color cell in the source display, not really used 
 *	for extents, but for linking with pure virtual functions. 
 * Arguments:
 *	szName - name of the color to allocate.
 *	pCol - allocated color.
 * Returns:
 *	TRUE if color was allocated successfully, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmExtentsClass::AllocNamedColor( const char 	*szName,
			       AmColorCellPtr 	pCol) 
{
    return( sourceDisplay->AllocNamedColor( szName, pCol));
}

/****************************************************************************
 *
 * Description:
 * 	draw text on the screen.  This is only for 8-bit text 
 * Arguments:
 *	szName - name of the color to allocate.
 *	pCol - allocated color.
 * Returns:
 *	TRUE if color was allocated successfully, FALSE otherwise.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawImageText( AmWindowClass *pWindow,
			      AmGCClass     *pGC,
			      int         x,
			      int         y,
			      char        *szText)
{
    AmExtentsClass::DrawText( pWindow, pGC, x, y, szText);
}

/****************************************************************************
 *
 * Description:
 * 	find the extents for rotated text.
 * Arguments:
 *	pWindow - window to draw to.
 *	pGC - gc to draw with.
 *	x, y - text starting position.
 *	dAngle - angle to draw at (in radians)
 *	szText - text string to draw.
 *
 ****************************************************************************/

void
AmExtentsClass::DrawRotatedText( AmWindowClass *pWindow,
				AmGCClass     *pGC,
				int         x,
				int         y,
				float       dAngle,
				char        *szText)
{ 
    int	iWidth, iAscent, iDescent, iSize;

    sourceDisplay->ExtentsText( pGC, szText, &iWidth, &iAscent, &iDescent);
    iSize = iWidth + iAscent + iDescent;
    addExtents( x, y - iAscent,
		( int)( x + cos( dAngle) * iSize), 
		( int)( y + sin( dAngle) * iSize));
}

/****************************************************************************
 *
 * Description:
 * 	find the extents for a filled polygon.
 * Arguments:
 *      pWindow - window to draw on.
 *      pGC - graphics context to use for drawing.
 *      pPoint - points to draw.
 *	nPoint - length of the points vector.
 *
 ****************************************************************************/

void
AmExtentsClass::FillPolygon( AmWindowClass *pWindow,
                             AmGCClass     *pGC,
                             AmPointPtr    pPoint,
                             int           nPoint)
{
    AmExtentsClass::DrawPoints( pWindow, pGC, pPoint, nPoint);
}

/****************************************************************************
 *
 * Description:
 * 	find the extents of a rectangle.
 * Arguments:
 *	x, y, width, height - dimensions of rectangle.
 *
 ****************************************************************************/

void AmExtentsClass::DrawRectangle( AmWindowClass *,
                                    AmGCClass     *,
                                    int           x,
                                    int           y,
                                    int           iWidth,
                                    int           iHeight)
{
    addExtents( x, y, x + iWidth, y + iHeight);
}

/****************************************************************************
 *
 * Description:
 * 	find the extents of a rectangle.
 * Arguments:
 *	x, y, width, height - dimensions of rectangle.
 *
 ****************************************************************************/

void
AmExtentsClass::FillRectangle( AmWindowClass *pWindow,
                              AmGCClass     *pGC,
                              int         x,
                              int         y,
                              int         iWidth,
                              int         iHeight)
{
    AmExtentsClass::DrawRectangle( pWindow, pGC, x, y, iWidth, iHeight);
}

/****************************************************************************
 *
 * Description:
 * 	wrapper function to create a buffer. hopefully, this shouldn't
 *	get called, because it allocates a buffer in the source display.
 * Arguments:
 *	cx, cy - dimensions of the buffer.
 * Returns:
 *	the created buffer as a AmWindowClass
 *
 ****************************************************************************/

AmWindowClass *
AmExtentsClass::CreateBuffer( AmWindowClass *pWindow,
                            int         cx,
                            int         cy)
{
    return( sourceDisplay->CreateBuffer( pWindow, cx, cy));
}

/****************************************************************************
 *
 * Description:
 * 	wrapper function to create a window. hopefully, this shouldn't
 *	get called, because it allocates a window in the source display.
 * Arguments:
 *	x, y, cx, cy - position and dimensions of the window.
 * Returns:
 *	the created window as a AmWindowClass
 *
 ****************************************************************************/

AmWindowClass *
AmExtentsClass::CreateWin( AmWindowClass *pWindow,
                         int         x,
                         int         y,
                         int         cx,
                         int         cy)
{
     return( sourceDisplay->CreateWin( pWindow, x, y, cx, cy)); 
}


/****************************************************************************
 *
 * Description:
 * 	create an image in the source display. hopefully, this shouldn't
 *	get called, because it allocates an image in the source display.
 * Arguments:
 *	iWidth, iHeight - dimensions of the image.
 *	iDepth - depth of the image.
 *	pData - data for the image (may be NULL)
 *	fSharedMemoryFlag - X-specific flag for shared memory allocation.
 * Returns:
 *	the created image as an AmImageClass.
 *
 ****************************************************************************/

AmImageClass *
AmExtentsClass::CreateImage( int     iWidth,
			   int     iHeight,
			   int     iDepth,
			   char    *pData)
{
    return( sourceDisplay->CreateImage( iWidth, iHeight, iDepth, pData));
}

/****************************************************************************
 *
 * Description:
 * 	create a glyph in the source display 
 * Arguments:
 *	pWindow - window glyph is created on.
 *	szFile - file to read the glyph from.
 *	lPixel - glyph foreground pixel.
 * Returns:
 *	The created glyph.
 *
 ****************************************************************************/

AmGlyphClass *
AmExtentsClass::CreateGlyph( AmWindowClass 	*pWindow,
			   char    	*szFile,
			   ulong_t 	lPixel) 
{
    return( sourceDisplay->CreateGlyph( pWindow, szFile, lPixel));
}


/****************************************************************************
 *
 * Description:
 * 	create a region in the source display 
 * Returns:
 *	The created region.
 *
 ****************************************************************************/

AmRegionClass *
AmExtentsClass::CreateRegion()
{
    return( sourceDisplay->CreateRegion());
}

/****************************************************************************
 *
 * Description:
 * 	get an image in the source display 
 * Arguments:
 *	pWindow - window image is on.
 *	pImage - image to retrieve.
 *	x,y - postion in the window to get the image from.
 *
 ****************************************************************************/

void 
AmExtentsClass::GetImage( AmWindowClass	*pWindow,
			AmImageClass	*pImage,
			int		x,
			int		y)
{
    sourceDisplay->GetImage( pWindow, pImage, x, y);
}
