#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <string.h>
#include <anvil/display.h>
#include <anvil/rect.h>

/****************************************************************************
 *
 * Description:
 * 	Do nothing default constructor. 
 *
 ****************************************************************************/

AmImageClass::AmImageClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	Do nothing default destructor. 
 *
 ****************************************************************************/

AmImageClass::~AmImageClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	Return the width of the image in virtual coordinates. 
 *
 ****************************************************************************/

int 
AmImageClass::GetWidth()
{
    return( _cxVirtual);
}

/****************************************************************************
 *
 * Description:
 * 	Return the height of the image in virtual coordinates. 
 *
 ****************************************************************************/

int 
AmImageClass::GetHeight()
{
    return( _cyVirtual);
}

/****************************************************************************
 *
 * Description:
 * 	Return the width of the image in screen coordinates. 
 *
 ****************************************************************************/

int 
AmImageClass::GetScreenWidth()
{
    return( _cxScreen);
}

/****************************************************************************
 *
 * Description:
 * 	Return the height of the image in screen coordinates.
 *
 ****************************************************************************/

int 
AmImageClass::GetScreenHeight()
{
    return( _cyScreen);
}

/****************************************************************************
 *
 * Description:
 * 	get the number of horizontal pixels in the image 
 *
 ****************************************************************************/

int
AmImageClass::GetNumHorizPixels()
{
    return( _cxPixels);
}

/****************************************************************************
 *
 * Description:
 * 	get the number of vertical pixels in the image 
 *
 ****************************************************************************/

int
AmImageClass::GetNumVerticalPixels()
{
    return( _cyPixels);
}

/****************************************************************************
 *
 * Description:
 * 	return the DPI that the image was created with 
 *
 ****************************************************************************/

int 
AmImageClass::GetDPI()
{
    return( _iDPI);
}

/****************************************************************************
 *
 * Description:
 * 	Basic image class constructor.  Set the virtual coordinates,
 *	screen coordinates, and pixels for the image.
 * Arguments:
 *  	pDisplay - display the image is created on.
 * 	cxVirtual, cyVirtual - size of the image in virtual coordinates.
 *
 ****************************************************************************/

AmImageClass::AmImageClass( AmDisplayClass *pDisplay,
			int          cxVirtual,
			int          cyVirtual)
{
    _pDisplay  = pDisplay;

    _cxVirtual = cxVirtual;
    _cyVirtual = cyVirtual;

    _cxScreen  = pDisplay->ScreenCoord( _cxVirtual);
    _cyScreen  = pDisplay->ScreenCoord( _cyVirtual);

    _cxPixels  = _pDisplay->DimInPixels( _cxScreen);
    _cyPixels  = _pDisplay->DimInPixels( _cyScreen);

    _iDPI      = _pDisplay->GetDPI();

    _iDepth           = 0;
    _iBitsPerPixel    = 0;
    _iBytesPerPixel   = 0;
    _iBytesPerLine    = 0;
    _iSize            = 0;

    _iDDDepth         = 0;
    _iDDBitsPerPixel  = 0;
    _iDDBytesPerPixel = 0;
    _iDDBytesPerLine  = 0;
    _iDDSize          = 0;
}

/****************************************************************************
 *
 * Description:
 * 	Return the depth of the image in bits.
 *
 ****************************************************************************/

int
AmImageClass::GetDepth()
{
    return( _iDepth);
}

/****************************************************************************
 *
 * Description:
 * 	Return the number of bits per pixel (usually the same as depth)
 *
 ****************************************************************************/

int
AmImageClass::GetBitsPerPixel()
{
    return( _iBitsPerPixel);
}

/****************************************************************************
 *
 * Description:
 * 	Return the number of bytes per line in the image, which might 
 *	include padding.
 *
 ****************************************************************************/

int
AmImageClass::GetBytesPerLine()
{
    return( _iBytesPerLine);
}

/****************************************************************************
 *
 * Description:
 * 	Returns the dirty "I need to be refilled" flag. 
 *
 ****************************************************************************/

AnvilBoolean
AmImageClass::GetDirty()
{
    return( _fDirty);
}

/****************************************************************************
 *
 * Description:
 * 	Sets the dirty "I need to be refilled" flag. 
 *
 ****************************************************************************/

void
AmImageClass::SetDirty( AnvilBoolean fDirty)
{
    _fDirty = fDirty;
}

/****************************************************************************
 *
 * Description:
 * 	return the AmDisplayClass that the image was created with 
 *
 ****************************************************************************/

AmDisplayClass *
AmImageClass::GetDisplay()
{
    return( _pDisplay);
}

/****************************************************************************
 *
 * Description:
 * 	Class-independent image clearing function.  Clears the rectangle
 *  	in virtual coordinstes in the image. 
 * Arguments:
 *	pColor - background color to clear with.
 *	pClearRect - rectangle to clear.
 *
 ****************************************************************************/

void 
AmImageClass::ClearImageRect( AmColorCellPtr       pColor,
			    AmRectangleRec const *pClearRect)
{
    char         *pcData, *pcTemp;
    int          ii, jj, x, y, cxPixels, cyPixels, cbBytes;
    AmRectangleRec rctScreen;

/* Don't clear if the rectangle is bogus. */

    if ( AmDegenRect( pClearRect)) { 
        return;
    }

/* If the rectangle is NULL, then it's the entire extents of the image. */

    if ( pClearRect == NULL) {

        x        = 0;
        y        = 0;
        cxPixels = _cxPixels;
        cyPixels = _cyPixels;

    } else {

        GetDisplay()->ScreenRect( pClearRect, &rctScreen);

        x        = rctScreen.x; 
        y        = rctScreen.y; 
        cxPixels = GetDisplay()->DimInPixels( rctScreen.width);
        cyPixels = GetDisplay()->DimInPixels( rctScreen.height);

    }

/* Get the data, translate the <x,y> coordinates to image coordinates, and
   clear out the subrectangle depending on the image depth. */

    pcData  = AmGetData();
    pcData += x * _iBytesPerPixel;

    if ( _iDepth == 1) {

        cbBytes = _iBytesPerLine - x * _iBytesPerPixel;

	for ( ii = y; ii < cyPixels; ii++) {
	    memset( pcData, 0, cbBytes);
            pcData += _iBytesPerLine;
	}

    } else {

	for ( ii = y; ii < cyPixels; ii++) {
            pcTemp = pcData;
	    for ( jj = x; jj < cxPixels; jj++) {
		ASSIGN_PIXEL( *pColor, pcTemp);
                pcTemp += _iBytesPerPixel;
	    }
            pcData += _iBytesPerLine;
	}

    }
}
