#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/rect.h>
#include <anvil/display.h>
#include <anvil/imageP.h>


/**************************************************************************
 * 
 * Description:
 *  	project a subrect of a source matrix to a subrect of destination 
 *	using a pixel function and client data for each element of the
 *	source matrix.
 * Arguments:
 *	srcData: source matrix. Actually can be of arbitrary type.
 *	srcSize: size of each element in the source matrix.
 *	rows, cols: rows, cols of source matrix.
 *	flipRows, flipCols: controls transposition during projection.
 *	srcRect: subrectangle of source matrix to project.  Note the
 *	source rectangle is in float coordinates for subpixel projection
 *	dstRect: subrectangle in destination image.
 *	pixelFn: pixel function to apply to source matrix.
 *	clientData: passed as second argument to the pixel function.
 *	transparency: transparency value 0.0 - 1.0
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED if the differencing vectors
 *	couldn't be allocated.
 * Recommended enhancements:
 *	change the malloced differencing vectors to use alloca.
 *
 *************************************************************************/

/* ARGSUSED */

AmErrorCode
AmImageClass::resizeMatrixImage( uchar_t            *pcSrcData,
                               int                cbSrcSize,
                               int                nRows,
                               int                nCols,
                               AnvilBoolean            fFlipRows,
                               AnvilBoolean            fFlipCols,
                               AmRealRectRec const  *pSrcRealRect,
                               AmRectangleRec const *pDstRect,
                               AmColorFn            fnPixel,
                               void               *pPixelClientData,
                               float              dTransparency)
{
    int                 nHorizPixels, nVertPixels,
                        *hIndex, *vIndex, *pHIndex, *pVIndex, ii, cx, cy;
    uchar_t		*pcSrc, *pcDst, *pcTemp;
    ulong_t             *plData, *plRow, *plEndRow, *plCol, *plEndCol;
    long		lPixel;
    static ulong_t      vSrcRed[ 256], vSrcGreen[ 256], vSrcBlue[ 256],   
			vDstRed[ 256], vDstGreen[ 256], vDstBlue[ 256];
    static float        dLastTransparency = 1.0;
    AnvilBoolean		fTransparency = ( dTransparency != 1.0);
    AmRectangleRec	rctScreen;

    if ( dTransparency == 0.0) {
	return( AmOK);
    }

/* Translate the destination rectangle from virtual to screen coordinates. */

    GetDisplay()->ScreenRect( pDstRect, &rctScreen);

    cx = GetDisplay()->DimInPixels( rctScreen.width);
    cy = GetDisplay()->DimInPixels( rctScreen.height);

/* Get the information out of the destination image. */

    plData = ( ulong_t *) AmGetData();

    nHorizPixels = _cxPixels;
    nVertPixels  = _cyPixels;

/* Assign up front for efficiency. */

    pcSrc = ( uchar_t *) &lPixel;

/* Create the forward differencing vectors, then scale the horizontal
   vector by the source pixel size, and the vertical vector by the source
   row length. */ 

    if (( hIndex = AmMakeRealFwdDiffVec( pSrcRealRect->x,
                                       pSrcRealRect->x + pSrcRealRect->width,
				       cx)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    for ( ii = 0; ii < cx; ii++) {
        hIndex[ ii] *= cbSrcSize;
    }

    if (( vIndex = AmMakeRealFwdDiffVec( pSrcRealRect->y,
                                       pSrcRealRect->y + pSrcRealRect->height,
				       cy)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    pVIndex = vIndex;
    for ( ii = 0; ii < cy; ii++) {
        vIndex[ ii] *= cbSrcSize * nCols;
    }

/* Create lookup tables to translate the rgb values from the original 
   value into the transparent value. */

    if ( fTransparency) {
        if ( dTransparency != dLastTransparency) {
            AmGenRGBLookups( dTransparency,
                           vSrcRed, vSrcGreen, vSrcBlue,
                           vDstRed, vDstGreen, vDstBlue);

        }
        dLastTransparency = dTransparency;
    }

/* Offset to the beginning of the source subimage. */

    pcSrcData += (( int) pSrcRealRect->x +
                  ( int) pSrcRealRect->y*nCols)*cbSrcSize;

/* Loop until the end of the destination subimage, setting pointers at the
   start and end of each row.  Use the vertical and horizontal forward
   differencing vectors to increment through the source image.  
   dstDataLong destLong point to the current pixel and end of the 
   destination image.  The boolean fFlipRows and fFlipCols parameters 
   control the increment and decrement of the destination loops. */

    if ( fFlipRows) {

    /* Iterate through the rows in decreasing order (start at the bottom
       and go to the top).  Note that we are adding the x offset to the
       row location for speedup (this way we only have to do the addition 
       once per row). */

	plEndRow = plData + rctScreen.x + rctScreen.y * nHorizPixels;
	plRow    = plData + rctScreen.x +
                   ( rctScreen.y + cy - 1) * nHorizPixels;

	while ( plRow >= plEndRow) {

	    pHIndex = hIndex;
	    pcTemp  = pcSrcData;

	    if ( fFlipCols) {

            /* Iterate through the cols in decreasing order (start at the
               right and go to the left). */

		plCol    = plRow + cx - 1;
		plEndCol = plRow;

		while ( plCol >= plEndCol) {

		    lPixel = ( fnPixel)(( void *) pcTemp, pPixelClientData);

                /* Take the resulting pixel and blend it with the destination
                   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {

			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
                                       vDstBlue[ pcDst[ BLUE]]) |
                                     ( vSrcGreen[ pcSrc[ GREEN]] +
                                       vDstGreen[ pcDst[ GREEN]]) |
                                     ( vSrcRed[ pcSrc[ RED]] +
                                       vDstRed[ pcDst[ RED]]);

			} else {
			    *plCol = lPixel;
			}
		    }

		    plCol--;
		    pcTemp += *pHIndex++;
		}

	    } else {

            /* Iterate through the cols in increasing order (start at the
               left and go to the right). */

		plCol    = plRow;
		plEndCol = plRow + cx - 1;

		while ( plCol <= plEndCol) {

		    lPixel = ( fnPixel)(( void *) pcTemp, pPixelClientData);

                /* Take the resulting pixel and blend it with the destination
                   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {

			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
                                       vDstBlue[ pcDst[ BLUE]]) |
                                     ( vSrcGreen[ pcSrc[ GREEN]] +
                                       vDstGreen[ pcDst[ GREEN]]) |
                                     ( vSrcRed[ pcSrc[ RED]] +
                                       vDstRed[ pcDst[ RED]]);

			} else {
			    *plCol = lPixel;
			}
		    }

		    plCol++;
		    pcTemp += *pHIndex++;
		}
	    }

	    plRow -= nHorizPixels;
	    pcSrcData += *pVIndex++;
	}

    } else {

    /* Iterate through the rows in increasing order (start at the top
       and go to the bottom).  Note that we are adding the x offset to the
       row location for speedup (this way we only have to do the addition 
       once per row). */

	plRow    = plData + rctScreen.x + rctScreen.y * nHorizPixels;
	plEndRow = plData + rctScreen.x +
                   ( rctScreen.y + cy - 1) * nHorizPixels;

	while ( plRow <= plEndRow) {

	    pHIndex = hIndex;
	    pcTemp  = pcSrcData;

	    if ( fFlipCols) {

            /* Iterate through the cols in decreasing order (start at the
               right and go to the left). */

		plCol    = plRow + cx - 1;
		plEndCol = plRow;

		while ( plCol >= plEndCol) {

		    lPixel = ( fnPixel)(( void *) pcTemp, pPixelClientData);

                /* Take the resulting pixel and blend it with the destination
                   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {

			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
                                       vDstBlue[ pcDst[ BLUE]]) |
                                     ( vSrcGreen[ pcSrc[ GREEN]] +
                                       vDstGreen[ pcDst[ GREEN]]) |
                                     ( vSrcRed[ pcSrc[ RED]] +
                                       vDstRed[ pcDst[ RED]]);

			} else {
			    *plCol = lPixel;
			}
		    }

                /* Decrement the row pointer, and increment the pointer
                   to the source data (by forward differencing). */

		    plCol--;
		    pcTemp += *pHIndex++;
		}

	    } else {

            /* Iterate through the cols in increasing order (start at the
               left and go to the right). */

		plCol    = plRow;
		plEndCol = plRow + cx - 1;

		while ( plCol <= plEndCol) {

		    lPixel = ( fnPixel)(( void *) pcTemp, pPixelClientData);

                /* Take the resulting pixel and blend it with the destination
                   if there's trasparency, otherwise, just set it directly. */

		    if ( lPixel != -1) {
			if ( fTransparency) {

			    pcDst = ( uchar_t *) plCol;
			    *plCol = ( vSrcBlue[ pcSrc[ BLUE]] +
                                       vDstBlue[ pcDst[ BLUE]]) |
                                     ( vSrcGreen[ pcSrc[ GREEN]] +
                                       vDstGreen[ pcDst[ GREEN]]) |
                                     ( vSrcRed[ pcSrc[ RED]] +
                                       vDstRed[ pcDst[ RED]]);

			} else {
			    *plCol = lPixel;
			}
		    }

                /* Increment the row pointer, and increment the pointer
                   to the source data (by forward differencing). */

		    plCol++;
		    pcTemp += *pHIndex++;
		}
	    }

	    plRow     += nHorizPixels;
	    pcSrcData += *pVIndex++;
	}
    } 

/* Free the indexing vectors. */

    free( hIndex);
    free( vIndex);

    return( AmOK);
}
