#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/project.h>
#include <anvil/display.h>
#include <anvil/imageP.h>

/**************************************************************************
 *
 * Description:
 * 	default read scanline function, for dealing with nice, square
 *	source matrices. 
 * Arguments:
 *	fd - file descriptor to read from.
 *	buffer - buffer to read into.
 *	size - size of each element.
 *	length - length of buffer.
 *	cols - number of columns in the matrix.
 *	x, y - xy start of scanline.
 * 	clientData - unused.
 *
 *************************************************************************/

AnvilBoolean 
AmReadScanline( int	fd,
	      uchar_t	*buffer,
	      int	size,
	      int	length,
	      int	cols,
	      int	x,
	      int	y,
	      void	*)
{
    if ( lseek( fd, ( cols*y + x)*size, SEEK_SET) < 0) {
	return( FALSE);
    }
    if ( read( fd, buffer, length*size) < length*size) { 
	return( FALSE);
    }
    return( TRUE);
}

/**************************************************************************
 *
 * Description:
 * 	take a source matrix read from a file, and scale the scanlines to the
 *  	display image, given source and destination rectangles.
 *	WARNING: this does not work for flipRows and flipCols
 * Arguments:
 *	fd - file descriptor to read from.
 *      cbSrcSize - size of each element in the source matrix.
 *      pixelFn - pixel function to apply to source matrix.
 *      clientData - passed as second argument to the pixel function.
 *      rows, cols - rows, cols of source matrix.
 *      srcRect - subrectangle of source matrix to project.  Note the
 *      source rectangle is in float coordinates for subpixel projection
 *      dstRect - subrectangle in destination image.
 *      transparency - transparency value 0.0 - 1.0
 *	readScanFn, readScanClientData - scanline read functions.
 * Returns:
 *      AmOK on success, AmMALLOC_FAILED if the differencing vectors
 *      couldn't be allocated.
 * Recommended enhancements:
 *      change the malloced differencing vectors to use alloca.
 *
 *************************************************************************/


AmErrorCode 
AmImageClass::imageReadRows( int		   	fd,
			   int		   	cbSrcSize,
			   AmColorFn		pixelFn,
			   void		   	*clientData,
			   int		   	rows,
			   int		   	cols,
			   AmRealRectRec const    *srcRect,
			   AmRectangleRec const   *dstRect,
			   float		transparency,
			   AmScanlineFn	   	readScanFn,
			   void		   	*readScanClientData)
{
    int		     *hIndex, *vIndex, *pHIndex, *pVIndex, nHorizPixels, 
		     nVertPixels, cx, cy, ii, lastIndex, pixel, bufWidth;
    ulong_t 	     *dstDataLong, *rowStartLong, *rowEndLong, *dstEndLong;
    uchar_t 	     *buffer, *srcDataPtrChar, *srcChar, *dstChar;
    AnvilBoolean	     okReadFlag=TRUE;
    static float     lastTransp = 1.0;
    static ulong_t   imgSrcRed[ 256], imgSrcGreen[ 256], imgSrcBlue[ 256],                           imgDstRed[ 256], imgDstGreen[ 256], imgDstBlue[ 256];
    AmRealRectRec	     srcScreenRect;
    AmRectangleRec     dstScreenRect;


/* translate the source and destination rectangles from virtual to screen
   coordinates */

    GetDisplay()->ScreenRealRect( srcRect, &srcScreenRect);
    GetDisplay()->ScreenRect( dstRect, &dstScreenRect);
    cx = GetDisplay()->DimInPixels( dstScreenRect.width);
    cy = GetDisplay()->DimInPixels( dstScreenRect.height);
    nHorizPixels = _cxPixels;
    nVertPixels = _cyPixels;

    dstDataLong = (( ulong_t *) AmGetData()) + 
			dstScreenRect.y*nHorizPixels + dstScreenRect.x;
    dstEndLong = dstDataLong + dstScreenRect.height*nHorizPixels;

/* assign the address up front for efficiency */

    srcChar = ( uchar_t *) &pixel;

/* allocate the forward differencing vectors, and the read buffer.  Scale
   the horizontal vector by the source size so we can avoid a multiply
   in the core loop */

    if (( hIndex = AmMakeRealFwdDiffVec( srcScreenRect.x, 
				       srcScreenRect.x + srcScreenRect.width,
				       cx)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ii = 0; ii < cx; ii++) {
	hIndex[ ii] *= cbSrcSize;
    }
    if ((vIndex = AmMakeRealIndexVec( srcScreenRect.y, 
				    srcScreenRect.y + srcScreenRect.height, 
				    cy)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    pVIndex = vIndex;
    for ( ii = 0; ii < cy; ii++) {
	vIndex[ ii] += ( int) srcScreenRect.y;
    }
    lastIndex = -1;
    bufWidth = ( int) srcScreenRect.width + 4;
    if (( buffer = ( uchar_t *) malloc( bufWidth*cbSrcSize)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* loop until the end of the destination subimage, setting pointers at the
   start and end of each row.  Use the vertical and horizontal forward
   differencing vectors to increment through the source image */
   
    if ( transparency < 1.0) {
        if ( transparency != lastTransp) {
            AmGenRGBLookups( transparency,
                           imgSrcRed, imgSrcGreen, imgSrcBlue,
                           imgDstRed, imgDstGreen, imgDstBlue);
        }
	lastTransp = transparency;

/* seek to the next source scanline, by the index vector */
	
        while ( dstDataLong < dstEndLong) {
	    if ( *pVIndex != lastIndex) {
		if ( !( readScanFn)( fd, buffer, cbSrcSize, bufWidth, cols,
				     ( int) srcScreenRect.x, *pVIndex, 
				     readScanClientData)) {
		    okReadFlag = FALSE;
		    break;
		}
		lastIndex = *pVIndex;
	    }
	    pVIndex++;

/* map the file's scanline to the image scanline */

            pHIndex = hIndex;
            rowStartLong = dstDataLong;
            rowEndLong = dstDataLong + cx;
            srcDataPtrChar = buffer;
            while ( rowStartLong < rowEndLong) {
		pixel = ( pixelFn)( srcDataPtrChar, clientData);
		if ( pixel != -1) {
		    dstChar = ( uchar_t *) rowStartLong;
                    *rowStartLong = ( imgSrcBlue[ srcChar[ BLUE]] +
                                      imgDstBlue[ dstChar[ BLUE]]) |
                                    ( imgSrcGreen[ srcChar[ GREEN]] +
                                      imgDstGreen[ dstChar[ GREEN]]) |
                                    ( imgSrcRed[ srcChar[ RED]] +
                                      imgDstRed[ dstChar[ RED]]);
		}
                rowStartLong++;
                srcDataPtrChar += *pHIndex++;
            }
            dstDataLong += nHorizPixels;
        }
    } else {

/* seek to the next source scanline, by the index vector */
	
        while ( dstDataLong < dstEndLong) {
	    if ( *pVIndex != lastIndex) {
		if ( !( readScanFn)( fd, buffer, cbSrcSize, bufWidth, cols,
				     ( int) srcScreenRect.x, *pVIndex, 
				     readScanClientData)) {
		    okReadFlag = FALSE;
		    break;
		}
		lastIndex = *pVIndex;
	    }
	    pVIndex++;

/* map the file's scanline to the image scanline */

            pHIndex = hIndex;
            rowStartLong = dstDataLong;
            rowEndLong = dstDataLong + cx;
            srcDataPtrChar = buffer;
            while ( rowStartLong < rowEndLong) {
		pixel = ( pixelFn)( srcDataPtrChar, clientData);
		if ( pixel != -1) {
		    *rowStartLong = pixel;
		}
                rowStartLong++;
                srcDataPtrChar += *pHIndex++;
            }
            dstDataLong += nHorizPixels;
        }
    }
    free( buffer);
    free( hIndex);
    free( vIndex);
    ANVIL_ERROR_CODE = ( okReadFlag) ? AmOK : AmREAD_FAILED;
    return( ANVIL_ERROR_CODE);
}
