#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <anvil/project.h>
#include <anvil/files.h>
#include <anvil/strlib.h>
#include <anvil/errlib.h>
#include <anvil/colorcell.h>

/* file format
 
   colors can be specified either as names hex values or as RGB 0..255 values
 
    numColors
    lowest color
      .
      .
    color
      .
    highest color
*/

/* read the colors in a color file into a vector of color cells to allocate */

AnvilBoolean AmReadColors( char 		*colorfile,
		    int 		*numColors,
		    AmColorCellPtr 	*colorRet)
{
    int 		kk, red, green, blue;
    AmColorCellPtr 	colorptr;
    FILE 		*fp;
    char 		string[ 256], *namePtr, *p;

    if (( fp = fopen( colorfile, "r")) == NULL) {
        AmPrintError( "Failed to open colorfile %s", colorfile);
        return( FALSE);
    }

/* find out if he put the number of colors at the head of the file.  If so,
   use it, otherwise, read the number of lines to determine the number of 
   colors */
    
    if ( fgets( string, sizeof( string), fp) == NULL) {
	AmPrintError( "Error: empty file %s", colorfile);
	fclose( fp);
	return( FALSE);
    }
    string[ strlen( string) - 1] = '\0';

/* make sure it's a number followed by white space and a newline */

    p = AmElimLeadWspace( string);
    for ( ; ( *p != '\0') && isdigit( *p); p++);
    for ( ; ( *p != '\0') && isspace( *p); p++);

/* AmFirst line is color -> no length info. count lines for number of colors */ 

    if ( *p != 0) {
	*numColors = AmNumLines( colorfile);
	rewind( fp);
    } else if ( sscanf( string, "%d\n", numColors) != 1) {
        AmPrintError( "Bad header in colorfile %s %s: %d (311)\n",
                 colorfile, __FILE__, __LINE__);
	fclose( fp);
        return( FALSE);
    }
    *colorRet = colorptr = 
	( AmColorCellPtr) calloc( sizeof( AmColorCellRec), *numColors);

/* read the colors from the file and save them in sampleSet */

    for ( kk = *numColors; kk; kk--) {
	colorptr->pixel = kk;
        if ( fgets( string, sizeof( string), fp) == NULL) {
	    AmPrintError( "Wrong # of records in file %s", colorfile);
	    fclose( fp);
	    *numColors = *numColors - kk;
	    return( TRUE);			/* we did read SOME colors */
	}
		    
/* crop the \n and move up the white space */

        string[ strlen( string) - 1] = '\0';
	namePtr = AmElimLeadWspace( string);

/* allocate any colors specified by RGB values */

        if( isdigit( *namePtr)) {
	    if ( sscanf( namePtr, "%d %d %d", &red, &green, &blue) != 3) {
		AmPrintError( "Bad record in file %s", colorfile);
                ( *numColors)--;
	    } else {
		colorptr->red = red;
		colorptr->green = green;
		colorptr->blue = blue;
		colorptr++;
	    }
        }
    }
    fclose( fp);
    return( TRUE);
}

/* read a string of RGB values into a color cell vector */

AmColorCellPtr AmReadColorCellStr( char *str)
{   
    int         	ii, length, *intvec;
    char        	*p;
    AmColorCellPtr	colorVec;

/* scan over each digit to find the length don't forget negatives */

    length = 0;
    for ( p = str; *p != '\0'; ) {
        while ( isdigit( *p++));
        while (( *p != '\0') &&  !isdigit( *p)) {
            p++;
        }
        length++;
    }
    intvec = AmReadIntStrVec( str);
    length /= 3;
    colorVec = ( AmColorCellPtr) malloc( sizeof( AmColorCellRec)*length);
    for ( ii = 0; ii < length; ii++) {
	colorVec[ ii].red = intvec[ ii*3];
	colorVec[ ii].green = intvec[ ii*3 + 1];
	colorVec[ ii].blue = intvec[ ii*3 + 2];
    }
    free( intvec);
    return( colorVec);
}

/* interpolate a set of colors */

void AmBetweenColorcell( double		frac,
		       AmColorCellPtr	color1,
		       AmColorCellPtr	color2,
		       AmColorCellPtr	retColor)
{
    retColor->red = 
	( unsigned char) (( 1.0 - frac)*color1->red + frac*color2->red);
    retColor->blue = 
	( unsigned char) (( 1.0 - frac)*color1->blue + frac*color2->blue);
    retColor->green = 
	( unsigned char) (( 1.0 - frac)*color1->green + frac*color2->green);
}

/* given a set of colorcells, memalloc and interpolate a new set of xcolors */
 
void AmInterpolateColorCells( AmColorCellPtr   colors,
                            int            numColors,
                            AmColorCellPtr   *newColors,
                            int            numNewColors)
{
    double      realIndex, step;
    int         ii, intIndex;
 
    *newColors = ( AmColorCellPtr) calloc( sizeof( AmColorCellRec), numNewColors);
    ( *newColors)[ 0] = colors[ 0];
    ( *newColors)[ numNewColors - 1] = colors[ numColors - 1];
    step = (( double) numColors - 1)/(( double) numNewColors - 1);
    for ( ii = 1, realIndex = step;
          ii < numNewColors - 1;
          ii++, realIndex += step) {
        intIndex = ( int) floor( realIndex);
        AmBetweenColorcell( realIndex - intIndex,
			  &(colors[ intIndex]), &(colors[ intIndex + 1]),
			  &((*newColors)[ ii]));
    }
}

/* take a set of colors and apply shades to them */

void AmShadeColors( AmColorCellPtr	colors,
		  AmColorCellPtr	shades,
		  int		numColors,
		  int		numShades)
{
    int	ii, jj;

    for ( ii = 0; ii < numShades; ii++) {
	for ( jj = 0; jj < numColors; jj++) {
	    shades[ ii*numColors + jj].red =
		(colors[ jj].red*ii + 1)/numShades;
	    shades[ ii*numColors + jj].green =
		(colors[ jj].green*ii + 1)/numShades;
	    shades[ ii*numColors + jj].blue =
		(colors[ jj].blue*ii + 1)/numShades;
	}
    }
}
