#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <anvil/macros.h>
#include <anvil/overlay.h>
#include <anvil/userevents.h>
#include <anvil/eventcodes.h>
#include <anvil/list.h>
#include <anvil/log.h>
#include <anvil/dispmgr.h>

/* Function wrapper to call delete, because C++ semantics suck. */

static void
freeObj( AmObjRectPtr objRect)
{
    delete objRect->obj;
    free( objRect);
}

/* True if an object has a parent */

static int
hasParent( void       *,
	   AmObjRectPtr objRect)
{
    return( objRect->obj->getParent() != NULL);
}

/***********************************************************************
 *
 * Function: AmDisplayManagerClass::AmDisplayManagerClass()
 *
 * Description:
 *   Do nothing default constructor.
 *
 ***********************************************************************/

AmDisplayManagerClass::AmDisplayManagerClass()
{
    _fNeedResize = FALSE;
}

/***********************************************************************
 *
 * Function: AmDisplayManagerClass( pDisplayClass, pWindowClass,
 *                                  fDoubleBuffer)
 * 
 * Description:
 *   Constructor for the DisplayManager.   It initializes the window,
 *   display, and creates the gc, and double buffer (if requested).
 *   Then it initializes the object and event lists, and sets up the
 *   callback handlers for the events sent by screen objects.
 *
 * Arguments:
 *   AmDisplayClass *pDisplayClass - Display to draw to.
 *   AmWindowClass  *pWindowClass  - Window to manage objects in.
 *   AnvilBoolean        fDoubleBuffer  - Whether or not to use double buffering.
 *
 ***********************************************************************/

AmDisplayManagerClass::AmDisplayManagerClass( AmDisplayClass *pDisplayClass,
                                              AmWindowClass  *pWindowClass,
                                              AnvilBoolean        fDoubleBuffer)
{
    int            x, y, cx, cy;
    AmColorCellRec color;

/* Initializations. */

    _pDisplay = pDisplayClass; 
    _pWindow  = pWindowClass;

/* Get the dimensions of the window (these are in "virtual coordinates")
   and keep them around. */

    _pWindow->GetWindowGeometry( &x, &y, &cx, &cy);
    
    _xOrigin = x;
    _yOrigin = y;
    _iWidth  = cx;
    _iHeight = cy;
    
/* Create the GC for use in drawing. */
    
    _pGC = _pWindow->CreateGC();

/* Create the cursor GC */

    color.red   = 0x7f;
    color.green = 0x7f;
    color.blue  = 0x7f;

    _pCursorGC = pWindowClass->CreateGC();
    _pCursorGC->SetRasterOp( AmXOR);
    _pCursorGC->SetForegroundColor( &color);

/* Initialize the object list, event list, deleted object list, valid flag
   and "just resized" flag. */

    _lstObj       = NULL;
    _lstEvent     = NULL;
    _lstDelete    = NULL;
    _fValid       = TRUE;
    _fJustResized = FALSE;
    _fNeedResize  = FALSE;

/* Handle double buffering. */

    if ( fDoubleBuffer == TRUE) {

    /* If it's double buffered, then create a pixmap double buffer 
       and clear it out. */

        _pWindowBuffer = _pDisplay->CreateBuffer( _pWindow, cx, cy);
        ClearArea();

    } else {

    /* Otherwise set the pixmap to NULL. */

        _pWindowBuffer = NULL;
    }
    
/* Add the callback which erases objects when they're deleted, and redraws
   the children when the parent is updated. */

    AmAddCallback( AM_PKG_ID, AM_DELETE_OBJECT, 
		 ( callbackProc) AmDeleteObjectCallback, this);
    AmAddCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT, 
		 ( callbackProc) AmUpdateObjectCallback, this);
    AmAddCallback( AM_PKG_ID, AM_COORDS_CHANGE, 
		 ( callbackProc) AmUpdateObjectCallback, this);
    AmAddCallback( AM_PKG_ID, AM_RECT_CHANGE, 
		 ( callbackProc) AmUpdateObjectCallback, this);
}

/***********************************************************************
 *
 * Function: ~AmDisplayManagerClass()
 * 
 * Description:
 *   Destructor for the AmDisplayManagerClass.  It removes the callback
 *   handlers, then deletes all of the objects with parents, all the
 *   non-parented objects. Finally it frees the display, windows and
 *   graphics contexts.
 *
 ***********************************************************************/

AmDisplayManagerClass::~AmDisplayManagerClass()
{
    
/* Remove the callback which erases the objects when they're deleted */
    
    AmRemoveCallback( AM_PKG_ID, AM_DELETE_OBJECT, 
		    ( callbackProc) AmDeleteObjectCallback, this);
    AmRemoveCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT, 
		    ( callbackProc) AmUpdateObjectCallback, this);
    AmRemoveCallback( AM_PKG_ID, AM_COORDS_CHANGE, 
		    ( callbackProc) AmUpdateObjectCallback, this);
    AmRemoveCallback( AM_PKG_ID, AM_RECT_CHANGE, 
		    ( callbackProc) AmUpdateObjectCallback, this);
    
/* Do the objects with parents AmFirst, then objects without parents, because
   parents try to delete their children. */ 
    
    _lstObj = AmDeleteItemLst( NULL, _lstObj, 
    			     ( AmSearchFn) hasParent, ( AmItemFn) freeObj);
    AmMapLst( _lstObj, ( AmItemFn) freeObj);
    AmJustFreeLst( _lstObj);

/* free any outstanding events added with AddEvent */

    AmFreeLst( _lstEvent, ( AmItemFn) free);
    
/* Delete the windows, gc's and stuff. */
   
    if ( _pGC != NULL) { 
        delete _pGC;
    }

    if ( _pCursorGC != NULL) { 
        delete _pCursorGC;
    }

    if ( _pWindowBuffer != NULL) {
        delete _pWindowBuffer;
    }

    if ( _pWindow != NULL) {
        delete _pWindow;
    }

    if ( _pDisplay != NULL) {
        delete _pDisplay;
    }
}

/***********************************************************************
 *
 * Function: GetExposeRects( vExposeRec, pnExposeRec)
 * 
 * Parameters:
 *   AmRectangleRec const *vExposeRec  : Vector of expose rectangles (may
 *                                     be NULL).
 *   int                *pnExposeRec : Pointer to number of expose rectangles
 *                                     (may be 0).
 *
 * Description:
 *   If the vector of expose rectangles is NULL, returns a single
 *   rectangle which is the extents of the display manager's
 *   drawing area, otherwise returns the rectangle vector.  Also
 *   updates the number of expose rectangles (side effec).
 *
 * Note:
 *   The function returns a reference to a local static variable.  That
 *   means everytime the function gets called the rectangle changes, so
 *   do not rely on the returned rectangle to be around.
 *
 ***********************************************************************/

AmRectangleRec const *
AmDisplayManagerClass::GetExposeRects( AmRectangleRec const *vExposeRec,
                                     int                *pnExposeRec)
{
    static AmRectangleRec	rec;

    if ( vExposeRec == NULL) {

    /* There is no expose rectangle, so return a rectangle with the
       extents of the window. */

        rec.x = _xOrigin;
        rec.y = _yOrigin;
        rec.width = _iWidth;
        rec.height = _iHeight;
        *pnExposeRec = 1;

        return( &rec);

    } else {

    /* The expose rectangles are already valid, just return them. */

    	return( vExposeRec);
    }
}

/***********************************************************************
 *
 * Function: ClearArea( vClearRects, nClearRects)
 * 
 * Parameters:
 *   const AmRectangleRec *vClearRecs : The rectangular regions to be cleared
 *                                    ( can be NULL).
 *   int                nClearRects : The number of rectangles ( can be 0).
 *
 * Description:
 *   ClearArea clear rectangular areas in the window or buffer.
 *
 * Returns:
 *   -nothing-
 *
 ***********************************************************************/

void
AmDisplayManagerClass::ClearArea( AmRectangleRec const *vClearRects,
                                int                nClearRects)
{
    AmRectangleRec const *vTempRects;
    int                ii;

/* Do nothing if the window isn't valid. */
 
    if (( !_pWindow->IsValid()) && ( _pWindowBuffer == NULL)) {
        return;
    }

/* Get the clip rectangles (in case vClearRects was NULL). */

    vTempRects = GetExposeRects( vClearRects, &nClearRects);

/* Set the clip rectangles and clear the areas. */
 
    _pGC->SetClipRectangles( vTempRects, nClearRects);

    for ( ii = 0; ii < nClearRects; ii++) {
        if ( _pWindowBuffer != NULL) {
            _pWindowBuffer->ClearRect( &vTempRects[ ii]);
        } else {
            _pWindow->ClearRect( &vTempRects[ ii]);
        }
    }

/* Reset the clip rectangle to the window's extents. */

    _pGC->SetClipRectangle( _xOrigin, _yOrigin, _iWidth, _iHeight);
}

/***********************************************************************
 *
 * Function: ReDraw( vExposeRects, nExposeRects)
 * 
 * Parameters:
 *   AmRectangleRec const *vExposeRects : Vector of expose rectangles (may
 *                                      be NULL).
 *   int                nExposeRects  : Number of expose rectangles (may
 *                                      be 0).
 *
 * Description:
 *   If the Display Manager is double buffered, then it does a
 *   copy area on each of the expose rectangles, otherwise, 
 *   it calls Draw( vExposeRects, nExposeRects).
 *
 * Returns:
 *   -nothing-
 *
 ***********************************************************************/

void
AmDisplayManagerClass::ReDraw( AmRectangleRec const *vExposeRects,
                             int                nExposeRects)
{
    AmRectangleRec const	*vTempRects;
    int                 ii, cx, cy;

/* If the window is invalid, there is nothing that we can do here. */

    if ( !_pWindow->IsValid()) {
        return;
    }
 
    if ( _pWindowBuffer == NULL) {

    /* If there is no double buffer, redraw the entire window contents. */

        Draw( vExposeRects, nExposeRects);

    } else {

    /* If there is an outstanding resize that couldn't be processed before
       the expose, then process it before drawing. */

	if ( _fNeedResize) {
	    _pWindow->GetDims( &cx, &cy);
	    Resize( cx, cy);
	    _fNeedResize = FALSE;
	}

        if ( _fJustResized == TRUE) {
            _fJustResized = FALSE;
            return;
        }

    	vTempRects = GetExposeRects( vExposeRects, &nExposeRects);
	
    /* It's double buffered so blit the buffer for each one of the
       exposed rectangles. */

        for ( ii = 0; ii < nExposeRects; ii++) {
            _pDisplay->CopyArea( _pWindowBuffer, _pWindow, _pGC,
                                 vTempRects[ ii].x,
                                 vTempRects[ ii].y,
                                 vTempRects[ ii].x,
                                 vTempRects[ ii].y,
                                 vTempRects[ ii].width,
                                 vTempRects[ ii].height);
        }
    }
}

/***********************************************************************
 *
 * Function: Draw( vExposeRects, nExposeRects)
 * 
 * Parameters:
 *   AmRectangleRec const *vExposeRects : Vector of expose rectangles (may
 *                                      be NULL).
 *   int                nExposeRects  :  Number of expose rectangles (may
 *                                      be 0).
 *
 * Description:
 *   Calls Draw on each of the screen objects which are exposed
 *   by the rectangles in vExposeRect.  The Draw functions are 
 *   called by the objects' stacking order.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::Draw( AmRectangleRec const *vExposeRects,
                           int                nExposeRects)
{
    AmWindowClass        *pWindow;
    AmListPtr            lstTemp;
    AmRectangleRec const *vTempRects;
    AmRectangleRec       *vClipRects;
    int                nClipRects, ii, cx, cy;
    AmObjRectPtr         pObjRect;

/* If the window is NULL, then bail. */

    if ( !_pWindow->IsValid()) {
        return;
    }

/* If there is an outstanding resize that couldn't be processed before
   the expose, then process it before drawing. */

    if ( _fNeedResize) {
	_pWindow->GetDims( &cx, &cy);
	Resize( cx, cy);
	_fNeedResize = FALSE;
    }

/* If the window has just resized, then we don't redraw it, but we DO
   turn the flag off. */

    if ( _fJustResized) {
    	_fJustResized = FALSE;
        return;
    }
    
/* Clear the areas which are to be redrawn. */

    ClearArea( vExposeRects, nExposeRects);

/* Figure out which window class to draw to. */

    if ( _pWindowBuffer != NULL) {
        pWindow = _pWindowBuffer;
    } else {
        pWindow = _pWindow;
    }

/*
 * If the expose rectangle is NULL, then it's assumed to be the entire 
 *   extents of the display.
 */

    vTempRects = GetExposeRects( vExposeRects, &nExposeRects);
    
/* Walk down the object list and draw each of the exposed objects. */

    vClipRects = ( AmRectanglePtr) alloca( sizeof( AmRectangleRec) * nExposeRects);
    for ( lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {

        pObjRect = ( AmObjRectPtr) lstTemp->data;
	
        /*
         * Copy the expose rectangles and clip them by the object's
         *   rectangle.
         */

        memcpy( vClipRects, vTempRects, sizeof( AmRectangleRec) * nExposeRects);
        nClipRects = nExposeRects;

        if ( AmClipRectVec( pObjRect->obj->GetExtentsRect(), vClipRects, 
                          &nClipRects)) {
            _pGC->SetClipRectangles( vClipRects, nClipRects);
            pObjRect->obj->Draw( pWindow, _pGC, vClipRects, nClipRects);
        }
    }
    
    /* Unset the clipping rectangle. */

    _pGC->SetClipRectangle( _xOrigin, _yOrigin, _iWidth, _iHeight);
     
    /* If it's double buffered, then blit the buffer. */
    
    if ( _pWindowBuffer != NULL) {
        for ( ii = 0; ii < nExposeRects; ii++) {
            _pDisplay->CopyArea( _pWindowBuffer, _pWindow, _pGC, 
                                 vTempRects[ ii].x, vTempRects[ ii].y,
                                 vTempRects[ ii].x, vTempRects[ ii].y,
                                 vTempRects[ ii].width, vTempRects[ ii].height);
        }
    }
}

/***********************************************************************
 *
 * Function: Update()
 * 
 * Description:
 *   Calls Update on each of the screen objects, then calls Draw()
 *   with the default exposeRect vector of NULL, redrawing all of the
 *   objects. AM_SET_FIELD_OBJECT is used for the event code in update()
 *   because that is the most thorough update request.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::Update()
{
    AmListPtr    lstTemp;
    AmObjRectPtr pObjRect;
    int          iEventCode = AM_SET_FIELD_OBJECT;
    
    for ( lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {
        pObjRect = ( AmObjRectPtr) lstTemp->data;
        if ( pObjRect->obj->getParent() == NULL) {
            pObjRect->obj->Update( &iEventCode, 1);
        }
    }

    Draw(); 
}

/* Comparison function for objects in the object-rectangle list */

static int
equalObjObjRect( AmScreenObjClass *pScreenObj,
                 AmObjRectPtr     pObjRect)
{
    return( pScreenObj == pObjRect->obj);
}

/***********************************************************************
 *
 * Function: Resize( newWidth, newHeight);
 * 
 * Description:
 *	creates the new pixmap double buffer (if the display is double
 *	buffered),resets the coordinates, and sets the new window sized
 *	for each of the screen objects.
 * Arguments:
 *	newWidth, newHeight: new dimensions of window.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::Resize( int cx,
                               int cy)
{
    AmListPtr    lstTemp;
    AmObjRectPtr pObjRect;
   
/* If the window isn't valid (because it hasn't been exposed), then
   defer the resize until the expose is called. */

    if ( !_pWindow->IsValid()) {
	_fNeedResize = TRUE;
	return;
    }

/* Save the new dimensions. */

    _xOrigin      = VirtualCoord( 0);
    _yOrigin      = VirtualCoord( 0);
    _iWidth       = cx;
    _iHeight      = cy;
    _fJustResized = TRUE;

/* Set the GC to reflect the resized extents */

    _pGC->SetClipRectangle( _xOrigin, _yOrigin, _iWidth, _iHeight); 

/* If the display is double buffered, then resize the buffer. */

    if ( _pWindowBuffer != NULL) {
        _pWindowBuffer->ResizeWindow( _iWidth, _iHeight);
    }

/* Clear the display, since we are redrawing everything anyways.
   This is not optimal, but it is how Flush() currently works.
   Probably due to the SetWindowSize() calls below. */

    ClearArea();

/* Loop through the list of screen objects and update their window
   sizes, then flush to refresh the screen to incoorperate any of
   the changes. */
    
    for ( lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {
        pObjRect = ( AmObjRectPtr) lstTemp->data;
        pObjRect->obj->SetWindowSize( _iWidth, _iHeight);
    }

    Flush();

// Signal that the display has changed.

    AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

}

/***********************************************************************
 *
 * Function: AmDeleteObjectCallback( iEventID, pDisplayManager, pScreenObj)
 * 
 * Description:
 * 	This function is called from a user event in the object's destructor.
 * 	it calls deleteObject on each of the children, which are deleted
 * 	themselves, but not on the object itself. 
 *
 * Arguments:
 *	iEventID: event ID callback is attached to.
 *	pDisplayManager: display manager controlling objects.
 *	pScreenObj: screen object to be deleted (and sending event).
 *
 ***********************************************************************/

/* ARGSUSED */
void
AmDisplayManagerClass::AmDeleteObjectCallback( int                 iEventID,
                                           AmDisplayManagerClass *pDisplayManager,
                                           AmScreenObjClass      *pScreenObj)
{
    pDisplayManager->deleteObject( pScreenObj);
}

/***********************************************************************
 *
 * Function: deleteObject( pScreenObj)
 * 
 * Description:
 * 	Delete an object from the display list.
 * Arguments:
 *	pScreenObj: screen object to be deleted (and sending event).
 *
 ***********************************************************************/

void
AmDisplayManagerClass::deleteObject( AmScreenObjClass *pScreenObj)
{
    AmScreenObjClass **ppScreenObj, *pScreenObjTemp;
    AnvilBoolean        fExists, fTopLevel;
    int            ii, iOrder;
    AmListPtr        lstDelete, lstTemp;

// If the object isn't anything in the object list, then we don't bother.

    inObjectList( pScreenObj, &fTopLevel, &fExists, &iOrder);

// Delete the object from the display list, if it's in there.

    if ( fExists) {

    // If it's a top level object then it needs to be removed from the
    //   display list.

        if ( fTopLevel) {

            AddDelete( pScreenObj, iOrder);
            RemoveObjectFromList( pScreenObj);

        }

    // Signal that something on the display has changed.

        AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    }

// Check if any objects in the display list depend on the object that
//   was just deleted.

    ppScreenObj = GetObjects(); 

    lstDelete = NULL;
    for ( ii = 0; ppScreenObj[ ii] != NULL; ii++) {
        if ( ppScreenObj[ ii]->getParent() == pScreenObj) {
	        lstDelete = AmInsertNodeLst( ppScreenObj[ ii], lstDelete);
        }
    }

// The delete function wil cause deleteObject to be called recursively,
//   because it kicks it off by user event.

    for ( lstTemp = lstDelete; 
          lstTemp != NULL; 
          lstTemp = lstTemp->next) {
        pScreenObjTemp = ( AmScreenObjClass *) lstTemp->data;
        delete pScreenObjTemp;
    }

// Free the temporary delete list and object vector.

    AmJustFreeLst( lstDelete);
    free( ppScreenObj);

}


/***********************************************************************
 *
 * Function: inObjectList( pScreenObj)
 * 
 * Description:
 * 	Given a screen object, return whether it is in the top level
 *	of the display list (or contained in an overlay), whether it
 *	exists at all, and the order in the display list (which determines
 *	stacking order).
 * Arguments/Returns:
 *	pScreenObj: object to search for.
 *	pfTopLevel: TRUE/FALSE object at display list top level
 *	pfExists: TRUE/FALSE object is in display list.
 *	piOrder: order in the display list of the object.
 *
 ***********************************************************************/

void 
AmDisplayManagerClass::inObjectList( AmScreenObjClass	*pScreenObj,
				   AnvilBoolean		*pfTopLevel,
				   AnvilBoolean		*pfExists,
				   int			*piOrder)
{
    AmListPtr	     lstTemp;
    AmOverlayClass     *pOverlay;
    AmObjRectPtr	     pObjRec;
    AmDataGraphicClass **ppDataGraphic;
    int		     nDataGraphic, ii;
    AnvilBoolean	     fFound;

/* Initializations: doesn't exist, not at top level, and 0 order */
   
    *pfExists = FALSE;
    *pfTopLevel = FALSE;
    *piOrder = 0;

/* walk the list of obj rect ptrs (which contain the screen objects) */

    for ( lstTemp = _lstObj, *piOrder = 0; 
	  lstTemp != NULL; 
	  lstTemp = lstTemp->next, (*piOrder)++) {
	pObjRec = ( AmObjRectPtr) lstTemp->data;

/* exists and is top level... */

	if ( pObjRec->obj == pScreenObj) {
	    *pfTopLevel = TRUE;
	    *pfExists = TRUE;
	    return;
	} else if ( pObjRec->obj->IsType( AmOVERLAY)) {

/* if the object is an overlay, see if the screen object is contained in
   the overlay */

	    pOverlay = ( AmOverlayClass *) pObjRec->obj;
	    pOverlay->GetLayers( &ppDataGraphic, &nDataGraphic);
	    if ( ppDataGraphic != NULL) {
		fFound = FALSE;
		for ( ii = 0; ii < nDataGraphic; ii++) {
		    if ( ppDataGraphic[ ii] == pScreenObj) {
			fFound = TRUE;
                        break;
		    }
		}
		free( ppDataGraphic);

/* found, but not top level */

		if ( fFound == TRUE) {
		    *pfTopLevel = FALSE;
		    *pfExists = TRUE;
		    return;
		}
	    }
	}
    }
}

/***********************************************************************
 *
 * Function: AmUpdateObjectCallback( iEventID, displayManager, pScreenObj)
 * 
 * Description:
 * 	Receives events sent when objects are changed, and adds them
 *	to the display manager's event list using AddEvent().  The
 *	function tests to ensure that the objects are managed by
 *	this display manager.
 * Arguments/Returns:
 *	iEventID: ID of the event being sent.
 *	pDisplayManager: this display manager.
 *	pScreenObj: object sending the event.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::AmUpdateObjectCallback( 
			int               	iEventID,
		        AmDisplayManagerClass 	*pDisplayManager,
		        AmScreenObjClass    	*pScreenObj)
{
    AnvilBoolean fTopLevel, fExists;
    int     iOrder;

    pDisplayManager->inObjectList( pScreenObj, &fTopLevel, &fExists, &iOrder);

    if ( fExists) {

	pDisplayManager->AddEvent( pScreenObj, iEventID, iOrder, fTopLevel);
        pDisplayManager->AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE,
                                          pDisplayManager);

    }
}

/***********************************************************************
 *
 * Function: RemoveObjectSet( ppScreenObj, nScreenObj)
 * 
 * Description:
 * 	Since removing an object will remove its children, this 
 * 	convienience function removes a set of objects and guarantees that
 * 	there are no conflicts. 
 * Arguments/Returns:
 *	ppScreenObject: vector of pointers to screen objects.
 *	nScreenObj: number of screen objects in the vector.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::RemoveObjectSet( AmScreenObjClass **ppScreenObj,
                                        int              nScreenObj)
{
    int		     ii, jj, kk, n, nDataGraphic;
    AnvilBoolean	     fFound;
    AmScreenObjClass   **ppScreenObjCopy, *pScreenObjParent;
    AmDataGraphicClass **ppDataGraphic;

    ppScreenObjCopy = ( AmScreenObjClass **) alloca(
                                                 nScreenObj * 
                                                 sizeof( AmScreenObjClass *));

    if ( ppScreenObjCopy == NULL) {
        DO_ERROR( AmALLOCA_FAILED);
        return;
    }

// Scan through the objects in the deletion set ...

    n = 0;
    for ( ii = 0; ii < nScreenObj; ii++) {

	fFound = FALSE;
        pScreenObjParent = ppScreenObj[ ii]->getParent();

	for ( jj = 0; jj < nScreenObj; jj++) {

        // If the parent of the object is in the deletion set, then remove
        //   the object from the deleting set ...

	    if ( pScreenObjParent == ppScreenObj[ jj]) {
		fFound = TRUE;
		break;
	    }
	    
	// Check to make sure the parent isn't a layer in an overlay
	//   to be removed.
	    
	    if ( ppScreenObj[ jj]->IsType( AmOVERLAY)) {
		(( AmOverlayClass *) ppScreenObj[ jj])->GetLayers(
                                                             &ppDataGraphic,
                                                             &nDataGraphic);
		for ( kk = 0; kk < nDataGraphic; kk++) {
		    if ( pScreenObjParent == ppDataGraphic[ kk]) {
			fFound = TRUE;
			break;
		    }
		}

		free( ppDataGraphic);
	    }
	}

	if ( fFound == FALSE) {
	    ppScreenObjCopy[ n++] = ppScreenObj[ ii];
	}
    }
    
// Deleting each of the objects will remove it from the motif display's
//   object list, and the object will cause its children to be deleted.
    
    for ( ii = 0; ii < n; ii++) {
	delete ppScreenObjCopy[ ii];
    }

    free( ppScreenObj);
}

/***********************************************************************
 *
 * Function: AddObject( pScreenObj)
 * 
 * Description:
 *   Add a screen object to the display manager's object list, and
 *   set the screen object's extents (if it has margins set.
 *
 * Arguments:
 *   AmScreenObjClass *pScreenObj - Screen object to add to list.
 *
 ***********************************************************************/

AmErrorCode
AmDisplayManagerClass::AddObject( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect;

    if (( pObjRect = ( AmObjRectPtr) malloc( sizeof( AmObjRectRec))) == NULL) {
        AmAddResourceToLog( "mallocFailed", sizeof( AmObjRectRec));
	DO_RETURN_ERROR( AmMALLOC_FAILED)
    }

    pObjRect->obj = pScreenObj;

    _lstObj = AmInsertNodeAtEndLst( pObjRect, _lstObj);

    pScreenObj->SetWindowSize( _iWidth, _iHeight);
    pObjRect->oldRect = *( pScreenObj->GetExtentsRect());

// Signal that something on the display has changed.

    AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    return( AmOK);
}

/* Matching funtction for screen objects and obj-rect lists (used for
   list processing functions). */

/***********************************************************************
 *
 * Function: eventRectObj( pScreenObj, pEventRec)
 * 
 * Description:
 *	Remove an object from the display manager's object list, and
 *	any outstanding references to it in the event list.  This does
 *	not actually delete the object itself.
 * Arguments:
 *      pScreenObj - screen object to remove.
 *
 ***********************************************************************/

static AnvilBoolean
eventRectObj( AmScreenObjClass *pScreenObj,
	      AmEventRectPtr   pEventRec)
{
    return( pEventRec->obj == pScreenObj);
}

/***********************************************************************
 *
 * Function: RemoveObjectFromList( AmScreenObjClass *pScreenObj)
 * 
 * Description:
 *	Remove an object from the display manager's object list, and
 *	any outstanding references to it in the event list.  This does
 *	not actually delete the object itself.
 * Arguments:
 *      pScreenObj - screen object to remove.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::RemoveObjectFromList( AmScreenObjClass	*pScreenObj)
{
    _lstEvent = AmDeleteItemLst( pScreenObj, _lstEvent,
                               ( AmSearchFn) eventRectObj, ( AmItemFn) free);
    _lstObj = AmDelPtrLst( pScreenObj, _lstObj, ( AmSearchFn) equalObjObjRect);

// Signal that something on the display has changed.

    AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

}

/***********************************************************************
 *
 * Function: RemoveObject( AmScreenObjClass *pScreenObj)
 * 
 * Description:
 *   Call the screen object's destructor.  This is a wrapper in
 *   case we need to do anything else.  Call the object's destructor, 
 *   which will kick off a user event which will delete all of its 
 *   children, and erase it. 
 *
 * Arguments:
 *   pScreenObj - screen object to remove.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::RemoveObject( AmScreenObjClass	*pScreenObj)
{
    delete pScreenObj;
}

/***********************************************************************
 *
 * Function: GetObjects()
 * 
 * Description:
 *   Returns a vector of screen objects maintained by the display
 *   manager.
 * 
 * Returns:
 *   NULL-terminated vector of pointers to screen objects.
 *
 ***********************************************************************/

AmScreenObjClass **
AmDisplayManagerClass::GetObjects()
{
    AmObjRectPtr	   pObjRec;
    AmScreenObjClass **ppScreenObj;
    AmListPtr        lstTemp;
    int            nScreenObj, ii;
    
    nScreenObj = AmLengthLst( _lstObj);
    ppScreenObj =  ( AmScreenObjClass **) malloc( sizeof( AmScreenObjClass *) *
                                                ( nScreenObj + 1));
    if ( ppScreenObj == NULL) {	
	DO_ERROR( AmMALLOC_FAILED);
        return( NULL);
    }
    
    for ( ii = 0, lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {
	pObjRec = ( AmObjRectPtr) lstTemp->data;
        ppScreenObj[ ii++] = pObjRec->obj;
    }

    ppScreenObj[ ii] = NULL;
    return( ppScreenObj);   
}


/***********************************************************************
 *
 * Function: FindObject( x, y)
 * 
 * Description:
 *      Given an x,y coordinate, return the top most enclosing screen object
 *
 * Arguments:
 *	 x,y: position in VIRTUAL COORDINATES
 *
 * Returns:
 *   Pointer to the AmFirst enclosing screen object, NULL if no object
 *   was found.
 *
 ***********************************************************************/

AmScreenObjClass *
AmDisplayManagerClass::FindObject( int x, 
			         int y)
{
    AmListPtr            lstTemp;
    AmObjRectPtr         pObjRec;
    const AmRectangleRec *pRec;
    AmScreenObjClass     *pScreenObj;
    
    pScreenObj = NULL;

    for ( lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {
	pObjRec = ( AmObjRectPtr) lstTemp->data;
        pRec = pObjRec->obj->GetRect();
	
	/* Check to see if the coordinate is within the extents of this object.
	   Note that the AmLast object in the list is selected, since it's
	   the topmost object */
	
	if ( AmPtInRect( pRec, x, y)) {
	    pScreenObj = pObjRec->obj;
        }
    }
    return( pScreenObj);
}

/***********************************************************************
 *
 * Function: IsTop( pScreenObj)
 * 
 * Description:
 *   Is an object on top of all the other objects in the display list?
 * 
 * Arguments:
 *   pScreenObj - object to test.
 *
 * Returns:
 *   TRUE/FALSE if object is frontmost.
 *
 ***********************************************************************/

AnvilBoolean
AmDisplayManagerClass::IsTopObject( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect;
    
    if ( _lstObj == NULL) {  
        return( FALSE);
    }
    pObjRect = ( AmObjRectPtr) AmLast( _lstObj);
    return( pScreenObj == pObjRect->obj);
}

/***********************************************************************
 *
 * Function: PopToTop( pScreenObj)
 * 
 * Description:
 *   Move an object to the end of the display list, so it displays
 *   on top.
 * 
 * Arguments:
 *	 pScreenObj - object to move.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::PopToTop( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect;
    
    pObjRect = ( AmObjRectPtr) AmFindDataLst( pScreenObj, _lstObj, 
                                          ( AmSearchFn) equalObjObjRect);

    if ( pObjRect != NULL) {
        _lstObj = AmDelPtrLst( pObjRect, _lstObj, ( AmSearchFn) AmEquals);
        _lstObj = AmInsertNodeAtEndLst( pObjRect, _lstObj);

        AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, pScreenObj);

    // Signal that something on the display has changed.

        AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    }
}

/***********************************************************************
 *
 * Function: PushToBottom( pScreenObj)
 * 
 * Description:
 *   Send an object to the front of the display list, so it displays
 *   on the bottom.
 * 
 * Parameters:
 *   AmScreenObjClass *pScreenObj - object to move.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::PushToBottom( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect;
    
    pObjRect = ( AmObjRectPtr) AmFindDataLst( pScreenObj, _lstObj, 
                                          ( AmSearchFn) equalObjObjRect);

    if ( pObjRect != NULL) {
        _lstObj = AmDelPtrLst( pObjRect, _lstObj, ( AmSearchFn) AmEquals);
        _lstObj = AmInsertNodeLst( pObjRect, _lstObj);

        AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, pScreenObj);

    // Signal that something on the display has changed.

        AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    }
}

/***********************************************************************
 *
 * Function: Pop( pScreenObj)
 * 
 * Description:
 *   Move the object up one layer in the display list.  Note, that
 *   that means it needs to move towards the back of the internal
 *   list.
 *
 * Parameters:
 *   AmScreenObjClass *pScreenObj - object to move.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::Pop( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect, pObjRectTemp;
    AmListPtr    lstTemp, lstPrev;

/* Find the list data that contains the screen object. */
    
    pObjRect = ( AmObjRectPtr) AmFindDataLst( pScreenObj, _lstObj, 
                                          ( AmSearchFn) equalObjObjRect);

/* Find the list element that contains the data and also find its 
   predecessor. */

    lstPrev = NULL;

    for ( lstTemp = _lstObj;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pObjRectTemp = ( AmObjRectPtr) lstTemp->data;

        if ( pObjRect == pObjRectTemp) {
            break;
        }

        lstPrev = lstTemp;
    }

    if (( lstTemp != NULL) && ( lstTemp->next != NULL)) {

    /* Unhook the list element with the screen object and
       move it back one in the list. */
 
        if ( lstPrev != NULL) {
            lstPrev->next = lstTemp->next;
        } else {

        /* It was at the beginning of the list. */

            _lstObj = lstTemp->next;
        }

        lstPrev = lstTemp->next;
        lstTemp->next = lstPrev->next;
        lstPrev->next = lstTemp;

        AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, pScreenObj);

    // Signal that something on the display has changed.

        AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    }
}

/***********************************************************************
 *
 * Function: Push( pScreenObj)
 * 
 * Description:
 *   Move the object down one layer in the display list.  Note, that
 *   that means it needs to move towards the front of the internal
 *   list.
 *
 * Parameters:
 *   AmScreenObjClass *pScreenObj - object to move.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::Push( AmScreenObjClass *pScreenObj)
{
    AmObjRectPtr pObjRect, pObjRectTemp;
    AmListPtr    lstTemp, lstPrev, lstPrevPrev;

// Find the list data that contains the screen object.
    
    pObjRect = ( AmObjRectPtr) AmFindDataLst( pScreenObj, _lstObj, 
                                            ( AmSearchFn) equalObjObjRect);

// Find the list element that contains the data, find its predecessor,
//   and find its predecessor's predecessor.

    lstPrev = NULL;
    lstPrevPrev = NULL;

    for ( lstTemp = _lstObj;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pObjRectTemp = ( AmObjRectPtr) lstTemp->data;

        if ( pObjRect == pObjRectTemp) {
            break;
        }

        lstPrevPrev = lstPrev;
        lstPrev = lstTemp;
    }

    if (( lstPrev != NULL) && ( lstTemp != NULL)) {

    // Unhook the list element with the screen object and move it forward
    //   one in the list.

        if ( lstPrevPrev != NULL) {
            lstPrevPrev->next = lstTemp;
        } else {

        // It's now at the beginning of the list.

            _lstObj = lstTemp;
        }

        lstPrev->next = lstTemp->next;
        lstTemp->next = lstPrev;

    // Signal that the object has changed and needs a redraw.

        AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, pScreenObj);

    // Signal that something on the display has changed.

        AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

    }
}

/* Translate a screen coordinate to a virtual coordinate. */

int
AmDisplayManagerClass::VirtualCoord( int iScreen)
{
    return( _pDisplay->VirtualCoord( iScreen));
}


/* Translate a virtual coordinate to a screen coordinate. */

int
AmDisplayManagerClass::ScreenCoord( int iVirtual)
{
    return( _pDisplay->ScreenCoord( iVirtual));
}

/***********************************************************************
 *
 * Function: ChangeDPI()
 * 
 * Description:
 * 	Wrapper function to change the DPI of the screen, resize
 *  	the drawing area, and redraw the screen objects. 
 * Arguments:
 *	pScreenObjBehind - object to put in back.
 *	pScreenObjAhead - object to put it in back of.
 *
 ***********************************************************************/

void
AmDisplayManagerClass::ChangeDPI( int iDPI,
                                int iWidth,
                                int iHeight)
{
    AmListPtr      lstTemp;
    AmObjRectPtr	 pObjRect;

    /* Change the DPI of the display. */

    _pDisplay->SetDPI( iDPI);

    /*
     * If the display is double buffered, then create a new buffer
     *   of the apropriate size.
     */

    if ( _pWindowBuffer != NULL) {
        _pWindowBuffer->ResizeWindow( iWidth, iHeight);
    }

    /* Save the new dimensions. */

    _xOrigin = VirtualCoord( 0);
    _yOrigin = VirtualCoord( 0);
    _iWidth = iWidth;
    _iHeight = iHeight;
    
    /* Loop through the list of screen objects. */
    
    for ( lstTemp = _lstObj; lstTemp != NULL; lstTemp = lstTemp->next) {

        pObjRect = ( AmObjRectPtr) lstTemp->data;
	
    /*
     * If the window is being resized smaller, then we need to clear
     *   out the two rectangles on the border of the plot.
     */
	
        pObjRect->obj->SetWindowSize( _iWidth, _iHeight);
        pObjRect->obj->Update();
    }

    /*
     * Remove any events from the display manager's queue since we
     *   have to redraw everything anyways.
     */

    Wipe();

    /* Set the GC to reflect the window's new extents */

    _pGC->SetClipRectangle( _xOrigin, _yOrigin, _iWidth, _iHeight); 

    /* Redraw all the objects on the screen. */

    Draw();

// Signal that something on the display has changed.

    AmSendUserEvent( AM_PKG_ID, AM_DISPLAY_CHANGE, this);

}

