#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/* functions to support compression, differencing, and unioning vectors
   of expose rectangles for AmDisplayManagerClass::Flush() to minimize the
   number of refreshes */

#include <anvil/dispmgr.h>
#include <anvil/dispmgrP.h>
#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/rect.h>

/****************************************************************************
 *
 * Description:
 * 	clip a set of exposure rectangles by a list of inferior objects. 
 * Arguments:
 *	rects - vector of expose rectangles.
 *	nRects - length of the expose rectangle vector.
 *	objList - list of objects.
 *	lastObject - AmLast object in front of the clip rectangles 
 *	(may be NULL).
 *	clipRectvec - resulting vector of clip rectangles.
 *	nClip - number of resulting clip rectangles.
 *	maxClip - length of the resulting vector of clip rectangles.
 *	ignoreTransparent - don't generate clip/expose rectangles underneath
 *	transparent objects (defined by the object's getOpaque() flag.
 * Modifications:
 *	should return an error code if the rectangle buffer is overflowed.
 *
 ****************************************************************************/

void 
AmClipByObjectList( AmRectangleRec const	*rects,
		  int			nRects,
		  AmListPtr		objList,
		  AmScreenObjClass	*lastObject,
		  AmRectanglePtr		clipRectVec,
		  int			*nClip,
		  int			maxClip,
		  AnvilBoolean		ignoreTransparent)
{
    AmObjRectPtr		objRect;
    AmRectangleRec	leftRect, aboveRect, rightRect, belowRect;
    ulong_t		flags;
    AmListPtr		clipList, newList, prevNode, newPrevNode, ptr,
			deadNode;
    AmRectanglePtr	clipRect;
    int			ii;
    AmExposeEnum	exposeCode;

/* start off with the list of rectangles */

    clipList = NULL;
    for ( ii = 0; ii < nRects; ii++) {
	clipList = AmInsertNodeLst( AmCopyVec( &rects[ ii], sizeof( AmRectangleRec)), 
				  clipList);
    }

/* walk down the object list, and find all of the objects which clip this
   object.  Right now, we make no discrimination AmBetween opaque objects
   (which clip this object), and transparent ones (which have to be redrawn) */

    for ( ; objList != NULL; objList = objList->next) {
	objRect = ( AmObjRectPtr) objList->data;
	if ( objRect->obj == lastObject) {
	    break;
	}

/* if the object is not opaque, then it doesn't clip its inferior if we're
   ignoring transparent objects */

	if ( ignoreTransparent && !objRect->obj->getOpaque()) {
	    continue;
	}
	prevNode = NULL;
	ptr = clipList;
	while ( ptr != NULL) {
	    clipRect = ( AmRectanglePtr) ptr->data;
	    exposeCode = 
		AmCreateExposureRects( objRect->obj->GetExtentsRect(), clipRect,
				     &leftRect, &aboveRect, 
				     &rightRect, &belowRect, &flags);

/* if the above intersects the below rectangle, then replace the below
   rectangle with the list generated as the intersection */

	    if ( exposeCode == AmINTERSECT) {

/* create a list of all the rectangles that intersected */

		newList = NULL;
		if ( flags & AmRECT_ABOVE) {
		    newList = AmInsertNodeLst( AmCopyVec( &aboveRect, 
						      sizeof( AmRectangleRec)), 
					     newList);
		}
		if ( flags & AmRECT_BELOW) {
		    newList = AmInsertNodeLst( AmCopyVec( &belowRect, 
						      sizeof( AmRectangleRec)), 
					     newList);
		}
		if ( flags & AmRECT_LEFT) {
		    newList = AmInsertNodeLst( AmCopyVec( &leftRect, 
						      sizeof( AmRectangleRec)), 
					     newList);
		}
		if ( flags & AmRECT_RIGHT) {
		    newList = AmInsertNodeLst( AmCopyVec( &rightRect, 
						      sizeof( AmRectangleRec)), 
					     newList);
		}

/* save the node that gets replaced.  If we're at the head of the list, 
   simply replace the head node with the new list, otherwise, take the
   previous node and set it's next pointer to the new list prepended to
   the rest of the list. In both cases, the previous node gets set to the
   AmLast element of the added list */

		if ( newList != NULL) {
		    newPrevNode = AmLastLst( newList);
		    if ( prevNode == NULL) {
			clipList = AmAppendLst( newList, clipList->next);
		    } else {
			prevNode->next = AmAppendLst( newList, ptr->next);
		    }
		    free( ptr->data);
		    free( ptr);
		    prevNode = newPrevNode;
		    ptr = newPrevNode->next;
		} else {
		    prevNode = ptr;
		    ptr = ptr->next;
		}

/* if the new rectangle entirely covers the below rectangle, then remove the
   below rectangle from the list */

	    } else if ( exposeCode == AmCOVER) {
		deadNode = ptr;
		if ( prevNode == NULL) {
		    clipList = clipList->next;
		    ptr = clipList;
		} else {
		    prevNode->next = ptr->next;
		    ptr = prevNode->next;
		}
		free( deadNode->data);
		free( deadNode);

/* if there's no intsersection, then just step the two pointers */

	    } else {		/* exposeCode == AmNO_OVERLAP */
		prevNode = ptr;
		ptr = ptr->next; 
	    }
	}
    }

/* copy the list to the vector, then free the clipList */ 

    for ( *nClip = 0, ptr = clipList; 
	  ( ptr != NULL) && ( *nClip < maxClip); 
	  ptr = ptr->next) {
	clipRectVec[ (*nClip)++] = *( AmRectanglePtr) ptr->data;
    }
    AmFreeLst( clipList, ( AmItemFn) free);
}
