#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* when Flush() is called from the display manager, these member functions
   are responsible for calculating which portions of which objects 
   have to be refreshed */

#include <stdio.h>
#include <anvil/rect.h>
#include <anvil/screenobj.h>
#include <anvil/userevents.h>
#include <anvil/eventcodes.h>
#include <anvil/list.h>
#include <anvil/dispmgr.h>
#include <anvil/dispmgrP.h>
#include <anvil/macros.h>

static const int MAXCLIPRECTS = 2048;
static const int MAXEVENTS = 128;

/* comparator function for drawClips which define the object to refresh,
   the screen rectangle to redraw, and its order in the display list */

static int 
byDrawClip( AmDrawClipPtr	a,
            AmDrawClipPtr	b)
{
    if ( a->order == b->order) {
	return( a->clipRect.x - b->clipRect.x);
    } else {
	return( a->order - b->order);
    }
}

/* comparator function for deleteRects which define the object to delete,
   and its order in the display list */

static int 
byDeleteRectOrder( AmDeleteRectPtr	a,
		   AmDeleteRectPtr	b)
{
    return( a->order - b->order);
}

/************************************************************************** 
 *
 * Description:
 * 	unify a vector of expose rectangles (drawClips) for a screen object.
 * Arguments:
 *	drawClips - a vector of structures containing pointers to the
 *	screen objects, the expose rectangles, and their expose ordering.
 *	The reduced vector is returned in place.
 *	numDrawClips - number of draw clips in the vector.
 *
 **************************************************************************/

static void 
unifyDrawClips( AmDrawClipPtr	drawClips,
		int		*numDrawClips)
{
    int			ii, count;
    AmRectangleRec	newRect;

    count = 0;
    for ( ii = 1; ii < *numDrawClips; ii++) {
	if ( AmUnifyRects( &drawClips[ count].clipRect,
			 &drawClips[ ii].clipRect, &newRect)) {
	    drawClips[ count].clipRect = newRect;
	} else {
	    drawClips[ ++count].clipRect = drawClips[ ii].clipRect;
	}
    }
    *numDrawClips = count + 1;
}
		  
/************************************************************************** 
 *
 * Description:
 * 	take a vector of drawClips and consolidate them. The drawclips
 *	contain the expose rectangle, screen object, and expose ordering. 
 * Arguments:
 *	drawClips: a vector of structures containing pointers to the
 *	screen objects, the expose rectangles, and their expose ordering.
 *	The consolidated vector is returned in-place.
 *	numDrawClips: number of draw clips in argument and result vector.
 *
 **************************************************************************/

static void 
consolidateDrawClips( AmDrawClipPtr	drawClips,
		      int		*numDrawClips)
{
    int	rectEnd, rectStart, rectCount, numRects;

    if ( *numDrawClips == 0) {
	return;
    }

/* sort the list by the object's order in the display list */

    qsort( drawClips, *numDrawClips, sizeof( AmDrawClipRec), 
	   ( AmSearchFn) byDrawClip);

/* loop over the list of drawClips, blocking by object */

    rectStart = 0;
    rectCount = 0;
    do {
	rectEnd = rectStart;
	while (( rectEnd < *numDrawClips) &&
	       ( drawClips[ rectStart].order == drawClips[ rectEnd].order)) {
	    rectEnd++;
	}
	numRects = rectEnd - rectStart;

/* unify all the draw clips from the same object, and scrunch them back down */

	unifyDrawClips( &drawClips[ rectStart], &numRects);
	memcpy( &drawClips[ rectCount], &drawClips[ rectStart], 
		numRects*sizeof( AmDrawClipRec));
	rectCount += numRects;
	rectStart = rectEnd;
    } while ( rectEnd < *numDrawClips);
    *numDrawClips = rectCount;
}

/************************************************************************** 
 *
 * Description:
 * 	execute the vector of draw clips created by an update.
 * Arguments:
 *	pWindow: window to draw to.
 *	gc: graphics context to draw with.
 *	drawClips: a vector of structures containing pointers to the
 *	screen objects, the expose rectangles, and their expose ordering.
 *	numDrawClips: number of draw clips in vector.
 *
 **************************************************************************/

void
AmDisplayManagerClass::executeDrawClips( AmWindowClass	*pWindow,
                                       AmGCClass		*gc,
                                       AmDrawClipPtr	drawClips,
                                       int		numDrawClips)
{
    int			rectEnd, rectStart, ii;
    AmRectangleRec	clipRects[ MAXCLIPRECTS];
    AmScreenObjClass	*obj;

    if ( numDrawClips == 0) {
	return;
    }

/* loop over the list of drawClips, blocking by object */

    rectStart = 0;
    do {
	rectEnd = rectStart;
	while (( rectEnd < numDrawClips) &&
	       ( drawClips[ rectStart].order == drawClips[ rectEnd].order)) {
	    rectEnd++;
	}

/* copy the clip rectangles to a vector of clip rects */

	for ( ii = 0; ii < rectEnd - rectStart; ii++) {
	    clipRects[ ii] = drawClips[ rectStart + ii].clipRect;
	}

	gc->SetClipRectangles( clipRects, rectEnd - rectStart);
	obj = drawClips[ rectStart].obj;
	obj->Draw( pWindow, gc, clipRects, rectEnd - rectStart);

	rectStart = rectEnd;
    } while ( rectEnd < numDrawClips);
    gc->SetClipRectangle( _xOrigin, _yOrigin, _iWidth, _iHeight);
}

/* comarator function for source objects of events and event Id's */

static int 
byObjAddrAndEventId( AmEventRectPtr	a,
		     AmEventRectPtr	b)
{
    if ( a->obj != b->obj) {
        return( a->obj - b->obj);
    } else {
	return( a->eventId - b->eventId);
    }
}

/* function for comparing source objects of events */

static int 
byObjAddr( AmEventRectPtr	a,
	   AmEventRectPtr	b)
{
    return( a->obj - b->obj);
}

/* function for comparing order of events */

static int 
byObjOrder( AmEventRectPtr	a,
	    AmEventRectPtr	b)
{
    return( a->order - b->order);
}

/************************************************************************** 
 *
 * Description:
 * 	find the old rectangle of the object from the object/rectangle list 
 * Arguments:
 *	obj - object to search for.
 *	objList - list of objRects.
 * Returns:
 *	The old rectangle associated with the object (before SetRect()
 *	was called
 *
 **************************************************************************/

static AmRectanglePtr 
findRect( AmScreenObjClass	*obj,
	  AmListPtr		objList)
{
   AmObjRectPtr	objRect;

   for ( ; objList != NULL; objList = objList->next) {
	objRect = ( AmObjRectPtr) objList->data;
	if ( objRect->obj == obj) {
	    return( &objRect->oldRect);
	}
    }
    return( NULL);
}

/************************************************************************** 
 *
 * Description:
 * 	draw an object clipped by its opaque inferiors, and redraw any 
 *	transparent objects that overlay it.
 * Arguments:
 *	obj - object to redraw.
 *	objList - list of screen objects under the display manager's control.
 *	pWindow - window to draw to.
 *	drawClips - draw clips generated by drawClippedObject()
 *	numDrawClips - number of draw clips returned in vector.
 *
 **************************************************************************/

void
AmDisplayManagerClass::drawClippedObject( AmScreenObjClass *obj,
                                        AmListPtr        objList,
                                        AmWindowClass    *pWindow,
                                        AmDrawClipPtr    drawClips,
                                        int            *numDrawClips)
{
    int			nClip, ii, order;
    AmListPtr		objPtr;
    AmObjRectPtr		objRect;
    AmRectangleRec	clipRects[ MAXCLIPRECTS], isectRect;

/* if this object is transparent, clear out the rectangle behind it except
   for any opaque objects, then redraw all the objects underneath the object's
   rectangle. If the object's opaque, just walk the list up to the object */

    if ( !obj->getOpaque()) {
	AmClipByObjectList( obj->GetExtentsRect(), 1, objList,
			  NULL, clipRects, &nClip, MAXCLIPRECTS, TRUE);

        ClearArea( clipRects, nClip);

	for ( objPtr = objList, order = 0; 
	      objPtr != NULL; 
	      objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( objRect->obj == obj) {
		break;
	    }

/* if the objects intersect then find the inferiors which clip the object
   and add them to the draw clip list */

	    if ( AmIntersectRect( objRect->obj->GetExtentsRect(), 
				obj->GetExtentsRect(), &isectRect)) {
		AmClipByObjectList( &isectRect, 1, objPtr,
				  NULL, clipRects, &nClip, MAXCLIPRECTS, TRUE);
		for ( ii = 0; ii < nClip; ii++) {
		    drawClips[ *numDrawClips + ii].clipRect = clipRects[ ii];
		    drawClips[ *numDrawClips + ii].order = order;
		    drawClips[ *numDrawClips + ii].obj = objRect->obj;
		}
		*numDrawClips += nClip;
	    }
	}
    } else {
	for ( objPtr = objList, order = 0; 
	      objPtr != NULL; 
	      objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( objRect->obj == obj) {
		break;
	    }
	}
    }

/* find all of the intersecting rectangles of the opaque objects which clip 
   this object, set the clip regions and draw the object */

    if ( objPtr != NULL) {
	objPtr = objPtr->next;
    }
    AmClipByObjectList( obj->GetExtentsRect(), 1, objPtr, NULL,
		      clipRects, &nClip, MAXCLIPRECTS, TRUE);
    for ( ii = 0; ii < nClip; ii++) {
	drawClips[ *numDrawClips + ii].clipRect = clipRects[ ii];
	drawClips[ *numDrawClips + ii].order = order;
	drawClips[ *numDrawClips + ii].obj = obj;
    }
    order++;
    *numDrawClips += nClip;

/* take all of the inferior objects which are transparent, and redraw them
   where they intersected the updated object.  Ignore any transparent objects 
   when generating the clip list, because they'll overdraw */

    for ( ; objPtr != NULL; objPtr = objPtr->next, order++) {
	objRect = ( AmObjRectPtr) objPtr->data;
	if ( !objRect->obj->getOpaque()) {
	    if ( AmIntersectRect( objRect->obj->GetExtentsRect(), 
				obj->GetExtentsRect(), &isectRect)) {
		AmClipByObjectList( &isectRect, 1, objPtr,
				  NULL, clipRects, &nClip, MAXCLIPRECTS, TRUE);
		for ( ii = 0; ii < nClip; ii++) {
		    drawClips[ *numDrawClips + ii].clipRect = clipRects[ ii];
		    drawClips[ *numDrawClips + ii].order = order;
		    drawClips[ *numDrawClips + ii].obj = objRect->obj;
		}
		*numDrawClips += nClip;
	    }
	}
    }
}

/************************************************************************** 
 *
 * Description:
 * 	walk through the list of deleted objects and generate the 
 *	corresponding vector of draw clips to execute.
 * Arguments:
 *	deleteList - list of deleteRectPtrs, which contain the extents
 *	rectangle and order of the object being deleted. 
 *	objList - list of screen objects under the display manager's control.
 *	pWindow - window to draw to.
 *	drawClips - draw clips generated by doDeleteList()
 *	numDrawClips - number of draw clips returned in vector.
 * Modifications:
 *	that #if 0 around the orderCount makes me uncomfortable.
 *
 **************************************************************************/

void
AmDisplayManagerClass::doDeleteList( AmListPtr     deleteList,
                                   AmListPtr     objList,
                                   AmWindowClass *pWindow, 
                                   AmDrawClipPtr drawClips,
                                   int         *numDrawClips)
{
    int			orderCount, ii, nClip, order;
    AmDeleteRectPtr	deleteRect;
    AmRectangleRec	clipRects[ MAXCLIPRECTS], isectRect;
    AmObjRectPtr		objRect;
    AmListPtr		objPtr;
   
    orderCount = 0;
    deleteList = AmMergesortLst(  deleteList, ( AmSearchFn) byDeleteRectOrder);
    for ( ; deleteList != NULL; deleteList = deleteList->next) {
	deleteRect = ( AmDeleteRectPtr) deleteList->data;

/* find the clip rectangles that were obscured by the object's rectangle 
   that don't intersect any opaque object and clear them.  The transparent
   objects will be redrawn */
    
	AmClipByObjectList( &deleteRect->rect, 1, objList, NULL,
			  clipRects, &nClip, MAXCLIPRECTS, TRUE);

	ClearArea( clipRects, nClip);

/* redraw all of the objects which were uncovered by the delete */

	for ( objPtr = objList, order = 0; 
	      objPtr != NULL; objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( order == deleteRect->order - orderCount) {
		break;
	    }

/* find the object's intersection with the deleted object's rectangle */

	    if ( AmIntersectRect( objRect->obj->GetExtentsRect(),
				&deleteRect->rect, &isectRect)) {

/* clip the intersecting rectangles by their inferiors, but ignore any 
   transparent objects, because they have to be redrawn */
		  
		AmClipByObjectList( &isectRect, 1,
				  objPtr->next, NULL, clipRects, &nClip, 
				  MAXCLIPRECTS, TRUE);
		for ( ii = 0; ii < nClip; ii++) {
		    drawClips[ *numDrawClips + ii].clipRect = clipRects[ ii];
		    drawClips[ *numDrawClips + ii].order = order;
		    drawClips[ *numDrawClips + ii].obj = objRect->obj;
		}
		*numDrawClips += nClip;
	    }
	}

/* find all of the transparent object which need to be redrawn because of 
   the delete if there aren't any objects on top of this one, then bail out */

	for ( ; objPtr != NULL; objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( !objRect->obj->getOpaque()) {

/* Find the intersection of the object's new rectangle with the transparent
   object and add the draw request */

		if ( AmIntersectRect( objRect->obj->GetExtentsRect(),
				    &deleteRect->rect, &isectRect)) {

/* clip the intersecting rectangles by their inferiors, but ignore any 
   transparent objects, because they have to be redrawn */
	      
		    AmClipByObjectList( &isectRect, 1,
				      objPtr->next, NULL, clipRects, &nClip,
				      MAXCLIPRECTS, TRUE);
		    for ( ii = 0; ii < nClip; ii++) {
			drawClips[ *numDrawClips + ii].clipRect = 
			    clipRects[ ii];
			drawClips[ *numDrawClips + ii].order = order;
			drawClips[ *numDrawClips + ii].obj = objRect->obj;
		    }
		    *numDrawClips += nClip;
		}
	    }
	}

#if 0
/* each deleted object reduces the order by 1 */

	orderCount++;
#endif
    }
}

/************************************************************************** 
 *
 * Description:
 * 	do the exposure and clip management for an object when it resizes. 
 * Arguments:
 *	eventRect - contains the order of the object, eventId, whether
 *	the object is top leve, and the object itself.
 *	objList - display list.
 *	pWindow - window stuff is drawn to.
 *	drawClips - generated vector of drawClips.
 *	numDrawClips - number of drawClips generated.
 * Modifications:
 *	should be passed the length of the drawClips vector, and return
 *	an error code when the buffer is overflowed.  The pWindow parameter
 *	is already in the display manager class, so it's unneeded.
 *
 **************************************************************************/

void
AmDisplayManagerClass::doResize( AmEventRectPtr eventRect, 
                               AmListPtr      objList, 
                               AmWindowClass  *pWindow, 
                               AmDrawClipPtr  drawClips,
                               int          *numDrawClips)
{
    AmRectanglePtr	oldRectPtr;
    AmRectangleRec	clipRects[ MAXCLIPRECTS], diffRects[ 20],
			diffClipRects[ 20], isectRect;
    int			nDiffRects, nDiffClipRects, nClip, ii, order;
    AmObjRectPtr		objRect;
    AmListPtr		objPtr;
    AnvilBoolean		resizeFlag;

/* find the old rectangle. if the object is opaque, then we only redraw 
   the difference AmBetween the old and new rectangles.  If the object is
   transparent, then we redraw the union of the old and new rectangles, and
   we clear out any areas underneath the object which aren't covered by
   opaque objects */

    if (( oldRectPtr = findRect( eventRect->obj, objList)) == NULL) {
	return;			/* this shouldn't ever happen */
    }
    if ( eventRect->obj->getOpaque()) {
	resizeFlag = AmDifferenceRects( oldRectPtr, 
				        eventRect->obj->GetExtentsRect(),
				        diffRects, &nDiffRects);
    } else {
	resizeFlag = 
	    AmDifferenceIncludeRects( oldRectPtr, 
				      eventRect->obj->GetExtentsRect(),
				      diffRects, &nDiffRects);
    }
    if ( resizeFlag) {

/* find the clip rectangles for the difference AmBetween the object's old 
   rectangle and new rectangle that don't intersect any opaque object 
   and clear them.  The transparent objects will be redrawn */

	AmClipByObjectList( diffRects, nDiffRects, objList, NULL,
			  clipRects, &nClip, MAXCLIPRECTS, TRUE);

	ClearArea( clipRects, nClip);

/* redraw all the objects which were uncovered by the resize */

	for ( objPtr = objList, order = 0; 
	      objPtr != NULL; 
	      objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( objRect->obj == eventRect->obj) {
		break;
	    }

/* loop over the difference rectangles and find their intersection with each
   object's rectangle */
    
	    nDiffClipRects = 0;
	    for ( ii = 0; ii < nDiffRects; ii++) {
		if ( AmIntersectRect( objRect->obj->GetExtentsRect(),
				    &diffRects[ ii], &isectRect)) {
		    diffClipRects[ nDiffClipRects++] = isectRect;
		}
	    }
	    if ( nDiffClipRects > 0) {

/* clip the intersecting rectangles by their inferiors, but ignore any 
   transparent objects, because they have to be redrawn */
		  
		AmClipByObjectList( diffClipRects, nDiffClipRects,
				  objPtr->next, NULL, clipRects, &nClip, 
				  MAXCLIPRECTS, TRUE);
		for ( ii = 0; ii < nClip; ii++) {
		    drawClips[ *numDrawClips + ii].clipRect = clipRects[ ii];
		    drawClips[ *numDrawClips + ii].order = order;
		    drawClips[ *numDrawClips + ii].obj = objRect->obj;
		}
		*numDrawClips += nClip;
	    }
	}
    } else {

/* if no resize flag, then there was nothing exposed by the resize, so walk
   up to the current object */

	for ( objPtr = objList, order = 0; 
	      objPtr != NULL; 
	      objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( objRect->obj == eventRect->obj) {
		break;
	    }
	}
    }

/* if there aren't any objects on top of this one, then bail out */

    if (( objPtr != NULL) && ( objPtr->next != NULL)) {

/* draw all of the transparent objects which were on top of the resize, 
   either in the object's old rectangle, or the object's new rectangle */

	objPtr = objPtr->next;
	order++;
	for ( ; objPtr != NULL; objPtr = objPtr->next, order++) {
	    objRect = ( AmObjRectPtr) objPtr->data;
	    if ( !objRect->obj->getOpaque()) {

/* Find the intersection of the object's new rectangle with the transparent
   object and add the draw request */

		if ( AmIntersectRect( objRect->obj->GetExtentsRect(),
				    eventRect->obj->GetExtentsRect(),
				    &isectRect)) {

/* clip the intersecting rectangles by their inferiors, but ignore any 
   transparent objects, because they have to be redrawn */
	      
		    AmClipByObjectList( &isectRect, 1,
				      objPtr->next, NULL, clipRects, &nClip,
				      MAXCLIPRECTS, TRUE);
		    for ( ii = 0; ii < nClip; ii++) {
			drawClips[ *numDrawClips + ii].clipRect = 
			    clipRects[ ii];
			drawClips[ *numDrawClips + ii].order = order;
			drawClips[ *numDrawClips + ii].obj = objRect->obj;
		    }
		    *numDrawClips += nClip;
		}

/* find the intersection of the object's old rectangle with the transparent
   object and add the draw request */

		if ( AmIntersectRect( objRect->obj->GetExtentsRect(),
				    oldRectPtr, &isectRect)) {

/* clip the intersecting rectangles by their inferiors, but ignore any 
   transparent objects, because they have to be redrawn */
	      
		    AmClipByObjectList( &isectRect, 1,
				      objPtr->next, NULL, clipRects, &nClip,
				      MAXCLIPRECTS, TRUE);
		    for ( ii = 0; ii < nClip; ii++) {
			drawClips[ *numDrawClips + ii].clipRect = 
			    clipRects[ ii];
			drawClips[ *numDrawClips + ii].order = order;
			drawClips[ *numDrawClips + ii].obj = objRect->obj;
		    }
		    *numDrawClips += nClip;
		}
	    }
	}
    }

/* reset the old rectangle, now that the object has been resized/moved */

    *oldRectPtr = *( eventRect->obj->GetExtentsRect());
}

/************************************************************************** 
 *
 * Description:
 * 	update the list of children for an object, given the eventRect
 *	for the object update, and the object list, send the event to
 *	all the object's children.
 * Arguments:
 *	eventRect - contains event and object to update.
 *	objList - display manager's update list.
 *
 **************************************************************************/

static void 
doUpdateChildren( AmEventRectPtr 	eventRect, 
		  AmListPtr	objList)
{
    AmObjRectPtr		objRect;
    AmListPtr		objPtr;

/* for the event, find the list of children it affects, and update them, 
   then send a user event telling that they've updated, so they're added
   to the event list */

    for ( objPtr = objList; objPtr != NULL; objPtr = objPtr->next) {
	objRect = ( AmObjRectPtr) objPtr->data;
	if ( objRect->obj->getParent() == eventRect->obj) {
AN_DEBUG( 40, ( "updating child object %s\n", objRect->obj->getObjectName()));
	    objRect->obj->Update( &eventRect->eventId, 1);
	    AmSendUserEvent( AM_PKG_ID, eventRect->eventId, objRect->obj);
	}
    }
}
       
/************************************************************************** 
 *
 * Description:
 * 	flush all of the outstanding events.
 *	Part #1:  if the object is resized, or moved, all the objects under
 *	the object's old rectangle have to be updated.
 *	Part #2:  if the object is updated in any way, it has to be drawn
 *	clipped by its inferiors.
 *	Part #3:  if the object has any children, then they have to be updated
 *	and drawn clipped by their inferiors 
 * Modifications:
 *	I don't know why the AmSendUserEvent() for the AM_FLUSH event has
 *	beem dyked out.
 *
 **************************************************************************/

void 
AmDisplayManagerClass::Flush()
{
   AmWindowClass        *pWindow;
   AmEventRectPtr	eventRect, nextEventRect;
   AmListPtr		eventPtr, eventListCopy, objPtr;
   AmDrawClipRec	drawClips[ MAXCLIPRECTS];
   int			numDrawClips, eventIds[ MAXEVENTS], numEvents;
   AmScreenObjClass	*currObj;
   AmObjRectPtr		objRect=NULL;
   AnvilBoolean		found;

/* set the drawing window to the double buffer if available, or set it
   to the drawing window if it's valid, otherwise flag that the events
   need to be flushed on the next expose */

    if ( _pWindowBuffer != NULL) {
        pWindow = _pWindowBuffer;
    } else {
	if ( _pWindow->IsValid()) {
	    pWindow = _pWindow;
	} else {
	    _fNeedFlush = TRUE;
	    return;
	}
    }

/* take care of the deletions AmFirst */

   numDrawClips = 0;
   doDeleteList( _lstDelete, _lstObj, pWindow, drawClips, &numDrawClips);
   AmFreeLst( _lstDelete, ( AmItemFn) free);
   _lstDelete = NULL;

/* sort the list of objects by object address and unique them to remove 
   duplicates */

   _lstEvent = AmMergesortLst( _lstEvent, ( AmSearchFn) byObjAddrAndEventId);
   eventListCopy = AmMakeUniqueFromSortedLst( _lstEvent, 
					    ( AmSearchFn) byObjAddrAndEventId);

/* next, take care of all of the updates, which may call SetRects.  
   Consolidate the updates, so we only update each object once */

   eventPtr = eventListCopy;
   while ( eventPtr != NULL) {
	numEvents = 0;
	eventRect = ( AmEventRectPtr) eventPtr->data;
	if ( eventRect->topLevel) {
	    eventIds[ numEvents++] = eventRect->eventId;
	    currObj = eventRect->obj;
	    for ( eventPtr = eventPtr->next; 
		  eventPtr != NULL; 
		  eventPtr = eventPtr->next) {
		nextEventRect = ( AmEventRectPtr) eventPtr->data;
		if ( nextEventRect->obj != currObj) {
		    break;
		}
		eventIds[ numEvents++] = nextEventRect->eventId;
	    }
AN_DEBUG( 40, ("updating object at top level %s\n", currObj->getObjectName()));
	    currObj->Update( eventIds, numEvents);
	} else {
	    eventPtr = eventPtr->next;
	}

/* update any children whether the object is topLevel or not */

	doUpdateChildren( eventRect, _lstObj);
    } 
    AmJustFreeLst( eventListCopy);

/* grab the event list, sort the list of objects by object address, unique 
   them to remove duplicates by object, ignoring the event code, because
   we only have to redraw once, no matter how many events received.  Resort
   them according to the order of the object in the display list.
   NOTE:  The eventList may have been changed because of objects updating
   their children */

    _lstEvent = AmMergesortLst( _lstEvent, ( AmSearchFn) byObjAddrAndEventId);
    eventListCopy = AmMakeUniqueFromSortedLst( _lstEvent, ( AmSearchFn) byObjAddr);
    eventListCopy = AmMergesortLst( eventListCopy, ( AmSearchFn) byObjOrder);
    objPtr = _lstObj;
    for ( eventPtr = eventListCopy; 
	  eventPtr != NULL;
	  eventPtr = eventPtr->next) {

/* find the object in the object list that corresponds to the object in the
   event */

	eventRect = ( AmEventRectPtr) eventPtr->data;
	if ( eventRect->topLevel) {
	    found = FALSE;
	    while ( objPtr != NULL) {
		objRect = ( AmObjRectPtr) objPtr->data;
		if ( objRect->obj == eventRect->obj) {
		    found = TRUE;
		    break;
		}
		objPtr = objPtr->next;
	    } 

/* if the eventRect object is in the object list ( I certainly hope it is),
   find out if its rectangle has changed.  If so, then perform the resize.
   Don't bother with the resize if there's no window */

	    if ( found) {
		if ( !AmEqualRect( &objRect->oldRect, 
				 objRect->obj->GetExtentsRect())) {
		    if ( pWindow != NULL) {
			doResize( eventRect, _lstObj, pWindow, 
				  drawClips, &numDrawClips);
		    }
		}

/* redraw the object clipped by its opaque inferiors, and redraw any
   transparent object in front of it */

		if ( pWindow != NULL) {		/* don't draw if no window */
		    drawClippedObject( eventRect->obj, _lstObj, pWindow, 
				       drawClips, &numDrawClips);
		}
	    }
	}
    }

/* consolidate all the draw/clip requests and execute them */

    if ( pWindow != NULL) {			/* don't draw if no window */
	consolidateDrawClips( drawClips, &numDrawClips);
	executeDrawClips( pWindow, _pGC, drawClips, numDrawClips);

        if (( _pWindowBuffer != NULL) && ( _pWindow != NULL) &&
	     _pWindowBuffer->IsValid() && _pWindow->IsValid()) {

            _pDisplay->CopyArea( _pWindowBuffer, _pWindow, _pGC,
                                  _xOrigin, _yOrigin,
                                  _xOrigin, _yOrigin,
                                  _iWidth, _iHeight);
        }
    }

/* free the event list copy, and wipe the event list, and set the 
   "we need to flush" flag to false */

    AmJustFreeLst( eventListCopy);
    _fNeedFlush = FALSE;
    Wipe();

/* send a user event to anyone who is listening that we have flushed the 
   outstanding events */

    this->AmSendUserEvent( AM_PKG_ID, AM_FLUSH, this);
}

/************************************************************************** 
 *
 * Description:
 *	Wipe the outstanding list of events by freeing and NULLing the
 *	list of events.
 *
 **************************************************************************/

void AmDisplayManagerClass::Wipe()
{
    AmFreeLst( _lstEvent, ( AmItemFn) free);
    _lstEvent = NULL;
}

/************************************************************************** 
 *
 * Description:
 * 	add an event to the outstanding event list to be processed 
 *	by Flush() 
 * Arguments:
 *	obj - object to add the event for
 *	eventId - event to add.
 *	order - object's order in the display list.
 *	topLevel - whether the object is at the top level display list
 *	or a component of an overlay.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure.
 *
 **************************************************************************/

AmErrorCode 
AmDisplayManagerClass::AddEvent( AmScreenObjClass	*obj,
			       int		eventId,
			       int		order,
			       AnvilBoolean		topLevel)
{
    AmEventRectPtr	eventRect;

    if (( eventRect = ( AmEventRectPtr) 
			    malloc( sizeof( AmEventRectRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    eventRect->obj = obj;
    eventRect->eventId = eventId;
    eventRect->order = order;
    eventRect->topLevel = topLevel;
    _lstEvent = AmInsertNodeLst( eventRect, _lstEvent);
    return( AmOK);
}

/************************************************************************** 
 *
 * Description:
 * 	add a delete to the delete list to be processed by Flush() 
 * Arguments:
 *	obj - object to add the delete event for.
 *	order - its order in the display list.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure.
 *
 **************************************************************************/

AmErrorCode 
AmDisplayManagerClass::AddDelete( AmScreenObjClass *obj,
				int	       order)
{
    AmDeleteRectPtr deleteRect;

    if (( deleteRect = ( AmDeleteRectPtr) 
		malloc( sizeof( AmDeleteRectRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    deleteRect->rect = *obj->GetExtentsRect();
    deleteRect->order = order;
    _lstDelete = AmInsertNodeLst( deleteRect, _lstDelete);
    return( AmOK);
}
