#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions for handling readouts for image plots.  This will likely
   be moved to the plot directory, since it has no dependencies on the
   display manager. */

#include <malloc.h>
#include <fcntl.h>
#if !defined( _WIN32)
#include <unistd.h>
#endif
#include <string.h>
#include <anvil/overlay.h>
#include <anvil/imageplot.h>


/****************************************************************************
 *
 * Description:
 * 	return the z value at a coordinate, defined by dataX, dataY.
 * Arguments: 
 *	dataX, dataY - point in the imagePlot's coordinate system.
 *	dataZ - returned z value.
 * Returns:
 *	TRUE if the point is in the extents of the plot, FALSE otherwise.
 * Modifications:
 *	should be better support for thematic vs. continuous values.
 *	this isn't the cleanest implementation.
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotClass::getZvalue( float        dataX,
			   float        dataY,        
			   float        *dataZ)
{
    void	*val;
    int		index;

    if ( getZptr( dataX, dataY, &val)) {
	if ( colorIndex.fn != NULL) {
	    *dataZ = ( colorIndex.fn)( val, colorIndex.clientData);
	} else if ( colorIndex.table != NULL) {
	    index = ( colorIndex.getIndexFn)( val, &colorIndex);
	    *dataZ = colorIndex.table[ index];
	}
	return( TRUE);
    } else {
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the pointer to the value at a coordinate, defined by 
 *	dataX, dataY. The value is a type-free pointer, which may be
 *	converted using the imageplot's accessing or lookup functions.
 * Arguments: 
 *	dataX, dataY - point in the imagePlot's coordinate system.
 *	ptr - pointer to the resulting value.
 * Returns:
 *	TRUE if the point is in the extents of the plot, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotClass::getZptr( float       dataX,
			 float       dataY,
			 void	     **ptr)
{
    int			row, col;
    static uchar_t	buffer[ 512];

    if (!IsValid()) {
        AN_DEBUG( 90, ( "getZPtr: not valid.\n"));
	return( FALSE);
    }
    if ( !_fKeepData) {
	return( FALSE);
    }

/* NOTE:this should actually handle RGB lookup values and other stuff like 
   that, but I'll worry about that later */

    if (( dataX > coords.xmax) || ( dataX < coords.xmin) ||
	( dataY > coords.ymax) || ( dataY < coords.ymin)) {
	return( FALSE);
    }

/* since the matrix is UL xmin, ymin, the rows are reversed. We work off
   of subrows and subcols, because we might be working with a disk read 
   image or a pyramid image */

    if ( _fFlipRows == FALSE) {
        row = (int)(( _iSubRows - 1) - ( _iSubRows - 1) *
              ( dataY - _dSubYmin) / ( _dSubYmax - _dSubYmin));
    } else {
        row = (int)(( _iSubRows - 1) *
              ( dataY - _dSubYmin) / ( _dSubYmax - _dSubYmin));
    }

    if ( _fFlipCols == FALSE) {
        col = (int)(( _iSubCols - 1) *
              ( dataX - _dSubXmin) / ( _dSubXmax - _dSubXmin));
    } else {
        col = (int)(( _iSubCols - 1) - ( _iSubCols - 1) *
              ( dataX - _dSubXmin) / ( _dSubXmax - _dSubXmin));
    }

/* if the matrix is being read from a file, then we seek to get the value */

    if (( dataInfo.sourceType == AmDIRECT_SOURCE) ||
        ( dataInfo.sourceType == AmFILE_SOURCE) ||
        ( dataInfo.sourceType == AmMMAP_SOURCE) ||
        (( dataInfo.sourceType == AmPYRAMID_SOURCE) && _fKeepData) ||
	(( dataInfo.sourceType == AmREAD_SOURCE) && _fKeepData)) {

/* if the data is kept around, then index the matrix, else if it's read from
   a file, then open and seek the file */

	*ptr = &dataInfo.data[( row*_iSubCols + col)*dataInfo.size];
    } else if ( dataInfo.sourceType == AmREAD_SOURCE)  {
	if (!( _fnScanline)( dataInfo.fd, buffer, dataInfo.size, 1, 
			     dataInfo.dims[ 1], col, row, 
			     _pScanlineClientData)) {
	    return( FALSE);
	}
	*ptr = buffer;
    } else {
	return( FALSE);	
    }
    return( TRUE);
}

/* given an x,y position, return the appropriate string */
/****************************************************************************
 *
 * Description:
 * 	return the string for the value at a coordinate, defined by 
 *	dataX, dataY. 
 * Arguments: 
 *	dataX, dataY - point in the imagePlot's coordinate system.
 *	buffer - resulting string.
 * Returns:
 *	TRUE if the point is in the extents of the plot, FALSE otherwise.
 *
 ****************************************************************************/
	
AnvilBoolean 
AmImagePlotClass::getZstring( float       dataX,
			    float       dataY,
			    char	*buffer)
{
    void	*val;
    int		index;
    float	num;
    char	*unitStr;

    if ( getZptr( dataX, dataY, &val)) {

/* if there's a string table, then use it. */

	if ( stringIndex.table != NULL) {
	    index = ( stringIndex.getIndexFn)( val, &stringIndex);
	    if ( strings != NULL) {
		if (( index >= 0) && ( index < stringIndex.length)) {
		    strcpy( buffer, strings[ index]);
		} else {
		    buffer[ 0] = '\0';
		}
	    } else if ( index != -1) {
		AmGetUnitString( stringIndex.units, &unitStr);
		sprintf( buffer, "%d %s", stringIndex.table[ index], unitStr);
	    } else {
                buffer[ 0] = '\0';
            }
	} else if ( nameFn != NULL) {

/* if there's a name function, then use that */

	    strcpy( buffer, ( nameFn)( val, nameClientData));
	} else {

/* otherwise, it's a floating point value and print with variable and units */

	    if ( colorIndex.fn != NULL) {
		num = ( colorIndex.fn)( val, colorIndex.clientData);
                if ( ( colorIndex.minClip != TRUE || num >= colorIndex.min) &&
                     ( colorIndex.maxClip != TRUE || num <= colorIndex.max)) {
                    AmPrintValue( colorIndex.units, num, buffer);
                } else {
                    buffer[ 0] = '\0';
                }
	    } else if ( colorIndex.getIndexFn != NULL) {

/* umm, is there a string table for the colors? */

		index = ( colorIndex.getIndexFn)( val, &colorIndex);
		if ( colorIndex.strings != NULL && index != -1 &&
                     index < colorIndex.length) {
		    strcpy( buffer, colorIndex.strings[ index]);
		} else if ( index != -1 && index < colorIndex.length) {

/* just show the index and units */

		    AmGetUnitString( colorIndex.units, &unitStr);
		    sprintf( buffer, "%d %s", 
			     colorIndex.table[ index], unitStr);
		} else {
                    buffer[ 0] = '\0';
                }
	    } else {

/* umm...how are we rendering an image if we get here...truecolor only.
   we need to add trueColorFn test for the pixel function, but trueColorFn
   is statically declared in image_fns.C */

		buffer[ 0] = '\0';
	    }
	}
	return( TRUE);
    } else {
	return( FALSE);
    }
}
