#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* display-independent base class implementation for magic windows,
   magic outlines, magic images, and magic plots. */

#include <stdarg.h>
#include <stdio.h>
#include <anvil/plot.h>
#include <anvil/resource.h>
#include <anvil/geometry.h>
#include <anvil/macros.h>
#include <anvil/list.h>
#include <anvil/copyobj.h>
#include <anvil/magicmgr.h>
#include <anvil/outline.h>
#include <anvil/magic.h>

static const int MAGIC_WIN_WIDTH = 128,
		 MAGIC_WIN_HEIGHT = 128,
		 MAGIC_WIN_RESIZE_TAB = 16;

/************************************************************************
 * 									*
 * Base Magic Class definition 						*
 *									*
 ************************************************************************/

/************************************************************************
 *
 * Description:
 * 	default destructor calls the client data destroy function and
 *  	removes the event handlers 
 *
 ************************************************************************/

AmMagicClass::~AmMagicClass()
{
    magicManager->RemoveMagic( this);
}

/************************************************************************
 *
 * Description:
 *	Default constructor.
 *
 ************************************************************************/

AmMagicClass::AmMagicClass()
{
}

/************************************************************************
 *
 * Description:
 *	is a point inside the magic window's resize rectangle?
 * Arguments:
 *	x,y - point to test.
 * Returns:
 *	TRUE if the point is inside the resize rectangle, FALSE otherwise.
 *
 ************************************************************************/

AnvilBoolean 
AmMagicClass::inResize( int x,
		        int y)
{
    return( AmPtInRect( &rect, x, y) &&
	    ( x > rect.x + rect.width - resizeTabSize) &&
	    ( y > rect.y + rect.height - resizeTabSize));
}

/************************************************************************
 *
 * Description:
 *	Abstract base magic class constructor, sets the magic manager,
 *	source plot and call on drag flags, sets the rectangle and 
 *	resize tab size, and adds the magic window to the magic manager's
 *	display list.
 * Arguments:
 *	magicManagerArg - display manager which controls the magic windows 
 *	sourcePlotArg - source plot for data for the magic window 
 *	callOnDragArg - call SetRectWrapper() every time the mouse is 
 *	moved, or just on release.
 *
 ************************************************************************/

AmMagicClass::AmMagicClass( AmMagicManagerClass *magicManagerArg,
			    AmDataGraphicClass  *sourcePlotArg,
			    AnvilBoolean             callOnDragArg)
{
    int				x, y, width, height;
    AmRectangleRec const	*srcRect = sourcePlotArg->GetRect();

    sourcePlot = sourcePlotArg;
    callOnDrag = callOnDragArg;
    magicManager = magicManagerArg;

/* figure out where the pointer is, and start the window at a reasonable 
   location */

    width = AmGetDefault( "magicPlot.width", MAGIC_WIN_WIDTH);
    height = AmGetDefault( "magicPlot.height", MAGIC_WIN_HEIGHT);
    x = srcRect->x + srcRect->width/2;
    y = srcRect->y + srcRect->height/2;

/* create the magic window */

    display = magicManagerArg->GetDisplay();
    window = magicManagerArg->GetWindow();
    gc = magicManagerArg->GetGC();
    rect.width = width;
    rect.height = height;
    rect.x = x - width/2;
    rect.y = y - height/2;

/* clip to the bounds of the source plot */

    if ( rect.x + rect.width > srcRect->x + srcRect->width) {
	rect.x = srcRect->x;
	rect.width = srcRect->width;
    } 
    if ( rect.y + rect.height > srcRect->y + srcRect->height) {
	rect.y = srcRect->y;
	rect.height = srcRect->height;
    } 
    resizeTabSize = AmGetDefault( "magic.resizeTabSize", MAGIC_WIN_RESIZE_TAB);
}

/************************************************************************
 *
 * Description:
 * 	determine the coordinates that the magic window represents in 
 *	the source plot's coordinate system
 * Arguments:
 *	pXmin, pYmin, pXmax, pYmax: coordinates in the source plot's
 *	coordinate system.
 *
 ************************************************************************/

void 
AmMagicClass::GetSourceCoords( float	*pXmin,
			       float	*pYmin,
			       float	*pXmax,
			       float	*pYmax)
{
    sourcePlot->GetDataPt( rect.x, rect.y + rect.height, pXmin, pYmin);
    sourcePlot->GetDataPt( rect.x + rect.width, rect.y, pXmax, pYmax);
}

/************************************************************************
 *
 * Description:
 * 	update function for the magic window class 
 * Returns: 
 *	AmOK
 *
 ************************************************************************/

AmErrorCode 
AmMagicWindowClass::Update()
{
#if 0
    DrawWrapper( &rect);
#endif
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 * 	rectangle set wrapper function for magic window.  We use a 
 *	wrapper function so that subclassed magic windows only have to
 *	implement the SetRect function, which hopefully ignores some
 *	of the magic window details.   
 * Arguments:
 *	rectArg - rectangle that the magic window is set to.
 *
 ************************************************************************/

void 
AmMagicWindowClass::SetRectWrapper( AmRectangleRec const *rectArg)
{

/* move the window if the position has changed */

    if (( rectArg->x != rect.x) || ( rectArg->y != rect.y)) {
  	window->MoveWindow( rectArg->x, rectArg->y);
    }

/* resize the window and create a new image if the size has changed */

    if (( rectArg->width != rect.width) || ( rectArg->height != rect.height)) {
  	window->ResizeWindow( rectArg->width, rectArg->height);
    }

/* set the rectangle ... the draw will be handled later... */

    SetRect( rectArg);
#if 0
    DrawWrapper( rectArg);
#endif
}


/************************************************************************
 *
 * Description:
 * 	default magic window class draw function does nothing
 * Arguments:
 *	exposeRect - rectangle to redraw.
 *
 ************************************************************************/

/* ARGSUSED */

void 
AmMagicWindowClass::Draw( AmRectangleRec const 	*exposeRect)
{
}

/************************************************************************
 *
 * Description:
 * 	Draw wrapper function for magic window.  We use a 
 *	wrapper function so that subclassed magic windows only have to
 *	implement the Draw() function, which hopefully ignores some
 *	of the magic window details.   The magic window draw wrapper
 *	clips the expose rectangle, copies the background from the
 *	underlying window, calls the Draw() function, then draws the
 *	resize rectangle.  This function may be called many times in
 *	sequence if there are multiple expose rectangles.
 * Arguments:
 *	exposeRect: exposed rectangle in virtual coordinates from the
 *	display manager.
 *
 ************************************************************************/

void 
AmMagicWindowClass::DrawWrapper( AmRectangleRec const	*exposeRect)
{
    AmRectangleRec	drawRect = *exposeRect, offsetRect = rect;
    AmWindowClass	*pManagerWin;

    offsetRect.x -= rect.x;
    offsetRect.y -= rect.y;
    drawRect.x -= rect.x;
    drawRect.y -= rect.y;
    AmClipRect( &offsetRect, &drawRect);
    if ( copySource) {
AN_DEBUG( 30, ( "copying area %d %d %d %d %d %d\n", 
		exposeRect->x, exposeRect->y,
		drawRect.x, drawRect.y, drawRect.width, drawRect.height));
	if ( magicManager->GetBuffer() != NULL) {
	    pManagerWin = magicManager->GetBuffer();
	} else {
	    pManagerWin = magicManager->GetWindow();
	}
	display->CopyArea( pManagerWin, window, gc, 
			   exposeRect->x, exposeRect->y,
			   drawRect.x, drawRect.y,
			   drawRect.width, drawRect.height);
    }
    Draw( exposeRect);

/* draw the resize tab */

    display->DrawRectangle( window, gc,
                            offsetRect.width - resizeTabSize, 
			    offsetRect.height - resizeTabSize,
                            resizeTabSize, resizeTabSize);
}

/************************************************************************
 *
 * Description:
 * 	destructor for magic windows.  Delete the private window and
 *	graphics context.
 *
 ************************************************************************/

AmMagicWindowClass::~AmMagicWindowClass()
{
    delete window;
    delete gc;
}

/************************************************************************
 *
 * Description:
 * 	constructor for magic windows.  This calls the base magic window
 *	constructor, and creates a private window and graphics context.
 * 	The copyFlag argument is used to control whether the background
 *	is copied to the window as it is dragged.
 * Arguments:
 *	magicManagerArg - display manager the magic window is attached to.
 *	sourcePlotArg - source plot to extract data from.
 *	callOnDragArg - update on mouse motion or mouse release only.
 *	drawSourceFlagArg - if true, the background is copied from the 
 *	underlying window, if not, the window is left blank as it is
 *	dragged (but should be cleared from the AmLast update).
 *
 ************************************************************************/

AmMagicWindowClass::AmMagicWindowClass( AmMagicManagerClass *magicManagerArg,
					AmDataGraphicClass  *sourcePlotArg,
					AnvilBoolean             callOnDragArg,
					AnvilBoolean	      	    drawSourceFlagArg) :
			  AmMagicClass( magicManagerArg,
					sourcePlotArg,
					callOnDragArg)
{

    copySource = drawSourceFlagArg;
    window = display->CreateWin( magicManager->GetWindow(),
    				 rect.x, rect.y, rect.height, rect.width);
    gc = window->CreateGC();
    magicManager->AddMagic( this);
    Update();
    DrawWrapper( &rect);
}

/************************************************************************
 *
 * Description:
 * 	constructor for magic plots.  Magic Plots are Magic Windows where
 *	a private plot is drawn.  Typically, this plot is a "trace" of
 *	the source plot where some parameter has been changed, such as
 *	the magnification factor, or transparency.  A "trace" copies the
 *	attributes of the source plot, but not the data, saving storage.
 *	The plot is re-initialized whenever any of the fields in the
 *	source plot is changed by calling the Init(() member function.
 * Arguments:
 *	magicManagerArg - display manager the magic window is attached to.
 *	sourcePlotArg - source plot to extract data from.
 *	callOnDragArg - update on mouse motion or mouse release only.
 *
 ************************************************************************/

AmMagicPlotClass::AmMagicPlotClass( AmMagicManagerClass *magicManagerArg,
				    AmDataGraphicClass  *sourcePlotArg,
				    AnvilBoolean             callOnDragArg) :
			    AmMagicWindowClass( magicManagerArg,
					        sourcePlotArg,
					        callOnDragArg, 
					        FALSE)
{
    plot = NULL;
}

/************************************************************************
 *
 * Description:
 * 	default magic plot initialize/update function calls AmTraceObject
 * Arguments:
 *	sourcePlot - underlying plot to "magic"
 * Returns:
 *	copied plot to update in the magic window
 * Modifications:
 *	eventually AmTraceObject will be replaced by AmCopyObject()
 *
 ************************************************************************/

AmDataGraphicClass *
AmMagicPlotClass::Init( AmDataGraphicClass        *sourcePlot)
{
    return(( AmDataGraphicClass *) AmTraceObject( GetWindow(), sourcePlot, NULL));
}

/************************************************************************
 *
 * Description:
 * 	magic plot class update re-initializes the plot and calls resize
 * Returns:
 *	AmOK
 *
 ************************************************************************/

AmErrorCode 
AmMagicPlotClass::Update()
{
    if ( plot != NULL) {
	delete plot;
    }
    plot = Init( sourcePlot);

/* SetRect calls resize */

    SetRectWrapper( &rect);
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 * 	default magic plot draw function.  Take the expose rectangle, 
 *	clip it, and translate its coordinates into the magic window's 
 *	coordinate space 
 * Arguments:
 *	exposeRect - expose rectangle to redraw.
 * 
 ************************************************************************/

void 
AmMagicPlotClass::Draw( AmRectangleRec const *exposeRect)
{
    AmRectangleRec drawRect = *exposeRect, offsetRect = rect;

    if ( plot == NULL) {
	return;
    }
    offsetRect.x = 0;
    offsetRect.y = 0;
    drawRect.x = 0;
    drawRect.y = 0;
    AmClipRect( &offsetRect, &drawRect);
    plot->Draw( window, gc, &drawRect, 1);
}

/************************************************************************
 *
 * Description:
 * 	SetRect for a magic plot resets the rectangle for the plot 
 *  	magic plot class resize gets the coordinates for the rectangle, 
 *  	calls the passed resize function and draws the resize tab rectangle 
 * Arguments:
 *	rectArg - rectangle to reset the magic plot to.
 * 
 ************************************************************************/

void 
AmMagicPlotClass::SetRectWrapper( AmRectangleRec const	*rectArg)
{
    plot->SetRect( 0, 0, rectArg->width, rectArg->height);
    AmMagicWindowClass::SetRectWrapper( rectArg);
}

/************************************************************************
 *
 * Description:
 * 	default magic plot SetRect() function gets the source coordinates
 *	for the magic window, and updates the magic plot's coordinates.
 * Arguments:
 *	rectArg - new rectangle for the magic window.
 * 
 ************************************************************************/

void 
AmMagicPlotClass::SetRect( AmRectangleRec const *rectArg)
{
    float		xmin, ymin, xmax, ymax;

/* a handly safety check, just in case the caller hasn't initialized the
   plot.. */

    rect = *rectArg;
    if ( plot == NULL) {
	Update();
    }
    GetSourceCoords( &xmin, &ymin, &xmax, &ymax);
    plot->SetCoords( xmin, ymin, xmax, ymax);
    plot->Update();
}

/************************************************************************
 *
 * Description:
 * 	destructor for the magic plot calls the magic destructor and 
 *	deletes the plot. The window, display, and gc are destroyed by 
 *	magic window, and the callbacks are removed by the Magic class 
 *	destructor.  Remove the callback which updates the magic plot 
 *	whenever the parent is changed 
 * 
 ************************************************************************/

AmMagicPlotClass::~AmMagicPlotClass()
{
    delete plot;
}

/************************************************************************
 *
 * Description:
 * 	constructor for a magic image contained in a window 
 * Arguments:
 *	magicManagerArg - display manger the magic window is displayed on.
 *	sourcePlotArg - source plot data in the magic window is derived from.
 *	callOnDrag - call when mouse is moved, or call on mouse release.
 * 
 ************************************************************************/

AmMagicImageClass::AmMagicImageClass( AmMagicManagerClass *magicManagerArg,
				      AmDataGraphicClass  *sourcePlotArg,
				      AnvilBoolean             callOnDragArg) :
			    AmMagicWindowClass( magicManagerArg,
					        sourcePlotArg,
					        callOnDragArg)
{
    image = display->CreateImage( rect.width, rect.height,
				  display->GetDepth(), NULL);
}

/************************************************************************
 *
 * Description:
 * 	magic image default update, basically call SetRectWrapper()
 *	on the current rectangle.
 * Returns:
 *	AmOK
 *
 ************************************************************************/

AmErrorCode 
AmMagicImageClass::Update()
{
    SetRectWrapper( GetRect());
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 * 	draw the image in a magic image.  take the expose rectangle and 
 *  	clip it to the magic window, call the draw member function to 
 *  	presumably fill the image, draw the image, and draw the resize 
 *	tab. 
 * Arguments:
 * 	exposeRect - rectangle to redraw.
 *
 ************************************************************************/

void 
AmMagicImageClass::DrawWrapper( AmRectangleRec const	*exposeRect)
{
    AmRectangleRec drawRect = *exposeRect,
		 offsetRect = rect;

    display->GetImage( magicManager->GetWindow(), image, rect.x, rect.y);
    offsetRect.x -= rect.x;
    offsetRect.y -= rect.y;
    drawRect.x -= rect.x;
    drawRect.y -= rect.y;
    AmClipRect( &offsetRect, &drawRect);
    Draw( &drawRect);
AN_DEBUG( 30, ( "drawing image %d %d %d %d\n", 
		drawRect.x, drawRect.y, rect.width, rect.height));
    display->DrawImage( window, gc, drawRect.x, drawRect.y, 
    				    drawRect.x, drawRect.y, 
				    drawRect.width, drawRect.height, image);

/* draw the resize tab */

    display->DrawRectangle( window, gc,
                            rect.width - resizeTabSize, 
			    rect.height - resizeTabSize,
                            resizeTabSize, resizeTabSize);
}

/************************************************************************
 *
 * Description:
 * 	default draw function for magic images does nothing. 
 * Arguments:
 *	drawArg - rectangle in the magic window to redraw.
 *
 ************************************************************************/

void AmMagicImageClass::Draw( AmRectangleRec const	*drawArg)
{
}

/************************************************************************
 *
 * Description:
 * 	SetRectWrapper for a magic image resizes the window and the image 
 * Arguments:
 *	rectArg - rectangle to reset the magic window to
 *
 ************************************************************************/

void 
AmMagicImageClass::SetRectWrapper( AmRectangleRec const	*rectArg)
{


/* create a new image if the size has changed */

    if (( rectArg->width != rect.width) || ( rectArg->height != rect.height)) {
	delete image;
	image = display->CreateImage( rectArg->width, rectArg->height, 
				      display->GetDepth(), NULL);
    }
#if 0
    display->GetImage( magicManager->GetWindow(), image, 
		       rectArg->x, rectArg->y);
#endif
    AmMagicWindowClass::SetRectWrapper( rectArg);
}

/************************************************************************
 *
 * Description:
 * 	destroys the image when blown away, The callbacks are removed by
 *  	AmMagicClass's destructor 
 *
 ************************************************************************/

AmMagicImageClass::~AmMagicImageClass()
{
    delete image;
}

/************************************************************************
 *
 * Description:
 * 	initialization for a magic outline class.  Creates a magic, 
 *	saves the update function, and creates a graphic context for 
 *	drawing the cursor 
 * Arguments:
 *	magicManagerArg - display manager the magic window is attached to.
 *	sourcePlotArg - source plot to extract data from.
 *	callOnDragArg - update on mouse motion or mouse release only.
 *
 ************************************************************************/

AmMagicOutlineClass::AmMagicOutlineClass( AmMagicManagerClass     *magicManagerArg,
					  AmDataGraphicClass      *sourcePlotArg,
					  AnvilBoolean                 callOnDragArg) :
			    AmMagicClass( magicManagerArg,
					  sourcePlotArg,
					  callOnDragArg)
{ 
    lastDrawn = FALSE;

    outline = new AmWindowOutlineClass( magicManagerArg,
				        rect.x, rect.y, rect.width, rect.height, 
				        resizeTabSize);
    magicManager->AddOutline( outline);
    SetRectWrapper( &rect);
    magicManager->AddMagic( this);
}		    

/************************************************************************
 *
 * Description:
 * 	magic outline default update, basically call SetRectWrapper()
 *	on the current rectangle.
 * Returns:
 *	AmOK
 *
 ************************************************************************/

AmErrorCode 
AmMagicOutlineClass::Update()
{

/* don't undraw, because the entire display just got wiped out ... */

    lastDrawn = FALSE;
    SetRectWrapper( &rect);
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 * 	DrawWrapper base implementation in AmMagicOutlineClass calls 
 *	clipRect for the expose rectangle, then Draw().
 * Arguments:
 *	exposeRect - expose rectangle to redraw.
 *
 ************************************************************************/

void 
AmMagicOutlineClass::DrawWrapper( AmRectangleRec const	*exposeRect)
{
    AmRectangleRec drawRect = *exposeRect;

    AmClipRect( &rect, &drawRect);
    Draw( exposeRect);
}

/************************************************************************
 *
 * Description:
 * 	default Magic outline undraw calls the draw function.
 * Arguments:
 *	exposeRect - expose rectangle to redraw.
 *
 ************************************************************************/

void 
AmMagicOutlineClass::Undraw( AmRectangleRec const *exposeRect)
{
    AmRectangleRec drawRect = *exposeRect;

    AmClipRect( &rect, &drawRect);
    Draw( exposeRect);
}

/************************************************************************
 *
 * Description:
 * 	default magic outline draw does nothing 
 * Arguments:
 *	exposeRect - expose rectangle to redraw.
 *
 ************************************************************************/

/* ARGSUSED */

void 
AmMagicOutlineClass::Draw( AmRectangleRec const *exposeRect)
{
}

/************************************************************************
 *
 * Description:
 * 	Set the magic outline's rectangle wrapper function to call the
 *	derived SetRect() function.
 * Arguments:
 *	rectArg - new rectangle to set the outline to.
 *
 ************************************************************************/

void 
AmMagicOutlineClass::SetRectWrapper( AmRectangleRec const	*rectArg)
{
    if ( lastDrawn) {
	Undraw( &rect);
    } 
    lastDrawn = TRUE;

/* set the rectangle, update the outline, and draw the new cursor */

    SetRect( rectArg);
#if 0
    DrawWrapper( rectArg);
#endif
    outline->SetRectangle( rect.x, rect.y, rect.width, rect.height);
}

/************************************************************************
 *
 * Description:
 * 	default magic outline set rect function just sets the rectangle 
 * Arguments:
 *	rectArg - new rectangle to set the outline to.
 *
 ************************************************************************/

void
AmMagicOutlineClass::SetRect( AmRectangleRec const	*rectArg)
{
    rect = *rectArg;
}

/************************************************************************
 *
 * Description:
 * 	destructor: close the magic outline, remove the outline handler 
 *	from the display manager.
 *
 ************************************************************************/

AmMagicOutlineClass::~AmMagicOutlineClass()
{

/* remove the outline and the event handler */

    magicManager->RemoveOutline( outline);
}

/************************************************************************
 * 
 * Description:
 * 	drag/resize a magic window or its outline from one mouse position 
 *	to the next.  This function handles the bounds checking and 
 *	resize rectangle testing.
 * Arguments:
 * 	magic - magic window to be moved.
 *	outline - associated outline (if any).
 *	lastX, lastY - AmLast mouse position.
 *	newX, newY - current mouse position.
 * Returns:
 *	TRUE if window size/position was changed, FALSE otherwise 
 *
 ************************************************************************/

AnvilBoolean
AmMoveMagic( AmMagicClass	 	*magic,
	     AmWindowOutlineClass 	*outline,
	     int		 	lastX,
	     int		 	lastY,
	     int		 	newX,
	     int			newY)
{
    AmRectangleRec const	*boundRect = magic->GetSourcePlot()->GetRect(),
    				*rect = magic->GetRect();
    AmRectangleRec		newRect;
    AnvilBoolean			inResizeFlag;
    int				tabSize;

/* if the AmLast mouse position was in the resize rectangle, then drag-resize
   the window */

    tabSize = magic->GetResizeTabSize();
    inResizeFlag = AmBetween( lastX, rect->x + rect->width - tabSize,
			           rect->x + rect->width) &&
		   AmBetween( lastY, rect->y + rect->height - tabSize,
				   rect->y + rect->height);
    newRect = *rect;
    if ( inResizeFlag) {
    	newRect.width += newX - lastX;
	newRect.height += newY - lastY;

/* make sure that the user didn't try to resize the window too small... */

	if (( newRect.width <= tabSize) || ( newRect.height <= tabSize)) {
	    return( FALSE);
	} 

/* or beyond the extents of the plot */

	if ( newRect.x + newRect.width > boundRect->x + boundRect->width) {
	    newRect.width = boundRect->x + boundRect->width - newRect.x;
	}
	if ( newRect.y + newRect.height > boundRect->y + boundRect->height) {
	    newRect.height = boundRect->y + boundRect->height - newRect.y;
	}
    } else {
    	newRect.x += newX - lastX;
	newRect.y += newY - lastY;

/* if the user dragged the magic window, make sure it's not beyond the
   extents of the source plot */

	if ( newRect.x < boundRect->x) {
	    newRect.x = boundRect->x;
	}
	if ( newRect.y < boundRect->y) {
	    newRect.y = boundRect->y;
	}
	if ( newRect.x + newRect.width > boundRect->x + boundRect->width) {
	    newRect.x = boundRect->x + boundRect->width - newRect.width;
	}
	if ( newRect.y + newRect.height > boundRect->y + boundRect->height) {
	    newRect.y = boundRect->y + boundRect->height - newRect.height;
	}
    }

/* if the magic window is updated on drags, then update it, otherwise,
   move the outline */

    if ( magic->GetCallOnDrag()) {
    	magic->SetRectWrapper( &newRect);
    } else {
    	outline->SetRectangle( newRect.x, newRect.y, 
			       newRect.width, newRect.height);
    }
    return( TRUE);
}
