#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdarg.h>
#include <stdio.h>
#include <anvil/plot.h>
#include <anvil/geometry.h>
#include <anvil/macros.h>
#include <anvil/list.h>
#include <anvil/magicmgr.h>
#include <anvil/magic.h>

/*************************************************************************
 *
 * Description:
 *	find the magic window that's closest to the passed virtual 
 *	coordinate 
 * Arguments:
 *	xpos, ypos: coordinate in virtual coordinates 
 *	inResize: flag for whether position was in rectangle's
 *	resize box
 * Returns:
 *	AmFirst magic window which contains the coordinate, or NULL
 *	if no magics contain it
 *
 *************************************************************************/

AmMagicClass *
AmMagicManagerClass::findMagic( int	xpos,
			      int	ypos,
			      AnvilBoolean	*inResize)
{
    AmMagicClass		*magic;
    AmListPtr		ptr;
    AmRectangleRec const	*rect;
    int			tabSize;

    for ( ptr = _lstMagic; ptr != NULL; ptr = ptr->next) {
	magic = ( AmMagicClass *) ptr->data;
	rect = magic->GetRect();
	if ( AmPtInRect( rect, xpos, ypos)) {
	    tabSize = magic->GetResizeTabSize();
	    *inResize = AmBetween( xpos, rect->x + rect->width - tabSize,
				       rect->x + rect->width) &&
			AmBetween( ypos, rect->y + rect->height - tabSize,
				       rect->y + rect->height);
	    return( magic);
	}
    }
    return( NULL);
}

/*************************************************************************
 * 
 * Description:
 *	handler to delete the magic window if the source plot is removed.
 *	The object has to be tested against the magic window's source plot,
 *	because another object may have sent the event.
 * Arguments:
 *	eventId - launching user event id.
 *	magic - magic window passed in the AmAddCallback call.
 *	obj - object passed as call data by the sendEvent call.
 *
 *************************************************************************/

/* ARGSUSED */

static void 
destroyMagicCallback( int		eventId,
		      AmMagicClass	*magic,
		      AmScreenObjClass	*obj)
{
    if ( obj == ( AmScreenObjClass *) magic->GetSourcePlot()) {
    	delete magic;
    }
}

/*************************************************************************
 * 
 * Description:
 * 	handler to resize the magic window if the source plot is resized 
 *	The object has to be tested against the magic window's source plot,
 *	because another object may have sent the event.
 * Arguments:
 *	eventId - launching user event id.
 *	magic - magic window passed in the AmAddCallback call.
 *	obj - object passed as call data by the sendEvent call.
 *
 *************************************************************************/

/* ARGSUSED */

static void 
resizeMagicCallback( int		eventId,
		     AmMagicClass	*magic,
		     AmScreenObjClass	*obj)
{
    if ( obj == ( AmScreenObjClass *) magic->GetSourcePlot()) {
    	magic->SetRectWrapper( magic->GetRect());
    }
}

/*************************************************************************
 * 
 * Description:
 *	handler to redraw the magic window if the source display is flushed 
 * Arguments:
 *	eventId - launching user event id.
 *	magic - magic window passed in the AmAddCallback call.
 *	displayManager - display manager that was flushed.
 *
 *************************************************************************/

/* ARGSUSED */

static void 
redrawMagicCallback( int			eventId,
		     AmMagicClass		*magic,
		     AmMagicManagerClass	*displayManager)
{
    if ( displayManager == magic->GetMagicManager()) {
    	magic->DrawWrapper( magic->GetRect());
    }
}

/*************************************************************************
 * 
 * Description:
 * 	handler to update the magic window if the source plot's fields 
 *	are set.  The object has to be tested against the magic window's 
 *	source plot, because another object may have sent the event.
 * Arguments:
 *	eventId - launching user event id.
 *	magic - magic window passed in the AmAddCallback call.
 *	obj - object passed as call data by the sendEvent call.
 *
 *************************************************************************/

/* ARGSUSED */

static void 
updateMagicCallback( int		eventId,
		     AmMagicClass		*magic,
		     AmScreenObjClass	*obj)
{
    if ( obj == ( AmScreenObjClass *) magic->GetSourcePlot()) {
    	magic->Update();
    }
}

/*************************************************************************
 * 
 * Description:
 * 	add the magic window to the magic manager's magic window list.
 *	Add user event handlers to receive events if the source plot is
 *	deleted, resized or has its fields changed, or if the display
 *	manager is flushed.
 * Arguments:
 *	magic - magic window to add to the magic manger's magic window
 *	list.
 *
 *************************************************************************/

void 
AmMagicManagerClass::AddMagic( AmMagicClass	*magic)
{
    AddUserCallback( AM_PKG_ID, AM_DELETE_OBJECT, 
		     ( callbackProc) destroyMagicCallback, magic);
    AddUserCallback( AM_PKG_ID, AM_RECT_CHANGE, 
		     ( callbackProc) resizeMagicCallback, magic);
    AddUserCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT, 
		     ( callbackProc) updateMagicCallback, magic);
    AddUserCallback( AM_PKG_ID, AM_COORDS_CHANGE,
		     ( callbackProc) updateMagicCallback, magic);
    AddUserCallback( AM_PKG_ID, AM_FLUSH, 
		     ( callbackProc) redrawMagicCallback, magic);
    _lstMagic = AmInsertNodeLst( magic, _lstMagic);
}

/*************************************************************************
 * 
 * Description:
 * 	remove the magic window from the magic manager's magic window list.
 *	Remove the user event handlers which were added when the magic
 *	window was added.
 * Arguments:
 *	magic - magic window to remove to the magic manger's magic window
 *	list.
 *
 *************************************************************************/

void 
AmMagicManagerClass::RemoveMagic( AmMagicClass	*magic)
{
    RemoveUserCallback( AM_PKG_ID, AM_DELETE_OBJECT, 
			( callbackProc) destroyMagicCallback, magic);
    RemoveUserCallback( AM_PKG_ID, AM_RECT_CHANGE, 
			( callbackProc) resizeMagicCallback, magic);
    RemoveUserCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT, 
			( callbackProc) updateMagicCallback, magic);
    RemoveUserCallback( AM_PKG_ID, AM_COORDS_CHANGE, 
			( callbackProc) updateMagicCallback, magic);
    RemoveUserCallback( AM_PKG_ID, AM_FLUSH, 
			( callbackProc) redrawMagicCallback, magic);
    _lstMagic = AmDelPtrLst( magic, _lstMagic, ( AmSearchFn) AmEquals);
}
    
/*************************************************************************
 * 
 * Description:
 * 	draw the map manager's contents with the grids and outlines, then 
 *	draw the contents of each exposed magic window.
 * Arguments:
 *	vExposeRect - vector of expose rectangles.
 *	nExposeRect - number of expose rectangles.
 *
 *************************************************************************/

void 
AmMagicManagerClass::ReDraw( AmRectangleRec const	*vExposeRect,
			   int			nExposeRect)
{
    AmListPtr		ptr;
    AmMagicClass		*magic;

    AmMapManagerClass::ReDraw( vExposeRect, nExposeRect);
    for ( ptr = _lstMagic; ptr != NULL; ptr = ptr->next) {
    	magic = ( AmMagicClass *) ptr->data;
	magic->DrawWrapper( magic->GetRect());
    }
}

/*************************************************************************
 * 
 * Description:
 * 	Magic manager constructor creates a map manager, then initialzes
 *	the magic window list.
 * Arguments:
 *	pDisplayClass - display for the magic manager.
 *	pWindowClass - magic manager window.
 *	fDoubleBuffer - double buffer flag.
 *
 *************************************************************************/

AmMagicManagerClass::AmMagicManagerClass( AmDisplayClass	*pDisplayClass,
				      AmWindowClass	*pWindowClass,
				      AnvilBoolean		fDoubleBuffer)
    : AmMapManagerClass( pDisplayClass, pWindowClass, fDoubleBuffer)
{
    _lstMagic = NULL;
}

/* do-nothing constructor */

AmMagicManagerClass::AmMagicManagerClass()
{
    _lstMagic = NULL;
}

/*************************************************************************
 * 
 * Description:
 * 	Magic manager destructor deletes the magic windows in the magic
 *	window list.
 * Arguments:
 *	pDisplayClass - display for the magic manager.
 *	pWindowClass - magic manager window.
 *	fDoubleBuffer - double buffer flag.
 *
 *************************************************************************/

AmMagicManagerClass::~AmMagicManagerClass()
{
    AmListPtr	ptr;
    AmMagicClass	*magic;

    for ( ptr = _lstMagic; ptr != NULL; ptr = ptr->next) {
    	magic = ( AmMagicClass *) ptr->data;
	delete magic;
    }

}
