#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/* functions for manipulating outlines, which are XOR cursors managed by
   the display manager.  The display manager takes care of exposure and
   update management for the outlines, so they are redrawn properly without
   "mouse droppings".  The outlines themselves undraw their AmLast position
   when they're updated. */

#include <stdio.h>
#include <anvil/errcode.h>
#include <anvil/list.h>
#include <anvil/defaults.h>
#include <anvil/mapmgr.h>
#include <anvil/outline.h>


/*******************************************************************
 *
 * Description:
 *   	Constructor which NULLs out the display manager, and 
 *	initializes the "active" flag to FALSE.
 *
 ******************************************************************/

AmOutlineClass::AmOutlineClass()
{
    _pMapManager = NULL;
    _fActive = FALSE;
}

/*******************************************************************
 *
 * Description:
 *	Adds outline to display manager's outline list, and turns
 *	on its active flag so it will be updated.
 * Arguments:
 *	pMapManager - display manager 
 *
 ******************************************************************/

AmOutlineClass::AmOutlineClass( AmMapManagerClass *pMapManager)
{
    _pMapManager = pMapManager;
    _fActive = TRUE;
}

/*******************************************************************
 *
 * Description:
 *	base outline class destructor.
 *
 ******************************************************************/

AmOutlineClass::~AmOutlineClass()
{
}

/*******************************************************************
 *
 * Description:
 *	base outline class comparator function, used to prevent 
 *	overlaid XOR duplicates, normally the derived member functions
 *	test the x,y position.
 * Arguments:
 *	Outline - outline to compare against.
 *
 ******************************************************************/

int
AmOutlineClass::operator == ( AmOutlineClass &Outline)
{
    return( this == &Outline);
}


/*************************************************************************
 *
 * Description:
 *	Base-class 'do-nothing' draw function 
 * Arguments:
 *	pDisplay - display to draw the line on.
 *	pWindow - window to draw on.
 *	pGC - GC to draw with (don't change this, the function should
 *	stay XOR).
 *
 ************************************************************************/

/* ARGSUSED */
AmErrorCode
AmOutlineClass::Draw( AmDisplayClass *pDisplay,
		    AmWindowClass  *pWindow,
		    AmGCClass      *pGC)
{
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *	return the outline's map manager.
 * Returns:
 *	the map manager the outline was instantiated with.
 *
 *************************************************************************/

AmMapManagerClass *
AmOutlineClass::GetMapManager()
{
    return( _pMapManager);
}

/*************************************************************************
 *
 * Description:
 *	return the outline's active flag.
 *
 *************************************************************************/

AnvilBoolean
AmOutlineClass::GetActive()
{
    return( _fActive);
}

/*************************************************************************
 *
 * Description:
 *	set the outline's active flag.
 *
 *************************************************************************/

void
AmOutlineClass::SetActive( AnvilBoolean fActive)
{
    _fActive = fActive;
}

/*************************************************************************
 *
 * Description:
 *	Base Object Outline constructor.
 *
 *************************************************************************/

AmObjectOutlineClass::AmObjectOutlineClass()
{
    _x = 0;
    _y = 0;
    _iWidth = 0;
    _iHeight = 0;
}

/************************************************************************
 *
 * Description:
 *	Create an object outline and add it to the display manager's 
 *	managed outline list, by calling the base AmOutlineClass 
 *	constructor.
 * Arguments:
 *	pMapManager: display manager to add the outline to
 *
 ************************************************************************/

AmObjectOutlineClass::AmObjectOutlineClass( AmMapManagerClass *pMapManager)
    :AmOutlineClass( pMapManager)
{
    _x = 0;
    _y = 0;
    _iWidth = 0;
    _iHeight = 0;
}

/************************************************************************
 *
 * Description:
 *	Object outline contrstructor: sets the base rectangle 
 *	parameters, then calls the outline class constructor, which
 *	adds the outline to the map manager's outline list.
 * Arguments:
 *	pMapManager: display manager to add the outline to,
 *	x, y, iWidth, iHeight: dimensions of the outline rectangle.
 *
 ************************************************************************/

AmObjectOutlineClass::AmObjectOutlineClass( AmMapManagerClass *pMapManager,
					int             x,
					int             y,
					int             iWidth,
					int             iHeight)
    :AmOutlineClass( pMapManager)
{
    _x = x;
    _y = y;
    _iWidth = iWidth;
    _iHeight = iHeight;
}

/************************************************************************
 *
 * Description:
 * 	destructor...calls the base outline constructor implicitly, which
 *  	removes the outline from the map manager's outline list 
 *
 ************************************************************************/

AmObjectOutlineClass::~AmObjectOutlineClass()
{
}

/*************************************************************************
 *
 * Description:
 *	Object outline comparison operator.  Make sure the map manager, 
 *	active flag, and coordinates match.
 *
 *************************************************************************/

int
AmObjectOutlineClass::operator == ( AmObjectOutlineClass &ObjectOutline)
{
    if (( _pMapManager == ObjectOutline._pMapManager) &&
        ( _fActive == ObjectOutline._fActive) &&
        ( _x == ObjectOutline._x) &&
        ( _y == ObjectOutline._y) &&
        ( _iWidth == ObjectOutline._iWidth) &&
        ( _iHeight == ObjectOutline._iHeight)) {

        return( TRUE);
    }
    return( FALSE);
}

/*************************************************************************
 *
 * Description:
 *	Draw the object outline, a rectangle.
 * Arguments:	
 *	pDisplay - display to draw the line on.
 *	pWindow - window to draw on.
 *	pGC - GC to draw with (don't change this, the function should
 *	stay XOR).
 * Returns:
 *	AmOK if everythings OK, AmNULL_ARGUMENT if any of the draw
 *	parameters are NULL.
 *
 *************************************************************************/

AmErrorCode
AmObjectOutlineClass::Draw( AmDisplayClass *pDisplay,
			  AmWindowClass  *pWindow,
			  AmGCClass      *pGC)
{
    if ( pDisplay == NULL || pWindow == NULL || pGC == NULL) {
	DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }
    if ( _iWidth == 0 || _iHeight == 0) {
	DO_RETURN_ERROR( AmZERO_DIMENSION);
    }
    pDisplay->DrawRectangle( pWindow, pGC, _x, _y, _iWidth, _iHeight);

    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *	Retrieve the current rectangle from the object outline 
 * Arguments:
 *	px, py, piWidth, piHeight: dimensions of the rectangle .
 *	if the parameters are NULL, then the values aren't set.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::GetRectangle( int *px,
				  int *py,
				  int *piWidth,
				  int *piHeight)
{
    if ( px != NULL) {
	*px = _x;
    }
    if ( py != NULL) {
	*py = _y;
    }
    if ( piWidth != NULL) {
	*piWidth = _iWidth;
    }
    if ( piHeight != NULL) {
	*piHeight = _iHeight;
    }
}

/************************************************************************
 *
 * Description:
 *	Retrieve the origin of the outline. Note that GetRectangle is
 *	called with NULL parameters for iWidth, iHeight.
 * Arguments:
 *	px, py: pointers to rectangle origin coord results.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::GetOrigin( int *px,
			       int *py)
{
    GetRectangle( px, py, NULL, NULL);
}

/************************************************************************
 *
 * Description:
 *	Retrieve the size of the outline. Note that GetRectangle is
 *	called with NULL parameters for x, y
 * Arguments:
 *	piWidth, piHeight: returned dimensions of the outline.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::GetSize( int *piWidth,
			     int *piHeight)
{
    GetRectangle( NULL, NULL, piWidth, piHeight);
}

/************************************************************************
 *
 * Description:
 *	Set the outline's rectangle.
 * Arguments:
 *	x, y, iWidth, iHeight: new rectangle dimensions.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::SetRectangle( int x,
				  int y,
				  int iWidth,
				  int iHeight)
{
    AnvilBoolean fRedraw;

    if (( _x != x) || ( _y != y) ||
        ( _iWidth != iWidth) || ( _iHeight != iHeight)) {

/* undraw the outline.  Note the fRedraw flag, because HideOutline
   may change the state of the outline */

        if ( _fActive) {
	    _pMapManager->HideOutline( this);
	    fRedraw = TRUE;
        } else {
	    fRedraw = FALSE;
	}
        _x = x;
        _y = y;
        _iWidth = iWidth;
        _iHeight = iHeight;

/* redraw the outline if needed... */

        if ( fRedraw) {
	    _pMapManager->ShowOutline( this);
        }
    }
}

/************************************************************************
 *
 * Description:
 *	Set the origin for the object outline (upper left x,y)
 * Arguments:
 *	x, y - new rectangle origin.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::SetOrigin( int x,
			       int y)
{
    SetRectangle( x, y, _iWidth, _iHeight);
}

/************************************************************************
 *
 * Description:
 *	Set the width and height for the object outline 
 * Arguments:
 *	iWidth, iHeight - new rectangle dimensions.
 *
 ***********************************************************************/

void
AmObjectOutlineClass::SetSize( int iWidth,
			     int iHeight)
{
    SetRectangle( _x, _y, iWidth, iHeight);
}

/************************************************************************
 *
 * Description:
 *	find out if an outline is duplicated by any of the active outlines
 *	in the map manager's outline list.
 * Arguments:
 *	lstOutlines: list of the map manager's outlines.
 *	outline: outline to test against.
 *	up: flag to stop when the passed outline is reached in the list.
 *
 ***********************************************************************/
   
static AnvilBoolean
duplicateOutline( AmListPtr 	lstOutlines, 
		  AmOutlineClass 	*outline, 
		  AnvilBoolean 	upTo)
{
    AmOutlineClass *outline2;
    AmListPtr      temp;

    for ( temp = lstOutlines; temp != NULL; temp = temp->next) {
        outline2 = ( AmOutlineClass *) temp->data;
        if ((( void *) outline2) == (( void *) outline)) {
            if ( upTo) {
                return( FALSE);
            } else {
                continue;
            }
        } else if ( *outline2 == *outline) {
	    break;
	}
    }
    return( temp != NULL);
}

/* Map manager outline-related functions */

/************************************************************************
 *
 * Description:
 *	Add an outline to the map manager's outline list.
 * Arguments:
 *	outline: outline to add to the list, and draw it.
 *
 ***********************************************************************/

AmErrorCode
AmMapManagerClass::AddOutline( AmOutlineClass *outline)
{
    _lstOutlines = AmInsertNodeLst( outline, _lstOutlines);
    
    if ( _fOutlinesActive && outline->GetActive()) {
	if ( !duplicateOutline( _lstOutlines, outline, FALSE)) {
	    outline->Draw( _pDisplay, _pWindow, _pGCOutline);
	}
    }
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *	Remove an outline from the map manager's outline list.
 * Arguments:
 *	outline: outline to remove from the list, and undraw if it
 *	isn't a duplicate.
 * Returns:
 *	AmBAD_ARGUMENT if the outline isn't in the map manager's outline
 *	list, AmOK on success.
 *
 ***********************************************************************/

AmErrorCode
AmMapManagerClass::RemoveOutline( AmOutlineClass *outline)
{
    AmListPtr temp, prev;

    /* Figure out if the outline is part of the outline list. */

    prev = NULL;
    for ( temp = _lstOutlines; temp != NULL; temp = temp->next) {
        if ( (( void *) outline) == temp->data) {
           break;
        }
	prev = temp;
    }

    /* Return if the outline does not exist. */
 
    if ( temp == NULL) {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if ( _fOutlinesActive && outline->GetActive()) {
        if ( !duplicateOutline( _lstOutlines, outline, FALSE)) {
            outline->Draw( _pDisplay, _pWindow, _pGCOutline);
        }
    }

    if ( prev != NULL) {
	prev->next = temp->next;
    } else {
	_lstOutlines = temp->next;
    }
    free( temp);

    delete outline;

    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *     Go through the outline list and draw each one of the outlines,
 *     according to the draw function associated with each.
 * Arguments:
 *	pDisplay - display the outlines are drawn on.
 *	pWindow - window the outlines are drawn to.
 *	pGC - graphics context outlines are drawn with.
 * Modifications:
 *	the display, window, and gc are actually part of the 
 *	display manager, and don't need to be passed as parameters.
 *
 ***********************************************************************/

void
AmMapManagerClass::drawActiveOutlines( AmDisplayClass *pDisplay,
			             AmWindowClass  *pWindow,
			             AmGCClass      *pGC)
{
    AmOutlineClass *newOutline;
    AmListPtr      temp;
    int          count;

    count = 0;

    for ( temp = _lstOutlines; temp != NULL; temp = temp->next) {
        newOutline = ( AmOutlineClass *) temp->data;
	if ( newOutline->GetActive()) {
	    if ( !duplicateOutline( _lstOutlines, newOutline, TRUE)) {
		newOutline->Draw( pDisplay, pWindow, pGC);
            }
	}
    }
}

/************************************************************************
 *
 * Description:
 *     Go through the outline list and draw each one of the outlines,
 *     according to the draw function associated with each.
 * Returns:
 *	AmOK
 *
 ***********************************************************************/

AmErrorCode
AmMapManagerClass::RemoveAllOutlines()
{
    AmListPtr      temp, next;
    AmOutlineClass *newOutline;

    if ( _fOutlinesActive == TRUE) {
        drawActiveOutlines( _pDisplay, _pWindow, _pGCOutline);
    }

    temp = _lstOutlines;
    while ( temp != NULL) {
        newOutline = ( AmOutlineClass *) temp->data;
        next = temp->next;

        delete newOutline;

        free( temp);
        temp = next;
    }

    _lstOutlines = NULL;

    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *     Draw the current set of outlines 
 *
 ***********************************************************************/

void
AmMapManagerClass::ShowOutlines()
{
/* Nothing needs to be done if the outlines are already active. */

    if ( _fOutlinesActive == TRUE) {
        return;
    }

/* Draw the outline. */

    drawActiveOutlines( _pDisplay, _pWindow, _pGCOutline);
    _fOutlinesActive = TRUE;
}

/************************************************************************
 *
 * Description:
 *     Unraw the current set of outlines 
 *
 ***********************************************************************/

void
AmMapManagerClass::HideOutlines()
{

/* Nothing needs to be done if the outlines are not active. */

    if ( _fOutlinesActive == FALSE) {
        return;
    }

/* Remove the outlines. */

    drawActiveOutlines( _pDisplay, _pWindow, _pGCOutline);
    _fOutlinesActive = FALSE;
}

/************************************************************************
 *
 * Description:
 *      Draw the passed outline and set its active flag to TRUE
 * Arguments:
 *	outline - outline to draw.
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT is the outline wasn't in the
 *	map manager's outline list.
 *
 ***********************************************************************/

AmErrorCode
AmMapManagerClass::ShowOutline( AmOutlineClass *outline)
{
    AmListPtr temp;

    for ( temp = _lstOutlines; temp != NULL; temp = temp->next) {
        if ( (( void *) outline) == temp->data) {
	    break;
	}
    }

    if ( temp == NULL) {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if (( _fOutlinesActive == TRUE) && ( outline->GetActive() == FALSE)) {
        if ( duplicateOutline( _lstOutlines, outline, FALSE) == FALSE) { 
            outline->Draw( _pDisplay, _pWindow, _pGCOutline);
        }
    }

    outline->SetActive( TRUE);

    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *      Unraw the passed outline and set its active flag to FALSE
 * Arguments:
 *	outline - outline to draw.
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT is the outline wasn't in the
 *	map manager's outline list.
 *
 ***********************************************************************/

AmErrorCode
AmMapManagerClass::HideOutline( AmOutlineClass *outline)
{
    AmListPtr temp;

    for ( temp = _lstOutlines; temp != NULL; temp = temp->next) {
	if ( (( void *) outline) == temp->data) {
	    break;
	}
    }

    if ( temp == NULL) {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

    if (( _fOutlinesActive == TRUE) && ( outline->GetActive() == TRUE)) {
        if ( duplicateOutline( _lstOutlines, outline, FALSE) == FALSE) { 
            outline->Draw( _pDisplay, _pWindow, _pGCOutline);
        }
    }

    outline->SetActive( FALSE);

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *	Drag outline constructor takes the start X,Y location and
 *	adds the outline to the display manager's outline list 
 * Arguments:
 *	pMapManager: display manager to add the outline to.
 *	startX, startY: starting X,Y location (may be overridden by
 *	SetStartPos().
 *
 *************************************************************************/

AmDragLineOutlineClass::AmDragLineOutlineClass( AmMapManagerClass	*pMapManager,
					    int			startX,
					    int			startY) :
  AmOutlineClass( pMapManager)
{
    _startX = -1;
    _startY = -1;
    _currX = -1;
    _currY = -1;
    SetLine( startX, startY, startX, startY);
}

AmDragLineOutlineClass::~AmDragLineOutlineClass()
{
}

/*************************************************************************
 *
 * Description:
 *	Set the starting X,Y position for the drag outline.
 * Arguments:
 *	startX, startY: starting X,Y location 
 *
 *************************************************************************/

void
AmDragLineOutlineClass::SetStartPos( int	startX,
				   int	startY)
{
    SetLine( startX, startY, _currX, _currY);
}

/*************************************************************************
 *
 * Description:
 *	Set the current X,Y position for the drag outline.
 * Arguments:
 *	currX, currY: current X,Y location 
 *
 *************************************************************************/

void
AmDragLineOutlineClass::SetCurrPos( int	currX,
				  int	currY)
{
    SetLine( _startX, _startY, currX, currY);
}

/*************************************************************************
 *
 * Description:
 *	Set the current x0, y0, x1, y1 position for the drag line.
 * Arguments:
 *	currX, currY: current X,Y location 
 *
 *************************************************************************/

void
AmDragLineOutlineClass::SetLine( int startX,
			       int startY,
			       int currX,
			       int currY)
{
    AnvilBoolean fRedraw;

    if (( startX != _startX) || ( startY != _startY) ||
	( currX != _currX) || ( currY != _currY)) {

        if ( _fActive) {
	    _pMapManager->HideOutline( this);
	    fRedraw = TRUE;
        } else {
	    fRedraw = FALSE;
        }
	_currX = currX;
	_currY = currY;
	_startX = startX;
	_startY = startY;
        if ( fRedraw) {
	    _pMapManager->ShowOutline( this);
        }
    }
}

/*************************************************************************
 *
 * Description:
 *	Set the starting X,Y position for the drag outline.
 * Arguments:
 *	startX, startY: starting X,Y location 
 *
 *************************************************************************/

void
AmDragLineOutlineClass::GetStartPos( int	*pStartX,
				   int	*pStartY)
{
    *pStartX = _startX;
    *pStartY = _startY;
}

/*************************************************************************
 *
 * Description:
 *	Get the current X,Y position for the drag outline.
 * Arguments:
 *	currX, currY: current X,Y location 
 *
 *************************************************************************/

void
AmDragLineOutlineClass::GetCurrPos( int	*pCurrX,
				  int	*pCurrY)
{
    *pCurrX = _currX;
    *pCurrY = _currY;
}

/*************************************************************************
 *
 * Description:
 *	Drag outline comparison operator.  Make sure the map manager, 
 *	active flag, and coordinates match.
 *
 *************************************************************************/

int
AmDragLineOutlineClass::operator == ( AmDragLineOutlineClass &DragOutline)
{
    return(( _pMapManager == DragOutline._pMapManager) &&
	   ( _fActive == DragOutline._fActive) &&
	   ( _currX == DragOutline._currX) &&
	   ( _currY == DragOutline._currY) &&
	   ( _startX == DragOutline._startX) &&
	   ( _startY == DragOutline._startY));
}

/*************************************************************************
 *
 * Description:
 *	Draw the drag outline, a single line 
 * Arguments:	
 *	pDisplay - display to draw the line on.
 *	pWindow - window to draw on.
 *	pGC - GC to draw with (don't change this, the function should
 *	stay XOR).
 * Returns:
 *	AmOK if everythings OK, AmNULL_ARGUMENT if any of the draw
 *	parameters are NULL.
 *
 *************************************************************************/

AmErrorCode
AmDragLineOutlineClass::Draw( AmDisplayClass *pDisplay,
			    AmWindowClass  *pWindow,
			    AmGCClass      *pGC)
{
    if ( pDisplay == NULL || pWindow == NULL || pGC == NULL) {
	DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }
    pDisplay->DrawLine( pWindow, pGC, _startX, _startY, _currX, _currY);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *	Draw the magic outline, which is the object outline with a
 *	resize rectangle in the lower right hand corner.
 * Arguments:	
 *	pDisplay - display to draw the line on.
 *	pWindow - window to draw on.
 *	pGC - GC to draw with (don't change this, the function should
 *	stay XOR).
 * Returns:
 *	AmOK if everythings OK, AmNULL_ARGUMENT if any of the draw
 *	parameters are NULL.
 *
 *************************************************************************/

AmErrorCode
AmWindowOutlineClass::Draw( AmDisplayClass *pDisplay,
			  AmWindowClass  *pWindow,
			  AmGCClass      *pGC)
{

    if ( pDisplay == NULL || pWindow == NULL || pGC == NULL) {
	DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }
    AmObjectOutlineClass::Draw( pDisplay, pWindow, pGC);
    pDisplay->DrawLine( pWindow, pGC, 
    			_x + _iWidth - _resizeTabSize,
    			_y + _iHeight - _resizeTabSize,
    			_x + _iWidth,
    			_y + _iHeight - _resizeTabSize);
    pDisplay->DrawLine( pWindow, pGC, 
    			_x + _iWidth - _resizeTabSize,
    			_y + _iHeight - _resizeTabSize,
    			_x + _iWidth - _resizeTabSize,
    			_y + _iHeight);
    return( AmOK);
}

/************************************************************************
 *
 * Description:
 *	Construct the magic outline, which calls the AmObjectOutlineClass
 *	constructor, then sets the resizeTabSize.
 * Arguments:
 *	pMapManager: display manager to add the outline to,
 *	x, y, iWidth, iHeight: dimensions of the outline rectangle.
 *
 ************************************************************************/

AmWindowOutlineClass::AmWindowOutlineClass( AmMapManagerClass *pMapManager,
				        int             x,
				        int             y,
				        int             iWidth,
				        int             iHeight,
				        int	        iResizeTabSize)
    :AmObjectOutlineClass( pMapManager, x, y, iHeight, iHeight)
{
    _resizeTabSize = iResizeTabSize;
}

/************************************************************************
 *
 * Description:
 * 	window outline class destructor 
 *
 ************************************************************************/

AmWindowOutlineClass::~AmWindowOutlineClass()
{
}

/************************************************************************
 *
 * Description:
 * 	test if the point is inside of the window outline's resize 
 *	rectangle?
 * Arguments:
 *	x, y - point to test.
 * Returns:
 *	TRUE if the point is inside the window outline's resize rectangle
 *	FALSE otherwise.
 *
 ************************************************************************/

AnvilBoolean
AmWindowOutlineClass::inResize( int	x,
			      int	y)
{
    return(( x < _x + _iWidth) && ( y < _y + _iHeight) &&
	   ( x >_x + _iWidth - _resizeTabSize) &&
	   ( y > _y + _iHeight - _resizeTabSize));
}
