#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


/* Tree search generalized from Knuth (6.2.2) Algorithm T just like
   the AT&T man page says.

   The node_t structure is for internal use only, lint doesn't grok it.

   Written by reading the System V Interface Definition, not the code.

   Totally public domain.  */

#include <stdlib.h>
#include <anvil/tsearch.h>

typedef struct node_t {
  const void *key;
  struct node_t *left;
  struct node_t *right;
} node;


void *
tsearch( const void 	*key,
         void	  	**vrootp,
         AmSearchFn	compar)
{
    node *q;
    node **rootp = (node **) vrootp;

    if (rootp == NULL) return NULL;

    while (*rootp != NULL)		/* Knuth's T1: */
    {
        int r;

        r = (*compar) (key, (*rootp)->key);
        if (r == 0)			/* T2: */
	  return *rootp;			/* we found it! */
        rootp = (r < 0)
	        ? &(*rootp)->left		/* T3: follow left branch */
	        : &(*rootp)->right;	/* T4: follow right branch */
    }

    q = (node *) malloc (sizeof (node));	/* T5: key not found */
    if (q != NULL)			/* make new node */
    {
        *rootp = q;			/* link new node to old */
        q->key = key;			/* initialize new node */
        q->left = q->right = NULL;
    }
    return q;
}


void *
tfind ( const void 	*key, 
	void * const 	*vrootp, 
	AmSearchFn	compar)
{
    node **rootp = (node **) vrootp;

    if (rootp == NULL)
	return NULL;

    while (*rootp != NULL)		/* Knuth's T1: */
    {
        int r;

        r = (*compar)(key, (*rootp)->key);
        if (r == 0)			/* T2: */
	    return *rootp;			/* we found it! */

        rootp = (r < 0)
		  ? &(*rootp)->left	/* T3: follow left branch */
		  : &(*rootp)->right;	/* T4: follow right branch */
    }
    return NULL;
}


void *
tdelete ( const void 	*key, 
	  void		**vrootp, 
	  AmSearchFn	compar)
{
  node *p;
  node *q;
  node *r;
  int cmp;
  node **rootp = (node **) vrootp;

  if (rootp == NULL || (p = *rootp) == NULL)
    return NULL;

  while ((cmp = (*compar) (key, (*rootp)->key)) != 0)
    {
      p = *rootp;
      rootp = (cmp < 0)
	      ? &(*rootp)->left		/* follow left branch */
	      : &(*rootp)->right;	/* follow right branch */
      if (*rootp == NULL)
	return NULL;			/* key not found */
    }

  r = (*rootp)->right;			/* D1: */
  q = (*rootp)->left;
  if (q == NULL)			/* Left NULL? */
    q = r;
  else if (r != NULL)			/* Right link is NULL? */
    {
      if (r->left == NULL)		/* D2: Find successor */
	{
	  r->left = q;
	  q = r;
	}
      else
	{				/* D3: Find (struct node_t *)0 link */
	  for (q = r->left; q->left != NULL; q = r->left)
	    r = q;
	  r->left = q->right;
	  q->left = (*rootp)->left;
	  q->right = (*rootp)->right;
	}
    }
  free ((struct node_t *) *rootp);	/* D4: Free node */
  *rootp = q;				/* link parent to new node */
  return p;
}


static void
trecurse (const void 	*vroot, 
	  AmWalkFn	action, 
	  int		level)
{
  node *root = (node *) vroot;

  if (root->left == NULL && root->right == NULL)
    (*action) (root, leaf, level);
  else
    {
      (*action) (root, preorder, level);
      if (root->left != NULL)
	trecurse (root->left, action, level + 1);
      (*action) (root, postorder, level);
      if (root->right != NULL)
	trecurse (root->right, action, level + 1);
      (*action) (root, endorder, level);
    }
}


void 
twalk ( const void 	*vroot, 
	AmWalkFn	action)
{
  const node *root = (node *) vroot;

  if (root != NULL && action != NULL)
    trecurse (root, action, 0);
}
