#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <math.h>
#include <anvil/macros.h>
#include <anvil/image.h>
#include <anvil/imageP.h>

/* a more accurate forward differencing vector, which uses the real 
   coordinates in the source matrix */

void AmRealFwdDiffVec( float	min,
		     float	max,
		     int	length,
		     int	*vec)
{
    float	sum, inc = ( max - min)/length, lastVal;
    int		ii;

    sum = min;
    lastVal = floor( min);
    for ( ii = 0; ii < length; ii++) {
	sum += inc;
	vec[ ii] = ( int) sum - ( int) lastVal;
	lastVal = sum;
    }
}

/* make a real forward differencing vector */

int *AmMakeRealFwdDiffVec( float	min,
			 float	max,
			 int	length)
{
    int	*vec;
    
    if (( vec = ( int *) malloc( length*sizeof( int))) == NULL) {
	return( NULL);
    }
    AmRealFwdDiffVec( min, max, length, vec);
    return( vec);
}

/* fill a remainder vector ( scaled by 256..) */

void AmRealRemainVec( float	min,
		    float	max,
		    int		length,
		    int		*vec)
{
    float	ratio = ( max - min)/length, temp;
    int		ii;

    temp = min; 
    for ( ii = 0; ii < length; ii++) {
	vec[ ii] = ( int) (( temp - ( int) temp)*256);
    	temp += ratio;
    }
}

/* create a remainder vector */

int *AmMakeRealRemainVec( float	min,
			float	max,
			int	length)
{
    int	*vec;
    
    if (( vec = ( int *) malloc( length*sizeof( int))) == NULL) {
	return( NULL);
    }
    AmRealRemainVec( min, max, length, vec);
    return( vec);
}

/* make a floating point index vector from a min/max into a length */

int *AmMakeRealIndexVec( float	min,
		       float	max,
		       int	length)
{
    int		*vec, ii;
    float	val, inc = ( max - min)/length;

    val = min - floor( min);
    if ((vec = ( int *) malloc( sizeof( int)*length)) == NULL) {
	return( NULL);
    }
    for ( ii = 0; ii < length; ii++) {
        vec[ ii] = ( int) val;
        val += inc;
    }
    return( vec);
}

/* take an already allocated array and make a forward differencing vector
   in it */

void AmFwdDiffVec( int	max,
		 int	length,
		 int	*vec)
{
    int         ii, lastVal, currVal, val;

    val = max;
    lastVal = 0;
    for ( ii = 0; ii < length; ii++) {
        currVal = val / length;
        val += max;
        vec[ ii] = currVal - lastVal;
        lastVal = currVal;
    }
}


/* create a forward differencing vector up to value "max" of "length" */
 
int *AmMakeFwdDiffVec( int    max,
		     int    length)
{
    int  *vector;

    if (( vector = ( int *) malloc( sizeof( int) * length)) == NULL) {
	return( NULL);
    }
    AmFwdDiffVec( max, length, vector);
    return( vector);
}

/* make a straight indexing vector */

int *AmMakeIndexVec( int	max,
		   int	length)
{
    int  ii, val, *vec;

    val = 0;
    if (( vec = ( int *) malloc( sizeof( int) * length)) == NULL) {
	return( NULL);
    }
    for ( ii = 0; ii < length; ii++) {
        vec[ ii] = val / length;
        val += max;
    }
    return( vec);
}

/* create a forward differencing vector up to value "max" of "length" */
 
int *AmMakeDitherVec( int	spacing,
		    int  length)
{
    int *vecPtr, *vec, ii;

    length += spacing;
    if (( vecPtr = vec = ( int *) malloc( sizeof( int)*length)) == NULL) {
	return( NULL);
    }
    for ( ii = 0; ii < length; ii++) {
	*vecPtr++ = ( ii%spacing == 0);
    }
    return( vec);
}

/* generate the lookup tables for transparency in 24-bit color */
 
void AmGenRGBLookups( float         srcWeight,
                    ulong_t       *srcRed,
                    ulong_t       *srcGreen,
                    ulong_t       *srcBlue,
                    ulong_t       *dstRed,
                    ulong_t       *dstGreen,
                    ulong_t       *dstBlue)
{
    int         ii, val;
    float       srcVal1 = 0.0,
                srcInc1 = srcWeight/256,
                srcVal2 = 0.0,
                srcInc2 = ( 1.0 - srcWeight)/256;
 
    for ( ii = 0; ii < 256; ii++, srcVal1 += srcInc1, srcVal2 += srcInc2) {
        val = ( int) ( srcVal1*256);
        srcRed[ ii] = val << RED_SHIFT;
        srcGreen[ ii] = val << GREEN_SHIFT;
        srcBlue[ ii] = val << BLUE_SHIFT;
        val = ( int) ( srcVal2*256);
        dstRed[ ii] = val << RED_SHIFT;
        dstGreen[ ii] = val << GREEN_SHIFT;
        dstBlue[ ii] = val << BLUE_SHIFT;
    }
}


/* interpolate four pixels with 256-scaled remainder vectors... */

ulong_t AmInterpPixels( ulong_t	pt1,
		      ulong_t	pt2,
		      ulong_t	pt3,
		      ulong_t	pt4,
		      int	remRow,
		      int	remCol)
{ 
    ulong_t	red, green, blue;
    uchar_t	*p1 = ( uchar_t *) &pt1, 
    		*p2 = ( uchar_t *) &pt2, 
		*p3 = ( uchar_t *) &pt3, 
		*p4 = ( uchar_t *) &pt4;

    red = (( p1[ RED]*( 256 - remCol) + p2[ RED]*remCol)*( 256 - remRow) +
	   ( p3[ RED]*( 256 - remCol) + p4[ RED]*remCol)*remRow) >> 16;
    green = (( p1[ GREEN]*( 256 - remCol) + p2[ GREEN]*remCol)*( 256 - remRow) +
	     ( p3[ GREEN]*( 256 - remCol) + p4[ GREEN]*remCol)*remRow) >> 16;
    blue = (( p1[ BLUE]*( 256 - remCol) + p2[ BLUE]*remCol)*( 256 - remRow) +
	    ( p3[ BLUE]*( 256 - remCol) + p4[ BLUE]*remCol)*remRow) >> 16;
    return(( red << RED_SHIFT) | 
     	   ( green << GREEN_SHIFT) |
	   ( blue << BLUE_SHIFT));
}
