#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/errcode.h>
#include <anvil/display.h>
#include <anvil/image.h>
#include <anvil/imageP.h>

/* copy a submatrix of a source matrix to a destination matrix */

AmErrorCode 
AmCopySubmatrix( uchar_t		*srcMatrix,
		 int			srcRows,
		 int			srcCols,
		 AmRectangleRec const	*srcRect,
		 uchar_t		*dstMatrix,
		 int			dstRows,
		 int			dstCols,
		 int			dstX,
		 int			dstY,
		 int			size,
		 AnvilBoolean		flipRows,
		 AnvilBoolean		flipCols)
{
    int		yy, xx, srcIndex, dstIndex;
    uchar_t	*srcRowPtr, *dstRowPtr;

/* check the source rectangle to make sure it's contained in the 
   destination */

    if (( srcRect->x < 0) || 
	( srcRect->x + srcRect->width > srcCols) ||
	( srcRect->y < 0) ||
	( srcRect->y + srcRect->height > srcRows) ||
	( dstX + srcRect->width > dstCols) ||
	( dstY + srcRect->height > dstRows)) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* the order of the loops changes depending on flipRows and flipCols */

    if ( flipRows) {
	for ( yy = 0; yy < srcRect->height; yy++) {
	    srcIndex = srcRect->x + ( srcRect->y + yy)*srcCols;
	    dstIndex = dstX + (dstY + ( srcRect->height - yy - 1)*dstCols);
	    srcRowPtr = &srcMatrix[ srcIndex*size];
	    dstRowPtr = &dstMatrix[ dstIndex*size];
	    if ( flipCols) {
		for ( xx = 0; xx < srcRect->width; xx++) {
		     memcpy( &dstRowPtr[ (srcRect->width - xx - 1)*size],
			     &srcRowPtr[ xx*size], size);
		}
	    } else {
		memcpy( dstRowPtr, srcRowPtr, srcRect->width*size);
	    }
	}
    } else {
	for ( yy = 0; yy < srcRect->height; yy++) {
	    srcIndex = srcRect->x + ( srcRect->y + yy)*srcCols;
	    dstIndex = dstX + (dstY + yy)*dstCols;
	    srcRowPtr = &srcMatrix[ srcIndex*size];
	    dstRowPtr = &dstMatrix[ dstIndex*size];
	    if ( flipCols) {
		for ( xx = 0; xx < srcRect->width; xx++) {
		     memcpy( &dstRowPtr[ (srcRect->width - xx - 1)*size],
			     &srcRowPtr[ xx*size], size);
		}
	    } else {
		memcpy( dstRowPtr, srcRowPtr, srcRect->width*size);
	    }
	}
    }
    return( AmOK);
}

/* resize a source matrix to a destination matrix */

AmErrorCode 
AmResizeMatrix( uchar_t			*srcMatrix,
		int			srcRows,
		int			srcCols,
		uchar_t			*dstMatrix,
		int			dstRows,
		int			dstCols,
		int			size,
		AmRectangleRec const	*srcRect,
		AmRectangleRec const	*dstRect)
{
    int		*hIndex, *vIndex, *pHIndex, *pVIndex, srcWidth, dstWidth, ii;
    uchar_t	*srcPtr, *dstPtr, *dstEnd, *rowPtr, *rowEnd, *srcRowPtr;

    if (( srcRect->x < 0) ||
	( srcRect->x + srcRect->width > srcCols) ||
	( srcRect->y < 0) ||
	( srcRect->y + srcRect->height > srcRows) ||
        ( dstRect->x < 0) ||
	( dstRect->x + dstRect->width > dstCols) ||
	( dstRect->y < 0) ||
	( dstRect->y + dstRect->height > dstRows)) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* create the horizontal and vertical difference vectors */

    if (( hIndex = AmMakeFwdDiffVec( srcRect->width, 
				     dstRect->width)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ii = 0; ii < dstRect->width; ii++) {
	hIndex[ ii] *= size;
    }
    if (( vIndex = AmMakeFwdDiffVec( srcRect->height, 
				     dstRect->height)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    srcWidth = srcCols*size;
    dstWidth = dstCols*size;
    for ( ii = 0; ii < dstRect->height; ii++) {
	vIndex[ ii] *= srcWidth;
    }
    pVIndex = vIndex;

/* offset the source and destination matrices by their rectangles x,y */

    srcPtr = srcMatrix + ( srcRect->y*srcCols + srcRect->x)*size;
    dstPtr = dstMatrix + ( dstRect->y*dstCols + dstRect->x)*size;
    dstEnd = dstPtr + dstRect->height*dstWidth;

/* loop over each row, setting pointers to the start of the source and
   destination matrix, and set the end row marker */

    while ( dstPtr < dstEnd) {
	pHIndex = hIndex;
	rowPtr = dstPtr;
	rowEnd = dstPtr + dstRect->width*size;
	srcRowPtr = srcPtr;

/* loop to the end of the row of the destination matrix */

	while ( rowPtr < rowEnd) {
	    memcpy( rowPtr, srcRowPtr, size);
	    rowPtr += size;
	    srcRowPtr += *pHIndex++;	/* premultiplied by size */
	}

/* step to the next row of the destination matrix, and the corresponding
   source matrix */

	dstPtr += dstWidth;
	srcPtr += *pVIndex++;	/* premultiplied by srcWidth */
    }
    free( hIndex);
    free( vIndex);
    return( AmOK);
}
