#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/display.h>
#include <anvil/image.h>
#include <anvil/imageP.h>

/* read and resize from a source file to a matrix */

AmErrorCode 
AmMatrixReadRows( int			fd,
		  int			srcRows,
		  int			srcCols,
		  AmRealRectRec const   *srcRect,
		  uchar_t		*matrix,
		  int			dstRows,
		  int			dstCols,
		  AmRectangleRec const	*dstRect,
		  int			size,
		  AnvilBoolean		flipRows,
		  AnvilBoolean		flipCols,
		  AmScanlineFn		readScanFn,
		  void			*readScanClientData)
{
    int		  *hIndex, *vIndex, *pHIndex, *pVIndex, ii, lastIndex, 
    		  bufWidth; 
    uchar_t	  *dstData, *rowStart, *rowEnd, *srcDataPtr, *dstEnd, *buffer;
    AnvilBoolean	  okReadFlag;

    if (( srcRect->x < 0) ||
	( srcRect->y < 0) ||
	( srcRect->x + srcRect->width > srcCols) ||
	( srcRect->y + srcRect->height > srcRows) ||
	( dstRect->x < 0) ||
	( dstRect->y < 0) ||
	( dstRect->x + dstRect->width > dstCols) ||
	( dstRect->y + dstRect->height > dstRows)) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* allocate the forward differencing vectors, and the read buffer.  Scale
   the horizontal vector by the source size so we can avoid a multiply
   in the core loop */

    if (( hIndex = AmMakeRealFwdDiffVec( srcRect->x, srcRect->x + srcRect->width, 
				       dstRect->width)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ii = 0; ii < dstRect->width; ii++) {
	hIndex[ ii] *= size;
    }
    if (( vIndex = AmMakeRealIndexVec( srcRect->y, srcRect->y + srcRect->height, 
				     dstRect->height)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    pVIndex = vIndex;
    for ( ii = 0; ii < dstRect->height; ii++) {
	vIndex[ ii] += ( int) srcRect->y;
    }
    lastIndex = -1;
    bufWidth = ( int) srcRect->width + 4;
    if ((buffer = ( unsigned char *) malloc( bufWidth*size)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* loop until the end of the destination subimage, setting pointers at the
   start and end of each row.  Use the vertical and horizontal forward
   differencing vectors to increment through the source image */

    okReadFlag = TRUE;

/* seek to the next source scanline, by the index vector */
   
    if ( flipRows) {
	dstEnd = &matrix[ (dstRect->y*dstCols + dstRect->x)*size];
	dstData = &matrix[ ( dstRect->x + 
			     ( dstRect->y + dstRect->height - 1)*dstCols)*size];
	while ( dstData >= dstEnd) {
	    if ( *pVIndex != lastIndex) {
		if ( !( readScanFn)( fd, buffer, size, bufWidth, srcCols,
				     ( int) srcRect->x, *pVIndex,
				     readScanClientData)) {
		    okReadFlag = FALSE;
		    break;
		}
		lastIndex = *pVIndex;
	    }
	    pVIndex++;

/* map the file's scanline to the image scanline, forward or reverse, 
   depending on the flipCols flag */

	    pHIndex = hIndex;
	    srcDataPtr = buffer;
	    if ( flipCols) {
		rowStart = dstData + ( dstRect->width - 1)*size;;
		rowEnd = dstData;
		while ( rowStart >= rowEnd) {
		    memcpy( rowStart, srcDataPtr, size);
		    srcDataPtr += *pHIndex++;
		    rowStart += size;
		}
	    } else {
		rowStart = dstData;
		rowEnd = dstData + dstRect->width*size;
		while ( rowStart < rowEnd) {
		    memcpy( rowStart, srcDataPtr, size);
		    srcDataPtr += *pHIndex++;
		    rowStart -= size;
		}
	    }
	    dstData -= dstCols*size;
	}
    } else { 
	dstData = &matrix[ (dstRect->y*dstCols + dstRect->x)*size];
	dstEnd = &matrix[ ( dstRect->x + 
			    ( dstRect->y + dstRect->height)*dstCols)*size];
	while ( dstData < dstEnd) {
	    if ( *pVIndex != lastIndex) {
		if ( !( readScanFn)( fd, buffer, size, bufWidth, srcCols,
				     ( int) srcRect->x, *pVIndex,
				     readScanClientData)) {
		    okReadFlag = FALSE;
		    break;
		}
		lastIndex = *pVIndex;
	    }
	    pVIndex++;

/* map the file's scanline to the image scanline, forward or reverse, 
   depending on the flipCols flag */

	    pHIndex = hIndex;
	    srcDataPtr = buffer;
	    if ( flipCols) {
		rowStart = dstData + ( dstRect->width - 1)*size;;
		rowEnd = dstData;
		while ( rowStart >= rowEnd) {
		    memcpy( rowStart, srcDataPtr, size);
		    srcDataPtr += *pHIndex++;
		    rowStart -= size;
		}
	    } else {
		rowStart = dstData;
		rowEnd = dstData + dstRect->width*size;
		while ( rowStart < rowEnd) {
		    memcpy( rowStart, srcDataPtr, size);
		    srcDataPtr += *pHIndex++;
		    rowStart += size;
		}
	    }
	    dstData += dstCols*size;
	}
    }
    free( buffer);
    free( hIndex);
    free( vIndex);
    ANVIL_ERROR_CODE = (okReadFlag) ? AmOK : AmREAD_FAILED;
    return( ANVIL_ERROR_CODE);
}
