#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdio.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/list.h>

/*************************************************************************
 *
 * Description:
 * 	insert an element into a sorted list and maintain the ordering 
 *	according to a comparison function.
 * Arguments:
 *	el - element to insert.
 *	lst - list to insert it into.
 *	cmpFn - comparison function to use for ordering.
 * Returns:
 *	The list with the element inserted.
 * Modifications:
 *	should return an error if the insert malloc() fails.
 *
 *************************************************************************/

AmListPtr AmInsertSortLst( void 	*el,
		       AmListPtr	lst,
		       AmSearchFn	cmpFn)
{
    AmListPtr	head, prevNode;

/* no list?, well that's easy. */

    if ( lst == NULL) {
	return( AmInsertNodeLst( el, NULL));
    }

/* if smaller than the head of the list, then just CONS it onto the head
   of the list, or if there's only one element in the list, then put it
   after the AmFirst element */

    if (( cmpFn)( el, lst->data) < 0) {
	return( AmInsertNodeLst( el, lst));
    } else if ( lst->next == NULL) {
	lst->next = AmInsertNodeLst( el, NULL);
	return( lst);
    }

/* save the head of the list, the previous node for hooking into the new node
   then step to the next node */

    head = lst;
    prevNode = lst;
    lst = lst->next;
    do {

/* is it AmBetween the previous element and the current element? */

	if ((( cmpFn)( el, prevNode->data) >= 0) &&
	    (( cmpFn)( el, lst->data) < 0)) {
	    prevNode = AmInsertNodeAfterLst( el, prevNode);
	    return( head);
	}
	lst = lst->next;
	prevNode = prevNode->next;
    } while ( lst != NULL);

/* bigger than the largest element? put it on the end of the list */

    prevNode = AmInsertNodeAfterLst( el, prevNode);
    return( head);
}

/* same thing, but don't insert if it's equal */
/*************************************************************************
 *
 * Description:
 * 	insert an element into a sorted list, but don't insert it if
 *	it's considered equal according to a comparison function.
 * Arguments:
 *	el - element to insert.
 *	lst - list to insert it into.
 *	cmpFn - comparison function to use for ordering.
 * Returns:
 *	The list with the element inserted (maybe)
 * Modifications:
 *	should return an error if the insert malloc() fails.
 *
 *************************************************************************/

AmListPtr AmInsertUniqLst( void 	*el,
		       AmListPtr	lst,
		       AmSearchFn	cmpFn)
{
    AmListPtr	head, prevNode;
    int		val;

/* no list?, well that's easy. */

    if ( lst == NULL) {
	head = ( AmListPtr) malloc( sizeof( AmListRec));
	head->next = NULL;
	head->data = el;
	return( head);
    }

/* if smaller than the head of the list, then just CONS it onto the head
   of the list */

    val = ( cmpFn)( el, lst->data);
    if ( val < 0) {
	head = ( AmListPtr) malloc( sizeof( AmListRec));
	head->next = lst;
	head->data = el;
	return( head);
    } else if ( val == 0) {
	return( lst);
    }

/* save the head of the list, the previous node for hooking into the new node
   then step to the next node */

    head = lst;
    prevNode = lst;
    for ( lst = lst->next; 
          lst != NULL; 
	  lst = lst->next, prevNode = prevNode->next) {

/* smaller than next element? put it after the previous element */

	val = ( cmpFn)( el, lst->data);
	if ( val < 0) {
	    prevNode = AmInsertNodeAfterLst( el, prevNode);
	    return( head);
	} else if ( val == 0) {
	    return( head);
	}
    } 

/* end of the list? put it on the end. */

    prevNode = AmInsertNodeAfterLst( el, prevNode);
    return( head);
}
