#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* utility functions for breaking up strings into string vectors */

#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/strlib.h>

/****************************************************************************
 *
 * Description:
 * 	chop up a string of tokens separated by delimiters into an allocated, 
 *	NULL-terminated vector of strings.  The delimiters are removed
 *	from AmBetween the tokens.
 * Arguments:
 *	string - string to break up.
 *	delim - delimiter AmBetween the tokens.
 * Returns:
 *	An allocated, NULL-terminated vector of strings, or NULL if the
 *	vector couldn't be allocated.
 *
 ****************************************************************************/

char **
AmChopString( char *string, 
	    int  delim)
{
    int		length;
    char	**vec, **vecptr, *p, *q;

    length = 0;

/* careful in this loop, because it can terminate either by the string
   terminator, or strchr() returning NULL */

    for ( p = string; ( p != NULL) && ( *p != '\0'); length++) {
        p = strchr( ( const char *)(p + 1), delim);
    }
    vecptr = vec = ( char **) malloc( sizeof( char *)*( length + 1));
    if ( vecptr == NULL) {
	return( NULL);
    }

/* careful in this loop, because it can terminate either by the string
   terminator, or strchr() returning NULL */

    for ( p = string; ( p != NULL) && ( *p != '\0');) {
        q = strchr( ( const char *)(p + 1), delim);  
	if ( q != NULL) {
	    *vecptr = ( char *) malloc( q - p + 1);
	    strncpy( *vecptr, p, q - p);
	    (*vecptr)[ q - p] = '\0';
            p = q + 1;
	} else {
	    *vecptr = strdup( p);
	    p = NULL;
	}
        vecptr++;
    }
    *vecptr = NULL;
    return( vec);
}

/****************************************************************************
 *
 * Description:
 *     Splits a buffer of strings into a vector of strings.  Quoted
 *     strings can be used to generate strings that contain white 
 *     space.
 * Arguments:
 *	string - string to break up.
 * Returns:
 *	An allocated, NULL-terminated vector of strings, or NULL if the
 *	vector couldn't be allocated.
 *
 ****************************************************************************/

char **
AmVectorQuotedStrings( char *string)
{
    char	*p, *start, *end, **vector;
    int		count, length, state, ii, jj;

    if ( string == NULL) {
	return( NULL);

#if 0
        if ( !(vector = ( char **)malloc( sizeof( char *)))) {
            ANVIL_ERROR_CODE = AmMALLOC_FAILED;
            return( NULL);
        }
        vector[ 0] = NULL;
        return( vector);
#endif
    }
    if ( !strcasecmp( string, "NULL")) {
	return( NULL);
    }

  /*
   * Figure out how many strings there are... strings can be surrounded
   *   by white spaces or by quotes.
   */

    count = 0;
    state = 0;
    for ( p = string; *p != '\0'; p++) {
        switch( state) {
            case 0: /* Outside a string. */
                if ( *p == '\"') {
                    state = 1;
                } else if ( !isspace( *p)) {
                    state = 2;
                }
                break;
            case 1: /* Inside a quoted string. */
                if ( *p == '\"') {
                    count++;
                    state = 0;
                }
                break;
            case 2: /* Inside a string surrounded by white space. */
                if ( isspace( *p)) {
                    count++;
                    state = 0;
                } else if ( *p == '\"') {
                    count++;
                    state = 1;
                }
                break;
        }
    }

  /*
   * If we are in the middle of a string at the end count that as an
   *   additional string.
   */

    if ( state != 0) {
        count++;
    }

    if ( !(vector = ( char **) malloc( sizeof( char *)*( count + 1)))) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    p = string;
    for ( ii = 0; ii < count; ii++) {

  /* Find the start of the next string. */

        while ( isspace( *p)) {
            p++;
        }

        if ( *p == '\"') {
            p++; /* Skip the quotes. */
            start = p;
            while ( *p && *p != '\"') {
                p++;
            }
            end = p; 
            if ( *p == '\"') {
                p++; /* Skip the quotes. */
            }
        } else {
            start = p;
            while ( *p && !isspace( *p) && *p != '\"') {
                p++;
            }
            end = p;
        }

    /* Duplicate the string. */

        length = end - start;
	if ( !(vector[ ii] = ( char *) malloc( length + 1))) {

            for ( jj = 0; jj < ii; jj++) {
                free( vector[ jj]);
            }
            free( vector);

            ANVIL_ERROR_CODE = AmMALLOC_FAILED;
            return( NULL);
        }

        if ( length) {
  	    memcpy( vector[ ii], start, length);
        }
	vector[ ii][ length] = '\0';
    }

/* Put in a NULL string as an indicator for the AmLast entry. */

    vector[ count] = NULL;
    return( vector);
}

/****************************************************************************
 *
 * Description:
 * 	copy a string to another string, replacing the newlines with '\n's'
 *	used for dumping multiline-strings.
 * Arguments:
 *	dst - destination string.
 *	src - source string.
 *
 ****************************************************************************/

void 
AmCopyNewlineString( char	*dst,
		   char	*src)
{
    while ( *src != '\0') {
	if ( *src == '\n') {
	    *dst++ = '\\';
	    *dst++ = 'n';
	} else {
	    *dst++ = *src;
	}
	src++;
    }
    *dst = '\0';
}

/****************************************************************************
 *
 * Description:
 * 	replace the '\n's in a string with newlines (inverse of 
 *	AmCopyNewlineString). 
 * Arguments:
 *	str - string to replace the newlines in
 *
 ****************************************************************************/

void 
AmReplaceNewlines( char *str)
{
    char	*p;
 
    p = str;
    while ( *str != '\0') {
	if (( str[ 0] == '\\') && ( str[ 1] == 'n')) {
	    *p++ = '\n';
	    str += 2;
	} else { 	
	    *p++ = *str++;
	}
    }
    *p = '\0';
}
