#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <anvil/choppath.h>
#include <anvil/files.h>

/* file path manipulation utilities. */

/****************************************************************************
 *
 * Description:
 * 	is p a path delimiting character? (for UNIX or Windows)
 * Arguments:
 *	c - character to test 
 * Returns:
 *	TRUE if it's a path delimiter, FALSE otherwise
 *
 ****************************************************************************/

AnvilBoolean 
AmPathDelim( char c)
{
    return(( c == '/') || ( c == '\\'));
}

/****************************************************************************
 *
 * Description:
 * 	Non-Destructively returns the root filename from a file path.  
 * Arguments:
 *	filename - full file path.
 * Returns:
 *	pointer into the file path at the start of the file name.
 *
 ****************************************************************************/

const char *
AmGetRootName( const char *filename)
{
    const char *ptr;

    if ( filename == NULL) {
	AN_DEBUG( 1, ( "NULL filename passed to AmGetRootName\n"));
	return( NULL);
    }
    ptr = filename + strlen( filename);
    while(( !AmPathDelim( *ptr)) &&  ptr > filename) {
        ptr--;
    }
    if ( AmPathDelim( *ptr)) {
        ptr++;
    }
    return( ptr);
}

/****************************************************************************
 *
 * Description:
 * 	Destructively chop the extension from a filename.
 * Arguments:
 *	filename - file to chop.
 *
 ****************************************************************************/

void 
AmChopExtension( char *filename)
{
    char *p;

    if ( filename == NULL) {
	AN_DEBUG( 1, ( "NULL filename passed to AmChopExtension\n"));
	return;
    }
    for ( p = filename + strlen( filename) -1; p >= filename; p--) {
	if ( *p == '.') {
	    *p = '\0';
	    return;
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	Return a pointer to the extension in a filename (after the '.')
 * Arguments:
 *	filename - file to get the extension from.
 * Returns:
 *	A pointer to the start of the extension at the end of the filename.
 *
 ****************************************************************************/

const char *
AmGetExtension( const char *filename)
{
    const char *p;
  
/* return NULL if there is no filename */

    if ( filename == NULL) {
	AN_DEBUG( 1, ( "NULL filename passed to AmChopExtension\n"));
	return( NULL);
    }
    for ( p = filename + strlen( filename) -1; p >= filename; p--) {
	if ( *p == '.') {
            if ( *(p + 1) != '\0') {
	        return( p + 1);
            } else {
                return( NULL);
            }
	}
    }
    return( NULL);
}

/****************************************************************************
 *
 * Description:
 * 	Return the directory path from a filename in a static pointer
 *	(don't call it twice...)
 * Arguments:
 *	filename - file to get the directory path from.
 * Returns:
 *	A static string containing the directory path of the file path.
 *
 ****************************************************************************/

const char *
AmGetPath( const char	*filename)
{
    static char path[ PATHLEN];
    char        *lastCharPtr;

    if ( filename == NULL || !AmAbsolutePath( filename)) {
	return( NULL);
    } else if ( filename[ 1] == '\0') {	/* can't chop '/' */
	return( NULL);
    } else {
        strcpy( path, filename);
	lastCharPtr = &path[ strlen( path) - 1];
        if ( AmPathDelim( *lastCharPtr == '/')) {
	    return( path);
        } else {
	    for ( ; !AmPathDelim( *lastCharPtr); lastCharPtr--) {};
            *( lastCharPtr + 1) = '\0';
            return( path);
        }
    }
}
