#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions for clipping lines and polygons */

#include <anvil/project.h>
#include <anvil/geometry.h>
#include <anvil/datafmt.h>

/****************************************************************************
 *
 * Description:
 *   	These clipping routines model lines as parametric equations to 
 *   	reduce the number of computations (which is why they make no sense).
 * Arguments:
 *   	denom can be (-,+)(x,y,z)
 *  	num is (x,y,z)0 distance to (x,y,z)(min,max)
 *   	tE, tL are from 0 to 1 (t for the parametric curves x(t), y(t), z(t)
 * Returns:
 *	TRUE if the either point is inside, FALSE if they're both outside.
 *
 ****************************************************************************/

static AnvilBoolean 
Clipt( float denom, 
       float num, 
       float *tE, 
       float *tL)
{
    float t;

    if ( denom > 0.0) {		
	t = num / denom;
	if ( t > *tL) {
	    return( FALSE);		/* both outside */
	} else if ( t > *tE) {
	    *tE = t;			/* t0 outside, t1 inside */
	}
    } else if ( denom < 0.0) {
	t = num / denom;	
	if ( t < *tE) {
	    return( FALSE);		/* both outside */
	} else if ( t < *tL) {
	    *tL = t;			/* t0 inside, t1 outside */
	}
    } else {
	if ( num > 0.0) {
	    return( FALSE);		/* parallel, outside */
	}
    }
    return( TRUE);			/* if there's ANY part inside */
}

/****************************************************************************
 *
 * Description:
 * 	clip a line against a rectangle using a parametric line 
 *	representation.
 * Arguments:
 *	x0, y0, x1, y1 - line to clip.
 *	xmin, ymin, xmax, ymax - extents of the region to clip.
 * Returns:
 *	TRUE if any part of the line is inside the bounds, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean AmClip2d( float 	*x0,
		float	*y0,
		float	*x1,
		float	*y1,
		float	xmin,
		float	ymin,
		float	xmax,
		float	ymax)
{
    float 	tE, tL, dx, dy, val;

/* find slopes and check for trivial case */

    dx = *x1 - *x0;
    dy = *y1 - *y0;
    if (( dx == 0) && ( dy == 0)) {
	return(( *x0 >= xmin) && ( *x0 <= xmax) && 
	       ( *y0 >= ymin) && ( *y0 >= ymax)); 
    } 
    tE = 0.0;
    tL = 1.0;

/* for some reason, float conversion is find for assignment, but
   not argument passing */

    val = xmin - *x0;
    if ( Clipt( dx, val, &tE, &tL)) {		/* inside left */
	val = *x0 - xmax;
	if ( Clipt( -dx, val, &tE, &tL)) {	/* in right */
	    val = ymin - *y0;
	    if ( Clipt( dy, val, &tE, &tL)) {	/* above bottom */
		val = *y0 - ymax;
		if ( Clipt( -dy, val, &tE, &tL)) {    /* below top */
		    if ( tL <= 1.0) {
			*x1 = *x0 + tL*dx;	/* new x1, y1 */
			*y1 = *y0 + tL*dy;
		    } 
		    if ( tE > 0.0) {
			*x0 += tE*dx;		/* new x0, y0 */
			*y0 += tE*dy;
		    }
		    return( TRUE);
		} 
	    } 
	} 
    } 
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	clip a polygon against an extents rectangle.  Since clipping
 *	a polygon can actually result in more points than the source
 *	polygon had, it generates a new array.  The tmpPts array is
 *	passed in as a temporary holder.  Since this gets called over
 *	and over again in a loop, it's better to do the allocations
 *	at the start.  The clip routine works by clipping the polygon
 *	by each edge consecutively.
 * Arguments:
 *	pts - points in the polygon to clip.
 *	tmpPts - temporary point holder.
 *	resPts - resulting clipped polygon.
 *	xmin, ymin, xmax, ymax - extents to clip against.
 *	numpts - number of points in the original polygon, returns the
 *	number of points in the clipped polygon.
 * Returns:
 *	TRUE if any points in the polygon are inside the clipping extents,
 *	FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmClipPoly( AmCoordPtr	pts,
	  AmCoordPtr	tmpPts,
	  AmCoordPtr	resPts,
	  float		xmin,
	  float		ymin,
	  float		xmax,
	  float		ymax,
	  int		*numpts)
{
    int		ii, jj, ptindex = 0;
    float	dx, dy, tE = 0.0, tL = 1.0;
    
/* now walk down each edge.  
   If the line is clipped, then the copy flag of its second point
   is set, since it creates a new line.  If the line is entirely inside, then
   the status quo remains.  If the line is entirely outside, then the copy
   flag for the AmFirst point is set to 0 */

    jj = *numpts - 1;
    for ( ii = 0; ii < *numpts; ii++) {
	dx = pts[ ii].x - pts[ jj].x;
	if ( Clipt( dx, xmin - pts[ jj].x, &tE, &tL)) {
	    if ( tE > 0.0) {
                dy = pts[ ii].y - pts[ jj].y;
		tmpPts[ ptindex].x = pts[ jj].x + tE*dx;
		tmpPts[ ptindex++].y = pts[ jj].y + tE*dy;
		tE = 0.0;
	    } 
	    if ( tL <= 1.0) {
                dy = pts[ ii].y - pts[ jj].y;
		tmpPts[ ptindex].x = pts[ jj].x + tL*dx;
		tmpPts[ ptindex++].y = pts[ jj].y + tL*dy;
		tL = 1.0;
	    } 
	}
	jj = ii;
    }

/* do the right edge and writing into the second point buffer */

    *numpts = ptindex;
    jj = *numpts - 1;
    ptindex = 0;
    for ( ii = 0; ii < *numpts; ii++) {
	dx = tmpPts[ ii].x - tmpPts[ jj].x;
	if ( Clipt( -dx, tmpPts[ jj].x - xmax, &tE, &tL)) {
	    if ( tE > 0.0) {
                dy = tmpPts[ ii].y - tmpPts[ jj].y;
		resPts[ ptindex].x = tmpPts[ jj].x + tE*dx;
		resPts[ ptindex++].y = tmpPts[ jj].y + tE*dy;
		tE = 0.0;
	    } 
	    if ( tL <= 1.0) {
                dy = tmpPts[ ii].y - tmpPts[ jj].y;
		resPts[ ptindex].x = tmpPts[ jj].x + tL*dx;
		resPts[ ptindex++].y = tmpPts[ jj].y + tL*dy;
		tL = 1.0;
	    } 
	}
	jj = ii;
    }
/* now clip against the horizontal edges - bottom */
	    
    *numpts = ptindex;
    jj = *numpts - 1;
    ptindex = 0;
    for ( ii = 0; ii < *numpts; ii++) {
	dy = resPts[ ii].y - resPts[ jj].y;
	if ( Clipt( dy, ymin - resPts[ jj].y, &tE, &tL)) {
	    if ( tE > 0.0) {
                dx = resPts[ ii].x - resPts[ jj].x;
		tmpPts[ ptindex].x = resPts[ jj].x + tE*dx;
		tmpPts[ ptindex++].y = resPts[ jj].y + tE*dy;
		tE = 0.0;
	    } 
	    if ( tL <= 1.0) {
                dx = resPts[ ii].x - resPts[ jj].x;
		tmpPts[ ptindex].x = resPts[ jj].x + tL*dx;
		tmpPts[ ptindex++].y = resPts[ jj].y + tL*dy;
		tL = 1.0;
	    } 
	}
	jj = ii;
    }

/* clip against top */

    *numpts = ptindex;
    jj = *numpts - 1;
    ptindex = 0;
    for ( ii = 0; ii < *numpts; ii++) {
	dy = tmpPts[ ii].y - tmpPts[ jj].y;
	if ( Clipt( -dy, tmpPts[ jj].y - ymax, &tE, &tL)) {
	    if ( tE > 0.0) {
                dx = tmpPts[ ii].x - tmpPts[ jj].x;
		resPts[ ptindex].x = tmpPts[ jj].x + tE*dx;
		resPts[ ptindex++].y = tmpPts[ jj].y + tE*dy;
		tE = 0.0;
	    } 
	    if ( tL <= 1.0) {
                dx = tmpPts[ ii].x - tmpPts[ jj].x;
		resPts[ ptindex].x = tmpPts[ jj].x + tL*dx;
		resPts[ ptindex++].y = tmpPts[ jj].y + tL*dy;
		tL = 1.0;
	    } 
	}
	jj = ii;
    }
    *numpts = ptindex;
    return( *numpts > 0);
}
