#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <anvil/project.h>
#include <anvil/colorhist.h>

int AmOnIndex( AmHistPtr    a,
             AmHistPtr    b)
{
    return( a->index - b->index);
}

/* largest histogram values AmFirst */

int AmBigHistFirst( AmHistPtr       a,
                  AmHistPtr       b)
{
    return( b->hist - a->hist);
}

/* reduce a pixel from a source to the destination resolution */

unsigned long AmReducePixel( ulong_t      pixel)
{
    uchar_t     *ptr;
    ulong_t     red, green, blue;

    ptr = ( uchar_t *) &pixel;
    ptr++;
    blue = (( ulong_t) *ptr++ >> ( SRC_RES - DST_RES)) & DST_MASK;
    green = (( ulong_t) *ptr++ >> ( SRC_RES - DST_RES)) & DST_MASK;
    red = (( ulong_t) *ptr++ >> ( SRC_RES - DST_RES)) & DST_MASK;
    return(( blue << ( DST_RES*2)) | ( green << DST_RES) | red);
}

void AmDecomposeReducedPixel( ulong_t     pixel,
                            uchar_t     *red,
                            uchar_t     *green,
                            uchar_t     *blue)
{
    ulong_t     redlong, greenlong, bluelong;

    redlong = (pixel & DST_MASK) << LOSS;
    greenlong = ((pixel >> DST_RES) & DST_MASK) << LOSS;
    bluelong = ((pixel >> DST_RES*2) & DST_MASK) << LOSS;
    *red = redlong;
    *green = greenlong;
    *blue = bluelong;
}

/* find the closest color in the histogram to a specified pixel */

int AmClosestColorIndex( AmHistPtr          hist,
                       AmColorCellPtr     color,
                       int              numHist)
{
    int         colordist = 768, ii, canddist, closest = 0;

    for ( ii = 0; ii < numHist; ii++) {
        canddist = ( int) ( 0.33*abs( hist[ ii].color.red - color->red) +
                            0.5*abs( hist[ ii].color.green - color->green) +
                            0.17*abs( hist[ ii].color.blue - color->blue));
        if ( canddist < colordist) {
            colordist = canddist;
            closest = ii;
        }
    }
    return( closest);
}

/* upgrade an image from 8 to 24 pixels */

void AmImage8to24( uchar_t         *imageData,
		 int             width,
		 int             height,
		 AmColorCellPtr    colors,
		 int             numColors,
		 ulong_t         **newImageData)
{
    int 	ii;
    uchar_t	*p;
    ulong_t	*ptr;

    *newImageData = ( ulong_t *) malloc( width*height*sizeof( ulong_t));
    for ( ii = 0, ptr = *newImageData; ii < width*height; ii++, ptr++) {
	*ptr = 0;			/* to prevent UMR's */
        p = ( uchar_t *) ptr;
        ASSIGN_PIXEL( colors[ imageData[ ii]], p);
    }
}

/* downgrade a 24 bit image to 8 bit */

void AmImage24to8( ulong_t        *imageData,
                 int            width,
                 int            height,
                 uchar_t        **newImageData,
                 AmColorCellPtr   *colors,
                 int            *numColors)     /* passed and returned */
{
    AmHistRec             histogram[ NUMHIST];
    AmColorAssignRec      colormap [ NUMHIST];
    int                 ii, end, index[ NUMHIST];

/* generate a histogram of the pixel shifted to 32768 colors */

    for ( ii = 0; ii < NUMHIST; ii++) {
        histogram[ ii].hist = 0;
        AmDecomposeReducedPixel(( ulong_t) ii,
                              &histogram[ ii].color.red,
                              &histogram[ ii].color.green,
                              &histogram[ ii].color.blue);
        histogram[ ii].index = ii;
    }

/* stuff the histogram */

    for ( ii = 0; ii < width*height; ii++) {
        histogram[ AmReducePixel( imageData[ ii])].hist++;
    }
    end = 0;
    for ( ii = 0; ii < NUMHIST; ii++) {
        index[ ii] = end;
        if ( histogram[ ii].hist > 0) {
            histogram[ end] = histogram[ ii];
            end++;
        }
    }
    if ( end < *numColors) {
        *numColors = end;
    }

/* sort the array by the most popular colors, assign the indexes, and
   set the rest of the histogram indexes to the closest popular color */

    qsort( histogram, end, sizeof( AmHistRec), ( AmSearchFn) AmBigHistFirst);
    for ( ii = 0; ii < *numColors; ii++) {
        histogram[ ii].cmapIndex = ii;
        colormap[ ii].assigned = TRUE;
        colormap[ ii].color = histogram[ ii].color;
    }
    for ( ii = *numColors; ii < end; ii++) {
        histogram[ ii].cmapIndex =
            AmClosestColorIndex( histogram, &histogram[ ii].color, *numColors);
    }

/* sort the array into the order indexed by the index array */

    qsort( histogram, end, sizeof( AmHistRec), ( AmSearchFn) AmOnIndex);
    *newImageData = ( uchar_t *) malloc( width*height);

/* assign the image pixels to the colormap index */

    for ( ii = 0; ii < width*height; ii++) {
        ( *newImageData)[ ii] = ( uchar_t)
            histogram[ index[ AmReducePixel( imageData[ ii])]].cmapIndex;
    }

/* copy the colors from the histogram */

    *colors = ( AmColorCellPtr) malloc( sizeof( AmColorCellRec)*( *numColors));
    for ( ii = 0; ii < *numColors; ii++) {
        ( *colors)[ ii] = histogram[ ii].color;
    }
}

