#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <anvil/project.h>
#include <anvil/files.h>
#include <anvil/strlib.h>
#include <anvil/colorcell.h>

/* 
   utility functions for reading ASCII files containing colormaps into 
   vectors of color cells.
 
   colors can be specified either as names hex values or as RGB 0..255 values
 
    numColors
    lowest color
      .
      .
    color
      .
    highest color
*/

/****************************************************************************
 *
 * Description:
 * 	read the colors in a color file into a vector of color cells to 
 *	allocate.
 * Arguments:
 *	colorfile - ASCII file containing RGB values.
 *	numColors - returned number of colors in the file.
 *	colorRet - pointer to the vector of returned colors.
 * Returns:
 *	TRUE on successful read, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmReadColors( char 		*colorfile,
	    int 		*numColors,
	    AmColorCellPtr 	*colorRet)
{
    int 		kk, red, green, blue;
    AmColorCellPtr 	colorptr;
    FILE 		*fp;
    char 		string[ 256], *namePtr, *p;

    if (( fp = fopen( colorfile, "r")) == NULL) {
        return( FALSE);
    }

/* find out if he put the number of colors at the head of the file.  If so,
   use it, otherwise, read the number of lines to determine the number of 
   colors */
    
    if ( fgets( string, sizeof( string), fp) == NULL) {
	fclose( fp);
	return( FALSE);
    }
    string[ strlen( string) - 1] = '\0';

/* make sure it's a number followed by white space and a newline */

    p = AmElimLeadWspace( string);
    for ( ; ( *p != '\0') && isdigit( *p); p++);
    for ( ; ( *p != '\0') && isspace( *p); p++);

/* AmFirst line is color -> no length info. count lines for number of colors */ 

    if ( *p != 0) {
	*numColors = AmNumLines( colorfile);
	rewind( fp);
    } else if ( sscanf( string, "%d\n", numColors) != 1) {
	fclose( fp);
        return( FALSE);
    }
    *colorRet = colorptr = 
	( AmColorCellPtr) calloc( sizeof( AmColorCellRec), *numColors);

/* read the colors from the file and save them in sampleSet */

    for ( kk = *numColors; kk; kk--) {
	colorptr->pixel = kk;
        if ( fgets( string, sizeof( string), fp) == NULL) {
	    fclose( fp);
	    *numColors = *numColors - kk;
	    return( TRUE);			/* we did read SOME colors */
	}
		    
/* crop the \n and move up the white space */

        string[ strlen( string) - 1] = '\0';
	namePtr = AmElimLeadWspace( string);

/* allocate any colors specified by RGB values */

        if( isdigit( *namePtr)) {
	    if ( sscanf( namePtr, "%d %d %d", &red, &green, &blue) != 3) {
                ( *numColors)--;
	    } else {
		colorptr->red = red;
		colorptr->green = green;
		colorptr->blue = blue;
		colorptr++;
	    }
        }
    }
    fclose( fp);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *	Given a string containing sets of 3-tuple RGB values, scan it into 
 *	a colorcell vector.
 * Arguments:
 *	str - string to convert.
 * Returns:
 *	vector of color cells, or NULL if there was an allocation problem.
 * Modifications:
 *	The algorithm is overly complicated, and the function should return
 *	the number of color cells read.
 *
 ****************************************************************************/

AmColorCellPtr 
AmReadColorCellStr( char *str)
{   
    int         	ii, length, *intvec;
    char        	*p;
    AmColorCellPtr	colorVec;

/* scan over each digit to find the length don't forget negatives */

    length = 0;
    for ( p = str; *p != '\0'; ) {
        while ( isdigit( *p++));
        while (( *p != '\0') &&  !isdigit( *p)) {
            p++;
        }
        length++;
    }

/* read the string into a vector of integers. */

    if (( intvec = AmReadIntStrVec( str)) == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }

/* allocate the vector of RGB values */

    length /= 3;
    if (( colorVec = ( AmColorCellPtr) 
		malloc( sizeof( AmColorCellRec)*length)) == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    for ( ii = 0; ii < length; ii++) {
	colorVec[ ii].red = intvec[ ii*3];
	colorVec[ ii].green = intvec[ ii*3 + 1];
	colorVec[ ii].blue = intvec[ ii*3 + 2];
    }
    free( intvec);
    return( colorVec);
}

/****************************************************************************
 *
 * Description:
 * 	interpolate AmBetween two colors based on a fraction value.
 * Arguments:
 *	frac - the weight to interpolate.
 *	color1, color2 - the two colors to interpolate.
 *	retColor - returned interpolated color.
 *
 ****************************************************************************/

void 
AmBetweenColorcell( double	frac,
		  AmColorCellPtr	color1,
		  AmColorCellPtr	color2,
		  AmColorCellPtr	retColor)
{
    retColor->red = 
	( unsigned char) (( 1.0 - frac)*color1->red + frac*color2->red);
    retColor->blue = 
	( unsigned char) (( 1.0 - frac)*color1->blue + frac*color2->blue);
    retColor->green = 
	( unsigned char) (( 1.0 - frac)*color1->green + frac*color2->green);
}

/****************************************************************************
 *
 * Description:
 * 	given a set of colorcells, stretch them to a new set of colorcells.
 * Arguments:
 *	colors - colors to interpolate.
 *	numColors - number of colors.
 *	newColors - newly allocated vector of color cells.
 *	numNewColors - number of allocated color cells.
 * Modifications:
 *	This should return an error code for the malloc failure.
 *
 ****************************************************************************/
 
void 
AmInterpolateColorCells( AmColorCellPtr   colors,
		       int            numColors,
		       AmColorCellPtr   *newColors,
		       int            numNewColors)
{
    double      realIndex, step;
    int         ii, intIndex;
 
    *newColors = ( AmColorCellPtr) calloc( sizeof( AmColorCellRec), numNewColors);
    if ( *newColors == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return;
    }
    ( *newColors)[ 0] = colors[ 0];
    ( *newColors)[ numNewColors - 1] = colors[ numColors - 1];
    step = (( double) numColors - 1)/(( double) numNewColors - 1);
    for ( ii = 1, realIndex = step;
          ii < numNewColors - 1;
          ii++, realIndex += step) {
        intIndex = ( int) floor( realIndex);
        AmBetweenColorcell( realIndex - intIndex,
			  &(colors[ intIndex]), &(colors[ intIndex + 1]),
			  &((*newColors)[ ii]));
    }
}

/****************************************************************************
 *
 * Description:
 * 	take a set of colors and apply shades to them. 
 * Arguments:
 *	colors - colors to shade.
 *	shades - resulting vector of shades (should be allocated to 
 *	numColors*numShades)
 *	numColors - number of colors in the colors array.
 *	numShades - number of shades to create.
 *
 ****************************************************************************/

void AmShadeColors( AmColorCellPtr	colors,
		  AmColorCellPtr	shades,
		  int		numColors,
		  int		numShades)
{
    int	ii, jj;

    for ( ii = 0; ii < numShades; ii++) {
	for ( jj = 0; jj < numColors; jj++) {
	    shades[ ii*numColors + jj].red =
		(colors[ jj].red*ii + 1)/numShades;
	    shades[ ii*numColors + jj].green =
		(colors[ jj].green*ii + 1)/numShades;
	    shades[ ii*numColors + jj].blue =
		(colors[ jj].blue*ii + 1)/numShades;
	}
    }
}
