#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/description.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/list.h>

/* function for parsing a binary file description, which is used for 
   converting data from big-endian to little-endian format by swapper.
   These are used in conjunction with the lexer descr.l and the parser
   descr.y */

/* symbol table is a vector of strings and associated types with a length.
   The file description fp has to be static because yyparse() blows, and
   statTypeList is static because yyparse() can't give a return value */

static const int 	NUMSYMS = 200;
static AmSymbolRec  	symbolTable[ NUMSYMS];
static int	  	numSymbols;
static FILE 	  	*descrFp;
static AmListPtr	statTypeList;
static AnvilBoolean 		errorFlag;

/* for tracking the line #'s for errors */

int AmDescrErrorLineNo;
char AmDescrErrorString[ 512];

/* yacck */

int descrparse( void);
void descrrestart( FILE *);

/****************************************************************************
 *
 * Description:
 * 	add an identifier to the symbol table.
 * Arguments:
 *	name - string name to add to the symbol table.
 *	type - type to add.
 *
 ****************************************************************************/

static void 
addSymbol( char		*name,
	   AmTypePtr	type)
{
    symbolTable[ numSymbols].name = name;
    symbolTable[ numSymbols].thing = type;
    numSymbols++;
}

/****************************************************************************
 *
 * Description:
 * 	retreive a symbol from the symbol table. Go from the back of the 
 *	symbol table so the AmLast reference is retreived.
 * Arguments:
 *	name - string to search for in the symbol table.
 * Returns:
 *	name with the associated type (symbol)
 *
 ****************************************************************************/

static AmTypePtr 
AmGetSymbol( char  *name)
{
    int	ii;

    for ( ii = numSymbols - 1; ii >= 0; ii--) {
	if ( !strcmp( symbolTable[ ii].name, name)) {
	    return( symbolTable[ ii].thing);
	}
    }
    return( NULL);
}

/****************************************************************************
 *
 * Description:
 * 	make a simple type from the type enumeration. 
 * Arguments:
 *	typeName - simple type name.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

static AmTypePtr 
makeSimple( char *typeName)
{
    AmTypePtr	type;
 
    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_SIMPLE;
    type->name = NULL;
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.simpleType.typeName = strdup( typeName);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	make a reference type from a string.
 * Arguments:
 *	reference - reference name.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

static AmTypePtr 
makeReference( char *reference)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_REFERENCE;
    type->name = NULL;
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.referenceType.typeName = strdup( reference);
    type->u.referenceType.type = NULL;
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	make a basic type reference, whether it be a simple type or
 *  	a reference to another type, depending on if it's a simple
 *  	type keyword 
 * Arguments:
 *	name - type name.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

static AmTypePtr 
makeBasicType( char *name)
{
    AmDataTypeEnum	dataType;

    dataType = AmStrToDataTypeEnum( name);
    if ( dataType != AmNO_TYPE) {
	return( makeSimple( name));
    } else {
	return( makeReference( name));
    }
}

/****************************************************************************
 *
 * Description:
 * 	make a scalar type from a typeName and a string name. 
 * Arguments:
 *	typeName - name of the type.
 *	name - string name identifier.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

AmTypePtr 
AmMakeScalar( char	*typeName,
	    char 	*name)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_SCALAR;
    type->name = strdup( name);
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.scalarType.type = makeBasicType( typeName);
    addSymbol( name, type);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	create a static array declaration from the type name, identifier
 *	name and array length.
 * Arguments:
 *	typeName - name of the type.
 *	name - string name identifier.
 *	length - length of the static array.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

AmTypePtr 
AmMakeStatic( char	*typeName,
	      char 	*name,
	      char	*length)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_STATIC_ARRAY;
    type->name = strdup( name);
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.staticArrayType.type = makeBasicType( typeName);
    type->u.staticArrayType.length = atoi( length);
    addSymbol( name, type);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	create a dynamic array declaration from the type name, identifier
 *	name and array length identifier name.
 * Arguments:
 *	typeName - name of the type.
 *	name - string name identifier.
 *	lengthname - name of the identifier containing the length of the 
 *	static array.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

AmTypePtr 
AmMakeDynamic( char	*typeName,
	     char 	*name,
	     char 	*lengthname)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_DYNAMIC_ARRAY;
    type->name = strdup( name);
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.dynamicArrayType.lengthname = strdup( lengthname);
    type->u.dynamicArrayType.lengthType = NULL;
    type->u.dynamicArrayType.type = makeBasicType( typeName);
    addSymbol( name, type);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	make an EOF array type, which is defined as an array that 
 *	continues to the end of the file.
 * Arguments:
 *	typeName - name of the type.
 *	name - string name identifier.
 * Returns:
 *	created symbol, or NULL if it couldn't be malloced.
 *
 ****************************************************************************/

AmTypePtr 
AmMakeEof( char	*typeName,
	 char 	*name)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    if ( type == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    type->tag = AmTYPE_EOF_ARRAY;
    type->name = strdup( name);
    type->typedefFlag = FALSE;
    type->u.eofArrayType.type = makeBasicType( typeName);
    addSymbol( name, type);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	add the field to the structures's list of fields.
 * Arguments:
 *	decl - declaration to add.
 *	declList - list of field declarations.
 * Returns:
 *	new list of declarations.
 *	
 ****************************************************************************/

AmListPtr 
AmAddDecl( AmTypePtr 	decl, 
	 AmListPtr	declList)
{
    declList = AmInsertNodeAtEndLst( decl, declList);
    return( declList);
}

/****************************************************************************
 *
 * Description:
 * 	add the field to the list of types 
 * Arguments:
 *	type - type to add
 *	typeList - type list to add it to.
 * Returns:
 *	new list of types.
 *	
 ****************************************************************************/

AmListPtr 
AmAddType( AmTypePtr 	type, 
	 AmListPtr	typeList)
{
    if ( typeList == NULL) {
	typeList = AmInsertNodeLst( type, typeList);
	statTypeList = typeList;
    } else {
	typeList = AmInsertNodeAtEndLst( type, typeList);
    }
    return( typeList);
}

/****************************************************************************
 *
 * Description:
 * 	Given the structure name and list of fields make a structure 
 *	declaration.
 * Arguments:
 *	name - name of the structure.
 *	fieldList - list of structure fields.
 * Returns:
 *	new list of types.
 *	
 ****************************************************************************/

AmTypePtr 
AmMakeStruct( char	*name,
	    AmListPtr	fieldList)
{
    AmTypePtr	type;

    type = ( AmTypePtr) malloc( sizeof( AmTypeRec));
    type->tag = AmTYPE_STRUCT;
    type->name = strdup( name);
    type->typedefFlag = FALSE;
    type->lineNo = AmDescrErrorLineNo;
    type->u.structType.numFields = AmLengthLst( fieldList);
    type->u.structType.fields = ( AmTypePtr *) AmMakeVectorLst( fieldList);
    addSymbol( name, type);
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	set the typedef flag in a type, so it's only used as 
 *	a reference type.
 * Arguments:
 *	type - type to create the typedef from.
 * Returns:
 *	the type with the flag set.
 *	
 ****************************************************************************/

AmTypePtr 
AmMakeTypedef( AmTypePtr	type)
{
    type->typedefFlag = TRUE;
    return( type);
}

/****************************************************************************
 *
 * Description:
 * 	inverse of isTypedef
 * Arguments:
 *	type - type to test.
 * Returns:
 *	FALSE if it's a typedef, TRUE if it's a declaration.
 *	
 ****************************************************************************/

static AnvilBoolean 
isNotTypedef( AmTypePtr	type)
{
    return( !type->typedefFlag);
}

/****************************************************************************
 *
 * Description:
 * 	take all the reference types in the parse tree and resolve them
 *	to their actual types.
 * Arguments:
 *	typeList - list of types to resolve.
 *	errorString - returned error string if there's a problem.
 * Returns:
 *	TRUE if all the types resolved OK, FALSE if there was an error.
 *	
 ****************************************************************************/

AnvilBoolean 
AmResolveTypeList( AmListPtr	typeList,
		 char		*errorString)
{
    AmTypePtr	type;

    for ( ; typeList != NULL; typeList = typeList->next) {
        type = ( AmTypePtr) typeList->data;
	if ( !AmResolveType( type, errorString)) {
	    return( FALSE);
	}
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	recursively resolve the subtypes within a type.
 * Arguments:
 *	type - type to resolve.
 *	errorString - error string returned if there was a problem.
 * Returns:
 *	TRUE if all the types resolved OK, FALSE if there was an error.
 *	
 ****************************************************************************/

AnvilBoolean 
AmResolveType( AmTypePtr	type,
	     char	*errorString)
{
    AmTypePtr	reference;
    int		ii;

    switch( type->tag) {
    case AmTYPE_SIMPLE:
	type->u.simpleType.primitiveType = 
	    AmStrToDataTypeEnum( type->u.simpleType.typeName);
	if ( type->u.simpleType.primitiveType == AmNO_TYPE) {
	    sprintf( errorString, "line %d: unknown type %s",
		     type->lineNo, type->u.simpleType.typeName);
	    return( FALSE);
	}
	break;
    case AmTYPE_REFERENCE:

/* resolve the type reference */

	reference = AmGetSymbol( type->u.referenceType.typeName);
	if ( reference == NULL) {
	    sprintf( errorString, "line %d: bad reference: %s",
		     type->lineNo, type->u.referenceType.typeName);
	    return( FALSE);
	}
	type->u.referenceType.type = reference;
	break;
    case AmTYPE_SCALAR:

/* resolve the type of the scalar */

	if ( !AmResolveType( type->u.scalarType.type, errorString)) {
	    return( FALSE);
	}
	break;
    case AmTYPE_STATIC_ARRAY:

/* resolve the array type */

	if ( !AmResolveType( type->u.staticArrayType.type, errorString)) {
	    return( FALSE);
	}
	break;
    case AmTYPE_DYNAMIC_ARRAY:

/* resolve the array type and array length */

	reference = AmGetSymbol( type->u.dynamicArrayType.lengthname);
	if ( reference == NULL) {
	    sprintf( errorString, "line %d: bad reference: %s",
		     type->lineNo, type->u.dynamicArrayType.lengthname);
	    return( FALSE);
	}
	type->u.dynamicArrayType.lengthType = reference;
	if ( !AmResolveType( type->u.dynamicArrayType.type, errorString)) {
	    return( FALSE);
	}
	break;
    case AmTYPE_EOF_ARRAY:

/* resolve the array type, since it is of infinite length */

	if ( !AmResolveType( type->u.eofArrayType.type, errorString)) {
	    return( FALSE);
	}
	break;
    case AmTYPE_STRUCT:

/* resolve each one of the struct field types */

	for ( ii = 0; ii < type->u.structType.numFields; ii++) {
	    if ( !AmResolveType( type->u.structType.fields[ ii], errorString)) {
	        return( FALSE);
	    }
	}
	break;
    default:
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	input function for the yacc() parser.
 * Arguments:
 *	buffer - buffer to read.
 *	maxSize - size of the buffer.
 * Returns:
 *	actual number of bytes read.
 *	
 ****************************************************************************/

int 
AmFpYYinput( char	*buffer,
	   int	maxSize)
{
    return( fread( buffer, 1, maxSize, descrFp));
}

/****************************************************************************
 *
 * Description:
 * 	take a description file and parse it into a type structure.
 * Arguments:
 *	file - file to parse.
 *	typeList - list of types to parse it into.
 *	errorString - error string if there was a problem parsing.
 * Returns:
 *	AmOK on success, AmOPEN_FAILED on open file error, 
 *	AmBAD_ARGUMENT on parse or type resolve error.
 *	
 ****************************************************************************/

AmErrorCode 
AmParseDescriptionFile( char 		*file,
		        AmListPtr	*typeList,
		        char		*errorString)
{
    if (( descrFp = fopen( file, "r")) == NULL) {
	sprintf( errorString, "failed to open %s", file);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    descrrestart( descrFp);
    AmDescrErrorLineNo = 1;
    errorFlag = FALSE;
    descrparse();
    fclose( descrFp);
    if ( errorFlag) {
	sprintf( errorString, AmDescrErrorString);
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    *typeList = statTypeList;
    if ( !AmResolveTypeList( *typeList, errorString)) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	does type a refer to type b?
 * Arguments:
 *	a - type to check the reference of.
 *	b - does a refer to this type?
 * Returns:
 *	TRUE if a references b, FALSE otherwise.
 *	
 ****************************************************************************/

static AnvilBoolean 
refersToType( AmTypePtr	a,
	      AmTypePtr	b)
{
    int	ii;

    switch ( a->tag) {
    case AmTYPE_REFERENCE:
	return( a->u.referenceType.type == b);
    case AmTYPE_SCALAR:
	return( refersToType( a->u.scalarType.type, b));
    case AmTYPE_STATIC_ARRAY:
	return( refersToType( a->u.staticArrayType.type, b));
    case AmTYPE_DYNAMIC_ARRAY:
	return( refersToType( a->u.dynamicArrayType.type, b));
    case AmTYPE_STRUCT:
	for ( ii = 0; ii < a->u.structType.numFields; ii++) {
	    if ( refersToType( a->u.structType.fields[ ii], b)) {
		return( TRUE);
	    }
	}
	return( FALSE);
    default:
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	scan the type list and return all of the types which are 
 *	declarations 
 * Arguments:
 *	typeList - list of types.
 * Returns:
 *	allocated list of types which are declarations 
 *	
 ****************************************************************************/

AmListPtr 
AmGetDeclarations( AmListPtr	typeList)
{
    return( AmFilterCopyLst( typeList, ( AmElementFn) isNotTypedef));
}

/****************************************************************************
 *
 * Description:
 * 	swap a pair of bytes.
 * Arguments:
 *	a, b - byte pointers.
 *	
 ****************************************************************************/

static void 
swapByte( uchar_t	*a,
	  uchar_t	*b)
{
    uchar_t	c;

    c = *a;
    *a = *b;
    *b = c;
}

/****************************************************************************
 *
 * Description:
 * 	swap an item based on its type, and return the number of bytes 
 *	swapped.
 * Arguments:
 *	type - declaration defining the data to be swapped.
 *	buffer - buffer containing data to swap.
 * Returns:
 *	number of bytes swapped.
 *	
 ****************************************************************************/

static int 
swapType( AmTypePtr	type,
	  uchar_t	*buffer)
{
    AmTypePtr		lengthType, lengthScalar;
    int			ii, fieldLen, arrayLength;
    uchar_t		*ptr;

    switch ( type->tag) {
    case AmTYPE_REFERENCE:
	return( swapType( type->u.referenceType.type, buffer));
    case AmTYPE_SCALAR:
	return( swapType( type->u.scalarType.type, buffer));
    case AmTYPE_STATIC_ARRAY:
	for ( ii = 0, ptr = buffer; 
	      ii < type->u.staticArrayType.length; 
	      ii++, ptr += swapType( type->u.staticArrayType.type, ptr));
	return( ptr - buffer);
    case AmTYPE_DYNAMIC_ARRAY:
	lengthType = type->u.dynamicArrayType.lengthType;
        lengthScalar = lengthType->u.scalarType.type;
        arrayLength = lengthScalar->u.simpleType.refValue;
	for ( ii = 0, ptr = buffer; ii < arrayLength;
	      ii++, ptr += swapType( type->u.dynamicArrayType.type, ptr));
	return( ptr - buffer);
    case AmTYPE_STRUCT:
	for ( ii = 0, ptr = buffer; ii < type->u.structType.numFields; ii++) {
	    fieldLen = swapType( type->u.structType.fields[ ii], ptr);

/* struct fields are padded to 4 byte boundaries */

	    if ( fieldLen % 4 != 0) {
		fieldLen = 4*( fieldLen/4 + 1);
	    }
	    ptr += fieldLen;
	}
	return( ptr - buffer);
    case AmTYPE_SIMPLE:
	switch ( type->u.simpleType.primitiveType) {
	case AmFLOAT:
	case AmINT:
	case AmLONG:
	case AmUNSIGNED_LONG:
	case AmUNSIGNED_INT:
	    swapByte( &buffer[ 0], &buffer[ 3]);
	    swapByte( &buffer[ 1], &buffer[ 2]);
	    return( 4);
	case AmDOUBLE:
	    swapByte( &buffer[ 0], &buffer[ 7]);
	    swapByte( &buffer[ 1], &buffer[ 6]);
	    swapByte( &buffer[ 2], &buffer[ 5]);
	    swapByte( &buffer[ 3], &buffer[ 4]);
	    return( 8);
	case AmCHAR:
	case AmUNSIGNED_CHAR:
	    return( 1);
	case AmSHORT:
	case AmUNSIGNED_SHORT:
	    swapByte( &buffer[ 0], &buffer[ 1]);
	    return( 2);
	default:
	    return( 0);
	}
    default:
	return( 0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	for array lengths, get the integer value from the type reference 
 * Arguments:
 *	type - declaration describing the integer value (in some reference)
 *	value - the value to scan.
 * Returns:
 *	the integer value, or 0 if it couldn't be parsed.
 *	
 ****************************************************************************/

static int 
getTypeInt( AmTypePtr	type, 
	    uchar_t	*value)
{
    if ( type->tag == AmTYPE_SIMPLE) {
	switch ( type->u.simpleType.primitiveType) {
	case AmINT:
	    return( *( int *) value);
	case AmUNSIGNED_INT:
	    return( *( uint_t *) value);
	case AmLONG:
	    return( *( long *) value);
	case AmUNSIGNED_LONG:
	    return( *( ulong_t *) value);
	case AmCHAR:
	    return( *( char *) value);
	case AmUNSIGNED_CHAR:
	    return( *( uchar_t *) value);
	default:
	    return( 0);
	}
    } else if ( type->tag == AmTYPE_REFERENCE) {
	return( getTypeInt( type->u.referenceType.type, value));
    } else if ( type->tag == AmTYPE_SCALAR) {
	return( getTypeInt( type->u.scalarType.type, value));
    } else {
	return( 0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	is a type an array type?
 * Arguments:
 *	type - type to test.
 * Returns:
 *	TRUE if it's an array type, FALSE otherwise.
 *	
 ****************************************************************************/

static AnvilBoolean 
arrayType( AmTypePtr	type)
{
    return(( type->tag == AmTYPE_STATIC_ARRAY) ||
	   ( type->tag == AmTYPE_DYNAMIC_ARRAY) ||
	   ( type->tag == AmTYPE_EOF_ARRAY));
}

/****************************************************************************
 *
 * Description:
 * 	read a type from a buffer,  return TRUE or FALSE depending on 
 *	whether the type was completely read.
 * Arguments:
 *	type - type declaration defining data to be parsed.
 *	buffer - buffer containing data.
 *	length - length of the buffer, returns number of unparsed bytes.
 *	swap - flag to control whether value should be swapped.
 * Returns:
 *	TRUE if the buffer was entirely parsed, FALSE otherwise.
 *	
 ****************************************************************************/

static AnvilBoolean 
readParseType( AmTypePtr	type, 
	       uchar_t	*buffer,
	       int	*length,
	       int	*numelParsed,
	       AnvilBoolean	swap)
{
    int			currLength, itemLength, readlen, arrayLength, ii,
			passNumelParsed = 0, numToRead;
    AmDataTypeEnum	dataType;
    AmTypePtr		lengthType, lengthScalar;
    uchar_t		*ptr, *bufend;

    switch( type->tag) {
    case AmTYPE_SIMPLE:

/* if it's a simple type, then scan the value if it's an integer, because it 
   might be used later as an array length */

	dataType = type->u.simpleType.primitiveType;
	type->u.simpleType.refValue = getTypeInt( type, buffer);
	readlen = AmDataTypeEnumToSize( dataType);
	if ( readlen > *length) {
	    return( FALSE);
	} else {
	    if ( swap) {
		swapType( type, buffer);
	    }
	    *length -= readlen;
	    return( TRUE);
	}
    case AmTYPE_REFERENCE:
	return( readParseType( type->u.referenceType.type, 
			       buffer, length, &passNumelParsed, swap));
    case AmTYPE_SCALAR:
	return( readParseType( type->u.scalarType.type, 
			       buffer, length, &passNumelParsed, swap));
    case AmTYPE_STATIC_ARRAY:

/* read and parse the elements one by one, tracking the number of elements
   and bytes read.  If the item can be read, then swap it, otherwise bail,
   and add the number of read elements to the number of elements parsed */

	arrayLength = type->u.staticArrayType.length;
	numToRead = arrayLength - *numelParsed;
	for ( ii = 0, ptr = buffer;
	      ii < numToRead;
	      ii++, ptr += itemLength) {
	    currLength = *length;
	    if ( readParseType( type->u.staticArrayType.type, 
				ptr, &currLength, &passNumelParsed, FALSE)) {
		if ( swap) {
		    swapType( type->u.staticArrayType.type, ptr);
		}
		itemLength = *length - currLength;
		*length -= itemLength;
	    } else {
		break;
	    }
	}
	*numelParsed += ii;
	return( ii == numToRead);
    case AmTYPE_DYNAMIC_ARRAY:

/* read and parse the elements one by one, tracking the number of elements
   and bytes read.  If the item can be read, then swap it, otherwise bail,
   and add the number of read elements to the number of elements parsed */

	lengthType = type->u.dynamicArrayType.lengthType;
	lengthScalar = lengthType->u.scalarType.type;
	arrayLength = lengthScalar->u.simpleType.refValue;
	numToRead = arrayLength - *numelParsed;
	for ( ii = 0, ptr = buffer;
	      ii < numToRead;
	      ii++, ptr += itemLength) {
	    currLength = *length;
	    if ( readParseType( type->u.dynamicArrayType.type, 
				ptr, &currLength, &passNumelParsed, FALSE)) {
		if ( swap) {
		    swapType( type->u.dynamicArrayType.type, ptr);
		}
		itemLength = *length - currLength;
		*length -= itemLength;
	    } else {
		break;
	    }
	}
	*numelParsed += ii;
	return( ii == numToRead);
    case AmTYPE_EOF_ARRAY:

/* a shorthand notation for saying "this type of data to the end of the file" */

	bufend = buffer + *length;
	for ( ii = 0, ptr = buffer; ptr < bufend; ii++, ptr += itemLength) {
	    currLength = *length;
	    if ( readParseType( type->u.eofArrayType.type, 
				ptr, &currLength, &passNumelParsed, FALSE)) {
		if ( swap) {
		    swapType( type->u.eofArrayType.type, ptr);
		}
		itemLength = *length - currLength;
		*length -= itemLength;
	    } else {
		break;
	    }
	}
	*numelParsed += ii;

/* always return true, because we're reading to the end of the file */

	return( TRUE);
    case AmTYPE_STRUCT:

/* parse each of the fields in the struct, and bail if any can't be read,
   Note the FALSE swap flag again */

	for ( ii = 0, ptr = buffer; 
	      ii < type->u.structType.numFields; 
	      ii++, ptr += itemLength) {
	    currLength = *length;
	    if ( !readParseType( type->u.structType.fields[ ii],
				 ptr, &currLength, &passNumelParsed, FALSE)) {
		return( FALSE);
	    }

/* struct fields are padded to word boundaries, so we round to 4 bytes */

	    itemLength = *length - currLength;
	    if ( itemLength % 4 != 0) {
		itemLength = 4*( itemLength/4 + 1);
	    }
	    *length -= itemLength;
	}

/* if all the struct fields were read, we now have to swap them */

	if ( swap) {
	    swapType( type, buffer);
	}
	return( TRUE);
    default:
	AN_DEBUG( 1, ("can't find tag %d\n", type->tag));
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	read and parse a buffer into a list of types.
 * Arguments:
 *	readType - list of type declarations defining data to be parsed.
 *	buffer - buffer containing data.
 *	length - length of the buffer. 
 *	numelParsed - returned number of elements parsed.
 *	swap - flag to control whether value should be swapped.
 * Returns:
 *	number of unparsed bytes 
 *	
 ****************************************************************************/

int 
AmReadParseTypeList( AmReadTypePtr	readType,
		   uchar_t	*buffer,
		   int		length,
		   int		*numelParsed,
		   AnvilBoolean	swap)
{
    int		currLength;
    uchar_t	*ptr, *bufend = buffer + length;
    AmListPtr	currTypePtr;

/* if we have multiple types, walk down the list of types, if we just have
   one type, then we figure out the length of that and read as many as we can,
   note the FALSE swap flag, so the item doesn't get swapped twice. 
   Note the numelParsed parameter, which if the top level structure is
   an array, tracks the number of elements read, so we read the correct
   number of elements in the next pass */
    
    if ( readType->typeList->next == NULL) {
	currLength = length;
	readParseType(( AmTypePtr) readType->currTypePtr->data, buffer, 
		      &length, numelParsed, swap);

/* return the number of bytes we have to back up to get to the head of the
   AmLast structure */

	return( length);
    } else {
	for ( currTypePtr = readType->currTypePtr, ptr = buffer;
	      currTypePtr != NULL; 
	      currTypePtr = currTypePtr->next, ptr += currLength - length) {
	    currLength = length;

/* read and parse each individual type, and walk down the buffer as they're
   being read */

	    if ( !readParseType(( AmTypePtr) currTypePtr->data, 
				ptr, &length, numelParsed, swap)) {

/* if it's an array, then we mark the bytes as read */

		if ( arrayType(( AmTypePtr) currTypePtr->data)) {
		    ptr += currLength - length;
		}
		break;
	    }
	}
	readType->currTypePtr = currTypePtr;
	return( bufend - ptr);
    }
}

/****************************************************************************
 *
 * Description:
 * 	recursive function to print the type structure 
 * Arguments:
 *	type - type structure to print.
 *	
 ****************************************************************************/

void 
AmPrintType( AmTypePtr	type)
{
    int	ii;

    switch ( type->tag) {
    case AmTYPE_SIMPLE:
	printf( "%s ", type->u.simpleType.typeName);
	break;
    case AmTYPE_REFERENCE:
	AmPrintType( type->u.referenceType.type);
	break;
    case AmTYPE_SCALAR:
	AmPrintType( type->u.scalarType.type);
	printf( "%s;\n", type->name);
	break;
    case AmTYPE_STATIC_ARRAY:
	AmPrintType( type->u.staticArrayType.type);
	printf( "%s[ %d];\n", 
		type->name,
		type->u.staticArrayType.length);
	break;
    case AmTYPE_DYNAMIC_ARRAY:
	AmPrintType( type->u.dynamicArrayType.type);
	printf( "%s[ %s];\n", 
		type->name,
		type->u.dynamicArrayType.lengthname);
	break;
    case AmTYPE_EOF_ARRAY:
	AmPrintType( type->u.eofArrayType.type);
	printf( "%s[];\n", type->name);
	break;
    case AmTYPE_STRUCT:
	printf( "struct %s {\n", type->name);
	for ( ii = 0; ii < type->u.structType.numFields; ii++) {
	    printf( "\t");
	    AmPrintType( type->u.structType.fields[ ii]);
	}
	printf( "};\n");
	break;
    default:
        break;
    }
}

/****************************************************************************
 *
 * Description:
 * 	print a list of types.
 * Arguments:
 *	typeList - typeList structure to print.
 *	
 ****************************************************************************/

void 
AmPrintDescription( AmListPtr	typeList)
{
    for ( ; typeList != NULL; typeList = typeList->next) {
	AmPrintType(( AmTypePtr) typeList->data);
    }
}

/****************************************************************************
 *
 * Description:
 * 	wrapper error function for lexer. 
 *	
 ****************************************************************************/

int descrerror( const char *errorString)
{
    fprintf( stderr, "parse error: %s at line %d\n", 
	     errorString, AmDescrErrorLineNo);
    errorFlag = TRUE;
    return( TRUE);
}
