#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


/* description:
   parsing and retreival functions for the enumerations database, usually
   stored in $ANVIL_CONFIG/lookups/AnvilEnumLookup.  The parsing functions
   are called from enums.y. 
   The AnvilEnumLookup file serves two purposes: AmFirst, it maps enumeration
   strings to enumeration values by creating an enumeration class, then
   defining the mapping of strings to values in the class.  It also
   maps tags from <anvil/tagstrs.h> to enumeration classes.

   "blotz" is a tag in <anvil/tagstrs.h>

   blotz.enumClass:		blotzEnums
   blotzEnums.frob:		1
   blotzEnums.foo:		2
   blotzEnums.retch:		3

   It also defines the subdirectories in the $ANVIL_CONFIG directory to 
   read data from for the different tags from <anvil/tagstrs.h> which 
   can read file data.   Each of these declarations are of the form:

   tagName.subdir:		subdirectory-name
*/

#include <string.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <anvil/tsearch.h>
#include <anvil/project.h>
#include <anvil/strlib.h>
#include <anvil/enums.h>
#include <anvil/enumsP.h>

/* mapping from tag names to enumeration classes */

typedef struct enumTagClass_ {
    char	*tagName;
    char	*className;
} enumTagClassRec, *enumTagClassPtr;

/* tree for enumeration classes containing search tables by enumeration
   name and enumeration value */

typedef struct enumClass_ {
    char	*className;
    void	*enumNameRoot;
    void	*enumValRoot;
} enumClassRec, *enumClassPtr;

/* enumeration name/value definition */

typedef struct enumDef_ {
    char	*enumName;
    int		enumVal;
} enumDefRec, *enumDefPtr;

/* subdirectory for tag definition */

typedef struct subdirDef_ {
    char	*tagName;
    char	*subdirName;
} subdirDefRec, *subdirDefPtr;

/* casting for twalk() */

typedef void ( *AmWalkFn)( const void *, VISIT order, int);

FILE			*enumsFp;
static AnvilBoolean		errorFlag;
static void 		*enumClassRoot = NULL, *enumTagClassRoot = NULL,
			*subdirRoot = NULL;

/* external functions declarations */

int 	AmEnumsErrorLineNo;
char	AmEnumsErrorString[ 512];

/* top level functions for the YACC parser */

int enumsparse( void);
void enumsrestart( FILE *);

/* buffer reading function for the enumerations parser */

int AmEnumsFpYYinput( char        *buffer,
                    int         maxSize)
{
    return( fread( buffer, 1, maxSize, enumsFp));
}

/*************************************************************************
 *
 * Description:
 *	set the global enumeration database 
 * Arguments:
 * 	dbFile - name of the enumeration database file, which contains the
 * 	enumeration tag/class definitions, enumeration definitions for each
 * 	class, and the subdirectory definitions for resources which may be
 * 	read from preferences files 
 *	errorString - string to write the error message to on parse
 *	failure.
 * Returns: 
 *	AmOK on success, AmOPEN_FAILED or AmBAD_ARGUMENT on failure
 *
 *************************************************************************/

AmErrorCode 
AmParseEnumerationFile( const char *dbFile,
			char 	   *errorString)
{

    if (( enumsFp = fopen( dbFile, "r")) == NULL) {
	sprintf( errorString, "failed to open %s", dbFile);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    enumsrestart( enumsFp);
    AmEnumsErrorLineNo = 1;
    errorFlag = FALSE;
    enumsparse();
    fclose( enumsFp);
    if ( errorFlag) {
	strcpy( errorString, AmEnumsErrorString);
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 * 	comparator function for the enumeration class tree 
 * Arguments:
 *	a, b - enumeration classes to compare.
 * Returns: 
 *	strcmp() of the class names.
 *
 *************************************************************************/

static int 
byEnumClassName( enumClassPtr	a,
		 enumClassPtr	b)
{
    return( strcmp( a->className, b->className));
}

/*************************************************************************
 *
 * Description:
 * 	comparator function for the tag->enumeration class tree 
 * Arguments:
 *	a, b - tag->enumeration classes to compare.
 * Returns: 
 *	strcmp() of the tag names
 *
 *************************************************************************/

static int 
byTagName( enumTagClassPtr	a,
	   enumTagClassPtr	b)
{
    return( strcmp( a->tagName, b->tagName));
}

/*************************************************************************
 *
 * Description:
 * 	comparator function for the enumeration tag/value tree by 
 *	enumeration name 
 * Arguments:
 *	a, b - tag->enumeration definitions to compare.
 * Returns: 
 *	strcmp() of the enumeration names names
 *
 *************************************************************************/

static int 
byEnumName( enumDefPtr	a,
	    enumDefPtr	b)
{
    return( strcmp( a->enumName, b->enumName));
}

/*************************************************************************
 *
 * Description:
 * 	comparator function for the enumeration tag/value tree by 
 *	enumeration value 
 * Arguments:
 *	a, b - tag->enumeration definitions to compare.
 * Returns: 
 *	strcmp() of the enumeration names names
 *
 *************************************************************************/

static int 
byEnumVal( enumDefPtr	a,
	   enumDefPtr	b)
{
    return( a->enumVal - b->enumVal);
}

/*************************************************************************
 *
 * Description:
 * 	given a tag/enumeration class pair, add it to the tag/enum class 
 *	search tree and the enum class search tree. Called from enums.y 
 * Arguments:
 *	tagName - enumeration tag name.
 *	className - enumeration class name.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure.
 *
 *************************************************************************/

AmErrorCode 
AmAddEnumClassDef( char 	*tagName, 
		 char	*className)
{
    enumTagClassPtr	enumTagClass;
    enumClassPtr	enumClass;
    enumClassRec	target;

/* add the tag/enumeration class defintion to the enumeration tag/class
   tree */

    if (( enumTagClass = ( enumTagClassPtr)
		malloc( sizeof( enumTagClassRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* the names are already strdup'd by the lexer from yytext */

    if (( enumTagClass->tagName = tagName) == NULL) {
	DO_RETURN_ERROR( AmSTRDUP_FAILED);
    }
    if (( enumTagClass->className = className) == NULL) {
	DO_RETURN_ERROR( AmSTRDUP_FAILED);
    }

AN_DEBUG( 200, ("adding tag/class %s %s\n", tagName, className));

/* the tag/enumeration class tree is ordered by tag name */

    tsearch( enumTagClass, &enumTagClassRoot, ( AmSearchFn) byTagName);

/* the enumeration class tree is ordered by class name.  Since there may
   be multiple tags for each class, try to find the class in the tree 
   before inserting it */

    target.className = className;
    if ( tfind( &target, ( const void **) &enumClassRoot, 
		( AmSearchFn) byEnumClassName) == NULL) {
AN_DEBUG( 200, ("adding class %s\n", className));
	if (( enumClass = ( enumClassPtr) 
		  malloc( sizeof( enumClassRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if (( enumClass->className = strdup( className)) == NULL) {
	    DO_RETURN_ERROR( AmSTRDUP_FAILED);
	}
	enumClass->enumNameRoot = NULL;
	enumClass->enumValRoot = NULL;
	tsearch( enumClass, &enumClassRoot, ( AmSearchFn) byEnumClassName);
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 * 	add an enumeration defintion to the enum string/value search 
 *	tress for the enumeration class. called from enums.y.
 * Arguments:
 *	className - enum class name to add.
 *	enumName - enumeration name to add.
 *	enumVal - enumeration value to add.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure,
 *	AmBAD_ARGUMENT if the enumeration class name couldn't be found.
 *
 *************************************************************************/

AmErrorCode 
AmAddEnumDef( char	*className,
	    char	*enumName,
	    int		enumVal)
{
    enumClassRec	target;
    enumDefPtr		enumDef;
    enumClassPtr	*enumClassRes, enumClass;

/* find the enum class to which the enumeration belongs */

    target.className = className;
    enumClassRes = ( enumClassPtr *) 
	tfind( &target, ( const void **) &enumClassRoot, 
	       ( AmSearchFn) byEnumClassName);
    if ( enumClassRes == NULL) {
AN_DEBUG( 99, ("failed to find class for %s %s\n", className, enumName));
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    enumClass = *enumClassRes;
    if (( enumDef = ( enumDefPtr) malloc( sizeof( enumDefRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* the enum name is already strdup'd by the lexer from yytext */

    if (( enumDef->enumName = enumName) == NULL) {
	DO_RETURN_ERROR( AmSTRDUP_FAILED);
    }
    enumDef->enumVal = enumVal;

/* add the enum name/value pair to the string-search tree and value-search
   tree for the enumeration class */

AN_DEBUG( 200, ("adding enum tag %s, value %d, to class %s\n", 
		enumName, enumVal, className));
    tsearch( enumDef, &enumClass->enumNameRoot, ( AmSearchFn) byEnumName);
    tsearch( enumDef, &enumClass->enumValRoot, ( AmSearchFn) byEnumVal);
    return( AmOK);
}

/******************************************************************************
 *
 * Description:
 * 	given a string tag (from strdefs.h), a string for an enumeration, and
 * 	an enumeration database, return the enumeration for that string in the
 *  	enumeration class for that tag 
 * Arguments:
 * 	tagStr - tag for which the enumeration is to found
 * 	enumName - name of the enumeration value 
 * 	enumVal - returned enumeration value (set to 0 if not found).
 * Returns: 
 *	TRUE if enumeration is found, FALSE if not 
 *
 ******************************************************************************/

AnvilBoolean 
AmGetStrEnum( const char	*tagStr,
	    char	*enumName,
	    int		*enumVal)
{
    enumTagClassRec	enumTagClassTarget;
    enumTagClassPtr	*enumTagClassRes;
    enumClassRec	enumClassTarget;
    enumClassPtr	*enumClassRes, enumClass;
    enumDefRec		enumDefTarget;
    enumDefPtr		*enumDefRes;
    const char        	*ptag;

    *enumVal = 0;
    if (( tagStr == NULL) || ( enumName == NULL)) {
	return( FALSE);
    }
 
/* AmFirst, we need to look up in the enumeration database what the enumeration
   type associated with the tag string is, then use the enumeration type
   and the value to get the enumeration value.  Note that since we are
   handling "field.tag" tags, we have to check for a "." in the tag. */
 
    if (( ptag = strchr( tagStr, '.')) != NULL) {
	ptag++;
    } else {
	ptag = tagStr;
    }

/* search for the class name corresponding to the tag */

    enumTagClassTarget.tagName = ( char *) ptag;
    enumTagClassRes = ( enumTagClassPtr *)
	tfind( &enumTagClassTarget, ( const void **) &enumTagClassRoot, 
	       ( AmSearchFn) byTagName);
    if ( enumTagClassRes == NULL) {
	return( FALSE);
    }

/* search for the enumeration class correspding to the class name */

    enumClassTarget.className = ( *enumTagClassRes)->className;
    enumClassRes = ( enumClassPtr *) 
	tfind( &enumClassTarget, ( const void **) &enumClassRoot, 
	       ( AmSearchFn) byEnumClassName);
    if ( enumClassRes == NULL) {
	return( FALSE);
    }
    enumClass = *enumClassRes;

/* search for the enumeration value corresponding to the enumeration name 
   in the enumeration class */

    enumDefTarget.enumName = enumName;
    enumDefRes = ( enumDefPtr *)
	tfind( &enumDefTarget, ( const void **) &enumClass->enumNameRoot, 
	       ( AmSearchFn) byEnumName);
    if ( enumDefRes == NULL) {
	return( FALSE);
    }
    *enumVal = ( *enumDefRes)->enumVal;
    return( TRUE);
}

/******************************************************************************
 *
 * Description:
 * 	given a string tag (from strdefs.h), a value for an enumeration, and
 * 	an enumeration database, return the enumeration string for that value 
 *  	in the enumeration class for that tag 
 * Arguments:
 * 	tagStr - tag for which the enumeration is to found
 * 	enumVal - enumeration value 
 * 	string - returned enumeration string (set to "" if not found)
 * Returns: 
 *	TRUE if value is found, FALSE if not 
 *
 ******************************************************************************/

AnvilBoolean 
AmGetEnumStr( const char	*tagStr,
	    int		enumVal,
	    char	*string)
{
    enumTagClassRec	enumTagClassTarget;
    enumTagClassPtr	*enumTagClassRes;
    enumClassRec	enumClassTarget;
    enumClassPtr	*enumClassRes, enumClass;
    enumDefRec		enumDefTarget;
    enumDefPtr		*enumDefRes;
    const char        	*ptag;

    if ( string == NULL) {
	return( FALSE);
    }
    string[ 0] = '0';
    if ( tagStr == NULL) {
	return( FALSE);
    }

/* AmFirst, we need to look up in the enumeration database what the enumeration
   type associated with the tag string is, then use the enumeration type
   and the value to get the enumeration value.  Note that since we are
   handling "field.tag" tags, we have to check for a "." in the tag. */
 
    if (( ptag = strchr( tagStr, '.')) != NULL) {
	ptag++;
    } else {
	ptag = tagStr;
    }

/* search for the class name corresponding to the tag */

    enumTagClassTarget.tagName = ( char *) ptag;
    enumTagClassRes = ( enumTagClassPtr *)
	tfind( &enumTagClassTarget, ( const void **) &enumTagClassRoot, 
	       ( AmSearchFn) byTagName);
    if ( enumTagClassRes == NULL) {
	return( FALSE);
    }

/* search for the enumeration class correspding to the class name */

    enumClassTarget.className = ( *enumTagClassRes)->className;
    enumClassRes = ( enumClassPtr *) 
	tfind( &enumClassTarget, ( const void **) &enumClassRoot, 
	       ( AmSearchFn) byEnumClassName);
    if ( enumClassRes == NULL) {
	return( FALSE);
    }
    enumClass = *enumClassRes;

/* search for the enumeration name corresponding to the enumeration value 
   in the enumeration class */

    enumDefTarget.enumVal = enumVal;
    enumDefRes = ( enumDefPtr *)
	tfind( &enumDefTarget, ( const void **) &enumClass->enumValRoot, 
	       ( AmSearchFn) byEnumVal);
    if ( enumDefRes == NULL) {
	return( FALSE);
    }
    strcpy( string, ( *enumDefRes)->enumName);
    return( TRUE);
}


static int 	COUNT;
static int	*VALUES;
static char	**STRINGS;

/******************************************************************************
 *
 * Description:
 * 	how many values are there in the tree, anyway?  used by twalk to
 *	find the number of enumerations in the tsearch() tree.
 * 
 ******************************************************************************/

static void 
countAction( const void *, VISIT order, int)
{
    if (( order == preorder) || ( order == leaf)) {
	COUNT++;
    }
}

/******************************************************************************
 *
 * Description:
 * 	called by twalk to set the values and strings from the enumerations
 *	tree.
 * 
 ******************************************************************************/

static void 
setEnumNameValueAction( enumDefPtr	*a,
			VISIT		order,
			int)
{
    if (( order == preorder) || ( order == leaf)) {
	VALUES[ COUNT] = ( *a)->enumVal;
	STRINGS[ COUNT] = strdup(( *a)->enumName);
	COUNT++;
    }
}

/******************************************************************************
 *
 * Description:
 * 	given a tag from the enumeration database, return all the strings
 *  	for the enumeration class, and the enumeration values 
 * Arguments:
 * 	tagStr - tag to return enumeration values for
 * 	strings - pointer to returned NULL-terminated vector of strings.  
 *		  Set to NULL if no values are found 
 * 	values -  vector of integer enumeration values, set to NULL if 
 *		  no values are found.
 * 	numValues -  number of values found, 0 if no values were found.
 * Returns: 
 *	TRUE if enumerations were found, FALSE if not 
 *
 ******************************************************************************/

AnvilBoolean
AmGetEnumStringValues( const char      *tagStr,
		     char            ***strings,
		     int             **values,
		     int             *numValues)
{
    enumTagClassRec	enumTagClassTarget;
    enumTagClassPtr	*enumTagClassRes;
    enumClassRec	enumClassTarget;
    enumClassPtr	*enumClassRes, enumClass;
    const char		*ptag;

    *strings = NULL;
    *values = NULL;
    *numValues = 0;
    if ( tagStr == NULL) {
	return( FALSE);
    }
 
/* AmFirst, we need to look up in the enumeration database what the enumeration
   type associated with the tag string is, then use the enumeration type
   and the value to get the enumeration value.  Note that since we are
   handling "field.tag" tags, we have to check for a "." in the tag. */
 
    if (( ptag = strchr( tagStr, '.')) != NULL) {
	ptag++;
    } else {
	ptag = tagStr;
    }

/* search for the class name corresponding to the tag */

    enumTagClassTarget.tagName = ( char *) ptag;
    enumTagClassRes = ( enumTagClassPtr *)
	tfind( &enumTagClassTarget, ( const void **) &enumTagClassRoot, 
	       ( AmSearchFn) byTagName);
    if ( enumTagClassRes == NULL) {
	return( FALSE);
    }

/* search for the enumeration class correspding to the class name */

    enumClassTarget.className = ( *enumTagClassRes)->className;
    enumClassRes = ( enumClassPtr *) 
	tfind( &enumClassTarget, ( const void **) &enumClassRoot, 
	       ( AmSearchFn) byEnumClassName);
    if ( enumClassRes == NULL) {
	return( FALSE);
    }
    enumClass = *enumClassRes;

/* how many values are in the enum-value tree for this class? */

    COUNT = 0;
    twalk( enumClass->enumNameRoot, ( AmWalkFn) countAction);
    if (( STRINGS = ( char **) malloc( sizeof( char *)*( COUNT + 1))) == NULL) {
	return( FALSE);
    }
    if (( VALUES = ( int *) malloc( sizeof( int)*( COUNT + 1))) == NULL) {
	return( FALSE);
    }
    *numValues = COUNT;
    COUNT = 0;
    twalk( enumClass->enumNameRoot, ( AmWalkFn) setEnumNameValueAction);

/* terminate the vectors, just in case... */

    STRINGS[ COUNT] = NULL;
    VALUES[ COUNT] = 0;
    *strings = STRINGS;
    *values = VALUES;
    return( TRUE);
}

/******************************************************************************
 *
 * Description:
 * 	comparator function for the tag/subdirectory search tree 
 *
 ******************************************************************************/

static int 
bySubdirTagName( subdirDefPtr	a,
		 subdirDefPtr	b)
{
    return( strcmp( a->tagName, b->tagName));
}

/******************************************************************************
 *
 * Description:
 * 	add a tag/subdirectory pair to the subdirectory/tag search tree.  
 *	called from enums.y 
 *
 ******************************************************************************/

AmErrorCode 
AmAddSubdirDef( char	*tagName,
	      char	*subdirName)
{
    subdirDefPtr	subdirDef;

    if (( subdirDef = ( subdirDefPtr) malloc( sizeof( subdirDefRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* the tag name and subdirectory name are already strdup'd by the lexer from
   yytext */

    if (( subdirDef->tagName = tagName) == NULL) {
	DO_RETURN_ERROR( AmSTRDUP_FAILED);
    }
    if (( subdirDef->subdirName = subdirName) == NULL) {
	DO_RETURN_ERROR( AmSTRDUP_FAILED);
    }
    tsearch( subdirDef, &subdirRoot, ( AmSearchFn) bySubdirTagName);
    return( AmOK);
}

/******************************************************************************
 *
 * Description:
 *	for a resource which is stored in the Config directory, find the
 *  	appropriate subdirectory for that resource.
 * Arguments:
 * 	resource -  resource for which a subdirectory is defined.
 * Returns: 
 *	the subdirectory name if value is found, NULL if not.  The value
 *	is returned directly from the resource tree, so it should not be
 *	freed.
 *
 ******************************************************************************/

const char *
AmResourceToSubdir( const char  *resource)
{
    subdirDefRec	subdirDefTarget;
    subdirDefPtr	*subdirDefRes;

    if ( resource == NULL) {
	return( NULL);
    }
    subdirDefTarget.tagName = ( char *) resource;
    subdirDefRes = ( subdirDefPtr *)
	tfind( &subdirDefTarget, ( const void **) &subdirRoot, 
	       ( AmSearchFn) bySubdirTagName);
    if ( subdirDefRes == NULL) {
	return( NULL);
    } else {
	return(( *subdirDefRes)->subdirName);
    }
}

/* error reporting function for the parser */ 

int enumserror( const char *errorString)
{
    sprintf( AmEnumsErrorString, "parse error: %s at line %d\n",
	     errorString, AmEnumsErrorLineNo);
    errorFlag = TRUE;
    return( TRUE);
}
