#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <search.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <anvil/envdefs.h>
#include <anvil/record.h>
#include <anvil/regexp.h>
#include <anvil/expr.h>

/* functions for parsing arithmetic expressions, used in the unit parser
   to parse unit conversions. This is used in combination with the lexer
   expr.l and the parser expr.y */

static char exprBuffer[ 2048], *exprBufferPtr;

/* wrapper functions for yyparse() and yyrestart() */

int exprparse( void);
void exprrestart( FILE *);

AmExprPtr	statExpr;

/****************************************************************************
 *
 * Description:
 * 	make a binary expression from two subexpressions and an operator.
 * Arguments:
 *	op - binary operator.
 *	leftExpr, rightExpr - left and right expressions.
 * Returns:
 *	resulting binary expression, or NULL if it couldn't be allocated.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeBinexpr( char	*op,
	     AmExprPtr	leftExpr,
	     AmExprPtr	rightExpr)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_BINEXPR;
    expr->u.binexpr.op = *op;
    expr->u.binexpr.left = leftExpr;
    expr->u.binexpr.right = rightExpr;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make a single-argument function expression, given a function name 
 *	and an expression as its argument.
 * Arguments:
 *	fnName - function name
 *	arg - expression to use as argument.
 * Returns:
 *	resulting function expression, or NULL if it couldn't be allocated.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeFunexpr( char	*fnName,
	     AmExprPtr	arg)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_FUNCTION;
    expr->u.funexpr.fn = ( AmFunFn) AmGetRegister( fnName);
    expr->u.funexpr.arg = arg;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make a 2-argument function expression, given a function name 
 *	and 2 expressions as its arguments.
 * Arguments:
 *	fnName - function name
 *	arg1, arg2 - expression arguments.
 * Returns:
 *	resulting function expression, or NULL if it couldn't be allocated.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeFunexpr2( char	*fnName,
	      AmExprPtr	arg1,
	      AmExprPtr	arg2)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_FUNCTION2;
    expr->u.funexpr2.fn = ( AmFunFn2) AmGetRegister( fnName);
    expr->u.funexpr2.arg1 = arg1;
    expr->u.funexpr2.arg2 = arg2;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make a 3-argument function expression, given a function name 
 *	and 3 expressions as its arguments.
 * Arguments:
 *	fnName - function name
 *	arg1, arg2, arg3 - expression arguments.
 * Returns:
 *	resulting function expression, or NULL if it couldn't be allocated.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeFunexpr3( char	*fnName,
	      AmExprPtr	arg1,
	      AmExprPtr	arg2,
	      AmExprPtr	arg3)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_FUNCTION3;
    expr->u.funexpr3.fn = ( AmFunFn3) AmGetRegister( fnName);
    expr->u.funexpr3.arg1 = arg1;
    expr->u.funexpr3.arg2 = arg2;
    expr->u.funexpr3.arg3 = arg3;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make a 4-argument function expression, given a function name 
 *	and 4 expressions as its arguments.
 * Arguments:
 *	fnName - function name
 *	arg1, arg2, arg3, arg4 - expression arguments.
 * Returns:
 *	resulting function expression, or NULL if it couldn't be allocated.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeFunexpr4( char	*fnName,
	      AmExprPtr	arg1,
	      AmExprPtr	arg2,
	      AmExprPtr	arg3,
	      AmExprPtr	arg4)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_FUNCTION4;
    expr->u.funexpr4.fn = ( AmFunFn4) AmGetRegister( fnName);
    expr->u.funexpr4.arg1 = arg1;
    expr->u.funexpr4.arg2 = arg2;
    expr->u.funexpr4.arg3 = arg3;
    expr->u.funexpr4.arg4 = arg4;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make an identifier expression, given a string identifier name
 * Arguments:
 *	ident - identifier to turn into an expression.
 * Returns:
 *	resulting identifier expression, or NULL if malloc failed.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeIdentifier( char	*ident)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->exprType = AmEXPR_IDENT;
    expr->u.ident = ( AmIdentPtr) malloc( sizeof( AmIdentRec));
    if ( expr->u.ident == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    expr->u.ident->name = ident;
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	make a numeric expression, given a string which is converted to
 *	a numeric value.
 * Arguments:
 *	numberStr - number to turn into an expression.
 * Returns:
 *	resulting numeric expression, or NULL if malloc failed.
 *
 ****************************************************************************/

AmExprPtr 
AmMakeNumber( char *numberStr)
{
    AmExprPtr	expr;

    statExpr = expr = ( AmExprPtr) malloc( sizeof( AmExprRec));
    if ( expr == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
    }
    expr->exprType = AmEXPR_NUMBER;
    expr->u.number = atof( numberStr);
    return( expr);
}

/****************************************************************************
 *
 * Description:
 * 	evaluate a simple expression with only one variable 
 * Arguments:
 *	expr - expression to evaluate.
 *	identVal - value to replace the identifier with
 * Returns:
 *	value resulting from evaluating the expression.
 *
 ****************************************************************************/

float 
AmEvalSimpleExpr( AmExprPtr	expr, 
		float	identVal)
{
    switch ( expr->exprType) {
    case AmEXPR_NUMBER:
	return( expr->u.number);
    case AmEXPR_IDENT:
	return( identVal);
    case AmEXPR_FUNCTION:
	return(( expr->u.funexpr.fn)
		    ( AmEvalSimpleExpr( expr->u.funexpr.arg, identVal)));
    case AmEXPR_FUNCTION2:
	return(( expr->u.funexpr2.fn)
		    ( AmEvalSimpleExpr( expr->u.funexpr2.arg1, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr2.arg2, identVal)));
    case AmEXPR_FUNCTION3:
	return(( expr->u.funexpr3.fn)
		    ( AmEvalSimpleExpr( expr->u.funexpr3.arg1, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr3.arg2, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr3.arg3, identVal)));
    case AmEXPR_FUNCTION4:
	return(( expr->u.funexpr4.fn)
		    ( AmEvalSimpleExpr( expr->u.funexpr4.arg1, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr4.arg2, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr4.arg3, identVal),
		      AmEvalSimpleExpr( expr->u.funexpr4.arg4, identVal)));
    case AmEXPR_BINEXPR:
	switch ( expr->u.binexpr.op) {
	case '+':
	    return( AmEvalSimpleExpr( expr->u.binexpr.left, identVal) +
		    AmEvalSimpleExpr( expr->u.binexpr.right, identVal));
	case '-':
	    return( AmEvalSimpleExpr( expr->u.binexpr.left, identVal) -
		    AmEvalSimpleExpr( expr->u.binexpr.right, identVal));
	case '/':
	    return( AmEvalSimpleExpr( expr->u.binexpr.left, identVal) /
		    AmEvalSimpleExpr( expr->u.binexpr.right, identVal));
	case '*':
	    return( AmEvalSimpleExpr( expr->u.binexpr.left, identVal) *
		    AmEvalSimpleExpr( expr->u.binexpr.right, identVal));
	default:
	    return( 0.0);
	}
    default:
	return( 0.0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	substitute the identPtrs for the idents in the expression 
 *	so the values can be substituted quickly 
 * Arguments:
 *	expr - expression to substitude values in.
 *	AmIdentPtr - vector of identifiers to replace.
 *	numIdents - length of the vector.
 *	errorMsg - error message returned if there were problems.
 * Returns:
 *	TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/
   
AnvilBoolean 
AmSubstituteExpr( AmExprPtr		expr,
		AmIdentPtr	identVec,
		int		numIdents,
		char		*errorMsg)
{
    int	ii;

    switch ( expr->exprType) {
    case AmEXPR_NUMBER:
	return( TRUE);
    case AmEXPR_IDENT:
	for ( ii = 0; ii < numIdents; ii++) {
	    if ( !strcmp( identVec[ ii].name, expr->u.ident->name)) {
		expr->u.ident = &identVec[ ii];
		return( TRUE);
	    }
	} 
	sprintf( errorMsg, "unassigned variable %s", expr->u.ident->name);
	return( FALSE);
    case AmEXPR_FUNCTION:
	return( AmSubstituteExpr( expr->u.funexpr.arg, identVec, numIdents,
				errorMsg));
    case AmEXPR_FUNCTION2:
	return( AmSubstituteExpr( expr->u.funexpr2.arg1, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr2.arg2, identVec, numIdents,
				errorMsg));
    case AmEXPR_FUNCTION3:
	return( AmSubstituteExpr( expr->u.funexpr3.arg1, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr3.arg2, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr3.arg3, identVec, numIdents,
				errorMsg));
    case AmEXPR_FUNCTION4:
	return( AmSubstituteExpr( expr->u.funexpr4.arg1, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr4.arg2, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr4.arg3, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.funexpr4.arg4, identVec, numIdents,
				errorMsg));
    case AmEXPR_BINEXPR:
	return( AmSubstituteExpr( expr->u.binexpr.left, identVec, numIdents,
				errorMsg) &&
		AmSubstituteExpr( expr->u.binexpr.left, identVec, numIdents,
				errorMsg));
    default:
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	assign the identifiers to values in a vector of identifiers
 * Arguments:
 *	identVec - vector of identifiers.
 *	values - vector of values to set.
 *	numIdents - number of identifiers.
 *
 ****************************************************************************/

void 
AmAssignIdents( AmIdentPtr	identVec,
	      float	*values,
	      int	numIdents)
{
    int	ii;

    for ( ii = 0; ii < numIdents; ii++) {
	identVec[ ii].value = values[ ii];
    }
}

/****************************************************************************
 *
 * Description:
 * 	evaluate an expression with the identifiers assigned. 
 * Arguments:
 *	expr - expression to evaluate.
 * Returns:
 *	resulting value of the expression.
 *
 ****************************************************************************/

float 
AmEvalAssignedExpr( AmExprPtr	expr)
{
    switch ( expr->exprType) {
    case AmEXPR_NUMBER:
	return( expr->u.number);
    case AmEXPR_IDENT:
	return( expr->u.ident->value);
    case AmEXPR_FUNCTION:
	return(( expr->u.funexpr.fn)
		    ( AmEvalAssignedExpr( expr->u.funexpr.arg)));
    case AmEXPR_FUNCTION2:
	return(( expr->u.funexpr2.fn)
		    ( AmEvalAssignedExpr( expr->u.funexpr2.arg1),
		      AmEvalAssignedExpr( expr->u.funexpr2.arg2)));
    case AmEXPR_FUNCTION3:
	return(( expr->u.funexpr3.fn)
		    ( AmEvalAssignedExpr( expr->u.funexpr3.arg1),
		      AmEvalAssignedExpr( expr->u.funexpr3.arg2),
		      AmEvalAssignedExpr( expr->u.funexpr3.arg3)));
    case AmEXPR_FUNCTION4:
	return(( expr->u.funexpr4.fn)
		    ( AmEvalAssignedExpr( expr->u.funexpr4.arg1),
		      AmEvalAssignedExpr( expr->u.funexpr4.arg2),
		      AmEvalAssignedExpr( expr->u.funexpr4.arg3),
		      AmEvalAssignedExpr( expr->u.funexpr4.arg4)));
    case AmEXPR_BINEXPR:
	switch ( expr->u.binexpr.op) {
	case '+':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) +
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '-':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) -
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '/':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) /
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '*':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) *
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	default:
	    return( 0.0);
	}
    default:
	return( 0.0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	evaluate a boolean expression.
 * Arguments:
 *	expresion to evaluate.
 * Results:
 *	value of the expression.
 *
 ****************************************************************************/

AnvilBoolean 
AmEvalBooleanExpr( AmExprPtr	expr)
{
    switch ( expr->exprType) {
    case AmEXPR_BINEXPR:
	switch ( expr->u.binexpr.op) {
	case '&':
	    return( AmEvalBooleanExpr( expr->u.binexpr.left) &&
		    AmEvalBooleanExpr( expr->u.binexpr.right));
	case '|':
	    return( AmEvalBooleanExpr( expr->u.binexpr.left) ||
		    AmEvalBooleanExpr( expr->u.binexpr.right));
	case '>':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) >
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '<':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) <
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '=':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) ==
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	case '^':
	    return( AmEvalAssignedExpr( expr->u.binexpr.left) !=
		    AmEvalAssignedExpr( expr->u.binexpr.right));
	default:
	    return( !!AmEvalAssignedExpr( expr));
	}
    default:
	return( 0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	because the interface to YACC *sucks*, we have to write a function
 *	which copies the input data to a static buffer.
 *
 ****************************************************************************/

AnvilBoolean 
AmCopyExprBuffer( char	*buffer)
{
    if ( strlen( buffer) < sizeof( exprBuffer)) {
	strcpy( exprBuffer, buffer);
	exprBufferPtr = exprBuffer;
	return( TRUE);
    } else {
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 *	YACC input function.
 * Arguments:
 *	buffer - buffer to read data into.
 *	maxSize - number of bytes in the buffer.
 * Returns:
 *	number of bytes read into the buffer.
 *
 ****************************************************************************/

int AmExprYYinput( char	*buffer,
		 int	maxSize)
{
    int	length;

    if ( (unsigned) maxSize < strlen( exprBufferPtr)) {
	strncpy( buffer, exprBufferPtr, maxSize);
	exprBufferPtr += maxSize;
	return( maxSize);
    } else {
	strcpy( buffer, exprBufferPtr);
	length = strlen( exprBufferPtr);
	exprBufferPtr += length;
	return( length);
    }
}

/****************************************************************************
 *
 * Description:
 *	parse an expression, given a string.
 * Arguments:
 *	exprStr - string containing the expression.
 * Returns:
 *	the parsed expression on success, NULL on failure.
 *
 ****************************************************************************/

AmExprPtr AmParseExpr( char	*exprStr)
{
    exprrestart( NULL);
    AmCopyExprBuffer( exprStr);
    exprparse();
    return( statExpr);
}

/****************************************************************************
 *
 * Description:
 *	yyerror() wrapper function.
 *
 ****************************************************************************/

int exprerror( const char *msg)
{
    return (0);
}
