#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <anvil/project.h>
#include <anvil/files.h>
#include <anvil/choppath.h>

/* different absolute paths for NT Versus UNIX */
/****************************************************************************
 *
 * Description:
 *      is pathname an absolute or relative path?
 * Arguments:
 *      pathname - path to test.
 * Returns:
 *      TRUE if it's an absolute path, FALSE otherwise.
 *
 ****************************************************************************/


#if defined(_WIN32)

AnvilBoolean 
AmAbsolutePath( const char *pathname)
{
    if ( pathname == NULL) {
	return( FALSE);
    } else if ((( pathname[ 1] == ':') && isalpha( pathname[ 0])) ||
	       AmPathDelim( pathname[ 0])) {
    	return( TRUE);
    } else {
    	return( FALSE);
    }
}

#else 

AnvilBoolean 
AmAbsolutePath( const char *pathname) 
{
    if ( pathname == NULL) {
	return( FALSE);
    } else {
    	return( pathname[ 0] == '/');
    }
}

#endif
/****************************************************************************
 *
 * Description:
 *      apply some sanity to a path by changing all '\' characters to 
 *	'/' characters.
 * Arguments:
 *      pathname - path to sanity.
 *
 ****************************************************************************/


void 
AmSanePath( char *pathname)
{
    if ( pathname == NULL) {
    	return;
    }
    for ( ; *pathname != '\0'; pathname++) {
    	if ( *pathname == '\\') {
	    *pathname = '/';
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	given an environment variable and a subdirectory rooted off 
 *	the directory pointed to by the environment variable, and a 
 *	filename, create a fully qualified path either by composing 
 *	them if the filename is relative, or from the file itself if 
 *	it's absolute.  Used for getting files from the $ANVIL_CONFIG
 *	directory. 
 * Arguments:
 *	envName - name of the environment variable.
 *	dirName - name of the subdirectory.
 * 	filename - file name 
 * Returns:
 *	the full path to the resulting file (returned as a static).
 *	or NULL if the file doesn't exist.
 *
 ****************************************************************************/

char *
AmGetEnvSubdirFile( const char *envName,
		    const char *dirName,
		    const char *filename)
{
    static char buffer[ 512], *envPath;
 
/* First, try the filename as given if it starts with "/" ...  */
 
    if ( AmAbsolutePath( filename)) {
        if ( access ( filename, R_OK ) == 0 ) {
            strcpy( buffer, filename);
            return( buffer);
        } else {
            return( NULL);
        }
    }
 
/* Next, try to append the filename to the envPath ... */
 
    envPath = getenv( envName);
    if ( envPath == NULL) {
        return( NULL);
    }
    sprintf( buffer, "%s/%s/%s", envPath, dirName, filename);
    if ( access( buffer, R_OK) == 0) {
        return( buffer);
    } else {
        return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 *  	pull a file specified by a filename with an environment variable,
 *  	and return a static buffer to the full path if the file exists, or
 *  	NULL if it doesn't 
 * Arguments:
 *  	envName - name of the environment variable
 *  	filename - name of the file
 * Returns:
 *	the full path name to the resulting file or NULL if the file 
 *	doesn't exist.
 *
 ****************************************************************************/
 
char *
AmGetEnvFile( const char *envName, 
	    const char *filename)
{
    static char buffer[ 512], *envPath;
 
/* First, try the filename as given if it starts with "/" ...  */
 
    if ( AmAbsolutePath( filename)) {
        if ( access ( filename, R_OK ) == 0 ) {
            strcpy( buffer, filename);
            return( buffer);
        } else {
            return( NULL);
        }
    }
 
/* Next, try to append the filename to the envPath ... */
 
    envPath = getenv(envName);
    if ( envPath == NULL) {
        return( NULL);
    }
    sprintf( buffer, "%s/%s", envPath, filename);
    if ( access( buffer, R_OK) == 0) {
        return( buffer);
    } else {
        return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	given a file paths with environment variables, evaluate the variables, 
 *  	and return the path 
 * Arguments:
 *	path - pathname with embedded environment variables prefixed
 *	with '$'.
 * Returns:
 *	the evaluated path.
 * Modifications:
 *	I don't think this tests for undefined environment variables
 *	properly.
 *
 ****************************************************************************/

char *
AmFileEnvPath( const char *path)
{
    static char	pathbuf[ PATHLEN];
    char	env[ ENVLEN], *p;
    int		ii;

    ii = 0;
    while ( *path != '\0') {
	if ( *path != '$') {
	    pathbuf[ ii++] = *path++;
	} else {
	    path++;				/* skip over the '$' */
	    p = strpbrk( path, "/\\");		/* copy to the 1st '/' or '\'*/
	    strncpy( env, path, p - path);
	    env[ p - path] = '\0';		/* and NULL terminate it */
	    path = p;				/* step path to after var */
	    p = getenv( env);			/* insert the environment */
	    if ( p != NULL) {			/* variable result if it's */
		strcpy( &pathbuf[ ii], p);	/* not NULL */
		ii += strlen( p);
	    }
	}
    }
    pathbuf[ ii] = '\0';
    return( pathbuf);
}
/******************************************************************
 *
 *  open a file prefixed with a path from an environment variable:
 *  envName -    name of the environment variable
 *  filename -   name of the file
 *  type -       type to pass to fopen, either "r", "w" or "a"
 *
 ******************************************************************/

FILE *AmOpenEnvFile ( const char  *envName,
                      const char  *filename,
                      const char  *type )
{
    FILE        *fp;
    char        *envPath, buffer[512];

    if ( *filename == '/' ) {
        fp = fopen ( filename, type );
        if ( fp != NULL ) return ( fp );
    }

    envPath = getenv ( envName );
    if ( envPath == NULL ) {
        strcpy ( buffer, filename );
    } else {
        if ( *filename != '/' ) sprintf ( buffer, "%s/%s", envPath, filename );
        if ( *filename == '/' ) sprintf ( buffer, "%s%s", envPath, filename );
    }

    fp = fopen ( buffer, type );

    return ( fp );

}

