#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


/* Get, set, print, and scan functions for tag-values used by records. */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>

#include <anvil/datatype.h>
#include <anvil/colorcell.h>
#include <anvil/strlib.h>
#include <anvil/mem_funcs.h>
#include <anvil/enums.h>
#include <anvil/record.h>
#include <anvil/fns.h>
#include <anvil/envdefs.h>
#include <anvil/files.h>
#include <anvil/getset.h>
#include <anvil/choppath.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *  	set a vector value, given the offset pointer, data type, storage type,
 *  	vector length 
 * Arguments:
 *  	offsetPtr - pointer to the value to be set.
 *  	dataType - enumerated data type of the value
 *  	storage - dynamically allocated (AmDYNAMIC), or static (AmSTATIC)
 *  	value - value to retrieve
 *  	length - length of vector to set (usually retrieved from structure 
 * Returns:
 *	TRUE if the data type was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmSetVector( uchar_t 		*offsetPtr, 
	     AmDataTypeEnum 	dataType, 
	     AmStorageTypeEnum	storage,
	     void		*value,
	     int		length)
{
    char	**srcPtr, **dstPtr;

    switch ( dataType) {
    case AmINT_PTR:
	if ( storage == AmDYNAMIC) {
	    if ( *( uchar_t **) offsetPtr != NULL) {
		free( *( uchar_t **) offsetPtr);
	    }
	    *( int **) offsetPtr = 
		( int *) AmCopyVec( value, length*sizeof( int));
	} else {
	    memcpy( offsetPtr, value, length*sizeof( int));
	}
	break;
    case AmFLOAT_PTR:
	if ( storage == AmDYNAMIC) {
	    if ( *( uchar_t **) offsetPtr != NULL) {
		free( *( uchar_t **) offsetPtr);
	    }
	    *( float **) offsetPtr = 
		( float *) AmCopyVec( value, length*sizeof( float));
	} else {
	    memcpy( offsetPtr, value, length*sizeof( float));
	}
	break;
    case AmSTRING_PTR:
	if ( storage == AmDYNAMIC) {
	    if ( *( char ***) offsetPtr != NULL) {
		AmFreeStrVec( *( char ***) offsetPtr);
	    }
	    *( char ***) offsetPtr = AmCopyStrVec(( const char **) value);
	} else {

/* free the old strings from the string array */

	    for ( dstPtr = ( char **) offsetPtr; *dstPtr != NULL; dstPtr++) {
		free( *dstPtr);
		*dstPtr = NULL;
	    }
	    for ( srcPtr = ( char **) value, dstPtr = ( char **) offsetPtr;
		  *srcPtr != NULL; srcPtr++, dstPtr++) {
		*dstPtr = strdup( *srcPtr);
		if ( *dstPtr == NULL) {
		    DO_ERROR( AmSTRDUP_FAILED);
		    return( FALSE);
		}
	    }
	}
	break;
    default:
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	set a scalar value, given the offset pointer, data type, storage 
 *	type and a void pointer to the value to set it with
 * Arguments:
 *  	offsetPtr - pointer to the value to be set.
 *  	dataType - enumerated data type of the value
 *  	storage - dynamically allocated (AmDYNAMIC), or static (AmSTATIC)
 *  	value - value to retrieve
 * Returns:
 *	TRUE if the data type was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmSetScalar( uchar_t 		*offsetPtr, 
	     AmDataTypeEnum 	dataType, 
	     AmStorageTypeEnum	storage,
	     void		*value)
{
    switch ( dataType) {
    case AmINT:
    case AmENUM:
    case AmUNITS:
    case AmVARIABLE:
	*( int *) offsetPtr = ( int) value;
	break;
    case AmBOOLEAN:
	*( AnvilBoolean *) offsetPtr = ( int) value;
	break;

/* floats are converted to double (pointers) by AmMakeTagValueArrayVararg */

    case AmFLOAT:	
	memcpy( offsetPtr, &value, sizeof( float));
	break;
    case AmDOUBLE:	/* note that doubles are copied as pointers */
	*( double *) offsetPtr = *( double *) value;
	break;
    case AmCHAR:
	*( char *) offsetPtr = ( int) value;
	break;
    case AmUNSIGNED_CHAR:
	*( uchar_t *) offsetPtr = ( uint_t) value;
	break;
    case AmSHORT:
	*( short *) offsetPtr = ( int) value;
	break;
    case AmUNSIGNED_SHORT:
	*( ushort_t *) offsetPtr = ( uint_t) value;
	break;
    case AmLONG:
	*( long *) offsetPtr = ( long) value;
	break;
    case AmUNSIGNED_LONG:
	*( ulong_t *) offsetPtr = ( ulong_t) value;
	break;
    case AmUNSIGNED_INT:
	*( uint_t *) offsetPtr = ( uint_t) value;
	break;
    case AmCOLOR_CELL:
	*( AmColorCellPtr) offsetPtr = *( AmColorCellPtr) value;
	break;
    case AmFILE:
    case AmSTRING:

/* if it's a static string, then copy it over, otherwise, free it if it's
   not NULL, and set the pointer to a copy of the new string */

	if (( storage == AmSTATIC) || ( storage == AmLOCAL)) {
	    if ( value != NULL) {
		strcpy( ( char *) offsetPtr, ( char *) value);
	    }
	} else { 
	    if ( *( char **) offsetPtr != NULL) {
		free( *( char **) offsetPtr);
	    }
	    if ( value != NULL) {
		*( char **) offsetPtr = strdup(( char *) value);
	    } else {	
		*( char **) offsetPtr = NULL;
	    }
	}
	break;
    case AmOBJECT:

/* delete the old object if it is not NULL */

	if ( *( void **) offsetPtr != NULL) {
	    delete ( AmRecordClass *) *( void **) offsetPtr;
	}
	*( void **) offsetPtr = value;
    case AmFUNC_PTR:
    case AmVOID_PTR:
	*( void **) offsetPtr = value;
	break;
    default:
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	given an offset pointer, a datatype, storage type, a return pointer 
 *	and a length, set the return pointers contents 
 * Arguments:
 *  	offsetPtr -pointer to the value to be retrieved 
 *  	dataType - enumerated data type of the value
 *  	storage - dynamically allocated (AmDYNAMIC), or static (AmSTATIC)
 *  	ptr - 	pointer to set with retreived value.
 *  	length - length of value to retrieve
 * Returns:
 *	TRUE if the data type was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetValue( uchar_t 		*offsetPtr,
	  AmDataTypeEnum 	dataType, 
	  AmStorageTypeEnum	storage,
	  void			**ptr,
	  int			length)
{
    switch ( dataType) {
    case AmINT:
    case AmENUM:
    case AmUNITS:
    case AmVARIABLE:
	*ptr = ( void *) *( int *) offsetPtr;
	break;
    case AmBOOLEAN:
	*( AnvilBoolean *) ptr = *( AnvilBoolean *) offsetPtr;
	break;
    case AmFLOAT:	

/* floats are converted to double (pointers) by AmMakeTagValueArrayVararg */

	memcpy( ptr, offsetPtr, sizeof( float));
	break;
    case AmDOUBLE:	
	memcpy( ptr, offsetPtr, sizeof( double));
	break;
    case AmCHAR:
	*( char *) ptr = *( char *) offsetPtr;
	break;
    case AmUNSIGNED_CHAR:
	*( uchar_t *) ptr =  *( uchar_t *) offsetPtr;
	break;
    case AmSHORT:
	*( short *)ptr = *( short *) offsetPtr;
	break;
    case AmUNSIGNED_SHORT:
	*( ushort_t *) ptr = *( ushort_t *) offsetPtr;
	break;
    case AmLONG:
	*( long *) ptr = *( long *) offsetPtr;
	break;
    case AmUNSIGNED_LONG:
	*( ulong_t *) ptr = *( ulong_t *) offsetPtr;
	break;
    case AmUNSIGNED_INT:
	*( uint_t *) ptr = *( uint_t *) offsetPtr;
	break;
    case AmCOLOR_CELL:
	*( AmColorCellPtr) ptr = *( AmColorCellPtr) offsetPtr;
	break;
    case AmFILE:
    case AmSTRING:
	if ( storage == AmDYNAMIC) {
	    if ( *( char **) offsetPtr != NULL) {
		*ptr = strdup(*( char **) offsetPtr);
	    } else {
		*ptr = NULL;
	    }
	} else {
	    *ptr = strdup(( char *) offsetPtr);
	}
	break;
    case AmFUNC_PTR:
    case AmVOID_PTR:
    case AmOBJECT:
	*ptr = *( void **) offsetPtr;
	break;
    case AmINT_PTR:
	if ( length > 0) {
	    if ( storage == AmDYNAMIC) {
		*ptr = AmCopyVec( *( int **) offsetPtr, length*sizeof( int));
	    } else {
		*ptr = AmCopyVec(( int *) offsetPtr, length*sizeof( int));
	    }
	} else {
	    *ptr = NULL;
	}
	break;
    case AmCOLOR_CELL_PTR:
	if ( length > 0) {
	    if ( storage == AmDYNAMIC) {	
		*ptr = AmCopyVec( *( AmColorCellPtr *) offsetPtr,
				length*sizeof( AmColorCellRec));
	    } else {
		*ptr = AmCopyVec( ( AmColorCellPtr) offsetPtr,
				length*sizeof( AmColorCellRec));
	    }
	} else {
	    *ptr = NULL;
	}
	break;
    case AmFLOAT_PTR:
	if ( length > 0) {
	    if ( storage == AmDYNAMIC) {
		*ptr = AmCopyVec( *( float **) offsetPtr, length*sizeof( float)); 
	    } else {
		*ptr = AmCopyVec(( float *) offsetPtr, length*sizeof( float));
	    }
	} else {
	    *ptr = NULL;
	}
	break;
    case AmSTRING_PTR:
	if ( length > 0) {
	    if ( storage == AmDYNAMIC) {
		*ptr = ( void *) AmCopyStrVec( *( const char ***) offsetPtr);
	    } else {
		*ptr = ( void *) AmCopyStrVec( ( const char **) offsetPtr);
	    }
	} else { 	
	    *ptr = NULL;
	}
	break;
    default:
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	given a tag value, print it out according to the type from the lookup
 *	table 
 * Arguments:
 *  	buffer - string buffer to print into.
 *	tagStr - tag string used for looking up units if they apply.
 *  	length - length of the vector if it's a vector value.
 *	value - pointer to the value to print.
 *  	dataType - the value's data type.
 * Returns:
 *	TRUE if the data type was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPrintValue( char		*buffer,
	    char		*tagStr,
	    int			length,
	    void		*value,
	    AmDataTypeEnum	dataType)
{
    char		*string, *p;
    int			ii;
    AmColorCellPtr	colorCell;

    if ( tagStr == NULL ) {
	AN_DEBUG( 30, ("tagStr pointer in AmPrintValue is NULL\n"));
	return( FALSE);
    }
    if ( value == NULL ) {
	AN_DEBUG( 30, ("value pointer in AmPrintValue is NULL\n"));
	return( FALSE);
    }

/* for the tag, find the corresponding datatype, and use it for printing the 
   data value string */

    switch ( dataType) {
    case AmFLOAT:
	sprintf( buffer, "%f", *( float *) value);
	break;
    case AmINT:
	sprintf( buffer, "%d", *( int *) value);
	break;
    case AmINT_PTR:
	if ( *( int **) value != NULL) {
	    for ( ii = 0, p = buffer; ii < length; ii++, p += strlen( p)) {
		sprintf( p, "%d ", (*( int **) value)[ ii]);
	    }
	} else {
	    return( FALSE);
	}
	break;
    case AmFLOAT_PTR:
	if ( *( float **) value != NULL) {
	    for ( ii = 0, p = buffer; ii < length; ii++, p += strlen( p)) {
		sprintf( p, "%f ", (*( float **) value)[ ii]);
	    }
	} else {
	    return( FALSE);
	}
	break;
    case AmSTRING_PTR:
	if ( *( char ***) value != NULL) {
	    for ( ii = 0, p = buffer; 
		  (( ii < length) && ((*( char ***) value)[ ii]) != NULL) ; 
		  ii++, p += strlen( p)) {
		sprintf( p, "\"%s\" ", (*( char ***) value)[ ii]);
	    }
	} else {
	    return( FALSE);
	}
	break;
    case AmENUM:
	if ( !AmGetEnumStr( tagStr, *( int *) value, buffer)) {
	    return( FALSE);
	}
	break;
    case AmDOUBLE:
	sprintf( buffer, "%f", *( double *) value);
	break;
    case AmCHAR:
	sprintf( buffer, "%c", *( char *) value);
	break;
    case AmUNSIGNED_CHAR:
	sprintf( buffer, "%c", *( uchar_t *) value);
	break;
    case AmSHORT:
	sprintf( buffer, "%hd", *( short *) value);
	break;
    case AmUNSIGNED_SHORT:
	sprintf( buffer, "%hu", *( ushort_t *) value);
	break;
    case AmLONG:
	sprintf( buffer, "%ld", *( long *) value);
	break;
    case AmUNSIGNED_LONG:
	sprintf( buffer, "%lu", *( unsigned long *) value);
	break;
    case AmUNSIGNED_INT:
	sprintf( buffer, "%u", *( uint_t *) value);
	break;

/* the scanner doesn't like newlines, so we convert them to '\n's.  UNKNOWNS
   default to strings. */

    case AmUNKNOWN:
    case AmSTRING:
	if ( *( char **) value != NULL) {
	    AmCopyNewlineString( buffer, *( char **) value);
	} else {
	    return( FALSE);
	}
	break;
    case AmFILE:
	if ( *( char **) value != NULL) {
	    sprintf( buffer, "%s", *( char **) value);
	} else {
	    return( FALSE);
	}
	break;
    case AmFUNC_PTR:
	string = AmGetRegisteredString( *( void **) value);
	if ( string != NULL) {
	    sprintf( buffer, "%s", string);
	} else {
	    return( FALSE);
	}
	break;
    case AmOBJECT:
    case AmVOID_PTR:
	sprintf( buffer, "%x", ( unsigned int) *( void **) value);
	break;
    case AmBOOLEAN:
        if ( *( AnvilBoolean *)value == TRUE) {
	    sprintf( buffer, "TRUE");
        } else {
            sprintf( buffer, "FALSE");
        }
	break;
    case AmCOLOR_CELL_PTR:
	colorCell = *( AmColorCellPtr *) value;
	if ( colorCell != NULL) {
	    for ( ii = 0, p = buffer; ii < length; ii++, p += strlen( p)) {
		sprintf( p, "%d %d %d ", 
			 colorCell->red, colorCell->green, colorCell->blue);
		colorCell++;
	    }
	} else {
	    return( FALSE);
	}
	break;
    case AmCOLOR_CELL:
	colorCell = ( AmColorCellPtr) value;
	if ( colorCell != NULL) {
	    sprintf( buffer, "%d %d %d", 
		     colorCell->red, colorCell->green, colorCell->blue);
	}
	break;
    case AmUNITS:
	if ( AmGetUnitString( *( unitType *) value, &string)) {
	    strcpy( buffer, string);
	} else {	
	    strcpy( buffer, "unknown");
	}
	break;
    default:
	AN_DEBUG( 10, ("Data type %d not found\n", dataType));
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	scan a single value from a string, given a tag (which will 
 *	determine the type from a lookup table) 
 * Arguments:
 *	tagStr - tag string used for looking up units if they apply.
 *	valueStr - value string to scan.
 *	value - pointer to the value to scan into.
 *  	copied - flag whether the value had to be allocated.
 * Returns:
 *	AmBAD_TYPE if the type for tagStr wasn't found.
 *	AmFILE_NO_EXIST if the tag was a "file." directive, and the 
 *	file doesn't exist.
 *	AmOPEN_FAILED if the file couldn't be opened. 
 *	AmMALLOC_FAILED or AmSTRDUP_FAILED if memory couldn't be allocated
 *	AmOK if everything was OK.
 *
 ****************************************************************************/

AmErrorCode 
AmScanValue( char	*tagStr,
	     char	*valueStr,
	     void	**value,
	     AnvilBoolean	*copied)
{
    char		*path, relPath[ PATHLEN];
    AmDataTypeEnum	dataType;
    const char		*subdir;

    dataType = AmGetTagType( tagStr);
    *copied = FALSE;

/* for the tag, find the corresponding datatype, and use it for scanning the 
   data value string */

    switch ( dataType) {

/* promote floats to doubles, and allocate them, because it's more than
   4 bytes */

    case AmFLOAT:
	sscanf( valueStr, "%f", ( float *) value);
	break;
    case AmINT:
	sscanf( valueStr, "%d", ( int *) value);
	break;
    case AmINT_PTR:
 
/* check to see if the resource is prefixed with ".file".  If so, the values
   are read from a file, otherwise, they're parsed from the string */

	if ( AmPrefixTag( FILE_PREFIX, tagStr)) {
	    if ( AmAbsolutePath( valueStr)) {
		subdir = NULL;
	    } else {
		subdir = AmResourceToSubdir( &tagStr[ 5]);
	    }
	    if ( subdir == NULL) {
		strcpy( relPath, valueStr);
	    } else {
		sprintf( relPath, "%s/%s", subdir, valueStr);
	    }
	    path = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( path == NULL) {
		*value = NULL;
		DO_RETURN_ERROR( AmFILE_NO_EXIST);
	    }
	    *value = ( void *) AmReadIntVec( path);
	} else {
	    *value = ( void *) AmReadIntStrVec( valueStr);
	}
	*copied = TRUE;
	break;
    case AmFLOAT_PTR:

/* check to see if the floats are in a file, or should be parsed directly */

	if ( AmPrefixTag( FILE_PREFIX, tagStr)) {
	    if ( AmAbsolutePath( valueStr)) {
		subdir = NULL;
	    } else {
		subdir = AmResourceToSubdir( &tagStr[ 5]);
	    }
	    if ( subdir == NULL) {
		strcpy( relPath, valueStr);
	    } else {
		sprintf( relPath, "%s/%s", subdir, valueStr);
	    }
	    path = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( path == NULL) {
		*value = NULL;
		DO_RETURN_ERROR( AmFILE_NO_EXIST);
	    }
	    *value = ( void *) AmReadFloatVec( path);
	} else { 
	    *value = ( void *) AmReadFloatStrVec( valueStr);
	}
	*copied = TRUE;
	break;
    case AmSTRING_PTR:

/* check to see if the strings are in a file, or should be parsed directly */

	if ( AmPrefixTag( FILE_PREFIX, tagStr)) {
	    if ( AmAbsolutePath( valueStr)) {
		subdir = NULL;
	    } else {
		subdir = AmResourceToSubdir( &tagStr[ 5]);
	    }
	    if ( subdir == NULL) {
		strcpy( relPath, valueStr);
	    } else {
		sprintf( relPath, "%s/%s", subdir, valueStr);
	    }
	    path = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( path == NULL) {
		*value = NULL;
		DO_RETURN_ERROR( AmFILE_NO_EXIST);
	    }
	    *value = ( void *) AmReadLines( path);
	} else {
	    *value = ( void *) AmVectorQuotedStrings( valueStr);
	}
	*copied = TRUE;
	break;
    case AmCOLOR_CELL_PTR:

/* check for the file prefix... */

	if ( AmPrefixTag( FILE_PREFIX, tagStr)) {
	    int	numColors;
	    
	    if ( AmAbsolutePath( valueStr)) {
		subdir = NULL;
	    } else {
		subdir = AmResourceToSubdir( &tagStr[ 5]);
	    }
	    if ( subdir == NULL) {
		strcpy( relPath, valueStr);
	    } else {
		sprintf( relPath, "%s/%s", subdir, valueStr);
	    }
	    path = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( path == NULL) {
		*value = NULL;
		DO_RETURN_ERROR( AmFILE_NO_EXIST);
	    }

/* if so, read from the colormap file */

	    if ( AmReadColors( path, &numColors, 
			     ( AmColorCellPtr *) value) != AmOK) {
		DO_RETURN_ERROR( AmOPEN_FAILED);
	    }	
	} else {

/* otherwise read striaght from the string */

	    *value = ( void *) AmReadColorCellStr( valueStr);
	}
	*copied = TRUE;
	break;
    case AmENUM:

/* use the enumeration database to convert the string into an enum */

	{
	    int		enumVal;

	    if ( AmGetStrEnum( tagStr, valueStr, &enumVal)) {
		*value = ( void *) enumVal;
	    } else {
		AmAddResourceToLog( "badEnumeration", tagStr, valueStr);
		*value = ( void *) 0;
	    }
	}
	break;
    case AmDOUBLE:
	*value = ( void *) malloc( sizeof( double));
	if ( *value == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	sscanf( valueStr, "%lf", ( double *) *value);
	break;
    case AmCHAR:
    case AmSHORT:
	sscanf( valueStr, "%d", ( int *) value);
	break;
    case AmUNSIGNED_CHAR:
    case AmUNSIGNED_SHORT:
    case AmUNSIGNED_INT:
	sscanf( valueStr, "%u", ( uint_t *) value);
	break;
    case AmLONG:
	sscanf( valueStr, "%ld", ( long *) value);
	break;
    case AmUNSIGNED_LONG:
	sscanf( valueStr, "%lu", ( unsigned long *) value);
	break;
    case AmFILE:
	if ( valueStr != NULL)  {
	    *value = ( void *) strdup( AmFileEnvPath( valueStr));
	    if ( *value == NULL) {
		DO_RETURN_ERROR( AmSTRDUP_FAILED);
	    }
	} else {
	    *value = NULL;
	}
	*copied = TRUE;
	break;
    case AmSTRING:
	if ( valueStr != NULL) {
	    *value = ( void *) strdup( valueStr);
	    if ( *value == NULL) {
		DO_RETURN_ERROR( AmSTRDUP_FAILED);
	    }
	} else {
	    *value = NULL;
	}
	*copied = TRUE;
	break;
    case AmBOOLEAN:
	*value = ( void *) !strncasecmp( valueStr, "TRUE", 4);
	break;
    case AmCOLOR_CELL:
	AmColorCellPtr	colorCell;
	uint_t		red, green, blue;

/* color cells are scanned as RGB unsigned integer values, then 
   assigned to the unsigned chars in the color cell structure */

	colorCell = ( AmColorCellPtr) malloc( sizeof( AmColorCellRec));
	if ( colorCell == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	sscanf( valueStr, "%u %u %u", &red, &green, &blue);
	colorCell->red = red;
	colorCell->green = green;
	colorCell->blue = blue;
	*value = ( void *) colorCell;
	*copied = TRUE;
	break;
    case AmOBJECT:
    case AmFUNC_PTR:
    case AmVOID_PTR:

/* use the name register for void pointers and function pointers */

	*value = AmGetRegister( valueStr);
	break;
    case AmUNITS:
	if ( !AmGetUnitCode( valueStr, ( int *) value)) {
	    AmAddResourceToLog( "badUnits", tagStr, valueStr);
	    *value = ( void *) 0;
	}
	break;
    default:
	return( AmBAD_TYPE);
    }   
    return( AmOK);
}
