#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <anvil/project.h>
#include <anvil/files.h>
#include <anvil/tagstrs.h>
#include <anvil/record.h>
#include <anvil/strlib.h>
#include <anvil/choppath.h>
#include <anvil/header.h>
#include <anvil/log.h>

/* utility functions for dealing with tag: value format header files */


/****************************************************************************
 *
 * Description:
 * 	test if a file is a reasonably legitimate header, and fill in the 
 *	type of file if it is
 * Arguments:
 *  	filename - name of file to test.
 *  	typename - returned typename of file, referred to by AmNtype.
 * Returns:
 *	TRUE if the file is a legit header, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmIsHeader( const char	*filename,
	  char		*typeName)
{
    FILE	*fp;
    char	buffer[ 1024], ident[ 512], value[ 1024];
    int		numLines;
    AnvilBoolean	okFile;

    if (( fp = fopen( filename, "r")) == NULL) {
	return( FALSE);
    }
    numLines = 0;
    okFile = TRUE;
    strcpy( typeName, "UNKNOWN");
    do {
	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
	    break;
	}

/* continue if it's a comment */

	if ( buffer[ 0] == '#') {
	    numLines++;
	} else {

/* make sure there's a '\n' at the end of the buffer */

	    if ( buffer[ strlen( buffer) - 1] != '\n') {
		okFile = FALSE;
		break;
	    }

/* scan for the identifier and value */

	    if ( sscanf( buffer, "%[A-Za-z0-9]:%s", ident, value) < 1) {
		okFile = FALSE;
		break;
	    }
	    if ( strchr( buffer, ':') == NULL) {
		okFile = FALSE;
		break;
	    }
	    if ( !strcmp( ident, AmNtype)) {
		strcpy( typeName, value);
	    }
	}
    } while ( TRUE);

/* close the file and return the flag */

    fclose( fp);
    return( okFile);
}

/****************************************************************************
 *
 * Description:
 * 	read a tag-value header into a vector of strings.
 * Arguments:
 *	filename - file header to read.
 * Returns:
 *	NULL-terminated vector of strings pairs, where the AmFirst string
 *	is the tag, and the second string is the value.
 *
 ****************************************************************************/

char **
AmReadHeader( const char *filename)
{
    int         length, count;
    char        **vector, buffer[ BUFFERLEN], tag[ FIELDLEN], *value,
		filepath[ PATHLEN], *p;
    const char	*path;
    FILE        *fp;

    length = AmNumLines( filename);
    if ( length <= 0) {
        AmAddResourceToLog( "emptyFileError", filename);
        return( NULL);
    }

    if (( fp = fopen( filename, "r")) == NULL) {
        AmAddResourceToLog( "openFileError", filename);
        return( NULL);
    }

/* even entries are tags, odd entries are values */

    vector = ( char **) malloc(( length*2 + 1)*sizeof( char *));
    count = 0;

/* read each one of the entries until the end of the file */

    while ( TRUE) {
        if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
            break;
        }
        buffer[ strlen( buffer) - 1] = '\0';    /* get rid of the \n */
        if ( buffer[ 0] != '#') {               /* comment delmiter */
            if ( sscanf( buffer, "%[^:]", tag) != 1) {
                continue;
            }

/* scan to the colon and the AmFirst non-whitespace character for the value */

            for ( value = buffer; *value != ':'; value++);
            value++;
            for ( ; isspace( *value); value++) {};

/* truncate any whitespace characters after the value */

	    p = &value[ strlen( value) - 1]; 
	    if ( isspace( *p)) {
		for ( ; isspace( *p) && ( p >= value); p--)  {};
		p++;
		*p = '\0';
	    }

/* if the tag type is AmFILE, then check to see if the filename is absolute.
   If not, then see if putting the pathname for the header file in
   front will access the file */

            vector[ count++] = strdup( tag);
	    if ((( AmGetTagType( tag) == AmFILE) ||
		 AmPrefixTag( FILE_PREFIX, tag)) && 
		 !AmAbsolutePath( value)) {
		path = AmGetPath( filename);

/* check to see if the file is in the same directory as the header.  If
   so, then prepend the header's path, otherwise, just use the filename */

		if ( path != NULL) {
		    sprintf( filepath, "%s/%s", path, value);
		    if ( access( filepath, R_OK) == 0) {
			vector[ count++] = strdup( filepath);
		    } else {
			vector[ count++] = strdup( value);
		    }
		} else {
		    vector[ count++] = strdup( value);
		}
	    } else {
		vector[ count++] = strdup( value);
	    }
        }
    }
    fclose( fp);
    vector[ count] = NULL;
    return( vector);
}

/****************************************************************************
 *
 * Description:
 * 	given a vector of header strings and a tag, return the value 
 *	associated with the tag 
 * Arguments:
 *	headerStrings - NULL terminated vector of pairs of tag-value strings
 *	tag - tag to search for.
 * Returns:
 *	pointer to the string value associated with the tag, or NULL if 
 *	it wasn't found
 *
 ****************************************************************************/

const char *
AmGetHeaderStringTagValue( const char     **headerStrings,
			 const char     *tag)
{
    int ii;

    for ( ii = 0; headerStrings[ ii] != NULL; ii += 2) {
        if ( AmMatchTag( headerStrings[ ii], tag)) {
            return( headerStrings[ ii + 1]);
        }
    }
    return( NULL);
}
