#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <malloc.h>
#include <anvil/project.h>
#include <anvil/fns.h>
#include <anvil/rect.h>
#include <anvil/histogram.h>

/****************************************************************************
 *
 * Description:
 * 	do a binary search on a vector of real values, which is a little
 *	different than the integer binary search, because it tests for
 *	the non-exact match, in which case it returns the LOWER of the
 *	two bounding values.
 * Arguments:
 *	target - value to search for.
 *	vector - vector of values to search.
 *	length - length of the vector.
 * Returns:
 *	index of the value which is lower than the search value and 
 *	closest to it.
 *
 ****************************************************************************/

static int
kindaBinarySearch( float target,
                   float *vector,
                   int	 length)
{
    int	lo = 0, mid, hi = length - 1;

    if ( target < vector[ lo]) {
        return( 0);
    }

    if ( target > vector[ hi]) {
        return( length - 1);
    }

    while ( lo < hi) {
	mid = ( lo + hi + 1) >> 1;
	if ( target < vector[ mid]) {
	    hi = mid - 1;
	} else {
	    lo = mid;
	}
    }
    return( lo);
}

/****************************************************************************
 *
 * Description:
 * 	do a binary search on a vector of integers 
 * Arguments:
 *	target - target value to search for.
 *	vector - vector of integers to search.
 *	length - length of the vector.
 * Returns:
 *	either the index of the matching value, or the index of the
 *	next lower value.
 *
 ****************************************************************************/

static int 
binarySearch( int	target,
	      int	*vector,
	      int	length)
{
    int	lo = 0, mid, hi = length - 1;

    while ( lo < hi) {
        mid = ( lo + hi + 1) >> 1;
        if ( target < vector[ mid]) {
	    hi = mid - 1;
	} else if ( target > vector[ mid]) {
	    lo = mid + 1;
	} else {
	    return( mid);
	}
    }
    return( lo);
}

/****************************************************************************
 *
 * Description:
 * 	take a matrix of continuous values and do a histogram on it 
 * Arguments:
 *      matrix -	type-free matrix to histogram.
 *      zFn -		accessing function
 *      clientData -	additional parameter for the zFn.
 *      offset -	offset of value in data element.
 *      size -		size of matrix elements:
 *      rows, cols -	dimensions of matrix.
 *      histogram -	resulting histogram count.
 *      numValues -	number of values in histogram.
 *      bins -		bins to histogram.
 *      minClip -	do not include values below lowest value in count.
 *      maxClip -	clip values above the maximum.
 *
 ****************************************************************************/

void 
AmHistogramContinuousMatrix( uchar_t	*matrix,
			   AmAccessFn	zFn,
			   void		*clientData,
			   int		offset,
			   int		size,
			   int		rows,
			   int		cols,
			   AmHistRec	*histogram,
			   int		numValues,
			   float	*bins,
			   AnvilBoolean	minClip,
			   AnvilBoolean	maxClip)
{
    int		ii, index;
    float       min = bins[ 0], max = bins[ numValues - 1], value;
    uchar_t	*matend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = bins[ ii];
    }
    matend = matrix + rows*cols*size;
    while ( matrix < matend) {
	value = ( zFn)( matrix + offset, clientData);
	if ((( value >= min) || !minClip) && 
	    (( value <= max) || !maxClip)) {
	    index = kindaBinarySearch( value, bins, numValues);
	    histogram[ index].histValue++;
	}
	matrix += size;
    }
}   

/****************************************************************************
 *
 * Description:
 * 	take a matrix of continuous values and do a histogram on it 
 *	a submatrix defined by a rectangle 
 * Arguments:
 *  	matrix -	type-free matrix to histogram.
 *  	zFn -		accessing function
 *  	clientData -	additional parameter for the zFn.
 *  	offset -	offset of value in data element
 *  	size -		size of matrix elements:
 *  	rows, cols -	dimensions of matrix.
 *  	rect -		dimensions of submatrix.
 *  	histogram -	resulting histogram count.
 *  	numValues -	number of values in histogram.
 *  	bins -		bins to histogram.
 *  	minClip -	do not include values below lowest value in count.
 *  	maxClip -	clip values above the maximum.
 *
 ****************************************************************************/

void 
AmHistogramContinuousSubmatrix( uchar_t			*matrix,
			      AmAccessFn			zFn,
			      void			*clientData,
			      int			offset,
			      int			size,
			      int			rows,
			      int			cols,
			      AmRectangleRec const	*rect,
			      AmHistRec			*histogram,
			      int			numValues,
			      float			*bins,
			      AnvilBoolean			minClip,
			      AnvilBoolean			maxClip)
{
    int		ii, index;
    float       min = bins[ 0], max = bins[ numValues - 1], value;
    uchar_t	*matptr, *matend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = bins[ ii];
    }
    matrix += ( rect->x + rect->y*cols)*size;
    for ( ii = 0; ii < rect->height; ii++) {
    	matptr = matrix;
	matend = matptr + rect->width*size;
	while ( matptr < matend) {
	    value = ( zFn)( matptr + offset, clientData);
	    if ((( value >= min) || !minClip) && 
		(( value <= max) || !maxClip)) {
		index = kindaBinarySearch( value, bins, numValues);
		histogram[ index].histValue++;
	    }
	    matptr += size;
	}
	matrix += cols*size;
    }
}

/****************************************************************************
 *
 * Description:
 * 	take a matrix of thematic values and do a histogram on it 
 * Arguments:
 *  	matrix -	type-free matrix to histogram.
 *  	AmIntFn -		integer accessing function
 *  	clientData -	additional parameter for the zFn.
 *  	offset -	offset of value in daeta element 
 *  	size -		size of matrix elements:
 *  	rows, cols -	dimensions of matrix.
 *  	rect -		dimensions of submatrix.
 *  	histogram -	resulting histogram count.
 *  	numValues -	number of values in histogram.
 *  	table -		table to histogram over.
 *
 ****************************************************************************/

void 
AmHistogramThematicMatrix( uchar_t		*matrix,
			 AmIntAccessFn		AmIntFn,
			 void			*clientData,
			 int			offset,
			 int			size,
			 int			rows,
			 int			cols,
			 AmThematicHistPtr	histogram,
			 int			numValues,
			 int			*table)
{
    int		ii, value, index;
    uchar_t	*matend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = table[ ii];
    }
    matend = matrix + rows*cols*size;
    while ( matrix < matend) {
	value = ( AmIntFn)( matrix + offset, clientData);
	index = binarySearch( value, table, numValues);
	histogram[ index].histValue++;
	matrix += size;
    }
}

/****************************************************************************
 *
 * Description:
 * 	take a matrix of thematic values and do a histogram on a 
 *	submatrix defined by a rectangle 
 * Arguments:
 *  	matrix -	type-free matrix to histogram.
 *  	AmIntFn -		integer accessing function
 *  	clientData -	additional parameter for the zFn.
 *  	offset -	offset of value in data element.
 *  	size -		size of matrix elements:
 *  	rows, cols -	dimensions of matrix.
 *  	rect -		dimensions of submatrix.
 *  	histogram -	resulting histogram count.
 *  	numValues -	number of values in histogram.
 *  	table -		table to histogram over.
 *
 ****************************************************************************/

void 
AmHistogramThematicSubmatrix( uchar_t		*matrix,
			    AmIntAccessFn		AmIntFn,
			    void		*clientData,
			    int			offset,
			    int			size,
			    int			rows,
			    int			cols,
			    AmRectangleRec const	*rect,
			    AmThematicHistPtr	histogram,
			    int			numValues,
			    int			*table)
{
    int		ii, value, index;
    uchar_t	*matptr, *matend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = table[ ii];
    }
    matrix += ( rect->x + rect->y*cols)*size;
    for ( ii = 0; ii < rect->height; ii++) {
	matptr = matrix;
	matend = matptr + rect->width*size;
	while ( matptr < matend) {
	    value = ( AmIntFn)( matptr + offset, clientData);
	    index = binarySearch( value, table, numValues);
	    histogram[ index].histValue++;
	    matptr += size;
	}
	matrix += cols*size;
    }
}
	   
/****************************************************************************
 *
 * Description:
 * 	take a vector of scatter points with a continuous z-value and 
 *	generate a histogram from them 
 * Arguments:
 *  vector -		type-free matrix to histogram.
 *  zFn -		accessing function
 *  clientData -	additional parameter for the zFn.
 *  zOffset -		offset of z value in data element 
 *  size -		size of matrix elements:
 *  rows, cols -	dimensions of matrix.
 *  histogram -		resulting histogram count.
 *  numValues -		number of values in histogram.
 *  bins -		bins to histogram.
 *  minClip -		do not include values below lowest value in count.
 *  maxClip -		clip values above the maximum.
 *
 ****************************************************************************/

void 
AmHistogramContinuousScatter( uchar_t		*vector,
			    AmAccessFn		zFn,
			    void		*zClientData,
			    int			zOffset,
			    int			size,
			    int			rows,
			    AmHistRec		*histogram,
			    int			numValues,
			    float		*bins,
			    AnvilBoolean		minClip,
			    AnvilBoolean		maxClip)
{
    int		ii, index;
    uchar_t	*vecend;
    float	min = bins[ 0], max = bins[ numValues - 1], value;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = bins[ ii];
    }
    vecend = vector + size*rows;
    while ( vector < vecend) {
	value = ( zFn)( vector + zOffset, zClientData); 
	if ((( value >= min) || !minClip) && 
	    (( value <= max) || !maxClip)) {
	    index = kindaBinarySearch( value, bins, numValues);
	    histogram[ index].histValue++;
	}
	vector += size;
    }
}
	   
/****************************************************************************
 *
 * Description:
 * 	take a vector of scatter points with a continuous z value and 
 *	generate a histogram from a subsection defined by a coordinate 
 *	rectangle 
 * Arguments: 
 *  	xFn, yFn -		x and y functions
 *  	xOffset, yOffset -	x and y data offsets
 *  	xyClientData -		x,y client data.
 *  	vector -		type-free matrix to histogram.
 *  	zFn -			accessing function
 *  	zClientData -		additional parameter for the zFn.
 *  	size -			size of matrix elements:
 *  	zOffset -		offset of z value in data element.
 *  	rows, cols -		dimensions of matrix.
 *  	rect -			dimensions of submatrix.
 *  	histogram -		resulting histogram count.
 *  	numValues -		number of values in histogram.
 *  	bins -			bins to histogram.
 *  	minClip -		do not include values below lowest value 
 *				in count.
 *  	maxClip -		clip values above the maximum.
 *
 ****************************************************************************/

void AmHistogramContinuousSubscatter( uchar_t		*vector,
				    AmAccessFn		xFn,
				    AmAccessFn		yFn,
				    int			xOffset,
				    int			yOffset,
				    void		*xyClientData,
				    AmAccessFn		zFn,
				    int			zOffset,
				    void		*zClientData,
				    int			size,
				    int			rows,
				    AmRealRectRec const	*rect,
				    AmHistRec		*histogram,
				    int			numValues,
				    float		*bins,
				    AnvilBoolean		minClip,
				    AnvilBoolean		maxClip)
{
    int		ii, index;
    uchar_t	*vecend;
    float	min = bins[ 0], max = bins[ numValues - 1], value;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = bins[ ii];
    }
    vecend = vector + size*rows;
    while ( vector < vecend) {
	if ( AmInRealRect( rect, ( xFn)( vector + xOffset, xyClientData),
			       ( yFn)( vector + yOffset, xyClientData))) {
	    value = ( zFn)( vector + zOffset, zClientData); 
	    if ((( value >= min) || !minClip) && 
		(( value <= max) || !maxClip)) {
		index = kindaBinarySearch( value, bins, numValues);
		histogram[ index].histValue++;
	    }
	}
	vector += size;
    }
}
	   
/****************************************************************************
 *
 * Description:
 * 	take a vector of scatter points with a thematic z value and generate 
 *	a histogram from them 
 * Arguments:
 *  	vector -	type-free matrix to histogram.
 *  	zFn -		accessing function
 *  	clientData -	additional parameter for the zFn.
 *  	size -		size of matrix elements:
 *  	rows, cols -	dimensions of matrix.
 *  	histogram -	resulting histogram count.
 *  	numValues -	number of values in histogram.
 *  	bins -		bins to histogram.
 *
 ****************************************************************************/

void 
AmHistogramThematicScatter( uchar_t		*vector,
			  AmIntAccessFn		AmIntFn,
			  int			offset,
			  void			*clientData,
			  int			size,
			  int			rows,
			  AmThematicHistPtr	histogram,
			  int			numValues,
			  int			*table)
{
    int		ii, index;
    uchar_t	*vecend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = table[ ii];
    }
    vecend = vector + size*rows;
    while ( vector < vecend) {
	index = binarySearch(( AmIntFn)( vector + offset, clientData), 
			     table, numValues);
	histogram[ index].histValue++;
	vector += size;
    }
}
	   
/****************************************************************************
 *
 * Description:
 * 	take a subset vector of scatter points defined with a thematic z 
 *	value and generate a histogram from them 
 * Arguments:
 *  	vector -		type-free matrix to histogram.
 *	xFn, yFn -		x and y functions
 * 	xOffset, yOffset -	x and y data offsets
 * 	xyClientData -		x,y client data.
 * 	AmIntFn -			z value integer lookup function
 * 	zClientData -		additional parameter for the AmIntFn.
 * 	size -			size of matrix elements:
 * 	zOffset -		offset of z value 
 * 	rows, cols -		dimensions of matrix.
 * 	rect -			dimensions of submatrix.
 * 	histogram -		resulting histogram count.
 * 	numValues -		number of values in histogram.
 * 	bins -			bins to histogram.
 *
 ****************************************************************************/

void 
AmHistogramThematicSubscatter( uchar_t		*vector,
			     AmAccessFn		xFn,
			     AmAccessFn		yFn,
			     int		xOffset,
			     int		yOffset,
			     void		*xyClientData,
			     AmIntAccessFn	AmIntFn,
			     int		intOffset,
			     void		*intClientData,
			     int		size,
			     int		rows,
			     AmRealRectRec const	*rect,
			     AmThematicHistPtr	histogram,
			     int		numValues,
			     int		*table)
{
    int		ii, index;
    uchar_t	*vecend;

    for ( ii = 0; ii < numValues; ii++) {
	histogram[ ii].histValue = 0;
	histogram[ ii].dataValue = table[ ii];
    }
    vecend = vector + size*rows;
    while ( vector < vecend) {
	if ( AmInRealRect( rect, ( xFn)( vector + xOffset, xyClientData),
			       ( yFn)( vector + yOffset, xyClientData))) {
	    index = binarySearch(( AmIntFn)( vector + intOffset, intClientData), 
				 table, numValues);
	    histogram[ index].histValue++;
	}
	vector += size;
    }
}
