#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* line intersection utilities. */

#include <anvil/project.h>
#include <anvil/geometry.h>

/****************************************************************************
 *
 * Description:
 * 	return the intersection  of two lines (x0,y0) to (x1,y1)
 *	and (x2,y2) to (x3,y3)
 * Arguments:
 *	x0, x1, x1, y1 - endpoints of AmFirst line.
 *	x2, y2, x3, y3 - endpoints of the second line.
 *	x, y - returned intersection, if the lines intersect, garbage
 *	otherwise.
 * Returns:
 *	TRUE if the lines intersect, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean AmLineIntersection( float x0,
			  float	y0,
			  float x1,
			  float	y1,
			  float x2,
			  float	y2,
			  float x3,
			  float	y3,
			  float *x,
			  float	*y)
{
    float 	m1, m2, b1, b2;

/* 1st line is vertical, get y value from other line if it isn't vertical
   as well */

    if ( x0 == x1) {
	if ( x2 == x3) {
	    return( FALSE);
	}
	m2 = ( y3 - y2)/ ( x3 - x2);

/* y - y2 = m*(x - x2) */

	*x = x0;
	*y = m2*(x0-x2) + y2;
	return( AmBetweenFloat( *x, x0, x1) && AmBetweenFloat( *x, x2, x3) &&
		AmBetweenFloat( *y, y0, y1) && AmBetweenFloat( *y, y2, y3));
    }

/* 2nd line is vertical, get y value from other line */

    if ( x2 == x3) {
	m1 = ( y1 - y0)/ ( x1 - x0);

/* y - y0 = m*( x - x0) */

	*x = x2;
	*y = m1*(x2-x0) + y0;
	return( AmBetweenFloat( *x, x0, x1) && AmBetweenFloat( *x, x2, x3) &&
		AmBetweenFloat( *y, y0, y1) && AmBetweenFloat( *y, y2, y3));
    }

/* normal case: change equations to form y = mx + b */

    m1 = ( y1 - y0)/( x1 - x0);
    m2 = ( y3 - y2)/( x3 - x2);

/* equal slopes implies parallel */

    if ( m1 == m2) {
	return( FALSE);
    }
    b1 = y0 - m1*x0;
    b2 = y2 - m2*x2;

/* y = m1*x + b1 && y = m2*x + b2 -> x = ( b2 - b1) / ( m1 - m2) */

    *x =  ( b2 - b1) / ( m1 - m2);
    *y = *x*m1 + b1;
    return( AmBetweenFloat( *x, x0, x1) && AmBetweenFloat( *x, x2, x3) &&
	    AmBetweenFloat( *y, y0, y1) && AmBetweenFloat( *y, y2, y3));
}

/****************************************************************************
 *
 * Description:
 * 	return the intersection  of two lines (x0,y0) to (x1,y1)
 *	and (x2,y2) to (x3,y3)
 * Arguments:
 *	x0, x1, x1, y1 - endpoints of AmFirst line.
 *	x2, y2, x3, y3 - endpoints of the second line.
 *	x, y - returned intersection, if the lines intersect, garbage
 *	otherwise.
 * Returns:
 *	TRUE if the lines intersect, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean AmIntLineIntersection( int x0,
			     int y0,
			     int x1,
			     int y1,
			     int x2,
			     int y2,
			     int x3,
			     int y3,
			     int *x,
			     int *y)
{
    float 	m1, m2, b1, b2;

/* 1st line is vertical, get y value from other line if it isn't vertical
   as well */

    if ( x0 == x1) {
	if ( x2 == x3) {
	    return( FALSE);
	}
	m2 = (( float) ( y3 - y2))/(( float) ( x3 - x2));

/* y - y2 = m*(x - x2) */

	*x = x0;
	*y = (int)(m2*( x0 - x2)) + y2;
	return( AmBetween( *x, x2, x3) &&
		AmBetween( *y, y0, y1) && AmBetween( *y, y2, y3));
    }

/* 2nd line is vertical, get y value from other line */

    if ( x2 == x3) {
	m1 = (( float) ( y1 - y0))/(( float) ( x1 - x0));

/* y - y0 = m*( x - x0) */

	*x = x2;
	*y = (int)(m1*( x2 - x0)) + y0;
	return( AmBetween( *x, x0, x1) && 
		AmBetween( *y, y0, y1) && AmBetween( *y, y2, y3));
    }

/* normal case: change equations to form y = mx + b */

    m1 = (( float) ( y1 - y0))/(( float) ( x1 - x0));
    m2 = (( float) ( y3 - y2))/(( float) ( x3 - x2));

/* equal slopes implies parallel */

    if ( m1 == m2) {
	return( FALSE);
    }
    b1 = y0 - m1*x0;
    b2 = y2 - m2*x2;

/* y = m1*x + b1 && y = m2*x + b2 -> x = ( b2 - b1) / ( m1 - m2) */

    *x =  (int)(( b2 - b1) / ( m1 - m2));
    *y = (int)((float)*x*m1 + b1);
    return( AmBetween( *x, x0, x1) && AmBetween( *x, x2, x3) &&
	    AmBetween( *y, y0, y1) && AmBetween( *y, y2, y3));
}
