#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <math.h>
#include <anvil/geometry.h>
#include <anvil/macros.h>

/****************************************************************************
 *
 * Description:
 *      given a line, and an x value, return the y value for x. 
 *	(integer form)
 * Arguments:
 *      x0, y0, x1, y1 - line endpoints.
 *	x - x value.
 * Returns:
 *      the y value that corresponds to the x value.
 *
 ****************************************************************************/

int 
AmLineat( int x0,
	int y0,
	int x1,
	int y1,
	int x)
{
    float m;

    if ( x1 == x0) {
	return(( y0 + y1) / 2);		/* as good a value as any */
    }
    m = ( y1 - y0)/( x1 - x0);
    return( (int)(m*( x - x0)) + y0);
}

/****************************************************************************
 *
 * Description:
 * 	find line at y ( inverse of AmLineat, also integer form) 
 * Arguments:
 *      x0, y0, x1, y1 - line endpoints.
 *	y - y value.
 * Returns:
 *      the x value that corresponds to the y value.
 *
 ****************************************************************************/

int 
AmLineaty( int x0,
	 int y0,
	 int x1,
	 int y1,
	 int y)
{
    float m;

    if ( y1 == y0) {
	return( ( x0 + x1) / 2);	/* as good a value as any */
    }
    m = ( x1 - x0) / ( y1 - y0);
    return( (int)(m*( y - y0)) + x0);
}

/****************************************************************************
 *
 * Description:
 * 	find slope intercept solution, given an integer endpoint and
 *	a floating point slope.
 * Arguments:
 *	x0, y0 - line endpoint.
 *	m - line slope.
 *	x - x value.
 * Returns:
 *      the y value at x, given the slope and intercept.
 *
 ****************************************************************************/

int 
AmSlopeat( int 	x0,
	 int 	y0,
	 float 	m,
	 int 	x)
{
    return( (int)(m*(x - x0)) + y0);
}

/****************************************************************************
 *
 * Description:
 * 	find the inverse slope intercept solution, given an integer 
 *	endpoint and *	a floating point slope.
 * Arguments:
 *	x0, y0 - line endpoint.
 *	m - line slope.
 *	y - y value
 * Returns:
 *      the x value at y, given the slope and intercept.
 *
 ****************************************************************************/

int 
AmSlopeaty( int 	x0,
	  int 	y0,
	  float	m,
	  int	y)
{
    return((int)((float)( y - y0)/m) + x0);
}

/****************************************************************************
 *
 * Description:
 * 	given an x,y point and a slope in a rectangle, return the xy points 
 *	that the line describes.
 * Arguments:
 *	x0, y0 - line endpoint.
 *	m - line slope.
 * 	rectX0, rectY0, rectWidth, rectHeight - rectangle position and
 *	dimensions.
 *	lineX0, lineY0, lineX1, lineY1 - line endpoints.
 *
 ****************************************************************************/

void 
AmLineSlopeRect( int 	x0,
	       int	y0,
	       float	m,
	       int	rectX0,
	       int	rectY0,
	       int	rectWidth,
	       int	rectHeight,
	       int	*lineX0,
	       int	*lineY0,
	       int	*lineX1,
	       int	*lineY1)
{

/* x0, y0 could be intersection with y = rectX0 axis */

    *lineX0 = rectX0;
    *lineY0 = AmSlopeat( x0, y0, m, *lineX0);

/* check to see if it intersects min or max horiz of rectangle */

    if ( *lineY0 < rectY0) {
	*lineY0 = rectY0;
	*lineX0 = AmSlopeaty( x0, y0, m, *lineY0);
    } else if ( *lineY0 > rectY0 + rectHeight) {
	*lineY0 = rectY0 + rectHeight;
	*lineX0 = AmSlopeaty( x0, y0, m, *lineY0);
    }

/* x1, y1, could be intersection with y = rectX0 + rectHeight */

    *lineX1 = rectX0 + rectWidth;
    *lineY1 = AmSlopeat( x0, y0, m, *lineX1);

/* check to see if it intersects min or max horiz of rectangle */

    if ( *lineY1 < rectY0) {
	*lineY1 = rectY0;
	*lineX1 = AmSlopeaty( x0, y0, m, *lineY1);
    } else if ( *lineY1 > rectY0 + rectHeight) {
	*lineY1 = rectY0 + rectHeight;
	*lineX1 = AmSlopeaty( x0, y0, m, *lineY1);
    }
}

/****************************************************************************
 *
 * Description:
 * 	given a line from (x0, y0) to ( x1, y1), what is the distance 
 *	to x2, y2?
 * Arguments:
 *	x0, y0, x1, y1 - line endpoints.
 *	x2, y2 - point to calculate the distance for.
 * Returns:
 *	the distance from the point to the line.
 *
 ****************************************************************************/

float 
AmLineDist( float	x0,
	  float	y0,
	  float	x1,
	  float	y1,
	  float	x2,
	  float	y2)
{
    float	m, x, y;

    if ( x1 == x0) {
	return( fabs( x2 - x1));
    } else if ( y1 == y0) {
	return( fabs( y2 - y1));
    } else {
	m = ( y1 - y0)/( x1 - x0);
	x = (x2/m + y2 - y0 + m*x0)/( m + 1/m);
	y = m*(x - x0) + y0;
	return( sqrt( SQRDIST( x, y, x2, y2)));
    }
}

/****************************************************************************
 *
 * Description:
 * 	given a line from (x0, y0) to (x1, y1), and a line from
 *	(x1, y1) to (x2, y2), what is the angle at (x1, y1) in radians?
 * Arguments:
 *	x0, y0 - point 1.
 *	x1, y1 - point 2.
 *	x2, y2 - point 3.
 * Returns:
 * 	The angle in radians at x1, y1, from 0 to M_PI.
 *
 ****************************************************************************/

float AmLineAngle( float	x0,
		 float	y0,
		 float	x1,
		 float	y1,
		 float	x2,
		 float	y2)
{
    float	angle1, angle2, angle;
	
    angle1 = atan2( x0 - x1, y0 - y1);
    angle2 = atan2( x2 - x1, y2 - y1);
    angle = fabs( angle1 - angle2);
    if ( angle > M_PI) {
	angle = 2*M_PI - angle;
    }
    return( angle);
}
