#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions for contouring matrices into polygons.  The algorithm is
   admittedly cryptic, because we did a lot of hand-hacking to get a 
   lot of speed */


#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <malloc.h>
#include <memory.h>
#include <anvil/ratio.h>
#include <anvil/list.h>
#include <anvil/macros.h>
#include <anvil/contour.h>
#include <anvil/tagstrs.h>

/* EDGE1, EDGE2 are horiz or vert (left or right hand), EDGE3 is diagonal */

#define EDGE1   0x1
#define EDGE2   0x2
#define EDGE3   0x4
#define SETMARK( m, c, i, j, id) (m[ (i)*(c) + j] = id)
#define ISMARK( m, c, i, j) (m[ (i)*(c) + j])
#define ISOK( m, r, c, i, j, id) (( i >= 0) && ( i < r - 1) && ( j >= 0) && \
				  ( j < c - 2) && ( m[ ( i)*( c) + j] != id))
#define NEXT( a, b) (((a) == (b) - 1) || ((a) == (b) + 1))

static float points[ 10];

static const char *descriptionFile =
        "$ANVIL_CONFIG/descriptions/contour_polygon_format";


/* local point list used to prevent memory fragmentation... */

static AmListPtr	PT_LST = NULL;

/****************************************************************************
 *
 * Description:
 * 	rescale a contour list according to a new coordinate system.  The
 *	original contour list is in the matrix coordinate system (rows,cols),
 *	and needs to be scaled to the matrices' world coordinate system.
 * Arguments:
 *	conList - list of contour polygons.
 *	rows, col - dimensions of the source matrix.
 *	xmin, ymin, xmax, ymax - coordinates to transform to
 *	flipRows, flipCols - control origin of source matrix .
 *
 ****************************************************************************/

void 
AmScaleContourList( AmListPtr	conList,
		    int		rows,
		    int		cols,
		    float	xmin,
		    float	ymin,
		    float	xmax,
		    float	ymax,
		    AnvilBoolean	flipRows,
		    AnvilBoolean	flipCols)
{
    float		xscale, yscale;
    AmPolygonPtr	contour;
    AmCoordPtr		pt, ptEnd;

    xscale = (xmax - xmin)/( cols - 1);
    yscale = (ymax - ymin)/( rows - 1);

/* calculate the polygon's extents while we're at it */

    for ( ;conList != NULL; conList = conList->next) {
	contour = ( AmPolygonPtr) conList->data;
	contour->xmin = contour->pts[ 0].x;
	contour->ymin = contour->pts[ 0].y;
	contour->xmax = contour->pts[ 0].x;
	contour->ymax = contour->pts[ 0].y;
	ptEnd = contour->pts + contour->numPts;
	for ( pt = contour->pts; pt < ptEnd; pt++) {
	    if ( flipCols) {
		pt->x = xscale*(( cols - 1) - pt->x) + xmin;
	    } else {
		pt->x = xscale*pt->x + xmin;
	    }
	    if ( flipRows) {
		pt->y = yscale*pt->y + ymin;
	    } else {
		pt->y = yscale*(( rows - 1) - pt->y) + ymin;  
	    }
	    if ( pt->x < contour->xmin) {
		contour->xmin = pt->x;
	    }
	    if ( pt->y < contour->ymin) {
		contour->ymin = pt->y;
	    }
	    if ( pt->x > contour->xmax) {
		contour->xmax = pt->x;
	    }
	    if ( pt->y > contour->ymax) {
		contour->ymax = pt->y;
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	free the storage used by a list of contour lists.
 * Arguments:
 *	conList - list of list of countour polygons.
 * 
 ****************************************************************************/

void 
AmFreeContourList( AmListPtr conList)
{
    AmListPtr	conSet, conPtr;
    AmPolygonPtr	Polygon;

    AN_DEBUG( 20, ("freeing %d levels\n", AmLengthLst( conList)));
    for ( conPtr = conList; conPtr != NULL; conPtr = conPtr->next) {
	AN_DEBUG( 20, ( "freeing %d polygons\n", 
			AmLengthLst(( AmListPtr) conPtr->data)));
	for ( conSet = ( AmListPtr) conPtr->data;
	      conSet != NULL;
	      conSet = conSet->next) {
	    Polygon = ( AmPolygonPtr) conSet->data;
	    free( Polygon->pts);
	    free( Polygon);
	}
	AmJustFreeLst(( AmListPtr) conPtr->data);
    }
    AmJustFreeLst( conList);
}

/****************************************************************************
 *
 * Description:
 * 	return the zpoints for an indexed triangle.
 * Arguments:
 *	matrix - source matrix.
 *	zFn, clientData - element accessing function and client data.
 *	size - element size.
 *	cols - number of columns in the matrix.
 *	ii, jj - row and column (jj times 2, because the "mark grid" of visited
 *	points is twice as wide as the matrix.
 * 
 ****************************************************************************/

static float *
createTri( uchar_t       *matrix,
	   AmAccessFn      zFn,
	   void	         *clientData,
	   int	         size,
	   int           cols,
	   int           ii,
	   int           jj)
{

/* j in markGrid is 2 times matrix dimensions */

    matrix += (( ii*cols) + ( jj >> 1))*size; 	
    if ( ODD( jj)) {
        points[ 0] = ( zFn)( matrix, clientData);
	matrix += size;
        points[ 1] = ( zFn)( matrix, clientData);
        matrix += cols*size;
        points[ 2] = ( zFn)( matrix, clientData);
    } else {
        points[ 0] = ( zFn)( matrix, clientData);
        matrix += cols*size;
        points[ 1] = ( zFn)( matrix, clientData);
	matrix += size;
        points[ 2] = ( zFn)( matrix, clientData);
    }
    return( points);
}

/****************************************************************************
 *
 * Description:
 * 	test whether a triangle (defined as a set of 3 z values) is 
 *	intersected by a zplane. 
 * Arguments:
 *	points - array of 3 z-values to test.
 *	z - z plane to test against.
 * Returns:
 *	a flag set which shows the edges which have been intersected by
 *	the plane (composed of EDGE1, EDGE2, EDGE3)
 * 
 ****************************************************************************/

static int 
testTri( float       *points,
	 float       z)
{
    int	flag = 0;

    if (( points[ 0] > z) != ( points[ 1] > z)) {
	flag |= EDGE1;
    }
    if (( points[ 1] > z) != ( points[ 2] > z)) {
	flag |= EDGE2;
    }
    if (( points[ 2] > z) != ( points[ 0] > z)) {
	flag |= EDGE3;
    }
    return( flag);
}

/****************************************************************************
 *
 * Description:
 * 	convienience function to create a point from an x,y pair.
 * Arguments:
 *	x,y - xy point to make up the coordinate.
 * Returns:
 *	An allocated coordPtr structure.
 * Modifications;
 *	better check code has to be added to this and its caller to check for
 *	malloc failures.
 * 
 ****************************************************************************/

static AmCoordPtr 
mkpt( float      x,
      float      y)
{
    AmCoordPtr      pt;

    if (( pt = ( AmCoordPtr) malloc( sizeof( AmCoordRec))) == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( NULL);
    }
    pt->x = x;
    pt->y = y;
    return( pt);
}

/****************************************************************************
 * 
 * Description:
 *	is the triangle defined by i1,j1 next to the triangle defined
 *	by i2,j2?
 * Arguments:
 *	i1, j1 - row, column in the mark grid of the first triangle
 *	i2, j2 - row, column in the mark grid of the second triangle
 * 
 ****************************************************************************/

static AnvilBoolean 
nextTo( int	i1,
        int	j1,
        int	i2,
        int	j2)
{
    return((( i1 == i2) && NEXT( j1, j2)) ||
	   ( ODD( j1) && ( i2 == i1 - 1) && ( j2 == j1 - 1)) ||
	   (( i2 == i1 + 1) && ( j2 == j1 + 1)));
}

/****************************************************************************
 * 
 * Description:
 * 	perform the triangle intersection with the plane and return 
 *	the flags for which edges were intersected.
 * Arguments:
 *	points - triangle z values 
 *	ii - UL corner of tri. 
 *	jj - odd(j) -> right else left tri 
 *	z -  intersecting plane 
 *	intpts - xy intersects. z is intersects plane 
 * Returns:
 *	edge flags for which edge intersected.
 * 
 ****************************************************************************/

static int 
intersectTri( float   	*points,
	      int     	ii,
	      int     	jj,
	      float   	z,
	      float   	*intpts,
	      int	lastFlag)
{
    float       t, sValues[ 3], s;
    int 	i0 = ( points[ 0] > z), 
		i1 = ( points[ 1] > z), 
		i2 = ( points[ 2] > z),
		indexCount, intIndex[ 3];

/* right and left handed triangles intersect differently.  */

    indexCount = 0;
    intIndex[ 0] = 0x0;
    if ( ODD( jj)) {
        jj >>= 1;                /* j is double col index from mark grid */
        if (( i0 != i1) && ( lastFlag != EDGE2)) {   /* edge intersects */
	    s = z - points[ 0];
            t = s/( points[ 1] - points[ 0]);
            intpts[ 0] = jj + t;            /* reduced from x0+(x1-x0)*t */
            intpts[ 1] = ii;                 /* reduced from y0+(y1-y0)*t */
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE1;
        }
        if (( i1 != i2) && ( lastFlag != EDGE1)) {   
	    s = z - points[ 1];
            t = s/( points[ 2] - points[ 1]);
            intpts[ 2] = jj + 1;
            intpts[ 3] = ii + t;
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE2;
        }
        if (( i2 != i0) && ( lastFlag != EDGE3)) {   
	    s = z - points[ 2];
            t = s/( points[ 0] - points[ 2]);
            intpts[ 4] = jj + 1 - t;         
            intpts[ 5] = ii + 1 - t;
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE3;
        }
    } else {
        jj >>= 1;                /* j is double col index from mark grid */
        if (( i0 != i1) && ( lastFlag != EDGE2)) {   /* edge intersects */
	    s = z - points[ 0];
            t = s/( points[ 1] - points[ 0]);
            intpts[ 0] = jj;            /* reduced from x0+(x1-x0)*t */
            intpts[ 1] = ii + t;         /* reduced from y0+(y1-y0)*t */
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE1;
        }
        if (( i1 != i2) && ( lastFlag != EDGE1)) {   
	    s = z - points[ 1];
            t = s/( points[ 2] - points[ 1]);
            intpts[ 2] = jj + t;         
            intpts[ 3] = ii + 1;
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE2;
        }
        if (( i2 != i0) && ( lastFlag != EDGE3)) {   
	    s = z - points[ 2];
            t = s/( points[ 0] - points[ 2]);
            intpts[ 4] = jj + 1 - t;         
            intpts[ 5] = ii + 1 - t;
	    sValues[ indexCount] = s;
	    intIndex[ indexCount++] = EDGE3;
        }
    }

/* if there's more than 1 intersection, make sure the index shows the
   best one */

    if ( indexCount > 1) {
        if ( fabs( sValues[ 1]) < fabs( sValues[ 0])) {
            return( intIndex[ 1]);
        }
    }
    return( intIndex[ 0]);
}
   
/****************************************************************************
 * 
 * Description:
 * 	given the current triangle, try and find the next triangle in
 *	the contour.
 * Arguments:
 *	lastflag - edge flag for the intersection of the last triangle,
 *	this dictates the next triangle to be tested.
 *	matrix - matrix to contour.
 *	zFn, clientData - accessing function and client data to convert
 *	the source matrix into float values.
 *	size - size of each element in the source matrix.
 *	rows, cols - dimensions of the source matrix.
 *	ii, jj - current row and column, where column is multiplied by
 *	two, because there are twice as many triangles in a row as source
 *	points.
 *	z - z plane value to test against.
 *	markGrid - grid of flags for tested points.
 *	nexti, nextj - row and column of the next triangle.
 *	contourId - unique contour identifier.
 * Returns:
 *	TRUE if there's a new triangle, FALSE if there isn't
 * 
 ****************************************************************************/

static AnvilBoolean 
nextTri( int	 	lastflag,
	 uchar_t     	*matrix,
	 AmAccessFn	zFn,
	 void		*clientData,
	 int	 	size,
	 int         	rows,
	 int         	cols,
	 int         	ii,
	 int         	jj,
	 float       	z,
	 uchar_t        *markGrid,
	 int         	*nexti,
	 int         	*nextj,
	 uchar_t	contourId)
{
    int		colst2 = cols << 1;

/* depending on the last point intersected, we have to test only one triangle */

    if ( ODD( jj)) {

/* last triangle was odd, new triangle will be even */

	switch ( lastflag) {
	case EDGE1:		/* horizontal edge: triangle above */
	    if ( ISOK( markGrid, rows, colst2, ii - 1, jj - 1, contourId) &&	
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii - 1, jj - 1), z)) {
		*nexti = ii - 1;
		*nextj = jj - 1;
		return( TRUE);
	    }
	    break;
	case EDGE2:		/* vertical edge: triangle right */
	    if ( ISOK( markGrid, rows, colst2, ii, jj + 1, contourId) &&
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii, jj + 1), z)) {
		*nexti = ii;
		*nextj = jj + 1;
		return( TRUE);
	    }
	    break;
	case EDGE3:		/* diagonal edge: triangle left */
	    if ( ISOK( markGrid, rows, colst2, ii, jj - 1, contourId) &&
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii, jj - 1), z)) {
		*nexti = ii;
		*nextj = jj - 1;
		return( TRUE);
	    }
	    break;
	default:
	    break;
	}
    } else {

/* last triangle was even, new triangle will be odd */

	switch ( lastflag) {
	case EDGE1:		/* vertical edge: triangle left */
	    if ( ISOK( markGrid, rows, colst2, ii, jj - 1, contourId) &&
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii, jj - 1), z)) {
		*nexti = ii;
		*nextj = jj - 1;
		return( TRUE);
	    }
	    break;
	case EDGE2:		/* horizontal edge: triangle below */
	    if ( ISOK( markGrid, rows, colst2, ii + 1, jj + 1, contourId) &&
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii + 1, jj + 1), z)) {
		*nexti = ii + 1;
		*nextj = jj + 1;
		return( TRUE);
	    }
	    break;
	case EDGE3:		/* diagonal edge: triangle right */
	    if ( ISOK( markGrid, rows, colst2, ii, jj + 1, contourId) &&
		 testTri( createTri( matrix, zFn, clientData, size, cols, 
				     ii, jj + 1), z)) {
		*nexti = ii;
		*nextj = jj + 1;
		return( TRUE);
	    }
	    break;
	default:
	    break;
	}
    }

/* mark where it ended */

    *nexti = ii;
    *nextj = jj;
    return( FALSE);
}
                
/****************************************************************************
 * 
 * Description:
 * 	given a starting matrix location, follow the contour back to the
 *  	starting point or the edge of the matrix.
 * Arguments:
 *	lastflag - edge flag for the intersection of the triangle,
 *	this dictates the next triangle to be tested.
 *	matrix - matrix to contour.
 *	zFn, clientData - accessing function and client data to convert
 *	the source matrix into float values.
 *	size - size of each element in the source matrix.
 *	rows, cols - dimensions of the source matrix.
 *	ii, jj - current row and column, where column is multiplied by
 *	two, because there are twice as many triangles in a row as source
 *	points.
 *	markGrid - grid of flags for tested points.
 *	z - z plane value to test against.
 *	closed - TRUE if the polygon is closed, FALSE if it terminated
 *	on an edge.
 *	contourId - contour identifier.
 * Returns:
 *	the polygon representing the contour.
 * 
 ****************************************************************************/

static AmPolygonPtr 
genContour( int       	lastflag,
	    uchar_t   	*matrix,   
	    AmAccessFn  zFn,
	    void      	*clientData,
	    int       	size,
	    int       	rows,
	    int       	cols,
	    int       	ii,        
	    int       	jj,
	    uchar_t     *markGrid, 
	    float     	z,         
	    AnvilBoolean   	*closed,
	    uchar_t	contourId)  
{
    float       	intpts[ 6];
    int         	intflags, colst2 = cols << 1, starti = ii, startj = jj,
			length = 0;
    AmListPtr     	ptlist = NULL, tmp;
    AmCoordRec		pt;
    AmPolygonPtr	Polygon;

    do {
        SETMARK( markGrid, colst2, ii, jj, contourId);
        intflags = intersectTri( createTri( matrix, zFn, clientData, size, 
					    cols, ii, jj),
				 ii, jj, z, intpts, lastflag);

/* the edge numbers 1 & 2 reverse for odd & even triangles */

        if ( intflags & EDGE1) {
	    pt.x = intpts[ 0];
	    pt.y = intpts[ 1];
            lastflag = EDGE1;
        } else if ( intflags & EDGE2) {
	    pt.x = intpts[ 2];
	    pt.y = intpts[ 3];
            lastflag = EDGE2;
        } else if ( intflags & EDGE3) {
	    pt.x = intpts[ 4];
	    pt.y = intpts[ 5];
            lastflag = EDGE3;
        }

/* if we've run out of pre-allocated points create a new one, otherwise,
   take one from our 'free list' */

	if ( PT_LST == NULL) {
	    ptlist = AmInsertNodeLst( mkpt( pt.x, pt.y), ptlist);
	} else {
	    *(( AmCoordPtr) PT_LST->data) = pt;
	    tmp = PT_LST;
	    PT_LST = PT_LST->next;
	    tmp->next = ptlist;
	    ptlist = tmp;
	}
	length++;

/* if we hit the first edge, go back to the start and walk to the other edge */

    } while ( nextTri( lastflag, matrix, zFn, clientData, size, rows, cols, 
		       ii, jj, z, markGrid, &ii, &jj, contourId));

/* mark the flag for whether the contour was closed */

    *closed = nextTo( ii, jj, starti, startj);

/* set the 're-use' list for the next call */

    if ( PT_LST == NULL) {
	PT_LST = ptlist;
    } else {
	for ( tmp = PT_LST; tmp->next != NULL; tmp = tmp->next);
	tmp->next = ptlist;
    }
    Polygon = ( AmPolygonPtr) calloc( 1, sizeof( AmPolygonRec));
    Polygon->pts = ( AmCoordPtr) AmCopyVectorLst( ptlist, sizeof( AmCoordRec));
    Polygon->numPts = length;
    return( Polygon);
}

/****************************************************************************
 * 
 * Description:
 * 	given a matrix, its dimensions and an intersecting z-plane, generate a 
 *  	list of contour lines. 
 * Arguments:
 *	matrix - matrix to contour.
 *	zFn, clientData - accessing function and client data to convert
 *	the source matrix into float values.
 *	size - size of each element in the source matrix.
 *	markGrid - grid of flags for tested points.
 *	rows, cols - dimensions of the source matrix.
 *	z - z plane value to test against.
 *	edgeList - returned list of edge contour polygons.
 *	inList - returned list of internal contour polygons.
 *	maxLength - maximum number of points in a returned contour polygon.
 * Returns:
 *	TRUE.
 * Modifications:
 *	should return an error code for malloc failure.
 * 
 ****************************************************************************/

AnvilBoolean 
AmMakeContours( uchar_t      	*matrix,
		AmAccessFn     	zFn,
		void	   	*clientData,
		int	   	size,
		uchar_t	   	*markGrid,
		int          	rows,
		int          	cols,
		float        	z,
		AmListPtr      	*edgeList,
		AmListPtr      	*inList,
		int	   	*maxLength)
{
    int         	ii, jj, colst2 = cols << 1, i0, i1, i2, i3,
			contourId;
    AnvilBoolean		closed;
    uchar_t        	*markPtr1, *markPtr2;
    AmPolygonPtr  	contour;
    uchar_t		*matPtr1, *matPtr2;

    *edgeList = NULL;
    *inList = NULL;
    *maxLength = 0;

/* test vertical edges first */

    markPtr1 = markGrid;
    markPtr2 = markGrid + colst2 - 1;
    contourId = 1;
    for ( ii = 0, markPtr1 = markGrid, markPtr2 = markGrid + colst2 - 1; 
	  ii < rows - 1; 
	  ii++, markPtr1 += colst2, markPtr2 += colst2) {

/* left side: certain it's even */

        if (( !*markPtr1) && 
	    ( testTri( createTri( matrix, zFn, clientData, 
				  size, cols, ii, 0), z) & EDGE1)) {
            contour = genContour( EDGE2, matrix, zFn, clientData, size, 
				  rows, cols, ii, 0, markGrid, z, &closed,
				  contourId);

/* if it didn't intersect at least two triangles, then it's not a contour... */

	    if ( contour->numPts > 2) {
		*edgeList = AmInsertNodeLst( contour, *edgeList);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}
	    } else {
		free( contour->pts);
		free( contour);
	    }
	    contourId++;
	    if ( contourId == 0) {
		contourId = 1;
	    }
        }

/* right side: certain it's odd */

        if (( !*markPtr2) &&
            ( testTri( createTri( matrix, zFn, clientData, size, 
				  cols, ii, colst2 - 3), z) & EDGE1)) {
            contour = genContour( EDGE2, matrix, zFn, clientData, size, 
				  rows, cols, ii, colst2 - 3, markGrid, 
				  z, &closed, contourId);
	    if ( contour->numPts > 2) {
		*edgeList = AmInsertNodeLst( contour, *edgeList);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}
	    } else {
		free( contour->pts);
		free( contour);
	    }
	    contourId++;
	    if ( contourId == 0) {
		contourId = 1;
	    }
        }
    }

/* test horizontal edges */

    for ( jj = 1, markPtr1 = markGrid+1, 
		 markPtr2 = markGrid+colst2*( rows - 2);
	  jj < colst2 - 2; 
	  jj += 2, markPtr1 += 2, markPtr2 += 2) {

/* take odd triangles on top */

        if ((!*markPtr1) &&
            ( testTri( createTri( matrix, zFn, clientData, size, 
				  cols, 0, jj), z) & EDGE1)) {
            contour = genContour( EDGE2, matrix, zFn, clientData, size, rows, 
				  cols, 0, jj, markGrid, z, &closed,
				  contourId);
	    if ( contour->numPts > 2) {
		*edgeList = AmInsertNodeLst( contour, *edgeList);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}
	    } else {
		free( contour->pts);
		free( contour);
	    }
	    contourId++;
	    if ( contourId == 0) {
		contourId = 1;
	    }
        }

/* bottom takes even triangles */

        if ((!*markPtr2) &&
            ( testTri( createTri( matrix, zFn, clientData, size, 
				  cols, rows - 2, jj - 1), z) & EDGE2)) {
            contour = genContour( EDGE1, matrix, zFn, clientData, size, 
				  rows, cols, rows - 2, jj - 1, markGrid, 
				  z, &closed, contourId);

/* if it's too small, don't throw it away, because it's stored in PT_LST */

	    if ( contour->numPts > 2) {
		*edgeList = AmInsertNodeLst( contour, *edgeList);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}
	    } else {
		free( contour->pts);
		free( contour);
	    }
	    contourId++;
	    if ( contourId == 0) {
		contourId = 1;
	    }
        }
    }

/* search for interior contour lines */

    for ( ii = 1, 
	  markPtr1 = markGrid + colst2 + 2, 
	  matPtr1 = matrix + (cols + 1)*size;
	  ii < rows - 2; 
	  ii++, markPtr1 += colst2, matPtr1 += cols*size) {

/* take the first two points and determine their orientation to the z plane */

	i0 = (( zFn)( matPtr1, clientData) > z);
	i1 = (( zFn)(matPtr1 + cols*size, clientData) > z);
        for ( jj = 2, markPtr2 = markPtr1, matPtr2 = matPtr1 + size; 
	      jj < colst2 - 2; 
	      jj += 2, markPtr2++, matPtr2 += size) {

/* take the next two points and find their orientation */

	    i2 = (( zFn)( matPtr2, clientData) > z);
	    i3 = (( zFn)( matPtr2 + cols*size, clientData) > z);

/* if lower triangle crosses, make a contour starting with it. Note that
   the mark grid pointer is incremented here as well as the outer loop */

            if ((!*markPtr2++) && (( i0 != i1) || ( i0 != i2))) {
		contour = genContour( 0, matrix, zFn, clientData, size, 
				      rows, cols, ii, jj, 
				      markGrid, z, &closed, contourId);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}

/* depending on whether the contor actually closed, add it to the interior
   or edge lists */
		
		if ( contour->numPts > 2) {
		    if ( closed) {
			*inList = AmInsertNodeLst( contour, *inList);
		    } else {
			*edgeList = AmInsertNodeLst( contour, *edgeList);
		    }
		} else {
		    free( contour->pts);
		    free( contour);
		}
		contourId++;
		if ( contourId == 0) {
		    contourId = 1;
		}
            }

/* if the upper triangle crosses, make a contour starting with it */

            if ((!*markPtr2) && ( i2 != i3)) {
		contour = genContour( 0, matrix, zFn, clientData, size, 
				      rows, cols, ii, jj + 1, markGrid, 
				      z, &closed, contourId);
		if ( contour->numPts > *maxLength) {
		    *maxLength = contour->numPts;
		}

/* depending on whether the contor actually closed, add it to the interior
   or edge lists */

		if ( contour->numPts > 2) {
		    if ( closed) {
			*inList = AmInsertNodeLst( contour, *inList);
		    } else {
			*edgeList = AmInsertNodeLst( contour, *edgeList);
		    }
		} else {
		    free( contour->pts);
		    free( contour);
		}
		contourId++;
		if ( contourId == 0) {
		    contourId = 1;
		}
            }
	    i0 = i2;
	    i1 = i3;
        }
    }
    return( TRUE);
}

/****************************************************************************
 * 
 * Description:
 *	Given a matrix, accessing function, dimensions, a min and max
 *	value, step value, and world-coordinate for the matrix,
 *	return a list of lists of edge contours and a list of lists
 *	of internal contours, scaled to the matrices' coordinate system.
 *	also return the length of the largest contour found.
 * Arguments:
 *	matrix - matrix to contour.
 *	zFn, clientData - accessing function and client data to convert
 *	the source matrix into float values.
 *	size - size of each element in the source matrix.
 *	rows, cols - dimensions of the source matrix.
 *	zmin, zmax - min and max contour values.
 *	step - step AmBetween contour planes
 *	edgeListList - returned list of lists of edge contour polygons.
 *	inListList - returned list of lists of internal contour polygons.
 *	maxLength - maximum number of points in a returned contour polygon.
 *	xmin, ymin, xmax, ymax - real world coordinate for the matrix
 *	being contoured.
 * Returns:
 *	TRUE.
 * Modifications:
 *	should return an error code for malloc failure.
 * 
 ****************************************************************************/

AmErrorCode 
AmMakeContourList( uchar_t      *matrix,	    
		   AmAccessFn   zFn,
		   void	     	*clientData,
		   int	     	size,
		   int	     	rows,	   
		   int	     	cols,
		   float	zmin,	  
		   float	zmax,
		   float	step,	  
		   AmListPtr    *edgeListList, 
		   AmListPtr    *inListList,  
		   int	     	*maxLength,
		   float	xmin,
		   float	ymin,
		   float	xmax,
		   float	ymax)
{
    AmListPtr	edgeList, inList;
    int		length, markSize = rows*cols*2;
    uchar_t     *markGrid;

    if (( markGrid = ( uchar_t *) malloc( markSize)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    *edgeListList = NULL;
    *inListList = NULL;
    *maxLength = 0;

/* we step backwards, because the lists are built in stacks, and people 
   like the low z  values first */

    for ( ; zmax >= zmin; zmax -= step) {
AN_DEBUG( 99, ( "contouring %.2f", zmax));
	memset( markGrid, '\0', markSize);
	AmMakeContours( matrix, zFn, clientData, size, markGrid, rows, cols, 
		      zmax, &edgeList, &inList, &length);
	AmScaleContourList( edgeList, rows, cols, xmin, ymin, xmax, ymax);
	AmScaleContourList( inList, rows, cols, xmin, ymin, xmax, ymax);
	*edgeListList = AmInsertNodeLst( edgeList, *edgeListList);
	*inListList = AmInsertNodeLst( inList, *inListList);
	if ( length > *maxLength) {
	    *maxLength = length;
	}

/* only loop once if the step is 0.0 */

	if ( step == 0.0) {
	    break;
	}
    }

/* free the markList and point list store, set it in case we get 
   called again */ 

    FREE( markGrid);
    AmFreeLst( PT_LST, ( AmItemFn) free);
    PT_LST = NULL;
    return( AmOK);
}


/****************************************************************************
 * 
 * Description:
 *	Given a matrix, accessing function, dimensions, a vector of
 *	contour values, step value, and world-coordinate for the matrix,
 *	return a list of lists of edge contours and a list of lists
 *	of internal contours, scaled to the matrices' coordinate system.
 *	also return the length of the largest contour found.
 * Arguments:
 *	matrix - matrix to contour.
 *	zFn, clientData - accessing function and client data to convert
 *	the source matrix into float values.
 *	size - size of each element in the source matrix.
 *	rows, cols - dimensions of the source matrix.
 *	binvec - list of values to contour against.
 *	step - step AmBetween contour planes
 *	edgeListList - returned list of lists of edge contour polygons.
 *	inListList - returned list of lists of internal contour polygons.
 *	maxLength - maximum number of points in a returned contour polygon.
 *	xmin, ymin, xmax, ymax - real world coordinate for the matrix
 *	being contoured.
 * Returns:
 *	TRUE.
 * Modifications:
 *	should return an error code for malloc failure.
 * 
 ****************************************************************************/

AmErrorCode 
AmBinnedContourList( uchar_t     *matrix,
		     AmAccessFn  zFn,
		     void	 *clientData,
		     int	 size,
		     int	 rows,
		     int	 cols,
		     float       *binvec,
		     int	 numbins,
		     AmListPtr   *edgeListList, 
		     AmListPtr   *inListList,
		     int	 *maxLength,
		     float       xmin,
		     float       ymin,
		     float       xmax,
		     float       ymax)
{
    AmListPtr	edgeList, inList;
    int		length, ii, markSize = rows*cols*2;
    float	level;
    uchar_t     *markGrid;

    if (( markGrid = ( uchar_t *) malloc( markSize)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    *edgeListList = NULL;
    *inListList = NULL;
    *maxLength = 0;

/* we step backwards, because the lists are built in stacks, and people 
   like the low z  values first */

    for ( ii = numbins - 1; ii >= 0; ii--) {
	level = binvec[ ii];
AN_DEBUG( 99, ( "contouring %.2f", level));
	memset( markGrid, '\0', markSize);
	AmMakeContours( matrix, zFn, clientData, size, markGrid, rows, cols, 
		      level, &edgeList, &inList, &length);
	AmScaleContourList( edgeList, rows, cols, xmin, ymin, xmax, ymax);
	AmScaleContourList( inList, rows, cols, xmin, ymin, xmax, ymax);
	*edgeListList = AmInsertNodeLst( edgeList, *edgeListList);
	*inListList = AmInsertNodeLst( inList, *inListList);
	if ( length > *maxLength) {
	    *maxLength = length;
	}
    }

/* free the markGrid and point list store, and re-set it in case we get 
   called again */ 

    FREE( markGrid);
    AmFreeLst( PT_LST, ( AmItemFn) free);
    PT_LST = NULL;
    return( AmOK);
}


/**************************************************************************
 * 
 * Description:
 * 	close a polygon which has intersected the extents of the drawing 
 * Arguments:
 *	poly - polygon to close
 *	xmin, ymin, xmax, ymax - extents of the drawing.
 *	numpts - input number of points, returns new number of points
 *	if new lines are generated for when contour intersects the 
 *	drawing edge.
 * Returns:
 *	static vector of coordinates representing the points in the polygon.
 * Modifications:
 *	There have been cases, I think opposite edge intersections, where
 *	it generated incorrect edges.
 *
 **************************************************************************/

AmCoordPtr 
AmClosePoly( AmPolygonPtr	poly,
	     float		xmin,
	     float		ymin,
	     float		xmax,
	     float		ymax,
	     int		*numpts)
{
    AmCoordPtr 		a = &poly->pts[ 0], b = &poly->pts[ poly->numPts - 1];
    int			aCode, bCode;
    float		aDist, bDist;
    static AmCoordRec	points[ 65536];

    memcpy( points, poly->pts, poly->numPts*sizeof( AmCoordRec));
    *numpts = poly->numPts;
    aDist = fabs( a->x - xmin);
    aCode = 1;
    if ( fabs( a->x - xmax) < aDist) {
	aDist = fabs( a->x - xmax);
	aCode = 2;
    }
    if ( fabs( a->y - ymin) < aDist) {
	aDist = fabs( a->y - ymin);
	aCode = 3;
    }
    if ( fabs( a->y - ymax) < aDist) {
	aDist = fabs( a->y - ymax);
	aCode = 4;
    }
    bDist = fabs( b->x - xmin);
    bCode = 1;
    if ( fabs( b->x - xmax) < bDist) {
	bDist = fabs( b->x - xmax);
	bCode = 2;
    }
    if ( fabs( b->y - ymin) < bDist) {
	bDist = fabs( b->y - ymin);
	bCode = 3;
    }
    if ( fabs( b->y - ymax) < bDist) {
	bDist = fabs( b->y - ymax);
	bCode = 4;
    }

/* if the codes are the same, then we don't have to close the polygon */

/* try the corner codes */

    if ((( aCode == 1) && ( bCode == 3)) ||
        (( bCode == 1) && ( aCode == 3))) {
	points[ *numpts].x = xmin;
	points[ *numpts].y = ymin;
	(*numpts)++;
    } else if ((( aCode == 2) && ( bCode == 3)) ||
	       (( bCode == 2) && ( aCode == 3))) {
	points[ *numpts].x = xmax;
	points[ *numpts].y = ymin;
	(*numpts)++;
    } else if ((( aCode == 2) && ( bCode == 4)) ||
	       (( bCode == 2) && ( aCode == 4))) {
	points[ *numpts].x = xmax;
	points[ *numpts].y = ymax;
	(*numpts)++;
    } else if ((( aCode == 4) && ( bCode == 1)) ||
	       (( bCode == 1) && ( aCode == 4))) {
	points[ *numpts].x = xmin;
	points[ *numpts].y = ymax;
	(*numpts)++;
    }

/* now try the opposing codes..for each code set, find out which side
   the line is closer to */

    if (( aCode == 3) && ( bCode == 4)) {
	if (( points[ 0].x - xmin) + (points[ *numpts - 1].x - xmin) <
	    ( xmax - points[ 0].x) + ( xmax - points[ *numpts - 1].x)) {
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymax;
	    (*numpts)++;				/* left side */
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymin;
	    (*numpts)++;
	} else {
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymax;
	    (*numpts)++;				/* right side */
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymin;
	    (*numpts)++;
	}
    } else if (( aCode == 4) && ( bCode == 3)) {
	if (( points[ 0].x - xmin) + (points[ *numpts - 1].x - xmin) <
	    ( xmax - points[ 0].x) + ( xmax - points[ *numpts - 1].x)) {
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymin;
	    (*numpts)++;				/* left side */
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymax;
	    (*numpts)++;
	} else {
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymin;
	    (*numpts)++;				/* right side */
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymax;
	    (*numpts)++;
	}
    } else if (( aCode == 1) && ( bCode == 3)) {
	if (( points[ 0].y - ymin) + (points[ *numpts - 1].y - ymin) <
	    ( ymax - points[ 0].y) + ( ymax - points[ *numpts - 1].y)) {
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymin;
	    (*numpts)++;				/* top side */
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymin;
	    (*numpts)++;
	} else {
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymax;
	    (*numpts)++;				/* bottom side */
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymax;
	    (*numpts)++;
	}
    } else if (( aCode == 3) && ( bCode == 1)) {
	if (( points[ 0].y - ymin) + (points[ *numpts - 1].y - ymin) <
	    ( ymax - points[ 0].y) + ( ymax - points[ *numpts - 1].y)) {
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymin;
	    (*numpts)++;				/* top side */
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymin;
	    (*numpts)++;
	} else {
	    points[ *numpts].x = xmin;
	    points[ *numpts].y = ymax;
	    (*numpts)++;				/* bottom side */
	    points[ *numpts].x = xmax;
	    points[ *numpts].y = ymax;
	    (*numpts)++;
	}
    }
    return( points);
}

/**************************************************************************
 *
 * Description:
 *      contour a matrix and write the polygons into a file to be read
 *	as a polygon plot 
 * Arguments:
 *	matrix - matrix to contour
 *	rows, cols - dimensions of the matrix
 *	size - size of each matrix element.
 *	zFn, clientData - matrix accessing function.
 *      xmin, ymin, xmax, ymax - extents of the drawing.
 *      headerPath - header output file.
 *      dataPath - data file.
 * Returns:
 *	error on open or malloc failure, AmOK otherwise.
 *
 **************************************************************************/
   

AmErrorCode
AmContourMatrixFile( uchar_t	*matrix,
		     int	rows,
		     int	cols,
		     int	size,
		     AmAccessFn	zFn,
		     void	*clientData,
		     float	xmin,
		     float	ymin,
		     float	xmax,
		     float	ymax,
		     AnvilBoolean	flipRows,
		     AnvilBoolean	flipCols,
		     float	min,
		     float	max,
		     float	step,
		     char	*headerPath,
		     char	*dataPath)
{
    int			fdOut, maxLength, polyRows, numPts;
    FILE		*hdrFp;
    AmListPtr		edgeList, inList, ptr;
    AmPolygonPtr	poly;
    AmCoordPtr		points;
    uchar_t		*markGrid;
    float		z;

/* open the input header, and output files */
 
    if (( fdOut = open( dataPath, O_WRONLY | O_CREAT, PERMS)) < 0) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    if (( hdrFp = fopen( headerPath, "w")) == NULL) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* allocate the flag grid for the contour */
    
    markGrid = ( uchar_t *) malloc( rows*cols*2);
    if ( markGrid == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* contour each level and write it out, then free the contour.  */

    polyRows = 0;
    for ( z = max; z >= min; z -= step) {
AN_DEBUG( 60, ("contouring %.2f\n", z));
	memset( markGrid, '\0', rows*cols*2);
	AmMakeContours( matrix, zFn, clientData, size, markGrid, rows, cols, 
		        z, &edgeList, &inList, &maxLength);
	AmScaleContourList( edgeList, rows, cols, xmin, ymin, xmax, ymax, 
			    flipRows, flipCols);
	AmScaleContourList( inList, rows, cols, xmin, ymin, xmax, ymax, 
			    flipRows, flipCols);
	polyRows += AmLengthLst( edgeList);
	polyRows += AmLengthLst( inList);

/* write out the polygons with the associated z values */

	for ( ptr = inList; ptr != NULL; ptr = ptr->next) {
	    poly = ( AmPolygonPtr) ptr->data;
	    write( fdOut, poly, sizeof( AmPolygonRec));
	    write( fdOut, &z, sizeof( z));
	    write( fdOut, poly->pts, poly->numPts*sizeof( AmCoordRec));
	    free( poly->pts);
	    free( poly);
	}
	AmJustFreeLst( inList);
	for ( ptr = edgeList; ptr != NULL; ptr = ptr->next) {
	    poly = ( AmPolygonPtr) ptr->data;

/* close the polygons */

	    points = AmClosePoly( poly, xmin, ymin, xmax, ymax, &numPts);
	    poly->numPts = numPts;
	    write( fdOut, poly, sizeof( AmPolygonRec));
	    write( fdOut, &z, sizeof( z));
	    write( fdOut, points, poly->numPts*sizeof( AmCoordRec));
	    free( poly->pts);
	    free( poly);
	}
	AmJustFreeLst( edgeList);
    }
    free( markGrid);

/* write out the header */

    fprintf( hdrFp, "filename:	%s\n", dataPath);
    fprintf( hdrFp, "rows:		%d\n", polyRows);
    fprintf( hdrFp, "size:		%d\n", sizeof( AmPolygonRec) + 
					   sizeof( float));
    fprintf( hdrFp, "type:		%s\n", AmNpolylinePlot);
    fprintf( hdrFp, "xmin:          %f\n", xmin);
    fprintf( hdrFp, "ymin:          %f\n", ymin);
    fprintf( hdrFp, "xmax:          %f\n", xmax);
    fprintf( hdrFp, "ymax:          %f\n", ymax);
    fprintf( hdrFp, "colorsMin:		%f\n", min);
    fprintf( hdrFp, "colorsMax:		%f\n", max);
    fprintf( hdrFp, "note:		contoured\n");
#if BYTE_ORDER == LITTLE_ENDIAN
    fprintf( hdrFp, "byteOrder: littleEndian\n");
#else
    fprintf( hdrFp, "byteOrder: bigEndian\n");
#endif
    fprintf( hdrFp, "description:       %s\n", descriptionFile);
    fprintf( hdrFp, "fileSource:        contour\n");
    fprintf( hdrFp, "appId:        	contour\n");
    fclose( hdrFp);
    close( fdOut);
    return( AmOK);
}
