#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* utility functions for opening files using pipes and compressed input.
   this will probably have to be modified to support Windows properly.
   these functions are duplicates of readfile.cpp */ 

#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <anvil/project.h>
#include <anvil/strlib.h>

/****************************************************************************
 *
 * Description:
 * 	open a file for reading, and if it's got a compressed extension, 
 *	then open a pipe to the uncompress command, and set the pipeFlag.
 * Arguments:
 *	filename - filename to open, or create a pipe to.
 *	pipeFlag - TRUE if a pipe was opened, FALSE if it's a straight file.
 * Returns:
 *	file descriptor to the file or the pipe.
 *
 ****************************************************************************/

int 
AmOpenReadFile( const char 	*filename, 
	      AnvilBoolean		*pipeFlag)
{
    FILE	*fp;
    char	cmd[ PATHLEN];

/* try uncompress */

    if ( AmEndswith( filename, ".Z")) {
	sprintf( cmd, "uncompress -c %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fileno( fp));
	}

/* try gunzip */

    } else if ( AmEndswith( filename, ".gz")) {
	sprintf( cmd, "gunzip -c %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fileno( fp));
	}

/* try unzip */

    } else if ( AmEndswith( filename, ".z")) {
	sprintf( cmd, "unzip -p %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fileno( fp));
	}
    } else {
	*pipeFlag = FALSE;
	return( open( filename, O_RDONLY, PERMS));
    }
}

/****************************************************************************
 *
 * Description:
 * 	close a file or pipe depending on the pipe flag 
 * Arguments:
 *	fd - file descriptor to close.
 *	pipeFlag - pipe flag returned from AmOpenReadFile()
 * Returns:
 *	the results of close() or pclose()
 *
 ****************************************************************************/

int 
AmCloseReadFile( int	fd,
	       AnvilBoolean	pipeFlag)
{
    FILE *fp;

    if ( pipeFlag) {
	fp = fdopen( fd, "r");
	return( pclose( fp));
    } else {
	return( close( fd));
    }
}

/****************************************************************************
 *
 * Description:
 *      keep reading bytes until done, error or buffer empty.  This is
 *      used for reading pipes and streams.
 * Arguments:
 *      fd - file descriptor of the open file.
 *      buf - buffer to read into.
 *      nbyte - number of bytes expected.
 * Return:
 *      number of total bytes read, or -1 on error
 *
 ****************************************************************************/

int 
AmReadLoopFile( int	fd,
	      void	*buf,
	      int	nbytes)
{
    int 	bytesRead, totalBytes;
    char 	*buffer = ( char *) buf;

    totalBytes = 0;
    do {

/* read as many bytes as we can get, and step down the buffer */

        bytesRead = read( fd, buffer, nbytes);
        if ( bytesRead < 0) {
            return( -1);
        }
        nbytes -= bytesRead;
        totalBytes += bytesRead;
        buffer += bytesRead;
    } while (( nbytes > 0) && ( bytesRead > 0));

/* return the number of bytes read */

    return( totalBytes);
}

/****************************************************************************
 *
 * Description:
 * 	open a file for reading, and if it's got a compressed extension, 
 *	then open a pipe to the uncompress command, and set the pipeFlag.
 * Arguments:
 *	filename - filename to open, or create a pipe to.
 *	pipeFlag - TRUE if a pipe was opened, FALSE if it's a straight file.
 * Returns:
 *	file pointer to the file or the pipe.
 *
 ****************************************************************************/

FILE *
AmOpenReadFilePointer( const char *filename, 
		     AnvilBoolean	*pipeFlag)
{
    FILE	*fp;
    char	cmd[ PATHLEN];

    if ( AmEndswith( filename, ".Z")) {
	sprintf( cmd, "uncompress -c %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fp);
	}
    } else if ( AmEndswith( filename, ".gz")) {
	sprintf( cmd, "gunzip -c %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fp);
	}
    } else if ( AmEndswith( filename, ".z")) {
	sprintf( cmd, "unzip -p %s", filename);
	if (( fp = popen( cmd, "r")) == NULL) {
	    return( FALSE);
	} else {
	    *pipeFlag = TRUE;
	    return( fp);
	}
    } else {
	*pipeFlag = FALSE;
	return( fopen( filename, "r"));
    }
}

/****************************************************************************
 *
 * Description:
 * 	close a file or pipe depending on the pipe flag 
 * Arguments:
 *	fd - file descriptor to close.
 *	pipeFlag - pipe flag returned from AmOpenReadFilePointer()
 * Returns:
 *	the results of fclose() or pclose()
 *
 ****************************************************************************/

int 
AmCloseReadFilePointer( FILE	*fp,
		      AnvilBoolean	pipeFlag)
{
    if ( pipeFlag) {
	return( pclose( fp));
    } else {
	return( fclose( fp));
    }
}

/****************************************************************************
 *
 * Description:
 * 	open a file for reading, and if it's got a compressed extension, 
 *	then open a pipe to the uncompress command, and set the pipeFlag.
 * Arguments:
 *	filename - filename to open, or create a pipe to.
 *	pipeFlag - TRUE if a pipe was opened, FALSE if it's a straight file.
 * Returns:
 *	file pointer to the file or the pipe.
 *
 ****************************************************************************/

int 
AmReadLoopFilePointer( FILE	*fp,
		     void	*buf,
		     int	nbytes)
{
    int 	bytesRead, totalBytes;
    char 	*buffer = ( char *) buf;

    totalBytes = 0;
    do {

/* read as many bytes as we can get, and step down the buffer */

        bytesRead = fread( buffer, 1, nbytes, fp);
        if ( bytesRead < 0) {
            return( -1);
        }
        nbytes -= bytesRead;
        totalBytes += bytesRead;
        buffer += bytesRead;
    } while (( bytesRead > 0) && ( nbytes > 0));

/* return the number of bytes read */

    return( totalBytes);
}
