#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdio.h>
#include <math.h>
#if defined(_WIN32)
#undef y0
#undef y1
#endif
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/errcode.h>
#include <anvil/geometry.h>
#include <anvil/datafmt.h>
#include <anvil/macros.h>
#include <anvil/quadnet.h>

/* Functions for reducing sets of disconnected lines and polygons by 
   regenerating the connectivity data, and reducing 2-connected vectors,
   and shared polygon edges.  There are bugs in the polygon edge
   reduction algorithm that have to be addressed dealing with shared
   edges AmBetween polygons. */

/* because the volume of data being processed by the quadtree algorithm
   is so large, and the depth of the quadtree can be substantial, 
   the nodes and arcs are stored in static variables, which are 
   realloced as needed */

static NodePtr		saveNodes = NULL;
static LineArcPtr	saveLineArcs = NULL;
static int		maxLines = 0;
static LineArcPtr	*saveNodeArcs = NULL;
static int		maxNodeArcs = 0;
static NodePtr		*saveQuadNodes = NULL;
static int		maxQuadNodes;

/* free the static node and arc data once the processing is completed */

void 
AmFreeQuadMultiCallData()
{
    if ( saveNodes != NULL) {
    	free( saveNodes);
    }
    if ( saveLineArcs != NULL) {
    	free( saveLineArcs);
    }
    if ( saveNodeArcs != NULL) {
    	free( saveNodeArcs);
    }
    if ( saveQuadNodes != NULL) {
    	free( saveQuadNodes);
    }
}

/****************************************************************************
 *
 * Description:
 * 	convert a set of polygons to a networked representation, where
 *	vertices and shared edges are converted into connected arcs 
 *	and nodes.
 * Arguments:
 *	polys - vector of polygons to turn into line arcs.
 *	numPolys - number of polygons.
 *	polysz - size of the polygons' header structures.
 *	lineArcsPtr - returned vector of arcs.
 *	numArcs - returned length of the arcs vector.
 *	nodesPtr - resulting vector of nodes.
 *	multiCallFlag - used to control whether information is allocated
 *	locally, or if the statics are used when this function is called
 *	repeatedly.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on memory allocation failure.
 *	 
 ****************************************************************************/

static AmErrorCode 
PolygonToLineArc( AmPolygonPtr	polys,
		  int		numPolys,
		  int		polysz,
		  LineArcPtr	*lineArcsPtr,
		  int		*numArcs,
		  NodePtr	*nodesPtr,
		  AnvilBoolean	multiCallFlag)
{
    int		ii, jj, count, start;
    LineArcPtr	lineArcs;
    NodePtr	nodes;
    char	*ptr;
    AmPolygonPtr	polyPtr;

/* how many points total are in all the polygons */

    *numArcs = 0;
    for ( ii = 0, ptr = ( char *) polys;
    	  ii < numPolys;
    	  ii++, ptr += polysz + (( AmPolygonPtr) ptr)->numPts*sizeof( AmCoordRec)) {
	*numArcs += (( AmPolygonPtr) ptr)->numPts;
    }

    if ( multiCallFlag) {

/* free the old buffers and allocate the new ones if we need more space, if
   we're getting called over and over */

    	if ( *numArcs > maxLines) {
    	    free( saveNodes);
    	    free( saveLineArcs);
	    maxLines = *numArcs;
	    if (( saveLineArcs = ( LineArcPtr) 
		     calloc( *numArcs, sizeof( LineArcRec))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	    if (( saveNodes = ( NodePtr) 
		     calloc( *numArcs, sizeof( NodeRec))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	} 
	*lineArcsPtr = lineArcs = saveLineArcs;
	*nodesPtr = nodes = saveNodes;
    } else {

/* otherwise, just allocate every time */

	if (( *lineArcsPtr = lineArcs = ( LineArcPtr) 
		 calloc( *numArcs, sizeof( LineArcRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if (( *nodesPtr = nodes = ( NodePtr) 
		 calloc( *numArcs, sizeof( NodeRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }

/* copy over the segments from the polygons to the network */

    count = 0;
    for ( ii = 0, ptr = ( char *) polys;
    	  ii < numPolys;
    	  ii++, ptr += polysz + (( AmPolygonPtr) ptr)->numPts*sizeof( AmCoordRec)) {
	polyPtr = ( AmPolygonPtr) ptr;
AN_DEBUG( 40, ("polygon %d, %d points\n", ii, polyPtr->numPts));
	start = count;
	for ( jj = 0; jj < polyPtr->numPts - 1; jj++) {
	    nodes[ count].x = polyPtr->pts[ jj].x;
	    nodes[ count].y = polyPtr->pts[ jj].y;
	    lineArcs[ count].leftNode = &nodes[ count];
	    lineArcs[ count].rightNode = &nodes[ count + 1];
	    if ( jj > 0) {
		nodes[ count].lineArcLeft = &lineArcs[ count - 1];
	    } else {
		nodes[ count].lineArcLeft = NULL;
	    }
	    nodes[ count].lineArcRight = &lineArcs[ count];
	    count++;
	}

/* copy the closing segment */

	nodes[ count].x = polyPtr->pts[ polyPtr->numPts - 1].x;
	nodes[ count].y = polyPtr->pts[ polyPtr->numPts - 1].y;
	nodes[ count].lineArcLeft = &lineArcs[ count - 1];
	nodes[ count].lineArcRight = &lineArcs[ count];
	lineArcs[ count].leftNode = &nodes[ count]; 
	lineArcs[ count].rightNode = &nodes[ start];

/* AmFirst node's left line points to the AmLast line arc */

	nodes[ start].lineArcLeft = &lineArcs[ count];
	count++;
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Take a set of disconnected segments, and by proximity testing
 *	turn them into a connected vector of arcs and nodes.
 * Arguments:
 *	lines - vector of segments to turn into line arcs.
 *	AmNumLines - number of segments.
 *	linesz - size of the line structures.
 *	lineArcsPtr - returned vector of arcs.
 *	numArcs - returned length of the arcs vector.
 *	nodesPtr - resulting vector of nodes.
 *	multiCallFlag - used to control whether information is allocated
 *	locally, or if the statics are used when this function is called
 *	repeatedly.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on memory allocation failure.
 *	 
 ****************************************************************************/

static AmErrorCode 
LineToLineArc( AmLinePtr		lines,
	       int		AmNumLines,
	       int		linesz,
	       LineArcPtr	*lineArcsPtr,
	       NodePtr		*nodesPtr,
	       AnvilBoolean		multiCallFlag)
{
    char		*lineBufPtr;
    AmLinePtr		linePtr;
    LineArcPtr		lineArcs;
    NodePtr		nodes;
    int			ii;

    if ( multiCallFlag) {

/* free the old buffers and allocate the new ones if we need more space, if
   we're getting called over and over */

    	if ( AmNumLines > maxLines) {
    	    free( saveNodes);
    	    free( saveLineArcs);
	    maxLines = AmNumLines;
	    if (( saveLineArcs = ( LineArcPtr) 
		     calloc( AmNumLines, sizeof( LineArcRec))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	    if (( saveNodes = ( NodePtr) 
		     calloc( AmNumLines*2, sizeof( NodeRec))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	} 
	*lineArcsPtr = lineArcs = saveLineArcs;
	*nodesPtr = nodes = saveNodes;
    } else {

/* otherwise, just allocate every time */

	if (( *lineArcsPtr = lineArcs = ( LineArcPtr) 
		 calloc( AmNumLines, sizeof( LineArcRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if (( *nodesPtr = nodes = ( NodePtr) 
		 calloc( AmNumLines*2, sizeof( NodeRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }

/* copy over the pointers */

    for ( ii = 0, lineBufPtr = ( char *) lines;
    	  ii < AmNumLines;
    	  ii++, lineBufPtr += linesz) {
        linePtr = ( AmLinePtr) lineBufPtr;
        nodes[ ii*2].x = linePtr->x0;
        nodes[ ii*2].y = linePtr->y0;
        nodes[ ii*2].lineArcLeft = &lineArcs[ ii];
        nodes[ ii*2].lineArcRight = NULL;
        nodes[ ii*2].arcs = NULL;
        nodes[ ii*2].narcs = 0;
        nodes[ ii*2 + 1].x = linePtr->x1;
        nodes[ ii*2 + 1].y = linePtr->y1;
        nodes[ ii*2 + 1].lineArcLeft = &lineArcs[ ii];
        nodes[ ii*2 + 1].lineArcRight = NULL;
        nodes[ ii*2 + 1].arcs = NULL;
        nodes[ ii*2 + 1].narcs = 0;
        lineArcs[ ii].leftNode = &nodes[ ii*2];
        lineArcs[ ii].rightNode = &nodes[ ii*2 + 1];
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	given a node and the extents we're working with, return
 *	a code showing which subquad the node belongs to.
 * Arguments:
 *	node - node containing x,y point to test.
 *	xmin, ymin, xmax, ymax - extents to test against.
 * Returns:
 *	0 - x,y in lower left,
 *	1 - x,y in lower right.
 *	2 - x,y in upper left
 *	3 - x,y in upper right.
 * Modifications:
 *	that assert is probably a very, very bad idea...
 *
 ****************************************************************************/

static int 
getCode( NodePtr	node,
	 float		xmin,
	 float		ymin,
	 float		xmax,
	 float		ymax)
{
    if (( node->x < xmin) || ( node->y < ymin) ||
        ( node->x > xmax) || ( node->y > ymax)) {
        ASSERT( FALSE);
    }
    if ( node->y < ( ymax + ymin)/2) {
	if ( node->x < ( xmax + xmin)/2) {
    	    return( 0);
	} else {
    	    return( 1);
	}
    } else {
	if ( node->x < ( xmax + xmin)/2) {
    	    return( 2);
	} else {
    	    return( 3);
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	Given a set of nodes and arcs which are within a leaf quad,
 *	set the arc left or right node pointers to the AmFirst node
 *	in the set, and set the arc pointers in the AmFirst node to
 *	point to the arcs which point to it.
 * Arguments:
 *	nodes - set of nodes which are in the leaf quad.
 *	numNodes - number of nodes in the set.
 *	lineArcs - lines which point to the set of nodes.
 *	numArcs - number of arcs in the set.
 *
 ****************************************************************************/

static void 
ReduceNodes( NodePtr	*nodes,
	     int	numNodes,
	     LineArcPtr	*lineArcs,
	     int	*numArcs)
{
    int		ii, code, count;

/* we are a leaf node.  Make all of the vectors that point to nodes in this
   quad point to our AmFirst node, and accumulate the arcs into the arclist
   for that node */

    nodes[ 0]->arcs = &lineArcs[ *numArcs];
    count = 0;
    for ( ii = 0; ii < numNodes; ii++) {
	if ( nodes[ ii]->lineArcLeft != NULL) {
	    code = 0;
	    if ( nodes[ ii] == nodes[ ii]->lineArcLeft->leftNode) {
		code += 1;
	    } 
	    if ( nodes[ ii] == nodes[ ii]->lineArcLeft->rightNode) {
		code += 2;
	    }

/* assign the arc's left or right node, depending on the code. If
   both map to the node, then mark them as NULL. */

	    if ( code == 1) {	
		nodes[ ii]->lineArcLeft->leftNode = nodes[ 0];
		nodes[ 0]->arcs[ count++] = nodes[ ii]->lineArcLeft;
	    } else if ( code == 2) {
		nodes[ ii]->lineArcLeft->rightNode = nodes[ 0];
		nodes[ 0]->arcs[ count++] = nodes[ ii]->lineArcLeft;
	    } else { 	/* code = 0 (never) or code = 3 */
AN_DEBUG( 1, ("at node %d, left line reduced out\n", ii));
		nodes[ ii]->lineArcLeft->leftNode = NULL;
		nodes[ ii]->lineArcLeft->rightNode = NULL;
	    }
	}
	if ( nodes[ ii]->lineArcRight != NULL) {
	    code = 0;
	    if ( nodes[ ii] == nodes[ ii]->lineArcRight->leftNode) {
		code += 1;
	    } 
	    if ( nodes[ ii] == nodes[ ii]->lineArcRight->rightNode) {
		code += 2;
	    }

/* assign the arc's left or right node, depending on the code. If
   both map to the node, then mark them as NULL. */

	    if ( code == 1) {	
		nodes[ ii]->lineArcRight->leftNode = nodes[ 0];
		nodes[ 0]->arcs[ count++] = nodes[ ii]->lineArcRight;
	    } else if ( code == 2) {
		nodes[ ii]->lineArcRight->rightNode = nodes[ 0];
		nodes[ 0]->arcs[ count++] = nodes[ ii]->lineArcRight;
	    } else { 	/* code = 0 (never) or code = 3 */
AN_DEBUG( 1, ("at node %d, right line reduced out\n", ii));
		nodes[ ii]->lineArcRight->leftNode = NULL;
		nodes[ ii]->lineArcRight->rightNode = NULL;
	    }
	}
    }
    nodes[ 0]->narcs = count;
    for ( ii = 1; ii < numNodes; ii++) {
	nodes[ ii]->narcs = 0;
    }
    *numArcs += count;
}

/****************************************************************************
 *
 * Description:
 * 	are all the nodes within the quad within a tolerance of the AmFirst 
 *	node?  If so, it's a leaf quad.
 * Arguments:
 *	nodes - set of nodes which are in the leaf quad.
 *	numNodes - number of nodes in the set.
 *	lineArcs - lines which point to the set of nodes.
 *	numArcs - number of arcs in the set.
 *
 ****************************************************************************/

static AnvilBoolean 
doneQuad( NodePtr	*nodes,
	  int		numNodes,
	  float		tolerance)
{
    int		ii;

    for ( ii = 1; ii < numNodes; ii++) {
    	if ( sqrt( SQRDIST( nodes[ 0]->x, nodes[ 0]->y,
			    nodes[ ii]->x, nodes[ ii]->y)) > tolerance) {
	     return( FALSE);
	}
    }
AN_DEBUG( 99, ("done at %d nodes\n", numNodes));
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	Given a set of line arcs and the nodes that they point to,
 *	recursively spit the set by a quadtree algorithm.  If all the
 *	nodes are within a tolerance of one another, then we're done,
 *	and we consolidate the nodes and arcs that point to them into
 *	a single node. Otherwise, split the vector of nodes into a
 *	set of four vectors (in place), and track the minimum bounding
 *	rectangles of each subset of nodes.  The MBR is to reduce the
 *	level of recursion by constraining the size of the rectangle
 *	at each recursive set.
 * Arguments:
 *	nodes - vector of nodes to split
 *	numNodes - number of nodes in the vector.
 *	xmin, ymin, xmax, ymax - current extents of the set of nodes.
 *	tolerance - distance to test for quadtree leaf termination.
 *	lineArcs - line arcs that point to the nodes being reduced.
 *	numArcs - length of the line arc vector.
 *
 ****************************************************************************/

static void 
splitTree( NodePtr	*nodes,
	   int		numNodes,
	   float	xmin,
	   float	ymin,
	   float	xmax,
	   float	ymax,
	   float	tolerance,
	   LineArcPtr	*lineArcs,
	   int		*numArcs)
{
    int		counts[ 4], indexes[ 4], code, ii, currCode;
    NodePtr	*otherNode, *currNode, tmpNode;
    float	xmins[ 4], ymins[ 4], xmaxs[ 4], ymaxs[ 4];

/* if all of the nodes within the quad node are the same, then we terminate */

    if ( doneQuad( nodes, numNodes, tolerance)) {
	ReduceNodes( nodes, numNodes, lineArcs, numArcs);
	return;
    }

/* figure out the counts in each of the quads */

    for ( ii = 0; ii < 4; ii++) {
    	counts[ ii] = 0;
    }
    for ( ii = 0; ii < numNodes; ii++) {
    	counts[ getCode( nodes[ ii], xmin, ymin, xmax, ymax)]++;
    }

/* step to the AmFirst non-NULL quad */

    currCode = 0;
    while ( counts[ currCode] == 0) {
    	currCode++;
    }

/* set the indexes to reflect the counts */

    indexes[ 0] = 0;
    for ( ii = 0; ii < 3; ii++) {
    	indexes[ ii + 1] = indexes[ ii] + counts[ ii];
    }
    for ( ii = 0; ii < 4; ii++) {
    	counts[ ii] = 0;
	xmins[ ii] = HUGE;
	ymins[ ii] = HUGE;
	xmaxs[ ii] = -HUGE;
	ymaxs[ ii] = -HUGE;
    }
    ii = 0; 
    while (( ii < numNodes) && ( currCode < 4)) {

/* if the node is in another quad, swap it with the AmFirst unprocessed 
   node in that quad.  If it's in the same quad, the increment the current
   counter.  If we've finished all the nodes in this quad, then step to the
   AmFirst unprocessed node in the next quad with unprocessed nodes */

    	code = getCode( nodes[ ii], xmin, ymin, xmax, ymax);

/* squeeze out the extents of the subtile to reflect the nodes it contains */

	if ( nodes[ ii]->x < xmins[ code]) {
	    xmins[ code] = nodes[ ii]->x;
	}
	if ( nodes[ ii]->y < ymins[ code]) {
	    ymins[ code] = nodes[ ii]->y;
	}
	if ( nodes[ ii]->x > xmaxs[ code]) {
	    xmaxs[ code] = nodes[ ii]->x;
	}
	if ( nodes[ ii]->y > ymaxs[ code]) {
	    ymaxs[ code] = nodes[ ii]->y;
	}
    	otherNode = &nodes[ indexes[ code] + counts[ code]];
    	currNode = &nodes[ ii];
    	ASSERT( otherNode != NULL);
    	ASSERT( currNode != NULL);
    	ASSERT( otherNode >= currNode);
	counts[ code]++;
    	if ( code != currCode) {
    	    tmpNode = *otherNode;
    	    *otherNode = *currNode;
    	    *currNode = tmpNode;
	} else {
	    ii++;
	    while (( currCode < 4) && ( ii == indexes[ currCode + 1])) {
AN_DEBUG( 99, ("stepping to quad %d\n", ii));
		if ( indexes[ currCode] + counts[ currCode] !=
		     indexes[ currCode + 1]) {
		    ASSERT( FALSE);
		}
	    	currCode++;
	    	ii = indexes[ currCode] + counts[ currCode];
	    }
	}
	if ( code < 3) {
	    if ( indexes[ code] + counts[ code] > indexes[ code + 1]) {
	    	ASSERT( FALSE);
	    }
	}
    }
    for ( ii = 0; ii < 4; ii++) {
    	if ( counts[ ii] != 0) {
    	    AN_DEBUG( 90, ("splitting %d nodes\n", counts[ ii]));

/* set up the extents for the subtree and subdivide it */

	    splitTree( &nodes[ indexes[ ii]], counts[ ii],
		       xmins[ ii], ymins[ ii], xmaxs[ ii], ymaxs[ ii],
		       tolerance, lineArcs, numArcs);
	} 
    }
}

/****************************************************************************
 *
 * Description:
 * 	Use a quadtree algorithm to reduce a vector of line arcs and the
 *	nodes that they point to into a connected graph of data with
 *	duplicate nodes removed (according to a distance tolerance).
 * Arguments:
 *	nodes - vector of nodes to split
 *	numNodes - number of nodes in the vector.
 *	xmin, ymin, xmax, ymax - current extents of the set of nodes.
 *	tolerance - distance to test for quadtree leaf termination.
 *	lineArcs - line arcs that point to the nodes being reduced.
 *	numArcs - length of the line arc vector.
 *	multiCallFlag - flag for locally allocating data or preserving
 *	vectors of data in statics to cut down on stack space for 
 *	recursive calls.
 *
 ****************************************************************************/

static AmErrorCode 
QuadTreeReduce( NodePtr		nodes,
	        int		numNodes,
	        float		xmin,
	        float		ymin,
	        float		xmax,
	        float		ymax,
	        float		tolerance,
	        LineArcPtr	*lineArcs,
	        int		*numArcs,
	        AnvilBoolean		multiCallFlag)
{
    NodePtr	*nodePtrVec;
    int		ii;

/* if we're being called repeatedly, then only allocate the buffer if it's
   larger than the AmLast one */

    if ( multiCallFlag) {
    	if ( numNodes > maxQuadNodes) {
    	    if ( saveQuadNodes == NULL) {
    	    	free( saveQuadNodes); 
	    }
	    maxQuadNodes = numNodes;
	    if (( saveQuadNodes = ( NodePtr *) 
	    	      malloc( numNodes*sizeof( NodePtr))) == NULL) {
	        DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	}
	nodePtrVec = saveQuadNodes;
    } else {
	if (( nodePtrVec = ( NodePtr *) 
		  malloc( numNodes*sizeof( NodePtr))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }
    for ( ii = 0; ii < numNodes; ii++) {
    	nodePtrVec[ ii] = &nodes[ ii];
	nodePtrVec[ ii]->visit = FALSE;
    }
    splitTree( nodePtrVec, numNodes, xmin, ymin, xmax, ymax, tolerance,
	       lineArcs, numArcs);
    if ( !multiCallFlag) {
	free( nodePtrVec);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	given a pair of connected arcs, is their connection within 
 *	the tolerance? compare each end node against each other.
 * Arguments:
 *	a, b - line arcs to compare.
 *	tolerance - tolerance to test against.
 * Returns:
 *	TRUE if either lines' endpoint is within the tolerance of the
 *	other lines, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmToleranceArc( LineArcPtr	a,
	      LineArcPtr	b,
	      float		vectorTolerance,
	      float		angleTolerance)
{
    float	distance, angle;

    if ( a->leftNode == b->leftNode) {
	distance = AmLineDist( a->rightNode->x, a->rightNode->y,
			     b->rightNode->x, b->rightNode->y,
			     a->leftNode->x, a->leftNode->y);
	angle = AmLineAngle( a->rightNode->x, a->rightNode->y,
			   a->leftNode->x, a->leftNode->y,
			   b->rightNode->x, b->rightNode->y);
    } else if ( a->leftNode == b->rightNode) {
	distance = AmLineDist( a->rightNode->x, a->rightNode->y,
			     b->leftNode->x, b->leftNode->y,
			     a->leftNode->x, a->leftNode->y);
	angle = AmLineAngle( a->rightNode->x, a->rightNode->y,
			   a->leftNode->x, a->leftNode->y,
			   b->leftNode->x, b->leftNode->y);
    } else if ( a->rightNode == b->leftNode) {
	distance = AmLineDist( a->leftNode->x, a->leftNode->y,
			     a->rightNode->x, a->rightNode->y,
			     b->rightNode->x, b->rightNode->y);
	angle = AmLineAngle( a->leftNode->x, a->leftNode->y,
			   a->rightNode->x, a->rightNode->y,
			   b->rightNode->x, b->rightNode->y);
    } else {
	distance = AmLineDist( a->leftNode->x, a->leftNode->y,
			     b->leftNode->x, b->leftNode->y,
			     a->rightNode->x, a->rightNode->y);
	angle = AmLineAngle( a->leftNode->x, a->leftNode->y,
			   a->rightNode->x, a->rightNode->y,
			   b->leftNode->x, b->leftNode->y);
    }
    return(( distance < vectorTolerance) || ( M_PI - angle < angleTolerance));
}

/****************************************************************************
 *
 * Description:
 * 	given an arc and a connected node, return the arc pointer in the
 *	node that doesn't point to the arc.
 * Arguments:
 *	arc - arc to test.
 *	node - node to test.
 * Returns:
 *	The line arc pointer that the node points to that isn't the passed
 *	arc.
 * Modifications:
 *	This doesn't seem too reliable, and should probably check that 
 *	there are only two arc pointers at the node.
 *
 ****************************************************************************/

static LineArcPtr 
nextNodeArc( LineArcPtr		arc,
	     NodePtr		node)
{
    if ( node->arcs[ 0] == arc) {
	return( node->arcs[ 1]);
    } else {
	return( node->arcs[ 0]);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Given a node, take all of the 2-connected polylines starting at
 *	that node, and reduce them according to a specified tolerance.
 *	for each of the arcs connected to the node, trace the arcs
 *	as long as they are 2-connected to the next node.  Eliminate
 *	the arcs which are below the tolerance in length.
 * Arguments:
 *	node - node that the 2-connected polylines are connected to.
 *	vectorTolerance - line length tolerance.
 *	angleTolerance - line connect angle tolerance.
 * Modifications:
 *	This function only reduces lines by their length, and the 
 *	length that's passed in is actually the angle tolerance used
 *	for reducing lines.  This should actually take two tolerances,
 *	one for the arc length, the other for the arc angle.
 *
 ****************************************************************************/

static void 
reduceNode( NodePtr	node,
	    float	vectorTolerance,
	    float	angleTolerance)
{

    LineArcPtr	trace[ MAXARCS], newArc, arc, nextArc;
    NodePtr	currNode, nextNode;
    int		ii, numtrace;

    node->visit =  TRUE;
    for ( ii = 0; ii < node->narcs; ii++) {
	arc = node->arcs[ ii];

/* follow a trace of 2-connected nodes either to the end, or to the next
   node with more than one arc */

	numtrace = 0;
	currNode = node;
	do {
	    if ( arc->leftNode == currNode) {
		nextNode = arc->rightNode;
	    } else {
		nextNode = arc->leftNode;
	    }

/* if the node has only one arc, then it's the end of the trace */

	    if ( nextNode->narcs == 1) {
		trace[ numtrace++] = arc;
		nextNode->visit = TRUE;
		break; 
	    } else if (( nextNode->narcs == 2) && !nextNode->visit) {
		nextNode->visit = TRUE;

/* if the node has only two arcs, and hasn't been visit, then it's a 
   candidate for removal, and get the next arc for the node */

		trace[ numtrace++] = arc;
		if ( nextNode->arcs[ 0] == arc) {
		    arc = nextNode->arcs[ 1];
		} else {
		    arc = nextNode->arcs[ 0];
		}
		currNode = nextNode;
	    } else {
		break;		/* node has more than 2 arcs */
	    }
	} while ( TRUE);

/* if the trace has 2 or more arcs, then we might be able to reduce it */
	
	if ( numtrace >= 2) {
	    arc = trace[ 0];
	    for ( ii = 1; ii < numtrace; ii++) {
		nextArc = trace[ ii];

/* if we can reduce the arc, set the current arc pointer's node to the 
   opposite end of the arc, and pick a new arc */

		if ( AmToleranceArc( arc, nextArc, 
				   vectorTolerance, angleTolerance)) {
		    if ( arc->leftNode == nextArc->leftNode) {
			arc->leftNode = nextArc->rightNode;
			newArc = nextNodeArc( nextArc, nextArc->rightNode);
		    } else if ( arc->leftNode == nextArc->rightNode) {
			arc->leftNode = nextArc->leftNode;
			newArc = nextNodeArc( nextArc, nextArc->leftNode); 
		    } else if ( arc->rightNode == nextArc->leftNode) { 
			arc->rightNode = nextArc->rightNode;
			newArc = nextNodeArc( nextArc, nextArc->rightNode);
		    } else if ( arc->rightNode == nextArc->rightNode) {
			arc->rightNode = nextArc->leftNode;
			newArc = nextNodeArc( nextArc, nextArc->leftNode);
		    }

/* mark the arc as deleted */

		    nextArc->leftNode = NULL;
		    nextArc->rightNode = NULL;
		    nextArc = newArc;
		} else {
		    arc = nextArc;
		}
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	reduce a connected network by removing all lines within a tolerance 
 *	of the length of the other lines.
 * Arguments:
 *	nodes - vector of nodes containing arc pointers to reduce.
 *	numNodes - number of nodes in the array.
 *	vectorTolerance - line length tolerance.
 *	angleTolerance - line angle tolerance.
 * Modifications:
 *	This function only reduces lines by their length, and the 
 *	length that's passed in is actually the angle tolerance used
 *	for reducing lines.  This should actually take two tolerances,
 *	one for the arc length, the other for the arc angle.
 *
 ****************************************************************************/

static void 
reduceNetwork( NodePtr	nodes,
	       int	numNodes,
	       float	vectorTolerance,
	       float	angleTolerance)
{
    int	ii;

    for ( ii = 0; ii < numNodes; ii++) {
	if (( nodes[ ii].narcs != 2) && !nodes[ ii].visit) {
	    reduceNode( &nodes[ ii], vectorTolerance, angleTolerance);
	} 
    }

/* second pass for all unreduced 2-connect nodes */

    for ( ii = 0; ii < numNodes; ii++) {
	if ( !nodes[ ii].visit) {
	    reduceNode( &nodes[ ii], vectorTolerance, angleTolerance);
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	do two arcs span AmBetween the same two nodes? Test all the node
 *	pointer combinations.
 * Arguments:
 *	a, b - line arcs to test.
 * Returns:
 *	TRUE if they point to the same nodes in any order, FALSE otherwise.
 *
 ****************************************************************************/

static AnvilBoolean 
colocatedArcs( LineArcPtr	a,
	       LineArcPtr	b)
{
    if ( a->leftNode == b->leftNode) {
	return( a->rightNode == b->rightNode);
    } else if ( a->rightNode == b->leftNode) {
	return( a->leftNode == b->rightNode);
    } else {
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	comparator function for sorting vectors pointed to by the same node 
 *	by the "other" node that they point to.
 * Arguments:
 *	a, b - line arcs to compare.
 *	(static) currNode - the node they radiate from.
 * Returns:
 *	the difference AmBetween the addresses of the node that they point
 *	to that isn't the radiating node.
 *
 ****************************************************************************/

static NodePtr	currNode;
static int 
byOtherNode( LineArcPtr	a,
	     LineArcPtr	b)
{
    if (( a->leftNode == currNode) && ( b->leftNode == currNode)) {
	return( a->rightNode - a->rightNode);
    } else if (( a->rightNode == currNode) &&( b->rightNode == currNode)) {
	return( a->leftNode - b->leftNode);
    } else if (( a->leftNode == currNode) && ( b->rightNode == currNode)) {
	return( a->rightNode - b->leftNode);
    } else if (( a->rightNode == currNode) && ( b->leftNode == currNode)) {
	return( a->leftNode - b->rightNode);
    } else {
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the number of arcs from this node which are not co-located.
 *	sort the arcs by their "opposing nodes", then find any arcs which
 *	point to the same node as this one.
 * Arguments:
 *	node - node to reduce duplicate arcs from.
 * Returns:
 * 	the number of reduced arcs.
 *
 ****************************************************************************/

static int 
numRealArcs( NodePtr	node)
{
    int 	count, ii;

    currNode = node;
    qsort( node->arcs, node->narcs, sizeof( LineArcPtr), 
	   ( AmSearchFn) byOtherNode);

/* reduce the count by the number of colocated arcs */

    count = node->narcs;
    for ( ii = 0; ii < node->narcs - 1; ii++) {
	if ( colocatedArcs( node->arcs[ ii], node->arcs[ ii + 1])) {
	    count--;
	}
    }
    return( count);
}

static LineArcPtr 
getArc( NodePtr	node1,
        NodePtr	node2)
{
    int		ii;
    LineArcPtr	arc;

    for ( ii = 0; ii < node1->narcs; ii++) {
	arc = node1->arcs[ ii];
	if (( arc->leftNode == node2) || ( arc->rightNode == node2)) {
	    return( arc);
	}
    }
    return( NULL);
}

static void 
removeNode( NodePtr leftNode, 
	    NodePtr deadNode,
	    NodePtr rightNode)
{
    int		ii, jj;
    LineArcPtr	arc;

/* take all of the arcs from the right node to the remove node and remove
   them. */
    
    jj = 0;
    for ( ii = 0; ii < rightNode->narcs; ii++) {
	arc = rightNode->arcs[ ii];
	if (( arc->leftNode == deadNode) || ( arc->rightNode == deadNode)) {
	    arc->rightNode = NULL;
	    arc->leftNode = NULL;
	} else {
	    rightNode->arcs[ jj++] = rightNode->arcs[ ii];
	}
    }
    if ( jj == rightNode->narcs) {
	AN_DEBUG( 1, ("no arcs removed!!\n"));
    }

/* take all of the arcs from the left node to the remove node and have
   them point to the right node. */

    for ( ii = 0; ii < leftNode->narcs; ii++) {
	arc = leftNode->arcs[ ii];
	if ( arc->leftNode == deadNode) {
	    arc->leftNode = rightNode;
	    rightNode->arcs[ jj++] = arc;
	} else if ( arc->rightNode == deadNode) {
	    arc->rightNode = rightNode;
	    rightNode->arcs[ jj++] = arc;
	}
    }
    if ( jj != rightNode->narcs) {
	AN_DEBUG( 1, ("bad number of ARCS!!\n"));
	printf( "%f %f 1\n", leftNode->x, leftNode->y);
	printf( "%f %f 2\n", deadNode->x, deadNode->y);
	printf( "%f %f 3\n", rightNode->x, rightNode->y);
	return;
    }
}

/****************************************************************************
 *
 * Description:
 * 	reduce all "strings" which are connected to a single node, applying 
 *	the same reduction to colocated arcs 
 * Arguments:
 *	node - that the arcs are connected to.
 *	vectorTolerance - line length to reduce
 *	angleTolerance - line angle to reduce.
 * Modifications:
 *	There's the same problem with the polygon reduction stuff as with
 *	the vectors, where the angle's aren't tested, and there's also
 *	a problem with shared vertices not being removed properly.
 * 
 ****************************************************************************/

static void 
reducePolygonNode( NodePtr	node,
		   float	vectorTolerance,
		   float	angleTolerance)
{
    LineArcPtr	arc, nextArc;
    NodePtr	currNode, nextNode, nextNextNode, trace[ MAXARCS];
    int		ii, jj, numtrace;

    node->visit =  TRUE;
    for ( ii = 0; ii < node->narcs; ii++) {
	arc = node->arcs[ ii];

/* follow a trace of 2-connected nodes either to the end, or to the next
   node with more than one arc */

	numtrace = 0;
	currNode = node;
	trace[ numtrace++] = currNode;
	do {
	    if ( arc->leftNode == currNode) {
		nextNode = arc->rightNode;
	    } else {
		nextNode = arc->leftNode;
	    }

/* if the node has only one arc, then it's the end of the trace */

	    if ( nextNode->narcs == 1) {
		trace[ numtrace++] = nextNode;
		nextNode->visit = TRUE;
		break; 
	    } else if (( numRealArcs( nextNode) == 2) && !nextNode->visit) {
		nextNode->visit = TRUE;

/* find the AmFirst arc in the next node which doesn't point back to the
   current node and doesn't point to a visited node */

		trace[ numtrace++] = nextNode;
		for ( jj = 0; jj < nextNode->narcs; jj++) {
		    arc = nextNode->arcs[ jj];
		    if (( arc->leftNode != currNode) && 
			( arc->rightNode != currNode))  {
			break;
		    }
		}
		if ( jj == nextNode->narcs) {
AN_DEBUG( 1, ("failed to find next arc!!\n"));
		    break;
		} 
		currNode = nextNode;
	    } else {
		break;		/* node has more than 2 non-colocated arcs */
	    }
	} while ( TRUE);

/* if the trace has 3 or more nodes, then we might be able to reduce it */
	
	if ( numtrace >= 3) {
AN_DEBUG( 40, ("removing trace of %d nodes\n", numtrace));
	    currNode = trace[ 0];
	    nextNode = trace[ 1];
	    arc = getArc( currNode, nextNode);
	    for ( jj = 2; jj < numtrace; jj++) {
		nextNextNode = trace[ jj];
		nextArc = getArc( nextNode, nextNextNode);
		if (( arc == NULL) || ( nextArc == NULL)) {
		    AN_DEBUG( 1, ("BAD ARC!!\n"));
		    break;
		}

/* if we can reduce the arc, set the current arc pointer's node to the 
   opposite end of the arc, and pick a new arc. While we're here, we reduce
   all of the colocated arcs. */

		if ( AmToleranceArc( arc, nextArc, vectorTolerance,
				   angleTolerance)) {
		    AN_DEBUG( 30, ("removing arc\n"));
		    removeNode( currNode, nextNode, nextNextNode);
		    arc = getArc( currNode, nextNextNode);
		    nextNode = nextNextNode;
		} else {
		    node = nextNode;
		    nextNode = nextNextNode;
		}
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	reduce a connected network by removing all lines within a tolerance 
 *	of the slope of other lines, except that co-located lines are 
 *	reduced in parallel.
 * Arguments:
 *	nodes - vector of nodes to reduce.
 *	numNodes - length of the vector.
 *	vectorTolerance - vector length to reduce.
 *	angleTolerance - arc tolerance to reduce (radians)
 * Modifications:
 *	yeah, well, that's the theory, but the co-located line reduction
 *	doesn't work, and there's gaps AmBetween the polygons, and there's
 *	also the angle-testing/length-testing problem that exists with
 *	the other reduction algorithms.
 * 
 ****************************************************************************/

static void 
reducePolygonNetwork( NodePtr	nodes,
		      int	numNodes,
		      float	vectorTolerance,
		      float	angleTolerance)
{
    int	ii;

    for ( ii = 0; ii < numNodes; ii++) {
	if (( numRealArcs( &nodes[ ii]) != 2) && !nodes[ ii].visit) {
	    reducePolygonNode( &nodes[ ii], vectorTolerance, angleTolerance);
	} 
    }

/* second pass for all unreduced 2-connect nodes */

    for ( ii = 0; ii < numNodes; ii++) {
	if ( !nodes[ ii].visit) {
	    reducePolygonNode( &nodes[ ii], vectorTolerance, angleTolerance);
	}
    }
}

/****************************************************************************
 *
 * Description:
 *	verify that for each line arc in the network that it belongs to
 *	the line arc list for its left and right nodes 
 *
 ****************************************************************************/

static void 
verifyNodesArcsAfter( LineArcPtr	*arcs,
		      int		numArcs,
		      NodePtr		nodes,
		      int		numNodes)
{
    int		ii, jj, kk;
    LineArcPtr	arc;
    NodePtr	leftNode, rightNode;

    for ( ii = 0; ii < numArcs; ii++) {
	leftNode = arcs[ ii]->leftNode;
	if ( leftNode != NULL) {
	    if ( leftNode->narcs != 0) {
		for ( jj = 0; jj < leftNode->narcs; jj++) {
		    if ( leftNode->arcs[ jj] == arcs[ ii]) {
			break;
		    }
		}
		if ( jj == leftNode->narcs) {
		    fprintf( stderr, "error at arc %d, leftNode\n", ii);
		}
	    }
	} else {
	    fprintf( stderr, "arc %d has a NULL left node\n", ii);
	}
	rightNode = arcs[ ii]->rightNode;
	if ( rightNode != NULL) {
	    if ( rightNode->narcs != 0) {
		for ( jj = 0; jj < rightNode->narcs; jj++) {
		    if ( rightNode->arcs[ jj] == arcs[ ii]) {
			break;
		    }
		}
		if ( jj == rightNode->narcs) {
		    fprintf( stderr, "error at arc %d, rightNode\n", ii);
		}
	    }
	} else {
	    fprintf( stderr, "arc %d has a NULL right node\n", ii);
	}
    }
    for ( ii = 0; ii < numNodes; ii++) {
	for ( jj = 0; jj < nodes[ ii].narcs; jj++) {
	    arc = nodes[ ii].arcs[ jj];
	    leftNode = arc->leftNode;
	    for ( kk = 0; kk < leftNode->narcs; kk++) {
		if ( leftNode->arcs[ kk] == arc) {
		    break;
		}
	    }
	    if (( leftNode->narcs != 0) && ( kk == leftNode->narcs)) {
		fprintf( stderr, "left node bad at node %d, arc %d\n", ii, jj);
	    }
	    rightNode = arc->rightNode;
	    for ( kk = 0; kk < rightNode->narcs; kk++) {
		if ( rightNode->arcs[ kk] == arc) {
		    break;
		}
	    }
	    if (( rightNode->narcs != 0) && ( kk == rightNode->narcs)) {
		fprintf( stderr, "right node bad at node %d, arc %d\n", ii, jj);
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	reduce a set of lines by a specified tolerance.  This is the top
 *	level function that calls the quadtree algorithm and the 2-connected
 *	vector reduction.
 * Arguments:
 *	lines - vector of lines to reduce.
 *	AmNumLines - length of the vector.
 *	linesz - size of each line in bytes.
 *	xmin, ymin, xmax, ymax - extents covered by the set of lines.
 *	vectorTolernace - vector length tolerance in the lines' 
 *	coordinate system.
 *	angleTolerance - arc tolerance in radians.
 *	nodeTolerance - node tolerance for node matching in the lines'
 *	coordinate system.
 *	multiCallFlag - flag to save buffers in statics, to prevent 
 *	memory fragmentation when the algorithm is called repeatedly.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED if memory couldn't be allocated.
 * 
 ****************************************************************************/

AmErrorCode 
AmReduceLines( AmLinePtr	lines,
	     int	*AmNumLines,
	     int	linesz,
	     float	xmin,
	     float	ymin,
	     float	xmax,
	     float	ymax,
	     float	vectorTolerance,
	     float	angleTolerance,
	     float	nodeTolerance,
	     AnvilBoolean	multiCallFlag)

{
    LineArcPtr	lineArcs, *nodeArcs;
    AmLinePtr	linePtr;
    char	*srcLinePtr, *dstLinePtr;
    NodePtr	nodes;
    int		numNodeArcs, count, ii, dataSize;
    AmErrorCode	errCode;

    if ( *AmNumLines == 0) {
	return( AmOK);
    }

/* convert the segments into a naive network representation */

    errCode = LineToLineArc( lines, *AmNumLines, linesz, 
    			     &lineArcs, &nodes, multiCallFlag);
    if ( errCode != AmOK) {
	return( errCode);
    }

/* if we're getting called repeatedly, then we want to minimize the # of
   times we're allocating blocks, so we preserve them in statics */

    if ( multiCallFlag) {
    	if ( *AmNumLines*2 > maxNodeArcs) {
    	    if ( saveNodeArcs != NULL) {
		free( saveNodeArcs);
	    }
	    maxNodeArcs = *AmNumLines*2;
	    if (( saveNodeArcs = ( LineArcPtr *) 
			malloc( *AmNumLines*2*sizeof( LineArcPtr))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	}
	nodeArcs = saveNodeArcs;
    } else {
	if (( nodeArcs = ( LineArcPtr *) 
		    malloc( *AmNumLines*2*sizeof( LineArcPtr))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }

/* call the quadtree algorithm to consolidate the nodes within the specified
   tolerance */

    numNodeArcs = 0;
    QuadTreeReduce( nodes, *AmNumLines*2, xmin, ymin, xmax, ymax, 
		    nodeTolerance, nodeArcs, &numNodeArcs, multiCallFlag);
		    
    AN_DEBUG( 80, ( "numNodeArcs = %d\n", numNodeArcs));

/* network reduce any 2-connected traces */

    reduceNetwork( nodes, numNodeArcs, vectorTolerance, angleTolerance);

/* take all of the arcs which still have left and right nodes, and copy 
   the data from those arcs */

    count = 0;
    dataSize = linesz - sizeof( AmLineRec);
    dstLinePtr = ( char *) lines;
    for ( ii = 0, srcLinePtr = ( char *) lines; 
	  ii < *AmNumLines; 
	  ii++, srcLinePtr += linesz) {
	if (( lineArcs[ ii].leftNode != NULL) && 
	    ( lineArcs[ ii].rightNode != NULL)) {
	     linePtr = ( AmLinePtr) dstLinePtr;
	     linePtr->x0 = lineArcs[ ii].leftNode->x;
	     linePtr->y0 = lineArcs[ ii].leftNode->y;
	     linePtr->x1 = lineArcs[ ii].rightNode->x;
	     linePtr->y1 = lineArcs[ ii].rightNode->y;

/* copy over the data, and step up the destination line pointer */ 

	     memcpy((( char *) dstLinePtr) + sizeof( AmLineRec),
		    (( char *) srcLinePtr) + sizeof( AmLineRec), dataSize);
	     count++;
	     dstLinePtr += linesz;
	}
    }

/* free up the temporary data now that we're done with it */

    if ( !multiCallFlag) {
	free( nodes);
	free( lineArcs);
	free( nodeArcs);
    }
    *AmNumLines = count;
    return( AmOK);
}



/****************************************************************************
 *
 * Description:
 * 	reduce a set of polygon by a specified tolerance.  This is the top
 *	level function that calls the quadtree algorithm and the 2-connected
 *	vector reduction.
 * Arguments:
 *	polys - vector of polygons to reduce.
 *	numPolys - number of polygons in the vector.
 *	polysz - size of each polygon.
 *	xmin, ymin, xmax, ymax - extents covered by the set of lines.
 *	angleTolerance - arc tolerance in radians.
 *	vectorTolerance - vector tolerance in the source coordinates.
 *	nodeTolerance - node tolerance for node matching in the lines'
 *	coordinate system.
 *	multiCallFlag - flag to save buffers in statics, to prevent 
 *	memory fragmentation when the algorithm is called repeatedly.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED if memory couldn't be allocated.
 * Modifications:
 *	There are problems because of polygons with shared edges.
 * 
 ****************************************************************************/

AmErrorCode AmReducePolygons( AmPolygonPtr	polys,
			    int		numPolys,
			    int		polysz,
			    float	xmin,
			    float	ymin,
			    float	xmax,
			    float	ymax,
			    float	vectorTolerance,
			    float	angleTolerance,
			    float	nodeTolerance,
			    AnvilBoolean	multiCallFlag)
{
    LineArcPtr	lineArcs, *nodeArcs;
    NodePtr	nodes;
    int		numNodeArcs, count, ii, jj, numArcs, newnumpts, oldnumpts;
    AmErrorCode	errCode;
    AmPolygonPtr	polyPtr;
    char	*ptr, *newptr;

/* convert the polygons to a naive network representation */

    errCode = PolygonToLineArc( polys, numPolys, polysz, 
				&lineArcs, &numArcs, &nodes, multiCallFlag);
    if ( errCode != AmOK) {
	return( errCode);
    }

/* if we're getting called repeatedly, then we want to minimize the # of
   times we're allocating blocks, so we preserve them in statics */

    if ( multiCallFlag) {
    	if ( numArcs > maxNodeArcs) {
    	    if ( saveNodeArcs != NULL) {
		free( saveNodeArcs);
	    }
	    maxNodeArcs = numArcs;
	    if (( saveNodeArcs = ( LineArcPtr *) 
			malloc( numArcs*sizeof( LineArcPtr))) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	}
	nodeArcs = saveNodeArcs;
    } else {
	if (( nodeArcs = ( LineArcPtr *) 
		    malloc( numArcs*2*sizeof( LineArcPtr))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }

/* use the quadtree algorithm to consolidate nodes which are within a 
   specified tolerance of each other */

    numNodeArcs = 0;
    QuadTreeReduce( nodes, numArcs, xmin, ymin, xmax, ymax, 
		    nodeTolerance, nodeArcs, &numNodeArcs, multiCallFlag);
    verifyNodesArcsAfter( nodeArcs, numNodeArcs, nodes, numArcs);

/* the reduction algorithm for polygons is a little different than for 
   vectors, because neighboring polygons will cause shared lines AmBetween
   nodes, which would force a non-reduction */

    AN_DEBUG( 80, ( "numNodeArcs = %d\n", numArcs));
    reducePolygonNetwork( nodes, numArcs, vectorTolerance, angleTolerance);

/* copy over the segments from the polygons to the network */

    count = 0;
    oldnumpts = 0;
    newnumpts = 0;

/* scruch the array as the polygons get reduced */

    for ( ii = 0, ptr = ( char *) polys, newptr = ( char *) polys;
    	  ii < numPolys;
    	  ii++, ptr += polysz + oldnumpts*sizeof( AmCoordRec), 
		newptr += polysz + newnumpts*sizeof( AmCoordRec)) {
	oldnumpts = (( AmPolygonPtr) ptr)->numPts;
	polyPtr = ( AmPolygonPtr) ptr;
	newnumpts = 0;
	for ( jj = 0; jj < polyPtr->numPts; jj++) {
	    if (( lineArcs[ count].leftNode != NULL) &&
		( lineArcs[ count].rightNode != NULL)) {
		polyPtr->pts[ newnumpts].x = lineArcs[ count].leftNode->x;
		polyPtr->pts[ newnumpts].y = lineArcs[ count].leftNode->y;
		newnumpts++;
	    }
	    count++;
	}
	polyPtr->numPts = newnumpts;
	memcpy( newptr, ptr, polysz + newnumpts*sizeof( AmCoordRec));
    }

/* free up the temporary data now that we're done with it */

    if ( !multiCallFlag) {
	free( nodes);
	free( lineArcs);
	free( nodeArcs);
    }
    return( AmOK);
}
