#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <anvil/resource.h>
#include <anvil/fns.h>
#include <anvil/strlib.h>
#include <anvil/tagstrs.h>
#include <anvil/datatype.h>
#include <anvil/datafmt.h>
#include <anvil/macros.h>
#include <anvil/geometry.h>
#include <anvil/raspoly.h>

/* polygon rasterization algorithm shamelessly plagarized from the 
   graphics gems books.. */

const int PTBUFFERSIZE = 16384;

/* parametric edge information */

typedef struct Edge {
    float	x;
    float	dx;
    int		ii;
} EdgeRec, *EdgePtr;

static int	ptIndex[ PTBUFFERSIZE], nact, nn;
static EdgeRec	active[ PTBUFFERSIZE];
static AmPtPtr	pt;

/****************************************************************************
 * 
 * Description:
 *	comparator function for comparing the y values of the static 
 *	points in the point array indexed by a and b.
 * Arguments:
 *	a, b - pointers to indices to be compared.
 * Returns:
 *	the difference of the y values of the points indexed by a and b.
 *
 ****************************************************************************/

static int 
compareInd( int	*a,
	    int	*b)
{ 
    return( pt[ *a].y - pt[ *b].y);
}

/****************************************************************************
 * 
 * Description:
 *	compare two active edges 
 *
 ****************************************************************************/

static int 
compareActive( EdgePtr	a,
	       EdgePtr	b)
{
    return(( a->x > b->x) ? 1 : -1);
}

/****************************************************************************
 * 
 * Description:
 *	delete an active edge from the static active list, by scrunching
 *	the array.
 * Arguments:
 *	ii - index of the active edge.
 *
 ****************************************************************************/
  
static void 
delActive( int	ii)
{
    int	jj;

    for ( jj = 0; ( jj < nact) && ( active[ jj].ii != ii); jj++);
    if ( jj >= nact) {
	return;
    }
    nact--;
    memcpy( &active[ jj], &active[ jj + 1], ( nact - jj)*sizeof( EdgeRec));
}

/****************************************************************************
 * 
 * Description:
 *	insert an active edge into the active list.
 * Arguments:
 *	ii index of edge to insert.
 *	y - starting y value of the edge.
 *
 ****************************************************************************/

static void 
insActive( int	ii,
	   int	y)
{
    AmPtPtr	p, q;
    float	dx;
    int		jj;

    jj = ( ii < nn - 1) ? ii + 1 : 0;
    if ( pt[ ii].y < pt[ jj].y) {
	p = &pt[ ii];
	q = &pt[ jj];
    } else {
	p = &pt[ jj];
	q = &pt[ ii];
    }
    dx = ( q->x - p->x)/( q->y - p->y);
    active[ nact].dx = dx;
    active[ nact].x = dx*( y - p->y) + p->x;
    active[ nact].ii = ii;
    nact++;
}

/****************************************************************************
 * 
 * Description:
 * 	write the scanline of a specified value into the matrix 
 * Arguments:
 *	matrix - target matrix.
 *	size - size of each matrix element.
 *	y - y index of the scanline.
 *	xmin, xmax - start and end indices of the scanline.
 *	val - integer value to write into the matrix.
 *	AmSetFn - set function to use.
 *  
 ****************************************************************************/

static void 
writeScan( uchar_t	*matrix, 
	   int		size,
	   int		y,
	   int		cols,
	   int		xmin,
	   int		xmax,
	   int		val,
	   AmIntSetFn	AmSetFn)
{
    xmin = MAX( xmin, 0);
    xmax = MIN( xmax, cols - 1);
    while ( xmin <= xmax) {
       ( AmSetFn)( &matrix[ (y*cols + xmin)*size], NULL, val);
       xmin++;
    }
}

/****************************************************************************
 * 
 * Description:
 * 	generate the extents values for a polygon.
 * Arguments:
 *	poly - polygon to set the extents for (they're stored in the
 *	polygon)
 * Modifications:
 *	This function is duplicated in a couple of places and is likely
 *	to be moved to another file. 
 *  
 ****************************************************************************/

void 
AmExtentsPoly( AmPolygonPtr	poly)
{
    int	ii;

    poly->xmin = MY_HUGE;
    poly->ymin = MY_HUGE;
    poly->xmax = -MY_HUGE;
    poly->ymax = -MY_HUGE;

    for ( ii = 0; ii < poly->numPts; ii++) {
	if ( poly->pts[ ii].x < poly->xmin) {
	    poly->xmin = poly->pts[ ii].x;
	}
	if ( poly->pts[ ii].y < poly->ymin) {
	    poly->ymin = poly->pts[ ii].y;
	}
	if ( poly->pts[ ii].x > poly->xmax) {
	    poly->xmax = poly->pts[ ii].x;
	}
	if ( poly->pts[ ii].y > poly->ymax) {
	    poly->ymax = poly->pts[ ii].y;
	}
    }
}

/****************************************************************************
 * 
 * Description:
 *	rasterize the points in a polygon into a matrix 
 * Arguments:
 *	points - integer scaled points from the polygon
 *	numpts - # of polygon points
 *	matrix - type-free matrix to write into 
 *	matrixDataType - data type for the matrix: determines set function
 *	and size of each element 
 *	rows, cols - dimensions of the matrix 
 *
 ****************************************************************************/

void 
AmRasterPoly( AmPtPtr		points,
	    int			numpts,
	    int			val,
	    uchar_t		*matrix,
	    AmDataTypeEnum	matrixDataType,
	    int			rows,
	    int			cols)
{
    int		ii, jj, kk, y0, y1, y,
		size = AmDataTypeEnumToSize( matrixDataType);
    AmIntSetFn	AmSetFn = AmDataTypeEnumToIntSetFn( matrixDataType);

    if ( numpts < 1 ) return;

    for ( kk = 0; kk < numpts; kk++) {	
	ptIndex[ kk] = kk;
    }
    pt = points;
    qsort( ptIndex, numpts, sizeof( int), ( AmSearchFn) compareInd);
    nact = 0;
    kk = 0;
    nn = numpts;

/* highest and lowest point */

    y0 = pt[ ptIndex[ 0]].y;
    y1 = pt[ ptIndex[ numpts - 1]].y;

/* bound the vertical component by the dimensions of the matrix, 
   so we don't have to bother with clipping */

    for ( y = MAX( y0, 0); y <= MIN( y1, rows - 1); y++) {	/* by scanline */

/* find the polygon edges which intersect the current scanline */

	for ( ; (kk < nn) && pt[ ptIndex[ kk]].y <= y; kk++) {
	    ii = ptIndex[ kk];
	    jj = (ii > 0) ? ii - 1 : nn - 1;
	    if ( pt[ jj].y <= y) {
		delActive( jj);
	    } else if( pt[ jj].y > y) {
		insActive( jj, y);
	    }
	    jj = (ii < nn - 1) ? ii + 1 : 0;
	    if ( pt[ jj].y <= y) {
		delActive( ii);
	    } else if ( pt[ jj].y > y) {
		insActive( ii, y);
	    }
	}
	qsort( active, nact, sizeof( EdgeRec), ( AmSearchFn) compareActive);

/* write the scanlines for the active edges */

	for ( jj = 0; jj < nact; jj += 2) {
	    writeScan( matrix, size, y, cols,
		       ( int) active[ jj].x, ( int) active[ jj + 1].x,
		       val, AmSetFn);
	    active[ jj].x += active[ jj].dx;
	    active[ jj + 1].x += active[ jj + 1].dx;
	}
    }
}

/****************************************************************************
 * 
 * Description:
 * 	take a polygon, clip it, and rasterize it into a matrix. 
 * Arguments:
 *	poly - poylgon to rastersize.
 *	offset - offset of the value of interest in the polygon structure.
 *	dataType - type of the value of interest.
 *	xmin, ymin, xmax, ymax - extents of the matrix in the polygon's
 *	coordinate system.
 *	matrix - type-free matrix to write into 
 *	matrixDataType - data type for the matrix: determines set function
 *	and size of each element 
 *	rows, cols - dimensions of the matrix 
 *
 ****************************************************************************/

void 
AmRasterizePolygon( AmPolygonPtr	poly,
		  int		offset,
		  AmDataTypeEnum	dataType,
		  float		xmin,
		  float		ymin,
		  float		xmax,
		  float		ymax,
		  uchar_t	*matrix,
		  AmDataTypeEnum	matrixDataType,
		  int		rows,
		  int		cols)
{
    int		jj, val, size;
    float	xscale, yscale;
    AmPtRec	points[ PTBUFFERSIZE];
    AmIntAccessFn	AmIntFn = AmDataTypeEnumToIntAccessFn( dataType);

    if ( poly->numPts < 1 ) {
	return;
    }
    size = AmDataTypeEnumToSize( matrixDataType);
    xscale = cols/( xmax - xmin);
    yscale = rows/( ymax - ymin);

/* scale each polygon into integer coordinates and scan convert it */

    for ( jj = 0; jj < poly->numPts; jj++) {
	points[ jj].x = ( int) ((poly->pts[ jj].x - xmin)*xscale);
	points[ jj].y = ( int) ((ymax - poly->pts[ jj].y)*yscale);
    }

/* get the value from the polygon, and write it into the matrix */ 

    val = ( AmIntFn)(( uchar_t *) poly + offset, NULL);
    AmRasterPoly( points, poly->numPts, val, 
		matrix, matrixDataType, rows, cols);
}
