#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* base record class member functions for setting members in derived classes
   according to tag-value pairs. 
   Modifications:
	  the functions which manipulate the tag-values arrays should have
	  more checks to ensure that the arrays aren't overflowed.
*/


#include <stdarg.h>
#include <search.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>

#include <anvil/tagstrs.h>
#include <anvil/strdefs.h>
#include <anvil/strlib.h>
#include <anvil/mem_funcs.h>
#include <anvil/files.h>
#include <anvil/list.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/envdefs.h>
#include <anvil/colorcell.h>
#include <anvil/units.h>
#include <anvil/enums.h>
#include <anvil/record.h>
#include <anvil/register.h>

#define BIGBUFLEN	32768
		
/****************************************************************************
 *
 * Description:
 *  	AmRecordClass Constructor:  Set the valid flag, and NULL out the
 *	extra fields.
 *
 ****************************************************************************/

AmRecordClass::AmRecordClass()
{
    valid = TRUE;
    extraFields = NULL;
    numExtraFields = 0;
}

/****************************************************************************
 *
 * Description:
 *  	AmRecordClass Destructor:  free the extra fields using freeTagValue.
 *
 ****************************************************************************/

AmRecordClass::~AmRecordClass()
{
    int	ii;

/* blow away any extra field definitions */

    if ( extraFields != NULL) {
	for ( ii = 0; ii < numExtraFields; ii++) {
	    freeTagValue( &extraFields[ ii]);
	}
	free( extraFields);
    }
}

/****************************************************************************
 *
 * Description:
 *  	Return the object's valid flag.  This is used throughout the
 *	entire screen object hierarchy to test validity.
 * Returns:
 *	TRUE if object is valid, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmRecordClass::IsValid()
{
    return( valid);
}

/****************************************************************************
 *
 * Description:
 * 	set the object's validity flag. 
 * Arguments:
 *	TRUE/FALSE valid flag
 *
 ****************************************************************************/

void 
AmRecordClass::SetValid( AnvilBoolean	flag)
{
    valid = flag;
}

/****************************************************************************
 *
 * Description:
 * 	wrapper function to initialize the track changes table (which
 *	allows us to find out which values have been changed in the class),
 *	calls the class's virtual setFields() function, then takes the
 *	tag-values that weren't parsed by setFields, and sets them as
 *	string pairs using setExtraFields.
 * Arguments:
 *	tagValues - tag-value pairs to set in the class.
 *	numValues - number of tag-value pairs.
 * Returns:
 *	number of fields which weren't parsed by the setFields() call.
 * Modifications:
 *	this should be changed to return the number of fields unparsed by
 *	setExtraFields(), which should be 0.
 *
 *
 ****************************************************************************/
 
int 
AmRecordClass::setFieldsWrapper( AmTagValuePtr       tagValues,
			         int                 numValues)
{
    AmTrackChanges();
    numValues = setFields( tagValues, numValues);
    setExtraFields( tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	Wrapper function to call the class's virtual getFields() function,
 *	then retrieve any fields which may be stored as "extra fields",
 *	which are string pairs stored in AmRecordClass.
 * Arguments:
 *	tagValues - tag-value pairs to retrieve in the class.
 *	numValues - number of tag-value pairs.
 * Returns:
 *	number of fields which weren't parsed by the setFields() call,
 * Modifications:
 *	this should be changed to return the number of fields unparsed by
 *	setExtraFields(), which should be 0.
 *
 ****************************************************************************/
 
int 
AmRecordClass::getFieldsWrapper( AmTagValuePtr       tagValues,
			         int                 numValues)
{
    numValues = getFields( tagValues, numValues);
    getExtraFields( tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	varargs wrapper member function for setting fields in an object
 *	from a varargs tag-value list.
 * Arguments:
 *	tag - the AmFirst tag in the tag-values table.
 *	varargs - NULL-terminated varargs list of tags and values.
 * Returns:
 *	TRUE if the created object is valid, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmRecordClass::SetVaFields( const char *tag, ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		count;

    va_start( args, tag);
    count = AmMakeTagValueArrayVararg( tagValues, tag, args);

/* call the object's virtual field getting function */

    setFieldsWrapper( tagValues, count);
    return( IsValid());
}

/****************************************************************************
 *
 * Description:
 * 	varargs wrapper member function for retrieving fields in an object
 *	from a varargs tag-value list.  This function assumes that the
 *	tags and values will be freed by the caller (note the tagCopied
 *	and copied flags in the tagValues array)
 * Arguments:
 *	tag - the AmFirst tag in the tag-values table.
 *	varargs - NULL-terminated varargs list of tags and values.
 * Returns:
 *	The number of fields parsed by the low-level getFields call
 *	(see getFieldsWrapper for modifications)
 *
 ****************************************************************************/

int 
AmRecordClass::GetVaFields( const char *tag, ...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         ii;
 
    va_start( args, tag);
    for ( ii = 0, tagValues[ ii].tag = ( char *) tag;
          tagValues[ ii].tag != NULL;
          ii++, tagValues[ ii].tag = va_arg( args, char *)) {
	tagValues[ ii].tagCopied = FALSE;
	tagValues[ ii].copied = FALSE;
        tagValues[ ii].value = va_arg( args, void *);
    }
    va_end( args);

/* call the object's virtual field setting function */

    return( getFieldsWrapper( tagValues, ii));
}

/****************************************************************************
 *
 * Description:
 * 	lower level function to free a non-atomic value, called by
 *	FreeNonAtomicValues().  This is used when tag-values
 *	arrays are copied for creating new objects in AmCopyObject() and
 *	AmTraceObject().
 * Arguments:
 *	dataType - data type of the object to free (maybe)
 *
 ****************************************************************************/

static void 
freeIfNotAtomic( AmDataTypeEnum	dataType,
		 void		*value)
{
    if ( !AmIsAtomic( dataType)) {
	if ( dataType == AmSTRING_PTR) {
	    AmFreeStrVec(( char **) value);
	} else {
	    free( value);
	}

/* floats get converted to doubles and allocated, strings and files 
   are allocated, even though they're atomic. AmUNKNOWN is equivalent
   to string. */

    } else if (( dataType == AmDOUBLE) || 
	       ( dataType == AmSTRING) ||
	       ( dataType == AmUNKNOWN) ||
	       ( dataType == AmFILE)) {
	free( value);
    } 
}

/****************************************************************************
 *
 * Description:
 * 	type-specific freeing function.  Given a dataType and a pointer 
 *	to a value, free the value.
 * Arguments:
 *	dataType - type of the value.
 *	value - value to free.
 *
 ****************************************************************************/

static void 
freeByType( AmDataTypeEnum   dataType,
	    void           *value)
{
    switch ( dataType) {
    case AmSTRING:
    case AmFILE:
    case AmINT_PTR:
    case AmFLOAT_PTR:
    case AmCOLOR_CELL_PTR:
    case AmCOLOR_CELL:
        if ( value != NULL) {
            free( value);
        }
        break;
    case AmSTRING_PTR:
        if ( value != NULL) {
            AmFreeStrVec(( char **) value);
        }
        break;
    default:
        break;
    }
}

/****************************************************************************
 *
 * Description:
 * 	free the non-atomic values in a tag-values array.  This is used 
 *	when tag-values arrays are copied for creating new objects in 
 *	AmCopyObject() and AmTraceObject().  doubles and strings are allocated
 *	in the copy function.
 * Arguments:
 *	tagValues - array of tags and values to free.
 *	numValues - number of values.
 *
 ****************************************************************************/

void 
AmFreeNonAtomicValues( AmTagValuePtr	tagValues,
		       int		numValues)
{
    int			ii;
    AmDataTypeEnum	dataType;

    for ( ii = 0; ii < numValues; ii++) {
	if ( tagValues[ ii].copied) {
	    dataType = AmGetTagType( tagValues[ ii].tag);
	    freeIfNotAtomic( dataType, tagValues[ ii].value);
	}
	if ( tagValues[ ii].tagCopied) {
	    free( tagValues[ ii].tag);
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	write the header strings for a single object in a simple tag: value 
 *	header format 
 * Arguments:
 *	fp - file pointer to write the fields to.
 * Returns:
 *	TRUE on success, FALSE on failure
 * Modifications:
 *	There's a memory leaks when the value is allocated for each 
 *	field in the AmFirst loop.
 *
 ****************************************************************************/

AnvilBoolean AmRecordClass::WriteHeader( FILE	*fp)
{
    AmFieldLengthRec      fieldVec[ MAXTAGS];
    int			  ii, numStrings;
    char		  valbuf[ BIGBUFLEN];
    void		  *value;
    AmDataTypeEnum	  dataType;

    numStrings = dumpFields( fieldVec, FALSE);
    value = ( void *) malloc( sizeof( ulong_t)*4);
    for ( ii = 0; ii < numStrings; ii++) {

/* find the tag and allocate enough memory to get a pointer to the value */

	if (( dataType = AmGetTagType( fieldVec[ ii].tag)) == AmUNKNOWN) {
            AN_DEBUG( 10, ( "Internal tag %s not found.\n", fieldVec[ ii].tag));
            continue;
        }
        GetVaFields( fieldVec[ ii].tag, value, NULL);
	if ( AmPrintValue( valbuf, fieldVec[ ii].tag, fieldVec[ ii].length, 
			   value, dataType)) {
	    fprintf( fp, "%s:\t", fieldVec[ ii].tag);
	    fputs( valbuf, fp);
	    fprintf( fp, "\n");
	}
	if (( dataType == AmSTRING) || ( dataType == AmFILE) || 
	    !AmIsAtomic( dataType)) {
	    freeByType( dataType, *( void **) value);
	}
    }
    free( value);

/* take any "extra" fields (data not parsed by the class fields), and
   dump them too. */

    for ( ii = 0; ii < numExtraFields; ii++) {
	if ( AmPrintValue( valbuf, extraFields[ ii].tag, 0, 
			   &extraFields[ ii].value, 
			   AmGetTagType( extraFields[ ii].tag))) {
	    fprintf( fp, "%s:\t", extraFields[ ii].tag);
	    fputs( valbuf, fp);
	    fprintf( fp, "\n");
	}
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	dump the fields for an object into a file, using the parseable
 *	dumpFields format, with the object name followed by the fields
 *	AmBetween {}'s and fields on <>'s.
 * Arguments:
 *	fp - file pointer to the output file, opened for writing.
 * Returns:
 *	TRUE.
 *
 ****************************************************************************/

AnvilBoolean
AmRecordClass::DumpFields( FILE *fp)
{
    void      		**value, *holder;
    AmFieldLengthRec      fieldVec[ MAXTAGS];
    int       		ii, numStrings, size;
    char		valbuf[ BIGBUFLEN];
    AmRecordClass		**ptr;
    AmDataTypeEnum	dataType;

    numStrings = dumpFields( fieldVec, FALSE);
    for ( ii = 0; ii < numStrings; ii++) {

/* find the tag and allocate enough memory to get a pointer to the value,
   then extract the value from the object */

	if (( dataType = AmGetTagType( fieldVec[ ii].tag)) == AmUNKNOWN) {
            AN_DEBUG( 10, ( "Internal tag %s not found.\n", fieldVec[ ii].tag));
            continue;
        }
        size = AmDataTypeEnumToSize( dataType); 
        if ( size != 0) {
            value = ( void **)malloc( size);
        }  else {
	    holder = NULL;
	    value = &holder;
	}
        GetVaFields( fieldVec[ ii].tag, value, NULL);

/* if the type for the data is AmOBJECT, then we dump that object 
   recursively.  If it's a vector of objects, then iterate the list.  The
   objects are enclosed by a set of brackets to identify the difference
   AmBetween atomic fields and object fields */

	if ( dataType == AmOBJECT) {
	    fprintf( fp, "%s:\t{\n", fieldVec[ ii].tag);
	    (*( AmRecordClass **) value)->DumpFields( fp);
	    fprintf( fp, "}\n");
	} else if ( dataType == AmOBJECT_PTR) {
	    fprintf( fp, "%s:\t{\n", fieldVec[ ii].tag);
	    for ( ptr = *( AmRecordClass ***) value; *ptr != NULL; ptr++) {
		( *ptr)->DumpFields( fp);
	    }
	    fprintf( fp, "}\n");
	} else {
	    if ( AmPrintValue( valbuf, fieldVec[ ii].tag, fieldVec[ ii].length, 
			     value, AmGetTagType( fieldVec[ ii].tag))) {
		fprintf( fp, "%s:\t<", fieldVec[ ii].tag);
		fputs( valbuf, fp);
		fprintf( fp, ">\n");
	    }
	}

/* If the value was atomic, we allocated it.  If not, then it was allocated
   by GetVaFields(), but we have to make sure GetVaFields() didn't return
   NULL.  In the case of an object, only the pointer is returned, not a 
   copy of the object.  Free the vector of strings if it's a returned
   string vector */

	if ( size != 0) { 
	    free( value);
	} else if (( holder != NULL) && ( dataType != AmOBJECT)) {
	    if ( dataType == AmSTRING_PTR) {
		AmFreeStrVec(( char **) holder);
	    } else {
		free( holder);
	    }
	}
    }

/* take any "extra" fields (data not parsed by the class fields), and
   dump them too. */

    for ( ii = 0; ii < numExtraFields; ii++) {
	if ( AmPrintValue( valbuf, extraFields[ ii].tag, 0, 
			 &extraFields[ ii].value, 
			 AmGetTagType( extraFields[ ii].tag))) {
	    fprintf( fp, "%s:\t<", extraFields[ ii].tag);
	    fputs( valbuf, fp);
	    fprintf( fp, ">\n");
	}
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	retreive the values for an object into a TagValue array, sumilar
 *	to dumpFields() except that it generates a tag-value array.
 * Arguments:
 *	tagValues - vector of tag-values to retrieve
 *	numValues - number of returned values.
 * Returns:
 *	TRUE.
 * Modifications:
 *	I think there's a memory leack with the values being malloced
 *	and the copied field being set to FALSE.
 *
 ****************************************************************************/

AnvilBoolean
AmRecordClass::RetreiveTagValues( AmTagValuePtr	tagValues,
				int		*numValues)
{
    AmFieldLengthRec	fieldVec[ MAXTAGS];
    AmTagValueRec	tagCopies[ MAXTAGS];
    int			ii, size, numNonNullTags;
    AmDataTypeEnum	dataType;

    *numValues = dumpFields( fieldVec, TRUE);
    for ( ii = 0; ii < *numValues; ii++) {
	tagValues[ ii].tag = fieldVec[ ii].tag;
	tagValues[ ii].tagCopied = FALSE;
	tagValues[ ii].copied = TRUE;

/* allocate the pointers to hold the returned data */

        if (( dataType = AmGetTagType( fieldVec[ ii].tag)) != AmUNKNOWN) {
	    size = AmDataTypeEnumToSize( dataType); 
	    if ( size != 0) {
		tagValues[ ii].value = ( void *)malloc( size);
	    } else {
		tagValues[ ii].value = ( void *)malloc( sizeof( void *));
	    }
	} else {
	    tagValues[ ii].value = NULL;
	    AN_DEBUG( 1, ( "cantFindTag %s", tagValues[ ii].tag));
	}
    }

/* get fields trashes the tagValues array, so we need to make a copy. */

    memcpy( tagCopies, tagValues, *numValues*sizeof( AmTagValueRec));
    getFields( tagCopies, *numValues);

/* for any pointer values that are NULL, drop the tag */

    numNonNullTags = 0;
    for ( ii = 0; ii < *numValues; ii++) {
	if (( dataType = AmGetTagType( tagValues[ ii].tag)) != AmUNKNOWN) {
	    if ((( dataType == AmSTRING_PTR) ||
		 ( dataType == AmINT_PTR) ||
		 ( dataType == AmFLOAT_PTR) ||
		 ( dataType == AmFUNC_PTR) ||
		 ( dataType == AmCOLOR_CELL_PTR) ||
		 ( dataType == AmFILE) ||
		 ( dataType == AmSTRING)) &&
		(* ( void **) tagValues[ ii].value == NULL)) {
		free( tagValues[ ii].value);
	    } else {
		tagValues[ numNonNullTags++] = tagValues[ ii];
	    }
	} else {
	    AN_DEBUG( 1, ( "cantFindTag %s", tagValues[ ii].tag));
	}
    }
    *numValues = numNonNullTags;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	compare the tag and the field definition tag.  Used in the bsearch()
 *	call to compare tag names against fields.  Used by getFieldDef()
 *	and AmScreenObjClass::ParseColorCellArgs().
 * Arguments:
 *	tag - tag to compare.
 *	fieldDef - field definition to compare against.
 * Results:
 *	strcmp() of the tag string and fieldDef tag string.
 *
 ****************************************************************************/

int 
AmCompareFieldTag( char               *tag,
		 AmFieldDefPtr        fieldDef)
{
    return( strcmp( tag, fieldDef->tag));
}

/****************************************************************************
 *
 * Description:
 * 	do a bsearch on the field definitions table to find the field
 *	definition for a tag.  The tag may have the "file." prefix,
 *	which means the data is read from an ASCII file, so if it's
 *	there, it's stripped off the tag before the bsearch().  This
 *	is why the field definitions for each of the objects have to
 *	be sorted on the field tag.
 * Arguments:
 *	tag - tag to compare.
 *	fieldDefs - field definition array to compare against.
 * Results:
 *	matching field definition pointer, or NULL if it wasn't found.
 *
 ****************************************************************************/

static AmFieldDefPtr 
getFieldDef( char		*tag,
	     AmFieldDefPtr	fieldDefs,
	     int		numFieldDefs)
{
    char	*ptag;

/* field defs may be NULL */
  
    if ( fieldDefs == NULL) {
	return( NULL);
    }

/* check to see if the tag is a ".", expression, and if so, then use the
   string after the "." as the tag for the type lookup */

    if (( ptag = strchr( tag, '.')) != NULL) { 
	ptag++;
    } else {
	ptag = tag;
    }
    return(( AmFieldDefPtr) bsearch(( const void *) ptag,
				  fieldDefs, numFieldDefs,
				  sizeof( AmFieldDefRec), 
				  ( AmSearchFn) AmCompareFieldTag));
}

/****************************************************************************
 *
 * Description:
 * 	free a tag-value pair depending on the flags and data type.
 * Arguments:
 *	tagValue - tag-value pair to free.
 *
 ****************************************************************************/

void 
AmRecordClass::freeTagValue( AmTagValuePtr	tagValue)
{
    AmDataTypeEnum	dataType;

    if ( tagValue->copied) {
	dataType = AmGetTagType( tagValue->tag);

/* if it was stored by a type, then free it by that type */

	if ( dataType != AmUNKNOWN) {
	    freeByType( dataType, tagValue->value);
	} else {

/* type defaults to string if it isn't found in string definition list */

	    if ( tagValue->value != NULL) {
		free( tagValue->value);
	    }
	}
    }
    if ( tagValue->tagCopied) {
    	free( tagValue->tag);
    }
}

/****************************************************************************
 *
 * Description:
 * 	given a field definition vector, and a vector of tag-value pairs, 
 *	set the values in the structure.  This mashes the tagValues vector, 
 *	and resets it to the non-found fields, and returns the number of 
 *	fields that weren't found.  The atomic values are parsed AmFirst,
 *	because they may contain the vector lengths for the vector values
 *	that are parsed in the second pass.
 * Arguments:
 *	fieldDefs - field definitions vector, sorted on the tag name.
 *	numFieldDefs - length of the field definitions vector.
 *	tagValues, numValues - tag-values to set in the class.
 * Returns:
 *	number of unparsed values.
 *
 ****************************************************************************/

int AmRecordClass::SetFieldByDefinitions( AmFieldDefPtr	fieldDefs,
					int		numFieldDefs,
					AmTagValuePtr	tagValues,
					int		numValues)
{
    uchar_t		*dataPtr = ( uchar_t *) this, *offsetPtr;
    AmTagValueRec		nonAtomicValues[ MAXTAGS], unusedValues[ MAXTAGS];
    int			numNonAtomicValues, ii, numUnusedValues;
    AmFieldDefPtr		fieldDef;
    AmDataTypeEnum	dataType;

/* set the atomic types AmFirst, because they will dictate how values will be
   set for non-atomic types */

    numNonAtomicValues = 0;
    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {

/* find out the field's data type */
	
	if (( dataType = AmGetTagType( tagValues[ ii].tag)) == AmUNKNOWN) {
	    unusedValues[ numUnusedValues++] = tagValues[ ii];
	    continue;
	}
	if ( AmIsAtomic( dataType)) {

/* and its field definition within the class */

	    fieldDef = getFieldDef( tagValues[ ii].tag, 
				    fieldDefs, numFieldDefs);
	    if ( fieldDef == NULL) {
		unusedValues[ numUnusedValues++] = tagValues[ ii];
	    } else {

/* get the data offset from the field definition, and set the value according
   to the data type.  Mark the address as changed */ 

		offsetPtr = dataPtr + fieldDef->offset;
		AmMarkChanged( offsetPtr);
		if ( !AmSetScalar( offsetPtr, dataType, 
				 fieldDef->storage, tagValues[ ii].value)) {
		    AN_DEBUG( 10, ( "Data type %d not found.\n", dataType));
		}

/* if the tag or value was copied, then free them here */

		if ( tagValues[ ii].copied) {
		    free( tagValues[ ii].value);
		}
		if ( tagValues[ ii].tagCopied) {
		    free( tagValues[ ii].tag);
		}
	    }
	} else {
	    nonAtomicValues[ numNonAtomicValues++] = tagValues[ ii];
	}
    }

/* now process the non-atomic values */

    for ( ii = 0; ii < numNonAtomicValues; ii++){

/* find out the field's data type */

	if (( dataType = AmGetTagType( nonAtomicValues[ ii].tag)) == AmUNKNOWN) {
	    unusedValues[ numUnusedValues++] = nonAtomicValues[ ii];
	    continue;
	}

/* get the data offset from the field definition, and set the value according
   to the data type */

	fieldDef = getFieldDef( nonAtomicValues[ ii].tag, 
				fieldDefs, numFieldDefs);
	if ( fieldDef == NULL) {
	    unusedValues[ numUnusedValues++] = nonAtomicValues[ ii];
	} else {
	    offsetPtr = dataPtr + fieldDef->offset;
	    AmMarkChanged( offsetPtr);
	    if ( !AmSetVector( offsetPtr, dataType,
			     fieldDef->storage, nonAtomicValues[ ii].value, 
			     *( int *) ( dataPtr + fieldDef->countOffset))) {
		AN_DEBUG( 10, ( "Data type %d not found.\n", dataType));
	    }

/* if the tag or value was copied, then free them here */

	    if ( nonAtomicValues[ ii].copied) {
		freeByType( dataType, nonAtomicValues[ ii].value);
	    }
	    if ( nonAtomicValues[ ii].tagCopied) {
		free( nonAtomicValues[ ii].tag);
	    }
	}
    }

/* copy the unused values back to the original tag array and return the number
   of unprocessed values */
    
    memcpy(( void *) tagValues, ( void *) unusedValues, 
	   sizeof( AmTagValueRec)*numUnusedValues);
    return( numUnusedValues);
} 

/****************************************************************************
 *
 * Description:
 * 	given a field definition vector, and a vector of tag-value pairs, 
 *	get the values from the structure.  This mashes the tagValues vector, 
 *	and resets it to the non-found fields, and returns the number of 
 *	fields that weren't found. Note that AmCOLOR_CELL and 
 *	AmCOLOR_CELL_PTR may be retreived by GetFieldByDefinitions(), 
 *	but aren't set by SetFieldByDefintions(), because 
 *	ParseColorCellArgs() is used for screen objects, for color 
 *	allocations.
 * Arguments:
 *	fieldDefs - field definitions vector, sorted on the tag name.
 *	numFieldDefs - length of the field definitions vector.
 *	tagValues, numValues - tag-values to retreive from the class.
 * Returns:
 *	number of unparsed values.
 *
 ****************************************************************************/

int AmRecordClass::GetFieldByDefinitions( AmFieldDefPtr	fieldDefs,
					int		numFieldDefs,
					AmTagValuePtr	tagValues,
					int		numValues)
{
    uchar_t		*dataPtr = ( uchar_t *) this, *offsetPtr;
    AmTagValueRec		unusedValues[ MAXTAGS];
    int			ii, numUnusedValues, length=0;
    AmFieldDefPtr		fieldDef;
    void		**ptr;
    AmDataTypeEnum	dataType;

/* set the atomic types AmFirst, because they will dictate how values will be
   set for non-atomic types */

    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {

/* find out the field's data type */

	if (( dataType = AmGetTagType( tagValues[ ii].tag)) == AmUNKNOWN) {
	    unusedValues[ numUnusedValues++] = tagValues[ ii];
	    continue;
	}

/* and its field definition within the class */

	fieldDef = getFieldDef( tagValues[ ii].tag, fieldDefs, numFieldDefs);
	if ( fieldDef == NULL) {
	    unusedValues[ numUnusedValues++] = tagValues[ ii];
	} else {

/* get the data offset from the field definition, and set the value according
   to the data type */

	    offsetPtr = dataPtr + fieldDef->offset;
	    ptr = ( void **) tagValues[ ii].value;
	    length = *( int *) ( dataPtr + fieldDef->countOffset);
	    if ( !AmGetValue( offsetPtr, dataType, fieldDef->storage,
			    ptr, length)) {
		AN_DEBUG( 10, ( "Data type %d not found\n", dataType));
	    }
	}

/* if the tag was copied, then free it here.  Set the value copied flag */

	if ( tagValues[ ii].tagCopied) {
	    free( tagValues[ ii].tag);
	}
	tagValues[ ii].copied = TRUE;
    } 

/* copy the unused values back to the original tag array and return the number
   of unprocessed values */
    
    memcpy(( void *) tagValues, ( void *) unusedValues, 
	   sizeof( AmTagValueRec)*numUnusedValues);
    return( numUnusedValues);
}

/****************************************************************************
 *
 * Description:
 * 	given the field definitions vector for a class, the number of 
 *	fields, and a vector of strings to except, dump the class 
 *	definition to a field/length array 
 * Arguments:
 *	fieldLengths - vector of field names with associated vector
 *	lengths for vector fields (returned by dumpFields())
 *	numStrings - length of the fieldLengths array.
 *	fieldDefs - field definitions vector, sorted on the tag name.
 *	numFieldDefs - length of the field definitions vector.
 *	exceptFields - fields to NOT dump, typically AmNdata, so a
 *	bogus data pointer doesn't get read in.
 * Returns:
 *	number of unparsed values.
 * Modifications:
 *	Hopefully, we'll be able to clean up the design and get rid of
 *	the exceptFields parameter.
 *
 ****************************************************************************/

int
AmRecordClass::DumpFieldsByDefinitions( AmFieldLengthPtr	fieldLengths,
                                      int         	numStrings,
                                      AmFieldDefPtr 	fieldDefs,
                                      int         	numFieldDefs,
				      char		**exceptFields)
{
    int		ii, jj, kk;
    uchar_t	*ptr;
    AnvilBoolean	writeTag;

    if ( numStrings + numFieldDefs > MAXTAGS) {
        AN_DEBUG( 9, ( "AmRecordClass::DumpFieldByDefinition - number of tags exceeds MAXTAG (= %d)\n", MAXTAGS));
        return ( numStrings);
    }

/* if the field isn't atomic, use the field definition count offset
   pointer to get the length */

    kk = 0;
    for ( ii = 0; ii < numFieldDefs; ii++) {

/* if the tag is in one of the exception fields, then we don't bother to 
   write it */

	writeTag = TRUE;
	if ( exceptFields != NULL) {
	    for ( jj = 0; exceptFields[ jj] != NULL; jj++) {
		if ( AmMatchTag( fieldDefs[ ii].tag, exceptFields[ jj])) {
		    writeTag = FALSE;
		    break;
		}
	    }
	}
	if ( writeTag) {

/* if the tag is for a vector, then set the vector's length from the structure
   offset. If it's atomic, set the length to 0 to avoid purify UMR's */ 

	    fieldLengths[ numStrings + kk].tag = fieldDefs[ ii].tag; 
	    if ( !AmIsAtomic( AmGetTagType( fieldDefs[ ii].tag))) {
		ptr = (( uchar_t *) this) + fieldDefs[ ii].countOffset;
		fieldLengths[ numStrings + kk].length = *( int *) ptr;
	    } else {
		fieldLengths[ numStrings + kk].length = 0;
	    }
            kk++;
	}
    }

/* add the number of fields */

    return( numStrings + kk);
}

/****************************************************************************
 *
 * Description:
 * 	base implementation for dumpFields() causes failed assertion 
 * Arguments:
 *	fieldLengths - vector of field names with associated vector
 *	allFields - a soon to be unused parameter, hopefully.
 *
 ****************************************************************************/

int 
AmRecordClass::dumpFields( AmFieldLengthPtr	fieldLengths,
			 AnvilBoolean	allFields)
{
    ASSERT( FALSE);
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	comparison function for sorting AmTagValuePtr's by tag.
 * Arguments:
 *	a, b - tagValuePtr's to compare.
 * Returns:
 *	strcmp() of the tagValuePtrs' tags.
 *
 ****************************************************************************/

static int 
byTag( AmTagValuePtr	a,
       AmTagValuePtr	b)
{
    return( strcmp( a->tag, b->tag));
}

/****************************************************************************
 *
 * Description:
 * 	utility function for setExtraFields to copy a tag-value pair. 
 *	the tag name is allocated, and the tagCopied flag is set to TRUE.
 * Arguments:
 *	dst - tag-value to copy to.
 *	src - tag-value to copy from.
 *
 ****************************************************************************/


static void 
copyTagValue( AmTagValuePtr	dst,
	      AmTagValuePtr	src)
{
   dst->tag = strdup( src->tag);
   dst->tagCopied = TRUE;
   dst->value = src->value;
   dst->copied = src->copied;
}

/****************************************************************************
 *
 * Description:
 * 	used by setExtraFields, this sets the tag value.  If the dataType
 *	for the tag is known (meaning it's in the tag-type lookup table),
 *	then set the tag-value according to the type, otherwise, assume
 *	it's a string.
 * Arguments:
 *	dst - tag-value to set.
 *	src - tag-value to copy from.
 *
 ****************************************************************************/

void 
AmRecordClass::AmSetTagValue( AmTagValuePtr	dst,
			      AmTagValuePtr	src)
{
    AmDataTypeEnum	dataType;

/* copy the tag, and find out if the value is in the string def list.  If not,
   then the type defaults to string */

    dst->tag = strdup( src->tag);
    if (( dataType = AmGetTagType( src->tag)) == AmUNKNOWN) {
	dataType = AmSTRING;
    }

/* since we can't do lengths, the only vector type we can handle is
   NULL-terminated dynamic.  By the way, if the data type is COLOR_CELL,
   this will choke, because the values aren't allocated large enough */

    if ( AmIsAtomic( dataType)) {
	if ( !AmSetScalar(( uchar_t *) &dst->value, dataType, 
			  AmDYNAMIC, src->value)) {
	    AN_DEBUG( 10, ( "Data type %d not found\n", dataType));
	}
	if (( dataType == AmSTRING) || ( dataType == AmFILE)) {
	    dst->copied = TRUE;
	} else {
	    dst->copied = FALSE;
	}
    } else {
	if ( !AmSetVector(( uchar_t *) &dst->value, dataType, AmDYNAMIC, 
			( uchar_t *) src->value, 0)) {
	    AN_DEBUG( 10, ( "Data type %d not found\n", dataType));
	}
	dst->copied = TRUE;
    }

/* free the source data if we have to */

    if ( src->copied) {
	freeByType( dataType, src->value);
    }
    if ( src->tagCopied) {
	free( src->tag);
    }
    dst->tagCopied = TRUE;
}

/****************************************************************************
 *
 * Description:
 * 	set the fields which are not parsed by setFields().  These are 
 *	set in the "extraFields" member of the record class, so user data
 *	can be added to an object.
 * Arguments:
 *	tagValues - tag-values to set in the object.
 *	numValues - number of values to set.
 *
 ****************************************************************************/

void 
AmRecordClass::setExtraFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    int			count, tagCnt, extraCnt, diff, ii;
    AmTagValuePtr 	newFields;

/* bail if there's nothing to do */

    if ( numValues == 0) {
	return;
    }

/* sort the new tags by field, so they can be zipper merged */

    qsort( tagValues, numValues, sizeof( AmTagValueRec), ( AmSearchFn) byTag);
    tagCnt = 0;
    extraCnt = 0;
    count = 0;

/* find out the resulting number of values from the merge */

    while (( tagCnt < numValues) && ( extraCnt < numExtraFields)) {
	diff = byTag( &tagValues[ tagCnt], &extraFields[ extraCnt]);
	if ( diff < 0) {
	    tagCnt++;
	} else if ( diff > 0) {
	    extraCnt++;
	} else {
	    tagCnt++;
	    extraCnt++;
	}
	count++;
    }
    count += ( numValues - tagCnt) + ( numExtraFields - extraCnt);

/* allocate a new extraFields vector for the soon to be merged values */

    newFields = ( AmTagValuePtr) calloc( count, sizeof( AmTagValueRec));
    tagCnt = 0;
    extraCnt = 0;
    count = 0;

/* needs to actually copy the fields */

    while (( tagCnt < numValues) && ( extraCnt < numExtraFields)) {
	diff = byTag( &tagValues[ tagCnt], &extraFields[ extraCnt]);
	if ( diff < 0) {
	    AmSetTagValue( &newFields[ count], &tagValues[ tagCnt++]);
	} else if ( diff > 0) {
	    copyTagValue( &newFields[ count], &extraFields[ extraCnt++]);
	} else {

/* if the same, then copy the new values */

	    AmSetTagValue( &newFields[ count], &tagValues[ tagCnt++]);
	    extraCnt++;
	}
	count++;
    }
    while ( tagCnt < numValues) {
	AmSetTagValue( &newFields[ count++], &tagValues[ tagCnt++]);
    } 
    while ( extraCnt < numExtraFields) {
	copyTagValue( &newFields[ count++], &extraFields[ extraCnt++]);
    }

/* free the old extra fields and set the new ones */
    
    if ( extraFields != NULL) {
	for ( ii = 0; ii < numExtraFields; ii++) {
	    if ( extraFields[ ii].tagCopied) {
		free( extraFields[ ii].tag);
	    }
	    if ( extraFields[ ii].copied) {
		free( extraFields[ ii].value);
	    }
	}
	free( extraFields);
    }
    numExtraFields = count;
    extraFields = newFields;
}

/****************************************************************************
 *
 * Description:
 * 	retreive the fields which are not parsed by getFields().  These are 
 *	pulled from the "extraFields" member of the record class, so user data
 *	can be retrieved from an object.
 * Arguments:
 *	tagValues - tag-values to retrieve from the object.
 *	numValues - number of values to retrieve.
 *
 ****************************************************************************/

void 
AmRecordClass::getExtraFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    AmTagValueRec		target;
    AmTagValuePtr		result;
    int			ii;
    AmDataTypeEnum	dataType;

    if ( extraFields != NULL) {
	for ( ii = 0; ii < numValues; ii++) {
	    target.tag = tagValues[ ii].tag;
	    result = ( AmTagValuePtr) bsearch( &target, extraFields, 
					     numExtraFields,
					     sizeof( AmTagValueRec), 
					     ( AmSearchFn) byTag);

/* get the string def for the tag to find the type and copy the value */

	    if ( result != NULL) {
		dataType = AmGetTagType( tagValues[ ii].tag);
		if ( dataType == AmUNKNOWN) {
		    dataType = AmSTRING;
		}
		if ( !AmGetValue(( uchar_t *) &result->value, dataType, AmDYNAMIC,
			       ( void **) tagValues[ ii].value, 0)) {
		    AN_DEBUG( 10, ("type %d not found\n", dataType));
		}
	    } else {
		tagValues[ ii].value = NULL;
	    }
	}
    } else {
	for ( ii = 0; ii < numValues; ii++) {
	    tagValues[ ii].value = NULL;
	}
    }
}

/*********************************************************************
 *
 * Description:
 *     Given a record, a string tag, a void pointer value, and a type
 *     the function calls SetVaFields on the passed-in record,
 *     casting the void pointer to the correct value.
 * Arguments:
 *	record - record to set the value in
 *	tag - tag to set the value for
 *	value - value to set.
 * Returns:
 *	TRUE if the tag's data type was found, FALSE otherwise.
 *
 *********************************************************************/

AnvilBoolean
AmSetRecordField( AmRecordClass  *record,
                const char   *tag,
                void         *value)
{
    AmDataTypeEnum        type;

    type = AmGetTagType( tag);
    switch( type) {
        case AmSHORT:
            record->SetVaFields( tag, *( short *)value, NULL);
            break;
        case AmUNSIGNED_SHORT:
            record->SetVaFields( tag, *( ushort_t *)value, NULL);
            break;
        case AmINT:
        case AmUNITS:
        case AmUNSIGNED_INT:
            record->SetVaFields( tag, *( uint_t *)value, NULL);
            break;
        case AmLONG:
            record->SetVaFields( tag, *( long *)value, NULL);
            break;
        case AmUNSIGNED_LONG:
            record->SetVaFields( tag, *( ulong_t *)value, NULL);
            break;
        case AmFLOAT:
            record->SetVaFields( tag, *( float *)value, NULL);
            break;
        case AmDOUBLE:
            record->SetVaFields( tag, *( double *)value, NULL);
            break;
        case AmBOOLEAN:
            record->SetVaFields( tag, *( AnvilBoolean *)value, NULL);
            break;
        case AmENUM:
            record->SetVaFields( tag, *( int *)value, NULL);
            break;
        case AmSTRING:
            record->SetVaFields( tag, ( char *)value, NULL);
            break;
        case AmFILE:
            record->SetVaFields( tag, ( char *)value, NULL);
            break;
        default:
            DO_ERROR( AmBAD_ARGUMENT);
            return( FALSE);
    }
    return( TRUE);
}
