#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions for manipulating rectangles in every way, shape and form
   imagineable.  There are two types of rectangles, RectangleRecs, which
   are integer rectangles, and RealRectRects, which are floating point
   rectangles */ 

#include <math.h>
#if defined(_WIN32) || defined (__BSD44__) 
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <anvil/project.h>
#include <anvil/ratio.h>
#include <anvil/macros.h>
#include <anvil/rect.h>

const ulong_t AmRECT_ABOVE = 0x1;
const ulong_t AmRECT_BELOW = 0x2;
const ulong_t AmRECT_LEFT = 0x4;
const ulong_t AmRECT_RIGHT = 0x8;

/****************************************************************************
 *
 * Description:
 * 	is a real point inside a real rectangle? 
 * Arguments:
 *	rect - rectangle to test against.
 *	x,y - point to test.
 * Returns:
 *	TRUE if the point is inside the rectangle, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmInRealRect( AmRealRectRec const	*rect,
	    float		x,
	    float		y)
{
    return(( x > rect->x) && ( y > rect->y) &&
	   ( x <= rect->x + rect->width) && ( y <= rect->y + rect->height));
}

/****************************************************************************
 *
 * Description:
 * 	clip a rectangle by another rectangle.
 * Arguments:
 *	pSrcRect - rectangle to clip.
 *	pClipRect - rectangle to clip it against.
 * Returns:
 *	TRUE if the rectangles intersect or one contains the other,
 *	FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmClipRect( AmRectangleRec const *pSrcRect,
	  AmRectanglePtr       pClipRect)
{
    int xSrc, ySrc, cxSrc, cySrc,
        xDst, yDst, cxDst, cyDst,
        x, y, cx, cy;

    xSrc  = pSrcRect->x;
    ySrc  = pSrcRect->y;
    cxSrc = pSrcRect->width;
    cySrc = pSrcRect->height;

    xDst  = pClipRect->x;
    yDst  = pClipRect->y;
    cxDst = pClipRect->width;
    cyDst = pClipRect->height;

/* Check for intersection first. */

    if ( xSrc + cxSrc < xDst) {
        return( FALSE);
    } else if ( xSrc > xDst + cxDst) {
        return (FALSE);
    } else if ( ySrc + cySrc < yDst) {
        return( FALSE);
    } else if ( ySrc > yDst + cyDst) {
        return (FALSE);
    }

/* There is some kind of intersection, so compute the new clip
   rectangle. */

    x = MAX( xSrc, xDst);
    cx = MIN( xSrc + cxSrc, xDst + cxDst) - x;

    y = MAX( ySrc, yDst);
    cy = MIN( ySrc + cySrc, yDst + cyDst) - y;

/* Copy the result into the clip rectangle. */

    pClipRect->x = x;
    pClipRect->y = y;
    pClipRect->width = cx;
    pClipRect->height = cy;

    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	clip a vector of rectangles in place.  and return the number 
 *	of resulting rectangles, and TRUE/FALSE if there's more than 0.
 * Arguments: 
 *	clipRect - rectangle to clip against.
 *	rectVec - rectangle vector to clip.
 *	nRect - number of rectangles, returns number of rectangles inside
 *	the clip rectangle.
 * Returns:
 *	TRUE if any of rectangles intersect the clip rectangle, 
 *	FALSE otherwise
 *
 ****************************************************************************/

AnvilBoolean 
AmClipRectVec( AmRectangleRec const	*clipRect,
	     AmRectanglePtr	rectVec,
	     int		*nRect)
{
    int	ii, count;

    count = 0;
    for ( ii = 0; ii < *nRect; ii++) {
	if ( AmClipRect( clipRect, &rectVec[ ii])) {
	    rectVec[ count++] = rectVec[ ii];
	}
    }

    *nRect = count;
    return( *nRect > 0);
}

/****************************************************************************
 *
 * Description:
 * 	is a rectangle degenerate? 
 * Arguments:
 *	rect - rect to test.
 * Returns:
 *	TRUE if the width or height are less than 0.
 *
 ****************************************************************************/

AnvilBoolean 
AmDegenRect( AmRectangleRec const *rect)
{

/* Note that a rectangle of width 0 is actually valid. */

    return(( rect->width < 0) || ( rect->height < 0));
}

/****************************************************************************
 *
 * Description:
 * 	is a point in a rectangle?
 * Arguments:
 *	rect - rect to test against.
 *	x, y - point to test. 
 * Returns:
 *	TRUE if the point is inside the rectangle, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPtInRect( AmRectangleRec const	*rect,
	  int			x,
	  int			y)
{
    return(( x >= rect->x) && ( y >= rect->y) &&
	   ( x <= rect->x + rect->width) && ( y <= rect->y + rect->height));
}

/****************************************************************************
 *
 * Description:
 * 	set the union of two rectangles, and return TRUE.
 * Arguments:
 *	rect1, rect2 - source rectangles.
 *	rect3 - their minimum bounding rectangle.
 * Returns:
 *	TRUE, really.
 *
 ****************************************************************************/

AnvilBoolean 
AmUnionRect( AmRectangleRec const  	*rect1,
	   AmRectangleRec const	*rect2,
	   AmRectanglePtr		rect3)
{
    int	x1, y1;
    
    rect3->x = MIN( rect1->x, rect2->x);
    rect3->y = MIN( rect1->y, rect2->y);
    x1 = MAX( rect1->x + rect1->width, rect2->x + rect2->width);
    y1 = MAX( rect1->y + rect1->height, rect2->y + rect2->height);
    rect3->width = x1 - rect3->x;
    rect3->height = y1 - rect3->y;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	take the union of a set of rectangles, and generate the minimum
 *	bounding rectangle.
 * Arguments:
 *	rectVec - vector of rectangles.
 *	nRect - length of the vector.
 *	rect - minimum bounding rectangle.
 *
 ****************************************************************************/

void 
AmUnionRectVec( AmRectangleRec const	*rectVec,
	      int			nRects,
	      AmRectanglePtr		rect)
{
    int	ii, x0, y0, x1, y1;

    x0 = rectVec[ 0].x;
    y0 = rectVec[ 0].y;
    x1 = rectVec[ 0].x + rectVec[ 0].width;
    y1 = rectVec[ 0].y + rectVec[ 0].height;
    for ( ii = 1; ii < nRects; ii++) {
	x0 = MIN( rectVec[ ii].x, x0);
	y0 = MIN( rectVec[ ii].y, y0);
	x1 = MAX( rectVec[ ii].x + rectVec[ ii].width, x1);
	y1 = MAX( rectVec[ ii].y + rectVec[ ii].height, y1);
    }
    rect->x = x0;
    rect->y = y0;
    rect->width = x1 - x0;
    rect->height = y1 - y0;
}


/****************************************************************************
 *
 * Description:
 * 	set the intersection of two rectangles, and return a boolean if 
 *	they intersect 
 * Arguments:
 *	rect1, rect2 - rectangles to test.
 *	rect3 - intersection rectangle ( degenerate if they don't)
 * Returns:
 *	TRUE if they intersect, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmIntersectRect( AmRectangleRec const  	*rect1,
	       AmRectangleRec const	*rect2,
	       AmRectanglePtr		rect3)
{
    int	x1, y1;

    rect3->x = MAX( rect1->x, rect2->x);
    rect3->y = MAX( rect1->y, rect2->y);
    x1 = MIN( rect1->x + rect1->width, rect2->x + rect2->width);
    y1 = MIN( rect1->y + rect1->height, rect2->y + rect2->height);
    rect3->width = x1 - rect3->x;
    rect3->height = y1 - rect3->y;
    return(( rect3->width >= 0) && ( rect3->height >= 0));
}

/****************************************************************************
 *
 * Description:
 * 	set the intersection of two real rectangles, and return a boolean if
 *  	they intersect 
 * Arguments:
 *	rect1, rect2 - rectangles to test.
 *	rect3 - intersection rectangle ( degenerate if they don't)
 * Returns:
 *	TRUE if they intersect, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmIntersectRect( AmRealRectRec const  	*rect1,
	       AmRealRectRec const	*rect2,
	       AmRealRectPtr		rect3)
{
    int	x1, y1;

    rect3->x = MAX( rect1->x, rect2->x);
    rect3->y = MAX( rect1->y, rect2->y);
    x1 = (int)MIN( rect1->x + rect1->width, rect2->x + rect2->width);
    y1 = (int)MIN( rect1->y + rect1->height, rect2->y + rect2->height);
    rect3->width = x1 - rect3->x;
    rect3->height = y1 - rect3->y;
    return(( rect3->width >= 0) && ( rect3->height >= 0));
}

/****************************************************************************
 *
 * Description:
 * 	translate a rectangle from its coordinate system to another 
 *	rectangle's coordinate coordinate system.
 * Arguments:
 *	fromRect - rectangle to translate from.
 *	toRect - rectangle to translate to.
 * Returns:
 *	static pointer to the newly translated rectangle.
 *
 ****************************************************************************/

AmRectangleRec const *
AmRelativeRect( AmRectangleRec const	*fromRect,
	      AmRectangleRec const	*toRect)
{
    static AmRectangleRec	returnRect;

/* if the rectangle is NULL, then return the destination rectangle */

    if ( fromRect == NULL) {
	returnRect = *toRect;
	returnRect.x = 0;
	returnRect.y = 0;
    } else {
	returnRect.x = MAX( fromRect->x - toRect->x, 0);
	returnRect.y = MAX( fromRect->y - toRect->y, 0);
	returnRect.width = MIN( toRect->width, 
				fromRect->width + toRect->x - fromRect->x);
	returnRect.height = MIN( toRect->height, 
				 fromRect->height + toRect->y - fromRect->y);
    }
    return( &returnRect);
}

/****************************************************************************
 *
 * Description:
 * 	given front and back rectangles, generate the difference AmBetween 
 *	the two rectangles, and create the flag vector
 * Arguments:
 *      frontRect - rectangle on top.
 *      backRect - backward rectangle.
 *      leftRect - generated leftRect (may be degenerate)
 *      topRect - generated topRect (may be degenerate)
 *      rightRect - generated rightRect (may be degenerate)
 *      bottomRect - generated bottomRect (may be degenerate)
 *      flags - returned flags composed of ABOVE, LEFT, RIGHT, BELOW.
 * Returns:
 *	AmNO_OVERLAP if the expose rectangles don't overlap.
 *	AmCOVER if the front rectangle covers the back rectangle entirely.
 *	AmINTERSECT if the rectangles intersect.
 * Modifications:
 *	this is almost a complete dumplicate of the AmCreateExposureRects()
 *	routine in lib/misc/rect.cpp, and will likely be consolidated.
 *
 ****************************************************************************/

AmExposeEnum 
AmCreateExposureRects( const AmRectangleRec	*frontRect,
		       const AmRectangleRec	*backRect,
		       AmRectanglePtr		leftRect,
		       AmRectanglePtr		topRect,
		       AmRectanglePtr		rightRect,
		       AmRectanglePtr		bottomRect,
		       ulong_t			*flags)
{
    int          x1, x2, x3, x4, y1, y2, y3, y4;

    *flags = 0x0;

/* Check if there's any overlap */

    if (( frontRect->x > backRect->x + backRect->width) ||
	( frontRect->y > backRect->y + backRect->height) ||
	( frontRect->x + frontRect->width < backRect->x) ||
	( frontRect->y + frontRect->height < backRect->y)) {
	return( AmNO_OVERLAP);
    }

/* see if the rectangle covers it completely */

    if (( frontRect->x <= backRect->x)  &&
	( frontRect->y <= backRect->y) &&
	( frontRect->x + frontRect->width >= backRect->x + backRect->width) &&
	( frontRect->y + frontRect->height >= backRect->y + backRect->height)) {
	return( AmCOVER);
    }

/* Figure out the 8 special points that need to be considered. */

    x1 = backRect->x;
    x2 = backRect->x + backRect->width;
    x3 = frontRect->x;
    x4 = frontRect->x + frontRect->width;

    y1 = backRect->y;
    y2 = backRect->y + backRect->height;
    y3 = frontRect->y;
    y4 = frontRect->y + frontRect->height;

/* Figure out the (at most) four exposed areas. */


    topRect->x = x1;
    topRect->y = y1;
    topRect->width = MIN( x4 - x1, x2 - x1);
    topRect->height = y3 - y1;

    leftRect->x = x1;
    leftRect->y = MAX( y1, y3);
    leftRect->width = x3 - x1;
    leftRect->height = y2 - leftRect->y;

    bottomRect->x = MAX( x1, x3);
    bottomRect->y = y4;
    bottomRect->width = x2 - bottomRect->x;
    bottomRect->height = y2 - y4;

    rightRect->x = x4;
    rightRect->y = y1;
    rightRect->width = x2 - x4;
    rightRect->height = MIN( y4 - y1, y2 - y1);

    if (( leftRect->width >= 0) && ( leftRect->height >= 0)) {
	*flags |= AmRECT_LEFT;
    }
    if (( topRect->width >= 0) && ( topRect->height >= 0)) {
	*flags |= AmRECT_ABOVE;
    }
    if (( rightRect->width >= 0) && ( rightRect->height >= 0)) {
	*flags |= AmRECT_RIGHT;
    }
    if (( bottomRect->width >= 0) && ( bottomRect->height >= 0)) {
	*flags |= AmRECT_BELOW;
    }
    return(( *flags != 0) ? AmINTERSECT : AmNO_OVERLAP);
}

/****************************************************************************
 *
 * Description:
 * 	same as AmCreateExposureRects, except that if there's no 
 *	intersection, it returns the old rectangle, and generates
 * 	a vector of rectangles.
 * Arguments:
 *	oldRect - behind rectangle (or AmLast position of object)
 *	newRect - front rectangle (or new position of object)
 * 	rectVec - returned vector of the exposed rectangles.
 *	nRects - returns number of exposed rectangles in the vector.
 * Returns:
 *	TRUE if the rectangles intersect or don't overlap, meaning
 *	that expose rectangles were generated, FALSE if the new rectangle
 *	covers the old rectangle entirely.
 * Modifications:
 *	this will probably move to lib/misc/rect.cpp as well
 *
 ****************************************************************************/

AnvilBoolean 
AmDifferenceRects( const AmRectangleRec	*oldRect,
		   const AmRectangleRec	*newRect,
		   AmRectanglePtr	rectVec,
		   int			*nRects)
{
    AmRectangleRec	leftRect, topRect, rightRect, bottomRect;
    ulong_t		flags;
    AmExposeEnum	exposeCode;

    exposeCode = AmCreateExposureRects( newRect, oldRect, &leftRect, &topRect,
				      &rightRect, &bottomRect, &flags);
    if ( exposeCode == AmCOVER) {	
	*nRects = 0;
	return( FALSE);
    } else if ( exposeCode == AmNO_OVERLAP) {
	rectVec[ 0] = *oldRect;
	*nRects = 1;
	return( TRUE);
    } else {
	*nRects = 0;
	if ( flags & AmRECT_LEFT) {
	    rectVec[ (*nRects)++] = leftRect;
	}
	if ( flags & AmRECT_ABOVE) {
	    rectVec[ (*nRects)++] = topRect;
	}
	if ( flags & AmRECT_RIGHT) {
	    rectVec[ (*nRects)++] = rightRect;
	}
	if ( flags & AmRECT_BELOW) {
	    rectVec[ (*nRects)++] = bottomRect;
	}
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	same as AmDifferenceRects(), but the new rectangle is included as well
 * Arguments:
 *	oldRect - behind rectangle (or AmLast position of object)
 *	newRect - front rectangle (or new position of object)
 * 	rectVec - returned vector of the exposed rectangles.
 *	nRects - returns number of exposed rectangles in the vector.
 * Returns:
 *	TRUE if the rectangles intersect or don't overlap, meaning
 *	that expose rectangles were generated, FALSE if the new rectangle
 *	covers the old rectangle entirely.
 * Modifications:
 *	this might move to lib/misc/rect.cpp as well
 *
 ****************************************************************************/

AnvilBoolean 
AmDifferenceIncludeRects( const AmRectangleRec	*oldRect,
			  const AmRectangleRec	*newRect,
			  AmRectanglePtr	rectVec,
			  int			*nRects)
{
    AmDifferenceRects( oldRect, newRect, rectVec, nRects);
    rectVec[ (*nRects)++] = *newRect;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	are two rectangles equal?
 * Arguments:
 *	a, b - rectangles to test.
 * Returns:
 *	TRUE if they're the same, FALSE otherwise. 
 *
 ****************************************************************************/

AnvilBoolean 
AmEqualRect( AmRectangleRec const   *a,
	   AmRectangleRec const   *b)
{
    return(( a->x == b->x) && ( a->y == b->y) && 
	   ( a->width == b->width) && ( a->height == b->height));
}

/****************************************************************************
 *
 * Description:
 * 	given a source matrix and coordinate system, and a destination 
 *	matrix and coordinate system, calculate the source and destination 
 *	rectangles in integer coordinates.
 * Arguments:
 *	srcRows, srcCols - dimensions of the source matrix.
 *	srcXmin, srcYmin, srcXmax, srcYmax - coordinates of the source matrix.
 *	dstRows, dstCols - dimensions of the destination matrix.
 *	dstXmin, dstYmin, dstXmax, dstYmax - coordinates of destination matrix.
 *	srcRect - returned source rectangle.
 *	dstRect - returned destination rectangle.
 * Returns:
 *	TRUE if the rectangles overlap, FALSE otherwise.
 * Modifications:
 *	I'm not sure the overlap test is correct.
 *
 ****************************************************************************/

AnvilBoolean 
AmSrcAndDestRects( int		srcRows,
		 int		srcCols,
		 float		srcXmin,
		 float		srcYmin,
		 float		srcXmax,
		 float		srcYmax,
		 int		dstRows,
		 int		dstCols,
		 float		dstXmin,
		 float		dstYmin,
		 float		dstXmax,
		 float		dstYmax,
		 AmRectanglePtr	srcRect,
		 AmRectanglePtr	dstRect)
{
    float	allXmin, allXmax, allYmin, allYmax;
    int		x1, y1;

/* find the intersection of the coordinate systems */

    allXmin = MAX( srcXmin, dstXmin);
    allYmin = MAX( srcYmin, dstYmin);
    allXmax = MIN( srcXmax, dstXmax);
    allYmax = MIN( srcYmax, dstYmax);
    if (( allXmin >= allXmax) || ( allYmin >= allYmax)) {
	return( FALSE);
    }

/* calculate the source rectangle */

    srcRect->x = (short int)(srcCols*( allXmin - srcXmin)/( srcXmax - srcXmin));
    srcRect->y = (short int)(srcRows*( srcYmax - allYmax)/( srcYmax - srcYmin));
    x1 = (int)(srcCols*( allXmax - srcXmin)/( srcXmax - srcXmin));
    y1 = (int)(srcRows*( srcYmax - allYmin)/( srcYmax - srcYmin));
    srcRect->width = x1 - srcRect->x;
    srcRect->height = y1 - srcRect->y;

/* and the destination rectangle */

    dstRect->x = (short int)(dstCols*( allXmin - dstXmin)/( dstXmax - dstXmin));
    dstRect->y = (short int)(dstRows*( dstYmax - allYmax)/( dstYmax - dstYmin));
    x1 = (int)(dstCols*( allXmax - dstXmin)/( dstXmax - dstXmin));
    y1 = (int)(dstRows*( dstYmax - allYmin)/( dstYmax - dstYmin));
    dstRect->width = x1 - dstRect->x;
    dstRect->height = y1 - dstRect->y;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	take a real rectangle and create the bounding integer rectangle.
 * Arguments:
 *	srcRect - real rectangle.
 *	dstRect - resulting bounding integer rectangle. 
 *
 ****************************************************************************/

void 
AmMakeIntRect( AmRealRectRec const	*srcRect,
	     AmRectanglePtr	dstRect)
{
    dstRect->x = (short int)floor( srcRect->x);
    dstRect->y = (short int)floor( srcRect->y);
    dstRect->width = (short int)ceil( srcRect->x + srcRect->width) - dstRect->x;
    dstRect->height = (short int)ceil( srcRect->y + srcRect->height) - dstRect->y;
}

/****************************************************************************
 *
 * Description:
 * 	take a real rectangle and create the contained integer rectangle.
 * Arguments:
 *	srcRect - real rectangle.
 *	dstRect - resulting bounding integer rectangle. 
 *
 ****************************************************************************/

void 
AmMakeContainedIntRect( AmRealRectRec const	*srcRect,
			AmRectanglePtr		dstRect)
{
    dstRect->x = (short int)ceil( srcRect->x);
    dstRect->y = (short int)ceil( srcRect->y);
    dstRect->width = (short int)floor( srcRect->x + srcRect->width) - dstRect->x;
    dstRect->height = (short int)floor( srcRect->y + srcRect->height) - dstRect->y;
}


/****************************************************************************
 *
 * Description:
 * 	apply flips to an integer rectangle 
 * Arguments:
 * 	rect - rectangle to flip.
 *	flipRows, flipCols - flip flags. 
 *
 ****************************************************************************/

void 
AmApplyFlips( AmRectanglePtr   rect,
	    AnvilBoolean        flipRows,
	    AnvilBoolean        flipCols)
{
    if ( flipCols) {
        rect->x += rect->width;
        rect->width = -rect->width;
    }
    if ( flipRows) {
        rect->y += rect->height;
        rect->height = -rect->height;
    }
}

/****************************************************************************
 *
 * Description:
 * 	apply flips to a real rectangle 
 * Arguments:
 * 	rect - rectangle to flip.
 *	flipRows, flipCols - flip flags. 
 *
 ****************************************************************************/

void 
AmApplyFlips( AmRealRectPtr    rect,
	    AnvilBoolean        flipRows,
	    AnvilBoolean        flipCols)
{
    if ( flipCols) {
        rect->x += rect->width;
        rect->width = -rect->width;
    }
    if ( flipRows) {
        rect->y += rect->height;
        rect->height = -rect->height;
    }
}

/****************************************************************************
 *
 * Description:
 * 	find out if two rectangles can be unified into a single rectangle
 *	if they have matching edges. 
 * Arguments:
 *	a,b - rects to unify.
 *	c  - resulting unified rectangle
 * Returns:
 *	TRUE if the rectangles can be unified, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmUnifyRects( AmRectangleRec const   *a,
	    AmRectangleRec const   *b,
	    AmRectanglePtr         c)
{

/* b inside a */

    if (( a->x <= b->x) && ( a->y <= b->y) &&
        ( a->x + a->width >= b->x + b->width) &&
        ( a->y + a->height >= b->y + b->height)) {
        *c = *a;
        return( TRUE);
    }

/* a inside b */

    if (( b->x <= a->x) && ( b->y <= a->y) &&
        ( b->x + b->width >= a->x + a->width) &&
        ( b->y + b->height >= a->y + a->height)) {
        *c = *b;
        return( TRUE);
    }

/* connected horizontally */

    if (( a->y == b->y) && ( a->y + a->height == b->y + b->height) &&
        (( a->x + a->width == b->x) || ( b->x + b->width == a->x))) {
        c->y = a->y;
        c->height = a->height;
        c->x = MIN( a->x, b->x);
        c->width = a->width + b->width;
        return( TRUE);
    }

/* connected vertically */

    if (( a->x == b->x) && ( a->x + a->width == b->x + b->width) &&
        (( a->y + a->height == b->y) || ( b->y + b->height == a->y))) {
        c->x = a->x;
        c->width = a->width;
        c->y = MIN( a->y, b->y);
        c->height = a->height + b->height;
        return( TRUE);
    }
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	take a vector of rectangles and unify any rectangles which are 
 *	contained or adjoining.
 * Arguments: 
 *	rects - vector of rectangles to unify.
 *	nRects - number of rectangles, returns resulting number of 
 *	unified rectangles 
 * Modificiations:
 *	should return an error code if the alloca failed.
 *
 ****************************************************************************/

void 
AmUnifyRectVec( AmRectanglePtr	rects,
	      int		*nRects)
{
    int			ii, jj, count;
    AmRectangleRec	newRect;
    AnvilBoolean		*deleted;

    deleted = ( AnvilBoolean *) alloca( *nRects*sizeof( AnvilBoolean));
    if ( deleted == NULL) {
	DO_ERROR( AmALLOCA_FAILED);
	return;
    }

/* for each rectangle, scan all the rectangles which follow it in the list.
   If the rectangles are joined, replace the rectangle with the joined 
   rectangle, and mark the other rectangle as deleted */

    for ( ii = 0; ii < *nRects; ii++) {
	for ( jj = ii + 1; jj < *nRects; jj++) {
	    if ( !deleted[ jj]) {
		if ( AmUnifyRects( &rects[ ii], &rects[ jj], &newRect)) {
		    rects[ ii] = newRect;
		}
	    }
	}
    }

/* remove the deleted rectangles */

    count = 0;
    for ( ii = 0; ii < *nRects; ii++) {
	if ( !deleted[ ii]) {
	    rects[ count++] = rects[ ii];
	}
    }
    *nRects = count;
}

/****************************************************************************
 *
 * Description:
 *	width, height to pixel and inverse functions to deal with
 *	off-by-1 problems.
 *
 ****************************************************************************/

int
AmWidthToPixels( int cx)
{
    return( cx + 1);
}


int
AmHeightToPixels( int cy)
{
    return( cy + 1);
}


int
AmPixelsToWidth( int nHorizPixels)
{
    return( nHorizPixels - 1);
}


int
AmPixelsToHeight( int nVertPixels)
{
    return( nVertPixels - 1);
}

