#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions to register client data and function pointers to string values
   so files can be written and read with function names */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#if !defined(__BSD44__) 
#include <search.h>
#endif
#include <malloc.h>
#include <sys/types.h>
#include <anvil/register.h>
#include <anvil/tsearch.h>

typedef struct register_ {
    char        *name;
    void        *value;
} AmRegisterRec, *AmRegisterPtr;

static void	*registerAddrRoot = NULL,
		*registerStrRoot = NULL;

/****************************************************************************
 *
 * Description:
 *	comparator function for comparing names in the registry.
 * Arguments:
 *	a, b - registrys to compare
 * Returns:
 *	strcmp of the names of the registrys
 *
 ****************************************************************************/

static int 
compareRegisterStr( AmRegisterPtr	a,
		    AmRegisterPtr	b)
{
    return( strcmp( a->name, b->name));
}

/****************************************************************************
 *
 * Description:
 *	comparator function for comparing addresses in the registry.
 * Arguments:
 *	a, b - registrys to compare
 * Returns:
 *	difference of their addresses 
 *
 ****************************************************************************/

static int 
compareRegisterAddr( AmRegisterPtr	a,
		     AmRegisterPtr	b)
{
    return(( uchar_t *) a->value - ( uchar_t *) b->value);
}

/****************************************************************************
 *
 * Description:
 *	add a name-value combination to the registry.
 * Arguments:
 *	name - unique string name of the value to register.
 *	value - value to register.
 * Modifications:
 *	should return an error code if the node couldn't be allocated.
 *	should also return an error code if the entry is a duplicate on
 *	the string name.
 *
 ****************************************************************************/

void 
AmAddRegister( char *name, void *value)
{
    AmRegisterPtr	reg;

    reg = ( AmRegisterPtr) malloc( sizeof( AmRegisterRec));
    if ( reg == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
 	return;
    }
    reg->name = name;
    reg->value = value;
    tsearch( reg, &registerStrRoot, ( AmSearchFn) compareRegisterStr);
    tsearch( reg, &registerAddrRoot, ( AmSearchFn) compareRegisterAddr);
}

/****************************************************************************
 *
 * Description:
 * 	find a name-value combination in the registry.
 * Arguments:
 *	name - name to search for.
 * Returns:
 *	either the value in the registry, or NULL if it isn't found.
 *
 ****************************************************************************/


void *
AmGetRegister( char *name)
{
    AmRegisterPtr	*found;
    AmRegisterRec	target;

    target.name = name;
    found = ( AmRegisterPtr *) tfind( &target, 
				    ( const void **) &registerStrRoot, 
				    ( AmSearchFn) compareRegisterStr);
    if ( found == NULL) {
	return( NULL);
    } else {
	return(( *found)->value);
    }
}

/****************************************************************************
 *
 * Description:
 * 	given an registered value, return the associated string name.
 * Arguments:
 *	value - value to search for.
 * Returns:
 *	either the string in the registry, or NULL if it isn't found.
 * Modifications:
 *	should return a const char *
 *
 ****************************************************************************/

char *
AmGetRegisteredString( void *value)
{
    AmRegisterPtr	*found;
    AmRegisterRec	target;

    target.value = value;
    found = ( AmRegisterPtr *) tfind( &target, 
				    ( const void **) &registerAddrRoot, 
			            ( AmSearchFn) compareRegisterAddr);
    if ( found == NULL) {
	return( NULL);
    } else {
	return(( *found)->name);
    }
}

/****************************************************************************
 *
 * Description:
 * 	remove a name and the associated value from the registry.
 * Arguments:
 *	name - unique string name to search on.
 * Returns:
 *	TRUE if the value was found, FALSE otherwise.
 * Modifications:
 *	should take a const char *
 *
 ****************************************************************************/

AnvilBoolean 
AmDeleteRegister( char *name)
{
    AmRegisterRec target;
    AmRegisterPtr *found;

    target.name = name;
    found = ( AmRegisterPtr *) tfind( &target, 
				    ( const void **) &registerAddrRoot, 
				     ( AmSearchFn) compareRegisterAddr);
    if ( found != NULL) {
	target.value = ( *found)->value;
	tdelete( &target, &registerStrRoot, ( AmSearchFn) compareRegisterStr);
	tdelete( &target, &registerAddrRoot, ( AmSearchFn) compareRegisterAddr);
	return( TRUE);
    } else {
	return( FALSE);
    }
}
