#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* regular expression parsing function.  This is not a perfect implementation
   of UNIX regular expressions, because it doesn't use the backup table
   graph method, but it works well enough for syntax checking. This file
   works in combination with the lexer regexp.l and the parser regexp.y */

#include <stdio.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <string.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/tsearch.h>
#include <anvil/regexp.h>
#include <anvil/regparse.h>

int regexpparse( void);
void regexprestart( FILE *);

static void 	*regexpRoot = NULL;	/* root of precompiled regexp tree */
static char     statBuffer[ 1024];	/* static buffer for yyparse */	
static int      statStrSize;		/* for copying buffer in yyinput() */

/* static regular expression used becuase yacc is broken, and doesn't
   allow return values from the parser */

AmREPtr	statRE;

/****************************************************************************
 *
 * Description:
 * 	add a single regular expression to a list of regular expressions 
 *	(called by the parser)
 * Arguments:
 *      a - regular expression to add.
 *	set - set to add it to.
 * Returns:
 *      realloced set of regular expressions.
 * Modofications:
 *	the regular expression set should be a linked list, and the 
 *	function should set an error code if the new node couldn't be
 *	allocated.
 *
 ****************************************************************************/


AmREPtr 
AmAddRegexRE( AmREPtr a, 
	    AmREPtr set)
{
    if ( set == NULL) {
	statRE = set = ( AmREPtr) malloc( sizeof( AmRERec));
	set->tag = AmREGEXP;
	set->REVal.fullRE.numSet = 0;
	set->REVal.fullRE.REset = NULL;
    }
    set->REVal.fullRE.numSet++;
    set->REVal.fullRE.REset = 
	( AmREPtr *) realloc( set->REVal.fullRE.REset, 
			    set->REVal.fullRE.numSet*sizeof( AmRERec *));
    set->REVal.fullRE.REset[ set->REVal.fullRE.numSet - 1] = a;
    return( set);
}

/****************************************************************************
 *
 * Description:
 * 	repeat arbitrary '*' expression, which is a regexp followed by '*'
 *	(called by the parser)
 * Arguments:
 *	a - regular expression to repeat.
 * Returns:
 *	new "star" regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated
 *
 ****************************************************************************/

AmREPtr 
AmMakeStarRE( AmREPtr a)
{
    AmREPtr	re;

    re = ( AmREPtr) malloc( sizeof( AmRERec));
    re->tag = AmSTAR_REGEXP;
    re->REVal.starRE.RE = a;
    return( re);
}

/****************************************************************************
 *
 * Description:
 * 	make an OR'ed regular expression from two regular expressions 
 *	(called by the parser) left '|' right.
 * Arguments:
 *	left, right - left and right regular expressions to create the
 *	OR expression from.
 * Returns:
 *	new "OR" regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated
 *
 ****************************************************************************/

AmREPtr 
AmOrRegexRE( AmREPtr	left,
	   AmREPtr	right)
{
    AmREPtr	re;

    re = ( AmREPtr) malloc( sizeof( AmRERec));
    re->tag = AmOR_REGEXP;
    re->REVal.orRE.REleft = left;
    re->REVal.orRE.REright = right;
    return( re);
}

/****************************************************************************
 *
 * Description:
 * 	make a character range sub-charset regular expression.
 *	(called by the parser) "A-Z"
 * Arguments:
 *	a - starting character.
 *	b - ending character
 * Returns:
 *	new range regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated,
 *	or if a and b aren't ordered.
 *
 ****************************************************************************/

AmREPtr 
AmMakeRangeRE( char *a,
	     char *b)
{
    AmREPtr	re;

    re = ( AmREPtr) malloc( sizeof( AmRERec));
    re->tag = AmRANGE_REGEXP;
    re->REVal.rangeRE.minChar = *a;
    re->REVal.rangeRE.maxChar = *b;
    return( re);
}

/****************************************************************************
 *
 * Description:
 * 	add a single-charset regular expression to a charset 
 *	(called by the parser)
 * Arguments:
 *	a - single charset regular expression.
 *	set - charset-set regular expression.
 * Returns:
 *	new range regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated,
 *
 ****************************************************************************/

AmREPtr 
AmAddCharsetRE( AmREPtr a, 
	      AmREPtr set)
{
    if ( set == NULL) {
	set = ( AmREPtr) malloc( sizeof( AmRERec));
	set->tag = AmCHARSET_REGEXP;
	set->REVal.charsetRE.numSet = 0;
	set->REVal.charsetRE.charset = NULL;
    }
    set->REVal.charsetRE.numSet++;
    set->REVal.charsetRE.charset = 
	( AmREPtr *) realloc( set->REVal.charsetRE.charset, 
			    set->REVal.charsetRE.numSet*sizeof( AmREPtr));
    set->REVal.charsetRE.charset[ set->REVal.charsetRE.numSet - 1] = a;
    return( set);
}

/****************************************************************************
 *
 * Description:
 * 	create a '.' character regular expression ( which accepts any
 *	single character, called by the parser) 
 * Arguments:
 *	a - single charset regular expression.
 *	set - charset-set regular expression.
 * Returns:
 *	new range regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated,
 *
 ****************************************************************************/

AmREPtr 
AmMakeDotRE( char *str)
{
    AmREPtr	re;

    re = ( AmREPtr) malloc( sizeof( AmRERec));
    re->tag = AmDOT_REGEXP;
    re->REVal.charRE.character = *str;
    return( re);
}

/****************************************************************************
 *
 * Description:
 * 	create a single character regular expression from a string
 *	(called by the parser)
 * Arguments:
 *	str - one character string expression.
 * Returns:
 *	one character string regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated,
 *
 ****************************************************************************/

AmREPtr 
AmMakeOneCharRE( char *str)
{
    AmREPtr	re;

    re = ( AmREPtr) malloc( sizeof( AmRERec));
    re->tag = AmCHAR_REGEXP;
    re->REVal.charRE.character = *str;
    return( re);
}

/****************************************************************************
 *
 * Description:
 *	make a "number of occurences" expression, of either a range of
 *	occurrences, minimum # of occurrences, or a fixed range of
 *	occurrences (called by the parser) \{1,2\}, \{1,\}, or \{1\}
 * Arguments:
 *	re - regular expression to repeat.
 *	str - string representing range of occurrences.
 * Returns:
 *	one character string regular expression.
 * Modifications:
 *	should set an error code if the new node couldn't be allocated,
 *
 ****************************************************************************/

AmREPtr 
AmMakeOccurRE( AmREPtr 		re,
	     char 		*str, 
	     AmRangeTypeEnum	type)
{
    AmREPtr	newRE;

    newRE = ( AmREPtr) malloc( sizeof( AmRERec));
    newRE->tag = AmOCCURRENCE_REGEXP;
    newRE->REVal.occurRE.RE = re;
    switch ( type) {
    case AmRANGE_OCCURRENCES:
	sscanf( str, "\\{%d,%d\\}", 
		&newRE->REVal.occurRE.min, &newRE->REVal.occurRE.max);
	break;
    case AmMIN_OCCURRENCES:
	sscanf( str, "\\{%d,\\}", &newRE->REVal.occurRE.min);
	newRE->REVal.occurRE.max = 255;
	break;
    case AmRANGE_FIXED:
	sscanf( str, "\\{%d\\}", &newRE->REVal.occurRE.min);
	newRE->REVal.occurRE.max = newRE->REVal.occurRE.min;
	break;
    default:
	break;
    }
    return( newRE);
}

/****************************************************************************
 *
 * Description:
 * 	determine whether a string matches a regular expression, with a
 *	flag to allow partial matches.
 * Arguments:
 *	re - regular expression to test against.
 *	str - string to test.
 *	partialFlag - flag for partial matches.
 * Returns:
 *	a pointer to a NULL string if the string matched, or the pointer 
 *	to start of the string where it didn't match.
 *
 ****************************************************************************/

static char *
REMatch( AmREPtr	re,
	 char	*str,
	 int	partialFlag)
{
    char	*pstr;
    int		i, numOccur;
    AmREPtr	firstCharRE;

    switch ( re->tag) {
    case AmOR_REGEXP:
	if ( *str == '\0') {
	    return( partialFlag ? str : NULL);
	}
	pstr = REMatch( re->REVal.orRE.REleft, str, partialFlag);
	if ( pstr != NULL) {
	    return( pstr);
	} 
	pstr = REMatch( re->REVal.orRE.REright, str, partialFlag);
	return( pstr);
    case AmDOT_REGEXP:		/* dot accepts anything */
	if ( *str == '\0') {
	    return( partialFlag ? str : NULL);
	}
	return( str + 1);
    case AmCHAR_REGEXP:

/* check for a matching character */

	if ( *str == '\0') {
	    return( partialFlag ? str : NULL);
	} else if ( *str == re->REVal.charRE.character) {
	    return( str + 1);
	} else {
	    return( NULL);
	}
    case AmRANGE_REGEXP:

/* check to see if the character is in range */

	if ( *str == '\0') {
	    return( partialFlag ? str : NULL);
	} else if (( *str >= re->REVal.rangeRE.minChar) &&
		   ( *str <= re->REVal.rangeRE.maxChar)) {
	    return( str + 1);
	} else {	
	    return( NULL);
	}
    case AmCHARSET_REGEXP:

/* test each member in the charset, and check for the '^' sign at the start to
   negate the whole set */

	if ( *str == '\0') {
	    return( partialFlag ? str : NULL);
	} else {
    	    firstCharRE = re->REVal.charsetRE.charset[ 0];
	    if ( re->REVal.charsetRE.numSet == 0) {
		return( str);
	    } else {
		if (( firstCharRE->tag == AmCHAR_REGEXP) && 
		    ( firstCharRE->REVal.charRE.character == '^')) {
		    for ( i = 0; i < re->REVal.charsetRE.numSet; i++) {
			pstr = REMatch( re->REVal.charsetRE.charset[ i], 
					str, partialFlag);
			if ( pstr != NULL) {
			    return( NULL);
			} 
		    } 
		    return( str + 1);
		} else {
		    for ( i = 0; i < re->REVal.charsetRE.numSet; i++) {
			pstr = REMatch( re->REVal.charsetRE.charset[ i], 
					str, partialFlag);
			if ( pstr != NULL) {
			    return( str + 1);
			}
		    } 
		    return( NULL);
		}
	    }
	}
    case AmSTAR_REGEXP:

/* star regular expression matches zero or more occurences */

	if ( *str == '\0') {
	    return( str);
	}
	do {
	    pstr = REMatch( re->REVal.starRE.RE, str, partialFlag);
	    if ( pstr != NULL) {
		str = pstr;
	    } 
	} while (( pstr != NULL) && ( *pstr != '\0'));
	return( str);
    case AmOCCURRENCE_REGEXP:

/* make sure we're above the minimum */

	if (( *str == '\0') && ( re->REVal.occurRE.min == 0)) {
	    return( str);
	} else {
	    for ( i = 0; i < re->REVal.occurRE.min; i++) {
		pstr = REMatch( re->REVal.occurRE.RE, str, partialFlag);
		if ( pstr == NULL) {
		    return( NULL);
		} else if ( *pstr == '\0') {
		    return( partialFlag ? pstr : NULL);
		} else {
		    str = pstr;
		}
	    }

/* and in the range.  We don't fail if a character doesn't pass, because
   it might match following expression */

	    numOccur = re->REVal.occurRE.max - re->REVal.occurRE.min; 
	    for ( i = 0; i < numOccur; i++) {
		pstr = REMatch( re->REVal.occurRE.RE, str, partialFlag);
		if ( pstr == NULL) {
		    return( str);		
		} else if ( *pstr == '\0') {
		    return( pstr);
		} else {
		    str = pstr;
		}
	    }
	}

/* if we're beyond the range, then we just return here, and maybe the
   rest of the RE will pick it up */

	return( str);
    case AmREGEXP:

/* walk through and match each regular expression */

	for ( i = 0; i < re->REVal.fullRE.numSet; i++) {
	    pstr = REMatch( re->REVal.fullRE.REset[ i], str, partialFlag);
	    if ( pstr == NULL) {
		return( NULL);
	    } else if ( *pstr == '\0') {	/* we terminated early */
		if ( partialFlag) {
		    return( pstr);
		} else if (i == re->REVal.fullRE.numSet - 1) {
		    return( pstr);
		} 
	    }
	    str = pstr; 
	}
	return( str);
    default:
	fprintf( stderr, "bad tag\n");
	return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	match a regular expression, and support partial matching 
 * Arguments:
 *	re - regular expression to test against.
 *	str - string to test.
 *	partialFlag - flag for partial matches.
 * Returns:
 *	TRUE if the string matched, FALSE otherwise.
 *
 ****************************************************************************/

static AnvilBoolean 
regexpMatch( AmREPtr	re,
	     char	*str,
	     AnvilBoolean	partialFlag)
{
    char 	*p;

    p = REMatch( re, str, partialFlag);
    if ( p == NULL) {
       return( FALSE);
    } else if ( *p == '\0') {
       return( TRUE);
    } else {
       return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	comparator function for matching regular expression strings
 *	in the cache.
 * Arguments:
 *	a, b - regular expressions and their source strings to compare.
 * Returns:
 *	strcmp() of the regular expression strings.
 *
 ****************************************************************************/

static int 
matchRegexpString( AmREStringPtr	a,
		   AmREStringPtr	b)
{
    return( strcmp( a->regexpString, b->regexpString));
}

/****************************************************************************
 *
 * Description:
 * 	top-level regular expression matching routine.  First, find
 *	if the regular expression string is in the cache and match
 *	against the cached expression, otherwise, parse the regular
 *	expression string, add it to the cache, and test against it.
 *	in the cache.
 * Arguments:
 *	re - regular expression to test against.
 *	str - string to test.
 *	partialFlag - flag for partial matches.
 * Returns:
 *	TRUE if the string matched, FALSE otherwise.
 * Modifications:
 *	The ASSERTS have to go, there should be error return codes for
 *	failure to parse the expression, as well as the return flag.
 *
 ****************************************************************************/

AnvilBoolean 
AmRegexpMatch( char 	*regexpStr,
	     char	*str,
	     AnvilBoolean	partialFlag)
{
    AmREStringRec	target;
    AmREStringPtr	node, *nodePtr;

/* if the node has been found in the tree, then use the pre-compiled
   regular expression for matching */

    target.regexpString = regexpStr;
    nodePtr = ( AmREStringPtr *) tfind( &target, (const void **) &regexpRoot, 
				     ( AmSearchFn) matchRegexpString);
    if ( nodePtr == NULL) {
	statStrSize = strlen( regexpStr);
	ASSERT( (unsigned)statStrSize < sizeof( statBuffer) - 1);
	regexprestart( NULL);
	strcpy( statBuffer, regexpStr);
	regexpparse();
	ASSERT( statRE != NULL);

/* create a node in the tree, compile the regular expression, insert the
   node, and try to match the string */

	node = ( AmREStringPtr) malloc( sizeof( AmREStringRec));
	node->regexpString = strdup( regexpStr);
	node->regexp = statRE;
	tsearch( node, &regexpRoot, ( AmSearchFn) matchRegexpString);
    } else {
	node = *nodePtr;
    }
    return( regexpMatch( node->regexp, str, partialFlag));
}
 
/****************************************************************************
 *
 * Description:
 * 	input function for strings for the regular expression parser 
 *	called by the parser.
 * Arguments:
 *	buffer - buffer to read from.
 *	maxSize - size of the buffer.
 * Returns:
 *	number of bytes read.
 *
 ****************************************************************************/
 
int 
AmStrYYinput( char    *buffer,
	    int     maxSize)
{
    int		length;

    length = statStrSize;
    if ( length > maxSize) {
        length = maxSize;
    }
    memcpy( buffer, statBuffer, length);
    statStrSize -= length;
    return( length);
}
 
/****************************************************************************
 *
 * Description:
 * 	for testing, print a compiled regular expression 
 * Arguments:
 *	re - regular expression to test.
 *
 ****************************************************************************/

void 
AmPrintRE( AmREPtr	re)
{
    int	i;

    switch ( re->tag) {
    case AmOR_REGEXP:
	AmPrintRE( re->REVal.orRE.REleft);
	printf( " | ");
	AmPrintRE( re->REVal.orRE.REright);
	break;
    case AmCHAR_REGEXP:
	printf( "%c ", re->REVal.charRE.character);
	break;
    case AmRANGE_REGEXP:
	printf( "%c-%c ", 
		re->REVal.rangeRE.minChar,
		re->REVal.rangeRE.maxChar);
	break;
    case AmCHARSET_REGEXP:
	printf( "[");
	for ( i = 0; i < re->REVal.charsetRE.numSet; i++) {
	    AmPrintRE( re->REVal.charsetRE.charset[ i]);
	}
	printf( "]");
	break;
    case AmSTAR_REGEXP:
	AmPrintRE( re->REVal.starRE.RE);
	printf( "*");
	break;
    case AmOCCURRENCE_REGEXP:
	AmPrintRE( re->REVal.occurRE.RE);
	printf( "\\{%d,%d\\}", re->REVal.occurRE.min, 
			       re->REVal.occurRE.max);
	break;
    case AmREGEXP:
	for ( i = 0; i < re->REVal.fullRE.numSet; i++) {
	    AmPrintRE( re->REVal.fullRE.REset[ i]);
	}
	break;
    default:
	fprintf( stderr, "bad tag\n");
	break;
    }
}

/****************************************************************************
 *
 * Description:
 * 	regular expression error code.
 * Arguments:
 *	msg - error message string.
 * Modifications:
 *	should set an error code and error string.
 *
 ****************************************************************************/

int 
regexperror( const char *msg)
{
    return (0);
}
