#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <anvil/project.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/geometry.h>
#include <anvil/sample.h>

/* octant 1, x,y > 0, x > y */

static void lineOct1( uchar_t 		*start, 
		      int		size,
		      int		width, 
		      int		dx,
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrE, d, incrNE;

/* initialize decision values */

    incrE = dy << 1;
    d = incrE - dx;
    incrNE = (dy - dx) << 1;
    width += size;

/* walk based on x-line always increment x, sometimes increment y */

    while ( dx--) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrE;
	    start += size;
	} else {
	    d += incrNE;
	    start += width;			/* actually width + 1 */
	}
    }
}

/* octant 2, x,y > 0, x < y */

static void lineOct2( uchar_t 		*start, 
		      int		size,
		      int		width,
		      int		dx,
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrN, d, incrNE, w1 = width + size;

    incrN = dx << 1;
    d = incrN - dy;
    incrNE = (dx - dy) << 1;

/* y is greater than x, so walk based on y-line */
/* always increment y, sometimes increment x */

    while ( dy--) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrN;
	    start += width;
	} else {
	    d += incrNE;
	    start += w1;
	} 
    }
}

/* third octant, x < 0, y > 0, -x < y */

static void lineOct3( uchar_t 		*start, 
		      int		size,
		      int		width, 
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrN, incrNW, d, w1;

    incrN = (-dx) << 1;
    d = incrN - dy;
    incrNW = ((-dx) - dy) << 1;
    w1 = width - size;

/* y is greater than negative x, so walk based on y */
/* always increment y, sometimes decrement x */

    while ( dy--) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrN;
	    start += width;
	} else {
	    d += incrNW;
	    start += w1;
	} 
    }
}

/* fourth octant, x < 0, y > 0, -x > y */

static void lineOct4( uchar_t 		*start, 
		      int		size,
		      int		width,
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrW, d, incrNW;

AN_DEBUG( 10, ("in octant 4\n"));
    incrW = dy << 1;
    d = incrW - (-dx);
    incrNW = (dy - (-dx)) << 1;
    width -= size;

/* negative x is greater than y, so walk based on x. */
/* always decrement x, sometimes increment y */

    while ( dx++) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrW;
	    start -= size;
	} else {
	    d += incrNW;
	    start += width;
	}
    }
}

/* fifth octant, x < 0, y < 0, -x > -y */

static void lineOct5( uchar_t 		*start, 
		      int		size,
		      int		width,
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrW, d, incrSW;

    incrW = (-dy) << 1;
    d = incrW - (-dx);
    incrSW = ((-dy) - (-dx)) << 1;
    width += size;

/* negative x is greater than y, so walk based on x. */
/* always decrement x, sometimes decrement y */

    while ( dx++) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrW;
	    start -= size;
	} else {
	    d += incrSW;
	    start -= width;
	}
    }
}

/* sixth octant, x < 0, y < 0, -x < -y */

static void lineOct6( uchar_t 		*start, 
		      int		size,
		      int		width,
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrS, incrSW, d, w1 = width + size;

    incrS = (-dx) << 1;
    d = incrS - (-dy);
    incrSW = ((-dx) - (-dy)) << 1;

/* negative y is greater than negative x, so walk based on y */

    while ( dy++) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrS;
	    start -= width;
	} else {

/* always increment y, sometimes decrement x */

	    d += incrSW;
	    start -= w1;
	} 
    }
}

/* octant 7, x > 0 y < 0, x < -y */

static void lineOct7( uchar_t 		*start, 
		      int		size,
		      int		width,
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrS, d, incrSE, w1 = width - size;

    incrS = dx << 1;
    d = incrS - (-dy);
    incrSE = (dx - -(dy)) << 1;

/* y is greater than x, so walk based on y-line */

    while ( dy++) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrS;
	    start -= width;
	} else {
	    d += incrSE;
	    start -= w1;
	} 
    }
}

/* octant 8, x,y > 0, x > -y */

static void lineOct8( uchar_t 		*start, 
		      int		size,
		      int		width, 
		      int		dx, 
		      int		dy,
		      uchar_t		*buffer)
{
    int		incrE, d, incrSE, w1 = width - size;

/* initialize decision values */

    incrE = (-dy) << 1;
    d = incrE - dx;
    incrSE = ((-dy) - dx) << 1;
  
/* walk based on x-line */

    while ( dx--) {
	memcpy( buffer, start, size);
	buffer += size;
	if ( d <= 0) {
	    d += incrE;
	    start += size;
	} else {
	    d += incrSE;
	    start -= w1;
	}
    }
}

/* sample a matrix from a line */

void AmSampleMatrix( uchar_t	*matrix,
		   int		size,
		   int		width,
		   int		height,
		   int		x0,
		   int		y0,
		   int		x1,
		   int		y1,
		   uchar_t	*buffer)
{
    int	dx = x1 - x0, dy = y1 - y0;

    width *= size;
    matrix += y0*width + x0*size;
    if ( dy > 0) {
    	if ( dx > 0) {
    	    if ( dx > dy) {
    	    	lineOct1( matrix, size, width, dx, dy, buffer);
	    } else {
    	    	lineOct2( matrix, size, width, dx, dy, buffer);
	    }
	} else {	
	    if ( abs( dx) < dy) {
    	    	lineOct3( matrix, size, width, dx, dy, buffer);
	    } else {
    	    	lineOct4( matrix, size, width, dx, dy, buffer);
	    }
	}
    } else {
    	if ( dx < 0) {
	    if ( abs( dx) > abs( dy)) {
		lineOct5( matrix, size, width, dx, dy, buffer);
	    } else {
		lineOct6( matrix, size, width, dx, dy, buffer);
	    } 
	} else {
	    if ( abs( dx) < abs( dy)) {
		lineOct7( matrix, size, width, dx, dy, buffer);
	    } else {
		lineOct8( matrix, size, width, dx, dy, buffer);
	    } 
	}
    }
}

/* given a block of scatter points, find the points within a specified
   rectangle and return them in the buffer */

void AmSampleScatter( uchar_t 	*vector,
		    int		length,
		    AmAccessFn	xFn,
		    int		xOffset,
		    AmAccessFn	yFn,
		    int		yOffset,
		    void	*clientData,
		    int		size,
		    float	x0,
		    float	y0,
		    float	x1,
		    float	y1,
		    float	lineWidth,
		    uchar_t	*buffer,
		    int		*numPts)
{
    float	x, y, AmDist;
    uchar_t	*vecend;

    *numPts = 0;
    vecend = vector + length*size;
    for ( ; vector < vecend; vector += size) {
	x = ( xFn)( vector + xOffset, clientData);
	y = ( yFn)( vector + yOffset, clientData);

/* trivial reject AmFirst... */

	if (( x >= x0 - lineWidth) && ( x <= x1 + lineWidth) && 
	    ( y >= y0 - lineWidth) && ( y <= y1 + lineWidth)) {
	    AmDist = AmLineDist( x0, y0, x1, y1, x, y);
	    if ( AmDist < lineWidth) {
		memcpy( buffer, vector, size);
		buffer += size;
		( *numPts)++;
	    }
	}
    }
}
