#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* utility functions for dealing with tag-value pairs in the AmRecordClass.
   Modifications:
	the functions which manipulate the tag-values arrays should have
	more checks to ensure that the arrays aren't overflowed.
*/

#include <stdarg.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>

#include <anvil/project.h>
#include <anvil/tagstrs.h>
#include <anvil/strdefs.h>
#include <anvil/strlib.h>
#include <anvil/mem_funcs.h>
#include <anvil/files.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/envdefs.h>
#include <anvil/colorcell.h>
#include <anvil/resource.h>
#include <anvil/units.h>
#include <anvil/enums.h>
#include <anvil/tsearch.h>
#include <anvil/record.h>
#include <anvil/register.h>

/****************************************************************************
 * 
 * Description:
 *	tag match comparison function wrapper.  If either of the tags is 
 *	prefixed with a "prefix." expression, then compare the tag after
 *	 the '.'
 * Arguments:
 *	tag1, tag2: tags to compare 
 * Returns:
 *	TRUE if tags match, FALSE if they don't.
 *
 ****************************************************************************/

AnvilBoolean 
AmMatchTag( const char	*tag1,
	  const char	*tag2)
{   
    char *ptag1, *ptag2;

    if (( ptag1 = strchr( tag1, '.')) != NULL) {
	ptag1++;
    } else {
	ptag1 = ( char *) tag1;
    }
    if (( ptag2 = strchr( tag2, '.')) != NULL) {
	ptag2++;
    } else {
	ptag2 = ( char *) tag2;
    }
    return( !strcmp( ptag1, ptag2));
}

/****************************************************************************
 * 
 * Description:
 *	prefix match comparison function wrapper.  Determine if tag "tag"
 *	is prefixed with the prefix "prefix".  This is used when reading
 *	values from files, rather than directly, such as "file.colorsBinvec"
 * Arguments:
 *	prefix:	prefix to search for, such as "file.".  It's best to add the
 *	".", to prevent ambiguity with tags such as "filename".
 *	tag: tag to compare against.
 * Returns:
 *	TRUE if tag is prefixed with "prefix", FALSE otherwise.
 *
 ***************************************************************************/

AnvilBoolean 
AmPrefixTag( const char	*prefix,
	   const char	*tag)
{
    return( !strncmp( prefix, tag, strlen( prefix)));
}

/****************************************************************************
 * 
 * Description:
 *	remove a tag value pair from an array of tag values if the tag 
 *	matches, scrunch the array, and decrement the passed number of
 *	tag values.
 * Arguments:
 *	tag - tag to search for.
 *	tagValues - vector of tag values to search.
 *	numValues - length of the vector, decremented if the value is
 *	found and removed.
 * Returns:
 *	TRUE if the tag is removed, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmRemoveTagValue( char		*tag,
		  AmTagValuePtr	tagValues,
		  int		*numValues)
{
    int		ii, count;
    AnvilBoolean	foundFlag;

    foundFlag = FALSE;
    count = 0;
    for ( ii = 0; ii < *numValues; ii++) {
	if ( !AmMatchTag( tagValues[ ii].tag, tag)) {
	    tagValues[ count++] = tagValues[ ii];
	} else {
	    AmFreeNonAtomicValues( &tagValues[ ii], 1);
	    foundFlag = TRUE;
	}
    }
    *numValues = count;
    return( foundFlag);
}

static void *strdefRoot = NULL;

/* comparison function for searching the table of tags and associated types. */
 
static int 
compareTag( AmStrdefPtr	a,
	    AmStrdefPtr   b)
{
    return( strcmp( a->tag, b->tag));
}

/***************************************************************************
 *
 * Description:
 *	add a tag and associated data type to the table of tags and 
 *	associated types.  If the tag already exists in the table, the
 *	new tag and type will replace it.
 * Arguments:
 *	tag: tag to add to the table of tags and associated types.
 *	dataType: data type to associate with the tag.
 * Returns:
 *	AmMALLOC_FAILED if memory couldn't be allocated.
 *	AmOK otherwise.
 *
 ***************************************************************************/

AmErrorCode 
AmAddTagType( const char 	*tag, 
	      AmDataTypeEnum	dataType)
{
    AmStrdefPtr	strdef, *resdef;

    if (( strdef = ( AmStrdefPtr) malloc( sizeof( AmStrdefRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* insert the string definition into the search tree */

    strdef->tag = strdup( tag);
    strdef->dataType = dataType;
    resdef = ( AmStrdefPtr *) tsearch( strdef, &strdefRoot, 
				     ( AmSearchFn) compareTag);
    if ( resdef != NULL) {
	if ( *resdef != strdef) {
	    if ( dataType != ( *resdef)->dataType) {
		AN_DEBUG( 1, ( "data type %d for tag %s does not match\n",
			       dataType, tag));
	    }
	    free( strdef->tag);
	    free( strdef);
	}
    }
    return( AmOK);
}

/***************************************************************************
 *
 * Description:
 *	Take the table from <anvil/strdefs.h> and add each entry to
 *	the tag type lookup table.
 * Returns:
 *	AmMALLOC_FAILED: if the function failed to allocate memory.
 *	AmOK if everything is OK.
 *
 ***************************************************************************/

AmErrorCode 
AmCreateLookupTable()
{
    AmErrorCode	errCode;
    unsigned int ii;

    for ( ii = 0; ii < VECLEN( strLookup); ii++) {
	errCode = AmAddTagType( strLookup[ ii].tag, strLookup[ ii].dataType);
	if ( errCode != AmOK) {
	    return( errCode);
	}
    }
    return( AmOK);
}

/***************************************************************************
 *
 * Description:
 * 	get the string definition for the specified tag, parsing tags with 
 *  	"." prefix expressions.
 * Arguments:
 *	tag - tag to search for in the string definitions table.
 * Returns:
 *	the tag-type definition, or NULL if the tag wasn't found.
 *
 ***************************************************************************/

AmStrdefPtr 
AmGetStrdef( const char	*tag)
{
    const char	*ptag;
    AmStrdefPtr	*result;
    AmStrdefRec	target;

/* check to see if the tag is a ".", expression, and if so, then use the
   string after the "." as the tag for the type lookup */

    if (( ptag = strchr( tag, '.')) != NULL) { 
	ptag++;
    } else {
	ptag = tag;
    }
    target.tag = ( char *) ptag;
    result = ( AmStrdefPtr *)
	tfind( &target, ( const void **) &strdefRoot, 
	       ( AmSearchFn) compareTag);
    if ( result == NULL) {
	return( NULL);
    } else {
	return( *result);
    }
}

/***************************************************************************
 *
 * Description:
 * 	get the type associated with the specified tag.
 * Arguments:
 *	tag - tag to search for in the string definitions table.
 * Returns:
 *	the tag's type, or AmUNKNOWN if it wasn't found. 
 *
 ***************************************************************************/

AmDataTypeEnum 
AmGetTagType( const char *tag)
{
    AmStrdefPtr	strdef;

    strdef = AmGetStrdef( tag);
    if ( strdef == NULL) {
	return( AmUNKNOWN);
    } else {
	return( strdef->dataType);
    }
}

/***************************************************************************
 *
 * Description:
 * 	get the string definition for the specified tag, NOT parsing tags 
 *	with prefixed "." expressions, except for expressions that start 
 *	with "file." 
 * Arguments:
 *	tag - tag to search for in the string definitions table.
 * Returns:
 *	the tag-type definition, or NULL if the tag wasn't found.
 *
 ***************************************************************************/

AmStrdefPtr 
AmGetLiteralStrdef( const char *tag)
{
    const char		*ptag;

    if ( !strncmp( tag, FILE_PREFIX, 5)) {
	ptag = &tag[ 5];
    } else {
	ptag = tag;
    }
    return( AmGetStrdef( ptag));
}

/***************************************************************************
 *
 * Description:
 * 	after getting a vector of tagValues, change them into an array 
 *  	suitable for setting the values, and free the allocated atomic 
 *	values before setting them...  The vector values (allocated values) 
 *	have to be freed after they are set in the object.   This is used
 *	when copying objects.
 * Arguments:
 *	tagValues - tag-values array to change over.
 *	numValues - length of the tag-values array.
 *
 ***************************************************************************/

void 
AmChangeGetToSet( AmTagValuePtr	tagValues,
		  int		numValues)
{
    int			ii, flag;
    double		*val;
    short		shortVal;
    ushort_t		ushortVal;
    char		charVal;
    uchar_t		ucharVal;
    void		*ptr;
    AmDataTypeEnum	dataType;

    for ( ii = 0; ii < numValues; ii++) {
	dataType = AmGetTagType( tagValues[ ii].tag);
	if ( dataType == AmDOUBLE) {
	    val = ( double *) malloc( sizeof( double));
	    *val = *( float *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = val;
	    tagValues[ ii].copied = TRUE;
	} else if ( dataType == AmFLOAT) {
	    float	floatVal;

	    floatVal = *( float *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    memcpy( &tagValues[ ii].value, ( float *) &floatVal, 
		    sizeof( float));
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmBOOLEAN) {
	    flag = *( AnvilBoolean *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ( void *) flag;
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmSHORT) {
	    shortVal = *( short *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ( void *) shortVal;
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmUNSIGNED_SHORT) {
	    ushortVal = *( ushort_t *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ( void *) ushortVal;
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmCHAR) {
	    charVal = *( char *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ( void *) charVal;
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmUNSIGNED_CHAR) {
	    ucharVal = *( uchar_t *) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ( void *) ucharVal;
	    tagValues[ ii].copied = FALSE;
	} else if ( dataType == AmCOLOR_CELL) {

/* color cells stay the same */

	} else if (( dataType == AmSTRING) || ( dataType == AmFILE)) {
	    ptr = *( void **) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ptr;
	    tagValues[ ii].copied = TRUE;
	} else if ( AmIsAtomic( dataType)) {

/* all the rest of the atomic types fit in 4-byte integers.  
   Free the old pointer before setting the new. */

	    ptr = *( void **) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ptr;
	    tagValues[ ii].copied = FALSE;
	} else {

/* the others are non-atomic values */

	    ptr = *( void **) tagValues[ ii].value;
	    free( tagValues[ ii].value);
	    tagValues[ ii].value = ptr;
	}
    }
}

/***************************************************************************
 *
 * Description:
 * 	take an array of strings which represent tags and values and parse
 *  	them into a tag value array to be read into the plot structures
 *  	for fields of string, double, or vector datatype, allocate them.
 * Arguments:
 *	table - NULL-terminated vector of tag-value pairs represented 
 *	as strings.
 *	tagValues - resulting tag-values array.
 * Returns:
 *	The number of tags in the resulting tagValues array.
 *
 ***************************************************************************/
 
int 
AmHeaderStringsTagValueArray( char	   **table,
			    AmTagValuePtr    tagValues)
{
    int        	ii, length, count;
    AmErrorCode	errCode;
 
    length = AmVeclen(( char const **) table);

/* scan each of the strings in the table, and if it's not valid, then don't
   copy it to the tagValue array */

    count = 0;
    for ( ii = 0; ii < length; ii += 2) {
	errCode = AmScanValue( table[ ii], table[ ii + 1],
			       &tagValues->value, &tagValues->copied);
	if ( errCode != AmBAD_TYPE) {
            tagValues->tag = strdup( table[ ii]);
	    tagValues->tagCopied = TRUE;
            tagValues++;
            count++;
        } else {

/* if the value isn't scanned, then assume a string and take it literally */

            tagValues->tag = strdup( table[ ii]);
	    tagValues->tagCopied = TRUE;
	    tagValues->value = strdup( table[ ii + 1]);
	    tagValues->copied = TRUE;
            tagValues++;
            count++;
	}
    }
    return( count);
}

/***************************************************************************
 *
 * Description:
 * 	take one of the resource files in tag value format and parse it 
 *	into a tag value array to be read into the plot structures for 
 *	fields of string, double, or vector datatype, allocate them 
 * Arguments:
 *	file - file containing the tags and values in tag: value format.
 *	tagValues - resulting tag-values array.
 * Returns:
 *	The number of tags in the resulting tagValues array.
 *
 ***************************************************************************/
 
int 
AmHeaderTagValueArray( char           *file,
		     AmTagValuePtr    tagValues)
{
    char       **lines;
    int        count;
 
/* read the header and find out how long it is */
 
    lines = AmReadHeader( file);
    if ( lines == 0) {
	DO_ERROR( AmOPEN_FAILED);
        return( 0);
    }
    count = AmHeaderStringsTagValueArray( lines, tagValues);
    AmFreeStrVec( lines);
    return( count);
}

/***************************************************************************
 *
 * Description:
 * 	search for any occurrences of AmNheader and AmNheaderStrings in
 *	a tag-values array and parse and expand them in place in the
 *	tag-values array.
 * Arguments:
 *	tagValues - tag-values array to expand.
 *	numValues - initially the number of values, returns the number
 *	of values in the expanded array.
 * Returns:
 *	AmOPEN_FAILED if any of the header files couldn't be read,
 *	AmOK on success.
 *
 ***************************************************************************/

static AmErrorCode 
expandHeaders( AmTagValuePtr	tagValues,
	       int		*numValues)
{
    AmTagValueRec	headerValues[ MAXTAGS], newValues[ MAXTAGS];
    int		ii, headerCount, numTagValues = 0, count;
    AnvilBoolean	headerFound;

    headerFound = FALSE;
    count = 0;
    for ( ii = 0; ii < *numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNheader)) {

/* read the header and insert the values in place */

	    headerCount = AmHeaderTagValueArray(( char *) tagValues[ ii].value,
					      headerValues);
	    if ( headerCount == 0) {
		DO_RETURN_ERROR( AmOPEN_FAILED);
	    }
	    if ( tagValues[ ii].tagCopied) {
		free( tagValues[ ii].tag);
	    }
	    if ( tagValues[ ii].copied) {
		free( tagValues[ ii].value);
	    }
	    memcpy( &newValues[ count], headerValues,
		    headerCount*sizeof( AmTagValueRec));
	    count += headerCount;
	    headerFound = TRUE;
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNheaderStrings)) {

/* read the vector of strings and insert the values in place */

	    headerCount = 
		AmHeaderStringsTagValueArray(( char **) tagValues[ ii].value,
					   headerValues);
	    memcpy( &newValues[ count], headerValues,
		    headerCount*sizeof( AmTagValueRec));
	    if ( tagValues[ ii].tagCopied) {
		free( tagValues[ ii].tag);
	    }
	    if ( tagValues[ ii].copied) {
		AmFreeStrVec(( char **) tagValues[ ii].value);
	    }
	    count += headerCount;
	    headerFound = TRUE;
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNnumTagValues)) {

/* save the number of tag values for when the tag values are read */

	    numTagValues = *( int *) &tagValues[ ii].value;
	    if ( tagValues[ ii].tagCopied) {
		free( tagValues[ ii].tag);
	    }
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNtagValues)) {

/* take the tag values and insert them in place */

	    memcpy( &newValues[ count], tagValues[ ii].value,
		    numTagValues*sizeof( AmTagValueRec));
	    if ( tagValues[ ii].tagCopied) {
		free( tagValues[ ii].tag);
	    }
	    count += numTagValues;
	    headerFound = TRUE;
	} else {
	    newValues[ count++] = tagValues[ ii];
	}
    }
    memcpy( tagValues, newValues, count*sizeof( AmTagValueRec));
    *numValues = count;
    if ( headerFound) {
	return( expandHeaders( tagValues, numValues));
    } else {
	return( AmOK);
    }
}
 
/***************************************************************************
 *
 * Description:
 * 	normal AmMakeTagValueArrayVararg, takes a NULL-terminated varargs list 
 *	of parameters, and generates a tag-values array.
 * Arguments:
 *	tagValues - resulting tag-values array.
 *	va_list args - NULL-terminated varargs list of tag-value pairs.
 * Returns:
 *	The number of tags in the resulting tag-values array.
 *
 ***************************************************************************/

int 
AmMakeTagValueArrayVararg( AmTagValuePtr	tagValues,
			 va_list	args)
{
    char *arg;

    arg = va_arg( args, char *);
    return( AmMakeTagValueArrayVararg( tagValues, arg, args));
    va_end( args);
}

/***************************************************************************
 *
 * Description:
 * 	take a varargs list of tags and values, and turn them into a TagValue
 *  	array.  This is for set fields, where the AmFirst tag is a positional
 *  	parameter. 
 * Arguments:
 *	tagValues - resulting tag-values array.
 *	tag - the AmFirst tag.
 *	va_list args - NULL-terminated varargs list of tag-value pairs.
 * Returns:
 *	The number of tags in the resulting tag-values array.
 * Modifications:
 *	we may have the varargs list start on tagValues, rather than
 *	tag, and remove the second parameter.
 *
 ***************************************************************************/

int 
AmMakeTagValueArrayVararg( AmTagValuePtr   	tagValues,
			 const char	*tag,
			 va_list       	args)
{
    int                 count;
    AmDataTypeEnum	dataType;
    AmTagValuePtr		tagValuesPtr = tagValues;
 
/* retreive the lookup table for the structure ( which has to be rooted off
   of the record class) */

    count = 0;
    for ( tagValues->tag = ( char *) tag;
          tagValues->tag != NULL;
          tagValues->tag = va_arg( args, char *)) {
	if (( dataType = AmGetTagType( tagValues->tag)) == AmUNKNOWN) {
	    dataType = AmSTRING;
	}

/* default is to not copy data */

	tagValues->copied = FALSE;
	tagValues->tagCopied = FALSE;
	switch ( dataType) {
	case AmBOOLEAN:
	case AmINT:
	case AmENUM:
	case AmUNITS:
	    tagValues->value = ( void *) va_arg( args, int);
	    break;
	case AmDOUBLE:

/* since doubles are 8 bytes, we have to have more than 4 bytes for them, 
   so we allocate and copy */

	    tagValues->value = ( void *) malloc( sizeof( double));
	    tagValues->copied = TRUE;
	    *( double *) tagValues->value = va_arg( args, double);
	    break;
	case AmFLOAT:

/* float values passed as varargs get promoted to doubles, but we'd
   rather deal with them as real floats */

	    float	floatVal;

	    floatVal = va_arg( args, double);
	    memcpy( &tagValues->value, &floatVal, sizeof( float));
	    break;
	case AmCHAR:
	    tagValues->value = ( void *) va_arg( args, char);
	    break;
	case AmUNSIGNED_CHAR:
	    tagValues->value = ( void *) va_arg( args, unsigned char);
	    break;
	case AmSHORT:
	    tagValues->value = ( void *) va_arg( args, short);
	    break;
	case AmUNSIGNED_SHORT:
	    tagValues->value = ( void *) va_arg( args, unsigned short);
	    break;
	case AmLONG:
	    tagValues->value = ( void *) va_arg( args, long);
	    break;
	case AmUNSIGNED_LONG:
	    tagValues->value = ( void *) va_arg( args, unsigned long);
	    break;
	case AmUNSIGNED_INT:
	    tagValues->value = ( void *) va_arg( args, unsigned int);
	    break;
	case AmFILE:
	case AmSTRING:
	    tagValues->value = ( void *) va_arg( args, char *);
	    break;
	case AmSTRING_PTR:
	    tagValues->value = ( void *) va_arg( args, char **);
	    break;
	case AmINT_PTR:
	    tagValues->value = ( void *) va_arg( args, int *);
	    break;
	case AmFLOAT_PTR:
	    tagValues->value = ( void *) va_arg( args, float *);
	    break;
	case AmFUNC_PTR:
	case AmVOID_PTR:
	case AmOBJECT:
	    tagValues->value = ( void *) va_arg( args, void *);
	    break;
	case AmCOLOR_CELL:
	case AmCOLOR_CELL_PTR:
	    tagValues->value = ( void *) va_arg( args, AmColorCellPtr);
	    break;
	default:
	    break;
	}
	tagValues++;
	count++;
    }
    expandHeaders( tagValuesPtr, &count);
    ASSERT( count <= MAXTAGS);
    return( count);
}

/***************************************************************************
 *
 * Description:
 * 	predicate function for dataTypes, returns whether they are 
 *	atomic or not.
 *
 ***************************************************************************/

AnvilBoolean 
AmIsAtomic( AmDataTypeEnum	dataType)
{
    return(( dataType == AmFLOAT) ||
	   ( dataType == AmCHAR) ||
	   ( dataType == AmUNSIGNED_CHAR) ||
	   ( dataType == AmINT) ||
	   ( dataType == AmUNSIGNED_INT) ||
	   ( dataType == AmSHORT) ||
	   ( dataType == AmUNSIGNED_SHORT) ||
	   ( dataType == AmLONG) ||
	   ( dataType == AmUNSIGNED_LONG) ||
	   ( dataType == AmENUM) ||
	   ( dataType == AmDOUBLE) ||
	   ( dataType == AmSTRING) ||
	   ( dataType == AmFILE) ||
	   ( dataType == AmFUNC_PTR) ||
	   ( dataType == AmBOOLEAN) ||
	   ( dataType == AmVOID_PTR) || 
	   ( dataType == AmOBJECT) ||
	   ( dataType == AmUNITS) ||
	   ( dataType == AmCOLOR_CELL));
}

/***************************************************************************
 *
 * Description:
 *	utility function for setting a bin vector, where it sets the vector
 *	to 1 less than the specified length.  If the copied flag is set
 *	in the tag value pair, the value is freed after copying.  If the
 *	copiedTag field is set, then the tag is freed.
 * Arguments:
 *	binvec - pointer to the resulting vector of floats (allocated)
 *	length - length (-1) to set the vector.
 *	tagValue - tag-value pair to set the vector from.
 *
 ***************************************************************************/
 
void 
AmSetBinvec( float           **binvec,
	   int             length,
	   AmTagValuePtr     tagValue)
{
    if ( *binvec != NULL) {
        FREE( *binvec);
    }
    *binvec = ( float *) AmCopyVec( tagValue->value,
                                  sizeof( float)*( length - 1));
    if ( tagValue->copied) {
        FREE( tagValue->value);
    }
    if ( tagValue->tagCopied) {
	FREE( tagValue->tag);
    }
}

/***************************************************************************
 *
 * Description:
 * 	Given a vector of tag-value pairs, return the value associated with
 *  	the tag.  Go through the list backwards to make sure that the AmLast
 *  	value is gotten and not the AmFirst (so that one construct a list of
 *  	tag values and over-write previous values in the list). 
 * Arguments:
 *	tagValues - tag value array to scan.
 *	numValues - length of the tag-values array.
 *	tag - tag to scan for
 *	value - pointer to the returned value. THIS VALUE IS NOT COPIED
 * Returns:
 *	TRUE if the tag was found, FALSE otherwise.
 *
 ***************************************************************************/

AnvilBoolean 
AmGetTagValueFromTags( AmTagValuePtr	tagValues,
		     int		numValues,
		     const char		*tag,
		     void		**value)
{
    int	ii;

    for ( ii = numValues - 1; ii >= 0; ii--) {
	if ( AmMatchTag( tag, tagValues[ ii].tag)) {
	    *value = tagValues[ ii].value;
	    return( TRUE);
	}
    }
    return( FALSE);
}
