#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */





#if !defined(__BSD44__) 
#include <search.h>
#endif
#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <anvil/register.h>
#include <anvil/envdefs.h>
#include <anvil/regexp.h>
#include <anvil/expr.h>
#include <anvil/projects.h>
#include <anvil/enums.h>
#include <anvil/tsearch.h>
#include <anvil/units.h>
#include <anvil/unitsP.h>

/* functions for parsing the units file, which contains unit names, unit
   aliases, printing precisions, printing string names, registered custom
   printing functions, registered custom scan functions, regular expressions
   for entry validation, and unit conversions.

   unit name:		string
   unit alias:		<unit name>:		<unit alias>
   printing precision:	<unit name>.precision:	<integer precision>
   printing string:	<unit name>.string:	<string>
   registered print:	<unit name>.print:	<registered function name>
   registered scan:	<unit name>.scan:	<registered function name>
   regular expression:	<unit name>.regexp:	"regular expression"
   unit converter:	<unit name>-><unit name>:	arithmetic expression

   comments are preceded with a '#'

   There are references in this file for parsing a variables file, but this
   will likely be removed in a future release.
   These functions are used in combination with the lexer units.l and
   parser units.y
*/


/* root trees for finding unit codes from unit names, and unit names from
   unit codes. Unit names are added to the unit name and code trees, and
   aliases are only added to the unit name tree */

static void *unitNameRoot = NULL,
	    *unitCodeRoot = NULL,
	    *unitConvertRoot = NULL,
	    *varNameRoot = NULL,
	    *varCodeRoot = NULL;

const unitType		AmUNIT_UNDEFINED = 0,
			AmVARIABLE_UNDEFINED = 0;
static unitType 	UNIT_CODE_COUNT = AmUNIT_UNDEFINED + 1;
static unitType		VAR_CODE_COUNT = AmVARIABLE_UNDEFINED + 1;
unitType		AmUNIT_METER, AmUNIT_LONGITUDE_ARCDEG, 
			AmUNIT_LATITUDE_ARCDEG, AmUNIT_ARCDEG;
static FILE	*unitsFp;
int		AmUnitsErrorLineNo;
static AnvilBoolean	errorFlag, unitInitFlag = FALSE;
char		AmUnitsErrorString[ 512];

/* top level functions for the YACC parser. */

int unitsparse( void);
void unitsrestart( FILE *);

/****************************************************************************
 *
 * Description:
 *      comparator function for unit names.
 * Arguments:
 *      a, b - units to compare.
 * Returns:
 *      strcmp() of the unit names.
 *
 ****************************************************************************/

static int 
compareUnitName( unitNamePtr	a,
		 unitNamePtr	b)
{
    return( strcasecmp( a->name, b->name));
}
/****************************************************************************
 *
 * Description:
 *      compare unit enumeration codes.
 * Arguments:
 *      a, b - units to compare.
 * Returns:
 *      difference of their codes.
 *
 ****************************************************************************/

static int 
compareUnitCode( unitNamePtr a,
		 unitNamePtr b)
{
    return( a->u.unitDef.code - b->u.unitDef.code);
}

/****************************************************************************
 *
 * Description:
 *      comparator function for variable names.
 * Arguments:
 *      a, b - units to compare.
 * Returns:
 *      strcmp() of the unit names.
 *
 ****************************************************************************/

static int 
compareVarName( AmVarDefPtr    a,
		AmVarDefPtr    b)
{
    return( strcasecmp( a->variableName, b->variableName));
}

/****************************************************************************
 *
 * Description:
 *      comparator function for variable codes.
 * Arguments:
 *      a, b - units to compare.
 * Returns:
 *      difference of the variable codes.
 *
 ****************************************************************************/

static int 
compareVarCode( AmVarDefPtr    a,
	        AmVarDefPtr    b)
{
    return( a->variableCode - b->variableCode);
}

/****************************************************************************
 *
 * Description:
 *      comparator function for unit converters.
 * Arguments:
 *      a, b - unit converters to compare.
 * Returns:
 *      difference of the source and destination units, with source units
 *	taking precedence.
 *
 ****************************************************************************/

static int 
compareConvert( AmUnitConvertPtr	a,
		AmUnitConvertPtr	b)
{
    int val;

    val = a->sourceUnits - b->sourceUnits;
    if ( val == 0) {
	return( a->destUnits - b->destUnits);
    } else {
	return( val);
    }
}

/****************************************************************************
 *
 * Description:
 *      add a unit name to the unit name table, a tsearch() tree, sorted
 *	on the unit name.
 * Arguments:
 *      unitName - unit name structure to add.
 *
 ****************************************************************************/

static void 
addToUnitNameTable( unitNamePtr	unitName)
{
    tsearch( unitName, &unitNameRoot, ( AmSearchFn) compareUnitName);
}

/****************************************************************************
 *
 * Description:
 *      add a unit name to the unit code table, a tsearch() tree, sorted
 *	on the unit enumeration code.
 * Arguments:
 *      unitName - unit name structure to add.
 *
 ****************************************************************************/

static void 
addToUnitCodeTable( unitNamePtr	unitName)
{
    tsearch( unitName, &unitCodeRoot, ( AmSearchFn) compareUnitCode);
}

/****************************************************************************
 *
 * Description:
 *      add a unit converter to the unit converter table.
 * Arguments:
 *      unitConvert - unit converter structure to add.
 *
 ****************************************************************************/

static void 
addToUnitConvertTable( AmUnitConvertPtr	unitConvert)
{
    tsearch( unitConvert, &unitConvertRoot, ( AmSearchFn) compareConvert);
}

/****************************************************************************
 *
 * Description:
 *      add a variable name structure to the variable table, sorted on
 *	variable name. 
 * Arguments:
 *      varDef - variable name structure to add.
 *
 ****************************************************************************/

static void 
addToVarNameTable( AmVarDefPtr        varDef)
{
    tsearch( varDef, &varNameRoot, ( AmSearchFn) compareVarName);
}

/****************************************************************************
 *
 * Description:
 *      add a variable name structure to the variable table, sorted on
 *	variable code. 
 * Arguments:
 *      varDef - variable name structure to add.
 *
 ****************************************************************************/

static void 
addToVarCodeTable( AmVarDefPtr        varDef)
{
    tsearch( varDef, &varCodeRoot, ( AmSearchFn) compareVarCode);
}

/****************************************************************************
 *
 * Description:
 * 	get the unit converter from a source to a destination unit.  There's 
 *  	a one-element caching to improve performance for large datasets
 * Arguments:
 *      sourceUnits - units to convert from.
 *	destUnits - units to convert to.
 * Returns:
 *	The unit conversion structure from sourceUnits to destUnits, or
 *	NULL if it wasn't found.
 *
 ****************************************************************************/

static AmUnitConvertPtr 
getUnitConverter( unitType    sourceUnits,
		  unitType    destUnits)
{
    AmUnitConvertRec	    target;
    AmUnitConvertPtr	    *ref;
    static unitType	    lastSourceUnits = AmUNIT_UNDEFINED, 
			    lastDestUnits = AmUNIT_UNDEFINED;
    static AmUnitConvertPtr   lastConverter = NULL;

    if (( sourceUnits == lastSourceUnits) && ( destUnits == lastDestUnits)) {
	return( lastConverter);
    }

    target.sourceUnits = sourceUnits;
    target.destUnits = destUnits;
    ref = ( AmUnitConvertPtr *) tfind( &target, 
				     ( const void **) &unitConvertRoot, 
				     ( AmSearchFn) compareConvert);
    if ( ref != NULL) {
	lastSourceUnits = sourceUnits;
	lastDestUnits = destUnits;
	lastConverter = *ref;
	return( *ref);
    } else {
	return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	return a unit structure based on the unit name 
 * Arguments:
 *      name - name of the units tosearch for.
 * Returns:
 *	The unit name structure for the name, or NULL if it wasn't found.
 *
 ****************************************************************************/

static unitNamePtr 
findUnitByName( const char	*name)
{
    unitNamePtr	*ref;
    unitNameRec	target;
    char	namebuf[ 256];

    strcpy( namebuf, name);
    target.name = namebuf;
    ref = ( unitNamePtr *) tfind( &target, 
				  ( const void **) &unitNameRoot, 
				  ( AmSearchFn) compareUnitName);

/* if it's an alias, return the name that was aliased */

    if ( ref != NULL) {
	if (( *ref)->kind == AmUNIT_ALIAS) {
	    return(( *ref)->u.alias);
	} else {
	    return( *ref);
	}
    } else {
	return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	return a unit structure based on the unit code
 * Arguments:
 *      code - code of the units to search for.
 * Returns:
 *	The unit name structure for the code, or NULL if it wasn't found.
 *
 ****************************************************************************/

static unitNamePtr 
findUnitByCode( unitType code)
{
    unitNamePtr	*ref;
    unitNameRec	target;

    target.u.unitDef.code = code;
    ref = ( unitNamePtr *) tfind( &target, 
				  ( const void **) &unitCodeRoot, 
			          ( AmSearchFn) compareUnitCode);

/* the unit code table has no aliases */

    if ( ref != NULL) {
	return( *ref);
    } else {
	return( NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a unit to the unit name database. The name is put in a structure
 *	with the unit code uniquely defined by the static counter
 *	UNIT_CODE_COUNT.
 * Arguments:
 *      name - name of the units to add.
 * Modifications:
 *	should return an error code if the structure couldn't be allocated.
 *	should take a const char *
 *
 ****************************************************************************/

void AmAddUnitName( char *name)
{
    unitNamePtr	unitName;

    unitName = ( unitNamePtr) malloc( sizeof( unitNameRec));
    if ( unitName == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return;
    }
    unitName->kind = AmUNIT_NAME;
    unitName->name = name;
    unitName->u.unitDef.code = UNIT_CODE_COUNT++;
    unitName->u.unitDef.precision = 0;
    unitName->u.unitDef.printFunc = NULL;
    unitName->u.unitDef.scanFunc = NULL;
    unitName->u.unitDef.regexp = strdup( RE_REAL_FMT_UNITS);
    unitName->u.unitDef.string = NULL;
    addToUnitNameTable( unitName);
    addToUnitCodeTable( unitName);
}

/****************************************************************************
 *
 * Description:
 * 	add a unit alias to the unit database. 
 * Arguments:
 *      alias - name of the unit alias.
 *	name - name the alias maps to.
 * Modifications:
 *	should return an error code if the structure couldn't be allocated.
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddUnitAlias( char	*alias,
	      char	*name)
{
    unitNamePtr	ref, unitName;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	unitName = ( unitNamePtr) malloc( sizeof( unitNameRec));
	unitName->kind = AmUNIT_ALIAS;
	unitName->name = alias;
	unitName->u.alias = ref;
	addToUnitNameTable( unitName);
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a printing precision for a unit to the unit database 
 * Arguments:
 *	name - name to add the printing precision for
 * Modifications:
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddPrecision( char	*name,
	      int	precision)
{
    unitNamePtr	ref;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	ref->u.unitDef.precision = precision;
	ref = findUnitByCode( ref->u.unitDef.code);
	if ( ref != NULL) {
	    ref->u.unitDef.precision = precision;
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a registered print function to the unit database 
 * Arguments:
 *	name - name to add the registered printing function for.
 *	printFnName - printing function added with AmAddRegister().
 * Modifications:
 *	should return an error code if the unit name wasn't found, or
 *	if the registered printing function wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddPrint( char *name,
          char *printFnName)
{
    unitNamePtr	ref;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	ref->u.unitDef.printFunc = ( AmPrintFn) AmGetRegister( printFnName);
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a print string to the unit database 
 * Arguments:
 *	name - name to add the printing precision for
 *	string - unit string to print values with.
 * Modifications:
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void AmAddString( char	*name,
		char	*string)
{
    unitNamePtr	ref;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	ref->u.unitDef.string = string;
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a NULL print string to the unit database, so the values are
 *	printed with no unit labels.
 * Arguments:
 *	name - name to set the empty unit printing string for.
 * Modifications:
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddString( char	*name)
{
    unitNamePtr	ref;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	ref->u.unitDef.string = strdup( "");
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a registered scan function to the unit database
 * Arguments:
 *	name - name to set the registered scanning function for.
 *	scanFnName - scanning function name to set.
 * Modifications:
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddScan( char *name,
	 char *scanFnName)
{
    unitNamePtr	ref;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	ref->u.unitDef.scanFunc = ( AmScanFn) AmGetRegister( scanFnName);
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a regular expression for the unit, so the unit will be checked
 *	with this regular expression when it's scanned.
 * Arguments:
 *	name - name to set the regular expression for.
 *	regexp - regular expression to set.
 * Modifications:
 *	should return an error code if the unit name wasn't found.
 *	should take a const char *
 *
 ****************************************************************************/

void 
AmAddRegexp( char *name,
	   char *regexp)
{
    unitNamePtr	ref;
    char	*q, *p;

    ref = findUnitByName( name);
    if ( ref != NULL) {
	if ( ref->u.unitDef.regexp != NULL) {
	    free( ref->u.unitDef.regexp);
	}

/* for some reason, the lexer leaves a \n at the end of the expression, so
   we have to get rid of it */

	p = regexp;
	q = regexp;

/* remove the quotes */

	while ( *p != '\0') {
	    if (( *p != '"') && ( *p != '\n')) {
		*q++ = *p;
	    } 
	    p++;
	}
	*q = '\0';
AN_DEBUG( 10, ("adding regexp %s\n", regexp));
	ref->u.unitDef.regexp = strdup( regexp);
    }
}

/****************************************************************************
 *
 * Description:
 * 	get the regular expression for a unit 
 * Arguments:
 *	unitCode - unit enumeration code to search on
 *	unitRegexp - pointer to the returned regular expression, or
 *	pointer NULL if it wasn't found.
 * Returns:
 *	TRUE if the regexp and unit was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetRegexp( unitType	unitCode,
	   char		**unitRegexp)
{
    unitNamePtr	ref;

    ref = findUnitByCode( unitCode);
    if ( ref != NULL) {
	*unitRegexp = ref->u.unitDef.regexp;
	return( TRUE);
    } else {
	*unitRegexp = NULL;
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	add a unit converter to the unit conversion database 
 * Arguments:
 *	sourceName - source unit name to convert from
 *	destName - destination unit name to convert to
 *	expr - pointer to an expression structure to use for the conversion
 *	(see exprparse.cpp)
 * Modifications:
 *	should return an error code if either of the units weren't found.
 *
 ****************************************************************************/

void 
AmAddConvert( char	*sourceName,
	    char	*destName,
	    AmExprPtr	expr)
{
    AmUnitConvertPtr	unitConvert;
    unitNamePtr		sourceRef, destRef;

    sourceRef = findUnitByName( sourceName);
    destRef = findUnitByName( destName);
    if (( sourceRef != NULL) && ( destRef != NULL)) {
	unitConvert = ( AmUnitConvertPtr) malloc( sizeof( AmUnitConvertRec));
	unitConvert->sourceUnits = sourceRef->u.unitDef.code;
	unitConvert->destUnits = destRef->u.unitDef.code;
	unitConvert->conversion = expr;
	addToUnitConvertTable( unitConvert);
    }
}

/****************************************************************************
 *
 * Description:
 * 	initialize the units and varaiable files.
 * Returns:
 *	TRUE on success, FALSE on failure, also sets the unitInitFlag
 * Modifications:
 *	should return an error code if the files didn't parse properly.
 *
 ****************************************************************************/

static AnvilBoolean 
doUnitStuff()
{
    char        unitsPath[ PATHLEN], varsPath[ PATHLEN], errorString[ MSGLEN];
    AmErrorCode	errCode;

    if ( !unitInitFlag) {
	sprintf( unitsPath, "%s/lookups/units", getenv( CONFIG_ENV));
	if (( errCode = AmParseUnitsFile( unitsPath, errorString)) != AmOK) {
	    return( FALSE);
	} else {
	    unitInitFlag = TRUE;
	}
	sprintf( varsPath, "%s/%s/vars", getenv( CONFIG_ENV), AM_LOOKUP_DIR);
	if (( errCode = AmParseVariableFile( varsPath)) != AmOK) {
	    DO_RETURN_ERROR( errCode);
	}
    }

    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	convert a value from one unit to another 
 * Arguments:
 *	sourceCode - unit code to convert the value from.
 *	destCode - unit code to convert to.
 *	value - value to convert.
 *	convertValue - resulting converted value.
 * Returns:
 *	TRUE on success, FALSE if either unit or the unit converted couldn't
 *	be found.
 *
 ****************************************************************************/

AnvilBoolean 
AmConvertValue( unitType	sourceCode,
	      unitType	destCode,
	      float	value,
	      float	*convertValue)
{
    AmUnitConvertPtr	converter;

    doUnitStuff();
    if ( sourceCode == destCode) {
	*convertValue = value;
	return( TRUE);
    }
    converter = getUnitConverter( sourceCode, destCode);
    if ( converter != NULL) {
	*convertValue = AmEvalSimpleExpr( converter->conversion, value);
	return( TRUE);
    } else {
	*convertValue = value;
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	is there a converter avaiable for this set of units? 
 * Arguments:
 *	sourceCode - unit code to convert the value from.
 *	destCode - unit code to convert to.
 * Returns:
 *	TRUE on success, FALSE if either unit or the unit converted couldn't
 *	be found.
 *
 ****************************************************************************/

AnvilBoolean 
AmConverterExists( unitType	sourceCode,
		 unitType	destCode)
{
    doUnitStuff();
    if ( sourceCode == destCode) {
	return( TRUE);
    } else {
	return( getUnitConverter( sourceCode, destCode) != NULL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	parse the variable file, which is of the form variable:      units
 * Arguments:
 *	file - file to parse.
 * Returns:
 *	AmOPEN_FAILED on open file failure.
 * Modifications:
 *	should have more informative error messages for dealing with
 *	parsing errors.
 *	should take a const char * for the file.
 *
 ****************************************************************************/

AmErrorCode 
AmParseVariableFile( char     *file)
{
    FILE        *fp;
    char        buffer[ RESLEN], varname[ RESLEN], unitname[ RESLEN];

    if (( fp = fopen( file, "r")) == NULL) {
        return( AmOPEN_FAILED);
    }
    do {
        if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
            break;
        }
        sscanf( buffer, "%[^:]: %[^\n]", varname, unitname);
        AmAddVariableUnits( varname, unitname);
    } while ( TRUE);
    fclose( fp);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	add a variable and the units are used for displaying the variable 
 * Arguments:
 *	varStr - variable name.
 *	unitStr - units to use for the variable.
 * Returns:
 *	TRUE if the units were found, FALSE otherwise.
 * Modifications:
 *	should take const char *'s
 *	should return an error code for malloc failure, or bad argument.
 *
 ****************************************************************************/

AnvilBoolean 
AmAddVariableUnits( char  *varStr,
		  char  *unitStr)
{
    AmVarDefPtr   varDef;
    unitNamePtr ref;

    ref = findUnitByName( unitStr);
    if ( ref != NULL) {
        varDef = ( AmVarDefPtr) malloc( sizeof( AmVarDefRec));
	if ( varDef == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    return( FALSE);
	}
        varDef->variableCode = VAR_CODE_COUNT++;
        varDef->variableName = strdup( varStr);
        varDef->unitCode = ref->u.unitDef.code;
        addToVarNameTable( varDef);
        addToVarCodeTable( varDef);
        return( TRUE);
    } else {
        return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	convert a variable string to a variable code
 * Arguments:
 *	varString - variable name.
 *	varCode - (hopefully) returned variable code.
 * Returns:
 *	TRUE if the variable code was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetVariableCode( const char     *varString,
		 int            *varCode)
{
    AmVarDefPtr   *ref;
    AmVarDefRec   target;
    char        varStringBuf[ 256];

    doUnitStuff();
    strcpy( varStringBuf, varString);
    target.variableName = varStringBuf;
    ref = ( AmVarDefPtr *) tfind( &target, 
				( const void **) &varNameRoot,
                                ( AmSearchFn) compareVarName);
    if ( ref !=  NULL) {
        *varCode = ( *ref)->variableCode;
        return( TRUE);
    } else {
        return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	convert a variable string to a variable code
 * Arguments:
 *	varCode - variable code.
 *	varString - (hopefully) returned variable string.
 * Returns:
 *	TRUE if the variable string was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetVariableString( varType      varCode,
		   char         **varString)
{
    AmVarDefPtr   *ref;
    AmVarDefRec   target;

    doUnitStuff();
    target.variableCode = varCode;
    ref = ( AmVarDefPtr *) tfind( &target, 
				( const void **) &varCodeRoot,
                                ( AmSearchFn) compareVarCode);
    if ( ref != NULL) {
        *varString = ( *ref)->variableName;
        return( TRUE);
    } else {
        *varString = "";
        return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Get the units for a variable.
 * Arguments:
 *	varCode - variable code.
 *	varString - (hopefully) returned variable string.
 * Returns:
 *	TRUE if the variable string was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean AmGetVariableUnits( varType       varCode,
                          unitType      *unitCode)
{
    AmVarDefPtr   *ref;
    AmVarDefRec   target;

    doUnitStuff();
    target.variableCode = varCode;
    ref = ( AmVarDefPtr *) tfind( &target, 
				( const void **) &varCodeRoot,
                                ( AmSearchFn) compareVarCode);
    if ( ref !=  NULL) {
        *unitCode = ( *ref)->unitCode;
        return( TRUE);
    } else {
        return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Get the units for a variable.
 * Arguments:
 *	varCode - variable code.
 *	varString - (hopefully) returned variable string.
 * Returns:
 *	TRUE if all the strings were found in the units table, FALSE
 *	otherwise.
 * Modifications:
 *	This should probably be in a separate configuration file, so
 *	people can add their own units.
 *
 ****************************************************************************/

AnvilBoolean 
AmInitUnits()
{
    doUnitStuff();
    return( AmGetUnitCode( "meter", &AmUNIT_METER) &&
	    AmGetUnitCode( "longitude_arcdeg", &AmUNIT_LONGITUDE_ARCDEG) &&
	    AmGetUnitCode( "latitude_arcdeg", &AmUNIT_LATITUDE_ARCDEG) &&
	    AmGetUnitCode( "arcdeg", &AmUNIT_ARCDEG));
}

/****************************************************************************
 *
 * Description:
 * 	perform rouding on a value depending on its print precision.
 * Arguments:
 *	value - value to round.
 *	precision - precision to round it to.
 *
 ****************************************************************************/

static void 
roundit( float	*value,
	 int	precision)
{
    int	ii,

    factor = 1;
    for ( ii = 0; ii < precision; ii++) {
	factor *= 10;
    }
    *value *= factor;
    *value = rint( *value);
    *value /= factor;
}

/****************************************************************************
 *
 * Description:
 * 	given a unit code and a value, format the value to a string. 
 *	Defaults to printing the value with 2 decimal places if the 
 *	unit precision or printing function isn't found.
 * Arguments:
 *	unitCode - unit code to search for in the unit lookup table to
 *	either get the unit string and precision, or custom print function 
 *	if there is one.  Defaults to printing the value with 2 decimal
 *	places.
 *	value - value to format.
 *	buffer - string to print it into.
 * Returns:
 *	TRUE if a print function or precision/string was found, else FALSE
 *
 ****************************************************************************/

AnvilBoolean 
AmPrintValue( unitType	unitCode,
	    float	value,
	    char	*buffer)
{
    char	fmt[ 64];
    unitNamePtr	ref;
	
    doUnitStuff();
    ref = findUnitByCode( unitCode);
    if ( ref != NULL) {
	if ( ref->u.unitDef.printFunc != NULL) {
	    ( ref->u.unitDef.printFunc)( value, buffer);
	} else {

/* use the print string if there is one */

	    if ( ref->u.unitDef.string == NULL) {
		sprintf( fmt, "%%.%df %s", 
			 ref->u.unitDef.precision, ref->name);
	    } else {
		sprintf( fmt, "%%.%df %s", 
			 ref->u.unitDef.precision, ref->u.unitDef.string);
	    }
	    roundit( &value, ref->u.unitDef.precision);
	    sprintf( buffer, fmt, value);
	}
	return( TRUE);
    } else {

/* default is 2 decimal places */

	roundit( &value, 2);
	sprintf( buffer, "%.2f", value);
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	given a value and units, scan it.  If there's a custom scanning
 *	function use it, otherwise scan as a float.
 * Arguments:
 *	unitCode - unit code to search for the scan function.
 *	buffer - buffer to scan
 *	value - returned value.
 * Returns:
 *	1 if value scanned, 0 otherwise.
 * Modifications:
 *	should actually test the scan and return TRUE/FALSE.
 *
 ****************************************************************************/

AnvilBoolean 
AmScanValue( unitType	unitCode,
	   char		*buffer,
	   float	*value)
{
    unitNamePtr	ref;

    doUnitStuff();
    ref = findUnitByCode( unitCode);
    if ( ref != NULL) {
	if ( ref->u.unitDef.scanFunc != NULL) {
	    return(( ref->u.unitDef.scanFunc)( buffer, value));
	} else {
	    return( sscanf( buffer, "%f", value) == 1);
	}
    } else {
	return( sscanf( buffer, "%f", value) == 1);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Print the value with no units and the precision, or use the print 
 *	function if there's a custom print function for the unit code.
 * Arguments:
 *	unitCode - to search for a print function or precision if available.
 *	value - value to print.
 *	buffer - buffer to print it into.
 * Returns:
 *	TRUE if a print function or precision was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPrintValueNoUnits( unitType	unitCode,
		   float	value,
		   char		*buffer)
{
    char	fmt[ 64];
    unitNamePtr	ref;

    doUnitStuff();
    ref = findUnitByCode( unitCode);
    if ( ref != NULL) {

/* see if there's a custom print function... */

	if ( ref->u.unitDef.printFunc != NULL) {
	    ( ref->u.unitDef.printFunc)( value, buffer);
	} else {

/* otherwise, use the precision specified for the units */

	    sprintf( fmt, "%%.%df", ref->u.unitDef.precision);
	    sprintf( buffer, fmt, value);
	}
	return( TRUE);
    } else {

/* default precision is 2 decimal places */

	sprintf( buffer, "%.2f", value);
	return( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 *	convert a unit string to a unit code by searching the units table.
 * Arguments:
 *	unitString - string to search for.
 *	unitCode - returned code AmUNIT_UNDEFINED if not found.
 * Returns:
 *	TRUE if the string was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetUnitCode( const char	*unitString, 
	     unitType	*unitCode)
{
    unitNamePtr	ref;

    doUnitStuff();
    ref = findUnitByName( unitString);
    if ( ref != NULL) {
	*unitCode = ref->u.unitDef.code;
	return( TRUE);
    } else {
	*unitCode = AmUNIT_UNDEFINED;
	return( FALSE);
    }
}
	
/****************************************************************************
 *
 * Description:
 *	convert a unit code to a unit string by searching the units table.
 * Arguments:
 *	unitCode - code to search for
 *	unitString - matching unit string, or "" if not found.
 * Returns:
 *	TRUE if the code was found, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGetUnitString( unitType	unitCode,
	       char	**unitString)
{
    unitNamePtr	ref;

    doUnitStuff();
    ref = findUnitByCode( unitCode);
    if ( ref != NULL) {
	*unitString = ref->name;
	return( TRUE);
    } else {
	*unitString = "";
	return( FALSE);
    }
}

/* static variables used for traversing the units table. */

static int COUNT; 
static unitType *CODES;
static char **STRINGS;

/****************************************************************************
 *
 * Description:
 *	count the number of unit codes.
 * Arguments:
 *	twalk() function arguments.
 *
 ****************************************************************************/

static void 
countAction( const void *, VISIT order, int)
{
    if (( order == preorder) || ( order == leaf)) {
	COUNT++;
    }
}

/****************************************************************************
 *
 * Description:
 *	set the code in the node in the table, and increment the count.
 * Arguments:
 *	twalk() function arguments.
 *
 ****************************************************************************/

static void 
setCodeAction( const void *a, VISIT order, int)
{
    unitNamePtr	ptr;

    if (( order == preorder) || ( order == leaf)) {
	ptr = *( unitNamePtr *) a;
	CODES[ COUNT++] = ptr->u.unitDef.code;
    }
}

/****************************************************************************
 *
 * Description:
 *	set the string in the node in the table, and increment the count.
 * Arguments:
 *	twalk() function arguments.
 *
 ****************************************************************************/

static void 
setStringAction( const void *a, VISIT order, int)
{
    unitNamePtr	ptr;

    if (( order == preorder) || ( order == leaf)) {
	ptr = *( unitNamePtr *) a;
	STRINGS[ COUNT++] = strdup( ptr->name);
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the list of available unit codes, by doing a twalk() on
 *	the unit code/name tree.
 * Arguments:
 *	codes - pointer to the returned vector of unit codes.
 *	numCodes - pointer to the length of the unit code vector.
 *
 ****************************************************************************/

void 
AmGetUnitCodeList( int	**codes,
	         int	*numCodes)
{
    doUnitStuff();
    COUNT = 0;
    twalk( unitCodeRoot, ( AmWalkFn) countAction);
    *numCodes = COUNT;
    COUNT = 0;
    CODES = *codes = ( unitType *) malloc( *numCodes*sizeof( unitType));
    twalk( unitCodeRoot, ( AmWalkFn) setCodeAction);
}

/****************************************************************************
 *
 * Description:
 * 	return the list of available unit strings, by doing a twalk() on
 *	the unit code/name tree.
 * Arguments:
 *	strings - pointer to the returned vector of unit strings.
 *	numStrings - pointer to the length of the unit string vector.
 *
 ****************************************************************************/

void 
AmGetUnitStringList( char	***strings,
		   int	*numStrings)
{
    doUnitStuff();
    COUNT = 0;
    twalk( unitCodeRoot, ( AmWalkFn) countAction);
    *numStrings = COUNT;
    COUNT = 0;

/* make the string list NULL terminated for convienience */

    STRINGS = *strings = ( char **) malloc(( *numStrings + 1)*sizeof( char *));
    twalk( unitCodeRoot, ( AmWalkFn) setStringAction);
    ( *strings)[ *numStrings] = NULL;
}

/****************************************************************************
 *
 * Description:
 * 	units file read function for the parser.
 * Arguments:
 *	buffer - buffer to read into.
 *	maxSize - size of the buffer.
 * Returns:
 *	number of bytes read.
 *
 ****************************************************************************/
   
int 
AmUnitsFpYYinput( char	*buffer,
		int	maxSize)
{
    return( fread( buffer, 1, maxSize, unitsFp));
}

/****************************************************************************
 *
 * Description:
 * 	parse the units file and store the information in the unit 
 *	code and unit string trees
 * Arguments:
 *	file - file to parse
 *	errorString - returned error string if there was a parse error.
 * Returns:
 *	AmOK on sucess, AmOPEN_FAILED if the file couldn't be opened,
 *	AmBAD_ARGUMENT if the file couldn't be parsed.
 * Modifications:
 *	file should be a const char *
 *
 ****************************************************************************/

AmErrorCode 
AmParseUnitsFile( char	*file,
		  char	*errorString)
{
    if (( unitsFp = fopen( file, "r")) == NULL) {
	sprintf( errorString, "failed to open %s", file);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    unitsrestart( unitsFp);
    AmUnitsErrorLineNo = 1;
    errorFlag = FALSE;
    unitsparse();
    fclose( unitsFp);
    if ( errorFlag) {
	strcpy( errorString, AmUnitsErrorString);
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	YACC parser error function.
 * Arguments:
 *	error string 
 * Modifications:
 *	This should probably be more informative.
 *
 ****************************************************************************/

int 
unitserror( const char *errorString)
{
    sprintf( AmUnitsErrorString, "parse error: %s at line %d\n", 
	     errorString, AmUnitsErrorLineNo);
    errorFlag = TRUE;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	scan a latitude or longitude value in DD MM SS NEWS format.
 * Arguments:
 *	buffer - buffer to scan.
 *	value - value to scan into.
 * Returns:
 *	TRUE if the value scanned properly, FALSE otherwise.
 * Modifications:
 *	doesn't return FALSE if the value didn't scan.
 *
 ****************************************************************************/

static AnvilBoolean 
scanLatLon( char	*buffer,
	    float	*value)
{
    int		degrees = 0, minutes = 0, seconds = 0;
    char	compass = 'E', pad1[ 64], pad2[ 64], pad3[ 64];

/* if it's a floating point value, then just convert it from float syntax */

    if ( AmRegexpMatch( RE_REAL_FMT, buffer, FALSE)) {
	*value = atof( buffer);
	return( TRUE);
    } else {

/* check if it's DD-MM-SS C format */

	if ( sscanf( buffer, "%d%[^0-9]%d%[^0-9]%d%[^NEWSnews]%c", 
		     &degrees, pad1, &minutes, pad2, &seconds, 
		     pad3, &compass) >= 5) {
	    *value = degrees + minutes/60.0 + seconds/3600.0;

/* how about DD-MM C format? */

	} else if ( sscanf( buffer, "%d%[^0-9]%d%[^NEWSnews]%c",
			    &degrees, pad1, &minutes, pad2, &compass) == 5) {
	    *value = degrees + minutes/60.0;

/* OK, maybe DD C format? */

	} else {
	    sscanf( buffer, "%d%[^NEWSnews]%c", &degrees, pad1, &compass);
	    *value = degrees;
	}

/* if the latitude or longitude is west or south, then return -value */

	if (( tolower( compass) == 'w') ||
	    ( tolower( compass) == 's')) {
	    *value = -*value;
	}
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	print a latitude longitude value in DD MM SS NEWS format, given 
 *	the compass direction
 * Arguments:
 *	val - value to print
 *	buffer - buffer to print into.
 *	dir - compass direction.
 *
 ****************************************************************************/

static void 
printLatlonVal( float	val,
		char	*buffer,
		char	dir)
{
    int          ideg, imin, isec;

/* in case the incoming value is negative, take the absolute 
   value (we assume that the "dir" character already takes into
   account whether this value is positive or negative) */

    val = fabs(val);

/* AmFirst, get the integer degrees ... */

    ideg = ( int) val;
    val -= ( float) ideg;
    val *= 60.0;

/* the, the integer minutes ... */

    imin = ( int) floor( val + 0.005);
    val -= ( float) imin;
    if ( val < 0.0) val = 0.0;
    val *= 60.0;

/* and finally, the integer seconds ... */

    isec = ( int) rint( val);

/* make sure we don't have 60 minutes or 60 seconds ... */

    if ( isec == 60) {
	isec = 0;
	imin++;
    }
    if ( imin == 60) {
	imin = 0;
	ideg++;
    }

/* and finally, print the string ... */

    if ( isec == 0) {
	if ( imin == 0) {
	    sprintf( buffer, "%3d%c %c", ideg, '\260', dir);
	} else {
	    sprintf( buffer, "%3d%c%2d' %c", ideg, '\260', imin, dir);
	}
    } else {
	sprintf( buffer, "%3d%c%2d'%2d\" %c", ideg, '\260', imin, isec, dir);
    }

    return;
}

/****************************************************************************
 *
 * Description:
 * 	print a latitude value in DD MM SS NEWS format.
 *	(used as a registered function)
 * Arguments:
 *	val - value to print
 *	buffer - buffer to print into.
 *
 ****************************************************************************/

static void 
printLatitude( float	val,
	       char	*buffer)
{
    char	dir;

    dir = ( val >= 0) ? 'N' : 'S';
    printLatlonVal( val, buffer, dir);
}

/****************************************************************************
 *
 * Description:
 * 	print a longitude value in DD MM SS NEWS format.
 *	(used as a registered function)
 * Arguments:
 *	val - value to print
 *	buffer - buffer to print into.
 *
 ****************************************************************************/

static void 
printLongitude( float	val,
		char	*buffer)
{
    char	dir;

    dir = ( val >= 0) ? 'E' : 'W';
    printLatlonVal( val, buffer, dir);
}

/****************************************************************************
 *
 * Description:
 * 	register widely used unit functions.
 *
 ****************************************************************************/

void 
AmRegisterUnitFns()
{
    AmAddRegister( "scanLatLon", scanLatLon);
    AmAddRegister( "printLatitude", printLatitude);
    AmAddRegister( "printLongitude", printLongitude);
}

/****************************************************************************
 *
 * Description:
 * 	general purpose wrapper for unit initialization. register the
 *	unit functions, parse the units file, parse the variables file,
 *	parse the projection units file, and set the enumeration 
 *	lookup databaes.
 * Returns:
 *	AmOK on success, system error on failure
 *
 ****************************************************************************/

AmErrorCode AmInitUnitStuff()
{
    char        unitsPath[ PATHLEN], varsPath[ PATHLEN], errorString[ MSGLEN],
                projUnitsPath[ PATHLEN], enumsPath[ PATHLEN];
    AmErrorCode errCode;

/* register the predefined unit functions */

    AmRegisterUnitFns();

/* parse the units file */

    sprintf( unitsPath, "%s/%s/units", getenv( CONFIG_ENV), AM_LOOKUP_DIR);
    if (( errCode = AmParseUnitsFile( unitsPath, errorString)) != AmOK) {
        DO_RETURN_ERROR( errCode);
    }
    if ( !AmInitUnits()) {
        DO_RETURN_ERROR( errCode);
    }

/* parse the variables file */

    sprintf( varsPath, "%s/%s/vars", getenv( CONFIG_ENV), AM_LOOKUP_DIR);
    if (( errCode = AmParseVariableFile( varsPath)) != AmOK) {
        DO_RETURN_ERROR( errCode);
    }

/* parse the file of units available for projections */

    sprintf( projUnitsPath, "%s/%s/projUnits", getenv( CONFIG_ENV),
             AM_LOOKUP_DIR);
    if (( errCode =  ReadProjUnits( projUnitsPath)) != AmOK) {
        DO_RETURN_ERROR( errCode);
    }

/* set up the enumeration lookup database */

    sprintf( enumsPath, "%s/%s/%s",
             getenv( CONFIG_ENV), AM_LOOKUP_DIR, AM_ENUM_LOOKUP);
    if (( errCode = AmParseEnumerationFile( enumsPath, errorString)) != AmOK) {
        return( errCode);
    }
    return( AmOK);
}
