#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <anvil/project.h>
#include <anvil/colorhist.h>


/*
 * Function: onIndex( a, b) 
 *
 * Purpose:
 *   Compares the index values of two histogram records.  This is used
 *   to sort a vector of histogram records by the index value.
 *
 * Parameters:
 *   AmHistPtr a : First histogram record. 
 *   AmHistPtr b : Second histogram record. 
 *
 * Returns:
 *   int : < 0, if a is less than b.
 *         = 0, if a is equals to b. 
 *         > 0, if a is greater than b.
 */

static int
onIndex( AmHistPtr a,
           AmHistPtr b)
{
    return( a->index - b->index);
}


/*
 * Function: bigHistFirst( a, b) 
 *
 * Purpose:
 *   Compares the histogram values of two histogram records.  This is used
 *   to sort a vector of histogram records by the histogram values. 
 *
 * Parameters:
 *   AmHistPtr a : First histogram record. 
 *   AmHistPtr b : Second histogram record. 
 *
 * Returns:
 *   int : < 0, if b is less than a.
 *         = 0, if b is equals to a. 
 *         > 0, if b is greater than a.
 */

static int
bigHistFirst( AmHistPtr a,
              AmHistPtr b)
{
    return( b->hist - a->hist);
}


/*
 * Function: AmReducePixel( ulPixel) 
 *
 * Purpose:
 *   Reduce a pixel from a source to the destination resolution.
 *
 * Parameters:
 *   ulong_t   ulPixel : Pixel to reduce from the source resolution.
 *
 * Returns:
 *   ulong_t : The pixel at the destination resolution.
 *
 * Comments:
 *   This assumes that the source pixel format is BGR.
 */

ulong_t
AmReducePixel( ulong_t ulPixel)
{
    uchar_t *pcPixel;
    ulong_t ulRed, ulGreen, ulBlue;

    pcPixel = ( uchar_t *) &ulPixel;
    pcPixel++;

    ulBlue  = (( ulong_t) *pcPixel++ >> ( SRC_RES - DST_RES)) & DST_MASK;
    ulGreen = (( ulong_t) *pcPixel++ >> ( SRC_RES - DST_RES)) & DST_MASK;
    ulRed   = (( ulong_t) *pcPixel++ >> ( SRC_RES - DST_RES)) & DST_MASK;

    return(( ulBlue << ( DST_RES*2)) | ( ulGreen << DST_RES) | ulRed);
}


/*
 * Function: AmDecomposeReducedPixel( ulPixel, pucRed, pucGreen, pucBlue) 
 *
 * Purpose:
 *   Reduce a pixel from a source to the destination resolution and
 *   decompose it into its red, green, and blue components.
 *
 * Parameters:
 *   ulong_t ulPixel   : Pixel to reduce and decompose.
 *   uchar_t *pucRed   : The red component of the reduced pixel. 
 *   uchar_t *pucGreen : The green component of the reduced pixel. 
 *   uchar_t *pucBlue  : The blue component of the reduced pixel. 
 *   
 * Returns:
 *   The RGB components of the pixel are passed back through the pucRed,
 *   pucGree, and pucBlue parameters.
 *
 * Comments:
 *   This assumes that the source pixel format is BGR.
 */

void
AmDecomposeReducedPixel( ulong_t ulPixel,
                         uchar_t *pucRed,
                         uchar_t *pucGreen,
                         uchar_t *pucBlue)
{
    ulong_t ulRed, ulGreen, ulBlue;

    ulRed   = ( ulPixel & DST_MASK) << LOSS;
    ulGreen = (( ulPixel >> DST_RES) & DST_MASK) << LOSS;
    ulBlue  = (( ulPixel >> DST_RES*2) & DST_MASK) << LOSS;

    *pucRed   = ulRed;
    *pucGreen = ulGreen;
    *pucBlue  = ulBlue;
}


/*
 * Function: AmClosestColorIndex( vHist, pColor, nHist) 
 *
 * Purpose:
 *   Find the closest color in the histogram to the specified pixel.
 *
 * Parameters:
 *   AmHistPtr      vHist  : A vector of histogram values. 
 *   AmColorCellPtr pColor : Color which to find the closest color for. 
 *   int            nHist  : Number of entries in the histogram vector. 
 *
 * Returns:
 *   The index of the closest color.
 */

int
AmClosestColorIndex( AmHistPtr      vHist,
                     AmColorCellPtr pColor,
                     int            nHist)
{
    int colordist = 768, ii, canddist, closest = 0;

    for ( ii = 0; ii < nHist; ii++) {
        canddist = ( int) ( 0.33*abs( vHist[ ii].color.red - pColor->red) +
                            0.5*abs( vHist[ ii].color.green - pColor->green) +
                            0.17*abs( vHist[ ii].color.blue - pColor->blue));
        if ( canddist < colordist) {
            colordist = canddist;
            closest = ii;
        }
    }

    return( closest);
}


/*
 * Function: AmImage8to24( pcImage, cx, cy, vColors, nColors, ppulImage) 
 *
 * Purpose:
 *   Upgrade an image from 8 to 24 pixels.
 *
 * Parameters:
 *   uchar_t        *pcImage    : The 8 bit image.
 *   int            cx          : Width of the image.
 *   int            cy          : Height of the image.
 *   AmColorCellPtr vColors     : The palette for the 8 bit image.
 *   int            nColors     : The number of colors in the 8 bit palette.
 *   ulong_t        **ppulImage : The new 24 bit image.
 *   
 * Returns:
 *   The new image is passed back through the ppulImage parameter.
 *
 * Comments:
 *   This assumes that the destination pixel format is BGR.
 */

/* ARGSUSED */
void
AmImage8to24( uchar_t        *pcImage,
              int            cx,
              int            cy,
              AmColorCellPtr vColors,
              int            nColors,
              ulong_t        **ppulImage)
{
    int     ii;
    uchar_t *pc;
    ulong_t *pul;

    *ppulImage = ( ulong_t *) malloc( cx * cy * sizeof( ulong_t));
    if ( *ppulImage != NULL) {

        for ( ii = 0, pul = *ppulImage; ii < cx * cy; ii++, pul++) {
            *pul = 0; /* to prevent UMR's */
            pc = ( uchar_t *) pul;
            ASSIGN_PIXEL( vColors[ pcImage[ ii]], pc);
        }

    }
}


/*
 * Function: AmImage24to8( pcImage, cx, cy, vColors, nColors, ppulImage) 
 *
 * Purpose:
 *   Downgrade an image from 24 to 8 pixels using close.
 *
 * Parameters:
 *   ulong_t        **pulImage  : The 24 bit image.
 *   int            cx          : Width of the image.
 *   int            cy          : Height of the image.
 *   uchar_t        *ppcImage   : The new 8 bit image.
 *   AmColorCellPtr vColors     : The palette for the 8 bit image.
 *   int            nColors     : The number of colors in the 8 bit palette.
 *   
 * Returns:
 *   The new image is passed back through the ppulImage parameter.
 *
 * Comments:
 *   This assumes that the destination pixel format is BGR.
 */

void AmImage24to8( ulong_t        *pulImage,
                   int            cx,
                   int            cy,
                   uchar_t        **ppcImage,
                   AmColorCellPtr *vColors,
                   int            *nColors)     /* passed and returned */
{
    AmHistRec        vHist[ NUMHIST];
    AmColorAssignRec colormap[ NUMHIST];
    int              ii, end, index[ NUMHIST];

/* Generate a histogram of the pixel shifted to 32768 colors */

    for ( ii = 0; ii < NUMHIST; ii++) {
        vHist[ ii].hist = 0;
        AmDecomposeReducedPixel(( ulong_t) ii,
                                &vHist[ ii].color.red,
                                &vHist[ ii].color.green,
                                &vHist[ ii].color.blue);
        vHist[ ii].index = ii;
    }

/* Stuff the histogram */

    for ( ii = 0; ii < cx * cy; ii++) {
        vHist[ AmReducePixel( pulImage[ ii])].hist++;
    }

    end = 0;
    for ( ii = 0; ii < NUMHIST; ii++) {
        index[ ii] = end;
        if ( vHist[ ii].hist > 0) {
            vHist[ end] = vHist[ ii];
            end++;
        }
    }

    if ( end < *nColors) {
        *nColors = end;
    }

/* Sort the array by the most popular colors, assign the indexes, and
   set the rest of the histogram indexes to the closest popular color */

    qsort( vHist, end, sizeof( AmHistRec), ( AmSearchFn) bigHistFirst);
    for ( ii = 0; ii < *nColors; ii++) {
        vHist[ ii].cmapIndex = ii;
        colormap[ ii].assigned = TRUE;
        colormap[ ii].color = vHist[ ii].color;
    }

    for ( ii = *nColors; ii < end; ii++) {
        vHist[ ii].cmapIndex =
            AmClosestColorIndex( vHist, &vHist[ ii].color, *nColors);
    }

/* Sort the array into the order indexed by the index array. */

    qsort( vHist, end, sizeof( AmHistRec), ( AmSearchFn) onIndex);
    *ppcImage = ( uchar_t *) malloc( cx * cy);

/* Assign the image pixels to the colormap index. */

    for ( ii = 0; ii < cx * cy; ii++) {
        ( *ppcImage)[ ii] = ( uchar_t)
            vHist[ index[ AmReducePixel( pulImage[ ii])]].cmapIndex;
    }

/* Copy the colors from the histogram. */

    *vColors = ( AmColorCellPtr) malloc( sizeof( AmColorCellRec)*( *nColors));
    for ( ii = 0; ii < *nColors; ii++) {
        ( *vColors)[ ii] = vHist[ ii].color;
    }
}

