#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdarg.h>
#include <stdio.h>

#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/Form.h>

#include <anvil/project.h>
#include <anvil/resource.h>
#include <anvil/errlib.h>
#include <anvil/motif.h>
#include <anvil/errdlog.h>
#include <anvil/utility.h>
#include <anvil/files.h>
#include <anvil/log.h>
#include <anvil/eventcodes.h>
#include <anvil/userevents.h>

#include <ToolKit/DialogClass.h>
#include <ToolKit/ScrolledTextClass.h>

#include <ToolKit/ErrorBoxClass.h>
#include <ToolKit/WarningBoxClass.h>
#include <ToolKit/InformationBoxClass.h>

static Widget wErrorShell = NULL;

/**********************************************************************
 *
 * Error reporting functions: 
 *
 **********************************************************************/

/*
 * Function: AmInitializeError( wShell)
 *
 * Purpose:
 *     Initializes the shell for the error dialogs.
 *
 * Parameters:
 *     Widget wShell: The widget that is supposed to act as the
 *                    shell for the error dialogs.
 */

void
AmInitializeError( Widget wShell)
{
    wErrorShell = wShell;
}


/*
 * Function: PrintErrorBuffer( wParent, sz)
 *
 * Purpose:
 *     Creates an error dialog using the passed in widget as a parent
 *     and displaying the passed in string as the message.
 *
 * Parameters:
 *     Widget     wParent: Parent widget to be used for the error dialog.     
 *     const char *sz    : Message to be displayed in the dialog.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

static Widget
PrintErrorBuffer( Widget wParent, const char *sz)
{
    AmErrorBoxClass *eb;
 
    eb = new AmErrorBoxClass( wParent, "ErrorBox", sz); 
    AmSendUserEvent( AM_PKG_ID, AM_MESSAGE_STRING, ( char *) sz);

/* Return the dialog widget */

    return(( Widget) *eb); 
}


/*
 * Function: AmPrintError( szFormat, ...)
 *
 * Purpose:
 *     Prints an error dialog using the internal error shell as the dialog's
 *     parent.  The error message is printed using the format string and the
 *     variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     const char *szFormat : Format string for the error message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintError( const char *szFormat, ...)
{
    va_list args;
    char    sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the error dialog using the internal error shell. */

    return( PrintErrorBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintError( wShell, szFormat, ...)
 *
 * Purpose:
 *     Prints an error dialog using the passed in shell as the dialog's
 *     parent.  The error message is printed using the format string and the
 *     variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     Widget     wParent   : Parent widget for the error dialog.
 *     const char *szFormat : Format string for the error message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintError( Widget wShell, const char *szFormat, ...)
{
    va_list	args;
    char	sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the error dialog using the passed in shell widget. */

    return( PrintErrorBuffer( wShell, sz));
}


/*
 * Function: AmPrintResourceError( szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintError except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The internal error shell is used as the dialog's parent widget.
 *
 * Parameters:
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintResourceError( const char *szResource, ...)
{
    va_list      args;
    char         sz[ MSGLEN];
    const char	 *szFormat;
 
/* Get the string from the resource file */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the error dialog using the internal error shell. */

    return( PrintErrorBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintResourceError( wShell, szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintError except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The passed in widget is used as the dialog's parent widget.
 *
 * Parameters:
 *     Widget     wShell      : Parent widget for the error dialog.
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget AmPrintResourceError( Widget wShell, const char *szResource, ...)
{
    va_list      args;
    char         sz[ MSGLEN]; 
    const char   *szFormat;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the error dialog using the passed in shell widget. */

    return( PrintErrorBuffer( wShell, sz));
}



/**********************************************************************
 *
 * Warning reporting functions: 
 *
 **********************************************************************/

/*
 * Function: PrintWarnBuffer( wParent, sz)
 *
 * Purpose:
 *     Creates a warning dialog using the passed in widget as a parent
 *     and displaying the passed in string as the message.
 *
 * Parameters:
 *     Widget     wParent: Parent widget to be used for the warning dialog.
 *     const char *sz    : Message to be displayed in the dialog.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

static Widget
PrintWarnBuffer( Widget wParent, const char *sz)
{
    AmWarningBoxClass *wb;

    wb = new AmWarningBoxClass( wParent, "WarningBox", sz);
    AmSendUserEvent( AM_PKG_ID, AM_MESSAGE_STRING, ( char *) sz);

/* Return the dialog widget */

    return(( Widget) *wb); 
}


/*
 * Function: AmPrintWarn( szFormat, ...)
 *
 * Purpose:
 *     Prints a warning dialog using the internal error shell as the dialog's
 *     parent.  The warning message is printed using the format string and the
 *     variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     const char *szFormat : Format string for the warning message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintWarn( const char *szFormat, ...)
{
    va_list	args;
    char	sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the warning dialog using the passed in shell widget. */

    return( PrintWarnBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintWarn( wShell, szFormat, ...)
 *
 * Purpose:
 *     Prints a warning dialog using the passed in shell as the dialog's
 *     parent.  The warning message is printed using the format string and
 *     the variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     Widget     wParent   : Parent widget for the warning dialog.
 *     const char *szFormat : Format string for the warning message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintWarn( Widget wShell, const char *szFormat, ...)
{
    va_list	args;
    char	sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the warning dialog using the passed in shell widget. */

    return( PrintWarnBuffer( wShell, sz));
}


/*
 * Function: AmPrintResourceWarn( szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintWarn except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The internal error shell is used as the dialog's parent widget.
 *
 * Parameters:
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintResourceWarn( const char *szResource, ...)
{
    va_list    args;
    char       sz[ MSGLEN];
    const char *szFormat;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the warning dialog using the internal error shell widget. */

    return( PrintWarnBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintResourceWarn( wShell, szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintError except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The passed in widget is used as the dialog's parent widget.
 *
 * Parameters:
 *     Widget     wShell      : Parent widget for the error dialog.
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintResourceWarn( Widget wShell, const char *szResource, ...)
{
    va_list      args;
    char         sz[ MSGLEN]; 
    const char   *szFormat;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the warning dialog using the passed in shell widget. */

    return( PrintWarnBuffer( wShell, sz));
}



/**********************************************************************
 *
 * Information reporting functions: 
 *
 **********************************************************************/

/*
 * Function: PrintInfoBuffer( wParent, sz)
 *
 * Purpose:
 *     Creates an information dialog using the passed in widget as a parent
 *     and displaying the passed in string as the message.
 *
 * Parameters:
 *     Widget     wParent: Parent widget to be used for the info dialog.     
 *     const char *sz    : Message to be displayed in the dialog.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

static Widget
PrintInfoBuffer( Widget wParent, const char *sz)
{
    AmInformationBoxClass *ib;

    ib = new AmInformationBoxClass( wParent, "InfoBox", sz); 
    AmSendUserEvent( AM_PKG_ID, AM_MESSAGE_STRING, ( char *) sz);

    return(( Widget) *ib); 
}


/*
 * Function: AmPrintInfo( szFormat, ...)
 *
 * Purpose:
 *     Prints an information dialog using the internal error shell as the
 *     dialog's parent.  The information message is printed using the format
 *     string and the variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     const char *szFormat : Format string for the information message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintInfo( const char *szFormat, ...)
{
    va_list	args;
    char	sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the information dialog using the internal error shell. */

    return( PrintInfoBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintInfo( wShell, szFormat, ...)
 *
 * Purpose:
 *     Prints an information dialog using the passed in shell as the dialog's
 *     parent.  The information message is printed using the format string and
 *     the variable arguments parameter.  This works just like like
 *     printf( szFormat, ... ).
 *
 * Parameters:
 *     Widget     wParent   : Parent widget for the information dialog.
 *     const char *szFormat : Format string for the information message.
 *     ...                  : Variable arguments that gives the argumetns
 *                            to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintInfo( Widget wShell, const char *szFormat, ...)
{
    va_list	args;
    char	sz[ MSGLEN];

/* Parse the variable arguments. */

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the information dialog using the passed in error shell. */

    return( PrintInfoBuffer( wShell, sz));
}


/*
 * Function: AmPrintResourceInfo( szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintInfo except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The internal error shell is used as the dialog's parent widget.
 *
 * Parameters:
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintResourceInfo( const char *szResource, ...)
{
    va_list      args;
    char         sz[ MSGLEN];
    const char   *szFormat;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the information dialog using the internal error shell. */

    return( PrintInfoBuffer( wErrorShell, sz));
}


/*
 * Function: AmPrintResourceError( wShell, szResource, ...)
 *
 * Purpose:
 *     Same as AmPrintInfo except that the format string is looked
 *     up in the resource database using the passed in resource string.
 *     The passed in widget is used as the dialog's parent widget.
 *
 * Parameters:
 *     Widget     wShell      : Parent widget for the error dialog.
 *     const char *szResource : Resource string that is used to look up
 *                              the format string in the database.
 *     ...                    : Variable arguments that gives the argumetns
 *                              to the format string.
 *
 * Returns:
 *     Widget : The dialog widget.
 */

Widget
AmPrintResourceInfo( Widget wShell, const char *szResource, ...)
{
    va_list      args;
    char         sz[ MSGLEN]; 
    const char   *szFormat;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

/* Create the information dialog using the passed in error shell. */

    return( PrintInfoBuffer( wShell, sz));
}
 

/**********************************************************************
 *
 * Wait dialog functions: 
 *
 **********************************************************************/

/* Create a wait dialog message... */

static AmWaitPtr
ShowWaitBuffer( Widget     wParent,
                const char *sz)
{
    return( AmMakeWaitDialog( wParent, sz, FALSE));
}

AmWaitPtr
AmShowWait( const char *szFormat,
          ...)
{
    va_list	args;
    char	sz[ MSGLEN];

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

    return( ShowWaitBuffer( wErrorShell, sz));
}

AmWaitPtr
AmShowWait( Widget     wShell,
          const char *szFormat,
          ...)
{
    va_list	args;
    char	sz[ MSGLEN];

    va_start( args, szFormat);
    vsprintf( sz, szFormat, args);
    va_end( args);

    return( ShowWaitBuffer( wShell, sz));
}

/* Same as PrintWait, but looks the string up in the resource database. */

AmWaitPtr
AmShowResourceWait( const char *szResource,
                  ...)
{
    va_list      args;
    char         sz[ MSGLEN];
    const char	 *szFormat;
 
/* get the string from the resource file */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);
    return( ShowWaitBuffer( wErrorShell, sz));
}


/* Same as AmShowResourceWait except with a supplied parent. */

AmWaitPtr
AmShowResourceWait( Widget     wShell,
                  const char *szResource, 
                  ...)
{
    va_list      args;
    char         sz[ MSGLEN]; 
    const char   *szFormat;
 
/* get the string from the resource file */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);
    return( ShowWaitBuffer( wShell, sz));
}

void
AmRemoveWait( AmWaitPtr wait)
{
    AmRemoveWaitDialog( wait);
}



/**********************************************************************
 *
 * Convenience functions: 
 *
 **********************************************************************/

/* set an error color for the widget */

void
AmSetErrorColor( Widget  w, 
               AnvilBoolean fError)
{
    const char  *szColor, *szResource;
    char 	sz[ RESLEN];
    int		istat;
    XColor	xCol, xColExact;
    Display	*pDisplay = XtDisplay( w);
    Colormap	colormap;

    XtVaGetValues( w, XmNcolormap, &colormap, NULL);

    if ( !fError) {
	szResource = AmGetResourceName( w);
	sprintf( sz, "%s.foreground", szResource);
	szColor = AmGetDefault( sz, "black");
	istat = XAllocNamedColor( pDisplay, colormap, 
			  ( char *) szColor, &xCol, &xColExact);
    } else {
        szColor = AmGetDefault( "error.color", "red");
	istat = XAllocNamedColor( pDisplay, colormap, 
			  ( char *) szColor, &xCol, &xColExact);
    }

    if ( istat != 0 ) {
        XtVaSetValues( w, XmNforeground, xCol.pixel, NULL);
    } else {
	AN_DEBUG (1,(" in AmSetErrorColor : XAllocNamedColor failed ... \n" ));
    }
}



/**********************************************************************
 *
 * Error detail reporting functions: 
 *
 **********************************************************************/

/* Callback to pop down the error detail dialog. */

/* ARGSUSED */
static void
HideErrorDetailDialogCallback( Widget    w,
                               XtPointer xpClientData,
                               XtPointer xpCallData)
{
    AmDialogClass *pwcDialog = ( AmDialogClass *) xpClientData;
    pwcDialog->Hide();
}

/* Callback to pop up the error detail dialog. */

/* ARGSUSED */
static void
ShowErrorDetailDialogCallback( Widget    w,
                               XtPointer xpClientData,
                               XtPointer xpCallData)
{
    AmDialogClass *pwcDialog = ( AmDialogClass *) xpClientData;
    pwcDialog->Show();
}

/* ARGSUSED */
static AmDialogClass *
MakeErrorDetailDialog( Widget w)
{
    AmDialogClass       *pwcDialog;
    AmScrolledTextClass *pwcText;
    int		      nRows;
    char	      *szLog;
    const char	      *szFile;

    szFile = AmGetLogFile();
    if ( szFile == NULL) {
	return( NULL);
    }

    szLog = AmReadFileString( szFile);
    if ( szLog == NULL) {
	return( NULL);
    }

    nRows = AmNumLines( szFile);

    pwcDialog = new AmDialogClass( w, "logDetail");
    pwcDialog->HideOnDelete();

    pwcText = new AmScrolledTextClass( pwcDialog->GetForm(), "logText");
    pwcText->AttachLeft();
    pwcText->AttachTop();
    pwcText->AttachRight();
    pwcText->AttachBottom();

    pwcText->ShowCursorPosition( FALSE);
    pwcText->AllowEditing( FALSE);
    pwcText->WrapWords( TRUE);
    pwcText->ShowHorizontalScrollBar( FALSE);

    pwcText->SetText( szLog);
    free( szLog);

    pwcDialog->AddButton( "okBtn", 
                          ( XtCallbackProc) HideErrorDetailDialogCallback, 
                          ( XtPointer) pwcDialog);

    return( pwcDialog);
}

/* Low-level function that takes a sz instead of varargs. */

static Widget
PrintLogErrorBuffer( Widget     wParent,
                     const char *sz)
{
    return( AmMakeLogErrorDialog( wParent, sz));
}

/* Same as print resource error, except that we add a callback for the 
   help button which brings up the log file */

Widget AmPrintLogError( Widget	 wShell,
		      const char *szResource,
		      ...)
{
    va_list     args;
    char        sz[ MSGLEN];
    const char  *szFormat;
    Widget      w, wDialog;
    AmDialogClass *pwcDialog;
 
/* Get the string from the resource file. */
 
    va_start( args, szResource);
    szFormat = AmGetDefault( szResource, szResource);
    vsprintf( sz, szFormat, args);
    va_end( args);

    wDialog = PrintLogErrorBuffer( wShell, sz);

/* Make the dialog that shows the error detail log. */

    pwcDialog = MakeErrorDetailDialog( wShell);

/* Attach a callback to the "help" button which has been changed
   to read "Details ...". */ 

    w = XmMessageBoxGetChild( wDialog, XmDIALOG_HELP_BUTTON);

    if ( pwcDialog != NULL) {
        XtAddCallback( w, XmNactivateCallback,
                       ( XtCallbackProc) ShowErrorDetailDialogCallback, 
		       pwcDialog);
    }

    return( wDialog);
}
