#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>

#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include <anvil/project.h>
#include <anvil/errcode.h>
#include <anvil/xcolor.h>
#include <anvil/motif.h>
#include <anvil/defaults.h>
#include <anvil/record.h>
#include <anvil/colorcell.h>
#include <anvil/xapp.h>
#include <anvil/visual.h>

static const int 	AmDEFAULT_HELP_TIMEOUT = 800;
static int 		AmHELP_TIMEOUT = 0;
static const char       *AmHELP_BACKGROUND = "#d2d2a5";
static const char       *AmHELP_FOREGROUND = "black";
static const char 	*AmHELP_FONT = "fixed";
static Display		*saveDisplay;
static Window		helpWin = 0;
static GC		helpGC = NULL;
static AmColorCellRec     fg, bg;

/* destroy the window, and remove this event handler */

static void removeHelpWindow( Widget    w,
			      Window	win,
			      XEvent	*,
			      AnvilBoolean	*)
{
    XDestroyWindow( XtDisplay( w), win);
    XtRemoveEventHandler( w, PointerMotionMask | EnterWindowMask |
    			     LeaveWindowMask | ButtonPressMask |
    			     ButtonReleaseMask | KeyPressMask, FALSE,
    			  ( XtEventHandler) removeHelpWindow, 
    			  ( XtPointer) win);
}

/* direct call to create a bubba-help window */

void AmHelpWindow( int		x, 
		 int		y,
		 Window		window,
		 const char	*string)
{
    Widget	w = XtWindowToWidget( saveDisplay, window);
    Window	rootWin = DefaultRootWindow( saveDisplay);

    if (( w != NULL) && XtIsSensitive( w)) {
	int			dir, ascent, descent, rootX, rootY;
	XCharStruct		overall;
	XGCValues		gcv;
	Window			child;
	XSetWindowAttributes	attr;

/* create a GC with the correct font.  We will set a foreground and background
   color later, but we'll have to integrate the color cube stuff. */

    	if ( helpGC == NULL) {
	    const char 	 *fontName;

	    fontName = AmGetDefault( "quickHelp.font", AmHELP_FONT);
	    bg = AmGetColorFromString( XtDisplay( w),
                     AmGetDefault( "quickHelp.background", AmHELP_BACKGROUND));
	    fg = AmGetColorFromString( XtDisplay( w),
	             AmGetDefault( "quickHelp.foreground", AmHELP_FOREGROUND));
	    gcv.font = XLoadFont( saveDisplay, fontName);
	    gcv.foreground = fg.pixel;
	    gcv.background = bg.pixel;
	    helpGC = XCreateGC( XtDisplay( w), rootWin,
	    			GCFont | GCForeground | GCBackground, &gcv);
	}

/* size the window by the size of the text string.  If there's no "quickHelp"
   string in the resource file, then bail */ 

    	XGetGCValues( saveDisplay, helpGC, GCFont, &gcv);
    	XQueryTextExtents( XtDisplay( w), gcv.font, string, 
    			   strlen( string), 
			   &dir, &ascent, &descent, &overall);

/* find the mouse's position to the AmFirst non-transient (menu) shell */

	XTranslateCoordinates( saveDisplay, window, rootWin, x, y,
			       &rootX, &rootY, &child);

/* override-redirect prevents the window manager from doing any decorations.
   SetTransientForHint is a little superfluous, but we want to make sure.
   Set saveunder, so the redraw is a little faster. We're small anyway */ 

        attr.override_redirect = TRUE;
        attr.background_pixel = bg.pixel;
        attr.save_under = TRUE;
    	helpWin = XCreateWindow( XtDisplay( w), rootWin,
			         rootX + 10, rootY + 10,
			         overall.width + 10, 
			         (( ascent + descent)*5)/4, 1,
                                 AmGetAppDepth(),
			         InputOutput,
                                 AmGetAppVisual(),
			         CWOverrideRedirect | CWBackPixel | CWSaveUnder,
			         &attr);
        XSetTransientForHint( XtDisplay( w), helpWin, 
        		      DefaultRootWindow( XtDisplay( w)));
        XSetWindowColormap( XtDisplay( w), helpWin, AmGetAppColormap());

/* map the window, flush the display to make sure it's mapped, and draw
   the help string */

        XMapWindow( XtDisplay( w), helpWin);
        XFlush( XtDisplay( w));
	XDrawString( XtDisplay( w), helpWin, helpGC, 5,
                     ( ascent * 9 + descent) / 8, string, strlen( string));

/* set up an event handler so that if the user does anything that the
   window gets destroyed */

	XtAddEventHandler( w, 
			   PointerMotionMask | EnterWindowMask |
			       LeaveWindowMask | ButtonPressMask |
			       ButtonReleaseMask | KeyPressMask, 
			   FALSE,
			   ( XtEventHandler) removeHelpWindow, 
			   ( XtPointer) helpWin);
    }
}

static XtIntervalId	helpId = 0;

/* timeout proc to bring up the help message */

static void AmHelpTimeOutProc( Window		window,	
			     XtIntervalId	*)
{
    Widget	w = XtWindowToWidget( saveDisplay, window);

    helpId = 0;

    if ( w != NULL) {
	Window		root, child;
	int		rootX, rootY, winX, winY;
	uint_t 		mask;
    	const char	*helpString, *resourceName;
    	char		helpRes[ RESLEN];

/* where's the pointer */

	XQueryPointer( XtDisplay( w), XtWindow( w), 
		       &root, &child, &rootX, &rootY, &winX, &winY,
		       &mask); 

/* get the string...if we're in show resource mode, then show the 
   resource string */

    	resourceName = AmGetResourceName( w);
    	sprintf( helpRes, "%s.quickHelp", resourceName);
    	helpString = AmGetDefault( helpRes, ( char *) NULL);
    	if ( helpString == NULL) {
    	    if ( !AmGetDefault( "quickHelp.showResources", ( AnvilBoolean) FALSE)) {
		return;
	    } else {
	    	helpString = resourceName;
	    	XStoreBytes( saveDisplay, resourceName, strlen( resourceName));
	    }
	}

/* kick off the help window */

	AmHelpWindow( winX, winY, window, helpString);
    }
}

/* help handler that unfortunately has to be inserted into the main loop
   after XtDispatchEvent.  Hopefully, we'll be able to use a workproc someday,
   but this'll have to do 

   Display:	connection to server.
   appContext:	Usually XtWidgetToApplicationContext( shell)
   event:	event returned from XtNextEvent 
*/

void AmHelpHandler( Display	*display,
		  XtAppContext	appContext,
		  XEvent	*event)
{
    XCrossingEvent	*crossEvent = ( XCrossingEvent *) event;
    XMotionEvent	*motionEvent = ( XMotionEvent *) event;

/* -1 turns off bubba help... */

    if ( AmHELP_TIMEOUT == -1) {
	return;
    }

/* if the user's done anything, remove the timeout */

    if (( event->type == MotionNotify) ||
    	( event->type == LeaveNotify) ||
    	( event->type == EnterNotify) ||
    	( event->type == ButtonPress) ||
    	( event->type == ButtonRelease) ||
    	( event->type == KeyPress) ||
    	( event->type == KeyRelease)) {
	if ( helpId != 0) {
	    XtRemoveTimeOut( helpId);
	    helpId = 0;
	}
    }

/* if the user's moved the mouse, or entered a window, then add the help
   timeout.  Use a default timeout, and see if the caller overrode it in
   the resource file */

    if ( event->type == MotionNotify) {
	saveDisplay = display;
        if ( AmHELP_TIMEOUT == 0) {
            AmHELP_TIMEOUT = AmGetDefault( "quickHelp.timeout",
            				 AmDEFAULT_HELP_TIMEOUT);
        }
    	helpId = XtAppAddTimeOut( appContext, AmHELP_TIMEOUT, 
    				  ( XtTimerCallbackProc) AmHelpTimeOutProc,
    				  ( XtPointer) motionEvent->window);
    } else if ( event->type == EnterNotify) {
	saveDisplay = display;
        if ( AmHELP_TIMEOUT == 0) {
            AmHELP_TIMEOUT = AmGetDefault( "quickHelp.timeout",
            				 AmDEFAULT_HELP_TIMEOUT);
        }
    	helpId = XtAppAddTimeOut( appContext, AmHELP_TIMEOUT, 
    				  ( XtTimerCallbackProc) AmHelpTimeOutProc,
    				  ( XtPointer) crossEvent->window);
    }
}

int
AmIsRunning()
{
    char        *homePath;
    char        lockFile[ MAXPATHLEN], realPath[ MAXPATHLEN];
    int         ii, pid;
    int         flag;

    flag = 0;

    homePath = getenv( "HOME");
    if ( homePath == NULL) {
        return( 0);
    }

    sprintf( lockFile, "%s/.netscape/lock", homePath);

    if ( realpath( lockFile, realPath) != NULL) {
        return( 0);
    }

    for ( ii = 0; realPath[ ii] != '\0'; ii++) {
        if ( realPath[ ii] == ':') {
            pid = atoi( &realPath[ ii + 1]);
            break;
        }
    }

    if ( realPath[ ii] == '\0') {
        return( 0);
    }

    return( 1);
}

void
AmHelpCback( Widget wButton, XtPointer, XtPointer)
{
    const char *helpString, *resourceName;
    char       helpRes[ RESLEN], buffer[ TEXTLEN];

    resourceName = AmGetResourceName( wButton);
    sprintf( helpRes, "%s.help", resourceName);
    helpString = AmGetDefault( helpRes, ( char *) NULL);

    if ( helpString) {
        if ( AmIsRunning()) {
            sprintf( buffer, "netscape -raise -remote \"openURL( %s, 'NetMentor Help' )\" &", helpString);
        } else {
            sprintf( buffer, "netscape %s &", helpString);
        }

        system( buffer);
    }
}
