#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <Xm/Xm.h>
#include <anvil/errlib.h>
#include <anvil/motif.h>
#include <anvil/list.h>
#include <anvil/state_hand.h>
#include <anvil/user_data.h>


static void
fnFreeHandler( AmHandlerDataPtr pHandlerData)
{
    if ( pHandlerData != NULL) {

        AmJustFreeLst( pHandlerData->handlerList);
        FREE( pHandlerData);

    }
}


/* Wrapper functions for turning widgets on and off again */

void
AmTurnOnWidget( Widget w)
{
    XtSetSensitive( w, TRUE);
}


void
AmTurnOffWidget( Widget	w)
{
    XtSetSensitive( w, FALSE);
}


/* If the launch widget gets destroyed, turn off the handler and set the
   launch widget to NULL */

/* ARGSUSED */
static void
turnOffHandler( Widget       w,
                AmHandlerPtr pHandler,
                XtPointer    xpCallData)
{
   pHandler->launchWidget = NULL;
   AmRemoveHandler( pHandler->handlerWidget, 
                    pHandler->event_mask,
                    pHandler->AmHandlerFn,
                    pHandler->clientData);
}


/* ARGSUSED */
static void
removeHandlerData( Widget       w,
                   AmHandlerPtr pHandler,
                   XtPointer    xpCallData)
{

    if ( pHandler->launchWidget) {
        XtRemoveCallback( pHandler->launchWidget, XmNdestroyCallback,
                          ( XtCallbackProc) turnOffHandler,
                          ( XtPointer) pHandler);
    }

    FREE( pHandler);
}


/*
 * Add an event handler to a widget, and store it in the handler stack. 
 *  Suspend or cancel the current handler depending on the relative
 *  priority.
 */

AmErrorCode
AmAddHandler( Widget           wLaunchWidget,
              Widget           w,
              EventMask        xEventMask,
              XtEventHandler   AmHandlerFn,
              XtPointer        xpClientData,
              unsigned int     iPriority,
              AmStateHandlerFn fnRemoveFn,
              AmStateHandlerFn fnSuspendFn,
              AmStateHandlerFn fnRestartFn)
{
    AmHandlerDataPtr pHandlerData;
    AmHandlerPtr     pHandler, pHandlerNew;
    AmListPtr        lstTemp;

/* The event handler information is stored in the widget. */

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( w, HANDLER_STRING);
    if ( pHandlerData == NULL) {

        pHandlerData = ( AmHandlerDataPtr) malloc( sizeof( AmHandlerDataRec));
        if ( pHandlerData == NULL) {
            DO_RETURN_ERROR( AmMALLOC_FAILED);
        }

        pHandlerData->magicNo     = HANDLER_MAGIC;
        pHandlerData->handlerList = NULL;

        if ( AmAddUserData( w, HANDLER_STRING, pHandlerData,
                            ( AmItemFn) fnFreeHandler) != AmOK) {
            fnFreeHandler( pHandlerData);
            return( ANVIL_ERROR_CODE);
        }
    }

/*
 * If there's an event handler currently being executed check to see if
 *   the handler has an equal or higher priority than the current handler.
 *   If so, then suspend the current handler, and put the new one on the
 *   stack.  If not, then remove the current handler, and turn its launch
 *   widget back on.
 */

    if ( pHandlerData->handlerList != NULL) {

        pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
        XtRemoveEventHandler( w, pHandler->event_mask, FALSE,
                              pHandler->AmHandlerFn,
                              pHandler->clientData);

        if ( iPriority >= pHandler->priority) {

        /*
         * Remove all the handlers that have lower priority than the handler
         *   that is getting added.
         */

	    lstTemp = pHandlerData->handlerList;
	    while ( lstTemp != NULL) {

                pHandler = ( AmHandlerPtr) lstTemp->data;
                if ( iPriority < pHandler->priority) {
                    break;
                }

                XtRemoveCallback( pHandler->handlerWidget, XmNdestroyCallback,
                                  ( XtCallbackProc) removeHandlerData,
                                  ( XtPointer) pHandler);

	        if ( pHandler->launchWidget != NULL) {

                    XtRemoveCallback( pHandler->launchWidget,
                                      XmNdestroyCallback,
                                      ( XtCallbackProc) turnOffHandler,
                                      ( XtPointer) pHandler);
		    AmTurnOnWidget( pHandler->launchWidget);

	        }

	        pHandlerData->handlerList =
                    AmFreeHead( pHandlerData->handlerList);
		lstTemp = pHandlerData->handlerList;

                /* Call the user defined remove function if needed. */

                if ( pHandler->removeFn != NULL) {
                    pHandler->removeFn( pHandler->clientData);
                }

	        FREE( pHandler);
            } 
	} else {

        /* Suspend the current handler. */

            if ( pHandler->suspendFn != NULL) {
                pHandler->suspendFn( pHandler->clientData);
            }
        }
    }

/* store the handler information, put it on the list, and turn it on. */

    pHandlerNew = ( AmHandlerPtr) malloc( sizeof( AmHandlerRec));
    pHandlerNew->launchWidget  = wLaunchWidget;
    pHandlerNew->handlerWidget = w;
    pHandlerNew->event_mask    = xEventMask;
    pHandlerNew->AmHandlerFn   = AmHandlerFn;
    pHandlerNew->clientData    = xpClientData;
    pHandlerNew->priority      = iPriority;
    pHandlerNew->removeFn      = fnRemoveFn;
    pHandlerNew->suspendFn     = fnSuspendFn;
    pHandlerNew->restartFn     = fnRestartFn;

    pHandlerData->handlerList = AmInsertNodeLst( pHandlerNew,
                                                 pHandlerData->handlerList);

/*
 * If there's a launch widget, then turn it off while the handler's 
 *   executing.
 */

    if ( wLaunchWidget != NULL) {
	XtAddCallback( pHandlerNew->launchWidget, XmNdestroyCallback,
		       ( XtCallbackProc) turnOffHandler,
                       ( XtPointer) pHandlerNew);
	AmTurnOffWidget( wLaunchWidget);
    }

    XtAddCallback( pHandlerNew->handlerWidget, XmNdestroyCallback,
		   ( XtCallbackProc) removeHandlerData,
                   ( XtPointer) pHandlerNew);

    XtAddEventHandler( w, xEventMask, FALSE, AmHandlerFn, xpClientData);

    return( AmOK);
}

/* Remove the event handler from the current stack */

AmErrorCode
AmRemoveHandler( Widget         w,
                 EventMask      xEventMask,
                 XtEventHandler AmHandlerFn,
                 XtPointer      xpClientData)
{
    AmHandlerDataPtr pHandlerData;
    AmHandlerPtr     pHandler, pHandlerNew;
    AmListPtr        lstTemp;

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( w, HANDLER_STRING);

/* If there is no handler data, then the handler wasn't installed, was it? */

    if ( pHandlerData == NULL) {
	return( ANVIL_ERROR_CODE);
    } else if ( pHandlerData->magicNo != HANDLER_MAGIC) {
	DO_RETURN_ERROR( AmUNKNOWN_USER_DATA);
    }

/* Search for the handler in the handler list, and turn off the handler. */ 

    for ( lstTemp = pHandlerData->handlerList;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

	pHandler = ( AmHandlerPtr) lstTemp->data;
	if (( pHandler->event_mask == xEventMask) &&
	    ( pHandler->AmHandlerFn == AmHandlerFn) &&
	    ( pHandler->clientData == xpClientData)) {

	    break;
	}
    }

    if ( lstTemp != NULL) {

	XtRemoveCallback( pHandler->handlerWidget, XmNdestroyCallback,
	                  ( XtCallbackProc) removeHandlerData,
                          ( XtPointer) pHandler);

    /*
     * If the handler was launched by a widget, turn it back on, then
     *   delete the handler from the handler list.
     */

	if ( pHandler->launchWidget != NULL) {

	    XtRemoveCallback( pHandler->launchWidget, XmNdestroyCallback,
                             ( XtCallbackProc) turnOffHandler,
                             ( XtPointer) pHandler);
	    AmTurnOnWidget( pHandler->launchWidget);

	}

    /* Call the user defined remove function if there is one. */

        if ( pHandler->removeFn != NULL) {
            pHandler->removeFn( pHandler->clientData);
        }

	if ( lstTemp == pHandlerData->handlerList) {

        /* 
         * The handler being removed is currently active, which means the
         *    X handler needs to be removed, 
         */

            XtRemoveEventHandler( w, xEventMask, FALSE, AmHandlerFn,
                                  xpClientData);

            if ( pHandlerData->handlerList->next) {
	        pHandlerNew =
                    ( AmHandlerPtr) pHandlerData->handlerList->next->data;

	        XtAddEventHandler( w, pHandlerNew->event_mask, FALSE,
                                   pHandlerNew->AmHandlerFn,
                                   pHandlerNew->clientData);

                if ( pHandlerNew->restartFn != NULL) {
                    pHandlerNew->restartFn( pHandlerNew->clientData);
                }
            }
	}

	pHandlerData->handlerList = 
	    AmDeleteItemLst( pHandler, pHandlerData->handlerList, 
			     ( AmSearchFn) AmEquals, ( AmItemFn) free);

        if ( pHandlerData->handlerList == NULL) {
            AmRemoveUserData( w, HANDLER_STRING);
            fnFreeHandler( pHandlerData);
        }

	return( AmOK);

    } else {
	return( AmMISSING_EVENT_HANDLER);
    }
}


AmErrorCode
AmFlushHandler( Widget       w,
                unsigned int iPriority)
{
    AmHandlerDataPtr pHandlerData;
    AmHandlerPtr     pHandler;
    AmListPtr        lstTemp;

/* the event handler information is stored in the widget */

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( w, HANDLER_STRING);
    if ( pHandlerData == NULL) {	
        return( ANVIL_ERROR_CODE);
    } else if ( pHandlerData->magicNo != HANDLER_MAGIC) {
	DO_RETURN_ERROR( AmUNKNOWN_USER_DATA);
    } 

/*
 * If there's an event handler currently being executed check to see if
 *   the handler has a higher priority than the flush priority.  If not,
 *   remove the current handler and any other handlers that have a lower 
 *   priority than the flush priority.
 */

    if ( pHandlerData->handlerList != NULL) {
	pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
	if ( iPriority >= pHandler->priority) {

	    XtRemoveEventHandler( w, pHandler->event_mask, FALSE,
                                  pHandler->AmHandlerFn, pHandler->clientData);

            /*
             * Remove all the handlers that have lower priority than the
             *   flush priority.
             */

            lstTemp = pHandlerData->handlerList;
            while( lstTemp != NULL) {

                pHandler = ( AmHandlerPtr) lstTemp->data;

                if ( iPriority < pHandler->priority) {
                    break;
                }

                XtRemoveCallback( pHandler->handlerWidget, XmNdestroyCallback,
	                          ( XtCallbackProc) removeHandlerData,
                                  ( XtPointer) pHandler);

            /* Turn on the launch widget if needed. */

	        if ( pHandler->launchWidget != NULL) {

                    XtRemoveCallback( pHandler->launchWidget,
                                      XmNdestroyCallback,
                                      ( XtCallbackProc) turnOffHandler,
                                      ( XtPointer) pHandler);
		    AmTurnOnWidget( pHandler->launchWidget);

	        }
              
	        pHandlerData->handlerList =
                    AmFreeHead( pHandlerData->handlerList);
                lstTemp = pHandlerData->handlerList;

            /* Call the user defined remove function if needed. */

                if ( pHandler->removeFn != NULL) {
                    pHandler->removeFn( pHandler->clientData);
                }

	        FREE( pHandler);
            }

        /* Start a new handler if there is one. */

            if ( pHandlerData->handlerList) {
	        pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
	        XtAddEventHandler( w, pHandler->event_mask, FALSE,
                    pHandler->AmHandlerFn, pHandler->clientData);

                if ( pHandler->restartFn != NULL) {
                    pHandler->restartFn( pHandler->clientData);
                }
            }
	}

        if ( pHandlerData->handlerList == NULL) {
            AmRemoveUserData( w, HANDLER_STRING);
            fnFreeHandler( pHandlerData);
        }

    }

    return( AmOK);
}
