#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <Xm/Xm.h>
#include <Xm/RowColumn.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/List.h>
#include <Xm/FileSB.h>
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/Frame.h>
#include <Xm/Separator.h>

#include <anvil/motif.h>
#include <anvil/errcode.h>
#include <anvil/utility.h>

/*****************************************************************************
 *
 * Function: AmMakeActionArea( parent, actionDefs, numDefs, action, makeSep)
 *
 *     Widget       parent     - The parent of the action area.
 *     AmActionDefPtr actionDefs - Description of the actions to be added to
 *                               the action area.
 *     int          numDefs    - How many definitions there are in theo
 *                               description.
 *     AmActionPtr    action     - Structure used to return values, i.e. 
 *                               the created widgets.
 *     AnvilBoolean      makeSep    - Whether or not to create a separator
 *                               above the buttons (TRUE means creating
 *                               a separator).
 *
 * Description:
 *
 *    Creates an action area which is a form with a separator and buttons
 *    below the separator.  Which buttons to add to the form is determined
 *    by an array of action definitions.  An action definition consists of a
 *    name, a callback function, and client data.  If the callback function
 *    is NULL no callback is added for that specific button.  A passed-in
 *    structure is filled with the created widgets so that some of the
 *    values can be changed externally if needed.
 * 
 *    The form that contains all the widgets is returned.
 *
 *****************************************************************************/

Widget
AmMakeActionArea( Widget       parent,
                AmActionDefPtr actionDefs,
                unsigned     numActions,
                AmActionPtr    action,
                AnvilBoolean      makeSep)
{
    Widget    form, separator, buttonArea, button;
    Dimension width; 
    unsigned  ii;

/* Sanity check. */

    if ( actionDefs == NULL) {
        ANVIL_ERROR_CODE = AmBAD_ARGUMENT;
        return( NULL);
    }

    if ( numActions > MAX_ACTIONS) {
        numActions = MAX_ACTIONS;
    }


/*
 * It is assumed that the parent is a form.  Since most dialogs
 *    have a form as a manager widget.
 */
   
    form = XtVaCreateManagedWidget( "actionForm",
                xmFormWidgetClass, parent,
                StickLeft, StickRight, StickBottom,
                NULL);

    buttonArea = XtVaCreateManagedWidget( "buttonArea",
                     xmFormWidgetClass, form,
                     StickLeft, StickRight, StickBottom,
                     XmNfractionBase, numActions * 2,
                     NULL);

    if ( makeSep == TRUE) {
        separator = XtVaCreateManagedWidget( "separator",
                        xmSeparatorWidgetClass, form,
                        StickLeft, StickTop, StickRight,
                        NULL);
        XtVaSetValues( buttonArea,
            BelowOf( separator),
            NULL);
    } else {
        separator = NULL;
        XtVaSetValues( buttonArea,
            StickTop,
            NULL);
    }

/*
 * Create the action buttons.  The buttons are attached to the right hand
 *   side.
 */

    if ( action != NULL) {
        action->form = form; 
        action->separator = separator; 
        action->buttonArea = buttonArea;
    }

    for ( ii = 0; ii < numActions; ii++) {
        button = XtVaCreateManagedWidget( actionDefs[ ii].name,
                     xmPushButtonWidgetClass, buttonArea,
                     StickTop, StickBottom, 
                     XmNleftAttachment, XmATTACH_POSITION,
                     XmNleftPosition, 1 + ( ii * 2),
                     NULL);
        XtVaGetValues( button,
            XmNwidth, &width,
            NULL);
        XtVaSetValues( button,
            XmNleftOffset, ( ( -1) * width) / 2,
            NULL);

    /* Setup the callback for the button. */

        if ( actionDefs[ ii].AmCallbackFn != NULL) {
            XtAddCallback( button, XmNactivateCallback,
                actionDefs[ ii].AmCallbackFn, actionDefs[ ii].clientData);
        }

        if ( action != NULL) {
            action->buttons[ ii] = button;
        }
    }

    return( form);
}


/*****************************************************************************
 *
 * Function: AmMakeFrame( parent, name, label, frame)
 *
 *     Widget    parent  - The parent of the frame.
 *     char      *name   - The name to give to the frame.
 *     AnvilBoolean   label   - A boolean that indicates whether a label 
 *                         gadget is to be added or not.
 *     AmFramePtr  frame   - Structure used to return values, i.e. the 
 *                         created widgets.
 *
 * Description:
 *
 *    Creates a frame with an optional title gadget.  The name of the frame
 *    widget is taken from the passed-in name parameter.  If a title gadget
 *    is to be created it gets the name "label" associated with it.  The
 *    passed-in frame structure is filled to contain all the created widgets.
 *
 *    The frame widget is returned.
 *
 *****************************************************************************/

Widget
AmMakeFrame( Widget    parent,
           char      *name,
           AnvilBoolean   label,
           AmFramePtr  frame)
{
    Widget frameWidget, labelWidget;

/* Sanity check. */

    if ( name == NULL) {
        return( NULL);
    }

    frameWidget = XtVaCreateManagedWidget( name,
                      xmFrameWidgetClass, parent,
                      NULL);

    if ( label == TRUE) {
        labelWidget = XtVaCreateManagedWidget( "label",
                          xmLabelGadgetClass, frameWidget,
                          XmNchildType, XmFRAME_TITLE_CHILD,
                          XmNchildVerticalAlignment, XmALIGNMENT_CENTER,
                          NULL);
    } else {
        labelWidget = NULL;
    }

    if ( frame != NULL) {
        frame->frame = frameWidget;
        frame->label = labelWidget;
    }

    return( frameWidget);
}
