#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <malloc.h>
#include <string.h>
#include <windows.h>
#include <anvil/dispatch.h>
#include <anvil/tsearch.h>
#include <anvil/list.h>

const int MAXEVENTS = 16;

/* top level node for list of handlers for a window */

typedef struct ntWindowHandlerList_ {
    HWND	hWnd;		 /* window handler list is assigned to */
    AmListPtr	handlerList;	 /* list of handlers */
} ntWindowHandlerListRec, *ntWindowHandlerListPtr;

typedef struct msgParam_ {
    UINT	iMsg;		/* message number */
    WPARAM	wParam;		/* match LOWORD if iMsg == WM_COMMAND */
} msgParamRec, *msgParamPtr;

typedef struct ntHandler_ {
    HWND	hWndChild;		/* child window to match for lParam */
    msgParamRec	msgVec[ MAXEVENTS];	/* match LOWORD if iMsg==WM_COMMAND */
    int		nMsg;			/* length of message and params vecs */
    AmNtHandlerFn	AmHandlerFn;		/* handler function */
    void	*clientData;		/* handler client data */
    AnvilBoolean	deleted;		/* deleted flag */
} ntHandlerRec, *ntHandlerPtr;

void *handlerRoot = NULL;

/* comparator function for window handler tree searching */

static int 
matchHandlerWindow( ntWindowHandlerListPtr	a,
		    ntWindowHandlerListPtr	b)
{
    return(( uchar_t *) a->hWnd - ( uchar_t *) b->hWnd);
}

/* for sorting messages by message code */

static int 
byMsg( msgParamPtr	a,
       msgParamPtr	b)
{
    return( a->iMsg - b->iMsg);
}

/* copy the vector of messages and wParams into the handler structure and
   sort them for correct comparisons */

static void 
copyMsgs( ntHandlerPtr	handler,
	  UINT		*pMsgVec,
	  WPARAM	*wParamVec,
	  int		nMsg)
{
    int	ii;

    handler->nMsg = nMsg;
    for ( ii = 0; ii < nMsg; ii++) {
    	handler->msgVec[ ii].iMsg = pMsgVec[ ii];
	if ( wParamVec != NULL) {
	    handler->msgVec[ ii].wParam = wParamVec[ ii];
	} else {
	    handler->msgVec[ ii].wParam = 0;
	}
    }
    qsort( handler->msgVec, handler->nMsg, sizeof( msgParamRec),
    	 ( AmSearchFn) byMsg);
}

/* comparison function to see if a single event matches an event handler */

static AnvilBoolean 
matchHandlerEvent( ntHandlerPtr	a,
		   UINT		iMsg,
		   WPARAM	wParam,
		   LPARAM	lParam)
{
    int	ii;

    for ( ii = 0; ii < a->nMsg; ii++) {

/* return TRUE if the event ID's match, and if they're both WM_COMMAND, that
   the wParam values match */

    	if ( iMsg == a->msgVec[ ii].iMsg) {
	    if ( iMsg == WM_COMMAND) {
		if ( LOWORD( a->msgVec[ ii].wParam) == LOWORD( wParam)) {
		    return( TRUE);
		} 
	    } else {
		if ( a->hWndChild != NULL) {
		    return( a->hWndChild == ( HWND) lParam);
		}
	    	return( TRUE);
	    }
	}
    }
    return( FALSE);
}

/* for comparing a pair of eventh handlers event code lists */

static AnvilBoolean
matchHandlers( ntHandlerPtr	a,
	       ntHandlerPtr	b)
{
    int	ii;

    if ( a->nMsg != b->nMsg) {
	return( FALSE);
    } 
    for ( ii = 0; ii < a->nMsg; ii++) {
    	if ( a->msgVec[ ii].iMsg == b->msgVec[ ii].iMsg) {
	    if ( a->msgVec[ ii].iMsg == WM_COMMAND) {
		if ( LOWORD( a->msgVec[ ii].wParam) != 
		     LOWORD( b->msgVec[ ii].wParam)) {
		    return( FALSE);
		}
	    }
	} else {
	    return( FALSE);
	}
    }
    if (( a->hWndChild != NULL) || ( b->hWndChild != NULL)) {
	return( a->hWndChild == b->hWndChild);
    }
    return( TRUE);
}

/* comparator function for handler list searching by event, handler and
   client data */

static int
matchFullHandler( ntHandlerPtr	a,
	          ntHandlerPtr	b)
{
    return( matchHandlers( a, b) && 
    	    ( a->AmHandlerFn == b->AmHandlerFn) &&
	    ( a->clientData == b->clientData));
}

/*************************************************************************
 *
 * Description:
 *	Add an event handler to hWnd, which will be called if 
 *	dispatchEvent is called with matching iMsg and wParam 
 *	wParam is matched if iMsg == WM_COMMAND.  The event handler
 *	is inserted at the start of the event list, so it will be
 *	called ahead of previously installed handlers.  
 * Arguments:
 *	hWnd - window to attach handler to.
 *	iMsg - message to match
 *	wParam - wParam to match
 *	handler - handler to call
 *	clientData - client data to pass to handler (must be malloced)
 *
 *************************************************************************/

AmErrorCode 
AmInsertHandler( HWND			hWnd,
				 UINT			iMsg,
				 WPARAM			wParam,
	             AmNtHandlerFn	handler,
	             void			*clientData)
{
    return( AmInsertHandler( hWnd, &iMsg, &wParam, 1, handler, clientData));
}

AmErrorCode 
AmInsertHandler( HWND			hWnd,
				 UINT			*pMsgVec,
				 WPARAM			*wParamVec,
				 int			nMsg,
				 AmNtHandlerFn	handler,
				 void			*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr			handlerNode;

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);

/* if not, create the node and insert it into the tree. */

    if ( result == NULL) {
    	if (( node = ( ntWindowHandlerListPtr) 
			    malloc( sizeof( ntWindowHandlerListRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	node->hWnd = hWnd;
	node->handlerList = NULL;
	tsearch( node, &handlerRoot, ( AmSearchFn) matchHandlerWindow);
    } else {
    	node = *result;
    }

/* add the new node to the handler list */

    if (( handlerNode = ( ntHandlerPtr) 
			    malloc( sizeof( ntHandlerRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    copyMsgs( handlerNode, pMsgVec, wParamVec, nMsg);
    handlerNode->AmHandlerFn = handler;
    handlerNode->clientData = clientData;
    handlerNode->deleted = FALSE;
    handlerNode->hWndChild = NULL;
    node->handlerList = AmInsertNodeLst( handlerNode, node->handlerList);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *	Add an event handler to hWnd, which will be called if 
 *	dispatchEvent is called with matching iMsg and lParam = hWndChild.
 *	The event handler is inserted at the start of the event list, 
 *	so it will be called ahead of previously installed handlers.  
 * Arguments:
 *	hWnd - window to attach handler to.
 *	hWndChild - child window to match.
 *	iMsg - message to match
 *	wParam - wParam to match
 *	handler - handler to call
 *	clientData - client data to pass to handler (must be malloced)
 *
 *************************************************************************/

AmErrorCode
AmInsertHandler( HWND		hWnd,
	         HWND		hWndChild,
	         UINT		iMsg,
	         WPARAM		wParam,
	         AmNtHandlerFn	handler,
	         void		*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		handlerNode;

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);

/* if not, create the node and insert it into the tree. */

    if ( result == NULL) {
    	if (( node = ( ntWindowHandlerListPtr) 
			    malloc( sizeof( ntWindowHandlerListRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	node->hWnd = hWnd;
	node->handlerList = NULL;
	tsearch( node, &handlerRoot, ( AmSearchFn) matchHandlerWindow);
    } else {
    	node = *result;
    }

/* add the new node to the handler list */

    if (( handlerNode = ( ntHandlerPtr) 
			    malloc( sizeof( ntHandlerRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    copyMsgs( handlerNode, &iMsg, &wParam, 1);
    handlerNode->AmHandlerFn = handler;
    handlerNode->clientData = clientData;
    handlerNode->deleted = FALSE;
    handlerNode->hWndChild = hWndChild;
    node->handlerList = AmInsertNodeLst( handlerNode, node->handlerList);
    return( AmOK);

}

/*************************************************************************
 *
 * Description:
 *	Add an event handler to hWnd, which will be called if 
 *	dispatchEvent is called with matching iMsg and wParam 
 *	wParam is matched if iMsg == WM_COMMAND.  The event handler
 *	is inserted at the end of the event list, so it will be
 *	called after previously installed handlers.
 * Arguments:
 *	hWnd - window to attach handler to.
 *	iMsg - message to match
 *	wParam - wParam to match
 *	handler - handler to call
 *	clientData - client data to pass to handler (must be malloced)
 * Returns:
 *	AmOK if everythings happy, system error otherwise.
 *
 *************************************************************************/

AmErrorCode 
AmAppendHandler( HWND			hWnd,
	             UINT			iMsg,
				 WPARAM			wParam,
				 AmNtHandlerFn	handler,
				 void			*clientData)
{
    return( AmAppendHandler( hWnd, &iMsg, &wParam, 1, handler, clientData));
}

AmErrorCode 
AmAppendHandler( HWND		hWnd,
	         UINT		*pMsgVec,
	         WPARAM		*wParamVec,
	         int		nMsg,
	         AmNtHandlerFn	handler,
	         void		*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		handlerNode;

    if ( nMsg > MAXEVENTS) {
    	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);

/* if not, create the node and insert it into the tree. */

    if ( result == NULL) {
    	if (( node = ( ntWindowHandlerListPtr) 
			    malloc( sizeof( ntWindowHandlerListRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	node->hWnd = hWnd;
	node->handlerList = NULL;
	tsearch( node, &handlerRoot, ( AmSearchFn) matchHandlerWindow);
    } else {
    	node = *result;
    }

/* add the new node to the handler list */

    if (( handlerNode = ( ntHandlerPtr) 
			    malloc( sizeof( ntHandlerRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    copyMsgs( handlerNode, pMsgVec, wParamVec, nMsg);
    handlerNode->AmHandlerFn = handler;
    handlerNode->clientData = clientData;
    handlerNode->deleted = FALSE;
    handlerNode->hWndChild = NULL;
    node->handlerList = AmInsertNodeAtEndLst( handlerNode, node->handlerList);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *	Append an event handler to hWndParent, which will be called if 
 *	dispatchEvent is called with matching iMsg and lParam = hWnd.
 *	The event handler is appended to the end of the event list, 
 *	so it will be called after of previously installed handlers.  
 * Arguments:
 *	hWnd - window to attach handler to.
 *	hWndChild - child window to match.
 *	iMsg - message to match
 *	wParam - wParam to match
 *	handler - handler to call
 *	clientData - client data to pass to handler (must be malloced)
 *
 *************************************************************************/


AmErrorCode
AmAppendHandler( HWND		hWnd,
	         HWND		hWndChild,
	         UINT		iMsg,
	         WPARAM		wParam,
	         AmNtHandlerFn	handler,
	         void		*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		handlerNode;

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);

/* if not, create the node and insert it into the tree. */

    if ( result == NULL) {
    	if (( node = ( ntWindowHandlerListPtr) 
			    malloc( sizeof( ntWindowHandlerListRec))) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	node->hWnd = hWnd;
	node->handlerList = NULL;
	tsearch( node, &handlerRoot, ( AmSearchFn) matchHandlerWindow);
    } else {
    	node = *result;
    }

/* add the new node to the handler list */

    if (( handlerNode = ( ntHandlerPtr) 
			    malloc( sizeof( ntHandlerRec))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    copyMsgs( handlerNode, &iMsg, &wParam, 1);
    handlerNode->AmHandlerFn = handler;
    handlerNode->clientData = clientData;
    handlerNode->deleted = FALSE;
    handlerNode->hWndChild = hWndChild;
    node->handlerList = AmInsertNodeLst( handlerNode, node->handlerList);
    return( AmOK);

}

/*************************************************************************
 *
 * Description:
 *	Remove an event handler from hWnd, 
 *	wParam is matched if iMsg == WM_COMMAND
 * Arguments:
 *	hWnd - window to attach handler to.
 *	iMsg - message to match
 *	wParam - wParam to match
 *	handler - handler to remoe
 *	clientData - client data passed to handler 
 *
 *************************************************************************/

AmErrorCode 
AmDeleteHandler( HWND			hWnd,
	             UINT			iMsg,
	             WPARAM			wParam,
	             AmNtHandlerFn	handler,
	             void			*clientData)
{
    return( AmDeleteHandler( hWnd, &iMsg, &wParam, 1, handler, clientData));
}

AmErrorCode 
AmDeleteHandler( HWND		hWnd,
		 UINT		*pMsgVec,
		 WPARAM		*wParamVec,
		 int		nMsg,
		 AmNtHandlerFn	handler,
		 void		*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		candNode;
    ntHandlerRec		matchNode;
    AmListPtr			ptr;
    AnvilBoolean			matchFound;

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);
    if ( result == NULL) {
    	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    node = *result;

/* mark all matching nodes in the list as deleted, and if any are found,
   set the flag */

    copyMsgs( &matchNode, pMsgVec, wParamVec, nMsg);
    matchNode.hWndChild = NULL;
    matchNode.AmHandlerFn = handler;
    matchNode.clientData = clientData;
    matchFound = FALSE;
    for ( ptr = node->handlerList; ptr != NULL; ptr = ptr->next) {
    	candNode = ( ntHandlerPtr) ptr->data;
	if ( matchFullHandler( &matchNode, candNode)) {
	    candNode->deleted = TRUE;
	    matchFound = TRUE;
	}
    }
    if ( matchFound) {
    	return( AmOK);
    } else {
    	return( AmBAD_ARGUMENT);
    }
}


AmErrorCode 
AmDeleteHandler( HWND		hWnd,
	         HWND		hWndChild,
	         UINT		iMsg,
	         WPARAM		wParam,
	         AmNtHandlerFn	handler,
	         void		*clientData)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		candNode;
    ntHandlerRec		matchNode;
    AmListPtr			ptr;
    AnvilBoolean			matchFound;

/* have any handlers been assigned to this window? */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);
    if ( result == NULL) {
    	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    node = *result;

/* mark all matching nodes in the list as deleted, and if any are found,
   set the flag */

    copyMsgs( &matchNode, &iMsg, &wParam, 1);
    matchNode.hWndChild = hWndChild;
    matchNode.AmHandlerFn = handler;
    matchNode.clientData = clientData;
    matchFound = FALSE;
    for ( ptr = node->handlerList; ptr != NULL; ptr = ptr->next) {
    	candNode = ( ntHandlerPtr) ptr->data;
	if ( matchFullHandler( &matchNode, candNode)) {
	    candNode->deleted = TRUE;
	    matchFound = TRUE;
	}
    }
    if ( matchFound) {
    	return( AmOK);
    } else {
    	return( AmBAD_ARGUMENT);
    }
}
/*************************************************************************
 *
 * Description:
 *	Event dispatcher.  First, find the list of events associated with
 *	the window argument.  If there are none, then call DefWindowProc().
 *	If there are any WM_CREATE or WM_NCCREATE handlers, call them
 *	AmFirst, then mark them as deleted.  This is irregardless of the
 *	event, because the WM_CREATE message is generated while the
 *	window is still NULL, and before any InsertHandler() calls can
 *	be made, so we just call the create handlers on the AmFirst 
 *	dispatch to the window, then delete them.  The function depends
 *	on a call counter to ensure re-entrancy.  Events are deleted 
 *	from the window's event list by marking them, then removing
 *	them from the list at the top call level. A WM_DESTROY event
 *	marks all event handlers as deleted.  If no handlers are executed,
 *	the window's DefWindowProc() is called.
 * Arguments:
 *	hWnd - window 
 *	iMsg - message to match
 *	wParam - wParam to match
 *	lParam - passed to handler 
 *
 *************************************************************************/

LRESULT CALLBACK
AmDispatchEvent( HWND	hWnd,
	         UINT	iMsg,
	         WPARAM	wParam,
	         LPARAM	lParam)
{
    ntWindowHandlerListRec 	target;
    ntWindowHandlerListPtr 	*result, node;
    ntHandlerPtr		candNode;
    AmListPtr			ptr, deadNode, prevNode;
    AnvilBoolean			foundFlag;
    static int			callCount = 0;

/* have any handlers been assigned to this window? If not, then call
   DefWindowProc() and exit */

    target.hWnd = hWnd;
    result = ( ntWindowHandlerListPtr *)
    	tfind( &target, ( void * const *) &handlerRoot,
	       ( AmSearchFn) matchHandlerWindow);
    if ( result == NULL) {
	return( DefWindowProc( hWnd, iMsg, wParam, lParam));
    }
    node = *result;

/* increment the call count to support reentrancy, since the dispatcher
   can be called from an event handler (via WM_DESTROY) */
 
    callCount++;

/* if there are any WM_NCCREATE or WM_CREATE messages, then process them
   immediately, and delete the handlers, since they should only be called
   once. */

    for ( ptr = node->handlerList; ptr != NULL; ptr = ptr->next) {
    	candNode = ( ntHandlerPtr) ptr->data;
	if (( matchHandlerEvent( candNode, WM_NCCREATE, 0, 0) ||
	      matchHandlerEvent( candNode, WM_CREATE, 0, 0)) &&
	    !candNode->deleted) {
	    ( candNode->AmHandlerFn)( hWnd, iMsg, wParam, lParam,
				    candNode->clientData);
	    candNode->deleted = TRUE;
	}
    }

/* call any event handlers with matching event codes which haven't
   been marked as deleted. If no handlers are found, then we
   call the DefWindowProc() */
   
    foundFlag = FALSE;
    for ( ptr = node->handlerList; ptr != NULL; ptr = ptr->next) {
    	candNode = ( ntHandlerPtr) ptr->data;
	if ( matchHandlerEvent( candNode, iMsg, wParam, lParam) && 
	     !candNode->deleted) {
	    foundFlag = TRUE;
	    ( candNode->AmHandlerFn)( hWnd, iMsg, wParam, lParam, 
	    		            candNode->clientData);
	}
    }

/* if the message is WM_DESTROY, then mark all of the event handlers 
   as deleted */

    if ( iMsg == WM_DESTROY) {
	for ( ptr = node->handlerList; ptr != NULL; ptr = ptr->next) {
	    candNode = ( ntHandlerPtr) ptr->data;
	    candNode->deleted = TRUE;
	}
    }

/* now that the event handlers have been called, decrement the call
   counter */

    callCount--;

/* delete all of the nodes which have been marked as deleted. Delete up
   the the AmFirst unmarked node. (or the end of the list if all the nodes
   have been deleted).  ONLY DELETE NODES WHEN WE ARE AT THE TOP LEVEL
   of the call count tree, since this function must be re-entrant... */

    if ( callCount == 0) {
	ptr = node->handlerList; 
	while ( ptr != NULL) {
	    candNode = ( ntHandlerPtr) ptr->data;
	    if ( candNode->deleted) {
		deadNode = ptr;
		ptr = ptr->next;
		free( deadNode->data);
		free( deadNode);
	    } else {
		break;
	    }
	}

/* delete the rest of the marked nodes after the head */

	node->handlerList = ptr;
	if ( ptr != NULL) {
	    prevNode = ptr;
	    ptr = ptr->next;
	    while ( ptr != NULL) {
		candNode = ( ntHandlerPtr) ptr->data;
		if ( candNode->deleted) {
		    deadNode = ptr;
		    ptr = ptr->next;

/* hook the preceding node's pointer to the next pointer */

		    prevNode->next = ptr;
		    free( deadNode->data);
		    free( deadNode);
		} else {
		    prevNode = ptr;
		    ptr = ptr->next;
		}
	    }
	}

/* if the handler list is NULL, then remove the window's handler list from
   the tree. */

	if ( node->handlerList == NULL) {
	    tdelete( node, &handlerRoot, ( AmSearchFn) matchHandlerWindow);
	    free( node);
	}
    }
    if ( !foundFlag) {
	return( DefWindowProc( hWnd, iMsg, wParam, lParam));
    } else {
	return( 0);
    }
}

static int compareDialog( HWND	hDlgA,
			  HWND	hDlgB)
{
    return(( int) hDlgA - ( int) hDlgB);
}

static AmListPtr dialogList = NULL;

/*************************************************************************
 *
 * Description:
 *	The joys of Microsoft windows.  Modeless Dialogs need their messages
 *	re-routed from the main event loop, so there has to be a test
 *	when messages are intended for them, so they're not sent to the
 *	main window. The register and unregister functions are used 
 *	when dialogs are enabled and disabled.  This should be called
 *	just after the CreateDialog() call.
 * Arguments:
 *	hDlg - dialog to be registered.
 *
 *************************************************************************/

void 
AmRegisterDialog( HWND	hDlg)
{
    dialogList = ( AmListPtr) AmInsertNodeLst( hDlg, dialogList);
}

/*************************************************************************
 *
 * Description:
 *	Remove a dialog from the registry.
 * Arguments:
 *	hDlg - dialog to be unregistered.
 *
 *************************************************************************/

void 
AmUnregisterDialog( HWND	hDlg)
{
    dialogList = ( AmListPtr) AmDelPtrLst( hDlg, dialogList, 
				       ( AmSearchFn) compareDialog);
}

/*************************************************************************
 *
 * Description:
 *	Test whether a message is supposed to be sent to a registered
 *	dialog.  This function should be called in the DispatchMessage()
 *	loop.
 * Arguments:
 *	msg - message to test whether it is being sent to a dialog in
 *	the registry.
 *
 *************************************************************************/

AnvilBoolean 
AmIsDialogEvent( MSG	*msg)
{
    AmListPtr	ptr;
    HWND	hDlg;

    for ( ptr = dialogList; ptr != NULL; ptr = ptr->next) {
    	hDlg = ( HWND) ptr->data;
	if ( IsDialogMessage( hDlg, msg)) {
	    return( TRUE);
	}
    }
    return( FALSE);
}
