#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdarg.h>
#include <stdio.h>
#include <windows.h>
#include <anvil/nt.h>
#include <anvil/project.h>
#include <anvil/resource.h>
#include <anvil/files.h>
#include <anvil/log.h>
#include <resources/resource.h>

/**************************************************************************
 *
 * Description:
 * 	low-level print error function brings up a modal error dialog, 
 *	pulling the caption from the resource file. 
 * Arguents:
 *	hWnd: parent window for the message box.
 *	errMsg: error message;
 * Resoruces: 
 *	error.caption: caption for the error dialog.
 *
 **************************************************************************/

void 
PrintErrorBuffer( HWND		hWnd,
	          const char	*errMsg)
{
    const char *caption = AmGetDefault( "error.caption", "Error");

    MessageBox( hWnd, errMsg, caption, MB_OK | MB_ICONERROR | MB_APPLMODAL);
}

/**************************************************************************
 *
 * Description:
 *	use a format from the resource database to popup a modal error 
 *	message from the varargs format arguments.  The hWnd parameter 
 *	is used as the parent of the MessageBox created by PrintErrorBuffer().
 * Arguments:
 *	hWnd: parent window for the message box.
 *	res: resource to use for vsprintf() format
 *	varargs: varargs list of parameters for the message.
 *
 *************************************************************************/

void 
AmPrintResourceError( HWND	hWnd,
		    const char	*res,
		    ...)
{
    va_list	args;
    char	buf[ MSGLEN];
    const char	*fmt;

/* get the string from the resource file */

    va_start( args, res);
    fmt = AmGetDefault( res, res);
    vsprintf( buf, fmt, args);
    va_end( args);
    PrintErrorBuffer( hWnd, res);
}

/**************************************************************************
 *
 * Description:
 *	dialog procedure to display the contents of the log file.  The
 *	log file is (hopefully) written to by AmAddResourceToLog().
 * Arguments:
 *	hDlg - dialog window.
 *	iMsg, wParam, lParam - event parameters:
 * Files:
 *	temporary log file created by AmInitLog().
 *
 **************************************************************************/

static BOOL CALLBACK logErrorFileDialogProc( HWND	hDlg,
					     UINT	iMsg,
					     WPARAM	wParam,
					     LPARAM	lParam)
{
    HWND	 hWndText;
    char	*logStrings;

    switch ( iMsg) {
    case WM_INITDIALOG:
	logStrings = AmReadFileString(( char *) AmGetLogFile()); 
	if ( logStrings == NULL) {
	    AmPrintResourceError( GetParent( hDlg), "noLogFile");
	    EndDialog( hDlg, 0);
	}
	hWndText = GetDlgItem( hDlg, LOG_LABEL);
	SetWindowText( hWndText, logStrings);
	free( logStrings);
	return( TRUE);
    case WM_COMMAND:
    	switch ( LOWORD( wParam)) {
	case IDCANCEL:
	    EndDialog( hDlg, 0);
	    return( TRUE);
	}
	break;
    }
    return( FALSE);
}

/* static buffer for log error callback to pass to the logErrorDialogProc */

static char	logMsgBuf[ MSGLEN];

/**************************************************************************
 *
 * Description:
 *	dialog proc for the log error dialog.  If the user hits dismiss, 
 *	blow the dialog away.  If they hit details, bring up the error 
 *	details dialog.
 * Arguments:
 *	hDlg - dialog window.
 *	iMsg, wParam, lParam - event parameters:
 *
 **************************************************************************/

static BOOL CALLBACK logErrorDialogProc( HWND	hDlg,
					 UINT	iMsg,
					 WPARAM	wParam,
					 LPARAM	lParam)
{
    int		errCode;
    HINSTANCE	hInstance;
    HWND	hWndParent = GetParent( hDlg), hWndText;

    switch ( iMsg) {
    case WM_INITDIALOG:
	hWndText = GetDlgItem( hDlg, LOG_MSG_LABEL);
	SetWindowText( hWndText, logMsgBuf);
    case WM_COMMAND:
    	switch ( LOWORD( wParam)) {
	case ID_DETAILS:
	    hWndParent = GetParent( hDlg);
	    hInstance = ( HINSTANCE) GetWindowLong( hWndParent, GWL_HINSTANCE);
	    errCode = DialogBox( hInstance, MAKEINTRESOURCE( LOG_FILE_DLOG),
				 hWndParent, ( DLGPROC) logErrorFileDialogProc);
	    if ( errCode == -1) {   
		AmPrintResourceError( hWndParent, "ErrorLogDialogFailed");
	    }
	case ID_DISMISS:
	case IDCANCEL:
	    EndDialog( hDlg, 0);
	    return( TRUE);
	}
	break;
    }
    return( FALSE);
}

/**************************************************************************
 *
 * Description:
 *	A variation on the error message dialog, which allows the user 
 *	to bring up a details dialog which shows the logfile written
 *	by PrintResourceLog().
 * Arguments:
 *	hWnd: parent window for the message box.
 *	res: resource to use for vsprintf() format
 *	varargs: varargs list of parameters for the message.
 *
 *************************************************************************/

void
AmPrintLogError( HWND		hWnd,
	       const char	*res,
	       ...)
{
    va_list	args;
    const char	*fmt;
    HINSTANCE	hInstance;
    int		errCode;

/* get the string from the resource file, then use it as a 
   format to print the varargs parameters into the static buffer
   which is set in the dialog by the dialog proc */

    va_start( args, res);
    fmt = AmGetDefault( res, res);
    vsprintf( logMsgBuf, fmt, args);
    va_end( args);
    hInstance = ( HINSTANCE) GetWindowLong( hWnd, GWL_HINSTANCE);
    errCode = DialogBox( hInstance, MAKEINTRESOURCE(ERROR_LOG_DLOG),
			 hWnd, ( DLGPROC) logErrorDialogProc);
    if ( errCode == -1) {   
	AmPrintResourceError( hWnd, "ErrorLogDialogFailed");
    }
}
