#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <anvil/tsearch.h>
#include <anvil/errcode.h>
#include <anvil/ntresource.h>

/**************************************************************************
 * 
 *  NT resource data base interface. 
 *	Currently, we're using straight ASCII files instead of the 
 *	registry for simplicity.  The AmNTResourceDbClass is inherited
 *	from the base AmResourceDbClass, and is used by the the AmGetDefault()
 *	functions by setting it as the global resource database with
 *	AmSetTheResourceDb()
 *
 **************************************************************************/


/* structure for saving resource name-value pairs in the tsearch() tree */

typedef struct resEntry_ {
    char	*resourceName;
    char	*resourceValue;
} resEntryRec, *resEntryPtr;

/* tsearch() comparison function for resource names */

static int byResName( resEntryPtr	a,
		      resEntryPtr	b)
{
    return( strcmp( a->resourceName, b->resourceName));
}

/* the destructor should walk through the tree and delete all the items */

AmNTResourceDbClass::~AmNTResourceDbClass()
{
}

/* used by SaveResourceDb to write the tree out into the file */

static FILE	*fp;

static void 
writeAction( resEntryPtr a,
	     VISIT	 order,
	     int	 depth)
{
    if (( order == preorder) || ( order == leaf)) {
    	fprintf( fp, "%s:\t%s\n", a->resourceName, a->resourceValue);
    }
}

/*************************************************************************
 * 
 * Description:
 *	walk through the resource tree and write out the name-value
 *	pairs into the original database file
 *
 *************************************************************************/

void
AmNTResourceDbClass::SaveResourceDb()
{
 
    if (( fp = fopen( filename, "w")) == NULL) {	
    	DO_ERROR( AmOPEN_FAILED);
	return;
    }
    twalk( resourceRoot, ( AmWalkFn) writeAction);
    fclose( fp);
}
		       
/*************************************************************************
 * 
 * Description:
 * 	constructor for the NT resource class.  Parses the file by hand 
 *	into a tree constructed by tsearch().  The file is of the format
 *	<resourceName>:	<resourceValue>.  leading spaces in the resource
 *	value are stripped, and lines starting with '#' are treated as
 *	comments and ignored.
 * Arguments:
 *	filenameArg: name of the file containing the resource entries.
 *
 *************************************************************************/
 
AmNTResourceDbClass::AmNTResourceDbClass( const char *filenameArg) :
			AmResourceDbClass( filenameArg) 
{
    FILE	*fp;
    char	buffer[ RESLEN], *p, *namePtr, *valPtr;
    resEntryPtr	resEntry;

    resourceRoot = NULL;

    if (( fp = fopen( filename, "r")) == NULL) {
    	return;
    }
    do {
    	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
	    break;
	} 

/* gid rid of that annoying '\n' that fgets puts at the end of the line */

	buffer[ strlen( buffer) - 1] = '\0';
	if ( buffer[ 0] == '#') {
	    continue;			/* skip comments */
	}

/* break up the buffer into a name and value, separated by ':' */

	namePtr = buffer;
	for ( p = buffer; (( *p != '\0') && ( *p != ':')); p++) {};
	if ( *p == '\0') {
	    continue;			/* skip over bad parses */
	}
	*p = '\0';
	p++;
	for ( ; isspace( *p); p++) {};
	valPtr = p;

	if (( resEntry = ( resEntryPtr) 
		    malloc( sizeof( resEntryRec))) == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    break;
	}
	if (( resEntry->resourceName = strdup( namePtr)) == NULL) {
	    DO_ERROR( AmSTRDUP_FAILED);
	    break;
	}
	if (( resEntry->resourceValue = strdup( valPtr)) == NULL) {
	    DO_ERROR( AmSTRDUP_FAILED);
	    break;
	}
	tsearch( resEntry, &resourceRoot, ( AmSearchFn) byResName);
    } while ( TRUE);
    fclose( fp);
}

/*************************************************************************
 * 
 * Description:
 *	is the NT resource database valid? are there any entries in the
 * 	resource tree? 
 * Returns:
 *	TRUE if there are entries in the resource database, FALSE if
 *	there are none.
 *
 ************************************************************************/

AnvilBoolean AmNTResourceDbClass::IsValid()
{
    return( resourceRoot != NULL);
}

/*************************************************************************
 * 
 * Description:
 *	Overloaded functions to retreive string, integer, float, and
 *	boolean values.  The values are pulled from the tree using tfind, 
 *	then converted from string values.  The string value is a direct 
 *	reference to the resource database, so it shouldn't be freed.
 *  Returns:
 *	TRUE if the resource was found in the resource database,
 *	FALSE if not.
 *
 *************************************************************************/

AnvilBoolean AmNTResourceDbClass::GetResource( const char *tag, char **value)
{
    resEntryPtr	*res;
    resEntryRec	target;

    target.resourceName = ( char *) tag;
    res = ( resEntryPtr *) tfind( &target, ( void * const *) &resourceRoot, 
   				  ( AmSearchFn) byResName);
    if ( res == NULL) {
        return( FALSE);
    }
    *value = ( *res)->resourceValue;
    return( TRUE);
}


AnvilBoolean AmNTResourceDbClass::GetResource( const char *tag, int *value)
{
    char	*resStr;

    if ( GetResource( tag, &resStr)) {
    	*value = atoi( resStr);
	return( TRUE);
    } else {
    	return( FALSE);
    }
}

AnvilBoolean AmNTResourceDbClass::GetResource( const char *tag, float *value)
{
    char	*resStr;

    if ( GetResource( tag, &resStr)) {
    	*value = atof( resStr);
	return( TRUE);
    } else {
    	return( FALSE);
    }
}

AnvilBoolean AmNTResourceDbClass::GetResource( const char *tag, AnvilBoolean *value)
{
    char	*resStr;

    if ( GetResource( tag, &resStr)) {
    	*value = !strcasecmp( resStr, "true");
	return( TRUE);
    } else {
    	return( FALSE);
    }
}

/*************************************************************************
 * 
 * Description:
 *	Overloaded functions to set string, integer, float, and
 *	boolean values in the resource tree.  
 *  Returns:
 *
 *************************************************************************/

AnvilBoolean AmNTResourceDbClass::PutResource( const char *tag, const char *value)
{
    resEntryPtr	resEntry;

    if (( resEntry = ( resEntryPtr) 
		malloc( sizeof( resEntryRec))) == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( FALSE);
    }
    if (( resEntry->resourceName = strdup( tag)) == NULL) {
	DO_ERROR( AmSTRDUP_FAILED);
	return( FALSE);
    }
    if (( resEntry->resourceValue = strdup( value)) == NULL) {
	DO_ERROR( AmSTRDUP_FAILED);
	return( FALSE);
    }
    tsearch( resEntry, &resourceRoot, ( AmSearchFn) byResName);
    return( TRUE);
}

/* put resources for ints, booleans, floats, and strings */


AnvilBoolean AmNTResourceDbClass::PutResource( const char *tag, int value)
{
    char	buffer[ 512];

    sprintf( buffer, "%d", value);
    return( PutResource( tag, buffer));
}

AnvilBoolean AmNTResourceDbClass::PutResource( const char *tag, float value)
{
    char	buffer[ 512];

    sprintf( buffer, "%f", value);
    return( PutResource( tag, buffer));
}

AnvilBoolean AmNTResourceDbClass::PutResource( const char *tag, AnvilBoolean value)
{
    char	buffer[ 512];

    if ( value) {
    	return( PutResource( tag, "true"));
    } else {
    	return( PutResource( tag, "false"));
    }
}
