#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <search.h>
#include <anvil/list.h>
#include <anvil/dispatch.h>
#include <anvil/nt_state_hand.h>
#include <anvil/nt_state_handP.h>
#include <anvil/nt_user_data.h>

/* used for sorting events to ensure correct comparisons */

static int 
byEventId( UINT	*eventA,
	   UINT	*eventB)
{
    return( *eventA - *eventB);
}

/* wrapper functions for turning buttons/controls on and off again */

void 
TurnOnControl( HWND	hWnd,
	       UINT	messageId)
{
    HMENU	hMenu;

/* if it's a button, the event is attached on a left button down, otherwise
   it's a menu item  and the message id is used to get the menu item */

    if ( messageId == WM_LBUTTONDOWN) {
	EnableWindow( hWnd, TRUE);
    } else {
    	hMenu = GetMenu( hWnd);
	EnableMenuItem( hMenu, messageId, MFS_ENABLED);
    }
}

void 
TurnOffControl( HWND	hWnd,
		UINT	messageId)
{
    HMENU	hMenu;

/* if it's a button, the event is attached on a left button down, otherwise
   it's a menu item  and the message id is used to get the menu item */

    if ( messageId == WM_LBUTTONDOWN) {
	EnableWindow( hWnd, FALSE);
    } else {
    	hMenu = GetMenu( hWnd);
	EnableMenuItem( hMenu, messageId, MFS_DISABLED);
    }
}

/* If the launch control gets destroyed, turn off the handler and set the
   launch control to NULL */

static void 
turnOffHandler( HWND		hWnd,
		UINT		iMsg,
		WPARAM		wParam,
		LPARAM		lParam,
		AmHandlerPtr	handler)
{
   handler->launchControl = NULL;
   AmDeleteHandler( handler->window, 
					handler->eventCodeVec, NULL, handler->numEvents,
					handler->AmHandlerFn, handler->clientData);
}

/* if the window that the handler is attached to is destroyed, then
   remove the destroy handler that's attached to the launch control. */

static void 
turnOffturnOffHandler( HWND		hWnd,
		       UINT		iMsg,
		       WPARAM		wParam,
		       LPARAM		lParam,
		       AmHandlerPtr	handler)
{
    AmDeleteHandler( handler->window, WM_DESTROY, 0,
					 ( AmNtHandlerFn) turnOffHandler, handler);
}

/*************************************************************************
 *
 * Description:
 *	Add an event handler to a window, kicked off by a launch control
 *	The event handler is assigned to an event code, and is also assigned
 *	a priority, which affects its behaviour in relation to other
 *	handlers added by AmAddHandler.  If a handler is currently active
 *	and a new handler is added with a lower priority, that handler is
 *	"suspended", then restored when the new handler is removed.  If 
 *	the new handler is of the same or higher priority, the current
 *	handler is cancelled outright.  Sometimes it's neccessary to 
 *	have actions occur when the handler is removed, suspended or
 *	restored, so "removeFn" is called when the handler is removed.
 *	"suspendFn" is called when the handler is suspended, and 
 *	"restartFn" is called when the handler is restored.  The handler
 *	is associated with a launch control, which is automatically
 *	disabled while the handler is active, and enabled when it is
 *	removed.  If the launch control is destroyed, the handler is removed.
 * 	Event handlers can only be added to event codes, not lParams.
 * Arguments:
 *	launchControl: control which launched the handler. (may be NULL)
 *	window: window that the handler is assigned to.
 *	eventCodeVec: event that activates the handler.
 *	AmHandlerFn: the event handler to add (standard event handler)
 *	clientData: passed to AmHandlerFn, removeFn, suspendFn, and restartFn
 *		as the clientData parameter.
 *	priority: priority assigned to the event handler.
 *	removeFn: called when the handler is removed. (may be NULL)
 *	suspendFn: called when the handler is suspended. (may be NULL)
 *	restartFn: called when the handler is restarted. (may be NULL)
 * Returns:
 *	AmOK if everything's hunky-dory.
 *	system error code if there were problems.
 *
 *************************************************************************/

AmErrorCode
AmAddHandler( HWND	   launchControl,
	    UINT	   messageId,
            HWND	   window,
	    UINT	   eventCode,
	    AmNtHandlerFn    AmHandlerFn,
	    void	   *clientData,
	    uint_t         priority,
            AmStateHandlerFn removeFn,
            AmStateHandlerFn suspendFn,
            AmStateHandlerFn restartFn)
{

    return( AmAddHandler( launchControl, messageId, window, &eventCode, 1,
			AmHandlerFn, clientData, priority, removeFn,
			suspendFn, restartFn));
}

/*************************************************************************
 *
 * Description:
 * 	Lower-level interface for the event handlers allows the caller
 *	to pass a vector of events and a length.
 * Parameters:
 *	eventCodeVec: vector of events which activate the handler.
 *	numEvents: length of the vector.
 *
 *************************************************************************/

AmErrorCode
AmAddHandler( HWND	   launchControl,
	    UINT	   messageId,
            HWND	   window,
	    UINT	   *eventCodeVec,
	    int		   numEvents,
	    AmNtHandlerFn    AmHandlerFn,
	    void	   *clientData,
	    uint_t         priority,
            AmStateHandlerFn removeFn,
            AmStateHandlerFn suspendFn,
            AmStateHandlerFn restartFn)
{
    AmHandlerDataPtr pHandlerData;
    AmHandlerPtr     pHandler, pHandlerNew;
    AmListPtr        lstTemp;
    AmErrorCode	   errCode;

/* the event handler information is stored in the window */

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( window, HANDLER_STRING);
    if ( pHandlerData == NULL) {
        pHandlerData = ( AmHandlerDataPtr) malloc( sizeof( AmHandlerDataRec));
        if ( pHandlerData == NULL) {
            DO_RETURN_ERROR( AmMALLOC_FAILED);
        }
        pHandlerData->magicNo = HANDLER_MAGIC;
        pHandlerData->handlerList = NULL;
        if (( errCode = AmAddUserData( window, HANDLER_STRING, 
				     pHandlerData)) != AmOK) {
            free( pHandlerData);
            return( ANVIL_ERROR_CODE);
        }
    }

/*
 * If there's an event handler currently being executed check to see if
 *   the handler has an equal or higher priority than the *   current
 *   handler.  If so, then suspend the current handler, and put the 
 *   new one on the stack.  If not, then remove the current handler,
 *   and turn its launch control back on.
 */

    if ( pHandlerData->handlerList != NULL) {
	pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
	AmDeleteHandler( window, pHandler->eventCodeVec, NULL, 
		       pHandler->numEvents, 
		       pHandler->AmHandlerFn, pHandler->clientData);
	if ( priority >= pHandler->priority) {

        /*
         * Remove all the handlers that have lower priority than the handler
         *   that is getting added.
         */

	    lstTemp = pHandlerData->handlerList;
	    while ( lstTemp != NULL) {
                pHandler = ( AmHandlerPtr) lstTemp->data;
                if ( priority < pHandler->priority) {
                    break;
                }
	        if ( pHandler->launchControl != NULL) {
		    TurnOnControl( pHandler->launchControl,
		    		   pHandler->messageId);
	        }
	        pHandlerData->handlerList =
                    AmFreeHead( pHandlerData->handlerList);
		lstTemp = pHandlerData->handlerList;

                /* Call the user defined remove function if needed. */

                if ( pHandler->removeFn != NULL) {
                    ( pHandler->removeFn)( pHandler->clientData);
                }
	        FREE( pHandler);
            } 
	} else {

        /* Suspend the current handler. */

            if ( pHandler->suspendFn != NULL) {
                ( pHandler->suspendFn)( pHandler->clientData);
            }
        }
    }

/* store the handler information, put it on the list, and turn it on. */

    if (( pHandlerNew = ( AmHandlerPtr) malloc( sizeof( AmHandlerRec))) == NULL) {
    	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    pHandlerNew->launchControl = launchControl;
    pHandlerNew->messageId = messageId;
    pHandlerNew->window = window;
    pHandlerNew->numEvents = numEvents;
    memcpy( pHandlerNew->eventCodeVec, eventCodeVec, sizeof( UINT)*numEvents);
    qsort( pHandlerNew->eventCodeVec, numEvents, sizeof( UINT),
    	   ( AmSearchFn) byEventId);
    pHandlerNew->AmHandlerFn = AmHandlerFn;
    pHandlerNew->clientData = clientData;
    pHandlerNew->priority = priority;
    pHandlerNew->removeFn = removeFn;
    pHandlerNew->suspendFn = suspendFn;
    pHandlerNew->restartFn = restartFn;
    pHandlerData->handlerList = AmInsertNodeLst( pHandlerNew,
                                                 pHandlerData->handlerList);

/*
 * If there's a launch control, then turn it off while the handler's 
 *   executing.
 */

    if ( launchControl != NULL) {
		AmInsertHandler( launchControl, WM_DESTROY, 0,
						 ( AmNtHandlerFn) turnOffHandler, pHandlerNew);
		AmInsertHandler( window, WM_DESTROY, 0,
						 ( AmNtHandlerFn) turnOffturnOffHandler, pHandlerNew);
		TurnOffControl( launchControl, messageId);
    }
    AmAppendHandler( window, eventCodeVec, NULL, numEvents, 
    				 AmHandlerFn, clientData);
    return( AmOK);
}

/* does a handler match vector of events? */

static AnvilBoolean 
matchHandlerEvent( AmHandlerPtr	pHandler,
		   AmNtHandlerFn	AmHandlerFn,
		   void		*clientData,
		   UINT		*eventCodeVec,
		   int		numEvents)
{
    if (( pHandler->AmHandlerFn == AmHandlerFn) &&
	( pHandler->clientData == clientData) &&
	( pHandler->numEvents == numEvents)) {
	return( memcmp( pHandler->eventCodeVec, eventCodeVec, 
			sizeof( UINT)*numEvents) == 0);
    } else { 
    	return( FALSE);
    }
}


/*************************************************************************
 *
 * Description:
 *	Remove the event handler described by the window it's assigned to
 *	the eventCode, handler function and client data.  If the handler is
 *	active, then the handlers' remove function is called.  The highest
 *	priority suspended handler is then restored.
 * Arguments:
 *	window: window that the handler is assigned to.
 *	eventCode: event that activates the handler.
 *	AmHandlerFn: the event handler to remove (standard event handler)
 *	clientData: client data that was passed to AmHandlerFn.
 * Returns:
 *	AmOK if everything's hunky-dory.
 *	system error code if there were problems,
 *	AmUNKNOWN_USER_DATA if the window user data used to hold the 
 *	    handler stack was bogus.
 *	AmMISSING_EVENT_HANDLER if the event handler wasn't in the stack.
 *
 *************************************************************************/

AmErrorCode
AmRemoveHandler( HWND		window,
	       UINT		eventCode,
	       AmNtHandlerFn	AmHandlerFn,
	       void		*clientData)
{
    return( AmRemoveHandler( window, &eventCode, 1, AmHandlerFn, clientData));
}

AmErrorCode 
AmRemoveHandler( HWND		window,
	       UINT		*eventCodeVec,
	       int		numEvents,
	       AmNtHandlerFn	AmHandlerFn,
	       void		*clientData)
{
    AmHandlerDataPtr pHandlerData;
    AmHandlerPtr     pHandler, pHandlerNew;
    AmListPtr        lstTemp;

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( window, HANDLER_STRING);

/* If there is no handler data, then the handler wasn't installed, was it? */

    if ( pHandlerData == NULL) {
	return( ANVIL_ERROR_CODE);
    } else if ( pHandlerData->magicNo != HANDLER_MAGIC) {
	DO_RETURN_ERROR( AmUNKNOWN_USER_DATA);
    }

/* sort the incoming event codes before comparing them */

    qsort( eventCodeVec, numEvents, sizeof( UINT), ( AmSearchFn) byEventId);

/* Search for the handler in the handler list, and turn off the handler. */ 

    for ( lstTemp = pHandlerData->handlerList;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {
	pHandler = ( AmHandlerPtr) lstTemp->data;
	if ( matchHandlerEvent( pHandler, AmHandlerFn, clientData,
				eventCodeVec, numEvents)) {
	    break;
	}
    }

    /*
     * If the handler was launched by a control, turn it back on, then
     *   delete the handler from the handler list.
     */

    if ( lstTemp != NULL) {
	if (pHandler->launchControl != NULL) {
	    TurnOnControl( pHandler->launchControl, pHandler->messageId);
	    AmDeleteHandler( pHandler->launchControl, WM_DESTROY, 0,
						 ( AmNtHandlerFn) turnOffHandler, pHandler);
	    AmDeleteHandler( pHandler->window, WM_DESTROY, 0,
						 ( AmNtHandlerFn) turnOffturnOffHandler, pHandler);
	}

    /* Call the user defined remove function if there is one. */

        if ( pHandler->removeFn != NULL) {
            ( pHandler->removeFn) ( pHandler->clientData);
        }

	if ( lstTemp == pHandlerData->handlerList) {

        /* 
         * The handler being removed is currently active, which means the
         *    NT handler needs to be removed, 
         */

            AmDeleteHandler( window, eventCodeVec, NULL, numEvents, 
	    					 AmHandlerFn, clientData);
            if ( pHandlerData->handlerList->next) {
				pHandlerNew =
                    ( AmHandlerPtr) pHandlerData->handlerList->next->data;
				AmAppendHandler( window, pHandlerNew->eventCodeVec, NULL,
								 pHandlerNew->numEvents,
								 pHandlerNew->AmHandlerFn, pHandlerNew->clientData);
                if ( pHandlerNew->restartFn != NULL) {
                    ( pHandlerNew->restartFn)( pHandlerNew->clientData);
                }
            }
	}
	pHandlerData->handlerList = 
	    AmDeleteItemLst( pHandler, pHandlerData->handlerList, 
						 ( AmSearchFn) AmEquals, ( AmItemFn) free);
	return( AmOK);
    } else {
	return( AmMISSING_EVENT_HANDLER);
    }
}

/*************************************************************************
 *
 * Description:
 *	Remove all handlers of priority "priority" or less from the
 *	window's handler stack.
 * Arguments:
 *	window: window that the handler are to be removed from
 *	priority: priority to remove up to.
 * Returns:
 *	AmOK if everything's hunky-dory.
 *	system error code if there were problems,
 *	AmUNKNOWN_USER_DATA if the window user data used to hold the 
 *	    handler stack was bogus.
 *
 *************************************************************************/

AmErrorCode
AmFlushHandler( HWND	   window,
              unsigned int priority)
{
    AmHandlerDataPtr	pHandlerData;
    AmHandlerPtr		pHandler;
    AmListPtr		lstTemp;

/* the event handler information is stored in the window */

    pHandlerData = ( AmHandlerDataPtr) AmGetUserData( window, HANDLER_STRING);
    if ( pHandlerData == NULL) {	
        return( ANVIL_ERROR_CODE);
    } else if ( pHandlerData->magicNo != HANDLER_MAGIC) {
	DO_RETURN_ERROR( AmUNKNOWN_USER_DATA);
    } 

/*
 * If there's an event handler currently being executed check to see if
 *   the handler has a higher priority than the flush priority.  If not,
 *   remove the current handler and any other handlers that have a lower 
 *   priority than the flush priority.
 */

    if ( pHandlerData->handlerList != NULL) {
	pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
	if ( priority >= pHandler->priority) {
	    AmDeleteHandler( window, pHandler->eventCodeVec, NULL,
	    		   pHandler->numEvents, pHandler->AmHandlerFn, 
			   pHandler->clientData);

            /*
             * Remove all the handlers that have lower priority than the
             *   flush priority.
             */

            lstTemp = pHandlerData->handlerList;
            while( lstTemp != NULL) {
                pHandler = ( AmHandlerPtr) lstTemp->data;
                if ( priority < pHandler->priority) {
                    break;
                }

	/* Turn on the launch control if needed. */

	        if ( pHandler->launchControl != NULL) {
		    TurnOnControl( pHandler->launchControl, 
		    		   pHandler->messageId);
	        }
	        pHandlerData->handlerList =
                    AmFreeHead( pHandlerData->handlerList);
                lstTemp = pHandlerData->handlerList;

	/* Call the user defined remove function if needed. */

                if ( pHandler->removeFn != NULL) {
                    ( pHandler->removeFn)( pHandler->clientData);
                }
	        FREE( pHandler);
            }

        /* Start a new handler if there is one. */

            if ( pHandlerData->handlerList) {
				pHandler = ( AmHandlerPtr) pHandlerData->handlerList->data;
				AmAppendHandler( window, pHandler->eventCodeVec, NULL, 
								 pHandler->numEvents, pHandler->AmHandlerFn, 
								 pHandler->clientData);
                if ( pHandler->restartFn != NULL) {
                    ( pHandler->restartFn)( pHandler->clientData);
                }
            }
	}
    }
    return( AmOK);
}
