#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <windows.h>
#include <search.h>
#include <anvil/project.h>
#include <anvil/tsearch.h>
#include <anvil/userevents.h>
#include <anvil/list.h>
#include <anvil/dispatch.h>

/* list of handlers assigned to a single event.  The top level tree is 
   ordered by the packageId and eventId. */ 

typedef struct userHandlerList {
    int		packageId;		/* package ID discriminator */
    int		eventId;		/* event id */
    AmListPtr	handlerList;		/* list of user event handlers */
} AmUserHandlerListRec, *AmUserHandlerListPtr;

typedef struct userHandler {
    int		packageId;	
    int		eventId;
    void	*clientData;			/* pass to func */
    void	(*proc)( int, void *, void *);	/* user procedure */
    AnvilBoolean	callOnce;			/* call once and remove? */
    HWND	hWnd;				/* window associated with */
    AnvilBoolean	deleted;			/* handler deleted mark flag */
} AmUserHandlerRec, *AmUserHandlerPtr;

static void 	*handlerRoot = NULL;		/* event handler root */
static void	*vaHandlerRoot = NULL;		/* varargs handler root */
char	*END_LIST = "END_LIST";			/* varargs terminator */


static AnvilBoolean 
descendentOf( HWND	hParent,
	      HWND	hChild)
{

/* everyone is a child of nobody, nobody is a child of everyone */

    if (( hParent == NULL) || ( hChild == NULL)) {
	return( TRUE);
    }
    while ( hChild != NULL) {
    	if ( hChild == hParent) {
	    return( TRUE);
	}
	hChild = GetParent( hChild);
    }
    return( FALSE);
}

/* find the handler list for a particular event and package id */

static int 
cmpEvent( AmUserHandlerListPtr	ah,
          AmUserHandlerListPtr	bh)
{
    if ( ah->packageId == bh->packageId) {
    	return( ah->eventId - bh->eventId);
    } else {
    	return( ah->packageId - bh->packageId);
    }
}

/* comparison function for user event handlers */

static int 
cmpHandlerWindow( AmUserHandlerPtr	a,
	          AmUserHandlerPtr	b)
{
    return(( a->proc == b->proc) &&
	   ( a->clientData == b->clientData) &&
	   ( a->hWnd == b->hWnd));
}

/* check the handler-called-once flag, and see if the window is the 
   descendent of the targeted parent */

static int 
deleteThisHandler( HWND			hWnd,
	           AmUserHandlerPtr	b)
{
    return( b->deleted);
}

/***************************************************************************
 * 
 * Description:
 * 	send a user event to everyone in the universe who's listening 
 *	to it.  the packageId and eventId uniquely identify the event.
 *	the call data is passed as the third argument to any event handlers
 *	which are attached to the event.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	callData - directly passed as third argument to event handler 
 * Returns:
 *	TRUE if the window the event caused event handlers to be called.
 *
 ***************************************************************************/

AnvilBoolean 
AmSendUserEvent( int	packageId,
	       int	eventId,
	       void	*callData)
{
    return( AmSendUserEventChildren( packageId, eventId, callData, NULL));
}

/***************************************************************************
 * 
 * Description:
 * 	send a user event to be immediately processed by all functions which
 * 	have registered an interest which are children of another window, so
 *  	that the event is not sent to multiple instantiations of the same 
 *	dialog.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	callData - directly passed as third argument to event handler 
 * 	hWnd - window to send event to all descendents of 
 * Returns:
 *	TRUE if the window the event caused event handlers to be called.
 *
 ***************************************************************************/

AnvilBoolean 
AmSendUserEventChildren( int	packageId,
		       int 	eventId,
		       void	*callData,
		       HWND	hWnd)
{
    AmUserHandlerListRec	target;
    AmUserHandlerListPtr	userHandlerList, *pUserHandlerList;
    AmUserHandlerPtr	userHandler;
    AmListPtr		ptr;
    static int		callCount = 0;
    AnvilBoolean		callFlag;

// AN_DEBUG( 99, ("AmSendUserEvent: %d, %x %x\n", eventId, callData, w));
/* try to find the event in the tree... */

    target.packageId = packageId;
    target.eventId = eventId;
    pUserHandlerList = ( AmUserHandlerListPtr *) 
	tfind( &target, ( void * const *) &handlerRoot, ( AmSearchFn) cmpEvent);
    if ( pUserHandlerList != NULL) {
	callFlag = FALSE;
	userHandlerList = *pUserHandlerList;
	callCount++;
	for ( ptr = userHandlerList->handlerList; 
	      ptr != NULL; 
	      ptr = ptr->next) {
	    userHandler = ( AmUserHandlerPtr) ptr->data;

/* if the event is not tagged to a parent window, then send it to all 
   listeners, otherwise only send it to descendents of the specified parent 
   (see descendentOf) */

	    if ( !userHandler->deleted && 
	         descendentOf( hWnd, userHandler->hWnd)) {
		( userHandler->proc)
		    ( eventId, userHandler->clientData, callData);
		callFlag = TRUE;
	    } 
	}
	callCount--;

/* delete all of the handlers which have been deleted. Only call this if 
   the event is at the top level.  If we call it from lower in the list 
   in another sendEvent call, then it will trash the list */

	if ( callCount == 0) {
	    userHandlerList->handlerList = 
		AmDelElLst( hWnd, userHandlerList->handlerList, 
			  ( AmSearchFn) deleteThisHandler);
	}

/* remove the node from the tree if there are no more associated handlers */

	if ( userHandlerList->handlerList == NULL) {
	    tdelete( userHandlerList, &handlerRoot, ( AmSearchFn) cmpEvent);
	}
	return( callFlag);
    } else {
	return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 * 	lower level engine function to insert an event handler into the
 *	handler tree under the correct window, eventId and package ID.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	callData - directly passed as third argument to event handler 
 * 	hWnd - window to send event to all descendents of 
 * Returns:
 *	Always TRUE.
 *
 ***************************************************************************/

static AnvilBoolean 
addCallbackHandler( int			packageId,
		    int			eventId,
		    callbackProc	proc,
		    void		*clientData,
		    HWND		hWnd)
{
    AmUserHandlerListRec	target;
    AmUserHandlerListPtr	*pUserHandlerList, userHandlerList;
    AmUserHandlerPtr	userHandler;

/* create a structure to hold the user event. Set the deleted flag to false.
   We use deleted flags, because a user event can be deleted from the list
   of handlers for that event while the list is being traversed in 
   AmSendUserEventChildren.  The deletions are applied after the list is
   traversed, and any handlers  */

    userHandler = ( AmUserHandlerPtr) malloc( sizeof( AmUserHandlerRec));
    userHandler->packageId = packageId;
    userHandler->eventId = eventId;
    userHandler->clientData = clientData;
    userHandler->proc = proc;
    userHandler->hWnd = hWnd;
    userHandler->deleted = FALSE;

/* try to find the event in the tree... */

    target.packageId = packageId;
    target.eventId = eventId;
    pUserHandlerList = ( AmUserHandlerListPtr *) 
    	tfind( &target, ( void * const *) &handlerRoot, ( AmSearchFn) cmpEvent);

/* if not, then insert it */

    if ( pUserHandlerList == NULL) {
	userHandlerList = ( AmUserHandlerListPtr) 
	    malloc( sizeof( AmUserHandlerListRec));
	userHandlerList->handlerList = NULL;
	userHandlerList->packageId = packageId;
	userHandlerList->eventId = eventId;
	tsearch(( void *) userHandlerList, &handlerRoot, ( AmSearchFn) cmpEvent);
    } else { 
	userHandlerList = *pUserHandlerList;
    }

/* and add the user event to the list of handlers */

    userHandlerList->handlerList = 
	AmInsertNodeLst( userHandler, userHandlerList->handlerList);
    return( TRUE);
}

/***************************************************************************
 * 
 * Description:
 *	remove an interest in a user event.  The packageId, eventId, 
 *	proc and client data must match for the event to be removed.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	proc - event handler procedure to call
 * 	clientData - client data registered with the event handler.
 * Returns:
 *	TRUE if the event handler was deleted, FALSE if it wasn't found.
 *
 ***************************************************************************/

AnvilBoolean 
AmRemoveCallback( int		packageId,
		int		eventId,
		callbackProc	proc,
		void		*clientData) 
{
    return( removeCallbackWindow( packageId, eventId, proc, clientData, NULL));
}

/***************************************************************************
 * 
 * Description:
 * 	remove a callback which is associated with a window 
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	proc - event handler procedure to call
 * 	clientData - client data registered with the event handler.
 *	hWnd - window event handler was installed with, NULL if the
 *	       interest is global.
 * Returns:
 *	TRUE if the event handler was deleted, FALSE if it wasn't found.
 *
 ***************************************************************************/

AnvilBoolean 
removeCallbackWindow( int		packageId,
		      int		eventId,
		      callbackProc	proc,
		      void		*clientData,
		      HWND		hWnd)
{
    AmUserHandlerListRec	target;
    AmUserHandlerListPtr	*pUserHandlerList;
    AmUserHandlerRec	userHandler;
    AmUserHandlerPtr	found;

/* try to find the event in the tree... */

    target.packageId = packageId;
    target.eventId = eventId;
    pUserHandlerList = ( AmUserHandlerListPtr *) 
    	tfind( &target, ( void * const *) &handlerRoot, ( AmSearchFn) cmpEvent);
    if ( pUserHandlerList == NULL) {
	return( FALSE);
    }

/* find the handler in the list and set its delete flag */

    userHandler.clientData = clientData;
    userHandler.proc = proc;
    userHandler.hWnd = hWnd;
    found = ( AmUserHandlerPtr) AmFindDataLst( &userHandler, 
    					   ( *pUserHandlerList)->handlerList, 
					   ( AmSearchFn) cmpHandlerWindow);
    if ( found != NULL) {
    	found->deleted = TRUE;
	return( TRUE);
    } else {
AN_DEBUG( 40,( "failed to delete %x( %x), eventId = %d\n",
	       proc, clientData, eventId));
    	return( FALSE);
    }
}

/* when the window is destroyed, remove the callback, and free the 
   call structure. */

static void 
removeCallbackDestroy( HWND		hWnd,
		       UINT		iMsg,
		       WPARAM		wParam,
		       LPARAM		lParam,
		       AmUserHandlerPtr	userHandler)
{
    removeCallbackWindow( userHandler->packageId, userHandler->eventId,
			  userHandler->proc, userHandler->clientData, 
			  userHandler->hWnd);
    free( userHandler);
}

/***************************************************************************
 * 
 * Description:
 * 	add a callback which will be called whether an event is sent 
 *	to everyone, or a window and it descendents. 
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	proc - event handler procedure to call
 * 	clientData - client data passed as the second argument to the
 *	event handler when the event is received.
 * Returns:
 *	Always TRUE.
 *
 ***************************************************************************/

AnvilBoolean 
AmAddCallback( int 		packageId,
	     int		eventId,
	     callbackProc	proc,
	     void		*clientData)
{
    return( addCallbackWindow( packageId, eventId, NULL, proc, clientData));
}

/***************************************************************************
 * 
 * Description:
 * 	add a callback which will be called when an event is sent 
 *	to a window and it descendents. The event handler is removed
 *	when the window receives a WM_DESTROY event.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	hWnd - window that the event will be registered. Any events
 *	sent to ancestors of this window will be received.
 *	proc - event handler procedure to call
 * 	clientData - client data passed as the second argument to the
 *	event handler when the event is received.
 * Returns:
 *	Always TRUE.
 *
 ***************************************************************************/

AnvilBoolean
addCallbackWindow( int		packageId,
		   int		eventId,
		   HWND		hWnd,
		   callbackProc	proc,
		   void		*clientData)
{
    AmUserHandlerPtr	userHandler;

    userHandler = ( AmUserHandlerPtr) malloc( sizeof( AmUserHandlerRec));
    userHandler->packageId = packageId;
    userHandler->eventId = eventId;
    userHandler->proc = proc;
    userHandler->clientData = clientData;
    userHandler->hWnd = hWnd;
    AmInsertHandler( hWnd, WM_DESTROY, 0, 
					 ( AmNtHandlerFn) removeCallbackDestroy, userHandler);
    return( addCallbackHandler( packageId, eventId, proc, clientData, hWnd));
}


/* when the window is destroyed, then we remove the callback from the
   tree of handlers we use for tracking */

static void 
destroyUserVaCallback( HWND	hWnd,
		       UINT	iMsg,
		       WPARAM	wParam,
		       LPARAM	lParam,
		       AmListPtr	fnDataLst)
{
    tdelete( fnDataLst, &vaHandlerRoot, ( AmSearchFn) AmCompareListAddrs);
}

/* wrapper function to call the varargs user callback with the correct #
   of arguments */

static void 
AmVaUserCallbackWrapper( int 		event,
		       AmListPtr		clientDataAndEventIdLst,
		       void 		*callData)
{
    AmListPtr	    clientDataLst;
    VarargsCallback proc = ( VarargsCallback) clientDataAndEventIdLst->data;

/* the format of the list is proc, pkg id, event id, arguments */

    clientDataLst = clientDataAndEventIdLst->next->next->next;

/* since varargs doesn't convert straight to a stack, we have to do a
   switch statement kludge for the user's function. */

    switch( AmLengthLst( clientDataLst)) {
    case 0:
        ( proc)( event, callData);
        break;
    case 1:
        ( proc)( event,
                       clientDataLst->data,
                       callData);
        break;
    case 2:
        ( proc)( event,
                       clientDataLst->data,
                       clientDataLst->next->data,
                       callData);
        break;
    case 3:
        ( proc)( event,
                       clientDataLst->data,
                       clientDataLst->next->data,
                       clientDataLst->next->next->data,
                       callData);
        break;
    case 4:
        ( proc)( event,
                       clientDataLst->data,
                       clientDataLst->next->data,
                       clientDataLst->next->next->data,
                       clientDataLst->next->next->next->data,
                       callData);
        break;
    case 5:
        ( proc)( event,
                       clientDataLst->data,
                       clientDataLst->next->data,
                       clientDataLst->next->next->data,
                       clientDataLst->next->next->next->data,
                       clientDataLst->next->next->next->next->data,
                       callData);
        break;
    case 6:
        ( proc)( event,
                       clientDataLst->data,
                       clientDataLst->next->data,
                       clientDataLst->next->next->data,
                       clientDataLst->next->next->next->data,
                       clientDataLst->next->next->next->next->data,
                       clientDataLst->next->next->next->next->next->data,
                       callData);
        break;
    case 7:
        ( proc) ( event,
                   clientDataLst->data,
                   clientDataLst->next->data,
                   clientDataLst->next->next->data,
                   clientDataLst->next->next->next->data,
                   clientDataLst->next->next->next->next->data,
                   clientDataLst->next->next->next->next->next->data,
                   clientDataLst->next->next->next->next->next->next->data,
                   callData);
        break;
    case 8:
        ( proc) ( event,
		  clientDataLst->data,
		  clientDataLst->next->data,
		  clientDataLst->next->next->data,
		  clientDataLst->next->next->next->data,
		  clientDataLst->next->next->next->next->data,
		  clientDataLst->next->next->next->next->next->data,
		  clientDataLst->next->next->next->next->next->next->data,
		  clientDataLst->next->next->next->next->next->next->next->data,
		  callData);
        break;
    default:
        AN_DEBUG( 1, ( "Too many arguments, callback not called"));
        break;
    }
}

/***************************************************************************
 * 
 * Description:
 * 	add a callback which will be called when an event is sent 
 *	to a window and it descendents. The client data is a varargs
 *	list of values (assumed to be void pointers) terminated with
 *	the keyword END_LIST.  A maximum of 8 arguments may be parsed.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	hWnd - window that the event will be registered. Any events
 *	sent to ancestors of this window will be received.
 *	proc - event handler procedure to call
 * 	clientData - varargs list of argument data which is passed to
 *	the callback when it's called.
 *
 ***************************************************************************/

AnvilBoolean 
addVaCallbackWindow( int		packageId,
		     int		eventId,
		     HWND		hWnd,
		     VarargsCallback	proc, 
		     ...)
{
    va_list	clientDataArgs;
    void 	*clientData;
    AmListPtr	clientDataLst = NULL, fnDataLst;
    AnvilBoolean	flag;

/* prepend the package id and event id as discriminators */

    clientDataLst = AmInsertNodeLst(( void *) eventId, clientDataLst);
    clientDataLst = AmInsertNodeLst(( void *) packageId, clientDataLst);

/* loop through the varargs list and create a generic list */

    va_start( clientDataArgs, proc);
    for ( ;;) {
        clientData = va_arg( clientDataArgs, void *);
        if ( clientData == END_LIST) {         /* better terminator */
            break;
        }
        clientDataLst = AmInsertNodeAtEndLst( clientData, clientDataLst);
    }
    va_end( clientDataArgs);
    fnDataLst = AmInsertNodeLst(( void *) proc, clientDataLst);

/* add the callback and the data list to the binary tree of all varargs
   callbacks */

    tsearch( fnDataLst, &vaHandlerRoot, ( AmSearchFn) AmCompareListAddrs);
    flag = addCallbackWindow( packageId, eventId, hWnd, 
			      ( callbackProc) AmVaUserCallbackWrapper, fnDataLst);
    AmInsertHandler( hWnd, WM_DESTROY, 0, 
					 ( AmNtHandlerFn) removeCallbackDestroy, fnDataLst);
    return( flag);
}

/***************************************************************************
 * 
 * Description:
 * 	remove a varargs callback which was assigned to a window.
 *	the packageId, eventId, window, proc and all the client data
 *	arguments must match to remove the callback.
 * Arguments:
 *	packageId - uniquely identifying package ID.
 *	eventId - unique event identifier.
 *	hWnd - window that the event will be registered. Any events
 *	sent to ancestors of this window will be received.
 *	proc - event handler procedure to call
 * 	clientData - varargs list of argument data which was passed to
 *	the callback when it was called.
 * Returns:
 *	Always TRUE.
 *
 ***************************************************************************/

AnvilBoolean 
removeVaCallbackWindow( int		packageId,
			int		eventId,
			HWND		hWnd,
			VarargsCallback	proc, 
			...)
{
    va_list     clientDataArgs;
    void	*clientData;
    AmListPtr     clientDataLst = NULL, fnDataLst, *target, ptr;
    AnvilBoolean	flag = FALSE;

/* prepend the package id and event id as discriminators */

    clientDataLst = AmInsertNodeLst(( void *) eventId, clientDataLst);
    clientDataLst = AmInsertNodeLst(( void *) packageId, clientDataLst);

/* loop through the varargs list and create a generic list */

    va_start( clientDataArgs, proc);
    for ( ;;) {
        clientData = va_arg( clientDataArgs, void *);
        if ( clientData == END_LIST) {         /* better terminator */
            break;
        }
        clientDataLst = AmInsertNodeAtEndLst( clientData, clientDataLst);
    }
    va_end( clientDataArgs);
    fnDataLst = AmInsertNodeLst(( void *) proc, clientDataLst);

/* remove the callback from the tree of all callbacks */

    target = ( AmListPtr *) tfind( fnDataLst, ( void * const *) &vaHandlerRoot,
                                 ( AmSearchFn) AmCompareListAddrs);
    if ( target != NULL) {
	flag = removeCallbackWindow( packageId, eventId, 
				     ( callbackProc) AmVaUserCallbackWrapper, 
				     *target, hWnd);

/* remove the destroy callback with the same pointer it was created with */

	AmDeleteHandler( hWnd, WM_DESTROY, 0, 
					 ( AmNtHandlerFn) destroyUserVaCallback, *target);
	ptr = *target;
	tdelete( *target, &vaHandlerRoot, ( AmSearchFn) AmCompareListAddrs);
	AmJustFreeLst( ptr);
    } else {
AN_DEBUG( 40, ( "failed to delete va proc %x event = %d\n", proc, eventId));
    }
    AmJustFreeLst( fnDataLst);
    return( flag);
}
