#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/* projection class: basically a wrapper for the USGS projection library */

#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <anvil/strlib.h>
#include <anvil/projects.h>
#include <anvil/projclass.h>
#include <anvil/tagstrs.h>
#include <anvil/macros.h>

/****************************************************************************
 *
 * Destructor:
 *	free the projection strings and the projection pointer for the 
 *	projection class 
 *
 ****************************************************************************/

AmProjectionClass::~AmProjectionClass()
{
    if ( projStrs != NULL) {
	AmFreeStrVec( projStrs);
    } 
    if ( pj != NULL) {
	pj_free( pj);
    }
}

/****************************************************************************
 *
 * Constructor:
 *	base constructor: initialize the units to NULL, unset the units,
 *	set the valid flag.
 *
 ****************************************************************************/

AmProjectionClass::AmProjectionClass()
{
    projStrs = NULL;
    numProjStrs = 0;
    pj = NULL;
    srcUnits = AmUNIT_UNDEFINED;
    dstUnits = AmUNIT_UNDEFINED;
    inverseFlag = FALSE;
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	given the set of strings from the USGS projection, make a 
 *	projection class 
 * Arguments:
 * 	args - NULL-terminated vector of strings of the form
 *	"tag=value", which initialize the USGS projection.
 *	inverse projections are prefixed with the string "inv-proj"
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT if the projection fails to
 *	initialize, AmMALLOC_FAILED if the malloc to copy the strings 
 *	failed.
 *
 ***************************************************************************/

AmErrorCode 
AmProjectionClass::SetProjection( const char **args)
{
    const char	**ptr; 
    char	unitStr[ 64], **ptr2;
    AmErrorCode	errCode = AmOK;

    if ( pj != NULL) {
	pj_free( pj);
        pj = NULL;
    }
    if ( projStrs != NULL) {
	AmFreeStrVec( projStrs);
	projStrs = NULL;
    }
    if ( args == NULL) {
        return( AmOK);
    }
    if ( args[ 0] == NULL) {
	return( AmOK);
    }
    for ( ptr = args; *ptr != NULL; ptr++) {
	if ( !strcmp( *ptr, "proj=none")) {
	    srcUnits = AmUNIT_ARCDEG;
	    dstUnits = AmUNIT_ARCDEG;
	    projStrs = AmCopyStrVec( args);
	    if ( projStrs == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    } else {
		return( AmOK);
	    }
	}
    }

/* find out if the projection is inverted.  
   find out if there's units specified in the projection */

    srcUnits = AmUNIT_ARCDEG;
    dstUnits = AmUNIT_METER;
    inverseFlag = FALSE;
    for ( ptr = args; *ptr != NULL; ptr++) {
	if ( !strncmp( *ptr, INV_PROJ_TAG, strlen( INV_PROJ_TAG))) {
	    inverseFlag = TRUE;
	} else if ( !strncmp( *ptr, "units=", strlen( "units="))) {
	    sscanf( *ptr, "units=%s", unitStr);
	    if ( !AmGetUnitCode( unitStr, &dstUnits)) {
		AN_DEBUG( 1, ( "projection: failed to find units %s\n", 
			       unitStr));
	    }
	}
    }

/* intialize the projection */

    pj = pj_init( AmVeclen( args), args);
    if ( pj == NULL) {
	AN_DEBUG( 1, ("failed to initialize projection %s\n", args[ 0]));
	errCode = AmBAD_ARGUMENT;
	SetValid( FALSE);
    } else {	
	SetValid( TRUE);
    }

/* we have to get rid of the "inv-" directive in front of the proj */

    numProjStrs = AmVeclen( args);
    projStrs = ( char **) malloc( sizeof( char *)*( numProjStrs + 1));
    if ( projStrs == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ptr = args, ptr2 = projStrs; *ptr != NULL; ptr++, ptr2++) {
	if ( !strncmp( *ptr, INV_PROJ_TAG, strlen( INV_PROJ_TAG))) {
	    *ptr2 = strdup( &((*ptr)[ 4]));
	} else {
	    *ptr2 = strdup( *ptr);
	}
    }
    *ptr2 = NULL;
    return( errCode);
}

/****************************************************************************
 *
 * Description:
 * 	constructor:  given the set of strings from the USGS projection, 
 *	make a projection class 
 * Arguments:
 * 	args - NULL-terminated vector of strings of the form
 *	"tag=value", which initialize the USGS projection.
 *	inverse projections are prefixed with the string "inv-proj"
 *
 ***************************************************************************/

AmProjectionClass::AmProjectionClass( const char **args)
{
    projStrs = NULL;
    numProjStrs = 0;
    pj = NULL;
    srcUnits = AmUNIT_UNDEFINED;
    dstUnits = AmUNIT_UNDEFINED;
    inverseFlag = FALSE;

    SetProjection( args);
}

/****************************************************************************
 *
 * Description:
 * 	constructor:  given the set of strings from the USGS projection, 
 *	make a projection class, and take an inverse flag instead of the
 *	"inv-proj" argument.
 * Arguments:
 * 	args - NULL-terminated vector of strings of the form
 *	"tag=value", which initialize the USGS projection.
 *	inverse projections are prefixed with the string "inv-proj"
 * Modifications: 
 *	should make better usage of the SetProjection() function.
 *
 ***************************************************************************/

AmProjectionClass::AmProjectionClass( const char	**args,
				  AnvilBoolean	inverseFlagArg)
{
    const char	**ptr; 
    char	unitStr[ 64];

    projStrs = NULL;
    numProjStrs = 0;
    pj = NULL;
    srcUnits = AmUNIT_UNDEFINED;
    dstUnits = AmUNIT_UNDEFINED;

/* find out if the projection is inverted.  
   find out if there's units specified in the projection */

    srcUnits = AmUNIT_ARCDEG;
    dstUnits = AmUNIT_METER;
    inverseFlag = inverseFlagArg;
    for ( ptr = args; *ptr != NULL; ptr++) {
	if ( !strncmp( *ptr, "units=", strlen( "units="))) {
	    sscanf( *ptr, "units=%s", unitStr);
	    AmGetUnitCode( unitStr, &dstUnits);
	}
    }
    pj = pj_init( AmVeclen( args), args);
    projStrs = AmCopyStrVec( args);
}

/****************************************************************************
 *
 * Description:
 * 	dump the strings for a projection.
 * Returns:
 *	the set of USGS strings that were used to create the projection,
 *	with "inv-proj" if the projection is inverted, or NULL if there
 *	is no projection.
 *
 ***************************************************************************/

const char **
AmProjectionClass::DumpProjection()
{
    static char	*dumpStrings[ 64], projTmpStr[ 64];
    char	**ptr;
    int		ii;

    if ( projStrs == NULL) {
	return( NULL);
    }
    for ( ptr = projStrs, ii = 0; *ptr != NULL; ptr++, ii++) {
	if ( inverseFlag && !strcmp( *ptr, "proj")) {
	    sprintf( projTmpStr, "inv-%s", *ptr);
	    dumpStrings[ ii] = projTmpStr;
	} else { 
	    dumpStrings[ ii] = projStrs[ ii];
	} 
    }
    dumpStrings[ ii] = NULL;
    return(( const char **) dumpStrings);
}

/****************************************************************************
 *
 * Description:
 * 	return expected input units for projection 
 *
 ***************************************************************************/

unitType 
AmProjectionClass::expectsUnits()
{
    return ( srcUnits );
}

/****************************************************************************
 *
 * Description:
 * 	return output units for projection 
 *
 ***************************************************************************/

unitType 
AmProjectionClass::resultUnits()
{
    return ( dstUnits );
}

/* is a layer projected? */

AnvilBoolean AmProjectionClass::IsProjected()
{
    return( pj != NULL);
}

/****************************************************************************
 *
 * Description:
 * 	apply a projection to a point.
 * Arguments:
 * 	dstX, dstY - source and returned projected point.
 * Returns:
 *	TRUE if the point projected properly, FALSE if not.
 *
 ***************************************************************************/

AnvilBoolean 
AmProjectionClass::ProjectPt( float	*dstX,
			    float	*dstY)
{
    if ( pj != NULL) {
	projectPt( pj, inverseFlag, dstX, dstY);
	return( TRUE);
    }
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	apply an inverse projection to a point.
 * Arguments:
 * 	dstX, dstY - source and returned projected point.
 * Returns:
 *	TRUE if the point projected properly, FALSE if not.
 *
 ***************************************************************************/

AnvilBoolean 
AmProjectionClass::InvProjectPt( float	*dstX,
			       float	*dstY)
{
    if ( pj != NULL) {
	projectPt( pj, !inverseFlag, dstX, dstY);
	return( TRUE);
    }
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	apply a projection to a point with units.  return FALSE if 
 *	either value can't be converted into the projection's 
 *	expected units 
 * Arguments:
 * 	dstX, dstY - source and returned projected point.
 *	xUnits, yUnits - units the values are in.
 * Returns:
 *	TRUE if the point projected properly, FALSE if not.
 *
 ***************************************************************************/

AnvilBoolean 
AmProjectionClass::ProjectPtUnits( float		*dstX,
				 float		*dstY,
				 unitType	xUnits,
				 unitType	yUnits)
{
    float	convX, convY;

    if ( !AmConvertValue( xUnits, srcUnits, *dstX, &convX)) {
	return( FALSE);
    }
    if ( !AmConvertValue( yUnits, srcUnits, *dstY, &convY)) {
	return( FALSE);
    }
    if ( ProjectPt( &convX, &convY) ) {
        *dstX = convX;
        *dstY = convY;
        return( TRUE);
    } else {
	return( FALSE);
    }
}
    
/****************************************************************************
 *
 * Description:
 * 	apply an inverse projection to a point with units.  Return 
 *	FALSE if either value can't be converted into the projection's 
 *	result units 
 * Arguments:
 * 	dstX, dstY - source and returned projected point.
 *	xUnits, yUnits - units the values are in.
 * Returns:
 *	TRUE if the point projected properly, FALSE if not.
 *
 ***************************************************************************/

AnvilBoolean 
AmProjectionClass::InvProjectPtUnits( float	*dstX,
				    float	*dstY,
				    unitType	xUnits,
				    unitType	yUnits)
{
    float	convX, convY;

    if ( !AmConvertValue( xUnits, dstUnits, *dstX, &convX)) {
	return( FALSE);
    }
    if ( !AmConvertValue( yUnits, dstUnits, *dstY, &convY)) {
	return( FALSE);
    }
    if ( InvProjectPt( &convX, &convY) ) {
        *dstX = convX;
        *dstY = convY;
        return( TRUE);
    } else {
	return( FALSE);
    }
}

/* projection calls can't use the AmRecordClass convienience functions
   because AmCoordClass multiple inherits this and screenobjclass */

/* get the field definitions for the projection class and set the fields
   which are specified by the tag-value array */

int AmProjectionClass::setFields( AmTagValuePtr  tagValues,
                                int          numValues)
{
    int		ii, numUnusedFields;

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
    	if ( AmMatchTag( tagValues[ ii].tag, AmNprojection)) {
    	    SetProjection(( const char **) tagValues[ ii].value);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNnumProjStrs)) {
	    numProjStrs = ( int) tagValues[ ii].value;
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/* get the field definisions for the projection class, and retreive the
   fields */

int AmProjectionClass::getFields( AmTagValuePtr  tagValues,
			        int          numValues)
{
    int		ii, numUnusedFields;

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
    	if ( AmMatchTag( tagValues[ ii].tag, AmNprojection)) {
    	    *( char ***) tagValues[ ii].value = 
    	    	AmCopyStrVec(( const char **) projStrs);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNnumProjStrs)) {
    	    *( int *) tagValues[ ii].value = AmVeclen(( const char **) projStrs);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}
