#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* base annotation class member functions.  This mostly defines a 
   screen rectangle, and doesn't define any new fields over screenObjClass.
   Mostly it's here to provide a clean class tree, and to provide a place
   for any future hooks which might be applied to annotations. */

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/display.h>
#include <anvil/tagstrs.h>
#include <anvil/files.h>
#include <anvil/annotation.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmAnnotationClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmANNOTATION) {
        return( TRUE);
    } else {
        return( AmScreenObjClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so annotations can be subclassed.
 *
 ****************************************************************************/

AmAnnotationClass::AmAnnotationClass()
{
}

/****************************************************************************
 *
 * Description:
 *      destructor: doesn't do anything right now, because annotations
 *      are basically just screen objects.
 *
 ****************************************************************************/

AmAnnotationClass::~AmAnnotationClass()
{
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for an annotation.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.  There isn't 
 *	much here, because annotations are basically just screen objects.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify plot for resources
 *
 ****************************************************************************/

void 
AmAnnotationClass::initializeFields( AmWindowClass 	*windowArg,
				   const char   *nameArg)
{

/* set the screen object lookup table and the database 
   for the enumeration lookup */

    AmScreenObjClass::initializeFields( windowArg, nameArg);

/* annotations don't have parents */

    setParent( NULL);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for an annotation, given the tag-values 
 *	stored in a tag values array.  First, the coordinates fields 
 *	are retrived, then the other fields are retrieved using 
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmAnnotationClass::getFields( AmTagValuePtr tagValues,
			    int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmScreenObjClass::getFields( tagValues, numValues);
    AmAnnotationClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for an annotation given the tag-values stored in
 *      a tag values array.  First, the AmScreenObjClass fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  Any of the tag-values which occur in the
 *      list of fields for tickmarks are copied to temporary arrays,
 *      then set in the tickmarks.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmAnnotationClass::setFields( AmTagValuePtr tagValues,
			    int         numValues)
{
    int         	numFieldDefs;
    AmFieldDefPtr 	fieldDefs;

    numValues = AmScreenObjClass::setFields( tagValues, numValues);
    AmAnnotationClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an annotation.  This is
 *      used in the map editor to dump the grid's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmAnnotationClass::dumpFields( AmFieldLengthPtr fieldLengths,
			     AnvilBoolean	    allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( screen object) */

    numStrings = AmScreenObjClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmAnnotationClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));

}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an annotation. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmAnnotationClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmAnnotationClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
                                           int		*numDefs)
{
    *fieldDefs = NULL;
    *numDefs = 0;
}
