#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* member functions for circle annotation class */

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/circleannot.h>
#include <anvil/rect.h>
#include <anvil/tagstrs.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmCircleAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmCIRCLE_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so circle annotations can be subclassed.
 *
 ****************************************************************************/


AmCircleAnnotClass::AmCircleAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      destructor: doesn't do anything right now.
 *
 ****************************************************************************/

AmCircleAnnotClass::~AmCircleAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a circle annotation.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.  Initialize the
 *	basic annotation fields, then set the fields which are specific
 *	to a circle annotation, including filled, start and end arc
 *	angles in degrees, and the line width.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify annotation for resources
 *
 ****************************************************************************/

void
AmCircleAnnotClass::initializeFields( AmWindowClass  *windowArg,
                                    const char   *nameArg)
{
    AmAnnotationClass::initializeFields( windowArg, nameArg);

    objectType = AmCIRCLE_ANNOTATION;

    opaque  = FALSE;
    _fFrame = FALSE;
    filled  = FALSE;
   
    startArc = 0;
    endArc = 360;
    _dLineWidth = 0.0;
}

/****************************************************************************
 *
 * Description:
 *      AmCircleAnnotClass constructor.  Creates a circle annotation using 
 *	the window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in 
 *	the tickmarks.
 * Arguments:
 *      windowArg - window that the circle annotation will be drawn on.
 *      nameArg - string name of the circle annotation, used for 
 *		  identification and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmCircleAnnotClass::AmCircleAnnotClass( AmWindowClass  *windowArg,
                                    const char   *nameArg,
                                    ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    numValues = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    setFieldsWrapper( tagValues, numValues);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for circle annotation given the tag-values stored in
 *      a tag values array.  First, the AmAnnotationClass fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmCircleAnnotClass::setFields( AmTagValuePtr tagValues,
			     int         numValues)
{
    int                numFieldDefs;
    AmFieldDefPtr        fieldDefs;
    const AmRectangleRec *tempRect;

    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmCircleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    if ( AmChanged( &_dLineWidth)) {
        tempRect = GetRect();
        SetRect( tempRect->x, tempRect->y, tempRect->width, tempRect->height);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for circle annotation, given the tag-values 
 *	stored in a tag values array.  First, the annotation fields are 
 *	retrived, then the other fields are retrieved using the
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

 
int 
AmCircleAnnotClass::getFields( AmTagValuePtr tagValues,
			     int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmCircleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	set the circle annotations's rectangle.  The extents rect 
 *	overlaps the rectangle by the circle's line width.
 * Arguments:
 *	x, y, width, height - dimensions of the new rectangle
 * Modifictions:
 *	should probably call AmScreenObjClass::SetRect() rather than
 *	setting the rect by hand.  
 *
 ****************************************************************************/

void 
AmCircleAnnotClass::SetRect( int  x,
			   int  y,
			   int  width,
			   int  height)
{
    rect.x = x;
    rect.y = y;
    rect.width = width;
    rect.height = height;

/* set the extents rectangle to reflect the linewidth */

    extentsRect.x = (short int)(rect.x - _dLineWidth);
    extentsRect.y = (short int)(rect.y - _dLineWidth);
    extentsRect.width = (short int)(rect.width + _dLineWidth * 2);
    extentsRect.height = (short int)(rect.height + _dLineWidth * 2);
    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *      Draw a circle annotation. 
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *	nRects - number of exposed rectangles.
 * Returns:
 *      AmOK if everything's OK, which it is. 
 *
 ****************************************************************************/

AmErrorCode 
AmCircleAnnotClass::draw( AmWindowClass	    	*w,
			AmGCClass			*gc,
			AmRectangleRec const  	*drawRects,
			int			nRects)
{
    const AmRectangleRec *drawRect = GetRect();

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, drawRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( drawRect)) {
        return( AmOK);
    }

/* clear the rectangle before drawing the legend.  Note that we use fill
   rectangle, because ClearRect() is not sensitive to the gc clip regions */

    if ( opaque == TRUE) {
	gc->SetForegroundColor( &bgColor);
        display->FillRectangle( w, gc, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }
    gc->SetForegroundColor( &fgColor);
    gc->SetLineWidth( _dLineWidth);

    if ( filled == TRUE) {
        display->FillArc( w, gc,
			  drawRect->x, drawRect->y, 
			  drawRect->width, drawRect->height,
			  DEGTORAD( startArc), DEGTORAD( endArc));
    } else {
        display->DrawArc( w, gc,
			  drawRect->x, drawRect->y, 
			  drawRect->width - 1, drawRect->height - 1,
			  DEGTORAD( startArc), DEGTORAD( endArc));
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in AmCircleAnnotClass.  This is
 *      used in the map editor to dump the circle annotation's fields into 
 *	a file.  the field name (and length, if it's a vector field) is 
 *	written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmCircleAnnotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			      AnvilBoolean		allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmCircleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmCircleAnnotClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmCircleAnnotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmCircleAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
				       int          *numDefs)
{
    static AmFieldDefRec CircleAnnotClassFieldDefinitions[] = {
        { AmNendArc, offsetof( AmCircleAnnotClass, endArc) },
        { AmNfilled, offsetof( AmCircleAnnotClass, filled) },
        { AmNlineWidth, offsetof( AmCircleAnnotClass, _dLineWidth) },
        { AmNstartArc, offsetof( AmCircleAnnotClass, startArc) },
    };
    *fieldDefs = CircleAnnotClassFieldDefinitions;
    *numDefs = VECLEN( CircleAnnotClassFieldDefinitions);
}
