#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/*
 * File: distancescale.cpp
 *
 * Description:
 *   This file contains the distance scale member functions.  Distance
 *   scales are used to show the coordinate system of their parent
 *   object.
 *
 * Modifications:
 *   AmDistanceScaleClass has to be modified to support map projections
 *   so it can show distances in meters for coordinate systems in
 *   latitude/longitude.  There should also be more checks for 
 *   unreasonable values.  It has also not been tested with unit
 *   conversions, such as when the source coordinate system is in
 *   meters and we want to show miles.
 */

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/errcode.h>
#include <anvil/ratio.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/scale.h>
#include <anvil/userevents.h>
#include <anvil/datagraphic.h>
#include <anvil/distscale.h>

/****************************************************************************
 *
 * Function: IsType( AmObjectTypeEnum eType)
 *
 * Purpose:
 *   To verify whether this object is, or is derived from the passed
 *   in object type.
 *
 * Parameters:
 *   AmObjectTypeEnum eType - Requested object type.
 *
 * Returns:
 *   AnvilBoolean - TRUE if the object is of this type or a subclass of it,
 *             FALSE if not.
 *
 * Comments:
 *   A recursive run-time type function which for each class determines
 *   if the requested type is of this class, or of the parent's
 *   class (and so on).  This should be used instead of GetObjectType()
 *   wherever possible, to allow subclassing.
 *
 ****************************************************************************/

AnvilBoolean 
AmDistanceScaleClass::IsType( AmObjectTypeEnum eType)
{
    if ( eType == AmDISTANCE_SCALE) {
        return( TRUE);
    } else {
        return( AmScaleClass::IsType( eType));
    }
}

/****************************************************************************
 *
 * Function: initializeFields( pWindow, pCoord, szName)
 *
 * Purpose:
 *   Initializes the member variables of the distance scale class.
 *
 * Parameters:
 *   AmWindowClass *pWindow - Window on which the distance scale will be drawn.
 *   AmCoordClass  *pCoord  - Coordinate plot for which the distance scale is
 *                          created.
 *   const char  *szName  - String name used to identify the distance scale for
 *                          resources.
 *
 * Comments:
 *   The intializeFields() functions are created for each class, and
 *   each one calls the one for the "parent" class.  The AmScreenObjClass
 *   intializeFields() function is the topmost one that is called.  We
 *   can't use two-parameter constructors, because since we use varargs
 *   constructors, there is a naming ambiguity.  The distance scale
 *   calls the AmScaleClass initializer, then initializes the the distance
 *   scale-specific fields, such as tick size, style and spacing.
 *
 * Modifications:
 *   The size values should be consts in a source file, not hardcoded.
 *
 ****************************************************************************/

void 
AmDistanceScaleClass::initializeFields( AmWindowClass *pWindow,
				      AmCoordClass  *pCoord,
				      const char  *szName)
{
    float dXMin, dYMin, dXMax, dYMax, dBase;

    AmScaleClass::initializeFields( pWindow, pCoord, szName);

    objectType    = AmDISTANCE_SCALE;
    _eScaleStyle  = AmTICK_STYLE;

    _nTicks       = 5;
    _fShowMinor   = FALSE;
    _nMinorTicks  = 5;
    _iTickSize    = 5;

    opaque        = FALSE;
    _fFrame       = FALSE;

    titleVertical = FALSE;

/* Grab the units and extents from the plot. */

    pCoord->GetVaFields( AmNxunits, &_uUnits, NULL);
    pCoord->GetCoords( &dXMin, &dYMin, &dXMax, &dYMax);

/* Try to pick a reasonable spacing, given the current coordinates
   of the coordinate system. */

    AmScaleBaseTenSpacing( dXMin, dXMax, 50, &_dSpacing, &dBase);
}

/****************************************************************************
 *
 * Function: AmDistanceScaleClass()
 *
 * Purpose:
 *   Default constructor, so distance scales can be subclassed.
 *
 ****************************************************************************/

AmDistanceScaleClass::AmDistanceScaleClass()
{
}

/****************************************************************************
 *
 * Function: ~AmDistanceScaleClass()
 *
 * Purpose:
 *   Destructor for the distance scale class.
 *
 * Comments:
 *   Nothing has to be freed which is specific to AmDistanceScaleClass.
 *
 ****************************************************************************/

AmDistanceScaleClass::~AmDistanceScaleClass()
{
}

/****************************************************************************
 *
 * Description:
 *   When the rectangle for the distance scale is set, resize the 
 *   distance scale components (namely, the tick size).
 *
 * Parameters:
 *   x, y, width, height - new rectangle dimensions 
 *
 * Modifictations:
 *   this doesn't really set the components properly, and will have
 *   to be redone.  It doesn't follow after the principle of 
 *   "least astonishment"
 *
 ****************************************************************************/

void
AmDistanceScaleClass::SetRect( int x,
			     int y,
			     int cx,
			     int cy)
{
    AmScaleClass::SetRect( x, y, cx, cy);
}

/****************************************************************************
 *
 * Description:
 *      AmDistanceScaleClass constructor.  Creates a distance scale using 
 *	the window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in the 
 *	distance scale.
 * Arguments:
 *      windowArg - window that the distance scale will be drawn on.
 * 	graphic - parent coordinate system that the distance scale is
 *		  tied to.
 *      nameArg - string name of the distance scale, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmDistanceScaleClass::AmDistanceScaleClass( AmWindowClass *pWindow,
					AmCoordClass  *pCoord,
					const char  *szName,
					...)
{
    va_list	args;
    AmTagValueRec	vTagValues[ MAXTAGS];
    int		nTagValues;

    if ( !pCoord->IsType( AmCOORD)) {

	DO_ERROR( AmBAD_ARGUMENT);
	SetValid( FALSE);

    } else {

	initializeFields( pWindow, pCoord, szName);

	va_start( args, szName);
	nTagValues = AmMakeTagValueArrayVararg( vTagValues, args);
	setFieldsWrapper( vTagValues, nTagValues);
	va_end( args);

	SetValid( TRUE);

    }
}

/****************************************************************************
 *
 * Function: setFields( vTagValues, nTagValues)
 *
 * Purpose:
 *   Set the fields for the distance scale given the tag-values stored in
 *   a tag values array.
 *
 * Parameters:
 *   AmTagValuePtr vTagValues - Vector of tag-value pairs (with tagCopied and
 *                            copied flags).  The unparsed arguments are
 *                            returned "scrunched" in this array.
 *   int         nTagValues - Length of the vector.
 *
 * Returns:
 *   int - Number of unparsed arguments.
 * 
 * Comments:
 *   First, the AmScaleClass fields are set, then the other fields are set
 *   using the standard SetFieldByDefintions() function which uses a lookup
 *   table from the tag strings to the structure offsets.
 *
 ****************************************************************************/

int 
AmDistanceScaleClass::setFields( AmTagValuePtr vTagValues,
			       int         nTagValues)
{
    AmFieldDefPtr vFieldDefs;
    int         nFieldDefs;

    nTagValues = AmScaleClass::setFields( vTagValues, nTagValues);
    AmDistanceScaleClass::GetFieldDefinitions( &vFieldDefs, &nFieldDefs);
    nTagValues = SetFieldByDefinitions( vFieldDefs, nFieldDefs,
                                        vTagValues, nTagValues);

    return( nTagValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for distance scale, given the tag-values stored in
 *      a tag values array.  First, the scale fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmDistanceScaleClass::getFields( AmTagValuePtr vTagValues,
			       int         nTagValues)
{
    AmFieldDefPtr vFieldDefs;
    int         nFieldDefs;
 
    nTagValues = AmScaleClass::getFields( vTagValues, nTagValues);
    AmDistanceScaleClass::GetFieldDefinitions( &vFieldDefs, &nFieldDefs);
    nTagValues = GetFieldByDefinitions( vFieldDefs, nFieldDefs,
                                        vTagValues, nTagValues);

    return( nTagValues);
}

/****************************************************************************
 *
 * Description:
 *      Draw a distance scale. 
 * Arguments:
 *      window - window that the distance scale will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *	nRects - number of exposed rectangles.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmDistanceScaleClass::draw( AmWindowClass        *pWindow,
			  AmGCClass            *pGC,
			  AmRectangleRec const *pExposeRects,
			  int                nExposeRects)	
{
    AmCoordClass   *pCoord = ( AmCoordClass *) parent;
    float        dXMin, dYMin, dXMax, dYMax, dDistance, dConvSpacing;
    AmRectanglePtr pClipRects;
    int          nClipRects;
    int          iScreenSpacing, x0, y0, x1, y1, ii, iStart,
                 iWidth, iAscent, iDescent, iXStart, iXEnd, x;
    unitType     uSourceUnits;
    char         szBuffer[ 64], *szUnit;

/* If the object is invalid, only do the most basic display. */

    if ( !IsValid()) {
	AmScreenObjClass::draw( pWindow, pGC, pExposeRects, nExposeRects);
        return( AmNOT_VALID);
    }

/* Set the clip rectangles for the text and title GC's which are copied from
   the GC passed down from the display manager */

    pGC->GetClipRectangles( &pClipRects, &nClipRects);

    textGC->SetClipRectangles( pClipRects, nClipRects);
    titleGC->SetClipRectangles( pClipRects, nClipRects);

    FREE( pClipRects);

/* Clear the rectangle before drawing the tickmarks, and get the line width
   Use fillrectangle, because clear rect isn't sensitive to gc clip regions */

    if ( opaque == TRUE) {

        pGC->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nExposeRects; ii++) {
	    display->FillRectangle( pWindow, pGC,
                                    pExposeRects[ ii].x,
                                    pExposeRects[ ii].y,
                                    pExposeRects[ ii].width,
                                    pExposeRects[ ii].height);
	}

    }

/* Set all GCs to the foreground color. */

    pGC->SetForegroundColor( &fgColor);
    textGC->SetForegroundColor( &fgColor);
    titleGC->SetForegroundColor( &fgColor);

/* Get the current coordinates for the source plot. */

    pCoord->GetCoords( &dXMin, &dYMin, &dXMax, &dYMax);

/* Get the units for the source plot. */

    pCoord->GetVaFields( AmNxunits, &uSourceUnits, NULL);

/* Figure out the conversion values. */ 

    AmConvertValue( _uUnits, uSourceUnits, _dSpacing, &dConvSpacing);
    pCoord->GetScreenPt( dXMin, dYMin, &x0, &y0);
    pCoord->GetScreenPt( dXMin + dConvSpacing, dYMin, &x1, &y1);

    iScreenSpacing = abs( x1 - x0);
   
    iXStart = rect.x + rect.width / 2 - ( _nTicks * iScreenSpacing) / 2;
    iXEnd = rect.x + rect.width / 2 + ( _nTicks * iScreenSpacing) / 2;

    display->DrawLine( pWindow, pGC,
                       iXStart, rect.y + rect.height / 2,
		       iXEnd, rect.y + rect.height / 2);

/* Show tickmarks for distances. */

    if ( _eScaleStyle == AmTICK_STYLE) {

	if ( _fShowMinor) {
	    dDistance = - _dSpacing;
	} else {
	    dDistance = 0;
	}

	for ( ii = 0; ii <= _nTicks; ii++) {

	    x = ii * iScreenSpacing;

	    display->DrawLine( pWindow, pGC,
                               iXStart + x,
                               rect.y + rect.height / 2,
			       iXStart + x,
                               rect.y + rect.height / 2 - _iTickSize);

	    if ( showText) {
		sprintf( szBuffer, "%.0f", fabs( dDistance));
		display->ExtentsText( pGC, szBuffer,
                                      &iWidth, &iAscent, &iDescent);
		display->DrawAlignedText( pWindow, textGC,
                                          iXStart + x,
				          rect.y + rect.height/2 - _iTickSize -
                                              textOffset - iDescent,
				          szBuffer, AmCENTER_HORIZONTAL);
	    }

	    dDistance += _dSpacing;
	}

    /* Show the minor ticks in the AmFirst tickmark */

	if ( _fShowMinor) {

	    for ( ii = 0; ii <= _nMinorTicks; ii++) {
		x = ( ii * iScreenSpacing) / _nMinorTicks;
		display->DrawLine( pWindow, pGC,
                                   iXStart + x,
                                   rect.y + rect.height / 2,
				   iXStart + x, 
				   rect.y + rect.height / 2 - _iTickSize);
	    }
	}

    } else {

    /* Format showing opposing rectangles for the scale */

	if ( _fShowMinor) {
	    iStart = 1;
	} else {
	    iStart = 0;
	}

	pGC->SetForegroundColor( &fgColor);
	dDistance = 0;

	for ( ii = iStart; ii < _nTicks; ii++) {

	    x = ii * iScreenSpacing;
	    if ( EVEN( ii)) {
		y0 = - _iTickSize;
		y1 = 0;
	    } else {
		y0 = 0;
		y1 = - _iTickSize;
	    }

	    display->FillRectangle( pWindow, pGC,
                                    iXStart + x,
				    rect.y + rect.height / 2 + y0,
				    iScreenSpacing, _iTickSize);
	    display->DrawRectangle( pWindow, pGC,
                                    iXStart + x,
				    rect.y + rect.height / 2 + y0,
				    iScreenSpacing, _iTickSize);

	    display->DrawRectangle( pWindow, pGC,
                                    iXStart + x,
				    rect.y + rect.height / 2 + y1,
				    iScreenSpacing, _iTickSize);

	    if ( showText) {
		sprintf( szBuffer, "%.0f", fabs( dDistance));
		display->ExtentsText( pGC, szBuffer,
                                      &iWidth, &iAscent, &iDescent);
		display->DrawAlignedText( pWindow, textGC,
                                          iXStart + x,
				          rect.y + rect.height / 2 - 
                                              _iTickSize - textOffset - iDescent,
				          szBuffer, AmCENTER_HORIZONTAL);
	    }

	    dDistance += _dSpacing;
	}

	if ( showText) {
	    x = _nTicks * iScreenSpacing;
	    sprintf( szBuffer, "%.0f", fabs( dDistance));
	    display->ExtentsText( pGC, szBuffer, &iWidth, &iAscent, &iDescent);
	    display->DrawAlignedText( pWindow, textGC,
                                      iXStart + x,
				      rect.y + rect.height / 2 - 
					    _iTickSize - textOffset - iDescent,
				      szBuffer, AmCENTER_HORIZONTAL);
	}

/* show the rectangles for the minor divisions */

	if ( _fShowMinor) {
	    for ( ii = 0; ii < _nMinorTicks; ii++) {
		x = ( ii * iScreenSpacing) / _nMinorTicks;
		x1 = (( ii + 1) * iScreenSpacing) / _nMinorTicks;

		if ( EVEN( ii)) {
		    y0 = - _iTickSize;
		    y1 = 0;
		} else {
		    y0 = 0;
		    y1 = - _iTickSize;
		}

		display->FillRectangle( pWindow, pGC,
                                        iXStart + x, 
					rect.y + rect.height / 2 + y0, 
					x1 - x,
                                        _iTickSize);
		display->DrawRectangle( pWindow, pGC,
                                        iXStart + x, 
					rect.y + rect.height / 2 + y0, 
					x1 - x,
                                        _iTickSize);

		display->DrawRectangle( pWindow, pGC,
                                        iXStart + x, 
					rect.y + rect.height / 2 + y1, 
					x1 - x,
                                        _iTickSize);
	    }

/* show the text for distance at the starting tickmark */

	    if ( showText) {
		sprintf( szBuffer, "%.0f", fabs( _dSpacing));
		display->ExtentsText( pGC, szBuffer,
                                      &iWidth, &iAscent, &iDescent);
		display->DrawAlignedText( pWindow, textGC, 
					  iXStart,
					  rect.y + rect.height/2 - _iTickSize -
                                              textOffset - iDescent,
					  szBuffer, AmCENTER_HORIZONTAL);
	    }
	}
    }

/* Draw the title below the scales. */

    if ( showTitle) {
	AmGetUnitString( _uUnits, &szUnit);
	sprintf( szBuffer, "%s (%s)", title, szUnit);
	display->ExtentsText( pGC, szBuffer, &iWidth, &iAscent, &iDescent);
	display->DrawAlignedText( pWindow, titleGC,
                                  rect.x + rect.width / 2,
				  rect.y + rect.height / 2 + titleOffset +
                                      iAscent,
				  szBuffer, AmCENTER_HORIZONTAL);
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in AmDistanceScaleClass.  This is
 *      used in the map editor to dump the distance scale's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmDistanceScaleClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				AnvilBoolean		  allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( scale) */

    numStrings = AmScaleClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmDistanceScaleClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmDistanceScaleClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmDistanceScaleClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmDistanceScaleClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					 int		*numDefs)
{
    static AmFieldDefRec DistanceScaleClassFieldDefinitions[] = {
	{ AmNdistanceUnits, offsetof( AmDistanceScaleClass, _uUnits) },
	{ AmNnumMinorTicks, offsetof( AmDistanceScaleClass, _nMinorTicks) },
	{ AmNnumTicks, offsetof( AmDistanceScaleClass, _nTicks) },
	{ AmNscaleStyle, offsetof( AmDistanceScaleClass, _eScaleStyle) },
	{ AmNshowMinorScale, offsetof( AmDistanceScaleClass, _fShowMinor) },
	{ AmNtickSize, offsetof( AmDistanceScaleClass, _iTickSize) },
	{ AmNtickSpacing, offsetof( AmDistanceScaleClass, _dSpacing) },
    };
    *fieldDefs = DistanceScaleClassFieldDefinitions;
    *numDefs = VECLEN( DistanceScaleClassFieldDefinitions);
}
