#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* AmGridClass member functions.  A grid is used to show the coordinate
   system of a parent object.  The AmGridClass contains four TickClases to
   show the tickmarks around the parent object, and also draws a grid
   for the parent object.
   Modifications:
	Some work needs to be done for projected grids, so latitude-longitude
	grids can be shown over projected plots, and vice-versa.  
*/
  

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#if defined (_WIN32) || defined (__BSD44__)
#include <malloc.h>
#else
#include <alloca.h>
#endif

#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/ratio.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/scale.h>
#include <anvil/datagraphic.h>
#include <anvil/userevents.h>
#include <anvil/tick.h>
#include <anvil/grid.h>
    
/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean
AmGridClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmGRID) {
        return( TRUE);
    } else {
        return( AmCoordClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a grid.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.  The grid 
 *	creates the four "subobject" tickmarks in its initializeFields()
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify plot for resources
 *
 ****************************************************************************/

void 
AmGridClass::initializeFields( AmWindowClass   *windowArg,
			     AmCoordClass    *graphicArg,
			     const char	   *nameArg)
{

    float	xmin, ymin, xmax, ymax, horizBase, horizSpacing, 
		vertBase, vertSpacing, spacing;
    int		maxTicks;

    AmScreenObjClass::initializeFields( windowArg, nameArg);
    parent = graphicArg;
    objectType = AmGRID;
    drawGrid = FALSE;
    rect = *graphicArg->GetRect();

    opaque  = FALSE;
    _fFrame = FALSE;

    gridStyle = AmGRID_PLUS;
    plusSize = 5;
    dashes = strdup( "\2\2");
    ndash = 2;
    lineStyle = AmLINE_SOLID;
    _dLineWidth = 1.0;
    sameTickSpacing = TRUE;

    upperTicks = new AmTickClass( windowArg, graphicArg, "upperTicks",
				AmNtickPosition, AmTOP,
				AmNtickStyle, AmSPACED,
				AmNfontSize, 10,
				AmNtextVertical, TRUE,
				AmNopaque, FALSE,
				AmNwidth, rect.width,
				NULL);
    leftTicks = new AmTickClass( windowArg, graphicArg, "leftTicks",
				AmNtickPosition, AmLEFT,
				AmNtickStyle, AmSPACED,
				AmNfontSize, 10,
				AmNtextVertical, FALSE,
				AmNopaque, FALSE,
				AmNheight, rect.height,
				NULL);
    rightTicks = new AmTickClass( windowArg, graphicArg, "rightTicks",
				AmNtickPosition, AmRIGHT,
				AmNtickStyle, AmSPACED,
				AmNfontSize, 10,
				AmNtextVertical, FALSE,
				AmNopaque, FALSE,
				AmNheight, rect.height,
				NULL);
    lowerTicks = new AmTickClass( windowArg, graphicArg, "lowerTicks",
				AmNtickPosition, AmBOTTOM,
				AmNtickStyle, AmSPACED,
				AmNfontSize, 10,
				AmNtextVertical, TRUE,
				AmNopaque, FALSE,
				AmNwidth, rect.width,
				NULL);

/* set the initial spacing for the tickmarks */

    (( AmCoordClass *) parent)->GetCoords( &xmin, &ymin, &xmax, &ymax);
    upperTicks->GetVaFields( AmNmaxTicks, &maxTicks,
			     NULL);
    AmScaleBaseTenSpacing( xmin, xmax, maxTicks, &horizSpacing, &horizBase);
    AmScaleBaseTenSpacing( ymin, ymax, maxTicks, &vertSpacing, &vertBase);
    spacing = MAX( horizSpacing, vertSpacing);
    upperTicks->SetVaFields( AmNtickBase, horizBase,
			     AmNtickSpacing, spacing,
			     NULL);
    lowerTicks->SetVaFields( AmNtickBase, horizBase,
			     AmNtickSpacing, spacing,
			     NULL);
    leftTicks->GetVaFields( AmNmaxTicks, &maxTicks,
			    NULL);
    leftTicks->SetVaFields( AmNtickBase, vertBase,
			    AmNtickSpacing, spacing,
			    NULL);
    rightTicks->SetVaFields( AmNtickBase, vertBase,
			     AmNtickSpacing, spacing,
			     NULL);
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so grids can be subclassed.
 *
 ****************************************************************************/

AmGridClass::AmGridClass()
{
}

/****************************************************************************
 *
 * Description:
 *      destructor: free the dash list if it's been allocated, and 
 *	delete the four "subobject" tickmarks.
 *
 ****************************************************************************/

AmGridClass::~AmGridClass()
{
    if ( dashes != NULL) {
    	free( dashes);
    }
    delete upperTicks;
    delete leftTicks;
    delete rightTicks;
    delete lowerTicks;
}

/****************************************************************************
 *
 * Description:
 *      AmGridClass constructor.  Creates a grid using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the grid.
 * Arguments:
 *      windowArg - window that the grid will be drawn on.
 *      nameArg - string name of the grid, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmGridClass::AmGridClass( AmWindowClass	*windowArg,
		      AmCoordClass  	*graphicArg, 
		      const char	*nameArg,
		      ...)
{
    
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		count;

    initializeFields( windowArg, graphicArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    setFieldsWrapper( tagValues, count);
    va_end( args);
    SetValid( TRUE);
}

/* fields accepted by AmGridClass which are actually passed on to the 
   constituent tickmarks. */

static char *tickFields[] = { AmNfont, AmNfontSize, AmNfontSlant,
			      AmNfontSlant, AmNfontWeight, AmNfontXScale,
			      AmNfontYScale, AmNtitleFont, 
			      AmNtitleFontSize, AmNtitleFontSlant, 
			      AmNtitleFontWeight, AmNtitleFontXScale, 
			      AmNtitleFontYScale,
			      NULL };

/****************************************************************************
 *
 * Description:
 *      Set the fields for a grid given the tag-values stored in
 *      a tag values array.  First, the generic coords fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  Any of the tag-values which occur in the
 *	list of fields for tickmarks are copied to temporary arrays, 
 *	then set in the tickmarks.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 * Modifications:
 *	probably the dumpFields() function should be used to get the
 *	available tickmarks fields to prevent synchronzation problems
 *	if more fields are added to AmTickClass.
 *
 ****************************************************************************/

int 
AmGridClass::setFields( AmTagValuePtr	tagValues,
		      int		numValues)
{
    int         ii, numFieldDefs, numTickValues, numUnusedValues;
    AmFieldDefPtr fieldDefs;
    AmTagValueRec	tickValues[ MAXTAGS], copyTickValues[ MAXTAGS];

    numValues = AmCoordClass::setFields( tagValues, numValues);
    AmGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* copy over the fields which set the ticks to a new array, but turn off the
   copied flags, because we're just setting the pointers */

    numTickValues = 0;
    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {
    	if ( AmInStrvec( tagValues[ ii].tag, ( const char **) tickFields)) {
    	    tickValues[ numTickValues] = tagValues[ ii];
	    tickValues[ numTickValues].copied = FALSE;
	    tickValues[ numTickValues++].tagCopied = FALSE;
	} else {
	    tagValues[ numUnusedValues++] = tagValues[ ii];
	}
    }

/* since setFields scrunches the array, copy the tag values before 
   setting them in each tickmark */

    if ( numTickValues > 0) {
	memcpy( copyTickValues, tickValues, 
		sizeof( AmTagValueRec)*numTickValues);
	leftTicks->setFields( copyTickValues, numTickValues);
	memcpy( copyTickValues, tickValues, 
		sizeof( AmTagValueRec)*numTickValues);
	rightTicks->setFields( copyTickValues, numTickValues);
	memcpy( copyTickValues, tickValues, 
		sizeof( AmTagValueRec)*numTickValues);
	upperTicks->setFields( copyTickValues, numTickValues);
	memcpy( copyTickValues, tickValues, 
		sizeof( AmTagValueRec)*numTickValues);
	lowerTicks->setFields( copyTickValues, numTickValues);
    }
    return( numUnusedValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a grid, given the tag-values stored in
 *      a tag values array.  First, the coordinates fields are retrived, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The fields which are specific to tickmarks
 *	are pulled from the leftmost tickmarks.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmGridClass::getFields( AmTagValuePtr   tagValues,
		      int           numValues)
{
    int         ii, numFieldDefs, numTickValues, numUnusedValues;
    AmFieldDefPtr fieldDefs;
    AmTagValueRec tickValues[ MAXTAGS];
 
    numValues = AmCoordClass::getFields( tagValues, numValues);
    AmGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
                                       tagValues, numValues);

/* copy over the fields which set the ticks to a new array */ 

    numTickValues = 0;
    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {
    	if ( AmInStrvec( tagValues[ ii].tag, ( const char **) tickFields)) {
    	    tickValues[ numTickValues++] = tagValues[ ii];
	} else {
	    tagValues[ numUnusedValues++] = tagValues[ ii];
	}
    }

/* since the values are set for all ticks, we just get them for the
   ones on the left */

    leftTicks->getFields( tickValues, numTickValues);
    return( numUnusedValues);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a grid.  This is
 *      used in the map editor to dump the grid's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmGridClass::dumpFields( AmFieldLengthPtr    fieldLengths,
                       AnvilBoolean           allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;
 
/* get the fields for the parent class ( screen object) */
 
    numStrings = AmScreenObjClass::dumpFields( fieldLengths, allFields);
 
/* get all the easy to specify fields for this class */
 
    AmGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/****************************************************************************
 *
 * Description:
 *      Set the rectangle for a grid.  The AmGridClass extents rectangle
 *	may extend outside the sizing rectangle, because the tickmarks
 *	may extend outside the rect.
 * Arguments:
 *      x, y, width, height:  new dimensions for the screen rectangle.
 *
 ****************************************************************************/

void 
AmGridClass::SetRect( int	x,
		    int	y,
		    int	width,
		    int	height)
{
    AmRectangleRec const	*leftRect, *rightRect, *upperRect, *lowerRect;

    rect.x = x;
    rect.y = y;
    rect.width = width;
    rect.height = height;
    upperRect = upperTicks->GetRect();
    leftRect = leftTicks->GetRect();
    rightRect = rightTicks->GetRect();
    lowerRect = lowerTicks->GetRect();
    upperTicks->SetRect( rect.x, rect.y - upperRect->height,
			 rect.width, upperRect->height);
    lowerTicks->SetRect( rect.x, rect.y + rect.height,
			 rect.width, lowerRect->height);
    leftTicks->SetRect( rect.x - leftRect->width, rect.y,
			leftRect->width, rect.height);
    rightTicks->SetRect( rect.x + rect.width, rect.y,
			 rightRect->width, rect.height);

/* set the extents rectangle to reflect the tickmarks */

    extentsRect.x = rect.x - leftRect->width;
    extentsRect.y = rect.y - upperRect->height;
    extentsRect.width = rect.width + leftRect->width + rightRect->width;
    extentsRect.height = rect.height + upperRect->height + lowerRect->height;

/* send the user event to the display manager */

    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *      Draw the grid.  Draw the 4 tickmarks, then draw the grid.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/


/* draw the grid */

AmErrorCode 
AmGridClass::draw( AmWindowClass		*w,
		 AmGCClass		*gc,
		 AmRectangleRec const	*exposeRects,
		 int			nRects)
{
    AmErrorCode		errCode;
    float		horizBase, horizSpacing, vertBase, vertSpacing,
			xmin, ymin, xmax, ymax, x, y, spacing;
    int			x0, y0, x1, y1, x2, y2, x3, y3, maxTicks;
    AmTickStyleEnum	horizStyle, vertStyle;
    AmCoordClass		*graphic = ( AmCoordClass *) parent;
    AmRectangleRec const	*tickRect;
    AmRectanglePtr	clipRects;

/* only draw the outline if the object is invalid */

    if ( !IsValid()) {
	return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }
    if ( sameTickSpacing) {

/* set the initial spacing for the tickmarks if it's dictated by tickmarks */ 

	(( AmCoordClass *) parent)->GetCoords( &xmin, &ymin, &xmax, &ymax);
	upperTicks->GetVaFields( AmNmaxTicks, &maxTicks,
				 NULL);
	AmScaleBaseTenSpacing( xmin, xmax, maxTicks, &horizSpacing, &horizBase);
	AmScaleBaseTenSpacing( ymin, ymax, maxTicks, &vertSpacing, &vertBase);
	spacing = MAX( horizSpacing, vertSpacing);
	upperTicks->SetVaFields( AmNtickBase, horizBase,
				 AmNtickSpacing, spacing,
				 NULL);
	lowerTicks->SetVaFields( AmNtickBase, horizBase,
				 AmNtickSpacing, spacing,
				 NULL);
	leftTicks->GetVaFields( AmNmaxTicks, &maxTicks,
				NULL);
	leftTicks->SetVaFields( AmNtickBase, vertBase,
				AmNtickSpacing, spacing,
				NULL);
	rightTicks->SetVaFields( AmNtickBase, vertBase,
				 AmNtickSpacing, spacing,
				 NULL);

    }

/* draw the tickmarks */

    tickRect = upperTicks->GetRect();
    if (( errCode = upperTicks->draw( w, gc, tickRect, 1)) != AmOK) {
	return( errCode);
    }
    tickRect = leftTicks->GetRect();
    if (( errCode = leftTicks->draw( w, gc, tickRect, 1)) != AmOK) {
	return( errCode);
    }
    tickRect = rightTicks->GetRect();
    if (( errCode = rightTicks->draw( w, gc, tickRect, 1)) != AmOK) {
	return( errCode);
    }
    tickRect = lowerTicks->GetRect();
    if (( errCode = lowerTicks->draw( w, gc, tickRect, 1)) != AmOK) {
	return( errCode);
    }
    if ( drawGrid) {
	gc->SetLineWidth( _dLineWidth);
	upperTicks->GetVaFields( AmNtickBase, &horizBase,
				 AmNtickSpacing, &horizSpacing,
				 AmNtickStyle, &horizStyle,
				 NULL);
	leftTicks->GetVaFields( AmNtickBase, &vertBase,
			        AmNtickSpacing, &vertSpacing,
			        AmNtickStyle, &vertStyle,
			        NULL);
	graphic->GetDataPt( rect.x, rect.y + rect.height - 1, &xmin, &ymin);
	graphic->GetDataPt( rect.x + rect.width - 1, rect.y, &xmax, &ymax);

/* get the expose rectangles for the exposed area, and clip them by the
   rectangle for the grid, otherwise the grid will be drawn on a non-exposed
   area on top of whatever objects are there, even if they're above it */

	clipRects = ( AmRectanglePtr) alloca( nRects*sizeof( AmRectangleRec));
	memcpy( clipRects, exposeRects, nRects*sizeof( AmRectangleRec));

/* bail if there's no overlap */

	if ( !AmClipRectVec( &rect, clipRects, &nRects)) {
	    return( AmOK);
	}
	gc->SetClipRectangles( clipRects, nRects);
	gc->SetForegroundColor( &fgColor);
	if ( gridStyle == AmGRID_GRID) {

/* set the line style and dashes if they're applied */

	    gc->SetLineStyle( lineStyle);
	    if ( lineStyle != AmLINE_SOLID) {
		gc->SetDashes( dashes, ndash, 0);
	    }
	    for ( y = vertBase - vertSpacing; 
		  y <= ymax; 
		  y += vertSpacing) {
		for ( x = horizBase - horizSpacing; 
		      x <= xmax; 
		      x += horizSpacing) {

/* draw the grid as individual boxes, since they might be map projected */

		    graphic->GetScreenPt( x, y, 
					  &x0, &y0);
		    graphic->GetScreenPt( x + horizSpacing, y, 
					  &x1, &y1);
		    graphic->GetScreenPt( x + horizSpacing, y + vertSpacing, 
					  &x2, &y2);
		    graphic->GetScreenPt( x, y + vertSpacing, 
					  &x3, &y3);

/* make sure at least one corner is in the box */

		    if (( x0 < rect.x + rect.width) && 
			( y2 < rect.y + rect.height) &&
			( x1 > rect.x) && ( y0 > rect.y)) {
			display->DrawLine( w, gc, x0, y0, x1, y1);
			display->DrawLine( w, gc, x1, y1, x2, y2);
		    }
		}
	    }

/* set the line style back to solid */

	    gc->SetLineStyle( AmLINE_SOLID);
	} else if ( gridStyle == AmGRID_PLUS) {

/* draw pluses where the grid boxes would be */

	    for ( y = vertBase - vertSpacing; 
		  y <= ymax; 
		  y += vertSpacing) {
		for ( x = horizBase - horizSpacing; 
		      x <= xmax; 
		      x += horizSpacing) {
		    graphic->GetScreenPt( x, y, &x0, &y0);
		    if (( x0 > rect.x) && 
			( y0 > rect.y) && 
			( x0 < rect.x + rect.width) && 
			( y0 < rect.y + rect.height)) {

/* make sure the plus is in the box */

			display->DrawLine( w, gc, x0 - plusSize, y0, 
						  x0 + plusSize, y0);
			display->DrawLine( w, gc, x0, y0 - plusSize,
						  x0, y0 + plusSize);
		    }
		}
	    }
	}
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a grid. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmGridClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *	BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmGridClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
			        int		*numDefs)
{
    static AmFieldDefRec GridClassFieldDefinitions[] = {
        { AmNbottomTicks, offsetof( AmGridClass, lowerTicks) },
	{ AmNdashes, offsetof( AmGridClass, dashes),
		     offsetof( AmGridClass, ndash), AmDYNAMIC },
	{ AmNdrawGrid, offsetof( AmGridClass, drawGrid) },
	{ AmNgridStyle, offsetof( AmGridClass, gridStyle) },
        { AmNleftTicks, offsetof( AmGridClass, leftTicks) },
	{ AmNlineStyle, offsetof( AmGridClass, lineStyle) },
	{ AmNlineWidth, offsetof( AmGridClass, _dLineWidth) },
	{ AmNnumDashes, offsetof( AmGridClass, ndash) },
        { AmNrightTicks, offsetof( AmGridClass, rightTicks) },
	{ AmNsameTickSpacing, offsetof( AmGridClass, sameTickSpacing) },
        { AmNtopTicks, offsetof( AmGridClass, upperTicks) }
    };

    *fieldDefs = GridClassFieldDefinitions;
    *numDefs = VECLEN( GridClassFieldDefinitions);
}
