#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* member function implementation for image annotation, which support
   GIF and TIFF images.  
   Modifications:
	Should also support JPEG and Windows Metafile format images,
	and other image types.
*/

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/tiffio.h>
#include <anvil/gif.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/rect.h>
#include <anvil/imageannot.h>
#include <anvil/image.h>
#include <anvil/userevents.h>
#undef NUMCOLORS		/* defined in WINGDI.H and colorhist.h */
#include <anvil/colorhist.h>
#include <anvil/tagstrs.h>


/* macro used by the TIFF library for stepping to the next value in an image */

#define MAXMAXVALUE 1023
#define NEXTSAMPLE \
{ \
    if ( bitsleft == 0) { \
        inp++; \
        bitsleft = 8; \
    } \
    bitsleft -= bitsPerPixel; \
    sample = ( *inp >> bitsleft) & maxval; \
}

/* RGB-unsigned long-pixel true color function */

static ulong_t 
trueColorFn( unsigned long *src, void *clientData)
{
    return( *src);
}

/***************************************************************************
 *
 * Description:
 *	Recursive run-time typing function
 * Arguments:
 *	requested class type 
 * Returns:
 *	TRUE if object is of requested type or a subclass of it
 *
 ***************************************************************************/

AnvilBoolean 
AmImageAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmCIRCLE_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so image annotations can be subclassed.
 *
 ****************************************************************************/

AmImageAnnotClass::AmImageAnnotClass()
{
}

/***************************************************************************
 *
 * Description:
 * 	destructor: delete the source image, display image, and free the
 *	filename if they've been set 
 *
 ***************************************************************************/

AmImageAnnotClass::~AmImageAnnotClass()
{
    if ( _plSrcData != NULL) {
	FREE( _plSrcData);
    }
    if ( _pImage != NULL) {
	delete _pImage;
    }
    if ( _szFile != NULL) {
	FREE( _szFile);
    }
}

/***************************************************************************
 *
 * Description:
 * 	field initializer.  Each class calls the initializer for its
 *	parent class.  These are not constructors because of argument
 *	conflicts with varargs 
 * Arguments:
 *	windowArg - window the image annotation will be drawn on.
 *	nameArg - resource name of the annotation.
 *
 ***************************************************************************/

void
AmImageAnnotClass::initializeFields( AmWindowClass  *windowArg,
                                   const char   *nameArg)
{
    AmAnnotationClass::initializeFields( windowArg, nameArg);

    objectType     = AmIMAGE_ANNOTATION;
    colors         = NULL;
    numColors      = 0;

    _plSrcData     = NULL;
    _pImage        = NULL;
    _szFile        = NULL;
    _fInterpolate  = FALSE;
    _cxSrcPixels   = 0;
    _cySrcPixels   = 0;
    _fFlipRows     = FALSE;
    _fFlipCols     = FALSE;
    _eColormapType = AmPSEUDO_COLOR;
}

/***************************************************************************
 *
 * Description:
 *	constructor to create an image annotation
 * Arguments:
 *	windowArg: window the image annotation will be drawn on.
 *	nameArg: resource name of the annotation.
 *
 ***************************************************************************/

AmImageAnnotClass::AmImageAnnotClass( AmWindowClass *pWindow,
				  const char  *szName,
				  ...)
{
    va_list	args;
    AmTagValueRec	vTagValues[ MAXTAGS];
    int		nTagValues;

    initializeFields( pWindow, szName);
    va_start( args, szName);
    nTagValues = AmMakeTagValueArrayVararg( vTagValues, args);
    va_end( args);
    setFieldsWrapper( vTagValues, nTagValues);
}

/***************************************************************************
 *
 * Description:
 *	read a tiff file.
 * Arugments:
 *	filename - file to read.
 *	width, height - returned image width and height.
 *	depth - returned image depth.
 *	imageData - returned image data.
 *	colors - returned colors used in the image if it's colormapped
 *	(may be a pointer to NULL)
 *	numColors - number of colors used in the image if it's colormapped.
 *	(may be 0)
 *	eColormapType - AmTRUE_COLOR or AmPSEUDO_COLOR.
 * Returns:
 *	TRUE if the image was read properly, FALSE otherwise.
 * Modifications:
 *	should return an error code, and check the malloc to allocate
 *	the image.  This should also probably be moved to the imagefile
 *	subdirectory.
 *
 ***************************************************************************/

AnvilBoolean 
AmReadTIFF( char			*filename,
	  int			*width,
	  int			*height,
	  int			*depth,
	  uchar_t		**imageData,
	  AmColorCellPtr		*colors,
	  int			*numColors,
	  AmColormapTypeEnum	*eColormapType)
{
    ushort_t	photometric, *redCmap, *greenCmap, *blueCmap,
		bitsPerPixel, samplesPerPixel;
    TIFF	*tif;
    int		maxval, row, col, ii, bitsleft, lastNonZeroMarker;
    uchar_t	*outp, *inp, sample, *inbuf;
    AnvilBoolean	greyscale;
    FILE	*fp = fopen( "tiffdebug", "w");

    if (( tif = TIFFOpen( filename, "r")) == NULL) {
	return( FALSE);
    }
    TIFFPrintDirectory( tif, fp, TIFFPRINT_COLORMAP);
    fclose( fp);

/* get the number of bits per pixel, samples per pixel, and the photometric */

    if  ( !TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel)) {
	bitsPerPixel = 1;
    }
    *depth = bitsPerPixel;
    if  ( !TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel)) {
	samplesPerPixel = 1;
    }
    if  ( !TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
	AN_DEBUG( 1, ( "errorGettingPhotometric %s", filename));
	return( FALSE);
    }

/* get the dimensions of the image */

    TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, width);
    TIFFGetField( tif, TIFFTAG_IMAGELENGTH, height);
    maxval = ( 1 << bitsPerPixel) - 1;
    if ( maxval == 1 && samplesPerPixel == 1) {
	greyscale = TRUE;
    } else {
	switch ( photometric) {
        case PHOTOMETRIC_MINISBLACK:
        case PHOTOMETRIC_MINISWHITE:
            greyscale = TRUE;
            break;
        case PHOTOMETRIC_PALETTE:

/* if it's colormapped, then read the colormap into R, G and B vectors */

            if ( !TIFFGetField( tif, TIFFTAG_COLORMAP,
                                &redCmap, &greenCmap, &blueCmap)) {
		AN_DEBUG( 1, ( "errorGettingCmap %s", filename));
                return( FALSE); 
            }
            *numColors = maxval + 1;

/* if the colormap trails with 0's, then copy to the AmLast non-zero value */

            lastNonZeroMarker = 0;
            for ( ii = 0; ii < *numColors; ii++) {
                if (( redCmap[ ii]  != 0) ||
                    ( greenCmap[ ii] != 0) ||
                    ( blueCmap[ ii] != 0)) {
                    lastNonZeroMarker = ii;
                }
            }
	    *numColors = lastNonZeroMarker + 1;
	    *colors = ( AmColorCellPtr) 
		malloc( sizeof( AmColorCellRec)*( *numColors));
	    for ( ii = 0; ii < *numColors; ii++) {
		(*colors)[ ii].red = redCmap[ ii];
		(*colors)[ ii].green = greenCmap[ ii];
		(*colors)[ ii].blue = blueCmap[ ii];
	    }
            maxval = MAXMAXVALUE;
            greyscale = FALSE;
	    *eColormapType = AmPSEUDO_COLOR;
            break;
        case PHOTOMETRIC_RGB:
            greyscale = FALSE;
	    *eColormapType = AmTRUE_COLOR;
            break;
        default:
            AN_DEBUG( 1, ( "unknownPhotometric %s", filename));
            return( FALSE);
        }
    }

/* allocate a buffer to read the scanline, then read the file scanline by
   scanline RGB images map to 32 bit values, all others to 8-bit values */

    if ( photometric == PHOTOMETRIC_RGB) {
	*imageData = ( uchar_t *) 
	    malloc(( *width)*( *height)*sizeof( unsigned long));
    } else {
	*imageData = ( uchar_t *) malloc(( *width)*( *height));
    }
    inbuf = ( uchar_t *) malloc( TIFFScanlineSize( tif));
    outp = *imageData;
    for ( row = 0; row < *height; row++) {
        if ( TIFFReadScanline( tif, inbuf, row, 0) < 0)  {
	    AN_DEBUG( 1, ( "badDataRow %s", filename));
            return( FALSE);
        }
        bitsleft = 8;
        inp = inbuf;
        switch ( photometric) {
        case PHOTOMETRIC_MINISBLACK:
            for ( col = 0; col < *width; col++, outp++) {
                NEXTSAMPLE
                *outp = sample;
            }
            break;
        case PHOTOMETRIC_MINISWHITE:
            for ( col = 0; col < *width; col++, outp++) {
                NEXTSAMPLE
                *outp = maxval - sample;
            }
            break;
        case PHOTOMETRIC_PALETTE:
            for ( col = 0; col < *width; col++, outp++) {
                NEXTSAMPLE
                *outp = sample;
            }
            break;
        case PHOTOMETRIC_RGB:
            for ( col = 0; col < *width; col++) {
                outp++;
                NEXTSAMPLE
                *outp++ = sample;
                NEXTSAMPLE
                *outp++ = sample;
                NEXTSAMPLE
                *outp++ = sample;
                if ( samplesPerPixel == 4) {
                    NEXTSAMPLE
                }
            }
            break;
        default:
            break;
        }
    }

/* free the scanline buffer and close the TIFF file descriptor */

    FREE( inbuf);
    TIFFClose( tif);
    return( TRUE);
}

/***************************************************************************
 *
 * Description:
 *	set the fields specific to an image annotation (subclassed from
 *	annotation)  If the file has changed, then read the source
 *	image in GIF or TIFF format and project it to the display.
 *	If the source image is in 8-bit form, is is turned into 24-bit
 *	data before it's projected.  This function is called by the 
 *	varargs constructor and SetVaFields().
 * Arguments:
 *	tagValues: vector of tag-value pairs to set members in the class.
 *		   unparsed pairs are left in the vector.
 *	numValues: length of the tag-value pair vector.
 * Returns:
 *	number of unparsed values.
 * Recommended changes:
 *	Annotation only supports 8 and 24-bit images.  It should support
 *	1, 4, 8, 12, 16-bit images and more formats.
 *
 ***************************************************************************/

int 
AmImageAnnotClass::setFields( AmTagValuePtr tagValues,
			    int         numValues)
{
    int             numFieldDefs, width, height, depth, ii, localNumColors;
    AmFieldDefPtr     fieldDefs;
    FILE	    *fp;
    uchar_t	    *imageData, localColorMap[ 3][ MAXCOLORMAPSIZE];
    ulong_t	    *longImageData;
    AmColorCellPtr    imageColors = NULL;
 
    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmImageAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* if the filename was set, then load the image file */

    if ( AmChanged( &_szFile)) {
	if (( fp = fopen( _szFile, "r")) == NULL) {
	    SetValid( FALSE);
	} else {

/* try a GIF file AmFirst, which becomes an 8-bit pseudocolor image */

	    if ( ReadGIF( fp, 1, &imageData, &width, &height, localColorMap,
			  &localNumColors)) {
		imageColors = ( AmColorCellPtr) 
				malloc( sizeof( AmColorCellRec)*localNumColors);
		for ( ii = 0; ii < localNumColors; ii++) {
		    imageColors[ ii].red = localColorMap[ 0][ ii];
		    imageColors[ ii].green = localColorMap[ 1][ ii];
		    imageColors[ ii].blue = localColorMap[ 2][ ii];
		}
		depth = 8;
		fclose( fp);
		_eColormapType = AmPSEUDO_COLOR;
	    } else {

/* now try a TIFF file, which can be 8 or 24-bit */ 

		fclose( fp);
		if ( !AmReadTIFF( _szFile, &width, &height, &depth, &imageData,
				&imageColors, &localNumColors,
                                &_eColormapType)) {
		    SetValid( FALSE);
		}
	    }
	    if ( IsValid()) {

/* testing for pseudo color isn't the best, and we shouldn't be hardcoded 
   to 8-bit images */

		if ( _eColormapType == AmPSEUDO_COLOR) {

/* we need to upgrade the image to a 24-bit image */

		    SetVaFields( AmNcolors, imageColors,
				 AmNnumColors, localNumColors,
				 NULL);
		    AmImage8to24( imageData, width, height, 
				colors, localNumColors, &longImageData);
		    FREE( imageData);
		    imageData = ( uchar_t *) longImageData;
		    if ( imageColors != NULL) {
			FREE( imageColors);
		    }
		} 

/* copy the new data to the source image and dimensions, and project the
   image */

		if ( _plSrcData != NULL) {
		    FREE( _plSrcData);
		} 

		_plSrcData = ( ulong_t *) imageData;
		_cxSrcPixels = width;
		_cySrcPixels = height;
		projectImage();
	    }
	}
    } else if ( AmChanged( &_fInterpolate, &_fFlipRows, &_fFlipCols)) {
	projectImage();
    }

    return( numValues);
}

/***************************************************************************
 *
 * Description:
 *	retrieve the values from the class specified by the tag-values
 *	array, where the values are pointers to the values to be returned.
 * Arguments:
 *	tagValues: vector of tag-value pairs to set members in the class.
 *		   unparsed pairs are left in the vector.
 *	numValues: length of the tag-value pair vector.
 * Returns:
 *	number of unparsed values.
 *
 ***************************************************************************/
 
int 
AmImageAnnotClass::getFields( AmTagValuePtr tagValues,
			    int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmImageAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/***************************************************************************
 *
 * Description:
 *	create the destintation display image and project the source image 
 *	to the display.
 *
 ***************************************************************************/

void 
AmImageAnnotClass::projectImage()
{
    AmRectangleRec	dstRect;
    AmRealRectRec		srcRect;

    if ( _pImage == NULL) {

	_pImage = display->CreateImage( rect.width, rect.height,
					display->GetDepth(), NULL);
	if ( _pImage == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    SetValid( FALSE);
	}
    }

    srcRect.x = 0.0;
    srcRect.y = 0.0;
    srcRect.width = ( float) _cxSrcPixels;
    srcRect.height = ( float) _cySrcPixels;

    dstRect.x = 0;
    dstRect.y = 0;
    dstRect.width = _pImage->GetWidth();
    dstRect.height = _pImage->GetHeight();

    if ( _fInterpolate) {
	_pImage->bilinMatrixImage(( uchar_t *) _plSrcData,
                                       sizeof( ulong_t), 
                                       _cySrcPixels, _cxSrcPixels, 
				       _fFlipRows, _fFlipCols, 
				       &srcRect, &dstRect, 
				       ( AmColorFn) trueColorFn,
				       NULL, 1.0);
    } else {
	_pImage->resizeMatrixImage(( uchar_t *) _plSrcData,
                                        sizeof( ulong_t),
					_cySrcPixels, _cxSrcPixels, 
					_fFlipRows, _fFlipCols, 
					&srcRect, &dstRect, 
					( AmColorFn) trueColorFn,
					NULL, 1.0);
    }
}

/****************************************************************************
 *
 * Description:
 * 	set the rectangle for the image, and resize the source image to 
 *	the destination screen rectangle.
 * Arguments:
 *      x, y, width, height - new rectangle dimensions
 *
 ****************************************************************************/

void 
AmImageAnnotClass::SetRect( int x,
			  int y,
			  int cx,
			  int cy)
{
    if ( !IsValid()) {
	return;
    }

/* Only reallocate and regenerate the image if the image is NULL or if the
   rectangle size has changed, or if the DPI has changed. */

    if ( _pImage != NULL) {

	if (( rect.width != cx) || ( rect.height != cy) || 
	    ( _pImage->GetDPI() != _pImage->GetDisplay()->GetDPI())) {
	    delete _pImage;
	    _pImage = NULL;
	}

    }

    rect.x = x;
    rect.y = y;
    rect.width = cx;
    rect.height = cy;

    extentsRect = rect;

    if ( _pImage == NULL) {
	projectImage();
    }

    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *      Draw the image annotation to the screen, do a putimage on
 *	the exposed rectangles.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - number of exposed rectangles
 * Returns:
 *      AmOK if everything's OK.
 *
 ****************************************************************************/

AmErrorCode 
AmImageAnnotClass::draw( AmWindowClass	    	*w,
		       AmGCClass			*gc,
		       AmRectangleRec const   	*exposeRects,
		       int			nRects)
{

/* If this object is invalid only draw the outline. */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

/* Return in the degenerate cases. */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

    display->DrawImage( w, gc, 0, 0,
                        rect.x, rect.y,
                        rect.width, rect.height,
                        _pImage);

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an image annotation.  This is
 *      used in the map editor to dump the ticks's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmImageAnnotClass::dumpFields( AmFieldLengthPtr     fieldLengths,
			     AnvilBoolean		allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmImageAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an image annotation. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmImageAnnotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmImageAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
				      int          *numDefs)
{
    static AmFieldDefRec ImageAnnotClassFieldDefinitions[] = {
	{ AmNfilename, offsetof( AmImageAnnotClass, _szFile), 0, AmDYNAMIC },
	{ AmNflipCols, offsetof( AmImageAnnotClass, _fFlipCols) },
	{ AmNflipRows, offsetof( AmImageAnnotClass, _fFlipRows) },
	{ AmNinterpolate, offsetof( AmImageAnnotClass, _fInterpolate) }
    };
    *fieldDefs = ImageAnnotClassFieldDefinitions;
    *numDefs = VECLEN( ImageAnnotClassFieldDefinitions);
}
