#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <malloc.h>
#include <math.h>
#include <stddef.h>
#include <string.h>
#include <anvil/files.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/legend.h>
#include <anvil/imageplot.h>
#include <anvil/imagelegend.h>
#include <anvil/colorcell.h>

/*
 * Function: IsType( AmObjectTypeEnum eType)
 *
 * Purpose:
 *   To verify whether this object is, or is derived from the passed
 *   in object type.
 * 
 * Paramters:
 *   objectType eType - Enumeration value of an Anvil object type. 
 *
 * Returns:
 *   AnvilBoolean - TRUE if the object is of this type or a sublcass of it,
 *             FALSE if not.
 *
 * Comments:
 *   A recursive run-time type function which for each class determines
 *   if the requested type is of this class, or of the parent's
 *   class (and so on).  This should be used instead of GetObjectType()
 *   wherever possible, to allow subclassing.
 *
 */

AnvilBoolean
AmImageLegendClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmIMAGE_LEGEND) {
        return( TRUE);
    } else {
        return( AmLegendClass::IsType( type));
    }
}

/*
 * Function: initializeFields( pWindow, pImagePlot, sNname)
 *
 * Purpose:
 *   Initializes the member variables of the legend class. 
 * 
 * Paramters:
 *   AmWindowClass    *pWindow    - Window on which the image legend will be drawn.
 *   AmImagePlotClass *pImagePlot - Image plot for which the legend is created.
 *   const char     *szName     - String name used to identify the image legend 
 *                                for resources.
 *
 * Returns:
 *   -nothing-
 *
 * Comments:
 *   The intializeFields() functions are created for each class, and
 *   each one calls the one for the "parent" class.  The AmScreenObjClass
 *   intializeFields() function is the topmost one that is called.  We
 *   can't use two-parameter constructors, because since we use varargs
 *   constructors, there is a naming ambiguity.  The image legend calls
 *   the AmLegendClass initializer, then initializes the the image legend
 *   specific fields, such as the legend type, and item images.
 *
 *   The values are initialized to some default values.  The image
 *   legend is by default a continuous legend.  That means, there is
 *   one main item which shows a continuous color bar as opposed to
 *   several items which show one color each.
 */

void 
AmImageLegendClass::initializeFields( AmWindowClass    *pWindow,
				    AmImagePlotClass *pPlot,
				    const char     *szName)
{

/* Initialize the fields of the parent class. */

    AmLegendClass::initializeFields( pWindow, pPlot, szName);

/* Initialize the fields specific to the image legend. */

    objectType = AmIMAGE_LEGEND;
    legendType = AmLEGEND_CONTINUOUS;
    images = NULL;
}

/*
 * Function: AmImageLegendClass( AmWindowClass *, ImageplotClass *,
 *                             const char *, ...)
 *
 * Purpose:
 *   Constructor for the image legend class. 
 * 
 * Paramters:
 *   pWindow    - A pointer to the window that the legend is created on. 
 *   pImagePlot - A pointer to the image plot that the legend is created
 *                for.
 *   szName     - The name of the image legend.
 *   ...        - An unspecified amount of initialization values.
 *
 * Return Value:
 *   -nothing-
 * 
 * Comments:
 *   The constructor makes sure that the passed in plot class is
 *   an image plot, or is NULL.  These are the only two allowed
 *   values.  If the plot argument is NULL the legend is a stand-
 *   alone legend and the display parameters have to be 
 *   explicitly supplied.
 */

AmImageLegendClass::AmImageLegendClass( AmWindowClass		*windowArg,
				    AmImagePlotClass	*plotArg,
				    const char		*nameArg,
				    ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    if ( plotArg == NULL || plotArg->IsType( AmIMAGE_PLOT)) {

    /* Initialize the image legend member values. */

        initializeFields( windowArg, plotArg, nameArg);

    /* Parse the supplied initialization arguments. */ 

	va_start( args, nameArg);
	numValues = AmMakeTagValueArrayVararg( tagValues, args);
	va_end( args);
	setFieldsWrapper( tagValues, numValues);

	SetValid( TRUE);
    } else {

    /* Invalid parent option. */

	SetValid( FALSE);
    }
}

/*
 * Function: ~AmImageLegendClass()
 *
 * Purpose:
 *   Destructor for the image legend class. 
 * 
 * Paramters:
 *   -none-
 *
 * Return Value:
 *   -nothing-
 * 
 * Comments:
 *   The destructor frees any allocated memory.  Most notably the
 *   vector if images created for each of the image legend items.
 */

AmImageLegendClass::~AmImageLegendClass()
{
    int	ii;

/* Delete the images from the NULL termianted vector, if they've been
   created */

    if ( images != NULL) {
	for ( ii = 0; images[ ii] != NULL; ii++) {
	    delete images[ ii];
	}
    }
    free( images);
}

/*
 * Function: Update( int *, int)
 *
 * Purpose:
 *   Update gets called when either the legend's parent plot or the
 *   legend's values get changed.  It update the values of the legend
 *   and lays out the legend's rectangles.
 * 
 * Paramters:
 *   viEventCodes - A vector of integer event codes.
 *   nEventCodes  - The number of events in the vector.
 *
 * Return Value:
 *   Returns a AmOK if no error was encountered and returns
 *   the error code of the error otherwise.
 * 
 * Comments:
 *   If the legend has a parent plot then the important values are
 *   retrieved from it (the binvec, table, colorIndex, and so on);
 *   however, if the legend is stand alone, the values have to be
 *   set before Update gets called.  In either case, Update checks
 *   to make sure that the right number of labels are supplied, and,
 *   if they are not, creates the right number of labels. Finally,
 *   the layout rectangles are re-generated by a call to layoutRects.
 *
 * Errors:
 *   AmMALLOC_FAILED - Failed to allocate the image class vector.
 *   *               - Any error code that can be generated by
 *                     AmIndexesToStrings().
 */

AmErrorCode
AmImageLegendClass::Update( int	*viEventCodes,
			  int	nEventCodes)
{
    AmDataGraphicClass *pGraphic = ( AmDataGraphicClass *) parent;
    int              ii, numWanted = 0;

/* Remove the previously created images. */

    if ( images != NULL) {
        for ( ii = 0; images[ ii] != NULL; ii++) {
            delete images[ ii];
        }
        free( images);
        images = NULL;
    }

/* If the image legend has a parent, reget all the values. */

    if ( pGraphic != NULL) {

        AmDestroyIndexData( &colorIndex);

        if ( colors != NULL) {
            free( colors);
            colors = NULL;
        }

    /* Free the NULL-terminated vector of labels if they're not NULL. */

        numLabels = 0;
        if ( labels != NULL) {
            AmFreeStrVec( labels);
            labels = NULL;
        }

    /* Get the new definitions from the parent. */

        pGraphic->GetVaFields( AmNcolorsUnits, &colorIndex.units, 
                               AmNcolorsTable, &colorIndex.table,
                               AmNcolorsBinvec, &colorIndex.binvec,
                               AmNcolorsStrings, &colorIndex.strings,
                               AmNnumColors, &numColors,
                               AmNcolors, &colors,
                               AmNmin, &colorIndex.min,
                               AmNmax, &colorIndex.max,
                               AmNminClip, &colorIndex.minClip,
                               AmNmaxClip, &colorIndex.maxClip,
                               NULL);
        colorIndex.length = numColors;
    } else {

    /* for a non-parented legend, Free the NULL-terminated vector of 
       labels if they're not NULL. */
	
        numLabels = 0;
        if ( labels != NULL) {
            AmFreeStrVec( labels);
            labels = NULL;
        }
    }

/*
 * Note that in the editable case all the information is supposedly 
 *   supplied; however, we make certain here that the labels
 *   do indeed exist (in the correct amount).  In the non-editable
 *   case, the labels get generated here.
 */

    minClip = colorIndex.minClip;
    maxClip = colorIndex.maxClip;

    if ( colorIndex.binvec != NULL) {

    /* There is a binvec associated with the legend. */

        if ( legendType == AmLEGEND_CONTINUOUS) {

        /* A continuous legend consists of 3 items.  The minimum
           cutoff, the main item, and the maximum cutoff. */

            numItems = 3;
            numWanted = 2 + labelsPerItem;
            if ( numLabels != numWanted) { 
                if ( labels != NULL) {
                    AmFreeStrVec( labels);
                }
                numLabels = numWanted;
                labels = AmIndexesToStrings( &colorIndex, numLabels, legendType);
            }
        } else {

        /* A non-continuous legend consists of as many items as there
           are colors. */

            numItems = colorIndex.length;
            numWanted = numItems;
            if ( numLabels != numWanted) {
                if ( labels != NULL) {
                    AmFreeStrVec( labels);
                }
                numLabels = numWanted;
                labels = AmIndexesToStrings( &colorIndex, numLabels, legendType);
            }
        }
    } else if ( colorIndex.table != NULL) {

    /* A table is associated with the legend.  This means the
       legend has to be distinct, since each color value has
       a different meaning.  This also means the number of
       items is equal to the number of colors.  */

	legendType = AmLEGEND_DISTINCT;
        numItems = colorIndex.length;
        numWanted = colorIndex.length;

        if ( numLabels != numWanted) {
            if ( labels != NULL) {
                AmFreeStrVec( labels);
            }
            numLabels = numWanted;
	    if ( colorIndex.strings == NULL) {
		labels = AmIndexesToStrings( &colorIndex, numLabels,
                                           AmLEGEND_DISTINCT);
	    } else {

            /* The assumption is that colorIndex.strings has 
               colorIndex.length labels, otherwise there is
               trouble ahead. */

		labels = AmCopyStrVec( (const char **)colorIndex.strings);
	    }
        }
    } else {

    /* A min and a max value are associated with the legend. */ 

        if ( legendType == AmLEGEND_CONTINUOUS) {

        /* A continuous legend consists of 3 items.  The minimum
           cutoff, the main item, and the maximum cutoff. */

            numItems = 3;
            numWanted = 2 + labelsPerItem;
        } else {

        /* A non-continuous legend consists of as many items as there
           are colors. */

            numItems = colorIndex.length;
            numWanted = colorIndex.length;
        }

        /* Make sure that there are enough labels.  If there aren't 
           generate the right number.  This even applies to stand-
           alone legends. */

        if ( numLabels != numWanted) { 
            if ( labels != NULL) {
                AmFreeStrVec( labels);
            }
            numLabels = numWanted;
            labels = AmIndexesToStrings( &colorIndex, numLabels, legendType);
        }
    }

/* Make sure that the labels exist, or were created. */

    if (( labels == NULL) && ( numLabels != 0)) {

    /* Since the labels don't exist, there must have been a problem
       creating them - the error code should be set already. */

        return( ANVIL_ERROR_CODE);
    }

/* Layout the legend's rectangles. */

    layoutRects();

/* Create the images needed to draw the legend. */

    images = ( AmImageClass **) 
	malloc( sizeof( AmImageClass *) * ( numItems + 1));
    if ( images == NULL) {
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    for ( ii = 0; ii < numItems; ii++) {
        images[ ii] = display->CreateImage( itemRects[ ii].width,
                                            itemRects[ ii].height,
                                            display->GetDepth(),
                                            NULL);
    }

    images[ numItems] = NULL;

    if (( legendType == AmLEGEND_CONTINUOUS) && ( colorIndex.binvec != NULL)) {
        fillImagesCB();
    } else {
        fillImages();
    }

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for an image legend, given the tag-values stored in
 *      a tag values array.  First, the legend fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  If the rectangle width or height have been
 *	changed, then set the rectangle.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int
AmImageLegendClass::setFields( AmTagValuePtr tagValues,
                             int         numValues)
{
    int         numFieldDefs, tempWidth, tempHeight;
    AmFieldDefPtr fieldDefs;

    numValues = AmLegendClass::setFields( tagValues, numValues);
    AmImageLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
                    tagValues, numValues);
    tempWidth = rect.width;
    tempHeight = rect.height;

    if ( AmChanged( &rect.width, &rect.height, NULL)) {
	Update();
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for an image legend, given the tag-values 
 *	stored in a tag values array.  First, the scale fields are 
 *	retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int
AmImageLegendClass::getFields( AmTagValuePtr tagValues,
                             int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmLegendClass::getFields( tagValues, numValues);
    AmImageLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an image legend.  This is
 *      used in the map editor to dump the image legend's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmImageLegendClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			      AnvilBoolean		allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* Get the fields for the parent class ( legend) */

    numStrings = AmLegendClass::dumpFields( fieldLengths, allFields);

/* Get all the easy to specify fields for this class */

    AmImageLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 * 	Draw a single legend item for an image legend.  The Draw()
 *	function is implemented in AmLegendClass, and calls DrawItems()
 *	member function as a virtual, which calls this.
 * Arguments:
 *	pWindow - window to draw the item to.
 *	pGC - GC to draw the item with.
 *	itemNumber - index of the item.
 *	pItemRect - item's screen rectangle.
 *
 *************************************************************************/

AmErrorCode
AmImageLegendClass::DrawItem( AmWindowClass        *pWindow,
                            AmGCClass            *pGC,
                            int                iItemNumber,
                            AmRectangleRec const *pItemRect)
{

/* Copy the legend item's image to the window. */

    display->DrawImage( pWindow, textGC,
                        0, 0,
                        pItemRect->x, pItemRect->y,
                        pItemRect->width, pItemRect->height,
                        images[ iItemNumber]);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 * 	Draw all the legend items. Called by AmLegendClass::DrawItems()
 *	as a virtual function.
 * Arguments:
 *	pWindow - window to draw to.
 *	pGC - GC to draw with.
 *	vItemRect - vector of item rectangles.
 * Returns:
 *	AmOK.
 *
 *************************************************************************/

AmErrorCode
AmImageLegendClass::DrawItems( AmWindowClass        *pWindow,
                             AmGCClass            *pGC,
                             AmRectangleRec const *vItemRect)
{
    int ii;

    for ( ii = 0; ii < numItems; ii++) {

    /* Skip the AmFirst and AmLast item if min- or max-clipping are
       set.  Take into account that the items can be in either
       increasing or decreasing order. */

        if ( ( ii == 0) && 
             ((( minClip == TRUE) && ( increaseItems == FALSE)) ||
              (( maxClip == TRUE) && ( increaseItems == TRUE)))) {
            continue;
        } else  if ( ( ii == numItems - 1) && 
             ((( minClip == TRUE) && ( increaseItems == TRUE)) ||
              (( maxClip == TRUE) && ( increaseItems == FALSE)))) {
            continue;
        }

        DrawItem( pWindow, pGC, ii, &vItemRect[ ii]);
    }

    return( AmOK);
}


/*************************************************************************
 *
 * Description:
 * 	Fill an area of an image with a single color.  Note that the 
 *	coordinates are in screen coordinates and not, as in most other 
 *	cases, in virtual coordinates. 
 * Arguments:
 *	pImage - image to fill.
 *	xStart, yStart - upper left hand corner to start fill.
 *	cx, cy - dimensions of rectangle to fill.
 *	col - color cell to fill the image with.
 * Returns:
 * 	AmOK
 *
 *************************************************************************/

static AmErrorCode
fillImage( AmImageClass   *pImage,
           int          xStart,
           int          yStart,
           int          cx,
           int          cy,
           AmColorCellPtr col)
{
    char *pcData;
    int  xOffset, x, y, cBitsPixel, cBytesLine, cBytesPixel;

    if ( pImage->GetDepth() != 1) {

        cBitsPixel = pImage->GetBitsPerPixel();
        cBytesPixel = ( cBitsPixel + 7) >> 3;    /* Round up */
        cBytesLine = pImage->GetBytesPerLine();

        pcData = pImage->AmGetData();
        pcData += yStart * cBytesLine + xStart * cBytesPixel;

        xOffset = cBytesLine - cx * cBytesPixel;

        for ( y = 0; y < cy; y++) {
            for ( x = 0; x < cx; x++) {
		ASSIGN_PIXEL( *col, pcData);
                pcData += cBytesPixel;
            }
            pcData += xOffset;
        }

        pImage->SetDirty( TRUE);

    }

    return( AmOK);
}

/*
 * Function: fillImages()
 *
 * Purpose:  
 *   Fills the images of the image legend with the appropriate colors.
 *
 * Parameters:
 *   -none-
 *
 * Return Value:
 *   An AmErrorCode indicating success (AmOK) or failure (the error code).
 *
 * Comments:
 *   This fill function is used for distinct image legends, as well as, 
 *   continuous linearly scaled legends.
 *
 *   Since FillImage takes screen coordinates everything needs to be
 *   converted from virtual coordinates to screen cordinates.
 */

AmErrorCode
AmImageLegendClass::fillImages()
{
    int          ii, iTotalSize, iSizePerColor=0, iColor,
                 iRemainder, iPos, iMaxPos, iSize,
                 x, y, cx, cy;
    double       dFractRemainder, dFractSize=0.0;

/*
 * Figure out the total size in pixels used to draw the items.
 *   We loop through all the items here in the case of items
 *   being of various sizes (as opposed to homogeneous sizes).
 */

    iTotalSize = 0;
    for ( ii = 1; ii < numItems - 1; ii++) {

        iSize = ( vertical ) ? itemRects[ ii].height : itemRects[ ii].width;

        iSize = display->ScreenCoord( iSize);
        iSize = display->DimInPixels( iSize);
      
        iTotalSize += iSize;
    }

/* Calculate the number of pixels per color, and the number of fractional
   pixels that get lost in the round off. */

    if ( colorIndex.length > 2) {
        iSizePerColor = iTotalSize / ( colorIndex.length - 2);
        dFractSize = iTotalSize / ( double) ( colorIndex.length - 2);
        dFractSize -= iSizePerColor;
    }
   
    iColor = 1;
    iRemainder = 0;
    dFractRemainder = 0;

/* Fill the AmFirst image - it is either the min- or max-cutoff item. */

    cx = display->DimInPixels( display->ScreenCoord( itemRects[ 0].width));
    cy = display->DimInPixels( display->ScreenCoord( itemRects[ 0].height));

    fillImage( images[ 0], 0, 0, cx, cy,
               &colors[ ( increaseItems) ? colorIndex.length - 1 : 0]);

/* Fill all the images except for the AmFirst and the AmLast image. */

    for ( ii = 1; ii < numItems - 1; ii++) {

        cx = display->DimInPixels( display->ScreenCoord( itemRects[ ii].width));
        cy = display->DimInPixels( display->ScreenCoord( itemRects[ ii].height));

        iPos = 0;
        iMaxPos = ( vertical) ? cy : cx;

        while (( iPos < iMaxPos) && iColor < ( colorIndex.length - 1)) {

            if ( iRemainder > 0) {

            /* The previous color bar spilled over into this legend item. */

                iSize = iRemainder;
                iRemainder = 0;
            } else {

            /* Get the size for the color bar. */

                iSize = iSizePerColor;

            /* Keep track of the fractional round-off, once the total
               roud-off exceeds 1.0 increase that particular color bar
               by 1 pixel. */

                dFractRemainder += dFractSize;
                if ( dFractRemainder >= 1.0) {
                    iSize += 1;
                    dFractRemainder -= 1;
                }
            } 

            /* If the size of this color bar doesn't entirely fit into
               the image item, spill the bar over into the next image
               item. */

            if ( iColor == ( colorIndex.length - 2)) {
                iSize = iMaxPos - iPos;
            } else if ( iPos + iSize > iMaxPos) {
                iRemainder = iPos + iSize - iMaxPos;
                iSize = iMaxPos - iPos;
            }

        /* Set the dimensions for the color bar. */ 

            if ( vertical == TRUE) {
                x = 0;
                y = iPos;
                /* cx is already set correctly. */
                cy = iSize;
            } else {
                x = iPos;
                y = 0;
                cx = iSize;
                /* cy is already set correctly. */
            }

            fillImage( images[ ii], x, y, cx, cy,
                       &colors[ ( increaseItems) ?
                                    colorIndex.length - 1 - iColor : iColor]);

        /* If we have drawn the entire color band, go to the next color. */

            if ( iRemainder == 0) {
                iColor++;
            }

            iPos += iSize; 
        }
    }

/* Fill the AmLast image - it is either the min- or max-cutoff item. */

    cx = display->ScreenCoord( itemRects[ numItems - 1].width);
    cx = display->DimInPixels( cx); 

    cy = display->ScreenCoord( itemRects[ numItems - 1].height);
    cy = display->DimInPixels( cy);

    fillImage( images[ numItems - 1], 0, 0, cx, cy,
               &colors[ ( increaseItems) ? 0 : colorIndex.length - 1]);

    return( AmOK);
}


/*
 * Function: fillImagesCB()
 *
 * Purpose:  
 *   Fills the images of the image legend with the appropriate colors.
 *   (CB = continuous binvec).
 *
 * Parameters:
 *   -none-
 *
 * Return Value:
 *   An AmErrorCode indicating success (AmOK) or failure (the error code).
 *
 * Comments:
 *   This fill function is used when drawing a continuous legend associated
 *   with a binvec.  That means the color bands do not have to be (and in most
 *   cases are not) of equal size; furthermore, since it is a continuous
 *   legend there are at most three items: the min cutoff item, the main item,
 *   and the max cutoff item.
 *
 *   Since FillImage takes screen coordinates everything needs to be
 *   converted from virtual coordinates to screen cordinates.
 */

AmErrorCode
AmImageLegendClass::fillImagesCB()
{
    int          ii, iSize, iPos, iMaxPos, iItem,
                 x, y, cx, cy;
    double       dFractRemainder, dRange, dPixelPerRange, dCurrentRange;

/*
 * A continuous legend is composed of three items.  The min cutoff item
 *   (< min) the range item (min to max), and the max cutoff item (> max).
 */

/* Draw the min- or max-cutoff item depending on whether the legend
   is increasing or decreasing. */

    cx = display->DimInPixels( display->ScreenCoord( itemRects[ 0].width));
    cy = display->DimInPixels( display->ScreenCoord( itemRects[ 0].height));

    fillImage( images[ 0], 0, 0, cx, cy,
               &colors[ ( increaseItems) ? colorIndex.length - 1 : 0]);

/* If it exists, draw the range item. */

    if ( numItems > 2 && colorIndex.length > 2) {

        cx = display->DimInPixels( display->ScreenCoord( itemRects[ 1].width));
        cy = display->DimInPixels( display->ScreenCoord( itemRects[ 1].height));

        dRange = colorIndex.binvec[ colorIndex.length - 2] -
                 colorIndex.binvec[ 0];
        dPixelPerRange = (( vertical) ? cy : cx) / dRange;

        iPos = 0;
        iMaxPos = ( vertical) ? cy : cx;
        dFractRemainder = 0;

        for ( ii = 1; ii < colorIndex.length - 1; ii++) {

        /* Get the size for the color bar. */

            iItem = ( increaseItems)  ? iItem = colorIndex.length - 1 - ii : ii;
            dCurrentRange = colorIndex.binvec[ iItem] -
                            colorIndex.binvec[ iItem - 1];

            iSize = ( int)( dCurrentRange * dPixelPerRange);

        /* Keep track of the fractional round-off, once the total
           roud-off exceeds 1.0 increase that particular color bar
           by 1 pixel. */

            dFractRemainder += ( dCurrentRange * dPixelPerRange) - iSize;
            if ( dFractRemainder >= 1) {
                iSize += 1;
                dFractRemainder -= 1;
            }

        /* Make sure that the color bars do not exceed the length of the
           image item and the AmLast color bar fills the image item 
           completely. */

            if (( ii == colorIndex.length - 2) &&
                ( iPos + iSize != iMaxPos)) {
                iSize = iMaxPos - iPos;
            } else if ( iPos + iSize > iMaxPos) {
                iSize = iMaxPos - iPos;
            }

            if ( vertical == TRUE) {
                x = 0;
                y = iPos;
                /* cx is already set. */
                cy = iSize;
            } else {
                x = iPos;
                y = 0;
                cx = iSize;
                /* cy is already set. */
            }

            fillImage( images[ 1], x, y, cx, cy,
                       &colors[ ( increaseItems) ?
                                    colorIndex.length - 1 - ii : ii]);

            iPos += iSize; 
        }
    }

/* Draw the min- or max-cutoff item depending on whether the legend
   is increasing or decreasing. */

    if ( numItems > 1 && colorIndex.length > 1) {

        cx = display->ScreenCoord( itemRects[ numItems - 1].width);
        cx = display->DimInPixels( cx);

        cy = display->ScreenCoord( itemRects[ numItems - 1].height);
        cy = display->DimInPixels( cy); 

        fillImage( images[ numItems - 1], 0, 0, cx, cy,
                   &colors[ ( increaseItems) ? 0 : colorIndex.length - 1]);
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an AmImageLegendClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmImageLegendClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmImageLegendClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				       int		*numDefs)
{
    static AmFieldDefRec ImageLegendClassFieldDefinitions[] = {
	{ AmNoutline, offsetof( AmImageLegendClass, outline) },
    };

    *fieldDefs = ImageLegendClassFieldDefinitions;
    *numDefs = VECLEN( ImageLegendClassFieldDefinitions);
}
