#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/*
 * File: legend.cpp
 * 
 * Purpose:
 *   This file contains the legend class member functions.  The legend class
 *   object is the base object for all the other legends, such as the image
 *   legend, the line legend, the polygon legend, the scatter legend, and so
 *   on.
 *
 *  The legend class provides a basic set of functionality to the 
 *  derived legends.  It contains functions to create the layout of the
 *  legends (layoutRects), to generate the strings for the legends
 *  (AmIndexesToStrings), to draw the legends (Draw), to update legends
 *  (Update), and so on.
 *
 *  The derived legends need to override the DrawItem function and possibly
 *  the Update function.
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/mem_funcs.h>
#include <anvil/resource.h>
#include <anvil/defaults.h>
#include <anvil/plot.h>
#include <anvil/legend.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>

/****************************************************************************
 *
 * Function: GetLabelExtents( pDisplay, pGC, szText,
 *                            pcx, pcy, piAscent, piDescent)
 *                         
 * Purpose:
 *   Retrieves the extents of a a multi line text string.
 *
 * Parameters:
 *   AmDisplayClass *pDisplay  - The display.
 *   AmGCClass      *pGC       - The GC specifying the font of the text.
 *   char         *szText    - The text to query. 
 *   int          *pcx       - The return value for the width of the text
 *   int          *pcy       - The return value for the height of the text.
 *   int          *piAscent  - The return value for the ascent. 
 *   int          *piDescent - The return value for the descent. 
 *
 * Return Value:
 *   -none-
 *
 * Comments:
 *   GetLabelExtents loops through all the lines of text in the label
 *   and the extents returned are the sum of the text's heights, the
 *   width of the widest line, and the font's descent and ascent.
 *
 ***************************************************************************/

static void 
GetLabelExtents( AmDisplayClass *pDisplay,
                 AmGCClass      *pGC,
                 char         *szText,
                 int          *pcx,
                 int          *piAscent,
                 int          *piDescent,
                 int          *pcy)
{
    int     cx, cy, cLines;
    char    *pcStart, *pcEnd;
    AnvilBoolean fLast;

/* Initializations. */

    *pcx = 0;
    *pcy = 0;
    cLines = 0;
    pcEnd = szText;
    fLast = FALSE;

    cy = pDisplay->FontHeight( pGC, piAscent, piDescent);

    while ( pcEnd && ( *pcEnd != '\0')) {

    /* Find the end of the current line of text. */

        pcStart = pcEnd;
        while (( *pcEnd != '\0') && ( *pcEnd != '\n')) {
            pcEnd++;
        }

    /* Temporarily NULL terminate the string (if needed). */

        if ( *pcEnd == '\0') {
            fLast = TRUE;
        } else {
            *pcEnd = '\0'; 
        }

    /* Retrieve the width of this line of text. */

        cx = pDisplay->TextWidth( pGC, pcStart, pcEnd - pcStart);

    /* Record the maximum width. */

        if ( cx > *pcx) { 
            *pcx = cx;
        }

   /* Set the AmLast character of the string to what it was before and 
      go on to the next line (if this wasn't already the AmLast line). */

        if ( fLast == FALSE) {
            *pcEnd = '\n';
            pcEnd += 1;
        }

        cLines++;
    }

/* Return the total height of the label. */ 

    *pcy = cLines * cy;
}


/****************************************************************************
 *
 * Function: GetMaxLabelExtents( pDisplay, pGC, vszText, pcx, pcy)
 *                         
 * Purpose:
 *   Retrieves the extents of the labels and returns the height of
 *   the "tallest" label and the width of the "widest" label.
 *
 * Parameters:
 *   AmDisplayClass *pDisplay  - The display.
 *   AmGCClass      *pGC       - The GC specifying the font of the text.
 *   char         **vszText  - The vector of labels.
 *   int          *pcx       - The return value for the width of the text
 *   int          *pcy       - The return value for the height of the text.
 *
 * Return Value:
 *   -none-
 *
 * Comments:
 *   GetMaxLabelExtents loops through all the labels of the vector of
 *   labels vszText.  It records the label with the maximum extents and
 *   returns the width throgh the pcx argument and the height through 
 *   the pcy argument.
 *
 ***************************************************************************/

static void
GetMaxLabelExtents( AmDisplayClass *pDisplay,
                    AmGCClass      *pGC,
                    char         **vszText,
                    int          *pcx,
                    int          *pcy)
{
    int ii, cx, iAscent, iDescent, cy;

/* Initializations. */

    *pcx = 0;
    *pcy = 0;

    if ( vszText != NULL) {
	for ( ii = 0; vszText[ ii] != NULL; ii++) {
            GetLabelExtents( pDisplay, pGC, vszText[ ii],
                             &cx, &iAscent, &iDescent, &cy);

	    if ( cx > *pcx) {
		*pcx = cx;
	    }
	    if ( cy > *pcy) {
                *pcy = cy;
            }
	}
    }
}

/****************************************************************************
 *
 * Function: DrawTextRect( pDisplay, pWindow, pGC, szText, pRect,
 *                         fVertical, eAlignH, eAlignV)
 *                         
 * Purpose:
 *   Draws aligned multi line text within a bounding rectangle.
 *
 * Parameters:
 *   AmDisplayClass *pDisplay - A pointer to a display that is associated with
 *                            the passed in window.
 *   AmWindowClass  *pWindow  - A pointer to a window that the text is to be
 *                            drawn on.
 *   AmGCClass      *pGC      - A pointer to a GC specifying the font and color
 *                            of the text.
 *   AmRectangleRec *pRect    - A pointer to a rectangle specifying the bounding
 *                            area for the text.
 *   AnvilBoolean      fVertical - Indicates whether the text is to be drawn
 *                            vertically (TRUE), or horizontally (FALSE).
 *   AmAlignEnum    eAlignH   - Horizontal alignment.
 *   AmAlignEnum    eAlignV   - Vertical alignment.
 *
 * Return Value:
 *   -none-
 *
 * Comments:
 *     Draws the multi line text given by the szText argument into the area
 *     given by the rect argument.  The text is drawn horizontal if fVertical
 *     is FALSE and vertical otherwise.  The text is aligned to the rectangle
 *     horizontally according to the eAlignH argument which can be either
 *     AmALIGN_LEFT, AmALIGN_CENTER, or AmALIGN_RIGHT, and vertically
 *     according to the eAlignV argument which can be either AmALIGN_TOP,
 *     AmALIGN_CENTER, or AmALIGN_BOTTOM.
 *
 *     Note that these alignments actually refer to the text as if it was
 *     horizontal text.  So if fVertical is TRUE the eAlignV actually gives
 *     the horizontal alignment and eAlignH gives the vertical alignment.
 *
 ***************************************************************************/

static void
DrawTextRect( AmDisplayClass *pDisplay,
              AmWindowClass  *pWindow,
              AmGCClass      *pGC,
              char         *szText,
              AmRectangleRec *pRect,
              AnvilBoolean      fVertical,
              AmAlignEnum    eAlignH,
              AmAlignEnum    eAlignV)
{
    int     cxRect, cyRect, cxLabel, cyLabel, iAscent, iDescent,
            iLabel, cyText, yAdjust, xAdjust, iTemp;
    char    *pcStart, *pcEnd;
    AnvilBoolean fLast;

    cxRect = pRect->width;
    cyRect = pRect->height;

/* Flip the width and height arguments if the text is to be
   displayed vertically. */

    if ( fVertical) {
	iTemp = cxRect;
	cxRect = cyRect;
	cyRect = iTemp;
    }

/* Calculate the extents of the multi line text. */

    GetLabelExtents( pDisplay, pGC, szText, &cxLabel,
                     &iAscent, &iDescent, &cyLabel);

/* Calculate the adjustment for the vertical alignment. */

    switch( eAlignV) {
        case AmALIGN_TOP:
            yAdjust = 0;
            break;
        default:
        case AmALIGN_CENTER:
            yAdjust = ( cyRect - cyLabel) / 2;
            break;
        case AmALIGN_BOTTOM:
            yAdjust = cyRect - cyLabel;
            break;
    }

    cyText = iAscent + iDescent;

    iLabel = 0;
    fLast = FALSE;
    pcEnd = szText;

/* Step through the text and align each line of the label according to
   the horizontal alignment. */

    while ( !fLast) {
        pcStart = pcEnd;

    /* Find the AmLast character of this line. */

        while ( ( *pcEnd != '\0') && ( *pcEnd != '\n')) {
            pcEnd++;
        }

        if ( *pcEnd == '\0') {
            fLast = TRUE;
        }

    /* Null terminate the string. */

        *pcEnd = '\0';

    /* Gets the extents of this line. */

        cxLabel = pDisplay->TextWidth( pGC, pcStart, strlen( pcStart));

    /* Calculate the horizontal adjustment for the text. */

        switch ( eAlignH) {
            case AmALIGN_LEFT:
                xAdjust = 0;
                break;
            default:
            case AmALIGN_CENTER:
                xAdjust = ( cxRect - cxLabel) / 2;
                break;
            case AmALIGN_RIGHT:
                xAdjust = cxRect - cxLabel;
                break;
        }

    /* Draw the text - either vertically or horizontally. */

        if ( szText != NULL && strcmp( szText, "") != 0) {
            if ( fVertical == TRUE) {

            /* Note that the yAdjustment is added to the x coordinate and
               the xAdjustment is taken into account for the y coordinate.
               This is because we calculated all the adjustments for the
               alignments as if the text was drawn horizontally and now
               we are drawing the text vertically. */

	        pDisplay->DrawRotatedText( pWindow, pGC,
                                           pRect->x + yAdjust +
                                           iAscent + cyText * iLabel,
                                           pRect->y + pRect->height - xAdjust,
                                           M_PI/2.0, pcStart);
            } else {
                pDisplay->DrawText( pWindow, pGC,
                                    pRect->x + xAdjust,
				    pRect->y + yAdjust + iAscent +
				    cyText * iLabel,
				    pcStart);
	    }
        }

        iLabel += 1;

    /* Reset the '\0' character to what it was before we drew this line
       of the label. */

        if ( fLast == FALSE) {
            *pcEnd = '\n';
            pcEnd++;
        }
    }
}

/*************************************************************************
 *
 * Description:
 *	recursive run-time typing function used for all screen objects
 * Arguments:
 *	type - requested type
 * Returns:
 *	TRUE if object or its ancestor class is of type "type", FALSE 
 *	otherwise
 *
 *************************************************************************/

AnvilBoolean
AmLegendClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmLEGEND) {
        return( TRUE);
    } else {
        return( AmScaleClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so legends can be subclassed.
 *
 ****************************************************************************/

AmLegendClass::AmLegendClass()
{
    itemRects  = NULL;
    labels     = NULL; 
    labelRects = NULL;
}

/****************************************************************************
 *
 * Description:
 *      Legend class destructor: free the color index data,
 *      the vector of item rectangles, label rectangles, and the vector
 *	of labels
 *
 ****************************************************************************/


AmLegendClass::~AmLegendClass()
{
    if ( IsValid()) {
        AmDestroyIndexData( &colorIndex);

        if ( itemRects) {
            free( itemRects);
        }

        if ( labels) {
            AmFreeStrVec( labels);
        }

        if ( labelRects) {
            free( labelRects);
        }
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a legend.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *	graphicArg - parent data graphic that legend is based on.
 *	(May be NULL for standalone legends)
 *      szName - string name used to identify legend for resources
 *
 ****************************************************************************/

void 
AmLegendClass::initializeFields( AmWindowClass        *windowArg,
			       AmDataGraphicClass   *graphicArg,
			       const char	  *nameArg)
{
    AmScaleClass::initializeFields( windowArg, graphicArg, nameArg);

    legendType = AmLEGEND_DISTINCT;
    numRows    = 10;
    outline    = TRUE;

/* Get the default values from the resource daetabase. */

    rect.height    = AmGetDefault( "legend.height", AmLEGEND_HEIGHT);
    rect.width     = AmGetDefault( "legend.width", AmLEGEND_WIDTH);
    vertical       = AmGetDefault( "legend.vertical", TRUE);
    increaseItems  = AmGetDefault( "legend.increaseItems", TRUE);
    editable       = AmGetDefault( "editable", FALSE);
    numItems       = AmGetDefault( "legend.numItems", AmLEGEND_NUM_ITEMS);
    labelsPerItem  = AmGetDefault( "legend.labelsPerItem",
                                 AmLEGEND_LABELS_PER_ITEM);
    mainItemHeight = AmGetDefault( "legend.mainItemHeight",
                                 AmLEGEND_MAIN_ITEM_HEIGHT);
    itemWidth      = AmGetDefault( "legend.itemWidth", AmLEGEND_ITEM_WIDTH);
    itemHeight     = AmGetDefault( "legend.itemHeight", AmLEGEND_ITEM_HEIGHT);
    itemSpacing    = AmGetDefault( "legend.itemSpacing", AmLEGEND_ITEM_SPACING);

    colors = NULL;
    labels = NULL;

    AmInitIndexData( &colorIndex);
    minClip = colorIndex.minClip;
    maxClip = colorIndex.maxClip;

    numItems   = 0;
    numLabels  = 0;
    itemRects  = NULL;
    labelRects = NULL;

    _dXScale   = 1.0;
    _dYScale   = 1.0;

    _iOriginalWidth  = rect.width;
    _iOriginalHeight = rect.height;
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a legend, given the tag-values stored in
 *      a tag values array.  First, the scale fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int
AmLegendClass::getFields( AmTagValuePtr tagValues,
			int         numValues)
{
    int         ii, numFieldDefs, numUnusedValues;
    AmFieldDefPtr fieldDefs;

    numValues = AmScaleClass::getFields( tagValues, numValues);
    AmLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNcolorsBinvec) ||
             !strcmp( tagValues[ ii].tag, AmNbinvec)) {
            if ( colorIndex.binvec != NULL) {
                *( void **) tagValues[ ii].value =
                    AmCopyVec( colorIndex.binvec,
                        ( colorIndex.length - 1) * sizeof( float));
            } else {
                *( void **) tagValues[ ii].value = NULL;
            }
        } else {
            tagValues[ numUnusedValues++] = tagValues[ ii];
        }
    }

    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a legend, given the tag-values stored in
 *      a tag values array.  First, the scale fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/


int
AmLegendClass::setFields( AmTagValuePtr tagValues,
                        int         numValues)
{
    int         ii, numFieldDefs, numUnusedValues;
    AmFieldDefPtr fieldDefs;

    numValues = AmScaleClass::setFields( tagValues, numValues);
    AmLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    colorIndex.length = numColors;
    numUnusedValues = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNbinvec) ||
             !strcmp( tagValues[ ii].tag, AmNcolorsBinvec)) {
            AmSetBinvec( &colorIndex.binvec, numColors, &tagValues[ ii]);
        } else {
            tagValues[ numUnusedValues++] = tagValues[ ii];
        }
    }
    return( numUnusedValues);
}

/****************************************************************************
 *
 * Description:
 *	Update member function for legends, calls the layoutRects()
 *	function.
 * Arguments:
 *	eventCodes - vector of events which cause the update.
 *	numEventCodes - number of events.
 *
 ****************************************************************************/

/* ARGSUSED */
AmErrorCode
AmLegendClass::Update( int *eventCodes,
                     int numEventCodes)
{
    layoutRects();

    extentsRect = rect; /* Reset the extents rectangle. */
    return( AmOK);
}


/* I WANT TO BE COMMENTED .... !!! */

char **
AmIndexesToStrings( AmIndexDataPtr   index,
                  int            num,
                  AmLegendTypeEnum legendType)
{
    char	**strings, *unitStr,
                buffer[ TEXTLEN], valbuf[ TEXTLEN], valbuf2[ TEXTLEN];
    float	lo=0, hi, min, max;
    int		ii, aa, bb;

    strings = ( char **) malloc( sizeof( char *)*( num + 1));

/* If there was a lookup table walk through it. */

    if ( index->table != NULL) {
	if ( index->length == num) {
	    for ( ii = 0; ii < index->length; ii++) {
                AmPrintValue( index->units, ( float) index->table[ ii], valbuf);
		sprintf( buffer, "%d %s", index->table[ ii], valbuf);
		strings[ ii] = strdup( buffer);
	    }
	} else {

        /* This seems very broken to me - Phil. */

            AmGetUnitString( index->units, &unitStr);
	    for ( ii = 0; ii < num; ii++) {
		aa = ii * ( index->length) / num;
		bb = ( ii + 1) * ( index->length) / num - 1;
		if ( lo != bb) {
		    sprintf( buffer, "%d - %d %s", 
			     index->table[ aa], index->table[ bb], unitStr);
		} else {
		    sprintf( buffer, "%d %s", index->table[ aa], unitStr);
		}
	    }
	}
    } else if (( index->binvec != NULL) && ( legendType == AmLEGEND_DISTINCT)) {
	for ( ii = 0; ii < num; ii++) {
	    if ( ii == 0) {
                AmPrintValue( index->units, index->binvec[ 0], valbuf);
		sprintf( buffer, "< %s", valbuf);
	    } else if ( ii == num - 1) {
                AmPrintValue( index->units, index->binvec[ index->length - 2],
                    valbuf);
		sprintf( buffer, "> %s",  valbuf);
	    } else {
                AmPrintValueNoUnits( index->units,
                    index->binvec[ ( ii - 1) * index->length / num], valbuf);
                AmPrintValue( index->units,
                    index->binvec[ ii * index->length / num], valbuf2);
		sprintf( buffer, "%s - %s", valbuf, valbuf2); 
	    }
	    strings[ ii] = strdup( buffer);
	}
    } else {

        if ( index->binvec != NULL) {
            min = index->binvec[ 0];
            max = index->binvec[ index->length - 2];
        } else {
            min = index->min;
            max = index->max;
        } 

/* it's a linear scaling table.  print either the min value, max value, or 
   the bounding values */
	    
	for ( ii = 0; ii < num; ii++) {
	    if ( ii == 0) {
                AmPrintValue( index->units, min, valbuf);
		sprintf( buffer, "< %s", valbuf);
	    } else if ( ii == num - 1) {
                AmPrintValue( index->units, max, valbuf);
		sprintf( buffer, "> %s", valbuf);
	    } else if (( legendType == AmLEGEND_DISTINCT) || ( num == 3)) {
		lo = min +
                         (( max - min)*( ii - 1)) / ( num - 2);
		hi = min + (( max - min)*ii) / ( num - 2);
                AmPrintValueNoUnits( index->units, lo, valbuf);
                AmPrintValue( index->units, hi, valbuf2);
		sprintf( buffer, "%s - %s", valbuf, valbuf2); 
	    } else {
		lo = min + (( max - min)*( ii - 1)) / ( num - 3);
                AmPrintValue( index->units, lo, valbuf);
		sprintf( buffer, "%s", valbuf);
            }
	    strings[ ii] = strdup( buffer);
	}
    }

/* NULL terminate the strings for AmFreeStrVec() */

    strings[ num] = NULL;
    return( strings);
}


/****************************************************************************
 *
 * Function: AmLayoutTitleRect( fShow, pDisplay, pGC, szText,
 *                            fVertical, iOffset, pRect, px, py)
 *                         
 * Purpose:
 *   Creates the rectangle for the title it takes into account the
 *   extents of the title, the orientation, and the origin.
 *
 * Parameters:
 *   AnvilBoolean      fShow     - Whether or not there is a title.
 *   AmDisplayClass *pDisplay - The display.
 *   AmGCClass      *pGC      - The GC specifying the font of the text.
 *   char         *vszText  - The title label.
 *   AnvilBoolean      fVertical - The orientation of the title, either
 *                            horizontal (FALSE) or vertical (TRUE). 
 *   int          iOffset   - The seperation AmBetween the title and
 *                            the follwing item. 
 *   AmRectangleRec *pRect    - The return value for the extents rectangle
 *   int          *px       - Both the current x coordinate (where the
 *                            title will be placed) and the updated x
 *                            coordinate (where the next itme is supposed
 *                            to be placed) adjusted for the title.
 *   int          *py       - Both the current y coordinate (where the
 *                            title will be placed) and the updated y
 *                            coordinate (where the next itme is supposed
 *                            to be placed) adjusted for the title.
 *
 * Return Value:
 *   AmErrorCode indicating success or failure.  At this point the
 *   only possible value that is returned is AmOK.
 *
 * Comments:
 *   This function is used from within LayoutRects to create the
 *   title layout rectangle.  The rectangle of the title is simply 
 *   the rectangle of the text extents of the, possibly multi line,
 *   label.  The rectangle is then adjusted for whether or not
 *   the title is horizontal or vertical.  The origin of the title
 *   rectangle is set according to the two passed in arguments px
 *   and py.  These two arguments are also updated so that they
 *   contain the location where the rest of the legend is to be
 *   placed.  This new location also depends on the iOffset argument 
 *   which gives the seperation AmBetween the title and the legend.
 *
 ***************************************************************************/

AmErrorCode
AmLayoutTitleRect( AnvilBoolean      fShow,
                 AmDisplayClass *pDisplay,
                 AmGCClass      *pGC,
                 char         *szText,
                 AnvilBoolean      fVertical,
                 int          iOffset,
                 AmRectanglePtr pRect,
                 int          *px,
                 int          *py)
                 
{
    int cx, cy, iAscent, iDescent;

/* Layout the title (center the whole thing later. */

    if (szText && ( strcmp( szText, "") != 0) && ( fShow == TRUE)) { 

    /* Get the extents of the title lable. */

        GetLabelExtents( pDisplay, pGC, szText,
                         &cx, &iAscent, &iDescent, &cy);

    /* Create the rectangle depending on the orientation of the title. */

        if ( fVertical) {

            pRect->x = *px;
            pRect->y = *py;
            pRect->width = cy;
            pRect->height = cx;

        /* Update the current x position. */

            *px += pRect->width + iOffset;
        } else {

            pRect->x = *px;
            pRect->y = *py; 
            pRect->width = cx;
            pRect->height = cy;

        /* Update the current y position. */

            *py += pRect->height + iOffset;
        }
    } else {

    /* Make sure that the rectangle is made up of values that indicate
       it does not exist. */

        pRect->x = *px;
        pRect->y = *py;
        pRect->width = 0;
        pRect->height = 0;
    }

    return( AmOK);    
}


/****************************************************************************
 *
 * Function: CenterAreas( x, y, cx, cy)
 *                         
 * Purpose:
 *   Centers the title and item rectangle in respect to each other.
 *
 * Parameters:
 *   int          x  - The item's x origin.
 *   int          y  - The item's y origin.
 *   int          cx - The item's widht.
 *   int          cy - The item's height.
 *
 * Return Value:
 *   AmErrorCode indicating success or failure.  At this point the
 *   only possible value that is returned is AmOK.
 *
 * Comments:
 *   Depending on which area is larger, either the title gets centered
 *   along the dimension of the "item" rectangle, or the "item" rectangle
 *   gets centered along the dimension of the title.  Also if the title
 *   is vertical, the centering is done vertically, otherwise the
 *   centering is done horizontally.
 *
 *   The overall extends of the title and item rectangle combined is
 *   aligned with respect to the legend's rectangle.
 *
 ***************************************************************************/

AmErrorCode
AmLegendClass::CenterAreas( int          x,
                          int          y,
                          int          cx,
                          int          cy)
{
    AnvilBoolean fUpdateRect;
    int     ii, xOffsetTitle, yOffsetTitle, xOffsetItem, yOffsetItem,
            cxTotal, cyTotal, cxOffset, cyOffset;

/* Calculate the offet of the title rectangle and the item rectangles
   respectively. */ 

    if ( titleVertical == TRUE) {
        if ( cy > titleRect.height) {
            xOffsetTitle = 0; 
            yOffsetTitle = ( cy - titleRect.height) / 2;
            xOffsetItem  = 0;
            yOffsetItem  = 0;

        /* Calculate the total dimensions. */

            cxTotal = x + cx;
            cyTotal = cy;
        } else {
            xOffsetTitle = 0;
            yOffsetTitle = 0;
            xOffsetItem  = 0;
            yOffsetItem  = ( titleRect.height - cy) / 2;

        /* Calculate the total dimensions. */

            cxTotal = x + cx;
            cyTotal = titleRect.height;
        }
    } else {
        if ( cx > titleRect.width) {
            xOffsetTitle = ( cx - titleRect.width) / 2;
            yOffsetTitle = 0; 
            xOffsetItem  = 0; 
            yOffsetItem  = 0; 

        /* Calculate the total dimensions. */

            cxTotal = cx;
            cyTotal = y + cy;
        } else {
            xOffsetTitle = 0;
            yOffsetTitle = 0;
            xOffsetItem  = ( titleRect.width - cx) / 2;
            yOffsetItem  = 0;

        /* Calculate the total dimensions. */

            cxTotal = titleRect.width;
            cyTotal = y + cy;
        }
    }


/* Compare against the current extents. */

    fUpdateRect = FALSE;

/* Save extents of the title and item rectangle combined (this
   is the minimum size of the legend). */

    _rectInterior.x      = 0;
    _rectInterior.y      = 0;
    _rectInterior.width  = cxTotal;
    _rectInterior.height = cyTotal;

    if ( rect.width < _rectInterior.width) {
       fUpdateRect = TRUE;
       rect.width = _rectInterior.width;
    }

    if ( rect.height < _rectInterior.height) {
       fUpdateRect = TRUE;
       rect.height = _rectInterior.height;
    }

/* Adjust the title rectangle. */

    cxOffset = rect.x + ( rect.width - cxTotal) / 2;
    cyOffset = rect.y + ( rect.height - cyTotal) / 2;

    titleRect.x += xOffsetTitle + cxOffset;
    titleRect.y += yOffsetTitle + cyOffset;

/* Adjust the legend items's rectangles. */

    for ( ii = 0; ii < numItems; ii++) {
        itemRects[ ii].x += xOffsetItem + cxOffset;
        itemRects[ ii].y += yOffsetItem + cyOffset;
    }

/* Adjust the legend labels's rectangles. */

    for ( ii = 0; ii < numLabels; ii++) {
        labelRects[ ii].x += xOffsetItem + cxOffset;
        labelRects[ ii].y += yOffsetItem + cyOffset;
    }

    if ( fUpdateRect == TRUE) {
        extentsRect = rect;
        AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
    }

    return( AmOK);
}


/****************************************************************************
 *
 * Function: layoutRects()
 *                         
 * Purpose:
 *   Lays out the three areas associated with legends, the title area,
 *   the symbol/color area, and the label/text area.
 *
 * Parameters:
 *   -none-
 *
 * Return Value:
 *   AmErrorCode indicating success or failure.
 *
 * Comments:
 *     The layoutRects function calculates and positions the three areas
 *     associated with legends:  The title area, the symbol/color area
 *     and the text area.  The areas are sized so that everything fits.
 *     Depending on the orientation of the title, it is either on top or
 *     to the left of the symbol area (horizontal or vertical).  Depending
 *     on the orientation of the legend the labels are to the right or on
 *     top of the symbol area (vertical or horizontal).  Also note that the
 *     symbol/color area and the text area consist of more than one layout
 *     rectangle (one rectangle each for each row).
 *
 *     The overall layout is centered in the extends rectangle.  It is
 *     possible that the layout rectangles will cover a larger area than
 *     the extends rectangle at which point some parts of the legend will
 *     be clipped.
 *
 ***************************************************************************/

AmErrorCode
AmLegendClass::layoutRects()
{
    int x, y, yMax, yAdjust, cxItems, cyItems, cxText, cyText, cyMainText,
        ii, iLabel, xStart, yStart, temp;
 
    int iMainItemHeight, iItemHeight, iItemWidth, iItemSpacing;

/* Calculate the new heights. */

    iMainItemHeight = (int)(mainItemHeight * _dYScale);
    iItemWidth      = (int)(itemWidth * _dXScale);
    iItemHeight     = (int)(itemHeight * _dYScale);
    iItemSpacing    = itemSpacing;

/* Initializations. */

    xStart  = 0;
    yStart  = 0;
    yMax    = 0;
    iLabel  = 0;
    yAdjust = 0;

/* Free memory used by previous layouts. */

    if ( itemRects != NULL) {
        free( itemRects);
        itemRects = NULL;
    }
  
    if ( labelRects != NULL) {
        free( labelRects);
        labelRects = NULL;
    }
 
/* Allocate memory for the new layout. */
 
    itemRects = ( AmRectanglePtr) malloc( sizeof( AmRectangleRec) * numItems);
    if ( itemRects == NULL) {
        RETURN_ERROR( AmMALLOC_FAILED);
    }

    labelRects = ( AmRectanglePtr) malloc( sizeof( AmRectangleRec) * numLabels);
    if ( labelRects == NULL) {
        RETURN_ERROR( AmMALLOC_FAILED);
    }

/* Setup the title rectangle. */

    AmLayoutTitleRect( showTitle, display, titleGC, title, titleVertical,
                       titleOffset, &titleRect, &xStart, &yStart);

    x = xStart;
    y = yStart;

/* Figure out the maximum extents of the text items. */

    GetMaxLabelExtents( display, textGC, labels, &cxText, &cyText);

/* If the text item is bigger than the legend item then we need to
   possibly start the legend item further down and centered on the text.
   Calculate that adjustment for that amount here. */

    if ( cyText > iItemHeight) {
        yAdjust = ( cyText - iItemHeight) / 2;
    }

/* Layout the legend's rectangles depending on the legend type. */

    if ( legendType == AmLEGEND_CONTINUOUS) {

    /* A continuous legend consists of at most three item rectangles.
       The min clipping, max clipping, and main item rectangles. */

        y += yAdjust;

    /* Define the min (or max) clipping rectangle AmFirst. */

        itemRects[ 0].x = x;
        itemRects[ 0].y = y;
        itemRects[ 0].width = iItemWidth;
        itemRects[ 0].height = iItemHeight;

    /* Define the associated label rectangle.  It is to the right
       and vertically centered. */

        labelRects[ iLabel].x = x + iItemWidth + textOffset;
        labelRects[ iLabel].y = y + ( iItemHeight - cyText)/2;
        labelRects[ iLabel].width = cxText;
        labelRects[ iLabel].height = cyText;

    /* If min clipping is off and this legend is in decreasing order 
       (which means the minimum value is at the top and the maximum
       value is at the bottom), or if max clipping is off and the legend
       is in increasing order (the reverse) then we actually use
       the rectangle defined above; however, otherwise, the rectangle
       is only defined so that later code doesn't break, but we 
       don't take it into account for the placement of the next
       item. */

        if ( (( minClip == FALSE) && ( increaseItems == FALSE)) ||
              ( maxClip == FALSE) && ( increaseItems == TRUE)) {
            y += iItemHeight;
        } else {

        /* We need to go back to the very beginning. */

            y -= yAdjust;
        }

    /* Go to the next label. */

        iLabel += 1;

        if ( labelsPerItem >= 2) {
  
        /* There are at least 2 labels. */

        /* Figure out the total (minimum) extents of the labels. */

            cyMainText = labelsPerItem * cyText;
            if (( cyMainText - cyText) > iMainItemHeight) {
                iMainItemHeight = cyMainText - cyText;
            } else {
                cyMainText = iMainItemHeight + cyText;
            }

        /* If the spacing AmBetween the min cutoff and the
           main item label is too small for the text, increase
           the spacing. */

            if ( yAdjust + cyText / 2 > iItemSpacing) {
                iItemSpacing = yAdjust + cyText / 2;
            }
 
            y += iItemSpacing;

        /* Define the label rectangles evenly spaced along the
           main item. */

            for ( ii = 0; ii < labelsPerItem; ii++) {
                labelRects[ iLabel].x = x + iItemWidth + textOffset;
                labelRects[ iLabel].y = y - cyText / 2 +
                    ii * iMainItemHeight / ( labelsPerItem - 1);
                labelRects[ iLabel].width = cxText;
                labelRects[ iLabel].height = cyText;

                iLabel += 1;
            }

        } else if ( labelsPerItem == 1) {

        /* There is exactly 1 label. */

            if ( yAdjust + ( cyText - iMainItemHeight) / 2 > iItemSpacing) {
                iItemSpacing = yAdjust + ( cyText - iMainItemHeight) / 2;
            }

            y += iItemSpacing;

            labelRects[ iLabel].x = x + iItemWidth + textOffset;
            labelRects[ iLabel].y = y + ( iMainItemHeight - cyText) / 2;
            labelRects[ iLabel].width = cxText;
            labelRects[ iLabel].height = cyText;

            iLabel += 1;
        } else {

        /* There are no labels. */

            if ( yAdjust > iItemSpacing) {
                iItemSpacing = yAdjust;
            }

            y += iItemSpacing;
        }

        itemRects[ 1].x = x;
        itemRects[ 1].y = y;
        itemRects[ 1].width = iItemWidth;
        itemRects[ 1].height = iMainItemHeight;

        y += iMainItemHeight;

    /* If the third legend item is to be drawn, add in the spacing. */ 

        if ( (( minClip == FALSE) && ( increaseItems == TRUE)) ||
              ( maxClip == FALSE) && ( increaseItems == FALSE)) {
            y += iItemSpacing;
        }

        itemRects[ 2].x = x;
        itemRects[ 2].y = y;
        itemRects[ 2].width = iItemWidth;
        itemRects[ 2].height = iItemHeight;

        labelRects[ iLabel].x = x + iItemWidth + textOffset;
        labelRects[ iLabel].y = y + ( iItemHeight - cyText) / 2;
        labelRects[ iLabel].width = cxText;
        labelRects[ iLabel].height = cyText;

        if ( (( minClip == FALSE) && ( increaseItems == TRUE)) ||
              ( maxClip == FALSE) && ( increaseItems == FALSE)) {
            y += iItemHeight + yAdjust;
        }

        cxItems = iItemWidth + textOffset + cxText; 
        cyItems = y - yStart;
    } else {

        if ( cyText > iItemHeight + iItemSpacing) {
            iItemSpacing = cyText - iItemHeight;
        }

        for ( ii = 0; ii < numItems; ii++) {
            itemRects[ ii].x = x;
            itemRects[ ii].y = y;
            itemRects[ ii].width = iItemWidth;
            itemRects[ ii].height = iItemHeight;

            labelRects[ ii].x = x+ iItemWidth + textOffset;
            labelRects[ ii].y = y+ ( iItemHeight - cyText) / 2;
            labelRects[ ii].width = cxText;
            labelRects[ ii].height = cyText;

            y += iItemHeight;
            if ( y > yMax) {
                yMax = y;
            }

            if ( numRows && !(( ii + 1) % numRows) && ( ii < ( numItems - 1))) {
                y = yStart;
                y += yAdjust;
                x += iItemWidth + textOffset + cxText + textOffset;
            } else {
                y += iItemSpacing;
            }
        }

        cxItems = x + iItemWidth + textOffset + cxText - xStart;
        cyItems = yMax - yStart + yAdjust;
    }

    if ( vertical == FALSE) {
        for ( ii = 0; ii < numItems; ii++) {
            temp = itemRects[ ii].x;
            itemRects[ ii].x = xStart + itemRects[ ii].y - yStart;
            itemRects[ ii].y = yStart + cxItems - ( temp - xStart) -
                               itemRects[ ii].width;

            temp = itemRects[ ii].width;
            itemRects[ ii].width = itemRects[ ii].height;
            itemRects[ ii].height = temp;
        }

        for ( ii = 0; ii < numLabels; ii++) {
            temp = labelRects[ ii].x;
            labelRects[ ii].x = xStart + labelRects[ ii].y - yStart;
            labelRects[ ii].y = yStart + cxItems - ( temp - xStart) -
                                labelRects[ ii].width;

            temp = labelRects[ ii].width;
            labelRects[ ii].width = labelRects[ ii].height;
            labelRects[ ii].height = temp;
        }

        temp = cxItems;
        cxItems = cyItems; 
        cyItems = temp;
    }

    CenterAreas( xStart, yStart, cxItems, cyItems);

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Draw a single legend item (i.e. color, symbol ... etc ... ). 
 *	this is a base implementation which is overridden by derived
 *	classes.
 * Arguments:
 * 	w - window to draw the item to
 *	gc - graphics context to draw the item with.
 *	item - index of the item to draw.
 *	itemRect - rectangle defining item's extents.
 *
 ****************************************************************************/

AmErrorCode
AmLegendClass::DrawItem( AmWindowClass        *w,
                       AmGCClass            *gc,
                       int                item,
                       AmRectangleRec const *itemRect)
{
    gc->SetForegroundColor( &colors[ item]);
    display->FillRectangle( w, gc, itemRect->x, itemRect->y,
			    itemRect->width, itemRect->height);

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Draw all the legend items (i.e. color, symbol ... etc ... ). 
 *	this is a base implementation which is overridden by derived
 *	classes.
 * Arguments:
 * 	w - window to draw the item to
 *	gc - graphics context to draw the item with.
 *	itemRects - vector of rectangles defining items' extents.
 *
 ****************************************************************************/

AmErrorCode
AmLegendClass::DrawItems( AmWindowClass       *w,
                        AmGCClass            *gc,
                        AmRectangleRec const *itemRects)
{
    int          ii;

    for ( ii = 0; ii < numItems - 1; ii++) {

        if ( ( ii == 0) && 
             ((( minClip == TRUE) && ( increaseItems == FALSE)) ||
              (( maxClip == TRUE) && ( increaseItems == TRUE)))) {
            continue;
        } else  if ( ( ii == numItems - 1) && 
                     ((( minClip == TRUE) && ( increaseItems == TRUE)) ||
                      (( maxClip == TRUE) && ( increaseItems == FALSE)))) {
            continue;
        }

        DrawItem( w, gc, ii, &itemRects[ ii]);
    }

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw a legend.  This is currently used by all the derived
 *	legend classes.
 * Arguments:
 *      window - window that the legend will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - number of exposed rectangles.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/


AmErrorCode
AmLegendClass::draw( AmWindowClass        *w,
                   AmGCClass            *gc,
                   AmRectangleRec const *exposeRects,
		   int		      nRects)
{
    int 		ii, index, nClipRects;
    AmRectanglePtr	clipRects;
    AmRectangleRec const	*drawRect = GetRect();

/* If the object is not valid just draw the outline. */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

/* Return in the degenerate cases. */

    if ( AmDegenRect( drawRect)) {
        return( AmOK);
    }

/* Get the clip rectangles from the gc passed by the display manager, and
   set them in the legend's private GC's. */

    gc->GetClipRectangles( &clipRects, &nClipRects);
    titleGC->SetClipRectangles( clipRects, nClipRects);
    textGC->SetClipRectangles( clipRects, nClipRects);
    free( clipRects);

/* Clear the rectangle before drawing the legend. We use FillRectangle 
   instead of clear rect, because we have to pay attention to the clip 
   rectangle passed in through the gc from the display manager */

    if ( opaque == TRUE) {
	gc->SetForegroundColor( &bgColor);
	display->FillRectangle( w, gc, drawRect->x, drawRect->y,
			        drawRect->width, drawRect->height);
    }
 
/* Draw the title. */

    if ( showTitle == TRUE && title != NULL && strcmp( title, "") != 0) { 
        titleGC->SetForegroundColor( &fgColor);
        DrawTextRect( display, w, titleGC, title, &titleRect,
                      titleVertical, AmALIGN_CENTER, AmALIGN_CENTER);
    }

/* Draw the legend items. */

    textGC->SetForegroundColor( &fgColor);
    DrawItems( w, textGC, itemRects);

    textGC->SetForegroundColor( &fgColor);

    if ( outline) { 
        for ( ii = 0; ii < numItems; ii++) {

            if ( ( ii == 0) && 
                 ((( minClip == TRUE) && ( increaseItems == FALSE)) ||
                  (( maxClip == TRUE) && ( increaseItems == TRUE)))) {
                continue;
            } else  if ( ( ii == numItems - 1) && 
                 ((( minClip == TRUE) && ( increaseItems == TRUE)) ||
                  (( maxClip == TRUE) && ( increaseItems == FALSE)))) {
                continue;
            }

            display->DrawRectangle( w, textGC, itemRects[ ii].x,
                         itemRects[ ii].y, itemRects[ ii].width,
                         itemRects[ ii].height);
        }
    }

    for ( ii = 0; ii < numLabels; ii++) {

	index = ( increaseItems == TRUE) ? numLabels - ii - 1 : ii;

        if ( ( ii == 0) && ( minClip == TRUE)) {
            continue;
        } else if ( ( ii == numLabels - 1) && ( maxClip == TRUE)) {
            continue;
        }

        if ( labels[ ii] == NULL) {
            break;
        }

        DrawTextRect( display, w, textGC, labels[ ii],
                      &labelRects[ index], ( vertical == FALSE),
                      AmALIGN_LEFT, AmALIGN_CENTER);
    }

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	When the rectangle changes the new layout needs to be determined. 
 * Arguments:
 *	x, y, width, height - dimensions of the legend's new rectangle.
 * Modifications:
 *	there's a problem with the SetRect/setFields calls which cause
 *	the legend size and legend rectangle to get out of sync in the
 *	map editor.
 *
 ***************************************************************************/

void
AmLegendClass::SetRect( int x,
                        int y,
                        int cx,
                        int cy)
{

/* Figure out the scaling. */

/*
    _dXScale = ( float) cx / ( float) _iOriginalWidth;
    _dYScale = ( float) cy / ( float) _iOriginalHeight;
*/

    AmScreenObjClass::SetRect( x, y, cx, cy);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a legend.  This is
 *      used in the map editor to dump the legend's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmLegendClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			 AnvilBoolean	   allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( scale) */

    numStrings = AmScaleClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmLegendClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmLegendClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmLegendClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				       int		*numDefs)
{
    static AmFieldDefRec LegendClassFieldDefinitions[] = {
	{ AmNcolorsMax, offsetof( AmLegendClass, colorIndex.max) },
	{ AmNcolorsMin, offsetof( AmLegendClass, colorIndex.min) },
	{ AmNcolorsStrings, offsetof( AmLegendClass, colorIndex.strings),
                            offsetof( AmLegendClass, colorIndex.length),
                            AmDYNAMIC },
	{ AmNcolorsTable, offsetof( AmLegendClass, colorIndex.table),
                          offsetof( AmLegendClass, colorIndex.length),
                          AmDYNAMIC },
	{ AmNeditable, offsetof( AmLegendClass, editable) },
	{ AmNincreaseItems, offsetof( AmLegendClass, increaseItems) },
	{ AmNitemHeight, offsetof( AmLegendClass, itemHeight) },
	{ AmNitemSpacing, offsetof( AmLegendClass, itemSpacing) },
	{ AmNitemWidth, offsetof( AmLegendClass, itemWidth) },
	{ AmNlabelsPerItem, offsetof( AmLegendClass, labelsPerItem) },
	{ AmNlegendType, offsetof( AmLegendClass, legendType) },
	{ AmNmainItemHeight, offsetof( AmLegendClass, mainItemHeight) },
	{ AmNmaxClip, offsetof( AmLegendClass, colorIndex.maxClip) },
	{ AmNminClip, offsetof( AmLegendClass, colorIndex.minClip) },
	{ AmNnumItems, offsetof( AmLegendClass, numItems) },
	{ AmNnumRows, offsetof( AmLegendClass, numRows) },
	{ AmNoutline, offsetof( AmLegendClass, outline) },
	{ AmNtable, offsetof( AmLegendClass, colorIndex.table),
                    offsetof( AmLegendClass, numColors), AmDYNAMIC },
	{ AmNvertical, offsetof( AmLegendClass, vertical) }
    };
    *fieldDefs = LegendClassFieldDefinitions;
    *numDefs = VECLEN( LegendClassFieldDefinitions);
}
