#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* polyline annotation class member function definitions.
   polyline annotations are generalized line annotations, with
   multiple connected line segments, options for arrows at either
   or both ends, and line width */


#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/lineannot.h>
#include <anvil/rect.h>
#include <anvil/tagstrs.h>

/***************************************************************************
 *
 * Description:
 *      Function for drawing arrow heads at the ends of the line annotation.
 * Arguments:
 *	win - window arrow is drawn to.
 *	display - display arrow is drawn to
 *	gc - graphics context arrow is drawn with.
 *	startP, endP - start and end points.
 * 	w1, i2 - arrow width and height 
 * Returns:
 *      TRUE if object is of requested type or a subclass of it
 * Modifications:
 *	the display parameter should be AmFirst for consistency.
 *
 ***************************************************************************/

static AmPointRec
drawArrow( AmWindowClass  	*win,
           AmDisplayClass 	*display,
           AmGCClass      	*gc,
           AmPointRec        	startP,
           AmPointRec        	endP,
           int          	w1,
           int          	l2)
{
    float 	dx, dy, l1, angle;
    AmPointRec  arrow[ 4];

    dx = endP.x - startP.x;
    dy = endP.y - startP.y;

    l1 = sqrt( dx*dx + dy*dy);

    arrow[ 0] = endP;

    endP.x = ( short)( startP.x + dx * ( l1 - l2)/l1);
    endP.y = ( short)( startP.y + dy * ( l1 - l2)/l1);

    angle = atan2( -1 * dx, dy);

    arrow[1 ].x = ( short)( endP.x + w1 * cos( angle)); 
    arrow[1 ].y = ( short)( endP.y + w1 * sin( angle)); 

    arrow[2 ].x = ( short)( endP.x - w1 * cos( angle)); 
    arrow[2 ].y = ( short)( endP.y - w1 * sin( angle)); 

    display->FillPolygon( win, gc, arrow, 3);

    return( endP);
}

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmLineAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmLINE_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so line annotations can be subclassed.
 *
 ****************************************************************************/

AmLineAnnotClass::AmLineAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Destructor: Free the vector of points that make up the polyline.
 *
 ****************************************************************************/

AmLineAnnotClass::~AmLineAnnotClass()
{
    if ( points != NULL) {
        free( points);
    }
}
/****************************************************************************
 *
 * Description:
 *      initialize the fields for a line annotation.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify object for resources
 *
 ****************************************************************************/

void
AmLineAnnotClass::initializeFields( AmWindowClass  *windowArg,
                                  const char   *nameArg)
{
    AmAnnotationClass::initializeFields( windowArg, nameArg);
    objectType = AmLINE_ANNOTATION;

    opaque  = FALSE;
    _fFrame = FALSE;

    _dLineWidth = 0.0;
    numPoints = 0;
    points = NULL;
    arrows = AmARROW_NONE;
}

/****************************************************************************
 *
 * Description:
 *      AmLineAnnotClass constructor.  Creates a line annotation.
 *      The proprerties for the line annotation are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      winwdowArg - window that the line annotation will be drawn on.
 *      nameArg - string name of the line annotation, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmLineAnnotClass::AmLineAnnotClass( AmWindowClass  *windowArg,
                                const char   *nameArg,
                                ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    numValues = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    setFieldsWrapper( tagValues, numValues);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a line annotation, given the tag-values stored in
 *      a tag values array.  First, the AmAnnotationClass fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  If the line width has been changed, then 
 *	reset the rectangle to make sure the extents rectangle contains
 *	the polyline and its line width.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmLineAnnotClass::setFields( AmTagValuePtr tagValues,
			   int         numValues)
{
    int                numFieldDefs;
    AmFieldDefPtr        fieldDefs;
    const AmRectangleRec *tempRect;

    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmLineAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    if ( AmChanged( &_dLineWidth)) {
        tempRect = GetRect();
        SetRect( tempRect->x, tempRect->y, tempRect->width, tempRect->height);
    }
    return( numValues);
}

/* get the fields which are specific to an image legend */
 
int 
AmLineAnnotClass::getFields( AmTagValuePtr tagValues,
			   int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmLineAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      set the rectangle for the line annotation, and set the extents
 *	rectangle to overlap the line width.
 * Arguments:
 *      x, y, width, height - new rectangle dimensions
 *
 ****************************************************************************/

void 
AmLineAnnotClass::SetRect( int	x,
			 int	y,
			 int	width,
			 int	height)
{
    int margin;

    rect.x = x;
    rect.y = y;
    rect.width = width;
    rect.height = height;

    if ( rect.width == 0) {
        rect.width = 1;
    }

    if ( rect.height == 0) {
        rect.height = 1;
    }

/* set the extents rectangle to reflect the linewidth */

    margin = (int)(( _dLineWidth * 2 + 10) *
                 ( ( float)display->GetDPI()/BASEDPI));

    extentsRect.x = rect.x - margin;
    extentsRect.y = rect.y - margin;
    extentsRect.width = rect.width + 2*margin;
    extentsRect.height = rect.height + 2*margin;

    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}
/****************************************************************************
 *
 * Description:
 *      Draw the line annotation to the screen.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - number of exposed rectangles
 * Returns:
 *      AmOK if everything's OK, AmMALLOC_FAILED if the malloc() failed.
 * Modifications:
 *	The arrow sizes should be settable attributes in the line
 *	annotation, not hardcoded.  The newPoints should be alloca'ed()
 *	instead of malloced.
 *
 ****************************************************************************/

AmErrorCode 
AmLineAnnotClass::draw( AmWindowClass	      	*w,
		      AmGCClass	      		*gc,
		      AmRectangleRec const  	*exposeRects,
		      int		      	nRects)
{
    const AmRectangleRec *drawRect = GetRect();
    int                	 ii, num, w1, l2;
    float                scaleX, scaleY;
    AmPointPtr           newPoints;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( drawRect)) {
        return( AmOK);
    }

/* clear the rectangle before drawing the legend */

    if ( opaque == TRUE) {
	gc->SetForegroundColor( &bgColor);
        display->FillRectangle( w, gc, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }
    gc->SetForegroundColor( &fgColor);
    gc->SetLineWidth( _dLineWidth);

    num = numPoints/2;
    if ( !(newPoints = ( AmPointPtr)malloc( sizeof( AmPointRec) * num))) {
        return( AmMALLOC_FAILED);
    }

    scaleX = ( float)drawRect->width / ( float)originalRect.width;
    scaleY = ( float)drawRect->height / ( float)originalRect.height;

/* Scale the original line points according to the rectangle */

    for ( ii = 0; ii < num; ii++) {
        newPoints[ ii].x = ( short)(( float)drawRect->x +
            ( float)(points[ ii * 2] - originalRect.x) * ( float)scaleX +
                0.5);
        newPoints[ ii].y = ( short)(( float)drawRect->y +
            ( float)(points[ ii * 2 + 1] - originalRect.y) * ( float)scaleY +
                0.5);
    }

/* Attach arrow heads if needed. */

    l2 = (int)((20 + _dLineWidth) * display->GetDPI() / BASEDPI);
    w1 = (int)((10 + _dLineWidth) * display->GetDPI() / BASEDPI);

    if (  (arrows == AmARROW_HEAD || arrows == AmARROW_BOTH) &&
          ( numPoints >= 2)) {
        newPoints[ num - 1] = drawArrow( w, display, gc, newPoints[ num - 2],
                            newPoints[ num - 1], w1, l2);
    }

    if (  (arrows == AmARROW_TAIL || arrows == AmARROW_BOTH) &&
          ( numPoints >= 2)) {
        newPoints[ 0] = drawArrow( w, display, gc, newPoints[ 1],
                            newPoints[ 0], w1, l2);
    }
    display->DrawLines( w, gc, newPoints, num);
    free( newPoints);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a line annotation.  This is
 *      used in the map editor to dump the ticks's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmLineAnnotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			    AnvilBoolean	      allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmLineAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *	Update a line legend, by finding the extents of the points in
 *	the line.
 * Arguments:
 *	eventCodes - vector of event codes which launched the update
 *	numCodes - number of codes.
 * Returns:
 *	AmOK, or AmNOT_VALID if the line annotation's not valid.
 *
 *************************************************************************/

AmErrorCode
AmLineAnnotClass::Update( int	*eventCodes,
			int	numCodes)
{
    int minX, maxX, minY, maxY, ii;

    if ( !IsValid()) {
        return( AmNOT_VALID);
    }

    if ( numPoints >= 2) {
        minX = points[ 0];
        maxX = points[ 0];
        minY = points[ 1];
        maxY = points[ 1];
    } else {
        minX = 0;
        maxX = 0;
        minY = 0;
        maxY = 0;
    }

    for ( ii = 2; ii < numPoints; ii += 2) {   
        if ( points[ ii] < minX) {
            minX = points[ ii];
        } else if ( points[ ii] > maxX) {
            maxX = points[ ii];
        }
        if ( points[ ii + 1] < minY) {
            minY = points[ ii + 1];
        } else if ( points[ ii + 1] > maxY) {
            maxY = points[ ii + 1];
        }
    }

    originalRect.x = minX;
    originalRect.y = minY;
    originalRect.width = maxX - minX + 1;
    originalRect.height = maxY - minY + 1;
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a line annotation. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmLineAnnotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmLineAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
                                          int          *numDefs)
{
    static AmFieldDefRec LineAnnotClassFieldDefinitions[] = {
        { AmNarrows, offsetof( AmLineAnnotClass, arrows) },
        { AmNlineWidth, offsetof( AmLineAnnotClass, _dLineWidth) },
        { AmNnumPoints, offsetof( AmLineAnnotClass, numPoints) },
        { AmNpoints, offsetof( AmLineAnnotClass, points),
                     offsetof( AmLineAnnotClass, numPoints), AmDYNAMIC },
    };
    *fieldDefs = LineAnnotClassFieldDefinitions;
    *numDefs = VECLEN( LineAnnotClassFieldDefinitions);
}
