#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <anvil/tagstrs.h>
#include <anvil/rect.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/resource.h>
#include <anvil/multitextannot.h>

/* multiline line text annotation, same as text annotation, except that
   it takes multiple lines, and has text alignment control.
   Modifications:
	The function should allow rotated text, and control for text
	spacing.  Some kind of RichText support would be nice too.
	Being able to read text from files would be nice.
*/ 

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmMultiTextAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmMULTILINE_TEXT_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so Multiline text annotations can be subclassed.
 *
 ****************************************************************************/

AmMultiTextAnnotClass::AmMultiTextAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Destructor: free the label and font name, and the private 
 *	graphics context.
 *
 ****************************************************************************/

AmMultiTextAnnotClass::~AmMultiTextAnnotClass()
{
    if ( label != NULL) {
	free( label);
    }
    if ( font != NULL) {
	free( font);
    }
    if ( textGC != NULL) {
	delete textGC;
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the size of a multiline text 
 * Arguments:
 *	display - display text will be drawn to.
 *	textGC - graphics context text will be drawn with.
 *	str - string to draw with possible embedded newlines.
 *	width, height - returned width and height.
 * Modifications:
 *	AmChopString might return a NULL, or an error, which should be 
 *	propogated up to this function
 *
 ****************************************************************************/

static void 
multiTextSize( AmDisplayClass	*display,
	       AmGCClass		*textGC,
	       char		*str,
	       int		*width,
	       int		*height)
{
    char	**strvec, **strptr;
    int		strWidth, strAscent, strDescent, numlines;

    strvec = AmChopString( str, '\n');
    numlines = AmVeclen(( const char **) strvec);

/* find the longest string to set the rectangle */

    *width = 0;
    for ( strptr = strvec; *strptr != NULL; strptr++) {
	display->ExtentsText( textGC, *strptr, &strWidth, 
			      &strAscent, &strDescent);
	if ( strWidth > *width) {
	    *width = strWidth;
	}
    }
    AmFreeStrVec( strvec);
    *height = ( strAscent + strDescent)*numlines;
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a multiline text annotation.  The 
 *	intializeFields() functions are created for each class, and each 
 *	one calls the one for the "parent" class.  The AmScreenObjClass 
 *	intializeFields() function is the topmost one that is called.  
 *	We can't use two-parameter constructors, because since we use 
 *	varargs constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify annotation for resources
 *
 ****************************************************************************/

void 
AmMultiTextAnnotClass::initializeFields( AmWindowClass  *windowArg,
				       const char   *nameArg)
{
    int	width, height;

    AmAnnotationClass::initializeFields( windowArg, nameArg);
    objectType = AmMULTILINE_TEXT_ANNOTATION;

    opaque  = FALSE;
    _fFrame = FALSE;

    label = strdup( AmGetDefault( "multitextannot.default", 
				"This is a test string."));
    labelAngle = 0;
    font = strdup( DEFAULT_FONT);
    fontSize = DEFAULT_SIZE;
    fontWeight = DEFAULT_WEIGHT;
    fontSlant = DEFAULT_SLANT;
    fontXScale = 1.0;
    fontYScale = 1.0;

/* a local gc is used for text annotation, because setting fonts is very
   expensive */

    textGC = window->CreateGC();
    textGC->SetFont( font, fontSize, fontWeight, fontSlant,
                     fontXScale, fontYScale);

/* set the rectangle size if the label isn't NULL */

    if ( label != NULL) {
	multiTextSize( display, textGC, label, &width, &height);
        rect.width = width;
        rect.height = height;
        extentsRect = rect;
    }
}

/****************************************************************************
 *
 * Description:
 *      AmMultiTextAnnotClass constructor.  Creates a multiline text annotation
 *	using the window that it will be drawn on and the name, followed 
 *	by NULL-terminated tag-value pairs for setting the fields in the 
 *	multiline text annotation.
 * Arguments:
 *      windowArg - window that the multiline text annotation will be 
 *	drawn on.
 *      nameArg - string name of the multiline text anntotation, used for 
 *		identification and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmMultiTextAnnotClass::AmMultiTextAnnotClass( AmWindowClass  *windowArg,
					  const char   *nameArg,
					  ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    numValues = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    setFieldsWrapper( tagValues, numValues);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a multiline text annotation, given the 
 *	tag-values stored in a tag values array.  First, the annotation  
 *	fields are set, then the other fields are set using the standard 
 *	SetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.  If the string or font
 *	has changed, then the rectangle size is recalculated.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmMultiTextAnnotClass::setFields( AmTagValuePtr tagValues,
				int         numValues)
{
    int         numFieldDefs, width, height, tempWidth, tempHeight;
    AmFieldDefPtr fieldDefs;
    AnvilBoolean	widthChanged, heightChanged;
	
    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmMultiTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    tempWidth = rect.width;
    tempHeight = rect.height;
    widthChanged = FALSE;
    heightChanged = FALSE;

    if ( AmChanged( &rect.width)) {
        widthChanged = TRUE;
    }
    if ( AmChanged( &rect.height)) {
        heightChanged = TRUE;
    }
    if ( AmChanged( &label, &font, &fontSize, &fontWeight, &fontSlant,
                  &fontXScale, &fontYScale, NULL)) {
        textGC->SetFont( font, fontSize, fontWeight, fontSlant,
                         fontXScale, fontYScale);

/* if the label's been set, then find out how big it is */

        if ( label != NULL) {
	    multiTextSize( display, textGC, label, &width, &height);
            rect.width = width;
            rect.height = height;
            extentsRect = rect;
        }
    }

/* if the width or height were changed explicitly, they override the string */

    if ( widthChanged || heightChanged) {
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }

    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	set the rectangle containing the multiline text, which scales
 *	the font to accomodate the rectangle, then resizes the rectangle
 *	because there might not quite be a perfect match.
 * Arguments:
 *	x, y, width, height - dimensions of the new rectangle
 *
 ***************************************************************************/ 

void 
AmMultiTextAnnotClass::SetRect( int x,
			      int y,
			      int width,
			      int height)
{
    int 	textWidth, textHeight;

    if (( width == 0) || ( height == 0)) {
	return;
    }
    rect.x = x;
    rect.y = y;

    fontXScale = ( float)width / ( float)rect.width * fontXScale;
    fontYScale = ( float)height / ( float)rect.height * fontYScale;

    textGC->SetFont( font, fontSize, fontWeight, fontSlant,
		     fontXScale, fontYScale);

/* how many lines of text are being drawn? */

    if ( label != NULL) {
	multiTextSize( display, textGC, label, &textWidth, &textHeight);
        rect.width = textWidth;
        rect.height = textHeight;
    } else {
        rect.width = width;
        rect.height = height;
    }
    extentsRect = rect;
    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a multiline text annotation, given the 
 *	tag-values stored in a tag values array.  First, the annotation 
 *	fields are retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions() which uses a lookup table from the tag 
 *	strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmMultiTextAnnotClass::getFields( AmTagValuePtr tagValues,
				int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmMultiTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Draw the multiline text annotation to the screen.  Chop
 *	up the string, and draw each string individually.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 * Modifications:
 *	should return a AmMALLOC_FAILED if the AmChopString() failed,
 *	should also have control for string alignment.
 *
 ****************************************************************************/

AmErrorCode 
AmMultiTextAnnotClass::draw( AmWindowClass	      	*w,
			   AmGCClass	      	*gc,
			   AmRectangleRec const	*exposeRects,
			   int		      	nRects)
{
    int                width, ascent, descent, nClipRects, y;
    const AmRectangleRec *drawRect = GetRect();
    AmRectanglePtr       clipRects;
    char	       **strvec, **p;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* copy the clip rectangles from the display manager to the private GC */
    
    gc->GetClipRectangles( &clipRects, &nClipRects);
    textGC->SetClipRectangles( clipRects, nClipRects);
    free( clipRects);

/* clear the rectangle before drawing the legend.  Use Fill Rectangle, because
   ClearRect isn't sensitive to clip rectangles */

    if ( opaque == TRUE) {
	textGC->SetForegroundColor( &bgColor);
        display->FillRectangle( w, textGC, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }

/* set the foreground and background colors */

    textGC->SetForegroundColor( &fgColor);
    textGC->SetBackgroundColor( &bgColor);

/* find the extents and draw the text */

    display->ExtentsText( textGC, label, &width, &ascent, &descent);
    strvec = AmChopString( label, '\n');
    y = drawRect->y + ascent;
    for ( p = strvec; *p != NULL; p++) {
	display->DrawText( w, textGC, drawRect->x, y, *p);
	y += ascent + descent;
    }
    AmFreeStrVec( strvec);
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a multiline text annotation.  
 *	This is used in the map editor to dump the annotation's fields 
 *	into a file.  the field name (and length, if it's a vector field) 
 *	is written into the vector.  Each dumpFields member function 
 *	calls the dumpFields() member function of the parent class.  
 *	This is also used for copying objects for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmMultiTextAnnotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				 AnvilBoolean	   allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmMultiTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a multiline text annotation. 
 *	This is used by SetFieldsByDefinitions() and 
 *	GetFieldsByDefintions() to set the fields in the 
 *	MultLineTextAnnotClass.  It returns a vector of field names 
 *	followed by offsets, offset of the length of the field if 
 *	it's a vector, and a storageType, which is AmSTATIC for static 
 *	arrays in the class, AmDYNAMIC for dynamically allocated fields.
 *	THIS TABLE MUST BE SORTED IN ALPHABETICAL ORDER ON THE FIELD NAME 
 *	FOR SEARCHING 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmMultiTextAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
					  int          *numDefs)
{
    static AmFieldDefRec MultiTextAnnotClassFieldDefinitions[] = {
        { AmNfont, offsetof( AmMultiTextAnnotClass, font), 0, AmDYNAMIC },
        { AmNfontSize, offsetof( AmMultiTextAnnotClass, fontSize) },
        { AmNfontSlant, offsetof( AmMultiTextAnnotClass, fontSlant) },
        { AmNfontWeight, offsetof( AmMultiTextAnnotClass, fontWeight) },
        { AmNfontXScale, offsetof( AmMultiTextAnnotClass, fontXScale) },
        { AmNfontYScale, offsetof( AmMultiTextAnnotClass, fontYScale) },
        { AmNlabel, offsetof( AmMultiTextAnnotClass, label), 0, AmDYNAMIC },
        { AmNlabelAngle, offsetof( AmMultiTextAnnotClass, labelAngle) },
    };
    *fieldDefs = MultiTextAnnotClassFieldDefinitions;
    *numDefs = VECLEN( MultiTextAnnotClassFieldDefinitions);
}
