#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* member function definition for AmPolyLineLegendClass */

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/plineplot.h>
#include <anvil/plinelegend.h>

/*************************************************************************
 *
 * Description:
 *      recursive run-time typing function used for all screen objects
 * Arguments:
 *      type - requested type
 * Returns:
 *      TRUE if object or its ancestor class is of type "type", FALSE
 *      otherwise
 *
 *************************************************************************/

AnvilBoolean 
AmPolyLineLegendClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmLINE_LEGEND) {
        return( TRUE);
    } else {
        return( AmLegendClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so line legends can be subclassed.
 *
 ****************************************************************************/

AmPolyLineLegendClass::AmPolyLineLegendClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Line legend class destructor: free the line widths, and clean up
 *	the line width index data.
 *
 ****************************************************************************/

AmPolyLineLegendClass::~AmPolyLineLegendClass()
{
    if ( _pdLineWidth != NULL) {
	FREE( _pdLineWidth);
    }
    if ( _pdLineWidthScale != NULL) {
	FREE( _pdLineWidthScale);
    }
    AmDestroyIndexData( &_indexLineWidth);
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a line legend.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *	plotArg - parent plot that legend is based on (may be NULL for
 *	standalone legends)
 *      szName - string name used to identify legend for resources
 *
 ****************************************************************************/

void 
AmPolyLineLegendClass::initializeFields( AmWindowClass		*windowArg,
				         AmPolyLinePlotClass	*plotArg,
				         const char		*nameArg)
{
    AmLegendClass::initializeFields( windowArg, plotArg, nameArg);
    objectType = AmPOLYLINE_LEGEND;
    legendType = AmLEGEND_DISTINCT;
    showColors = TRUE;
    _fShowLineWidth = FALSE;
    _dLineWidth = 0;
    _pdLineWidth = NULL;
    _pdLineWidthScale = NULL;
    AmInitIndexData( &_indexLineWidth); 
}

/****************************************************************************
 *
 * Description:
 *      AmPolyLineLegendClass constructor.  Creates a poly line legend using 
 *	the window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in the 
 *	polyline legend.
 * Arguments:
 *      windowArg - window that the line legend will be drawn on.
 *	plotArg - parent plot that legend is based on (may be NULL for
 *	standalone legends)
 *      nameArg - string name of the line legend, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolyLineLegendClass::AmPolyLineLegendClass( AmWindowClass	    *windowArg,
					      AmPolyLinePlotClass   *plotArg,
					      const char	    *nameArg,
					      ...)
{
    va_list		args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int			numValues;

    if ( plotArg == NULL || plotArg->IsType( AmPOLYLINE_PLOT)) {
	AmPolyLineLegendClass::initializeFields( windowArg, plotArg, nameArg);
	va_start( args, nameArg);
	numValues = AmMakeTagValueArrayVararg( tagValues, args);
	va_end( args);
	setFieldsWrapper( tagValues, numValues);
	SetValid( TRUE);
    } else {
	SetValid( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for line legend, given the tag-values stored in
 *      a tag values array.  First, the legend fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolyLineLegendClass::setFields( AmTagValuePtr	tagValues,
				  int		numValues)
{
    int         ii, numFieldDefs, numUnusedFields;
    int         tempWidth, tempHeight;
    AmFieldDefPtr fieldDefs;

    numValues = AmLegendClass::setFields( tagValues, numValues);
    AmPolyLineLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNlineWidthBinvec)) {
            AmSetBinvec( &_indexLineWidth.binvec, _indexLineWidth.length,
                &tagValues[ ii]);
        } else { 
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }

    tempWidth = rect.width;
    tempHeight = rect.height;

    if ( AmChanged( &rect.width, &rect.height, NULL)) {
	Update();
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a poly line legend, given the tag-values 
 *	stored in a tag values array.  First, the generic legend fields 
 *	are retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolyLineLegendClass::getFields( AmTagValuePtr      tagValues,
				  int                numValues)
{
    int         numFieldDefs, numUnusedFields, ii;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmLegendClass::getFields( tagValues, numValues);
    AmPolyLineLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNlineWidthBinvec)) {
            if ( _indexLineWidth.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( _indexLineWidth.binvec,
                        ( _indexLineWidth.length - 1) * sizeof( float));
            } else {
                *( void **) tagValues[ ii].value = NULL;
            }
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 * 	Update a poly line legend when its parent has changed, or if it's
 *	been updated standalone.
 * Arguments:
 *	eventCodes - vector of events that caused the update.
 *	numEventCodes - number of events.
 * Returns:
 * 	AmOK on success, error on system error.
 * Modifications:
 *	There should be more checks for allocated vectors, and definitely
 *	more comments.
 *
 ***************************************************************************/

AmErrorCode
AmPolyLineLegendClass::Update( int	*eventCodes,
			       int	numEventCodes)
{
    AmDataGraphicClass	*graphic = ( AmDataGraphicClass *) parent;
    AmIndexDataPtr      index;
    float		plotXmax, plotXmin;
    int			ii;
    short		plotWidth;	

    if ( ( graphic != NULL) && !editable) {

        numColors = 0;
        if ( colors != NULL) {
            free( colors);
            colors = NULL;
        }
        AmDestroyIndexData( &colorIndex);

        if ( _pdLineWidth != NULL) {
            free( _pdLineWidth);
            _pdLineWidth = NULL;
        }
        AmDestroyIndexData( &_indexLineWidth);

        if ( _pdLineWidthScale != NULL) {
            free( _pdLineWidthScale);
            _pdLineWidthScale = NULL;
        }

        numLabels = 0;
        if ( labels != NULL) {
            AmFreeStrVec( labels);
            labels = NULL;
        }

/* plot width is a short, because the screen object rectangle values are
   shorts */

        graphic->GetVaFields( AmNcurrXmax, &plotXmax,
			      AmNcurrXmin, &plotXmin,
			      AmNwidth, &plotWidth,
			      NULL);

        itemExtents = ( float) plotWidth / ( float)( plotXmax - plotXmin);

        graphic->GetVaFields( AmNcolors, &colors, 
			      AmNnumColors, &numColors,
			      AmNcolorsMin, &colorIndex.min,
			      AmNcolorsMax, &colorIndex.max,
			      AmNcolorsBinvec, &colorIndex.binvec,
			      AmNcolorsUnits, &colorIndex.units,
			      AmNcolorsTable, &colorIndex.table,
			      AmNcolorsStrings, &colorIndex.strings,
			      NULL);

        colorIndex.length = numColors;

        graphic->GetVaFields( AmNlineWidthList, &_pdLineWidth,
			      AmNlineWidthLength, &_indexLineWidth.length,
			      AmNlineWidthMin, &_indexLineWidth.min,
			      AmNlineWidthMax, &_indexLineWidth.max,
			      AmNlineWidthBinvec, &_indexLineWidth.binvec,
			      AmNlineWidthUnits, &_indexLineWidth.units,
			      AmNlineWidthTable, &_indexLineWidth.table,
			      AmNlineWidthStrings, &_indexLineWidth.strings,
			      AmNscaleLineWidthList, &_pdLineWidthScale,
			      AmNlineWidth, &_dLineWidth,
			      NULL);
    }


    if (( _pdLineWidthScale != NULL) && ( _pdLineWidth == NULL)) {
        _pdLineWidth = ( float *) malloc( sizeof( int) *
                                          _indexLineWidth.length);
        for ( ii = 0; ii < _indexLineWidth.length; ii++) {
            _pdLineWidth[ ii] = _pdLineWidthScale[ ii] * itemExtents;
        }
    }

    if ( showColors && colors != NULL) {
        index = &colorIndex;
    } else if ( _fShowLineWidth && _pdLineWidth != NULL) {
        index = &_indexLineWidth;
    } else {

        _fShowLineWidth = showColors = FALSE;

        if ( _pdLineWidth != NULL) {
            _fShowLineWidth = TRUE;
            index = &_indexLineWidth;
        } else if ( colors != NULL) {
            showColors = TRUE;
            index = &colorIndex;
        } else {
            index = NULL;
        }
    }

    if ( index != NULL) {
        numItems = index->length;
 
        if ( ( graphic != NULL && !editable) || ( numLabels != numItems)) {
            if ( labels) {
                AmFreeStrVec( labels);
            }
            numLabels = index->length;
            if ( index->strings == NULL) {
                labels = AmIndexesToStrings( index, numLabels);
            } else {
                labels = AmCopyStrVec(( const char **) index->strings);
            }
        }
    }

    layoutRects();
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Draw a single legend item for a line legend.  The Draw()
 *      function is implemented in AmLegendClass, and calls DrawItems()
 *      member function as a virtual, which calls this.
 * Arguments:
 *      pWindow - window to draw the item to.
 *      pGC - GC to draw the item with.
 *      itemNumber - index of the item.
 *      pItemRect - item's screen rectangle.
 *
 *************************************************************************/

AmErrorCode
AmPolyLineLegendClass::DrawItem( AmWindowClass        *w,
			         AmGCClass            *gc,
			         int                  item,
			         AmRectangleRec const *drawRect)
{
    AmPointRec pts[ 2];

    if ( _fShowLineWidth && _pdLineWidth != NULL) {
        textGC->SetLineWidth( _pdLineWidth[ item]);
    } else {
        textGC->SetLineWidth( _dLineWidth);
    }

    if ( showColors && colors != NULL) {
        textGC->SetForegroundColor( &colors[ item]);
    } else {
        textGC->SetForegroundColor( &fgColor);
    }

    pts[ 0].x = drawRect->x;
    pts[ 0].y = drawRect->y + drawRect->height / 2;
    pts[ 1].x = drawRect->x + drawRect->width;
    pts[ 1].y = drawRect->y + drawRect->height / 2;
    display->DrawLines( w, textGC, pts, 2);

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Draw all the legend items. Called by AmLegendClass::DrawItems()
 *      as a virtual function.
 * Arguments:
 *      pWindow - window to draw to.
 *      pGC - GC to draw with.
 *      vItemRect - vector of item rectangles.
 * Returns:
 *      AmOK.
 *
 *************************************************************************/

AmErrorCode
AmPolyLineLegendClass::DrawItems( AmWindowClass        *w,
				  AmGCClass            *gc,
				  AmRectangleRec const *itemRects)
{
    int          ii;

    for ( ii = 0; ii < numItems - 1; ii++) {
        DrawItem( w, gc,
            ( increaseItems == TRUE) ? numItems - 1 - ii : ii,
            &itemRects[ ii]);
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a line legend.  This is
 *      used in the map editor to dump the line legend's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolyLineLegendClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				   AnvilBoolean	       allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmLegendClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmPolyLineLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmPolyLineLegendClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmPolyLineLegendClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmPolyLineLegendClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
					    int            *numDefs)
{
    static AmFieldDefRec LineLegendClassFieldDefinitions[] = {
        { AmNitemExtents, offsetof( AmPolyLineLegendClass, itemExtents) },
        { AmNlineWidth, offsetof( AmPolyLineLegendClass, _dLineWidth) },
        { AmNlineWidthLength,
            offsetof( AmPolyLineLegendClass, _indexLineWidth.length) },
        { AmNlineWidthList,
            offsetof( AmPolyLineLegendClass, _pdLineWidth),
            offsetof( AmPolyLineLegendClass, _indexLineWidth.length),
            AmDYNAMIC },
        { AmNlineWidthMax, 
	   offsetof( AmPolyLineLegendClass, _indexLineWidth.max) },
        { AmNlineWidthMin, 
	   offsetof( AmPolyLineLegendClass, _indexLineWidth.min) },
        { AmNlineWidthStrings,
            offsetof( AmPolyLineLegendClass, _indexLineWidth.strings),
            offsetof( AmPolyLineLegendClass, _indexLineWidth.length),
            AmDYNAMIC },
        { AmNlineWidthTable,
            offsetof( AmPolyLineLegendClass, _indexLineWidth.table),
            offsetof( AmPolyLineLegendClass, _indexLineWidth.length),
            AmDYNAMIC },
        { AmNlineWidthUnits,
            offsetof( AmPolyLineLegendClass, _indexLineWidth.units) },
        { AmNscaleLineWidthList,
            offsetof( AmPolyLineLegendClass, _pdLineWidthScale),
            offsetof( AmPolyLineLegendClass, _indexLineWidth.length),
            AmDYNAMIC },
        { AmNshowColors, 
	    offsetof( AmPolyLineLegendClass, showColors) },
        { AmNshowLineWidths, 
	    offsetof( AmPolyLineLegendClass, _fShowLineWidth) },
    };
    *fieldDefs = LineLegendClassFieldDefinitions;
    *numDefs = VECLEN( LineLegendClassFieldDefinitions);
}
