#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/* grid for polar graphs */

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/errcode.h>
#include <anvil/ratio.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/scale.h>
#include <anvil/userevents.h>
#include <anvil/datagraphic.h>
#include <anvil/polargraph.h>
#include <anvil/polargrid.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmPolarGridClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmTICKS) {
        return( TRUE);
    } else {
        return( AmScaleClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for polar grids.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.  The polar grid
 *      calls the AmScaleClass initializer, then initializes the the
 *      polar grid-specific fields, such as tick size, style and
 *      spacing.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify object for resources
 * Modifications:
 *      the size values should be consts in a source file, not hardcoded
 *
 ****************************************************************************/

void
AmPolarGridClass::initializeFields( AmWindowClass	   *windowArg,
				  AmPolarGraphClass  *graphicArg,
				  const char	   *nameArg)
{
    AmScaleClass::initializeFields( windowArg, graphicArg, nameArg);

    opaque  = FALSE;
    _fFrame = FALSE;

    numSpokes = 8;
    style = AmBASE_TEN;
    spacing = 1.0;
    base = 0.0;
    numTicks = 5;
    maxTicks = 10;
    showAngles = FALSE;
    showRadii = FALSE;
    navigational = FALSE;
    clockwise = FALSE;
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so polar grids can be subclassed.
 *
 ****************************************************************************/

AmPolarGridClass::AmPolarGridClass()
{
}

/****************************************************************************
 *
 * Description:
 *      destructor: nothing has to be freed which is specific to
 *      AmPolarGridClass.
 *
 ****************************************************************************/

AmPolarGridClass::~AmPolarGridClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmPolarGridClass constructor.  Creates a polar grid using
 *      the window that it will be drawn on and the name, followed by
 *      NULL-terminated tag-value pairs for setting the fields in the
 *      distance scale.
 * Arguments:
 *      windowArg - window that the polar grid will be drawn on.
 *      graphic - parent polar graph that the polar grid is tied to.
 *      nameArg - string name of the polar grid, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/
 
AmPolarGridClass::AmPolarGridClass( AmWindowClass	*windowArg,
			        AmPolarGraphClass	*graphic,
			        const char	*nameArg,
			        ...)
{
    
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		count;

    if ( !graphic->IsType( AmPOLAR_GRAPH)) {
	DO_ERROR( AmBAD_ARGUMENT);
	SetValid( FALSE);
    } else {
	initializeFields( windowArg, graphic, nameArg);
	va_start( args, nameArg);
	count = AmMakeTagValueArrayVararg( tagValues, args);
	setFieldsWrapper( tagValues, count);
	va_end( args);
	SetValid( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for the polar grid given the tag-values stored in
 *      a tag values array.  First, the AmScaleClass fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolarGridClass::setFields( AmTagValuePtr	tagValues,
			   int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmScaleClass::setFields( tagValues, numValues);
    AmPolarGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a polar grid, given the tag-values stored in
 *      a tag values array.  First, the scale fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolarGridClass::getFields( AmTagValuePtr	tagValues,
			   int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmScaleClass::getFields( tagValues, numValues);
    AmPolarGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Draw a polar graph grid.
 * Arguments:
 *      window - window the the polar grid will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - number of exposed rectangles.
 * Returns:
 *      AmOK if everything's OK.
 *
 ****************************************************************************/

AmErrorCode 
AmPolarGridClass::draw( AmWindowClass		*w,
		        AmGCClass		*gc,
		        AmRectangleRec const	*exposeRects,
		        int			nRects)	
{
    AmRectangleRec const	*drawRect = GetRect();
    AmPolarGraphClass	*graphic = ( AmPolarGraphClass *) parent;
    float		minVal, maxVal, angle, radius, radscale, realAngle;
    int			x0, y0, x1, y1, ii, nClipRects;
    AmRectanglePtr	clipRects;
    char		buffer[ 128];
    unitType		units;

/* only draw the outline if the object is invalid */

    if ( !IsValid()) {
	return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

    if ( AmDegenRect( drawRect)) {
	return( AmOK);
    }

/* set the clip rectangles for the text and title GC's which are copied from
   the GC passed down from the display manager */

    gc->GetClipRectangles( &clipRects, &nClipRects);
    textGC->SetClipRectangles( clipRects, nClipRects);
    titleGC->SetClipRectangles( clipRects, nClipRects);
    textGC->SetForegroundColor( &fgColor);
    titleGC->SetForegroundColor( &fgColor);
    gc->SetForegroundColor( &fgColor);
    free( clipRects);

/* set the ranging information for the circles */

    graphic->GetVaFields( AmNheightMin, &minVal,
			  AmNheightMax, &maxVal,
			  AmNheightUnits, &units,
			  NULL);
    if ( style == AmBASE_TEN) {
        spacing = pow( 10.0, floor( log10( maxVal - minVal)) - 1.0);
        if (( maxVal - minVal)/spacing > maxTicks) {
            spacing *= 10.0;
        }
        while (( maxVal - minVal)/spacing < maxTicks/2) {
            spacing /= 2.0;
        }
        base = (( int) ( minVal/spacing))*spacing;
    } else if ( style == AmEND_END) {
        spacing = ( maxVal - minVal)/numTicks;
        base = minVal;
    }

/* draw the concentric circles */

    for ( radius = base; radius <= maxVal; radius += spacing) {
	radscale = (radius - minVal)/( maxVal - minVal);
	x0 = (int)(rect.x + rect.width/2 - radscale*rect.width/2);
	y0 = (int)(rect.y + rect.height/2 - radscale*rect.height/2);
	x1 = (int)(rect.x + rect.width/2 + radscale*rect.width/2);
	y1 = (int)(rect.y + rect.height/2 + radscale*rect.height/2);
	display->DrawArc( w, gc, x0, y0, x1 - x0, y1 - y0, 0.0, M_PI*2);
	if ( showRadii) {
	    AmPrintValue( units, radius, buffer);
	    display->DrawAlignedText( w, gc, x1, rect.y + rect.height/2, 
				      buffer, AmLEFT_UP_VERTICAL);
	}
    }

/* draw the spokes */

    x0 = rect.x + rect.width/2;
    y0 = rect.y + rect.height/2;
    for ( ii = 0; ii < numSpokes; ii++) {
	angle = ( ii*M_PI*2)/numSpokes;
	realAngle = angle;
	if ( navigational == TRUE) {
	    realAngle = M_PI/2 - angle;
	} else if ( clockwise == TRUE) {
	    realAngle = -angle;
	}
	x1 = (int)(rect.x + rect.width/2 + cos( realAngle)*rect.width/2);
	y1 = (int)(rect.y + rect.height/2 - sin( realAngle)*rect.height/2);
	display->DrawLine( w, gc, x0, y0, x1, y1);
	if ( showAngles == TRUE) {
	    sprintf( buffer, "%.0f", angle*180.0/M_PI);
	    display->DrawRotatedText( w, gc, x1, y1, angle, buffer);
	}
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in AmPolarGridClass.  This is
 *      used in the map editor to dump the polar grid's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolarGridClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			    AnvilBoolean		 allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( scale) */

    numStrings = AmScaleClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmPolarGridClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmPolarGridClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmPolarGridClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmPolarGridClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				     int		*numDefs)
{
    static AmFieldDefRec PolarGridClassFieldDefinitions[] = {
	{ AmNclockwise, offsetof( AmPolarGridClass, clockwise) },
	{ AmNnavigational, offsetof( AmPolarGridClass, navigational) },
	{ AmNnumTicks, offsetof( AmPolarGridClass, numTicks) },
	{ AmNshowAngles, offsetof( AmPolarGridClass, showAngles) },
	{ AmNshowRadii, offsetof( AmPolarGridClass, showRadii) },
	{ AmNtickBase, offsetof( AmPolarGridClass, base) },
	{ AmNtickNumSpokes, offsetof( AmPolarGridClass, numSpokes) },
	{ AmNtickSpacing, offsetof( AmPolarGridClass, spacing) },
	{ AmNtickStyle, offsetof( AmPolarGridClass, style) }
    };
    *fieldDefs = PolarGridClassFieldDefinitions;
    *numDefs = VECLEN( PolarGridClassFieldDefinitions);
}
