#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* polygon legend class member function definitions. */

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/plot.h>
#include <anvil/polygonplot.h>
#include <anvil/polylegend.h>
#include <anvil/hatchpoly.h>


const float AmDEFAULT_HATCH_WIDTH = 0.0;
const int   AmDEFAULT_HATCH_ANGLE = 45;
const float AmDEFAULT_ITEM_EXTENTS = 100;

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean
AmPolyLegendClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmPOLYGON_LEGEND) {
        return( TRUE);
    } else {
        return( AmLegendClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so Polygon legends can be subclassed.
 *
 ****************************************************************************/

AmPolyLegendClass::AmPolyLegendClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Destructor: free the hatch width and hatch angle lists, and the
 *	hatchWidth and hatchAngle IndexData's
 *
 ****************************************************************************/


AmPolyLegendClass::~AmPolyLegendClass()
{
    if ( IsValid()) {

        if ( hatchWidthList != NULL) {
            free( hatchWidthList);
        }

        AmDestroyIndexData( &hatchWidthIndex);

        if ( hatchAngleList != NULL) {
            free( hatchAngleList);
        }

        AmDestroyIndexData( &hatchAngleIndex);
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a polygon legend.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify legend for resources
 *
 ****************************************************************************/

void 
AmPolyLegendClass::initializeFields( AmWindowClass		*windowArg,
				   AmPolygonPlotClass	*plotArg,
				   const char		*nameArg)
{
    AmLegendClass::initializeFields( windowArg, plotArg, nameArg);
    objectType = AmPOLYGON_LEGEND;
    legendType = AmLEGEND_DISTINCT;
    showColors = TRUE;
    showHatchAngles = FALSE;
    showHatchWidths = FALSE;
    hatched = FALSE;
    hatchWidth = AmDEFAULT_HATCH_WIDTH;
    AmInitIndexData( &hatchWidthIndex);
    hatchWidthList = NULL;
    hatchAngle = AmDEFAULT_HATCH_ANGLE;
    AmInitIndexData( &hatchAngleIndex);
    hatchAngleList = NULL;
    itemExtents = AmDEFAULT_ITEM_EXTENTS;
}

/****************************************************************************
 *
 * Description:
 *      AmPolyLegendClass constructor.  Creates a polygon legend using the 
 *	window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in the 
 *	polygon legend.
 * Arguments:
 *      windowArg - window that the polygon legend will be drawn on.
 *      nameArg - string name of the polygon legend, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolyLegendClass::AmPolyLegendClass( AmWindowClass		*windowArg,
				  AmPolygonPlotClass	*plotArg,
				  const char		*nameArg,
				  ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    if ( plotArg == NULL || plotArg->IsType( AmPOLYGON_PLOT)) {
	initializeFields( windowArg, plotArg, nameArg);
	va_start( args, nameArg);
	numValues = AmMakeTagValueArrayVararg( tagValues, args);
	va_end( args);
	setFieldsWrapper( tagValues, numValues);
	SetValid( TRUE);
    } else {
	SetValid( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a polygon legend, given the tag-values stored in
 *      a tag values array.  First, the generic legend fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/


int
AmPolyLegendClass::setFields( AmTagValuePtr tagValues,
                            int         numValues)
{
    int         numFieldDefs, numUnusedFields, ii;
    int         tempWidth, tempHeight;
    AmFieldDefPtr fieldDefs;

    numValues = AmLegendClass::setFields( tagValues, numValues);
    AmPolyLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNhatchAngleBinvec)) {
            AmSetBinvec( &hatchAngleIndex.binvec, hatchAngleIndex.length,
                &tagValues[ ii]);
        } else if ( !strcmp( tagValues[ ii].tag, AmNhatchWidthBinvec)) {
            AmSetBinvec( &hatchWidthIndex.binvec, hatchWidthIndex.length,
                &tagValues[ ii]);
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }

    tempWidth = rect.width;
    tempHeight = rect.height;

    if ( AmChanged( &rect.width, &rect.height)) {
	Update();
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a polygon legend, given the tag-values 
 *	stored in a tag values array.  First, the legend fields are 
 *	retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions() which uses a lookup table from the tag 
 *	strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

/* get the fields which are specific to a polygon plot legend */
 
int AmPolyLegendClass::getFields( AmTagValuePtr        tagValues,
				int                numValues)
{
    int         numFieldDefs, numUnusedFields, ii;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmLegendClass::getFields( tagValues, numValues);
    AmPolyLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
                    tagValues, numValues);

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNhatchAngleBinvec)) {
            if ( hatchAngleIndex.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( hatchAngleIndex.binvec,
                             ( hatchAngleIndex.length - 1) * sizeof( float));
            } else {
                * ( void **) tagValues[ ii].value = NULL;
            }
        } else if ( !strcmp( tagValues[ ii].tag, AmNhatchWidthBinvec)) {
            if ( hatchAngleIndex.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( hatchWidthIndex.binvec,
                             ( hatchWidthIndex.length - 1) * sizeof( float));
            } else {
                * ( void **) tagValues[ ii].value = NULL;
            }
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Update a polygon legend when its parent has changed, or if it's
 *      been updated standalone.
 * Arguments:
 *      eventCodes - vector of events that caused the update.
 *      numEventCodes - number of events.
 * Returns:
 *      AmOK on success, error on system error.
 * Modifications:
 *      There should be more checks for allocated vectors, and definitely
 *      more comments.
 *
 ***************************************************************************/

AmErrorCode
AmPolyLegendClass::Update( int	*eventCodes,
			 int	numEventCodes)
{
    AmDataGraphicClass *graphic = ( AmDataGraphicClass *) parent;
    AmIndexDataPtr     index;  
    short            plotWidth;
    float            plotXmax, plotXmin;

    if ( graphic != NULL) {

    /* Free the memory used up by the old definitions. */ 

        AmDestroyIndexData( &hatchAngleIndex);

        if ( hatchAngleList != NULL) {
            free( hatchAngleList);
            hatchAngleList = NULL;
        }

        AmDestroyIndexData( &hatchWidthIndex);

        if ( hatchWidthList != NULL) {
            free( hatchWidthList);
            hatchWidthList = NULL;
        }

        AmDestroyIndexData( &colorIndex);
        numColors = 0;

        if ( colors != NULL) {
            free( colors);
            colors = NULL;
        }

        numLabels = 0;
        if ( labels != NULL) {
            AmFreeStrVec( labels);
            labels = NULL;
        }

        graphic->GetVaFields( AmNdrawSolid, &solid,
                     AmNcurrXmax, &plotXmax,
                     AmNcurrXmin, &plotXmin,
                     AmNhatched, &hatched,
                     AmNwidth, &plotWidth,
                     NULL);

        itemExtents = ( float)plotWidth / ( float)( plotXmax - plotXmin);

    /* Get all the values from the parent. */

        graphic->GetVaFields( AmNhatchAngle, &hatchAngle,
                     AmNhatchAngleBinvec, &hatchAngleIndex.binvec,
                     AmNhatchAngleClientData, &hatchAngleIndex.clientData,
                     AmNhatchAngleFn, &hatchAngleIndex.fn,
                     AmNhatchAngleLength, &hatchAngleIndex.length,
                     AmNhatchAngleList, &hatchAngleList,
                     AmNhatchAngleMax, &hatchAngleIndex.max,
                     AmNhatchAngleMin, &hatchAngleIndex.min,
                     AmNhatchAngleOffset, &hatchAngleIndex.offset,
                     AmNhatchAngleStrings, &hatchAngleIndex.strings,
                     AmNhatchAngleTable, &hatchAngleIndex.table,
                     AmNhatchAngleType, &hatchAngleIndex.dataType,
                     NULL);

        graphic->GetVaFields( AmNhatchWidth, &hatchWidth,
                     AmNhatchWidthBinvec, &hatchWidthIndex.binvec,
                     AmNhatchWidthClientData, &hatchWidthIndex.clientData,
                     AmNhatchWidthFn, &hatchWidthIndex.fn,
                     AmNhatchWidthLength, &hatchWidthIndex.length,
                     AmNhatchWidthList, &hatchWidthList,
                     AmNhatchWidthMax, &hatchWidthIndex.max,
                     AmNhatchWidthMin, &hatchWidthIndex.min,
                     AmNhatchWidthOffset, &hatchWidthIndex.offset,
                     AmNhatchWidthStrings, &hatchWidthIndex.strings,
                     AmNhatchWidthTable, &hatchWidthIndex.table,
                     AmNhatchWidthType, &hatchWidthIndex.dataType,
                     NULL);

        graphic->GetVaFields( AmNcolors, &colors,
                     AmNnumColors, &numColors,
                     AmNcolorsMin, &colorIndex.min,
                     AmNcolorsMax, &colorIndex.max,
                     AmNcolorsBinvec, &colorIndex.binvec,
                     AmNcolorsUnits, &colorIndex.units,
                     AmNcolorsTable, &colorIndex.table,
                     AmNcolorsStrings, &colorIndex.strings,
                     NULL);
        colorIndex.length = numColors;
    }

/*
 * If no variable has been selected to display the legend by
 *   try displaying using hatch widhts.  If that doesn't work
 *   try displaying by hatch angles, finally if that doens't
 *   work try using colors.
 */

    if ( showHatchWidths && hatchWidthList != NULL) {
        index = &hatchWidthIndex;
    } else if ( showHatchAngles && hatchAngleList != NULL) {
        index = &hatchAngleIndex;
    } else if ( showColors && colors != NULL) {
        index = &colorIndex;
    } else {

        showHatchWidths = showHatchAngles = showColors = FALSE;

        if ( hatchWidthList != NULL) {
            showHatchWidths = TRUE;
            index = &hatchWidthIndex;
        } else if ( hatchAngleList != NULL) {
            showHatchAngles = TRUE;
            index = &hatchAngleIndex;
        } else if ( colors != NULL) {
            showColors = TRUE;
            index = &colorIndex;
        } else {
            index = NULL;
        }
    }

    if ( index != NULL) {
        numItems = index->length;

        if ( graphic != NULL || numLabels != numItems) {
            if ( labels) {
                AmFreeStrVec( labels);
            }
            numLabels = index->length;
            if ( index->strings == NULL) {
                labels = AmIndexesToStrings( index, numLabels);
            } else {
                labels = AmCopyStrVec(( const char **) index->strings);
            }
        }
    }

    if ( hatchWidth == 0.0) {
        if ( graphic) {
            hatchWidth = ( plotXmax - plotXmin) / 100.0;
        } else {
            hatchWidth =  1 / 100.0;
        }
    }

    AN_DEBUG( 30, ( "Laying out rectangles.\n"));

    layoutRects();

    AN_DEBUG( 30, ( "Done updating.\n"));
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Draw a single legend item for a polygon legend.  The Draw()
 *      function is implemented in AmLegendClass, and calls DrawItems()
 *      member function as a virtual, which calls this.
 * Arguments:
 *      pWindow - window to draw the item to.
 *      pGC - GC to draw the item with.
 *      itemNumber - index of the item.
 *      pItemRect - item's screen rectangle.
 *
 *************************************************************************/

AmErrorCode
AmPolyLegendClass::DrawItem( AmWindowClass        *w,
                           AmGCClass            *gc,
                           int                item,
                           AmRectangleRec const *drawRect)
{
    
    AmPointRec poly[ 5];

/* Set the color for the rectangle. */

    if ( showColors) {
        textGC->SetForegroundColor( &colors[ item]);
    } else {
        textGC->SetForegroundColor( &fgColor);
    }

/* Outline or fill the rectangle */

    if ( solid) {
        display->FillRectangle( w, textGC, drawRect->x, drawRect->y,
                     drawRect->width, drawRect->height);
    } else {
        display->DrawRectangle( w, textGC, drawRect->x, drawRect->y,
                     drawRect->width, drawRect->height);

        if ( hatched || showHatchAngles || showHatchWidths) {
	    AmGCClass	*hatchGC = w->CreateGC();

        /* Setup the rectangle to be hatched. */

            poly[ 0].x = drawRect->x;
            poly[ 0].y = drawRect->y;
            poly[ 1].x = drawRect->x + drawRect->width;
            poly[ 1].y = drawRect->y;
            poly[ 2].x = drawRect->x + drawRect->width;
            poly[ 2].y = drawRect->y + drawRect->height;
            poly[ 3].x = drawRect->x;
            poly[ 3].y = drawRect->y + drawRect->height;

            if ( showHatchWidths) {
                if ( hatchWidthList != NULL) {
                    hatchWidth = hatchWidthList[ item];
                } else {
                    AN_DEBUG( 20, (
                        "PolyLegend: No hatch widht list, using initial.\n"));
                }
            } else if ( showHatchAngles) {
	        if ( hatchAngleList != NULL) {
		    hatchAngle = hatchAngleList[ item];
	        } else {
                    AN_DEBUG( 20, (
                        "PolyLegend: No hatch angle list, using inital.\n"));
                }
            }

        /* This will give the extents of the legend with respect
           to the coordinate system of the plot. */

            AmHatchPoly( display, w, textGC, hatchGC, poly, 4, drawRect,
		       DEGTORAD( hatchAngle), 
		       (int)MAX( itemExtents * hatchWidth, 1));
	    delete hatchGC;
        }
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Draw all the legend items. Called by AmLegendClass::DrawItems()
 *      as a virtual function.
 * Arguments:
 *      pWindow - window to draw to.
 *      pGC - GC to draw with.
 *      vItemRect - vector of item rectangles.
 * Returns:
 *      AmOK.
 *
 *************************************************************************/

AmErrorCode
AmPolyLegendClass::DrawItems( AmWindowClass        *w,
                            AmGCClass            *gc,
                            AmRectangleRec const *itemRects)
{
    int          ii;

    for ( ii = 0; ii < numItems - 1; ii++) {
        DrawItem( w, gc,
            ( increaseItems == TRUE) ? numItems - 1 - ii : ii,
            &itemRects[ ii]);
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a polygon legend.  This is
 *      used in the map editor to dump the polygon legend's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolyLegendClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			     AnvilBoolean	       allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( plot) */

    numStrings = AmLegendClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmPolyLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmPolyLegendClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmPolyLegendClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmPolyLegendClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					   int		*numDefs)
{
    static AmFieldDefRec PolyLegendClassFieldDefinitions[] = {
        { AmNdrawSolid, offsetof( AmPolyLegendClass, solid) },
        { AmNhatchAngle, offsetof( AmPolyLegendClass, hatchAngle) },
        { AmNhatchAngleBinvec,
          offsetof( AmPolyLegendClass, hatchAngleIndex.binvec),
          offsetof( AmPolyLegendClass, hatchAngleIndex.length),
          AmDYNAMIC },
        { AmNhatchAngleLength,
          offsetof( AmPolyLegendClass, hatchAngleIndex.length) },
        { AmNhatchAngleList, offsetof( AmPolyLegendClass, hatchAngleList),
          offsetof( AmPolyLegendClass, hatchAngleIndex.length),
          AmDYNAMIC },
        { AmNhatchAngleMax,
          offsetof( AmPolyLegendClass, hatchAngleIndex.max) },
        { AmNhatchAngleMin,
          offsetof( AmPolyLegendClass, hatchAngleIndex.min) },
        { AmNhatchAngleStrings,
          offsetof( AmPolyLegendClass, hatchAngleIndex.strings),
          offsetof( AmPolyLegendClass, hatchAngleIndex.length),
          AmDYNAMIC },
        { AmNhatchAngleTable,
          offsetof( AmPolyLegendClass, hatchAngleIndex.table) ,
          offsetof( AmPolyLegendClass, hatchAngleIndex.length),
          AmDYNAMIC },
        { AmNhatchAngleUnits,
          offsetof( AmPolyLegendClass, hatchAngleIndex.units) },
        { AmNhatchWidth, offsetof( AmPolyLegendClass, hatchWidth) },
        { AmNhatchWidthBinvec,
          offsetof( AmPolyLegendClass, hatchWidthIndex.binvec),
          offsetof( AmPolyLegendClass, hatchWidthIndex.length),
          AmDYNAMIC },
        { AmNhatchWidthLength,
          offsetof( AmPolyLegendClass, hatchWidthIndex.length) },
        { AmNhatchWidthList,
          offsetof( AmPolyLegendClass, hatchWidthList),
          offsetof( AmPolyLegendClass, hatchWidthIndex.length),
          AmDYNAMIC },
        { AmNhatchWidthMax,
          offsetof( AmPolyLegendClass, hatchWidthIndex.max) },
        { AmNhatchWidthMin,
          offsetof( AmPolyLegendClass, hatchWidthIndex.min) },
        { AmNhatchWidthStrings,
          offsetof( AmPolyLegendClass, hatchWidthIndex.strings),
          offsetof( AmPolyLegendClass, hatchWidthIndex.length),
          AmDYNAMIC },
        { AmNhatchWidthTable,
          offsetof( AmPolyLegendClass, hatchWidthIndex.table),
          offsetof( AmPolyLegendClass, hatchWidthIndex.length),
          AmDYNAMIC },
        { AmNhatchWidthUnits,
          offsetof( AmPolyLegendClass, hatchWidthIndex.units) },
        { AmNhatched, offsetof( AmPolyLegendClass, hatched) },
        { AmNitemExtents, offsetof( AmPolyLegendClass, itemExtents) },
	{ AmNshowColors, offsetof( AmPolyLegendClass, showColors) },
	{ AmNshowHatchAngles, offsetof( AmPolyLegendClass, showHatchAngles) },
	{ AmNshowHatchWidths, offsetof( AmPolyLegendClass, showHatchWidths) }
    };

    *fieldDefs = PolyLegendClassFieldDefinitions;
    *numDefs = VECLEN( PolyLegendClassFieldDefinitions);
}
