#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* member functions for rectangle annotation class: supports open and
   filled rectangles with line widths. */

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <math.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/rectangleannot.h>
#include <anvil/tagstrs.h>
#include <anvil/rect.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmRectangleAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmRECTANGLE_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so rectangle annotations can be subclassed.
 *
 ****************************************************************************/

AmRectangleAnnotClass::AmRectangleAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Destructor: doesn't need to free anything.
 *
 ****************************************************************************/

AmRectangleAnnotClass::~AmRectangleAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a rectangle annotation.  The 
 *	intializeFields() functions are created for each class, and each 
 *	one calls the one for the "parent" class.  The AmScreenObjClass 
 *	intializeFields() function is the topmost one that is called.  We 
 *	can't use two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify object for resources
 *
 ****************************************************************************/

void
AmRectangleAnnotClass::initializeFields( AmWindowClass *windowArg,
                                       const char   *nameArg)
{
    AmAnnotationClass::initializeFields( windowArg, nameArg);
    objectType = AmRECTANGLE_ANNOTATION;

    opaque  = FALSE;
    _fFrame = FALSE;

    filled = FALSE;
    _dLineWidth = 0.0;
}

/****************************************************************************
 *
 * Description:
 *      AmRectangleAnnotClass constructor.  Creates a rectangle annotation 
 *	using the window that it will be drawn on and the name, followed 
 *	by NULL-terminated tag-value pairs for setting the fields in the 
 *	rectangle annotation.
 * Arguments:
 *      windowArg - window that the rectangle annotation will be drawn on.
 *      nameArg - string name of the rectangle annotation, used for 
 *	identification and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmRectangleAnnotClass::AmRectangleAnnotClass( AmWindowClass *windowArg,
                                          const char   *nameArg,
                                          ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    numValues = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    setFieldsWrapper( tagValues, numValues);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a rectangle annotation, given the tag-values 
 *	stored in a tag values array.  First, the annotation fields are 
 *	set, then the other fields are set using the standard 
 *	SetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.  If the line width
 *	has changed, then set the object's extents rectangle to cover it.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmRectangleAnnotClass::setFields( AmTagValuePtr tagValues,
				int         numValues)
{
    int                numFieldDefs;
    AmFieldDefPtr        fieldDefs;
    const AmRectangleRec *tempRect;

    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmRectangleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
                    tagValues, numValues);
    if ( AmChanged( &_dLineWidth)) {
        tempRect = GetRect();
        SetRect( tempRect->x, tempRect->y, tempRect->width, tempRect->height);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a rectangle annotation, given the 
 *	tag-values stored in a tag values array.  First, the annotation
 *	fields are retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions(), which uses a lookup table from the tag 
 *	strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmRectangleAnnotClass::getFields( AmTagValuePtr tagValues,
				int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmRectangleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	set the rectangle annotation's rectangle.  Make the extents
 *	rect large enough to cover the line width.
 * Arguments:
 *	x, y, width, height - rectangle dimensions.
 * 
 ****************************************************************************/ 

void 
AmRectangleAnnotClass::SetRect( int	x,
			      int	y,
			      int	width,
			      int	height)
{
    int iBorder;

    rect.x = x;
    rect.y = y;
    rect.width = width;
    rect.height = height;

/* set the extents rectangle to reflect the linewidth */
 
    iBorder = (int)MAX( ceil( _dLineWidth ), 1);

    extentsRect.x = rect.x - iBorder;
    extentsRect.y = rect.y - iBorder;
    extentsRect.width = rect.width + iBorder * 2;
    extentsRect.height = rect.height + iBorder * 2; 

    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *      Draw the rectangle annotation to the screen. 
 * Arguments:
 *      window - window that the annotation will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmRectangleAnnotClass::draw( AmWindowClass	   	*w,
			   AmGCClass		*gc,
			   AmRectangleRec const   *exposeRect,
			   int		   	nRects)
{
    const AmRectangleRec *drawRect = GetRect();

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRect, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( drawRect)) {
        return( AmOK);
    }

/* clear the rectangle before drawing the legend Use Fill Rectangle, because
   ClearRect isn't sensitive to gc clip regions */

    if ( opaque == TRUE) {
	gc->SetForegroundColor( &bgColor);
        display->FillRectangle( w, gc, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }

/* fill or draw the rectangle */

    gc->SetForegroundColor( &fgColor);
    gc->SetLineWidth( _dLineWidth);
    if ( filled == TRUE) {
        display->FillRectangle( w, gc,
                     drawRect->x, drawRect->y,
                     drawRect->width, drawRect->height);
    } else {
        display->DrawRectangle( w, gc,
                     drawRect->x, drawRect->y,
                     drawRect->width, drawRect->height);
    }

    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a rectangle annotation.  
 *	This is used in the map editor to dump the annotation's fields 
 *	into a file.  The field name (and length, if it's a vector field) 
 *	is written into the vector.  Each dumpFields member function 
 *	calls the dumpFields() member function of the parent class.  
 *	This is also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmRectangleAnnotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				 AnvilBoolean	   allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmRectangleAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a rectangle annotation. 
 *	This is used by SetFieldsByDefinitions() and 
 *	GetFieldsByDefintions() to set the fields in the RectAnnotClass.  
 *	It returns a vector of field names followed by offsets, offset 
 *	of the length of the field if it's a vector, and a storageType, 
 *	which is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  THIS TABLE MUST BE SORTED
 *  	IN ALPHABETICAL ORDER ON THE FIELD NAME FOR SEARCHING 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmRectangleAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
					  int          *numDefs)
{
    static AmFieldDefRec RectangleAnnotClassFieldDefinitions[] = {
        { AmNfilled, offsetof( AmRectangleAnnotClass, filled) },
        { AmNlineWidth, offsetof( AmRectangleAnnotClass, _dLineWidth) },
    };
    *fieldDefs = RectangleAnnotClassFieldDefinitions;
    *numDefs = VECLEN( RectangleAnnotClassFieldDefinitions);
}
