#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* scale class member function definitions.  Scales are the base class
   for legends, ticks and grid, and define resources for text and titles. */

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <anvil/macros.h>
#include <anvil/display.h>
#include <anvil/files.h>
#include <anvil/scale.h>
#include <anvil/defaults.h>
#include <anvil/userevents.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmScaleClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmSCALE) {
        return( TRUE);
    } else {
        return( AmScreenObjClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so scales can be subclassed.
 *
 ****************************************************************************/

AmScaleClass::AmScaleClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Default destructor: free the title and fonts, and the private
 *	GC's for the title and text.
 *
 ****************************************************************************/

AmScaleClass::~AmScaleClass()
{
    if ( title != NULL) {
        free( title);
    }
    if ( titleFont != NULL) {
        free( titleFont);
    }
    if ( font != NULL) {
	free( font);
    }

    delete textGC;
    delete titleGC;
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a scale.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify scale for resources
 *
 ****************************************************************************/

void 
AmScaleClass::initializeFields( AmWindowClass	 *windowArg,
			      AmScreenObjClass     *graphicArg,
			      const char	 *nameArg)
{

/* set the screen object lookup table and the database 
   for the enumeration lookup */

    AmScreenObjClass::initializeFields( windowArg, nameArg);
    setParent( graphicArg);

    _rectInterior.x      = 0;
    _rectInterior.y      = 0;
    _rectInterior.width  = 0;
    _rectInterior.height = 0;

    _dLineWidth = 0.0;
    titleGC = window->CreateGC();

    showTitle = FALSE;
    title = NULL;
    titleFont = strdup( DEFAULT_FONT);
    titleFontSize = DEFAULT_SIZE;
    titleFontWeight = AmWEIGHT_BOLD;
    titleFontSlant = DEFAULT_SLANT;
    titleFontXScale = 1.0;
    titleFontYScale = 1.0;
    titleVertical = TRUE;
    titleOffset = 5;

    textGC = window->CreateGC();

    font = strdup( DEFAULT_FONT);
    fontSize = 10;
    fontWeight = DEFAULT_WEIGHT;
    fontSlant = DEFAULT_SLANT;
    fontXScale = 1.0;
    fontYScale = 1.0;
    textVertical = FALSE;
    textOffset = 5;
    showText = TRUE; 

    min = 0.0;		/* set the min and max to something reasonable */
    max = 1.0;

    titleGC->SetForegroundColor( &fgColor);
    titleGC->SetFont( titleFont, titleFontSize, titleFontWeight,
                      titleFontSlant, titleFontXScale, titleFontYScale);

    textGC->SetForegroundColor( &fgColor);
    textGC->SetFont( font, fontSize, fontWeight, fontSlant,
                     fontXScale, fontYScale);

    if ( graphicArg != NULL) {
        graphicArg->GetVaFields( AmNmin, &min, AmNmax, &max, NULL);
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a scale, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScaleClass::getFields( AmTagValuePtr	tagValues,
		       int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmScreenObjClass::getFields( tagValues, numValues);
    AmScaleClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a scale, given the tag-values stored in
 *      a tag values array.  First, the screen object fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScaleClass::setFields( AmTagValuePtr	tagValues,
		       int		numValues)
{
    int         	numFieldDefs;
    AmFieldDefPtr 	fieldDefs;

    AmScaleClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = AmScreenObjClass::setFields( tagValues, numValues);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    if ( AmChanged( &font, &fontSize, &fontWeight, &fontSlant,
		  &fontXScale, &fontYScale, 
		  NULL)) {
	textGC->SetFont( font, fontSize, fontWeight, fontSlant,
			 fontXScale, fontYScale);
    }
    if ( AmChanged( &titleFont, &titleFontSize, &titleFontWeight,
		  &titleFontSlant, &titleFontXScale, &titleFontYScale, 
		  NULL)) {
	titleGC->SetFont( titleFont, titleFontSize, titleFontWeight,
			  titleFontSlant, titleFontXScale, titleFontYScale);
    }
    return( numValues);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields for a scale.  This is
 *      used in the map editor to dump the scale's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmScaleClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			AnvilBoolean	 	  allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* Get the fields for the parent class ( screen object). */

    numStrings = AmScreenObjClass::dumpFields( fieldLengths, allFields);

/* Get all the easy to specify fields for this class. */

    AmScaleClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 * 	DumpFields() wrapper function.
 * Modifications:
 *	this function should need to exist. 
 *
 *************************************************************************/

AnvilBoolean
AmScaleClass::DumpFields( FILE *fp)
{
    const char *objectName;

    objectName = getObjectTypeName();
    if ( objectName == NULL) {
        return( FALSE);
    }

    fprintf( fp, "%s Plot%p ( Plot%p ) {\n", objectName, this, parent);
    AmRecordClass::DumpFields( fp);
    fprintf( fp, "}\n");

    return( TRUE);
}

/*************************************************************************
 *
 * Description:
 * 	utility function for figuring out base 10 and scaling 
 * Arguments:
 *	min, max - ranget to scale over.
 *	maxTicks - maximum number of ticks to display.
 *	spacing - returned spacing.
 *	base - returned starting value.
 *
 *************************************************************************/

void 
AmScaleBaseTenSpacing( float	min,
		     float	max,
		     int	maxTicks,
		     float	*spacing,
		     float	*base)
{
    *spacing = pow( 10.0, floor( log10( max - min)) - 1.0);
    if (( max - min)/( *spacing) > maxTicks) {
	*spacing *= 10.0;
    }
    while (( max - min)/( *spacing) < maxTicks/2) {
	*spacing /= 2.0;
    }
    *base = (( int) ( min/( *spacing)))*( *spacing);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmScaleClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmScaleClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void
AmScaleClass::GetFieldDefinitions( AmFieldDefPtr *fieldDefs,
                                 int         *numDefs)
{
    static AmFieldDefRec ScaleClassFieldDefinitions[] = {
	{ AmNfont, offsetof( AmScaleClass, font), 0, AmDYNAMIC },
	{ AmNfontSize, offsetof( AmScaleClass, fontSize) },
	{ AmNfontSlant, offsetof( AmScaleClass, fontSlant) },
	{ AmNfontWeight, offsetof( AmScaleClass, fontWeight) },
	{ AmNfontXScale, offsetof( AmScaleClass, fontXScale) },
	{ AmNfontYScale, offsetof( AmScaleClass, fontYScale) },
	{ AmNlineWidth, offsetof( AmScaleClass, _dLineWidth) },
	{ AmNmax, offsetof( AmScaleClass, max) },
	{ AmNmin, offsetof( AmScaleClass, min) },
	{ AmNshowText, offsetof( AmScaleClass, showText) },
	{ AmNshowTitle, offsetof( AmScaleClass, showTitle) },
	{ AmNtextOffset, offsetof( AmScaleClass, textOffset) },
	{ AmNtextVertical, offsetof( AmScaleClass, textVertical) },
	{ AmNtitle, offsetof( AmScaleClass, title), 0, AmDYNAMIC },
	{ AmNtitleFont, offsetof( AmScaleClass, titleFont), 0, AmDYNAMIC },
	{ AmNtitleFontSize, offsetof( AmScaleClass, titleFontSize) },
	{ AmNtitleFontSlant, offsetof( AmScaleClass, titleFontSlant) },
	{ AmNtitleFontWeight, offsetof( AmScaleClass, titleFontWeight) },
	{ AmNtitleFontXScale, offsetof( AmScaleClass, titleFontXScale) },
	{ AmNtitleFontYScale, offsetof( AmScaleClass, titleFontYScale) },
	{ AmNtitleOffset, offsetof( AmScaleClass, titleOffset) },
	{ AmNtitleVertical, offsetof( AmScaleClass, titleVertical) },
    };
    *fieldDefs = ScaleClassFieldDefinitions;
    *numDefs = VECLEN( ScaleClassFieldDefinitions);
}
