#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stddef.h>
#include <anvil/defaults.h>
#include <anvil/files.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/pixthing.h>
#include <anvil/userevents.h>
#include <anvil/scatterplot.h>
#include <anvil/scatterlegend.h>

/***************************************************************************** 
 *
 * Description:
 *   Creates a symbol containing information on how it is to be displayed. 
 *   The data variable points to the symbol data where one byte AmEquals one
 *   pixel (if the data is to be freed when the symbol gets destroyed the
 *   dataCopied variable should be set to TRUE).  The width and height give
 *   the dimensions of the symbol data.  The scale gives the factor by
 *   which to scale the symbol by.
 * Arguments:
 *   data - data for the symbol
 *   dataCopied - "I've been copied, free me! Flag"
 *   width, height - dimensions of the symbol
 *   scale - scale factor to apply when drawing symbols.
 *	
 ****************************************************************************/

static AmSymbolPtr
createSymbol( uchar_t *data,
              AnvilBoolean dataCopied,
              int     width,
              int     height,
              float   scale)
{
    AmSymbolPtr newSymbol;

    newSymbol = ( AmSymbolPtr) malloc( sizeof( AmSymbolRec));

    if ( newSymbol == NULL) {
        DO_ERROR( AmMALLOC_FAILED);
        return( NULL);
    }

    newSymbol->data = data;
    newSymbol->dataCopied = dataCopied;
    newSymbol->width = width;
    newSymbol->height = height;
    newSymbol->scale = scale;

    return( newSymbol);
}


/***************************************************************************** 
 *
 * Description:
 *   Releases the memory used by a symbol.
 *	
 ****************************************************************************/

void
AmFreeSymbol( AmSymbolPtr symbol)
{
    if ( symbol != NULL) {
        if (( symbol->data != NULL) && ( symbol->dataCopied == TRUE)) {
            free( symbol->data);
        }
        free( symbol);
    }
}

/*************************************************************************
 *
 * Description:
 *      recursive run-time typing function used for all screen objects
 * Arguments:
 *      type - requested type
 * Returns:
 *      TRUE if object or its ancestor class is of type "type", FALSE
 *      otherwise
 *
 *************************************************************************/

AnvilBoolean
AmScatterLegendClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmSCATTER_LEGEND) {
        return( TRUE);
    } else {
        return( AmLegendClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so scatter legends can be subclassed.
 *
 ****************************************************************************/

AmScatterLegendClass::AmScatterLegendClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Scatter legend class destructor: free the symbol file, the image
 *	used for drawing, the symbols, and symbol name list, and free
 *	the symbol and symbolSize index structures.
 *
 ****************************************************************************/

AmScatterLegendClass::~AmScatterLegendClass()
{
    int ii;

    if ( symfile != NULL) {
        free( symfile);
    }

    if ( image != NULL) {
        delete image;
        image = NULL;
    }

    if ( symbols != NULL) {
        for ( ii = 0; ii < numSymbols; ii++) {
             AmFreeSymbol( symbols[ ii]);
        }
        free( symbols);
    }

    if ( symbolList != NULL) {
        for ( ii = 0; ii < symbolIndex.length; ii++) {
            free( symbolList[ ii]);
        }
        free( symbolList);
    }
    AmDestroyIndexData( &symbolIndex);

    if ( symbolSizeList != NULL) {
        free( symbolSizeList);
    }
    AmDestroyIndexData( &symbolSizeIndex);
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a scatter legend.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      plotArg - parent plot that legend is based on (may be NULL for
 *      standalone legends)
 *      szName - string name used to identify legend for resources
 *
 ****************************************************************************/

void 
AmScatterLegendClass::initializeFields( AmWindowClass		*windowArg,
				        AmScatterPlotClass	*plotArg,
				        const char		*nameArg)
{
    AmLegendClass::initializeFields( windowArg, plotArg, nameArg);
    objectType = AmSCATTER_LEGEND;
    showColors = FALSE;
    showSymbols = FALSE;
    showSymbolSizes = FALSE;
    symfile = strdup("x");
    symbolList = NULL;
    symbolSizeList = NULL;
    image = NULL;
    symbols = NULL;
    numSymbols = 0;
    AmInitIndexData( &symbolIndex);
    AmInitIndexData( &symbolSizeIndex);
}

/****************************************************************************
 *
 * Description:
 *      AmScatterLegendClass constructor.  Creates a scatter legend using the 
 *	window that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the scatter legend.
 * Arguments:
 *      windowArg - window that the scatter legend will be drawn on.
 *      plotArg - parent plot that legend is based on (may be NULL for
 *      standalone legends)
 *      nameArg - string name of the scatter legend, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

/* constructor to create a scatter legend */

AmScatterLegendClass::AmScatterLegendClass( AmWindowClass	*windowArg,
					    AmScatterPlotClass	*plotArg,
					    const char		*nameArg,
					    ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    if ( plotArg == NULL || plotArg->IsType( AmSCATTER_PLOT)) { 
	initializeFields( windowArg, plotArg, nameArg);
	va_start( args, nameArg);
	numValues = AmMakeTagValueArrayVararg( tagValues, args);
	va_end( args);
	setFieldsWrapper( tagValues, numValues);
	SetValid( TRUE);
    } else {
	SetValid( FALSE);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a scatter legend, given the tag-values stored in
 *      a tag values array.  First, the legend fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterLegendClass::setFields( AmTagValuePtr	tagValues,
			         int		numValues)
{
    int         numFieldDefs, numUnusedFields, ii;
    int         tempWidth, tempHeight;
    AmFieldDefPtr fieldDefs;

    numValues = AmLegendClass::setFields( tagValues, numValues);
    AmScatterLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNsymbolBinvec)) {
            AmSetBinvec( &symbolIndex.binvec, symbolIndex.length,
                       &tagValues[ ii]);
        } else if ( !strcmp( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
            AmSetBinvec( &symbolSizeIndex.binvec, symbolSizeIndex.length,
                       &tagValues[ ii]);
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }

    tempWidth = rect.width;
    tempHeight = rect.height;

    if ( AmChanged( &rect.width, &rect.height, NULL)) {
	Update();
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a scatter legend, given the tag-values stored in
 *      a tag values array.  First, the generic legend fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterLegendClass::getFields( AmTagValuePtr      tagValues,
				 int                numValues)

{
    int         numFieldDefs, numUnusedFields, ii;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmLegendClass::getFields( tagValues, numValues);
    AmScatterLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
                    tagValues, numValues);

    numUnusedFields = 0;

/* retrieve the bin vectors by hand, because they're one shorter than the 
   value used to reference their length */

    for ( ii = 0; ii < numValues; ii++) {
        if ( !strcmp( tagValues[ ii].tag, AmNsymbolBinvec)) {
            if ( symbolIndex.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( symbolIndex.binvec,
                             ( symbolIndex.length - 1) * sizeof( float));
            } else {
                tagValues[ ii].value = NULL;
            }
        } else if ( !strcmp( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
            if ( symbolSizeIndex.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( symbolSizeIndex.binvec,
                             ( symbolSizeIndex.length - 1) * sizeof( float));
            } else {
                * ( void **) tagValues[ ii].value = NULL;
            }
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Update a scatter legend when its parent has changed, or if it's
 *      been updated standalone.
 * Arguments:
 *      eventCodes - vector of events that caused the update.
 *      numEventCodes - number of events.
 * Returns:
 *      AmOK on success, error on system error.
 *
 ***************************************************************************/

AmErrorCode 
AmScatterLegendClass::Update( int	*eventCodes,
			      int	numEventCodes)
{
    AmDataGraphicClass *graphic = ( AmDataGraphicClass *) parent;
    float            size, maxSize, scale;
    char             *sympath, relPath[ PATHLEN];
    uchar_t	     *symData;
    int              ii, symWidth, symHeight, minWidth, minHeight,
                     maxWidth, maxHeight, symMaxWidth, symMaxHeight;

/* Free some previously allocated data that gets overwritten in
 *   both the non-editable legend case, and the editable legend
 *   case.
 */

    if ( image != NULL) {
        delete image;
        image = NULL;
    }

    if ( symbols != NULL) {
        for ( ii = 0; ii < numSymbols; ii++) {
            AmFreeSymbol( symbols[ ii]);
        }
        free( symbols);
        symbols = NULL;
        numSymbols = 0;
    }

    if ( !editable && graphic != NULL) {

    /* Free previously allocated data. */

        if ( colors != NULL) {
            free( colors);
            colors = NULL;
        }
        AmDestroyIndexData( &colorIndex);

        if ( symbolList != NULL) {
            free( symbolList);
            symbolList = NULL;
        }
        AmDestroyIndexData( &symbolIndex);

        if ( symbolSizeList != NULL) {
            free( symbolSizeList);
            symbolSizeList = NULL;
        }
        AmDestroyIndexData( &symbolSizeIndex);

        numLabels = 0;

        graphic->GetVaFields( AmNsymbol, &symfile,
                     NULL);

    /* Get color related information. */

        graphic->GetVaFields( AmNcolors, &colors,
                     AmNnumColors, &numColors,
                     AmNcolorsMin, &colorIndex.min,
                     AmNcolorsMax, &colorIndex.max,
                     AmNcolorsBinvec, &colorIndex.binvec,
                     AmNcolorsUnits, &colorIndex.units,
                     AmNcolorsTable, &colorIndex.table,
                     AmNcolorsStrings, &colorIndex.strings,
                     NULL);
        colorIndex.length = numColors;

    /* Get the symbol related information. */

        graphic->GetVaFields( AmNsymbolList, &symbolList,
                     AmNsymbolLength, &symbolIndex.length,
                     AmNsymbolMin, &symbolIndex.min,
                     AmNsymbolMax, &symbolIndex.max,
                     AmNsymbolBinvec, &symbolIndex.binvec,
                     AmNsymbolUnits, &symbolIndex.units,
                     AmNsymbolTable, &symbolIndex.table,
                     AmNsymbolStrings, &symbolIndex.strings,
                     NULL);

    /* Get the symbol size related information. */

        graphic->GetVaFields( AmNsymbolSizeList, &symbolSizeList,
                     AmNsymbolSizeLength, &symbolSizeIndex.length,
                     AmNsymbolSizeMin, &symbolSizeIndex.min,
                     AmNsymbolSizeMax, &symbolSizeIndex.max,
                     AmNsymbolSizeBinvec, &symbolSizeIndex.binvec,
                     AmNsymbolSizeUnits, &symbolSizeIndex.units,
                     AmNsymbolSizeTable, &symbolSizeIndex.table,
                     AmNsymbolSizeStrings, &symbolSizeIndex.strings,
                     NULL);
     }

/* Figure out by which variable(color, symbol type, or symbol size) to
   draw the legend. */

    if ( showSymbolSizes && symbolSizeList != NULL) { 
        showColors = showSymbols = FALSE;
        index = &symbolSizeIndex;
    } else if ( showSymbols && symbolList != NULL) {
        showColors = showSymbolSizes = FALSE;
        index = &symbolIndex;
    } else if ( showColors && colors != NULL) {
        showSymbols = showSymbolSizes = FALSE;
        index = &colorIndex;
    } else {
        showColors = showSymbols = showSymbolSizes = FALSE;
        if ( symbolSizeList != NULL) {
            showSymbolSizes = TRUE;
            index = &symbolSizeIndex; 
        } else if ( symbolList != NULL) {
            showSymbols = TRUE;
            index = &symbolIndex; 
        } else if ( colors != NULL) {
            showColors = TRUE;
            index = &colorIndex; 
        } else {
            index = NULL;
        }
    }

/* Generate labels if needed.  This happens when either the legend
 *   is attached to a parent, or when the number of labels does not
 *   match the number of items to be printed.
 */

    if ( index != NULL) {
        numItems = index->length;
        if ( ( !editable && graphic != NULL) || ( numLabels != numItems)) {
            numLabels = numItems;
            if ( labels != NULL) {
                AmFreeStrVec( labels);
            }
            if ( index->strings == NULL) {
                labels = AmIndexesToStrings( index, numLabels);
            }  else {
                labels = AmCopyStrVec(( const char **) index->strings);
            }
        }
    }

    layoutRects();

/* Find the smallest and largest extends of the legend items. */

    maxWidth = 0;
    maxHeight = 0;
    minWidth = -1;
    minHeight = -1;

    for ( ii = 0; ii < numItems; ii++) {
        if ( ( minWidth == -1) || ( itemRects[ ii].width < minWidth)) {
            minWidth = itemRects[ ii].width;
        }
        if ( itemRects[ ii].width > maxWidth) {
            maxWidth = itemRects[ ii].width;
        }
        if ( ( minHeight == -1) || ( itemRects[ ii].height < minHeight)) {
            minHeight = itemRects[ ii].height;
        }
        if ( itemRects[ ii].height > maxHeight) {
            maxHeight = itemRects[ ii].height;
        }
    }

/* Create an image - size it according to the largest legend item. */

    image = display->CreateImage( maxWidth, maxHeight, display->GetDepth(), 
                         NULL);

    symbols = ( AmSymbolPtr *) malloc( sizeof( AmSymbolPtr) * numItems);
    for ( ii = 0; ii < numItems; ii++) {
        symbols[ ii] = NULL;
    }
    numSymbols = numItems;

/* If the legend is to display symbols, read them in and store them in 
 *   the symbols list.
 */

    if ( showSymbols && symbolList != NULL) {
	symMaxWidth = 1;
	symMaxHeight = 1;
	for ( ii = 0; ii < numItems; ii++) {
	    sprintf( relPath, "%s/%s", AM_SYMBOL_DIR, symbolList[ ii]);
	    sympath = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( sympath != NULL) {
		if ( AmReadBitmapFileData( sympath, &symWidth, &symHeight, 
                         &symData, 1, 0)) {
                    symbols[ ii] = createSymbol( symData, TRUE, symWidth,
                        symHeight, 1.0);
                    if ( symbols[ ii]->width > symMaxWidth) {
                        symMaxWidth = symbols[ ii]->width;
                    }
                    if ( symbols[ ii]->height > symMaxHeight) {
                        symMaxHeight = symbols[ ii]->height;
                    }
                } else {
                    symbols[ ii] = NULL;
		    AN_DEBUG( 1, ( "Failed to load symbol %s", sympath));
		}
	    } else {
                symbols[ ii] = NULL;
		AN_DEBUG( 1, (  "Failed to find symbol %s", symbolList[ ii]));
	    }
	}

    /* Figure out the scaling needed. */

        scale = 1.0;
        if ( symMaxWidth > minWidth) {
            scale = minWidth / ( float) symMaxWidth;
        }

        if (( symMaxHeight * scale) > minHeight) {
            scale = minHeight / ( float) symMaxHeight;
        }

    /* Use the scaling parameters to set the symbol dimensions. */

	for ( ii = 0; ii < numItems; ii++) {
            if ( symbols[ ii] != NULL) {
                symbols[ ii]->scale = scale;
            }
        }

    } else if ( showSymbolSizes && symbolSizeList != NULL) {

    /* We are not showing different symbols, so we need to read
       a default symbol and then scale it. */

        maxSize = 0.0;
        for ( ii = 0; ii < numItems; ii++) {
            size = symbolSizeList[ ii];
            if ( size > maxSize) {
                maxSize = size;
            }
        }

	sprintf( relPath, "%s/%s", AM_SYMBOL_DIR, symfile);
	sympath = AmGetEnvFile( CONFIG_ENV, relPath);
        if ( sympath != NULL) {
	    if ( AmReadBitmapFileData( sympath, &symWidth, 
                     &symHeight, &symData, 1, 0)) {

                for ( ii = 0; ii < numItems; ii++) {
                    if ( ii == 0) {
                        symbols[ ii] = createSymbol( symData, TRUE, symWidth,
                                            symHeight, 1.0);
                    } else {
                        symbols[ ii] = createSymbol( symData, FALSE, symWidth,
                                            symHeight, 1.0);
                    }
                }
	    } else {
		AN_DEBUG( 1, ( "Failed to load symbol %s", sympath));
	    }
	} else {
	    AN_DEBUG( 1, ( "Failed to find symbol %s", symfile));
	}


    /* Figure out the scaling needed. */

        symMaxWidth = (int)(maxSize * symWidth);
        symMaxHeight = (int)(maxSize * symHeight);

        scale = 1.0;
        if ( symMaxWidth > minWidth) {
            scale = minWidth / ( float) symMaxWidth;
        }

        if (( symMaxHeight * scale) > minHeight) {
            scale = minHeight / ( float) symMaxHeight;
        }

    /* Use the scaling parameters to set the symbol dimensions. */

	for ( ii = 0; ii < numItems; ii++) {
            if ( symbols[ ii] != NULL) {
                symbols[ ii]->scale = symbolSizeList[ ii] * scale;
            }
        }

    } else {

    /* We are not showing different symbols, so we need to read
       a default symbol. */

	sprintf( relPath, "%s/%s", AM_SYMBOL_DIR, symfile);
	sympath = AmGetEnvFile( CONFIG_ENV, relPath);
	if ( sympath != NULL) {
	    if ( AmReadBitmapFileData( sympath, &symWidth, 
                                     &symHeight, &symData, 1, 0)) {
                for ( ii = 0; ii < numItems; ii++) {
                    if ( ii == 0) {
                        symbols[ ii] = createSymbol( symData, TRUE, symWidth,
                                            symHeight, 1.0);
                    } else {
                        symbols[ ii] = createSymbol( symData, FALSE, symWidth,
                                            symHeight, 1.0);
                    }
                }
	    } else {
		AN_DEBUG( 1, ( "Failed to load symbol %s", sympath));
	    }
	} else {
	    AN_DEBUG( 1, (  "Failed to find symbol %s", symfile));
	}

    /* Figure out the scaling needed. */

        scale = 1.0;
        if ( symWidth > minWidth) {
            scale = minWidth / ( float) symWidth;
        }

        if (( symHeight * scale) > minHeight) {
            scale = minHeight / ( float) symHeight;
        }

    /* Use the scaling parameters to set the symbol dimensions. */

	for ( ii = 0; ii < numItems; ii++) {
            if ( symbols[ ii] != NULL) {
                symbols[ ii]->scale = scale;
            }
        }
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Draw a single legend item for a scatter legend.  The Draw()
 *      function is implemented in AmLegendClass, and calls DrawItems()
 *      member function as a virtual, which calls this.
 * Arguments:
 *      pWindow - window to draw the item to.
 *      pGC - GC to draw the item with.
 *      itemNumber - index of the item.
 *      pItemRect - item's screen rectangle.
 *
 *************************************************************************/

AmErrorCode
AmScatterLegendClass::DrawItem( AmWindowClass        *w,
                                AmGCClass            *gc,
                                int                  item,
                                AmRectangleRec const *drawRect)
{
    AmRectangleRec itemRect;
    AmColorCellPtr color;
    ulong_t      *data, pixel;
    int          width, height;
    uchar_t	 *p = ( uchar_t *) &pixel;
    float	 dpiScale = display->GetDPI() / BASEDPI;

/* convert the color cell to an unsigned long pixel for AmScalePix */

    if ( showColors) {
        color = &colors[ item];
    } else {
        color = &fgColor;
    }
    ASSIGN_PIXEL( *color, p);

    data = ( ulong_t *) image->AmGetData();
    width = image->GetNumHorizPixels();
    height = image->GetNumVerticalPixels();

    itemRect.x = 0;
    itemRect.y = 0;
    itemRect.width = image->GetWidth();
    itemRect.height = image->GetHeight();

/* clear out the background image, scale the pixel and draw the image */

    image->ClearImageRect( &bgColor, &itemRect);
    AmScalePix( data, width, height,
        symbols[ item]->data, symbols[ item]->width, symbols[ item]->height,
        pixel, drawRect->width >> 1,  drawRect->height >> 1,
        symbols[ item]->scale*dpiScale);
    display->DrawImage( w, gc, 0, 0, drawRect->x, drawRect->y,
                 drawRect->width, drawRect->height, image);

    return( AmOK);
} 

/*************************************************************************
 *
 * Description:
 *      Draw all the legend items. Called by AmLegendClass::DrawItems()
 *      as a virtual function.
 * Arguments:
 *      pWindow - window to draw to.
 *      pGC - GC to draw with.
 *      vItemRect - vector of item rectangles.
 * Returns:
 *      AmOK.
 *
 *************************************************************************/

AmErrorCode
AmScatterLegendClass::DrawItems( AmWindowClass        *w,
                               AmGCClass            *gc,
                               AmRectangleRec const *itemRects)
{
    int          ii;
 
    for ( ii = 0; ii < numItems; ii++) {
        DrawItem( w, gc,
            ( increaseItems == TRUE) ? numItems - 1 - ii : ii,
            &itemRects[ ii]);
    }
 
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a scatter legend.  This is
 *      used in the map editor to dump the scatter legend's fields into
 *      a file.  the field name (and length, if it's a vector field) is
 *      written into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmScatterLegendClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				  AnvilBoolean	      allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmLegendClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmScatterLegendClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmScatterLegendClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmScatterLegendClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/


void 
AmScatterLegendClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					   int			*numDefs)
{
    static AmFieldDefRec ScatterLegendClassFieldDefinitions[] = {
	{ AmNshowColors, offsetof( AmScatterLegendClass, showColors) },
	{ AmNshowSymbolSizes, offsetof( AmScatterLegendClass, showSymbolSizes) },
	{ AmNshowSymbols, offsetof( AmScatterLegendClass, showSymbols) },
        { AmNsymbol, offsetof( AmScatterLegendClass, symfile), 0, AmDYNAMIC },
        { AmNsymbolLength, offsetof( AmScatterLegendClass, symbolIndex.length) },
        { AmNsymbolList, offsetof( AmScatterLegendClass, symbolList),
              offsetof( AmScatterLegendClass, symbolIndex.length),
              AmDYNAMIC },
        { AmNsymbolMax, offsetof( AmScatterLegendClass, symbolIndex.max) },
        { AmNsymbolMin, offsetof( AmScatterLegendClass, symbolIndex.min) },
        { AmNsymbolSizeLength,
              offsetof( AmScatterLegendClass, symbolSizeIndex.length) },
        { AmNsymbolSizeList,
              offsetof( AmScatterLegendClass, symbolSizeList),
              offsetof( AmScatterLegendClass, symbolSizeIndex.length),
              AmDYNAMIC },
        { AmNsymbolSizeMax,
              offsetof( AmScatterLegendClass, symbolSizeIndex.max) },
        { AmNsymbolSizeMin,
          offsetof( AmScatterLegendClass, symbolSizeIndex.min) },
        { AmNsymbolSizeStrings,
              offsetof( AmScatterLegendClass, symbolSizeIndex.strings),
              offsetof( AmScatterLegendClass, symbolSizeIndex.length),
              AmDYNAMIC },
        { AmNsymbolSizeTable,
              offsetof( AmScatterLegendClass, symbolSizeIndex.table),
              offsetof( AmScatterLegendClass, symbolSizeIndex.length),
              AmDYNAMIC },
        { AmNsymbolSizeUnits,
              offsetof( AmScatterLegendClass, symbolSizeIndex.units) },
        { AmNsymbolStrings, offsetof( AmScatterLegendClass, symbolIndex.strings),
              offsetof( AmScatterLegendClass, symbolIndex.length),
              AmDYNAMIC },
        { AmNsymbolTable, offsetof( AmScatterLegendClass, symbolIndex.table),
              offsetof( AmScatterLegendClass, symbolIndex.length),
              AmDYNAMIC },
        { AmNsymbolUnits, offsetof( AmScatterLegendClass, symbolIndex.units) },

    };
    *fieldDefs = ScatterLegendClassFieldDefinitions;
    *numDefs = VECLEN( ScatterLegendClassFieldDefinitions);
}
