#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* single line text annotation class member function definitions.  Text
   Annotations have a private GC to speed the setting of font attributes */ 

#include <malloc.h>
#include <string.h>
#include <stddef.h>
#include <anvil/tagstrs.h>
#include <anvil/rect.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/userevents.h>
#include <anvil/textannot.h>
#include <anvil/resource.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmTextAnnotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmTEXT_ANNOTATION) {
        return( TRUE);
    } else {
        return( AmAnnotationClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so text annotations can be subclassed.
 *
 ****************************************************************************/

AmTextAnnotClass::AmTextAnnotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Destructor: free the text annotation's string and private graphics
 *	context.
 *
 ****************************************************************************/

AmTextAnnotClass::~AmTextAnnotClass()
{
    if ( label != NULL) {
	free( label);
    }
    if ( font != NULL) {
	free( font);
    }
    delete textGC;
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a text annotation.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify object for resources
 *
 ****************************************************************************/

void
AmTextAnnotClass::initializeFields( AmWindowClass  *windowArg,
                                  const char   *nameArg)
{
    int width, ascent, descent;

    AmAnnotationClass::initializeFields( windowArg, nameArg);
    objectType = AmTEXT_ANNOTATION;

    opaque  = FALSE;
    _fFrame = FALSE;

    label = strdup( AmGetDefault( "textannot.default", 
				  "This is a test string."));
    labelAngle = 0;
    font = strdup( DEFAULT_FONT);
    fontSize = DEFAULT_SIZE;
    fontWeight = DEFAULT_WEIGHT;
    fontSlant = DEFAULT_SLANT;
    fontXScale = 1.0;
    fontYScale = 1.0;

/* a local gc is used for text annotation, because setting fonts is very
   expensive */

    textGC = window->CreateGC();
    textGC->SetFont( font, fontSize, fontWeight, fontSlant,
		     fontXScale, fontYScale);
    if ( label != NULL) {
        display->ExtentsText( textGC, label, &width, &ascent, &descent);
        rect.width = width;
        rect.height = ascent + descent;
        extentsRect = rect;
    }
}

/****************************************************************************
 *
 * Description:
 *      AmTextAnnotClass constructor.  Creates a text annotation using the 
 *	window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in the 
 *	text annotation
 * Arguments:
 *      windowArg - window that the text annotation will be drawn on.
 *      nameArg - string name of the text annotation, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmTextAnnotClass::AmTextAnnotClass( AmWindowClass  *windowArg,
                                const char   *nameArg,
                                ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		numValues;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    numValues = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    setFieldsWrapper( tagValues, numValues);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a text annotation, given the tag-values stored in
 *      a tag values array.  First, the annotation fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmTextAnnotClass::setFields( AmTagValuePtr tagValues,
			   int         numValues)
{
    int         numFieldDefs, width, ascent, descent,
                tempWidth, tempHeight;
    AnvilBoolean     widthChanged, heightChanged;
    AmFieldDefPtr fieldDefs;

    widthChanged = FALSE;
    heightChanged = FALSE;

    numValues = AmAnnotationClass::setFields( tagValues, numValues);
    AmTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
					tagValues, numValues);

/* Reset the extents rectangle to reflect the new text and or font
   if it's been changed. */

    tempWidth = rect.width;
    tempHeight = rect.height;

    if ( AmChanged( &rect.width)) {
        widthChanged = TRUE;
    }

    if ( AmChanged( &rect.height)) {
        heightChanged = TRUE;
    }

    if ( AmChanged( &label, &font, &fontSize, &fontWeight, &fontSlant,
		  &fontXScale, &fontYScale, NULL)) {

        textGC->SetFont( font, fontSize, fontWeight, fontSlant,
			 fontXScale, fontYScale);

        if ( label != NULL) {
            display->ExtentsText( textGC, label, &width, &ascent, &descent);
            rect.width = width;
            rect.height = ascent + descent;
            extentsRect = rect;
        }
    }

/*
 * If the rectangle has been specifically set in the set values call 
 *   then we try to scale the font to fit it as closely as possible.
 *   This helps aleviate the problem of fonts not being 100% compatible
 *   across output devices.
 */

    if ( widthChanged || heightChanged) {
        SetRect( rect.x, rect.y, tempWidth, tempHeight);
    }

    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	set the rectangle for a text annotation, and set the font scaling
 *	to reflect the new rectangle.
 * Arguments:
 *	x, y, width, height - rectangle dimensions.
 * 
 ****************************************************************************/

void 
AmTextAnnotClass::SetRect( int x,
			 int y,
			 int width,
			 int height)
{
    int ascent, descent;

    if (( width == 0) || ( height == 0)) {
	return;
    }
    rect.x = x;
    rect.y = y;

    fontXScale = ( float)width / ( float)rect.width * fontXScale;
    fontYScale = ( float)height / ( float)rect.height * fontYScale;

    textGC->SetFont( font, fontSize, fontWeight, fontSlant,
		     fontXScale, fontYScale);

    if ( label != NULL) {
        display->ExtentsText( textGC, label, &width, &ascent, &descent);
        rect.width = width;
        rect.height = ascent + descent;
    } else {
        rect.width = width;
        rect.height = height;
    }
    extentsRect = rect;
    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}
 
/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a text annotation, given the tag-values 
 *	stored in a tag values array.  First, the annotation fields are 
 *	retrieved, then the other fields are retrieved using 
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmTextAnnotClass::getFields( AmTagValuePtr tagValues,
			   int         numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmAnnotationClass::getFields( tagValues, numValues);
    AmTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Draw the text annotation to the screen.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK
 *
 ****************************************************************************/

AmErrorCode 
AmTextAnnotClass::draw( AmWindowClass	      	*w,
		      AmGCClass	      		*gc,
		      AmRectangleRec const  	*exposeRects,
		      int		      	nRects)
{
    const AmRectangleRec *drawRect = GetRect();
    int                width, ascent, descent, nClipRects;
    AmRectanglePtr       clipRects;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( drawRect)) {
        return( AmOK);
    }

/* copy the clip rectangles from the display manager to the private GC */
    
    gc->GetClipRectangles( &clipRects, &nClipRects);
    textGC->SetClipRectangles( clipRects, nClipRects);
    free( clipRects);

/* clear the rectangle before drawing the legend.  Use Fill Rectangle, because
   ClearRect isn't sensitive to clip rectangles */

    if ( opaque == TRUE) {
	textGC->SetForegroundColor( &bgColor);
        display->FillRectangle( w, textGC, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }

/* set the foreground and background colors */

    textGC->SetForegroundColor( &fgColor);
    textGC->SetBackgroundColor( &bgColor);

/* find the extents and draw the text */

    display->ExtentsText( textGC, label, &width, &ascent, &descent);
    if ( labelAngle == 0) {
        display->DrawText( w, textGC, drawRect->x, drawRect->y + ascent, label);
    } else {
        display->DrawRotatedText( w, textGC, drawRect->x, drawRect->y,
				  DEGTORAD( labelAngle), label);
    }
    return( AmOK);
}


/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a text annotation.  This is
 *      used in the map editor to dump the annotation's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmTextAnnotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			    AnvilBoolean	      allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( legend) */

    numStrings = AmAnnotationClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmTextAnnotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a text annotation . This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmImagePlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  THIS TABLE MUST BE SORTED
 *      IN ALPHABETICAL ORDER ON THE FIELD NAME FOR SEARCHING 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmTextAnnotClass::GetFieldDefinitions( AmFieldDefPtr  *fieldDefs,
				     int          *numDefs)
{
    static AmFieldDefRec TextAnnotClassFieldDefinitions[] = {
	{ AmNfont, offsetof( AmTextAnnotClass, font), 0, AmDYNAMIC },
        { AmNfontSize, offsetof( AmTextAnnotClass, fontSize) },
        { AmNfontSlant, offsetof( AmTextAnnotClass, fontSlant) },
        { AmNfontWeight, offsetof( AmTextAnnotClass, fontWeight) },
        { AmNfontXScale, offsetof( AmTextAnnotClass, fontXScale) },
        { AmNfontYScale, offsetof( AmTextAnnotClass, fontYScale) },
	{ AmNlabel, offsetof( AmTextAnnotClass, label), 0, AmDYNAMIC },
	{ AmNlabelAngle, offsetof( AmTextAnnotClass, labelAngle) },
    };
    *fieldDefs = TextAnnotClassFieldDefinitions;
    *numDefs = VECLEN( TextAnnotClassFieldDefinitions);
}
