#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* AmTickClass member functions: Tickmarks are used to show the coordinate
   system of a parent object along a single axis.
   Modifications:
	AmTickClass needs to be expanded to support map projections, and
	ticks for string/thematic values, for bargraphs and the like.
*/

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/ratio.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/scale.h>
#include <anvil/datagraphic.h>
#include <anvil/tick.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmTickClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmTICKS) {
        return( TRUE);
    } else {
        return( AmScaleClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for tickmarks.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.  The tickmarks
 *      calls the AmScaleClass initializer, then initializes the the 
 *	tickmark-specific fields, such as tick size, style and 
 *	spacing.
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify plot for resources
 * Modifications:  
 *	the size values should be consts in a source file, not hardcoded
 *
 ****************************************************************************/

void 
AmTickClass::initializeFields( AmWindowClass   *windowArg,
			     AmCoordClass    *graphicArg,
			     const char	   *nameArg)
{
    AmScaleClass::initializeFields( windowArg, graphicArg, nameArg);
    objectType = AmTICKS;
    tickPos = AmLEFT;
    offset = 0;
    style = AmBASE_TEN;
    spacing = 1.0;
    base = 0.0;
    size = 5;
    outside = TRUE;
    maxTicks = 10;
    numTicks = 5;

    opaque  = FALSE;
    _fFrame = FALSE;
}

/****************************************************************************
 *
 * Description:
 *      Default constructor, so tickmarks can be subclassed.
 *
 ****************************************************************************/

AmTickClass::AmTickClass()
{
}

/****************************************************************************
 *
 * Description:
 *      destructor: nothing has to be freed which is specific to AmTickClass
 *
 ****************************************************************************/

AmTickClass::~AmTickClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmTickClass constructor.  Creates a grid using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the tickmarks.
 * Arguments:
 *      windowArg - window that the tickmarks will be drawn on.
 *      nameArg - string name of the tickmarks, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmTickClass::AmTickClass( AmWindowClass	*windowArg,
		      AmCoordClass  	*graphicArg, 
		      const char	*nameArg,
		      ...)
{
    
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		count;

    initializeFields( windowArg, graphicArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    setFields( tagValues, count);
    va_end( args);
    SetValid( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for tickmarks given the tag-values stored in
 *      a tag values array.  First, the AmScaleClass fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 * Modifications:
 *	that AmTrackChanges() call shouldn't be in there.
 *
 ****************************************************************************/

int 
AmTickClass::setFields( AmTagValuePtr	tagValues,
		      int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    AmTrackChanges();
    numValues = AmScaleClass::setFields( tagValues, numValues);
    AmTickClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( SetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for tickmarks, given the tag-values stored in
 *      a tag values array.  First, the coordinates fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmTickClass::getFields( AmTagValuePtr	tagValues,
		      int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmScaleClass::getFields( tagValues, numValues);
    AmTickClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 * 	given the tick and text orientation, return the alignment 
 *	enumeration for the text strings.
 * Arguments:
 *	tickPos - tick left,right,top, bottom orientation.
 *	vertical - whether text is purpendicular relative to the tickmark
 *	orientation.
 *
 ****************************************************************************/

static AmTextAlignEnum 
tickTextAlign( AmTickPosEnum	tickPos,
	       AnvilBoolean		vertical)
{
    if ( tickPos == AmLEFT) {
	return(( vertical) ? AmCENTER_UP_VERTICAL : AmRIGHT_HORIZONTAL);
    } else if ( tickPos == AmRIGHT) {
	return(( vertical) ? AmCENTER_DOWN_VERTICAL : AmLEFT_HORIZONTAL);
    } else if ( tickPos == AmTOP) {
	return(( vertical) ? AmLEFT_UP_VERTICAL : AmCENTER_HORIZONTAL);
    } else {
	return(( vertical) ? AmLEFT_DOWN_VERTICAL : AmCENTER_HORIZONTAL);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Draw a set of tickmarks. Tickmarks have two private graphics
 *	contexts, textGC and titleGC, which are used for the text and
 *	title fonts, mostly to improve font loading performance in X
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *	nRects - number of exposed rectangles
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmTickClass::draw( AmWindowClass		*w,
		 AmGCClass		*gc,
		 AmRectangleRec const	*exposeRects,
		 int			nRects)	
{
    AmCoordClass		*graphic = ( AmCoordClass *) parent;
    int			ii, localLineWidth, scrX, scrY, sgn,
			x0, y0, x1, y1, width, ascent, descent, actualOffset,
			nClipRects;
    float		minVal, maxVal, x, y, 
			currXmin, currYmin, currXmax, currYmax;
    char		*unitsTag, valbuf[ TEXTLEN];
    unitType		units;
    AmTextAlignEnum	textAlign, titleAlign;
    AmRectanglePtr	clipRects;
    AmRectangleRec const	*drawRect = GetRect();

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( w, gc, exposeRects, nRects));
    }

    localLineWidth = (int)MAX( _dLineWidth, 1);
AN_DEBUG( 90, ("drawRect = (%d, %d, %d, %d)\n", 
	       drawRect->x, drawRect->y,
	       drawRect->width, drawRect->height));

/* set the clip rectangles for the text and title GC's which are copied from
   the GC passed down from the display manager */

    gc->GetClipRectangles( &clipRects, &nClipRects);
    textGC->SetClipRectangles( clipRects, nClipRects);
    titleGC->SetClipRectangles( clipRects, nClipRects);
    textGC->SetForegroundColor( &fgColor);
    titleGC->SetForegroundColor( &fgColor);
    free( clipRects);

/* clear the rectangle before drawing the tickmarks, and get the line width 
   Use fillrectangle, because clear rect isn't sensitive to gc clip regions */

    if ( opaque == TRUE) {
	gc->SetForegroundColor( &bgColor);
	display->FillRectangle( w, gc, drawRect->x, drawRect->y,
				drawRect->width, drawRect->height);
    }

/* restore the GC's foreground color */

    gc->SetForegroundColor( &fgColor);

/* if the display is base ten, then use that to calculate the base and 
   spacing.  Modify the number of ticks to make sure that it sure that a
   reasonable number are displayed */

    graphic->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);

/* set the min, max, variable, and units depending on the tick orientation.
   find the start and end of the tick base line */

    if (( tickPos == AmLEFT) || ( tickPos == AmRIGHT)) {
	graphic->GetDataPt( rect.x, rect.y + rect.height - 1, &x, &minVal);
	graphic->GetDataPt( rect.x, rect.y, &x, &maxVal);
	unitsTag = AmNyunits;
	x0 = ( tickPos == AmLEFT) ? rect.x + rect.width  - 1: rect.x;
	x1 = x0;
	y0 = rect.y;
	y1 = rect.y + rect.height - 1;
    } else {
	graphic->GetDataPt( rect.x, rect.y, &minVal, &y);
	graphic->GetDataPt( rect.x + rect.width - 1, rect.y, &maxVal, &y);
	unitsTag = AmNxunits;
	x0 = rect.x;
	x1 = rect.x + rect.width - 1;
	y0 = ( tickPos == AmTOP) ? rect.y + rect.height - 1: rect.y;
	y1 = y0;
    }
    graphic->GetVaFields( unitsTag, &units, 
			  NULL);

/* don't bother drawing if the max is less than the min */

    if ( maxVal <= minVal) {
	return( AmOK);
    }
    if ( style == AmBASE_TEN) {	
	AmScaleBaseTenSpacing( minVal, maxVal, maxTicks, &spacing, &base);
    } else if ( style == AmEND_END) {
	spacing = (maxVal - minVal)/numTicks;
	base = minVal;
    }

/* get the text and title alignments */

    textAlign = tickTextAlign( tickPos, textVertical);
    titleAlign = tickTextAlign( tickPos, titleVertical);

/* dominant choices are vertical and horizontal */

    if (( tickPos == AmLEFT) || ( tickPos == AmRIGHT)) {

/* get x position and tick orientation depending on left or right */

	x = ( tickPos == AmLEFT) ? currXmin : currXmax;
	sgn = ( tickPos == AmLEFT) ? -1 : 1;
	for ( ii = 0, y = base; y <= currYmax; ii++, y += spacing) {

/* find where the tick value is on the screen, and draw the tickmark with
   appropriately aligned text */

	    graphic->GetScreenPt( x, y, &scrX, &scrY);
	    if (( scrY > rect.y) && ( scrY < rect.y + rect.height)) {
		display->DrawLine( w, textGC, x0, scrY, x0 + sgn*size, scrY);
		AmPrintValue( units, y, valbuf);
		display->DrawAlignedText( w, textGC, 
					  x0 + sgn*(size + textOffset), scrY,
					  valbuf, textAlign);
	    }
	}

/* draw the baseline and the title */

/*
 * Incomplete:
 *     Because of the localLineWidth here the ticks are off by one in each
 *     direction.
 *
 *     We need to come up with a concrete definition of what line width
 *     means.
 *
 *     -Phil
 */

	display->DrawLine( w, textGC, x0 + sgn*localLineWidth, y0, 
				      x0 + sgn*localLineWidth, y1);
	if (( title != NULL) && ( title[ 0] != '\0')) {
	    display->DrawAlignedText( w, titleGC,
				      x0 + sgn*titleOffset, (y0 + y1)/2,
				      title, titleAlign);
	}
    } else if (( tickPos == AmTOP) || ( tickPos == AmBOTTOM)) {

/* same thing, but for horizontal tickmarks. If the tickmarks are on the 
   bottom, then fold the font size into the text spacing */

	y = ( tickPos == AmLEFT) ? currYmin : currYmax;
	sgn = ( tickPos == AmTOP) ? -1 : 1;
	if (( tickPos == AmBOTTOM) && !textVertical) {
	    display->ExtentsText( textGC, "TEST", &width, &ascent, &descent);
	    actualOffset = textOffset + ascent + descent;
	} else {
	    actualOffset = textOffset;
	}
	for ( ii = 0, x = base; x <= currXmax; ii++, x += spacing) {
	    graphic->GetScreenPt( x, y, &scrX, &scrY);
	    if (( scrX >= rect.x) && ( scrX <= rect.x + rect.width)) {
		display->DrawLine( w, textGC, scrX, y0, scrX, y0 + sgn*size);
		AmPrintValue( units, x, valbuf);
		display->DrawAlignedText( w, textGC, 
					  scrX, y0 + sgn*(size + actualOffset), 
					  valbuf, textAlign);
	    }
	}
	display->DrawLine( w, textGC, x0, y0 + sgn*localLineWidth, 
				      x1, y0 + sgn*localLineWidth);
	if (( title != NULL) && ( title[ 0] != '\0')) {
	    display->DrawAlignedText( w, titleGC,
				      ( x0 + x1)/2, y0 + sgn*titleOffset,
				      title, titleAlign);
	}
    } 
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in AmTickClass.  This is
 *      used in the map editor to dump the ticks's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmTickClass::dumpFields( AmFieldLengthPtr    fieldLengths,
		       AnvilBoolean		 allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( scale) */

    numStrings = AmScaleClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmTickClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a AmTickClass. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmTickClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  NOTE THAT THIS TABLE MUST
 *      BE SORTED ON THE TAG NAME FOR SEARCHING.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmTickClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
			        int		*numDefs)
{
    static AmFieldDefRec TickClassFieldDefinitions[] = {
	{ AmNmaxTicks, offsetof( AmTickClass, maxTicks) },
	{ AmNnumTicks, offsetof( AmTickClass, numTicks) },
	{ AmNtickBase, offsetof( AmTickClass, base) },
	{ AmNtickOutside, offsetof( AmTickClass, outside) },
	{ AmNtickPosition, offsetof( AmTickClass, tickPos) },
	{ AmNtickSpacing, offsetof( AmTickClass, spacing) },
	{ AmNtickStyle, offsetof( AmTickClass, style), },
    };
    *fieldDefs = TickClassFieldDefinitions;
    *numDefs = VECLEN( TickClassFieldDefinitions);
}
