#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <windows.h>
#include <math.h>
#include <anvil/project.h>

#if defined( _WIN32)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <anvil/ntdisplay.h>
#include <anvil/errcode.h>
#include <anvil/ntcolor.h>
#include <anvil/macros.h>
#include <anvil/dispatch.h>

#define SCREEN_COORD( iVirtual) ( iVirtual * _iDPI / _iDPIBase)
#define VIRTUAL_COORD( iScreen) ( iScreen * _iDPIBase / _iDPI)

/******************************************************************** 
 *                                                                     
 * AmNTDisplayClass:                                                     
 *                                                                     
 ********************************************************************/

/* NTDisplay (Microsoft Windows NT Display) do nothing constructor. */

AmNTDisplayClass::AmNTDisplayClass()
    : AmDisplayClass()
{
}

/* NTDisplay (Microsoft Windows NT Display) destructor. */

AmNTDisplayClass::~AmNTDisplayClass()
{
}

/* Retrieve the depth of the display. */

int
AmNTDisplayClass::GetDepth()
{

    /* The display's depth is the same as the applications's depth. */

    return( AmGetAppDepth());
}


/*
 * Function: void DrawLines( pWindow, pGC, pPoint, nPoint)
 *
 * Parameters:
 *   AmWindowClass *pWindow : The window to draw on. 
 *   AmGCClass     *pGC     : How to draw the lines (color, etc...)
 *   Point       *pPoint  : The end points of the lines.
 *   int         nPoint   : The number of points passed in.  This value
 *                          should be number of lines + 1. 
 * Description:
 *   DrawLines draws a set of non-distinct (joined) lines on the
 *   window using the passed in GC for the line style, color, width,
 *   clipping, etc ...
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::DrawLines( AmWindowClass *pWindow,
                           AmGCClass     *pGC,
                           Point       *pPoint,
                           int         nPoint)
{
    AmNTGCClass     *pNTGC = ( AmNTGCClass *) pGC;
    HDC           hDC;
    POINT         *pPOINT;
    int           ii;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    /* Create storage for NT points on the stack. */

    pPOINT = ( POINT *) alloca( sizeof( POINT) * nPoint);
    if ( pPOINT == NULL) {
        DO_ERROR( AmMALLOC_FAILED);
        return;
    }

    /* Convert from Anvil points structure to NT windows points. */

    for ( ii = 0; ii < nPoint; ii++) { 
        pPOINT[ ii].x = SCREEN_COORD( pPoint[ ii].x);
        pPOINT[ ii].y = SCREEN_COORD( pPoint[ ii].y);
    }

    /* Draw the lines. */

    MoveToEx( hDC, pPOINT[ 0].x, pPOINT[ 0].y, NULL);
    PolylineTo( hDC, &pPOINT[ 1], nPoint - 1);
}


/*
 * Function: void DrawSegments( pWindow, pGC, pSegment, nSegment)
 *
 * Parameters:
 *   AmWindowClass *pWindow  : The window to draw on. 
 *   AmGCClass     *pGC      : How to draw the segments (color, etc...)
 *   AmSegmentRec  *pSegment : The segments.
 *   int         nSegment  : The number of segments passed in.
 *
 * Description:
 *   DrawSegments draws a set of distinct (non-joined) lines, also
 *   refered to as segments.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::DrawSegments( AmWindowClass *pWindow,
                              AmGCClass     *pGC,
                              AmSegmentRec  *pSegment,
                              int         nSegment) 
{
    AmNTGCClass *pNTGC = ( AmNTGCClass *) pGC;
    HDC       hDC;
    POINT     *pPOINT;
    ulong_t   *plVertix;
    int       nPOINT, ii;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    /*
     * Allocate storage on the stack for the NT points and the associated 
     *   vertix counts.
     */

    nPOINT = nSegment * 2;
    pPOINT = ( POINT *) alloca( sizeof( POINT) * nPOINT * 2);
    if ( pPOINT == NULL) {
        DO_ERROR( AmMALLOC_FAILED);
        return;
    }

    plVertix = ( ulong_t *) alloca( sizeof( ulong_t) * nSegment);
    if ( plVertix == NULL) {
        DO_ERROR( AmMALLOC_FAILED);
        return;
    }

    /*
     * Convert from Anvil segment structure to NT point structure
     *   and vertix counts.
     */

    for ( ii = 0; ii < nSegment; ii++) { 
        pPOINT[ ii * 2].x = SCREEN_COORD( pSegment[ ii].x0);
        pPOINT[ ii * 2].y = SCREEN_COORD( pSegment[ ii].y0);
        pPOINT[ ii * 2 + 1].x = SCREEN_COORD( pSegment[ ii].x1);
        pPOINT[ ii * 2 + 1].y = SCREEN_COORD( pSegment[ ii].y1);
        plVertix[ ii] = 2;
    }

    /* Draw the segments. */
    
    PolyPolyline( hDC, pPOINT, plVertix, nSegment);
}


/*
 * Function: void DrawPoints( pWindow, pGC, pPoint, nPoint)
 *
 * Parameters:
 *   AmWindowClass *pWindow : The window to draw on. 
 *   AmGCClass     *pGC     : How to draw the points (color, etc...)
 *   Point       *pPoint  : The points.
 *   int         nPoint   : The number of points passed in.
 *
 * Description:
 *   Draw a set of points in the foreground color.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::DrawPoints( AmWindowClass *pWindow,
                            AmGCClass     *pGC,
                            Point       *pPoint,
                            int         nPoint)
{
    AmNTGCClass    *pNTGC = ( AmNTGCClass *) pGC;
    HDC          hDC;
    AmColorCellRec col;
    int          ii;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    /* Get the color. */

    pGC->GetForegroundColor( &col);

    /* Draw the points. */

    for ( ii = 0; ii < nPoint; ii++) {
        SetPixel( hDC, SCREEN_COORD( pPoint[ ii].x),
                  SCREEN_COORD( pPoint[ ii].y), col.pixel);
    }
}


/*
 * Function: void DrawArc( pWindow, pGC, x, y, cx, cy, dAngleStart, dAngelEnd)
 *
 * Parameters:
 *   AmWindowClass *pWindow    : The window to draw on. 
 *   AmGCClass     *pGC        : How to draw the arc (color, etc...)
 *   int         x           : The upper left x coordinate of the bounding
 *                             rectangle.
 *   int         y           : The upper left y coordinate of the bounding
 *                             rectangle
 *   int         cx          : The width of the bounding rectangle. 
 *   int         cy          : The height of the bounding rectangle.
 *   float       dAngleStart : The start position given in radians.
 *   float       dAngleEnd   : The end position given in radians.
 *
 * Description:
 *   Draw an elliptical arc on the screen starting at the passed in
 *   angle dAngleStart to the angle dAngleEnd.  The ellipse fits into 
 *   the rectangle which has the upper left corner at <x,y> and which 
 *   has a width of cx and has a height of cy.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::DrawArc( AmWindowClass *pWindow,
                         AmGCClass     *pGC,
                         int         x,
                         int         y,
                         int         cx,
                         int         cy,
                         float       dAngleStart,
                         float       dAngleEnd)
{
    AmNTGCClass *pNTGC = ( AmNTGCClass *) pGC;
    HDC       hDC;
    double    dSin, dCos, dLength;
    int       x0, y0, x1, y1;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);
    cx = SCREEN_COORD( cx);
    cy = SCREEN_COORD( cy);

    /* 
     * Figure out the intersection of the arc with the surrounding
     *   rectangle.  <x0, y0> is the line drawn from the center of
     *   the arc to the point intersecting the arc at the end.
     *   <x1, y1> is the line drawn from the center of the arc to
     *   the point intersecting the arc at the beginning.
     */

    dSin = sin( dAngleEnd);
    dCos = cos( dAngleEnd);

    dLength = ( cx > cy ) ? cx / 2.0 : cy / 2.0;

    x0 = ( int)(( x + cx) / 2.0 + dLength * dCos + 0.5);
    y0 = ( int)(( y + cy) / 2.0 - dLength * dSin + 0.5);

    dSin = sin( dAngleStart);
    dCos = cos( dAngleStart);

    x1 = ( int)(( x + cx) / 2.0 + dLength * dCos + 0.5);
    y1 = ( int)(( y + cy) / 2.0 - dLength * dSin + 0.5);

    /* Draw the arc. */

    Arc( hDC, x, y, x + cx + 1, y + cy + 1, x0, y0, x1, y1);
}


/*
 * Function: void FillArc( pWindow, pGC, x, y, cx, cy, dAngleStart, dAngelEnd)
 *
 * Parameters:
 *   AmWindowClass *pWindow    : The window to draw on. 
 *   AmGCClass     *pGC        : How to draw the arc (color, etc...)
 *   int         x           : The upper left x coordinate of the bounding
 *                             rectangle.
 *   int         y           : The upper left y coordinate of the bounding
 *                             rectangle
 *   int         cx          : The width of the bounding rectangle. 
 *   int         cy          : The height of the bounding rectangle.
 *   float       dAngleStart : The start position given in radians.
 *   float       dAngleEnd   : The end position given in radians.
 *
 * Description:
 *   Draw an elliptical filled arc on the screen starting at the passed
 *   in angle dAngleStart to the angle dAngleEnd.  The ellipse fits into 
 *   the rectangle which has the upper left corner at <x,y> and which 
 *   has a width of cx and has a height of cy.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::FillArc( AmWindowClass *pWindow,
                         AmGCClass     *pGC,
                         int         x,
                         int         y,
                         int         cx,
                         int         cy,
                         float       dAngleStart,
                         float       dAngleEnd)
{
    AmNTGCClass    *pNTGC = ( AmNTGCClass *) pGC;
    HBRUSH       hBrush;
    HDC          hDC;
    AmColorCellRec col;
    double       dSin, dCos, dLength;
    int          x0, y0, x1, y1;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);
    cx = SCREEN_COORD( cx);
    cy = SCREEN_COORD( cy);

    /* 
     * Figure out the intersection of the arc with the surrounding
     *   rectangle.  <x0, y0> is the line drawn from the center of
     *   the arc to the point intersecting the arc at the end.
     *   <x1, y1> is the line drawn from the center of the arc to
     *   the point intersecting the arc at the beginning.
     */

    dSin = sin( dAngleEnd);
    dCos = cos( dAngleEnd);

    dLength = ( cx > cy ) ? cx / 2.0 : cy / 2.0;

    x0 = ( int)(( x + cx) / 2.0 + dLength * dCos + 0.5);
    y0 = ( int)(( y + cy) / 2.0 - dLength * dSin + 0.5);

    dSin = sin( dAngleStart);
    dCos = cos( dAngleStart);

    x1 = ( int)(( x + cx) / 2.0 + dLength * dCos + 0.5);
    y1 = ( int)(( y + cy) / 2.0 - dLength * dSin + 0.5);

    pGC->GetForegroundColor( &col);

    /* Create a brush to fill the arc with. */

    hBrush = SelectObject( hDC, CreateSolidBrush( col.pixel));

    /* Draw the filled arc. */

    Pie( hDC, x, y, x + cx + 1, y + cy + 1, x0, y0, x1, y1);

    /* Free the brush that we created before. */

    DeleteObject( SelectObject( hDC, hBrush));
}


/*
 * Function: void DrawRectangle( pWindow, pGC, x, y, cx, cy)
 *
 * Parameters:
 *   AmWindowClass *pWindow : The window to draw on. 
 *   AmGCClass     *pGC     : How to draw the rectangle (color, etc...)
 *   int         x        : The upper left x coordinate of the rectangle.
 *   int         y        : The upper left y coordinate of the rectangle
 *   int         cx       : The width of the rectangle. 
 *   int         cy       : The height of the rectangle.
 *
 * Description:
 *   DrawRectangle draws a rectangle with the upper left corner at <x,y>
 *   and with a width <cx> and a height <cy>.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::DrawRectangle( AmWindowClass *pWindow,
                               AmGCClass     *pGC,
                               int         x,
                               int         y,
                               int         cx,
                               int         cy)
{
    AmNTGCClass *pNTGC = ( AmNTGCClass *) pGC;
    HDC       hDC;
    HBRUSH    hBrush;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);
    cx = SCREEN_COORD( cx);
    cy = SCREEN_COORD( cy);

    hBrush = SelectObject( hDC, GetStockObject( NULL_BRUSH));
    Rectangle( hDC, x, y, x + cx + 1, y + cy + 1);
    SelectObject( hDC, hBrush);
}


/*
 * Function: void FillRectangle( pWindow, pGC, x, y, cx, cy)
 *
 * Parameters:
 *   AmWindowClass *pWindow : The window to draw on. 
 *   AmGCClass     *pGC     : How to draw the rectangle (color, etc...)
 *   int         x        : The upper left x coordinate of the rectangle.
 *   int         y        : The upper left y coordinate of the rectangle
 *   int         cx       : The width of the rectangle. 
 *   int         cy       : The height of the rectangle.
 *
 * Description:
 *   DrawRectangle draws a filled rectangle with the upper left corner
 *   at <x,y> and with a width <cx> and a height <cy>.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::FillRectangle( AmWindowClass *pWindow,
                               AmGCClass     *pGC,
                               int         x,
                               int         y,
                               int         cx,
                               int         cy)
{
    AmNTGCClass    *pNTGC = ( AmNTGCClass *) pGC;
    HDC          hDC;
    HBRUSH       hBrush;
    AmColorCellRec col;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);
    cx = SCREEN_COORD( cx);
    cy = SCREEN_COORD( cy);

    pGC->GetForegroundColor( &col);
    hBrush = SelectObject( hDC, CreateSolidBrush( col.pixel));
    Rectangle( hDC, x, y, x + cx + 1, y + cy + 1);
    DeleteObject( SelectObject( hDC, hBrush));
}


/*
 * Function: void FillPolygon( pWindow, pGC, pPoint, nPoint) 
 *
 * Parameters:
 *   AmWindowClass *pWindow : The window to draw on. 
 *   AmGCClass     *pGC     : How to draw the polygon (color, etc...)
 *   Point       *pPoint  : The vertices. 
 *   int         nPoint   : The number of vertices in the polygon.
 *
 * Description:
 *   FillPolygon draws a filled polygon.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::FillPolygon( AmWindowClass *pWindow,
                             AmGCClass     *pGC,
                             Point       *pPoint,
                             int         nPoint)
{
    AmNTGCClass    *pNTGC = ( AmNTGCClass *) pGC;
    POINT        *pPOINT;
    HDC          hDC;
    HBRUSH       hBrush;
    AmColorCellRec col;
    int          ii;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    /* Convert from Anvil points structure to NT windows points. */

    pPOINT = ( POINT *) alloca( sizeof( POINT) * nPoint);
    if ( pPOINT == NULL) {
        DO_ERROR( AmMALLOC_FAILED);
        return;
    }

    for ( ii = 0; ii < nPoint; ii++) { 
        pPOINT[ ii].x = SCREEN_COORD( pPoint[ ii].x);
        pPOINT[ ii].y = SCREEN_COORD( pPoint[ ii].y);
    }

    pGC->GetForegroundColor( &col);
    hBrush = SelectObject( hDC, CreateSolidBrush( col.pixel));
    Polygon( hDC, pPOINT, nPoint);
    DeleteObject( SelectObject( hDC, hBrush));
}


/*
 * Function: void CopyArea( pWindowSrc, pWindowDst, pGC, xSrc, ySrc,
 *                          xDst, yDst, cxSrc, cySrc)
 *
 * Parameters:
 *   AmWindowClass *pWindowSrc : The window to copy from. 
 *   AmWindowClass *pWindowDst : The window to copy to. 
 *   AmGCClass     *pGC        : How to draw the rectangle (raster op, etc...)
 *   int         xSrc        : The upper left x coordinate of the source
 *                             rectangle.
 *   int         ySrc        : The upper left y coordinate of the source
 *                             rectangle
 *   int         xDst        : The upper left x coordinate of the destination
 *                             rectangle.
 *   int         yDst        : The upper left y coordinate of the destination
 *                             rectangle.
 *   int         cxSrc       : The width of the rectangle. 
 *   int         cySrc       : The height of the rectangle.
 *
 * Description:
 *   CopyArea copies an area from the source window to a location in
 *   the destination window.  The passed in GC determines what kind of
 *   raster operation is to be used, and it also determines the  
 *   clipping region of the operation.  The GC has to be created for
 *   the destination window otherwise the clip region in the GC will not
 *   have any effect; however, the raster operation will take effect
 *   regardless.
 *  
 *   Note that the source window and the destination window can be the
 *   same.
 *
 * Returns:
 *   -nothing-
 */

void
AmNTDisplayClass::CopyArea( AmWindowClass *pWindowSrc,
                          AmWindowClass *pWindowDst,
                          AmGCClass     *pGC,
                          int         xSrc,
                          int         ySrc,
                          int         xDst,
                          int         yDst,
                          int         cxSrc,
                          int         cySrc)
{
    AmNTWindowClass *pNTWindowSrc = ( AmNTWindowClass *) pWindowSrc;
    AmNTWindowClass *pNTWindowDst = ( AmNTWindowClass *) pWindowDst;
    HDC           hDCSrc, hDCDst;
    AmRasterOpEnum  eRasterOp;
    int           iDrawMode;

    hDCDst = pNTWindowDst->getHDC( pGC);

    xDst = SCREEN_COORD( xDst);
    yDst = SCREEN_COORD( yDst);
    xDst = SCREEN_COORD( xSrc);
    yDst = SCREEN_COORD( ySrc);
    xDst = SCREEN_COORD( cxSrc);
    yDst = SCREEN_COORD( cySrc);

    if ( pNTWindowSrc == pNTWindowDst) {
        hDCSrc = hDCDst;
    } else {
        hDCSrc = pNTWindowSrc->getHDC( NULL);
    }

    eRasterOp = pGC->GetRasterOp();

    switch( eRasterOp) {
    case AmSET:
        iDrawMode = WHITENESS;
        break;
    case AmCLEAR:
        iDrawMode = BLACKNESS;
        break;
    case AmAND:
        iDrawMode = SRCAND;
        break;
    case AmAND_REVERSE:
        iDrawMode = SRCERASE;
        break;
    case AmAND_INVERTED:
        iDrawMode = 0x00220326; 
        break;
    case AmNAND:
        iDrawMode = 0x007700E6;
        break;
    case AmOR:
        iDrawMode = SRCPAINT;
        break;
    case AmOR_REVERSE:
        iDrawMode = 0x00DD0228;
        break;
    case AmOR_INVERTED:
        iDrawMode = MERGEPAINT;
        break;
    case AmNOR:
        iDrawMode = NOTSRCERASE;
        break;
    case AmXOR:
        iDrawMode = SRCINVERT;
        break;
    case AmEQUIV:
        iDrawMode = 0x00990066;
        break;
    case AmCOPY:
        iDrawMode = SRCCOPY;
        break;
    case AmCOPY_INVERTED:
        iDrawMode = NOTSRCCOPY;
        break;
    case AmNOOP:
        iDrawMode = 0x00AA0029;
        break;
    case AmINVERT:
        iDrawMode = DSTINVERT;
        break;
    default:
        iDrawMode = SRCCOPY;
    }

    BitBlt( hDCDst, xDst, yDst, cxSrc, cySrc, hDCSrc, xSrc, ySrc, iDrawMode);
}


void
AmNTDisplayClass::GetImage( AmWindowClass *pWindow,
                          AmImageClass  *pImage,
                          int         x,
                          int         y)
{
}


void
AmNTDisplayClass::DrawGlyph( AmWindowClass *pWindow,
                           AmGlyphClass  *pGlyph,
                           int         x,
                           int         y)
{
    pGlyph->DrawGlyph( this, pWindow, x, y);
}


void
AmNTDisplayClass::DrawImage( AmWindowClass *pWindow,
                           AmGCClass     *pGC,
                           int         xSrc,
                           int         ySrc,
                           int         xDst,
                           int         yDst,
                           int         cx,
                           int         cy,
                           AmImageClass  *pImage)
{
    AmNTGCClass     *pNTGC = ( AmNTGCClass *) pGC;
    AmNTImageClass  *pNTImage = ( AmNTImageClass *) pImage;
    HBITMAP       hBitmap;
    HDC		  hDC;
#if 0
    HDC           hDC = GetDC((( AmNTWindowClass *) pWindow)->getHWND());
#endif
    hDC = pNTGC->getHDC();
    /* Get a handle to a device context. */

    xSrc = SCREEN_COORD( xSrc);
    ySrc = SCREEN_COORD( ySrc);
    xDst = SCREEN_COORD( xDst);
    yDst = SCREEN_COORD( yDst);
    cx = MIN( SCREEN_COORD( cx) + 1, pNTImage->GetScreenWidth());
    cy = MIN( SCREEN_COORD( cy) + 1, pNTImage->GetScreenHeight());

    /* Create the device dependent bitamp and draw it to the screen. */

    hBitmap = pNTImage->GetBitmap( hDC);
    DrawBitmap( hDC, xSrc, ySrc, xDst, yDst, cx, cy, hBitmap);
#if 0
    ReleaseDC((( AmNTWindowClass *) pWindow)->getHWND(), hDC);
#endif
}


void
AmNTDisplayClass::DrawText( AmWindowClass *pWindow,
                          AmGCClass     *pGC,
                          int         x,
                          int         y,
                          char        *szText)
{
    AmNTGCClass  *pNTGC = ( AmNTGCClass *) pGC;
    TEXTMETRIC *ptm;
    HDC        hDC;
    int        iModeOld;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();
    ptm = pNTGC->GetNTFont();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);

    iModeOld = SetBkMode( hDC, TRANSPARENT);
    TextOut( hDC, x, y - ptm->tmAscent, szText, strlen( szText));
    SetBkMode( hDC, iModeOld);
}


void
AmNTDisplayClass::DrawImageText( AmWindowClass *pWindow,
                               AmGCClass     *pGC,
                               int         x,
                               int         y,
                               char        *szText)
{
    AmNTGCClass  *pNTGC = ( AmNTGCClass *) pGC;
    TEXTMETRIC *ptm;
    HDC        hDC;
    int        iModeOld;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();
    ptm = pNTGC->GetNTFont();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);

    iModeOld = SetBkMode( hDC, OPAQUE);
    TextOut( hDC, x, y - ptm->tmAscent, szText, strlen( szText));
    SetBkMode( hDC, iModeOld);
}


void
AmNTDisplayClass::DrawRotatedText( AmWindowClass   *pWindow,
                                 AmGCClass       *pGC,
                                 int           x,
                                 int           y,
                                 float         dRadianAngle,
                                 char          *szText)
{ 
    AmNTGCClass  *pNTGC = ( AmNTGCClass *) pGC;
    TEXTMETRIC *ptm;
    HDC        hDC;
    HFONT      hFont, hFontOld;
    int        iModeOld;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();
    ptm = pNTGC->GetNTFont();

    x = SCREEN_COORD( x);
    y = SCREEN_COORD( y);

    hFont = pNTGC->GetRotatedFont( RADTODEG( dRadianAngle));
    hFontOld = SelectObject( hDC, hFont);

    iModeOld = SetBkMode( hDC, TRANSPARENT);
    TextOut( hDC, x - ptm->tmAscent, y, szText, strlen( szText));
    SetBkMode( hDC, iModeOld);

    SelectObject( hDC, hFontOld);
    DeleteObject( hFont);
}


void
AmNTDisplayClass::ExtentsText( AmGCClass *pGC,
                             char    *szText,
                             int     *piWidth,
                             int     *piAscent,
                             int     *piDescent)
{  
    AmNTGCClass     *pNTGC = ( AmNTGCClass *) pGC;
    TEXTMETRIC    *ptm;   
    HDC           hDC;
    SIZE          size;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();
	ptm = pNTGC->GetNTFont();

    GetTextExtentPoint32( hDC, szText, strlen( szText), &size);

    *piWidth = VIRTUAL_COORD( size.cx);
    *piAscent = VIRTUAL_COORD( ptm->tmAscent);
    *piDescent = VIRTUAL_COORD( ptm->tmDescent);
}


int
AmNTDisplayClass::TextWidth( AmGCClass *pGC,
                           char    *szText,
                           int     iLength)
{
    AmNTGCClass     *pNTGC = ( AmNTGCClass *) pGC;
    HDC           hDC;
    SIZE          size;

    /* Get a handle to a device context. */

    hDC = pNTGC->getHDC();

    if ( iLength != 0) {
        GetTextExtentPoint32( hDC, szText, iLength, &size);
    } else {
        GetTextExtentPoint32( hDC, szText, strlen( szText), &size);
    }

    return( VIRTUAL_COORD( size.cx));
}


int
AmNTDisplayClass::FontHeight( AmGCClass *pGC,
                            int     *piAscent,
                            int     *piDescent)
{
    AmNTGCClass  *pNTGC = ( AmNTGCClass *) pGC;
    TEXTMETRIC *ptm = pNTGC->GetNTFont();   

    *piAscent = VIRTUAL_COORD( ptm->tmAscent);
    *piDescent = VIRTUAL_COORD( ptm->tmDescent);

    return( *piAscent + *piDescent);
}



AnvilBoolean
AmNTDisplayClass::AllocColor( AmColorCellPtr pcol)
{
    pcol->pixel = RGB( pcol->red, pcol->green, pcol->blue); 
    return( TRUE);
}

AnvilBoolean
AmNTDisplayClass::AllocNamedColor( const char   *szCol,
                                 AmColorCellPtr pcol)
{
    return( FALSE);
}

AmWindowClass *
AmNTDisplayClass::CreateBuffer( AmWindowClass *pWindow,
                              int         cx,
                              int         cy)
{
    return( new AmNTWindowClass( this, pWindow, cx, cy));
}


AmWindowClass *
AmNTDisplayClass::CreateWin( AmWindowClass *pWindow,
                           int         x,
                           int         y,
                           int         cx,
                           int         cy)
{
    AmNTWindowClass *pNTWindow;

    pNTWindow = new AmNTWindowClass( this, pWindow, x, y, cx, cy);
    return( pNTWindow);
}



/* Create an image (virtual function semi-constructor) */

AmImageClass *
AmNTDisplayClass::CreateImage( int     cx,
                             int     cy,
                             int     iDepth,
                             char    *pData)
{
    AmNTImageClass *pNTImage;

    pNTImage = new AmNTImageClass( this, cx, cy, iDepth, pData);
    return( pNTImage);
}

/* Virtual function semi-contstructor to create glyphs. */

AmGlyphClass *
AmNTDisplayClass::CreateGlyph( AmWindowClass *pWindow,
                             char        *szSymbolFile,
                             ulong_t     lPixel)
{
    return( NULL);
}

/**********************************************************************
 *                                                                  
 * AmNTWindowClass:                                                   
 *                                                                  
 **********************************************************************/

/* NTWindow (Microsoft Windows NT Window) do nothing constructor. */

AmNTWindowClass::AmNTWindowClass()
    : AmWindowClass()
{
    _eWindowType = AmWINDOW_ERROR;
    _hWindow     = NULL;
    _lstGC       = NULL;
    _hBitmap     = NULL;
    _hDC         = NULL;           /* Cached DC.                    */
    _pNTGC       = NULL;           /* GC associated with cached DC. */

    _cx          = 0;              /* Width of the buffer.          */
    _cy          = 0;              /* Height of the buffer.         */
    _fPrivate    = FALSE;
}

/* validity test function for windows NT windows */

AnvilBoolean AmNTWindowClass::IsValid()
{
    return( _hWindow != NULL);
}

/* NTWindow (Microsoft Windows NT Window) constructor. */

AmNTWindowClass::AmNTWindowClass( AmDisplayClass *pDisplay,
                              HWND         hWnd)
    : AmWindowClass( pDisplay)
{
    AmNTDisplayClass *pNTDisplay = ( AmNTDisplayClass *) pDisplay;

    /* Initializations. */

    _eWindowType = AmWINDOW_WINDOW;
    _hWindow     = hWnd;
    _lstGC       = NULL;
    _hBitmap     = NULL;
    _hDC         = NULL;           /* Cached DC.                    */
    _pNTGC       = NULL;           /* GC associated with cached DC. */

    _cx          = 0;              /* Only used for a buffer.       */
    _cy          = 0;              /* Ditto.                        */
    _fPrivate    = FALSE;

    _colBackground.red   = 0xff;
    _colBackground.green = 0xff;
    _colBackground.blue  = 0xff;
    pDisplay->AllocColor( &_colBackground);
}

static AnvilBoolean instanceFlag = TRUE;
static char szWinName[] = "AnvilWindow";

/* NTWindow (Microsoft Windows NT Window) constructor. */

AmNTWindowClass::AmNTWindowClass( AmDisplayClass *pDisplay,
                              AmWindowClass  *pWindow,
			      int	   x, 
			      int	   y,
			      int	   width,
			      int	   height)
    : AmWindowClass( pDisplay)
{
    HINSTANCE	    hInstance;
    AmNTWindowClass  *pNTWindow = ( AmNTWindowClass *) pWindow;
    AmNTDisplayClass *pNTDisplay = ( AmNTDisplayClass *) pDisplay;

    /* Initializations. */

    _eWindowType = AmWINDOW_WINDOW;
    _lstGC       = NULL;
    _hBitmap     = NULL;
    _hDC         = NULL;           /* Cached DC.                    */
    _pNTGC       = NULL;           /* GC associated with cached DC. */
    _fPrivate    = TRUE;

    hInstance = ( HINSTANCE) GetWindowLong( pNTWindow->getHWND(), 
					    GWL_HINSTANCE);

/* register the window class if it's the AmFirst time through */

    if ( instanceFlag) {
    	WNDCLASSEX	wc;

	wc.cbSize 	 = sizeof( wc);
	wc.style	 = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc   = ( WNDPROC) dispatchEvent;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = hInstance;
	wc.hIcon	 = LoadIcon( hInstance, szWinName);
	wc.hCursor 	 = LoadCursor( NULL, IDC_ARROW);
	wc.hbrBackground = ( HBRUSH)( COLOR_WINDOW + 1);
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = szWinName;
	wc.hIconSm	 = LoadIcon( hInstance, szWinName);
	instanceFlag  	 = FALSE;

	if ( RegisterClassEx( &wc) == NULL) {
	    _hWindow = NULL;
	    return;
	}
    }

    _hWindow = CreateWindow( szWinName, szWinName,
			     WS_CHILD | WS_VISIBLE | WS_BORDER,
			     x, y, width, height, pNTWindow->getHWND(),
			     NULL, hInstance, dispatchEvent);
}

/* NTWindow (Microsoft Windows NT Window) constructor for a buffer. */

AmNTWindowClass::AmNTWindowClass( AmDisplayClass *pDisplay,
                              AmWindowClass  *pWindow,
                              int          cx,
                              int          cy)
    : AmWindowClass( pDisplay)
{
    AmNTDisplayClass *pNTDisplay = ( AmNTDisplayClass *) _pDisplay;
    AmNTWindowClass  *pNTWindow = ( AmNTWindowClass *) pWindow;
    HDC            hDC;

    /* Initializations. */

    _eWindowType = AmWINDOW_BUFFER;
    _hWindow     = pNTWindow->getHWND();
    _lstGC       = NULL;
    _hDC         = NULL;           /* Cached DC.                    */
    _pNTGC       = NULL;           /* GC associated with cached DC. */
    _fPrivate    = TRUE;

    _cx          = cx;             /* Width of the buffer.          */
    _cy          = cy;             /* Height of the buffer.         */

    /* Create the internal bitmap used as a buffer. */

    hDC      = GetDC( _hWindow);
    _hBitmap = CreateCompatibleBitmap( hDC,
                                       _pDisplay->ScreenCoord( _cx),
                                       _pDisplay->ScreenCoord( _cy));
    ReleaseDC( _hWindow, hDC);

    _colBackground.red   = 0;
    _colBackground.green = 0;
    _colBackground.blue  = 0;
    pDisplay->AllocColor( &_colBackground);
}


/* NTWindow (Windows NT Window) destructor. */

AmNTWindowClass::~AmNTWindowClass()
{
    /* if the window isn't NULL, then destroy it */

    if ( _fPrivate && ( _hWindow != NULL)) {
	DestroyWindow( _hWindow);
    }

    /* Release any resources used by the window. */

    if ( _hDC != NULL) {
        releaseHDC( NULL);
    }

    if ( _hBitmap != NULL) {
        DeleteObject( _hBitmap);
    }
}

/*
 * Invalidate a region of a window.  A NULL parameter indicates that
 *   the entire window is to be invalidated..
 */

void
AmNTWindowClass::InvalidateRect( AmRectangleRec const *pRect)
{
    RECT rect;

    if ( pRect == NULL) {
        ::InvalidateRect( _hWindow, NULL, TRUE);
    } else {
        rect.left   = _pDisplay->VirtualCoord( pRect->x); 
        rect.top    = _pDisplay->VirtualCoord( pRect->y); 
        rect.right  = rect.left + _pDisplay->VirtualCoord( pRect->width); 
        rect.bottom = rect.top + _pDisplay->VirtualCoord( pRect->height); 

        ::InvalidateRect( _hWindow, &rect, TRUE);
    }
}

/* Clear a section of a window. */

void
AmNTWindowClass::ClearRect( AmRectangleRec const *pRect)
{
    HDC          hDC;
    HPEN         hPen;
    HBRUSH       hBrush;
    int          x, y, cx, cy;

    /* Get a handle to a device context. */

    hDC = getHDC( NULL);

    x = _pDisplay->ScreenCoord( pRect->x);
    y = _pDisplay->ScreenCoord( pRect->y);
    cx = _pDisplay->ScreenCoord( pRect->width);
    cy = _pDisplay->ScreenCoord( pRect->height);

    hPen = SelectObject( hDC,
                         CreatePen( PS_SOLID, 0, _colBackground.pixel));

    hBrush = SelectObject( hDC, CreateSolidBrush( _colBackground.pixel));

    Rectangle( hDC, x, y, x + cx + 1, y + cy + 1);
    DeleteObject( SelectObject( hDC, hBrush));
    DeleteObject( SelectObject( hDC, hPen));
}


/* Return the geometry of a window. */

void
AmNTWindowClass::GetWindowGeometry( int *px,
                                  int *py,
                                  int *pcx,
                                  int *pcy)
{
    RECT rect;

    if ( _hBitmap != NULL) {
        if ( px != NULL) {
            *px = 0;
        }
        if ( pcy != NULL) { 
            *py = 0;
        }
        if ( pcx != NULL) {
            *pcx = _cx;
        }
        if ( pcy != NULL) { 
            *pcy = _cy; 
        }
    } else {
        GetClientRect( _hWindow, &rect);

        if ( px != NULL) {
            *px = _pDisplay->VirtualCoord( rect.top);
        }
        if ( pcy != NULL) { 
            *py = _pDisplay->VirtualCoord( rect.left); 
        }
        if ( pcx != NULL) {
            *pcx = _pDisplay->VirtualCoord( rect.right - rect.left);
        }
        if  ( pcy != NULL) {
            *pcy = _pDisplay->VirtualCoord( rect.bottom - rect.top);
        }
    }
}


/* Return the extents of a window. */

void
AmNTWindowClass::GetDims( int *pcx, int *pcy) 
{
    RECT rect;

    if ( _hBitmap != NULL) {
        if ( pcx != NULL) {
            *pcx = _cx;
        }
        if ( pcy != NULL) { 
            *pcy = _cy; 
        }
    } else {
        GetClientRect( _hWindow, &rect);

        if ( pcx != NULL) {
            *pcx = _pDisplay->VirtualCoord( rect.right - rect.left);
        }
        if ( pcy != NULL) { 
            *pcy = _pDisplay->VirtualCoord( rect.bottom - rect.top);
        }
    }
}


/* Resize the window.  This only has an effect when the window is
   a buffer. */

void
AmNTWindowClass::ResizeWindow( int cx,
                             int cy)
{
    HDC      hDC;

    /* If the window is a buffer and the extents have changed, get
       rid of the old buffer and create a new one with the apropriate
       dimensions. */

    if (( _hBitmap != NULL) && (( cx != _cx) || ( cy != _cy))) {

        _cx = cx;
        _cy = cy;

        /* Release any currently used DC in case they have the bitmap
           selected. */

        releaseHDC( NULL);

        /* Get rid of the old bitmap. */

        DeleteObject( _hBitmap);

        /* And create a new bitmap of the desired size. */

        hDC      = GetDC( _hWindow);
        _hBitmap = CreateCompatibleBitmap( hDC,
                                           _pDisplay->ScreenCoord( _cx),
                                           _pDisplay->ScreenCoord( _cy));
        ReleaseDC( _hWindow, hDC);
    } else if ( _hWindow != NULL) {
	RECT	rect;

	GetWindowRect( _hWindow, &rect);
	::MoveWindow( _hWindow, rect.left, rect.top, cx, cy, TRUE);
	_cx = cx;
	_cy = cy;
    }
}


void
AmNTWindowClass::MoveWindow( int x,
                           int y)
{
    if ( _hWindow != NULL) {
	RECT	rect;

	GetWindowRect( _hWindow, &rect);
	::MoveWindow( _hWindow, x, y, 
		      rect.right - rect.left, rect.bottom - rect.top, TRUE);
    }
}

/* Semi-virtual NTGC creation function. */

AmGCClass *
AmNTWindowClass::CreateGC()
{
    AmNTGCClass *pNTGC;

    pNTGC = new AmNTGCClass( this);
    return( pNTGC);
}


/**********************************************************************
 *
 * Microsoft Windows NT window specific functions:
 * (functions that are not inherited from AmWindowClass)
 *
 **********************************************************************/

/* Return the window's handle. */

HWND
AmNTWindowClass::getHWND()
{ 
    return( _hWindow);
}

/* Return a handle to a device context that is filled with the values
   of the passed in GC.  A cache is used to speed up consecutive usage
   of a GC. */

HDC
AmNTWindowClass::getHDC( AmGCClass *pGC)
{
    AmNTGCClass *pNTGC = ( AmNTGCClass *) pGC;
    HDC       hDC, hDCMem;

    /* If we have never created a device context, do so now. */

	if (( _hDC != NULL) && ( pNTGC == NULL)) {
        releaseHDC( NULL);
    }

    if ( _hDC == NULL) {

        /* Get a device context from the window handle. */

        hDC = GetDC( _hWindow);

        /* If this window is a "buffer" instead of a "real" window, we need
           to create a memory device context. */

        if ( _hBitmap != NULL) {
            hDCMem = CreateCompatibleDC( hDC);
            SelectObject( hDCMem, _hBitmap);
            ReleaseDC( _hWindow, hDC);
            hDC = hDCMem;
        }

        _hDC = hDC;

        /* If there is a GC set the values of the device context
           accordingly. */

        if ( pNTGC != NULL) {
            pNTGC->setHDC( _hDC);
        }
    } else {

        /* If the cached GC is the same as the passed in GC we just
           need to update the device context's values.  Otherwise we
           need to reset all of the values. */

        if ( pNTGC != NULL) {
            if ( _pNTGC == pNTGC) {
                pNTGC->cleanHDC( _hDC);
            } else {
                pNTGC->setHDC( _hDC);
            }
        }
    }

    /* Cache the passed-in GC. */

    _pNTGC = pNTGC;

    return( _hDC);
}

/* Release the device context used by the window.  If the window is
   a buffer, the device context previously created is deleted, otherwise
   the device context is simply released. */

void
AmNTWindowClass::releaseHDC( AmGCClass *pGC)
{

    /* Check whether the cached GC is the GC for which we want to release
       the device context, or if the device context in use is to be released
       regardless (the passed in GC is set to NULL). */

    if (( pGC == NULL) || ( pGC == _pNTGC)) {

        /*
         * Delete the DC if it was created using CreateCompatibleDC(),
         *   otherwise release the DC since it was gotten through GetDC().
         */

        if ( _hBitmap != NULL) {
            DeleteDC( _hDC);
        } else {
            ReleaseDC( _hWindow, _hDC);
        }

        _hDC = NULL;
    }
}
