#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <windows.h>
#include <anvil/project.h>
#include <anvil/ntdisplay.h>
#include <anvil/ntcolor.h>

/*
 * Function: void DrawBitmap( hDC, xSrc, ySrc, xDst, yDst, cx, cy,
 *                            hBitmap)
 * Description:
 *   DrawBitmap draws a section of a bitmap onto a device context. 
 *   The section is specified through the xSrc, ySrc and cx, cy
 *   parameters (the starting location and the dimensions respectively).
 *   The location of the section is specified by xDst, yDst and cx, cy.
 */

void
DrawBitmap( HDC     hDC,
            int     xSrc,
            int     ySrc,
            int     xDst,
            int     yDst,
            int     cx,
            int     cy,
            HBITMAP hBitmap)
{
    HDC    hDCMem;

    /* Create the memory device context. */

    hDCMem = CreateCompatibleDC( hDC); 
    SelectObject( hDCMem, hBitmap);
    
	/* Make sure the mapping mode of the memory device context matches
       the one of the passed in device context. */

    SetMapMode( hDCMem, GetMapMode( hDC)); 

    /* Copy the bitmap onto the device. */

	GdiFlush();
    BitBlt( hDC, xDst, yDst, cx, cy, hDCMem, xSrc, ySrc, SRCCOPY);

    DeleteDC( hDCMem);
}

/*
 * Function: AmNTImageClass( pDisplay, cx, cy, iDepth, pcData)
 *  
 * Parameters:
 *   pDisplayClass pDisplay         : The display to create the image with.
 *   int           cx               : The width of the image (in virtual
 *                                    coordinates).
 *   int           cy               : The height of the image (in virtual
 *                                    coordinates).
 *   int           iDepth           : The depth of the image (1 or 32).
 *   char          *pcData          : The data to initialize the image with.
 *
 * Description:
 *   The constuctor for the AmNTImageClass.  The image is created for the
 *   passed in display (pDisplay).  The dimensions of the new image are
 *   given by cx and cy (width and height respectively).  The dimensions
 *   are in virtual coordinates.  The passed in data is used to initialize
 *   the image - it consists of 32 bit pixels.  That is, each pixel is four
 *   bytes long.  The AmFirst byte is the BLUE pixel, the second byte is the
 *   GREEN pixel, the third byte is the RED pixel, and finally the fourth
 *   byte is filler data and should be set to 0.
 *
 */

AmNTImageClass::AmNTImageClass( AmDisplayClass *pDisplay,
                            int          cx,
                            int          cy,
                            int          iDepth,
                            char         *pcData)
    : AmImageClass( pDisplay, cx, cy) 
{
    AnvilBoolean  fMonochrome = ( iDepth == 1);
	uint_t   ii;
	ushort_t *pus;
    ulong_t  *pul;

    /* Initialize device independent information. */

    _hBitmap = NULL;

    /*
     * The internal image is either a monochrome (1-bit) image or
     *   a true color (32-bit) image.  The conversion to a device
     *   dependent bitmap occurs later.
     */

    if ( fMonochrome == TRUE) {
        _iDepth = 1;
        _iBitsPerPixel = 8;
        _iBytesPerPixel = 1;
        _iBytesPerLine = _iBytesPerPixel * _cxPixels;
    } else {
        _iDepth = 32;
        _iBitsPerPixel = 32;
        _iBytesPerPixel = 4;
        _iBytesPerLine = _iBytesPerPixel * _cxPixels;
    }

    _iSize = _cyPixels * _iBytesPerLine;

    /* Initialize device dependent information. */

    if ( fMonochrome == TRUE) {
        _iDDBitsPerPixel = 8;
        _nDDColors = 2;
    } else {

        /*
         * If the application is running on a true color display then we
         *   use a 32-bit bitmap which is represented in RGB values.
         *   Otherwise we use an 8-bit bitmap which uses dithering to 
         *   aproximate the RGB values with the 256 colors allowed.
         *   We let windows handle any necessary conversions from the
         *   32-bit bitmap to the 24 or 16 bit displays, and from the
         *   8-bit bitmap to the 4, 2, and 1 bit displays.  This means
         *   images run faster on 32-bit and on 8-bit displays than on
         *   other displays; however, 24-bit and 16-bit displays have the
         *   advantage of not being dithered, so they might be faster than
         *   8-bit displays).
         */

        if ( AmGetAppDepth() < 16) {
            _iDDBitsPerPixel = 8;
            _nDDColors = GetAppNumColors();
        } else {
            _iDDBitsPerPixel = 32;
            _nDDColors = 0;
        }
    }

    /* Bytes per pixels is bits per pixel rounded up to the nearest 8. */

    _iDDBytesPerPixel = ( _iDDBitsPerPixel + 7) >> 3;

    /* Bytes per line is bytes per pixel rounded up to the nearest 4. */

    _iDDBytesPerLine = ((( _iDDBytesPerPixel * _cxPixels) + 3) >> 2) << 2; 

    _iDDSize = _cyPixels * _iDDBytesPerLine;

    /* Initialize the BITMAPINFOHEADER structure that is used
       to convert the device independent data to device dependent
       data. */

    _bmih.biSize          = sizeof( BITMAPINFOHEADER);
    _bmih.biWidth         = _cxPixels;
    _bmih.biHeight        = - _cyPixels;
    _bmih.biPlanes        = 1;
    _bmih.biBitCount      = _iDDBitsPerPixel;
    _bmih.biCompression   = BI_RGB;
    _bmih.biSizeImage     = _iDDSize;
    _bmih.biXPelsPerMeter = 0;
    _bmih.biYPelsPerMeter = 0;
    _bmih.biClrUsed       = 0;
    _bmih.biClrImportant  = 0;

    /* Initialize the BITMAPINFO structure that is used to
       convert the data. */

    if ( _iDDBitsPerPixel == 8) {
        _pbmi = ( BITMAPINFO *) malloc( sizeof( BITMAPINFOHEADER) +
                                        sizeof( ushort_t) * _nDDColors);

        /* If colors are used, fill the color structure with the appropriate
           color indecies. */

        _pbmi->bmiHeader = _bmih;
	    pus = ( ushort_t *) &_pbmi->bmiColors[ 0];

        for ( ii = 0; ii < _nDDColors; ii++) {
	        pus[ ii] = ii;
        }
    } else {
        _pbmi = ( BITMAPINFO *) malloc( sizeof( BITMAPINFOHEADER) +
                                        sizeof( ulong_t));
        _pbmi->bmiHeader = _bmih;
        pul = ( ulong_t *) &_pbmi->bmiColors[ 0];
        pul = NULL;
	}

    /* Initialize the device dependent data (for 8-bit images). */

    _pcDDData = NULL;
	
    /* Set the internal image. */

	_pcData = NULL;
    if ( pcData != NULL) {
        SetData( pcData);
    } else {
        _pcData = ( uchar_t *) calloc( 1, sizeof( char) * _iSize); 
    }
}

/*
 * Function: ~AmNTImageClass()
 * 
 * Description:
 *   The AmNTImageClass destructor frees up any resources used by the image.
 *   This includes the internal device independent and device dependent
 *   data as well as the bitmap info structure and the cached bitmap.
 */

AmNTImageClass::~AmNTImageClass()
{
    if ( _pcData != NULL) {
        free( _pcData);
    }

    if ( _pbmi != NULL) {
        free( _pbmi);
    }

    if ( _hBitmap != NULL) {
        GdiFlush();
        DeleteObject( _hBitmap);
    }
}

/* 
 * Set the internal image according to the data passed in. The 
 *   representation of the data and the internal image have to
 *   match perfectly (32-bit BBGGRR00).
 */

AnvilBoolean
AmNTImageClass::SetData( char *pcData)
{
    if (( _iDDBitsPerPixel == 8) || ( _pcDDData == NULL)) {
        if ( _pcData == NULL) {
            _pcData = ( uchar_t *) malloc( sizeof( char) * _iSize);
            if ( _pcData == NULL) {
                ANVIL_ERROR_CODE = AmMALLOC_FAILED;
                return( FALSE);
            }
        }
  
        memcpy( _pcData, pcData, _iSize);
        _fDirty = TRUE;

    } else {

        GdiFlush();
        memcpy( _pcDDData, pcData, _iDDSize);

    }

    return( TRUE);
}


AnvilBoolean
AmNTImageClass::CopyData8Bit()
{
    const ulong_t *plRedDitherMatrix, *plGreenDitherMatrix,
                  *plBlueDitherMatrix, *plPaletteOffset;
    uchar_t       *pcSrc, *pcDst;
    int           xMod, x, yMod, y, i, iDstEOLOffset,
                  iRedOffset, iGreenOffset, iBlueOffset;

    if ( _pcDDData == NULL) {
        return( FALSE);
    }

    if ( _pcData == NULL) {
        return( TRUE);
    }

    iDstEOLOffset = _iDDBytesPerLine - _iDDBytesPerPixel * _cxPixels;

    plRedDitherMatrix = GetNTRedDitherMatrix();
    plGreenDitherMatrix = GetNTGreenDitherMatrix();
    plBlueDitherMatrix = GetNTBlueDitherMatrix();

	plPaletteOffset = GetAppPaletteOffsets();

	GdiFlush();
	pcDst = _pcDDData;
    pcSrc = _pcData;

    yMod = 0;
    for ( y = 0; y < _cyPixels; y++, pcDst += iDstEOLOffset) {

        xMod = 0; 
        for ( x = 0;
              x < _cxPixels;
              x++, pcSrc += _iBytesPerPixel, pcDst += _iDDBytesPerPixel) {

            i = plRedDitherMatrix[ pcSrc[ RED]];
            iRedOffset = ( pcDecisionMatrix[ SECONDWORD( i)][ yMod][ xMod]) ?
                             SECONDBYTE( i) : FIRSTBYTE( i);

            i = plGreenDitherMatrix[ pcSrc[ GREEN]];
            iGreenOffset = ( pcDecisionMatrix[ SECONDWORD( i)][ yMod][ xMod]) ?
                               SECONDBYTE( i) : FIRSTBYTE( i);

            i = plBlueDitherMatrix[ pcSrc[ BLUE]];
            iBlueOffset = ( pcDecisionMatrix[ SECONDWORD( i)][ yMod][ xMod]) ?
                              SECONDBYTE( i) : FIRSTBYTE( i);
		
	    *pcDst = ( uchar_t) 
		plPaletteOffset[ iRedOffset + iGreenOffset + iBlueOffset];

            xMod++;
            xMod &= 0x07;
        }

        yMod++;
        yMod &= 0x07;
    }

    _fDirty = FALSE;

    return( TRUE);
}

AnvilBoolean
AmNTImageClass::CopyData32Bit()
{
    if ( _pcDDData == NULL) {
        return( FALSE);
    }

	if ( _pcData != NULL) {

        GdiFlush();
        memcpy( _pcDDData, _pcData, _iDDSize);

        free( _pcData);
	    _pcData = NULL;
    }

    _fDirty = FALSE;

    return( TRUE);
}

/* 
 * Copy the internal image to the actual image, translation from the
 *   different depth is done here.
 */

AnvilBoolean
AmNTImageClass::CopyData()
{
    if ( _iDDBitsPerPixel == 8) {
        return( CopyData8Bit());
    } else {
        return( CopyData32Bit());
    }
}

/*
 * Retreive the data for the image.  Note that if the data is changed SetDirty( TRUE)
 *   should be called to tell the image that it needs to update its' internal data
 *   structures before it gets drawn the next time.
 */

char *
AmNTImageClass::AmGetData()
{
    _fDirty = TRUE;
    if (( _iDDBitsPerPixel == 8) || ( _pcDDData == NULL)) {
        return(( char *) _pcData);
    } else {

        GdiFlush();
        return(( char *) _pcDDData);

    }
}

HBITMAP
AmNTImageClass::CreateBitmap8Bit( HDC hDC)
{
    _hBitmap = CreateDIBSection( hDC, _pbmi, DIB_PAL_COLORS,
                                ( void **) &_pcDDData, NULL, 0);

    CopyData8Bit();

    return( _hBitmap);
}

HBITMAP
AmNTImageClass::CreateBitmap32Bit( HDC hDC)
{
    DIBSECTION ds;

    _hBitmap = CreateDIBSection( hDC, _pbmi, DIB_RGB_COLORS,
                                ( void **) &_pcDDData, NULL, 0);


    if ( _pcData != NULL) {

		GdiFlush();
        memcpy( _pcDDData, _pcData, _iDDSize);

		free( _pcData);
		_pcData = NULL;
    }

	AmGetObject( _hBitmap, sizeof( DIBSECTION), &ds);

    return( _hBitmap);
}

HBITMAP
AmNTImageClass::CreateBitmap( HDC hDC)
{
    if ( _hBitmap != NULL) {
        GdiFlush();
        DeleteObject( _hBitmap);
    }

    if ( _iDDBitsPerPixel == 8) {
        _hBitmap = CreateBitmap8Bit( hDC);
    } else {
        _hBitmap = CreateBitmap32Bit( hDC);
    }

    _fDirty = FALSE;
    return( _hBitmap);
}


HBITMAP
AmNTImageClass::GetBitmap( HDC hDC)
{
    if ( _hBitmap == NULL) {
        CreateBitmap( hDC);
    } else if ( _fDirty == TRUE) {
        CopyData();
    }

    return( _hBitmap);
}
