#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdarg.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <anvil/plot.h>
#include <anvil/resource.h>
#include <anvil/copyobj.h>
#include <anvil/ntdisplay.h>
#include <anvil/ntdispmgr.h>
#include <anvil/magic.h>
#include <anvil/dispatch.h>
#include <anvil/nt.h>
#include <anvil/nt_user_data.h>
#include <resources/resource.h>

class MagnifyClass : public AmMagicPlotClass {
private:
    float	_factor;
    HWND	_hDlg;
public:
    MagnifyClass( AmNTDisplayManagerClass	*displayManager,
		  AmOverlayClass		*sourcePlot,
		  AnvilBoolean		callOnDrag);
    void 		SetMagnify( float value) { _factor = value; };
    float		GetMagnify() { return( _factor); };
    virtual void	SetRect( AmRectangleRec const *rectArg);
    virtual 		~MagnifyClass();
    void		SetDialog( HWND hDlg) { _hDlg = hDlg; };
    HWND		GetDialog() { return( _hDlg); };
};

static const int MAGNIFY_MAX = 10;
static char *MAGNIFY_STRING = "Magnify";

/***********************************************************************
 *
 * Description:
 * 	update the magnified plot when the rectangle is changed.
 * Arguments:
 *	rectArg - rectangle to move or resize to.
 *
 ***********************************************************************/

void MagnifyClass::SetRect( AmRectangleRec const *rectArg)
{
    float	xmin, ymin, xmax, ymax, cx, cy, wx, wy;
    int		eventId = AM_COORDS_CHANGE;

AN_DEBUG( 50, ( "calling MagnifyClass::SetRect %d, %d, %d, %d\n",
		rectArg->x, rectArg->y, rectArg->width, rectArg->height));

    rect = *rectArg;
    GetSourceCoords( &xmin, &ymin, &xmax, &ymax);
    cx = ( xmax + xmin)/2;
    cy = ( ymax + ymin)/2;
    wx = ( xmax - xmin)/( _factor*2);
    wy = ( ymax - ymin)/( _factor*2);
    plot->SetCoords( cx - wx, cy - wy, cx + wx, cy + wy);
    plot->Update( &eventId, 1);
}

/***********************************************************************
 *
 * Description:
 * 	Magnifying Glass magic window destructor.  Blow the dialog away.
 * 	The default MagicWindow destructor removes the plot, and 
 *	~MagicWindow removes the window.
 *
 ***********************************************************************/

MagnifyClass::~MagnifyClass()
{
    if ( _hDlg != NULL) {
	EndDialog( _hDlg, 0);
	_hDlg = NULL;
    }
}

/***********************************************************************
 *
 * Description:
 *	Dialog procedure for the magnify magic window.
 * Arguments:
 *	hDlg - handle to the dialog.
 *	iMsg - message sent by dispatcher.
 *	wParam, lParam - message arguments.
 *
 ***********************************************************************/

static BOOL CALLBACK magnifyMagicDialogProc( HWND	hDlg,
					     UINT	iMsg,
					     WPARAM	wParam,
					     LPARAM	lParam)
{
    HWND		hWndTrackBar;
    int			min, max, factor;
    MagnifyClass	*magnify;

    switch ( iMsg) {

/* on init dialog, set the values for the trackbar */

    case WM_INITDIALOG:
	min = 1;
	max = AmGetDefault( "magnify.max", MAGNIFY_MAX);
	factor = 2;
	hWndTrackBar = GetDlgItem( hDlg, MAG_SLIDER);
	SendMessage( hWndTrackBar, TBM_SETRANGEMIN, 
		     ( WPARAM) FALSE, ( LPARAM) min);
	SendMessage( hWndTrackBar, TBM_SETRANGEMAX, 
		     ( WPARAM) FALSE, ( LPARAM) max);
	SendMessage( hWndTrackBar, TBM_SETPOS,
		     ( WPARAM) TRUE, ( LPARAM) factor);
	return( TRUE);

/* listen for the commands to the dialog and slider. */

    case WM_COMMAND:

/* try and pull the user data from the dialog */

	if (( magnify = ( MagnifyClass *) 
			    AmGetUserData( hDlg, MAGNIFY_STRING)) == NULL) {
	    AmPrintResourceError( hDlg, "getUserDataFailed");
	    return( FALSE);
	}
        switch ( LOWORD( wParam)) {
	case ID_DISMISS:
	    DestroyWindow( hDlg);
	    magnify->SetDialog( NULL);
	    delete magnify;
	    return( TRUE);
	}
	break;
    case WM_HSCROLL:
	AmNTWindowClass  		*pMagicWin;
	RECT	   		lpRect;
	AmRectangleRec const	*magicRect;

	if (( magnify = ( MagnifyClass *) 
			    AmGetUserData( hDlg, MAGNIFY_STRING)) == NULL) {
	    AmPrintResourceError( hDlg, "getUserDataFailed");
	    return( FALSE);
	}
	pMagicWin = ( AmNTWindowClass *) magnify->GetWindow();
	magicRect = magnify->GetRect();
	hWndTrackBar = GetDlgItem( hDlg, MAG_SLIDER);
	factor = SendMessage( hWndTrackBar, TBM_GETPOS,
			      ( WPARAM) 0, ( LPARAM) 0);
	magnify->SetMagnify( factor);
	lpRect.top = 0;
	lpRect.left = 0;
	lpRect.right = magicRect->width;
	lpRect.bottom = magicRect->height;
	InvalidateRect( pMagicWin->getHWND(), &lpRect, FALSE);
	magnify->Update();
	break;
    }
    return( FALSE);
}

/***********************************************************************
 *
 * Description:
 *	drawMagic paint handler draws the window contents when it's
 *	exposed.
 * Arguments:
 *	hWnd - handle to the magic window.
 *	iMsg - message sent by dispatcher.
 *	wParam, lParam - message arguments.
 *	magnify - passed magnifying glass magic window data.
 *
 ***********************************************************************/

void drawMagic( HWND		hWnd,
		UINT		iMsg,
		WPARAM		wParam,
		LPARAM		lParam,
		MagnifyClass	*magnify)
{
    HDC			hDC;
    AmRectangleRec	rect;
    PAINTSTRUCT		ps;

    hDC = BeginPaint( hWnd, &ps); 
    rect.x = ps.rcPaint.left;
    rect.y = ps.rcPaint.top;
    rect.width = ps.rcPaint.right - ps.rcPaint.left;
    rect.height = ps.rcPaint.bottom - ps.rcPaint.top;
    magnify->DrawWrapper( &rect);
    EndPaint( hWnd, &ps);
}

/***********************************************************************
 *
 * Description:
 * 	derived constructor from AmMagicPlotClass for Magnify cursor. 
 * Arguments:
 *	displayManager - display manager the magnifying glass is 
 *	associated with.
 *	sourcePlot - plot to magnify.
 *	callOnDrag - call when dragged, or only at end of reposition.
 *
 ***********************************************************************/

MagnifyClass::MagnifyClass( AmNTDisplayManagerClass	*displayManager,
			    AmOverlayClass		*sourcePlot,
			    AnvilBoolean			callOnDrag) :
			AmMagicPlotClass( displayManager, 
					( AmDataGraphicClass *) sourcePlot,
					callOnDrag)
{
    int		    errCode;
    HINSTANCE	    hInstance;
    AmNTWindowClass   *pWindow = ( AmNTWindowClass *) displayManager->GetWindow(),
		    *pMagicWin =  ( AmNTWindowClass *) GetWindow();

    _factor = 2.0;
    Update();

/* create a modeless dialog box */

    hInstance = ( HINSTANCE) GetWindowLong( pWindow->getHWND(), GWL_HINSTANCE);
    _hDlg = CreateDialog( hInstance, MAKEINTRESOURCE( MAGNIFY_MAGIC_DLOG),
			 pWindow->getHWND(), ( DLGPROC) magnifyMagicDialogProc);
    if ( _hDlg == NULL) {
    	AmPrintResourceError( pWindow->getHWND(), "magnifyDialogFailed");
    }
    if ( AmAddUserData( _hDlg, MAGNIFY_STRING, this) != AmOK) {
    	AmPrintResourceError( pWindow->getHWND(), "addUserDataFailed");
    }
    AmRegisterDialog( _hDlg);
    ShowWindow( _hDlg, SW_SHOW);
    AmInsertHandler( pMagicWin->getHWND(), WM_PAINT, 0, 
					 ( AmNtHandlerFn) drawMagic, this);
}

/***********************************************************************
 *
 * Description:
 * 	callback to set up a magnifying glass for an object 
 * Arguments:
 *	hWnd - handle to window sending message.
 *	iMsg - message sent by dispatcher.
 *	wParam, lParam - message arguments.
 *	displayManager - display manager magnifying glass is to be
 *	associated with.
 *	overlay - overlay plot to magnify.
 *
 ***********************************************************************/

void AmMagnifyGlassCallback( HWND				hWnd,
			   UINT				iMsg,
			   WPARAM			wParam,
			   LPARAM			lParam,
			   AmNTDisplayManagerClass	*displayManager,
			   AmOverlayClass			*overlay)
{
    new MagnifyClass( displayManager, overlay, 
		      AmGetDefault( "magnify.drag", ( AnvilBoolean) FALSE));
}
