#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <malloc.h>
#include <stdlib.h>
#include <windows.h>
#include <anvil/ntdispmgr.h>
#include <anvil/dispmgrargs.h>
#include <anvil/coordinates.h>
#include <anvil/dispatch.h>
#include <anvil/userevents.h>
#include <anvil/nt_state_hand.h>
#include <anvil/outline.h>
#include <anvil/macros.h>
#include <anvil/plot_handlers.h>

/* list of events listened for by the zoom2xHandler */

static UINT zoom2xEvents[] = { WM_LBUTTONDOWN, WM_RBUTTONDOWN, WM_MOUSEMOVE };

/* cursor suspend function, simply hides the outline */

static void
suspendZoom2xHandler( AmDispMgrArgsPtr dispMgrArgs)
{
    AmMapManagerClass	    *displayManager = dispMgrArgs->displayManager;
    AmObjectOutlineClass    *outline = ( AmObjectOutlineClass *)
					dispMgrArgs->outline;

    displayManager->HideOutline( outline);
}

/* cursor restore function, draws the outline and displays a restart message */

static void
restartZoom2xHandler( AmDispMgrArgsPtr dispMgrArgs)
{
    AmMapManagerClass	    *displayManager = dispMgrArgs->displayManager;
    AmObjectOutlineClass    *outline = ( AmObjectOutlineClass *)
					dispMgrArgs->outline;

    displayManager->ShowOutline( outline);
    displayManager->WriteResourceMessage( "restartZoom2xHandler");
}

/* remove function removes the outline from the outline list, and frees
   the plot arguments */

static void
removeZoom2xHandler( AmDispMgrArgsPtr dispMgrArgs)
{
    AmMapManagerClass	    *displayManager = dispMgrArgs->displayManager;
    AmObjectOutlineClass    *outline = ( AmObjectOutlineClass *)
					dispMgrArgs->outline;

    displayManager->RemoveOutline( outline);
    AmFreeDispMgrArgs( dispMgrArgs);
}

/***************************************************************************
 *
 * Description:
 * 	handler activated by AmZoom2xCallback, this listens for left button
 *	press (which zooms the object), right button press (which cancels)
 * 	and mouse motion, which drags the outline with the mouse in the
 *	center.  The outline is restricted to the extents of the object
 *	being zoomed.
 * Arguments:
 *	hWnd: window the object is displayed on, and the handler is 
 *	activated for.
 *	iMsg: either WM_RBUTTONDOWN, WM_LBUTTONDOWN, or WM_MOUSEMOVE
 *	wParam: contains the mouse x,y position.
 *	dispMgrArgs: contstructed by the callback, contains the display
 *	manager, the screen object and the outline to be dragged.
 *
 ***************************************************************************/

static void 
zoom2xHandler( HWND		hWnd,
	       UINT		iMsg,
	       WPARAM		wParam,
	       LPARAM		lParam,
	       AmDispMgrArgsPtr	dispMgrArgs)
{
    AmCoordClass		*coords = dispMgrArgs->coords;
    AmMapManagerClass		*displayManager = dispMgrArgs->displayManager;
    AmObjectOutlineClass        *outline = ( AmObjectOutlineClass *)
						dispMgrArgs->outline;
    AmRectangleRec const	*rect = coords->GetRect();
    float			currXmin, currXmax, currYmin, currYmax;
    int                 	x, y;

    x = displayManager->VirtualCoord( LOWORD( lParam)); 
    y = displayManager->VirtualCoord( HIWORD( lParam));

/*
 * Make sure the zoom rectangle doesn't go beyond the extents of the
 * object.
 */

    if ( x + rect->width/4 > rect->x + rect->width) {
        x = rect->x + rect->width - rect->width/4;
    } else if ( x - rect->width/4 < rect->x) {
        x = rect->x + rect->width/4;
    }
    if ( y + rect->height/4 > rect->y + rect->height) {
        y = rect->y + rect->height - rect->height/4;
    } else if ( y - rect->height/4 < rect->y) {
        y = rect->y + rect->height/4;
    }
    if ( iMsg == WM_LBUTTONDOWN) {

/* Project the rectangle's coordinates into data space, reset the
   plot's coordinate system, update, and send a user event to notify
   that the plot's coordinate system has changed. */

	coords->GetDataPt( x - rect->width/4, y + rect->height/4,
			   &currXmin, &currYmin);
	coords->GetDataPt( x + rect->width/4, y - rect->height/4,
			   &currXmax, &currYmax);
	coords->SetCoords( currXmin, currYmin, currXmax, currYmax);

/* remove the event handler before the flush because it removes the
   outline. Display the zoom completion message */

	AmRemoveHandler( hWnd, zoom2xEvents, VECLEN( zoom2xEvents),
		       ( AmNtHandlerFn) zoom2xHandler, dispMgrArgs);
	displayManager->WriteResourceMessage( "zoomedMsg");
	displayManager->Flush();
    } else if ( iMsg == WM_MOUSEMOVE) {
	outline->SetOrigin( x - rect->width/4, y - rect->height/4);
	outline->SetSize( rect->width/2, rect->height/2);
    } else if ( iMsg == WM_RBUTTONDOWN) {

/* remove the event handler before the flush because it removes the
   outline. Display the zoom cancel message */

	AmRemoveHandler( hWnd, zoom2xEvents, VECLEN( zoom2xEvents),
		       ( AmNtHandlerFn) zoom2xHandler, dispMgrArgs);
	displayManager->WriteResourceMessage( "zoomCancelMsg");
    }
}

/***************************************************************************
 *
 * Description:
 * 	Zoom2X by allowing the user to drag a window half the size of the
 *	object by its center.  The event handler is added with 
 * 	ZOOM_2X_PRIORITY.  Left button press zooms the object, 
 *	right button press cancels, and mouse motion, which drags the 
 *	outline with the mouse in the center.  The outline is restricted 
 *	to the extents of the object being zoomed.
 * Arguments:
 *	hWnd: The control that launches the callback (there are problems
 *	with this...)
 *	iMsg, wParam, lParam: the message id that launched the zoom callback.
 *	displayManager: display manager the object is displayed on.
 *	coords: The object to zoom.
 *
 ***************************************************************************/

void AmZoom2xCallback( HWND			hWnd,
		       UINT			iMsg,
		       WPARAM			wParam,
		       LPARAM			lParam,
		       AmNTDisplayManagerClass	*displayManager,
		       AmCoordClass		*coords)
{
    AmRectangleRec const		*rect;
    AmObjectOutlineClass        	*outline;
    AmNTWindowClass		*window = ( AmNTWindowClass *) 
					displayManager->GetWindow();
    UINT			msgCode;
    
    displayManager->WriteResourceMessage( "zoomInstructionMsg");
    rect = coords->GetRect();
    outline = new AmObjectOutlineClass( displayManager,
				      rect->x + rect->width/2,
				      rect->y + rect->height/2,
				      rect->width/2, rect->height/2);
    displayManager->AddOutline( outline);
    msgCode = ( iMsg == WM_COMMAND) ? wParam : iMsg;
    AmAddHandler( hWnd, msgCode, window->getHWND(),
    		zoom2xEvents, VECLEN( zoom2xEvents),
		( AmNtHandlerFn) zoom2xHandler,
	        AmMakeDispMgrArgs( displayManager, coords, outline), 
		ZOOM_2X_PRIORITY,
		( AmStateHandlerFn) removeZoom2xHandler,
		( AmStateHandlerFn) suspendZoom2xHandler,
		( AmStateHandlerFn) restartZoom2xHandler);
}

/***************************************************************************
 *
 * Description:
 * 	Fully unzoom the object to its extents.  Any outstanding
 *	event handlers at or below UNZOOM_PRIORITY are removed.
 * Arguments:
 *	hWnd: The control that launches the callback (there are problems
 *	with this...)
 *	iMsg, wParam, lParam: the message id that launched the zoom callback.
 *	displayManager: display manager the object is displayed on.
 *	coords: The object to unzoom.
 *
 ***************************************************************************/

void AmUnzoomCallback( HWND			hWnd,
		     UINT			iMsg,
		     WPARAM			wParam,
		     LPARAM			lParam,
		     AmNTDisplayManagerClass	*displayManager,
		     AmCoordClass			*coords)
{
    float 		xmin, ymin, xmax, ymax;
    AmNTWindowClass	*pWindow  = ( AmNTWindowClass *) 
    				displayManager->GetWindow();

    displayManager->WriteResourceMessage( "unzoomSuccessMsg");
    AmFlushHandler( pWindow->getHWND(), UNZOOM_PRIORITY);
    coords->GetRange( &xmin, &ymin, &xmax, &ymax);
    coords->SetCoords( xmin, ymin, xmax, ymax);
    displayManager->Flush();
}

/***************************************************************************
 *
 * Description:
 * 	Unzoom the object to its half its current extents.  Any outstanding
 *	event handloers at or below UNZOOM_PRIORITY are removed.
 * Arguments:
 *	hWnd: The control that launches the callback (there are problems
 *	with this...)
 *	iMsg, wParam, lParam: the message id that launched the zoom callback.
 *	displayManager: display manager the object is displayed on.
 *	coords: The object to unzoom.
 *
 ***************************************************************************/

void AmUnzoomHalfCallback( HWND			hWnd,
			 UINT			iMsg,
			 WPARAM			wParam,
		         LPARAM			lParam,
			 AmNTDisplayManagerClass	*displayManager,
			 AmCoordClass		*coords)
{
    float		xmin, ymin, xmax, ymax,
			currXmin, currYmin, currXmax, currYmax,
			newXmin, newYmin, newXmax, newYmax;
    AmNTWindowClass	*pWindow  = ( AmNTWindowClass *) 
    				displayManager->GetWindow();

    AmFlushHandler( pWindow->getHWND(), UNZOOM_PRIORITY);

    coords->GetRange( &xmin, &ymin, &xmax, &ymax);
    coords->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
    newXmin = (currXmax + currXmin)/2 - (currXmax - currXmin);
    newYmin = (currYmax + currYmin)/2 - (currYmax - currYmin);
    newXmax = (currXmax + currXmin)/2 + (currXmax - currXmin);
    newYmax = (currYmax + currYmin)/2 + (currYmax - currYmin);
    if ( newXmin < xmin) {
        newXmin = xmin;
    }
    if ( newYmin < ymin) {
        newYmin = ymin;
    }
    if ( newXmax > xmax) {
        newXmax = xmax;
    }
    if ( newYmax > ymax) {
        newYmax = ymax;
    }
    coords->SetCoords( newXmin, newYmin, newXmax, newYmax);
    displayManager->WriteResourceMessage( "unzoomHalfSuccessMsg");
    displayManager->Flush();
}

/***************************************************************************
 *
 * Description:
 * 	Refresh all of the objects managed by the display manager.
 * Arguments:
 *	hWnd: The control that launches the callback (there are problems
 *	with this...)
 *	iMsg, wParam, lParam: the message id that launched the zoom callback.
 *	displayManager: display manager the object is displayed on.
 *
 ***************************************************************************/

void AmRefreshCallback( HWND			hWnd,
		      UINT			iMsg,
		      WPARAM			wParam,
		      LPARAM			lParam,
		      AmNTDisplayManagerClass	*displayManager)
{
    displayManager->Draw( NULL, 0);
    displayManager->WriteResourceMessage( "refreshMsg");
}
