#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* display bargraphs, where y values are shown as heights for bars */

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <ctype.h>
#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/resource.h>
#include <anvil/log.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/list.h>
#include <anvil/display.h>
#include <anvil/bargraph.h>
#include <anvil/tagstrs.h>
#include <anvil/fns.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean AmBarGraphClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmBAR_GRAPH) {
        return( TRUE);
    } else {
        return( AmGraphClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Empty Constructor so bar graphs can be subclassed.
 *
 ****************************************************************************/

AmBarGraphClass::AmBarGraphClass() 
{
}

/****************************************************************************
 *
 * Description:
 *      AmBarGraphClass constructor.  Creates a bar graph using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the bar graph.
 * Arguments:
 *      windowArg - window that the bar graph will be drawn on.
 *      nameArg - string name of the bar graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmBarGraphClass::AmBarGraphClass( AmWindowClass  *windowArg, 
			      const char   *nameArg,
			      ...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    AmBarGraphClass::initializeFields( windowArg, nameArg);
    objectType = AmBAR_GRAPH;
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    coords.currXmin = coords.xmin;
    coords.currYmin = coords.ymin;
    coords.currXmax = coords.xmax;
    coords.currYmax = coords.ymax;
}

/****************************************************************************
 *
 * Description:
 *      AmBarGraphClass constructor.  Creates a bar graph as part of an
 *      overlay.  The proprerties for the bar graph are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the bar graph will be added to.
 *      nameArg - string name of the bar graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmBarGraphClass::AmBarGraphClass( AmOverlayClass	*overlayArg,
			      const char	*nameArg,
			      ...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* initialize the fields for a bar graph and set the object type */

    AmBarGraphClass::initializeFields( overlayArg->GetWindow(), nameArg);
    objectType = AmBAR_GRAPH;

/* parse the tag-value pairs, set the fields and add the layer to the overlay
   if it's valid */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a bar graph.  call the
 *      AmGraphClass initialize fieldsm then set the bar graph specific fields
 * Arguments:
 *      windowArg - window that the line graph will be drawn on.
 *      nameArg - string name of the line graph, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmBarGraphClass::initializeFields( AmWindowClass  *windowArg,
			         const char   *nameArg)
{
    AmGraphClass::initializeFields( windowArg, nameArg);

/* set all the bar graph attributes */

    cmapName = NULL;
    barSpacing = 0;
    barHeight = 20;
    barStyle = AmBAR_HEIGHT;

/* set the fields for the color and string index */

    AmInitIndexData( &colorIndex);
}
/****************************************************************************
 *
 * Description:
 *      check values for a bar graph.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmBarGraphClass::checkValues()
{
    if ( !AmGraphClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmBarGraphClass destructor.
 *
 ****************************************************************************/

AmBarGraphClass::~AmBarGraphClass()
{    
    AmDestroyIndexData( &colorIndex);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for bar graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are retrieved, 
 *	then the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The colors bin vector is retrieved by hand
 *	because it's 1 less than the length of the colors array.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmBarGraphClass::getFields( AmTagValuePtr	tagValues,
			  int		numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, numUnusedFields, ii;

/* get the fields for the parent class ( graph), then get the fields for
   this class */

    numValues = AmGraphClass::getFields( tagValues, numValues);
    colorIndex.length = numColors;
    AmBarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( AmMatchTag( tagValues[ ii].tag, AmNcolorsBinvec)) {
            if ( colorIndex.binvec != NULL) {
                * ( void **) tagValues[ ii].value =
                    AmCopyVec( colorIndex.binvec,
                        ( colorIndex.length - 1) * sizeof( float));
            } else {
               * ( void **) tagValues[ ii].value = NULL;
            }
        } else {
            tagValues[ numUnusedFields++] = tagValues[ ii];
        }
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a bar graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The colorIndex bin vector is set by hand
 *	because it's 1 less in length than the colors array.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmBarGraphClass::setFields( AmTagValuePtr	tagValues,
			  int		numValues)
{
    int			ii, numFieldDefs, numUnusedFields;
    AmFieldDefPtr		fieldDefs;
    
/* get the fields for the parent class ( Data graphic object) */

    numValues = AmGraphClass::setFields( tagValues, numValues);
    colorIndex.length = numColors;

    AmBarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNcolorsBinvec)) {
	    AmSetBinvec( &colorIndex.binvec, colorIndex.length, &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    } 
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 * 	once the fields in the bar graph have been set using setFields(),
 *	update the index data structure 
 *
 ****************************************************************************/

void 
AmBarGraphClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmGraphClass::updateFields();
    if ( IsValid()) {
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colors") != AmOK) {	
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* set the yFn to be the same as the heightIndex function */

	dataInfo.yFn = heightIndex.fn;

/* perform the xy ranging, if there's an extents function to set the range */

        if ( dataInfo.sourceType != AmREAD_SOURCE) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }
    }
}

/****************************************************************************
 *
 * Description:
 *      Draw the bar graph to the screen. 
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK 
 *
 ****************************************************************************/

AmErrorCode
AmBarGraphClass::draw( AmWindowClass        *window,
                     AmGCClass            *gc,
                     AmRectangleRec const *exposeRects,
		     int		nRects)
{
    AmRectangleRec const  *dispRect = GetRect();
    int			barWidth, barHeight=0, ii, index, lastBarX, barX;
    float		scale, barTotal, barTab, zval;
    uchar_t		*ptr;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) { 
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }
 
/* if there's no overlay data, then we clear the area that we're going to
   draw to.  We have to use FillRectangle(), because we have to pay
   attention to the gc's clip rectangle which is set by the parent */
 
    if (( parent == NULL) && opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, exposeRects[ ii].x, 
						exposeRects[ ii].y,
						exposeRects[ ii].width, 
						exposeRects[ ii].height);
	}
    }
    gc->SetForegroundColor( &fgColor);
    if ( barStyle == AmBAR_HEIGHT) {

/* show evenly spaced bars of the appropriate height */

	barWidth = ( dispRect->width - ( ( dataInfo.dims[ 0] - 1)*barSpacing))/
			dataInfo.dims[ 0];
	scale = dispRect->height/( heightIndex.max - heightIndex.min);
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    if ( colors != NULL) {
		index = ( colorIndex.getIndexFn)( ptr, &colorIndex);
		if (( index >= 0) && ( index < colorIndex.length)) {
		    gc->SetForegroundColor( &colors[ index]);
		}
	    }
	    zval = ( heightIndex.fn)( ptr + heightIndex.offset, 
				      heightIndex.clientData);
	    barHeight = ( int)( ( zval - heightIndex.min)*scale);
	    display->FillRectangle( window, gc, 
				    dispRect->x + ii*( barWidth + barSpacing),
				    dispRect->y + dispRect->height - barHeight,
				    barWidth, barHeight);
	}
    } else if ( barStyle == AmBAR_WIDTH) {

/* find the total width of the graph */

	barTotal = 0.0;
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    barTotal += ( heightIndex.fn)( ptr + heightIndex.offset, 
					   heightIndex.clientData);
	}

/* values are a succession of widths.  In order for the user to see anything,
   the colors have to be set */

	lastBarX = 0;
	barTab = 0.0;
	scale = dispRect->width/barTotal;
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    if ( colors != NULL) {
		index = ( colorIndex.getIndexFn)( ptr, &colorIndex);
		gc->SetForegroundColor( &colors[ index]);
	    }
	    barTab += ( heightIndex.fn)( ptr + heightIndex.offset, 
					 heightIndex.clientData);
	    barX = ( int) ( scale*barTab);
	    display->FillRectangle( window, gc,
				    dispRect->x + lastBarX, 
				    dispRect->y + dispRect->height - barHeight,
				    barX - lastBarX, barHeight);
	}
    } else if ( barStyle == AmBAR_WIDTH_AND_HEIGHT) {
	AN_DEBUG( 1, ( "currently unimplemented"));
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw the current selection set.  The selection set is a buffer
 *      containing copies of the selected bars.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the bar graph's invalid.
 *
 ****************************************************************************/

AmErrorCode
AmBarGraphClass::DrawSelection( AmWindowClass        *window,
			      AmGCClass            *gc,
			      AmRectangleRec const *exposeRects,
			      int		 nRects,
			      AnvilBoolean		 fHighlight)
{
    AmRectangleRec const  *dispRect = GetRect();
    int			barWidth, barHeight=0, ii, lastBarX, barX, nSel;
    float		scale, barTotal, barTab, zval;
    uchar_t		*ptr, *selectionData;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) { 
        DO_RETURN_ERROR( AmNOT_VALID);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }
    GetSelection(( void **) &selectionData, &nSel);
    if ( nSel <= 0) {
	return( AmOK);
    }
    if ( fHighlight) {
	gc->SetForegroundColor( &highlightColor);
    } else {
	gc->SetForegroundColor( &fgColor);
    }
    if ( barStyle == AmBAR_HEIGHT) {

/* show evenly spaced bars of the appropriate height */

	barWidth = ( dispRect->width - ( ( dataInfo.dims[ 0] - 1)*barSpacing))/
			dataInfo.dims[ 0];
	scale = dispRect->height/( heightIndex.max - heightIndex.min);
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    zval = ( heightIndex.fn)( ptr + heightIndex.offset, 
				      heightIndex.clientData);
	    barHeight = ( int)( ( zval - heightIndex.min)*scale);
	    display->FillRectangle( window, gc, 
				    dispRect->x + ii*( barWidth + barSpacing),
				    dispRect->y + dispRect->height - barHeight,
				    barWidth, barHeight);
	}
    } else if ( barStyle == AmBAR_WIDTH) {

/* find the total width of the graph */

	barTotal = 0.0;
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    barTotal += ( heightIndex.fn)( ptr + heightIndex.offset, 
					   heightIndex.clientData);
	}

/* values are a succession of widths.  In order for the user to see anything,
   the colors have to be set */

	lastBarX = 0;
	barTab = 0.0;
	scale = dispRect->width/barTotal;
	for ( ii = 0, ptr = dataInfo.data; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.size) {
	    barTab += ( heightIndex.fn)( ptr + heightIndex.offset, 
					 heightIndex.clientData);
	    barX = ( int) ( scale*barTab);
	    display->FillRectangle( window, gc,
				    dispRect->x + lastBarX, 
				    dispRect->y + dispRect->height - barHeight,
				    barX - lastBarX, barHeight);
	}
    } else if ( barStyle == AmBAR_WIDTH_AND_HEIGHT) {
	AN_DEBUG( 1, ( "currently unimplemented"));
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a bar graph.  This is
 *      used in the map editor to dump the graph's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmBarGraphClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			   AnvilBoolean	     allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( graph) */

    numStrings = AmGraphClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmBarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
				     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a bar graph. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmImagePlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  THE LIST MUST BE SORTED BY 
 *	ALPHABETICAL ORDER ON THE FIELD NAME. 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmBarGraphClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDef,
				    int		*numFieldDef)
{
    static AmFieldDefRec BarGraphClassFieldDefinitions[] = {
	{ AmNbarHeight, offsetof( AmBarGraphClass, barHeight) },
	{ AmNbarSpacing, offsetof( AmBarGraphClass, barSpacing) },
	{ AmNbarStyle, offsetof( AmBarGraphClass, barStyle) },
	{ AmNcolorsClientData, 
	  offsetof( AmBarGraphClass, colorIndex.clientData) },
	{ AmNcolorsClientDataFreeFn, 
	  offsetof( AmBarGraphClass, colorIndex.clientDataFreeFn) },
	{ AmNcolorsFn, offsetof( AmBarGraphClass, colorIndex.fn) },
	{ AmNcolorsIntFn, offsetof( AmBarGraphClass, colorIndex.AmIntFn) },
	{ AmNcolorsMax, offsetof( AmBarGraphClass, colorIndex.max) },
	{ AmNcolorsMin, offsetof( AmBarGraphClass, colorIndex.min) },
	{ AmNcolorsName, 
	  offsetof( AmBarGraphClass, colorIndex.name), 0, AmDYNAMIC },
	{ AmNcolorsOffset, 
	  offsetof( AmBarGraphClass, colorIndex.offset) },
	{ AmNcolorsStrings, offsetof( AmBarGraphClass, colorIndex.strings),
			    offsetof( AmBarGraphClass, colorIndex.length), 
			    AmDYNAMIC },
	{ AmNcolorsTable, offsetof( AmBarGraphClass, colorIndex.table),
			  offsetof( AmBarGraphClass, colorIndex.length), 
			  AmDYNAMIC },
	{ AmNcolorsType, 
	  offsetof( AmBarGraphClass, colorIndex.dataType) },
	{ AmNcolorsUnits, offsetof( AmBarGraphClass, colorIndex.units) },
    };
    *fieldDef = BarGraphClassFieldDefinitions;
    *numFieldDef = VECLEN( BarGraphClassFieldDefinitions);
}
