#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#if !defined(__BSD44__)
#include <search.h>
#endif
#include <string.h>
#include <stdio.h>
#include <anvil/cache.h>
#include <anvil/tsearch.h>
#include <anvil/errcode.h>

/* the data cache file is used by all data graphics so that files aren't
   re-loaded when they're being redisplayed.  It maintains a global tree
   which maps filenames to data pointers */

/* structure and root pointer for data/filename caching. */

typedef struct dataCache_ {
    char        *filename;
    void        *data;
    int         refCnt;
} dataCacheRec, *dataCachePtr;

static void *dataCacheRoot = NULL;

/***************************************************************************
 *
 * Description:
 * 	comparison function for the data cache.  If the filenames are NULL, 
 *	then test the data pointers, otherwise test the filenames.
 * Arguments:
 *	a,b - dataCachePtr's to compare.
 *
 ***************************************************************************/
  

static int 
compareDataCache( dataCachePtr       a,
		  dataCachePtr       b)
{
    if ( a->filename == NULL) {
        if ( b->filename == NULL) {
            return(( uchar_t *) a->data - ( uchar_t *) b->data);
        } else {
            return( -1);
        }
    } else if ( b->filename == NULL) {
        return( 1);
    } else {
        return( strcmp( a->filename, b->filename));
    }
}

/***************************************************************************
 *
 * Description:
 *  	If the filename is in the cache, return the associated data pointer,
 *	If not found, return NULL.
 * Arguments:
 *	filename - filename to search for in the cache.
 * Returns:
 *	returns the associated data pointer.
 *
 ***************************************************************************/

void *
AmGetCache( char    *filename)
{
    dataCacheRec        target;
    dataCachePtr        *result;

    if ( filename == NULL) {
	return( NULL);
    }

/* find out if the data is in the cache already */

    target.filename = filename;
    result = ( dataCachePtr *) tfind( &target, 
				      ( const void **) &dataCacheRoot,
                                      ( AmSearchFn) compareDataCache);
    if ( result != NULL) {
        return(( *result)->data);
    } else {
        return( NULL);
    }
}

/***************************************************************************
 *
 * Description:
 *  	If the filename is in the cache, return the associated data pointer,
 *	If not found, return NULL.  same as AmGetCache, except that the
 *	reference count pointer isn't incremented.
 * Arguments:
 *	filename - filename to search for in the cache.
 * Returns:
 *	returns the associated data pointer.
 *
 ***************************************************************************/

AnvilBoolean 
AmFindCache( char    	*filename,
	     void	**ptr,
	     int	*refCnt)
{
    dataCacheRec        target;
    dataCachePtr        *result;

    if ( filename == NULL) {
	return( FALSE);
    }

/* find out if the data is in the cache already */

    target.filename = filename;
    result = ( dataCachePtr *) tfind( &target, 
				      ( const void **) &dataCacheRoot,
                                      ( AmSearchFn) compareDataCache);
    if ( result != NULL) {
	*ptr = ( *result)->data;
	*refCnt = ( *result)->refCnt;
AN_DEBUG( 20, ("AmFindCache:reference count %s %d\n", filename, ( *result)->refCnt));
        return( TRUE);
    } else {
        return( FALSE);
    }
}

/***************************************************************************
 *
 * Description:
 * 	add a file and the associated data pointer to the cache.
 * Arguments:
 *	filename - file to add to the cache.
 *	data - associated data segment.
 * Returns:
 *	AmOK if everything's OK, AmMALLOC_FAILED if the node couldnt'
 *	be allocated.
 *
 ***************************************************************************/

AmErrorCode 
AmAddToCache( char    *filename,
	      void    *data)
{
    dataCacheRec        target;
    dataCachePtr        *result, newNode;

    if ( filename == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* find out if the data is in the cache already */

    target.filename = filename;
    target.data = data;
    result = ( dataCachePtr *) tfind( &target, 
				      ( const void **) &dataCacheRoot,
                                      ( AmSearchFn) compareDataCache);
    if ( result != NULL) {
        ( *result)->refCnt++;
	AN_DEBUG( 20, ( "addToCache: reference count %s %d\n",
			filename, ( *result)->refCnt));
    } else {
        newNode = ( dataCachePtr) malloc( sizeof( dataCacheRec));
        if ( newNode == NULL) {
            DO_RETURN_ERROR( AmMALLOC_FAILED);
        }
        newNode->filename = strdup( filename);
	if ( newNode->filename == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
        newNode->refCnt = 1;
        newNode->data = data;
	AN_DEBUG( 20, ( "addToCache: reference count %s %d\n",
			filename, newNode->refCnt));
        tsearch( newNode, &dataCacheRoot, ( AmSearchFn) compareDataCache);
    }
    return( AmOK);
}

/***************************************************************************
 *
 * Description:
 * 	remove a file and associated data pointer from the cache 
 * Arguments:
 *	filename - file to remove to the cache.
 *	data - associated data segment.
 * Returns:
 *	AmOK if everything's OK, AmBAD_ARGUMENT if the filename was NULL
 *	or if it couldn't be found.
 *
 ***************************************************************************/

AmErrorCode 
AmRemoveFromCache( char       *filename,
		   void       *data)
{
    dataCacheRec        target;
    dataCachePtr        *result;
    void		*dataPtr;
    char		*filenamePtr;

    if ( filename == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* find out if the data is in the cache and if the reference count
   is greater than 0 */

    target.filename = filename;
    target.data = data;
    result = ( dataCachePtr *) tfind( &target, 
				      ( const void **) &dataCacheRoot,
                                      ( AmSearchFn) compareDataCache);
    if ( result != NULL) {
AN_DEBUG( 20, ( "AmRemoveFromCache:reference count %s %d\n", 
		filename, ( *result)->refCnt));
        ( *result)->refCnt--;
        if (( *result)->refCnt == 0) {
	    dataPtr = ( *result)->data;
	    filenamePtr = ( *result)->filename;
            tdelete( *result, &dataCacheRoot, ( AmSearchFn) compareDataCache);
            free( dataPtr);
            free( filenamePtr);
        }
        return( AmOK);
    } else {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
}

/***************************************************************************
 *
 * Description:
 * 	remove the entry from the cache, but leave the responsibility
 *	to the caller to delete the actual data pointed to.
 * Arguments:
 *	filename - file to remove to the cache.
 *	data - associated data segment.
 * Returns:
 *	AmOK if everything's OK, AmBAD_ARGUMENT if the filename was NULL
 *	or if it couldn't be found.
 *
 ***************************************************************************/

AmErrorCode 
AmRemoveFromCacheTable( char       *filename,
		        void       *data)
{
    dataCacheRec        target;
    dataCachePtr        *result;
    char		*ptr;

    if ( filename == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* find out if the data is in the cache and if the reference count
   is greater than 0 */

    target.filename = filename;
    target.data = data;
    result = ( dataCachePtr *) tfind( &target, 
				      ( const void **) &dataCacheRoot,
                                      ( AmSearchFn) compareDataCache);
    if ( result != NULL) {
AN_DEBUG( 20, ( "AmRemoveFromCacheTable:reference count %s %d\n", 
		filename, ( *result)->refCnt));
        ( *result)->refCnt--;
        if (( *result)->refCnt == 0) {
	    ptr = ( *result)->filename;
            tdelete( *result, &dataCacheRoot, ( AmSearchFn) compareDataCache);
            free( ptr);
        }
        return( AmOK);
    } else {
        DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
}
