#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <anvil/macros.h>
#include <anvil/plot.h>
#include <anvil/overlay.h>
#include <anvil/cadplot.h>
#include <anvil/cad/segment.h>
#include <anvil/cad/segmentProto.h>
#include <anvil/cad/object.h>
#include <anvil/cad/objectProto.h>
#include <anvil/cad/mem.h>

#define POLYGON_OUTLINED TRUE

#define NUM_SEGMENTS	1000
#define OBJ_SIZE	10000
#define CACHE_SIZE	5
#define HASH_SIZE	23

#define NUM_POINTS	2048

#define GET_POINTS( p)                                       \
    ( long *) ((( uchar_t *) p) + sizeof( AmCadPolyLineRec))
    
#define GET_LINES( p)                                        \
    ( long *) ((( uchar_t *) p) + sizeof( AmCadPolygonRec))
    
#define GET_REFERENCES( p)                                   \
    ( long *) ((( uchar_t *) p) + sizeof( AmCadRefListRec))
    
typedef enum locationEnum_ {
    AmLOCATION_ERROR = 0,
    AmLOCATION_OUTSIDE = 1,
    AmLOCATION_ABOVE = 2,
    AmLOCATION_BELOW = 3,
    AmLOCATION_ON = 4
} locationEnum;


static AmPointRec      pointCache[ NUM_POINTS];

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/
/*                                                                         */
/* Private drawing functions:                                              */
/*                                                                         */
/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

static AmSegmentRec segments[ MAXLINES];
static long       nSegments = 0;

static void
drawLineCache( AmDisplayClass *display,
	       AmWindowClass  *window,
	       AmGCClass      *gc,
	       int          x0,
	       int          y0,
	       int          x1,
	       int          y1)
{

    if ( x0 != x1 || y0 != y1) {
    
	/* Set the segment's endpoints and draw if the buffer is exceeded */
	
	segments[ nSegments].x0 = x0;
	segments[ nSegments].y0 = y0;
	segments[ nSegments].x1 = x1;
	segments[ nSegments].y1 = y1;
	( nSegments)++;
	
	if ( nSegments == MAXLINES) {
	    display->DrawSegments( window, gc, segments, nSegments);
	    nSegments = 0;
	}
    }
}

static void
flushLineCache( AmDisplayClass *display,
		AmWindowClass  *window,
		AmGCClass      *gc)
{
    if ( nSegments != 0) {
	display->DrawSegments( window, gc, segments, nSegments);
	nSegments = 0;
    }
}


/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/
/*                                                                         */
/* Private Utility functions:                                              */
/*                                                                         */
/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: AmListPtr filterListByList( lstSource, lstFilter)
 *
 * Paramters:
 *   AmListPtr lstSource: The list of items to filter.
 *   AmListPtr lstFilter: The items to filter the source by.
 *
 * Description:
 *   filterListByList removes any items in lstSource that are also
 *   in lstFilter.  The result is a list of elements unique to
 *   lstSource.
 *
 *   filterListByList changes lstSource so the original structure
 *   is NOT retained.
 *
 * Returns:
 *   AmListPtr: The new header of lstSource.
 */

static AmListPtr
filterListByList( AmListPtr lstSource, AmListPtr lstFilter)
{
    AmListPtr lstPrev, lstISource, lstIFilter;
    long    idSource, idFilter;
    
    lstPrev = NULL;
    lstISource = lstSource;

    /* Loop through the elements in the source. */
    
    while ( lstISource != NULL) {
       idSource = ( long) lstISource->data;

       /* Loop through the elements in the filter */
       
       for ( lstIFilter = lstFilter;
	     lstIFilter != NULL;
	     lstIFilter = lstIFilter->next) {
	   
	   /*
	    * If the source element matches an element in the filter,
	    *   it needs to be removed.
	    */
	     
	   idFilter = ( long) lstIFilter->data;
	   
	   if ( idSource == idFilter) {
	       if ( lstPrev != NULL) {
		   lstPrev->next = lstISource->next;
		   free( lstISource);
		   lstISource = lstPrev->next;
	       } else {
		   lstSource = lstISource->next;
		   free( lstISource);
		   lstISource = lstSource;
	       }
	       break;
	   }
       }

       /* If the element wasn't filtered out, continue to the next element. */
       
       if ( lstIFilter == NULL) {
	   lstPrev = lstISource;
	   lstISource = lstISource->next;
       }
    }

    return( lstSource);
}

/*
 * Function: AmListPtr filterListByItem( lstSource, idFilter)
 *
 * Paramters:
 *   AmListPtr lstSource: The list of items to filter.
 *   long    idFilter : The item to filter the source by.
 *
 * Description:
 *   filterListByItem removes any occurence of idFilter from the
 *   source lstSource.
 *
 *   filterListByList changes lstSource so the original structure
 *   is NOT retained.
 *
 * Returns:
 *   AmListPtr: The new header of lstSource.
 */

static AmListPtr
filterListByItem( AmListPtr lstSource, long idFilter)
{
    AmListPtr lstPrev, lstISource;
    long    idSource;
    
    lstPrev = NULL;
    lstISource = lstSource;

    /* Loop through the elements in the source. */
    
    while ( lstISource != NULL) {
       idSource = ( long) lstISource->data;

       if ( idSource == idFilter) {
	   if ( lstPrev != NULL) {
	       lstPrev->next = lstISource->next;
	       free( lstISource);
	       lstISource = lstPrev->next;
	   } else {
	       lstSource = lstISource->next;
	       free( lstISource);
	       lstISource = lstSource;
	   }
       } else {
       	   lstPrev = lstISource;
	   lstISource = lstISource->next;
       }
    }

    return( lstSource);
}

/*
 * Function: AmListPtr addUniqueLst( lstSource, id)
 *
 * Paramters:
 *   AmListPtr lstSource: The list of items to filter.
 *   long    id       : The item to insert.
 *
 * Description:
 *   addUniqueLst inserts an item at the end of the list if and only
 *   if an item with the same value does not already exist.
 
 *   addUniqueLst changes lstSource so the original structure
 *   is NOT retained.
 *
 * Returns:
 *   AmListPtr: The new header of lstSource.
 */

static AmListPtr
addUniqueLst( AmListPtr lstSource, long id)
{
    AmListPtr lstPrev, lstISource;
    long    idSource;
    
    lstPrev = NULL;
    lstISource = lstSource;

    /* Loop through the elements in the source. */
    
    while ( lstISource != NULL) {
       idSource = ( long) lstISource->data;

       if ( idSource == id) {
	   break;
       }

       lstPrev = lstISource;
       lstISource = lstISource->next;
    }

    if ( lstISource == NULL) {
	lstISource = ( AmListPtr) malloc( sizeof( AmListRec));

	lstISource->data = ( void *) id;
	lstISource->next = NULL;

	if ( lstPrev != NULL) {
	    lstPrev->next = lstISource;
	} else {
	    lstSource = lstISource;
	}
    }

    return( lstSource);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: AnvilBoolean lineOutsideRectangles( x0, y0, x1, y1,
 *                                          clipRects, nClipRects)
 *
 * Parameters:
 *   int          x0        : X coordinate for 1st point in line.
 *   int          y0        : Y coordinate for 1st point in line.
 *   int          x1        : X coordinate for 2nd point in line.
 *   int          y1        : Y coordinate for 2nd point in line.
 *   AmRectangleRec *clipRects: Rectangles to set for trivial intersection.
 *   int          nClipRects: Number of rectangles.
 *
 * Description:
 *   lineOutisdeRectangles tests whether a lines endpoints (x0, y0) and
 *   (x1, y1) are outside of the rectangles given by clipRects.  The
 *   function only performs a trivial check.  That is it only checks
 *   whether the two points are on one side of the rectangle.  i.e.
 *   it checks if both points are either above, below, left of, or right
 *   of the rectangles.
 *
 * Returns:
 *   AnvilBoolean: TRUE when the line is outside the rectangles, FALSE otherwise.
 *            Note that FALSE doesn't mean the line intersects any of the
 *            rectangles, it just means the trivial check didn't succeed.
 *
 *
 * NOTE: This needs to be changed so that it doesn't use integer
 *       coordinates, but floating point real world coordinates.
 *       Then before calling this function the clipRects need to
 *       be translated to real world coordinates also.
 */

static AnvilBoolean
lineOutsideRectangles( int          x0,
		       int          y0,
		       int          x1,
		       int          y1,
		       AmRectangleRec *clipRects,
		       int          nClipRects)
{
    int ii, left, right, top, bottom;

    for ( ii = 0; ii < nClipRects; ii++) {

	left = clipRects[ ii].x;
	right = left + clipRects[ ii].width;
	top = clipRects[ ii].y;
	bottom = top + clipRects[ ii].height;

	/*
	 * Check for trivial reject.  If both points are on the same side
	 *   (i.e. left, right, above, or below) of the rectangle then
	 *   the line is outside of the rectangle and we can go check the
	 *   next rectangle.
	 */
	
	if (( x0 < left && x1 < left) ||
	    ( x0 > right && x1 > right) ||
	    ( y0 < top && y1 < top) ||
	    ( y0 > bottom && y1 > bottom)) {
	    continue;
	}

	/*
	 * Otherwise, the line might intersect the rectangle.  More
	 *   intersection testing should be done here.
	 */
	
	break;
    }

    /*
     * If we have gone through the entire loop then the line is outside
     *   of all rectangles.  Otherwise, there was an intersection.
     */
    
    return( ii == nClipRects);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: getParentObjects( memory, objectID, objects)
 *
 * Paramters:
 *   AmMemoryPtr memory  : The CAD data.
 *   long      objectID: The object to find the parents for.
 *   AmListPtr   objects : The list to place the parent IDs into.
 *
 * Description:
 *   Get the objects that are parents of the object given by objectID.
 *   The result is placed at the end of the objects list.
 *
 * Returns:
 *   AmListPtr : The expanded list of parent objects.
 */

static AmListPtr
getParentObjects( AmMemoryPtr memory,
		  long      objectID,
		  AmListPtr   objects)
{
    AmListPtr       temp;
    AmCadRefListPtr refList;
    long          *refs;
    int           ii;

    for ( temp = objects; temp != NULL; temp = temp->next) {
        if ( objectID == ( long) temp->data) {
	    return( objects);
	}
    }

/* Get the reference list for the object. */

    GetReferenceListLOCAL( memory, objectID, refList);
    if ( refList == NULL) {

	/* No references means this is a parent object. */
	
	objects = AmInsertNodeLst( ( void *) objectID, objects);
	return ( objects);
    }

    /*
     * A reference list exists, so this is not a parent object.  Instead
     *  get the parents for all the objects that are referencing this
     *  object.
     */
    
    refs = GET_REFERENCES( refList);
    for ( ii = 0; ii < refList->numRefs; ii++) {
        objects = getParentObjects( memory, refs[ ii], objects);
    }

    return( objects);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: AnvilBoolean findObjectLst( objects, objectID)
 *
 * Parameters:
 *   AmListPtr objects : The list to search for the objectID. 
 *   long    objectID: The object to search for.
 *
 * Description:
 *   findObjectLst searches the objects list for the object given
 *   by objectID.
 *
 * Returns:
 *   AnvilBoolean: TRUE if the object exists in the list, FALSE otherwise.
 */

static AnvilBoolean
findObjectLst( AmListPtr objects, long objectID)
{
    AmListPtr temp;

    for ( temp = objects; temp != NULL; temp = temp->next) {
	if ( objectID == ( long) temp->data) {
	    return( TRUE);
	}
    }

    return( FALSE);
}


#if 0
/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: long numPointsInPolygon( memory, polygon)
 *
 * Parameters:
 *   AmMemoryPtr     memory:  The CAD data.
 *   AmCadPolygonPtr polygon: The polygon to find the # of points for.
 *
 * Description:
 *   numPointsInPolygon counts the number of points in a polygon.
 *
 * Returns:
 *   long: The number of points.
 */

static long
numPointsInPolygon( AmMemoryPtr memory, AmCadPolygonPtr polygon)
{
    int            ii;
    long           numPoints,*lines;
    AmCadPolyLinePtr polyLine;
    
    /* Initializations */
    
    numPoints = 0;
    lines = GET_LINES( polygon);

    /* Find the number of points in each polyline. */
    
    for ( ii = 0; ii < polygon->AmNumLines; ii++) {
	polyLine = GET_POLYLINE( AmGetData( memory, lines[ ii]));
	numPoints += polyLine->numPoints;
    }
    
    return( numPoints);
}
#endif

/*
 * Function: AmErrorCode getPolygonPoints( memory, polygon, points, numPoints)
 *
 * Parameters:
 *   AmMemoryPtr     memory    : The CAD data.
 *   AmCadPolygonPtr polygon   : The polygon to get the points for.
 *   long          **points  : The point array return variable.
 *   long          *numPoints: The number of points return variable.
 *
 * Description:
 *   getPolygonPoints retrieves the points from a polygon and stuffs them
 *   into a newly allocated array.  The array and the number of points in
 *   the array are passed out through the points and numPoints arguments
 *   respectively.
 *
 * Returns:
 *   AmErrorCode: AmOK if everything went fine, the error code otherwise.
 *
 *   Also: long **points gets the point array.
 *         long *numPoints gets the number of points in the array. 
 *
 * Errors:
 *   AmMALLOC_FAILED
 *
 */

static AmListPtr
getPolygonPoints( AmMemoryPtr     memory,
		  AmCadPolygonPtr polygon)
{
    AmListPtr        lstTemp;
    int            ii, jj;
    long           *points, *lines;
    AmCadPolyLinePtr polyLine;
    
    /* Initializations */

    lstTemp = NULL;
    
    lines = GET_LINES( polygon);

    for ( ii = 0; ii < polygon->AmNumLines; ii++) {
	polyLine = GET_POLYLINE( AmGetData( memory, lines[ ii]));

	points = GET_POINTS( polyLine);
	
	for ( jj = 0; jj < polyLine->numPoints; jj++) {
	    lstTemp = addUniqueLst( lstTemp, points[ jj]);
	}
    }
    
    return( lstTemp);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: locationEnum pointLocationLine( x, y, x0, y0, x1, y1)
 *
 * Parameters:
 *   float x : X coordinate of the point.
 *   float y : Y coordinate of the point.
 *   float x0: X coordinate of the AmFirst point in the line.
 *   float y0: Y coordinate of the AmFirst point in the line.
 *   float x1: X coordinate of the second point in the line.
 *   float y1: Y coordinate of the second point in the line.
 *
 * Description:
 *   pointLocatedLine checks where the point (x, y) is located with
 *   respect to the line (x0, y0) (x1, y1).  Note that the left
 *   most point on a line is treated as excluded, while the right
 *   most point is treated as included.  Also note that a vertical
 *   line is treated in a special way.  Anything that is not ON
 *   the line is treated as outside of the line.
 *
 * Returns:
 *   int: AmLOCATION_ON if the point is on the line.
 *        AmLOCATION_BELOW if the point is point is below the line.
 *        AmLOCATION_OUTSIDE if the point is not in an interesting
 *          region.
 *
 * Errors:
 *   -none-
 */

static locationEnum
pointLocationLine( float x,
		   float y,
		   float x0,
		   float y0,
		   float x1,
		   float y1)
{
    float temp, slope;

    /* Does the line overlap the point (along the x axis)? */
    
    if ((( x0 < x) && ( x1 >= x)) || (( x0 >= x) && ( x1 < x))) {
	slope = ( y1 - y0) / ( x1 - x0);
	temp = slope * ( x - x0) + y0;
	    
	if ( temp > y) {
	    return( AmLOCATION_BELOW);
	} else if ( temp == y) {
	    return( AmLOCATION_ON);
	}
    } else if (( x0 == x1) && ( x == x0)) {
	if ((( y0 < y) && ( y1 >= y)) || (( y0 >= y) && ( y1 < y))) {
	    return( AmLOCATION_ON);
	}
    }
    
    return( AmLOCATION_OUTSIDE);
}


/*
 * Function: AnvilBoolean pointInPolygon( memory, x, y, polygon)
 *
 * Parameters:
 *   AmMemoryPtr     memory    : CAD data.
 *   float         x         : X coordinate of the point.
 *   float         y         : Y coordinate of the point. 
 *   AmCadPolygonPtr polygon   : The polygon to test for inclusion.
 *
 * Description:
 *   pointInPolygon checks whether a point is inside a polygon or
 *   outside.  Any point on the lines of the polygon are treated
 *   as included.
 *
 * Returns:
 *   AnvilBoolean: TRUE if the point is in the polygon, FALSE otherwise.
 *
 * Errors:
 *   AmNULL_ARGUMENT
 *
 */

static AnvilBoolean
pointInPolygon( AmMemoryPtr     memory,
	        float         x,
		float         y,
		AmCadPolygonPtr polygon)
{
    int            ii, jj;
    long           *points, *lines, numPoints, numBelow;
    float          x0=0, y0=0, x1=0, y1=0, x2=0, y2=0;
    AmCadPolyLinePtr polyLine=NULL;
    AmCadPointPtr    point;
    locationEnum   location;
    
    /* Sanity checking. */
    
    if ( polygon == NULL) {
	DO_ERROR( AmNULL_ARGUMENT);
	return( FALSE);
    }
    
    /* Initializations. */

    numBelow = 0;
    numPoints = 0;

    lines = GET_LINES( polygon);

    /* Find the number of points in each polyline. */
    
    for ( ii = 0; ii < polygon->AmNumLines; ii++) {
	GetPolyLineLOCAL( memory, lines[ ii], polyLine);
	points = GET_POINTS( polyLine);

	for ( jj = 0; jj < polyLine->numPoints; jj++) {
	    point = GET_POINT( AmGetData( memory, points[ jj]));
	    x2 = point->x;
	    y2 = point->y;

	    if ( numPoints != 0) {
		location = pointLocationLine( x, y, x1, y1, x2, y2);
		if ( location == AmLOCATION_ON) {
		    return( TRUE);
		} else if ( location == AmLOCATION_BELOW) {
		    numBelow++;
		}
	    } else {
		x0 = x2;
		y0 = y2;
	    }

	    numPoints++;
	    
	    x1 = x2;
	    y1 = y2;
	}
    }

    if ( numPoints > 2) {
	location = pointLocationLine( x, y, x2, y2, x0, y0);
	if ( location == AmLOCATION_ON) {
	    return( TRUE);
	} else if ( location == AmLOCATION_BELOW) {
	    numBelow++;
	}
    }

    return( ODD( numBelow));
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: AmErrorCode initMemory()
 *
 * Parameters:
 *   -none-
 *
 * Description:
 *   initMemory attempts to load the cad data specified by the
 *   filename given in the dataInfo record.  If initMemory can't
 *   load the data it tries to create some empty memory.
 *
 * Returns:
 *   AmOK if sucessfull.  The error code otherwise.
 * 
 * Errors:
 *   AmMISSING_FILENAME;
 *   AmBAD_ARGUMENT
 *   AmNULL_ARGUMENT
 *   AmMALLOC_FAILED
 *   AmOPEN_FAILED
 *   AmREAD_FAILED
 */   

AmErrorCode
AmCadPlotClass::initMemory()
{
    if ( _memory == NULL) {
	
	/* Make sure there is a filename for the data. */
	
	if ( dataInfo.filename == NULL) {
	    DO_RETURN_ERROR( AmMISSING_FILENAME);
	}
	
	/* Try to load the file into memory. */

	_memory = AmLoadMemory( dataInfo.filename, OBJ_SIZE, CACHE_SIZE,
			      HASH_SIZE);

	/* If we still don't have any memory, try to create something. */
	
        if ( _memory == NULL) {
            _memory = AmMakeMemory( NUM_SEGMENTS, OBJ_SIZE, CACHE_SIZE,
				  HASH_SIZE, dataInfo.filename);
        }
	
        if ( _memory == NULL) {
	    return( ANVIL_ERROR_CODE);
	}
    }

    return( AmOK);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/* Get the list of parent objects for an object. */
    
AmListPtr
AmCadPlotClass::GetParents( long objectID)
{
    AmListPtr lstTemp = NULL;

    if ( initMemory() != AmOK) {
	return( NULL);
    }

    lstTemp = getParentObjects( _memory, objectID, lstTemp);
    lstTemp = filterListByItem( lstTemp, objectID); 

    return( lstTemp);
}

/* Get the parents for a collection of objects. */

AmListPtr
AmCadPlotClass::GetParents( AmListPtr lstObjects)
{
    AmListPtr lstResult, lstIObjects;

    if ( initMemory() != AmOK) {
	return( NULL);
    }
    
    /* Initializations. */
    
    lstResult = NULL;
    lstIObjects = lstObjects;
    
    /* Step through the list of objects and get the parents for each. */
    
    for ( lstIObjects = lstObjects;
	  lstIObjects != NULL;
	  lstIObjects = lstIObjects->next) {

	lstResult = getParentObjects( _memory,
				      ( long) lstIObjects->data,
				      lstResult);
    }

    /*
     * Filter the list so that none of the passed-in objects show up in the
     *   parent list.
     */
    
    lstResult = filterListByList( lstResult, lstObjects);
    
    return( lstResult);
}


/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/* Utility function. */

AnvilBoolean
AmCadPlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmCAD_PLOT) {
        return( TRUE);
    } else {
        return( AmPlotClass::IsType( type));
    }
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/* Initialize fields which are specific to a CAD plot. */

void
AmCadPlotClass::initializeFields( AmWindowClass *windowArg,
                                const char  *nameArg)
{
    AmPlotClass::initializeFields( windowArg, nameArg);
    objectType = AmCAD_PLOT;

    _memory = NULL;
    _fChanged = FALSE;
    _szCmdHistory = strdup( "/tmp/CAD.history.XXXXXX");
    mktemp( _szCmdHistory);
    
    /*
     * The data doesn't get read in a standard way.  The source type
     *   means that the plot is responsible for loading its own data.
     */
    
    dataInfo.sourceType = AmCUSTOM;

    lstAttTypes = NULL;
}

/****************************************************************************
 *
 * Description:
 *      check values for a CAD plot.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag. 
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmCadPlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
        return( FALSE);
    } else {
	return( TRUE);
    }
}

/* Create a standalone CAD plot. */

AmCadPlotClass::AmCadPlotClass( AmWindowClass *windowArg,
                            const char   *nameArg,
                            ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}


/* Create a CAD plot as a component of an overlay plot. */

AmCadPlotClass::AmCadPlotClass( AmOverlayClass *overlayArg,
                            const char   *nameArg,
                            ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

    /* Then override the fields for the overlay plot with the fields passed in
       for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}


/*
 * Function: ~AmCadPlotClass()
 *
 * Paramters:
 *   -none-
 *
 * Description:
 *   Destructor for the cad plot class.  Frees up any memory that was
 *   used by the cad plot.
 *
 * Returns:
 *   -nothing-
 */

AmCadPlotClass::~AmCadPlotClass()
{
    if ( _memory != NULL) {
	AmFreeMemory( _memory);
    }

    if ( _szCmdHistory != NULL) {
	unlink( _szCmdHistory);
	free( _szCmdHistory);
    }
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/*
 * Function: int setFields( tagValues, numValues)
 *
 * Parameters:
 *   AmTagValuePtr tagValues: A list of tags followed by the values to be
 *                            set.
 *   int         numValues: How many pairs there are.
 *
 * Description:
 *   setFields sets the fields for a cad plot according to the tag
 *   value pairs passed in.  The tag specifies which field is to be
 *   set, the corresponding value specifies what the field is to be
 *   to.
 *
 * Returns:
 *   int: The number of fields not set.
 */

int
AmCadPlotClass::setFields( AmTagValuePtr tagValues,
                         int         numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmCadPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmCadPlotClass::updateFields()
{
    AmPlotClass::updateFields();
}

/* Get the fields for a cad plot */

int
AmCadPlotClass::getFields( AmTagValuePtr tagValues,
                         int         numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmCadPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=*=-=-=*=-=*/

/* Low level point drawing function. */

AmErrorCode
AmCadPlotClass::drawPoint( AmWindowClass *window,
			 AmGCClass     *gc,
			 AmCadPointPtr point)
{
    int x, y;
    
    GetScreenPt( point->x, point->y, &x, &y);
    display->DrawRectangle( window, gc, x - 1, y - 1, 3, 3);

    return( AmOK);
}

/* Low level polyline drawing function. */

AmErrorCode
AmCadPlotClass::drawPolyLine( AmWindowClass        *window,
                            AmGCClass            *gc,
                            AmCadPolyLinePtr     polyLine)
{
    AmCadPointPtr point;
    long        *pts, ii;
    int         x0=0, y0=0, x1, y1;

    if ( polyLine->numPoints > 0) {
    
	pts = GET_POINTS( polyLine); 
	
	for ( ii = 0; ii < polyLine->numPoints; ii++) {
	    point = GET_POINT( AmGetData( _memory, pts[ ii]));
	    GetScreenPt( point->x, point->y, &x1, &y1);
  
	    if ( ii != 0 && (( x0 != x1) || ( y0 != y1))) {
		drawLineCache( display, window, gc, x0, y0, x1, y1);
	    }
	    
	    x0 = x1;
	    y0 = y1;
	}

	flushLineCache( display, window, gc);
    }

    return( AmOK);
}

/* Low level polygon drawing function. */

AmErrorCode
AmCadPlotClass::drawPolygon( AmWindowClass   *window,
                           AmGCClass       *gc, 
                           AmCadPolygonPtr polygon)
{
    AmCadPointPtr     point=NULL;
    AmCadPolyLinePtr  polyLine=NULL;
    long            *points, *lines;
    AmRectangleRec    *clipRects;
    int             ii, jj, x, y, numPts,
	            x0=0, y0=0, x1=0, y1=0, x2=0, y2=0, nClipRects;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    gc->GetClipRectangles( &clipRects, &nClipRects);

    numPts = 0;
    if ( polygon->AmNumLines > 0) {

        lines = GET_LINES( polygon);

        for ( ii = 0; ii < polygon->AmNumLines; ii++) {

            GetPolyLineLOCAL( _memory, lines[ ii], polyLine);

	    points = GET_POINTS( polyLine);

            for ( jj = 0; jj < polyLine->numPoints; jj++) {
                GetPointLOCAL( _memory, points[ jj], point);
                GetScreenPt( point->x, point->y, &x, &y);

		if ( POLYGON_OUTLINED == TRUE) {

		    x1 = x;
		    y1 = y;
		    
		    if ( numPts != 0 && (( x1 != x2) || ( y1 != y2))) {

			if ( lineOutsideRectangles( x1, y1, x2, y2,
						    clipRects,
						    nClipRects) == FALSE) {
			    drawLineCache( display, window,
					   gc, x1, y1, x2, y2);
			}
			
		    } else if ( numPts == 0) {
			x0 = x1;
			y0 = y1;
		    }

		    numPts++;
		    x2 = x1;
		    y2 = y1;
		    
		} else if (( numPts == 0) ||
			   (( pointCache[ numPts - 1].x != x) ||
			    ( pointCache[ numPts - 1].y != y))) {

                    pointCache[ numPts].x = x;
                    pointCache[ numPts].y = y;
                    numPts++;
                }
            }
        }
    }

    if ( POLYGON_OUTLINED == TRUE) {

	if ( numPts > 2 && (( x2 != x0) || (y2 != y0))) {

	    if ( lineOutsideRectangles( x2, y2, x0, y0,
					clipRects, nClipRects) == FALSE) {
		drawLineCache( display, window, gc, x2, y2, x0, y0);
	    }
	}
	flushLineCache( display, window, gc);
	
    } else {

	while (( numPts > 0) &&
	       ( pointCache[ numPts - 1].x != x) &&
	       ( pointCache[ numPts - 1].y != y)) {
	    numPts--;
	}

	display->FillPolygon( window, gc, pointCache, numPts);
    }
    
    return( AmOK);
}

/* Low level polygon outline drawing function. */

AmErrorCode
AmCadPlotClass::drawOutlinePolygon( AmWindowClass   *window,
				  AmGCClass       *gc, 
				  long          polygonID,
				  AmCadPolygonPtr polygon,
				  AmListPtr       selected,
				  int           shiftX,
				  int           shiftY)
{
    AmCadPointPtr     point=NULL;
    AmCadPolyLinePtr  polyLine=NULL;
    long            *points, *lines;
    AnvilBoolean         shiftPolygon;
    AmRectangleRec    *clipRects;
    int             ii, jj, x0=0, y0=0, x1=0, y1=0, x2=0, y2=0, numPoints, nClipRects;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    gc->GetClipRectangles( &clipRects, &nClipRects);

    numPoints = 0;

    if ( polygon->AmNumLines > 0) {

	if ( findObjectLst( selected, polygonID)) {
	    shiftPolygon = TRUE;
	} else {
	    shiftPolygon = FALSE;
	}

        lines = GET_LINES( polygon);

        for ( ii = 0; ii < polygon->AmNumLines; ii++) {

            GetPolyLineLOCAL( _memory, lines[ ii], polyLine);
            points = GET_POINTS( polyLine);

            for ( jj = 0; jj < polyLine->numPoints; jj++) {
                GetPointLOCAL( _memory, points[ jj], point);
                GetScreenPt( point->x, point->y, &x1, &y1);

		if ( shiftPolygon || findObjectLst( selected, points[ jj])) {
		    x1 += shiftX;
		    y1 += shiftY;
		}

		if ( numPoints == 0) {
		    display->FillRectangle( window, gc, x1 - 2, y1 - 2, 5, 5);
		    x0 = x1;
		    y0 = y1;
		} else if (( x1 != x2) || ( y1 != y2)) {
		    if ( lineOutsideRectangles( x1, y1, x2, y2,
						clipRects,
						nClipRects) == FALSE) {
			drawLineCache( display, window, gc, x1, y1, x2, y2);
			
			display->FillRectangle( window, gc,
						x1 - 2, y1 - 2, 5, 5);
		    }
		}
		
		x2 = x1;
		y2 = y1;
		
		numPoints++;
            }
        }

	if (( numPoints > 2) && (( x1 != x0) || ( y1 != y0))) {
	    if ( lineOutsideRectangles( x1, y1, x0, y0,
					clipRects, nClipRects) == FALSE) {
		drawLineCache( display, window, gc, x1, y1, x0, y0);
	    }
	}
    }

    flushLineCache( display, window, gc);
    
    if ( clipRects != NULL) {
	free( clipRects);
    }

    return( AmOK);
}

/* Draw a CAD point to the screen */

AmErrorCode
AmCadPlotClass::DrawPoint( AmWindowClass *window,
                         AmGCClass     *gc, 
                         long        pointID)
{
    AmCadPointPtr     point=NULL;
    AmColorCellRec    fgSaveColor, bgSaveColor;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    GetPointLOCAL( _memory, pointID, point);
    if ( point == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    gc->GetForegroundColor( &fgSaveColor);
    gc->GetBackgroundColor( &bgSaveColor);

    gc->SetForegroundColor( &fgColor);
    gc->SetBackgroundColor( &bgColor);

    drawPoint( window, gc, point);

    gc->SetForegroundColor( &fgSaveColor);
    gc->SetBackgroundColor( &bgSaveColor);
    
    return( AmOK);
}


/* Draw a CAD point outline to the screen */

AmErrorCode
AmCadPlotClass::DrawOutlinePoint( AmWindowClass *window,
				AmGCClass     *gc, 
				long        pointID,
				AmListPtr     selected,
				int         shiftX,
				int         shiftY)
{
    AmCadPointPtr     point=NULL;
    int             x, y;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    GetPointLOCAL( _memory, pointID, point);
    if ( point == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    GetScreenPt( point->x, point->y, &x, &y);

    if ( findObjectLst( selected, pointID)) {
	x += shiftX;
	y += shiftY;
    }

    display->DrawArc( window, gc, x - 4, y - 4, 9, 9, 0, M_PI * 2);

    return( AmOK);
}

/* Draw a CAD poly line to the screen */

AmErrorCode
AmCadPlotClass::DrawPolyLine( AmWindowClass *window,
                            AmGCClass     *gc, 
                            long        polyLineID)
{
    AmCadPolyLinePtr  polyLine=NULL;
    AmColorCellRec    fgSaveColor, bgSaveColor;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    GetPolyLineLOCAL( _memory, polyLineID, polyLine);
    if ( polyLine == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    gc->GetForegroundColor( &fgSaveColor);
    gc->GetBackgroundColor( &bgSaveColor);

    gc->SetForegroundColor( &fgColor);
    gc->SetBackgroundColor( &bgColor);

    drawPolyLine( window, gc, polyLine);

    gc->SetForegroundColor( &fgSaveColor);
    gc->SetBackgroundColor( &bgSaveColor);

    return( AmOK);
}

/* Draw a CAD poly line outline to the screen */

AmErrorCode
AmCadPlotClass::DrawOutlinePolyLine( AmWindowClass *window,
				   AmGCClass     *gc, 
				   long        polyLineID,
				   AmListPtr     selected,
				   int         shiftX,
				   int         shiftY)
{
    AmCadPolyLinePtr  polyLine=NULL;
    AmCadPointPtr     point=NULL;
    long            *points;
    int             ii, x1=0, y1=0, x2=0, y2=0;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    GetPolyLineLOCAL( _memory, polyLineID, polyLine);
    if ( polyLine == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    if ( polyLine->numPoints > 0) {

        points = GET_POINTS( polyLine);

        for ( ii = 0; ii < polyLine->numPoints; ii++) {
            GetPointLOCAL( _memory, points[ ii], point);

            GetScreenPt( point->x, point->y, &x1, &y1);

	    if ( findObjectLst( selected, points[ ii])) {
		x1 += shiftX;
		y1 += shiftY;
	    }

            if (( ii != 0) && (( x1 != x2) || ( y1 != y2))) {
		display->FillRectangle( window, gc, x1 - 2, y1 - 2, 5, 5);
		drawLineCache( display, window, gc, x1, y1, x2, y2);
            } else if ( ii == 0) {
		display->FillRectangle( window, gc, x1 - 2, y1 - 2, 5, 5);
	    }

            x2 = x1;
            y2 = y1;
        }

	flushLineCache( display, window, gc);
    }

    return( AmOK);
}


/* Draw a CAD polygon to the screen */

AmErrorCode
AmCadPlotClass::DrawPolygon( AmWindowClass *window,
                           AmGCClass     *gc, 
                           long        polygonID)
{
    AmCadPolygonPtr   polygon;
    AmErrorCode     errorCode;
    AmColorCellRec    fgSaveColor, bgSaveColor;
    
    GetPolygonLOCAL( _memory, polygonID, polygon);
    if ( polygon == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    gc->GetForegroundColor( &fgSaveColor);
    gc->GetBackgroundColor( &bgSaveColor);

    gc->SetForegroundColor( &fgColor);
    gc->SetBackgroundColor( &bgColor);
    
    errorCode = drawPolygon( window, gc, polygon);

    gc->SetForegroundColor( &fgSaveColor);
    gc->SetBackgroundColor( &bgSaveColor);

    return( errorCode);
}

/* Draw a CAD polygon to the screen */

AmErrorCode
AmCadPlotClass::DrawOutlinePolygon( AmWindowClass *window,
				  AmGCClass     *gc, 
				  long        polygonID,
				  AmListPtr     selected,
				  int         shiftX,
				  int         shiftY)
{
    AmCadPolygonPtr   polygon;

    GetPolygonLOCAL( _memory, polygonID, polygon);
    if ( polygon == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    return( drawOutlinePolygon( window, gc, polygonID, polygon, selected,
				shiftX, shiftY));
}

/* Draw a CAD object to the screen */

AmErrorCode
AmCadPlotClass::DrawObject( AmWindowClass *window,
			  AmGCClass     *gc,
			  long        objectID)
{
    uchar_t *data;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    if ( window == NULL || gc == NULL) {
	DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }

    data = AmGetData( _memory, objectID);

    switch( HTYPE((( AmHeaderPtr) data))) {
    
    case AmTYPE_POINT:
	return( DrawPoint( window, gc, objectID));

	//  case AmTYPE_LINE:
	//	DrawLine( window, gc, objectID);
	//	break;

    case AmTYPE_POLYLINE:
	return( DrawPolyLine( window, gc, objectID));

    case AmTYPE_POLYGON:
	return( DrawPolygon( window, gc, objectID));
    }

    return( AmOK);
}

/* Draw a CAD object to the screen */

AmErrorCode
AmCadPlotClass::DrawOutlineObject( AmWindowClass *window,
				 AmGCClass     *gc,
				 long        objectID,
				 AmListPtr     selected,
				 int         shiftX,
				 int         shiftY)
{
    uchar_t *data;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    if ( window == NULL || gc == NULL) {
	DO_RETURN_ERROR( AmNULL_ARGUMENT);
    }

    data = AmGetData( _memory, objectID);

    switch( HTYPE((( AmHeaderPtr) data))) {
    
    case AmTYPE_POINT:
	return( DrawOutlinePoint( window, gc, objectID,
				  selected, shiftX, shiftY));

	//  case AmTYPE_LINE:
	//	DrawLine( window, gc, objectID);
	//	break;

    case AmTYPE_POLYLINE:
	return( DrawOutlinePolyLine( window, gc, objectID,
				     selected, shiftX, shiftY));

    case AmTYPE_POLYGON:
	return( DrawOutlinePolygon( window, gc, objectID,
				    selected, shiftX, shiftY));
    }

    return( AmOK);
}


/* Draw the CAD objects to the screen */

AmErrorCode
AmCadPlotClass::draw( AmWindowClass        *window,
                    AmGCClass            *gc, 
                    AmRectangleRec const *exposeRects,
		    int			nRects)
{
    AmCadListPtr      	list;
    AmCadNodePtr      	node;
    AmObjectTablePtr  	objTable;
    AmSegmentTablePtr 	segTable;
    AmMemorySegmentPtr  segment;
    AmHeaderPtr       	header;
    AmCadObjectPtr    	object;
    AmCadPolygonPtr   	polygon;
    AmCadPolyLinePtr  	polyline;
    AmCadPointPtr     	point;
    uchar_t         	*data;
    long            	ii, index;
  
    AmRectanglePtr    	saveClipRects;
    int             	nSaveClipRects, saveLineWidth;
    AmColorCellRec    	saveForegroundColor, saveBackgroundColor;

    int             numObjs = 0, numPolys = 0, AmNumLines = 0, numPoints = 0;
        
    /* If this object is invalid only draw the outline. */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

    /* Return in the degenerate cases. */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    /* Save the clip rectangles and other GC components. */

    gc->GetClipRectangles( &saveClipRects, &nSaveClipRects);
    saveLineWidth = (int)(gc->GetLineWidth());
    gc->GetForegroundColor( &saveForegroundColor);
    gc->GetBackgroundColor( &saveBackgroundColor);

    /* Make sure the cad data is actually loaded into memory. */

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
	
    /* Set the clip rectangle to what it is supposed to be. */
    
    gc->SetClipRectangles( exposeRects, nRects);

    /* Clear out the background. */

    if (( parent == NULL) && opaque) {
        gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width, 
				    exposeRects[ ii].height);
	}
    }

    segTable = _memory->segTable;
    objTable = _memory->objTable;
    
    gc->SetLineWidth( 0);
    gc->SetBackgroundColor( &bgColor);
    gc->SetForegroundColor( &fgColor);

    /* Go through the list of polygons and draw them. */
    
    list = segTable->map[ AmTYPE_POLYGON]; 
    if ( list != NULL) { 
        node = list->head;
    } else {
        node = NULL;
    }
    
    while ( node != NULL) {
        segment = &segTable->segments[ ( long) node->data];
        AmActivateSegment( _memory, ( long) node->data); 
        data = segment->data;

        index = 0;
        while ( index < segment->AmLast) {
            header = ( AmHeaderPtr) &data[ index];
            index += HSIZE( header);

            if ( HTYPE( header) == AmTYPE_POLYGON) {

		/*
		 * Note: This is dangerous.  It should be implemented using
		 *   an alloca and a memcpy.  Functions should only be called
		 *   with persistent memory.  Anything returned by GET_ is
		 *   -NOT- persistent.
		 */
		
                polygon = GET_POLYGON( header);

                drawPolygon( window, gc, polygon);

		/* Make sure the polygon segment is in memory. */
		
                AmActivateSegment( _memory, ( long) node->data);
		
		numObjs++;
		numPolys++;
		
            }
        }
	
        node = node->next;
    }
    
    /*
     * Go through the list of polylines and draw all of those that don't have
     *   parents.
     */
    
    list = segTable->map[ AmTYPE_POLYLINE]; 
    if ( list != NULL) { 
        node = list->head;
    } else {
        node = NULL;
    }
    
    while ( node != NULL) {
        segment = &segTable->segments[ ( long) node->data];
        AmActivateSegment( _memory, ( long) node->data); 
        data = segment->data;
	
        index = 0;
        while ( index < segment->AmLast) {
            header = ( AmHeaderPtr) &data[ index];
            index += HSIZE( header);
	    
            if ( HTYPE( header) == AmTYPE_POLYLINE) {

		AmNumLines++;
		
		object = GET_OBJECT( header);
		if ( object->references == AmINVALID) {

		    /*
		     * Note: Again this is dangerous, see note written
		     *   above.
		     */
		    
		    polyline = GET_POLYLINE( header);
		    drawPolyLine( window, gc, polyline);

		    /* Make sure the polyline segment is in memory. */

		    numObjs++;
		    
		    AmActivateSegment( _memory, ( long) node->data);
		}
            }
        }

        node = node->next;
    }
    
    /*
     * Go through the list of points and draw all of those that don't have
     *   parents.
     */

    /* Get the list of segments that contain objects of type AmTYPE_POINT. */
    
    list = segTable->map[ AmTYPE_POINT]; 
    if ( list != NULL) { 
        node = list->head;
    } else {
        node = NULL;
    }

    /* Loop through all these segments. */
    
    while ( node != NULL) {
        segment = &segTable->segments[ ( long) node->data];
        AmActivateSegment( _memory, ( long) node->data); 
        data = segment->data;
	
        index = 0;
        while ( index < segment->AmLast) {
            header = ( AmHeaderPtr) &data[ index];
            index += HSIZE( header);
	    
            if ( HTYPE( header) == AmTYPE_POINT) {

		numPoints++;
		
		object = GET_OBJECT( header);

		/* Only draw the object ff it has no references (parents). */
		
		if ( object->references == AmINVALID) {

		    /*
		     * Note: Again this is dangerous, see note written
		     *   above.
		     */
		    
		    point = GET_POINT( header);
		    drawPoint( window, gc, point);

		    numObjs++;
		}
            }
        }

        node = node->next;
    }
    
    /* Print out some debugging information. */
    
    AN_DEBUG( 20, ( "Number of (parent) objects: %d\n", numObjs));
    AN_DEBUG( 20, ( "    # of polygons: %d\n", numPolys));
    AN_DEBUG( 20, ( "    # of lines: %d\n", AmNumLines));
    AN_DEBUG( 20, ( "    # of points: %d\n", numPoints));
    AN_DEBUG( 20, ( "    ------------------------------\n"));
    AN_DEBUG( 20, ( "    # of total objects: %d\n",
	     numPolys + AmNumLines + numPoints));
    
    /* Restore the GC values that we might have stomped on. */

    gc->SetClipRectangles( saveClipRects, nSaveClipRects);
    gc->SetLineWidth( saveLineWidth);
    gc->SetForegroundColor( &saveForegroundColor);
    gc->SetBackgroundColor( &saveBackgroundColor);

    return( AmOK);
}


long
AmCadPlotClass::AddPoint( float x, float y)
{
    long point;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }
    
    point = AmMakePoint( _memory, x, y);
    if ( point == AmINVALID) {
        ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
        return( AmINVALID);
    }

    return( point);
}


long
AmCadPlotClass::AddPoint( int x, int y)
{
    float projX, projY;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }
    
    GetDataPt( x, y, &projX, &projY);
    return( AddPoint( projX, projY));
}


AmErrorCode
AmCadPlotClass::AmRemovePoint( long point)
{

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }

    return( ::AmRemovePoint( _memory, point));
}

AmCadPointPtr
AmCadPlotClass::AmGetPoint( long pointID)
{
    if ( initMemory() != AmOK) {
	return( NULL);
    }

    return( ::AmGetPoint( _memory, pointID));
}

AmErrorCode
AmCadPlotClass::AmSetPoint( long pointID, int x, int y)
{
    float       projX, projY;
    uchar_t     *data;
    AmCadPointPtr point;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    GetDataPt( x, y, &projX, &projY);

    data = AmGetData( _memory, pointID);
    if ( data == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    if ( HTYPE((( AmHeaderPtr) data)) != AmTYPE_POINT) {
        return( ANVIL_ERROR_CODE);
    }

    point = GET_POINT( data);

    point->x = projX;    
    point->y = projY;    

    AmChangedObject( _memory, pointID);
    
    return( AmOK);
}

AmErrorCode
AmCadPlotClass::MovePoint( long pointID, int shiftX, int shiftY)
{
    int         screenX, screenY;
    float       projX, projY;
    uchar_t     *data;
    AmCadPointPtr point;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    data = AmGetData( _memory, pointID);
    if ( data == NULL) {
	return( ANVIL_ERROR_CODE);
    }

    if ( HTYPE((( AmHeaderPtr) data)) != AmTYPE_POINT) {
	return( ANVIL_ERROR_CODE);
    }
       
    point = GET_POINT( data);

    GetScreenPt( point->x, point->y, &screenX, &screenY);
    GetDataPt( screenX + shiftX, screenY + shiftY, &projX, &projY);

    point->x = projX;
    point->y = projY;

    AmChangedObject( _memory, pointID);

    return( AmOK);
}
    
AmErrorCode
AmCadPlotClass::GetPointLocation( long pointID, int *x, int *y)
{
    AmCadPointPtr point;
    uchar_t     *data;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    data = AmGetData( _memory, pointID);
    if ( data == NULL) {
        return( ANVIL_ERROR_CODE);
    }

    if ( HTYPE((( AmHeaderPtr) data)) != AmTYPE_POINT) {
        return( ANVIL_ERROR_CODE);
    }

    point = GET_POINT( data);

    GetScreenPt( point->x, point->y, x, y);

    return( AmOK);
}


long
AmCadPlotClass::AddLine( long p1, long p2)
{
    long line, points[ 2];

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }

    points[ 0] = p1;
    points[ 1] = p2;

    line = AmMakePolyLine( _memory, 2, points);
    if ( line == AmINVALID) {
        ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
        return( AmINVALID);
    }

    return( line);
}


AmErrorCode
AmCadPlotClass::AmRemoveLine( long line)
{

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }

    return( ::AmRemoveLine( _memory, line));
}


long
AmCadPlotClass::AddPolyLine( long numPts, long *pts)
{
    long line;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }

    line = AmMakePolyLine( _memory, numPts, pts);
    if ( line == AmINVALID) {
        ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
        return( AmINVALID);
    }

    return( line);
}


AmErrorCode
AmCadPlotClass::AmRemovePolyLine( long polyLine)
{
    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }

    return( ::AmRemovePolyLine( _memory, polyLine));
}


long
AmCadPlotClass::AddPolygon( long AmNumLines, long *lines)
{
    long polygon;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }
    
    polygon = AmMakePolygon( _memory, AmNumLines, lines);
    if ( polygon == AmINVALID) {
        ANVIL_ERROR_CODE = ANVIL_ERROR_CODE;
        return( AmINVALID);
    }

    return( polygon);
}

AmErrorCode
AmCadPlotClass::AmRemovePolygon( long polygon)
{
    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }

    return( ::AmRemovePolygon( _memory, polygon));
}

AmErrorCode
AmCadPlotClass::MovePolygon( long polygonID, int shiftX, int shiftY)
{
    AmCadPolygonPtr  polygon;
    AmListPtr        lstTemp, lstPoints;

    fprintf( stderr, "Moving polygon...\n");
    
    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }

    GetPolygonLOCAL( _memory, polygonID, polygon);
    if ( polygon == NULL) {
	return( ANVIL_ERROR_CODE);
    }

    lstPoints = getPolygonPoints( _memory, polygon);

    for ( lstTemp = lstPoints; lstTemp != NULL; lstTemp = lstTemp->next) {
	MovePoint( ( long) lstTemp->data, shiftX, shiftY);
    }

    AmJustFreeLst( lstPoints);
    
    return( AmOK);
}


AmErrorCode
AmCadPlotClass::MoveObject( long objectID, int shiftX, int shiftY)
{
    uchar_t *data;

    if ( initMemory() != AmOK) {
	return( ANVIL_ERROR_CODE);
    }
    
    data = AmGetData( _memory, objectID);
    if ( data == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

    switch( HTYPE((( AmHeaderPtr) data))) {
    case AmTYPE_POINT:
	return( MovePoint( objectID, shiftX, shiftY));
	
    case AmTYPE_LINE:
	break;
	
    case AmTYPE_POLYLINE:
	break;

    case AmTYPE_POLYGON:
	return( MovePolygon( objectID, shiftX, shiftY));

    default:
	break;
	
    }

    return( AmOK);
}
/* Find a point in all the points.*/

long
AmCadPlotClass::FindPoint( int x,
                         int y) 
{
    AmCadListPtr      	list;
    AmCadNodePtr      	node;
    AmSegmentTablePtr 	segTable;
    AmMemorySegmentPtr  segment;
    AmHeaderPtr       	header;
    AmCadPointPtr     	point;
    uchar_t         	*data;
    long            	index, minPoint;
    float           	projX, projY, dx, dy, distance, minDistance;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }
    
    GetDataPt( x, y, &projX, &projY);

    segTable = _memory->segTable;

    minPoint = AmINVALID;
    minDistance = 0;

    list = segTable->map[ AmTYPE_POINT]; 
    if ( list != NULL) { 
        node = list->head;
    } else {
        node = NULL;
    }

    while ( node != NULL) {
        segment = &segTable->segments[ ( long) node->data];
        AmActivateSegment( _memory, ( long) node->data);
        data = segment->data;
 
        index = 0;
        while ( index < segment->AmLast) {
            header = ( AmHeaderPtr) &data[ index];
            index += HSIZE( header);
 
            if ( HTYPE( header) == AmTYPE_POINT) {
                point = GET_POINT( header);

                dx = point->x - projX;
                dy = point->y - projY;

                distance = dx * dx + dy * dy;

                if ( minPoint == AmINVALID || distance < minDistance) {
                    minDistance = distance;
                    minPoint = HID( header);
                }
            }
        }

        node = node->next;
    }

    return( minPoint);
}

/* Find a polygon in all the polygons.*/

long
AmCadPlotClass::FindPolygon( int x,
                           int y) 
{
    AmCadListPtr      	list;
    AmCadNodePtr      	node;
    AmSegmentTablePtr 	segTable;
    AmMemorySegmentPtr  segment;
    AmHeaderPtr       	header;
    AmCadPolygonPtr   	polygon;
    uchar_t         	*data;
    long            	index;
    float           	projX, projY;

    if ( initMemory() != AmOK) {
	return( AmINVALID);
    }
    
    GetDataPt( x, y, &projX, &projY);

    segTable = _memory->segTable;

    list = segTable->map[ AmTYPE_POLYGON]; 
    if ( list != NULL) { 
        node = list->head;
    } else {
        node = NULL;
    }

    while ( node != NULL) {
        segment = &segTable->segments[ ( long) node->data];
        AmActivateSegment( _memory, ( long) node->data);
        data = segment->data;
 
        index = 0;
        while ( index < segment->AmLast) {
            header = ( AmHeaderPtr) &data[ index];
 
            if ( HTYPE( header) == AmTYPE_POLYGON) {
                polygon = GET_POLYGON( header);

		if ( pointInPolygon( _memory, projX, projY, polygon)) {
		    return( HID( header));
		}
	    }

            index += HSIZE( header);
        }
	node = node->next;
    }
    
    return( AmINVALID);
}

/* Dump the settable fields for a CAD plot. */

int
AmCadPlotClass::dumpFields( AmFieldLengthPtr fieldLengths,
                          AnvilBoolean        allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* Get the fields for the parent class ( plot). */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

/* Get all the easy to specify fields for this class. */

    AmCadPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}


/* Return the field definitions for a CAD plot. */

void
AmCadPlotClass::GetFieldDefinitions( AmFieldDefPtr *fieldDefs,
                                   int         *numDefs)
{
    *fieldDefs = NULL;
    *numDefs = 0;
}

AmErrorCode
AmCadPlotClass::AddAttributeType( char         *szAttType,
				AmIndexDataPtr index)
{
    AmListPtr   lstTemp;
    AmAttribPtr attTemp;

    for ( lstTemp = lstAttTypes; lstTemp != NULL; lstTemp = lstTemp->next) {
        attTemp = ( AmAttribPtr) lstTemp->data;

        if ( strcmp( attTemp->szType, szAttType) == 0) {
            return( AmUNKNOWN_ERROR);
        }
    }

    attTemp = ( AmAttribPtr) malloc( sizeof( AmAttribRec));
    if ( attTemp == NULL) {
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    attTemp->szType = strdup( szAttType);
    attTemp->indexData = index;
    attTemp->lstDpyParams = NULL;

    lstAttTypes = AmInsertNodeLst( attTemp, lstAttTypes);
  
    return( AmOK);
}

AmErrorCode
AmCadPlotClass::AddAttributeDisplayParameter( char        *szAttType,
					    char        *szDpyParam,
					    void        *pData,
					    long        iElementSize,
					    long        iLength)
{
    AmListPtr             lstTemp;
    AmAttribPtr           attTemp;
    AmDisplayParameterPtr dpTemp;

    for ( lstTemp = lstAttTypes; lstTemp != NULL; lstTemp = lstTemp->next) {
        attTemp = ( AmAttribPtr) lstTemp->data;

        if ( strcmp( attTemp->szType, szAttType) == 0) {
            break;
        }
    }

    if ( lstTemp == NULL) {
        return( AmUNKNOWN_ERROR);
    }

    for ( lstTemp = attTemp->lstDpyParams;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

	dpTemp = ( AmDisplayParameterPtr) lstTemp->data;

	if ( strcmp( dpTemp->szType, szDpyParam) == 0) {
	    return( AmUNKNOWN_ERROR);
	}
    }

    dpTemp = ( AmDisplayParameterPtr) malloc( sizeof( AmDisplayParameterRec));
    if ( dpTemp == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    dpTemp->szType = strdup( szDpyParam);
    dpTemp->pData = ( void *) malloc( iElementSize * iLength);
    if ( dpTemp->pData == NULL) {
	free( dpTemp);
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    
    memcpy( dpTemp->pData, pData, iElementSize * iLength);
    
    dpTemp->iElementSize = iElementSize;
    dpTemp->iLength = iLength;

    attTemp->lstDpyParams = AmInsertNodeLst( dpTemp, attTemp->lstDpyParams);

    return( AmOK);
}

/*
 * Get the index into the display parameter tables from the attribute
 *   value.
 */

int
AmCadPlotClass::GetAttributeIndex( char *szAttType,
                                 char *szDpyParam,
				 void *pValue)
{
    AmListPtr      lstTemp;
    AmAttribPtr    attTemp;

    /* Look for the attribute type. */

    for ( lstTemp = lstAttTypes; lstTemp != NULL; lstTemp = lstTemp->next) {
        attTemp = ( AmAttribPtr) lstTemp->data;

        if ( strcmp( attTemp->szType, szAttType) == 0) {
            break;
        }
    }

    /* Did we find anything? */

    if ( lstTemp == NULL) {
        return( -1);
    }

//    pIndexData = att->indexData;
//    pIndexData = att->indexData;
    return( 1);
}
