#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <math.h>
#include <string.h>
#include <stddef.h>
#include <anvil/project.h>
#include <anvil/tagstrs.h>
#include <anvil/ratio.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/projects.h>
#include <anvil/geometry.h>
#include <anvil/screenobj.h>
#include <anvil/coordinates.h>
#include <anvil/eventcodes.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmCOORD) {
        return( TRUE);
    } else {
        return( AmScreenObjClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a a coordinate system.  turn off
 *	the "set" flags, set the aspect ratio, set the coordinates and
 *	extents to something reasonable, unudefine the units and 
 *	intialize the projection information.
 *
 ****************************************************************************/

void
AmCoordClass::setCoordClass()
{
    coordsSet       = FALSE;
    rangeSet        = FALSE;
    rectSet         = FALSE;
    preserveAspect  = FALSE;
    aspectFactor    = 1.0;
    coords.currXmin = coords.xmin = -1.0;
    coords.currYmin = coords.ymin = -1.0;
    coords.currXmax = coords.xmax = 1.0;
    coords.currYmax = coords.ymax = 1.0;
    xUnits          = AmUNIT_UNDEFINED;
    yUnits          = AmUNIT_UNDEFINED;
    inProjStrs      = NULL;
    numInProjStrs   = 0;
    projStrs        = NULL;
    numProjStrs     = 0;
}

/****************************************************************************
 *
 * Description:
 * 	Do nothing constructor. 
 *
 ****************************************************************************/

AmCoordClass::AmCoordClass() 
{ 
    setCoordClass();
}

/****************************************************************************
 *
 * Description:
 * 	Return the flags for coord set, range set and rectangle set. 
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::CoordsSet()
{
    return( coordsSet);
}

AnvilBoolean
AmCoordClass::RangeSet()
{
    return( rangeSet);
}

AnvilBoolean
AmCoordClass::RectSet()
{
    return( rectSet);
}

/****************************************************************************
 *
 * Description:
 * 	Return the projection class structures for forward and reverse 
 *	projections. 
 *
 ****************************************************************************/

AmProjectionClass *
AmCoordClass::GetForwardProjection()
{
    return( &fwdProject);
}

AmProjectionClass *
AmCoordClass::GetReverseProjection()
{
    return( &rvsProject);
}

/****************************************************************************
 *
 * Description:
 *      check values for coordiante system.  Each check values function 
 *	calls the check values function for the parent class and is used to 
 *	set the object's IsValid() flag
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::checkValues()
{
    AnvilBoolean fOK;
    char    sz[ 1024];

    fOK = TRUE;

/* check to make sure the coordinates are in the correct ordering. */

    if ( coords.xmax < coords.xmin) {
	AmAddResourceToLog( "xmaxLessThanXmin", coords.xmax, coords.xmin);
	fOK = FALSE;
    }

    if ( coords.ymax < coords.ymin) {
	AmAddResourceToLog( "ymaxLessThanYmin", coords.ymax, coords.ymin);
        fOK = FALSE;
    }

    if ( coords.currXmax < coords.currXmin) {
	AmAddResourceToLog( "currXmaxLessThanCurrXmin", 
			  coords.currXmax, coords.currXmin);
	fOK = FALSE;
    }

    if ( coords.currYmax < coords.currYmin) {
	AmAddResourceToLog( "currYmaxLessThanCurrYmin", 
			  coords.currYmax, coords.currYmin);
	fOK = FALSE;
    }

/* make sure the projections are OK */

    if (( projStrs != NULL) && !fwdProject.IsValid()) {
	AmStrvecToString(( const char **) projStrs, sz, ' ');
	AmAddResourceToLog( "badForwardProjection", sz);
	fOK = FALSE;
    }

    if (( inProjStrs != NULL) && !rvsProject.IsValid()) {
	AmStrvecToString(( const char **) inProjStrs, sz, ' ');
	AmAddResourceToLog( "badReverseProjection", sz);
	fOK = FALSE;
    }

    return( fOK);
}

/****************************************************************************
 *
 * Description:
 *	Coordinate system class destructor.  Free the projection 
 *	string vectors.
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AmCoordClass::~AmCoordClass()
{
    if ( inProjStrs != NULL) {
	AmFreeStrVec( inProjStrs);
    }
    if ( projStrs != NULL) {
	AmFreeStrVec( projStrs);
    }
}

/****************************************************************************
 *
 * Description:
 *	Set the rectangle for a coordinate system.  Set the screen object
 *	rectangle, then set the coordinates to reflect the new rectangle.
 * Arguments:
 *	x, y, width, height:  new dimensions for the screen rectangle.
 *
 ****************************************************************************/

void AmCoordClass::SetRect( int	x,
			  int	y,
			  int	width,
			  int	height)
{
    AmScreenObjClass::SetRect( x, y, width, height);
    rectSet = TRUE;

/* Under aspect preservation, if the coordinate system is taller than 
   the rectangle, then clip the coordinate system to match the rectangle.
   Do not fiddle with the coordinates if the rectangle is degenerate, or
   if the coordinates have not been set yet. */

    if ( preserveAspect && coordsSet && !AmDegenRect( &rect)) {
	SetCoords( coords.currXmin, coords.currYmin, 
		   coords.currXmax, coords.currYmax);
    }
}

/****************************************************************************
 *
 * Description:
 *	Set the current extents for a coordinate system. The actual 
 *	coordinates set may be different than the passed values because
 *	of rectangle sizing, aspect preservation and aspect factor.
 * Arguments:
 *	currXmin, currYmin, currXmax, currYmax: new current extents.
 *
 ****************************************************************************/

void
AmCoordClass::SetCoords( float currXmin,
                       float currYmin,
                       float currXmax,
                       float currYmax)
{
    float	coordHeight, center, coordWidth;

/* Under aspect preservation, if the coordinate system is taller than 
   the rectangle, then clip the coordinate system to match the rectangle.
   Don't do the aspect preserve stuff if the rectangle hasn't been set yet */

    if ( preserveAspect && rectSet && !AmDegenRect( &rect)) {

	if (( currYmax - currYmin)/( currXmax - currXmin) >= 
	    (( float) rect.height)/(( float) rect.width)) {
	    coordWidth = ((( float) rect.width)/(( float) rect.height))*
			    ( currYmax - currYmin);
	    center = ( currXmax + currXmin)/2;
	    currXmax = center + ( 1.0/aspectFactor)*coordWidth/2;
	    currXmin = center - ( 1.0/aspectFactor)*coordWidth/2;
	} else {
	    coordHeight = ((( float) rect.height)/(( float) rect.width))*
			    ( currXmax - currXmin);
	    center = ( currYmax + currYmin)/2;
	    currYmax = center + aspectFactor*coordHeight/2;
	    currYmin = center - aspectFactor*coordHeight/2;
	}
    }

/* re-set the coords so that the map projected coordinates are preserved */

    coordsSet = TRUE;
    coords.currXmin = currXmin;
    coords.currYmin = currYmin;
    coords.currXmax = currXmax;
    coords.currYmax = currYmax;

    AN_DEBUG( 99, ( "setting coordinates to (%.2f, %.2f, %.2f, %.2f)\n",
		    coords.currXmin, coords.currYmin, 
		    coords.currXmax, coords.currYmax));

    projCoords = coords;
    fwdProject.ProjectPt( &projCoords.currXmin, &projCoords.currYmin);
    fwdProject.ProjectPt( &projCoords.currXmax, &projCoords.currYmax);
    fwdProject.ProjectPt( &projCoords.xmin, &projCoords.ymin);
    fwdProject.ProjectPt( &projCoords.xmax, &projCoords.ymax);

    AN_DEBUG( 99, ( "projecting coordinates to (%.2f, %.2f, %.2f, %.2f\n",
		    projCoords.currXmin, projCoords.currYmin, 
		    projCoords.currXmax, projCoords.currYmax));
}

/****************************************************************************
 *
 * Description:
 * 	When the projection class is set for a coordinate system, set the 
 *	projected coordinates as well. 
 * Arguments:
 *	USGS projection strings (NULL terminated).
 *
 ****************************************************************************/

void
AmCoordClass::SetProjection( const char **args) 
{
    if ( projStrs != NULL) {
	AmFreeStrVec( projStrs);
    }
    projStrs = AmCopyStrVec( args);
    numProjStrs = AmVeclen( args);
    fwdProject.SetProjection( args);
    projCoords = coords;
    fwdProject.ProjectPt( &projCoords.currXmin, &projCoords.currYmin);
    fwdProject.ProjectPt( &projCoords.currXmax, &projCoords.currYmax);
    fwdProject.ProjectPt( &projCoords.xmin, &projCoords.ymin);
    fwdProject.ProjectPt( &projCoords.xmax, &projCoords.ymax);
}

/****************************************************************************
 *
 * Description:
 * 	Set the  range for a coordinate system, which sets the projected
 *	coordinates as well.  if the current extents haven't been set yet,
 *	then this sets them.
 * Arguments:
 *	xmin, ymin, xmax, ymax: new extents for the coordinate system.
 *
 ****************************************************************************/

void
AmCoordClass::SetRange( float xmin,
                      float ymin,
                      float xmax,
                      float ymax)
{
    coords.xmin = xmin;
    coords.ymin = ymin;
    coords.xmax = xmax;
    coords.ymax = ymax;
    projCoords  = coords;

    fwdProject.ProjectPt( &projCoords.currXmin, &projCoords.currYmin);
    fwdProject.ProjectPt( &projCoords.currXmax, &projCoords.currYmax);
    fwdProject.ProjectPt( &projCoords.xmin, &projCoords.ymin);
    fwdProject.ProjectPt( &projCoords.xmax, &projCoords.ymax);

    AN_DEBUG( 99, ( "setting range to (%.2f, %.2f, %.2f, %.2f)\n",
		    coords.xmin, coords.ymin, 
		    coords.xmax, coords.ymax));
    AN_DEBUG( 99, ( "projecting range to (%.2f, %.2f, %.2f, %.2f\n",
		    projCoords.xmin, projCoords.ymin, 
		    projCoords.xmax, projCoords.ymax));

    rangeSet = TRUE;

/* If the coordinates haven't been set, then set them to the range. */

    if ( !coordsSet) {
	SetCoords( xmin, ymin, xmax, ymax);
    } 
}

/****************************************************************************
 *
 * Description:
 * 	Given a point in screen space, return the point in the original 
 *	data space (i.e. unprojected, but the point has the projection 
 *	applied to it when it's projected from screen coordinates to data 
 *	coordinates. 
 * Arguments:
 *	scrX, scrY: screen X,Y coordinates.
 *	dataX, dataY: point in the object's real coordinate system.
 *
 ****************************************************************************/

void
AmCoordClass::GetDataPt( int   scrX,
                       int   scrY,
                       float *dataX,
                       float *dataY)
{
    if ( fwdProject.IsProjected()) {
	*dataX = AmIntScaleToFloat( rect.x, rect.x + rect.width, scrX, 
				  projCoords.currXmin, projCoords.currXmax);
    	*dataY = AmIntScaleToFloat( rect.y, rect.y + rect.height, scrY,
				  projCoords.currYmax, projCoords.currYmin);

/* We inverse project the point, to return it to the source coordinate
   system. */

	fwdProject.InvProjectPt( dataX, dataY);
    } else {
	*dataX = AmIntScaleToFloat( rect.x, rect.x + rect.width, scrX, 
				  coords.currXmin, coords.currXmax);
    	*dataY = AmIntScaleToFloat( rect.y, rect.y + rect.height, scrY,
				  coords.currYmax, coords.currYmin);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Given a point in data space (unprojected), find out where it is in 
 * 	screen space. 
 * Arguments:
 *	dataX, dataY: point in the object's real coordinate system.
 *	scrX, scrY: screen X,Y coordinates.
 *
 ****************************************************************************/

void AmCoordClass::GetScreenPt( float	dataX,
			      float	dataY,
			      int	*scrX,
			      int	*scrY)
{
    if ( fwdProject.IsProjected()) {
	fwdProject.ProjectPt( &dataX, &dataY);
	*scrX = AmFloatScaleToInt( projCoords.currXmin, projCoords.currXmax, 
				 dataX, rect.x, rect.x + rect.width);
	*scrY = AmFloatScaleToInt( projCoords.currYmin, projCoords.currYmax, 
				 dataY, rect.height + rect.y, rect.y);
    } else {
	*scrX = AmFloatScaleToInt( coords.currXmin, coords.currXmax, 
				 dataX, rect.x, rect.x + rect.width);
	*scrY = AmFloatScaleToInt( coords.currYmin, coords.currYmax, 
				 dataY, rect.height + rect.y, rect.y);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Zoom the screen extents of the coordinate system given a zoom 
 *	rectangle in screen coordinates.
 * Arguments:
 *	x, y, width, height: dimensions of the zoom rectangle in 
 *	screen coordinates.
 *
 ****************************************************************************/

void
AmCoordClass::ZoomRect( int x,
                      int y,
                      int width,
                      int height)
{
    float currXmin, currYmin, currXmax, currYmax;

    GetDataPt( x, y + height, &currXmin, &currYmin);
    GetDataPt( x + width, y, &currXmax, &currYmax);
    SetCoords( currXmin, currYmin, currXmax, currYmax);
}

/****************************************************************************
 *
 * Description:
 * 	Zoom out to the extents of the coordinate system. 
 *
 ****************************************************************************/

void
AmCoordClass::Unzoom()
{
    SetCoords( coords.xmin, coords.ymin, coords.xmax, coords.ymax);
}

/****************************************************************************
 *
 * Description:
 * 	Is the point in the object's screen rectangle? 
 * Arguments:
 *	scrX, scrY: point to test in screen coordinates.
 * Returns:
 *	true if the point is in the screen rectangle, false if outside.
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::InRect( int scrX,
                    int scrY)
{
    return( AmPtInRect( &rect, scrX, scrY));
}

/****************************************************************************
 *
 * Description:
 * 	Is the point inside the object's current extents?
 * Arguments:
 *	dataX, dataY: point to test in coordinate system.
 * Returns:
 *	true if the point is in the coordinate extents, false if outside.
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::InCoords( float dataX,
                      float dataY)
{
    return(( dataX >= coords.currXmin) && ( dataY >= coords.currYmin) &&
	   ( dataX <= coords.currXmax) && ( dataY <= coords.currYmax));
}

/****************************************************************************
 *
 * Description:
 * 	does the rectangle defined by xmin, ymin, xmax, ymax overlap
 *	the current coordinates?
 * Arguments:
 *	xmin, ymin, xmax, ymax - extents of the rectangle
 * Returns:
 *	TRUE if the rectangle overlaps the current coordinates, FALSE
 *	otherwise.
 *
 ****************************************************************************/

 AnvilBoolean
 AmCoordClass::OverlapRectCoords( float	xmin, 
				float	ymin,
				float	xmax,
				float	ymax)
{
    return( AmOverlapCoords( xmin, ymin, xmax, ymax,
			     coords.currXmin, coords.currYmin,
			     coords.currXmax, coords.currYmax));
}
/****************************************************************************
 *
 * Description:
 * 	does the rectangle defined by xminA, yminA, xmaxA, ymaxA overlap
 *	xminB, yminB, xmaxB, ymaxB?
 * Arguments:
 *	xmin, ymin, xmax, ymax (A,B)- extents of the rectangles
 * Returns:
 *	TRUE if the rectangle overlaps, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmOverlapCoords( float	xminA,
		 float	yminA,
		 float	xmaxA,
		 float	ymaxA,
		 float	xminB,
		 float	yminB,
		 float	xmaxB,
		 float	ymaxB)
{
    return(( xmaxA > xminB) && ( ymaxA > yminB) &&
	   ( xminA < xmaxB) && ( yminA < ymaxB));
}

/****************************************************************************
 *
 * Description:
 * 	Clip a point to the range of a coordinate system.
 * Arguments:
 *	dataX, dataY: point to clip in coordinate system.
 * Returns:
 *	true if the point was clipped 
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::ClipRange( float *dataX,
                       float *dataY)
{
    AnvilBoolean fClip = FALSE;

    if ( *dataX < coords.xmin) {
	*dataX = coords.xmin;
	fClip = TRUE;
    }

    if ( *dataX > coords.xmax) {
	*dataX = coords.xmax;
	fClip = TRUE;
    }

    if ( *dataY < coords.ymin) {
	*dataY = coords.ymin;
	fClip = TRUE;
    }

    if ( *dataY > coords.ymax) {
	*dataY = coords.ymax;
	fClip = TRUE;
    }

    return( fClip);
}

/****************************************************************************
 *
 * Description:
 * 	Clip a point to the current extents of a coordinate system.
 * Arguments:
 *	dataX, dataY: point to clip in coordinate system.
 * Returns:
 *	true if the point was clipped 
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::ClipCoords( float *dataX,
                        float *dataY)
{
    AnvilBoolean fClip = FALSE;

    if ( *dataX < coords.currXmin) {
	*dataX = coords.currXmin;
	fClip = TRUE;
    }
    if ( *dataX > coords.currXmax) {
	*dataX = coords.currXmax;
	fClip = TRUE;
    }
    if ( *dataY < coords.currYmin) {
	*dataY = coords.currYmin;
	fClip = TRUE;
    }
    if ( *dataY > coords.currYmax) {
	*dataY = coords.currYmax;
	fClip = TRUE;
    }

    return( fClip);
}
	
/****************************************************************************
 *
 * Description:
 * 	Clip a point to the screen rectangle.  Return TRUE if it was clipped. 
 * Arguments:
 *	scrX, scrY: screen coordinate to clip.
 * Returns:
 *	true if the point was clipped 
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::BoundPt( int *scrX,
                     int *scrY)
{
    AnvilBoolean fClip = FALSE;

    if ( *scrX < rect.x) {
	*scrX = rect.x;
	fClip = TRUE;
    }
    if ( *scrY < rect.y) {
	*scrY = rect.y;
	fClip = TRUE;
    }
    if ( *scrX > rect.x + rect.width) {
	*scrX = rect.x + rect.width;
	fClip = TRUE;
    }
    if ( *scrY > rect.y + rect.height) {
	*scrY = rect.y + rect.height;
	fClip = TRUE;
    }

    return( fClip);
}
	
/****************************************************************************
 *
 * Description:
 * 	For a matrix whose extents are represented by mat(X,Y)(min,max),
 *	find out what the current screen coordinates represent in the matrix
 *	(rows, cols) coordinate system 
 * Arguments:
 *  	mat(Xmin,Ymin,Xmax,Ymax): extents of source matrix.
 *  	dst(Xmin,Ymin,Xmax,Ymax): extents of exposed rectangle.
 *  	rows, cols: dimensions of source matrix.
 *  	flipRows,flipCols: flip rows and columns on display.
 *  	drawRect: screen coordinates of exposed rectangle.
 *  	screenDataRect: returned extents in source matrix.
 * Returns:
 *	true if the resulting rectangle is not degenerate.
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::MatrixRealExtents( float        matXmin,
                               float        matYmin,
                               float        matXmax,
                               float        matYmax,
                               float        dstXmin,
                               float        dstYmin,
                               float        dstXmax,
                               float        dstYmax,
                               int          rows,
                               int          cols,
                               AnvilBoolean      flipRows,
                               AnvilBoolean      flipCols,
                               AmRectanglePtr drawRect,
                               AmRealRectPtr  screenDataRect)
{
    int     scrX0, scrY0, scrX1, scrY1;
    float   xmin, ymin, xmax, ymax, dataX0, dataY0, dataX1, dataY1;
    AnvilBoolean screenFlag;

/* Clip the draw rectangle to the screen. */

    AmClipRect( GetRect(), drawRect);

/* Clip the requested draw area to the source matrix extents. */

    if ( dstXmin < matXmin) {
	dstXmin = matXmin;
    }
    if ( dstYmin < matYmin) {
	dstYmin = matYmin;
    }
    if ( dstXmax > matXmax) {
	dstXmax = matXmax;
    }
    if ( dstYmax > matYmax) {
	dstYmax = matYmax;
    }

/* find out what the matrix represents in screen space.  */

    GetScreenPt( dstXmin, dstYmin, &scrX0, &scrY1);
    GetScreenPt( dstXmax, dstYmax, &scrX1, &scrY0);

/* are we on the screen at all? */

    screenFlag = (( scrX0 <= drawRect->x + drawRect->width) &&
		  ( scrY0 <= drawRect->y + drawRect->height) &&
		  ( scrX1 >= drawRect->x) &&
		  ( scrY1 >= drawRect->y));

/* clip to the screen */

    BoundPt( &scrX0, &scrY0);
    BoundPt( &scrX1, &scrY1);

/* reset the draw rectangle to reflect reality */

    drawRect->x = scrX0;
    drawRect->y = scrY0;
    drawRect->width = scrX1 - scrX0;
    drawRect->height = scrY1 - scrY0;

/* project the draw rectangle into the coordinate system, then project 
   those points into matrix space */

    GetDataPt( scrX0, scrY1, &xmin, &ymin);
    GetDataPt( scrX1, scrY0, &xmax, &ymax);

/* get the source coordinates, using the flip flags for rows and columns
   orientation */

    if ( flipCols) {
    	dataX0 = cols*( matXmax - xmax)/( matXmax - matXmin);
    	dataX1 = cols*( matXmax - xmin)/( matXmax - matXmin);
    } else {
	dataX0 = cols*( xmin - matXmin)/( matXmax - matXmin);
	dataX1 = cols*( xmax - matXmin)/( matXmax - matXmin);
    } 
    if ( flipRows) {
	dataY0 = rows*( ymin - matYmin)/( matYmax - matYmin);
	dataY1 = rows*( ymax - matYmin)/( matYmax - matYmin);
    } else {
	dataY0 = rows*( matYmax - ymax)/( matYmax - matYmin);
	dataY1 = rows*( matYmax - ymin)/( matYmax - matYmin);
    }

/* Set the coordinates in a rectangle. */

    screenDataRect->x      = MAX( dataX0, 0.0);
    screenDataRect->y      = MAX( dataY0, 0.0);
    screenDataRect->width  = MIN( dataX1 - dataX0, ( float) cols - 1);
    screenDataRect->height = MIN( dataY1 - dataY0, ( float) rows - 1);

/* Return extents are on screen, and size not degenerate. */

    return( screenFlag && 
	    ( screenDataRect->width >= 0) && ( screenDataRect->height >= 0));
}

/****************************************************************************
 *
 * Description:
 * 	find out what a screen rectangle in the current coordinate system 
 *  	represents in the data matrix, except that the data rectangle is in
 * 	floating point coordinates. 
 * Arguments:
 *	rows, col: dimensions of the matrix.
 *	flipRows, flipCols: flip controls to transpose the matrix.
 *	drawRect: rectangle to draw to.
 *	screenDataRect: what the screen represents in the source matrix
 *
 ****************************************************************************/

AnvilBoolean
AmCoordClass::ScreenRealExtents( int          rows,
                               int          cols,
                               AnvilBoolean      flipRows,
                               AnvilBoolean      flipCols,
                               AmRectanglePtr drawRect,
                               AmRealRectPtr  screenDataRect)
{
    return( MatrixRealExtents( coords.xmin, coords.ymin, 
			       coords.xmax, coords.ymax,
			       coords.currXmin, coords.currYmin, 
			       coords.currXmax, coords.currYmax, 
			       rows, cols, flipRows, flipCols,
			       drawRect, screenDataRect));
}

/****************************************************************************
 *
 * Description:
 * 	for a real-world line width, change it to a screen line width 
 * Arguments:
 *	dLineWidth: line widht in the coordinate system.
 * Returns:
 *	floating-point line width in pixels
 *
 ****************************************************************************/

float
AmCoordClass::ScreenLineWidth( float dLineWidth)
{
    return( rect.width * dLineWidth / (coords.currXmax - coords.currXmin));
}

/****************************************************************************
 *
 * Description:
 * 	how big is a pixel on the screen in the real world? 
 * Returns:
 *	size of a screen pixel in coordinate system.
 *
 ****************************************************************************/

float
AmCoordClass::PixelSize()
{
    return(( coords.currXmax - coords.currXmin)/rect.width);
}

/****************************************************************************
 *
 * Description:
 * 	return the extents of a coordinate system 
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned range of the coordinate system
 *
 ****************************************************************************/

void
AmCoordClass::GetRange( float *xmin,
                      float *ymin,
                      float *xmax,
                      float *ymax)
{
    *xmin = coords.xmin;
    *ymin = coords.ymin;
    *xmax = coords.xmax;
    *ymax = coords.ymax;
}

/****************************************************************************
 *
 * Description:
 * 	return the extents of a coordinate system to which a projection
 *	has been applied
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned range of the coordinate system
 *	with the projection applied.
 *
 ****************************************************************************/

void
AmCoordClass::GetProjectRange( float *xmin,
                             float *ymin,
                             float *xmax,
                             float *ymax)
{
    *xmin = projCoords.xmin;
    *ymin = projCoords.ymin;
    *xmax = projCoords.xmax;
    *ymax = projCoords.ymax;
}

/****************************************************************************
 *
 * Description:
 * 	return the current extents of a coordinate system.
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned current extents of the coordinate 
 *	system
 *
 ****************************************************************************/

void
AmCoordClass::GetCoords( float *currXmin,
                       float *currYmin,
                       float *currXmax,
                       float *currYmax)
{
    *currXmin = coords.currXmin;
    *currYmin = coords.currYmin;
    *currXmax = coords.currXmax;
    *currYmax = coords.currYmax;
}

/****************************************************************************
 *
 * Description:
 * 	return the current extents of a coordinate system with the system's
 *	projection applied.
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned current extents of the coordinate 
 *	system with the projection applied.
 *
 ****************************************************************************/

void
AmCoordClass::GetProjectCoords( float *currXmin,
                              float *currYmin,
                              float *currXmax,
                              float *currYmax)
{
    *currXmin = projCoords.currXmin;
    *currYmin = projCoords.currYmin;
    *currXmax = projCoords.currXmax;
    *currYmax = projCoords.currYmax;
}

/****************************************************************************
 *
 * Description:
 * 	Apply a translation in the screen coordinate system to the coordinate
 *	system, but restrict it to the coordinate system's range.
 * Arguments:
 *	startX, startY: start position of the pan in screen coordinates.
 *	lastX, lastY: end position of the pan in screen coordinates.
 *	restrictFlag: flag to restrict the pan to the bounds of the 
 *	coordinate system.
 *
 ****************************************************************************/

void
AmCoordClass::Pan( int	startX,
                 int     startY,
                 int     lastX,
                 int     lastY,
                 AnvilBoolean restrictFlag)
{
    float	x0, y0, x1, y1, newXmin, newYmin, newXmax, newYmax;

    GetDataPt( lastX, lastY, &x0, &y0);
    GetDataPt( startX, startY, &x1, &y1);

/* restrict the coordinates to the extents of the data */

    if ( restrictFlag) {
	if ( coords.currXmin + x1 - x0 < coords.xmin) {

AN_DEBUG( 99, ("xmin extents reached\n"));

	    newXmin = coords.xmin;
	    newXmax = coords.xmin + coords.currXmax - coords.currXmin;
	} else if ( coords.currXmax + x1 - x0 > coords.xmax) {

AN_DEBUG( 99, ("xmax extents reached\n"));

	    newXmin = coords.xmax - ( coords.currXmax - coords.currXmin);
	    newXmax = coords.xmax;
	} else {
	    newXmin = coords.currXmin + x1 - x0;
	    newXmax = coords.currXmax + x1 - x0;
	}
	if ( coords.currYmin + y1 - y0 < coords.ymin) {

AN_DEBUG( 99, ("ymin extents reached\n"));

	    newYmin = coords.ymin;
	    newYmax = coords.ymin + coords.currYmax - coords.currYmin;
	} else if ( coords.currYmax + y1 - y0 > coords.ymax) {

AN_DEBUG( 99, ("ymax extents reached\n"));

	    newYmin = coords.ymax - ( coords.currYmax - coords.currYmin);
	    newYmax = coords.ymax;
	} else {
	    newYmin = coords.currYmin + y1 - y0;
	    newYmax = coords.currYmax + y1 - y0;
	}
    } else {
	newXmin = coords.currXmin + x1 - x0;
	newXmax = coords.currXmax + x1 - x0;
	newYmin = coords.currYmin + y1 - y0;
	newYmax = coords.currYmax + y1 - y0;
    }

    SetCoords( newXmin, newYmin, newXmax, newYmax);
}


/****************************************************************************
 *
 * Description:
 * 	when the rectangle has been panned, return the newly invalidated 
 * 	rectangles.
 * Arguments:
 *	x,y: relative pan amounts in the screen coordinate system.
 *	horizRect, vertRect: resulting rectangles which have to be
 *	redrawn.  If the rectangle doesn't have to be redrawn, then the
 *	width or height is set to 0
 *
 ****************************************************************************/

void AmCoordClass::PanInvalRects( int		x,	/* relative */
			        int		y,
				AmRectanglePtr	vertRect,
				AmRectanglePtr	horizRect)
{
    float	newXmin, newYmin, oldXmin, oldYmin, xShift, yShift;

/* do the vertical rectangle */

    if ( x > rect.width) {
        x = rect.width;
    } else if ( x < -rect.width) {
        x = -rect.width;
    }

    if ( y > rect.height) {
        y = rect.height;
    } else if ( y < -rect.height) {
        y = -rect.height;
    }

    if ( x > 0) {	/* panned right */
	vertRect->x = 0;
	vertRect->width = x;
    } else {		/* panned left */
	vertRect->x = rect.width + x;
	vertRect->width = -x;
    }

    vertRect->y = 0;
    vertRect->height = rect.height;

/* do the horizontal rectangle */

    if ( y > 0) {	/* panned down */
	horizRect->y = 0;
	horizRect->height = y;
    } else {		/* panned up */
	horizRect->y = rect.height + y;
	horizRect->height = -y;
    }

    if ( x > 0) {	/* and right */
	horizRect->x = x;
	horizRect->width = rect.width - x;
    } else {		/* and left */
	horizRect->x = 0;
	horizRect->width = rect.width;
    }

/* translate the rectangles into the screen coordinate system of the plot */

    vertRect->x += rect.x;
    vertRect->y += rect.y;
    horizRect->x += rect.x;
    horizRect->y += rect.y;

/* change the coordinates of the plot to reflect the pan, but we don't
   want to update, because that's separate */

    GetDataPt( -x, -y, &newXmin, &newYmin);
    GetDataPt( 0, 0, &oldXmin, &oldYmin);
    xShift = newXmin - oldXmin;
    yShift = newYmin - oldYmin;
    SetCoords( coords.currXmin + xShift, coords.currYmin + yShift,
	       coords.currXmax + xShift, coords.currYmax + yShift);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a coordinate system, given the tag-values stored 
 *	in a tag values array.  First, the screen object fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int
AmCoordClass::setFields( AmTagValuePtr tagValues,
                       int         numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs;
    unitType	oldXunits = xUnits, oldYunits = yUnits;

    numValues = AmScreenObjClass::setFields( tagValues, numValues);
    AmCoordClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* if the projection has changed, then reset the projected coordinates
   and extents */

    if ( AmChanged( &projStrs)) {
	numProjStrs = AmVeclen(( const char **) projStrs);
	fwdProject.SetProjection(( const char **) projStrs);
	projCoords = coords;
	fwdProject.ProjectPt( &projCoords.currXmin, &projCoords.currYmin);
	fwdProject.ProjectPt( &projCoords.currXmax, &projCoords.currYmax);
	fwdProject.ProjectPt( &projCoords.xmin, &projCoords.ymin);
	fwdProject.ProjectPt( &projCoords.xmax, &projCoords.ymax);
    }
    if ( AmChanged( &inProjStrs)) {
	numInProjStrs = AmVeclen(( const char **) inProjStrs);
	rvsProject.SetProjection(( const char **) inProjStrs);
    }

/* if the current coordinates were set, then set the coords changed flag */

    if ( AmChanged( &coords.currXmin) && AmChanged( &coords.currYmin) &&
	 AmChanged( &coords.currXmax) && AmChanged( &coords.currYmax)) {
	coordsSet = TRUE;
    }

/* if the range was set, then set the coords changed flag */

    if ( AmChanged( &coords.xmin) && AmChanged( &coords.ymin) &&
	 AmChanged( &coords.xmax) && AmChanged( &coords.ymax)) {
	rangeSet = TRUE;
    }

/* if the unit type has changed, then convert the extents values */

    if ( AmChanged( &xUnits) && coordsSet) {
	AmConvertValue( oldXunits, xUnits, coords.xmin, &coords.xmin);
	AmConvertValue( oldXunits, xUnits, coords.xmax, &coords.xmax);
    }
    if ( AmChanged( &yUnits) && coordsSet) {
	AmConvertValue( oldYunits, yUnits, coords.ymin, &coords.ymin);
	AmConvertValue( oldYunits, yUnits, coords.ymax, &coords.ymax);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a coordinate system, given the tag-values 
 *	stored in a tag values array.  First, the screen object fields are 
 *	retrived, then the other fields are retrieved using the
 *	GetFieldByDefintions() function which uses a lookup table from 
 *	the tag strings to the structure offsets.  The line width bin 
 *	vector is retrieved by hand because the length is one less than 
 *	the line width vector.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int
AmCoordClass::getFields( AmTagValuePtr tagValues,
                       int         numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs;

    numValues = AmScreenObjClass::getFields( tagValues, numValues);
    AmCoordClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* we have to set the projection strings by hand, because for some reason
   offsetof doesn't work properly with an inherited class */

    return( numValues);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a coordinate system.  This is
 *      used in the map editor to dump the objects's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmCoordClass::dumpFields( AmFieldLengthPtr	fieldLengths,
			AnvilBoolean		allFields)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, numStrings;

    numStrings = AmScreenObjClass::dumpFields( fieldLengths, allFields);
    AmCoordClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = AmCoordClass::DumpFieldsByDefinitions( fieldLengths, numStrings,
						      fieldDefs, numFieldDefs);
    return( numStrings);
}

/*************************************************************************
 *
 * Description:
 * 	If the coordinate extents represent a matrix, return the correct 
 *	row and column. 
 * Arguments:
 *	x,y: position in the coordinate system.
 *	rows, cols: dimensions of the matrix the coordinate system
 *	represents.
 *	flipRows, flipCols: flags to transpose the matrix
 *	xpos, ypos: column, row position in the matrix.
 * Returns:
 *	TRUE if the point is in the matrix bounds, FALSE otherwise.
 *
 *************************************************************************/

AnvilBoolean
AmCoordClass::GetRowcol( float   x,
                       float   y,
                       int     rows,
                       int     cols,
                       AnvilBoolean flipRows,
                       AnvilBoolean flipCols,
                       int     *xpos,
                       int     *ypos)
{
    return( AmGetRowcolumn( x, y, rows, cols, flipRows, flipCols, 
		          coords.xmin, coords.ymin,
                          coords.xmax, coords.ymax,
			  xpos, ypos));
}

/*************************************************************************
 *
 * Description:
 * 	Lower level get row/col function, where the matrix extents 
 *	(in coordinate space) are passed as arguments. 
 * Arguments:
 *	x,y: position in the coordinate system.
 *	rows, cols: dimensions of the matrix the coordinate system
 *	represents.
 *	flipRows, flipCols: flags to transpose the matrix
 *	xmin, ymin, xmax, ymax: extents of the coordiante system.
 *	xpos, ypos: column, row position in the matrix.
 * Returns:
 *	TRUE if the point is in the matrix bounds, FALSE otherwise.
 *
 *************************************************************************/

AnvilBoolean
AmGetRowcolumn( float   x,
              float   y,
              int     rows,
              int     cols,
              AnvilBoolean flipRows,
              AnvilBoolean flipCols,
              float   xmin,
              float   ymin,
              float   xmax,
              float   ymax,
              int     *xpos,
              int     *ypos)
{
    if ( flipCols) {
	*xpos = AmFloatScaleToInt( xmin, xmax, x, cols - 1, 0);
    } else {
	*xpos = AmFloatScaleToInt( xmin, xmax, x, 0, cols - 1);
    }
    if ( flipRows) {
	*ypos = AmFloatScaleToInt( ymin, ymax, y, 0, rows - 1);
    } else {
	*ypos = AmFloatScaleToInt( ymin, ymax, y, rows - 1, 0);
    }
    return(( *xpos >= 0) && ( *xpos < cols) &&
	   ( *ypos >= 0) && ( *ypos < rows));
}

/*************************************************************************
 *
 * Description:
 * 	If the image hasn't been initialized, then create it, or destroy 
 *	and recreate it if resized. It's in here, because AmOverlayClass 
 *	roots from AmCoordClass.
 * Arguments:
 *	pImage: current display image.
 * Returns:
 *	new display image
 *
 *************************************************************************/

AmImageClass *
AmCoordClass::ResizeDisplayImage( AmImageClass *pImage)
{
    AmRectangleRec const *pRect;

/* If the image hasn't been initialized, then create it, or destroy and
   recreate it if resized, or if the DPI has changed. */

    pRect = GetRect();

    if ( pImage != NULL) {
        if (( pImage->GetWidth() != pRect->width) ||
            ( pImage->GetHeight() != pRect->height) ||
	    ( pImage->GetDPI() != pImage->GetDisplay()->GetDPI())) {
            delete pImage;
            pImage = NULL;
        }
    }

/* If the image has been resized, then reset the exposure to cover the
   entire plot rectangle. */

    if (( pImage == NULL) && ( pRect->width > 0) && ( pRect->height > 0)) {
        pImage = display->CreateImage( pRect->width, pRect->height,
                                       display->GetDepth(), NULL);
    }

    return( pImage);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a coordinate system. This is 
 *	used by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmCoordClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.
 *	THIS TABLE MUST BE SORTED ALPHABETICALLY ON THE TAG NAME FOR
 *	SEARCHING. 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmCoordClass::GetFieldDefinitions( AmFieldDefPtr	*fields,
				      int		*numDefs)
{
    static AmFieldDefRec CoordClassFieldDefinitions[] = {
	{ AmNaspectFactor, offsetof( AmCoordClass, aspectFactor) },
	{ AmNcurrXmax, offsetof( AmCoordClass, coords.currXmax) },
	{ AmNcurrXmin, offsetof( AmCoordClass, coords.currXmin) },
	{ AmNcurrYmax, offsetof( AmCoordClass, coords.currYmax) },
	{ AmNcurrYmin, offsetof( AmCoordClass, coords.currYmin) },
	{ AmNinProjection, offsetof( AmCoordClass, inProjStrs), 
			   offsetof( AmCoordClass, numInProjStrs), AmDYNAMIC },
	{ AmNnumInProjStrs, offsetof( AmCoordClass, numInProjStrs) },
	{ AmNnumProjStrs, offsetof( AmCoordClass, numProjStrs) },
	{ AmNpreserveAspect, offsetof( AmCoordClass, preserveAspect) },
	{ AmNprojCurrXmax, offsetof( AmCoordClass, projCoords.currXmax) },
	{ AmNprojCurrXmin, offsetof( AmCoordClass, projCoords.currXmin) },
	{ AmNprojCurrYmax, offsetof( AmCoordClass, projCoords.currYmax) },
	{ AmNprojCurrYmin, offsetof( AmCoordClass, projCoords.currYmin) },
	{ AmNprojXmax, offsetof( AmCoordClass, projCoords.xmax) },
	{ AmNprojXmin, offsetof( AmCoordClass, projCoords.xmin) },
	{ AmNprojYmax, offsetof( AmCoordClass, projCoords.ymax) },
	{ AmNprojYmin, offsetof( AmCoordClass, projCoords.ymin) },
	{ AmNprojection, offsetof( AmCoordClass, projStrs),
			 offsetof( AmCoordClass, numProjStrs), AmDYNAMIC },
	{ AmNxmax, offsetof( AmCoordClass, coords.xmax) },
	{ AmNxmin, offsetof( AmCoordClass, coords.xmin) },
	{ AmNxunits, offsetof( AmCoordClass, xUnits) },
	{ AmNymax, offsetof( AmCoordClass, coords.ymax) },
	{ AmNymin, offsetof( AmCoordClass, coords.ymin) },
	{ AmNyunits, offsetof( AmCoordClass, yUnits) }
    };

    *numDefs = VECLEN( CoordClassFieldDefinitions);
    *fields = CoordClassFieldDefinitions;
}
