#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/log.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/display.h>
#include <anvil/userevents.h>
#include <anvil/tagstrs.h>
#include <anvil/fns.h>
#include <anvil/datagraphic.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmDataGraphicClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmDATA_GRAPHIC) {
        return( TRUE);
    } else {
        return( AmCoordClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 * 	return the length of the data being read.  product the
 *	vector of dimensions in the dataInfo structure.
 * Arguments:
 *	dataInfo structure, containing the vector of dimensions for the
 *	datagraphic.
 * Returns:
 *	the length of the data.
 *
 ****************************************************************************/

int 
AmDataLength( AmDataInfoPtr	dataInfo)
{
    int	ii, length;

    length = 1;
    for ( ii = 0; ii < dataInfo->numdims; ii++) {	
	length *= dataInfo->dims[ ii];
    }
    return( length);
}

/****************************************************************************
 *
 * Description:
 *      Empty Constructor
 *
 ****************************************************************************/


AmDataGraphicClass::AmDataGraphicClass() 
{
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a data graphic.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify data graphic for resources
 *
 ****************************************************************************/

void
AmDataGraphicClass::initializeFields( AmWindowClass	*windowArg,
				    const char	 *nameArg)
{
    const char	*szColor;
    int		ii;

    AmScreenObjClass::initializeFields( windowArg, nameArg);
    rect.x = 0;
    rect.y = 0;
    rect.width = 100;
    rect.height = 100;

/* set all of the fields for the coordinate system and display rectangle */

    setCoordClass();

/* set all the basic plot attributes */

    fixedExtents = FALSE;
    appId = NULL;
    priority = 0;
    selectionSet = NULL;
    numSelected = 0;
    title = NULL;

/* default the text string stuff */

    textOffset = 0;
    textSize = 0;
    showText = FALSE;

/* turn off tiles... */

    tileHeader = NULL;
    detailLevel = 1;			/* 1:1 detail level */

/* set the selection set color */

    szColor = AmGetDefault( "select.color", "red");
    if ( !display->AllocNamedColor( szColor, &highlightColor)) {
	AN_DEBUG( 10, ( "failed to allocate highlight color %s\n", szColor));
    }

/* set all the default fields for dataInfo */

    dataInfo.sourceType = AmNO_SOURCE;	    /* default is no data */
    dataInfo.skipBytes = 0;
    dataInfo.filename = NULL;
    dataInfo.fd = -1;
    dataInfo.numdims = 1;
    dataInfo.data = NULL;
    dataInfo.shared = FALSE;
    dataInfo.xFn = ( AmAccessFn) NULL;
    dataInfo.yFn = ( AmAccessFn) NULL;
    dataInfo.xOffset = 0;
    dataInfo.yOffset = sizeof( float);
    dataInfo.xType = AmFLOAT;
    dataInfo.yType = AmFLOAT;
    dataInfo.size = 0;
    dataInfo.descripFile = NULL;
    dataInfo.clientData = NULL;
    dataInfo.dataType = AmNO_TYPE;
    dataInfo.enableCacheFlag = TRUE;
    for ( ii = 0; ii < MAXDIMS; ii++) {
	dataInfo.dims[ ii] = 0;
    }
}

/****************************************************************************
 *
 * Description:
 *      check values for a data graphic.  Each check values function calls 
 *	the check values function for the parent class and is used to set
 *      the object's IsValid() flag.  The check values function writes
 *	details about the errors to the log file.
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmDataGraphicClass::checkValues()
{
    AnvilBoolean	okFlag;

    okFlag = TRUE;
    if ( !AmCoordClass::checkValues()) {
	okFlag = FALSE;
    }
    if ( dataInfo.sourceType == AmERROR_SOURCE) {
	AmAddResourceToLog( "sourceTypeError");
	okFlag = FALSE;
    }
    if ( dataInfo.size <= 0) {
	AmAddResourceToLog( "sizeLessThanEqualZero", dataInfo.size);
	okFlag = FALSE;
    }

/* if the source type was set to read source, file source, or pyramid source
   and there's no filename, there's an error */

    if ((( dataInfo.sourceType == AmREAD_SOURCE) ||
	 ( dataInfo.sourceType == AmFILE_SOURCE) ||
	 ( dataInfo.sourceType == AmMMAP_SOURCE) ||
	 ( dataInfo.sourceType == AmPYRAMID_SOURCE)) && 
        ( dataInfo.filename == NULL)) {
	AmAddResourceToLog( "noFilename", dataInfo.sourceType);
	okFlag = FALSE;
    }
    return( okFlag);
}

/****************************************************************************
 *
 * Description:
 * 	Destructor: free the pertinent information in a data graphic, 
 *	close any open files, free strings, remove the data from
 *	the cache, and free any temporary buffers maintained by the
 *	pyramiding algorithm.
 *
 ****************************************************************************/

AmDataGraphicClass::~AmDataGraphicClass()
{ 

/* close the file descriptor if it's still open and it's not shared with 
   another display graphic */

    if (( dataInfo.fd != -1) && !dataInfo.shared) {
	close( dataInfo.fd);
    }

/* free the filename, appId, and title, and description */ 

    if ( appId != NULL) {
	FREE( appId);
    }
    if ( title != NULL) {
	FREE( title);
    }
    if ( dataInfo.descripFile != NULL) {
	FREE( dataInfo.descripFile);
    }

/* free the selection set if it's not NULL */
   
    if ( selectionSet != NULL) {
	FREE( selectionSet);
    }

/* if the data isn't shared, and we've still got a copy of it, then 
   free it here */

    if ( dataInfo.sourceType == AmFILE_SOURCE) {
        if (( dataInfo.data != NULL) && !dataInfo.shared) {
            AmRemoveFromCache( dataInfo.filename, dataInfo.data);
        }
    } else if ( dataInfo.sourceType == AmDIRECT_SOURCE) {
        if (( dataInfo.data != NULL) && !dataInfo.shared) {
	    FREE( dataInfo.data);
	}
    } else if (( dataInfo.sourceType == AmPYRAMID_SOURCE) ||
	       ( dataInfo.sourceType == AmREAD_SOURCE)) {

/* if we're keeping around a temporary data pointer, blow it away */

	if ( dataInfo.data != NULL) {
	    FREE( dataInfo.data);
	}
    }

    if ( dataInfo.filename != NULL) {
	FREE( dataInfo.filename);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Perform autoscaling on the data used for an index data structure.
 *	This uses the accessing function, offset and client data from
 *	the index data structure, and the size from dataInfo, and returns the
 *	min and max for the index data.  This should only be applied to
 *	data which is read in its entirety, i.e. AmFILE_SOURCE or 
 *	AmDIRECT_SOURCE>
 * Arguments:
 *	indexData: indexData structure to set the min and max for.
 *	dataInfo: describes the data to be scaled.
 *	min, max: returned min and max values.
 *
 ****************************************************************************/

static void 
scaleData( AmIndexDataPtr	indexData, 
	   AmDataInfoPtr	dataInfo,	
	   float 	*min,
	   float	*max)
{
    uchar_t	*ptr;
    int		length, ii;
    float	val;
    
    *min = MY_HUGE; 
    *max = -MY_HUGE;
    length = AmDataLength( dataInfo);
    for ( ii = 0, ptr = dataInfo->data; 
	  ii < length; 
	  ii++, ptr += dataInfo->size) {
	val = ( indexData->fn)( ptr + indexData->offset, indexData->clientData);
	if ( val < *min) {
	    *min = val;
	} else if ( val > *max) {
	    *max = val;
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	ranging function for data which has x,y coordinates, such as
 *	scatter plots, line plots, and polygon plots.  It uses the
 *	xFn, yFn, and clientData from the dataInfo structure, as well
 *	as the size and data length.
 * Arguments:
 *	xmin, ymin, xmax, ymax - returned data extents 
 *
 ****************************************************************************/
 
void 
AmDataGraphicClass::GetDataExtents( float    *xmin,
				  float    *ymin,
				  float    *xmax,
				  float    *ymax)
{
    int         ii, length = AmDataLength( &dataInfo);
    uchar_t     *ptr;
    float       x, y;

    if (( dataInfo.xFn == NULL) || ( dataInfo.yFn == NULL)) {
	return;
    }
    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0, ptr = dataInfo.data;
          ii < length;
          ii++, ptr += dataInfo.size) {
        x = ( dataInfo.xFn)( ptr + dataInfo.xOffset, dataInfo.clientData);
        y = ( dataInfo.yFn)( ptr + dataInfo.yOffset, dataInfo.clientData);
        if ( x < *xmin) {
            *xmin = x;
        }
        if ( y < *ymin) {
            *ymin = y;
        }
        if ( x > *xmax) {
            *xmax = x;
        }
        if ( y > *ymax) {
            *ymax = y;
        }
    }
}

/****************************************************************************
 *
 * Description:
 * 	read data from a file into a block of memory.  First, check if
 *	the file is in the cache, and if so, then use the cache pointer
 *	otherwise, open the file, seek over any header bytes, 
 *	allocate the data and read it.
 * Arguments:
 *	uses the dataInfo substructure.
 * Returns:
 *	AmOK if everything's OK, system error otherwise 
 *
 ****************************************************************************/

AmErrorCode 
AmDataGraphicClass::readData()
{
    int		actual, length, fd;
    uchar_t	*data;

/* find out if the data is in the cache already, but only if caching is
   enabled */

    if ( dataInfo.enableCacheFlag) {
	if (( data = ( uchar_t *) AmGetCache( dataInfo.filename)) != NULL) {
	    dataInfo.data = data;
	    return( AmOK);
	}
    }

/* wasn't in the cache */

    if (( fd = open( dataInfo.filename, O_RDONLY)) < 0) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    length = AmDataLength( &dataInfo)*dataInfo.size;
    if (( dataInfo.data = ( uchar_t *) malloc( length)) == NULL) {
	close( fd);
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if ( lseek( fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if (( actual = read( fd, dataInfo.data, length)) < length) {
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    close( fd);
    if ( dataInfo.enableCacheFlag) {
	AmAddToCache( dataInfo.filename, dataInfo.data);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      retreive values for a data graphic.  Each getFields() function calls
 *      getFields() for the parent class, then most of the fields are
 *      retrived using the GetFieldByDefinitions() function.  
 * Arguments:
 *      tagValues - vector of tag-values to retrieve.  The values are
 *      pointers to the results.  The array is "scrunched" in place
 *      as each value is retreived.  Any unprocessed values are returned
 *      in place in the array.
 *      numValues - number of values in the array.
 * Returns:
 *      number of unprocessed values.
 *
 ****************************************************************************/

int 
AmDataGraphicClass::getFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;

/* get the fields for the parent class ( coordinate system) */

    numValues = AmCoordClass::getFields( tagValues, numValues);

/* get all the easy to specify fields for this class */

    AmDataGraphicClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *      Dump the field names used in a datagraphic.  dumpFields() is called
 *      recursively by each object and its parent.  The fieldLengths
 *      array returns the name of each field, and if it's a vector
 *      field, the vector's length.
 * Arguments:
 *      fieldLengths - vector of tags and vector lengths.
 *      allFields - an obsolete tag which was used to differentiate
 *      AmBetween dumping to a file or copying the object.
 * Returns:
 *      The number of fields dumped
 *
 ***************************************************************************/

int
AmDataGraphicClass::dumpFields( AmFieldLengthPtr	fieldLengths,
			      AnvilBoolean		allFields)
{
    int		numFieldDefs, numStrings;
    AmFieldDefPtr	fieldDefs;
    char	*exceptFields[] = { AmNdata, AmNclientData, NULL },
		**fields;

/* get the fields for the parent class ( coordinate system) */

    numStrings = AmCoordClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmDataGraphicClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    if ( allFields) {
	fields = NULL;
    } else {
    	fields = exceptFields;
    }
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
				          fieldDefs, numFieldDefs, fields);
    return( numStrings);
}

/****************************************************************************
 *
 * Description:
 *      Read function, uses the fields in the data info structure
 *      to read the data.  Handles AmFILE_SORUCE, AmREAD_SOURCE and
 *	AmDIRECT_SOURCE.
 * Returns:
 *      AmOK if everything was read OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmDataGraphicClass::getDataFromDescription()
{
    int		length;
    AmErrorCode	errCode;

/* if the sourceType is direct source, then we're already taken care of.  We
   now deal with the file source */

    length = AmDataLength( &dataInfo)*dataInfo.size;
    if ( dataInfo.sourceType == AmFILE_SOURCE) {

/* read the data as specified by dataInfo.  We maintain a global list
   of all the files read in, so that we can keep a reference count, and not 
   duplicate data */

	if (( errCode = readData()) != AmOK) {
	    return( errCode);
	}

/* if we read it on the fly, then we just open the file descriptor here */

    } else if ( dataInfo.sourceType == AmREAD_SOURCE) {

/* set the data segment to NULL, because we're going to read it on init */

    	dataInfo.data = NULL;
	if (( dataInfo.fd = open( dataInfo.filename, O_RDONLY)) < 0) {
	    DO_RETURN_ERROR( AmOPEN_FAILED);
	}

/* if it's a data pointer, then we add the datapointer to the cache */

    } else if (( dataInfo.sourceType == AmDIRECT_SOURCE) &&
	       ( dataInfo.filename != NULL)) {
	if ( dataInfo.enableCacheFlag) {
	    AmAddToCache( dataInfo.filename, dataInfo.data);
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	initialize the index data to some reasonable values.  This 
 *	should be used for each indexData structure which maps the data
 *	through a function to a graphics attribute.
 * Arguments:
 *	indexData - structure to initialize. 
 *
 ****************************************************************************/

void 
AmInitIndexData( AmIndexDataPtr	indexData)
{
    indexData->indexType = AmINDEX_LINEAR;
    indexData->min = 0.0;
    indexData->max = 1.0;
    indexData->scale = 1.0;
    indexData->fn = NULL;
    indexData->AmIntFn = NULL;
    indexData->clientData = NULL;
    indexData->clientDataFreeFn = NULL;
    indexData->length = 0;
    indexData->binvec = NULL;
    indexData->table = NULL;
    indexData->offset = 0;
    indexData->dataType = AmNO_TYPE;
    indexData->getIndexFn = NULL;
    indexData->units = AmUNIT_UNDEFINED;
    indexData->strings = NULL;
    indexData->minFixed = FALSE;
    indexData->maxFixed = FALSE;
    indexData->minClip = FALSE;
    indexData->maxClip = FALSE;
    indexData->name = NULL;
}

/****************************************************************************
 *
 * Description:
 * 	trash any allocated fields in an indexData structure.  Free the
 *	binvector, table and strings if they aren't NULL  
 * Arguments:
 *	indexData - structure to destroy. 
 *
 ****************************************************************************/

void 
AmDestroyIndexData( AmIndexDataPtr	indexData)
{
    if ( indexData->binvec != NULL) {
	free( indexData->binvec);
    }
    if ( indexData->table != NULL) {
	free( indexData->table);
    }
    if ( indexData->strings != NULL) {
	AmFreeStrVec( indexData->strings);
    }
    if ( indexData->name != NULL) {
	free( indexData->name);
    }
    if (( indexData->clientDataFreeFn != NULL) && 
	( indexData->clientData != NULL)) {
	( indexData->clientDataFreeFn)( indexData->clientData);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Update the dependent fields in an indexData structure.  If 
 *	there's an integer table, then find the integer lookup function.
 *	If the accessing function isn't defined, and the type is, then
 *	set the accessing function from the type.  If the bin vector
 *	is NULL. then set the fixed flags for the min and max if they've
 *	been set, otherwise, auto-range the data if it's been read in,
 *	and set the scale for speed.
 * Arguments:
 *	indexData - structure to update.
 * Returns:
 *	AmOK if everything's OK, errors if functions couldn't be found
 *	or if the data's unavailable for ranging.
 *
 ****************************************************************************/

AmErrorCode 
AmDataGraphicClass::updateIndexData( AmIndexDataPtr	indexData,
				   const char	*tag)
{
    float	min, max;

/* if there's a table, then it's an integer table lookup function */

    if ( indexData->table != NULL) {
	if ( indexData->AmIntFn == NULL) {

/* AmN..Fn is used for integer functions as well as floats... */

	    if ( indexData->fn != NULL) {
		indexData->AmIntFn = ( AmIntAccessFn) indexData->fn;
	    } else {
		indexData->AmIntFn = 
		    AmDataTypeEnumToIntAccessFn( indexData->dataType);
	    }
	    AmMarkChanged( &indexData->AmIntFn);
	}
	if ( indexData->AmIntFn == NULL) {
	    AmAddResourceToLog( "badAccessFunction", tag);
	    DO_RETURN_ERROR( AmBAD_ACCESS_FUNCTION);
	} 
	indexData->getIndexFn = AmTableLookupFn;
	indexData->indexType = AmINDEX_LOOKUP;
    } else {

/* if there's no accessing function defined, then try to divine one from
   the data type */

	if ( indexData->fn == NULL) {
	    indexData->fn = AmDataTypeEnumToAccessFn( indexData->dataType);
	    AmMarkChanged( &indexData->fn);
	}
	if ( indexData->fn == NULL) {
	    AmAddResourceToLog( "badAccessFunction", tag);
	    DO_RETURN_ERROR( AmBAD_ACCESS_FUNCTION);
	}

/* if there's a bin vector, set the indexing function to non-linear.  If 
   there's a table, then use a table lookup indexing function */

	if ( indexData->binvec != NULL) {
	    indexData->getIndexFn = AmNonlinearIndexFn;
	    indexData->indexType = AmINDEX_NONLINEAR;
	} else {

/* if the min and max values aren't already fixed, then see if they've been set.
   If set, then set the fixed flags as well. Set the linear indexing function
   and its scale */

	    indexData->getIndexFn = AmLinearIndexFn;
	    indexData->indexType = AmINDEX_LINEAR;
	    if ( !indexData->minFixed) {
		indexData->minFixed = AmChanged( &indexData->min);
	    }
	    if ( !indexData->maxFixed) {
		indexData->maxFixed = AmChanged( &indexData->max);
	    }

/* auto-scaling can only be applied if the data is in RAM from AmFILE_SOURCE
   or AmDIRECT_SOURCE, and the data is not NULL */

	    if ( !( indexData->minFixed || indexData->maxFixed) &&
		 ( dataInfo.sourceType != AmREAD_SOURCE) && 
		 ( dataInfo.sourceType != AmPYRAMID_SOURCE) &&
		 ( dataInfo.data != NULL)) {
		scaleData( indexData, &dataInfo, &min, &max);
		indexData->min = min;
		AmMarkChanged( &indexData->min);		/* mark for caller */
		indexData->max = max;
		AmMarkChanged( &indexData->max);		/* mark for caller */
	    } 

/* set the scale anyway, because we try to set min & max to something OK.
   the top and bottom colors are for values which are out of range, so the
   values index from 1 to numColors - 2, below min indexes 0, and above max
   indexes numColors - 1 */

	    indexData->scale = ( indexData->length - 2)/
				    ( indexData->max - indexData->min);
	}
    } 
    return( AmOK);
}

/* set fields wrapper for data graphics */

int 
AmDataGraphicClass::setFieldsWrapper( AmTagValuePtr       tagValues,
				    int               numValues)
{

/* call the record class setFieldsWrapper function, which sets
   up the track changes, and sets the extra fields.  WE DO NOT
   call the AmScreenObjClass::setFieldsWrapper() because it sends user
   events that we don't want received until all of our fields are
   updated */

    numValues = AmRecordClass::setFieldsWrapper( tagValues, numValues);

/* call the type-recursive checkValues() function to ensure that
   everything is set properly. */

    if ( !checkValues()) {
	SetValid( FALSE);
	return( numValues);
    } 

/* call getDataFromDescription(), a virtual function for the subclass,
   to read the data. */

    if ( dataInfo.sourceType != AmNO_SOURCE &&
         dataInfo.sourceType != AmCUSTOM) {

/* read the data if any of the pertinent fields have changed */

	if ( AmChanged( &dataInfo.data, &dataInfo.dims, &dataInfo.dims[ 0],
		      &dataInfo.dims[ 1], &dataInfo.size, &dataInfo.filename, 
		      &dataInfo.sourceType,
		      NULL)) {

	  /* get the data, and make sure everything is OK */

	    if ( getDataFromDescription() != AmOK) {
		SetValid( FALSE);
		if ( dataInfo.filename != NULL) {
		    AmAddResourceToLog( "badFile", dataInfo.filename);
		} else {
		    AmAddResourceToLog( "badInput");
		}
		return( numValues);
	    }
	} 

    } 

/* this is a hack put in because of copying pyramid files.  If the data 
   is specified and the source type is AmPYRAMID_SOURCE, we want to re-set
   the data to NULL, otherwise it will be freed whenever the pyramid is
   changed, reallocated or resized.  It's basically shared from the 
   other plot... (this is for magic windows)... */

    if (( dataInfo.sourceType == AmPYRAMID_SOURCE) && 
	AmChanged( &dataInfo.data)) {
	dataInfo.data = NULL;
    }

/* call the updateFields functions, which updates indexData structures 
   and other information that was set. */

    updateFields();

/* send an event for the object, and one for the parent if it is an overlay,
   and the object is a data graphic contained inside of it */

    AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, this);
    if ( IsType( AmDATA_GRAPHIC) && ( getParent() != NULL) &&
        ( getParent()->GetObjectType() == AmOVERLAY)) {
        AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, getParent());
    }
    return( numValues);
}

/* Data GraphicClass updates no fields. */

void 
AmDataGraphicClass::updateFields()
{
}

/****************************************************************************
 *
 * Description:
 *      set the fields in the datagraphic from the tag-value argument list.  
 *	Each class calls setFields() for the parent class, then uses
 *      SetFieldByDefinitions() to set its own members.  Members which
 *      can't be processed using the SetFieldsBy...() are processed by
 *      hand, such as the palette and bin vector.  The datagraphic
 *	setFields function uses the AmChanged() directive (which tracks
 *	the changed fields) to update fields which are dependent on
 *	the values which have been set.
 * Arguments:
 *      tagValues: tag-value pairs to be parsed (this array is reduced in
 *                 place, with the parsed arguments removed).
 *      numValues: length of the tagValues array.
 * Return:
 *      number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int 
AmDataGraphicClass::setFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    int			numFieldDefs;
    AmFieldDefPtr		fieldDefs;

/* get the fields for the parent class ( coordinate system) */

    numValues = AmCoordClass::setFields( tagValues, numValues);

/* get all the easy to specify fields for this class 
   take all the values which weren't found and copy them to the head of
   the array and keep a count, so they can be processed by the plot
   specific functions afterwards */

    AmDataGraphicClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* if the filename or data has been set, and the source type is still 
   AmNO_SOURCE, then set it appropriately.
   NOTE: I still need a way to support AmMMAP_SOURCE.  AmDIRECT_SOURCE
   takes precedence over AmFILE_SOURCE */

    if ( dataInfo.sourceType == AmNO_SOURCE) {
	if ( AmChanged( &dataInfo.data) && ( dataInfo.filename == NULL)) {
	    dataInfo.sourceType = AmDIRECT_SOURCE;
	} else if ( AmChanged( &dataInfo.filename) && 
		    ( dataInfo.filename != NULL)) {
	    dataInfo.sourceType = AmFILE_SOURCE;
	    dataInfo.data = NULL;		
	}
    }

/* convert the types to access functions */

    if ( AmChanged( &dataInfo.xType) && ( dataInfo.xFn == NULL)) {
	dataInfo.xFn = AmDataTypeEnumToAccessFn( dataInfo.xType);
    }
    if ( AmChanged( &dataInfo.yType) && ( dataInfo.yFn == NULL)) {
	dataInfo.yFn = AmDataTypeEnumToAccessFn( dataInfo.yType);
    }

/* if the coordinates have been specified, then set them explicitly here */

    if ( AmChanged( &coords.xmin, &coords.ymin, 
		  &coords.xmax, &coords.ymax, NULL)) {
	SetRange( coords.xmin, coords.ymin, coords.xmax, coords.ymax);
	fixedExtents = TRUE;
    }
    if ( AmChanged( &coords.currXmin, &coords.currYmin, 
		  &coords.currXmax, &coords.currYmax, NULL)) {
	SetCoords( coords.currXmin, coords.currYmin, 
		   coords.currXmax, coords.currYmax);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	return the plot's priority (used for layer ordering in overlays)
 * Returns:
 *	the priority value.
 *
 ****************************************************************************/

int 
AmDataGraphicClass::GetPriority()
{
    return( priority);
}

/****************************************************************************
 *
 * Description:
 * 	assign the selection set and send a user event.  The user event is 
 *	received by Flush() and the plot's draw routine draws the selection 
 *	set 
 * Arguents:
 *	items: the items to set in the selection set.
 *	numItems: the number of items set.
 *
 ****************************************************************************/

void 
AmDataGraphicClass::SetSelection( void 	*items,
				int	numItems)
{
    if ( selectionSet != NULL) {
	free( selectionSet);
    }
    selectionSet = items;
    numSelected = numItems;
    AN_DEBUG( 60, ( "%d objects selected\n", numItems));
}

/****************************************************************************
 *
 * Description:
 * 	return the current selection set.
 * Arguents:
 *	items: the items to retreive from the selection set.
 *	numItems: the number of items returned.
 *
 ****************************************************************************/

void 
AmDataGraphicClass::GetSelection( void 	**items,
				int	*numItems)
{
    *items = selectionSet;
    *numItems = numSelected;
}

/****************************************************************************
 *
 * Description:
 * 	delete the current selection from the data graphic.  Walk through
 *	the selection set (which should be in the same ordering as the 
 *	data), and delete any matching values, as tested by memcmp().
 *
 ****************************************************************************/

void 
AmDataGraphicClass::DeleteSelection()
{
    uchar_t	*ptr, *procPtr, *unprocPtr;
    int		ii;

    if ( numSelected > dataInfo.dims[ 0]) {
	return;
    }
    ptr = ( uchar_t *) selectionSet;
    procPtr = ( uchar_t *) dataInfo.data;
    for ( ii = 0, unprocPtr = dataInfo.data;
	  ii < dataInfo.dims[ 0];
	  ii++, unprocPtr += dataInfo.size) {
	if ( !memcmp( unprocPtr, ptr, dataInfo.size)) {
	    ptr += dataInfo.size;
	} else {
	    memcpy( procPtr, unprocPtr, dataInfo.size);
	    procPtr += dataInfo.size;
	}
    }
    dataInfo.dims[ 0] -= numSelected;
}

/****************************************************************************
 *
 * Description:
 * 	replace the current selection from the data graphic with a new
 *	selection set.  Walk through the selection set (which should be in 
 *	the same ordering as the data), and delete any matching values, as 
 *	tested by memcmp().
 * Arguents:
 *	replaceSelection: the items to replace in the selection set.
 *	numItems: the number of items to replace.
 *
 ****************************************************************************/

AmErrorCode 
AmDataGraphicClass::ReplaceSelection( void	*replaceSelection,
				    int		numItems)
{
    uchar_t	*ptr, *replacePtr, *procPtr, *unprocPtr;
    int		ii;

    if ( numItems > dataInfo.dims[ 0]) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    ptr = ( uchar_t *) selectionSet;
    replacePtr = ( uchar_t *) replaceSelection;
    procPtr = ( uchar_t *) dataInfo.data;
    for ( ii = 0, unprocPtr = dataInfo.data;
	  ii < dataInfo.dims[ 0];
	  ii++, unprocPtr += dataInfo.size) {
	if ( !memcmp( unprocPtr, ptr, dataInfo.size)) {
	    memcpy( unprocPtr, replacePtr, dataInfo.size);
	    ptr += dataInfo.size;
	} else {
	    memcpy( procPtr, unprocPtr, dataInfo.size);
	    procPtr += dataInfo.size;
	}
    }
    dataInfo.dims[ 0] -= numSelected;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Draw the current selection set.  The base function doesn't do
 *	anything. (it should probably be pure virtual, be we don't
 *	want to force people to implement things).
 * Arguents:
 *	w - window to draw to
 *	gc - graphics context to draw with.
 *	exposeRects - expose rectangles to draw
 *	nRects - number of expose rectangles.
 *	fHighlight - whether to draw in highlight or normal colors
 *
 ****************************************************************************/

AmErrorCode 
AmDataGraphicClass::DrawSelection( AmWindowClass		*w,
				 AmGCClass		*gc,
				 AmRectangleRec const	*exposeRects,
				 int			nRects,
				 AnvilBoolean		fHighlight)
{
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	clear the current selection and send a user event 
 *
 ****************************************************************************/

void 
AmDataGraphicClass::ClearSelection()
{
    if ( selectionSet != NULL) {
	free( selectionSet);
    }
    selectionSet = NULL;
    numSelected = 0;
}

/****************************************************************************
 *
 * Description:
 * 	set the range for a datagraphic and send an AM_COORDS_CHANGE user 
 *	event.
 *
 ****************************************************************************/

void 
AmDataGraphicClass::SetRange( float	xmin,
			    float	ymin,
			    float	xmax,
			    float	ymax)
{
    AmCoordClass::SetRange( xmin, ymin, xmax, ymax);
    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 * 	set the current extents for a datagraphic and send an 
 *	AM_COORDS_CHANGE user event.
 *
 ****************************************************************************/

void 
AmDataGraphicClass::SetCoords( float	currXmin,
			     float	currYmin,
			     float	currXmax,
			     float	currYmax)
{
    AmCoordClass::SetCoords( currXmin, currYmin, currXmax, currYmax);
    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
}

/* retrieve the field definitions for the plot class. THE LIST MUST BE
   SORTED BY ALPHABETICAL ORDER ON THE FIELD NAME.  This is used by
   SetFieldsByDefinitions() */ 

void 
AmDataGraphicClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDef,
				       int		*numFieldDef)
{
    static AmFieldDefRec DataGraphicClassFieldDefinitions[] = {
	{ AmNappId, offsetof( AmDataGraphicClass, appId), 0, AmDYNAMIC },
	{ AmNclientData, offsetof( AmDataGraphicClass, dataInfo.clientData) },
	{ AmNcols, offsetof( AmDataGraphicClass, dataInfo.dims[ 1]) },
	{ AmNdata, offsetof( AmDataGraphicClass, dataInfo.data) },
	{ AmNdataType, offsetof( AmDataGraphicClass, dataInfo.dataType) },
	{ AmNdescription, offsetof( AmDataGraphicClass, dataInfo.descripFile),
		       0, AmDYNAMIC },
	{ AmNdetailLevel, offsetof( AmDataGraphicClass, detailLevel) },
	{ AmNdims, offsetof( AmDataGraphicClass, dataInfo.dims), 
		   offsetof( AmDataGraphicClass, dataInfo.numdims), AmSTATIC },
	{ AmNenableCacheFlag, 
		   offsetof( AmDataGraphicClass, dataInfo.enableCacheFlag) },
	{ AmNfilename, offsetof( AmDataGraphicClass, dataInfo.filename), 
		       0, AmDYNAMIC },
	{ AmNhighlightColor, offsetof( AmDataGraphicClass, highlightColor) },
	{ AmNnumdims, offsetof( AmDataGraphicClass, dataInfo.numdims) },
	{ AmNpriority, offsetof( AmDataGraphicClass, priority) },
	{ AmNrows, offsetof( AmDataGraphicClass, dataInfo.dims[ 0]) },
	{ AmNshared, offsetof( AmDataGraphicClass, dataInfo.shared) },
	{ AmNshowText, offsetof( AmDataGraphicClass, showText) },
	{ AmNsize, offsetof( AmDataGraphicClass, dataInfo.size) },
	{ AmNskipBytes, offsetof( AmDataGraphicClass, dataInfo.skipBytes) },
	{ AmNsourceType, offsetof( AmDataGraphicClass, dataInfo.sourceType) },
	{ AmNtextOffset, offsetof( AmDataGraphicClass, textOffset) },
	{ AmNtextSize, offsetof( AmDataGraphicClass, textSize) },
	{ AmNtitle, offsetof( AmDataGraphicClass, title), 0, AmDYNAMIC },
	{ AmNxFn, offsetof( AmDataGraphicClass, dataInfo.xFn) },
	{ AmNxOffset, offsetof( AmDataGraphicClass, dataInfo.xOffset) },
	{ AmNxType, offsetof( AmDataGraphicClass, dataInfo.xType) },
	{ AmNyFn, offsetof( AmDataGraphicClass, dataInfo.yFn) },
	{ AmNyOffset, offsetof( AmDataGraphicClass, dataInfo.yOffset) },
	{ AmNyType, offsetof( AmDataGraphicClass, dataInfo.yType) },
    };
    *fieldDef = DataGraphicClassFieldDefinitions;
    *numFieldDef = VECLEN( DataGraphicClassFieldDefinitions);
}
