#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* graphs are similar to plots, except that they allow different coordinate
   systems on the X and Y axes.  The base graph class is rooted off of 
   data graphic. */

#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/resource.h>
#include <anvil/record.h>
#include <anvil/graph.h>
#include <anvil/log.h>

extern const int AmGRAPH_HEIGHT = 200,
		 AmGRAPH_WIDTH = 300;

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmGraphClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmGRAPH) {
        return( TRUE);
    } else {
        return( AmDataGraphicClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Free the heightIndex index data structure.
 *
 ****************************************************************************/

AmGraphClass::~AmGraphClass()
{
    AmDestroyIndexData( &heightIndex);
}

/****************************************************************************
 *
 * Description:
 *      Empty Constructor
 *
 ****************************************************************************/

AmGraphClass::AmGraphClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	get data extents for graphs is a little different than for data 
 *	graphics, because the heightIndex is used for the y coordinate 
 * Arguments:
 *	xmin, ymin, xmax, ymax: resulting range of the data.
 *
 ****************************************************************************/

void 
AmGraphClass::GetDataExtents( float	*xmin,
			    float	*ymin,
			    float	*xmax,
			    float	*ymax)
{
    int		ii;
    float	x, y, length = AmDataLength( &dataInfo);
    uchar_t	*ptr;

    if ( heightIndex.fn == NULL) {
	return;
    }

/* if the data is set up in x,y pairs, then the x function is non-NULL, and
   the x and y values should be tested */

    if ( dataInfo.xFn != NULL) {
	*xmin = MY_HUGE;
	*ymin = MY_HUGE;
	*xmax = -MY_HUGE;
	*ymax = -MY_HUGE;
	for ( ii = 0, ptr = dataInfo.data;
	      ii < length;
	      ii++, ptr += dataInfo.size) {
	    x = ( dataInfo.xFn)( ptr + dataInfo.xOffset, 
				 dataInfo.clientData);
	    y = ( heightIndex.fn)( ptr + heightIndex.offset, 
				   dataInfo.clientData);
	    if ( x < *xmin) {
		*xmin = x;
	    }
	    if ( y < *ymin) {
		*ymin = y;
	    }
	    if ( x > *xmax) {
		*xmax = x;
	    }
	    if ( y > *ymax) {
		*ymax = y;
	    }
	}
    } else {

/* if there's no X function, then just test the y values, and set the xmin,
   xmax to the number of points */

	*xmin = 0;
	*ymin = MY_HUGE;
	*xmax = dataInfo.dims[ 0];
	*ymax = -MY_HUGE;
	for ( ii = 0, ptr = dataInfo.data;
	      ii < length;
	      ii++, ptr += dataInfo.size) {
	    y = ( heightIndex.fn)( ptr + heightIndex.offset, 
				   heightIndex.clientData);
	    if ( y < *ymin) {
		*ymin = y;
	    }
	    if ( y > *ymax) {
		*ymax = y;
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a graph.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify graph for resources
 *
 ****************************************************************************/

void 
AmGraphClass::initializeFields( AmWindowClass	*windowArg,
			      const char	*nameArg)
{
    AmDataGraphicClass::initializeFields( windowArg, nameArg);
    rect.width = AmGetDefault( "graph.width", AmGRAPH_WIDTH);
    rect.height = AmGetDefault( "graph.height", AmGRAPH_HEIGHT);
    AmInitIndexData( &heightIndex);
}

/****************************************************************************
 *
 * Description:
 *      check values for a graph.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag. The graph class check values function
 *	checks to make sure the heightIndex is initialized properly.
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmGraphClass::checkValues()
{
    if ( !AmDataGraphicClass::checkValues()) {
    	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      set the fields in the graph from the tag-value argument list.  Each
 *      class calls setFields() for the parent class, then uses
 *      SetFieldByDefinitions() to set its own members.  The setFields 
 *	function then updates the heightIndex structure (which is the 
 *	function for the y coordinate), then sets the coordindate system
 *	y range.
 * Arguments:
 *      tagValues: tag-value pairs to be parsed (this array is reduced in
 *                 place, with the parsed arguments removed).
 *      numValues: length of the tagValues array.
 * Return:
 *      number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int 
AmGraphClass::setFields( AmTagValuePtr       tagValues,
		       int               numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

/* Data graphic is set after the fields for graph class, because the extents
   function depends on the fields being set from AmGraphClass, but it's called
   from AmDataGraphicClass(). */

    AmGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs, 
				       tagValues, numValues);
    numValues = AmDataGraphicClass::setFields( tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	once the fields have been set using setFields(), update the 
 *	indexData structures and other dependent fields.
 *
 ****************************************************************************/

void 
AmGraphClass::updateFields()
{
    AmDataGraphicClass::updateFields();
    if ( IsValid()) {

/* set the index data client structures for the y function */

	if ( dataInfo.sourceType != AmNO_SOURCE) {
	    if ( updateIndexData( &heightIndex, "height") != AmOK) {
		AmAddResourceToLog( "failedToInitializeHeightIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( AmChanged( &heightIndex.min, &heightIndex.max, NULL)) {
	    coords.currYmin = coords.ymin = heightIndex.min;
	    coords.currYmax = coords.ymax = heightIndex.max;
	}
    }
}
 
/****************************************************************************
 *
 * Description:
 *      retreive values for a plot.  Each getFields() function calls
 *      getFields() for the parent class, then most of the fields are
 *      retrived using the GetFieldByDefinitions() function.  The
 *      bin vectors are retrieved by hand, because their length is
 *      1 less than the length value.
 * Arguments:
 *      tagValues - vector of tag-values to retrieve.  The values are
 *      pointers to the results.  The array is "scrunched" in place
 *      as each value is retreived.  Any unprocessed values are returned
 *      in place in the array.
 *      numValues - number of values in the array.
 * Returns:
 *      number of unprocessed values.
 *
 ****************************************************************************/

int 
AmGraphClass::getFields( AmTagValuePtr       tagValues,
		       int               numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmDataGraphicClass::getFields( tagValues, numValues);
    AmGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues =  GetFieldByDefinitions( fieldDefs, numFieldDefs,
				        tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Dump the field names used in a plot.  dumpFields() is called
 *      recursively by each object and its parent.  The fieldLengths
 *      array returns the name of each field, and if it's a vector
 *      field, the vector's length.
 * Arguments:
 *      fieldLengths - vector of tags and vector lengths.
 *      allFields - an obsolete tag which was used to differentiate
 *      AmBetween dumping to a file or copying the object.
 * Returns:
 *      The number of fields dumped
 *
 ***************************************************************************/

int 
AmGraphClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			AnvilBoolean	      	  allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( coordinate system) */

    numStrings = AmDataGraphicClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/****************************************************************************
 *
 * Description:
 *      Return the field definitions for an image plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmGraphClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.
 * 	THE LIST MUST BE SORTED BY ALPHABETICAL ORDER ON THE FIELD NAME.  
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/
 
void 
AmGraphClass::GetFieldDefinitions( AmFieldDefPtr   *fieldDefs,
				 int           *numFieldDefs)
{
    static AmFieldDefRec GraphClassFieldDefinitions[] = {
        { AmNheightClientData, offsetof( AmGraphClass, heightIndex.clientData) },
        { AmNheightClientDataFreeFn, 
		offsetof( AmGraphClass, heightIndex.clientDataFreeFn) },
        { AmNheightFn, offsetof( AmGraphClass, heightIndex.fn) },
        { AmNheightMax, offsetof( AmGraphClass, heightIndex.max) },
        { AmNheightMin, offsetof( AmGraphClass, heightIndex.min) },
	{ AmNheightName, offsetof( AmGraphClass, heightIndex.name), 0, AmDYNAMIC },
	{ AmNheightOffset, offsetof( AmGraphClass, heightIndex.offset) },
        { AmNheightType, offsetof( AmGraphClass, heightIndex.dataType) },
	{ AmNheightUnits, offsetof( AmGraphClass, heightIndex.units) }
    };
    *fieldDefs = GraphClassFieldDefinitions;
    *numFieldDefs = VECLEN( GraphClassFieldDefinitions);
}
