#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* functions for drawing hatchmarks inside polygons */

#include <malloc.h>
#include <string.h>
#if !defined(_WIN32)  && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdlib.h>
#include <math.h>
#include <anvil/geometry.h>
#include <anvil/macros.h>
#include <anvil/datafmt.h>
#include <anvil/hatchpoly.h>

/****************************************************************************
 *
 * Description:
 * 	draw the hatchmarks in a polygon with a specfied angle and spacing.
 *	This is a public function because it is used by polygon legends.
 * Arguments:
 *	display - display to draw to.
 *      win - window the the selection set will be drawn on.
 *      drawGC - graphics context used for drawing (passed to ::draw())
 *	hatchGC - adds the polygon to the clip region specified by the
 *	gc's clip rectangles. 
 *	poly - vector of screen coordinate space points representing   
 *	the polygon.
 *	rect - bounding rectangle of the polygon.
 *	numpts - number of points in the polygon.
 *	angle - hatch angle in radians.
 *	spacing - spacing AmBetween hatch marks in screen pixels.
 *
 ****************************************************************************/
 
AmErrorCode 
AmHatchPoly( AmDisplayClass		*display,
	     AmWindowClass		*win,
	     AmGCClass      		*drawGC,
	     AmGCClass			*hatchGC,
	     AmPointPtr     		poly,
	     int          		numpts,
	     AmRectangleRec const	*rect,
	     float        		angle,
	     int          		spacing)
{
    int         	vertFlag, x0, x1, nRects, ii;
    float       	slope = 0.0;
    AmRectanglePtr	clipRectVec;
    AmRegionClass	*polyClipRegion;
    AmColorCellRec	color;

    if ( numpts == 0) {
	return( AmOK);
    }

/* intersect the polygon with the clip rectangles passed to the 
   draw routine's GC */

    drawGC->GetClipRectangles( &clipRectVec, &nRects);
    if ( clipRectVec == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    polyClipRegion = display->CreateRegion();
    polyClipRegion->AddPolygon( poly, numpts, AmFILL_WINDING);
    for ( ii = 0; ii < nRects; ii++) {
	AmRegionClass	*rectRegion;

	rectRegion = display->CreateRegion();
	rectRegion->AddRectangle( &clipRectVec[ ii]);
	polyClipRegion->IntersectRegion( rectRegion);
	delete rectRegion;
    }
    free( clipRectVec);

/* set the clip region and the foreground and background colors */

    hatchGC->SetClipRegion( polyClipRegion);
    drawGC->GetForegroundColor( &color);
    hatchGC->SetForegroundColor( &color);

    vertFlag = ( angle == M_PI/2);
    if ( !vertFlag) {
        slope = tan( angle);
    }

/* loop forwards */

    for ( x0 = rect->x; ; x0 += spacing) {
	if ( vertFlag) {
	   x1 = x0;
	}
	x1 = AmSlopeaty( x0, rect->y, slope, rect->y + rect->height);
	if (( x1 > rect->x + rect->width) &&
	    ( x0 > rect->x + rect->width)) {
	    break;
	}
	display->DrawLine( win, hatchGC, 
			   x0, rect->y, x1, rect->y + rect->height);
    }

/* loop backwards */

    for ( x0 = rect->x - spacing; ; x0 -= spacing) {
	if ( vertFlag) {
	   x1 = x0;
	}
	x1 = AmSlopeaty( x0, rect->y, slope, rect->y + rect->height);
	if ( x1 < rect->x) {
	    break;
	}
	display->DrawLine( win, hatchGC, 
			   x0, rect->y, x1, rect->y + rect->height);
    }
    delete polyClipRegion;
    return( AmOK);
}
