#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* Pixel functions used for image plots to convert data to colors */

#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/imageplot.h>
#include <anvil/imageplotP.h>

/**************************************************************************
 *
 * Description:
 * 	24-bit true color pixel function.  This assumes that the RGB 
 *	values are stored in an unsigned long.  The low-level imaging
 *	routines do the appropriate translation to the device
 * Arguments:
 *	val - pixel value
 *	clientData - client data (unused).
 * Returns:
 *	returns val directly as an unsigned long.
 *
 **************************************************************************/

/* ARGSUSED */

static unsigned long
trueColorFn( void *val,
             void *clientData)
{   
    return( * ( ulong_t *) val); 
}

/**************************************************************************
 *
 * Description:
 * 	colormapped pixel function.  Maps the value to a vector of
 *	pixels using the indexData structure stored in the argData.
 *	The index function is either linear, non linear, or table 
 *	lookup.  The function also uses the minClip and maxClip flags,
 *	which cause it to return -1, which is interpreted by the
 *	image routines to not draw the pixel using a threshold.
 * Arguments:
 *	val - value to convert to a pixel.
 *	arg - arg data containing colormap client data, the indexData
 *	structure, and and user client data passed in as the 
 *	AmNcolorsClientData resource.
 * Returns:
 *	unsigned long pixel value, -1 indicates it shouldn't be drawn
 *
 **************************************************************************/

static unsigned long
colormapFn( void *val,
            void *arg)
{
    AmColormapArgPtr	argData = ( AmColormapArgPtr) arg;
    int			index;
    ulong_t		lReturn = 0,
                        *pl;

    pl = &lReturn;

    index = ( argData->indexData->getIndexFn) ( val, argData->indexData);
    if ( index <= 0) {
	if ( argData->indexData->minClip) {
	    return(( unsigned long) -1);
	} else {
	    ASSIGN_PIXEL( argData->colors[ 0], pl);;
	    return( lReturn);
	}
    } else if ( index >= argData->indexData->length - 1) {
    	if ( argData->indexData->maxClip) {
	    return(( unsigned long) -1);
	} else {
	    ASSIGN_PIXEL( argData->colors[ argData->indexData->length - 1], pl);
	    return( lReturn);
	}
    } else {
	ASSIGN_PIXEL( argData->colors[ index], pl);
	return( lReturn);
    }
}

/**************************************************************************
 *
 * Description:
 * 	light shaded pixel function.  This is a little more complicated
 *	than the standard pixel functions, because it has to look at the
 *	values to the right and below the pixel.  The shade value is
 *	calculcated as the normal of the two slopes.
 * Arguments:
 *	val - value to convert to a pixel.
 *	arg - arg data containing light shade client data, including the
 *	source matrix width and pixel offsets. (argData->hOffset and
 *	argData->vOffset).
 * Returns:
 *	unsigned long pixel value.
 *
 **************************************************************************/
 
static unsigned long
lightShadeFn( void *val,
              void *arg)
{
    AmLightShadeArgPtr    argData = (AmLightShadeArgPtr) arg;
    AmAccessFn		zFn = argData->indexData->fn;
    float               z0, z1, z2, normalX, normalY, AmDist, shade;
    int                 index;
    ulong_t		lReturn = 0;
    ulong_t		*pl;
 
/* get the pixel, the pixel to the right, and the down pixel */
 
    z0 = ( zFn)( val, argData->indexData->clientData);
    z1 = ( zFn)(( uchar_t *) val + argData->hOffset, 
		 argData->indexData->clientData);

/* make sure we don't sample beyond the end of the matrix */

    if (( uchar_t *) val + argData->vOffset < argData->matrixEnd) {
	z2 = ( zFn)(( uchar_t *) val + argData->vOffset, 
		     argData->indexData->clientData);
    } else {
	z2 = z1;
    }
    normalX = ( z1 - z0)*argData->binSize;
    normalY = ( z2 - z0)*argData->binSize;
    AmDist = sqrt( normalX*normalX + normalY*normalY + argData->normalZ2);
 
/* shade value goes from -1 to 1 */
 
    shade = ( normalX * argData->lightX + normalY * argData->lightY +
              argData->normalZ * argData->lightZ) / AmDist;
 
/* index into the colormap */
 
    index = MIN(( int) (argData->indexData->length*( 1.0 + shade)),
		argData->indexData->length*2 - 1);
    pl = &lReturn;
    ASSIGN_PIXEL( argData->colors[ index >> 1], pl);
    return( lReturn);
}

/**************************************************************************
 *
 * Description:
 * 	make the client data for the light shaded pixel function.  It
 *	precalcuates the light source angle and normals,  then uses the
 * 	matrix dimensions to calculate the horizontal and vertical
 *	pixel offsets.  Because of this, the imagePlot keepData resource
 *	MUST BE SET, otherwise the next row won't be available, and it
 *	will come up garbage.
 * Arguments:
 *	lightAngle - light horizontal angle in degrees.
 *	lightHeight - light vertical angle in degrees.
 *	colors - vector of colors.
 *	colorIndex - color index mapping structure.
 *	data - source matrix pointer (to calculate the bounds of the
 *	matrix.
 *	rows, cols -  dimensions of the source matrix.
 *	size - size of each matrix element.
 * Returns:
 *	clientData pointer to pass to the light shaded pixel function.
 *
 **************************************************************************/

AmLightShadeArgPtr AmMakeLightShadeArg( float		lightAngle,
				    float		lightHeight,
				    AmColorCellPtr	colors,
				    AmIndexDataPtr	colorIndex,
				    uchar_t		*data,
				    int			rows,
				    int			cols,
				    int			size)
{
    AmLightShadeArgPtr	lightShadeArg;
    float		lightX, lightY, lightZ, AmDist;

    lightShadeArg = ( AmLightShadeArgPtr) malloc( sizeof( AmLightShadeArgRec));
    lightX = cos( DEGTORAD( 90.0 - lightAngle));
    lightY = sin( DEGTORAD( 90.0 - lightAngle));
    lightZ = sin( DEGTORAD( lightHeight));
    AmDist = sqrt( lightX*lightX + lightY*lightY + lightZ*lightZ);
    lightX /= AmDist;
    lightY /= AmDist;
    lightZ /= AmDist;
 
/* h bin size and v bin size have to be in the same units as the z value */
 
    lightShadeArg->binSize = 5.0;
 
/* save the z normal and square to remove a multiply in the pixel function */
 
    lightShadeArg->normalZ = lightShadeArg->binSize*lightShadeArg->binSize;
    lightShadeArg->normalZ2 = lightShadeArg->normalZ*lightShadeArg->normalZ;
    lightShadeArg->lightX = lightX;
    lightShadeArg->lightY = lightY;
    lightShadeArg->lightZ = lightZ;
 
/* pointers to pixel down and pixel right */
 
    lightShadeArg->vOffset = cols*size;
    lightShadeArg->hOffset = size;
    lightShadeArg->matrixEnd = data + rows*cols*size;
    lightShadeArg->colors = colors;
    lightShadeArg->indexData = colorIndex;
    return( lightShadeArg);
}

/**************************************************************************
 *
 * Description:
 * 	Set the pixel function and client data for an image plot.
 * Arguments:
 *	data - source matrix pointer
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each matrix element.
 *
 **************************************************************************/

void AmImagePlotClass::setPixelFnClientData( uchar_t	*data,
					   int		rows,
					   int		cols,
					   int		size)
{
    AmColormapArgPtr	colormapArg;

/* Update the pixel function all the time.. create either the true color, 
   light shade or colormap pixel function depending on the arguments */

    if ( _pPixelClientData != NULL) {
	free( _pPixelClientData);
    }
    if ( _fTrueColor) {
	_fnPixel = trueColorFn;
	_pPixelClientData = NULL;
    } else if ( _fLightShaded) { 
	_fnPixel = lightShadeFn;
	_pPixelClientData = AmMakeLightShadeArg( _dLightAngle, _dLightHeight, 
					       colors, &colorIndex,
					       data, rows, cols, size);
    } else {
	colormapArg = ( AmColormapArgPtr) malloc( sizeof( AmColormapArgRec));
	colormapArg->indexData = &colorIndex;
	colormapArg->colors = colors;
	_fnPixel = colormapFn;
	_pPixelClientData = colormapArg;
    }
}
