#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <malloc.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <anvil/macros.h>
#include <anvil/files.h>
#include <anvil/log.h>
#include <anvil/mem_funcs.h>
#include <anvil/overlay.h>
#include <anvil/userevents.h>
#include <anvil/image.h>
#include <anvil/geometry.h>
#include <anvil/imageplot.h>
#include <anvil/imageplotP.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmIMAGE_PLOT) {
        return( TRUE);
    } else {
        return( AmPlotClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for an image plot.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use
 *      two-parameter constructors, because since we use varargs
 *      constructors, there is a naming ambiguity.
 * Arguments:
 *      pWindow - object is created on, and will be drawn on this window
 *      szName - string name used to identify plot for resources
 *
 ****************************************************************************/

void 
AmImagePlotClass::initializeFields( AmWindowClass	*pWindow,
				  const char	*szName)
{

/* call the parent initializeFields() member function */

    AmPlotClass::initializeFields( pWindow, szName);

    objectType           = AmIMAGE_PLOT;
    dataInfo.numdims     = 2; /* Images are 2-dimensional. */

    _pImage              = NULL;
    _dTransparency       = 1.0;
    _fKeepData           = TRUE;

    _fTrueColor          = FALSE;

    _fLightShaded        = FALSE;
    _dLightHeight        = 0.0;
    _dLightAngle         = 0.0;

    _fTileDetail         = TRUE;

    _fnPixel             = NULL;
    _pPixelClientData    = NULL;

    _fInterpolate        = FALSE;

/* set up the default scanline reading function */

    _fnScanline          = AmReadScanline;
    _pScanlineClientData = NULL;
    _scanlineDataFreeFn   = NULL;

/* set the sub rows and sub columns for read-from disk subsets. */

    _iSubRows            = 0;
    _iSubCols            = 0;

    _fFlipRows           = FALSE;
    _fFlipCols           = FALSE;

    _dSubXmin            = 0.0;
    _dSubXmax            = 0.0;
    _dSubYmin            = 0.0;
    _dSubYmax            = 0.0;

    _fScreenRes          = FALSE;
}

/****************************************************************************
 *
 * Description:
 *      check values for an image plot.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotClass::checkValues()
{
    AnvilBoolean fOK;

    fOK = TRUE;

/* First make sure that our parent class thinks everything is
   in good working order. */

    if ( !AmPlotClass::checkValues()) {
	fOK = FALSE;
    }

/* Check fields specific to the image plot.  First check the
   dataInfo specific fields. */

    if ( dataInfo.dims[ 0] <= 0) {
	AmAddResourceToLog( "rowsLessThanEqualZero", dataInfo.dims[ 0]);
	fOK = FALSE;
    }

    if ( dataInfo.dims[ 1] <= 0) {
	AmAddResourceToLog( "colsLessThanEqualZero", dataInfo.dims[ 1]);
	fOK = FALSE;
    }

    if ( dataInfo.size <= 0) {
	AmAddResourceToLog( "sizeLessThanEqualZero", dataInfo.size);
	fOK = FALSE;
    }

/* Check the fields pertaining to color. */

   if (( colors == NULL) && ( _fTrueColor == FALSE)) {
	AmAddResourceToLog( "nonTrueColorNoColormap");
	fOK = FALSE;
    } else if (( colorIndex.fn == NULL)  && 
	       ( colorIndex.AmIntFn == NULL) &&
	       ( colorIndex.dataType == AmNO_TYPE) &&
	       ( _fTrueColor == FALSE) &&
	       ( _fLightShaded == FALSE)) {
	AmAddResourceToLog( "colorIndexFnNULL");
	fOK = FALSE;
    }

/* make sure that the keepData flag is set, or the source type is file source
   or memory source if we're using light shading */

    if ( _fLightShaded && !( _fKeepData || 
			     ( dataInfo.sourceType == AmFILE_SOURCE) ||
			     ( dataInfo.sourceType == AmDIRECT_SOURCE))) {
	AmAddResourceToLog( "lightShadeMustHaveKeepData");
	fOK = FALSE;
    }

/* Check the relationship AmBetween the dataInfo fields and the color fields. */

    if (( dataInfo.size > 0)  && (colorIndex.offset >= dataInfo.size)) {
	AmAddResourceToLog( "lookupOffsetGreaterThanSize", 
			  colorIndex.offset, dataInfo.size);
	fOK = FALSE;
    }

    return( fOK);
}

/****************************************************************************
 *
 * Description:
 *	Default constructor, so Image plots can be subclassed. 
 *
 ****************************************************************************/

AmImagePlotClass::AmImagePlotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmImagePlotClass constructor.  Creates an image plot using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the image plot.
 * Arguments:
 *      windowArg - window that the image plot will be drawn on.
 *      nameArg - string name of the image plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmImagePlotClass::AmImagePlotClass( AmWindowClass  *pWindow, 
				const char   *szName,
				...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         n;

/* initialize all the basic fields for a plot */

    initializeFields( pWindow, szName);

/* set the default image client data and pixel function */

    va_start( args, szName);
    n = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    n = setFieldsWrapper( tagValues, n);
}

/****************************************************************************
 *
 * Description:
 *      AmImagePlotClass constructor.  Creates a image plot as part of an
 *      overlay.  The proprerties for the image plot are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the image plot will be added to.
 *      nameArg - string name of the image plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmImagePlotClass::AmImagePlotClass( AmOverlayClass *pOverlay,
				const char   *szName,
				...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         n;

/* Initialize the fields for a plot.  The coordinate system and rectangle
   will be set by AmOverlayClass::AddPlot(). */

    initializeFields( pOverlay->GetWindow(), szName);
    setParent( pOverlay);

/* Then override the fields for the overlay plot with the fields passed in
   for initialization. */

    va_start( args, szName);
    n = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    n = setFieldsWrapper( tagValues, n);
    if ( IsValid() && ( pOverlay != NULL)) {
	pOverlay->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Destructor: Free the images associated with the image plot, and
 *	the pixel function client data if there is any. 
 *
 ****************************************************************************/


AmImagePlotClass::~AmImagePlotClass()
{

/* Don't delete the display image if we're part of an overlay plot. */

    if ( parent == NULL) {
	if ( _pImage != NULL) {
	    delete _pImage;
	} 
    }

/* Free the pixel function client data if it isn't NULL. */

    if ( _pPixelClientData != NULL) {
	FREE( _pPixelClientData);
    }
    if (( _scanlineDataFreeFn != NULL) && ( _pScanlineClientData != NULL)) {
	( _scanlineDataFreeFn) ( _pScanlineClientData);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for an image plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmImagePlotClass::setFields( AmTagValuePtr	tagValues,
			   int		numValues)
{
    int		numFieldDefs; 
    AmFieldDefPtr	fieldDefs;

/* Initially, it's valid. */

    SetValid( TRUE);

/* Parse all the standard arguments. */

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmImagePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* If the source type is AmFILE_SOURCE, AmMMAP_SOURCE or AmDIRECT_SOURCE,
   _fKeepData is true. */

    if (( dataInfo.sourceType == AmFILE_SOURCE) ||
	( dataInfo.sourceType == AmMMAP_SOURCE) ||
	( dataInfo.sourceType == AmDIRECT_SOURCE)) {
	_fKeepData = TRUE;
    }

/* If the dataType is set, and the colorIndex dataType or function
   aren't set, then default the colorIndex to the same as the
   dataType. */

    if ( AmChanged( &dataInfo.dataType) && ( dataInfo.size == 0)) {
        dataInfo.size = AmDataTypeEnumToSize( dataInfo.dataType);
	colorIndex.dataType = dataInfo.dataType;
        AmMarkChanged( &dataInfo.size);
        AmMarkChanged( &colorIndex.dataType);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmImagePlotClass::updateFields()
{
    AmPlotClass::updateFields();
    if ( IsValid()) {
	if (( colors != NULL) && !_fTrueColor) {
	    if ( updateIndexData( &colorIndex, "colors") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for an image plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmImagePlotClass::getFields( AmTagValuePtr	tagValues,
			   int		numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs;

    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmImagePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	Update the image plot, based on the events which have been sent
 *	using AmSendUserEvent().  This stuffs the displayed image with the
 *	appropriate values.  This calls UpdateReadSource() if the source
 *	type reads from a flat file, UpdateTileSource() if reading from
 *	a pyramid, and UpdateMemorySource() if the data is in-RAM.
 * Arguments:
 *	eventCodes - list of event codes sent by AmSendUserEvent()
 *	numCodes - number of event codes.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::Update( int	*eventCodes,
			int	numCodes)
{
    AmRectangleRec const	*pRect = GetRect();

    if ( !IsValid() || !RectSet()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }

    if ( AmDegenRect( pRect)) {
	return( AmOK);
    }

/* If the image is out of sync, then resize it, and reset the rectangle
   to the screen extents. We have to support the pyramidal images, where
   the data image is resized according to the number of tiles needed to 
   cover the screen. */

    if ( parent == NULL) {
	if (( _pImage = ResizeDisplayImage( _pImage)) == NULL) {
	    DO_RETURN_ERROR( ANVIL_ERROR_CODE);
	}
    } else {

    /* The overlay may not have an image for us, so we bail. */

	_pImage = (( AmOverlayClass *) parent)->GetImage();
	if ( _pImage == NULL) {
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
    }

    if ( dataInfo.sourceType == AmREAD_SOURCE) {
	if (( ANVIL_ERROR_CODE = UpdateReadSource()) != AmOK) {
	    return( ANVIL_ERROR_CODE);
	}
    } else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	if (( ANVIL_ERROR_CODE = UpdateTileSource()) != AmOK) {
	    return( ANVIL_ERROR_CODE);
	}
    } else if( dataInfo.sourceType != AmNO_SOURCE) {
	if (( ANVIL_ERROR_CODE = UpdateMemorySource()) != AmOK) {
	    return( ANVIL_ERROR_CODE);
	}
    }

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Update the image plot when the data is re-read from a flat file.
 *	If the _fKeepData flag is TRUE, then a temporary matrix is allocated
 *	to hold the data, otherwise the matrix is read directly into the
 *	image and the values are thrown away.  Note that UpdateReadSource()
 *	can't use the lightshading pixel function if _fKeepData is not
 *	used, because the pixel function requires more than one line in
 *	the matrix.  Bilinear interpolation also only works when _fKeepData
 *	is true.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::UpdateReadSource()
{
    AmRealRectRec		drctSrc;
    AmRectangleRec const	*pScreenRect = GetRect();
    AmRectangleRec	rctMatrix, rctDst = *pScreenRect;

    float		x0, y0, x1, y1, dxScale, dyScale;
    int			subRectX, subRectY, subX0, subY0, subX1, subY1,
			oldSubRows = _iSubRows, oldSubCols = _iSubCols;

    if ( !IsValid()) {
        return( AmNOT_VALID);
    }

/* Set the pixel function to reflect the submatrix that we're working with. */

/* ERROR ERROR ERROR ERROR ERROR --- subRows and subCols aren't necessarily
   set at this point (the AmFirst time through at least --- so this needs to
   be called at a later time, too... */

    setPixelFnClientData( dataInfo.data, _iSubRows, _iSubCols, dataInfo.size);

/* Find out what part of the data image is used for the update. */

    if ( !ScreenRealExtents( dataInfo.dims[ 0], dataInfo.dims[ 1],
    			     _fFlipRows, _fFlipCols, &rctDst, &drctSrc)) {
	return( AmOK);
    }

    if ( _fKeepData == FALSE) {

	rctDst.x -= pScreenRect->x;
	rctDst.y -= pScreenRect->y;

	return( _pImage->imageReadRows( dataInfo.fd, dataInfo.size, 
                                        _fnPixel, _pPixelClientData,
                                        dataInfo.dims[ 0], dataInfo.dims[ 1],
                                        &drctSrc, &rctDst, _dTransparency,
                                        _fnScanline, _pScanlineClientData));
    } else {

    /* Project the screen coordinates of the update rectangle to the data
       coordinate system. */

	GetDataPt( rctDst.x, rctDst.y + rctDst.height, &x0, &y0);
	GetDataPt( rctDst.x + rctDst.width, rctDst.y, &x1, &y1);
	ClipRange( &x0, &y0);
	ClipRange( &x1, &y1);

	dxScale = coords.xmax - coords.xmin;
	dyScale = coords.ymax - coords.ymin;

	_dSubXmin = coords.currXmin;
	_dSubYmin = coords.currYmin;
	_dSubXmax = coords.currXmax;
	_dSubYmax = coords.currYmax;

	_fScreenRes = ((( int) drctSrc.width < rect.width) || 
		       (( int) drctSrc.height < rect.height));

    /* Recalculate the submatrix rows and columns. */

	if ( _fScreenRes) {

	    _iSubRows = (int)(dataInfo.dims[ 0] * ( _dSubYmax - _dSubYmin) / dyScale);
	    _iSubCols = (int)(dataInfo.dims[ 1] * ( _dSubXmax - _dSubXmin) / dxScale);

	} else {

	    GetScreenPt( _dSubXmin, _dSubYmax, &subX0, &subY0);
	    GetScreenPt( _dSubXmax, _dSubYmin, &subX1, &subY1);

	    _iSubRows = subY1 - subY0;
	    _iSubCols = subX1 - subX0;

	}

     /* If we're keeping a submatrix around from read data, then we resize
        the matrix if the screen extents have changed. */

	if (( dataInfo.data != NULL) && 
	    (( _iSubRows != oldSubRows) || ( _iSubCols != oldSubCols))) {
	    FREE( dataInfo.data);
	    dataInfo.data = NULL;
	}

	if ( dataInfo.data == NULL) {
	    dataInfo.data = ( uchar_t *) malloc( dataInfo.size * _iSubRows *
                                                 _iSubCols);
	    if ( dataInfo.data == NULL) {
                AmAddResourceToLog( "mallocFailed",
                                  dataInfo.size * _iSubRows * _iSubCols);
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	}

    /* Project from the source matrix to the interim submatrix. */

	if ( !_fScreenRes) {

    /* Since the interim submatrix is a projection of the source data,
       we project the border cells into matrix space and submatrix space
       separately.  Read from the source file into the submatrix. */

	    rctDst.x -= pScreenRect->x;
	    rctDst.y -= pScreenRect->y;

	    AmMatrixReadRows( dataInfo.fd, 
	    		      dataInfo.dims[ 0], dataInfo.dims[ 1],
			      &drctSrc, dataInfo.data,
			      _iSubRows, _iSubCols, &rctDst, dataInfo.size,
			      _fFlipRows, _fFlipCols,
			      _fnScanline, _pScanlineClientData);
	} else {

	    subRectX = (int)(_iSubCols * ( x0 - _dSubXmin) / ( _dSubXmax - _dSubXmin));
	    subRectY = (int)(_iSubRows * ( _dSubYmax - y1) / ( _dSubYmax - _dSubYmin));
	    rctMatrix.x = (int)(dataInfo.dims[ 1] * ( x0 - coords.xmin) / dxScale);
	    rctMatrix.y = (int)(dataInfo.dims[ 0] * ( coords.ymax - y1) / dyScale);
	    rctMatrix.width = (int)(dataInfo.dims[ 1] * ( x1 - x0) / dxScale);
	    rctMatrix.height = (int)(dataInfo.dims[ 0] * ( y1 - y0) / dyScale);

	    AmReadSubmatrix( dataInfo.fd, dataInfo.dims[ 0], dataInfo.dims[ 1], 
			     &rctMatrix, dataInfo.data,
                             _iSubRows, _iSubCols, subRectX, subRectY, 
			     dataInfo.size, _fnScanline, _pScanlineClientData);

	}

    /* Project from the submatrix to the current coordinates. */

	MatrixRealExtents( _dSubXmin, _dSubYmin, _dSubXmax, _dSubYmax,
                           x0, y0, x1, y1,
			   _iSubRows, _iSubCols, _fFlipRows, _fFlipCols, 
			   &rctDst, &drctSrc);

    /* Take the matrix and resize it to fit the screen. If the matrix is
       alreay been projected, then it's pretty much a 1:1 copy. */

	rctDst.x -= pScreenRect->x;
	rctDst.y -= pScreenRect->y;

	if ( _fInterpolate) {
	    return( _pImage->bilinMatrixImage( 
                                 dataInfo.data, dataInfo.size, 
				 _iSubRows, _iSubCols, _fFlipRows, _fFlipCols,
				 &drctSrc, &rctDst,
				 _fnPixel, _pPixelClientData, _dTransparency));
	} else {
	    return( _pImage->resizeMatrixImage( 
				 dataInfo.data, dataInfo.size, 
				 _iSubRows, _iSubCols, _fFlipRows, _fFlipCols,
				 &drctSrc, &rctDst,
				 _fnPixel, _pPixelClientData, _dTransparency));
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	Update the image plot when the data stored in memory.  Simply
 *	projects the submatrix into the image.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::UpdateMemorySource()
{
    AmRectangleRec	dstCopy = *GetRect();
    AmRectangleRec const	*screenRect = GetRect();
    AmRealRectRec		srcRect;

/* set the "sub range" and "sub dimensions"... */

    _dSubXmin = coords.xmin;
    _dSubYmin = coords.ymin;
    _dSubXmax = coords.xmax;
    _dSubYmax = coords.ymax;

    _iSubRows = dataInfo.dims[ 0];
    _iSubCols = dataInfo.dims[ 1];

    setPixelFnClientData( dataInfo.data, dataInfo.dims[ 0], 
			  dataInfo.dims[ 1], dataInfo.size);

/* find out what part of the data image is used for the update */

    if ( !ScreenRealExtents( dataInfo.dims[ 0], dataInfo.dims[ 1],
			     _fFlipRows, _fFlipCols, &dstCopy, &srcRect)) {
	return( AmOK);
    }
    dstCopy.x -= screenRect->x;
    dstCopy.y -= screenRect->y;

    if ( _fInterpolate) {
	return( _pImage->bilinMatrixImage( 
			     dataInfo.data, dataInfo.size,
			     dataInfo.dims[ 0], dataInfo.dims[ 1], 
			     _fFlipRows, _fFlipCols, &srcRect, &dstCopy, 
			     _fnPixel, _pPixelClientData, _dTransparency));
    } else {
	return( _pImage->resizeMatrixImage( 
			     dataInfo.data, dataInfo.size,
			     dataInfo.dims[ 0], dataInfo.dims[ 1], 
			     _fFlipRows, _fFlipCols, &srcRect, &dstCopy, 
			     _fnPixel, _pPixelClientData, _dTransparency));
    }
}

/****************************************************************************
 *
 * Description:
 * 	Update the image plot when the data is read from a pyramid format
 *	file.  There are separate functions for when the tiles are 
 *	read into an interim matrix or directly into the display image.
 * 	THE TILE STUFF IS BROKEN FOR FLIP ROWS and FLIP COLUMNS.  The
 *	pyramidal update doesn't support lightshading pixel functions 
 *	when _fKeepData is not set.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::UpdateTileSource()
{
    if ( _fKeepData) {
	return( UpdateTileSourceKeepData());
    } else {
	return( UpdateTileSourceNoKeepData());
    }
}

/****************************************************************************
 *
 * Description:
 *	Draw the image to the screen. If the image plot is standalone,
 *	then it draws the image, otherwise it defers it to the parent
 *	overlay.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::draw( AmWindowClass		*window,
	              AmGCClass			*gc, 
	              AmRectangleRec const	*exposeRects,
	              int			nRects)
{
    AmRectangleRec const	*dispRect = GetRect();
    int			ii;

    if ( !IsValid()) {
        return( AmOK);
    }
    if ( AmDegenRect( dispRect)) {
	return( AmOK);
    }

/* Since the image has a 0,0 origin, we shift the expose rectangle from the
   window coordinate system to the image coordinate system. 
   if we're standalone, draw the image, otherwise, it will get drawn by the
   overlay plot */

    if ( parent == NULL) {	
	if ( _pImage == NULL) {
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	for ( ii = 0; ii < nRects; ii++) {

	    display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
				exposeRects[ ii].y - dispRect->y,
				exposeRects[ ii].x, exposeRects[ ii].y,
				exposeRects[ ii].width,
                                exposeRects[ ii].height,
				_pImage);
	}
    }

    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	return the image plot's display image, or the parent's display
 *	image if it's part of an overlay.
 * Returns:
 *	The display image.
 *
 ****************************************************************************/

AmImageClass *
AmImagePlotClass::GetImage() 
{
    if ( parent != NULL) {
	return((( AmOverlayClass *) parent)->GetImage());
    } else {
	return( _pImage);
    }
}


/****************************************************************************
 *
 * Description:
 *	Set the image plot's display image.
 * Arguments:
 *	The display image.
 *
 ****************************************************************************/

void 
AmImagePlotClass::SetImage( AmImageClass *imageArg)
{
    if ( parent != NULL) {
	(( AmOverlayClass *) parent)->SetImage( imageArg);
    } else {
	_pImage = imageArg;
    }
}

/****************************************************************************
 *
 * Description:
 *	Given a x,y position in the image plot's real coordinate system,
 *	return the row and column in the interim matrix.
 * Arguments:
 *	xpos, ypos: position in the plot's coordinate systemA
 *	row, col: row and column position from the submatrix.
 * Returns:
 *	TRUE if row,col are in the bounds of the submatrix, FALSE
 *	if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotClass::GetRowcol( float	xpos,
			   float	ypos,
			   int		*row,
			   int		*col)
{
    return( AmGetRowcolumn( xpos, ypos,
                          _iSubRows, _iSubCols,
                          _fFlipRows, _fFlipCols,
			  _dSubXmin, _dSubYmin, _dSubXmax, _dSubYmax,
                          col, row));
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an image plot.  This is
 *      used in the map editor to dump the plot's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmImagePlotClass::dumpFields( AmFieldLengthPtr fieldLengths,
			    AnvilBoolean        allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;
 
/* Get the fields for the parent class ( plot). */
 
    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);
 
/* Get all the easy to specify fields for this class. */
 
    AmImagePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);

    return( numStrings);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an image plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmImagePlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmImagePlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				     int		*numFieldDefs)
{
    static AmFieldDefRec ImagePlotClassFieldDefinitions[] = {
        { AmNflipCols, offsetof( AmImagePlotClass, _fFlipCols) },
        { AmNflipRows, offsetof( AmImagePlotClass, _fFlipRows) },
	{ AmNinterpolate, offsetof( AmImagePlotClass, _fInterpolate) },
	{ AmNkeepData, offsetof( AmImagePlotClass, _fKeepData) },
	{ AmNlightShadeAngle, offsetof( AmImagePlotClass, _dLightAngle) },
	{ AmNlightShadeFlag, offsetof( AmImagePlotClass, _fLightShaded) },
	{ AmNlightShadeHeight, offsetof( AmImagePlotClass, _dLightHeight) },
	{ AmNscanLineClientData, 
              offsetof( AmImagePlotClass, _pScanlineClientData) },
	{ AmNscanLineClientDataFreeFn, 
              offsetof( AmImagePlotClass, _scanlineDataFreeFn) },
	{ AmNscanLineFn, offsetof( AmImagePlotClass, _fnScanline) },
	{ AmNsubCols, offsetof( AmImagePlotClass, _iSubCols) },
	{ AmNsubRows, offsetof( AmImagePlotClass, _iSubRows) },
	{ AmNsubXmax, offsetof( AmImagePlotClass, _dSubXmax) },
	{ AmNsubXmin, offsetof( AmImagePlotClass, _dSubXmin) },
	{ AmNsubYmax, offsetof( AmImagePlotClass, _dSubYmax) },
	{ AmNsubYmin, offsetof( AmImagePlotClass, _dSubYmin) },
	{ AmNtileDetailFlag, offsetof( AmImagePlotClass, _fTileDetail) },
	{ AmNtransparency, offsetof( AmImagePlotClass, _dTransparency) },
	{ AmNtrueColor, offsetof( AmImagePlotClass, _fTrueColor) },
    };
    
    *fieldDefs = ImagePlotClassFieldDefinitions;
    *numFieldDefs = VECLEN(ImagePlotClassFieldDefinitions);
}
