#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* these are the update functions for image plots using the hierarchical
   data format. */

#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stddef.h>
#include <stdarg.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/files.h>
#include <anvil/overlay.h>
#include <anvil/image.h>
#include <anvil/geometry.h>
#include <anvil/imageplot.h>
#include <anvil/neighbor.h>
#include <anvil/imageplotP.h>
#include <anvil/compression.h>

static const int MAXTILES = 1024;

/****************************************************************************
 *
 * Description:
 *	Find the appropriate tiles in the hierarchical file for the 
 *	current coordinates, then allocate a matrix to read them into, 
 *	and display the matrix in an image on the screen.
 * Returns:
 *	AmOK if the tiles were displayed OK or if there were no tiles to
 *	display, error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::UpdateTileSourceKeepData()
{
    AmRectangleRec	dispRect = *GetRect(), screenRect = *GetRect(), 
			tileSrcRect;
    AmRealRectRec		srcRect;
    int			ii, oldSubRows, oldSubCols, tileSize, dstX, dstY;
    uchar_t		*memptr;
    AmErrorCode		errCode;
    AmTileRec		tiles[ MAXTILES];
    AmTilePtr		tile;
    int			numTiles;

    oldSubRows = _iSubRows;
    oldSubCols = _iSubCols;

/* Get the tiles needed to stuff the display image. */

    numTiles = 0;
    errCode = AmFindImageTiles( coords.currXmin, coords.currYmin, 
			      coords.currXmax, coords.currYmax,
			      dispRect.height, dispRect.width, 
			      dataInfo.fd, &tileHeader->rootTile, tiles,
			      &numTiles, MAXTILES, detailLevel);

    if ( errCode != AmOK) {
	return( errCode);
    }

/* Bail if there are no tiles to chow down. */

    if ( numTiles <= 0) {
	return( AmOK);
    }

    AmTileVectorExtents( tiles, numTiles, &_iSubRows, &_iSubCols,
		       &_dSubXmin, &_dSubYmin, &_dSubXmax, &_dSubYmax);

/* Sort the tiles for performance. */

    qsort( tiles, numTiles, sizeof( AmTileRec), ( AmSearchFn) AmCompareTileSeekAddr);

/* Reset the pixel function client data. */

    setPixelFnClientData( dataInfo.data, _iSubRows, _iSubCols, dataInfo.size);

/* If we're keeping around a copy of the submatrix and the size has changed, 
   reallocate it. */

    if (( _iSubCols != oldSubCols) || ( _iSubRows != oldSubRows)) {
	if ( dataInfo.data != NULL) {
	    FREE( dataInfo.data);
	    dataInfo.data = NULL;
	}
    }
    if ( dataInfo.data == NULL) {

AN_DEBUG( 50, ("resizing data to %d %d\n", _iSubRows, _iSubCols));

	dataInfo.data = ( uchar_t *) malloc( dataInfo.size *
                                             _iSubRows * _iSubCols);
	if ( dataInfo.data == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    }

/* overallocate the tile by just a little bit to compensate for roundoff
   errors in the pyramid algorithm */

    tileSize = ( tiles[ 0].rows + 2)*( tiles[ 0].cols + 2)*dataInfo.size;
    if (( memptr = ( uchar_t *) alloca( tileSize)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }

/* loop over the tiles and copy them into the submatrix. */
   
    for ( ii = 0; ii < numTiles; ii++) {
	tile = &tiles[ ii];

/* set the source rectangle, which may change for each tile, because of 
   roundoff errors in the pyramid algorithm... */

	tileSrcRect.x = 0;
	tileSrcRect.y = 0;
	tileSrcRect.width = tile->cols;
	tileSrcRect.height = tile->rows;
	errCode = AmReadTile( tileHeader, tile, memptr);
	if ( errCode != AmOK) {
	    return( errCode);
	}

/* copy the tile into the submatrix */

	dstX = (int)rint( _iSubCols * ( tile->xmin - _dSubXmin) /
                     ( _dSubXmax - _dSubXmin));
	dstY = (int)rint( _iSubRows * ( _dSubYmax - tile->ymax) /
                     ( _dSubYmax - _dSubYmin));

AN_DEBUG( 40, ( "reading tile %ld at %d to %d %d\n", 
		tile->seekAddr, ii, dstX, dstY));

	AmCopySubmatrix( memptr, tile->rows, tile->cols, &tileSrcRect, 
		         dataInfo.data, _iSubRows, _iSubCols, 
			 dstX, dstY, dataInfo.size);
    }
  
/* project from the submatrix to the current coordinates */

    MatrixRealExtents( _dSubXmin, _dSubYmin, _dSubXmax, _dSubYmax,
		       coords.currXmin, coords.currYmin,
		       coords.currXmax, coords.currYmax,
		       _iSubRows, _iSubCols, _fFlipRows, _fFlipCols,
		       &dispRect, &srcRect);
 
/* take the matrix and resize it to fit the screen. If the matrix is alreay
   been projected, then it's pretty much a 1:1 copy */

AN_DEBUG( 10, ("dispRect = (%d, %d, %d, %d)\n",
	       dispRect.x, dispRect.y, dispRect.width, dispRect.height));
AN_DEBUG( 10, ("srcRect = (%.2f, %.2f, %.2f, %.2f)\n",
	       srcRect.x, srcRect.y, srcRect.width, srcRect.height));
    dispRect.x -= screenRect.x;
    dispRect.y -= screenRect.y;
    if ( _fInterpolate) { 
	return( _pImage->bilinMatrixImage(
                             dataInfo.data, dataInfo.size,
			     _iSubRows, _iSubCols, _fFlipRows, _fFlipCols,
			     &srcRect, &dispRect,
			     _fnPixel, _pPixelClientData, _dTransparency));
    } else {
	return( _pImage->resizeMatrixImage( 
			     dataInfo.data, dataInfo.size,
			     _iSubRows, _iSubCols, _fFlipRows, _fFlipCols,
			     &srcRect, &dispRect,
			     _fnPixel, _pPixelClientData, _dTransparency));
    }
}

/****************************************************************************
 *
 * Description:
 *	Find the appropriate tiles in the hierarchical file for the 
 *	current coordinates, then project each tile into the display image.
 *	This function does not retain the data from the tiles in a temporary
 *	matrix.
 * Returns:
 *	AmOK if the tiles were displayed OK or if there were no tiles to
 *	display, error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmImagePlotClass::UpdateTileSourceNoKeepData()
{
    AmRectangleRec	dispRect = *GetRect(), screenRect = *GetRect(),
			tileDispRect;
    AmRealRectRec		tileRect;
    int			ii, tileSize, oldSubRows, oldSubCols, numTiles;
    AmTilePtr		tile;
    AmTileRec		tiles[ MAXTILES];
    uchar_t		*memptr;
    AmErrorCode		errCode;

    oldSubRows = _iSubRows;
    oldSubCols = _iSubCols;

/* get the tiles needed to stuff the display image */

    numTiles = 0;
    errCode = AmFindImageTiles( coords.currXmin, coords.currYmin, 
			      coords.currXmax, coords.currYmax,
                              dispRect.height, dispRect.width,
                              dataInfo.fd, &tileHeader->rootTile, tiles,
                              &numTiles, MAXTILES, detailLevel);

/* bail if there are no tiles to chow down. */

    if ( numTiles <= 0) {
	return( AmOK);
    }
    if ( errCode != AmOK) {
        return( errCode);
    }
    AmTileVectorExtents( tiles, numTiles, &_iSubRows, &_iSubCols,
                       &_dSubXmin, &_dSubYmin, &_dSubXmax, &_dSubYmax);

/* sort the tiles for performance */

    qsort( tiles, numTiles, sizeof( AmTileRec), ( AmSearchFn) AmCompareTileSeekAddr);

/* reset the pixel function client data */

    setPixelFnClientData( dataInfo.data, _iSubRows, _iSubCols, dataInfo.size);

/* overallocate the tile by just a little bit to compensate for roundoff
   errors in the pyramid algorithm */

    tileSize = ( tiles[ 0].rows + 2)*( tiles[ 0].cols + 2)*dataInfo.size;
    if (( memptr = ( uchar_t *) alloca( tileSize)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    for ( ii = 0; ii < numTiles; ii++) {
	tile = &tiles[ ii];
AN_DEBUG( 40, ("loading tile %d, addr = %ld, size = %ld\n",
		ii, tile->seekAddr, tile->numBytes));
	errCode = AmReadTile( tileHeader, tile, memptr);
	if ( errCode != AmOK) {
	    return( errCode);
	}

/* if we're not keeping the data around in a submatrix, find out what the 
   tile represents on the screen, and draw it directly to the screen image */

	tileDispRect = dispRect;
	if ( MatrixRealExtents( tile->xmin, tile->ymin, 
				tile->xmax, tile->ymax,
				coords.currXmin, coords.currYmin, 
				coords.currXmax, coords.currYmax, 
				tile->rows, tile->cols,
				_fFlipRows, _fFlipCols,
				&tileDispRect, &tileRect)) {
	    tileDispRect.x -= dispRect.x;
	    tileDispRect.y -= dispRect.y;
	    if ( _fInterpolate) {
		_pImage->bilinMatrixImage( memptr, dataInfo.size,
					   tile->rows, tile->cols,
					   _fFlipRows, _fFlipCols, 
					   &tileRect, &tileDispRect, 
					   _fnPixel, _pPixelClientData, 
					   _dTransparency);
	    } else {
		_pImage->resizeMatrixImage( memptr, dataInfo.size,
					    tile->rows, tile->cols,
					    _fFlipRows, _fFlipCols, 
					    &tileRect, &tileDispRect, 
					    _fnPixel, _pPixelClientData, 
					    _dTransparency);
	    }
	}
    }
    return( AmOK);
}
