#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdio.h>

/* unfortunately, windows #define's the bessel functions */

#include <math.h>
#if defined(_WIN32)
#undef y0
#undef y1
#undef yn
#endif
#include <string.h>
#include <sys/types.h>
#include <anvil/fns.h>
#include <anvil/macros.h>

/*****************************************************************************
 * 
 * Description
 *	linear index from min to max to length. The second parameter is
 *	a void pointer which is cast to an indexData structure which
 *      contains the pertinent indexing information, including the 
 *	accessing function, its client data, and the scale and min and max.
 * Arguments:
 *	val - a pointer to the value to scale (type free)
 *	arg - pointer to the indexData structure initialzed by 
 *	updateIndexData.
 * Returns:
 *	index into the vector of graphics attributes (line width, colors, etc)
 *
 *****************************************************************************/

int AmLinearIndexFn( void  *val, 
		   void	 *arg)
{
    AmIndexDataPtr indexData = ( AmIndexDataPtr) arg;

/* since the scale is from min-max to numColors - 2, we add 1 to the index
   so the min value is only indexed if the value's out of range */

    return(( int) floor((( indexData->fn)
				((( uchar_t *) val) + indexData->offset,
				 indexData->clientData) - 
		        indexData->min)*indexData->scale) + 1);
}

/*****************************************************************************
 * 
 * Description
 *	non linear index over a float bin vector.  The bin vector is 1 bin 
 *	shorter than the graphics attribute vector.  indexData->length 
 *	indexes the color vector. The second parameter is
 *	a void pointer which is cast to an indexData structure which
 *      contains the pertinent indexing information, including the 
 *	accessing function, its client data, and the bin vector.
 * Arguments:
 *	val - a pointer to the value to scale (type free)
 *	arg - pointer to the indexData structure initialzed by 
 *	updateIndexData.
 * Returns:
 *	index into the vector of graphics attributes (line width, colors, etc)
 *
 *****************************************************************************/

int AmNonlinearIndexFn( void 	*val, 
		      void	*arg)
{
    AmIndexDataPtr indexData = ( AmIndexDataPtr) arg;
    float	 fval;
    int		 lo = 0, mid, hi = indexData->length - 2;

/* the colors vector is indexData->length, and the bin vector is
   indexData->length - 1 */

    fval = ( indexData->fn)((( uchar_t *) val) + indexData->offset,
			    indexData->clientData);
    if ( fval < indexData->binvec[ 0]) {
	return( -1);
    } else if ( fval > indexData->binvec[ hi]) {
	return( indexData->length);
    }
    while ( lo < hi) {
	mid = ( lo + hi + 1) >> 1;
	if ( fval < indexData->binvec[ mid]) {
	    hi = mid - 1;
	} else {
	    lo = mid;
	}
    }
    return( lo);
}

/*****************************************************************************
 * 
 * Description
 *	table lookup index over a sorted integer vector.  The second
 *	parameter is a void pointer which is cast to an indexData structure 
 *	which contains the pertinent indexing information, including the 
 *	accessing function, its client data, and the integer table.  There
 *	are separate indexing functions for each of the integer types
 *	(char, short, int). AmLookupVal is the basic binary search function.
 * Arguments: (AmLookupVal)
 *	ival - integer lookup val
 *	table - sorted integer table from indexData.
 *	length - length of the lookup table.
 *	(lookup___Fn):
 *	val - a pointer to the value to scale (type free)
 *	arg - pointer to the indexData structure initialzed by 
 *	updateIndexData.
 * Returns:
 *	index into the vector of graphics attributes (line width, colors, etc)
 *
 *****************************************************************************/

int AmLookupVal( int    ival,
	       int    *table,
	       int    length)
{
    int          lo = 0, hi = length - 1, mid;

    while ( lo <= hi) {
        mid = ( lo + hi) >> 1;
        if ( ival < table[ mid]) {
            hi = mid - 1;
        } else if ( ival > table[ mid]) {
            lo = mid + 1;
        } else {
            return( mid);
        }
    }
    return( lo);
}

/* table lookup function for character values */

int AmLookupCharFn( void	*val, 
		  void	*arg)
{
    AmIndexDataPtr indexData = ( AmIndexDataPtr) arg;
    int		 ival = *( char *) (( char *) val + indexData->offset);

    if (( ival < indexData->table[ 0]) && indexData->minClip) {
	return( -1);
    } else if (( ival > indexData->table[ indexData->length - 1]) && 
	       indexData->maxClip) {
	return( -1);
    }
    return( AmLookupVal( ival, indexData->table, indexData->length));
}

/* table lookup function for short integer values */

int AmLookupShortFn( void	*val, 
		   void	*arg)
{
    AmIndexDataPtr indexData = ( AmIndexDataPtr) arg;
    int		 ival = *( short *)(( char *) val + indexData->offset);

    if (( ival < indexData->table[ 0]) && indexData->minClip) {
	return( -1);
    } else if (( ival > indexData->table[ indexData->length - 1]) && 
	       indexData->maxClip) {
	return( -1);
    }
    return( AmLookupVal( ival, indexData->table, indexData->length));
}

/* table lookup function for integer values */

int AmTableLookupFn( void	*val, 
		   void	*arg)
{
    AmIndexDataPtr indexData = ( AmIndexDataPtr) arg;
    int		 ival = ( indexData->AmIntFn)(( char *) val + indexData->offset,
					    indexData->clientData);

    if (( ival < indexData->table[ 0]) && indexData->minClip) {
	return( -1);
    } else if (( ival > indexData->table[ indexData->length - 1]) && 
	       indexData->maxClip) {
	return( -1);
    }
    return( AmLookupVal( ival, indexData->table, indexData->length));
}

#if 0
/* 
    This lookup table and associated search function converts a dataType
    enumeration for an integer type to an integer table search function.
*/

typedef struct enumLookupFnDef_ {
    AmDataTypeEnum	dataType;
    AmIndexFn		fn;
} enumLookupFnDefRec, *enumLookupFnDefPtr;

static enumLookupFnDef_ lookupFns[] = {
    { AmINT, lookupIntFn },
    { AmUNSIGNED_INT, lookupIntFn },
    { AmCHAR, AmLookupCharFn },
    { AmUNSIGNED_CHAR, AmLookupCharFn },
    { AmSHORT, AmLookupShortFn },
    { AmUNSIGNED_SHORT, AmLookupShortFn },
    { AmLONG, lookupIntFn },
    { AmUNSIGNED_LONG, lookupIntFn }
};

/* translate a data type to the corresponding lookup function */

AmIndexFn dataTypeEnumToLookupFn( AmDataTypeEnum	dataType)
{
    unsigned int ii;
 
    for ( ii = 0; ii < VECLEN( lookupFns); ii++) {
        if ( lookupFns[ ii].dataType == dataType) {
            return( lookupFns[ ii].fn);
        }
    }
    return( NULL);
}
#endif
