#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* line graph class member functions.  Line graphs draw sets of points with
   connected lines with control for line color.  Since no new fields are
   (currently) defined over AmGraphClass, mostly the AmGraphClass functions are
   called within wrappers.  This is likely to change as the AmLineGraphClass
   becomes more sophisticated. */

#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/ratio.h>
#include <anvil/defaults.h>
#include <anvil/linegraph.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmLineGraphClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmLINE_GRAPH) {
        return( TRUE);
    } else {
        return( AmGraphClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      AmLineGraphClass constructor.  Creates a line graph using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the line graph.
 * Arguments:
 *      windowArg - window that the line graph will be drawn on.
 *      nameArg - string name of the line graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmLineGraphClass::AmLineGraphClass( AmWindowClass 	*windowArg, 
				const char	*nameArg,
				...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* initialize all the basic fields for a line graph, and set the object type */ 

    AmLineGraphClass::initializeFields( windowArg, nameArg);
    objectType = AmLINE_GRAPH;

/* parse the arguments and set the fields */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 *      AmLineGraphClass constructor.  Creates a line graph as part of an
 *      overlay.  The proprerties for the line graph are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the line graph will be added to.
 *      nameArg - string name of the line graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmLineGraphClass::AmLineGraphClass( AmOverlayClass *overlayArg,
				const char   *nameArg,
				...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* initialize all the basic fields for a graph, and set the object type */ 

    AmLineGraphClass::initializeFields( overlayArg->GetWindow(), nameArg);
    objectType = AmLINE_GRAPH;

/* parse the arguments and set the fields, then add the layer to the overlay
   if it's valid */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmLineGraphClass destructor.  
 *
 ****************************************************************************/

AmLineGraphClass::~AmLineGraphClass()
{
}

/****************************************************************************
 *
 * Description:
 *      do-nothing constructor so line plots can be subclassed.
 *
 ****************************************************************************/

AmLineGraphClass::AmLineGraphClass()
{
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a line graph.  Since line
 *	graph doesn't create any new members over graph, this function
 *	just calls the AmGraphClass constructor.
 * Arguments:
 *      windowArg - window that the line graph will be drawn on.
 *      nameArg - string name of the line graph, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmLineGraphClass::initializeFields( AmWindowClass	*windowArg,
			          const char	*nameArg)
{
    AmGraphClass::initializeFields( windowArg, nameArg);
}

/****************************************************************************
 *
 * Description:
 *      check values for a line graph.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmLineGraphClass::checkValues()
{
    if ( !AmGraphClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a line graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are set.
 *	Since AmLineGraphClass doesn't set any fields, it returns after this.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmLineGraphClass::setFields( AmTagValuePtr       tagValues,
			   int               numValues)
{
    numValues = AmGraphClass::setFields( tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmLineGraphClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmGraphClass::updateFields();
    if ( IsValid()) {

/* set the yFn to be the same as the heightIndex function */

        dataInfo.yFn = heightIndex.fn;

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a line graph, given the tag-values stored in
 *      a tag values array.  The generic graph fields are retrived, since
 *      line graph doesn't define any new members.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmLineGraphClass::getFields( AmTagValuePtr       tagValues,
			   int               numValues)
{
    numValues = AmGraphClass::getFields( tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Draw the lines in the line graph.  
 * Arguments:
 *      window - window the the line graph will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmLineGraphClass::draw( AmWindowClass		*window,
		      AmGCClass			*gc, 
		      AmRectangleRec const	*exposeRects,
		      int			nRects)
{
    int			ii, jj, kk, numleft;
    AmRectangleRec const	*dispRect = GetRect();
    AmPointRec		points[ AmMAX_SCREENPTS];
    uchar_t		*ptr;
    float		xscale, x;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmOK);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* clear the exposed rectangle.  We have to use fill rectangle because
   it pays attention to the gc clip rectangle set by the parent */

    if ( opaque && ( parent == NULL)) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width, 
				    exposeRects[ ii].height);
	}
    }

/* each column is represented by a different color if there are colors.  We
   cheat a little bit and assume there is at least 1 column */

    gc->SetForegroundColor( &fgColor);
    xscale = (coords.xmax - coords.xmin)/dataInfo.dims[ 0];
    if ( dataInfo.numdims == 2) {
	for ( ii = 0; ii < dataInfo.dims[ 1]; ii++) {
	    if ( numColors != 0) {
		gc->SetForegroundColor( &colors[ ii % numColors]);
	    }

/* draw the points in blocks for performance */

	    for ( jj = 0; jj < dataInfo.dims[ 0]; jj += AmMAX_SCREENPTS) {
		numleft = dataInfo.dims[ 0] - jj;
		for ( kk = 0, ptr = &dataInfo.data[ jj*dataInfo.size]; 
		      kk < MIN( AmMAX_SCREENPTS, numleft); 
		      kk++, ptr += dataInfo.size) {

/* scale the coordinates by the current coordinate system */

		    if ( dataInfo.xFn == NULL) {
			points[ kk].x = 
			    AmFloatScaleToInt( coords.currXmin, 
					     coords.currXmax,
					     coords.xmin + xscale*( jj + kk),
					     dispRect->x, 
					     dispRect->x + dispRect->width);
		    } else {
			x = ( dataInfo.xFn) ( ptr + dataInfo.xOffset,
					      dataInfo.clientData);
			points[ kk].x = 
			    AmFloatScaleToInt( coords.currXmin, 
					     coords.currXmax, x,
					     dispRect->x, 
					     dispRect->x + dispRect->width);
		    }
		    points[ kk].y = 
			AmFloatScaleToInt( coords.currYmin,
					 coords.currYmax,
		    			 ( heightIndex.fn)
					    ( ptr + heightIndex.offset,
                                              heightIndex.clientData),
					 dispRect->y + dispRect->height,
					 dispRect->y);
		}
		display->DrawLines( window, gc, points, numleft);
	    }
	} 
    } else {

/* draw the points in blocks for efficiency */

	for ( jj = 0; jj < dataInfo.dims[ 0]; jj += AmMAX_SCREENPTS) {
	    numleft = dataInfo.dims[ 0] - jj;
	    for ( kk = 0, ptr = &dataInfo.data[ jj*dataInfo.size]; 
		  kk < MIN( AmMAX_SCREENPTS, numleft); 
		  kk++, ptr += dataInfo.size) {

/* scale the coordinates by the current coordinate system */

		if ( dataInfo.xFn == NULL) {
		    points[ kk].x = 
			AmFloatScaleToInt( coords.currXmin, 
					 coords.currXmax,
					 coords.xmin + xscale*( jj + kk),
					 dispRect->x, 
					 dispRect->x + dispRect->width);
		} else {
		    x = ( dataInfo.xFn) ( ptr + dataInfo.xOffset,
					  dataInfo.clientData);
		    points[ kk].x = 
			AmFloatScaleToInt( coords.currXmin, 
					 coords.currXmax, x,
					 dispRect->x, 
					 dispRect->x + dispRect->width);
		}
		points[ kk].y = 
		    AmFloatScaleToInt( coords.currYmin,
				     coords.currYmax,
				     ( heightIndex.fn)
				       ( ptr + heightIndex.offset, 
                                         heightIndex.clientData),
				     dispRect->y + dispRect->height,
				     dispRect->y);
	    }
	    display->DrawLines( window, gc, points, numleft);
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw the current selection set.  The selection set is a buffer
 *      containing copies of the selected lines.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 *	fHighlight - controls whether lines are drawn in the highlight
 *	color or their normal color 
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the line graphs's invalid.
 *
 ****************************************************************************/

AmErrorCode 
AmLineGraphClass::DrawSelection( AmWindowClass		*window,
			       AmGCClass			*gc, 
			       AmRectangleRec const	*exposeRects,
			       int			nRects,
			       AnvilBoolean			fHighlight)
{
    int			ii, jj=0, numleft, nSel;
    AmRectangleRec const	*dispRect = GetRect();
    AmPointRec		points[ AmMAX_SCREENPTS];
    uchar_t		*ptr, *selectionData;
    float		xscale;
    AmColorCellRec	highlightColor;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmNOT_VALID);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }
    GetSelection(( void **) &selectionData, &nSel);
    if ( nSel <= 0) {
	return( AmOK);
    }
    if ( fHighlight) {
	gc->SetForegroundColor( &highlightColor);
    } else {
	gc->SetForegroundColor( &fgColor);
    }
    xscale = ( coords.xmax - coords.xmin)/dataInfo.dims[ 0];
    for ( ii = 0; ii < nSel; ii += AmMAX_SCREENPTS) {
	numleft = nSel - jj;
	for ( jj = 0, ptr = &selectionData[ ii*dataInfo.size];
	      jj < MIN( AmMAX_SCREENPTS, numleft); 
	      jj++, ptr += dataInfo.size) {

/* scale the coordinates by the current coordinate system */

	    points[ jj].x = 
		AmFloatScaleToInt( coords.currXmin, coords.currXmax,
				 coords.xmin + xscale*( ii + jj), dispRect->x, 
				 dispRect->x + dispRect->width);
	    points[ jj].y = 
		AmFloatScaleToInt( coords.currYmin, coords.currYmax,
				 ( heightIndex.fn)
				    ( ptr + heightIndex.offset,
				      heightIndex.clientData),
				 dispRect->y + dispRect->height, dispRect->y);
	}
	display->DrawLines( window, gc, points, numleft);
    }
    return( AmOK);
}
