#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#if ( defined(_WIN32) || defined(__BSD44__)) 
#include <malloc.h>
#else 
#include <alloca.h>
#endif
#include <stddef.h>
#include <stdarg.h>
#include <anvil/macros.h>
#include <anvil/resource.h>
#include <anvil/mem_funcs.h>
#include <anvil/geometry.h>
#include <anvil/overlay.h>
#include <anvil/lineplot.h>
#include <anvil/datafmt.h>
#include <anvil/compression.h>
#include <anvil/log.h>

static const int MAXTILES = 1024;

/****************************************************************************
 *
 * Description:
 * 	range function for line plots. This overrides the virtual function
 *	for AmDataGraphicClass, which only tests x,y coordinates.
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned data extents.
 *
 ****************************************************************************/

void 
AmLinePlotClass::GetDataExtents( float        *xmin,
			         float        *ymin,
			         float        *xmax,
			         float        *ymax)
{
    uchar_t	*ptr;
    AmLinePtr	line;
    int		ii;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;

/* Compare both endpoints against the x,y min,max for the ranging */

    for ( ii = 0, ptr = dataInfo.data; 
   	  ii < dataInfo.dims[ 0]; 
	  ii++, ptr += dataInfo.size) {
	line = ( AmLinePtr) ptr;
        if ( line->x0 < *xmin) {
            *xmin = line->x0;
        }
        if ( line->x1 < *xmin) {
            *xmin = line->x1;
        }
        if ( line->y0 < *ymin) {
            *ymin = line->y0;
        }
        if ( line->y1 < *ymin) {
            *ymin = line->y1;
        }
        if ( line->x0 > *xmax) {
            *xmax = line->x0;
        }
        if ( line->x1 > *xmax) {
            *xmax = line->x1;
        }
        if ( line->y0 > *ymax) {
            *ymax = line->y0;
        }
        if ( line->y1 > *ymax) {
            *ymax = line->y1;
        }
    }
}

/****************************************************************************
 *
 * Description:
 *	Recursive type function which tests the object type against the
 *	passed type and recursively tests the parent class's type.
 * Arguments:
 *	type: object type for testing.
 * Returns:
 *	TRUE if object class or any of its ancestors is of requested type, 
 *	FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmLinePlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmLINE_PLOT) {
       return( TRUE); 
    } else {
       return( AmPlotClass::IsType( type)); 
    }
}

/****************************************************************************
 *
 * Description:
 * 	initialize fields which are specific to a line plot.  Call the 
 *	basic plot class initializations, then initialize the line width
 *	tables, the line buffer size and the object.  We don't use
 *	constructors for this because of ambiguities with varargs
 * Arguments:
 *	windowArg - window that the line plot will be drawn on.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 *
 ****************************************************************************/

void
AmLinePlotClass::initializeFields( AmWindowClass	*windowArg,
				   const char		*nameArg)
{
    AmPlotClass::initializeFields( windowArg, nameArg);
    _dLineWidth = 0.0;
    _pdLineWidth = NULL;
    _pdLineWidthScale = NULL;
    AmInitIndexData( &_indexLineWidth);
    _indexLineWidth.offset = sizeof( AmLineRec);

    lineBufSize = LINEBUFSIZE;
    objectType = AmLINE_PLOT;

    /*
     * Color function has to be set explicitly. Default doesn't work 
     *   the range function changes to take the vector endpoints for
     *   finding the extents of the plot.
     */

    colorIndex.offset = sizeof( AmLineRec);
    stringIndex.offset = sizeof( AmLineRec);
}

/****************************************************************************
 *
 * Description:
 *      check values for a line plot.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag. 
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmLinePlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}


/****************************************************************************
 *
 * Description:
 *	do-nothing constructor so line plots can be subclassed. 
 *
 ****************************************************************************/

AmLinePlotClass::AmLinePlotClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	AmLinePlotClass constructor.  Creates a line plot using the window
 *	that it will be drawn on and the name, followed by NULL-terminated
 *	tag-value pairs for setting the fields in the line plot.
 * Arguments:
 *	windowArg - window that the line plot will be drawn on.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 * 	argList - NULL-teriminated tag value pairs, with tags defined in 
 *		  <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmLinePlotClass::AmLinePlotClass( AmWindowClass  *windowArg,
				  const char     *nameArg, 
				  ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 * 	AmLinePlotClass constructor.  Creates a line plot as part of an 
 *	overlay.  The proprerties for the line plot are set by a 
 *	NULL-terminated vector of tag-value pairs.
 * Arguments:
 *	overlayArg - overlay that the line plot will be added to.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 * 	argList - NULL-teriminated tag value pairs, with tags defined in 
 *		  <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmLinePlotClass::AmLinePlotClass( AmOverlayClass  *overlayArg,
				  const char      *nameArg,
				  ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* we want all of the parameters which are set for the overlay plot to be
   set for the line plot as well. Set the overlay plot back pointer to 
   reference in the draw and destroy routines 
   color function has to be set explicitly. Default doesn't work */

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 * 	AmLinePlotClass destructor.  Free the line width vectors if they've
 *	been allocated.
 *
 ****************************************************************************/

AmLinePlotClass::~AmLinePlotClass()
{
    AmDestroyIndexData( &_indexLineWidth);
    if ( _pdLineWidth != NULL) {
	free( _pdLineWidth);
    }
    if ( _pdLineWidthScale != NULL) {
	free( _pdLineWidthScale);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Set the fields for a line plot, given the tag-values stored in
 *	a tag values array.  First, the generic plot fields are set, then
 *	the other fields are set using the standard SetFieldByDefintions()
 *	function which uses a lookup table from the tag strings to the
 *	structure offsets.  The line width bin vector is set by hand because
 *	the length is one less than the line width vector.
 * Arguments:
 *	tagValues - vector of tag-value pairs (with tagCopied and copied
 *	flags).  The unparsed arguments are returned "scrunched" in this
 *	array.
 *	numValues - length of the vector.
 * Returns:
 *	number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmLinePlotClass::setFields( AmTagValuePtr	tagValues,
			    int			numValues)
{
    int		numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr	fieldDefs;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNlineWidthBinvec)) {
	    AmSetBinvec( &_indexLineWidth.binvec, _indexLineWidth.length,
		       &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmLinePlotClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmPlotClass::updateFields();

/* update the index data for line width and color functions */

    if ( IsValid()) {
	if (( _pdLineWidth != NULL) || ( _pdLineWidthScale != NULL)) {
	    if ( updateIndexData( &_indexLineWidth, "lineWidth") != AmOK) {
		AmAddResourceToLog( "failedToInitializeLineWidthIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colorIndex") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }
    }
}

/****************************************************************************
 *
 * Description:
 * 	Retreive the fields for a line plot, given the tag-values stored in
 *	a tag values array.  First, the generic plot fields are retrived, then
 *	the other fields are retrieved using GetFieldByDefintions()
 *	function which uses a lookup table from the tag strings to the
 *	structure offsets.  The line width bin vector is retrieved by hand 
 *	because the length is one less than the line width vector.
 * Arguments:
 *	tagValues - vector of tag-value pairs (with tagCopied and copied
 *	flags).  The unparsed arguments are returned "scrunched" in this
 *	array.  The values are pointers to the returned values.
 *	numValues - length of the vector.
 * Returns:
 *	number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmLinePlotClass::getFields( AmTagValuePtr	tagValues,
			    int			numValues)
{
    int		numFieldDefs, numUnusedFields, ii;
    AmFieldDefPtr	fieldDefs;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNlineWidthBinvec)) {
	    if ( _indexLineWidth.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( _indexLineWidth.binvec,
			     ( _indexLineWidth.length - 1) * sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}

    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *	Draw the current selection set.  The selection set is a buffer
 *	containing copies of the selected lines.
 * Arguments:
 *	window - window that the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 *	exposeRects - vector of exposed rectangles.
 *	nRects - length of the expose rectangle vector.
 *	fHighlight - TRUE: draw in the highlight color, FALSE, draw
 *	in the normal color.
 * Returns:
 *	AmOK if everything's OK, AmNOT_VALID if the line plot's invalid.
 *
 ****************************************************************************/

AmErrorCode
AmLinePlotClass::DrawSelection( AmWindowClass        *window,
                                AmGCClass            *gc, 
                                AmRectangleRec const *exposeRects,
                                int                  nRects,
			        AnvilBoolean		     fHighlight)
{
    float		xmin, ymin, xmax, ymax; 
    uchar_t		*selection;
    int			nSel, nseg;
    AmSegmentRec	segments[ MAXLINES];
    AmRectangleRec	surroundRect;

    /* Sanity checking. */

    if ( !IsValid()) {
	DO_RETURN_ERROR( AmNOT_VALID);
    }

    /* Return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
	return( AmOK);
    }

    /* Get the selection. */

    GetSelection(( void **) &selection, &nSel);
 
/* Take the expose rectangle and translate it to data coordinates to clip
   lines. */

    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
    GetDataPt( surroundRect.x + surroundRect.width,
	       surroundRect.y + surroundRect.height, &xmax, &ymin);

/* if highlighted, draw in the highlight color */

    nseg = 0;
    drawLineVector( window, gc, selection, nSel, segments, &nseg, fHighlight,
		    xmin, ymin, xmax, ymax);
    display->DrawSegments( window, gc, segments, nseg);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Draw the lines in the line plot.  Clear out the background if the
 *	plot's opaque, and draw the lines from the buffer, read them block
 *	by block, or read them pyramidided into tiles.   The lines are
 *	buffered until one of the graphics attributes (color, linewidth)
 *	is changed.
 * Arguments:
 *	window - window the the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 *	exposeRects - vector of exposed rectangles.
 *	nRects - length of the expose rectangle vector.
 * Returns:
 *	AmOK if everything's OK, system error if there was a memory
 *	allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode
AmLinePlotClass::draw( AmWindowClass        *window,
                       AmGCClass	    *gc, 
                       AmRectangleRec const *exposeRects,
                       int                  nRects)
{
    int			nseg, ii, numel;
    AmSegmentRec	segments[ MAXLINES];
    uchar_t		*lineBuffer;
    float		xmin, ymin, xmax, ymax;
    AmRectangleRec	surroundRect;

    /* If this object is invalid only draw the outline. */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

    /* Return in the degenerate case */

    if ( AmDegenRect( GetRect())) {
	return( AmOK);
    }

    /*
     * If there's no overlay data, then we clear the area that we're going
     *   to draw to.  We have to use FillRectangle(), because we have to
     *   pay attention to the gc's clip rectangle which is set by the parent.
     */

    if (( parent == NULL) && opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width, 
				    exposeRects[ ii].height);
	}
    }

    /*
     * Set the line width and line color, which may be overridden by functions
     *   in the loop.
     */

    _dLineWidthLast = _dLineWidth;
    gc->SetLineWidth( _dLineWidthLast);
    gc->SetLineCap( AmCAP_ROUND);

    _col = fgColor;
    _colLast = fgColor;
    gc->SetForegroundColor( &_colLast);
 
    /*
     * Take the expose rectangles, find the containing rectangle, and project
     *   it into the coordinate system for clipping the lines.
     */

    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
    GetDataPt( surroundRect.x + surroundRect.width,
               surroundRect.y + surroundRect.height, &xmax, &ymin);
    nseg = 0;
	
    /*
     * Cast each element to a data segment and draw it.  Buffer the line
     *   segments on width and color to try to speed things up.
     */

    if ( dataInfo.sourceType == AmREAD_SOURCE) {

        /*
         * Allocate a buffer to draw a couple of lines, and read the data
         *   in chunks to display it.
         */

        lineBuffer = ( uchar_t *) alloca( dataInfo.size * lineBufSize);
	if ( lineBuffer == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}
	if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	for ( ii = 0; ii < dataInfo.dims[ 0]; ii += lineBufSize) {
	    numel = MIN( lineBufSize, dataInfo.dims[ 0] - ii);
	    if ( read( dataInfo.fd, lineBuffer, dataInfo.size*numel) <
                       dataInfo.size * numel) {
		DO_RETURN_ERROR( AmREAD_FAILED);
	    }
	    drawLineVector( window, gc, lineBuffer, lineBufSize, segments, &nseg, 
			    FALSE, xmin, ymin, xmax, ymax);
	}
    } else if (( dataInfo.sourceType == AmDIRECT_SOURCE) ||
	       ( dataInfo.sourceType == AmFILE_SOURCE)) {
	drawLineVector( window, gc, dataInfo.data, dataInfo.dims[ 0],
			segments, &nseg, FALSE, xmin, ymin, xmax, ymax);
    } else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	AmTileRec	tiles[ MAXTILES];
	AmTilePtr	tile;
	AmErrorCode	errCode;
	int		numTiles, maxRows;

	numTiles = 0;
	errCode = AmFindLayerTiles( coords.currXmin, coords.currYmin,
				    coords.currXmax, coords.currYmax,
				    ( float) detailLevel,
				    dataInfo.fd, &tileHeader->rootTile, tiles, 
				    &numTiles, MAXTILES);
	if ( errCode != AmOK) {
	    return( errCode);
	}

        /* Sort the tiles for performance. */

	qsort( tiles, numTiles, sizeof( AmTileRec), 
	       ( AmSearchFn) AmCompareTileSeekAddr);

        /* Allocate the line buffer to draw the largest tile. */

	maxRows = 0;
	for ( ii = 0; ii < numTiles; ii++) {
	    if ( tiles[ ii].rows > maxRows) {
		maxRows = tiles[ ii].rows;
	    }
	}
	if (( lineBuffer = ( uchar_t *) 
		alloca( maxRows*dataInfo.size)) == NULL) {	
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}

/* search to each tile, uncompress it if the tile is compressed, or read
   it if it isn't, then draw the lines in the tile. */

	for ( ii = 0; ii < numTiles; ii++) {
	    tile = &tiles[ ii];
	    errCode = AmReadTile( tileHeader, tile, lineBuffer);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	    drawLineVector( window, gc, lineBuffer, tile->rows, 
			    segments, &nseg, FALSE, xmin, ymin, xmax, ymax);
	}
    }

    /* Since it's likely that the very AmLast call didn't empty the buffer,
       we do it here. */

    display->DrawSegments( window, gc, segments, nseg);

    /* If the object plot isn't inside of an overlay, then highlight the
       current selection. */

    if ( parent == NULL) {
	return( DrawSelection( window, gc, exposeRects, nRects, TRUE));
    } else {
	return( AmOK);
    }
}

/****************************************************************************
 *
 * Description:
 * 	draw a single line, using the segments buffer as a GC caching 
 *	mechanism.  The segments vector is a buffer which is flushed to
 *	the screen when it's filled, or when a graphics attribute has
 *	changed for the current line.
 * Arguments:
 *	window - window the the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 * 	line - line to draw.
 *	segments - vector of lines in the draw buffer.
 *	nseg - number of segments in the buffer.
 *
 ****************************************************************************/

void
AmLinePlotClass::drawLine( AmWindowClass *window,
			   AmGCClass     *gc, 
			   AmLinePtr     line,
			   AmSegmentPtr  segments,
			   int           *nseg,
			   AnvilBoolean	 fHighlight)
{
    int x0, y0, x1, y1, index;

    /*
     * For speed purposes, only set the line width when it changes, and
     *   draw the AmLast set of lines in the old line width.
     */

    if ( _pdLineWidth != NULL) {
	index = ( _indexLineWidth.getIndexFn)( line, &_indexLineWidth);
	_dLineWidth = _pdLineWidth[ index];
    }
    if ( _pdLineWidthScale != NULL) {
	index = ( _indexLineWidth.getIndexFn)( line, &_indexLineWidth);
	_dLineWidth = ScreenLineWidth( _pdLineWidthScale[ index]);
    }

    /* Only set the color when it changes */

    if (( colors != NULL) && ( fHighlight == FALSE)) {
	_col = colors[ ( colorIndex.getIndexFn)( line, &colorIndex)];
    } else {
	_col = highlightColor;
    }

    if (( _dLineWidth != _dLineWidthLast) ||
        ( AmCompareColor( &_col, &_colLast) == FALSE)) {

	display->DrawSegments( window, gc, segments, *nseg);
	*nseg = 0;

        _colLast = _col;
        _dLineWidthLast = _dLineWidth;

	gc->SetForegroundColor( &_colLast);
	gc->SetLineWidth( _dLineWidthLast);

    }

    /* Set the segment's endpoints and draw if the buffer is exceeded. */

    GetScreenPt( line->x0, line->y0, &x0, &y0);
    GetScreenPt( line->x1, line->y1, &x1, &y1);
    segments[ *nseg].x0 = x0;
    segments[ *nseg].y0 = y0;
    segments[ *nseg].x1 = x1;
    segments[ *nseg].y1 = y1;
    ( *nseg)++;
}

/* draw a vector of lines... */

void
AmLinePlotClass::drawLineVector( AmWindowClass	*window, 
				 AmGCClass	*gc, 
				 uchar_t	*lineBuffer, 
				 int		numLines,
				 AmSegmentPtr	segments, 
				 int		*nseg,
				 AnvilBoolean	fHighlight,
			         float		xmin, 
				 float		ymin, 
				 float		xmax, 
				 float		ymax)
{
    float	x0, y0, x1, y1;
    int		ii;
    AmLinePtr	line;
    uchar_t	*ptr;

    for ( ii = 0, ptr = lineBuffer; ii < numLines; ii++, ptr += dataInfo.size) {
	line = ( AmLinePtr) ptr;
	x0 = line->x0;
	y0 = line->y0;
	x1 = line->x1;
	y1 = line->y1;
	if ( AmClip2d( &x0, &y0, &x1, &y1, xmin, ymin, xmax, ymax)) {
	    drawLine( window, gc, line, segments, nseg, fHighlight);
	}
	if ( *nseg == MAXLINES) {
	    display->DrawSegments( window, gc, segments, *nseg);
	    *nseg = 0;
	}
    }
}

/*************************************************************************
 * 
 * Description
 * 	Return the extents of a pointer cast to a line. 
 * Arguments:
 *	ptr - void pointer to a line to get the extents of.
 *	xmin, ymin, xmax, ymax - returns extents of line.
 *
 *************************************************************************/

void
AmLinePlotClass::ObjectExtents( void  *ptr,
                                float *xmin,
                                float *ymin,
                                float *xmax,
                                float *ymax) 
{
    AmLinePtr	line = ( AmLinePtr) ptr;

    *xmin = MIN( line->x0, line->x1);
    *ymin = MIN( line->y0, line->y1);
    *xmax = MAX( line->x0, line->x1);
    *ymax = MAX( line->y0, line->y1);
}

/*************************************************************************
 * 
 * Description:
 * 	return the list of settable fields in a line plot.  This is
 *	used in the map editor to dump the plot's fields into a file.
 *	the field name (and length, if it's a vector field) is written
 *	into the vector.  Each dumpFields member function calls the
 *	dumpFields() member function of the parent class.  This is
 *	also used for copying plots for magic windows.
 * Arguments:
 *	fieldLengths: vector of field names and lengths.
 *	allFields: TRUE when dumping to a file, FALSE when copying
 *	in memory.  This is an obsolete parameter that may be removed
 *	in the future.
 * Returns:
 *	number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmLinePlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			     AnvilBoolean	         allFields)
{
    int           numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

    /* Get the fields for the parent class ( plot). */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

    /* Get all the easy to specify fields for this class. */

    AmLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);

    if ( _indexLineWidth.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNlineWidthBinvec;
	fieldLengths[ numStrings++].length = _indexLineWidth.length - 1;
    }

    return( numStrings);
}

/*************************************************************************
 * 
 * Description:
 * 	Return the field definitions for a line plot. This is used
 *	by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *	set the fields in the AmLinePlotClass.  It returns a vector
 *	of field names followed by offsets, offset of the length
 *	of the field if it's a vector, and a storageType, which 
 *	is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *	dynamically allocated fields.
 * Arugments:
 *	fieldDefs: returned vector of field definitions.
 *	numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmLinePlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				    int		*numDefs)
{
    static AmFieldDefRec LinePlotClassFieldDefintions[] = {
	{ AmNlineWidth, offsetof( AmLinePlotClass, _dLineWidth) },
	{ AmNlineWidthClientData, 
	    offsetof( AmLinePlotClass, _indexLineWidth.clientData) },
	{ AmNlineWidthClientDataFreeFn, 
	    offsetof( AmLinePlotClass, _indexLineWidth.clientDataFreeFn) },
	{ AmNlineWidthFn, offsetof( AmLinePlotClass, _indexLineWidth.fn) },
	{ AmNlineWidthIntFn, 
	    offsetof( AmLinePlotClass, _indexLineWidth.AmIntFn) },
	{ AmNlineWidthLength,
            offsetof( AmLinePlotClass, _indexLineWidth.length) },
	{ AmNlineWidthList, offsetof( AmLinePlotClass, _pdLineWidth),
            offsetof( AmLinePlotClass, _indexLineWidth.length),
            AmDYNAMIC },
	{ AmNlineWidthMax, offsetof( AmLinePlotClass, _indexLineWidth.max) },
	{ AmNlineWidthMin, offsetof( AmLinePlotClass, _indexLineWidth.min) },
	{ AmNlineWidthName,
            offsetof( AmLinePlotClass, _indexLineWidth.name), 
            0, AmDYNAMIC },
	{ AmNlineWidthOffset,
            offsetof( AmLinePlotClass, _indexLineWidth.offset) },
	{ AmNlineWidthStrings,
            offsetof( AmLinePlotClass, _indexLineWidth.strings),
            offsetof( AmLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
	{ AmNlineWidthTable,
            offsetof( AmLinePlotClass, _indexLineWidth.table),
            offsetof( AmLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
	{ AmNlineWidthType,
            offsetof( AmLinePlotClass, _indexLineWidth.dataType) },
	{ AmNlineWidthUnits,
            offsetof( AmLinePlotClass, _indexLineWidth.units) },
	{ AmNscaleLineWidthList, 
            offsetof( AmLinePlotClass, _pdLineWidthScale),
            offsetof( AmLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
    };
    *fieldDefs = LinePlotClassFieldDefintions;
    *numDefs = VECLEN( LinePlotClassFieldDefintions);
}
