#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#if !defined (_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdlib.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <anvil/tsearch.h>
#include <anvil/errcode.h>
#include <anvil/display.h>
#include <anvil/list.h>
#include <anvil/fns.h>
#include <anvil/image.h>		/* for AmScanlineFn */
#include <anvil/neighbor.h>

/* functions to support the neighboring algorithm to reduce the number of
   colors needed to represent matrices which represent thematic values */

/*  compare integers... */

static int 
compareInt( void	*a,
	    void	*b)
{
    return(( int) a - ( int) b);
}

/*  compare integer pointers... */

static int 
compareIntPtr( void	*a,
	       void	*b)
{
    return( *( int *) a - *( int *) b);
}

static int 
binarySearch( int    ival,
              int    *table,
              int    length)
{
    int          lo = 0, hi = length - 1, mid;

    while ( lo <= hi) {
        mid = ( lo + hi) >> 1;
        if ( ival < table[ mid]) {
            hi = mid - 1;
        } else if ( ival > table[ mid]) {
            lo = mid + 1;
        } else {
            return( mid);
        }
    }
    return( -1);
}

/****************************************************************************
 *
 * Description:
 * 	Given the neighbor list, allocate color indices to each node.
 *	for each node's neighbor list, assign a flags vector according
 *	to the index of each neighbor, then assign the color for that 
 *	node to the AmFirst flag in the vector that was not set.
 * Arguments:
 *	length - number of indexes to assign.
 *	nabors - vector of neighbor lists.
 *	indexes - pointer to returned vector of assigned indexes (allocated)
 *	maxnabor - maximum number of neighbors for a node.
 * Returns:
 *	AmOK if it's OK, AmMALLOC_FAILED, or AmALLOCA_FAILED if arrays
 *	couldn't be allocated.
 *
 ****************************************************************************/

static AmErrorCode 
setColorIndexes( int		length,
		 AmListPtr	*nabors, 	
		 int		**indexes,
		 int		maxnabor)
{
    int		ii, jj, index, dist, bestDist, bestDistIndex,
		stringLength, bestStringLength;
    AnvilBoolean	*flags;
    AmListPtr	naborptr;

    if (( *indexes = ( int *) malloc( sizeof( int)*length)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( flags = ( AnvilBoolean *) alloca( sizeof( AnvilBoolean)*maxnabor)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }

/* de-assign everyone's colors */

    for ( ii = 0; ii < length; ii++) {
	( *indexes)[ ii] = -1;
    }

/* for each neighbor list, find if a color index has been assigned for that 
   neighbor.  If so, set the flags array for that color's index */

    for ( ii = 0; ii < length; ii++) {
	memset( flags, 0, maxnabor*sizeof( AnvilBoolean));
AN_DEBUG( 70, ("index = %d: ", ii));
	for ( naborptr = nabors[ ii]; 
	      naborptr != NULL; 
	      naborptr = naborptr->next) {
	    index = ( int) naborptr->data;
AN_DEBUG( 70, ("%d ", index));
	    if (( *indexes)[ index] != -1)  {
		flags[ ( *indexes)[ index]] = TRUE;
	    }
	}
AN_DEBUG( 70, ("\n"));

/* use the first color index which isn't used by one of this node's neighbors */

	bestDist = 0;
	bestDistIndex = 0;
	dist = 0;
	stringLength = 0;
	bestStringLength = 0;
	for ( jj = 0; jj < maxnabor; jj++) {
	    if ( !flags[ jj]) {
		dist++;
		stringLength++;
		if ( dist > bestDist) {
		    bestDist = dist;
		    bestDistIndex = jj;
		    bestStringLength = stringLength;
		}
	    } else {
		stringLength = 0;
		dist = 0;
	    }
	}
	( *indexes)[ ii] = bestDistIndex - bestStringLength/2;
    }
    return( AmOK);
}

static int COUNT, *TABLE;

/****************************************************************************
 *
 * Description:
 * 	count the number of values in the neighbor tree. used by the
 *	neighbor table functions to generate a histogram of values.
 * Arguments:
 *	a - value node (ignored)
 *	order - twalk() visit order
 *	level - twalk() level
 *	uses the COUNT static variable.
 *
 ****************************************************************************/

/* ARGSUSED */

static void 
countVals( const void	*a,	
	   VISIT	order,
	   int		level)		
{
    if (( order == preorder) || ( order == leaf)) {
	COUNT++;
    }
}

/****************************************************************************
 *
 * Description:
 * 	set the values in the neighbor tree into a vector and
 *	maintain the vector count in a static. Used by the
 *	neighbor table functions to generate a histogram of values.
 * Arguments:
 *	a - value node (ignored)
 *	order - twalk() visit order
 *	level - twalk() level
 *	uses the COUNT and TABLE static variables.
 *
 ****************************************************************************/

/* ARGSUSED */

static void 
setVals( const void	*a,
	 VISIT		order,
	 int		level)	
{
    if (( order == preorder) || ( order == leaf)) {
	TABLE[ COUNT++] = **( int **) a;
    }
}

/****************************************************************************
 *
 * Description:
 * 	from a disk file, using a scanline function (AmREAD_SOURCE), create a 
 *  	sorted vector of all the different integer values. basically a 
 *	histogram without the value count.
 * Arguments:
 *	fd - file descriptor to read the matrix from.
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each source matrix element.
 *	skipBytes - # of bytes to skip before reading the matrix.
 *	AmScanFn, scanData - scanline read function and client data.
 *	valFn, clientData - converts source matrix values to integer values
 *	suitable for histogramming.
 *	offset - offset of the value in the source element (usually 0)
 *	table - returned pointer to a vector of integers representing
 *	generated histogram of values in the source matrix (allocated)
 *	length - returned pointer to the length of the returned vector.
 * Returns:
 *	AmOK on success, AmREAD_FAILED on read failure, AmMALLOC_FAILED
 *	on malloc failure.
 *
 ****************************************************************************/

AmErrorCode 
AmDiskBasedNeighborTable( int		fd,
			  int		rows,
			  int		cols,
			  int		size,
			  int		skipBytes,
			  AmScanlineFn	AmScanFn,
			  void		*scanData,
			  AmIntAccessFn	valFn,
			  void		*clientData,
			  int		offset,
			  int		**table,
			  int		*length)
{
    int		ii, jj, *val, **res;
    void	*root;
    uchar_t	*currLine;

    if (( val = ( int *) malloc( sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if ( lseek( fd, skipBytes, SEEK_SET) < 0) {
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if (( currLine = ( uchar_t *) malloc( cols*size)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* create a tree with all the different values in the file */

    root = NULL;
    for ( ii = 0; ii < rows; ii++) {
	if ( !( AmScanFn)( fd, currLine, size, cols, cols, 0, ii, scanData))  {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	for ( jj = 0; jj < cols; jj++) {
	    *val = ( valFn)( &currLine[ jj*size + offset], clientData);
	    if (( res = ( int **) tsearch(( void *) val, &root, 
					  ( AmSearchFn) compareIntPtr)) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }

/* if the value was inserted in the tree as a new value, then allocate a new
   pointer (local, because it's hard to pull it out of the tree */
	
	    if ( *res == val) {
		if (( val = ( int *) alloca( sizeof( int))) == NULL) {
		    DO_RETURN_ERROR( AmALLOCA_FAILED);
		}
	    }
	}
    }
    free( val);

/* find out how many values in the tree, create the table, and assign the 
   values from the tree to the table */

    COUNT = 0;
    twalk( root, ( AmWalkFn) countVals);
    *length = COUNT;
    if (( TABLE = *table = ( int *) malloc( *length*sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    COUNT = 0;
    twalk( root, ( AmWalkFn) setVals);

/* delete all the nodes to free the storage */

    for ( ii = 0; ii < *length; ii++) {
	tdelete( &TABLE[ ii], ( void **) &root, ( AmSearchFn) compareIntPtr);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	given a matrix, return a sorted vector representing the set of 
 *	values contained in the matrix.
 * Arguments:
 *	matrix - source matrix to get the values for.
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each source matrix element.
 *	valFn, clientData - converts source matrix values to integer values
 *	suitable for histogramming.
 *	offset - offset of the value in the source element (usually 0)
 *	table - returned pointer to a vector of integers representing
 *	generated histogram of values in the source matrix (allocated)
 *	length - returned pointer to the length of the returned vector.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure,
 *	AmALLOCA_FAILED on alloca() failure.
 *
 ****************************************************************************/

AmErrorCode 
AmMakeNeighborTable( uchar_t		*matrix,
		     int		rows,
		     int		cols,
		     int		size,
		     AmIntAccessFn	valFn,
		     void		*clientData,
		     int		offset,
		     int		**table,
		     int		*length)
{
    int		ii, *val, **res;
    void	*root;

/* create a tree with all the different values in the matrix */

    root = NULL;
    if (( val = ( int *) alloca( sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    for ( ii = 0; ii < rows*cols; ii++) {
	*val = ( valFn)( &matrix[ ii*size + offset], clientData);
	if (( res = ( int **) tsearch(( void *) val, &root, 
				      ( AmSearchFn) compareIntPtr)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}

/* if the value was inserted in the tree as a new value, then allocate a new
   pointer (on the stack, because it's hard to free from the tree. */
	
	if ( *res == val) {
	    if (( val = ( int *) alloca( sizeof( int))) == NULL) {
		DO_RETURN_ERROR( AmALLOCA_FAILED);
	    }
	}
    }

/* find out how many values in the tree, create the table, and assign the 
   values from the tree to the table */

    COUNT = 0;
    twalk( root, ( AmWalkFn) countVals);
    *length = COUNT;
    if (( TABLE = *table = ( int *) malloc( *length*sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    COUNT = 0;
    twalk( root, ( AmWalkFn) setVals);

/* delete all the nodes to free the storage */

    for ( ii = 0; ii < *length; ii++) {
	tdelete( &TABLE[ ii], ( void **) &root, ( AmSearchFn) compareIntPtr);
    }

/* sort the table so it can be searched with bsearch() */

    qsort( *table, *length, sizeof( int), ( AmSearchFn) compareIntPtr);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	from a disk file, using a scanline function (AmREAD_SOURCE), create a 
 *  	vector of lists of neighbors for each table entry in the table.  
 * Arguments:
 *	fd - file descriptor to read the data from
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each element in the source matrix.
 *	skipBytes - number of bytes to skip over at the start of the source
 *	matrix.
 *	AmScanFn - a function to read scanlines from the source file.
 *	scanData - clientData for the scanlines read function.
 *	intFn - integer lookup function,
 *	clientData - lookup function client data
 *	length - length of table.
 *	naborlst - returned pointer to a vector of lists for each 
 *	value in the table in the indexData structure.
 * Returns:
 *	AmOK if the neighbor list was generated properly, system error 
 *	if there were problems.
 *
 ****************************************************************************/

static AmErrorCode 
diskBasedNeighborList( int		fd,
		       int		rows,
		       int		cols,
		       int		size,
		       int		skipBytes,
		       AmScanlineFn	AmScanFn,
		       void		*scanData,
		       AmIntAccessFn	intFn,
		       void		*clientData,
		       int		*idTable,
		       int		numIds,
		       AmListPtr	**naborlst)
{
    uchar_t	*lastLine, *currLine, *nextLine, *tmp;
    int		candval, ii, jj, kk, ival, ivalIndex, candIndices[ 4];
    AmListPtr	*naborptr;

/* allocate the list of neighbor lists */

    if (( *naborlst = ( AmListPtr *) 
		calloc( numIds, sizeof( AmListPtr))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    naborptr = *naborlst;
    if (( lastLine = ( uchar_t *) malloc( cols*size)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( currLine = ( uchar_t *) malloc( cols*size)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( nextLine = ( uchar_t *) malloc( cols*size)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* seek to the head of the file and read the AmLast and current lines */

    if ( lseek( fd, skipBytes, SEEK_SET) < 0) {
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if ( !( AmScanFn)( fd, lastLine, size, cols, cols, 0, 0, scanData))  {
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    if ( !( AmScanFn)( fd, currLine, size, cols, cols, 0, 1, scanData))  {
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    for ( ii = 1; ii < rows - 1; ii++) {

/* read the next line */

	if ( !( AmScanFn)( fd, nextLine, size, cols, cols, 
			 0, ii + 1, scanData))  {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	for ( jj = 1; jj < cols - 1; jj++) {
	    ival = ( intFn)( &currLine[ jj*size], clientData);
	    ivalIndex = binarySearch( ival, idTable, numIds);
	    if ( ivalIndex < 0) {
		continue;
	    }

/* test the neighbors left, right, above and below */

	    candval = ( intFn) ( &currLine[ ( jj + 1)*size], clientData);
	    candIndices[ 0] = binarySearch( candval, idTable, numIds);
	    candval = ( intFn) ( &currLine[ ( jj - 1)*size], clientData);
	    candIndices[ 1] = binarySearch( candval, idTable, numIds);
	    candval = ( intFn) ( &lastLine[ jj*size], clientData);
	    candIndices[ 2] = binarySearch( candval, idTable, numIds);
	    candval = ( intFn) ( &nextLine[ jj*size], clientData);
	    candIndices[ 3] = binarySearch( candval, idTable, numIds);
	    for ( kk = 0; kk < 4; kk++) {
		if (( candIndices[ kk] != ivalIndex) && ( candIndices[ kk] >= 0)) {
		    naborptr[ ivalIndex] = 
			AmInsertUniqLst(( void *) candIndices[ kk],
				        naborptr[ ivalIndex], 
				        ( AmSearchFn) compareInt);
		    if ( naborptr[ ivalIndex] == NULL) {
			DO_RETURN_ERROR( AmMALLOC_FAILED);
		    }
		}
	    }
	}

/* rotate the three buffers */

	tmp = lastLine;
	lastLine = currLine;
	currLine = nextLine;
	nextLine = tmp;
    }

/* free the buffers and return */

    free( lastLine);
    free( currLine);
    free( nextLine);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	given a file, its dimensions, scanline function, scan client data, 
 *	and size, an index data structure, which contains a lookup function, 
 *	table and length, return a list of indexes into the resulting 
 *	colormap, and the number of colors that have to be allocated. 
 * Arguments:
 *	fd - file descriptor to read the matrix from.
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each source matrix element.
 *	skipBytes - # of bytes to skip before reading the matrix.
 *	AmScanFn, scanData - scanline read function and client data.
 *	valFn, clientData - converts source matrix values to integer values
 *	suitable for histogramming.
 *	offset - offset of the value in the source element (usually 0)
 *	table - returned pointer to a vector of integers representing
 *	generated histogram of values in the source matrix (allocated)
 *	length - returned pointer to the length of the returned vector.
 * Returns:
 *	AmOK on success, AmREAD_FAILED on read failure, AmMALLOC_FAILED
 *	on malloc failure.
 *
 ****************************************************************************/

AmErrorCode 
AmDiskBasedNeighborColorIndexes( int		fd,
			         int		rows,
			         int		cols,
			         int		size,
			         int		skipBytes,
			         AmScanlineFn	AmScanFn,
			         void		*scanData,
				 AmIntAccessFn	intFn,
				 void		*clientData,
				 int		*idTable,
				 int		numIds,
			         int		**indexes,
			         int		*maxnabor)
{
    AmErrorCode	errCode;
    AmListPtr	*naborlst;
    int		naborlen, ii;

/* create the neighbor list */

    errCode = diskBasedNeighborList( fd, rows, cols, size, skipBytes,
				     AmScanFn, scanData, intFn, clientData, 
				     idTable, numIds, &naborlst);
    if ( errCode != AmOK) {
	return( errCode);
    }

/* find the sorry-ass value with most mooching neighbors */

    *maxnabor = 0;
    for ( ii = 0; ii < numIds; ii++) {
	naborlen = AmLengthLst( naborlst[ ii]);
	if ( naborlen > *maxnabor) {
	    *maxnabor = naborlen;
	}
    }

/* create an array of indexes which reflect the neighbor map */

    errCode = setColorIndexes( numIds, naborlst, indexes, *maxnabor);

/* free the neighbor map and bail */

    for ( ii = 0; ii < numIds; ii++) {
	AmJustFreeLst( naborlst[ ii]);
    }
    free( naborlst);
    return( errCode);
}

/****************************************************************************
 *
 * Description:
 * 	from a matrix, create a vector of lists of neighbors for each table 
 *	entry in indexData.  
 * Arguments:
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each element in the source matrix.
 *	intFn - integer lookup function,
 *	clientData - lookup function client data
 *	numIds - length of table.
 *	naborlst - returned pointer to a vector of lists for each 
 *	value in the table in the indexData structure.
 * Returns:
 *	AmOK if the neighbor list was generated properly, system error 
 *	if there were problems.
 *
 ****************************************************************************/

static AmErrorCode 
makeNeighborList( uchar_t		*matrix,
		  int			rows,
		  int			cols,
		  int			size,
		  AmIntAccessFn		intFn,
		  void			*clientData,
		  int			*idTable,
		  int			numIds,
		  AmListPtr		**naborlst)
{
    AmListPtr	*naborptr;
    int		ii, jj, kk, candval, candIndex, offlist[ 4], index, ival,
		ivalIndex;

    offlist[ 0] = -1;
    offlist[ 1] = 1;
    offlist[ 2] = -cols;
    offlist[ 3] = cols;

/* allocate the list of neighbor lists */

    if (( *naborlst = ( AmListPtr *) 
		calloc( numIds, sizeof( AmListPtr))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    naborptr = *naborlst;

/* walk through the matrix and test the neighbors and add them to the lists */

    for ( ii = 1; ii < rows - 1; ii++) {
	for ( jj = 1; jj < cols - 1; jj++) {
	    index = ii*cols + jj;
	    ival = ( intFn)( &matrix[ index*size], clientData);
	    ivalIndex = binarySearch( ival, idTable, numIds);
	    if ( ivalIndex < 0) {
		continue;
	    }

/* test the neighbors above, below, left, and right */

	    for ( kk = 0; kk < 4; kk++) {
		candval = ( intFn) ( &matrix[( index + offlist[ kk])*size], 
				     clientData);
		candIndex = binarySearch( candval, idTable, numIds);
		if (( candIndex != ivalIndex) && ( candIndex >= 0)) {
		    naborptr[ ivalIndex] = 
			AmInsertUniqLst(( void *) candIndex,
				        naborptr[ ivalIndex], 
					( AmSearchFn) compareInt);
		    if ( naborptr[ ivalIndex] == NULL) {
			DO_RETURN_ERROR( AmMALLOC_FAILED);
		    }
		}
	    }
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	given a matrix, its dimensions, and size, a index data structure, 
 *	which contains a lookup function, table and numIds, return a list 
 *	of indexes into the resulting colormap, and the number of colors that 
 *	have to be allocated given a matrix, return a sorted vector 
 *	representing the set of values contained in the matrix.
 * Arguments:
 *	matrix - source matrix to get the values for.
 *	rows, cols - dimensions of the source matrix.
 *	size - size of each source matrix element.
 *	valFn, clientData - converts source matrix values to integer values
 *	suitable for histogramming.
 *	offset - offset of the value in the source element (usually 0)
 *	table - returned pointer to a vector of integers representing
 *	generated histogram of values in the source matrix (allocated)
 *	maxnabor - returned pointer to the length of the returned vector.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure,
 *	AmALLOCA_FAILED on alloca() failure.
 *
 ****************************************************************************/

AmErrorCode 
AmMakeNeighborColorIndexes( uchar_t		*matrix,
			    int			rows,
			    int			cols,
			    int			size,
			    AmIntAccessFn	intFn,
			    void		*clientData,
			    int			*idTable,
			    int			numIds,
			    int			**indexes,
			    int			*maxnabor)
{
    AmErrorCode	errCode;
    AmListPtr	*naborlst;
    int		naborlen, ii;

/* create the neighbor list */

    errCode = makeNeighborList( matrix, rows, cols, size, 
				intFn, clientData, idTable, numIds, &naborlst);
    if ( errCode != AmOK) {
	return( errCode);
    }

/* find the sorry-ass value with most mooching neighbors */

    *maxnabor = 0;
    for ( ii = 0; ii < numIds; ii++) {
	naborlen = AmLengthLst( naborlst[ ii]);
	if ( naborlen > *maxnabor) {
	    *maxnabor = naborlen;
	}
    }

/* create an array of indexes which reflect the neighbor map */

    errCode = setColorIndexes( numIds, naborlst, indexes, *maxnabor);

/* free the neighbor map and bail */

    for ( ii = 0; ii < numIds; ii++) {
	AmJustFreeLst( naborlst[ ii]);
    }
    free( naborlst);
    return( errCode);
}
