#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



/* member functions for object plots, which are similar to scatter plots,
   but with a user-defined object drawing function */

#include <malloc.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <anvil/macros.h>
#include <anvil/plot.h>
#include <anvil/extents.h>
#include <anvil/overlay.h>
#include <anvil/objectplot.h>
#include <anvil/datafmt.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmObjectPlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmOBJECT_PLOT) {
        return( TRUE);
    } else {
        return( AmPlotClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 * 	default object drawing function draws a circle.  The function
 *	takes the screen coordinates of the object being drawn, the
 *	pixel size of the display (in the plot's coordinate sytem),
 *	a pointer to the object, and a client data pointer.
 * Arguments:
 *      display - display that the object will be drawn to.
 *	win - window the object will be drawn on.
 *	gc - graphics context to use for drawing.
 *	x,y - screen position of the object.
 *	pixelSize - size of a screen pixel in the plot's coordinate system.
 *	object - pointer to the object to draw.
 *	clientData - passed client data.
 *
 ****************************************************************************/

void 
AmDefaultDrawObjectFn( AmDisplayClass	*display,
		     AmWindowClass	*win,
		     AmGCClass		*gc,
		     int		x,
		     int		y,
		     float		pixelSize,
		     void		*object,
		     void		*clientData)
{
    display->DrawArc( win, gc, x - 5, y - 5, 10, 10, 0.0, 6.28);
}


/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to an object plot.  Call the
 *      basic plot class initializations, then initialize the drawing 
 *	function and client data, and the xy functions.
 * Arguments:
 *      windowArg - window that the line plot will be drawn on.
 *      nameArg - string name of the line plot, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmObjectPlotClass::initializeFields( AmWindowClass	*windowArg,
				     const char		*nameArg)
{
    AmPlotClass::initializeFields( windowArg, nameArg);
    objectType = AmOBJECT_PLOT;
    drawFn = AmDefaultDrawObjectFn;
    drawFnClientData = NULL;
    drawFnClientDataFreeFn = NULL;
    setXFn = NULL;
    dataInfo.xOffset = 0;
    setYFn = NULL;
    dataInfo.yOffset = 0;
    setXYFn = NULL;
    maxWidth = 0;
    maxHeight = 0;

/* initialize the font information */

    font = strdup( DEFAULT_FONT);
    fontSize = 10;
    fontWeight = DEFAULT_WEIGHT;
    fontSlant = DEFAULT_SLANT;
    fontXScale = 1.0;
    fontYScale = 1.0;
    _dLineWidth = 0.0;
}

/****************************************************************************
 *
 * Description:
 *      check values for an object plot.  Each check values function 
 *	calls the check values function for the parent class and is used 
 *	to set the object's IsValid() flag. The object plot class check 
 *	values function checks to make sure the heightIndex is initialized 
 *	properly.
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmObjectPlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
        return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      do-nothing constructor so object plots can be subclassed.
 *
 ****************************************************************************/

AmObjectPlotClass::AmObjectPlotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmObjectPlotClass constructor.  Creates an object plot using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the line plot.
 * Arguments:
 *      windowArg - window that the object plot will be drawn on.
 *      nameArg - string name of the object plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmObjectPlotClass::AmObjectPlotClass( AmWindowClass  *windowArg, 
				  const char   *nameArg,
				  ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 *      AmObjectPlotClass constructor.  Creates an object plot as part of an
 *      overlay.  The proprerties for the object plot are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the object plot will be added to.
 *      nameArg - string name of the object plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmObjectPlotClass::AmObjectPlotClass( AmOverlayClass  *overlayArg,
				  const char	*nameArg,
				  ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmObjectPlotClass destructor.  
 *
 ****************************************************************************/

AmObjectPlotClass::~AmObjectPlotClass()
{
    if (( drawFnClientDataFreeFn != NULL) && ( drawFnClientData != NULL)) {
	( drawFnClientDataFreeFn)( drawFnClientData);
    }
    if ( font != NULL) {
	free( font);
    }
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for an object plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmObjectPlotClass::setFields( AmTagValuePtr	tagValues,
			    int		numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmObjectPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields. ObjectPlot
 *	updates the color index field.
 *
 ****************************************************************************/

void
AmObjectPlotClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmPlotClass::updateFields();
    if ( IsValid()) {
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colors") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for an object plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are retrived, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmObjectPlotClass::getFields( AmTagValuePtr	tagValues,
			    int		numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmObjectPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	return the extents of an object in screen coordinates.  This uses
 *	the extents display to figure out the extents of the object by
 *	"recording" the extents of the drawing functions used.  This is
 *	used when erasing and dragging objects to find the area that needs
 *	to be cleared behind them.
 * Arguments:
 *	object - pointer to the object to find the extents for.
 *	win - window the object is drawn to.
 *	gc - graphics context the object is drawn with.
 *	rect - returned object extents in screen coordinates.
 *
 ****************************************************************************/

void 
AmObjectPlotClass::GetObjectExtents( void 	*object,
				   AmWindowClass	*win,
				   AmGCClass	*gc,
				   AmRectanglePtr	rect)
{
    AmExtentsClass	extents( display);

    extents.StartExtents();
    DrawObject( &extents, win, gc, ( uchar_t *) object, GetRect());
    *rect = *extents.GetExtents();
    rect->x--;
    rect->y--;
    rect->width += 2;
    rect->height += 2;
}

/****************************************************************************
 *
 * Description:
 * 	walk through the vector of objects displayed by the object plot
 *	and find the maximum extents.
 * Arguments:
 *	win - window the objects are drawn to.
 *	gc - graphics context the objects are drawn with.
 *
 ****************************************************************************/

void 
AmObjectPlotClass::SetMaxObjectExtents( AmWindowClass	*win,
				      AmGCClass		*gc)
{
    int			ii;
    AmRectangleRec	extents;

    maxWidth = 0;
    maxHeight = 0;
    for ( ii = 0; ii < dataInfo.dims[ 0]; ii++) {
	GetObjectExtents( &dataInfo.data[ ii*dataInfo.size], win, gc, &extents);
	if ( extents.width > maxWidth) {
	    maxWidth = extents.width;
	}
	if ( extents.height > maxHeight) {
	    maxHeight = extents.height;
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	take the maximum object extents and set them in the class.
 * Arguments:
 *	maxWidthArg, maxHeightArg - maximum object extents.
 *
 ****************************************************************************/

void 
AmObjectPlotClass::GetMaxObjectExtents( int	*maxWidthArg,
				      int	*maxHeightArg)
{
    *maxWidthArg = maxWidth;
    *maxHeightArg = maxHeight;
}

/****************************************************************************
 *
 * Description:
 *      Draw the current selection set.  The selection set is a buffer
 *      containing copies of the selected objects.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 *	fHighlight - TRUE: draw in highlight color, FALSE, draw as normal.
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the object plot's invalid.
 *
 ****************************************************************************/

AmErrorCode 
AmObjectPlotClass::DrawSelection( AmWindowClass		*window,
				AmGCClass			*gc, 
				AmRectangleRec const 	*exposeRects,
				int			nRects,
				AnvilBoolean			fHighlight)
{
    uchar_t		*ptr, *selected;
    int			ii, nSel;
    AmRectangleRec 	surroundRect;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
	DO_RETURN_ERROR( AmNOT_VALID);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* get the selection */

    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetSelection(( void **) &selected, &nSel);
    surroundRect.x -= maxWidth/2;
    surroundRect.y -= maxHeight/2;
    surroundRect.width += maxWidth;
    surroundRect.height += maxHeight;

/* get the pixel size for scaling, the line width, and the highlight color */

    gc->SetLineWidth( _dLineWidth);
    if ( fHighlight) {
	gc->SetForegroundColor( &highlightColor);
    } else {
	gc->SetForegroundColor( &fgColor);
    }

/* draw the items in the selected set */

    for ( ii = 0, ptr = selected; ii < nSel; ii++, ptr += dataInfo.size) {
	DrawObject( display, window, gc, ptr, &surroundRect);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	virtual member function to draw a single object 
 * Arguments:
 *	display - display that the selection set will be drawn to.
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      obj - object to draw.
 *      boundRect - bounding rectangle to clip the object to.
 *
 ****************************************************************************/


void 
AmObjectPlotClass::DrawObject( AmDisplayClass	*display,
			     AmWindowClass	*window,
			     AmGCClass		*gc,
			     uchar_t		*obj,
			     AmRectangleRec const	*boundRect)
{
    float	x, y;
    int		screenX, screenY;

    x = ( dataInfo.xFn)( obj + dataInfo.xOffset, dataInfo.clientData);
    y = ( dataInfo.yFn)( obj + dataInfo.yOffset, dataInfo.clientData);
    GetScreenPt( x, y, &screenX, &screenY);

/* only draw the object if it's in the bounding rectangle */

    if (( boundRect == NULL) || AmPtInRect( boundRect, screenX, screenY)) {
	( drawFn)( display, window, gc, screenX, screenY,
		   PixelSize(), obj, drawFnClientData);
    }
}

/****************************************************************************
 *
 * Description:
 *      Draw the objects in the object plot.  Clear out the background if the
 *      plot's opaque, and draw the objects.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, as it should be.
 *
 ****************************************************************************/

AmErrorCode 
AmObjectPlotClass::draw( AmWindowClass		*window,
		       AmGCClass			*gc, 
		       AmRectangleRec const 	*exposeRects,
		       int			nRects)
{
    uchar_t		*ptr;
    int			ii;
    AmRectangleRec	surroundRect;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* if there's no overlay data, then we clear the area that we're going to
   draw to.  We have to use FillRectangle(), because we have to pay
   attention to the gc's clip rectangle which is set by the parent */

    if (( parent == NULL) && opaque) {
        gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width, 
				    exposeRects[ ii].height);
	}
    }

/* set the drawing attributes */

    gc->SetForegroundColor( &fgColor);
    gc->SetFont( font, fontSize, fontWeight, fontSlant, 
		 fontXScale, fontYScale);
    gc->SetLineWidth( _dLineWidth);
    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    surroundRect.x -= maxWidth/2;
    surroundRect.y -= maxHeight/2;
    surroundRect.width += maxWidth;
    surroundRect.height += maxHeight;
    for ( ii = 0, ptr = dataInfo.data; 
	  ii < dataInfo.dims[ 0]; 
	  ii++, ptr += dataInfo.size) {
	DrawObject( display, window, gc, ptr, &surroundRect);
    }

/* restore the line width */

    gc->SetLineWidth( 0);

/* if the object plot isn't inside of an overlay, then highlight the current
   selection */

    if ( parent == NULL) {
	return( DrawSelection( window, gc, exposeRects, nRects, TRUE));
    } else {
	return( AmOK);
    }
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an object plot.  This is
 *      used in the map editor to dump the plot's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmObjectPlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			     AnvilBoolean	       allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( plot) */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmObjectPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an object plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmObjectPlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  The vector must be sorted 
 *	in alphabetical order on the tag name for searching.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmObjectPlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				      int		*numDefs)
{
    static AmFieldDefRec ObjectPlotClassFieldDefintions[] = {
	{ AmNdrawFn, offsetof( AmObjectPlotClass, drawFn) },
	{ AmNdrawFnClientData, offsetof( AmObjectPlotClass, drawFnClientData) },
	{ AmNdrawFnClientDataFreeFn, 
		offsetof( AmObjectPlotClass, drawFnClientDataFreeFn) },
	{ AmNfont, offsetof( AmObjectPlotClass, font), 0, AmDYNAMIC },
	{ AmNfontSize, offsetof( AmObjectPlotClass, fontSize) },
	{ AmNfontSlant, offsetof( AmObjectPlotClass, fontSlant) },
	{ AmNfontWeight, offsetof( AmObjectPlotClass, fontWeight) },
	{ AmNfontXScale, offsetof( AmObjectPlotClass, fontXScale) },
	{ AmNfontYScale, offsetof( AmObjectPlotClass, fontYScale) },
	{ AmNlineWidth, offsetof( AmObjectPlotClass, _dLineWidth) },
	{ AmNsetXFn, offsetof( AmObjectPlotClass, setXFn) },
	{ AmNsetXYFn, offsetof( AmObjectPlotClass, setXYFn) },
	{ AmNsetYFn, offsetof( AmObjectPlotClass, setYFn) }
    };
    *fieldDefs = ObjectPlotClassFieldDefintions;
    *numDefs = VECLEN( ObjectPlotClassFieldDefintions);
}
