#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <malloc.h>
#include <anvil/strlib.h>
#include <anvil/list.h>
#include <anvil/macros.h>
#include <anvil/overlay.h>
#include <anvil/userevents.h>
#include <anvil/eventcodes.h>

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmOverlayClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmOVERLAY) {
        return( TRUE);
    } else {
        return( AmCoordClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 * 	does the layer use an image? This should probably be changed to
 *	use a virtual function which returns TRUE/FALSE whether the layer
 *	uses an image.  This is used so that image-based plots are 
 *	on the bottom of the stacking order, and also to cause Overlay
 *	to allocate the image.
 * Arguments:
 *      a: layer in the overlay.
 * Returns: 
 *	TRUE if the layer uses an image, FALSE otherwise.
 *
 ****************************************************************************/


AnvilBoolean 
AmImageTypeLayer( AmDataGraphicClass *a)
{
    AmObjectTypeEnum objectType;

    objectType = a->GetObjectType();
    return( a->IsType( AmIMAGE_PLOT) ||
	    a->IsType( AmSCATTER_PLOT) ||
	    a->IsType( AmSCATTER_GRAPH));
}

/****************************************************************************
 *
 * Description:
 *      AmOverlayClass constructor.  Creates an overlay using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the overlay
 * Arguments:
 *      windowArg - window that the overlay will be drawn on.
 *      nameArg - string name of the overlay, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmOverlayClass::AmOverlayClass( AmWindowClass	 *windowArg,
			    char	 *nameArg,
			    ...)
{
    va_list	args;
    AmTagValueRec	tagValues[ MAXTAGS];
    int		count;

    AmScreenObjClass::initializeFields( windowArg, nameArg);
    AmCoordClass::setCoordClass();
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    displayImage = NULL;	/* the image is empty because size unknown */
    layerList = NULL;		/* the plot list is added to later */
    objectType = AmOVERLAY;
    currLayer = NULL;		/* there is no current layer */
    maintainRange = TRUE;
    maintainXRange = TRUE;
    setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 *      AmOverlayClass destructor.  Walk through the layer list and call
 *	each of the layers' virtual destructors. Delete the display 
 *	image if one was used.
 *
 ****************************************************************************/

AmOverlayClass::~AmOverlayClass()
{
    AmListPtr		ptr;
    AmDataGraphicClass	*layer;

    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;
	delete layer;
    }
    AmJustFreeLst( layerList);
    if ( displayImage != NULL) {
	delete displayImage;
    }
}

/****************************************************************************
 *
 * Description:
 * 	retreive the display image from the overlay 
 * Returns:
 *	the display image used int the overlay (may be NULL).
 *
 ****************************************************************************/

AmImageClass *
AmOverlayClass::GetImage()
{
    return( displayImage);
}

/****************************************************************************
 *
 * Description:
 * 	set the display image for the overlay 
 * Arguments:
 *	the display image used int the overlay (may be NULL).
 *
 ****************************************************************************/

void 
AmOverlayClass::SetImage( AmImageClass *image)
{
    displayImage = image;
}

static int 
comparePriority( AmDataGraphicClass    *a, 
		 AmDataGraphicClass    *b)
{
    return( a->GetPriority() - b->GetPriority());
}

/****************************************************************************
 *
 * Description:
 * 	Test whether a layer is within the overlay.
 * Arguments:
 *	layer - data graphic to test against the overlay's layer list.
 *
 ****************************************************************************/

AnvilBoolean 
AmOverlayClass::IsLayer( AmDataGraphicClass	*layer)
{
    AmListPtr	ptr;

    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	if (( AmDataGraphicClass *) ptr->data == layer) {
	    return( TRUE);
	}
    }
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	Add a layer to the overlay.  Most of the work deals with the
 *	coordinate systems to ensure that projections, ranges and
 *	current extents are maintained properly:
 *	(with maintain range set)
 *	Extents:  If the layer is the AmFirst layer added, and the range
 *	for the overlay is uninitialized, the extents of the overlay are
 *	set to the layer.  If the overlay already contains layers, the
 *	extents are grown to contain the new layer.
 *	Current Extents:  If the layer is the AmFirst layer added, and
 *	the overlay's extents are unset, then the current extents of
 *	the overlay are set to the current extents of the layer (if
 *	they are set), otherwise the overlay's current extents are
 *	set to the range of the new layer.
 *	If maintainRange is not set, then the current extents for
 *	all the layers are forced to the same coordinate system.
 *	Projections:  If the layer has a projection applied, or the
 *	overlay has a projection to apply, and the layer is not already
 *	in that projection, then the layer is drawn with the projection
 *	applied.  If it is the AmFirst layer added, then that projection
 *	is applied to the overlay and all successive plots.
 *	The layer is inserted sorted on the AmNpriority resource, but
 *	raster layers are always below vector layers.
 *	The function sends AM_COORDS_CHANGE, AM_SET_FIELD_OBJECT, and
 *	AM_SELECT_LAYER user events on completion.
 * Arguments:
 *	layer - data graphic to test against the overlay's layer list.
 *
 ****************************************************************************/

void 
AmOverlayClass::AddLayer( AmDataGraphicClass	*layer)
{
    float		layerXmin, layerYmin, layerXmax, layerYmax,
			overXmin, overYmin, overXmax, overYmax,
			currXmin, currYmin, currXmax, currYmax;
    unitType		layerXunits, layerYunits, 
			overlayXunits, overlayYunits;
    char		**layerProjStrs, **layerInProjStrs;
    AnvilBoolean		layerProjected;

    if ( !layer->IsValid()) {
	return;
    }

/* TURN OFF ASPECT PRESERVATION.  THE OVERLAY WILL PERFORM ASPECT 
   PRESERVATION AND TELL THE LAYERS WHAT TO DO.. */

    layer->SetVaFields( AmNpreserveAspect, FALSE,
			NULL);

/* set the layer rectangle to the rectangle of the overlay. Get the layer's
   projection strings, if any */

    layer->SetRect( rect.x, rect.y, rect.width, rect.height);
    layer->GetVaFields( AmNprojection, &layerProjStrs,
			AmNinProjection, &layerInProjStrs,
    			AmNxunits, &layerXunits,
			AmNyunits, &layerYunits,
			NULL);
    GetVaFields( AmNxunits, &overlayXunits,
		 AmNyunits, &overlayYunits,
		 NULL);

/* the layer is projected if the overlay is "in" a projection, the layer is
   projected and the layer is not "in" a projection, and the units don't
   match */

    if (( layerXunits == AmUNIT_UNDEFINED) ||
	( layerYunits == AmUNIT_UNDEFINED)) {
	AN_DEBUG( 1, ("layer units are undefined\n"));
    }
    layerProjected = (( inProjStrs != NULL) || 
		      (( layerProjStrs != NULL) && 
		       strcmp( *layerProjStrs, "proj=none"))) &&
		     ( layerInProjStrs == NULL) &&
		     (( layerXunits != overlayXunits) && 
		      ( layerYunits != overlayYunits));

/* if the data is in a projection, then set the inProjection strings for
   the overlay. If the overlay is in a projection, and the data is not
   projected, then apply the projection to the overlay */

    if (( layerInProjStrs != NULL) && ( inProjStrs == NULL)) {
	SetVaFields( AmNinProjection, layerInProjStrs,
		     AmNnumInProjStrs, 
			AmVeclen(( const char **) layerInProjStrs), 
		     NULL);
    } else if ( layerProjected && ( layerProjStrs == NULL)) {
	layer->SetVaFields( AmNprojection, inProjStrs, 
			    AmNnumProjStrs, 
				AmVeclen(( const char **) inProjStrs),
			    NULL);
    }

/* if this is the AmFirst layer being set, then get its current coordinates 
   and set them for the overlay range if the overlay range isn't set.
   Also set the x and y units for the overlay layer.  Don't bother with
   any of this if the overlay doesn't maintain range for the layers */

    if ( maintainRange) {
	if ( layerList == NULL) {

/* set the range for the overlay to reflect the AmFirst added layer */

	    if ( layer->RangeSet()) {
		layer->GetRange( &layerXmin, &layerYmin, 
				 &layerXmax, &layerYmax);
		if ( layerProjected) {
		    layer->GetForwardProjection()->ProjectPt( 
			    &layerXmin, &layerYmin);
		    layer->GetForwardProjection()->ProjectPt( 
			    &layerXmax, &layerYmax);
		}

/* SetRange calls set coords, which sets the coords set flag, so we need 
   to know it before setting the range */

		SetRange( layerXmin, layerYmin, layerXmax, layerYmax);
	    }

/* if the layer has coordinates set, then set them in the overlay */

	    if ( layer->CoordsSet()) {
		layer->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
	    } else if ( layer->RangeSet()) {
		layer->GetRange( &currXmin, &currYmin, &currXmax, &currYmax);
	    }
	    if ( layerProjected) {
		layer->GetForwardProjection()->ProjectPt( 
				&currXmin, &currYmin);
		layer->GetForwardProjection()->ProjectPt( 
				&currXmax, &currYmax);
	    }

/* set the coordinates for all the overlay and for the layer, which hasn't
   been added to the layer list yet... */

	    AmCoordClass::SetCoords( currXmin, currYmin, currXmax, currYmax);
	    GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
	    if ( layerProjected) {
		layer->GetForwardProjection()->InvProjectPt( 
				&currXmin, &currYmin);
		layer->GetForwardProjection()->InvProjectPt( 
				&currXmax, &currYmax);
	    }
	    layer->SetCoords( currXmin, currYmin, currXmax, currYmax);

/* if the layer is projected, then set the overlay units to the result of
   the projection, otherwise set them to the layer's units. */ 

	    if ( layerProjected) {
		layerXunits = layer->GetForwardProjection()->resultUnits();
		layerYunits = layer->GetForwardProjection()->resultUnits();
	    } else {
		layer->GetVaFields( AmNxunits, &layerXunits,
				    AmNyunits, &layerYunits,
				    NULL);
	    }
	    SetVaFields( AmNxunits, layerXunits,
			 AmNyunits, layerYunits,
			 NULL);
	    currLayer = layer;		/* set the current layer to 1st layer */
	} else {

/* take the extents of the layer, and add it to the extents of the overlay. */

	    GetRange( &overXmin, &overYmin, &overXmax, &overYmax);
	    layer->GetRange( &layerXmin, &layerYmin, 
			     &layerXmax, &layerYmax);
	    if ( layerProjected) {
		layer->GetForwardProjection()->ProjectPt( 
			&layerXmin, &layerYmin);
		layer->GetForwardProjection()->ProjectPt( 
			&layerXmax, &layerYmax);
	    } 
	    if ( layerXmin < overXmin) {
		overXmin = layerXmin;
	    } 
	    if ( layerYmin < overYmin) {
		overYmin = layerYmin;
	    }
	    if ( layerXmax > overXmax) {
		overXmax = layerXmax;
	    }
	    if ( layerYmax > overYmax) {
		overYmax = layerYmax;
	    }
	    AmCoordClass::SetRange( overXmin, overYmin, overXmax, overYmax);

/* set the layer's coordinate system to the overlay's */

	    if ( CoordsSet()) {
		GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
		if ( layerProjected) {
		    layer->GetForwardProjection()->InvProjectPt( 
			    &currXmin, &currYmin);
		    layer->GetForwardProjection()->InvProjectPt( 
			    &currXmax, &currYmax);
		}
		layer->AmCoordClass::SetCoords( currXmin, currYmin, 
					      currXmax, currYmax);
	    }
	 
/* if the overlay is in a projection, and the layer doesn't have a projection,
   apply the overlay's projection here */
 
	    if ( layerProjected) {
		layerXunits = layer->GetForwardProjection()->resultUnits();
		layerYunits = layer->GetForwardProjection()->resultUnits();
	    } else {
		layer->GetVaFields( AmNxunits, &layerXunits,
				    AmNyunits, &layerYunits,
				    NULL);
	    }
	    if (( layerXunits != overlayXunits) || 
		( layerYunits != overlayYunits)) {
		char	*xunitStr, *yunitStr, *overlayXunitStr, 
			    *overlayYunitStr;

		AmGetUnitString( layerXunits, &xunitStr);
		AmGetUnitString( layerYunits, &yunitStr);
		AmGetUnitString( overlayXunits, &overlayXunitStr);
		AmGetUnitString( overlayYunits, &overlayYunitStr);
		AN_DEBUG( 20, ( "layerOverlayUnitsNotSame %s %s %s %s",
			        xunitStr, yunitStr, 
			        overlayXunitStr, overlayYunitStr));
	    }
	}

/* free projection string vectors if they're not NULL */

	if ( layerProjStrs != NULL) {
	    AmFreeStrVec( layerProjStrs);
	}
	if ( layerInProjStrs != NULL) {
	    AmFreeStrVec( layerInProjStrs);
	}
    }

/* set the parent of the layer to be this overlay */

    layer->setParent( this);

/* add the layer to the list of overlays */

    layerList = AmInsertSortLst( layer, layerList, 
				 ( AmSearchFn) comparePriority);

/* send a user event to the display manager saying that the coordinates
   have changed, that the overlay's fields have changed */

    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
    AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, this);

    /* Make this the selected layer. */

    currLayer = layer; 
    AmSendUserEvent( AM_PKG_ID, AM_SELECT_LAYER, currLayer);
    AmSendUserEvent( AM_PKG_ID, AM_ADD_LAYER, currLayer);
}

/****************************************************************************
 *
 * Description:
 * 	remove a layer from the layer list.  The range of the overlay
 *	is recalculated with the layer removed.
 * Arguments:
 *	layer: layer to remove.
 *	The function sends AM_COORDS_CHANGE, AM_SET_FIELD_OBJECT, 
 *	AM_DELETE_LAYER, and AM_SELECT_LAYER user events on completion.
 * Modifications:
 *	The function should return TRUE/FALSE depending on whether the
 *	layer was in the overlay.
 *
 ****************************************************************************/

void 
AmOverlayClass::RemoveLayer( AmDataGraphicClass	*layer)
{
    AmListPtr		ptr;
    AmDataGraphicClass	*layerPtr, *lastImageLayer;
    float		xmin, ymin, xmax, ymax, 
			overXmin, overYmin, overXmax, overYmax;

/* delete the layer from the layer list */

    layerList = AmDelPtrLst( layer, layerList, ( AmSearchFn) AmEquals);

/* reset the current layer to the AmFirst layer if this was the current layer,
   then delete the layer */

    if (( currLayer == layer) && ( layerList != NULL)) {
	currLayer = ( AmDataGraphicClass *) layerList->data;
    } else if ( layerList == NULL) {
	currLayer = NULL;
    }

/* if there are any layers left, take the ranges of all the layers, 
   and reset the extents of the overlay to reflect them */

    if ( layerList != NULL) {
	xmin = MY_HUGE;
	ymin = MY_HUGE;
	xmax = -MY_HUGE;
	ymax = -MY_HUGE;
	for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	    layerPtr = ( AmDataGraphicClass *) ptr->data;
	    layerPtr->GetRange( &overXmin, &overYmin, &overXmax, &overYmax);
	    if ( layerPtr->GetForwardProjection()->IsProjected()) {
		layerPtr->GetForwardProjection()->ProjectPt( 
			&overXmin, &overYmin);
		layerPtr->GetForwardProjection()->ProjectPt( 
			&overXmax, &overYmax);
	    }
	    if ( overXmin < xmin) {
		xmin = overXmin;
	    }
	    if ( overYmin < ymin) {
		ymin = overYmin;
	    }
	    if ( overXmax > xmax) {
		xmax = overXmax;
	    }
	    if ( overYmax > ymax) {
		ymax = overYmax;
	    }
	}
	AmCoordClass::SetRange( xmin, ymin, xmax, ymax);
    }

/* if there are no image layers left, then remove the display image */

    lastImageLayer = ( AmDataGraphicClass *) 
	AmLastOfLst(( AmElementFn) AmImageTypeLayer, layerList);
    if ( lastImageLayer == NULL) {
	if ( displayImage != NULL) {
	    delete displayImage;
	    displayImage = NULL;
	}
    }

/* send a user event to the display manager saying that the coordinates
   have changed, that the overlay's fields have been set, and that the
   layer has been deleted */

    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
    AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, this);
    AmSendUserEvent( AM_PKG_ID, AM_DELETE_LAYER, layer);

    if ( currLayer != NULL) {
        AmSendUserEvent( AM_PKG_ID, AM_SELECT_LAYER, currLayer);
    }
}

/****************************************************************************
 *
 * Description:
 * 	resize the display image (if there is one) to be in sync with the 
 *  	displayRectangle.  This is used when the overlay is resized.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmOverlayClass::syncDisplayImage()
{
    AmDataGraphicClass	*lastImageLayer;

/* If the screen has resized, then resize the image, reset the 
   update rectangle to the new screen extents, by setting it to NULL 
   ( which it may already be) */

    lastImageLayer = ( AmDataGraphicClass *) 
	AmLastOfLst(( AmElementFn) AmImageTypeLayer, layerList);
    if ( lastImageLayer != NULL) {
	if (( displayImage = ResizeDisplayImage( displayImage)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    } else if ( displayImage != NULL) {

/* if there are no plots using the display image, then set it to NULL */

	delete displayImage;
	displayImage = NULL;
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Override member function from coord class, which takes the coordinates 
 *	which were set in the overlay and applies them to everyone in the 
 *	layer list.  If the maintainRange flag is set, then the coordinates
 *	are kept in sync, otherwise, it's used relative to the overlay's
 *	range to set the zoom factor.
 * Arguments:
 *	currXmin, currYmin, currXmax, currYmax: current extents to set the
 *	overlay to.
 *
 ****************************************************************************/

void 
AmOverlayClass::SetCoords( float	currXmin,
			 float 	currYmin,
			 float 	currXmax,
			 float 	currYmax)
{
    AmListPtr		ptr;
    AmDataGraphicClass	*layer;
    float		layerXmin, layerYmin, layerXmax, layerYmax,
			layerCurrXmin, layerCurrYmin, layerCurrXmax, 
			layerCurrYmax, xmin, ymin, xmax, ymax;

    AmCoordClass::SetCoords( currXmin, currYmin, currXmax, currYmax);
    GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);

/* if we maintain range for the layers, set the coordinates for each layer */

    if ( maintainRange) {
	for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	    layer = ( AmDataGraphicClass *) ptr->data;
	    if ( layer->GetForwardProjection()->IsProjected()) {
		layerCurrXmin = currXmin;
		layerCurrXmax = currXmax;
		layerCurrYmin = currYmin;
		layerCurrYmax = currYmax;
		layer->GetForwardProjection()->InvProjectPt( 
			&layerCurrXmin, &layerCurrYmin);
		layer->GetForwardProjection()->InvProjectPt( 
			&layerCurrXmax, &layerCurrYmax);
		layer->SetCoords( layerCurrXmin, layerCurrYmin, 
				  layerCurrXmax, layerCurrYmax);
	    } else {
		layer->SetCoords( currXmin, currYmin, currXmax, currYmax);
	    }
	}
    } else {

/* otherwise, scale the layers relative to their own range, depending on
   the overlay's scale */

	GetRange( &xmin, &ymin, &xmax, &ymax);
	for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	    layer = ( AmDataGraphicClass *) ptr->data;
	    layer->GetRange( &layerXmin, &layerYmin, &layerXmax, &layerYmax);
	    layerCurrXmin = layerXmin + 
		( layerXmax - layerXmin)*( currXmin - xmin)/( xmax - xmin);
	    layerCurrXmax = layerXmin + 
		( layerXmax - layerXmin)*( currXmax - xmin)/( xmax - xmin);
	    layerCurrYmin = layerYmin + 
		( layerYmax - layerYmin)*( currYmin - ymin)/( ymax - ymin);
	    layerCurrYmax = layerYmin + 
		( layerYmax - layerYmin)*( currYmax - ymin)/( ymax - ymin);
	    layer->SetCoords( layerCurrXmin, layerCurrYmin, 
			      layerCurrXmax, layerCurrYmax);
	}
    }

/* send a user event saying that the coordinates have changed */
    
    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 * 	Overrides the SetRect() member function for AmCoordClass. set the 
 *	rectangles for the overlay and all the layers 
 * Arguments:
 *	x, y, width, height: screen dimensions to set the overlay to.
 *
 ****************************************************************************/

void 
AmOverlayClass::SetRect( int	x,
		       int	y,
		       int	width,
		       int	height)
{
    AmListPtr		ptr;
    AmDataGraphicClass	*layer;

/* we have to set the rectangle and sync the display image before setting
   the rectangle, because SetRect() can call SetCoords(), which can update
   the layer before the rectangle is set */

    AmScreenObjClass::SetRect( x, y, width, height);
    if ( syncDisplayImage() != AmOK) {
	return;
    }
    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;

/* call the low level set rect, because we don't want to update until all
   the rectangles are in sync, which we do with a final set recta call */

	layer->AmScreenObjClass::SetRect( x, y, width, height);
    }
    AmCoordClass::SetRect( x, y, width, height);

/* send a user event saying that the coordinates have changed */
    
    AmSendUserEvent( AM_PKG_ID, AM_COORDS_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 * 	draw function for overlay.  Draw each of the raster layers, 
 *	then draw the vector layers (sorted by priority).
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmOverlayClass::draw( AmWindowClass	    	*window,
		    AmGCClass		*gc,
		    AmRectangleRec const  *exposeRects,
		    int		    	nRects)
{
    AmDataGraphicClass	*layer;
    AmListPtr		ptr;
    AmRectangleRec const	*dispRect = GetRect();
    AnvilBoolean		imageFlag;
    int			ii;

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }
    if ( AmDegenRect( dispRect)) {
	return( AmOK);
    }
    if ( layerList == NULL) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* walk through the list of layers, and draw all the image layers.  */

    imageFlag = FALSE;
    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;
	if ( AmImageTypeLayer( layer)) {
	    imageFlag = TRUE;
	    layer->draw( window, gc, exposeRects, nRects);
	}
    }

/* since the image has a 0,0 origin, we shift the expose rectangle from the
   window coordinate system to the image coordinate system */
 
    if ( imageFlag && ( displayImage != NULL)) {
	for ( ii = 0; ii < nRects; ii++) {
	    display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
				exposeRects[ ii].y - dispRect->y,
				exposeRects[ ii].x, exposeRects[ ii].y,
				exposeRects[ ii].width,
                                exposeRects[ ii].height,
				displayImage);
	}
    } else if ( opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width,
				    exposeRects[ ii].height);
	}
    }

/* now draw all the non-image layers */

    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;
	if ( !AmImageTypeLayer( layer)) {
	    layer->draw( window, gc, exposeRects, nRects);
	}
    }

/* now draw the selection for each layer */

    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;
	layer->DrawSelection( window, gc, exposeRects, nRects, TRUE);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Update the overlay, based on the events which have been sent
 *      using AmSendUserEvent().  This calls Update() for each of the layers.
 * Arguments:
 *      eventCodes - list of event codes sent by AmSendUserEvent()
 *      numCodes - number of event codes.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/


AmErrorCode 
AmOverlayClass::Update( int	*eventCodes,
		      int	numEvents)
{
    AmListPtr		ptr;
    AmDataGraphicClass	*layer;
    AmRectangleRec const  *dispRect = GetRect();
    AmRectangleRec	clearRect;
    AmErrorCode		errCode;

    if ( !IsValid()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }
AN_DEBUG( 20, ("in update call for overlay plot %s\n", getObjectName()));

/* If the screen has resized, then resize the image, reset the 
   update rectangle to the new screen extents, by setting it to NULL 
   ( which it may already be) */

    errCode = syncDisplayImage();
    if ( errCode != AmOK) {
	return( errCode);
    }

/* clear out the background image */

    if ( displayImage != NULL) {
	clearRect = *dispRect;
	clearRect.x = 0;
	clearRect.y = 0;
	displayImage->ClearImageRect( &bgColor, &clearRect);
    }
    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	layer = ( AmDataGraphicClass *) ptr->data;
	layer->Update( eventCodes, numEvents);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	set the layers in an overlay...we're presuming that the layers are
 *  	simply the old layers resorted. 
 * Arguments: 
 *	layers - vector of pointers to datagraphics to set in the layer
 *	list.  This is expected to be the same length as the overlay's
 *	layer list.
 *	numLayers - number of layers to set (unused).
 * 
 ****************************************************************************/

/* ARGSUSED */

void 
AmOverlayClass::SetLayers( AmDataGraphicClass	**layers,
			 int			numLayers)
{
    AmListPtr	ptr;

    for ( ptr = layerList; ptr != NULL; ptr = ptr->next) {
	ptr->data = *layers++;
    }
    AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, this);
}

/****************************************************************************
 *
 * Description:
 * 	return the vector of layers in the overlay 
 * Arguments:
 *	layers - pointer to a vector of pointers to the layers in the overlay.
 *	numlayers - number of layers returned.
 * Returns:
 *	AmOK if everything's OK, AmMALLOC_FAILED if the vector couldn't
 *	be allocated.
 * 
 ****************************************************************************/

AmErrorCode 
AmOverlayClass::GetLayers( AmDataGraphicClass	***layers,
			 int			*numLayers)
{
    *numLayers = AmLengthLst( layerList);
    *layers = ( AmDataGraphicClass **) AmMakeVectorLst( layerList);
    if ( *layers == NULL) {
        *numLayers = 0;
    }
    return(( *layers != NULL) ? AmOK : AmMALLOC_FAILED);
}

/****************************************************************************
 *
 * Description:
 * 	return the vector of layers in the overlay which match the 
 *	selected appId.
 * Arguments:
 *	layers - pointer to a vector of pointers to the layers in the overlay.
 *	with the same AppID as the one passed.
 *	numlayers - number of layers returned. 
 *	appId - application ID to search for.
 * Returns:
 *	AmOK if everything's OK, AmMALLOC_FAILED if the vector couldn't
 *	be allocated.
 * 
 ****************************************************************************/

AmErrorCode 
AmOverlayClass::GetLayersByAppId( AmDataGraphicClass	***layers,
				int			*numLayers,
				const char		*appId)
{
    int	 	count, ii;
    char 	*candAppId;
    AmErrorCode	errCode;

    if (( errCode = GetLayers( layers, numLayers)) != AmOK) {
	return( errCode);
    }
    count = 0;
    for ( ii = 0; ii < *numLayers; ii++) {
	( *layers)[ ii]->GetVaFields( AmNappId, &candAppId, NULL);
	if ( candAppId != NULL) {
	    if ( AmMatchTag( appId, candAppId)) {
		( *layers)[ count++] = ( *layers)[ ii];
	    }
	    free( candAppId);
	}
    }
    *numLayers = count;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for an overlay, given the tag-values stored in
 *      a tag values array.  First, the generic coord fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  Note the exception case for handling 
 *	preserveAspect, which affects all layers 
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmOverlayClass::setFields( AmTagValuePtr	tagValues,
			 int		numValues)
{
    AmFieldDefPtr		fieldDefs;
    int			numFieldDefs;

    numValues = AmCoordClass::setFields( tagValues, numValues);
    AmOverlayClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for an overlay, given the tag-values stored in
 *      a tag values array.  First, the coordinate fields are retrieved, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/
 
int 
AmOverlayClass::getFields( AmTagValuePtr      tagValues,
			 int              numValues)
{
    AmFieldDefPtr fieldDefs;
    int         numFieldDefs;
 
    numValues = AmCoordClass::getFields( tagValues, numValues);
    AmOverlayClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
                                   tagValues, numValues));
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in an overlay.  This is
 *      used in the map editor to dump the overlay's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmOverlayClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			  AnvilBoolean	    allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( datagraphic) */

    numStrings = AmCoordClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmOverlayClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for an overlay. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmOverlayClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmOverlayClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				   int		*numFieldDefs)
{
    static AmFieldDefRec OverlayClassFieldDefintions[] = {
	{ AmNcurrLayer, offsetof( AmOverlayClass, currLayer) },
	{ AmNmaintainRange, offsetof( AmOverlayClass, maintainRange) },
	{ AmNmaintainXRange, offsetof( AmOverlayClass, maintainXRange) }
    };
    *fieldDefs = OverlayClassFieldDefintions;
    *numFieldDefs = VECLEN( OverlayClassFieldDefintions);
}
