#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#if ( defined(_WIN32) || defined(__BSD44__)) 
#include <malloc.h>
#else 
#include <alloca.h>
#endif
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <anvil/macros.h>
#include <anvil/resource.h>
#include <anvil/mem_funcs.h>
#include <anvil/geometry.h>
#include <anvil/overlay.h>
#include <anvil/plineplot.h>
#include <anvil/datafmt.h>
#include <anvil/compression.h>
#include <anvil/log.h>
#include <anvil/units.h>

static const int MAXTILES = 1024;

/****************************************************************************
 *
 * Description:
 * 	range function for polylines. This overrides the virtual function
 *	for AmDataGraphicClass, which only tests x,y coordinates.
 * Arguments:
 *	xmin, ymin, xmax, ymax: returned data extents.
 *
 ****************************************************************************/

void 
AmPolyLinePlotClass::GetDataExtents( float        *xmin,
				     float        *ymin,
				     float        *xmax,
				     float        *ymax)

{
    uchar_t     	*ptr;
    AmPolylinePtr       pline;
    int         	ii;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0, ptr = dataInfo.data;
          ii < dataInfo.dims[ 0];
          ii++, ptr += dataInfo.size) {
        pline = ( AmPolylinePtr) ptr;
        if ( pline->xmin < *xmin) {
            *xmin = pline->xmin;
        }
        if ( pline->ymin < *ymin) {
            *ymin = pline->ymin;
        }
        if ( pline->xmax > *xmax) {
            *xmax = pline->xmax;
        }
        if ( pline->ymax > *ymax) {
            *ymax = pline->ymax;
        }
    }
}

/****************************************************************************
 *
 * Description:
 *      find the extents of a polyline, and set them in the polyline.  This
 *      also sets the "minimum bounding rectangle has been set" flag.  If
 *      the flag's already been set, this function does nothing.
 * Arguments:
 *      poly - polygon to find the extents of.
 *
 ****************************************************************************/

static void
setPolylineExtents( AmPolylinePtr pline)
{
    int ii;

    if ( !( pline->flags & AM_MBR_FLAG)) {
        pline->xmin = MY_HUGE;
        pline->ymin = MY_HUGE;
        pline->xmax = -MY_HUGE;
        pline->ymax = -MY_HUGE;

        for ( ii = 0; ii < pline->numPts; ii++) {
            if ( pline->pts[ ii].x < pline->xmin) {
                pline->xmin = pline->pts[ ii].x;
            }
            if ( pline->pts[ ii].y < pline->ymin) {
                pline->ymin = pline->pts[ ii].y;
            }
            if ( pline->pts[ ii].x > pline->xmax) {
                pline->xmax = pline->pts[ ii].x;
            }
            if ( pline->pts[ ii].y > pline->ymax) {
                pline->ymax = pline->pts[ ii].y;
            }
        }
        pline->flags |= AM_MBR_FLAG;
    }
}


/****************************************************************************
 *
 * Description:
 *	Recursive type function which tests the object type against the
 *	passed type and recursively tests the parent class's type.
 * Arguments:
 *	type: object type for testing.
 * Returns:
 *	TRUE if object class or any of its ancestors is of requested type, 
 *	FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPolyLinePlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmPOLYLINE_PLOT) {
       return( TRUE); 
    } else {
       return( AmPlotClass::IsType( type)); 
    }
}

/****************************************************************************
 *
 * Description:
 * 	initialize fields which are specific to a polyline plot.  Call the 
 *	basic plot class initializations, then initialize the line width
 *	tables, the line buffer size and the object.  We don't use
 *	constructors for this because of ambiguities with varargs
 * Arguments:
 *	windowArg - window that the line plot will be drawn on.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 *
 ****************************************************************************/

void
AmPolyLinePlotClass::initializeFields( AmWindowClass	*windowArg,
				       const char	*nameArg)
{
    AmPlotClass::initializeFields( windowArg, nameArg);
    _dLineWidth = 0.0;
    _pdLineWidth = NULL;
    _pdLineWidthScale = NULL;
    AmInitIndexData( &_indexLineWidth);
    _indexLineWidth.offset = sizeof( AmLineRec);

    plineBufSize = PLINEBUFSIZE;
    objectType = AmPOLYLINE_PLOT;
    _textGridDensity = 100;

    /*
     * Color function has to be set explicitly. Default doesn't work 
     *   the range function changes to take the vector endpoints for
     *   finding the extents of the plot.
     */

    colorIndex.offset = sizeof( AmLineRec);
    stringIndex.offset = sizeof( AmLineRec);
}

/****************************************************************************
 *
 * Description:
 *      check values for a line plot.  Each check values function calls the
 *      check values function for the parent class and is used to set
 *      the object's IsValid() flag. 
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmPolyLinePlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *	do-nothing constructor so line plots can be subclassed. 
 *
 ****************************************************************************/

AmPolyLinePlotClass::AmPolyLinePlotClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	AmPolyLinePlotClass constructor.  Creates a line plot using the window
 *	that it will be drawn on and the name, followed by NULL-terminated
 *	tag-value pairs for setting the fields in the line plot.
 * Arguments:
 *	windowArg - window that the line plot will be drawn on.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 * 	argList - NULL-teriminated tag value pairs, with tags defined in 
 *		  <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolyLinePlotClass::AmPolyLinePlotClass( AmWindowClass  *windowArg,
					  const char     *nameArg, 
					  ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 * 	AmPolyLinePlotClass constructor.  Creates a pline plot as part of an 
 *	overlay.  The proprerties for the line plot are set by a 
 *	NULL-terminated vector of tag-value pairs.
 * Arguments:
 *	overlayArg - overlay that the line plot will be added to.
 *	nameArg - string name of the line plot, used for identification
 *		  and pulling resources.
 * 	argList - NULL-teriminated tag value pairs, with tags defined in 
 *		  <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolyLinePlotClass::AmPolyLinePlotClass( AmOverlayClass  *overlayArg,
					  const char      *nameArg,
					  ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* we want all of the parameters which are set for the overlay plot to be
   set for the line plot as well. Set the overlay plot back pointer to 
   reference in the draw and destroy routines 
   color function has to be set explicitly. Default doesn't work */

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 * 	AmPolyLinePlotClass destructor.  Free the line width vectors if 
 *	they've been allocated.
 *
 ****************************************************************************/

AmPolyLinePlotClass::~AmPolyLinePlotClass()
{
    AmDestroyIndexData( &_indexLineWidth);
    if ( _pdLineWidth != NULL) {
	free( _pdLineWidth);
    }
    if ( _pdLineWidthScale != NULL) {
	free( _pdLineWidthScale);
    }
}

/****************************************************************************
 *
 * Description:
 * 	Set the fields for a polyline plot, given the tag-values stored in
 *	a tag values array.  First, the generic plot fields are set, then
 *	the other fields are set using the standard SetFieldByDefintions()
 *	function which uses a lookup table from the tag strings to the
 *	structure offsets.  The line width bin vector is set by hand because
 *	the length is one less than the line width vector.
 * Arguments:
 *	tagValues - vector of tag-value pairs (with tagCopied and copied
 *	flags).  The unparsed arguments are returned "scrunched" in this
 *	array.
 *	numValues - length of the vector.
 * Returns:
 *	number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolyLinePlotClass::setFields( AmTagValuePtr	tagValues,
				int		numValues)
{
    int		numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr	fieldDefs;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmPolyLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNlineWidthBinvec)) {
	    AmSetBinvec( &_indexLineWidth.binvec, _indexLineWidth.length,
		       &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmPolyLinePlotClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmPlotClass::updateFields();

/* update the index data for line width and color functions */

    if ( IsValid()) {
	if (( _pdLineWidth != NULL) || ( _pdLineWidthScale != NULL)) {
	    if ( updateIndexData( &_indexLineWidth, "lineWidth") != AmOK) {
		AmAddResourceToLog( "failedToInitializeLineWidthIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colorIndex") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }
    }
}

/****************************************************************************
 *
 * Description:
 * 	Retreive the fields for a line plot, given the tag-values stored in
 *	a tag values array.  First, the generic plot fields are retrived, then
 *	the other fields are retrieved using GetFieldByDefintions()
 *	function which uses a lookup table from the tag strings to the
 *	structure offsets.  The line width bin vector is retrieved by hand 
 *	because the length is one less than the line width vector.
 * Arguments:
 *	tagValues - vector of tag-value pairs (with tagCopied and copied
 *	flags).  The unparsed arguments are returned "scrunched" in this
 *	array.  The values are pointers to the returned values.
 *	numValues - length of the vector.
 * Returns:
 *	number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolyLinePlotClass::getFields( AmTagValuePtr	tagValues,
				int		numValues)
{
    int		numFieldDefs, numUnusedFields, ii;
    AmFieldDefPtr	fieldDefs;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmPolyLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNlineWidthBinvec)) {
	    if ( _indexLineWidth.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( _indexLineWidth.binvec,
			     ( _indexLineWidth.length - 1) * sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}

    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *	Draw the current selection set.  The selection set is a buffer
 *	containing copies of the selected lines.
 * Arguments:
 *	window - window that the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 *	exposeRects - vector of exposed rectangles.
 *	nRects - length of the expose rectangle vector.
 *	fHighlight - TRUE: draw in the highlight color, FALSE, draw
 *	in the normal color.
 * Returns:
 *	AmOK if everything's OK, AmNOT_VALID if the line plot's invalid.
 *
 ****************************************************************************/

AmErrorCode
AmPolyLinePlotClass::DrawSelection( AmWindowClass        *window,
				    AmGCClass            *gc, 
				    AmRectangleRec const *exposeRects,
				    int                  nRects,
				    AnvilBoolean		 fHighlight)
{
    float		xmin, ymin, xmax, ymax; 
    uchar_t		*selection;
    int			nSel;
    AmRectangleRec	surroundRect;

    /* Sanity checking. */

    if ( !IsValid()) {
	DO_RETURN_ERROR( AmNOT_VALID);
    }

    /* Return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
	return( AmOK);
    }

    /* Get the selection. */

    GetSelection(( void **) &selection, &nSel);
 
/* Take the expose rectangle and translate it to data coordinates to clip
   lines. */

    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
    GetDataPt( surroundRect.x + surroundRect.width,
	       surroundRect.y + surroundRect.height, &xmax, &ymin);

/* if highlighted, draw in the highlight color */

    drawPolylineVector( window, gc, selection, nSel, fHighlight,
			xmin, ymin, xmax, ymax);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      read polylines into a buffer, and use/set the number of bytes if
 *      there's a partial polyline.  This is used for reading polyline
 *      from files into buffers for display.
 * Arguments:
 *      fd - file descriptor to read from
 *      size - size of the polyline header.
 *      buffer - buffer to read into.
 *      buflen - length of the buffer.
 *      leftOverBytes - bytes left over from the previous read, and
 *      returns bytes left over from the current read.  This should be
 *      initialized to 0 the first time this function is called.
 *      bytesRead - number of bytes actually read.
 *      count - number of polylines actually read
 * Returns:
 *      AmREAD_FAILED if there was a read error.
 *      AmOK if everything's hunky-dory.
 *
 ****************************************************************************/

static AmErrorCode
readPolylineBuffer( int          fd,
                    int          size,
                    uchar_t      *buffer,
                    int          buflen,
                    int          *leftOverBytes,
                    int          *bytesRead,
                    int          *count)
{
    int         	 numBytes, numRead, plineSize;
    uchar_t     	 *ptr, *bufend;
    AmPolylinePtr        poly;

    numBytes = buflen - *leftOverBytes;
    if (( numRead = read( fd, &buffer[ *leftOverBytes], numBytes)) < 0) {
        return( AmREAD_FAILED);
    }
    bufend = &buffer[ numRead + *leftOverBytes];
    plineSize = 0;
    *count = 0;
    for ( ptr = buffer; ptr < bufend; ptr += plineSize) {
        poly = ( AmPolylinePtr) ptr;
        plineSize = size + poly->numPts*sizeof( AmCoordRec);
        if (( ptr + size > bufend) || ( ptr + plineSize > bufend)) {
            break;
        }
        poly->pts = ( AmCoordPtr)( ptr + size);
        ( *count)++;
    }
    *leftOverBytes = bufend - ptr;
    *bytesRead = ptr - buffer;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      read data from a file into a block of memory.  First, check if
 *      the file is in the cache, and if so, then use the cache pointer
 *      otherwise, open the file, seek over any header bytes,
 *      allocate the data and read it.  When reading the polygons, this
 *      function sets the pointer to the vector of vertices which follow
 *      the polygon.
 * Arguments:
 *      uses the dataInfo substructure.
 * Returns:
 *      AmOK if everything's OK, system error otherwise
 *
 ****************************************************************************/

AmErrorCode
AmPolyLinePlotClass::readData()
{
    int                 ii, length, fd, plineSize;
    AmPolylinePtr       pline;
    uchar_t             *ptr, *data, *dataEnd;
    struct stat         statbuf;

    if ( dataInfo.enableCacheFlag) {
	if (( data = ( uchar_t *) AmGetCache( dataInfo.filename)) != NULL) {
	    dataInfo.data = data;
	    return( AmOK);
	}
    }
    if (( fd = open( dataInfo.filename, O_RDONLY)) < 0) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* find out if the data is in the cache already */

    if ( fstat( fd, &statbuf) < 0) {
	DO_RETURN_ERROR( AmSTAT_FAILED);
    }
    length = statbuf.st_size - dataInfo.skipBytes;
    AN_DEBUG( 20, ("file size = %d\n", length));

/* polygons AmFirst, points second */

    if (( dataInfo.data = ( uchar_t *) malloc( length)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if ( lseek( fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if ( read( fd, dataInfo.data, length) < length) {
	DO_RETURN_ERROR( AmREAD_FAILED);
    }

/* step the data pointer by the size of each polygon */

    plineSize = 0;
    dataEnd = dataInfo.data + length;
    for ( ii = 0, ptr = dataInfo.data;
	  ii < dataInfo.dims[ 0];
	  ii++, ptr += plineSize) {
	pline = ( AmPolylinePtr) ptr;
	pline->pts = ( AmCoordPtr) ( ptr + dataInfo.size);
	plineSize = dataInfo.size + pline->numPts*sizeof( AmCoordRec);
	if ( ptr + plineSize > dataEnd) {
	    fprintf( stderr, "polyline %d, size %d overflows buffer\n", 
		     ii, plineSize);
	    fflush( stderr);
	    ASSERT( 0);
	}
	if ( !( pline->flags & AM_MBR_FLAG)) {
	    setPolylineExtents( pline);
	}
    }
    if ( dataInfo.enableCacheFlag) {
	AmAddToCache( dataInfo.filename, dataInfo.data);
    }
    close( fd);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Draw the lines in the line plot.  Clear out the background if the
 *	plot's opaque, and draw the lines from the buffer, read them block
 *	by block, or read them pyramidided into tiles.   The lines are
 *	buffered until one of the graphics attributes (color, linewidth)
 *	is changed.
 * Arguments:
 *	window - window the the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 *	exposeRects - vector of exposed rectangles.
 *	nRects - length of the expose rectangle vector.
 * Returns:
 *	AmOK if everything's OK, system error if there was a memory
 *	allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode
AmPolyLinePlotClass::draw( AmWindowClass        *window,
			   AmGCClass	    	*gc, 
			   AmRectangleRec const *exposeRects,
			   int                  nRects)
{
    int			ii, numPlines, count, leftOverBytes, bytesRead,
			densityRows=0, densityCols=0;
    uchar_t		*plineBuffer;
    float		xmin, ymin, xmax, ymax;
    AmRectangleRec	surroundRect;
    AmErrorCode		errCode;
    AnvilBoolean		*densityGrid=NULL;

    /* If this object is invalid only draw the outline. */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

    /* Return in the degenerate case */

    if ( AmDegenRect( GetRect())) {
	return( AmOK);
    }

    /*
     * If there's no overlay data, then we clear the area that we're going
     *   to draw to.  We have to use FillRectangle(), because we have to
     *   pay attention to the gc's clip rectangle which is set by the parent.
     */

    if (( parent == NULL) && opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, 
				    exposeRects[ ii].x, 
				    exposeRects[ ii].y,
				    exposeRects[ ii].width, 
				    exposeRects[ ii].height);
	}
    }

    /* create the density grid for spacing the text on the polylines */

    if ( showText) {
	densityRows = rect.height/_textGridDensity + 1;
	densityCols = rect.width/_textGridDensity + 1;
	densityGrid = ( AnvilBoolean *) alloca( densityRows*densityCols);
	if ( densityGrid == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	for ( ii = 0; ii < densityRows*densityCols; ii++) {
	    densityGrid[ ii] = FALSE;
	}
    }

    /*
     * Set the line width and line color, which may be overridden by functions
     *   in the loop.
     */

    _dLineWidthLast = _dLineWidth;
    gc->SetLineWidth( _dLineWidthLast);
    gc->SetLineCap( AmCAP_ROUND);

    _col = fgColor;
    _colLast = fgColor;
    gc->SetForegroundColor( &_colLast);
 
    /*
     * Take the expose rectangles, find the containing rectangle, and project
     *   it into the coordinate system for clipping the lines.
     */

    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
    GetDataPt( surroundRect.x + surroundRect.width,
               surroundRect.y + surroundRect.height, &xmax, &ymin);
	
    /*
     * Cast each element to a data segment and draw it.  Buffer the line
     *   segments on width and color to try to speed things up.
     */

    if ( dataInfo.sourceType == AmREAD_SOURCE) {

        /*
         * Allocate a buffer to draw a couple of lines, and read the data
         *   in chunks to display it.
         */

        plineBuffer = ( uchar_t *) alloca( plineBufSize);
	if ( plineBuffer == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}
	if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	leftOverBytes = 0;
	count = 0;
	while ( count < dataInfo.dims[ 0]) {
	    errCode = readPolylineBuffer( dataInfo.fd, dataInfo.size, 
					  plineBuffer, plineBufSize, 
					  &leftOverBytes, &bytesRead,
					  &numPlines);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	    errCode = drawPolylineVector( window, gc, plineBuffer,
					  numPlines, FALSE, 
					  xmin, ymin, xmax, ymax);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	    if ( showText) {
		drawPolylineVectorText( window, gc, plineBuffer, numPlines,
					xmin, ymin, xmax, ymax, 
					densityGrid, densityRows, densityCols);
	    }
	    count += numPlines;
	}
	if ( leftOverBytes > 0) {
	    memcpy( plineBuffer, &plineBuffer[ bytesRead], leftOverBytes);
	}
    } else if (( dataInfo.sourceType == AmDIRECT_SOURCE) ||
	       ( dataInfo.sourceType == AmFILE_SOURCE)) {
	drawPolylineVector( window, gc, dataInfo.data, dataInfo.dims[ 0], FALSE,
			    xmin, ymin, xmax, ymax);
	if ( showText) {
	    drawPolylineVectorText( window, gc, dataInfo.data, 
				    dataInfo.dims[ 0],
				    xmin, ymin, xmax, ymax, 
				    densityGrid, densityRows, densityCols);
	}
    } else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	AmTileRec	tiles[ MAXTILES];
	AmTilePtr	tile;
	AmErrorCode	errCode;
	int		numTiles, maxBytes;

	numTiles = 0;
	errCode = AmFindLayerTiles( coords.currXmin, coords.currYmin,
				    coords.currXmax, coords.currYmax,
				    ( float) detailLevel,
				    dataInfo.fd, &tileHeader->rootTile, tiles, 
				    &numTiles, MAXTILES);
	if ( errCode != AmOK) {
	    return( errCode);
	}

        /* Sort the tiles for performance. */

	qsort( tiles, numTiles, sizeof( AmTileRec), 
	       ( AmSearchFn) AmCompareTileSeekAddr);

        /* Allocate the line buffer to draw the largest tile. */

	maxBytes = 0;
	for ( ii = 0; ii < (signed)numTiles; ii++) {
	    if ( tiles[ ii].numBytes > (unsigned)maxBytes) {
		maxBytes = tiles[ ii].numBytes;
	    }
	}
	if (( plineBuffer = ( uchar_t *) alloca( maxBytes)) == NULL) {	
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}

/* search to each tile, uncompress it if the tile is compressed, or read
   it if it isn't, then draw the lines in the tile. */

	for ( ii = 0; ii < numTiles; ii++) {
	    tile = &tiles[ ii];
	    errCode = AmReadTile( tileHeader, tile, plineBuffer);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	    drawPolylineVector( window, gc, plineBuffer, 
				tile->rows, FALSE, xmin, ymin, xmax, ymax);
	    if ( showText) {
		drawPolylineVectorText( window, gc, plineBuffer, tile->rows,
					xmin, ymin, xmax, ymax, 
					densityGrid, densityRows, densityCols);
	    }
	    count += numPlines;
	}
    }

/* If the object plot isn't inside of an overlay, then highlight the
   current selection. */

    if ( parent == NULL) {
	return( DrawSelection( window, gc, exposeRects, nRects, TRUE));
    } else {
	return( AmOK);
    }
}

/****************************************************************************
 *
 * Description:
 * 	draw a single line, using the segments buffer as a GC caching 
 *	mechanism.  The segments vector is a buffer which is flushed to
 *	the screen when it's filled, or when a graphics attribute has
 *	changed for the current line.
 * Arguments:
 *	window - window the the selection set will be drawn on.
 *	gc - graphics context used for graphics attributes.
 * 	line - line to draw.
 *	segments - vector of lines in the draw buffer.
 *	nseg - number of segments in the buffer.
 *
 ****************************************************************************/

AmErrorCode
AmPolyLinePlotClass::drawPline( AmWindowClass 	*window,
			        AmGCClass     	*gc, 
			        AmPolylinePtr   pline,
			        AnvilBoolean	 	fHighlight)
{
    int 	x, y, index, ii;
    AmPointPtr	points;

    /*
     * For speed purposes, only set the line width when it changes, and
     *   draw the AmLast set of lines in the old line width.
     */

    if ( _pdLineWidth != NULL) {
	index = ( _indexLineWidth.getIndexFn)( pline, &_indexLineWidth);
	_dLineWidth = _pdLineWidth[ index];
    }
    if ( _pdLineWidthScale != NULL) {
	index = ( _indexLineWidth.getIndexFn)( pline, &_indexLineWidth);
	_dLineWidth = ScreenLineWidth( _pdLineWidthScale[ index]);
    }

    /* Only set the color when it changes */

    if (( colors != NULL) && ( fHighlight == FALSE)) {
	_col = colors[ ( colorIndex.getIndexFn)( pline, &colorIndex)];
    } else {
	_col = highlightColor;
    }

    if ( _dLineWidth != _dLineWidthLast) {
	gc->SetLineWidth( _dLineWidth);
        _dLineWidthLast = _dLineWidth;
    }
    if ( AmCompareColor( &_col, &_colLast) == FALSE) {
	gc->SetForegroundColor( &_col);
        _colLast = _col;
    }

    points = ( AmPointPtr) alloca( sizeof( AmPointRec)*pline->numPts);
    if ( points == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    for ( ii = 0; ii < pline->numPts; ii++) {
	GetScreenPt( pline->pts[ ii].x, pline->pts[ ii].y, &x, &y);
	points[ ii].x = x;
	points[ ii].y = y;
    }
    display->DrawLines( window, gc, points, pline->numPts);
    return( AmOK);
}

AmErrorCode
AmPolyLinePlotClass::drawPolylineVectorText( AmWindowClass	*window, 
					     AmGCClass		*gc, 
					     uchar_t		*plineBuffer, 
					     int		numPlines,
					     float		xmin,
					     float		ymin,
					     float		xmax,
					     float		ymax,
					     AnvilBoolean		*densityGrid,
					     int		densityRows,
					     int		densityCols)
{
    char		buffer[ FIELDLEN];
    int			textRow, textCol, x, y, index, nextIndex, textIndex,
			ii, plineSize, bestDist, bestIndex=0, candDist;
    AnvilBoolean		incFlag, foundFlag;
    uchar_t		*ptr;
    AmPolylinePtr	pline;
    float		angle, xscale, yscale;

    xscale = ( densityRows - 1)/( xmax - xmin);
    yscale = ( densityCols - 1)/( ymax - ymin);
    gc->SetForegroundColor( &fgColor);
    for ( ii = 0, ptr = plineBuffer; ii < numPlines; ii++, ptr += plineSize) {
	pline = ( AmPolylinePtr) ptr;

/* if the polyline bounding box overlaps the current viewport */

	if ( AmOverlapCoords( pline->xmin, pline->ymin, 
			      pline->xmax, pline->ymax, 
			      xmin, ymin, xmax, ymax)) {
	    index = pline->numPts/2;
	    incFlag = TRUE;
	    foundFlag = FALSE;
	    bestDist = MY_HUGEINT;
	    do {

/* start at the center of the polyline and search for a point which
   projects to a cell in the density matrix which hasn't been filled yet */

		textRow = (int)(( ymax - pline->pts[ index].y)*yscale);
		textCol = (int)(( pline->pts[ index].x - xmin)*xscale);
		if (( textRow >= 0) && ( textCol >= 0) &&
		    ( textRow < densityRows) && ( textCol < densityCols) &&
		    InCoords( pline->pts[ index].x, pline->pts[ index].y)) {
		    if ( !densityGrid[ textRow*densityCols + textCol]) {
			densityGrid[ textRow*densityCols + textCol] = TRUE;
			candDist = SQRDIST( textRow, textCol,
					    densityRows/2, densityCols/2);
			if ( candDist < bestDist) {
			    bestIndex = index;
			    bestDist = candDist;
			}
			foundFlag = TRUE;
		    }
		}

/* keep walking down the polyline to the end, then walk backwards from
   the center */

		if ( incFlag) {
		    index++;
		    if ( index >= pline->numPts) {
			index = pline->numPts/2;
			incFlag = FALSE;
		    }
		} else {
		    index--;
		    if ( index <= 0) {
			break;
		    }
		}
	    } while ( index >= 0);

/* if found, then figure the angle of the segment that it's over, 
   make sure it gets drawn upright, then draw the text */

	    if ( foundFlag) {
		AmPrintValue( colorIndex.units, 
			      ( colorIndex.fn)
				    ((( uchar_t *) pline) + colorIndex.offset,
				      colorIndex.clientData),
			      buffer);
		if ( index >= 0)  {
		    nextIndex = bestIndex - 1;
		} else {
		    nextIndex = bestIndex + 1;
		}
		angle = atan2( pline->pts[ bestIndex].y - 
				  pline->pts[ nextIndex].y,
			       pline->pts[ bestIndex].x - 
				  pline->pts[ nextIndex].x);
		if ( angle < 0.0) {
		    angle += 2*M_PI;
		}
		textIndex = bestIndex;
		if (( angle > M_PI/2) && ( angle < 3*M_PI/2)) {
		    angle += M_PI;
		    if ( angle > 2*M_PI) {
			angle -= 2*M_PI;
		    } 
		    textIndex = nextIndex;
		}
		if ( angle < 0.0) {
		    angle += 2*M_PI;
		}
		GetScreenPt( pline->pts[ nextIndex].x, 
			     pline->pts[ nextIndex].y, 
			     &x, &y);
		display->DrawRotatedText( window, gc, x, y, angle, buffer);
	    }
	}
	plineSize = dataInfo.size + pline->numPts*sizeof( AmCoordRec);
    }
    return( AmOK);
}

/* draw a vector of lines... */

AmErrorCode
AmPolyLinePlotClass::drawPolylineVector( AmWindowClass	*window, 
					 AmGCClass	*gc, 
					 uchar_t	*plineBuffer, 
					 int		numPlines,
					 AnvilBoolean	fHighlight,
					 float		xmin,
					 float		ymin,
					 float		xmax,
					 float		ymax)
{
    int			ii, plineSize;
    AmPolylinePtr	pline;
    uchar_t		*ptr;
    AmErrorCode		errCode;

    for ( ii = 0, ptr = plineBuffer; ii < numPlines; ii++, ptr += plineSize) {
	pline = ( AmPolylinePtr) ptr;
	if ( AmOverlapCoords( pline->xmin, pline->ymin, 
			      pline->xmax, pline->ymax, 
			      xmin, ymin, xmax, ymax)) {
	    errCode = drawPline( window, gc, pline, fHighlight);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	}
	plineSize = dataInfo.size + pline->numPts*sizeof( AmCoordRec);
    }
    return( AmOK);
}

/*************************************************************************
 * 
 * Description
 * 	Return the extents of a pointer cast to a line. 
 * Arguments:
 *	ptr - void pointer to a line to get the extents of.
 *	xmin, ymin, xmax, ymax - returns extents of line.
 *
 *************************************************************************/

void
AmPolyLinePlotClass::ObjectExtents( void  *ptr,
				    float *xmin,
				    float *ymin,
				    float *xmax,
				    float *ymax) 
{
    AmPolylinePtr	pline = ( AmPolylinePtr) ptr;

    *xmin = pline->xmin;
    *ymin = pline->ymin;
    *xmax = pline->xmax;
    *ymax = pline->ymax;
}

/*************************************************************************
 * 
 * Description:
 * 	return the list of settable fields in a line plot.  This is
 *	used in the map editor to dump the plot's fields into a file.
 *	the field name (and length, if it's a vector field) is written
 *	into the vector.  Each dumpFields member function calls the
 *	dumpFields() member function of the parent class.  This is
 *	also used for copying plots for magic windows.
 * Arguments:
 *	fieldLengths: vector of field names and lengths.
 *	allFields: TRUE when dumping to a file, FALSE when copying
 *	in memory.  This is an obsolete parameter that may be removed
 *	in the future.
 * Returns:
 *	number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolyLinePlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
				 AnvilBoolean	     allFields)
{
    int           numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

    /* Get the fields for the parent class ( plot). */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

    /* Get all the easy to specify fields for this class. */

    AmPolyLinePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);

    if ( _indexLineWidth.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNlineWidthBinvec;
	fieldLengths[ numStrings++].length = _indexLineWidth.length - 1;
    }

    return( numStrings);
}

/*************************************************************************
 * 
 * Description:
 * 	Return the field definitions for a line plot. This is used
 *	by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *	set the fields in the AmPolyLinePlotClass.  It returns a vector
 *	of field names followed by offsets, offset of the length
 *	of the field if it's a vector, and a storageType, which 
 *	is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *	dynamically allocated fields.
 * Arugments:
 *	fieldDefs: returned vector of field definitions.
 *	numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmPolyLinePlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				    int		*numDefs)
{
    static AmFieldDefRec LinePlotClassFieldDefintions[] = {
	{ AmNlineWidth, 
	    offsetof( AmPolyLinePlotClass, _dLineWidth) },
	{ AmNlineWidthClientData, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.clientData) },
	{ AmNlineWidthClientDataFreeFn, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.clientDataFreeFn) },
	{ AmNlineWidthFn, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.fn) },
	{ AmNlineWidthIntFn, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.AmIntFn) },
	{ AmNlineWidthLength,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.length) },
	{ AmNlineWidthList, offsetof( AmPolyLinePlotClass, _pdLineWidth),
            offsetof( AmPolyLinePlotClass, _indexLineWidth.length),
            AmDYNAMIC },
	{ AmNlineWidthMax, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.max) },
	{ AmNlineWidthMin, 
	    offsetof( AmPolyLinePlotClass, _indexLineWidth.min) },
	{ AmNlineWidthName,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.name), 
            0, AmDYNAMIC },
	{ AmNlineWidthOffset,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.offset) },
	{ AmNlineWidthStrings,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.strings),
            offsetof( AmPolyLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
	{ AmNlineWidthTable,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.table),
            offsetof( AmPolyLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
	{ AmNlineWidthType,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.dataType) },
	{ AmNlineWidthUnits,
            offsetof( AmPolyLinePlotClass, _indexLineWidth.units) },
	{ AmNscaleLineWidthList, 
            offsetof( AmPolyLinePlotClass, _pdLineWidthScale),
            offsetof( AmPolyLinePlotClass, _indexLineWidth.length), 
            AmDYNAMIC },
	{ AmNtextGridDensity, 
            offsetof( AmPolyLinePlotClass, _textGridDensity) }
    };
    *fieldDefs = LinePlotClassFieldDefintions;
    *numDefs = VECLEN( LinePlotClassFieldDefintions);
}
