#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/display.h>
#include <anvil/plot.h>
#include <anvil/tagstrs.h>
#include <anvil/log.h>
#include <anvil/fns.h>

/* default plot width and height */

const int AmPLOT_WIDTH = 200, AmPLOT_HEIGHT = 200;

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmPlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmPLOT) {
        return( TRUE);
    } else {
        return(AmDataGraphicClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      Empty Constructor
 *
 ****************************************************************************/

AmPlotClass::AmPlotClass() 
{
}

/****************************************************************************
 *
 * Description:
 *      initialize the fields for a plot.  The intializeFields()
 *      functions are created for each class, and each one calls the one
 *      for the "parent" class.  The AmScreenObjClass intializeFields()
 *      function is the topmost one that is called.  We can't use 
 *	two-parameter constructors, because since we use varargs 
 *	constructors, there is a naming ambiguity 
 * Arguments:
 *      windowArg - object is created on, and will be drawn on this window
 *      nameArg - string name used to identify plot for resources
 *
 ****************************************************************************/

void 
AmPlotClass::initializeFields( AmWindowClass	*windowArg,
			       const char	*nameArg)
{
    AmDataGraphicClass::initializeFields( windowArg, nameArg);
    rect.width = AmGetDefault( "plot.width", AmPLOT_WIDTH);
    rect.height = AmGetDefault( "plot.height", AmPLOT_HEIGHT);

/* set all the basic plot attributes */

    showStrings = FALSE;
    strings = NULL;
    colors = NULL;
    cmapName = NULL;
    nameFn = NULL;
    nameClientData = NULL;
    nameClientDataFreeFn = NULL;
    parent = NULL;

/* set the fields for the color index and string index */

    AmInitIndexData( &colorIndex);
    AmInitIndexData( &stringIndex);
}

/****************************************************************************
 *
 * Description:
 *	Plot class destructor.  Most of the information is destroyed by
 *	the AmDataGraphicClass destructor.  This frees the tile information,
 *	and the string and color index information
 *
 ****************************************************************************/

AmPlotClass::~AmPlotClass()
{
    int	ii;

/* close the pointers to each of the tile files */

    if (( dataInfo.sourceType == AmPYRAMID_SOURCE) && ( tileHeader != NULL)) {
	if ( tileHeader->fds != NULL) {
	    for ( ii = 0; ii < tileHeader->numFiles; ii++) {
		close( tileHeader->fds[ ii]);
		free( tileHeader->files[ ii]);
	    }
	    free( tileHeader->fds);
	    free( tileHeader->files);
	}
	free( tileHeader);
    }
    if (( nameClientDataFreeFn != NULL) && ( nameClientData != NULL)) {
	( nameClientDataFreeFn)( nameClientData);
    }
    if ( strings != NULL) {
	AmFreeStrVec( strings);
    }

/* destroy the color and string indexes */

    AmDestroyIndexData( &colorIndex);
    AmDestroyIndexData( &stringIndex);
}

/****************************************************************************
 *
 * Description:
 * 	check values for a plot.  Each check values function calls the
 *	check values function for the parent class and is used to set
 *	the object's IsValid() flag
 * Returns:
 *	TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPlotClass::checkValues()
{
    if ( !AmDataGraphicClass::checkValues()) {
	return( FALSE);
    }
    if (( colorIndex.offset >= dataInfo.size) && ( colors != NULL)) {
	AmAddResourceToLog( "colorIndexOffsetMustLessThanSize", 
			  colorIndex.offset, dataInfo.size);
	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	retreive values for a plot.  Each getFields() function calls
 *	getFields() for the parent class, then most of the fields are
 *	retrived using the GetFieldByDefinitions() function.  The
 *	bin vectors are retrieved by hand, because their length is 
 *	1 less than the length value.
 * Arguments:
 *	tagValues - vector of tag-values to retrieve.  The values are
 *	pointers to the results.  The array is "scrunched" in place
 *	as each value is retreived.  Any unprocessed values are returned
 *	in place in the array.
 *	numValues - number of values in the array.
 * Returns:
 *	number of unprocessed values.
 *
 ****************************************************************************/

int 
AmPlotClass::getFields( AmTagValuePtr	tagValues,
		        int		numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, ii, numUnusedFields;

/* get the fields for the parent class ( coordinate system) */

    numValues = AmDataGraphicClass::getFields( tagValues, numValues);

/* get all the easy to specify fields for this class */

    AmPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* get all of the harder to specify fields */

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if (( AmMatchTag( tagValues[ ii].tag, AmNcolorsBinvec)) ||
	    ( AmMatchTag( tagValues[ ii].tag, AmNbinvec))) {
	    if ( colorIndex.binvec != NULL) {
		* ( void **) tagValues[ ii].value = 
		    AmCopyVec( colorIndex.binvec,
			    ( colorIndex.length - 1)*sizeof( float));
	    } else {
		* ( void **) tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }

/* return the number of unset fields, which are scrunched in TagValue array */

    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *	Dump the field names used in a plot.  dumpFields() is called
 *	recursively by each object and its parent.  The fieldLengths
 *	array returns the name of each field, and if it's a vector 
 *	field, the vector's length.
 * Arguments:
 *	fieldLengths - vector of tags and vector lengths.
 *	allFields - an obsolete tag which was used to differentiate
 *	AmBetween dumping to a file or copying the object.
 * Returns:
 *	The number of fields dumped 
 *
 ***************************************************************************/

int
AmPlotClass::dumpFields( AmFieldLengthPtr	fieldLengths,
			 AnvilBoolean		allFields)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, numStrings;

/* Get the fields for the parent class. */

    numStrings = AmDataGraphicClass::dumpFields( fieldLengths, allFields);

/* Get the easy to specify fields for this class. */

    AmPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
                                          fieldDefs, numFieldDefs);

/* Get all of the harder to specify fields. */

    if ( colorIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNcolorsBinvec;
	fieldLengths[ numStrings++].length = colorIndex.length - 1;
    }
    if ( stringIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNstringsBinvec;
	fieldLengths[ numStrings++].length = stringIndex.length - 1;
    }

/* Return the number of strings put into the table. */

    return( numStrings);
}

/****************************************************************************
 *
 * Description:
 *	Base read function, uses the fields in the data info structure
 *	to read the data.  If it's pyramid format, it reads the tile
 *	header, otherwise it calls the default getDataFrom..() from
 *	AmDataGraphicClass
 * Returns:
 *	AmOK if everything was read OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmPlotClass::getDataFromDescription()
{

/* if it's a pyramid tile file, then open the file descriptor, and read the
   tile information */

    if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	return( AmReadTileHeader( dataInfo.filename, &tileHeader,
				&dataInfo.fd));
    } else {
	return( AmDataGraphicClass::getDataFromDescription());
    }
}

/****************************************************************************
 *
 * Description:
 *	Read a "palette" file, which maps integer pixel values to RGB colors.
 * Arguments:
 *	paletteFile: file containing integer values and RGB values in the
 *	form pixel R G B.
 *	table: resulting vector of pixels.
 *	colors: resulting vector of colors.
 *	numColors: length of table and colors vectors.
 * Returns:
 *	AmOK if things were OK, system error otherwise.
 *
 ****************************************************************************/

static AmErrorCode 
readPalette( char            *paletteFile,
	     int             **table,
	     AmColorCellPtr  *colors,
	     int             *numColors)
{
    FILE        *fp;
    int         ii, pixel, red, green, blue;

/* get the number of lines in the file, then try to open it */

    *numColors = AmNumLines( paletteFile);
    if (( fp = fopen( paletteFile, "r")) == NULL) {
        DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* allocate the color and pixel arrays */

    if (( *colors = ( AmColorCellPtr)
                malloc( *numColors*sizeof( AmColorCellRec))) == NULL) {
	fclose( fp);
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( *table = ( int *) malloc( *numColors*sizeof( int))) == NULL) {
	fclose( fp);
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* read the values */

    for ( ii = 0; ii < *numColors; ii++) {
        if ( fscanf( fp, "%d %d %d %d", &pixel, &red, &green, &blue) != 4) {
            break;
        }
        ( *table)[ ii] = pixel;
        ( *colors)[ ii].red = red;
        ( *colors)[ ii].green = green;
        ( *colors)[ ii].blue = blue;
    }
    if ( ii < *numColors) {
        *numColors = ii;
    }

/* close the file and return AmOK */

    fclose( fp);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	set the fields in the plot from the tag-value argument list.  Each
 *	class calls setFields() for the parent class, then uses 
 *	SetFieldByDefinitions() to set its own members.  Members which
 *	can't be processed using the SetFieldsBy...() are processed by
 *	hand, such as the palette and bin vector.
 * Arguments:
 *      tagValues: tag-value pairs to be parsed (this array is reduced in
 *                 place, with the parsed arguments removed).
 *      numValues: length of the tagValues array.
 * Return:
 *      number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int 
AmPlotClass::setFields( AmTagValuePtr	tagValues,
		        int		numValues)
{
    int			ii, numUnusedFields, numFieldDefs;
    AmFieldDefPtr		fieldDefs;
    AmColorCellPtr	paletteColors;

    SetValid( TRUE);
    
/* get the fields for the parent class ( Data graphic object), 
   and set the color index length.  Note it has to be set before colorsBinvec 
   and colorsTable, so they get the proper length */

    numValues = AmDataGraphicClass::setFields( tagValues, numValues);
    colorIndex.length = numColors;
    AmPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);

/* tiles, palette, and colorsBinvec are handled explicitly by AmPlotClass */

    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNpalette)) {
	    if ( readPalette( ( char *) tagValues[ ii].value,
			      &colorIndex.table, &paletteColors,
			      &colorIndex.length) != AmOK) {
		AmAddResourceToLog( "readPaletteFailed", 
				    ( char *) tagValues[ ii].value);
		SetValid( FALSE);
	    } else {
		numColors = colorIndex.length;
		if ( colors != NULL) {
		    free( colors);
		} 
		colors = paletteColors;
	    }

/* copy the binvecs by hand, because they're one shorter than the colors */

	} else if (( AmMatchTag( tagValues[ ii].tag, AmNcolorsBinvec)) ||
		   ( AmMatchTag( tagValues[ ii].tag, AmNbinvec))) {
	    AmSetBinvec( &colorIndex.binvec, colorIndex.length, 
			 &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNstringsBinvec)) {
	    AmSetBinvec( &stringIndex.binvec, stringIndex.length, 
			 &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmPlotClass::updateFields()
{
    AmDataGraphicClass::updateFields();

/* set the index data client structures for the color function and the
   string function */

    if ( IsValid()) {
	if ( strings != NULL) {
	    if ( updateIndexData( &stringIndex, "strings") != AmOK) {
		AmAddResourceToLog( "failedToInitializeStringIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
    } 
}

/* retrieve the field definitions for the plot class. THE LIST MUST BE
   SORTED BY ALPHABETICAL ORDER ON THE FIELD NAME.  This is used by
   SetFieldByDefinitions() */ 

void 
AmPlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDef,
				  int		*numFieldDef)
{
    static AmFieldDefRec PlotClassFieldDefinitions[] = {
	{ AmNcolorsClientData, offsetof( AmPlotClass, colorIndex.clientData) },
	{ AmNcolorsClientDataFreeFn, 
		offsetof( AmPlotClass, colorIndex.clientDataFreeFn) },
	{ AmNcolorsFn, offsetof( AmPlotClass, colorIndex.fn) },
	{ AmNcolorsIntFn, offsetof( AmPlotClass, colorIndex.AmIntFn) },
	{ AmNcolorsMax, offsetof( AmPlotClass, colorIndex.max) },
	{ AmNcolorsMin, offsetof( AmPlotClass, colorIndex.min) },
	{ AmNcolorsName, offsetof( AmPlotClass, colorIndex.name), 
		0, AmDYNAMIC },
	{ AmNcolorsOffset, offsetof( AmPlotClass, colorIndex.offset) },
	{ AmNcolorsStrings, offsetof( AmPlotClass, colorIndex.strings),
			    offsetof( AmPlotClass, colorIndex.length), 
			    AmDYNAMIC },
	{ AmNcolorsTable, offsetof( AmPlotClass, colorIndex.table),
			  offsetof( AmPlotClass, colorIndex.length), 
			  AmDYNAMIC },
	{ AmNcolorsType, offsetof( AmPlotClass, colorIndex.dataType) },
	{ AmNcolorsUnits, offsetof( AmPlotClass, colorIndex.units) },
	{ AmNmax, offsetof( AmPlotClass, colorIndex.max) },
	{ AmNmaxClip, offsetof( AmPlotClass, colorIndex.maxClip) },
	{ AmNmin, offsetof( AmPlotClass, colorIndex.min) },
	{ AmNminClip, offsetof( AmPlotClass, colorIndex.minClip) },
	{ AmNnameClientDataFreeFn, 
		offsetof( AmPlotClass, nameClientDataFreeFn) },
	{ AmNnameFn, offsetof( AmPlotClass, nameFn) },
	{ AmNoffset, offsetof( AmPlotClass, colorIndex.offset) },
	{ AmNstringList, offsetof( AmPlotClass, strings), 
			 offsetof( AmPlotClass, stringIndex.length), 
			 AmDYNAMIC },
	{ AmNstringsFn, offsetof( AmPlotClass, stringIndex.fn) },
	{ AmNstringsIntFn, offsetof( AmPlotClass, stringIndex.AmIntFn) },
	{ AmNstringsLength, offsetof( AmPlotClass, stringIndex.length) },
	{ AmNstringsMax, offsetof( AmPlotClass, stringIndex.max) },
	{ AmNstringsMin, offsetof( AmPlotClass, stringIndex.min) },
	{ AmNstringsName, offsetof( AmPlotClass, stringIndex.name), 
		0, AmDYNAMIC },
	{ AmNstringsOffset, offsetof( AmPlotClass, stringIndex.offset) },
	{ AmNstringsTable, offsetof( AmPlotClass, stringIndex.table),
			    offsetof( AmPlotClass, stringIndex.length), 
			    AmDYNAMIC },
	{ AmNstringsType, offsetof( AmPlotClass, stringIndex.dataType) },
	{ AmNtable, offsetof( AmPlotClass, colorIndex.table), 
		    offsetof( AmPlotClass, colorIndex.length), AmDYNAMIC },
	{ AmNzFn, offsetof( AmPlotClass, colorIndex.fn) },
	{ AmNzunits, offsetof( AmPlotClass, colorIndex.units) }
    };
    *fieldDef = PlotClassFieldDefinitions;
    *numFieldDef = VECLEN( PlotClassFieldDefinitions);
}
