#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* class member functions for a polar graph. */

#if defined(_WIN32) || defined (__BSD44__)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <stddef.h>
#include <stdarg.h>
#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/polargraph.h>
#include <anvil/datafmt.h>
#include <anvil/defaults.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPolarGraphClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmPOLAR_GRAPH) {
       return( TRUE); 
    } else {
       return( AmGraphClass::IsType( type)); 
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a polar graph.  call the
 *      AmGraphClass initialize fieldsm then set the polar graph specific fields
 * Arguments:
 *      windowArg - window that the polar graph will be drawn on.
 *      nameArg - string name of the polar graph, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmPolarGraphClass::initializeFields( AmWindowClass	*windowArg,
				   const char	*nameArg)
{
    AmGraphClass::initializeFields( windowArg, nameArg);

/* default values take x,y,z tuples (or r,t,z).  */

    AmInitIndexData( &angleIndex);
    objectType = AmPOLAR_GRAPH;
    fixedExtents = TRUE;		/* extents only used for zoom ref */
    navigational = FALSE;
    clockwise = FALSE;

    _dLineWidth = 0.0;

    clipAtMin = TRUE;
}

/****************************************************************************
 *
 * Description:
 *	Check the values for a polar graph.  These are suppoesd to be
 *	implemented for a graph/plot objects, but they haven't all be
 *	done yet. 
 * Returns:
 *	TRUE if the graph has been initialized properly, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPolarGraphClass::checkValues()
{
    if ( !AmGraphClass::checkValues()) {
    	return( FALSE);
    } else {
    	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmPolarGraphClass constructor.  Creates a polar graph using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the polar graph.
 * Arguments:
 *      windowArg - window that the polar graph will be drawn on.
 *      nameArg - string name of the polar graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolarGraphClass::AmPolarGraphClass( AmWindowClass *windowArg,
				  const char   *nameArg, 
				  ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( checkValues());
}

/****************************************************************************
 *
 * Description:
 *      AmPolarGraphClass constructor.  Creates a polar graph as part of an
 *      overlay.  The proprerties for the polar graph are set by a
 *      NULL-terminated vector of tag-value pairs.  This doesn't really
 *	apply, since polar graphs have different coordinate systems from
 *	the rest of the plots and graphs.
 * Arguments:
 *      overlayArg - overlay that the bar graph will be added to.
 *      nameArg - string name of the bar graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolarGraphClass::AmPolarGraphClass( AmOverlayClass  *overlayArg,
				  const char    *nameArg,
				  ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* we want all of the parameters which are set for the overlay plot to be
   set for the polar plot as well. Set the overlay plot back pointer to 
   reference in the draw and destroy routines 
   color function has to be set explicitly. Default doesn't work */

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( checkValues());
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmPolarGraphClass destructor.
 *
 ****************************************************************************/

AmPolarGraphClass::~AmPolarGraphClass()
{
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a polar graph, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolarGraphClass::setFields( AmTagValuePtr	tagValues,
			    int		numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;

    numValues = AmGraphClass::setFields( tagValues, numValues);
    AmPolarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmPolarGraphClass::updateFields()
{
    AmGraphClass::updateFields();
    if ( IsValid()) {
	if (( angleIndex.fn != NULL) || ( angleIndex.dataType != AmNO_TYPE)) {
	    if ( updateIndexData( &angleIndex, "angle") != AmOK) {
		AmAddResourceToLog( "failedToInitializeAngleIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a polar graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are gotten, then
 *      the other fields are pulled using the standard GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolarGraphClass::getFields( AmTagValuePtr	tagValues,
			    int		numValues)
{
    int		numFieldDefs;
    AmFieldDefPtr	fieldDefs;
 
    numValues = AmGraphClass::getFields( tagValues, numValues);
    AmPolarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 * 	private function to translate a block of angle, radius points 
 *	into a point array 
 * Arguments:
 *	polarBuffer - source data to convert.
 *	Points - points in screen coordinate system.
 *	numel - number of elements in the polar buffer.
 *
 ****************************************************************************/

void 
AmPolarGraphClass::polarPoints( uchar_t		*polarBuffer, 
			        AmPointPtr	points, 
			        int		numel)
{
    int		jj;
    uchar_t	*ptr;
    float	angle, radius, radscale;

    for ( jj = 0, ptr = polarBuffer; 
	  jj < numel; 
	  jj++, ptr += dataInfo.size, points++) {

/* calculate the angle and radius */

	if ( angleIndex.fn != NULL) {
	    angle = DEGTORAD(( angleIndex.fn)
				( ptr + angleIndex.offset, 
				  angleIndex.clientData));
	} else {
	    angle = ( M_PI*2*jj)/numel;
	}
	if ( navigational == TRUE) {
	    angle = M_PI/2 - angle;
	} else if ( clockwise == TRUE) {
	    angle = -angle;
	}
	radius = ( heightIndex.fn) ( ptr + heightIndex.offset, 
				     heightIndex.clientData);

/* scale it to the screen and go to rectangular coordinates.  
   If we clip at min, then all negative values go to zero */

	radscale = ( radius - heightIndex.min)/
			( heightIndex.max - heightIndex.min);
	if ( clipAtMin && ( radscale < 0.0)) {
	    radscale = 0.0;
	}
	points->x = (short int)(rect.x + rect.width/2 + cos( angle)*radscale*rect.width/2);
	points->y = (short int)(rect.y + rect.height/2 - sin( angle)*radscale*rect.height/2);
    }
}

/****************************************************************************
 *
 * Description:
 *      Draw the polar graph to the screen.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK
 *
 ****************************************************************************/

AmErrorCode 
AmPolarGraphClass::draw( AmWindowClass		*window,
		       AmGCClass			*gc, 
		       AmRectangleRec const 	*exposeRects,
		       int			nRects)
{
    int		ii, polarBufSize;
    uchar_t	*ptr, *polarBuffer=NULL;
    AmPointPtr	points;

/* if this object is invalid only draw the outpolar */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
	return( AmOK);
    }

/* if there's no overlay data, then we clear the area that we're going to
   draw to.  We have to use FillRectangle(), because we have to pay 
   attention to the gc's clip rectangle which is set by the parent */

    if (( parent == NULL) && opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, exposeRects[ ii].x, 
						exposeRects[ ii].y,
						exposeRects[ ii].width, 
						exposeRects[ ii].height);
	}
    }

/* set the polar width and polar color, which may be overridden by functions
   in the loop */

    gc->SetLineWidth( _dLineWidth);
    gc->SetLineCap( AmCAP_ROUND);
    gc->SetForegroundColor( &fgColor);
	
/* cast each element to a data segment and draw it.  Buffer the polar segments 
   on width and color to try to speed things up */

    if ( dataInfo.sourceType == AmREAD_SOURCE) {

/* allocate a buffer to store all the data on the stack.  People don't have 
   billion point polar plots... */
	
	polarBufSize = dataInfo.size*AmDataLength( &dataInfo);
	if (( polarBuffer = ( uchar_t *) alloca( polarBufSize)) == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}
	if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	if ( read( dataInfo.fd, polarBuffer, polarBufSize) < polarBufSize) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}

/* otherwise just set the pointer */

    } else if (( dataInfo.sourceType == AmFILE_SOURCE) ||
    	       ( dataInfo.sourceType == AmDIRECT_SOURCE)) {
	polarBuffer = dataInfo.data;

    } else {
	AN_DEBUG (1,("  in AmPolarGraphClass::draw ... invalid dataInfo.sourceType ??? (%d) \n",
				dataInfo.sourceType ));
	DO_RETURN_ERROR( AmNULL_ARG_PTR);
    }

/* if there are two dimensions, we're plotting a set of polar plots, otherwise,
   just a single one.  If there's a colormap, then use the colors for each of
   the polar plots */

    if ( dataInfo.numdims == 2) {
	points = ( AmPointPtr) alloca( sizeof( AmPointRec)*dataInfo.dims[ 1]);
	for ( ii = 0, ptr = polarBuffer; 
	      ii < dataInfo.dims[ 0]; 
	      ii++, ptr += dataInfo.dims[ 1]*dataInfo.size) {
	    if ( colors != NULL) {
		gc->SetForegroundColor( &colors[ ii % numColors]);
	    }
	    polarPoints( ptr, points, dataInfo.dims[ 1]);
	    display->DrawLines( window, gc, points, dataInfo.dims[ 1]);
	}
    } else {
	points = ( AmPointPtr) alloca( sizeof( AmPointRec)*dataInfo.dims[ 0]);
	polarPoints( polarBuffer, points, dataInfo.dims[ 0]);
	display->DrawLines( window, gc, points, dataInfo.dims[ 0]);
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a polar graph.  This is
 *      used in the map editor to dump the graph's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolarGraphClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			     AnvilBoolean	       allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( plot) */

    numStrings = AmGraphClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmPolarGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
				     fieldDefs, numFieldDefs));
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a bar graph. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmOverlayClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  THE LIST MUST BE SORTED BY
 *      ALPHABETICAL ORDER ON THE FIELD NAME.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmPolarGraphClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				      int		*numDefs)
{
    static AmFieldDefRec PolarGraphClassFieldDefintions[] = {
        { AmNangleClientData, offsetof( AmPolarGraphClass, 
					angleIndex.clientData) },
        { AmNangleClientDataFreeFn, offsetof( AmPolarGraphClass, 
					      angleIndex.clientDataFreeFn) },
        { AmNangleFn, offsetof( AmPolarGraphClass, angleIndex.fn) },
        { AmNangleMax, offsetof( AmPolarGraphClass, angleIndex.max) },
        { AmNangleMin, offsetof( AmPolarGraphClass, angleIndex.min) },
        { AmNangleType, offsetof( AmPolarGraphClass, angleIndex.dataType) },
        { AmNangleUnits, offsetof( AmPolarGraphClass, angleIndex.units) },
	{ AmNclipAtMin, offsetof( AmPolarGraphClass, clipAtMin) },
	{ AmNclockwise, offsetof( AmPolarGraphClass, clockwise) },
	{ AmNnavigational, offsetof( AmPolarGraphClass, navigational) },
	{ AmNlineWidth, offsetof( AmPolarGraphClass, _dLineWidth) },
    };
    *fieldDefs = PolarGraphClassFieldDefintions;
    *numDefs = VECLEN( PolarGraphClassFieldDefintions);
}
