#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <malloc.h>
#include <string.h>
#if !defined(_WIN32)  && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdarg.h>
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <anvil/strlib.h>
#include <anvil/plot.h>
#include <anvil/macros.h>
#include <anvil/resource.h>
#include <anvil/mem_funcs.h>
#include <anvil/overlay.h>
#include <anvil/polygonplot.h>
#include <anvil/hatchpoly.h>
#include <anvil/datafmt.h>
#include <anvil/compression.h>
#include <anvil/log.h>
#include <anvil/files.h>

static const int MAXTILES = 1024;

/****************************************************************************
 *
 * Description:
 * 	is a point below a line? 
 * Arguments:
 *	a, b: endpoints of the line to test against.
 *	x, y: point to test.
 * Returns:
 *	TRUE if the point is below the line, FALSE otherwise.
 *
 ****************************************************************************/
 
static AnvilBoolean 
tstLinePt( AmCoordPtr       a,
	   AmCoordPtr       b,
	   float          x,
	   float          y)
{
    float      m;

    if ((( a->x < x) && ( x <= b->x)) || (( b->x < x) && ( x <= a->x))) {
        m = ( b->y - a->y)/( b->x - a->x);
        return( m*( x - a->x) + a->y > y);
    }
    return( FALSE);
}

/****************************************************************************
 *
 * Description:
 * 	is a point inside of a polygon? 
 * Arguments:
 *	polygon - polygon to test against.
 *	x,y - x,y point to test against the polygon.
 * Returns:
 *	TRUE if the point is inside the polygon, FALSE otherwise.
 *
 ****************************************************************************/
 
static AnvilBoolean 
inPolygon( AmPolygonPtr     polygon,
	   float          x,
	   float          y)
{
    int ii, icount = 0;
 
    for ( ii = 0; ii < polygon->numPts - 1; ii++) {
        if ( tstLinePt( &polygon->pts[ ii], &polygon->pts[ ii + 1], x, y)) {
            icount++;
        }
    }
    if ( tstLinePt( &polygon->pts[ 0], &polygon->pts[ polygon->numPts - 1],
                    x, y)) {
        icount++;
    }
    return( ODD( icount));
}

/****************************************************************************
 *
 * Description:
 * 	find the extents of a polygon, and set them in the polygon.  This
 *	also sets the "minimum bounding rectangle has been set" flag.  If
 *	the flag's already been set, this function does nothing.
 * Arguments:
 *	poly - polygon to find the extents of.
 *
 ****************************************************************************/

static void 
setPolygonExtents( AmPolygonPtr	poly)
{
    int	ii;

    if ( !( poly->flags & AM_MBR_FLAG)) {
	poly->xmin = MY_HUGE;
	poly->ymin = MY_HUGE;
	poly->xmax = -MY_HUGE;
	poly->ymax = -MY_HUGE;

	for ( ii = 0; ii < poly->numPts; ii++) {
	    if ( poly->pts[ ii].x < poly->xmin) {
		poly->xmin = poly->pts[ ii].x;
	    }
	    if ( poly->pts[ ii].y < poly->ymin) {
		poly->ymin = poly->pts[ ii].y;
	    }
	    if ( poly->pts[ ii].x > poly->xmax) {
		poly->xmax = poly->pts[ ii].x;
	    }
	    if ( poly->pts[ ii].y > poly->ymax) {
		poly->ymax = poly->pts[ ii].y;
	    }
	}
	poly->flags |= AM_MBR_FLAG;
    } 
}

/****************************************************************************
 *
 * Description:
 *  	draw an individual polygon.  This function calculates the foreground
 *	color, hatch angle and hatch width, and solid or open drawing.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *	polygon - polygon to draw to the screen.
 *	screenPts - buffer to hold the array of screen points to draw. 
 *
 ****************************************************************************/

void 
AmPolygonPlotClass::drawPolygon( AmWindowClass	*window,
			         AmGCClass	*gc,
			         AmPolygonPtr	polygon,
			         AmPointPtr	screenPts)
{
    int			numpts, scrX, scrY, scrXmin, scrYmin, scrXmax, scrYmax,
			ascent, descent, width, height, index, x0, y0, x1, y1;
    float		pxmin, pxmax, pymin, pymax;
    char		*string;
    AmRectangleRec	boundRect;
    AnvilBoolean		drawFlag;

    numpts = polygon->numPts;
    scalePoints( polygon->pts, &numpts, screenPts);
    screenPts[ numpts] = screenPts[ 0];		/* close it */

/* if there is a hatch and angle function, then hatch the polygon */

    if ( hatched) {
	drawFlag = TRUE;
	if ( hatchAngleList != NULL) {
	    index = ( hatchAngleIndex.getIndexFn)( polygon, &hatchAngleIndex);
	    if (( index < 0) && !hatchAngleIndex.minClip) {
		index = 0;
	    } else if (( index >= hatchAngleIndex.length) && 
		       !hatchAngleIndex.maxClip) {
		index = hatchAngleIndex.length - 1;
	    }
	    if (( index >= 0) && ( index < hatchAngleIndex.length)) {
		hatchAngle = hatchAngleList[ index];
	    } else {
		drawFlag = FALSE;
	    }
	}
	if ( hatchWidthList != NULL) {
	    index = ( hatchWidthIndex.getIndexFn)( polygon, &hatchWidthIndex);
	    if (( index < 0) && !hatchWidthIndex.minClip) {
		index = 0;
	    } else if (( index >= hatchWidthIndex.length) && 
		       !hatchWidthIndex.maxClip) {
		index = hatchWidthIndex.length - 1;
	    }
	    if (( index >= 0) && ( index < hatchWidthIndex.length)) {
		hatchWidth = hatchWidthList[ index];
	    } else {
		drawFlag = FALSE;
	    }
	}
	if ( drawFlag) {

/* draw the bounds of the polygon and hatch it */
      
	    display->DrawLines( window, gc, screenPts, numpts + 1);

/* note that the integer spacing must be greater than 1.  The bounding
   rectangle is calculated, because the polygon is hatched by clipping
   with a region */

	    GetScreenPt( polygon->xmin, polygon->ymax, &x0, &y0);
	    GetScreenPt( polygon->xmax, polygon->ymin, &x1, &y1);
	    boundRect.x = x0;
	    boundRect.y = y0;
	    boundRect.width = x1 - x0;
	    boundRect.height = y1 - y0;
	    AmHatchPoly( display, window, gc, hatchGC, screenPts, numpts, 
			 &boundRect, DEGTORAD( hatchAngle), 
			 (int)MAX( ScreenLineWidth( hatchWidth), 1));
	}
    } else if ( solid) {

/* fill the polygon */

	display->FillPolygon( window, gc, screenPts, numpts + 1);
    } else {

/* just draw the outline of the polygon */

	display->DrawLines( window, gc, screenPts, numpts + 1);
    }

/* if there's a string function, draw the string in the center of the 
   polygon's extents if possible.  Use either the string function or the
   name function if it's valid */

    if ((( strings != NULL) || ( nameFn != NULL)) && showStrings) {
	pxmin = polygon->xmin;
	pymin = polygon->ymin;
	pxmax = polygon->xmax;
	pymax = polygon->ymax;
	if ( inPolygon( polygon, ( pxmax + pxmin)/2, ( pymax + pymin)/2)) {
	    GetScreenPt(( pxmax + pxmin)/2, ( pymax + pymin)/2, 
			&scrX, &scrY);
	    GetScreenPt( pxmin, pymin, &scrXmin, &scrYmax);
	    GetScreenPt( pxmax, pymax, &scrXmax, &scrYmin);
	    if ( nameFn != NULL) {
		string = ( nameFn)( polygon, nameClientData);
	    } else {
		index = ( stringIndex.getIndexFn)( polygon, &stringIndex);
		string = strings[ index];
	    }

/* check to make sure that the text isn't bigger than the polygon */

	    display->ExtentsText( gc, string, &width, &ascent, &descent);
	    height = ascent + descent;
	    if (( width < scrXmax - scrXmin) &&
		( height < scrYmax - scrYmin)) {
		display->DrawText( window, gc, scrX - width/2, 
				   scrY - descent + height/2, string);
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	draw a block of polygons.  Check to make sure the extents have been
 *	set for the polygon, or set them here.  If the polygon vertex list
 *	is longer than the buffer allocated for drawing points, rubberband
 *	the buffer to contain them.  Call the drawPolygon funtion to
 *	actually draw the polygon.
 * Arguments:
 *	display - display to draw to. (this is unneeded).
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *	screenPts - buffer to hold the array of screen points to draw. 
 *	this is used to prevent multiple allocations, and is retained
 *	then freed by the parent draw function.
 *	polyBuffer - vector containing a block of polygons
 *	endPtr - marker end pointer to the vector containing the polygon block.
 *	numPolys - number of polygons in the buffer.
 *	xmin, ymin, xmax, ymax: bounding rectangle to draw to in the 
 *	plot's coordinate system.
 *	fHighlight - controls whether the polygons are drawn in their
 *	normal color or the highlight color.
 * Returns:
 *	AmOK if everything's copacetic, AmALLOCA_FAILED if there's no
 *	more RAM to be had.
 *
 ****************************************************************************/

AmErrorCode 
AmPolygonPlotClass::drawPolygonVector( AmDisplayClass	*display,
				       AmWindowClass	*window,
				       AmGCClass	*gc,
				       uchar_t		*polyBuffer,
				       uchar_t		**endPtr,
				       int		numPolys,
				       AnvilBoolean		fHighlight)
{
    int			ii, polySize, maxPts, index;
    AmPolygonPtr	polygon;
    AmPointPtr		screenPts;
    uchar_t		*ptr;
    char		patternPath[ PATHLEN], *path;
    AnvilBoolean		drawFlag;

/* if they're highlighted, then draw in the highlight, otherwise set the
   foreground color */

    if ( fHighlight) {
	gc->SetForegroundColor( &highlightColor);
    } else {
	gc->SetForegroundColor( &fgColor);
    }

/* set the fill style appropriately */

    if ( hatchPattern != NULL) {
	sprintf( patternPath, "patterns/%s", hatchPattern);
	path = AmGetEnvFile( CONFIG_ENV, patternPath);
	if ( path != NULL) {
	    gc->SetFillPattern( path);
	}
	gc->SetFillStyle( AmFILLSTYLE_STIPPLED);
    } else if ( hatchPatternList != NULL) {
	gc->SetFillStyle( AmFILLSTYLE_STIPPLED);
    } else {
	gc->SetFillStyle( AmFILLSTYLE_SOLID);
    }

/* find the max # of points in a polygon */

    maxPts = 0; 
    for ( ii = 0, ptr = polyBuffer; ii < numPolys; ii++, ptr += polySize) {
	polygon = ( AmPolygonPtr) ptr;
	if ( polygon->numPts > maxPts) {
	    maxPts = polygon->numPts;
	}
	polySize = dataInfo.size + polygon->numPts*sizeof( AmCoordRec);
    }
    screenPts = ( AmPointPtr) alloca( sizeof( AmPointRec)*( maxPts + 1));
    if ( screenPts == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }

/* draw the polygons in the buffer, stepping by the polygon size */ 

    for ( ii = 0, ptr = polyBuffer; ii < numPolys; ii++, ptr += polySize) {
	polygon = ( AmPolygonPtr) ptr;
	if ( !( polygon->flags & AM_MBR_FLAG)) {
	    setPolygonExtents( polygon);
	}
	if ( OverlapRectCoords( polygon->xmin, polygon->ymin,
				polygon->xmax, polygon->ymax)) {

/* set the foreground color if the pixel function is set */

	    drawFlag = TRUE;
	    if (( colors != NULL) && !fHighlight) {
		index = ( colorIndex.getIndexFn)( polygon, &colorIndex);
		if (( index < 0) && !colorIndex.minClip) {
		    index = 0;
		} else if (( index >= colorIndex.length) && 
			   !colorIndex.maxClip) {
		    index = colorIndex.length - 1;
		}
		if (( index >= 0) && ( index < colorIndex.length)) {
		    gc->SetForegroundColor( &colors[ index]);
		} else {
		    drawFlag = FALSE;
		}
	    }

/* set the stipple if the hatch pattern list is not NULL */

	    if ( hatchPatternList != NULL) {
		index = ( hatchPatternIndex.getIndexFn)
				( polygon, &hatchPatternIndex);
		if (( index < 0) && !hatchPatternIndex.minClip) {
		    index = 0;
		} else if (( index >= hatchPatternIndex.length) && 
			   !hatchPatternIndex.maxClip) {
		    index = hatchPatternIndex.length - 1;
		}
		if (( index >= 0) && ( index < hatchPatternIndex.length)) {
		    sprintf( patternPath, "patterns/%s", 
			     hatchPatternList[ index]);
		    path = AmGetEnvFile( CONFIG_ENV, patternPath);
		    if ( path != NULL) {
			gc->SetFillPattern( path);
		    }
		} else {
		    drawFlag = FALSE;
		}
	    }
	    if ( drawFlag) {
		drawPolygon( window, gc, polygon, screenPts);
	    }
	}
	polySize = dataInfo.size + polygon->numPts*sizeof( AmCoordRec);
    }
    *endPtr = ptr;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Take the vector of vertices in a polygon and scale them to
 *	screen coordinates.  If there are any duplicate points, remove
 *	them.
 * Arguments:
 *	points - vector of points in the polygon plot's coordinate system.
 *	numpts - number of points in the polygon.  This will decrease if
 *	there are any points eliminated.
 *  	screenpts - vector of points to draw to the screen
 *
 ****************************************************************************/

void 
AmPolygonPlotClass::scalePoints( AmCoordPtr	points,
			         int		*numpts,
			         AmPointPtr	screenpts)
{
    int	scrX, scrY, ii, count;

/* while scaling the points to the screen, get rid of the duplicate points,
   which can cause all sorts of unpleasantness */

    count = 0;
    for ( ii = 0; ii < *numpts; ii++) {
	GetScreenPt( points[ ii].x, points[ ii].y, &scrX, &scrY);
	if ( ii > 0) {
	    if (( screenpts[ count - 1].x != scrX) || 
		( screenpts[ count - 1].y != scrY)) {
		screenpts[ count].x = scrX;
		screenpts[ count++].y = scrY;
	    }
	} else {
	    screenpts[ count].x = scrX;
	    screenpts[ count++].y = scrY;
	}
    }

/* if the closing point is duplicated, then reduce that too */

    while (( count > 0) &&
	   ( screenpts[ count - 1].x == screenpts[ 0].x) &&
	   ( screenpts[ count - 1].y == screenpts[ 0].y)) {
	count--;
    }
    *numpts = count;
}

/****************************************************************************
 *
 * Description:
 *      range function for polygon plots. This overrides the virtual function
 *      for AmDataGraphicClass, which only tests x,y coordinates.
 * Arguments:
 *      xmin, ymin, xmax, ymax: returned data extents.
 *
 ****************************************************************************/

void 
AmPolygonPlotClass::GetDataExtents( float          *xmin,
				  float          *ymin,
				  float          *xmax,
				  float          *ymax)
{
    uchar_t	*ptr;
    AmPolygonPtr	poly;
    int		ii;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0, ptr = dataInfo.data; 
   	  ii < dataInfo.dims[ 0]; 
	  ii++, ptr += dataInfo.size) {
	poly = ( AmPolygonPtr) ptr;
	if ( poly->xmin < *xmin) {
	    *xmin = poly->xmin;
	}
	if ( poly->ymin < *ymin) {
	    *ymin = poly->ymin;
	}
	if ( poly->xmax > *xmax) {
	    *xmax = poly->xmax;
	}
	if ( poly->ymax > *ymax) {
	    *ymax = poly->ymax;
	}
    }
}

/* find the polygon with the most points. this is useful for allocating the
   point buffer for drawing the polygons, so we're not constantly reallocing */

static int 
maxPolyPts( AmDataInfoPtr dataInfo) 
{
    AmPolygonPtr	poly;
    int		ii, maxpts;
    uchar_t	*ptr;

    maxpts = 0;
    for ( ii = 0, ptr = dataInfo->data; 
	  ii < dataInfo->dims[ 0]; 
	  ii++, ptr += dataInfo->size) {
	poly = ( AmPolygonPtr) ptr;
	if ( poly->numPts > maxpts) {
	    maxpts = poly->numPts;
	}
    }
    return( maxpts);
}

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmPolygonPlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmPOLYGON_PLOT) {
        return( TRUE);
    } else {
        return( AmPlotClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a line plot.  Call the
 *      basic plot class initializations, then initialize the line width
 *      tables, the line buffer size and the object.  We don't use
 *      constructors for this because of ambiguities with varargs
 * Arguments:
 *      windowArg - window that the line plot will be drawn on.
 *      nameArg - string name of the line plot, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmPolygonPlotClass::initializeFields( AmWindowClass	*windowArg,
				      const char 	*nameArg)
{

    AmPlotClass::initializeFields( windowArg, nameArg);

/* color function has to be set explicitly. Default doesn't work 
   The range function changes to take the vector endpoints for finding the 
   extents of the plot */

    objectType = AmPOLYGON_PLOT;
    hatchAngle = 45.0;			/* hatch angles are in degrees */
    hatchWidth = 0.0;
    hatchPattern = NULL;
    hatchGC = window->CreateGC();
    AmInitIndexData( &hatchWidthIndex);
    hatchWidthIndex.offset = sizeof( AmPolygonRec);
    hatchWidthList = NULL;
    AmInitIndexData( &hatchAngleIndex);
    hatchAngleIndex.offset = sizeof( AmPolygonRec);
    hatchAngleList = NULL;
    AmInitIndexData( &hatchPatternIndex);
    hatchPatternIndex.offset = sizeof( AmPolygonRec);
    hatchPatternList = NULL;
    colorIndex.offset = sizeof( AmPolygonRec);
    stringIndex.offset = sizeof( AmPolygonRec);
    hatched = FALSE;
    solid = FALSE;
    mostPolyPts = MOSTPOLYPTS;
    polyBufSize = POLYBUFSIZE;
}

/****************************************************************************
 *
 * Description:
 *      check values for a polygon plot.  Each check values function 
 *	calls the check values function for the parent class and is used 
 *	to set the object's IsValid() flag. 
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmPolygonPlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmPolygonPlotClass constructor.  Creates a line plot using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the polygon plot.
 * Arguments:
 *      windowArg - window that the polygon plot will be drawn on.
 *      nameArg - string name of the polygon plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolygonPlotClass::AmPolygonPlotClass( AmWindowClass  *windowArg,
				        const char     *nameArg, 
				        ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( IsValid() && ( dataInfo.sourceType != AmNO_SOURCE));
}

/****************************************************************************
 *
 * Description:
 *      AmPolygonPlotClass constructor.  Creates a polygon plot as part of an
 *      overlay.  The proprerties for the polygon plot are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the line plot will be added to.
 *      nameArg - string name of the line plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmPolygonPlotClass::AmPolygonPlotClass( AmOverlayClass  *overlayArg,
					const char	*nameArg,
					...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);

/* call check values to make sure the plot's valid, and that there's a 
   source */

    if ( dataInfo.sourceType == AmNO_SOURCE) {
    	AmAddResourceToLog( "noSourceType");
	SetValid( FALSE);
    }
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      do-nothing constructor so polygon plots can be subclassed.
 *
 ****************************************************************************/

AmPolygonPlotClass::AmPolygonPlotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmPolygonPlotClass destructor.  free the hatch width and hatch angle
 *	index structures, and the hatch width and hatch angle lists if
 *	they've been allocated.
 *
 ****************************************************************************/


AmPolygonPlotClass::~AmPolygonPlotClass()
{
    AmDestroyIndexData( &hatchWidthIndex);
    AmDestroyIndexData( &hatchAngleIndex);
    AmDestroyIndexData( &hatchPatternIndex);
    if ( hatchWidthList != NULL) {
	free( hatchWidthList);
    }
    if ( hatchAngleList != NULL) {
	free( hatchAngleList);
    }
    if ( hatchPatternList != NULL) {
	AmFreeStrVec( hatchPatternList);
    }
    delete hatchGC;
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a polygon plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are retrived, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.   The hatch angle and hatch width bin vectors
 *	are retreived by hand, because they're one less in length than the
 *	hatch angle and hatch width vectors.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolygonPlotClass::getFields( AmTagValuePtr	tagValues,
			       int		numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, ii, numUnusedFields;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmPolygonPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs); 
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNhatchAngleBinvec)) {
	    if ( hatchAngleIndex.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( hatchAngleIndex.binvec, 
			     ( hatchAngleIndex.length - 1)*sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNhatchWidthBinvec)) {
	    if ( hatchAngleIndex.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( hatchWidthIndex.binvec, 
			     ( hatchWidthIndex.length - 1)*sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a polygon plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The hatch angle and hatch width bin vectors
 *	are retreived by hand, because they're one less in length than the
 *	hatch angle and hatch width vectors.  The hatch width and hatch
 *	angle index structures are updated after the initial values have
 *	been set. The function also finds the maximum number of points
 *	in a polygon in the data set to allocate drawing buffers.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmPolygonPlotClass::setFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    AmFieldDefPtr	fieldDefs;
    int		numFieldDefs, ii, numUnusedFields;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmPolygonPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
        if ( AmMatchTag( tagValues[ ii].tag, AmNhatchAngleBinvec)) {
            AmSetBinvec( &hatchAngleIndex.binvec, hatchAngleIndex.length, 
		       &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag,  AmNhatchWidthBinvec)) {
            AmSetBinvec( &hatchWidthIndex.binvec, hatchWidthIndex.length, 
		       &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag,  AmNhatchPatternBinvec)) {
            AmSetBinvec( &hatchPatternIndex.binvec, hatchPatternIndex.length, 
		       &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmPolygonPlotClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmPlotClass::updateFields();

/* set the index data structures for the colors, hatch width and hatch angle */

    if ( IsValid()) {
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colors") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( hatchWidthList != NULL) {
	    if ( updateIndexData( &hatchWidthIndex, "hatchWidth") != AmOK) {
		AmAddResourceToLog( "failedToInitializeHatchWidthIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( hatchAngleList != NULL) {
	    if ( updateIndexData( &hatchAngleIndex, "hatchAngle") != AmOK) {
		AmAddResourceToLog( "failedToInitializeHatchAngleIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( hatchPatternList != NULL) {
	    if ( updateIndexData( &hatchPatternIndex, "hatchPattern") != AmOK) {
		AmAddResourceToLog( "failedToInitializeHatchPatternIndex", 
				    GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* If the data pointer has changed find the polygon with the most points, 
   and take the list of polygons and set their extents */

	if ( AmChanged( &dataInfo.dims[ 0], &dataInfo.size, &dataInfo.data, 
		      NULL)) {
	    if (( dataInfo.sourceType != AmPYRAMID_SOURCE) && 
		( dataInfo.sourceType != AmREAD_SOURCE)) {
		mostPolyPts = maxPolyPts( &dataInfo);
	    } else { 
		mostPolyPts = 100;		/* arbirtrary value */
	    }
	}


/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }

/* set the polygon hatch angle and hatch width functions.
   the default hatch width is the coordinate extents/100.0 if the hatchWidth
   is 0.0 */

	if ( hatchWidth == 0.0) {
	    GetRange( &xmin, &ymin, &xmax, &ymax);
	    hatchWidth = ( xmax - xmin)/100.0;
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	read polygons into a buffer, and use/set the number of bytes if 
 *	there's a partial polygon.  This is used for reading polygons 
 *	from files into buffers for display.  
 * Arguments:
 *	fd - file descriptor to read from
 *	size - size of the polygon header.
 *	buffer - buffer to read into.
 *	buflen - length of the buffer.
 *	leftOverBytes - bytes left over from the previous read, and 
 *	returns bytes left over from the current read.  This should be
 *	initialized to 0 the AmFirst time this function is called.
 *	bytesRead - number of bytes actually read.
 *	count - number of polygons actually read
 * Returns:
 *	AmREAD_FAILED if there was a read error.
 *	AmOK if everything's hunky-dory.
 *
 ****************************************************************************/

static AmErrorCode 
readPolygonBuffer( int		fd,
		   int		size,
		   uchar_t	*buffer,
		   int		buflen,
		   int		*leftOverBytes,
		   int		*bytesRead,
		   int		*count)
{
    int		numBytes, numRead, polySize;
    uchar_t	*ptr, *bufend;
    AmPolygonPtr	poly;

    numBytes = buflen - *leftOverBytes;
    if (( numRead = read( fd, &buffer[ *leftOverBytes], numBytes)) < 0) {
	return( AmREAD_FAILED);
    }
    bufend = &buffer[ numRead + *leftOverBytes];
    polySize = 0;
    *count = 0;
    for ( ptr = buffer; ptr < bufend; ptr += polySize) {
	poly = ( AmPolygonPtr) ptr;
	polySize = size + poly->numPts*sizeof( AmCoordRec);
	if (( ptr + size > bufend) || ( ptr + polySize > bufend)) {
	    break;
	}
	poly->pts = ( AmCoordPtr)( ptr + size);
	( *count)++;
    }
    *leftOverBytes = bufend - ptr;
    *bytesRead = ptr - buffer;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw the current selection set.  The selection set is a buffer
 *      containing copies of the selected polygons.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 *	fHighlight - TRUE: draw in the highlight color, FALSE, draw in
 *	the normal color.
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the line plot's invalid.
 *
 ****************************************************************************/

AmErrorCode 
AmPolygonPlotClass::DrawSelection( AmWindowClass	*window,
				   AmGCClass		*gc, 
				   AmRectangleRec const	*exposeRects,
				   int			nRects,
				   AnvilBoolean		fHighlight)
{
    uchar_t		*selected, *ptr;
    int			nSel;
    AmRectangleRec	surroundRect;
    float		xmin, ymin, xmax, ymax;
    AmErrorCode		errCode;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* get the selection */

    GetSelection(( void **) &selected, &nSel);
    if ( nSel > 0) {

/* take the expose rectangle and translate it to data coordinates to clip 
   polygons */
   
	AmUnionRectVec( exposeRects, nRects, &surroundRect);
	GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
	GetDataPt( surroundRect.x + surroundRect.width, 
		   surroundRect.y + surroundRect.height, &xmax, &ymin);
	gc->SetLineWidth( 0);
	errCode = drawPolygonVector( display, window, gc, 
				     selected, &ptr, nSel, fHighlight);
	return( errCode);
    } else {
	return( AmOK);
    }
}

/****************************************************************************
 *
 * Description:
 * 	when polygons are read from a file, they're read as a block of
 *	data followed by a vector of vertices.  There's a pointer in 
 *	the polygon structure which is supposed point to the vector of
 *	vertices, and this function sets the pointer for a set of 
 *	polygons which have been read into a buffer.
 * Arguments:
 *	polyBuffer - buffer of polygon structures followed by vertices.
 *	numPolys - number of polygons in the buffer.
 *	size - size of each polygon header structure.
 *
 ****************************************************************************/

static void 
setPolygonPtPtrs( uchar_t	*polyBuffer,
		   int		numPolys,
		   int		size)
{
    AmPolygonPtr	poly;
    int		ii, polySize;
    uchar_t	*ptr;

    polySize = 0;
    for ( ii = 0, ptr = polyBuffer; ii < numPolys; ii++, ptr += polySize) {
	poly = ( AmPolygonPtr) ptr;
	poly->pts = ( AmCoordPtr) ( ptr + size);
	polySize = size + poly->numPts*sizeof( AmCoordRec);
    }
}

/****************************************************************************
 *
 * Description:
 *	Draws the polygons to the screen. Clear out the background if the
 *	plot's opaque, and draw the polygons from the buffer, read them
 *	block by block, or read the tiles from the pyramid. 
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmPolygonPlotClass::draw( AmWindowClass		*window,
			  AmGCClass		*gc, 
			  AmRectangleRec const	*exposeRects,
			  int			nRects)
{
    int			ii, numPolys, count, leftOverBytes, bytesRead;
    uchar_t		*ptr, *polyBuffer;
    float		xmin, ymin, xmax, ymax;
    AmErrorCode		errCode;
    AmRectangleRec	surroundRect;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* if there's no overlay data, then we clear the area that we're drawing to.
   We use FillRectangle, because we have to pay attention to the gc's clip
   region, which is set by the caller */

    if (( parent == NULL) && opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( window, gc, exposeRects[ ii].x, 
						exposeRects[ ii].y,
						exposeRects[ ii].width, 
						exposeRects[ ii].height);
	}
    }

/* take the expose rectangle and translate it to data coordinates to clip 
   polygons */
   
    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    GetDataPt( surroundRect.x, surroundRect.y, &xmin, &ymax);
    GetDataPt( surroundRect.x + surroundRect.width, 
	       surroundRect.y + surroundRect.height, &xmax, &ymin);

/* set the line width which may be overridden by functions in the loop */

    gc->SetLineWidth( 0);

/* cast each element to a data segment and draw it.  Buffer the polygons 
   by color to try to speed things up */

    if ( dataInfo.sourceType == AmREAD_SOURCE) {

/* allocate a buffer to read the polygons in chunks, and seek to the head
   of the file */

	if (( polyBuffer = ( uchar_t *) alloca( polyBufSize)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0)  {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	leftOverBytes = 0;
	count = 0;
	while ( count < dataInfo.dims[ 0]) {
	    if (( errCode = readPolygonBuffer( 
				dataInfo.fd, dataInfo.size, polyBuffer,
				polyBufSize, &leftOverBytes, &bytesRead,
				&numPolys)) != AmOK) {
		return( errCode);
	    }
printf( "read leftover = %d, bytesRead = %d\n", leftOverBytes, bytesRead);
	    errCode = drawPolygonVector( display, window, gc, polyBuffer, &ptr,
					 numPolys, FALSE);
	    if ( errCode != AmOK) {
		return( errCode);
	    }

/* if we're not at the end, copy the last chunk read into the head of
   the buffer */

	    count += numPolys;
	    if ( leftOverBytes > 0) {
		memcpy( polyBuffer, &polyBuffer[ bytesRead], leftOverBytes);
	    }
	}
    } else if (( dataInfo.sourceType == AmFILE_SOURCE) ||
	       ( dataInfo.sourceType == AmDIRECT_SOURCE)) {
	errCode = drawPolygonVector( display, window, gc, dataInfo.data, &ptr,
				     dataInfo.dims[ 0], FALSE);
	if ( errCode != AmOK) {
	    return( errCode);
	}
    } else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
        AmTilePtr         tile;
	AmTileRec		tiles[ MAXTILES];
        AmErrorCode     errCode;
        int             numTiles, maxBytes;
	uchar_t		*polyBuffer;

        numTiles = 0;
        errCode = AmFindLayerTiles( coords.currXmin, coords.currYmin,
                                  coords.currXmax, coords.currYmax,
				  ( float) detailLevel,
                                  dataInfo.fd, &tileHeader->rootTile, tiles,
                                  &numTiles, MAXTILES);
        if ( errCode != AmOK) {
            return( errCode);
        }

/* find the size of the largest tile, and allocate a buffer to hold it */ 
	
	maxBytes = 0;
	for ( ii = 0; ii < numTiles; ii++) {
	    if ( tiles[ ii].numBytesUncompressed > (unsigned)maxBytes) {
		maxBytes = tiles[ ii].numBytesUncompressed;
	    }
	}

/* allocate data to hold the largest tile */

	if (( polyBuffer = ( uchar_t *) alloca( maxBytes)) == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}

/* read each of the tiles */

	for ( ii = 0; ii < numTiles; ii++) {
	    tile = &tiles[ ii];
	    errCode = AmReadTile( tileHeader, tile, polyBuffer);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	    setPolygonPtPtrs( polyBuffer, tile->rows, dataInfo.size);
	    errCode = drawPolygonVector( display, window, gc, polyBuffer, &ptr,
				         tile->rows, FALSE);
	    if ( errCode != AmOK) {
		return( errCode);
	    }
	}
    }

/* restore the fill style to solid */

    gc->SetFillStyle( AmFILLSTYLE_SOLID);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      read data from a file into a block of memory.  First, check if
 *      the file is in the cache, and if so, then use the cache pointer
 *      otherwise, open the file, seek over any header bytes,
 *      allocate the data and read it.  When reading the polygons, this
 *	function sets the pointer to the vector of vertices which follow
 *	the polygon.
 * Arguments:
 *      uses the dataInfo substructure.
 * Returns:
 *      AmOK if everything's OK, system error otherwise
 *
 ****************************************************************************/

AmErrorCode 
AmPolygonPlotClass::readData()
{
    int			ii, length, fd, polySize;
    AmPolygonPtr	polygon;
    uchar_t		*ptr, *data;
    struct stat		statbuf;

    if ( dataInfo.enableCacheFlag) {
	if (( data = ( uchar_t *) AmGetCache( dataInfo.filename)) != NULL) {
	    dataInfo.data = data;
	    return( AmOK);
	}
    }
    if (( fd = open( dataInfo.filename, O_RDONLY)) < 0) {
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* find out if the data is in the cache already */

    if ( fstat( fd, &statbuf) < 0) {
	DO_RETURN_ERROR( AmSTAT_FAILED);
    }
    length = statbuf.st_size - dataInfo.skipBytes;

/* polygons AmFirst, points second */

    if (( dataInfo.data = ( uchar_t *) malloc( length)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if ( lseek( fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if ( read( fd, dataInfo.data, length) < length) {
	DO_RETURN_ERROR( AmREAD_FAILED);
    }

/* step the data pointer by the size of each polygon */

    polySize = 0;
    for ( ii = 0, ptr = dataInfo.data; 
	  ii < dataInfo.dims[ 0]; 
	  ii++, ptr += polySize) {
	polygon = ( AmPolygonPtr) ptr;
	polygon->pts = ( AmCoordPtr) ( ptr + dataInfo.size);
	polySize = dataInfo.size + polygon->numPts*sizeof( AmCoordRec);
	if ( !( polygon->flags & AM_MBR_FLAG)) {
	    setPolygonExtents( polygon);
	}
    }
    if ( dataInfo.enableCacheFlag) {
	AmAddToCache( dataInfo.filename, dataInfo.data);
    }
    close( fd);
    return( AmOK);
}

/*************************************************************************
 *
 * Description
 *      Return the extents of a pointer cast to a polygon.  Each polygon
 *	contains the extents of its minimum bounding rectangle.
 * Arguments:
 *      ptr - void pointer to a line to get the extents of.
 *      xmin, ymin, xmax, ymax - returns extents of line.
 *
 *************************************************************************/


void 
AmPolygonPlotClass::ObjectExtents( void      *ptr,
				   float     *xmin,
				   float     *ymin,
				   float     *xmax,
				   float     *ymax)
{
    AmPolygonPtr	poly = ( AmPolygonPtr) ptr;

    *xmin = poly->xmin;
    *ymin = poly->ymin;
    *xmax = poly->xmax;
    *ymax = poly->ymax;
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a polygon plot.  This is
 *      used in the map editor to dump the plot's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmPolygonPlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			        AnvilBoolean		    allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( coordinate system) */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmPolygonPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);
    if ( hatchAngleIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNhatchAngleBinvec;
	fieldLengths[ numStrings++].length = hatchAngleIndex.length - 1;
    }
    if ( hatchWidthIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNhatchWidthBinvec;
	fieldLengths[ numStrings++].length = hatchWidthIndex.length - 1;
    }
    if ( hatchPatternIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNhatchPatternBinvec;
	fieldLengths[ numStrings++].length = hatchPatternIndex.length - 1;
    }
    return( numStrings);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a line plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmPolygonPlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.  THIS TABLE MUST BE SORTED ON 
 *	ALPHABETICAL ORDER FOR SEARCHING. 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void AmPolygonPlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					    int		*numDefs)
{
    static AmFieldDefRec PolygonPlotClassFieldDefinitions[] = {
	{ AmNdrawSolid, offsetof( AmPolygonPlotClass, solid) },
	
	{ AmNhatchAngle, offsetof( AmPolygonPlotClass, hatchAngle) },
        { AmNhatchAngleClientData,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.clientData) },
        { AmNhatchAngleClientDataFreeFn,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.clientDataFreeFn) },
        { AmNhatchAngleFn, offsetof( AmPolygonPlotClass, hatchAngleIndex.fn) },
        { AmNhatchAngleIntFn, 
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.AmIntFn) },
        { AmNhatchAngleLength, 
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.length) },
        { AmNhatchAngleList, offsetof( AmPolygonPlotClass, hatchAngleList),
                            offsetof( AmPolygonPlotClass, hatchAngleIndex.length),
                            AmDYNAMIC },
        { AmNhatchAngleMax,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.max) },
        { AmNhatchAngleMin,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.min) },
        { AmNhatchAngleName,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.name), 0, AmDYNAMIC },
        { AmNhatchAngleOffset,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.offset) },
        { AmNhatchAngleStrings, 
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.strings),
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.length),
	  AmDYNAMIC },
        { AmNhatchAngleTable, 
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.table),
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.length),
	  AmDYNAMIC },
        { AmNhatchAngleType,
          offsetof( AmPolygonPlotClass, hatchAngleIndex.dataType) },
        { AmNhatchAngleUnits, 
	  offsetof( AmPolygonPlotClass, hatchAngleIndex.units) },
	{ AmNhatchPattern, offsetof( AmPolygonPlotClass, hatchPattern) },
        { AmNhatchPatternClientData,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.clientData) },
        { AmNhatchPatternClientDataFreeFn,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.clientDataFreeFn) },
        { AmNhatchPatternFn, offsetof( AmPolygonPlotClass, hatchPatternIndex.fn) },
        { AmNhatchPatternIntFn, 
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.AmIntFn) },
        { AmNhatchPatternLength, 
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.length) },
        { AmNhatchPatternList, offsetof( AmPolygonPlotClass, hatchPatternList),
                            offsetof( AmPolygonPlotClass, hatchPatternIndex.length),
                            AmDYNAMIC },
        { AmNhatchPatternMax,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.max) },
        { AmNhatchPatternMin,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.min) },
        { AmNhatchPatternName,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.name), 0, AmDYNAMIC },
        { AmNhatchPatternOffset,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.offset) },
        { AmNhatchPatternStrings, 
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.strings),
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.length),
	  AmDYNAMIC },
        { AmNhatchPatternTable, 
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.table),
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.length),
	  AmDYNAMIC },
        { AmNhatchPatternType,
          offsetof( AmPolygonPlotClass, hatchPatternIndex.dataType) },
        { AmNhatchPatternUnits, 
	  offsetof( AmPolygonPlotClass, hatchPatternIndex.units) },
	{ AmNhatchWidth, offsetof( AmPolygonPlotClass, hatchWidth) },
        { AmNhatchWidthClientData,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.clientData) },
        { AmNhatchWidthClientDataFreeFn,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.clientDataFreeFn) },
        { AmNhatchWidthFn, offsetof( AmPolygonPlotClass, hatchWidthIndex.fn) },
        { AmNhatchWidthIntFn, 
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.AmIntFn) },
        { AmNhatchWidthLength, 
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.length) },
        { AmNhatchWidthList, 
	  offsetof( AmPolygonPlotClass, hatchWidthList),
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.length),
	  AmDYNAMIC },
        { AmNhatchWidthMax,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.max) },
        { AmNhatchWidthMin,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.min) },
        { AmNhatchWidthName,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.name), 0, AmDYNAMIC },
        { AmNhatchWidthOffset,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.offset) },
        { AmNhatchWidthStrings, 
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.strings),
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.length),
	  AmDYNAMIC },
        { AmNhatchWidthTable, 
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.table),
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.length),
	  AmDYNAMIC },
        { AmNhatchWidthType,
          offsetof( AmPolygonPlotClass, hatchWidthIndex.dataType) },
        { AmNhatchWidthUnits, 
	  offsetof( AmPolygonPlotClass, hatchWidthIndex.units) },
	{ AmNhatched, offsetof( AmPolygonPlotClass, hatched) },
    };
    *fieldDefs = PolygonPlotClassFieldDefinitions;
    *numDefs = VECLEN( PolygonPlotClassFieldDefinitions);
}
