#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* reference windows are used to show a parent coordiante system zoomed
   out to its full extents.  They are typically used in a dialog where
   an outline rectangle draws the current extents of the parent, and
   is updated when the parent's coordinate system is changed, or can
   be dragged to pan the parent's coordinate system */

#include <stdarg.h>
#include <stdio.h>
#include <anvil/xdisplay.h>
#include <anvil/mem_funcs.h>
#include <anvil/va_cback.h>
#include <anvil/plot.h>
#include <anvil/copyobj.h>
#include <anvil/reference.h>

/* reference windows are used to show the full extents of a parent plot, 
   usually in a reference window dialog */

/****************************************************************************
 *
 * Description:
 * 	draw the reference window at full extents, with a rectangle to show 
 *	the current extents of the parent plot.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, which it is.
 *
 ****************************************************************************/

AmErrorCode 
AmReferenceWindowClass::draw( AmWindowClass	   	*w,
			    AmGCClass		*drawGC,
			    AmRectangleRec const 	*exposeRects,
			    int		   	nRects)
{
    float		xmin, ymin, xmax, ymax;
    int			x0, y0, x1, y1;
    AmRectangleRec const	*plotRect = plot->GetRect();

    display->CopyArea( pixWin, w, drawGC, 
		       0, 0, rect.x, rect.y, rect.width, rect.height);
    (( AmCoordClass *) parent)->GetCoords( &xmin, &ymin, &xmax, &ymax);
    plot->GetScreenPt( xmin, ymax, &x0, &y0);
    plot->GetScreenPt( xmax, ymin, &x1, &y1);
    x0 += rect.x - plotRect->x;
    y0 += rect.y - plotRect->y;
    x1 += rect.x - plotRect->x;
    y1 += rect.y - plotRect->y;

/* draw the rectangle in black and white dashes so we can always see it, 
   and restore to a normal line style, because it's pretty uncommon for 
   draw routines to set line style. */

    drawGC->SetForegroundColor( &fgColor);
    drawGC->SetBackgroundColor( &bgColor);
    drawGC->SetDashes( "\2\2", 2, 0);
    drawGC->SetLineStyle( AmLINE_DOUBLE_DASH);
    drawGC->SetLineWidth( 2);
    display->DrawRectangle( w, drawGC, x0, y0, x1 - x0, y1 - y0);
    drawGC->SetLineStyle( AmLINE_SOLID);
    drawGC->SetLineWidth( 0);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	if the source plot has been updated, then trace it.  If it's been 
 *	updated or the coordinates have been changed, set the coordinates 
 *	and range.
 * Arguments:
 *      eventCodes - list of event codes sent by AmSendUserEvent()
 *      numCodes - number of event codes.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmReferenceWindowClass::Update( int	*eventCodes,
			      int	numCodes)
{
    float		xmin, ymin, xmax, ymax;
    AnvilBoolean		newPlot;
    AmRectangleRec const	*plotRect;

    if ( !IsValid()) {
	DO_RETURN_ERROR( AmNOT_VALID);
    }
    newPlot = ( AmInIntVector( AM_SET_FIELD_OBJECT, eventCodes, numCodes) ||
		( numCodes == 0));
    if ( newPlot) {
	if ( plot != NULL) {
	    delete plot;
	}
	plot = ( AmPlotClass *) AmTraceObject( window, parent, NULL);
	opaque = plot->getOpaque();
	plot->SetVaFields( AmNpreserveAspect, FALSE, NULL);
	(( AmCoordClass *) parent)->GetRange( &xmin, &ymin, &xmax, &ymax);
	plot->SetCoords( xmin, ymin, xmax, ymax);
    }
    if ( newPlot || AmInIntVector( AM_RECT_CHANGE, eventCodes, numCodes)) {
	plot->SetRect( 0, 0, rect.width, rect.height);
	plot->Update( eventCodes, numCodes);
	plotRect = plot->GetRect();
	plot->Draw( pixWin, gc, plotRect, 1);
	return( AmOK);
    } else {
	return( AmOK);
    }
}

/****************************************************************************
 *
 * Description:
 *	This overrides the SetRect() member function, because the 
 *	reference window is actually drawn to a buffer which is then
 *	blitted to the screen.  This allocates a new buffer, then
 *	sets the rectangle for the plot.
 * Arguments:
 *      x, y, width, height:  new dimensions for the screen rectangle.
 *
 ****************************************************************************/

void 
AmReferenceWindowClass::SetRect( int	x,
			       int	y,
			       int	width,
			       int	height)
{
    if ( pixWin != NULL) {
        delete pixWin;
	pixWin = NULL;
    }
    if (( width != 0) && ( height != 0)) {
	pixWin = display->CreateBuffer( window, width, height);
    }
    AmScreenObjClass::SetRect( x, y, width, height);
}

/****************************************************************************
 *
 * Description:
 * 	Destructor: delete the private plot, buffer and graphics context 
 *
 ****************************************************************************/

AmReferenceWindowClass::~AmReferenceWindowClass()
{
    delete plot;
    if ( pixWin != NULL) {
	delete pixWin;
    }
    delete gc;
}

/****************************************************************************
 *
 * Description:
 * 	constructor: create a reference window, which shows the fully 
 *	zoomed out extents of the plot.
 * Arguments:
 *      windowArg - window that the reference window will be drawn on.
 *      parentArg - parent object which will be drawn at full extents of
 *		    its coordinate system.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmReferenceWindowClass::AmReferenceWindowClass( AmWindowClass		*windowArg,
					    AmCoordClass		*parentArg,
					    const char		*nameArg,
					    ...)
{
    va_list     	args;
    AmTagValueRec 	tagValues[ MAXTAGS];
    int         	count;
    float		xmin, ymin, xmax, ymax;

    AmScreenObjClass::initializeFields( windowArg, nameArg);
    objectType = AmREFERENCE_WINDOW;
    parent = parentArg;
    plot = ( AmPlotClass *) AmTraceObject( window, parent, NULL);
    plot->GetRange( &xmin, &ymin, &xmax, &ymax);
    plot->SetCoords( xmin, ymin, xmax, ymax);
    gc = window->CreateGC();
    pixWin = NULL;

/* set the default image client data and pixel function */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    plot->SetRect( 0, 0, rect.width, rect.height);
    opaque = plot->getOpaque();
    plot->SetVaFields( AmNpreserveAspect, FALSE, NULL);
    SetRect( 0, 0, rect.width, rect.height);
}

/****************************************************************************
 *
 * Description:
 *      set the fields in the referece window from the tag-value argument 
 *	list.  Each class calls setFields() for the parent class, then uses
 *      SetFieldByDefinitions() to set its own members.  
 * Arguments:
 *      tagValues: tag-value pairs to be parsed (this array is reduced in
 *                 place, with the parsed arguments removed).
 *      numValues: length of the tagValues array.
 * Return:
 *      number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int AmReferenceWindowClass::setFields( AmTagValuePtr	tagValues,
				     int		numValues)
{
    return( AmScreenObjClass::setFields( tagValues, numValues));
}
/****************************************************************************
 *
 * Description:
 *      retreive values for a reference window.  Each getFields() function 
 *	calls getFields() for the parent class, then most of the fields are
 *      retrived using the GetFieldByDefinitions() function.  
 * Arguments:
 *      tagValues - vector of tag-values to retrieve.  The values are
 *      pointers to the results.  The array is "scrunched" in place
 *      as each value is retreived.  Any unprocessed values are returned
 *      in place in the array.
 *      numValues - number of values in the array.
 * Returns:
 *      number of unprocessed values.
 *
 ****************************************************************************/


int AmReferenceWindowClass::getFields( AmTagValuePtr	tagValues,
				     int		numValues)
{
    return( AmScreenObjClass::getFields( tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *      Dump the field names used in a reference window.  dumpFields() 
 *	is called recursively by each object and its parent.  The 
 *	fieldLengths array returns the name of each field, and if it's 
 *	a vector field, the vector's length.
 * Arguments:
 *      fieldLengths - vector of tags and vector lengths.
 *      allFields - an obsolete tag which was used to differentiate
 *      AmBetween dumping to a file or copying the object.
 * Returns:
 *      The number of fields dumped
 *
 ***************************************************************************/


int AmReferenceWindowClass::dumpFields( AmFieldLengthPtr  fieldLengths,
				      AnvilBoolean         allFields)
{
    return( AmScreenObjClass::dumpFields( fieldLengths, allFields));
}

AnvilBoolean AmReferenceWindowClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmREFERENCE_WINDOW) {
	return( TRUE);
    } else {
	return( AmScreenObjClass::IsType( type));
    }
}

/* reference window is just a screen object, so there are no extra 
   field definitions */

void AmReferenceWindowClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
						int		*numFieldDefs)
{
    *numFieldDefs = 0;
}

/****************************************************************************
 *
 * Description:
 * 	return the reference plot created for the reference window 
 * Returns:
 *	The private reference plot
 *
 ***************************************************************************/

AmCoordClass *AmReferenceWindowClass::GetPlot()
{
    return( plot);
}
