#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <anvil/macros.h>
#include <anvil/mem_funcs.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/display.h>
#include <anvil/image.h>
#include <anvil/log.h>
#include <anvil/fns.h>
#include <anvil/overlay.h>
#include <anvil/copyobj.h>
#include <anvil/coordinates.h>
#include <anvil/imageplot.h>
#include <anvil/tagstrs.h>
#include <anvil/render3d.h>

static float 
verticalExaggerationFn( void			*data,
			AmVerticalClientDataPtr	clientData)
{
    return( clientData->verticalExaggeration*
		( clientData->zFn)( data, clientData->clientData));
}

AnvilBoolean
Render3DClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmRENDER_3D) {
        return( TRUE);
    } else {
        return( AmScreenObjClass::IsType( type));
    }
}


void 
Render3DClass::initializeFields( AmWindowClass	*windowArg,
				 const char	*nameArg)
{
    AmScreenObjClass::initializeFields( windowArg, nameArg);
    objectType = AmRENDER_3D;
    displayImage = NULL;
    texImage = NULL;
    elevationPlot = NULL;
    elevationMatrix = NULL;
    ex = 0.0;
    ey = 0.0;
    ez = 0.0;
    focal = 1.0;
    theta = 0.0;
    phi = 0.0;
    screenWidth = 1.0;
    screenHeight = 1.0;
    texWidth = 256;
    texHeight = 256;
    matrixRows = 0;
    matrixCols = 0;
    verticalClientData.verticalExaggeration = 1.0;
    verticalClientData.clientData = NULL;
    verticalClientData.zFn = NULL;
}

Render3DClass::Render3DClass()
{
}

Render3DClass::Render3DClass( AmWindowClass	*windowArg,
			      AmOverlayClass	*overlayArg,
			      const char	*nameArg,
			      ...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         n;

/* initialize all the basic fields for a plot */

    initializeFields( windowArg, nameArg);
    setParent( overlayArg);

/* set the default image client data and pixel function */

    va_start( args, nameArg);
    n = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    n = setFieldsWrapper( tagValues, n);
}

void
Render3DClass::updateFields()
{
}

int 
Render3DClass::setFields( AmTagValuePtr	tagValues,
			  int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

/* Initially, it's valid. */

    SetValid( TRUE);

/* Parse all the standard arguments. */

    numValues = AmScreenObjClass::setFields( tagValues, numValues);
    Render3DClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
                                       tagValues, numValues);
    return( numValues);
}

int
Render3DClass::getFields( AmTagValuePtr  tagValues,
                           int          numValues)
{
    AmFieldDefPtr fieldDefs;
    int         numFieldDefs;

    numValues = AmScreenObjClass::getFields( tagValues, numValues);
    Render3DClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
                                       tagValues, numValues);
    return( numValues);
}

AnvilBoolean Render3DClass::checkValues()
{
    return( AmScreenObjClass::checkValues());
}

Render3DClass::~Render3DClass()
{
}

AmErrorCode 
Render3DClass::Update( int	*eventCodes,
		       int	numCodes)
{
    AmWindowClass		*texBuffer;
    AmOverlayClass	*parentOverlay = ( AmOverlayClass *) parent;
    AmGCClass		*tempGC;
    float		subXmin, subYmin, subXmax, subYmax,
			currXmin, currYmin, currXmax, currYmax;
    int			subRows, subCols, size, x0, y0, x1, y1;
    uchar_t		*matrix;
    AmRectangleRec	copyRect;
    AnvilBoolean		flipRows, flipCols;
    AmColorCellRec	bgColor;

    if ( IsValid()) {

/* copy the parent overlay, and the elevation plot, make sure the coordinates
   are the same, then set their rectangles to the size of the texture map and 
   update them.  Create a temporary buffer and draw the overlay into it */

	parentOverlay = ( AmOverlayClass *) AmTraceObject( window, parent, NULL);
	elevationPlot = ( AmImagePlotClass *) 
	    AmTraceObject( window, sourceElevationPlot, NULL);
	parentOverlay->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
	elevationPlot->SetCoords( currXmin, currYmin, currXmax, currYmax);
	parentOverlay->SetRect( 0, 0, texWidth, texHeight);
	parentOverlay->Update();
	elevationPlot->SetRect( 0, 0, texWidth, texHeight);
	elevationPlot->Update();
	texBuffer = display->CreateBuffer( window, texWidth, texHeight);
	tempGC = window->CreateGC();
	parentOverlay->draw( texBuffer, tempGC, parentOverlay->GetRect(), 1);
	if ( texImage != NULL) {
	    if (( texImage->GetWidth() != texWidth) ||
		( texImage->GetHeight() != texHeight)) {
		delete texImage;
	    }
	}

/* copy the texture map image from the temporary buffer */

	texImage = display->CreateImage( texWidth, texHeight, 
					 display->GetDepth(), 
					 NULL);
	display->GetImage( texBuffer, texImage, 0, 0);
	delete texBuffer;

/* get the information from the elevation plot */

	elevationPlot->GetVaFields( AmNcolorsFn, 
					&verticalClientData.zFn,
				    AmNcolorsClientData, 
					&verticalClientData.clientData,
				    AmNsubXmin, &subXmin,
				    AmNsubYmin, &subYmin,
				    AmNsubXmax, &subXmax,
				    AmNsubYmax, &subYmax,
				    AmNsubRows, &subRows,
				    AmNsubCols, &subCols,
				    AmNflipRows, &flipRows,
				    AmNflipCols, &flipCols,
				    AmNdata, &matrix,
				    AmNsize, &size,
				    NULL);
	elevationPlot->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);

/* copy the elevation data into a temporary matrix */

	AmGetRowcolumn( currXmin, currYmin, subRows, subCols,
		      flipRows, flipCols, subXmin, subYmin, subXmax, subYmax,
		      &x0, &y0);
	AmGetRowcolumn( currXmax, currYmax, subRows, subCols,
		      flipRows, flipCols, subXmin, subYmin, subXmax, subYmax,
		      &x1, &y1);
	copyRect.x = MIN( x0, x1);
	copyRect.y = MIN( y0, y1);
	copyRect.width = MAX( x0, x1) - copyRect.x;
	copyRect.height = MAX( y0, y1) - copyRect.y;
	if ( elevationMatrix != NULL) {
	    FREE( elevationMatrix);
	}
	matrixRows = copyRect.height;
	matrixCols = copyRect.width;
	elevationMatrix = ( uchar_t *) malloc( matrixRows*matrixCols*size);
	AmCopySubmatrix( matrix, subRows, subCols, &copyRect, elevationMatrix, 
		         matrixRows, matrixCols, 0, 0, size, flipRows, flipCols);
	if ( displayImage != NULL) {
	    delete displayImage;
	}
	displayImage = display->CreateImage( rect.width, rect.height,
					     display->GetDepth(), 
					     NULL);

/* draw the 3D image */

	bgColor.red = 0;
	bgColor.green = 0;
	bgColor.blue = 0;
	display->AllocColor( &bgColor);
	displayImage->ClearImageRect( &bgColor, &rect);
	AmImage3D( displayImage, ex, ey, ez, focal, theta, phi,
		 screenWidth, screenHeight, elevationMatrix, 
		 matrixRows, matrixCols, size, 
		 ( AmAccessFn) verticalExaggerationFn, &verticalClientData, 
		 currXmin, currYmin, currXmax, currYmax, texImage);
    }
    return( AmOK);
}

void 
Render3DClass::updateViewpoint()
{
    float 		currXmin, currYmin, currXmax, currYmax;
    int			size;
    AmColorCellRec	bgColor;

    elevationPlot->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
    elevationPlot->GetVaFields( AmNsize, &size,
				NULL);
    bgColor.red = 0;
    bgColor.green = 0;
    bgColor.blue = 0;
    display->AllocColor( &bgColor);
    displayImage->ClearImageRect( &bgColor, &rect);
    AmImage3D( displayImage, ex, ey, ez, focal, theta, phi,
	     screenWidth, screenHeight, elevationMatrix, 
	     matrixRows, matrixCols, size, 
	     ( AmAccessFn) verticalExaggerationFn, &verticalClientData, 
	     currXmin, currYmin, currXmax, currYmax, texImage);
}

AmErrorCode
Render3DClass::draw( AmWindowClass	*window,
		     AmGCClass		*gc,
		     AmRectangleRec const	*exposeRects,
		     int		nRects)
{
    AmRectangleRec const  *dispRect = GetRect();
    int                 ii;

    if ( !IsValid()) {
        return( AmOK);
    }
    if ( AmDegenRect( dispRect)) {
        return( AmOK);
    }
    if ( displayImage == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }
    for ( ii = 0; ii < nRects; ii++) {
	display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
                                exposeRects[ ii].y - dispRect->y,
                                exposeRects[ ii].x, exposeRects[ ii].y,
                                exposeRects[ ii].width,
                                exposeRects[ ii].height,
                                displayImage);
    }
    return( AmOK);
}

int
Render3DClass::dumpFields( AmFieldLengthPtr fieldLengths,
                           AnvilBoolean        allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* Get the fields for the parent class ( screenobject). */

    numStrings = AmScreenObjClass::dumpFields( fieldLengths, allFields);

/* Get all the easy to specify fields for this class. */

    AmScreenObjClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
                                          fieldDefs, numFieldDefs);
    return( numStrings);
}


void
Render3DClass::GetFieldDefinitions( AmFieldDefPtr        *fieldDefs,
                                     int                *numFieldDefs)
{
    static AmFieldDefRec Render3DClassFieldDefinitions[] = {
	{ AmNelevationPlot, offsetof( Render3DClass, sourceElevationPlot) },
	{ AmNex, offsetof( Render3DClass, ex) },
	{ AmNey, offsetof( Render3DClass, ey) },
	{ AmNez, offsetof( Render3DClass, ez) },
	{ AmNfocal, offsetof( Render3DClass, focal) },
	{ AmNphi, offsetof( Render3DClass, phi) },
	{ AmNscreenHeight, offsetof( Render3DClass, screenHeight) },
	{ AmNscreenWidth, offsetof( Render3DClass, screenWidth) },
	{ AmNtexHeight, offsetof( Render3DClass, texHeight) },
	{ AmNtexWidth, offsetof( Render3DClass, texWidth) },
	{ AmNtheta, offsetof( Render3DClass, theta) },
	{ AmNverticalExaggeration, 
	  offsetof( Render3DClass, verticalClientData.verticalExaggeration) },
    };
    *fieldDefs = Render3DClassFieldDefinitions;
    *numFieldDefs = VECLEN( Render3DClassFieldDefinitions);
}
