#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <anvil/image.h>
#include <anvil/tsearch.h>
#include <anvil/files.h>
#include <anvil/envdefs.h>
#include <anvil/display.h>
#include <anvil/pixthing.h>
#include <anvil/imageP.h>
#include <anvil/macros.h>
#include <anvil/scatter.h>
#include <anvil/colorcell.h>

/* scatter plot drawing functions which draw the symbols into an image for
   performance */


/*************************************************************************
 * 
 * Description:
 * 	given a value and an index data structure, return the pixel value
 *	of the indexed color.
 * Arguments:
 *	val - value to access and index.
 *	indexData - index data information.
 *	colors - vector of colors to index into.
 * Returns:
 *	pixel value, or -1 if the value is clipped and out of range.
 *
 *************************************************************************/

ulong_t 
AmColormapPixelFn( void            *val,
		 AmIndexDataPtr    indexData,
		 AmColorCellPtr    colors)
{
    int                 index;
    ulong_t             returnVal = 0;
    uchar_t             *p;
 
    p = ( uchar_t *) &returnVal;
    index = ( indexData->getIndexFn) ( val, indexData);
    if ( index <= 0) {
        if ( indexData->minClip) {
            return(( unsigned long) -1);
        } else {
            ASSIGN_PIXEL( colors[ 0], p);
            return( returnVal);
        }
    } else if ( index >= indexData->length - 1) {
        if ( indexData->maxClip) {
            return(( unsigned long) -1);
        } else {
            ASSIGN_PIXEL( colors[ indexData->length - 1], p);
            return( returnVal);
        }
    } else {
        ASSIGN_PIXEL( colors[ index], p);
        return( returnVal);
    }
}

/*************************************************************************
 * 
 * Description:
 * 	comparison function for the symbol cache. 
 * Arguments:
 *	a, b - symbol cache structures to compare.
 * Returns:
 *	strcmp() of their filenames.
 *
 ************************************************************************/

static int 
compareSymfile( AmSymCachePtr	a, 
		AmSymCachePtr	b)
{
    return( strcmp( a->symfile, b->symfile));
}

/*************************************************************************
 * 
 * Description:
 * 	find a symbol in the symbol cache, or load it into the cache
 *	and return it.
 * Arguments:
 *	symfile - symbol name to search for.
 *	symbolCache - the symbol cache tree to search.
 * Returns:
 *	the symbol cache pointer to the returned symbol.
 *
 ************************************************************************/

AmSymCachePtr 
AmGetSymbol( char 	*symfile, 
	   void 	**symbolCache)
{
    char	*sympath, relPath[ PATHLEN];
    AmSymCacheRec	target;
    AmSymCachePtr	sym, *foundSym;

    target.symfile = symfile;
    foundSym = ( AmSymCachePtr *) tfind( &target, (const void **)symbolCache, 
				       ( AmSearchFn) compareSymfile);

/* if the symbol wasn't found, then load it from the file and store it in
   the cache. */

    if ( foundSym == NULL) {
	sprintf( relPath, "%s/%s", AM_SYMBOL_DIR, symfile);
	sympath = AmGetEnvFile( CONFIG_ENV, relPath);
	if ( sympath == NULL) {
	    return( NULL);
	} else {
	    sym = ( AmSymCachePtr) malloc( sizeof( AmSymCacheRec));
	    sym->symfile = strdup( symfile);
	    if ( !AmReadBitmapFileData( sympath, &sym->symWidth, 
				      &sym->symHeight, &sym->symbol, 1, 0)) {
		sym->symbol = ( uchar_t *) malloc( 1);
		sym->symbol[ 0] = 1;
		sym->symWidth = 1;
		sym->symHeight = 1;
	    }
	    tsearch( sym, symbolCache, ( AmSearchFn) compareSymfile);
	    return( sym);
	}
    } else {
	return( *foundSym);
    }
}

/* for accumulating the list of symbol names for deletion */

static AmListPtr  symbolList;
 
static void getSymbolNames( const void *a, VISIT order, int)
{
    AmSymCachePtr ptr;
 
    if (( order == preorder) || ( order == leaf)) {
        ptr = *( AmSymCachePtr *) a;
        symbolList = AmInsertNodeLst( ptr->symfile, symbolList);
    }
}

/* clean up the symbol cache. */
 
void AmFreeSymbolCache( void      **symbolCache)
{
    AmListPtr     ptr;
    AmSymCacheRec target;
    AmSymCachePtr *foundSym, foundPtr;

/* remember that twalk takes a void *, where tfind, tdelete and tsearch
   take a void ** */

    symbolList = NULL;
    twalk( *symbolCache, getSymbolNames);

/* walk through the list of symbols, find each of them in the tree,
   delete it, and free the symbol, filename and structure, then
   free the temporary list */

    for ( ptr = symbolList; ptr != NULL; ptr = ptr->next) {
        target.symfile = ( char *) ptr->data;
        foundSym = ( AmSymCachePtr *) tfind( &target, (const void **)symbolCache,
                                           ( AmSearchFn) compareSymfile);
        if ( foundSym != NULL) {
	    foundPtr = *foundSym;
            tdelete( foundPtr, symbolCache, ( AmSearchFn) compareSymfile);
	    free( foundPtr->symbol);
	    free( foundPtr->symfile); 
	    free( foundPtr);
	}
    }
    AmJustFreeLst( symbolList);
}
        

static int hIndex[ MAXSYM], vIndex[ MAXSYM], hFrac[ MAXSYM], vFrac[ MAXSYM];

/* copy the symbol into the image */

void
AmBilinPix( ulong_t	*image,
          int		imageWidth,
          int		imageHeight,	
          uchar_t	*pix,
          int		pixWidth,
          int		pixHeight,
	  ulong_t	pixel,
          int		x,
          int		y,
          float		scale)
{
    int 	newWidth = MAX(( int)( scale*pixWidth), 1),
		newHeight = MAX(( int)( scale*pixHeight), 1),
		*pHindex, *pVindex = vIndex, *pVFrac, *pHFrac, ii;
    ulong_t	*imageRowEnd, red, green, blue;
    uchar_t	*p1, *p2, *p3, *p4, *pixRow;
    float	vRatio, hRatio, temp;

/* make sure the symbol is in the destination image's bounds */

    if (( x < ( newWidth >> 1)) || ( y < ( newHeight >> 1)) || 
	(( x + 1 + ( newWidth >> 1))  >= imageWidth) ||
        (( y + 1 + ( newHeight >> 1)) >= imageHeight)) {
	return;
    }

    AmFwdDiffVec( pixWidth, newWidth, hIndex);
    AmFwdDiffVec( pixHeight, newHeight, vIndex);
    vRatio = (( float) pixWidth)/(( float) newWidth);
    hRatio = (( float) pixHeight)/(( float) newHeight);
    for ( ii = 0; ii < newWidth; ii++) {
	temp = ii*hRatio;
	hFrac[ ii] = ( int) ((temp - ( int) temp)*256);
    }
    for ( ii = 0; ii < newHeight; ii++) {
	temp = ii*vRatio;
	vFrac[ ii] = ( int) ((temp - ( int) temp)*256);
    }
	

/* reset the image pointer and walk through our fake pixmap. Width loses
   pixmap width for stepping to next row, for a little bit of efficiency. */ 

    image += ( y - ( newHeight >> 1))*imageWidth + ( x - ( newWidth >> 1));
    pVFrac = vFrac;
    while ( newHeight--) {
	pHindex = hIndex;
	imageRowEnd = image + newWidth;
	pixRow = pix;
	pHFrac = hFrac;

/* only draw the pixels which are 'lit' */

	while ( image < imageRowEnd) {

/* we want to make sure that we don't step over the lit pixels in the symbol,
   so we test all of the pixels in the symbol which are mapped to the image */

	    p1 = ( uchar_t *) (( *pixRow) ? 
			       &pixel : image);
	    p2 = ( uchar_t *) (( *(pixRow + 1)) ? 
			       &pixel : image + 1);
	    p3 = ( uchar_t *) (( *(pixRow + pixWidth)) ? 
			       &pixel : image + imageWidth);
	    p4 = ( uchar_t *) ((*(pixRow + pixWidth + 1)) ? 
			       &pixel : image + imageWidth + 1);

            red = (( p1[ RED]*( 256 - *pHFrac) + p2[ RED]*( *pHFrac))*
                        ( 256 - *pVFrac) +
                   ( p3[ RED]*( 256 - *pHFrac) + p4[ RED]*( *pHFrac))*
                        ( *pVFrac)) >> 16;
            green = (( p1[ GREEN]*( 256 - *pHFrac) + p2[ GREEN]*( *pHFrac))*
                        ( 256 - *pVFrac) +
                   ( p3[ GREEN]*( 256 - *pHFrac) + p4[ GREEN]*( *pHFrac))*
                        ( *pVFrac)) >> 16;
            blue = (( p1[ BLUE]*( 256 - *pHFrac) + p2[ BLUE]*( *pHFrac))*
                        ( 256 - *pVFrac) +
                   ( p3[ BLUE]*( 256 - *pHFrac) + p4[ BLUE]*( *pHFrac))*
                        ( *pVFrac)) >> 16;
	    *image++ = ( red << RED_SHIFT) |
		       ( green << GREEN_SHIFT) |
		       ( blue << BLUE_SHIFT);
	    pHFrac++;
	    pixRow += *pHindex++;
	}
	pix += pixWidth*( *pVindex++);
	image += imageWidth - newWidth;
	pVFrac++;
    }
}

/* copy the symbol into the image */

void
AmScalePix( ulong_t	*image,
          int		imageWidth,
          int		imageHeight,	
          uchar_t	*pix,
          int		pixWidth,
          int		pixHeight,
	  ulong_t	pixel,
          int		x,
          int		y,
          float		scale)
{
    int 	newWidth = MAX(( int)( scale*pixWidth), 1),
		newHeight = MAX(( int)( scale*pixHeight), 1),
		*pHindex, *pVindex = vIndex;
    uchar_t 	*pixRow, *p, *q, *pEnd;
    ulong_t	*imageRowEnd;

/* make sure the symbol is in the destination image's bounds */

    if (( x < ( newWidth >> 1)) || ( y < ( newHeight >> 1)) || 
	(( x + ( newWidth >> 1) + 1)  >= imageWidth) ||
        (( y + ( newHeight >> 1) + 1) >= imageHeight)) {
	return;
    }

    AmFwdDiffVec( pixWidth, newWidth, hIndex);
    AmFwdDiffVec( pixHeight, newHeight, vIndex);

/* reset the image pointer and walk through our fake pixmap. Width loses
   pixmap width for stepping to next row, for a little bit of efficiency. */ 

    image += ( y - ( newHeight >> 1))*imageWidth + ( x - ( newWidth >> 1));
    while ( newHeight--) {
	pHindex = hIndex;
	imageRowEnd = image + newWidth;
	pixRow = pix;

/* only draw the pixels which are 'lit' */

	while ( image < imageRowEnd) {

/* we want to make sure that we don't step over the lit pixels in the symbol,
   so we test all of the pixels in the symbol which are mapped to the image */

	    if ( *pVindex > 0) {
		pEnd = pixRow + ( *pVindex)*pixWidth;
		for ( p = pixRow; p < pEnd; p += pixWidth) {
		    if ( *pHindex > 0) {
			for ( q = p; q < p + *pHindex; q++) {
			    if ( *q) {
				*image = pixel;
			    }
			}
		    } else {
			if ( *p) {
			    *image = pixel;
			}
		    }
		}
	    } else {
		if ( *pixRow) {
		    *image = pixel;
		}
	    }
	    image++;
	    pixRow += *pHindex++;
	}
	pix += pixWidth*( *pVindex++);
	image += imageWidth - newWidth;
    }
}

/**********************************************************************
 *
 * Description:
 *      scatter plot image drawing function using pixel replication
 * Arguments:
 *      display - display to draw to.
 *      coords - coordinate system to draw with.
 *      drawRect - bounding draw rectangle.
 *      srcData - source data to render.
 *      srcSize - size of each source data element.
 *      length - number of data elements.
 *      xFn, xOffset, yFn, yOffset - accessing functions and offsets
 *      for x,y values in the scatter points.
 *      colorIndexData - color indexing structure.
 *      colorList - list of colors.
 *      symbolIndexData - symbol list indexing structure.
 *      symbolList - list of symbols to render.
 *      symbolSizeIndexData - symbol size indexing structure.
 *      symbolSizeList - list of symbol sizes.
 *      clientData - clientDat
 *      color - color to draw the glyph with.
 *      glyphCache - the glyph cache tree.
 *      clientData - client data for the x and y accessing functions
 *	image - image to draw to.
 *      sym - default symbol to draw.
 *      symbolCache - the symbol cache for fast loading.
 *      fgColor - color to draw in if the color list is NULL.
 *
 **********************************************************************/

AmErrorCode 
AmSymbolDraw( AmDisplayClass		*display,	 
	      AmCoordClass		*coords,	 
	      AmRectangleRec const 	*drawRect,	 
	      uchar_t			*srcData,	 
	      int			srcSize,	 
	      int			length,		 
	      AmAccessFn		xFn,		 
	      int			xOffset,  	 
	      AmAccessFn		yFn,
	      int			yOffset,  	 
	      AmIndexDataPtr		colorIndexData,	 
	      AmColorCellPtr		colorList,	 
	      AmIndexDataPtr		symbolIndexData, 
	      char	        	**symbolList,	 
	      AmIndexDataPtr		symbolSizeIndexData,     
	      float			*symbolSizeList, 
	      void			*clientData,	 
	      AmImageClass		*image,		 
	      AmSymCachePtr		sym,		 
	      void			**symbolCache,	 
	      AmColorCellPtr		fgColor)
{
    AmRectangleRec 	screenRect;
    uchar_t		*dataEnd = srcData + length*srcSize, *p;
    char		*imageData = image->AmGetData();
    int			x, y, screenX, screenY, index;
    float		scale = 1.0, dpiScale = display->GetDPI()/BASEDPI;
    ulong_t		pixel;
    AnvilBoolean		drawFlag;

    if ( image == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* translate the bounding and expose rectangles from virtual to screen 
   coordindates */

    display->ScreenRect( coords->GetRect(), &screenRect);
    p = ( uchar_t *) &pixel;
    ASSIGN_PIXEL( *fgColor, p);

/* walk through the source data */

    for ( ; srcData < dataEnd; srcData += srcSize) {

/* scale and clip the x, y values */

	coords->GetScreenPt(( xFn)( srcData + xOffset, clientData),
			    ( yFn)( srcData + yOffset, clientData), &x, &y);
	if ( AmPtInRect( drawRect, x, y)) {
	    drawFlag = TRUE;

/* if the pixel function is non-NULL, set the new foreground color */

	    if ( colorList != NULL) {
		pixel = AmColormapPixelFn( srcData, colorIndexData, colorList);
	    }

/* if the symbol list is non-NULL, set the new symbol */

	    if ( symbolList != NULL) {
		index = ( symbolIndexData->getIndexFn)
			    ( srcData, symbolIndexData);
		if ( index < 0) {
		    if ( symbolIndexData->minClip) {
			drawFlag = FALSE;
		    } else {
			index = 0;
		    }
		} else if ( index >= symbolIndexData->length) {
		    if ( symbolIndexData->maxClip) {
			drawFlag = FALSE;
		    } else {
			index = symbolIndexData->length - 1;
		    }
		} 
		if ( drawFlag) {
		    sym = AmGetSymbol( symbolList[ index], symbolCache);
		    if ( sym == NULL) {
			continue;
		    }
		}
	    }

/* if the size list is non-NULL, set the symbol size */

	    if ( symbolSizeList != NULL) {
		index = ( symbolSizeIndexData->getIndexFn)
				( srcData, symbolSizeIndexData);
		if ( index < 0) {
		    if ( symbolSizeIndexData->minClip) {
			drawFlag = FALSE;
		    } else {
			index = 0;
		    }
		} else if ( index >= symbolSizeIndexData->length) {
		    if ( symbolSizeIndexData->maxClip) {
			drawFlag = FALSE;
		    } else {
			index = symbolSizeIndexData->length - 1;
		    }
		} 
		if ( drawFlag) {
		    scale = symbolSizeList[ index];
		}
	    }
	    if ( drawFlag && ( sym != NULL)) {

/* translate x,y from virtual to screen coordinates */

		screenX = display->ScreenCoord( x);
		screenY = display->ScreenCoord( y);
		AmScalePix(( ulong_t *) imageData, 
			  image->GetNumHorizPixels(), 
			  image->GetNumVerticalPixels(), 
			  sym->symbol, sym->symWidth, sym->symHeight, pixel, 
			  screenX - screenRect.x, screenY - screenRect.y, 
			  scale*dpiScale);
	    }
	}
    } 
    return( AmOK);
}

/**********************************************************************
 *
 * Description:
 *      scatter plot image drawing function using bilinear interpolation
 *	to smooth the symbols.
 * Arguments:
 *      display - display to draw to.
 *      coords - coordinate system to draw with.
 *      drawRect - bounding draw rectangle.
 *      srcData - source data to render.
 *      srcSize - size of each source data element.
 *      length - number of data elements.
 *      xFn, xOffset, yFn, yOffset - accessing functions and offsets
 *      for x,y values in the scatter points.
 *      colorIndexData - color indexing structure.
 *      colorList - list of colors.
 *      symbolIndexData - symbol list indexing structure.
 *      symbolList - list of symbols to render.
 *      symbolSizeIndexData - symbol size indexing structure.
 *      symbolSizeList - list of symbol sizes.
 *      clientData - clientDat
 *      color - color to draw the glyph with.
 *      glyphCache - the glyph cache tree.
 *      clientData - client data for the x and y accessing functions
 *	image - image to draw to.
 *      sym - default symbol to draw.
 *      symbolCache - the symbol cache for fast loading.
 *      fgColor - color to draw in if the color list is NULL.
 *
 **********************************************************************/

AmErrorCode 
AmBilinSymbolDraw( AmDisplayClass	*display,	 
		   AmCoordClass	 	*coords,	 
		   AmRectangleRec const *drawRect,	 
		   uchar_t		*srcData,	 
		   int		 	srcSize,	 
		   int		 	length,		 
		   AmAccessFn		xFn,		 
		   int		 	xOffset,
		   AmAccessFn		yFn,
		   int		 	yOffset,
		   AmIndexDataPtr	colorIndexData,	 
		   AmColorCellPtr	colorList,	 
		   AmIndexDataPtr	symbolIndexData, 
		   char	         	**symbolList,	 
		   AmIndexDataPtr	symbolSizeIndexData,     
		   float		*symbolSizeList, 
		   void		 	*clientData,	 
		   AmImageClass	 	*image,		 
		   AmSymCachePtr	sym,		 
		   void		 	**symbolCache,	 
		   AmColorCellPtr	fgColor)
{
    AmRectangleRec 	screenRect;
    uchar_t		*dataEnd = srcData + length*srcSize, *p;
    char		*imageData = image->AmGetData();
    int			x, y, index, screenX, screenY;
    float		scale = 1.0, dpiScale = display->GetDPI()/BASEDPI;
    ulong_t		pixel;
    AnvilBoolean		drawFlag;

    if ( image == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* translate the bounding and expose rectangles from virtual to screen
   coordindates */

    display->ScreenRect( coords->GetRect(), &screenRect);

/* preassign the foreground pixel if we're only drawing in one color */

    p = ( uchar_t *) &pixel;
    ASSIGN_PIXEL( *fgColor, p);

/* walk through the source data */

    for ( ; srcData < dataEnd; srcData += srcSize) {

/* scale and clip the x, y values */

	coords->GetScreenPt(( xFn)( srcData + xOffset, clientData),
			    ( yFn)( srcData + yOffset, clientData), &x, &y);
	if ( AmPtInRect( drawRect, x, y)) {
	    drawFlag = TRUE;

/* if the pixel function is non-NULL, set the new foreground color */

	    if ( colorList != NULL) {
		pixel = AmColormapPixelFn( srcData, colorIndexData, colorList);
		if ( pixel == ( ulong_t) -1) {
		    drawFlag = FALSE;
		}
	    }

/* if the symbol list is non-NULL, set the new symbol */

	    if ( symbolList != NULL) {
		index = ( symbolIndexData->getIndexFn)
			    ( srcData, symbolIndexData);
		if ( index < 0) {
		    if ( symbolIndexData->minClip) {
			drawFlag = FALSE;
		    } else {
			index = 0;
		    }
		} else if ( index >= symbolIndexData->length) {
		    if ( symbolIndexData->maxClip) {
			drawFlag = FALSE;
		    } else {
			index = symbolIndexData->length - 1;
		    }
		} 
		if ( drawFlag) {
		    sym = AmGetSymbol( symbolList[ index], symbolCache);
		}
	    }

/* if the size list is non-NULL, set the symbol size */

	    if ( symbolSizeList != NULL) {
		index = ( symbolSizeIndexData->getIndexFn)
				( srcData, symbolSizeIndexData);
		if ( index < 0) {
		    if ( symbolSizeIndexData->minClip) {
			drawFlag = FALSE;
		    } else {
			index = 0;
		    }
		} else if ( index >= symbolSizeIndexData->length) {
		    if ( symbolSizeIndexData->maxClip) {
			drawFlag = FALSE;
		    } else {
			index = symbolSizeIndexData->length - 1;
		    }
		} 
		if ( drawFlag) {
		    scale = symbolSizeList[ index];
		}
	    }
	    if ( drawFlag && ( sym != NULL)) {

/* translate x,y from virtual to screen coordinates */

                screenX = display->ScreenCoord( x);
                screenY = display->ScreenCoord( y);
		AmBilinPix(( ulong_t *) imageData, 
			  image->GetNumHorizPixels(), 
			  image->GetNumVerticalPixels(), 
			  sym->symbol, sym->symWidth, sym->symHeight, pixel, 
			  screenX - screenRect.x, screenY - screenRect.y, 
			  scale*dpiScale);
	    }
	}
    } 
    return( AmOK);
}

/**********************************************************************
 *
 * Description:
 *      scatter plot image drawing function.
 * Arguments:
 *      display - display to draw to.
 *      coords - coordinate system to draw with.
 *      drawRect - bounding draw rectangle.
 *      srcData - source data to render.
 *      srcSize - size of each source data element.
 *      length - number of data elements.
 *      xFn, xOffset, yFn, yOffset - accessing functions and offsets
 *      for x,y values in the scatter points.
 *      colorIndexData - color indexing structure.
 *      colorList - list of colors.
 *      symbolIndexData - symbol list indexing structure.
 *      symbolList - list of symbols to render.
 *      symbolSizeIndexData - symbol size indexing structure.
 *      symbolSizeList - list of symbol sizes.
 *      clientData - clientDat
 *      color - color to draw the glyph with.
 *      glyphCache - the glyph cache tree.
 *      clientData - client data for the x and y accessing functions
 *	image - image to draw to.
 *      glyph - default glyph to draw.
 *      glyphCache - the glyph cache for fast loading.
 *      fgColor - color to draw in if the color list is NULL.
 *	interpolate - whether to smooth the symbols when they are drawn.
 *
 **********************************************************************/

AmErrorCode 
AmDrawSymbolImage( AmDisplayClass	 	*display,	 
		 AmCoordClass	 	*coords,	 
		 AmRectangleRec const 	*drawRect,	 
		 uchar_t		*srcData,	 
		 int		 	srcSize,	 
		 int		 	length,		 
		 AmAccessFn		xFn,		 
		 int		 	xOffset,
		 AmAccessFn		yFn,
		 int		 	yOffset,
		 AmIndexDataPtr	 	colorIndexData,	 
		 AmColorCellPtr	 	colorList,	 
		 AmIndexDataPtr	 	symbolIndexData, 
		 char	         	**symbolList,	 
		 AmIndexDataPtr	 	symbolSizeIndexData,     
		 float		 	*symbolSizeList, 
		 void		 	*clientData,	 
		 AmImageClass	 	*image,		 
		 AmSymCachePtr	 	sym,		 
		 void		 	**symbolCache,	 
		 AmColorCellPtr	 	fgColor,
		 AnvilBoolean		interpolate)
{
    if ( interpolate) {
	return( AmBilinSymbolDraw( 
			display, coords, drawRect, srcData,	 
		        srcSize, length, xFn, xOffset, yFn, yOffset,
			colorIndexData, colorList,
			symbolIndexData, symbolList, symbolSizeIndexData,     
			symbolSizeList, clientData, image, sym, 
			symbolCache, fgColor));
    } else {
	return( AmSymbolDraw( 
		    display, coords, drawRect, srcData,	 
		    srcSize, length, xFn, xOffset, yFn, yOffset, 
		    colorIndexData, colorList,	
		    symbolIndexData, symbolList, symbolSizeIndexData,     
		    symbolSizeList, clientData, image, sym, 
		    symbolCache, fgColor));
    }
}
