#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#if !defined(__BSD44__)
#include <search.h>
#endif
#include <anvil/image.h>
#include <anvil/files.h>
#include <anvil/envdefs.h>
#include <anvil/display.h>
#include <anvil/pixthing.h>
#include <anvil/imageP.h>
#include <anvil/tsearch.h>
#include <anvil/macros.h>
#include <anvil/scatter.h>
#include <anvil/colorcell.h>

/* functions for drawing scatter symbols as a set of glyphs, which are
   drawn similar to raster data */

static float scaleTol = 1.0/300.0;

/**********************************************************************
 * 
 * Description:
 * 	compare glyphs based on symbol name, color, and size.  This
 *	is used for searching the glyph cache.
 * Arguments:
 *	a, b - glyphs to compare.
 * Returns:
 *	sort value based on name, color, and size.
 *
 **********************************************************************/

static int 
compareGlyphAndSize( AmGlyphCachePtr	a,
		     AmGlyphCachePtr	b)
{
    int		diffName, diffColor;
    float	diffScale;

    diffName = strcmp( a->symfile, b->symfile);
    if ( diffName == 0) {
	diffColor = ( int) a->color - ( int) b->color;
	if ( diffColor == 0) {
	    diffScale = a->scale - b->scale;
	    if ( diffScale < -scaleTol) {
		return( -1);
	    } else if ( diffScale > scaleTol) {
		return( 1);
	    } else {
		return( 0);
	    }
	} else {
	    return( diffColor);
	}
    } else {
	return( diffName);
    }
}

/**********************************************************************
 * 
 * Description:
 * 	compare glyphs based on symbol name and color 
 * Arguments:
 *	a, b - glyphs to compare.
 * Returns:
 *	sort value based on name and color.
 *
 **********************************************************************/

static int 
compareGlyph( AmGlyphCachePtr	a,
	      AmGlyphCachePtr	b)
{
    int		diffName;

    diffName = strcmp( a->symfile, b->symfile);
    if ( diffName == 0) {
	return(( int) a->color - ( int) b->color);
    } else {
	return( diffName);
    }
}

/**********************************************************************
 * 
 * Description:
 * 	get or load a glyph from the cache, depending on glyph name, 
 *	scale, and color 
 * Arguments:
 *	display - display to draw to.
 *	window - window to draw to.
 *	scale - scale to apply to the glyph.
 *	color - color to draw the glyph with.
 *	glyphCache - the glyph cache tree.
 * Returns:
 * 	either the pointer to the glyph in the cache, or the loaded
 *	glyph.
 *
 **********************************************************************/

AmGlyphCachePtr 
AmGetGlyph( AmDisplayClass	*display,
	  AmWindowClass	*window,
	  char		*symfile,
	  float		scale,
	  ulong_t	color,
	  void		**glyphCache)
{
    AmGlyphCacheRec	target;
    AmGlyphCachePtr	sym, *foundSym;
    AmGlyphClass		*glyph;
    char		relPath[ PATHLEN], *sympath;
    int			width, height;

/* see if exactly the glyph we want is in the cache, sized and everything... */

    target.symfile = symfile;
    target.scale = scale;
    target.color = color;
    foundSym = ( AmGlyphCachePtr *) tfind( &target, 
					 ( const void **) glyphCache,
				         ( AmSearchFn) compareGlyphAndSize);

/* if it's not sized properly, then see if there's one with a matching name and
   color, which we have to store in the glyph cache */ 

    if ( foundSym == NULL) {

/* create the structure for the glyph cache */

	if (( sym = ( AmGlyphCachePtr) malloc( sizeof( AmGlyphCacheRec))) == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    return( NULL);
	} 

/* assign the symbol file scale and color */

	if (( sym->symfile = strdup( symfile)) == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    return( NULL);
	}
	sym->scale = scale;
	sym->color = color;

/* search the cache */

	foundSym = ( AmGlyphCachePtr *) tfind( &target, 
					     ( const void **) glyphCache,
					     ( AmSearchFn) compareGlyph);
	if ( foundSym != NULL) {
	    if (( sym = ( AmGlyphCachePtr) 
			malloc( sizeof( AmGlyphCacheRec))) == NULL) {
		DO_ERROR( AmMALLOC_FAILED);
		return( NULL);
	    } 
	    if (( sym->symfile = strdup( symfile)) == NULL) {
		DO_ERROR( AmMALLOC_FAILED);
		return( NULL);
	    }
	    width = (int)(( *foundSym)->glyph->GetWidth()*scale);
	    height = (int)(( *foundSym)->glyph->GetHeight()*scale);
	    sym->glyph = ( *foundSym)->glyph->ResizeGlyph( width, height);
	} else {
	    sprintf( relPath, "%s/%s", AM_SYMBOL_DIR, symfile);
	    sympath = AmGetEnvFile( CONFIG_ENV, relPath);
	    if ( sympath == NULL) {
		DO_ERROR( AmOPEN_FAILED);
		return( NULL);
	    } else {

/* read the glyph. If the scale isn't 1:1, then resize it */

		glyph = display->CreateGlyph( window, sympath, color);
		if ( scale != 1.0) {
		    width = (int)(glyph->GetWidth()*scale);
		    height = (int)(glyph->GetHeight()*scale);
		    sym->glyph = glyph->ResizeGlyph( width, height);
		    delete glyph;
		} else {
		    sym->glyph = glyph;
		}
	    }
	}
	tsearch( sym, glyphCache, ( AmSearchFn) compareGlyphAndSize);
	return( sym);
    } else {
	return( *foundSym);
    }
}


/**********************************************************************
 * 
 * Description:
 * 	glyph scatter plot drawing function.
 * Arguments:
 *	display - display to draw to.
 *	window - window to draw to.
 *	coords - coordinate system to draw with.
 *	drawRect - bounding draw rectangle.
 *	srcData - source data to render.
 *	srcSize - size of each source data element.
 *	length - number of data elements.
 *	xFn, xOffset, yFn, yOffset - accessing functions and offsets
 *	for x,y values in the scatter points.
 *	colorIndexData - color indexing structure.
 *	colorList - list of colors.
 *	symbolIndexData - symbol list indexing structure.
 *	symbolList - list of symbols to render.
 *	symbolSizeIndexData - symbol size indexing structure.
 *	symbolSizeList - list of symbol sizes.
 *	clientData - clientDat
 *	color - color to draw the glyph with.
 *	glyphCache - the glyph cache tree.
 *	clientData - client data for the x and y accessing functions
 *	glyph - default glyph to draw.
 *	glyphCache - the glyph cache for fast loading.
 *	fgColor - color to draw in if the color list is NULL.
 *
 **********************************************************************/

AmErrorCode 
AmGlyphDraw( AmDisplayClass		*display,
	     AmWindowClass		*window,
	     AmCoordClass		*coords,
	     AmRectangleRec const	*drawRect,
	     uchar_t			*srcData,
	     int			srcSize,
	     int			length,
	     AmAccessFn       		xFn,  
	     int			xOffset,
	     AmAccessFn       		yFn,
	     int			yOffset,
	     AmIndexDataPtr   		colorIndexData, 
	     AmColorCellPtr   		colorList,
	     AmIndexDataPtr   		symbolIndexData,
	     char           		**symbolList,
	     AmIndexDataPtr   		symbolSizeIndexData,
	     float          		*symbolSizeList,
	     void           		*clientData,
	     AmGlyphCachePtr    	glyph,
	     void           		**glyphCache,
	     AmColorCellPtr   		fgColor)
{
    uchar_t		*dataEnd = srcData + length*srcSize;
    ulong_t		pixel;
    int			x, y, index;
    AnvilBoolean		drawFlag;
    float		scale = 1.0;
    char		*symfile = glyph->symfile;
    AmGlyphCachePtr	sym;

/* set the scale tolerance for scale comparisons */

    scaleTol = 1.0/display->GetDPI();

/* assign the pixel */

    pixel = fgColor->pixel;

/* walk through the source data */

    for ( ; srcData < dataEnd; srcData += srcSize) {

/* scale and clip the x,y values */

	coords->GetScreenPt(( xFn)( srcData + xOffset, clientData),
			    ( yFn)( srcData + yOffset, clientData), &x, &y);
	if ( AmPtInRect( drawRect, x, y)) {
	    drawFlag = TRUE;

/* if the pixel function is non-NULL, set the new foreground color */

            if ( colorList != NULL) {
                pixel = AmColormapPixelFn( srcData, colorIndexData, colorList);
                if ( pixel == ( ulong_t) -1) {
                    drawFlag = FALSE;
                }
            }

/* if the size list is non-NULL, set the symbol size */

            if ( symbolSizeList != NULL) {
                index = ( symbolSizeIndexData->getIndexFn)
                                ( srcData, symbolSizeIndexData);
                if ( index < 0) {
                    if ( symbolSizeIndexData->minClip) {
                        drawFlag = FALSE;
                    } else {
                        index = 0;
                    }
                } else if ( index >= symbolSizeIndexData->length) {
                    if ( symbolSizeIndexData->maxClip) {
                        drawFlag = FALSE;
                    } else {
                        index = symbolSizeIndexData->length - 1;
                    }
                }
                if ( drawFlag) {
                    scale = symbolSizeList[ index];
                }
            }

/* if the symbol list is non-NULL, set the new symbol */

            if ( symbolList != NULL) {
                index = ( symbolIndexData->getIndexFn)
                            ( srcData, symbolIndexData);
                if ( index < 0) {
                    if ( symbolIndexData->minClip) {
                        drawFlag = FALSE;
                    } else {
                        index = 0;
                    }
                } else if ( index >= symbolIndexData->length) {
                    if ( symbolIndexData->maxClip) {
                        drawFlag = FALSE;
                    } else {
                        index = symbolIndexData->length - 1;
                    }
                }
                if ( drawFlag) {
                    symfile = symbolList[ index];
                }
            }

/* retreive the glyph from the cache and draw it */

	    sym = AmGetGlyph( display, window, symfile, scale, 
			      pixel, glyphCache);
	    if ( sym != NULL) {
		sym->glyph->DrawGlyph( display, window, 
				       x - sym->glyph->GetWidth()/2,
				       y - sym->glyph->GetHeight()/2);
	   }
	}
    }
    return( AmOK);
}
