#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


/* ScatterGraph class member function definitions.  Scatter Graphs are 
   similar to scatter plots, except that they allow different x and y 
   axes, and use the heightIndex index structure for the y axis */

#include <malloc.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <sys/types.h>
#include <anvil/resource.h>
#include <anvil/strlib.h>
#include <anvil/mem_funcs.h>
#include <anvil/macros.h>
#include <anvil/overlay.h>
#include <anvil/image.h>
#include <anvil/datafmt.h>
#include <anvil/scattergraph.h>
#include <anvil/defaults.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmScatterGraphClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmSCATTER_PLOT) {
        return( TRUE);
    } else {
        return( AmGraphClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a scatter graph.  Call the
 *      basic graph class initializations, then initialize the color, symbol
 *      and symbol list related fields.
 * Arguments:
 *      windowArg - window that the scatter graph will be drawn on.
 *      nameArg - string name of the scatter graph, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmScatterGraphClass::initializeFields( AmWindowClass	*windowArg,
				     const char		*nameArg)
{
    AmGraphClass::initializeFields( windowArg, nameArg);
    interpolate = FALSE;
    displayImage = NULL;
    objectType = AmSCATTER_GRAPH;
    symbolCache = NULL;
    glyphCache = NULL;
    glyph = NULL;
    symfile = strdup( "x");
    sym = AmGetSymbol( "x", &symbolCache);
    AmInitIndexData( &symbolIndex);
    AmInitIndexData( &symbolSizeIndex);
    AmInitIndexData( &colorIndex);
    colorIndex.offset = sizeof( AmPointRec);
    symbolList = NULL;
    symbolSizeList = NULL;
    scatterBufSize = AmSCATTERBUFSIZE;
    dataInfo.xFn = ( AmAccessFn) AmFloatFn;
    dataInfo.xOffset = 0;
    dataInfo.yFn = ( AmAccessFn) AmFloatFn;
    dataInfo.yOffset = sizeof( float);
}
/****************************************************************************
 *
 * Description:
 *      check values for a scatter graph.  Each check values function 
 *	calls the check values function for the parent class and is used 
 *	to set the object's IsValid() flag. 
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmScatterGraphClass::checkValues()
{
    if ( !AmGraphClass::checkValues()) {
        return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmScatterGraphClass constructor.  Creates a scatter graph using the 
 *	window that it will be drawn on and the name, followed by 
 *	NULL-terminated tag-value pairs for setting the fields in the 
 *	scatter graph.
 * Arguments:
 *      windowArg - window that the scatter graph will be drawn on.
 *      nameArg - string name of the scatter graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmScatterGraphClass::AmScatterGraphClass( AmWindowClass  *windowArg,
				      const char   *nameArg, 
				      ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* set the general fields for a plot, and initialize the fields specific
   to a scatter plot */

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( IsValid() && ( dataInfo.sourceType != AmNO_SOURCE));
}

/****************************************************************************
 *
 * Description:
 *      do-nothing default constructor, so scatter plots can be subclassed.
 *
 ****************************************************************************/

AmScatterGraphClass::AmScatterGraphClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmScatterGraphClass constructor.  Creates a scatter graph using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the scatter graph.
 * Arguments:
 *      windowArg - window that the scatter graph will be drawn on.
 *      nameArg - string name of the scatter graph, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmScatterGraphClass::AmScatterGraphClass( AmOverlayClass *overlayArg,
				      const char   *nameArg,
				      ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    objectType = AmSCATTER_GRAPH;
    count = setFieldsWrapper( tagValues, count);
    SetValid( IsValid() && ( dataInfo.sourceType != AmNO_SOURCE));
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }

}

/****************************************************************************
 *
 * Description:
 *      AmScatterGraphClass destructor: destroy the symbolIndex, colorIndex 
 *      and symbolSizeIndex indexData structures, the symbol size list
 *      and the symbol list if they've been allocated.  If the plot's
 *      standalone, then free the display image if it's been allocated,
 *      then free the scatter plot's symbol cache.
 *
 ****************************************************************************/


AmScatterGraphClass::~AmScatterGraphClass()
{
    AmDestroyIndexData( &symbolIndex);
    AmDestroyIndexData( &symbolSizeIndex);
    AmDestroyIndexData( &colorIndex);
    if ( symbolSizeList != NULL) {
	free( symbolSizeList);
    }
    if ( symbolList != NULL) {
	AmFreeStrVec( symbolList);
    }
    if ( parent == NULL) {
	if ( displayImage != NULL) {
	    delete displayImage;
	} 
    }
    if ( symfile != NULL) {
	free( symfile);
    }

/* clean up the symbol cache */

    AmFreeSymbolCache( &symbolCache);
}

/****************************************************************************
 *
 * Description:
 *      draw the current selection set as glyphs, which are drawn on
 *      top of the display, rather than directly into the image.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 *	fHighlight - controls whether objects are drawn in the highlight
 *	color or their normal color.
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the scatter graphs is invalid.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterGraphClass::DrawSelection( AmWindowClass         	*window,
				  AmGCClass             	*gc,
				  AmRectangleRec const  	*exposeRects,
				  int		  	nRects,
				  AnvilBoolean		fHighlight)
{
    uchar_t             *selectionData;
    int                 nSel;
    AmColorCellPtr	color;
    AmRectangleRec	surroundRect;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* get the selection */

    GetSelection(( void **) &selectionData, &nSel);
    AmUnionRectVec( exposeRects, nRects, &surroundRect);
    if ( nSel > 0) {

/* if the base glyph hasn't been loaded, then restore it from the cache */

        if ( glyph == NULL) {
            glyph = AmGetGlyph( display, window, symfile, 1.0,
                              highlightColor.pixel, &glyphCache);
        }
	if ( fHighlight) {
	    color = &highlightColor;
	} else {
	    color = &fgColor;
	}
        AmGlyphDraw( display, window, ( AmCoordClass *) this, &surroundRect,
                   selectionData, dataInfo.size, nSel,
                   dataInfo.xFn, dataInfo.xOffset,
                   dataInfo.yFn, dataInfo.yOffset,
                   &colorIndex, NULL, &symbolIndex, symbolList,
                   &symbolSizeIndex, symbolSizeList,
                   dataInfo.clientData, glyph, &glyphCache, color);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a scatter graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.   Since the bin vectors for the symbol list
 *      and symbolSize list are 1 less than the length of the vectors,
 *      they're set by hand.  The symbolIndex, colorIndex, and symbolSizeIndex
 *      structures are initialized after the fields are set.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterGraphClass::setFields( AmTagValuePtr	tagValues,
			      int		numValues)
{
    int         numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr fieldDefs;

/* set the fields for the graph class parent object,  and set the color index 
   length.  Note it has to be set before colorsBinvec and colorsTable, so 
   they get the proper length */

    numValues = AmGraphClass::setFields( tagValues, numValues);
    colorIndex.length = numColors;
    AmScatterGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolBinvec)) {
	    AmSetBinvec( &symbolIndex.binvec, symbolIndex.length, 
		       &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
	    AmSetBinvec( &symbolSizeIndex.binvec, symbolSizeIndex.length, 
		       &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNcolorsBinvec)) {
	    AmSetBinvec( &colorIndex.binvec, colorIndex.length, 
		       &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmScatterGraphClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmGraphClass::updateFields();

/* set the index data client struct for the symbol list and symbol size list */
   
    if ( IsValid()) {
	if ( symbolList != NULL) {
	    if ( updateIndexData( &symbolIndex, "symbol") != AmOK) {
		AmAddResourceToLog( "failedToInitializeSymbolIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( symbolSizeList != NULL) {
	    if ( updateIndexData( &symbolSizeIndex, "symbolSize") != AmOK) {
		AmAddResourceToLog( "failedToInitializeSymbolSizeIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "color") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}


/* set the yFn to be the same as the heightIndex function */

	dataInfo.yFn = heightIndex.fn;

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }

/* read the symbol file */

	if ( AmChanged( &symfile)) {
	    sym = AmGetSymbol( symfile, &symbolCache);
	}
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a scatter graph, given the tag-values stored in
 *      a tag values array.  First, the generic graph fields are retrived, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The symbol list and symbol size list are
 *      retrieved by hand.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterGraphClass::getFields( AmTagValuePtr	tagValues,
			      int		numValues)
{
    int         numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmGraphClass::getFields( tagValues, numValues);
    AmScatterGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolBinvec)) {
	    if ( symbolIndex.binvec != NULL) {
		tagValues[ ii].value = 
		    AmCopyVec( symbolIndex.binvec, 
			     ( symbolIndex.length - 1)*sizeof( float));
	    } else {
		tagValues[ ii].value = NULL;
	    }
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
	    if ( symbolSizeIndex.binvec != NULL) {
		tagValues[ ii].value = 
		    AmCopyVec( symbolSizeIndex.binvec, 
			     ( symbolSizeIndex.length - 1)*sizeof( float));
	    } else {
		tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      for scatter graph update, redraw the scatter graph into the
 *      image buffer, after allocating a new image (if it's not an overlay
 *      plot, where the resize is already done)
 * Arguments:
 *      eventCodes - vector of event codes from AmSendUserEvent()
 *      numEventCodes - number of event codes.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterGraphClass::Update( int	*eventCodes,
			   int	numCodes)
{
    AmRectangleRec const  *dispRect = GetRect();
    AmRectangleRec	clearRect;
    int			ii, numel;
    uchar_t		*scatterBuffer;

/* if this object is invalid return */

    if ( !IsValid()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }

/* if the image is out of sync, then resize it, and reset the rectangle
   to the screen extents */

    if ( parent == NULL) {
	if (( displayImage = ResizeDisplayImage( displayImage)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    } else { 
	displayImage = (( AmOverlayClass *) parent)->GetImage();
    }

/* clear out the updated rectangle */

    clearRect = *dispRect;
    clearRect.x = 0;
    clearRect.y = 0;
    displayImage->ClearImageRect( &bgColor, &clearRect);
	
/* only draw the scatter plot if the coordinates have been set as well as
   the rectangle */
 
    if ( CoordsSet() && RectSet()) {
	if ( dataInfo.sourceType == AmREAD_SOURCE) {
	    if (( scatterBuffer = 
		( uchar_t *) malloc( dataInfo.size*scatterBufSize)) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	    if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
		DO_RETURN_ERROR( AmSEEK_FAILED);
	    }
	    for ( ii = 0; ii < dataInfo.dims[ 0]; ii += scatterBufSize) {
		numel = MIN( scatterBufSize, dataInfo.dims[ 0] - ii);
		if ( read( dataInfo.fd, scatterBuffer, 
			   dataInfo.size*numel) < dataInfo.size*numel) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		AmDrawSymbolImage( display, ( AmCoordClass *) this, dispRect,
				 scatterBuffer, dataInfo.size, scatterBufSize,
				 dataInfo.xFn, dataInfo.xOffset,
				 heightIndex.fn, heightIndex.offset,
				 &colorIndex, colors,
				 &symbolIndex, symbolList,
				 &symbolSizeIndex, symbolSizeList,
				 dataInfo.clientData,
				 displayImage, sym, &symbolCache, &fgColor,
				 interpolate);
	    }
	    free( scatterBuffer);
	} else {
	    AmDrawSymbolImage( display, ( AmCoordClass *) this, dispRect,
			     dataInfo.data, dataInfo.size, dataInfo.dims[ 0], 
			     dataInfo.xFn, dataInfo.xOffset,
			     heightIndex.fn, heightIndex.offset,
			     &colorIndex, colors,
			     &symbolIndex, symbolList,
			     &symbolSizeIndex, symbolSizeList,
			     dataInfo.clientData,
			     displayImage, sym, &symbolCache, &fgColor,
			     interpolate);
	} 
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw the scatter graph.  just blit the image to the screen if 
 *	we're not part of an overlay.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, which it is.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterGraphClass::draw( AmWindowClass		*window,
			 AmGCClass		*gc, 
			 AmRectangleRec const 	*exposeRects,
			 int			nRects)
{
    AmRectangleRec const	*dispRect = GetRect();
    int			ii;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* since the image has a 0,0 origin, we shift the expose rectangle from the
   window coordinate system to the image coordinate system 
   if we're standalone, draw the image, otherwise, it will get drawn by the
   overlay plot */

    if ( parent == NULL) {	
	for ( ii = 0; ii < nRects; ii++) {
	    display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
				exposeRects[ ii].y - dispRect->y, 
				exposeRects[ ii].x, exposeRects[ ii].y,
				exposeRects[ ii].width, exposeRects[ ii].height,
				displayImage);
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Return the extents of a scatter point.  This is used for moving
 *      and selecting points.
 * Arguments:
 *      ptr - data point.
 *      xmin, ymin, xmax, ymax: location and extents of the point.
 *
 ****************************************************************************/

void 
AmScatterGraphClass::ObjectExtents( void 	*ptr,
				  float	*xmin,
				  float	*ymin,
				  float	*xmax,
				  float	*ymax)
{
    AmCoordPtr	point = ( AmCoordPtr) ptr;

    *xmin = point->x;
    *ymin = point->y;
    *xmax = point->x;
    *ymax = point->y;
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a scatter graph.  This is
 *      used in the map editor to dump the graph's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmScatterGraphClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			       AnvilBoolean		 allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( plot) */

    numStrings = AmGraphClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmScatterGraphClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);
    if ( symbolIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNsymbolBinvec;
	fieldLengths[ numStrings++].length = symbolIndex.length - 1;
    }
    if ( symbolSizeIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNsymbolSizeBinvec;
	fieldLengths[ numStrings++].length = symbolSizeIndex.length - 1;
    }
    return( numStrings);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a scatter plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmScatterGraphClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields. THIS MUST BE IN ALPHABETICAL
 *      ORDER ON THE FIELD NAME FOR SEARCHING
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmScatterGraphClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					int		*numDefs)
{
    static AmFieldDefRec ScatterGraphClassFieldDefinitions[] = {
	{ AmNcolorsClientData, offsetof( AmScatterGraphClass, 
					 colorIndex.clientData) },
	{ AmNcolorsClientDataFreeFn, offsetof( AmScatterGraphClass, 
					       colorIndex.clientDataFreeFn) },
	{ AmNcolorsFn, offsetof( AmScatterGraphClass, colorIndex.fn) },
	{ AmNcolorsIntFn, offsetof( AmScatterGraphClass, colorIndex.AmIntFn) },
	{ AmNcolorsMax, offsetof( AmScatterGraphClass, colorIndex.max) },
	{ AmNcolorsMin, offsetof( AmScatterGraphClass, colorIndex.min) },
	{ AmNcolorsName, offsetof( AmScatterGraphClass, colorIndex.name), 
			 0, AmDYNAMIC },
	{ AmNcolorsOffset, offsetof( AmScatterGraphClass, colorIndex.offset) },
	{ AmNcolorsStrings, offsetof( AmScatterGraphClass, colorIndex.strings),
			    offsetof( AmScatterGraphClass, colorIndex.length),
			    AmDYNAMIC },
	{ AmNcolorsTable, offsetof( AmScatterGraphClass, colorIndex.table),
			  offsetof( AmScatterGraphClass, colorIndex.length),
			  AmDYNAMIC },
	{ AmNcolorsType, offsetof( AmScatterGraphClass, colorIndex.dataType) },
	{ AmNcolorsUnits, offsetof( AmScatterGraphClass, colorIndex.units) },
	{ AmNinterpolate, offsetof( AmScatterGraphClass, interpolate) },
	{ AmNsymbol, offsetof( AmScatterGraphClass, symfile), 0, AmDYNAMIC },
        { AmNsymbolClientData,
          offsetof( AmScatterGraphClass, symbolIndex.clientData) },
        { AmNsymbolClientDataFreeFn,
          offsetof( AmScatterGraphClass, symbolIndex.clientDataFreeFn) },
        { AmNsymbolFn, offsetof( AmScatterGraphClass, symbolIndex.fn) },
        { AmNsymbolIntFn, offsetof( AmScatterGraphClass, symbolIndex.AmIntFn) },
        { AmNsymbolLength, offsetof( AmScatterGraphClass, symbolIndex.length) },
        { AmNsymbolList, offsetof( AmScatterGraphClass, symbolList),
			 offsetof( AmScatterGraphClass, symbolIndex.length),
			 AmDYNAMIC },
        { AmNsymbolMax,
          offsetof( AmScatterGraphClass, symbolIndex.max) },
        { AmNsymbolMin,
          offsetof( AmScatterGraphClass, symbolIndex.min) },
        { AmNsymbolName,
          offsetof( AmScatterGraphClass, symbolIndex.name), 0, AmDYNAMIC },
        { AmNsymbolOffset,
          offsetof( AmScatterGraphClass, symbolIndex.offset) },
        { AmNsymbolSizeClientData,
          offsetof( AmScatterGraphClass, symbolSizeIndex.clientData) },
        { AmNsymbolSizeClientDataFreeFn,
          offsetof( AmScatterGraphClass, symbolSizeIndex.clientDataFreeFn) },
        { AmNsymbolSizeFn, offsetof( AmScatterGraphClass, symbolSizeIndex.fn) },
        { AmNsymbolSizeIntFn, 
	  offsetof( AmScatterGraphClass, symbolSizeIndex.AmIntFn) },
        { AmNsymbolSizeLength, 
  	  offsetof( AmScatterGraphClass, symbolSizeIndex.length) },
        { AmNsymbolSizeList, 
	      offsetof( AmScatterGraphClass, symbolSizeList),
	      offsetof( AmScatterGraphClass, symbolSizeIndex.length),
	      AmDYNAMIC },
        { AmNsymbolSizeName,
          offsetof( AmScatterGraphClass, symbolSizeIndex.name), 0, AmDYNAMIC },
        { AmNsymbolSizeOffset,
          offsetof( AmScatterGraphClass, symbolSizeIndex.offset) },
        { AmNsymbolSizeType,
          offsetof( AmScatterGraphClass, symbolSizeIndex.dataType) },
        { AmNsymbolSizeMax,
          offsetof( AmScatterGraphClass, symbolSizeIndex.max) },
        { AmNsymbolSizeMin,
          offsetof( AmScatterGraphClass, symbolSizeIndex.min) },
        { AmNsymbolSizeStrings, 
	  offsetof( AmScatterGraphClass, symbolSizeIndex.strings),
	  offsetof( AmScatterGraphClass, symbolSizeIndex.length),
	  AmDYNAMIC },
        { AmNsymbolSizeTable, 
	      offsetof( AmScatterGraphClass, symbolSizeIndex.table),
	      offsetof( AmScatterGraphClass, symbolSizeIndex.length),
	      AmDYNAMIC },
        { AmNsymbolSizeUnits, 
	      offsetof( AmScatterGraphClass, symbolSizeIndex.units) },
        { AmNsymbolStrings, offsetof( AmScatterGraphClass, symbolIndex.strings),
			  offsetof( AmScatterGraphClass, symbolIndex.length),
			  AmDYNAMIC },
        { AmNsymbolTable, offsetof( AmScatterGraphClass, symbolIndex.table),
			  offsetof( AmScatterGraphClass, symbolIndex.length),
			  AmDYNAMIC },
        { AmNsymbolType, offsetof( AmScatterGraphClass, symbolIndex.dataType) },
        { AmNsymbolUnits, offsetof( AmScatterGraphClass, symbolIndex.units) },
    };
    *fieldDefs = ScatterGraphClassFieldDefinitions;
    *numDefs = VECLEN( ScatterGraphClassFieldDefinitions);
}
