#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#if defined(_WIN32) || defined(__BSD44__)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <sys/types.h>
#include <anvil/resource.h>
#include <anvil/strlib.h>
#include <anvil/mem_funcs.h>
#include <anvil/macros.h>
#include <anvil/overlay.h>
#include <anvil/image.h>
#include <anvil/datafmt.h>
#include <anvil/scatterplot.h>
#include <anvil/defaults.h>
#include <anvil/compression.h>
#include <anvil/log.h>

/* scatter plot class member function definitions. */

static const int MAXTILES = 1024;

/****************************************************************************
 *
 * Description:
 *      Recursive type function which tests the object type against the
 *      passed type and recursively tests the parent class's type.
 * Arguments:
 *      type: object type for testing.
 * Returns:
 *      TRUE if object class or any of its ancestors is of requested type,
 *      FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmScatterPlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmSCATTER_PLOT) {
        return( TRUE);
    } else {
        return( AmPlotClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a scatter plot.  Call the
 *      basic plot class initializations, then initialize the symbol
 *	and symbol list related fields.
 * Arguments:
 *      windowArg - window that the scatter plot will be drawn on.
 *      nameArg - string name of the scatter plot, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmScatterPlotClass::initializeFields( AmWindowClass	*windowArg,
				    const char	*nameArg)
{
    AmPlotClass::initializeFields( windowArg, nameArg);
    interpolate = FALSE;
    colorIndex.offset = sizeof( AmPointRec);
    displayImage = NULL;
    objectType = AmSCATTER_PLOT;
    symbolCache = NULL;
    glyphCache = NULL;
    glyph = NULL;
    symfile = strdup("x");
    sym = AmGetSymbol("x", &symbolCache);
    AmInitIndexData( &symbolIndex);
    AmInitIndexData( &symbolSizeIndex);
    symbolList = NULL;
    symbolSizeList = NULL;
    scatterBufSize = AmSCATTERBUFSIZE;
    dataInfo.xFn = ( AmAccessFn) AmFloatFn;
    dataInfo.xOffset = 0;
    dataInfo.yFn = ( AmAccessFn) AmFloatFn;
    dataInfo.yOffset = sizeof( float);
}

/****************************************************************************
 *
 * Description:
 *      check values for a scatter plot.  Each check values function calls
 *      the check values function for the parent class and is used to set
 *      the object's IsValid() flag.  The check values function writes
 *      details about the errors to the log file.
 * Returns:
 *      TRUE if everything's OK, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean
AmScatterPlotClass::checkValues()
{
    if ( !AmPlotClass::checkValues()) {
	return( FALSE);
    } else {
	return( TRUE);
    }
}

/****************************************************************************
 *
 * Description:
 * 	do-nothing default constructor, so scatter plots can be subclassed. 
 *
 ****************************************************************************/

AmScatterPlotClass::AmScatterPlotClass()
{
}

/****************************************************************************
 *
 * Description:
 *      AmScatterPlotClass constructor.  Creates a scatter plot using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the scatter plot.
 * Arguments:
 *      windowArg - window that the scatter plot will be drawn on.
 *      nameArg - string name of the scatter plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmScatterPlotClass::AmScatterPlotClass( AmWindowClass  *windowArg,
				    const char	 *nameArg, 
				    ...) 
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* set the general fields for a plot, and initialize the fields specific
   to a scatter plot */

    initializeFields( windowArg, nameArg);
    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( IsValid() && ( dataInfo.sourceType != AmNO_SOURCE));
}

/****************************************************************************
 *
 * Description:
 *      AmScatterPlotClass constructor.  Creates a scatter plot as part of an
 *      overlay.  The proprerties for the scatter plot are set by a
 *      NULL-terminated vector of tag-value pairs.
 * Arguments:
 *      overlayArg - overlay that the scatter plot will be added to.
 *      nameArg - string name of the scatter plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmScatterPlotClass::AmScatterPlotClass( AmOverlayClass 	*overlayArg,
				    const char		*nameArg,
				    ...)
{  
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

    initializeFields( overlayArg->GetWindow(), nameArg);
    setParent( overlayArg);

/* then override the fields for the overlay plot with the fields passed in
   for initialization */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
    SetValid( IsValid() && ( dataInfo.sourceType != AmNO_SOURCE));
    if ( IsValid() && ( overlayArg != NULL)) {
	overlayArg->AddLayer( this);
    }
}

/****************************************************************************
 *
 * Description:
 *      AmScatterPlotClass destructor: destroy the symbolIndex and 
 *	symbolSizeIndex indexData structures, the symbol size list 
 *	and the symbol list if they've been allocated.  If the plot's
 *	standalone, then free the display image if it's been allocated,
 *	then free the scatter plot's symbol cache.
 *
 ****************************************************************************/

AmScatterPlotClass::~AmScatterPlotClass()
{
    AmDestroyIndexData( &symbolIndex);
    AmDestroyIndexData( &symbolSizeIndex);
    if ( symbolSizeList != NULL) {
	free( symbolSizeList);
    }
    if ( symbolList != NULL) {
	AmFreeStrVec( symbolList);
    }
    if ( parent == NULL) {
	if ( displayImage != NULL) {
	    delete displayImage;
	} 
    }

/* free the symbol cache */

    AmFreeSymbolCache( &symbolCache);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a scatter plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are set, then
 *      the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.   Since the bin vectors for the symbol list
 *	and symbolSize list are 1 less than the length of the vectors,
 *	they're set by hand.  The symbolIndex and symbolSizeIndex 
 *	structures are initialized after the fields are set.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterPlotClass::setFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    int         numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr fieldDefs;

    numValues = AmPlotClass::setFields( tagValues, numValues);
    AmScatterPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolBinvec)) {
	    AmSetBinvec( &symbolIndex.binvec, symbolIndex.length, 
		       &tagValues[ ii]);
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
	    AmSetBinvec( &symbolSizeIndex.binvec, symbolSizeIndex.length, 
		       &tagValues[ ii]);
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmScatterPlotClass::updateFields()
{
    float	xmin, ymin, xmax, ymax;

    AmPlotClass::updateFields();

/* set the index data client struct for the symbol list and symbol size list */
    
    if ( IsValid()) {
	if ( symbolList != NULL) {
	    if ( updateIndexData( &symbolIndex, "symbolIndex") != AmOK) {
		AmAddResourceToLog( "failedToInitializeSymbolIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( symbolSizeList != NULL) {
	    if ( updateIndexData( &symbolSizeIndex, 
				  "symbolSizeIndex") != AmOK) {
		AmAddResourceToLog( "failedToInitializeSymbolSizeIndex", 
				  GetName());
		SetValid( FALSE);
		return;
	    }
	}
	if ( colors != NULL) {
	    if ( updateIndexData( &colorIndex, "colorIndex") != AmOK) {
		AmAddResourceToLog( "failedToInitializeColorIndex", GetName());
		SetValid( FALSE);
		return;
	    }
	}

/* perform the xy ranging, if there's an extents function to set the range */

        if (( dataInfo.sourceType != AmREAD_SOURCE) &&
            ( dataInfo.sourceType != AmPYRAMID_SOURCE)) {
            if ( !fixedExtents) {
                GetDataExtents( &xmin, &ymin, &xmax, &ymax);
                SetRange( xmin, ymin, xmax, ymax);
            }
        } else if ( !fixedExtents) {
            SetValid( FALSE);
            AmAddResourceToLog( "extentsFnNoRead", dataInfo.sourceType);
            return;
        }

/* read the symbol file */

	if ( AmChanged( &symfile)) {
	    sym = AmGetSymbol( symfile, &symbolCache);
	    if ( sym == NULL) {
		AmAddResourceToLog( "symbolNotFound", symfile);
		SetValid( FALSE);
		return;
	    }
	}
    }
}

/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a scatter plot, given the tag-values stored in
 *      a tag values array.  First, the generic plot fields are retrived, then
 *      the other fields are retrieved using GetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The symbol list and symbol size list are
 *	retrieved by hand.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.  The values are pointers to the returned values.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmScatterPlotClass::getFields( AmTagValuePtr	tagValues,
			     int		numValues)
{
    int         numFieldDefs, ii, numUnusedFields;
    AmFieldDefPtr fieldDefs;
 
    numValues = AmPlotClass::getFields( tagValues, numValues);
    AmScatterPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = GetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    numUnusedFields = 0;
    for ( ii = 0; ii < numValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolBinvec)) {
	    if ( symbolIndex.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( symbolIndex.binvec, 
			     ( symbolIndex.length - 1)*sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else if ( AmMatchTag( tagValues[ ii].tag, AmNsymbolSizeBinvec)) {
	    if ( symbolSizeIndex.binvec != NULL) {
		*( void **) tagValues[ ii].value = 
		    AmCopyVec( symbolSizeIndex.binvec, 
			     ( symbolSizeIndex.length - 1)*sizeof( float));
	    } else {
		*( void **) tagValues[ ii].value = NULL;
	    }
	} else {
	    tagValues[ numUnusedFields++] = tagValues[ ii];
	}
    }
    return( numUnusedFields);
}

/****************************************************************************
 *
 * Description:
 * 	draw the strings for a vector of symbols.  This is used when the
 *	showStrings flag is set to true.
 * Arguments:
 *      window - window that the strings are to be drawn on.
 *      gc - graphics context used for graphics attributes.
 *	ptr - pointer to the data to extract the strings from.
 *      numPts - number of points in the vector.
 *      boundRect - bounding rect to clip to.
 *
 ****************************************************************************/

void 
AmScatterPlotClass::drawSymbolStrings( AmWindowClass	*window,
				     AmGCClass		*gc,
				     uchar_t		*ptr,
				     int		numPts,
				     AmRectangleRec const	*boundRect)
{
    float	x, y;
    int		screenX, screenY, ii;
    char	str[ TEXTLEN];

    for ( ii = 0; ii < numPts; ii++, ptr += dataInfo.size) {
	x = ( dataInfo.xFn)( ptr + dataInfo.xOffset, dataInfo.clientData);
	y = ( dataInfo.yFn)( ptr + dataInfo.yOffset, dataInfo.clientData);
	GetScreenPt( x, y, &screenX, &screenY);
	if ( AmPtInRect( boundRect, screenX, screenY)) {
	    strncpy( str, ( char *) ( ptr + textOffset), textSize);
	    AmElimTrailWspace( str);

/* need to track max symbol size for offset */

	    display->DrawAlignedText( window, gc, screenX, screenY, str, 
				      AmCENTER_HORIZONTAL);
	}
    }
}

/****************************************************************************
 *
 * Description:
 * 	for scatter plot set rectangle, redraw the scatter plot into the 
 *	image buffer, after allocating a new image (if it's not an overlay 
 *	plot, where the resize is already done) 
 * Arguments:
 *	eventCodes - vector of event codes from AmSendUserEvent()
 *	numEventCodes - number of event codes.
 * Returns:
 *	AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterPlotClass::Update( int	*eventCodes,
			  int	numEventCodes)
{
    AmRectangleRec const  *dispRect = GetRect(); 
    AmRectangleRec	clearRect;
    int			ii, numel;
    uchar_t		*pointBuffer;

/* if this object is invalid return */

    if ( !IsValid()) {
        return( AmNOT_VALID);
    }
 
/* if the image is out of sync, then resize it, and reset the rectangle
   to the screen extents */

    if ( parent == NULL) {
	if (( displayImage = ResizeDisplayImage( displayImage)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
    } else { 
	displayImage = (( AmOverlayClass *) parent)->GetImage();
    }

/* clear out the updated rectangle */

    if ( parent == NULL) {
	clearRect = *dispRect;
	clearRect.x = 0;
	clearRect.y = 0;
	displayImage->ClearImageRect( &bgColor, &clearRect);
    }
    if ( dataInfo.sourceType == AmREAD_SOURCE) {

/* if the source type is read-from-file, then read the points in buffers
   and write each buffer into the image */

	if (( pointBuffer = 
	    ( uchar_t *) alloca( dataInfo.size*scatterBufSize)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	for ( ii = 0; ii < dataInfo.dims[ 0]; ii += scatterBufSize) {
	    numel = MIN( scatterBufSize, dataInfo.dims[ 0] - ii);
	    if ( read( dataInfo.fd, pointBuffer, 
		       dataInfo.size*numel) < dataInfo.size*numel) {
		DO_RETURN_ERROR( AmREAD_FAILED);
	    }
	    AmDrawSymbolImage( display, ( AmCoordClass *) this, dispRect,
			     pointBuffer, dataInfo.size, 
			     scatterBufSize,
			     dataInfo.xFn, dataInfo.xOffset,
			     dataInfo.yFn, dataInfo.yOffset, 
			     &colorIndex, colors,
			     &symbolIndex, symbolList,
			     &symbolSizeIndex, symbolSizeList,
			     dataInfo.clientData,
			     displayImage, sym, &symbolCache, &fgColor,
			     interpolate);
	}
    } else if (( dataInfo.sourceType == AmFILE_SOURCE) ||
	       ( dataInfo.sourceType == AmDIRECT_SOURCE)) {

/* if the source type is direct-from-memory, then just write the 
   memory segment into the image */

	AmDrawSymbolImage( display, ( AmCoordClass *) this, dispRect,
			 dataInfo.data, dataInfo.size, 
			 dataInfo.dims[ 0], 
			 dataInfo.xFn, dataInfo.xOffset,
			 dataInfo.yFn, dataInfo.yOffset,
			 &colorIndex, colors,
			 &symbolIndex, symbolList,
			 &symbolSizeIndex, symbolSizeList,
			 dataInfo.clientData,
			 displayImage, sym, &symbolCache, &fgColor,
			 interpolate);
    } else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	AmTileRec         tiles[ MAXTILES];
	AmTilePtr         tile;
	AmErrorCode     errCode;
	int             numTiles, maxRows;

/* find the tiles which overlap the current coordinates */

	numTiles = 0;
	errCode = AmFindLayerTiles( coords.currXmin, coords.currYmin,
				  coords.currXmax, coords.currYmax,
				  ( float) detailLevel,
				  dataInfo.fd, &tileHeader->rootTile, tiles,
				  &numTiles, MAXTILES);
	if ( errCode != AmOK) {
	    return( errCode);
	}

/* allocate the point buffer to draw the largest tile */

	maxRows = 0;
	for ( ii = 0; ii < numTiles; ii++) {
	    if ( tiles[ ii].rows > maxRows) {
		maxRows = tiles[ ii].rows;
	    }
	}
	if (( pointBuffer = ( uchar_t *)
		alloca( maxRows*dataInfo.size)) == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}
	for ( ii = 0; ii < numTiles; ii++) {
	    tile = &tiles[ ii];
	    if ( tile->numBytes == 0) {
		continue;
	    }
	    errCode = AmReadTile( tileHeader, tile, pointBuffer);
	    if ( errCode != AmOK) {
		return( errCode);
	    }

/* write the symbols from the tile into the image */

	    AmDrawSymbolImage( display, ( AmCoordClass *) this, dispRect,
			     pointBuffer, dataInfo.size, tile->rows,
			     dataInfo.xFn, dataInfo.xOffset,
			     dataInfo.yFn, dataInfo.yOffset,
			     &colorIndex, colors,
			     &symbolIndex, symbolList,
			     &symbolSizeIndex, symbolSizeList,
			     dataInfo.clientData,
			     displayImage, sym, &symbolCache, &fgColor,
			     interpolate);
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	draw the current selection set as glyphs, which are drawn on
 *	top of the display, rather than directly into the image.
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 *	fHighlight - controls whether the scatter points are drawn in
 *	the highlight color or normal 
 * Returns:
 *      AmOK if everything's OK, AmNOT_VALID if the scatter plot's invalid.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterPlotClass::DrawSelection( AmWindowClass	 	*window,
				 AmGCClass		*gc, 
				 AmRectangleRec const  	*exposeRects,
				 int		 	nRects,
				 AnvilBoolean		fHighlight)
{
    uchar_t		*selected;
    int			nSel;
    AmColorCellPtr	color;
    AmRectangleRec	surroundRect;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* get the selection */

    GetSelection(( void **) &selected, &nSel);
    if ( nSel > 0) {

/* if the base glyph hasn't been loaded, then restore it from the cache */

	if ( glyph == NULL) {
	    glyph = AmGetGlyph( display, window, symfile, 1.0, 
			      highlightColor.pixel, &glyphCache);
	}
	AmUnionRectVec( exposeRects, nRects, &surroundRect);
	if ( fHighlight) {
	    color = &highlightColor;
	} else {
	    color = &fgColor;
	}
	AmGlyphDraw( display, window, ( AmCoordClass *) this, &surroundRect,
		   selected, dataInfo.size, nSel,
		   dataInfo.xFn, dataInfo.xOffset, 
		   dataInfo.yFn, dataInfo.yOffset,
		   &colorIndex, NULL, &symbolIndex, symbolList,
		   &symbolSizeIndex, symbolSizeList,
		   dataInfo.clientData, glyph, &glyphCache, color);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	Draw the scatter plot.  If the scatter plot is from memory, then
 *	we just blit the image to the screen.  If it's AmREAD_SOURCE
 *	or AmPYRAMID_SOURCE, then read the flat file or the pyramid
 *	file on the fly.
 * Arguments:
 *      window - window the the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error if there was a memory
 *      allocation or I/O problem.
 *
 ****************************************************************************/

AmErrorCode 
AmScatterPlotClass::draw( AmWindowClass		*window,
			AmGCClass			*gc, 
			AmRectangleRec const  	*exposeRects,
			int			nRects)
{
    AmRectangleRec const	*dispRect = GetRect();
    int			ii;

/* if this object is invalid only draw the outline */

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

/* return in the degenerate cases */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }
 
/* since the image has a 0,0 origin, we shift the expose rectangle from the
   window coordinate system to the image coordinate system 
   if we're standalone, draw the image, otherwise, it will get drawn by the
   overlay plot */

    if ( parent == NULL) {	
	for ( ii = 0; ii < nRects; ii++) {
	    display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
				exposeRects[ ii].y - dispRect->y, 
				exposeRects[ ii].x, exposeRects[ ii].y,
				exposeRects[ ii].width, exposeRects[ ii].height,
				displayImage);
	}
    }

/* if we show the text strings for the scatter points, we do it in a 
   second pass */

    if ( showText) {
	AmRectangleRec	boundRect;
	uchar_t		*pointBuffer;
	int		numel;

	AmUnionRectVec( exposeRects, nRects, &boundRect);
	if ( dataInfo.sourceType == AmREAD_SOURCE) {
	    if (( pointBuffer = 
		( uchar_t *) alloca( dataInfo.size*scatterBufSize)) == NULL) {
		DO_RETURN_ERROR( AmMALLOC_FAILED);
	    }
	    if ( lseek( dataInfo.fd, dataInfo.skipBytes, SEEK_SET) < 0) {
		DO_RETURN_ERROR( AmSEEK_FAILED);
	    }
	    for ( ii = 0; ii < dataInfo.dims[ 0]; ii += scatterBufSize) {
		numel = MIN( scatterBufSize, dataInfo.dims[ 0] - ii);
		if ( read( dataInfo.fd, pointBuffer, 
			   dataInfo.size*numel) < dataInfo.size*numel) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		drawSymbolStrings( window, gc, pointBuffer, numel,
				   &boundRect);
	    }
	} else if (( dataInfo.sourceType == AmFILE_SOURCE) ||
		   ( dataInfo.sourceType == AmDIRECT_SOURCE)) {
	    drawSymbolStrings( window, gc, dataInfo.data, dataInfo.dims[ 0],
			       &boundRect);
	} else if ( dataInfo.sourceType == AmPYRAMID_SOURCE) {
	    AmTileRec         tiles[ MAXTILES];
	    AmTilePtr         tile;
	    AmErrorCode     errCode;
	    int             numTiles, maxRows;

	    numTiles = 0;
	    errCode = AmFindLayerTiles( coords.currXmin, coords.currYmin,
				      coords.currXmax, coords.currYmax,
				      ( float) detailLevel,
				      dataInfo.fd, &tileHeader->rootTile, tiles,
				      &numTiles, MAXTILES);
	    if ( errCode != AmOK) {
		return( errCode);
	    }

/* allocate the point buffer to draw the largest tile */

	    maxRows = 0;
	    for ( ii = 0; ii < numTiles; ii++) {
		if ( tiles[ ii].rows > maxRows) {
		    maxRows = tiles[ ii].rows;
		}
	    }
	    if (( pointBuffer = ( uchar_t *)
		    alloca( maxRows*dataInfo.size)) == NULL) {
		DO_RETURN_ERROR( AmALLOCA_FAILED);
	    }
	    for ( ii = 0; ii < numTiles; ii++) {
		tile = &tiles[ ii];
		if ( tile->numBytes == 0) {
		    continue;
		}
		errCode = AmReadTile( tileHeader, tile, pointBuffer);
		if ( errCode != AmOK) {
		    return( errCode);
		}
		drawSymbolStrings( window, gc, pointBuffer, tile->rows,
				   &boundRect);
	    }
	}
    }
    return( DrawSelection( window, gc, exposeRects, nRects, TRUE));
}

/****************************************************************************
 *
 * Description:
 * 	Return the extents of a scatter point.  This is used for moving
 *	and selecting points.
 * Arguments:
 *	ptr - data point.
 *	xmin, ymin, xmax, ymax: location and extents of the point.
 *
 ****************************************************************************/

void 
AmScatterPlotClass::ObjectExtents( void 	*ptr,
				 float	*xmin,
				 float	*ymin,
				 float	*xmax,
				 float	*ymax)
{
    AmCoordPtr	point = ( AmCoordPtr) ptr;

    *xmin = point->x;
    *ymin = point->y;
    *xmax = point->x;
    *ymax = point->y;
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a scatter plot.  This is
 *      used in the map editor to dump the plot's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int
AmScatterPlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			      AnvilBoolean		allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( plot) */

    numStrings = AmPlotClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmScatterPlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numStrings = DumpFieldsByDefinitions( fieldLengths, numStrings,
					  fieldDefs, numFieldDefs);
    if ( symbolIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNsymbolBinvec;
	fieldLengths[ numStrings++].length = symbolIndex.length - 1;
    }
    if ( symbolSizeIndex.binvec != NULL) {
	fieldLengths[ numStrings].tag = AmNsymbolSizeBinvec;
	fieldLengths[ numStrings++].length = symbolSizeIndex.length - 1;
    }
    return( numStrings);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a scatter plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmScatterPlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields. THIS MUST BE IN ALPHABETICAL 
 *	ORDER ON THE FIELD NAME FOR SEARCHING 
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmScatterPlotClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
				       int		*numDefs)
{
    static AmFieldDefRec ScatterPlotClassFieldDefinitions[] = {
	{ AmNinterpolate, offsetof( AmScatterPlotClass, interpolate) },
	{ AmNsymbol, offsetof( AmScatterPlotClass, symfile), 0, AmDYNAMIC },
        { AmNsymbolClientData,
          offsetof( AmScatterPlotClass, symbolIndex.clientData) },
        { AmNsymbolClientDataFreeFn,
          offsetof( AmScatterPlotClass, symbolIndex.clientDataFreeFn) },
        { AmNsymbolFn, offsetof( AmScatterPlotClass, symbolIndex.fn) },
        { AmNsymbolIntFn, offsetof( AmScatterPlotClass, symbolIndex.AmIntFn) },
        { AmNsymbolLength, offsetof( AmScatterPlotClass, symbolIndex.length) },
        { AmNsymbolList, offsetof( AmScatterPlotClass, symbolList),
			 offsetof( AmScatterPlotClass, symbolIndex.length),
			 AmDYNAMIC },
        { AmNsymbolMax,
          offsetof( AmScatterPlotClass, symbolIndex.max) },
        { AmNsymbolMin,
          offsetof( AmScatterPlotClass, symbolIndex.min) },
        { AmNsymbolName,
          offsetof( AmScatterPlotClass, symbolIndex.name), 0, AmDYNAMIC },
        { AmNsymbolOffset,
          offsetof( AmScatterPlotClass, symbolIndex.offset) },
        { AmNsymbolSizeClientData,
          offsetof( AmScatterPlotClass, symbolSizeIndex.clientData) },
        { AmNsymbolSizeClientDataFreeFn,
          offsetof( AmScatterPlotClass, symbolSizeIndex.clientDataFreeFn) },
        { AmNsymbolSizeFn, offsetof( AmScatterPlotClass, symbolSizeIndex.fn) },
        { AmNsymbolSizeIntFn, 
	  offsetof( AmScatterPlotClass, symbolSizeIndex.AmIntFn) },
        { AmNsymbolSizeLength, 
  	  offsetof( AmScatterPlotClass, symbolSizeIndex.length) },
        { AmNsymbolSizeList, 
	      offsetof( AmScatterPlotClass, symbolSizeList),
	      offsetof( AmScatterPlotClass, symbolSizeIndex.length),
	      AmDYNAMIC },
        { AmNsymbolSizeName,
          offsetof( AmScatterPlotClass, symbolSizeIndex.name), 0, AmDYNAMIC },
        { AmNsymbolSizeOffset,
          offsetof( AmScatterPlotClass, symbolSizeIndex.offset) },
        { AmNsymbolSizeType,
          offsetof( AmScatterPlotClass, symbolSizeIndex.dataType) },
        { AmNsymbolSizeMax,
          offsetof( AmScatterPlotClass, symbolSizeIndex.max) },
        { AmNsymbolSizeMin,
          offsetof( AmScatterPlotClass, symbolSizeIndex.min) },
        { AmNsymbolSizeStrings, 
	  offsetof( AmScatterPlotClass, symbolSizeIndex.strings),
	  offsetof( AmScatterPlotClass, symbolSizeIndex.length),
	  AmDYNAMIC },
        { AmNsymbolSizeTable, 
	      offsetof( AmScatterPlotClass, symbolSizeIndex.table),
	      offsetof( AmScatterPlotClass, symbolSizeIndex.length),
	      AmDYNAMIC },
        { AmNsymbolSizeUnits, 
	      offsetof( AmScatterPlotClass, symbolSizeIndex.units) },
        { AmNsymbolStrings, offsetof( AmScatterPlotClass, symbolIndex.strings),
			  offsetof( AmScatterPlotClass, symbolIndex.length),
			  AmDYNAMIC },
        { AmNsymbolTable, offsetof( AmScatterPlotClass, symbolIndex.table),
			  offsetof( AmScatterPlotClass, symbolIndex.length),
			  AmDYNAMIC },
        { AmNsymbolType, offsetof( AmScatterPlotClass, symbolIndex.dataType) },
        { AmNsymbolUnits, offsetof( AmScatterPlotClass, symbolIndex.units) },
    };
    *fieldDefs = ScatterPlotClassFieldDefinitions;
    *numDefs = VECLEN( ScatterPlotClassFieldDefinitions);
}

/* utility and support functions */

/*************************************************************************
 *
 * Description:
 * 	do a linear search for the closest scatter point to a coordinate 
 * Arguments:
 *	ptX, ptY - point in the scatter plot's coordinate system.
 *	scatterPlot - scatter plot to examine.
 * Modifications:
 *	this should probably be made into a member function.
 *
 *************************************************************************/

int 
AmClosestScatterPointIndex( float             ptX,
			  float             ptY,
			  AmScatterPlotClass  *scatterPlot)
{
    int         ii, xOffset, yOffset, bestIndex;
    uchar_t     *scatterData;
    AmAccessFn    xFn, yFn;
    void        *clientData;
    float       scatterX, scatterY, AmDist, bestDist;

/* extract the fields */

    scatterPlot->GetVaFields( AmNdata, &scatterData,
                              AmNxFn, &xFn,
                              AmNxOffset, &xOffset,
                              AmNyFn, &yFn,
                              AmNyOffset, &yOffset,
                              AmNclientData, &clientData,
                              NULL);

/* generate the distance for the AmFirst point */

    bestIndex = 0;
    bestDist = SQRDIST(( xFn)( scatterData + xOffset, clientData),
                       ( yFn)( scatterData + yOffset, clientData), ptX, ptY);
    scatterData += scatterPlot->GetSize();

/* linear search for the closest point */

    for ( ii = 1; 
          ii < scatterPlot->GetRows(); 
          ii++, scatterData += scatterPlot->GetSize()) {
        scatterX = ( xFn)( scatterData + xOffset, clientData);
        scatterY = ( yFn)( scatterData + yOffset, clientData);
        AmDist = SQRDIST( ptX, ptY, scatterX, scatterY);
        if ( AmDist < bestDist) {
            bestIndex = ii;
            bestDist = AmDist;
        }   
    }
    return( bestIndex);
}
