#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */





#include <search.h>
#include <malloc.h>
#include <stddef.h>
#include <string.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/tagstrs.h>
#include <anvil/files.h>
#include <anvil/envdefs.h>
#include <anvil/mem_funcs.h>
#include <anvil/screenobj.h>
#include <anvil/coordinates.h>
#include <anvil/userevents.h>
#include <anvil/eventcodes.h>
#include <anvil/enums.h>
#include <anvil/log.h>

/****************************************************************************
 *
 * Description:
 *	Empty Constructor
 *
 ****************************************************************************/

AmScreenObjClass::AmScreenObjClass()
{
}

/****************************************************************************
 *
 * Description:
 * 	initialize the fields for a screen object.  The intializeFields()
 *	functions are created for each class, and each one calls the one
 *	for the "parent" class.  The AmScreenObjClass intializeFields() 
 *	function is the topmost one that is called.
 * Arguments:
 *	windowArg - object is created on, and will be drawn on this window
 * 	nameArg - string name used to identify screen object for resources
 *
 ****************************************************************************/

void AmScreenObjClass::initializeFields( AmWindowClass	*windowArg,
				       const char	*nameArg)
{
    window = windowArg;
    display = window->GetDisplay();

/* an actual screen object is not valid */

    objectType = AmERROR_OBJECT;

/* initialize the parent to NULL */

    parent = NULL;

/* turn off the margins */

    leftMargin = -1;
    rightMargin = -1;
    topMargin = -1;
    bottomMargin = -1;

/* set the rectangle to something reasonable */

    rect.x = 0;
    rect.y = 0;
    rect.width = 100;
    rect.height = 100;
    extentsRect = rect;

/* copy the name if there is one */

    if ( nameArg != NULL) {
	name = strdup( nameArg);
    } else {
	name = NULL;
    }

/* alllocate the foreground and background colors. Default foreground is
   black, default background is white. */

    fgColor.red = 0;
    fgColor.green = 0;
    fgColor.blue = 0;
    display->AllocateColors( &fgColor, 1);
    bgColor.red = 255;
    bgColor.green = 255;
    bgColor.blue = 255;
    display->AllocateColors( &bgColor, 1);
    colors = NULL;
    cmapName = NULL;
    numColors = 0;

    opaque = TRUE;
    _fFrame = FALSE;
}

/****************************************************************************
 *
 * Description:
 *	Check the values for a screen object.  This function should be
 *	defined for all screen objects, and call the checkValues() function
 *	for the parent class.  AmScreenObjClass() objects start as valid 
 * Returns:
 *	TRUE for valid object, false for invalid object.
 *
 ****************************************************************************/

AnvilBoolean AmScreenObjClass::checkValues()
{
    return( TRUE);
}
/****************************************************************************
 *
 * Description:
 *	AmScreenObjClass constructor.  Calls initialize fields, then parses
 *	the varargs argument list.  This function is unused.
 * Arguments:
 *	windowArg - object is created on, and will be drawn on this window
 * 	nameArg - string name used to identify screen object for resources
 *	varargs - a paired list of tag-value pairs, with the tags from
 *		  <anvil/tagstrs.h>, and the types of the values dictated
 *		  from <anvil/strdefs.h>.  The list is NULL-terminated
 *
 ****************************************************************************/

AmScreenObjClass::AmScreenObjClass( AmWindowClass	*windowArg,
				const char	*nameArg,
				...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* initialize all the basic fields for a plot */

    initializeFields( windowArg, nameArg);

/* set the default image client data and pixel function */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 *	Screen object destructor.  Each object calls the destructor for its
 *	parent class. This destructor sends the user event to the display 
 *	manager (and whoever else is listening) that this object has been
 *	deleted.  It also frees any other malloced object fields.
 *
 ****************************************************************************/

AmScreenObjClass::~AmScreenObjClass()
{

/* free the colormap name, object name, and free the foreground and background 
   pixels and the colors */

    if ( cmapName != NULL) {
	FREE( cmapName);
    }
    if ( name != NULL) {
	FREE( name);
    }
    if ( colors != NULL) {
        FREE( colors);
    }

/* send a user event saying that this object's been deleted */

    AmSendUserEvent( AM_PKG_ID, AM_DELETE_OBJECT, this);
}

/****************************************************************************
 *
 * Description:
 *	return the object's screen rectangle.  This rectangle is used for
 *	sizing the object.
 * Returns:
 *	A pointer to the object's screen rectangle.  
 *
 ****************************************************************************/

AmRectangleRec const *AmScreenObjClass::GetRect()
{
    return( &rect);
}

/****************************************************************************
 *
 * Description:
 *	return the object's extents rectangle.  This rectangle is used for
 *	determining the screen real-estate of an object.
 * Returns:
 *	A pointer to the object's extents rectangle.  
 *
 ****************************************************************************/

AmRectangleRec const *AmScreenObjClass::GetExtentsRect()
{
    return( &extentsRect);
}

/****************************************************************************
 *
 * Description:
 *	set the object's screen rectangle.  The default is to copy the 
 *	rectangle to the extents
 * Arguments:
 *	x,y,width,height: dimensions of the screen object's new screen
 *	rectangle
 *
 ****************************************************************************/

void AmScreenObjClass::SetRect( int	x,
			      int	y,
			      int	width,
			      int	height)
{
    rect.x = x;
    rect.y = y;
    rect.width = width;
    rect.height = height;
    extentsRect = rect;
    AmSendUserEvent( AM_PKG_ID, AM_RECT_CHANGE, this);
}

/****************************************************************************
 *
 * Description:
 *	set the rectangle, and change the margins to reflect the new 
 *	rectangle, so when it's resized by SetWindowSize(), the new screen
 *	rectangle affects the margins.
 * Arguments:
 *	x,y,width,height: dimensions of the screen object's new screen
 *	rectangle
 *	winWidth, winHeight: dimensions of the window that the screen
 *	object is offset to.
 *
 ****************************************************************************/

void AmScreenObjClass::SetRectAndMargins( int	x,
					int	y,
					int	width,
					int	height,
					int	winWidth,
					int	winHeight)
{
    SetRect( x, y, width, height);
    if ( leftMargin != -1) {
	leftMargin = rect.x;
    }
    if ( rightMargin != -1) {
	rightMargin = winWidth - (rect.x + rect.width);
    }
    if ( topMargin != -1) {
	topMargin = rect.y;
    }
    if ( bottomMargin != -1) {
	bottomMargin = winHeight - (rect.y + rect.height);
    }
}


/****************************************************************************
 *
 * Description:
 *	Default screen object draw function.  The draw function takes 
 *	a vector of expose rectangles which define the region to be
 *	refreshed.  The screen object draw function simply draws a 
 *	rectangle
 * Arguments:
 *	w - window to draw to.
 *	gc - graphics context for drawing attributes.
 *	exposeRects - vector of NON-OVERLAPPING expose rectangles to
 *	be refreshed
 *	nRects - number of expose rectangles.
 *
 ****************************************************************************/

AmErrorCode AmScreenObjClass::Draw( AmWindowClass               *w,
				    AmGCClass                   *gc,
				    AmRectangleRec const        *exposeRects,
				    int			        nRects)
{

    int	ii;

/* Return if the object's size is degenerate. */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }

/* Clear out the exposed rectangles if the object is opaque. */

    if ( opaque) {
	gc->SetForegroundColor( &bgColor);
	for ( ii = 0; ii < nRects; ii++) {
	    display->FillRectangle( w, gc, exposeRects[ ii].x, 
					   exposeRects[ ii].y,
					   exposeRects[ ii].width,
					   exposeRects[ ii].height);
	}
    }

    draw( w, gc, exposeRects, nRects);

    if ( _fFrame == TRUE) {

	gc->SetForegroundColor( &fgColor);
	gc->SetLineWidth( 1.0);

        display->DrawRectangle( w, gc,
                                rect.x, rect.y,
                                rect.width, rect.height);
    }

    return( AmOK);
}

AmErrorCode AmScreenObjClass::draw( AmWindowClass               *w,
				    AmGCClass                   *gc,
				    AmRectangleRec const        *exposeRects,
				    int			        nRects)
{


/* Return if the object's size is degenerate. */

    if ( AmDegenRect( GetRect())) {
        return( AmOK);
    }
    return( AmOK);
}


/****************************************************************************
 *
 * Description:
 *	draw an outline rectangle with an X through it. 
 * Arguments:
 *	w - window to draw to.
 *	gc - graphics context for drawing attributes.
 *	drawRect - rectangle to draw.
 *
 ****************************************************************************/

void AmScreenObjClass::DrawOutline( AmWindowClass               *w,
                                    AmGCClass                   *gc,
                                    AmRectangleRec const        *drawRect)
{

/* Return if the extends rectangle is degenerate. */ 

    if ( AmDegenRect( GetRect())) {
        return;
    }

/* Draw an outline around the rectangle. */

    display->DrawRectangle( w, gc,
			    rect.x, rect.y,
			    rect.width, rect.height);
}

/****************************************************************************
 *
 * Description:
 * 	return the enumerated type of the object.  The object type is
 *	a run-time identifier for screen objects, and can be extended 
 *	for subclassed objects.
 * Returns:
 *	The object's enumerated type
 *
 ****************************************************************************/

AmObjectTypeEnum AmScreenObjClass::GetObjectType()
{
    return( objectType);
}

/****************************************************************************
 *
 * Description:
 * 	A recursive run-time type function which for each class determines
 *	if the requested type is of this class, or of the parent's
 *	class (and so on).  This should be used instead of GetObjectType()
 *	wherever possible, to allow subclassing.
 * Arguments:
 *	type - requested object type
 * Returns:
 *	TRUE if the object is of this type or a subclass of it.
 *	FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean AmScreenObjClass::IsType(AmObjectTypeEnum type)
{
    return( type == AmSCREEN_OBJECT);
}

/****************************************************************************
 *
 * Description:
 * 	Update is called by the display manager whenever the attributes 
 *	of an object has changed.  This is different from draw, in that
 *	there are objects which have processing which can be done outside
 *	of a redraw.  For most objects except image-related objects and
 *	legends/scales, this function does nothing.
 * Arguments:
 *	eventCodes - vector of event codes sent by AmSendUserEvent() to
 *	be processed by this object.  This vector may be NULL, and if it
 *	is, then all the attributes for the object should be assumed to
 *	have changed.
 *	numEvents - number of event codes.
 * Return:
 *	Error code if there are any problems on the update, AmOK if everything
 *	went OK.
 *
 ****************************************************************************/

AmErrorCode AmScreenObjClass::Update( int	*eventCodes,
				    int	numEvents)
{
    return( IsValid() ? AmOK : AmNOT_VALID);
}

/****************************************************************************
 *
 * Description:
 * 	return the display for the screen object 
 * Return:
 *	The object's display pointer.
 *
 ****************************************************************************/

AmDisplayClass *AmScreenObjClass::GetDisplay()
{
    return( display);
}

/****************************************************************************
 *
 * Description:
 * 	return the window pointer for the screen object.
 * Return:
 *	The object's window pointer.
 *
 ****************************************************************************/

AmWindowClass *AmScreenObjClass::GetWindow()
{
    return( window);
}


/****************************************************************************
 *
 * Description:
 * 	allocate the colors for the plot, whether they're set explicitly, or
 *  	read from a file 
 * Arguments:
 *	cmapName - name of the colormap file.  If the file does not start
 *	with '/', it is assumed to reside int the $ANVIL_CONFIG/colormaps
 *	directory.
 *	numColors - pointer to the number of returned colors 
 *	colors - pointer to vector of returned color cells.
 * Return:
 *	AmErrorCode: AmOK if OK, otherwise system error or AmFILE_NO_EXIST.
 * Possible Modifications:
 *	Make the function try read the colormap from the current directory
 *	if it doesn't reside in $ANVIL_CONFIG/colormaps.
 *
 ****************************************************************************/

AmErrorCode AmReadColormap( char          *cmapName,
                          int           *numColors,
                          AmColorCellPtr  *colors)
{
    char                *temp, cmapPath[ PATHLEN], relPath[ PATHLEN];
    AmColorCellPtr        fileCells;
    int                 numReadColors;

/* read the colors from the color file, and if the user has specified the
   number of colors, then interpolate to them, otherwise, use the number
   of colors in the file for the number of colors in the plot */

    if ( cmapName[ 0] != '/') {
        sprintf( relPath, "%s/%s", AM_COLORMAP_DIR, cmapName);
        temp = AmGetEnvFile( CONFIG_ENV, relPath);
        if ( temp == NULL) {
            DO_RETURN_ERROR( AmFILE_NO_EXIST);
        }
        strcpy( cmapPath, temp);
    } else {
        strcpy( cmapPath, cmapName);
    }
    if ( cmapPath == NULL) {
        DO_RETURN_ERROR( AmFILE_NO_EXIST);
    }
    if ( !AmReadColors( cmapPath, &numReadColors, &fileCells)) {
        DO_RETURN_ERROR( AmREAD_FAILED);
    }
    if ( *numColors != 0) {
        AmInterpolateColorCells( fileCells, numReadColors, colors, *numColors);
        free( fileCells);
        if ( *colors == NULL) {
            DO_RETURN_ERROR( AmMALLOC_FAILED);
        }
    } else {
        *colors = fileCells;
        *numColors = numReadColors;
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	take any AmCOLOR_CELL or AmCOLOR_CELL_PTR arguments and de-allocate 
 *	the old values, and allocate the new ones.  These fields are parsed
 *	by hand in screen object, because of the vagaries of reading,
 *	allocating, and interpolating colormaps.
 * Arguments:
 *	fieldDefs: field definitions for the screen object class, hopefully
 *	with some of them containing data of the type AmCOLOR_CELL or
 *	AmCOLOR_CELL_PTR.
 *	numFieldDefs: length of the fieldDefs array.
 *	tagValues: tag-value pairs to be parsed (this array is reduced in
 *	place, with the parsed arguments removed).
 *	numValues: length of the tagValues array.
 * Returns:
 *	number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int AmScreenObjClass::ParseColorCellArgs( AmFieldDefPtr	fieldDefs,
					int		numFieldDefs,
					AmTagValuePtr	tagValues,
					int		numValues)
{
    int			ii, numUnusedValues, numNonColorValues,
			oldNumColors = numColors; 
    AmColorCellPtr	cell, newColors;
    AmStrdefPtr		strdef;
    char		*offsetPtr;
    AmFieldDefPtr		fieldDef;

/* in the AmFirst loop, read the single color cells, and vector lengths for the
   colors */

    numNonColorValues = 0;
    for ( ii = 0; ii < numValues; ii++) {

/* find out the field's data type */
 
        strdef = AmGetStrdef( tagValues[ ii].tag);

/* tag not found ... */

        if ( strdef == NULL) {
	    tagValues[ numNonColorValues++] = tagValues[ ii];
            continue;
        }
	fieldDef = ( AmFieldDefPtr)
	    bsearch( tagValues[ ii].tag, fieldDefs, numFieldDefs,
		     sizeof( AmFieldDefRec), ( AmSearchFn) AmCompareFieldTag);
	if ( fieldDef != NULL) {
	    offsetPtr = ( char *) this + fieldDef->offset;
	    if ( strdef->dataType == AmCOLOR_CELL) {

/* if it's a single color cell, deallocate the old one, and allocate the
   new one */

		AmMarkChanged( offsetPtr);
		cell = ( AmColorCellPtr) tagValues[ ii].value;
		display->AllocateColors( cell, 1);
		*( AmColorCellPtr) offsetPtr = *cell;
		if ( tagValues[ ii].copied) {
		    free( tagValues[ ii].value);
		}
		if ( tagValues[ ii].tagCopied) {
		    free( tagValues[ ii].tag);
		}

/* set num colors */

	    } else if ( AmMatchTag( strdef->tag, AmNnumColors)) {

/* if the # of colors has been changed, then re-interpolate the color vector,
   and set the oldNumColors to reflect the new colors */

		numColors = ( int) tagValues[ ii].value;
		if ( colors != NULL) {
		    AmInterpolateColorCells( colors, oldNumColors, 
					   &newColors, numColors);
		    FREE( colors);
		    display->AllocateColors( newColors, numColors);
		    colors = newColors;
		    oldNumColors = numColors;
		}
		if ( tagValues[ ii].tagCopied) {
		    free( tagValues[ ii].tag);
		}
	    } else {
		tagValues[ numNonColorValues++] = tagValues[ ii];
	    }
	} else {
	    tagValues[ numNonColorValues++] = tagValues[ ii];
	}
    }

/* now do the non-atomic values */

    numUnusedValues = 0;
    for ( ii = 0; ii < numNonColorValues; ii++) {
	if ( AmMatchTag( tagValues[ ii].tag, AmNcolors) ||
	     AmMatchTag( tagValues[ ii].tag, AmNcolormap)) {
	    AmMarkChanged( &colors);

/* if it's a vector of color cells or a colormap file, deallocate the old ones,
   find the length for the new ones, assign it, and allocate them.  */

	    if ( colors != NULL) {
		FREE( colors);
	    }

/* if it's a colormap file, then read the colors from it, otherwise, take them
   directly from the value */

	    if ( AmMatchTag( tagValues[ ii].tag, AmNcolormap)) {
		if ( cmapName != NULL) {
		    FREE( cmapName);
		}
		cmapName = strdup(( char *) tagValues[ ii].value);
		if ( cmapName != NULL) {
		    if ( AmReadColormap( cmapName, &numColors, 
				       &newColors) != AmOK) {
			AmAddResourceToLog( "readColormapFailed", cmapName);
			SetValid( FALSE);
		    }
		}
	    } else {
		newColors = ( AmColorCellPtr) 
		    AmCopyVec( tagValues[ ii].value, 
			     sizeof( AmColorCellRec)*numColors);
	    }

/* allocate and assign the color vector and set the colors changed flag */
	
	    if ( IsValid()) {
		display->AllocateColors( newColors, numColors);
		colors = newColors;
	    }

/* free the value and tag if they're flagged */

	    if ( tagValues[ ii].copied) {
	        free( tagValues[ ii].value);
	    }
	    if ( tagValues[ ii].tagCopied) {
		free( tagValues[ ii].tag);
	    }
	} else {
	    tagValues[ numUnusedValues++] = tagValues[ ii];
	}
    }

/* copy the unused values back to the tag values array, and return the 
   number of unprocessed values */ 

    return( numUnusedValues);
}
 
/****************************************************************************
 *
 * Description:
 * 	wrapper function to do send events, set extra fields, and track changes
 *  	for screen objects.  This should be called by all the object
 *	constructors so user events are sent consistently.
 * Arguments:
 *	tagValues: tag-value pairs to be parsed (this array is reduced in
 *		   place, with the parsed arguments removed).
 *	numValues: length of the tagValues array.
 * Return:
 *	number of unparsed values remaining in the tagValues array.
 * Modifications:
 *	should call the AmRecordClass::setFieldsWrapper() function by 
 *	a :: syntax, rahter than calling AmTrackChanges() and setExtraFields()
 *	explicitly.
 *
 ****************************************************************************/
 
int AmScreenObjClass::setFieldsWrapper( AmTagValuePtr       tagValues,
				      int               numValues)
{
    numValues = AmRecordClass::setFieldsWrapper( tagValues, numValues);
 
/* send a user event that we've been updated to anybody who cares to
   listen */
 
    AmSendUserEvent( AM_PKG_ID, AM_SET_FIELD_OBJECT, this);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *	set the fields for a screen object.  The color cell arguments
 *	are parsed by hand, the rest of the arguments are parsed by
 *	the SetFieldsByDefinitions() interface.
 * Arguments:
 *	tagValues: tag-value pairs to be parsed (this array is reduced in
 *		   place, with the parsed arguments removed).
 *	numValues: length of the tagValues array.
 * Return:
 *	number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int AmScreenObjClass::setFields( AmTagValuePtr       tagValues,
			       int               numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;
  
    AmScreenObjClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = ParseColorCellArgs( fieldDefs, numFieldDefs, 
				    tagValues, numValues);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs,
				       tagValues, numValues);
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *	retreive the fields for a screen object.
 * Arguments:
 *	tagValues: tag-value pairs to be parsed (this array is reduced in
 *		   place, with the parsed arguments removed).  The values
 *		   are treated as void pointers to the returned values.
 *		   Note that all strings and vector values are allocated
 *		   when they are returned.
 *	numValues: length of the tagValues array.
 * Return:
 *	number of unparsed values remaining in the tagValues array.
 *
 ****************************************************************************/

int AmScreenObjClass::getFields( AmTagValuePtr	tagValues,
			       int		numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    AmScreenObjClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/****************************************************************************
 *
 * Description:
 *	dump the field names which are used by a screen object.
 * Arguments:
 *	fieldLengths: the name of each field in the screen object, and
 *		      the length to allocate by elements (which has to
 *		      be multiplied by the element size.
 *	allFields: used to cull specific fields for copying objects and
 *		   tracing objects. This field will hopefully become
 *		   obsolete.
 * Return:
 *	number of fieldLength structures returned in the fieldLengths vector.
 *
 ****************************************************************************/

int
AmScreenObjClass::dumpFields( AmFieldLengthPtr	fieldLengths,
			    AnvilBoolean		allFields)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    AmScreenObjClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, 0, fieldDefs, numFieldDefs));
}

/****************************************************************************
 *
 * Description:
 *	Set the window size that the object is drawn to, which will call
 *	SetRect() if any of the object's margins are set.
 * Arguments:
 *	winWidth, winHeight: dimensions window is set to 
 *
 ****************************************************************************/

void AmScreenObjClass::SetWindowSize( int		winWidth,
				    int		winHeight)
{
    int	width, height, x, y;

/* if the left and right margins are set, then the width is the window
   width minus the margins, otherwise it's the rectangle width.  
   Same for the height */

    width = (( rightMargin != -1) && ( leftMargin != -1)) ?
		     winWidth - (rightMargin + leftMargin) : rect.width;
    height = (( topMargin != -1) && ( bottomMargin != -1)) ?
		     winHeight - (topMargin + bottomMargin) : rect.height;

/* if the right margin is set and the left margin isn't then x is the 
   window width minus the width of the object and the right margin, otherwise
   if the leftMargin is set, it's the left margin, otherwise it's the 
   rectangle's x. Same thing for y. */

    if ( leftMargin != -1) {
	x = leftMargin;
    } else if ( rightMargin != -1) {
	x = winWidth - rightMargin - width;
    } else { 
	x = rect.x;
    }
    if ( topMargin != -1) {
	y = topMargin;
    } else if ( bottomMargin != -1) {
	y = winHeight - bottomMargin - height;
    } else {
	y = rect.y;
    }
    SetRect( x, y, width, height);
}

/****************************************************************************
 *
 * Description:
 *	write the data for a screen object in a header file format.  This 
 *  	function assumes that the object is a primitive, i.e. non-composite
 *  	object 
 * Arguments:
 *	fp - file pointer to write data out to.
 *
 ****************************************************************************/

AnvilBoolean AmScreenObjClass::WriteHeader( FILE *fp)
{
    char	typeStr[ RESLEN];

/* write out the object's type */

    if ( AmGetEnumStr( AmNobjectType, GetObjectType(), typeStr)) {
	fprintf( fp, "%s:	%s\n", AmNtype, typeStr);
    }

/* use the recordClass writeHeader to write out the rest of the fields */

    (( AmRecordClass *) this)->WriteHeader( fp);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      return the name of the screen object, set in its constructor. 
 *
 ****************************************************************************/

const char *AmScreenObjClass::GetName()
{ 
    return( name);
}

/****************************************************************************
 *
 * Description:
 *	write the data for a screen object in a header file format, with
 *	a structural syntax.  This is used for printing maps and for
 *	cutting and pasting screen objects.
 * Arguments:
 *	fp - file pointer to write data out to.
 * Returns:
 *	Always returns TRUE.
 *
 ****************************************************************************/

AnvilBoolean
AmScreenObjClass::DumpFields( FILE *fp)
{
    const char *objectName;

    objectName = getObjectTypeName();
    if ( objectName == NULL) {
        return( FALSE);
    }

/* slightly different syntax if there's a parent */

    if ( parent != NULL) {
        fprintf( fp, "%s Plot%p ( Plot%p ) {\n", objectName, this, parent);
    } else {
        fprintf( fp, "%s Plot%p {\n", objectName, this);
    }

    AmRecordClass::DumpFields( fp);
    fprintf( fp, "}\n");

    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      return a pointer to the object's parent 
 *
 ****************************************************************************/

AmScreenObjClass *AmScreenObjClass::getParent()
{
    return( parent);
}

/****************************************************************************
 *
 * Description:
 *      returns the flag controlling whether the object is opaque 
 *	(covers the objects behind it) This is used by the display
 *	manager for redrawing and clipping objects when they change
 *
 ****************************************************************************/

AnvilBoolean AmScreenObjClass::getOpaque()
{
    return( opaque);
}

/* set the object's parent */

void AmScreenObjClass::setParent( AmScreenObjClass	*parentArg)
{
    parent = parentArg;
}

/* return the object's name */

const char *
AmScreenObjClass::getObjectName()
{
    if ( name != NULL) {
	return( name);
    } else {
	return( getObjectTypeName());
    }
}

/* get the object's type name */

const char *
AmScreenObjClass::getObjectTypeName()
{
    static char string[ RESLEN];

    AmGetEnumStr( "objectType", ( int) GetObjectType(), string);
    return( string);
}
			    
/* return the field definitions which are specific to a screen object.
   NOTE THAT ALL OF THE FIELDS HAVE TO BE IN ALPHABETICAL ORDER FOR
   SEARCHING. The structures contain the tag name, offset in the
   structure, offset to the integer value definining the value's
   length if it's a vector, and a storage type, which is either 
   AmSTATIC for embedded arrays, or AmDYNAMIC, for dynamically allocated
   arrays */

void AmScreenObjClass::GetFieldDefinitions( AmFieldDefPtr	*fieldDefs,
					  int		*numFieldDefs)
{
    static AmFieldDefRec ScreenObjFieldDefinitions[] = {
        { AmNbackground, offsetof( AmScreenObjClass, bgColor) },
	{ AmNbottomMargin, offsetof( AmScreenObjClass, bottomMargin) },
	{ AmNcolormap, offsetof( AmScreenObjClass, cmapName), 0, AmDYNAMIC },
	{ AmNcolors, offsetof( AmScreenObjClass, colors),
		     offsetof( AmScreenObjClass, numColors), AmDYNAMIC },
        { AmNforeground, offsetof( AmScreenObjClass, fgColor) },
        { AmNframe, offsetof( AmScreenObjClass, _fFrame) },
	{ AmNheight, offsetof( AmScreenObjClass, rect.height) },
	{ AmNleftMargin, offsetof( AmScreenObjClass, leftMargin) },
	{ AmNname, offsetof( AmScreenObjClass, name), 0, AmDYNAMIC },
	{ AmNnumColors, offsetof( AmScreenObjClass, numColors) },
	{ AmNopaque, offsetof( AmScreenObjClass, opaque) },
	{ AmNrightMargin, offsetof( AmScreenObjClass, rightMargin) },
	{ AmNtopMargin, offsetof( AmScreenObjClass, topMargin) },
	{ AmNwidth, offsetof( AmScreenObjClass, rect.width) },
	{ AmNx, offsetof( AmScreenObjClass, rect.x) },
	{ AmNy, offsetof( AmScreenObjClass, rect.y) }
    };
    *fieldDefs = ScreenObjFieldDefinitions;
    *numFieldDefs = VECLEN( ScreenObjFieldDefinitions);
}
