#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <math.h>
#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/sphere.h>

/*************************************************************************
 *
 * Description:
 *	compute the cosines and sines for the theta and phi angles in 
 *	a spherical plot, scaled by the spherical plot's radius.
 * Arugments:
 *	rows, cols: dimensions of the source matrix.
 *	radius: radius of the sphere in screen coordinates.
 *	sinphis: pointer to a vector of sins of the phi angles, of rows
 *	length.
 *	cosphis: pointer to a vector of cosines of the phi angles, of rows
 *	length.
 *	sinthetas: 
 *	costhetas: pointer to a vector of sines of the theta angles, of cols
 *	length.
 *	sinthetas: pointer to a vector of sines of the theta angles, of cols
 *	length.
 *
 *************************************************************************/

AmErrorCode 
AmComputeSphericalAngles( int	rows,
			int	cols,
			int	radius,
			float	**sinphis,
			float	**cosphis,
			float	**sinthetas,
			float	**costhetas)
{
    int		ii;
    float	phiDegStep, thetaDegStep, angle;

    if (( *sinphis = ( float *) malloc( rows*sizeof( float))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( *cosphis = ( float *) malloc( rows*sizeof( float))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( *sinthetas = ( float *) malloc( cols*sizeof( float))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( *costhetas = ( float *) malloc( cols*sizeof( float))) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    phiDegStep = M_PI/rows;
    angle = -M_PI/2;
    for ( ii = 0; ii < rows; ii++) {
	( *sinphis)[ ii] = radius*sin( angle);
	( *cosphis)[ ii] = radius*cos( angle);
	angle += phiDegStep;
    }
    thetaDegStep = ( M_PI*2)/cols;
    angle = 0;
    for ( ii = 0; ii < cols; ii++) {
	( *sinthetas)[ ii] = sin( angle);
	( *costhetas)[ ii] = cos( angle);
	angle += thetaDegStep;
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 * 	take a matrix of values interpreted as the radii for a spherical 
 *	plot.  The phis and thetas are precomputed by the 
 *	AmComputeSphericalAngles() routine.
 * Arguments:
 *	image - image that the spherical plot will be drawn to.
 *	points - matrix of values which compose the sphere.
 *	size - size of each element in the source matrix.
 *	rows, cols - dimensions of the source matrix.
 *	colorsFn - coverts source values to floats.
 *	colorsClientData - client data for colorsFn.
 *	colorsMin, colorsMax - min and max scaling values
 *	colors - vector of colors to render.
 *	numColors - length of the colors vector.
 *	sinphis, cosphis, sinthetas, costhetas - precomputed sin and cosine
 *	phi and theta vectors for the sphere.
 *	theta, gamma - sphere initial rotation.
 *
 *************************************************************************/

void AmSphericalPlotTables( 
    AmImageClass		*image,		/* destination image */
    uchar_t		*points,	/* 2-D data matrix */
    int			size,
    int			rows,		/* data matrix dimensions */
    int			cols,
    AmAccessFn		colorsFn,
    void		*colorsClientData,
    float		colorsMin,
    float		colorsMax,
    AmColorCellPtr	colors,
    int			numColors,
    float		*sinphis, 	/* sin( phi)*radius */
    float		*cosphis, 	/* cos( phi)*radius */
    float		*sinthetas,	/* sin( theta) */
    float		*costhetas,	/* cos( theta) */
    float		theta,
    float		gamma)
{
    float	r, *leftTheta, *rightTheta, *cosphiPtr, *sinphiPtr,
		*endTheta, *endphi, *leftCostheta, *rightCostheta, xval,
		singamma, cosgamma;
    int		x, y, center, index, width, height, cx, cy, colcenter;
    ulong_t	*imageData;
    uchar_t	*leftPt, *leftPtPtr, *rightPt, *rightPtPtr, *endPt;

    width = image->GetWidth();
    height = image->GetHeight();
    cx = width/2;
    cy = height/2;
    imageData = ( ulong_t *) image->AmGetData();
    singamma = sin( gamma);
    cosgamma = cos( gamma);
    colcenter = (int)(cols/2 - (( cols/2)*theta)/ M_PI);

    if ( colcenter < 0) {
	colcenter = 0;
    } else if ( colcenter >= cols) {
	colcenter = rows - 1;
    }

/* 
   sin(g) >= 0, cos( g) >= 0 => back to front, bottom to top
   sin(g) >= 0, cos( g) < 0 => back to front, top to bottom
   sin(g) < 0, cos( g) >= 0 => front to back, bottom to top
   sin(g) < 0, cos( g) < 0 =>  front to back, top to bottom
*/
    center = (cols + 1) >> 1;
    if ( cosgamma >= 0.0) {
	if (EVEN( cols)) {
	    rightTheta = leftTheta = sinthetas + center;
	    rightCostheta = leftCostheta = costhetas + center;
	    rightPt = leftPt = points + colcenter*size;
	} else {
	    rightTheta = sinthetas + center;
	    rightCostheta = costhetas + center;
	    rightPt = points + colcenter*size;
	    leftPt = points + ( colcenter - 1)*size;
	    leftTheta = sinthetas + center - 1;
	    leftCostheta = costhetas + center - 1;
	}
	endTheta = sinthetas + cols;
    } else {

/* since we're reversed, the column center is rotated 180 degrees, and
   the thetas are set at the edges, rather than the center */

	colcenter = ( colcenter + (( cols + 1) >> 1)) % cols;
	rightCostheta = costhetas;
	leftCostheta = costhetas + cols - 1;
	rightTheta = sinthetas;
	leftTheta = sinthetas + cols - 1;
	if ( EVEN( cols)) {
	    leftPt = rightPt = points + colcenter*size;
	} else {
	    leftPt = points + colcenter*size;
	    rightPt = points + ( colcenter - 1)*size;
	}
	endTheta = sinthetas + center + 1;
    }
    endPt = points + cols*size;

/* if the gamma rotation is away from the viewer, draw the points from 
   top to bottom */

    if ( singamma < 0.0) {
	endphi = cosphis + rows;
	while ( rightTheta < endTheta) {

/* left and right pointers start in at points in top row, 
   are incremented row by row */

	    leftPtPtr = leftPt;
	    rightPtPtr = rightPt;
	    cosphiPtr = cosphis;
	    sinphiPtr = sinphis;
	    for ( cosphiPtr = cosphis, sinphiPtr = sinphis; 
		  cosphiPtr < endphi; 
		  cosphiPtr++, sinphiPtr++) {

/* do right hand side */

		r = (( colorsFn)( rightPtPtr, colorsClientData) - colorsMin) /
			( colorsMax - colorsMin);
		rightPtPtr += cols;
		xval = r*( *cosphiPtr);
		x = (int)(*rightTheta*xval);
		y = (int)(cosgamma*r*( *sinphiPtr) - singamma*(*rightCostheta)*xval);

/* continue of the data is out of the image */

		if (( x >= width/2) || ( x < -width/2)) {
		    continue;
		}
		if (( y >= height/2) || ( y < -height/2)) {
		    continue;
		}
		index = (int)(r*numColors);

/* make sure the index is in the numColors range */

		if ( index < 0) {
		    index = 0;
		} else if ( index > numColors - 1) {
		    index = numColors - 1;
		}
		imageData[ cx + x + (cy - y)*width] = colors[ index].pixel;

/* do left hand side */

		r = (( colorsFn)( leftPtPtr, colorsClientData) - colorsMin) /
			( colorsMax - colorsMin);
		leftPtPtr += cols;
		xval = r*( *cosphiPtr);
		x = (int)(*leftTheta*xval);
		y = (int)(cosgamma*r*( *sinphiPtr) - singamma*(*leftCostheta)*xval);

/* continue of the data is out of the image */

		if (( x >= width/2) || ( x < -width/2)) {
		    continue;
		}
		if (( y >= height/2) || ( y < -height/2)) {
		    continue;
		}
		index = (int)(r*numColors);

/* make sure the index is in the numColors range */

		if ( index < 0) {
		    index = 0;
		} else if ( index > numColors - 1) {
		    index = numColors - 1;
		}
		imageData[ cx + x + (cy - y)*width] = colors[ index].pixel;
	    }

/* move the left and right points and carrying angles */

	    rightTheta++;
	    rightCostheta++;
	    leftCostheta--;
	    leftTheta--;
	    leftPt -= size;
	    rightPt += size;
	    if ( leftPt < points) {	/* wraparound */
		leftPt = endPt;
	    }
	    if ( rightPt >= endPt) {
		rightPt = points;
	    }
	}
    } else {		/* gamma rotation towards -> points bottom to top */
	while ( rightTheta < endTheta) {

/* left and right pointers start in at points in bottom row, 
   are decremented row by row */

	    leftPtPtr = leftPt + (rows-1)*cols;
	    rightPtPtr = rightPt + (rows-1)*cols;
	    for ( cosphiPtr = cosphis + rows - 1, sinphiPtr = sinphis + rows - 1;
		  cosphiPtr > cosphis; 
		  cosphiPtr--, sinphiPtr--) {

/* do right hand side */

		r = (( colorsFn)( rightPtPtr, colorsClientData) - colorsMin) /
			( colorsMax - colorsMin);
		rightPtPtr -= cols;
		xval = r*( *cosphiPtr);
		x = (int)(*rightTheta*xval);
		y = (int)(cosgamma*r*( *sinphiPtr) - singamma*(*rightCostheta)*xval);

/* continue of the data is out of the image */

		if (( x >= width/2) || ( x < -width/2)) {
		    continue;
		}
		if (( y >= height/2) || ( y < -height/2)) {
		    continue;
		}
		index = (int)(r*numColors);

/* make sure the index is in the numColors range */

		if ( index < 0) {
		    index = 0;
		} else if ( index > numColors - 1) {
		    index = numColors - 1;
		}
		imageData[ cx + x + (cy - y)*width] = colors[ index].pixel;

/* do left hand side */

		r = (( colorsFn)( leftPtPtr, colorsClientData) - colorsMin) /
			( colorsMax - colorsMin);
		leftPtPtr -= cols;
		xval = r*( *cosphiPtr);
		x = (int)(*leftTheta*xval);
		y = (int)(cosgamma*r*( *sinphiPtr) - singamma*(*leftCostheta)*xval);

/* continue of the data is out of the image */

		if (( x >= width/2) || ( x < -width/2)) {
		    continue;
		}
		if (( y >= height/2) || ( y < -height/2)) {
		    continue;
		}
		index = (int)(r*numColors);

/* make sure the index is in the numColors range */

		if ( index < 0) {
		    index = 0;
		} else if ( index > numColors - 1) {
		    index = numColors - 1;
		}
		imageData[ cx + x + (cy - y)*width] = colors[ index].pixel;
	    }

/* move the left and right points and carrying angles */

	    rightTheta++;
	    rightCostheta++;
	    leftCostheta--;
	    leftTheta--;
	    leftPt -= size;
	    rightPt += size;
	    if ( leftPt < points) {	/* wraparound */
		leftPt = endPt;
	    }
	    if ( rightPt >= endPt) {
		rightPt = points;
	    }
	} 
    }
}
