#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include <stddef.h>
#include <anvil/macros.h>
#include <anvil/overlay.h>
#include <anvil/sphereplot.h>
#include <anvil/sphere.h>

/****************************************************************************
 *
 * Description:
 *      A recursive run-time type function which for each class determines
 *      if the requested type is of this class, or of the parent's
 *      class (and so on).  This should be used instead of GetObjectType()
 *      wherever possible, to allow subclassing.
 * Arguments:
 *      type - requested object type
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

AnvilBoolean 
AmSpherePlotClass::IsType( AmObjectTypeEnum type)
{
    if ( type == AmSPHERE_PLOT) {
        return( TRUE);
    } else {
        return( AmDataGraphicClass::IsType( type));
    }
}

/****************************************************************************
 *
 * Description:
 *      AmSpherePlotClass destructor.
 *
 ****************************************************************************/

AmSpherePlotClass::~AmSpherePlotClass()
{
    delete displayImage;
    if (( colorsClientDataFreeFn != NULL) && ( colorsClientData != NULL)) {
	( colorsClientDataFreeFn)( colorsClientData);
    }
    if ( sinthetas != NULL) {
	free( sinthetas);
    }
    if ( costhetas != NULL) {
	free( costhetas);
    }
    if ( sinphis != NULL) {
	free( sinphis);
    }
    if ( cosphis != NULL) {
	free( cosphis);
    }
}

/****************************************************************************
 *
 * Description:
 *      initialize fields which are specific to a spherical plot.  call the
 *      AmDataGraphicClass initialize fields, then set the spherical 
 *	plot specific fields.
 * Arguments:
 *      windowArg - window that the spherical plot will be drawn on.
 *      nameArg - string name of the spherical plot, used for identification
 *                and pulling resources.
 *
 ****************************************************************************/

void 
AmSpherePlotClass::initializeFields( AmWindowClass	*windowArg,
				   const char	*nameArg)
{
    AmDataGraphicClass::initializeFields( windowArg, nameArg);
    theta = 0.0;
    phi = 0.0;
    colorsFn = AmFloatFn;
    colorsMin = 0.0;
    colorsMax = 0.0;
    colorsClientData = NULL;
    colorsClientDataFreeFn = NULL;
    displayImage = NULL;
    dataInfo.numdims = 2;
    bgColor.red = 0;
    bgColor.green = 0;
    bgColor.blue = 0;
    display->AllocateColors( &bgColor, 1);

/* initialize the coordinates to prevent problems in AmCoordClass::checkValues(),
   even though coordinates aren't really relevant here */

    coords.currXmin = -1.0;
    coords.currXmax = 1.0;
    coords.currYmin = -1.0;
    coords.currYmax = 1.0;
    coords.xmin = -1.0;
    coords.xmax = 1.0;
    coords.ymin = -1.0;
    coords.ymax = 1.0;
    fixedExtents = TRUE;
}

/****************************************************************************
 *
 * Description:
 *      AmSpherePlotClass constructor.  Creates a spherical plot using the window
 *      that it will be drawn on and the name, followed by NULL-terminated
 *      tag-value pairs for setting the fields in the bar graph.
 * Arguments:
 *      windowArg - window that the spherical plot will be drawn on.
 *      nameArg - string name of the spherical plot, used for identification
 *                and pulling resources.
 *      argList - NULL-teriminated tag value pairs, with tags defined in
 *                <anvil/tagvalues.h>
 *
 ****************************************************************************/

AmSpherePlotClass::AmSpherePlotClass( AmWindowClass  *windowArg,
                                  const char   *nameArg,
                                  ...)
{
    va_list     args;
    AmTagValueRec tagValues[ MAXTAGS];
    int         count;

/* initialize all the basic fields for a plot */

    initializeFields( windowArg, nameArg);

/* set the default image client data and pixel function */

    va_start( args, nameArg);
    count = AmMakeTagValueArrayVararg( tagValues, args);
    va_end( args);
    count = setFieldsWrapper( tagValues, count);
}

/****************************************************************************
 *
 * Description:
 *      Check the values for a spherical plot.  These are suppoesd to be
 *      implemented for a graph/plot objects, but they haven't all been
 *      done yet.
 * Returns:
 *      TRUE if the plot has been initialized properly, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmSpherePlotClass::checkValues()
{
    if ( !AmDataGraphicClass::checkValues()) {
    	return( FALSE);
    }
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 *      Set the fields for a spherical plot, given the tag-values stored in
 *      a tag values array.  First, the generic data graphic fields are set, 
 *	then the other fields are set using the standard SetFieldByDefintions()
 *      function which uses a lookup table from the tag strings to the
 *      structure offsets.  The data accessing function is set if the 
 *	source dataType has been changed.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmSpherePlotClass::setFields( AmTagValuePtr       tagValues,
			    int               numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmDataGraphicClass::setFields( tagValues, numValues);
    AmSpherePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    numValues = SetFieldByDefinitions( fieldDefs, numFieldDefs, 
				       tagValues, numValues);
    if ( AmChanged( &dataInfo.dataType)) {
	dataInfo.size = AmDataTypeEnumToSize( dataInfo.dataType);
	colorsFn = AmDataTypeEnumToAccessFn( dataInfo.dataType);
    }
    return( numValues);
}

/****************************************************************************
 *
 * Description:
 *      once the fields have been set using setFields(), update the
 *      indexData structures and other dependent fields.
 *
 ****************************************************************************/

void
AmSpherePlotClass::updateFields()
{
    AmDataGraphicClass::updateFields();
}
 
/****************************************************************************
 *
 * Description:
 *      Retreive the fields for a spherical plot, given the tag-values stored 
 *	in a tag values array.  First, the generic graph fields are gotten, 
 *	then the other fields are pulled using the standard 
 *	GetFieldByDefintions() function which uses a lookup table from  
 *	the tag strings to the structure offsets.
 * Arguments:
 *      tagValues - vector of tag-value pairs (with tagCopied and copied
 *      flags).  The unparsed arguments are returned "scrunched" in this
 *      array.
 *      numValues - length of the vector.
 * Returns:
 *      number of unparsed arguments.
 *
 ****************************************************************************/

int 
AmSpherePlotClass::getFields( AmTagValuePtr       tagValues,
			    int               numValues)
{
    int         numFieldDefs;
    AmFieldDefPtr fieldDefs;

    numValues = AmDataGraphicClass::getFields( tagValues, numValues);
    AmSpherePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( GetFieldByDefinitions( fieldDefs, numFieldDefs,
				   tagValues, numValues));
}

/*************************************************************************
 *
 * Description:
 *      return the list of settable fields in a spherical plot.  This is
 *      used in the map editor to dump the graph's fields into a file.
 *      the field name (and length, if it's a vector field) is written
 *      into the vector.  Each dumpFields member function calls the
 *      dumpFields() member function of the parent class.  This is
 *      also used for copying plots for magic windows.
 * Arguments:
 *      fieldLengths: vector of field names and lengths.
 *      allFields: TRUE when dumping to a file, FALSE when copying
 *      in memory.  This is an obsolete parameter that may be removed
 *      in the future.
 * Returns:
 *      number of strings dumped in the fieldLengths array.
 *
 *************************************************************************/

int 
AmSpherePlotClass::dumpFields( AmFieldLengthPtr    fieldLengths,
			     AnvilBoolean	       allFields)
{
    int         numFieldDefs, numStrings;
    AmFieldDefPtr fieldDefs;

/* get the fields for the parent class ( coordinate system) */

    numStrings = AmDataGraphicClass::dumpFields( fieldLengths, allFields);

/* get all the easy to specify fields for this class */

    AmSpherePlotClass::GetFieldDefinitions( &fieldDefs, &numFieldDefs);
    return( DumpFieldsByDefinitions( fieldLengths, numStrings,
                                     fieldDefs, numFieldDefs));
}

/****************************************************************************
 *
 * Description:
 *      Update the spherical plot, based on the events which have been sent
 *      using AmSendUserEvent().  This stuffs the displayed image with the
 *      appropriate values.  
 * Arguments:
 *      eventCodes - list of event codes sent by AmSendUserEvent()
 *      numCodes - number of event codes.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmSpherePlotClass::Update( int 	*eventCodes,
			 int	numCodes)
{
    AmRectangleRec const	*currRect = GetRect();
    AmRectangleRec	clearRect;
    int			radius;
    AmErrorCode		errCode;

    if ( !IsValid() || !RectSet()) {
        DO_RETURN_ERROR( AmNOT_VALID);
    }
    if ( AmDegenRect( currRect)) {
        return( AmOK);
    }

/* if the image is out of sync, then resize it, and reset the rectangle
   to the screen extents. We have to support the pyramidal images, where
   the data image is resized according to the number of tiles needed to
   cover the screen */

    if ((displayImage = ResizeDisplayImage( displayImage)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* clear out the background window */

    clearRect = *currRect;
    clearRect.x = 0;
    clearRect.y = 0;
    displayImage->ClearImageRect( &bgColor, &clearRect);

/* create the precompute vectors */

    if ( sinthetas != NULL) {
	free( sinthetas);
    }
    if ( costhetas != NULL) {
	free( costhetas);
    }
    if ( sinphis != NULL) {
	free( sinphis);
    }
    if ( cosphis != NULL) {
	free( cosphis);
    }
    radius = 2*MIN( displayImage->GetWidth(), displayImage->GetHeight())/5;
    errCode = AmComputeSphericalAngles( dataInfo.dims[ 0], dataInfo.dims[ 1],
				      radius, &sinphis, &cosphis, 
				      &sinthetas, &costhetas);
    if ( errCode != AmOK) {
	DO_RETURN_ERROR( errCode);
    }
    AmSphericalPlotTables( displayImage, dataInfo.data, dataInfo.size,
			 dataInfo.dims[ 0], dataInfo.dims[ 1], colorsFn,
			 colorsClientData, colorsMin, colorsMax, 
			 colors, numColors,
			 sinphis, cosphis, sinthetas, costhetas, 
			 theta, phi);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      Draw the image containing the spherical plot to the screen. 
 * Arguments:
 *      window - window that the selection set will be drawn on.
 *      gc - graphics context used for graphics attributes.
 *      exposeRects - vector of exposed rectangles.
 *      nRects - length of the expose rectangle vector.
 * Returns:
 *      AmOK if everything's OK, system error otherwise.
 *
 ****************************************************************************/

AmErrorCode 
AmSpherePlotClass::draw( AmWindowClass           *window,
		       AmGCClass               *gc,
		       AmRectangleRec const    *exposeRects,
		       int                   nRects)
{
    AmRectangleRec const  *dispRect = GetRect();
    int                 ii;

    if ( !IsValid()) {
        return( AmScreenObjClass::draw( window, gc, exposeRects, nRects));
    }

    if ( AmDegenRect( dispRect)) {
        return( AmOK);
    }

/* since the image has a 0,0 origin, we shift the expose rectangle from the
   window coordinate system to the image coordinate system.
   if we're standalone, draw the image, otherwise, it will get drawn by the
   overlay plot */

    if ( parent == NULL) {
        if ( displayImage == NULL) {
            DO_RETURN_ERROR( AmBAD_ARGUMENT);
        }
        for ( ii = 0; ii < nRects; ii++) {
            display->DrawImage( window, gc, exposeRects[ ii].x - dispRect->x,
                                exposeRects[ ii].y - dispRect->y,
                                exposeRects[ ii].x, exposeRects[ ii].y,
                                exposeRects[ ii].width, exposeRects[ ii].height,
                                displayImage);
        }
    }
    return( AmOK);
}

/*************************************************************************
 *
 * Description:
 *      Return the field definitions for a spherical plot. This is used
 *      by SetFieldsByDefinitions() and GetFieldsByDefintions() to
 *      set the fields in the AmSpherePlotClass.  It returns a vector
 *      of field names followed by offsets, offset of the length
 *      of the field if it's a vector, and a storageType, which
 *      is AmSTATIC for static arrays in the class, AmDYNAMIC for
 *      dynamically allocated fields.
 * Arugments:
 *      fieldDefs: returned vector of field definitions.
 *      numDefs: returned length of the fieldDefs array.
 *
 *************************************************************************/

void 
AmSpherePlotClass::GetFieldDefinitions( AmFieldDefPtr   *fieldDefs,
				      int           *numFieldDefs)
{
    static AmFieldDefRec SpherePlotClassFieldDefinitions[] = {
	{ AmNcolorsClientData, offsetof( AmSpherePlotClass, colorsClientData) },
	{ AmNcolorsClientDataFreeFn, 
		offsetof( AmSpherePlotClass, colorsClientDataFreeFn) },
	{ AmNcolorsFn, offsetof( AmSpherePlotClass, colorsFn) },
	{ AmNcolorsMax, offsetof( AmSpherePlotClass, colorsMax) },
	{ AmNcolorsMin, offsetof( AmSpherePlotClass, colorsMin) },
        { AmNphi, offsetof( AmSpherePlotClass, phi) },
        { AmNtheta, offsetof( AmSpherePlotClass, theta) },
    };
    *fieldDefs = SpherePlotClassFieldDefinitions;
    *numFieldDefs = VECLEN( SpherePlotClassFieldDefinitions);
}
