#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* basic pyramid utility functions to read the tile headers, and get the
   tiles for a specified layer */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#if defined (_WIN32) || defined(__BSD44__)
#include <malloc.h>
#else
#include <alloca.h>
#endif
#include <fcntl.h>
#include <stdlib.h>

#include <anvil/fns.h>
#include <anvil/errcode.h>
#include <anvil/choppath.h>
#include <anvil/newtile.h>
#include <anvil/macros.h>
#include <anvil/log.h>
#include <anvil/choppath.h>
#include <anvil/compression.h>
#include <anvil/files.h>

/****************************************************************************
 *
 * Description:
 *	Comparison function to be used by qsort() to sort tiles by their
 *	seek addresses, so they can be loaded faster.
 * Arguments:
 *	a,b - tiles to compare.
 * Returns:
 *	difference AmBetween the tile seek addresses. 
 *
 ****************************************************************************/

int 
AmCompareTileSeekAddr( AmTilePtr	a,
		     AmTilePtr	b)
{
    return( a->seekAddr - b->seekAddr);
}


/****************************************************************************
 *
 * Description:
 *	AnvilBoolean function to test whether a tile is a leaf tile or not.
 *	It tests to make sure all the child seek addresses are 0.
 * Arguments:
 *	tile - tile to be tested.
 * Returns:
 *	TRUE if the tile is a leaf, FALSE otherwise.
 *
 ****************************************************************************/

AnvilBoolean 
AmLeafTile( AmTilePtr	tile)
{
    return(( tile->urSeekAddr == 0) && ( tile->lrSeekAddr == 0) &&
	   ( tile->lrSeekAddr == 0) && ( tile->llSeekAddr == 0));
}

/****************************************************************************
 *
 * Description:
 *	Read the tile header from the tiler header file.  This checks
 *	the magic number in the tile file, then reads the vector of
 *	files which contain the tiles.  It then reads the topmost tile.
 * Arguments:
 *	tileHeaderFile - usually extension ".tileHdr", contains the
 *	files and tiling structure for the hierarchical file.
 *	tileHeader - contains the vector of files and their file descriptors,
 *	and the topmost tile in the hierarchy.
 *	fdReturn - the file descriptor for the tile header file.
 * Returns:
 *	AmOK if s'il march, system error s'il ne marche pas.
 *
 ****************************************************************************/

AmErrorCode 
AmReadTileHeader( const char	*tileHeaderFile,
		AmTileHeaderPtr	*tileHeader,
		int		*fdReturn)
{
    int		fd, numFiles, numFileBytes, ii;
    char	*fileNameBuffer, filePath[ PATHLEN],
		*file, *p; 
    const char	*path;
    ulong_t	magicNumber;

    *fdReturn = -1;

/* is the filename absolute?  if so, pull the path */

    path = AmGetPath( tileHeaderFile);
    if (( fd = open( tileHeaderFile, O_RDONLY)) < 0) {
	AmAddResourceToLog( "failedToOpenTileHeader", tileHeaderFile);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    *tileHeader = ( AmTileHeaderPtr) malloc( sizeof( AmTileHeaderRec));
    if ( read( fd, &magicNumber, sizeof( magicNumber)) < 0) {
	AmAddResourceToLog( "tileMagicNumberReadFailed", tileHeaderFile);
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    if ( magicNumber != AmTILE_MAGIC) {
	close( fd);
	DO_RETURN_ERROR( AmMAGIC_FAILED);
    }
    if ( read( fd, &numFiles, sizeof( numFiles)) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    if ( read( fd, &numFileBytes, sizeof( numFileBytes)) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    if (( fileNameBuffer = ( char *) alloca( numFileBytes)) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    if ( read( fd, fileNameBuffer, numFileBytes) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    ( *tileHeader)->numFiles = numFiles;
    if ((( *tileHeader)->files = 
	    ( char **) malloc( sizeof( char *)*( numFiles + 1))) == NULL) {
	close( fd);
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if ((( *tileHeader)->fds = 
	    ( int *) malloc( sizeof( int)*numFiles)) == NULL) {
	close( fd);
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* step through and copy the filenames, separated by NULL terminators */

    p = fileNameBuffer;
    for ( ii = 0; ii < numFiles; ii++) {
	( *tileHeader)->files[ ii] = strdup( p);
	while ( *p++ != '\0');
    }
    ( *tileHeader)->files[ numFiles] = NULL;

/* read the root tile */

    if ( read( fd, &( *tileHeader)->rootTile, sizeof( AmTileRec)) < 0) {
	close( fd);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }

/* open pointers to each of the files referenced by the master tile file */

    for ( ii = 0; ii < numFiles; ii++) {
	file = ( *tileHeader)->files[ ii];
	if (( path != NULL) && !AmAbsolutePath( file)) {
	    sprintf( filePath, "%s/%s", path, file);
	} else {	
	    strcpy( filePath, file);
	}
	if ((( *tileHeader)->fds[ ii] = open( filePath, O_RDONLY)) < 0) {
	    close( fd);
	    AmAddResourceToLog( "failedToOpenTileFile", filePath);
	    DO_RETURN_ERROR( AmOPEN_FAILED);
	}
    }
    *fdReturn = fd;
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	find the tiles which overlap the area of interest, and copy them
 *	into a vector.  The detailLevel parameter controls how low in the
 *	pyramid to go to extract the tiles.
 * Arguments:
 *	xmin, ymin, xmax. ymax - area of interest to overlap.
 *	detailLevel - detail level factor.
 *	tileFd - file descriptor for the tile hierarchy file.
 *	tile - top level tile to test.
 *	tileVec - returned vector of tiles.
 *	nTiles - returned number of tiles in the vector.
 *	tileLength - length of the tile vector (to prevent buffer overflow)
 * Returns:
 *	AmOK on success, AmBUFFER_OVERFLOW if there were more tiles than
 *	the buffer could hold, or a system error if the tiles couldn't
 *	be read or seeked.
 * Modifications:
 *	We're considering a change to the tile structure so that tiles 
 *	will contain a backpointer to their parents, and maintaining the
 *	current thread in the tile tree so tiles can be seeked faster.
 *	This would likely modify the parameters to this function.
 *
 ****************************************************************************/

AmErrorCode AmFindLayerTiles( float	xmin,
			    float	ymin,
			    float	xmax,
			    float	ymax,
			    float	detailLevel,
			    int		tileFd,
			    AmTilePtr	tile,
			    AmTilePtr	tileVec,
			    int		*nTiles,
			    int		tileLength)
{
    float	dispExtents, tileExtents, cx, cy;
    AmTileRec	newTile;
    AmErrorCode	errCode;

    if ( tile == NULL) {
	return( AmOK);
    }
    if (( tile->xmin <= xmax) && ( tile->ymin <= ymax) &&
	( tile->xmax >= xmin) && ( tile->ymax >= ymin)) {
	dispExtents = MIN( ymax - ymin, xmax - xmin);
	tileExtents = MIN( tile->ymax - tile->ymin, tile->xmax - tile->xmin);
	if ((( dispExtents*detailLevel > tileExtents) || AmLeafTile( tile)) &&
	    ( tile->rows > 0)) {
	    if ( *nTiles >= tileLength - 1) {
		DO_RETURN_ERROR( AmBUFFER_OVERFLOW);
	    }
AN_DEBUG( 40, ("found tile (%.2f, %.2f, %.2f, %.2f\n",
	       tile->xmin, tile->ymin, tile->xmax, tile->ymax));
	    tileVec[ *nTiles] = *tile;
	    ( *nTiles)++;
	} else {
	    cx = ( tile->xmax + tile->xmin)/2;
	    cy = ( tile->ymax + tile->ymin)/2;
	    
	    if (( xmax >= cx) && ( ymax >= cy) && ( tile->urSeekAddr != 0)) {
		if ( lseek( tileFd, tile->urSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindLayerTiles( xmin, ymin, xmax, ymax, detailLevel,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	    if (( xmax >= cx) && ( ymin <= cy) && ( tile->lrSeekAddr != 0)) {
		if ( lseek( tileFd, tile->lrSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindLayerTiles( xmin, ymin, xmax, ymax, detailLevel,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	    if (( xmin <= cx) && ( ymax >= cy) && ( tile->ulSeekAddr != 0)) {
		if ( lseek( tileFd, tile->ulSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindLayerTiles( xmin, ymin, xmax, ymax, detailLevel,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    }
	    if (( xmin <= cx) && ( ymin <= cy) && ( tile->llSeekAddr != 0)) {
		if ( lseek( tileFd, tile->llSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindLayerTiles( xmin, ymin, xmax, ymax, detailLevel,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Similar to AmFindLayerTiles(), except that it's intended for use
 *	with raster pyramids.
 * Arguments:
 *	xmin, ymin, xmax. ymax - area of interest to overlap.
 *	rows, cols - size of the display rectangle.
 *	tileFd - file descriptor for the tile hierarchy file.
 *	tile - top level tile to test.
 *	tileVec - returned vector of tiles.
 *	nTiles - returned number of tiles in the vector.
 *	tileLength - length of the tile vector (to prevent buffer overflow)
 * Returns:
 *	AmOK on success, AmBUFFER_OVERFLOW if there were more tiles than
 *	the buffer could hold, or a system error if the tiles couldn't
 *	be read or seeked.
 * Modifications:
 *	We're considering a change to the tile structure so that tiles 
 *	will contain a backpointer to their parents, and maintaining the
 *	current thread in the tile tree so tiles can be seeked faster.
 *	This would likely modify the parameters to this function.  We're
 *	also considering a detail control for images, which would allow
 *	tiles higher in the pyramid to be loaded for speed.
 *
 ****************************************************************************/

AmErrorCode 
AmFindImageTiles( float		xmin,
		float		ymin,
		float		xmax,
		float		ymax,
		int		rows,
		int		cols,
		int		tileFd,
		AmTilePtr		tile,
		AmTilePtr		tileVec,
		int		*nTiles,
		int		tileLength,
		int		detailLevel)
{
    float	dispRes, tileRes, cx, cy;
    AmTileRec	newTile;
    AmErrorCode	errCode;

    if ( tile == NULL) {
	return( AmOK);
    }

/* only bother pursing things if the current tile overlaps the extents of
   the area of interest */

    if (( tile->xmin <= xmax) && ( tile->ymin <= ymax) &&
	( tile->xmax >= xmin) && ( tile->ymax >= ymin)) {

	dispRes = MIN(( ymax - ymin)/rows, ( xmax - xmin)/cols);
	tileRes = MIN(( tile->ymax - tile->ymin)/tile->rows,
		      ( tile->xmax - tile->xmin)/tile->cols);
/* if the display resolution is looser than the tile resolution, then we've
   hit the right layer */

	if (( dispRes*detailLevel > tileRes) || AmLeafTile( tile)) {
	    if ( *nTiles >= tileLength - 1) {
		DO_RETURN_ERROR( AmBUFFER_OVERFLOW);
	    }
AN_DEBUG( 40, ("found tile (%.2f, %.2f, %.2f, %.2f) ( %.2f, %.2f) ( %d, %d)\n",
	       tile->xmin, tile->ymin, tile->xmax, tile->ymax,
	       tile->xmax - tile->xmin, tile->ymax - tile->ymin,
	       tile->rows, tile->cols));
	    tileVec[ *nTiles] = *tile;
	    ( *nTiles)++;
	} else {
	    cx = ( tile->xmax + tile->xmin)/2;
	    cy = ( tile->ymax + tile->ymin)/2;

/* depending on the overlap of the area of interest with the four subtiles,
   recurse down those sections of the tile tree. */

	    if (( xmax >= cx) && ( ymax >= cy) && ( tile->urSeekAddr != 0)) {
		if ( lseek( tileFd, tile->urSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindImageTiles( xmin, ymin, xmax, ymax, rows, cols,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength, detailLevel);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	    if (( xmax >= cx) && ( ymin <= cy) && ( tile->lrSeekAddr != 0)) {
		if ( lseek( tileFd, tile->lrSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindImageTiles( xmin, ymin, xmax, ymax, rows, cols,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength, detailLevel);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	    if (( xmin <= cx) && ( ymax >= cy) && ( tile->ulSeekAddr != 0)) {
		if ( lseek( tileFd, tile->ulSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindImageTiles( xmin, ymin, xmax, ymax, rows, cols,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength, detailLevel);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    }
	    if (( xmin <= cx) && ( ymin <= cy) && ( tile->llSeekAddr != 0)) {
		if ( lseek( tileFd, tile->llSeekAddr, SEEK_SET) < 0) {
		    DO_RETURN_ERROR( AmSEEK_FAILED);
		}
		if ( read( tileFd, &newTile, sizeof( newTile)) < 0) {
		    DO_RETURN_ERROR( AmREAD_FAILED);
		}
		errCode = AmFindImageTiles( xmin, ymin, xmax, ymax, rows, cols,
					  tileFd, &newTile, tileVec, nTiles,
					  tileLength, detailLevel);
		if ( errCode != AmOK) {
		    return( errCode);
		}
	    } 
	}
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	Given a vector of tiles, return the extents of their collective
 *	bounding rectangles.
 * Arguments:
 *	tiles - vector of tiles to test.
 *	numTiles - length of the vector.
 *	rows, cols - resulting rows and columns of the tiles
 *	xmin, ymin, xmax, ymax - resulting extents of the tiles.
 * Possible problems:
 *	because of roundoff errors, the rows and columns calculation for
 *	the tiles may be inaccurate.
 *
 ****************************************************************************/

void 
AmTileVectorExtents( AmTilePtr	tiles,
		   int		numTiles,
		   int		*rows,
		   int		*cols,
		   float	*xmin,
		   float	*ymin,
		   float	*xmax,
		   float	*ymax)
{
    float	minXres, minYres, tileXres, tileYres;
    int		ii;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    minXres = MY_HUGE;
    minYres = MY_HUGE;

/* test the vector of tiles against the extents */

    for ( ii = 0; ii < numTiles; ii++) {
	if ( tiles[ ii].xmin < *xmin) {
	    *xmin = tiles[ ii].xmin;
	}
	if ( tiles[ ii].ymin < *ymin) {
	    *ymin = tiles[ ii].ymin;
	}
	if ( tiles[ ii].xmax > *xmax) {
	    *xmax = tiles[ ii].xmax;
	}
	if ( tiles[ ii].ymax > *ymax) {
	    *ymax = tiles[ ii].ymax;
	}
	tileXres = ( tiles[ ii].xmax - tiles[ ii].xmin)/tiles[ ii].cols,
	tileYres = ( tiles[ ii].ymax - tiles[ ii].ymin)/tiles[ ii].rows;
	if ( tileXres < minXres) {
	    minXres = tileXres;
	}
	if ( tileYres < minYres) {
	    minYres = tileYres;
	}
    }

/* set the rows and columns based on the minimum resolution and coordinate
   extents of the tiles. */

    *rows = (int)rint(( *ymax - *ymin)/minYres);
    *cols = (int)rint(( *xmax - *xmin)/minXres);
}

/****************************************************************************
 *
 * Description:
 * 	utility function for reading a tile.  Uses the compressed flag
 *	in the tile to decide on a straight or compressed read.
 * Arguments:
 *	tileHeader - top level tile header, contains vector of file
 *	descriptors.
 *	tile - to to read.
 *	size - size of each data element.
 *	buffer - preallocated buffer to read into.
 * Returns:
 *	AmOK on success, AmSEEK_FAILED or AmREAD_FAILED on failure.
 *
 ****************************************************************************/

AmErrorCode
AmReadTile( AmTileHeaderPtr	tileHeader,
	  AmTilePtr	tile,
	  uchar_t	*buffer)
{
    int	fd, tileSize;

    if ( tile->numBytes == 0) {
	return( AmOK);
    }

/* get the file descriptor from the vector of file descriptors. */

    fd = tileHeader->fds[ tile->fileIndex];
    if ( LSEEK( fd, tile->seekAddr, SEEK_SET) < 0) {
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }

/* read the tile compressed if the flag is set. */

    if ( tile->flags & AmTILE_COMPRESSED_FLAG) {
	if ( uncompressBlock( fd, ( char *) buffer, tile->numBytes) < 0) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
    } else {

/* read the tile uncooked */

	tileSize = tile->numBytesUncompressed;
	if ( read( fd, ( char *) buffer, tileSize) < tileSize) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
    }
    return( AmOK);
}
