#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdio.h>
#include <math.h>
#include <string.h>
#if defined(__BSD44__)
#include <unistd.h>
#endif

#include <time.h>
#include <sys/utsname.h>

#include <anvil/macros.h>
#include <anvil/envdefs.h>
#include <anvil/errlib.h>
#include <anvil/postscript.h>
#include <anvil/postscriptP.h>
#include <anvil/postscriptcolor.h>

#define SCREEN_COORD( iVirtual) ( iVirtual * _iDPI / _iDPIBase)
#define VIRTUAL_COORD( iScreen) ( iScreen * _iDPIBase / _iDPI)

/**********************************************************************
 *
 * Class: AmPostScriptClass
 *
 **********************************************************************/

/* Constructor */ 

AmPostScriptClass::AmPostScriptClass( int                iDepth,
				      AmColormapTypeEnum eColormapType,
				      int                iDPI)
    : AmDisplayClass()
{
    _eDisplayType = AmPOSTSCRIPT_DISPLAY;

    _iDepth = iDepth;
    _eColormapType = eColormapType;
    _iDPI = iDPI;

    if ( _eColormapType == AmPSEUDO_COLOR) {
        AmCreatePostScriptColorCube( &_pCol, 8, 7, 4);
    } else {
        _pCol = NULL;
    }
}

/* Destructor */

AmPostScriptClass::~AmPostScriptClass()
{
    fprintf( stderr, "In ~AmPostScriptClass()\n");

    if ( _pCol != NULL) {
        free( _pCol);
    }
}

int
AmPostScriptClass::DimInPixels( int iDimension)
{
    return( iDimension);
}

/* Return the depth of the postscript class */

int
AmPostScriptClass::GetDepth()
{
    return( _iDepth);
}

/* Generate the postscript code for drawing lines. */

void
AmPostScriptClass::DrawLines( AmWindowClass *w,
                              AmGCClass     *gc,
                              AmPointPtr    pts,
                              int           npts)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    int                   ii;

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes. */

    pw->ApplyGCAttributes(( AmPostScriptGCClass *) pgc);

    /* Start a new path and walk along the points. */

    fprintf( pw->getFp(), "newpath\n");
    fprintf( pw->getFp(), "%d %d moveto\n",
             ( int) SCREEN_COORD( pts[ 0].x),
             ( int) SCREEN_COORD( pts[ 0].y));

    for ( ii = 1; ii < npts; ii++) {
        fprintf( pw->getFp(), "%d %d lineto\n",
                ( int) SCREEN_COORD( pts[ ii].x),
                ( int) SCREEN_COORD( pts[ ii].y));
    }

    fprintf( pw->getFp(), "stroke\n");
}

/* Generate the postscript code for drawing segments. */

void
AmPostScriptClass::DrawSegments( AmWindowClass *w,
                               AmGCClass     *gc,
                               AmSegmentRec  *segs,
                               int         nseg)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    int                   ii;

/* if there's a new GC, or any of the fields have changed, then apply the
   attributes */

    pw->ApplyGCAttributes(( AmPostScriptGCClass *) pgc);

/* start a new path and walk along the points */

    fprintf( pw->getFp(), "newpath\n");

    for ( ii = 0; ii < nseg; ii++) {
        fprintf( pw->getFp(), "%d %d moveto\n", 
             SCREEN_COORD( segs[ ii].x0),
             SCREEN_COORD( segs[ ii].y0));
        fprintf( pw->getFp(), "%d %d lineto\n", 
             SCREEN_COORD( segs[ ii].x1),
             SCREEN_COORD( segs[ ii].y1));
    }

    fprintf( pw->getFp(), "stroke\n");
}

/* Generate the postscript code for drawing points. */

void
AmPostScriptClass::DrawPoints( AmWindowClass *w,
                               AmGCClass     *gc,
                               AmPointPtr    pts,
                               int           npts)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    int             ii;

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes. */

    pw->ApplyGCAttributes( pgc);

    /* Draw the points as very, very short lines. */

    for ( ii = 0; ii < npts; ii++) {
        fprintf( pw->getFp(), "%d %d moveto\n", 
                SCREEN_COORD( pts[ ii].x),
                SCREEN_COORD( pts[ ii].y));
        fprintf( pw->getFp(), "1 1 rlineto\n");
    }

    fprintf( pw->getFp(), "stroke\n");
}

/* Draw a filled arc starting at angle1, of angle2 width. */

void
AmPostScriptClass::FillArc( AmWindowClass *pWindow,
                          AmGCClass     *pGC,
                          int         x,
                          int         y,
                          int         cx,
                          int         cy,
                          float       angle1,
                          float       angle2)
{
    AmPostScriptOutputClass *pw  = ( AmPostScriptOutputClass *) pWindow;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) pGC;
    float                 dNewAngle;
    AmRectangleRec          rctVirtual, rctScreen;

    rctVirtual.x      = x;
    rctVirtual.y      = y;
    rctVirtual.width  = cx;
    rctVirtual.height = cy;

    ScreenRect( &rctVirtual, &rctScreen);

/* If there's a new GC, or any of the fields have changed, then apply
   the attributes. */

    pw->ApplyGCAttributes( pgc);

    /* If the radius is greater than 360, then postscript starts, and we
       want to wrap fully. */

    if ( fabs( angle2) < M_PI*2) {
        dNewAngle = angle2 + angle1;
        fprintf( pw->getFp(), "gsave\n");
        fprintf( pw->getFp(), "newpath\n");
        fprintf( pw->getFp(), "%d %d %d %d %f %f elliparc\n",
                 rctScreen.x, rctScreen.y,
                 rctScreen.width, rctScreen.height,
                 -RADTODEG( angle1), -RADTODEG( dNewAngle));
        fprintf( pw->getFp(), "closepath fill\n");
        fprintf( pw->getFp(), "grestore\n");
    } else {
        fprintf( pw->getFp(), "gsave\n");
        fprintf( pw->getFp(), "newpath\n");
        fprintf( pw->getFp(), "%d %d %d %d 0 360 elliparc\n",
                 rctScreen.x, rctScreen.y,
                 rctScreen.width, rctScreen.height);
        fprintf( pw->getFp(), "closepath fill\n");
        fprintf( pw->getFp(), "grestore\n");
    }
}

/* Draw an arc starting at angle1, of angle2 width */

void
AmPostScriptClass::DrawArc( AmWindowClass *pWindow,
                          AmGCClass     *pGC,
                          int         x,
                          int         y,
                          int         cx,
                          int         cy,
                          float       angle1,
                          float       angle2)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) pWindow;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) pGC;
    float                 dNewAngle;
    AmRectangleRec          rctVirtual, rctScreen;

    rctVirtual.x      = x;
    rctVirtual.y      = y;
    rctVirtual.width  = cx;
    rctVirtual.height = cy;

    ScreenRect( &rctVirtual, &rctScreen);

    /* If there's a new GC, or any of the fields have changed, then apply
       the attributes. */

    pw->ApplyGCAttributes( pgc);
    if ( fabs( angle2) < M_PI*2) {
        dNewAngle = angle2 + angle1;
        fprintf( pw->getFp(), "gsave\n");
        fprintf( pw->getFp(), "newpath\n");
        fprintf( pw->getFp(), "%d %d %d %d %f %f elliparc\n",
                 rctScreen.x, rctScreen.y,
                 rctScreen.width, rctScreen.height,
                 -RADTODEG( angle1), -RADTODEG( dNewAngle));
        fprintf( pw->getFp(), "stroke\n");
        fprintf( pw->getFp(), "grestore\n");
    } else {
        fprintf( pw->getFp(), "gsave\n");
        fprintf( pw->getFp(), "newpath\n");
        fprintf( pw->getFp(), "%d %d %d %d 0 360 elliparc\n",
                 rctScreen.x, rctScreen.y,
                 rctScreen.width, rctScreen.height);
        fprintf( pw->getFp(), "stroke\n");
        fprintf( pw->getFp(), "grestore\n");
    }
}

/* do nothing...it's not supported */

void
AmPostScriptClass::CopyArea( AmWindowClass *,
                           AmWindowClass *,
                           AmGCClass     *,
                           int,
                           int,
                           int,
                           int,
                           int,
                           int)
{
    fprintf( stderr, "CopyArea is unsupported in PostScript");
}

/* doesn't work for postscript */

void AmPostScriptClass::GetImage( AmWindowClass       *w,
                                AmImageClass        *image,
                                int               x,
                                int               y)
{
    fprintf( stderr, "GetImage is unsupported in PostScript");
}

/* Place an image in the output. */

void
AmPostScriptClass::DrawImage( AmWindowClass *pWindow,
                            AmGCClass     *pGC,
                            int         xSrc,
                            int         ySrc,
                            int         xDst,
                            int         yDst,
                            int         cxSrc,
                            int         cySrc,
                            AmImageClass  *pImage)
{
    AmPostScriptOutputClass *pw  = ( AmPostScriptOutputClass *) pWindow;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) pGC;
    AmRectangleRec          rctVirtual, rctScreen;
    int                   nColors;

/* If there's a new GC, or any of the fields have changed, then apply
   the attributes. */

    pw->ApplyGCAttributes( pgc);

    rctVirtual.x      = xSrc;
    rctVirtual.y      = ySrc;
    rctVirtual.width  = cxSrc;
    rctVirtual.height = cySrc;

    ScreenRect( &rctVirtual, &rctScreen);

    if ( pImage->GetDepth() == 1) {

        AmPrintPostScriptMaskImage(( uchar_t *) pImage->AmGetData(),
                                 rctScreen.x, rctScreen.y,
                                 SCREEN_COORD( xDst),
                                 SCREEN_COORD( yDst),
                                 DimInPixels( rctScreen.width),
                                 DimInPixels( rctScreen.height),
                                 pImage->GetNumHorizPixels(),
				 pImage->GetNumVerticalPixels(),
                                 pw->getFp());

    } else if ( _eColormapType == AmPSEUDO_COLOR) {

        nColors = AmGetPostScriptNumRedShades() *
                  AmGetPostScriptNumBlueShades() *
                  AmGetPostScriptNumGreenShades();

        AmPrintPostScriptPseudoColorImage(( ulong_t *) pImage->AmGetData(),
                                        rctScreen.x, rctScreen.y,
                                        SCREEN_COORD( xDst),
                                        SCREEN_COORD( yDst),
                                        DimInPixels( rctScreen.width),
                                        DimInPixels( rctScreen.height),
                                        pImage->GetNumHorizPixels(),
				        pImage->GetNumVerticalPixels(),
                                        _iDepth,
                                        _pCol,
                                        nColors,
                                        pw->getFp());

    /* Return to the standard colorspace. */

        fprintf( pw->getFp(), "[/DeviceRGB] setcolorspace\n");

    } else {

        AmPrintPostScriptTrueColorImage(( ulong_t *) pImage->AmGetData(),
                                      rctScreen.x, rctScreen.y,
                                      SCREEN_COORD( xDst),
                                      SCREEN_COORD( yDst),
                                      DimInPixels( rctScreen.width),
                                      DimInPixels( rctScreen.height),
                                      pImage->GetNumHorizPixels(),
				      pImage->GetNumVerticalPixels(),
                                      _iDepth,
                                      pw->getFp());

    }
}

/* Draw a glyph. */

void
AmPostScriptClass::DrawGlyph( AmWindowClass *w,
                            AmGlyphClass  *glyph,
                            int         x,
                            int         y)
{
}

/* In a string, all characters which are in the set of characters to
   be quoted have to quote character inserted before them. */
 
static char *
insertQuoteChar( char *str,
                 char *charsToQuote,
                 int  quoteChar)
{
    static char resStr[ 512];
    char        *p;
 
    p = resStr;
    while ( *str != '\0') {
        if ( strchr( charsToQuote, *str) != NULL) {
            *p++ = quoteChar;
        }
        *p++ = *str++;
    }
    *p = '\0';
    return( resStr);
}

/* Draw text on the screen.  This is only for 8-bit text. Note that
   we've reversed the PostScript drawing coordinate system, so we have
   to suspend it, and scale it -1 in y for drawing text. */

void
AmPostScriptClass::DrawText( AmWindowClass *w,
                           AmGCClass     *gc,
                           int         x,
                           int         y,
                           char        *text)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    AmFontWeightEnum        weight;
    AmFontSlantEnum         slant;
    char                  *name;  
    int                   size, width, ascent, descent;
    float                 xScale, yScale;

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes */

    pw->ApplyGCAttributes( pgc);
    pgc->GetFont( &name, &size, &weight, &slant, &xScale, &yScale);
    free( name);

    ExtentsText( pgc, text, &width, &ascent, &descent); 

    fprintf( pw->getFp(), "%d %d moveto\n",
             SCREEN_COORD( x), SCREEN_COORD( y));
    fprintf( pw->getFp(), "gsave\n");
    fprintf( pw->getFp(), "%f -%f scale\n",
             xScale, yScale);

    /* Since PostScript uses () for delimiters, we have to quote them before
       printing the string */

    fprintf( pw->getFp(), "(%s) show\n", insertQuoteChar( text, "()\\", '\\'));
    fprintf( pw->getFp(), "grestore\n");
}

/* Draw text on the screen.  This is only for 8-bit text. Note that
   we've reversed the PostScript drawing coordinate system, so we have
   to suspend it, and scale it -1 in y for drawing text. */

void
AmPostScriptClass::DrawImageText( AmWindowClass  *w,
                                AmGCClass      *gc,
                                int          x,
                                int          y,
                                char         *text)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    AmFontWeightEnum        weight;
    AmFontSlantEnum         slant;
    char                  *name;  
    float                 xScale, yScale;
    int                   size;

    /* If there's a new GC, or any of the fields have changed, then apply
       the attributes */

    pw->ApplyGCAttributes( pgc);
    pgc->GetFont( &name, &size, &weight, &slant, &xScale, &yScale);
    free( name);

    fprintf( pw->getFp(), "%d %d moveto\n",
             SCREEN_COORD( x), SCREEN_COORD( y));
    fprintf( pw->getFp(), "gsave\n");
    fprintf( pw->getFp(), "%f -%f scale\n",
             xScale, yScale);

    /* Since PostScript uses () for delimiters, we have to quote them before
       printing the string. */

    fprintf( pw->getFp(), "(%s) show\n", insertQuoteChar( text, "()\\", '\\'));
    fprintf( pw->getFp(), "grestore\n");
}

/* Draw rotated text on the screen.  This is only for 8-bit text. Note that
   we've reversed the PostScript drawing coordinate system, so we have
   to suspend it, and scale it -1 in y for drawing text. */

void
AmPostScriptClass::DrawRotatedText( AmWindowClass *w,
                                  AmGCClass     *gc,
                                  int         x,
                                  int         y,
                                  float       radianAngle,
                                  char        *text)
{ 
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass      *pgc = ( AmPostScriptGCClass *) gc;
    AmFontWeightEnum        weight;
    AmFontSlantEnum         slant;
    char                  *name;  
    int                   size;
    float                 xScale, yScale;

/* NOTE:  I have to set the text path, but I don't have the postscript 
   manual back yet... */

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes */

    pw->ApplyGCAttributes( pgc);
    pgc->GetFont( &name, &size, &weight, &slant, &xScale, &yScale);

    /* free the unused font name */

    free( name);

    fprintf( pw->getFp(), "%d %d moveto\n",
             SCREEN_COORD( x), SCREEN_COORD( y));
    fprintf( pw->getFp(), "gsave\n");
    fprintf( pw->getFp(), "%.3f rotate\n", -RADTODEG( radianAngle));
    fprintf( pw->getFp(), "%f -%f scale\n",
             xScale, yScale);

    /* Since PostScript uses () for delimiters, we have to quote them before
       printing the string. */

    fprintf( pw->getFp(), "(%s) show\n", insertQuoteChar( text, "()\\", '\\'));
    fprintf( pw->getFp(), "grestore\n");
}

/* Fill a polygon */

void
AmPostScriptClass::FillPolygon( AmWindowClass *w,
                                AmGCClass     *gc,
                                AmPointPtr    pts,
                                int           numpts)
{
    AmPostScriptOutputClass *pw = ( AmPostScriptOutputClass *) w;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) gc;
    int                   ii;

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes. */

    pw->ApplyGCAttributes( pgc);
    fprintf( pw->getFp(), "newpath\n");
    fprintf( pw->getFp(), "%d %d moveto\n", 
	     SCREEN_COORD( pts[ 0].x), SCREEN_COORD( pts[ 0].y));
    for ( ii = 1; ii < numpts; ii++) {
    fprintf( pw->getFp(), "%d %d lineto\n",
             SCREEN_COORD( pts[ ii].x), SCREEN_COORD( pts[ ii].y));
    }
    fprintf( pw->getFp(), "closepath fill\n");
}

/* Draw a rectangle */

void
AmPostScriptClass::DrawRectangle( AmWindowClass *pWindow,
                                AmGCClass     *pGC,
                                int         x,
                                int         y,
                                int         cx,
                                int         cy)
{
    AmPostScriptOutputClass *pw  = ( AmPostScriptOutputClass *) pWindow;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) pGC;
    AmRectangleRec          rctVirtual, rctScreen;

/* Convert the virtual rectangle to the screen rectangle. */

    rctVirtual.x      = x;
    rctVirtual.y      = y;
    rctVirtual.width  = cx;
    rctVirtual.height = cy;

    ScreenRect( &rctVirtual, &rctScreen);

 /* If there's a new GC, or any of the fields have changed, then apply the
    attributes. */

    pw->ApplyGCAttributes( pgc);

/* Draw the rectangle. */    

    fprintf( pw->getFp(), "%d %d %d %d rectstroke\n",
             rctScreen.x, rctScreen.y,
             rctScreen.width, rctScreen.height);
}

/* Fill a rectangle */

void
AmPostScriptClass::FillRectangle( AmWindowClass *pWindow,
                                AmGCClass     *pGC,
                                int         x,
                                int         y,
                                int         cx,
                                int         cy)
{
    AmPostScriptOutputClass *pw  = ( AmPostScriptOutputClass *) pWindow;
    AmPostScriptGCClass     *pgc = ( AmPostScriptGCClass *) pGC;
    AmRectangleRec          rctVirtual, rctScreen;

/* Convert the virtual rectangle to the screen rectangle. */

    rctVirtual.x      = x;
    rctVirtual.y      = y;
    rctVirtual.width  = cx;
    rctVirtual.height = cy;

    ScreenRect( &rctVirtual, &rctScreen);

    /* If there's a new GC, or any of the fields have changed, then apply the
       attributes. */

    pw->ApplyGCAttributes( pgc);
    fprintf( pw->getFp(), "%d %d %d %d rectfill\n",
             rctScreen.x, rctScreen.y,
             rctScreen.width, rctScreen.height);
}

/* Always 24-bit... */

AnvilBoolean
AmPostScriptClass::AllocColor( AmColorCellPtr cell)
{
    if ( _iDepth == 24) {
        cell->pixel = ( cell->blue << 16) | ( cell->green << 8) | cell->red;
        return( TRUE);
    } else {
        return( TRUE);
    }
}

/* create an image (virtual function semi-constructor */
 
AmImageClass *
AmPostScriptClass::CreateImage( int     cx,
                              int     cy,
                              int     iDepth,
                              char    *pData)
{
    AmPostScriptImageClass *pPostScriptImage;

    pPostScriptImage = new AmPostScriptImageClass( this, cx, cy, 
                                                 iDepth, pData);
    return(( AmImageClass *) pPostScriptImage);
}

/* Glyph creation virtual semi-constructor. */

AmGlyphClass *
AmPostScriptClass::CreateGlyph( AmWindowClass *pWindow,
                              char        *szSymbol,
                              ulong_t     pixel)
{
    AmPostScriptGlyphClass *pPostScriptGlyph;

    pPostScriptGlyph = new AmPostScriptGlyphClass( this, pWindow, szSymbol,
                                                 pixel);
    return(( AmGlyphClass *) pPostScriptGlyph);
}

AmRegionClass *
AmPostScriptClass::CreateRegion()
{
    return(( AmRegionClass *) new AmPostScriptRegionClass());
}

/* Return the extents of a text string if it were to be drawn to the screen.
   Unfortunately, this requires the Adobe font format, which is called in the
   GC. */

void
AmPostScriptClass::ExtentsText( AmGCClass *pGC,
                              char    *szText,
                              int     *piWidth,
                              int     *piAscent,
                              int     *piDescent)
{  
    AmPostScriptGCClass    *pgc = ( AmPostScriptGCClass *) pGC;
    int                  iWidth, iAscent, iDescent;

    pgc->GetTextExtents( szText, &iWidth, &iAscent, &iDescent);

    *piWidth = iWidth;
    *piAscent = iAscent;
    *piDescent = iDescent;
}

int
AmPostScriptClass::TextWidth( AmGCClass *pGC,
                            char    *szText,
                            int     iLength)
{
    AmPostScriptGCClass *pPostScriptGC = ( AmPostScriptGCClass *) pGC;
    int               iWidth, iAscent, iDescent;

    ExtentsText( pPostScriptGC, szText, &iWidth, &iAscent, &iDescent); 

    return( iWidth);
}

int
AmPostScriptClass::FontHeight( AmGCClass *pGC,
                             int     *piAscent,
                             int     *piDescent)
{
    AmPostScriptGCClass *pPostScriptGC = ( AmPostScriptGCClass *) pGC;
    int               iWidth, iAscent, iDescent;

    ExtentsText( pPostScriptGC, "TEST", &iWidth, &iAscent, &iDescent); 

    *piAscent = iAscent;
    *piDescent = iDescent;

    return( *piAscent + *piDescent);
}

AnvilBoolean
AmPostScriptClass::AllocNamedColor( const char   *szName,
                                  AmColorCellPtr pCol)
{
    return( TRUE);
}        


/**********************************************************************
 *
 * Class: AmPostScriptOutputClass
 *
 **********************************************************************/

void
AmPostScriptOutputClass::PrintHeader()
{
    time_t t;
    char   *szTime, *szUser, szHost[ TEXTLEN];
    struct utsname hostInfo;

    /*
     * Print the comment header with information about postscript type,
     *   creator, number of pages, page orientation, and so on.
     */
 
    fprintf( _pFile, "%%!PS-Adobe-3.0\n");

    /* The bounding box for each page in the document. */

    fprintf( _pFile, "%%%%BoundingBox: 0 0 ");
    if ( _fLandscape == TRUE) {
        fprintf( _pFile, "%d %d\n", _iHeightPage, _iWidthPage);
    } else {
        fprintf( _pFile, "%d %d\n", _iWidthPage, _iHeightPage);
    }

    fprintf( _pFile, "%%%%Copyright: 1997-2000 Carnegie Mellon University\n");
    fprintf( _pFile, "%%%%Creator: ANVIL Software library\n");

    t = time( NULL);
    szTime = asctime( localtime( &t));

    fprintf( _pFile, "%%%%CreationDate: %s", szTime);

#if defined(__BSD44__)
    szUser = getlogin();
#else
    szUser = cuserid( NULL);
#endif

    uname( &hostInfo);
    strcpy( szHost, hostInfo.nodename);
    fprintf( _pFile, "%%%%For: %s@%s\n",
             ( szUser[ 0] == '\0') ? "unknown" : szUser, 
             ( szHost[ 0] == '\0') ? "unknown" : szHost);

    fprintf( _pFile, "%%%%LanguageLevel: 2\n");

    fprintf( _pFile, "%%%%Orientation: %s\n",
             ( _fLandscape == TRUE) ? "Landscape" : "Portrait");

    fprintf( _pFile, "%%%%Pages: 1\n");
    fprintf( _pFile, "%%%%PageOrder: Ascend\n");

    fprintf( _pFile, "%%%%Title: %s\n", _szTitle);

    /* Where do we get our version from - should be somewhere? */

    fprintf( _pFile, "%%%%Version: 0.6.0\n");

    fprintf( _pFile, "%%%%EndComments\n");
}

void
AmPostScriptOutputClass::PrintDefaults()
{
    fprintf( _pFile, "%%%%BeginDefaults\n");
    fprintf( _pFile, "%%%%EndDefaults\n");
}

void
AmPostScriptOutputClass::PrintProlog()
{
    fprintf( _pFile, "%%%%BeginProlog\n");
    fprintf( _pFile, "%%%%EndProlog\n");
}

void
AmPostScriptOutputClass::PrintSetup()
{
#if 0
    int   n;
    char  sz[ 4096], szPath[ PATHLEN];
    FILE  *pFilePrinter;
#endif

    fprintf( _pFile, "%%%%BeginSetup\n");

    /* Setup the page that we want. */

    if ( _fLandscape) {
        fprintf( _pFile, "<< /PageSize [%d %d] >> setpagedevice\n",
                 _iHeightPage, _iWidthPage);
    } else {
        fprintf( _pFile, "<< /PageSize [%d %d] >> setpagedevice\n",
                 _iWidthPage, _iHeightPage);
    }

    /* 
     * A printer file is used for setting the printer-specific page size,
     *   like for the HP 650C or HP 750C.
     */


#if 0

    /* For now let's not use this, since it actually introduces problems,
       and doesn't give us anything new: */

    if ( _szPrinter != NULL) {
        sprintf( szPath, "%s/%s/%s", getenv( CONFIG_ENV), AM_PRINTER_DIR,
                 _szPrinter);
  
        pFilePrinter = fopen( szPath, "r");

        if ( pFilePrinter == NULL) {
            AmPrintWarn( "Could not read printer file \"%s\"", szPath);
        } else {
            n = fread( sz, 1, sizeof( sz), pFilePrinter);
            sz[ n] = '\0';
            fclose( pFilePrinter);

        /* Switch the page sizes if the output is in landscape mode. */

            if ( _fLandscape) {
                fprintf( _pFile, sz, _iHeightPage, _iWidthPage);
            } else {
                fprintf( _pFile, sz, _iWidthPage, _iHeightPage);
            }
        }
    }

#endif

    /* Set the default line width. */

    fprintf( _pFile, "0 setlinewidth\n");
 
    /* Set the default font. */
  
    fprintf( _pFile, "/Times-Roman findfont 7 scalefont setfont\n");
 
    /* Definition for image drawing function. */
 
    fprintf( _pFile, "/imagedraw\n");
    fprintf( _pFile, "{\n");
    fprintf( _pFile, "  /matrix exch cvlit def\n");
    fprintf( _pFile, "  /depth exch cvlit def\n");
    fprintf( _pFile, "  /height exch cvlit def\n");
    fprintf( _pFile, "  /width exch cvlit def\n");
    fprintf( _pFile, "  /picstr width string def\n");
    fprintf( _pFile, "  width height depth matrix\n");
    fprintf( _pFile, "  { currentfile picstr readhexstring pop }\n");
    fprintf( _pFile, "  false 3 colorimage\n");
    fprintf( _pFile, "} def\n");
 
    /* Definition of arc drawing function */
 
    fprintf( _pFile, "/elliparc\n");
    fprintf( _pFile, "{\n");
    fprintf( _pFile, "  /angle2 exch cvlit def\n");
    fprintf( _pFile, "  /angle1 exch cvlit def\n");
    fprintf( _pFile, "  /height exch cvlit def\n");
    fprintf( _pFile, "  /width exch cvlit def\n");
    fprintf( _pFile, "  /y exch cvlit def\n");
    fprintf( _pFile, "  /x exch cvlit def\n");
    fprintf( _pFile, "  x y translate\n");
    fprintf( _pFile, "  1 height width div scale\n");
    fprintf( _pFile, "  width 2 div\n");       /* X: center */
    fprintf( _pFile, "  width 2 div\n");       /* Scale turns it into height. */
    fprintf( _pFile, "  width 2 div\n");       /* Radius */
    fprintf( _pFile, "  angle1 angle2 arc\n"); /* Ang1 Ang2 Arc */
    fprintf( _pFile, "} def\n");

    fprintf( _pFile, "%%%%EndSetup\n");
}

void
AmPostScriptOutputClass::PrintPageSetup( int iPage)
{
    fprintf( _pFile, "%%%%Page: (Page %d) %d\n", iPage, iPage);
    fprintf( _pFile, "%%%%BeginPageSetup\n");
    fprintf( _pFile, "/pgsave save def\n");

    /*
     * Use the window and postscript output dimensions to determine the
     *   scaling to use for the drawing to this window.
     */
 
    if ( _fLandscape == FALSE) {

        /*
         * Shift the coordinate system from the lower left to the upper
         *   rigth corner.
         */

        fprintf( _pFile, "0 %d translate\n", _iHeightPage);

    } else {

        /* Rotate the page so that it is in landscape mode. */

        fprintf( _pFile, "4 dict begin\n");
        fprintf( _pFile, "0 0 %d %d\n", _iWidthPage, _iHeightPage);
        fprintf( _pFile, "/ury exch def /urx exch def\n");
        fprintf( _pFile, "/lly exch def /llx exch def\n");
        fprintf( _pFile, "90 rotate\n");
        fprintf( _pFile, "llx neg llx urx sub lly sub translate\n");
        fprintf( _pFile, "end\n");

        /*
         * Shift the coordinate system from the lower left to the upper
         *   rigth corner.
         */

        fprintf( _pFile, "0 %d translate\n", _iWidthPage);
    }

    /*
     * Set up the scaling from our coordinates into the printers
     *   coordinates.
     */

    fprintf( _pFile, "%d %d div ", BASEDPI, _pDisplay->GetDPI());
    fprintf( _pFile, "%d %d div neg ", BASEDPI, _pDisplay->GetDPI()); 
    fprintf( _pFile, "scale\n");

    /* Turn on stroke adjustment for uniform line widths. */

    fprintf( _pFile, "true setstrokeadjust\n");

    fprintf( _pFile, "%%%%EndPageSetup\n");
}

void
AmPostScriptOutputClass::PrintPageTrailer()
{
    fprintf( _pFile, "pgsave restore\n");
    fprintf( _pFile, "showpage\n");
    fprintf( _pFile, "%%%%PageTrailer\n");
}

void
AmPostScriptOutputClass::PrintTrailer()
{
    fprintf( _pFile, "%%%%Trailer\n");
    fprintf( _pFile, "%%%%EOF\n");
}

/* Create an PostScript window class. */

AmPostScriptOutputClass::AmPostScriptOutputClass( FILE            *pFile,
                                              float           dWidthInInches,
                                              float           dHeightInInches,
                                              AmPostScriptClass *pPostScript,
                                              AnvilBoolean         fLandscape,
                                              char            *szPrinter)
    : AmWindowClass( pPostScript)
{
    /* Initializations. */

    _pFile = pFile;
    _dWidthInInches = dWidthInInches;
    _dHeightInInches = dHeightInInches;
    _fLandscape = fLandscape;
    _szPrinter = strdup( szPrinter);

    _pPostScriptGC = NULL;

    _iWidthPage = (int)(_dWidthInInches * BASEDPI);
    _iHeightPage = (int)(_dHeightInInches * BASEDPI);

    /* The title should be user settable. */

    _szTitle = strdup( "unknown");

    PrintHeader();      /* Start the document. */
    PrintDefaults();
    PrintProlog();
    PrintSetup();
    PrintPageSetup( 1); /* Start on the AmFirst (and only) page. */
}

/* Destructor */

AmPostScriptOutputClass::~AmPostScriptOutputClass()
{
    fprintf( stderr, "In ~AmPostScriptOutputClass()\n");

    PrintPageTrailer(); /* Finish the page.     */
    PrintTrailer();     /* Finish the document. */

    fclose( _pFile);

    /* Free some of the resources used. */

    if ( _szPrinter != NULL) {
        FREE( _szPrinter);
    }

    if ( _szTitle != NULL) {
        FREE( _szTitle);
    }
}


/* Return the file pointer for the output file */

FILE *
AmPostScriptOutputClass::getFp()
{
    return( _pFile);
}

/* Return the output dimensions for a postscript output */

void AmPostScriptOutputClass::GetDims( int *piWidth,
                                     int *piHeight)
{
    *piWidth = (int)(_dWidthInInches * _pDisplay->GetDPIBase());
    *piHeight = (int)(_dHeightInInches * _pDisplay->GetDPIBase());
}

/* create a new GC */

AmGCClass *
AmPostScriptOutputClass::CreateGC()
{
    return(( AmGCClass *) new AmPostScriptGCClass( this));
}

/* get the geometry of a postscript window */

void
AmPostScriptOutputClass::GetWindowGeometry( int *px,
                                          int *py,
                                          int *piWidth,
                                          int *piHeight)
{
    *px = 0;
    *py = 0;
    *piWidth = (int)(_dWidthInInches * _pDisplay->GetDPIBase());
    *piHeight = (int)(_dHeightInInches * _pDisplay->GetDPIBase());
}

/* If the GC being called with is different from the current GC, then output
   the changed values.  If it's the same, then just output the dirty values */

void
AmPostScriptOutputClass::ApplyGCAttributes( AmPostScriptGCClass *pPostScriptGC)
{
    AmColorCellRec   oldFg, newFg, oldBg, newBg;
    int            oldLineWidth, newLineWidth, oldFontSize, newFontSize,
                   oldDashOffset, newDashOffset,
                   oldNumDashes, newNumDashes, ii;
    float          oldFontXScale, newFontXScale, oldFontYScale,
                   newFontYScale;
    AmFontWeightEnum oldFontWeight, newFontWeight;
    AmFontSlantEnum  oldFontSlant, newFontSlant;
    AmLineStyleEnum  oldLineStyle, newLineStyle;
    AmLineJoinEnum   oldLineJoin, newLineJoin;
    char           *oldFont, *newFont, *oldDashes, *newDashes;
    AmRasterOpEnum   oldRasterOp, newRasterOp;
    AmRectanglePtr   oldClipRectVec, newClipRectVec;
    int            numOldRects, numNewRects;
    AnvilBoolean        fClip;
    AmRegionClass	*oldRegion, *newRegion;

    if ( _pPostScriptGC == NULL) {

        _pPostScriptGC = pPostScriptGC;
        pPostScriptGC->OutputDirtyGCchanges( this);

    } else if ( pPostScriptGC != _pPostScriptGC) {

        /* Reset the fields, even though the values are the same, so the
           dirty flags get set, then output the dirty GC changes, and re-
           assign the GC */

        _pPostScriptGC->GetForegroundColor( &oldFg);
        pPostScriptGC->GetForegroundColor( &newFg);
        if ( !AmCompareColor( &newFg, &oldFg)) {
            pPostScriptGC->setFgDirty();
        }

        _pPostScriptGC->GetBackgroundColor( &oldBg);
        pPostScriptGC->GetBackgroundColor( &newBg);
        if ( !AmCompareColor( &newBg, &oldBg)) { 
            pPostScriptGC->setBgDirty();
        }

        /* Do the line width. */

        oldLineWidth = (int)(_pPostScriptGC->GetLineWidth());
        newLineWidth = (int)(pPostScriptGC->GetLineWidth());
        if ( oldLineWidth != newLineWidth) {
            pPostScriptGC->setLineWidthDirty();
        }
    
        /* Check the dashes and number of dashes. */

        _pPostScriptGC->GetDashes( &oldDashes, &oldNumDashes, &oldDashOffset);
        pPostScriptGC->GetDashes( &newDashes, &newNumDashes, &newDashOffset);
        if (( oldNumDashes != newNumDashes) ||
            ( oldDashOffset != newDashOffset) ||
            ( !memcmp(( void *) oldDashes,
                      ( void *) newDashes, oldNumDashes))) {
            pPostScriptGC->setDashDirty();
        }

        /* Do the line style. */

        oldLineStyle = _pPostScriptGC->GetLineStyle();
        newLineStyle = pPostScriptGC->GetLineStyle();
        if ( oldLineStyle != newLineStyle) {
            pPostScriptGC->setLineStyleDirty();
        }

        /* Do the line join. */

        oldLineJoin = _pPostScriptGC->GetLineJoin();
        newLineJoin = pPostScriptGC->GetLineJoin();
        if ( oldLineJoin != newLineJoin) {
            pPostScriptGC->setLineJoinDirty();
        }

        /* Do the font. */

        _pPostScriptGC->GetFont( &oldFont, &oldFontSize, &oldFontWeight,
                                 &oldFontSlant, &oldFontXScale, &oldFontYScale);
        pPostScriptGC->GetFont( &newFont, &newFontSize, &newFontWeight,
                                 &newFontSlant, &newFontXScale, &newFontYScale);
        if (( strcmp( oldFont, newFont)) ||
            ( oldFontSize != newFontSize) ||
            ( oldFontWeight != newFontWeight) ||
            ( oldFontSlant != newFontSlant) ||
            ( oldFontXScale != newFontXScale) ||
            ( oldFontYScale != newFontYScale)) {
            pPostScriptGC->setFontDirty();
        }
        free( oldFont);
        free( newFont);

	/* do the clip region */

	oldRegion = _pPostScriptGC->GetClipRegion();
	newRegion = pPostScriptGC->GetClipRegion();
	if (( oldRegion != NULL) && ( newRegion != NULL)) { 
	    if ( !oldRegion->EqualRegion( newRegion)) {
		pPostScriptGC->setClipRegionDirty();
	    }
	} else if (( oldRegion != NULL) || ( newRegion != NULL)) {
	    pPostScriptGC->setClipRegionDirty();
	}

        /* Do the clip rectangle. */

        _pPostScriptGC->GetClipRectangles( &oldClipRectVec, &numOldRects);
        pPostScriptGC->GetClipRectangles( &newClipRectVec, &numNewRects);
        fClip = ( numNewRects == numOldRects);
        for ( ii = 0; ( ii < numNewRects) && fClip; ii++) {
            fClip = (( oldClipRectVec[ ii].x == newClipRectVec[ ii].x) &&
                     ( oldClipRectVec[ ii].y == newClipRectVec[ ii].y) &&
                     ( oldClipRectVec[ ii].width ==
                       newClipRectVec[ ii].width) &&
                     ( oldClipRectVec[ ii].height == 
                       newClipRectVec[ ii].height));
        }

        /* Free the temporary clip rectangle vectors. */

        free( oldClipRectVec);
        free( newClipRectVec);
        if ( !fClip) {
            pPostScriptGC->setClipRectDirty();
        }

        /* Check the raster operation. */

        oldRasterOp = _pPostScriptGC->GetRasterOp();
        newRasterOp = pPostScriptGC->GetRasterOp();
        if ( oldRasterOp != newRasterOp) {
            pPostScriptGC->setRasterOpDirty();
        }

        /* Output all the fields which have dirty flags set, and set the
           internal GC to the current GC. */

        pPostScriptGC->OutputDirtyGCchanges( this);
        _pPostScriptGC = pPostScriptGC;

    } else {

        /* Same GCs output dirty fields. */

        pPostScriptGC->OutputDirtyGCchanges( this);
    }
}


/* Don't do anything for an invalidate */

/* ARGSUSED */
void
AmPostScriptOutputClass::InvalidateRect( AmRectangleRec const *pRect)
{
}

/* or a clear... */

/* ARGSUSED */
void
AmPostScriptOutputClass::ClearRect( AmRectangleRec const *pRect)
{
}

/* or an update... */ 

void
AmPostScriptOutputClass::Update()
{
}

/* ARGSUSED */
AmWindowClass *
AmPostScriptClass::CreateBuffer( AmWindowClass	*pWindow,
			       int		cx,
			       int		cy)
{
    return( NULL);
}

  
/* ARGSUSED */
AmWindowClass *
AmPostScriptClass::CreateWin( AmWindowClass	*pWindow,
			    int		x,
			    int		y,
			    int		cx,
			    int		cy)
{
    return( NULL);
}

/* ARGSUSED */
void
AmPostScriptOutputClass::ResizeWindow( int	cx,
				     int	cy)
{
}

/* ARGSUSED */
void
AmPostScriptOutputClass::MoveWindow( int	x,
				   int	y)
{
}
