#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Scale.h>
#include <Xm/PushB.h>
#include <stdarg.h>
#include <stdio.h>
#include <anvil/plot.h>
#include <anvil/motif.h>
#include <anvil/resource.h>
#include <anvil/plot_cback.h>
#include <anvil/copyobj.h>
#include <anvil/motifdisplay.h>
#include <anvil/magic.h>

/* magnifying glass magic cursor.  Derived from magic plot class, allows
   the user to set the magnification factor for the magic plot using 
   a slider in a pop-up dialog */

class MagnifyClass : public AmMagicPlotClass {
private:
    float		factor;
    Widget		dialog;
public:
    MagnifyClass( AmMotifDisplayClass	*motifDisplay,
		  AmPlotClass		*sourcePlot,
		  AnvilBoolean		callOnDrag);
    void 		SetMagnify( float value) { factor = value; };
    float		GetMagnify() { return( factor); };
    Widget		GetDialog() { return( dialog); };
    void		SetDialog( Widget dialogArg) { dialog = dialogArg; };
    virtual void	SetRect( AmRectangleRec const *rectArg);
    virtual 		~MagnifyClass();
};

/*************************************************************************
 * 
 * Description:
 * 	update the magnified plot when the rectangle is changed. 
 *	virtual functiob overrides the magic plot implementation.
 * Arguments:
 * 	rectArg - rectangle that the magnifying glass is being dragged
 *	or resized to.
 * 
 *************************************************************************/

void 
MagnifyClass::SetRect( AmRectangleRec const *rectArg)
{
    float	xmin, ymin, xmax, ymax, cx, cy, wx, wy;
    int		eventId = AM_COORDS_CHANGE;

AN_DEBUG( 50, ( "calling MagnifyClass::SetRect %d, %d, %d, %d\n",
		rectArg->x, rectArg->y, rectArg->width, rectArg->height));

    rect = *rectArg;
    GetSourceCoords( &xmin, &ymin, &xmax, &ymax);
    cx = ( xmax + xmin)/2;
    cy = ( ymax + ymin)/2;
    wx = ( xmax - xmin)/( factor*2);
    wy = ( ymax - ymin)/( factor*2);
    plot->SetCoords( cx - wx, cy - wy, cx + wx, cy + wy);
    plot->Update( &eventId, 1);
}

/*************************************************************************
 * 
 * Description:
 * 	callback to set the magnification scale from the scale widget
 *	in the dialog.
 * Arguments:
 * 	scale - scale widget in the dialog.
 *	magnify - magnifying glass cursor magic window.
 *	callData - unused.
 * 
 *************************************************************************/

/* ARGSUSED */

static void 
setMagnifyScale( Widget		scale,
		 MagnifyClass	*magnify,
		 XtPointer	callData)
{
    int	value;

    XmScaleGetValue( scale, &value);
    magnify->SetMagnify( value);
    magnify->Update();
    magnify->DrawWrapper( magnify->GetRect());
}

/*************************************************************************
 * 
 * Description:
 * 	destructor: blow away the dialog if it wasn't set to NULL already
 *	by the its destroy callback.
 * 
 *************************************************************************/

MagnifyClass::~MagnifyClass()
{
    if ( dialog != NULL) {
	XtDestroyWidget( dialog);
    }
}

/*************************************************************************
 * 
 * Description:
 * 	dialog destroy callback.  Set the dialog to NULL and call
 *	the magnifying glass's destructor
 * Arguments:
 *	dlog - dialog being destroyed.
 *	magnify - magnifying glass magic window 
 *	callData - unused.
 * 
 *************************************************************************/

/* ARGSUSED */

static void 
destroyMagnifyCallback( Widget		dlog,	
			MagnifyClass	*magnify,
			XtPointer	callData)
{
    magnify->SetDialog( NULL);
    delete magnify;
}

/*************************************************************************
 * 
 * Description:
 * 	derived constructor from AmMagicPlotClass for Magnifying glass cursor 
 * Arguments:
 *	motifDisplay - display the magic window is attached to.
 *	sourcePlot - plot being magnified.
 *	callOnDrag - update on drag, or only on placement.
 * 
 *************************************************************************/

/* ARGSUSED */

MagnifyClass::MagnifyClass( AmMotifDisplayClass	*motifDisplay,
			    AmPlotClass		*sourcePlot,
			    AnvilBoolean		callOnDrag) :
				AmMagicPlotClass( motifDisplay, 
						sourcePlot,
						callOnDrag)
{
    Widget	scale, closeBtn;
    Arg		args[ 20];
    int		n;

    factor = 2.0;
    Update();
    DrawWrapper( &rect);

/* create a dialog for setting the zoom factor */

    n = AmMakeArgList( args,
                     XmNdeleteResponse, XmDESTROY,
                     XmNautoUnmanage, FALSE,
                     NULL);
    dialog = XmCreateFormDialog( motifDisplay->GetShell(), "magnify", args, n);
    scale = XtVaCreateManagedWidget( "magnifyScale", xmScaleWidgetClass,
				     dialog,
				     XmNorientation, XmHORIZONTAL,
				     XmNminimum, 1,
				     XmNmaximum, 20,
				     XmNvalue, 2,
				     StickLeft, StickTop, StickBottom,
				     XmNshowValue, TRUE,
				     NULL);
    closeBtn = XtVaCreateManagedWidget( "closeBtn", xmPushButtonWidgetClass,
					dialog,
					RightOf( scale),
					StickTop, StickBottom,
					NULL);
    XtManageChild( dialog);

/* add callbacks to destroy the dialog and set the scale */

    XtAddCallback( closeBtn, XmNactivateCallback,
		   ( XtCallbackProc) AmPopdownCback, dialog);
    XtAddCallback( scale, XmNvalueChangedCallback,
		   ( XtCallbackProc) setMagnifyScale, this);
    XtAddCallback( dialog, XmNdestroyCallback,
		   ( XtCallbackProc) destroyMagnifyCallback, this);
}

/*************************************************************************
 * 
 * Description:
 * 	callback to set up a magnifying glass for an object 
 * Arguments:
 *	btn - button to call callback.
 *	motifDisplay - display the magic window is attached to.
 *	plot - plot being magnified.
 *	callData - unused.
 * 
 *************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmMagnifyGlassCallback( Widget		btn,
		      AmMotifDisplayClass	*motifDisplay,
		      AmPlotClass		*plot,
		      XtPointer		callData)
{
    new MagnifyClass( motifDisplay, plot, 
		      AmGetDefault( "magnify.drag", ( AnvilBoolean) FALSE));
    return( TRUE);
}
