#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdio.h>
#include <math.h>
#include <Xm/Xm.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Scale.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/DrawingA.h>
#include <anvil/errlib.h>
#include <anvil/va_cback.h>
#include <anvil/events.h>
#include <anvil/motif.h>
#include <anvil/plot_cback.h>
#include <anvil/macros.h>
#include <anvil/utility.h>
#include <anvil/colorsel.h>
#include <anvil/slick.h>
#include <anvil/slickP.h>

typedef struct colorData_ {
    AmMotifDisplayClass   *motifDisplay;
    AmPlotClass     	*plot;
} colorDataRec, *colorDataPtr;


static AmActionDefRec actionAreaDef[] = {
    { "okBtn", NULL, NULL },
    { "closeBtn", ( XtCallbackProc) AmPopdownCback, NULL },
};

static void 
setPlotFlagToggleCallback( Widget		tgl,
			   AmMotifDisplayClass	*motifDisplay,
			   AmPlotClass		*plot,
			   char			*fieldName,
			   XtPointer)
{
    plot->SetVaFields( fieldName, XmToggleButtonGetState( tgl), NULL);
    motifDisplay->Flush();
}

/**************************************************************************
 *
 * Description:
 *	Given a field name, and a toggle, add a callback which sets
 *	the field to a boolean value and flushes the passed motifDisplay.
 * Arguments:
 *	tgl - xmToggleButtonWidgetClass toggle.
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 *	fieldName - field to set 
 * Returns:
 *	TRUE
 * Modifications:
 *	should take a screen object, not a plot.
 *
 ***************************************************************************/

AnvilBoolean
AmSetPlotFlagToggle( Widget	     tgl,
  		   AmMotifDisplayClass *motifDisplay,
		   AmPlotClass	     *plot,
		   char		     *fieldName)
{
    AnvilBoolean	flag;

    plot->GetVaFields( fieldName, &flag, NULL);
    XmToggleButtonSetState( tgl, flag, FALSE);
    AmAddVaCallback( tgl, XmNvalueChangedCallback,
		   ( VarargsCallback) setPlotFlagToggleCallback,
		   motifDisplay, plot, fieldName, 
		   END_LIST);
    return( TRUE);
}

static void 
setPlotEnumToggleCallback( Widget		tgl,
			   AmMotifDisplayClass	*motifDisplay,
			   AmPlotClass		*plot,
			   char			*fieldName,
			   int			enumValue,
			   XtPointer)
{
    if ( XmToggleButtonGetState( tgl)) {
	plot->SetVaFields( fieldName, enumValue, NULL);
    }
    motifDisplay->Flush();
}

/**************************************************************************
 *
 * Description:
 * 	set the enumeration value for the field if the toggle is depressed.  
 *  	Intended for use in radio boxes 
 * Arguments:
 *	tgl - xmToggleButtonWidgetClass toggle.
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 *	fieldName - field to set 
 *	enumValue - enumeration value to set for this toggle.
 * Returns:
 *	TRUE
 * Modifications:
 *	should take a screen object, not a plot.
 *
 ***************************************************************************/

AnvilBoolean
AmSetPlotEnumToggle( Widget	     		tgl,
  		     AmMotifDisplayClass 	*motifDisplay,
		     AmPlotClass	     	*plot,
		     char		     	*fieldName,
		     int		     	enumValue)
{
    int	lastValue;

    plot->GetVaFields( fieldName, &lastValue, NULL);
    if ( lastValue == enumValue) {
	XmToggleButtonSetState( tgl, TRUE, FALSE);
    }
    AmAddVaCallback( tgl, XmNvalueChangedCallback,
		   ( VarargsCallback) setPlotEnumToggleCallback,
		   motifDisplay, plot, fieldName, enumValue,
		   END_LIST);
    return( TRUE);
}

static void 
setPlotValueScaleCallback( Widget		scl,
			   AmMotifDisplayClass	*motifDisplay,
			   AmPlotClass		*plot,
			   char			*fieldName,
			   XtPointer)
{
    int	value;

    XmScaleGetValue( scl, &value);
    plot->SetVaFields( fieldName, value, NULL);
    motifDisplay->Flush();
}

/**************************************************************************
 *
 * Description:
 * 	set the value for an attribute of a plot based on a scale's value 
 * Arguments:
 *	scale - xmScaleWidgetClass 
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 *	fieldName - field to set 
 * Returns:
 *	TRUE
 * Modifications:
 *	should take a screen object, not a plot.
 *
 ***************************************************************************/

AnvilBoolean
AmSetPlotValueScale( Widget            scl,
 		   AmMotifDisplayClass *motifDisplay,
		   AmPlotClass         *plot,
		   char              *fieldName)
{
    int	value;

    plot->GetVaFields( fieldName, &value, NULL);
    XmScaleSetValue( scl, value);
    AmAddVaCallback( scl, XmNvalueChangedCallback,
		   ( VarargsCallback) setPlotValueScaleCallback,
		   motifDisplay, plot, fieldName, 
		   END_LIST);

    return( TRUE);
}

static void 
setPlotValueFloatScaleCallback( Widget			scl,
			        AmMotifDisplayClass	*motifDisplay,
			        AmPlotClass		*plot,
			        char			*fieldName,
			        XtPointer)
{
    int		value;
    short	decimalPoints;
    float	plotValue;

    XmScaleGetValue( scl, &value);
    XtVaGetValues( scl, 
		   XmNdecimalPoints, &decimalPoints,
		   NULL);
    plotValue = (( float) value)/pow( 10.0, ( float) decimalPoints);
    plot->SetVaFields( fieldName, plotValue, NULL);
    motifDisplay->Flush();
}

/**************************************************************************
 *
 * Description:
 * 	set the value for an float attribute of a plot based on a 
 *	scale's value 
 * Arguments:
 *	scale - xmScaleWidgetClass 
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 *	fieldName - field to set 
 * Returns:
 *	TRUE
 * Modifications:
 *	should take a screen object, not a plot.
 *
 ***************************************************************************/

AnvilBoolean
AmSetPlotFloatValueScale( Widget            	scl,
			  AmMotifDisplayClass 	*motifDisplay,
			  AmPlotClass         	*plot,
			  char              	*fieldName)
{
    float	value;
    int		scaleValue;
    short	decimalPoints;

    plot->GetVaFields( fieldName, &value, NULL);
    XtVaGetValues( scl, 
		   XmNdecimalPoints, &decimalPoints,
		   NULL);
    scaleValue = (int)(value*pow( 10.0, ( float) decimalPoints));
    XmScaleSetValue( scl, scaleValue);
    AmAddVaCallback( scl, XmNvalueChangedCallback,
		   ( VarargsCallback) setPlotValueFloatScaleCallback,
		   motifDisplay, plot, fieldName, 
		   END_LIST);

    return( TRUE);
}

/**************************************************************************
 *
 * Description:
 * 	create a scale for an image plot's transparency, and set the callback 
 * Arguments:
 *	parent - parent widget.
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 * Returns:
 *	TRUE
 *
 ***************************************************************************/

Widget
AmTransparencyScale( Widget			parent,
		     AmMotifDisplayClass	*motifDisplay,
		     AmImagePlotClass		*plot)
{
    Widget	scale;
    float	transparency;
    int		transval;

    plot->GetVaFields( AmNtransparency, &transparency, NULL);
    transval = (int)transparency*100;
    scale = XtVaCreateManagedWidget( "transparencyScl", xmScaleWidgetClass, 
				     parent,
				     XmNorientation, XmHORIZONTAL,
				     XmNdecimalPoints, 2,
				     XmNminimum, 0,
				     XmNmaximum, 100,
				     XmNvalue, transval,
				     XmNshowValue, TRUE,
				     NULL);
    AmAddVaCallback( scale, XmNvalueChangedCallback, 
		   ( VarargsCallback) AmScaleTransparencyCallback,
		   motifDisplay, plot, END_LIST);
    return( scale);
}
				     
/**************************************************************************
 *
 * Description:
 * 	scale for the transparency for a plot 
 *	motifDisplay - motif display containing the object.
 *	plot - plot to set the value for.
 * Returns:
 *	TRUE
 *
 ***************************************************************************/

AnvilBoolean
AmScaleTransparencyCallback( Widget              scale,
			     AmMotifDisplayClass *motifDisplay,
			     AmPlotClass         *plot,
			     XtPointer)
{
    int		min, max, val;
    float	transparency;

    XtVaGetValues( scale, 
		   XmNminimum, &min,
		   XmNmaximum, &max,
		   XmNvalue, &val,
		   NULL);
    transparency = (( float) val - min)/(( float) max - min);
AN_DEBUG( 70, ("setting transparency = %.2f\n", transparency));
    plot->SetVaFields( AmNtransparency, transparency, NULL);
    motifDisplay->Flush();

    return( TRUE);
}

/* figure out the horizontal and vertical angles from the mouse position,
   update the readout, and send an event to refresh the drawing */

static void 
lightShadeDragHandler( Widget		area,
		       float		*horizAngle,
		       float		*vertAngle,
		       Widget		readoutLbl,
		       XMotionEvent	*event,
		       AnvilBoolean		*)
{
    Dimension	width, height;
    float	AmDist, maxdist;
    char	buffer[ 256];
    XmString	mstr;

    XtVaGetValues( area,
		   XmNwidth, &width,
		   XmNheight, &height,
		   NULL);
    *horizAngle = 90.0 - RADTODEG( atan2( height/2 - event->y, 		
					  event->x - width/2));
    AmDist = sqrt(( event->x - width/2)*( event->x - width/2) +
		( event->y - height/2)*( event->y - height/2));
    maxdist = MIN( width/2, height/2);
    *vertAngle = 90.0 - 90.0*AmDist/maxdist;
    sprintf( buffer, "horiz = %.0f\260, vert = %.0f\260\n",
	     *horizAngle, *vertAngle);
    mstr = XmStringCreateLtoR( buffer, XmFONTLIST_DEFAULT_TAG);
    XtVaSetValues( readoutLbl, XmNlabelString, mstr, NULL);
    XmStringFree( mstr);
    AmSendExpose( XtDisplay( area), XtWindow( area));
}

/* remove the handlers on the release */

static void 
lightShadeReleaseHandler( Widget	area,
			  float		*horizAngle,
			  float		*vertAngle,
			  Widget	readoutLbl,
			  XButtonEvent	*event,
			  AnvilBoolean	*)
{
    AmRemoveVaHandler( area, PointerMotionMask, FALSE,
		     ( VarargsHandler) lightShadeDragHandler, 
		     horizAngle, vertAngle, readoutLbl, END_LIST);
    AmRemoveVaHandler( area, ButtonReleaseMask, FALSE,
		     ( VarargsHandler) lightShadeReleaseHandler, 
		     horizAngle, vertAngle, readoutLbl, END_LIST);
}

/* kick off the drag and release handlers when the user presses the 
   mouse in the light shade drawing area */

static void 
lightShadePressHandler( Widget		area,
			float		*horizAngle,
			float		*vertAngle,
			Widget		readoutLbl,
			XButtonEvent	*event,
			AnvilBoolean		*)
{
    AmAddVaHandler( area, PointerMotionMask, FALSE,
		  ( VarargsHandler) lightShadeDragHandler, 
		  horizAngle, vertAngle, readoutLbl, END_LIST);
    AmAddVaHandler( area, ButtonReleaseMask, FALSE,
		  ( VarargsHandler) lightShadeReleaseHandler, 
		  horizAngle, vertAngle, readoutLbl, END_LIST);
}

/* callback for the standalone dialog.  Set the light shade horizontal
   and vertical angles. */

static void 
lightShadeOkCallback( int			eventId,
		      Widget			formDlog,
		      AmImagePlotClass		*plot,
		      float			*horizAngle,
		      float			*vertAngle,
		      XtPointer)
{
    XtDestroyWidget( XtParent( formDlog));
    plot->SetVaFields( AmNlightShadeAngle, *horizAngle,
		       AmNlightShadeHeight, *vertAngle,
		       NULL);
}

/* draw the stuff for the light shade horizontal and vertical angle */

static void
AmDrawLightShadeCallback( Widget	area,
			  float		*horizAngle,
			  float		*vertAngle,
			  XtPointer)
{
    Dimension	width, height;
    Display	*display = XtDisplay( area);
    GC		gc = DefaultGC( display, DefaultScreen( display));
    int		x, y;
    float	maxdist;

    XtVaGetValues( area,
		   XmNwidth, &width,
		   XmNheight, &height,
		   NULL);
    XClearWindow( display, XtWindow( area));
    XDrawArc( display, XtWindow( area), gc,
	      width/10, height/10, 8*width/10, 8*height/10, 0, 360*64);
    XDrawArc( display, XtWindow( area), gc,
	      3*width/10, 3*height/10, 4*width/10, 4*height/10, 0, 360*64);
    maxdist = MIN( width/2, height/2);
    x = (int)(width/2 + cos( DEGTORAD( 90.0 - *horizAngle))*
		  (( 90.0 - *vertAngle)/90.0)*maxdist);
    y = (int)(height/2 - sin( DEGTORAD( 90.0 - *horizAngle))*
		   (( 90.0 - *vertAngle)/90.0)*maxdist);
    XFillRectangle( display, XtWindow( area), gc, x - 2, y - 2, 4, 4);
    XDrawLine( display, XtWindow( area), gc, 0, height/2, width, height/2);
    XDrawLine( display, XtWindow( area), gc, width/2, 0, width/2, height);
}

/*************************************************************************
 * 
 * Description:
 *	create the light shade dialog.
 * Arguments:
 *	parent - parent widget.
 *	plot - image plot to change the light shading for.
 *	dialog - returned dialog.
 *	okBtn - returned OK button
 *
 *************************************************************************/

void 
AmCreateLightShadeDialog( Widget		parent,
			  AmImagePlotClass	*plot,
			  Widget		*dialog,
			  Widget		*okBtn)
{
    AmActionRec         actionArea;
    Widget		area, actionForm, readoutLbl;
    int			n;
    Arg			args[ 20];
    static float	horizAngle, vertAngle;
    char		buffer[ 128];
    XmString		mstr;

    *dialog = NULL;
    *okBtn = NULL;
    if ( plot->GetObjectType() != AmIMAGE_PLOT) {
	AmAddResourceToLog( "lightShadeNotImagePlot");
	return;
    }


/* since we're modal, we can get away with static variables */

    plot->GetVaFields( AmNlightShadeAngle, &horizAngle,
		       AmNlightShadeHeight, &vertAngle,
		       NULL);

/* create a form dialog with a drawing area and OK and cancel buttons */

    n = AmMakeArgList( args, 
		       XmNdeleteResponse, XmDESTROY,
		       XmNdialogStyle, XmDIALOG_PRIMARY_APPLICATION_MODAL,
		       NULL);
    *dialog = XmCreateFormDialog( parent, "lightShadeForm", args, n);

/* set up the client data for the popdown callback, and the pointer
   for the OK button */

    actionAreaDef[ 1].clientData = *dialog;
    actionForm = AmMakeActionArea( *dialog,
                     actionAreaDef, ELEMENTS( actionAreaDef), &actionArea);
    XtVaSetValues( *dialog, StickLeft, StickRight, StickBottom, NULL);
    *okBtn = actionArea.buttons[ 0];

/* create readout label with the current horizontal and vertical angles */

    sprintf( buffer, "horiz = %.0f\260, vert = %.0f\260\n",
	     horizAngle, vertAngle);
    mstr = XmStringCreateLtoR( buffer, XmFONTLIST_DEFAULT_TAG);
    readoutLbl = XtVaCreateManagedWidget( "readoutLbl", xmLabelWidgetClass,
					  *dialog,
					  AboveOf( actionForm),
					  StickLeft, StickRight,
					  XmNlabelString, mstr,
					  NULL);
    XmStringFree( mstr);

/* and the drawing area to show the angles */

    area = XtVaCreateManagedWidget( "lightShadeArea", xmDrawingAreaWidgetClass,
				    *dialog,
				    StickLeft, StickRight, StickTop,
                                    AboveOf( readoutLbl),
				    XmNwidth, 150, 
				    XmNheight, 150,
				    NULL);

/* add callbacks and handlers for the drawing area and buttons */

    AmAddVaHandler( area, ButtonPress, FALSE,
		  ( VarargsHandler) lightShadePressHandler, 
		  &horizAngle, &vertAngle, readoutLbl,
		  END_LIST);
    AmAddVaCallback( area, XmNexposeCallback,
		   ( VarargsCallback) AmDrawLightShadeCallback,
		   &horizAngle, &vertAngle, 
		   END_LIST);
    AmAddVaCallbackWidget( AM_PKG_ID, AM_OK, *dialog,
			   ( VarargsCallback) lightShadeOkCallback,
		   	   *dialog, plot, &horizAngle, &vertAngle, 
			   END_LIST);
    XtManageChild( *dialog);
}

/*************************************************************************
 * 
 * Description:
 *	create the light shade dialog.
 * Arguments:
 *	parent - parent widget.
 *	plot - image plot to change the light shading for.
 *	dialog - returned dialog.
 *	okBtn - returned OK button
 *
 *************************************************************************/

AnvilBoolean
AmLightShadeCallback( Widget			btn,
		      AmMotifDisplayClass	*motifDisplay,
		      AmImagePlotClass		*plot,
		      XtPointer)
{
    Widget	dialog, okBtn;

    AmCreateLightShadeDialog( AmGetWidgetShell( btn), plot, &dialog, &okBtn);
    if ( dialog != NULL) {
	AmSetOkButton( okBtn, plot);
	XtAddCallback( okBtn, XmNactivateCallback,
		       ( XtCallbackProc) AmUpdateMotifDisplayCallback,
		       motifDisplay);
	return( TRUE);
    } else {
	AmAddResourceToLog( "cantCreateLightShadeDialog");
	return( FALSE);
    }
}

/*************************************************************************
 * 
 * Description:
 *	restore the callback to bring up the light shade dialog
 *	when the dialog is destroyed 
 *
 *************************************************************************/

static void
setLightShadeDialogCallback( Widget		w,
			     Widget		button,
			     AmPlotClass	*plot,
			     char		**args,
			     XtPointer		callData)
{
    AmAddVaCallback( button, XmNactivateCallback,
		     ( VarargsCallback) AmLightShadeDialogCallback,
		     plot, args,
		     END_LIST);
}

/*************************************************************************
 * 
 * Description:
 *	light shade dialog intended for embedding in a slick subform.
 * Arguments:
 *	w - launching button
 *	plot - image plot to change the light shading for.
 *	args - string arguments (unused)
 *	callData - Widget call data.
 * Returns:
 *	TRUE if dialog was created successfully, FALSE otherwise.
 *
 *************************************************************************/

AnvilBoolean
AmLightShadeDialogCallback( Widget		w,
			    AmImagePlotClass	*plot,
			    char		**args,
			    XtPointer		callData)
{
    Widget	dialog, okBtn, parent = AmGetWidgetShell( w);

    AmCreateLightShadeDialog( parent, plot, &dialog, &okBtn);
    if ( dialog != NULL) {
	AmAddVaCallback( okBtn, XmNactivateCallback,
			 ( VarargsCallback) AmUnmanageSubDialogCallback,
			 w, dialog, 
			 END_LIST);

/* We remove and replace the callback to launch the dialog, because we
   keep the dialog around on OK for our parent slick. */

        AmRemoveVaCallback( w, XmNactivateCallback,
			    ( VarargsCallback) AmLightShadeDialogCallback,
			    plot, args, 
			    END_LIST);
	AmAddVaCallback( dialog, XmNdestroyCallback,
			 ( VarargsCallback) setLightShadeDialogCallback,
			 w, plot, args,
			 END_LIST);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateLightShadeDialog");
        return( FALSE);
    }
}
