#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/ToggleB.h>
#include <Xm/List.h>
#include <Xm/TextF.h>
#include <Xm/Scale.h>
#include <Xm/DrawingA.h>
#include <anvil/files.h>
#include <anvil/strlib.h>
#include <anvil/vector.h>
#include <anvil/errlib.h>
#include <anvil/motif.h>
#include <anvil/motifdisplay.h>
#include <anvil/va_cback.h>
#include <anvil/envdefs.h>
#include <anvil/plot_cback.h>
#include <anvil/linegraph.h>
#include <anvil/tick.h>
#include <anvil/textcheck.h>
#include <anvil/interp.h>
#include <anvil/userevents.h>
#include <anvil/slick.h>
#include <anvil/slickP.h>
#include <anvil/utility.h>

static AmActionDefRec actionAreaDef[] = {
    { "okBtn", NULL, NULL },
    { "closeBtn", ( XtCallbackProc) AmPopdownCback, NULL },
    { "helpBtn", NULL, NULL }
};

typedef struct rangeDialogData_ {
    Widget		minText;
    Widget		maxText;
    Widget		nonLinearTgl;
    Widget		formDlog;
    AmLineGraphClass	*scaleGraph;
    char		*minTag;
    char		*maxTag;
    char		*binvecTag;
    char		*lengthTag;
    int			numColors;
} rangeDialogDataRec, *rangeDialogDataPtr;

/* get the min and max values from the text widgets */

static void 
minmaxValues( Widget	minText,
	      Widget	maxText,
	      float	*min,
	      float	*max)
{
    char	*minStr, *maxStr;

    minStr = XmTextFieldGetString( minText);
    maxStr = XmTextFieldGetString( maxText);
    *min = atof( minStr);
    *max = atof( maxStr);
    XtFree( minStr);
    XtFree( maxStr);
}

/***************************************************************************
 * 
 * Description:
 * 	update the scale graph whenever the min or max text values change 
 * Arguments:
 *	text - one of the text widgets 
 *	minText, maxText - min and max textfields
 *	scaleDisplay - display manager to containing range graph.
 *	scaleGraph - range graph.
 *	minmaxMin, minmaxMax - pointer to current min and max values.
 *	callData - unused 
 *
 **************************************************************************/

/* ARGSUSED */

static void 
minmaxValueCallback( Widget		text,
		     Widget		minText,
		     Widget		maxText,
		     AmMotifDisplayClass	*scaleDisplay,
		     AmLineGraphClass	*scaleGraph,
		     float		*minmaxMin,
		     float		*minmaxMax,
		     XtPointer		callData)
{
    float	min, max, *binvec;
    int		numbins;

    scaleGraph->GetVaFields( AmNdata, &binvec, 
			     AmNrows, &numbins,
			     NULL);
    minmaxValues( minText, maxText, &min, &max);
    if ( min < max) {
	AmSetErrorColor( minText, FALSE);
	AmSetErrorColor( maxText, FALSE);

/* reset the range check values for the min and max value, since they update
   as the min and max are changed. */

	minmaxMin[ 1] = max;
	minmaxMax[ 0] = min;
	AmStretchVec( min, max, binvec, numbins);
	scaleGraph->SetVaFields( AmNdata, binvec, 
				 AmNmin, min,
				 AmNmax, max,
				 AmNcurrYmin, min,
				 AmNcurrYmax, max,
				 NULL);
	scaleDisplay->Update();
    } else {
	AmSetErrorColor( minText, TRUE);
	AmSetErrorColor( maxText, TRUE);
    }
}

/***************************************************************************
 * 
 * Description:
 * 	select a non-linear mapping and update the display appropriately 
 * Arguments:
 *	binvecList - XmList containing non-linear mappings.
 *	minText, maxText - min and max textfields
 *	scaleDisplay - display manager to containing range graph.
 *
 ***************************************************************************/

static void 
selectNonLinearMapping( Widget			binvecList,
			Widget			minText,
			Widget			maxText,
			AmMotifDisplayClass	*scaleDisplay,
			AmLineGraphClass		*scaleGraph)
{
    XmString	*items;
    int		count, numbins, numlines;
    char	*filename, *path, relPath[ PATHLEN];
    float	*binvec, *binvec2, *oldBinvec, min, max;

/* get the min and max values from the text fields */

    minmaxValues( minText, maxText, &min, &max);

/* if he selected a map file, then load it, 
   get the old bin vector, because we free it once its replaced, or we 
   stretch it to the new min max values */

    XtVaGetValues( binvecList,
		   XmNselectedItems, &items,
		   XmNselectedItemCount, &count,
		   NULL);
    scaleGraph->GetVaFields( AmNdata, &oldBinvec, 
			     AmNrows, &numbins,
			     NULL);
    if ( count != 0) {
	XmStringGetLtoR( items[ 0], XmFONTLIST_DEFAULT_TAG, &filename);
	sprintf( relPath, "%s/%s", AM_BINVEC_DIR, filename);
	path = AmGetEnvFile( CONFIG_ENV, relPath);
	XtFree( filename);
	binvec = AmReadFloatVec( path);
	if ( binvec == NULL) {
	    AmPrintError( "failed to open map file %s", filename);
	    return;
	}
	numlines = AmNumLines( path);

/* stretch it out to the number of displayed values */

	binvec2 = ( float *) malloc( sizeof( float)*numbins);

/* interpolate to the displayed min and max, then set it as the data 
   in the graph, and free the old vector, and the pre-interpolated vector */

	AmInterpolateVec( binvec, numlines, binvec2, numbins);
	AmStretchVec( min, max, binvec2, numbins);
	scaleGraph->SetVaFields( AmNdata, binvec2, NULL);
	free( oldBinvec);
	free( binvec);
    } else {

/* if there's no selection, just stretch the current vector in place */

	AmStretchVec( min, max, oldBinvec, numbins);
	scaleGraph->SetVaFields( AmNdata, oldBinvec, NULL);
    }
    scaleDisplay->Update();
}

/***************************************************************************
 * 
 * Description:
 *	toggle AmBetween linear and non-linear mappings.  Controls 
 *	sensitivity of the text and list widgets.
 * Arguments:
 *	binvecList - list of non-linear mappings.
 *	minText, maxText - text fields for min and max values.
 *	scaleDisplay - display manager to containing range graph.
 *	scaleGraph - range graph.
 *	callData - unused.
 *
 ***************************************************************************/

static void 
nonLinearTglCallback( Widget		tgl,
		      Widget		binvecList,
		      Widget		minText,
		      Widget		maxText,
		      AmMotifDisplayClass	*scaleDisplay,
		      AmLineGraphClass	*scaleGraph,
		      XtPointer		callData)
{
    float	min, max, *binvec, *oldBinvec;
    int		numbins;

/* depending on whether we're in linear or non-linear mode */

    if ( XmToggleButtonGetState( tgl)) {
	selectNonLinearMapping( binvecList, minText, maxText, 
				scaleDisplay, scaleGraph);
    } else {

/* get the min and max values from the text fields, and create a linear 
   vector */

	scaleGraph->GetVaFields( AmNdata, &oldBinvec,
				 AmNrows, &numbins,
				 NULL);
	minmaxValues( minText, maxText, &min, &max);
	binvec = AmMakeRangeVector( min, max, numbins);
	scaleGraph->SetVaFields( AmNdata, binvec,
				 NULL);
	free( oldBinvec);
	scaleDisplay->Update();
    }
}

/***************************************************************************
 * 
 * Description:
 *	callback to update the display when a non-linear mapping is
 *	selected from the list.
 * Arguments;
 *	binvecList - list of non-linear mapping vectors.
 *	minText, maxText - text fields for min and max values.
 *	scaleDisplay - display manager showing the range.
 *	scaleGraph - graph showing the current range.
 *	callData - unused
 *
 ****************************************************************************/

/* ARGSUSED */

static void 
selectNonLinearMappingCallback( Widget			binvecList,
				Widget			minText,
				Widget			maxText,
				AmMotifDisplayClass	*scaleDisplay,
				AmLineGraphClass		*scaleGraph,
				XtPointer		callData)
{
    selectNonLinearMapping( binvecList, minText, maxText, 
			    scaleDisplay, scaleGraph);
}

/***************************************************************************
 * 
 * Description:
 * 	user event handler to set the range values from the binvec editor 
 * Arguments:
 *	eventId - OK event.
 *	minText, maxText - TextFields containing min & max values.
 *	nonLinearTgl - linear/non-linear toggle, dictates whether values
 *	are taken from min and max texts, or non-linear binvecs.
 *	formDlog - range dialog.
 *	scaleGraph - displays the linear/non-linear range.
 *	minTag, maxTag - plot tags which the range refers to as a graphics
 *	attribute.
 *	binvecTag - binvec tag for the graphics attribue
 *	okArgs - slick OK arguments.
 *
 ***************************************************************************/

/* ARGSUSED */
static void 
setRangeDialogOkCallback( int			eventId,
			  rangeDialogDataPtr	rangeDialogData,
			  AmOkArgsPtr		okArgs)
{
    float	min, max, *binvec;

/* non-linear scaling, reset the bin vector, and filename or data depending
   on if the data is still around.  Get the min max values and set them too */ 

    minmaxValues( rangeDialogData->minText, rangeDialogData->maxText, 
		  &min, &max);
    AmSetTagValue( &min, AmFLOAT, rangeDialogData->minTag, okArgs);
    AmSetTagValue( &max, AmFLOAT, rangeDialogData->maxTag, okArgs);

/* if it's linear scaling, take the values from the text fields. Set the
   binvec to NULL to turn off non-linear scaling */
       
    if ( XmToggleButtonGetState( rangeDialogData->nonLinearTgl)) {
	rangeDialogData->scaleGraph->GetVaFields( AmNdata, &binvec, 
						  NULL);
	AmSetTagValue( &rangeDialogData->numColors, AmINT, 
		       rangeDialogData->lengthTag, okArgs);
	AmSetTagValue( binvec, AmFLOAT_PTR, rangeDialogData->binvecTag, okArgs);
    } else {
	AmSetTagValue( NULL, AmFLOAT_PTR, rangeDialogData->binvecTag, okArgs);
    }
    XtDestroyWidget( XtParent( rangeDialogData->formDlog));
    free( rangeDialogData);
}

/***************************************************************************
 * 
 * Description:
 * 	dialog for setting the range of data.  Since range can be applied
 *	to different graphics attributes, this function takes tags for
 *	the graphics attributes for which the range is to be applied.  The
 *	higher level functions pass the tags for which the range applies.
 * Arguments:
 *	plot - plot whose range is to be edited.
 *	minTag, maxTag - min and max tags.
 *	binvecTag - tag for non-linear range attribute.
 *	unitsTag - units for the range.
 *	lengthTag - length tag for the bin vector.
 *	showTag - unused.. this argument may be removed in the future.
 *	dialog - returned dialog (to be embedded in a slick subform)
 *	okBtn - returned ok button (to be embedded in a slick subform)
 *
 ****************************************************************************/

void 
AmCreateRangeDialog( Widget	parent,
		   AmPlotClass	*plot,
		   char		*minTag,
		   char		*maxTag,
		   char		*binvecTag,
		   char		*unitsTag,
		   char		*lengthTag,
		   char		*showTag,
		   Widget	*dialog,
		   Widget	*okBtn)
{
    AmActionRec           actionArea;
    Widget		formDlog, topForm, minLbl, minText, maxLbl, maxText, 
			nonLinearTgl, binvecLbl, binvecList, graphArea, 
			actionForm;
    Arg			args[ 20];
    char		**mapNameList, *mapFile, relPath[ PATHLEN];
    XmString		*motifStrings;
    int			numStrings, n, numcolors;
    float		min, max;
    unitType		units, xunits;
    float		*minmaxMin, *minmaxMax, *binvec;
    AmMotifDisplayClass	*scaleDisplay;
    AmLineGraphClass	*scaleGraph;
    AmTickClass		*hTicks, *vTicks;
    AmXDisplayClass	*disp;
    AmXWindowClass	*win;
    rangeDialogDataPtr	rangeDialogData;

/* also serves as failure flag */

    *dialog = NULL;

/* get the list of non-linear numeric mappings */

    sprintf( relPath, "%s/%s", AM_BINVEC_DIR, AM_MAP_LIST_FILE);
    mapFile = AmGetEnvFile( CONFIG_ENV, relPath);
    if ( mapFile == NULL) {
	AmPrintError( "cantOpenMapFile", relPath);
	return;
    }
    mapNameList = AmReadStrvec( mapFile); 
    if ( mapNameList == NULL) {
	AmPrintResourceError( "cantReadMapFile", mapFile);
	return;
    }

/* get the pertinent fields from the plot we want to set */

    plot->GetVaFields( minTag, &min,
		       maxTag, &max,
		       unitsTag, &units,
		       binvecTag, &binvec,
		       lengthTag, &numcolors,
		       NULL);
    if ( numcolors == 0) {
	AmPrintResourceError( "noColormapRange");
	return;
    }

/* create a form dialog to contain everything */

    n = AmMakeArgList( args,
		     XmNdeleteResponse, XmDESTROY,
		     XmNautoUnmanage, FALSE,
		     XmNdialogStyle, XmDIALOG_PRIMARY_APPLICATION_MODAL,
		     NULL);
    formDlog = XmCreateFormDialog( AmGetWidgetShell( parent), "setRange", 
				   args, n);
    *dialog = formDlog;

/* put the scale related widgets in a bar across the top */

    topForm = XtVaCreateManagedWidget( "rangeForm", xmFormWidgetClass, formDlog,
				       StickLeft, StickRight, StickTop, 
				       NULL);
    minLbl = XtVaCreateManagedWidget( "minLbl", xmLabelWidgetClass, topForm,
				      StickLeft, StickTop, StickBottom,
				      NULL);
    minText = XtVaCreateManagedWidget( "minText", xmTextFieldWidgetClass, 
				       topForm,
				       RightOf( minLbl), StickTop, StickBottom,
				       NULL);
    maxLbl = XtVaCreateManagedWidget( "maxLbl", xmLabelWidgetClass, topForm,
				      RightOf( minText), StickTop, StickBottom,
				      NULL);
    maxText = XtVaCreateManagedWidget( "maxText", xmTextFieldWidgetClass, 
				       topForm,
				       RightOf( maxLbl), StickTop, StickBottom,
				       NULL);
    nonLinearTgl = XtVaCreateManagedWidget( "nonLinearTgl", 
					    xmToggleButtonWidgetClass,
				            topForm,
				            StickRight, StickTop, 
					    StickBottom,
					    RightOf( maxText),
					    XmNset, ( binvec != NULL),
				            NULL);

/* NOTE: add a destroy callback to free these */

    minmaxMin = ( float *) malloc( 2*sizeof( float));
    minmaxMax = ( float *) malloc( 2*sizeof( float));
    minmaxMin[ 0] = -HUGE;
    minmaxMin[ 1] = max;
    minmaxMax[ 0] = min;
    minmaxMax[ 1] = HUGE;
    AmUnitVarEntry( minText, &min, AmFLOAT, units, minmaxMin);
    AmUnitVarEntry( maxText, &max, AmFLOAT, units, minmaxMax);
    XtAddCallback( formDlog, XmNdestroyCallback,
		   ( XtCallbackProc) AmFreeStructPtrCallback, minmaxMin);
    XtAddCallback( formDlog, XmNdestroyCallback,
		   ( XtCallbackProc) AmFreeStructPtrCallback, minmaxMax);

/* create a list of the available non-linear binning files on the right */

    binvecLbl = XtVaCreateManagedWidget( "binvecLbl", xmLabelWidgetClass,
					 formDlog,
					 StickRight, BelowOf( topForm),
					 NULL);
    motifStrings = AmMotifStringList( mapNameList, &numStrings);

/* there's a bug with scrolled list sensitivity that the resource has to
   be set on the list returned by the convienience function */

    n = AmMakeArgList( args,	
		     XmNitemCount, numStrings,
		     XmNitems, motifStrings,
		     BelowOf( binvecLbl), 
		     StickRight,
		     StickBottom,
		     XmNbottomOffset, 40,
		     XmNselectionPolicy, XmSINGLE_SELECT,
		     XmNvisibleItemCount, 12,
		     XmNscrollBarDisplayPolicy, XmSTATIC,
		     NULL);
    binvecList = XmCreateScrolledList( formDlog, "binvecList", args, n);
    XtVaSetValues( binvecList, 
		   XmNsensitive, ( binvec != NULL), 
		   NULL);
    XtManageChild( binvecList);
    AmFreeMotifStringList( motifStrings, numStrings);
    AmFreeStrVec( mapNameList);

    actionAreaDef[ 1].clientData = formDlog,
    actionForm = AmMakeActionArea( formDlog,
                     actionAreaDef, ELEMENTS( actionAreaDef),
                     &actionArea);
    XtVaSetValues( actionForm, StickLeft, StickRight, StickBottom, NULL);
    XtSetSensitive( actionArea.buttons[ 2], False);

    graphArea = XtVaCreateManagedWidget( "graphArea",
                    xmDrawingAreaWidgetClass, formDlog,
                    StickLeft, BelowOf( topForm),
                    LeftOf( binvecList), AboveOf( actionForm),
                    XmNwidth, 400,
                    NULL);

    *okBtn = actionArea.buttons[ 0];

/* create a graph to show the values for scaling */

    if ( binvec == NULL) {
	binvec = AmMakeRangeVector( min, max, numcolors - 1);
    }
    AmGetUnitCode( "colors", &xunits);

/* create a display, window, and display manager */

    disp = new AmXDisplayClass( graphArea);
    win = new AmXWindowClass( disp, graphArea);
    scaleDisplay = new AmMotifDisplayClass( disp, win);

/* create a graph and tickmarks to show the current range */

    scaleGraph = new AmLineGraphClass( win, "scaleGraph",
				     AmNleftMargin, 80,
				     AmNbottomMargin, 40,
				     AmNtopMargin, 20,
				     AmNrightMargin, 20,
				     AmNdata, binvec,
				     AmNrows, numcolors - 1,
				     AmNheightFn, AmFloatFn,
				     AmNheightUnits, units,
				     AmNheightType, AmFLOAT,
				     AmNsize, sizeof( float),
				     AmNyunits, units,
				     AmNxunits, xunits,
				     AmNxmin, 0.0,
				     AmNxmax, ( float) numcolors,
				     NULL);
    hTicks = new AmTickClass( win, scaleGraph, "scaleHticks",
			    AmNtickPosition, AmBOTTOM,
			    AmNtickStyle, AmBASE_TEN,
			    AmNtextVertical, FALSE,
			    AmNtitle, "colors",
			    AmNtitleOffset, 40,
			    AmNtitleVertical, FALSE,
			    AmNleftMargin, 80,
			    AmNrightMargin, 20,
			    AmNbottomMargin, 0,
			    AmNheight, 40,
			    AmNtitleOffset, 30,
			    AmNresizeFont, FALSE,
			    NULL);
    vTicks = new AmTickClass( win, scaleGraph, "scaleVticks",
			    AmNtickPosition, AmLEFT,
			    AmNtickStyle, AmBASE_TEN,
			    AmNtextVertical, FALSE,
			    AmNtitleOffset, 10,
			    AmNleftMargin, 0,
			    AmNwidth, 80,
			    AmNtopMargin, 20,
			    AmNbottomMargin, 40,
			    AmNtitleOffset, 30,
			    AmNresizeFont, FALSE,
			    NULL);
    scaleDisplay->AddObject( scaleGraph);
    scaleDisplay->AddObject( hTicks);
    scaleDisplay->AddObject( vTicks);
    scaleDisplay->Flush();

/* add the callbacks for the different widgets */

    XtAddCallback( nonLinearTgl, XmNvalueChangedCallback, 
		   ( XtCallbackProc) AmToggleSensitiveCallback, binvecList);
    AmAddVaCallback( binvecList, XmNsingleSelectionCallback, 
		   ( VarargsCallback) selectNonLinearMappingCallback,
		   minText, maxText, scaleDisplay, scaleGraph, END_LIST);
    AmAddVaCallback( nonLinearTgl, XmNvalueChangedCallback,
		   ( VarargsCallback) nonLinearTglCallback,
		   binvecList, minText, maxText, scaleDisplay, scaleGraph,
		   END_LIST);
    
/* add callbacks to the min and max text to reset the min and max, and
   redisplay the binvector */

    AmAddVaCallback( minText, XmNactivateCallback,
		   ( VarargsCallback) minmaxValueCallback,
		   minText, maxText, scaleDisplay, scaleGraph, 
		   minmaxMin, minmaxMax,
		   END_LIST);
    AmAddVaCallback( maxText, XmNactivateCallback,
		   ( VarargsCallback) minmaxValueCallback,
		   minText, maxText, scaleDisplay, scaleGraph, 
		   minmaxMin, minmaxMax,
		   END_LIST);
    AmAddVaCallback( minText, XmNlosingFocusCallback,
		   ( VarargsCallback) minmaxValueCallback,
		   minText, maxText, scaleDisplay, scaleGraph, 
		   minmaxMin, minmaxMax,
		   END_LIST);
    AmAddVaCallback( maxText, XmNlosingFocusCallback,
		   ( VarargsCallback) minmaxValueCallback,
		   minText, maxText, scaleDisplay, scaleGraph, 
		   minmaxMin, minmaxMax,
		   END_LIST);
    rangeDialogData = ( rangeDialogDataPtr) 
	malloc( sizeof( rangeDialogDataRec));
    if ( rangeDialogData == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return;
    }
    rangeDialogData->minText = minText;
    rangeDialogData->maxText = maxText;
    rangeDialogData->nonLinearTgl = nonLinearTgl;
    rangeDialogData->formDlog = formDlog;
    rangeDialogData->scaleGraph = scaleGraph;
    rangeDialogData->minTag = minTag;
    rangeDialogData->maxTag = maxTag;
    rangeDialogData->binvecTag = binvecTag;
    rangeDialogData->lengthTag = lengthTag;
    rangeDialogData->numColors = numcolors;
    AmAddCallbackWidget( AM_PKG_ID, AM_OK, *dialog,
			 ( callbackProc) setRangeDialogOkCallback,
			 rangeDialogData);
    XtManageChild( formDlog);
}

/***************************************************************************
 * 
 * Description:
 * 	used in a slick subform, this restores a range dialog callback
 *	when the dialog is ok'ed.  it is attached to the dialog's destroy 
 *	callback.
 * Arguments:
 *	dialog - slick subform.
 *	button - button which launched the range editor dialog
 *	plot - plot to edit the range of.
 *	args - vector of arguments to pass.
 *	callData - unused
 *
 **************************************************************************/

static AnvilBoolean
setRangeDialogCallback( Widget,		/* dialog */
			Widget          	button,
			AmPlotClass       	*plot,
			char		*args,
			XtPointer)
{
    AmAddVaCallback( button, XmNactivateCallback,
                   ( VarargsCallback) AmRangeDialogCallback, 
		   plot, args, END_LIST);
    return( TRUE);
}

/***************************************************************************
 * 
 * Description:
 * 	callback for embedding in a slick subform.  It unmanages the dialog
 * 	on the OK, and sets it up to be remanaged again when the button is hit.
 * Arguments:
 *	w - widget to launch the dialog from.
 *	plot - plot to edit the range of.
 *	args - vector of arguments to construct the range dialog with:
 *	minTag, maxTag, binvecTag, unitsTag, lengthTag, showTag.
 *	callData - unused.
 *
 **************************************************************************/
 
/* ARGSUSED */ 
AnvilBoolean
AmRangeDialogCallback( Widget    w,
		     AmPlotClass *plot,
		     char      **args,
		     XtPointer xpCallData)
{
    Widget      	dialog, okBtn, parent = AmGetWidgetShell( w);
 
    AmCreateRangeDialog( parent, plot, args[ 0], args[ 1], args[ 2],
		       args[ 3], args[ 4], args[ 5], 
		       &dialog, &okBtn);

    if ( dialog != NULL) {
	AmAddVaCallback( okBtn, XmNactivateCallback,
		       ( VarargsCallback) AmUnmanageSubDialogCallback,
		       w, dialog, END_LIST);

/* we remove and replace the callback to launch the dialog, because we 
   keep the dialog around on OK for our parent slick */

	AmRemoveVaCallback( w, XmNactivateCallback,
			  ( VarargsCallback) AmRangeDialogCallback, 
			  plot, args, END_LIST);

	AmAddVaCallback( dialog, XmNdestroyCallback,
		       ( VarargsCallback) setRangeDialogCallback,
		       w, plot, args, END_LIST);

        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}
 
/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the colors used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetRangeDialogCallback( Widget            w,
			AmMotifDisplayClass *motifDisplay,
			AmPlotClass         *plot,
			XtPointer)
 
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNmin, AmNmax, AmNcolorsBinvec, 
		       AmNzunits, AmNnumColors, AmNshowColors,
		       &dialog, &okBtn);

    if ( dialog != NULL) {
        AmSetOkButton( okBtn, plot);
        XtAddCallback( okBtn, XmNactivateCallback,
		       ( XtCallbackProc) AmUpdateMotifDisplayCallback,
		       motifDisplay);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the hatch angles used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetHatchAngleRangeDialogCallback( Widget                  w,
				  AmMotifDisplayClass       *,
				  AmPlotClass               *plot,
				  XtPointer)
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNhatchAngleMin, AmNhatchAngleMax, AmNhatchAngleBinvec, 
		       AmNhatchAngleUnits, AmNhatchAngleLength, 
		       AmNshowHatchAngles, &dialog, &okBtn);

    if ( dialog != NULL) {
	AmSetOkButton( okBtn, plot);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the hatch width used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetHatchWidthRangeDialogCallback( Widget                  w,
				  AmMotifDisplayClass       *,
				  AmPlotClass               *plot,
				  XtPointer)
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNhatchWidthMin, AmNhatchWidthMax, AmNhatchWidthBinvec, 
		       AmNhatchWidthUnits, AmNhatchWidthLength, 
		       AmNshowHatchWidths,
		       &dialog, &okBtn);
    if ( dialog != NULL) {
        AmSetOkButton( okBtn, plot);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the line width used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetLineWidthRangeDialogCallback( Widget                  w,
				 AmMotifDisplayClass       *,
				 AmPlotClass               *plot,
				 XtPointer)
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNlineWidthMin, AmNlineWidthMax, AmNlineWidthBinvec, 
		       AmNlineWidthUnits, AmNlineWidthLength, AmNshowLineWidths,
		       &dialog, &okBtn);

    if ( dialog != NULL) {
        AmSetOkButton( okBtn, plot);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the symbols used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetSymbolRangeDialogCallback( Widget                  w,
			      AmMotifDisplayClass       *,
			      AmPlotClass               *plot,
		              XtPointer)
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNsymbolMin, AmNsymbolMax, AmNsymbolBinvec, 
		       AmNsymbolUnits, AmNsymbolLength, AmNshowSymbols,
		       &dialog, &okBtn);

    if ( dialog != NULL) {
        AmSetOkButton( okBtn, plot);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}

/***************************************************************************
 * 
 * Description:
 *	Top level callback to allow the user to set the range for 
 *	the symbol size used in a plot.
 * Arguments:
 *	w - widget to attach the dialog to.
 *	motifDisplay - display manager the plot is displayed on 
 *	plot - object to edit the range for
 *	callData - unused.
 *
 **************************************************************************/

/* ARGSUSED */
AnvilBoolean
AmSetSymbolSizeRangeDialogCallback( Widget                  w,
				  AmMotifDisplayClass       *,
				  AmPlotClass               *plot,
				  XtPointer)
{
    Widget      dialog, okBtn;
 
    AmCreateRangeDialog( AmGetWidgetShell( w), plot, 
		       AmNsymbolSizeMin, AmNsymbolSizeMax, AmNsymbolSizeBinvec, 
		       AmNsymbolSizeUnits, AmNsymbolSizeLength, 
		       AmNshowSymbolSizes,
		       &dialog, &okBtn);

    if ( dialog != NULL) {
        AmSetOkButton( okBtn, plot);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeDialog");
        return( FALSE);
    }
}
