#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/ToggleB.h>
#include <Xm/List.h>
#include <Xm/TextF.h>
#include <Xm/Scale.h>
#include <Xm/DrawingA.h>
#include <anvil/files.h>
#include <anvil/strlib.h>
#include <anvil/vector.h>
#include <anvil/errlib.h>
#include <anvil/motif.h>
#include <anvil/motifdisplay.h>
#include <anvil/va_cback.h>
#include <anvil/envdefs.h>
#include <anvil/plot_cback.h>
#include <anvil/linegraph.h>
#include <anvil/tick.h>
#include <anvil/textcheck.h>
#include <anvil/interp.h>
#include <anvil/userevents.h>
#include <anvil/slick.h>
#include <anvil/slickP.h>
#include <anvil/utility.h>
#include <anvil/colorlegendP.h>

static AmActionDefRec actionAreaDef[] = {
    { "okBtn", NULL, NULL },
    { "closeBtn", ( XtCallbackProc) AmPopdownCback, NULL },
    { "helpBtn", NULL, NULL }
};

typedef struct rangeColorDialogData_ {
    Widget		minText;
    Widget		maxText;
    Widget		nonLinearTgl;
    Widget		formDlog;
    Widget		numColorsScl;
    Widget		binvecList;
    AmLineGraphClass	*scaleGraph;
    AmMotifDisplayClass	*scaleDisplay;
    AmLegendClass	*legend;
    AmMotifDisplayClass	*legendDisplay;
    char		*colormapFile;
    AmPlotClass		*plot;
} rangeColorDialogDataRec, *rangeColorDialogDataPtr;

/* get the min and max values from the text widgets */

static void 
minmaxValues( Widget	minText,
	      Widget	maxText,
	      float	*min,
	      float	*max)
{
    char	*minStr, *maxStr;

    minStr = XmTextFieldGetString( minText);
    maxStr = XmTextFieldGetString( maxText);
    *min = atof( minStr);
    *max = atof( maxStr);
    XtFree( minStr);
    XtFree( maxStr);
}

/***************************************************************************
 * 
 * Description:
 * 	update the scale graph whenever the min or max text values change 
 * Arguments:
 *	text - one of the text widgets 
 *	minText, maxText - min and max textfields
 *	scaleDisplay - display manager to containing range graph.
 *	scaleGraph - range graph.
 *	minmaxMin, minmaxMax - pointer to current min and max values.
 *	callData - unused 
 *
 **************************************************************************/

/* ARGSUSED */

static void 
minmaxValueCallback( Widget			text,
		     rangeColorDialogDataPtr	rangeDialogData,
		     float			*minmaxMin,
		     float			*minmaxMax,
		     XtPointer			callData)
{
    float	min, max, *binvec;
    int		numbins;

    rangeDialogData->scaleGraph->GetVaFields( AmNdata, &binvec, 
					      AmNrows, &numbins,
					      NULL);
    minmaxValues( rangeDialogData->minText, rangeDialogData->maxText, &min, &max);
    if ( min < max) {
	AmSetErrorColor( rangeDialogData->minText, FALSE);
	AmSetErrorColor( rangeDialogData->maxText, FALSE);

/* reset the range check values for the min and max value, since they update
   as the min and max are changed. */

	minmaxMin[ 1] = max;
	minmaxMax[ 0] = min;
	AmStretchVec( min, max, binvec, numbins);
	rangeDialogData->scaleGraph->SetVaFields( AmNdata, binvec, 
						  AmNmin, min,
						  AmNmax, max,
						  AmNcurrYmin, min,
						  AmNcurrYmax, max,
						  NULL);
	rangeDialogData->legend->SetVaFields( AmNcolorsMin, min,
					      AmNcolorsMax, max,
					      AmNcolorsBinvec, binvec,
					      NULL);
	rangeDialogData->scaleDisplay->Update();
	rangeDialogData->legendDisplay->Update();
    } else {
	AmSetErrorColor( rangeDialogData->minText, TRUE);
	AmSetErrorColor( rangeDialogData->maxText, TRUE);
    }
}

/***************************************************************************
 * 
 * Description:
 * 	select a non-linear mapping and update the display appropriately 
 * Arguments:
 *	rangeDialogData - contains widgets and screen objects associated
 *	with the range/color dialog.
 *	binvecList - XmList containing non-linear mappings.
 *	minText, maxText - min and max textfields
 *	scaleDisplay - display manager to containing range graph.
 *
 ***************************************************************************/

static void 
selectNonLinearMapping( rangeColorDialogDataPtr	rangeDialogData)
{
    XmString		*items;
    int			count, numbins, numlines, oldNumColors;
    char		*filename, *path, relPath[ PATHLEN];
    float		*binvec, *binvec2, *oldBinvec, min, max, *binvecPtr;
    AmColorCellPtr	oldColors, newColors;

/* get the min and max values from the text fields */

    minmaxValues( rangeDialogData->minText, rangeDialogData->maxText, 
		  &min, &max);

/* if he selected a map file, then load it, 
   get the old bin vector, because we free it once its replaced, or we 
   stretch it to the new min max values */

    XtVaGetValues( rangeDialogData->binvecList,
		   XmNselectedItems, &items,
		   XmNselectedItemCount, &count,
		   NULL);
    rangeDialogData->scaleGraph->GetVaFields( AmNdata, &oldBinvec, 
					      AmNrows, &numbins,
					      NULL);
    if ( count != 0) {
	XmStringGetLtoR( items[ 0], XmFONTLIST_DEFAULT_TAG, &filename);
	sprintf( relPath, "%s/%s", AM_BINVEC_DIR, filename);
	path = AmGetEnvFile( CONFIG_ENV, relPath);
	binvec = AmReadFloatVec( path);
	if ( binvec == NULL) {
	    AmPrintError( "failed to open map file %s", filename);
	    return;
	}

/* stretch it out to the number of displayed values */

	numlines = AmNumLines( path);
	binvec2 = ( float *) malloc( sizeof( float)*numbins);
	if ( binvec == NULL) {
	    AmPrintError( "mallocFailed");
	    return;
	}

/* interpolate to the displayed min and max, then set it as the data 
   in the graph, and free the old vector, and the pre-interpolated vector */

	AmInterpolateVec( binvec, numlines, binvec2, numbins);
	AmStretchVec( min, max, binvec2, numbins);
	rangeDialogData->scaleGraph->SetVaFields( AmNdata, binvec2, 
						  AmNrows, numbins,
						  NULL);
	binvecPtr = binvec2;
	XtFree( filename);
	free( oldBinvec);
	free( binvec);
    } else {

/* if there's no selection, just stretch the current vector in place */

	AmStretchVec( min, max, oldBinvec, numbins);
	rangeDialogData->scaleGraph->SetVaFields( AmNdata, oldBinvec, NULL);
	binvecPtr = oldBinvec;
    }

/* re-interpolate the colors in the colormap */

    rangeDialogData->legend->GetVaFields( AmNnumColors, &oldNumColors,
					  AmNcolors, &oldColors,
					  NULL);
    AmInterpolateColorCells( oldColors, oldNumColors, &newColors, numbins + 1);
    rangeDialogData->legend->SetVaFields( AmNnumColors, numbins + 1,
					  AmNcolors, newColors,
					  AmNcolorsBinvec, binvecPtr,
					  NULL);
    free( oldColors);
    free( newColors);
    rangeDialogData->scaleDisplay->Update();
    rangeDialogData->legendDisplay->Update();
}

/***************************************************************************
 * 
 * Description:
 *	toggle between linear and non-linear mappings.  Controls 
 *	sensitivity of the text and list widgets.
 * Arguments:
 *	binvecList - list of non-linear mappings.
 *	minText, maxText - text fields for min and max values.
 *	scaleDisplay - display manager to containing range graph.
 *	scaleGraph - range graph.
 *	callData - unused.
 *
 ***************************************************************************/

static void 
nonLinearTglCallback( Widget			tgl,
		      rangeColorDialogDataPtr	rangeDialogData,
		      XtPointer			callData)
{
    float	min, max, *binvec, *oldBinvec;
    int		numbins;

/* depending on whether we're in linear or non-linear mode */

    if ( XmToggleButtonGetState( tgl)) {
	selectNonLinearMapping( rangeDialogData);
    } else {

/* get the min and max values from the text fields, and create a linear 
   vector */

	rangeDialogData->scaleGraph->GetVaFields( AmNrows, &numbins,
						  AmNdata, &oldBinvec,
						  NULL);
	minmaxValues( rangeDialogData->minText, rangeDialogData->maxText, 
		      &min, &max);
	binvec = AmMakeRangeVector( min, max, numbins);
	rangeDialogData->scaleGraph->SetVaFields( AmNdata, binvec,
						  NULL);

/* free the old bin vector */

	if ( oldBinvec != NULL) {
	    free( oldBinvec);
	}
	rangeDialogData->scaleDisplay->Update();

/* set the min and max for the legend */

	rangeDialogData->legend->SetVaFields( AmNcolorsMin, min,
					      AmNcolorsMax, max,
					      NULL);
	rangeDialogData->legendDisplay->Update();
    }
}

/***************************************************************************
 * 
 * Description:
 *	callback to update the display when a non-linear mapping is
 *	selected from the list.
 * Arguments;
 *	binvecList - list of non-linear mapping vectors.
 *	minText, maxText - text fields for min and max values.
 *	scaleDisplay - display manager showing the range.
 *	scaleGraph - graph showing the current range.
 *	callData - unused
 *
 ****************************************************************************/

/* ARGSUSED */

static void 
selectNonLinearMappingCallback( Widget			binvecList,
				rangeColorDialogDataPtr	rangeDialogData,
				XtPointer		callData)
{
    selectNonLinearMapping( rangeDialogData);
}

/***************************************************************************
 * 
 * Description:
 * 	user event handler to set the range values from the binvec editor 
 * Arguments:
 *	eventId - OK event.
 *	rangeColorDialogData - structure contains dialog information
 *	about colors and range. 
 *	okArgs - slick OK arguments.
 *
 ***************************************************************************/

/* ARGSUSED */
static void 
setRangeDialogOkCallback( int				eventId,
			  rangeColorDialogDataPtr	rangeDialogData,
			  AmOkArgsPtr			okArgs)
{
    float		min, max, *binvec;
    int			numBinvec, numColors;
    AmColorCellPtr	colors;

/* non-linear scaling, reset the bin vector, and filename or data depending
   on if the data is still around.  Get the min max values and set them too */ 

    minmaxValues( rangeDialogData->minText, rangeDialogData->maxText, 
		  &min, &max);
    AmSetTagValue( &min, AmFLOAT, AmNcolorsMin, okArgs);
    AmSetTagValue( &max, AmFLOAT, AmNcolorsMax, okArgs);

/* if it's linear scaling, take the values from the text fields. Set the
   binvec to NULL to turn off non-linear scaling */
       
    if ( XmToggleButtonGetState( rangeDialogData->nonLinearTgl)) {
	rangeDialogData->scaleGraph->GetVaFields( AmNdata, &binvec, 
						  AmNrows, &numBinvec,
						  NULL);
	AmSetTagValue( &numBinvec, AmINT, AmNnumColors, okArgs);
	AmSetTagValue( binvec, AmFLOAT_PTR, AmNcolorsBinvec, okArgs);
    } else {
	AmSetTagValue( NULL, AmFLOAT_PTR, AmNcolorsBinvec, okArgs);
    }
    rangeDialogData->legend->GetVaFields( AmNnumColors, &numColors,
					  NULL);

/* set the colormap file if it exists, otherwise, set the colors directly. */

    if ( rangeDialogData->colormapFile != NULL) {
	AmSetTagValue( rangeDialogData->colormapFile, AmSTRING, 
		       AmNcolormap, okArgs);
    } else {
	rangeDialogData->legend->GetVaFields( AmNcolors, &colors,
					      NULL);
	AmSetTagValue( colors, AmCOLOR_CELL_PTR, AmNcolors, okArgs);
    }
    AmSetTagValue( &numColors, AmINT, AmNnumColors, okArgs);
    XtDestroyWidget( XtParent( rangeDialogData->formDlog));
    free( rangeDialogData);
}

/*************************************************************************
 *
 * Description:
 *      change the colormap in the legend when the user picks a new colormap.
 * Arguments:
 *	w - XmList widget
 *	rangeDialogData - contains widget and screen object information
 *	from the range dialog.
 *      cb - XmList callback struct
 *
 **************************************************************************/

/* ARGSUSED */

static void
changeColormapCallback( Widget                  w,
			rangeColorDialogDataPtr	rangeDialogData,
                        XmListCallbackStruct    *cb)
{
    char                *filename, *filepath, relPath[ PATHLEN];
    AmColorCellPtr      fileColors, newColors;
    int                 numFileColors, numColors;

    XmStringGetLtoR( cb->item, XmFONTLIST_DEFAULT_TAG, &filename);
    sprintf( relPath, "%s/%s", AM_COLORMAP_DIR, filename);
    filepath = AmGetEnvFile( CONFIG_ENV, relPath);
    if ( filepath == NULL) {
        AmPrintResourceError( "cantFindColormapFile", relPath);
        return;
    }

/* save the colormap file name */

    if ( rangeDialogData->colormapFile != NULL) {
        free( rangeDialogData->colormapFile);
    }
    rangeDialogData->colormapFile = strdup( filename);
    XtFree( filename);
    rangeDialogData->legend->GetVaFields( AmNnumColors, &numColors, NULL);

/* read the colors from the file and interpolate them to the number of
   colors currently in the legend */

    if ( !AmReadColors( filepath, &numFileColors, &fileColors)) {
        AmPrintResourceError( "cantReadColormapFile", filepath);
        return;
    }
    AmInterpolateColorCells( fileColors, numFileColors, &newColors, numColors);
    rangeDialogData->legend->SetVaFields( AmNcolors, newColors, NULL);
    rangeDialogData->legendDisplay->Flush();
    free( newColors);
    free( fileColors);
}

/*************************************************************************
 *
 * Description:
 *      change the colormap in the legend when the user picks a new colormap.
 * Arguments:
 *	w - XmList widget
 *	rangeDialogData - contains widget and screen object information
 *	from the range dialog.
 *      cb - XmList callback struct
 *
 **************************************************************************/

/* ARGSUSED */

static void
setNumColorsCallback( Widget                  	scale,
		      rangeColorDialogDataPtr	rangeDialogData,
		      XtPointer)
{
    int			newNumColors, oldNumBinvec, numColors;
    AmColorCellPtr	currColors, newColors;
    float		*oldBinvec, *newBinvec;
    char		relPath[ PATHLEN], *cmapPath;

    XmScaleGetValue( scale, &newNumColors);
    rangeDialogData->scaleGraph->GetVaFields( AmNdata, &oldBinvec, 
					      AmNrows, &oldNumBinvec,
					      NULL);

/* read the colors from the color file, then interpolate to them, if there's
   a color file.  If not, get the colors from the plot (because we haven't
   selected a color file), and interpolate them */

    if ( rangeDialogData->colormapFile != NULL) {
	sprintf( relPath, "%s/%s", AM_COLORMAP_DIR, 
		 rangeDialogData->colormapFile);
printf( "attempting to pull colormap file %s\n", relPath);
	cmapPath = AmGetEnvFile( CONFIG_ENV, relPath);
	if ( cmapPath == NULL) {
            AmPrintResourceError( "cantFindColormapFile", 
				  rangeDialogData->colormapFile);
            return;
        }
        if ( !AmReadColors( cmapPath, &numColors, &currColors)) {
            AmPrintResourceError( "cantReadColormapFile", cmapPath);
            return;
        }
    } else {
printf( "pulling colormap file from plot\n");
        rangeDialogData->plot->GetVaFields( AmNnumColors, &numColors,
					    AmNcolors, &currColors,
					    NULL);
    }

/* interpolate the new colors and bin vector */

printf( "interpolating from %d to %d\n", numColors, newNumColors);
    AmInterpolateColorCells( currColors, numColors, &newColors, newNumColors);
    free( currColors);
    newBinvec = ( float *) malloc((newNumColors - 1)*sizeof( float));
    if ( newBinvec == NULL) {
	AmPrintResourceError( AmGetWidgetShell( scale), "mallocFailed");
	return;
    }
    AmInterpolateVec( oldBinvec, oldNumBinvec, newBinvec, newNumColors - 1);

/* set the fields for the range graph and legend */

    rangeDialogData->scaleGraph->SetVaFields( AmNdata, newBinvec,
					      AmNrows, newNumColors - 1,
					      NULL);
    rangeDialogData->legend->SetVaFields( AmNnumColors, newNumColors,
					  AmNcolors, newColors,
					  AmNcolorsBinvec, newBinvec,
					  NULL);
    rangeDialogData->legendDisplay->Flush();
    rangeDialogData->scaleDisplay->Flush();
    free( newColors);
    free( oldBinvec);
}

/***************************************************************************
 * 
 * Description:
 * 	dialog for setting the range of data.  Since range can be applied
 *	to different graphics attributes, this function takes tags for
 *	the graphics attributes for which the range is to be applied.  The
 *	higher level functions pass the tags for which the range applies.
 * Arguments:
 *	parent - parent of the dialog.
 *	plot - plot whose range is to be edited.
 *	dialog - returned dialog (to be embedded in a slick subform)
 *	okBtn - returned ok button (to be embedded in a slick subform)
 *
 ****************************************************************************/

void 
AmCreateRangeColorDialog( Widget	parent,
			  AmPlotClass	*plot,
			  Widget	*dialog,
			  Widget	*okBtn)
{
    AmActionRec         	actionArea;
    Widget			formDlog, topForm, minLbl, minText, maxLbl, 
				maxText, nonLinearTgl, binvecLbl, binvecList, 
				graphArea, actionForm, legendArea, numColorsScl,
				cmapList;
    Arg				args[ 20];
    char			**mapNameList, *mapFile, relPath[ PATHLEN],
				**cmapNameList, *cmapListFile, *colormapFile;
    XmString			*motifStrings;
    int				numStrings, n, numcolors;
    float			min, max;
    unitType			units, xunits;
    float			*minmaxMin, *minmaxMax, *binvec;
    AmMotifDisplayClass		*scaleDisplay, *legendDisplay;
    AmLineGraphClass		*scaleGraph;
    AmTickClass			*hTicks, *vTicks;
    AmXDisplayClass		*scaleDisp, *legendDisp;
    AmXWindowClass		*scaleWin, *legendWin;
    rangeColorDialogDataPtr	rangeDialogData;
    AmLegendClass		*legend;

/* also serves as failure flag */

    *dialog = NULL;

/* get the list of non-linear numeric mappings */

    sprintf( relPath, "%s/%s", AM_BINVEC_DIR, AM_MAP_LIST_FILE);
    mapFile = AmGetEnvFile( CONFIG_ENV, relPath);
    if ( mapFile == NULL) {
	AmPrintError( "cantOpenMapFile", relPath);
	return;
    }
    mapNameList = AmReadStrvec( mapFile); 
    if ( mapNameList == NULL) {
	AmPrintResourceError( "cantReadMapFile", mapFile);
	return;
    }

/* read the list of colormaps */

    sprintf( relPath, "%s/%s", AM_COLORMAP_DIR, AM_CMAP_LIST_FILE);
    cmapListFile = AmGetEnvFile( CONFIG_ENV, relPath);
    if ( cmapListFile == NULL) {
        AmPrintResourceError( "colormapListOpenFailed", relPath);
        return;
    }
    cmapNameList = AmReadLines( cmapListFile);
    if ( cmapNameList == NULL) {
        AmPrintResourceError( "colormapListReadFailed", cmapListFile);
        return;
    }

/* get the pertinent fields from the plot we want to set */

    plot->GetVaFields( AmNcolorsMin, &min,
		       AmNcolorsMax, &max,
		       AmNcolorsUnits, &units,
		       AmNcolorsBinvec, &binvec,
		       AmNnumColors, &numcolors,
		       AmNcolormap, &colormapFile,
		       NULL);
    if ( numcolors == 0) {
	AmPrintResourceError( "noColormapRange");
	return;
    }

/* create a form dialog to contain everything */

    n = AmMakeArgList( args,
		       XmNdeleteResponse, XmDESTROY,
		       XmNautoUnmanage, FALSE,
		       XmNdialogStyle, XmDIALOG_PRIMARY_APPLICATION_MODAL,
		       NULL);
    formDlog = XmCreateFormDialog( AmGetWidgetShell( parent), "setRange", 
				   args, n);
    *dialog = formDlog;

/* put the scale related widgets in a bar across the top */

    topForm = XtVaCreateManagedWidget( "rangeForm", xmFormWidgetClass, formDlog,
				       StickRight, StickTop, 
				       NULL);
    minLbl = XtVaCreateManagedWidget( "minLbl", xmLabelWidgetClass, topForm,
				      StickLeft, StickTop, StickBottom,
				      NULL);
    minText = XtVaCreateManagedWidget( "minText", xmTextFieldWidgetClass, 
				       topForm,
				       RightOf( minLbl), StickTop, StickBottom,
				       NULL);
    maxLbl = XtVaCreateManagedWidget( "maxLbl", xmLabelWidgetClass, topForm,
				      RightOf( minText), StickTop, StickBottom,
				      NULL);
    maxText = XtVaCreateManagedWidget( "maxText", xmTextFieldWidgetClass, 
				       topForm,
				       RightOf( maxLbl), StickTop, StickBottom,
				       NULL);
    nonLinearTgl = XtVaCreateManagedWidget( "nonLinearTgl", 
					    xmToggleButtonWidgetClass,
				            topForm,
				            StickTop, 
					    StickBottom,
					    RightOf( maxText),
					    XmNset, ( binvec != NULL),
				            NULL);

/* create the ranges for entering the min and max values */

    minmaxMin = ( float *) malloc( 2*sizeof( float));
    minmaxMax = ( float *) malloc( 2*sizeof( float));
    minmaxMin[ 0] = -HUGE;
    minmaxMin[ 1] = max;
    minmaxMax[ 0] = min;
    minmaxMax[ 1] = HUGE;
    AmUnitVarEntry( minText, &min, AmFLOAT, units, minmaxMin);
    AmUnitVarEntry( maxText, &max, AmFLOAT, units, minmaxMax);
    XtAddCallback( formDlog, XmNdestroyCallback,
		   ( XtCallbackProc) AmFreeStructPtrCallback, minmaxMin);
    XtAddCallback( formDlog, XmNdestroyCallback,
		   ( XtCallbackProc) AmFreeStructPtrCallback, minmaxMax);

/* create the action area at the bottom of the dialog */

    actionAreaDef[ 1].clientData = formDlog;
    actionForm = AmMakeActionArea( formDlog,
				   actionAreaDef, ELEMENTS( actionAreaDef),
				   &actionArea);
    XtVaSetValues( actionForm, StickLeft, StickRight, StickBottom, NULL);
    XtSetSensitive( actionArea.buttons[ 2], False);
    *okBtn = actionArea.buttons[ 0];

/* create a list of the available non-linear binning files on the right */

    binvecLbl = XtVaCreateManagedWidget( "binvecLbl", xmLabelWidgetClass,
					 formDlog,
					 StickRight, BelowOf( topForm),
					 NULL);
    motifStrings = AmMotifStringList( mapNameList, &numStrings);

/* there's a bug with scrolled list sensitivity that the resource has to
   be set on the list returned by the convienience function */

    n = AmMakeArgList( args,	
		       XmNitemCount, numStrings,
		       XmNitems, motifStrings,
		       BelowOf( binvecLbl), 
		       StickRight,
		       AboveOf( actionForm),
		       XmNselectionPolicy, XmSINGLE_SELECT,
		       XmNvisibleItemCount, 12,
		       XmNscrollBarDisplayPolicy, XmSTATIC,
		       NULL);
    binvecList = XmCreateScrolledList( formDlog, "binvecList", args, n);
    XtVaSetValues( binvecList, 
		   XmNsensitive, ( binvec != NULL), 
		   NULL);
    XtManageChild( binvecList);
    AmFreeMotifStringList( motifStrings, numStrings);
    AmFreeStrVec( mapNameList);

/* create a scale to control the number of colors in the legend */

    numColorsScl = XtVaCreateManagedWidget( "numColorsScl",
					    xmScaleWidgetClass, formDlog,
					    BelowOf( topForm), StickLeft,
					    AboveOf( actionForm),
					    XmNshowValue, TRUE,
					    XmNminimum, 3,
					    XmNmaximum, 128,
					    XmNvalue, numcolors,
					    NULL);

/* create a list of the available colormaps */

    motifStrings = AmMotifStringList( cmapNameList, &numStrings);
    n = AmMakeArgList( args,
		       XmNitemCount, numStrings,
		       XmNitems, motifStrings,
		       BelowOf( topForm), 
		       RightOf( numColorsScl),
		       AboveOf( actionForm),
		       XmNselectionPolicy, XmSINGLE_SELECT,
		       XmNvisibleItemCount, 12,
		       XmNscrollBarDisplayPolicy, XmSTATIC,
		       NULL);
    cmapList = XmCreateScrolledList( formDlog, "cmapList", args, n);
    XtManageChild( cmapList);
    AmFreeMotifStringList( motifStrings, numStrings);
    AmFreeStrVec( cmapNameList);

/* create the area for drawing the legend */
    
    legendArea = XtVaCreateManagedWidget( "legendArea",
					   xmDrawingAreaWidgetClass, 
					   formDlog,
					   RightOf( cmapList), 
					   BelowOf( topForm),
					   AboveOf( actionForm),
					   XmNwidth, 160,
					   XmNheight, 240,
					   NULL);

/* create a display, window, and display manager */

    legendDisp = new AmXDisplayClass( legendArea);
    legendWin = new AmXWindowClass( legendDisp, legendArea);
    legendDisplay = new AmMotifDisplayClass( legendDisp, legendWin);

/* create the legend for the plot */

    legend = AmMakeStandaloneLegend( legendDisplay, plot);
    legendDisplay->AddObject( legend);
    legendDisplay->Flush();
			
/* create the area for drawing the binvec */

    graphArea = XtVaCreateManagedWidget( "graphArea",
					 xmDrawingAreaWidgetClass, 
					 formDlog,
					 RightOf( legendArea), 
					 BelowOf( topForm),
					 LeftOf( binvecList), 
					 AboveOf( actionForm),
					 XmNwidth, 400,
					 NULL);


/* create a graph to show the values for scaling */

    if ( binvec == NULL) {
	binvec = AmMakeRangeVector( min, max, numcolors - 1);
    }
    AmGetUnitCode( "colors", &xunits);

/* create a display, window, and display manager */

    scaleDisp = new AmXDisplayClass( graphArea);
    scaleWin = new AmXWindowClass( scaleDisp, graphArea);
    scaleDisplay = new AmMotifDisplayClass( scaleDisp, scaleWin);

/* create a graph and tickmarks to show the current range */

    scaleGraph = new AmLineGraphClass( scaleWin, "scaleGraph",
				       AmNleftMargin, 80,
				       AmNbottomMargin, 40,
				       AmNtopMargin, 20,
				       AmNrightMargin, 20,
				       AmNdata, binvec,
				       AmNrows, numcolors - 1,
				       AmNheightFn, AmFloatFn,
				       AmNheightUnits, units,
				       AmNheightType, AmFLOAT,
				       AmNsize, sizeof( float),
				       AmNyunits, units,
				       AmNxunits, xunits,
				       AmNxmin, 0.0,
				       AmNxmax, ( float) numcolors,
				       NULL);
    hTicks = new AmTickClass( scaleWin, scaleGraph, "scaleHticks",
			      AmNtickPosition, AmBOTTOM,
			      AmNtickStyle, AmBASE_TEN,
			      AmNtextVertical, FALSE,
			      AmNtitle, "colors",
			      AmNtitleOffset, 40,
			      AmNtitleVertical, FALSE,
			      AmNleftMargin, 80,
			      AmNrightMargin, 20,
			      AmNbottomMargin, 0,
			      AmNheight, 40,
			      AmNtitleOffset, 30,
			      AmNresizeFont, FALSE,
			      NULL);
    vTicks = new AmTickClass( scaleWin, scaleGraph, "scaleVticks",
			      AmNtickPosition, AmLEFT,
			      AmNtickStyle, AmBASE_TEN,
			      AmNtextVertical, FALSE,
			      AmNtitleOffset, 10,
			      AmNleftMargin, 0,
			      AmNwidth, 80,
			      AmNtopMargin, 20,
			      AmNbottomMargin, 40,
			      AmNtitleOffset, 30,
			      AmNresizeFont, FALSE,
			      NULL);
    scaleDisplay->AddObject( scaleGraph);
    scaleDisplay->AddObject( hTicks);
    scaleDisplay->AddObject( vTicks);
    scaleDisplay->Flush();

/* create the structure to pass arguments to the different callbacks */

    rangeDialogData = ( rangeColorDialogDataPtr) 
			    malloc( sizeof( rangeColorDialogDataRec));
    if ( rangeDialogData == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return;
    }
    rangeDialogData->minText = minText;
    rangeDialogData->maxText = maxText;
    rangeDialogData->nonLinearTgl = nonLinearTgl;
    rangeDialogData->formDlog = formDlog;
    rangeDialogData->scaleGraph = scaleGraph;
    rangeDialogData->scaleDisplay = scaleDisplay;
    rangeDialogData->legend = legend;
    rangeDialogData->legendDisplay = legendDisplay;
    rangeDialogData->colormapFile = colormapFile;
    rangeDialogData->binvecList = binvecList;
    rangeDialogData->plot = plot;

/* add the callbacks for the different widgets */

    XtAddCallback( numColorsScl, XmNvalueChangedCallback,
		   ( XtCallbackProc) setNumColorsCallback, rangeDialogData);
    XtAddCallback( nonLinearTgl, XmNvalueChangedCallback, 
		   ( XtCallbackProc) AmToggleSensitiveCallback, binvecList);
    XtAddCallback( binvecList, XmNsingleSelectionCallback, 
		   ( XtCallbackProc) selectNonLinearMappingCallback,
		   rangeDialogData); 
    XtAddCallback( nonLinearTgl, XmNvalueChangedCallback,
		   ( XtCallbackProc) nonLinearTglCallback, rangeDialogData);
    XtAddCallback( cmapList, XmNsingleSelectionCallback,
		   ( XtCallbackProc) changeColormapCallback, rangeDialogData);
    
/* add callbacks to the min and max text to reset the min and max, and
   redisplay the binvector */

    AmAddVaCallback( minText, XmNactivateCallback,
		     ( VarargsCallback) minmaxValueCallback,
		     rangeDialogData, minmaxMin, minmaxMax,
		     END_LIST);
    AmAddVaCallback( maxText, XmNactivateCallback,
		     ( VarargsCallback) minmaxValueCallback,
		     rangeDialogData, minmaxMin, minmaxMax,
		     END_LIST);
    AmAddVaCallback( minText, XmNlosingFocusCallback,
		     ( VarargsCallback) minmaxValueCallback,
		     rangeDialogData, minmaxMin, minmaxMax,
		     END_LIST);
    AmAddVaCallback( maxText, XmNlosingFocusCallback,
		     ( VarargsCallback) minmaxValueCallback,
		     rangeDialogData, minmaxMin, minmaxMax,
		     END_LIST);
    AmAddCallbackWidget( AM_PKG_ID, AM_OK, *dialog,
			 ( callbackProc) setRangeDialogOkCallback,
			 rangeDialogData);
    XtManageChild( formDlog);
}

/*************************************************************************
 *
 * Description:
 *      reassign the colors/range dialog creation callback if the dialog is
 *      destroyed.  This is used in the map editor slick interface
 * Arguments:
 *      w - launching widget.
 *      button - button colors/range dialog callback is atteached to.
 *      plot - plot whose colors/range is being edited
 *      callData - unused
 *
 **************************************************************************/

/* ARGSUSED */

static void
setRangeColorDialogCallback( Widget               w,
                             Widget               button,
                             AmPlotClass          *plot,
			     char		  **args,
                             XtPointer            callData)
{
    AmAddVaCallback( button, XmNactivateCallback,
                     ( VarargsCallback) AmRangeColorDialogCallback, 
		     plot, args,
		     END_LIST);
}

/***************************************************************************
 *
 * Description:
 *      callback for embedding in a slick subform.  It unmanages the dialog
 *      on the OK, and sets it up to be remanaged again when the button is
 *      hit.
 * Arguments:
 *      w - launching widget.
 *      plot - plot whose colormap is being edited
 *      callData - unused
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmRangeColorDialogCallback( Widget          	w,
			    AmPlotClass     	*plot,
			    char		**args,
			    XtPointer       	callData)
{
    Widget      dialog, okBtn;

    AmCreateRangeColorDialog( AmGetWidgetShell( w), plot, &dialog, &okBtn);
    if ( dialog != NULL) {
        AmAddVaCallback( okBtn, XmNactivateCallback,
                         ( VarargsCallback) AmUnmanageSubDialogCallback,
                         w, dialog, 
			 END_LIST);
        AmRemoveVaCallback( w, XmNactivateCallback,
                          ( VarargsCallback) AmRangeColorDialogCallback, 
			  plot, args, 
			  END_LIST);
        AmAddVaCallback( dialog, XmNdestroyCallback,
                         ( VarargsCallback) setRangeColorDialogCallback,
                         w, plot, args,
			 END_LIST);
        return( TRUE);
    } else {
        AmAddResourceToLog( "cantCreateRangeColorDialog");
        return( FALSE);
    }
}
