#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




/* selection set tools. */

#include <malloc.h>
#if ( !defined( _WIN32) && !defined(__BSD44__)) 
#include <alloca.h>
#endif
#include <stdlib.h>
#include <Xm/Xm.h>
#include <X11/cursorfont.h>
#include <anvil/list.h>
#include <anvil/datafmt.h>
#include <anvil/polytest.h>
#include <anvil/va_cback.h>
#include <anvil/userevents.h>
#include <anvil/state_hand.h>
#include <anvil/plot.h>
#include <anvil/objectplot.h>
#include <anvil/scatterplot.h>
#include <anvil/polygonplot.h>
#include <anvil/lineplot.h>
#include <anvil/motifdisplay.h>
#include <anvil/errlib.h>
#include <anvil/geometry.h>
#include <anvil/macros.h>
#include <anvil/selectutil.h>


typedef struct selectData_ {
     AmMotifDisplayClass	*motifDisplay;	/* motif display manager */
     AmPlotClass		*plot;		/* source plot to select from */
     AnvilBoolean		*selectFlagVec;	/* vector of selected flags */
     AnvilBoolean		*lastSelectFlagVec;	/* AmLast selection set */
     AmPolyTestFn		polyFunc;	/* polygon containment test function */
     AmDistFn		distFunc;	/* distance evaluation function */
     void		*clientData;	/* client arg for poly & AmDist funcs */
} selectDataRec, *selectDataPtr;

typedef struct pointClientData_ {
    AmAccessFn		xFn;		/* point x,y accessing functions */
    AmAccessFn		yFn;
    int			xOffset;	/* x and y offsets */
    int			yOffset;
    void		*clientData;	/* accessing function client data */
} pointClientDataRec, *pointClientDataPtr;

static void 
selectHandler( Widget		area,
	       selectDataPtr	selectData,
	       XButtonEvent	*event,
	       AnvilBoolean		*);

static void
closeSelectorCallback( int		eventId,
		       selectDataPtr	selectData,
		       XtPointer	callData);

/*************************************************************************
 * 
 * Description:
 * 	take the selection set flag vector, use the flags to generate
 *	a selection set, and store it in the plot.
 * Arguments:
 *	selectData - selection data including the select flag vector
 *	which is parallel to the plot's data vector.
 * 	count - number of selected elements.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure
 *
 **************************************************************************/

static AmErrorCode 
applySelectionSet( AmWindowClass	*window,
		   AmGCClass	 	*gc,
		   selectDataPtr 	selectData,
		   int	    	 	*count)
{
    int         ii, rows, index, size, diffCount;
    uchar_t     *data, *selectVec, *diffData;
    AmPlotClass	*plot = selectData->plot;

/* get the pertinent attributes from the source plot */

    selectData->plot->GetVaFields( AmNrows, &rows,
                                   AmNsize, &size,
                                   AmNdata, &data,
                                   NULL);

/* find the elements which were in the old selection set that aren't
   in the new selection set */

    diffCount = 0;
    for ( ii = 0; ii < rows; ii++) {
	if ( selectData->lastSelectFlagVec[ ii] &&
	     !selectData->selectFlagVec[ ii]) {
	  diffCount++;
	}
    }
    if ( diffCount > 0) {
	diffData = ( uchar_t *) malloc( diffCount*size);
	if ( diffData == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	diffCount = 0;
	for ( ii = 0; ii < rows; ii++) {
	    if ( selectData->lastSelectFlagVec[ ii] &&
		 !selectData->selectFlagVec[ ii]) {
		memcpy( &diffData[ diffCount*size], &data[ ii*size], size);
		diffCount++;
	    }
	}

/* undraw the elements in the old selection set that aren't in the new
   selection set */

	plot->SetSelection( diffData, diffCount);
	plot->DrawSelection( window, gc, plot->GetRect(), 1, FALSE);
    }

    for ( ii = 0, *count = 0;
          ii < rows;
          *count += selectData->selectFlagVec[ ii++]) {};

/* if there's objects selected, copy them to the selection array, set the
   selection, and draw the new selection */

    if ( *count > 0) {
        if (( selectVec = ( uchar_t *) malloc( *count*size)) == NULL) {
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
        index = 0;
        for ( ii = 0; ii < rows; ii++) {
            if ( selectData->selectFlagVec[ ii]) {
                memcpy( &selectVec[ size*index], &data[ size*ii], size);
                index++;
            }
        }
        selectData->plot->SetSelection( selectVec, *count);
	plot->DrawSelection( window, gc, plot->GetRect(), 1, TRUE);
    } 

/* copy the current selection set to the old selection set */

    memcpy( selectData->lastSelectFlagVec, 
	    selectData->selectFlagVec, rows);
    return( AmOK);
}

/*************************************************************************
 * 
 * Description:
 * 	take the generic list of points from the lasso, and generate
 *	a polygon from it.
 * Arguments:
 *	dragList - generic list of points
 *	poly - generated polygon (points are allocated in the function 
 * 	coords - coordinate system for scaling.
 * Returns:
 *	AmOK on success, AmMALLOC_FAILED on malloc failure
 *
 **************************************************************************/

static AnvilBoolean 
makeLassoPolygon( AmListPtr	dragList,
		  AmPolygonPtr	poly, 
		  AmCoordClass	*coords)
{
    float	x, y;
    AmPointPtr	pt;
    int		ii;

    poly->numPts = AmLengthLst( dragList);
    poly->pts = ( AmCoordPtr) malloc( poly->numPts*sizeof( AmCoordRec));
    if ( poly->pts == NULL) {
	DO_ERROR( AmMALLOC_FAILED);
	return( FALSE);
    }
    poly->xmin = HUGE;
    poly->ymin = HUGE;
    poly->xmax = -HUGE;
    poly->ymax = -HUGE;
    if ( poly->pts == NULL) {
	return( FALSE);
    }
    for ( ii = 0; dragList != NULL; dragList = dragList->next, ii++) {
	pt = ( AmPointPtr) dragList->data;

/* get the lasso point and scale it to the plot's coordinate system */

	coords->GetDataPt( pt->x, pt->y, &x, &y);
	poly->pts[ ii].x = x;
	poly->pts[ ii].y = y;
	if ( poly->pts[ ii].x < poly->xmin) {
	    poly->xmin = poly->pts[ ii].x;
	}
	if ( poly->pts[ ii].y < poly->ymin) {
	    poly->ymin = poly->pts[ ii].y;
	}
	if ( poly->pts[ ii].x > poly->xmax) {
	    poly->xmax = poly->pts[ ii].x;
	}
	if ( poly->pts[ ii].y > poly->ymax) {
	    poly->ymax = poly->pts[ ii].y;
	}
    }
    poly->flags = AM_MBR_FLAG;
    return( TRUE);
}

/*************************************************************************
 * 
 * Description:
 * 	take the polygon and the selection data (containing the selection
 *	flag vector), and set the selection flag for each item contained
 *	in the polygon.
 * Arguments:
 *	selectData - selection data including the select flag vector
 *	which is parallel to the plot's data vector.
 * 	poly - polygon to test against.
 *
 **************************************************************************/

static void 
polygonSelectFlags( selectDataPtr	selectData,
		    AmPolygonPtr		poly)
{
    int		rows, size, ii;
    uchar_t	*data;

    selectData->plot->GetVaFields( AmNrows, &rows, 
				   AmNsize, &size,
				   AmNdata, &data,
				   NULL);
    for ( ii = 0; ii < rows; ii++, data += size) {
	if (( selectData->polyFunc)( poly, data, selectData->clientData)) {
	    selectData->selectFlagVec[ ii] = TRUE;
	}
    }
}

/*************************************************************************
 * 
 * Description:
 * 	set the selection flag for the object closest to the point 
 * Arguments:
 *	selectData - selection data including the select flag vector
 *	which is parallel to the plot's data vector.
 * 	x, y - point to "nearest" test.
 *
 **************************************************************************/

static void 
pointSelectFlags( selectDataPtr	selectData,
		  float		x,
		  float		y)
{
    float	AmDist = HUGE, candDist;
    int		index, ii, rows, size;
    uchar_t	*data;
	
    selectData->plot->GetVaFields( AmNrows, &rows, 
				   AmNsize, &size,
				   AmNdata, &data,
				   NULL);
    index = 0;
    for ( ii = 0; ii < rows; ii++, data += size) {
	candDist = ( selectData->distFunc)( x, y, data, selectData->clientData);
	if ( candDist < AmDist) {
	    AmDist = candDist;
	    index = ii;
	}
    }
    if ( selectData->selectFlagVec[ index] == FALSE) {
	selectData->selectFlagVec[ index] = TRUE;
    } else {
	selectData->selectFlagVec[ index] = FALSE;
    }
}
	    
/*************************************************************************
 * 
 * Description:
 * 	lasso drag event handler. Listens to MotionNotify, ButtonPress,
 *	and ButtonRelease events.
 * Arguments:
 *	area - drawing area to draw the lasso on.
 *	selectData - selection data.
 *	dragList - generated generic list of points (allocated)
 *	event - X event
 *
 **************************************************************************/

static void 
lassoHandler( Widget		area,
	      selectDataPtr	selectData,
	      AmListPtr		*dragList,
	      XEvent		*event)
{
    XMotionEvent	*motionEvent;
    AmPointPtr		point;
    AmPlotClass     	*plot = selectData->plot;
    AmMotifDisplayClass *motifDisplay = selectData->motifDisplay;
    AmWindowClass       *window = motifDisplay->GetWindow();
    AmDisplayClass	*display = motifDisplay->GetDisplay();
    AmPointRec		pts[ 2];
    float		x, y;
    AmPolygonRec	poly;
    AmListPtr		dragPtr;
    int			count;
 
    if ( event->type == MotionNotify) {
	motionEvent = ( XMotionEvent *) event;
	point = ( AmPointPtr) malloc( sizeof( AmPointRec));
	point->x = motionEvent->x;
	point->y = motionEvent->y;
	*dragList = AmInsertNodeLst( point, *dragList);
	pts[ 1] = *point;
	point = ( AmPointPtr) (*dragList)->next->data;
	pts[ 0] = *point;
	display->DrawLines( window, motifDisplay->GetCursorGC(), pts, 2);

/* XOR undraw the closure for the lasso */

	point = ( AmPointPtr) AmLast( *dragList);
	pts[ 1] = *point;
	display->DrawLines( window, motifDisplay->GetCursorGC(), pts, 2);

/* XOR draw the closure for the lasso */
		
	point = ( AmPointPtr) (*dragList)->data;
	pts[ 0] = *point;
	display->DrawLines( window, motifDisplay->GetCursorGC(), pts, 2);

    } else if ( event->type == ButtonRelease) {

/* remove the event handler.  It will be reinstated on the next button press */

	AmRemoveVaHandler( area, PointerMotionMask | ButtonReleaseMask, FALSE,
			 ( VarargsHandler) lassoHandler, 
			 selectData, dragList, END_LIST);

/* undraw the lasso */

	for ( dragPtr = *dragList; 
	      dragPtr->next != NULL; 
	      dragPtr = dragPtr->next) {
	    point = ( AmPointPtr) dragPtr->data;
	    pts[ 0] = *point;
	    point = ( AmPointPtr) dragPtr->next->data;
	    pts[ 1] = *point;
	    display->DrawLines( window, motifDisplay->GetCursorGC(), pts, 2);
	}

/* XOR undraw the closure for the lasso */

	point = ( AmPointPtr) (*dragList)->data;
	pts[ 0] = *point;
	point = ( AmPointPtr) AmLast( *dragList);
	pts[ 1] = *point;
	display->DrawLines( window, motifDisplay->GetCursorGC(), pts, 2);

/* if there's more than two elements in the drag list, treat it as a 
   polygon, and do a containment, otherwise, do a distance test for the 
   closest object. Free the points from the polygon, because they're
   allocated by makeLassoPolygon */

	if ((( *dragList)->next == NULL) || 
	    (( *dragList)->next->next == NULL)) {
	    point = ( AmPointPtr) ( *dragList)->data;
	    plot->GetDataPt( point->x, point->y, &x, &y);
	    pointSelectFlags( selectData, x, y);
	} else {
	    if ( !makeLassoPolygon( *dragList, &poly, plot)) {
		return;
	    }
	    polygonSelectFlags( selectData, &poly);
	    free( poly.pts);
	} 
	
/* free the drag list and reset it */

	AmFreeLst( *dragList, ( AmItemFn) free);
	*dragList = NULL;

/* apply the selection set to the plot and draw it */

	if ( applySelectionSet( motifDisplay->GetWindow(),
				motifDisplay->GetGC(),
				selectData, &count) != AmOK) {
	    AmPrintResourceError( motifDisplay->GetShell(),
				"applySelectionFailed");
	}
    }
}

/* close the selector.  Remove the event handler, apply the selection
   set, and send the user event */

static void
closeSelector( selectDataPtr	selectData)
{
    char	        resbuf[ RESLEN];
    AmMotifDisplayClass *motifDisplay = selectData->motifDisplay;
    Widget		area = motifDisplay->GetArea();
    int			count;

/* remove the event handler and user callback, and reset the cursor */ 

    AmRemoveHandler( area, ButtonPressMask, 
		     ( XtEventHandler) selectHandler, selectData);
    AmRemoveCallbackWidget( AM_PKG_ID, AM_CLOSE_SELECTION, 
			    ( callbackProc) closeSelectorCallback, selectData,
			    motifDisplay->GetShell());
    motifDisplay->SetCursor( XC_arrow);

/* apply the selection set on button 2 */

    if ( applySelectionSet( motifDisplay->GetWindow(),
			    motifDisplay->GetGC(),
			    selectData, &count) != AmOK) {
	AmPrintResourceError( motifDisplay->GetShell(), 
			      "applySelectionFailed");
    }

/* send a selection event if more than 0 items have been selected, and
write an informative message */

    if ( count > 0) {
	AmSendUserEventChildren( AM_PKG_ID, AM_SET_SELECTION, 
				 selectData->plot, motifDisplay->GetShell());
	sprintf( resbuf, "%s.selectString", 
		 selectData->plot->getObjectName());
	motifDisplay->WriteResourceMessage( resbuf, count);
	motifDisplay->SetCursor( XC_arrow);
    }
}

/* ARGSUSED */

static void
closeSelectorCallback( int		eventId,
		       selectDataPtr	selectData,
		       XtPointer	callData)
{
    closeSelector( selectData);
}

/* free the selection data */

static void
freeSelectData( selectDataPtr	selectData)
{
    free( selectData->selectFlagVec);
    free( selectData->lastSelectFlagVec);
    free( selectData);
}

/*************************************************************************
 * 
 * Description:
 * 	button press handler to kick off event handlers for selecting 
 *	and completing selection of objects. Button 1 does either a 
 *	point or lasso selection, Button 2 completes the selection,
 *	and button 3 cancels the selection.
 * Arguments:
 *	area - drawing area to draw the lasso on.
 *	selectData - selection data.
 *	dragList - generated generic list of points (allocated)
 *	event - X event
 *
 **************************************************************************/

static void 
selectHandler( Widget		area,
	       selectDataPtr	selectData,
	       XButtonEvent	*event,
	       AnvilBoolean		*)
{
    AmListPtr			*dragList;
    AmPointPtr			point;
    AmMotifDisplayClass   	*motifDisplay = selectData->motifDisplay;
    char			resbuf[ RESLEN];

    if ( event->button == Button1) {

/* start another subselection */

	point = ( AmPointPtr) malloc( sizeof( AmPointRec));
	point->x = event->x;
	point->y = event->y;
	dragList = ( AmListPtr *) malloc( sizeof( AmListPtr));
	*dragList = AmInsertNodeLst( point, NULL);
	AmAddVaHandler( area, PointerMotionMask | ButtonReleaseMask, FALSE,
		      ( VarargsHandler) lassoHandler, 
		      selectData, dragList, END_LIST);
    } else if ( event->button == Button2) {
	closeSelector( selectData);
	freeSelectData( selectData);
    } else if ( event->button == Button3) {

/* cancel */

	AmRemoveHandler( area, ButtonPressMask, 
		       ( XtEventHandler) selectHandler, selectData);
        AmRemoveCallbackWidget( AM_PKG_ID, AM_CLOSE_SELECTION, 
			    ( callbackProc) closeSelectorCallback, selectData,
			    motifDisplay->GetShell());
	sprintf( resbuf, "%s.cancelString", 
		 selectData->plot->getObjectName());
	motifDisplay->WriteResourceMessage( resbuf);
	motifDisplay->SetCursor( XC_arrow);
	freeSelectData( selectData);
    }
}

/*************************************************************************
 * 
 * Description:
 * 	general purpose selector interface, takes a polygon containment
 *	testing function for lasso selecting, and a point proximity testing
 *	function for point selecting.
 * Arguments:
 *     motifDisplay -	display manager to control refreshes and cursors 
 *     plot -		plot to select objects from
 *     polyTest -	polygon containment test function 
 *     pointDist -	point distance function (used for comparisons)
 *     clientData -	client data for polygon AND point functions 
 *
 *************************************************************************/

static void 
selector( Widget		btn, 
	  AmMotifDisplayClass	*motifDisplay,
	  AmPlotClass		*plot,
	  AmPolyTestFn		polyTest,
	  AmDistFn		pointDist,
	  void			*clientData)
{
    selectDataPtr	selectData;
    int			nItems;

/* unset the current selection set if there is one */

    plot->SetSelection( NULL, 0);
    motifDisplay->Flush();
    if ( plot->getParent() == NULL) {
	plot->Draw( motifDisplay->GetWindow(), motifDisplay->GetGC(),
		    plot->GetRect(), 1);
    } else {
	plot->getParent()->Draw( motifDisplay->GetWindow(), 
				 motifDisplay->GetGC(),
				 plot->GetRect(), 1);
    }

/* create a passing structure with the motif display, plot, and a vector of
   flags which show items that have been selected, then add a button
   press handler which does most of the work */

    plot->GetVaFields( AmNrows, &nItems, NULL);
    selectData = ( selectDataPtr) malloc( sizeof( selectDataRec));
    selectData->motifDisplay = motifDisplay;
    selectData->plot = plot;
    selectData->selectFlagVec = ( AnvilBoolean *) calloc( sizeof( AnvilBoolean), nItems);
    selectData->lastSelectFlagVec = 
	( AnvilBoolean *) calloc( sizeof( AnvilBoolean), nItems);
    selectData->polyFunc = polyTest;
    selectData->distFunc = pointDist;
    selectData->clientData = clientData;
    AmAddHandler( btn, motifDisplay->GetArea(), ButtonPressMask, 
		( XtEventHandler) selectHandler, selectData,
		SELECT_PRIORITY);

/* add a user event handler which closes the select handler.  Typically
   called by callbacks which operate on the selection set */

    AmAddCallbackWidget( AM_PKG_ID, AM_CLOSE_SELECTION, motifDisplay->GetShell(),
			 ( callbackProc) closeSelectorCallback, selectData);

    motifDisplay->SetCursor( XC_crosshair);
}

/****************************************************************************
 * 
 * Description:
 * 	polygon containment and distance functions for point data 
 * Arguments:
 *	poly - polygon to contain with
 *
 ****************************************************************************/

AnvilBoolean AmPointPolyTestFn( AmPolygonPtr		poly,
		         uchar_t		*data,
		         pointClientDataPtr	pointClientData)
{
    return( AmPtInPoly(( pointClientData->xFn)
			( data + pointClientData->xOffset, 
			  pointClientData->clientData),
		     ( pointClientData->yFn)	
			( data + pointClientData->yOffset, 
			  pointClientData->clientData),
		     poly));
}

float AmPointDistFn( float		x,
		   float		y,
		   uchar_t		*data,
		   pointClientDataPtr	pointClientData)
{
    return( AmDist( x, y,
		  ( pointClientData->xFn)
		    ( data + pointClientData->xOffset, 
		      pointClientData->clientData),
		  ( pointClientData->yFn)	
		    ( data + pointClientData->yOffset, 
		      pointClientData->clientData)));
}

/* select objects to perform an operation on */

void AmSelectObjectCallback( Widget		btn,
			   AmMotifDisplayClass	*motifDisplay,
			   AmPlotClass		*plot,
			   XtPointer)
{
    pointClientDataPtr pointClientData;

    if ( !plot->IsType( AmOBJECT_PLOT)) {
	AmPrintResourceError( motifDisplay->GetShell(),
			    "objectSelectPlotWrongType");
	return;
    }
    if (( pointClientData = ( pointClientDataPtr)
		malloc( sizeof( pointClientDataRec))) == NULL) {
	AmPrintResourceError( "mallocFailed");
	return;
    }
    plot->GetVaFields( AmNxFn, &pointClientData->xFn,
		       AmNyFn, &pointClientData->yFn,
		       AmNxOffset, &pointClientData->xOffset,
		       AmNyOffset, &pointClientData->yOffset,
		       AmNclientData, &pointClientData->clientData,
		       NULL);

/* set up the selector to work with point data */

    selector( btn, motifDisplay, plot, 
	      ( AmPolyTestFn) AmPointPolyTestFn, ( AmDistFn) AmPointDistFn, 
	      pointClientData);
}

/* select scatter points to perform an operation on */

void AmSelectScatterCallback( Widget		btn,
			    AmMotifDisplayClass	*motifDisplay,
			    AmPlotClass		*plot,
			    XtPointer)
{
    pointClientDataPtr pointClientData;

    if ( !plot->IsType( AmSCATTER_PLOT)) {
	AmPrintResourceError( motifDisplay->GetShell(),
			    "scatterSelectPlotWrongType");
	return;
    }
    if (( pointClientData = ( pointClientDataPtr)
		malloc( sizeof( pointClientDataRec))) == NULL) { 
	AmPrintResourceError( "mallocFailed");
	return;
    }
    plot->GetVaFields( AmNxFn, &pointClientData->xFn,
		       AmNyFn, &pointClientData->yFn,
		       AmNxOffset, &pointClientData->xOffset,
		       AmNyOffset, &pointClientData->yOffset,
		       AmNclientData, &pointClientData->clientData,
		       NULL);

/* set up the selector to work with point data */

    selector( btn, motifDisplay, plot, 
	      ( AmPolyTestFn) AmPointPolyTestFn, ( AmDistFn) AmPointDistFn, 
	      pointClientData);
}

/* polygon containment and distance test functions for polygons */

AnvilBoolean AmPolygonPolyTestFn( AmPolygonPtr	poly,
		           AmPolygonPtr	candPoly,
		           void		*clientData)
{
    return( AmPolyInPoly( candPoly, poly, FALSE));
}

float AmPolygonDistFn( float	x,
		     float	y,
		     AmPolygonPtr	poly,
		     void	*clientData)
{
    int		ii;

/* calculate the polygon's extents if we have to, 
   and the distance from the center of the extents */
  
    if ( !( poly->flags & AM_MBR_FLAG)) {
	poly->xmin = HUGE;
	poly->xmax = -HUGE;
	poly->ymin = HUGE;
	poly->ymax = -HUGE;
	for ( ii = 0; ii < poly->numPts; ii++) {
	    if ( poly->pts[ ii].x < poly->xmin) {
		poly->xmin = poly->pts[ ii].x;
	    }
	    if ( poly->pts[ ii].y < poly->ymin) {
		poly->ymin = poly->pts[ ii].y;
	    }
	    if ( poly->pts[ ii].x > poly->xmax) {
		poly->xmax = poly->pts[ ii].x;
	    }
	    if ( poly->pts[ ii].y > poly->ymax) {
		poly->ymax = poly->pts[ ii].y;
	    }
	}
    }
    return( AmDist( x, y, ( poly->xmax + poly->xmin)/2,
		  ( poly->ymax + poly->ymin)/2));
}

/* select polygons to perform an operation on */

void AmSelectPolygonCallback( Widget		btn,
			    AmMotifDisplayClass	*motifDisplay,
			    AmPlotClass		*plot,
			    XtPointer)
{
    if ( !plot->IsType( AmPOLYGON_PLOT)) {
	AmPrintResourceError( motifDisplay->GetShell(),
			    "polygonSelectPlotWrongType");
	return;
    }

/* set up the selector to work with polygon information */

    selector( btn, motifDisplay, plot, 
	      ( AmPolyTestFn) AmPolygonPolyTestFn, ( AmDistFn) AmPolygonDistFn, NULL);
}

/* polygon containment and distance test functions for lines */

AnvilBoolean AmLinePolyTestFn( AmPolygonPtr	poly,
		        AmLinePtr		line,
		        void		*clientData)
{
    return( AmPtInPoly( line->x0, line->y0, poly) &&
	    AmPtInPoly( line->x1, line->y1, poly));
}

float AmLineDistFn( float		x,
		  float		y,
		  AmLinePtr	line,
		  void		*clientData)
{
    return( AmDist( x, y, ( line->x0 + line->x1)/2, ( line->y0 + line->y1)/2));
}


/* select lines to perform an operation on */

void AmSelectLineCallback( Widget			btn,
			 AmMotifDisplayClass	*motifDisplay,
			 AmPlotClass		*plot,
			 XtPointer)
{
    if ( !plot->IsType( AmLINE_PLOT)) {
	AmPrintResourceError( motifDisplay->GetShell(),
			    "lineSelectPlotWrongType");
	return;
    }
    selector( btn, motifDisplay, plot, 
	      ( AmPolyTestFn) AmLinePolyTestFn, ( AmDistFn) AmLineDistFn, NULL);
}

/* return the index of the item nearest to the screen coordinate */

int AmNearestIndex( AmPlotClass	*plot,
		  int		x,
		  int		y)
{
    int			size, rows, ii;
    AmAccessFn		xFn, yFn;
    uchar_t		*data, *ptr;
    AmSourceTypeEnum	sourceType;
    void		*clientData;
    int			index;
    float		bestDist, candDist, xpos, ypos;

    plot->GetVaFields( AmNdata, &data,
		       AmNrows, &rows,
		       AmNsize, &size,
		       AmNxFn, &xFn,
		       AmNyFn, &yFn,
		       AmNclientData, &clientData,
		       AmNsourceType, &sourceType,
		       NULL);

/* translate the screen point to the plot's coordinate system, and find
   the closest point in the data */

    plot->GetDataPt( x, y, &xpos, &ypos);
    index = 0;
    bestDist = AmDist( xpos, ypos, ( xFn)( data, clientData), 
				 ( yFn)( data, clientData));
    for ( ii = 1, ptr = &data[ size]; ii < rows; ii++, ptr += size) {
	candDist = AmDist( xpos, ypos, ( xFn)( ptr, clientData), 
				     ( yFn)( ptr, clientData));
	if ( candDist < bestDist) {
	    bestDist = candDist;
	    index = ii;
	}
    }
    return( index);
} 
