#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdarg.h>
#include <stdio.h>
#include <stddef.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Scale.h>
#include <Xm/Label.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>
#include <anvil/strlib.h>
#include <anvil/plot.h>
#include <anvil/tagstrs.h>
#include <anvil/motif.h>
#include <anvil/resource.h>
#include <anvil/bargraph.h>
#include <anvil/tick.h>
#include <anvil/plot_cback.h>
#include <anvil/errlib.h>
#include <anvil/userevents.h>
#include <anvil/motifdisplay.h>
#include <anvil/histogram.h>
#include <anvil/magic.h>
#include <anvil/macros.h>
#include <anvil/utility.h>
#include <anvil/errcode.h>

/* histogram magic window.  Allows the user to drag an outline window 
   and show a histogram of its contents in a dialog display */

const int HISTOGRAM_FONT_SIZE = 8;

class HistogramMagicClass : public AmMagicOutlineClass {
private:
    Widget		dialog;
    AmMotifDisplayClass	*histDisplay;
    AmBarGraphClass	*histGraph;
public:
    HistogramMagicClass( AmMotifDisplayClass     *motifDisplay,
                         AmPlotClass             *sourcePlot,
                         AnvilBoolean               callOnDrag);
    Widget              GetDialog() { return( dialog); };
    void                SetDialog( Widget dialogArg) { dialog = dialogArg; };
    AmMotifDisplayClass   *GetPrivateDisplay() { return( histDisplay); };
    AmBarGraphClass	*GetHistGraph() { return( histGraph); };
    virtual void        SetRect( AmRectangleRec const *rect);
    virtual             ~HistogramMagicClass();
};

static AmActionDefRec actionAreaDef[] = {
    { "closeBtn", ( XtCallbackProc) AmPopdownCback, NULL },
    { "helpBtn", NULL, NULL }
};

/****************************************************************************
 *
 * Description:
 *	user event handler ties to the AM_SET_FIELDS user event which
 *	adjusts the range of the histogram graph if the source plot 
 *	changes.  
 * Arguments:
 *      eventId - event from the source plot.
 *	histMagic - histogram magic window data structure.
 *	source - source object of the event (might not be the one of
 *	interest)
 * Modifications:
 *	Should also handle just about any other change to
 *	the source plot.  Should also handle the histogram vector
 *	length change currently dealt with in SetRect().
 *
 ****************************************************************************/

static void 
updateHistGraph( int			eventId,
		 HistogramMagicClass	*histMagic,
		 AmScreenObjClass		*source)
{
    AmDataGraphicClass	*sourcePlot = histMagic->GetSourcePlot();
    AmBarGraphClass	*histGraph = histMagic->GetHistGraph();
    int			numColors, *colorsTable;
    float		min, max;
    AmColorCellPtr	colors;

    if ( source == ( AmScreenObjClass *) sourcePlot) {
	sourcePlot->GetVaFields( AmNnumColors, &numColors,
			         AmNcolorsMin, &min,
			         AmNcolorsMax, &max,
			         AmNcolorsTable, &colorsTable,
				 AmNcolors, &colors,
			         NULL);
	if ( colorsTable == NULL) {
	    histGraph->SetVaFields( AmNxmin, min,
				    AmNxmax, max,
				    AmNcurrXmin, min,
				    AmNcurrXmax, max,
				    AmNnumColors, numColors,
				    AmNcolors, colors,
				    NULL);
	} else {
	    histGraph->SetVaFields( AmNxmin, 0.0,
				    AmNxmax, ( float) numColors,
				    AmNcurrXmin, 0.0,
				    AmNcurrXmax, ( float) numColors,
				    AmNnumColors, numColors,
				    AmNcolors, colors,
				    NULL);
	    free( colorsTable);
	}
	free( colors);
	histMagic->GetPrivateDisplay()->Flush();
    }
}

/****************************************************************************
 *
 * Description:
 *	destroy callback for the histogram dialog.  Sets the dialog
 *	to NULL, because the callbacks are removed automagically when
 *	the dialog is destroyed, and the delete calls XtDestroyWidget()
 *	on the dialog if it's not NULL.
 * Arguments:
 *      dialog - dialog which is being destroyed.
 *	histMagic - histogram magic window data structure.
 * Modifications:
 *	Should also handle just about any other change to
 *	the source plot.
 *
 ****************************************************************************/

/* ARGSUSED */

static void 
destroyHistogramCallback( Widget		dialog,
			  HistogramMagicClass	*histMagic,
			  XtPointer		callData)
{
    histMagic->SetDialog( NULL);
    delete histMagic;
}

/****************************************************************************
 *
 * Description:
 *	destructor: remove the user callback to update the histogram,
 *	the destroy callback to delete the histogram magic window,
 *	and destroy the dialog.  This may be called by the magic window
 *	destructor, or by the dialog delete callback.
 *
 ****************************************************************************/

HistogramMagicClass::~HistogramMagicClass()
{
    magicManager->RemoveUserCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT,
				      ( callbackProc) updateHistGraph,
				      this);
    if ( dialog != NULL) {
	XtRemoveCallback( dialog, XmNdestroyCallback,
			  ( XtCallbackProc) destroyHistogramCallback, this);
	XtDestroyWidget( dialog);
    }
}

/****************************************************************************
 *
 * Description:
 * 	take a plot, a rectangle, a histogram vector and a length, and
 *  	fill the histogram vector for a thematic source image plot 
 * Arguments:
 *      imagePlot - source image plot to histogram.
 *	x0, y0, x1, x1 - extents of the rectangle in the source plot's
 *	coordinate space.
 *	histogram - vector of values and histogram counts.
 *	length - length of the histogram vector.
 * Returns:
 *	TRUE
 *
 ****************************************************************************/

AnvilBoolean
AmImagePlotThematicHistogram( AmImagePlotClass  	*imagePlot,
			      AmRectangleRec const	*rect,
                              float               	x0,
                              float               	y0,
                              float               	x1,
                              float               	y1,
                              AmThematicHistPtr   	histogram,
                              int                 	length)
{
    AmRectangleRec      screenRect, sourceIntRect;
    AmRealRectRec       sourceRect;
    AmIntAccessFn       AmIntFn;
    int                 subRows, subCols, *table, size, colorsOffset;
    void                *colorsClientData;
    uchar_t             *data;
    float               subXmin, subYmin, subXmax, subYmax;
    AnvilBoolean             flipRows, flipCols;

/* the histogram is different if we're profiling an image or a
   scatter plot */

    imagePlot->GetVaFields( AmNcolorsIntFn, &AmIntFn,
                            AmNcolorsOffset, &colorsOffset,
                            AmNcolorsClientData, &colorsClientData,
                            AmNcolorsTable, &table,
                            AmNsubXmin, &subXmin,
                            AmNsubYmin, &subYmin,
                            AmNsubXmax, &subXmax,
                            AmNsubYmax, &subYmax,
                            AmNsubRows, &subRows,
                            AmNsubCols, &subCols,
                            AmNflipRows, &flipRows,
                            AmNflipCols, &flipCols,
                            AmNdata, &data,
                            AmNsize, &size,
                            NULL);

    screenRect = *rect;
    if ( !imagePlot->MatrixRealExtents( subXmin, subYmin, subXmax, subYmax,
                                        x0, y0, x1, y1,
                                        subRows, subCols, flipRows, flipCols,
                                        &screenRect, &sourceRect)) {
        return( FALSE);
    }
    AmMakeContainedIntRect( &sourceRect, &sourceIntRect);

    AN_DEBUG( 20, ( "sample = (%d, %d, %d, %d)\n",
                    sourceIntRect.x, sourceIntRect.y,
                    sourceIntRect.width, sourceIntRect.height));

/* perform a histogram on the matrix with an integer table */

    AmHistogramThematicSubmatrix( data, AmIntFn, colorsClientData,
                                  colorsOffset, size, subRows, subCols,
                                  &sourceIntRect, histogram,
                                  length, table);
    FREE( table);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	take a plot, a rectangle, a histogram vector and a length, and
 *  	fill the histogram vector for a thematic source scatter plot 
 * Arguments:
 *      imagePlot - source image plot to histogram.
 *	x0, y0, x1, x1 - extents of the rectangle in the source plot's
 *	coordinate space.
 *	histogram - vector of values and histogram counts.
 *	length - length of the histogram vector.
 * Returns:
 *	TRUE
 *
 ****************************************************************************/

AnvilBoolean 
AmScatterPlotThematicHistogram( AmScatterPlotClass	*scatterPlot,
			        float			x0,
			        float			y0,
			        float			x1,
			        float			y1,
			        AmThematicHistPtr	histogram,
			        int			length)
{
    AmRealRectRec	coordRect;
    AmAccessFn		xFn, yFn;
    AmIntAccessFn	AmIntFn;
    int			rows, *table, size, colorsOffset, xOffset, yOffset;
    void		*xyClientData, *colorsClientData;
    uchar_t		*data;

    scatterPlot->GetVaFields( AmNxFn, &xFn,
			      AmNyFn, &yFn,
			      AmNxOffset, &xOffset,
			      AmNyOffset, &yOffset,
			      AmNclientData, &xyClientData,
			      AmNcolorsIntFn, &AmIntFn,
			      AmNcolorsOffset, &colorsOffset,
			      AmNcolorsClientData, &colorsClientData,
			      AmNcolorsTable, &table,
			      AmNdata, &data,
			      AmNsize, &size,
			      AmNrows, &rows,
			      NULL);
    coordRect.x = x0;
    coordRect.y = y0;
    coordRect.width = x1 - x0;
    coordRect.height = y1 - y0;
    AmHistogramThematicSubscatter( data, xFn, yFn, xOffset, yOffset, 
				 xyClientData, AmIntFn, colorsOffset,
				 colorsClientData, size, rows, 
				 &coordRect, histogram, length, table);
    FREE( table);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	take a plot, a rectangle, a histogram vector and a length, and
 *  	fill the histogram vector for a continuous image plot.
 * Arguments:
 *      imagePlot - source image plot to histogram.
 *	x0, y0, x1, x1 - extents of the rectangle in the source plot's
 *	coordinate space.
 *	histogram - vector of values and histogram counts.
 *	length - length of the histogram vector.
 * Returns:
 *	TRUE
 *
 ****************************************************************************/

AnvilBoolean 
AmImagePlotContinuousHistogram( AmImagePlotClass	*imagePlot,
				AmRectangleRec const	*rect,
			        float			x0,
			        float			y0,
			        float			x1,
			        float			y1,
			        AmHistRec		*histogram,
			        int			length)
{
    AmRectangleRec	screenRect, sourceIntRect;
    AmRealRectRec	sourceRect;
    AmAccessFn		zFn;
    int			subRows, subCols, size, colorsOffset, ii;
    float		min, max, *binvec;
    AmDataTypeEnum	dataType;
    void		*colorsClientData;
    uchar_t		*data;
    float               subXmin, subYmin, subXmax, subYmax;
    AnvilBoolean		minClip, maxClip, flipRows, flipCols;

/* the histogram is different if we're profiling an image or a 
   scatter plot */

    imagePlot->GetVaFields( AmNcolorsFn, &zFn,
			    AmNcolorsType, &dataType,
			    AmNcolorsOffset, &colorsOffset,
			    AmNcolorsClientData, &colorsClientData,
			    AmNcolorsBinvec, &binvec,
			    AmNcolorsMin, &min,
			    AmNcolorsMax, &max,
                            AmNsubXmin, &subXmin,
                            AmNsubYmin, &subYmin,
                            AmNsubXmax, &subXmax,
                            AmNsubYmax, &subYmax,
                            AmNsubRows, &subRows,
                            AmNsubCols, &subCols,
                            AmNflipRows, &flipRows,
                            AmNflipCols, &flipCols,
			    AmNdata, &data,
			    AmNsize, &size,
			    AmNminClip, &minClip,
			    AmNmaxClip, &maxClip,
			    NULL);

/* transform the coordinates to matrix dimension space */

    screenRect = *rect;
    if ( !imagePlot->MatrixRealExtents( subXmin, subYmin, subXmax, subYmax,
                                        x0, y0, x1, y1,
                                        subRows, subCols, flipRows, flipCols,
                                        &screenRect, &sourceRect)) {
        return( FALSE);
    }
    AmMakeContainedIntRect( &sourceRect, &sourceIntRect);

    AN_DEBUG( 20, ( "sample = (%d, %d, %d, %d)\n",
                    sourceIntRect.x, sourceIntRect.y,
                    sourceIntRect.width, sourceIntRect.height));

/* use continuous values.  If there's no binvec, then create one */

    if ( binvec == NULL) {
	binvec = ( float *) malloc( length*sizeof( float));
	if ( binvec == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    return( FALSE);
	}
	for ( ii = 0; ii < length; ii++) {
	    binvec[ ii] = min + ii * ( max - min)/( length - 1);
	}
    }
    AmHistogramContinuousSubmatrix( data, zFn, colorsClientData,
				    colorsOffset, size, subRows, subCols, 
				    &sourceIntRect, histogram,
				    length, binvec, minClip, maxClip);
    FREE( binvec);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	take a plot, a rectangle, a histogram vector and a length, and
 *  	fill the histogram vector for a continuous scatter plot.
 * Arguments:
 *      imagePlot - source image plot to histogram.
 *	x0, y0, x1, x1 - extents of the rectangle in the source plot's
 *	coordinate space.
 *	histogram - vector of values and histogram counts.
 *	length - length of the histogram vector.
 * Returns:
 *	TRUE
 *
 ****************************************************************************/

AnvilBoolean 
AmScatterPlotContinuousHistogram( AmScatterPlotClass	*scatterPlot,
				float			x0,
				float			y0,
				float			x1,
				float			y1,
				AmHistRec			*histogram,
				int			length)
{
    AmRealRectRec		coordRect;
    AmAccessFn		xFn, yFn, zFn;
    int			rows, size, ii, colorsOffset, xOffset, yOffset;
    float		min, max, *binvec;
    AmDataTypeEnum	dataType;
    void		*xyClientData, *colorsClientData;
    uchar_t		*data;
    AnvilBoolean		minClip, maxClip;

    scatterPlot->GetVaFields( AmNxFn, &xFn,
			      AmNyFn, &yFn,
			      AmNxOffset, &xOffset,
			      AmNyOffset, &yOffset,
			      AmNclientData, &xyClientData,
			      AmNcolorsFn, &zFn,
			      AmNcolorsType, &dataType,
			      AmNcolorsOffset, &colorsOffset,
			      AmNcolorsClientData, &colorsClientData,
			      AmNdata, &data,
			      AmNsize, &size,
			      AmNrows, &rows,
			      AmNmin, &min,
			      AmNmax, &max,
			      AmNcolorsBinvec, &binvec,
			      AmNminClip, &minClip,
			      AmNmaxClip, &maxClip,
			      NULL);
    coordRect.x = x0;
    coordRect.y = y0;
    coordRect.width = x1 - x0;
    coordRect.height = y1 - y0;
    if ( binvec == NULL) {
	binvec = ( float *) malloc( length*sizeof( float));
	if ( binvec == NULL) {
	    DO_ERROR( AmMALLOC_FAILED);
	    return( FALSE);
	}
	for ( ii = 0; ii < length; ii++) {
	    binvec[ ii] = min + ii * ( max - min)/( length - 1);
	}
    }
    AmHistogramContinuousSubscatter( data, xFn, yFn, xOffset, yOffset,
				   xyClientData, zFn, colorsOffset,
				   colorsClientData, size, rows,
				   &coordRect, histogram, length, 
				   binvec, minClip, maxClip);
    FREE( binvec);
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	magic class SetRect virtual function is called whenever
 *	the histogram needs to be updated.  Check to make sure 
 *	that the histogram vector length is correct.  
 * Arguments:
 *	rectArg - rectangle that the magic window is being dragged
 *	or resized to.
 *
 ****************************************************************************/

void HistogramMagicClass::SetRect( AmRectangleRec const *rectArg)
{
    int		rows, cols, length, numColors, *colorsTable;
    float	x0, y0, x1, y1;

/* Set the rectangle. Where are we in the source plot's coordinate system? */

    rect = *rectArg;
    sourcePlot->GetDataPt( rect.x, rect.y + rect.height, &x0, &y0);
    sourcePlot->GetDataPt( rect.x + rect.width, rect.y, &x1, &y1);
    sourcePlot->GetVaFields( AmNrows, &rows,
			     AmNcols, &cols,
			     AmNnumColors, &numColors,
			     AmNcolorsTable, &colorsTable,
			     NULL);

/* if the source plot has a colors table, then it's a thematic plot */

    if ( colorsTable != NULL) {
	AmThematicHistPtr	histVec;

	histGraph->GetVaFields( AmNdata, &histVec, 
				AmNrows, &length,
				NULL);
	if ( length != numColors) {
	    length = numColors;
	    free( histVec);
	    if (( histVec = ( AmThematicHistPtr) 
		    calloc( length, sizeof( AmThematicHistRec))) == NULL) {
		DO_ERROR( AmMALLOC_FAILED);
		return;
	    }
	}

/* different histograms for image and scatter plot data */
	
	if ( sourcePlot->IsType( AmIMAGE_PLOT)) {
	    AmImagePlotThematicHistogram(( AmImagePlotClass *) sourcePlot,
					 &rect, x0, y0, x1, y1, 
					 histVec, length);
	} else if ( sourcePlot->IsType( AmSCATTER_PLOT)) {
	    AmScatterPlotThematicHistogram(( AmScatterPlotClass *) sourcePlot,
					   x0, y0, x1, y1, histVec, length);
	}

/* reset the data to update the graph */

	histGraph->SetVaFields( AmNdata, histVec,
				AmNrows, length,
				NULL);
	FREE( colorsTable);
    } else {
	AmHistPtr	histVec;

	histGraph->GetVaFields( AmNdata, &histVec, 
				AmNrows, &length,
				NULL);
	if ( length != numColors - 1) {
	    length = numColors - 1;
	    free( histVec);
	    if (( histVec = ( AmHistPtr) 
		    calloc( length, sizeof( AmHistRec))) == NULL) {
		DO_ERROR( AmMALLOC_FAILED);
		return;
	    }
	}

/* different histograms for image and scatter plot data */
	
	if ( sourcePlot->IsType( AmIMAGE_PLOT)) {
	    AmImagePlotContinuousHistogram(( AmImagePlotClass *) sourcePlot,
					   &rect, x0, y0, x1, y1, 
					   histVec, length);
	} else if ( sourcePlot->IsType( AmSCATTER_PLOT)) {
	    AmScatterPlotContinuousHistogram(( AmScatterPlotClass *) sourcePlot,
					     x0, y0, x1, y1, histVec, length);
	}

/* reset the data to update the graph */

	histGraph->SetVaFields( AmNdata, histVec,
				AmNrows, length,
				NULL);
    } 
    histDisplay->Flush();
}

/****************************************************************************
 *
 * Description:
 * 	magic window derived class for a histogram cursor.  Creates
 *	a magic outline window and an associated dialog to display
 *	the histogram as it updates.
 * Arguments:
 *	motifDisplay - motif display that the magic window's going 
 *	to be dragged around on.
 *	plot - plot that's going to be histogrammed.
 *	callOnDrag - whether to update when dragging, or just on placement.
 *
 ****************************************************************************/

HistogramMagicClass::HistogramMagicClass( AmMotifDisplayClass     *motifDisplay,
					  AmPlotClass             *plot,
					  AnvilBoolean               callOnDrag) :
				AmMagicOutlineClass( motifDisplay,
						   plot,
						   callOnDrag)

{
    AmActionRec           actionArea;
    Widget		title, actionForm, frame, form, area;
    Arg			args[ 20];
    int			n, numColors, fontSize, offset, *colorsTable;
    AmAccessFn		zFn;
    AmDataTypeEnum	dataType;
    void		*clientData;
    AmColorCellPtr        colors;
    AmTickClass		*leftTicks, *bottomTicks;
    float		min, max;
    unitType		units;
    const char		*fontName;
    AmXDisplayClass	*pXDisplay;
    AmXWindowClass	*pXWindow;

/* histograms can only be applied to image plots and scatter plots */

    if (( !plot->IsType( AmIMAGE_PLOT)) && !plot->IsType( AmSCATTER_PLOT)) {
	AmPrintResourceError("histogramOnlyOnScatterOrImage");
        return;
    }

/* create a dialog to display the histogram. */

    n = AmMakeArgList( args,
                     XmNdeleteResponse, XmDESTROY,
                     XmNautoUnmanage, FALSE,
                     NULL);
    dialog = XmCreateFormDialog( motifDisplay->GetShell(), "showHistogram", 
				 args, n);
    actionAreaDef[ 0].clientData = dialog;
    actionForm = AmMakeActionArea( dialog,
				 actionAreaDef, ELEMENTS( actionAreaDef),
				 &actionArea);
    XtVaSetValues( actionForm, StickLeft, StickRight, StickBottom, NULL);
    XtSetSensitive( actionArea.buttons[ 1], False);

    title = XtVaCreateManagedWidget( "histogramTitle",
				     xmLabelWidgetClass, dialog,
				     StickLeft, StickTop, 
				     NULL);

    frame = XtVaCreateManagedWidget( "histogramFrame",
				     xmFrameWidgetClass, dialog, 
				     StickLeft, BelowOf( title), 
				     StickRight, AboveOf( actionForm),
				     NULL);
    form = XtVaCreateManagedWidget( "histogramForm",
				    xmFormWidgetClass, frame,
				    NULL);
    area = XtVaCreateManagedWidget( "histogramArea",
				    xmDrawingAreaWidgetClass, form,
				    StickLeft, StickTop, 
				    StickRight, StickBottom,
				    XmNwidth, 320, 
				    XmNheight, 320,
				    NULL);

/* create the motif display where the histogram and tickmarks are going
   to be drawn */

    pXDisplay = new AmXDisplayClass( area);
    pXWindow = new AmXWindowClass( pXDisplay, area);
    histDisplay = new AmMotifDisplayClass( pXDisplay, pXWindow, TRUE);

/* get the number of colors in the source plot for the histogram length */

    plot->GetVaFields( AmNnumColors, &numColors,
                       AmNcolors, &colors,
		       AmNcolorsMin, &min,
		       AmNcolorsMax, &max,
		       AmNcolorsUnits, &units,
		       AmNcolorsFn, &zFn,
		       AmNcolorsOffset, &offset,
		       AmNcolorsClientData, &clientData,
		       AmNcolorsType, &dataType,
		       AmNcolorsTable, &colorsTable,
		       NULL);

/* if the table is NULL, then the source plot is continuous, otherwise
   it's thematic. Create the bargraph to display the histogram */

    if ( colorsTable == NULL) {
	AmHistPtr	histValues;

	if (( histValues = ( AmHistPtr) 
		    calloc( numColors - 1, sizeof( AmHistRec))) == NULL) {
	    AmPrintResourceError( motifDisplay->GetShell(), "mallocFailed");
	    return;
	}
	histGraph = 
	    new AmBarGraphClass( histDisplay->GetWindow(),
			       "histBarGraph", 
			       AmNdata, histValues,
			       AmNsize, sizeof( AmHistRec),
			       AmNrows, numColors - 1,
			       AmNheightType, AmINT,
			       AmNheightOffset, offsetof( AmHistRec, histValue),
			       AmNleftMargin, 80,
			       AmNbottomMargin, 80,
			       AmNrightMargin, 20,
			       AmNtopMargin, 20,
			       AmNx, 80,
			       AmNy, 20,
			       AmNwidth, 220,
			       AmNheight, 220,
			       AmNxmin, min,
			       AmNxmax, max,
			       AmNxunits, units,
			       AmNnumColors, numColors,
			       AmNcolors, colors,
			       AmNcolorsOffset, offsetof( AmHistRec, dataValue),
			       AmNcolorsType, AmFLOAT,
			       AmNcolorsMin, min,
			       AmNcolorsMax, max,
			       NULL);
    } else {
	AmThematicHistPtr	histValues;

	if (( histValues = ( AmThematicHistPtr) 
		    calloc( numColors, sizeof( AmThematicHistRec))) == NULL) {
	    AmPrintResourceError( motifDisplay->GetShell(), "mallocFailed");
	    return;
	}
	histGraph = 
	    new AmBarGraphClass( histDisplay->GetWindow(),
			       "histBarGraph", 
			       AmNdata, histValues,
			       AmNsize, sizeof( AmThematicHistRec),
			       AmNrows, numColors,
			       AmNheightType, AmINT,
			       AmNheightOffset, offsetof( AmHistRec, histValue),
			       AmNleftMargin, 80,
			       AmNbottomMargin, 80,
			       AmNrightMargin, 20,
			       AmNtopMargin, 20,
			       AmNx, 80,
			       AmNy, 20,
			       AmNwidth, 220,
			       AmNheight, 220,
			       AmNxmin, 0.0,
			       AmNxmax, ( float) numColors,
			       AmNnumColors, numColors,
			       AmNcolors, colors,
			       AmNcolorsOffset, offsetof( AmHistRec, dataValue),
			       AmNcolorsType, AmINT,
			       AmNcolorsTable, colorsTable,
			       NULL);
	free( colorsTable);
    }
    free( colors);

/* add the user callback to update the histogram when the source object
   has changed */

    motifDisplay->AddUserCallback( AM_PKG_ID, AM_SET_FIELD_OBJECT,
				   ( callbackProc) updateHistGraph,
				   this);


/* create tickmarks on the left and bottom.  The horizontal tickmarks should 
   be expanded to support names for thematic maps */

    fontName = AmGetDefault( "histogram.ticks.font", DEFAULT_FONT);
    fontSize = AmGetDefault( "histogram.ticks.fontSize", HISTOGRAM_FONT_SIZE);
    leftTicks = new AmTickClass( histDisplay->GetWindow(),
			       histGraph, "leftTicks",
			       AmNtickPosition, AmLEFT,
			       AmNtextVertical, FALSE,
			       AmNtopMargin, 20,
			       AmNleftMargin, 0,
			       AmNbottomMargin, 80,
			       AmNwidth, 80,
			       AmNheight, 220,
			       AmNx, 0,
			       AmNy, 20,
			       AmNfont, fontName,
			       AmNfontSize, fontSize,
			       NULL);
    bottomTicks = new AmTickClass( histDisplay->GetWindow(),
			         histGraph, "bottomTicks",
			         AmNtickPosition, AmBOTTOM,
			         AmNtextVertical, TRUE,
				 AmNleftMargin, 80,
				 AmNheight, 80,
				 AmNrightMargin, 20,
				 AmNbottomMargin, 0,
				 AmNx, 80,
				 AmNy, 240,
			         AmNwidth, 220,
			         AmNfont, fontName,
			         AmNfontSize, fontSize,
			         NULL);
    histDisplay->AddObject( histGraph);
    histDisplay->AddObject( leftTicks);
    histDisplay->AddObject( bottomTicks);
    XtManageChild( dialog);
    Update();

/* add callbacks to destroy the dialog. */

    XtAddCallback( dialog, XmNdestroyCallback,
		   ( XtCallbackProc) destroyHistogramCallback, this);
}

/****************************************************************************
 *
 * Description:
 * 	magic window callback for a histogram cursor.
 * Arguments:
 *	btn - button that kicks off the callback.
 *	motifDisplay - motif display that the magic window's going 
 *	to be dragged around on.
 *	plot - plot that's going to be histogrammed.
 *
 ****************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmShowHistogramCallback( Widget            btn,
		       AmMotifDisplayClass *motifDisplay,
		       AmPlotClass         *plot,
		       XtPointer)
{
    new HistogramMagicClass( motifDisplay, plot, TRUE);
    return( TRUE);
}
