#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <malloc.h>
#include <stdlib.h>
#include <Xm/Xm.h>
#include <X11/cursorfont.h>
#include <anvil/xdisplay.h>
#include <anvil/motif.h>
#include <anvil/motifdisplay.h>
#include <anvil/userevents.h>
#include <anvil/macros.h>
#include <anvil/plot_cback.h>
#include <anvil/events.h>
#include <anvil/state_hand.h>
#include <anvil/outline.h>
#include <anvil/dispmgrargs.h>

/* functions for zooming and panning around plots */

/***************************************************************************
 * 
 * Description:
 *	remove the zoom2x handler and restore the cursor to the arrow
 * Arguments:
 *	dispMgrArgs - contains the display manager, outline, and the
 *	coordinate system being zoomed 
 *
 **************************************************************************/

static void
removeZoom2xHandler( AmDispMgrArgsPtr dispMgrArgs)
{
    AmMotifDisplayClass     *displayManager = ( AmMotifDisplayClass *) 
				dispMgrArgs->displayManager;
    AmOutlineClass        *outline = dispMgrArgs->outline;

    displayManager->SetCursor( XC_arrow);
    displayManager->RemoveOutline( outline);
    AmFreeDispMgrArgs( dispMgrArgs);
}

/***************************************************************************
 * 
 * Description:
 * 	This button handler is activated on button press, which zooms 
 *	the object and removes the handler, and draws the outline rectangle 
 *	as the user moves the mouse.
 * Arguments:
 *	w - widget that the cursor's being dragged on.
 *	dispMgrArgs - contains the display manager, outline, and the
 *	coordinate system being zoomed 
 *	event - ButtonPress or MotionNotify event.
 *	continueDispatch - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

static void 
zoom2xHandler( Widget		w,
	       AmDispMgrArgsPtr	dispMgrArgs,
	       XEvent		*event,
	       AnvilBoolean		*continueDispatch)
{
    AmCoordClass		*coords = dispMgrArgs->coords;
    AmMotifDisplayClass   *displayManager = ( AmMotifDisplayClass *)
				dispMgrArgs->displayManager;
    AmObjectOutlineClass  *outline = ( AmObjectOutlineClass *) dispMgrArgs->outline;
    AmRectangleRec const	*rect;
    XButtonEvent	*buttonEvent=NULL;
    XMotionEvent	*motionEvent;
    float		currXmin, currXmax, currYmin, currYmax;
    int                 x=0, y=0;

/*
 * Check if the zoom operation was canceled.  And if not figure out
 *   the coordinates of the cursor.
 */
    
    if ( event->type == ButtonPress) {
        buttonEvent = ( XButtonEvent *)event;

        x = displayManager->VirtualCoord( buttonEvent->x); 
        y = displayManager->VirtualCoord( buttonEvent->y); 
        if ( buttonEvent->button == Button3) {
 
        /* Cancel, remove the old outline and the zoom handler. */ 
 
	    displayManager->SetCursor( XC_watch);
            displayManager->WriteResourceMessage( "zoomCancelMsg");

	    AmRemoveHandler( w, ButtonPressMask | PointerMotionMask, 
			  ( XtEventHandler) zoom2xHandler, dispMgrArgs);
            return;
        }
    } else if ( event->type == MotionNotify) {
        motionEvent = ( XMotionEvent *)event;
        x = displayManager->VirtualCoord( motionEvent->x); 
        y = displayManager->VirtualCoord( motionEvent->y); 
    }

/*
 * Make sure the zoom rectangle doesn't go beyond the extends of the
 *   object.
 */

    rect = coords->GetRect();
    if ( x + rect->width/4 > rect->x + rect->width) {
        x = rect->x + rect->width - rect->width/4;
    } else if ( x - rect->width/4 < rect->x) {
        x = rect->x + rect->width/4;
    }

    if ( y + rect->height/4 > rect->y + rect->height) {
        y = rect->y + rect->height - rect->height/4;
    } else if ( y - rect->height/4 < rect->y) {
        y = rect->y + rect->height/4;
    }

    if ( event->type == ButtonPress) {
        if ( buttonEvent->button == Button1) {

        /* Display a busy cursor and remove the old outline. */

	    displayManager->SetCursor( XC_watch);
	    displayManager->RemoveOutline( outline);

        /* Project the rectangle's coordinates into data space, reset the
           plot's coordinate system, update, and send a user event to notify
           that the plot's coordinate system has changed. */


	    coords->GetDataPt( x - rect->width/4, y + rect->height/4,
			       &currXmin, &currYmin);
	    coords->GetDataPt( x + rect->width/4, y - rect->height/4,
			       &currXmax, &currYmax);
	    coords->SetCoords( currXmin, currYmin, currXmax, currYmax);

	    AmRemoveHandler( w, ButtonPressMask | PointerMotionMask, 
		           ( XtEventHandler) zoom2xHandler, dispMgrArgs);
	    displayManager->Flush();
        }
    } else {
	outline->SetRectangle( x - rect->width/4, y - rect->height/4,
			       rect->width/2, rect->height/2);
    }
}
	
/***************************************************************************
 * 
 * Description:
 * 	Callback to do zoom 2x on a coordinate system, centered on 
 *	a cursor press, which drags an outline rectangle of the zoom
 *	area.
 * Arguments:
 *	w - widget that the cursor's being dragged on.
 *	motifDisplay - display manager containing the coordinate system
 *	to be zoomed.
 *	coords - coordinate system being zoomed 
 *	callData - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmZoom2xCallback( Widget			w,
		AmMotifDisplayClass	*motifDisplay,
		AmCoordClass		*coords,
		XtPointer		callData)
{
    AmRectangleRec const	  *rect;
    AmOutlineClass          *outline;

    motifDisplay->SetCursor( XC_crosshair);
    motifDisplay->WriteResourceMessage( "zoomInstructionMsg");

    rect = coords->GetRect();
    outline = new AmObjectOutlineClass( motifDisplay, rect->x, rect->y,
				      rect->width/2, rect->height/2);
    motifDisplay->AddOutline( outline);

    AmAddHandler( w, motifDisplay->GetArea(), 
        ButtonPressMask | PointerMotionMask,
        ( XtEventHandler) zoom2xHandler, 
        AmMakeDispMgrArgs( motifDisplay, coords, outline), ZOOM_2X_PRIORITY,
        ( AmStateHandlerFn) removeZoom2xHandler,
        ( AmStateHandlerFn) AmSuspendCursorHandler,
        ( AmStateHandlerFn) AmRestartCursorHandler);

    return( TRUE);
}

/***************************************************************************
 * 
 * Description:
 *	remove the drag zoom handler and restore the cursor to the arrow
 * Arguments:
 *	dispMgrArgs - contains the display manager, outline, and the
 *	coordinate system being zoomed 
 * Modifications:
 *	the remove zoom handlers should be consolidated.
 *
 **************************************************************************/

static void
removeDragZoomHandler( AmDispMgrArgsPtr dispMgrArgs)
{
    AmMotifDisplayClass 	*displayManager = ( AmMotifDisplayClass *) 
						dispMgrArgs->displayManager;
    AmOutlineClass        *outline = dispMgrArgs->outline;

    displayManager->SetCursor( XC_arrow);
    if ( outline != NULL) {
        displayManager->RemoveOutline( outline);
    }
    AmFreeDispMgrArgs( dispMgrArgs);
}

/***************************************************************************
 * 
 * Description:
 * 	drag zoom rectangle handler.  listens for the initial button
 *	press, drag and release events, and draws the outline rectangle 
 *	as the user moves the mouse.
 * Arguments:
 *	w - widget that the cursor's being dragged on.
 *	dispMgrArgs - contains the display manager, outline, and the
 *	coordinate system being zoomed 
 *	event - ButtonPress or MotionNotify event.
 *	continueDispatch - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

static void 
dragZoomHandler( Widget		w,
		 AmDispMgrArgsPtr	dispMgrArgs,
		 XEvent		*event,
		 AnvilBoolean	*continueDispatch)
{
    AmMotifDisplayClass	*displayManager = ( AmMotifDisplayClass *)
						dispMgrArgs->displayManager;
    AmCoordClass		*coords = dispMgrArgs->coords;
    AmObjectOutlineClass  *outline = ( AmObjectOutlineClass *)
					dispMgrArgs->outline;
    AnvilBoolean             buttonPressFlag = dispMgrArgs->flag;
    XButtonEvent	*buttonEvent;
    XMotionEvent	*motionEvent;
    int                 x, y, startX, startY;

    if ( event->type == ButtonPress) {
	buttonEvent = ( XButtonEvent *) event;

        x = displayManager->VirtualCoord( buttonEvent->x); 
        y = displayManager->VirtualCoord( buttonEvent->y); 

	if ( buttonEvent->button == Button3) {
 
        /* Cancel, remove the zoom handler. */ 

            displayManager->WriteResourceMessage( "zoomrectCancelMsg");

	    AmRemoveHandler( w, ButtonPressMask | ButtonReleaseMask |
			   PointerMotionMask, 
			   ( XtEventHandler) dragZoomHandler, dispMgrArgs);
	    displayManager->SetCursor( XC_arrow);

        } else if ( buttonEvent->button == Button1) {

        /* Switch cursors and set the rectangle origin. */

            displayManager->SetCursor( XC_bottom_right_corner);
            dispMgrArgs->flag = TRUE;
	    outline = new AmObjectOutlineClass( displayManager, x, y, 0, 0);
	    displayManager->AddOutline( outline);
	    dispMgrArgs->outline = outline;
        }
    } else if ( event->type == ButtonRelease) {
	buttonEvent = ( XButtonEvent *) event;


/* uh...he might have hit a button that isn't button 1, or button 3... */

	if ( buttonEvent->button == Button1) {
	    x = displayManager->VirtualCoord( buttonEvent->x); 
	    y = displayManager->VirtualCoord( buttonEvent->y); 

    /* Make sure the rectangle is within the plot's extents. */

	    coords->BoundPt( &x, &y);
	    outline->GetOrigin( &startX, &startY);

    /* Unset the cursor and zoom on the button release, and send a user event 
       that the plot's coordinate system has changed */

	    displayManager->SetCursor( XC_watch);
	    displayManager->RemoveOutline( outline);

	    if (( abs( x - startX) < 10) || ( abs( y - startY) < 10)) {
		displayManager->WriteResourceMessage( "zoomTooSmallMsg");
	    } else {
		displayManager->WriteResourceMessage( "zoomRectSuccessMsg");
		coords->ZoomRect( MIN( startX, x), MIN( startY, y),
				  abs( x - startX), abs( y - startY));
		displayManager->Flush();
	    }

	    AmRemoveHandler( w, ButtonPressMask | ButtonReleaseMask | 
			   PointerMotionMask, ( XtEventHandler) dragZoomHandler,
			   dispMgrArgs);
	    displayManager->SetCursor( XC_arrow);
	}

    } else {
        motionEvent = ( XMotionEvent *) event;

        x = displayManager->VirtualCoord( motionEvent->x); 
        y = displayManager->VirtualCoord( motionEvent->y); 

    /* Undraw the AmLast rectangle and draw the new one. */

	if ( buttonPressFlag) {
	    coords->BoundPt( &x, &y);
	    outline->GetOrigin( &startX, &startY);
            if ( outline != NULL) {
                outline->SetRectangle( MIN( startX, x), MIN( startY, y),
				       abs( x - startX), abs( y - startY));
            } 
	}
    }
}

/***************************************************************************
 * 
 * Description:
 * 	Callback to do drag zoom rectangle on a coordinate system, 
 * Arguments:
 *	w - widget causing the callback.
 *	motifDisplay - display manager containing the coordinate system
 *	to be zoomed.
 *	coords - coordinate system being zoomed 
 *	callData - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmZoomRectCallback( Widget		w,
		  AmMotifDisplayClass	*displayManager,
		  AmCoordClass		*coords,
		  XtPointer		callData)
{
    displayManager->WriteResourceMessage( "zoomRectInstructionMsg");
    displayManager->SetCursor( XC_crosshair);

    AmAddHandler( w, displayManager->GetArea(), 
        ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
        ( XtEventHandler) dragZoomHandler, 
        AmMakeDispMgrArgs( displayManager, coords, NULL), 
	ZOOM_RECTANGLE_PRIORITY,
        ( AmStateHandlerFn) removeDragZoomHandler,
        ( AmStateHandlerFn) AmSuspendCursorHandler,
        ( AmStateHandlerFn) AmRestartCursorHandler);

    return( TRUE);
}

/***************************************************************************
 * 
 * Description:
 * 	Callback to unzoom a coordinate system to its full extents.
 * Arguments:
 *	w - widget calling the callback.
 *	motifDisplay - display manager containing the coordinate system
 *	to be unzoomed.
 *	coords - coordinate system being unzoomed 
 *	callData - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmUnzoomCallback( Widget			w,
		AmMotifDisplayClass	*motifDisplay,
		AmCoordClass		*coords,
		XtPointer		callData)		
{
    float xmin, ymin, xmax, ymax;

    motifDisplay->WriteResourceMessage( "unzoomSuccessMsg");
    AmFlushHandler( motifDisplay->GetArea(), UNZOOM_PRIORITY);
    coords->GetRange( &xmin, &ymin, &xmax, &ymax);
    coords->SetCoords( xmin, ymin, xmax, ymax);
    motifDisplay->Flush();

    return( TRUE);
}

/***************************************************************************
 * 
 * Description:
 * 	Callback to unzoom a coordinate system to twice its current
 *	extents.
 * Arguments:
 *	w - widget calling the callback.
 *	motifDisplay - display manager containing the coordinate system
 *	to be unzoomed.
 *	coords - coordinate system being unzoomed 
 *	callData - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmUnzoomHalfCallback( Widget		w,	
		    AmMotifDisplayClass 	*motifDisplay,
		    AmCoordClass        	*coords,
		    XtPointer		callData)		
{
    float	currXmin, currYmin, currXmax, currYmax,
		newXmin, newYmin, newXmax, newYmax;

    motifDisplay->WriteResourceMessage( "unzoomHalfSuccessMsg");
    AmFlushHandler( motifDisplay->GetArea(), UNZOOM_PRIORITY);

    coords->GetCoords( &currXmin, &currYmin, &currXmax, &currYmax);
    newXmin = (currXmax + currXmin)/2 - (currXmax - currXmin);
    newYmin = (currYmax + currYmin)/2 - (currYmax - currYmin);
    newXmax = (currXmax + currXmin)/2 + (currXmax - currXmin);
    newYmax = (currYmax + currYmin)/2 + (currYmax - currYmin);
    coords->ClipRange( &newXmin, &newYmin);
    coords->ClipRange( &newXmax, &newYmax);
    coords->SetCoords( newXmin, newYmin, newXmax, newYmax);
    motifDisplay->Flush();

    return( TRUE);
}


/***************************************************************************
 * 
 * Description:
 * 	Callback refresh the display
 * Arguments:
 *	w - widget that the cursor's being dragged on.
 *	motifDisplay - display manager to refresh.
 *	callData - unused.
 *
 ***************************************************************************/

/* ARGSUSED */

AnvilBoolean
AmRefreshCallback( Widget			w,
		 AmMotifDisplayClass	*motifDisplay,
		 XtPointer		callData)
{
    int iOldCursor;

    iOldCursor = motifDisplay->SetCursor( XC_watch);

    motifDisplay->Draw( NULL, 0);

    motifDisplay->WriteResourceMessage( "refreshMsg");
    motifDisplay->SetCursor( iOldCursor);

    return( TRUE);
}
