#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/tagstrs.h>
#include <anvil/files.h>
#include <anvil/colorcell.h>
#include <anvil/resource.h>
#include <util/bilin.h>

/* find the distance AmBetween two colors */

float AmColorDist( AmColorCellPtr	a,
	         AmColorCellPtr	b)
{
    return( 0.33*abs( a->red - b->red) + 
	    0.5*abs( a->green - b->green) + 
	    0.17*abs( a->blue - b->blue));
}

/* 
   bilinear interpolation using s and t as parametric values 

	       x
	0,0 +-----+ 
	    |     |
	    |     | y
	    |     |
	    +-----+ 1,1

*/

/* interpolate 4 true color pixels on the RGB values */

void AmInterpolateTrueColor( uchar_t	*src,
			   void		*,		/* client data */
			   int		size,
			   float	x,
			   float	y,
			   uchar_t	*dst)
{
    int		red0, red1, red2, red3, redA, redB, red,
		green0, green1, green2, green3, greenA, greenB, green,
		blue0, blue1, blue2, blue3, blueA, blueB, blue;

    red0 = src[ RED]; 
    red1 = src[ sizeof( ulong_t) + RED];
    red2 = src[ sizeof( ulong_t)*2 + RED];
    red3 = src[ sizeof( ulong_t)*3 + RED];
    redA = (int)(red0*( 1.0 - x) + red1*x);
    redB = (int)(red2*( 1.0 - x) + red3*x);
    red = (int)(redA*( 1.0 - y) + redB*y);

    green0 = src[ GREEN]; 
    green1 = src[ sizeof( ulong_t) + GREEN];
    green2 = src[ sizeof( ulong_t)*2 + GREEN];
    green3 = src[ sizeof( ulong_t)*3 + GREEN];
    greenA = (int)(green0*( 1.0 - x) + green1*x);
    greenB = (int)(green2*( 1.0 - x) + green3*x);
    green = (int)(greenA*( 1.0 - y) + greenB*y);

    blue0 = src[ BLUE]; 
    blue1 = src[ sizeof( ulong_t) + BLUE];
    blue2 = src[ sizeof( ulong_t)*2 + BLUE];
    blue3 = src[ sizeof( ulong_t)*3 + BLUE];
    blueA = (int)(blue0*( 1.0 - x) + blue1*x);
    blueB = (int)(blue2*( 1.0 - x) + blue3*x);
    blue = (int)(blueA*( 1.0 - y) + blueB*y);
	
    *( ulong_t *) dst = red | ( green << GREEN_SHIFT) | ( blue << BLUE_SHIFT);
}

/* interpolate a pseudo color image by finding the closest color to the
   interpolated colors */

void AmInterpolatePseudoColor( uchar_t			*src,
			     pseudoColorInterpPtr	fnData,
			     int			size,
			     float			x,
			     float			y,
			     uchar_t			*dst)
{
    int			redA, redB, greenA, greenB, blueA, blueB, 
			bestIndex, ii;
    AmColorCellRec	interpColor;
    float		AmDist, mindist;
   
/* decompose the palette indices into the RGB components and
   interpolate AmBetween them */

    redA = (int)(fnData->palette[ src[ 0]].red*( 1.0 - x) +
	   fnData->palette[ src[ 1]].red*x);
    redB = (int)(fnData->palette[ src[ 2]].red*( 1.0 - x) +
	   fnData->palette[ src[ 3]].red*x);
    interpColor.red = (uchar_t)(redA*( 1.0 - y) + redB*y);
    greenA = (int)(fnData->palette[ src[ 0]].green*( 1.0 - x) +
	     fnData->palette[ src[ 1]].green*x);
    greenB = (int)(fnData->palette[ src[ 2]].green*( 1.0 - x) +
	     fnData->palette[ src[ 3]].green*x);
    interpColor.green = (uchar_t)(greenA*( 1.0 - y) + greenB*y);
    blueA = (int)(fnData->palette[ src[ 0]].blue*( 1.0 - x) +
	     fnData->palette[ src[ 1]].blue*x);
    blueB = (int)(fnData->palette[ src[ 2]].blue*( 1.0 - x) +
	     fnData->palette[ src[ 3]].blue*x);
    interpColor.blue = (uchar_t)(blueA*( 1.0 - y) + blueB*y);

/* find the closest color in the palette to the interpolated RGB value */

    mindist = AmColorDist( &interpColor, &fnData->palette[ 0]);
    bestIndex = 0;
    for ( ii = 1; ii < fnData->numColors; ii++) {
	AmDist = AmColorDist( &interpColor, &fnData->palette[ ii]);
	if ( AmDist < mindist) {
	    mindist = AmDist;
	    bestIndex = ii;
	}
    }
    *dst = fnData->palette[ bestIndex].pixel;
}

/* interpolate an intensity image in the standard fashion */

void AmInterpolateIntensity( uchar_t		*src,
			   intensityInterpPtr	fnData,
			   int			size,
			   float		x,
			   float		y,
			   uchar_t		*dst)
{
    float	a, b, c;

    a = ( fnData->fn)( src, NULL)*( 1.0 - x) + 
	( fnData->fn)( &src[ size], NULL)*x;
    b = ( fnData->fn)( &src[ size*2], NULL)*( 1.0 - x) + 
	( fnData->fn)( &src[ size*3], NULL)*x;
    c = a*( 1.0 - y) + b*y;
    ( fnData->setValFn)( dst, NULL, c);
}

/* interpolate a thematic image by finding the most popular value */

void AmInterpolateThematic( uchar_t		*src,
			  thematicInterpPtr	fnData,
			  int			size,
			  float			x,
			  float			y,
			  uchar_t		*dst)
{
    float	weights[ 4], bestWeight;
    int		ii, jj, values[ 4], bestIndex;

/* figure out the weights for each value */

    weights[ 0] = sqrt( SQR( 1.0 - x) + SQR( 1.0 - y));
    weights[ 1] = sqrt( SQR( x) + SQR( 1.0 - y));
    weights[ 2] = sqrt( SQR( 1.0 - x) + SQR( y));
    weights[ 3] = sqrt( SQR( x) + SQR( y));
    for ( ii = 0; ii < 4; ii++) {
	values[ ii] = ( fnData->fn)( &src[ ii*size], NULL);

/* accumulate the weights for the values if they're the same */

	for ( jj = 0; jj < ii; jj++) {
	    if ( values[ ii] == values[ jj]) {
		weights[ jj] += weights[ ii];
		break;
	    }
	}
    }

/* figure out the value with the best weight */

    bestWeight = 0.0;
    bestIndex = 0;
    for ( ii = 0; ii < 4; ii++) {
	if ( weights[ ii] > bestWeight) {
	    bestIndex = ii;
	    bestWeight = weights[ ii];
	}
    }

/* set the value */

    ( fnData->setValFn)( dst, NULL, values[ bestIndex]);
}

/* create the arguments for the interpolation routines based on a header file */

AnvilBoolean AmCreateInterpolationArgs( AmResourceDbClass	*db,
				 void			**clientData,
				 interpolateFn		*interpFn)
{
    AnvilBoolean		    trueColor;
    char		    *paletteFile = NULL, *colormapFile = NULL,
			    *tableFile, *dataTypeStr;
    int			    numColors=0, ii, red, green, blue, pixel;
    intensityInterpPtr	    intensityInterp;
    thematicInterpPtr	    thematicInterp;
    pseudoColorInterpPtr    pseudoColorInterp;
    AmColorCellPtr	    palette=NULL;
    FILE		    *fp;
    AmDataTypeEnum	    dataType;

    if ( db->GetResource( AmNtrueColor, &trueColor)) {

/* TrueColor interpolation */

	*interpFn = AmInterpolateTrueColor;
	*clientData = NULL;
	return( TRUE);
    } else if ( db->GetResource( AmNpalette, &paletteFile) ||
	        db->GetResource( AmNcolormap, &colormapFile)) {

/* PseudoColor interpolation */

	*interpFn = ( interpolateFn) AmInterpolatePseudoColor;
	if ( db->GetResource( AmNpalette, &paletteFile)) {

/* the palette file is the RGB colors and the RGB lookup table */

	    numColors = AmNumLines( paletteFile);
	    palette = ( AmColorCellPtr) 
		malloc( numColors*sizeof( AmColorCellRec));
	    if (( fp = fopen( paletteFile, "r")) == NULL) {
		return( FALSE);
	    }
	    for ( ii = 0; ii < numColors; ii++) {
		if ( fscanf( fp, "%d %d %d %d",
			     &pixel, &red, &green, &blue) != 4) {
		    fclose( fp);
		    return( FALSE);
		}
		palette[ ii].pixel = pixel;
	        palette[ ii].red = red;
	        palette[ ii].green = green;
	        palette[ ii].blue = blue;
	    }
	    fclose( fp);
	} else if ( !db->GetResource( AmNcolormap, &colormapFile) ||
		    !db->GetResource( AmNtable, &tableFile)) {

/* look for separate colormap and table files */

	    numColors = AmNumLines( colormapFile);
	    palette = ( AmColorCellPtr) 
		malloc( numColors*sizeof( AmColorCellRec));
	    if (( fp = fopen( colormapFile, "r")) == NULL) {
		return( FALSE);
	    }
	    for ( ii = 0; ii < numColors; ii++) {
		if ( fscanf( fp, "%d %d %d", &red, &green, &blue) != 3) {
		    fclose( fp);
		    return( FALSE);
		}
	        palette[ ii].red = red;
	        palette[ ii].green = green;
	        palette[ ii].blue = blue;
	    }
	    fclose( fp);
	    if (( fp = fopen( tableFile, "r")) == NULL) {
		return( FALSE);
	    }
	    for ( ii = 0; ii < numColors; ii++) {
		if ( fscanf( fp, "%d", &pixel) != 1) {
		    fclose( fp);
		    return( FALSE);
		}
		palette[ ii].pixel = pixel;
	    }
	}
	pseudoColorInterp = ( pseudoColorInterpPtr)
	    malloc( sizeof( pseudoColorInterpRec));
	pseudoColorInterp->palette = palette;
	pseudoColorInterp->numColors = numColors;
	*clientData = pseudoColorInterp;
	return( TRUE);
    } else if ( db->GetResource( AmNtable, &tableFile)) {

/* Thematic interpolation */
	
	*interpFn = ( interpolateFn) AmInterpolateThematic;
	thematicInterp = ( thematicInterpPtr) 
	    malloc( sizeof( thematicInterpRec));
	if ( !db->GetResource( AmNdataType, &dataTypeStr)) {
	    return( FALSE);
	}
	if ( !db->GetResource( AmNtable, &tableFile)) {
	    return( FALSE);
	}

/* convert data type string to enum, and init the integer get and
   set functions */

	dataType = AmStrToDataTypeEnum( dataTypeStr);
	thematicInterp->fn = AmDataTypeEnumToIntAccessFn( dataType);
	thematicInterp->setValFn = AmDataTypeEnumToIntSetFn( dataType);
	*clientData = thematicInterp;
	return( TRUE);
    } else { 

/* Intensity interpolation.  Just put in the get and set functions */

	*interpFn = ( interpolateFn) AmInterpolateIntensity;
	intensityInterp = ( intensityInterpPtr) 
	    malloc( sizeof( intensityInterpRec));
	if ( !db->GetResource( AmNdataType, &dataTypeStr)) {
	    return( FALSE);
	}
	dataType = AmStrToDataTypeEnum( dataTypeStr);
	intensityInterp->fn = AmDataTypeEnumToAccessFn( dataType);
	intensityInterp->setValFn = AmDataTypeEnumToSetFn( dataType);
	*clientData = intensityInterp;
	return( TRUE);
    } 
}
