#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */






#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <fcntl.h>
#include <anvil/endian.h>
#include <anvil/project.h>
#include <anvil/contour.h>
#include <anvil/files.h>
#if defined(_WIN32)
#include <anvil/ntresource.h>
#else 
#include <anvil/xresource.h>
#endif
#include <anvil/fns.h>
#include <anvil/envdefs.h>
#include <anvil/tagstrs.h>
#include <anvil/choppath.h>
#include <util/utils.h>

/* program to convert an image file to a polygon file using a contouring
   algorithm */

static const char *descriptionFile = 
	"$ANVIL_CONFIG/descriptions/contour_polygon_format";

static const int MAXCONTOUR = 1024;


int main( int argc, char *argv[])
{
    char 		dataPath[ PATHLEN], hdrPath[ PATHLEN], 
			*kind, *dataType; 
#if defined (_WIN32)
    AmNTResourceDbClass	*db;
#else 
    AmXResourceDbClass	*db;
#endif
    int			fdIn, fdOut, rows, cols, size, maxLength,
			polyRows, ii, count, numbytes, numPts;
    FILE		*hdrFp, *fp;
    float		binvec[ MAXCONTOUR];
    AmDataTypeEnum	dataTypeCode;
    AmAccessFn		zFn;
    float		step, min, max, xmin, ymin, xmax, ymax;
    AmListPtr		edgeList, inList, ptr;
    AmPolygonPtr	poly;
    uchar_t		*matrix, *markGrid;
    AmCoordPtr		points;
    AnvilBoolean		pipeFlag;
    char *fields[] = { AmNfilename, AmNrows, AmNcols, AmNsize, 
		       AmNtype, AmNdataType, AmNxmin, 
		       AmNymin, AmNxmax, AmNymax, AmNfileSource, 
		       NULL };

    if ( argc < 5) {
	fprintf( stderr, "usage %s inheader infile outfile ", argv[ 0]);
	fprintf( stderr, "[min max step] | binvec\n");
	exit( -1);
    }
    if ( getenv( CONFIG_ENV) == NULL) {
	fprintf( stderr, "environment variable %s not defined", CONFIG_ENV);
        exit( -1);
    }
    AmInitUnits();

/* construct the output file paths */

    sprintf( dataPath, "%s.dat", argv[ 3]);
    sprintf( hdrPath, "%s.hdr", argv[ 3]);

/* open the input header, and output files */
 
    if (( fdIn = AmOpenReadFile( argv[ 2], &pipeFlag)) < 0) {
	fprintf( stderr, "failed to open input file %s\n", argv[ 2]);
	exit( -1);
    }
    if (( fdOut = open( dataPath, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open output file %s\n", dataPath);
	exit( -1);
    }
    if (( hdrFp = fopen( hdrPath, "w")) == NULL) {
	fprintf( stderr, "failed to open output header %s\n", hdrPath);
	exit( -1);
    }

#if defined(_WIN32)
    db = new AmNTResourceDbClass( argv[ 1]);
#else
    db = new AmXResourceDbClass( argv[ 1]);
#endif
    if ( !db->IsValid()) {
	fprintf( stderr, "failed to read header file %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db->GetResource( AmNtype, &kind)) {
	fprintf( stderr, "failed to get file type from header %s\n", 
		 argv[ 2]);
	exit( -1);
    }
    if ( !strcmp( kind, AmNimagePlot)) {
	fprintf( stderr, "only image files may be contoured\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNrows, &rows)) {
	fprintf( stderr, "failed to read rows from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNcols, &cols)) {
	fprintf( stderr, "failed to read cols from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNsize, &size)) {
	fprintf( stderr, "failed to read size from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNdataType, &dataType)) {
	fprintf( stderr, "failed to read dataType from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNxmin, &xmin)) {
	fprintf( stderr, "failed to read xmin from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNymin, &ymin)) {
	fprintf( stderr, "failed to read ymin from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNxmax, &xmax)) {
	fprintf( stderr, "failed to read xmax from image file\n");
	exit( -1);
    }
    if ( !db->GetResource( AmNymax, &ymax)) {
	fprintf( stderr, "failed to read ymax from image file\n");
	exit( -1);
    }
    dataTypeCode = AmStrToDataTypeEnum( dataType);
    if ( dataTypeCode == AmNO_TYPE) {
	fprintf( stderr, "data type %s wasn't found\n", dataType);
	exit( -1);
    }
    zFn = AmDataTypeEnumToAccessFn( dataTypeCode);
    if ( zFn == NULL) {
	fprintf( stderr, "data type %s has no accessing function\n",
		 AmDataTypeEnumToStr( dataTypeCode));
	exit( -1);
    }

/* read the data */

    matrix = ( uchar_t *) malloc( size*rows*cols);
    if ( matrix == NULL) {
	fprintf( stderr, "failed to malloc %d bytes\n", size*rows*cols);
	exit( -1);
    }
    if (( numbytes = AmReadLoopFile( fdIn, matrix, size*rows*cols)) < 
	  size*rows*cols) {
	fprintf( stderr, "failed to read matrix %d expected, %d read\n",
		 size*rows*cols, numbytes);
	exit( -1);
    }
    AmCloseReadFile( fdIn, pipeFlag);

/* allocate the flag grid for the contour */
    
    markGrid = ( uchar_t *) malloc( rows*cols*2);
    if ( markGrid == NULL) {
	fprintf( stderr, "failed to malloc %d bytes\n", rows*cols*2);
	exit( -1);
    }

/* take the matrix and contour it into polygons. */ 
    
    if ( argc == 7) {
	min = atof( argv[ 4]);
	max = atof( argv[ 5]);
	if ( min >= max) {
	    fprintf( stderr, "min %.2f is greater than max %.2f\n", min, max);
	    exit( -1);
	}
	step = atof( argv[ 6]);
	if ( step <= 0.0) {
	    fprintf( stderr, "step %.2f, must be greater than 0\n", step);
	    exit( -1);
	}
	polyRows = 0;

/* contour each level and write it out, then free the contour.  We write
it out from the lowest to the highest, so the big polygons go AmFirst */

	for ( ; min < max; min += step) {
AN_DEBUG( 60, ("contouring %.2f", min));
	    memset( markGrid, '\0', rows*cols*2);
	    AmMakeContours( matrix, zFn, NULL, size, markGrid, rows, cols, 
			  min, &edgeList, &inList, &maxLength);
	    AmScaleContourList( edgeList, rows, cols, xmin, ymin, xmax, ymax);
	    AmScaleContourList( inList, rows, cols, xmin, ymin, xmax, ymax);
	    polyRows += AmLengthLst( edgeList);
	    polyRows += AmLengthLst( inList);

/* when the polygons are read in, there's a z-value associated with each
polygon. */
	    for ( ptr = inList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		write( fdOut, poly, sizeof( AmPolygonRec));
		write( fdOut, &max, sizeof( float));
		write( fdOut, poly->pts, poly->numPts*sizeof( AmCoordRec));
	    }
	    for ( ptr = inList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		free( poly->pts);
		free( poly);
	    } 
	    AmJustFreeLst( inList);

/* for the edge list, things are more complicated, because we have
to close the polygons */

	    for ( ptr = edgeList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		numPts = poly->numPts;
		points = AmClosePoly( poly, xmin, ymin, xmax, ymax, &numPts);
		poly->numPts = numPts;
		write( fdOut, poly, sizeof( AmPolygonRec));
		write( fdOut, &max, sizeof( float));
		write( fdOut, points, poly->numPts*sizeof( AmCoordRec));
	    }
	    for ( ptr = edgeList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		free( poly->pts);
		free( poly);
	    } 
	    AmJustFreeLst( edgeList);
	}
    } else {	

/* read the bin vector file in ASCII form and check to make sure it
   read OK, and that it didn't overflow the buffer */

	if (( fp = fopen( argv[ 4], "r")) == NULL) {
	    fprintf( stderr, "failed to open bin vector file %s\n", argv[ 4]);
	    exit( -1);
	}
	for ( count = 0; count < MAXCONTOUR; count++) {
	    if ( fscanf( fp, "%f", &binvec[ count]) != 1) {	
		break;
	    }
	}
	fclose( fp);
	if ( count == 0) {
	    fprintf( stderr, "failed to read bin vector file %s\n", argv[ 4]);
	    exit( -1);
	}
	if ( count == MAXCONTOUR) {
	    fprintf( stderr, "bin vector file %s has more than %d entries\n", 
		     argv[4], MAXCONTOUR);
	    exit( -1);
	}
	polyRows = 0;

/* now contour each of the bins */

	for ( ii = 0; ii < count; ii++) {
AN_DEBUG( 60, ("contouring %.2f", binvec[ ii]));
	    memset( markGrid, '\0', rows*cols*2);
	    AmMakeContours( matrix, zFn, NULL, size, markGrid, rows, cols, 
			  binvec[ ii], &edgeList, &inList, &maxLength);
	    AmScaleContourList( edgeList, rows, cols, xmin, ymin, xmax, ymax);
	    AmScaleContourList( inList, rows, cols, xmin, ymin, xmax, ymax);
	    polyRows += AmLengthLst( edgeList);
	    polyRows += AmLengthLst( inList);

/* when the polygons are read in, there's a z-value associated with each
polygon. */
	    for ( ptr = inList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		write( fdOut, poly, sizeof( AmPolygonRec));
		write( fdOut, &binvec[ ii], sizeof( float));
		write( fdOut, poly->pts, poly->numPts*sizeof( AmCoordRec));
	    }
	    for ( ptr = inList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		free( poly->pts);
		free( poly);
	    } 
	    AmJustFreeLst( inList);

/* for the edge list, things are more complicated, because we have
   to close the polygons */

	    for ( ptr = edgeList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		numPts = poly->numPts;
		points = AmClosePoly( poly, xmin, ymin, xmax, ymax, &numPts);
		poly->numPts = numPts;
		write( fdOut, poly, sizeof( AmPolygonRec));
		write( fdOut, &binvec[ ii], sizeof( float));
		write( fdOut, points, poly->numPts*sizeof( AmCoordRec));
	    }
	    for ( ptr = edgeList; ptr != NULL; ptr = ptr->next) {
		poly = ( AmPolygonPtr) ptr->data;
		free( poly->pts);
		free( poly);
	    } 
	    AmJustFreeLst( edgeList);
	}
    }

/* write out the header */

    fprintf( hdrFp, "filename:	%s\n", AmGetRootName( dataPath));
    fprintf( hdrFp, "rows:		%d\n", polyRows);
    fprintf( hdrFp, "size:		%d\n", sizeof( AmPolygonRec) + 
					   sizeof( float));
    fprintf( hdrFp, "type:		%s\n", AmNpolygonPlot);
    fprintf( hdrFp, "xmin:          %f\n", xmin);
    fprintf( hdrFp, "ymin:          %f\n", ymin);
    fprintf( hdrFp, "xmax:          %f\n", xmax);
    fprintf( hdrFp, "ymax:          %f\n", ymax);
    fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    fprintf( hdrFp, "note:		contoured\n");
#if BYTE_ORDER == LITTLE_ENDIAN
    fprintf( hdrFp, "byteOrder: littleEndian\n");
#else
    fprintf( hdrFp, "byteOrder: bigEndian\n");
#endif
    fprintf( hdrFp, "description:       %s\n", descriptionFile);
    fprintf( hdrFp, "fileSource:        contour\n");
    fprintf( hdrFp, "sourceFile:        %s\n", argv[ 1]);
    AmDumpHeaderExcept( argv[ 1], fields, hdrFp);
    fclose( hdrFp);
    close( fdOut);
}
