#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <anvil/tagstrs.h>
#include <anvil/macros.h>
#if defined(_WIN32)
#include <anvil/ntresource.h>
#else
#include <anvil/xresource.h>
#endif
#include <anvil/projclass.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/choppath.h>
#include <util/utils.h>


int main( int argc, char **argv)
{
    int		fd, rows, cols, size, fdOut, ii, x0, y0, x1, y1; 
    float	xmin, ymin, xmax, ymax, cropXmin, cropYmin, cropXmax, cropYmax,
		roundoff=0;
    FILE	*hdrFp;
    char	*fileType, *buffer, outpath[ 1024], hdrpath[ 1024], **projStrs; 
    char	*fields[] = { AmNfilename, AmNrows, AmNcols,
			      AmNxmin, AmNymin, AmNxmax, AmNymax, NULL };
    AnvilBoolean	pipeFlag;

    if ( argc < 5) {
	fprintf( stderr, 
		 "usage: %s inhdr infile proj outfile\n", argv[ 0]);
	exit( -1);
    }

/* get the extents of the original file */

#if defined(_WIN32)
    AmNTResourceDbClass db( argv[ 1]);
#else
    AmXResourceDbClass db( argv[ 1]);
#endif
    if ( !db.GetResource( AmNxmin, &xmin)) {
	fprintf( stderr, "failed to get xmin from %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db.GetResource( AmNymin, &ymin)) {
	fprintf( stderr, "failed to get ymin from %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db.GetResource( AmNxmax, &xmax)) {
	fprintf( stderr, "failed to get xmax from %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db.GetResource( AmNymax, &ymax)) {
	fprintf( stderr, "failed to get ymax from %s\n", argv[ 1]);
	exit( -1);
    }
    projStrs = AmReadStrvec( argv[ 3]);
    AmProjectionClass project(( const char **) projStrs);


/* find the extents that we're going to crop to */

    cropXmin = xmin;
    cropYmin = ymin;
    cropXmax = xmax;
    cropYmax = ymax;
    project.InvProjectPt( &cropXmin, &cropYmin);
    project.InvProjectPt( &cropXmax, &cropYmax);
    if ( fabs( cropXmax - cropXmin) > 1.0) {
	roundoff = 2.0;
    } else if ( fabs( cropXmax - cropXmin) > 0.5) {
	roundoff = 4.0;
    } else if ( fabs( cropXmax - cropXmin) > 0.25) {
	roundoff = 8.0;
    } else if ( fabs( cropXmax - cropXmin) > 0.125) {
	roundoff = 16.0;
    }
    cropXmin = ceil( cropXmin*roundoff)/roundoff;
    cropYmin = ceil( cropYmin*roundoff)/roundoff;
    cropXmax = floor( cropXmax*roundoff)/roundoff;
    cropYmax = floor( cropYmax*roundoff)/roundoff;
    AN_DEBUG( 10, ( "cropped latlon = %.3f, %.3f, %.3f, %.3f\n",
		    cropXmin, cropYmin, cropXmax, cropYmax));
    project.ProjectPt( &cropXmin, &cropYmin);
    project.ProjectPt( &cropXmax, &cropYmax);

/* find out what kind of file we're dealing with and open it */

    if ( !db.GetResource( AmNtype, &fileType)) {
	fprintf( stderr, "failed to get fileType from %s\n", argv[ 1]);
	exit( -1);
    }
    if (( fd = AmOpenReadFile( argv[ 2], &pipeFlag)) < 0) {
	fprintf( stderr, "failed to open file %s\n", argv[ 2]);
	exit(-1);
    }

/* if it's a relative path for the output file, then prepend the current
   working directory to it */

    sprintf( outpath, "%s.dat", argv[ 4]);
    sprintf( hdrpath, "%s.hdr", argv[ 4]);
    if (( fdOut = open( outpath, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open file %s\n", argv[ 4]);
	exit(-1);
    }
    if (( hdrFp = fopen( hdrpath, "w")) == NULL) {
	 fprintf( stderr, "failed to open header file %s\n", hdrpath);
	 exit( -1);
    }
    if ( !db.GetResource( AmNrows, &rows)) {
	fprintf( stderr, "failed to get rows from %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db.GetResource( AmNcols, &cols)) {
	fprintf( stderr, "failed to get cols from %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !db.GetResource( AmNsize, &size)) {
	fprintf( stderr, "failed to get size from %s\n", argv[ 1]);
	exit( -1);
    }

/* find out where the cropped coordinates are in the image coords */

    x0 = ( int) ( cols*(cropXmin - xmin)/( xmax - xmin));
    y0 = ( int) ( rows - rows*(cropYmax - ymin)/( ymax - ymin));
    x1 = ( int) ( cols*(cropXmax - xmin)/( xmax - xmin));
    y1 = ( int) ( rows - rows*(cropYmin - ymin)/( ymax - ymin));

/* allocate a row buffer, seek to the start of the subimage and read by rows */

    if (( buffer = ( char *) malloc( cols*size)) == NULL) {
	fprintf( stderr, "failed to malloc %d bytes\n", cols*size);
	exit( -1);
    }
    if ( !pipeFlag) {
	if ( lseek( fd, cols*y0*size, SEEK_SET) < 0) {
	    fprintf( stderr, "failed to seek %d bytes\n", cols*y0*size);
	    exit( -1);
	}
    } else {
	for ( ii = 0; ii < y0; ii++) {
	    if ( AmReadLoopFile( fd, buffer, cols*size) < 
			       cols*size) {
		fprintf( stderr, "failed to read %d bytes row %d\n", 
			 ii, cols*size);
		exit( -1);
	    }
	}
    }
    for ( ii = 0; ii < y1 - y0; ii++) {
	AmPrintHatch( ii, y1 - y0, NHATCH);
	if ( AmReadLoopFile( fd, buffer, cols*size) < cols*size) {
	    fprintf( stderr, "failed to read %d bytes row %d\n", 
		     ii, cols*size);
	    exit( -1);
	}
	if ( write( fdOut, &buffer[ x0*size], (x1 - x0)*size) < 
	     ( x1 - x0)*size) {
	    fprintf( stderr, "failed to write %d bytes row %d\n", 
		     ii, ( x1 - x0)*size);
	    exit( -1);
	}
    }

/* create the header file which describes the cropped output */
 
    AmCloseReadFile( fd, pipeFlag);
    close( fdOut);
    fprintf( hdrFp, "filename:      %s\n", AmGetRootName( outpath));
    fprintf( hdrFp, "xmin:          %f\n", cropXmin);
    fprintf( hdrFp, "ymin:          %f\n", cropYmin);
    fprintf( hdrFp, "xmax:          %f\n", cropXmax);
    fprintf( hdrFp, "ymax:          %f\n", cropYmax);
    fprintf( hdrFp, "rows:          %d\n", y1 - y0);
    fprintf( hdrFp, "cols:          %d\n", x1 - x0);
    fprintf( hdrFp, "sourceFile:    %s\n", argv[ 1]);
    AmDumpHeaderExcept( argv[ 1], fields, hdrFp);
    fclose( hdrFp);
}
