#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */






#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#include <fcntl.h>
#include <sys/types.h>
#include <anvil/endian.h>
#include <anvil/macros.h>
#include <anvil/strlib.h>
#include <anvil/list.h>
#include <anvil/files.h>
#include <anvil/choppath.h>
#include <util/utils.h>
#include <util/dem.h>

const char *descriptionFile = "$ANVIL_CONFIG/descriptions/short_image";

/* print the old DEM header, which contains most of the pertinent information
   anyway. */

void printHeader( oldDemHeaderPtr oldHead) 
{
    int	i, j;

    AN_DEBUG (3,(" \n\n" ));

    AN_DEBUG (3,( "filename .................... %.40s\n", oldHead->fileName ));
    AN_DEBUG (3,( "text ........................ %.40s\n", oldHead->text ));
    AN_DEBUG (3,( "section ..................... %.40s\n", oldHead->section ));
    AN_DEBUG (3,( "mc_origin ................... %.4s\n", oldHead->mcOrigin ));
    AN_DEBUG (3,( "dem_level ................... %.6s\n", oldHead->demLevel ));
    AN_DEBUG (3,( "pattern ..................... %.6s\n", oldHead->pattern ));
    AN_DEBUG (3,( "reference ................... %.6s\n", oldHead->reference ));
    AN_DEBUG (3,( "zone ........................ %.6s\n", oldHead->zone ));

    AN_DEBUG (3,( "parameters : \n" ));
    for ( i = 0; i < 15; i++) {
	AN_DEBUG (3,( "      %.24s \n", oldHead->parameters[ i] ));
    }

    AN_DEBUG (3,( "ground_unit ................. %.6s\n", oldHead->groundUnit ));
    AN_DEBUG (3,( "elev_unit ................... %.6s\n", oldHead->elevUnit ));
    AN_DEBUG (3,( "nsides ...................... %.6s\n", oldHead->nsides ));

    AN_DEBUG (3,( "corners : \n" ));
    for ( i = 0; i < 4; i++) {
	for ( j = 0; j < 2; j++) {
	    AN_DEBUG (3,( "      [%d,%d] : %.24s\n", i, j, oldHead->corners[ i][ j] ));
	}
    }

    AN_DEBUG (3,( "min ......................... %.24s\n", oldHead->min ));
    AN_DEBUG (3,( "max ......................... %.24s\n", oldHead->max ));
    AN_DEBUG (3,( "angle ....................... %.24s\n", oldHead->angle ));
    AN_DEBUG (3,( "accuracy .................... %.6s\n", oldHead->accuracy ));
    AN_DEBUG (3,( "x-axis resolution ........... %.12s\n", oldHead->resolution[0] ));
    AN_DEBUG (3,( "y-axis resolution ........... %.12s\n", oldHead->resolution[1] ));
    AN_DEBUG (3,( "z-axis resolution ........... %.12s\n", oldHead->resolution[2] ));
    AN_DEBUG (3,( "nrows ....................... %.6s\n", oldHead->nrows ));
    AN_DEBUG (3,( "ncols ....................... %.6s\n", oldHead->ncols ));

    AN_DEBUG (3,(" \n\n" ));
}

/* read a set of DEM files into a matrix */

int readDemFile( char 		*filename,
		 int	        *width,
		 int	        *height,
		 short		**data,
		 float		*xmin,		/* extents of DEM file */
		 float		*ymin,
		 float		*xmax,
		 float		*ymax,
		 float		*xres,		/* x and y resolutions */
		 float		*yres,		/* x and y resolutions */
		 unitType       *xyUnits,
		 int		*projZone, 	/* UTM projection zone */
		 char		*ellipse,
		 int		*nadCode)	/* 0=none,1=nad27,2=nad83 */
{
    oldDemHeaderPtr     header;
    newDemHeaderPtr	newHeader;
    demColHeadPtr	demColHead;
    char		buffer[ 8192], *ptr;
    int                 entries, reclen = 1024, ii, jj, kk, x0, y0, ncols, utmCode,
			elev, ellipseCode;
    AnvilBoolean		CRFlag, pipeFlag, feetFlag;
    float		cx, cy;
    short		*colPtr;
    FILE		*fp;

    AN_DEBUG (3,(" in readDemFile ... \n" ));

    if (( fp = AmOpenReadFilePointer( filename, &pipeFlag)) == NULL) {
	return( FALSE);
    }

    AN_DEBUG (3,(" --> reading %d bytes into header ... \n", reclen*2 ));
    fread( buffer, reclen*2, 1, fp);
    header = ( oldDemHeaderPtr) buffer;
    newHeader = ( newDemHeaderPtr) buffer;

    AN_DEBUG (3,(" --> header information : \n" ));
    printHeader( header);

    /* test the units code field in the header to find out whether this
       is in arc-seconds or meters (UTM projection)
       if the file is in UTM projection, then the extents are in meters
       otherwise, they're in arc-seconds.
       Find out if the elevations are in feet */
	
    utmCode = AmAtoi2( header->reference, sizeof( header->reference));
    *xyUnits = ( utmCode == 1) ? AmUNIT_METER : AmUNIT_ARCDEG;
    AN_DEBUG (3,("     UTM code  : %d \n", utmCode ));
    AN_DEBUG (3,("     x-y units : %d \n", *xyUnits ));

    feetFlag = ( AmAtoi2( header->elevUnit, sizeof( header->elevUnit)) == 1);
    if ( feetFlag ) {
	AN_DEBUG (3,("     z units   : FEET \n" ));
    } else {
	AN_DEBUG (3,("     z units   : METERS \n" ));
    }

    if ( *xyUnits == AmUNIT_METER) {
	*projZone = AmAtoi2( header->zone, sizeof( header->zone));
    } else {
	*projZone = 0;
    }
    AN_DEBUG (3,("     projection zone : %d \n", *projZone ));

    ellipseCode = AmAtoi2( newHeader->horzDatum, 
			 sizeof( newHeader->horzDatum)); 
    *nadCode = 0;
    if ( ellipseCode == 1) {
	strcpy( ellipse, "clrk66");
	*nadCode = 1;
    } else if ( ellipseCode == 2) {
	strcpy( ellipse, "WGS72");
    } else if ( ellipseCode == 3) {
	strcpy( ellipse, "WGS84");
    } else if ( ellipseCode == 4) {
	strcpy( ellipse, "GRS80");
	*nadCode = 2;
    } else {
	strcpy( ellipse, "");
    }

    AN_DEBUG (3,("     ellipse  : <%s> \n", ellipse ));
    AN_DEBUG (3,("     NAD code : %d \n", *nadCode ));

/* determine the four corners of the final rectange */

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0; ii < 4; ii++) {
	cx = AmAtof2( header->corners[ ii][ 0], 24);
	cy = AmAtof2( header->corners[ ii][ 1], 24);
	if ( cx < *xmin) *xmin = cx;
	if ( cy < *ymin) *ymin = cy;
	if ( cx > *xmax) *xmax = cx;
	if ( cy > *ymax) *ymax = cy;
    }

    AN_DEBUG (3,("     original extents : %.2f -> %.2f \n", *xmin, *xmax ));
    AN_DEBUG (3,("                        %.2f -> %.2f \n", *ymin, *ymax ));

/* if the ground units are in meters, round the extents to the sampling
   resolution (which is usually 30 meters) */
   
/* NOTE !!! occasionally, the maximum x-value will be something like 
   624119.44 but the AmLast point in the DEM file will have an x-coordinate 
   of 624120, so in order not to lose any points, we'll actually make an 
   extra couple of rows and columns ... (SMR 2/18/97) */

    *xres = AmAtof2( header->resolution[ 0], 12);
    *yres = AmAtof2( header->resolution[ 1], 12);

    AN_DEBUG (3,("     bin size : %.2f x %.2f \n", *xres, *yres ));

    if ( *xyUnits == AmUNIT_METER) {
	*xmin = ceil( *xmin/(*xres))*(*xres);
	*ymin = ceil( *ymin/(*yres))*(*yres);
	*xmax = floor( *xmax/(*xres))*(*xres);
	*ymax = floor( *ymax/(*yres))*(*yres);

	*xmin -= *xres;		*xmax += *xres;
	*ymin -= *yres;		*ymax += *xres;
    }

    AN_DEBUG (3,("          new extents : %.2f -> %.2f \n", *xmin, *xmax ));
    AN_DEBUG (3,("                        %.2f -> %.2f \n", *ymin, *ymax ));

/* we're assuming here that the DEM files have the same resolution */

    *width = ( int) floor(( *xmax - *xmin)/(*xres)) + 1;
    *height = ( int) floor(( *ymax - *ymin)/(*yres)) + 1;

    AN_DEBUG (3,(" --> matrix size : %d x %d \n", *width, *height ));

/* initialize matrix to -9999 so that we'll be able to distinguish un-set
   values later on ... */

    *data  = ( short *) calloc(( *width)*( *height), sizeof( short));
    for ( ii=0, kk=0; ii<*width; ii++ ) {
	for ( jj=0; jj<*height; jj++, kk++ ) {
	    (*data)[kk] = -9999;
	}
    }

/* if there's a carriage return, then rewind to the start of the file, and 
   read the header using fgets */

    CRFlag = ( strchr( buffer, '\n') != NULL);

    AN_DEBUG (3,(" closing DEM file ... \n" ));
    AmCloseReadFilePointer( fp, pipeFlag);
    AN_DEBUG (3,(" re-opening DEM file ... \n" ));
    if (( fp = AmOpenReadFilePointer( filename, &pipeFlag)) == NULL) {
	return( FALSE);
    }

    if ( CRFlag) {
	AN_DEBUG (3,(" CR flag is TRUE \n" ));
	AN_DEBUG (3,(" --> reading header using fgets (%d bytes) \n", 
				sizeof(buffer) ));
	fgets( buffer, sizeof( buffer), fp);
    } else {
	AN_DEBUG (3,(" CR flag is FALSE \n" ));
	AN_DEBUG (3,(" --> reading header using fread (%d bytes) \n", 
				reclen ));
	fread( buffer, reclen, 1, fp);
    }
	
    ncols = AmAtoi2( header->ncols, 6);
    AN_DEBUG (3,(" --> number of columns : %d \n", ncols ));

    if ( CRFlag) {
	for ( ii = 0; ii < ncols; ii++) {
	    AmPrintHatch( ii, ncols, NHATCH);
	    AN_DEBUG (3,("         COLUMN #%d \n", ii ));

/* read the buffer, extract the column header, and the start x & y. let's 
   round those suckers, shall we? */
		
	    fgets( buffer, sizeof( buffer), fp);
	    reclen = strlen( buffer) - 1;
	    demColHead = ( demColHeadPtr) buffer;
	    x0 = ( int) rint(( AmAtof2( demColHead->xGo, 24) - *xmin)/(*xres));
	    y0 = ( int) rint(( AmAtof2( demColHead->yGo, 24) - *ymin)/(*yres));
	    entries = AmAtoi2( demColHead->m, 6) * AmAtoi2( demColHead->n, 6);

	    AN_DEBUG (3,("             (%.2f, %.2f) \n", AmAtof2(demColHead->xGo,24),
							 AmAtof2(demColHead->yGo,24) ));
	    AN_DEBUG (3,("             x0 : %d   y0 : %d   entries : %d \n", x0, y0, entries ));

/* if the column starts outside the matrix, read and discard all its entries */

	    if (( x0 < 0) || ( y0 < 0) || ( x0 >= *width) || ( y0 >= *height)) {
		AN_DEBUG (3,(" OUT OF RANGE ???  SKIPPING ENTIRE COLUMN ??? \n" ));
		fprintf( stderr, "column %d row position out of range %d %d\n",
			 ii, x0, y0);
		entries -= ( reclen - sizeof( demColHeadRec))/6;
		while ( entries > 0) {
		    fgets( buffer, sizeof( buffer), fp);
		    entries -= strlen( buffer)/6;
		}
	    } else {
		colPtr = *data + (( *height) - y0 - 1)*( *width) + x0;
		ptr = buffer + sizeof( demColHeadRec);

/* read in CR-delimited blocks, and keep reading until we've got all the entries.
   Since the file is in column major order starting from the bottom, we step
   by -rows for each value read */

		for ( kk = 0; kk < entries; kk++) {
		    if ( colPtr >= *data) {
			elev = AmAtoi2( ptr, 6);
			if ( feetFlag) {
			    if ( ii==0 && kk==0 ) AN_DEBUG (3,(" APPLYING FEET-to-METERS CONVERSION !!! \n" ));
			    *colPtr = (short)(elev * 0.3048);
			} else {
			    *colPtr = elev;
			}
		    }
		    colPtr -= *width;
		    ptr += 6;
		    if (( kk < entries - 1) && ( ptr - buffer >= reclen - 5)) {
			fgets( buffer, sizeof( buffer), fp);
			reclen = strlen( buffer) - 1;
			ptr = buffer;
		    }
		}

	    }
	}
    } else {
	for ( ii = 0; ii < ncols; ii++) {
	    AmPrintHatch( ii, ncols, NHATCH);

/* figure out where it goes in the matrix */

	    fread( buffer, 1, reclen, fp);
	    demColHead = ( demColHeadPtr) buffer;
	    x0 = ( int) rint(( AmAtof2( demColHead->xGo, 24) - *xmin)/(*xres));
	    y0 = ( int) rint(( AmAtof2( demColHead->yGo, 24) - *ymin)/(*yres));
	    entries = AmAtoi2( demColHead->m, 6)*AmAtoi2( demColHead->n, 6);

	    AN_DEBUG (3,("             (%.2f, %.2f) \n", AmAtof2(demColHead->xGo,24),
							 AmAtof2(demColHead->yGo,24) ));
	    AN_DEBUG (3,("             x0 : %d   y0 : %d   entries : %d \n", x0, y0, entries ));

/* and if it's in range */

	    if (( x0 < 0) || ( y0 < 0) || ( x0 >= *width) || ( y0 >= *height)) {
		AN_DEBUG (3,(" OUT OF RANGE ???  SKIPPING ENTIRE COLUMN ??? \n" ));
		fprintf( stderr, "column %d row position out of range %d %d\n",
			 ii, x0, y0);
		entries -= (reclen - sizeof( demColHeadRec))/6;
		while ( entries > 0) {
		    fread( buffer, 1, reclen, fp);
		    entries -= reclen/6;
		}
	    } else {

/* read the in 1024-byte blocks, until we've got all the entries in the column,
   Since the file is in column major order starting from the bottom, we
   step by -rows for each value */

		colPtr = *data + (( *height) - y0 - 1)*( *width) + x0;
		ptr = buffer + sizeof( demColHeadRec);
		for ( kk = 0; kk < entries; kk++) {
		    if ( colPtr >= *data) {
			elev = AmAtoi2( ptr, 6);
			if ( feetFlag) {
			    if ( ii==0 && kk==0 ) AN_DEBUG (3,(" APPLYING FEET-to-METERS CONVERSION !!! \n" ));
			    *colPtr = (int)(elev*0.3048);
			} else {
			    *colPtr = elev;
			}
		    }
		    colPtr -= *width;
		    ptr += 6;
		    if (( kk < entries - 1) && ( ptr - buffer >= reclen - 5)) {
			fread( buffer, 1, reclen, fp);
			ptr = buffer;
		    }
		}
	    }
	}
    }
    AmCloseReadFilePointer( fp, pipeFlag);

/* if the units are in degrees, then the corners are really in arcseconds.
   We do this at the end, because there are other variables in arcseconds */

    if ( *xyUnits != AmUNIT_METER) {
	*xmin /= 3600.0;
	*ymin /= 3600.0;
	*xmax /= 3600.0;
	*ymax /= 3600.0;
    }

    return( TRUE);
}

int main( int argc, char *argv[]) 
{
    int   	width, height, fd, ii, projZone, nadCode;
    short 	*data;
    float 	xmin, xmax, ymin, ymax, xres, yres, min, max;
    unitType    xyUnits;
    char	dataPath[ PATHLEN], headerPath[ PATHLEN], ellipse[ PATHLEN];
    FILE	*fp;
  
/* initialize units */

    AmInitUnits();

/* verify arguments */

    if ( argc < 3) {
	fprintf( stderr, "usage: %s demfile outfile\n", argv[ 0]);
	exit( -1);
    }

/* derive data and header filenames from "path" argument, and open
   these two files */

    sprintf( dataPath, "%s.dat", argv[ 2]);
    sprintf( headerPath, "%s.hdr", argv[ 2]);
    if (( fp = fopen( headerPath, "w")) == NULL) {
	fprintf( stderr, "failed to open header file %s\n", headerPath);
	exit( -1);
    }
    if (( fd = open( dataPath, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open data file %s\n", dataPath);
	exit( -1);
    }

/* read the DEM file */

    if ( !readDemFile( argv[ 1], &width, &height, &data,
		       &xmin, &ymin, &xmax, &ymax, &xres, &yres,
		       &xyUnits, &projZone, ellipse, &nadCode)) {
	fprintf( stderr, "failed to read DEM file %s\n", argv[ 1]);
	exit( -1);
    }

/* write out the binary matrix to the binary data file, and close
   that file */

    write( fd, data, width*height*2);
    close( fd);

/* find the vertical range of the data */

    min = MY_HUGE;
    max = -MY_HUGE;
    for ( ii = 0; ii < width*height; ii++) {
	if ((( float) data[ ii]) < min) {
	    min = ( float) data[ ii];
	}
	if ((( float) data[ ii]) > max) {
	    max = ( float) data[ ii];
	}
    }

/* write out the header information to the ASCII header file */

    fprintf( fp, "filename:	%s\n", AmGetRootName( dataPath));
    fprintf( fp, "xmin: 		%f\n", xmin);
    fprintf( fp, "ymin: 		%f\n", ymin);
    fprintf( fp, "xmax: 		%f\n", xmax);
    fprintf( fp, "ymax: 		%f\n", ymax);
    fprintf( fp, "rows: 		%d\n", height);
    fprintf( fp, "cols: 		%d\n", width);
    fprintf( fp, "min: 		%f\n", min);
    fprintf( fp, "max: 		%f\n", max);
    fprintf( fp, "dataType: 	short\n");
    fprintf( fp, "size: 		2\n");

    if ( xyUnits == AmUNIT_METER) { 
	fprintf( fp, "xunits: 	meter\n");
	fprintf( fp, "xvariable: 	distance\n");
	fprintf( fp, "yunits: 	meter\n");
	fprintf( fp, "yvariable: 	distance\n");
    } else if ( xyUnits == AmUNIT_ARCDEG) {
	fprintf( fp, "xunits: 	arcdeg\n");
	fprintf( fp, "xvariable: 	longitude\n");
	fprintf( fp, "yunits: 	arcdeg\n");
	fprintf( fp, "yvariable: 	latitude\n");
    }

    fprintf( fp, "zunits: 	meter\n");
    fprintf( fp, "zvariable: 	elevation\n");
    fprintf( fp, "numdims: 	2\n");
    fprintf( fp, "type: 		imagePlot\n");
    fprintf( fp, "fileSource: 	DEM\n");

    if ( xyUnits == AmUNIT_METER) {
	fprintf( fp, 
		 "inProjection:	\"proj=utm\" \"zone=%d\" \"ellps=%s\"\n",
		 projZone, "WGS84");
    }

#if BYTE_ORDER == LITTLE_ENDIAN
    fprintf( fp, "byteOrder: littleEndian\n");
#else
    fprintf( fp, "byteOrder: bigEndian\n");
#endif
    fprintf( fp, "description:       %s\n", descriptionFile);

    fclose( fp);

}
