#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */






#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <anvil/macros.h>
#include <anvil/project.h>
#include <anvil/tiffio.h>
#include <anvil/choppath.h>
#include <util/utils.h>

const char *ucharDescriptionFile = 
	"$ANVIL_CONFIG/descriptions/unsignedChar_image";

const char *ulongDescriptionFile = 
	"$ANVIL_CONFIG/descriptions/unsignedLong_image";


#define MAXMAXVALUE 1023
#define NEXTSAMPLE \
{ \
    if ( bitsleft == 0) { \
	inp++; \
	bitsleft = 8; \
    } \
    bitsleft -= bitsPerPixel; \
    sample = ( *inp >> bitsleft) & maxval; \
}


int main( int argc, char *argv[])
{
    TIFF	*tif;
    ushort_t	bitsPerPixel, samplesPerPixel, photometric,
		*redCmap, *greenCmap, *blueCmap;
    int		row, rows, col, cols, maxval, outbuflen, fdOut, numcolors, 
		bitsleft, ii, lastNonZeroMarker=0;
    uchar_t	*inbuf, *outbuf, *inp, *outp, sample;
    char	dataPath[ PATHLEN], hdrpath[ PATHLEN], palettePath[ PATHLEN];
    AnvilBoolean	greyscale;
    FILE	*paletteFp, *inHdrFp, *hdrFp;
    float	xspacing, yspacing, xmin, ymin, xmax, ymax, unused1, unused2;

    if ( argc < 3) {
	fprintf( stderr, "usage %s inheader infile outfile\n", argv[ 0]);
	exit( -1);
    }
    if  (( inHdrFp = fopen( argv[ 1], "r")) == NULL) {
	fprintf( stderr, "failed to open header file %s\n", argv[ 1]);
	exit( -1);
    }

/* read the .tfw header file to get the coordinates for the DRG file */

    if ( fscanf( inHdrFp, "%f\n", &xspacing) != 1) {
	fprintf( stderr, "failed to scan xspacing from %s\n", argv[ 1]);
    }
    if ( fscanf( inHdrFp, "%f\n", &unused1) != 1) {
	fprintf( stderr, "failed to scan second line from %s\n", argv[ 1]);
    }
    if ( fscanf( inHdrFp, "%f\n", &unused2) != 1) {
	fprintf( stderr, "failed to scan third line from %s\n", argv[ 1]);
    }
    if ( fscanf( inHdrFp, "%f\n", &yspacing) != 1) {
	fprintf( stderr, "failed to scan yspacing from %s\n", argv[ 1]);
    }
    if ( fscanf( inHdrFp, "%f\n", &xmin) != 1) {
	fprintf( stderr, "failed to scan xmin from %s\n", argv[ 1]);
    }
    if ( fscanf( inHdrFp, "%f\n", &ymax) != 1) {
	fprintf( stderr, "failed to scan ymax from %s\n", argv[ 1]);
    }
    fclose( inHdrFp);
    xspacing = fabs( xspacing);
    yspacing = fabs( yspacing);

/* open the tiff file for reading */

    if (( tif = TIFFOpen( argv[ 2], "r")) == NULL) {
	fprintf( stderr, "failed to open TIFF file %s", argv[ 2]);
	exit( -1);
    }

/* create the fully qualified file paths, and file names without paths */

    sprintf( dataPath, "%s.dat", argv[ 3]);
    sprintf( hdrpath, "%s.hdr", argv[ 3]);
    sprintf( palettePath, "%s.palette", argv[ 3]);

/* open the files */

    if (( fdOut = open( dataPath, O_CREAT | O_WRONLY, PERMS)) < 0) {
	fprintf( stderr, "failed to open output file %s", dataPath);
	exit( -1);
    }
    if  (( hdrFp = fopen( hdrpath, "w")) == NULL) {
	fprintf( stderr, "failed to open header file %s", hdrpath);
	exit( -1);
    }
    if  (( paletteFp = fopen( palettePath, "w")) == NULL) {
	fprintf( stderr, "failed to open palette file %s", palettePath);
	exit( -1);
    }

/* get the number of bits per pixel, samples per pixel, and the photometric */

    if  ( !TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel)) {
	bitsPerPixel = 1;
    }
    if  ( !TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel)) {
	samplesPerPixel = 1;
    }
    if  ( !TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
	fprintf( stderr, "error getting photometric from file %s\n", argv[ 1]);
	exit( -1);
    }

/* get the dimensions of the image */

    TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &cols);
    TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &rows);
    maxval = ( 1 << bitsPerPixel) - 1;
    if ( maxval == 1 && samplesPerPixel == 1) {
	greyscale = TRUE;
    } else {
	switch ( photometric) {
	case PHOTOMETRIC_MINISBLACK:
	case PHOTOMETRIC_MINISWHITE:
	    greyscale = TRUE;
	    break;
	case PHOTOMETRIC_PALETTE:
	
/* if it's colormapped, then read the colormap into R, G and B vectors */

	    if ( !TIFFGetField( tif, TIFFTAG_COLORMAP, 
			        &redCmap, &greenCmap, &blueCmap)) {
		 fprintf( stderr, "error getting colormaps from file %s\n", 
			  argv[ 1]);
		 exit( -1);
	    }
	    numcolors = maxval + 1;
	    if  (( paletteFp = fopen( palettePath, "w")) == NULL) {
		fprintf( stderr, "failed to open palette file %s", palettePath);
		exit( -1);
	    }
	    lastNonZeroMarker = 0;
	    for ( ii = 0; ii < numcolors; ii++) {
		if ((( redCmap[ ii] >> 8) != 0) ||
		    (( greenCmap[ ii] >> 8) != 0) ||
		    (( blueCmap[ ii] >> 8) != 0)) {
		    lastNonZeroMarker = ii;
		}
	    }
	    lastNonZeroMarker++;
	    for ( ii = 0; ii < lastNonZeroMarker; ii++) {
		fprintf( paletteFp, "%d %d %d %d\n", 
			 ii, redCmap[ ii] >> 8, 
			 greenCmap[ ii] >> 8, blueCmap[ ii] >> 8);
	    }
	    fclose( paletteFp);
	    maxval = MAXMAXVALUE;
	    greyscale = FALSE;
	    break;
	case PHOTOMETRIC_RGB:
	    greyscale = FALSE;
	    break;
	default:
	    fprintf( stderr, "unrecognized photometric %d in file %s\n", 
	     	     photometric, argv[ 1]);
	    exit( -1);
	}
    }

/* allocate a buffer to read the scanline, then read the file scanline by
   scanline */

    inbuf = ( uchar_t *) malloc( TIFFScanlineSize( tif));

/* RGB images map to 32 bit values, all others to 8-bit values */

    if ( photometric == PHOTOMETRIC_RGB) {
	outbuflen = cols*sizeof( unsigned long);
    } else {
	outbuflen = cols;
    }
    outbuf = ( uchar_t *) malloc( outbuflen);
    for ( row = 0; row < rows; row++) {
	AmPrintHatch( row, rows, NHATCH);
	if ( TIFFReadScanline( tif, inbuf, row, 0) < 0)  {
	    fprintf( stderr, "bad data row %d in file %s\n", row, argv[ 1]);
	    exit(  -1);
	}
	bitsleft = 8;
	outp = outbuf;
	inp = inbuf; 
	switch ( photometric) {
	case PHOTOMETRIC_MINISBLACK:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = sample;
	    }
	    break;
	case PHOTOMETRIC_MINISWHITE:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = maxval - sample;
	    }
	    break;
	case PHOTOMETRIC_PALETTE:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = sample;
	    }
	    break;
	case PHOTOMETRIC_RGB:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = sample;
		NEXTSAMPLE
		*outp = sample;
		NEXTSAMPLE
		*outp = sample;
		if ( samplesPerPixel == 4) {
		    NEXTSAMPLE
		}
	    }
	    break;
	default:
	    break;
	}
	write( fdOut, outbuf, outbuflen);
    }
    xmax = xmin + xspacing*cols;
    ymin = ymax - yspacing*rows;
    fprintf( hdrFp, "filename:	%s\n", AmGetRootName( dataPath));
    if ( photometric == PHOTOMETRIC_PALETTE) {
	fprintf( hdrFp, "palette:	%s\n", AmGetRootName( palettePath));
    } else if ( greyscale) {
	fprintf( hdrFp, "colormap:	Grey-Scale\n");
	fprintf( hdrFp, "numcolors:	%d\n", lastNonZeroMarker);
    }
    fprintf( hdrFp, "type:	imagePlot\n");
    fprintf( hdrFp, "rows:	%d\n", rows);
    fprintf( hdrFp, "cols:	%d\n", cols);
    if ( photometric == PHOTOMETRIC_RGB) {
	fprintf( hdrFp, "size:		%d\n", sizeof( unsigned long));
	fprintf( hdrFp, "dataType:	unsignedLong\n");
	fprintf( hdrFp, "description:	%s\n", ulongDescriptionFile);
	fprintf( hdrFp, "trueColor:	true\n");
    } else {
	fprintf( hdrFp, "size:		%d\n", sizeof( uchar_t));
	fprintf( hdrFp, "dataType:	unsignedChar\n");
	fprintf( hdrFp, "colorsType:	unsignedChar\n");
	fprintf( hdrFp, "description:	%s\n", ucharDescriptionFile);
    }
    fprintf( hdrFp, "xmin:		%.2f\n", xmin);
    fprintf( hdrFp, "ymin:		%.2f\n", ymin);
    fprintf( hdrFp, "xmax:		%.2f\n", xmax);
    fprintf( hdrFp, "ymax:		%.2f\n", ymax);
    fprintf( hdrFp, "xunits:		meter\n");
    fprintf( hdrFp, "xvariable:		distance\n");
    fprintf( hdrFp, "yunits:		meter\n");
    fprintf( hdrFp, "yvariable:		distance\n");
    fprintf( hdrFp, "fileSource:	DRG\n");
    fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    close( fdOut);
}
