#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <anvil/project.h>
#include <anvil/files.h>
#if defined (_WIN32)
#include <anvil/ntresource.h>
#else
#include <anvil/xresource.h>
#endif
#include <anvil/datafmt.h>
#include <anvil/imageP.h>
#include <anvil/geometry.h>
#include <anvil/datatype.h>
#include <anvil/macros.h>
#include <anvil/fns.h>
#include <anvil/tagstrs.h>
#include <anvil/files.h>
#include <anvil/envdefs.h>
#include <anvil/choppath.h>
#include <util/utils.h>

/* program for concatenating together source files.  This version is 
   faster than "concat", but doesn't handle overlapping image files */

const int MAXFILES = 2048,
	  BUFROWS = 128,
	  POLYBUFFERSIZE = 1048576;

typedef struct fileInfo_ {
    char		*header;	/* header filename */
    char		*file;		/* filename */
#if defined(_WIN32)
    AmNTResourceDbClass	*db;		/* file header resource file */
#else
    AmXResourceDbClass	*db;		/* file header resource file */
#endif
    float		xmin;		/* file extents */
    float		ymin;
    float		xmax;
    float		ymax;
    float		xres;
    float		yres;
    int			startX;
    int			startY;
    int			endX;
    int			endY;
    int			rows;		/* file rows and columns */
    int			cols;
    int			fd;		/* file descriptor (may be -1) */
    AnvilBoolean		pipeFlag;	/* pipe or file */
    AmDataTypeEnum        dataType;       /* data element type */
    AmAccessFn            fn;             /* function to access data */
    int                 *xMap;          /* x, y mapping vectors */
    int                 xMapLen;
    int                 *yMap;
    int                 yMapLen;
    char                *buffer;        /* row buffer */
    int                 yIndex;         /* current y index */
} fileInfoRec, *fileInfoPtr;


/*************************************************************************
 * 
 * Description:	
 *	comparator function for sorting file information structures 
 *	by ymin value
 * Arguments:
 *	a,b - file information structures.
 * Returns:
 *	1 if a > b, -1 if a < b, 0 if a == b
 *
 *************************************************************************/

static int 
byYmin( fileInfoPtr  a,
        fileInfoPtr  b)
{
    if ( a->ymin > b->ymin) {
	return( 1);
    } else if ( a->ymin < b->ymin) {
	return( -1);
    } else {
	return( 0);
    }
}

/****************************************************************************
 * 
 * Description:	
 *	do the extents of the two files overlap?
 * Arguments:
 *	a, b - file info structures to compare the extents of
 * Returns:
 *	TRUE if they overlap, FALSE otherwise
 *
 ****************************************************************************/

static AnvilBoolean 
overlapFileInfo( fileInfoPtr	a,
		 fileInfoPtr	b)
{
    return(( AmInsideFloat( a->xmin, b->xmin, b->xmax) &&
	     AmInsideFloat( a->ymin, b->ymin, b->ymax)) ||
	   ( AmInsideFloat( a->xmax, b->xmin, b->xmax) &&
	     AmInsideFloat( a->ymax, b->ymin, b->ymax)));
}

/****************************************************************************
 * 
 * Description:	
 *	do the extents of any of the files in a set of files overlap?
 * Arguments:
 *	fileInfo - vector of file info structs to test.
 *	length - number of structures 
 * Returns:
 *	TRUE if any of the files overlap, FALSE otherwise.
 *
 *****************************************************************************/

static AnvilBoolean
anyOverlapFileInfo( fileInfoPtr	fileInfo,
		    int		length)
{
    int	ii, jj;

    for ( ii = 0; ii < length - 1; ii++) {
	for ( jj = ii + 1; jj < length; jj++) {
	    if ( overlapFileInfo( &fileInfo[ ii], &fileInfo[ jj])) {
		return( TRUE);
	    }
	}
    }
    return( FALSE);
}

/****************************************************************************
 * 
 * Description:
 *	get the extents of a set of files defined by a vector of
 *	fileInfo structures.  Also check to make sure the units match
 *	properly.
 * Arguments:
 * 	fileInfo - vector of fileInfo structures.
 *	length - number of files to concatenate 
 *	xmin, ymin, xmax, ymax - returned data extents.
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT if any of the files are bad.
 *
 ****************************************************************************/

AmErrorCode 
fileInfoExtents( fileInfoPtr	fileInfo,
		 int		length,
		 float		*xmin, 
		 float		*ymin, 
		 float		*xmax, 
		 float		*ymax)
{
    int	ii;
    char	*xunits, *yunits, *allxunits=NULL, *allyunits=NULL;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0; ii < length; ii++) {

/* get the files' extents, and expand the overall extents */
	
	if ( !fileInfo[ ii].db->GetResource( AmNxmin, &fileInfo[ ii].xmin)) {
	    fprintf( stderr, "failed to get xmin from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNymin, &fileInfo[ ii].ymin)) {
	    fprintf( stderr, "failed to get ymin from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNxmax, &fileInfo[ ii].xmax)) {
	    fprintf( stderr, "failed to get xmax from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNymax, &fileInfo[ ii].ymax)) {
	    fprintf( stderr, "failed to get ymax from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( fileInfo[ ii].xmin < *xmin) {
	    *xmin = fileInfo[ ii].xmin;
	}
	if ( fileInfo[ ii].ymin < *ymin) {
	    *ymin = fileInfo[ ii].ymin;
	}
	if ( fileInfo[ ii].xmax > *xmax) {
	    *xmax = fileInfo[ ii].xmax;
	}
	if ( fileInfo[ ii].ymax > *ymax) {
	    *ymax = fileInfo[ ii].ymax;
	}

/* check to make sure the units match while we're at it. */

	if ( !fileInfo[ ii].db->GetResource( AmNxunits, &xunits)) {
	    fprintf( stderr, "failed to get xunits from file %s\n",
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNyunits, &yunits)) {
	    fprintf( stderr, "failed to get yunits from file %s\n",
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( ii == 0) {
	    allxunits = xunits;
	    allyunits = yunits;
	} else {
	    if ( strcmp( xunits, allxunits) || strcmp( yunits, allyunits)) {
		fprintf( stderr, "units don't match for file %s\n",
			 fileInfo[ ii].header);
		DO_RETURN_ERROR( AmBAD_ARGUMENT);
	    }
	}
    }
    return( AmOK);
}

/****************************************************************************
 * 
 * Description:
 *	concatenate a series of scatter or line plot files, defined by the
 *	fileInfo structures.
 * Arguments:
 * 	fileInfo - vector of fileInfo structures.
 *	length - number of files to concatenate 
 *	outFd - output file descriptor 
 *	rows - returned data rows.
 *	size - returned data size
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT if any of the files are bad,
 *	or system error
 *
 ****************************************************************************/

AmErrorCode
concatItemFiles( fileInfoPtr	fileInfo, 
		 int		length, 
		 int		outFd,
		 int		*rows,
		 int		*size)
{
    int		ii, fd, firstSize=0;
    uchar_t	*newbuf;

/* read each of the source files in blocks, and write it out as blocks */

    *rows = 0;
    for ( ii = 0; ii < length; ii++) {
	AmPrintHatch( ii, length, NHATCH);

/* open the file */

	if (( fd = AmOpenReadFile( fileInfo[ ii].file, 
				 &fileInfo[ ii].pipeFlag)) < 0) {
	    fprintf( stderr, "failed to open input file %s\n",
		     fileInfo[ ii].file);
	    DO_RETURN_ERROR( AmOPEN_FAILED);
	}

/* get the rows and size */

	if ( !fileInfo[ ii].db->GetResource( AmNrows, 
					     &fileInfo[ ii].rows)) {
	    fprintf( stderr, "failed to get rows from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNsize, size)) {
	    fprintf( stderr, "failed to get size from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}

/* make sure all the sizes are the same */

	if ( ii == 0) {
	    firstSize = *size;
	} else if ( firstSize != *size) {
	    fprintf( stderr, "size %d in file %s different from size %d\n",
		     *size, fileInfo[ ii].header, firstSize);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}

/* read the scatter data file by file */

	newbuf = ( uchar_t *) malloc( *size*fileInfo[ ii].rows);
	if ( newbuf == NULL) {
	    fprintf( stderr, "failed to malloc %d bytes!\n", 
		     *size*fileInfo[ ii].rows);
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if ( AmReadLoopFile( fd, newbuf, *size*fileInfo[ ii].rows) < 
	     *size*fileInfo[ ii].rows) {
	    fprintf( stderr, "failed to read points from %s\n",
		     fileInfo[ ii].file);
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	write( outFd, newbuf, *size*fileInfo[ ii].rows);
	free( newbuf);
	*rows += fileInfo[ ii].rows;
	AmCloseReadFile( fd, fileInfo[ ii].pipeFlag);
    }
    close( outFd);
    return( AmOK);
}

/****************************************************************************
 * 
 * Description:
 *	concatenate a series of polygon files, defined by fileInfo structures.
 * Arguments:
 * 	fileInfo - vector of fileInfo structures.
 *	length - number of files to concatenate 
 *	outFd - output file descriptor 
 *	rows - returned data rows.
 *	size - returned data size
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT if any of the files are bad,
 *	or system error
 *
 ****************************************************************************/

AmErrorCode
concatPolygonFiles( fileInfoPtr	fileInfo, 
		    int		length, 
		    int		outFd,
		    int		*rows,
		    int		*size)
{
    int		fd, ii, leftOver, numBytes, polySize;
    uchar_t	*polyBuffer, *polyBufferPtr, *polyBufferEnd;
    AmPolygonPtr	polygon;

    *rows = 0;

/* find out the number of polygons in each file and their size */

    for ( ii = 0; ii < length; ii++) {	
	if ( !fileInfo[ ii].db->GetResource( AmNrows, 
					     &fileInfo[ ii].rows)) {
	    fprintf( stderr, "failed to get rows from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNsize, size)) {
	    fprintf( stderr, "failed to get size from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	*rows += fileInfo[ ii].rows;
    }

/* allocate a buffer to read the polygons into... */

    polyBuffer = ( uchar_t *) alloca( POLYBUFFERSIZE);

/* loop over each file and read it in blocks */

    for ( ii = 0; ii < length; ii++) {
	if (( fd = AmOpenReadFile( fileInfo[ ii].file, 
				 &fileInfo[ ii].pipeFlag)) < 0) {
	    fprintf( stderr, "failed to open input file %s\n",
		     fileInfo[ ii].file);
	    DO_RETURN_ERROR( AmOPEN_FAILED);
	}

/* read the polygons followed by their points and write them out again */

	leftOver = 0;
	do {
	    numBytes = AmReadLoopFile( fd, polyBuffer + leftOver, 
				     POLYBUFFERSIZE - leftOver);
	    if ( numBytes <= 0) {
		break;
	    }
	    polyBufferPtr = polyBuffer;
	    polyBufferEnd = polyBuffer + numBytes + leftOver;
	    do {
		polygon = ( AmPolygonPtr) polyBufferPtr;
		polySize = polygon->numPts*sizeof( AmCoordRec) + *size;

/* did we read in the entire polygon? */

		if (((( uchar_t *) polygon) + *size < polyBufferEnd) &&
		    ((( uchar_t *) polygon) + polySize <= polyBufferEnd)) {

/* track the  new extents of the data */

		    polyBufferPtr += polySize;
		} else {

/* write out the polygon buffer, and take the incomplete polygon and put it
   at the head of the buffer */

		    write( outFd, polyBuffer, polyBufferPtr - polyBuffer);
		    leftOver = polyBufferEnd - polyBufferPtr;
		    memcpy( polyBuffer, polyBufferPtr, leftOver);
		    break;
		}
	    } while ( TRUE);
	} while ( TRUE);
	AmCloseReadFile( fd, fileInfo[ ii].pipeFlag);
    }
    close( outFd);
    return( AmOK);
}

/****************************************************************************
 * 
 * Description:
 *	concatenate a set of non-overlapping image files, defined by
 *	fileInfo structures.
 * Arguments:
 * 	fileInfo - vector of fileInfo structures.
 *	length - number of files to concatenate 
 *	outFd - output file descriptor 
 *	xmin, ymin, xmax, ymax - extents of the concatented file.
 *	rows, cols - returned data rows, cols.
 *	size - returned data size
 * Returns:
 *	AmOK on success, AmBAD_ARGUMENT if any of the files are bad,
 *	or system error
 *
 ****************************************************************************/

AmErrorCode
concatNonOverlapImageFiles( fileInfoPtr	fileInfo, 
			    int		length, 
			    int		outFd,
			    float	xmin,
			    float	ymin,
			    float	xmax,
			    float	ymax,
			    int		*rows,
			    int		*cols,
			    int		*size)
{
    int		ii, jj, mm, dstIndex, numrows, buflen, rowlen;
    float	xres, yres;
    uchar_t	*readRow, *writeRow;

/* find the combined extents of all the images, and the highest resolution
   image */

    for ( ii = 0; ii < length; ii++) {
	if ( !fileInfo[ ii].db->GetResource( AmNrows, &fileInfo[ ii].rows)) {
	    fprintf( stderr, "failed to get rows from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNcols, &fileInfo[ ii].cols)) {
	    fprintf( stderr, "failed to get cols from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNsize, size)) {
	    fprintf( stderr, "failed to get size from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	fileInfo[ ii].fd = 0;
    }
    xres = ( fileInfo[ 0].xmax - fileInfo[ 0].xmin)/( fileInfo[ 0].cols - 1);
    yres = ( fileInfo[ 0].ymax - fileInfo[ 0].ymin)/( fileInfo[ 0].rows - 1);

/* add one for centerpixel offset */

    *rows = ( int) (( ymax - ymin)/xres) + 1;
    *cols = ( int) (( xmax - xmin)/yres) + 1;
AN_DEBUG( 50, ("xmin=%f, ymin=%f, xmax=%f, ymax=%f, rows=%d, cols=%d\n",
	    xmin, ymin, xmax, ymax, *rows, *cols));

/* sort the matrices by y min */

    qsort( fileInfo, length, sizeof( fileInfoRec), ( AmSearchFn) byYmin);

/* allocate read buffers for each source matrix, and set the element size */ 

    for ( ii = 0; ii < length; ii++) {
	fileInfo[ ii].startX = (int)(*cols*( fileInfo[ ii].xmin - xmin)/( xmax - xmin));
	fileInfo[ ii].endX = (int)(*cols*( fileInfo[ ii].xmax - xmin)/( xmax - xmin));
	fileInfo[ ii].startY = (int)(*rows*( ymax - fileInfo[ ii].ymax)/( ymax - ymin));
	fileInfo[ ii].endY = (int)(*rows*( ymax - fileInfo[ ii].ymin)/( ymax - ymin));
AN_DEBUG( 75, ("%s = (%.2f, %.2f, %.2f, %.2f) = (%d, %d, %d, %d)\n",
	   fileInfo[ ii].file, 
	   fileInfo[ ii].xmin, fileInfo[ ii].ymin, 
	   fileInfo[ ii].xmax, fileInfo[ ii].ymax, 
	   fileInfo[ ii].startX, fileInfo[ ii].startY, 
	   fileInfo[ ii].endX, fileInfo[ ii].endY));
    }

/* allocate a row for the write matrix */

    writeRow = ( uchar_t *) alloca( BUFROWS*( *cols)*( *size));
    if ( writeRow == NULL) {
	fprintf( stderr, "failed to allocate %d bytes\n", 
		 BUFROWS*( *cols)*( *size));
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    readRow = ( uchar_t *) alloca( BUFROWS*( *cols)*( *size));
    if ( readRow == NULL) {
	fprintf( stderr, "failed to allocate %d bytes\n", 
		 BUFROWS*( *cols)*( *size));
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
	    
/* step by the new resolution. clear out the write buffer each row */

    numrows = BUFROWS;
    for ( mm = 0; mm < *rows; mm += numrows) {
	AmPrintHatch( mm/BUFROWS, ( *rows)/BUFROWS, NHATCH);
	memset( writeRow, 0, ( *cols)*( *size)*BUFROWS);
	for ( ii = 0; ii < length; ii++) {

/* open files that should be opened, close files that should be closed */

	    if (( mm >= fileInfo[ ii].startY) && ( mm < fileInfo[ ii].endY)) {
		if ( fileInfo[ ii].fd == 0) {
		    fileInfo[ ii].fd = AmOpenReadFile( fileInfo[ ii].file, 
						     &fileInfo[ ii].pipeFlag);
		}
		if ( fileInfo[ ii].fd < 0) {
		    fprintf( stderr, "failed to open file %s\n",
			     fileInfo[ ii].file);
		    DO_RETURN_ERROR( AmOPEN_FAILED);
		}

/* read the buffer if we're at the start row, or if the mapped row index
   has changed */

		numrows = MIN( BUFROWS, fileInfo[ ii].endY - mm);
		rowlen = fileInfo[ ii].cols*( *size);
		buflen = rowlen*numrows;
		if ( AmReadLoopFile( fileInfo[ ii].fd, readRow, buflen) < buflen) {
		    fprintf( stderr, "failed to read from file %s at %d\n",
			     fileInfo[ ii].file, mm);
		}

/* map the buffer into the write row. If there's an accessing function, then
   use it to only write the value if it's higher than the current value */

		dstIndex = fileInfo[ ii].startX*( *size);
		for ( jj = 0; jj < numrows; jj++) {
		    memcpy( &writeRow[ dstIndex + jj*(*cols)*( *size)],
			    &readRow[ jj*rowlen], rowlen);
		}
	    }
	    if (( mm > fileInfo[ ii].endY) && ( fileInfo[ ii].fd != 0)) {
		AmCloseReadFile( fileInfo[ ii].fd, fileInfo[ ii].pipeFlag);
		fileInfo[ ii].fd = 0;
	    }
	}
	write( outFd, writeRow, ( *cols)*( *size)*numrows);
    }
    close( outFd);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *      concatenate a set of overlapping image files, defined by
 *      fileInfo structures.
 * Arguments:
 *      fileInfo - vector of fileInfo structures.
 *      length - number of files to concatenate
 *      outFd - output file descriptor
 *      xmin, ymin, xmax, ymax - extents of the concatented file.
 *      rows, cols - returned data rows, cols.
 *      size - returned data size
 * Returns:
 *      AmOK on success, AmBAD_ARGUMENT if any of the files are bad,
 *      or system error
 *
 ****************************************************************************/

AmErrorCode
concatOverlapImageFiles( fileInfoPtr fileInfo,
			 int         length,
			 int         outFd,
			 float       xmin,
			 float       ymin,
			 float       xmax,
			 float       ymax,
			 int         *rows,
			 int         *cols,
			 int         *size)
{
    int		ii, jj, mm, buflen, start, dstIndex, srcIndex;
    float	yval, a, b, minres;
    char	*dataTypeStr;
    uchar_t	*writeRow;

/* find the combined extents of all the images, and the highest resolution
image */

    minres = MY_HUGE;

    for ( ii = 0; ii < length; ii++) {
	if ( !fileInfo[ ii].db->GetResource( AmNrows, &fileInfo[ ii].rows)) {
	    fprintf( stderr, "failed to get rows from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNcols, &fileInfo[ ii].cols)) {
	    fprintf( stderr, "failed to get cols from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNsize, size)) {
	    fprintf( stderr, "failed to get size from file %s\n", 
		     fileInfo[ ii].header);
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	}
	if ( !fileInfo[ ii].db->GetResource( AmNdataType, &dataTypeStr)) {
	    fprintf( stderr, "failed to get data type from file %s\n", 
		     fileInfo[ ii].header);
	    fileInfo[ ii].fn = NULL;
	    DO_RETURN_ERROR( AmBAD_ARGUMENT);
	} else {
	    fileInfo[ ii].dataType = AmStrToDataTypeEnum( dataTypeStr);
	    fileInfo[ ii].fn = AmDataTypeEnumToAccessFn( fileInfo[ ii].dataType);
	}

/* x,y is in center of pixel.. */

	fileInfo[ ii].xres = ( fileInfo[ ii].xmax - fileInfo[ ii].xmin)/
				( fileInfo[ ii].cols - 1);
	if ( fileInfo[ ii].xres < minres) {
	    minres = fileInfo[ ii].xres;
	}
	fileInfo[ ii].yres = ( fileInfo[ ii].ymax - fileInfo[ ii].ymin)/
				( fileInfo[ ii].rows - 1);
	if ( fileInfo[ ii].yres < minres) {
	    minres = fileInfo[ ii].yres;
	}
	fileInfo[ ii].fd = 0;
    }

/* add one for centerpixel offset */

    *rows = ( int) (( ymax - ymin)/minres) + 1;
    *cols = ( int) (( xmax - xmin)/minres) + 1;

    AN_DEBUG (3,("  x-axis : %.2f -> %.2f \n", xmin, xmax ));
    AN_DEBUG (3,("  y-axis : %.2f -> %.2f \n", ymin, ymax ));
    AN_DEBUG (3,("  resolution : %.2f x %.2f \n", minres, minres ));
    AN_DEBUG (3,("  --> output matrix : %d x %d \n", *rows, *cols ));

/* sort the matrices by y min */

    qsort( fileInfo, length, sizeof( fileInfoRec), ( AmSearchFn) byYmin);

/* allocate read buffers for each source matrix, and 
   generate horizontal and vertical mapping vectors for each file */

    for ( ii = 0; ii < length; ii++) {
	fileInfo[ ii].buffer = ( char *) malloc( fileInfo[ ii].cols*( *size));
	if ( fileInfo[ ii].buffer == NULL) {
	    fprintf( stderr, "failed to allocate %d bytes\n",
		     fileInfo[ ii].cols*( *size));
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	fileInfo[ ii].yIndex = 0;
	fileInfo[ ii].xMapLen = 
	    ( int) ( fileInfo[ ii].cols*fileInfo[ ii].xres/minres);
	fileInfo[ ii].xMap = 
	    AmMakeIndexVec( fileInfo[ ii].cols, fileInfo[ ii].xMapLen);
	fileInfo[ ii].yMapLen = 
	    ( int) ( fileInfo[ ii].rows*fileInfo[ ii].yres/minres);
	fileInfo[ ii].yMap = 
	    AmMakeIndexVec( fileInfo[ ii].rows, fileInfo[ ii].yMapLen);
    }

/* allocate a row for the write matrix */

    writeRow = ( uchar_t *) malloc(( *cols)*( *size));
    if ( writeRow == NULL) {
	fprintf( stderr, "failed to malloc %d bytes\n", *cols*( *size));
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* step by the new resolution. clear out the write buffer each row */

    AN_DEBUG (3,(" looping over %d rows ... \n", *rows ));
    for ( yval = ymax - minres, mm = 0; mm < *rows; mm++, yval -= minres) {
	AmPrintHatch( mm, *rows, NHATCH);
	memset( writeRow, 0, ( *cols)*( *size));

	for ( ii = 0; ii < length; ii++) {

/* open files that should be opened, close files that should be closed */

	    if (( yval >= fileInfo[ ii].ymin) &&
		( yval <= fileInfo[ ii].ymax) && 
		( fileInfo[ ii].fd == 0)) {
		fileInfo[ ii].fd = 
		    AmOpenReadFile( fileInfo[ ii].file, &fileInfo[ ii].pipeFlag);
		if ( fileInfo[ ii].fd < 0) {
		    fprintf( stderr, "failed to open file %s\n",
			     fileInfo[ ii].file);
		    DO_RETURN_ERROR( AmOPEN_FAILED);
		}
	    }

	    if (( yval < fileInfo[ ii].ymin) && ( fileInfo[ ii].fd != 0)) {
		AmCloseReadFile( fileInfo[ ii].fd, fileInfo[ ii].pipeFlag);
		fileInfo[ ii].fd = 0;
	    }

	    if (( yval >= fileInfo[ ii].ymin) && ( yval <= fileInfo[ ii].ymax)) {

/* read the buffer if we're at the start row, or if the mapped row index
has changed */

		if (( fileInfo[ ii].yIndex == 0) ||
		    ( fileInfo[ ii].yMap[ fileInfo[ ii].yIndex] !=
		      fileInfo[ ii].yMap[ fileInfo[ ii].yIndex - 1])){
		    buflen = fileInfo[ ii].cols*( *size);
		    if ( AmReadLoopFile( fileInfo[ ii].fd, 
				       fileInfo[ ii].buffer, buflen) < buflen) {
			fprintf( stderr, "failed to read from file %s ",
				 fileInfo[ ii].file);
			fprintf( stderr, "at ymin = %f\n", yval);
		    }
		}

/* map the buffer into the write row. If there's an accessing function, then
use it to only write the value if it's higher than the current value */

		start = ( int) (( *cols)*( fileInfo[ ii].xmin - xmin)/
					    ( xmax - xmin));
		for ( jj = 0; jj < fileInfo[ ii].xMapLen; jj++) {
		    dstIndex = (start + jj)*( *size);
		    srcIndex = fileInfo[ ii].xMap[ jj]*( *size);

		    if ( fileInfo[ ii].fn != NULL) {
			a = ( fileInfo[ ii].fn)
				( &fileInfo[ ii].buffer[ srcIndex], NULL);
			b = ( fileInfo[ ii].fn)
				( &writeRow[ dstIndex], NULL);
			if ( a > b) {
			    memcpy(( void *) &writeRow[ dstIndex],
				   ( void *) &fileInfo[ ii].buffer[ srcIndex],
				   ( *size));
			}
		    } else {
			memcpy(( void *) &writeRow[ dstIndex],
			       ( void *) &fileInfo[ ii].buffer[ srcIndex], ( *size));
		    }
		}

/* increment the y index for the source file */

		fileInfo[ ii].yIndex++;
	    }
	}
	write( outFd, writeRow, ( *cols)*( *size));
    }
    close( outFd);
    return( AmOK);
}

/************************************************************************
 * 
 * Description:
 * 	program for concatenating non-overlapping files.  This takes
 *	an ASCII list of file headers and data files, with the 
 *	header and data filenames on each row.  The output file is the
 *	files concatenated together.  If the files are image files, then
 *	'0' is written into all parts of the file that the source files
 *	aren't written into.
 * Arguments:
 *	concat file-and-header-list-file output-file
 *
 ************************************************************************/

int 
main( int argc, char *argv[])
{
    FILE		*listFp, *hdrFp;
    int			outFd, ii, length, rows, cols, size;
    fileInfoRec		fileInfo[ MAXFILES];
    char		headerBuf[ PATHLEN], fileBuf[ PATHLEN], 
   			*kind, *otherKind, header[ PATHLEN], outfile[ PATHLEN];
    float		xmin, ymin, xmax, ymax;
    char *fields[] = { AmNfilename, AmNrows, AmNcols, AmNsize, AmNtype, 
		       AmNnumdims, AmNxmin, AmNymin, AmNxmax, AmNymax, NULL };

    if ( argc < 3) {
	fprintf( stderr, "usage: %s file-and-header-list-file outfile\n", 
		 argv[ 0]);
	exit( -1);
    }
    if (( listFp = fopen( argv[ 1], "r")) == NULL) {
	fprintf( stderr, "failed to open list file %s\n", argv[1]);
	exit( -1);
    }
    if ( getenv( CONFIG_ENV) == NULL) {
	fprintf( stderr, "environment variable %s not defined", CONFIG_ENV);
	exit( -1);
    }
    AmInitUnits();
    length = AmNumLines( argv[ 1]);

/* open a path to the header file and data file */

    sprintf( header, "%s.hdr", argv[ 2]);
    if (( hdrFp = fopen( header, "w")) == NULL) {
	fprintf( stderr, "failed to write header file %s\n", header);
	exit( -1);
    }
    sprintf( outfile, "%s.dat", argv[ 2]);
    if (( outFd = open( outfile, O_WRONLY | O_CREAT, PERMS)) < 0){
	fprintf( stderr, "failed to open output file %s\n", outfile);
	exit( -1);
    }

/* read the headers and files */

    for ( ii = 0; ii < length; ii++) {
	fscanf( listFp, "%s %s", headerBuf, fileBuf);
	fileInfo[ ii].header = strdup( headerBuf);
	fileInfo[ ii].file = strdup( fileBuf);
	fileInfo[ ii].fd = 0;
    }
    fclose( listFp);

/* read the resource databases from the headers */

    for ( ii = 0; ii < length; ii++) {
#if defined(_WIN32)
	fileInfo[ ii].db = new AmNTResourceDbClass( fileInfo[ ii].header);
#else
	fileInfo[ ii].db = new AmXResourceDbClass( fileInfo[ ii].header);
#endif
    }

/* find out what kind of files we're dealing with, and if they're all the
   same type */

    if ( !fileInfo[ 0].db->GetResource( AmNtype, &kind)) {	
	fprintf( stderr, "failed to read file type from %s\n", 
		 fileInfo[ 0].header);
	exit( -1);
    }
    for ( ii = 1; ii < length; ii++) {
	if ( !fileInfo[ ii].db->GetResource( AmNtype, &otherKind)) {	
	    fprintf( stderr, "failed to read file type from %s\n", 
		     fileInfo[ ii].header);
	    exit( -1);
	}
	if ( strcmp( kind, otherKind)) {
	    fprintf( stderr, "different file type %s != %s, file:%s\n", 
		     kind, otherKind, fileInfo[ ii].header);
	    exit( -1);
	}
    }

/* get the extents of all the files */

    if ( fileInfoExtents( fileInfo, length,
			  &xmin, &ymin, &xmax, &ymax) != AmOK) {
	fprintf( stderr, "failed to get file extents\n");
	exit( -1);
    }
    if ( !strcmp( kind, AmNimagePlot)) {
        if ( anyOverlapFileInfo( fileInfo, length)) {
	    printf( "image files overlap\n");
	    if ( concatOverlapImageFiles( fileInfo, length, outFd,
					  xmin, ymin, xmax, ymax,
					  &rows, &cols, &size) != AmOK) {
		fprintf( stderr, "failed to concatenate image files\n");
		exit( -1);
	    }
	} else {
	    printf( "image files don't overlap\n");
	    if ( concatNonOverlapImageFiles( fileInfo, length, outFd, 
					     xmin, ymin, xmax, ymax,
					     &rows, &cols, &size) != AmOK) {
		fprintf( stderr, "failed to concatenate image files\n");
		exit( -1);
	    }
	}

/* create the header file which describes the concatenated output */
	
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( outfile));
	fprintf( hdrFp, "xmin:		%f\n", xmin);
	fprintf( hdrFp, "ymin:		%f\n", ymin);
	fprintf( hdrFp, "xmax:		%f\n", xmax);
	fprintf( hdrFp, "ymax:		%f\n", ymax);
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "cols:		%d\n", cols);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		%s\n", AmNimagePlot);
    } else if ( !strcmp( kind, AmNpolygonPlot)) {
	if ( concatPolygonFiles( fileInfo, length, outFd, &rows, &size) != AmOK) {
	    fprintf( stderr, "failed to concatenate files\n");
	    exit( -1);
	}
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( outfile));
	fprintf( hdrFp, "xmin:		%f\n", xmin);
	fprintf( hdrFp, "ymin:		%f\n", ymin);
	fprintf( hdrFp, "xmax:		%f\n", xmax);
	fprintf( hdrFp, "ymax:		%f\n", ymax);
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		%s\n", AmNpolygonPlot);
    } else if ( !strcmp( kind, AmNscatterPlot) ||
		!strcmp( kind, AmNlinePlot)) {
	if ( concatItemFiles( fileInfo, length, outFd, &rows, &size) != AmOK) {
	    fprintf( stderr, "failed to concatenate files\n");
	    exit( -1);
	}
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( outfile));
	fprintf( hdrFp, "xmin:		%f\n", xmin);
	fprintf( hdrFp, "ymin:		%f\n", ymin);
	fprintf( hdrFp, "xmax:		%f\n", xmax);
	fprintf( hdrFp, "ymax:		%f\n", ymax);
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		%s\n", kind);
    } else {
	fprintf( stderr, "unrecognized file type %s\n", kind);
	exit( -1);
    }

/* take all the fields from the AmFirst header which don't get changed and
   dump them to the new header */

    AmDumpHeaderExcept( fileInfo[ 0].header, fields, hdrFp);
    fclose( hdrFp);
    close( outFd);
}
