#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */





/* program to convert USGS landuse/landcover files to polygon format output */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdlib.h>
#include <anvil/endian.h>
#include <anvil/project.h>
#include <anvil/envdefs.h>
#include <anvil/strlib.h>
#include <anvil/macros.h>
#include <anvil/geocodes.h>
#include <anvil/projects.h>
#include <anvil/choppath.h>
#include <anvil/files.h>
#include <anvil/raspoly.h>
#include <util/utils.h>
#include <util/landuse.h>

const char *descriptionFile = 
	"$ANVIL_CONFIG/descriptions/landuse_polygon_format";

/* for sorting vectors by arc number. */

int 
byArcId( larcPtr  a,
	 larcPtr  b)
{
    return( b->arcNo - a->arcNo);
}

/* for sorting polygons by polygon number */

int 
byPolyId( lpolyPtr  a,
	  lpolyPtr  b)
{
    return( b->polyNo - a->polyNo);
}

/* read the attribute codes from a landuse mapping file */
 
AmErrorCode 
readLanduseAttrCodes( char         		*attrCodeFile,
		      int          		*numCodes,
		      landuseAttrCodePtr  	*attrCodes)
{
    FILE        *fp;
    int         ii;
    char	buffer[ 512];

    if (( fp = fopen( attrCodeFile, "r")) == NULL) {
        DO_RETURN_ERROR( AmOPEN_FAILED);
    }
    if ( fscanf( fp, "%d\n", numCodes) != 1) {
	fclose( fp);
	DO_RETURN_ERROR( AmREAD_FAILED);
    }
    *attrCodes = ( landuseAttrCodePtr) 
	malloc( sizeof( landuseAttrCodeRec)*( *numCodes));
    if ( *attrCodes == NULL) {
	fclose( fp);
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

/* let's scan softly into a buffer, then scan the values */

    for ( ii = 0; ii < *numCodes; ii++) {
	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
            fprintf( stderr, "failed to read attribute code %d\n", ii);
	    fclose( fp);
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
        if ( sscanf( buffer, "%d %d",
                     &( *attrCodes)[ ii].code, &( *attrCodes)[ ii].attr) != 2) {
            fprintf( stderr, "failed to scan attribute code %d\n", ii);
	    fclose( fp);
	    DO_RETURN_ERROR( AmREAD_FAILED);
        }
    }
    fclose( fp);
    return( AmOK);
}

/* copy a landuse polygon and all of its islands to the polygon vector */

void copyPolygonIslands( PJ		*pj,
			 lpolyPtr	polyPtr,
		         lulcPolyPtr	polyVec,
		         int		*index,
			 float		swLat,
			 float		swLon,
			 AnvilBoolean	AmFirst)
{
    int	ii;
    UV	uv;

    if (( polyPtr->surroundPolyNo == 0) || !AmFirst) {
 
/* copy the number of points, points and attribute from the polygon */
 
	polyVec[ *index].geoPoly.polygon.numPts = polyPtr->npts;
	polyVec[ *index].geoPoly.polygon.pts = ( AmCoordPtr)
	    malloc( sizeof( AmCoordRec)*polyPtr->npts);
	polyVec[ *index].geoPoly.attributeCode = polyPtr->attr;
	polyVec[ *index].origCode = polyPtr->origAttr;

/* re-align the points to the UTM start point and scale by 10 (according to
   the USGS landuse manual */

	for ( ii = 0; ii < polyPtr->npts; ii++) {
	    uv.u = swLon + polyPtr->pts[ ii].x*10.0;
	    uv.v = swLat + polyPtr->pts[ ii].y*10.0;
	    uv = pj_inv( uv, pj);
	    polyVec[ *index].geoPoly.polygon.pts[ ii].x = RADTODEG( uv.u);
	    polyVec[ *index].geoPoly.polygon.pts[ ii].y = RADTODEG( uv.v);
	}
	(*index)++;

/* call it recursively for each of the islands, and turn off the flag so
   they're actually copied */

	for ( ii = 0; ii < polyPtr->numIslands; ii++) {
	    copyPolygonIslands( pj, polyPtr->islands[ ii], polyVec, index, 
				swLat, swLon, FALSE);
	}
    }
}

/* read a GIRAS land use data file */

int readLanduseData( char		*landuseFile,
		     char		*attrCodeFile,
		     char		*outfile)
{
    lulcPolyPtr		landPolys;
    char		*projVec[ 4];
    mapHeaderRec	mapHeader;
    sectionHeaderRec	sectionHeader;
    arcRecordRec	arcRecord;
    coordRecordRec	coordRecord;
    polyRecordRec	polyRecord;
    fapRecordRec	fapRecord;
    landuseDataPtr	landusePtr;
    landuseSectionPtr	section;
    lpolyPtr		polyPtr, foundPoly;
    lpolyRec		targPoly;
    larcPtr		arcPtr;
    larcRec		targetArc;
    fpointPtr		coordPtr;
    ltextPtr		textPtr;
    short		*fapPtr;
    int			i, j, k, m, n, padSize, lastplc, npts, lastpla,
			narcs, start, index, intlat, intlon, deg, min, sec,
			SWctlPtX, SWctlPtY, attr, numCodes, zone, fd, fdOut,
    			numPolys, ii, jj;
    char		padBuf[ 80], buffer[ 512], *title,
			polyfile[ 512], polyhdr[ 512];
    float		swLat, swLon, xmin, ymin, xmax, ymax;
    UV			uv;
    PJ			*pj;
    FILE		*fpOut;
    landuseAttrCodePtr	attrCodes;
    AnvilBoolean		pipeFlag;

    if (( fd = AmOpenReadFile( landuseFile, &pipeFlag)) < 0) {
	fprintf( stderr, "failed to open landuse file %s\n", landuseFile);
	return( FALSE);
    }
/* try and get the attribute codes file for mapping to our  attributes */
 
    if ( readLanduseAttrCodes( attrCodeFile, &numCodes, &attrCodes) != AmOK) {
        fprintf( stderr, "failed to read attribute codes file\n");
        return( FALSE);
    }
    landusePtr = ( landuseDataPtr) malloc( sizeof( landuseDataRec));
    AmReadLoopFile( fd, &mapHeader, sizeof( mapHeader));

/* read the lat and lon values from the file, and scale them to 
   doubles/10000 the latitiude value isn't properly negated in the file */

    deg =  AmAtoi2( &mapHeader.latlonControlPts[ 0][ 3], 3);
    min =  AmAtoi2( &mapHeader.latlonControlPts[ 0][ 6], 2);
    sec =  AmAtoi2( &mapHeader.latlonControlPts[ 0][ 8], 2);
    intlat = deg*3600 + min*60 + sec;
    deg =  AmAtoi2( &mapHeader.latlonControlPts[ 1][ 3], 3);
    min =  AmAtoi2( &mapHeader.latlonControlPts[ 1][ 6], 2);
    sec =  AmAtoi2( &mapHeader.latlonControlPts[ 1][ 8], 2);
    intlon = -(deg*3600 + min*60 + sec);
    uv.u = (( float) intlon)/3600.0;
    uv.v = (( float) intlat)/3600.0;

/* get the projection function for this zone */

    projVec[ 0] = "proj=utm";
    projVec[ 2] = "ellps=WGS72";
    zone = (int)( 1 + ( 180 + uv.u)/6);
    sprintf( buffer, "zone=%d", zone);
    projVec[ 1] = buffer;
    if (( pj = pj_init( 3, ( const char **) projVec)) == NULL) {
        fprintf( stderr, "projection initialization failed\n");
        return( FALSE);
    }

/* now, apparently, there are also some control points in the header which
   affect the positioning in the map */

    SWctlPtX = 10*AmAtoi2( mapHeader.controlPts[ 0], 
			 sizeof( mapHeader.controlPts[ 0]));
    SWctlPtY = 10*AmAtoi2( mapHeader.controlPts[ 1], 
			 sizeof( mapHeader.controlPts[ 1]));

/* convert the latitude and longitude to utm coordinates */
     
    uv.u = DEGTORAD( uv.u);
    uv.v = DEGTORAD( uv.v);
    uv = pj_fwd( uv, pj);
    swLon = uv.u - SWctlPtX;
    swLat = uv.v - SWctlPtY;

/* scan the critical sections of the header */

    landusePtr->numSections = AmAtoi2( mapHeader.numMapSections, 
				       sizeof( mapHeader.numMapSections));
    landusePtr->sections = ( landuseSectionPtr) 
	malloc( sizeof( landuseSectionRec)*landusePtr->numSections);
    landusePtr->numTexts = AmAtoi2( mapHeader.textSubfileLen,
				    sizeof( mapHeader.textSubfileLen));
    textPtr = landusePtr->texts = 
	( ltextPtr) malloc( sizeof( ltextPtr)*landusePtr->numTexts);

/* read each section of the land-use file */

    for ( i = 0; i < landusePtr->numSections; i++) {
	section = landusePtr->sections + i;
	AmReadLoopFile( fd, &sectionHeader, sizeof( sectionHeader));

/* read the number of coordinates, arcs, faps, and polygons, and allocate
   the arrays */

	section->numCoords = AmAtoi2( sectionHeader.numArcCoords,
				     sizeof( sectionHeader.numArcCoords))/2;
	coordPtr = section->coords = 
	    ( fpointPtr) malloc( sizeof( fpointRec)*section->numCoords);
	section->numArcs = AmAtoi2( sectionHeader.numArcs,
				   sizeof( sectionHeader.numArcs));
	arcPtr = section->arcs = 
	    ( larcPtr) malloc( sizeof( larcRec)*section->numArcs);
	section->numPolys = AmAtoi2( sectionHeader.numPolys,
				    sizeof( sectionHeader.numPolys));
	polyPtr = section->lpolys = 
	    ( lpolyPtr) malloc( sizeof( lpolyRec)*section->numPolys);
	section->numFaps = AmAtoi2( sectionHeader.numFaps,
				   sizeof( sectionHeader.numFaps));
	fapPtr = section->faps = 
	    ( short *) malloc( sizeof( short)*section->numFaps);

/* read the arcs */

	lastplc = 0;
	for ( j = 0; j < section->numArcs; j++) {
	    AmReadLoopFile( fd, &arcRecord, sizeof( arcRecord));
	    arcPtr->arcNo = abs( AmAtoi2( arcRecord.arcNo,
					sizeof( arcRecord.arcNo)));
            arcPtr->plc = abs( AmAtoi2( arcRecord.lastCoordPos, 
				      sizeof( arcRecord.lastCoordPos)))/2;
            arcPtr->npts = arcPtr->plc - lastplc;
	    lastplc = arcPtr->plc;
	    arcPtr++;
	}

/* read the end to pad out to 80 columns */

	padSize = section->numArcs*sizeof( arcRecord) % 80;
	AmReadLoopFile( fd, padBuf, padSize);

/* read the coordinates */

	for ( j = 0; j < section->numCoords; j++) {
	    AmReadLoopFile( fd, &coordRecord, sizeof( coordRecord));
	    coordPtr->x = AmAtoi2( coordRecord.x, sizeof( coordRecord.x));
	    coordPtr->y = AmAtoi2( coordRecord.y, sizeof( coordRecord.y));
	    coordPtr++;
	}

/* read the end to pad out to 80 columns */

	padSize = section->numCoords*sizeof( coordRecord) % 80;
	if ( padSize != 0) {
	    padSize = 80 - padSize;
	    AmReadLoopFile( fd, padBuf, padSize);
	}

/* read the polygons */

	for ( j = 0; j < section->numPolys; j++) {
	    AmReadLoopFile( fd, &polyRecord, sizeof( polyRecord));
	    attr = AmAtoi2( &polyRecord.attr[ 8], 2);
	    polyPtr->origAttr = attr;

/* convert from landuse attributes to our private attributes */

	    for ( m = 0; m < numCodes; m++) {
		if ( attr == attrCodes[ m].code) {
		    polyPtr->attr = attrCodes[ m].attr;
		    break;
		}
	    }
	    if ( m == numCodes) {
		fprintf( stderr, "attribute code %d not found\n", attr);
	    }
            polyPtr->polyNo = abs( AmAtoi2( polyRecord.polyNo,
                                          sizeof( polyRecord.polyNo)));
            polyPtr->pla = abs( AmAtoi2( polyRecord.lastArcNo,
				       sizeof( polyRecord.lastArcNo))) - 1;
            polyPtr->surroundPolyNo = 
		abs( AmAtoi2( polyRecord.surroundPoly,
			    sizeof( polyRecord.surroundPoly)));
            polyPtr->numIslands = 0;
	    polyPtr++;
	}

/* read the end to pad out to 80 columns */

	padSize = section->numPolys*sizeof( polyRecord) % 80;
	if ( padSize != 0) {
	    padSize = 80 - padSize;
	    AmReadLoopFile( fd, padBuf, padSize);
	}

/* read the FAP ( file of arcs by polygon) subfile */

	fapPtr = section->faps = 
	    ( short *) malloc( sizeof( short)*section->numFaps);
	for ( j = 0; j < section->numFaps; j++) {
	    AmReadLoopFile( fd, &fapRecord, sizeof( fapRecordRec));
	    section->faps[ j] = 
		AmAtoi2( fapRecord.arcNo, sizeof( fapRecord.arcNo));
	}

/* read the end to pad out to 80 columns */

	padSize = section->numFaps*sizeof( fapRecord) % 80;
	if ( padSize != 0) {
	    padSize = 80 - padSize;
	    AmReadLoopFile( fd, padBuf, padSize);
	}
	lastpla = 0;
        for ( j = 0, polyPtr = section->lpolys;
              j < section->numPolys; j++, polyPtr++) {

/* find out how many arcs are in the permiter of the polygon.  Ignore  
   arcs in the islands, which are marked with 0's in the FAP subfile */

	    for ( k = lastpla; k <= abs( polyPtr->pla); k++) {
		if ( section->faps[ k] == 0) {
		    break;
		}
	    }
	    narcs = k - lastpla;
            polyPtr->narcs = 0;
	    polyPtr->npts = 0;
	    polyPtr->arcs = ( larcPtr *) malloc( sizeof( larcPtr *)*narcs);

/* now take each of the arc ID's and search for it in the vector of arcs and
   return a pointer to the arc. The FAP's are positive and negative for
   the polygon being to the left or right of the arc, which we don't really
   care about */

	    for ( k = lastpla; k < lastpla + narcs; k++) {
		targetArc.arcNo = abs( section->faps[ k]);
		for ( m = 0; m < section->numArcs; m++) {
		    if ( section->arcs[ m].arcNo == abs( section->faps[ k])) {
			polyPtr->arcs[ polyPtr->narcs++] = &section->arcs[ m];
			polyPtr->npts += section->arcs[ m].npts;
		    }
		}
	    }

/* allocate the vectors for the points. */

	    polyPtr->pts = ( fpointPtr) 
		malloc( sizeof( fpointRec)*polyPtr->npts);

/* for each one of the arcs, we've stored the number of points in the arc, 
   and also its starting coordinate in the coordinate subfile.  We now
   take each arc, and copy the coordinate to the polygon's points */

	    npts = 0;
	    for ( m = 0; m < polyPtr->narcs; m++) {
		index = polyPtr->arcs[ m] - section->arcs;
		if ( index == 0) {
		    start = 0;
		} else {
		    start = section->arcs[ index - 1].plc;
		}
		if ( section->faps[ m + lastpla] > 0) {
		    for ( n = start; n < polyPtr->arcs[ m]->plc; n++) {
			polyPtr->pts[ npts++] = section->coords[ n];
		    }
		} else {
		    for ( n = polyPtr->arcs[ m]->plc - 1; n >= start; n--) {
			polyPtr->pts[ npts++] = section->coords[ n];
		    }
		}
	    }
	    lastpla = abs( polyPtr->pla) + 1;
        }

/* sort the polygons so we can search on them for the islands */

	qsort( section->lpolys, section->numPolys, 
	       sizeof( lpolyRec), ( AmSearchFn) byPolyId);

/* now, we accumulate islands, but AmFirst we allocate the island arrays */ 
	
	for ( k = 0, polyPtr = section->lpolys; 
	      k < section->numPolys;
	      k++, polyPtr++) {
	    if ( polyPtr->surroundPolyNo != 0) {
		targPoly.polyNo = abs( polyPtr->surroundPolyNo);
		foundPoly = ( lpolyPtr) bsearch( &targPoly, 
						 section->lpolys, 
						 section->numPolys, 
						 sizeof( lpolyRec), 
						 ( AmSearchFn) byPolyId);
		if ( foundPoly == NULL) {
		    fprintf( stderr, "failed to find polygon %d\n",
			     targPoly.polyNo);
		} else {
		    foundPoly->numIslands++;
		}
	    }
	}

/* allocate island vectors and zero out accumulators */

	for ( k = 0, polyPtr = section->lpolys; 
	      k < section->numPolys;
	      k++, polyPtr++) {
	    if ( polyPtr->numIslands != 0) {
		polyPtr->islands = ( lpolyPtr *) 
		    malloc( sizeof( lpolyPtr)*polyPtr->numIslands);
	    }
	    polyPtr->numIslands = 0;
	}

/* fill the island vectors, and increment the island count accumulators back
   to their original values */

	for ( k = 0, polyPtr = section->lpolys; 
	      k < section->numPolys;
	      k++, polyPtr++) {
	    if ( polyPtr->surroundPolyNo != 0) {
		targPoly.polyNo = polyPtr->surroundPolyNo;
		foundPoly = ( lpolyPtr) bsearch( &targPoly, 
						 section->lpolys, 
						 section->numPolys, 
						 sizeof( lpolyRec), 
						 ( AmSearchFn) byPolyId);
		if ( foundPoly == NULL) {
		    fprintf( stderr, "failed to find polygon %d\n",
			     targPoly.polyNo);
		} else {
		    foundPoly->islands[ foundPoly->numIslands++] = polyPtr;
		}
	    }
	}
    }

/* done reading, close the file */

    AmCloseReadFile( fd, pipeFlag);

/* allocate the resulting vector of polygons */

    numPolys = 0;
    for ( i = 0; i < landusePtr->numSections; i++) {
	numPolys += landusePtr->sections[ i].numPolys;
    }
    landPolys = ( lulcPolyPtr) malloc( sizeof( lulcPolyRec)*( numPolys));

/*  copy from the old to the new */

    index = 0;
    for ( i = 0; i < landusePtr->numSections; i++) {

/* set a pointer from the old polygons to the new polygons */

	section = landusePtr->sections + i;
	for ( j = 0, polyPtr = landusePtr->sections[ i].lpolys; 
	      j < landusePtr->sections[ i].numPolys; j++, polyPtr++) {
	    copyPolygonIslands( pj, polyPtr, landPolys, &index, 
				swLat, swLon, TRUE);
	}
    }

/* find the extents of the polygons to generate the extents of the file */

    xmin = MY_HUGE;
    ymin = MY_HUGE;
    xmax = -MY_HUGE;
    ymax = -MY_HUGE;
    for ( ii = 0; ii < numPolys; ii++) {
	for ( jj = 0; jj < landPolys[ ii].geoPoly.polygon.numPts; jj++) {
	    if ( landPolys[ ii].geoPoly.polygon.pts[ jj].x < xmin) {
		xmin = landPolys[ ii].geoPoly.polygon.pts[ jj].x;
	    }
	    if ( landPolys[ ii].geoPoly.polygon.pts[ jj].y < ymin) {
		ymin = landPolys[ ii].geoPoly.polygon.pts[ jj].y;
	    }
	    if ( landPolys[ ii].geoPoly.polygon.pts[ jj].x > xmax) {
		xmax = landPolys[ ii].geoPoly.polygon.pts[ jj].x;
	    }
	    if ( landPolys[ ii].geoPoly.polygon.pts[ jj].y > ymax) {
		ymax = landPolys[ ii].geoPoly.polygon.pts[ jj].y;
	    }
	}
    }

/* create the data file and header for the polygon files.  */

    sprintf( polyfile, "%s.dat", outfile);
    sprintf( polyhdr, "%s.hdr", outfile);
    if (( fdOut = open( polyfile, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open output file %s\n", polyfile);
	return( FALSE);
    }
    if (( fpOut = fopen( polyhdr, "w")) == NULL) {
	fprintf( stderr, "failed to open header file %s\n", polyhdr);
	return( FALSE);
    }
    strncpy( buffer, mapHeader.title, sizeof( mapHeader.title));
    buffer[ sizeof( mapHeader.title) - 1]  = '\0';
    title = AmElimLeadWspace( buffer);
    title = AmElimTrailWspace( title);
    fprintf( fpOut, "filename:	%s\n", AmGetRootName( polyfile));
    fprintf( fpOut, "title:		%s polygon data\n", title);
    fprintf( fpOut, "rows:		%d\n", numPolys);
    fprintf( fpOut, "xunits:		arcdeg\n");
    fprintf( fpOut, "xvariable:	longitude\n");
    fprintf( fpOut, "yunits:		arcdeg\n");
    fprintf( fpOut, "yvariable:	latitude\n");
    fprintf( fpOut, "type:		polygonPlot\n");
    fprintf( fpOut, "size:		%d\n", sizeof( lulcPolyRec));
    fprintf( fpOut, "xmin:		%f\n", xmin);
    fprintf( fpOut, "ymin:		%f\n", ymin);
    fprintf( fpOut, "xmax:		%f\n", xmax);
    fprintf( fpOut, "ymax:		%f\n", ymax);
    fprintf( fpOut, "attributeMap:	%s\n", attrCodeFile);
    fprintf( fpOut, "fileSource:	LULC\n");
    fprintf( fpOut, "colorsType:	int\n");
    fprintf( fpOut, "colorsOffset:	%d\n", sizeof( AmPolygonRec));
    fprintf( fpOut, "sourceFile:	%s\n", landuseFile);
    fprintf( fpOut, "description:	%s\n", descriptionFile);
#if BYTE_ORDER == LITTLE_ENDIAN
    fprintf( fpOut, "byteOrder: 	littleEndian\n");
#else
    fprintf( fpOut, "byteOrder: 	bigEndian\n");
#endif
    fclose( fpOut);
    for ( ii = 0; ii < numPolys; ii++) {
	AmExtentsPoly( &landPolys[ ii].geoPoly.polygon);
	landPolys[ ii].geoPoly.polygon.flags = AM_MBR_FLAG;
	write( fdOut, &landPolys[ ii], sizeof( landPolys[ ii]));
	write( fdOut, landPolys[ ii].geoPoly.polygon.pts, 
		      landPolys[ ii].geoPoly.polygon.numPts*sizeof( AmCoordRec));
    }
    close( fdOut);
    return( TRUE);
}


int main( int argc, char *argv[])
{
    char	attrpath[ PATHLEN], *env;

    if ( argc < 3) {
	fprintf( stderr, "usage %s infile outfile [attrfile]\n", argv[ 0]);
	exit( -1);
    }
    if ( argc == 3) {
        env = getenv( CONFIG_ENV);
        if ( env == NULL) {
            fprintf( stderr, "%s environment variable not set\n", CONFIG_ENV);
            exit( -1);
        }
        sprintf( attrpath, "%s/mappings/lulc_mapping", env);
    } else {
        strcpy( attrpath, argv[ 3]);
    }
    if ( !readLanduseData( argv[ 1], attrpath, argv[ 2])) {
	fprintf( stderr, "failed to read landuse data file\n");
	exit( -1);
    }
}
