#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */




#include <stdio.h>
#include <math.h>
#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <anvil/datatype.h>
#include <anvil/rect.h>
#include <anvil/fns.h>
#include <anvil/strlib.h>
#if defined(_WIN32)
#include <anvil/ntresource.h>
#else
#include <anvil/xresource.h>
#endif
#include <anvil/newtile.h>
#include <anvil/macros.h>
#include <anvil/tagstrs.h>
#include <anvil/image.h>
#include <anvil/imageP.h>
#include <anvil/compression.h>
#include <anvil/files.h>
#include <anvil/choppath.h>
#include <util/utils.h>
#include <util/tile.h>

/* this file implements the recursive quadtree algorithm for generating
   hierarchical format pyramid files.  The program may take a list of
   input files of the same type of varying resolutions, and generates
   an ASCII header file, a tile header file containing the tile filenames
   and tile data structure with seek addresses, and one or more tile
   data files 
   First pass:  an in-memory representation of the tile tree is created
   which reflects the structure and distribution of the output tiles.
   Second pass:  the source files are written to the tile tree in blocks,
   which are recursively distributed amongst the tiles.
   Third pass:  if the -compress option is selected, the tiles are 
   compressed in-place using LZW compression, the tile seek pointers
   are reset and the tile data files are truncated.
*/

#define FUDGE 	0.1

/* input file description structure */

typedef struct fileInfo_ {
    char		*header;	/* source header file */
    char		*file;		/* source data file */
    float		xmin;		/* extents of file */
    float		ymin;
    float		xmax;
    float		ymax;
    int			rows;		/* data dimensions of file */
    int			cols;
    float		xres;		/* file x and y resolutions */
    float		yres;
    int			fd;		/* file descriptor */
    int			size;		/* file element size */
    AmDataTypeEnum	dataType;	/* file data type */
} fileInfoRec, *fileInfoPtr;

/* because shorts aren't large enough the hold the dimensions of the
   images that we deal with!!! */

typedef struct BigRect_ {
    int		x;
    int		y;
    int		width;
    int		height;
} BigRectRec, *BigRectPtr;


const int MAXFILES = 256;
const int BLOCKSIZE = 8192*1024;

/****************************************************************************
 *
 * Description:
 *      comparator function to sort input file description structures
 *      by the files' resolutions.
 * Arguments:
 *      a, b - input file description structure 
 * Returns:
 *      TRUE if the object is of this type or a subclass of it.
 *      FALSE if not.
 *
 ****************************************************************************/

static int 
byResolution( fileInfoPtr	a,
	      fileInfoPtr	b)
{ 
    float	ares, bres;

    ares = MIN(( a->xmax - a->xmin)/a->cols, ( a->ymax - a->ymin)/a->rows);
    bres = MIN(( b->xmax - b->xmin)/b->cols, ( b->ymax - b->ymin)/b->rows);
    if ( ares < bres) {
	return( 1);
    } else if ( ares == bres) {
	return( 0);
    } else {
	return( -1);
    }
}

#if 0
/****************************************************************************
 *
 * Description:
 *      predicate function which returns TRUE if the file completely contains
 *      the requested extents.
 * Arguments:
 *	fileInfo - input file description structure.
 *      xmin, ymin, xmax, ymax - extents to test against
 * Returns:
 *      TRUE if the file contains the extents, FALSE if it doesn't.
 *
 ****************************************************************************/

static AnvilBoolean 
containFile( fileInfoPtr fileInfo,
	     float       xmin,
	     float       ymin,
	     float       xmax,
	     float       ymax)
{
    return(( fileInfo->xmin >= xmin) && ( fileInfo->ymin >= ymin) &&
           ( fileInfo->xmax <= xmax) && ( fileInfo->ymax <= ymax));
}
#endif


/****************************************************************************
 *
 * Description:
 *      predicate function which returns TRUE if the file overlaps
 *      the requested extents.
 * Arguments:
 *	fileInfo - input file description structure.
 *      xmin, ymin, xmax, ymax - extents to test against
 * Returns:
 *      TRUE if the file overlaps the extents, FALSE if it doesn't.
 *
 ****************************************************************************/

static AnvilBoolean 
overlapFile( fileInfoPtr fileInfo,
	     float       xmin,
	     float       ymin,
	     float       xmax,
	     float       ymax)
{
    return(( fileInfo->xmin < xmax) && ( fileInfo->ymin < ymax) &&
           ( fileInfo->xmax > xmin) && ( fileInfo->ymax > ymin));
}

/****************************************************************************
 *
 * Description:
 *      given a source matrix and coordinate system, and a destination
 *      matrix and coordinate system, calculate the source and destination
 *      rectangles in integer coordinates.  This is the same as 
 *	AmSrcAndDestRects, except that it takes rectangles with integer values
 *	instead of shorts.
 * Arguments:
 *      srcRows, srcCols - dimensions of the source matrix.
 *      srcXmin, srcYmin, srcXmax, srcYmax - coordinates of the source matrix.
 *      dstRows, dstCols - dimensions of the destination matrix.
 *      dstXmin, dstYmin, dstXmax, dstYmax - coordinates of destination matrix.
 *      srcRect - returned source rectangle.
 *      dstRect - returned destination rectangle.
 * Returns:
 *      TRUE if the rectangles overlap, FALSE otherwise.
 * Modifications:
 *      I'm not sure the overlap test is correct.
 *
 ****************************************************************************/

static AnvilBoolean
srcAndDestBigRects( int            srcRows,
                    int            srcCols,
                    float          srcXmin,
                    float          srcYmin,
                    float          srcXmax,
                    float          srcYmax,
                    int            dstRows,
                    int            dstCols,
                    float          dstXmin,
                    float          dstYmin,
                    float          dstXmax,
                    float          dstYmax,
                    BigRectPtr     srcRect,
                    BigRectPtr     dstRect)
{
    float       allXmin, allXmax, allYmin, allYmax;
    int         x1, y1;

/* find the intersection of the coordinate systems */

    allXmin = MAX( srcXmin, dstXmin);
    allYmin = MAX( srcYmin, dstYmin);
    allXmax = MIN( srcXmax, dstXmax);
    allYmax = MIN( srcYmax, dstYmax);
    if (( allXmin >= allXmax) || ( allYmin >= allYmax)) {
        return( FALSE);
    }

/* calculate the source rectangle */

    srcRect->x = (int)(srcCols*( allXmin - srcXmin)/( srcXmax - srcXmin));
    srcRect->y = (int)(srcRows*( srcYmax - allYmax)/( srcYmax - srcYmin));
    x1 = (int)(srcCols*( allXmax - srcXmin)/( srcXmax - srcXmin));
    y1 = (int)(srcRows*( srcYmax - allYmin)/( srcYmax - srcYmin));
    srcRect->width = x1 - srcRect->x;
    srcRect->height = y1 - srcRect->y;

/* and the destination rectangle */

    dstRect->x = (int)(dstCols*( allXmin - dstXmin)/( dstXmax - dstXmin));
    dstRect->y = (int)(dstRows*( dstYmax - allYmax)/( dstYmax - dstYmin));
    x1 = (int)(dstCols*( allXmax - dstXmin)/( dstXmax - dstXmin));
    y1 = (int)(dstRows*( dstYmax - allYmin)/( dstYmax - dstYmin));
    dstRect->width = x1 - dstRect->x;
    dstRect->height = y1 - dstRect->y;
    return( TRUE);
}

/****************************************************************************
 *
 * Description:
 * 	resize a source matrix to a destination matrix 
 * Arguments:
 *	srcMatrix - source matrix
 *	srcRows, srcCols - dimensions of the source matrix 
 *	dstMatrix - destination matrix.
 *	dstRows, dstCols - dimensions of the destination matrix.
 *	size - element size.
 *	srcRect - rectangle to sample from the source matrix.
 *	dstRect - rectangle to sample from the destination matrix.
 *
 ***************************************************************************/

static AmErrorCode 
resizeBigMatrix( uchar_t               *srcMatrix,
		 int                   srcRows,
		 int                   srcCols,
		 uchar_t               *dstMatrix,
		 int                   dstRows,
		 int                   dstCols,
		 int                   size,
		 BigRectRec const      *srcRect,
		 BigRectRec const      *dstRect)
{
    int         *hIndex, *vIndex, *pHIndex, *pVIndex, srcWidth, dstWidth, ii;
    uchar_t     *srcPtr, *dstPtr, *dstEnd, *rowPtr, *rowEnd, *srcRowPtr;

/* create the horizontal and vertical difference vectors */

    if (( hIndex = AmMakeFwdDiffVec( srcRect->width, dstRect->width)) == NULL) {
	fprintf( stderr, "failed to allocate horiz difference vector %d\n", 
		 dstRect->width);
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    for ( ii = 0; ii < dstRect->width; ii++) {
        hIndex[ ii] *= size;
    }
    if (( vIndex = AmMakeFwdDiffVec( srcRect->height, dstRect->height)) == NULL) {
	fprintf( stderr, "failed to allocate vert difference vector %d\n", 
		 dstRect->height);
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    srcWidth = srcCols*size;
    dstWidth = dstCols*size;
    for ( ii = 0; ii < dstRect->height; ii++) {
        vIndex[ ii] *= srcWidth;
    }
    pVIndex = vIndex;

/* offset the source and destination matrices by their rectangles x,y */

    srcPtr = srcMatrix + ( srcRect->y*srcCols + srcRect->x)*size;
    dstPtr = dstMatrix + ( dstRect->y*dstCols + dstRect->x)*size;
    dstEnd = dstPtr + dstRect->height*dstWidth;

/* loop over each row, setting pointers to the start of the source and
   destination matrix, and set the end row marker */

    while ( dstPtr < dstEnd) {
        pHIndex = hIndex;
        rowPtr = dstPtr;
        rowEnd = dstPtr + dstRect->width*size;
        srcRowPtr = srcPtr;

/* loop to the end of the row of the destination matrix */

        while ( rowPtr < rowEnd) {
            memcpy( rowPtr, srcRowPtr, size);
            rowPtr += size;
            srcRowPtr += *pHIndex++;    /* premultiplied by size */
        }

/* step to the next row of the destination matrix, and the corresponding
   source matrix */

        dstPtr += dstWidth;
        srcPtr += *pVIndex++;   /* premultiplied by srcWidth */
    }
    free( hIndex);
    free( vIndex);
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 * 	take a block of the source data with its dimensions and extents
 *	and recursively write the block to the tiles pointed to by the
 *	tile pyramid
 * Arguments:
 *	fds - file descriptors to each of the output tile files.
 *	block - block to write to the tile pyramid.
 *	tile - tile to write to the tree
 *	blockRows - number of rows in the block.
 *	blockCols - number of columns in the block
 *	xmin, ymin, xmax, ymax - extents of the block in the coordinate
 *	system.
 * Returns:
 *	AmOK on OK, error otherwise.
 *
 ****************************************************************************/

static AmErrorCode 
writeBlockRecursive( int	*fds,
		     uchar_t	*block,
		     memTilePtr	tile,
		     int	blockRows,
		     int	blockCols,
		     int	elementSize,
		     float	xmin,
		     float	ymin,
		     float	xmax,
		     float	ymax)
{
    AmTilePtr		fileTile;
    float		tileRes, blockRes;
    uchar_t		*tileBlock=NULL;
    BigRectRec		srcRect, dstRect;
    seekAddress		tileOffset=0;
    AmErrorCode		errCode;
    int			tileSize=0;

/* if the tile is NULL (past the leaf), then return */

    if ( tile == NULL) {
	return( AmOK);
    }
    fileTile = &tile->fileTile;
    if ( fileTile == NULL) {
	DO_RETURN_ERROR( AmBAD_ARGUMENT);
    }

/* does the block overlap the current tile? */

    if (( fileTile->xmax < xmin) || ( fileTile->ymax < ymin) ||
	( fileTile->xmin > xmax) || ( fileTile->ymin > ymax)) {
	return( AmOK);
    }

/* is the current tile resolution higher than the block resolution? */

    tileRes = MIN(( fileTile->xmax - fileTile->xmin)/fileTile->cols,
		  ( fileTile->ymax - fileTile->ymin)/fileTile->rows);
    blockRes = MIN(( xmax - xmin)/blockCols, ( ymax - ymin)/blockRows);

/* go down 1 extra layer, but don't write the filled flag... */

    if ( blockRes > tileRes*2) {
	return( AmOK);
    }

/* recurse, but do not write the tile if it has already been filled */

    if ( !tile->filled) {

/* we're in...  figure out how the block projects into the tile */

	srcAndDestBigRects( blockRows, blockCols, xmin, ymin, xmax, ymax,
			    fileTile->rows, fileTile->cols, 
			    fileTile->xmin, fileTile->ymin,
			    fileTile->xmax, fileTile->ymax,
			    &srcRect, &dstRect);
AN_DEBUG( 10, ("from (%d, %d, %d, %d) to (%d, %d, %d, %d)\n",
		srcRect.x, srcRect.y, srcRect.width, srcRect.height,
		dstRect.x, dstRect.y, dstRect.width, dstRect.height));
AN_DEBUG( 10, ("file index = %d, seek addr = %ld\n", 
		fileTile->fileIndex, fileTile->seekAddr));

/* seek to the tile, and the correct row in the tile */
  
	tileOffset = dstRect.y*fileTile->cols*elementSize;
	if ( LSEEK( fds[ fileTile->fileIndex], 
		    fileTile->seekAddr + tileOffset, SEEK_SET) < 0) {
	    fprintf( stderr, "failed to seek to %ld in file %d\n",
		     fileTile->seekAddr + tileOffset, fileTile->fileIndex);
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}

/* allocate a block for the tile, and read it.  We re-read the block
   because of integer roundoff problems that might cause the tile to
   overflow.  We malloc and de-malloc the blocks, because they're of 
   the same size, and since this is recursive, we don't want to do 
   allocas() */

	tileSize = dstRect.height*fileTile->cols*elementSize;
	tileBlock = ( uchar_t *) malloc( tileSize);
	if ( tileBlock == NULL) {
	    fprintf( stderr, "failed to allocate tile block %d bytes\n",
			     tileSize);
	    fprintf( stderr, "tile extents = %.2f, %.2f, %.2f, %.2f\n",
			     fileTile->xmin, fileTile->ymin,
			     fileTile->xmax, fileTile->ymax);
	    DO_RETURN_ERROR( AmMALLOC_FAILED);
	}
	if ( read( fds[ fileTile->fileIndex], tileBlock, tileSize) < 0) {
	    fprintf( stderr, "failed to read tile block %d bytes\n",
			     tileSize);
	    fprintf( stderr, "tile extents = %.2f, %.2f, %.2f, %.2f\n",
			     fileTile->xmin, fileTile->ymin,
			     fileTile->xmax, fileTile->ymax);
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}

/* since we read to the start of the block, reset the destination rectangle
   start row to 0 */

	dstRect.y = 0;
	resizeBigMatrix( block, blockRows, blockCols, 
		         tileBlock, dstRect.height, fileTile->cols,
		         elementSize, &srcRect, &dstRect);
    }

/* seek back to the start of the block in the tile and write it out again. */

    if ( LSEEK( fds[ fileTile->fileIndex], 
	        fileTile->seekAddr + tileOffset, SEEK_SET) < 0) {
	fprintf( stderr, "failed to seek to %ld in file %d\n",
		 fileTile->seekAddr + tileOffset, fileTile->fileIndex);
	DO_RETURN_ERROR( AmSEEK_FAILED);
    }
    if ( write( fds[ fileTile->fileIndex], tileBlock, tileSize) < 0) {
	fprintf( stderr, "failed to write %d bytes in file %d\n",
		 tileSize, fileTile->fileIndex);
	DO_RETURN_ERROR( AmWRITE_FAILED);
    }
    free( tileBlock);
    errCode = writeBlockRecursive( fds, block, tile->llChild, 
				   blockRows, blockCols, elementSize, 
				   xmin, ymin, xmax, ymax);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeBlockRecursive( fds, block, tile->ulChild, 
				   blockRows, blockCols, elementSize, 
				   xmin, ymin, xmax, ymax);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeBlockRecursive( fds, block, tile->lrChild, 
				   blockRows, blockCols, elementSize, 
				   xmin, ymin, xmax, ymax);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeBlockRecursive( fds, block, tile->urChild, 
				   blockRows, blockCols, elementSize, 
				   xmin, ymin, xmax, ymax);
    if ( errCode != AmOK) {
	return( errCode);
    }
    return( AmOK);
}

/****************************************************************************
 *
 * Description:
 *	given a source file, and the tile tree, set the "filled" flags
 *	for all the tiles which are wholly contained within the file,
 *	and of lower resolution than the file.
 * Arguments:
 *	sourceFile - source file description structure.
 *	tile - pointer to the tile tree
 *
 ****************************************************************************/

static void 
setFilledFlags( fileInfoPtr	sourceFile,
		memTilePtr	tile)
{
    float	tileRes, fileRes;

/* leaf tile ends the recursion */

    if ( tile == NULL) {
	return;
    }

/* if the tile is wholly contained in the file's extents... */

    if (( tile->fileTile.xmin >= sourceFile->xmin) &&
	( tile->fileTile.ymin >= sourceFile->ymin) &&
	( tile->fileTile.xmax <= sourceFile->xmax) &&
	( tile->fileTile.ymax <= sourceFile->ymax)) {

/* and the tile's resolution is lower than the files... */

	tileRes = MIN(( tile->fileTile.xmax - 
			    tile->fileTile.xmin)/tile->fileTile.cols,
		      ( tile->fileTile.ymax - 
			    tile->fileTile.ymin)/tile->fileTile.rows);
	fileRes = MIN(( sourceFile->xmax - sourceFile->xmin)/sourceFile->cols,
		      ( sourceFile->ymax - sourceFile->ymin)/sourceFile->rows);

/* set the filled flag and recurse... */

	if ( fileRes < tileRes) {
	    tile->filled = TRUE;
	} else {
	    return;
	}
	setFilledFlags( sourceFile, tile->llChild);
	setFilledFlags( sourceFile, tile->ulChild);
	setFilledFlags( sourceFile, tile->lrChild);
	setFilledFlags( sourceFile, tile->urChild);
    }
}

/****************************************************************************
 *
 * Description:
 * 	take a file and recursively write it to the pyramid tile tree. 
 * Arguments:
 *	sourceFile - source file description structure.
 *	elementSize - size of the elements in the source file 
 *	rootTile - root tile of the pyramid.
 *	outputFds - vector of output file descriptors.
 * Returns:
 *	AmOK on success, system error on open, seek, read or malloc failure.
 * Modifications:
 *	the element size is contained in the file description structure
 *	and is redundant.
 *
 ****************************************************************************/

static AmErrorCode 
writeFileToTree( fileInfoPtr	sourceFile,
		 int		elementSize,
		 memTilePtr	rootTile,
		 int		*outputFds)
{
    uchar_t	*block;
    int		rowsize, blockRows, sourceFd, numBytes, rowsRead, ii;
    float	ymax, rowHeight;
    AmErrorCode	errCode;
    AnvilBoolean	pipeFlag;

    errCode = AmOK;

/* how big is a row, how many rows can we read in a block, and how tall
   is a row in the coordinate system? */

    rowsize = sourceFile->cols*elementSize;
    blockRows = MIN( BLOCKSIZE/rowsize, sourceFile->rows);
    rowHeight = ( sourceFile->ymax - sourceFile->ymin)/sourceFile->rows;

/* allocate the read-block, then open the file. */

    if (( block = ( uchar_t *) malloc( rowsize*blockRows)) == NULL) {
	DO_RETURN_ERROR( AmMALLOC_FAILED);
    }
    if (( sourceFd = AmOpenReadFile( sourceFile->file, &pipeFlag)) < 0) {
	fprintf( stderr, "failed to open source file %s\n", sourceFile->file);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* read in blocks */

    rowsRead = 0;
    for ( ii = 0, ymax = sourceFile->ymax; 
	  ii < sourceFile->rows; 
	  ii += blockRows, ymax -= rowHeight*rowsRead) {
	numBytes = AmReadLoopFile( sourceFd, block, rowsize*blockRows);
	rowsRead = numBytes/rowsize;

/* write the block to the pyramid */

	errCode = writeBlockRecursive( 
			outputFds, block, rootTile, rowsRead, 
			sourceFile->cols, elementSize,
			sourceFile->xmin, ymax - rowHeight*rowsRead, 
			sourceFile->xmax, ymax);
	if ( errCode != AmOK) {
	    fprintf( stderr, "failed to write block at y = %.2f, ii = %d\n",
		     ymax, ii);
	    break;
	}
    }
    free( block);

/* set the filled flags in the tree for this file, so the tiles won't be
   overwritten */

    setFilledFlags( sourceFile, rootTile);
    AmCloseReadFile( sourceFd, pipeFlag);
    return( AmOK);
}
    
/****************************************************************************
 *
 * Description:
 * 	recursively create the in-memory tile tree from the source files
 *	and extents.
 * Arguments:
 *	xmin, ymin, xmax, ymax - extents to cover for the tile tree.
 *	res - resolution at this level.
 *	tileRows, tileCols - dimensions of the matrix for each tile.
 *	fileInfo - vector of source file information structures.
 *	numFiles - number of source files.
 *	depth - desired depth of the pyramid.
 *	elementSize - size of each element.
 *	seekAddrs - current seek addresses in each of the output tile
 *	files, maintained during the recursion.
 *	numOutputFiles - number of output tile files, and seek addresses.
 *	tileCount - total number of tiles.
 * Returns:
 *	an in-memory pointer containing the tile tree. 
 *
 ****************************************************************************/

memTilePtr 
makeTileTree( float		xmin,
	      float		ymin,
	      float		xmax,
	      float		ymax,
	      float		res,
	      int		tileRows,
	      int		tileCols,
	      fileInfoPtr	fileInfo,
	      int		numFiles,
	      int		depth,
	      int		elementSize,
	      seekAddress	*seekAddrs,
	      int		numOutputFiles,
	      int		*tileCount)
{
    float	fileRes, bestRes, cx, cy;
    int		bestIndex, ii, fileIndex;
    memTilePtr	tile;
    ulong_t	numBytes;

/* we're done if we've reached the maximum depth */

    if ( depth == 0) {
	return( NULL);
    }

/* see if there is a file which contains the extents, and is of higher 
   resolution than the requested resolution.  Find the file with the 
   closest resolution to the requested resolution, which contains the
   requested area. */

    bestRes = -MY_HUGE;
    bestIndex = -1;
    for ( ii = 0; ii < numFiles; ii++) {
	if ( overlapFile( &fileInfo[ ii], xmin, ymin, xmax, ymax)) {
	    fileRes = MIN( fileInfo[ ii].xres, fileInfo[ ii].yres);
	    if (( fileRes <= res + FUDGE) && ( fileRes > bestRes)) {
		bestIndex = ii;
		bestRes = fileRes;
	    }
	}
    }

/* if we didn't find a file with a tighter resolution than requested, then
   we don't create a tile, because there's no sense in supersampling */

    if ( bestIndex == -1) {
	return( NULL);
    }

/* distribute the tiles amongst the different files */

    fileIndex = *tileCount % numOutputFiles;

/* otherwise, create the tile pointer and recurse */
	
    tile = ( memTilePtr) malloc( sizeof( memTileRec));
#if 0
    tile->fileTile.rows = ( ymax - ymin)/res;
    tile->fileTile.cols = ( xmax - xmin)/res;
#endif
    tile->fileTile.rows = tileRows;
    tile->fileTile.cols = tileCols;
    numBytes = tile->fileTile.rows*tile->fileTile.cols*elementSize;
    tile->fileTile.numBytes = numBytes;
    tile->fileTile.numBytesUncompressed = numBytes;
    tile->fileTile.xmin = xmin;
    tile->fileTile.ymin = ymin;
    tile->fileTile.xmax = xmax;
    tile->fileTile.ymax = ymax;
    tile->filled = FALSE;
    tile->fileTile.flags = 0;
#if defined(SEEK64)
    tile->flags |= AmTILE_64BIT_SEEK_FLAG;
#endif

/* set the tile's seek address, and increment the seek address */

    tile->fileTile.seekAddr = seekAddrs[ fileIndex];
    tile->fileTile.fileIndex = fileIndex;

/* right now, the file indices are set to 0, but later, they'll index the 
   list of output files. */

    tile->fileTile.fileIndex = fileIndex;
    seekAddrs[ fileIndex] += numBytes;

/* recurse, dividing the tile by 4 from the center. */

    cx = ( xmax + xmin)/2;
    cy = ( ymax + ymin)/2;
    ( *tileCount)++;
    AN_DEBUG( 10, ( "file index = %d, tile Count = %d rows = %d, cols = %d\n", 
		    fileIndex, *tileCount, 
		    tile->fileTile.rows, tile->fileTile.cols));
    tile->llChild = makeTileTree( xmin, ymin, cx, cy, res/2.0, 
    				  tileRows, tileCols, fileInfo,
				  numFiles, depth - 1, elementSize, seekAddrs,
				  numOutputFiles, tileCount);
    tile->ulChild = makeTileTree( xmin, cy, cx, ymax, res/2.0, 
    				  tileRows, tileCols, fileInfo,
				  numFiles, depth - 1, elementSize, seekAddrs,
				  numOutputFiles, tileCount);
    tile->lrChild = makeTileTree( cx, ymin, xmax, cy, res/2.0, 
    				  tileRows, tileCols, fileInfo,
				  numFiles, depth - 1, elementSize, seekAddrs,
				  numOutputFiles, tileCount);
    tile->urChild = makeTileTree( cx, cy, xmax, ymax, res/2.0, 
    				  tileRows, tileCols, fileInfo,
				  numFiles, depth - 1, elementSize, seekAddrs,
				  numOutputFiles, tileCount);
    return( tile);
}

/****************************************************************************
 *
 * Description:
 * 	take the ASCII list of files from the list file, and turn them into
 *  	a vector of fileInfo structs 
 * Arguments:
 *	listFile - ascii file of headers and data files of the form
 *	blotz.hdr blotz.dat.
 *	fileInfo - vector of source file information structures.
 *	numFiles - number of source files.
 *	elementSize - size of the elements.
 * Returns:
 *	TRUE if all the files are of the same format and type, and parse
 *	correctly, FALSE otherwise.
 *
 ****************************************************************************/

static AnvilBoolean 
extractFileInfo( char		*listFile,
		 fileInfoPtr	fileInfo,
		 int		*numFiles,
		 int		*elementSize)
{
    FILE	*fp;
    char	buffer[ BUFFERLEN], header[ PATHLEN], file[ PATHLEN],
		*dataTypeStr, *kind;
    AnvilBoolean	okFlag;

    *numFiles = 0;
    if (( fp = fopen( listFile, "r")) == NULL) {
	fprintf( stderr, "failed to open %s\n", listFile);
	return( FALSE);
    }
    okFlag = TRUE;
    do {
	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
	    break;
	}
	sscanf( buffer, "%s %s", header, file);
	fileInfo[ *numFiles].header = strdup( header);
	fileInfo[ *numFiles].file = strdup( file);
	fileInfo[ *numFiles].fd = 0;

#if defined(_WIN32)
	AmNTResourceDbClass db( header);
#else
	AmXResourceDbClass db( header);
#endif

	if ( !db.IsValid()) {
	    fprintf( stderr, "header %s is not valid\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNtype, &kind)) {
	    fprintf( stderr, "failed to get kind from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNxmin, &fileInfo[ *numFiles].xmin)) {
	    fprintf( stderr, "failed to get xmin from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNymin, &fileInfo[ *numFiles].ymin)) {
	    fprintf( stderr, "failed to get ymin from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNxmax, &fileInfo[ *numFiles].xmax)) {
	    fprintf( stderr, "failed to get xmax from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNymax, &fileInfo[ *numFiles].ymax)) {
	    fprintf( stderr, "failed to get ymax from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNrows, &fileInfo[ *numFiles].rows)) {
	    fprintf( stderr, "failed to get rows from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNcols, &fileInfo[ *numFiles].cols)) {
	    fprintf( stderr, "failed to get cols from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNsize, &fileInfo[ *numFiles].size)) {
	    fprintf( stderr, "failed to get size from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( *numFiles == 0) {
	    *elementSize = fileInfo[ *numFiles].size;
	} else if ( fileInfo[ *numFiles].size != *elementSize) {
	    fprintf( stderr, "element size for %s is %d, not %d\n",
		     header, fileInfo[ *numFiles].size, *elementSize);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNdataType, &dataTypeStr)) {
	    fprintf( stderr, "failed to get data type from %s\n", header);
	    okFlag = FALSE;
	    break;
	}

/* find the data type enumeration */

	fileInfo[ *numFiles].dataType = AmStrToDataTypeEnum( dataTypeStr);

/* calculate the file's resolution */

	fileInfo[ *numFiles].xres = 
		( fileInfo[ *numFiles].xmax - 
		  fileInfo[ *numFiles].xmin)/fileInfo[ *numFiles].cols;
	fileInfo[ *numFiles].yres = 
		( fileInfo[ *numFiles].ymax - 
		  fileInfo[ *numFiles].ymin)/fileInfo[ *numFiles].rows;
	( *numFiles)++;
    } while ( okFlag);
    fclose( fp);
    return( okFlag);
}

int main( int argc, char *argv[])
{
    fileInfoRec fileInfo[ MAXFILES];
    int		numSrcFiles, ii, depth, maxCols, maxRows,
		topRows, topCols, elementSize, maxResIndex, numDstFiles,
		tileCount, outputFds[ MAXFILES];
    float	xmin, ymin, xmax, ymax, minRes, maxRes, topRes;
    memTilePtr	rootTile;
    seekAddress	seekAddrs[ MAXFILES], losses[ MAXFILES];
    char	tileHeader[ PATHLEN], *filenames[ MAXFILES], filebuf[ PATHLEN],
		header[ PATHLEN], 
		*imageFields[] = { AmNfilename, AmNsourceType,
				   AmNdescription, 
				   AmNxmin, AmNymin, AmNxmax, AmNymax,
				   AmNrows, AmNcols, NULL };
    FILE	*hdrFp;
    AnvilBoolean	compressFlag = FALSE;


    if ( argc < 4) {
	fprintf( stderr, 
	     "usage %s file-list outfile depth [numDstFiles=1] -compress\n", 
		 argv[ 0]);
	exit( -1);
    }
    if ( !extractFileInfo( argv[ 1], fileInfo, &numSrcFiles, &elementSize)) {
	fprintf( stderr, "failed to read header files\n");
	exit( -1);
    }

/* sort the file so the ones with the loosest resolution are generated AmFirst */

    qsort( fileInfo, numSrcFiles, sizeof( fileInfoRec), 
	   ( AmSearchFn) byResolution);

    depth = atoi( argv[ 3]);
    numDstFiles = 1;
    if ( argc >= 5) {

/* test if the 4th argument is -compress */

	if ( !strcmp( argv[ 4], "-compress")) {
	    compressFlag = TRUE;
	} else {
	    numDstFiles = atoi( argv[ 4]);
	}
    } 

/* test if the 5th argument is -compress */
   
    if ( argc >= 6) {
	if ( !strcmp( argv[ 5], "-compress")) {
	    compressFlag = TRUE;
	}
    }
    if (( numDstFiles < 1) || ( numDstFiles > 10)) {
	fprintf( stderr, "bad number of destination files %d\n", numDstFiles);
	exit( -1);
    }

/* create the output filenames */

    sprintf( tileHeader, "%s.tileHdr", argv[ 2]);
    for ( ii = 0; ii < numDstFiles; ii++) {
	sprintf( filebuf, "%s.%d.dat", argv[ 2], ii);
	filenames[ ii] = strdup( filebuf);
    }
    sprintf( header, "%s.hdr", argv[ 2]);

/* find the extents of all the files, and the min and max resolution. */ 

    xmin = MY_HUGE;
    ymin = MY_HUGE;
    xmax = -MY_HUGE;
    ymax = -MY_HUGE;
    minRes = MY_HUGE;
    maxRes = -MY_HUGE;
    maxResIndex = 0;
    for ( ii = 0; ii < numSrcFiles; ii++) {
	if ( fileInfo[ ii].xmin < xmin) {
	    xmin = fileInfo[ ii].xmin;
	}
	if ( fileInfo[ ii].ymin < ymin) {
	    ymin = fileInfo[ ii].ymin;
	}
	if ( fileInfo[ ii].xmax > xmax) {
	    xmax = fileInfo[ ii].xmax;
	}
	if ( fileInfo[ ii].ymax > ymax) {
	    ymax = fileInfo[ ii].ymax;
	}
	if ( MAX( fileInfo[ ii].xres, fileInfo[ ii].yres) > maxRes) {
	    maxRes = MAX( fileInfo[ ii].xres, fileInfo[ ii].yres);
	    maxResIndex = ii;
	}
	if ( MIN( fileInfo[ ii].xres, fileInfo[ ii].yres) < minRes) {
	    minRes = MAX( fileInfo[ ii].xres, fileInfo[ ii].yres);
	}
    }

/* figure out the resolution at the top of the pyramid */

    topRes = minRes;
    for ( ii = 0; ii < depth; ii++) {
	topRes *= 2;
    }

/* is the top resolution greater than the largest file's resolution? */

    if ( topRes < maxRes) {
	fprintf( stderr, "warning:insufficient depth to section top file\n");
	fprintf( stderr, "file res = %.2f, topRes = %.2f\n", maxRes, topRes);
    }

/* so, how big is it? */

    maxCols = (int)rint(( xmax - xmin)/minRes);
    maxRows = (int)rint(( ymax - ymin)/minRes);
    topCols = (int)rint(( xmax - xmin)/topRes);
    topRows = (int)rint(( ymax - ymin)/topRes);

    AN_DEBUG( 10, ("top dims = (%d, %d), bottom dims = (%d, %d)\n",
		   topRows, topCols, maxRows, maxCols));
    AN_DEBUG( 10, ("tile rows = %d, tileCols = %d\n", 
		   maxRows/( 1 << depth), maxCols/(1 << depth)));

/* initialize the seek addresses for each of the output files */

    for ( ii = 0; ii < numDstFiles; ii++) {
	seekAddrs[ ii] = 0;
    }
    tileCount = 0;
    rootTile = makeTileTree( xmin, ymin, xmax, ymax, topRes, topRows, topCols,
			     fileInfo, numSrcFiles, depth + 1, elementSize, 
			     seekAddrs, numDstFiles, &tileCount);
#if 0
    AmPrintTileTree( rootTile);
#endif

/* open the tile data file */

    for ( ii = 0; ii < numDstFiles; ii++) {
	if (( outputFds[ ii] = open( filenames[ ii], 
				   O_RDWR | O_CREAT, PERMS)) < 0) {
	    fprintf( stderr, "failed to open output file %s\n", filenames[ ii]);
	    exit( -1);
	}
    }

/* write each of the source files to the tile tree. */

    for ( ii = 0; ii < numSrcFiles; ii++) {
	AN_DEBUG( 10, ( "writing %s to tree\n", fileInfo[ ii].file));
	if ( writeFileToTree( &fileInfo[ ii], elementSize, 
			      rootTile, outputFds) != AmOK) {
	    fprintf( stderr, "writeFileToTreeFailed, code = %d\n", 
		     ANVIL_ERROR_CODE);
	}
    }

/* if compression is set, then compress the tree, and re-write the 
   tile to the header */

    if ( compressFlag) {
	STATBUF 	buf;
	seekAddress	length;

	for ( ii = 0; ii < numDstFiles; ii++) {
	    losses[ ii] = 0;
	}
	AmCompressTree( rootTile, losses, outputFds);

/* find the size for each of the files and truncate it by the number
   of bytes lost */

	for ( ii = 0; ii < numDstFiles; ii++) {
	    if ( FSTAT( outputFds[ ii], &buf) < 0) {
		fprintf( stderr, "failed to stat %s\n", filenames[ ii]);
		exit( -1);
	    }
	    length = buf.st_size - losses[ ii];
	    if ( ftruncate( outputFds[ ii], length) < 0) {
		fprintf( stderr, "failed to truncate %s\n", filenames[ ii]);
		exit( -1);
	    }
	}
    } 

/* write the tile information out to the tile header file */

    AmWriteTileInfo( tileHeader, filenames, numDstFiles, rootTile);
    AmReadAndPrintTileInfo( tileHeader);

/* close the file descriptors for the output files */

    for ( ii = 0; ii < numDstFiles; ii++) {
	close( outputFds[ ii]);
    }

/* write out the header */

    if (( hdrFp = fopen( header, "w")) == NULL) {
	fprintf( stderr, "failed to open output header %s\n", header);
	exit( -1);
    }
    fprintf( hdrFp, "xmin:		%.2f\n", xmin);
    fprintf( hdrFp, "ymin:		%.2f\n", ymin);
    fprintf( hdrFp, "xmax:		%.2f\n", xmax);
    fprintf( hdrFp, "ymax:		%.2f\n", ymax);
    fprintf( hdrFp, "rows:		%d\n", maxRows);
    fprintf( hdrFp, "cols:		%d\n", maxCols);
    fprintf( hdrFp, "sourceType:	pyramidSource\n");
    fprintf( hdrFp, "filename:		%s\n", AmGetRootName( tileHeader));
    if ( compressFlag) {
	fprintf( hdrFp, "note:		compressed\n");
    }
    fprintf( hdrFp, "note:		pyramid\n");
    AmDumpHeaderExcept( fileInfo[ 0].header, imageFields, hdrFp);
    fclose( hdrFp);
}
