#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <anvil/datatype.h>
#include <anvil/rect.h>
#include <anvil/fns.h>
#include <anvil/strlib.h>
#if defined(_WIN32)
#include <anvil/ntresource.h>
#else
#include <anvil/xresource.h>
#endif
#include <anvil/newtile.h>
#include <anvil/macros.h>
#include <anvil/tagstrs.h>
#include <anvil/image.h>
#include <anvil/imageP.h>
#include <anvil/compression.h>
#include <anvil/datafmt.h>
#include <anvil/files.h>
#include <anvil/quadnet.h>
#include <anvil/envdefs.h>
#include <anvil/choppath.h>
#include <util/utils.h>
#include <util/misc.h>
#include <util/tile.h>

#define FUDGE 	0.1

typedef struct fileInfo_ {
    char		*header;	/* source header file */
    char		*file;		/* source data file */
    float		xmin;		/* extents of file */
    float		ymin;
    float		xmax;
    float		ymax;
    int			rows;		/* data dimensions of file */
    int			fd;		/* file descriptor */
    AmDataTypeEnum	lookupType;	/* lookup value for filtering vector */
    int			lookupOffset;	/* lookup value offset */
    int			size;		/* file element size */
} fileInfoRec, *fileInfoPtr;

const int MAXFILES = 256;
#if defined (_WIN32)
const int BLOCKSIZE = 512*1024;
#else /* if it's a *REAL* machine */
const int BLOCKSIZE = 2048*1024;
#endif

static int byExtents( fileInfoPtr	a,
		      fileInfoPtr	b)
{ 
    float	aextents, bextents;

    aextents = MIN(( a->xmax - a->xmin), ( a->ymax - a->ymin));
    bextents = MIN(( b->xmax - b->xmin), ( b->ymax - b->ymin));
    if ( aextents < bextents) {
	return( 1);
    } else if ( aextents == bextents) {
	return( 0);
    } else {
	return( -1);
    }
}

#if 0
/* does a fileInfo contain the requested extents? */

static AnvilBoolean containFile( fileInfoPtr fileInfo,
                            float       xmin,
                            float       ymin,
                            float       xmax,
                            float       ymax)
{
    return(( fileInfo->xmin >= xmin) && ( fileInfo->ymin >= ymin) &&
           ( fileInfo->xmax <= xmax) && ( fileInfo->ymax <= ymax));
}
#endif


/* does a fileInfo overlap the requested extents? */

static AnvilBoolean overlapFile( fileInfoPtr fileInfo,
                            float       xmin,
                            float       ymin,
                            float       xmax,
                            float       ymax)
{
    return(( fileInfo->xmin < xmax) && ( fileInfo->ymin < ymax) &&
           ( fileInfo->xmax > xmin) && ( fileInfo->ymax > ymin));
}

/* figure out which tile to assign an x,y coordinate to. */

void allocateVectorRecursive( float		xmin,
			      float		ymin,
			      float		xmax,
			      float		ymax,
			      float		x,
			      float		y,
			      memTilePtr	tile,
			      int		depth)
{
    if ( depth == 0) {
	return;
    } else if ( tile == NULL) {
	return;
    } else if ( depth == 1) {
	tile->fileTile.rows++;
    } else {
	float	cx = ( xmax + xmin)/2, cy = ( ymax + ymin)/2;

/* recurse based on the quadrant that the line goes in */

	if ( y > cy) {
	    if ( x > cx) {
		allocateVectorRecursive( cx, cy, xmax, ymax, x, y, 
					 tile->urChild, depth - 1);
	    } else {
		allocateVectorRecursive( xmin, cy, cx, ymax, x, y,
					 tile->ulChild, depth - 1);
	    }
	} else {
	    if ( x > cx) {
		allocateVectorRecursive( cx, ymin, xmax, cy, x, y, 
					 tile->lrChild, depth - 1);
	    } else {
		allocateVectorRecursive( xmin, ymin, cx, cy, x, y,
					 tile->llChild, depth - 1);
	    }
	} 
    }
}

/* find out how big each of the tiles has to be to contain the filtered 
   vectors */

AmErrorCode sizeTiles( fileInfoPtr	sourceFile,
		       int		**layerFilterIdVec,
		       int		*numFilterIdsVec,
		       int		numFilters,
		       memTilePtr	rootTile)
{
    int		blockSize, numBytes, numRead, ii, jj, sourceFd, filterCount;
    uchar_t	*block;
    float	x, y;

/* allocate the read-block, then open the file. */

    blockSize = sourceFile->size*( BLOCKSIZE/sourceFile->size);
    if (( block = ( uchar_t *) alloca( blockSize)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    if (( sourceFd = open( sourceFile->file, O_RDONLY)) < 0) {
	fprintf( stderr, "failed to open source file %s\n", sourceFile->file);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* read in blocks */

    do {
	numBytes = read( sourceFd, block, blockSize);
AN_DEBUG( 20, ( "read %d bytes\n", numBytes));
	if ( numBytes < 0) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	if ( numBytes == 0) {
	    break;
	}
	numRead = numBytes/sourceFile->size;

/* go backwards, because the successive filters cull the set. */

	for ( ii = numFilters - 1; ii >= 0; ii--) {

/* if we're not at the bottom of the tree, and the filter is the same as the 
   AmLast filter, then don't bother allocating the tile layer, because it will
   get written to lower layers, and the current layer will simply point to
   them */

	    if (( ii == numFilters - 1) || 
		!AmSameFilter( layerFilterIdVec[ ii], 
			     numFilterIdsVec[ ii],
			     layerFilterIdVec[ ii + 1], 
			     numFilterIdsVec[ ii + 1])) {
		if ( layerFilterIdVec[ ii] != NULL) {
		    filterCount = AmFilterVec( block, numRead, sourceFile->size,
					     sourceFile->lookupType, 
					     sourceFile->lookupOffset,
					     layerFilterIdVec[ ii], 
					     numFilterIdsVec[ ii]);
AN_DEBUG( 20, ("filtered from %d to %d\n", numRead, filterCount));
		    numRead = filterCount;
		}

AN_DEBUG( 20, ("assigning %d vectors\n", numRead));

/* loop over each of the lines and assign it to the appropriate tile... */

		for ( jj = 0; jj < numRead; jj++) {
		    x = (( AmLinePtr) &block[ jj*sourceFile->size])->x0;
		    y = (( AmLinePtr) &block[ jj*sourceFile->size])->y0;
		    allocateVectorRecursive( sourceFile->xmin, 
					     sourceFile->ymin,
					     sourceFile->xmax, 
					     sourceFile->ymax,
					     x, y, rootTile, ii + 1);
		}
	    }
	} 
    } while ( TRUE);
    return ( AmOK);
}

/* take the block of vectors, and use the index vector to write the tiles
   into the sorted block of vectors */

void reindexVectorRecursive( float 	xmin, 
			     float	ymin,
			     float	xmax,
			     float	ymax,
			     int	index,
			     int	*indices,
			     int	size,
			     uchar_t	*block,
			     uchar_t	*sortBlock,
			     float	x,
			     float	y,
			     memTilePtr	tile,
			     int	depth)
{
    if ( depth == 0) {
	return;
    } else if ( tile == NULL) {
	return;
    } else if ( depth == 1) {
	memcpy( &sortBlock[ indices[ tile->tileIndex]*size], 
		&block[ index*size], size);
	indices[ tile->tileIndex]++;
    } else {
	float	cx = ( xmax + xmin)/2, cy = ( ymax + ymin)/2;

/* recurse based on the quadrant that the line goes in */

	if ( y > cy) {
	    if ( x > cx) {
		reindexVectorRecursive( cx, cy, xmax, ymax, index, indices,
					size, block, sortBlock, x, y, 
				        tile->urChild, depth - 1);
	    } else {
		reindexVectorRecursive( xmin, cy, cx, ymax, index, indices,
					size, block, sortBlock, x, y, 
				        tile->ulChild, depth - 1);
	    }
	} else {
	    if ( x > cx) {
		reindexVectorRecursive( cx, ymin, xmax, cy, index, indices,
					size, block, sortBlock, x, y, 
				        tile->lrChild, depth - 1);
	    } else {
		reindexVectorRecursive( xmin, ymin, cx, cy, index, indices,
					size, block, sortBlock, x, y, 
				        tile->llChild, depth - 1);
	    }
	} 
    }
}

/* recursively assign a vector to a tile, and find out which sum it will
   be assigned to */

void assignVectorRecursive( float 	xmin, 
			    float	ymin,
			    float	xmax,
			    float	ymax,
			    int		*sums,
			    float	x,
			    float	y,
			    memTilePtr	tile,
			    int		depth)
{
    if ( depth == 0) {
	return;
    } else if ( tile == NULL) {
	return;
    } else if ( depth == 1) {
	sums[ tile->tileIndex]++;
    } else {
	float	cx = ( xmax + xmin)/2, cy = ( ymax + ymin)/2;

/* recurse based on the quadrant that the line goes in */

	if ( y > cy) {
	    if ( x > cx) {
		assignVectorRecursive( cx, cy, xmax, ymax, sums, x, y, 
				       tile->urChild, depth - 1);
	    } else {
		assignVectorRecursive( xmin, cy, cx, ymax, sums, x, y,
				       tile->ulChild, depth - 1);
	    }
	} else {
	    if ( x > cx) {
		assignVectorRecursive( cx, ymin, xmax, cy, sums, x, y, 
				       tile->lrChild, depth - 1);
	    } else {
		assignVectorRecursive( xmin, ymin, cx, cy, sums, x, y,
				       tile->llChild, depth - 1);
	    }
	} 
    }
}


/* take the tree, sorted block, element size, indices into the sort block 
   (based on tile index), sums to be written to each tile, and the
   list of output file descriptors, and write the tile tree to the output
   file */

AmErrorCode writeVectorsRecursive( memTilePtr	tile,
				   uchar_t	*sortBlock,
				   int		size,
				   int		*indices,
				   int		*sums,
				   int		*outputFds)
{
    AmErrorCode	errCode;

    if ( tile == NULL) {
	return( AmOK);
    }
    if ( sums[ tile->tileIndex] != 0) {
	if ( LSEEK( outputFds[ tile->fileTile.fileIndex], tile->currSeekAddr,
		    SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
AN_DEBUG( 99, ("writing %d bytes from index %d\n", 
		sums[ tile->tileIndex]*size, indices[ tile->tileIndex]));
	if ( write( outputFds[ tile->fileTile.fileIndex], 
		    &sortBlock[ indices[ tile->tileIndex]*size],
		    sums[ tile->tileIndex]*size) < 0) {
	    DO_RETURN_ERROR( AmWRITE_FAILED);
	}
	tile->currSeekAddr += sums[ tile->tileIndex]*size;
    } 
    errCode = writeVectorsRecursive( tile->llChild, sortBlock, size,
				     indices, sums, outputFds);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeVectorsRecursive( tile->ulChild, sortBlock, size,
				     indices, sums, outputFds);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeVectorsRecursive( tile->lrChild, sortBlock, size,
				     indices, sums, outputFds);
    if ( errCode != AmOK) {
	return( errCode);
    }
    errCode = writeVectorsRecursive( tile->urChild, sortBlock, size,
				     indices, sums, outputFds);
    if ( errCode != AmOK) {
	return( errCode);
    }
    return( AmOK);
}

/* find out how big each of the tiles has to be to contain the filtered 
   vectors */

AmErrorCode writeDataToTiles( fileInfoPtr	sourceFile,
			      int		*outputFds,
			      int		**layerFilterIdVec,
			      int		*numFilterIdsVec,
			      int		numFilters,
			      int		tileCount,
			      memTilePtr	rootTile)
{
    int		numRead, ii, jj, sourceFd, numBytes, filterCount,
    		blockSize, *indices, *runIndices, *sums;
    uchar_t	*block, *sortBlock;
    float	x, y;

/* open the source file */

    if (( sourceFd = open( sourceFile->file, O_RDONLY)) < 0) {
	fprintf( stderr, "failed to open source file %s\n", sourceFile->file);
	DO_RETURN_ERROR( AmOPEN_FAILED);
    }

/* allocate the read-block, then open the file. */

    blockSize = sourceFile->size*( BLOCKSIZE/sourceFile->size);
    if (( block = ( uchar_t *) alloca( blockSize)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    if (( sortBlock = ( uchar_t *) alloca( blockSize)) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }

/* allocate a vector to hold the number of vectors to store in each tile,
   and another vector to store the indices within the source block for
   the tiles. */

    if (( sums = ( int *) alloca( tileCount*sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    if (( indices = ( int *) alloca( tileCount*sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }
    if (( runIndices = ( int *) alloca( tileCount*sizeof( int))) == NULL) {
	DO_RETURN_ERROR( AmALLOCA_FAILED);
    }

    do {

/* read in blocks */

	numBytes = read( sourceFd, block, blockSize);
AN_DEBUG( 20, ( "read %d bytes\n", numBytes));
	if ( numBytes < 0) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	if ( numBytes == 0) {
	    break;
	}
	numRead = numBytes/sourceFile->size;

/* go backwards, because the successive filters cull the set. */

	for ( ii = numFilters - 1; ii >= 0; ii--) {

/* clear out the indices and sums */
	
	    memset( sums, 0, tileCount*sizeof( int));
	    memset( indices, 0, tileCount*sizeof( int));
	    memset( runIndices, 0, tileCount*sizeof( int));

/* apply the filter if there is one */

	    if ( layerFilterIdVec[ ii] != NULL) {
		filterCount = AmFilterVec( block, numRead, sourceFile->size,
					 sourceFile->lookupType, 
					 sourceFile->lookupOffset,
					 layerFilterIdVec[ ii], 
					 numFilterIdsVec[ ii]);
AN_DEBUG( 20, ("filtered from %d to %d\n", numRead, filterCount));
		numRead = filterCount;
	    }

/* if we're not at the bottom of the tree, and the filter is the same as the 
   AmLast filter, then don't bother allocating the tile layer, because it will
   get written to lower layers, and the current layer will simply point to
   them */

	    if (( ii == numFilters - 1) || 
		!AmSameFilter( layerFilterIdVec[ ii], 
			     numFilterIdsVec[ ii],
			     layerFilterIdVec[ ii + 1], 
			     numFilterIdsVec[ ii + 1])) {

/* loop over each of the lines and assign it to the appropriate tile... */

		for ( jj = 0; jj < numRead; jj++) {
		    x = (( AmLinePtr) &block[ jj*sourceFile->size])->x0;
		    y = (( AmLinePtr) &block[ jj*sourceFile->size])->y0;
		    assignVectorRecursive( sourceFile->xmin, sourceFile->ymin,
					   sourceFile->xmax, sourceFile->ymax,
					   sums, x, y, rootTile, ii + 1);
		}
		AmSumsToIndices( sums, indices, tileCount);
		memcpy( runIndices, indices, tileCount*sizeof( int));
		for ( jj = 0; jj < numRead; jj++) {
		    x = (( AmLinePtr) &block[ jj*sourceFile->size])->x0;
		    y = (( AmLinePtr) &block[ jj*sourceFile->size])->y0;
		    reindexVectorRecursive( sourceFile->xmin, sourceFile->ymin,
					    sourceFile->xmax, sourceFile->ymax,
					    jj, runIndices, sourceFile->size,
					    block, sortBlock,
					    x, y, rootTile, ii + 1);
		}
		writeVectorsRecursive( rootTile, sortBlock, sourceFile->size, 
				       indices, sums, outputFds);
	    }
	}
    } while ( TRUE);
    return ( AmOK);
}

/* now that we know the number of elements per tile, assign the seek addresses 
   and calculate the number of bytes */

void setTileSeekAddrs( memTilePtr	tile,
		       int		elementSize,
		       seekAddress	*seekAddrs,
		       int		*count,
		       int		numOutputFiles)
{
    if ( tile == NULL) {
	return;
    }
    tile->fileTile.fileIndex = *count % numOutputFiles;
    tile->fileTile.seekAddr = seekAddrs[ tile->fileTile.fileIndex];
    tile->currSeekAddr = tile->fileTile.seekAddr;
    tile->fileTile.numBytes = tile->fileTile.rows*elementSize;
    tile->fileTile.numBytesUncompressed = tile->fileTile.rows*elementSize;
    tile->tileIndex = *count;
    seekAddrs[ tile->fileTile.fileIndex] += tile->fileTile.numBytes;
    ( *count)++;
    setTileSeekAddrs( tile->llChild, elementSize, seekAddrs, 
		      count, numOutputFiles);
    setTileSeekAddrs( tile->ulChild, elementSize, seekAddrs, 
		      count, numOutputFiles);
    setTileSeekAddrs( tile->lrChild, elementSize, seekAddrs, 
		      count, numOutputFiles);
    setTileSeekAddrs( tile->urChild, elementSize, seekAddrs, 
		      count, numOutputFiles);
}

/* take the ASCII list of files from the list file, and turn them into
   a vector of fileInfo structs */

static AnvilBoolean extractFileInfo( char		*listFile,
				fileInfoPtr	fileInfo,
				int		*numFiles,
				int		*elementSize)
{
    FILE	        *fp;
    char	        buffer[ BUFFERLEN], header[ PATHLEN], file[ PATHLEN],
		        *lookupTypeStr, *kind;
    AnvilBoolean	        okFlag;

    *numFiles = 0;
    if (( fp = fopen( listFile, "r")) == NULL) {
	fprintf( stderr, "failed to open %s\n", listFile);
	return( FALSE);
    }
    okFlag = TRUE;
    do {
	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
	    break;
	}
	sscanf( buffer, "%s %s", header, file);
	fileInfo[ *numFiles].header = strdup( header);
	fileInfo[ *numFiles].file = strdup( file);
	fileInfo[ *numFiles].fd = 0;

#if defined(_WIN32)
	AmNTResourceDbClass db( header);
#else
	AmXResourceDbClass db( header);
#endif

	if ( !db.IsValid()) {
	    fprintf( stderr, "header %s is not valid\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNtype, &kind)) {
	    fprintf( stderr, "failed to get kind from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNxmin, &fileInfo[ *numFiles].xmin)) {
	    fprintf( stderr, "failed to get xmin from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNymin, &fileInfo[ *numFiles].ymin)) {
	    fprintf( stderr, "failed to get ymin from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNxmax, &fileInfo[ *numFiles].xmax)) {
	    fprintf( stderr, "failed to get xmax from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNymax, &fileInfo[ *numFiles].ymax)) {
	    fprintf( stderr, "failed to get ymax from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNrows, &fileInfo[ *numFiles].rows)) {
	    fprintf( stderr, "failed to get rows from %s\n", header);
	    okFlag = FALSE;
	    break;
	}
	if ( !db.GetResource( AmNsize, &fileInfo[ *numFiles].size)) {
	    fprintf( stderr, "failed to get size from %s\n", header);
	    okFlag = FALSE;
	    break;
	}

/* get the size from the header and make sure everyone's the same size */

	if ( *numFiles == 0) {
	    *elementSize = fileInfo[ *numFiles].size;
	} else if ( fileInfo[ *numFiles].size != *elementSize) {
	    fprintf( stderr, "element size for %s is %d, not %d\n",
		     header, fileInfo[ *numFiles].size, *elementSize);
	    okFlag = FALSE;
	    break;
	}

/* get the lookup type and lookup offset from the header */

        if ( !db.GetResource( AmNcolorsType, &lookupTypeStr)) {
            fprintf( stderr, "failed to read lookup type from file\n");
	    okFlag = FALSE;
	    break;
        }
        fileInfo[ *numFiles].lookupType = AmStrToDataTypeEnum( lookupTypeStr);
        if ( !db.GetResource( AmNcolorsOffset, 
			      &fileInfo[ *numFiles].lookupOffset)) {
            fprintf( stderr, "failed to read lookup offset from file\n");
	    okFlag = FALSE;
	    break;
        }
	( *numFiles)++;
    } while ( okFlag);
    fclose( fp);
    return( okFlag);
}

/* expand the tiles to reflect their contents */

AmErrorCode expandTiles( memTilePtr	tile,
			 int		elementSize,
			 int		*outputFds)
{
    if ( tile == NULL) {
	return( AmOK);
    } else {
	int		fd, ii;
	uchar_t		*buffer;
	AmErrorCode	errCode;
	AmLinePtr		line;

	fd = outputFds[ tile->fileTile.fileIndex];
	if ( LSEEK( fd, tile->fileTile.seekAddr, SEEK_SET) < 0) {
	    DO_RETURN_ERROR( AmSEEK_FAILED);
	}
	if (( buffer = ( uchar_t *) 
			alloca( tile->fileTile.numBytes)) == NULL) {
	    DO_RETURN_ERROR( AmALLOCA_FAILED);
	}
	if ( read( fd, buffer, tile->fileTile.numBytes) < 0) {
	    DO_RETURN_ERROR( AmREAD_FAILED);
	}
	for ( ii = 0; ii < tile->fileTile.rows; ii++) {
	    line = ( AmLinePtr) &buffer[ ii*elementSize];
	    if ( line->x0 < tile->fileTile.xmin) {
		tile->fileTile.xmin = line->x0;
	    }
	    if ( line->y0 < tile->fileTile.ymin) {
		tile->fileTile.ymin = line->y0;
	    }
	    if ( line->x1 < tile->fileTile.xmin) {
		tile->fileTile.xmin = line->x1;
	    }
	    if ( line->y1 < tile->fileTile.ymin) {
		tile->fileTile.ymin = line->y1;
	    }
	    if ( line->x0 > tile->fileTile.xmax) {
		tile->fileTile.xmax = line->x0;
	    }
	    if ( line->y0 > tile->fileTile.ymax) {
		tile->fileTile.ymax = line->y0;
	    }
	    if ( line->x1 > tile->fileTile.xmax) {
		tile->fileTile.xmax = line->x1;
	    }
	    if ( line->y1 > tile->fileTile.ymax) {
		tile->fileTile.ymax = line->y1;
	    }
	}
	errCode = expandTiles( tile->llChild, elementSize, outputFds);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = expandTiles( tile->ulChild, elementSize, outputFds);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = expandTiles( tile->lrChild, elementSize, outputFds);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = expandTiles( tile->urChild, elementSize, outputFds);
	return( errCode);
    }
}

/* recursively reduce the vectors in the tile tree */

AmErrorCode reduceVectorsRecursive( memTilePtr	tile,
				    int		*outputFds,
				    int		elementSize,
				    seekAddress	*losses,
				    float	vectorTolerance,
				    float	arcTolerance,
				    float	nodeTolerance)
{
    if ( tile == NULL) {
	return( AmOK);
    } else {
	int		fd, AmNumLines;
	uchar_t		*buffer;
	AmErrorCode	errCode;

/* seek to and read the tile */

	fd = outputFds[ tile->fileTile.fileIndex];
	if ( tile->fileTile.rows > 0) {
	    if ( LSEEK( fd, tile->fileTile.seekAddr, SEEK_SET) < 0) {
		DO_RETURN_ERROR( AmSEEK_FAILED);
	    }
	    if (( buffer = ( uchar_t *) 
			    alloca( tile->fileTile.numBytes)) == NULL) {
		DO_RETURN_ERROR( AmALLOCA_FAILED);
	    }
	    if ( read( fd, buffer, tile->fileTile.numBytes) < 0) {
		DO_RETURN_ERROR( AmREAD_FAILED);
	    }

/* reduce the tile */

	    AmNumLines = tile->fileTile.rows;
	    AmReduceLines(( AmLinePtr) buffer, &AmNumLines, elementSize,
			tile->fileTile.xmin, tile->fileTile.ymin,
			tile->fileTile.xmax, tile->fileTile.ymax,
			vectorTolerance, arcTolerance, nodeTolerance, TRUE);

/* seek to the tile's new location and re-write it */

	    tile->fileTile.seekAddr -= losses[ tile->fileTile.fileIndex];
	    if ( LSEEK( fd, tile->fileTile.seekAddr, SEEK_SET) < 0) {
		DO_RETURN_ERROR( AmSEEK_FAILED);
	    }

/* add the tile reduction to the losses, and reset the # of rows and bytes */

	    losses[ tile->fileTile.fileIndex] += 
		( tile->fileTile.rows - AmNumLines)*elementSize;
	    tile->fileTile.rows = AmNumLines;
	    tile->fileTile.numBytes = AmNumLines*elementSize;
	    tile->fileTile.numBytesUncompressed = AmNumLines*elementSize;
	    if ( write( fd, buffer, tile->fileTile.numBytes) < 0) {
		DO_RETURN_ERROR( AmWRITE_FAILED);
	    }
	}

/* recursively reduce each of the child tiles */

	errCode = reduceVectorsRecursive( tile->llChild, outputFds,
					  elementSize, losses,
					  vectorTolerance/2, arcTolerance/2, 
					  nodeTolerance/2);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = reduceVectorsRecursive( tile->ulChild, outputFds,
					  elementSize, losses,
					  vectorTolerance/2, arcTolerance/2, 
					  nodeTolerance/2);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = reduceVectorsRecursive( tile->lrChild, outputFds,
					  elementSize, losses,
					  vectorTolerance/2, arcTolerance/2, 
					  nodeTolerance/2);
	if ( errCode != AmOK) {
	    return( errCode);
	}
	errCode = reduceVectorsRecursive( tile->urChild, outputFds,
					  elementSize, losses,
					  vectorTolerance/2, arcTolerance/2, 
					  nodeTolerance/2);
	return( errCode);
    }
}

/* recursively create the in-memory tile tree */

memTilePtr makeTileTree( float          xmin,
                         float          ymin,
                         float          xmax,
                         float          ymax,
                         fileInfoPtr    fileInfo,
                         int            numFiles,
                         int            depth,
                         seekAddress    *seekAddrs,
                         int            numOutputFiles,
                         int            *tileCount)
{
    float       fileExtents, bestExtents, cx, cy, extents;
    int         bestIndex, ii, fileIndex;
    memTilePtr  tile;

/* we're done if we've reached the maximum depth */

    if ( depth == 0) {
        return( NULL);
    }

/* find the smallest file which contains the required extents */

    bestExtents = MY_HUGE;
    bestIndex = -1;
    extents = MIN( xmax - xmin, ymax - ymin);
    for ( ii = 0; ii < numFiles; ii++) {
        if ( overlapFile( &fileInfo[ ii], xmin, ymin, xmax, ymax)) {
            fileExtents = MIN( fileInfo[ ii].xmax - fileInfo[ ii].xmin,
                               fileInfo[ ii].ymax - fileInfo[ ii].ymin);
            if (( fileExtents >= extents) && ( fileExtents < bestExtents)) {
                bestIndex = ii;
                bestExtents = fileExtents;
            }
        }
    }

/* if we didn't find a file with a tighter resolution than requested, then
   we don't create a tile, because there's no sense in supersampling */

    if ( bestIndex == -1) {
        return( NULL);
    }

/* distribute the tiles amongst the different files */

    fileIndex = *tileCount % numOutputFiles;

/* otherwise, create the tile pointer and recurse */

    tile = ( memTilePtr) malloc( sizeof( memTileRec));
    tile->fileTile.rows = 0;
    tile->fileTile.cols = 0;
    tile->fileTile.numBytes = 0;
    tile->fileTile.numBytesUncompressed = 0;
    tile->fileTile.xmin = xmin;
    tile->fileTile.ymin = ymin;
    tile->fileTile.xmax = xmax;
    tile->fileTile.ymax = ymax;
    tile->filled = FALSE;
    tile->fileTile.flags = 0;
#if defined(SEEK64)
    tile->flags |= AmTILE_64BIT_SEEK_FLAG;
#endif

/* the tile's seek address is set in the AmFirst filtering pass */

    tile->fileTile.seekAddr = 0;
    tile->fileTile.fileIndex = fileIndex;

/* right now, the file indices are set to 0, but later, they'll index the
   list of output files. */

    tile->fileTile.fileIndex = fileIndex;

/* recurse, dividing the tile by 4 from the center. */

    cx = ( xmax + xmin)/2;
    cy = ( ymax + ymin)/2;
    ( *tileCount)++;
    AN_DEBUG( 10, ( "file index = %d, tile Count = %d\n",
                    fileIndex, *tileCount));
    tile->llChild = makeTileTree( xmin, ymin, cx, cy, fileInfo,
                                  numFiles, depth - 1, seekAddrs,
                                  numOutputFiles, tileCount);
    tile->ulChild = makeTileTree( xmin, cy, cx, ymax, fileInfo,
                                  numFiles, depth - 1, seekAddrs,
                                  numOutputFiles, tileCount);
    tile->lrChild = makeTileTree( cx, ymin, xmax, cy, fileInfo,
                                  numFiles, depth - 1, seekAddrs,
                                  numOutputFiles, tileCount);
    tile->urChild = makeTileTree( cx, cy, xmax, ymax, fileInfo,
                                  numFiles, depth - 1, seekAddrs,
                                  numOutputFiles, tileCount);
    return( tile);
}

	   

int main( int argc, char *argv[])
{
    fileInfoRec fileInfo[ MAXFILES];
    int		numSrcFiles, ii, depth, elementSize, numDstFiles, tileCount, 
		outputFds[ MAXFILES], **filterIdVec, *numFilterIdVec;
    float	xmin, ymin, xmax, ymax, vectorTolerance=0, arcTolerance=0, 
		nodeTolerance=0;
    memTilePtr	rootTile;
    seekAddress	seekAddrs[ MAXFILES], losses[ MAXFILES];
    char	tileHeader[ PATHLEN], *filenames[ MAXFILES], filebuf[ PATHLEN],
		header[ PATHLEN], 
		*imageFields[] = { AmNfilename, AmNsourceType,
				   AmNdescription, 
				   AmNxmin, AmNymin, AmNxmax, AmNymax,
				   AmNrows, AmNcols, NULL },
		**filterFiles;
    const char	*path;
    FILE	*hdrFp;
    AnvilBoolean	compressFlag = FALSE, reduceFlag = FALSE;

    if ( argc < 4) {
	fprintf( stderr, 
"usage %s file-list outfile depth filter-files [numDstFiles=1] [-compress] ", 
		 argv[ 0]);
	fprintf( stderr, 
"[-reduce vector_tolerance arc_tolerance node_tolerance]\n");
	exit( -1);
    }
    if ( !extractFileInfo( argv[ 1], fileInfo, &numSrcFiles, &elementSize)) {
	fprintf( stderr, "failed to read header files\n");
	exit( -1);
    }

/* sort the file so the ones with the loosest resolution are generated AmFirst */

    qsort( fileInfo, numSrcFiles, sizeof( fileInfoRec), ( AmSearchFn) byExtents);
    depth = atoi( argv[ 3]);
    if (( filterFiles = AmReadStrvec( argv[ 4])) == NULL) {
	fprintf( stderr, "failed to read filter files from %s\n", argv[ 4]);
	exit( -1);
    }
    numDstFiles = 1;

/* test the rest of the optional arguments */

    ii = 5;
    while ( ii < argc) {
	
/* test if the argument is -compress */

	if ( !strcmp( argv[ ii], "-compress")) {
	    compressFlag = TRUE;
	    ii++;
	} else if ( !strcmp( argv[ ii], "-reduce")) {
	    if ( argc - ii < 3) {
		fprintf( stderr, "-reduce must be followed by 3 tolerances\n");
		exit( -1);
	    }
	    vectorTolerance = atof( argv[ ii + 1]);
	    nodeTolerance = atof( argv[ ii + 2]);
	    arcTolerance = atof( argv[ ii + 3]);
	    reduceFlag = TRUE;
	    ii += 4;
	} else {
	    numDstFiles = atoi( argv[ 5]);
	    ii++;
	}
    } 
    if (( numDstFiles < 1) || ( numDstFiles > 10)) {
        fprintf( stderr, "bad number of destination files %d\n", numDstFiles);
        exit( -1);
    }


/* create the output filenames */

    sprintf( tileHeader, "%s.tileHdr", argv[ 2]);
    for ( ii = 0; ii < numDstFiles; ii++) {
	sprintf( filebuf, "%s.%d.dat", argv[ 2], ii);
	filenames[ ii] = strdup( filebuf);
    }
    sprintf( header, "%s.hdr", argv[ 2]);

/* find the extents of the input files */

    xmin = MY_HUGE;
    ymin = MY_HUGE;
    xmax = -MY_HUGE;
    ymax = -MY_HUGE;
    for ( ii = 0; ii < numSrcFiles; ii++) {
	if ( fileInfo[ ii].xmin < xmin) {
	    xmin = fileInfo[ ii].xmin;
	}
	if ( fileInfo[ ii].ymin < ymin) {
	    ymin = fileInfo[ ii].ymin;
	}
	if ( fileInfo[ ii].xmax > xmax) {
	    xmax = fileInfo[ ii].xmax;
	}
	if ( fileInfo[ ii].ymax > ymax) {
	    ymax = fileInfo[ ii].ymax;
	}
    }

/* if the user specified the tolerances for node and vector reduction, then
   make sure that they're something reasonable... */

    if ( reduceFlag) {
	if ( nodeTolerance > ( xmax - xmin)/1000.0) {
	    fprintf( stderr, "a node tolerance of %f is probably too large\n",
		     nodeTolerance);
	    nodeTolerance = ( xmax - xmin)/10000.0;
	    fprintf( stderr, "resetting node tolerance to %f\n", nodeTolerance);
	} else if ( nodeTolerance < ( xmax - xmin)/1000000.0) {
	    fprintf( stderr, "a node tolerance of %f is probably too small\n",
		     nodeTolerance);
	    nodeTolerance = ( xmax - xmin)/10000.0;
	    fprintf( stderr, "resetting node tolerance to %f\n", nodeTolerance);
	}
	if ( arcTolerance < 1.0) {
	    fprintf( stderr, "an arc tolerance of %f is probably too small\n",
		     arcTolerance);
	    arcTolerance = 1.0;
	    fprintf( stderr, "resetting arc tolerance to %f\n", arcTolerance);
	} else if ( arcTolerance > 10.0) {
	    fprintf( stderr, "an arc tolerance of %f is probably too large\n",
		     arcTolerance);
	    arcTolerance = 10.0;
	    fprintf( stderr, "resetting arc tolerance to %f\n", arcTolerance);
	}
    }

/* initialize the seek addresses for each of the output files */

    for ( ii = 0; ii < numDstFiles; ii++) {
	seekAddrs[ ii] = 0;
    }
    tileCount = 0;
    rootTile = makeTileTree( xmin, ymin, xmax, ymax, fileInfo, numSrcFiles, 
			     depth, seekAddrs, numDstFiles, &tileCount);

/* read the vectors of filter id's and their lengths */

    filterIdVec = ( int **) malloc( sizeof( int *)*depth);
    numFilterIdVec = ( int *) malloc( sizeof( int)*depth);
    for ( ii = 0; ii < AmVeclen(( const char **) filterFiles); ii++) {
	path = AmGetEnvSubdirFile( CONFIG_ENV, AM_FILTER_DIR, filterFiles[ ii]);
	if (( filterIdVec[ ii] = AmReadIntVec( path)) == NULL) {
	    fprintf( stderr, "failed to read filter ID file %s\n", 
		     filterFiles[ ii]);
	    exit( -1);
	}
	numFilterIdVec[ ii] = AmNumLines( path);
    }
    for ( ii = AmVeclen(( const char **) filterFiles); ii < depth; ii++) {
	filterIdVec[ ii] = NULL;
	numFilterIdVec[ ii] = 0;
    }

/* figure out how big the tiles have to be... */

    for ( ii = 0; ii < numSrcFiles; ii++) {
	if ( sizeTiles( &fileInfo[ ii], filterIdVec, numFilterIdVec, depth,
			rootTile) != AmOK) {
	    fprintf( stderr, "failed to size tiles\n");
	    exit( -1);
	}
    }
    for ( ii = 0; ii < numDstFiles; ii++) {
	seekAddrs[ ii] = 0;
    }
    tileCount = 0;
    setTileSeekAddrs( rootTile, fileInfo[ 0].size, seekAddrs, &tileCount,
		      numDstFiles);
    for ( ii = 0; ii < numDstFiles; ii++) {
	if (( outputFds[ ii] = 
		    open( filenames[ ii], O_CREAT | O_RDWR, PERMS)) < 0) {
	    fprintf( stderr, "failed to open output file %s\n", filenames[ii]);
	}
    }
    for ( ii = 0; ii < numSrcFiles; ii++) {
	if ( writeDataToTiles( &fileInfo[ ii], outputFds,
			       filterIdVec, numFilterIdVec, 
			       depth, tileCount, rootTile) != AmOK) {
	    fprintf( stderr, "failed to write data to tiles\n");
	}
    }

/* expand the tiles to contain their contents */

    if ( expandTiles( rootTile, fileInfo[ 0].size, outputFds) != AmOK) {
	fprintf( stderr, "expandTiles failed\n");
	exit( -1);
    }

/* reduce the vectors if requested */

    if ( reduceFlag) {
	struct stat	buf;
	seekAddress	length;

	for ( ii = 0; ii < numDstFiles; ii++) {
	    losses[ ii] = 0;
	}
	if ( reduceVectorsRecursive( rootTile, outputFds, fileInfo[ 0].size,
				     losses, vectorTolerance, 
				     DEGTORAD( arcTolerance), 
				     nodeTolerance) != AmOK) {
	    fprintf( stderr, "reduceVectorsRecursive failed\n");
	    exit( -1);
	}

/* find the size for each of the files and truncate it by the number
   of bytes lost */

	for ( ii = 0; ii < numDstFiles; ii++) {
	    if ( fstat( outputFds[ ii], &buf) < 0) {
		fprintf( stderr, "failed to stat %s\n", filenames[ ii]);
		exit( -1);
	    }
	    length = buf.st_size - losses[ ii];
	    if ( ftruncate( outputFds[ ii], length) < 0) {
		fprintf( stderr, "failed to truncate %s\n", filenames[ ii]);
		exit( -1);
	    }
	}
    }

/* if compression is set, then compress the tree, and re-write the 
   tile to the header */

    if ( compressFlag) {
	struct stat	buf;
	seekAddress	length;

	for ( ii = 0; ii < numDstFiles; ii++) {
	    losses[ ii] = 0;
	}
	AmCompressTree( rootTile, losses, outputFds);

/* find the size for each of the files and truncate it by the number
   of bytes lost */

	for ( ii = 0; ii < numDstFiles; ii++) {
	    if ( fstat( outputFds[ ii], &buf) < 0) {
		fprintf( stderr, "failed to stat %s\n", filenames[ ii]);
		exit( -1);
	    }
	    length = buf.st_size - losses[ ii];
	    if ( ftruncate( outputFds[ ii], length) < 0) {
		fprintf( stderr, "failed to truncate %s\n", filenames[ ii]);
		exit( -1);
	    }
	}
    } 

/* write the tile information out to the tile header file */

    AmWriteTileInfo( tileHeader, filenames, numDstFiles, rootTile);
    AmReadAndPrintTileInfo( tileHeader);

/* close the file descriptors for the output files */

    for ( ii = 0; ii < numDstFiles; ii++) {
	close( outputFds[ ii]);
    }

/* write out the header */

    if (( hdrFp = fopen( header, "w")) == NULL) {
	fprintf( stderr, "failed to open output header %s\n", header);
	exit( -1);
    }
    fprintf( hdrFp, "xmin:		%.2f\n", xmin);
    fprintf( hdrFp, "ymin:		%.2f\n", ymin);
    fprintf( hdrFp, "xmax:		%.2f\n", xmax);
    fprintf( hdrFp, "ymax:		%.2f\n", ymax);
    fprintf( hdrFp, "sourceType:	pyramidSource\n");
    fprintf( hdrFp, "filename:		%s\n", AmGetRootName( tileHeader));
    if ( compressFlag) {
	fprintf( hdrFp, "note:		compressed\n");
    }
    fprintf( hdrFp, "note:		pyramid\n");
    AmDumpHeaderExcept( fileInfo[ 0].header, imageFields, hdrFp);
    fclose( hdrFp);
}
