#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <anvil/ratio.h>
#include <anvil/macros.h>
#include <anvil/datafmt.h>
#include <anvil/strlib.h>
#include <anvil/files.h>
#include <anvil/project.h>
#include <anvil/projects.h>
#include <anvil/tagstrs.h>
#if defined(_WIN32)
#include <anvil/ntresource.h>
#else
#include <anvil/xresource.h>
#endif
#include <anvil/envdefs.h>
#include <anvil/choppath.h>
#include <util/utils.h>
#include <util/bilin.h>

/* return the resulting units from the projection */

char *getProjectionUnits( char **args, AnvilBoolean invFlag)
{
   char	*units;

    if ( invFlag) {
	return( "arcdeg");
    }
    for ( ; *args != NULL; args++) {
        if ( !strncmp( *args, "units", strlen( "units"))) {
	    units = strchr( *args, '=') + 1;
	    if ( !strncasecmp( units, "m", 1)) {
		return( "meters");
	    } else if ( !strncasecmp( units, "km", 2)) {
		return( "km");
	    } else if ( !strncasecmp( units, "ft", 2)) {
		return( "ft");
	    } else if ( !strncasecmp( units, "yd", 2)) {
		return( "yd");
	    } else if ( !strncasecmp( units, "mi", 2)) {
		return( "mi");
	    }
	}
    }
    return( "meters");
}

static int MAXROW = -1, MINROW = 0;
static uchar_t **rowvec = NULL;

void initRowManager( int	numRows)
{ 
    MAXROW = -1;
    MINROW = 0;
    rowvec = ( uchar_t **) calloc( numRows, sizeof( uchar_t *));
}

/* get a specified row from the source matrix */

uchar_t *getRow( int	fd,
	         int	rownum,
	         int	rowlen)
{
    int		ii;
   
/* if the row is greater than the current largest row, then for 
   each row AmBetween the requested row and the current largest row,
   allocate the buffer, read the data, and insert it into the search table, 
   then return the requested row buffer.  */

    if ( rownum > MAXROW) {
	for ( ii = MAXROW + 1; ii <= rownum; ii++) {
	    rowvec[ ii] = ( uchar_t *) malloc( rowlen);
	    if ( AmReadLoopFile( fd, rowvec[ ii], rowlen) < rowlen) {
		fprintf( stderr, "failed to read row %d\n", ii);
	    }
	}
	MAXROW = rownum;
    } 
    return( rowvec[ rownum]);
}

/* free any of the buffers that are no longer in use */

void manageRows( int	currMinRow, 
		 int	pad)
{
    int		ii;

    for ( ii = MINROW; ii < currMinRow - pad; ii++) {
	if ( rowvec[ ii] != NULL) {
	    free( rowvec[ ii]);
	    rowvec[ ii] = NULL;
	}
    }
    if ( currMinRow - pad - 1 > MINROW) {
	MINROW = currMinRow - pad - 1;
    }
}

/* interpolate a point from the 4 neighboring pixels */

void interpIt( int		srcFd,
	       int		size,
	       int		rows,
	       int		cols,
	       float		x,
	       float		y,
	       interpolateFn	interpFn,
	       void		*clientData,
	       uchar_t		*dst)
{
    uchar_t	localBuf[ 2048], *ptr, *srcRowBuf;
    int		srcRow, srcCol;

    srcRow = (int)floor( y);
    srcCol = (int)floor( x);
    if (( srcRow >= 0) && ( srcCol >= 0) &&
	( srcRow < rows - 1) && ( srcCol < cols - 1)) {

/* get (x0, y0), (x1, y1) from the AmFirst row */

	srcRowBuf = getRow( srcFd, srcRow, cols*size);
	ptr = localBuf;
	memcpy( ptr, &srcRowBuf[ srcCol*size], size);
	ptr += size;
	memcpy( ptr, &srcRowBuf[ ( srcCol + 1)*size], size);
	ptr += size;

/* get (x2, y2), (x3, y3) from the second row */

	srcRowBuf = getRow( srcFd, srcRow + 1, cols*size);
	memcpy( ptr, &srcRowBuf[ srcCol*size], size);
	ptr += size;
	memcpy( ptr, &srcRowBuf[ ( srcCol + 1)*size], size);
	ptr += size;
	( interpFn)( localBuf, clientData, size, 
		     x - srcCol, y - srcRow, dst);
    }
}

/* apply a map projection to an image */

void reprojectImage( int		srcFd,
		     int		rows,
		     int		cols,
		     int		size,
		     AnvilBoolean		invFlag,    /* inverse projection? */
		     PJ			*pj,
		     int		samples,    /* for interpolation */
		     float		*xmin,	    /* replaced by projected */
		     float		*ymin,	    /* values on return */
		     float		*xmax,
		     float		*ymax,
		     int		dstFd,
		     int		*dstRows,
		     int		*dstCols,
		     interpolateFn	interpFn,	/* pixel interp fn */
		     void		*clientData)
{
    float       pxmin1, pymin1, pxmax1, pymax1, pxmin2, pymin2, pxmax2, pymax2,
		pxmin,  pymin,  pxmax,  pymax, x, y, s, t, avgBin, xmid, ymid,
		xa, ya, xb, yb, xc, yc, xstep, ystep, srcX, srcY;
    int		ii, jj, mm, nn, x0, y0, x1, y1, x2, y2, x3, y3, srcRow, srcCol,
		length, minRow;
    uchar_t	*rowPtr, *rowStart, *dstBuffer;

/* project the matrix extents */

    pxmin1 = *xmin;
    pymin1 = *ymin;
    projectPt( pj, invFlag, &pxmin1, &pymin1);
    pxmin2 = *xmin;
    pymax1 = *ymax;
    projectPt( pj, invFlag, &pxmin2, &pymax1);
    pxmax1 = *xmax;
    pymax2 = *ymax;
    projectPt( pj, invFlag, &pxmax1, &pymax2);
    pxmax2 = *xmax;
    pymin2 = *ymin;
    projectPt( pj, invFlag, &pxmax2, &pymin2);

/* the matrix turns into a polygon, so we have to find it's minimum bounding
   rectangle in coordinate space. If it's a forward projection, then we
   round the values, to give more reasonable extents */ 

    if ( invFlag) {
	pxmin = MIN( pxmin1, pxmin2);
	pymin = MIN( pymin1, pymin2);
	pxmax = MAX( pxmax1, pxmax2);
	pymax = MAX( pymax1, pymax2);
    } else {
	pxmin = floor( MIN( pxmin1, pxmin2));
	pymin = floor( MIN( pymin1, pymin2));
	pxmax = floor( MAX( pxmax1, pxmax2));
	pymax = floor( MAX( pymax1, pymax2));
    }

/*  now figure out the width and height of the destination matrix, so the
    pixels have the same aspect ratio of the source matrix */

    avgBin = sqrt(( pxmax - pxmin)*( pymax - pymin)/( rows*cols));

/* if inverse projecting, round it to the nearest 1/10th arc second, if
   forward projecting, round it to the nearest 10 meters */

    if ( invFlag) {
	avgBin = rint( avgBin*36000)/36000;
    } else {
	avgBin = rint( avgBin/10)*10;
    }

/* calculate the size of the destination matrix */

    *dstRows = ( int) (( pymax - pymin)/avgBin);
    *dstCols = ( int) (( pxmax - pxmin)/avgBin);
    if ( ODD( *dstRows)) {
	(*dstRows)++;
    }
    if ( ODD( *dstCols)) {
	(*dstCols)++;
    }

/* find the center of the projected coordinates, and round it if we're 
   forward projecting, then reset the extents of the matrix to reflect
   the new center */

    xmid = ( pxmax + pxmin)/2.0;
    ymid = ( pymax + pymin)/2.0;
    if ( !invFlag) {
	xmid = rint( xmid/10.0)*10.0;
	ymid = rint( ymid/10.0)*10.0;
    }
    pxmin = xmid - (*dstCols/2)*avgBin;
    pxmax = xmid + (*dstCols/2)*avgBin;
    pymin = ymid - (*dstRows/2)*avgBin;
    pymax = ymid + (*dstRows/2)*avgBin;
    length = ( *dstRows)*( *dstCols)*size;
    xstep = (( pxmax - pxmin)*samples)/( *dstCols);
    ystep = (( pymax - pymin)*samples)/( *dstRows);

/* initialize the row manager to reflect the number of rows in the source 
   matrix.  Initialize the write buffer to reflect a row of tiles */
 
    initRowManager( rows);
    dstBuffer = ( uchar_t *) malloc( *dstCols*samples*size);

/* loop over the rows in the destination matrix, stepping by the bilinear
   interpolation size */

    for ( ii = 0, y = pymax - ystep; 
	  ii < *dstRows + samples; 
	  ii += samples, y -= ystep) {
	for ( jj = ii; jj < ii + samples; jj++) {
	    AmPrintHatch( jj, *dstRows + samples*2, NHATCH);
	}

/* clear out the destination buffer */

	memset( dstBuffer, '\0', *dstCols*samples*size);
 
/* get the starting x,y0-3 by casting the four corners back to the image:
 
        x0,y0 +----+ x1,y1
              |    |
              |    |
        x2,y2 +----+ x3,y3
*/
	minRow = rows + 1;
        for ( jj = 0, x = pxmin; jj < *dstCols; jj += samples, x += xstep) {

/* new tile: find the points to interpolate by using the inverse projection
   to go back to the source matrix */
 
            srcX = x;
            srcY = y + ystep;
            projectPt( pj, !invFlag, &srcX, &srcY);
            x0 = AmFloatScaleToInt( *xmin, *xmax, srcX, 0, cols - 1);
            y0 = AmFloatScaleToInt( *ymin, *ymax, srcY, rows - 1, 0);
            srcX = x + xstep;
            srcY = y + ystep;
            projectPt( pj, !invFlag, &srcX, &srcY);
            x1 = AmFloatScaleToInt( *xmin, *xmax, srcX, 0, cols - 1);
            y1 = AmFloatScaleToInt( *ymin, *ymax, srcY, rows - 1, 0);
            srcX = x;
            srcY = y;
            projectPt( pj, !invFlag, &srcX, &srcY);
            x2 = AmFloatScaleToInt( *xmin, *xmax, srcX, 0, cols - 1);
            y2 = AmFloatScaleToInt( *ymin, *ymax, srcY, rows - 1, 0);
            srcX = x + xstep;
            srcY = y;
            projectPt( pj, !invFlag, &srcX, &srcY);
            x3 = AmFloatScaleToInt( *xmin, *xmax, srcX, 0, cols - 1);
            y3 = AmFloatScaleToInt( *ymin, *ymax, srcY, rows - 1, 0);
 
/* perform bilinear interpolation on the points in the tile. s and t 
   go from 0.0 to 1.0, and scale from the tile's endpoints. Make sure
   that the source and desination matrices aren't overstepped */
 
            for ( mm = ii, s = 0.0, rowStart = dstBuffer + jj*size;
                  ( mm < *dstRows) && ( mm < ii+samples);
                  mm++, s += 1.0/samples, rowStart += *dstCols*size) {
                xa = x0 + s*( x2 - x0);
                ya = y0 + s*( y2 - y0);
                xb = x1 + s*( x3 - x1);
                yb = y1 + s*( y3 - y1);
                for ( nn = jj, t = 0.0, rowPtr = rowStart;
                      ( nn < *dstCols) && ( nn < jj + samples);
                      nn++, t += 1.0/samples, rowPtr += size) {
		    xc = xa + t*( xb - xa);
		    yc = ya + t*( yb - ya);
                    srcCol = (int)floor( xc);
                    srcRow = (int)floor( yc);
                    if (( srcCol >= 0) && ( srcCol < cols) &&
                        ( srcRow >= 0) && ( srcRow < rows)) {

/* get the row that the point is from, and maintain the lowest row index */ 

			if ( srcRow < minRow) {
			    minRow = srcRow;
			}
			interpIt( srcFd, size, rows, cols, 
				  xc, yc, interpFn, clientData, rowPtr);
                    }
                }
            }
        }

/* write the buffer to the output file, and free any rows that have been 
   read that are behind the lowest row by a certain pad */

	write( dstFd, dstBuffer, (*dstCols)*size*samples);
	manageRows( minRow, samples);
    }
    *xmin = pxmin;
    *ymin = pymin;
    *xmax = pxmax;
    *ymax = pymax;
}


int main( int argc, char *argv[])
{
    char 		**projArgs, *kind, *polyBuf, ptsBuf[ 65536],
			path[ PATHLEN], hdrPath[ PATHLEN], 
			*lineBuf, *pointBuf;
    PJ			*pj;
    AnvilBoolean		invFlag;
#if defined(_WIN32)
    AmNTResourceDbClass	*db;
#else
    AmXResourceDbClass	*db;
#endif
    int			fdIn, fdOut, rows, cols, size, ptsize, ii, jj,
			dstRows, dstCols, samples = 10;
    FILE		*hdrFp;
    float		xmin = MY_HUGE, ymin = MY_HUGE, 
			xmax = -MY_HUGE, ymax = -MY_HUGE;
    AmPolygonPtr		poly;
    AmCoordPtr		pts, point;
    AmLinePtr		line;
    AnvilBoolean		pipeFlag;
    char *fields[] = { AmNfilename, AmNrows, AmNcols, AmNsize, AmNtype, 
		       AmNxmin, AmNymin, AmNxmax, AmNymax, AmNxunits, 
		       AmNyunits, AmNinProjection, NULL };

    if ( argc < 5) {
	fprintf( stderr, "usage %s proj inheader infile outfile\n", argv[ 0]);
	exit( -1);
    }

/* get the file which stores all the enumerations that we'll ever need */

    if ( getenv( CONFIG_ENV) == NULL) {
        fprintf( stderr, "environment variable %s not defined", CONFIG_ENV);
        exit( -1);
    }
    AmInitUnits();

/* read the projection strings */

    if (( projArgs = AmReadStrvec( argv[ 1])) == NULL) {
	fprintf( stderr, "failed to read projection file %s\n", argv[ 1]);
	exit( -1);
    }

/* generate the output pathnames */

    sprintf( path, "%s.dat", argv[ 4]);
    sprintf( hdrPath, "%s.hdr", argv[ 4]);

/* open the input header, and output files */
 
    if (( fdIn = AmOpenReadFile( argv[ 3], &pipeFlag)) < 0) {
	fprintf( stderr, "failed to open input file %s\n", argv[ 3]);
	exit( -1);
    }
    if (( fdOut = open( path, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open output file %s\n", path);
	exit( -1);
    }
    if (( hdrFp = fopen( hdrPath, "w")) == NULL) {
	fprintf( stderr, "failed to open output header %s\n", hdrPath);
	exit( -1);
    }

/* check for the "inv-" flag for inverse projections, and set a flag and
   return the projection to normal */

    invFlag = FALSE;
    for ( ii = 0; ii < AmVeclen(( const char **) projArgs); ii++) { 
	if ( !strncmp( projArgs[ ii], "inv-", 4)) {
	    invFlag = TRUE;
	    projArgs[ ii] = &projArgs[ ii][ 4];
	} 
    }
    if (( pj = pj_init( AmVeclen(( const char **) projArgs), 
			( const char **) projArgs)) == NULL) {
	fprintf( stderr, "projection %s failed\n", argv[ 1]);
	exit( -1);
    }
#if defined(_WIN32)
    db = new AmNTResourceDbClass( argv[ 2]);
#else
    db = new AmXResourceDbClass( argv[ 2]);
#endif
    if ( !db->IsValid()) {
	fprintf( stderr, "failed to read header file %s\n", argv[ 2]);
	exit( -1);
    }
    if ( !db->GetResource( "type", &kind)) {
	fprintf( stderr, "failed to get file type from header %s\n", argv[ 2]);
	exit( -1);
    }
	
    if ( !strcmp( kind, AmNpolygonPlot)) {
	if ( !db->GetResource( AmNrows, &rows)) {
	    fprintf( stderr, "failed to read rows from polygon file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNsize, &size)) {
	    fprintf( stderr, "failed to read size from polygon file\n");
	    exit( -1);
	}
	polyBuf = ( char *) malloc( size);
	for ( ii = 0; ii < rows; ii++) {
	    AmPrintHatch( ii, rows, NHATCH);

/* read the polygon and its points */

	    if ( AmReadLoopFile( fdIn, polyBuf, size) < size) {
		fprintf( stderr, "failed to read polygon %d\n", ii);
		exit( -1);
	    }
	    poly = ( AmPolygonPtr) polyBuf;
	    ptsize = poly->numPts*sizeof( AmCoordRec);
	    if ( AmReadLoopFile( fdIn, ptsBuf, ptsize) < ptsize) {
		fprintf( stderr, "failed to read points from poly %d\n", ii);
		exit( -1);
	    }

/* project each of the points in the polygon */

	    pts = ( AmCoordPtr) ptsBuf;
	    for ( jj = 0; jj < poly->numPts; jj++) {
		projectPt( pj, invFlag, &pts[ jj].x, &pts[ jj].y);
	    }

/* and don't forget to project the extents of the polygon if they've been
   set... */

	    if ( poly->flags & AM_MBR_FLAG) {
		projectPt( pj, invFlag, &poly->xmin, &poly->ymin);
		projectPt( pj, invFlag, &poly->xmax, &poly->ymax);
	    }

/* write out the polygon and the projected points */

	    write( fdOut, polyBuf, size);
	    write( fdOut, ptsBuf, ptsize);
 
/* track the  new extents of the data */

	    for ( jj = 0; jj < poly->numPts; jj++) {
		if ( pts[ jj].x < xmin) {
		    xmin = pts[ jj].x;
		}
		if ( pts[ jj].y < ymin) {
		    ymin = pts[ jj].y;
		}
		if ( pts[ jj].x > xmax) {
		    xmax = pts[ jj].x;
		}
		if ( pts[ jj].y > ymax) {
		    ymax = pts[ jj].y;
		}
	    }
	}
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( path));
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		polygonPlot\n");
        fprintf( hdrFp, "xmin:          %f\n", xmin);
        fprintf( hdrFp, "ymin:          %f\n", ymin);
        fprintf( hdrFp, "xmax:          %f\n", xmax);
        fprintf( hdrFp, "ymax:          %f\n", ymax);
        fprintf( hdrFp, "xunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
        fprintf( hdrFp, "yunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
	fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    } else if ( !strcmp( kind, AmNlinePlot)) {
	if ( !db->GetResource( AmNrows, &rows)) {
	    fprintf( stderr, "failed to read rows from vector file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNsize, &size)) {
	    fprintf( stderr, "failed to read size from vector file\n");
	    exit( -1);
	}
	lineBuf = ( char *) malloc( size);
	for ( ii = 0; ii < rows; ii++) {
	    AmPrintHatch( ii, rows, NHATCH);
	    if ( AmReadLoopFile( fdIn, lineBuf, size) < size) {
		fprintf( stderr, "failed to read line %d\n", ii);
		exit( -1);
	    }
	    line = (AmLinePtr) lineBuf;

/* project the line's endpoints */

	    projectPt( pj, invFlag, &line->x0, &line->y0);
	    projectPt( pj, invFlag, &line->x1, &line->y1);

/* write out the projected line */

	    write( fdOut, lineBuf, size);

/* track the extents of the lines */

	    if ( line->x0 < xmin) {
		xmin = line->x0;
	    }
	    if ( line->x1 < xmin) {
		xmin = line->x1;
	    }
	    if ( line->x0 > xmax) {
		xmax = line->x0;
	    }
	    if ( line->x1 > xmax) {
		xmax = line->x1;
	    }
	    if ( line->y0 < ymin) {
		ymin = line->y0;
	    }
	    if ( line->y1 < ymin) {
		ymin = line->y1;
	    }
	    if ( line->y0 > ymax) {
		ymax = line->y0;
	    }
	    if ( line->y1 > ymax) {
		ymax = line->y1;
	    }
	}
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( path));
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		linePlot\n");
        fprintf( hdrFp, "xmin:          %f\n", xmin);
        fprintf( hdrFp, "ymin:          %f\n", ymin);
        fprintf( hdrFp, "xmax:          %f\n", xmax);
        fprintf( hdrFp, "ymax:          %f\n", ymax);
        fprintf( hdrFp, "xunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
        fprintf( hdrFp, "yunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
	fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    } else if (!strcmp( kind, AmNscatterPlot)) {
	if ( !db->GetResource( AmNrows, &rows)) {
	    fprintf( stderr, "failed to read rows from vector file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNsize, &size)) {
	    fprintf( stderr, "failed to read size from vector file\n");
	    exit( -1);
	}
	pointBuf = ( char *) malloc( size);
	for ( ii = 0; ii < rows; ii++) {
	    AmPrintHatch( ii, rows, NHATCH);
	    if ( AmReadLoopFile( fdIn, pointBuf, size) < size) {
		fprintf( stderr, "failed to read point %d\n", ii);
		exit( -1);
	    }
	    point = ( AmCoordPtr) pointBuf;

/* project the point's coordinate */

	    projectPt( pj, invFlag, &point->x, &point->y);

/* write out the projected point */

	    write( fdOut, pointBuf, size);

/* maintain the extents of the projected file */

	    if ( point->x < xmin) {
		xmin = point->x;
	    }
	    if ( point->y < ymin) {
		ymin = point->y;
	    }
	    if ( point->x > xmax) {
		xmax = point->x;
	    }
	    if ( point->y > ymax) {
		ymax = point->y;
	    }
	}
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( path));
	fprintf( hdrFp, "rows:		%d\n", rows);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		scatterPlot\n");
        fprintf( hdrFp, "xmin:          %f\n", xmin);
        fprintf( hdrFp, "ymin:          %f\n", ymin);
        fprintf( hdrFp, "xmax:          %f\n", xmax);
        fprintf( hdrFp, "ymax:          %f\n", ymax);
        fprintf( hdrFp, "xunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
        fprintf( hdrFp, "yunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
	fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    } else if ( !strcmp( kind, AmNimagePlot)) {
	void 		*clientData;
	interpolateFn	interpFn;

	if ( !db->GetResource( AmNrows, &rows)) {
	    fprintf( stderr, "failed to read rows from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNcols, &cols)) {
	    fprintf( stderr, "failed to read cols from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNsize, &size)) {
	    fprintf( stderr, "failed to read size from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNxmin, &xmin)) {
	    fprintf( stderr, "failed to read xmin from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNymin, &ymin)) {
	    fprintf( stderr, "failed to read ymin from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNxmax, &xmax)) {
	    fprintf( stderr, "failed to read xmax from image file\n");
	    exit( -1);
	}
	if ( !db->GetResource( AmNymax, &ymax)) {
	    fprintf( stderr, "failed to read ymax from image file\n");
	    exit( -1);
	}
	if ( !AmCreateInterpolationArgs( db, &clientData, &interpFn)) {
	    fprintf( stderr, "failed to create intepolation function\n");
	    exit( -1);
	}
	reprojectImage( fdIn, rows, cols, size, invFlag, pj, samples,
			&xmin, &ymin, &xmax, &ymax, fdOut, 
			&dstRows, &dstCols, interpFn, clientData);
	fprintf( hdrFp, "filename:	%s\n", AmGetRootName( path));
	fprintf( hdrFp, "rows:		%d\n", dstRows);
	fprintf( hdrFp, "cols:		%d\n", dstCols);
	fprintf( hdrFp, "size:		%d\n", size);
	fprintf( hdrFp, "type:		imagePlot\n");
        fprintf( hdrFp, "xmin:          %f\n", xmin);
        fprintf( hdrFp, "ymin:          %f\n", ymin);
        fprintf( hdrFp, "xmax:          %f\n", xmax);
        fprintf( hdrFp, "ymax:          %f\n", ymax);
        fprintf( hdrFp, "xunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
        fprintf( hdrFp, "yunits:        %s\n", 
		 getProjectionUnits( projArgs, invFlag));
	fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    }

/* write the projection and the number of projection args in the header file.
   If the inverse flag is set, then remove the inProjection flag, because
   we're going to latitude/longitude.  The inProjection flag has an interaction
   with the overlay engine, which applies the inProjection to added layers. */

    if ( !invFlag) {
	fprintf( hdrFp, "inProjection:	");
	for ( ii = 0; ii < AmVeclen(( const char **) projArgs); ii++) {
	    fprintf( hdrFp, "\"%s\"", projArgs[ ii]);
	}
	fprintf( hdrFp, "\n");
	fprintf( hdrFp, "numProjArgs:	%d\n", 
			AmVeclen(( const char **) projArgs));
    }

/* copy the unchanged header fields from the source file */

    AmDumpHeaderExcept( argv[ 2], fields, hdrFp);
    close( fdIn);
    AmCloseReadFile( fdOut, pipeFlag);
    fclose( hdrFp);
}

