#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <anvil/project.h>
#include <anvil/datafmt.h>
#include <util/shapefile.h>
#include <util/dbf.h>
#include <anvil/streetpline.h>
#include <anvil/choppath.h>
#include <anvil/macros.h>

#define MAXRECORD	65536
#define MAXPARTS	2048
#define MAXPTS		16384
#define MAXFIELDS	128
#define NAME_FIELD	"NAME"
#define CFCC_FIELD	"CFCC"

typedef struct tigerMap_ {
    char	codeName[ 64];
    int		code;
} tigerMapRec, *tigerMapPtr;

int 
byTigerCodeName( tigerMapPtr	a,
		 tigerMapPtr	b)
{
    return( strcmp( a->codeName, b->codeName));
}

/* read the mapping from TIGER codes to our code numbers */

AnvilBoolean 
readTigerMap( char		*filename,
	      tigerMapPtr	tigerMap,
	      int		*numMaps)
{
    FILE	*fp;
    char	buffer[ 1024];

    if (( fp = fopen( filename, "r")) == NULL) {
	return( FALSE);
    } 
    *numMaps = 0;
    do {
	if ( fgets( buffer, sizeof( buffer), fp) == NULL) {
	    break;
	}
	if ( sscanf( buffer, "%s %d", tigerMap[ *numMaps].codeName,
				      &tigerMap[ *numMaps].code) != 2) {
	    break;
	}
	( *numMaps)++;
    } while ( TRUE);
    fclose( fp);
    qsort( tigerMap, *numMaps, sizeof( tigerMapRec), 
	   ( AmSearchFn) byTigerCodeName);
    return( TRUE);
}

/* return the integer code which corresponds to the TIGER string code */

AnvilBoolean 
getTigerCode( tigerMapPtr	tigerMap,
	      int		numMaps,
	      char		*codeStr,
	      int		*resCode)
{
    tigerMapRec	target;
    tigerMapPtr	result;

    strcpy( target.codeName, codeStr);
    result = ( tigerMapPtr) bsearch( &target, tigerMap, numMaps,
				     sizeof( tigerMapRec),
				     ( AmSearchFn) byTigerCodeName);
    if ( result == NULL) {
	*resCode = -1;
	return( FALSE);
    } else {
	*resCode = result->code;
	return( TRUE);
    }
}

void 
pointsExtents( AmCoordPtr	pts,
	       int		numPts,
	       float		*xmin,
	       float		*ymin,
	       float		*xmax,
	       float		*ymax)
{
    int	ii;

    *xmin = MY_HUGE;
    *ymin = MY_HUGE;
    *xmax = -MY_HUGE;
    *ymax = -MY_HUGE;
    for ( ii = 0; ii < numPts; ii++) {
	if ( pts[ ii].x < *xmin) {
	    *xmin = pts[ ii].x;
	}
	if ( pts[ ii].y < *ymin) {
	    *ymin = pts[ ii].y;
	}
	if ( pts[ ii].x > *xmax) {
	    *xmax = pts[ ii].x;
	}
	if ( pts[ ii].y > *ymax) {
	    *ymax = pts[ ii].y;
	}
    }
}

int main( int argc, char *argv[])
{
    shapeFileHeaderRec		shapeFileHeader;
    int				fd, dbfd, fdOut, pos, ii, rows, reclen, 
				numFields, nameIndex, typeIndex, nameOffset, 
				typeOffset, nameLength, typeLength, 
				numMaps, code;
    shapeRecordHeaderRec	shapeRecordHeader;
    char			*arcRecord, dbRecord[ MAXRECORD], 
				hdrPath[ PATHLEN], dataPath[ PATHLEN],
				codeStr[ FIELDLEN];
    shapePointRec		pts[ MAXPTS];
    arcRecordRec		arc;
    FILE			*hdrFp;
    AmStreetPlineRec		streetPline;
    AmCoordRec			streetPts[ MAXPTS];
    dbfRecHeaderRec		dbfRecHeaders[ MAXFIELDS];
    dbfFileHeaderRec		dbfFileHeader;
    tigerMapRec			tigerMap[ 256];

    if ( argc < 5) {
	fprintf( stderr, "usage: %s infile dbfile mapfile outfile\n", 
		 argv[ 0]);
	exit( -1);
    }

/* open the shapefile and .DBF files. */

    if (( fd = open( argv[ 1], O_RDONLY)) < 0) {
	fprintf( stderr, "failed to open %s\n", argv[ 1]);
	exit( -1);
    }
    if (( dbfd = open( argv[ 2], O_RDONLY)) < 0) {
	fprintf( stderr, "failed to open %s\n", argv[ 1]);
	exit( -1);
    }
    if ( !readTigerMap( argv[ 3], tigerMap, &numMaps)) {
	fprintf( stderr, "failed to read TIGER mapping file\n");
	exit( -1);
    }
    sprintf( hdrPath, "%s.hdr", argv[ 4]);
    sprintf( dataPath, "%s.dat", argv[ 4]);
    if (( fdOut = open( dataPath, O_WRONLY | O_CREAT, PERMS)) < 0) {
	fprintf( stderr, "failed to open %s\n", dataPath);
	exit( -1);
    }
    if (( hdrFp = fopen( hdrPath, "w")) == NULL) {
	fprintf( stderr, "failed to open %s\n", hdrPath);
	exit( -1);
    }

/* because of the alignment for doubles, and the file starting with 9 ints, 
   the header size doesn't wind up being exactly 100 bytes, so we force it */

    if ( read( dbfd, &dbfFileHeader, DBF_FILE_HEADER_LENGTH) < 
	       DBF_FILE_HEADER_LENGTH) {
        fprintf( stderr, "failed to read header from %s\n", argv[ 1]);
        exit( -1);
    }
    LITTLE_LONG( dbfFileHeader.numRecords);

/* we need to pull the name and type from each street in the .DBF file.. */

    reclen = 0;
    numFields = 0;
    nameIndex = -1;
    nameOffset = -1;
    nameLength = -1;
    typeIndex = -1;
    typeOffset = -1;
    typeLength = -1;
    do {
        if ( read( dbfd, dbRecord, 2) < 2) {
            fprintf( stderr, "failed to read record header from %s\n", 
		     argv[ 1]);
            exit( -1);
        }
        if ( dbRecord[ 0] == '\r') {
            break;
        }
        if ( read( dbfd, &dbRecord[ 2], DBF_REC_HEADER_LENGTH - 2) < 
	     DBF_REC_HEADER_LENGTH - 2) {
            fprintf( stderr, "failed to read record header from %s\n", 
		     argv[ 1]);
            exit( -1);
        }
        memcpy( &dbfRecHeaders[ numFields], dbRecord, 
		sizeof( dbfRecHeaderRec));
        LITTLE_LONG( dbfRecHeaders[ numFields].length);
        AN_DEBUG( 40, ( "name = %s, length = %d, type = %c\n",
			dbfRecHeaders[ numFields].name, 
			dbfRecHeaders[ numFields].length,
			dbfRecHeaders[ numFields].type));

/* look for the name and type indices. */

	if ( !strncmp( dbfRecHeaders[ numFields].name, NAME_FIELD,
		       strlen( NAME_FIELD))) {
	    nameIndex = numFields;
	    nameOffset = reclen;
	    nameLength = dbfRecHeaders[ nameIndex].length;
	} else if ( !strncmp( dbfRecHeaders[ numFields].name, CFCC_FIELD,
			      strlen( CFCC_FIELD))) {
	    typeIndex = numFields;
	    typeOffset = reclen;
	    typeLength = dbfRecHeaders[ nameIndex].length;
	}

        reclen += dbfRecHeaders[ numFields].length;
        numFields++;
    } while ( TRUE);
    if ( ODD( reclen)) {
        reclen++;
    }
    AN_DEBUG( 40, ( "record length = %d\n", reclen));
    AN_DEBUG( 40, ( "name offset = %d, type offset = %d\n", 
		    nameOffset, typeOffset));

/* we allocate the record to ensure that it starts on a 
   double word boundary */

    if (( arcRecord = ( char *) malloc( MAXRECORD)) == NULL) {
	fprintf( stderr, "failed to allocate %d bytes\n", MAXRECORD);
	exit( -1);
    }

/* because of the alignment for doubles, and the file starting with 9 
   ints, the size of the header doesn't wind up being exactly 100 bytes, 
   so we force it */

    if ( read( fd, &shapeFileHeader, SHP_FILE_HEADER_LENGTH) < 
	 SHP_FILE_HEADER_LENGTH) {
	fprintf( stderr, "failed to read header from %s\n", argv[ 1]);
	exit( -1);
    }
    BIG_LONG( shapeFileHeader.fileCode);
    LITTLE_LONG( shapeFileHeader.version);
    LITTLE_LONG( shapeFileHeader.shapeType);

/* get the extents and write them out to the header */

    LITTLE_DOUBLE( shapeFileHeader.xmin);
    LITTLE_DOUBLE( shapeFileHeader.ymin);
    LITTLE_DOUBLE( shapeFileHeader.xmax);
    LITTLE_DOUBLE( shapeFileHeader.ymax);

    AN_DEBUG( 20, ( "fileCode = %ld, length = %d, version = %d\n", 
		    shapeFileHeader.fileCode, shapeFileHeader.fileLength, 
		    shapeFileHeader.version));
    AN_DEBUG( 20, ( "xmin = %.2f, ymin = %.2f, xmax = %.2f, ymax = %.2f\n",
		    shapeFileHeader.xmin, shapeFileHeader.ymin, 
		    shapeFileHeader.xmax, shapeFileHeader.ymax));

/* we assume that the .SHP file is arcs, for roads */

    pos = SHP_FILE_HEADER_LENGTH;
    rows = 0;
    do {

/* read a record from the .SHP file and from the .DBF file */

	if ( read( fd, &shapeRecordHeader, sizeof( shapeRecordHeaderRec)) < 
		   (signed)sizeof( shapeRecordHeaderRec)) {
	    break;
	}
	BIG_LONG( shapeRecordHeader.recordNumber);
	BIG_LONG( shapeRecordHeader.contentLength);
	AN_DEBUG( 90, ( "record = %ld, length = %ld\n",
			shapeRecordHeader.recordNumber, 
			shapeRecordHeader.contentLength));
	if ( read( fd, arcRecord, shapeRecordHeader.contentLength*2) < 
		   (signed)shapeRecordHeader.contentLength*2) {
	    fprintf( stderr, "failed to read %ld bytes at %d\n", 
		     shapeRecordHeader.contentLength*2, pos);
	    break;
	}
	if ( read( dbfd, dbRecord, reclen) < (signed)reclen) {
	    fprintf( stderr, "failed to read record from dbase file\n");
	    break;
	}
	pos += shapeRecordHeader.contentLength*2;

/* copy to the arc structure to ensure double word alignment */

	memcpy( &arc, arcRecord, sizeof( arcRecordRec));
	AN_DEBUG( 90, ( "type = %d\n", arc.shapeType));
	LITTLE_LONG( arc.numPoints);
	LITTLE_LONG( arc.numParts);

/* memcpy the points records to ensure double word alignment. */ 

	memcpy( pts, &arcRecord[ ARC_HEADER_LENGTH], 
		arc.numPoints*sizeof( shapePointRec));
	for ( ii = 0; ii < arc.numPoints; ii++) {
	    LITTLE_DOUBLE( pts[ ii].x);
	    LITTLE_DOUBLE( pts[ ii].y);
	}

/* copy the points to the points array and find the extents */

	for ( ii = 0; ii < arc.numPoints; ii++) {
	    streetPts[ ii].x = pts[ ii].x;
	    streetPts[ ii].y = pts[ ii].y;
	}
	pointsExtents( streetPts, arc.numPoints, 
		       &streetPline.pline.xmin, &streetPline.pline.ymin,
		       &streetPline.pline.xmax, &streetPline.pline.ymax);

/* set the flags, number of points, name and code */

	streetPline.pline.flags = AM_MBR_FLAG;
	streetPline.pline.numPts = arc.numPoints;
	strncpy( streetPline.name, &dbRecord[ nameOffset], nameLength);
	strncpy( codeStr, &dbRecord[ typeOffset], 3);
	codeStr[ 3] = '\0';
	if ( !getTigerCode( tigerMap, numMaps, codeStr, &code)) {
	    fprintf( stderr, "failed to find code for %s\n", codeStr);
	}
	streetPline.code = code;
	write( fdOut, &streetPline, sizeof( AmStreetPlineRec));
	write( fdOut, streetPts, arc.numPoints*sizeof( AmCoordRec));
	rows++;
    } while ( TRUE);

/* write out the header information */

    fprintf( hdrFp, "filename:		%s\n", AmGetRootName( dataPath));
    fprintf( hdrFp, "xmin:		%f\n", shapeFileHeader.xmin);
    fprintf( hdrFp, "ymin:		%f\n", shapeFileHeader.ymin);
    fprintf( hdrFp, "xmax:		%f\n", shapeFileHeader.xmax);
    fprintf( hdrFp, "ymax:		%f\n", shapeFileHeader.ymax);
    fprintf( hdrFp, "xunits:		longitude_arcdeg\n");
    fprintf( hdrFp, "yunits:		latitude_arcdeg\n");
    fprintf( hdrFp, "type:		polylinePlot\n");
    fprintf( hdrFp, "size:		%d\n", sizeof( AmStreetPlineRec));
    fprintf( hdrFp, "rows:		%d\n", rows);
    fprintf( hdrFp, "fileSource:	TIGER\n");
    fprintf( hdrFp, "colorsOffset:	%d\n", 
	     offsetof( AmStreetPlineRec, code));
    fprintf( hdrFp, "colorsType:	int\n");
    fclose( hdrFp);
    close( fdOut);
    close( dbfd);
    close( fd);
}
