#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <anvil/tiffio.h>
#include <anvil/macros.h>
#include <anvil/choppath.h>
#include <anvil/project.h>
#include <anvil/tiff.h>

#define MAXMAXVALUE 1023
#define NEXTSAMPLE \
{ \
    if ( bitsleft == 0) { \
	inp++; \
	bitsleft = 8; \
    } \
    bitsleft -= bitsPerPixel; \
    sample = ( *inp >> bitsleft) & maxval; \
}


int main( int argc, char *argv[])
{
    TIFF		*tif;
    unsigned short	bitsPerPixel, samplesPerPixel, photometric,
			*redCmap, *greenCmap, *blueCmap;
    int			row, rows, col, cols, maxval, outbuflen, 
			fdOut, numcolors, bitsleft, ii, lastNonZeroMarker;
    uchar_t		*inbuf, *outbuf, *inp, *outp, sample;
    char		datapath[  PATHLEN], hdrpath[  PATHLEN], palettepath[ PATHLEN];
    AnvilBoolean		greyscale;
    FILE		*hdrFp, *paletteFp, *inHdrFp;

    if ( argc < 3) {
	fprintf( stderr, "usage %s infile outfile\n", argv[ 0]);
	exit( -1);
    }
    if  (( inHdrFp = fopen( argv[ 1], "r")) == NULL) {
	fprintf( stderr, "failed to open header file %s", argv[ 1]);
	exit( -1);
    }

/* open the tiff file for reading */

    if (( tif = TIFFOpen( argv[ 1], "r")) == NULL) {
	fprintf( stderr, "failed to open TIFF file %s", argv[ 2]);
	exit( -1);
    }
 
    sprintf( datapath, "%s.dat", argv[ 2]);
    sprintf( hdrpath, "%s.hdr", argv[ 2]);
    sprintf( palettepath, "%s.palette", argv[ 2]);
    if (( fdOut = open( datapath, O_CREAT | O_WRONLY, PERMS)) < 0) {
	fprintf( stderr, "failed to open output file %s", datapath);
	exit( -1);
    }
    if  (( hdrFp = fopen( hdrpath, "w")) == NULL) {
	fprintf( stderr, "failed to open header file %s", hdrpath);
	exit( -1);
    }

/* get the number of bits per pixel, samples per pixel, and the photometric */

    if  ( !TIFFGetField( tif, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel)) {
	bitsPerPixel = 1;
    }
    if  ( !TIFFGetField( tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel)) {
	samplesPerPixel = 1;
    }
    if  ( !TIFFGetField( tif, TIFFTAG_PHOTOMETRIC, &photometric)) {
	fprintf( stderr, "error getting photometric from file %s\n", argv[ 1]);
	exit( -1);
    }

/* get the dimensions of the image */

    TIFFGetField( tif, TIFFTAG_IMAGEWIDTH, &cols);
    TIFFGetField( tif, TIFFTAG_IMAGELENGTH, &rows);
    maxval = ( 1 << bitsPerPixel) - 1;
    if ( maxval == 1 && samplesPerPixel == 1) {
	greyscale = TRUE;
    } else {
	switch ( photometric) {
	case PHOTOMETRIC_MINISBLACK:
	case PHOTOMETRIC_MINISWHITE:
	    greyscale = TRUE;
	    break;
	case PHOTOMETRIC_PALETTE:
	
/* if it's colormapped, then read the colormap into R, G and B vectors */

	    if ( !TIFFGetField( tif, TIFFTAG_COLORMAP, 
			        &redCmap, &greenCmap, &blueCmap)) {
		 fprintf( stderr, "error getting colormaps from file %s\n", 
			  argv[ 1]);
		 exit( -1);
	    }
	    numcolors = maxval + 1;

/* if the colormap trails with 0's, then write to the AmLast non-zero value */

            lastNonZeroMarker = 0;
            for ( ii = 0; ii < numcolors; ii++) {
                if ((( redCmap[ ii] >> 8) != 0) ||
                    (( greenCmap[ ii] >> 8) != 0) ||
                    (( blueCmap[ ii] >> 8) != 0)) {
                    lastNonZeroMarker = ii;
                }
            }
	    lastNonZeroMarker++;
	    if  (( paletteFp = fopen( palettepath, "w")) == NULL) {
		fprintf( stderr, "failed to open palette file %s", palettepath);
		exit( -1);
	    }
	    for ( ii = 0; ii < lastNonZeroMarker; ii++) {
		fprintf( paletteFp, "%d %d %d %d\n", 
			 ii, redCmap[ ii] >> 8, greenCmap[ ii] >> 8, 
			 blueCmap[ ii] >> 8);
	    }
	    fclose( paletteFp);
	    maxval = MAXMAXVALUE;
	    greyscale = FALSE;
	    break;
	case PHOTOMETRIC_RGB:
	    greyscale = FALSE;
	    break;
	default:
	    fprintf( stderr, "unrecognized photometric %d in file %s\n", 
	     	     photometric, argv[ 1]);
	    exit( -1);
	}
    }

/* allocate a buffer to read the scanline, then read the file scanline by
   scanline */

    inbuf = ( uchar_t *) malloc( TIFFScanlineSize( tif));

/* RGB images map to 32 bit values, all others to 8-bit values */

    if ( photometric == PHOTOMETRIC_RGB) {
	outbuflen = cols*sizeof( unsigned long);
    } else {
	outbuflen = cols;
    }
    outbuf = ( uchar_t *) malloc( outbuflen);
    for ( row = 0; row < rows; row++) {
	if ( TIFFReadScanline( tif, inbuf, row, 0) < 0)  {
	    fprintf( stderr, "bad data row %d in file %s\n", row, argv[ 1]);
	    exit(  -1);
	}
	bitsleft = 8;
	outp = outbuf;
	inp = inbuf; 
	switch ( photometric) {
	case PHOTOMETRIC_MINISBLACK:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = sample;
	    }
	    break;
	case PHOTOMETRIC_MINISWHITE:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = maxval - sample;
	    }
	    break;
	case PHOTOMETRIC_PALETTE:
	    for ( col = 0; col < cols; col++, outp++) {
		NEXTSAMPLE
		*outp = sample;
	    }
	    break;
	case PHOTOMETRIC_RGB:
	    for ( col = 0; col < cols; col++) {
		outp++;
		NEXTSAMPLE
		*outp++ = sample;
		NEXTSAMPLE
		*outp++ = sample;
		NEXTSAMPLE
		*outp++ = sample;
		if ( samplesPerPixel == 4) {
		    NEXTSAMPLE
		}
	    }
	    break;
	default:
	    break;
	}
	write( fdOut, outbuf, outbuflen);
    }
    fprintf( hdrFp, "filename:	%s\n", AmGetRootName( datapath));
    if ( photometric == PHOTOMETRIC_PALETTE) {
	fprintf( hdrFp, "palette:	%s\n", AmGetRootName( palettepath));
    } else if ( greyscale) {
	fprintf( hdrFp, "colormap:	Grey-Scale\n");
	fprintf( hdrFp, "numcolors:	64\n");
    }
    fprintf( hdrFp, "type:	imagePlot\n");
    fprintf( hdrFp, "rows:	%d\n", rows);
    fprintf( hdrFp, "cols:	%d\n", cols);
    if ( photometric == PHOTOMETRIC_RGB) {
	fprintf( hdrFp, "size:		%d\n", sizeof( unsigned long));
	fprintf( hdrFp, "trueColor:	true\n");
	fprintf( hdrFp, "dataType:	int\n");
    } else {
	fprintf( hdrFp, "size:		%d\n", sizeof( uchar_t));
	fprintf( hdrFp, "dataType:	char\n");
    }
    fprintf( hdrFp, "sourceFile:	%s\n", argv[ 1]);
    fprintf( hdrFp, "fileSource:	TIFF\n");
    close( fdOut);
}
