#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#if !defined(_WIN32) && !defined(__BSD44__)
#include <alloca.h>
#endif
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <anvil/resource.h>
#include <anvil/datafmt.h>
#include <anvil/macros.h>
#include <anvil/files.h>
#include <anvil/newtile.h>
#include <anvil/compression.h>
#include <anvil/polyarea.h>
#include <anvil/choppath.h>
#include <util/utils.h>
#include <util/tile.h>
#include <util/misc.h>

/* extra buffer space in case the compression fails */

static const int COMPRESS_SLOP = 256;

/* filter a vector of objects by a lookup table, lookup type, lookup offset */

int AmFilterVec( uchar_t          *buffer,
               int              numread,
               int              size,
               AmDataTypeEnum     lookupType,
               int              lookupOffset,
               int              *filterIds,
               int              numIds)
{
    int         kk, jj, ival;
    uchar_t     *ptr;

    kk = 0;
    for ( jj = 0, ptr = buffer; jj < numread; jj++, ptr += size) {
        ival = AmGetLookupVal(( char *) ptr, lookupType, lookupOffset);
        if ( AmInTable( ival, filterIds, numIds)) {
            memcpy( &buffer[ kk*size], ptr, size);
            kk++;
        }
    }
    return( kk);
}

/* filter a vector of objects by a lookup table, lookup type, lookup offset */

int AmFilterPolygonVec( uchar_t          *buffer,
		      int              numread,
		      int              size,
		      AmDataTypeEnum     lookupType,
		      int              lookupOffset,
		      int              *filterIds,
		      float	       *filterSizes,
		      int              numIds)
{
    int         kk, jj, ival, elsize, index, dstIndex;
    uchar_t     *ptr;

    kk = 0;
    dstIndex = 0;
    elsize = size + (( AmPolygonPtr) buffer)->numPts*sizeof( AmCoordRec);
    for ( jj = 0, ptr = buffer; jj < numread; jj++, ptr += elsize) {
	elsize = size + (( AmPolygonPtr) ptr)->numPts*sizeof( AmCoordRec);

/* copy the polygon if it's attribute is in the table, and the area is
   larger than the filter size */

        ival = AmGetLookupVal(( char *) ptr, lookupType, lookupOffset);
	index = AmTableIndex( ival, filterIds, numIds);
	if ( index != -1) {
	    if ( AmPolygonArea(( AmPolygonPtr) ptr) > filterSizes[ index]) {
		memcpy( &buffer[ dstIndex], ptr, elsize);
		dstIndex += elsize;
		kk++;
	    }
        }
    }
    return( kk);
}

AnvilBoolean AmSameFilter( int *filter1,
                    int length1,
                    int *filter2,
                    int length2)
{
    int	ii;

    if ( filter1 == NULL) {
    	if ( filter2 == NULL) {
	    return( TRUE);
	} else {
	    return( FALSE);
	}
    } else if ( filter2 == NULL) {
	return( FALSE);
    } else if ( length1 != length2) {
	return( FALSE);
    } else {
	for ( ii = 0; ii < length1; ii++) {
	    if ( filter1[ ii] != filter2[ ii]) {
		return( FALSE);
	    }
	}
	return( TRUE);
    }
}

/* compress a pyramid on the fly */
   
AmErrorCode AmCompressTree( memTilePtr    tile,
                          seekAddress   *losses,
                          int           *fds)
{
    int         compressedLength, fileIndex, allocBytes;
    seekAddress seekAddr;

/* terminate on a NULL tile */

    if ( tile == NULL) {
        return( AmOK);
    }

    if ( !( tile->fileTile.flags & AmTILE_COMPRESSED_FLAG)) {
        char    *block;

/* if the tile has data, seek to the current tile's seek index in the 
   tile's file */

	fileIndex = tile->fileTile.fileIndex;
	seekAddr = tile->fileTile.seekAddr;
	if ( tile->fileTile.numBytes > 0) {
	    if ( LSEEK( fds[ fileIndex], seekAddr, SEEK_SET) < 0) {
		return( AmSEEK_FAILED);
	    }

/* duplicate the file descriptor, and seek to the tile's location minus
   the amount lost due to compression.  Allocate a little extra space
   in the buffer in case the compression fails */

	    allocBytes = tile->fileTile.numBytes + COMPRESS_SLOP;
	    if (( block = ( char *) alloca( allocBytes)) == NULL) {
		return( AmALLOCA_FAILED);
	    }
	    compressedLength = compressBlock( fds[ fileIndex], block,
					      tile->fileTile.numBytes);

/* only compress the tile if the compressed version is shorter */

	    if ( compressedLength < (signed)tile->fileTile.numBytes) {
		if ( LSEEK( fds[ fileIndex], seekAddr - losses[ fileIndex],
			    SEEK_SET) < 0) {
		    return( AmSEEK_FAILED);
		}
		if ( write( fds[ fileIndex], block, compressedLength) < 0) {
		    return( AmWRITE_FAILED);
		}

AN_DEBUG( 40, ("compressing tile at %ld->%ld from %ld to %d\n",
               seekAddr, seekAddr - losses[ fileIndex],
               tile->fileTile.numBytes, compressedLength));

/* reset the seek address before adding the current tile reduction to the
   loss for this file.  Reset the numBytes and set the flag for compression */

		tile->fileTile.flags |= AmTILE_COMPRESSED_FLAG;
		tile->fileTile.seekAddr -= losses[ fileIndex];
		losses[ fileIndex] += tile->fileTile.numBytes - 
						compressedLength;
		tile->fileTile.numBytes = compressedLength;
	    } else {

AN_DEBUG( 40, ("tile at %ld->%ld failed to compress\n",
		seekAddr, seekAddr - losses[ fileIndex]));

/* even though the tile isn't compressed, we still have to shift it on disk */

		if ( LSEEK( fds[ fileIndex], seekAddr, SEEK_SET) < 0) {
		    return( AmSEEK_FAILED);
		}
		if ( read( fds[ fileIndex], block, 
			   tile->fileTile.numBytes) < 0) {
		    return( AmREAD_FAILED);
		}
		tile->fileTile.seekAddr -= losses[ fileIndex];
		if ( LSEEK( fds[ fileIndex], tile->fileTile.seekAddr, 
			    SEEK_SET) < 0) {
		    return( AmSEEK_FAILED);
		}
		if ( write( fds[ fileIndex], block, 
			    tile->fileTile.numBytes) < 0) {
		    return( AmWRITE_FAILED);
		}
	    }
	} else {

AN_DEBUG( 40, ("tile at %ld->%ld has 0 bytes\n", 
		seekAddr, seekAddr - losses[ fileIndex]));

/* even though there's no data, reset the seek address anyway */

	    tile->fileTile.seekAddr -= losses[ fileIndex];
	}
    }
    AmCompressTree( tile->llChild, losses, fds);
    AmCompressTree( tile->ulChild, losses, fds);
    AmCompressTree( tile->lrChild, losses, fds);
    AmCompressTree( tile->urChild, losses, fds);
    return( AmOK);
}

/* testing program to read back the tile header file and print the results */

static void recursiveReadAndPrintTileInfo( int fd)
{
    AmTileRec     tile;

    read( fd, &tile, sizeof( tile));
    printf( "%.2f %.2f %.2f %.2f\n",
            tile.xmin, tile.ymin, tile.xmin, tile.ymax);
    printf( "%.2f %.2f %.2f %.2f\n",
            tile.xmin, tile.ymax, tile.xmax, tile.ymax);
    printf( "%.2f %.2f %.2f %.2f\n",
            tile.xmax, tile.ymax, tile.xmax, tile.ymin);
    printf( "%.2f %.2f %.2f %.2f\n",
            tile.xmax, tile.ymin, tile.xmin, tile.ymin);
    printf( "seekAddrs = %ld, %ld, %ld, %ld, numBytes = %ld, seekAddr = %ld\n",
            tile.ulSeekAddr, tile.llSeekAddr,
            tile.urSeekAddr, tile.lrSeekAddr, tile.numBytes, tile.seekAddr);
    if ( tile.llSeekAddr != 0) {
        LSEEK( fd, tile.llSeekAddr, SEEK_SET);
        recursiveReadAndPrintTileInfo( fd);
    }
    if ( tile.ulSeekAddr != 0) {
        LSEEK( fd, tile.ulSeekAddr, SEEK_SET);
        recursiveReadAndPrintTileInfo( fd);
    }
    if ( tile.lrSeekAddr != 0) {
        LSEEK( fd, tile.lrSeekAddr, SEEK_SET);
        recursiveReadAndPrintTileInfo( fd);
    }
    if ( tile.urSeekAddr != 0) {
        LSEEK( fd, tile.urSeekAddr, SEEK_SET);
        recursiveReadAndPrintTileInfo( fd);
    }
}

/* debug routine to make sure the tiles were written OK */

AnvilBoolean AmReadAndPrintTileInfo( char      *tileFile)
{
    int         fd, numFileBytes, numFiles, ii, magicNumber;
    char        *fileStrings, *p;

    if (( fd = open( tileFile, O_RDONLY)) < 0) {
        fprintf( stderr, "failed to open tile output header %s\n", tileFile);
        return( FALSE);
    }

/* read the number of file bytes and the # of files AmFirst */

    read( fd, &magicNumber, sizeof( magicNumber));
    read( fd, &numFiles, sizeof( numFiles));
    read( fd, &numFileBytes, sizeof( numFileBytes));
    fileStrings = ( char *) alloca( numFileBytes);
    read( fd, fileStrings, numFileBytes);
    printf( "numfiles = %d, numFileBytes = %d\n", numFiles, numFileBytes);
    p = fileStrings;
    for ( ii = 0; ii < numFiles; ii++) {
        printf( "file %d: %s\n", ii, p);
        while ( *p++ != '\0');
    }
    recursiveReadAndPrintTileInfo( fd);
    close( fd);
    return( TRUE);
}


/* write the tiles out into a file */

static AnvilBoolean writeTiles( int         fd,
			   memTilePtr  tile)
{
    if ( tile == NULL) {
        return( FALSE);
    } else {
        write( fd, &tile->fileTile, sizeof( tile->fileTile));
        writeTiles( fd, tile->llChild);
        writeTiles( fd, tile->ulChild);
        writeTiles( fd, tile->lrChild);
        writeTiles( fd, tile->urChild);
        return( TRUE);
    }
}

/* assign the tile seek addresses to the tile pyramid */

static AnvilBoolean makeTileSeekAddrs( memTilePtr   memTile,
				  seekAddress  *seekAddr)
{
    if ( memTile == NULL) {
        return( FALSE);
    }
    *seekAddr += sizeof( AmTileRec);
    memTile->fileTile.llSeekAddr = *seekAddr;
    if ( !makeTileSeekAddrs( memTile->llChild, seekAddr)) {
        memTile->fileTile.llSeekAddr = 0;
    }
    memTile->fileTile.ulSeekAddr = *seekAddr;
    if ( !makeTileSeekAddrs( memTile->ulChild, seekAddr)) {
        memTile->fileTile.ulSeekAddr = 0;
    }
    memTile->fileTile.lrSeekAddr = *seekAddr;
    if ( !makeTileSeekAddrs( memTile->lrChild, seekAddr)) {
        memTile->fileTile.lrSeekAddr = 0;
    }
    memTile->fileTile.urSeekAddr = *seekAddr;
    if ( !makeTileSeekAddrs( memTile->urChild, seekAddr)) {
        memTile->fileTile.urSeekAddr = 0;
    }
    return( TRUE);
}

/* write the tile information out to the output file */

AnvilBoolean AmWriteTileInfo( char             *outputFile,
                       char             **filenames,
                       int              numFiles,
                       memTilePtr       tiles)
{
    int                 numFileBytes, ii, fd;
    seekAddress         tileSeekAddr, seekAddr;
    ulong_t             magicNumber = AmTILE_MAGIC;
    const char		*rootName;

    if (( fd = open( outputFile, O_CREAT | O_RDWR, PERMS)) < 0) {
        fprintf( stderr, "failed to open tile output header %s\n", outputFile);
        return( FALSE);
    }

/* find out the number of bytes needed to write each of the output files */

    numFileBytes = 0;
    for ( ii = 0; ii < numFiles; ii++) {
        numFileBytes += strlen( AmGetRootName( filenames[ ii])) + 1;
    }

/* write the magic number, then write the number of bytes and number of
   files, then write each of the files with an extra byte for the string
   terminator */

    write( fd, &magicNumber, sizeof( magicNumber));
    write( fd, &numFiles, sizeof( numFiles));
    write( fd, &numFileBytes, sizeof( numFileBytes));
    for ( ii = 0; ii < numFiles; ii++) {
	rootName = AmGetRootName( filenames[ ii]);
        write( fd, rootName, strlen( rootName) + 1);
    }
    seekAddr = sizeof( magicNumber) + sizeof( numFiles) +
               sizeof( numFileBytes) + numFileBytes;
    tileSeekAddr = seekAddr;

/* calculate the seek addresses for the tiles and write them out */

    makeTileSeekAddrs( tiles, &tileSeekAddr);
    writeTiles( fd, tiles);
    close( fd);
    return( TRUE);
}

/* debug function to print the contents of the tile tree */

void AmPrintTileTree( memTilePtr  tile)
{
    if ( tile == NULL) {
        return;
    }
    printf( "%.2f %.2f %.2f %.2f\n",
            tile->fileTile.xmin, tile->fileTile.ymin,
            tile->fileTile.xmax, tile->fileTile.ymax);
    printf( "seekAddr = %ld, numBytes = %ld, rows = %d, cols = %d\n", 
	    tile->fileTile.seekAddr, tile->fileTile.numBytes, 
	    tile->fileTile.rows, tile->fileTile.cols);
    printf( "child seekAddrs = %p, %p, %p, %p\n",
	    tile->llChild, tile->ulChild, tile->lrChild, tile->urChild);
    AmPrintTileTree( tile->llChild);
    AmPrintTileTree( tile->ulChild);
    AmPrintTileTree( tile->lrChild);
    AmPrintTileTree( tile->urChild);
}

/* convert the vector of sums into a vector of relative indices. */

void AmSumsToIndices( int *sums,
                    int *indices,
                    int length)
{
    int sum, ii;

    sum = 0;
    for ( ii = 0; ii < length; ii++) {
        indices[ ii] = sum;
        sum += sums[ ii];
    }
}
