#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#include <anvil/project.h>
#include <anvil/macros.h>
#include <anvil/colorcell.h>
#include <anvil/visual.h>
#include <anvil/dither.h>
#include <anvil/xcolor.h>

/* Color information for the application. */

static Colormap rgbColormap;
static XColor   rgbColors[ 256];

/* Number of shades used for each primary color in the color cube. */

static ushort_t numRedShades;
static ushort_t numGreenShades;
static ushort_t numBlueShades;

/* Offset tables into the color cube for specific shades of a color. */ 

static ulong_t redOffset[ 256];
static ulong_t greenOffset[ 256];
static ulong_t blueOffset[ 256];

uchar_t XdecisionMatrix[ NUM_DITHER_MATRICES][ DITHER_LEVEL][ DITHER_LEVEL];

static ulong_t redDitherMatrix[ 256];
static ulong_t greenDitherMatrix[ 256];
static ulong_t blueDitherMatrix[ 256];

ulong_t *
AmGetXRedDitherMatrix()
{
    return( redDitherMatrix);
}

ulong_t *
AmGetXGreenDitherMatrix()
{
    return( greenDitherMatrix);
}

ulong_t *
AmGetXBlueDitherMatrix()
{
    return( blueDitherMatrix);
}

/*
 * Function: AmGetAppColormap()
 *
 * Description:
 *     This function returns the global colormap to be used by the
 *     application.  In the case of a true color visual the colormap
 *     is empty.  While in the case of the pseudo color visual the
 *     colormap is a private colormap with the allocations for a color
 *     cube.
 */

Colormap
AmGetAppColormap()
{
    return( rgbColormap);
}

/*
 * Function: AmGetAppColors()
 *
 * Description:
 *     This function is only important for the pseudo color visual case.
 *     It returns the color cells used in the colorcube.
 */

const XColor *
AmGetAppColors()
{
    return( rgbColors);
}

/*
 * Function: GetNumRedShades(), GetNumBlueShades(), GetNumGreenShades()
 *
 * Description:
 *     These functions return the number of shades allocated per color in
 *     the color cube.  The total number of colors is the product of the
 *     three values.
 */

ushort_t
AmGetXNumRedShades()
{
    return( numRedShades);
}

ushort_t
AmGetXNumGreenShades()
{
    return( numGreenShades);
}

ushort_t
AmGetXNumBlueShades()
{
    return( numBlueShades);
}


/*
 * Function: GetRedOffsetLookup(), GetGreenOffsetLookup(),
 *           GetBlueOffsetLookup()
 *
 * Description:
 *     These functions return a lookup table that is used in translation
 *     from a 32-bit image to an n-bit image.  The lookup tables are only
 *     used for pseudo color visuals.  The pixel value to be used is generated
 *     by adding the values gotten by indexing the red, green, and blue
 *     component of a color into the offset lookup tables.  The resulting
 *     value (the sum) is then used as an index into the static color cube.
 *
 * For example:
 *
 *     redOffset = GetRedOffsetLookup();
 *     greenOffset = GetGreenOffsetLookup();
 *     blueOffset = GetBlueOffsetLookup();
 *     color = AmGetAppColors() + redOffset[ red] + greenOffset[ green] +
 *             blueOffset[ blue];
 *     pixel = color->pixel; 
 */

const ulong_t *
AmGetXRedOffsetLookup()
{
    return( redOffset);
}

const ulong_t *
AmGetXGreenOffsetLookup()
{
    return( greenOffset);
}

const ulong_t *
AmGetXBlueOffsetLookup()
{
    return( blueOffset);
}

/*
 * Function: createColormap( display, window, visual, size)
 *
 * Description:
 *     This function creates a colormap with the specified number of
 *     colors allocated as read/write colors.  If for some reason
 *     the number of specified colors can not be allocated 0 is
 *     returned.
 */

static Colormap
createColormap( Display *display,
                Window  window,
                Visual  *visual,
                ulong_t size) 
{
    Colormap cmap;
    ulong_t  pixels[ 256];

    cmap = XCreateColormap( display, window, visual, AllocNone);
    if ( !XAllocColorCells( display, cmap, False, NULL, 0, pixels, size)) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( 0);
    }

    return( cmap);
}


/*
 * Function: copyColormap( display, source, dest, start, size)
 *
 * Description:
 *     This function copies the colors from one colormap to the start
 *     of another colormap.
 */

static void
copyColormap( Display  *display,
              Colormap source,
              Colormap dest,
              int      start,
              int      size)
{
    int    ii;
    XColor colors[ 256];

/* Setup which colors are to be returned by the XQueryColors call. */

    for ( ii = 0; ii < size; ii++) {
        colors[ ii].pixel = start + ii;
    }

/*
 * Copy the colors from the source colormap to the beginning of the
 *   destination colormap.
 */
  
    XQueryColors( display, source, colors, size);
    XStoreColors( display, dest, colors, size);
}


/*
 * Function: removeUnusedColors( display, cmap, start)
 *
 * Description:
 *     This function frees the unused colors from the new colormap. 
 */

static void
removeUnusedColors( Display  *display,
                    Colormap cmap,
                    int      start)
{
    int     ii;
    ulong_t colors[ 256];
 
    for ( ii = 0; ii < start; ii++) {
        colors[ ii] = ii;
    }

    XFreeColors( display, cmap, colors, start, 0);
}


/*
 * Function: createColorcube( display, window, visual, pallette, cube,
 *                            numRed, numGreen, numBlue)
 *
 * Description:
 *     This function creates and fills the color cube used for 8 bit 
 *     displays in conjunction with ordered dithering.
 */

static AmErrorCode
createXColorCube( Display  	 *display, 
                  Window   	 window,
                  Visual   	 *visual,
                  Colormap 	 *cube, 
                  XColor   	 *colors, 
                  uint_t   	 numRed,
                  uint_t   	 numGreen,
                  uint_t   	 numBlue)
{
    uint_t   		appMaxColors, numTotal, startColor, ii;
    Colormap 		defaultCmap, cmap;
    AmColorCellRec	cubeColors[ 256];
    XColor		localColors[ 256];

    AmCreateColorCube( cubeColors, numRed, numGreen, numBlue, 
		     redOffset, greenOffset, blueOffset);
		     
/* Save some information about the color-cube. */

    numRedShades = numRed;
    numGreenShades = numGreen;
    numBlueShades = numBlue;

    appMaxColors = AmGetAppMaxColors();

    numTotal = numRed * numGreen * numBlue;
    cmap = XCreateColormap( display, window, visual, AllocAll);

/* Get the pixel values for the default colormap of the screen. */

    defaultCmap = DefaultColormap( display, DefaultScreen( display));
    for ( ii = 0; ii < appMaxColors; ii++) {
        localColors[ ii].pixel = ii;
    }
    XQueryColors( display, defaultCmap, localColors, appMaxColors);
    startColor = appMaxColors - numTotal;
    for ( ii = 0; ii < numTotal; ii++) {
	localColors[ ii + startColor].red = 
	    (( uint_t) cubeColors[ ii].red) << 8;
	localColors[ ii + startColor].green = 
	    (( uint_t) cubeColors[ ii].green) << 8;
	localColors[ ii + startColor].blue = 
	    (( uint_t) cubeColors[ ii].blue) << 8;
	localColors[ ii + startColor].flags = DoRed | DoGreen | DoBlue;
    }
    XStoreColors( display, cmap, localColors, appMaxColors);

/* once the colors have been allocated, shift the returned colors down so
   we index from 0 */

    for ( ii = 0; ii < numTotal; ii++) {
	colors[ ii] = localColors[ ii + startColor];
    }

    *cube = createColormap( display, window, visual, appMaxColors);
    copyColormap( display, cmap, *cube, 0, appMaxColors);
    removeUnusedColors( display, *cube, appMaxColors - numTotal);
    XFreeColormap( display, cmap);

    AmInitDitherMatrices( numRedShades, numGreenShades, numBlueShades,
			XdecisionMatrix, redDitherMatrix, greenDitherMatrix,
			blueDitherMatrix);

    return( AmOK);
}

/*
 * Function: AmInitColormaps( display, window, visual)
 *
 * Description:
 *     This function creates the standard colormaps to be used by
 *     the application.
 */

AmErrorCode
AmInitColormaps( Display *display,
               Window  window,
               Visual  *visual)
{
    switch( AmGetAppVisualClass()) {
        case PseudoColor:
            RETURN_NOT_OK( createXColorCube( display, window, visual,
                               &rgbColormap, rgbColors, 7, 8, 4));
            break;
        case TrueColor:
            rgbColormap = XCreateColormap( display, window, visual, AllocNone);
            break;
        default:
            RETURN_ERROR( AmNO_APPLICABLE_VISUAL);
    }

    return( AmOK);
}


/*
 * Function: AmFreeColormap( display)
 *
 * Description:
 *     This function frees the colormap currently in use by a display.
 */

void
AmFreeColormap( Display *display)
{
    if ( rgbColormap) {
        XFreeColormap( display, rgbColormap);
    }
}


/*
 * Function: AmSetPixelValue( cell)
 *
 * Description:
 *     This function sets the pixel value in a color cell record
 *     according to the visual used.  If it is a pseudo color
 *     display, the index into the color cube is used.  If it is
 *     a true color display, the pixel that represents the RGB
 *     values is used.
 */

void
AmSetPixelValue( AmColorCellPtr  cell)
{
    const XColor  *color;
    const ulong_t *redLookup, *greenLookup, *blueLookup;

    if ( AmGetAppVisualClass() == PseudoColor) {
        color = rgbColors + redOffset[ cell->red] + greenOffset[ cell->green] +
                blueOffset[ cell->blue];
        cell->pixel = color->pixel;
    } else {
        redLookup = AmGetRedPixelLookup();
        greenLookup = AmGetGreenPixelLookup();
        blueLookup = AmGetBluePixelLookup();
        cell->pixel = redLookup[ cell->red] | greenLookup[ cell->green] |
                      blueLookup[ cell->blue];
    }
}


AmColorCellRec
AmGetColorFromString( Display    *display,
                    const char *colorName)
{
    XColor              xColor;
    static AmColorCellRec color;

    XParseColor( display, AmGetAppColormap(), colorName, &xColor);
 
    color.red = xColor.red >> 8;
    color.green = xColor.green >> 8;
    color.blue = xColor.blue >> 8;
    AmSetPixelValue( &color);

    return( color);
}
