#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* Includes: */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <X11/Xlib.h>

#include <anvil/project.h>
#include <anvil/envdefs.h>
#include <anvil/resource.h>
#include <anvil/units.h>
#include <anvil/strlib.h>
#include <anvil/display.h>
#include <anvil/xfonts.h>

/* Evil global variables. */

static AnvilBoolean    processed = FALSE;
static AmFontPtr    fontCache = NULL;
static int        numFonts = 0;
static const char **fontFamilies;

typedef struct enumLookupRec_
{
    const char *string;
    int        enumeration;
} enumLookupRec, *enumLookupPtr;

static enumLookupRec weightTable[] = {
    { "*", AmWEIGHT_ANY, },
    { "bold", AmWEIGHT_BOLD, },
    { "medium", AmWEIGHT_MEDIUM, },
    { NULL, AmWEIGHT_NONE, }
};

static enumLookupRec slantTable[] = {
    { "*", AmSLANT_ANY, },
    { "i", AmSLANT_ITALIC, },
    { "o", AmSLANT_OBLIQUE, },
    { "r", AmSLANT_ROMAN, },
    { NULL, AmSLANT_NONE, }
};


int
AmFontCmp( AmFontPtr font1, AmFontPtr font2)
{
    int result;

    if ( font1->szFamily == NULL && font2->szFamily == NULL) {
        return( 0);
    } else if ( font1->szFamily == NULL) {
        return( -1);
    } else if ( font2->szFamily == NULL) {
        return( 1);
    }

    if ( (result = strcmp( font1->szFamily, font2->szFamily)) != 0) {
        return( result);
    }

    if (( font1->eWeight == AmWEIGHT_ANY) ||
        ( font2->eWeight == AmWEIGHT_ANY) ||
        ( font1->eWeight == font2->eWeight)) {

        if (( font1->eSlant == AmSLANT_ANY) ||
            ( font2->eSlant == AmSLANT_ANY) ||
            ( font1->eSlant == font2->eSlant)) {

            return( 0);

        } else {
            return( ( int)font1->eSlant - ( int)font2->eSlant);
        }

    } else {
       return( ( int)font1->eWeight - ( int)font2->eWeight);
    }
}


int
AmFindEnum( enumLookupPtr table, char *string) {
    int ii;

    if ( table == NULL || string == NULL) {
        return( 0);
    }

    for ( ii = 0; table[ ii].string != NULL; ii++) {
        if ( !strcmp( table[ ii].string, string)) {
            return( table[ ii].enumeration);
        }
    }

    return( 0);
}


/* This function returns all font families found in the font cache. */

static const char **
parseFamilies( AmFontPtr fonts, int num)
{
    int        ii, jj;
    const char *temp;
    const char **families;

    if ( processed == FALSE) {
        return( NULL);
    }

/* Find out how many unique font families there are. */

    temp = NULL;
    for ( ii = 0, jj = 0; ii < num; ii++) {
        if ( temp == NULL || strcmp( temp, fonts[ ii].szFamily)) {
            temp = fonts[ ii].szFamily;
            jj++;
        }
    }

/* Generate the vector to return. */

    if ( !( families = ( const char **) malloc( sizeof( char *) * ( jj + 1)))) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    temp = NULL;
    for ( ii = 0, jj = 0; ii < num; ii++) {
        if ( temp == NULL || strcmp( temp, fonts[ ii].szFamily)) {
            families[ jj] = fonts[ ii].szFamily;
            temp = fonts[ ii].szFamily;
            jj++;
        }
    }

/* Mark the end of the list of font families. */

    families[ jj] = NULL;

    return( families);
}


AmErrorCode
AmInitFontCache( Display *display)
{
    AmFontPtr cache;
    char    **list, **strings, buffer[ TEXTLEN];
    int     num, ii, jj;

    if ( processed == TRUE) {
        return( AmOK);
    }

    list = XListFonts( display, SCALEABLEFONTS, MAXFONTS, &num);

    if ( num == 0) {
        DO_RETURN_ERROR( AmNO_FONTS);
    }

    if ( !( cache = ( AmFontPtr) malloc( sizeof( AmFontRec) * num))) {
        XFreeFontNames( list);
        DO_RETURN_ERROR( AmMALLOC_FAILED);
    }

    for ( ii = 0; ii < num; ii++) {
        strings = AmChopString( list[ ii] + 1, '-');
       
        if ( strings != NULL) { 

            sprintf( buffer, "%s-%s", strings[ 0], strings[ 1]);
            cache[ ii].szFont = strdup( list[ ii]);
            cache[ ii].szFamily = strdup( buffer);
            cache[ ii].eWeight = ( AmFontWeightEnum) AmFindEnum( weightTable,
                                                      strings[ 2]);
            cache[ ii].eSlant = ( AmFontSlantEnum) AmFindEnum( slantTable,
                                                    strings[ 3]);
            for ( jj = 0; strings[ jj] != NULL; jj++) {
                free( strings[ jj]);
            }
            free( strings);
        } else {
            cache[ ii].szFont = NULL;
            cache[ ii].szFamily = NULL;
        }
    }

    qsort( cache, num, sizeof( AmFontRec), ( AmSearchFn) AmFontCmp);

    XFreeFontNames( list);

  /* Set the cache parameters. */

    processed = TRUE;
    numFonts = num;
    fontCache = cache;
    fontFamilies = parseFamilies( cache, num);

    return( AmOK);
}


const char *
AmGetFontFromCache( const char     *family,
                  AmFontWeightEnum weight, 
                  AmFontSlantEnum  slant)
{
    AmFontPtr font;
    AmFontRec fontKey;

    if ( family == NULL) {
        return( NULL);
    }

    if ( processed == FALSE) {
        return( NULL);
    }

    /* Try to match the font on the family, weight, and slant values. */ 

    fontKey.szFont = NULL;
    fontKey.szFamily = ( char *) family;
    fontKey.eWeight = weight;
    fontKey.eSlant = slant;

    if ( (font = ( AmFontPtr) bsearch( &fontKey, fontCache, numFonts,
                                sizeof( AmFontRec), ( AmSearchFn) AmFontCmp))) {
        return( font->szFont);
    }

    /* Ignore weight and try again. */ 

    fontKey.eWeight = AmWEIGHT_ANY;

    if ( (font = ( AmFontPtr) bsearch( &fontKey, fontCache, numFonts,
                                sizeof( AmFontRec), ( AmSearchFn) AmFontCmp))) {
        return( font->szFont);
    }

    /* Ignore slant and try again. */ 

    fontKey.eWeight = weight;
    fontKey.eSlant = AmSLANT_ANY;

    if ( (font = ( AmFontPtr) bsearch( &fontKey, fontCache, numFonts,
                                sizeof( AmFontRec), ( AmSearchFn) AmFontCmp))) {
        return( font->szFont);
    }

    /* Ignore weight and slant ... maybe now we'll find something. */

    fontKey.eWeight = AmWEIGHT_ANY;

    if ( (font = ( AmFontPtr) bsearch( &fontKey, fontCache, numFonts,
                                sizeof( AmFontRec), ( AmSearchFn) AmFontCmp))) {
        return( font->szFont);
    }

    return( fontCache[ 0].szFont);
}


Font
AmLoadFont( Display        *display,
          const char     *family,
          int            size,
          AmFontWeightEnum weight, 
          AmFontSlantEnum  slant,
          float          xScale,
          float          yScale,
          int            DPI)
{
    const char *name;
    char  newname[ TEXTLEN];
    int   resX, resY, ii, jj, field;

    if ( processed == FALSE) {
        if ( AmInitFontCache( display) != AmOK) {
            return( 0);
        }
    }

    name = AmGetFontFromCache( family, weight, slant);

    if ( name == NULL) {
        return( 0);
    }

    if ( xScale == 0.0) {
        xScale = 1.0;
    }

    if ( yScale == 0.0) {
        yScale = 1.0;
    }

    resX = ( int)( DPI * xScale);
    resY = ( int)( DPI * yScale);

    for ( ii = jj = field = 0; name[ ii] != '\0' && field <= 14; ii++) {
        newname[ jj++] = name[ ii];
        if ( name[ ii] == '-') {
            field++;
            switch( field) {
                case 7:	/* Pixel size. */
                case 12: /* Average width. */

                    newname[ jj++] = '*';
                    if ( name[ ii + 1] != '\0') {
                        ii++;
                    }
                    break;

                case 8:	/* Point size. */
                    sprintf( &newname[ jj], "%d", size*10);
                    while( newname[ jj] != '\0') {
                        jj++;
                    }
                    if ( name[ ii + 1] != '\0') {
                        ii++;
                    }
                    break;

                case 9:	/* X resolution. */
                case 10: /* Y resolution. */
                    sprintf( &newname[ jj], "%d", ( field == 9) ? resX : resY);
                    while( newname[ jj] != '\0') {
                        jj++;
                    }
                    while(( name[ ii + 1] != '-') && (name[ ii + 1] != '\0')) {
                        ii++;
                    }
                    break;
            }
        }
    }

    newname[ jj] = '\0';
    if ( field != 14) {
        return( 0);
    }

    return( XLoadFont( display, newname));
}

XFontStruct *
AmLoadQueryFont( Display        *display,
               const char     *family,
               int            size,
               AmFontWeightEnum weight, 
               AmFontSlantEnum  slant,
               float          xScale,
               float          yScale,
               int            DPI)
{
    Font fid; 

    fid = AmLoadFont( display, family, size, weight, slant, xScale, yScale, DPI);

    if ( fid == 0) {
        return( NULL);
    }

    return( XQueryFont( display, fid));
}

const char **
AmGetFontFamilies()
{
    return( fontFamilies);
}

