#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */



#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <anvil/project.h>
#include <anvil/shmimg.h>
#include <anvil/shmimgP.h>
#if defined(sparc)
#include <sys/byteorder.h>
#define BYTE_ORDER 4321
#elif defined(__linux__)
#include <endian.h>
#elif defined(__BSD44__) 
#else
#include <sys/endian.h>
#endif

/*
 * Function: swap32MSBFirst( pixel)
 *           swap32LSBFirst( pixel)
 *           swap16( pixel)
 *
 * Description:
 *     Reverses the byte order of a pixel.  In the case of the
 *     32 bit functions the function actually only looks at 3 bytes
 *     the red, green, and blue value.  The swap32MSB function ignores
 *     the LSB (least significant byte) while the LSB function ignores
 *     the MSB (most significant byte).
 */

static void
swap32MSBFirst( ulong_t *pixel)
{
    uchar_t     *bytes = ( uchar_t *) pixel;

    *pixel = bytes[ 3] << 24 | bytes[ 2] << 16 | bytes[ 1] << 8;
}

static void
swap32LSBFirst( ulong_t *pixel)
{
    uchar_t     *bytes = ( uchar_t *) pixel;

    *pixel = bytes[ 0] | bytes[ 1] << 8 | bytes[ 2] << 16;
}

static void
swap16( ushort_t *pixel)
{
    uchar_t     *bytes = ( uchar_t *) pixel, temp;

    temp = bytes[ 0];
    bytes[ 0] = bytes[ 1];
    bytes[ 1] = temp;
}


/*
 * Function: swapImage( image)
 *
 * Description:
 *     Reverse the byte ordering for an image.  The function takes
 *     an X image in any depth ( 8, 12, 15, 16, 24, 32) and does
 *     the apropriate byte swapping.  Note that in the 8 bit
 *     case there is really nothing to be done.
 *
 * Note:
 *     Since this function is only called specifically from the image
 *     draw routines, the depth of the image is already known, so it
 *     is possible to just split this up into a swapImage32 and
 *     swapImage16 and do away with the switch statement.
 */

static void
swapImage( XImage	*image)
{
    switch( image->depth) {
        case 8:
            break;
        case 12:
        case 15:
        case 16:
        {
            ushort_t *end, *ptr;

            end = ( ushort_t *)
                      ( image->data + image->height*image->bytes_per_line);

            for ( ptr = ( ushort_t *) image->data; ptr < end; ptr++) {
                swap16( ptr);
            }
            break;
        }
        case 24:
        case 32:
        {
            ulong_t *end, *ptr;

            end = ( ulong_t *) 
	        ( image->data + image->height*image->bytes_per_line);

            if ( image->byte_order == MSBFirst) {
                for ( ptr = ( ulong_t *) image->data; ptr < end; ptr++) {
                    swap32MSBFirst( ptr);
                }
            } else {
                for ( ptr = ( ulong_t *) image->data; ptr < end; ptr++) {
                    swap32LSBFirst( ptr);
                }
            }
            break;
        }
    }
}


/*
 * Function: unswapImage( image)
 *
 * Description:
 *     Reverse the byte ordering for an image.  The function takes
 *     an X image in any depth ( 8, 12, 15, 16, 24, 32) and does
 *     the apropriate byte swapping.  Note that in the 8 bit
 *     case there is really nothing to be done.  This function
 *     differs from the swapImage function in that it reverses the
 *     effect.  This is needed because there is a slight trick in
 *     the swap32 function in that it actually only looks at 3 bytes
 *     instead of all 4.
 */

static void
unswapImage( XImage *image)
{
    switch( image->depth) {
        case 8:
            break;
        case 12:
        case 15:
        case 16:
        {
            ushort_t *end, *ptr;

            end = ( ushort_t *)
                      ( image->data + image->height*image->bytes_per_line);

            for ( ptr = ( ushort_t *) image->data; ptr < end; ptr++) {
                swap16( ptr);
            }
            break;
        }
        case 24:
        case 32:
        {
            ulong_t *end, *ptr;

            end = ( ulong_t *) 
	        ( image->data + image->height*image->bytes_per_line);

            if ( image->byte_order == MSBFirst) {
                for ( ptr = ( ulong_t *) image->data; ptr < end; ptr++) {
                    swap32LSBFirst( ptr);
                }
            } else {
                for ( ptr = ( ulong_t *) image->data; ptr < end; ptr++) {
                    swap32MSBFirst( ptr);
                }
            }
            break;
        }
    }
}


/*
 * Function: AmShmShowImage( display, window, gc, image, src_x, src_y,
 *                         dest_x, dest_y, src_width, src_height,
 *                         shmInfoPtr)
 *
 * Description:
 *     Puts an image on the screen.  Depending on the depth of the
 *     image the appropriate put image function is called.
 */

void AmShmShowImage( Display 		*display,
		   Window		window,
		   GC			gc,
		   XImage 		*image,
		   int			src_x,
		   int			src_y,
		   int			dst_x,
		   int			dst_y,
		   int			src_width,
		   int			src_height,
		   XShmSegmentInfo	*shminfoPtr)
{
    int swapFlag, byteOrder;

    if ( shminfoPtr == NULL) {

    /* Make sure the image is in the correct byte order. */

        if ( BYTE_ORDER == 4321) {
            byteOrder = MSBFirst;
        } else {
            byteOrder = LSBFirst;
        }

        swapFlag = ( byteOrder != image->byte_order);
        if ( swapFlag) {
            swapImage( image);
        }

	XPutImage( display, window, gc, image,
                   src_x, src_y, dst_x, dst_y, src_width, src_height);

    /* Undo any swapping done previously. */

	if ( swapFlag) {
	    unswapImage( image);
	}
    } else {

/* Use the shared memory extension to put the image on the screen. */

	XShmPutImage( display, window, gc, image,
		      src_x, src_y, dst_x, dst_y, src_width, src_height, FALSE);
	XSync( display, FALSE);
    }
}
