#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>
#include <anvil/project.h>
#include <anvil/macros.h>
#include <anvil/visual.h>
#include <anvil/xcolor.h>
#if defined( sparc)
#include <sys/byteorder.h>
#define BYTE_ORDER 4321
#elif defined(__linux__)
#include <endian.h>
#elif defined(__BSD44__)
#else
#include <sys/endian.h>
#endif

void AmInitDitherMatrices();

/* Some information that gets re-used throughout the application. */

/* Info about the visual. */

static Visual   *appVisual;
static uint_t   appVisualDepth;
static uint_t   appVisualClass;
static uint_t   appMaxColors;

/* Translation tables from 24-bit to n-bit (where n = 8, 12, 15, 16, 24, 32) */

static ulong_t redPixelLookup[ 256];
static ulong_t greenPixelLookup[ 256];
static ulong_t bluePixelLookup[ 256];

static ulong_t redLookup[ 256];
static ulong_t greenLookup[ 256];
static ulong_t blueLookup[ 256];

/* Accessing functions for the static global variables. */

Visual *
AmGetAppVisual()
{
    return( appVisual);
}

int
AmGetAppVisualClass()
{
    return( appVisualClass);
}

int
AmGetAppDepth()
{
    return( appVisualDepth);
}

int
AmGetAppMaxColors()
{
    return( appMaxColors);
}

/*
 * Function: AmGetRedLookup(), AmGetGreenLookup(), AmGetBlueLookup()
 *
 * Description:
 *     These functions return a lookup table that is used in translation
 *     from a 32-bit image to an n-bit image.  The lookup tables are only
 *     used for true color visuals.  The pixel value to be used is generated
 *     by or'ing together the values gotten by indexing the red, green, and
 *     blue component of a color into the lookup tables.
 *
 * For example:
 *
 *     redLookup = AmGetRedLookup();
 *     greenLookup = AmGetGreenLookup();
 *     blueLookup = AmGetBlueLookup();
 *     pixel = redLookup[ red] | greenLookup[ green] | blueLookup[ blue];
 */

const ulong_t *
AmGetRedLookup()
{
    return( redLookup);
}

const ulong_t *
AmGetGreenLookup()
{
    return( greenLookup);
}

const ulong_t *
AmGetBlueLookup()
{
    return( blueLookup);
}

const ulong_t *
AmGetRedPixelLookup()
{
    return( redPixelLookup);
}

const ulong_t *
AmGetGreenPixelLookup()
{
    return( greenPixelLookup);
}

const ulong_t *
AmGetBluePixelLookup()
{
    return( bluePixelLookup);
}


/*
 * Function: getColorInfo( mask, shift, prec)
 *
 * Description:
 *     Calculates the amount to shift and the precision from the
 *     color mask.
 */

static void
getColorInfo( ulong_t mask, ulong_t *shift, ulong_t *prec)
{
    *shift = 0;
    *prec = 8;

    while ( !( mask & 0x1)) {
        *shift += 1;
        mask >>= 1;
    }

    while ( mask & 0x1) {
        *prec -= 1;
        mask >>= 1;
    }
}


/*
 * Function: generateLookup( table, shift, prec)
 *
 * Description:
 *     This function fills a color lookup table (256 entries) that
 *     map from 24 bit precision to the precision given.
 *
 */

static void
generateLookup( ulong_t *table,
                int     shift,
                int     prec)
{
    int ii;

    for ( ii = 0; ii < 256; ii++) {
        table[ ii] = ( ii >> prec) << shift;
    }
}


/*
 * Function: AmInitVisual( display)
 *
 * Description:
 *     This function determines the default visual for the display and
 *     the screen (usually the default screen).
 *
 *
 */

AmErrorCode
AmInitVisual( Display *display) {
    XVisualInfo  visualInfo;
    Visual       *defaultVisual;
    ulong_t      prec, shift;
    int          swapFlag, numBits, screenNum, defaultDepth;

/*
 * Scan through the visuals that are legal for the application and
 *   set the application's visual to the best match.
 */

    screenNum = DefaultScreen( display);
    defaultDepth = DefaultDepth( display, screenNum);
    defaultVisual = DefaultVisual( display, screenNum);

    if ( !XMatchVisualInfo( display, screenNum, defaultDepth,
              defaultVisual->c_class, &visualInfo)) {
        RETURN_ERROR( AmNO_APPLICABLE_VISUAL);
    }

    appVisual = visualInfo.visual;
    appVisualDepth = visualInfo.depth;
    appVisualClass = visualInfo.c_class;
    appMaxColors = visualInfo.colormap_size;

    if ( appVisualClass == TrueColor) {

    /*
     * Generate lookup tables for fast translation from a 24 bit
     *   image to the depth of the visual.
     */

        swapFlag = (( BYTE_ORDER == 1234) &!
            ( ImageByteOrder( display) == LSBFirst)) ||
            (( BYTE_ORDER == 4321) &!
            ( ImageByteOrder( display) == MSBFirst));

        numBits = visualInfo.depth;

        getColorInfo( visualInfo.red_mask, &shift, &prec);
        generateLookup( redPixelLookup, shift, prec);
        if ( swapFlag) {
            shift = numBits - prec - shift;
        }
        generateLookup( redLookup, shift, prec);
        getColorInfo( visualInfo.green_mask, &shift, &prec);
        generateLookup( greenPixelLookup, shift, prec);
        if ( swapFlag) {
            shift = numBits - prec - shift;
        }
        generateLookup( greenLookup, shift, prec);
        getColorInfo( visualInfo.blue_mask, &shift, &prec);
        generateLookup( bluePixelLookup, shift, prec);
        if ( swapFlag) {
            shift = numBits - prec - shift;
        }
        generateLookup( blueLookup, shift, prec);
    }

    return( AmOK);
}
