#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <anvil/xdisplay.h>
#include <anvil/shmimg.h>
#include <anvil/visual.h>
#include <anvil/xcolor.h>
#include <anvil/dither.h>
#include <anvil/log.h>

/*
 * Creates an instantiation of the Anvil XImage class.  This includes
 *   creating a standard X image according to the screen visual and
 *   if needed a seperate data area which is used for the 24 bit
 *   internal image.
 *
 */

AmXImageClass::AmXImageClass( AmDisplayClass	*pDisplay,
			  int		cxVirtual,
			  int		cyVirtual,
			  int		iDepth,
			  char		*pcData)
    : AmImageClass( pDisplay, cxVirtual, cyVirtual)
{
    XShmSegmentInfo **ppShmSegment;
    Display	    *pxDisplay;
    Visual	    *pxVisual;

    pxDisplay     = (( AmXDisplayClass *) _pDisplay)->GetXDisplay();
    pxVisual      = (( AmXDisplayClass *) _pDisplay)->GetXVisual();

    _pShmSegment  = NULL;

    ppShmSegment  =  &_pShmSegment;

/*
 * Create an image according to the passed in depth.  An image of depth
 *   one gets created as a bimodal image.
 */
  
    _pcInternal   = NULL;

    if ( iDepth == 1) {

        _iDepth         = 1;
        _iBytesPerPixel = 1;
        _iBitsPerPixel  = _iBytesPerPixel * 8;
        _iBytesPerLine  = _iBytesPerPixel * _cxPixels;

	_pxImage        = AmShmMakeImage( pxDisplay, pxVisual, iDepth, XYBitmap, 
                                        _cxPixels, _cyPixels, ppShmSegment);

        _iSize          = 0;

    } else {

        _iDepth         = 32;
        _iBytesPerPixel = 4;
        _iBitsPerPixel  = _iBytesPerPixel * 8;
        _iBytesPerLine  = _iBytesPerPixel * _cxPixels;

	_pxImage        = AmShmMakeImage( pxDisplay, pxVisual, iDepth, ZPixmap, 
                                        _cxPixels, _cyPixels, ppShmSegment);

        _iSize          = _cyPixels * _iBytesPerLine;

    /*
     * Set up internal data space for the 24-bit internal representation.
     *   Notice that the 32 bit depth corresponds to the internal data
     *   structure for now.  So there is no need for a different internal 
     *   representation.
     */

	if ( AmGetAppDepth() != 32) {

	    _pcInternal = ( char *) malloc( _iSize);
	    if ( _pcInternal == NULL) {
                AmAddResourceToLog( "mallocFailed", _iSize);
	    } else {
		memset( _pcInternal, 0, _iSize);
	    }

	} else {

	    _pcInternal = NULL;

	}
    }

    _iDDDepth         = _pxImage->depth;
    _iDDBitsPerPixel  = _pxImage->bits_per_pixel;
    _iDDBytesPerPixel = ( _iDDBitsPerPixel + 7 ) >> 3;
    _iDDBytesPerLine  = _pxImage->bytes_per_line;
    _iDDSize          = _cyPixels * _iDDBytesPerLine;

/* Copy over the data to the image. */

    if ( pcData != NULL) {
	SetData( pcData);
    }
}

/* 
 * Set the internal image according to the data passed in. The 
 *   representation of the data and the internal image have to
 *   match perfectly (32-bit 00BBGGRR).
 */

AnvilBoolean
AmXImageClass::SetData( char *pcData)
{
    char *pcTemp;
    int  ii;

/*
 * Copy over the data to the internal image.  Notice that there is no need
 *   for special processing since the data that gets passed in has to be in
 *   the same 32-bit format as the internal representation.
 */

    if (( _pcInternal != NULL) && ( _iDepth != 1)) {
        memcpy( _pcInternal, pcData, _iSize);
    } else if ( _iDepth == 1) {

        pcTemp = _pxImage->data;
        for ( ii = 0; ii < _cyPixels; ii++) {
            memcpy( ( void *) pcTemp, ( void *) pcData, _iBytesPerLine);
            pcData += _iBytesPerLine;
            pcTemp += _iDDBytesPerLine;
        }

    }

    return( TRUE);
}

/* Pseudo color copy data function...currently only supports 8 bit data. */

AnvilBoolean
AmXImageClass::copyDataPseudoColor()
{
    switch( _iDDDepth) {
        case 1:

        /* Nothing to be done in the 1 bit case. */

            return( TRUE);

        case 8:
            return( copyData8Bit());

        case 12:
        case 15:
        case 16:
            return( FALSE);

        default:
            return( FALSE);
    }
}

/* copy data from the internal image to an 8-bit dithered destination image */

AnvilBoolean
AmXImageClass::copyData8Bit()
{
    uchar_t       *pcSrc, *pcDst, *pcTemp;
    const XColor  *pxColor, *pxColorCube;
    ulong_t       r, g, b, ra, ga, ba;
    ushort_t      xMod, yMod;
    const ulong_t *vlRedDither, *vlGreenDither, *vlBlueDither;
    int           x, y;

/* Get the information needed for dithering. */

    pxColorCube   = AmGetAppColors();

    vlRedDither   = AmGetXRedDitherMatrix();
    vlGreenDither = AmGetXGreenDitherMatrix();
    vlBlueDither  = AmGetXBlueDitherMatrix();

/* Copy over the data to the image. */

    pcDst = ( uchar_t *) _pxImage->data;

    yMod = 0;
    for ( y = 0; y < _cyPixels; y++) {

        pcSrc  = ( uchar_t *) _pcInternal + y * _iBytesPerLine;
        pcTemp = pcDst + y * _iDDBytesPerLine;

        xMod = 0;
        for ( x = 0; x < _cxPixels; x++) {

    /* Use the red,green,blue dither indices to index the color cube for
       the destination pixel.  The xMod and yMod values stagger the dither
       index. */

            r  = vlRedDither[ pcSrc[ RED]];
            g  = vlGreenDither[ pcSrc[ GREEN]];
            b  = vlBlueDither[ pcSrc[ BLUE]];

            ra = (XdecisionMatrix[ SECONDWORD( r)][ yMod][ xMod]) ?
                     SECONDBYTE( r) : FIRSTBYTE( r);
            ga = (XdecisionMatrix[ SECONDWORD( g)][ yMod][ xMod]) ?
                     SECONDBYTE( g) : FIRSTBYTE( g);
            ba = (XdecisionMatrix[ SECONDWORD( b)][ yMod][ xMod]) ?
                     SECONDBYTE( b) : FIRSTBYTE( b);

            pxColor = pxColorCube + ra + ga + ba;
            *pcTemp = pxColor->pixel;

            xMod++;
            xMod &= 0x7;

            pcSrc  += _iBytesPerPixel;
            pcTemp += _iDDBytesPerPixel;
        }

        yMod++;
        yMod &= 0x7;
    }

    return( TRUE);
}

/* copy the internal image to a truecolor destination image */

AnvilBoolean
AmXImageClass::copyDataTrueColor()
{
    const ulong_t *vlRed, *vlGreen, *vlBlue;
    uchar_t       *pcDst, *pcSrc, *pcTemp;
    int           x, y;

    vlRed   = AmGetRedLookup();
    vlGreen = AmGetGreenLookup();
    vlBlue  = AmGetBlueLookup();

/* Copy over the data to the image. */

    pcDst = ( uchar_t *) _pxImage->data; 

    if ( _iDDBytesPerPixel == 4) {

        for ( y = 0; y < _cyPixels; y++) {

            pcSrc  = ( uchar_t * ) _pcInternal + y * _iBytesPerLine;
            pcTemp = pcDst + y * _iDDBytesPerLine;

            for ( x = 0; x < _cxPixels; x++) {

                * ( ulong_t *) pcTemp = vlRed[ pcSrc[ RED]] |
                                        vlGreen[ pcSrc[ GREEN]] |
                                        vlBlue[ pcSrc[ BLUE]];

                pcSrc  += _iBytesPerPixel;
                pcTemp += 4;

            }
        }

    } else if ( _iDDBytesPerPixel == 2) {

        for ( y = 0; y < _cyPixels; y++) {

            pcSrc  = ( uchar_t * ) _pcInternal + y * _iBytesPerLine;
            pcTemp = pcDst + y * _iDDBytesPerLine;

            for ( x = 0; x < _cxPixels; x++) {

                * ( ushort_t *) pcTemp = vlRed[ pcSrc[ RED]] |
                                         vlGreen[ pcSrc[ GREEN]] |
                                         vlBlue[ pcSrc[ BLUE]];

                pcSrc  += _iBytesPerPixel;
                pcTemp += 2;

            }
        }

    }

    return( TRUE);
}

/* 
 * Copy the internal image to the actual image, translation from the
 *   different depth is done here.
 */

AnvilBoolean
AmXImageClass::CopyData()
{
    if ( _pcInternal == NULL) {
	return( TRUE);
    }

/* Call the routines dealing with the correct visual. */ 

    switch( AmGetAppVisualClass()) {
        case StaticGray:
        case GrayScale:
        case StaticColor:
            AmAddResourceToLog( "visualNotSupported", AmGetAppVisualClass());
            return( FALSE);

        case PseudoColor:
            return( copyDataPseudoColor());

        case TrueColor:
        case DirectColor:
            return( copyDataTrueColor());

        default:
            return( FALSE);
    }
}

/* copy data from the display image back to the internal image */

AnvilBoolean 
AmXImageClass::RevCopyData()
{
    if ( _pcInternal == NULL) {
	return( TRUE);
    }

/* Call the routines dealing with the correct visual. */ 

    switch( AmGetAppVisualClass()) {
        case StaticGray:
        case GrayScale:
        case StaticColor:
            AmAddResourceToLog( "visualNotSupported", AmGetAppVisualClass());
            return( FALSE);

        case PseudoColor:
            return( revCopyDataPseudoColor());

        case TrueColor:
        case DirectColor:
            return( revCopyDataTrueColor());

        default:
            return( FALSE);
    }
}

/* switch for the different pseudo color depths */

AnvilBoolean
AmXImageClass::revCopyDataPseudoColor()
{
    switch( _iDDDepth) {
        case 1:

        /* Nothing to be done in the 1 bit case. */

            return( TRUE);

        case 8:
            return( revCopyData8Bit());

        case 12:
        case 15:
        case 16:
            return( FALSE);

        default:
            return( FALSE);
    }
}


/* copy data from 8-bit dithered destination _pxImage to the internal image */

AnvilBoolean
AmXImageClass::revCopyData8Bit()
{
    const XColor *pxColor, *pxColorCube;
    uchar_t      *pcSrc, *pcDst, *pcTemp;
    int          x, y;

    pxColorCube = AmGetAppColors();

/* Copy over the data to the image. */

    pcDst = ( uchar_t *) _pcInternal;

/* loop over y and x, and use the color values from the color cube to get 
   the RGB values for the internal image */
  
    for ( y = 0; y < _cyPixels; y++) {
        pcSrc = ( uchar_t *) _pxImage->data + y * _iDDBytesPerLine; 
        pcTemp = pcDst + y * _iBytesPerLine;

        for ( x = 0; x < _cxPixels; x++) {

	    pxColor = &pxColorCube[ *pcSrc];

	    * ( ulong_t *) pcTemp = (( pxColor->red >> 8) << RED_SHIFT) |
                                    (( pxColor->green >> 8) << GREEN_SHIFT) |
                                    (( pxColor->blue >> 8) << BLUE_SHIFT);


            pcTemp += _iBytesPerPixel;
            pcSrc  += _iDDBytesPerPixel;
        }
    }

    return( TRUE);
}

/* copy truecolor destination image back to the internal image */

AnvilBoolean
AmXImageClass::revCopyDataTrueColor()
{
    const ulong_t *vlRed, *vlGreen, *vlBlue;
    uchar_t       *pcSrc, *pcDst, *pcTemp;
    int           x, y;

    vlRed   = AmGetRedLookup();
    vlGreen = AmGetGreenLookup();
    vlBlue  = AmGetBlueLookup();

/* Copy over the data to the image. */

    pcDst = ( uchar_t *) _pcInternal;

    for ( y = 0; y < _cyPixels; y++) {

        pcSrc = ( uchar_t *) _pxImage->data + y * _iDDBytesPerLine;
        pcTemp = pcDst + y * _iBytesPerLine;

        for ( x = 0; x < _cxPixels; x++) {

            * ( ulong_t *) pcTemp = vlRed[ pcSrc[ RED]] |
                                    vlGreen[ pcSrc[ GREEN]] |
                                    vlBlue[ pcSrc[ BLUE]];

            pcTemp += _iBytesPerPixel; 
            pcSrc  += _iDDBytesPerPixel; 

        }
    }

    return( TRUE);
}

/* retreive the data for the image */

char *
AmXImageClass::AmGetData()
{
    if (( _iDepth != 1) && ( _pcInternal != NULL)) {
        return( _pcInternal);
    } else {
        return( _pxImage->data);
    }
}


/* Destroy the image. */

AmXImageClass::~AmXImageClass()
{
    if ( _pxImage != NULL) {
        if (( _pcInternal != _pxImage->data) && ( _pcInternal != NULL)) {
            free( _pcInternal);
        }
        _pcInternal = NULL;

        AmShmDestroyImage((( AmXDisplayClass *) _pDisplay)->GetXDisplay(), 
		        _pxImage, _pShmSegment);

        _pShmSegment = NULL;
        _pxImage    = NULL;
    }
}

/* Return the X Image. */

XImage *
AmXImageClass::GetXImage()
{
    return( _pxImage);
}

/* Return the X Image shared memory pointer. */

XShmSegmentInfo *
AmXImageClass::getShmInfo()
{
    return( _pShmSegment);
}
