#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>
#include <Xm/DrawingA.h>

#include <anvil/macros.h>
#include <anvil/strlib.h>
#include <anvil/events.h>
#include <anvil/va_cback.h>
#include <anvil/userevents.h>
#include <anvil/eventcodes.h>
#include <anvil/textcheck.h>

#include <editor/AlignDialogClass.h>
#include <editor/EditorClass.h>

const int AM_SELECTION_SET_CHANGED = 8001;
const int AmMARGIN = 5;


/*
 * Function: AmAlignDialogClass::initializeValues()
 *
 * Purpose:
 *   Initialize the member variables to sane values.
 *
 * Comments:
 *   The _pEditor variable does not (!) get set here.
 */

void
AmAlignDialogClass::initializeValues()
{
    _eAlignType       = AmALIGN_OBJECTS_EACH_OTHER;

    _dHorizSpacing    = 0.0;
    _dVertSpacing     = 0.0;

    _dHorizMargin     = 0.0;
    _dVertMargin      = 0.0;

    _pXDisplay        = NULL;
    _pXGC             = NULL;
    _pXWindow         = NULL;
}


/*
 * Function: AmAlignDialogClass::createInterface()
 *
 * Purpose:
 *   Creates the widgets for the alignment dialog.
 *
 * Comments:
 *   The alignment dialog consists of several areas.  The three main
 *   areas are the display, which is put in the upper left hand
 *   corner, the vertical and horizontal alignment options which,
 *   are to the right and below of the display respectively, and 
 *   the alignment type selection which is below the horizontal
 *   alignment options.
 *
 *   Additionally, the interface sets up callbacks to "listen" to
 *   changes in object rectangles, object positions, and the screen
 *   size in the editor so that the display can be updated on the
 *   fly.
 */


void
AmAlignDialogClass::createInterface()
{
    AmLabelClass *pwcLabel;
    Colormap   colormap;
    XColor     xColor;
    float      dTemp;
    int        iBackground;

/* Create a label that indicates which alignment is selected. */

    _pwcLabel = new AmLabelClass( _wForm, "alignLabel");
    _pwcLabel->AttachLeft();
    _pwcLabel->AttachTop();
    _pwcLabel->AttachRight();

/* Put a separator under the label. */

    _pwcTopSeparator = new AmSeparatorClass( _wForm, "separator");
    _pwcTopSeparator->AttachLeft();
    _pwcTopSeparator->AttachTop( _pwcLabel);
    _pwcTopSeparator->AttachRight();

/* Create the form that holds the vertical alignment options. */

    _pwcVertForm = new AmFormClass( _wForm, "verticalForm");
    _pwcVertForm->AttachTop( _pwcTopSeparator);
    _pwcVertForm->AttachRight();

/*
 * Create the radio buttons which allow the user to select which
 *   kind of alignment they want to perform.  I.e. alignment
 *   of the objects to each other, sequentially, or to the page.
 *   These buttons are placed at the bottom above the action area.
 */

    _pwcRadioPane = new AmRadioPaneClass( _wForm, "optionRowCol");
    _pwcRadioPane->AttachLeft();
    _pwcRadioPane->AttachBottom();

/* Each other: */

    _pwcObjects = _pwcRadioPane->AddButton( "objectsBtn",
                                            ( XtCallbackProc) AmToggleCallback,
                                            ( XtPointer) this);

/* Sequentially: */

    _pwcSequential = _pwcRadioPane->AddButton( "sequentialBtn",
                                               ( XtCallbackProc) AmToggleCallback,
                                               ( XtPointer) this);

/* Page: */

    _pwcPage = _pwcRadioPane->AddButton( "pageBtn",
                                         ( XtCallbackProc) AmToggleCallback,
                                         ( XtPointer) this);

/* Make the AmFirst button the default. */

    _pwcObjects->SetState( TRUE, FALSE);

/* Put a separator above the pane of radio buttons. */

    _pwcSeparator = new AmSeparatorClass( _wForm, "separator");
    _pwcSeparator->AttachLeft();
    _pwcSeparator->AttachRight();
    _pwcSeparator->AttachBottom( _pwcRadioPane);

/* Create the form that holds the horizontal alignment options. */

    _pwcHorizForm = new AmFormClass( _wForm, "horizontalForm");
    _pwcHorizForm->AttachLeft();
    _pwcHorizForm->AttachBottom( _pwcSeparator);

/*
 * Create the vertical alignment options for aligning objects to each
 *   other.
 */

    _pwcVertPane = new AmPaneClass( _pwcVertForm, "verticalRowCol");
    _pwcVertPane->AttachLeft();
    _pwcVertPane->AttachTop();
    _pwcVertPane->AttachRight();

    _pwcTop = new AmToggleButtonClass( _pwcVertPane, "topBtn",
                                     ( XtCallbackProc) AmToggleCallback,
                                     ( XtPointer) this);
    _pwcVertCenter = new AmToggleButtonClass( _pwcVertPane, "centerBtn",
                                            ( XtCallbackProc) AmToggleCallback,
                                            ( XtPointer) this);
    _pwcBottom = new AmToggleButtonClass( _pwcVertPane, "bottomBtn",
                                        ( XtCallbackProc) AmToggleCallback,
                                        ( XtPointer) this);

/*
 * Create the horizontal alignment options for aligning objects to each
 *   other.
 */

    _pwcHorizPane = new AmPaneClass( _pwcHorizForm, "horizontalRowCol");
    XtVaSetValues( *_pwcHorizPane, XmNorientation, XmHORIZONTAL, NULL);

    _pwcHorizPane->AttachLeft();
    _pwcHorizPane->AttachTop();
    _pwcHorizPane->AttachBottom();

    _pwcLeft = new AmToggleButtonClass( _pwcHorizPane, "leftBtn",
                                      ( XtCallbackProc) AmToggleCallback,
                                      ( XtPointer) this);
    _pwcHorizCenter = new AmToggleButtonClass( _pwcHorizPane, "centerBtn",
                                             ( XtCallbackProc) AmToggleCallback,
                                             ( XtPointer) this);
    _pwcRight = new AmToggleButtonClass( _pwcHorizPane, "rightBtn",
                                       ( XtCallbackProc) AmToggleCallback,
                                       ( XtPointer) this);

/*
 * Create the vertical alignment options for aligning objects
 *   to the page.
 */

    _pwcVMarginForm = new AmFormClass( _pwcVertForm, "vMarginForm");
    _pwcVMarginForm->AttachLeft();
    _pwcVMarginForm->AttachTop( _pwcVertPane);
    _pwcVMarginForm->AttachRight();

    pwcLabel = new AmLabelClass( _pwcVMarginForm, "marginLbl");
    pwcLabel->AttachLeft();
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

    _pwcVMarginText = new AmTextFieldClass( _pwcVMarginForm, "vMargin");
    _pwcVMarginText->AttachLeft( pwcLabel);
    _pwcVMarginText->AttachTop();
    _pwcVMarginText->AttachBottom();

    dTemp = 0.0;
    AmUnitVarEntry( *_pwcVMarginText, &dTemp, AmFLOAT,
                  AmUNIT_UNDEFINED, NULL);

    XtAddCallback( *_pwcVMarginText, XmNactivateCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);
    XtAddCallback( *_pwcVMarginText, XmNlosingFocusCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);

    pwcLabel = new AmLabelClass( _pwcVMarginForm, "unitsLbl");
    pwcLabel->AttachLeft( _pwcVMarginText);
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

/* Do not map them at startup, we only want to show the toggle buttons
   for one of the alignment types. */

    XtVaSetValues( *_pwcVMarginForm, XmNmappedWhenManaged, FALSE, NULL);

/*
 * Create the horizontal alignment options for aligning objects
 *   to the page.
 */

    _pwcHMarginForm = new AmFormClass( _pwcHorizForm, "hMarginForm");
    _pwcHMarginForm->AttachLeft( _pwcHorizPane);
    _pwcHMarginForm->AttachTop(); 
    _pwcHMarginForm->AttachBottom();

    pwcLabel = new AmLabelClass( _pwcHMarginForm, "marginLbl");
    pwcLabel->AttachLeft();
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

    _pwcHMarginText = new AmTextFieldClass( _pwcHMarginForm, "vMargin");
    _pwcHMarginText->AttachLeft( pwcLabel);
    _pwcHMarginText->AttachTop();
    _pwcHMarginText->AttachBottom();

    dTemp = 0.0;
    AmUnitVarEntry( *_pwcHMarginText, &dTemp, AmFLOAT,
                  AmUNIT_UNDEFINED, NULL);

    XtAddCallback( *_pwcHMarginText, XmNactivateCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);
    XtAddCallback( *_pwcHMarginText, XmNlosingFocusCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);

    pwcLabel = new AmLabelClass( _pwcHMarginForm, "unitsLbl");
    pwcLabel->AttachLeft( _pwcHMarginText);
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

/* Do not map them at startup, we only want to show the toggle buttons
   for one of the alignment types. */

    XtVaSetValues( *_pwcHMarginForm, XmNmappedWhenManaged, FALSE, NULL);

/*
 * Create the vertical alignment options for aligning objects
 *   sequentially.
 */

    _pwcVertSeqPane = new AmPaneClass( _pwcVertForm, "sequentialVRowCol");
    _pwcVertSeqPane->AttachLeft();
    _pwcVertSeqPane->AttachTop();
    _pwcVertSeqPane->AttachRight();

    _pwcVert = new AmToggleButtonClass( _pwcVertSeqPane, "vSequential",
                                     ( XtCallbackProc) AmToggleCallback,
                                     ( XtPointer) this);

    _pwcVertSeqForm = new AmFormClass( _pwcVertForm, "vSpacingForm");
    _pwcVertSeqForm->AttachLeft();
    _pwcVertSeqForm->AttachTop( _pwcVertSeqPane);
    _pwcVertSeqForm->AttachRight();

    pwcLabel = new AmLabelClass( _pwcVertSeqForm, "spacingLbl");
    pwcLabel->AttachLeft();
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

    _pwcVertSeqText = new AmTextFieldClass( _pwcVertSeqForm, "vSpacing");
    _pwcVertSeqText->AttachLeft( pwcLabel);
    _pwcVertSeqText->AttachTop();
    _pwcVertSeqText->AttachBottom();

    dTemp = 0.0;
    AmUnitVarEntry( *_pwcVertSeqText, &dTemp, AmFLOAT, AmUNIT_UNDEFINED, NULL);

    XtAddCallback( *_pwcVertSeqText, XmNactivateCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);
    XtAddCallback( *_pwcVertSeqText, XmNlosingFocusCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);

    pwcLabel = new AmLabelClass( _pwcVertSeqForm, "unitsLbl");
    pwcLabel->AttachLeft( _pwcVertSeqText);
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

/* Do not map them at startup, we only want to show the toggle buttons
   for one of the alignment types. */

    XtVaSetValues( *_pwcVertSeqPane, XmNmappedWhenManaged, FALSE, NULL);
    XtVaSetValues( *_pwcVertSeqForm, XmNmappedWhenManaged, FALSE, NULL);

/*
 * Create the horizontal alignment options for aligning objects
 *   sequentially.
 */

    _pwcHorizSeqPane = new AmPaneClass( _pwcHorizForm, "sequentialHRowCol");
    _pwcHorizSeqPane->AttachLeft();
    _pwcHorizSeqPane->AttachTop();
    _pwcHorizSeqPane->AttachBottom();

    XtVaSetValues( *_pwcHorizForm, XmNorientation, XmHORIZONTAL, NULL);

    _pwcHoriz = new AmToggleButtonClass( _pwcHorizSeqPane, "hSequential",
                                      ( XtCallbackProc) AmToggleCallback,
                                      ( XtPointer) this);

    _pwcHorizSeqForm = new AmFormClass( _pwcHorizForm, "hSpacingForm");
    _pwcHorizSeqForm->AttachLeft( _pwcHorizSeqPane);
    _pwcHorizSeqForm->AttachTop();
    _pwcHorizSeqForm->AttachRight();
    _pwcHorizSeqForm->AttachBottom();

    pwcLabel = new AmLabelClass( _pwcHorizSeqForm, "spacingLbl");
    pwcLabel->AttachLeft();
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

    _pwcHorizSeqText = new AmTextFieldClass( _pwcHorizSeqForm, "hSpacing");
    _pwcHorizSeqText->AttachLeft( pwcLabel);
    _pwcHorizSeqText->AttachTop();
    _pwcHorizSeqText->AttachBottom();

    dTemp = 0.0;
    AmUnitVarEntry( *_pwcHorizSeqText, &dTemp, AmFLOAT, AmUNIT_UNDEFINED, NULL);

    XtAddCallback( *_pwcHorizSeqText, XmNactivateCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);
    XtAddCallback( *_pwcHorizSeqText, XmNlosingFocusCallback,
                   ( XtCallbackProc) AmTextCallback,
                   ( XtPointer) this);

    pwcLabel = new AmLabelClass( _pwcHorizSeqForm, "unitsLbl");
    pwcLabel->AttachLeft( _pwcHorizSeqText);
    pwcLabel->AttachTop();
    pwcLabel->AttachBottom();

/* Do not map them at startup, we only want to show the toggle buttons
   for one of the alignment types. */

    XtVaSetValues( *_pwcHorizSeqPane, XmNmappedWhenManaged, FALSE, NULL);
    XtVaSetValues( *_pwcHorizSeqForm, XmNmappedWhenManaged, FALSE, NULL);

/* Add a frame to be displayed around the drawing area. */

    _pwcFrame = new AmFrameClass( _wForm, "frame", FALSE); 
    _pwcFrame->AttachLeft();
    _pwcFrame->AttachTop( _pwcTopSeparator);
    _pwcFrame->AttachRight( _pwcVertForm);
    _pwcFrame->AttachBottom( _pwcHorizForm);

/* Add a drawing area that will show the objects to be aligned. */

    _pwcArea = new AmDrawingAreaClass( _pwcFrame, "alignArea");

    XtAddCallback( *_pwcArea, XmNexposeCallback,
                   ( XtCallbackProc) AmExposeCallback,
                   ( XtPointer) this);
    XtAddCallback( *_pwcArea, XmNresizeCallback,
                   ( XtCallbackProc) AmExposeCallback,
                   ( XtPointer) this);

/* Add an apply button (cancel and help are handled outside of
   the dialog. */

    AddButton( "applyBtn",
               ( XtCallbackProc) AmApplyCallback,
               ( XtPointer) this);


/* Create the classes to draw with. */

    _pXDisplay = new AmXDisplayClass( *_pwcArea);
    _pXGC      = NULL;
    _pXWindow  = NULL;

/* Get the foreground and background colors of the drawing area and
   store them for future use. */

    _pDisplay = XtDisplay( *_pwcArea);

    XtVaGetValues( *_pwcArea,
                   XmNcolormap, &colormap,
                   XmNbackground, &iBackground,
                   NULL);

    xColor.pixel = iBackground;
    XQueryColor( _pDisplay, colormap, &xColor);

    _colBackground.red = xColor.red >> 8;
    _colBackground.green = xColor.green >> 8;
    _colBackground.blue = xColor.blue >> 8;

    _pXDisplay->AllocColor( &_colBackground);

/* Hook in user events that notify us when any of the objects change
   in the editor. */

    AmAddCallback( AM_PKG_ID, AM_SELECTION_SET_CHANGED,
                 ( callbackProc) AmRedrawCallback, this);
    AmAddCallback( AM_PKG_ID, AM_RECT_CHANGE,
                 ( callbackProc) AmRedrawCallback, this);

/* Hook in the expose callback when the drawing area of the editor
   gets resized. */

    XtAddCallback( _pEditor->GetMotifDisplay()->GetArea(), XmNresizeCallback,
                   ( XtCallbackProc) AmExposeCallback, this);

}

/*
 * Function: AmAlignDialogClass::AmAlignDialogClass()
 *
 * Purpose:
 *   Default constructor.
 */

AmAlignDialogClass::AmAlignDialogClass()
    : AmDialogClass()
{
}


/*
 * Functions: AmAlignDialogClass::AmAlignDialogClass( wParent, szName, pEditor)
 *            AmAlignDialogClass::AmAlignDialogClass( bwcParent, szName, pEditor)
 *
 * Purpose:
 *   Constructor for the AmAlignDialogClass.  Both functions work
 *   the same way except that one takes a widget and the other
 *   takes a "wrapped" widget.
 *
 * Parameters: 
 *   Widget          wParent    : The parent of the dialog.
 *   BaseWidgetclass *bwcParent : The "wrapped" parent of the dialog.
 *   const char      *szName    : The name of the dialog.
 *   AmEditorClass     *pEditor   : The editor to tie the alignment dialog to.
 *
 * Comments:
 *   Creates the alignment dialog.  The parent's constructor is called
 *   with the widget and name arguments so that everything gets
 *   initialized correctly.
 */

AmAlignDialogClass::AmAlignDialogClass( Widget      wParent,
                                    const char  *szName,
                                    AmEditorClass *pEditor)
    : AmDialogClass( wParent, szName)
{
    _pEditor = pEditor;

    initializeValues();
    createInterface();

}


/* See above comment. */

AmAlignDialogClass::AmAlignDialogClass( AmBaseWidgetClass *bwcParent,
                                    const char      *szName,
                                    AmEditorClass     *pEditor)
    : AmDialogClass( bwcParent, szName)
{
    initializeValues();
    createInterface();

    _pEditor = pEditor;
}


/*
 * Function: AmAlignDialogClass::~AmAlignDialogClass()
 *
 * Purpose:
 *   Destructor for the AmAlignDialogClass.
 */

AmAlignDialogClass::~AmAlignDialogClass()
{

/* Unhook in user events that notify us when any of the objects change
   in the editor. */

    AmRemoveCallback( AM_PKG_ID, AM_SELECTION_SET_CHANGED,
                    ( callbackProc) AmRedrawCallback, this);
    AmRemoveCallback( AM_PKG_ID, AM_RECT_CHANGE,
                    ( callbackProc) AmRedrawCallback, this);
    if ( _pXGC != NULL) {
	delete _pXGC;
    }
    if ( _pXWindow != NULL) {
	delete _pXWindow;
    }
    if ( _pXDisplay != NULL) {
	delete _pXDisplay;
    }
}


/*
 * Function: AmAlignDialogClass::AmExposeCallback( w, xpClientData, xpCallData)
 *           AmAlignDialogClass::expose( w, xpCallData)
 *
 * Purpose:
 *   The expose callback is responsbile for redrawing the contents
 *   of the alignment dialog's drawing area.
 *
 * Parameters:
 *   Widget    w            : Drawing area widget.
 *   XtPointer xpClientData : Client data which is a pointer to an
 *                            AmAlignDialogClass object.
 *   XtPointer xpCallData   : Data send along by X for the expose event.
 * 
 * Comments:
 *   The AmExposeCallback gets called when a section of the drawing
 *   area gets unobscured.  It is also hooked into the resize
 *   callback for the editor's drawing area.  That is the expose
 *   callback gets called when the editor's drawing area gets
 *   resized.
 */

void
AmAlignDialogClass::AmExposeCallback( Widget    w,
                                  XtPointer xpClientData,
                                  XtPointer xpCallData)
{
    AmAlignDialogClass *pAlignDialog;
    pAlignDialog = ( AmAlignDialogClass *) xpClientData;

    pAlignDialog->expose( w, xpCallData);
}

/* ARGSUSED */
void
AmAlignDialogClass::expose( Widget    w,
                          XtPointer xpCallData)
{
    draw();
}


/*
 * Function: AmAlignDialogClass::AmRedrawCallback( iID, xpClientData, xpCallData)
 *           AmAlignDialogClass::redraw( iID, xpCallData)
 *
 * Purpose:
 *   The redraw callback is a user event callback that is called when
 *   the editor's objects are either moved or resized.  It is responsible
 *   for redrawing the drawing area to represent the new location and
 *   sizes of the editor's objects.
 *
 * Parameters:
 *   int       iID          : Event ID. 
 *   XtPointer xpClientData : Client data which is a pointer to an
 *                            AmAlignDialogClass object.
 *   XtPointer xpCallData   : Data send along by X for the expose event.
 *
 */

void
AmAlignDialogClass::AmRedrawCallback( int       iID,
                                  XtPointer xpClientData,
                                  XtPointer xpCallData)
{
    AmAlignDialogClass *pAlignDialog = ( AmAlignDialogClass *) xpClientData;
    pAlignDialog->redraw( iID, xpCallData);
}


/* ARGSUSED */
void
AmAlignDialogClass::redraw( int       iID,
                          XtPointer xpCallData)
{
    if ( XtIsRealized( *_pwcArea)) {
        AmSendExpose( XtDisplay( *_pwcArea), XtWindow( *_pwcArea));
    }
}


/*
 * Function: AmAlignDialogClass::draw()
 *
 * Purpose:
 *   Draws outlines representing the editor's objects at the locations
 *   that they would be at if the alignment options were applied.  The
 *   pivoting object (AmFirst selected object) is drawn in red, any other
 *   selected object is drawn in green, and any object that is not
 *   selected, and thus does not play a role in the alignment, is drawn
 *   in grey.
 */

void
AmAlignDialogClass::draw()
{
    AmMotifDisplayClass  *pMotifDisplay;
    const AmRectangleRec *pRect;
    AmRectangleRec       *vRects;
    AmScreenObjClass     **vpScreenObj;
    Dimension          dimWidth, dimHeight;
    int                iWidth, iHeight, x, y,
                       iWidthTemp, iHeightTemp, nScreenObj, ii,
                       iWidthEditor, iHeightEditor;
    float              dScale;
    AmColorCellRec       colBlack, colWhite, colRed, colGreen, colGrey;

/* Don't do anything if the drawing area isn't realized. */

    if ( XtIsRealized( *_pwcArea)) {

        pMotifDisplay = _pEditor->GetMotifDisplay();

    /* Create a new window if we haven't done so already. */

        if ( _pXWindow == NULL) {
            _pXWindow = new AmXWindowClass( _pXDisplay, *_pwcArea);
            _pXGC = new AmXGCClass( _pXWindow);
        }

    /* Figure out the extends of our drawing area in virtual coordinates. */

        XtVaGetValues( *_pwcArea,
                       XmNwidth, &dimWidth,
                       XmNheight, &dimHeight,
                       NULL);

        iWidthTemp = _pXDisplay->VirtualCoord( dimWidth);
        iHeightTemp = _pXDisplay->VirtualCoord( dimHeight);

    /* Get the extends of the editor's drawing area in virtual coordinates. */

        iWidthEditor = pMotifDisplay->GetWidth();
        iHeightEditor = pMotifDisplay->GetHeight();

    /* Figure out the biggest scaling factor that allows us to draw the
       "fake" layout page in our drawing area.  That is, figure out the 
       biggest scaling that preserves the aspect ratio of the editor's
       drawing area. */

        if ( iHeightEditor > iWidthEditor) { 
            iHeight = iHeightTemp - 2 * AmMARGIN;
            iWidth = iHeight * iWidthEditor / iHeightEditor;
        } else {
            iWidth = iWidthTemp - 2 * AmMARGIN;
            iHeight = iWidth * iHeightEditor / iWidthEditor;
        }

        if ( iHeight > ( iHeightTemp - 2 * AmMARGIN)) {

            dScale = ( float)( iHeightTemp - 2 * AmMARGIN) / ( float) iHeight;
            iHeight = iHeightTemp - 2 * AmMARGIN; 
            iWidth = (int)(iWidth * dScale);

        } else if ( iWidth > ( iWidthTemp - 2 * AmMARGIN)) {

            dScale = ( float)( iWidthTemp - 2 * AmMARGIN) / ( float) iWidth;
            iHeight = (int)(iHeight * dScale);
            iWidth = iWidthTemp - 2 * AmMARGIN;

        }

   /* Center the "fake" page on our drawing area. */

        x = ( int)( ( iWidthTemp - iWidth) / 2.0 + 0.5),
        y = ( int)( ( iHeightTemp - iHeight) / 2.0 + 0.5),

   /* Draw the "fake" page in white. */

        colWhite.red = 0xff;
        colWhite.green = 0xff;
        colWhite.blue = 0xff;
  
        _pXDisplay->AllocColor( &colWhite);
        _pXGC->SetForegroundColor( &colWhite);

        _pXDisplay->FillRectangle( _pXWindow, _pXGC,
                                   x, y, iWidth, iHeight);

   /* Draw a border around it. */

        colBlack.red = 0x0;
        colBlack.green = 0x0;
        colBlack.blue = 0x0;
  
        _pXDisplay->AllocColor( &colBlack);

        _pXGC->SetForegroundColor( &colBlack);

        _pXDisplay->DrawRectangle( _pXWindow, _pXGC,
                                   x - 1, y - 1, iWidth + 2, iHeight + 2);

   /* Clear the four areas that are not filled by the "fake" page. */

        _pXGC->SetForegroundColor( &_colBackground);
        _pXDisplay->FillRectangle( _pXWindow, _pXGC, 0, 0,
                                   iWidthTemp - 1, y - 2);
        _pXDisplay->FillRectangle( _pXWindow, _pXGC, 0, 0,
                                   x - 2, iHeightTemp - 1);
        _pXDisplay->FillRectangle( _pXWindow, _pXGC, x + iWidth + 2, 0,
                                   x - 2, iHeightTemp - 1);
        _pXDisplay->FillRectangle( _pXWindow, _pXGC, 0, y + iHeight + 2,
                                   iWidthTemp - 1, y - 2);

    /* Allocate the colors used by the object outlines. */

        colRed.red = 0x96;
        colRed.green = 0x00;
        colRed.blue = 0x32;
  
        _pXDisplay->AllocColor( &colRed);
        _pXGC->SetForegroundColor( &colRed);

        colGreen.red = 0x00;
        colGreen.green = 0x64;
        colGreen.blue = 0x32;
  
        _pXDisplay->AllocColor( &colGreen);
        _pXGC->SetForegroundColor( &colGreen);

        colGrey.red = 0x64;
        colGrey.green = 0x64;
        colGrey.blue = 0x64;
  
        _pXDisplay->AllocColor( &colGrey);
        _pXGC->SetForegroundColor( &colGrey);

    /* Either the width or the height can be used to calculate the
       scaling for the objects. */

        dScale = iWidth / ( float) iWidthEditor;

    /* Grab all the screen objects. */

        vpScreenObj = pMotifDisplay->GetObjects();
        nScreenObj = AmVeclen( ( const char **) vpScreenObj);

    /* Draw the objects, that are not (!) selected, in grey. */

        _pXGC->SetForegroundColor( &colGrey);
        for ( ii = nScreenObj - 1; ii >= 0; ii--) {
            if ( _pEditor->IsObjectSelected( vpScreenObj[ ii]) == FALSE) {
                pRect = vpScreenObj[ ii]->GetRect();

                _pXDisplay->DrawRectangle( _pXWindow, _pXGC,
                                           x + pRect->x * dScale + 0.5,
                                           y + pRect->y * dScale + 0.5,
                                           pRect->width * dScale + 0.5,
                                           pRect->height * dScale + 0.5);
            }
        }

        free( vpScreenObj);

    /* Grab all the selected screen objects. */

        vpScreenObj = _pEditor->GetSelectedObjectsV();
        nScreenObj = AmVeclen( ( const char **) vpScreenObj);

    /* Create the outlines to be drawn. */

        vRects = createAlignRects( vpScreenObj, nScreenObj);

    /* Draw the pivoting (AmFirst selected) object in red, and the rest
       in green. */

        for ( ii = nScreenObj - 1; ii >= 0; ii--) {

            if ( ii == nScreenObj - 1) {
                _pXGC->SetForegroundColor( &colRed);
            } else {
                _pXGC->SetForegroundColor( &colGreen);
            }

            pRect = &vRects[ ii];

            _pXDisplay->DrawRectangle( _pXWindow, _pXGC,
                                       x + pRect->x * dScale + 0.5,
                                       y + pRect->y * dScale + 0.5,
                                       pRect->width * dScale + 0.5,
                                       pRect->height * dScale + 0.5);
        }

        free( vRects);
        free( vpScreenObj);
    }
}


/*
 * Function: AmAlignDialogClass::alignRect( left, top, right, bottom,
 *                                        pRectSrc, pRectDst)
 *
 * Purpose:
 *   Aligns a rectangle according to the selected alignment mode. 
 *   The left, top, right, and bottom values must be appropriately
 *   set (either to the AmFirst selected item, the previously aligned
 *   item, or the page size).
 *
 * Parameters:
 *   int                left      : The left corner of the rectangle to
 *                                  align the object to.
 *   int                top       : The top coordinate of the rectangle to
 *                                  align the object to.
 *   int                right     : The right coordinate of the rectangle to
 *                                  align the object to.
 *   int                bottom    : The bottom coordinate of the rectangle to
 *                                  align the object to.
 *   const AmRectangleRec *pRectSrc : The original rectangle of the object.
 *   AmRectangleRec       *pRectDst : The result of the alignment.
 */

void
AmAlignDialogClass::alignRect( int                  left, 
                               int                  top, 
                               int                  right, 
                               int                  bottom, 
                               const AmRectangleRec *pRectSrc,
                               AmRectangleRec       *pRectDst)
{
    int iVertMargin, iHorizMargin, iDPI;

/* Assign the default (result = source). */

    *pRectDst    = *pRectSrc;

/* Figure out the DPI used so that we can translate the inches,
   given in the margin, or the spacing/gap fields, into pixels. */

    iDPI = _pEditor->GetMotifDisplay()->GetDisplay()->GetDPIBase(); 

    iVertMargin  = 0;
    iHorizMargin = 0;


    if ( _eAlignType == AmALIGN_OBJECTS_SEQUENTIAL) {

    /* Objects are aligned sequentially to each other, so figure out the
       spacing AmBetween each object. */

        iVertMargin = (int)(_dVertSpacing * iDPI);
        iHorizMargin = (int)(_dHorizSpacing * iDPI);

        if ( _pwcVert->GetState()) {
            pRectDst->y = bottom + iVertMargin;
        }

        if ( _pwcHoriz->GetState()) {
            pRectDst->x = right + iHorizMargin;
        }
    } else {

        if ( _eAlignType == AmALIGN_OBJECTS_PAGE) {

        /* Objects are aligned to the page, so figure out the margin 
           to the page borders. */

            iVertMargin = (int)(_dVertMargin * iDPI);
            iHorizMargin = (int)(_dHorizMargin * iDPI);
        }

        if ( _pwcTop->GetState()) { 

        /* Tops aligned. */

            pRectDst->y = top + iVertMargin;
            if ( _pwcBottom->GetState()) {

            /* Also bottoms aligned (sretch the object). */

                pRectDst->height = bottom - top - 2 * iVertMargin;
            }
        } else if ( _pwcBottom->GetState()) {

        /* Only bottoms aligned (move the object). */

            pRectDst->y = bottom - pRectSrc->height - iVertMargin;
        } else if ( _pwcVertCenter->GetState()) {

        /* Center the object. */

            pRectDst->y = ( bottom + top - pRectSrc->height) / 2;
        }

        if ( _pwcLeft->GetState()) { 

        /* Lefts aligned. */

            pRectDst->x = left + iHorizMargin;
            if ( _pwcRight->GetState()) {

            /* Also rights aligned (stretch the object). */

                pRectDst->width = right - left - 2 * iHorizMargin;
            }
        } else if ( _pwcRight->GetState()) {

        /* Only rights aligned (move the object). */

            pRectDst->x = right - pRectSrc->width - iHorizMargin;
        } else if ( _pwcHorizCenter->GetState()) {

        /* Center the object. */

            pRectDst->x = ( right + left - pRectSrc->width) / 2;
        }
    }
}


/*
 * Function: AmAlignDialogClass::createAlignRects( vpScreenObj, nScreenObj)
 *
 * Purpose:
 *   Creates the outline rectangles of the passed in screen objects
 *   alignining them either to each other, or to the page extends.
 *
 * Parameters:
 *   AmScreenObjClass **vpScreenObj: Vector of pointers to screen objects to
 *                                 be aligned.
 *   int            nScreenObj   : The number of screen objects in the vector.
 *
 * Returns:
 *   A vector of rectangles each rectangle representing the outline of
 *   the object in the same position as the object in the vector of
 *   objects.
 */

AmRectangleRec *
AmAlignDialogClass::createAlignRects( AmScreenObjClass **vpScreenObj,
                                      int              nScreenObj)
{
    int                left, right, top, bottom, ii;
    const AmRectangleRec *pRect;
    AmRectangleRec       *vRects;

/* If there are no screen objects, we are done. */

    if ( nScreenObj == 0) {
        return( NULL);
    }

    vRects = ( AmRectangleRec *) malloc( sizeof( AmRectangleRec) * nScreenObj);
    if ( vRects == NULL) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

/* Get the rectangle that everything gets aligned to. */

    if ( _eAlignType == AmALIGN_OBJECTS_PAGE) {

    /* The objects are aligned to the page. */

        left = 0;
        top = 0;
        right = _pEditor->GetMotifDisplay()->GetWidth();
        bottom = _pEditor->GetMotifDisplay()->GetHeight();
    } else {

    /* The objects are aligned to each other, so start with the AmFirst
       selected object. */

        pRect = vpScreenObj[ nScreenObj - 1]->GetRect();

        left   = pRect->x;
        top    = pRect->y;
        right  = pRect->x + pRect->width;
        bottom = pRect->y + pRect->height;
    }

/* Align each of the selected objects. */

    for ( ii = nScreenObj - 1; ii >= 0; ii--) {
        pRect = vpScreenObj[ ii]->GetRect();

        if (( ii == nScreenObj - 1) && ( _eAlignType != AmALIGN_OBJECTS_PAGE)) {

        /* Nothing to be done. */

            vRects[ ii] = *pRect;

        } else {
            alignRect( left, top, right, bottom, pRect, &vRects[ ii]); 
        }

    /* If we are aligning the rectangles sequentially, then the next rectangle
       depends on the current one. */


        if ( _eAlignType == AmALIGN_OBJECTS_SEQUENTIAL) {
            left   = vRects[ ii].x;
            top    = vRects[ ii].y;
            right  = left + vRects[ ii].width;
            bottom = top + vRects[ ii].height;
        }
    }

    return( vRects);
}


/*
 * Function: AmAlignDialogClass::AmToggleCallback( w, xpClientData, xpCallData)
 *           AmAlignDialogClass::toggle( w, xpCallData)
 *
 * Purpose:
 *   The toggle callback is called when one of the toggle buttons 
 *   in the alignment dialog changes state.   The callback updates
 *   the other toggle button and redraws the display.
 *
 * Parameters:
 *   Widget    w            : Toggle button area widget.
 *   XtPointer xpClientData : Client data which is a pointer to an
 *                            AmAlignDialogClass object.
 *   XtPointer xpCallData   : Data send along by X for the toggle event.
 */

void
AmAlignDialogClass::AmToggleCallback( Widget    w,
                                  XtPointer xpClientData,
                                  XtPointer xpCallData)
{
    AmAlignDialogClass *pAlignDialog = ( AmAlignDialogClass *) xpClientData;
    pAlignDialog->toggle( w, xpCallData);
}

/* ARGSUSED */
void
AmAlignDialogClass::toggle( Widget    w,
                          XtPointer xpCallData)
{
    XmString xsz;

/* First check the vertical toggle buttons. */

    if ( w == (Widget) *_pwcTop) {
        _pwcVertCenter->SetState( FALSE);
    } else if ( w == (Widget) *_pwcBottom) {
        _pwcVertCenter->SetState( FALSE);
    } else if ( w == (Widget) *_pwcVertCenter) {
        _pwcTop->SetState( FALSE);
        _pwcBottom->SetState( FALSE);
    }

/* Check the horizontal toggle buttons. */

    if ( w == (Widget) *_pwcLeft) {
        _pwcHorizCenter->SetState( FALSE);
    } else if ( w == (Widget) *_pwcRight) {
        _pwcHorizCenter->SetState( FALSE);
    } else if ( w == (Widget) *_pwcHorizCenter) {
        _pwcLeft->SetState( FALSE);
        _pwcRight->SetState( FALSE);
    }

/* Check the alignment type selection. */

    if ( w == (Widget) *_pwcObjects) {
        if ( _pwcObjects->GetState()) {
            _eAlignType = AmALIGN_OBJECTS_EACH_OTHER;

        /* Update the alignment label. */

            xsz = XmStringCreate( "Aligning objects to each other:",
                                  XmFONTLIST_DEFAULT_TAG);
            XtVaSetValues( *_pwcLabel, XmNlabelString, xsz, NULL); 
            XmStringFree( xsz);

        /* Only display the toggle buttons for aligning objects 
           to each other. */
             
            XtMapWidget( *_pwcVertPane);
            XtMapWidget( *_pwcHorizPane);
            XtUnmapWidget( *_pwcHMarginForm);
            XtUnmapWidget( *_pwcVertSeqPane);
            XtUnmapWidget( *_pwcHorizSeqPane);
            XtUnmapWidget( *_pwcVertSeqForm);
            XtUnmapWidget( *_pwcVMarginForm);
            XtUnmapWidget( *_pwcHorizSeqForm);

        }
    } else if ( w == (Widget) *_pwcSequential) {
        if ( _pwcSequential->GetState()) {
            _eAlignType = AmALIGN_OBJECTS_SEQUENTIAL;

        /* Update the alignment label. */

            xsz = XmStringCreate( "Aligning objects sequentially:",
                                  XmFONTLIST_DEFAULT_TAG);
            XtVaSetValues( *_pwcLabel, XmNlabelString, xsz, NULL); 
            XmStringFree( xsz);

        /* Only display the toggle buttons and text fields for aligning
           objects sequentially. */

            XtUnmapWidget( *_pwcVertPane);
            XtUnmapWidget( *_pwcHorizPane);
            XtMapWidget( *_pwcVertSeqPane);
            XtMapWidget( *_pwcHorizSeqPane);
            XtMapWidget( *_pwcVertSeqForm);
            XtMapWidget( *_pwcHorizSeqForm);
            XtUnmapWidget( *_pwcVMarginForm);
            XtUnmapWidget( *_pwcHMarginForm);
        }
    } else if ( w == (Widget) *_pwcPage) {
        if ( _pwcPage->GetState()) {
            _eAlignType = AmALIGN_OBJECTS_PAGE;

        /* Update the alignment label. */

            xsz = XmStringCreate( "Aligning objects to page:",
                                  XmFONTLIST_DEFAULT_TAG);
            XtVaSetValues( *_pwcLabel, XmNlabelString, xsz, NULL); 
            XmStringFree( xsz);

        /* Only display the toggle buttons and text fields for aligning
           objects to the page. */

            XtMapWidget( *_pwcVertPane);
            XtMapWidget( *_pwcHorizPane);
            XtUnmapWidget( *_pwcVertSeqPane);
            XtUnmapWidget( *_pwcHorizSeqPane);
            XtUnmapWidget( *_pwcVertSeqForm);
            XtUnmapWidget( *_pwcHorizSeqForm);
            XtMapWidget( *_pwcVMarginForm);
            XtMapWidget( *_pwcHMarginForm);
        }
    }

/* Update the outlines. */

    draw();
}


/*
 * Function: AmAlignDialogClass::AmTextCallback( w, xpClientData, xpCallData)
 *           AmAlignDialogClass::text( w, xpCallData)
 *
 * Purpose:
 *   The text callback is called when one of the text fields buttons 
 *   in the alignment dialog changes.  It effects the margins to the
 *   page, as well as the spacing AmBetween items.
 *   
 *
 * Parameters:
 *   Widget    w            : Text field widget.
 *   XtPointer xpClientData : Client data which is a pointer to an
 *                            AmAlignDialogClass object.
 *   XtPointer xpCallData   : Data send along by X for the text event.
 */

void
AmAlignDialogClass::AmTextCallback( Widget    w,
                                XtPointer xpClientData,
                                XtPointer xpCallData)
{
    AmAlignDialogClass *pAlignDialog = ( AmAlignDialogClass *) xpClientData;
    pAlignDialog->text( w, xpCallData);
}


/* ARGSUSED */
void
AmAlignDialogClass::text( Widget    w,
                        XtPointer xpCallData)
{
    char  *sz;
    float dTemp;

/* Get the value of the text field .*/

    XtVaGetValues( w, XmNvalue, &sz, NULL);
    AmScanType( sz, &dTemp, AmFLOAT, AmUNIT_UNDEFINED);
    XtFree( sz);

/* Assign it to the appropriate member variable. */

    if ( w == (Widget) *_pwcVMarginText) {
        _dVertMargin = dTemp;
    } else if ( w == (Widget) *_pwcHMarginText) {
        _dHorizMargin = dTemp;
    } else if ( w == (Widget) *_pwcVertSeqText) {
        _dVertSpacing = dTemp;
    } else if ( w == (Widget) *_pwcHorizSeqText) {
        _dHorizSpacing = dTemp;
    }

/* Update the display. */

    draw();
}


void
AmAlignDialogClass::AmApplyCallback( Widget    w,
                                 XtPointer xpClientData,
                                 XtPointer xpCallData)
{
    AmAlignDialogClass *pAlignDialog = ( AmAlignDialogClass *) xpClientData;
    pAlignDialog->apply( w, xpCallData);
}


/* ARGSUSED */
void
AmAlignDialogClass::apply( Widget    w,
                         XtPointer xpCallData)
{
    AmMotifDisplayClass *pMotifDisplay;
    AmScreenObjClass    **vpScreenObj;
    AmRectangleRec      *vRects;
    int               nScreenObj, cx, cy, ii;

/* Figure out the extends of the drawing area in the editor. */

    pMotifDisplay = _pEditor->GetMotifDisplay();

    cx = pMotifDisplay->GetWidth();
    cy = pMotifDisplay->GetHeight();
    
/* Grab all the selected screen objects. */

    vpScreenObj = _pEditor->GetSelectedObjectsV();
    if ( vpScreenObj == NULL) {
        return;
    }

    nScreenObj = AmVeclen( ( const char **) vpScreenObj);

/* Create the objects to be drawn. */

    vRects = createAlignRects( vpScreenObj, nScreenObj);
    if ( vRects == NULL) {
        free( vpScreenObj);
        return;
    }

/* First we unselect all the objects, only to reselect them. */

    _pEditor->UnselectAll();

/* Change the objects' rectangles. */ 

    for ( ii = nScreenObj - 1; ii >= 0; ii--) {
        vpScreenObj[ ii]->SetRectAndMargins( vRects[ ii].x,
                                             vRects[ ii].y,
                                             vRects[ ii].width,
                                             vRects[ ii].height,
                                             cx, cy);
        _pEditor->SelectObject( vpScreenObj[ ii]);
    }

/* Make it all happen on screen. */

    pMotifDisplay->Flush();

    free( vpScreenObj);
    free( vRects);
}
