#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <X11/cursorfont.h>

#include <anvil/defaults.h>
#include <anvil/strlib.h>
#include <anvil/state_hand.h>
#include <anvil/user_data.h>
#include <anvil/userevents.h>
#include <anvil/events.h>
#include <anvil/display.h>
#include <anvil/screenobj.h>
#include <anvil/datagraphic.h>
#include <anvil/overlay.h>
#include <anvil/postscript.h>
#include <anvil/tif_ximage.h>
#include <anvil/template.h>
#include <anvil/createobject.h>
#include <anvil/errlib.h>

#include <editor/handlers.h>
#include <editor/EditorClass.h>


#define ISTAG( buffer, tag)                                       \
    ( !strncmp( &buffer[ 1], tag, strlen( tag)))
#define GETINT( buffer, pad, temp)                                \
    ( sscanf( &buffer[ 1], "%[_A-Za-z]=%d", pad, &temp) == 2)
#define GETFLOAT( buffer, pad, temp)                              \
    ( sscanf( &buffer[ 1], "%[_A-Za-z]=%f", pad, &temp) == 2)
#define GETSTRING( buffer, pad, temp)                             \
    ( sscanf( &buffer[ 1], "%[_A-Za-z]=%s", pad, temp) == 2)
#define INRANGE( value, min, max)                                 \
    (( value >= min) && ( value <= max))



AnvilBoolean
AmIsMap( const char *szFilename)
{
    AnvilBoolean fResult;
    FILE    *pFile;
    char    sz[ FIELDLEN];
   
    fResult = TRUE;

    pFile = fopen( szFilename, "r");
    if ( pFile == NULL) {
        return( FALSE);
    }

    if ( fgets( sz, sizeof( sz), pFile) == NULL) {
        fResult = FALSE;
    }
    if ( fResult) {
	if ( sz[ 0] != '#') {
	    fResult = FALSE;
	}
	if ( !ISTAG( sz, AM_HEADER_STRING)) {
	    fResult = FALSE;
	}
    }
    fclose( pFile);

    return( fResult);
}


/*********************************************************************
 *
 * Function: AmErrorCode SaveObjectToFile( szFileName, pScreenObj,
 *                                         fAppend)
 *
 * Parameters:
 *     char           *szFileName : The file to write the object to.
 *     AmScreenObjClass *pScreenObj : The object to save.
 *     AnvilBoolean        fAppend     : Whether to append the object to
 *                                  the exisiting objects in the file, or
 *                                  whether to clear the file AmFirst. 
 *
 * Description:
 *     SaveObjectToFile writes the fields of the object given by
 *     pScreenObj to the file given by the file name szFileName.
 *     If fAppend is TRUE the object gets written to the end of
 *     the file, otherwise the old file gets cleared and the
 *     object gets written to the beginning of the file.
 *
 * Returns:
 *     AmErrorCode : AmOK if everything went as planned, the error
 *                   code otherwise. 
 * 
 * Errors:
 *     AmOPEN_FAILED
 *
 *********************************************************************/

AmErrorCode
AmEditorClass::SaveObjectToFile( char             *szFileName, 
                                 AmScreenObjClass *pScreenObj,
                                 AnvilBoolean          fAppend)
{
    AmOverlayClass     *pOverlay;
    AmDataGraphicClass **ppDataGraphic;
    FILE               *pFile;
    int                ii, nDataGraphic;

    /* Open the file for writing or appending. */
    
    if ( fAppend == FALSE) {

        pFile = fopen( szFileName, "w");
        if ( pFile == NULL) {
            AmAddResourceToLog( "openFailed", szFileName);
            DO_RETURN_ERROR( AmOPEN_FAILED);
        }

    } else {

        pFile = fopen( szFileName, "a");
        if ( pFile == NULL) {
            AmAddResourceToLog( "openFailed", szFileName);
            DO_RETURN_ERROR( AmOPEN_FAILED);
        }

    }

    /* Write the fields of the object to the file. */

    /* NOTE: there currently is no error code returned by
             DumpFields - should there be? */

    pScreenObj->DumpFields( pFile);

    /* 
     * If the object is an overlay, write the fields of all
     *   its' children, too.
     */

    if ( pScreenObj->IsType( AmOVERLAY) == TRUE) {
        pOverlay = ( AmOverlayClass *) pScreenObj;
        pOverlay->GetLayers( &ppDataGraphic, &nDataGraphic);
        for ( ii = 0; ii < nDataGraphic; ii++) {
            ppDataGraphic[ ii]->DumpFields( pFile);
        }
        free( ppDataGraphic);
    }

    fclose( pFile);
 
    return( AmOK); 
}


/*********************************************************************
 *
 * Function: AmListPtr LoadObjectsFromFile( szFileName)
 *
 * Parameters:
 *     char           *szFileName : The file to read the objects from.
 *
 * Description:
 *
 * Returns:
 * 
 * Errors:
 *
 *********************************************************************/

AmListPtr
AmEditorClass::LoadObjectsFromFile( char *szFileName)
{
    AmListPtr        lstObjects;
    AmScreenObjClass **vpScreenObj;
    int            ii;

    lstObjects = NULL;

/* Restore the objects from the file. */

    vpScreenObj = AmParseTemplateFile( szFileName,
                                     _pMotifDisplay->GetWindow(),
                                     ( AmCreateObjectFn) AmCreateObject,
                                     ( XtPointer) this);

    if ( vpScreenObj == NULL) {
        return( NULL);
    }

    /* Put the objects into a list. */

    for ( ii = 0;
          vpScreenObj[ ii] != NULL;
          ii++) {

        /* Remove any invalid objects. */

        if ( vpScreenObj[ ii]->IsValid() == FALSE) {
            delete vpScreenObj[ ii];
            continue;
        }

        lstObjects = AmInsertNodeAtEndLst( vpScreenObj[ ii], lstObjects);
    }

    free( vpScreenObj);
    return( lstObjects); 
}

/*********************************************************************
 *
 * Function: WriteHeaderInfo( szFileName)
 *
 * Parameters:
 *
 * Description:
 *
 * Returns:
 * 
 * Errors:
 *
 *********************************************************************/

void
AmEditorClass::WriteHeaderInfo( FILE *pFile)
{
    fprintf( pFile, "#%s\n", AM_HEADER_STRING);
    fprintf( pFile, "#%s=%f\n", AM_PAGE_WIDTH_STRING, _dWidthInInches);
    fprintf( pFile, "#%s=%f\n", AM_PAGE_HEIGHT_STRING, _dHeightInInches); 
    fprintf( pFile, "#%s=%d\n", AM_PRINT_DPI_STRING, _iPrintDPI);
    fprintf( pFile, "#%s=%d\n", AM_SCREEN_DPI_STRING, _iScreenDPI);
    fprintf( pFile, "#%s=%d\n", AM_LANDSCAPE_STRING, _fLandscape);
    fprintf( pFile, "#%s=%d\n", AM_TRUECOLOR_STRING, _fTrueColor);
    fprintf( pFile, "#%s=%s\n", AM_PROLOG_FILE_STRING, _szPrologFile);
    fprintf( pFile, "#%s\n", AM_END_HEADER_STRING);

}

AnvilBoolean
AmEditorClass::ReadHeaderInfo( FILE *pFile)
{
    char  sz[ FIELDLEN], szPad[ FIELDLEN], szTemp[ FIELDLEN];
    float dTemp;
    int   iTemp;

    if ( fgets( sz, sizeof( sz), pFile) == NULL) {
        AmAddResourceToLog( "errorEmptyMapFile");
        return( FALSE);
    }

    if ( sz[ 0] != '#') {
        AmAddResourceToLog( "errorMissingHeader");
        return( FALSE);
    }

    if ( !ISTAG( sz, AM_HEADER_STRING)) {
        AmAddResourceToLog( "errorMissingHeader");
        return( FALSE);
    }

    while ( fgets( sz, sizeof( sz), pFile) != NULL) {

        if ( sz[ 0] != '#') {
            break;
        }

        if ( ISTAG( sz, AM_END_HEADER_STRING)) {
            break;
        } else if ( ISTAG( sz, AM_PAGE_WIDTH_STRING)) {

            if ( !GETFLOAT( sz, szPad, dTemp)) {
                AmAddResourceToLog( "errorPageWidthParse");
                return( FALSE);
            }

            if ( !INRANGE( dTemp, AM_MIN_PAGE_WIDTH, AM_MAX_PAGE_WIDTH)) {
                AmAddResourceToLog( "errorPageWidthRange");
                return( FALSE);
            }

            _dWidthInInches = dTemp;

        } else if ( ISTAG( sz, AM_PAGE_HEIGHT_STRING)) {

            if ( !GETFLOAT( sz, szPad, dTemp)) {
                AmAddResourceToLog( "errorPageHeightParse");
                return( FALSE);
            }

            if ( !INRANGE( dTemp, AM_MIN_PAGE_HEIGHT, AM_MAX_PAGE_HEIGHT)) {
                AmAddResourceToLog( "errorPageHeightRange");
                return( FALSE);
            }

            _dHeightInInches = dTemp;

        } else if ( ISTAG( sz, AM_PRINT_DPI_STRING)) {

            if ( !GETINT( sz, szPad, iTemp)) {
                AmAddResourceToLog( "errorPrintDpiParse");
                return( FALSE);
            }

            if ( !INRANGE( iTemp, AM_MIN_PRINT_DPI, AM_MAX_PRINT_DPI)) {
                AmAddResourceToLog( "errorPrintDpiRange");
                return( FALSE);
            }

            _iPrintDPI = iTemp;

        } else if ( ISTAG( sz, AM_SCREEN_DPI_STRING)) {

            if ( !GETINT( sz, szPad, iTemp)) {
                AmAddResourceToLog( "errorScreenDpiParse");
                return( FALSE);
            }

            if ( !INRANGE( iTemp, AM_MIN_SCREEN_DPI, AM_MAX_SCREEN_DPI)) {
                AmAddResourceToLog( "errorScreenDpiRange");
                return( FALSE);
            }

            _iScreenDPI = iTemp;

        } else if ( ISTAG( sz, AM_LANDSCAPE_STRING)) {

            if ( !GETINT( sz, szPad, iTemp)) {
                AmAddResourceToLog( "errorLandscapeParse");
                return( FALSE);
            }

            if ( !INRANGE( iTemp, AM_MIN_BOOLEAN, AM_MAX_BOOLEAN)) {
                AmAddResourceToLog( "errorLandscapeRange");
                return( FALSE);
            }

            _fLandscape = iTemp;

        } else if ( ISTAG( sz, AM_TRUECOLOR_STRING)) {

            if ( !GETINT( sz, szPad, iTemp)) {
                AmAddResourceToLog( "errorTrueColorParse");
                return( FALSE);
            }

            if ( !INRANGE( iTemp, AM_MIN_BOOLEAN, AM_MAX_BOOLEAN)) {
                AmAddResourceToLog( "errorTrueColorRange");
                return( FALSE);
            }

            _fTrueColor = iTemp;

        } else if ( ISTAG( sz, AM_PROLOG_FILE_STRING)) {

            if ( !GETSTRING( sz, szPad, szTemp)) {
                AmAddResourceToLog( "errorPrologStringParse");
                return( FALSE);
            }
  
            if ( _szPrologFile != NULL) {
                free( _szPrologFile);
            }

            _szPrologFile = strdup( szTemp);

        }

    }

    return( TRUE);
}

AmErrorCode
AmEditorClass::SaveMapToFile( char *szFileName)
{
    AmOverlayClass     *pOverlay;
    AmScreenObjClass   **vpScreenObj;
    AmDataGraphicClass **ppDataGraphic;
    FILE             *pFile;
    int              ii, jj, nDataGraphic;

    /* Open the file for writing. */
    
    pFile = fopen( szFileName, "w");
    if ( pFile == NULL) {
        AmAddResourceToLog( "openFailed", szFileName);
        DO_RETURN_ERROR( AmOPEN_FAILED);
    }

    WriteHeaderInfo( pFile);

    vpScreenObj = _pMotifDisplay->GetObjects();
    for ( ii = 0; vpScreenObj[ ii] != NULL; ii++) {

        /* Write the fields of the object to the file. */

        vpScreenObj[ ii]->DumpFields( pFile);

        /* 
         * If the object is an overlay, write the fields of all
         *   its' children, too.
         */

        if ( vpScreenObj[ ii]->IsType( AmOVERLAY) == TRUE) {
            pOverlay = ( AmOverlayClass *) vpScreenObj[ ii];
            pOverlay->GetLayers( &ppDataGraphic, &nDataGraphic);
            for ( jj = 0; jj < nDataGraphic; jj++) {
                ppDataGraphic[ jj]->DumpFields( pFile);
            }
            free( ppDataGraphic);
        }
    }

    fclose( pFile);
    free( vpScreenObj);

    return( AmOK);
}


AnvilBoolean
AmEditorClass::LoadMapFromFile( char *szFileName)
{
    AmScreenObjClass **vpScreenObj, **vpScreenObjOld;
    FILE           *pFile;
    int            ii, nOld, iScreenWidth, iScreenHeight;

// Open the file for writing.
    
    pFile = fopen( szFileName, "r");
    if ( pFile == NULL) {
        AmAddResourceToLog( "openFailed", szFileName);
        return( FALSE);
    }

// Read the map header information.

    if ( !ReadHeaderInfo( pFile)) {
        AmAddResourceToLog( "errorReadHeader", szFileName);
        return( FALSE);
    }

    rewind( pFile);

    vpScreenObj = AmParseTemplateFp( pFile,
                                   _pMotifDisplay->GetWindow(),
                                   ( AmCreateObjectFn) AmCreateObject,
                                   ( XtPointer) this);

    fclose( pFile);

    if ( vpScreenObj == NULL) {
        AmAddResourceToLog( "errorNoObjects", szFileName);
        return( FALSE);
    }

    if  ( vpScreenObj[ 0] == NULL) {
        AmAddResourceToLog( "errorNoObjects", szFileName);
        return( FALSE);
    }

// Remove the old objects.

    vpScreenObjOld = _pMotifDisplay->GetObjects();
    nOld = AmVeclen(( const char **) vpScreenObjOld); 

    _pMotifDisplay->RemoveObjectSet( vpScreenObjOld, nOld);

// 
// Update the rulers (don't redraw them yet, since they will get
//   drawn when the drawing area gets resized.
//

    _pwcVerticalRuler->SetDPI( _iScreenDPI, FALSE);
    _pwcHorizontalRuler->SetDPI( _iScreenDPI, FALSE);

// Resize the drawing area according to the new values.

    iScreenWidth = (int)(_dWidthInInches * _iScreenDPI);
    iScreenHeight = (int)(_dHeightInInches * _iScreenDPI);

    _pMotifDisplay->GetDisplay()->SetDPI( _iScreenDPI);

    XtVaSetValues( _wArea,
                   XmNwidth, iScreenWidth,
                   XmNheight, iScreenHeight,
                   NULL);

    for ( ii = 0; vpScreenObj[ ii] != NULL; ii++) { 
        _pMotifDisplay->AddObject( vpScreenObj[ ii]);
    }

    free( vpScreenObj);

// Update the screen to show the new stuff.

    _pMotifDisplay->Flush();

    return( TRUE);
}


AmErrorCode
AmEditorClass::PrintMapToFile( char           *szFileName,
                             AmOutputTypeEnum eOutputType)
{
    AmWindowClass        *pWindow=NULL;
    AmDisplayClass       *pDisplay=NULL;
    AmGCClass            *pGC=NULL;

    Display            *display = XtDisplay( _wArea);
    Window             window = XtWindow( _wArea);
    GC                 gc;

    XImage             *xImage; 
    Pixmap             pixmap=0;
    Colormap           colormap;
    XColor             xColor;
    XGCValues          xGCV;

    AmColorCellRec       colBackground;

    FILE               *pFile;

    AmScreenObjClass     **vpScreenObj;
    const AmRectangleRec *pRec;

    int                iOutputWidth, iOutputHeight,
                       ii, x, y, iWidth, iHeight,
                       iEventCode = AM_SET_FIELD_OBJECT;
    char               szTempFile[] = "/tmp/AmTempFileXXXXXX"; 

    mktemp( szTempFile);

    if ( SaveMapToFile( szTempFile) != AmOK) {
        return( ANVIL_ERROR_CODE);
    }

    iOutputWidth = (int)(_dWidthInInches * _iPrintDPI);
    iOutputHeight = (int)(_dHeightInInches * _iPrintDPI);

    if ( eOutputType == AmOUTPUT_POSTSCRIPT) {

        pFile = fopen( szFileName, "w");
        if ( pFile == NULL) {
            AmAddResourceToLog( "openFailed", szFileName);
            DO_RETURN_ERROR( AmOPEN_FAILED);
        }

        if ( _fTrueColor == TRUE) {
            pDisplay = new AmPostScriptClass( 24, AmTRUE_COLOR, _iPrintDPI);
        } else {
            pDisplay = new AmPostScriptClass( 8, AmPSEUDO_COLOR, _iPrintDPI);
        }

        pWindow = new AmPostScriptOutputClass( pFile,
                                             _dWidthInInches,
                                             _dHeightInInches,
                                             ( AmPostScriptClass *) pDisplay,
                                             _fLandscape,
                                             _szPrologFile);
        pGC = pWindow->CreateGC(); 

    } else if ( eOutputType == AmOUTPUT_TIFF) {

        pDisplay = new AmXDisplayClass( _wArea);

        /* Create a pixmap the size of the resulting image. */

        pixmap = XCreatePixmap( display, window,
                                iOutputWidth, iOutputHeight,
                                DefaultDepth( display,
                                              DefaultScreen( display)));

        /* Get color information to fill the pixmap with. */ 

        XtVaGetValues( _wArea,
                       XmNbackground, &xColor.pixel,
                       XmNcolormap, &colormap,
                       NULL);

        XQueryColor( display, colormap, &xColor);

        /* Fill the pixmap in the widget's background color. */

        xGCV.foreground = xColor.pixel; 
        gc = XCreateGC( display, window, GCForeground, &xGCV);
        XFillRectangle( display, pixmap, gc, 0, 0, iOutputWidth + 1,
                        iOutputHeight + 1);
        XFreeGC( display, gc);

        /* Convert from X color to the anvil color representation. */

        colBackground.red = xColor.red >> 8;
        colBackground.green = xColor.green >> 8;
        colBackground.blue = xColor.blue >> 8;

        pWindow = new AmXWindowClass(( AmXDisplayClass *) _pMotifDisplay->GetDisplay(),
                                    pixmap, &colBackground); 
        pGC = new AmXGCClass(( AmXWindowClass *) pWindow);

    }

    vpScreenObj = AmParseTemplateFile( szTempFile, pWindow,
                                     ( AmCreateObjectFn) AmCreateObject,
                                     ( XtPointer) this);

    if ( vpScreenObj == NULL) {

        delete pWindow;
        delete pGC;
        delete pDisplay;
  
        unlink( szTempFile);

        return( ANVIL_ERROR_CODE);
    }

    /* All of the object's have to be updated. */

    for ( ii = 0; vpScreenObj[ ii] != NULL; ii++) {
        pRec = vpScreenObj[ ii]->GetRect();
        x = pRec->x;
        y = pRec->y;
        iWidth = pRec->width;
        iHeight = pRec->height;
        vpScreenObj[ ii]->SetRect( x, y, iWidth, iHeight);
        vpScreenObj[ ii]->Update( &iEventCode, 1);
    }

    /* Now we are ready to draw the objects. */

    for ( ii = 0; vpScreenObj[ ii] != NULL; ii++) {
        pRec = vpScreenObj[ ii]->GetExtentsRect();
        pGC->SetClipRectangle( pRec->x, pRec->y, pRec->width, pRec->height);
        vpScreenObj[ ii]->Draw( pWindow, pGC, pRec, 1);       
    }

    /* Get rid of all the objects we created. */

    for ( ii = 0; vpScreenObj[ ii] != NULL; ii++) {
        delete vpScreenObj[ ii];
    }
    free( vpScreenObj);

    /*
     * If we were printing to the pixmap we need to grab the contents
     *   and print them to a tiff file.
     */

    if ( eOutputType == AmOUTPUT_TIFF) {
        xImage = XGetImage( display, pixmap, 0, 0, iOutputWidth, iOutputHeight, 
                            0xffffff, ZPixmap);
        AmWriteXImageToTIFF( display, colormap, xImage, szFileName);
        XDestroyImage( xImage);

    }
 
    /* Free any memory used by the new window, display, and gc. */

    delete pWindow;
    delete pGC;
    delete pDisplay;

    unlink( szTempFile);

    return( AmOK);
}



AmErrorCode
AmEditorClass::ScaleMap( int iDPI)
{
    int iBaseDPI, cx, cy;
    int iOldCursor;

    _iScreenDPI = iDPI;

    iBaseDPI = _pXDisplay->GetDPIBase();
    cx       = (int)(_dWidthInInches * iBaseDPI);
    cy       = (int)(_dHeightInInches * iBaseDPI);

    iOldCursor = _pMotifDisplay->SetCursor( XC_watch);
    _pXWindow->ClearRect( NULL);

/* Change the DPI for the rulers, don't update them yet, since they
   will get a resize event from the drawing area anyways. */

    _pwcVerticalRuler->SetDPI( _iScreenDPI, FALSE);
    _pwcHorizontalRuler->SetDPI( _iScreenDPI, FALSE);
 
    cx = (int)(_dWidthInInches * _iScreenDPI);
    cy = (int)(_dHeightInInches * _iScreenDPI);

    _pXDisplay->SetDPI( _iScreenDPI);

    XtVaSetValues( _wArea,
                   XmNwidth, cx,
                   XmNheight, cy,
                   NULL);

    _pMotifDisplay->SetCursor( iOldCursor);

    return( AmOK);
}
