#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <X11/cursorfont.h>

#include <anvil/defaults.h>
#include <anvil/state_hand.h>
#include <anvil/user_data.h>
#include <anvil/userevents.h>

#include <editor/handlers.h>
#include <editor/EditorClass.h>

#define EDITOR_STRING "EditorUserData"

const int AM_SELECTION_SET_CHANGED = 8001;

/*********************************************************************
 *
 * Function: AmScreenObjClass *AmEditorClass::FindSelectedObject( x, y)
 *
 * Parameters:
 *     int x: The x location of the cursor. 
 *     int y: The y location of the cursor.
 *
 * Description:
 *     Finds the top most object that the coordinate <x, y> falls 
 *     within. 
 *
 * Returns:
 *     AmScreenObjClass *: If a screen object exists under the coordinate 
 *                       a pointer to it is returned, otherwise NULL
 *                       is returned.
 *
 *********************************************************************/

AmScreenObjClass *
AmEditorClass::FindSelectedObject( int x, int y)
{
    AmSelectObjClass     *pSelectObj;
    AmScreenObjClass     *pScreenObj;
    const AmRectangleRec *pRec;
    AmListPtr            lstTemp;
   
    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj= ( AmSelectObjClass *) lstTemp->data;
        pScreenObj = pSelectObj->GetScreenObj();
        pRec = pScreenObj->GetRect();

        /* Check to see if the mouse is within the extends of this object. */ 

        if ( x >= pRec->x && x <= pRec->x + pRec->width &&
             y >= pRec->y && y <= pRec->y + pRec->height) {
            return( pScreenObj);
        }
    }

    return( NULL);
}

/*********************************************************************
 *
 * Function:
 *
 * Parameters:
 *
 * Description:
 *
 * Returns:
 *
 *********************************************************************/


AmErrorCode
AmEditorClass::SelectObject( AmScreenObjClass *pScreenObj)
{
    AmSelectObjClass     *pSelectObj;
    AmObjectOutlineClass *pOutline;
    const AmRectangleRec *pRec;

    pRec = pScreenObj->GetRect();   

    if ( _lstSelectedObjects == NULL) {
        pOutline = new AmFirstSelectOutlineClass( _pMotifDisplay,
                                                pRec->x, pRec->y,
                                                pRec->width, pRec->height);
    } else {
        pOutline = new AmSelectOutlineClass( _pMotifDisplay,
                                           pRec->x, pRec->y,
                                           pRec->width, pRec->height);
    }

    pSelectObj = new AmSelectObjClass( pScreenObj, pOutline);
    _lstSelectedObjects = AmInsertNodeLst( pSelectObj, _lstSelectedObjects);

    _pMotifDisplay->AddOutline( pOutline);

    AmSendUserEvent( AM_PKG_ID, AM_SELECTION_SET_CHANGED, _lstSelectedObjects);
    return( AmOK);
}


AmErrorCode
AmEditorClass::UnselectObject( AmScreenObjClass *pScreenObj) 
{
    AmSelectObjClass *pSelectObj;
    AmListPtr        lstTemp, lstPrev;

    lstPrev = NULL;
    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        if ( pScreenObj == pSelectObj->GetScreenObj()) {

            _pMotifDisplay->RemoveOutline( pSelectObj->GetOutline());
            delete pSelectObj;
            
            if ( lstPrev != NULL) {    
                lstPrev->next = lstTemp->next;
            } else {
                _lstSelectedObjects = lstTemp->next;
            }

            free( lstTemp); 

            AmSendUserEvent( AM_PKG_ID, AM_SELECTION_SET_CHANGED,
                           _lstSelectedObjects);
            break;
        }

        lstPrev = lstTemp;
    }

    return( AmOK);
}


AmErrorCode
AmEditorClass::UnselectAll()
{
    AmSelectObjClass *pSelectObj;
    AmListPtr        lstTemp, lstNext;

    lstTemp = _lstSelectedObjects;
    while ( lstTemp != NULL) {
        lstNext = lstTemp->next;
 
        pSelectObj = ( AmSelectObjClass *) lstTemp->data;

        _pMotifDisplay->RemoveOutline( pSelectObj->GetOutline());
        delete pSelectObj;
        free( lstTemp);

        lstTemp = lstNext;
    }

    _lstSelectedObjects = NULL;
    AmSendUserEvent( AM_PKG_ID, AM_SELECTION_SET_CHANGED, _lstSelectedObjects);

    return( AmOK);
}


AnvilBoolean
AmEditorClass::IsObjectSelected( AmScreenObjClass *pScreenObj)
{
    AmSelectObjClass *pSelectObj;
    AmListPtr        lstTemp;

    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;

        if ( pSelectObj->GetScreenObj() == pScreenObj) {
            return( TRUE);
        }
    }

    return( FALSE);
}

/*********************************************************************
 *
 * Function: void AmEditorClass::GetSelectedObjectsExtents( pRec)
 *
 * Parameters:
 *     pRec:    The return parameter.
 *
 * Description:
 *     Figures out the rectangle that encompasses all the 
 *     selected objects. 
 *
 * Note:
 *     This function uses the GetRect function in the objects and so
 *     doesn't really get the entire extent of each screen object.
 *
 *********************************************************************/

void
AmEditorClass::GetSelectedObjectsExtents( AmRectangleRec *pRec)
{
    AmSelectObjClass     *pSelectObj;
    const AmRectangleRec *pRecTemp;
    AmListPtr            lstTemp;
    AnvilBoolean            fFirst;
    int                x1, y1, x2, y2;

    fFirst = TRUE;
    x1 = 0;
    y1 = 0;
    x2 = 0;
    y2 = 0;

    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        pRecTemp = pSelectObj->GetScreenObj()->GetRect();

        if ( fFirst || (( pRecTemp->x + pRecTemp->width) > x2)) {
            x2 = pRecTemp->x + pRecTemp->width;
        }

        if ( fFirst || (( pRecTemp->y + pRecTemp->height) > y2)) {
            y2 = pRecTemp->y + pRecTemp->height;
        }

        if ( fFirst || ( pRecTemp->x < x1)) {
            x1 = pRecTemp->x;
        }

        if ( fFirst || ( pRecTemp->y < y1)) {
            y1 = pRecTemp->y;
        }
 
        fFirst = FALSE;
    }

    pRec->x = x1;
    pRec->y = y1;
    pRec->width = x2 - x1;
    pRec->height = y2 - y1;
}


/*********************************************************************
 *
 * Function: void AmEditorClass::GetSelectedObjectsRealExtents( pRec)
 *
 * Parameters:
 *     pRec:    The return parameter.
 *
 * Description:
 *     Figures out the rectangle that encompasses all the 
 *     selected objects. 
 *
 * Note:
 *     This function uses the GetExtentsRect function in the objects
 *     and so really gets the entire extent of each screen object.
 *
 *********************************************************************/

void
AmEditorClass::GetSelectedObjectsRealExtents( AmRectangleRec *pRec)
{
    AmSelectObjClass     *pSelectObj;
    const AmRectangleRec *pRecTemp;
    AmListPtr            lstTemp;
    AnvilBoolean            fFirst;
    int                x1, y1, x2, y2;

    fFirst = TRUE;
    x1 = 0;
    y1 = 0;
    x2 = 0;
    y2 = 0;

    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        pRecTemp = pSelectObj->GetScreenObj()->GetExtentsRect();

        if ( fFirst || (( pRecTemp->x + pRecTemp->width) > x2)) {
            x2 = pRecTemp->x + pRecTemp->width;
        }

        if ( fFirst || (( pRecTemp->y + pRecTemp->height) > y2)) {
            y2 = pRecTemp->y + pRecTemp->height;
        }

        if ( fFirst || ( pRecTemp->x < x1)) {
            x1 = pRecTemp->x;
        }

        if ( fFirst || ( pRecTemp->y < y1)) {
            y1 = pRecTemp->y;
        }
 
        fFirst = FALSE;
    }

    pRec->x = x1;
    pRec->y = y1;
    pRec->width = x2 - x1;
    pRec->height = y2 - y1;
}

void
AmEditorClass::MoveSelectedObjects( int xMove,
                                  int yMove)
{
    AmSelectObjClass     *pSelectObj;
    AmScreenObjClass     *pScreenObj;
    AmObjectOutlineClass *pOutline;
    const AmRectangleRec *pRec;
    AmListPtr            lstTemp;
    int                x, y, iWidth, iHeight;

    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        pScreenObj = pSelectObj->GetScreenObj();
        pOutline = pSelectObj->GetOutline();

        pRec = pScreenObj->GetRect();

    /*
     * Need to copy the values from the rectangle because it is only
     *   a reference to the rectangle in the screen object.  When
     *   we change the screen object coordinates the rectangle
     *   changes, too.
     */

        x = pRec->x + xMove;
        y = pRec->y + yMove;
        iWidth = pRec->width;
        iHeight= pRec->height;

        pScreenObj->SetRectAndMargins( x, y, iWidth, iHeight,
                                       _pMotifDisplay->GetWidth(),
                                       _pMotifDisplay->GetHeight());
        pOutline->SetOrigin( x, y);
    }

    _pMotifDisplay->Flush();

}


AmScreenObjClass **
AmEditorClass::GetSelectedObjectsV()
{
    AmScreenObjClass **vpScreenObj;
    AmSelectObjClass *pSelectObj;
    AmListPtr        lstTemp;
    int            nScreenObj, ii;

    nScreenObj = AmLengthLst( _lstSelectedObjects);

    vpScreenObj = ( AmScreenObjClass **) malloc( sizeof( AmScreenObjClass *) *
                                               ( nScreenObj + 1));
    if ( vpScreenObj == NULL) {
        ANVIL_ERROR_CODE = AmMALLOC_FAILED;
        return( NULL);
    }

    for ( lstTemp = _lstSelectedObjects, ii = 0;
          lstTemp != NULL;
          lstTemp = lstTemp->next, ii++) {
        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        vpScreenObj[ ii] = pSelectObj->GetScreenObj(); 
    }

    /* NULL terminate the vector. */

    vpScreenObj[ ii] = NULL;

    return( vpScreenObj);
}

/*********************************************************************
 *
 * Function: int AmEditorClass::ActionToCursor( eAction)
 *
 * Description:
 *     Figures out the cursor that is to be displayed according to 
 *     the action given.
 *
 * Returns:
 *     int: The X cursor id.
 *
 *********************************************************************/

int
AmEditorClass::ActionToCursor( AmActionEnum eAction)
{
    int iCursor;

    switch( eAction) {
        case AmACTION_RESIZE_UP_LEFT:
            iCursor = XC_top_left_corner;
            break;
        case AmACTION_RESIZE_UP:
            iCursor = XC_top_side;
            break;
        case AmACTION_RESIZE_UP_RIGHT:
            iCursor = XC_top_right_corner;
            break;
        case AmACTION_RESIZE_LEFT:
            iCursor = XC_left_side;
            break;
        case AmACTION_RESIZE_RIGHT:
            iCursor = XC_right_side;
            break;
        case AmACTION_RESIZE_DOWN_LEFT:
            iCursor = XC_bottom_left_corner;
            break;
        case AmACTION_RESIZE_DOWN:
            iCursor = XC_bottom_side;
            break;
        case AmACTION_RESIZE_DOWN_RIGHT:
            iCursor = XC_bottom_right_corner;
            break;
        case AmACTION_MOVE:
            iCursor = XC_fleur;
            break;
        case AmACTION_DELETE:
            iCursor = XC_pirate;
            break;
        case AmACTION_RAISE:
        case AmACTION_LOWER:
            iCursor = XC_dot;
            break;
        case AmACTION_PICK:
            iCursor = XC_tcross;
            break;
        case AmACTION_NONE:
        default:
            iCursor = XC_arrow;
            break;
    }

    return( iCursor);
}


AmActionEnum
AmEditorClass::FindAction( int            x,
                         int            y,
                         AmScreenObjClass **ppScreenObj)
{
    AmListPtr            lstTemp;
    AmSelectObjClass     *pSelectObj;
    AmScreenObjClass     *pScreenObj=NULL;
    const AmRectangleRec *pRec;
    int                xDelta, yDelta, wDelta, hDelta, mwDelta, mhDelta,
                       iAnchor, iHalfAnchor;
    AmActionEnum         eAction;
   
/* Initializations. */
 
    iAnchor     = _pMotifDisplay->VirtualCoord( AmANCHOR_SIZE);
    iHalfAnchor = iAnchor / 2; 
    eAction     = AmACTION_NONE;

    if ( ppScreenObj != NULL) {
        pScreenObj   = NULL;   /* Previous object that had an action. */
        *ppScreenObj = NULL;   /* Object that has an action.          */
    }

    for ( lstTemp = _lstSelectedObjects;
          lstTemp != NULL;
          lstTemp = lstTemp->next) {

        pSelectObj = ( AmSelectObjClass *) lstTemp->data;
        pRec = pSelectObj->GetScreenObj()->GetRect();

        xDelta = x - pRec->x;
        yDelta = y - pRec->y;
        wDelta = xDelta - pRec->width;
        hDelta = yDelta - pRec->height;
        mwDelta = xDelta - pRec->width/2;
        mhDelta = yDelta - pRec->height/2;

    /* Check to see if the mouse is within the extents of this object. */ 

        if (( xDelta >= - iHalfAnchor) && ( wDelta <= iHalfAnchor) &&
            ( yDelta >= - iHalfAnchor) && ( hDelta <= iHalfAnchor)) {

            if ( ppScreenObj != NULL) {
                pScreenObj   = *ppScreenObj;
                *ppScreenObj = pSelectObj->GetScreenObj();
            }
 
            /* Check the resize areas in the left hand side. */

            if ( abs( xDelta) <= iHalfAnchor) {

                if ( abs( yDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_UP_LEFT);
                } else if ( abs( hDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_DOWN_LEFT);
                } else if ( abs( mhDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_LEFT);
                }

            /* Check the resize areas in the right hand side. */

            } else if ( abs( wDelta) <= iHalfAnchor) {

                if ( abs( yDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_UP_RIGHT);
                } else if ( abs( hDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_DOWN_RIGHT);
                } else if ( abs( mhDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_RIGHT);
                }

            /* Check the resize areas in the middle. */

            } else if ( abs( mwDelta) <= iHalfAnchor) {

                if ( abs( yDelta) <= iHalfAnchor) {
                    return( AmACTION_RESIZE_UP);
                } else if ( abs( hDelta) <= iHalfAnchor) { 
                    return( AmACTION_RESIZE_DOWN);
                }

            /* Check if it's a move. */

            }
 
            if (( xDelta >= 0) && ( wDelta <= 0) &&
                ( yDelta >= 0) && ( hDelta <= 0)) {

                eAction = AmACTION_MOVE;

            } else if ( ppScreenObj != NULL) {

                *ppScreenObj = pScreenObj;

            }
        }
    }

    return( eAction);
}


/*********************************************************************
 *
 * Function: AmActionEnum PositionToAction( x, y)
 *
 * Purpose:
 *   Figures out the action that could be done on an object depending
 *   on a a screen coordinate.
 *
 * Comments: 
 *   The possible actions that can be done on a screen object are:
 *
 *       AmACTION_NONE
 *       AmACTION_PICK
 *
 *********************************************************************/

AmActionEnum
AmEditorClass::PositionToAction( int x,
                               int y)
{
    AmScreenObjClass     *pScreenObj;   
    AmActionEnum         eAction;
    const AmRectangleRec *pRec;
    AnvilBoolean            fSetOutline;

    eAction = AmACTION_NONE;
    fSetOutline = TRUE;

    if ( _lstSelectedObjects != NULL) {

        /*
         * There is at least one active object, so check if one of the
         *     handler shortcuts exists.
         */

        eAction = FindAction( x, y, NULL);

    } else {

    /* There is no selected object, so check if the cursor is over one. */

        pScreenObj = _pMotifDisplay->FindObject( x, y);

        if ( pScreenObj != NULL) {

            eAction = AmACTION_PICK;
            pRec = pScreenObj->GetRect();
            fSetOutline = FALSE;

            if ( _pOutline == NULL) {
                _pOutline = new AmHighlightOutlineClass( _pMotifDisplay,
                                                       pRec->x, pRec->y,
                                                       pRec->width,
                                                       pRec->height);
                _pMotifDisplay->AddOutline( _pOutline); 
            } else {
                _pOutline->SetRectangle( pRec->x, pRec->y,
                                         pRec->width, pRec->height);
                if ( _pOutline->GetActive() != TRUE) {
                    _pMotifDisplay->ShowOutline( _pOutline);
                }
            }

            _pMotifDisplay->WriteResourceMessage( "%s", pScreenObj->GetName());

        }
    }

    /*
     * If the temporary outline is not to be shown, hide it here (if it
     *   exists).
     */
   
    if ( fSetOutline == TRUE) {
        if ( _pOutline != NULL && ( _pOutline->GetActive() == TRUE)) {
            _pMotifDisplay->HideOutline( _pOutline);
            _pMotifDisplay->WriteResourceMessage( " ");
        }
    }

    return( eAction);
}

/*********************************************************************
 *
 * AmMasterHandler
 *
 * Description:
 *     This handler is responsible for objects getting put in the
 *     active object set.  And for objects getting resized and
 *     moved.
 *
 *********************************************************************/

void
AmEditorClass::AmRemoveMasterCallback( XtPointer xpClientData)
{
    AmEditorClass *ec = ( AmEditorClass *) xpClientData;

    ec->removeMaster();
}

void
AmEditorClass::removeMaster()
{
    if ( _pOutline != NULL) {
        _pMotifDisplay->RemoveOutline( _pOutline);
    }

    UnselectAll();
}

void
AmEditorClass::AmSuspendMasterCallback( XtPointer xpClientData)
{
    AmEditorClass *ec = ( AmEditorClass *) xpClientData;

    ec->suspendMaster();
}

void
AmEditorClass::suspendMaster()
{
    if ( _pOutline != NULL) {
        _pMotifDisplay->HideOutline( _pOutline);
    }
}

void
AmEditorClass::AmRestartMasterCallback( XtPointer xpClientData)
{
    AmEditorClass *ec = ( AmEditorClass *) xpClientData;

    ec->restartMaster();
}

void
AmEditorClass::restartMaster()
{
    Window               window, windowRoot, windowChild;
    Display              *pDisplay;
    AmActionEnum           eAction;
    int                  xRoot, yRoot, x, y;
    unsigned int         iKeys;

    pDisplay = (( AmXDisplayClass *) _pMotifDisplay->GetDisplay())->GetXDisplay();
    window = (( AmXWindowClass *) _pMotifDisplay->GetWindow())->GetXWindow();

    XQueryPointer( pDisplay, window,
                   &windowRoot, &windowChild,
                   &xRoot, &yRoot, &x, &y, &iKeys);

    x = _pMotifDisplay->VirtualCoord( x);
    y = _pMotifDisplay->VirtualCoord( y);

    eAction = PositionToAction( x, y);
    _pMotifDisplay->SetCursor( ActionToCursor( eAction));
}


void
AmEditorClass::AmMasterHandler( Widget    w,
                            XtPointer xpClientData,
                            XEvent    *pxEvent,
                            AnvilBoolean   *pfContinueDispatch)
{
    AmEditorClass *ec = ( AmEditorClass *) xpClientData;

    ec->master( w, pxEvent, pfContinueDispatch);
}


/* ARGSUSED */
void
AmEditorClass::master( Widget  w,
                     XEvent  *pxEvent,
                     AnvilBoolean *pfContinueDispatch)
{
    XButtonEvent         *xButtonEvent;
    XMotionEvent         *xMotionEvent;
    AmScreenObjClass       *pScreenObj;
    AmActionEnum           eAction;
    int                  x=0, y=0;

    eAction = AmACTION_NONE;

/* First attempt to handle any button presses. */

    if ( pxEvent->type == ButtonPress) {

        xButtonEvent = ( XButtonEvent *) pxEvent;

        if ( xButtonEvent->button == Button3) {

        /* Right mouse button cancels selection. */

            UnselectAll();
	    x = xButtonEvent->x;
	    y = xButtonEvent->y;
            eAction = PositionToAction( x, y);

        } else if ( xButtonEvent->button == Button1) {

        /* Get the location of the mouse and convert it into virtual
           coordinates. */

            x = xButtonEvent->x;
            y = xButtonEvent->y;

            x = _pMotifDisplay->GetDisplay()->VirtualCoord( x);
            y = _pMotifDisplay->GetDisplay()->VirtualCoord( y);

            if ( _pOutline != NULL) {
                _pMotifDisplay->HideOutline( _pOutline);
            }

            eAction = FindAction( x, y, &pScreenObj);
            if ( pScreenObj != NULL) { 

                if (( eAction == AmACTION_MOVE) &&
                    ( xButtonEvent->state & ShiftMask)) {

                    UnselectObject( pScreenObj);
                    _pMotifDisplay->WriteResourceMessage( "unselectSuccessMsg");
     
                } else {

                /*
                 * The user clicked somewhere in the object -> setup
                 *   the secondary master handler that takes care of 
                 *   double clicks.
                 */
                      
                    AmSetupSecondaryHandler( this, pScreenObj, x, y, eAction);

                }
   
            } else {

                pScreenObj = _pMotifDisplay->FindObject( x, y);
                if ( pScreenObj != NULL) {
                  
                    /* A new object was selected. */

                    if ( !( xButtonEvent->state & ShiftMask)) {
                        UnselectAll();
                    }

                    SelectObject( pScreenObj);
                    _pMotifDisplay->WriteResourceMessage( "selectSuccessMsg");

                    if ( !( xButtonEvent->state & ShiftMask)) {
                        eAction = AmACTION_MOVE;
                        AmSetupSecondaryHandler( this,
                                               pScreenObj,
                                               x, y,
                                               eAction);
                    }
                } else {

                /* The user clicked somewhere where there was no object. */

                    if ( !( xButtonEvent->state & ShiftMask)) {
                        UnselectAll();
                    }

                    AmSetupOutlineHandler( this, x, y);
                }
            }

        }
    } else {

        if ( pxEvent->type == MotionNotify) {
            xMotionEvent = ( XMotionEvent *) pxEvent;
            x = xMotionEvent->x;
            y = xMotionEvent->y;
        } else if ( pxEvent->type == ButtonRelease) {
            xButtonEvent = ( XButtonEvent *) pxEvent;
            x = xButtonEvent->x;
            y = xButtonEvent->y;
        }

        x = _pMotifDisplay->GetDisplay()->VirtualCoord( x);
        y = _pMotifDisplay->GetDisplay()->VirtualCoord( y);

        eAction = PositionToAction( x, y);
    }

    _pMotifDisplay->SetCursor( ActionToCursor( eAction));
}
