#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _anvil_datagraphic_h_
#define _anvil_datagraphic_h_

#include <anvil/project.h>
#include <anvil/coordinates.h>
#include <anvil/display.h>
#include <anvil/units.h>
#include <anvil/list.h>
#include <anvil/fns.h>
#include <anvil/tagstrs.h>
#include <anvil/envdefs.h>
#include <anvil/newtile.h>
#include <anvil/cache.h>

#define MAXDIMS		8	
#define FILESZ  	256

/* user attribute definition structure */

typedef struct AmAttribute_ {
    char		*name;		/* name of user attribute */
    int			offset;		/* offset into binary record */
    AmDataTypeEnum	dataType;	/* user attribute data type */
    unitType		units;		/* units */
    float		min;		/* attribute data range */
    float		max;
    int			*table;		/* table of attribute values */
    float		*binvec;	/* non-linear scaling vector */
    int			length;		/* table or binvec length */
} AmAttributeRec, *AmAttributePtr;

/* structure describing the data */

typedef struct AmDataInfo_ AmDataInfoRec, *AmDataInfoPtr;

struct AmDataInfo_ {
    AmSourceTypeEnum	sourceType;	/* data source type */
    int         	skipBytes;      /* # bytes to skip in header */
    char        	*filename;      /* full path to read data */
    int			fd;		/* data file descriptor */
    int         	dims[ MAXDIMS]; /* list of data dimensions */
    int         	numdims;        /* number of dimensions */
    uchar_t     	*data;          /* actual data */
    AnvilBoolean		shared;		/* data is shared, don't free it */
    AmAccessFn		xFn;		/* x,y accessing functions */
    int			xOffset;	/* offset of x value */
    AmDataTypeEnum	xType;
    AmAccessFn		yFn;		/* y accessing function */
    int			yOffset;	/* offset of y value */
    AmDataTypeEnum	yType;
    void		*clientData;	/* to pass to access fns */
    int			size;		/* data element size */
    AmDataTypeEnum	dataType;	/* actual data type for plot (if set) */
    char		*descripFile;	/* syntax-based description of data */
    AnvilBoolean		enableCacheFlag;	/* flag to enable caching */

};

/* combination structure of the common information for the plots. */

class AmDataGraphicClass : public AmCoordClass {    /* inherits coord transform */
protected:
    char		*title;
    AmTileHeaderPtr	tileHeader;	/* pyramid info (may be NULL) */
    AnvilBoolean		fixedExtents;	/* extents fixed at start */
    AmDataInfoRec         dataInfo;   	/* data description */
    char		*appId;		/* applicaton specific ID */
    int			priority;	/* priority in an overlay */
    void		*selectionSet;	/* currently selected objects */
    int			numSelected;	/* # of selected objects */
    int			textOffset;	/* data structure string offset */
    int			textSize;	/* data structure string length */
    int			detailLevel;	/* detail level to show for tiles */
    AnvilBoolean		showText;
    AmColorCellRec	highlightColor;	/* selection set color */
    virtual AmErrorCode readData();	/* data read fn (default read()) */
    virtual AmErrorCode	getDataFromDescription();	/* get dataInfo */
    void GetFieldDefinitions( AmFieldDefPtr       *fieldDefs,
			      int               *numDefs);
    AmErrorCode		updateIndexData( AmIndexDataPtr	indexData,
					 const char	*tag);
    virtual AnvilBoolean	checkValues();
    virtual void	GetDataExtents( float	*xmin,
					float	*ymin,
					float	*xmax,
					float	*ymax);
public:
    const char		*getAppId() { return( appId); };
    virtual int		setFieldsWrapper( AmTagValuePtr	tagValues,
					  int		numValues);
    virtual int 	setFields( AmTagValuePtr  tagValues,
				   int          numValues);
    virtual int 	getFields( AmTagValuePtr      tagValues,
				   int              numValues);
    virtual void 	updateFields();
    virtual int 	dumpFields( AmFieldLengthPtr	fieldLengths,
				    AnvilBoolean		allFields = FALSE);
			AmDataGraphicClass(); 
    virtual 		~AmDataGraphicClass();
    void		initializeFields( AmWindowClass	*windowArg,
					  const char 	*nameArg);
    virtual AnvilBoolean	IsType( AmObjectTypeEnum type);
    int 		GetPriority();
    void		SetSelection( void 	*items, 
				      int	numItems);
    void		GetSelection( void 	**items, 
				      int	*numItems);
    void 		DeleteSelection();
    void		ClearSelection();
    AmErrorCode		ReplaceSelection( void  *replaceSelection,
					  int	numItems);
    virtual AmErrorCode	DrawSelection( AmWindowClass		*window,
				       AmGCClass			*gc,
				       AmRectangleRec const	*exposeRects,
				       int			nRects,
				       AnvilBoolean			fHighlight);
    virtual void	SetRange( float		xmin,
				  float		ymin,
				  float		xmax,
				  float		ymax);
    virtual void	SetCoords( float	currXmin,
				   float	currYmin,
				   float	currXmax,
				   float	currYmax);
    int		GetRows() { return( dataInfo.dims[ 0]); };
    int		GetCols() { return( dataInfo.dims[ 1]); };
    int		GetSize() { return( dataInfo.size); };
    AnvilBoolean	GetShared() { return( dataInfo.shared); };
    void	SetShared( AnvilBoolean sharedArg) { dataInfo.shared = sharedArg; };
    const char *GetAppId() { return( appId); };
    void	SetAppId( char *appIdArg) { appId = appIdArg; };
    uchar_t	*GetDataPointer() { return( dataInfo.data); };
    const char *GetTitle() { return( title); };
    const char *GetFilename() { return( dataInfo.filename); };
};

void AmInitIndexData( AmIndexDataPtr	indexData);
void AmDestroyIndexData( AmIndexDataPtr	indexData);

void xyExtentsFn( AmDataInfoPtr      dataInfo,
                  float            *xmin,
                  float            *ymin,
                  float            *xmax,
                  float            *ymax);

AmErrorCode AmReadColormap( AmDisplayClass  *display,
                          char          *cmapName,
                          int           *numColors,
                          AmColorCellPtr  *colors);

void AmSetBinvec( float           **binvec,
                int             length,
                AmTagValuePtr     tagValue);


int AmDataLength( AmDataInfoPtr      dataInfo);

#endif /* _anvil_datagraphic_h_ */
