#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _anvil_display_h_
#define _anvil_display_h_

#include <anvil/project.h>
#include <anvil/list.h>
#include <anvil/graphstructs.h>
#include <anvil/colorcell.h>
#include <anvil/fonts.h>
#include <anvil/fns.h>
#include <anvil/errcode.h>

/* Default font values. */

#define DEFAULT_FONT      "adobe-helvetica"
#define DEFAULT_FULL_FONT "-adobe-helvetica-medium-r-normal--10-100-75-75-p-56-iso8859-1"
#define DEFAULT_SIZE      14
#define DEFAULT_WEIGHT    AmWEIGHT_MEDIUM
#define DEFAULT_SLANT     AmSLANT_ROMAN
#define DEFAULT_POST_FONT "Helvetica"

/* Default resolution. */

#define BASEDPI           72

/* Default values. */

#define BYTES_PER_PIXEL   4
#define BITS_PER_PIXEL    32
#define DEPTH             32

/* Enumeration for line join styles. */

typedef enum lineJoin_ {
    AmLINEJOIN_ERROR = 0,
    AmJOIN_MITER     = 1,
    AmJOIN_ROUND     = 2,
    AmJOIN_BEVEL     = 3
} AmLineJoinEnum;

/* Enumeration for line cap styles. */

typedef enum lineCap_ {
    AmLINECAP_ERROR  = 0,
    AmCAP_NOT_LAST   = 1,
    AmCAP_BUTT       = 2,
    AmCAP_ROUND      = 3,
    AmCAP_PROJECTING = 4
} AmLineCapEnum;

/* Enumeration for line drawing styles. */

typedef enum lineStyle_ {
    AmLINESTYLE_ERROR  = 0,
    AmLINE_SOLID       = 1,
    AmLINE_ON_OFF_DASH = 2,
    AmLINE_DOUBLE_DASH = 3
} AmLineStyleEnum;

/* Enumeration for fill styles for polygons and rectangles */

typedef enum AmFillStyle_ {
    AmFILLSTYLE_ERROR = 	  0,
    AmFILLSTYLE_SOLID = 	  1,
    AmFILLSTYLE_TILED = 	  2,
    AmFILLSTYLE_STIPPLED = 	  3,
    AmFILLSTYLE_OPAQUE_STIPPLED = 4
} AmFillStyleEnum;

/* Enumeration for text alignment. */

typedef enum textAlignEnum_ {
     AmTEXTALIGN_ERROR      = 0,
     AmCENTER_HORIZONTAL    = 1,
     AmLEFT_HORIZONTAL      = 2,
     AmRIGHT_HORIZONTAL     = 3,
     AmCENTER_UP_VERTICAL   = 4,
     AmLEFT_UP_VERTICAL     = 5,
     AmRIGHT_UP_VERTICAL    = 6,
     AmCENTER_DOWN_VERTICAL = 7,
     AmLEFT_DOWN_VERTICAL   = 8,
     AmRIGHT_DOWN_VERTICAL  = 9
} AmTextAlignEnum;

/* Enumeration for raster operations */

typedef enum rasterOp_ {
    AmRASTEROP_ERROR = 0,
    AmCLEAR          = 1,
    AmAND            = 2,
    AmAND_REVERSE    = 3,
    AmCOPY           = 4,
    AmAND_INVERTED   = 5,
    AmNOOP           = 6,
    AmXOR            = 7,
    AmOR             = 8,
    AmNOR            = 9,
    AmEQUIV          = 10,
    AmINVERT         = 11,
    AmOR_REVERSE     = 12,
    AmCOPY_INVERTED  = 13,
    AmOR_INVERTED    = 14,
    AmNAND           = 15,
    AmSET            = 16
} AmRasterOpEnum;

/* Enumeration for colormap classes. */

typedef enum colormapType_ {
    AmCOLORMAPTYPE_ERROR = 0,
    AmPSEUDO_COLOR       = 1,    /* Colormapped. */
    AmTRUE_COLOR         = 2     /* Direct (24-bit) color. */
} AmColormapTypeEnum;

/* Enumeration for display classes. */

typedef enum displayType_ {
     AmDISPLAYTYPE_ERROR  = 0,
     AmX_DISPLAY          = 1,
     AmPOSTSCRIPT_DISPLAY = 2,
     AmNT_DISPLAY         = 3
} AmDisplayTypeEnum;

/* enumeration for polygon fill rules */

typedef enum fillRule_ {
    AmFILL_ERROR 	= 0,
    AmFILL_EVEN_ODD 	= 1,
    AmFILL_WINDING 	= 2
} AmFillRuleEnum;

/* Forward references for the classes. */

class AmDisplayClass;
class AmGCClass;
class AmImageClass;
class AmGlyphClass;
class AmRegionClass;

/* function pointer for reading scanlines from a source file. Used for 
   reading files by scanline into images */ 

typedef AnvilBoolean ( *AmScanlineFn)( int      fd,
                                uchar_t  *buffer,
                                int      size,
                                int      length,
                                int      cols,
                                int      x,
                                int      y,
                                void     *clientData);

/*
 * Class: AmWindowClass
 *
 * Description:
 *   The window class is mostly virtual functions implemented by each of
 *   the derived window (output) classes.  The class is used as the
 *   output device for drawing functions (i.e. the canvas to draw
 *   upon).
 */

class AmWindowClass {

protected:
    AmDisplayClass *_pDisplay;
    AmListPtr      _lstGC;     /* GCs created for this window. */

public:

    /* Constructor/Destructor. */

    AmWindowClass();
    AmWindowClass( AmDisplayClass *pDisplay);
    virtual ~AmWindowClass();

    /* Validity test function */

    virtual AnvilBoolean	IsValid() = 0;

    /* Accessor functions. */

    virtual void         GetDims( int *width, int *height) = 0;
    virtual void         GetWindowGeometry( int *x,
                                            int *y,
                                            int *width,
                                            int *height) = 0;
    virtual AmDisplayClass *GetDisplay();

    /* Mutator functions. */

    virtual void         ClearRect( AmRectangleRec const *rect) = 0;
    virtual void         ResizeWindow( int cx,
                                       int cy) = 0;
    virtual void         MoveWindow( int x,
	                             int y) = 0;

    /* GC related functions. */

    void                 AddGC( AmGCClass *pGC);
    void                 RemoveGC( AmGCClass *pGC);

/* device-independent GC creation function */

    virtual AmGCClass	 *CreateGC() = 0;

    /* Display related functions. */

    void                 RemoveDisplay( AmDisplayClass *pDisplay);
    virtual void         Update();
};

/*
 * Class: AmDisplayClass
 *
 * Description:
 *   The display class is mostly virtual functions implemented by each of
 *   the derived display classes.  The display class contains the drawing
 *   functions used to draw onto windows.
 */

class AmDisplayClass {

protected:
    AmDisplayTypeEnum _eDisplayType;     /* Display type (i.e. AmX_DISPLAY).  */ 
    int             _iDPIBase;         /* Dots per inch of the "reference"  */
                                       /*   resolution.                     */
    int             _iDPI;             /* Dots per inch of the display.     */

    AmListPtr         _lstWindow;        /* Windows created for this display. */
    
public:

    /* Constructors/Destructors. */

    AmDisplayClass();
    virtual ~AmDisplayClass() = 0;

    /* Accessor functions. */

    AmDisplayTypeEnum GetType();
    int             GetDPI();
    int             GetDPIBase();

    virtual int     GetDepth() = 0;
    virtual void    GetImage( AmWindowClass *pWindow,
                              AmImageClass  *pImage,
                              int         x,
                              int         y) = 0;

    /* Conversion functions. */

    int             ScreenCoord( int iVirtual);
    int             VirtualCoord( int iScreen); 
    float	    ScreenFloat( float iVirtual);
    float	    VirtualFloat( float iScreen);

    void	    ScreenRect( AmRectangleRec const *pVirtualRect,
                                AmRectanglePtr       pScreenRect);
    void	    VirtualRect( AmRectangleRec const *pScreenRect,
                                 AmRectanglePtr 	    pVirtualRect);
    void	    ScreenRealRect( AmRealRectRec const *pVirtualRealRect,
                                    AmRealRectPtr       pScreenRealRect);
    void	    VirtualRealRect( AmRealRectRec const *pScreenRealRect,
                                     AmRealRectPtr       pVirtualRealRect);

    virtual int     DimInPixels( int iDimension);

    /* Mutator functions. */

    void            SetDPI( int iDPI);

    void            AddWindow( AmWindowClass *pWindow);
    void            RemoveWindow( AmWindowClass *pWindow);

    /* Drawing functions. */

    virtual void    DrawLines( AmWindowClass *pWindow, 
                               AmGCClass     *pGC, 
                               AmPointPtr    pPoint, 
                               int         nPoint) = 0;
    void            DrawLine( AmWindowClass *pWindow,
                              AmGCClass     *pGC,
                              int         x0,
                              int         y0,
                              int         x1, 
                              int         y1);
    virtual void    DrawPoints( AmWindowClass 	*pWindow, 
                                AmGCClass     	*pGC, 
                                AmPointPtr    	pPoint, 
                                int     	nPoint) = 0;
    virtual void    DrawSegments( AmWindowClass *pWindow, 
                                  AmGCClass     *pGC, 
                                  AmSegmentRec  *pSegment,
                                  int         nSegment) = 0;
    virtual void    FillArc( AmWindowClass *pWindow, 
                             AmGCClass     *pGC, 
                             int         x, 
                             int         y, 
                             int         w, 
                             int         h, 
                             float       dAngelStart,
                             float       dAngleEnd) = 0;
    virtual void    DrawArc( AmWindowClass *pWindow, 
                             AmGCClass     *pGC, 
                             int         x, 
                             int         y, 
                             int         w, 
                             int         h, 
                             float       dAngleStart,
                             float       dAngleEnd) = 0;
    virtual void    DrawImage( AmWindowClass    *w,
                               AmGCClass    *gc,
                               int        srcX, 
                               int        srcY, 
                               int        dstX, 
                               int        dstY, 
                               int        srcWidth, 
                               int        srcHeight, 
                               AmImageClass    *image) = 0;
    virtual void    DrawGlyph( AmWindowClass    *w,
                               AmGlyphClass    *glyph,
                               int        x,
                               int        y) = 0;
    virtual void    CopyArea( AmWindowClass *src,
                              AmWindowClass *dst,
                              AmGCClass     *gc,
                              int         srcX,
                              int         srcY,
                              int         dstX,
                              int         dstY,
                              int         srcWidth,
                              int         srcHeight) = 0;
    virtual void    DrawText( AmWindowClass *w,
                              AmGCClass     *gc,
                              int         x, 
                              int         y,
                              char        *text) = 0;
    virtual void    DrawImageText( AmWindowClass *w,
                                   AmGCClass     *gc,
                                   int         x, 
                                   int         y,
                                   char        *text) = 0;
    virtual void    DrawRotatedText( AmWindowClass *w,
                                     AmGCClass     *gc,
                                     int         x, 
                                     int         y,
                                     float       radianAngle,
                                     char        *text) = 0;
    void            DrawAlignedText( AmWindowClass      *w,
                                     AmGCClass          *gc,
                                     int              x, 
                                     int              y,
                                     char             *text,
                                     AmTextAlignEnum    align);
    void            DrawMultilineText( AmWindowClass      *w,
                                       AmGCClass          *gc,
                                       int              x, 
                                       int              y,
                                       char             *text,
                                       AmTextAlignEnum    align);
    virtual void    ExtentsText( AmGCClass        *gc,
                                 char        *text,
                                 int        *width,
                                 int        *ascentReturn,
                                 int        *descentReturn) = 0;
    virtual int     TextWidth( AmGCClass              *gc,
                               char                 *text,
                               int                  length = 0) = 0;
    virtual int     FontHeight( AmGCClass             *gc,
                                int                 *ascentReturn,
                                int                 *descentReturn) = 0;
    virtual void    FillPolygon( AmWindowClass    *w,
                                 AmGCClass        *gc,
                                 AmPointPtr       pts,
                                 int        numpts) = 0;
    virtual void    DrawRectangle( AmWindowClass    *win,
                                   AmGCClass        *gc,
                                   int        x, 
                                   int        y, 
                                   int        w, 
                                   int        h) = 0;
    virtual void    FillRectangle( AmWindowClass    *win,
                                   AmGCClass        *gc,
                                   int        x, 
                                   int        y, 
                                   int        w, 
                                   int        h) = 0;
    void            DrawRotatedEllipse( AmWindowClass *win,
                                        AmGCClass     *gc,
                                        int     x0,
                                        int     y0,
                                        int     x1,
                                        int     y1,
                                        int     axisB,
                                        int     npts);

    /* Color functions. */

    virtual AnvilBoolean AllocColor( AmColorCellPtr pCol) = 0;
    virtual AnvilBoolean AllocNamedColor( const char   *szName,
                                     AmColorCellPtr pCol) = 0;
    void            AllocateColors( AmColorCellPtr pCol,
                                    int          nCol);
    AnvilBoolean         readAndAllocColors( char         *szFile, 
                                        int          *pnCol, 
                                        int          *pnFileCol,
                                        AmColorCellPtr *pCol);

    /* Creation functions. */

    virtual AmImageClass  *CreateImage( int     cx,
                                      int     cy,
                                      int     iDepth,
                                      char    *pData) = 0;
    virtual AmGlyphClass  *CreateGlyph( AmWindowClass *pWindow,
                                      char    *szFile,
                                      ulong_t lPixel) = 0;
    virtual AmWindowClass *CreateBuffer( AmWindowClass *pWindow,
                                       int         cx,
                                       int         cy) = 0;
    virtual AmWindowClass *CreateWin( AmWindowClass *pWindow,
	                            int         x,
				    int         y,
				    int         cx,
				    int         cy) = 0;
    virtual AmRegionClass *CreateRegion() = 0;
};

/*
 * Class:: AmGCClass
 *
 * Description:
 *   The gc class is mostly of virtual functions implemented by each of
 *   the derived gc classes.  The gc class is used to describe how drawing
 *   is to be done (i.e. thick lines, thin lines).
 */

class AmGCClass {

protected:
    AmWindowClass  	*_pWindow;
    AmDisplayClass 	*_pDisplay;
    char	   	*_fillPatFile;
    AmFillStyleEnum	_fillStyle;

public:

    /* Constructors/Destructor. */

    AmGCClass();
    AmGCClass( AmWindowClass *pWindow);
    virtual ~AmGCClass() = 0;

    /* Accessor functions. */

    virtual void          GetForegroundColor( AmColorCellPtr pCol) = 0;
    virtual void          GetBackgroundColor( AmColorCellPtr pCol) = 0;
    virtual float         GetLineWidth() = 0;
    virtual AmLineJoinEnum  GetLineJoin() = 0;
    virtual AmLineStyleEnum GetLineStyle() = 0;
    virtual AmLineCapEnum   GetLineCap() = 0;
    virtual AmRasterOpEnum  GetRasterOp() = 0;
    virtual AnvilBoolean 	    SetFillPattern( const char *fillPatFile) = 0;
    virtual char	    *GetFillPattern() = 0;
    virtual void	    SetFillStyle( AmFillStyleEnum fillStyle) = 0;
    virtual AmFillStyleEnum GetFillStyle() = 0;
    virtual void          GetDashes( char **ppcDashPattern, 
                                     int  *pnDashPattern,
                                     int  *piDashOffset) = 0;

    virtual void GetFont( char           **pszFont,
                          int            *piSize,
                          AmFontWeightEnum *peWeight = NULL,
                          AmFontSlantEnum  *peSlant = NULL,
                          float          *pdxScale = NULL,
                          float          *pdyScale = NULL) = 0;
    virtual void GetFont( AmFontRec *pFont) = 0;

    virtual void GetClipRectangles( AmRectanglePtr *clipRectVec,
                                    int         *nRects) = 0;
    AmWindowClass *GetWindow() { return( _pWindow); };

    /* Mutator functions. */

    virtual void SetForegroundColor( AmColorCellPtr pCol) = 0;
    virtual void SetBackgroundColor( AmColorCellPtr pCol) = 0;
    virtual void SetLineWidth( float dLineWidth) = 0;
    virtual void SetLineJoin( AmLineJoinEnum    lineJoin) = 0;
    virtual void SetLineStyle( AmLineStyleEnum    lineStyle) = 0;
    virtual void SetLineCap( AmLineCapEnum    lineCap) = 0;
    virtual void SetRasterOp( AmRasterOpEnum    rasterOp) = 0;
    virtual void SetDashes( char     *dashPat, 
                            int        ndash,
                            int        dashOffset) = 0;

    virtual AnvilBoolean SetFont( char           *name,
                             int            size,
                             AmFontWeightEnum weight = AmWEIGHT_MEDIUM,
                             AmFontSlantEnum  slant = AmSLANT_ROMAN,
                             float          xScale = 1.0,
                             float          yScale = 1.0) = 0;
    virtual AnvilBoolean SetFont( AmFontRec *pFont) = 0;

    virtual void SetClipRectangles( AmRectangleRec const *clipRectVec,
                                    int               nrects) = 0;
    virtual void SetClipRegion( AmRegionClass	*region) = 0;
    virtual AmRegionClass *GetClipRegion() = 0;
    void         SetClipRectangle( int    x,
                                   int    y,
                                   int    w,
                                   int    h);

    virtual void Update() = 0;
    void         RemoveWindow( AmWindowClass *pWindow);
};

/*
 * Class: AmImageClass
 *
 * Description:
 *   The image class is mostly virtual functions implemented by each of
 *   the derived image classes.
 */

class AmImageClass {
protected:
    AmDisplayClass        *_pDisplay;
    int		        _iDPI;             /* DPI image was created with.     */

    int		        _cxVirtual;        /* Keep virtual width and height   */
    int		        _cyVirtual;        /*   to prevent roundoff errors.   */

    int		        _cxScreen;         /* Keep screen width and height    */
    int		        _cyScreen;         /*   for future reference.         */
 
    int                 _cxPixels;         /* Number of pixels along x axis.  */
    int                 _cyPixels;         /* Number of pixels along y axis.  */

    AnvilBoolean             _fDirty;           /* Indicates whether the image     */
                                           /*   data has been changed or not. */

    int                 _iDepth;           /* Stats for the internal image.   */
    int                 _iBytesPerPixel;   /*                                 */
    int                 _iBitsPerPixel;    /*                                 */
    int                 _iBytesPerLine;    /*                                 */
    int                 _iSize;            /*                                 */

    int                 _iDDDepth;         /* Stats for the device dependent  */
    int                 _iDDBytesPerPixel; /*   image (such as X, Postscript, */
    int                 _iDDBitsPerPixel;  /*   or NT).                       */
    int                 _iDDBytesPerLine;  /*                                 */
    int                 _iDDSize;          /*                                 */

public: 
    AmImageClass();
    AmImageClass( AmDisplayClass *pDisplayClass,
		int          cxVirtual,
		int          cyVirtual);
    virtual 		~AmImageClass() = 0;

/* Accessor functions. */

    int 		GetWidth();
    int 		GetHeight();
    int 		GetScreenWidth();
    int 		GetScreenHeight();
    int 		GetNumHorizPixels();
    int 		GetNumVerticalPixels();

    int			GetDPI();
    AnvilBoolean             GetDirty();

    virtual int 	GetDepth();
    virtual int 	GetBitsPerPixel();
    virtual int 	GetBytesPerLine();

    virtual char 	*AmGetData() = 0;

    AmDisplayClass	*GetDisplay();

/* Mutator functions .*/

    void                SetDirty( AnvilBoolean fDirty);
    virtual AnvilBoolean 	SetData( char *data) = 0;
    void		ClearImageRect( AmColorCellPtr		colorCell,
					AmRectangleRec const	*clearRect);
    AmErrorCode 	resizeMatrixImage( uchar_t              *srcData,
				   	   int                  srcSize,
				   	   int                  rows,
				   	   int                  cols,
				   	   AnvilBoolean              flipRows,
				   	   AnvilBoolean              flipCols,
				   	   AmRealRectRec const    *srcRect,
				   	   AmRectangleRec const   *dstRect,
				   	   AmColorFn              pixelFn,
				   	   void                 *clientData,
				   	   float                transparency);
    AmErrorCode 	bilinMatrixImage( uchar_t              *srcData,
				   	  int                  srcSize,
				   	  int                  rows,
				   	  int                  cols,
				   	  AnvilBoolean              flipRows,
				   	  AnvilBoolean              flipCols,
				   	  AmRealRectRec const    *srcRect,
				   	  AmRectangleRec const   *dstRect,
				   	  AmColorFn              pixelFn,
				   	  void                 *clientData,
				   	  float                transparency);
    AmErrorCode 	imageReadRows(
				int                  fd,
				int                  srcSize,
				AmColorFn              pixelFn,
				void                 *clientData,
				int                  rows,
				int                  cols,
				AmRealRectRec const    *srcRect,
				AmRectangleRec const   *dstRect,
				float                transparency,
				AmScanlineFn           readScanFn,
				void                 *readScanClientData);
};

/* base class for glyphs, which are see-through images */

class AmGlyphClass {
protected:
    int        width;
    int        height;
    int        depth;
    AnvilBoolean    valid;
    ulong_t    color;
public:
    AnvilBoolean     IsValid();
    AmGlyphClass();
    virtual      ~AmGlyphClass() = 0;
    int         GetWidth();
    int         GetHeight();
    ulong_t     GetColor();
    virtual void     DrawGlyph( AmDisplayClass *pDisplay,
                                AmWindowClass  *pWindow,
                                int          x,
                                int          y) = 0;
    virtual AmGlyphClass    *ResizeGlyph( int cx,
                                        int cy) = 0;
};

class AmRegionClass {
protected:
    AmDisplayClass	*_pDisplay;
public:
    			AmRegionClass( AmDisplayClass *display);
    			AmRegionClass();
    virtual 		~AmRegionClass();
    virtual AnvilBoolean 	PointInRegion( int	x,
				       int	y) = 0;
    virtual AnvilBoolean 	EqualRegion( AmRegionClass	*otherRegion) = 0;
    virtual void	AddPolygon( AmPointPtr		pt,
				    int			n,
				    AmFillRuleEnum	fillRule) = 0;
    virtual void	AddRectangle( AmRectangleRec const *rect) = 0;
    virtual void	UnionRegion( AmRegionClass	*otherRegion) = 0;
    virtual void	IntersectRegion( AmRegionClass	*otherRegion) = 0;
    virtual void	XorRegion( AmRegionClass	*otherRegion) = 0;
    virtual void 	SubtractRegion( AmRegionClass	*otherRegion) = 0;
    virtual AmRegionClass 	*CopyRegion() = 0;
}; 

#endif /* _anvil_display_h_ */
