#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _anvil_dispmgr_h_
#define _anvil_dispmgr_h_

#include <anvil/list.h>
#include <anvil/screenobj.h>
#include <anvil/display.h>

/* used within the Flush() routine for tracking the list of objects and 
   clip rectangles which are generated by an update.  These are consolidated
   into draw requests for each affected object */ 

typedef struct AmDrawClip_ {
    AmScreenObjClass	*obj;
    AmRectangleRec	clipRect;
    int			order;
} AmDrawClipRec, *AmDrawClipPtr;
 
/* elements of objList.  Contains object and AmLast rectangle.  When the
   events are flushed, the old rectangle is needed if the object is resized
   to calculate whatever objects need to be cleared */
 
typedef struct AmObjRect_ {
    AmRectangleRec        oldRect;
    AmScreenObjClass      *obj;
} AmObjRectRec, *AmObjRectPtr;

/* elements of the delete list.  Contains the rectangle and order of the
   deleted object */

typedef struct AmDeleteRect_ {
    AmRectangleRec	rect;
    int			order;
} AmDeleteRectRec, *AmDeleteRectPtr;

/* when events are sent from objects to the motif display,
   AmUpdateObjectCallback stuffs one of these in the eventList, which is
   then processed and interpreted by Flush().  The order is the object's
   stacking order, the eventId is the event that was sent. topLevel is
   a flag for whether it's a top level object, or something in an overlay,
   and obj is the object itself. */
 
typedef struct AmEventRect_ {
    int             order;
    int             eventId;
    AnvilBoolean         topLevel;
    AmScreenObjClass  *obj;
} AmEventRectRec, *AmEventRectPtr;

/* base system-independent display manager class */

class AmDisplayManagerClass {
protected:
    AmDisplayClass  *_pDisplay;       /* The display to draw with.           */
    AmWindowClass   *_pWindow;        /* Window to draw to.                  */
    AmWindowClass   *_pWindowBuffer;  /* Double buffer AmXWindowClass wrapper. */
    AmGCClass       *_pGC;            /* The GC to use in drawing.           */
    AmGCClass	  *_pCursorGC;	    /* The XOR GC to use for cursors 	   */

    AmListPtr       _lstObj;          /* Objects and old rects (objRect).    */
    AnvilBoolean       _fValid;          /* Successful construction flag.       */

    int           _xOrigin;         /* The x and y coordinate of the       */
    int           _yOrigin;         /*   origin (virtual).                 */
    int           _iWidth;          /* Save old size when resizing.        */
    int           _iHeight;

    AmListPtr       _lstEvent;        /* Outstanding events (eventRect).     */
    AmListPtr       _lstDelete;       /* Outstanding deletes (eventRect).    */
    AnvilBoolean	  _fJustResized;    /* Display just resized, don't redraw. */

    AnvilBoolean	  _fNeedResize;     /* Outstanding resize request.         */
    AnvilBoolean	  _fNeedFlush;	    /* flush couldn't be processed 	   */

/* Functions used by Flush(). */

    void          executeDrawClips( AmWindowClass *win,
                                    AmGCClass     *gc,
                                    AmDrawClipRec *drawClips,
                                    int         numDrawClips);
    void          drawClippedObject( AmScreenObjClass *obj,
                                     AmListPtr        objList,
                                     AmWindowClass    *win,
                                     AmDrawClipPtr    drawClips,
                                     int            *numDrawClips);
    void          doDeleteList( AmListPtr     deleteList,
                                AmListPtr     objList,
                                AmWindowClass *window,
                                AmDrawClipPtr drawClips,
                                int         *numDrawClips);
    void          doResize( AmEventRectPtr eventRect,
                            AmListPtr      objList,
                            AmWindowClass  *window,
                            AmDrawClipPtr  drawClips,
                            int          *numDrawClips);
    static void    AmDeleteObjectCallback( int		     iEventID,
					 AmDisplayManagerClass *pDisplayManager,
					 AmScreenObjClass	     *pScreenObj);
    void	   deleteObject( AmScreenObjClass	*pScreenObj);
    static void	   AmUpdateObjectCallback( int		     iEventID,
    					 AmDisplayManagerClass *pDisplayManager,
					 AmScreenObjClass	     *pScreenObj);
    AmRectangleRec const	*GetExposeRects( AmRectangleRec const *vExposeRect,
    					 int		    *nExposeRect);
public:

    AmDisplayManagerClass();
    AmDisplayManagerClass( AmDisplayClass	*display,
			 AmWindowClass	*window,
                         AnvilBoolean 	fDoubleBuffer = FALSE);
    virtual 	    ~AmDisplayManagerClass();

    void            Update();
    void            Draw( AmRectangleRec const *pExposeRects = NULL,
                          int                nExposeRects = 0);
    void            ReDraw( AmRectangleRec const *pExposeRects = NULL,
                            int                nExposeRects = 0);
    virtual void    ClearArea( AmRectangleRec const *pClearRects = NULL,
                               int                nClearRects = 0);

    virtual void    Resize( int newWidth,
			    int newHeight);

    int             ScreenCoord( int iVirtual);
    int             VirtualCoord( int iScreen);

    void            ChangeDPI( int iDpi,
                               int iWidth,
                               int iHeight);

/* functions to return class members */

    AnvilBoolean         IsValid() { return( _fValid); };
    AmWindowClass     *GetWindow() { return( _pWindow); };
    AmWindowClass     *GetBuffer() { return( _pWindowBuffer); };
    AmGCClass         *GetGC() { return( _pGC); };
    AmGCClass	    *GetCursorGC() { return( _pCursorGC); };
    AmDisplayClass    *GetDisplay() { return( _pDisplay); };

    virtual int     GetWidth() { return( _iWidth); };
    virtual int     GetHeight() { return( _iHeight); };

/* get, add, remove, and control stacking of objects */

    AmScreenObjClass **GetObjects();
    AmErrorCode    AddObject( AmScreenObjClass *obj);
    void           RemoveObject( AmScreenObjClass *obj);
    void           RemoveObjectFromList( AmScreenObjClass *obj);
    void           RemoveObjectSet( AmScreenObjClass **objVec,
                                    int            numObjs);

    AnvilBoolean        IsTopObject( AmScreenObjClass *obj);

    void           PopToTop( AmScreenObjClass *obj);
    void           Pop( AmScreenObjClass *obj);
    void           PushToBottom( AmScreenObjClass *obj);
    void           Push( AmScreenObjClass *obj);

    void           updateObject( AmScreenObjClass *obj,
                                 int            eventId);

    AmScreenObjClass *FindObject( int x,
                                int y);

    virtual void    Flush();    /* flush event list */
    void            Wipe();        /* wipe event list */

    AmErrorCode     AddEvent( AmScreenObjClass *obj,
                              int            eventId,
                              int            order,
                              AnvilBoolean        topLevel);
    AmErrorCode     AddDelete( AmScreenObjClass    *obj,
                               int            order);
    void            inObjectList( AmScreenObjClass *obj,
                                  AnvilBoolean        *topLevel,
                                  AnvilBoolean        *exists,
                                  int            *order);

    virtual void    AmSendUserEvent( int  iPackageID,
                                   int  iEventID, 
                                   void *pClientData) = 0;
    void 	    SetNeedResize( AnvilBoolean flag) { _fNeedResize = flag; };
    AnvilBoolean	    GetNeedResize() { return( _fNeedResize); };
    void 	    SetNeedFlush( AnvilBoolean flag) { _fNeedFlush = flag; };
    AnvilBoolean	    GetNeedFlush() { return( _fNeedFlush); };
};

#endif /* _anvil_dispmgr_h_ */
