#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* General projections header file */
#ifndef anvil_projects_h
#define anvil_projects_h

    /* standard inclusions */
#include <math.h>

/* unfortunately, windows #defines the bessel functions in math.h */

#if defined(_WIN32)
#undef y0
#undef y1
#undef yn
#endif
#include <stdlib.h>
#include <anvil/project.h>
#include <anvil/errcode.h>
#include <anvil/units.h>
	/* prototype hypot for systems where absent */
extern double hypot(double, double);
	/* some useful constants */

typedef struct { double u, v; }	UV;
typedef struct { double r, i; }	COMPLEX;

#ifndef PJ_LIB__
#define XY UV
#define LP UV
#else
typedef struct { double x, y; }     XY;
typedef struct { double lam, phi; } LP;
#endif

#if defined( __cplusplus) || defined( c_plusplus) 
extern "C" {
#endif

extern int		/* global error return code */
pj_errno;

/* because of naming conflicts with Windows, we have to change the name
   PVALUE to PROJVALUE, because PVALUE is defined in WINREG.H */

typedef union { double  f; int  i; char *s; } PROJVALUE;

struct DERIVS {
	double x_l, x_p; /* derivatives of x for lambda-phi */
	double y_l, y_p; /* derivatives of y for lambda-phi */
};

struct FACTORS {
        struct DERIVS der;
        double h, k;    /* meridinal, parallel scales */
        double omega, thetap;   /* angular distortion, theta prime */
        double conv;    /* convergence */
        double s;               /* areal scale factor */
        double a, b;    /* max-min scale error */
        int code;               /* info as to analytics, see following */
};

typedef struct ARG_list {
        struct ARG_list *next;
        char used;
        char param[1]; } paralist;
        /* base projection data structure */
 


/* base projection data structure */
typedef struct PJconsts {
        XY  (*fwd)(LP, struct PJconsts *);
        LP  (*inv)(XY, struct PJconsts *);
        void (*spc)(LP, struct PJconsts *, struct FACTORS *);
        void (*pfree)(struct PJconsts *);
        paralist *params;   /* parameter list */
        int over;   /* over-range flag */
        int geoc;   /* geocentric latitude flag */
        double
                a,  /* major axis or radius if es==0 */
                e,  /* eccentricity */
                es, /* e ^ 2 */
                ra, /* 1/A */
                one_es, /* 1 - e^2 */
                rone_es, /* 1/one_es */
                lam0, phi0, /* central longitude, latitude */
                x0, y0, /* easting and northing */
                k0,     /* general scaling factor */
                to_meter, fr_meter; /* cartesian scaling */
#ifdef PROJ_PARMS__
PROJ_PARMS__
#endif /* end of optional extensions */
} PJ;



typedef struct { float lam, phi; } FLP;
typedef struct { int lam, phi; } ILP;

#define MAX_TAB_ID 80

struct CTABLE {
	char id[MAX_TAB_ID]; /* ascii info */
	LP ll;      /* lower left corner coordinates */
	LP del;     /* size of cells */
	ILP lim;    /* limits of conversion matrix */
	FLP *cvs;   /* conversion matrix */
};

/* Approximation structures and procedures */
typedef struct {	/* Chebyshev or Power series structure */
	UV a, b;		/* power series range for evaluation */
					/* or Chebyshev argument shift/scaling */
	struct PW_COEF {/* row coefficient structure */
		int m;		/* number of c coefficients (=0 for none) */
		double *c;	/* power coefficients */
	} *cu, *cv;
	int mu, mv;		/* maximum cu and cv index (+1 for count) */
	int power;		/* != 0 if power series, else Chebyshev */
} Tseries;

/* procedure prototypes */

double  dmstor(const char *, char **);
void set_rtodms(int, int);
char *rtodms(char *, double, int, int);
double adjlon(double);
double aacos(double), aasin(double), asqrt(double);
PROJVALUE pj_param(paralist *, char *);
paralist *pj_mkparam( const char *);
int pj_ell_set(paralist *, double *, double *);
double *pj_enfn(double);
double pj_mlfn(double, double, double, double *);
double pj_inv_mlfn(double, double, double *);
double pj_qsfn(double, double, double);
double pj_tsfn(double, double, double);
double pj_msfn(double, double, double);
double pj_phi2(double, double);
double pj_qsfn_(double, PJ *);
double *pj_authset(double);
double pj_authlat(double, double *);
COMPLEX pj_zpoly1(COMPLEX, COMPLEX *, int);
COMPLEX pj_zpolyd1(COMPLEX, COMPLEX *, int, COMPLEX *);
int pj_deriv(LP, double, PJ *, struct DERIVS *);
int pj_factors(LP, PJ *, double, struct FACTORS *);
XY pj_fwd(LP, PJ *);
LP pj_inv(XY, PJ *);
void pj_pr_list(PJ *);
void pj_free(PJ *);
PJ *pj_init(int, char const **);
void *pj_malloc(size_t);
void pj_dalloc(void *);
char *pj_strerrno(int);

Tseries *mk_cheby(UV, UV, double, UV *, UV (*)(UV), int, int, int);
UV bpseval(UV, Tseries *);
UV bcheval(UV, Tseries *);
UV biveval(UV, Tseries *);
void *vector1(int, int);
void **vector2(int, int, int);
int bchgen(UV, UV, int, int, UV **, UV(*)(UV));
int bch2bps(UV a, UV b, UV **, int, int);

/* nadcon related protos */
LP nad_interp(LP, struct CTABLE *);
LP nad_cvt(LP, int, struct CTABLE *);
struct CTABLE *nad_init(char *);
void nad_free(struct CTABLE *);

void projectPt( PJ      *pj,
                AnvilBoolean invFlag,
                float   *x,
                float   *y);

/* function for reading projection units lookup file */

AmErrorCode ReadProjUnits( char *filename);

#if defined( __cplusplus) || defined( c_plusplus) 
};
#endif

AnvilBoolean UTMdist( float          x0, 
                 float          y0,
                 float          x1,     
                 float          y1,
                 unitType       units,
                 float          *utmdist);


#endif /* end of basic projections header */
