/* Simple program that does an adaptive loop for the map app -- we use 
   three levels of filtering: no filtering, remove all but primary/secondary
   roads, and remove all but primary roads */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <include/odytypes.h>
#include <include/odyssey.h>
#include "kernel/vfs/includes/odyapi.h"
#include "odyssey/wardens/map/libodymap_tsops.h"
#include "odyssey/wardens/map/mapwarden.h"

/* The filters corresponding to the fidelity levels */

#define MAX_FILTERIDS 5
#define MAX_BW (10000000 * 8)
#define MIN_BW 0

static struct {
    int numIds;
    int filterids[MAX_FILTERIDS];
    int min_bw;
}
filters[] = {
    {2, {5003, 5004}, 0}, /* low fidelity */
    {1, {5004}, 50000}, /* intermediate fidelity */
    {0, {}, 250000}, /* high fidelity */
    {0, {}, MAX_BW} /* dummy entry */
};


static int cur_bw = MIN_BW;
static int bw_reqid = -1;

static void adapt_new_bw(int reqid, unsigned int rsrc, int val)
{
    ASSERT(reqid == bw_reqid);
    ASSERT(rsrc = ODY_GR_NETBW);
    cur_bw = val;
    fprintf(stderr,"adapt_new_bw: New bandwidth is %d\n", cur_bw);
    bw_reqid = -1;
}

int main (void)
{
    int retsz = 0;
    int rc;
    struct ody_req_des or_des;
    int fd = open("/ody/map", O_RDONLY);

    if (fd < 0) {
	perror("/ody/map");
	exit(1);
    }

    or_des.ord_resource = ODY_GR_NETBW;
    or_des.ord_version.ov_gs = ODY_GENERIC_INTERFACE;
    or_des.ord_version.ov_type = MAP_TTYPE;
    or_des.ord_version.ov_tsvnum = MAPWARDEN_VERSION;
  
    /* The adaptive loop */
    while (1) {
	/* Pick a filter according to the current bandwidth */
	int cur_fidelity, bw = cur_bw;
	int i;

	fprintf(stderr,"Picking fidelity for bandwidth %d\n", bw);

	for (cur_fidelity = 1;
	     cur_fidelity < sizeof(filters)/sizeof(filters[0]);
	     cur_fidelity++) {
	    if (bw < filters[cur_fidelity].min_bw) break;
	}

	ASSERT(cur_fidelity < sizeof(filters)/sizeof(filters[0]));

	/* We found the lowest fidelity that we could NOT support, so the next
	   lower one is the highest that we CAN support */
	cur_fidelity--; 

	fprintf(stderr,"Setting fidelity to %d, filters", cur_fidelity);
	for (i = 0; i < filters[cur_fidelity].numIds; i++)
	    fprintf(stderr," %d", filters[cur_fidelity].filterids[i]);
	fprintf(stderr,"\n");
      
	/* Set the filter */
	rc = ody_ftsop(fd, &or_des.ord_version, MAPW_SET_FILTER,
		       filters[cur_fidelity].numIds * sizeof(int),
		       filters[cur_fidelity].filterids, &retsz, NULL);

	if (rc) {
	    if (errno == EINTR) /* we probably got an upcall */
		continue;
	    perror("ody_ftsop");
	    exit(1);
	}
    
	/* Set up the adaptation request */

	or_des.ord_low = filters[cur_fidelity].min_bw;
	or_des.ord_high = filters[cur_fidelity+1].min_bw;
	or_des.ord_fp = adapt_new_bw;
    
	fprintf(stderr,"Request [%d, %d]\n", or_des.ord_low, or_des.ord_high);

	rc = ody_frequest(fd, &or_des, &bw_reqid);
	if (rc) {
	    if (errno == ODY_ENOTINWINDOW) {
		cur_bw = bw_reqid;
		bw_reqid = -1;
		fprintf(stderr, "Request not in window, bandwidth is %d\n",
			cur_bw);
	    }
	    else {
		perror("ody_frequest");
		exit(1);
	    }
	}
	else {
	    /* Block indefinitely */
	    rc = select(1, NULL, NULL, NULL, NULL);
	    
	    /* The only way we can leave select is if we got a signal */
	    ASSERT(rc == -1);
	    ASSERT(errno == EINTR);
	}
	/* Repeat the adaptive loop */
    }
}
