#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <dirent.h>

#include "include/odytypes.h"

#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <lwp/lock.h>

#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include <db.h>
#include <include/viceroy_onode.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/onode-manager.h"
#include "odyssey/wardens/map/mapw_private.h"
#include "ody_map.h"
#include "mapserver.h"

extern int distill(char* filename, int* filterIds, int numIds);

static int PacketSize = SFTP_MAXPACKETSIZE;

/* Table of looked up maps */
struct map map[MAP_TABLE_SIZE];

/* The current filter */
int* filter = NULL;
int filter_num = 0;

/* The server functions */

/* checks to see if the specified map exists on the server */
long map_lookup(RPC2_Handle _cid, RPC2_CountedBS *nameBS,
		RPC2_Integer *vnump, RPC2_Integer *vlen)
{
  struct stat sbuf;
  int i;
  char *name = nameBS->SeqBody;
  int max = nameBS->SeqLen;

  if ((max < 1) || (name[max-1] != '\0'))
    return EINVAL;

  if (stat(name,&sbuf) < 0) 
    return errno;
  
  /* if we've already looked it up, it's in the table */
  for (i = 0; i < MAP_TABLE_SIZE; i++)
    if (!strcmp(map[i].name, name)) {
      *vnump = i;
      *vlen = sbuf.st_size;
      return RPC2_SUCCESS;
    }
  
  /* It exists, so put it in the first empty slot in the table */
  for (i = 0; (i < MAP_TABLE_SIZE) && map[i].name[0]; i++);
  
  ASSERT(i < MAP_TABLE_SIZE); /* assert that we never fill up the table */
  
  strcpy(map[i].name, name);
  map[i].filter = NULL;
  map[i].filter_num = 0;
  *vnump = i;
  *vlen = sbuf.st_size;
  return RPC2_SUCCESS;
}

/* returns the specified map */
long map_open (RPC2_Handle _cid, RPC2_Integer vnum, SE_Descriptor* dummy_sftp)
{
  struct map* phdr;
  char filename[256];
  int rc, i;
  SE_Descriptor map_sftp;
  int fd;

  /* make sure it's a valid and open map */
  if ((vnum < 0) || (vnum >= MAP_TABLE_SIZE))
    return EINVAL;
  
  if (map[vnum].name[0] == '\0') 
    return ENOENT;

  printf ("Request for file %s\n", map[vnum].name);

  if (!strcmp (&map[vnum].name[strlen(map[vnum].name)-4], ".hdr")) {

    /* If this is a header file, we need to decide on the distillation
       quality.  This is the value of the current filter. */
    if (filter == NULL) {

      /* No filter specified.  In this case, use the actual header */
      fd = open (map[vnum].name, O_RDONLY, 0);
      if (fd < 0) {
	perror ("open header");
	return fd;
      }
      printf ("Serving file %s\n", map[vnum].name);
      
    } else {

      /* Distill the header and data files - use the distilled header */
      strcpy (filename, map[vnum].name);
      filename[strlen(filename)-4] = '\0';
      distill (filename, filter, filter_num);
      
      /* Open the distilled header */
      strcpy (filename, map[vnum].name);
      strcat (filename, ".dist");
      fd = open (filename, O_RDONLY, 0);
      if (fd < 0) {
	perror ("open header");
	return fd;
      }
      printf ("Serving file %s\n", filename);

    }

    /* Save filter for when data file is opened - they must match! */
    map[vnum].filter = filter;
    map[vnum].filter_num = filter_num;

  } else if (!strcmp (&map[vnum].name[strlen(map[vnum].name)-4], ".dat")) {

    /* If this is a data file, find the matching header file.  It should
       already have been looked up and opened. */
    strcpy (filename, map[vnum].name);
    strcpy (&filename[strlen(filename)-4], ".hdr");
    phdr = NULL;
    for (i = 0; i < MAP_TABLE_SIZE; i++) {
      if (!strcmp (map[i].name, filename)) {
	phdr = &map[i];
	break;
      }
    }
    if (phdr == NULL) {
      printf ("Header file %s not found for data file %s\n", filename, map[vnum].name);
      return (ENOENT);
    }
    if (phdr->filter == NULL) {

      /* No filter used for header - use the actual data file */
      fd = open (map[vnum].name, O_RDONLY, 0);
      if (fd < 0) {
	perror ("open data");
	return fd;
      }
      printf ("Serving file %s\n", map[vnum].name);

    } else {

      /* Header was distilled - use the distilled data file */
      strcpy (filename, map[vnum].name);
      strcat (filename, ".dist");
      fd = open (filename, O_RDONLY, 0);
      if (fd < 0) {
	perror ("open data");
	return fd;
      }
      printf ("Serving file %s\n", filename);

    }

  } else {

    /* Filename doesn't conform to conventions */
    printf ("WARNING: Unrecognized file type: %s\n", map[vnum].name);
    fd = open (map[vnum].name, O_RDONLY, 0);
    if (fd < 0) {
      perror ("open data");
      return fd;
    }

  }

  bzero(&map_sftp,sizeof(map_sftp));
  map_sftp.Tag = SMARTFTP;
  map_sftp.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
  map_sftp.Value.SmartFTPD.ByteQuota = -1;
  map_sftp.Value.SmartFTPD.Tag = FILEBYFD;
  map_sftp.Value.SmartFTPD.FileInfo.ByFD.fd = fd;
  fflush (stdout);

  if ((rc = RPC2_InitSideEffect(_cid, &map_sftp)) != RPC2_SUCCESS) {
    fprintf(stderr,"Getmap_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
	    RPC2_ErrorMsg(rc));
    return ETIMEDOUT;
  }
  
  if ((rc = RPC2_CheckSideEffect(_cid, &map_sftp,SE_AWAITLOCALSTATUS)) 
      != RPC2_SUCCESS) {
    fprintf(stderr,"GetMap_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
	    RPC2_ErrorMsg(rc));
    return ETIMEDOUT;
  }
  
  close (fd);

  return RPC2_SUCCESS;
}

/* returns the root directory */
long map_getdir (RPC2_Handle _cid, RPC2_Integer* maxlen, 
		 SE_Descriptor* dummy_sftp)
{
  SE_Descriptor map_sftp;
  struct dirent* dp;
  struct ody_dirent odp;
  char* data;
  DIR* dir;
  int size = sizeof(struct ody_dirent)-(ODY_MAXNAMLEN+1);
  int count = 0;
  int uniq = 1;
  int rc;

  dir = opendir (".");
  if (!dir) {
    perror ("open directory");
    return EINVAL;
  }

  /* Create a buffer to contain directory info */
  NALLOC (data, char, *maxlen);

  /* Read the directory */
  while ((dp = readdir (dir)) != NULL) {
    /* Only bother to send back regular files */
    /* Also hide distilled file since they are Odyssey-specific */
    if (dp->d_type == DT_REG && 
	strcmp(&dp->d_name[strlen(dp->d_name)-5], ".dist")) {
      odp.d_fileno = uniq++;
      odp.d_type = ODT_REG;
      odp.d_namlen = strlen(dp->d_name);
      strcpy (odp.d_name, dp->d_name);
      odp.d_reclen = size + ((odp.d_namlen+1 + 3) &~ 3);
      if (count + odp.d_reclen > *maxlen) {
	printf ("Not enough space to send back all files\n");
      } else {
	memcpy (&data[count], &odp, odp.d_reclen);
	count += odp.d_reclen;
      }
    }
  }

  closedir (dir);

  bzero(&map_sftp,sizeof(map_sftp));
  map_sftp.Tag = SMARTFTP;
  map_sftp.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
  map_sftp.Value.SmartFTPD.ByteQuota = -1;
  map_sftp.Value.SmartFTPD.Tag = FILEINVM;
  map_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = count;
  map_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = data;
  fflush (stdout);

  if ((rc = RPC2_InitSideEffect(_cid, &map_sftp)) != RPC2_SUCCESS) {
    fprintf(stderr,"Getdir_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
	    RPC2_ErrorMsg(rc));
    return ETIMEDOUT;
  }
  
  if ((rc = RPC2_CheckSideEffect(_cid, &map_sftp, SE_AWAITLOCALSTATUS)) 
      != RPC2_SUCCESS) {
    fprintf(stderr,"Getdir_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
	    RPC2_ErrorMsg(rc));
    return ETIMEDOUT;
  }

  *maxlen = count;
  return RPC2_SUCCESS;
}

int cmpInt (const void *a, const void *b)
{
  return (*((int *) a) - *((int *)b));
}

/* update the filter */
long map_setfilter (RPC2_Handle _cid, RPC2_BoundedBS* filterBBS)
{
  if (filter) {
    FREE (filter);
  }

  filter_num = filterBBS->SeqLen / sizeof (int);
  if (filter_num > 0) {
    
    NALLOC (filter, int, filter_num);
    memcpy (filter, filterBBS->SeqBody, filterBBS->SeqLen);
    qsort (filter, filter_num, sizeof(int), cmpInt);
  } else {
    filter = NULL;
  }

  return RPC2_SUCCESS;
}

/* initialize the RPC2 subsystem */
void Init_RPC(void)
{
  PROCESS dummy;
  RPC2_PortIdent pid;
  RPC2_SubsysIdent sid;
  SFTP_Initializer sftpi;
  int rc;

  if (LWP_Init(LWP_VERSION, LWP_NORMAL_PRIORITY, &dummy) != LWP_SUCCESS) {
    printf ("Can't initialize RPC2\n");
    exit (1);
  } /* endif */

  SFTP_SetDefaults(&sftpi);
  sftpi.EnforceQuota = 1;
  sftpi.DoPiggy = 0;
  sftpi.PacketSize = PacketSize;
  SFTP_Activate(&sftpi);

  pid.Tag = RPC2_PORTBYINETNUMBER;
  pid.Value.InetPortNumber = htons(MAPW_SERVER_PORT);
  rc = RPC2_Init(RPC2_VERSION, NULL, &pid, -1, NULL);
  ASSERT (rc == RPC2_SUCCESS);
  
  sid.Tag = RPC2_SUBSYSBYID;
  sid.Value.SubsysId = MAP_SUBSYSID;
  rc = RPC2_Export(&sid);
  ASSERT (rc == RPC2_SUCCESS);
}

/* set up threads, RPC, and other stuff */
static void Init(void)
{
  struct stat root_sbuf; /* the root of the volume */

  if (chdir(SERVER_ROOT) < 0) {
    perror(SERVER_ROOT);
    exit(1);
  }
  
  if (stat(SERVER_ROOT,&root_sbuf) < 0) {
    perror(SERVER_ROOT);
    exit(1);
  }
  
  Init_RPC();
}

int main(int argc, char **argv)
{
  RPC2_Handle cid;
  RPC2_RequestFilter reqfilter;
  RPC2_PacketBuffer *reqbuffer;
  int rc;
  static void Init(void);

  Init();

  reqfilter.FromWhom = ANY;
  reqfilter.OldOrNew = OLDORNEW;
  reqfilter.ConnOrSubsys.SubsysId = MAP_SUBSYSID;

  for ( ; ; ) {
    rc = RPC2_GetRequest(&reqfilter, &cid, &reqbuffer, NULL, NULL, 0, NULL);
    if (rc != RPC2_SUCCESS)
      fprintf(stderr, RPC2_ErrorMsg(rc));
    rc = ody_map_ExecuteRequest(cid, reqbuffer, NULL);
    if (rc != RPC2_SUCCESS)
      fprintf(stderr, RPC2_ErrorMsg(rc));
  }
}

void iopen ()
{
  ASSERT(0);
}
