/* xa_qt_srv_stub.c
   
   server main loop and RPC2 procedures
*
* Dushyanth Narayanan
*
* January 20 1997
*
* TODO
*
* ---- threading for handling multiple RPCs, using IOmgr for file ops and
*      eliminating a lot of global state that prevents concurrent movie opens.
*/

#include "xanim.h"
#include <strings.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <limits.h>

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <lwp/lock.h>

#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "include/odytypes.h"

#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"

#include <db.h>
#include "odyssey/wardens/qt/qt_server.h"
#include "ody_qt.h"
#include "xa_qt_srv.h"

static int PacketSize = SFTP_MAXPACKETSIZE;
static int num_workers = 5; /* number of server threads */

/* This lock is held by all threads whenever they access server structures;
   the idea is NOT to hold it when doing transfers to the client, so that
   we don't hold up other server threads */
/* NOTE: this is not very efficient, but it is quick and easy. If there are
   performance problems, then we have to go through and use multiple locks,
   and perhaps use readlocks instead of writelocks where appropriate */
static ds_hwlock_t qtsrv_global_lock;
#define GLOCK ds_hwlock_write(&qtsrv_global_lock)
#define GUNLOCK ds_hwlock_release_write(&qtsrv_global_lock);

/* This is the main server loop, executed by each worker thread */
static int
QT_Serve(argp)
    char *argp;
{
  RPC2_Handle cid;
  RPC2_RequestFilter reqfilter;
  RPC2_PacketBuffer *reqbuffer;
  int rc;

  reqfilter.FromWhom = ANY;
  reqfilter.OldOrNew = OLDORNEW;
  reqfilter.ConnOrSubsys.SubsysId = QTSUBSYSID;

  for ( ; ; ) {
    rc = RPC2_GetRequest(&reqfilter, &cid, &reqbuffer, NULL, NULL, 0, NULL);
    if (rc != RPC2_SUCCESS)
      fprintf(stderr, RPC2_ErrorMsg(rc));
    rc = QT_ExecuteRequest(cid, reqbuffer, NULL);
    if (rc != RPC2_SUCCESS)
      fprintf(stderr, RPC2_ErrorMsg(rc));
  }

  /* NOTREACHED */
  return 0;
}


/* Initialization for RPC2 subsystem; called by Init() */
static void Init_RPC(void)
{
  RPC2_PortIdent pid;
  RPC2_SubsysIdent sid;
  SFTP_Initializer sftpi;
  int rc;

  SFTP_SetDefaults(&sftpi);
  sftpi.EnforceQuota = 1;
  sftpi.DoPiggy = 0;
  sftpi.PacketSize = PacketSize;
  SFTP_Activate(&sftpi);

  pid.Tag = RPC2_PORTBYINETNUMBER;
  pid.Value.InetPortNumber = htons(QTPORTAL);
  rc = RPC2_Init(RPC2_VERSION, NULL, &pid, -1, NULL);
  ASSERT (rc == RPC2_SUCCESS);

  sid.Tag = RPC2_SUBSYSBYID;
  sid.Value.SubsysId = QTSUBSYSID;
  rc = RPC2_Export(&sid);
  ASSERT (rc == RPC2_SUCCESS);
}

/* Initialization routine: set up threads, RPC, and other stuff */

static void Init(void)
{
    struct stat root_sbuf; /* the root of the volume */
    int i;

#ifdef INSTRUMENTED
    oldcwd_fd = open(".",O_RDONLY);
    signal(SIGINT, call_exit);
    signal(SIGUSR1, pstats);
#endif /* INSTRUMENTED */

  if (chdir(SERVER_ROOT) < 0) {
      perror(SERVER_ROOT);
      exit(1);
  }

 if (stat(SERVER_ROOT,&root_sbuf) < 0) {
      perror(SERVER_ROOT);
      exit(1);
  }

  ds_thread_init(32768,0);

  Init_RPC();

  /* Initialized the global lock and start up the worker threads */
  ds_hwlock_init(&qtsrv_global_lock);
  for (i = 0; i < num_workers; i++)
      ASSERT(ds_thread_create(QT_Serve, (char *)i, DS_THREAD_NORMAL,
			      "qt_server", 0));
}

int main(int argc, char **argv)
{
  xaLONG i,j;
  char *in;

  /* parse command line */
  for(i=1; i < argc; i++) {
    in = argv[i];
    if ( (in[0] == '-') || (in[0] == '+') )
    {
      xaLONG len,opt_on;
      
      if (in[0] == '-') opt_on = xaFALSE;
      else opt_on = xaTRUE;
      /* if + turns off then reverse opt_on */
      if (DEFAULT_PLUS_IS_ON == xaFALSE) opt_on = (opt_on==xaTRUE)?xaFALSE:xaTRUE;
	
      len = strlen(argv[i]);
      j = 1;
     while( (j < len) && (in[j] != 0) )
     {
       switch(in[j])
       {
       case 'v':
	 xa_verbose = opt_on;	j++;	break;
       case 'p': {
	   char *end = in + j + 1;
	   PacketSize = strtol(end, &end, 10); j = end - in; break;
       }
       case 'd': 
	 RPC2_DebugLevel = 10; j++; break;
       case 'n': {
	   char *end = in + j + 1;
	   num_workers = strtol(end, &end, 10); j = end - in; break;
       }
       default:
	 fprintf(stderr,"Usage: qt_server [+v] [+p<PacketSize>] [+n<numthreads>]\n");
	 exit(1);
       } /* end of option switch */
     } /* end of loop through options */
    } /* end of if - */
  }

  Init();

  /* Now the main thread just sleeps, on an event that no one will signal */
  ds_thread_wait((char *)&main);
  /* NOTREACHED */
  return 0;
}

/* Table of looked-up movies */

#define MOVIE_TABLE_SIZE 256
static struct movie
{
  char name[256];
  XA_ANIM_HDR **tracks;
  int num_tracks;
}
movie[MOVIE_TABLE_SIZE];

/* The server functions */

/* lookup -- verifies if a given file name exists under the root */
/* check that the file exists, put it in the movie table (if it isn't already
   there), and return its position in the table as the vnum:
   NOTE: this means the vnum names it, so you can't ever evict things from
   the table. Might want to change that sometime */

long QT_Lookup_RPC(RPC2_Handle _cid, RPC2_CountedBS *nameBS,
		   RPC2_Integer *vnump)
{
  struct stat sbuf;
  int i;
  char *name = nameBS->SeqBody;
  int max = nameBS->SeqLen;

  if ((max < 1) || (name[max-1] != '\0'))
    return EINVAL;

  GLOCK;

  /* if we've already looked it up, it's in the table */
  for (i = 0; i < MOVIE_TABLE_SIZE; i++)
      if (!strcmp(movie[i].name,name)) {
	  *vnump = i;
	  GUNLOCK;
	  return RPC2_SUCCESS;
      }

  if (stat(name,&sbuf) < 0) {
      GUNLOCK;
      return errno;
  }

  /* It exists, so put it in the first empty slot in the table */
  for (i = 0; (i < MOVIE_TABLE_SIZE) && movie[i].name[0]; i++);

  ASSERT(i < MOVIE_TABLE_SIZE); /* assert that we never fill up the table */

  strcpy(movie[i].name, name);
  *vnump = i;
  GUNLOCK;
  return RPC2_SUCCESS;
}

/* open movie and return array of headers, one per track */

long QT_OpenMovie_RPC(RPC2_Handle _cid, RPC2_Integer vnum,
		      RPC2_Integer *max_tracks, 
		      RPC2_Integer *thumbnail_track,
		      RPC2_BoundedBS *hdr_info)
{
  XA_ANIM_HDR **QT_OpenMovie();
  int i;
  XA_ANIM_HDR **tracks;
  int num_tracks;
  QT_TrackHdr *all_hdrs;

  GLOCK;

  /* make sure we know about this movie */
  if ((vnum < 0) || (vnum >= MOVIE_TABLE_SIZE) ||
      (movie[vnum].name[0] == '\0')) {
      GUNLOCK;
      return EINVAL;
  }

  /* Check if we've already opened the movie */
  if (movie[vnum].tracks) {
      tracks = movie[vnum].tracks;
      *max_tracks = movie[vnum].num_tracks;
  }
  else {
      /* the name passed in should be a directory containing the track files */
      if (chdir(movie[vnum].name) < 0) {
	  GUNLOCK;
	  return errno;
      }

      /* open all the tracks */
      
      tracks = QT_OpenMovie(max_tracks,0 /* no audio attempt */);

      if (chdir(SERVER_ROOT) < 0) {
	  perror(SERVER_ROOT);
	  ASSERT(0);
      }
      if (*max_tracks <= 0)
	  return ENOENT; /* no available tracks */

      movie[vnum].tracks = tracks;
      movie[vnum].num_tracks = *max_tracks;

      if (xa_verbose)
	  fprintf(stderr,"Opened %d tracks of movie %s\n",
		  movie[vnum].num_tracks, movie[vnum].name);

  }


  /* make sure client sent us enough space to fit all the anim_hdrs */
  if (hdr_info->MaxSeqLen < *max_tracks*sizeof(QT_TrackHdr))
    *max_tracks = hdr_info->MaxSeqLen/sizeof(QT_TrackHdr);

  num_tracks = *max_tracks;

  hdr_info->SeqLen = (RPC2_Integer) num_tracks * sizeof(QT_TrackHdr);

  /* copy all the anim_hdrs and acts into one big array */

  all_hdrs = (QT_TrackHdr *)hdr_info->SeqBody;

  for (i = 0; i < num_tracks; i++) {

    /* copy anim_hdr info */
    memcpy(&(all_hdrs[i].codec_name),&(tracks[i]->codec_name),
	   sizeof(all_hdrs[i].codec_name));

    all_hdrs[i].compression_type = htonl(tracks[i]->qt_id);
    all_hdrs[i].depth = htonl(tracks[i]->qt_depth);
    all_hdrs[i].imagex = htonl(tracks[i]->imagex);
    all_hdrs[i].imagey = htonl(tracks[i]->imagey);
    all_hdrs[i].ave_fvid_size = htonl(tracks[i]->ave_fvid_size);
    all_hdrs[i].max_fvid_size = htonl(tracks[i]->max_fvid_size);
    all_hdrs[i].anim_type = htonl(tracks[i]->anim_type);
    all_hdrs[i].anim_flags = htonl(tracks[i]->anim_flags);
    all_hdrs[i].total_time = htonl(tracks[i]->total_time);
    all_hdrs[i].last_frame = htonl(tracks[i]->last_frame);
    all_hdrs[i].act_type = htonl(tracks[i]->acts->type);
    all_hdrs[i].cmap_rev = htonl(tracks[i]->acts->cmap_rev);
  }
  
  if ((num_tracks > 0) && !strcmp(all_hdrs[0].codec_name,"thumbnail")) {
      *thumbnail_track = 1;
      if (xa_verbose)
	  fprintf(stderr,"Track 0 is thumbnail\n");
  }
  else
      *thumbnail_track = 0;

  GUNLOCK;
  return RPC2_SUCCESS;
}  

/* return track info for a specific track */
long QT_OpenTrack_RPC(RPC2_Handle _cid, RPC2_Integer vnum,
		      RPC2_Integer track_id, SE_Descriptor *dummy_sed)
{
    XA_ANIM_HDR *track = NULL;
    QT_FrameInfo *all_frames = NULL;
    SE_Descriptor frame_list_sftp;
    int i;
    int rc = 0;

    GLOCK;

    /* make sure it's a valid and open movie */
    if ((vnum < 0) || (vnum >= MOVIE_TABLE_SIZE) ||
	(movie[vnum].tracks == NULL)) {
	GUNLOCK;
	return EINVAL;
    }

    /* and a valid track number */
    if ((track_id < 0) || (track_id >= movie[vnum].num_tracks) ||
	(movie[vnum].tracks[track_id] == NULL)) {
	GUNLOCK;
	return EINVAL;
    }

    track = movie[vnum].tracks[track_id];

    NALLOC(all_frames, QT_FrameInfo, track->last_frame + 1);

    /* copy the frame_lst and dlta_hdr information */
    for (i = 0; i <= track->last_frame; i++) {
	ACT_DLTA_HDR *dlta_hdr =
	    (ACT_DLTA_HDR *) track->frame_lst[i].act->data;

	all_frames[i].track_ID = htonl(track_id);
	all_frames[i].frame_no = htonl(i);
	all_frames[i].time_dur = htonl(track->frame_lst[i].time_dur);
	all_frames[i].zztime = htonl(track->frame_lst[i].zztime);
	all_frames[i].xapi_rev = htonl(dlta_hdr->xapi_rev);
	all_frames[i].flags = htonl(dlta_hdr->flags);
	all_frames[i].xpos = htonl(dlta_hdr->xpos);
	all_frames[i].ypos = htonl(dlta_hdr->ypos);
	all_frames[i].xsize = htonl(dlta_hdr->xsize);
	all_frames[i].ysize = htonl(dlta_hdr->ysize);
	all_frames[i].special = htonl(dlta_hdr->special);
	all_frames[i].size = htonl(dlta_hdr->fsize);
	all_frames[i].bufpos = htonl(dlta_hdr->fpos);
    }

    /* We release the global lock BEFORE doing the SFTP transfer */
    GUNLOCK;

    /* return the list via SFTP */
    bzero(&frame_list_sftp,sizeof(frame_list_sftp));
    frame_list_sftp.Tag = SMARTFTP;
    frame_list_sftp.Value.SmartFTPD.Tag = FILEINVM;
    frame_list_sftp.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
    frame_list_sftp.Value.SmartFTPD.ByteQuota = (track->last_frame + 1)
	* sizeof(QT_FrameInfo);
    frame_list_sftp.Value.SmartFTPD.SeekOffset = 0;
    frame_list_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen =
	(track->last_frame + 1) * sizeof(QT_FrameInfo);
    frame_list_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = 
	(RPC2_ByteSeq) all_frames;


    if ((rc = RPC2_InitSideEffect(_cid,&frame_list_sftp)) != 
	RPC2_SUCCESS) {
	fprintf(stderr,"QT_OpenTrack_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	rc = ETIMEDOUT;
	goto bad;
    }
    
    if ((rc = RPC2_CheckSideEffect(_cid,&frame_list_sftp,
				   SE_AWAITLOCALSTATUS)) != RPC2_SUCCESS) {
	fprintf(stderr,"QT_OpenTrack_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	rc = ETIMEDOUT;
	goto bad;
    }

    FREE(all_frames);

    GLOCK;

    if (xa_verbose)
	fprintf(stderr,"Returned header of track %s of movie %s\n",
		track->codec_name, movie[vnum].name);

    GUNLOCK;
		
    return RPC2_SUCCESS;

 bad: /* we should never come here with the global lock held */
    ASSERT(rc);
    if (all_frames) FREE(all_frames);
    return rc;
}

/* return a whole track */
long QT_FetchTrack_RPC(RPC2_Handle _cid, RPC2_Integer vnum,
		       RPC2_Integer track_id, SE_Descriptor *dummy_sed)
{
    XA_ANIM_HDR *track = NULL;
    int rc;
    SE_Descriptor track_sftp;
    off_t cur_fpos;

    GLOCK;

    /* make sure it's a valid and open movie */
    if ((vnum < 0) || (vnum >= MOVIE_TABLE_SIZE) ||
	(movie[vnum].tracks == NULL)) {
	GUNLOCK;
	return EINVAL;
    }

    /* and a valid track number */
    if ((track_id < 0) || (track_id >= movie[vnum].num_tracks) ||
	(movie[vnum].tracks[track_id] == NULL)) {
	GUNLOCK;
	return EINVAL;
    }

    track = movie[vnum].tracks[track_id];

    cur_fpos = lseek(track->vid_fd, 0, SEEK_SET);

    bzero(&track_sftp,sizeof(track_sftp));
    track_sftp.Tag = SMARTFTP;
    track_sftp.Value.SmartFTPD.Tag = FILEBYFD;
    track_sftp.Value.SmartFTPD.TransmissionDirection =
	SERVERTOCLIENT;
    track_sftp.Value.SmartFTPD.ByteQuota = -1;
    track_sftp.Value.SmartFTPD.SeekOffset = 0;
    track_sftp.Value.SmartFTPD.FileInfo.ByFD.fd = track->vid_fd;

    if (xa_verbose)
	fprintf(stderr,"Will SFTP from track %s/%s\n",
		movie[vnum].name, track->codec_name);

    /* We release the global lock before doing the transfer; since we
       are reading from a file that no one should ever write, and we
       aren't accessing any other shared structure, we should be ok */
    GUNLOCK;

    if ((rc = RPC2_InitSideEffect(_cid,&track_sftp)) != 
	RPC2_SUCCESS) {
	fprintf(stderr,"QT_FetchTrack_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	lseek(track->vid_fd, cur_fpos, SEEK_SET);
	return ETIMEDOUT;

    }

    if ((rc = RPC2_CheckSideEffect(_cid,&track_sftp,
				   SE_AWAITLOCALSTATUS)) != RPC2_SUCCESS) {
	fprintf(stderr,"QT_FetchTrack_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	lseek(track->vid_fd, cur_fpos, SEEK_SET);
	return ETIMEDOUT;
    }

    GLOCK;
    lseek(track->vid_fd, cur_fpos, SEEK_SET);
    GUNLOCK;

    if (xa_verbose) 
	fprintf(stderr,"SFTPed %ld bytes\n",
		track_sftp.Value.SmartFTPD.BytesTransferred);
    return RPC2_SUCCESS;
}


/* return a specified frame and related data to the client
   the actual image data is served via SFTP, the rest as RPC2 data structures
*/

long QT_GetFrame_RPC(RPC2_Handle _cid, RPC2_Integer vnum,
		     RPC2_Integer track_id, RPC2_Integer zztime,
		     RPC2_Integer maxbytes, RPC2_BoundedBS *frame_infoBS,
		     SE_Descriptor *dummy_sed)
{ 
    int rc;
    SE_Descriptor frame_data_sftp;
    XA_ANIM_HDR *track = NULL;
    XA_FRAME *frame_list = NULL;
    ACT_DLTA_HDR *dlta_hdr = NULL;
    QT_FrameInfo *frame_info = (QT_FrameInfo *) frame_infoBS->SeqBody;
    int maxframes = frame_infoBS->MaxSeqLen/sizeof(QT_FrameInfo);
    long offset, size, frame_no, last_frame;
    int i;
    struct timeval tv1, tv2;

    GLOCK;

    ASSERT(frame_infoBS->MaxSeqLen == maxframes * sizeof(QT_FrameInfo));
    ASSERT(frame_infoBS->SeqLen == 0);
       
    /* make sure it's a valid and open movie */
    if ((vnum < 0) || (vnum >= MOVIE_TABLE_SIZE) ||
	(movie[vnum].tracks == NULL)) {
        fprintf (stderr, "Oops: not a valid and open movie\n");
	GUNLOCK;
	return EINVAL;
    }

    /* and a valid track number */
    if ((track_id < 0) || (track_id >= movie[vnum].num_tracks) ||
	(movie[vnum].tracks[track_id] == NULL)) {
        fprintf (stderr, "Oops: not a valid track number\n");
	GUNLOCK;
	return EINVAL;
    }

    track = movie[vnum].tracks[track_id];
    frame_list = track->frame_lst;
    last_frame = track->last_frame;

    /* and a valid frame # */
    if ((zztime < 0) || (zztime > frame_list[last_frame].zztime)) {
        fprintf (stderr, "Oops: not a valid frame number: %ld\n", zztime);
	GUNLOCK;
	return EINVAL;
    }

    /* find the first frame # at >= zztime
       first approximation by assuming constant frame rate,
       then linear search */

    /* JNF - for big movies, this overflows.  Therefore use floats
       to do this calculation */
    frame_no = (long) ((float) last_frame * (float) zztime/ (float) frame_list[last_frame].zztime);

    ASSERT(frame_no >= 0);
    ASSERT(frame_no <= last_frame);
    
    /* if it's behind, push it forward */
    while ((frame_no < last_frame) && (frame_list[frame_no].zztime < zztime))
	frame_no++;
    
    /* if it's ahead, push it back */
    while ((frame_no > 0) && (frame_list[frame_no-1].zztime >= zztime))
	frame_no--;

    dlta_hdr = (ACT_DLTA_HDR *) frame_list[frame_no].act->data;
    offset = dlta_hdr->fpos;

    size = 0;

    /* copy frame meta data */
    for (i = 0; i < maxframes; i++) {
	ACT_DLTA_HDR *dlta_hdr;

	if (frame_no + i > last_frame)
	    break;
	    
	dlta_hdr = (ACT_DLTA_HDR *) frame_list[frame_no+i].act->data;

	if (dlta_hdr->fpos - offset + dlta_hdr->fsize > maxbytes)
	    break;

	size = dlta_hdr->fpos - offset + dlta_hdr->fsize;
	frame_info[i].track_ID = htonl(track_id);
	frame_info[i].frame_no = htonl(frame_no+i);
	frame_info[i].time_dur = htonl(frame_list[frame_no+i].time_dur);
	frame_info[i].zztime = htonl(frame_list[frame_no+i].zztime);
	frame_info[i].xapi_rev = htonl(dlta_hdr->xapi_rev);
	frame_info[i].flags = htonl(dlta_hdr->flags);
	frame_info[i].xpos = htonl(dlta_hdr->xpos);
	frame_info[i].ypos = htonl(dlta_hdr->ypos);
	frame_info[i].xsize = htonl(dlta_hdr->xsize);
	frame_info[i].ysize = htonl(dlta_hdr->ysize);
	frame_info[i].special = htonl(dlta_hdr->special);
	frame_info[i].size = htonl(dlta_hdr->fsize);
	frame_info[i].bufpos = htonl(dlta_hdr->fpos-offset);
    }

    frame_infoBS->SeqLen = i * sizeof(QT_FrameInfo);

    /* copy frame data */

    bzero(&frame_data_sftp,sizeof(frame_data_sftp));
    frame_data_sftp.Tag = SMARTFTP;
    frame_data_sftp.Value.SmartFTPD.Tag = FILEBYFD;
    frame_data_sftp.Value.SmartFTPD.TransmissionDirection =
	SERVERTOCLIENT;
    frame_data_sftp.Value.SmartFTPD.ByteQuota = size;
    frame_data_sftp.Value.SmartFTPD.SeekOffset = offset;
    frame_data_sftp.Value.SmartFTPD.FileInfo.ByFD.fd = track->vid_fd;

    if (xa_verbose)
	fprintf(stderr,"QT_GetFrame_RPC: will SFTP %ld (frames %ld-%ld) bytes "
		"from %s/%s\n",	size, frame_no, frame_no + i-1,
		movie[vnum].name, track->codec_name);

    gettimeofday(&tv1, NULL);

    /* We release the global lock before doing the transfer; since we
       are reading from a file that no one should ever write, and we
       aren't accessing any other shared structure, we should be ok */
    GUNLOCK;

    if ((rc = RPC2_InitSideEffect(_cid,&frame_data_sftp)) != 
	RPC2_SUCCESS) {
	fprintf(stderr,"QT_GetFrame_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	return ETIMEDOUT;

    }

    if ((rc = RPC2_CheckSideEffect(_cid,&frame_data_sftp,
				   SE_AWAITLOCALSTATUS)) != RPC2_SUCCESS) {
	fprintf(stderr,"QT_GetFrame_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	return ETIMEDOUT;
    }

    gettimeofday(&tv2, NULL);
    timersub(&tv2,&tv1,&tv2);
    
    ASSERT(size == frame_data_sftp.Value.SmartFTPD.BytesTransferred);
    if (xa_verbose) {
	double bw = (double) size / (tv2.tv_sec + (double)tv2.tv_usec/1000000);
	
	fprintf(stderr,"QT_GetFrame_RPC: SFTPed %ld bytes in %ld.%06ld sec "
		"(%f bytes/sec)\n", size, tv2.tv_sec, tv2.tv_usec, bw);
    }

    return RPC2_SUCCESS;
}

/* close the movie and free up all the storage (but DON'T evict from
   the lookup table) */

long QT_CloseMovie_RPC(RPC2_Handle _cid, RPC2_Integer vnum)
{
    int i;

    GLOCK;

    if ((vnum < 0) || (vnum >= MOVIE_TABLE_SIZE) ||
	(movie[vnum].tracks == NULL)) {
	GUNLOCK;
	return EINVAL;
    }

    for (i = 0; i < movie[vnum].num_tracks; i++) {
	XA_ANIM_HDR *track = movie[vnum].tracks[i];
	if (track) {
	    XA_FRAME *frame_lst = track->frame_lst;
	    void Free_Actions();

	    Free_Actions(track->acts);
	    if (frame_lst) {
		int j;
		for (j = 0; j < track->last_frame; j++)
		    if (frame_lst[i].dlta_hdr)
			free(frame_lst[i].dlta_hdr);
		free(frame_lst);
	    }

	    /* this stuff could probably go away, since we never use
	       these fields anyway */

	    if (track->next_file == track)
		track->next_file = NULL;
	    
	    if (track->prev_file == track)
		track->prev_file = NULL;
	    
	    if (track->next_file)
		track->next_file->prev_file = track->prev_file;
	    
	    if (track->prev_file)
		track->prev_file->next_file = track->next_file;
	    
	    /* close the file descriptor and free it all up */
	    if (track->vid_fd >= 0) {
		close(track->vid_fd);
		track->vid_fd = -1;
	    }
	    
	    free(track);
	}
    }

    free(movie[vnum].tracks);
    movie[vnum].tracks = NULL;
    movie[vnum].num_tracks = 0;

    GUNLOCK;
    return RPC2_SUCCESS;

}

/* Respond to a "ping" by the client by sending back some bytes via SFTP */
/* "Ping" responses do not require us to hold the global lock, since we
   never do anything with server data structures */

long QT_Ping_RPC(RPC2_Handle _cid, RPC2_Integer numbytes,
		 SE_Descriptor *dummy_sed)
{
    SE_Descriptor ping_data_sftp;
    void *buf;
    int rc;

    BALLOC(buf, numbytes);

    bzero(&ping_data_sftp,sizeof(ping_data_sftp));
    ping_data_sftp.Tag = SMARTFTP;
    ping_data_sftp.Value.SmartFTPD.Tag = FILEINVM;
    ping_data_sftp.Value.SmartFTPD.TransmissionDirection =
	SERVERTOCLIENT;
    ping_data_sftp.Value.SmartFTPD.ByteQuota = numbytes;
    ping_data_sftp.Value.SmartFTPD.SeekOffset = 0;
    ping_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen =
	(RPC2_Integer) numbytes;
    ping_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = 
	(RPC2_ByteSeq) buf;

    if (xa_verbose)
	fprintf(stderr,"QT_Ping_RPC: will SFTP %ld bytes\n", numbytes);

    if ((rc = RPC2_InitSideEffect(_cid,&ping_data_sftp)) != 
	RPC2_SUCCESS) {
	fprintf(stderr,"QT_Ping_RPC: RPC2_InitSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	FREE(buf);
	return ETIMEDOUT;

    }

    if ((rc = RPC2_CheckSideEffect(_cid,&ping_data_sftp,
				   SE_AWAITLOCALSTATUS)) != RPC2_SUCCESS) {
	fprintf(stderr,"QT_Ping_RPC: RPC2_CheckSideEffect returned %d: %s\n",rc,
		RPC2_ErrorMsg(rc));
	FREE(buf);
	return ETIMEDOUT;
    }

    FREE(buf);

    fprintf(stderr,"SFTPed %ld bytes\n", numbytes);
    return RPC2_SUCCESS;
}


#ifdef INSTRUMENTED

static int oldcwd_fd = -1;

static void pstats(int sig)
{
}

/* interrupt handler calls exit(), to make sure profiling information is
   written out */


static void call_exit(int sig)
{
    fchdir(oldcwd_fd);
    close(oldcwd_fd);
    pstats(SIGUSR1);
    exit(sig);
}

#endif /* INSTRUMENTED */
