#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <kernel/upcall/includes/upcall.h>

#include "odyssey/wardens/speech/libodyspeech_tsops.h"

/* Default filename to use if not specified by user */
#define DEFAULT_FILENAME "/ody/speech/test"

/* Make this global so it can be accessed by signal handlers */
int fd = -1;

/* copyFile() copies all data from an input file to an output file.
   It expects two open file descriptors.  It returns zero on success,
   non-zero on error. */
int copyFile (int from_fd, int to_fd)
{
  struct stat fs;
  char* pBuffer = 0;
  int bytesRead, bytesWritten;
  int totalBytesRead = 0;
  int error = 0; /* Set if an error occurs while copying the file */

  if (fstat (from_fd, &fs) == 0) {

    /* Allocate buffer to contain file data */
    pBuffer = malloc (fs.st_size);
    assert (pBuffer);

    do {
      
      /* Read data into the buffer */
      bytesRead = read (from_fd, &pBuffer[totalBytesRead], fs.st_size);
      if (bytesRead > 0) {
	totalBytesRead += bytesRead;
      } else {
	perror ("Error reading data file\n");
	error = 1;
      } /* endif */
      
    } while (totalBytesRead < fs.st_size && !error);

    /* Write the data to the destination file */
    bytesWritten = write (to_fd, pBuffer, totalBytesRead);
    if (bytesWritten != totalBytesRead) {
      perror ("Error copying file");
      error = 1;
    } /* endif */

    free (pBuffer);

  } else {
    perror ("Error getting file statistics");
  } /* endif */

  return (error);
}

/* displayFile() reads all data from a file, displays it to the
   screen, and appends a new-line character.  It expects an open file
   descriptor.  It returns zero on success, non-zero on error. */
int displayFile (int fd)
{
  char buffer[4096];
  int bytesRead;
  int error = 0; /* Set if an error occurs while copying the file */

  do {

    /* Read a chunk of data from the source file */
    bytesRead = read (fd, buffer, sizeof(buffer)-1);
    if (bytesRead > 0) {

      /* Write the data to the screen */
      buffer[bytesRead] = '\0';
      printf ("%s", buffer);

    } else if (bytesRead < 0) {
      perror ("Error reading data file\n");
      error = 1;
    } /* endif */
    
  } while (bytesRead > 0 && !error);

  printf ("\n");

  return (error);
}

/* recognizeFile() recognizes the speech in a data file by copying the file
   to the file in the Odyssey name space, then reading the translation from
   the Odyssey speech file.  It return s zero on success, non-zero
   otherwise. */
int recognizeFile (int ody_fd, char* filename) 
{
  int rc = 0;

  /* Open the specified data file */
  int data_fd = open (filename, O_RDONLY);
  if (data_fd >= 0) {

    /* Write the raw speech data to the file */
    rc = copyFile (data_fd, ody_fd);
    if (rc == 0) {

      /* Read the translation from the file */
      rc = displayFile (ody_fd);

    } /* endif */
	  
  } else {
    perror ("Unable to open specified file");
    rc = 1;
  } /* endif */

  return (rc);
}

/* ShowFormat() displays the format string and exits when the user enters
   incorrect arguments on the command line. */
void ShowFormat () 
{
  printf ("format: speechw_estimate [ADC file] [configuration] [log]\n");
  exit (1);
}

/* Open a file in the speech warden's name space.  Then, invoke the UI.  
   After the user has exited the UI, close the file (to be nice). */
int main (int argc, char* argv[])
{
  char* filename = DEFAULT_FILENAME;
  char* utterance = 0;
  char* logfile = 0;
  int config = 0;
  int rc;

  if (argc != 4) {
    ShowFormat ();
  } /* endif */

  utterance = argv[1];
  config = atoi(argv[2]);
  logfile = argv[3];

  /* Open the file */
  fd = open (filename, O_RDWR);
  if (fd < 0) {
    printf ("Unable to open ody file, perhaps the viceroy isn't started\n");
    exit (3);
  } /* endif */

  /* Set the configuration as specified */
  rc = speechSetStrategy (fd, config);
  if (rc) {
    printf ("Error setting configuration to: %d, rc = %d\n", config, rc);
    exit (2);
  } /* endif */ 

  /* Recognize the utterance */
  if (recognizeFile (fd, utterance)) {
    printf ("Unable to recognize utterance %s\n", utterance);
    exit (4);
  } /* endif */

  /* Output results to the log file */
  if (dumpStats (fd, logfile)) {
    printf ("Unable to log data to file: %s\n", logfile);
    exit (5);
  } /* endif */

  /* Close the file */
  close (fd);
  fd = -1;

  return (0);
}
