#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <kernel/upcall/includes/upcall.h>
#ifdef NETTR_UPCALL
#include <nettr/replay/ntrp_upcall.h>
#endif
#ifdef PSCOPE 
#include <linux/pscope.h>
#endif

#include "odyssey/wardens/speech/libodyspeech_tsops.h"

/* Default filename to use if not specified by user */
#define DEFAULT_FILENAME "/ody/speech/test"
#define DEFAULT_MODEL "map"

/* Simple hack to represent current network connectivity */
enum strategyType connection;

/* Make this global so it can be accessed by signal handlers */
int fd = -1;

/* Variables used to coordinate speech replay with ntrace replay */
static int ntrace_started = 0;
static int ntrace_first_transition = 0;
static int ntrace_ended = 0;

#ifdef PACING
static int alarm_triggered = 0;

/* Called to "handle" sigalrm */
void discard_signal(int signo)
{
    alarm_triggered = 1;
}
#endif

/* networkUp ups the connectivity when a USR1 signal is received -
   will be replaced eventually with real bandwidth estimation */
void networkUp ()
{
  int rc;

  if (connection < SPEECHW_STRATEGY_REMOTE && fd >= 0) {
    connection++;
    printf ("Network connection strengthened\n");
    rc = speechSetStrategy (fd, connection);
    if (rc) {
      printf ("Error setting connection level to: %d, rc = %d\n",
	      connection, rc);
    } /* endif */
  } /* endif */
}

/* networkDown lowers the connectivity when a USR2 signal is received
   - will be replaced eventually with real bandwidth estimation */
void networkDown ()
{
  int rc;

  if (connection > SPEECHW_STRATEGY_LOCAL && fd >= 0) {
    connection--;
    printf ("Network connection weakened\n");
    rc = speechSetStrategy (fd, connection);    if (rc) {
      printf ("Error setting connection level to: %d, rc = %d\n",
	      connection, rc);
    } /* endif */
  } /* endif */
}

/* handleUpcall handles upcalls generated by trace replay start, stop,
   and transition.  Each event will genereate a message to the screen. */
int handleUpcall(upclass_t upclass, upcall_t upcall, size_t argsz, void* argp, 
		 size_t* retszp, void* retp, int flags)
{
#ifdef NETTR_UPCALL
  if (upclass == UC_NTRP) {
    
    switch (upcall) {
    case NTRP_NOTIFY_START:
      printf ("Replay started\n");
      ntrace_started = 1;
      break;
    case NTRP_NOTIFY_END:
      printf ("Replay completed\n");
      ntrace_ended = 1;
      break;
    case NTRP_NOTIFY_TSEV:
      printf ("Replay transition\n");
      ntrace_first_transition = 1;
      break;
    default:
      printf("Unkown UC_NTRP upcall %d\n", upcall);
    } /* end case */

  } else {
#endif NETTR_UPCALL
    printf ("Unknown upclass %d\n", upclass);
#ifdef NETTR_UPCALL
  } /* endif */
#endif NETTR_UPCALL

  return (0);
}

/* Invokes TSOP to set the configuration for an o-node */
void configure (int configuration)
{
  int rc;

  rc = speechSetStrategy (fd, configuration);
  if (rc) {
    printf ("Error setting connection level to: %d, rc = %d\n",
	    configuration, rc);
  } else {
    printf ("Connection set to: %d\n", configuration);
  } /* endif */
}

/* copyFileToBuffer() copies all data from an input file to a buffer.
   It allocates the buffer before copying the data.  The function returns
   zero on success, non-zero on error. */
int copyFileToBuffer (int from_fd, char** pBuffer, int* pLen)
{
  struct stat fs;
  int bytesRead;
  int totalBytesRead = 0;
  int error = 0; /* Set if an error occurs while copying the file */

  if (fstat (from_fd, &fs) == 0) {

    /* Allocate buffer to contain file data */
    *pLen = fs.st_size;
    *pBuffer = malloc (*pLen);
    assert (*pBuffer);

    do {
      
      /* Read data into the buffer */
      bytesRead = read (from_fd, &(*pBuffer)[totalBytesRead], *pLen);
      if (bytesRead > 0) {
	totalBytesRead += bytesRead;
      } else {
	perror ("Error reading data file\n");
	error = 1;
      } /* endif */
      
    } while (totalBytesRead < *pLen && !error);

  } else {
    perror ("Error getting file statistics");
  } /* endif */

  return (error);
}

/* recognizeBuffer() recognizes an utterance in a data buffer using an
   Odyssey TSOP. It returns zero on success, non-zero otherwise. */
int recognizeBuffer (int ody_fd, char* model, char* buffer, int len) 
{
  char translation[1024];
  int rc = 0;

#ifdef PSCOPE
  pscope_start (1, PSCOPE_INTR_LPT);
  if (rc) perror ("pscope_start");
#endif

  rc = speechRecognize (ody_fd, model, buffer, len, translation,
			sizeof (translation));
  if (rc == 0) {
    printf ("Translation: %s\n", translation);
  } else {
    printf ("Error reading translation, rc = %d\n", rc);
  } /* endif */

#ifdef PSCOPE
  rc = pscope_stop ();
  if (rc) perror ("pscope_stop");
#endif

  return (rc);
}

/* recognizeFile() recognizes the speech in a data file.  It returns
   zero on success, non-zero otherwise. */ 
int recognizeFile (int ody_fd, char* model, char* filename) 
{
  char* buffer;
  int len;
  int rc = 0;

  /* Open the specified data file */
  int data_fd = open (filename, O_RDONLY);
  if (data_fd >= 0) {

    /* Write the file into a memory buffer */
    rc = copyFileToBuffer (data_fd, &buffer, &len);
    if (rc == 0) {

      /* Recognize the buffer */
      recognizeBuffer (ody_fd, model, buffer, len);

      free (buffer);

    } /* endif */
	  
    close (data_fd);

  } else {
    perror ("Unable to open specified file");
    rc = 1;
  } /* endif */

  return (rc);
}

/* promptUser() simply prompts the user for an ADC file, recognizes the file,
   displays the output, and prompts the user again. */
/* NOTE: I'd like to replace this with a nice GUI */
void promptUser (int fd, char* model)
{
  char filename[255];
  int done = 0;

  do {

    printf ("ADC File name (enter 'Q' to quit): ");
    fgets (filename, 254, stdin);
    filename[strlen(filename)-1] = '\0';

    if (!strcmp (filename, "Q") || !strcmp (filename, "q")) {
      done = 1;
    } else if (!strcmp (filename, "small")) {
      speechSetFidelity (fd, SPEECHW_FIDELITY_REDUCED);
    } else if (!strcmp (filename, "full")) {
      speechSetFidelity (fd, SPEECHW_FIDELITY_FULL);
    } else {

      /* translate the file */
      if (recognizeFile (fd, model, filename) > 0) {
	printf ("Error recognizing speech data\n"); 
      } /* endif */

    } /* endif */

  } while (!done);
}

void benchmark (int fd, char* model, char* utterance, int config, 
#ifdef pacing
		float delay, 
#endif
		char* logfile)
{
  char* buffer;
  int len;
  int data_fd;

#ifdef PACING
  struct timeval currentTS, idealTS;
  double idealTime, dwellTime;
  struct itimerval interval;

  /* We'll be using timers to pace ourselves */
  signal(SIGALRM, discard_signal);
#endif

  /* Open the specified data file */
  data_fd = open (utterance, O_RDONLY);
  if (data_fd < 0) {
    printf ("Couldn't open utterance %s\n", utterance);
    return;
  } /* endif */

  /* Read the utterance */
  if (copyFileToBuffer (data_fd, &buffer, &len)) {
    printf ("Couldn't read utterance %s\n", utterance);
    return;
  } /* endif */
  close (data_fd);

  /* Set the configuration */
  configure (config);

  /* Wait for start of trace */
  while (!ntrace_started) {
    select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL,
	   (struct timeval *)NULL);
  } /* endwhile */

  /* Do a recognition before start of "official" trace */
  if (recognizeBuffer (fd, model, buffer, len)) {
    exit (1);
  } /* endif */

  /* Wait for 1st transition to start benchmark */
  while (!ntrace_first_transition) {
    select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL,
	   (struct timeval *)NULL);
  } /* endwhile */

  /* Perfrom recognitions until trace has ended */
  if (dumpStats (fd, logfile)) {
    printf ("Unable to log data to file: %s\n", logfile);
    return;
  } /* endif */

#ifdef PACING
  /* Wait for a random amount of time between 0 and delay.  This 
     avoids always missing trace events due to timing considerations. */
  dwellTime = (double) random() * delay / (double) RAND_MAX;
  timerclear(&interval.it_interval);
  interval.it_value.tv_sec = dwellTime;
  interval.it_value.tv_usec = 
    (dwellTime - (double) interval.it_value.tv_sec) * 1000000.0;

  alarm_triggered = 0;
  setitimer(ITIMER_REAL, &interval, NULL);
  while (!alarm_triggered) {
    pause();
  } /* endwhile */
  
  /* Trace has started - save this time */
  gettimeofday (&idealTS, NULL);
  idealTime = (double) idealTS.tv_sec + ((double) idealTS.tv_usec) / 1000000.0;
#endif

  do {

    if (recognizeBuffer (fd, model, buffer, len)) {
      exit (1);
    } /* endif */

    if (!ntrace_ended) {

      if (dumpStats (fd, logfile)) {
	printf ("Unable to log data to file: %s\n", logfile);
      } /* endif */

#ifdef PACING
      gettimeofday (&currentTS, NULL);
      idealTime += delay;
      idealTS.tv_sec = (int) idealTime;
      idealTS.tv_usec = ((idealTime - (double) idealTS.tv_sec) * 1000000.0);

      if (timercmp(&currentTS, &idealTS, <)) {
	/* we're still before the ideal time, so wait for it...*/
	timerclear(&interval.it_interval);
	timersub(&idealTS, &currentTS, &interval.it_value);
	alarm_triggered = 0;
	setitimer(ITIMER_REAL, &interval, NULL);
	while (!alarm_triggered) {
	  pause();
	} /* end while */
      } else {
	timersub(&currentTS, &idealTS, &interval.it_value);
	printf("LATE by %ld.%06ld\n", interval.it_value.tv_sec,
	       interval.it_value.tv_usec);
      }
#endif

    } /* endif */
  } while (!ntrace_ended);
}

void runScript (int fd, char* model, char* script, char* prime, 
		int handleUpcalls) 
{
  FILE* file;
  char* parm;
  char line[256];
  int done = 0;
  int config;
  char utterance[255] = "";
  char log_file[255] = "";

#ifdef PACING
  float delay;
#endif

  /* Open the script */
  file = fopen (script, "r");
  if (file) {

    if (handleUpcalls) {

      /* Wait for ntrace replay to start */
      while (!ntrace_started) {
	select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL,
	       (struct timeval *)NULL);
      } /* endwhile */

      if (prime) {
	/* Do 2 recognitions before start of "official" trace */
	if (recognizeFile (fd, model, prime)) {
	  exit (1);
	} /* endif */
	if (recognizeFile (fd, model, prime)) {
	  exit (1);
	} /* endif */

	/* Wait for 1st transition to start benchmark */
	while (!ntrace_first_transition) {
	  select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL,
		 (struct timeval *)NULL);
	} /* endwhile */

      } /* endif */

    } /* endif */

    while (!feof (file) && !done) {
      
      /* Read a line from the file */
      if (fgets (line, 256, file)) {

	if (line[strlen(line)-1] == '\n') {
	  line[strlen(line)-1] = '\0';
	} /* endif */

	if (!strncmp (line, "Recognize", strlen ("Recognize"))) {
	  parm = line + strlen ("Recognize") + 1;
	  if (recognizeFile (fd, model, parm)) {
	    printf ("Unable to recognize file: %s", parm);
	  } /* endif */
	} else if (!strncmp (line, "Configure", strlen ("Configure"))) {
	  parm = line + strlen ("Configure") + 1;
	  configure (atoi(parm));
	} else if (!strncmp (line, "DumpStats", strlen ("DumpStats"))) {
	  parm = line + strlen ("DumpStats") + 1;
	  if (dumpStats (fd, parm)) {
	    printf ("Unable to log data to file: %s\n", parm);
	  } /* endif */
	} else if (!strncmp (line, "Repeat", strlen ("Repeat"))) {
	  rewind (file);
	} else if (!strncmp (line, "Benchmark", strlen ("Benchmark"))) {
	  parm = line + strlen ("Benchmark") + 1;
	  sscanf (parm, "%s %d %s", utterance, &config,
		  log_file);
	  benchmark (fd, model, utterance, config, log_file);

#ifdef PACING
	  sscanf (parm, "%s %d %f %s", utterance, &config,
		  &delay, log_file);
	  benchmark (fd, model, utterance, config, delay, log_file);
#endif

	} else {
	  printf ("Invalid command: %s\n", line);
	} /* endif */

      } /* endif */

      if (handleUpcalls) {
	/* If trace replay has ended then we're done */
	if (ntrace_ended) {
	  done = 1;
	} /* endif */
      } /* endif */

    } /* endwhile */

    fclose (file);

  } else {
    printf ("Unable to open script %s\n", script);
  } /* endif */
}

/* ShowFormat() displays the format string and exits when the user enters
   incorrect arguments on the command line. */
void ShowFormat () 
{
  printf ("format: speechw_driver [-f warden file] [-r script] [-m model]\n");
  printf ("        [-b utterance config interval log] [-d]\n\n");
  printf ("        [warden file] is the optional specification of an Odyssey\n");
  printf ("        file name to use, i.e /ody/speech/dev\n");
  printf ("        [script] is an optional script to run\n");
  printf ("        [model] specified which Janus vocabulary to use\n");
  printf ("        [utterance] is an utterance to benchmark at\n");
  printf ("        configuration [config].  Results are written to [log]\n"); 
  printf ("        -d is used to display a final stat dump\n");
  exit (1);
}

/* Open a file in the speech warden's name space.  Then, invoke the UI.  
   After the user has exited the UI, close the file (to be nice). */
int main (int argc, char* argv[])
{
  StrategyModel strategyModel;
  char* filename = DEFAULT_FILENAME;
  char* model = DEFAULT_MODEL;
  char* script = 0;
  char* utterance = 0;
  char* logfile = 0;
  char* prime = 0;
  char* dumpFile = 0;
  int config = 0;
  int handleUpcalls = 0;
  int i;

#ifdef PACING
  float delay = 0.0;
  struct timeval currentTS;

  /* Initialize the random number generator */
  gettimeofday (&currentTS, NULL);
  srandom (currentTS.tv_usec);
#endif

  i = 1;
  while (i < argc) {
    if (!strcmp (argv[i], "-f")) {
      if (i+1 < argc) {	
	filename = argv[i+1];
	i += 2;
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-m")) {
      if (i+1 < argc) {
	model = argv[i+1];
	i += 2;
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-r")) {
      if (i+1 < argc) {
	script = argv[i+1];
	i += 2;
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-p")) {
      if (i+1 < argc) {
	prime = argv[i+1];
	i+= 2;
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-b")) {
      if (i+2 < argc) {
	utterance = argv[i+1];
	config = atoi (argv[i+2]);
#ifdef PACING
	delay = atof (argv[i+3]);
	logfile = argv[i+4];
#else
	logfile = argv[i+3];
#endif
	handleUpcalls = 1;
#ifdef PACING
	i += 5;
#else
	i += 4;
#endif
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-d")) {
      if (i+1 < argc) {
	dumpFile = argv[i+1];
	i+= 2;
      } else {
	ShowFormat ();
      } /* endif */
    } else if (!strcmp (argv[i], "-u")) {
      handleUpcalls = 1;
      i += 1;
    } else {
      ShowFormat ();
    } /* endif */
  } /* end while */
	
  /* Assume no connection to start */
  connection = SPEECHW_STRATEGY_LOCAL;

  /* Set handlers for SIGUSR1 and SIGUSR2 to represent connectivity changes */
  signal (SIGUSR1, networkUp);
  signal (SIGUSR2, networkDown);

  if (handleUpcalls) {
    printf ("Handling upcalls\n");
#ifdef NETTR_UPCALL
    if (upcall_reg (UC_NTRP, NTRP_NOTIFY_START, handleUpcall) == UC_ERR) {
      printf ("Unable to register upcall handler for trace begin\n");
    } /* endif */
    if (upcall_reg (UC_NTRP, NTRP_NOTIFY_END, handleUpcall) == UC_ERR) {
      printf ("Unable to register upcall handler for trace end\n");
    } /* endif */
    if (upcall_reg (UC_NTRP, NTRP_NOTIFY_TSEV, handleUpcall) == UC_ERR) {
      printf ("Unable to register upcall handler for transitions\n");
    } /* endif */
#endif
  } /* endif */

  /* Open the file */
  fd = open (filename, O_RDWR);
  if (fd >= 0) {

    /* Initialize the Janus vocabulary */
    speechInitModel (fd, model, &strategyModel);

    if (utterance) {
      /* Perform the benchmark test */
#ifdef PACING
      benchmark (fd, model, utterance, config, delay, logfile);
#endif
      benchmark (fd, model, utterance, config, logfile);
    } else if (script) {
      /* Run the specified script */
      runScript (fd, model, script, prime, handleUpcalls);
    } else {
      /* Invoke interactive user interface */
      promptUser (fd, model);
    } /* endif */
    
    if (dumpFile) {
      if (dumpStats (fd, dumpFile)) {
	printf ("Unable to log data to file: %s\n", dumpFile);
      } /* endif */
    } /* endif */

    speechCloseModel (fd, model);

    close (fd);
    fd = -1;

  } else {
    printf ("Unable to open /ody/speech/... file\n");
    printf ("Perhaps the Viceroy is not started\n");
    return (-1);
  } /* endif */

  return (0);
}
