#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "odyssey/wardens/speech/odyjanus.h"
#include "ody_speech.h"

#include "include/odytypes.h"
#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/sftp.h>

#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"

/* Represents a specific loaded speech model. */
typedef struct SpeechModel {
  char*               sm_model;
  int                 sm_refcount;
  char*               sm_fname;
  char*               sm_fidelity;
  char*               sm_recognizePerByteCost;
  char*               sm_VQPerByteCost;
  char*               sm_recNoVQPerByteCost;
  struct SpeechModel* sm_next;
} SpeechModel;

/* A list of all models currently loaded for this server. */
SpeechModel* pModels;

/* This is data needed by the remote procedures.  There is no way to get this 
   to the procedures other than making the data global */
struct {
  int cont;
  int inPipe;
  int outPipe;
} RPCClientData;

static int num_workers = 5; /* number of server threads */

/* Global lock held by threads when doing speech stuff, but NOT when doig
   xfers to/from client */
static ds_hwlock_t speechsrv_global_lock;
#define GLOCK ds_hwlock_write(&speechsrv_global_lock);
#define GUNLOCK ds_hwlock_release_write(&speechsrv_global_lock);

/* findModel scans the list of loaded speech models and returns a pointer
   to a model with the specified name if it exists in the list.  If the
   model is not loaded, findModel returns a NULL pointer. */
SpeechModel* findModel (char* model) 
{
  SpeechModel* pModel;
  
  for (pModel = pModels; pModel; pModel = pModel->sm_next) {
    if (!strcmp (pModel->sm_model, model)) {
      return (pModel);
    } /* endif */
  } /* endfor */

  return (NULL);
}

/* addModel adds a new model to the list of loaded speech model and
   sets its reference count to 1.  It returns the new model. */
SpeechModel* addModel (char* model)
{
  SpeechModel* pModel;

  ALLOC (pModel, SpeechModel);
  pModel->sm_model = strdup (model);
  pModel->sm_refcount = 1;
  pModel->sm_next = pModels;
  pModels = pModel;

  return (pModel);
}

/* deleteModel deletes a model from the list of speech models.  It returns
   zero if successful, non-zero otherwise. */
int deleteModel (char* model)
{
  SpeechModel* pModel;
  SpeechModel* pLastModel = 0;

  for (pModel = pModels; pModel; pModel = pModel->sm_next) {
    if (!strcmp(pModel->sm_model, model)) {
      if (pLastModel) {
	pLastModel->sm_next = pModel->sm_next;
      } else {
	pModels = pModel->sm_next;
      } /* endif */
      free (pModel->sm_model);
      FREE (pModel);
      return (0);
    } else {
      pLastModel = pModel;
    } /* endif */
  } /* endfor */

  /* Oops, should have found the model */
  return (-1);
}
  
/* readBytes reads the specified number of bytes from a socket.  It keeps
   reading from the socket until the number of bytes specified has been read
   or an error occurs.  It returns the number of bytes read on success, or
   the value returned by read on an error */
int readBytes (int socket, char* buffer, int bytesToRead)
{
  int bytesRead = 0;
  int len;

  do {
    len = read (socket, &buffer[bytesRead], bytesToRead - bytesRead);
    if (len > 0) {
      bytesRead += len;

#ifdef ODYDEBUG
      printf ("Read %d bytes on descriptor %d\n", len, socket);
#endif

    } else if (len == 0) {
      bytesRead = 0;
    } else {
      perror ("Error reading from socket");
      bytesRead = -1;
    } /* endif */
  } while (bytesRead > 0 && bytesRead < bytesToRead);

  return (bytesRead);
}

/* speech_haltServer() is called to shut down the server.  It sets a
   global flag which causes the child process to close its pipes and
   exit.  Since the Janus server is listening on a pipe, it will note
   the closing of the pipe and shut iself down. */
long speech_haltServer (RPC2_Handle cid)
{
  /* Global lock not required */
  RPCClientData.cont = 0;
  return (RPC2_SUCCESS);
}

/* speech_pingServer() is called when active estimation of bandwidth is 
   necessary.  It transfers the specified number of bytes to the speech 
   warden, hoping to update the bandwidth estimation. */
long speech_pingServer (RPC2_Handle cid, RPC2_Integer size, SE_Descriptor* bp)
{
  char buffer[size]; /* Temporary buffer for ping data */
  SE_Descriptor sed;

  /* Global lock not required */

  /* Set up a descriptor to initiate a ping transfer */
  bzero(&sed, sizeof(sed));
  sed.Tag = SMARTFTP;
  sed.Value.SmartFTPD.Tag = FILEINVM;
  sed.Value.SmartFTPD.ByteQuota = -1;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buffer;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.MaxSeqLen = size;
  sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;

#ifdef ODYDEBUG
  printf ("Ping initiated\n");
#endif

  /* Transfer the data */
  if (RPC2_InitSideEffect(cid, &sed) != RPC2_SUCCESS) {
    return (-1);
  } /* endif */
  if (RPC2_CheckSideEffect(cid, &sed, SE_AWAITLOCALSTATUS) != RPC2_SUCCESS) {
    return (-1);
  } /* endif */

#ifdef ODYDEBUG
  printf ("Ping completed\n");
#endif

  return (RPC2_SUCCESS);
}

/* callJanus() makes a generic call to Janus.  It returns zero if the
   call was successfully made, and non-zero otherwise. */
int callJanus (InputHeader* pInputHeader, void* pInputData, 
	       OutputHeader* pOutputHeader, void** ppOutputData)
{
  int fds; 
  int rc = 0;

  *ppOutputData = 0;

  /* Write the header to the pipe */
  if (write (RPCClientData.outPipe, pInputHeader, sizeof(InputHeader)) 
      == sizeof(InputHeader)) {
    
    /* Write input data to the pipe */
    if (write (RPCClientData.outPipe, pInputData, 
	       pInputHeader->inputDataLength) 
	== pInputHeader->inputDataLength) {

      /* Wait for a response from the Janus process - this will allow RPC pings
	 to be handled in the meantime */
      ASSERT (RPCClientData.inPipe < 32);
      fds = 0;
      fds |= 1 << RPCClientData.inPipe; 
      if (IOMGR_Select (RPCClientData.inPipe, &fds, NULL, NULL, NULL) >= 0) {
	
	/* Since we only waited on input pipe, we must have some data ready */
	if (readBytes (RPCClientData.inPipe, (char *) pOutputHeader, 
		       sizeof(OutputHeader))
	    == sizeof(OutputHeader)) {

	  /* Allocate space for the output data and read it */
	  if (pOutputHeader->outputDataLength > 0) {

	    NALLOC (*ppOutputData, char, pOutputHeader->outputDataLength);
	    if (readBytes (RPCClientData.inPipe, *ppOutputData, 
			   pOutputHeader->outputDataLength) 
		== pOutputHeader->outputDataLength) {

	      /* Everything is OK */
	      rc  = 0;

	    } else {
	      printf ("Unable to read output data\n");
	      rc = -101;
	    } /* endif */

	  } /* endif */

	} else {
	  printf ("Unable to read header\n");
	  rc = -102;
	} /* endif */

      } else {
	printf ("Select failed\n");
	rc = -103;
      } /* endif */

    } else {
      printf ("Unable to write temporary filename\n");
      rc = -104;
    } /* endif */

  } else {
    printf ("Unable to write input header\n");
    rc = -105;
  } /* endif */

  return (rc);
}

/* speech_initModel initializes a speech model for a client.  If the 
   model is already loaded, it incremenents a reference count for the model.
   If the model has not been loaded, it loads the model and sets the 
   reference count to one.  The model parameters are returned, along
   with a return code. */
long speech_initModel (RPC2_Handle cid, RPC2_String model, RPC2_Integer* pRC, 
		       ModelParms* pModelParms)
{ 
  SpeechModel* pModel;
  InputHeader inputHeader;
  OutputHeader outputHeader;
  char* pOutputData;
  char* pch[5];
  char** ap;
  char* val;
  char* p;
  int results;

  GLOCK;

  pModel = findModel (model);
  if (pModel) {

    /* Model already loaded - increment its reference count */
    pModel->sm_refcount++;
#ifdef ODYDEBUG
    printf ("Reference count for model %s incremented to %d\n", 
	    pModel->sm_model, pModel->sm_refcount);
#endif

  } else {

    /* Try to load the specified model */
    inputHeader.inputType = 'L';
    strcpy (inputHeader.model, model);
    inputHeader.inputDataLength = 0;

    *pRC = callJanus (&inputHeader, 0, &outputHeader, (void **) &pOutputData);
    if (*pRC) {
      GUNLOCK;
      return (*pRC);
    }

    /* Check Janus return code */
    *pRC = outputHeader.rc;
    if (*pRC) {
      GUNLOCK;
      return (*pRC);
    }
	
    /* Check output data */
    if (!pOutputData) {
      *pRC = -110;
      GUNLOCK;
      return (*pRC);
    } /* endif */

#ifdef ODYDEBUG
    printf ("Output data is: %s\n", pOutputData);
#endif

    /* Create a new model */
    pModel = addModel (model);
    
    /* Save model parameters */
    results = 0;
    ap = pch;
    for (p = pOutputData; p != NULL; ) {
      while ((val = strsep(&p, "\n")) != NULL && *val == '\0');
      if (val) {
	*ap++ = val;
	results++;
      } /* endif */
    } /* endfor */

    if (results != 5) {
      printf ("Output data had only %d parameters\n", results);
      *pRC = -109;
      GUNLOCK;
      return (*pRC);
    } /* endif */

    pModel->sm_fname = strdup (pch[0]);
    pModel->sm_fidelity = strdup (pch[1]);
    pModel->sm_recognizePerByteCost = strdup (pch[2]);
    pModel->sm_VQPerByteCost = strdup (pch[3]);
    pModel->sm_recNoVQPerByteCost = strdup (pch[4]);
    FREE (pOutputData);
    
  } /* endif */

  /* Return the model parameters */
  pModelParms->fidelityBBS.SeqLen = strlen (pModel->sm_fidelity) + 1;
  if (pModelParms->fidelityBBS.SeqLen <= pModelParms->fidelityBBS.MaxSeqLen) {
    strcpy (pModelParms->fidelityBBS.SeqBody, pModel->sm_fidelity);
  } else {  
    printf ("Fidelity BBS of getModel RPC is too small\n");
    *pRC = -100;
    GUNLOCK;
    return (*pRC);
  } /* endif */
  
  pModelParms->recBBS.SeqLen = strlen (pModel->sm_recognizePerByteCost) + 1;
  if (pModelParms->recBBS.SeqLen <= pModelParms->recBBS.MaxSeqLen) {
    strcpy (pModelParms->recBBS.SeqBody, pModel->sm_recognizePerByteCost);
  } else {  
    printf ("Recognition time BBS of getModel RPC is too small\n");
    *pRC = -100;
    GUNLOCK;
    return (*pRC);
  } /* endif */
  
  pModelParms->VQBBS.SeqLen = strlen (pModel->sm_VQPerByteCost) + 1;
  if (pModelParms->VQBBS.SeqLen <= pModelParms->VQBBS.MaxSeqLen) {
    strcpy (pModelParms->VQBBS.SeqBody, pModel->sm_VQPerByteCost);
  } else {  
    printf ("VQ time BBS of getModel RPC is too small\n");
    *pRC = -100;
    GUNLOCK;
    return (*pRC);
  } /* endif */
  
  pModelParms->recNoVQBBS.SeqLen = strlen (pModel->sm_recNoVQPerByteCost)+1;
  if (pModelParms->recNoVQBBS.SeqLen <= pModelParms->recNoVQBBS.MaxSeqLen) {
    strcpy (pModelParms->recNoVQBBS.SeqBody, pModel->sm_recNoVQPerByteCost);
  } else {  
    printf ("Recogition (w/o VQ) time BBS of getModel RPC is too small\n");
    *pRC = -100;
    GUNLOCK;
    return (*pRC);
  } /* endif */

  *pRC = 0;
  GUNLOCK;
  return (RPC2_SUCCESS);
}

/* speech_closeModel closes a speech model for a client.  The reference
   count for the model is decremented.  If the reference count reaches 
   zero, the model is unloaded. */
long speech_closeModel (RPC2_Handle cid, RPC2_String model, RPC2_Integer* pRC)
{ 
  SpeechModel* pModel;
  InputHeader inputHeader;
  OutputHeader outputHeader;
  char* pOutputData;

  GLOCK;

  pModel = findModel (model);
  if (pModel) {

    /* Model already loaded - increment its reference count */
#ifdef BSD44
    /* Unloading a model in Linux coredumps Janus - so don't do it. */
    pModel->sm_refcount--;
#endif

#ifdef ODYDEBUG
    printf ("Reference count for model %s decremented to %d\n", 
	    pModel->sm_model, pModel->sm_refcount);
#endif

    if (pModel->sm_refcount <= 0) {

      /* delete the model */
      deleteModel (model);

      /* unload the model in Janus */
      inputHeader.inputType = 'U';
      strcpy (inputHeader.model, model);
      inputHeader.inputDataLength = 0;

      *pRC = callJanus (&inputHeader, 0, &outputHeader, 
			(void **) &pOutputData);
      if (*pRC) {
	GUNLOCK;
	return (*pRC);
      }

      /* Check Janus return code */
      *pRC = outputHeader.rc;
      if (*pRC) {
	GUNLOCK;
	return (*pRC);
      }
	
    } /* endif */

  } else {
    *pRC = -1; /* Unable to find the specified model */
    GUNLOCK;
    return (*pRC);
  } /* endif */

  *pRC = 0;
  GUNLOCK;
  return (RPC2_SUCCESS);
}

/* speech_getTranslation() is a remote procedure invoked by the warden to
   translate speech data.  It SFTPs data to a temporary file, then sends a
   translation request to the Janus server on its output pipe.  It selects on
   the output pipe for a response from the Janus server, allowing RPC2 to 
   keep the connection alive in the meantime.  When a response is received on
   its input pipe, it returns the output data to the warden. */
long speech_getTranslation(RPC2_Handle cid, 
			   RecognitionParms* pRecognitionParms,
			   RPC2_Integer *janusRC, 
			   RPC2_BoundedBS *pTranslationBBS, SE_Descriptor *bp,
			   RPC2_Integer *start_sec, RPC2_Integer* start_usec,
			   RPC2_Integer *end_sec, RPC2_Integer* end_usec)
{
  SE_Descriptor sed;
  InputHeader inputHeader;
  OutputHeader outputHeader;
  SpeechModel* pModel;
  char* outputData = 0;
  struct timeval tv;
  char temp_filename[256];

  /* We create a temporary file whose name is unique to this thread */
  sprintf(temp_filename,"%s.%s",TEMP_FILENAME, ds_thread_name());

  /* Set up descriptor to put data in temporary local file */
  bzero(&sed, sizeof(sed));
  sed.Tag = SMARTFTP;
  sed.Value.SmartFTPD.Tag = FILEBYNAME;
  sed.Value.SmartFTPD.ByteQuota = -1;
  strcpy (sed.Value.SmartFTPD.FileInfo.ByName.LocalFileName, temp_filename);
  sed.Value.SmartFTPD.FileInfo.ByName.ProtectionBits = 0777;
  sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;

  /* Transfer the data */
  if (RPC2_InitSideEffect(cid, &sed) != RPC2_SUCCESS) {
    return (-1);
  } /* endif */
  if (RPC2_CheckSideEffect(cid, &sed, SE_AWAITLOCALSTATUS) != RPC2_SUCCESS) {
    return (-1);
  } /* endif */

  /* File has been transfered - this marks the beginning of processing */
  GLOCK;

  gettimeofday (&tv, NULL);
  *start_sec = tv.tv_sec;
  *start_usec = tv.tv_usec;

  /* Format an input header */
  inputHeader.inputType = pRecognitionParms->inputType;
  inputHeader.outputType = pRecognitionParms->outputType;
  /* inputHeader.searchEffortPercent = pRecognitionParms->searchEffortPercent; */
  strcpy (inputHeader.model, pRecognitionParms->model);

  /* Only pass on fidelity if both models were loaded */
  pModel = findModel(pRecognitionParms->model);
  if (pModel && pModel->sm_fname && !strcmp(pModel->sm_fname,"Both")) {
    strcpy (inputHeader.fidelity, pRecognitionParms->fidelity);
  } else {
    strcpy (inputHeader.fidelity , "");
  }

  inputHeader.inputDataLength = strlen (temp_filename) + 1;

  /* Write the header to the pipe */
  *janusRC = callJanus (&inputHeader, temp_filename, &outputHeader, 
			(void **) &outputData);
  if (*janusRC) {
    GUNLOCK;
    return (*janusRC);
  }

  *janusRC = outputHeader.rc;
  if (*janusRC) {
    GUNLOCK;
    return (*janusRC);
  }

  pTranslationBBS->SeqLen = strlen (outputData) + 1;
  if (pTranslationBBS->SeqLen <= pTranslationBBS->MaxSeqLen) {
    strcpy (pTranslationBBS->SeqBody, outputData);
  } else {
    printf ("Transaltion buffer size: %d\n", (int) pTranslationBBS->MaxSeqLen);
    printf ("Translation size: %d\n", (int) pTranslationBBS->SeqLen);
    printf ("Translation buffer is too small\n");
    pTranslationBBS->SeqLen = 0;
    *janusRC = -100;
  } /* endif */

  FREE (outputData);

  /* About to return - this marks the end of processing */
  gettimeofday (&tv, NULL);
  *end_sec = tv.tv_sec;
  *end_usec = tv.tv_usec;

  GUNLOCK;
  return (RPC2_SUCCESS);
}

/* I have no idea what this does - it has to do with RPC2 */
void iopen () {}

/* error_report is used to report RPC2 error messages */
void error_report(char *message)
{
  fprintf(stderr, message);
  fprintf(stderr, "\n");
  exit(1);
}

/* Init_RPC() initializes the RPC2 server */
void Init_RPC (int port)
{
  RPC2_PortIdent pid;
  RPC2_SubsysIdent sid;
  SFTP_Initializer sftpi;
  int rc;
  char error_msg[100];

  SFTP_SetDefaults(&sftpi);
  sftpi.EnforceQuota = 0;
  SFTP_Activate(&sftpi);

  pid.Tag = RPC2_PORTBYINETNUMBER;
  pid.Value.InetPortNumber = htons(port);
  rc = RPC2_Init(RPC2_VERSION, NULL, &pid, -1, NULL);
  
  if (rc != RPC2_SUCCESS) {
    sprintf(error_msg, "%s\nCan't initialize RPC2", 
	    RPC2_ErrorMsg(rc));
    error_report(error_msg);
  } /* endif */

  sid.Tag = RPC2_SUBSYSBYID;
  sid.Value.SubsysId = SPEECH_SUBSYSID;
  rc = RPC2_Export(&sid);
  if (rc != RPC2_SUCCESS) {
    sprintf(error_msg, "%s\nCan't export the speech subsystem", 
	    RPC2_ErrorMsg(rc));
    error_report(error_msg);
  } /* endif */
  
  printf ("Initialized port %d\n", port);
}

/* serviceRPCs() is invoked by the child process to service RPC requests made
   by the speech warden.  It continues to service requests until explicitly 
   halted by a speech_haltServer RPC. */
static int
serviceRPCs(argp)
     char *argp;
{
  RPC2_Handle the_connection;
  RPC2_RequestFilter reqfilter;
  RPC2_PacketBuffer *reqbuffer;
  int rc;

  reqfilter.FromWhom = ANY;
  reqfilter.OldOrNew = OLDORNEW;
  reqfilter.ConnOrSubsys.SubsysId = SPEECH_SUBSYSID;

  while (RPCClientData.cont) {
    rc = RPC2_GetRequest(&reqfilter, &the_connection, &reqbuffer,
			 NULL, NULL, 0, NULL);
    if (rc != RPC2_SUCCESS) {
      fprintf(stderr, RPC2_ErrorMsg(rc));
    } /* endif */
    rc = ody_speech_ExecuteRequest(the_connection, reqbuffer, NULL);
    if (rc != RPC2_SUCCESS) {
      fprintf(stderr, RPC2_ErrorMsg(rc));
    } /* endif */
  } /* end while */

  /* We're done, wake up the main thread so it can shut down */
  ds_thread_wakeup((char *)&RPCClientData);

  return 0;
}

int main (int argc, char* argv[]) 
{
  int inPipe, outPipe, port;
  int i;

  if (argc != 4) {
    printf ("Format: speechsrv [inPipe] [outPipe] [port]\n");
    printf ("This really shouldn't be run from the command line\n");
    exit (1);
  } /* endif */
  
  inPipe = atoi (argv[1]);
  outPipe = atoi (argv[2]);
  port = atoi (argv[3]);

#ifdef ODYDEBUG
  printf ("inPipe = %d\n", inPipe);
  printf ("outPipe = %d\n", outPipe);
  printf ("port = %d\n", port);
#endif

  /* Set up global RPC data */
  RPCClientData.cont = 1;
  RPCClientData.inPipe = inPipe;
  RPCClientData.outPipe = outPipe;

  ds_thread_init(32768, 0);
  Init_RPC(port);
  
  /* Initialize the global lock and create the worker threads */
  ds_hwlock_init(&speechsrv_global_lock);
  for (i = 0; i < num_workers; i++)
      ASSERT(ds_thread_create(serviceRPCs, (char *)i, DS_THREAD_NORMAL,
			      "speechsrv", 0));

  /* We sleep until someone halts the server, at which point we exit. */
  while (RPCClientData.cont)
      ds_thread_wait((char *)&RPCClientData);

  /* We grab the lock; this ensures that no speech stuff will be happening
     when we shutdown. Some thread could be doing data xfer to/from a client,
     and it will be rudely killed, too bad */
  GLOCK;

#ifdef ODYDEBUG
  printf ("Odyssey server is halted\n");
#endif

  close (inPipe);
  close (outPipe);

  return (0);
}

