#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * ody_webserver_html.c
 *
 * code for distilling html
 *
 * TODO: this should all really have some kind of tokenizer, as opposed
 *       to the ad-hoc garbage I'm currently using.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <base/ds/ds_safeq.h>
#include <string.h>
#include "base/simplewww/swww.h"
#include <lwp/lwp.h>
#include <lwp/lock.h>
#include <rpc2/rpc2.h>
#include "external/jpeg-6a/jpeglib.h"
#include "distiller_main.h"
#include "distiller_html.h"

/* get_bytes_to_char gets a chunk of bytes from the ds_safeq_t, and
   passes along the requested number of bytes to its caller. if it can
   fulfill that request from the existing buffer, more power to it.

   it has the interesting semantics that it continues to fill a buffer
   up until it reaches the char 'c', and then returns that buffer.
   the buffer does *not* include 'c'. if 'disposable' is true,
   NULL is returned in lieu of a buffer.

   'size' is always returned. if a buffer is returned, 'size' reflects
   the size of that buffer. if no buffer is returned (because
   'disposable' is set to DISPOSE), 'size' is the number of bytes
   which were discarded.

   if the size returned == -1, there are no more bytes available in
   the stream.

   the returned buffer must be freed by the calling function. */
char *get_bytes_to_char(char                c,
			struct html_buffer *hbuf,
			enum html_dispose   disposable,
			int                *size)
{
    /* XXX main thing in here which needs fixed is that I add 1 to
           the size in some places for the null char, and then
	   I subtract one from it later to get rid of it, because
	   I realized I didn't need it. I need to recompute all of
	   the off-by-one thingies so that I don't have to play this
	   game. */
    struct data_chunk *dcp           = NULL;
    char              *bookmark      = NULL;
    char              *return_buffer = NULL;
    int                temp_size     = 0;

    /* fill the buffer, if need be */
    if (hbuf->remaining == 0 || hbuf->size == 0) {
	if (hbuf->buffer != NULL)
	    free(hbuf->buffer);
	dcp = ds_safeq_deq(hbuf->queue);

	if (dcp->type == END) {
	    /* that's all, folks */
	    ds_safeq_destroy(hbuf->queue);
	    FREE_CHUNK(dcp);
	    *size = -1;
	    return(return_buffer);
	}

	hbuf->buffer    = dcp->data;
	hbuf->size      = dcp->size;
	hbuf->index     = hbuf->buffer;
	hbuf->remaining = hbuf->size;
    }

    /* find the character in the buffer, if possible. if found, copy
       up to that point into a new buffer. if not found, copy the
       buffer over, and get another chunk (and append if needed) */
    bookmark  = index(hbuf->index, c);
    if (bookmark == NULL)
	bookmark = index(hbuf->index, '\0');
    temp_size = bookmark - hbuf->index;

    if (disposable == KEEP) {
	return_buffer = (char *) malloc(temp_size + 1);
	memcpy(return_buffer, hbuf->index, temp_size);
    }

    hbuf->remaining -= temp_size;
    hbuf->index     += temp_size;

    if (disposable == KEEP) {
	return_buffer[temp_size] = '\0';
    }

    *size = temp_size + 1;

    while (*bookmark == '\0') {
	/* we didn't actually get the whole thing; we ran out
	   of data in our hbuf->buffer. get another chunk for the
	   hbuf->buffer and try again. */

	/* fill the buffer */
	FREE_CHUNK(dcp);
	dcp = ds_safeq_deq(hbuf->queue);

	if (dcp->type == END) {
	    /* that's all, folks */
	    ds_safeq_destroy(hbuf->queue);
	    FREE_CHUNK(dcp);
	    *size = -1;
	    return(return_buffer);
	}

	hbuf->buffer    = dcp->data;
	hbuf->size      = dcp->size;
	hbuf->index     = hbuf->buffer;
	hbuf->remaining = hbuf->size;

	/* find the character in the buffer, and append to the
	   current return_buffer */
	bookmark  = index(hbuf->index, c);
	if (bookmark == NULL)
	    bookmark = index(hbuf->index, '\0');
	temp_size = bookmark - hbuf->index;
	
	if (disposable == KEEP) {
	    return_buffer = (char *) realloc(return_buffer,
					     temp_size + *size);
	    memcpy(return_buffer + *size - 1, hbuf->index, temp_size);
	}

	hbuf->remaining -= temp_size;
	hbuf->index     += temp_size;
	
	if (disposable == KEEP) {
	    return_buffer[temp_size + *size - 1] = '\0';
	}
    }

    *size -= 1;
    return(return_buffer);
}

void html_main(ds_safeq_t *readq, int reqid) {
    struct html_buffer  hbuf;
    char               *buffer;
    int                 size;
    int                 done;
    enum html_state     state;
    enum html_state     previous;

    INIT_HTML_BUFFER(&hbuf, readq);

    state    = NONE;
    previous = NONE;
    done     = 0;

    /* read data from pipe, and strip out appropriate data. */
    while(!done) {
	switch(state) {
	case NONE:
	    /* can go to IN_BEGIN_TAG */
	    get_bytes_to_char('<', &hbuf, DISPOSE, &size);
	    if (size == -1)  {
		state = DONE;
	    } else
		state = IN_BEGIN_TAG;

	    break;
	    
	case IN_BEGIN_TAG:
	    /* can go to NONE, IN_ANCHOR, IN_HEADER */
	    buffer = get_bytes_to_char('>', &hbuf, KEEP, &size);
	    if (size == -1) {
		state = DONE;
		break;
	    }

	    /* check what kind of tag this is */
	    if (strncasecmp("<IMG", buffer, 4) == 0) {
		if (isspace(buffer[4])) {
		    state = NONE; /* IMG has no end tag */
		    SEND_ALONG("%s>\n", buffer);
		}
	    } else if (strncasecmp("<A", buffer, 2) == 0) {
		if (isspace(buffer[2])) {
		    state = IN_ANCHOR;
		    SEND_ALONG("<P> %s>", buffer);
		}
	    } else if (strncasecmp("<H", buffer, 2) == 0) {
		if (isdigit(buffer[2])) {
		    state = IN_HEADER;
		    SEND_ALONG("%s>", buffer);
		}
	    } else {
		state = NONE;
	    }

	    free(buffer);

	    break;
	    
	case IN_ANCHOR:
	    /* can go to IN_END_TAG */
	    buffer = get_bytes_to_char('<', &hbuf, KEEP, &size);
	    if (size == -1) {
		state = DONE;
		break;
	    }

	    if (size > 1)
		SEND_ALONG("%s", buffer + 1);

	    free(buffer);
	    
	    previous = IN_ANCHOR;
	    state    = IN_END_TAG;

	    break;
	    
	case IN_HEADER:
	    /* can go to IN_END_TAG */
	    buffer = get_bytes_to_char('<', &hbuf, KEEP, &size);
	    if (size == -1) {
		state = DONE;
		break;
	    }

	    if (size)
		SEND_ALONG("%s", buffer + 1);
	    
	    free(buffer);
	    
	    previous = IN_HEADER;
	    state    = IN_END_TAG;

	    break;
	    
	case IN_END_TAG:
	    /* can go to NONE, IN_ANCHOR, IN_HEADER */
	    buffer = get_bytes_to_char('>', &hbuf, KEEP, &size);
	    if (size == -1) {
		state = DONE;
		break;
	    }

	    state = UNKNOWN;

	    if (strncmp(buffer, "</", 2) == 0) {
		if (previous == IN_ANCHOR &&
			   (strncasecmp(buffer, "</A", 3) == 0)) {
		    if (isspace(buffer[3]) || buffer[3] == '\0') {
			state = NONE;
			SEND_ALONG("%s>\n", buffer);
		    }
		} else if (previous == IN_HEADER &&
			   (strncasecmp(buffer, "</H", 3) == 0)) {
		    if (isdigit(buffer[3]) || buffer[3] == '\0') {
			state = NONE;
			SEND_ALONG("%s>\n", buffer);
		    }
		}
	    }

	    if (state == UNKNOWN) {
		state = previous;
		SEND_ALONG("%s>", buffer);
	    }

	    free (buffer);
	    previous = NONE;

	    break;
	    
	case DONE:
	    /* do any remaining clean up and get out */
	    done = 1;
	    break;
	    
	default:
	    fprintf(stderr, "broken state machine in html_main!\n");
	    exit(-1);
	}
    }
}
