#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * ody_webserver.c -- streaming web server for odyssey
 *
 * QUESTIONS
 *   * perhaps "look up req. ID" will lock req. ID, so we don't have to
 *     worry about buffer manipulation? (gotta be careful, since accept_data
 *     wakes up process) [is this still a valid question? i need to
 *     re-check this. -et Feb 13 1997]
 * 
 */

/* system/C headers */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <signal.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <assert.h>

/* ds headers */
#include <base/ds/ds_safeq.h>

/* RPC2 headers */
#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <lwp/lock.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

/* JPEG headers */
#include "distiller_cdjpeg.h"
#include "external/jpeg-6a/jpeglib.h"

/* HTML headers */
#include "distiller_html.h"

/* my headers */
#include "base/simplewww/swww.h"
#include "distiller.h"
#include "distiller_main.h"

/* globals */

struct request request_table[TABLE_SIZE];
struct Lock    table_lock;
struct alias_entry *alias_list;
int            debugging = 0;
int            aliasing = 0;
int            limiting = 0;
ds_safeq_t    *fetch_q;
ds_safeq_t    *compress_q;

/* setup and support */

int main(int argc, char **argv)
{
    /* Here we set up libwww and RPC2, and enter the main event loop.
       We don't exit unless there is some fundamental error. */
    RPC2_Handle cid;
    RPC2_RequestFilter reqfilter;
    RPC2_PacketBuffer *reqbuffer;
    PROCESS pid;
    int rc;
    int loop, loop2;
    char *alias_fname = NULL;
    char *exec_fname;

    cont_sw_threshold.tv_sec = 0;
    cont_sw_threshold.tv_usec = 500000;
    
    exec_fname = argv[0];
    while (argc > 1) {
	argv++; argc--;
	if (strcmp("-debug", argv[0]) == 0) {
	    debugging = 1;
	} else if (strcmp("-f", argv[0]) == 0) {
	    alias_fname = argv[1];
	    aliasing = 1;
	    argv++; argc--;
	} else if (strcmp("-l", argv[0]) == 0) {
	    /* avoid accessing non-aliased hosts */
	    limiting = 1;
	} else {
	    fprintf(stderr, "usage: %s [-debug] [-f alias filename] [-l]\n",
		    exec_fname);
	    exit(-1);
	}
    }
    SWWW_debugging(debugging);
    if (aliasing) {
	process_alias_file(alias_fname);
    }
    
    /* XXX we have to ignore sigpipe for swww to work properly */
    signal(SIGPIPE, SIG_IGN);
    
    if(SWWW_version(SWWW_VERSION)) {
	/* if library version doesn't match header version, abort */
	exit(1);
    }

    /* Initialize RPC2 */
    Init_RPC();

    /* Initialize the request table */
    Lock_Init(&table_lock);
    for (loop = 0; loop < TABLE_SIZE; loop++) {
	request_table[loop].state = UNUSED;
	Lock_Init(&(request_table[loop].lock));
	for (loop2 = 0; loop2 < 2; loop2++) {
	    request_table[loop].buffer[loop2]   = NULL;
	    request_table[loop].size[loop2]     = 0;
	    request_table[loop].capacity[loop2] = 0;
	    request_table[loop].bufstate[loop2] = READY;
	}
    }

    reqfilter.FromWhom = ANY;
    reqfilter.OldOrNew = OLDORNEW;
    reqfilter.ConnOrSubsys.SubsysId = ODYWEBSUBSYSID;

    fetch_q = ds_safeq_create();
    compress_q = ds_safeq_create();
    assert(fetch_q && compress_q);

    for (loop=0; loop<NUM_FETCH_THREADS; loop++) {
	LWP_CreateProcess(thread_shell, 32768, LWP_NORMAL_PRIORITY,
			  (char *) NULL, "Opening Task", &pid);
    }
    
    for (loop=0; loop<NUM_COMPRESS_THREADS; loop++) {
	LWP_CreateProcess(compression_shell, 65536,
			  LWP_NORMAL_PRIORITY - 1,
			  (char *) NULL, "Compression Task", &pid);
    }

    /* Enter main event loop */
    while(1) {			/* wait for client calls to service */
	rc = RPC2_GetRequest(&reqfilter, &cid, &reqbuffer,
			     NULL, NULL, 0, NULL);
	/* extra NULL required for SE descriptor */
	if (rc != RPC2_SUCCESS)
	    fprintf(stderr, RPC2_ErrorMsg(rc));
	rc = odyssey_web_ExecuteRequest(cid, reqbuffer, NULL);
	if(rc != RPC2_SUCCESS)
	    fprintf(stderr, RPC2_ErrorMsg(rc));
    }
    
    return 0;
}

void error_report(char *message)
{
    /* Report errors to standard error. 'message' is error message, and
       does not need to be null-terminated */
    
    fprintf(stderr, message);
    fprintf(stderr, "\n");
    exit(1);
}

void Init_RPC(void)
{
    /* Initializes the LWP and RPC2 packages. Must be invoked by main
       shell before using RPC2 */

    PROCESS mylpid;
    RPC2_PortIdent pid;
    RPC2_SubsysIdent sid;
    SFTP_Initializer sftpi;
    int rc;
    char msg[100];

    /* Initialize LWP Package */
    if (LWP_Init(LWP_VERSION, LWP_NORMAL_PRIORITY, &mylpid) != LWP_SUCCESS)
	error_report("Can't Initialize LWP");

    /* Initialize SFTP side effect package */
    SFTP_SetDefaults(&sftpi);
    SFTP_Activate(&sftpi);

    /* Initialize RPC2 Package */
    pid.Tag = RPC2_PORTBYINETNUMBER;
    pid.Value.InetPortNumber = htons(ODYWEBPORTAL);
    rc = RPC2_Init(RPC2_VERSION, NULL, &pid, -1, NULL);
    if (rc != RPC2_SUCCESS) {
	sprintf(msg, "%s\nCan't Initialize RPC2", RPC2_ErrorMsg(rc));
	error_report(msg);
    }

    sid.Tag = RPC2_SUBSYSBYID;
    sid.Value.SubsysId = ODYWEBSUBSYSID;
    rc = RPC2_Export(&sid) != RPC2_SUCCESS;
    if (rc != RPC2_SUCCESS) {
	sprintf(msg, "%s\nCan't export the ody_web subsystem",
		RPC2_ErrorMsg(rc));
	error_report(msg);
    }
}

void iopen(void) {}

/* RPC2 procedures */

long open_url(RPC2_Handle     _cid,
	      RPC2_Integer    method,      /* The method to retrieve the
					      URL with */
	                                   /* XXX this should be an
					      http_method_t, but
					      there's no good way to
					      turn the swww.h
					      enum into an RPC2 enum */
	      RPC2_String     url,         /* the URL identifying the
					      HTTP object */
	      RPC2_Integer    img_quality, /* minimum image object quality
					      requested */
	      RPC2_Integer    html_quality,/* minimum HTML object quality */
	                                   /* XXX eventually, should be
					      quality vector... */
	      RPC2_Integer    last_modified, /* 0 or modification date
						of cache object */
	      RPC2_Integer   *the_reqid,   /* ID of request */
	      RPC2_Integer   *status_code, /* HTTP status code of fetch */
              RPC2_BoundedBS *status,	   /* HTTP status string */
	      RPC2_Integer   *header_size, /* number of HTTP headers */
	      RPC2_BoundedBS *headers)     /* HTTP headers */
{
    int intercepted = 0;
    int new_reqid;

    /*
     * if aliasing, check for substitution on this url
     */
    if (aliasing) {
	char *scheme=NULL, *host=NULL, *path=NULL;
	int port;
	if (SWWW_parse_url(url, &scheme, &host, &port, &path) == 0 &&
	    host != NULL) {
	    struct alias_entry *alias = alias_list;

	    /* XXX should include port number in match for completeness */
	    while (alias) {
		if (strcmp(host, alias->request_url) == 0) {
		    /* found a match -- munge the url */
		    char *new_url = (char *)malloc(strlen("http://") +
						   strlen(alias->alias_url) +
						   strlen(alias->request_url) +
						   strlen(path) +
						   2 /* '/' */);
		    strcpy(new_url, "http://");
		    strcat(new_url, alias->alias_url);
		    strcat(new_url, "/");
		    strcat(new_url, alias->request_url);
		    if (limiting) {
			char *cgi_start = strstr(path, "cgi-bin/");
			char *q_mark;
			char *path_start = path;

/*
			if (cgi_start) {
			    strncat(new_url, path, cgi_start-path);
			    path_start = cgi_start+8;
			}
			*/
			q_mark = index(path_start, (int)'?');
			if (q_mark) {
			    strncat(new_url, path_start,
				    q_mark-path_start);
			    strcat(new_url, q_mark+1);
			} else
			    strcat(new_url, path_start);
			if (cgi_start)
			    strcat(new_url, ".html");
		    } else {
			strcat(new_url, path);
		    }
		    url = new_url; /* XXX free(url)? */
		    alias = NULL;
		    intercepted = 1;
		} else
		    alias = alias->next;
	    }
	    free(host);
	    if (scheme) free(scheme);
	    if (path) free(path);
	}
	if (limiting && !intercepted) {
	    /* this URL would force an access to the Net, which we are
	     * disallowing. */
	    return ODYWEB_NODOC;
	}
    }

    /*
     * create a request thread and get request ID
     */
    new_reqid = create_request(method, /* XXX if reqid < 0, handle error */
			       url,
			       img_quality,
			       html_quality,
			       last_modified);

    /* XXX check result code??  If it's -1, the WaitProcess will never
       be Signalled */
    
    DEBUG(fprintf(stderr, "opening request %d (img_qual %d)\n", new_reqid,
                          (int)img_quality));

    /*
     * wait for headers to be available
     */

    /* XXX this wouldn't work in a pre-emptive environment... */
    if (request_table[new_reqid].state != GOTHEADERS) {
	LWP_WaitProcess((char *) &(request_table[new_reqid].state));
    }
    if (request_table[new_reqid].state != GOTHEADERS) {
        fprintf(stderr, "woke up not in GOTHEADERS (%d)\n", 
		request_table[new_reqid].state);
	assert(0);
    }

    /*
     * return status, headers and req. ID to client
     */
    *the_reqid   = new_reqid;

    /* copy out status data */
    *status_code = request_table[new_reqid].status_code;

    if(strlen(request_table[new_reqid].status) > status->MaxSeqLen) {
	/* if it's too big, truncate it so it will fit,
	   and print an error message */
	fprintf(stderr, "fetch_url: "
		"Status line too long (%d chars);"
		" truncated to %ld chars.\n",
		strlen(request_table[new_reqid].status),
		       status->MaxSeqLen - 1);
	*((request_table[new_reqid].status)
	  + status->MaxSeqLen - 1)           = '\0';
    }
    
    /* NOTE: when using a BoundedBS to send a string, don't forget the \0! */
    status->SeqLen  = strlen(request_table[new_reqid].status) + 1;
    strcpy(status->SeqBody, request_table[new_reqid].status);

    /* pack up headers */
    *header_size = request_table[new_reqid].header_size;
    {
	SWWW_headers_t *hbookmark = request_table[new_reqid].headers;
	RPC2_ByteSeq   bookmark   = headers->SeqBody;
	int            size, temp;
	
	size = 0;
	
	if(hbookmark == NULL) {
	    /* there are no headers */
	    headers->SeqLen = 0;	/* send nothing back */
	} else {
	    while(hbookmark->next != NULL) {
		size += (temp = strlen(hbookmark->name) + 1);
		if (size < headers->MaxSeqLen) {
		    strcpy(bookmark, hbookmark->name);
		    bookmark += temp;
		} else {
		    fprintf(stderr, "fetch_url: headers (%d) are longer than "
			    "allocated space (%ld)\n",
			    size, headers->MaxSeqLen);
		    size = headers->MaxSeqLen;
		    goto errhandle;
		}
		
		size += (temp = strlen(hbookmark->value) + 1);
		if (size < headers->MaxSeqLen) {
		    strcpy(bookmark, hbookmark->value);
		    bookmark += temp;
		} else {
		    fprintf(stderr, "fetch_url: headers (%d) are longer than "
			    "allocated space (%ld)\n",
			    size, headers->MaxSeqLen);
		    size = headers->MaxSeqLen;
		    goto errhandle;
		}
		
		hbookmark = hbookmark->next;
	    }
      
	    headers->SeqLen = size;
	}
    }

    return RPC2_SUCCESS;
    
    /* errhandle routine; currently only invoked from within the
       header marshalling, because of boundedBS overflow */
 errhandle:
    return (-1);
}

long fetch_buffer(RPC2_Handle    _cid,
		  RPC2_Integer   reqid,  /* ID of request */
		  RPC2_Integer  *eoho,   /* 1 if this is the last buffer
					    (End Of HTTP Object),
					    0 otherwise */
		  SE_Descriptor *sp)     /* SFTP is used for sending
					    the buffer */
{
    /* N.B. if the assumptions about the buffer states (see the asserts
            below) are violated, it probably means that two different
            clients are using the same request ID. */
    int           bufid;
    int           other_bufid;
    SE_Descriptor buffer;
    int           rc;

    *eoho = 0;			/* innocent until proven guilty */

    if (request_table[reqid].state >= PRECLOSE) {
	/* all data has been sent; End of HTTP Object */
	*eoho = 1;
	return -10;		/* XXX need meaningful error */
    }

    /* check if existing buffer is big enough */
    bufid = request_table[reqid].bufinuse;

    if (!request_table[reqid].fetch_done &&
	request_table[reqid].size[bufid] < BUFFER_THRESH) {
	/* if not big enough, wait for that to happen */
	LWP_WaitProcess((char *)&(request_table[reqid]));
    }

    /* LOCK */
    ObtainWriteLock(&(request_table[reqid].lock));

    /* Set filled buffer to SENDING. */
    if (request_table[reqid].bufstate[bufid] == CURRENT) {
	request_table[reqid].bufstate[bufid] = SENDING;
    } else {
	/* there's bad trouble in ancient egypt, max */
	assert(0);
    }

    if (request_table[reqid].fetch_done) {
	/* if this is the last buffer, do the right stuff */
	*eoho = 1;
	request_table[reqid].state = PRECLOSE;
    } else {
	/* set READY buffer (the other buffer) to the CURRENT buffer, so
	   we can send the filled buffer via SFTP (and fill the other
	   buffer at the same time). */
	other_bufid = bufid + 1;
	other_bufid = other_bufid % 2;
	
	if (request_table[reqid].bufstate[other_bufid] == READY) {
	    /* change which buffer is the current one, and also wake
	       up accept_data if it's waiting for that buffer to become
	       current because of throttling */
	    request_table[reqid].bufstate[other_bufid] = CURRENT;
	    request_table[reqid].bufinuse = other_bufid;
	    LWP_SignalProcess((char *)
			      &(request_table[reqid].buffer[other_bufid]));
	} else {
	    /* there's bad trouble in outer space, sam */
	    assert(0);
	}
    }

    /* UNLOCK */
    ReleaseWriteLock(&(request_table[reqid].lock));


    /* send the buffer via SFTP */
    setupVMSE(&buffer,
	      request_table[reqid].buffer[bufid],
	      request_table[reqid].size[bufid]);

    if ((rc = RPC2_InitSideEffect(_cid, &buffer)) != RPC2_SUCCESS)
	goto errhandle;
    if ((rc = RPC2_CheckSideEffect(_cid, &buffer, SE_AWAITLOCALSTATUS))
	!= RPC2_SUCCESS)
	goto errhandle;

    /* set sent buffer to READY, set size back to 0 */
    request_table[reqid].size[bufid] = 0;
    request_table[reqid].bufstate[bufid] = READY;

    return RPC2_SUCCESS;

 errhandle:
    /* set sent buffer to READY, set size back to 0 */
    request_table[reqid].size[bufid] = 0;
    request_table[reqid].bufstate[bufid] = READY;

    close_url(_cid, reqid);
    RPC2_Unbind(_cid);

    return RPC2_FAIL;	/* XXX is this the right err code? */
}

/* calling this marks a request as closed, so that the server can
   clean up the resources associated with it. note that a client
   need not fetch all of the data associated with an HTTP object
   before calling close_url (allowing for aborting a connection, such
   as when the user presses the STOP button on a Web browser) */
long close_url(RPC2_Handle  _cid,
	       RPC2_Integer reqid)
{
    if (request_table[reqid].state == UNUSED)
	return -1;		/* XXX invalid connection; need error code */

    DEBUG(fprintf(stderr, "closing request %d\n", (int) reqid));

    ObtainWriteLock(&(request_table[reqid].lock));

    request_table[reqid].state = CLOSED;
    if (request_table[reqid].fetch_done == 1)
	cleanup_request(reqid);

    ReleaseWriteLock(&(request_table[reqid].lock));

    return RPC2_SUCCESS;
}

/* data pumps for SWWW */

int accept_headers(int             reqid,
		   int             status_code,
		   char           *status, 
		   int             header_size,
		   SWWW_headers_t *headers)
{
    char *type;
    struct compress_req *req;

    /*
     * decide whether to convert the headers
     */

    /*
     * XXX this currently uses the naive "slob" approach to adaptation.
           to wit, I know the minimum acceptable quality, so I'll provide
	   that. eventually, it'll be interesting to explore doing some
	   bandwidth measurements on the server, to try do impedance-matching
	   kinds of things. */
    
    if (request_table[reqid].img_quality < 100 ||
	request_table[reqid].html_quality < 100) {
	char *expected_size_str;
	int   expected_size;

	type              = SWWW_gethvalue(headers, "Content-type");
	expected_size_str = SWWW_gethvalue(headers, "Content-length");
	if (expected_size_str != NULL)
	    expected_size = atoi(expected_size_str);
	else
	    expected_size = 16384; /* XXX average size? */

	/* IMAGE/GIF */
	if ( request_table[reqid].img_quality < 100 &&
	     (type != NULL) &&
	     (expected_size >= 2048) &&
	     (strcasecmp("image/gif", type) == 0) ) {
	    /* we've satisfied the compression criteria; to wit, that
	       we want to compress (since the requested quality is < 100),
	       and that we know how to compress. In addition, the
	       image is large enough that compression is worthwhile. */

	    request_table[reqid].fordist = ds_safeq_create();
	    header_size += SWWW_clearhvalue(headers, "Content-length");
	    header_size += SWWW_sethvalue(headers,
					  "Content-type",
					  "image/jpeg");
	    request_table[reqid].compressing = TO_JPEG;

	    req = (struct compress_req *) malloc(sizeof(struct compress_req));
	    if (req == NULL) {
		/* we're out of memory, so die */
		assert(0);
	    }

	    req->reqid   = reqid;

	    ds_safeq_enq(compress_q, req);
/*
	    LWP_CreateProcess(compression_shell, 65536,
			      LWP_NORMAL_PRIORITY - 1,
			      (char *) reqid, "Compression Task", &pid);
			      */
	}

	/* TEXT/HTML */
	if ( request_table[reqid].html_quality < 100 && 
	     (type != NULL) &&
	     (expected_size >= 2048) &&
	     (strcasecmp("text/html", type) == 0) ) {
	    /* we've satisfied the compression criteria; to wit, that
	       we want to compress (since the requested quality is < 100),
	       and that we know how to compress.

	       I decided to leave in the "bigger than 2K" criteria
	       because, anecdotally, pages that rely solely on pages
	       for slick layout tend to be very small. Also, we already
	       know how to deal with images, so we may as well pay a
	       small extra price to not irritate the user. */

	    request_table[reqid].fordist = ds_safeq_create();
	    header_size += SWWW_clearhvalue(headers, "Content-length");
	    request_table[reqid].compressing = TO_HTML;
	    req = (struct compress_req *) malloc(sizeof(struct compress_req));
	    if (req == NULL) {
		/* we're out of memory, so die */
		assert(0);
	    }

	    req->reqid   = reqid;

	    ds_safeq_enq(compress_q, req);
/*
	    LWP_CreateProcess(compression_shell, 65536,
			      LWP_NORMAL_PRIORITY - 1,
			      (char *) reqid, "Compression Task", &pid);
			      */
	}

	if (type != NULL)
	    free(type);
	if (expected_size_str != NULL)
	    free(expected_size_str);
    }

    /*
     * add metadata to request_table
     */

    request_table[reqid].status_code = status_code;
    request_table[reqid].status      = strdup(status);
    request_table[reqid].header_size = header_size;
    request_table[reqid].headers     = SWWW_duplicate_headerlist(header_size,
								 headers);

    /* 
     * wake up open_url
     */
    /* XXX in a pre-emptive world, this would need some add'l locking;
           this whole section should probably write-lock the req. table
	   entry, and the open_url function should have a
	   lock/test/yield/unlock loop */
    request_table[reqid].state = GOTHEADERS;
    LWP_SignalProcess((char *)&(request_table[reqid].state));

    return(0);
}

int accept_web_data(int         reqid,
		    const void *data,
		    int         size)
{
    /* turn an SWWW accept_data call into an invocation of my
       accept_data (to disambiguate it from calls by the JPEG library) */
    return(accept_data(reqid, data, size, FROM_WEB));
}

int accept_jpeg_data(int         reqid,
		     const void *data,
		     int         size)
{
    DEBUG(fprintf(stderr, "accept_jpeg_data accepting %d bytes "
	                  "for request %d\n", size, reqid));
    return(accept_data(reqid, data, size, FROM_JPEG));
}

int accept_html_data(int         reqid,
		     const void *data,
		     int         size)
{
    DEBUG(fprintf(stderr, "accept_html_data accepting %d bytes "
	                  "for request %d\n", size, reqid));
    return(accept_data(reqid, data, size, FROM_HTML));
}

int accept_data(int            reqid,
		const void    *data,
		int            size,
		enum datatype  type)
{
    int   buffer;
    int   new_size;
    char *bp;
    int   done;

    /* if the connection is no longer accepting data, indicate this */
    if (request_table[reqid].state >= CLOSED) {
	errno = EPIPE;
	return -1;
    }

    if (type == FROM_WEB && request_table[reqid].compressing) {
	/* we're compressing, and this is raw data, so send it to the pipe */
	int                rc;
	struct data_chunk *chunk;

	CREATE_CHUNK(chunk, data, size);
	rc = ds_safeq_enq(request_table[reqid].fordist, chunk);

	return (rc ? -1 : size);
    }

    do {
	done = 1;		/* don't repeat unless necessary */

	/* LOCK */
	ObtainWriteLock(&(request_table[reqid].lock));
	
	buffer = request_table[reqid].bufinuse;
	
	/* place (converted, if apropo) data into buffer */
	/* check if new data + current data in buffer
	   will exceed capacity if buffer. if so, increase size of
	   buffer as needed. */
	new_size = size + request_table[reqid].size[buffer];
	if (new_size > request_table[reqid].capacity[buffer]) {
	    char *new_buffer;
	    int   requested_size = new_size + 1024;
	    
	    /* if the requested size is larger than BUFFER_MAX, then
	       release the lock and wait for the other buffer to become
	       the current buffer, so that we can fill it */
	    if (requested_size > BUFFER_MAX) {
		int other_buffer = buffer;
		other_buffer += 1;
		other_buffer = other_buffer % 2;

		done = 0;	/* we'll have to repeat this */
		ReleaseWriteLock(&(request_table[reqid].lock));
		LWP_WaitProcess((char *)
				&(request_table[reqid].buffer[other_buffer]));
		continue;	/* re-acquire the lock and start again */
	    }
	    
	    /* XXX probably better to realloc by doubling the size */
	    new_buffer = (char *) realloc(request_table[reqid].buffer[buffer],
					  requested_size);
	    if (new_buffer == NULL) {
		/* out of memory, so crash and burn */
		assert(0);
	    } else {
		request_table[reqid].buffer[buffer]   = new_buffer;
		request_table[reqid].capacity[buffer] = requested_size;
	    }
	}
    } while (!done);

    /* then, copy new data into buffer */
    bp = request_table[reqid].buffer[buffer];
    bp += request_table[reqid].size[buffer];
    memcpy(bp, data, size);
    request_table[reqid].size[buffer] += size;

    /* UNLOCK */
    ReleaseWriteLock(&(request_table[reqid].lock));

    /* if the ready data is big enough to send, wake up anyone
       who's waiting on it */
    if (request_table[reqid].size[buffer] >= BUFFER_THRESH) {
	LWP_SignalProcess((char *)&(request_table[reqid]));
    }

    return(size);
}

/* request table management */

/*
 * create_request -- allocates and initializes a new request, and
 *                   returns the request ID. If a new request cannot
 *                   be allocated, -1 is returned.
 */
int create_request(RPC2_Integer method,
		   RPC2_String  url,
		   RPC2_Integer img_quality,
		   RPC2_Integer html_quality,
		   RPC2_Integer last_modified)
{
    int                   loop;
    int                   reqid = -1;
    struct fetch_req     *req;

    /* find an unused request structure */
    ObtainWriteLock(&table_lock);
    for (loop = 0; loop < TABLE_SIZE; loop++) {
	if (request_table[loop].state == UNUSED) {
	    request_table[loop].state         = STARTING;
	    request_table[loop].fetch_done    = 0;
	    request_table[loop].img_quality   = img_quality;
	    request_table[loop].html_quality  = html_quality;
	    request_table[loop].last_modified = last_modified;
	    request_table[loop].compressing   = NO_COMPRESSION;
	    reqid = loop;
	    break;
	}
    }
    ReleaseWriteLock(&table_lock);

    if (reqid < 0)
	return reqid;		/* no free request structures */

    /* we've got a free request structure, so initialize it as needed */
    for (loop = 0; loop < 2; loop++) {
	if (request_table[reqid].buffer[loop] == NULL) {
	    request_table[reqid].buffer[loop] = (char *) malloc(BUFFER_INIT *
								sizeof(char));
	    if (request_table[reqid].buffer[loop] == NULL) {
		/* we're out of memory, so die */
		assert(0);
	    }

	    request_table[reqid].capacity[loop] = BUFFER_INIT;
	    request_table[reqid].size[loop]     = 0;
	    request_table[reqid].bufstate[loop]    = READY;
	}
    }

    request_table[reqid].bufinuse    = 0;       /* set buff 0 to be current */
    request_table[reqid].bufstate[0] = CURRENT;	/* buffer */


    /* clear out metadata, just in case */
    request_table[reqid].status_code = -1;
    request_table[reqid].status      = NULL;
    request_table[reqid].header_size = 0;
    request_table[reqid].headers     = NULL;

    /* queue a message for a fetcher thread */
    req = (struct fetch_req *) malloc(sizeof(struct fetch_req));
    if (req == NULL) {
	/* we're out of memory, so die */
	assert(0);
    }

    req->reqid   = reqid;
    req->method  = method;
    req->url     = url;
    req->date    = last_modified;

    ds_safeq_enq(fetch_q, req);

    /*
    LWP_CreateProcess(thread_shell, 32768, LWP_NORMAL_PRIORITY,
		      (char *) pp, url, &pid);
		      */

    return reqid;
}

/*
 * cleanup_request -- clean up a request after use.
 *
 * N.B.: assumes that a writelock is already held on the request before
 *       it is called.
 *
 */
void cleanup_request(int reqid)
{
    int loop;

    DEBUG(fprintf(stderr, "asked to clean up request %d\n", reqid));

    /* do sanity checking */
    if ( (request_table[reqid].state != CLOSED) &&
	 (request_table[reqid].fetch_done != 1) ) {
	/* if neither of these conditions are true, we should NOT
	   be cleaning things up; in fact, we shouldn't have been called */
	assert(0);
    }

    /* clean out buffers and reset them */
    for (loop = 0; loop < 2; loop++) {
	request_table[reqid].size[loop]     = 0;
	request_table[reqid].bufstate[loop] = READY;
	memset(request_table[reqid].buffer[loop],
	       0,
	       request_table[reqid].capacity[loop]);
    }

    request_table[reqid].bufinuse    = 0;       /* set buff 0 to be current */
    request_table[reqid].bufstate[0] = CURRENT;	/* buffer */

    /* deallocate metadata */
    request_table[reqid].status_code = -1;
    request_table[reqid].header_size = 0;

    if (request_table[reqid].status != NULL) {
	free(request_table[reqid].status);
	request_table[reqid].status = NULL;
    }

    SWWW_free_headerlist(request_table[reqid].headers);

    request_table[reqid].state = UNUSED;
}

/* LWP thread creation support */

/*
 * thread_shell -- creates a new LWP process to manage the call to
 *                 SWWW_fetch_url.
 */
int thread_shell(void *pp) {
    struct fetch_req     *req;
    int                   reqid;
    struct SWWW_fetch_in  in;
    struct SWWW_fetch_out out;
    int                   status_code;
    
    while (1) {
	req = (struct fetch_req *)ds_safeq_deq(fetch_q);
	reqid = req->reqid;

	/*
	 * call SWWW_fetch_url
	 */
	SWWW_SFI_INIT(&in);
	in.sfi_method        = GET;
	in.sfi_url           = req->url;
	in.sfi_reqid         = req->reqid;
	in.sfi_threading     = LWP;
	in.sfi_acceptheaders = accept_headers;
	in.sfi_acceptdata    = accept_web_data;
	
	if (req->date) {
	    struct tm *ftm;
	    char *date_string;
	    
	    ftm = localtime(&req->date);
	    date_string = asctime(ftm);
	    date_string[24] = '\0';	/* get rid of '\n' */
	    
	    in.sfi_headers = SWWW_create_headerlist();
	    SWWW_sethvalue(in.sfi_headers, "If-Modified-Since",
			   date_string);
	}
	status_code = SWWW_fetch_url(&in, &out);
	
	DEBUG(fprintf(stderr, "done fetching request %d\n", reqid));
	
	/* close off compression stream */
	if (request_table[reqid].compressing) {
	    struct data_chunk *tail;
	    CREATE_END_CHUNK(tail);
	    ds_safeq_enq(request_table[reqid].fordist, tail);
	}
	
	/* N.B.: the info in 'out' (and the status_code) is unnecessary,
	   since accept_headers already got it */
	
	/* XXX is there a failure mode in which we need to look at the
           status code to mark info about a failed fetch? */
	
	/*
	 * check if cleanup is necessary
	 */
	
	ObtainWriteLock(&(request_table[reqid].lock));
#ifdef OLD_STUFF
	if (!request_table[reqid].compressing)
#endif
	    request_table[reqid].fetch_done = 1;
	LWP_SignalProcess((char *)&(request_table[reqid]));
	
	/* if state == CLOSED, call cleanup_request */
	if (request_table[reqid].state == CLOSED) {
	    cleanup_request(reqid);
	}
	
	ReleaseWriteLock(&(request_table[reqid].lock));
	
	/* free parameters */
	free(req);
    }

    return (LWP_SUCCESS);
}

/*
 * compression_shell -- creates a new LWP process to manage the compression
 *                      of an HTTP object.
 */
int compression_shell(void *id) {
    struct compress_req        *req;
    int                         reqid;
    ds_safeq_t                 *readq;
    struct jpeg_destination_mgr dest;
    struct dest_mgr_params      dest_params;

    while (1) {
	req = (struct compress_req *)ds_safeq_deq(compress_q);
	reqid = req->reqid;

	readq = request_table[reqid].fordist;

	if (request_table[reqid].compressing == TO_JPEG) {
	    /* set up destination manager */
	    dest.init_destination    = my_init_destination;
	    dest.empty_output_buffer = my_empty_output_buffer;
	    dest.term_destination    = my_term_destination;
	    
	    dest_params.buffer       = (char *) malloc(4096);
	    dest_params.reqid        = reqid;
	    dest.user_data           = (char *) &dest_params;
	    
	    /* XXX hey, use a real content-type! */
	    cjpeg_main(readq,
		       "image/gif",
		       request_table[reqid].img_quality,
		       &dest);
	    
	    free(dest_params.buffer);
	} else if (request_table[reqid].compressing == TO_HTML) {
	    /* XXX do something useful here */
	    html_main(readq, reqid);
	} else {
	    fprintf(stderr, "Fatal error: compression type #%d is not "
		    "yet implemented.\n",
		    request_table[reqid].compressing);
	}
	
#ifdef OLD_STUFF
	request_table[reqid].fetch_done = 1;
	/* XXX should prolly be logic in here about calling cleanup_request */
	LWP_SignalProcess((char *)&(request_table[reqid]));
#endif    
	free(req);
    }

    return (LWP_SUCCESS);
}

/* destination manager for jpeg library */

void my_init_destination(j_compress_ptr cinfo)
{
    struct jpeg_destination_mgr *me = cinfo->dest;
    struct dest_mgr_params *params =
	(struct dest_mgr_params *) me->user_data;

    me->next_output_byte = params->buffer;
    me->free_in_buffer = 4096;
}

boolean my_empty_output_buffer(j_compress_ptr cinfo)
{
    struct jpeg_destination_mgr *me = cinfo->dest;
    struct dest_mgr_params *params =
	(struct dest_mgr_params *) me->user_data;

    accept_jpeg_data(params->reqid, params->buffer, 4096);

    me->next_output_byte = params->buffer;
    me->free_in_buffer = 4096;
    
    return TRUE;
}

void my_term_destination(j_compress_ptr cinfo)
{
    struct jpeg_destination_mgr *me = cinfo->dest;
    struct dest_mgr_params *params =
	(struct dest_mgr_params *) me->user_data;
    int size;

    size = 4096 - me->free_in_buffer;

    accept_jpeg_data(params->reqid, params->buffer, size);
}

void process_alias_file(char *alias_fname)
{
    FILE *alias_fp;
    int done = 0;
    
    /* read in alias table */
    alias_fp = fopen(alias_fname, "r");
    if (alias_fp == NULL) {
	aliasing = 0;
	fprintf(stderr, "Disabling aliasing: %s not found.\n",
		alias_fname);
    } else {
	alias_list = NULL;
	do {
	    char tmp_str[256];
	    int num_source_urls;
	    char *alias_url;
	    int result;
	    
	    result = fscanf(alias_fp, "%s %d\n", tmp_str, &num_source_urls);
	    if (result < 2) {
		done = 1;
		if (result == 1)
		    fprintf(stderr, "Couldn't completely parse aliases.\n");
	    } else {
		int i;
		alias_url = (char *)malloc(strlen(tmp_str)+1);
		strcpy(alias_url, tmp_str);
		for (i=0; i<num_source_urls; i++) {
		    if (fscanf(alias_fp, "%s", tmp_str) != 1) {
			done = 1;
			i = num_source_urls;
		    } else {
			char *request_url;
			struct alias_entry *new_alias;
			
			new_alias = (struct alias_entry *)malloc(sizeof(struct alias_entry));
			request_url = (char *)malloc(strlen(tmp_str)+1);
			strcpy(request_url, tmp_str);
			new_alias->request_url = request_url;
			new_alias->alias_url = alias_url;
			new_alias->next = alias_list;
			alias_list = new_alias;
		    }
		}
	    }
	} while (!done);
    }
    fclose(alias_fp);
}
