#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * various defines and macros
 */

#define TABLE_SIZE    32
#define BUFFER_INIT   25000	/* initial space alloc'd to buffers */
#define BUFFER_THRESH 24576	/* once buffer has more data than this,
				   it's a candidate for sending. */
#define BUFFER_MAX    65536     /* buffer cannot get larger than this;
				   if it does, throttle */

#define NUM_COMPRESS_THREADS 4
#define NUM_FETCH_THREADS    4

#define DEBUG(cmd) \
do { \
    if (debugging) \
        cmd; \
} while (0)

/*
 * data_chunk record -- used to funnel chunks of data in "streams"
 *                      via safeqs.
 */

enum chunk_type {
    DATA = 43,			/* contains data */
    END  = 59			/* end of "stream" */
};

struct data_chunk {
    enum chunk_type  type;	/* magic number indicating chunk type */
    char            *data;	/* the data */
    int              size;	/* number of bytes in chunk */
    char            *index;	/* index into chunk (if we've already
				   eaten part of it) */
    int              remaining;	/* number of bytes remaining, including
				   the byte pointed to by 'index' */
};

#define CREATE_CHUNK(chunk, buf, bsize) \
do { \
    chunk          = (struct data_chunk *) malloc(sizeof(struct data_chunk)); \
    (chunk)->data  = (char *) malloc(bsize+1); \
    memcpy((chunk)->data, buf, bsize); \
    (chunk)->data[bsize] = '\0'; \
    (chunk)->size        = bsize; \
    (chunk)->index       = (chunk)->data; \
    (chunk)->remaining   = bsize; \
    (chunk)->type        = DATA; \
} while (0)

#define CREATE_END_CHUNK(chunk) \
do { \
    chunk = (struct data_chunk *) malloc(sizeof(struct data_chunk)); \
    (chunk)->data      = NULL; \
    (chunk)->size      = 0; \
    (chunk)->index     = NULL; \
    (chunk)->remaining = 0; \
    (chunk)->type      = END; \
} while (0)

#define FREE_CHUNK(chunk) \
do { \
    if (chunk != NULL) { \
	if ((chunk)->type == DATA) \
            free((chunk)->data); \
        (chunk)->data      = NULL; \
        (chunk)->index     = NULL; \
        (chunk)->size      = 0; \
        (chunk)->remaining = 0; \
        free(chunk); \
        chunk = NULL; \
    } \
} while (0)

/*
 * request record
 */

/* buffers can make the transition between states in the following way:
 *    STARTING -> GOTHEADERS -> CLOSED -> UNUSED
 *
 * NB: the total order on these states is important; specifically, states
 *     in which we accept data should be less than states in which we
 *     don't accept new data (i.e. < CLOSED )
 */

enum request_state {
    STARTING = 14,		/* connection has been started, but
				   the headers have not yet been fetched */
    GOTHEADERS = 25,		/* connection is in use, and the
				   headers are available */
    PRECLOSE = 30,		/* all data has been sent via fetch_buffer */
    CLOSED = 37,		/* connection has been killed, but
				   data may still be coming in from
				   the network (and should be discarded) */
    UNUSED = 65			/* connection is not in use */
};

enum buffer_state {
    READY,			/* currently unused buffer */
    CURRENT,			/* buffer currently being filled */
    SENDING			/* buffer currently being sent */
};

enum datatype {
    FROM_WEB,			/* data from the SWWW library */
    FROM_JPEG,			/* data from the JPEG library */
    FROM_HTML			/* data from the HTML compressor */
};

enum compress_type {
    NO_COMPRESSION = 0,		/* don't compress */
    TO_JPEG = 15,		/* compress to JPEG */
    TO_HTML = 23		/* compress to html */
};

struct request {
    enum request_state state;	    /* state of this connection */
    int                fetch_done;  /* 0 while fetch_url is still executing,
				       1 when it is done */
				    /* N.B. if the state is CLOSED,
				            and fetch_done == 1, the
					    state can become UNUSED */

    struct Lock        lock;	    /* lock associated with this connection */
    int                img_quality; /* minimum acceptable image quality */
    int                html_quality;/* minimum acceptable HTML quality */
    time_t             last_modified;/* mod date for cached object */

    /* buffers for streaming */
    int                bufinuse;    /* buffer currently in use */
    char              *buffer[2];   /* the buffers */
    int                size[2];	    /* amount of data in the buffer */
    int                capacity[2]; /* current capacity of buffer */
    enum buffer_state  bufstate[2]; /* state of buffer */

    /* buffers for compression */
    enum compress_type compressing; /* value indicates whether compression
				       is occuring, and if so, to what. */
    ds_safeq_t        *fordist;	    /* safeq used to send data to
				       the appropriate distiller. */
    /* XXX OLD, TO BE REMOVED
    int                fordist[2]; */  /* two ends of a pipe used to send
				       data to the appropriate distiller. */

    /* request metadata */
    int                status_code;
    char              *status;
    int                header_size;
    SWWW_headers_t    *headers;
};

struct fetch_req {
    int   reqid;
    int   method;
    char *url;
    time_t date;
};

struct compress_req {
    int   reqid;
};

struct dest_mgr_params {
    char *buffer;
    int   reqid;
};

struct alias_entry {
    char *request_url;
    char *alias_url;
    struct alias_entry *next;
};

/*
 * macros
 */

/* this is a server-only, non-general macro (not to be confused with a
   similar one used by the client). I assume that the server is the source */
#define setupVMSE(descriptor, vmaddr, vmlength) \
        bzero(descriptor, sizeof(SE_Descriptor)); \
        (descriptor)->Tag = SMARTFTP; \
        (descriptor)->Value.SmartFTPD.Tag = FILEINVM; \
        (descriptor)->Value.SmartFTPD.hashmark = 1; /* verbose transfer */ \
        (descriptor)->Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT; \
        (descriptor)->Value.SmartFTPD.ByteQuota = -1; \
        (descriptor)->Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = vmaddr;\
        (descriptor)->Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = vmlength;

/*
 * prototypes
 */

void error_report(char *message);

void Init_RPC(void);

int accept_headers(int             reqid,
		   int             status_code,
		   char           *status, 
		   int             header_size,
		   SWWW_headers_t *headers);

int accept_web_data(int         reqid,
		    const void *data,
		    int         size);

int accept_jpeg_data(int         reqid,
		     const void *data,
		     int         size);

int accept_html_data(int         reqid,
		     const void *data,
		     int         size);

int accept_data(int        reqid,
		const void *data,
		int        size,
		enum datatype  type);

int create_request(RPC2_Integer method,
		   RPC2_String  url,
		   RPC2_Integer img_quality,
		   RPC2_Integer html_quality,
		   RPC2_Integer last_modified);

void cleanup_request(int reqid);
int thread_shell(void *pp);
int compression_shell(void *reqid);
void my_init_destination(j_compress_ptr cinfo);
boolean my_empty_output_buffer(j_compress_ptr cinfo);
void my_term_destination(j_compress_ptr cinfo);
void process_alias_file(char *alias_fname);
