#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/uio.h>
#include "include/odyssey.h"
#include "include/odytypes.h"
#include "SetDialog.h"
#include "DigestDialog.h"
#include "WWWSets.h"
#include "SetsMenu.h"
#include "sets.h"
#include "xutils.h"
#include <Xm/Xm.h>
#include <Xm/Frame.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/SeparatoG.h>

DialogLL* pSetDialogList = 0;

/* findSetDialog() finds the Set Dialog assocaited with the specified window.
   A window is associated with a set dialog if it is the dialog window or
   one of its subchildren.  The function returns a pointer to the dialog
   data structure if an associated dialog is found, and NULL otherwise. */
Set_ClientData* findSetDialog (Display* display, Window w)
{
  DialogLL* pNode;

  for (pNode = pSetDialogList; pNode; pNode = pNode->pNext) {
    if (isDescendant (display, w, XtWindow(pNode->pClientData->dialog))) {
      return pNode->pClientData;
    } /* endif */
  } /* endfor */

  return (NULL);
}
    
/* firstSetDialog() returns the data structure associated with the first
   set dialog in the list.  If no set dialog is in the list, the function
   reuturns NULL.  Since the list is maintained in order of usage, the
   first dialog in the list is the most recent dialog touched by the user. */
Set_ClientData* firstSetDialog ()
{
  if (pSetDialogList) {
    return (pSetDialogList->pClientData);
  } else {
    return (NULL);
  } /* endif */
}

/* RemoveFromDialogList() is called to remove a set dialog from the linked
   list when a set is closed. */
static void RemoveFromDialogList (Set_ClientData* pClientData)
{
  DialogLL* pNode;
  DialogLL* pLast = 0;
  
  for (pNode = pSetDialogList; pNode; pNode = pNode->pNext) {
    if (pNode->pClientData == pClientData) {

      if (pLast) {
	pLast->pNext = pNode->pNext;
      } else {
	pSetDialogList = pNode->pNext;
      } /* endif */
      FREE (pNode);
      return;

    } else {
      pLast = pNode;
    } /* endif */
  } /* endfor */

  /* There should always be a dialog to delete */
  ASSERT (0);
}

/* AdjustDialogList() is called whenever a dialog is touched or created
   by the user.  If the specified set dialog data structure already 
   exists in the list, the dialog is moved to the front of the list.  If
   the data structure is not in the list, it is inserted at the front of
   the list. */
static void AdjustDialogList (Set_ClientData* pClientData) 
{
  DialogLL* pNode;
  DialogLL* pLast = 0;
  
  for (pNode = pSetDialogList; pNode; pNode = pNode->pNext) {
    if (pNode->pClientData == pClientData) {
      if (pLast) {
	pLast->pNext = pNode->pNext;
	pNode->pNext = pSetDialogList;
	pSetDialogList = pNode;
      } /* endif */
      return;
    } /* endif */
  } /* endfor */

  ALLOC (pNode, DialogLL);
  pNode->pClientData = pClientData;
  pNode->pNext = pSetDialogList;
  pSetDialogList = pNode;
}
      
/* This function is called when the user selects the Iterate push
   button on the dialog.  It iterates on the set and displays the next
   url (as returned by setIterate) */
void Set_Iterate_Callback (Widget w, XtPointer client_data, 
			   XtPointer call_data)
{
  char buffer[512]; /* Sets requires preallocated buffer */
  Set_ClientData* pClientData;
  char msg[100];
  int rc;

  /* Iterate on the set.  This returns the name of the next URL in the 
     buffer */
  pClientData = (Set_ClientData *) client_data;  
  AdjustDialogList (pClientData);
  rc = IterateOnSet (pClientData->set, buffer, 512);
  if (rc == 0) {

    /* Display the retured URL */
    netscapeRemoteOpenURL (buffer, pClientData->display, 
			   pClientData->pNetscapeWindow, 0);  

  } else if (rc == 1) {
    /* Display a message that all elements have been yielded */
    InfoMessage (w, "No More Members", "All set members have been yielded");
  } else {
    /* Display a message that all elements are in error */
    sprintf (msg, "An error occurred iterating on the set.  The associated "
	     "return code is %d", rc);
    ErrorMessage (w, "Set Iterate Error", msg);
  } /* endif */
}

void Set_Digest_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  Set_ClientData* pClientData;
  Digest digest;
  DigestElem* pElem;

  /* Get a digest of the set */
  pClientData = (Set_ClientData *) client_data;  
  AdjustDialogList (pClientData);
  GetDigestOfSet (pClientData->set, &digest);
  Create_Digest_Dialog (w, pClientData->set, 
			pClientData->setName, &digest, pClientData->display,
			pClientData->pNetscapeWindow);

  /* Deallocate the digest members*/
  /* Deallocate the digest members*/
  while (digest.pList) {
    pElem = digest.pList;
    digest.pList = pElem->pNext;
    free (pElem);
  } /* end while */
}

/* This function is called when the user selects the Size push
   button on the dialog.  It displays the size of the set in the
   Netscape text output window. */
void Set_Size_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  Set_ClientData* pClientData;
  int size;
  char str[512];

  pClientData = (Set_ClientData *) client_data;  
  AdjustDialogList (pClientData);
  size = GetSetSize (pClientData->set);
  sprintf (str, "The set has %d elements", size); 
  netscapeString (str, pClientData->display, pClientData->pNetscapeWindow);
}

/* This function is called when the user selects the Rewind push
   button on the dialog.  It rewinds the set and displays a message
   to that effect in the Netscape text output window */
void Set_Rewind_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  Set_ClientData* pClientData;

  pClientData = (Set_ClientData *) client_data;  
  AdjustDialogList (pClientData);
  RewindSetIterator (pClientData->set);
  netscapeString ("The iterator has been reinitialized", pClientData->display,
		  pClientData->pNetscapeWindow);
}

/* This function is called when the user selects the Close Set push
   button on the dialog.  It first closes the set, then closes the
   dialog. */
void Set_Close_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  Set_ClientData* pClientData;

  /* First, close the set */
  pClientData = (Set_ClientData *) client_data;  
  CloseSet (pClientData->set);
  RemoveFromDialogList (pClientData);

  /* Then, destroy the menu-item associated with the set */
  XtDestroyWidget (pClientData->menuItem);
  
  /* If there are less than 2 open sets, make sure the union and intersect
     buttons are not sensitive */
  if (GetNumberOfOpenSets() < 2) {
    XtSetSensitive (pClientData->unionWidget, False);
    XtSetSensitive (pClientData->intersectWidget, False);
  } /* endif */

  /* Now close the dialog */
  XtDestroyWidget (pClientData->dialog);
  if (pClientData) {
    if (pClientData->setName) {
      free (pClientData->setName);
    } /* endif */
    free (pClientData);
  } /* endif */
}

/* This function is called when the user selects the Cancel push
   button on the dialog.  It deallocates the call-data structure and
   unmaps the dialog. */
void Set_Cancel_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
  Set_ClientData* pClientData;

  /* Just unmap the window, since the set is still open */
  pClientData = (Set_ClientData *) client_data;
  AdjustDialogList (pClientData);
  XtUnmanageChild (pClientData->dialog);
}

void Set_Help_Callback (Widget w, XtPointer client_data, XtPointer call_data)
{
}

void Set_MenuItem_Callback (Widget w, XtPointer client_data, 
			    XtPointer call_data)
{
  Set_ClientData* pClientData;

  /* Map the specified window */
  pClientData = (Set_ClientData *) client_data;
  XtUnmanageChild (pClientData->dialog);
  XtManageChild (pClientData->dialog);
}

/* This function creates a dialog for an open set.  The set id and
   name should be passed into the function, along with the widget for
   the root application window. */
void Create_Set_Dialog (Widget parent, Widget menu, Widget unionWidget, 
			Widget intersectWidget, int sp, char* url, 
			Display* display, Window* pNetscapeWindow)
{
  Widget dialog, dialog_frame, sets_form, dialog_sep, buttons_form;
  Widget label, text, iterateButton, digestButton, rewindButton, cancelButton;
  Widget sizeButton, closeButton, helpButton, setMenuItem;
  Set_ClientData* pClientData;  
  XmString xmstr;
  Arg args[10];

  /* Create a new client data structure for this dialog */
  pClientData = (Set_ClientData *) malloc (sizeof(Set_ClientData));

  /* Create dialog box */
  xmstr = XmStringCreateLtoR ("Open Set Window", XmSTRING_DEFAULT_CHARSET);
  XtSetArg(args[0], XmNautoUnmanage, (XtArgVal) False);
  XtSetArg(args[1], XmNdialogTitle, (XtArgVal) xmstr);
  dialog = XmCreateFormDialog (parent, "dialog", args, 2);
  XmStringFree (xmstr);

  /* Create a frame for the dialog window */
  XtSetArg(args[0], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[1], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[2], XmNleftAttachment, XmATTACH_FORM);
  XtSetArg(args[3], XmNrightAttachment, XmATTACH_FORM);
  XtSetArg(args[4], XmNshadowType, XmSHADOW_OUT);
  dialog_frame = XtCreateManagedWidget ("frame", xmFrameWidgetClass,
					dialog, args, 5);
      
  /* Create the main form for the dialog */
  sets_form = XtCreateManagedWidget ("form", xmFormWidgetClass, dialog_frame,
				     NULL, 0);

  /* Create a label for the entry field */
  XtSetArg(args[0], XmNtopOffset, (XtArgVal) 14);
  XtSetArg(args[1], XmNleftOffset, (XtArgVal) 4);
  XtSetArg(args[2], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[3], XmNleftAttachment, XmATTACH_FORM);
  label = XtCreateManagedWidget ("Set Name: ", xmLabelWidgetClass,
                                 sets_form, args, 4); 

  /* Create a read-only text field to show the set's name. */
  XtSetArg(args[0], XmNwidth, (XtArgVal) 310);
  XtSetArg(args[1], XmNtopOffset, (XtArgVal) 10);
  XtSetArg(args[2], XmNleftOffset, (XtArgVal) 5);
  XtSetArg(args[3], XmNrightOffset, (XtArgVal) 10);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNleftAttachment, XmATTACH_WIDGET);
  XtSetArg(args[6], XmNrightAttachment, XmATTACH_FORM);
  XtSetArg(args[7], XmNleftWidget, (XtArgVal) label);
  XtSetArg(args[8], XmNeditable, 0); /* Don't allow user to edit this field */
  XtSetArg(args[9], XmNvalue, url);
  text = XmCreateTextField (sets_form, "textfield", args, 10);
  XtManageChild (text);

  /* Create a separator between entry field and push buttons */
  XtSetArg(args[0], XmNtopOffset, (XtArgVal) 10);
  XtSetArg(args[1], XmNtopAttachment, XmATTACH_WIDGET);
  XtSetArg(args[2], XmNleftAttachment, XmATTACH_FORM);
  XtSetArg(args[3], XmNrightAttachment, XmATTACH_FORM);
  XtSetArg(args[4], XmNtopWidget, (XtArgVal) text);
  dialog_sep = XmCreateSeparatorGadget (sets_form, "separator", args, 5);
  XtManageChild (dialog_sep);

  /* Create form for push buttons */
  XtSetArg(args[2], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[3], XmNleftAttachment, XmATTACH_FORM);
  XtSetArg(args[4], XmNrightAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNtopAttachment, XmATTACH_WIDGET);	/* top */
  XtSetArg(args[6], XmNtopWidget, dialog_sep);
  XtSetArg(args[7], XmNtopOffset, 4);
  buttons_form = XtCreateManagedWidget ("form", xmFormWidgetClass, sets_form,
					args, 8);

  /* Create big iterate button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_FORM);
  XtSetArg(args[1], XmNleftOffset, 8);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_POSITION);
  XtSetArg(args[3], XmNrightPosition, 26);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNtopOffset, 4);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[7], XmNbottomOffset, 4);
  iterateButton = XtCreateManagedWidget ("Iterate", xmPushButtonWidgetClass,
					 buttons_form, args, 8);
  XtAddCallback(iterateButton, XmNactivateCallback, Set_Iterate_Callback, 
		pClientData);
    
  /* Create and place the digest button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 29);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_POSITION);
  XtSetArg(args[3], XmNrightPosition, 51);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNtopOffset, 4);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_POSITION);
  XtSetArg(args[7], XmNbottomPosition, 48);
  digestButton = XtCreateManagedWidget ("Digest", xmPushButtonWidgetClass,
					buttons_form, args, 8);
  XtAddCallback(digestButton, XmNactivateCallback, Set_Digest_Callback, 
		pClientData);
		  
  /* Create and place the rewind button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 29);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_POSITION);
  XtSetArg(args[3], XmNrightPosition, 51);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_POSITION);
  XtSetArg(args[5], XmNtopPosition, 53);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[7], XmNbottomOffset, 4);
  rewindButton = XtCreateManagedWidget ("Rewind", xmPushButtonWidgetClass,
					buttons_form, args, 8);
  XtAddCallback(rewindButton, XmNactivateCallback, Set_Rewind_Callback, 
		pClientData);
    
  /* Create and place the size button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 54);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_POSITION);
  XtSetArg(args[3], XmNrightPosition, 76);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNtopOffset, 4);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_POSITION);
  XtSetArg(args[7], XmNbottomPosition, 48);
  sizeButton = XtCreateManagedWidget ("Size", xmPushButtonWidgetClass,
				      buttons_form, args, 8);
  XtAddCallback(sizeButton, XmNactivateCallback, Set_Size_Callback, 
pClientData);
    
  /* Create and place the close button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 54);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_POSITION);
  XtSetArg(args[3], XmNrightPosition, 76);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_POSITION);
  XtSetArg(args[5], XmNtopPosition, 53);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[7], XmNbottomOffset, 4);
  closeButton = XtCreateManagedWidget ("Close Set", xmPushButtonWidgetClass,
				       buttons_form, args, 8);
  XtAddCallback(closeButton, XmNactivateCallback, Set_Close_Callback, 
		pClientData);

  /* Create and place the cancel button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 79);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_FORM); 
  XtSetArg(args[3], XmNrightOffset, 8);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_FORM);
  XtSetArg(args[5], XmNtopOffset, 4);
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_POSITION);
  XtSetArg(args[7], XmNbottomPosition, 48);
    
  cancelButton = XtCreateManagedWidget ("Cancel", xmPushButtonWidgetClass,
					buttons_form, args, 8);    
  XtAddCallback(cancelButton, XmNactivateCallback, Set_Cancel_Callback, 
		pClientData);
    
  /* Create and place the help button */
  XtSetArg(args[0], XmNleftAttachment, XmATTACH_POSITION);
  XtSetArg(args[1], XmNleftPosition, 79);
  XtSetArg(args[2], XmNrightAttachment, XmATTACH_FORM); 
  XtSetArg(args[3], XmNrightOffset, 8);
  XtSetArg(args[4], XmNtopAttachment, XmATTACH_POSITION);
  XtSetArg(args[5], XmNtopPosition, 53);    
  XtSetArg(args[6], XmNbottomAttachment, XmATTACH_FORM);
  XtSetArg(args[7], XmNbottomOffset, 4);
  helpButton = XtCreateManagedWidget ("Help...", xmPushButtonWidgetClass,
				      buttons_form, args, 8);
  XtAddCallback(helpButton, XmNactivateCallback, Set_Help_Callback,
		pClientData);

  /* Add a new item to the Sets menu */
  setMenuItem = XtCreateManagedWidget(url, xmPushButtonWidgetClass,
				      menu, NULL, 0);
  XtAddCallback(setMenuItem, XmNactivateCallback, Set_MenuItem_Callback, 
		pClientData);

  /* If there are 2 or more open sets, then union and intersct should
     be sensitive (selecctable) */
  if (GetNumberOfOpenSets() >= 2) {
    XtSetSensitive (unionWidget, True);
    XtSetSensitive (intersectWidget, True);
  } /* endif */

  /* Populate Client Data structure */
  pClientData->parent = parent;
  pClientData->dialog = dialog;
  pClientData->menuItem = setMenuItem;
  pClientData->unionWidget = unionWidget;
  pClientData->intersectWidget = intersectWidget;
  pClientData->set = sp;
  pClientData->setName = strdup (url);
  pClientData->display = display;
  pClientData->pNetscapeWindow = pNetscapeWindow;

  XtManageChild (dialog);

  AdjustDialogList (pClientData);
}


