#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "xutils.h"
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

/* getParent() returns the parent of the specified window. */
Window getParent (Display* display, Window window)
{
  Window root, parent, *kids;
  unsigned int nkids;
  
  if (!XQueryTree (display, window, &root, &parent, &kids, &nkids)) {
    printf ("ERROR: Unable to query tree\n");
    return (0);
  } /* endif */

  /* Deallocate list of child windows */
  XFree (kids);

  return (parent);
}

/* getRoot() returns the root of the specified window. */
Window getRoot (Display* display, Window window)
{
  Window root, parent, *kids;
  unsigned int nkids;
  
  if (!XQueryTree (display, window, &root, &parent, &kids, &nkids)) {
    printf ("ERROR: Unable to query tree\n");
    return (0);
  } /* endif */

  /* Deallocate list of child windows */
  XFree (kids);

  return (root);
}

/* isWindowMapped() returns true if the specified window is currnetly mapped */
int isWindowMapped (Display* display, Window window)
{
  XWindowAttributes atts;

  if (XGetWindowAttributes (display, window, &atts)) {
    return (atts.map_state == IsViewable);
  } else {
    return (0);
  } /* emdif */
}

/* getChildOffset() calculates the x and y offset of a specified child
   (or descendant) window relative to the specified parent window.  If
   the parent is not an ancestor of the child, the function returns -1.
   If the function succeeds, it returns zero.  */
int getChildOffset (Display* display, Window child, Window ancestor, 
		    int* x_ret, int* y_ret)
{
  Window root;
  int x, y;
  unsigned int width, height, border, depth;

  *x_ret = 0;
  *y_ret = 0;

  /* Loop until the specified ancestor is found */
  while (child != ancestor) {
    
    if (child == root) {
      printf ("The specified windows are not related\n");
      return (-1);
    } /* endif */
    
    if (XGetGeometry (display, child, &root, &x, &y, &width, &height, 
		      &border, &depth)) {
      *x_ret += x;
      *y_ret += y;
      child = getParent (display, child);
      if (!child) {
	printf ("Unable to get parent\n");
	return (-1);
      } /* endif */
      
    } else {
      printf ("Unable to get geometry\n");
      return (-1);
    } /* endif */
  }

  return (0);
}

/* isDecendant() returns true if the first window specified is a 
   decendant of the second window specified, and false otherwise. */
int isDescendant (Display* display, Window descendant, Window ancestor)
{
  Window root2, parent, *kids;
  unsigned int nkids;
  int i;
  int found = 0; /* Assume failure */

  if (!XQueryTree (display, ancestor, &root2, &parent, &kids, &nkids)) {
    printf ("ERROR: Unable to query tree\n");
    return (0);
  } /* endif */

  for (i = 0; i < nkids && !found; i++) {
    if (kids[i] == descendant || isDescendant (display, descendant, kids[i])) {
      found = 1;
    } /* endif */
  } /* endfor */

  XFree (kids);
  return (found);
}

/* getChildCoord() determines which child window and at which x and y
   coordinates relative to its origin a particular x, y coordinate at 
   an ancestor window is equivalent to.  This can be used to do cool
   things like send events to the child window. */
int getChildCoord (Display* display, Window ancestor, int x, int y,
		   Window* child, int* x_ret, int* y_ret)
{
  Window root, parent, *kids;
  int gx, gy, nkids, i;
  unsigned int width, height, border, depth;
  int found;

  *x_ret = 0;
  *y_ret = 0;

  do {

    if (!XQueryTree (display, ancestor, &root, &parent, &kids, &nkids)) {
      printf ("ERROR: Unable to query tree\n");
      return (-1);
    } /* endif */

    found = 0;
    for (i = nkids-1; i >= 0 && !found; i--) {

      if (XGetGeometry (display, kids[i], &root, &gx, &gy, &width, &height, 
			&border, &depth)) {

	if (gx <= x && gy <= y && gx+width >= x && gy+height >= y) {
	  
	  /* Event occurred within this child window */
	  found = 1;
	  ancestor = kids[i];
	  x -= gx;
	  y -= gy;

	} /* endif */

      } else {
	printf ("Unable to get geometry\n");
	return (-1);
      } /* endif */

    } /* endfor */
    
    XFree (kids);
      
  } while (found);

  *child = ancestor;
  *x_ret = x;
  *y_ret = y;

  return (0);
}

int sortByGeometry (const void* a, const void* b) 
{
  if (((WindowGeom *)a)->dist < ((WindowGeom *)b)->dist) {
    return (-1);
  } else if (((WindowGeom *)a)->dist > ((WindowGeom *)b)->dist) {
    return (1);
  } else {
    return (0);
  } /* endif */
}

/* getChildByGeometry() return the child window in the nth position
   when the list of children are sorted in the direction specified.
   Results of this function are undetermined when more than one child
   window can be at the specified position because their geometrical
   distances are equal. */
Window getChildByGeometry (Display* display, Window parent, 
			   enum directionType dir, int num)
{
  Window root, ancestor, win, *kids;
  unsigned int nkids;
  int i;
  int gx, gy, width, height, border, depth;
  WindowGeom* sortList;

  if (!XQueryTree (display, parent, &root, &ancestor, &kids, &nkids)) {
    printf ("ERROR: Unable to query tree\n");
    return (-1);
  } /* endif */
 
  if (nkids < num) {
    return (-1);
  } /* endif */
  
  /* Sort by geometry */
  sortList = (WindowGeom *) malloc (nkids * sizeof(WindowGeom));
  for (i = nkids-1; i >= 0; i--) {
    sortList[i].win = kids[i];
    if (XGetGeometry (display, kids[i], &root, &gx, &gy, &width, &height, 
		      &border, &depth)) {
      if (dir == DIR_NORTH) {
	sortList[i].dist = gy;
      } else if (dir == DIR_SOUTH) {
	sortList[i].dist = -gy;
      } else if (dir == DIR_EAST) {
	sortList[i].dist = gx;
      } else if (dir == DIR_WEST) {
	sortList[i].dist = -gx;
      } else if (dir == DIR_NORTHEAST) {
	sortList[i].dist = gx + gy;
      } else if (dir == DIR_NORTHWEST) {
	sortList[i].dist = gy - gx;
      } else if (dir == DIR_SOUTHEAST) {
	sortList[i].dist = gx - gy;
      } else if (dir == DIR_SOUTHWEST) {
	sortList[i].dist = -gx - gy;
      } else {
	printf ("Invalid direction specified");
      } /* endif */
    } else {
      printf ("Unable to get geometry\n");
      return (-1);
    } /* endif */
  } /* endfor */
  XFree (kids);
  qsort (sortList, nkids, sizeof(WindowGeom), sortByGeometry);

  win = sortList[num-1].win;
  free (sortList);

  return (win);
}

/* pressButton () presses a button in the specified window by sending to
   that window a button-press event for button 1, followed by a button-
   release event for button 2 */
void pressButton (Display* display, Window w)
{
  /* Initialize button event structures */
  XEvent buttonPress, buttonRelease;
  memset (&buttonPress, 0, sizeof (buttonPress));
  memset (&buttonRelease, 0, sizeof (buttonRelease));

  /* Populate structures with appropriate data */
  buttonPress.xbutton.type = ButtonPress;
  buttonPress.xbutton.send_event = True;
  buttonPress.xbutton.display = display;
  buttonPress.xbutton.window = w;
  buttonPress.xbutton.root = getRoot (display, w);
  buttonPress.xbutton.state = 0x0100;
  buttonPress.xbutton.button = 1;
  buttonPress.xbutton.same_screen = True;
  buttonRelease.xbutton.type = ButtonRelease;
  buttonRelease.xbutton.send_event = True;
  buttonRelease.xbutton.display = display;
  buttonRelease.xbutton.window = w;
  buttonRelease.xbutton.root = buttonPress.xbutton.root;
  buttonRelease.xbutton.state = 0x0100;
  buttonRelease.xbutton.button = 1;
  buttonRelease.xbutton.same_screen = True;

  /* Now send the events to the window */
  XSendEvent (display, w, False, ButtonPressMask, &buttonPress);
  XSendEvent (display, w, False, ButtonReleaseMask, &buttonRelease);
}

/* pressKey () presses a key in the specified window by sending to
   that window a key-press event for button 1 followed by a key-release
   event. */
void pressKey (Display* display, Window w, unsigned int state, 
	       unsigned int keycode)
{
  /* Initialize button event structures */
  XEvent keyPress, keyRelease;
  memset (&keyPress, 0, sizeof (keyPress));
  memset (&keyRelease, 0, sizeof (keyRelease));

  /* Populate structures with appropriate data */
  keyPress.xkey.type = KeyPress;
  keyPress.xkey.send_event = True;
  keyPress.xkey.display = display;
  keyPress.xkey.window = w;
  keyPress.xkey.root = getRoot(display, w);
  keyPress.xkey.state = state;
  keyPress.xkey.keycode = keycode;
  keyPress.xkey.same_screen = True;
  keyRelease.xkey.type = KeyRelease;
  keyRelease.xkey.send_event = True;
  keyRelease.xkey.display = display;
  keyRelease.xkey.window = w;
  keyRelease.xkey.root = keyPress.xkey.root;
  keyRelease.xkey.state = state;
  keyRelease.xkey.keycode = keycode;
  keyRelease.xkey.same_screen = True;

  /* Now send the events to the window */
  XSendEvent (display, w, False, KeyPressMask, &keyPress);
  XSendEvent (display, w, False, KeyReleaseMask, &keyRelease);
}
