#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "include/odytypes.h"
#include "ds_list.h"
#include "ds_hash.h"
#include "ds_thread.h"
#include "ds_hwlock.h"

#ifndef ODYDEBUG
#error This module is meant for use only with ODYDEBUG
#endif

/* maintain dependency graph for hwlocks.

   if I am blocked on a ds_hwlock_read(), then I have a dependency on
   the wpo of that graph. If I am blocked on a ds_hwlock_write(), then
   i have a dependency on all the readers of that lock.

   So, every time a thread blocks, we add it to a list, which maps
   threads to the locks. When the thread unblocks, we remove it.

   We define the graph by a primitive find_pred: thread -> set of threads,
   which finds all the threads that a particular thread is dependent on.

   So now, to detect a cycle in this digraph, we go through each thread
   in it, and find the transitive closure of find_pred. if this contains
   the thread itself, we have a cycle == deadlock.

   find_deadlock() will return true if there is a deadlock, false if
   not.  we could, for example hook SIGINT to call find_deadlock, so
   that if the program hangs, we could check for deadlock.
*/

/* structure that maps a thread to a lock it's blocked on */
struct dep_info {
    ds_thread_t t;  /* the blocked thread */
    ds_hwlock_t *l; /* the lock it's blocked on */
};

static long dh_dep_comp(e1, e2)
    struct dep_info *e1, *e2;
{
    /* two elements are equal if they refer to the same thread */
    if (e1->t == e2->t)
	return 0;

    /* if not equal, some arbitrary total order */
    return ((long) e1->t - (long) e2->t);
}

/* list of struct dep_info for threads blocked on ds_hwlock_read() */
static ds_list_t *read_blocked_list = NULL;

/* list of struct dep_info for threads blocked on ds_hwlock_write() */
static ds_list_t *write_blocked_list = NULL;

void dh_add_dep(thread, lock, whichop)
    ds_thread_t thread;
    ds_hwlock_t *lock;
    int whichop;
{
    struct dep_info dh_di, *dh_dip;

    if ((whichop == READ) && (read_blocked_list == NULL))
	read_blocked_list = ds_list_create(dh_dep_comp, TRUE, FALSE);

    if ((whichop == WRITE) && (read_blocked_list == NULL))
	write_blocked_list = ds_list_create(dh_dep_comp, TRUE, FALSE);

    dh_di.t = thread;

    /* a thread cannot block on two things at the same time! */
    if (read_blocked_list && ds_list_member(read_blocked_list,&dh_di))
	ASSERT(0);

    if (write_blocked_list && ds_list_member(write_blocked_list,&dh_di))
	ASSERT(0);

    if (whichop == READ)
	the_list = read_blocked_list;
    else if (whichop == WRITE)
	the_list = write_blocked_list;
    else
	ASSERT(0);

    ALLOC(dh_dip, struct dep_info);

    dh_dip->t = thread;
    dh_dip->l = lock;

    if (!ds_list_insert(the_list, dh_dip))
	ASSERT(0);
}


void dh_remove_dep(thread, lock)
    ds_thread_t thread;
    ds_hwlock_t *lock;
{
    int count = 0;
    struct dep_info dh_di, *dh_dip;

    dh_di.t = thread;

    /* when the function is called, the thread should be in exactly
       one of the blocked lists, and it should be blocked on the lock
       specified */

    if (read_blocked_list && (dh_dip = ds_list_member(read_blocked_list,
						      &dh_di)))
	count++;

    if (read_blocked_list && (dh_dip = ds_list_member(write_blocked_list,
						      &dh_di)))
	count++;


