#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "include/odytypes.h"
#include "ds_list.h"
#include "ds_hash.h"
#include "ds_thread.h"
#include "ds_hwlock.h"

int ds_hwlock_debug = 0;

/* element that goes into hash table: for each thread that has a write-lock,
   contains the number of times that thread has read-locked that lock */

struct rl_info {
    ds_thread_t reader;
    int count;
};

/* comparison function 
 */

static long ds_hwlock_comp(e1, e2)
    struct rl_info *e1, *e2;
{
    /* two elements are equal if they refer to the same thread */
    if (e1->reader == e2->reader)
	return 0;

    /* if not equal, some arbitrary total order */
    return ((long) e1->reader - (long) e2->reader);
}

/* size of our hash table -- some prime that's not close to a power of 2 */

#define ds_hwlock_nbuckets 5

/* hash function for our hash table */

static long ds_hwlock_hash(rlp)
    struct rl_info *rlp;
{
    long hval = (long) rlp->reader;
    /* add in a for-real hashing algorithm here */

    return hval;
}

/* Initialize a ds_hwlock_t structure */

void ds_hwlock_init(lock)
    ds_hwlock_t *lock;
{
    if (ds_hwlock_debug)
	ds_thread_log(0, "Initializing hwlock %p", lock);

    ds_lock_init(&((lock)->dh_wlock));
    ds_lock_write(&((lock)->dh_wlock));
    (lock)->dh_rpo = (lock)->dh_wpo = NULL;
    (lock)->dh_rpc = (lock)->dh_wpc = (lock)->dh_rc = 0;
    (lock)->dh_readers = ds_hash_create(ds_hwlock_comp, ds_hwlock_hash,
					ds_hwlock_nbuckets, TRUE, FALSE);
    (lock)->dh_rwaiters = 0;
    (lock)->dh_wwaiters = 0;
    ds_lock_release(&((lock)->dh_wlock));
}

/* Get the calling thread a read lock, blocking if necessary. Assumes
   that the lock structure itself has been handed in write-locked, and
   will be handed back the same way.  */

void
_ds_hwlock_read_slow(lock,me)
    ds_hwlock_t *lock;
    ds_thread_t me;
{

    if (ds_hwlock_debug)
	ds_thread_log(0, "Slow-readlocking hwlock %p",lock);

    while (lock->dh_wpo) { /* wait for whoever to give up their write lock */
	void *wait_event = READ_EVENT(lock);
	(lock->dh_rwaiters)++;
	ds_lock_release(&(lock->dh_wlock));
	ds_thread_wait(wait_event);
	ds_lock_write(&(lock->dh_wlock));
	(lock->dh_rwaiters--);
    }

    if (lock->dh_rc == 0) { /* no one has a lock */
	lock->dh_rpo = me;
	lock->dh_rc = 1;
	lock->dh_rpc = 1;
    }
    else if (lock->dh_rpo == me)
	(lock->dh_rpc)++;
    else {
	struct rl_info rli, *rlip;

	rli.reader = me;

	/* see if this thread has any read locks already */
	rlip = ds_hash_member(lock->dh_readers,&rli);
    
	if (rlip == NULL) {	/* we have no locks */

	    (lock->dh_rc)++;

	    if (lock->dh_rpo == NULL) { /* we can become the rpo */
		lock->dh_rpo = me;
		lock->dh_rpc = 1;
	    }

	    else { /* we have to add ourselves to the hash table */
		ALLOC(rlip, struct rl_info);
		rlip->reader = me;
		rlip->count = 1;
		ds_hash_insert(lock->dh_readers, rlip);
	    }
      
	}
	else { /* we've read-locked this object already */
      
	    (rlip->count)++;
      
	    if (lock->dh_rpo == NULL) { 
		/* we can become the rpo, remove hash entry */
		lock->dh_rpo = me;
		lock->dh_rpc = rlip->count;
		ds_hash_remove(lock->dh_readers, rlip);
		FREE(rlip);
	    }
	}
    }

    if (ds_hwlock_debug)
	ds_thread_log(0, "Slow-readlocked hwlock %p",lock);

}

/* Release a read lock held by the calling thread.  Assumes that the
   lock structure itself has been handed in write-locked, and will be
   handed back the same way.  */

void _ds_hwlock_release_read_slow(lock, me)
    ds_hwlock_t *lock;
    ds_thread_t me;
{

    if (ds_hwlock_debug)
	ds_thread_log(0, "Slow-releasing readlock on hwlock %p",lock);

    if (lock->dh_rpo == me) {
	(lock->dh_rpc)--;
	if (lock->dh_rpc == 0) {
	    (lock->dh_rc)--;
	    lock->dh_rpo = NULL;
	}
    }
    else { /* we are not rpo, see if we have a read lock */
	struct rl_info rli, *rlip;

	rli.reader = me;
    
	rlip = ds_hash_member(lock->dh_readers,&rli);

	if (rlip == NULL) /* you can't release a lock you don't hold! */
	    ODY_ASSERT(rlip);
    
	(rlip->count)--;
    
	if (rlip->count > 0) {
	    /* see if we can become the rpo */
	    if (lock->dh_rpo == NULL) {
		lock->dh_rpo = me;
		lock->dh_rpc = rlip->count;
		ds_hash_remove(lock->dh_readers, rlip);
	    }
	}
	else { /* we released our last read-lock */
	    ds_hash_remove(lock->dh_readers, rlip);
	    (lock->dh_rc)--;
	}
    }

    /* if we don't have a read lock, and not more than one other thread
       has a read lock, then some write waiter could be woken up */

    if ((lock->dh_wwaiters > 0) && (lock->dh_rpo != me) && (lock->dh_rc <= 1))
	ds_thread_wakeup(WRITE_EVENT(lock));
}

/* Get the calling thread a write lock, blocking if necessary. Assumes
   that the lock structure itself has been handed in write-locked, and
   will be handed back the same way. */

void _ds_hwlock_write_slow(lock, me)
    ds_hwlock_t *lock;
    ds_thread_t me;
{

    if (ds_hwlock_debug)
	ds_thread_log(0, "Slow-writelocking hwlock %p", lock);

    while (1) {
	void *wait_event = WRITE_EVENT(lock);
    
	if (lock->dh_wpo == me) { /* I already have a write lock */
	    (lock->dh_wpc)++;
	    break;
	}

	else if (lock->dh_wpo != NULL) 
	    ; /* someone else has a write lock, we block */

	/* there are no readers or I am the rpo and sole reader */
	else if ((lock->dh_rc == 0) ||
		 ((lock->dh_rc == 1) && (lock->dh_rpo == me))) {
	    lock->dh_wpo = me;
	    lock->dh_wpc = 1;
	    break;
	}

	/* I might be the sole reader even if I'm not the rpo */
	else if ((lock->dh_rc == 1) && (lock->dh_rpo == NULL)) {
	    struct rl_info rli, *rlip;
	    rli.reader = me;
	    rlip = ds_hash_member(lock->dh_readers, &rli);
	    if (rlip) {		/* we are the sole reader */
		/* become the rpo */
		lock->dh_rpo = me;
		lock->dh_rpc = rlip->count;
		ds_hash_remove(lock->dh_readers,rlip);
		FREE(rlip);

		/* and grab the write lock */
		lock->dh_wpo = me;
		lock->dh_wpc = 1;
		break;
	    }
	}

	/* We couldn't grab the write lock: block, wait for the WRITE_EVENT,
	   and then try again */

	wait_event = WRITE_EVENT(lock);
	(lock->dh_wwaiters)++;
	ds_lock_release(&(lock->dh_wlock));
	ds_thread_wait(wait_event);
	ds_lock_write(&(lock->dh_wlock));
	(lock->dh_wwaiters)--;
    }

    if (ds_hwlock_debug)
	ds_thread_log(0, "Slow-writelocked hwlock %p", lock);
}






