#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Viceroy/Warden heavyweight lock abstraction. Builds on top of ds_thread
 * and ds_hash.
 */

#ifndef _DS_HWLOCK_H_
#define _DS_HWLOCK_H_

#define TRUE  1
#define FALSE 0

typedef struct ds_hwlock {

    ds_lock_t dh_wlock;    /* to write lock the struct ds_hwlock itself */
    
    ds_thread_t dh_rpo;	   /* the read principal owner, can be NULL */
    int dh_rpc;		   /* number of times rpo has read-locked this lock */

    ds_thread_t dh_wpo;	   /* the write principal owner, NULL if no writer */
    int dh_wpc;		   /* number of times wpo has write-locked this lock */

    int dh_rc;		   /* number of threads that have read locks */
    ds_hash_t *dh_readers; /* the threads (except rpo) that have read locks,
			      and how many times each has read-locked */

    int dh_rwaiters;	   /* number of blocked would-be readers */
    int dh_wwaiters;	   /* number of blocked would-be writers */
}
ds_hwlock_t;

/* events that blocked readers/writers wait for */

#define READ_EVENT(lock) ((void *)(&((lock)->dh_rwaiters)))
#define WRITE_EVENT(lock) ((void *)(&((lock)->dh_wwaiters)))

/* These functions are NOT to be called directly; use the macros instead */

extern void _ds_hwlock_read_slow(ds_hwlock_t *lock, ds_thread_t me);
extern void _ds_hwlock_release_read_slow(ds_hwlock_t *lock, ds_thread_t me);
extern void _ds_hwlock_write_slow(ds_hwlock_t *lock, ds_thread_t me);

extern  int ds_hwlock_debug;

#ifdef ODYDEBUG
#define ODY_ASSERT(x) ASSERT(x)
#else
#define ODY_ASSERT(x)
#endif

/* Macros and functions to use for locking/unlocking a heavyweight lock */

extern void ds_hwlock_init(ds_hwlock_t *lock);

#define ds_hwlock_read(lock)   \
do {  \
    register ds_thread_t _dh_me;  \
	      \
    if (ds_hwlock_debug)  \
	ds_thread_log(0, "Readlocking hwlock %p", lock);  \
			  \
    ds_lock_write(&((lock)->dh_wlock));  \
    _dh_me = ds_thread_current();  \
 \
    /* can't read lock if someone has a write lock */ \
    if ((lock)->dh_wpo == NULL) { \
	if ((lock)->dh_rc == 0) { /* no one has a lock */  \
	    (lock)->dh_rpo = _dh_me;  \
	    (lock)->dh_rc = 1;  \
	    (lock)->dh_rpc = 1;  \
	}  \
	else if ((lock)->dh_rpo == _dh_me) { /* I am the rpo */  \
	    ((lock)->dh_rpc)++;  \
	}  \
	else  \
	    _ds_hwlock_read_slow((lock),_dh_me);  \
    }  \
    else  \
	_ds_hwlock_read_slow((lock),_dh_me);  \
    ds_lock_release(&((lock)->dh_wlock));  \
} while(0)

#define ds_hwlock_release_read(lock)  \
do {  \
    register ds_thread_t _dh_me;  \
	      \
    if (ds_hwlock_debug)  \
	ds_thread_log(0, "Releasing readlock on hwlock %p", lock);  \
			  \
    ds_lock_write(&((lock)->dh_wlock));  \
    _dh_me = ds_thread_current();  \
				      \
    /* Are we the principal owner, and can we drop our lock without  \
       needing to wake up anyone? */  \
						 \
    if ( ((lock)->dh_rpo == _dh_me)  &&  \
	(((lock)->dh_wwaiters == 0) || ((lock)->dh_rpc > 1) ||  \
	 ((lock)->dh_rc > 2))) {  \
 \
	/* did we just give up our last read-lock? */  \
	if (--((lock)->dh_rpc) == 0) { \
	    ((lock)->dh_rc)--;  \
	    (lock)->dh_rpo = NULL;  \
	}  \
    }  \
    else  \
	_ds_hwlock_release_read_slow((lock),_dh_me);  \
    ds_lock_release(&((lock)->dh_wlock)); \
\
} while(0)

#define ds_hwlock_write(lock)  \
do {  \
    register ds_thread_t _dh_me;  \
	      \
    if (ds_hwlock_debug)  \
	ds_thread_log(0, "Writelocking hwlock %p", lock);  \
			  \
    ds_lock_write(&((lock)->dh_wlock));  \
				  \
    _dh_me = ds_thread_current();  \
					  \
    if ((lock)->dh_wpo == _dh_me) /* I already have a write lock */  \
	((lock)->dh_wpc)++;  \
						      \
    /* there are no writers, and either no readers or we are the rpo and \
       sole reader */  \
    else if (((lock)->dh_wpo == NULL) &&  \
	     (((lock)->dh_rc == 0) || (((lock)->dh_rc == 1) &&  \
				       ((lock)->dh_rpo == _dh_me)))) {  \
	(lock)->dh_wpo = _dh_me;  \
	(lock)->dh_wpc = 1;  \
    }  \
    else  \
	_ds_hwlock_write_slow((lock),_dh_me);  \
																	        \
    ds_lock_release(&((lock)->dh_wlock));  \
} while(0)

#define ds_hwlock_release_write(lock)  \
do {  \
    ds_thread_t _dh_me; \
    ds_lock_write(&((lock)->dh_wlock));  \
	      \
    if (ds_hwlock_debug)  \
	ds_thread_log(0, "Releasing writelock on hwlock %p", lock);  \
 \
    _dh_me = ds_thread_current(); \
 \
    if ((lock)->dh_wpo == _dh_me) { /* do we hold a write lock? */  \
	((lock)->dh_wpc)--;  \
	if ((lock)->dh_wpc == 0) {  \
	    (lock)->dh_wpo = NULL;  \
	    if ((lock)->dh_rwaiters > 0) /* wake up any waiting readers */  \
		ds_thread_wakeup(READ_EVENT(lock));  \
	    else if ((lock)->dh_wwaiters > 0) /* wake up any waiting writers */ \
 		ds_thread_wakeup(WRITE_EVENT(lock));  \
	}  \
    } \
    else \
	ODY_ASSERT((lock)->dh_wpo == _dh_me); \
     \
    ds_lock_release(&((lock)->dh_wlock));  \
} while(0)

#endif /* _DS_HWLOCK_H */ \
