#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
   ds_rrlist.c: implementation of ds_rrlist's.
*/

#include "include/odytypes.h"

#include "ds_list.h"
#include "ds_hash.h"
#include "ds_rrlist.h"
#include "ds_rrlist.private.h"

/* magic numbers */

const magic_t ds_request_magic = 0xb696e46d;
const magic_t ds_rrlist_magic = 0x2b5407cf;
const magic_t ds_rrlist_reqtab_elt_magic = 0x97d66ff5;
const magic_t ds_rrlist_pidtab_elt_magic = 0x3fad469a;


static long       lastreqid=0;      /* last request id assigned */
static int        inited=0;
static ds_hash_t *reqtab=NULL;      /* request=>rrlist mapping */
static ds_hash_t *pidtab=NULL;      /* request=>rrlist mapping */

/* static functions: hash and comparison functions */

/* Request table */

static long
hash_reqtab_elt(void *anon) {
    ds_rrlist_reqtab_elt_t *elt = anon;
    ASSERT(DS_RRLIST_REQTAB_ELT_VALID(elt));
    return elt->reqid;
}

static long
comp_reqtab_elt(void *a1, void *a2) {
    ds_rrlist_reqtab_elt_t *e1 = a1;
    ds_rrlist_reqtab_elt_t *e2 = a2;
    ASSERT(DS_RRLIST_REQTAB_ELT_VALID(e1));
    ASSERT(DS_RRLIST_REQTAB_ELT_VALID(e2));
    return (e1->reqid - e2->reqid);
}

/* Process-id table */

static long
hash_pidtab_elt(void *anon) {
    ds_rrlist_pidtab_elt_t *elt = anon;
    ASSERT(DS_RRLIST_PIDTAB_ELT_VALID(elt));
    return elt->pid;
}

static long
comp_pidtab_elt(void *a1, void *a2) {
    ds_rrlist_pidtab_elt_t *e1 = a1;
    ds_rrlist_pidtab_elt_t *e2 = a2;
    ASSERT(DS_RRLIST_PIDTAB_ELT_VALID(e1));
    ASSERT(DS_RRLIST_PIDTAB_ELT_VALID(e2));
    return (e1->pid - e2->pid);
}

/* 
   pids and reqids are unique in each structure by invariant,
   so they are sufficient for distinction
*/

static long
hashpid(void *anon) {
    ds_request_t *request = anon;
    ASSERT(DS_REQUEST_VALID(request));
    return request->pid;
}

static long
hashreq(void *anon) {
    ds_request_t *request = anon;
    ASSERT(DS_REQUEST_VALID(request));
    return request->reqid;
}

static long
comppid(void *a1, void *a2) {
    ds_request_t *r1 = a1;
    ds_request_t *r2 = a2;

    ASSERT(DS_REQUEST_VALID(r1));
    ASSERT(DS_REQUEST_VALID(r2));

    return (r1->pid - r2->pid);
}

static long
compreq(void *a1, void *a2) {
    ds_request_t *r1 = a1;
    ds_request_t *r2 = a2;

    ASSERT(DS_REQUEST_VALID(r1));
    ASSERT(DS_REQUEST_VALID(r2));

    return (r1->reqid - r2->reqid);
}

/* 
   lists can contain many elements with the same low/high values.  As
   long as they are ordered correctly with respect to those values, we
   can use the pointers themselves to distinguish equal bounds values.
   This is necessary since otherwise equality tests for ds_list_remove
   calls would be too weak.
*/

static long
complow(void *a1, void *a2) {
    ds_request_t *r1 = a1;
    ds_request_t *r2 = a2;
    long result;

    ASSERT(DS_REQUEST_VALID(r1));
    ASSERT(DS_REQUEST_VALID(r2));

    result = r1->low - r2->low;
    if (!result)
	result = (r1->reqid - r2->reqid);
    return result;
}

static long
comphigh(void *a1, void *a2) {
    ds_request_t *r1 = a1;
    ds_request_t *r2 = a2;
    long result;

    ASSERT(DS_REQUEST_VALID(r1));
    ASSERT(DS_REQUEST_VALID(r2));

    result = r1->high - r2->high;
    if (!result)
	result = (r1->reqid - r2->reqid);
    return result;
}

/* Some helper functions to simplify the calls below. */
static inline void
remove_request(ds_request_t *reqp, ds_rrlist_t *list)
{
    ds_rrlist_reqtab_elt_t *rte;
    ds_rrlist_reqtab_elt_t dummy_rte;
    ds_rrlist_pidtab_elt_t *pte;
    ds_rrlist_pidtab_elt_t dummy_pte;
    

    /* remove from reqtab */
    dummy_rte.magic = ds_rrlist_reqtab_elt_magic;
    dummy_rte.reqid = reqp->reqid;
    ASSERT((rte = ds_hash_remove(reqtab,&dummy_rte)) != NULL);
    DS_RRLIST_REQTAB_ELT_DESTROY(rte);

    /* Find the appropriate pidtab entry */
    dummy_pte.magic = ds_rrlist_pidtab_elt_magic;
    dummy_pte.pid = reqp->pid;
    ASSERT((pte = ds_hash_member(pidtab, &dummy_pte)) != NULL);
    /* Remove request from that pid's list of requests */
    ASSERT(ds_list_remove(pte->reqlist, reqp) != NULL);

    /* remove from rrlist */
    ASSERT(ds_hash_remove(list->pidhash,reqp) != NULL);
    ASSERT(ds_hash_remove(list->reqhash,reqp) != NULL);
    ASSERT(ds_list_remove(list->less,reqp) != NULL);
    ASSERT(ds_list_remove(list->greater,reqp) != NULL);
}

static inline void
insert_request(ds_request_t *reqp, ds_rrlist_t *list)
{
    ds_rrlist_reqtab_elt_t *rte;
    ds_rrlist_pidtab_elt_t *pte;
    ds_rrlist_pidtab_elt_t dummy;

    /* Generate the request id */
    lastreqid++;
    reqp->reqid = lastreqid;

    /* Insert it into the reqid->req table */
    DS_RRLIST_REQTAB_ELT_CREATE(rte,lastreqid,list);
    ASSERT(ds_hash_insert(reqtab,rte) != NULL);

    /* Insert it into the pid->list-of-req table */

    /* See if we have any other requests for this pid. */
    dummy.magic = ds_rrlist_pidtab_elt_magic;
    dummy.pid = reqp->pid;
    pte = ds_hash_member(pidtab,&dummy);

    /* If we don't, create and insert an entry for this pid. */
    if (!pte) {
	ALLOC(pte, ds_rrlist_pidtab_elt_t);
	pte->magic = ds_rrlist_pidtab_elt_magic;
	pte->pid = reqp->pid;
	pte->reqlist = ds_list_create(compreq, TRUE, FALSE);
	ASSERT(ds_hash_insert(pidtab, pte) != NULL);
    }

    /* Do the insert */
    ASSERT(ds_list_insert(pte->reqlist, reqp) != NULL);

    /* Insert it in the appropriate list. */
    ASSERT(ds_hash_insert(list->pidhash,reqp) != NULL);
    ASSERT(ds_hash_insert(list->reqhash,reqp) != NULL);
    ASSERT(ds_list_insert(list->less,reqp) != NULL);
    ASSERT(ds_list_insert(list->greater,reqp) != NULL);
}


/* Observers */

bool
ds_rrlist_valid(ds_rrlist_t *l) {
    if (DS_RRLIST_VALID(l))
	return TRUE;
    else
	return FALSE;
}

int
ds_rrlist_value(ds_rrlist_t *l) {
    ASSERT(DS_RRLIST_VALID(l));
    return l->value;
}

/* Mutators */

ds_rrlist_t *
ds_rrlist_create(int value) {
    ds_rrlist_t *result;
    
    /* Is the universal hashtab setup yet? */
    /* Check here b/c it is cheap. */
    /* XXX change to inited.  Add in creation of pidtab */
    if (!inited) {
	/* not yet, do it. */
	reqtab = ds_hash_create(comp_reqtab_elt,
				hash_reqtab_elt,
				DS_RRLIST_HASHSIZE,
				TRUE,FALSE);

	pidtab = ds_hash_create(comp_pidtab_elt,
				hash_pidtab_elt,
				DS_RRLIST_HASHSIZE,
				TRUE,FALSE);
	inited = 1;
    }

    ALLOC(result,ds_rrlist_t);
    result->magic = ds_rrlist_magic;
    result->value = value;
    result->pidhash = ds_hash_create(comppid,hashpid,DS_RRLIST_HASHSIZE,
				     TRUE,FALSE);
    result->reqhash = ds_hash_create(compreq,hashreq,DS_RRLIST_HASHSIZE,
				     TRUE,FALSE);
    result->less = ds_list_create(complow,TRUE,FALSE);
    result->greater = ds_list_create(comphigh,TRUE,FALSE);

    return result;
}

void
ds_rrlist_destroy(ds_rrlist_t *l) {
    /* these are all safe-destroy, so they must be empty */
    ASSERT(DS_RRLIST_VALID(l));
    l->magic = 0;
    l->value = 0;
    ds_hash_destroy(l->pidhash);
    ds_hash_destroy(l->reqhash);
    ds_list_destroy(l->less);
    ds_list_destroy(l->greater);
    FREE(l);
}

ds_rrlist_return_t
ds_rrlist_request(ds_rrlist_t *l, ds_request_t *r, int *value,
		  ds_request_t **old_req) {
    ds_rrlist_return_t      result = DS_RRLIST_SUCCESS;

    ASSERT(DS_RRLIST_VALID(l));
    ASSERT(DS_REQUEST_VALID(r));
    ASSERT(value);
    ASSERT(old_req);

    r->reqid = 0;

    /* do we have a current request for this pid/resource pair? */
    *old_req = ds_hash_member(l->pidhash,r);
    if (*old_req != NULL) {
	result |= DS_RRLIST_DUPLICATE;
	remove_request(*old_req, l);
    }
    
    /* Is the request out-of-bounds already? */
    if ((r->low > l->value) || (r->high < l->value)) {
	result |= DS_RRLIST_OUTOFWINDOW;
	*value = l->value;
    } else {
	insert_request(r, l);
    }
    return result;
}

ds_rrlist_return_t
ds_rrlist_cancel(int reqid, ds_request_t **req) {
    ds_rrlist_return_t      result = DS_RRLIST_SUCCESS;
    ds_request_t            req_dummy;
    ds_rrlist_t            *l;
    ds_rrlist_reqtab_elt_t *rte;
    ds_rrlist_reqtab_elt_t  rte_dummy;

    ASSERT(req);

    /* First, we need to get the list this request is on */
    rte_dummy.magic = ds_rrlist_reqtab_elt_magic;
    rte_dummy.reqid = reqid;
    rte = ds_hash_member(reqtab,&rte_dummy);

    if (rte == NULL) {      /* didn't find the list */
	result |= DS_RRLIST_NOSUCHREQ;
	*req = NULL;
    } else {
	/* we have the list, continue */
	l = rte->list;
	ASSERT(DS_RRLIST_VALID(l));
	/* See if we can find the item to remove. */
	req_dummy.magic = ds_request_magic;
	req_dummy.reqid = reqid;
	*req = ds_hash_member(l->reqhash,&req_dummy);
	if (*req) {
	    /* found it, remove it */
	    ASSERT(DS_REQUEST_VALID(*req));
	    remove_request(*req, l);
	    /* Caller of cancel must free contents */
	} else {
	    result |= DS_RRLIST_NOSUCHREQ;
	}
    }
    return result;
}

ds_rrlist_return_t
ds_rrlist_purge(pid_t pid)
{
    ds_rrlist_pidtab_elt_t *pte;
    ds_rrlist_pidtab_elt_t dummy_pte;
    ds_rrlist_reqtab_elt_t *rte;
    ds_rrlist_reqtab_elt_t dummy_rte;
    ds_rrlist_return_t     result=DS_RRLIST_SUCCESS;
    ds_request_t           *reqp;
    
    dummy_pte.magic = ds_rrlist_pidtab_elt_magic;
    dummy_pte.pid = pid;
    pte = ds_hash_member(pidtab, &dummy_pte);

    if (!pte) {
	result |= DS_RRLIST_NOSUCHPID;
	return result;
    }
    
    while ((reqp = ds_list_first(pte->reqlist)) != NULL) {
	/* Find it's list.  It has to be there */
	dummy_rte.magic = ds_rrlist_reqtab_elt_magic;
	dummy_rte.reqid = reqp->reqid;
	ASSERT((rte = ds_hash_member(reqtab, &dummy_rte)) != NULL);
	remove_request(reqp, rte->list);
	if (reqp->argp) {
	    (reqp->argp->free_fn)(reqp->argp->argval);
	    FREE(reqp->argp);
	}
	FREE(reqp);
    }
    /* Remove this pid's entry from the pidtab */
    ASSERT(ds_hash_remove(pidtab, pte) != NULL);
    ds_list_destroy(pte->reqlist);
    FREE(pte);
    return result;
}

ds_rrlist_return_t
ds_rrlist_set_value(ds_rrlist_t *l, int newval, ds_list_t **to_notify) {
    ds_rrlist_return_t      result = DS_RRLIST_SUCCESS;
    ds_request_t           *req = NULL;

    ASSERT(DS_RRLIST_VALID(l));
    ASSERT(to_notify);

    /* zero the list pointer */
    *to_notify = NULL;

    /* 
       check quickly if we have to do anything, and get out if not.
       (the common case)
    */
    if (newval > l->value) {
	req = (ds_request_t *) ds_list_first(l->greater);
	if ((req == NULL) || (newval <= req->high)) {
	    l->value = newval;
	    return result;
	}
    } else {
	req = (ds_request_t *) ds_list_last(l->less);
	if ((req == NULL) || (newval >= req->low)) {
	    l->value = newval;
	    return result;
	}
    }

    /* 
       there is at least one bound that has been exceeded: req
       currently points to it.  l->value has not yet been set,
       and the to_notify list must be constructed
    */

    *to_notify = ds_list_create(NULL,FALSE,FALSE);

    if (newval > l->value) {
	/* check upper bounds */
	while (req != NULL) {
	    ASSERT(DS_REQUEST_VALID(req));
	    if (req->high >= newval) {
		break;
	    } else {
		/* Sanity check: ensure it is the first in greater */
		ASSERT(ds_list_first(l->greater) == req);
		/* remove it */
		remove_request(req, l);
		/* insert into result list */
		ASSERT(ds_list_insert(*to_notify,req) == req);
		/* step to next pointer in list. */
		req = (ds_request_t *) ds_list_first(l->greater);
	    }
	}
    } else {
	/* check lower bounds */
	while (req != NULL) {
	    ASSERT(DS_REQUEST_VALID(req));
	    if (req->low <= newval)  {
		break;
	    } else {
		/* Sanity check: ensure it is the last in less */
		ASSERT(ds_list_last(l->less) == req);
		/* remove it */
		remove_request(req, l);
		/* insert into result list */
		ASSERT(ds_list_insert(*to_notify,req) == req);
		/* step to next pointer in list. */
		req = (ds_request_t *) ds_list_last(l->less);
	    }
	}
    }
    l->value = newval;
    return result;
}

static void
ds_request_print(ds_request_t *req, FILE *f) {
    ASSERT(DS_REQUEST_VALID(req));
    fprintf(f,"pid %d\treqid %d\tlow %d\thigh %d\n",
	    req->pid, req->reqid, req->low, req->high);
}

void
ds_rrlist_dump(ds_rrlist_t *l, FILE *f, char *name) {
    ds_hash_iter_t *hi;
    ds_list_iter_t *li;
    ds_request_t   *req;

    ASSERT(DS_RRLIST_VALID(l));
    ASSERT(f);
    ASSERT(name);
    
    fprintf(f,"Printing resource request list %s\n",name);
    fprintf(f,"========================================================\n");
    fprintf(f,"--------value = %ld\n",l->value);
    fprintf(f,"--------last requid assigned = %ld\n",lastreqid);
    fprintf(f,"--------pidhash table\n");

    hi = ds_hash_iter_create(l->pidhash);
    while ((req = (ds_request_t *) ds_hash_iter_next(hi)) != NULL) {
	ds_request_print(req,f);
    }
    ds_hash_iter_destroy(hi);

    fprintf(f,"--------reqhash table\n");

    hi = ds_hash_iter_create(l->reqhash);
    while ((req = (ds_request_t *) ds_hash_iter_next(hi)) != NULL) {
	ds_request_print(req,f);
    }
    ds_hash_iter_destroy(hi);

    fprintf(f,"--------lower bound list\n");

    li = ds_list_iter_create(l->less);
    while ((req = (ds_request_t *) ds_list_iter_next(li)) != NULL) {
	ds_request_print(req,f);
    }
    ds_list_iter_destroy(li);

    fprintf(f,"--------upper bound list\n");

    li = ds_list_iter_create(l->greater);
    while ((req = (ds_request_t *) ds_list_iter_next(li)) != NULL) {
	ds_request_print(req,f);
    }
    ds_list_iter_destroy(li);

    fprintf(f,"\n\n");
}

int ds_rrlist_get_reqid ()
{
    /* Asmittedly a hack - JNF */
    return (++lastreqid);
}
