#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * ds_safeq.c: implementation of ds_safeq_t
 */

#include <stdlib.h>
#include "include/odytypes.h"

#include "ds_list.h"
#include "ds_thread.h"
#include "ds_safeq.h"
#include "ds_safeq.private.h"

/* magic numbers */

const magic_t ds_safeq_magic = 1574524372;

bool
ds_safeq_valid(ds_safeq_t *q) {
    if (DS_SAFEQ_VALID(q))
	return TRUE;
    else 
	return FALSE;
}

int
ds_safeq_count(ds_safeq_t *q)
{
    int count;

    ds_lock_read(&q->sq_lock);
    count = ds_list_count(q->sq_list);
    ds_lock_release(&q->sq_lock);

    return count;
}

ds_safeq_t *
ds_safeq_create() 
{
    ds_safeq_t *result;

    ALLOC(result,ds_safeq_t);
    result->sq_magic = ds_safeq_magic;
    result->sq_dying = 0;
    result->sq_waiters = 0;
    ds_lock_init(&result->sq_lock);
    result->sq_list = ds_list_create(NULL, FALSE, TRUE);
    ASSERT(result->sq_list);
    
    return result;
}

void
ds_safeq_destroy(q)
    ds_safeq_t *q;
{
    ASSERT(DS_SAFEQ_VALID(q));

    ds_lock_write(&q->sq_lock);

    /* Flag the safeq as dying */
    q->sq_dying = 1;

    /* Let all readers who might care get a crack at the entries */
    while (ds_list_count(q->sq_list)) {
	ds_lock_release(&q->sq_lock);
	ds_thread_wait(&q->sq_MTsignal);
	ds_lock_write(&q->sq_lock);
    }

    /* Signal any remaining waiters that it's time to go. */
    while (q->sq_waiters) {
	ds_lock_release(&q->sq_lock);
	ds_thread_schedule(&q->sq_signal);
	ds_lock_write(&q->sq_lock);
    }
    
    /* Now, go ahead and destroy it */
    q->sq_magic = 0;
    ds_list_destroy(q->sq_list);
    /* XXX 
     * do we want to leave the lock locked? 
     * At the moment, I think not, just so that anyone else will know
     * that they've goofed up.
     */
    ds_lock_release(&q->sq_lock);
    FREE(q);
}

int
ds_safeq_enq(q, i)
    ds_safeq_t *q;
    void       *i;
{
    int result;

    ASSERT(DS_SAFEQ_VALID(q));
    ds_lock_write(&q->sq_lock);
    if (!q->sq_dying) { 
	ASSERT(ds_list_append(q->sq_list,i) == i);
	if (q->sq_waiters) {
	    ds_thread_wakeup(&q->sq_signal);
	}
	result = 0;
    } else {
	result = 1;
    }
    ds_lock_release(&q->sq_lock);
    return result;
}

void *
ds_safeq_deq(q)
    ds_safeq_t *q;
{
    void *result;

    ASSERT(DS_SAFEQ_VALID(q));
    ds_lock_write(&q->sq_lock);
    while (1) {
	if (!ds_list_count(q->sq_list)) {
	    if (q->sq_dying) {
		ds_lock_release(&q->sq_lock);
		return NULL;
	    }
	    /* Need to wait */
	    q->sq_waiters++;
	    ds_lock_release(&q->sq_lock);
	    ds_thread_wait(&q->sq_signal);
	    ds_lock_write(&q->sq_lock);
	    q->sq_waiters--;
	    continue; /* Try again */
	} else {
	    result = ds_list_get_first(q->sq_list);
	    ASSERT(result);
	    if (!ds_list_count(q->sq_list)) {
		ds_thread_wakeup(&q->sq_MTsignal);
	    }
	    ds_lock_release(&q->sq_lock);
	    return result;
	}
    }
}
