#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <stdarg.h>

#include "include/odytypes.h"

#include "ds_log.h"
#include "ds_thread.h"
#include "ds_thread.private.h"

static int tnum;
static int stsz;
static int logl;

#ifndef __GNUC__
#define __inline__
#endif

int ds_lock_debug =0;
int ds_event_debug =0;

static __inline__ char *
generate_tname(class)
    char *class;
{
    char *result;

    NALLOC(result, char, strlen(class)+10);
    sprintf(result, "%s:%d", class, tnum++);
    return result;
}

static __inline__ struct threadinfo_t *
prologue_ti(name, body, param, loglevel, tid)
    char        *name;
    PFIC        body;
    void        *param;
    int         loglevel;
    ds_thread_t tid;
{
    struct threadinfo_t *ti;

    ALLOC(ti, struct threadinfo_t);
    ti->ti_body = body;
    ti->ti_arg = param;
    ti->ti_name = name;
    ti->ti_ptid = tid;
    ti->ti_log = ds_log_create(loglevel, stdout, 100, ti->ti_name);
    ASSERT(ti->ti_log);
    return (ti);
}

static __inline__ void
postlogue_ti(ti)
    struct threadinfo_t *ti;
{
    ds_thread_t tid;

    ASSERT(LWP_CurrentProcess(&tid) == LWP_SUCCESS);
    ti->ti_tid = tid;
    ASSERT(!LWP_NewRock(TINFO_ROCK, (char*)ti));
}

void
ds_thread_init(stacksz, loglevel)
    int stacksz;
    int loglevel;
{
    ds_thread_t         main_thread;
    char                *name;
    struct threadinfo_t *ti;

    /* Inisialize LWP */
    ASSERT(LWP_Init(LWP_VERSION, LWP_NORMAL_PRIORITY, &main_thread)
	   == LWP_SUCCESS);

    /* IOMGR init */
    ASSERT(IOMGR_Initialize() == LWP_SUCCESS);

    /* Set first threadid */
    tnum = 0;
    
    /* Save the stacksz */
    stsz = stacksz;

    /* Set up global loglevel */
    logl=loglevel;

    /* Stash the thread info under a rock. */
    name = generate_tname(mainthread);
    ti = prologue_ti(name, NULL, NULL, loglevel, main_thread);
    postlogue_ti(ti);

    /* Brag about it. */
    ds_thread_log(10,"threads initialized.  Main thread = %p", main_thread);
}

static int
thread_prologue(arg)
    void *arg;
{
    int                 result;
    struct threadinfo_t *ti = (struct threadinfo_t *)arg;

    /* Set up the threadinfo structure, and then call the body. */
    postlogue_ti(ti);
    ds_thread_log(100, "Thread %s (%p) created", ti->ti_name, ti->ti_tid);

    /* Execute the thread body. */
    result = (ti->ti_body)(ti->ti_arg);

    /* Deallocate the threadinfo bits */
    ds_thread_log(100, "Thread %s (%p) finished", ti->ti_name, ti->ti_tid);
    ds_log_destroy(ti->ti_log);
    FREE(ti->ti_name);
    FREE(ti);
    
    return result;
}
    

ds_thread_t
ds_thread_create(ep, param, prio, class, loglevel)
    PFIC        ep;
    void        *param;
    int         prio;
    char        *class;
    int         loglevel;
{
    ds_thread_t         ptid;
    ds_thread_t         tid;
    int                 result;
    char                *name;
    struct threadinfo_t *ti;

    /* Set up the threadinfo struct for the prologue */
    name = generate_tname(class);
    ptid = ds_thread_current();
    ti = prologue_ti(name, ep, param, loglevel, ptid);

    result = LWP_CreateProcess(thread_prologue, stsz, prio, (char *)ti, 
			       name, &tid);
    if (result) {
	FREE(name);
	return NULL;
    }
    
    return tid;
}

ds_thread_t
ds_thread_current()
{
    ds_thread_t tid;

    ASSERT(LWP_CurrentProcess(&tid) == LWP_SUCCESS);
    return tid;
}

const char *
ds_thread_name()
{
    return LWP_Name();
}

void
ds_thread_log(level, fmt)
    int level;
    char *fmt;
{
    va_list ap;
    struct threadinfo_t *ti;

    ASSERT(LWP_GetRock(TINFO_ROCK, (char **)&ti) == LWP_SUCCESS);
    
    va_start(ap, fmt);
    if (level <= logl) {
	ds_log_doprint(ti->ti_log, 0, fmt, ap);
    } else {
	ds_log_doprint(ti->ti_log, level, fmt, ap);
    }
    va_end(ap);
}

void
ds_thread_lflush()
{
    struct threadinfo_t *ti;

    ASSERT(LWP_GetRock(TINFO_ROCK, (char **)&ti) == LWP_SUCCESS);
    ds_log_flush(ti->ti_log);
}

int
ds_thread_setll(int level)
{
    struct threadinfo_t *ti;

    ASSERT(LWP_GetRock(TINFO_ROCK, (char **)&ti) == LWP_SUCCESS);
    return(ds_log_setlevel(ti->ti_log, level));
}
	
