#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Viceroy/Warden thread abstraction.  Builds on top of the 
 * LWP thread abstraction.
 */

#ifndef _DS_THREAD_H_
#define _DS_THREAD_H_

#include <stdarg.h>
#include <lwp/lwp.h>
#include <lwp/lock.h>
#include "ds_log.h"

#define DS_THREAD_SUCCESS    LWP_SUCCESS
#define DS_THREAD_ENOWAIT    LWP_ENOWAIT

/*
 * There are three priorities we care about:
 * 
 * NORMAL: most things use normal -- they are in some path that determines
 * response times
 *
 * CRITICAL: things that *have* to happen ASAP.   I'm not convinced anything
 * will need to go here, so think before you use it.
 *
 * DAEMON: background processing, periodic events that are not time-critical.
 */

#define DS_THREAD_NORMAL   LWP_NORMAL_PRIORITY
#define DS_THREAD_CRITICAL LWP_MAX_PRIORITY
#define DS_THREAD_DAEMON   LWP_NORMAL_PRIORITY-1

#define DS_THREAD_STACKSZ  32*1024

#define ds_thread_t  PROCESS
#define ds_lock_t    struct Lock

/*
 * Initialize ds_threads package.  Initializes underlying threads
 * package, sets up the log (which will be stdio) and set up the
 * static info that we need.  Passed the stack size to use for
 * threads, and the loglevel to use.  We don't use ds_log's idea of
 * loglevel, since if we aren't going to use inlining for the log
 * message function.  
 */
extern void   ds_thread_init __P((int stacksize, int loglevel));

/*
 * Create a thread.  Given a pointer to a function to run as the
 * thread body, a pointer to an argument structure to pass to the
 * thread body, a priority, and a string meant to name the type of
 * thread created.  Returns the ds_thread_t so created, or NULL if
 * something went wrong (which is reported to the log.)
 */

extern ds_thread_t ds_thread_create __P((PFIC ep, void *param, int prio, 
					 char *class, int loglevel));

/*
 * WARNING: don't destroy threads.  Don't even think about it.  See
 * Satya if you would like to so much as ponder destroying a thread
 * you've created. -bnoble
 *
 * Seriously; rather than create and destroy threads willy-nilly,
 * create a pool of them, and dispatch work to them through something
 * like a ds_safeq_t.  You can even create more of them if you find
 * that the queue depth grows too large.  But it costs way too much to
 * be creating them and destroying them on the fly.
 */

#if 0
/*
 * Destroy a thread.  Returns zero on success, some
 * thread-system-specific error code (that's guaranteed to not be zero
 * on failure.  
 */
extern int ds_thread_destroy __P((ds_thread_t p));
#endif

/*
 * Log a message; I have no idea if this can be properly inlined or
 * not.  I guess if worse comes to worse, we spend a null procedure
 * call if we want it undefined.  I'd prefer to not use the macro
 * trick, I suppose.
 */

extern void ds_thread_log __P((int level, char *fmt, ...));

/*
 * Explicitly flush the log
 */
extern void ds_thread_lflush __P(());

/*
 * Change a log's level.  Return the old one
 */
extern int ds_thread_setll __P((int level));

/* 
 * Get the current thread's id.
 */
extern ds_thread_t ds_thread_current __P(());

/*
 * Get the current thread's unique name
 */
extern const char *ds_thread_name __P(());



/*
 * These are effectively just the lock package macros/event calls if
 * ODYDEBUG is not set, but if it is this also checks to see if we'd
 * like to log the messgae that a lock was acquired, or an event was
 * waited on/signalled.  
 */

#ifdef ODYDEBUG

extern ds_lock_debug;
extern ds_event_debug;

#define ds_lock_init(lock) \
do { \
    if (ds_lock_debug) { \
	ds_thread_log(0, "Initializing lock %p", lock); \
        ds_thread_lflush(); \
    } \
    Lock_Init(lock); \
} while (0)

#define ds_lock_read(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "readlocking %p", lock); \
        ds_thread_lflush(); \
    } \
    ObtainReadLock(lock); \
} while (0) 

#define ds_lock_write(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "writelocking %p", lock); \
        ds_thread_lflush(); \
    } \
    ObtainWriteLock(lock); \
} while (0) 

#define ds_lock_shared(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "sharelocking %p", lock); \
        ds_thread_lflush(); \
    } \
    ObtainSharedLock(lock); \
} while (0) 

#define ds_lock_boost(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "boosting %p", lock); \
        ds_thread_lflush(); \
    } \
    BoostSharedLock(lock); \
} while (0) 

#define ds_lock_unboost(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "unboosting %p", lock); \
        ds_thread_lflush(); \
    } \
    UnboostSharedLock(lock); \
} while (0) 

#define ds_lock_release(lock) \
do { \
    if (ds_lock_debug) {  \
	ds_thread_log(0, "releasing %p", lock); \
        ds_thread_lflush(); \
    } \
    if ((lock)->excl_locked) { \
	ReleaseSharedLock(lock); \
    } else { \
	ReleaseReadLock(lock); \
    } \
} while (0) 

#define ds_thread_wait(event) \
do { \
    if (ds_event_debug) { \
	ds_thread_log(0, "waiting on %p",event); \
        ds_thread_lflush(); \
    } \
    LWP_WaitProcess(event); \
} while (0)

#define ds_thread_mwait(wcount, elist) \
do { \
    if (ds_event_debug) { \
        int i; \
	ds_thread_log(0, "waiting on event-list, count %d:",wcount);  \
	for (i=0; i<count; i++) { \
            ds_thread_log(0,"\t%d) %p",i,elist[i]); \
        } \
        ds_thread_lflush(); \
    } \
    LWP_MWaitProcess(wcount,elist); \
} while (0)

#define ds_thread_schedule(event) \
do { \
    if (ds_event_debug) { \
	ds_thread_log(0, "Scheduling on %p",event); \
        ds_thread_lflush(); \
    } \
    LWP_SignalProcess(event); \
} while (0)

#define ds_thread_wakeup(event) \
do { \
    if (ds_event_debug) { \
	ds_thread_log(0, "Waking on %p",event); \
        ds_thread_lflush(); \
    } \
    LWP_NoYieldSignal(event); \
} while (0)

#define ds_thread_yield() \
do { \
    if (ds_event_debug) { \
	ds_thread_log(0, "Yielding"); \
        ds_thread_lflush(); \
    } \
    LWP_DispatchProcess(); \
} while (0)

#else

#define ds_lock_init           Lock_Init
#define ds_lock_read           ObtainReadLock
#define ds_lock_write          ObtainWriteLock
#define ds_lock_shared         ObtainSharedLock
#define ds_lock_boost          BoostSharedLock
#define ds_lock_unboost        UnboostSharedLock

/* 
 * We use only one release call to simplify our lives.  I believe
 * using ReleaseSharedLock is kosher only because we are guaranteed
 * not to have a Shared and a Write holding at the same time, so
 * clearing both bits is okay.
 */
#define ds_lock_release(lock)    \
do {                             \
    if ((lock)->excl_locked) {   \
	ReleaseSharedLock(lock); \
    } else {                     \
        ReleaseReadLock(lock);   \
    }                            \
} while (0)

#define ds_thread_wait         LWP_WaitProcess
#define ds_thread_mwait        LWP_MWaitProcess
#define ds_thread_schedule     LWP_SignalProcess
#define ds_thread_wakeup       LWP_NoYieldSignal
#define ds_thread_yield        LWP_DispatchProcess

#endif

#endif /* _DS_THREAD_H_ */
