#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Odyssey resource requests
 *
 * Implements the user-level support bits for resource requests.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>

#include "kernel/upcall/includes/upcall.h"
#include "include/odyssey.h"
#include "kernel/vfs/includes/ody_upcall.h"
#include "kernel/vfs/includes/odyapi.h"

struct ody_reqent_t {
    LIST_ENTRY(ody_reqent_t) or_list;
    int                      or_reqid;
    ody_req_fn_t             or_fp;
};

LIST_HEAD(ody_reqlist_t, ody_reqent_t);

static char                 inited = 0; /* not initialized yet */
static const int            rtabsz = 23; 
static struct ody_reqlist_t *rtab;

/* 
 * We don't need to block upcalls here: that implicitly happens.
 */
static int
ody_uchdlr(ucl, uca, argsz, argp, retszp, retp, flags)
    upclass_t ucl;
    upcall_t  uca;
    size_t    argsz;
    void      *argp;
    size_t    *retszp;
    void      *retp;
    int       flags;
{
    struct ody_ucargs_t    *args = argp;
    struct ody_reqlist_t   *bucket;
    struct ody_reqent_t    *orep;
    
#ifdef ODYDEBUG
    assert(ucl == UC_ODYSSEY);
    assert(uca == ODY_NOTIFICATION);
#endif
    assert(argsz == sizeof(struct ody_ucargs_t));
    
    /*  
     * Put in an upcall_block().  It will be cleared by the return
     * from this signal handler, but no other intermediate calls to
     * upcall_unblock (providing that they are properly matched) will
     * clear the mask.
     */
    assert(!upcall_block());
    
    bucket = &rtab[args->oua_reqid % rtabsz];
    for (orep = bucket->lh_first;
	 orep; 
	 orep = orep->or_list.le_next) {
	if (orep->or_reqid == args->oua_reqid)
	    break;
    }
    
    if (!orep) {
#ifdef ODYDEBUG
	fprintf(stderr,"ody_uchdlr: bad upcall number:");
	fprintf(stderr,"%d (%d, %d, %d)\n", uca, args->oua_reqid,
		args->oua_rsrc, args->oua_rval);
#endif
	return ENOENT;
    }
    
    (orep->or_fp)(args->oua_reqid, args->oua_rsrc, args->oua_rval);

    LIST_REMOVE(orep, or_list);
    free(orep);
    
    return 0;
}

static int
init()
{
    int rc;
    int i;
    
    rc = upcall_block();

    if (rc) {
#ifdef ODYDEBUG
	perror("odyssey: upcall_block");
#endif
	errno = EOPNOTSUPP;
	return -1;
    }
    
    rtab = malloc(sizeof(struct ody_reqlist_t) * rtabsz);
    for (i=0; i<rtabsz; i++) {
	LIST_INIT(&rtab[i]);
    }
    
    if (upcall_reg(UC_ODYSSEY, ODY_NOTIFICATION, ody_uchdlr)
	== UC_ERR)
    {
#ifdef ODYDEBUG
	perror("odyssey: upcall_reg");
#endif
	rc = -1;
	goto exit; /* Need to unblock upcalls */
    }
    
    inited = 1;

 exit:
    /* If it gets here, unblock has to succeed */
    assert(!upcall_unblock());
    
    return rc;
}


#define REMOVE_ENTRY(reqid, entp)           \
do {                                        \
    struct ody_reqlist_t *bucket;           \
    bucket = &rtab[(reqid) % rtabsz];       \
    for ((entp) = bucket->lh_first;         \
	 (entp);                            \
	 (entp) = (entp)->or_list.le_next)  \
    {                                       \
	if ((entp)->or_reqid == (reqid)) {  \
	    LIST_REMOVE((entp), or_list);   \
	    break;                          \
	}                                   \
    }                                       \
} while (0)

#define INSERT_ENTRY(reqid, fp, entp)           \
do {                                            \
    struct ody_reqlist_t *bucket;               \
    bucket = &rtab[(reqid) % rtabsz];           \
    (entp)->or_reqid = (reqid);                 \
    (entp)->or_fp = (fp);                       \
    LIST_INSERT_HEAD(bucket, (entp), or_list);  \
} while (0)

int
ody_request(path, reqp, resp)
    char               *path;
    struct ody_req_des *reqp;
    int                *resp;
{
    int                  error;
    struct ody_reqent_t  *orep;
    struct ody_reqent_t  *victim;
    struct ody_reqargs   ora;
    
    if (!inited) {
	error = init();
	if (error) {
	    return error;
	}
    }

    assert(!upcall_block());
    
    /* Speculatively malloc the ofep */
    orep = malloc(sizeof(struct ody_reqent_t));
    if (!orep) {
	errno = ENOMEM;
	error = -1;
	goto exit;
    }

    ora.ora_resource = reqp->ord_resource;
    ora.ora_gen_intf = reqp->ord_version.ov_gs;
    ora.ora_obj_type = reqp->ord_version.ov_type;
    ora.ora_type_intf = reqp->ord_version.ov_tsvnum;
    ora.ora_low = reqp->ord_low;
    ora.ora_high = reqp->ord_high;
    ora.ora_cancelled = 0;
    ora.ora_value = 0;
    ora.ora_result = 0;

    error = __ody_req(path, &ora);

    if (error && (errno != ODY_ENOTINWINDOW)) {
	goto exit;
    }

    if (ora.ora_cancelled) {
	REMOVE_ENTRY(ora.ora_cancelled, victim);
#ifdef ODYDEBUG
	if (!victim) {
	    fprintf(stderr,"Odyssey: can't remove request %d\n",
		    ora.ora_cancelled);
	}
#endif
	free(victim);
    }
    
    if (ora.ora_result) {
	/* We have a valid, placed request */
	INSERT_ENTRY(ora.ora_result, reqp->ord_fp, orep);
	*resp = ora.ora_result;
    } else {
	/* We don't have a result, we have a value out-of-window */
	assert(errno == ODY_ENOTINWINDOW);
	assert(error == -1);
	*resp = ora.ora_value;
    }

 exit:
    assert(!upcall_unblock());
    if (error) {
	/* We should free the orep, since we never entered it */
	free(orep);
    }
    return error;
}

int
ody_frequest(fd, reqp, resp)
    int                fd;
    struct ody_req_des *reqp;
    int                *resp;
{
    int                  error;
    struct ody_reqent_t  *orep;
    struct ody_reqent_t  *victim;
    struct ody_reqargs   ora;

    if (!inited) {
	error = init();
	if (error) return error;
    }

    assert(!upcall_block());
    
    /* Speculatively malloc the orep */
    orep = malloc(sizeof(struct ody_reqent_t));
    if (!orep) {
	errno = ENOMEM;
	error = -1;
	goto exit;
    }

    ora.ora_resource = reqp->ord_resource;
    ora.ora_gen_intf = reqp->ord_version.ov_gs;
    ora.ora_obj_type = reqp->ord_version.ov_type;
    ora.ora_type_intf = reqp->ord_version.ov_tsvnum;
    ora.ora_low = reqp->ord_low;
    ora.ora_high = reqp->ord_high;
    ora.ora_cancelled = 0;
    ora.ora_value = 0;
    ora.ora_result = 0;

    error = __ody_freq(fd, &ora);

    if (error && (errno != ODY_ENOTINWINDOW)) {
	goto exit;
    }

    if (ora.ora_cancelled) {
	REMOVE_ENTRY(ora.ora_cancelled, victim);
#ifdef ODYDEBUG
	if (!victim) {
	    fprintf(stderr,"Odyssey: can't remove request %d\n",
		    ora.ora_cancelled);
	}
#endif
	free(victim);
    }
    
    if (ora.ora_result) {
	/* We have a valid, placed request */
	INSERT_ENTRY(ora.ora_result, reqp->ord_fp, orep);
	*resp = ora.ora_result;
    } else {
	/* We don't have a result, we have a value out-of-window */
	assert(errno == ODY_ENOTINWINDOW);
	assert(error == -1);
	*resp = ora.ora_value;
    }

 exit:
    assert(!upcall_unblock());
    if (error) {
	/* We should free the orep, since we never entered it */
	free(orep);
    }
    return error;
    
}


int
ody_cancel(reqid)
    int reqid;
{
    char                *fname = "/ody/.CONTROL";
    int                 rc;
    struct ody_reqent_t *victim;

    if (!inited) {
	    rc = init();
	    if (rc) return rc;
    }

    assert(!upcall_block());

    rc = __ody_cancel(fname, reqid);
    
    if (!rc) {
	    REMOVE_ENTRY(reqid, victim);
#ifdef ODYDEBUG
	    if (!victim) {
		    fprintf(stderr,"Odyssey: __ody_cancel=>0, but no req\n");
	    }
#endif
	    free (victim);
    }
    
    assert(!upcall_unblock());

    return rc;
}
