#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <errno.h>

#include <sys/uio.h>

#include "include/odyssey_multres.h"
#include "odyssey_multres.private.h"
#include "include/odytypes.h"

/* The following is a linked list of outstanding multi-requests */
static ody_mreq_res_prvt_t* pRequestList = 0;

/* addRequest adds a request structure to the head of the
   global list */
void addRequest (ody_mreq_res_prvt_t* pRequest)
{
  pRequest->omrp_next = pRequestList;
  pRequestList = pRequest;
}

/* removeRequest removes a request structure from the global 
   list and deallocates it */
void removeRequest (ody_mreq_res_prvt_t* pRequest) 
{
  ody_mreq_res_prvt_t* pPrevRequest;                 

  FREE (pRequest->omrp_resources);                 
  if (pRequestList == pRequest) {                    
    pRequestList = pRequest->omrp_next;              
  } else {                                           
    /* Find the previous request in the list */      
    for (pPrevRequest = pRequestList;                
	 pPrevRequest->omrp_next &&                  
	   pPrevRequest->omrp_next != pRequest;      
	 pPrevRequest = pPrevRequest->omrp_next);    
    ASSERT (pPrevRequest->omrp_next == pRequest);    
    pPrevRequest->omrp_next = pRequest->omrp_next;   
  } /* endif */                                      
  FREE (pRequest);                                   
} 

/* scaleWindow scales the specified current resource values by the
   specified multiplier, and places the results in the output
   value structure. */
void scaleWindow (int resnum, ody_req_res_prvt_t* pResData,
		  ody_req_res_val_t* pInValues, 
		  ody_req_res_val_t* pOutValues, float scale)
{
  int i;

  for (i = 0; i < resnum; i++) {
    if (pResData[i].orrp_incr_effect == INCR_EFFECT_POSITIVE) {
      if ((float) pInValues[i].orrv_value * scale > MAX_RES_VALUE) {
	pOutValues[i].orrv_value = MAX_RES_VALUE;
      } else {
	pOutValues[i].orrv_value = (float) pInValues[i].orrv_value * scale;
      } /* endif */
    } else {
      if ((float) pInValues[i].orrv_value / scale > MAX_RES_VALUE) {
	pOutValues[i].orrv_value = MAX_RES_VALUE;
      } else {
	pOutValues[i].orrv_value = (float) pInValues[i].orrv_value / scale;
      } /* endif */
    } /* endif */
  } /* endif */  
}

/* setResourceWindows() uses Newton's methods to determine appropriate
   high and low windows for each resource based upon the mult-resource
   request evaluation function and window supplied by the user. */
int setResourceWindows (ody_mreq_res_prvt_t* pRequest, 
			float current_low_value, float current_high_value)
{
  struct ody_req_des or_des;
  ody_req_res_val_t* pScaledValues;
  float low_mult, high_mult, new_mult;
  float low_value, high_value, new_value;
  int offTheScaleFlag;
  int requestId;
  int i, j;
  int ody_rc;

#ifdef SUPER_DEBUG
  printf ("Testing for low window values\n");
#endif
  
  /* Create a structure to hold scaled values used for to probe for 
     the appropriate window */
  NALLOC (pScaledValues, ody_req_res_val_t, pRequest->omrp_resnum);
  memcpy (pScaledValues, pRequest->omrp_values, 
	  pRequest->omrp_resnum * sizeof (ody_req_res_val_t));

  /* Use Newton's method to determine where windows should be set */
  /* First, calculate windows for low window value */
  if (pRequest->omrp_low_eval_fp) {
 
    offTheScaleFlag = 0;
    low_mult = 1.0;
    high_mult = 1.0;
    high_value = current_low_value;
    do {
      
      /* If the multiplier is more than the maximum value of a resource,
	 then we are in an infinite loop, since resources take on
	 integer values only. Therefore, the resources are scaled
	 as much as possible. */
      if (low_mult <= (1.0 / (float) MAX_RES_VALUE)) {
	offTheScaleFlag = 1;
	break;
      } /* endif */
      
      low_mult = low_mult / 2.0;
      
      scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		   pRequest->omrp_values, pScaledValues, low_mult);
      low_value = pRequest->omrp_low_eval_fp (pRequest->omrp_resnum, 
					      pScaledValues);
      
#ifdef SUPER_DEBUG
      printf ("Low multiplier is %f\n", low_mult);
      printf ("Low value is %f\n", low_value);
#endif
      
    } while (pRequest->omrp_low < low_value);
    
    while (low_value != high_value && !offTheScaleFlag) {
      
      /* Calculate new multiplier using Newton's method */
      new_mult = ((float) (pRequest->omrp_low - low_value) / 
		  (float) (high_value - low_value)) * (high_mult - low_mult)
	+ low_mult;
      
      scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		   pRequest->omrp_values, pScaledValues, new_mult);
      new_value = pRequest->omrp_low_eval_fp (pRequest->omrp_resnum, 
					      pScaledValues);
      
#ifdef SUPER_DEBUG
      printf ("Low multiplier is %f\n", low_mult);
      printf ("High multiplier is %f\n", high_mult);
      printf ("Low value is %f\n", low_value);
      printf ("High value is %f\n", high_value);
      printf ("New multiplier is %f\n", new_mult);
      printf ("New value is %f\n", new_value);
#endif
      
      if (new_value > pRequest->omrp_low) {
	/* New value should not be greater than previous high value */
	ASSERT (new_value <= high_value);
	if (new_value == high_value) {
	  /* New value is equal to a previously calculated value, so
	     we're no longer making progress. */
	  break; 
	} else {
	  high_value = new_value;
	  high_mult = new_mult;
	} /* endif */
      } else if (new_value < pRequest->omrp_low) {
	/* New value should not be lower than previous low value */
	ASSERT (new_value >= low_value);     
	if (new_value == low_value) {
	  /* New value is equal to a previously calculated value, so
	     we're no longer making progress. */
	  break; 
	} else {
	  low_value = new_value;
	  low_mult = new_mult;
	} /* endif */
      } else {
	/* New value is equal to a desired value */
	break;
      } /* endif */
      
    } /* endwhile */
  
    /* Save window values for each resource - be conservative, so
       use lowest possible bound for high window, etc. */
    for (i = 0; i < pRequest->omrp_resnum; i++) {
      if (pRequest->omrp_resources[i].orrp_incr_effect == 
	  INCR_EFFECT_POSITIVE) {
	scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		     pRequest->omrp_values, pScaledValues, high_mult);
	pRequest->omrp_resources[i].orrp_low_window_value = 
	  pScaledValues[i].orrv_value; 
      } else {
	scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		     pRequest->omrp_values, pScaledValues, low_mult);
	pRequest->omrp_resources[i].orrp_high_window_value = 
	  pScaledValues[i].orrv_value; 
      } /* endif */
    } /* endfor */

  } else {

    /* Save maximum window values for each resource */
    for (i = 0; i < pRequest->omrp_resnum; i++) {
      if (pRequest->omrp_resources[i].orrp_incr_effect == 
	  INCR_EFFECT_POSITIVE) {
	pRequest->omrp_resources[i].orrp_low_window_value = 0;
      } else {
	pRequest->omrp_resources[i].orrp_high_window_value = MAX_RES_VALUE;
      } /* endif */
    } /* endfor */

  } /* endif */

#ifdef SUPER_DEBUG
  printf ("Testing for high window values\n");
#endif
  
  /* Then, calculate windows for high window value */
  if (pRequest->omrp_high_eval_fp) {

    offTheScaleFlag = 0;
    low_mult = 1.0;
    high_mult = 1.0;
    low_value = current_high_value;
    do {
      
      /* If the multiplier is more than the maximum value of a resource,
	 then we are in an infinite loop, since resources take on
	 integer values only. Therefore, the resources are scaled
	 as much as possible. */
      if (high_mult >= MAX_RES_VALUE) {
	offTheScaleFlag = 1;
	break;
      } /* endif */
      
      high_mult = high_mult * 2.0;
      
      scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		   pRequest->omrp_values, pScaledValues, high_mult);
      high_value = pRequest->omrp_high_eval_fp (pRequest->omrp_resnum, 
						pScaledValues);
      
#ifdef SUPER_DEBUG
      printf ("High multiplier is %f\n", high_mult);
      printf ("High value is %f\n", high_value);
#endif
      
    } while (pRequest->omrp_high > high_value);
    
    while (low_value != high_value && !offTheScaleFlag) {
      
      /* Calculate new multiplier using Newton's method */
      new_mult = ((float) (pRequest->omrp_high - low_value) / 
		  (float) (high_value - low_value)) * (high_mult - low_mult)
	+ low_mult;
      
      scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		   pRequest->omrp_values, pScaledValues, new_mult);
      new_value = pRequest->omrp_high_eval_fp (pRequest->omrp_resnum, 
					       pScaledValues);
      
#ifdef SUPER_DEBUG
      printf ("Low multiplier is %f\n", low_mult);
      printf ("High multiplier is %f\n", high_mult);
      printf ("Low value is %f\n", low_value);
      printf ("High value is %f\n", high_value);
      printf ("New value is %f\n", new_value);
      printf ("New multiplier is %f\n", new_mult);
#endif
      
      if (new_value > pRequest->omrp_high) {
	/* New value should not be greater than previous high value */
	ASSERT (new_value <= high_value);
	if (new_value == high_value) {
	  /* New value is equal to a previously calculated value, so
	     we're no longer making progress. */
	  break;
	} else {
	  high_value = new_value;
	  high_mult = new_mult;
	} /* endif */
      } else if (new_value < pRequest->omrp_high) {
	/* New value should not be lower than previous low value */
	fflush (stdout);
	ASSERT (new_value >= low_value);     
	if (new_value == low_value) {
	  /* New value is equal to a previously calculated value, so
	     we're no longer making progress. */
	  break;
	} else {
	  low_value = new_value;
	  low_mult = new_mult;
	} /* endif */
      } else {
	/* New value is equal to a desired value */
	break;
      } /* endif */
      
    } /* endwhile */
    
    /* Save window values for each resource - be conservative, so
       use lowest possible bound for high window, etc. */
    for (i = 0; i < pRequest->omrp_resnum; i++) {
      if (pRequest->omrp_resources[i].orrp_incr_effect == 
	  INCR_EFFECT_POSITIVE) {
	scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		     pRequest->omrp_values, pScaledValues, low_mult);
	pRequest->omrp_resources[i].orrp_high_window_value = 
	  pScaledValues[i].orrv_value;
      } else {
	scaleWindow (pRequest->omrp_resnum, pRequest->omrp_resources, 
		     pRequest->omrp_values, pScaledValues, high_mult);
	pRequest->omrp_resources[i].orrp_low_window_value =
	  pScaledValues[i].orrv_value; 
      } /* endif */
    } /* endfor */
  
  } else {

    /* Save maximum window values for each resource */
    for (i = 0; i < pRequest->omrp_resnum; i++) {
      if (pRequest->omrp_resources[i].orrp_incr_effect == 
	  INCR_EFFECT_POSITIVE) {
	pRequest->omrp_resources[i].orrp_high_window_value = MAX_RES_VALUE;
      } else {
	pRequest->omrp_resources[i].orrp_low_window_value = 0;
      } /* endif */
    } /* endfor */

  } /* endif */

  /* Set the windows.  If any resource is out of bounds, we need to repeat
     the whole process */
  for (i = 0; i < pRequest->omrp_resnum; i++) {
    
    or_des.ord_resource = pRequest->omrp_resources[i].orrp_resource;
    or_des.ord_version = pRequest->omrp_version;
    or_des.ord_fp = outOfWindow;
    or_des.ord_low = pRequest->omrp_resources[i].orrp_low_window_value;
    or_des.ord_high = pRequest->omrp_resources[i].orrp_high_window_value;
    
#ifdef ODYDEBUG
    printf ("Setting window for resource %d at (%d,%d)\n", 
	    or_des.ord_resource, or_des.ord_low, or_des.ord_high);
#endif
    
    if (pRequest->omrp_fd >= 0) {
      ody_rc = ody_frequest(pRequest->omrp_fd, &or_des, &requestId);
    } else {
      ody_rc = ody_request(pRequest->omrp_path, &or_des, &requestId);
    } /* endif */
    if (ody_rc == 0) {
      pRequest->omrp_resources[i].orrp_request_id = requestId;
#ifdef ODYDEBUG
      printf ("Request id is %d\n", requestId);
#endif
    } else {

      if (errno != ODY_ENOTINWINDOW) {
	fflush (stdout);
	perror("ody_frequest failed");
	ASSERT (0);
      } else {

#ifdef ODYDEBUG
	printf ("Resource out of bounds at value %d\n", requestId);
#endif
	pRequest->omrp_resources[i].orrp_request_id = -1;
	pRequest->omrp_values[i].orrv_value = requestId;

	/* Get the current values of the other resources */
	for (j = 0; j < pRequest->omrp_resnum; j++) {
	  if (j != i) {

	    pRequest->omrp_values[j].orrv_value = 
	      ody_getResourceValue (pRequest->omrp_fd, pRequest->omrp_path,
				    pRequest->omrp_resources[j].orrp_resource,
				    pRequest->omrp_version);

	    pRequest->omrp_resources[j].orrp_request_id = -1;
#ifdef DEBUG
	    printf ("Current value of resource %d is %d\n",
		    pRequest->omrp_values[j].orrv_resource,
		    pRequest->omrp_values[j].orrv_value);
#endif

	  } /* endif */
	} /* endfor */

	FREE (pScaledValues);
	return (ODY_ENOTINWINDOW);

      } /* endif */
    } /* endif */
    
  } /* endfor */

  FREE (pScaledValues);
  return (0);
}

/* outOfWindow is called whenever a bound on an individual resource is
   exceeded.  It first must locate the specific resource request, then
   it will either call the multi-resource handler (if the multi-resource
   window is exceeded), or recalculate the individual resource windows. */
void outOfWindow(int reqid, unsigned int rsrc, int val)
{
  ody_mreq_res_prvt_t* pRequest;
  ody_req_res_val_t* pValues;
  ody_req_res_prvt_t* pResource = 0;
  ody_mreq_fn_t req_fp;
  float low_value = 0, high_value = 0;
  int resnum;
  int i;

#ifdef ODYDEBUG
  printf ("Out of window, request id = %d\n", reqid);
#endif

  /* Find the request which generated this callback */
  for (pRequest = pRequestList; pRequest; pRequest = pRequest->omrp_next) {
    for (i = 0; i < pRequest->omrp_resnum; i++) {
      if (pRequest->omrp_resources[i].orrp_request_id == reqid) {
	pResource = &pRequest->omrp_resources[i];
	break;
      } /* endif */
    } /* endfor */
    if (pResource) {
      break;
    } /* endif */
  } /* endfor */

  /* If the request id was not found, ignore this callback */
  if (!pResource) {
    printf ("Request id %d was not recognized\n", reqid);
    return;
  } /* endif */

  /* The resource type should be the same as the request */
  ASSERT (rsrc == pResource->orrp_resource);

#ifdef ODYDEBUG
  printf ("Current value of resource %d is %d\n", rsrc, val);
#endif

  /* Get the current value of all other resources */
  resnum = pRequest->omrp_resnum;
  for (i = 0; i < resnum; i++) {
    if (&pRequest->omrp_resources[i] == pResource) {
      pRequest->omrp_values[i].orrv_value = val;
    } else {
      pRequest->omrp_values[i].orrv_value = 
	ody_getResourceValue (pRequest->omrp_fd, pRequest->omrp_path,
			      pRequest->omrp_resources[i].orrp_resource,
			      pRequest->omrp_version);
      pRequest->omrp_resources[i].orrp_request_id = -1;
#ifdef ODYDEBUG
  printf ("Current value of resource %d is %d\n", 
	  pRequest->omrp_resources[i].orrp_resource, 
	  pRequest->omrp_values[i].orrv_value);
#endif
    } /* endif */
  } /* endfor */

  do {

    /* Check if the multi-request is out of bounds. */
    if (pRequest->omrp_low_eval_fp) {
      low_value = pRequest->omrp_low_eval_fp (resnum, pRequest->omrp_values);
    } /* endif */
    if (pRequest->omrp_high_eval_fp) {
      high_value = pRequest->omrp_high_eval_fp (resnum, pRequest->omrp_values);
    } /* endif */

    if ((pRequest->omrp_high_eval_fp && high_value > pRequest->omrp_high) || 
	(pRequest->omrp_low_eval_fp && low_value < pRequest->omrp_low)) {
      
#ifdef ODYDEBUG
      printf ("Value is out of bounds\n");
#endif

      /* Out of the window! Cleanup and call back the specified function */
      req_fp = pRequest->omrp_req_fp;
      reqid = pRequest->omrp_request_id;
      pValues = pRequest->omrp_values;

      removeRequest (pRequest);

      /* Call back the specified user function */
      req_fp (reqid, resnum, pValues);
      return;

    } /* endif */

#ifdef ODYDEBUG
    printf ("Trying to set windows\n");
#endif

    /* Try to set resource windows.  On failure, need to repeat the process */
  } while (setResourceWindows (pRequest, low_value, high_value) 
	   == ODY_ENOTINWINDOW);
}

/* ody_getResourceValue() is called to determine the current value of
   an individual resource.  Because there is no system call to do this,
   it accomplishes its task by setting a small resource window, expecting
   to be told the resource value when the resource value is out of
   bounds.  If, by some small chance, the resource value is inside the
   window, it cancels the request.  The function has the side effect of
   cancelling any outstanding resource request on the resource for
   the application. */
int ody_getResourceValue (int fd, char* path, int resource, 
			  struct ody_vers version)
{
  struct ody_req_des or_des;
  int value = 0;
  int requestId;
  int ody_rc;

  /* Set a request for a window with high and low values of zero */
  or_des.ord_resource = resource;
  or_des.ord_version = version;
  or_des.ord_fp = outOfWindow;
  or_des.ord_low = 0;
  or_des.ord_high = 0;

  if (fd >= 0) {
    ody_rc = ody_frequest(fd, &or_des, &requestId);
  } else {
    ody_rc = ody_request (path, &or_des, &requestId);
  } /* endif */

  if (ody_rc == 0) {
    /* The resource value must be 0 */
    ody_cancel (requestId);
    value = 0;
  } else {
    if (errno != ODY_ENOTINWINDOW) {
      perror("ody_frequest failed");
      ASSERT (0);
    } else {
      value = requestId; 
    } /* endif */
  } /* endif */

  return (value);
}

/* ody_multi_request() sets a multi-resource request.  It first verifies
   that the resources are withing the specified bounds.  Then, it sets
   the individual resource windows. */
int ody_multi_request (int fd, char* path, struct ody_req_mres* req, 
		       int* reqid, ody_req_res_val_t** val)
{
  ody_req_res_val_t* pValues;
  ody_req_res_prvt_t* pResData;
  ody_mreq_res_prvt_t* pRequestData;
  ody_eval_fn_t eval_fp = 0;
  float low_value = 0, high_value = 0, incr_value = 0, orig_value = 0;
  int saved_value;
  int i;

  /* Need at least one input resource */
  ASSERT (req->orm_resnum > 0);

  /* Verify that at least one bound function is set */
  if (!req->orm_low_eval_fp && !req->orm_high_eval_fp) {
    errno = EINVAL;
    return (EINVAL);
  } /* endif */

  /* Get the current value of each resource */
  NALLOC (pValues, ody_req_res_val_t, req->orm_resnum);
  NALLOC (pResData, ody_req_res_prvt_t, req->orm_resnum);
  for (i = 0; i < req->orm_resnum; i++) {

#ifdef ODYDEBUG
    /* Verify that application is using the current resource interface */
    ASSERT (req->orm_resources[i].orr_version == ODY_MULTRES_VERSION);
#endif

    pValues[i].orrv_version = ODY_MULTRES_VERSION;
    pValues[i].orrv_resource = req->orm_resources[i].orr_resource;
    pValues[i].orrv_value = ody_getResourceValue (fd, path,
						  pValues[i].orrv_resource,
						  req->orm_version);
    pResData[i].orrp_resource = pValues[i].orrv_resource;
    pResData[i].orrp_request_id = -1; /* No active request yet */

#ifdef ODYDEBUG
    printf ("Current value of resource %d is %d\n", pValues[i].orrv_resource,
	    pValues[i].orrv_value);
#endif
  } /* endifor */
    
  /* Verify that we are inside the window */
  if (req->orm_low_eval_fp) {
    low_value = req->orm_low_eval_fp (req->orm_resnum, pValues);
    orig_value = low_value;
    eval_fp = req->orm_low_eval_fp;
  } /* endif */
  if (req->orm_high_eval_fp) {
    high_value = req->orm_high_eval_fp (req->orm_resnum, pValues);
    orig_value = high_value;
    eval_fp = req->orm_high_eval_fp;
  } /* endif */

  if ((req->orm_high_eval_fp && high_value > req->orm_high) || 
      (req->orm_low_eval_fp && low_value < req->orm_low)) {

    /* Oops, we're outside the window already */
    errno = ODY_ENOTINWINDOW;
    *val = pValues;
    FREE (pResData);
    return (ODY_ENOTINWINDOW);

  } /* endif */

  /* Calculate whether incremental effect of each resource is positive 
     or negative. */
  for (i = 0; i < req->orm_resnum; i++) {
    
    saved_value = pValues[i].orrv_value;
    pValues[i].orrv_value *= 1.1;
    if (pValues[i].orrv_value == saved_value) {
      pValues[i-1].orrv_value++;
    } /* endif */
    incr_value = eval_fp (req->orm_resnum, pValues);
    if (incr_value > orig_value) {
      pResData[i].orrp_incr_effect = INCR_EFFECT_POSITIVE;
    } else {
      pResData[i].orrp_incr_effect = INCR_EFFECT_NEGATIVE;
    } /* endif */
    pValues[i].orrv_value = saved_value;

  } /* endfor */

  /* Create a multi-request object and add it to the linked list */
  ALLOC (pRequestData, ody_mreq_res_prvt_t);
  pRequestData->omrp_fd = fd;
  pRequestData->omrp_path = path;
  pRequestData->omrp_resnum = req->orm_resnum;
  pRequestData->omrp_low = req->orm_low;
  pRequestData->omrp_high = req->orm_high;
  pRequestData->omrp_low_eval_fp = req->orm_low_eval_fp;
  pRequestData->omrp_high_eval_fp = req->orm_high_eval_fp;
  pRequestData->omrp_req_fp = req->orm_req_fp;
  pRequestData->omrp_version = req->orm_version;
  pRequestData->omrp_values = pValues;
  pRequestData->omrp_resources = pResData;

  /* Calcualte resource windows using Newton's method */
  while (setResourceWindows (pRequestData, low_value, high_value) 
	 == ODY_ENOTINWINDOW) {

#ifdef ODYDEBUG
    printf ("Unable to set windows, resource must have moved\n");
#endif

    /* Recheck if we're still inside the multi-window */
    if (req->orm_low_eval_fp) {
      low_value = req->orm_low_eval_fp (req->orm_resnum, pValues);
    } /* endif */
    if (req->orm_high_eval_fp) {
      high_value = req->orm_high_eval_fp (req->orm_resnum, pValues);
    } /* endif */
    if ((req->orm_high_eval_fp && high_value > req->orm_high) || 
	(req->orm_low_eval_fp && low_value < req->orm_low)) {

      /* Oops, we're outside the window now */
      FREE (pRequestData);
      FREE (pResData);
      errno = ODY_ENOTINWINDOW;
      *val = pValues;
      return (ODY_ENOTINWINDOW);

    } /* endif */    

  } /* endwhile */

  ASSERT (!REQUEST_BLOCK());
  addRequest (pRequestData);
  ASSERT (!REQUEST_UNBLOCK());

  pRequestData->omrp_request_id = pResData[0].orrp_request_id;
  *reqid = pRequestData->omrp_request_id;

  return (0);
}

/* ody_fmrequest() is a small wrapper function for multi-resource requests
   which pass file descriptors */
int ody_fmrequest (int fd, struct ody_req_mres* req, int* reqid,     
		   ody_req_res_val_t** val)
{
  if (fd < 0) {
    errno = EINVAL;
    return (EINVAL);
  } /* endif */

  return (ody_multi_request (fd, 0, req, reqid, val));
} /* endif */

/* ody_mrequest() is a small wrapper function for multi-resource requests
   which pass file paths */
int ody_mrequest (char* path, struct ody_req_mres* req, int* reqid,     
		  ody_req_res_val_t** val)
{
  if (!path) {
    errno = EINVAL;
    return (EINVAL);
  } /* endif */

  return (ody_multi_request (-1, path, req, reqid, val));
} /* endif */

/* ody_mcancel() cancels a multi-resource request.  It finds the request
   specified, cancels all associated individual resource request, and
   deallocates the request structure. */
int ody_mcancel (int reqid)
{
  ody_mreq_res_prvt_t* pRequest;
  int i;
  
  ASSERT(!REQUEST_BLOCK());

  /* Find the specified request */
  for (pRequest = pRequestList; pRequest; pRequest = pRequest->omrp_next) {
    if (pRequest->omrp_request_id == reqid) {
      break;
    } /* endif */
  } /* endfor */ 
  
  if (!pRequest) {
    ASSERT(!REQUEST_UNBLOCK());
    printf ("Cancel: No outstanding request with id %d\n" , reqid);
    return (0);
  } /* endif */
    
  /* Cancel all outstanding requests */
  for (i = 0; i < pRequest->omrp_resnum; i++) {
    if (pRequest->omrp_resources[i].orrp_request_id >= 0) {
      ody_cancel (pRequest->omrp_resources[i].orrp_request_id);
#ifdef ODYDEBUG
      printf ("Cancelled request %d\n", 
	      pRequest->omrp_resources[i].orrp_request_id);
#endif
    } /* endif */
  } /* endfor */

  /* Deallocate the request structure */
  removeRequest (pRequest);
  ASSERT(!REQUEST_UNBLOCK());
  return (0);
}
