#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * client.c -- web client functionality for SimpleWWW
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This is the client interface for the SimpleWWW library. See
 * swww.h for a detailed description of the library API.
 *
 * History
 *   28 Jun 1996 -- 0.2 -- With a sudden burst of enthusiasm, most
 *                         of the client code has been written.
 *    3 Jul 1996 -- 0.3 -- client.h has been incorporated into swww.h,
 *                         and client-private.h has become client.h
 *    5 Jul 1996 -- 0.4 -- header parsing finally moved into separate
 *                         function (in util.c)
 *   17 Jul 1996 -- 0.5 -- added LWP IOMGR_Select() support
 *
 * TODO:
 *   * XXX acknowledge fact that there may be HTTP/0.9 servers out there
 *   * XXX can probably abstract out very similar error handling code
 *     to make SWWW_fetch_url() more readable.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <assert.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <lwp/lwp.h>

#include "swww.h"
#include "errors.h"
#include "util.h"
#include "client.h"

/*
 * Public functions:
 */

/*
 * fetch_url -- given a URL and a method, initiates an HTTP request.
 *              Returns the response code, once the server has been contacted.
 */
int SWWW_fetch_url(struct SWWW_fetch_in  *inp,
		   struct SWWW_fetch_out *outp)
{
    char *scheme  = NULL;
    char *host    = NULL;
    char *path    = NULL;
    int  errcode  = 0;		/* set before jumping to errhandle:
				   "exception handler" at end */
    char *errbody = NULL;	/* also set before jumping to errhandle: */
    int  port;
    int  result;
    int  status_code=0;
    int  size;
    char buffer[1024];		/* buffer for reading from network */
    char *bookmark;
    int  sockfd;
    int  retbuffer_size;
    char *retbuffer;
    SWWW_headers_t *headers;

    /* General setup */
    outp->sfo_header_size = 0;
    outp->sfo_headers         = NULL;

    /*
     * Parse the URL into (scheme, host, path)
     */
    result = SWWW_parse_url(inp->sfi_url,
			    &scheme, &host, &port, &path);
    
    if(result != 0) {
	/* the URL was malformed */
	outp->sfo_status = strdup("HTTP/1.0 400 Bad Request (URL malformed)");
	errcode = 400;
	errbody = SWWW_400;
	goto errhandle;
    }

    if(strcmp(scheme, "http")) {
	/* if scheme != "http", complain bitterly */
	outp->sfo_status = strdup("HTTP/1.0 501 Not Implemented");
	errcode = 501;
	errbody = SWWW_501;
	goto errhandle;
    }

    /*
     * Set up the headers to be sent
     */
    
    /* XXX nothing here yet */

    /*
     * Open TCP connection to the server
     */
    sockfd = open_tcp(host, port, inp->sfi_threading);
    
    if(sockfd < 0) {		/* deal with errors */
	switch(sockfd) {
	case OT_NOHOST:
	    outp->sfo_status =
		strdup("HTTP/1.0 500 Internal Server Error (no such host)");
	    errcode = 500;
	    errbody = SWWW_500_no_such_host;
	    goto errhandle;
	case OT_NOSOCKET:
	    outp->sfo_status = strdup("HTTP/1.0 500 Internal Server Error "
				      "(couldn't create socket)");
	    errcode = 500;
	    errbody = SWWW_500_couldnt_create_socket;
	    goto errhandle;
	case OT_NOCONNECT:
	    outp->sfo_status = strdup("HTTP/1.0 500 Internal Server Error "
				      "(unable to connect to server)");
	    errcode = 500;
	    errbody = SWWW_500_unable_to_connect;
	    goto errhandle;
	default:
	    fprintf(stderr, "SWWW_fetch_url: Nonsense error from open_tcp!\n");
	    assert(0);
	}
    }

    /*
     * Send the request, and wait for response (with some suitable timeout)
     */
    switch(inp->sfi_method) {
    case GET:
	snprintf(buffer, 1024, "GET %s HTTP/1.0\r\n", path);
	break;
    case HEAD:
	snprintf(buffer, 1024, "HEAD %s HTTP/1.0\r\n\r\n", path);
	break;
    case POST:
	snprintf(buffer, 1024, "POST %s HTTP/1.0\r\n"
		               "Content-length: %d\r\n\r\n",
		               path, inp->sfi_postinfo->spi_length);
	break;
    default:
	outp->sfo_status = strdup("HTTP/1.0 501 Not Implemented");
	errcode = 501;
	errbody = SWWW_501;
	goto errhandle;
    }

    /* XXX is Eric purposely splitting the write into GET and then
       headers?  Since I don't understand, I'll just put them together
       for now. */
    /* XXX need to prevent buffer overflow! XXX */
    headers = inp->sfi_headers;
    while (headers && headers->name) {
	strcat(buffer, headers->name);
	strcat(buffer, ": ");
	strcat(buffer, headers->value);
	strcat(buffer, "\r\n");
	headers = headers->next;
    }
    strcat(buffer, "\r\n");
    
    result = writen(sockfd, buffer, strlen(buffer));
    if (result < 0) {
	/* Return a connection failure error */
	perror("SWWW_fetch_url/send request");
	outp->sfo_status = strdup("HTTP/1.0 500 Internal Server Error "
			 "(unable to connect to server)");
	errcode = 500;
	errbody = SWWW_500_unable_to_connect;
	goto errhandle;
    }

    /* XXX when general header support is added, put it here
           (and don't forget to take final "\r\n" out of above */

    if (inp->sfi_method == POST) {
	/* If this is a POST method, also send the POST data */
	result = send_postinfo(sockfd, inp->sfi_postinfo);
	if (result < 0) {
	    /* Return a connection failure error */
	    perror("SWWW_fetch_url/send request");
	    outp->sfo_status = strdup("HTTP/1.0 500 Internal Server Error "
				      "(unable to connect to server)");
	    errcode = 500;
	    errbody = SWWW_500_unable_to_connect;
	    goto errhandle;
	}
    }

    /*
     * Read the headers, and fill in response, header_size, and headers
     */

    result = parse_headers(sockfd,
			   inp->sfi_threading,
			   &(outp->sfo_status),
			   &(outp->sfo_header_size),
			   &(outp->sfo_headers),
			   &retbuffer_size,
			   &retbuffer);

#ifdef HEADER_DEBUGGING
    fprintf(stderr, "rep headers (%s/%s):\n", inp->sfi_url, outp->sfo_status);
    SWWW_print_headerlist(outp->sfo_headers);
#endif
    
    if (result < 0) {
	/* Return a connection failure error */
	perror("SWWW_fetch_url/parse headers");
	outp->sfo_status = strdup("HTTP/1.0 500 Internal Server Error "
				  "(unable to connect to server)");
	errcode = 500;
	errbody = SWWW_500_unable_to_connect;
	goto errhandle;
    }

    /*
     * deduce response code from response header
     */

    bookmark = index(outp->sfo_status, ' ');
    
    if(bookmark != NULL) {
	char *scratch;

	while(isspace(*bookmark)) {	/* be tolerant of whitespace */
	    bookmark++;
	}
	scratch = strdup(bookmark);
	bookmark = index(scratch, ' ');
	if(bookmark != NULL) {
	    *bookmark = '\0';
	    status_code = atoi(scratch);
	} else {
	    status_code = 500;
	}
	free(scratch);
    } else {
	status_code = 500;
    }

    if(inp->sfi_acceptheaders != NULL) {
	result = (*(inp->sfi_acceptheaders))(inp->sfi_reqid,
					     status_code,
					     outp->sfo_status,
					     outp->sfo_header_size,
					     outp->sfo_headers);
	if (result < 0)
	    goto cleanup;
    }
  
    if(retbuffer != NULL) {
	/* spit out remainder of buffer to datapump function */
	result = (*(inp->sfi_acceptdata))(inp->sfi_reqid,
					  retbuffer,
					  retbuffer_size);
	free(retbuffer);
	if (result < 0)
	    goto cleanup;
    }

    /*
     * Set up a read loop to read the remainder of the data, and pass it
     * along to the datapump function.
     */

    do {
	
	if(inp->sfi_threading == LWP) {
	    /* before doing the read, do an IOMGR_Select to
	       determine whether it's safe to do the read. */
	    int            rc;
	    fd_set         readset;
	    
	    FD_ZERO(&readset);
	    FD_SET(sockfd, &readset);
	    
	    rc = IOMGR_Select(sockfd+1, (int *) &readset, NULL, NULL, NULL);
	    
	    if (!FD_ISSET(sockfd, &readset)) {
		fprintf(stderr, "SWWW_fetch_url/select: timed out on select. "
			        "Aborting.\n");
		break;
	    }
	    
	    if(rc != 1) {
		fprintf(stderr, "SWWW_fetch_url/select: got %d back from "
			"IOMGR_Select, for no good reason. "
			"(Did you init IOMGR?)\n", rc);
		/* XXX should assert() */
	    }
	}
	
	size = read(sockfd, buffer, 1024);
	if(size > 0) {
	    result = (*(inp->sfi_acceptdata))(inp->sfi_reqid,
					      buffer,
					      size);
	    if(result == -1) {
		fprintf(stderr,
			"SWWW_fetch_url: acceptdata refusing data "
			"(transfer aborted?)\n");
		perror("acceptdata");
		break;		/* acceptdata_fn is no longer accepting
				   data, so stop sending to it */
	    }
	}
    } while(size > 0);
    
    /*
     * Clean up
     */
 cleanup:

    free(scheme);
    free(host);
    free(path);
    close(sockfd);
    
    return(status_code);
    
    /* This error handler simulates exception handling; if
       I encounter an error above, I set the following variables,
       and then jump here:
       
       errcode   set to the appropriate HTTP status code.
                 this is also what is returned by this function
       *status   set to the appropriate HTTP status line.
                 (this is expected to be set by the function
		  before it returns)
       errbody   set to a constant string which is sent as
                 the HTTP object body to the acceptdata_fn.
		 (this is not dealloacted, and is expected to
		 be from errors.h)

	This handler also frees scheme, host, and path, if they have
        been set (i.e., if they are non-NULL).
     */
 errhandle:

    outp->sfo_header_size = 0;		/* no headers returned */
    outp->sfo_headers     = NULL;
    
    if (inp->sfi_acceptheaders != NULL) {
	result = (*(inp->sfi_acceptheaders))(inp->sfi_reqid,
					     status_code,
					     outp->sfo_status,
					     outp->sfo_header_size,
					     outp->sfo_headers);
    }
    
    (*(inp->sfi_acceptdata))(inp->sfi_reqid,
			     errbody,
			     strlen(errbody));
    
    if (scheme != NULL)
	free(scheme);
    
    if (host   != NULL)
	free(host);
    
    if (path   != NULL)
	free(path);
    
    return(errcode);
}

/*
 * (private) local functionality
 */

int open_tcp(const char            *host,
	     int                   port,
	     SWWW_threading_type_t threading)
{
    /*
     * Opens a TCP connection to host on port. If port is negative,
     * port is assumed to be 80. Note that this is totally WWW (and
     * Internet) -centric, and we should probably at least check /etc/services.
     *
     * Returns an open file descriptor, or the following errors:
     *
     *  OT_NOHOST    No such host. The host could not be looked up.
     *  OT_NOSOCKET  Could not create socket. (errno contains specific error)
     *  OT_NOCONNECT Could not connect to host. (errno contains specific error)
     */
    struct sockaddr_in servaddr;
    struct hostent     hostinfo;
    struct hostent     *hp;
    unsigned long      inaddr;
    int                sockfd, result;

    /* Fill in address to server */
    if(port < 0)
	port = 80;		/* default WWW port;
				   XXX should I check /etc/services? */
    
    bzero((char *) &servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(port);
    
    if((inaddr = inet_addr(host)) != INADDR_NONE) {
	bcopy((char *) &inaddr, (char *) &servaddr.sin_addr, sizeof(inaddr));
	hostinfo.h_name = NULL;
    } else {
	if((hp = gethostbyname(host)) == NULL) {
	    herror("open_tcp/gethostbyname");
	    return(OT_NOHOST);	/* no such host */
	}
	hostinfo = *hp;
	bcopy(hp->h_addr, (char *) &servaddr.sin_addr, hp->h_length);
    }
    
    /*
     * Open a connection to the server
     */
    sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sockfd < 0) {
	perror("open_tcp/socket");
	return(OT_NOSOCKET);	/* couldn't create socket */
    }

    /*
     * if threading is being used, make the connection non-blocking
     */
    if (threading == LWP) {
	fcntl(sockfd, F_SETFL, O_NONBLOCK);
    }

    result = connect(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr));
    
    if (errno == EINPROGRESS && threading == LWP) {
	/* connect would have blocked, so wait for it. */
	int    rc;
	fd_set writeset;
	
	SWWW_DEBUG(fprintf(stderr, "waiting for connect\n"));
	
	FD_ZERO(&writeset);
	FD_SET(sockfd, &writeset);
	
	rc = IOMGR_Select(sockfd+1, NULL, (int *) &writeset, NULL, NULL);
	SWWW_DEBUG(fprintf(stderr, "got connect\n"));
    } else if(result < 0) {
	perror("open_tcp/connect");
	return(OT_NOCONNECT);
    }
    
    return(sockfd);
}


/*
 * write "n" bytes to a descriptor.
 * use in place of write() when fd is a stream socket.
 * taken from Unix Network Programming (Stevens, p. 279)
 */

int writen(register int fd, register char *ptr, register int nbytes)
{
    int nleft, nwritten;
    
    nleft = nbytes;
    while(nleft > 0) {
	nwritten = write(fd, ptr, nleft);
	if(nwritten <= 0)
	    return(nwritten);
	
	nleft -= nwritten;
	ptr   += nwritten;
    }
    return(nbytes - nleft);
}

/*
 * send_postinfo -- given the SWWW_postinfo structure, sends the POST
 *                  object along to the open socket.
 *
 *                  Returns 0 if successful, or -1 if there is an error
 *                  (with errno set).
 */
int send_postinfo(int sockfd, struct SWWW_postinfo *pip)
{
    char buffer[1024];
    int  size;
    int  count;
    int  result;

    /* Read in POST data (up the the length of the POST data),
       and send it back out to the reqid */
    if (pip->spi_bufsize > 0) {
	result = writen(sockfd, pip->spi_buffer, pip->spi_bufsize);

	if(result < 0)
	    return(result);
    }

    count = pip->spi_bufsize;
    if(count >= pip->spi_length)
	return(0);

    size = (*(pip->spi_fetchdata)) (pip->spi_postid,
				    buffer,
				    1024);
    while(size > 0) {
	result = writen(sockfd, buffer, size);
	if(result < 0)
	    return(result);
	count += size;
	if(count >= pip->spi_length)
	    break;
	size = (*(pip->spi_fetchdata)) (pip->spi_postid,
					buffer,
					1024);
    }

    return(0);
}
