#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * ploink -- a simple Web server which uses the SimpleWWW library
 *
 * Several configuration options are described in the file "config.h";
 * these are all compile-time options.
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/stat.h>
#include <base/ds/ds_safeq.h>
#include <base/simplewww/swww.h>
#include <lwp/lwp.h>
#include "main.h"
#include "config.h"

int main(int argc, char **argv) {
    int response_code;
    PROCESS pid;
    ds_safeq_t *request_q;
    
    if(SWWW_version(SWWW_VERSION)) {
	/* if library version doesn't match header file, we're hosed */
	exit(1);
    }
    
    /* XXX For now, leave this here. I'll be moving it into
       SimpleWWW at some later point, once I figure out the
       proper way to do it. */
    signal(SIGPIPE, SIG_IGN);
    
    if(LWP_Init(LWP_VERSION, LWP_NORMAL_PRIORITY, &pid) != LWP_SUCCESS) {
	fprintf(stderr, "main: can't initialize LWP!\n");
	exit(-1);
    }
    
    if(IOMGR_Initialize() != LWP_SUCCESS) {
	fprintf(stderr, "main: can't initialize IOMGR!\n");
	exit(-1);
    }

    SWWW_server_init(8, &request_q);
    response_code = SWWW_accept_requests(SERVERPORT,
					 process_requests,
					 LWP,
					 request_q);
    
    return(-1);			/* shouldn't ever reach here */
}

/*
 * REQUEST HANDLING
 */

/*
 * process_requests -- workhorse routine for a simple web server
 *     (using the SimpleWWW library). This routine is invoked whenever
 *     a request comes in, and does one of three things, depending
 *     on the method:
 *     
 *     GET     Retrieves the file requested, relative to the directory
 *             "/afs/cs.cmu.edu/user/tilt/public_html"
 *
 *     POST    The POST object data is returned.
 *
 *     Others  For any other method, an appropriate error message is
 *             returned.
 *
 * (This routine actually acts a demux, and invokes the appropriate
 * function for the given method.)
 *
 * Things that would be nice, but aren't currently supported:
 *    -- HEAD should provide a subset of GET functionality
 *    -- The root directory should be specifiable.
 *    -- Very limited access control is provided by throwing out
 *       any file containing ".." in the path. A more sophisticated
 *       approach would be better.
 *
 */
int process_requests(struct SWWW_fetch_in  *inp,
		     struct SWWW_fetch_out *outp)
{
    int            result=0;
    time_t         the_time;
    struct tm      *tp;
    char           buffer[100];

    /* Set up default headers */
    outp->sfo_headers     = SWWW_create_headerlist();
    outp->sfo_header_size = 0;

    outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					    "Server",
					    PLOINK_VERSION_STRING " "
					    SWWW_VERS_STRING " "
					    PLOINK_COMMENTS);
    the_time = time(NULL);
    tp = gmtime(&the_time);
    strftime(buffer, 100, "%a, %d %b %Y %H:%M:%S GMT", tp);

    outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					    "Date", buffer);
					    

    switch(inp->sfi_method) {
    case GET:
    case HEAD:
	result = process_get_request(inp, outp);
	break;
    case POST:
	result = process_post_request(inp, outp);
	break;
    default:
	result = process_unknown_request(inp, outp);
	break;
    }

    return(result);
}

/*
 * process_get_request -- handles GET requests
 * 
 * Retrieves the file requested, relative to the directory
 * "/afs/cs.cmu.edu/user/tilt/public_html"
 */

int process_get_request (struct SWWW_fetch_in  *inp,
			 struct SWWW_fetch_out *outp)
{
    char        *path;
    int         result=0;
    char        *errbody = NULL;
    struct stat sb;

    /* Find the file in question */
    path = file_lookup(inp->sfi_url, &sb);

    /* Set up and send headers */
    if (path == NULL) {
	/* Bogus pathname, return error */
	outp->sfo_status = strdup("HTTP/1.0 403 Forbidden");
	outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					       "Content-type", "text/html");
	errbody = FORBIDDEN;
	result = 403;
    } else {
	/* Got a real pathname */
	char      buffer[100];
	char      *type = deduce_type(path);
	struct tm *tp;

	outp->sfo_status = strdup("HTTP/1.0 200 Document follows");
	outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
						"Content-type", type);
	sprintf(buffer, "%d", (int) sb.st_size);
	outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
						"Content-length", buffer);
#ifdef HAVE_STAT_TIMESPEC
	tp = gmtime(&sb.st_mtimespec.tv_sec);
#else
	tp = gmtime(&sb.st_mtime);
#endif
	strftime(buffer, 100, "%a, %d %b %Y %H:%M:%S GMT", tp);
	outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
						"Last-modified", buffer);
	free(tp);
	free(type);
    }

    if (inp->sfi_acceptheaders != NULL) {
	(*(inp->sfi_acceptheaders)) (inp->sfi_reqid,
				     result,
				     outp->sfo_status,
				     outp->sfo_header_size,
				     outp->sfo_headers);
    }

    /* If this is HEAD, we can quit now. This would also be a good
       place to quit on a conditional get, once that gets implemented */
    if (inp->sfi_method == HEAD)
	return(result);


    /* Send data */
    if (errbody == NULL) {
	int  size;
	int  fd;
	char buffer[8192];

	fd = open(path, O_RDONLY);
	size = read(fd, buffer, 8192);
	while(size > 0) {
	    (*(inp->sfi_acceptdata)) (inp->sfi_reqid,
				      buffer,
				      size);
	    size = read(fd, buffer, 8192);
	}
	close(fd);
    } else {
	/* There was an error, so return an error message */
	(*(inp->sfi_acceptdata)) (inp->sfi_reqid,
				  errbody,
				  strlen(errbody));
    }

    return result;
}

/*
 * process_post_request -- handles POST requests
 * 
 * For any path, the POST object data is returned. This provides a
 * trivial "ping" that fits inside of HTTP.
 *
 */

int process_post_request(struct SWWW_fetch_in  *inp,
			 struct SWWW_fetch_out *outp)
{
    char                 buffer[1024];
    int                  size;
    int                  count;
    struct SWWW_postinfo *pip;

    /* Setup and initialization */
    pip = inp->sfi_postinfo;

    /* Compose and send headers */
    outp->sfo_status      = strdup("HTTP/1.0 200 OK");
    outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					    "Content-type", "text/plain");
    sprintf(buffer, "%d", pip->spi_length);
    outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					    "Content-length", buffer);

    if (inp->sfi_acceptheaders != NULL) {
	(*(inp->sfi_acceptheaders)) (inp->sfi_reqid,
				     200,
				     outp->sfo_status,
				     outp->sfo_header_size,
				     outp->sfo_headers);
    }

    /* Read in POST data (up the the length of the POST data),
       and send it back out to the reqid */

    (*(inp->sfi_acceptdata)) (inp->sfi_reqid,
			      pip->spi_buffer,
			      pip->spi_bufsize);

    count = pip->spi_bufsize;
    if(count >= pip->spi_length)
	return(200);

    size = (*(pip->spi_fetchdata)) (pip->spi_postid,
				    buffer,
				    1024);
    while(size > 0) {
	(*(inp->sfi_acceptdata)) (inp->sfi_reqid,
				  buffer,
				  size);
	count += size;
	if(count >= pip->spi_length)
	    break;
	size = (*(pip->spi_fetchdata)) (pip->spi_postid,
					buffer,
					1024);
    }

    return 200;
}

/*
 * process_unknown_request -- handles requests that are unknown or unsupported.
 *
 * Returns an appropriate error.
 *
 * (This is a good example of the minimum requirements for processing
 * a request.)
 *
 */

int process_unknown_request(struct SWWW_fetch_in  *inp,
			 struct SWWW_fetch_out *outp)
{
    /* Set up and send headers */
    outp->sfo_status = strdup("HTTP/1.0 501 Not Implemented");
    outp->sfo_header_size += SWWW_sethvalue(outp->sfo_headers,
					    "Content-type", "text/html");

    if (inp->sfi_acceptheaders != NULL) {
	(*(inp->sfi_acceptheaders)) (inp->sfi_reqid,
				     501,
				     outp->sfo_status,
				     outp->sfo_header_size,
				     outp->sfo_headers);
    }

    /* Send data */
    (*(inp->sfi_acceptdata)) (inp->sfi_reqid,
			      METHOD_NOT_SUPP,
			      strlen(METHOD_NOT_SUPP));

    return 501;
}

/*
 *  UTILITY FUNCTIONS
 */

/*
 * file_lookup -- given a path, finds it relative to ROOTDIR.
 *                checks to insure that final path is not outside
 *                of ROOTDIR. If it is, returns NULL, otherwise
 *                returns the full and canonical path of the object.
 *
 *                Also, if the the final path is for a directory,
 *                it tries (once) to find a file named DIRINDEX
 *                in that directory. If it finds such a file, it
 *                will return that instead. If it does not (or if
 *                it that file is also a directory) it returns NULL.
 *
 *                The result of stat()ing the file returned in path
 *                is returned in sb.
 *
 *  N.B. The access control mechanism is currently both lazy and
 *       paranoid; specifically, it just checks to make sure the
 *       pathname doesn't contain a ".." component. This is overly
 *       restrictive, since something in a subdirectory might (validly)
 *       contain a relative URL of the form "../foo.html", and a lazy
 *       browser might just tack that on to a relative URL. (Although
 *       none that I know of do.)
 *
 *       Also, no attempt is made to ensure that a symlink doesn't lead
 *       outside of ROOTDIR.
 */

char *file_lookup(char *path, struct stat *sb)
{
    char        *temp;
    int         result;

    if (*path != '/')
	/* This isn't an absolute path by any strech of the imagination
	   (and we aren't a proxy server, so we don't check for
	   absolute URIs) */
	return(NULL);

    if ((temp = strstr(path, "..")) != NULL) {
	/* The absolute path contains "..", so check to make sure it's
	   not either "../" or "..\0" */
	if ( (*(temp + 2) == '/') ||
	     (*(temp + 2) == '\0') ) {
	    /* This is an attempt to go up a level, which might lead
	       out of ROOTDIR. Be both paranoid and lazy, and deny
	       the request */
	    return(NULL);
	}
    }

    temp = (char *) malloc(strlen(ROOTDIR) + strlen(path) + 1);

    strcpy(temp, ROOTDIR);
    strcat(temp, path);

    result = stat(temp, sb);
    if (result) {
	free(temp);
	return(NULL);
    }

    if (S_ISDIR(sb->st_mode)) {
	/* is a directory */
	char *temp2;

	temp2 = (char *) malloc(strlen(ROOTDIR) + strlen(path) +
				strlen(DIRINDEX) + 1);
	temp2 = strcpy(temp2, temp);
	temp2 = strcat(temp2, DIRINDEX);
	free(temp);
	temp = temp2;

	result = stat(temp, sb);
	if (result || S_ISDIR(sb->st_mode)) {
	    free(temp);
	    return(NULL);
	}
    }

    return(temp);
}

/*
 * deduce_type -- given a path, deduces the MIME type of the file
 *                (from the extension)
 */

char *deduce_type(char *path)
{
    char *bookmark;

    bookmark = rindex(path, '.');
    if (bookmark != NULL) {
	if (strcasecmp(bookmark, ".html") == 0) {
	    return(strdup("text/html"));
	} else if (strcasecmp(bookmark, ".htm") == 0) {
	    return(strdup("text/html"));
	} else if (strcasecmp(bookmark, ".gif") == 0) {
	    return(strdup("image/gif"));
	} else if (strcasecmp(bookmark, ".jpg") == 0) {
	    return(strdup("image/jpeg"));
	}
    }

    return(DEFAULT_TYPE);
}
