#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * server.c -- web server functionality for SimpleWWW
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This is the server interface for the SimpleWWW library. See
 * swww.h for a detailed description of the library API.
 *
 * History
 *    3 Jul 1996 -- 0.1 -- colorful first attempt
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <lwp/lwp.h>

#include "swww.h"
#include "util.h"
#include "server.h"

/*
 * SERVER FUNCTIONS
 *
 * SWWW_server_init
 *
 */
int SWWW_server_init(int num_fetchers,
				/* how many fetch
				   threads to create */
		     ds_safeq_t **request_q)
				/* queue for requests -
				   for passing to accept_req */
{
    ds_safeq_t *req_q;
    int i, rc;
    PROCESS pid;

    req_q = ds_safeq_create();
    assert(req_q);

    for (i=0; i<num_fetchers; i++) {
	rc = LWP_CreateProcess(fetch_url_LWP_shell, 65536,
			       LWP_NORMAL_PRIORITY,
			       (char *) req_q, "FETCH_URL", &pid);
	if (rc)
	    return rc;
    }

    *request_q = req_q;
    return 0;
}



/*
 * SWWW_accept_requests -- this function listens at the specified port,
 *                         and hands requests off to the specified
 *                         fetch_url_fn. If it returns, it will return
 *                         an error code indicating why it returned.
 *
 */
int SWWW_accept_requests(int port,        /* Port number to listen
					     for requests on */
			 SWWW_fetch_url_fn fetch_url,
                                          /* Function which should be used
					     to fetch URLs needed to
					     fulfill requests */
			 SWWW_threading_type_t threading,
                                          /* type of threading to be
					     used (currently either
					     NO_THREADING or LWP) */
			 ds_safeq_t *request_q)
				          /* request queue for workers */
{
    int                 sockfd;
    struct sockaddr_in  serv_addr;

    if(SWWW_version(101)) {
	/* if library version isn't 0.1.1, squawk */
	fprintf(stderr,
		"SWWW_accept_requests: "
		"not updated for correct SimpleWWW API!\n");
    }
    
    /* set up server connection */
    if((sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	perror("socket");
	return(-1);		/* XXX a general question throughout
				   is whether STDERR is an apropo
				   place to put errors. */
    }

    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port        = htons(port);

    if(bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	perror("bind");
	return(-1);			/* XXX again, dang error handling */
    }

    /* wait for requests */
    listen(sockfd, 5);
    
    while(1) {
	/* pass of requests to fetch_url_fn, with acceptdata_fn being write(),
	   to the sockfd of the request */
	char                *therequest, *retbuffer;
	int                 response_code, header_size, retbuffer_size;
	char                *the_url, *version;
	SWWW_method_t       the_method;
	SWWW_headers_t      *in_headers;
	struct sockaddr_in  cli_addr;
	int                 newsockfd, clilen;
	int                 result;
	
	if(threading == LWP) {
	    /* before doing the accept(), do an IOMGR_Select to
	       determine whether it's safe to do the accept(). */
	    int    rc;
	    fd_set readset;
	    
	    FD_ZERO(&readset);
	    FD_SET(sockfd, &readset);
	    
	    rc = IOMGR_Select(sockfd+1, (int *) &readset, NULL, NULL, NULL);
	    
	    if(rc != 1) {
		fprintf(stderr, "SWWW_accept_requests/select: got %d back "
			"from IOMGR_Select, for no good reason. "
		        "(Did you init IOMGR?)\n", rc);
		/* XXX should assert() */
	    }
	}
	
	clilen = sizeof(cli_addr);
	newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
	
	/* Parse request and send to fetch_url_fn */
	/* XXX eventually want to support POST, in which case retbuffer
           is important */
	result = parse_headers(newsockfd, threading,
			       &therequest, &header_size, &in_headers,
			       &retbuffer_size, &retbuffer);
	if (result < 0) {
	    /* there was some problem reading the headers,
	       probably a premature socket closure */
	    close(newsockfd);
	    goto cleanup;
	}

	result = parse_request(therequest, &the_method, &the_url, &version);
	if (result < 0) {
	    /* the request is malformed, and probably not from an HTTP
	       client. discard it. */
	    /* XXX is this the right response?
	       better response is 400 bad request */
	    close(newsockfd);
	    goto cleanup;
	}

	SWWW_DEBUG(fprintf(stderr,
			   "%d|%s|%s\n", the_method, the_url, version));
	
	if(threading == LWP) {
	    char procname[32];
	    fetch_params_t *parameters;
	    
	    /* set up parameters. this is allocated here, but deallocated by
	       the newly called function. */
	    parameters = (fetch_params_t *) malloc(sizeof(fetch_params_t));
	    
	    parameters->fetch_url     = fetch_url;
	    parameters->method        = the_method;
	    parameters->url           = strdup(the_url);
	    parameters->reqid         = newsockfd;
	    parameters->headers       = in_headers;
	    parameters->acceptheaders = (SWWW_acceptheaders_fn) send_headers;
	    parameters->acceptdata    = (SWWW_acceptdata_fn)    write;
	    parameters->buffer        = retbuffer; /* freed by LWP shell */

	    if (the_method == POST) {
		char *temp;

		temp = SWWW_gethvalue(in_headers, "Content-Length");
		if (temp != NULL) {
		    parameters->length = atoi(temp);
		    free(temp);
		} else
		    parameters->length = 0;

		parameters->bufsize   = retbuffer_size;
		parameters->fetchdata = (SWWW_fetchdata_fn) read;
	    }

	    sprintf(procname, "FETCH_URL (SOCKET %d)", newsockfd);
	    /* XXX I have no clue whether 65536 is a big enough stack. */
	    ds_safeq_enq(request_q, parameters);
	} else {
	    /* if we're not using LWP, just execute this here: */
	    struct SWWW_fetch_in  in;
	    struct SWWW_fetch_out out;

	    SWWW_SFI_INIT(&in);
	    in.sfi_method        = the_method;
	    in.sfi_url           = the_url;
	    in.sfi_reqid         = newsockfd;
	    in.sfi_headers       = in_headers;
	    in.sfi_threading     = threading;
	    in.sfi_acceptheaders = (SWWW_acceptheaders_fn) send_headers;
	    in.sfi_acceptdata    = (SWWW_acceptdata_fn) write;

	    if (the_method == POST) {
		char *temp;

		in.sfi_postinfo = (struct SWWW_postinfo *)
		    malloc(sizeof(struct SWWW_postinfo));
		in.sfi_postinfo->spi_postid = newsockfd;

		temp = SWWW_gethvalue(in_headers, "Content-Length");

		if (temp != NULL) {
		    in.sfi_postinfo->spi_length = atoi(temp);
		    free(temp);
		} else
		    in.sfi_postinfo->spi_length = 0;

		in.sfi_postinfo->spi_buffer    = retbuffer;
		in.sfi_postinfo->spi_bufsize   = retbuffer_size;
		in.sfi_postinfo->spi_fetchdata = (SWWW_fetchdata_fn) read;
	    }

	    response_code = (*fetch_url)(&in, &out);
	    
	    free(out.sfo_status);	/* not used */

	    SWWW_free_headerlist(in_headers);
	    SWWW_free_headerlist(out.sfo_headers);

	    if (the_method == POST)
		free(in.sfi_postinfo);

	    if (retbuffer != NULL)
		free(retbuffer);
	    
	    close(newsockfd);
	}
	
    cleanup:
	if (therequest != NULL)
	    free(therequest);
	
	if (the_url != NULL)
	    free(the_url);
	
	if (version != NULL)
	    free(version);
	
    } /* end while(1) */
}

/*
 * PRIVATE FUNCTIONS
 *
 * fetch_url_LWP_shell -- invoked when SWWW_accept_headers calls
 *                        LWP_CreateProcess. The scope of the LWP is
 *                        the duration of this function.
 *
 *                        this function invokes the fetch_url_fn.
 */
int fetch_url_LWP_shell(void *q) {
    int                   response_code;
    ds_safeq_t            *request_q = (ds_safeq_t *)q;
    fetch_params_t        *parameters;
    struct SWWW_fetch_in  in;
    struct SWWW_fetch_out out;

    while (1) {
	parameters = (fetch_params_t *)ds_safeq_deq(request_q);

	SWWW_DEBUG(fprintf(stderr,
			   "LWP fetching with reqid %d\n", parameters->reqid));
	
	SWWW_SFI_INIT(&in);
	in.sfi_method        = parameters->method;
	in.sfi_url           = parameters->url;
	in.sfi_reqid         = parameters->reqid;
	in.sfi_headers       = parameters->headers;
	in.sfi_threading     = LWP;
	in.sfi_acceptheaders = parameters->acceptheaders;
	in.sfi_acceptdata    = parameters->acceptdata;
	
	if (parameters->method == POST) {
	    in.sfi_postinfo = (struct SWWW_postinfo *)
		malloc(sizeof(struct SWWW_postinfo));
	    in.sfi_postinfo->spi_postid    = parameters->reqid;
	    in.sfi_postinfo->spi_length    = parameters->length;
	    in.sfi_postinfo->spi_buffer    = parameters->buffer;
	    in.sfi_postinfo->spi_bufsize   = parameters->bufsize;
	    in.sfi_postinfo->spi_fetchdata = parameters->fetchdata;
	}
	
	response_code = (*(parameters->fetch_url)) (&in, &out);
	
	/* free parameters */
	SWWW_DEBUG(fprintf(stderr, "LWP closing reqid %d\n",
			   parameters->reqid));
	close(parameters->reqid);
	
	free(parameters->url);
	if (parameters->buffer != NULL)
	    free(parameters->buffer);
	free(parameters);
	free(out.sfo_status);
	
	if (parameters->method == POST)
	    free(in.sfi_postinfo);
	
	SWWW_free_headerlist(parameters->headers);
	SWWW_free_headerlist(out.sfo_headers);
    }

    return(LWP_SUCCESS);
}

/*
 * parse_request -- given a request line, returns the method,
 *                  the url, and the version
 *
 */

int parse_request(const char *the_request, SWWW_method_t *the_method,
		  char **the_url, char **version) {
    char *scratch = strdup(the_request);
    char *bookmark, *bookmark2;

    /* sanity initializations */
    *the_method = METHOD_ERROR;
    *the_url    = NULL;
    *version    = NULL;

    bookmark = scratch;
    
    /* determine the method */
    while( (*bookmark != '\0') && (!isspace(*bookmark)) ) {
	bookmark++;
    }

    if(*bookmark != '\0') {
	*bookmark = '\0';		/* scratch now points to
					   the method string */
	
	if(strcmp(scratch, "GET") == 0) {
	    *the_method = GET;
	} else if(strcmp(scratch, "HEAD") == 0) {
	    *the_method = HEAD;
	} else if(strcmp(scratch, "POST") == 0) {
	    *the_method = POST;
	} else {
	    *the_method = METHOD_ERROR;
	    return(PR_UNKNOWN_METHOD); /* unknown method */
	}
    } else {
	return(PR_BAD_REQUEST);	 /* bad request */
    }

    /* determine the URL */
    
    while( (*bookmark != '\0') && (isspace(*bookmark)) ) {
	bookmark++;
    }
    bookmark2 = ++bookmark;
    
    while( (*bookmark != '\0') && (!isspace(*bookmark)) ) {
	bookmark++;
    }

    if(*bookmark != '\0') {
	*bookmark = '\0';
    } else {
	/* if we end here, this is an HTTP/0.9 request */
	*version = strdup("HTTP/0.9");
    }

    *the_url = strdup(bookmark2);
    
    /* determine the version */
    if(*version != NULL) {
	free(scratch);
	return(0);
    } else {
	while( (*bookmark != '\0') && (isspace(*bookmark)) ) {
	    bookmark++;
	}
	bookmark2 = ++bookmark;
	
	if(*bookmark != '\0') {
	    *version = strdup(bookmark2);
	    return(0);
	} else {
	    return(PR_BAD_REQUEST);	/* bad request */
	}
    }
}

int send_headers(int reqid, int status_code, char *status,
		 int header_size, SWWW_headers_t *headers)
{
    SWWW_headers_t *bookmark;
    static char    *crlf = "\r\n";
    int            result;
    
    write(reqid, status, strlen(status));
    write(reqid, crlf, 2);
    
    bookmark = headers;
    if(header_size) {		/* make sure we have headers */
	while(bookmark->next != NULL) {
	    write(reqid, bookmark->name, strlen(bookmark->name));
	    write(reqid, ": ", 2);
	    write(reqid, bookmark->value, strlen(bookmark->value));
	    write(reqid, crlf, 2);
	    bookmark = bookmark->next;
	}
    }

    result = write(reqid, crlf, 2);
    
    if (result < 0)
	return(-1);		/* attempt to send headers failed */
    
    return(0);
}
