#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * util.c -- general utilities for simplewww
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * History
 *   30 Jun 1996 -- 0.1 -- colorful first attempt
 *    3 Jul 1996 -- 0.2 -- util.h has been moved into swww.h.
 *                         SWWW_version() added.
 *    5 Jul 1996 -- 0.3 -- added parse_headers()
 *   15 Jul 1996 -- 0.4 -- modified parse_headers() to allocate a header list,
 *                         and added SWWW_free_headerlist()
 *   23 Jul 1996 -- 0.5 -- added LWP support to parse_headers()
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <lwp/lwp.h>

#include "swww.h"

int SWWW_debuglevel = 0;
#define SWWW_DEBUGLEVEL

#include "util.h"

/* from the "cs" library */
void cs_parsedate(char *, struct tm *, int, int, int);

/*
 * PUBLIC FUNCTIONS:
 *
 * SWWW_version -- accepts an expected version number from the application,
 *                 and compares it to the actual version number. If they
 *                 differ, a complaint is printed to stderr, and the
 *                 actual version number is returned. Otherwise, 0 is returned.
 */
int SWWW_version(int expectedversion) {
    int rc;
    
    rc = (expectedversion == SWWW_VERSION ? 0 : SWWW_VERSION);
    if(rc) {
	fprintf(stderr, "SWWW_version: SimpleWWW library version (%d) differs "
		"from expected version (%d)\n",
		SWWW_VERSION, expectedversion);
    }
    return(rc);
}

/*
 * SWWW_debugging -- sets the SWWW debugging level.
 *                   Level 0 prints no messages,
 *                   level 1 prints all messages.
 */
void SWWW_debugging(int debuglevel) {
    SWWW_debuglevel = debuglevel;
}

/*
 * SWWW_create_headerlist -- returns an empty header list which can be
 *                           manipulated with the other header list
 *                           functions.
 */
SWWW_headers_t *SWWW_create_headerlist(void)
{
    SWWW_headers_t *new;

    new = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
    if (new == NULL) {
	/* XXX should assert() */
	return(new);
    }

    new->name  = NULL;
    new->value = NULL;
    new->next  = NULL;

    return(new);
}

/*
 * SWWW_gethvalue -- accepts a list of headers (such as is generated by
 *                   SWWW_gethvalue), and the name of a header, and
 *                   returns the value of that header. If there is no
 *                   such header, NULL is returned. Calling code is
 *                   responsible for freeing the returned value.
 */
char *SWWW_gethvalue(SWWW_headers_t *headers,
		     const char *name)
{
    SWWW_headers_t *hbookmark = headers;
    
    if(hbookmark == NULL)	/* if there is no header list, there */
	return(NULL);		/* are no name/value pairs */
    
    while(hbookmark->next != NULL) {
	if(strcasecmp(name, hbookmark->name) == 0) {
	    return(strdup(hbookmark->value));	/* return the value */
	}
	hbookmark = hbookmark->next;
    }
    
    /* no luck */
    return(NULL);
}

/* SWWW_clearhvalue -- accepts a list of headers (such as is generated by
 *                     SWWW_fetch_url), and the name of a header.
 *                     The header is removed from the list of headers
 *                     if it exists.
 *
 *                     -1 is returned if the headers is removed, and
 *                      0 is returned if no headers match name.
 */

int SWWW_clearhvalue(SWWW_headers_t *headers, /* list of headers */
		     const char     *name)    /* name of the header which
						 we are clearing */
{
    SWWW_headers_t *hbookmark = headers;
    SWWW_headers_t *previous;
    
    if(hbookmark == NULL)	/* if there is no header list, there */
	return(0);		/* are no name/value pairs */
    
    previous = NULL;

    while(hbookmark->next != NULL) {
	if(strcasecmp(name, hbookmark->name) == 0) {
	    /* this is the one to remove */
	    if (hbookmark->next == NULL) {
		if (previous != NULL) {
		    previous->next = NULL;
		} else {
		    /* this is the only entry. best approach is to
		       just clear out entries and hope for the best. */
		    free(hbookmark->name);
		    hbookmark->name = NULL;
		    free(hbookmark->value);
		    hbookmark->value = NULL;
		}
	    } else {
		SWWW_headers_t *temp;
		free(hbookmark->name);
		hbookmark->name  = hbookmark->next->name;
		free(hbookmark->value);
		hbookmark->value = hbookmark->next->value;
		temp = hbookmark->next;
		hbookmark->next = hbookmark->next->next;
		free(temp);
	    }
	    return(-1);
	}
	previous  = hbookmark;
	hbookmark = hbookmark->next;
    }
    
    /* no luck */
    return(0);
}

/*
 * SWWW_sethvalue -- accepts a list of headers (such as is generated by
 *                   SWWW_fetch_url), the name of a header, and
 *                   the new value for that header. The header is
 *                   added if it does not already exist.
 *
 *                   1 is returned if the header is added, and
 *                   0 is returned if an existing header is modified.
 *
 */

int SWWW_sethvalue(SWWW_headers_t *headers,   /* list of headers */
		   const char     *name,      /* name of the header
						 which we are changing */
		   const char     *value)     /* new value for the header */
{
    SWWW_headers_t *hbookmark = headers;
    
    if(hbookmark == NULL)	/* if there is no header list, there */
	return(0);		/* are no name/value pairs */
    
    while(hbookmark->next != NULL) {
	if(strcasecmp(name, hbookmark->name) == 0) {
	    /* found it, now change it */
	    free(hbookmark->value);
	    hbookmark->value = strdup(value);
	    return(0);
	}
	hbookmark = hbookmark->next;
    }

    /* no luck, so add new header */
    hbookmark->name  = strdup(name);
    hbookmark->value = strdup(value);
    hbookmark->next  = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
    
    hbookmark        = hbookmark->next;
    hbookmark->name  = NULL;
    hbookmark->value = NULL;
    hbookmark->next  = NULL;
    return(1);
}

/*
 * SWWW_duplicate_headerlist -- given a pointer to the start of a list of
 *                              headers, it allocates space for a new copy
 *                              of the list, and returns a pointer to the
 *                              new list.
 */

SWWW_headers_t *SWWW_duplicate_headerlist(int            header_size,
					  SWWW_headers_t *headers)
{
    SWWW_headers_t *hbookmark = headers;
    SWWW_headers_t *hbookmark2;
    SWWW_headers_t *head;
    int            loop;

    if(headers == NULL)
	return(NULL);

    head       = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
    hbookmark2 = head;
    
    for(loop = 0; loop < header_size; loop++) {
	hbookmark2->name  = strdup(hbookmark->name);
	hbookmark2->value = strdup(hbookmark->value);
	hbookmark2->next  = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
	
	hbookmark  = hbookmark->next;
	hbookmark2 = hbookmark2->next;
    }
    
    hbookmark2->name  = NULL;
    hbookmark2->value = NULL;
    hbookmark2->next  = NULL;
    
    return(head);
}

/*
 * SWWW_free_headerlist -- given a pointer to the start of a list of headers,
 *                         it frees all memory used by that list (including
 *                         strings).
 */

void SWWW_free_headerlist(SWWW_headers_t *headers)
{
    SWWW_headers_t *hbookmark = headers;
    SWWW_headers_t *temp;
    
    if(headers == NULL)
	return;
    
    while(hbookmark->next != NULL) {
	if(hbookmark->name != NULL)
	    free(hbookmark->name);
	if(hbookmark->value != NULL)
	    free(hbookmark->value);
	
	temp = hbookmark;
	hbookmark = hbookmark->next;
	
	free(temp);
    }
}

void SWWW_print_headerlist(SWWW_headers_t *headers)
{
    SWWW_headers_t *hbookmark = headers;
    
    if(headers == NULL)
	return;
    
    while(hbookmark->next != NULL) {
	if(hbookmark->name != NULL && hbookmark->value != NULL)
	    fprintf(stderr, " %s: %s\n", hbookmark->name, hbookmark->value);
	hbookmark = hbookmark->next;
    }
}

/*
 * SWWW_pack_headers -- given a list of headers, return a packed buffer.
 *                      NOTE: the caller must free the returned buffer.
 */

void SWWW_pack_headers(SWWW_headers_t  *headers,
	               char           **packed_headers,
		       int             *size)
{
    char           *bookmark;
    SWWW_headers_t *hbookmark;

    /* figure out how big the headers are */
    hbookmark = headers;
    *size      = 0;

    while (hbookmark->next != NULL) {
	*size += strlen(hbookmark->name) + 1;
	*size += strlen(hbookmark->value) + 1;
	hbookmark = hbookmark->next;
    }

    *size += 1;

    /* allocate an appropriately size chunk of memory */
    *packed_headers = (char *) malloc(sizeof(char) * (*size));

    /* run down the list and wrap 'em up */
    hbookmark = headers;
    bookmark  = *packed_headers;

    while (hbookmark->next != NULL) {
	strcpy(bookmark, hbookmark->name);
	bookmark += strlen(hbookmark->name) + 1;
	strcpy(bookmark, hbookmark->value);
	bookmark += strlen(hbookmark->value) + 1;
	hbookmark = hbookmark->next;
    }

    *bookmark = 0;		/* trailing NULL to indicate end,
				   if number of entries isn't known */
}

/*
 * SWWW_unpack_headers -- given packed headers, returns a list of headers
 *                        header_size is the number of headers packed;
 *                        if unknown, use -1. (there may be compatability
 *                        problems if -1 is used but SWWW_pack_headers was
 *                        not used to pack the headers, since some legacy
 *                        code this package is used doesn't indicate the
 *                        end clearly.)
 */

SWWW_headers_t *SWWW_unpack_headers(int header_size,
	                            char *packed_headers)
{
    SWWW_headers_t *head, *hbookmark;
    char           *bookmark;
    int            loop;
    
    head = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
    hbookmark = head;
    
    bookmark  = packed_headers;

    if (header_size != -1) {
	for(loop = 0; loop < header_size; loop++) {
	    hbookmark->name  = strdup(bookmark);
	    bookmark += strlen(hbookmark->name) + 1;
	    hbookmark->value = strdup(bookmark);
	    bookmark += strlen(hbookmark->value) + 1;
	    hbookmark->next = (SWWW_headers_t *)
		                             malloc(sizeof(SWWW_headers_t));
	    hbookmark = hbookmark->next;
	}
    } else {
	while(*bookmark) {
	    hbookmark->name  = strdup(bookmark);
	    bookmark += strlen(hbookmark->name) + 1;
	    hbookmark->value = strdup(bookmark);
	    bookmark += strlen(hbookmark->value) + 1;
	    hbookmark->next = (SWWW_headers_t *)
		                             malloc(sizeof(SWWW_headers_t));
	    hbookmark = hbookmark->next;
	}
    }
    
    hbookmark->name  = NULL;
    hbookmark->value = NULL;
    hbookmark->next  = NULL;
    
    return(head);
}

/*
 * SWWW_parse_url -- given a url, parses the URL, returning pointers to the
 *                   scheme, the host (and the port number, if given), and
 *                   the path. Returns 0 if successful, and -1 if not.
 *                   Caller is responsible for deallocating scheme, host, and
 *                   path.
 */

int SWWW_parse_url(const char *url, char **scheme,
	      char **host, int *port, char **path)
{
    char *bookmark, *bookmark2, *scratch;
    
    /*
     * Parse the URL into (scheme, host, path)
     */
    scratch = strdup(url);
    
    /* First, extract scheme */
    bookmark = index(scratch, ':');
    if(bookmark == NULL) {
	free(scratch);		/* was malformed */
	return(-1);
    } else {
	*(bookmark++) = '\0';
	*scheme = strdup(scratch);
    }
    
    /* Next, extract the host */
    if(*bookmark != '\0' &&
       *(bookmark++) == '/' &&
       *(bookmark++) == '/' &&
       *bookmark != '\0') {
	/* That atrocity above ensures that the next two characters are '//',
	   and that the string doesn't terminate after the '//'. */
	bookmark2 = bookmark;	/* keep this around until we null-terminate
				   it, and can copy it out */
    } else {
	free(scratch);		/* was malformed */
	return(-1);
    }

    bookmark = index(bookmark, '/');
    if(bookmark == NULL) {
	*path = strdup("/");
    } else {
	*path = strdup(bookmark);
	*bookmark = '\0';
    }

    /* Now we can make a copy of the host info, and make sure it
       doesn't have any port info */
    *host = strdup(bookmark2);
    bookmark = index(*host, ':');
    if(bookmark != NULL) {
	*(bookmark++) = '\0';
	*port = atoi(bookmark);
    } else {
	*port = -1;		/* no port information */
    }

  free(scratch);
  return(0);
}

void SWWW_parse_date(char      *datestr,
                    struct tm *tm,
		    int        currtm,
		    int        past,
		    int        error)
{
    cs_parsedate(datestr, tm, currtm, past, error);
}

/*
 * PRIVATE FUNCTIONS
 *
 * parse_headers -- given an open file descriptor, it returns back list of
 *                  headers, and a buffer containing any data following the
 *                  headers which it may have inadvertantly read. (the calling
 *                  function will have to deal with the contents of that
 *                  buffer, if any, before getting further data off of the
 *                  open file descriptor.)
 *
 * TODO:
 *   * rename "status" references to something like "firstline", since
 *     in requests, it's actually the request.
 */
int parse_headers(int                   openfd,
		  SWWW_threading_type_t threading,
		  char                  **status,
		  int                   *header_size,
		  SWWW_headers_t        **headers,
		  int                   *retbuffer_size,
		  char                  **retbuffer)
{
    char           buffer[1024];	/* buffer for reading from network */
    char           hbuffer[1024];	/* buffer for header parsing */
    int            hbuffer_size;
    char           *bookmark, *bookmark2;
    char           *hbookmark = hbuffer;
    SWWW_headers_t *lbookmark;
    int            got_status, headers_parsed;
    int            remaining;

    /*
     * Allocate header list
     */
    *headers          = (SWWW_headers_t *) malloc(sizeof(SWWW_headers_t));
    (*headers)->name  = NULL;
    (*headers)->value = NULL;
    (*headers)->next  = NULL;
    lbookmark = *headers;
    
    /*
     * Read the headers, and fill in response, header_size, and headers
     */

    /* initialization */
    got_status = 0;		/* flag is unset until status line has
				   been dealt with */
    headers_parsed = 0;		/* flag is unset until we're through
				   the headers */
    *header_size = 0;		/* start count from 0 */
    hbuffer_size = 0;

    *retbuffer_size = 0;
    *retbuffer      = NULL;
    
    *status = NULL;
    
    /* get the first chunk of data */
    if(threading == LWP) {
	/* before doing the read, do an IOMGR_Select to determine whether
	   it's safe to do the read. */
	int    rc;
	fd_set readset;
	
	FD_ZERO(&readset);
	FD_SET(openfd, &readset);
	
	rc = IOMGR_Select(openfd+1, (int *) &readset, NULL, NULL, NULL);
	
	if(rc != 1) {
	    fprintf(stderr, "parse_headers/select: got %d back from "
		            "IOMGR_Select, for no good reason. "
		            "(Did you init IOMGR?)\n", rc);
	    /* XXX should assert() */
	}
    }

    remaining = read(openfd, buffer, 1023);
    if (remaining < 0) {
	return(PH_BROKEN_CONNECTION);
    }
    buffer[remaining] = '\0';	/* just in case */

    bookmark = buffer;
    
    while(!headers_parsed) {
	bookmark2 = index(bookmark, '\n'); /* despite the fact that headers
					      are terminated by CR LF, there's
					      a lot of servers that terminate
					      the line with just CR. So, we're
					      "tolerant". Still, there's brain
					      damage here, and heads should
					      roll. */
	if(bookmark2 == NULL) {
	    /* This header doesn't end in this buffer, so copy it over to
	       hbuffer, and fetch another set. Make sure we don't overflow
	       hbuffer. */
	    hbuffer_size += remaining;
	    if(hbuffer_size > 1024) {
		fprintf(stderr, "overflowed hbuffer!\n");
				/* XXX handle this more gracefully */
		exit(-1);
		/* XXX should assert() */
	    } else {
		memcpy(hbookmark, bookmark, remaining);
		hbookmark += remaining;
	    }
	    remaining = 0;
	} else {
	    /* found end of header. check for \r preceding the \n, and
	       then copy the header out */
	    if(*(bookmark2 - 1) == '\r') {
		*(bookmark2 - 1) = '\0';
		remaining -= 1;		/* if we don't do this, we'll never
					   count the \n */
	    } else {
		*bookmark2 = '\0';
	    }
      
	    hbuffer_size += strlen(bookmark);
	    if(hbuffer_size > 1024) {
		fprintf(stderr, "overflowed hbuffer!\n");
				/* XXX handle this more gracefully */
		exit(-1);
		/* XXX should assert() */
	    } else {
		strcpy(hbookmark, bookmark);
		remaining    -= (strlen(bookmark) + 1);
	    }
      
	    bookmark = bookmark2 + 1;
	    
	    /* OK. At this point, we've (finally) got a header in hbuffer.
	       Now, do something with it. */
	    
	    if(hbuffer_size == 0) {
		headers_parsed = 1;
	    } else if(got_status) {
		/* this isn't the status line, so treat as a regular header */
		char *temp, *tempbookmark;
		
		temp = strdup(hbuffer);
		
		tempbookmark = index(temp, ':');
		
		if(tempbookmark) {
		    /* if this is non-null, we're golden */
		    *tempbookmark = '\0';                    /* make it
								easy to
								pickup
								name */
		    while(isspace(*(++tempbookmark)));       /* skip past
								whitespace */
		    lbookmark->value = strdup(tempbookmark); /* get the
								value */
		} else {
		    lbookmark->value = strdup("SWWW FOUND MALFORMED HEADER");
		}
		
		lbookmark->name   = strdup(temp);
		lbookmark->next   = (SWWW_headers_t *)
		                        malloc(sizeof(SWWW_headers_t));
		
		/* set up next entry */
		
		lbookmark         = lbookmark->next;
		lbookmark->name   = NULL;
		lbookmark->value  = NULL;
		lbookmark->next   = NULL;
		(*header_size)++;
		
		free(temp);
	    } else {
		*status = strdup(hbuffer);
		got_status = 1;
	    }
	    
	    /* Clean up */
	    hbookmark = hbuffer;
	    hbuffer_size = 0;
	}
	
	if(headers_parsed) {
	    /* set up buffer to be returned, and return it */
	    if(remaining > 0) {
		*retbuffer = (char *) malloc(remaining * sizeof(char));
		memcpy(*retbuffer, bookmark, remaining);
		*retbuffer_size = remaining;
		return(0);
	    } else {
		*retbuffer = NULL;
		*retbuffer_size = remaining;
		return(0);
	    }
	}
	
	if(remaining == 0) {
	    /* we've chomped through the buffer, so get another
	       buffer's worth */
	    if(threading == LWP) {
		/* before doing the read, do an IOMGR_Select to
		   determine whether it's safe to do the read. */
		int    rc;
		fd_set readset;
		
		FD_ZERO(&readset);
		FD_SET(openfd, &readset);
		
		rc = IOMGR_Select(openfd+1, (int *) &readset,
				  NULL, NULL, NULL);
		
		if(rc != 1) {
		    fprintf(stderr,
			    "parse_headers/select: got %d back from "
			    "IOMGR_Select, for no good reason. "
			    "(Did you init IOMGR?)\n", rc);
		    /* XXX should assert() */
		}
	    }

	    remaining = read(openfd, buffer, 1023);
	    if (remaining < 0) {
		return(PH_BROKEN_CONNECTION);
	    }
	    buffer[remaining] = '\0';	/* just in case */
	    
	    bookmark = buffer;
	}
    } /* end while(!headers_parsed) */
    
    /* XXX shouldn't be able to get here; should assert() */
    return(-1);
}
