#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * cm.h -- public interface for Odyssey connection manager
 *
 * This package abstracts the kinds of connections that Odyssey
 * wardens might want to use, into virtual connections that are
 * managed by the package.
 *
 * The goal of the package is to centralize several common issues
 * associated with connections:
 *
 *   * bandwidth/latency measurements of the connection
 *   * failure recovery
 *   * mobility? (have the other end of a connection possibly migrate,
 *      or at least go over a different interface, in a transparent
 *      way?)
 *   * possibly other issues which we haven't thought of yet.
 */

#ifndef _CONN_MGR_H_
#define _CONN_MGR_H_


/*
 * DEFINITIONS/MACROS
 */

/* version */

#define CMGR_VERSION 2

/* error codes */

#define CMGR_SUCCESS       0	           /* success */
#define CMGR_NOTSUPP       (CMGR_SUCCESS-1)    /* unknown connection type */
#define CMGR_BADCONN       (CMGR_NOTSUPP-1)    /* invalid server id */
#define CMGR_BADSERVER     (CMGR_BADCONN-1)    /* invalid server spec to create */
#define CMGR_CONNDOWN      (CMGR_BADSERVER-1)  /* connection is down */
#define CMGR_ERROR         (CMGR_CONNDOWN-1)  /* generic error */

/* defines for cmgr_{get,put} lock arg */
#define CMGR_LOCK   1
#define CMGR_NOLOCK 0

/* unique handle to a connection */
typedef long cmgr_id_t;

#define CMGR_CID_NONE      ((cmgr_id_t) -1)   /* to indicate error */

/* connection types supported */

typedef enum cmgr_type {
    CMGR_RPC2  = 13
} cmgr_type_t;

/* status of connection */
typedef enum cmgr_status {
    CMGR_ALIVE = 31,
    CMGR_DEAD = 37
}
cmgr_status_t;

/* connection direction */

typedef enum cmgr_direction {
    CMGR_IN    = 17,
    CMGR_OUT   = 23,
    CMGR_INOUT = 29
} cmgr_direction_t;

/* Information about a connection, including a real physical handle that has
   meaning to the underlying layers (such as RPC2)
*/

/*
 * Flags: used here and in cmgr_initparams_t 
 */
#define CMGR_CONNISLOCAL  0x1   /* Connection is local.  Don't estimate. */

typedef struct cmgr_connection {
    cmgr_type_t    cc_type;
    unsigned int   cc_flags;
    union {
	cmgr_rpc2_handle_t rpc2;
    }
    cc_handle;
    ds_rrlist_t  *cc_bwlist;
}
cmgr_connection_t;

/* signature for connstatus function, which is invoked when the
   status of a connection is changed (say, from CMGR_UP to CMGR_DOWN, or
   vice versa) */
typedef void  (*cmgr_connstatus_fn) __P((cmgr_connection_t *ccp,
				       cmgr_status_t      status));

/* abstract representation of initialization parameters */

typedef struct cmgr_initparams {
    cmgr_type_t ci_type;	/* connection type */
    struct timeval ci_restart;  /* min time to elapse between reconnection
				   attempts */
    int            ci_flags;    /* See below */
    union {			/* type-specific initialization parameters */
	cmgr_rpc2_initparams_t ci_rpc2;   /* RPC2 */
    } ci_params;
} cmgr_initparams_t;

#if 0
/* information about the connection */

typedef struct cmgr_conninfo {
    int            ci_bandwidth; /* in bytes/sec */
    struct timeval ci_latency;
    int            ci_traffic;   /* bytes transferred during measurement
				    window */
    struct timeval ci_winstart;
    struct timeval ci_winstop;   /* the measurement window */
    cmgr_status_t  ci_status;    /* status of connection */
} cmgr_conninfo_t;
#endif

/* Recent connection history information */
#define CMGR_MAXLOGLENGTH    (RPC2_MAXLOGLENGTH)
#define CMGR_MAXQUANTUM      (RPC2_MAXQUANTUM)

typedef enum NetLogTag    cmgr_logtag_t;
typedef RPC2_NetLogEntry  cmgr_logent_t;
typedef RPC2_NetLog       cmgr_log_t;

/*
 * GENERAL UTILITY FUNCTIONS
 *
 * General utility functions for the connection manager, such as
 * initialization and cleanup.
 */

/*
 * cmgr_init -- performs any necessary initialization. must be called
 *              before using package. Takes the version number, to
 *              make sure interface matches implementation.
 */

void cmgr_init __P((int expected_version));

/*
 * CONNECTION MANAGEMENT
 *
 * Functions for the creation/destruction of connections.
 */

/*
 * cmgr_create -- creates a connection entity of the specified type, using
 *                the supplied initialization parameters.
 *
 *                returns a unique positive connection ID if successful.
 */

cmgr_id_t cmgr_create __P((cmgr_initparams_t *cip));

/*
 * cmgr_destroy -- will attempt to destroy a given connection
 *
 *                 returns 0 if the connection was destroyed, 1 if it was only
 *                 marked for destruction, -1 on error.
 */

int cmgr_destroy __P((cmgr_id_t connid));

/*
 * cmgr_register -- register a callback function to be invoked when the
 *                  connection changes state (say, from CMGR_UP to CMGR_DOWN).
 *                  more than one such function can be registered.
 *                  the function must be explicitly un-registered to
 *                  be removed (that is, it does not disappear once it
 *                  has been invoked).
 */

int cmgr_register __P((cmgr_id_t connid, cmgr_connstatus_fn *fp));

/*
 * cmgr_unregister -- removes the named callback function on the named
 *                    connection.
 *
 *                    returns CMGR_SUCCESS if successful. returns CMGR_NOTCONN
 *                    if it is invoked on an invalid connection. returns
 *                    CMGR_NOTFUNC if there is no such callback for the
 *                    connection.
 */

int cmgr_unregister __P((cmgr_id_t connid, cmgr_connstatus_fn *fp));

/*
 * cmgr_get -- given a connection ID, returns a connection
 *             description for it (if it already exists), or returns NULL
 *             (if it does not). This connection description can be used
 *             for connection-specific operations (such as obtaining a
 *             valid RPC2 handle to use for in making an RPC2 call).
 *
 *             If cmgr_get() returns a non-NULL value, then the caller must
 *             call cmgr_put() when done (i.e., after the RPC is made).
 */

cmgr_connection_t *cmgr_get __P((cmgr_id_t connid, int lock));

/*
 * cmgr_put -- indicates that the connection is not currently being used,
 *             must be called after a succesful cmgr_get()
 */

int cmgr_put __P((cmgr_id_t connid, int unlock));

/*
 * CONNECTION INFORMATION
 */

#if 0
/*
 * cmgr_getinfo -- returns information about the specified connection
 */

int cmgr_getinfo __P((cmgr_id_t connid, cmgr_direction_t dir, cmgr_conninfo_t  *cip));
#endif

/*
 * cmgr_getlogs
 * 
 * Get the round-trip-time and throughput logs from recent past
 * 
 * 
 */
int cmgr_getlogs __P((cmgr_id_t connid, 
		      cmgr_log_t *rttlog, 
		      cmgr_log_t *tplog));

/*
 * cmgr_isvalid -- check on current status of connection
 *               returns 1 if the connection is currently marked up, 0 if not,
 *               asserts if passed an invalid conn id.
 */

int cmgr_isvalid __P((cmgr_id_t connid));

/*
 * cmgr_validate -- ask connection manager to attempt to restart connection
 *                  different from cmgr_isvalid() in that it may block, trying
 *                  to restart connection, or if someone else is calling
 *                  cmgr_validate()
 *
 *                  returns 1 if the connection is up on return,
 *                  0 if it is down, asserts if passed an invalid conn id.
 */

int cmgr_validate __P((cmgr_id_t connid));

/*
 * cmgr_markdown -- mark a connection as down (unusable). Typically called
 *                  after an RPC on that connection has failed.
 *
 *                  returns 0 on success, -1 on error.
 */

int cmgr_markdown __P((cmgr_id_t connid));

/********** Iterators ************/

struct cmgr_iter_t; /* Opaque */

/*
 * cmgri_make
 * 
 * Get an iterator on the connection manager table.  
 *
 * XXX could be some nastiness about missing connections mid-iterator.
 * for the moment, we won't worry about this.
 */

struct cmgr_iter_t *cmgri_make __P(());

/*
 * cmgri_destroy
 *
 * Destroy an iterator.  Assert if not a valid iterator
 */
void cmgri_destroy __P((struct cmgr_iter_t *));

/*
 * cmgri_next
 *
 * Yeild the next valid connection, if one exists.  Returns zero on
 * success, ENOENT if there are no more connections for this iterator.
 * Assert if not a valid iterator.
 */
int cmgri_next __P((struct cmgr_iter_t *, cmgr_id_t *));

/*
 * cmgr_settimeout
 * 
 * Set the timeout value for a connection
 * 
 * 
 */
int cmgr_settimeout __P((cmgr_id_t connid,
		      struct timeval tv));

/*
 * cmgr_islocal
 *
 * Returns true if the specified connection is local 
 *
 */
int cmgr_islocal __P((cmgr_id_t connid));

#endif /* _CONN_MGR_H_ */

