#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _ODYSSEY_MULTRES_H_
#define _ODYSSEY_MULTRES_H_

/*
   odyssey_multres.h - multiple-resource user-library definitions

   Include file for applications that monitor multiple resources 
   simultaneously.  It provides an interface to a user-level library
   which allows applications to register requests for windows on
   allowable functions over multiple resources.  Any function
   that is monotonically increasing (or decreasing) with respect to
   any input resource is allowable as an evaluation function. 

*/

/* Include odyssey.h for resource monitoring definitions. */
#include "include/odyssey.h"

/* Include sys/upcall.h for upcall blocking definitions */
#include "kernel/upcall/includes/upcall.h"

/* Current version of multiple-resource interface */
#define ODY_MULTRES_VERSION 0

/* This structure is used to specify information about one of the
   resources that will be monitored as part of a multiple-=resource
   request.  It contains a version numnber and the resource id of the
   resource to be monitored. */

typedef struct ody_req_res {
    int              orr_version;       /* which version of the interface */
    unsigned int     orr_resource;      /* resource number */
} ody_req_res_t;

/* This structure is used to specify the current value of a resource
   to a multiple-resource evaluation function.  It contains a version
   number, as well as the resource id and current value of a
   resource. */

typedef struct ody_req_res_val {
    int              orrv_version;      /* which version of the interface */
    unsigned int     orrv_resource;     /* resource number */
    int              orrv_value;        /* current value of the resource */
} ody_req_res_val_t;

/* Prototype for a user-supplied multiple-resource evaluation
   function.  The function takes a list of resources as input and
   returns the evaluation as a floating-point value.  Only functions
   which are monotonically increasing (or decreasing) with respect to
   each individual resource may be specified as an evaluation function. */

typedef float (*ody_eval_fn_t) __P((int resnum, ody_req_res_val_t* resources));

/* Prototype for the multiple-resource function to be called when the
   window is exceeded.  The function takes a list of resources and
   their respective values that must be deallocated by the application
   programmer after use. */
typedef void (*ody_mreq_fn_t) __P((int reqid, int resnum, 
				   ody_req_res_val_t* resources));

/* This structure is used to specify a multiple-resource request.  It
   differs from the single-resource request in that the user must specify
   an array of resources to be monitored as well as an evaluation function
   that will take the current values of the specified resources as input
   and return a numberic value.  The low and high bounds are evaluated
   with respect to the evaluation function output. */

typedef struct ody_req_mres {
    struct ody_vers    orm_version;     /* which version of the interface */
    int                orm_resnum;      /* number of resources specified */
    ody_req_res_t*     orm_resources;   /* array of resources to monitor */
    float              orm_low;         /* lower bound on function value */
    float              orm_high;        /* upper bound on function value */
    ody_eval_fn_t      orm_low_eval_fp; /* user-supplied evaluation func. */
    ody_eval_fn_t      orm_high_eval_fp; /* user-supplied evaluation func. */
    ody_mreq_fn_t      orm_req_fp;      /* call this when bound is violated */
} ody_req_mres_t;

/* Function prototypes for multi-resource requests.  These are
   essentially parallel to the single-resource request functions. */

int ody_mrequest  __P((char* path,      /* filename for odyssey object */
		       ody_req_mres_t* req, /* request data */
		       int* reqid,      /* request id returned on success */
		       ody_req_res_val_t** val /* resource values */
		       ));

int ody_fmrequest __P((int fd,          /* descriptor for odyssey object */
		       ody_req_mres_t* req, /* request data */
		       int* reqid,      /* request id returned on success */
		       ody_req_res_val_t** val /* resource values */
		       ));

int ody_mcancel   __P((int reqid));

#define REQUEST_BLOCK() upcall_block()
#define REQUEST_UNBLOCK() upcall_unblock()

#endif





