#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
** odytypes.h
**
** the every-odyssey-file-should-have-them definitions
**
** Types: bool        boolean variables
**        magic_t     magic number tags for structures
**        CMPFN       comparison function of two anonymous objects
**        TIMEVAL     a time value; currently a struct timeval
**                    might change with better machines.  Full precision.
**
** Macros: ASSERT     use instead of assert.h
**         ALLOC      use instead of malloc for objects
**         NALLOC     use instead of malloc for arrays
**         BALLOC     use instead of malloc for (void *) blocks of memory
**         ZALLOC, ZNALLOC, ZBALLOC   zero the allocated memory
**         FREE       use instead of free
**         TIME       get current time
**         TIMEDIFF   difference in time, in TIMEVAL units.
*/

#ifndef _ODYTYPES_H_
#define _ODYTYPES_H_

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>

/**************************************** prototypes */
#ifndef __P
#ifdef __STDC__
#define __P(args)     args
#else /* __STDC__ */
#define __P(args)     ()
#endif /* __STDC__ */
#endif /* __P */


/**************************************** booleans */
#undef TRUE
#undef FALSE
typedef enum {TRUE =1, FALSE=0} bool;

/*************************************** magic number tags */
typedef unsigned long magic_t;

/*************************************** Common function pointer types */

#ifdef __STDC__

typedef long (*COMPFN)(void*,void*);  /* used by data structure routines */

#else /* __STDC__ */

typdef int (*COMPFN)();

#endif /* __STDC__ */

/*************************************** TIMEVAL */

typedef struct timeval TIMEVAL;

/*************************************** assertions */
#ifndef __FILE__
#define __FILE__ "()"
#endif

#ifndef __LINE__
#define __LINE__ 0
#endif

#ifdef __STDC__

#define ASSERT(cond) 						      \
do {                                                                  \
    if (!(cond)) {                                                    \
	int *j = 0;                                                   \
	fprintf(stderr, "Assert (%s) in %s:%d\n",                     \
		#cond, __FILE__, __LINE__);                           \
	fflush(stderr);                                               \
	*j = 1; /* cause a real SIGTRAP to happen: can be caught */   \
    }                                                                 \
} while (0)

#else /* __STDC__ */

#define ASSERT(cond) 						      \
do {                                                                  \
    if (!(cond)) {                                                    \
	int *j = 0;                                                   \
	fprintf(stderr, "Assert failed in %s:%d\n",                   \
		__FILE__, __LINE__);                                  \
	fflush(stderr);                                               \
	*j = 1; /* cause a real SIGTRAP to happen: can be caught */   \
    }                                                                 \
} while (0)

#endif /* __STDC__ */

/*************************************** Allocation, deallocation, validity */

#define ALLOC(X,T)                                                \
do {                                                              \
    if (((X) = (T *)malloc(sizeof(T))) == NULL)                   \
	ASSERT(0);                                                \
} while (0)

#define ZALLOC(X,T)                                               \
do {                                                              \
    if (((X) = (T *)calloc(1, sizeof(T))) == NULL)                \
	ASSERT(0);                                                \
} while (0)

#define NALLOC(X,T,S)                                  \
do {                                                   \
    if (((X) = (T *)malloc(sizeof(T)*(S))) == NULL) {  \
	ASSERT(0);                                     \
    }                                                  \
} while (0)

#define ZNALLOC(X,T,S)                                 \
do {                                                   \
    if (((X) = (T *)calloc((S),sizeof(T))) == NULL) {  \
	ASSERT(0);                                     \
    }                                                  \
} while (0)

#define BALLOC(X,S)                                    \
do {                                                   \
    if (((X) = (void *)malloc((S))) == NULL) {          \
	ASSERT(0);                                     \
    }                                                  \
} while (0)

#define ZBALLOC(X,S)                                    \
do {                                                   \
    if (((X) = (void *)calloc((S),1)) == NULL) {          \
	ASSERT(0);                                     \
    }                                                  \
} while (0)

#define FREE(X)           \
do {                      \
    if ((X) != NULL) {    \
	free((X));        \
	(X) = NULL;       \
    }                     \
} while (0)

/*************************************** time operations */

/* tvp should be of type TIMEVAL */
#define TIME(tv)    (gettimeofday(&(tv),NULL))

/* tvb, tve, tvr should be of type TIMEVAL */
/* tve should be >= tvb. */
#define TIMEDIFF(tvb,tve,tvr)                                    \
do {                                                             \
    ASSERT(((tve).tv_sec > (tvb).tv_sec)                         \
	   || (((tve).tv_sec == (tvb).tv_sec)                    \
	       && ((tve).tv_usec > (tvb).tv_usec)));             \
    if ((tve).tv_usec < (tvb).tv_usec) {                         \
	(tvr).tv_usec = 1000000 + (tve).tv_usec - (tvb).tv_usec; \
	(tvr).tv_sec = (tve).tv_sec - (tvb).tv_sec - 1;          \
    } else {                                                     \
	(tvr).tv_usec = (tve).tv_usec - (tvb).tv_usec;           \
	(tvr).tv_sec = (tve).tv_sec - (tvb).tv_sec;              \
    }                                                            \
} while (0)


/* Operations on timevals. */

#undef timerclear
#undef timerisset
#undef timercmp
#undef timeradd
#undef timersub

#define timerclear(tvp)         (tvp)->tv_sec = (tvp)->tv_usec = 0
#define timerisset(tvp)         ((tvp)->tv_sec || (tvp)->tv_usec)
#define timercmp(tvp, uvp, cmp)                                         \
        (((tvp)->tv_sec == (uvp)->tv_sec) ?                             \
            ((tvp)->tv_usec cmp (uvp)->tv_usec) :                       \
            ((tvp)->tv_sec cmp (uvp)->tv_sec))
#define timeradd(tvp, uvp, vvp)                                         \
        do {                                                            \
                (vvp)->tv_sec = (tvp)->tv_sec + (uvp)->tv_sec;          \
                (vvp)->tv_usec = (tvp)->tv_usec + (uvp)->tv_usec;       \
                if ((vvp)->tv_usec >= 1000000) {                        \
                        (vvp)->tv_sec++;                                \
                        (vvp)->tv_usec -= 1000000;                      \
                }                                                       \
        } while (0)
#define timersub(tvp, uvp, vvp)                                         \
        do {                                                            \
                (vvp)->tv_sec = (tvp)->tv_sec - (uvp)->tv_sec;          \
                (vvp)->tv_usec = (tvp)->tv_usec - (uvp)->tv_usec;       \
                if ((vvp)->tv_usec < 0) {                               \
                        (vvp)->tv_sec--;                                \
                        (vvp)->tv_usec += 1000000;                      \
                }                                                       \
        } while (0)

#endif /* _ODYTYPES_H_ */
