#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Onode Manager
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This is the public interface for an onode management package for use
 * by wardens. The underlying assumption here is that most wardens will
 * want very similar mechanisms for creating, looking up, and deallocating
 * onodes.
 *
 * History
 *   31 Jul 1996 --  0. 0. 1 -- Colorful first attempt
 *
 *   16 Aug 1996 --  0. 0. 2 -- Simplified locks to read and write;
 *                              if we want more later, we can always
 *                              add them. Changed lock functions into
 *                              macros.
 *
 *                              Removed reference count from
 *                              onode_record, since it's already
 *                              in the onode itself (and this way,
 *                              there's no need to try to refind the
 *                              onode_record from the onode).
 *
 *                              Added onode_open and onode_close macros.
 *
 *   19 Aug 1995 --  0. 0. 3 -- Added critical sections (by hand, since
 *                              I had *no* desire to muck with the code.)
 *                              Changed type signatures slightly to conserve
 *                              stack space.
 *
 *                              Got a first cut at onode creation/deletion
 *                              based on struct onode contents, not just
 *                              bookkeeping.  This involved a slight change
 *                              to onode_free.
 *
 * TODO
 *   * failure modes/error codes?
 *   * better TRUE/FALSE than 1/0?
 *   * when things can be gc'd, need to tell kernel is gone
 * 
 * PUSHED ON THE STACK
 *   * should lock/promote be able to return immediately if lock cannot
 *     be obtained (probably yes). what are the correct semantics?
 *     This will probably be highly influenced by LWP's locking semantics.
 *   * LWP lock package does not have facilities for promotion/demotion;
 *     if I use it, these calls will make no sense.
 */

#ifndef _ONODE_MANAGER_H_
#define _ONODE_MANAGER_H_

/*
 * Common defines, enums and structs
 */

#define ONODE_MGR_VERSION 000003 /* 0.0.3 */

struct onode_create_in {
    /* Ofid is already part of the argument list */
    u_int32_t       oc_flags; /* Same os o_flags */
    enum ody_vtype  oc_vtype;
    ofree_t         oc_ofree;
    void            *oc_data;
};

/*
 * MACROS
 */

/*
 * Locking macros
 *
 * The locking semantics are that there can be multiple read locks,
 * xor a single write lock.
 *
 * onode_readlock -- given an onode pointer, obtains a read lock on it.
 *                   
 */

#define onode_readlock(op) \
do { \
    ds_hwlock_read(&op->o_lock); \
} while (0)

/*
 * onode_writelock -- given an onode pointer, obtains a write lock on it.
 *                   
 */

#define onode_writelock(op) \
do { \
    ds_hwlock_write(&op->o_lock); \
} while (0)

/*
 * onode_release_read -- given an onode pointer, release the lock held on it.
 *                   
 */

#define onode_release_read(op) \
do { \
    ds_hwlock_release_read(&op->o_lock); \
} while (0)

/*
 * onode_release_read -- given an onode pointer, release the lock held on it.
 *                   
 */

#define onode_release_write(op) \
do { \
    ds_hwlock_release_write(&op->o_lock); \
} while (0)

/*
 * Open/Close reference count macros
 */

/*
 * onode_open -- given an onode pointer, increments the open count
 *               (and the reference count).
 * Must be called with locks held
 */

#define onode_open(op, mode) \
do { \
  onode_writelock(op); \
  op->o_refcnt += 1;   \
  op->o_ocount += 1;   \
  if (mode & (O_O_WRITE | O_TRUNC)) { \
    op->o_owrite += 1; \
  } \
  onode_release_write(op); \
} while (0)

/*
 * onode_close -- given an onode pointer, decrements the open count
 *                (and the reference count).
 * Must be called with locks held
 */

#define onode_close(op, mode) \
do { \
  onode_writelock(op); \
  op->o_refcnt -= 1;   \
  op->o_ocount -= 1;   \
  if (mode & (O_O_WRITE | O_TRUNC)) { \
    op->o_owrite -= 1; \
  } \
  onode_release_write(op); \
} while (0)

/*
 * GENERAL UTILITY FUNCTIONS
 *
 * The following functions are of general utility.
 */

/*
 * onode_init -- initializes the Onode Manager. Must be called at least
 *               once before the package is used. Invocations after
 *               the package is initialized have no effect.
 * XXX - shouldn't we trip on that instead?
 */

int onode_init(void);

/*
 * onode_finalize -- cleans up the Onode Manager.
 */

int onode_finalize(void);

/*
 * onode_version -- takes an expected version number and compares it against
 *                  the current version number. If they differ, it complains
 *                  to stderr and returns the correct version number.
 *                  If they match, 0 is returned.
 */

int onode_version(int expectedversion);

/*
 * ONODE STORAGE MANAGEMENT
 *
 * The following functions provide a general mechanism for allocating,
 * retrieving, and freeing onodes.
 */

/*
 * onode_create -- given an ofid, allocates storage for it and returns
 *                 a pointer to that storage. The second argument is
 *                 a list of required parameters for onodes -- by
 *                 supplying these at creation time, we can be
 *                 relatively confident that all onodes are valid onodes.
 *                 if the onode already exists, return NULL.
 */
struct viceroy_onode *onode_create(struct ofid *new_ofid,
				   struct onode_create_in *in);

/*
 * onode_get -- given an ofid, returns the onode for it. if no such onode,
 *              returns NULL.
 */
struct viceroy_onode *onode_get(struct ofid *some_ofid);

/*
 * onode_put -- given an onode, indicates that the calling function
 *              is no longer interested in it. should be called after
 *              any call to onode_create or onode_lookup.
 *
 *              (n.b. this decreases a reference count which, when it
 *              reaches zero, means that the onode is fair game for
 *              garbage collection.)
 */
int onode_put(struct viceroy_onode *some_onode);

#endif _ONODE_MANAGER_H_
