#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _WARDEN_H_
#define _WARDEN_H_

/*
 * This file defines the interfaces that a warden is expected to
 * provide and adhere to.  For now, this is a minimal set; it will
 * grow substantially over time.
 *
 * The basic idea: each warden provides a reference to a struct
 * warden, which defines it's function entry points, etc.  As you
 * write wardens, and find things that you think should properly be
 * the domain of the viceroy, let me know and we can add things to the
 * viceroy-interface that is available to wardens.  
 */

/* Error codes that warden methods could return, other than the standard
 error codes */

#define ODYW_MAX_ERROR -5000
#define ODYW_MIN_ERROR -4500

#define EMOUNTPT (ODYW_MAX_ERROR)    /* tried to cross a mount point for a
					tome */
#define EMOUNTEDON (ODYW_MAX_ERROR-1)/* tried to access ".." from root of a
					mounted tome */

/*
 * Need a forward decln.
 */
struct warden;

/*
 * The funcitons that a warden is expected to provide.  
 */

/*
 * Parse arguments
 *
 * At startup time, if the viceroy sees an argument that it doesn't
 * understand, it passes argc, argv[], and the index of the current
 * argument under consideration.  Is expected to return the number of
 * arguments eaten, if any, or -1 if some of the arguments were
 * intended for it but were malformed somehow.  (This will be
 * interpreted as a fatal viceroy error, and propt it to spit out a
 * usage string.)  The default implementation returns zero without
 * looking at anything.
 */
typedef int (*w_parseargs_fn)   __P((int cur, int argc, char *argv[]));
extern  int wdflt_parseargs     __P((int cur, int argc, char *argv[]));

/*
 * Initialization
 *
 * Takes a pointer-to-warden as an argument; this should be the warden
 * being initialized.  Returns zero if successful, nonzero if things
 * are bad.  A nonzero return code will be evaulated as an errno
 * number.  If there's no initialization to be done, use the default
 * implementation.  The only warden function that can be called before
 * init is parseargs.  The init function is responsible for setting up
 * anything needed in w_data.  
 */
typedef int (*w_init_fn)    __P((struct warden *));
extern  int wdflt_init      __P((struct warden *));

/*
 * Grab the connection id to a server storing the given onode.  Used
 * to check bandwidth and latency numbers by the default routine that
 * knows about RPC2 logs.
 *
 * If the onode has no connection id associated with it, this function
 * should return -1.
 * 
 * This call has been re-invigorated for use by the connection manager.
 */

typedef cmgr_id_t (*w_getcid_fn) __P((struct warden *wp,
				      struct viceroy_onode  *op));

/*
 * Print some summary statistics collected by the warden.  Can be
 * anything the warden wants them to be.  Should send the output to
 * the passed FILE *, and can ingore errors, etc.
 */
typedef void (*w_pstats_fn) __P((struct warden *wp, FILE *));

/*
 * Root:
 *
 * returns the onode that is the root onode of the named volume.  This
 * root onode should have been set up by the rootwarden's init
 * function.  Returns zero on success, some useful error otherwise.
 *
 * Locking/Reference discipline:
 *
 * The onode should be returned read-locked and referenced: the
 * viceroy promises to unlock and put the reference.  
 */

typedef int (*w_root_fn)    __P((struct warden *, tome_t,
				 struct viceroy_onode **));

/*
 * Lookup
 *
 * XXX - for this to work correctly, we need to change the OdyMiniCache
 * lookup routine.
 *
 * Pass in odp, a "directory" onode, and a string, name.  The lookup
 * routine is expected to look for some prefix of name in the
 * directory odp.  How it chooses to do that is up to it.  (One can
 * imagine using the same lookup fn for many different types.)
 *
 * The lookup function is free to try to cope with as many components
 * of the remaining name as possible, but it is obligated to try to
 * resolve at least one such component.  The handle remain should be
 * left pointing at the first character of an unresolved component.
 *
 * If one, (or more) components can be successfully resolved, the
 * function should return zero.  Should lookup encounter an
 * unresolvable name after resolving one or more entries, it should
 * leave remain pointing to the first character of that component, and
 * return success.  The next attempt at lookup will fail.  (We do this
 * to simplify error handling.)  Lookup is only obligated to return
 * nonzero when the very first component of name cannot be resolved in
 * the context of odp.
 *
 * The *last* onode that is discoverd by name resolution should be
 * returned in the op handle.  It should be returned read-locked and
 * referenced; the viceroy will unlock and put the onode.
 *
 * Locking/Reference discipline:
 * 
 * odp is passed in readlocked.  The called warden is free to boost
 * the lock if necessary.  opp should be readlocked, and odp unlocked
 * and put, before returning.  The viceory will release opp's readlock
 * and put opp.  In general, it should be possible to drop any directory
 * lock before obtaining a child lock, but you are only required to do so
 * when resolving ".." in the context of some directory, and you do
 * not also hold the tome lock.
 */

typedef int (*w_lookup_fn) __P((struct warden *wp,
				struct viceroy_onode *odp,
				struct ody_cred *cred,
				char          *name, 
				char          **remain,
				struct viceroy_onode **opp));

/*
 * Access:
 *
 * Given an onode and a user credential, check to see if user cred has
 * mode access on onode op.  Return either 0, meaning the user has the
 * requested rights, or EACCESS, meaning the user does not.  Mode can
 * be any of: VREAD, VWRITE, or VEXEC (on 'files' means "can run this",
 * on 'directories' means "can perform lookup in this.")
 *
 * The onode passed in should be readlocked: the caller is entirely
 * responsible for manipulating reference counts.
 * 
 * Note that this means wardens are free to implement their own access
 * control schemes: we may want a unified way (kerberlike) to
 * authenticate users' identities.  
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and readlocked, and the viceroy
 * will release the readlock and put the onode.  
 */

typedef int (*w_access_fn) __P((struct warden *wp,
				struct viceroy_onode *op,
				struct ody_cred  *cred,
				int           mode));

/* 
 * Open:
 *
 * Given an onode, open it with mode/flags in mode (These include
 * FREAD, FWRITE, and any of the modes O_NONBLOCK, O_APPEND, O_CREAT,
 * O_TRUNC, O_EXCL, O_SHLOCK, and O_EXLOCK.  The latter two should be
 * ignored.
 * 
 * If the file is marked as ON_UNCACHEABLE, then the warden is
 * entirely responsible for performing bookkeeping on its own.
 * Otherwise, the open will be redirected to the device/inode pair
 * marked in the onode.  An onode should either always be marked
 * as UNCACHEABLE, or it should never be.  If it is ever both, the kernel
 * will panic if it catches you.
 *
 * The onode to be opened will be referenced and write-locked by the
 * viceroy before calling the warden's open method, and the viceroy
 * will unlock and put the reference.  The viceroy will also take
 * responsibility for incrementing the openers count (and, if
 * necessary the writers count).
 *
 * If the open cannot be performed, return some sensible error
 * message, otherwise return 0.  
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  
 */

typedef int (*w_open_fn) __P((struct warden *wp,
			      struct viceroy_onode *op,
			      struct ody_cred  *cred,
			      int           mode));

/* 
 * Close:
 *
 * Given a previously opened onode, close it.
 * mode should be passing the mode/flags it was opened with.
 * (These include FREAD, FWRITE, and any of the modes O_NONBLOCK,
 * O_APPEND, O_CREAT, O_TRUNC, O_EXCL, O_SHLOCK, and O_EXLOCK.  
 * I think for now they can all be ignored.
 * 
 * If the file is not marked as ON_UNCACHEABLE, then the close will be
 * redirected to the device/inode pair marked in the onode (the warden
 * should still decrement o_ocount/o_owrite), otherwise the warden is
 * entirely responsible for performing bookkeeping on its own.
 *
 * The onode to be closed will be referenced and write-locked by the
 * viceroy before calling the warden's close method, and the viceroy
 * will unlock and put the reference.  The viceroy will also take
 * responsibility for decrementing the openers count (and, if
 * necessary the writers count).
 *
 * If the close cannot be performed, return some sensible error
 * message, otherwise return 0.  
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  
 */

typedef int (*w_close_fn) __P((struct warden *wp,
			       struct viceroy_onode  *op,
			       struct ody_cred  *cred,
			       int           mode));


/*
 * GetAttr
 *
 * Return a struct vattr for an onode
 *
 * vap is the attribute structure to fill.  You are allowed to fill in
 * a -1 if the field makes no sense, but try to fill it in rather than
 * just ignoring it.
 * 
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and readlocked, and the viceroy
 * will release the readlock and put the onode.  
 */
typedef int (*w_getattr_fn) __P((struct warden *wp, 
				 struct viceroy_onode  *op,
				 struct ody_vattr *vap,
				 struct ody_cred  *cred));

/*
 * Read-Write.
 * 
 * Read/Write bytes to an onode
 * 
 * rwflag is O_RW_WRITE or O_RW_READ, depending on the requested operation.
 * 
 * count is the number of bytes requested to be read/written.  The
 * number of bytes actually read/written should be returned in count.  
 * 
 * offset is the first byte in the file to be read/written.
 *
 * ioflag can have any of the IO flags defined in vnode.h.  The two of
 * interest are IO_UNIT and IO_NDELAY.  At the moment, neither of
 * these are supported by the underlying software.  We'll look into
 * supporting at least the latter soon.  (It looks easy.)
 *
 * data is a handle to a buffer which must be used as the source of
 * the data for a write, and may be used as the destination of data in
 * a read.  If the warden prefers, it can substitute a pointer to its
 * own buffer for reading, avoiding a copy.  (Confession: at the
 * moment it's copied anyway, but this should change soon.)
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  
 */

typedef int (*w_rdwr_fn) __P((struct warden *wp,
			      struct viceroy_onode  *op,
			      struct ody_cred  *cred,
			      int           rwflag,
			      int           *count,
			      int           offset,
			      int           ioflag,
			      void          **data));

/*
 * rdwr_dispose
 *
 * Declaration that the viceroy is no longer interested in a buffer
 * that was passed back in a previous rdwr call.  This is only called
 * if the previous read passed back a buffer different from that
 * allocated by the viceroy.
 * 
 * count is the number of bytes the warden actually returned in the buffer. 
 * 
 * offset is the offset within op where the run of count bytes begins.
 *
 * data is the actual buffer returned by the warden.
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  
 */

typedef int (*w_rdwr_dispose_fn) __P((struct warden *wp,
				      struct viceroy_onode  *op,
				      struct ody_cred  *cred,
				      int           count,
				      int           offset,
				      void          *data));

/*
 * ReadDir
 * 
 * Read directory contents.
 * 
 * count is the number of bytes requested to be read.  The number of
 * bytes actually read should be returned in count.
 * 
 * data is a buffer which must be used as the destination of data. The
 * data should consist of a series of dirent structures such as that
 * defined by getdirentries(2)
 *
 * offset is the byte position at which to start reading: can we rely
 * on this being a multiple of sizeof(struct dirent)????
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and readlocked, and the viceroy
 * will release the readlock and put the onode.  */

typedef int (*w_readdir_fn) __P((struct warden *wp,
			      struct viceroy_onode  *op,
			      struct ody_cred  *cred,
			      int           *count,
			      int           offset,
			      void          *data));

/*
 * rdwr_dispose
 *
 * Declaration that the viceroy is no longer interested in a buffer
 * that was passed back in a previous readdir call.  This is only called
 * if the previous read passed back a buffer different from that
 * allocated by the viceroy.
 * 
 * count is the number of bytes the warden actually returned in the buffer. 
 * 
 * offset is the offset within op where the run of count bytes begins.
 *
 * data is the actual buffer returned by the warden.
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and readlocked, and the viceroy
 * will release the readlock and put the onode.  
 */

typedef int (*w_readdir_dispose_fn) __P((struct warden *wp,
				      struct viceroy_onode  *op,
				      struct ody_cred  *cred,
				      int           count,
				      int           offset,
				      void          *data));

/*
 * Type-specific operations.
 *
 * A type-specific operation passes the following arguments:
 *
 *      The version number of the type's interface the client
 *        is using.
 *      A pointer to the operand onode
 *      The operation number
 *      The size of the argument buffer
 *      A pointer to the argument buffer
 *      A pointer to the maximum size the return buffer can be.
 *      A handle to put the return buffer into.
 *
 * The onode named in tsop will be referenced and write-locked by the
 * viceroy before calling the warden's tsop method, and the viceroy
 * will unlock and put the reference.
 *
 * The warden is responsible for allocating the return buffer.  It can
 * be no bigger than *retsz, but can be smaller.  The size of the
 * return buffer so allocated should be placed in *retsz.  There is no
 * default for this one.  
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  */

typedef int (*w_tsop_fn) __P((struct warden *wp, 
			      int           tsvnum, 
			      struct viceroy_onode  *op, 
			      struct ody_cred  *cred,
			      int           opnum, 
			      size_t        argsz, 
			      void          *argp,
			      size_t        *retsz, 
			      void          **retp));
    
/*
 * Since the wardens are responsible for allocation of return buffers,
 * they have to be responsible for deallocation too.  We pass the
 * opnum, arguments, and onode along in case that made a difference in
 * allocation.  
 *
 * The onode named in dispose will be referenced and write-locked by
 * the viceroy before calling the warden's dispose method, and the
 * viceroy will unlock and put the reference.  
 *
 * At the moment I can't imagine a warden returning other than zero,
 * so I'll assert if it happens.
 *
 * Locking/Reference discipline
 *
 * The onode is passed in referenced and writelocked, and the viceroy
 * will release the writelock and put the onode.  
 */

typedef int (*w_tsop_dispose_fn) __P((struct warden *wp, 
				      struct viceroy_onode  *op, 
				      struct ody_cred  *cred,
				      int           opnum, 
				      size_t        argsz, 
				      void          *argp, 
				      size_t        retsz, 
				      void          *retp));

/*
 */    
typedef int (*w_expand_fn) __P((struct warden *wp,
				struct viceroy_onode  *odp,
				struct ody_cred  *cred,
				char          *name, 
				char          **remain,
				struct viceroy_onode  **opp));

/*
 */    
typedef int (*w_cursexpand_fn) __P((struct warden *wp,
				    struct viceroy_onode  *op,
				    struct ody_cred  *cred,
				    size_t        size,
				    size_t        *retsz,
				    char          *buf));
	
/*
 */    
typedef int (*w_prefetch_fn) __P((struct warden *wp,
				  struct viceroy_onode  *op,
				  struct ody_cred  *cred,
				  int            mode));

/*
 * Tell me your best guess about the bandwidth/latency on the
 * connection to the server storing onode foo.  Returns zero on
 * success, ENOENT if that onode isn't known.  We need to think about
 * how this is supposed to work.  
 */

/* XXX - not spec'd yet */
				     
struct warden {
    unsigned int       w_typenum;     /* Type identifier */
    char               *w_name;       /* Name of warden, for diagnostics */
    unsigned int       w_flags;       /* Flags */

    /************** Administration */

    w_parseargs_fn     w_parseargs;
    char               *w_usage;       /* Usage string */
    /* Initialize the warden */
    w_init_fn          w_init;
    /* Get a connection id for a particular vnode */
    w_getcid_fn        w_getcid;
    /* Print a set of statistics to a file */
    w_pstats_fn        w_pstats;

    /*************** vnode operations */

    /* Give me the root onode (filled only by the root warden) */
    w_root_fn          w_root;
    /* Looking for onodes (Waiting for godot?)*/
    w_lookup_fn        w_lookup;
    /* Access control */
    w_access_fn        w_access;
    /* Status retrieval */
    w_getattr_fn       w_getattr;
    /* Open/close */
    w_open_fn          w_open;
    w_close_fn         w_close;
    /* RdWr */
    w_rdwr_fn          w_rdwr;
    w_rdwr_dispose_fn  w_rdwrdisp;
    /* ReadDir */
    w_readdir_fn       w_readdir;
    w_readdir_dispose_fn w_readdirdisp;

    /************** Odyssey Extensions */
    /* Type-specific operations */
    w_tsop_fn          w_tsop;
    w_tsop_dispose_fn  w_tsopdisp;
    /* Looking for cursor onodes (Waiting for godot?)*/
#ifdef SETS
    w_expand_fn        w_expand;
    w_cursexpand_fn    w_cursexpand;
    w_prefetch_fn      w_prefetch;
#endif

    /************** Fields for cache bookkeeping */
    /* XXX objection to putting these in a warden-accessible place?
       XXX problem was use of indexing another structure given "wp" */
    long               w_total_size;
    long               w_current_size;

    /************** A handle for hiding data... */
    void              *w_data;
};

#ifdef VICEROY

extern int            warden_parseargs __P((int cur, int argc, char *argv[]));
extern void           warden_init __P(());
extern struct warden *warden_get  __P((unsigned int type));
extern void           warden_usage __P((FILE *fp));
extern void           warden_stats __P((FILE *fp));

#endif /* VICEROY */

#endif /* _WARDEN_TAB_H_ */

