package odyssey.map;

import  java.awt.*;
import  java.io.*;
import  odyssey.awt.*;

/**
 * A map is a two-dimensional object which has two parts:
 *   1) an Image which is the more traditional geographical
 *      "map" of the area in question, and
 *   2) a two-dimensional array of values which represent
 *      attributes for each pixel in the Image.
 *
 * The intended use is that the values in (2) represent nominal
 * bandwidth for the locations depicted in (1). (2) need not be
 * at the granularity of 1-1 correspondence between each pixel
 * and each value; the value can be coarser grained.
 */

public class Map extends Canvas implements java.io.Serializable {
    static int        NORESULT    = -2;
    static int        TRANSPARENT = -1;

    BlockingImage     image;
    transient Image   overlay;
    int               values[][];
    int               vw, vh;
    int               iw, ih;

    String            regionNames[];
    int               regions[][];

    transient Point   pointer = null;
    transient Image   pointold;
    boolean           used = false;
    transient boolean recovered = false;

    /**
     * Creates a new Map with default values.
     */
    public Map(Image image, String name, String config)
    throws IOException
    {
	this(new BlockingImage(image, name), config);
    }

    public Map(BlockingImage image, String config)
    throws IOException
    {
	this(image, config, 1, 1);
    }

    public Map(Image image, String name, String config, int width, int height)
    throws IOException
    {
	this(new BlockingImage(image, name), config, width, height);
    }

    /**
     * Creates a new Map with a value grid with width * height regions.
     */
    public Map(BlockingImage bimage, String config, int width, int height)
    throws IOException
    {
	this.image = bimage;

	this.iw = image.getWidth();
	this.ih = image.getHeight();

	this.vw = width;
	this.vh = height;

	// initialize values
	values = new int[iw / vw][ih / vh];
	for (int x = 0; x < values.length; x++) {
	    for (int y = 0; y < values[0].length; y++) {
		values[x][y] = TRANSPARENT; // initially transparent
	    }
	}

        // initialize regions
        FileInputStream f    = new FileInputStream(config);
	BufferedReader  file = new BufferedReader(new InputStreamReader(f));

	String line;
	line = file.readLine();
	int size = Integer.parseInt(line);

	regionNames = new String[size];
	regions     = new int[size][4];

	for (int loop = 0; loop < size; loop++) {
	    line = file.readLine();
	    int index = 0;

	    int end = line.indexOf(' ');
	    regionNames[loop] = line.substring(index, end);
	    index = end + 1;

	    for (int loop2 = 0; loop2 < 4; loop2++) {
		end       = line.indexOf(' ', index);
		int value;
		if (end >= 0) {
		    value = Integer.parseInt(line.substring(index, end));
		}  else {
		    value = Integer.parseInt(line.substring(index));
		}
		regions[loop][loop2] = value;
		index = end + 1;
	    }
	}

	file.close();

	setSize(iw, ih);
    }

    private synchronized void recover()
    {
	if (recovered) {
	    return;
	}
	
	if (overlay == null) {
	    overlay = createImage(iw, ih);
	}

	Graphics g = overlay.getGraphics();
	g.drawImage(image.theImage, 0, 0, this);

	int ix, iy, value;
	for (int vx = 0; vx < values.length; vx++) {
	    for (int vy = 0; vy < values[0].length; vy++) {
		ix = vx * vw;
		iy = vy * vh;

		value = values[vx][vy];
		if (value != TRANSPARENT) {
		    g.setColor(new Color(255, 255 - value, 255 - value));
		    g.fillRect(ix, iy, vw, vh);
		}
	    }
	}

	g.dispose();

	recovered = true;
    }

    /**
     * Color an area with a value. If the granularity
     * of the values map is coarser than the granularity
     * of the image, then the value may cover more pixels
     * then the specifically indicated one.
     *
     * @returns the old color value
     */
    public int color(int x, int y, int value)
    {
	Graphics g;
	int      result = -1;

	// if it's bigger than the image than forget it
	if (x >= iw || y >= ih) {
	    return result;
	}

	// find where this lays on the values grid
	int vx = x / vw;
	int vy = y / vh;

	int ix = vx * vw;
	int iy = vy * vh;

	if (overlay == null) {
	    overlay = createImage(iw, ih);
	    g = overlay.getGraphics();
	    g.drawImage(image.theImage, 0, 0, this);

	    if (used) {
		recover();	// recover the contents of overlay
	    }
	} else {
	    g = overlay.getGraphics();
	}

	try {
	    result = values[vx][vy];
	    if (result == value) {
		// if we're recoloring it with the same color,
		// don't bother
		g.dispose();
		return NORESULT;
	    }

	    values[vx][vy] = value;
	    if (!used) {
		recovered = true;
		used = true;
	    }

	    if (value >= 0) {
		g.setColor(new Color(255, 255 - value, 255 - value));
		g.fillRect(ix, iy, vw, vh);
	    } else if (value == TRANSPARENT) {
		// transparent (no modulation)
		g.drawImage(image.theImage,
			    ix, iy, ix + vw, iy + vy,
			    ix, iy, ix + vw, iy + vy,
			    this);
	    }
	} catch (ArrayIndexOutOfBoundsException e) {
	}

	g.dispose();

	repaint(ix, iy, vw, vh);

	return result;
    }

    public int getColor(int x, int y)
    {
	int result = -1;

	// if it's bigger than the image than forget it
	if (x >= iw || y >= ih) {
	    return -1;
	}

	// find where this lays on the values grid
	int vx = x / vw;
	int vy = y / vh;

	try {
	    result = values[vx][vy];
	} catch (ArrayIndexOutOfBoundsException e) {
	}

	return result;
    }

    public boolean setPointer(int x, int y)
    {
	Image   background;

	if (pointer != null) {
	    if (x == pointer.x && y == pointer.y) {
		return false;	// hasn't changed
	    }
	}

	/*
	if (overlay == null) {
	    background = image.theImage;
	} else {
	    background = overlay;
	}

	Graphics g = background.getGraphics();

	if (pointold == null) {
	    pointold = createImage(21, 21);
	} else {
	    // redraw old image
	    g.drawImage(pointold, pointer.x - 10, pointer.y - 10, this);
	}

	*/
	pointer = new Point(x, y);

	/*
	// get old image
	Graphics old = pointold.getGraphics();
	old.drawImage(background, 0, 0, 21, 21,
		      x - 10, y - 10, x + 10, y + 10, this);

	// draw new image
	g.setColor(new Color(160, 32, 240)); // purple
	g.drawOval(x - 10, y - 10, 21, 21);
	g.drawLine(x - 10, y, x + 10, y);
	g.drawLine(x, y - 10, x, y + 10);

	// clean up
	old.dispose();
        g.dispose();

	repaint(x - 10, y - 10, 21, 21);
	*/

	return true;
    }

    /**
     * compute the value for a region
     */
    public int regionValue(int index)
    {
	int sum          = 0;
	int totalColored = 0;
	int totalClear   = 0;

	// find bounds in value grid
	int left   = regions[index][0] / vw;
	int top    = regions[index][1] / vh;
	int right  = regions[index][2] / vw;
	int bottom = regions[index][3] / vh;

	/* System.out.println("in0: " + regions[index][0] +
			  " in1: " + regions[index][1] +
			  " in2: " + regions[index][2] +
			  " in3: " + regions[index][3] +
			  " vw: "  + vw +
			  " vh: "  + vh); */

	/* System.out.println("left: "   + left +
			  " top: "    + top +
			  " right: "  + right +
			  " bottom: " + bottom); */

	// loop amongst the daffodils
	for (int x = left; x < right; x++) {
	    for (int y = top; y < bottom; y++) {
		if (values[x][y] == TRANSPARENT) {
		    totalClear++;
		} else {
		    totalColored++;
		    sum += values[x][y];
		}
	    }
	}

	// if more than 2/3 is transparent, then treat the block
	// as transparent
	/* System.out.println("clear: " + totalClear +
			  " color: " + totalColored +
			  " sum: "   + sum); */

	if (totalClear > (totalColored * 2)) {
	    return TRANSPARENT;
	} else {
	    return (sum / totalColored);
	}
    }

    /**
     * Create a config file for BWMGR
     */
    public void dumpConfigFile() throws IOException
    {
	FileWriter     f   = new FileWriter("bw-levels");
	BufferedWriter out = new BufferedWriter(f);

	for (int loop = 0; loop < regions.length; loop++) {
	    out.write(regionNames[loop]);
	    out.write("\t");
	    int value = regionValue(loop);
	    if (value == TRANSPARENT) {
		// unmodulated
		out.write("0");
	    } else if (value > 200) {
		// high
		out.write("667");
	    } else if (value > 150) {
		// medium
		out.write("1000");
	    } else {
		out.write("2500");
	    }
	    out.newLine();
	}

	out.close();
    }

    /**
     * Repaint the regions
     */
    public void repaintRegions()
    {
	Graphics g;

	if (overlay == null) {
	    overlay = createImage(iw, ih);
	    g = overlay.getGraphics();
	    g.drawImage(image.theImage, 0, 0, this);

	    /* if (used) {
		recover();	// recover the contents of overlay
	    } */
	} else {
	    g = overlay.getGraphics();
	}

	g.setColor(new Color(160, 32, 240)); // purple
	for (int loop = 0; loop < regions.length; loop++) {
	    g.drawRect(regions[loop][0], 
		       regions[loop][1], 
		       regions[loop][2] - regions[loop][0],
		       regions[loop][3] - regions[loop][1]);
	}
	g.dispose();
	repaint();
    }

    /**
     * Draw the map, with the values grid overlayed on top
     */
    public void paint(Graphics g)
    {
        if (recovered == false && used == true) {
	    recover();
	}

	if (overlay == null) {
	    g.drawImage(image.theImage, 0, 0, this);
	} else {
	    g.drawImage(overlay, 0, 0, this);
	}
    }
}
