#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <string.h>

#include <sys/param.h>
#include <sys/queue.h>
#include <sys/types.h>

#include <unistd.h>

#define _UPCALL_REG_IMPL  /* We want the Special Definitions */
#include "kernel/upcall/includes/upcall.h"

LIST_HEAD(upcall_chain_t, upcall_ent_t);

struct upcall_ent_t {
	LIST_ENTRY(upcall_ent_t)    ue_next;
	upclass_t                   ue_upclass;
	upcall_t                    ue_upcall;
	uc_hdlr_t                   ue_fn;
};

static char            inited=0; /* nonzero only if other statics meaningful */

extern unsigned sig_upcall; /* Signal on which upcalls are received */

/*
 * We expect O(64) upcalls to be registered for by any process.  I
 * suppose we could keep track, but I'm not going to just yet.
 */
#define START_SIZE            79      /* a prime not close to a power of 2 */
static struct upcall_chain_t  *htab;  /* Hash table for functions */
static unsigned int           hsize;  /* Size of same */

/* 
 * How big the default argument buffer should be.
 */
static int bufsz     =1024;

#define NEW_ENTRY(ep, cl, ca)                          \
do {                                                   \
	(ep) = (struct upcall_ent_t *)                 \
		malloc(sizeof(struct upcall_ent_t));   \
	if (ep) {                                      \
		ep->ue_upclass = (cl);                 \
		ep->ue_upcall = (ca);                  \
	}                                              \
} while (0)

#define FREE_ENTRY(ep)                 \
do {                                   \
	if (ep) {                      \
		(ep)->ue_fn = NULL;    \
		free(ep);              \
	}                              \
} while (0)

/* This is meant to be used by handle_sigupcall */
/*
 * If we crap out, and we are currently handling an upcall, we try to
 * terminate it as gracefully as possible.  We also shoot the program
 * that sees the error.
 */
#define ERROR(s, serial)                                   \
do {                                                       \
	char  msg[MAXPATHLEN];                             \
	snprintf(msg, MAXPATHLEN-1,                        \
		 "Process %d: upcall: %s", getpid(), (s)); \
	perror(msg);                                       \
        if (serial)                                        \
                __upcall_ret(serial, 0, NULL, errno);      \
	exit(errno);                                       \
} while (0)

#define hash(cl, ca)   ((((cl) << 8) + (ca)) % hsize)

/*
 * Basic idea:
 * loop until done:
 *         grab the next upcall
 *         fill the argument buffer if it is too big
 *         lookup the function; if none, return a code to panic
 *         dispatch
 *         blow in the return values
 *
 * If anything here fails, we need to toast the process. (?)
 */
static void
catch_sigupcall(sig, code, scp)
	int sig;
	int code;
	struct sigcontext *scp;
{
	upclass_t            upclass;
	upcall_t             upcall;
	size_t               argsz = bufsz;
	char                 argbuf[bufsz];
	size_t               retsz;
	serial_t             serial;
	void                *retbuf;
	void                *bigargbuf;
	int                  returned;
	int                  bucket;
	int                  flags;
	struct upcall_ent_t *uep;


	/* There are one or more upcalls waiting for us */
	do {
		if (__upcall_next(&upclass, &upcall, &argsz, argbuf, &retsz, 
				  &serial))
		{
			ERROR("__upcall_next", 0);
		}
		if (upclass == UC_UNUSED)
			return;

		/* Did we get the argument? */
		if (argsz > bufsz) {
			bigargbuf = malloc(argsz);
			if (!bigargbuf) {
				errno = ENOMEM;
				ERROR("allocate bigargbuf", serial);
			}
			if (__upcall_args(serial, argsz, bigargbuf)) {
				ERROR("__upcall_args", serial);
			}
		} else {
			bigargbuf = argbuf;
		}
		
		if (retsz) {
			retbuf = malloc(retsz);
			if (!retbuf) {
				errno = ENOMEM;
				ERROR("allocate retbuf", serial);
			}
		} else {
			retbuf = NULL;
		}

		bucket = hash(upclass, upcall);
		for (uep = htab[bucket].lh_first;
		     uep;
		     uep = uep->ue_next.le_next)
		{
			if (uep->ue_upclass == upclass &&
			    uep->ue_upcall == upcall)
			{
				break;
			}
		}
		if (!uep) {
			errno = EOPNOTSUPP; 
			/* 
			 * The kernel thought we had a handler for this
			 * upcall, but we don't.  We should probably
			 * panic in this case.
			 */
			ERROR("looking for handler", serial);
		}
		
		flags = 0;
		if (!retbuf) flags |= UPCALL_ASYNC;

		returned = (uep->ue_fn)(upclass, upcall, argsz, bigargbuf, 
					&retsz, retbuf, flags);

		if (__upcall_ret(serial, retsz, retbuf, returned)) {
			ERROR("__upcall_ret", 0);
		}
		
		if (bigargbuf != argbuf) free(bigargbuf);
		if (retbuf) free(retbuf);
			
	} while (1);

}


/*
 * set up the hash table for call/class -> function mapping, and
 * install the signal handler for sig_upcall.  On successful
 * completion, we know that the kernel understands (and supports)
 * sig_uipcall, so if it ever trips on it later, we should assert.  
 */

static int
init() 
{
	int              result=0;
	int              i;
	struct sigaction sa;

	/* 
	 * Block signals while we are working: we don't want some bozo
	 * sending us a sig_upcall until we're really ready to deal
	 * with one.  */
	if (upcall_block())
	{
		return errno;  /* Likely that sig_upcall not recognized */
	}

	/* Below this point goto exit on exception. */

	hsize = START_SIZE;
	htab = (struct upcall_chain_t *) 
		malloc (sizeof(struct upcall_chain_t) * hsize);

	if (!htab) {
		result = ENOMEM;
		goto exit;
	}

	/* Below this point goto exit2 on exception */

	for (i=0; i<hsize; i++) {
		LIST_INIT(&htab[i]);
	}

	sa.sa_handler = catch_sigupcall;
	if (sigemptyset(&sa.sa_mask)) {
		result = errno;
		goto exit2;
	}

	/* I think we always want to restart on this signal... */
	sa.sa_flags = 0;
	sa.sa_flags |= SA_RESTART;
	if (sigaction(sig_upcall, &sa, NULL) == -1) {
		result = errno;
		goto exit2;
	}

	/* To fall through here, we must have at least blocked sig_upcall */
 exit:
	if (upcall_unblock())
		return errno;
	if (!result) inited = 1;
	return result;

 exit2:
	free(htab);
	goto exit;
}

/* 
 * XXX - should we leave the signal blocked until there is (at least)
 * one upcall handler installed?  Should we block it long-term if
 * we remove all of them?  I think this will be the uncommon case, but
 * am not sure.
 */

void *
upcall_reg(upclass, upcall, fn)
	upclass_t upclass;
	upcall_t  upcall;
	uc_hdlr_t fn;
{
	int                  result=0;
	int                  bucket;
	struct upcall_ent_t *uep;
	void                *ofn;
	int                  regflag;

	/* handle initialization */
	if (!inited) {
		if ((result = init())) {
			errno = result;
			return (void*) -1;
		}
	}

	/* Block sig_upcall until we've set this one up. */
	assert(!upcall_block());

	/* 
	 * Grab the entry we'd like to get, if it exists. 
	 * Remember the bucket from where it came.
	 */
	bucket = hash(upclass, upcall);
	for (uep = htab[bucket].lh_first; uep ; uep = uep->ue_next.le_next)
	{
		if (uep->ue_upclass == upclass && 
		    uep->ue_upcall == upcall)
		{
			break;
		}
	}
		
	if (uep) {
		/* We already have a handler for this fn */
		ofn = uep->ue_fn;
	} else {
		/* We don't have a handler yet. */
		ofn = NULL;
		NEW_ENTRY(uep, upclass, upcall);
		if (!uep) {
			result = ENOMEM; /* Couldn't alloc handler */
			goto exit;
		}
	}

	/* Invariants:
	 * if ofn && fn, we are changing handlers. 
	 * if !ofn && fn, we are installing a new handler. 
	 * if ofn && !fn, we are removing a handler. 
	 * if !ofn && !fn, we aren't doing anything.
	 * uep points to a valid, but perhaps uninstalled, handler; it
	 *   needs to have it's fn pointer changed if we succeed.
	 */

	/* Check with the kernel to see if this is okay. */
	regflag = 0;
	if (fn) {
		regflag |= UPCALL_REGISTER;
	}

	/* 
	 * Catch the case where we are installing NULL, and used to 
	 * have NULL installed 
	 */
	if (ofn || fn) {
		if (__upcall_reg(upclass, upcall, regflag)) {
			result = errno;
			goto exit;
		}
	}
	
	/* 
	 * We are now going to succeed.  If there is a new handler, we
	 * need to set uep's fn, and possibly insert uep in the bucket.
	 * If there isn't a new handler, we need to possibly remove
	 * uep from the bucket, and then destroy it.
	 */
	if (fn) {
		/* New function; remember it. */
		uep->ue_fn = fn;
		if (!ofn) {
			LIST_INSERT_HEAD(&htab[bucket], uep, ue_next);
		}
	} else {
		if (ofn) {
			LIST_REMOVE(uep, ue_next);
		}
		FREE_ENTRY(uep);
	}

 exit:
	/* un-block sig_upcall */
	assert(!upcall_unblock());

	if (!result) {
		return ofn;
	} else {
		errno = result;
		return (void*)-1;
	}
}

void
upcall_print()
{
    int                    i;
    struct upcall_ent_t   *uep;
    

    printf("Printing registrations\n");
    for (i = 0; i < hsize; i++) {
	if (htab[i].lh_first) 
	    printf("Bucket %d\n", i);
	for (uep = htab[i].lh_first; uep; uep = uep->ue_next.le_next) {
	    printf("\tupclass %d, upcall %d, fn 0x%lx\n", uep->ue_upclass,
		   uep->ue_upcall, (unsigned long)uep->ue_fn);
	}	    
    }
    printf("-----------------------------------------------\n");
}
