#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_hash.h"
#include "kernel/sets/common/sets_domain.h"

/* NOTE: called via system call, so curproc is valid */
int setup_fd(struct setType *set, int setnum, int elem)
{
	extern struct fileops vnops;
	FILEOBJ_t f;
	int fd, err;

	err = falloc(curproc, &f, &fd);
	if (err) {
		return -err;
	}
	f->f_flag = FREAD;
	f->f_type = DTYPE_VNODE;
	f->f_ops = &vnops;
	f->f_data = (void *)set->resolved[elem].fsp;

	set->resolved[elem].fp = (FILEOBJ_t)curproc->p_fd->fd_ofiles[fd];
	set->resolved[elem].fd   = fd;

	/* get an extra ref to the file * so that we we'll have a
	 * valid reference to it until we releaseSet */
	((FILEOBJ_t)set->resolved[elem].fp)->f_count++;
	vref((struct vnode *)f->f_data);

	printf(" yield fp %p vp %p\n", f, f->f_data);
	printf(" yield fp_ref %d vp_ref %d\n", f->f_count,
	       ((struct vnode *)f->f_data)->v_usecount);

	return fd;
}

/* To allow rwip() to be able to pin pages before SETS can get to them, 
 * have another freelist on which to put pinned pages until SETS can place 
 * them on the sets' free list. This is important to prevent the last 
 * buffers of large files from flushing the first buffers before SETS can 
 * pin them.
 */

#include <sys/buf.h>

#ifdef __NetBSD__
TAILQ_HEAD(bqueues, buf) PinnedForSets;
#endif

void init_bufferManagement()
{
#ifdef __NetBSD__
	TAILQ_INIT(&PinnedForSets);
#endif
}

/*
 * Since prefetch buffers don't get reclaimed, sometime we may need to force
 * them out. For now, I do that when that last open set is closed.
 */
void purgePrefetchBuffers()
{
}

/* Routine to read a page from a directory. */
int sets_get_dir_page(FSOBJ_t dir, char *buf, int *diroffset,
		      PROCESS_t p)
{
	struct uio auio;
	struct iovec aiov[2];
	int error, eof;

	auio.uio_iov = aiov;
	auio.uio_iovcnt = 1;
	auio.uio_offset = *diroffset;
	auio.uio_resid = DEV_BSIZE;
	auio.uio_segflg = UIO_SYSSPACE;
	auio.uio_rw = UIO_READ;
	auio.uio_procp = p;

	aiov[0].iov_base = buf;
	aiov[0].iov_len = DEV_BSIZE;

	/* Assume the worker thread assumed the id of the caller */
#ifdef	__NetBSD__
	error = VOP_READDIR(dir, &auio, p->p_ucred, &eof, (off_t *)0, 0);
#elif	defined(__FreeBSD__)
	error = VOP_READDIR(dir, &auio, p->p_ucred, &eof, (int *)0,
			    (u_int **)0);
#endif
    
	if (error) {
		return -error;
	}

	*diroffset = auio.uio_offset;
	return DEV_BSIZE - auio.uio_resid; /* Number of bytes read. */
}

/* NOTE: open_object errors are returned as positive, because they
 * get to the user without going through the errno mechanism 
 * (via the error argument to setIterate) */

/* This does work similar to vn_open.  Can't just use vn_open because we may
 * need to do ody_prefetch instead of VOP_OPEN in the middle. */
int open_object(struct setType *myset, elemType *elem, int myindex)
{
	FSOBJ_t vp;
	int err;
	struct sets_domain *dom;
	struct nameidata nd;
	
	LOG_ENTRY(myindex, OPEN_MEMBER);

	/* need LOCKLEAF for VOP_ACCESS */
	NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_SYSSPACE, elem->name,
	       myset->owner);
	err = namei(&nd); /* side effects? need to unlock/put inode? */
	/* if this is ENOSPC and we're in ODYSSEY, don't want
	   to skip this element -- retry later. */
	if (err)
	    goto out;
	
	vp = nd.ni_vp;
	err = VOP_ACCESS(vp, VREAD, myset->owner->p_ucred,
			 myset->owner);
	if (err) {
	    vput(vp);
	    goto out;
	}
	
	dom = sets_find_domain(vp);
	if (PREFETCHING && OpenPrefetches(myset) && dom)
	    err = (dom->prefetch_fn)(vp, FREAD, myset->owner, NULL);
	else
	    err = OPEN_FSOBJ(vp, FREAD, curproc);
	
	if (err) {
	    vput(vp);
	    goto out;
	} else {
	    VOP_UNLOCK(vp);
	    elem->fsp = vp;
	}

	printf("vp_ref %d\n", vp->v_usecount);
out:
	LOG_EXIT(myindex, OPEN_MEMBER);
	return err;
}

/*
 *  notes
 *   - keep vp with 1 ref
 *   - if next_vp valid, return it with the 1 ref made by namei
 *   - if dvp non-null, return ref'ed parent vp
 *   - does the check for null return vp, and gives ENOENT -- so the
 *     caller doesn't need to check for null-vp
 */
int expand_part(FSOBJ_t vp, char *name, FSOBJ_t *vpp,
		workerMsgType *msg)
{
	struct nameidata nd;
	struct filedesc *fdp = msg->set->owner->p_fd;
	FSOBJ_t save_cdir;
	int error = 0;

	if (!vp)
	    if (name[0] == '/')
		vp = msg->set->rootdir;
	    else
		vp = msg->set->curdir;

	DEBUG(PATHNAME, printf("expand_part in %p for %s\n", vp, name););

	if (name[0] == '\0') {
		DEBUG(PATHNAME, printf("null expand_part\n"););
		*vpp = vp;
		FSOBJ_REF(*vpp);
		return 0;
	}
	save_cdir = fdp->fd_cdir;
	fdp->fd_cdir = vp;
	
	NDINIT(&nd, LOOKUP, FOLLOW|LOCKLEAF, UIO_SYSSPACE,
	       name, msg->set->owner);
	error = namei(&nd);
	fdp->fd_cdir = save_cdir;

	if (error)
		return SETS_EXP_BADLOOKUP;
	
	if (nd.ni_vp == NULL)
		return SETS_EXP_NOMATCH;
	
	*vpp = nd.ni_vp;	/* already VREF'ed by namei */

	return error;
}

/*
 * expand_wildcard
 *  in
 *   vp:      FSOBJ for the directory where matching will occur
 *   pattern: beginning of the component with the wildcard
 *  out
 *   matches: list of strings that match the expression
 *
 *  notes
 *   - pattern will be past the '/'
 *   - the end of the component is marked by '\0'
 *   - matched strings include no wildcards or '/'
 *   - REF the vp while using, then RELE
 */
int expand_wildcard(FSOBJ_t dir, char *pattern, 
		    workerMsgType *msg, int myindex)
{
	char *word;		/* Name of currently examined dir entry */
	char *wptr;		/* Current location in processing word */
	char *pptr;		/* Current location in the pattern string */
	char *wild_wptr;	/* Last char in word to match wildcard */
	char *wild_pptr;	/* Char after wildcard in pattern string */
	char doWildcard;	/* True if currently matching to a wildcard. */
	char *suffix;
	enum { notdone, matched, notmatched } result;
	struct dirent *dp;
	char *endptr=NULL, *buf;
	int error = 0;
	int diroffset;

	if (!dir) {
		if (pattern[0] == '/') {
			dir = msg->set->rootdir;
			pattern++;
			msg->msg.expand.pathname[0] = '/';
			msg->msg.expand.pathname[1] = '\0';
		} else
			dir = msg->set->curdir;
		FSOBJ_REF(dir);
	}

	if (!IS_DIR_OBJ(dir))
		return SETS_EXP_NOTDIR;

	diroffset = 0;		/* This is updated by get_dir_page */
  
	/* isolate pattern, point at suffix */
	for (suffix = pattern; *suffix && *suffix != '/'; suffix++)
		;
	if (*suffix)
		*suffix++ = '\0';

	MYALLOC(buf, char *, DEV_BSIZE);
	bzero(buf, DEV_BSIZE);
  
	DEBUG(PATHNAME, printf("(%d) lookup on dvp %p\n", myindex, dir););

	/* Endless loop. A return is called if the READDIR returns 0 bytes. */
	for (dp = NULL; ; dp = (struct dirent *)((caddr_t)dp + dp->d_reclen)) {
		if ((dp == NULL) || ((caddr_t)dp >= endptr)) {
			/* Read a page into buf, point dp at start of buf,
			 * endptr at end */
			error = sets_get_dir_page(dir, buf, &diroffset,
						  msg->set->owner);
			if (error <= 0) { /* Error condition or end-of-dir */
				DEBUG(PATHNAME, 
				      printf("get_dir_page returns %d\n", 
					     error););
				break;
			}
      
			dp = (struct dirent *)buf;

			/* Error holds the number of bytes read. */
			endptr = &buf[error]; 
		}
    
		DEBUG(PATHNAME, printf("dp %p buf %p endptr %p reclen %x name %s\n",
				       dp, buf, endptr, dp->d_reclen,
				       dp->d_name););
    
		if (dp->d_reclen <= 0 || 
		    dp->d_reclen > (u_short)(endptr - (char *)dp)) {
			DEBUG(PATHNAME, printf("Acck, wierd looking directory entry\n"); );
			error = SETS_EXP_NOTDIR;
			break;
		}
    
#if 0				/* does anybody use this? */
		if (dp->d_ino == 0) 
			continue; /* Skip to next entry */
#endif
    
		wptr = &dp->d_name[0]; /* In Mach it's a char *, 
					* in linux a char[n] */
		word = &dp->d_name[0];
    
		DEBUG(PATHNAME, printf("(%d): glob %s word %s\n",
				       myindex, pattern, word););
    
		/* For reasons of unix compatibility, a leading '*'
		 * won't match a "." */
		if (*pattern == '*' && *word == '.')
			continue;
    
		result = notdone;		
		doWildcard = 0;
		wild_wptr = wild_pptr = 0;
    
		/* Go until we've determine if the word matches the pattern */
		for (pptr = pattern; result == notdone; pptr++) {
			switch (*pptr) {
			case '?' : { /* Match on any single character */
				if (*wptr) {
					wptr++;	/* Matched this one,
						 * skip to next */
	  
					/* "*?" == "?*", as an optimization to
					 * avoid backtracking skip this char
					 * but keep wildcard on */
					if (doWildcard) {
						/* Treat wptr as consumed */
						wild_wptr = wptr;
						/* Treat the "?" as the "*" */
						wild_pptr = pptr;
					}
				} else
					/* Word ended prematurely */
					result = notmatched;
				break;
			}
	
			case '[' : { /* list of chars to match with 1 
				      * char from w */
				char *p, *dash, match = 0;
	
				if (*wptr == 0) {
					/* Word ended prematurely */
					result = notmatched;
				} else {
	  
					for (p = pptr + 1;
					     *p && *p != ']';
					     p++) {
						dash = p + 1;
	    
						if (*dash == '-') { 
							/* Next char is a '-': 
							 * a sequence of
							 * matches */
							if (!(*wptr < *p ||
							      *wptr > *(++dash))) {
								match = 1;
								break;
							}
							p++; /* Skip the dash */
						} else if (*p == *wptr) {
							match = 1;
							break;
						}
					}
	  
					if (match) {
						/* Skip ahead to the end of
						 * the selection */
						while (*pptr && *pptr != ']')
							pptr++;
						wptr++;	/* Matched this one,
							 * skip to next */
						/* No longer processing a
						 * '*' */
						doWildcard = 0;
					} else 
						result = notmatched;
				}
				break;
			}
	
			case '*' : {
				doWildcard = 1;
				wild_pptr = pptr;
				wild_wptr = wptr;
				break;
			}
	
			case 0 : { /* Exhausted the pattern string */
				if (*wptr == 0 || doWildcard) {
					result = matched;
				} else {
					result = notmatched;
				}
				break;
			}
	
			default : {
				if (*pptr == *wptr) {
					wptr++;	/* Definate match */
					/* No longer processing a '*' */
					doWildcard = 0;
				} else
					result = notmatched;
				break;
			}
			}
      
			/* If a mismatch, yet word wasn't fully
			 * processed, apply '*' */
			if (result == notmatched && *wptr) {
				if (doWildcard) {
					/* '*' ate another char */
					wild_wptr = ++wptr;
					pptr = wild_pptr;
					result = notdone;
				}	
	
				/* Fall back on last wildcard if can */
				else if (wild_wptr) {	
					/* Eat up one more char via '*' */
					wptr = ++wild_wptr;
					/* Restart pattern after '*' */
					pptr = wild_pptr;
					/* Start using '*' again */
					doWildcard = 1;
					result = notdone;
				}
			}
		}
    
		/* it is a true match, build a new message for it. */
		/* It's a match if the pattern is satisfied and either
		 * wildcard was in effect (which would consume the remainder
		 * of the word) or the word was completely processed. */
		if (result == matched) {
			DEBUG(PATHNAME, printf("(%d): glob matched %s\n",
					       myindex, word););
			if (*suffix != '\0') {
				workerMsgType *newmsg;
				
				MAKE_NEW_MSG_FROM_OLD(newmsg, msg);
				newmsg->operation = WM_EXPAND;
				STRALLOC(newmsg->msg.expand.pathname);
				strcpy(newmsg->msg.expand.pathname,
				       msg->msg.expand.pathname);
				STRALLOC(newmsg->msg.expand.query);
				strcpy(newmsg->msg.expand.query, word);
				strcat(newmsg->msg.expand.query, "/");
				strcat(newmsg->msg.expand.query, suffix);
				newmsg->msg.expand.vp = dir;
				FSOBJ_REF(dir);
				msg->set->expands++;
				enqueueMessage(newmsg, myindex);
			} else {
				/* don't expand - just insert it! */
				char *newmember;
				
				STRALLOC(newmember);
				strcpy(newmember, msg->msg.expand.pathname);
				strcat(newmember, word);
				insertNewMember(msg->set, newmember, myindex);
			}
		}
	}
  
	/* Clean up. */
	FSOBJ_RELE(dir);
	MYFREE(buf, DEV_BSIZE);

	return error;
}

void close_elem(elemType *elem, struct proc *p)
{
    printf("close_elem fp_ref %d vp_ref %d\n", elem->fp->f_count,
	   elem->fsp->v_usecount);
    if (elem->fp != NULL)
	closef(elem->fp, p);
    printf("close_elem vp_ref %d\n", elem->fsp->v_usecount);
    if (elem->fsp != 0 && elem->fsp != (FSOBJ_t)-1)
        vrele(elem->fsp);
}
