#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */


#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_domain.h"

static int expand_query(workerMsgType *, FSOBJ_t, char *, char *, int);
static int expand_brace(char *, workerMsgType *, int);

/*
 * expand_pathname
 *  important arguments
 *   msg->msg.expand.pathname: portion of pathname already expanded
 *   msg->msg.expand.vp:       FSOBJ corresponding to pathname
 *   msg->msg.expand.query:    portion of pathname to be expanded,
 *                             added as a suffix to pathname
 *
 *  notes
 *   - vp and pathname are either both valid or both NULL.  if valid,
 *     they refer to the same object
 *   - if vp is non-NULL and query starts with a '/', forget vp
 *   - if pathname is non-NULL, it ends in a '/'
 *   - vp has been VREF'ed once for this message (if non-NULL)
 *     purgeMessage will FSOBJ_RELE(vp) if non-NULL -- DON'T DO IT HERE
 *   - query is non-NULL
 */
int expand_pathname(workerMsgType *msg, int myindex)
{
	FSOBJ_t vp;
#ifndef WIN32
	FSOBJ_t next_vp;
#endif
	char *curptr, *prefix, *query;
	char *component;
	int error = 0;

	DEBUG(PATHNAME, S_PRINTF("(%d): Expanding %s in %s (%p)\n",
			       myindex, msg->msg.expand.query,
			       msg->msg.expand.pathname,
			       msg->msg.expand.vp););
	
	vp = msg->msg.expand.vp;
	prefix = msg->msg.expand.pathname;
	query = msg->msg.expand.query;
	
	/*** Sanity checks ***/

	if (query[0] == '\0')	/* empty query? */
		return SETS_EXP_BADSPEC;

	/* vp better be paired with a prefix
	 * -- when might this not happen? */
	if (vp) {
	    if (prefix[0] == '\0')
		S_PRINTF("sets BOGON? vp == %p but prefix is NULL\n", vp);
	    if (!IS_DIR_OBJ(vp))
		return SETS_EXP_NOTDIR;
	}

	/*** Start munching ***/
	
	curptr = component = query;

	for (;;) {
		switch(*curptr) {

		case '\0':
			/* Nothing interesting found --
			 * insert member for cat(prefix,query) */
			strcat(prefix, query);
			insertNewMember(msg->set, prefix, myindex);
			msg->msg.expand.pathname = 0; /* prevent FREE */
			return 0;

		case '{':
			/* generate new EXPANDs for each item in the list */
			return expand_brace(curptr, msg, myindex);

#ifndef WIN32
		case QUERY_MARKER:
			/* split out the query and suffix */
			if (curptr != query && component+1 != curptr) {
				S_PRINTF("sets query doesn't start component\n");
				return SETS_EXP_BADSPEC;
			}

			if (component != query) {
				*component = '\0'; /* lose the trailing '/' */
			}
			component = ++curptr;
			for (; *curptr && *curptr != QUERY_MARKER; curptr++)
				;
			if (!*curptr)
				return SETS_EXP_BADSPEC;
			*curptr++ = '\0';

			error = expand_part(vp, query, &next_vp, msg);
			if (error)
				return error;

			strcat(prefix, query);
			strcat(prefix, COMPONENT_STRING);
			return expand_query(msg, next_vp, component,
					    curptr, myindex);
#endif
		case '*':
		case '?':
#ifndef WIN32
		case '[':  /* not handled by win32 at the moment */
#endif
			/* match wildcard expressions */

			/* first look for braces within */
			for (; *curptr && *curptr != COMPONENT_MARKER; curptr++) {
				if (*curptr == '{')
					return expand_brace(curptr, msg,
							    myindex);
			}

			/* Note: Win32 doesn't do any expand_part's since I
			   can't do relative lookups on a per-thread basis (?) */
#ifndef WIN32
			if (component != query) {
				/* need to update vp */
				*component = '\0';
				error = expand_part(vp, query, &next_vp, msg);
				if (error)
					return error;
				strcat(prefix, query);
				strcat(prefix, COMPONENT_STRING);
				*component = COMPONENT_MARKER;
				FSOBJ_UNLOCK(next_vp);
				return expand_wildcard(next_vp, component+1,
						       msg, myindex);
			} else
				return expand_wildcard(vp, query, msg,
						       myindex);
#else
			return expand_wildcard(query, component, msg, myindex);
#endif

		case COMPONENT_MARKER:
			component = curptr++;
			break;

		default:
			curptr++;

		}
	}

	return error;
}

/*
 * expand_brace
 *  in
 *   start:   the beginning of the brace pattern
 *
 *  notes
 *   - start points to a '{'
 *   - each yielded string may contain '/', or wildcards
 */
static int expand_brace(char *start, workerMsgType *msg, int myindex)
{
	register char *c;
	char *elem, *end, c2;
	register unsigned int level;
	int error = 0, done;

	LOG_ENTRY(myindex, DOSQUIGGLY);

	/* 1. Ensure the glob is really a list. */
	if (*start != '{')
//		panic("sets: expand_brace called with bad start\n");
		exit(-1);

	/* 2. Find the end of the start (avoid nested starts) */
	level = 1;		
	end = start;
	do {
		end++;		/* Skip first '{' */
		if (*end == '{')
			level++;
		else if (*end == '}')
			level--;
	} while (level && *end);
  
	if (*end != '}') {	/* Bad deal, missing '}' */
		S_PRINTF("AcccK! missing } at line %d\n", __LINE__);
		error = SETS_EXP_BADSPEC;
		goto out;
	}
  
	/* Isolate the glob, end points to rest of name */
	*(end++) = 0;
  
	/* 3. Isolate the prefix from the glob */
	*start = 0;
  
	/* 4. look for commas, creating new msgs for each,
	 * let each case incr c */
	for (done = 0, c = elem = start + 1; !done; ) {
		switch (*c) {
		case '{' : {	/* Skip this nested list
				 * (skip to matching '}') */
			for (level = 1; level && *c;) {
				c++; /* Here to keep incrementing beyond 
				      * closing '}' */
				if (*c == '{')
					level++;
				else if (*c == '}')
					level--;
			}	
      
			if (*c != '}') { /* Bad deal, missing '}' */
				S_PRINTF("AcccK! missing } at line %d\n",
				       __LINE__);
				error = SETS_EXP_BADSPEC;
				goto out;
			} else
				c++; /* Skip over closing '}' */
			break;
		}
      
		case ',' :
		case '\0' : {
			workerMsgType *newmsg;
			
			/* We found the end of a member, create an
			 * expand msg for it. */
			c2 = *c;
			*c = 0;
			
			MAKE_NEW_MSG_FROM_OLD(newmsg, msg);
			newmsg->operation = WM_EXPAND;
			STRALLOC(newmsg->msg.expand.pathname);
			strcpy(newmsg->msg.expand.pathname,
			       msg->msg.expand.pathname);
			STRALLOC(newmsg->msg.expand.query);
			strcpy(newmsg->msg.expand.query,
			       msg->msg.expand.query);
			strcat(newmsg->msg.expand.query, elem);
			strcat(newmsg->msg.expand.query, end);
			newmsg->msg.expand.vp = msg->msg.expand.vp;
			if (newmsg->msg.expand.vp)
				FSOBJ_REF(newmsg->msg.expand.vp);
			CRITICAL(msg->set->lock, msg->set->expands++);
			enqueueMessage(newmsg, myindex);
      
			/* reset for next element. */
			if (c2 == ',') {
				*c = ',';
				elem = ++c;
			} else
				done = 1;
      
			break;
		}
      
		default : {
			c++;	/* Skip over this character.
				 * It'll get copied later */
		}
			}
	}

 out:
	LOG_EXIT(myindex, DOSQUIGGLY);
	return error;
}

/* assumptions:
 *   dvp is the dir in which the cursor (query) lives
 *  notes
 *   -a single ref+lock follows the dvp from the calling expansion into the 
 *    query message and 1 follows cursvp
 */
static int expand_query(workerMsgType *msg, FSOBJ_t dvp,
			char *query, char *suffix, int myindex)
{
#ifndef WIN32
	FSOBJ_t   cvp;
	CURSOBJ_t curs;
	struct sets_domain *dom;
	workerMsgType *newmsg;
	int error;

	/* find the parent's domain */
	dom = sets_find_domain(dvp);
	if (!dom) {
		FSOBJ_UNLOCK(dvp);
		return SETS_BADDOMAIN;
	}
	/* use the expand function to get a cursor vp */
	error = (dom->expand_fn)(dvp, &cvp, query,
				 msg->set->owner);
	if (error)
		return SETS_EXP_BADEXPAND;
	if (!cvp)
		return SETS_EXP_NOMATCH;

	error = OPEN_CURSOBJ(cvp,
			     InterpretPages(msg->set) ? SETS_WRITE : SETS_READ,
			     msg->set->owner,
			     &curs);
	FSOBJ_UNLOCK(cvp);
	if (error)
	    return SETS_EXP_BADCURSOR;

	/* Instead of reusing the message, I make a new one.  Although
	 * less efficient, it simplifies the code because now the daemon
	 * can ALWAYS free "msg" after the call the expand_pathname. */
	MAKE_NEW_MSG_FROM_OLD(newmsg, msg);
	newmsg->operation = WM_QUERY;
	newmsg->msg.query.handle = curs;
	newmsg->msg.query.pathname = msg->msg.expand.pathname;
	newmsg->msg.query.vp = dvp; /* carry reference from namei */
	STRALLOC(newmsg->msg.query.suffix);
	strcpy(newmsg->msg.query.suffix, suffix);
	CRITICAL(msg->set->lock, msg->set->expands++);
	enqueueMessage(newmsg, myindex);
	
	msg->msg.expand.pathname = 0; /* reusing this string for newmsg */
#endif
	return 0;
}
