#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"

/*
 * These globals limit prefetching's usage of the buffer cache 
 * in terms of bytes.
 */
int sets_max = 0x100000;	/* Maximum allowed usage of the buffer cache */
int sets_count = 0;		/* Current usage of buffer cache */
int sets_restart = 0xf0000;	/* Restart stalled prefetches when usage 
				 * drops past this point */
int pin_count = 0;		/* Current amount of data pinned. */
int pin_max = 0xc0000;		/* Max amount of data to be pinned. */

int aggressive = 0;
int nprefix = 2;

int numWorkers = 0;

int do_set_daemon()
{
	workerMsgType *message;
	int j;
#ifndef WIN32
	int free_msg;
#endif
	int myindex;

	myindex = ++numWorkers;
	DEBUG(UTIL, S_PRINTF("In setDaemon [%d]\n", myindex););

	while (1) {
		struct setType *myset;
		int error, i;
		
		message = dequeueMessage(myindex);

		if (message->operation == WM_EXIT) {
			/* daemon should return */
			purgeMessage(message, myindex);
			numWorkers--;
			S_PRINTF("SetDaemon %i exitting\n", myindex);
			return 0;
		}

		myset = message->set;

		if (message->unique != myset->unique) {
			/* Set has died, and may be reallocated */
			DEBUG(UTIL, 
				S_PRINTF("(%d) Got a dead message (op == %d)\n",
					myindex, message->operation););
			purgeMessage(message, myindex);
			continue;
		}
		
		switch (message->operation) {
		case WM_EXPAND:
			error = expand_pathname(message, myindex);
			if (error > 0) {
				i = find_set_elem_slot(myset);
				
					DEBUG(UTIL, S_PRINTF("(%d): Error %d at "
							   "resolved[%d]\n", 
							   myindex, error, i););
				
				CRITICAL(myset->lock, {
					/* Don't try to open this! */
					myset->resolved[i].state = ELEM_FAILED;
					myset->resolved[i].fsp = (FSOBJ_t)-1;
					myset->resolved[i].error = error;
					
					/* We finished an expand */
					if (--myset->expands == 0) { 
						/* Finished all expands */
						myset->flags |= DONEEXPANDING;
					}
				});
				/* What should we report as the name?
				 * Issue here is a *set* failure, vs a
				 * *member* failure...
				 * Used to be a CONDITION_SIGNAL here, not 
				 * sure it's needed. */
			} else {
				CRITICAL(myset->lock, {
					/* We finished an expand */
					if (--myset->expands == 0) { 
						/* We finished all expands */
						myset->flags |= DONEEXPANDING;
					}
				});
			}
			
			/* Not sure it's safe to do this in the 
			 * critical section. */
			if (myset->expands == 0)
				CONDITION_SIGNAL(&(myset->cond));

			purgeMessage(message, myindex);
			break;
		case WM_QUERY:
#ifndef WIN32
			error = expand_cursor(message, &free_msg, myindex);
			if (error > 0) {
				i = find_set_elem_slot(myset);
				
				DEBUG(UTIL, S_PRINTF("(%d): Error %d at "
						   "resolved[%d]\n", 
						   myindex, error, i););
				myset->resolved[i].state = ELEM_FAILED;
				myset->resolved[i].fsp = (FSOBJ_t)-1;
				myset->resolved[i].error = error;
				/* What should we report as the name? Issue 
				 * here is a *set* failure, vs a *member*
				 * failure... */
				CONDITION_SIGNAL(&(myset->cond));
			}
			
			CRITICAL(myset->lock, {
				/* We finished an expand */
				if (--myset->expands == 0) {
					/* We finished all expands */
					myset->flags |= DONEEXPANDING;
				}
			});
			
			/* Not sure it's safe to do this in the
			 * critical section. */
			if (myset->expands == 0)
				CONDITION_SIGNAL(&(myset->cond));

			if (free_msg) /* message was not reused */
				purgeMessage(message, myindex);
#endif
			break;
		case WM_OPEN:
			i = message->msg.open.index;
			if (i > myset->resSize) {
				S_PRINTF("Help, bogus index %x, > %x\n",
				       i, myset->resSize);
				purgeMessage(message, myindex);
				break;
			}

			if (myset->resolved[i].state != ELEM_OPENED) {
				S_PRINTF("Help, bogus index %x, "
				       "state %d not OPENED\n", i, 
				       myset->resolved[i].state);
				purgeMessage(message, myindex);
				break;
			}
			
			/* Obtain the file object */
			error = open_object(myset, &myset->resolved[i],
					    myindex);
      			
			if (error) {
				CRITICAL(myset->lock, {
					/* Don't try to yield this! */
					myset->resolved[i].state = ELEM_FAILED;
					myset->resolved[i].fsp = (FSOBJ_t)-1;
					myset->resolved[i].error = error;
				})
			} else {
				DEBUG(UTIL, 
				      S_PRINTF("(%d): Opened %s\n", myindex,
					     myset->resolved[i].name););
				/* Try to pre-read the file's prefix into 
				 * the cache  (if there is pin space */
#ifdef PRE_READ
				if (0 && PREFETCHING && pin_count < pin_max )
					preReadPrefix(myset, i, nprefix,
						      myindex);
				else
					CRITICAL(myset->lock, {
						myset->resolved[i].toyield = 
							myset->toyield;
						myset->toyield = i;
					});
#endif
				}

			DEBUG(UTIL, S_PRINTF("done with open, myset->opens %d\n",
					   myset->opens););

			/* Put this in a critical section to avoid a race
			 * where we wakeup before setiterate goes to sleep. */
			CRITICAL(myset->lock, 
				 CONDITION_SIGNAL(&(myset->cond)););

			/* Because there was an error, start another prefetch
			 * without bumping opens. This way we ensure that
			 * there are limitOpen good/valid files to be had */
			if (PREFETCHING && (aggressive || error)) {
				/* Queue another open message to take this 
				 * one's spot. */
				for (i++, j = 1; j < myset->resSize; 
				     j++ , i = (i + 1) % myset->resSize) {
					if (myset->resolved[i].state ==
					    ELEM_TAKEN) {
						if (myset->resolved[i].name == 0)
//WIN32							panic("Prefetching a member %d with no name\n", i);
							exit(-1);
						CRITICAL(myset->lock, myset->resolved[i].state = ELEM_OPENED);
						DEBUG(ITERATE, S_PRINTF("Worker spooling open for %s(%d)\n", \
								      myset->resolved[i].name, myset->opens););

						/* Reuse the message. */
						message->msg.open.index = i;
						enqueueMessage(message,
							       myindex);
						break;
					}
				}
				
				if (j >= myset->resSize) 
					/* Huh, no opens to do. */
					purgeMessage(message, myindex);
			} else {
				/* Clean up */
				purgeMessage(message, myindex);
			}
			break;
		case WM_UNION:
			/* reuses or frees message */
			do_union(message, myindex);
			break;
		case WM_INTERSECT:
			/* reuses or frees message */
			do_intersect(message, myindex);
			break;
		default:
//WIN32			panic("sets: wacky message\n");
				exit(-1);
		}
	}
}

void kill_daemons(void)
{
	int i, workers;
	workerMsgType *msg;

	workers = numWorkers;	/* numWorkers will decrement as they die */
	for (i=0; i<workers; i++) {
		MYALLOC(msg, workerMsgType *, sizeof(workerMsgType));
		msg->operation = WM_EXIT;
		msg->set = NULL;
		enqueueMessage(msg, -1);
	}
}
