#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* Produce a digest of an openset. Fills a buffer (given as a param), returns
 * number of bytes put into buffer.
 */

#include "kernel/sets/common/sets_internal.h"

int do_set_digest(struct procSetInfo *settbl, SETID handle,
		  char *buf, unsigned long size, int *retsz)
{
	struct setType *myset;
	unsigned short *myOffset;
	unsigned long nbytes;
	char *lastEntry;
	int result;		/* set by one of the final clauses */
    
	LOG_ENTRY(0, SETDIGEST);
    
	DEBUG(DIGEST, S_PRINTF("Entering setDigest(%d, %p, %ld)\n", handle,
			     buf, size););
    
	if (handle < 0 || handle >= NOPENSETS) {
		result = -EBADF;
		goto out;
	}

	if (buf == 0) {
		result = -EINVAL;
		goto out;
	}
    
	if ((myset = settbl->mySets[handle]) == NULL) {
		result = -EBADF;
		goto out;
	}

	DEBUG(DIGEST, setPrint(myset););

	/* iterator == -1 means the set hasn't yet begun to be resolved */
	if (myset->iterator == (short)-1) {
		workerMsgType *msg;

		MAKE_NEW_MSG(msg, WM_EXPAND, myset);

		STRALLOC(msg->msg.expand.pathname);
		*msg->msg.expand.pathname = 0; /* zero length string */

		msg->msg.expand.vp = (FSOBJ_t)NULL; /* no explicit starting vp */

		STRALLOC(msg->msg.expand.query);
		strcpy(msg->msg.expand.query, myset->spec);
	
		/* Initialize the iterator */
		CRITICAL(myset->lock, {
			settbl->mySets[handle]->iterator++;
			/* Indicate an expansion needs to be done. */
			myset->expands++;
		});

		enqueueMessage(msg, -1); /* And yield */
	}

	/* Read through the array, starting from myOffset. For each entry, copy
	 * the name of that member into the buffer (adding room for the length)
	 * until we run out of room in the buffer. If the set is not fully
	 * expanded, but there are no more members available now, add the
	 * special name "..." which means more to come. Since names are likely
	 * to be inserted consecutively, this is reasonable. (not true for
	 * iterate) */
	/* On another note, can we achieve starvation by repeatedly calling 
	 * setDigest? Perhaps we should force a yield before returning
	 * just "..."? */

	nbytes = size;
	myOffset = &settbl->mySets[handle]->digestOffset;
	lastEntry = NULL;
    
	/* Also exit if we run out of buffer space */
	while (*myOffset < myset->setSize) {
		unsigned long len, struct_len;

		if (myset->resolved[*myOffset].state != ELEM_FREE) {
			/* We have resolved an element, but there
			 * could be an error. */
			if (myset->resolved[*myOffset].error) {
				DEBUG(DIGEST,
				      S_PRINTF("Found error %d at slot %d\n", 
					     myset->resolved[*myOffset].error,
					     *myOffset););
				if (myset->resolved[*myOffset].name == NULL) {
					/* Go on to next member */
					CRITICAL(myset->lock,
					 {(*myOffset)++;});
					continue;
				}
			}

			/* account for '\0' */
			len = strlen(myset->resolved[*myOffset].name) + 1;
			/* Round to nearest longword, account for
			   pointer, flag byte. */
			struct_len = ((len + 4) &~ 3) + 4;
			
			/* We have room to add this structure, and
			 * potentially the '...' entry) */
			if (nbytes >= struct_len + 12) {
				char flags;

			        /* Hopefully this will panic if name
                                   is bogus */
			        DEBUG(DIGEST, S_PRINTF("digesting %s\n", myset->resolved[*myOffset].name););

				/* Have previous entry point to this one */
				if (lastEntry) 
					PUT_USER_LONG((long)buf, lastEntry);
				lastEntry = buf;
				PUT_USER_LONG(0, buf); /* assume end of list */
				/* should I check "state" too? */
				flags = (myset->resolved[*myOffset].fsp != 0 &&
					 myset->resolved[*myOffset].fsp != (FSOBJ_t)-1);
				PUT_USER_BYTE(flags, buf+4);
				COPYTOUSERSPACE(myset->resolved[*myOffset].name, buf+5, len);
				buf += struct_len;
				/* Buf should now point to start of
				 * next slot */
				nbytes -= struct_len;
		
				/* Go on to next member */
				CRITICAL(myset->lock, {(*myOffset)++;});
			} else
				break;
		} else {
//			panic("digest < size, but elements are free! Asserting...\n");
			exit(-1);
		}
	}

	/* At this point, either we ran out of members or buffer space. */
	DEBUG(DIGEST, S_PRINTF("doneexpanding %d, myoffset %d, size %d\n",
			     DoneExpanding(myset), *myOffset,
			     myset->setSize););

	/* IMPORTANT: we must have saved space for the "..." entry. */
	if (!DoneExpanding(myset) || (*myOffset < myset->setSize)) {
		/* We've shown all there is to show right now, but
                   there is some left. */
		if (lastEntry)	/* Have previous entry point to this one */
			PUT_USER_LONG((long)buf, lastEntry);   
		PUT_USER_LONG(0, buf); /* No next element. */
		PUT_USER_BYTE(0, buf+4);
		buf += 5;
		PUT_USER_LONG(0x002e2e2e, buf); /* Numerical "..." */
		nbytes -= 12;
	}

	*retsz = size - nbytes;
	result = 0;

 out:
	LOG_EXIT(0, SETDIGEST);
	return result;
}


int do_set_rewind_dig(struct procSetInfo *settbl, int handle)
{
	int result = 0;

	LOG_ENTRY(0, REWINDDIGEST);
    
	if (handle < 0 || handle >= NOPENSETS) {
		result = -EBADF;
		goto out;
	}

	if (settbl->mySets[handle] == NULL) {
		result = -EBADF;
		goto out;
	}

	CRITICAL(settbl->mySets[handle]->lock,
		 settbl->mySets[handle]->digestOffset = 0;);

	DEBUG(DIGEST, S_PRINTF("rewound digest for %p\n", 
			     settbl->mySets[handle]););

 out:
	LOG_EXIT(0, REWINDDIGEST);
	return result;
}	
