#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_hash.h"

int do_set_intersect(struct procSetInfo *settbl, SETID s1, SETID s2,
		     int flags, SETID *handle)
{
	struct setType *set1, *set2, *myset;
	workerMsgType *msg;
	int result = 0;
    
	LOG_ENTRY(0, SETINTERSECT);
    
	DEBUG(INTERSECT, S_PRINTF("Entering setIntersect(%d, %d, %d)\n",
				s1, s2, flags););

	if (settbl->nOpen >= NOPENSETS) {
		result = -EMFILE;
		goto out;
	}

	/* Make sure both s1 and s2 are valid */
	if ((s1 < 0 || s1 >= NOPENSETS) ||
	    (s2 < 0 || s2 >= NOPENSETS) ||
	    (settbl->mySets[s1] == NULL) ||
	    (settbl->mySets[s2] == NULL))
		{
			result = -EBADF;
			goto out;
		}

	/* Create a set to hold the intersection. */
	/* createSet will return a negative value on error,
	 * or the open file handle */
	result = createSet(0, settbl);
    
	if (result < 0) {
		goto out;
	}

	myset = settbl->mySets[result];
	set1 = settbl->mySets[s1];
	set2 = settbl->mySets[s2];
	
	myset->s1 = set1;
	myset->s2 = set2;
    
	/* Don't want iterate or digest to spool an expand msg for 
	 * this type of set.*/
	CRITICAL(myset->lock, {
		(settbl->mySets[result]->iterator)++;
		myset->expands = 1; /* we are going to queue 1 union msg. */
	} );

	/* Create a message to start the processing of the set. */
	MAKE_NEW_MSG(msg, WM_INTERSECT, myset);
	
	MYALLOC(msg->msg.merge.bm1, char *, set1->resSize);
	memset(msg->msg.merge.bm1, 0, set1->resSize);
	msg->msg.merge.bm1Size = set1->resSize;
	
	enqueueMessage(msg, -1); /* And yield */
	
	DEBUG(INTERSECT, setPrint(settbl->mySets[result]););

	*handle = result;
	result = 0;

out:
	LOG_EXIT(0, SETINTERSECT);
	return result;
}

void do_intersect(workerMsgType *message, int myindex)
{
	int i, j;
	struct setType *myset = message->set;
	
	LOG_ENTRY(myindex, DOINTERSECT);
	
	DEBUG(INTERSECT, S_PRINTF("Entering doIntersect on %p (set %p)\n",
				message, myset); );
    
	/* If the resolved arrays in the sets have grown,
	 * grow the bitmap too. */
	if (message->msg.merge.bm1Size < myset->s1->resSize) {
		char *old = message->msg.merge.bm1;
		MYALLOC(message->msg.merge.bm1, char *, myset->s1->resSize);
		memcpy(message->msg.merge.bm1, old,
		       message->msg.merge.bm1Size);
		MYFREE(old, message->msg.merge.bm1Size);
		message->msg.merge.bm1Size = myset->s1->resSize;
	}
	
	for (i = 0; i < myset->s1->resSize; i++) {
		if ((message->msg.merge.bm1[i] == 0) &&
		    myset->s1->resolved[i].name) {
			/* Try to find a matching member in s2 */
			for (j = 0; j < myset->s2->resSize; j++) {
				if (myset->s2->resolved[j].name)
					if (strcmp(myset->s1->resolved[i].name,
						   myset->s2->resolved[j].name) == 0) {
						char *name;
						
						/* found a new member */
						message->msg.merge.bm1[i] = 1;
						
						/* Need a new string to add to
						 * the new element. */
						STRALLOC(name);
						strcpy(name, myset->s1->resolved[i].name);
						
						insertNewMember(myset, name,
								myindex);
						
						/* What the heck, requeue the
						 * intersect message. */
						enqueueMessage(message,
							       myindex);
						
						/* Notify waiting users that
						 * another entry has been
						 * added. */
						CONDITION_SIGNAL(&(myset->cond));
						
						goto out;
					}
			}
		}
	}
	
	/* Maybe there's nothing more to get? */
	if (DoneExpanding(myset->s1) && DoneExpanding(myset->s2)) {
		/* Indicate that this set is fully expanded. */
		CRITICAL(myset->lock, {
			if (--myset->expands == 0) {
				/* We finished all expands */
				myset->flags |= DONEEXPANDING;
			}	
		});
		CONDITION_SIGNAL(&(myset->cond));
		
		/* Free up our message */
		purgeMessage(message, myindex);
		goto out;
	}
	
	/* What do we do? Condition_wait won't work because either set may be
	 * modified, and we can only wait on 1 event. Simple solution: requeue
	 * the message. Perhaps sleep more?
	 */
	enqueueMessage(message, myindex);
out:
	LOG_EXIT(myindex, DOINTERSECT);
	return;
}
