#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* Iterate on an openset. The return code is the equivalent of an open file
 * descriptor.
 */

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_hash.h"

/* Basically, we need to decide how to get multiple threads. for now
 * try the same thing used in the prototype -- multiple worker threads that
 * process messages.
 *
 *  alternative: try doing the expansion here if it is easy, or perhaps in
 *  setopen. Just do it if it is basic -- {a,b,c}
 *  Nahh, setiterate is almost always called right after setopen...
 *	      
 *  perhaps the easiest thing to do is to do all the work directly...
 *  rewrite doopen or expandpathname to return the "message" which is
 *  managed by sets.
 *
 *  could also have a background task which asks coda to fetch files, but
 *  doesn't actually open anything. the application does an open and hopefully
 *  stuff will still be in the name/buffer cache.
 *  the advantage is that the lower level can get rid of stuff if it isn't
 *  used. So the set won't be tying up resources to much.
 */

extern int limitOpens;
extern int aggressive;

int do_set_iterate(struct procSetInfo *settbl, SETID handle, u_int flags,
		   char *buf, u_int bufsize, u_int *error, FILEID *fd)
{
    struct setType *myset;
    short *myiterator;
    workerMsgType *workerMessage;
    int best = -1;
    register int i, j, minIncore = 0x7fffffff;
	FILEID fid;
    int result = -ENOENT;
    
    LOG_ENTRY(0, SETITERATE);
    
    DEBUG(ITERATE, S_PRINTF("Entering setIterate(%d, %d)\n", 
			  handle, flags););
    
    if (flags & SETS_GETNAME && ((buf == NULL) || bufsize == 0)) {
	result = -EINVAL;
	goto out;
    }
    
    if (flags & SETS_GETERR) {
	if (error == NULL) {	/* return an elem's error in error. */
	    result = -EINVAL;
	    goto out;
	} else
	    *error = 0;
    }
    
    if (handle < 0 || handle >= NOPENSETS) {
	result = -EBADF;
	goto out;
    }
    
    if ((myset = settbl->mySets[handle]) == NULL) {
	result = -EBADF;
	goto out;
    }
    
    myiterator = &(myset->iterator);

    DEBUG(ITERATE, setPrint(settbl->mySets[handle]););
    
    if (Exhausted(myset)) {
	goto out;		/* result already -ENOENT */
    }
    
    /* We want opens to happen, so force it to be so. */
    myset->flags &= ~DELAYOPENS;
    
    if (*myiterator == (short)-1) { /* Haven't resolved anything yet. */
	MAKE_NEW_MSG(workerMessage, WM_EXPAND, myset);
	
	STRALLOC(workerMessage->msg.expand.pathname);
	*workerMessage->msg.expand.pathname = 0; /* null string */
	
	/* no explicit starting vp */	
	workerMessage->msg.expand.vp = NULL;

	STRALLOC(workerMessage->msg.expand.query);
	strcpy(workerMessage->msg.expand.query, myset->spec);
	
	/* Initialize the iterator */
	CRITICAL(myset->lock, {
	    (*myiterator)++;
	    /* Indicate an expansion needs to be done. */
	    myset->expands++;
	});
	
	enqueueMessage(workerMessage, -1); /* And yield */

	DEBUG(ITERATE, S_PRINTF("setiterate waiting\n"););
	CONDITION_WAIT(&(myset->cond), NULL);
	DEBUG(ITERATE, S_PRINTF("setiterate awakened\n"););
    }
    
    
    /*
     * Determine the next (ready, unyielded, highest priority) member to
     * yield.
     * 	- Assume the array's order is in decreasing priority.
     *	- Scan array looking for the first ready member, yield it 
     *        when found.
     * If no ready member is found, wait for the next member to become
     * ready.  However, since there may be a lower priority member
     * ready, examine all members before waiting. Do the loop in a
     * critical section primarily to avoid a race condition where a
     * worker does a wakeup while we are in the loop. */
    
    /* Scan the resolved array looking for unyielded results. */
    CRITICAL(myset->lock, {
	while (myset->toyield == -1) {
	    /* Go and try to find something to yield. */
	    for (i = *myiterator; i < myset->setSize; i++) {
		DEBUG(ITERATE_LONG,
		      if (myset->resolved[i].state !=
			  ELEM_FREE) {
			  S_PRINTF("%d: %d fsp %p fp %p name %s error %d incore %d\n",
				 i,
				 myset->resolved[i].state,
				 myset->resolved[i].fsp,
				 myset->resolved[i].fp,
				 myset->resolved[i].name ?
				 myset->resolved[i].name :
				 "UNEXPANDED",
				 myset->resolved[i].error,
				 myset->resolved[i].nincore);
		      });	
		
		switch (myset->resolved[i].state) {
		case ELEM_FREE:	/* Ignore it */
		    S_PRINTF("Yikes, how did iterate find a free elem?\n");
		    break;
		    
		case ELEM_FAILED: {
		    if (flags & SETS_GETERR &&
			myset->resolved[i].name) {
			/* App wants to know about errors fetching a
			 * member.  So treat this as a valid member
			 * yield, but report the error.  Note:
			 * expansion errors won't get reported, as
			 * they have no member name associated with
			 * their entry.  */
			*error = myset->resolved[i].error;
			
			myset->resolved[i].toyield = myset->toyield;
			myset->toyield = i;
			/* Drop out of this loop! */
			i = myset->setSize;
			/* Iterator is updated below. */
		    } else {
			S_PRINTF("Error %d on opening (%s) (slot %d)\n", 
			       myset->resolved[i].error,
			       myset->resolved[i].name ? 
			       myset->resolved[i].name :
			       "UNEXPANDED", i);
			/* Update iterator if everything above it has
			 * been seen. */
			if (*myiterator == i) 
			    (*myiterator)++;
		    }
		    break;
		}
		
		case ELEM_SEEN: { 
		    /* Update iterator if everything above it has been
		     * seen. */
		    if (*myiterator == i) 
			(*myiterator)++;
		    break;
		}
		
		default: {
		    /* This is a legitimate error, we should know
		     * about it. */
		    S_PRINTF("Help!, state of slot %d is %d\n",
			   i, myset->resolved[i].state);
		    break;
		}
		
		case ELEM_TAKEN: {
		    if (!PREFETCHING ||
			myset->opens < limitOpens) {
			myset->resolved[i].state = ELEM_OPENED;
			myset->opens++;
			
			DEBUG(ITERATE, S_PRINTF("Iterate spooling open "
					      "for %s(%d)\n",
					      myset->resolved[i].name,
					      myset->opens););
			
			/* create a new message for
			 * the open */
			MAKE_NEW_MSG_NO_LOCK(workerMessage, WM_OPEN, myset);
			workerMessage->msg.open.index = i;
			enqueueMessage(workerMessage, -1); /* Yield */
			
			if (!PREFETCHING) {
			    i = myset->setSize;
			    /* Drop out of loop! */
			}
		    }
		    break;
		}
		
		case ELEM_OPENED : {
		    if (myset->resolved[i].fsp == (FSOBJ_t)-1) {
			S_PRINTF("Help, set %s has an error element %p "
			       "marked as opened\n.",
			       myset->spec, &myset->resolved[i]);
			// panic("Help, bogus OPENED set member.");
		    exit(-1);
			}
		    else if (myset->resolved[i].fsp) {
			/* There is a valid element. */
			if (PREFETCHING) {
			    if (myset->resolved[i].nincore == 0) {
				/* Go with this one, highest priority
				 * and raring to go! */
				myset->resolved[i].toyield = myset->toyield;
				myset->toyield = i;
				i = myset->setSize; /* Drop out of loop! */
			    }
			    else if (minIncore > myset->resolved[i].nincore) {
				/* It is the best one so far. (least
				 * out of core) */
				minIncore = myset->resolved[i].nincore;
				best = i;
			    }
			}
			else {
			    /* we found something to yield! */
			    myset->resolved[i].toyield = myset->toyield;
			    myset->toyield = i;
			    /* Drop out of loop! */
			    i = myset->setSize;
			}
		    }
		    break; /* for switch statement */
		}
		}
	    }
	    
	    /* If nothing was perfect, the "best" entry is better
	     * than "myset->toyield" */
	    if (PREFETCHING && (best != -1)) {	
		myset->resolved[best].toyield = myset->toyield;
		myset->toyield = best;
	    }
	    
	    if (DoneExpanding(myset) &&
		*myiterator == myset->setSize) {
		/* We've yielded everything, and no new things */
		DEBUG(ITERATE, S_PRINTF("Set must be empty\n"););
		myset->flags |= EXHAUSTED;
		/* Force us to drop out of loop and critical
		 * section. */
		break;
	    }
	    
	    if (myset->toyield == -1) {
		/* Pause for the open we just did, another open, or
		 * set expansion */
		LOG_ENTRY(0, SLEEP); 
		CONDITION_WAIT(&myset->cond, &myset->lock);
		LOG_EXIT(0, SLEEP); 
	    }
	}
	
	if (myset->toyield != -1) {
		/* Pop "best" from the start of the queue, make
	     * myset->toyield point to next best */
	    best = myset->toyield;
	    myset->toyield = myset->resolved[best].toyield;
	    
	    if (myset->resolved[best].error == 0) {
		/* Don't mark erroroneous members as seen. */
		/* Yield this element. */
		myset->resolved[best].state = ELEM_SEEN;
	    }
	    
	    /* Update iterator if everything above it has been
	     * seen. */
	    if (*myiterator == best)
		(*myiterator)++;
	    
	    if (PREFETCHING && !aggressive) {
		/* To keep things moving, queue another open if one
		 * remains to be done, otherwise try prereading
		 * something that needs it.  */
		/* KIP: might there be more expansion happening, but
		 * nothing to open yet?  we would want to
		 * myset->opens-- since we can't replace this prefetch
		 * at the moment.  check DoneExpanding(myset) */
		for (j = *myiterator; j < myset->setSize;
		     j++) {
			if (myset->resolved[j].state == ELEM_TAKEN) {
			myset->resolved[j].state = ELEM_OPENED;		
			DEBUG(ITERATE, S_PRINTF("Iterate doing open for %s\n",
					      myset->resolved[j].name););
			
			/* create a new msg from msg */
			MAKE_NEW_MSG_NO_LOCK(workerMessage, WM_OPEN, myset);
			workerMessage->msg.open.index = j;
			enqueueMessage(workerMessage, -1); /* Yield */
			break;
		    }
		    
		    else if (myset->resolved[j].state == ELEM_OPENED) {
				if (myset->resolved[j].fsp > (FSOBJ_t)0) {
			    if (0 && pin_count + myset->resolved[j].nincore <= pin_max) {
				/* start a preread. */
				break;
			    }
			}
			}
		}
	    }
	}	
    });			/* END OF CRITICAL SECTION */
    
    if (Exhausted(myset)) {
	goto out;
    }
    
    if (flags & SETS_GETNAME) { /* Means return the name too. */
	char tmp = 0;
	u_int len = strlen(myset->resolved[best].name) + 1;
	DEBUG(ITERATE,S_PRINTF("giving name %s\n", 
			     myset->resolved[best].name););
	if (len > bufsize) {
	    len = bufsize - 1;
	    tmp = myset->resolved[best].name[len];
	    myset->resolved[best].name[len] = 0;
	    DEBUG(ITERATE, S_PRINTF("clipped name %s\n",
				  myset->resolved[best].name););
	}
	COPYTOUSERSPACE(myset->resolved[best].name, buf, len);
	
	if (tmp != 0)
	    myset->resolved[best].name[len] = tmp;
    }
    
    if (myset->resolved[best].error && (flags & SETS_GETERR)) {
	/* Reporting an error. */
	DEBUG(ITERATE,S_PRINTF("Iterate saw error %d on %s\n",
			     myset->resolved[best].error,
			     myset->resolved[best].name););
	PUT_USER_INT(myset->resolved[best].error, error);
	result = -ENOENT;
	/* XXX Should this really indicate EOS?? */
	goto out;
    }
    
    /* make an open file descriptor for the vnode */
    fid = setup_fd(myset, handle, best);
    
    /* Trouble. Error means no more descriptors (ENFILE == in
     * process's table; EMFILE == in systems's table). Ideally we'd
     * like to keep looking for a valid member, but since this problem
     * won't go away in the short term, best bet is to inform the
     * application so it can close files to clear up the
     * problem. Don't leave this entry as SEEN, since it really hasn't
     * been. */
    if (FID_INVALID(fid)) {
	DEBUG(ITERATE, S_PRINTF("Iterate got error %x from alloc_filed\n", fid););
	CRITICAL(myset->lock, 
		 myset->resolved[best].state = ELEM_OPENED;);
	result = -ENFILE;
	goto out;
    }
    
    DEBUG(ITERATE, S_PRINTF("Iterate yielding %s, fd %x\n",
			  myset->resolved[best].name, fid););
    
    *fd = fid;
    result = 0;

 out:
    LOG_EXIT(0, SETITERATE);
    return result;
}


int do_set_rewind_iter(struct procSetInfo *settbl, int handle)
{
	struct setType *myset;
	PROCESS_t p;
	int i, result = 0;

        p = settbl->p;

	LOG_ENTRY(0, REWINDITERATOR);
    
	if (handle < 0 || handle >= NOPENSETS) {
		result = -EBADF;
		goto out;
	}

	if ((myset = settbl->mySets[handle]) == NULL) {
		result = -EBADF;
		goto out;
	}

	CRITICAL(myset->lock, myset->iterator = 0;);

	for (i = 0; i < myset->setSize; i++) {
		if (myset->resolved[i].state == ELEM_SEEN &&
		    myset->resolved[i].fsp != (FSOBJ_t)-1) {
			CRITICAL(myset->lock,
			 {
				/* Remove one ref, since we are
				 * downgrading this element to a
				 * vnode.  Process may or may not have
				 * recycled the fd */
				 RELEASE_FP(myset->resolved[i], p);
				 myset->resolved[i].state = ELEM_OPENED;
			 });
		}
	}
	myset->flags &= ~EXHAUSTED;

 out:
	LOG_EXIT(0, REWINDITERATOR);
	return result;
}	
