#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#ifndef _SETS_PRIVATE_H_
#define _SETS_PRIVATE_H_

/*------------------------------------------------------+
 | Pull in the appropriate system-dependent definitions |
 +------------------------------------------------------*/

#ifdef LINUX
#include "kernel/sets/linux21/sets_linux.h"
#endif
#ifdef __BSD44__
#include "kernel/sets/bsd44/sets_bsd44.h"
#endif
#ifdef WIN32
#include "kernel/sets/win32/sets_win32.h"
#endif

/*------------------------------------------------------+
 | System-independent definitions                       |
 +------------------------------------------------------*/

#include "kernel/sets/common/sets.h"
#include "kernel/sets/common/sets_message.h"

#define QUERY_MARKER '!'	/* shortcut to 1st char in marker (for case) */
#define MAXOPENSETS 300		/* Max on number of open sets in system. */
#define NOPENSETS 30		/* Max on number of open sets per process. */
#define DEFAULT_SET_SIZE 25	/* Default starting size of element member */

extern long lastOpenSet;	/* 0 if no sets open, or highest used index */
extern int  numWorkers;		/* number of daemons running */

/*------------------------------------------------------+
 | Set element details                                  |
 +------------------------------------------------------*/

typedef enum {
    ELEM_FREE,			/* this slot is not reserved. */
    ELEM_TAKEN,			/* this slot has been reserved. */
    ELEM_FETCHING, 		/* A prefetch is pending. 
				 * Inferred by fsp == 0 and OPENED */
    ELEM_OPENED, 		/* state after insertion, before yield.
				 * fsp != 0 --> has been opened */
    ELEM_FAILED, 		/* The prefetch failed. 
				 * Inferred by fsp == -1 and OPENED */
    ELEM_SEEN			/* The element has been yielded */
} elem_state;

typedef struct elemType {
    char       *name;		/* element's name */
    elem_state  state;		/* State of this element */
    u_short     nincore;	/* number of bytes (of prefix) NOT in
				 * the buffer cache */
    u_short     error;		/* The error that occurred getting a
				 * known member */
    int         toyield;	/* linked list of members ready to go. */
    FSOBJ_t     fsp;		/* pointer to system object */
    FILEOBJ_t   fp;		/* file pointer */
    FILEID      fd;		/* system-specifier file identifier
				 * (valid after yielded) */
} elemType;


/*------------------------------------------------------+
 | Set details                                          |
 +------------------------------------------------------*/

struct setType {
    u_long	 unique;	/* Uniquifier for this set. */
    u_short	 refct;		/* Number of pointers to this set. */
    u_short	 flags;		/* status of the set */
    char        *spec;		/* unresolved portion of query. */
    elemType    *resolved;	/* To be grown ala amortized analysis tables */
    u_short       resSize;	/* Current size of the resolved array */
    u_short       setSize;	/* number of members known to be in the set */
    u_short       expands;	/* Number of expansions to be done. */
    u_short       opens;		/* Highwater of opens (<=limitOpens) */
    int          toyield;	/* linked list of members ready to go. */
    short        iterator;	/* Current place of iterator in set */
    u_short       digestOffset;	/* Current place of digest in set */
    char        *playpen;	/* place for string manipulation */
    struct setType *s1, *s2;	/* Sets from which set is derived (union) */

    PROCESS_t    owner;		/* process of set creator */
    FSOBJ_t      curdir;	/* current working dir of creator. */
    FSOBJ_t      rootdir;	/* file system root of creator. */
    SETS_MUTEX   lock;		/* Should this be per resolved? */
    CONDITION    cond;		/* condition on which to sleep */
};

/* setStruct->flags values */
#define DONEEXPANDING   0x01	/* Set when a set is done producing members */
#define DELAYOPENS      0x02	/* Set if only digest. (don't wait for open) */
#define EXHAUSTED       0x04	/* Set if all elems have been yielded. */
#define DUP_ON_FORK     0x10	/* Duplicate the set on fork. */
#define SHARE_ON_FORK   0x20	/* Share set between processes */
#define OPEN_PREFETCHES 0x40	/* OPEN_OBJECT should prefetch */
#define INTERPRET_PAGES 0x80    /* tell warden to try to filter members */

/* shortcuts to check flag field */
#define DoneExpanding(s)	((s)->flags & DONEEXPANDING)
#define DelayOpens(s)		((s)->flags & DELAYOPENS)
#define Exhausted(s)		((s)->flags & EXHAUSTED)
#define Duplicate(s)		((s)->flags & DUP_ON_FORK)
#define Share(s)		((s)->flags & SHARE_ON_FORK)
#define OpenPrefetches(s)       ((s)->flags & OPEN_PREFETCHES)
#define InterpretPages(s)       ((s)->flags & INTERPRET_PAGES)

/*------------------------------------------------------+
 | Per-process structure (set table)                    |
 +------------------------------------------------------*/

struct procSetInfo {
    struct procSetInfo *next;
    struct procSetInfo **prev;
    struct setType     *mySets[NOPENSETS];
    PROCESS_t           p;	/* process owning this procSetInfo */
    int                 nOpen;	/* counter of open sets */
};

/* protect a critical section
   lck:  address of mutex
   body: the critical section code
   */
#define CRITICAL(lock,body)    \
	do {                       \
		MUTEX_LOCK(&(lock));   \
		body;                  \
		MUTEX_UNLOCK(&(lock)); \
	} while (0);


/*------------------------------------------------------+
 | Debugging                                            |
 +------------------------------------------------------*/

extern int setsDebugLevel;
#define DEBUG(mask, body)	if ((setsDebugLevel) & (mask)) {body}

#define OPEN		0x0001
#define ITERATE		0x0002
#define MYCLOSE		0x0004
#define UTIL		0x0008
#define PATHNAME	0x0010
#define QUERY		0x0020
#define MESSAGES_SMALL	0x0040
#define DIGEST		0x0080
#define THREAD		0x0100
#define MESSAGES	0x0200
#define MEM_ALLOC	0x0400
#define UNION		0x0800
#define INTERSECT	0x1000
#define ITERATE_LONG	0x2000
#define PREFETCH	0x4000


/*------------------------------------------------------+
 | Miscellaneous Prototypes                             |
 +------------------------------------------------------*/

int expand_pathname __P((workerMsgType *message, /* pathname.c */
			 int myindex));
int expand_cursor   __P((workerMsgType *message, /* query.c */
			 int *free_msg, int myindex));
void do_union     __P((workerMsgType *message,   /* setunion.c */
		       int myindex));
void do_intersect __P((workerMsgType *message,   /* setintersect.c */
		       int myindex));
void setInit         __P((void));                /* sets_util.c */
void setExit         __P((void));
int  find_set_elem_slot __P((struct setType *));
int  createSet       __P((char *name, struct procSetInfo *settbl));
void releaseSet      __P((struct setType *myset, int myindex));
int  insertNewMember __P((struct setType *set, char *name, int myindex));
void setPrint        __P((struct setType *set));

void sets_fork        __P((PROCESS_t, PROCESS_t));  /* sets_util.c */
void sets_proc_exit   __P((struct procSetInfo *));  /* setclose.c */

/* Operating system specific, but same interface. */
FILEID setup_fd        __P((struct setType *, int, int));
void init_bufferManagement __P((void));
void purgePrefetchBuffers  __P((void));
int  sets_get_dir_page __P((FSOBJ_t, char *, int *, PROCESS_t));
int  open_object       __P((struct setType *, elemType *, int));
void close_elem        __P((elemType *elem, PROCESS_t p));
int  expand_part       __P((FSOBJ_t vp, char *name, FSOBJ_t *vpp,
			    workerMsgType *msg));
#ifndef WIN32
int  expand_wildcard   __P((FSOBJ_t, char *, workerMsgType *, int));
#else
int  expand_wildcard   __P((char *, char *, workerMsgType *, int));
#endif

/* main API routines */
int do_set_open        __P((struct procSetInfo *, const char *, int, SETID *));
int do_set_close       __P((struct procSetInfo *, int));
int do_set_daemon      __P((void));
void kill_daemons      __P((void));
int do_set_iterate     __P((struct procSetInfo *, SETID, u_int, char *,
						   u_int, u_int *, FILEID *));
int do_set_digest      __P((struct procSetInfo *, SETID, char *,
						   unsigned long, int *));
int do_set_rewind_iter __P((struct procSetInfo *, int));
int do_set_rewind_dig  __P((struct procSetInfo *, int));
int do_set_member      __P((struct procSetInfo *, int, char *));
int do_set_size        __P((struct procSetInfo *, int, int *));
int do_set_intersect   __P((struct procSetInfo *, SETID, SETID, int, SETID *));
int do_set_union       __P((struct procSetInfo *, SETID, SETID, int, SETID *));
int do_set_restrict    __P((struct procSetInfo *, SETID, char *,
			    int, SETID *));

/*------------------------------------------------------+
 | Buffer Cache Management (currently ignored)          |
 +------------------------------------------------------*/

extern int sets_max;
extern int sets_restart;
extern int sets_count;
extern int pin_count;
extern int pin_max;
#define PREFETCHING (sets_max != 0)

#endif /* _SETS_PRIVATE_H_ */
