#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"

struct {
	struct workerMsg *head;	/* from NetBSD TAILQ */
	struct workerMsg **tail;
	SETS_MUTEX lock;
	CONDITION cond;
} messageQueue;

void printMessageQueue(void)
{
	workerMsgType *msg;
	int i;
    
	for (i = 0, msg = messageQueue.head; msg; i++, msg = msg->next) {
		S_PRINTF("    %d: (%p)> set %x", i, 
		       msg, (int)msg->set);
		if (msg->operation == WM_OPEN) 
			S_PRINTF(" OPN index %d\n", msg->msg.open.index);
		else if (msg->operation == WM_EXPAND)
			S_PRINTF(" EXP path %s query %s (fsobj = %p)\n", 
			       msg->msg.expand.pathname, msg->msg.expand.query,
			       msg->msg.expand.vp);
		else if (msg->operation == WM_QUERY)
			S_PRINTF(" QUE path %s suffix %s cursobj %p\n",
			       msg->msg.query.pathname,
			       msg->msg.query.suffix,
			       msg->msg.query.handle);
		else if (msg->operation == WM_UNION)
			S_PRINTF(" UNI bm1 %p bm2 %p\n", msg->msg.merge.bm1,
			       msg->msg.merge.bm1);
		else if (msg->operation == WM_INTERSECT)
			S_PRINTF(" INT bm1 %p\n", msg->msg.merge.bm1);
		else if (msg->operation == WM_EXIT)
			S_PRINTF(" EXIT\n");
	}
}

void initMessages(void)
{
	MUTEX_INIT(&messageQueue.lock);
	CONDITION_INIT(&messageQueue.cond);
	messageQueue.head = NULL;
	messageQueue.tail = &messageQueue.head;
}

void enqueueMessage(workerMsgType *msg, int myindex)
{
	LOG_ENTRY(myindex, ENQUEUE);
	DEBUG(MESSAGES_SMALL, S_PRINTF("Queueing msg %p (%d)\n", msg, msg->operation););

	CRITICAL(messageQueue.lock,
	 {
		 msg->next = NULL;
		 msg->prev = messageQueue.tail;
		 *messageQueue.tail = msg;
		 messageQueue.tail = &msg->next;
		 CONDITION_SIGNAL(&messageQueue.cond);
		 DEBUG(MESSAGES, { S_PRINTF("(%d): postEnqueue: \n", myindex);
				   printMessageQueue();
			   });
	 });
	LOG_EXIT(myindex, ENQUEUE);
}

workerMsgType *dequeueMessage(int myindex)
{
	workerMsgType *result = NULL;

	LOG_ENTRY(myindex, DEQUEUE);
	CRITICAL(messageQueue.lock, {
		while (messageQueue.head == NULL) {
			DEBUG(UTIL, S_PRINTF("(%d) waiting\n", myindex););
			CONDITION_WAIT(&messageQueue.cond, &messageQueue.lock);
			DEBUG(UTIL, S_PRINTF("(%d) unwaiting\n", myindex););
		}

		DEBUG(MESSAGES, S_PRINTF("(%d): preDequeue:\n", myindex););
		DEBUG(MESSAGES, printMessageQueue(););

		/* Will be NULL if nothing is there. */
		result = messageQueue.head;
		if (result->next != NULL)
			result->next->prev = result->prev;
		else
			messageQueue.tail = result->prev;
		*result->prev = result->next;
		DEBUG(MESSAGES_SMALL, S_PRINTF("deQ result = %p\n", result););
	});

	LOG_EXIT(myindex, DEQUEUE);
	return result;
}

/* Do the right thing when purging a message.  NOTE: Assume most
 * members are valid and set, as msg is bogus if not. */
void purgeMessage(workerMsgType *message, int myindex) {
	LOG_ENTRY(myindex, PURGEMESSAGE);
    
	switch(message->operation) {
	case WM_OPEN:
	case WM_READ:
	case WM_EXIT:
		break;		/* No specific cleanup */
	case WM_EXPAND: {
		if (message->msg.expand.pathname) /* Reused in reslvd array. */
			STRFREE(message->msg.expand.pathname);
		if (message->msg.expand.query)
		        STRFREE(message->msg.expand.query);
                if (message->msg.expand.vp)
			FSOBJ_RELE(message->msg.expand.vp);
		break;
	}
	case WM_QUERY:
		STRFREE(message->msg.query.pathname);
		STRFREE(message->msg.query.suffix);
		CLOSE_CURSOBJ(message->msg.query.handle,
			      message->set->owner);
		FSOBJ_RELE(message->msg.query.vp);
		break;
	case WM_UNION: {
		MYFREE(message->msg.merge.bm1, message->msg.merge.bm1Size);
		MYFREE(message->msg.merge.bm2, message->msg.merge.bm1Size);
		break;
	}
	case WM_INTERSECT: {
		MYFREE(message->msg.merge.bm1, message->msg.merge.bm1Size);
		break;
	}
	}

	if (message->set)
	    CRITICAL(message->set->lock, {
		if (--message->set->refct == 0)
		    /* Reclaim set if no references */
		    releaseSet(message->set, myindex);
	    });
	MYFREE(message, sizeof(workerMsgType));
	LOG_EXIT(myindex, PURGEMESSAGE);
}
