#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_hash.h"

/* Globals */
struct setType openSets[MAXOPENSETS];
long lastOpenSet = 0;
int limitOpens = 5;
long SetsUnique = 0;
int numWorkers;
SETS_MUTEX descriptorLock;

/* Debugging flag */
int setsDebugLevel = (int)0xfcbf;
/* OPEN|ITERATE|DIGEST|UTIL|QUERY|MYCLOSE|UNION; */

void setInit(void)
{
#ifndef WIN32
	sets_hashinit();
#endif

	memset(openSets, 0, MAXOPENSETS * sizeof(struct setType));

	numWorkers = 0;

	MUTEX_INIT(&descriptorLock);
	
	init_bufferManagement();
	
	initMessages();
}

void setExit(void)
{
#ifndef WIN32
	sets_hashexit();
#endif
}

int find_set_elem_slot(struct setType *set)
{
    int index;
    CRITICAL(set->lock, {
	for (index = set->setSize;
	     set->resolved[index].state && (index < set->resSize);
	     index++)
	    ;
	if (index == set->resSize) { /* Need to grow the array */
	    char *new;
	    MYALLOC(new, char *, set->resSize * sizeof(elemType) * 2);
	    memcpy(new, (char *)set->resolved,
		  set->resSize * sizeof(elemType));
	    memset(new + set->resSize * sizeof(elemType), 0,
		  set->resSize * sizeof(elemType));
	    MYFREE(set->resolved, set->resSize * sizeof(elemType));
	    set->resSize *= 2;
	    set->resolved = (elemType *)new;
	}
	set->resolved[index].state = ELEM_TAKEN;
	set->resolved[index].fsp = (FSOBJ_t)NULL;
	set->resolved[index].fp = (FILEOBJ_t)NULL;
	set->resolved[index].name = NULL;
	set->setSize++;
    });
    return index;
}

/* Common routine to create a set, returns index into openSets[] if no 
 * error.  Should assume the set is garbage (init all data) */
int createSet(char *name, struct procSetInfo *settbl)
{
	struct setType *set;
	int i, j, result;

	LOG_ENTRY(0, CREATESET);

	/* Allocate an entry in this process's open set table */
	for (i = 0, j = 0; /* Find an unused slot */
	     (j < NOPENSETS) && (settbl->mySets[i] != 0);
	     i = (i + 1) % NOPENSETS, j++)
		;
  
	if (j == NOPENSETS)	{
		/* Shouldn't be lower, can't be higher. */	
		if (settbl->nOpen != NOPENSETS)
//			panic("sets: settbl->nOpen!=NOPENSETS");
			exit(-1);
		result = -EMFILE;
		goto out;
	}
	result = i;
  
	/* Allocate a new entry in the system-wide set table */
	CRITICAL(descriptorLock, {
		for (i = lastOpenSet, j = 0; /* Find an unused descriptor */
		     (j < MAXOPENSETS);
		     i = (i + 1) % MAXOPENSETS, j++) 
			if (openSets[i].refct == 0) {
				lastOpenSet = i;
				openSets[i].refct = 1; /* Mark as used. */
				break;
			}
	});
  
	if (j == MAXOPENSETS) {	/* No free descriptors */
		result = -ENFILE;
		goto out;
	}
  
	settbl->mySets[result] = &openSets[i];
	settbl->nOpen++;
  
	/* Initialize the iterator and digest offsets. */
	openSets[i].iterator = (short)-1; /* indicate no expand was queued. */
	openSets[i].digestOffset = (short)0;
  
	set = &openSets[i];	/* Assume it was zeroed */
	set->unique = ++SetsUnique;
  
	set->spec = name; /* Either null or a valid name I hope! */
	
	set->resSize = DEFAULT_SET_SIZE;
	set->setSize = 0;
	set->expands = 0;
	set->opens = 0;
	set->flags = 0;
	STRALLOC(set->playpen);
	
	/* Stuff for management of buffer cache. */
	set->toyield = -1;
  
	MYALLOC(set->resolved, elemType *, set->resSize * sizeof(elemType));
	memset(set->resolved, 0, set->resSize * sizeof(elemType));
  
	MUTEX_INIT(&set->lock);
	CONDITION_INIT(&set->cond);
	
#ifndef WIN32
	/* All operations on this set are run as if by caller to createSet */
	set->curdir = Process_cdir;
	FSOBJ_REF(set->curdir);
	set->rootdir = Process_root;
	if (Process_root) 
		FSOBJ_REF(set->rootdir);
#endif
  
	/* Return the index of the slot we used. */
 out:
	LOG_EXIT(0, CREATESET);
	return result;
}

#if 0
/* XXX currently UNCALLED since I'm not hooking fork, and haven't
 * figured out another way XXX */
/* Used by fork, copies the procSetInfo of the old process for the new process. */
void sets_fork(PROCESS_t parent, PROCESS_t child)
{
	struct procSetInfo *oldstuff, *newstuff;
	int i;

	LOG_ENTRY(0, COPYOPENSET);

	if ((oldstuff = sets_hashlookup(PID(parent))) == 0) {
		LOG_EXIT(0, COPYOPENSET);
		return;
	}

	MYALLOC(newstuff, struct procSetInfo *, sizeof(struct procSetInfo));
	memset(newstuff, 0, sizeof(struct procSetInfo));
	/*     memcpy(newstuff, oldstuff, sizeof(struct procSetInfo)); */
  
	for (i = 0; i < NOPENSETS; i++) {
		if (oldstuff->mySets[i]) {
			struct setType *myset = oldstuff->mySets[i];
      
			if (Share(myset)) {
				/* Bump the count on any sets we inherited. */
				CRITICAL(oldstuff->mySets[i]->lock, 
					 oldstuff->mySets[i]->refct++);
				DEBUG(OPEN, S_PRINTF("Sharing set %s on fork\n",
						   myset->spec););

				newstuff->mySets[i] = oldstuff->mySets[i];
			}
      
			else if (Duplicate(myset)) {
				char *name;
				int result;
		
				STRALLOC(name);
				strcpy(name, myset->spec);
		
				result = createSet(name, newstuff);
				myset->flags |= DUP_ON_FORK;

				if (result < 0)
					S_PRINTF("AccckkK! on fork got error %d dup-ing %s\n",
					       result, myset->spec);
				else {

					/* We need to force the set to be in
					 * the same place. Rather than do it
					 * right, I put this hack here (yes 
					 * I'm lazy). If result > i, createSet
					 * skipped i, and something must be is
					 * wrong.  DCS 11/20/94
					 */
					if (result > i) 
						panic("createSet must be wrong");
					else if (result < i) {
						newstuff->mySets[i] = newstuff->mySets[result];
						newstuff->mySets[result] = NULL;
					}

					DEBUG(OPEN, S_PRINTF("Copying set %p on fork\n", myset);
					      setPrint(newstuff->mySets[result]););
				}
			}

			else {	/* Don't copy or share the set. */
				DEBUG(OPEN,
				      S_PRINTF("Not copying set %s on fork\n", 
					     myset->spec););
			}
		}
	}
	DEBUG(UTIL, S_PRINTF("Done with sets_fork\n"););
	LOG_EXIT(0, COPYOPENSET);
	newstuff->p = child;
	sets_hashinsert(newstuff);
	return;
}
#endif

/* Common routine to add a new member to the set. Since it is a set, we
 * must also be careful to eliminate duplicates at this point.
 * XXX the pathname might be relative -- need to remove assumption that
 *     cdir is the same when iterate/prefetch actually opens this guy
 * XXX Maybe I should use the fact that expandPathname already located a vnode
 *     for this guy?
 */
int insertNewMember(struct setType *set, char *name, int myindex)
{
	int i;
	workerMsgType *newmsg;

	LOG_ENTRY(myindex, INSERTMEMBER);

	/* Allocate an entry in the set to hold this member */
	CRITICAL(set->lock, { 
		for (i = 0; i < set->resSize; i++) {
			if (set->resolved[i].state == ELEM_FREE) {
				set->resolved[i].state = ELEM_TAKEN;
				set->resolved[i].name = name;
				set->setSize++;
				break;
			}

			else if (strcmp(set->resolved[i].name, name) == 0) {
				/* We found a duplicate. */
				DEBUG(UTIL, 
				      S_PRINTF("duplicate %s at index %d\n",
					     name, i););
				STRFREE(name);
				name = 0;
				break;
			}
		}
	
		if (i == set->resSize) { /* Need to grow the array */ 
			char *new; 
	    
			DEBUG(UTIL, S_PRINTF("Growing resolved array\n"););
			MYALLOC(new, char *, 
				set->resSize * sizeof(elemType) * 2); 
			memcpy(new, (char *)set->resolved,
			      set->resSize * sizeof(elemType)); 
			memset(new + set->resSize * sizeof(elemType), 0,
			      set->resSize * sizeof(elemType)); 
			MYFREE(set->resolved, 
			       set->resSize * sizeof(elemType)); 
			set->resSize *= 2; 
			set->resolved = (elemType *)new;
			/* Since we just grew the array, 
			 * the slot i must be free. */
			set->resolved[i].state = ELEM_TAKEN;
			set->resolved[i].name = name;
			set->setSize++;
		} 
	});

	if (name) {
		DEBUG(UTIL, S_PRINTF("(%d): Inserting %s at resolved[%d]\n", 
				   myindex, name, i););
	}

	if (!DelayOpens(set) && PREFETCHING && 
	    set->resolved[i].state == ELEM_TAKEN) {
		/* create a new message from msg */
		CRITICAL(set->lock,
			 if (set->opens < limitOpens) {
				 set->resolved[i].state = ELEM_OPENED;
				 set->opens++;
			 });
	
		if (set->resolved[i].state == ELEM_OPENED) {
			MAKE_NEW_MSG(newmsg, WM_OPEN, set);
	    
			newmsg->msg.open.index = i;
	    
			DEBUG(UTIL, 
			      S_PRINTF("(%d): (%d) Spooling open for %x:%s\n",
				     myindex, set->opens, (int)name, name););
	    
			enqueueMessage(newmsg, myindex); /* And Yield */
		}
	}

	LOG_EXIT(myindex, INSERTMEMBER);
	return i;
}

/* Common routine to purge an open set.  Don't bzero the set here,
 * assume it's garbage in createSet.
 * !!assume this is called within a critical section!! */
void releaseSet(struct setType *myset, int myindex)
{
	int i;
	elemType *elem;

	LOG_ENTRY(myindex, RELEASESET);
    
	/* Need to deallocate the set, release all vnodes/fileps
	 * referenced by resolved array */
	DEBUG(MYCLOSE, S_PRINTF("Deallocating set %p\n", myset););

#ifdef PRE_READ
	freePinnedBuffers(myset, myindex);
#endif
	
	for (i = 0; i < myset->setSize; i++) {
	        elem = &(myset->resolved[i]);
		/* If element isn't free, it must have a name. */
		if (elem->state > ELEM_FREE) {
			/* May be zero if an error occurred during expansion */
			if (elem->name) 
				STRFREE(elem->name);

			if (elem->state >= ELEM_OPENED)
			        close_elem(elem, myset->owner);
		}	
	}	
	
	if (myset->spec)	/* Unions don't have a name. */
		STRFREE(myset->spec);
	STRFREE(myset->playpen);
    
	MYFREE(myset->resolved, myset->resSize * sizeof(elemType));

	FSOBJ_RELE(myset->curdir);
	if (myset->rootdir)
	        FSOBJ_RELE(myset->rootdir);

	/* Reset the "lastopenset" counter, this uses a different lock -- 
	 * no deadlock */
	if (myset == &openSets[lastOpenSet]) {
		CRITICAL(descriptorLock, {
			for (i = lastOpenSet; i; i--) {
				if (openSets[i].refct == 0)
					lastOpenSet--;
				else
					break;
			}
		});

		if (lastOpenSet == 0) {
			purgePrefetchBuffers();
		}
		DEBUG(UTIL, S_PRINTF("Reset lastOpenSet\n"););
	}

	LOG_EXIT(myindex, RELEASESET);
}

void setPrint(struct setType *set)
{
	int i;
    
	S_PRINTF("Open set: %p iterator %d digest %d\n", set,
	       set->iterator, set->digestOffset);
	S_PRINTF(" count %d resolved %p resSize %d\n",
	       set->refct, set->resolved, set->resSize);
	S_PRINTF(" setSize %d e %d o %d flags %d\n", set->setSize,
	       set->expands, set->opens, set->flags);
	S_PRINTF(" unique %ld curdir %p rootdir %p\n", set->unique, 
	       set->curdir, set->rootdir);

	/* Print resolveds? */
	DEBUG(MESSAGES_SMALL, {
	    S_PRINTF("Resolved:\n");
	    for(i = 0; i < set->setSize; i++)
		S_PRINTF("  %d: state %d fsp %p fp %p "
		       "error %d name *(%p) == %s\n",
		       i, set->resolved[i].state,
		       set->resolved[i].fsp,
		       set->resolved[i].fp,
		       set->resolved[i].error,
		       set->resolved[i].name,
		       (set->resolved[i].name ? set->resolved[i].name :
			"UNEXPANDED"));
	});
}
