#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_hash.h"

int do_set_union(struct procSetInfo *settbl, SETID s1, SETID s2,
		 int flags, SETID *handle)
{
	struct setType *set1, *set2, *myset;
	workerMsgType *msg;
	int result;
    
	LOG_ENTRY(0, SETUNION);
	
	DEBUG(UNION, S_PRINTF("Entering setUnion(%d, %d, %d)\n",
			    s1, s2, flags););
    
	if (settbl->nOpen >= NOPENSETS) {
		result = -EMFILE;
		goto out;
	}
    
	/* Make sure both s1 and s2 are valid */
	if ((s1 < 0 || s1 >= NOPENSETS) ||
	    (s2 < 0 || s2 >= NOPENSETS) ||
	    (settbl->mySets[s1] == NULL) ||
	    (settbl->mySets[s2] == NULL))
	{
		result = -EBADF;
		goto out;
	}
    
	/* Create a set to hold the union. */
	/* createSet will return a negative value on error,
	 * or the open file handle */
	result = createSet(0, settbl);
	
	if (result < 0) {
		goto out;
	}
	
	myset = settbl->mySets[result];
	set1 = settbl->mySets[s1];
	set2 = settbl->mySets[s2];
    
	myset->s1 = set1;
	myset->s2 = set2;
    
	/* Don't want iterate or digest to spool an expand msg
	 * for this type of set.*/
	CRITICAL(myset->lock, {
		(settbl->mySets[result]->iterator)++;
		myset->expands = 1; /* we are going to queue 1 union msg. */
	} );
	
	/* Create a message to start the processing of the set. */
	MAKE_NEW_MSG(msg, WM_UNION, myset);
	
	MYALLOC(msg->msg.merge.bm1, char *, set1->resSize);
	memset(msg->msg.merge.bm1, 0, set1->resSize);
	msg->msg.merge.bm1Size = set1->resSize;
	
	MYALLOC(msg->msg.merge.bm2, char *, set2->resSize);
	memset(msg->msg.merge.bm2, 0, set2->resSize);
	msg->msg.merge.bm2Size = set2->resSize;
	
	enqueueMessage(msg, -1); /* And yield */
	
	DEBUG(UNION, setPrint(settbl->mySets[result]););

	*handle = result;
	result = 0;

out:
	LOG_EXIT(0, SETUNION);
	return result;
}

void do_union(workerMsgType *message, int myindex)
{
	int i, max;
	struct setType *myset = message->set;
	struct setType *set = NULL;
	char *old;

	LOG_ENTRY(myindex, DOUNION);
	
	DEBUG(UNION, S_PRINTF("Entering doUnion on %p (set %p)\n",
			     message, myset); );
    
	/* If the rslvd arrays in the sets have grown, grow the bitmap too. */
	if (message->msg.merge.bm1Size < myset->s1->resSize) {
		old = message->msg.merge.bm1;
		MYALLOC(message->msg.merge.bm1, char *, myset->s1->resSize);
		memcpy(message->msg.merge.bm1, old,
		       message->msg.merge.bm1Size);
		MYFREE(old, message->msg.merge.bm1Size);
		message->msg.merge.bm1Size = myset->s1->resSize;
	}
	if (message->msg.merge.bm2Size < myset->s2->resSize) {
		old = message->msg.merge.bm2;
		MYALLOC(message->msg.merge.bm2, char *, myset->s2->resSize);
		memcpy(message->msg.merge.bm2, old,
		       message->msg.merge.bm2Size);
		MYFREE(old, message->msg.merge.bm2Size);
		message->msg.merge.bm2Size = myset->s2->resSize;
	}

	max = (myset->s1->resSize > myset->s2->resSize) ? myset->s1->resSize :
		myset->s2->resSize;
	for (i = 0; i < max; i++) {
		if (i < myset->s1->resSize)
			if ((message->msg.merge.bm1[i] == 0) &&
			    myset->s1->resolved[i].name) {
				/* found a new member */
				message->msg.merge.bm1[i] = 1;
				set = myset->s1;
				break;
			}

		if (i < myset->s2->resSize)
			if ((message->msg.merge.bm2[i] == 0) &&
			    myset->s2->resolved[i].name) {
				/* found a new member */
				message->msg.merge.bm2[i] = 1;
				set = myset->s2;
				break;
			}
	}

	if (set == NULL) {
		/* Maybe there's nothing more to get? */
		if (DoneExpanding(myset->s1) && DoneExpanding(myset->s2)) {
			/* Indicate that this set is fully expanded. */
			DEBUG(UNION, S_PRINTF("Done expanding union %p\n",
					     myset); );
			CRITICAL(myset->lock, {
				if (--myset->expands == 0) {
					/* We finished all expands */
					myset->flags |= DONEEXPANDING;
				}
			});
			CONDITION_SIGNAL(&(myset->cond));
	    
			/* Free up our message */
			purgeMessage(message, myindex);
			goto out;
		}
		
		/* What do we do? Condition_wait won't work because either set
		 * may be modified, and we can only wait on 1 event. Simple
		 * solution: requeue the message.
		 */
		enqueueMessage(message, myindex);
		goto out;
	}
	
	/* Use old 'cause it is there and unused. */
	STRALLOC(old);
	strcpy(old, set->resolved[i].name);
	
	/* Allocate an entry in the set to hold this member */
	insertNewMember(myset, old, myindex);
	
	/* What the heck, requeue the union message. */
	enqueueMessage(message, myindex);
	
	/* Notify any waiting user that another entry has been added. */
	CONDITION_SIGNAL(&(myset->cond));
	
out:
	LOG_EXIT(myindex, DOUNION);
}
