#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <linux/string.h>

#include <kernel/sets/common/sets_internal.h>
#include <kernel/sets/common/sets_domain.h>
#include <kernel/sets/common/sets_hash.h>

static int wc_rd_callback(void *buf, const char *word, int namlen,
			  off_t offset, ino_t ino);

int expand_part(struct dentry *vp, char *name, struct dentry **vpp,
		workerMsgType *msg)
{
    struct dentry *target_de;
    
    if (name[0] == '\0') {
	DEBUG(PATHNAME, printf("null expand_part\n"););
	*vpp = vp;
	FSOBJ_REF(*vpp);
	return 0;
    }
    
    DEBUG(PATHNAME, printf("expand_part in %p for %s\n", vp, name););
    
    if (name[0] == '/')
	target_de = lookup_dentry(&name[1], msg->set->rootdir, 1);
    else
	if (!vp)
	    target_de = lookup_dentry(name, msg->set->curdir, 1);
	else
	    target_de = lookup_dentry(name, vp, 1);
    
    if (IS_ERR(target_de)) {
	DEBUG(PATHNAME, printf("expand_part: lookup fails.\n"););
	*vpp = NULL;
	return SETS_EXP_BADLOOKUP;
    }

    *vpp = target_de;

    return 0;
}

struct wildcard_info {
    char *pattern;
    char *suffix;
    char *scratch;
    workerMsgType *msg;
    int   myindex;
    struct dentry *dir;
};

/*
 * expand_wildcard
 *  in
 *   vp:      dentry for the directory where matching will occur
 *   pattern: beginning of the component with the wildcard
 *  out
 *   matches: list of strings that match the expression
 *
 *  notes
 *   - pattern will be past the '/'
 *   - the end of the component is marked by '\0'
 *   - matched strings include no wildcards or '/'
 *   - REF the vp while using, then RELE
 */
int expand_wildcard(struct dentry *dir, char *pattern, 
		    workerMsgType *msg, int myindex)
{
	char *suffix;
	struct file *f;
	struct inode * inode;
	struct wildcard_info wc_info;
	int error;

	if (!dir) {
		if (pattern[0] == '/') {
			dir = msg->set->rootdir;
			pattern++;
			msg->msg.expand.pathname[0] = '/';
			msg->msg.expand.pathname[1] = '\0';
		} else
			dir = msg->set->curdir;
		FSOBJ_REF(dir);
	}

	/* isolate pattern, point at suffix */
	for (suffix = pattern; *suffix && *suffix != '/'; suffix++)
		;
	if (*suffix)
		*suffix++ = '\0';
  
	DEBUG(PATHNAME, printf("(%d) lookup on dvp %p\n", myindex, dir););

	if (!IS_DIR_OBJ(dir))
		return SETS_EXP_NOTDIR;

	inode = dir->d_inode;
	f = get_empty_filp();
	if (!f) {
		FSOBJ_RELE(dir);
		return SETS_EXP_NOFP;
	}

	f->f_flags = 0;
	f->f_mode = 0;
	f->f_dentry = dir;
	f->f_pos = 0;
	f->f_reada = 0;
	if (!inode->i_op) {
		put_filp(f);
		FSOBJ_RELE(dir);
		return SETS_EXP_BADDIR;
	}
	f->f_op = inode->i_op->default_file_ops;
	if (!f->f_op->readdir) {
		put_filp(f);
		FSOBJ_RELE(dir);
		return SETS_EXP_BADDIR;
	}

	wc_info.pattern = pattern;
	wc_info.suffix = suffix;
	wc_info.myindex = myindex;
	wc_info.msg = msg;
	wc_info.dir = dir;
	wc_info.scratch = msg->set->playpen;
	do {
		down(&inode->i_sem);
		error = f->f_op->readdir(f, &wc_info, wc_rd_callback);
		up(&inode->i_sem);
		printk("readdir gives %d\n", error);
		printk("fpos = %d  i_size = %d\n", (int)f->f_pos,
		       (int)f->f_dentry->d_inode->i_size);
		if (error < 0) {
			put_filp(f);
			FSOBJ_RELE(dir);
			return SETS_EXP_BADDIR;
		}
	} while (error >= 0 && f->f_pos < f->f_dentry->d_inode->i_size);
  
	/* Clean up. */
	FSOBJ_RELE(dir);

	return 0;
}

static int wc_rd_callback(void *buf, const char *word, int namlen,
			  off_t offset, ino_t ino)
{
	struct wildcard_info *wc_info = (struct wildcard_info *)buf;
	workerMsgType *msg = wc_info->msg;
	enum { notdone, matched, notmatched } result;
	char *word_save;	/* copy of the work (null-term) */
	char *wptr;		/* Current location in processing word */
	char *pptr;		/* Current location in the pattern string */
	char *wild_wptr;	/* Last char in word to match wildcard */
	char *wild_pptr;	/* Char after wildcard in pattern string */
	char doWildcard;	/* True if currently matching to a wildcard. */

	/* For reasons of unix compatibility, a leading '*'
	 * won't match a "." */
	if (*(wc_info->pattern) == '*' && *word == '.')
		return 0;

	/* XXX sorry for the overhead.  the better solution is to use namlen
	 *     within the regexp matching code.  but that is Yuck. */
	strncpy(wc_info->scratch, word, namlen);
	word_save = wptr = wc_info->scratch;
	wptr[namlen] = '\0';

	DEBUG(PATHNAME, printf("(%d): glob %s word %s\n",
			       wc_info->myindex, wc_info->pattern, wptr););

	result = notdone;
	doWildcard = 0;
	wild_wptr = wild_pptr = 0;

	/* Go until we've determine if the word matches the pattern */
	for (pptr = wc_info->pattern; result == notdone; pptr++) {
		switch (*pptr) {
		case '?' : { /* Match on any single character */
			if (*wptr) {
				wptr++;	/* Matched this one,
					 * skip to next */
				
				/* "*?" == "?*", as an optimization to
				 * avoid backtracking skip this char
				 * but keep wildcard on */
				if (doWildcard) {
					/* Treat wptr as consumed */
					wild_wptr = wptr;
					/* Treat the "?" as the "*" */
					wild_pptr = pptr;
				}
			} else
				/* Word ended prematurely */
				result = notmatched;
			break;
		}
		
		case '[' : { /* list of chars to match with 1 
				      * char from w */
			char *p, *dash, match = 0;
			
			if (*wptr == 0) {
				/* Word ended prematurely */
				result = notmatched;
			} else {
				
				for (p = pptr + 1;
				     *p && *p != ']';
				     p++) {
					dash = p + 1;
					
					if (*dash == '-') { 
						/* Next char is a '-': 
						 * a sequence of
						 * matches */
						if (!(*wptr < *p ||
						      *wptr > *(++dash))) {
							match = 1;
							break;
						}
						p++; /* Skip the dash */
					} else if (*p == *wptr) {
						match = 1;
						break;
					}
				}
				
				if (match) {
					/* Skip ahead to the end of
					 * the selection */
					while (*pptr && *pptr != ']')
						pptr++;
					wptr++;	/* Matched this one,
						 * skip to next */
					/* No longer processing a
					 * '*' */
					doWildcard = 0;
				} else 
					result = notmatched;
			}
			break;
		}
		
		case '*' : {
			doWildcard = 1;
			wild_pptr = pptr;
			wild_wptr = wptr;
			break;
		}
		
		case 0 : { /* Exhausted the pattern string */
			if (*wptr == 0 || doWildcard) {
				result = matched;
			} else {
				result = notmatched;
			}
			break;
		}
		
		default : {
			if (*pptr == *wptr) {
				wptr++;	/* Definate match */
				/* No longer processing a '*' */
				doWildcard = 0;
			} else
				result = notmatched;
			break;
		}
		}
		
		/* If a mismatch, yet word wasn't fully
		 * processed, apply '*' */
		if (result == notmatched && *wptr) {
			if (doWildcard) {
				/* '*' ate another char */
				wild_wptr = ++wptr;
				pptr = wild_pptr;
				result = notdone;
			}	
			
				/* Fall back on last wildcard if can */
			else if (wild_wptr) {	
				/* Eat up one more char via '*' */
				wptr = ++wild_wptr;
				/* Restart pattern after '*' */
				pptr = wild_pptr;
				/* Start using '*' again */
				doWildcard = 1;
				result = notdone;
			}
		}
	}
	
	/* it is a true match, build a new message for it. */
	/* It's a match if the pattern is satisfied and either
	 * wildcard was in effect (which would consume the remainder
	 * of the word) or the word was completely processed. */
	if (result == matched) {
		DEBUG(PATHNAME, printf("(%d): glob matched %s\n",
				       wc_info->myindex, word_save););
		if (*(wc_info->suffix) != '\0') {
			workerMsgType *newmsg;

			MAKE_NEW_MSG_FROM_OLD(newmsg, msg);
			newmsg->operation = WM_EXPAND;
			STRALLOC(newmsg->msg.expand.pathname);
			strcpy(newmsg->msg.expand.pathname,
			       msg->msg.expand.pathname);
			STRALLOC(newmsg->msg.expand.query);
			strncpy(newmsg->msg.expand.query, word_save, namlen);
			newmsg->msg.expand.query[namlen] = '\0';
			strcat(newmsg->msg.expand.query, "/");
			strcat(newmsg->msg.expand.query, wc_info->suffix);
			newmsg->msg.expand.vp = wc_info->dir;
			FSOBJ_REF(wc_info->dir);
			msg->set->expands++;
			enqueueMessage(newmsg, wc_info->myindex);
		} else {
			/* don't expand - just insert it! */
			char *newmember;
			int len;

			STRALLOC(newmember);
			strcpy(newmember, msg->msg.expand.pathname);
			len = strlen(newmember);
			strncat(newmember, word_save, namlen);
			newmember[len+namlen] = '\0';
			insertNewMember(msg->set, newmember, wc_info->myindex);
		}
	}
	return 0;
}

void init_bufferManagement(void)
{
}

void purgePrefetchBuffers(void)
{
}

/* NOTE: critical that this is called directly from a system call - "current"
 *       is used by get_unused_fd */
int setup_fd(struct setType *set, int setnum, int elem)
{
    int fd;

    printk("setup_fd: set %p elem %d fsp %p fp %p\n",
	   set, elem, set->resolved[elem].fsp, set->resolved[elem].fp);

    fd = get_unused_fd();
    if (fd < 0)
	return -1;

    fd_install(fd, set->resolved[elem].fp);

    set->resolved[elem].fd = fd;
    set->resolved[elem].fp->f_count++; /* extra ref - live as long as set */

    printk("setup_fd: fsp->d_count: %d  fp->f_count: %d\n",
	   set->resolved[elem].fsp->d_count, set->resolved[elem].fp->f_count);

    return fd;
}

/*
 * OPEN_FSOBJ -- (doubles as OPEN_CURSOBJ) takes a dentry, opens and
 *               returns a file * for it
 *
 *               called by expand_query for cursor objects and by open_object
 *               for non-prefetchable files
 *               de assumed to carry a reference for this use
 */
int OPEN_FSOBJ(struct dentry *de, int flags, struct task_struct *p,
	       struct file **fpp)
{
    struct file *f;
    struct inode *inode = de->d_inode;
    int error = 0;

    ENTRY;

    f = get_empty_filp();
    if (!f) {
	error = -ENFILE;
	goto out;
    }
    
    f->f_flags = flags;
    f->f_mode = (flags+1) & O_ACCMODE;
    f->f_dentry = de;
    f->f_pos = 0;
    f->f_reada = 0;
    f->f_op = NULL;
    if (inode->i_op)
	f->f_op = inode->i_op->default_file_ops;
    if (f->f_op && f->f_op->open) {
	error = f->f_op->open(inode, f);
	if (error)
	    goto cleanup_all;
    }
    
    *fpp = f;
    goto out;
    
 cleanup_all:
    put_filp(f);
 out:
    EXIT;
    return error;
}

/* 
 * open_object -- takes an elemType and gets a dentry and file * for the
 *                named file.  Uses the prefetch_fn of a registered
 *                domain if possible and desired.
 */
int open_object(struct setType *myset, elemType *elem, int myindex)
{
    struct dentry *de;
    struct file *f;
    struct sets_domain *dom;
    int error = 0;

    ENTRY;

    de = lookup_dentry(elem->name, NULL, O_RDONLY);
    if (IS_ERR(de))
	goto out;

    dom = sets_find_domain(de);
    if (PREFETCHING && OpenPrefetches(myset) && dom)
	error = (dom->prefetch_fn)(de, O_RDONLY, myset->owner, &f);
    else
	error = OPEN_FSOBJ(de, O_RDONLY, myset->owner, &f);

    if (error)
	goto cleanup_dentry;

    elem->fsp = de;
    elem->fp = f;
    printk("open_object: fsp %p fp %p\n", elem->fsp, elem->fp);
    printk("open_object: fsp->d_count: %d  fp->f_count: %d\n",
	   elem->fsp->d_count, elem->fp->f_count);
    goto out;

 cleanup_dentry:
    dput(de);
 out:
    EXIT;
    return error;
}

/*
 * close_elem: clean up state for a set member (fsp, fp)
 *             the fd is cleaned up by the app itself
 */
void close_elem(elemType *elem, struct task_struct *p)
{
    if (elem->state < ELEM_OPENED || !elem->fsp || !elem->fp)
	return;

    /* the set held one reference to the file *, so that it won't go
     * away as long as the set is open, regardless of what the app
     * might do.  any references given-to/obtained-by the app will be
     * dealt with by close(2) or process cleanup.  note our ref to the
     * fp also protected the fsp. */
    printk("close_elem: fputting %p\n", elem->fp);
    fput(elem->fp);
    
    printk("close_elem: fsp->d_count: %d  fp->f_count: %d fd->fp %p\n",
	   elem->fsp->d_count, elem->fp->f_count, fcheck(elem->fd));
}
