#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * uc_kern.c -- kernel interface to the upcall system.
 */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/signal.h>
#include <sys/signalvar.h>
#include <sys/malloc.h>
#ifdef __FreeBSD__
#include <machine/spl.h>
#endif

#include "kernel/upcall/includes/upcall.h"
#include "upcall.private.h"
#include "uc_upclass.h"
#include "uc_pending.h"
#include "kernel/upcall/includes/uc_hash.h"

static int uc_kern_debug = 0;
static int uc_inited = 0;

#define PROCMAGIC 0x75AEAC78

#define ENTRY  if (uc_kern_debug) printf("Entered %s\n", __FUNCTION__)

/*
 * Prototypes for static inlined function 
 */
static void inline uc_post __P
        ((struct proc         *p,
          struct uc_pending_t *ucpp));

void
uc_init(void)
{
	int               i;
	struct uc_ucld_t *ucldp;

	if (uc_inited) 
	  return;
	
	uc_inited = 1;
	  
	ENTRY;

	for (i=1; i<upcl_tabsz; i++) {
		ucldp = upcl_tab[i];
		if (!ucldp) {
			if (!(ucldp = malloc(sizeof(struct uc_ucld_t),
					     M_TEMP, M_NOWAIT))) {
				panic("uc_init: can't malloc");
			}
			ucldp->ucld_inited = 0;
			upcl_tab[i] = ucldp;
		} else {
			ucldp->ucld_table = (ucldp->ucld_init)(ucldp);
			if (ucldp->ucld_table)
				ucldp->ucld_inited = 1;
			else
				ucldp->ucld_inited = 0;
		}
	}
	uc_hashinit();
}

int
uc_exist_registrations()
{
	struct uc_ucld_t* ucldp;
	int               regs, i;

	ENTRY;

	for (i = 1; i < upcl_tabsz; i++) {
	    ucldp = upcl_tab[i];
	    if (ucldp->ucld_inited) {
		if ((ucldp->ucld_exist)(ucldp, ucldp->ucld_table)) {
		    return 1;
		}
	    }
	}

	return 0;
}

void
uc_procnotify(p, pep, isreg)
	struct proc         *p;
	struct uc_procent_t *pep;
	int                  isreg;
{
	int                  s;
#ifdef DIAGNOSTIC
	struct uc_procent_t *probe;
#endif
	struct uc_info_t *ucip;

	ENTRY;

	/* Only we can fiddle with the list */
	s = splimp();

	ucip = uc_hashlookup(p->p_pid);

#ifdef DIAGNOSTIC
	for (probe = ucip->uci_upcalls.lh_first;
	     probe;
	     probe = probe->ucpe_list.le_next)
	{
		if (probe->ucpe_upclass == pep->ucpe_upclass
		    && probe->ucpe_upcall == pep->ucpe_upcall)
		{
			break;
		}
	}
#endif

	/* 
	 * If we are registering, and this procentry is already known,
	 * or deregistering, and we never knew about it, then we don't
	 * have to do anything.
	 */
	if ((isreg && (pep->ucpe_magic == PROCMAGIC))
	    || (!isreg && (pep->ucpe_magic != PROCMAGIC)))
	{
#ifdef DIAGNOSTIC
		/* Check to see that ucpe_magic was correct. */
		if ((isreg && !probe)
		     || (!isreg && probe)) 
		{
			printf("Magic = 0x%x, probe = %p", pep->ucpe_magic,
			       probe);
			panic("pep: magic number lied");
		}
#endif
		splx(s);
		return;
	}

	if (isreg) {
#ifdef DIAGNOSTIC
		if (probe) 
			panic("pep: registered but mortal");
#endif
		pep->ucpe_magic = PROCMAGIC;
		LIST_INSERT_HEAD(&ucip->uci_upcalls, pep, ucpe_list);
	} else {
#ifdef DIAGNOSTIC
		if (!probe) 
			panic("pep: unregistered but magic");
#endif
		pep->ucpe_magic = 0xbeefdead;
		LIST_REMOVE(pep, ucpe_list);
	}
	splx(s);
}

#if 0
void
uc_sigreturn(p)
	struct proc *p;
{
	if (uc_hashlookup(p->p_pid)->uci_pending.tqh_first) 
		psignal(p, SIGUPCALL);
}

void
uc_fork(p1, p2)
	struct proc *p1;
	struct proc *p2;
{
	struct uc_procent_t  *pep;

	p2->p_spare->p_ucinfo = malloc(sizeof(struct uc_info_t), M_TEMP,
				       M_WAITOK);

#ifdef DIAGNOSTIC
	if (!p2->p_spare->p_ucinfo) panic("uc_fork: malloc failed\n");
#endif

	/* set up pending list of child */
	TAILQ_INIT(&p2->p_pending);

	LIST_INIT(&p2->p_upcalls);
	/* 
	 * Walk through the registered upcalls of p1, registering them for p2
	 */
	for (pep = p1->p_upcalls.lh_first;
	     pep;
	     pep = pep->ucpe_list.le_next)
	{
		if (uc_doreg(p2, pep->ucpe_upclass, pep->ucpe_upcall, 1))
			panic("uc_fork: upcall_doreg failed");
	}
}

#endif /* 0 */

/*
 * Warning: uc_exit() calls this directly.
 */

/* Deregisters all upcalls and deallocates the ucinfo structure -- since
 * the upcall device was closed one exec, the process should now have no
 * upcall-related state at all
 */

void
uc_exec(p)
	struct proc *p;
{
	struct uc_procent_t *cprp;
	struct uc_procent_t *nprp;
	struct uc_pending_t *cpep;
	struct uc_pending_t *npep;
	struct uc_info_t *ucip;
	int                 s;
	int                 freeargs;

	ucip = uc_hashlookup(p->p_pid);
	/* 
	 * Walk down list of registered-for upcalls, deregistering
	 * things.  Beware: the procent's will be freed out from under
	 * us as a side-effect of deregistration 
	 */

	cprp = ucip->uci_upcalls.lh_first;
	while (cprp) {
		nprp = cprp->ucpe_list.le_next;
		if (uc_doreg(p, cprp->ucpe_upclass, cprp->ucpe_upcall, 0))
			panic("uc_exec: upcall_doreg failed");
		cprp = nprp;
	}

	/* 
	 * Walk down list of pending upcalls, toasting them.  If they
	 * are synchronous, be sure to return ETIMEDOUT thorugh the
	 * errno field.
	 */
	
	cpep = ucip->uci_pending.tqh_first;
	while (cpep) {
		npep = cpep->ucp_list.tqe_next;

		/* Dequeue it, and decide whether we should free arguments */
		s = splimp();
		cpep->ucp_shared--;
		freeargs = !(cpep->ucp_shared);
		TAILQ_REMOVE(&uc_hashlookup(p->p_pid)->uci_pending, cpep, ucp_list);
		splx(s);
		
		if (freeargs) {
			if (cpep->ucp_flags & UCP_FREEARGS) {
				free(cpep->ucp_argd->uca_argp,
				     cpep->ucp_argd->uca_mtype);
			}
			free(cpep->ucp_argd, M_TEMP);
		}

		if (cpep->ucp_flags & UCP_SYNC) {
			/* We have to return an error */
			cpep->ucp_errno = ETIMEDOUT;
			wakeup(cpep);
		} else {
			free(cpep, M_TEMP);
		}
		
		cpep = npep;
	}
	uc_hashdelete(ucip);
	free(ucip, M_TEMP);
}

/*
 * On exit, we deregister all previous registrations and deallocate the
 * ucinfo structure.  This is basically what uc_exec does. 
 */
void
uc_exit(p)
    struct proc *p;
{
	uc_exec(p);
}

/* Used by sys___upcall_reg, uc_fork(), and uc_exec() */
int uc_doreg(p, upclass, upcall, isreg)
	struct proc *p;
	upclass_t    upclass;
	upcall_t     upcall;
	int          isreg;
{
	int                    result;
	struct uc_ucld_t      *ucldp;
	struct uc_procent_t   *pep;

	ENTRY;

	if (upclass <= 0 || upclass > upcl_tabsz) {
		return EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		return EINVAL;
	}
	
	result = ((ucldp->ucld_notify)(ucldp, ucldp->ucld_table, p, 
				       upcall, isreg, &pep));

	if (!result && pep) {
		uc_procnotify(p, pep, isreg);
		if (!isreg) {
			if (ucldp->ucld_freepep) {
				(ucldp->ucld_freepep)(ucldp, 
						      ucldp->ucld_table,
						      pep);
			} else {
				free(pep, M_TEMP);
			}
		}
	}

	return result;
}

/* 
 * Common code to post upcall to a process.  Must be called at splimp().
 */
static void inline
uc_post(struct proc         *p,
	struct uc_pending_t *ucpp)
{
        struct uc_info_t    *ucip;
#ifdef DIAGNOSTIC
	struct uc_procent_t *ucpep;

	ENTRY;

	for (ucpep = uc_hashlookup(p->p_pid)->uci_upcalls.lh_first; ucpep;
	     ucpep = ucpep->ucpe_list.le_next)
	{
		if (ucpep->ucpe_upclass == ucpp->ucp_upclass
		    && ucpep->ucpe_upcall == ucpp->ucp_upcall)
		{
			break;
		}
	}
	if (!ucpep)
		panic("uc_post: upcall not registered for");
#endif
	
	ucip = uc_hashlookup(p->p_pid);
	TAILQ_INSERT_TAIL(&ucip->uci_pending, ucpp, ucp_list);
	psignal(p, ucip->uci_sig);
}

int
uc_post_sync(p, upclass, upcall, argsz, argp, argmtype, retszp, retp)
	struct proc *p;
	upclass_t   upclass;
	upcall_t    upcall;
	size_t      argsz;
	void        *argp;
	int         argmtype;
	size_t      *retszp;
	void        *retp;
{
	struct uc_args_t     *ucap;
	struct uc_pending_t  *ucpp;
	int                  flags;
	int                  s;
	int                  result;

	ENTRY;

	flags = (argmtype) ? UCP_FREEARGS : 0;
	flags |= UCP_SYNC;

	ucap = uc_args_create(argsz, argp, argmtype, 1);
	ucpp = uc_pend_create(upclass, upcall, ucap, *retszp, retp, flags);

	/* Atomically enqueue this request, post SIGUPCALL, and sleep */

	s = splimp();
	uc_post(p, ucpp);
	/* 
	 * This sleep is neither interruptible, nor timed.  If you
	 * don't want to blcok forever, use uc_post_async
	 */
	splx(s);
	tsleep(ucpp, PWAIT, "upcall", 0);

	/* We've been awakened.  Grab the answer. */
	result = ucpp->ucp_errno;

	if (result) {
		*retszp = 0;
	} else {
		*retszp = ucpp->ucp_retsz;
	}

	/* 
	 * sys___upcall_ret has already disposed of the argument if it
	 * was supposed to be freed, and has disposed of the ucap.
	 */

	free(ucpp, M_TEMP);
	return result;
}

int
uc_post_async(ucplsz, ucplp, upclass, upcall, argsz, argp, argmtype)
	int ucplsz;
	struct uc_plist_t *ucplp;
	upclass_t upclass;
	upcall_t upcall;
	size_t argsz;
	void *argp;
	int argmtype;
{
	struct uc_args_t     *ucap;
	struct uc_pending_t  *ucpp;
	struct uc_plent_t    *thisp;
	struct uc_plent_t    *nextp;
	int                  flags;
	int                  s;

	ENTRY;

	flags = (argmtype) ? UCP_FREEARGS : 0;
	ucap = uc_args_create(argsz, argp, argmtype, ucplsz);
	
	/* 
	 * We are going to turn interrupts off and back on inside the
	 * iteration across the process list.  This is so we don't
	 * have to worry about cases of sending an upcall to the
	 * entire universe taking up more time than we'd like.  The
	 * uc_pend_create also needs to be outside of the critical
	 * region, since it might block.
	 */
	
	/* XXX - need to free as we go */
	thisp = ucplp->lh_first;

	while (thisp) {	
		nextp = thisp->ucpl_list.le_next;
		ucpp = uc_pend_create(upclass, upcall, ucap, 0, NULL, flags);

		s = splimp();
		uc_post(thisp->ucpl_p, ucpp);
		splx(s);
		
		LIST_REMOVE(thisp, ucpl_list);
		free(thisp, M_TEMP);
		thisp = nextp;
#ifdef DIAGNOSTIC
		ucplsz--;
#endif
	}
#ifdef DIAGNOSTIC
	if (ucplsz) 
		panic("uc_post_async: ucplsz lied");
#endif
	return 0;
}

