#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * uc_pending.h
 *
 * Data structures/constants associated with pending upcalls
 */

#ifndef _UC_PENDING_H_
#define _UC_PENDING_H_

/*
 * Prototypes for functions declared in this file
 */
static inline struct uc_args_t * uc_args_create __P
                          ((size_t argsz,
                            void  *argp,
                            int    mtype,
                            int    shared));

static inline struct uc_pending_t * uc_pend_create __P
                          ((upclass_t         upclass,
                            upcall_t          upcall,
                            struct uc_args_t *argdp,
                            size_t            retsz,
                            void             *retp,
                            int               flags));

/*
 * An argument descriptor.  One of these is shared by all pending
 * upcalls from a single invocation.  The uca_shared field is set to
 * the number of processes we expect to use the arguments, and the
 * argument descriptor (and optionally, the argument buffer itself),
 * is only freed after the last such process exits or fails.
 */
struct uc_args_t {
	int                         uca_mtype;
	size_t                      uca_argsz;
	int                         uca_shared; /* How many procs need args */
	void                       *uca_argp;
};	

/*
 * Processes waiting on the result wait on the address of the
 * uc_pending_t structure added to the called process.
 */
struct uc_pending_t {
	TAILQ_ENTRY(uc_pending_t)   ucp_list;
	upclass_t                   ucp_upclass;
	upcall_t                    ucp_upcall;
	int                         ucp_flags;
	struct uc_args_t           *ucp_argd;
	int                         ucp_retsz;  
	void                       *ucp_retp;
	int                         ucp_errno;
	int                         ucp_serial;
};

#define ucp_mtype  ucp_argd->uca_mtype
#define ucp_argsz  ucp_argd->uca_argsz
#define ucp_shared ucp_argd->uca_shared
#define ucp_argp   ucp_argd->uca_argp

/* 
 * Flags
 * 
 * If a pending upcall entry is UCP_SYNC, then there is someone
 * waiting for the answer to this upcall.
 * 
 * If UCP_FREEARGS is set, we are supposed to free the actual argument
 * buffer after we last use it.
 *
 * We set UCP_INPROGRESS if this upcall is currently being serviced by
 * the process it is destined for.  
 */

#define UCP_SYNC       0x1  /* Otherwise async. */
#define UCP_FREEARGS   0x2  /* Free argument buffer when we are done. */
#define UCP_INPROGRESS 0x4  /* This pending call is in progress */

/* 
 * The structure in a struct proc that contains needed upcall info. 
 */
struct uc_info_t {
    LIST_HEAD(, uc_procent_t)   uci_upcalls;
    TAILQ_HEAD(, uc_pending_t)  uci_pending;
    LIST_ENTRY(uc_info_t) uc_hash;  /* hash chain on pid */
    pid_t uc_pid;   /* process that owns this structure */
    unsigned uci_sig;   /* signal to use when notifying process */
};

static inline struct uc_args_t *
uc_args_create(size_t argsz,
	       void  *argp,
	       int    mtype,
	       int    shared) 
{
	struct uc_args_t *result;

	result = malloc(sizeof(struct uc_args_t), M_UPCALL, M_WAITOK);

#ifdef DIAGNOSTIC
	if (!result)
		panic("uc_args_create: malloc failed");
	if (argsz && !argp)
		panic("uc_args_create: argp bogus");
	if (shared <= 0)
		panic("uc_args_create: no one needs argp");
	if (mtype < M_FREE || mtype >= M_LAST)
		panic("uc_args_create: bad mtype");
#endif
	result->uca_mtype = mtype;
	result->uca_argsz = argsz;
	result->uca_shared = shared;
	result->uca_argp = argp;
	return result;
}

static inline struct uc_pending_t *
uc_pend_create(upclass_t         upclass,
	       upcall_t          upcall,
	       struct uc_args_t *argdp,
	       size_t            retsz,
	       void             *retp,
	       int               flags)
{
	struct uc_pending_t *result;

	result = malloc(sizeof(struct uc_pending_t), M_UPCALL, M_WAITOK);

#ifdef DIAGNOSTIC
	if (!result)
		panic("uc_pend_create: malloc failed");
	if (!argdp)
		panic("uc_pend_create: bad argdp");
	if ((flags & UCP_SYNC) && retsz && !retp) 
		panic("uc_pend_create: sync entry, no retp");
	if ((flags & UCP_FREEARGS) && !argdp->uca_mtype)
		panic("uc_pend_create: mtype doesn't match flags");
	if (flags & UCP_INPROGRESS) 
		panic("uc_pend_create: can't create inprogress");
#endif
	result->ucp_upclass = upclass;
	result->ucp_upcall = upcall;
	result->ucp_flags = flags;
	result->ucp_argd = argdp;
	result->ucp_retsz = retsz;
	result->ucp_retp = retp;
	result->ucp_errno = 0;
	result->ucp_serial = 0;
	
	return result;
}

#endif /* _UC_PENDING_H_ */
