#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/proc.h>
#include <sys/mount.h>
#include <sys/malloc.h>
#include <sys/signal.h>
#ifdef __FreeBSD__
#include <machine/spl.h>
#endif

#include "kernel/upcall/includes/upcall.h"
#include "upcall.private.h"
#include "uc_upclass.h"
#include "uc_pending.h"
#include "uc_ioctl.h"
#include "kernel/upcall/includes/uc_hash.h"

static int uc_sys_debug = 0;
static int uc_serial = 1;

#define ENTRY  if (uc_sys_debug) printf("Entered %s\n", __FUNCTION__)

int
uci_sync(p, uap, retval)
	struct proc *p;
	struct uci_sync_args *uap;
	int *retval;
{
	upclass_t            upclass;
	upcall_t             upcall;
	pid_t                pid;
	struct uc_ucld_t     *ucldp;
	int                  plsz;
	struct uc_plist_t    *plp=0;
	struct uc_plent_t    *thisp;
	struct uc_plent_t    *nextp;
	struct proc          *targp;
	size_t               argsz;
	void                 *argp=0;
	size_t               retsz;
#ifdef DIAGNOSTIC
	size_t               origret;
#endif
	void                 *retp=0;
	int                  result=0;
	
	ENTRY;

	/* Step 1: get the process for which this is intended */
	upclass = uap->upclass;
	upcall = uap->upcall;
	pid = uap->pid;

	if (pid <= 0) {
		return EINVAL;
	}

	if (upclass <= 0 || upclass > upcl_tabsz) {
		return EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		return EINVAL;
	}

	
	/* 
	 * Returns a negative number on error, the negation of which
	 * is the errno code. 
	 */
	plsz = (ucldp->ucld_rcpt)(ucldp, ucldp->ucld_table, pid, upcall, &plp);
	if (plsz < 0) 
		return -plsz;

	/* From this point on, we will need to free things */
	if (plsz != 1) {
		result = ESRCH;
		goto exit;
	}

	targp = plp->lh_first->ucpl_p;

	/* Get argument buffer, allocate return buffer */
	argsz = uap->argsz;
	if (argsz) {
		argp = malloc(argsz, M_TEMP, M_WAITOK);
		result = copyin(uap->argp, argp, argsz);
		if (result) goto exit;
	}

	result = copyin(uap->retszp, &retsz, sizeof(retsz));
	if (result) goto exit;

#ifdef DIAGNOSTIC
	origret = retsz;
#endif
	if (retsz) {
		retp = malloc(retsz, M_TEMP, M_WAITOK);
	}

	/* Post the call */
	result = uc_post_sync(targp, upclass, upcall, argsz, argp, 0,
			      &retsz, retp);
	if (result) goto exit;

#ifdef DIAGNOSTIC
	if (retsz > origret) 
	    panic("uci_sync: return size exceeded");
#endif

	/* send out the results */
	result = copyout(retp, uap->retp, retsz);
	if (!result)
	    if (suword(uap->retszp, retsz) < 0)
		    result = EFAULT;

 exit:
	/* Free the argument & return buffers */
	if (argp) free(argp, M_TEMP);
	if (retp) free(retp, M_TEMP);

	/* Free the plist */
	if (plp) {
		thisp = plp->lh_first;
		while (thisp) {
			nextp = thisp->ucpl_list.le_next;
			free(thisp, M_TEMP);
			thisp = nextp;
		}
	}

	return result;
}


int
uci_async(p, uap, retval)
	struct proc *p;
	struct uci_async_args *uap;
	int *retval;
{
	upclass_t             upclass;
	upcall_t              upcall;
	pid_t                 pid;
	struct uc_ucld_t      *ucldp;
	int                   plsz;
	struct uc_plist_t     *plp=0;
	size_t                argsz;
	void                  *argp=0;
	int                   result=0;

	ENTRY;

	upclass = uap->upclass;
	upcall = uap->upcall;
	pid = uap->pid;

	if (upclass <= 0 || upclass > upcl_tabsz) {
	    return EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		return EINVAL;
	}

	/* 
	 * Returns a negative number on error, the negation of which
	 * is the errno code. 
	 */
	plsz = (ucldp->ucld_rcpt)(ucldp, ucldp->ucld_table, pid, upcall, &plp);
	if (plsz < 0) 
		return -plsz;

	if (plsz == 0)
		return ESRCH;

	/* Get argument buffer, allocate return buffer */
	argsz = uap->argsz;
	if (argsz) {
		argp = malloc(argsz, M_TEMP, M_WAITOK);
		result = copyin(uap->argp, argp, argsz);
		if (result) return result;
	}

	result = uc_post_async(plsz, plp, upclass, upcall, argsz, 
			       argp, M_TEMP);

	return result;
}

int
uci_reg(p, uap, retval)
	struct proc *p;
	struct uci_reg_args *uap;
	int *retval;
{
	upclass_t              upcls_arg;
	upcall_t               upca_arg;
	int                    flags_arg;
	int                    isreg;
	int                    result=0;
	
	ENTRY;

	upcls_arg = uap->upclass;
	upca_arg = uap->upcall;
	flags_arg = uap->flags;

	if (flags_arg & UPCALL_REGISTER)
		isreg = 1;
	else 
		isreg = 0;

	result = uc_doreg(p, upcls_arg, upca_arg, isreg);

	return result;
}

#define SUWORDCHK(addr, val, result, stmt) \
do {                                       \
      (result) = suword((addr),(val));   \
      if (result) {                      \
                result = EFAULT;           \
              stmt                       \
      }                                  \
} while (0)

int
uci_next(p, uap, retval)
	struct proc *p;
	struct uci_next_args *uap;
	int *retval;
{
	size_t               argsz;
	int                  s;
	struct uc_pending_t  *ucpp;
	int                  result=0;

	ENTRY;

	/* 
	 * Grab a pointer to the first entry.  Lock out inserting
	 * processes while we are at it.  Check that no one else is
	 * using it, and if not, mark it in progress.  
	 *
	 * Now, we are the only process that cares about anything in
	 * this entry other than the links entry, which we won't touch
	 * until we dequeue in __upcall_ret
	 */
	s = splimp();
	ucpp = uc_hashlookup(p->p_pid)->uci_pending.tqh_first;
	if (!ucpp) {
		result = ENOENT;
	} else if (ucpp->ucp_flags & UCP_INPROGRESS) {
		result = EALREADY;
	} else {
		ucpp->ucp_serial = uc_serial++;
		ucpp->ucp_flags |= UCP_INPROGRESS; /* lock out */
	}
	splx(s);

	if (result) goto exit;

	argsz = fuword(uap->argszp);

       /* Any of these could cause a jump to exit if they fail */
	SUWORDCHK(uap->upclassp, ucpp->ucp_upclass, result, goto exit;);
	SUWORDCHK(uap->upcallp, ucpp->ucp_upcall, result, goto exit;);
	SUWORDCHK(uap->argszp, ucpp->ucp_argsz, result, goto exit;);
	SUWORDCHK(uap->retszp, ucpp->ucp_retsz, result, goto exit;);
	SUWORDCHK(uap->serialp, ucpp->ucp_serial, result, goto exit;);
	
	/* argsz is valid, since we've stored to it. */
	if (argsz >= ucpp->ucp_argsz) {
		/* We can do the copyout; it's big enough. */
		result = copyout(ucpp->ucp_argp, uap->argp,
				 ucpp->ucp_argsz);
		if (result) goto exit;
	}
	
	/* We're done */
 exit:	
	if (result) {
		/* Store zero's in upcall/upclass */
		(void)suword(uap->upclassp,0);
		(void)suword(uap->upcallp,0);
		/* Clear the in-progress flag if necessary */
		s = splimp();
		if (ucpp) {
			ucpp->ucp_flags &= ~UCP_INPROGRESS;
			ucpp->ucp_serial = 0;
		}
		splx(s);
		/* If we were ENOENT, then just return 0 */
		if (result == ENOENT) result = 0;
	}
	return result;
}

int
uci_args(p, uap, retval)
	struct proc *p;
	struct uci_args_args *uap;
	int *retval;
{
	struct uc_pending_t *ucpp;
	int                 s;
	int                 result=0;

	ENTRY;

	s = splimp();
	ucpp = uc_hashlookup(p->p_pid)->uci_pending.tqh_first;
	if (!ucpp 
	    || !(ucpp->ucp_flags & UCP_INPROGRESS) 
	    || (ucpp->ucp_serial != uap->serial))
	{
		result = ENOENT;
	}
	splx(s);
	if (result) return result;

	if (ucpp->ucp_argsz > uap->bufsz) {
		return EINVAL;
	}

	/* Copy the arguments out */
	result = copyout(ucpp->ucp_argp, uap->buf, ucpp->ucp_argsz);
	return result;
}

int
uci_ret(p, uap, retval)
	struct proc *p;
	struct uci_ret_args *uap;
	int *retval;
{
	struct uc_pending_t *ucpp;
	int                 s;
	int                 result=0;
	size_t              retsz;
	int                 freeargs;
	int                 err;

	ENTRY;

	/* 
	 * Grab the pending entry; we can't remove it until after the
	 * copyout if we owe one; the copyout has to complete before
	 * dequeuing 
	 */
	s = splimp();
	ucpp = uc_hashlookup(p->p_pid)->uci_pending.tqh_first;
	if (!ucpp
	    || !(ucpp->ucp_flags & UCP_INPROGRESS)
	    || (ucpp->ucp_serial != uap->serial))
	{
		result = ENOENT;
	}
	splx(s);
	
	if (result) return result;

	if (ucpp->ucp_flags & UCP_SYNC) {
		err = uap->errno;
		if (err) {
			ucpp->ucp_errno = err;
		} else {
			retsz = uap->retsz;
			if (retsz > ucpp->ucp_retsz)
				return E2BIG;
			result = copyin(uap->retp, ucpp->ucp_retp, 
					retsz);

			if (result) return result;
			ucpp->ucp_errno = 0;
			ucpp->ucp_retsz = retsz;
		}
	}

	/* Can now remove the entry and dispose of it */
	s = splimp();
	TAILQ_REMOVE(&uc_hashlookup(p->p_pid)->uci_pending, ucpp, ucp_list);
	ucpp->ucp_shared--;
	freeargs = !(ucpp->ucp_shared);
	splx(s);

	if (freeargs) {
		if (ucpp->ucp_flags & UCP_FREEARGS) {
			free(ucpp->ucp_argd->uca_argp, 
			     ucpp->ucp_argd->uca_mtype);
		}
		free(ucpp->ucp_argd, M_TEMP);
	}

	if (ucpp->ucp_flags & UCP_SYNC)
		wakeup(ucpp);
	else
		free(ucpp, M_TEMP);
	return 0;
}

int
uci_set_sig(p, uap, retval)
        struct proc *p;
	struct uci_set_sig_args *uap;
	int *retval;
{
	struct uc_info_t *ucip;
	int               result=0;
	int               s;

	ENTRY;

	/* Is this a valid signal? */
	if (uap->sig >= NSIG) {
	    printf("uci_set_sig: %d >= %d\n", uap->sig, NSIG);
	    return EINVAL;
	}

	/* Set upcall signal for this process */
	s = splimp(); 
	ucip = uc_hashlookup(p->p_pid);
	ucip->uci_sig = uap->sig;
	splx(s);

	return result;
}


