#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * upclass.h -- how we keep track of classes.
 */

#ifndef _UPCLASS_H_
#define _UPCLASS_H_

/*
 * Entry points a class has to provide:
 *
 * An "init" function to call to initialize the class
 * A "notify" function to call on a register/deregister request.
 * A "rcpt" function to get a list of recipients for a call/pid list.
 *
 * The init call sets up uc_table, and is called at boot time.
 * 
 * Notify is responsible for adding/removing the pid from the table of
 * known acceptees for the given upcall.  This is managed per-upclass,
 * since different upclasses may want to optimize the structure
 * differently depending on their access patterns.  The upclass may
 * optionally check the validity of an upcall number to see if it is a
 * known upcall.  We provide default implementations of uc_table
 * management, which can be used either directly or by a thin veneer
 * which checks the upcall number and passes the request through.
 *
 * Finally, rcpt is responsible for returning a list of proc *'s given
 * a particular pid_t.  Again, we provide a default implementation for
 * this function.  
 *
 * The last two take, as their first two arguments, both the
 * uc_upcall_t entry that they are being called for, as well as an
 * explicit pointer to the opaque table.  This is to allow veneer
 * procedures to lay atop the default implementations and add to the
 * opaque structure used by such implementations.  Not for the fait of
 * heart, though.  
 */

/* Forward def: upclass descriptor */
struct uc_ucld_t;

/*
 * initialization
 *
 * Called at boot time.  Returns a pointer to an initialized opaque
 * table on success, zero on failure.  Is not expected to modify any
 * of the bits in the uc_ucld_t.
 */
typedef void * (*uc_init_t)  __P((struct uc_ucld_t *));

/*
 * existance
 * 
 * Called when the module is unloaded to determine if there exist any 
 * oustanding registrations for an upcall of this upclass.  If the function
 * returns a non-zero value, the module will not be allowed to unload.
 */
typedef int (*uc_exist_t) __P((struct uc_ucld_t *uclp, /* class */
                               void* tabp));           /* table */

/*
 * notification
 *
 * Called whenever a handler is installed/removed on process p.  This
 * includes any instances where a handler is changed, though it
 * appears just as a regular install.  Returns zero on success.
 * Should return ENOENT if asked to deregister an unregistered
 * handler.  The argument "reg" is nonzero if this is a registration,
 * zero if it is a deregistration.  If this is a new registration, or
 * a deregistration of a previous registration, the notification
 * routine is obligated to return a pointer to a uc_procent_t via the
 * final argument, peh.  Otherwise it may return NULL, or a valid
 * poniter-to-uc_procent_t; the upcall infrastructure will take care
 * of not re-inserting or re-removing it in the process' individual
 * list
 *
 * The default implementations guarantee not to play with the
 * uc_procent_t's data field; it extends the structure instead.  This
 * is done so that upclasses who layer on top of the default
 * implementation can store away some data.
 */
typedef int (*uc_notify_t) __P((struct uc_ucld_t *uclp,      /* class */
				void *tabp,                  /* table */
				struct proc *p,              /* process */
				upcall_t upcall,             /* call */
				int reg,                     /* r/d */
				struct uc_procent_t **peh)); /* see above */

/*
 * Since uc_notify has to return a pointer-to-uc_procent_t, (and it
 * may well have extended the structure), it should have a chance to
 * free it, as well.  This fuction can be NULL, in which case the
 * caller will just free it with a tag of M_UPCALL.
 */

typedef void (*uc_freepep_t) __P((struct uc_ucld_t *uclp,
				  void *tabp,
				  struct uc_procent_t *pep));

/* 
 * For upclasses that layer atop the default implementation, a way to
 * get at uc_procent_t's for specific upcall, proc pairs.  Also useful
 * for upclasses that may need to change the state stored in the
 * uc_procent_t's data field over time.  Returns NULL if it can't find
 * a uc_procent_t.  
 */

typedef struct uc_procent_t * (*uc_getpep_t) __P((struct uc_ucld_t *uclp,
						  void *tabp,
						  struct proc *p,
						  upcall_t upcall));

/*
 * rcpt
 *
 * Given a pid, creates a list of all processes that match the pid
 * criteria and have registered for the given upcall.  Positive pids
 * match only the exact same process, negative pids match any process
 * in the process group numbered |pid|, zero matches all registered
 * processes.  Returns the number of such processes found, or -(error
 * code) if error.  The caller is responsible for deallocating the list
 * returned in the final argument.  */
typedef int (*uc_rcpt_t) __P((struct uc_ucld_t *ucldp,   /* uc_ucld_t */
			      void *tabp,                /* opaque table */
			      pid_t pid,                 /* pid to look for */
			      upcall_t upcall,           /* Which upcall */
			      struct uc_plist_t **plh)); /* process list*/

/*
 * Print out the entries in an upclass table.  Used only for debugging
 */
typedef void (*uc_print_t) __P((struct uc_ucld_t *ucldp,
				void *tabp));

/* Upclass descriptors */
struct uc_ucld_t {
	upclass_t      ucld_upclass; /* class number */
	int            ucld_inited;  /* !0 <-> inited; only change at boot */
	char          *ucld_name;    /* class name */
	uc_init_t      ucld_init;    /* Set up lookup table */
        uc_exist_t     ucld_exist;   /* Are there any registrations? */
	uc_notify_t    ucld_notify;  /* notfiy class: intent to (not) handle */
	uc_freepep_t   ucld_freepep; /* Free notify-returned uc_procent_t */
	uc_getpep_t    ucld_getpep;  /* Get access to a particular pep */
	uc_rcpt_t      ucld_rcpt;    /* Check to see if call,proc pair valid */
	uc_print_t     ucld_print;   /* Print out table for debugging */
	void          *ucld_table;   /* Opaque struct used by class */
};

extern struct uc_ucld_t   *upcl_tab[]; /* Table of known upclasses */
extern size_t              upcl_tabsz; /* Number of entries in upcl_tab */

/*
 * Default implementations for upclass functions.  Implemented in
 * uc_upclass.c
 */

extern void *uc_init_default __P((struct uc_ucld_t *));

extern int uc_exist_default __P((struct uc_ucld_t *, void *));

extern int uc_notify_default __P((struct uc_ucld_t *, void *, 
				  struct proc *, upcall_t, int,
				  struct uc_procent_t **));

extern struct uc_procent_t *uc_getpep_default __P((struct uc_ucld_t *, 
						   void *, struct proc *,
						   upcall_t));

extern int uc_rcpt_default   __P((struct uc_ucld_t *, void *, pid_t, 
				  upcall_t, struct uc_plist_t **));
extern void uc_print_default  __P((struct uc_ucld_t *, void *));

/*
 * in-kernel asynchronous upcalls 
 */
extern int upcall_async  __P((upclass_t, upcall_t, pid_t, size_t, void *,
			      int));

#endif /* _UPCLASS_H_ */
