#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <linux/param.h>
#include <linux/time.h>
#include <linux/malloc.h>
#include <linux/sem.h>
#include <linux/kernel.h>

#include "ody_queue.h"
#include "kernel/upcall/includes/upcall.h"
#include "upcall.private.h"
#include "uc_pending.h"
#include "uc_upclass.h"
#include "uc_linux.h"

/*
 * uc_default.c
 *
 * Default implementation for the uc_ucld_t descriptor table
 */

/*
 * The default implementation is a hash table keyed on upcall number.
 * Each bucket in this table is a list (chain) of
 *
 * struct uc_callent_t
 *
 * For each upcall that is known, we keep a small list of the
 * processes that have requested that upcall.  We don't bother hashing
 * them; I expect there to be ~10 processes at most for each upcall
 * posted.  (After all, there are only ~512 processes in the system.)
 * We'll add some stats, though and check high-water marks/averages.
 */

/* 
 * The entries in the list of processes for each upcall.  They all
 * have a uc_procent_t header prepended to them, so that they can also
 * be kept on the per-process fork/exec notification list.
 */
struct uc_defprocent_t {
	/* this must be the first entry in this structure */
	struct uc_procent_t              udpe_procent;
	ODY_LIST_ENTRY(uc_defprocent_t)  udpe_list;
	struct task_struct              *udpe_t;
};

#define udpe_magic       udpe_procent.ucpe_magic
#define udpe_upclass     udpe_procent.ucpe_upclass
#define udpe_upcall      udpe_procent.ucpe_upcall

ODY_LIST_HEAD(uc_defproclist_t, uc_defprocent_t);

/* Entries in each hash chain */

struct uc_callent_t {
	upcall_t                      ucae_upcall;
	ODY_LIST_ENTRY(uc_callent_t)  ucae_list;
	unsigned int                  ucae_numprocs;
	struct uc_defproclist_t       ucae_plist;
	unsigned int                  ucae_lookups;
	unsigned int                  ucae_worstcase;
	struct semaphore              ucae_sem; /* Protect ucae_plist */
};

/* Head of a hash chain */
ODY_LIST_HEAD(uc_call_list_t, uc_callent_t);

/* 
 * The hash table is small; we expect about 128 or fewer calls per
 * class at any one time, so to give us an expected probe of 3, pick a
 * prime somewhere between 32 and 64.  It's only got the size of the
 * table (I may want to change it while debugging) and the table
 * itself.  
 */

static const int uc_calltabsz = 47;

struct uc_calltab_t {
	int                    cata_size;
	struct uc_call_list_t *cata_tab;
	struct semaphore       cata_sem; /* protect cata_tab */
};


/*
 * Now, for some helper functions.  These should probably be inline,
 * since they are itty bitty.  They all raise to splimp() since
 * multiple processes may be trying to register/invoke, etc.  
 */

/* 
 * Looks for a given uc_callent_t in a uc_calltab_t.  Returns a
 * pointer to it if found, NULL otherwise.
 */
static inline struct uc_callent_t *
calltab_find(struct uc_calltab_t *catap, upcall_t uc)
{
	struct uc_callent_t  *ucaep;
	int                   bucket;

	ENTRY;

	bucket = uc % catap->cata_size;

	down (&catap->cata_sem);
	for (ucaep = catap->cata_tab[bucket].lh_first;
	     ucaep;
	     ucaep = ucaep->ucae_list.le_next)
	{
		if (ucaep->ucae_upcall == uc) {
			up (&catap->cata_sem);
			return ucaep;
		}
	}
	up (&catap->cata_sem);
	return NULL;
}

/*
 * Insert a uc_callent_t into a uc_calltab_t.  Precondition: no entry
 * for that upcall already exists in the calltab.  DIAGNOSTIC kernels
 * check this to make sure, and panic if the invariant is broken.
 */
static inline void
calltab_insert(struct uc_calltab_t *catap, struct uc_callent_t *ucaep)
{
#ifdef DIAGNOSTIC
	struct uc_callent_t *probe;
#endif
	int             bucket;

	ENTRY;

	bucket = ucaep->ucae_upcall % catap->cata_size;

	down (&catap->cata_sem);
#ifdef DIAGNOSTIC
	for (probe = catap->cata_tab[bucket].lh_first;
	     probe;
	     probe = probe->ucae_list.le_next) 
	{
		if (probe->ucae_upcall == ucaep->ucae_upcall)
			panic("calltab_insert: upcall already present.");
	}
#endif
	ODY_LIST_INSERT_HEAD(&catap->cata_tab[bucket], ucaep, ucae_list);
	up (&catap->cata_sem);
}

/*
 * Remove a uc_callent_t from a uc_calltab_t.  Fast path just uses
 * LIST_REMOVE, but DIAGNOSTIC kernels check to see if (a) it  is
 * actually in the uc_calltab_t and (b) that it is currently empty.
 */
static inline void
calltab_remove(struct uc_calltab_t *catap, struct uc_callent_t *ucaep)
{
	ENTRY;
	
	down (&catap->cata_sem);
#ifdef DIAGNOSTIC
	if (calltab_find(catap, ucaep->ucae_upcall) != ucaep) 
		panic("calltab_remove: not in calltab");
	if (ucaep->ucae_numprocs != 0) 
		panic("calltab_remove: callent not empty");
#endif
	ODY_LIST_REMOVE(ucaep, ucae_list);
	up (&catap->cata_sem);
}



/*
 * callent_create
 *
 * Given an upcall number, create a callent for it.  Can block; be
 * careful about calling from an interrupt routine.  Won't return
 * NULL, will panic (or trap, on a non-diagnostic kernel) if the
 * malloc fails.
 */

static inline struct uc_callent_t *
callent_create(upcall_t uc)
{
	struct uc_callent_t *ucaep;

	ENTRY;

	ucaep = kmalloc(sizeof(struct uc_callent_t), GFP_KERNEL);

#ifdef DIAGNOSTIC
	if (!ucaep)
		panic("callent_create: kmalloc failed");
#endif

	memset (ucaep, 0, sizeof(struct uc_callent_t)); 
	sema_init (&ucaep->ucae_sem, 1);
	ucaep->ucae_upcall = uc;
	ODY_LIST_INIT(&ucaep->ucae_plist);
	return (ucaep);
}

/*
 * Looks for a given uc_defprocent_t in a uc_callent_t.  Returns a pointer
 * to it if found, NULL otherwise.
 */
 
static inline struct uc_defprocent_t *
callent_find(struct uc_callent_t *ucaep, struct task_struct* t)
{
	struct uc_defprocent_t *udpep;

	ENTRY;

	down (&ucaep->ucae_sem);
	for (udpep = ucaep->ucae_plist.lh_first;
	     udpep;
	     udpep = udpep->udpe_list.le_next)
	{
		if (udpep->udpe_t == t) {
			up (&ucaep->ucae_sem);
			return udpep;
		}
	}
	
	up (&ucaep->ucae_sem);
	return NULL;
}

/*
 * Insert a uc_defprocent_t into a uc_callent_t.  Precondition: no entry
 * for that process already exists in the callent.  DIAGNOSTIC kernels
 * check this to make sure, and panic if the invariant doesn't hold.  
 */
static inline void
callent_insert(struct uc_callent_t *ucaep, struct uc_defprocent_t *udpep)
{
#ifdef DIAGNOSTIC
	struct uc_defprocent_t *probe;
#endif

	ENTRY;

	down (&ucaep->ucae_sem);
#ifdef DIAGNOSTIC
	for (probe = ucaep->ucae_plist.lh_first;
	     probe;
	     probe = probe->udpe_list.le_next)
	{
		if (probe->udpe_p == udpep->udpe_p)
			panic("callent_insert: process already present");
	}
#endif
	ucaep->ucae_numprocs++;
	ODY_LIST_INSERT_HEAD(&ucaep->ucae_plist, udpep, udpe_list);	
	up (&ucaep->ucae_sem);
}

/*
 * Remove a uc_defprocent_t from a uc_callent_t.  Fast path just uses
 * LIST_REMOVE, but DIAGNOSTIC kernels check to see if the process is
 * actually in the uc_callent_t.  
 */
static inline void
callent_remove(struct uc_callent_t *ucaep, struct uc_defprocent_t *udpep)
{
	ENTRY;

	down (&ucaep->ucae_sem);
#ifdef DIAGNOSTIC
	if (callent_find(ucaep, udpep->udpe_p) != udpep)
		panic("callent_remove: not in callent");
#endif
	ucaep->ucae_numprocs--;
	ODY_LIST_REMOVE(udpep, udpe_list);
	up (&ucaep->ucae_sem);
}

/*
 * defprocent_create
 *
 * Given a process, create a plent for it.  Can block; be careful
 * about calling from an interrupt routine.  Won't return NULL, will
 * panic (or trap, on a non-diagnostic kernel) if the malloc fails.  
 */
static inline struct uc_defprocent_t *
defprocent_create(upclass_t uc, upcall_t ua, struct task_struct* t)
{
	struct uc_defprocent_t *udpep;

	ENTRY;

	udpep = kmalloc(sizeof(struct uc_defprocent_t), GFP_KERNEL);
	
#ifdef DIAGNOSTIC
	if (!udpep)
		panic("plent_create: kmalloc failed");
#endif
	
	memset (udpep, 0, sizeof(struct uc_defprocent_t));
	udpep->udpe_upclass = uc;
	udpep->udpe_upcall = ua;
       	udpep->udpe_magic = 0xbeefdead;  /* Not allocated */
	udpep->udpe_t = t;
	return udpep;
}

/* 
 * Create a uc_plent_t 
 */
static inline struct uc_plent_t *
plent_create(struct task_struct *t)
{
	struct uc_plent_t *plentp;

	ENTRY;

	plentp = kmalloc(sizeof(struct uc_plent_t), GFP_KERNEL);

#ifdef DIAGNOSTIC
	if (!plentp)
		panic("plent_create: kmalloc failed");
#endif
	memset (plentp, 0, sizeof(struct uc_plent_t));
	plentp->ucpl_t = t;
	return plentp;
}


/*
 * Check a pid, process pair to see if t is named by pid.
 * If pid > 0, pid's must match.
 * If pid < 0, pgroup must be -pid
 * otherwise true
 */

static inline int
pid_compare(pid_t pid, struct task_struct *t)
{
	ENTRY;

	if (!pid) return 1;
	if (pid < 0) {
		return (-pid == t->gid);
	}
	return (pid == t->pid);
}

   

/*
 * default initialization routine
 */
void *
uc_init_default(upcp)
	struct uc_ucld_t *upcp;
{
	struct uc_calltab_t *catap;
	int                  i;

	ENTRY;

	catap = kmalloc(sizeof(struct uc_calltab_t), GFP_KERNEL);

	if (!catap) {
		return NULL;
	}

	sema_init (&catap->cata_sem, 1);
	catap->cata_size = uc_calltabsz;
	catap->cata_tab = kmalloc(sizeof(struct uc_call_list_t) * uc_calltabsz,
				  GFP_KERNEL);

	if (!catap->cata_tab) {
		kfree(catap);
		return NULL;
	}

	for (i = 0; i < uc_calltabsz; i++) {
		ODY_LIST_INIT(&catap->cata_tab[i]);
	}

	return catap;
}

/* 
 * Default notificatin routine
 */

int
uc_notify_default(ucldp, tabp, t, upcall, isreg, peh)
	struct uc_ucld_t     *ucldp;
	void                 *tabp;
	struct task_struct   *t;
	upcall_t              upcall;
	int                   isreg;
	struct uc_procent_t **peh;
{
	struct uc_calltab_t    *catap;
	struct uc_callent_t    *ucaep;
	struct uc_defprocent_t *udpep;

	ENTRY;

	*peh = NULL;

	if (!ucldp->ucld_inited) 
		return -ENXIO;

	catap = (struct uc_calltab_t *)tabp;
	ucaep = calltab_find(catap, upcall);
	if (ucaep)
		udpep = callent_find(ucaep, t);
	else
		udpep = NULL;

	if (isreg) {
		/* 
		 * If we don't know about this upcall, create an entry
		 * for it and insert it into the calltab
		 */
		if (!ucaep) {
			ucaep = callent_create(upcall);
			calltab_insert(catap, ucaep);
		}
		/* 
		 * If this process wasn't in that upcall's list,
		 * create it and insert it.
		 */
		if (!udpep) {
			udpep = defprocent_create(ucldp->ucld_upclass,
						  upcall, t);
			callent_insert(ucaep, udpep);
		}
	} else {
		/* If we had an entry for this process, remove it. */
		if (udpep) {
			callent_remove(ucaep, udpep);
		}
		/* 
		 * If that was the last process in this callent,
		 * remove and free it.  XXX slight race condition on
		 * numprocs
		 */
		if (!ucaep->ucae_numprocs) {
			calltab_remove(catap, ucaep);
			kfree(ucaep);
		}
	}

	*peh = (struct uc_procent_t *)udpep;
#ifdef DIAGNOSTIC
	if (*peh != &udpep->udpe_procent)
		panic("uc_notify_default: first field of udpe not pe");
#endif
	return 0;
}

struct uc_procent_t *
uc_getpep_default(ucldp, tabp, t, uc)
	struct uc_ucld_t   *ucldp;
	void               *tabp;
	struct task_struct *t;
	upcall_t            uc;
{
	struct uc_calltab_t    *catap;
	struct uc_callent_t    *ucaep;
	struct uc_defprocent_t *udpep;
	
	ENTRY;

	if (!ucldp->ucld_inited) {
		return NULL;
	}

	catap = (struct uc_calltab_t *)tabp;
	ucaep = calltab_find(catap, uc);

	if (!ucaep)
		return NULL;

	down (&ucaep->ucae_sem); /* JNF - inserted ups - must be good ??? */
	for (udpep = ucaep->ucae_plist.lh_first;
	     udpep;
	     udpep = udpep->udpe_list.le_next)
	{
		if (t == udpep->udpe_t) {
			up (&ucaep->ucae_sem); 
			return (struct uc_procent_t *)udpep;
		}
	}
	up (&ucaep->ucae_sem); 
	return NULL;
}


int
uc_rcpt_default(ucldp, tabp, pid, upcall, plh)
	struct uc_ucld_t   *ucldp;
	void               *tabp;
	pid_t               pid;
	upcall_t            upcall;
	struct uc_plist_t **plh;
{
	struct uc_plist_t      *plp;
	struct uc_calltab_t    *catap;
	struct uc_callent_t    *ucaep;
	struct uc_defprocent_t *udpep;
	struct uc_plent_t      *newp;
	int                     result =0;

	ENTRY;

	if (!ucldp->ucld_inited) {
		*plh = NULL;
		return -ENXIO;
	}

	plp = kmalloc(sizeof(struct uc_plist_t), GFP_KERNEL);
#ifdef DIAGNOSTIC
	if (!plp)
		panic("uc_rcpt_default: cannot allocate plist");
#endif
	ODY_LIST_INIT(plp);
	*plh = plp;

	catap = (struct uc_calltab_t *)tabp;
	ucaep = calltab_find(catap, upcall);
	
	if (!ucaep)
		return 0;

	down (&ucaep->ucae_sem); 
	for (udpep = ucaep->ucae_plist.lh_first;
	     udpep;
	     udpep = udpep->udpe_list.le_next)
	{
		if (pid_compare(pid, udpep->udpe_t)) {
			newp = plent_create(udpep->udpe_t);
			ODY_LIST_INSERT_HEAD(plp, newp, ucpl_list);
			result++;
		}
	}
	up (&ucaep->ucae_sem); 
	return result;	
}

void
uc_print_default(ucldp, tabp)
	struct uc_ucld_t   *ucldp;
	void               *tabp;
{
	struct uc_calltab_t    *catap;
	struct uc_callent_t    *ucaep;
	struct uc_defprocent_t *udpep;
	int                    i;

	ENTRY;

	catap = (struct uc_calltab_t *) tabp;

	printk("Printing table for upclass %s\n", ucldp->ucld_name);

	for (i=0; i<catap->cata_size; i++) {
		for (ucaep = catap->cata_tab[i].lh_first;
		     ucaep;
		     ucaep = ucaep->ucae_list.le_next)
		{
			printk("\tCall %d (procs %d)\n", ucaep->ucae_upcall,
			       ucaep->ucae_numprocs);
			for (udpep = ucaep->ucae_plist.lh_first;
			     udpep;
			     udpep = udpep->udpe_list.le_next)
			{
				printk("\t\tProcess %d\n",
				       udpep->udpe_t->pid);
			}
		}
	}
		
	printk("---------------------------------------\n");
}
