#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * uc_kern.c -- kernel interface to the upcall system.
 */

#include <linux/param.h>
#include <linux/time.h>
#include <linux/malloc.h>
#include <linux/kernel.h>

#include "ody_queue.h"
#include "kernel/upcall/includes/upcall.h"
#include "upcall.private.h"
#include "uc_upclass.h"
#include "uc_pending.h"
#include "kernel/upcall/includes/uc_hash.h"
#include "uc_linux.h"

#define PROCMAGIC 0x75AEAC78

/*
 * Prototypes for static inlined function 
 */
static void inline uc_post __P ((struct task_struct* t, 
				 struct uc_pending_t *ucpp));

void
uc_init(void)
{
	int               i;
	struct uc_ucld_t *ucldp;

	ENTRY;
	
	for (i=1; i<upcl_tabsz; i++) {
		ucldp = upcl_tab[i];
		if (!ucldp) {
			if (!(ucldp = kmalloc(sizeof(struct uc_ucld_t), 
					      GFP_KERNEL))) {
				panic("uc_init: can't malloc");
			}
			ucldp->ucld_inited = 0;
			upcl_tab[i] = ucldp;
		} else {
			ucldp->ucld_table = (ucldp->ucld_init)(ucldp);
			if (ucldp->ucld_table)
				ucldp->ucld_inited = 1;
			else
				ucldp->ucld_inited = 0;
		}
	}
	uc_hashinit();

	EXIT;
}

void
uc_init_upclass (int uc_id, struct uc_ucld_t* ucldp)
{
	struct uc_ucld_t* tabp;

	ENTRY;

	tabp = upcl_tab[uc_id];
	if (tabp->ucld_inited) {
		/* Already inited */
		printk ("Superfulous upcall init of upclass %d\n", uc_id);
		return;
	}

	upcl_tab[uc_id] = ucldp;
	ucldp->ucld_table = (ucldp->ucld_init)(tabp);
	if (ucldp->ucld_table)
		ucldp->ucld_inited = 1;
	else
		ucldp->ucld_inited = 0;

	EXIT;
	return;
}
	
void
uc_procnotify(pep, isreg)
	struct uc_procent_t *pep;
	int                  isreg;
{
#ifdef DIAGNOSTIC
	struct uc_procent_t *probe;
#endif
	struct uc_info_t *ucip;

	ENTRY;

	/* Only we can fiddle with the list */
	ucip = uc_hashlookup(current->pid);
	down (&ucip->uci_sem);

#ifdef DIAGNOSTIC
	for (probe = ucip->uci_upcalls.lh_first;
	     probe;
	     probe = probe->ucpe_list.le_next)
	{
		if (probe->ucpe_upclass == pep->ucpe_upclass
		    && probe->ucpe_upcall == pep->ucpe_upcall)
		{
			break;
		}
	}
#endif

	/* 
	 * If we are registering, and this procentry is already known,
	 * or deregistering, and we never knew about it, then we don't
	 * have to do anything.
	 */
	if ((isreg && (pep->ucpe_magic == PROCMAGIC))
	    || (!isreg && (pep->ucpe_magic != PROCMAGIC)))
	{
#ifdef DIAGNOSTIC
		/* Check to see that ucpe_magic was correct. */
		if ((isreg && !probe)
		     || (!isreg && probe)) 
		{
			printf("Magic = 0x%x, probe = %p", pep->ucpe_magic,
			       probe);
			panic("pep: magic number lied");
		}
#endif
		up (&ucip->uci_sem);
		return;
	}

	if (isreg) {
#ifdef DIAGNOSTIC
		if (probe) 
			panic("pep: registered but mortal");
#endif
		pep->ucpe_magic = PROCMAGIC;
		ODY_LIST_INSERT_HEAD(&ucip->uci_upcalls, pep, ucpe_list);
	} else {
#ifdef DIAGNOSTIC
		if (!probe) 
			panic("pep: unregistered but magic");
#endif
		pep->ucpe_magic = 0xbeefdead;
		ODY_LIST_REMOVE(pep, ucpe_list);
	}
	up (&ucip->uci_sem);

	EXIT;
}

#if 0
void
uc_sigreturn(p)
	struct proc *p;
{
	if (uc_hashlookup(p->p_pid)->uci_pending.tqh_first) 
		psignal(p, SIGUPCALL);
}

void
uc_fork(p1, p2)
	struct proc *p1;
	struct proc *p2;
{
	struct uc_procent_t  *pep;

	p2->p_spare->p_ucinfo = malloc(sizeof(struct uc_info_t), M_UPCALL,
				       M_WAITOK);

#ifdef DIAGNOSTIC
	if (!p2->p_spare->p_ucinfo) panic("uc_fork: malloc failed\n");
#endif

	/* set up pending list of child */
	TAILQ_INIT(&p2->p_pending);

	ODY_LIST_INIT(&p2->p_upcalls);
	/* 
	 * Walk through the registered upcalls of p1, registering them for p2
	 */
	for (pep = p1->p_upcalls.lh_first;
	     pep;
	     pep = pep->ucpe_list.le_next)
	{
		if (uc_doreg(p2, pep->ucpe_upclass, pep->ucpe_upcall, 1))
			panic("uc_fork: upcall_doreg failed");
	}
}

#endif /* 0 */

/*
 * Warning: uc_exit() calls this directly.
 */

/* Deregisters all upcalls and deallocates the ucinfo structure -- since
 * the upcall device was closed one exec, the process should now have no
 * upcall-related state at all
 */

void
uc_exec()
{
	struct uc_procent_t *cprp;
	struct uc_procent_t *nprp;
	struct uc_pending_t *cpep;
	struct uc_pending_t *npep;
	struct uc_info_t *ucip;
	int                 freeargs;

	ucip = uc_hashlookup(current->pid);
	/* 
	 * Walk down list of registered-for upcalls, deregistering
	 * things.  Beware: the procent's will be freed out from under
	 * us as a side-effect of deregistration 
	 */

	cprp = ucip->uci_upcalls.lh_first;
	while (cprp) {
		nprp = cprp->ucpe_list.le_next;
		if (uc_doreg(cprp->ucpe_upclass, cprp->ucpe_upcall, 0))
			panic("uc_exec: upcall_doreg failed");
		cprp = nprp;
	}

	/* 
	 * Walk down list of pending upcalls, toasting them.  If they
	 * are synchronous, be sure to return ETIMEDOUT thorugh the
	 * errno field.
	 */
	
	cpep = ucip->uci_pending.tqh_first;
	while (cpep) {
		npep = cpep->ucp_list.tqe_next;

		/* Dequeue it, and decide whether we should free arguments */
		down (&ucip->uci_sem);
		cpep->ucp_shared--;
		freeargs = !(cpep->ucp_shared);
		TAILQ_REMOVE(&ucip->uci_pending, cpep, ucp_list);
		up (&ucip->uci_sem);
		
		if (freeargs) {
			if (cpep->ucp_flags & UCP_FREEARGS) {
				kfree(cpep->ucp_argd->uca_argp);
			}
			kfree(cpep->ucp_argd);
		}

		if (cpep->ucp_flags & UCP_SYNC) {
			/* We have to return an error */
			cpep->ucp_errno = -ETIMEDOUT;
			wake_up(&cpep->ucp_wq);
		} else {
			kfree(cpep);
		}
		
		cpep = npep;
	}
	uc_hashdelete(ucip);
	kfree(ucip);
}

/*
 * On exit, we deregister all previous registrations and deallocate the
 * ucinfo structure.  This is basically what uc_exec does. 
 */
void
uc_exit()
{
	uc_exec();
}

/* Used by sys___upcall_reg, uc_fork(), and uc_exec() */
int uc_doreg(upclass, upcall, isreg)
	upclass_t           upclass;
	upcall_t            upcall;
	int                 isreg;
{
	int                    result;
	struct uc_ucld_t      *ucldp;
	struct uc_procent_t   *pep;

	ENTRY;

	if (upclass <= 0 || upclass > upcl_tabsz) {
		ODEBUG (D_SYSUPCALL, "Invalid Upclass\n");
		return -EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		ODEBUG (D_SYSUPCALL, "Not inited\n");
		return -EINVAL;
	}
	
	result = ((ucldp->ucld_notify)(ucldp, ucldp->ucld_table, current,
				       upcall, isreg, &pep));

	if (!result && pep) {
		ODEBUG (D_SYSUPCALL, "Result is %d\n", result);
		uc_procnotify(pep, isreg);
		if (!isreg) {
			if (ucldp->ucld_freepep) {
				(ucldp->ucld_freepep)(ucldp, 
						      ucldp->ucld_table,
						      pep);
			} else {
				kfree(pep);
			}
		}
	}

#ifdef MODULE
	if (result == 0) {
  	        change_use_count(isreg);
	}
#endif
	    
	EXIT;

	return result;
}

/* 
 * Common code to post upcall to a process.  
 */
static void inline
uc_post(struct task_struct* t, struct uc_pending_t *ucpp)
{
        struct uc_info_t    *ucip;

	ucip = uc_hashlookup(t->pid);
#ifdef DIAGNOSTIC
	{
	struct uc_procent_t *ucpep;

	ENTRY;


	down (&ucip->uci_sem);
	for (ucpep = ucip->uci_upcalls.lh_first; ucpep;
	     ucpep = ucpep->ucpe_list.le_next)
	{
		if (ucpep->ucpe_upclass == ucpp->ucp_upclass
		    && ucpep->ucpe_upcall == ucpp->ucp_upcall)
		{
			break;
		}
	}
	if (!ucpep)
		panic("uc_post: upcall not registered for");
	}
#endif
	
	TAILQ_INSERT_TAIL(&ucip->uci_pending, ucpp, ucp_list);
#if UC_SIGNAL
	send_sig_info(ucip->uci_sig, 0, t);
#else	
	wake_up (&ucip->uci_wq);
#endif
	up (&ucip->uci_sem);

	EXIT;
}

int
uc_post_sync(t, upclass, upcall, argsz, argp, argmtype, retszp, retp)
        struct task_struct *t;
	upclass_t           upclass;
	upcall_t            upcall;
	size_t              argsz;
	void               *argp;
	int                 argmtype;
	size_t             *retszp;
	void               *retp;
{
	struct wait_queue     wait = { current, NULL };
	struct uc_args_t     *ucap;
	struct uc_pending_t  *ucpp;
	int                  flags;
	int                  result;

	ENTRY;

	flags = (argmtype) ? UCP_FREEARGS : 0;
	flags |= UCP_SYNC;

	ucap = uc_args_create(argsz, argp, 1);
	ucpp = uc_pend_create(upclass, upcall, ucap, *retszp, retp, flags);

	/* Atomically enqueue this request, post SIGUPCALL, and sleep */

	uc_post(t, ucpp);
	/* 
	 * This sleep is neither interruptible, nor timed.  If you
	 * don't want to blcok forever, use uc_post_async
	 */

	/* Add self to wait queue - wait for wakeup */
	add_wait_queue(&ucpp->ucp_wq, &wait);
	current->state = TASK_UNINTERRUPTIBLE;
	schedule();

	/* We've been awakened.  Grab the answer. */
	remove_wait_queue(&ucpp->ucp_wq, &wait);
	current->state = TASK_RUNNING;
	result = ucpp->ucp_errno;

	if (result) {
		*retszp = 0;
	} else {
		*retszp = ucpp->ucp_retsz;
	}

	/* 
	 * sys___upcall_ret has already disposed of the argument if it
	 * was supposed to be freed, and has disposed of the ucap.
	 */

	kfree(ucpp);

	EXIT;

	return result;
}

int
uc_post_async(ucplsz, ucplp, upclass, upcall, argsz, argp, argmtype)
	int ucplsz;
	struct uc_plist_t *ucplp;
	upclass_t upclass;
	upcall_t upcall;
	size_t argsz;
	void *argp;
	int argmtype;
{
	struct uc_args_t     *ucap;
	struct uc_pending_t  *ucpp;
	struct uc_plent_t    *thisp;
	struct uc_plent_t    *nextp;
	int                  flags;

	ENTRY;

	flags = (argmtype) ? UCP_FREEARGS : 0;
	ucap = uc_args_create(argsz, argp, ucplsz);
	
	/* 
	 * We are going to turn interrupts off and back on inside the
	 * iteration across the process list.  This is so we don't
	 * have to worry about cases of sending an upcall to the
	 * entire universe taking up more time than we'd like.  The
	 * uc_pend_create also needs to be outside of the critical
	 * region, since it might block.
	 */
	
	/* XXX - need to free as we go */
	thisp = ucplp->lh_first;

	while (thisp) {	
		nextp = thisp->ucpl_list.le_next;
		ucpp = uc_pend_create(upclass, upcall, ucap, 0, NULL, flags);

		uc_post(thisp->ucpl_t, ucpp);
		
		ODY_LIST_REMOVE(thisp, ucpl_list);
		kfree(thisp);
		thisp = nextp;
#ifdef DIAGNOSTIC
		ucplsz--;
#endif
	}
#ifdef DIAGNOSTIC
	if (ucplsz) 
		panic("uc_post_async: ucplsz lied");
#endif

	EXIT;

	return 0;
}

