#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <linux/param.h>
#include <linux/time.h>
#include <linux/malloc.h>
#include <asm/uaccess.h>
#ifdef UC_SIGNAL
#include <linux/signal.h>
#endif

#include "ody_queue.h"
#include "kernel/upcall/includes/upcall.h"
#include "upcall.private.h"
#include "uc_upclass.h"
#include "uc_pending.h"
#include "uc_ioctl.h"
#include "kernel/upcall/includes/uc_hash.h"
#include "uc_linux.h"

static int uc_serial = 1;

int
uci_sync(uap)
	struct uci_sync_args *uap;
{
	upclass_t            upclass;
	upcall_t             upcall;
	pid_t                pid;
	struct uc_ucld_t     *ucldp;
	int                  plsz;
	struct uc_plist_t    *plp=0;
	struct uc_plent_t    *thisp;
	struct uc_plent_t    *nextp;
	struct task_struct   *targp;
	size_t               argsz;
	void                 *argp=0;
	size_t               retsz;
#ifdef DIAGNOSTIC
	size_t               origret;
#endif
	void                 *retp=0;
	int                  result=0;
	
	ENTRY;

	/* Step 1: get the process for which this is intended */
	upclass = uap->upclass;
	upcall = uap->upcall;
	pid = uap->pid;

	if (pid <= 0) {
		return -EINVAL;
	}

	if (upclass <= 0 || upclass > upcl_tabsz) {
		return -EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		return -EINVAL;
	}

	
	/* 
	 * Returns a negative number on error, the negation of which
	 * is the errno code. 
	 */
	plsz = (ucldp->ucld_rcpt)(ucldp, ucldp->ucld_table, pid, upcall, &plp);
	if (plsz < 0) 
		return -plsz;

	/* From this point on, we will need to free things */
	if (plsz != 1) {
		result = -ESRCH;
		goto exit;
	}

	targp = plp->lh_first->ucpl_t;

	/* Get argument buffer, allocate return buffer */
	argsz = uap->argsz;
	if (argsz) {
		argp = kmalloc(argsz, GFP_KERNEL);
		result = copy_from_user(argp, uap->argp, argsz);
		if (result) goto exit;
	}

	result = copy_from_user(&retsz, uap->retszp, sizeof(retsz));
	if (result) goto exit;

#ifdef DIAGNOSTIC
	origret = retsz;
#endif
	if (retsz) {
		retp = kmalloc(retsz, GFP_KERNEL);
	}

	/* Post the call */
	result = uc_post_sync(targp, upclass, upcall, argsz, argp, 0,
			      &retsz, retp);
	if (result) goto exit;

#ifdef DIAGNOSTIC
	if (retsz > origret) 
	    panic("uci_sync: return size exceeded");
#endif

	/* send out the results */
	result = copy_to_user(uap->retp, retp, retsz);
	if (!result)
	    if (put_user(retsz,uap->retszp) < 0)
		    result = -EFAULT;

 exit:
	/* Free the argument & return buffers */
	if (argp) kfree(argp);
	if (retp) kfree(retp);

	/* Free the plist */
	if (plp) {
		thisp = plp->lh_first;
		while (thisp) {
			nextp = thisp->ucpl_list.le_next;
			kfree(thisp);
			thisp = nextp;
		}
	}

	EXIT;

	return result;
}


int
uci_async(uap)
	struct uci_async_args *uap;
{
	upclass_t             upclass;
	upcall_t              upcall;
	pid_t                 pid;
	struct uc_ucld_t      *ucldp;
	int                   plsz;
	struct uc_plist_t     *plp=0;
	size_t                argsz;
	void                  *argp=0;
	int                   result=0;

	ENTRY;

	upclass = uap->upclass;
	upcall = uap->upcall;
	pid = uap->pid;

	if (upclass <= 0 || upclass > upcl_tabsz) {
	    return -EINVAL;
	}

	ucldp = upcl_tab[upclass];
	if (!ucldp->ucld_inited) {
		return -EINVAL;
	}

	ODEBUG (D_SYSUPCALL, "About to call default rcpt function\n");

	/* 
	 * Returns a negative number on error, the negation of which
	 * is the errno code. 
	 */
	plsz = (ucldp->ucld_rcpt)(ucldp, ucldp->ucld_table, pid, upcall, &plp);
	ODEBUG (D_SYSUPCALL, "default rcpt function returns %d\n", plsz);
	if (plsz < 0) 
		return -plsz;

	if (plsz == 0)
		return -ESRCH;

	/* Get argument buffer, allocate return buffer */
	argsz = uap->argsz;
	if (argsz) {
		argp = kmalloc(argsz, GFP_KERNEL);
		result = copy_from_user(argp, uap->argp, argsz);
		if (result) return result;
	}

	ODEBUG (D_SYSUPCALL, "About to call uc_post_async\n");
	result = uc_post_async(plsz, plp, upclass, upcall, argsz, 
			       argp, 1);

	EXIT;

	return result;
}

int
uci_reg(uap)
	struct uci_reg_args *uap;
{
	upclass_t              upcls_arg;
	upcall_t               upca_arg;
	int                    flags_arg;
	int                    isreg;
	int                    result=0;
	
	ENTRY;

	upcls_arg = uap->upclass;
	upca_arg = uap->upcall;
	flags_arg = uap->flags;

	if (flags_arg & UPCALL_REGISTER)
		isreg = 1;
	else 
		isreg = 0;

	result = uc_doreg(upcls_arg, upca_arg, isreg);

	EXIT;

	return result;
}

#define SUWORDCHK(addr, val, result, stmt) \
do {                                       \
      (result) = put_user((val),(addr));   \
      if (result) {                      \
                result = -EFAULT;           \
              stmt                       \
      }                                  \
} while (0)

int
uci_next(uap)
	struct uci_next_args *uap;
{
#ifndef UC_SIGNAL
	struct wait_queue    wait = { current, NULL };
#endif
	size_t               argsz;
	struct uc_info_t     *ucip;
	struct uc_pending_t  *ucpp;
	int                  result=0;

	ENTRY;

	/* 
	 * Grab a pointer to the first entry.  Lock out inserting
	 * processes while we are at it.  Check that no one else is
	 * using it, and if not, mark it in progress.  
	 *
	 * Now, we are the only process that cares about anything in
	 * this entry other than the links entry, which we won't touch
	 * until we dequeue in __upcall_ret
	 */
	ucip = uc_hashlookup(current->pid);
	down (&ucip->uci_sem);
	ucpp = ucip->uci_pending.tqh_first;
#ifndef UC_SIGNAL		
	if (!ucpp) {
		/* No upcalls pending, so wait until one arrives */
		add_wait_queue(&ucip->uci_wq, &wait);
		current->state = TASK_INTERRUPTIBLE;
		up (&ucip->uci_sem);
		schedule();

		/* We've been awakened.  There ought to be an upcall */
		down (&ucip->uci_sem);
		remove_wait_queue(&ucip->uci_wq, &wait);
		current->state = TASK_RUNNING;
		ucpp = ucip->uci_pending.tqh_first;
	}
#endif
	if (!ucpp) {
		result = -ENOENT;
	} else if (ucpp->ucp_flags & UCP_INPROGRESS) {
		result = -EALREADY;
	} else {
		ucpp->ucp_serial = uc_serial++;
		ucpp->ucp_flags |= UCP_INPROGRESS; /* lock out */
	}
	up (&ucip->uci_sem);

	if (result) goto exit;

	get_user(argsz, uap->argszp);

       /* Any of these could cause a jump to exit if they fail */
	SUWORDCHK(uap->upclassp, ucpp->ucp_upclass, result, goto exit;);
	SUWORDCHK(uap->upcallp, ucpp->ucp_upcall, result, goto exit;);
	SUWORDCHK(uap->argszp, ucpp->ucp_argsz, result, goto exit;);
	SUWORDCHK(uap->retszp, ucpp->ucp_retsz, result, goto exit;);
	SUWORDCHK(uap->serialp, ucpp->ucp_serial, result, goto exit;);
	
	/* argsz is valid, since we've stored to it. */
	if (argsz >= ucpp->ucp_argsz) {
		/* We can do the copyout; it's big enough. */
		result = copy_to_user(uap->argp, ucpp->ucp_argp, 
				      ucpp->ucp_argsz);
		if (result) goto exit;
	}
	
	/* We're done */
 exit:	
	if (result) {
		/* Store zero's in upcall/upclass */
		(void)put_user(0,uap->upclassp);
		(void)put_user(0,uap->upcallp);
		/* Clear the in-progress flag if necessary */
		down (&ucip->uci_sem);
		if (ucpp) {
			ucpp->ucp_flags &= ~UCP_INPROGRESS;
			ucpp->ucp_serial = 0;
		}
		up (&ucip->uci_sem);
		/* If we were ENOENT, then just return 0 */
		if (result == -ENOENT) result = 0;
	}

	EXIT;

	return result;
}

int
uci_args(uap)
	struct uci_args_args *uap;
{
	struct uc_info_t    *ucip;
	struct uc_pending_t *ucpp;
	int                 result=0;

	ENTRY;

	ucip = uc_hashlookup(current->pid);
	down (&ucip->uci_sem);
	ucpp = ucip->uci_pending.tqh_first;
	if (!ucpp 
	    || !(ucpp->ucp_flags & UCP_INPROGRESS) 
	    || (ucpp->ucp_serial != uap->serial))
	{
		result = -ENOENT;
	}
	up (&ucip->uci_sem);
	if (result) return result;

	if (ucpp->ucp_argsz > uap->bufsz) {
		return -EINVAL;
	}

	/* Copy the arguments out */
	result = copy_to_user(uap->buf, ucpp->ucp_argp, ucpp->ucp_argsz);

	EXIT;

	return result;
}

int
uci_ret(uap)
	struct uci_ret_args *uap;
{
	struct uc_info_t    *ucip;
	struct uc_pending_t *ucpp;
	int                 result=0;
	size_t              retsz;
	int                 freeargs;
	int                 err;

	ENTRY;

	/* 
	 * Grab the pending entry; we can't remove it until after the
	 * copyout if we owe one; the copyout has to complete before
	 * dequeuing 
	 */
	ucip = uc_hashlookup(current->pid);
	down (&ucip->uci_sem);
	ucpp = ucip->uci_pending.tqh_first;
	if (!ucpp
	    || !(ucpp->ucp_flags & UCP_INPROGRESS)
	    || (ucpp->ucp_serial != uap->serial))
	{
		result = -ENOENT;
	}
	up (&ucip->uci_sem);
	
	if (result) return result;

	if (ucpp->ucp_flags & UCP_SYNC) {
		err = uap->errno;
		if (err) {
			ucpp->ucp_errno = err;
		} else {
			retsz = uap->retsz;
			if (retsz > ucpp->ucp_retsz)
				return -E2BIG;
			result = copy_from_user(ucpp->ucp_retp, uap->retp, 
						retsz);

			if (result) return result;
			ucpp->ucp_errno = 0;
			ucpp->ucp_retsz = retsz;
		}
	}

	/* Can now remove the entry and dispose of it */
	down (&ucip->uci_sem);
	TAILQ_REMOVE(&ucip->uci_pending, ucpp, ucp_list);
	ucpp->ucp_shared--;
	freeargs = !(ucpp->ucp_shared);
	up (&ucip->uci_sem);

	if (freeargs) {
		if (ucpp->ucp_flags & UCP_FREEARGS) {
			kfree(ucpp->ucp_argd->uca_argp);
		}
		kfree(ucpp->ucp_argd);
	}

	if (ucpp->ucp_flags & UCP_SYNC)
		/* If synchronous, wake up thread waiting for reply */
		wake_up(&ucpp->ucp_wq);
	else
		kfree(ucpp);

	EXIT;

	return 0;
}

#ifdef UC_SIGNAL
int
uci_set_sig(uap)
	struct uci_set_sig_args *uap;
{
	struct uc_info_t *ucip;
	int               result=0;

	ENTRY;

	/* Is this a valid signal? */
	if (uap->sig >= SIGRTMIN) {
	    ODEBUG (D_SYSUPCALL, "oops, uap->sig = %d, NSIG = %d\n",
		    uap->sig, SIGRTMIN);
	    return EINVAL;
	}

	/* Set upcall signal for this process */
	ucip = uc_hashlookup(current->pid);
	down (&ucip->uci_sem);
	ucip->uci_sig = uap->sig;
	up (&ucip->uci_sem);

	EXIT;

	return result;
}
#endif


