#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Stolen shamelessly from the MiniCache
 */

/*
 * This module contains the routines to implement the CFS name cache. The
 * purpose of this cache is to reduce the cost of translating pathnames 
 * into Vice FIDs. Each entry in the cache contains the name of the file,
 * the vnode (FID) of the parent directory, and the cred structure of the
 * user accessing the file.
 *
 * The first time a file is accessed, it is looked up by the local Venus
 * which first insures that the user has access to the file. In addition
 * we are guaranteed that Venus will invalidate any name cache entries in
 * case the user no longer should be able to access the file. For these
 * reasons we do not need to keep access list information as well as a
 * cred structure for each entry.
 *
 * The table can be accessed through the routines cnc_init(), cnc_enter(),
 * cnc_lookup(), cnc_rmfidcred(), cnc_rmfid(), cnc_rmcred(), and cnc_purge().
 * There are several other routines which aid in the implementation of the
 * hash table.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>
#include <sys/select.h>

#include "kernel/vfs/includes/ody.h"
#include "kernel/vfs/includes/onode.h"
#include "kernel/vfs/includes/odync.h"

#ifdef	__FreeBSD__
#include <vm/vm.h>
#include <vm/vm_object.h>
#endif

/* 
 * Declaration of the name cache data structure.
 */

int 	odync_use = 1;			 /* Indicate use of Ody Name Cache */

int	odync_size = ODYNC_CACHESIZE;	 /* size of the cache */
int	odync_hashsize = ODYNC_HASHSIZE; /* size of the primary hash */

struct 	odycache *odyncheap;	/* pointer to the cache entries */
struct	odyhash  *odynchash;	/* hash table of cfscache pointers */
struct	odylru   odync_lru;	/* head of lru chain */

struct odync_statistics odync_stat;	/* Keep various stats */

/* 
 * for testing purposes
 */
int odync_debug = 0;


/*
 * Entry points for the Odyssey Name Cache
 */

/*  
 * Initialize the cache, the LRU structure and the Hash structure(s)
 */

#define TOTAL_CACHE_SIZE 	(sizeof(struct odycache) * odync_size)
#define TOTAL_HASH_SIZE 	(sizeof(struct odyhash)  * odync_hashsize)

int odync_initialized = 0;      /* Initially the cache has not been initialized */

void
odync_init()
{
    register int i;

    /* zero the statistics structure */
    
    bzero(&odync_stat, (sizeof(struct odync_statistics)));
    
    ODY_ALLOC(odyncheap, struct odycache *, TOTAL_CACHE_SIZE);
    ODY_ALLOC(odynchash, struct odyhash *, TOTAL_HASH_SIZE);
    
    odync_lru.lru_next = 
	odync_lru.lru_prev = (struct odycache *)LRU_PART(&odync_lru);
    
    
    for (i=0; i < odync_size; i++) {	/* initialize the heap */
	ODYNC_LRUINS(&odyncheap[i], &odync_lru);
	ODYNC_HSHNUL(&odyncheap[i]);
	odyncheap[i].op = odyncheap[i].dop = (struct onode *)0;
    }
    
    for (i=0; i < odync_hashsize; i++) {	/* initialize the hashtable */
	ODYNC_HSHNUL((struct odycache *)&odynchash[i]);
    }
    
    odync_initialized++;
}


/*
 * Auxillary routines -- shouldn't be entry points
 */

static struct odycache *
odync_find(struct onode *dop,
	   const char   *name,
	   int           namelen,
	   struct ucred *cred,
	   int           hash)
{
	/* 
	 * hash to find the appropriate bucket, look through the chain
	 * for the right entry (especially right cred, unless cred == 0) 
	 */
	register struct odycache *oncp;
	int count = 1;

	ODYNC_DEBUG(ODYNC_FIND, 
		    myprintf(("odync_find(dop %p, name %s, len %d, cred %p, hash %d\n",
			   dop, name, namelen, cred, hash));)

	for (oncp = odynchash[hash].hash_next; 
	     oncp != (struct odycache *)&odynchash[hash];
	     oncp = oncp->hash_next, count++) 
	{

	    if ((ODY_NAMEMATCH(oncp, name, namelen, dop)) &&
		((cred == 0) || (oncp->cred == cred) || 
		 CRED_EQ(oncp->cred, cred)))
	    { 
		/* compare cr_uid instead */
		odync_stat.Search_len += count;
		return(oncp);
	    }
	}
	
	return((struct odycache *)NULL);
}

/*
 * Enter a new (dir cnode, name) pair into the cache, updating the
 * LRU and Hash as needed.
 */

void
odync_enter(dop, name, namelen, cred, op)
    struct onode *dop;
    register const char *name;
    int namelen;
    struct ucred *cred;
    struct onode *op;
{
    register struct odycache *oncp;
    register int hash;
    
    if (odync_use == 0)			/* Cache is off */
	return;
    
    ODYNC_DEBUG(ODYNC_ENTER, 
		myprintf(("Enter: dop %p op %p name %s cred %p \n",
			  dop, op, name, cred)); )
	
    hash = ODYNC_HASH(name, namelen, dop);
    oncp = odync_find(dop, name, namelen, cred, hash);
    if (oncp != (struct odycache *) 0) {	
	odync_stat.dbl_enters++;		/* duplicate entry */
	return;
    }
    
    odync_stat.enters++;		/* record the enters statistic */
    
    /* Grab the next element in the lru chain */
    oncp = ODYNC_LRUGET(odync_lru);
    
    ODYNC_LRUREM(oncp);	/* remove it from the lists */
    
    if (ODYNC_VALID(oncp)) {
	/* Seems really ugly, but we have to decrement the appropriate
	   hash bucket length here, so we have to find the hash bucket
	   */
	odynchash[ODYNC_HASH(ONCP_NAME(oncp), oncp->namelen,
			     oncp->dop)].length--;
	
	odync_stat.lru_rm++;	/* zapped a valid entry */
	ODYNC_HSHREM(oncp);
	vrele(OTOV(oncp->dop)); 
	vrele(OTOV(oncp->op));
	crfree(oncp->cred);
	if (LONG_NAME(oncp->namelen))
	    ODY_FREE(oncp->n.longname, oncp->namelen);
    }
    
    /*
     * Put a hold on the current vnodes and fill in the cache entry.
     */
    vref(OTOV(op));
    vref(OTOV(dop));
    crhold(cred); 
    oncp->dop = dop;
    oncp->op = op;
    oncp->namelen = namelen;
    oncp->cred = cred;
    
    if (LONG_NAME(oncp->namelen)) {
	ODYNC_DEBUG(ODYNC_ENTER, 
		    myprintf(("long name enter %s\n",name));)
	odync_stat.long_name_enters++;      /* record stats */
	ODY_ALLOC(oncp->n.longname, char *, oncp->namelen);
	bcopy(name, oncp->n.longname, (unsigned)namelen);
    } else
	bcopy(name, oncp->n.name, (unsigned)namelen);
    
    /* Insert into the lru and hash chains. */
    
    ODYNC_LRUINS(oncp, &odync_lru);
    ODYNC_HSHINS(oncp, &odynchash[hash]);
    odynchash[hash].length++;                      /* Used for tuning */
    
    ODYNC_DEBUG(ODYNC_PRINTODYNC, print_odync(); )
}

/*
 * Find the (dir cnode, name) pair in the cache, if it's cred
 * matches the input, return it, otherwise return 0
 */

struct onode *
odync_lookup(dop, name, namelen, cred)
	struct onode *dop;
	register const char *name;
        int namelen;
	struct ucred *cred;
{
	register int hash;
	register struct odycache *oncp;

	if (odync_use == 0)			/* Cache is off */
		return((struct onode *) 0);

	if (LONG_NAME(namelen)) {
	    ODYNC_DEBUG(ODYNC_LOOKUP, 
			myprintf(("long name lookup %s\n",name));)
	    odync_stat.long_name_lookups++;		/* record stats */
	}

	/* Use the hash function to locate the starting point,
	   then the search routine to go down the list looking for
	   the correct cred.
 	 */

	hash = ODYNC_HASH(name, namelen, dop);
	oncp = odync_find(dop, name, namelen, cred, hash);
	if (oncp == (struct odycache *) 0) {
		odync_stat.misses++;			/* record miss */
		return((struct onode *) 0);
	}

	odync_stat.hits++;

	/* put this entry at the end of the LRU */
	ODYNC_LRUREM(oncp);
	ODYNC_LRUINS(oncp, &odync_lru);

	/* move it to the front of the hash chain */
	/* don't need to change the hash bucket length */
	ODYNC_HSHREM(oncp);
	ODYNC_HSHINS(oncp, &odynchash[hash]);

	ODYNC_DEBUG(ODYNC_LOOKUP, 
		printf("lookup: dop %p, name %s, cred %p = op %p\n",
			dop, name, cred, oncp->op); )

	return(oncp->op);
}

static void
odync_remove(struct odycache *oncp,
	     enum dc_status dcstat)
{
	/* 
	 * remove an entry -- VN_RELE(oncp->dop, op), crfree(cred),
	 * remove it from it's hash chain, and
	 * place it at the head of the lru list.
	 */
        ODYNC_DEBUG(ODYNC_REMOVE,
		    myprintf(("odync_remove %s from parent %lx.%lx.%lx.%lx\n",
			      ONCP_NAME(oncp),
			      (oncp->dop)->o_fid.ofid_ttype,
			      (oncp->dop)->o_fid.ofid_ttype,
			      (oncp->dop)->o_fid.ofid_ttype,
			      (oncp->dop)->o_fid.ofid_ttype)););

  	ODYNC_HSHREM(oncp);

	ODYNC_HSHNUL(oncp);		/* have it be a null chain */
	if ((dcstat == IS_DOWNCALL) && (OTOV(oncp->dop)->v_usecount == 1)) {
		oncp->dop->o_flags |= ON_PURGING;
	}
	vrele(OTOV(oncp->dop)); 

	if ((dcstat == IS_DOWNCALL) && (OTOV(oncp->op)->v_usecount == 1)) {
		oncp->op->o_flags |= ON_PURGING;
	}
	vrele(OTOV(oncp->op)); 

	crfree(oncp->cred); 
	bzero(DATA_PART(oncp),DATA_SIZE);

	if (LONG_NAME(oncp->namelen))
	    ODY_FREE(oncp->n.longname, oncp->namelen);

	/* Put the null entry just after the least-recently-used entry */
	/* LRU_TOP adjusts the pointer to point to the top of the structure. */
	ODYNC_LRUREM(oncp);
	ODYNC_LRUINS(oncp, LRU_TOP(odync_lru.lru_prev));
}

/*
 * Remove all entries with a parent which has the input fid.
 */
void
odync_zapParentfid(fid, dcstat)
    struct ofid *fid;
    enum dc_status dcstat;
{
    /* To get to a specific fid, we might either have another hashing
       function or do a sequential search through the cache for the
       appropriate entries. The later may be acceptable since I don't
       think callbacks or whatever Case 1 covers are frequent occurences.
       */
    register struct odycache *oncp, *noncp;
    register int i;
    
    if (odync_use == 0)			/* Cache is off */
	return;
    
    ODYNC_DEBUG(ODYNC_ZAPPFID, 
		myprintf(("ZapParent: fid 0x%lx, 0x%lx, 0x%lx, 0x%lx \n",
			  fid->ofid_ttype, fid->ofid_tome, fid->ofid_vnum, 
			  fid->ofid_uniq)); )
	
    odync_stat.zapPfids++;
    
    for (i = 0; i < odync_hashsize; i++) {

	/*
	 * Need to save the hash_next pointer in case we remove the
	 * entry. remove causes hash_next to point to itself.
	 */
	
	for (oncp = odynchash[i].hash_next; 
	     oncp != (struct odycache *)&odynchash[i];
	     oncp = noncp) {
	    noncp = oncp->hash_next;
	    if (oncp->dop->o_fid.ofid_ttype == fid->ofid_ttype &&
		oncp->dop->o_fid.ofid_tome == fid->ofid_tome &&
		oncp->dop->o_fid.ofid_vnum == fid->ofid_vnum &&
		oncp->dop->o_fid.ofid_uniq == fid->ofid_uniq)
	    {
		
		odynchash[i].length--;      /* Used for tuning */
		odync_remove(oncp, dcstat); 
	    }
	}
    }
}

/*
 * Remove all entries which have the same fid as the input
 */
void
odync_zapfid(fid, dcstat)
    struct ofid *fid;
    enum dc_status dcstat;
{
    /* 
     * See comment for zapParentfid. This routine will be used if
     * attributes are being cached.  
     */
    register struct odycache *oncp, *noncp;
    register int i;
    
    if (odync_use == 0)			/* Cache is off */
	return;
    
    ODYNC_DEBUG(ODYNC_ZAPFID, 
		myprintf(("Zapfid: fid 0x%lx, 0x%lx, 0x%lx, 0x%lx \n",
			  fid->ofid_ttype, fid->ofid_tome, 
			  fid->ofid_vnum, fid->ofid_uniq));)
	
    odync_stat.zapFids++;

    for (i = 0; i < odync_hashsize; i++) {
	for (oncp = odynchash[i].hash_next; 
	     oncp != (struct odycache *)&odynchash[i];
	     oncp = noncp) {
	    noncp = oncp->hash_next;
	    if (oncp->op->o_fid.ofid_ttype == fid->ofid_ttype &&
		oncp->op->o_fid.ofid_tome == fid->ofid_tome &&
		oncp->op->o_fid.ofid_vnum == fid->ofid_vnum &&
		oncp->op->o_fid.ofid_uniq == fid->ofid_uniq)
	    {
		odynchash[i].length--;     /* Used for tuning */
		odync_remove(oncp, dcstat);
	    }
	}
    }
}

/* 
 * Remove all entries which match the fid and the cred
 */
void
odync_zapvnode(fid, cred, dcstat)
    struct ofid  *fid;
    struct ucred *cred;
    enum dc_status dcstat;
{
	/* See comment for zapfid. I don't think that one would ever
	   want to zap a file with a specific cred from the kernel.
	   We'll leave this one unimplemented.
	 */
	if (odync_use == 0)			/* Cache is off */
		return;

	ODYNC_DEBUG(ODYNC_ZAPVNODE, 
		myprintf(("Zapvnode: fid 0x%lx, 0x%lx, 0x%lx, 0x%lx cred %p\n",
			  fid->ofid_ttype, fid->ofid_tome, fid->ofid_vnum,
			  fid->ofid_uniq, cred)););
}

/*
 * Remove all entries which have the (dir vnode, name) pair
 */
void
odync_zapfile(dop, name, namelen)
    struct onode *dop;
    register const char *name;
    int namelen;
{
	/* use the hash function to locate the file, then zap all
 	   entries of it regardless of the cred.
	 */
	register struct odycache *oncp;
	int hash;

	if (odync_use == 0)			/* Cache is off */
		return;

	ODYNC_DEBUG(ODYNC_ZAPFILE, 
		myprintf(("Zapfile: dop %p name %s \n",
			  dop, name)); )

	if (namelen > ODYNC_NAMELEN) {
		odync_stat.long_remove++;		/* record stats */
	}

	odync_stat.zapFile++;

	hash = ODYNC_HASH(name, namelen, dop);
	oncp = odync_find(dop, name, namelen, 0, hash);

	while (oncp) {
	  odynchash[hash].length--;                 /* Used for tuning */
/* 1.3 */
	  odync_remove(oncp, NOT_DOWNCALL);
	  oncp = odync_find(dop, name, namelen, 0, hash);
	}
}

/* 
 * Remove all the entries for a particular user. Used when tokens expire.
 * A user is determined by his/her effective user id (id_uid).
 */

void
odync_purge_user(uid, dcstat)
	ouid_t uid;
	enum dc_status dcstat;
{
    /* 
     * I think the best approach is to go through the entire cache via
     * HASH or whatever and zap all entries which match the input
     * cred. Or just flush the whole cache.  It might be best to go
     * through on basis of LRU since cache will almost always be full
     * and LRU is more straightforward.  
     */

    register struct odycache *oncp, *noncp;
    int hash;

    if (odync_use == 0)			/* Cache is off */
	return;
    
    ODYNC_DEBUG(ODYNC_PURGEUSER, 
		myprintf(("ZapDude: uid 0x%lx\n",uid)); )
    odync_stat.zapUsers++;
    
    for (oncp = ODYNC_LRUGET(odync_lru);
	 oncp != (struct odycache *)(&odync_lru);
	 oncp = noncp) 
    {
	noncp = ODYNC_LRUGET(*oncp);
	
	if ((ODYNC_VALID(oncp)) &&
	    ((oncp->cred)->cr_uid == uid)) {
	    /* 
	     * Seems really ugly, but we have to decrement the
	     * appropriate hash bucket length here, so we have to find
	     * the hash bucket 
	     */
	    hash = ODYNC_HASH(ONCP_NAME(oncp), oncp->namelen, oncp->dop);
	    odynchash[hash].length--;     /* For performance tuning */
	    
	    odync_remove(oncp, dcstat); 
	}
    }
}

/*
 * Flush the entire name cache. In response to a flush of the Venus cache.
 */

void
odync_flush(dcstat)
     enum dc_status dcstat;
{
	/* 
	 * One option is to deallocate the current name cache and call
	 * init to start again. Or just deallocate, then rebuild.  Or
	 * again, we could just go through the array and zero the
	 * appropriate fields.  
	 */
	
	/* 
	 * Go through the whole lru chain and kill everything as we go.
	 * I don't use remove since that would rebuild the lru chain
	 * as it went and that seemed unneccesary.
	 */
	register struct odycache *oncp;
	int i;

	if (odync_use == 0)			/* Cache is off */
		return;

	odync_stat.Flushes++;

	for (oncp = ODYNC_LRUGET(odync_lru);
	     oncp != (struct odycache *)&odync_lru;
	     oncp = ODYNC_LRUGET(*oncp)) {
		if (ODYNC_VALID(oncp)) {
			ODYNC_HSHREM(oncp);	/* only zero valid nodes */
			ODYNC_HSHNUL(oncp);
			if ((dcstat == IS_DOWNCALL)
			    && (OTOV(oncp->dop)->v_usecount == 1))
			  {
			    oncp->dop->o_flags |= ON_PURGING;
			  }
			vrele(OTOV(oncp->dop)); 
			/* ody_vmflush doesn't do anything -- skip it */
#if 0
			if (CTOV(cncp->cp)->v_flag & VTEXT)) {
			    if (cfs_vmflush(cncp->cp))
				CFSDEBUG(CFS_FLUSH, myprintf(("odync_flush: (%x.%x.%x) busy\n", cncp->cp->c_fid.Volume, cncp->cp->c_fid.Vnode, cncp->cp->c_fid.Unique)); )
			}
#endif
			if ((dcstat == IS_DOWNCALL) 
			    && (OTOV(oncp->op)->v_usecount == 1))
			{
				oncp->op->o_flags |= ON_PURGING;
			}
			vrele(OTOV(oncp->op));  

			crfree(oncp->cred); 
			bzero(DATA_PART(oncp),DATA_SIZE);
		}
	}

	for (i = 0; i < odync_hashsize; i++)
	    odynchash[i].length = 0;
}

/*
 * Debugging routines
 */

/* 
 * This routine should print out all the hash chains to the console.
 */
void
print_odync()
{
	int hash;
	register struct odycache *oncp;

	for (hash = 0; hash < odync_hashsize; hash++) {
		myprintf(("\nhash %d\n",hash));

		for (oncp = odynchash[hash].hash_next; 
		     oncp != (struct odycache *)&odynchash[hash];
		     oncp = oncp->hash_next) {
			myprintf(("op 0x%x dop 0x%x cred 0x%x name %s\n",
				  (int)oncp->op, (int)oncp->dop,
				  (int)oncp->cred, ONCP_NAME(oncp)));
		     }
	}
}

void
odync_gather_stats()
{
    int i, max = 0, sum = 0, temp, zeros = 0, ave, n;

	for (i = 0; i < odync_hashsize; i++) {
	  if (odynchash[i].length) {
	    sum += odynchash[i].length;
	  } else {
	    zeros++;
	  }

	  if (odynchash[i].length > max)
	    max = odynchash[i].length;
	}

/*
 * When computing the Arithmetic mean, only count slots which 
 * are not empty in the distribution.
 */
        odync_stat.Sum_bucket_len = sum;
        odync_stat.Num_zero_len = zeros;
        odync_stat.Max_bucket_len = max;

	if ((n = odync_hashsize - zeros) > 0) 
	  ave = sum / n;
	else
	  ave = 0;

	sum = 0;
	for (i = 0; i < odync_hashsize; i++) {
	  if (odynchash[i].length) {
	    temp = odynchash[i].length - ave;
	    sum += temp * temp;
	  }
	}
        odync_stat.Sum2_bucket_len = sum;
}

/*
 * The purpose of this routine is to allow the hash and cache sizes to be
 * changed dynamically. This should only be used in controlled environments,
 * it makes no effort to lock other users from accessing the cache while it
 * is in an improper state (except by turning the cache off).
 */
int
odync_resize(hashsize, heapsize, dcstat)
    int hashsize; 
    int heapsize;
    enum dc_status dcstat;
{
    if ((hashsize % 2) || (heapsize % 2)) { /* Illegal hash or cache sizes */
	return(EINVAL);
    }                 
    
    odync_use = 0;                       /* Turn the cache off */
    
    odync_flush(dcstat);	/* free any cnodes in the cache */
    
    /* WARNING: free must happen *before* size is reset */
    ODY_FREE(odyncheap,TOTAL_CACHE_SIZE);
    ODY_FREE(odynchash,TOTAL_HASH_SIZE);
    
    odync_hashsize = hashsize;
    odync_size = heapsize;
    
    odync_init();                        /* Set up a cache with the new size */
    
    odync_use = 1;                       /* Turn the cache back on */
    return(0);
}

#define DEBUG
#ifdef	DEBUG
char odync_name_buf[ODY_MAXNAMLEN+1];

void
odync_name(struct onode *op)
{
	struct odycache *oncp, *noncp;
	int i;

	if (odync_use == 0)			/* Cache is off */
		return;

	for (i = 0; i < odync_hashsize; i++) {
		for (oncp = odynchash[i].hash_next; 
		     oncp != (struct odycache *)&odynchash[i];
		     oncp = noncp) {
			noncp = oncp->hash_next;
			if (oncp->op == op) {
			    if (oncp->namelen > ODYNC_NAMELEN)
				bcopy(oncp->n.longname, odync_name_buf,
				      oncp->namelen);
			    else
				bcopy(oncp->n.name, odync_name_buf,
				      oncp->namelen);
			    odync_name_buf[oncp->namelen] = 0;
				printf(" is %s (%p,%p)@%p",
				       odync_name_buf, oncp->op,
				       oncp->dop, oncp);
			}

		}
	}
}
#endif
