#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * ody_sys.c
 *
 * Implement the Odyssey extensions to the NetBSD/UNIX API.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/vnode.h>
#include <sys/ucred.h>
#include <sys/mount.h>
#include <sys/mbuf.h>
#include <sys/select.h>
#include <sys/namei.h>
#include <sys/proc.h>
#include <sys/file.h>
#include <sys/filedesc.h>

#include "include/odyssey.h"
#include "kernel/vfs/includes/ody.h"
#include "kernel/vfs/includes/onode.h"
#include "kernel/vfs/includes/odyapi.h"
#include "kernel/vfs/includes/ody_viceroy.h"
#include "kernel/vfs/includes/ody_sys.h"

int
ody_req (p, uap, retval)
	struct proc *p;
	struct ody_req_args *uap;
	int *retval;
{
	struct ody_reqargs  ora;
	struct nameidata    nd;
	int                 error;
	int                 rc;

	error = copyin(uap->ora, &ora, sizeof(struct ody_reqargs));
	if (error) return error;

	NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_USERSPACE, uap->path, p);
	
	if ((error = namei(&nd)))
		return (error);

	if (nd.ni_vp->v_tag != VT_ODYSSEY)
		return (EOPNOTSUPP);

	error = viceroy_request(vtomi(nd.ni_vp), &(VTOO(nd.ni_vp)->o_fid),
				ora.ora_resource, 
				ora.ora_gen_intf,
				ora.ora_obj_type, 
				ora.ora_type_intf, 
				ora.ora_low, 
				ora.ora_high, 
				p->p_ucred, p,
				&ora.ora_cancelled, 
				&ora.ora_value, 
				&ora.ora_result);
	vput(nd.ni_vp);

	if (error && (error != ODY_ENOTINWINDOW))
	    return (error);

	rc = copyout(&ora, uap->ora, sizeof(struct ody_reqargs));
	if (rc) panic("__ody_req: copyin'd, but can't copyout");

	return error;
}

int
ody_freq (p, uap, retval)
	struct proc *p;
	struct ody_freq_args *uap;
	int *retval;
{
	struct ody_reqargs  ora;
	int              error;
	int              rc;
	int              fd = uap->fd;
	struct filedesc  *fdp = p->p_fd;
	struct file      *fp;
	
	error = copyin(uap->ora, &ora, sizeof(struct ody_reqargs));
	if (error) return error;

	if ((u_int)fd >= fdp->fd_nfiles ||
	    (fp = fdp->fd_ofiles[fd]) == NULL)
		return (EBADF);

	switch (fp->f_type) {
	case DTYPE_VNODE: {
		struct vnode *vp = (struct vnode *)fp->f_data;

		if (vp->v_tag != VT_ODYSSEY)
			error = EOPNOTSUPP;
		else
			error = viceroy_request(vtomi(vp), &(VTOO(vp)->o_fid),
						ora.ora_resource, 
						ora.ora_gen_intf,
						ora.ora_obj_type, 
						ora.ora_type_intf, 
						ora.ora_low, 
						ora.ora_high, 
						p->p_ucred, p,
						&ora.ora_cancelled, 
						&ora.ora_value, 
						&ora.ora_result);
		break;
	}
	default:
#ifdef DIAGNOSTIC
		if (fp->f_type != DTYPE_SOCKET)
			panic("__ody_freg: unkonwn f_type");
#endif
		return (ESPIPE);
		break;
	}

	if (error && (error != ODY_ENOTINWINDOW))
		return(error);

	rc = copyout(&ora, uap->ora, sizeof(struct ody_reqargs));
	if (rc) panic("__ody_req: copied in, but can't out");

	return error;
}

int 
ody_cancel(p, uap, retval)
	struct proc *p;
	struct ody_cancel_args *uap;
	int *retval;
{
	struct nameidata nd;
	int              reqid = uap->reqid;
	int              rc;

	NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_USERSPACE, uap->path, p);
	
	if ((rc = namei(&nd)))
		return rc;

	if (nd.ni_vp->v_tag != VT_ODYSSEY)
		return (EOPNOTSUPP);
	
	/* Needs to be on control object */
	if (!IS_ODY_CTL_VP(nd.ni_vp)) {
		ODYDEBUG(ODY_IOCTL, 
			 myprintf(("ody_cancel error: vp != ctlvp")););
		rc = EOPNOTSUPP;
		goto exit;
	}

	rc = viceroy_cancel(vtomi(nd.ni_vp), reqid, p->p_ucred, p);

exit:
	vput(nd.ni_vp);
	return rc;
}

int
ody_tsop (p, uap, retval)
	struct proc *p;
	struct ody_tsop_args *uap;
	int *retval;
{
	struct nameidata nd;
	struct ody_vers  vers;
        unsigned int     opnum = uap->op;
	size_t           argsz = uap->argsz;
	void             *argp =NULL;
	size_t           retsz =0;
	void             *retp =NULL;
	int              rc;
	caddr_t          probe;
	
	NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_USERSPACE, uap->path, p);
	
	rc = copyin(uap->vers, &vers, sizeof(struct ody_vers));
	if (rc) goto exit;
	
	if (argsz) {
#if 0
		/* Let the VOP_TSOP routine handle this */
		ODY_ALLOC(argp, void *, argsz);
		rc = copyin(uap->arg, argp, argsz);
		if (rc) goto exit;
#else
		/* So pass it the user-space pointer instead */
		argp = uap->arg;
#endif
	}
	
	rc = copyin(uap->retsz, &retsz, sizeof(retsz));
	if (rc) goto exit;

	/* 
	 * Spot check the pending copyout on retsz: for now, copy back
	 * what we just read. 
	 */
	if (suword(uap->retsz, retsz) < 0) {
		rc = EFAULT;
		goto exit;
	}
	
	if (retsz) {
#if 0
		/* We don't want to do an extra internal copy */
		ODY_ALLOC(retp, void *, retsz);
#else
		/* So pass the user-space pointer instead */
		retp = uap->ret;
#endif
		/* Spot check for return buffer's validity */
		probe = uap->ret;
		if (subyte(probe, '\0') < 0) {
			rc = EFAULT;
			goto exit;
		}
		probe += retsz-1;
		if (subyte(probe, '\0') < 0) {
			rc = EFAULT;
			goto exit;
		}
	}
	
	if ((rc = namei(&nd)))
		goto exit;

	if (nd.ni_vp->v_tag != VT_ODYSSEY)
		return (EOPNOTSUPP);

	rc = viceroy_tsop(vtomi(nd.ni_vp), &(VTOO(nd.ni_vp)->o_fid),
			  vers.ov_gs, vers.ov_type, vers.ov_tsvnum,
			  opnum, argsz, argp, p->p_ucred, p,
			  &retsz, retp);
	vput(nd.ni_vp);
	
	if (rc) goto exit;
	
	if (suword(uap->retsz, retsz) < 0) {
		panic("sys_ody_tsop: we could suword, now we can't");
	}

#if 0
	/* We don't do this anymore; VOP_TSOP does */
	rc = copyout(retp, uap->ret, retsz);
#endif
	
 exit:
#if 0
	/* And, we don't malloc these in the kernel */
	if (retp) ODY_FREE(retp, retsz);
	if (argp) ODY_FREE(argp, argsz);
#endif
	return rc;
}

int
ody_ftsop (p, uap, retval)
	struct proc *p;
	struct ody_ftsop_args *uap;
	int *retval;
{
	int              fd = uap->fd;
	struct filedesc  *fdp = p->p_fd;
	struct file      *fp;
	struct ody_vers  vers;
        unsigned int     opnum = uap->op;
	size_t           argsz = uap->argsz;
	void             *argp =NULL;
	size_t           retsz =0;
	void             *retp =NULL;
	int              rc;
	caddr_t          probe;

	if ((u_int)fd >= fdp->fd_nfiles ||
	    (fp = fdp->fd_ofiles[fd]) == NULL)
	{
		rc = EBADF;
		goto exit;
	}

	rc = copyin(uap->vers, &vers, sizeof(struct ody_vers));
	if (rc) goto exit;

	if (argsz) {
#if 0
		/* Let the VOP_TSOP routine handle this */
		ODY_ALLOC(argp, void *, argsz);
		rc = copyin(uap->arg), argp, argsz);
		if (rc) goto exit;
#else
		/* So pass it the user-space pointer instead */
		argp = uap->arg;
#endif
	}

	rc = copyin(uap->retsz, &retsz, sizeof(retsz));
	if (rc) goto exit;

	/* 
	 * Spot check the pending copyout on retsz: for now, copy back
	 * what we just read. 
	 */
	if (suword(uap->retsz, retsz) < 0) {
		rc = EFAULT;
		goto exit;
	}
	
	if (retsz) {
#if 0
		/* We don't want to do an extra internal copy */
		ODY_ALLOC(retp, void *, retsz);
#else
		/* So pass the user-space pointer instead */
		retp = uap->ret;
#endif
		/* Spot check for return buffer's validity */
		probe = uap->ret;
		if (subyte(probe, '\0') < 0) {
			rc = EFAULT;
			goto exit;
		}
		probe += retsz-1;
		if (subyte(probe, '\0') < 0) {
			rc = EFAULT;
			goto exit;
		}
	}

	switch (fp->f_type) {
	case DTYPE_VNODE: {
		struct vnode *vp = (struct vnode *)fp->f_data;

		if (vp->v_tag != VT_ODYSSEY)
			rc = EOPNOTSUPP;
		else
			rc = viceroy_tsop(vtomi(vp), &(VTOO(vp)->o_fid),
					  vers.ov_gs, vers.ov_type,
					  vers.ov_tsvnum, opnum,
					  argsz, argp, p->p_ucred, p,
					  &retsz, retp);
		break;
	}
	default:
#ifdef DIAGNOSTIC
		if (fp->f_type != DTYPE_SOCKET)
			panic("ody_ftsop: unkonwn f_type");
#endif
		rc = ESPIPE;
		break;
	}
	if (rc) goto exit;

	if (suword(uap->retsz, retsz) < 0) {
		panic("sys_ody_tsop: we could suword, now we can't");
	}
		
#if 0
	/* We don't do this anymore; VOP_TSOP does */
	rc = copyout(retp, uap->ret, retsz);
#endif
	
 exit:
#if 0
	/* And, we don't malloc these in the kernel */
	if (retp) ODY_FREE(retp, retsz);
	if (argp) ODY_FREE(argp, argsz);
#endif
	return rc;
}

