#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Directory operations for Coda filesystem
 * Original version: (C) 1996 P. Braam and M. Callahan
 * Rewritten for Linux 2.1. (C) 1997 Carnegie Mellon University
 * 
 * Carnegie Mellon encourages users to contribute improvements to
 * the Coda project. Contact Peter Braam (coda@cs.cmu.edu).
 */

/*
 * stolen for Odyssey
 */

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/locks.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <linux/string.h>

#include <linux/ody_fs_i.h>
#include <linux/ody_linux.h>
#include <linux/ody_psdev.h>
#include <linux/ody_cache.h>

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

/* Return code munging for lookup, which now can accept *either*
 * directory entries *or* error codes.
 */

#if LINUX_VERSION_CODE >= 0x020207
#define RETURN_TYPE(value)   ERR_PTR(value)
#else
#define RETURN_TYPE(value)   value
#endif

/* dir inode-ops */
static int ody_create(struct inode *dir, struct dentry *new, int mode);
#if LINUX_VERSION_CODE >= 0x020207
static struct dentry *ody_lookup(struct inode *dir, struct dentry *entry);
#else
static int ody_lookup(struct inode *dir, struct dentry *entry);
#endif
static int ody_link(struct dentry *old_dentry, struct inode *dir_inode, 
		     struct dentry *entry);
static int ody_unlink(struct inode *dir_inode, struct dentry *entry);
static int ody_symlink(struct inode *dir_inode, struct dentry *entry,
			const char *symname);
static int ody_mkdir(struct inode *dir_inode, struct dentry *entry, int mode);
static int ody_rmdir(struct inode *dir_inode, struct dentry *entry);
static int ody_rename(struct inode *old_inode, struct dentry *old_dentry, 
                       struct inode *new_inode, struct dentry *new_dentry);

/* dir file-ops */
static int ody_readdir(struct file *file, void *dirent, filldir_t filldir);

/* dentry ops */
#if LINUX_VERSION_CODE >= 0x020208
static int ody_dentry_revalidate(struct dentry *de, int flags);
#else
static int ody_dentry_revalidate(struct dentry *de);
#endif
static void ody_dentry_delete(struct dentry *);

/* support routines */
static int ody_viceroy_readdir(struct file *filp, void *dirent, 
			       filldir_t filldir);
int ody_fsync(struct file *, struct dentry *dentry);
static int ody_refresh_inode(struct dentry *dentry);

struct dentry_operations ody_dentry_operations =
{
	ody_dentry_revalidate, /* revalidate */
	NULL, /* hash */
	NULL, /* compare */
	ody_dentry_delete /* delete */
};

struct inode_operations ody_dir_inode_operations =
{
	&ody_dir_operations,
	ody_create,	        /* create */
	ody_lookup,	        /* lookup */
	ody_link,	        /* link */
	ody_unlink,            /* unlink */
	ody_symlink,	        /* symlink */
	ody_mkdir,	        /* mkdir */
	ody_rmdir,   	        /* rmdir */
	NULL,		        /* mknod */
	ody_rename,	        /* rename */
	NULL,	                /* readlink */
	NULL,	                /* follow_link */
	NULL,	                /* readpage */
	NULL,		        /* writepage */
	NULL,		        /* bmap */
	NULL,	                /* truncate */
	ody_permission,        /* permission */
	NULL,                   /* smap */
	NULL,                   /* update page */
        ody_revalidate_inode   /* revalidate */
};

struct file_operations ody_dir_operations = {
        NULL,                   /* lseek */
        NULL,                   /* read -- bad  */
        NULL,                   /* write */
        ody_readdir,		/* readdir */
        NULL,                   /* select */
        NULL,                   /* ioctl */
        NULL,                   /* mmap */
        ody_open,		/* open */
	NULL,			/* flush */
        ody_release,		/* release */
	ody_fsync,		/* fsync */
        NULL,                   /* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL			/* lock */
};


/* inode operations for directories */
/* acces routines: lookup, readlink, permission */
#if LINUX_VERSION_CODE >= 0x020207
static struct dentry *ody_lookup(struct inode *dir, struct dentry *entry)
#else
static int ody_lookup(struct inode *dir, struct dentry *entry)
#endif
{
        struct ody_inode_info *dironp;
	struct inode *res_inode = NULL;
	struct ofid resfid;
	int dropme = 0; /* to indicate entry should not be cached */
	int type;
	int error = 0;
	const char *name = entry->d_name.name;
	size_t length = entry->d_name.len;
	
        ENTRY;
        ODEBUG(D_INODE, "name %s, len %d in ino %lx\n", 
	       name, length, dir->i_ino);

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("ody_lookup: inode is NULL or not a directory\n");
		return RETURN_TYPE(-ENOTDIR);
	}

	dironp = ITOO(dir);

	if ( length > ODY_MAXNAMLEN ) {
	        printk("name too long: lookup, %s (%*s)\n", 
		       ody_f2s(&dironp->o_fid), length, name);
		return RETURN_TYPE(-ENAMETOOLONG);
	}
	
	ODEBUG(D_INODE, "%*s in %s\n", length, name, 
	       ody_f2s(&dironp->o_fid));

        /* control object, create inode on the fly */
        if (ody_isroot(dir) && ody_iscontrol(name, length)) {
	        error = ody_onode_makectl(&res_inode, dir->i_sb);
		ODEBUG(D_SPECIAL, 
		       "Lookup on CTL object; dir ino %lx, count %d\n", 
		       dir->i_ino, dir->i_count);
                goto exit;
        }

        error = viceroy_lookup(dir->i_sb, &(dironp->o_fid), 
			       (const char *)name, length, &type, &resfid);

	if (!error) {
		if (type & ODY_NOCACHE) {
			type &= (~ODY_NOCACHE);
			ODEBUG(D_INODE, "dropme set for %s\n", 
			       ody_f2s(&resfid));
			dropme = 1;
		}
	    	error = ody_onode_make(&res_inode, &resfid, dir->i_sb);
		if (error) {
			/* XXX: some are positive, some negative */
			return RETURN_TYPE(-error);
		}
	} else if (error != -ENOENT) {
	        ODEBUG(D_INODE, "error for %s(%*s)%d\n",
		       ody_f2s(&dironp->o_fid), length, name, error);
		return RETURN_TYPE(error);
	}
	ODEBUG(D_INODE, "%s is (%s), type %d result %d, dropme %d\n",
	       name, ody_f2s(&resfid), type, error, dropme);

exit:
	entry->d_time = 0;
	entry->d_op = &ody_dentry_operations;
	d_add(entry, res_inode);
	if ( dropme ) {
		d_drop(entry);
		ITOO(res_inode)->o_flags |= ON_VATTR;
	}
        EXIT;
        return RETURN_TYPE(0);
}

int ody_permission(struct inode *inode, int mask)
{
        struct ody_inode_info *cp;
        int error;
	int omask = 0;
 
        ENTRY;

	mask &= O_ACCMODE;
	ODEBUG(D_INODE, "fid: %s mask: %d\n", ody_f2s(&(ITOO(inode)->o_fid)),
	       mask);

	/* no permission needed for this operation */
	if ( mask == 0) {
		EXIT;
		return 0;
	}

	if ( ody_access_cache == 1 ) {
		if ( ody_cache_check(inode, mask) ) {
			return 0; 
		}
	}

        cp = ITOO(inode);
        CHECK_ONODE(cp);

	/* these values are derived from namei.c:open_namei */
	if (mask & 0x1)
		omask = O_A_READ;
	if (mask & 0x2)
		omask = O_A_WRITE;

        ODEBUG(D_INODE, "omask is %o\n", omask);
        error = viceroy_access(inode->i_sb, &(cp->o_fid), mask);
    
        ODEBUG(D_INODE, "fid: %s, ino: %lx (mask: %o) error: %d\n", 
	       ody_f2s(&(cp->o_fid)), inode->i_ino, mask, error);

	if ( error == 0 ) {
		ody_cache_enter(inode, mask);
	}

        return error; 
}



/* creation routines: create, mkdir, link, symlink */

static int ody_create(struct inode *dir, struct dentry *de, int mode)
{
        int error=0;
        struct ody_inode_info *dironp;
	const char *name=de->d_name.name;
	int length=de->d_name.len;
	struct inode *result = NULL;
	struct ofid newfid;
	struct ody_vattr attrs;

	ENTRY;
	ODEBUG(D_INODE, "name: %s, length %d, mode %o\n",name, length, mode);

        if (!dir || !S_ISDIR(dir->i_mode)) {
                printk("ody_create: inode is null or not a directory\n");
                return -ENOENT;
        }

	if (ody_isroot(dir) && ody_iscontrol(name, length))
		return -EPERM;

	dironp = ITOO(dir);
        CHECK_ONODE(dironp);

        if ( length > ODY_MAXNAMLEN ) {
		printk("name too long: create, %s(%s)\n", 
		       ody_f2s(&dironp->o_fid), name);
		return -ENAMETOOLONG;
        }

	error = viceroy_create(dir->i_sb, &(dironp->o_fid), name, length, 
				0, mode, &newfid, &attrs);

        if ( error ) {
		ODEBUG(D_INODE, "create: %s, result %d\n",
		       ody_f2s(&newfid), error); 
		d_drop(de);
		return error;
	}

	error = ody_onode_make(&result, &newfid, dir->i_sb);
	if ( error ) {
		d_drop(de);
		result = NULL;
		return error;
	}

	/* invalidate the directory onode's attributes */
	dironp->o_flags |= ON_VATTR;
	d_instantiate(de, result);
        return 0;
}			     

static int ody_mkdir(struct inode *dir, struct dentry *de, int mode)
{
        struct ody_inode_info *dironp;
	struct inode *inode;
	struct ody_vattr attr;
	const char *name = de->d_name.name;
	int len = de->d_name.len;
	int error;
	struct ofid newfid;


	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("ody_mkdir: inode is NULL or not a directory\n");
		return -ENOENT;
	}

        if ( len > ODY_MAXNAMLEN )
                return -ENAMETOOLONG;

	if (ody_isroot(dir) && ody_iscontrol(name, len))
		return -EPERM;

        dironp = ITOO(dir);
        CHECK_ONODE(dironp);

	ODEBUG(D_INODE, "mkdir %s (len %d) in %s, mode %o.\n", 
	       name, len, ody_f2s(&(dironp->o_fid)), mode);

	attr.va_mode = mode;
	error = viceroy_mkdir(dir->i_sb, &(dironp->o_fid), 
			      name, len, &newfid, &attr);
        
        if ( error ) {
	        ODEBUG(D_INODE, "mkdir error: %s result %d\n", 
		       ody_f2s(&newfid), error); 
		d_drop(de);
                return error;
        }
         
	ODEBUG(D_INODE, "mkdir: new dir has fid %s.\n", 
	       ody_f2s(&newfid)); 

	error = ody_onode_make(&inode, &newfid, dir->i_sb);
	if ( error ) {
		d_drop(de);
		return error;
	}
	
	/* invalidate the directory onode's attributes */
	dironp->o_flags |= ON_VATTR;
	dir->i_nlink++;
	d_instantiate(de, inode);
        return 0;
}

/* try to make de an entry in dir_inodde linked to source_de */ 
static int ody_link(struct dentry *source_de, struct inode *dir_inode, 
		    struct dentry *de)
{
	struct inode *inode = source_de->d_inode;
        const char * name = de->d_name.name;
	int len = de->d_name.len;
        struct ody_inode_info *dir_onp, *onp;
	int error;

        ENTRY;

	if (ody_isroot(dir_inode) && ody_iscontrol(name, len))
		return -EPERM;

        dir_onp = ITOO(dir_inode);
        CHECK_ONODE(dir_onp);
        onp = ITOO(inode);
        CHECK_ONODE(onp);

	ODEBUG(D_INODE, "old: fid: %s\n", ody_f2s(&(onp->o_fid)));
	ODEBUG(D_INODE, "directory: %s\n", ody_f2s(&(dir_onp->o_fid)));

        if ( len > ODY_MAXNAMLEN ) {
                printk("ody_link: name too long. \n");
                return -ENAMETOOLONG;
        }

        error = viceroy_link(dir_inode->i_sb,&(onp->o_fid), &(dir_onp->o_fid), 
			   (const char *)name, len);

	if (  ! error ) { 
		dir_onp->o_flags |= ON_VATTR;
		inode->i_nlink++;
		d_instantiate(de, inode);
	} else {
		d_drop(de);
	}

        ODEBUG(D_INODE, "link result %d\n",error);
	EXIT;
        return(error);
}


static int ody_symlink(struct inode *dir_inode, struct dentry *de,
		       const char *symname)
{
        const char *name = de->d_name.name;
	int len = de->d_name.len;
        struct ody_inode_info *dir_onp = ITOO(dir_inode);
	int symlen;
        int error=0;
        
        ENTRY;

	if (ody_isroot(dir_inode) && ody_iscontrol(name, len))
		return -EPERM;

	if ( len > ODY_MAXNAMLEN )
                return -ENAMETOOLONG;

	symlen = strlen(symname);
	if ( symlen > ODY_MAXPATHLEN )
                return -ENAMETOOLONG;

        ODEBUG(D_INODE, "symname: %s, length: %d\n", symname, symlen);

	/*
	 * This entry is now negative. Since we do not create
	 * an inode for the entry we have to drop it. 
	 */
	d_drop(de);

	error = viceroy_symlink(dir_inode->i_sb, &(dir_onp->o_fid), name, len, 
			      symname, symlen);

	if ( !error ) {
		dir_onp->o_flags |= ON_VATTR;
	}

        ODEBUG(D_INODE, "in symlink result %d\n",error);
        EXIT;
        return error;
}

/* destruction routines: unlink, rmdir */

int ody_unlink(struct inode *dir, struct dentry *de)
{
        struct ody_inode_info *dironp;
        int error;
	const char *name = de->d_name.name;
	int len = de->d_name.len;

	ENTRY;

        dironp = ITOO(dir);
        CHECK_ONODE(dironp);

        ODEBUG(D_INODE, " %s in %s, ino %lx\n", name , 
	       ody_f2s(&(dironp->o_fid)), dir->i_ino);

        /* this file should no longer be in the namecache! */

        error = viceroy_remove(dir->i_sb, &(dironp->o_fid), name, len);

        if ( error ) {
                ODEBUG(D_INODE, "upc returned error %d\n", error);
                return error;
        }

        /* cache management */
	dironp->o_flags |= ON_VATTR;
	
	de->d_inode->i_nlink--;
	d_delete(de);

        return 0;
}

int ody_rmdir(struct inode *dir, struct dentry *de)
{
        struct ody_inode_info *dironp;
	const char *name = de->d_name.name;
	int len = de->d_name.len;
        int error, rehash = 0;

	if (!dir || !S_ISDIR(dir->i_mode)) {
		printk("ody_rmdir: inode is NULL or not a directory\n");
		return -ENOENT;
	}
        dironp = ITOO(dir);
        CHECK_ONODE(dironp);

	if (len > ODY_MAXNAMLEN)
		return -ENAMETOOLONG;

	error = -EBUSY;
	if (de->d_count > 1) {
		/* Attempt to shrink child dentries ... */
		shrink_dcache_parent(de);
		if (de->d_count > 1)
			return error;
	}
	/* Drop the dentry to force a new lookup */
	if (!list_empty(&de->d_hash)) {
		d_drop(de);
		rehash = 1;
	}

	/* update i_nlink and free the inode before unlinking;
	   if rmdir fails a new lookup set i_nlink right.*/
	if (de->d_inode->i_nlink)
		de->d_inode->i_nlink --;
	d_delete(de);

	error = viceroy_rmdir(dir->i_sb, &(dironp->o_fid), name, len);

        if ( error ) {
                ODEBUG(D_INODE, "upc returned error %d\n", error);
                return error;
        }

	if (rehash)
		d_add(de, NULL);
	/* XXX how can mtime be set? */

        return 0;
}

/* rename */
static int ody_rename(struct inode *old_dir, struct dentry *old_dentry, 
		      struct inode *new_dir, struct dentry *new_dentry)
{
        const char *old_name = old_dentry->d_name.name;
        const char *new_name = new_dentry->d_name.name;
	int old_length = old_dentry->d_name.len;
	int new_length = new_dentry->d_name.len;
	struct inode *old_inode = old_dentry->d_inode;
	struct inode *new_inode = new_dentry->d_inode;
        struct ody_inode_info *new_onp, *old_onp;
        int error, rehash = 0, update = 1;

	ENTRY;

        if ( (old_length > ODY_MAXNAMLEN) || new_length > ODY_MAXNAMLEN ) {
                return -ENAMETOOLONG;
        }

        old_onp = ITOO(old_dir);
        new_onp = ITOO(new_dir);

	/* make sure target is not in use */
	if ( new_dentry->d_count > 1 && !list_empty(&new_dentry->d_subdirs) )
		shrink_dcache_parent(new_dentry);
	if ( new_dentry->d_count >1 )
		return -EBUSY;

        ODEBUG(D_INODE, "old: %s, (%d length, %d strlen), new: %s (%d length, %d strlen).\n", old_name, old_length, strlen(old_name), new_name, new_length, strlen(new_name));

	/* XXX need to force creation of new inode in some cases XXX */
	/*     figure out how this matters in odyssey */


	/* if moving a directory, clean the dcache */
	if ((S_ISDIR(old_inode->i_mode) & old_dentry->d_count) > 1) 
		shrink_dcache_parent(old_dentry);

	if (old_dentry->d_count > 1) {
		return -EBUSY;
	}

	if (new_dentry->d_count > 1) {
		return -EBUSY;
	}

	d_drop(old_dentry);

	if (!list_empty(&new_dentry->d_hash)) {
	        d_drop(new_dentry);
		rehash = update;
	}

	if (new_inode)
		d_delete(new_dentry);

        error = viceroy_rename(old_dir->i_sb, &(old_onp->o_fid), 
			     &(new_onp->o_fid), old_length, new_length, 
			     (const char *) old_name, (const char *)new_name);

	if (rehash) {
		d_add(new_dentry, new_inode);
	}
	if (update)
		d_move(old_dentry, new_dentry);

	ODEBUG(D_INODE, "result %d\n", error);

	EXIT;
	return 0;
}



/* file operations for directories */
int ody_readdir(struct file *file, void *dirent,  filldir_t filldir)
{
        int result = 0;
        struct ody_inode_info *onp;
        struct file open_file;
	struct dentry open_dentry;
	struct inode *inode=file->f_dentry->d_inode;

        ENTRY;

        ODEBUG(D_FILE, "READDIR inode number: %lx.\n", inode->i_ino);

        if (!inode || !inode->i_sb || !S_ISDIR(inode->i_mode)) {
                printk("ody_readdir: inode is NULL or not a directory\n");
                return -EBADF;
        }

        onp = ITOO(inode);
        CHECK_ONODE(onp);
        
	if (!IS_UNCACHEABLE(onp)) {
		ody_prepare_openfile(inode, file, onp->o_ovp, &open_file,
				     &open_dentry);
		if ( S_ISREG(onp->o_ovp->i_mode) ) {
			printk("ody_readdir on S_ISREG file!\n");
			return -EINVAL;
		} else {
			/* potemkin case: we are handed a directory inode */
			result = open_file.f_op->readdir(&open_file, dirent,
							 filldir);
		}
		ody_restore_odyfile(inode, file, onp->o_ovp, &open_file);
	} else {
		/* we must read get the ody_dirents from the viceroy */
		result = ody_viceroy_readdir(file, dirent, filldir);
	}

	return result;
        EXIT;
}

/* ask venus to cache the file and return the inode of the container file,
   put this inode pointer in the onode for future read/writes */
int ody_open(struct inode *i, struct file *f)
{
        ino_t ino;
	dev_t dev;
	int vflags;
        struct ody_inode_info *onp;
        int error = 0;
	int uncacheable;
        struct inode *cont_inode = NULL;
        unsigned short flags = f->f_flags & (~O_EXCL);
	unsigned short ody_flags = ody_flags_to_oflags(flags);

        ENTRY;
        
        ODEBUG(D_SPECIAL, "OPEN inode number: %lx, flags %o.\n", 
	       f->f_dentry->d_inode->i_ino, flags);

        onp = ITOO(i);

	error = viceroy_open(i->i_sb, &(onp->o_fid), ody_flags, &ino, &dev,
			     &vflags);
	if (error) {
		ODEBUG(D_FILE,
		       "venus: dev %d, inode %ld, flags %d, out->result %d\n",
		       dev, ino, vflags, error);
		return error;
	}

	uncacheable = vflags & ON_UNCACHEABLE;

        /* ody_upcall returns ino number of cached object, get inode */
	if (uncacheable) {
		ODEBUG(D_FILE, "uncacheable onode opened: %s\n",
		       ody_f2s(&onp->o_fid));
	} else {
		ODEBUG(D_FILE, "cache file dev %d, ino %ld\n", dev, ino);
	}
	
	if (onp->o_ocount > 0) {
		if (uncacheable != (onp->o_flags & ON_UNCACHEABLE)) {
			printk("ody_open: UNCACHEABLE clash fid %s\n",
			       ody_f2s(&onp->o_fid));
			return -EINVAL;
		}
	} else
		if (uncacheable)
			onp->o_flags |= ON_UNCACHEABLE;
		else
			onp->o_flags &= ~ON_UNCACHEABLE;

	if (!uncacheable) {
		error = ody_inode_grab(dev, ino, &cont_inode);
			
		if ( error || !cont_inode ){
			printk("ody_open: ody_inode_grab error %d.", error);
			if (cont_inode)
				iput(cont_inode);
			return error;
		}

		if (onp->o_ovp) {
			iput(onp->o_ovp);
			onp->o_ovp = NULL;
		}

		onp->o_ovp = cont_inode; 
		onp->o_ocount++;
		
		/* XXX coda don't have this! XXX */
		if (flags & (O_WRONLY | O_RDWR)) {
			onp->o_owrite++;
			onp->o_flags &= ~ON_VATTR;
		}

		ODEBUG(D_FILE, "result %d, ody i->i_count is %d for ino %lx\n",
		       error, i->i_count, i->i_ino);
		ODEBUG(D_FILE, "cache ino: %lx, count %d, ops %x\n", 
		       onp->o_ovp->i_ino, onp->o_ovp->i_count,
		       (int)(onp->o_ovp->i_op));
	} else {
		i->i_count++;
		onp->o_ocount++;

		if (flags & (O_WRONLY | O_RDWR)) {
			onp->o_owrite++;
			onp->o_flags &= ~ON_VATTR;
		}
	}

        EXIT;
        return 0;
}

int ody_release(struct inode *i, struct file *f)
{
        struct ody_inode_info *onp;
        int error;
        unsigned short flags = f->f_flags & (~O_EXCL);
	unsigned short oflags = ody_flags_to_oflags(flags);

        ENTRY;

        onp =ITOO(i);
        CHECK_ONODE(onp);
        ODEBUG(D_FILE,  
	       "RELEASE ody (ino %lx, ct %d) cache (ino %lx, ct %d)\n",
               i->i_ino, i->i_count, (onp->o_ovp ? onp->o_ovp->i_ino : 0),
               (onp->o_ovp ? onp->o_ovp->i_count : -99));


        /* even when o_ocount=0 we cannot put o_ovp to
         * NULL since the file may be mmapped.
	 * See code in inode.c (ody_put_inode) for
	 * further handling of close.
	 */

        --onp->o_ocount;

	if (IS_UNCACHEABLE(onp))
		i->i_count--;

        if (flags & (O_WRONLY | O_RDWR)) {
                --onp->o_owrite;
        }

	error = viceroy_release(i->i_sb, &(onp->o_fid), oflags);

        ODEBUG(D_FILE, "ody_release: result: %d\n", error);
        return error;
}

/* support routines */
/* 
 * this structure is manipulated by filldir in vfs layer.
 * the count holds the remaining amount of space in the getdents buffer,
 * beyond the current_dir pointer.
 */

struct getdents_callback {
	struct linux_dirent * current_dir;
	struct linux_dirent * previous;
	int count;
	int error;
};

static int ody_viceroy_readdir(struct file *filp, void *getdent, 
			       filldir_t filldir)
{
        int result = 0,  offset, count, pos, error = 0;
	int errfill;
        caddr_t buff = NULL;
        struct ody_dirent *vdirent;
        struct getdents_callback *dents_callback;
        int string_offset;
	int size, len;
        char debug[255];
	struct inode *ino = filp->f_dentry->d_inode;

        ENTRY;        

        ODEBUG(D_FILE, "ody_viceroy_readdir: inode %lx.\n", ino->i_ino);

        /* we also need the ofset of the string in the dirent struct */
        string_offset = sizeof ( char )* 2  + sizeof(unsigned int) + 
                        sizeof(unsigned short);

        dents_callback = (struct getdents_callback *) getdent;

        size = count =  dents_callback->count;
        ODY_ALLOC(buff, void *, size);
        if ( ! buff ) { 
                printk("ody_viceroy_readdir: out of memory.\n");
                return -ENOMEM;
        }

        /* we use this routine to read the ody_dirents into our buffer */
	result = viceroy_readdir(ino->i_sb, &(ITOO(ino)->o_fid),
				 count, filp->f_pos, buff, &len);
        if ( result ) {
                printk("ody_viceroy_readdir: cannot read directory %d.\n",
		       result);
                error = result;
                goto exit;
        }
	result = len;

        /* Parse and write into user space. Filldir tells us when done! */
        offset = filp->f_pos;
        pos = 0;
        ODEBUG(D_FILE, "offset %d, count %d.\n", offset, count);

        while ( pos + string_offset < result ) {
                vdirent = (struct ody_dirent *) (buff + pos);

                /* test if the name is fully in the buffer */
                if ( pos + string_offset + (int) vdirent->d_namlen >= result ){
                        break;
                }
                
                /* now we are certain that we can read the entry from buff */

                /* for debugging, get the string out */
                memcpy(debug, vdirent->d_name, vdirent->d_namlen);
                *(debug + vdirent->d_namlen) = '\0';

                /* if we don't have a null entry, copy it */
                if ( vdirent->d_fileno ) {
                        int namlen  = vdirent->d_namlen;
                        off_t offs  = filp->f_pos; 
                        ino_t ino   = vdirent->d_fileno;
                        char *name  = vdirent->d_name;
                        /* adjust count */
                        count = dents_callback->count;

			errfill = filldir(dents_callback,  name, namlen, 
					  offs, ino); 
ODEBUG(D_FILE, "ino %lx, namlen %d, reclen %d, type %d, pos %d, string_offs %d, name %s, offset %d, count %d.\n", vdirent->d_fileno, vdirent->d_namlen, vdirent->d_reclen, vdirent->d_type, pos,  string_offset, debug, (u_int) offs, dents_callback->count);

		      /* errfill means no space for filling in this round */
                      if ( errfill < 0 ) break;
                }
                /* next one */
                filp->f_pos += (unsigned int) vdirent->d_reclen;
                pos += (unsigned int) vdirent->d_reclen;
        } 

exit:
        ODY_FREE(buff, size);
        return error;
}

/* called when a cache lookup succeeds */
#if LINUX_VERSION_CODE >= 0x020208
static int ody_dentry_revalidate(struct dentry *de, int flags)
#else
static int ody_dentry_revalidate(struct dentry *de)
#endif
{
	int valid = 1;
	struct inode *inode = de->d_inode;
	struct ody_inode_info *oii;
	ENTRY;

	if (inode) {
		if (is_bad_inode(inode))
			return 0;
		oii = ITOO(de->d_inode);
		if (oii->o_flags & ON_PURGE)
			valid = 0;
	}
	return valid ||  ody_isroot(de->d_inode);
}

/*
 * This is the callback from dput() when d_count is going to 0.
 * We use this to unhash dentries with bad inodes.
 */
static void ody_dentry_delete(struct dentry * dentry)
{
	int flags;

	if (!dentry->d_inode) 
		return ;

	flags =  (ITOO(dentry->d_inode)->o_flags) & ON_PURGE;
	if (is_bad_inode(dentry->d_inode) || flags) {
		ODEBUG(D_DOWNCALL, "bad inode, unhashing %s/%s, %ld\n", 
		       dentry->d_parent->d_name.name, dentry->d_name.name,
		       dentry->d_inode->i_ino);
		d_drop(dentry);
	}
}

static int ody_refresh_inode(struct dentry *dentry)
{
	struct ody_vattr attr;
	int error;
	int old_mode;
	ino_t old_ino;
	struct inode *inode = dentry->d_inode;
	struct ody_inode_info *oii = ITOO(inode);
	
	ENTRY;
	error = viceroy_getattr(inode->i_sb, &(oii->o_fid), &attr);
	if ( error ) { 
		make_bad_inode(inode);
		return -EIO;
	}

	/* this inode may be lost if:
	   - its type changed
	   - its ino changed 
	*/
	old_mode = inode->i_mode;
	old_ino = inode->i_ino;
	ody_vattr_to_iattr(inode, &attr);

	if ((inode->i_ino != old_ino) ||
	    ((old_mode & S_IFMT) != (inode->i_mode & S_IFMT))) {
		make_bad_inode(inode);
		inode->i_mode = old_mode;
		return -EIO;
	}
	
	oii->o_flags &= ~ON_VATTR;
	return 0;
}


/*
 * This is called when we want to check if the inode has
 * changed on the server.  Ody makes this easy since the
 * cache manager Venus issues a downcall to the kernel when this 
 * happens 
 */

int ody_revalidate_inode(struct dentry *dentry)
{
	int error = 0;
	struct ody_inode_info *oii = ITOO(dentry->d_inode);

	ENTRY;
	ODEBUG(D_INODE, "revalidating: %*s/%*s\n", 
	       dentry->d_name.len, dentry->d_name.name,
	       dentry->d_parent->d_name.len, dentry->d_parent->d_name.name);

	if ( oii->o_flags & (ON_VATTR | ON_PURGE )) {
		error = ody_refresh_inode(dentry);
	}

	return error;
}

