#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <linux/fs.h>

#include <linux/ody_fs_i.h>
#include <linux/ody_linux.h>
#include <linux/ody_psdev.h>
#include <linux/ody_cache.h>

#include "kernel/sets/common/sets_internal.h"
#include "kernel/sets/common/sets_domain.h"

extern struct dentry_operations ody_dentry_operations;

int ody_expand(struct dentry *dvp,
	       struct dentry **vpp,
	       char *nm,
	       struct task_struct *p)
{
	struct inode *inode = dvp->d_inode;
	struct ody_inode_info *dironp = ITOO(inode);
	int len = strlen(nm);
	struct dentry *entry;
	struct ofid resfid;
	struct inode *res_inode = NULL;
	struct qstr this;
	unsigned long hash;
	char c;
	int type;
	int error;

	if (!S_ISDIR(inode->i_mode)) {
		printk("ody_expand: inode is not a directory\n");
		return -ENOTDIR;
	}

	if (len > ODY_MAXNAMLEN ) {
	        printk("name too long: expand, %s (%*s)\n", 
		       ody_f2s(&dironp->o_fid), len, nm);
		return -ENAMETOOLONG;
	}
	
	ODEBUG(D_INODE, "%*s in %s\n", len, nm, ody_f2s(&dironp->o_fid));

        error = viceroy_expand(inode->i_sb, &(dironp->o_fid), 
			       (const char *)nm, len, &type, &resfid);

	if (!error) {
		error = ody_onode_make(&res_inode, &resfid, inode->i_sb);
		if (error)
			return -error;
	} else {
	        ODEBUG(D_INODE, "expand error for %s(%*s)%d\n",
		       ody_f2s(&dironp->o_fid), len, nm, error);
		return error;
	}

	this.name = nm;
	c = *nm;

	hash = init_name_hash();
	do {
		hash = partial_name_hash(c, hash);
		c = *++nm;
	} while (c && (c != '/'));
	this.len = nm - (const char *) this.name;
	this.hash = end_name_hash(hash);
	entry = d_alloc(dvp, &this);
	entry->d_time = 0;
	entry->d_op = &ody_dentry_operations;
	d_add(entry, res_inode);
	d_drop(entry);		/* don't cache cursors */
	ITOO(res_inode)->o_flags |= ON_VATTR;
	ITOO(res_inode)->o_flags |= ON_CURSOR;
	*vpp = entry;

	EXIT;
	return 0;
}

int ody_cursexpand(struct file *f,
		   size_t *bufsize,
		   char **buf,
		   struct task_struct *p)
{
	struct ody_inode_info *onode = ITOO(f->f_dentry->d_inode);
	int error;

	if (!onode->o_flags & ON_CURSOR) {
		printk("ody_cursexpand called on non-cursor.\n");
		return -EINVAL;
	}

	error = viceroy_cursexpand(f->f_dentry->d_inode->i_sb,
				   &(onode->o_fid),
				   buf, bufsize);

	return -error;
}

int ody_prefetch(struct dentry *de, int flags, struct task_struct *p,
		 struct file **fpp)
{
    struct file *f;
    struct inode *inode = de->d_inode;
    struct ody_inode_info *onp = ITOO(inode);
    unsigned short ody_flags = ody_flags_to_oflags(flags&~O_EXCL);
    struct inode *cont_inode = NULL;
    int error = 0;
    int vflags;
    ino_t ino;
    dev_t dev;

    ENTRY;

    f = get_empty_filp();
    if (!f) {
	error = -ENFILE;
	goto out;
    }
    
    f->f_flags = flags;
    f->f_mode = (flags+1) & O_ACCMODE;
    f->f_dentry = de;
    f->f_pos = 0;
    f->f_reada = 0;
    f->f_op = NULL;
    if (inode->i_op)
	f->f_op = inode->i_op->default_file_ops;

    error = viceroy_prefetch(de->d_inode->i_sb, &(onp->o_fid),
			     ody_flags, &ino, &dev, &vflags);

    if (error) {
	    put_filp(f);
	    printk("viceroy_prefetch error dev %d, inode %ld, flags %d, "
		   "out->result %d\n",
		   dev, ino, vflags, error);
	    return error;
    }

    if (vflags & ON_UNCACHEABLE) {
	    onp->o_flags |= ON_UNCACHEABLE;
	    inode->i_count++;
	    onp->o_ocount++;
	    
	    if (flags & (O_WRONLY | O_RDWR)) {
		    onp->o_owrite++;
		    onp->o_flags &= ~ON_VATTR;
	    }
    } else {
	    onp->o_flags &= ~ON_UNCACHEABLE;
	    error = ody_inode_grab(dev, ino, &cont_inode);
	    
	    if ( error || !cont_inode ){
		    printk("ody_prefetch: ody_inode_grab error %d.", error);
		    if (cont_inode)
			    iput(cont_inode);
		    return error;
	    }

	    if (onp->o_ovp) {
		    iput(onp->o_ovp);
		    onp->o_ovp = NULL;
	    }
	    
	    onp->o_ovp = cont_inode; 
	    onp->o_ocount++;
	    
	    /* XXX coda don't have this! XXX */
	    if (flags & (O_WRONLY | O_RDWR)) {
		    onp->o_owrite++;
		    onp->o_flags &= ~ON_VATTR;
	    }
	    
	    ODEBUG(D_FILE, "res %d, ody inode->i_ct is %d for ino %lx\n",
		   error, inode->i_count, inode->i_ino);
	    ODEBUG(D_FILE, "cache ino: %lx, count %d, ops %x\n", 
		   onp->o_ovp->i_ino, onp->o_ovp->i_count,
		   (int)(onp->o_ovp->i_op));
    }

    *fpp = f;

 out:
    EXIT;
    return 0;
}
