#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* eval.c

   Find the best fidelity level that will satisfy our resource constraints
   */

#include <stdio.h>
#include <math.h>
#include "multif.h"

double eval_rsrc_func(rfunc_t func, double size, double fidelity)
{
    double f1 = func.coeff;
    double f2 = pow(fidelity, func.f_exp);
    double f3 = pow(size, func.n_exp);
    double f4 = pow(log(size), func.logn_exp);
    return f1 * f2 * f3 * f4;
}

/* Find the max fidelity (if one exists between fmin and fmax) for which
   resource usage <= rmax. */

/* NOTE: We assume that rfunc is monotonically
   increasing, and do a binary search + linear interpolation, until
   the fidelity is within our error limits */

/* NOTE 2: Currently we only deal with a single resource usage function
   for each resource; later we might want to extend to multiple ones
   which are summed (for CPU, etc) or max'ed (latency of concurrent ops)
   */
   
int eval_rsrc_constraint(rfunc_t rfunc, double rmax, double size,
			 double fmin, double fmax, double ferr, double *fout)
{
    double rhigh = eval_rsrc_func(rfunc, size, fmax);
    double rlow = eval_rsrc_func(rfunc, size, fmin);

    if (rlow > rmax) return -1; /* can't be done */

    if (rhigh <= rmax) { /* We already satisfy the constraint */
	*fout = fmax;
	return 0;
    }

    while (fmax - fmin > ferr) {
	double newf = (fmin + fmax) / 2;
	double rnew = eval_rsrc_func(rfunc, size, newf);
	if (rnew < rmax) {
	    rlow = rnew;
	    fmin = newf;
	}
	else {
	    rhigh = rnew;
	    fmax = newf;
	}
    }

    *fout = fmin;
    return 0;
}

/* A dirt-simple function to estimate the available CPU fraction: 
   1/(1+loadavg), i.e. we assume that the loadavg represents the number of
   runnable processes (not including us) and that we share the CPU equally.
   */
double avail_cpu_fraction(void)
{
    FILE *fp = fopen("/proc/loadavg", "r");
    double loadavg;
    fscanf(fp, "%lf", &loadavg);
    fclose(fp);
    loadavg = 1.0/(1.0+loadavg);
    fprintf(stderr, "CPU availability is %f\n", loadavg);
    return loadavg;
}

/* Find the best fidelity attainable in a particular fidelity mode, subject
   to ALL the resource constraints.
  
   NOTE: Currently we assume there is only one resource usage entry
   per resource per fidelity mode (see previous comment).

   Currently we only support latency and CPU constraints
   */

int eval_fidelity_mode(rsrc_constraint *rsrcp, int numconstraints,
		       fidval *fp, double size, double *bestf)
{
    int retval;

    /* Do each constraint in turn */
    /* For now we make some very restrictive assumptions */
    /* We only handle latency constraints, and CPU usage functions */
    ASSERT(numconstraints == 1);
    ASSERT(rsrcp->rc_id == RU_LATENCY);
    ASSERT(fp->fv_numrsrcs == 1);
    ASSERT(fp->fv_rusages->ru_id == RU_LCPU);

    fprintf(stderr,"Evaluating mode %s with latency constraint %f ms\n",
	    fp->fv_name, rsrcp->rc_max);

    /* Maximum allowed usage of CPU is the maximum allowed latency * the
       percentage of CPU that we have */
    retval = eval_rsrc_constraint(fp->fv_rusages->ru_consumption,
				rsrcp->rc_max * avail_cpu_fraction(),
				size, fp->fv_utility[0], fp->fv_utility[1],
				0.001, bestf);

    if (retval < 0)
	fprintf(stderr,"Fidelity mode %s cannot meet constraints\n",
		fp->fv_name);
    else
	fprintf(stderr,"Fidelity mode %s can offer a fidelity of %f\n",
		fp->fv_name, *bestf);

    return retval;
}

int find_best_fidelity(fidval *fp, int numfids, rsrc_constraint *rsrcp,
		       int numconstraints, double size, double *bestf)
{
    double best;
    int result = -1;
    int i;

    for (i = 0; i < numfids; i++)
	if (eval_fidelity_mode(rsrcp, numconstraints, fp+i, size, &best) == 0)
	    if ((result < 0) || (best > *bestf)) {
		result = i;
		*bestf = best;
	    }

    return result;
}
