#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
  multif.h
 
  Data type and function definitions for the multifidelity API
*/

#ifndef _MULTIF_H_
#define _MULTIF_H_

#define MULTIF_VERSION 1

#define IN
#define OUT

/* a fidelity parameter: could be either a knob or a switch */
/* knob values all range from 0 to 1; num_values is the number
   of values if discrete, 0 if continuous */;
/* switches are just a set of integers (0 .. num_values-1), representing
   an unordered set of values */
/* Some knobs/switches may be outside our control -- i.e. they are set
   for us (for example, input data size). These are marked with
   is_settable == 0 */
/* Currently the interface assumes that they are all orthogonal; i.e.
   any combination of values is legal. When we extend to have arbitrary
   mappings from fidelity params to fidelities, we can fix this by having
   all illegal combinations map to an illegal fidelity which maps to
   an infinitely negative utility. */

typedef struct {
    char fp_name[32];
    int is_knob;
    int is_settable;
    double min, max; /* for continuous variables */
    int num_values;  /* for discrete variables */
}
fid_param_t;

/* Value of a fidelity parameter: real number in [0...1] for knobs, integer
   in [0...(n-1)] for switches */
typedef union {
    double knob_val;
    int switch_val;
}
fid_param_val_t;

/* register our fidelity parameters and fidelity metric; returns an
integer which is the optype_id. In the list of fidelity parameters,
all discrete parameters must precede all continuous parameters.
The last parameter is the number of (output) fidelity dimensions.
In this implementation, they are assumed to be identical to the
input fidelity parameters.
*/

int register_fidelity(IN int num_params, IN fid_param_t *params,
		      IN int num_dims, OUT int *opid);

/* all the quality metrics we care about */
typedef enum {
    Q_LATENCY,
    Q_BATTERYLIFE,
    Q_MONEY,
    Q_DISTRACTION,
    Q_LAST
} quality_t;


#define MAX_CONTINUOUS_VARS 5
#define MAX_DISCRETE_VARS 5
#define MAX_POLY_TERMS 5

/* A term in n continuous variables. The value is
   power[0] * var1^power[1] * var2^power[2] *  ...
   * (log var1) ^ power[n] + (log var2 ^ power[n+1])

   So the array should be of length 2n+1 at least
 */
typedef double term_t[2 * MAX_CONTINUOUS_VARS + 1];

/* A polynomial in n variables -- a sum of n terms */

typedef struct {
    int num_terms;
    term_t terms[MAX_POLY_TERMS];
} poly_t;

/* A "case" is a set of values for m discrete variables associated with
   a function of n continuous variables */

typedef struct {
    int discrete_vals[MAX_DISCRETE_VARS];
    poly_t continuous_func;
} casefunc_t;

/* A function of both discrete AND continuous vars is just a list of cases */

typedef struct {
    int num_discrete, num_continuous, num_cases;
    casefunc_t *cases;
} func_t;
    
/* Hint a utility function: the first four parameters are the four
 quality metrics defined above, followed by all the dimensions of
 fidelity */
int hint_utility(IN int optype_id, IN func_t utility_func);

/* Specialized form of the utility function: specify  a constraint on
   one quality metric. Optionally register a callback for when the
   constraint is violated. Constraints are upper bounds for latency,
   money, user distraction and lower bound for battery lifetime */

typedef void (*quality_callback_t)(quality_t which_quality, double val);

int hint_constraint(IN int optype_id, IN quality_t which_quality,
		    IN double constraint, IN quality_callback_t callback_fn);

/* rsrc_id enumerates the different types of resources */
typedef enum {
    R_NETBW, R_NETBW_XMIT, R_NETBW_RCV,
    R_LCPU, R_RCPU,
    R_PHYSMEM,
    R_DISKBW, R_DISKSPACE,
    R_CACHESTATE,
    R_LAST
}
rsrc_t;

/* Hint our resource consumption (for everything other than cache state) as
   a function of input size and fidelity parameters */
int hint_resource(IN int optype_id, IN rsrc_t which_resource,
		  IN func_t rsrc_func);


/* Signal beginning of an operation, and query for the values to set the
   fidelity parameters to; returns an integer opid */
int begin_fidelity_op(IN int optype_id, OUT int *opid, IN int num_params,
		      IN OUT fid_param_val_t *fidparams);


/* Signal end of an operation, and the fidelity achieved */

int end_fidelity_op(IN int optype_id, IN int opid,
		    IN int num_params, IN fid_param_val_t *fidparams,
		    IN int num_fids, IN fid_param_val_t *output_fids);

#endif /* _MULTIF_H_ */
