#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <errno.h>
#include <stdlib.h>

#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "kernel/vfs/includes/ody.h"

#include "multif.h"
#include "multif_internal.h"


static struct optype *op_types[MAX_OPTYPES];

int multif_register_fidelity(num_params, params, num_fids, optype_idp, cred,
			     pid)
     int num_params;
     fid_param_t *params;
     int num_fids;
     int *optype_idp;
     struct ody_cred *cred;
     u_short pid;

{
    int i;
    struct optype *opp;
    int optype_id;
    int num_continuous = 0;

    for (i = 0; i < num_params; i++) {
	if (params[i].is_knob) num_continuous++;
	else if (num_continuous) {
	    ds_thread_log(0,"multif_register_fidelity: All discrete params "
			  "must precede continuous ones\n");
	    FREE(params);
	    return EINVAL;
	}
    }

    /* We restrict the fidelity parameters to be identical to the output
       fidelity */

    if (num_params != num_fids) {
	ds_thread_log(0,"multif_register_fidelity: num_fids != num_params\n");
	FREE(params);
	return EINVAL;
    }

    /* NOTDONE: see if this fp_name already exists in our table */

    /* find a free entry */
    for (optype_id = 0; optype_id < MAX_OPTYPES; optype_id++)
	if (!op_types[optype_id]) break;

    if (optype_id == MAX_OPTYPES) {
	FREE(params);
	return ENOMEM;
    }

    ZALLOC(opp, struct optype);
    opp->pid = pid;
    opp->num_params = num_params;
    opp->num_fids = num_fids;
    opp->num_continuous = num_continuous;
    opp->num_discrete = num_params - num_continuous;
    opp->discrete_params = params;
    opp->continuous_params = params + opp->num_discrete;

    op_types[optype_id] = opp;

    ds_thread_log(0, "multif_register_fidelity(%d, %p, %d) returns %d",
		  num_params, params, num_fids, optype_id);
    for (i = 0; i < num_params; i++) {
	if (params[i].is_knob) {
	    if (params[i].num_values)
		ds_thread_log(0,"     knob %s (%lf-%lf continuous%s)",
			      params[i].fp_name, params[i].min, params[i].max,
			      params[i].is_settable?",settable":"");
	    else
		ds_thread_log(0,"     knob %s (%lf-%lf, %d steps%s)",
			      params[i].fp_name, params[i].min, params[i].max,
			      params[i].num_values,
			      params[i].is_settable?",settable":"");
	}
	else
	    ds_thread_log(0,"     switch %s (%d settings%s)",
			  params[i].fp_name, params[i].num_values,
			  params[i].is_settable?",settable":"");
    }
			      
    *optype_idp = optype_id;
    return 0;
}

int multif_hint_constraint(optype_id, which_quality, constraint, cred, pid)
     int optype_id;
     quality_t which_quality;
     double constraint;
     struct ody_cred *cred;
     u_short pid;
{
    if ((optype_id < 0) || (optype_id >= MAX_OPTYPES) || !op_types[optype_id])
	return EINVAL;
    if (op_types[optype_id]->pid != pid) return EPERM;
    if ((which_quality < 0) || (which_quality >= Q_LAST)) return EINVAL;

    /* Currently we only support latency constraints */
    if (which_quality != Q_LATENCY) {
	ds_thread_log(0, "Constraints on quality %d not supported",
		      which_quality);
	return EOPNOTSUPP;
    }

    op_types[optype_id]->constraints[which_quality] = constraint;
    return 0;
}

    
int multif_hint_resource(optype_id, which_resource, rsrc_func, cred, pid)
     int optype_id;
     rsrc_t which_resource;
     func_t rsrc_func;
     struct ody_cred *cred;
     u_short pid;
{
    func_t *funcp;

    if ((optype_id < 0) || (optype_id >= MAX_OPTYPES) || !op_types[optype_id])
	return EINVAL;
    if (op_types[optype_id]->pid != pid) return EPERM;
    if ((which_resource < 0) || (which_resource >= R_LAST)) return EINVAL;

    /* Currently we only support the L_CPU resource */
    if (which_resource != R_LCPU) {
	ds_thread_log(0, "Resource %d not supported", which_resource);
	return EOPNOTSUPP;
    }

    funcp = op_types[optype_id]->rsrc_hints[which_resource];
    if (funcp) {
	ds_thread_log(0, "multif_hint_resource: removing old hint");
	FREE(funcp->cases);
    }
    else
	ALLOC(funcp, func_t);

    *funcp = rsrc_func;
    NALLOC(funcp->cases, casefunc_t, funcp->num_cases);
    memcpy(funcp->cases, rsrc_func.cases,
	   funcp->num_cases * sizeof(casefunc_t));
    op_types[optype_id]->rsrc_hints[which_resource] = funcp;
    return 0;
}

/* XXX test implementation  -- only one op at a time, trivial fidelity
   decisions, only one process */

#if 0
/* return total CPU time (user+system) used by a process, in seconds */
#define JIFFY_TIME 0.01 /* seconds per JIFFY */
static
int
get_cputime(pid, cputime)
     u_short pid;
     double *cputime;
{
    char proc_fname[32];
    FILE *fp;
    int utime, stime;

    sprintf(proc_fname,"/proc/%hu/stat", pid);
    fp = fopen(proc_fname, "r");
    if (!fp) return -1;
    fscanf(fp, "%*d %*s %*c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %d %d",
	   &utime, &stime);
    fclose(fp);
    *cputime = (utime + stime) * JIFFY_TIME;
    return 0;
}

static struct app_info *cur_app = NULL;

static void init_cur_app(u_short pid)
{
    if (!cur_app) {
	int i;
	ZALLOC(cur_app, struct app_info);
	cur_app->ai_opid = -1;
	cur_app->ai_pid = pid;
	for (i = 0; i < RU_NUM_RSRCS; i++)
	    cur_app->ai_rc[i] = -1;
    }
}

#endif

int multif_begin_fidelity_op(optype_id, opidp, num_params, fidparams, cred,
			     pid)
     int optype_id;
     int *opidp;
     int num_params;
     fid_param_val_t *fidparams;
     struct ody_cred *cred;
     u_short pid;
{

    if ((optype_id < 0) || (optype_id >= MAX_OPTYPES) || !op_types[optype_id])
	return EINVAL;
    if (op_types[optype_id]->pid != pid) return EPERM;

    /* XXX: this part is just hacked up for the demo of radiator */
    {
	double latency =op_types[optype_id]->constraints[Q_LATENCY];
	FILE *fp = fopen("/usr/odyssey/etc/rad_fid_lat", "r");
	if (!fp) return ENOENT;
	while (!feof(fp)) {
	    int lat_bound;
	    fscanf(fp, "%d %d %lf", &lat_bound, &fidparams[0].switch_val,
		   &fidparams[1].knob_val);
	    if (latency <= lat_bound) break;
	}
	fclose(fp);
	return 0;
    }

#if 0

    if (op_types[optype_id]->cur_op) return EBUSY;

    

    retval = find_best_fidelity(cur_app->ai_fmp, cur_app->ai_nmodes,
				cur_app->ai_rc, cur_app->ai_rhints,
				input_size, &fidelity);
    if (retval < 0) return EDOM;

    ds_thread_log(0, "multif_begin_fidelity_op: mode %s, fidelity %f",
		  cur_app->ai_fmp[retval].fm_name, fidelity);
    *opid = cur_app->ai_opid = cur_app->ai_next_opid++;
    *fm_id = retval;
    *fm_fidelity = fidelity;

    gettimeofday(&cur_app->ai_op_begin_time, NULL);
    if (get_cputime(pid, &cur_app->ai_op_begin_cputime) < 0) {
	ds_thread_log(0, "Could not read starting CPU time for process %hu\n",
		      pid);
	cur_app->ai_op_begin_cputime = 0;
    }
		      
    return 0;
#endif
}

int multif_end_fidelity_op(optype_id, opid, num_params, fidparams, num_fids,
			   fids, cred, pid)
     int optype_id;
     int opid;
     int num_params;
     fid_param_val_t *fidparams;
     int num_fids;
     fid_param_val_t *fids;
     struct ody_cred *cred;
     u_short pid;
{
    /* XXX for the demo, end_fidelity_op does absolutely nothing */
    return 0;
#if 0
    struct timeval op_end_time, op_time;
    double op_end_cputime, op_cputime;

    if (!cur_app) return EINVAL;
    if (cur_app->ai_pid != pid) return EINVAL;
    if (cur_app->ai_opid != opid) return EINVAL;

    if (get_cputime(pid, &op_end_cputime) < 0) {
	ds_thread_log(0, "Could not read finishing CPU time for process %hu\n",
		      pid);
    }
    op_cputime = op_end_cputime - cur_app->ai_op_begin_cputime;
    gettimeofday(&op_end_time, NULL);
    timersub(&op_end_time, &cur_app->ai_op_begin_time, &op_time);

    ds_thread_log(0,"end_fidelity_op(%d, %f, \"%s\", %f): time %ld.%06ld,"
		  "cpu %.6lf\n", opid, input_size, 
		  cur_app->ai_fmp[fm_id].fm_name, fm_fidelity, op_time.tv_sec,
		  op_time.tv_usec, op_cputime);
    cur_app->ai_opid = -1;
    return 0;
#endif
}




