#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/time.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"

#include <include/viceroy_onode.h>
#include "rsrc.h"
#include "bmod.h"

struct bmod_data bmod;

void
bmod_init ()
{
    ds_hwlock_init (&bmod.bm_lock); /* Initialize lock to guard bmod data */
    bmod.bm_is_modulated = 0;
    bmod.bm_recharge_time.tv_sec = 1000000000;
    bmod.bm_recharge_time.tv_usec = 0;
}

int
bmod_start (int joules, float charge) 
{
    if (joules < 0) return (EINVAL);
    if (charge < 0.0 || charge > 1.0) return (EINVAL);

    if (rsrc_loglevel > 5) {
	ds_thread_log (0, "bmod_start: joules %d charge %f", joules, charge);
    }

    ds_hwlock_write (&bmod.bm_lock);
    bmod.bm_is_modulated = 1;
    bmod.bm_joules_total = (float) joules;
    bmod.bm_joules_left = (float) joules * charge;
    ds_hwlock_release_write (&bmod.bm_lock);

    return (0);
}
		
int 
bmod_stop ()
{
    ds_hwlock_write (&bmod.bm_lock);
    bmod.bm_is_modulated = 0;
    ds_hwlock_release_write (&bmod.bm_lock);

    return (0);
}

int
bmod_drain (float joules)
{
    ds_hwlock_write (&bmod.bm_lock);
    if (bmod.bm_is_modulated) {
	bmod.bm_joules_left -= joules;
	if (bmod.bm_joules_left < 0) {
	    bmod.bm_joules_left = 0;
	}
    }
    if (rsrc_loglevel >= 10) {
	ds_thread_log (0, "battery: joules left %f", bmod.bm_joules_left);
    }
    ds_hwlock_release_write (&bmod.bm_lock);
    
    return (0);
}

int
bmod_set_time_to_recharge (int seconds)
{
    struct timeval tv;

    if (rsrc_loglevel >= 1) {
	ds_thread_log (0, "bmod_set_time_to_recharge: seconds = %d\n", 
		       seconds);
    }
    if (seconds <= 0) return (EINVAL);
	
    ds_hwlock_write (&bmod.bm_lock);
    gettimeofday(&tv, NULL);
    tv.tv_sec += seconds;
    bmod.bm_recharge_time = tv;
    ds_hwlock_release_write (&bmod.bm_lock);
    
    return (0);
}

float 
bmod_get_charge ()
{
    float retval;

    ds_hwlock_read (&bmod.bm_lock);
    if (bmod.bm_is_modulated) {
	retval = bmod.bm_joules_left / bmod.bm_joules_total;
    } else {
	/* XXX - For now - return 1 if unmodulated.  In actuality should make
	   BIOS call to retrieve actual charge. */
	retval = 1.0;
    }

    ds_hwlock_release_read (&bmod.bm_lock);

    return (retval);
}

float 
bmod_get_remaining_joules ()
{
    float retval;

    ds_hwlock_read (&bmod.bm_lock);
    if (bmod.bm_is_modulated) {
	retval = bmod.bm_joules_left;
    } else {
	/* XXX - For now - return large value if unmodulated.  In
	   actuality should make BIOS call to retrieve actual
	   charge. */
	retval = 999999999.0;
    }
    ds_hwlock_release_read (&bmod.bm_lock);

    return (retval);
}

float 
bmod_get_total_joules ()
{
    float retval;

    ds_hwlock_read (&bmod.bm_lock);
    if (bmod.bm_is_modulated) {
	retval = bmod.bm_joules_total;
    } else {
	/* XXX - For now - return large value if unmodulated.  In
	   actuality should make BIOS call to retrieve actual
	   charge. */
	retval = 999999999.0;
    }
    ds_hwlock_release_read (&bmod.bm_lock);

    return (retval);
}

struct timeval 
bmod_get_time_to_recharge ()
{
    struct timeval retval, cur_tv;
    
    ds_hwlock_read (&bmod.bm_lock);
    gettimeofday (&cur_tv, NULL);
    timersub (&bmod.bm_recharge_time, &cur_tv, &retval);
    ds_hwlock_release_read (&bmod.bm_lock);

    return (retval);
}
