#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/stat.h>
#include <sys/queue.h>

#include "include/viceroy_onode.h"
#include "include/odytypes.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/cache_mgr.h"

/* Note: since cache files are never returned to the viceroy, this is
   monotonically increasing. */
unsigned long next_cache_file = 0;

/*
 * cache_init --  initialize type-independent cache-related stuff
 */
void cache_init()
{
    return;
}

/*
 * cache_request -- returns a list of num_cfiles initialized cache_entry
 *                  structures.  Only the tqe_next pointer is valid.
 */
struct cache_entry *cache_request(struct warden *requester,
				  long kilobytes,
				  int num_cfiles)
{
    struct cache_entry *cache_list = NULL, *new_entry;

    /* XXX This early implementation doesn't do "admission control" --
       everyone gets what they ask for. XXX */
    while (num_cfiles--) {
	ALLOC(new_entry, struct cache_entry);
	ASSERT(new_entry);
	new_entry->ce_key = NULL;
	sprintf(new_entry->ce_fname, "V%05lu", next_cache_file++);
	truncate(new_entry->ce_fname, 0);
	new_entry->ce_flags = 0;
	new_entry->ce_data = NULL;
	new_entry->ce_owner = requester;
	new_entry->ce_link.tqe_next = cache_list;
	cache_list = new_entry;
    }

    requester->w_total_size += kilobytes;

    return cache_list;
}

/*
 * cache_open -- mediate access to container files on disk
 */
int cache_open(struct cache_entry *centry,
	       int flags)
{
    int fd;
    struct stat sb;
    int rc;

    /* Do some spying if desired... */
    if (flags & O_O_TRUNC) {
	rc = stat(centry->ce_fname, &sb);
	if (!rc)
	    centry->ce_owner->w_current_size -= (sb.st_blocks >> 1); /* KB */
    }
    fd = open(centry->ce_fname, flags,
	      S_IRWXU | S_IRGRP | S_IROTH);
    return fd;
}

/*
 * cache_close -- maintain bookkeeping info on warden, using the file
 *                size at close time.  If the size violates the disk
 *                partitioning, fail the close and truncate the file.
 */
int cache_close(struct cache_entry *centry,
		int cfile_fd,
		int modified)
{
    struct stat sb;
    long new_current_size;

    /* Only do bookkeeping if this was opened for write.
       XXX trusting the warden here XXX */
    if (modified) {
	/* stat the thing check against warden's limit */
	if (fstat(cfile_fd, &sb) < 0)
	    return errno;
	new_current_size = centry->ce_owner->w_current_size +
	                   (sb.st_blocks >> 1);
	if (new_current_size > centry->ce_owner->w_total_size) {
	    /* XXX This is a little bogus.  This guy might have been open
	       for reading, had another cachefile fill up the space, then
	       leave this one to pay for it.  Leave it for now. XXX */
	    if (close(cfile_fd) < 0)
		return errno;
	    if (truncate(centry->ce_fname, 0) < 0)
		return errno;
	    return ENOSPC;
	} else {
	    centry->ce_owner->w_current_size = new_current_size;
	}
    }
    if (close(cfile_fd) < 0)
	return errno;
    else
	return 0;
}

/*
 * cache_truncate -- delete the contents of the container file
 */
int cache_truncate(struct cache_entry *centry)
{
    struct stat sb;

    if (stat(centry->ce_fname, &sb) < 0)
	return errno;
    if (truncate(centry->ce_fname, 0) < 0)
	return errno;
    centry->ce_owner->w_current_size -= (sb.st_blocks >> 1);

    return 0;
}

