#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * cmgr_main.c -- implementation of public API of Connection Manager
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <lwp/lwp.h>
#include <rpc2/rpc2.h>

#include "include/viceroy_onode.h"
#include "include/odytypes.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "cm.private.h"

#include "include/warden.h"
#include "odyssey/viceroy/rsrc_en.h" 

/*
 * GLOBAL INFORMATION
 *   pieces of shared state for the connection manager
 */

static int cmgr_inited = 0;            /* has the cm package been initialized? */
static ds_hash_t *cmgr_hashp = NULL;   /* hash table to store connections */
static ds_hwlock_t cmgr_lock;          /* exclusive access to hash table */
static int cmgr_next_connid = 0;       /* each cmgr_create() returns a unique id */

/* hash and comparison functions */
static long hash_cmgr_elt(void *anon)
{
    cmgr_elt_t *elt = anon;
    cmgr_id_t connid = elt->ct_connid;
    if (connid < 0)
	connid = -connid; /* make sure we never hash into something -ve */
    return connid;
}

static long comp_cmgr_elt(void *a1, void *a2)
{
    cmgr_elt_t *e1 = a1, *e2 = a2;
    return (e1->ct_connid - e2->ct_connid);
}


/*
 * GENERAL UTILITY FUNCTIONS
 */

/*
 * cmgr_init -- performs any necessary initialization. must be called
 *            before using package. Takes the version number, to
 *            make sure interface matches implementation.
 */

void
cmgr_init(expected_version)
    int expected_version;
{

    ASSERT(!cmgr_inited);    /* don't let people init more than once */
    ASSERT(expected_version == CMGR_VERSION); /* or have version skew */

    cmgr_hashp = ds_hash_create(comp_cmgr_elt, hash_cmgr_elt, CMGR_HASHSIZE,
				TRUE, FALSE);
    ds_hwlock_init(&cmgr_lock);

    ASSERT(cmgr_hashp);
    cmgr_inited = 1;
    cmgr_next_connid = 0;
}

/*
 * CONNECTION MANAGEMENT
 */

/*
 * cmgr_create -- creates a connection of the specified type, to the
 *              specified endpoint, using the supplied initialization
 *              parameters.
 *
 *              returns a positive connection ID if successful,
 *              otherwise returns the appropriate error.
 */

cmgr_id_t
cmgr_create(cip)
    cmgr_initparams_t *cip;
{
    cmgr_elt_t        *eltp;
    int rc;

    ASSERT(cmgr_inited);

    ALLOC(eltp, cmgr_elt_t);
    
    /* Copy in initparams */
    memcpy(&eltp->ct_initparams, cip, sizeof(cmgr_initparams_t));

    /* Copy in bogus name, to be properly filled in later. */
    strcpy(eltp->ct_name, "Bogus Connection");

    /* Grab a copy of the flags for export */
    eltp->ct_export.cc_flags = eltp->ct_initparams.ci_flags;

    /* Set the timeout value to 0; use the default value of the underlying
       connection type until the user changes the timeout value */
    eltp->ct_timeout.tv_sec = 0;
    eltp->ct_timeout.tv_usec = 0;

    ds_hwlock_init(&eltp->ct_lock);

    if (!(eltp->ct_initparams.ci_flags & CMGR_CONNISLOCAL)) 
	APM_NW_UP();

    /* create the connection (type-specific) */
    switch(cip->ci_type) {
    case CMGR_RPC2:
	eltp->ct_export.cc_type = CMGR_RPC2;
	rc = cmgr_rpc2_create(&(eltp->ct_export.cc_handle.rpc2),
			    &(eltp->ct_initparams.ci_params.ci_rpc2),
			    &(eltp->ct_status));
	if (rc == CMGR_SUCCESS) {
	    cmgr_rpc2_name(&(eltp->ct_initparams.ci_params.ci_rpc2),
			   eltp->ct_name, CMGR_NAME_LEN);
	}
	break;
    default:
	rc = CMGR_NOTSUPP;
	goto bad;
    }

    if (!(eltp->ct_initparams.ci_flags & CMGR_CONNISLOCAL)) 
	APM_NW_DOWN();

    /* 
     * Create a resource-request list.  Start at a generous 2MB.
     * XXX Would be nice to have a better guess.
     */
    eltp->ct_export.cc_bwlist = ds_rrlist_create(2*1024*1024/8);

    if (gettimeofday(&(eltp->ct_last_valid),NULL) < 0) {
	ds_thread_log(0,"cmgr_create: gettimeofday failed! : %s",
		      strerror(errno));
	ASSERT(0);
    }


    eltp->ct_destroyed = 0;
    eltp->ct_magic = CT_MAGIC;
    eltp->ct_refcount = 0;
    eltp->ct_validating = 0;

    /* create a unique connection id */
    eltp->ct_connid = cmgr_next_connid++;
    
    ds_thread_log(0,"Created conn %d (%s)",eltp->ct_connid,eltp->ct_name);
			  
    /* add the connection to the table */

    ds_hwlock_write(&cmgr_lock);
    ASSERT(ds_hash_insert(cmgr_hashp, eltp));
    ds_hwlock_release_write(&cmgr_lock);

    return eltp->ct_connid;

 bad:
    ASSERT(rc);
    FREE(eltp);
    errno = rc;
    return -1;
}

/*
 * cmgr_destroy -- destroys the connection to the specified destination.
 *
 *               returns CMGR_SUCCESS if the connection to that destination
 *               is successfully destroyed, or CMGR_NOTCONN if there is
 *               no connection to that destination.
 */

int cmgr_destroy(id)
    cmgr_id_t id;
{
    int retval;
    int islocal = 1;
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);

    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (!eltp) {
	errno = CMGR_BADCONN;
	ds_hwlock_release_read(&cmgr_lock);
	return -1;
    }

    ASSERT(CMGR_VALID_ELT(eltp));

    ds_hwlock_write(&cmgr_lock);

    islocal = (eltp->ct_initparams.ci_flags & CMGR_CONNISLOCAL);
    if (!islocal) APM_NW_UP();

    if (!eltp->ct_validating && (eltp->ct_refcount == 0)) {
	switch(eltp->ct_export.cc_type) {
	case CMGR_RPC2:
	    ds_thread_log(10,"Destroying connection %ld (%s)",
			  eltp->ct_connid, eltp->ct_name);
	    cmgr_rpc2_destroy(&(eltp->ct_export.cc_handle.rpc2));
	    break;
	default:
	    ASSERT(0);
	}

/* NOTE: Cancel all of the requests in the resource request list */

	ASSERT(ds_hash_remove(cmgr_hashp, eltp));
	eltp->ct_magic = 0;
	FREE(eltp);
	retval = 0;
    }
    else {
	ds_thread_log(10,"Marking connection %ld (%s) for destruction",
		      eltp->ct_connid, eltp->ct_name);
	eltp->ct_destroyed = 1;
	eltp->ct_status = CMGR_DEAD;
	retval = 1;
    }

    if (!islocal) APM_NW_DOWN();

    ds_hwlock_release_write(&cmgr_lock);
    ds_hwlock_release_read(&cmgr_lock);

    return retval;
}

/*
 * cmgr_get -- given a connection ID, returns a connection
 *           description for it (if it already exists), or returns NULL
 *           (if it does not). This connection description can be used
 *           for connection-specific operations (such as obtaining a
 *           valid RPC2 handle to use for in making an RPC2 call).
 *
 *           N.B. this function makes some weak guarantees about the
 *           state of the connection it returns; specifically, it will
 *           make some effort to ensure that the connection will be
 *           up (i.e., usable) before returning it, but it will not
 *           definitely guarantee that the connection will be up.
 */

cmgr_connection_t *cmgr_get(id, lock)
    cmgr_id_t id;
    int lock;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);
    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (eltp)
	ASSERT(CMGR_VALID_ELT(eltp));

    if (eltp && !(eltp->ct_destroyed)) {
	ds_hwlock_write(&cmgr_lock);
	eltp->ct_refcount++;
	ds_hwlock_release_write(&cmgr_lock);
	ds_hwlock_release_read(&cmgr_lock);
	if (lock) {
	    ds_thread_log(100, "want to lock conn %d", id);
	    ds_hwlock_write(&eltp->ct_lock);
	    ds_thread_log(100, "locked conn %d", id);
	}
	ds_thread_log(100, "cmgr_get: %d refcnt %d", id, eltp->ct_refcount);
	
	/* If nw APM is enabled, bring up the network now */
	APM_NW_CONN_UP (eltp->ct_export.cc_handle.rpc2);

	return &(eltp->ct_export);
    }
    else {
	ds_hwlock_release_read(&cmgr_lock);
	errno = CMGR_BADCONN;
	return NULL;
    }
}

int cmgr_put(id, unlock)
    cmgr_id_t id;
    int unlock;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);
    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (!eltp) {
	ds_hwlock_release_read(&cmgr_lock);
	errno = CMGR_BADCONN;
	return -1;
    }

    ASSERT(CMGR_VALID_ELT(eltp));
    ASSERT(eltp->ct_refcount); /* don't allow more puts than gets */

    ds_hwlock_write(&cmgr_lock);
    eltp->ct_refcount--;
    if (unlock)
	ds_hwlock_release_write(&eltp->ct_lock);
    ds_thread_log(100, "cmgr_put: %d refcnt %d", id, eltp->ct_refcount);

    /* if it was marked for destruction, maybe it's destructible now,
       check */

    if ((eltp->ct_refcount == 0) && eltp->ct_destroyed)
	cmgr_destroy(id);

    ds_hwlock_release_write(&cmgr_lock);
    ds_hwlock_release_read(&cmgr_lock);

    /* If nw APM is enabled, shut down the network now */
    APM_NW_CONN_DOWN (eltp->ct_export.cc_handle.rpc2);

    return CMGR_SUCCESS;
}

int cmgr_isvalid(id)
    cmgr_id_t id;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);
    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    ASSERT(CMGR_VALID_ELT(eltp));

    if (eltp->ct_status == CMGR_ALIVE) {
	ds_hwlock_release_read(&cmgr_lock);
	return 1;
    }
    ASSERT(eltp->ct_status == CMGR_DEAD);
    ds_hwlock_release_read(&cmgr_lock);
    return 0;
}


int cmgr_markdown(id)
    cmgr_id_t id;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_write(&cmgr_lock);
    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (!eltp) {
	ds_hwlock_release_write(&cmgr_lock);
	errno = CMGR_BADCONN;
	return -1;
    }

    ASSERT(CMGR_VALID_ELT(eltp));

    switch (eltp->ct_initparams.ci_type) {
    case CMGR_RPC2:
	/* If this is the first time, we want to unbind the RPC connection */
	if (eltp->ct_status != CMGR_DEAD) {
	    if (cmgr_rpc2_destroy(&(eltp->ct_export.cc_handle.rpc2)) ==
		CMGR_SUCCESS)
		ds_thread_log(0, "cmgr_markdown: Destroyed "
			      "old RPC2 handle %ld for conn %d",
				  eltp->ct_export.cc_handle.rpc2, 
			      eltp->ct_connid);
	    else
		ds_thread_log(0, "cmgr_markdown: Could not destroy "
			      "old RPC2 handle %ld for conn %d",
			      eltp->ct_export.cc_handle.rpc2, 
			      eltp->ct_connid);
	}
		break;
    default:
	ASSERT(0);
    }

    eltp->ct_status = CMGR_DEAD;

    ds_thread_log(0,"Connection %ld (%s) marked DOWN", eltp->ct_connid,
		  eltp->ct_name);

    ds_hwlock_release_write(&cmgr_lock);
    return 0;
}
    
int cmgr_validate(id)
    cmgr_id_t id;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    dummy.ct_connid = id;

    /* loop until we get a valid connection that no one else is validating */

    while(1) {
	ds_hwlock_read(&cmgr_lock);

	eltp = ds_hash_member(cmgr_hashp, &dummy);
	ASSERT(CMGR_VALID_ELT(eltp));
    
	if (eltp->ct_destroyed) {
	    ds_hwlock_release_read(&cmgr_lock);
	    return 0;
	}
	
	if (eltp->ct_validating) {
	    ds_hwlock_release_read(&cmgr_lock);
	    ds_thread_yield();
	}
	else
	    break;
    }

    if (eltp->ct_status == CMGR_ALIVE) {
	ds_hwlock_release_read(&cmgr_lock);
	return 1;
    }
    else { /* now we actually go and try to validate the connection */
	struct timeval time_now;
	struct timeval last_valid;
	struct timeval nowminuslast;
	struct timeval restart;
	cmgr_initparams_t ci;
	cmgr_status_t new_status = eltp->ct_status;
	struct timeval tv;

	ds_hwlock_write(&cmgr_lock);
	eltp->ct_validating = 1;

	restart = eltp->ct_initparams.ci_restart;
	last_valid = eltp->ct_last_valid;
	ci = eltp->ct_initparams;
	tv = eltp->ct_timeout;

	/* we can release all the locks since we set ct_validating,
	   no on else should be touching the same fields that we touch */

	ds_hwlock_release_write(&cmgr_lock);
	ds_hwlock_release_read(&cmgr_lock);

    
	if (gettimeofday(&time_now,NULL) < 0) {
	    ds_thread_log(0,"cmgr_get: gettimeofday failed! : %s",
			  strerror(errno));
	    ASSERT(0);
	}

	/* if the last validate was too recent, then don't try just yet. */

	timersub(&time_now, &last_valid, &nowminuslast);
	if timercmp(&nowminuslast, &restart, >= ) {

	    APM_NW_CONN_UP(id);
	    
	    switch (ci.ci_type) {
	    case CMGR_RPC2:
	    {
		cmgr_rpc2_handle_t crh;
		
		/* 1st destroy the old connection (if one exists) */
		if (eltp->ct_export.cc_handle.rpc2) {
		    int rc;

		    rc = cmgr_rpc2_destroy(&eltp->ct_export.cc_handle.rpc2);
		    ds_thread_log (0, "destroying connection %d, rc = %d\n",
				   eltp->ct_export.cc_handle.rpc2, rc);
		    if (rc == CMGR_SUCCESS) {
			eltp->ct_export.cc_handle.rpc2 = 0;
		    } 
		}

		cmgr_rpc2_create(&crh, &ci.ci_params.ci_rpc2, &new_status);
		
		if (new_status == CMGR_ALIVE) {
		    ds_thread_log(0,"cmgr_validate: restarted conn %ld (%s)",
				  eltp->ct_connid, eltp->ct_name);
		    if (tv.tv_sec || tv.tv_usec) {
			ds_thread_log(0,"cmgr_validate: set conn %ld (%s) "
				      "timeout to %ld.%06ld",
				      eltp->ct_connid, eltp->ct_name,
				      tv);
			cmgr_rpc2_settimeout(&crh, tv);
		    }
		}
		else
		    ds_thread_log(0,"cmgr_validate: failed to restart %ld (%s)",
				  eltp->ct_connid, eltp->ct_name);

		ds_hwlock_write(&cmgr_lock);
		ASSERT(CMGR_VALID_ELT(eltp));
		if (gettimeofday(&time_now,NULL) < 0) {
		    ds_thread_log(0,"cmgr_get: gettimeofday failed! : %s",
				  strerror(errno));
		    ASSERT(0);
		}

		eltp->ct_last_valid = time_now; 
		eltp->ct_export.cc_handle.rpc2 = crh;
		eltp->ct_validating = 0;
		eltp->ct_status = new_status;
		ds_hwlock_release_write(&cmgr_lock);
		break;
	    }
	    default:
		ASSERT(0);
	    }

	    APM_NW_CONN_DOWN(id);

	}
	else {
	    /* it's too soon since the last try, leave the connection down */
	    ds_thread_log(1000, "cmgr_validate: too soon to restart %ld", id);
	    ds_hwlock_write(&cmgr_lock);
	    ASSERT(CMGR_VALID_ELT(eltp));
	    eltp->ct_validating = 0;
	    ds_hwlock_release_write(&cmgr_lock);
	}

	ds_hwlock_read(&cmgr_lock);

	/* check if it was marked for destruction while validating */
	if (eltp->ct_destroyed) {
	    cmgr_destroy(eltp->ct_connid);
	    new_status = CMGR_DEAD;
	}
	else
	    new_status = eltp->ct_status;

	ds_hwlock_release_read(&cmgr_lock);

	if (new_status == CMGR_ALIVE)
	    return 1;
	ASSERT(new_status == CMGR_DEAD);
	return 0;
    }
}

#if 0
/*
 * CONNECTION INFORMATION
 */

/*
 * cmgr_getinfo -- returns information about the specified connection
 */

int cmgr_getinfo(id, dir, cip)
    cmgr_id_t id;
    cmgr_direction_t dir;
    cmgr_conninfo_t *cip;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;
    int rc;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);
    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (eltp)
	ASSERT(CMGR_VALID_ELT(eltp));
    else {
	ds_hwlock_release_read(&cmgr_lock);
	errno = CMGR_BADCONN;
	return -1;
    }

    cip->ci_status = eltp->ct_status;

    if (cip->ci_status == CMGR_ALIVE) {
	/* get the type-specific connection information */
	switch (eltp->ct_export.cc_type) {
	case CMGR_RPC2:
	{
	    cmgr_rpc2_conninfo_t crc;
	    cmgr_rpc2_handle_t crh = eltp->ct_export.cc_handle.rpc2;
	    
	    ds_hwlock_release_read(&cmgr_lock);
	    
	    rc = cmgr_rpc2_getinfo(&crh, &crc);
	    
	    if (rc != CMGR_SUCCESS) {
		ds_thread_log(100, "cmgr_getinfo: cmgr_rpc2_getinfo failed (%d)", id);
		errno = rc;
		return -1;
	    }
	    cip->ci_bandwidth = crc.crc_bandwidth;
	    cip->ci_latency = crc.crc_latency;
	    cip->ci_traffic = crc.crc_traffic;
	    cip->ci_winstart = crc.crc_winstart;
	    cip->ci_winstop = crc.crc_winstop;
	}
	    break;
	default:
	    ASSERT(0);
	}
    }
    else { /* if the connection is down, there is no other relevant info */
	ASSERT(cip->ci_status == CMGR_DEAD);
	ds_hwlock_release_read(&cmgr_lock);
	errno = CMGR_CONNDOWN;
	return -1;
    }	

    return 0;
}
#endif

int
cmgr_getlogs(connid, rttlog, tplog)
    cmgr_id_t connid;
    cmgr_log_t *rttlog;
    cmgr_log_t *tplog;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;
    int rc=0;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    /* Hold until exit */
    ds_hwlock_read(&cmgr_lock);

    dummy.ct_connid = connid;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (eltp) {
	ASSERT(CMGR_VALID_ELT(eltp)); 
    } else {
	rc = CMGR_BADCONN;
	goto exit;
    }

    if (eltp->ct_status == CMGR_DEAD) {
	rc = CMGR_CONNDOWN;
	goto exit;
    }

    /* Ask the underlying connection type layer to get the logs for us */
    switch (eltp->ct_export.cc_type) {
    case CMGR_RPC2:
    {
	cmgr_rpc2_handle_t crh;

	crh = eltp->ct_export.cc_handle.rpc2;
	rc = cmgr_rpc2_getlogs(&crh, rttlog, tplog);
    }
	break;
    default:
	ASSERT(0);
	break;
    }

 exit:
    ds_hwlock_release_read(&cmgr_lock);
    if (rc) {
	errno = rc;
	return -1;
    } else {
	return 0;
    }
}


/*********** Iterators */

struct cmgr_iter_t *cmgri_make()
{
    struct cmgr_iter_t *result;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));
    ALLOC(result, struct cmgr_iter_t);
    result->cit_magic = CIT_MAGIC;
    ds_hwlock_write(&cmgr_lock);
    result->cit_hash_iter = ds_hash_iter_create(cmgr_hashp);
    ds_hwlock_release_write(&cmgr_lock);
    ASSERT(result->cit_hash_iter);
    return result;
}

void cmgri_destroy(iter)
    struct cmgr_iter_t *iter;
{
    ASSERT(cmgr_inited);
    ODY_ASSERT(iter && (iter->cit_magic == CIT_MAGIC) && iter->cit_hash_iter);
    iter->cit_magic = 0;
    ds_hwlock_write(&cmgr_lock);
    ds_hash_iter_destroy(iter->cit_hash_iter);
    ds_hwlock_release_write(&cmgr_lock);
    FREE(iter);
}

int cmgri_next(iter, cidp)
    struct cmgr_iter_t *iter;
    cmgr_id_t        *cidp;
{
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ODY_ASSERT(cidp && iter && (iter->cit_magic == CIT_MAGIC) &&
	       iter->cit_hash_iter);
    ds_hwlock_write(&cmgr_lock);
    eltp = ds_hash_iter_next(iter->cit_hash_iter);
    ds_hwlock_release_write(&cmgr_lock);

    if (eltp) {
	*cidp = eltp->ct_connid;
	return 0;
    }
    else {
	errno = ENOENT;
	return -1;
    }
}

/* Change the timeout associated with a given connection */
int
cmgr_settimeout(connid, tv)
     cmgr_id_t connid;
     struct timeval tv;
{
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;
    int rc=0;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    if ((tv.tv_sec < 0) || (tv.tv_usec < 0)) {
	rc = EINVAL;
	goto exit;
    }

    if ((tv.tv_sec == 0) && (tv.tv_usec == 0)) {
	rc = EINVAL;
	goto exit;
    }

    /* Hold until exit */
    ds_hwlock_read(&cmgr_lock);

    dummy.ct_connid = connid;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (eltp) {
	ASSERT(CMGR_VALID_ELT(eltp)); 
    } else {
	rc = CMGR_BADCONN;
	goto exit;
    }

    /* Ask the underlying connection type layer to change the timeout */
    switch (eltp->ct_export.cc_type) {
    case CMGR_RPC2:
    {
	cmgr_rpc2_handle_t crh;
	crh = eltp->ct_export.cc_handle.rpc2;
	rc = cmgr_rpc2_settimeout(&crh, tv);
	if (!rc) {
	    eltp->ct_timeout = tv;
	    ds_thread_log(0,"cmgr_settimeout: set conn %ld (%s) "
			  "timeout to %ld.%06ld",
			  eltp->ct_connid, eltp->ct_name,
			  tv);
	}
	else {
	    ds_thread_log(0,"cmgr_settimeout: failed to set conn %ld (%s) "
			  "timeout to %ld.%06ld",
			  eltp->ct_connid, eltp->ct_name,
			  tv);
	}
	       
	break;
    }
    default:
	ASSERT(0);
	break;
    }

exit:
    ds_hwlock_release_read(&cmgr_lock);
    if (rc) {
	errno = rc;
	return -1;
    } else {
	return 0;
    }
}

/*
 * cmgr_islocal
 *
 *               rReturns True if the connection is local.  Returns False
 *               if the connection is remote or non-existant.
 */

int cmgr_islocal(id)
    cmgr_id_t id;
{
    int islocal;
    cmgr_elt_t dummy;
    cmgr_elt_t *eltp;

    ASSERT(cmgr_inited);
    ASSERT(ds_hash_valid(cmgr_hashp));

    ds_hwlock_read(&cmgr_lock);

    dummy.ct_connid = id;
    eltp = ds_hash_member(cmgr_hashp, &dummy);

    if (!eltp) {
	errno = CMGR_BADCONN;
	ds_hwlock_release_read(&cmgr_lock);
	return 0;
    }

    ASSERT(CMGR_VALID_ELT(eltp));

    islocal = (eltp->ct_initparams.ci_flags & CMGR_CONNISLOCAL);

    ds_hwlock_release_read(&cmgr_lock);

    return (islocal);
}



