#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#ifdef __linux__
#include <sys/signal.h>
#if !defined(__GLIBC__) || __GLIBC__ < 2
#include <linux/fs.h>
#endif
#include <mntent.h>
#endif

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/uio.h>
#include <sys/time.h>
#ifdef BW_TOOL
#include <sys/socket.h>
#include <netinet/in.h>
#include <include/bw_tool.h>
#endif

#include "include/odytypes.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"
#include "base/ds/ds_thread.h"

#include <kernel/vfs/includes/ody.h>
#include <kernel/vfs/includes/odyapi.h>

#include "include/viceroy_onode.h"
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"

#include "include/viceroy.h"
#include "include/odyssey.h"
#include "vfscalls.h"
#include "rsrc.h"
#include "bmod.h"
#include "odyssey/multif/multif_vic.h"

#define OUT_IOVEC   0
#define WARDEN_IOVEC 1
#define MAX_IOVEC   2

/* 
 * We normally only use one of the iov entries in requests.
 * Occasionally, we need the second one if the warden supplies its own
 * buffer for something.  The two functions that are capable of this
 * are rdwr (in the read case) and tsop.  
 */
struct request {
    size_t       insz;
    char         *inbuf;
    int          iovnum;
    struct iovec iov[MAX_IOVEC];
};

#ifdef BW_TOOL
static int        bwfd;		/* socket for getting requests from bwmon */
#endif
static int        kfd;
ds_safeq_t       *workq;
static int        downcall_uniq;

static void
kern_write(iovnum, iov)
    size_t iovnum;
    struct iovec *iov;
{
    size_t total = 0;
    int    i;
    int    rc;

    for (i=0; i<iovnum; i++) {
	total += iov[i].iov_len;
    }

#ifdef __linux__
    {
      if (iovnum == 1)
	rc = write(kfd, iov[0].iov_base, iov[0].iov_len);
      else {
	char *buffer, *place;
	NALLOC(buffer, char, total);
	place = buffer;
	for (i=0; i<iovnum; i++) {
	  memcpy(place, iov[i].iov_base, iov[i].iov_len);
	  place += iov[i].iov_len;
	}
	rc = write(kfd, buffer, total);
	FREE (buffer);
      }
    }
#else
    rc = writev(kfd, iov, iovnum);
#endif

    if (rc != total) {
      ds_thread_log(0, "%d != %d, errno is %d\n", rc, total, errno);
      ASSERT(errno == ESRCH); /* The only possible error is that the process
				 that made the odyssey request went away */
      /* XXX This error code should be propagated back up and dealt with
	 at a higher level -- this is a stopgap. It _should_ work since
	 kern_write is only used for responses to upcalls (not for downcalls)
	 and ESRCH should be the only error possible on those. */
      ds_thread_log(0, "kern_write: requesting process must have died");
    }
}

static void
kern_root(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *fidp = &out->ody_root.fid;
    
    out->oh.result = vfs_root(fidp);
    iov[OUT_IOVEC].iov_len = sizeof(struct ody_root_out);
    kern_write(iovnum, iov);
}

static void
kern_lookup(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *dirp = &in->ody_lookup.fid;
    char        *name = (char*)in + (int)in->ody_lookup.name;
    struct ofid *fidp = &out->ody_lookup.fid;
    enum vtype  *vtp  = (enum vtype *)&out->ody_lookup.vtype;
    
    ASSERT((int)in->ody_lookup.name == sizeof (struct ody_lookup_in));
    
    out->oh.result = vfs_lookup(dirp, name, fidp, vtp, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_lookup_out);
    kern_write(iovnum, iov);
}

static void
kern_access(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid  *fidp = &in->ody_access.fid;
    int          mode = in->ody_access.flags;

    out->oh.result = vfs_access(fidp, mode, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    kern_write(iovnum, iov);
}

static void
kern_open(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *fidp = &in->ody_open.fid;
    int         flags = in->ody_open.flags;
    dev_t       *devp = &out->ody_open.dev;
    ino_t       *inop = &out->ody_open.inode;
    int         uncacheable = 0;
    
    out->oh.result = vfs_open(fidp, flags, devp, inop, cred, &uncacheable);
    out->ody_open.flags = 0;
    if (uncacheable) {
	out->ody_open.flags |= ON_UNCACHEABLE;
    }
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_open_out);
    kern_write(iovnum, iov);
}

static void
kern_close(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *fidp = &in->ody_close.fid;
    int         flags = in->ody_close.flags;
    
    out->oh.result = vfs_close(fidp, flags, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    kern_write(iovnum, iov);
}

static void
kern_getattr(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid  *fidp = &in->ody_getattr.fid;
    struct ody_vattr *attrp = &out->ody_getattr.attr;
    
    out->oh.result = vfs_getattr(fidp, attrp, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_getattr_out);
    kern_write(iovnum, iov);
}

static void
kern_rdwr(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *fidp = &in->ody_rdwr.fid;
    int rwflag = in->ody_rdwr.rwflag;
    int *count = &out->ody_rdwr.count;
    int offset = in->ody_rdwr.offset;
    int ioflag = in->ody_rdwr.ioflag;
    void *buf;

    ASSERT(iov[WARDEN_IOVEC].iov_base == NULL);

    if (rwflag == O_RW_WRITE) {
	if (in->ody_rdwr.data != sizeof (struct ody_rdwr_in)) {
	    ds_thread_log(0, "Warning: rdwr in->data bogus!");
	    out->oh.result = EINVAL;
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
	    goto exit;
	}
	buf = (void*)in + (int)in->ody_rdwr.data;
    } else {
	out->ody_rdwr.data = sizeof (struct ody_rdwr_out);
	buf = (void*)out + (int)out->ody_rdwr.data;
    }

    /* Set up the rwflag */
    out->ody_rdwr.rwflag = rwflag;

    /* Set up the initial value for count; warden returns final val */
    *count = in->ody_rdwr.count;
    
    out->oh.result = vfs_rdwr(fidp, rwflag, count, offset, ioflag, &buf, cred);
    
    if (!out->oh.result && (rwflag == O_RW_READ)) {
	/* Check if response buffer is the same or not */
	if (buf != ((void*)out + (int)out->ody_rdwr.data)) {
	    iovnum++;
	    iov[WARDEN_IOVEC].iov_len = *count;
	    iov[WARDEN_IOVEC].iov_base = buf;
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_rdwr_out);
	} else {
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_rdwr_out) + *count;
	}
    } else {
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_rdwr_out);
    }

 exit:
    kern_write(iovnum, iov);

    if (iov[WARDEN_IOVEC].iov_base) {
	vfs_rdwrdisp(fidp, *count, offset, iov[WARDEN_IOVEC].iov_base, cred);
	iov[WARDEN_IOVEC].iov_base = NULL;
	iov[WARDEN_IOVEC].iov_len = 0;
    }
}

static void
kern_readdir(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid *fidp = &in->ody_readdir.fid;
    int *count = &out->ody_readdir.size;
    int offset = in->ody_readdir.offset;
    void *buf;

    ASSERT(iov[WARDEN_IOVEC].iov_base == NULL);
    
    out->ody_readdir.data = (char *)sizeof (struct ody_readdir_out);
    buf = (struct dirent *)
	((char *)out + (int)out->ody_readdir.data);

    /* Set up the initial value for count; warden returns final val */
    *count = in->ody_readdir.count;

    out->oh.result = vfs_readdir(fidp, count, offset, buf, cred);
    
    if (!out->oh.result) {
	/* Check if response buffer is the same or not */
	if (buf != ((void*)out + (int)out->ody_readdir.data)) {
	    iovnum++;
	    iov[WARDEN_IOVEC].iov_len = *count;
	    iov[WARDEN_IOVEC].iov_base = buf;
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_readdir_out);
	} else {
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_readdir_out) + *count;
	}
    } else {
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_readdir_out);
    }

    kern_write(iovnum, iov);

    if (iov[WARDEN_IOVEC].iov_base) {
	vfs_readdirdisp(fidp, *count, offset, iov[WARDEN_IOVEC].iov_base, cred);
	iov[WARDEN_IOVEC].iov_base = NULL;
	iov[WARDEN_IOVEC].iov_len = 0;
    }
}

static void
kern_tsop(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid  *fidp = &in->ody_tsop.fid;
    unsigned int gs = in->ody_tsop.gs;
    unsigned int type = in->ody_tsop.type;
    unsigned int ts = in->ody_tsop.ts;
    unsigned int opnum = in->ody_tsop.op;
    size_t       argsz = in->ody_tsop.argsz;
    size_t       retsz = in->ody_tsop.retsz;
    void         *argp = NULL;
    void         *retp;
    
    ASSERT(iov[WARDEN_IOVEC].iov_base == NULL);

    if ((int)in->ody_tsop.data != sizeof (struct ody_tsop_in)) {
	ds_thread_log(0, "ody_tsop_in malformed! (%d should be %d)",
		      in->ody_tsop.data, sizeof (struct ody_tsop_in));
	out->oh.result = EINVAL;
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
	goto exit;
    }
    
    if (argsz) {
	argp = (char*)in + (int)in->ody_tsop.data;
    } else {
	argp = NULL;
    }

    retp = NULL;

    out->oh.result = vfs_tsop(fidp, gs, type, ts, opnum, argsz, argp, &retsz,
			   &retp, cred);
	    
    if (!out->oh.result) {
	/* Fill in the results */
	out->ody_tsop.retsz = retsz;
	(int)out->ody_tsop.data = sizeof (struct ody_tsop_out);
	if (retsz) {
	    iovnum++;
	    iov[WARDEN_IOVEC].iov_len = retsz;
	    iov[WARDEN_IOVEC].iov_base = retp;
	}
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_tsop_out);
    } else {
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    }

 exit:
    kern_write(iovnum, iov);

    if (iov[WARDEN_IOVEC].iov_base) {
	vfs_dispose(fidp, opnum, argsz, argp, retsz, 
		    iov[WARDEN_IOVEC].iov_base, cred);
	iov[WARDEN_IOVEC].iov_base = NULL;
	iov[WARDEN_IOVEC].iov_len = 0;
    }
}

static void
kern_request(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    struct ofid  *fidp = &in->ody_request.fid;
    unsigned int rsrc = in->ody_request.rsrc;
    unsigned int gs = in->ody_request.gs;
    unsigned int type = in->ody_request.type;
    unsigned int ts = in->ody_request.ts;
    int          low = in->ody_request.low;
    int          high = in->ody_request.high;
    int          cancelled=0;
    int          value=0;
    int          reqid=0;

    out->oh.result = rsrc_request(fidp, rsrc, gs, type, ts, low, high,
				  in->ih.pid, &cancelled, &value, &reqid);
    
    if (!out->oh.result) {
	out->ody_request.cancelled = cancelled;
	out->ody_request.result = reqid;
    }
    else if (out->oh.result == ODY_ENOTINWINDOW) {
	out->ody_request.cancelled = cancelled;
	out->ody_request.value = value;
	out->ody_request.result = 0;
    }

    iov[OUT_IOVEC].iov_len = sizeof (struct ody_request_out);
    kern_write(iovnum, iov);
}

static void
kern_cancel(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    int reqid = in->ody_cancel.reqid;
    
    out->oh.result = rsrc_cancel(reqid);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    kern_write(iovnum, iov);
}

static void
kern_cancpid(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    out->oh.result = rsrc_cancpid(in->ih.pid);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    kern_write(iovnum, iov);
}

static void
kern_bmod_start(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    int joules   = in->ody_bmod_start.joules;
    float charge = in->ody_bmod_start.charge;

    out->oh.result = bmod_start (joules, charge);
    
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_bmod_start_out);
    kern_write(iovnum, iov);
}

static void
kern_bmod_stop(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    out->oh.result = bmod_stop ();
    
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_bmod_stop_out);
    kern_write(iovnum, iov);
}

static void
kern_time_to_recharge(in, out, iovnum, iov, cred)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
{
    int seconds   = in->ody_time_to_recharge.seconds;

    out->oh.result = bmod_set_time_to_recharge (seconds);
    
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_time_to_recharge_out);
    kern_write(iovnum, iov);
}

static void
kern_multif(in, out, iovnum, iov, cred, pid)
    union ody_inputArgs   *in;
    union ody_outputArgs  *out;
    size_t                iovnum;
    struct iovec          *iov;
    struct ody_cred       *cred;
    u_short               pid;
{
    unsigned int gs = in->ody_multif.gs;
    unsigned int type = in->ody_multif.type;
    unsigned int ts = in->ody_multif.ts;
    unsigned int opnum = in->ody_multif.op;
    size_t       argsz = in->ody_multif.argsz;
    size_t       retsz = in->ody_multif.retsz;
    void         *argp = NULL;
    void         *retp;
    
    if ((gs != ODY_GENERIC_INTERFACE)
	|| (type != ODY_NO_TYPE)) {
	ds_thread_log(0, "vfs_tsop: bad version (%d not %d) or (%d not %d)",
		      gs, ODY_GENERIC_INTERFACE, type, ODY_NO_TYPE);
	out->oh.result = ODY_EBADVERSION;
	goto exit;
    }

    ASSERT(iov[WARDEN_IOVEC].iov_base == NULL);
    
    if ((int)in->ody_multif.data != sizeof (struct ody_multif_in)) {
	ds_thread_log(0, "ody_multif_in malformed! (%d should be %d)",
		      in->ody_multif.data, sizeof (struct ody_multif_in));
	out->oh.result = EINVAL;
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
	goto exit;
    }
    
    if (argsz) {
	argp = (char*)in + (int)in->ody_multif.data;
    } else {
	argp = NULL;
    }

    retp = NULL;


    out->oh.result = multif_call(ts, opnum, argsz, argp, &retsz,
				&retp, cred, pid);
	    
    if (!out->oh.result) {
	/* Fill in the results */
	out->ody_multif.retsz = retsz;
	(int)out->ody_multif.data = sizeof (struct ody_multif_out);
	if (retsz) {
	    iovnum++;
	    iov[WARDEN_IOVEC].iov_len = retsz;
	    iov[WARDEN_IOVEC].iov_base = retp;
	}
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_multif_out);
    } else {
	iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
    }

 exit:
    kern_write(iovnum, iov);

    if (iov[WARDEN_IOVEC].iov_base) {
	multif_dispose(opnum, argsz, argp, retsz, 
		       iov[WARDEN_IOVEC].iov_base, cred);
	iov[WARDEN_IOVEC].iov_base = NULL;
	iov[WARDEN_IOVEC].iov_len = 0;
    }
}

#ifdef SETS
static void
kern_expand(in, out, iovnum, iov, cred)
    union ody_inputArgs  *in;
    union ody_outputArgs *out;
    size_t               iovnum;
    struct iovec         *iov;
    struct ody_cred      *cred;
{
    struct ofid *dirp = &in->ody_expand.fid;
    char        *name = (char*)in + (int)in->ody_expand.name;
    struct ofid *fidp = &out->ody_expand.fid;
    enum vtype  *vtp  = (enum vtype *)&out->ody_expand.vtype;
    int         *curs = (int *)&out->ody_expand.cursor;
    
    ASSERT((int)in->ody_expand.name == sizeof (struct ody_expand_in));
    
    out->oh.result = vfs_expand(dirp, name, fidp, vtp, curs, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_expand_out);
    kern_write(iovnum, iov);
}

static void
kern_cursexpand(in, out, iovnum, iov, cred)
    union ody_inputArgs  *in;
    union ody_outputArgs *out;
    size_t               iovnum;
    struct iovec         *iov;
    struct ody_cred      *cred;
{
    struct ofid *fidp  = &in->ody_cursexpand.fid;
    size_t      size   = in->ody_cursexpand.bufsz;
    size_t      *retsz = &out->ody_cursexpand.retsz;
    void        *buf;
    
    out->ody_cursexpand.data = sizeof (struct ody_cursexpand_out);
    buf = (void *)out + (int)out->ody_cursexpand.data;
    out->oh.result = vfs_cursexpand(fidp, size, retsz, buf, cred);
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_cursexpand_out)+*retsz;
    kern_write(iovnum, iov);
}

static void
kern_prefetch(in, out, iovnum, iov, cred)
    union ody_inputArgs  *in;
    union ody_outputArgs *out;
    size_t               iovnum;
    struct iovec         *iov;
    struct ody_cred      *cred;
{
    struct ofid *fidp  = &in->ody_prefetch.fid;
    int          flags = in->ody_prefetch.flags;
    dev_t       *devp = &out->ody_prefetch.dev;
    ino_t       *inop = &out->ody_prefetch.inode;
    int          uncacheable = 0;
    
    out->oh.result = vfs_prefetch(fidp, flags, devp, inop, cred, &uncacheable);
    out->ody_prefetch.flags = 0;
    if (uncacheable) {
	out->ody_prefetch.flags |= ON_UNCACHEABLE;
    }
    iov[OUT_IOVEC].iov_len = sizeof (struct ody_prefetch_out);
    kern_write(iovnum, iov);
}
#endif

static int
Worker(argp)
    char *argp;
{
    struct request    *reqp;
    struct ody_in_hdr *in;
    struct ody_out_hdr *out;
    struct iovec      *iov;
    size_t            iovnum;
    struct ody_cred   *cred;
    u_short           pid;

    while (1) {
	ds_thread_log(1000, "waiting for message\n");
	reqp = (struct request *)ds_safeq_deq(workq);
	in = (struct ody_in_hdr *)reqp->inbuf;
	out = (struct ody_out_hdr *)reqp->iov[OUT_IOVEC].iov_base;
	iov = reqp->iov;
	iovnum = reqp->iovnum;
	cred = &in->cred;
	pid = in->pid;

	out->opcode = in->opcode;
	out->unique = in->unique;

	ds_thread_log(100, "Opcode %d\tUniqe %d", in->opcode, in->unique);

	switch(in->opcode) {

	case ODY_ROOT: 
	    kern_root(in, out, iovnum, iov, cred);
	    break;
	case ODY_LOOKUP:
	    kern_lookup(in, out, iovnum, iov, cred);
	    break;
	case ODY_ACCESS:
	    kern_access(in, out, iovnum, iov, cred);
	    break;
	case ODY_OPEN:
	    kern_open(in, out, iovnum, iov, cred);
	    break;
	case ODY_CLOSE:
	    kern_close(in, out, iovnum, iov, cred);
	    break;
	case ODY_GETATTR:
	    kern_getattr(in, out, iovnum, iov, cred);
	    break;
	case ODY_RDWR:
	    kern_rdwr(in, out, iovnum, iov, cred);
	    break;
	case ODY_READDIR:
	    kern_readdir(in, out, iovnum, iov, cred);
	    break;
	case ODY_TSOP:
	    kern_tsop(in, out, iovnum, iov, cred);
	    break;
	case ODY_REQUEST:
	    kern_request(in, out, iovnum, iov, cred);
	    break;
	case ODY_CANCEL:
	    kern_cancel(in, out, iovnum, iov, cred);
	    break;
	case ODY_CANCPID:
	    kern_cancpid(in, out, iovnum, iov, cred);
	    break;
	case ODY_BMOD_START:
	    kern_bmod_start(in, out, iovnum, iov, cred);
	    break;
	case ODY_BMOD_STOP:
	    kern_bmod_stop(in, out, iovnum, iov, cred);
	    break;
	case ODY_TIME_TO_RECHARGE:
	    kern_time_to_recharge(in, out, iovnum, iov, cred);
	    break;
	case ODY_MULTIF:
	    kern_multif(in, out, iovnum, iov, cred, pid);
	    break;
#ifdef SETS
	case ODY_EXPAND:
	    kern_expand(in, out, iovnum, iov, cred);
	    break;
	case ODY_CURSEXPAND:
	    kern_cursexpand(in, out, iovnum, iov, cred);
	    break;
	case ODY_PREFETCH:
	    kern_prefetch(in, out, iovnum, iov, cred);
	    break;
#endif
	case ODY_SIGNAL:
	    /* XXX what else to do here? */
	    break;
	default:
	    /* Changed from 0 to 1 to run experiments with minimal logging */
	    ds_thread_log(/* XXX */ 1, "Opcode %d ** not supported **", 
			  in->opcode);
	    out->result = EOPNOTSUPP;
	    iov[OUT_IOVEC].iov_len = sizeof (struct ody_out_hdr);
	    kern_write(iovnum, iov);
	}
	
	/* 
	 * Deallocate the reqp.  If anyone added to the iovec, they
	 * must already have removed it again before returning to us.
	 * This is because we don't know who we should be giving it
	 * back to.
	 */
	ASSERT(reqp->iov[WARDEN_IOVEC].iov_base == NULL);
	free(reqp->inbuf);
	free(reqp->iov[OUT_IOVEC].iov_base);
	free(reqp);
    }
    return 0;
}


void
kern_loop()
{
    int                  readfds;
    int                  writefds;
    int                  exceptfds;
    int                  rc;
    struct request       *reqp;
    union ody_inputArgs  *in;
    int                  k_mask = (1 << kfd);
#ifdef BW_TOOL
    int                  bw_mask = (1 << bwfd);
    char                 req_buf[256];
#endif

    while (1) {
	exceptfds = readfds = (1 << kfd);
#ifdef BW_TOOL
	exceptfds = readfds = readfds | (1 << bwfd);
#endif
	writefds = 0;
	rc = IOMGR_Select(kfd+1, &readfds, &writefds, &exceptfds, NULL);
	if (rc <= 0) continue;

	if (readfds & k_mask) {
	    reqp = malloc(sizeof(struct request));
	    ASSERT(reqp);
	    
	    reqp->insz = OC_MAXMSGSIZE;
	    reqp->inbuf = malloc(reqp->insz);
	    ASSERT(reqp->inbuf);
	    
	    reqp->iov[OUT_IOVEC].iov_len = OC_MAXMSGSIZE;
	    reqp->iov[OUT_IOVEC].iov_base = malloc(reqp->iov[OUT_IOVEC].iov_len);
	    ASSERT(reqp->iov[OUT_IOVEC].iov_base);
	    
	    reqp->iov[WARDEN_IOVEC].iov_len = 0;
	    reqp->iov[WARDEN_IOVEC].iov_base = NULL;
	    
	    reqp->iovnum = 1;
	    
	    while (1) {
		rc = read(kfd, reqp->inbuf, reqp->insz);
		if (rc <=0)
		  ds_thread_log(0, "kern_loop: read failed- ret %d errno %d\n",
				rc, errno);
		ASSERT(rc>=sizeof(struct ody_in_hdr));
		in = (union ody_inputArgs *)reqp->inbuf;
		/* Check to see if this is an unusually large message */
		if (in->ih.opcode == ODY_TOOSMALL) {
		    reqp->insz = in->ody_toosmall.bufsiz;
		    free(reqp->inbuf);
		    reqp->inbuf = malloc(reqp->insz);
		} else {
		    break;
		}
	    }
	    ds_safeq_enq(workq, reqp);
	}
#ifdef BW_TOOL
	if (readfds & bw_mask) {
	    int err;
	    int vice_cid;
	    struct sockaddr_in from_addr;
	    int from_len = sizeof(from_addr);
	    int conn_bw;

	    /* detect sender */
	    err = recvfrom(bwfd, req_buf, 256, 0,
			   (struct sockaddr *)&from_addr, &from_len);
	    if (err < 0) {
		perror("read");
		/* ASSERT(0); not sure if this is a fatal condition */
	    }
	    else {
		req_buf[err] = '\0';
		sscanf(req_buf, "%d", &vice_cid);
		conn_bw = vr_rsrc_getbw(vice_cid);
		sprintf(req_buf, "%f", (float)conn_bw);
		err = sendto(bwfd, req_buf, strlen(req_buf), 0,
			     (struct sockaddr *)&from_addr, from_len);
	    }
	}
#endif
    }
}

void
kern_init(dev, mnt, numworkers, loglevel)
    char *dev;
    char *mnt;
    int numworkers;
    int loglevel;
{
    struct ody_out_hdr msg;
    int                result;
    int                i;
#ifdef BW_TOOL
    struct sockaddr_in serv_addr;
#endif


    /* Try to unmount the directory if it is already mounted */
#ifdef __BSD44__
    unmount(mnt, 0);
#endif
#ifdef __linux__
    umount(mnt);
#endif

    /* Test the kernel */
    kfd = open(dev, O_RDWR, 0);
    if (kfd < 0) {
	perror("opening kernel device");
	exit(errno);
    }
    
#ifdef BW_TOOL
    /* open a socket to receive packets from bwmon.
     * Note: in error cases, let viceroy go ahead and run. */
    bwfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (bwfd < 0) {
	perror("socket");
    } else {
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(BW_VICE_PORT);
	result = bind(bwfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr));
	if (result < 0) {
	    perror("bind");
	}
    }
#endif
    msg.opcode = ODY_FLUSH;
    msg.unique = 0;
    result = write(kfd, (char*)&msg, sizeof (struct ody_out_hdr));
    if (result < 0) {
	perror("writing to kernel device");
	exit(errno);
    }
    if (result != sizeof (struct ody_out_hdr)) {
	fprintf(stderr,"Incomplete write to kernel device, wanted %d got %d\n",
		sizeof(unsigned long)*2, result);
	exit(-2);
    }

#ifdef __BSD44__
    if (mount(MOUNT_ODY, mnt, 0, dev)) {
	perror("mount");
	exit(errno);
    }
#endif
#ifdef __linux__
    {
	struct sigaction sa;
	sa.sa_handler = SIG_IGN;
	sigemptyset(&(sa.sa_mask));
	sa.sa_flags = 0;
	if ( sigaction(SIGCHLD, &sa, NULL) ) {
	    perror("Cannot set signal handler for SIGCHLD");
	    ASSERT(0);
	}
    }
    if (fork() == 0) {
	int error;
	error = close(kfd);
	if (error) {
	    pid_t parent;
	    perror("child: can't close kfd: killing parent.");
	    parent = getppid();
	    kill(parent, SIGKILL);
	    exit(1);
	}
	error = mount("ody", mnt, "ody", MS_MGC_VAL, &dev);
	if (error) {
	    pid_t parent;
	    perror("child: mount system call failed. Killing parent.");
	    parent = getppid();
	    kill(parent, SIGKILL);
	    exit(1);
	} else {
	    FILE *fd;
	    struct mntent ent;
	    fd = setmntent("/etc/mtab", "a");
	    if ( fd > 0 ) { 
		ent.mnt_fsname="Odyssey";
		ent.mnt_dir=mnt;
		ent.mnt_type= "ody";
		ent.mnt_opts = "rw";
		ent.mnt_freq = 0;
		ent.mnt_passno = 0;
		error = addmntent(fd, & ent);
		error = endmntent(fd);
		exit(0);
	    }
	}
	exit(1);
    }
#endif

    workq = ds_safeq_create();
    ASSERT(workq);

    /* Fire off worker threads */
    for (i=0; i<numworkers; i++) {
	ASSERT(ds_thread_create(Worker, (char *)i, DS_THREAD_NORMAL,
				"worker", loglevel));
    }

    downcall_uniq = 1;
}

int
purge_fid(ofidp)
    struct ofid *ofidp;
{
    union ody_outputArgs msg;
    int                  result;

    msg.oh.opcode = ODY_PURGEFID;
    msg.oh.unique = downcall_uniq++;
    msg.ody_purgefid.fid = *ofidp;
    result = write(kfd, (char *)&msg, sizeof (struct ody_purgefid_out));
    if (result < 0) {
	perror("writing to kernel device");
	exit(errno);
    }
    if (result != sizeof (struct ody_purgefid_out)) {
	fprintf(stderr,"Incomplete write to kernel device, wanted %d got %d\n",
		sizeof(unsigned long)*2, result);
	exit(-2);
    }
    return 0;
}


int
flush_nc()
{
    struct ody_out_hdr msg;
    int                result;

    ds_thread_log(0, "flushing namecache");
    msg.opcode = ODY_FLUSH;
    msg.unique = downcall_uniq++;
    result = write(kfd, (char *)&msg, sizeof (struct ody_out_hdr));
    if (result < 0) {
	perror("writing to kernel device");
	exit(errno);
    }
    if (result != sizeof (struct ody_out_hdr)) {
	fprintf(stderr,"Incomplete write to kernel device, wanted %d got %d\n",
		sizeof(unsigned long)*2, result);
	exit(-2);
    }
    return 0;
}
