#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/queue.h>

#if defined(INSTRUMENTED) && !defined(__linux__)
#include <sys/dkstat.h>
#include <sys/vmmeter.h>
#endif

#include <sys/errno.h>
#include <netinet/in.h>
#ifdef NETTR_UPCALL
#include <nettr/replay/ntrp_upcall.h>
#endif
#ifdef RESLAN
#include <netinet/reslan_upcall.h>
#endif

#include "kernel/upcall/includes/upcall.h"
#include "include/odytypes.h"
#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>
#ifndef NOFAIL
#include <rpc2/fail.h>
#endif

#include "kernel/upcall/includes/upcall.h"

#include "base/ds/ds_log.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include "include/viceroy_onode.h"
#include "include/onode-manager.h"
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/cache_mgr.h"

#include "include/viceroy.h"
#include "include/downcall.h"
#include "vcroy.h"
#include "kernel.h"
#include "rsrc.h"
#include "rsrc_en.h"

#if defined(INSTRUMENTED) && !defined(__linux__)
#include "dkstats.h"
#endif

#ifdef T_MALLOC
#include <malloc.h>
#endif

/* Where to find wardens */
char *WardenDir = "/usr/odyssey/lib";

/* Cache parameters */
static char *RootDir = "/usr/odyssey/viceroy.cache";

/* Device for communication */
static char *kerndev = "/dev/ody0";
static char *mountpt = "/ody";

/* RPC2 & LWP parameters */
static int timeout_val = 15;      /* seconds */
static int retries = 5;           /* five retires XXX */
static unsigned short portal = VCROY_PORTAL_DFLT;
static int stacksize = 32768; /* default lwp stack size */
static int do_rpc2setlog = 0;

/* Global APM parameter */
int ViceroyAPMMode = 0; /* Default is do no apm */
float apm_halflife = 0.01; /* Default half-life */

/* Number of workers */
static int numworkers = 8;

/* logging level */
static int loglevel = 0;

/* odyssey, bo, or lf */
report_mode mode = odyssey;


#ifdef INSTRUMENTED

/* whether to dump statistics on each ntrace upcall, or not */
int dump_stats = 1;

#if defined(INSTRUMENTED) && !defined(__linux__)
/* whether to include the dkstats */
static int logdkstats = 0;
#endif

#define BLIND_OPTIMISM_BEST_BW 600000
int blind_optimism_bw = BLIND_OPTIMISM_BEST_BW;	/* starting value */
#endif /* INSTRUMENTED */

#ifdef RESLAN
unsigned int reslan_bandwidth = BLIND_OPTIMISM_BEST_BW; /* starting value */
#endif

void
usage() {
    fprintf(stderr,
    "viceroy\t[-timeout <seconds>] [-retries <num>] [-port <pnum>]\n");
    fprintf(stderr,
    "\t[-stacksz <bytes>] [-rootdir <dir>] [-wardendir <dir>]\n");
    fprintf(stderr,
    "\t[-kerndev <device>] [-mountpt <directory>] [-workers <num>]\n");
    fprintf(stderr,
    "\t[-loglevel <level>] [-rloglev <level>]\n");
    fprintf(stderr,
    "\t[-rinterval <sec> <usec>] [-rrho <percent>]\n");
    fprintf(stderr,
    "\t[-apmmode <level>] [-apmhalflife <value>]\n");
    fprintf(stderr,
    "\t[-rsrc_en_strategy [normal|min|max]\n");
    fprintf(stderr,
    "\t[-rpc2debug <level>] [-rpc2log]\n");
#ifdef INSTRUMENTED
    fprintf(stderr,
    "\t[-blind-optimism] [-laissez]\n");
    fprintf(stderr,
    "\t[-nodump]\n");
#ifndef __linux__
    fprintf(stderr,
    "\t[-logdkstats]\n");
#endif
#endif /* INSTRUMENTED */
#ifdef T_MALLOC
    fprintf(stderr,
    "\t[-maldebug <debug level>]\n");
#endif
    fprintf(stderr,
    "\t[-logbw] [-logmins <minutes>]\n");
    
    warden_usage(stderr);
    exit(-1);
}

void
ParseArgs(argc, argv) 
    int argc;
    char *argv[];
{
    int i;
    int result;
    int rho_pct;
    for (i = 1; i<argc; i++) {
	if (!strcmp(argv[i],"-help")) {
	    usage();
#ifdef INSTRUMENTED
	} else if (!strcmp(argv[i],"-nodump")) {
	    dump_stats = 0;
#ifndef __linux__
	} else if (!strcmp(argv[i],"-logdkstats")) {
	    logdkstats = 1;
#endif
#endif /* INSTRUMENTED */
	} else if (!strcmp(argv[i],"-timeout")) {
	    i++; if (i>=argc) usage();
	    timeout_val = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-retries")) {
	    i++; if (i>=argc) usage();
	    retries = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-port")) {
	    i++; if (i>=argc) usage();
	    portal = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-stacksz")) {
	    i++; if (i>=argc) usage();
	    stacksize = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-loglevel")) {
	    i++; if (i>=argc) usage();
	    loglevel = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rootdir")) {
	    i++; if (i>=argc) usage();
	    RootDir = argv[i];
	} else if (!strcmp(argv[i],"-wardendir")) {
	    i++; if (i>=argc) usage();
	    WardenDir = argv[i];
	} else if (!strcmp(argv[i],"-kerndev")) {
	    i++; if (i>=argc) usage();
	    kerndev = argv[i];
	} else if (!strcmp(argv[i],"-mountpt")) {
	    i++; if (i>=argc) usage();
	    mountpt = argv[i];
	} else if (!strcmp(argv[i],"-workers")) {
	    i++; if (i>=argc) usage();
	    numworkers = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rloglev")) {
	    i++; if (i>=argc) usage();
	    rsrc_loglevel = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rpc2log")) {
	    do_rpc2setlog = 1;
#if 0
	} else if (!strcmp(argv[i],"-eloglev")) {
	    i++; if (i>=argc) usage();
	    cmgr_rpc2_estimlevel = atoi(argv[i]);
#endif
	} else if (!strcmp(argv[i],"-apmmode")) {
	    i++; if (i>=argc) usage();
	    ViceroyAPMMode = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rsrc_en_strategy")) {
	    i++; if (i>=argc) usage();
	    if (rsrc_en_set_strategy(argv[i]) < 0) usage();
	} else if (!strcmp(argv[i],"-apm_halflife")) {
	    i++; if (i>=argc) usage();
	    apm_halflife = strtod(argv[i], NULL);
	} else if (!strcmp(argv[i],"-rpc2debug")) {
	    i++; if (i>=argc) usage();
	    RPC2_DebugLevel = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rinterval")) {
	    i++; if (i>=argc) usage();
	    rsrc_interval.tv_sec = atoi(argv[i]);
	    i++; if (i>=argc) usage();
	    rsrc_interval.tv_usec = atoi(argv[i]);
	} else if (!strcmp(argv[i],"-rrho")) {
	    i++; if (i>=argc) usage();
	    rho_pct = atoi(argv[i]);
	    rsrc_rho = (((double)rho_pct)/100.0);
#ifdef INSTRUMENTED
	} else if (!strcmp(argv[i],"-blind-optimism")) {
	    mode = blind_optimism;
	} else if (!strcmp(argv[i],"-laissez")) {
	    mode = laissez;
#endif /* INSTRUMENTED */
	} else if (!strcmp(argv[i],"-logbw")) {
	    rsrc_logbandwidth = 1;
	} else if (!strcmp(argv[i],"-logmins")) {
	    i++; if (i>=argc) usage();
	    rsrc_log_minutes = atoi(argv[i]);
#ifdef T_MALLOC
	} else if (!strcmp(argv[i],"-maldebug")) {
	    int mal_debug_lev;

	    i++; if (i>=argc) usage();
	    mal_debug_lev = atoi(argv[i]);
	    if (mal_debug_lev >= 0 && mal_debug_lev <= 3)
		mal_debug(mal_debug_lev);
#endif
	} else {
	    result = warden_parseargs(i, argc, argv);
	    if (result <= 0) usage();
	    i += (result - 1);
	}
    }
}

#ifdef INSTRUMENTED

void
do_dump(s)
    char *s;
{
    static int uniq = 0;
    FILE *fp;
    char filename[64];
    time_t t;
    struct tm *tp;

    if (time(&t) == -1) {
	fprintf(stderr,"dump_stats: time(): %s",strerror(errno));
	return;
    }
	
    t = time(NULL);
    tp = localtime(&t);
    sprintf(filename,"/tmp/warden.stats.%d.%d.%d:%d",uniq++,
	    tp->tm_yday, tp->tm_hour, tp->tm_min);

    if (!(fp = fopen(filename,"w"))) {
	fprintf(stderr,"dump_stats: %s: %s",filename,strerror(errno));
	fflush(stderr);
	return;
    }

    fprintf(fp, "%s", s);

    fprintf(fp, "RPCSend: %ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
	    rpc2_Sent.Total, rpc2_Sent.Retries, rpc2_Sent.Cancelled,
	    rpc2_Sent.Multicasts, rpc2_Sent.Busies, rpc2_Sent.Naks,
	    rpc2_Sent.Bytes);

    fprintf(fp, "RPCRcvd: %ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
	    rpc2_Recvd.Total, rpc2_Recvd.Giant, rpc2_Recvd.Replies,
	    rpc2_Recvd.Requests, rpc2_Recvd.GoodReplies,
	    rpc2_Recvd.GoodRequests, rpc2_Recvd.Multicasts,
	    rpc2_Recvd.GoodMulticasts, rpc2_Recvd.Busies,
	    rpc2_Recvd.GoodBusies, rpc2_Recvd.Bogus,
	    rpc2_Recvd.Naks, rpc2_Recvd.Bytes);
    
    fprintf(fp, "SFTPSent: %ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
	    sftp_Sent.Total, sftp_Sent.Starts, sftp_Sent.Datas,
	    sftp_Sent.DataRetries, sftp_Sent.Acks, sftp_Sent.Naks,
	    sftp_Sent.Busies, sftp_Sent.Bytes, sftp_Sent.Timeouts);

	    
    fprintf(fp, "SFTPRcvd: %ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
	    sftp_Recvd.Total, sftp_Recvd.Starts, sftp_Recvd.Datas,
	    sftp_Recvd.DataRetries, sftp_Recvd.Acks, sftp_Recvd.Naks,
	    sftp_Recvd.Busies, sftp_Recvd.Bytes, sftp_Recvd.Timeouts);

    
    warden_stats(fp);
    if (rsrc_logbandwidth)
	rsrc_dumplog();

    fclose(fp);
    
}

static void
user_dump()
{
    do_dump("Dumping on user request\n");
}    

#ifdef NETTR_UPCALL
static char tsev[2048];

#if defined(INSTRUMENTED) && !defined(__linux__)
static char dkbuf[1024];
#endif

static int
handle_upcall(upclass, upcall, argsz, argp, retszp, retp, flags)
    upclass_t upclass;
    upcall_t  upcall;
    size_t    argsz;
    void     *argp;
    size_t   *retszp;
    void     *retp;
    int       flags;
{
#ifdef NETTR_UPCALL
    struct ntrp_tev_arg *targp;
#endif
#if defined(INSTRUMENTED) && !defined(__linux__)
    struct dkstats dk;
#endif

    switch (upclass) {
#ifdef NETTR_UPCALL
    case UC_NTRP:
	break;
#endif
#ifdef RESLAN
    case UC_RESLAN:
        if (upcall == RL_BANDWIDTH_IS) {
	    reslan_bandwidth = ((struct rl_bwis_arg *) argp)->rlbw_bandwidth;
	    /* An upcall with a 0 bandwidth specifies no max. reservation */
	    if (!reslan_bandwidth) reslan_bandwidth = BLIND_OPTIMISM_BEST_BW;
	    return 0;
	} else {
	    fprintf(stderr,"handle_upcall: unkown UC_RESLAN upcall %d\n", 
		    upcall);
	    goto exit;	
	} /* endif */  
#endif
    default:
	fprintf(stderr,"handle_upcall: unknown upclass %d\n", upclass);
	goto exit;
	break;
    }

#ifdef NETTR_UPCALL
    switch (upcall) {
    case NTRP_NOTIFY_START:
	targp = (struct ntrp_tev_arg *)argp;
	if (logdkstats) {
	    dkreadstats(&dk);
	    dkprintstats(&dk, dkbuf);
	}
	sprintf(tsev,"Replay start event received: %ld.%06ld %s",
		targp->nta_tv.tv_sec, targp->nta_tv.tv_usec,
		ctime(&targp->nta_tv.tv_sec));
	if (rsrc_logbandwidth) {
	    rsrc_clearlog();
	}
	break;
    case NTRP_NOTIFY_END:
	targp = (struct ntrp_tev_arg *)argp;
	if (logdkstats) {
	    dkreadstats(&dk);
	    dkprintstats(&dk,dkbuf);
	}
	sprintf(tsev,"Replay end event received: %ld.%06ld %s",
		targp->nta_tv.tv_sec, targp->nta_tv.tv_usec,
		ctime(&targp->nta_tv.tv_sec));
	if (rsrc_logbandwidth) {
	    rsrc_dumplog();
	    rsrc_clearlog();
	}
	break;
    case NTRP_NOTIFY_TSEV:
	targp = (struct ntrp_tev_arg *)argp;
	if (logdkstats) {
	    dkreadstats(&dk);
	    dkprintstats(&dk, dkbuf);
	}
	sprintf(tsev,"Replay timestamp event received: %ld.%06ld %s", 
		targp->nta_tv.tv_sec, targp->nta_tv.tv_usec,
		ctime(&targp->nta_tv.tv_sec));
	break;
    default:
	fprintf(stderr,"handle_upcall: unkown UC_NTRP upcall %d\n", upcall);
	goto exit;
    }
#endif

#ifdef INSTRUMENTED
    if (mode == blind_optimism) {
	/* calculate the nominal bandwidth for the blind optimism case */
#ifdef NETTR_UPCALL
	if (upcall != NTRP_NOTIFY_END)
	    blind_optimism_bw = 100000000 / targp->nta_ser_ibt;
	else
#endif
	    blind_optimism_bw = BLIND_OPTIMISM_BEST_BW;
	fprintf(stderr, "Replay upcall: The gods declare the new"
		        " bandwidth to be %d. Huzzah!\n",
		        blind_optimism_bw);
    }
#endif /* INSTRUMENTED */
    
#if defined(INSTRUMENTED) && !defined(__linux__)
    if (dump_stats) {
        if (logdkstats) {
	    strcat(tsev,dkbuf);
        }
        do_dump(tsev);
    }
#endif

 exit:
    fflush(stderr);
    return 0;
}

#endif
#endif /* INSTRUMENTED */

/* exit cleanly so profiler will flush its output file */
void
exit_for_prof() {
    exit(0);
}

static void
sigs_and_upcalls()
{
#ifdef INSTRUMENTED
    int rc;
#endif

    /* make ^C trigger an exit() */
    if ((int)signal(SIGINT, exit_for_prof) == -1)
	perror("sigs_and_upcalls: signal SIGINT");

    if ((int)signal(SIGTERM, exit_for_prof) == -1)
	perror("sigs_and_upcalls: signal SIGTERM");

#if !defined(SIGUPCALL) && !defined(__linux__)
    /* If SIGUPCALL is not defined, receive upcalls on SIGPIPE */
    upcall_set_sig (SIGPIPE);
#endif

#ifdef INSTRUMENTED
    rc = (int)signal(SIGUSR1, rsrc_testfire);
    if (rc == -1) {
	perror("sigs_and_upcalls: signal USR1");
	exit(errno);
    }

    rc = (int)signal(SIGUSR2, rsrc_testreset);
    if (rc == -1) {
	perror("sigs_and_upcalls: signal USR2");
	exit(errno);
    }

#ifdef __linux__
    rc = (int)signal(SIGURG, user_dump);
    if (rc == -1) {
	perror("sigs_and_upcalls: signal SIGURG");
	exit(errno);
    }
#else
    rc = (int)signal(SIGINFO, user_dump);
    if (rc == -1) {
	perror("sigs_and_upcalls: signal SIGINFO");
	exit(errno);
    }
#endif


#ifdef NETTR_UPCALL
    /* Register upcalls for the same thing */
     if (upcall_reg(UC_NTRP, NTRP_NOTIFY_TSEV, handle_upcall) == UC_ERR)
	 fprintf(stderr,"upcall_reg: handle_upcall: can't register NTR_NOTIFY_TSEV upcall\n");
     if (upcall_reg(UC_NTRP, NTRP_NOTIFY_START, handle_upcall) == UC_ERR)
	 fprintf(stderr,"upcall_reg: handle_upcall: can't register NTR_NOTIFY_START upcall\n");
     if (upcall_reg(UC_NTRP, NTRP_NOTIFY_END, handle_upcall) == UC_ERR)
	 fprintf(stderr,"upcall_reg: handle_upcall: can't register NTR_NOTIFY_END upcall\n");
#endif

#endif /* INSTRUMENTED */

#ifdef RESLAN
     /* Register for reslan upcall */
     if (upcall_reg(UC_RESLAN, RL_BANDWIDTH_IS, handle_upcall) == UC_ERR)
	 fprintf(stderr,"upcall_reg: handle_upcall: can't register RL_BANDWIDTH_IS upcall\n");
#endif
}

/* 
 * Must be called after ds_thread_init()
 */
static void
VRPC_Init() {
    RPC2_PortIdent pid;
    SFTP_Initializer sftpi;
    struct timeval tov;

    SFTP_SetDefaults(&sftpi);
    SFTP_Activate(&sftpi); /* Initialize side effects SFTP package */

    pid.Tag = RPC2_PORTBYINETNUMBER;
    pid.Value.InetPortNumber = htons(portal);

    tov.tv_sec = timeout_val;
    tov.tv_usec = 0;

    ASSERT(RPC2_Init(RPC2_VERSION, NULL, &pid, retries, &tov)
	     == RPC2_SUCCESS);

#ifndef NOFAIL
    Fail_Initialize("Prototype viceroy", 0);
    Fcon_Init();
#endif

}

void
idle() {
    
}

int
main(argc, argv)
    int argc;
    char *argv[];
{
#ifdef T_MALLOC
    mal_debug(0);             /* make the Toronto malloc stay out of
                                 the way */
#endif

    ParseArgs(argc, argv);
    ds_thread_init(stacksize, loglevel);

#if ODYDEBUG
    /* Set up maximum thread wait time */
    cont_sw_threshold.tv_sec = 0;
    cont_sw_threshold.tv_usec = 100000;
#endif

    ASSERT(!chdir(RootDir));
    sigs_and_upcalls();
#if defined(INSTRUMENTED) && !defined(__linux__)
    if (logdkstats)
      dkinit();
#endif
    VRPC_Init();

    if (do_rpc2setlog)
      RPC2_SetLog(stderr, 10);

    APM_NW_INIT();
    onode_init();
    cmgr_init(CMGR_VERSION);
    warden_init(numworkers);
    kern_init(kerndev, mountpt, numworkers, loglevel);
    cache_init();
    rsrc_init();
    kern_loop();

    return 0;
}

