#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * storage.c -- storage and retrieval facilities for Onode Manager
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This is the allocation and retrieval code for an onode management
 * package for use by wardens.
 *
 * History
 *   31 Jul 1996 --  0. 0. 1 -- Colorful first attempt
 *
 * TODO
 *   * the initial implementation will be very brain dead, and just use
 *     a linked list. once it works, I'll shift over to a hash table, or
 *     some other more sophisticated data structure.
 *
 *   * pay attention to critical regions. specifically, pay attention to
 *     locks. do we need to observe locks before deallocing, or do we
 *     assume the programmer is allowed to shoot him/herself?
 *
 */

#define _OM_STORAGE_C

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "include/odytypes.h"
#include "include/viceroy_onode.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"

#include "include/onode-manager.h"
#include "include/downcall.h"

#include "om_storage.h"

ds_lock_t om_lock;		/* effectively, a monitor for the
				   onode manager. initialized in
				   onode_init() */

/* XXX -maybe we should use a hash table here? */
onode_record *the_onode_list = NULL;

/*
 * PUBLIC FUNCTIONS
 */

/*
 * onode_create -- given an ofid, allocates storage for it and returns
 *                 a pointer to that storage. The second argument is
 *                 a list of required parameters for onodes -- by
 *                 supplying these at creation time, we can be
 *                 relatively confident that all onodes are valid onodes.
 *                 if the onode already exists, return NULL.
 */
struct viceroy_onode *onode_create(struct ofid *new_ofid,
				   struct onode_create_in *in)
{
    onode_record *new_entry, *result;
    struct viceroy_onode *op;
    
    /* verify the ofid does not already exist */
    ds_lock_read(&om_lock);
    result = onode_lookup(new_ofid);
    ds_lock_release(&om_lock);
    if(result != NULL) {
	ds_thread_log(0, "onode_create called for valid onode! (%x.%x.%x.%x)",
		      new_ofid->ofid_ttype, new_ofid->ofid_tome, 
		      new_ofid->ofid_vnum, new_ofid->ofid_uniq);
	return(NULL);
    }
    
    /* allocate space for the new onode */
    ALLOC(new_entry, onode_record);
    ALLOC(op, struct viceroy_onode);
    
    /* fill in the new onode */
    bzero(op, sizeof(struct viceroy_onode));
    op->o_fid = *new_ofid;
    op->o_flags = in->oc_flags;
    op->o_vtype = in->oc_vtype;
    op->o_refcnt = 1;	 /* at least one reference to it, since it exists. */
    ds_hwlock_init(&op->o_lock);
    op->o_ofree = in->oc_ofree;
    op->o_data = in->oc_data;
    
    new_entry->onode = op;
    
    ds_lock_write(&om_lock);
    new_entry->next = the_onode_list; /* put it at the head of the list */
    new_entry->previous = NULL;	      /* since it's at head, there's
					 nothing before it */
    the_onode_list = new_entry;
    
    if(new_entry->next != NULL) {
	/* Set the old head of the list to point to what now precedes it */
	new_entry->next->previous = new_entry;
    }
    ds_lock_release(&om_lock);
    
    /* return the onode */
    ds_thread_log(100,"Onode (%x.%x.%x.%x) created: refcnt %d",	    
		  new_ofid->ofid_ttype, new_ofid->ofid_tome, 
		  new_ofid->ofid_vnum, new_ofid->ofid_uniq,
		  op->o_refcnt);
    
    return(new_entry->onode);
}

/*
 * onode_get -- given an ofid, returns the onode for it. if no such onode,
 *              returns NULL.
 */
struct viceroy_onode *onode_get(struct ofid *some_ofid)
{
    onode_record *entry;
    struct viceroy_onode *op;
    
    ds_lock_read(&om_lock);
    entry = onode_lookup(some_ofid);
    ds_lock_release(&om_lock);
    if(entry == NULL) {
	ds_thread_log(1000, "onode_get: ofid (%x.%x.%x.%x) NOT found",
		      some_ofid->ofid_ttype, some_ofid->ofid_tome,
		      some_ofid->ofid_vnum, some_ofid->ofid_uniq);
	return(NULL);
    }
    
    op = entry->onode;
    
    /* Here, we writelock om_lock instead of readlocking the onode,
       so that we don't violate the locking discipline. Since
       no one increments or decrements the reference count except
       for the onode manager, this is fine. */
    ds_lock_write(&om_lock);
    op->o_refcnt++;
    ds_lock_release(&om_lock);
    
    ds_thread_log(1000, "onode_get: ofid (%x.%x.%x.%x) found: refcnt %d",
		  some_ofid->ofid_ttype, some_ofid->ofid_tome,
		  some_ofid->ofid_vnum, some_ofid->ofid_uniq,
		  op->o_refcnt);
    return(op);
}

/*
 * onode_put -- given an onode, indicates that the calling function
 *              is no longer interested in it. should be called after
 *              any call to onode_create or onode_get.
 *
 *              (n.b. this decreases a reference count which, when it
 *              reaches zero, means that the onode is fair game for
 *              garbage collection.)
 */
int onode_put(struct viceroy_onode *some_onode)
{
    onode_record *entry;
    
    if(some_onode == NULL)
	return(-1);		/* XXX bad onode */
    
    ds_lock_write(&om_lock);	/* BEGIN critical region */

    entry = onode_lookup(&some_onode->o_fid);
    if(entry == NULL)
	ASSERT(0);		/* onode not in table! */
    
    ASSERT(some_onode == entry->onode);
    some_onode->o_refcnt--;
    
    ds_thread_log(1000, "onode_put: ofid (%x.%x.%x.%x) found: refcnt %d",
		  some_onode->o_fid.ofid_ttype, some_onode->o_fid.ofid_tome,
		  some_onode->o_fid.ofid_vnum, some_onode->o_fid.ofid_uniq,
		  some_onode->o_refcnt);
    
    if(some_onode->o_refcnt == 0) {
	/* Before calling onode_free, should hold both om_lock
	   and onode->o_lock */
	ds_hwlock_write(&some_onode->o_lock);
	purge_fid(&some_onode->o_fid);
	onode_free(entry);
	/* onode_free promises to unlock onode->o_lock */
    }

    ds_lock_release(&om_lock);	/* END critical region */
    
    return 0;
}

/*
 * PRIVATE FUNCTIONS
 */

/*
 * onode_lookup -- given an ofid, return a pointer to the onode. if
 *                 the onode does not currently exist, return NULL.
 * 
 * Must be called with om_lock held either read or write.
 */

onode_record *onode_lookup(struct ofid *someofid)
{
    onode_record *bookmark = the_onode_list;
    /* check to see if list is empty */
    if(the_onode_list == NULL) {
	return(NULL);
    }
    
    /* if not, search the list for a match */
    while(bookmark != NULL) {
	if(ofid_eq(someofid, &bookmark->onode->o_fid))
	    return(bookmark);
	else
	    bookmark = bookmark->next;
    } /* end while */
    
    return(NULL);
}

/*
 * onode_free -- Given an onode_entry, frees the space it uses.
 *
 * Must be called with om_lock held write.
 * Must be called with orp->o_lock held write.
 *
 * On exit, releases orp->o_lock (which is good, since orp will cease
 * to exist.)
 *
 * Returns the onode that this onode was mounted on top of, if such an
 * onode exists, otherwise returns null.  This is so that the caller,
 * who holds om_lock, can call put on the orphan, which will in turn
 * try to call om_lock.
 * 
 * XXX - should call the viceroy's kpurge for this onode as well.  
 *
 */

void onode_free(onode_record *orp)
{
    struct viceroy_onode *op;
    
    if(orp == NULL) {
	/* XXX - is there a good reason this should be harmless? */
	/* No lock to free */
	return;		/* fail silently, but harmlessly */
    }
    
    /* Alter the doubly linked list to remove this record from it */
    if(orp->previous != NULL) {
	orp->previous->next = orp->next;
    } else {
	/* this must be the current head of the list, so change the
	   head appropriately */
	if(the_onode_list == orp) {
	    the_onode_list = orp->next;
	} else {
	    /* corrupted doubly linked list. */
	    ASSERT(0);
	}
    }
    
    if(orp->next != NULL)
	orp->next->previous = orp->previous;
    
    /* Free the record */
    op = orp->onode;
    if ( (op->o_ofree) != NULL ) {
	/* only free the data if a freeing function has been supplied */
	(op->o_ofree)(op->o_data);
    }
    
    /* XXX - what about o_cachefile */
    ASSERT(op->o_owrite == 0);
    ASSERT(op->o_ocount == 0);
    
    /* Free the lock before returning */
    ds_hwlock_release_write(&op->o_lock);
    FREE(op);
    FREE(orp);
}

/*
 * ofid_eq -- given two fids, returns 1 if they are equal, and
 *            0 if they are not.
 */
int ofid_eq(struct ofid *ofid1, struct ofid *ofid2)
{
    if(ofid1->ofid_ttype != ofid2->ofid_ttype)
	return(0);
    
    if(ofid1->ofid_tome != ofid2->ofid_tome)
	return(0);
    
    if(ofid1->ofid_uniq != ofid2->ofid_uniq)
	return(0);
    
    if(ofid1->ofid_vnum != ofid2->ofid_vnum)
	return(0);
    
    return(1);
}
