#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * Root warden
 *
 * This handles accesses on the root of the Odyssey subtree, on which
 * all other tomes will be mounted.
 *
 * 19 Aug 1996 -- Supports open() and close() by redirecting to a location
 *                on local disk
 *
 * Specs for the functions provided here are in warden.h
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/fcntl.h>

#include <kernel/vfs/includes/ody.h>

#include "include/viceroy_onode.h"
#include "include/odytypes.h"
#include "base/ds/ds_list.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <lwp/lwp.h>
#include <rpc2/rpc2.h>

#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/onode-manager.h"

#include "root_warden.h"

static char *rootdir  = "/usr/odyssey/viceroy.cache/rootloc";

size_t tsopbufsize=32*1024;
static void *tsopbuf;

static uniq_t rw_uniq=1;

static int         rw_init       __P((struct warden *wp));
static int         rw_root       __P((struct warden *wp,
				      tome_t tome,
				      struct viceroy_onode **root_onodep));
static int         rw_lookup     __P((struct warden *wp,
				      struct viceroy_onode  *odp,
				      struct ody_cred  *cred,
				      char          *name,
				      char          **remain,
				      struct viceroy_onode  **opp));

static int         rw_access    __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         rw_open      __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         rw_close     __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         rw_tsop      __P((struct warden *wp, 
				     int           tsvnum, 
				     struct viceroy_onode  *op, 
				     struct ody_cred  *cred,
				     int           opnum, 
				     size_t        argsz, 
				     void          *argp,
				     size_t        *retsz, 
				     void          **retp));

static int     rw_tsop_dispose  __P((struct warden *wp, 
				     struct viceroy_onode  *op,
				     struct ody_cred  *cred,
				     int           opnum, 
				     size_t        argsz, 
				     void          *argp, 
				     size_t        retsz,
				     void          *retp));

static int         rw_getattr   __P((struct warden *wp,
				     struct viceroy_onode  *op,
				     struct ody_vattr  *vap,
				     struct ody_cred  *cred));

struct warden root_warden = {
    ROOT_TTYPE,
    "Root",
    0,
    wdflt_parseargs,
    "",
    rw_init,
    (w_getcid_fn) NULL,
    (w_pstats_fn) NULL,

    rw_root,
    rw_lookup,
    rw_access,
    rw_getattr,
    rw_open,
    rw_close,
    (w_rdwr_fn)NULL,
    (w_rdwr_dispose_fn)NULL,
    (w_readdir_fn)NULL,
    (w_readdir_dispose_fn)NULL,
    rw_tsop,
    rw_tsop_dispose,
#ifdef SETS
    (w_expand_fn)rw_lookup,
    (w_cursexpand_fn)NULL,
    (w_prefetch_fn)NULL,
#endif
    0,
    0,
    (void*)NULL			/* void *w_data */
};

enum ody_vtype
onode_type(struct stat *sbuf) {
    switch (sbuf->st_mode & S_IFMT) {
    case S_IFDIR:
	return O_VDIR;
	break;
    case S_IFREG:
	return O_VREG;
	break;
    case S_IFLNK:
	return O_VLNK;
	break;
    case S_IFSOCK:		/* Can't be any of these */
    case S_IFIFO:
    case S_IFCHR:
    case S_IFBLK:
	return O_VBAD;
	break;
    default:
	ASSERT(0);
	break;
    }
    return O_VBAD;
}

/* Warden-specific data for an onode */

struct onode_data {
    struct viceroy_onode *od_onode;
    char *od_name;		/* Component name */
    ds_list_t *od_kids;		/* List of children we know about */
    struct viceroy_onode *od_parent;	/* parent; if root, can be NULL */
};

/* free the warden-specific data associated with an onode */
static void rw_o_data_free(o_data)
    struct onode_data *o_data;
{
    if (o_data) {
	if (o_data->od_name)
	    free(o_data->od_name);
	if (o_data->od_kids)
	    ds_list_destroy(o_data->od_kids);
	free(o_data);
    }
}

int
rw_init(wp)
    struct warden *wp;
{
    struct ofid root_ofid;
    struct viceroy_onode *root_onode;
    struct onode_data *o_data;
    struct stat sbuf;
    struct onode_create_in root_onode_params;
    int result = 0;
    int root_fd;


    /* 
     * Can we actually get a handle on this object?
     */

    if (stat(rootdir, &sbuf)) {
	result = errno;
	ds_thread_log(10,"rw_init: root location '%s': %s",rootdir,
		      strerror(result));
	goto bad;
    }

    if ((root_fd = open(rootdir,O_RDONLY,0)) < 0) {
	result = errno;
	ds_thread_log(10,"rw_init: error: root location '%s': %s",rootdir,
		      strerror(result));
	goto bad;
    }
    close(root_fd);

    /* Is it a directory? */
    if (onode_type(&sbuf) != O_VDIR) {
	result = ENOTDIR;
	ds_thread_log(10,"rw_init: error: root location '%s' is not a directory",
		      rootdir);
	goto bad;
    }
  
    /* Construct the ofid  */

    root_ofid.ofid_ttype = ROOT_TTYPE;
    root_ofid.ofid_tome = ROOT_TOME;
    root_ofid.ofid_vnum = 0; /* sbuf.st_ino; */
    root_ofid.ofid_uniq = rw_uniq++;

    /* Construct bookkeeping data for the root onode */

    ALLOC(o_data, struct onode_data);
    o_data->od_name = NULL;	/* no component name for root */

    /* unordered, safe-destroy, possibly duplicated list of children */
    /* might have to introduce ordering later */
    o_data->od_kids = ds_list_create(NULL,TRUE,TRUE);
    o_data->od_parent = NULL;	/* no parent for root */

    /* Create the root onode */
    root_onode_params.oc_flags = ON_ROOTOFVOL;
    root_onode_params.oc_vtype = O_VDIR;
    root_onode_params.oc_ofree = rw_o_data_free;
    root_onode_params.oc_data = o_data;

    if ((root_onode = onode_create(&root_ofid, &root_onode_params)) == 0) {
	result = errno;
	ds_thread_log(0,"rw_init: onode_create() failed for root onode: %s",
		      strerror(result));
	goto bad;
    }
  
    /* Grab a lock on the root onode so we can play with it */
    onode_writelock(root_onode);
  
    o_data->od_onode = root_onode;

    /* Since the file itself resides on local disk, make it the cache file */
    root_onode->o_cachefile = strdup(rootdir);
    ASSERT(root_onode->o_cachefile);

    /* We're done fiddling with the onode, unlock it */
    onode_release_write(root_onode);
  
    /* Squirrel away root onode for rw_root() */
    wp->w_data = root_onode;

    /* Create the tsop buffer */
    BALLOC(tsopbuf,tsopbufsize);

    ds_thread_log(100,"rw_init: created root (%x.%x.%x.%x)",
		  root_ofid.ofid_ttype,root_ofid.ofid_tome,
		  root_ofid.ofid_vnum,root_ofid.ofid_uniq);


    return 0;			/* Everything is fine. */
  
 bad:
    ASSERT(result);

    /* We don't deallocate the onode because it doesn't hurt anyone:
       besides, if we come here it's an error condition anyway! */

    wp->w_data = NULL;
    return result;
}

int
rw_root(wp, tome, root_onodep)
    struct warden *wp;
    tome_t tome;
    struct viceroy_onode **root_onodep;
{
    struct viceroy_onode *root_onode = NULL;
    struct ofid root_ofid;
    int result = 0;

    ASSERT(tome == ROOT_TOME); /* we support only the one volume */

    /* wp->w_data should point to the root onode */
    root_onode = wp->w_data;

    /* Bump the refcount, because the caller is going to do a put */
    root_ofid = root_onode->o_fid;
    root_onode = onode_get(&root_ofid);

    if (root_onode == NULL) {
	result = errno;
	ds_thread_log(0,"rw_root: onode_get() failed for root onode: %s",
		      strerror(result));
	goto bad;
    }

    /* Return the onode to the viceroy read-locked */

    onode_readlock(root_onode);

    /* Everything's fine -- return locked, referenced root onode */
    *root_onodep = root_onode;

    ds_thread_log(100,"rw_root: success: ofid (%x.%x.%x.%x)",
		  root_ofid.ofid_ttype,root_ofid.ofid_tome,
		  root_ofid.ofid_vnum,root_ofid.ofid_uniq);

    return 0;
  
 bad:
    ASSERT(result);
    return result;
}


/* Given an onode to start from and a "pathname" string, do a lookup.
   The onode is pssed in read-locked and referenced, should be unlocked
   and put before returning. The onode that is returned should be returned
   read-locked and referenced */

/* Character that separates components */
const char c_sepr = '/';

int
rw_lookup(wp, odp, cred, name, remain, opp)
    struct warden *wp;
    struct viceroy_onode *odp;
    struct ody_cred *cred;
    char *name;
    char **remain;
    struct viceroy_onode **opp;
{
    unsigned long ttype, tome;
    int nresolved = 0;
    int result = 0;

    *opp = NULL;

    ds_thread_log(1000,"rw_lookup: '%s' on fid (%x.%x.%x.%x)",name,
		  odp->o_fid.ofid_ttype,odp->o_fid.ofid_tome,
		  odp->o_fid.ofid_vnum,odp->o_fid.ofid_uniq);
		

    /* Might get this, i.e. /ody/<warden>/.. */
    if (name == NULL || name[0] == '\0') {
        *remain = name;
	*opp = odp;
        return (0);
    }

    /* Try to resolve as many components as possible */

    /* At the beginning of this loop, we always have a read lock and a
       reference on odp. When exiting the loop also, we have a read
       lock and reference on odp. Thus, if the next component cannot
       be resolved, we break out of the loop without releasing the
       lock or reference on odp. If the next component can be resolved,
       we drop the lock and reference on odp, readlock and reference the
       new onode, point odp to this, and continue the loop */

    while (name && *name) {
	char *next_sep;
	struct ofid next_comp_ofid;
	struct onode_data *o_data = odp->o_data;

	*remain = name;

	/* We don't cross mount points, viceroy must do that */
	if ((odp->o_flags) & ON_MOUNTPT) {
	    result = EMOUNTPT;
	    break;
	}

	/* Make sure that this onode is a directory */
	if (odp->o_vtype != O_VDIR) {
	    result = ENOTDIR;
	    break;
	}

	/* get the name of the next component */
	next_sep = strchr(name,c_sepr);
	if (next_sep)
	    *next_sep = '\0';

	/* Step 0: Special case '.' and '..' */
	if (!strcmp(name,".")) {
	    next_comp_ofid = odp->o_fid;
	    /* we no longer nead the parent */
	    onode_release_read(odp);
	    onode_put(odp);
	}
	else if (!strcmp(name,"..")) {
	    struct viceroy_onode *o_parent = o_data->od_parent;
	    if (o_parent) {
		/* we no longer nead the parent */
		onode_release_read(odp);
		onode_put(odp);
		next_comp_ofid = o_parent->o_fid;
	    }

	    else {
		result = EMOUNTEDON;
		break;
	    }
	}
	else {			/* begin normal pathname lookup */
	    ds_list_iter_t *iter;
	    struct viceroy_onode *childp;

	    /* Step 1: have we resolved it before? */
	    iter = ds_list_iter_create(o_data->od_kids);
	    while ((childp = ds_list_iter_next(iter)) != NULL) {
		o_data = childp->o_data;
		if (!strcmp(o_data->od_name, name))
		    break;
	    }
	    ds_list_iter_destroy(iter);

	    if (childp) {
		next_comp_ofid = childp->o_fid;
		/* we no longer nead the parent */
		onode_release_read(odp);
		onode_put(odp);
	    }

	    else {		/* need to go to disk to find it */
		int name_len, ismountpt = 0;
		struct stat sbuf;
		struct onode_create_in child_onode_params;
		char path[ODY_MAXPATHLEN];

		/* Didn't find it */
		/* Step 2: look in the actual directory for the child */
		/* We cheat here: since the file is on disk, the
		   o_cachefile will be its pathname */

		strcpy(path,odp->o_cachefile);
		strcat(path,"/");
		strcat(path,name);

		/* If we can stat it, it exists */
		if (lstat(path,&sbuf) < 0) {
		    result = errno;
		    break;
		}

		/* Find if it's a mount point; we represent mount points
		   as symlinks to "#<ttype>.<tome>" (max length 15 bytes)
		   where <ttype> and <tome> are decimal integers */

		if (onode_type(&sbuf) == O_VLNK) {
		    char mountname[16]; /* cannot exceed this */

		    /* get contents of the symlink */

		    name_len = readlink(path,mountname,15);

		    /* check that it has valid mount point name syntax */
		    if (name_len < 16) { 
		    
			mountname[name_len] = '\0';

			if (sscanf(mountname,"#%lu.%lu", &ttype,&tome) == 2) {

			    ismountpt = 1;

			    /* Set up ofid and onode */

			    child_onode_params.oc_flags = ON_MOUNTPT;
			    child_onode_params.oc_vtype = O_VDIR;
			}
		    }
		}

		next_comp_ofid.ofid_ttype = ROOT_TTYPE;
		next_comp_ofid.ofid_tome = ROOT_TOME;
		next_comp_ofid.ofid_vnum = 0; /* sbuf.st_ino;*/
		next_comp_ofid.ofid_uniq = rw_uniq++;
		child_onode_params.oc_ofree = rw_o_data_free;
			    
		if (!ismountpt) {

		    /* Set up ofid and onode */

		    child_onode_params.oc_flags = 0;
		    child_onode_params.oc_vtype = onode_type(&sbuf);
	
		}

		ds_thread_log(100,"rw_lookup: create fid (%x.%x.%x.%x) ('%s')",
			      next_comp_ofid.ofid_ttype,
			      next_comp_ofid.ofid_tome,
			      next_comp_ofid.ofid_vnum,
			      next_comp_ofid.ofid_uniq,
			      name);

		/* Construct bookkeeping data for the onode */
	
		ALLOC(o_data, struct onode_data);
		o_data->od_name = strdup(name);

		/* unordered, safe-destroy, possibly duplicated list of
		   children */

		/* might have to introduce ordering later */
		o_data->od_kids = ds_list_create(NULL,TRUE,TRUE);
		o_data->od_parent = odp; /* no parent for root */

		child_onode_params.oc_data = o_data;

		if ((childp = onode_create(&next_comp_ofid,
					   &child_onode_params))
		    == 0) {
		    result = errno;
		    ds_thread_log(0,"rw_lookup(): onode_create() of child: %s",
				  strerror(result));
		    goto bad;
		}
  
		/* Add the child to parents list */

		onode_writelock(odp);
		o_data = odp->o_data;
		if (ds_list_insert(o_data->od_kids,childp) == NULL) {
		    result = errno;
		    ds_thread_log(0,"rw_lookup: ds_list_insert failed: %s",
				  strerror(result));
		    goto bad;
		}
		onode_release_write(odp);

		/* We're done with the parent */
		onode_release_read(odp);
		onode_put(odp);

		/* Lock and reference the new onode */

		onode_writelock(childp);

		o_data = childp->o_data;
		o_data->od_onode = childp;
		
		if (ismountpt) {
		    childp->o_u.ou_root.ofid_ttype = ttype;
		    childp->o_u.ou_root.ofid_tome = tome;
		}

		/* Since the file itself resides on local disk,
		   make it the cache file */

		childp->o_cachefile = strdup(path);
		ASSERT(childp->o_cachefile);
	
		/* We're done with fiddling with the onode, unlock it */
		onode_release_write(childp);

	    }

	}
    
	/* get the child (thus referencing it), and read-lock it */

	odp = onode_get(&next_comp_ofid);
	if (odp == NULL) {
	    result = errno;
	    ds_thread_log(0,"rw_root: onode_get() failed for root onode: %s",
			  strerror(result));
	    goto bad;
	}

	onode_readlock(odp);

	nresolved++;

	if (next_sep) {
	    *next_sep = c_sepr;
	    name = next_sep+1;
	}
	else			/* we resolved the last component */
	    name += strlen(name);
    }

    if (nresolved) {		/* we resolved at least one component */
	*remain = name;
	*opp = odp;
	ds_thread_log(100,"rw_lookup: resolved component #%d, (%x.%x.%x.%x)",
		      nresolved,
		      odp->o_fid.ofid_ttype,odp->o_fid.ofid_tome,
		      odp->o_fid.ofid_vnum,odp->o_fid.ofid_uniq);

	return 0;
    }


    /* If we couldn't resolve even the first component, unlock and put the
     onode that was handed to us*/

 bad:
    ASSERT(result);
    onode_release_read(odp);
    onode_put(odp);
    ds_thread_log(0,"rw_lookup: error: returning (%d)",result);
    return result;
}


/* Does user 'cred' have 'mode' permissions on object 'op' */

int
rw_access(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    /* 'Demo-only' code */
    /* O_A_READ/O_A_EXEC for all, O_A_WRITE for none.
       (note that O_A_READ is equiv. to the vnode mode
       VREAD, O_A_EXEC -> VEXEC, and O_A_WRITE -> VWRITE. */

    switch (mode) {
    case O_A_READ:
    case O_A_EXEC:
	return 0;
    case O_A_WRITE:
	return EACCES;
    }
    return EINVAL;
}

int rw_getattr(wp, op, vap, cred)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_vattr *vap;
    struct ody_cred *cred;
{
    struct stat  sbuf;
    int          rc;
    int          level=100;
    
    ds_thread_log(100,"rw_getattr on fid (%x.%x.%x.%x)",
		  op->o_fid.ofid_ttype,op->o_fid.ofid_tome,
		  op->o_fid.ofid_vnum,op->o_fid.ofid_uniq);

    rc = lstat(op->o_cachefile, &sbuf);
    if (rc) {
	goto exit;
    }

    /* 
     * XXX - if this is a link, we should examine the parent 
     * explicitly, but we don't, since we depend on the underlying UFS
     * to get this right.
     */
    vap->va_type = onode_type(&sbuf);
    vap->va_mode = sbuf.st_mode & 07777;
    vap->va_nlink = sbuf.st_nlink;
    vap->va_uid = sbuf.st_uid;
    vap->va_gid = sbuf.st_gid;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = sbuf.st_size;
    vap->va_blocksize = sbuf.st_blksize;
    vap->va_atime = sbuf.st_atime;
    vap->va_mtime = sbuf.st_mtime;
    vap->va_ctime = sbuf.st_ctime;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = sbuf.st_size;
    vap->va_filerev = op->o_fid.ofid_uniq;
    
 exit:
    if (rc) {
	level = 1;
    }
    ds_thread_log(level,"rw_getattr returns %d", rc);

    return rc;
}

/* Open an onode in the specified mode */

int
rw_open(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    struct stat sbuf;
    int result = 0;

    /* Current no writes are allowed */
    if (mode & O_O_WRITE) {
	result = EACCES;
	goto bad;
    }

    /* Currently, all files are cacheable, and the cache file name was
       stored in the onode at creation (i.e., first lookup) time */

    if (stat(op->o_cachefile,&sbuf) < 0) {
	result = errno;
	goto bad;
    }

    /* Tell the viceroy/kernel where the cache file is. Now, reads and writes
       will go to the cache file, we don't need them */

    op->o_device = sbuf.st_dev;
    op->o_inode = sbuf.st_ino;

    return 0;
 bad:
    ASSERT(result);
    return result;
}

int rw_close(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    /* Nothing to do: since all files are read-only and cacheable */
    return 0;
}

int rw_tsop(wp, tsvnum, op, cred, opnum, argsz, argp, retsz, retp)
    struct warden *wp;
    int           tsvnum;
    struct viceroy_onode  *op;
    struct ody_cred  *cred;
    int           opnum; 
    size_t        argsz; 
    void          *argp;
    size_t        *retsz; 
    void          **retp;
{
    /* Is this a read or write? */
    if (opnum == ROOT_TSOP_WRITE) {
	/* Do nothing */
	return 0;
    } else if (opnum == ROOT_TSOP_READ) {
	*retsz = (*(int*)argp);
	if (*retsz > tsopbufsize) {
	    *retsz = tsopbufsize;
	}
	*retp = tsopbuf;
	return 0;
    } else {
	return EOPNOTSUPP;
    }
}

int
rw_tsop_dispose(wp, op, cred, opnum, argsz, argp, retsz, retp)
    struct warden *wp; 
    struct viceroy_onode  *op;
    struct ody_cred  *cred;
    int           opnum; 
    size_t        argsz; 
    void          *argp; 
    size_t        retsz; 
    void          *retp;
{
    return 0;
}


