#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#ifdef __BSD44__
#include <sys/dkio.h>
#include <net/if.h>
#else
#include <asm/types.h>
#include <linux/hdreg.h>
#include <linux/wireless.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>
#include <string.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_list.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "rsrc.h"
#include "rsrc_en.h"
#include "bmod.h"

#define ATAPI_STANDBY_IMMEDIATE     0xe0
#define ATAPI_SLEEP         0xe6

static int s;
static int wl_inited = 0;
#ifdef __BSD44__
static struct ifreq ifr;
#else
static struct iwreq rq;
#endif

/* These variables are used for communication with the Energy Monitor */
static int               rsrc_en_socket = -1;
static int               rsrc_en_connection = -1;
static struct em_data_t* rsrc_en_pemdata = NULL;
static int               rsrc_en_hz = 1;

static int en_init = 0;

static float wa_power = 0.0;
static float wa_power_raw = 0.0; /* This is used for debugging */

static struct power_req_list_t* preqlist = 0;

extern float apm_halflife;

#define RSRC_EN_NORMAL  0
#define RSRC_EN_MINIMAL 1
#define RSRC_EN_MAXIMAL 2
static int rsrc_en_strategy = RSRC_EN_NORMAL;;

/* This is the extra overhead imposed by the interface to the data 
   collection computer.  We should subtract this out from the power
   measurements since it is an implementation artifact. */
#define RSRC_EN_OVERHEAD 0.333
#define RSRC_EN_CAP 15

/* Sets the power resource strategy.  The minimal and maximal strategies
   are provided mostly for experimental comparison purposes only. */
int
rsrc_en_set_strategy (char* arg) 
{
    if (!strcmp (arg, "normal")) {
	rsrc_en_strategy = RSRC_EN_NORMAL;
    } else if (!strcmp (arg, "min")) {
	rsrc_en_strategy = RSRC_EN_MINIMAL;
    } else if (!strcmp (arg, "max")) {
	rsrc_en_strategy = RSRC_EN_MAXIMAL;
    } else {
	return (-1);
    }
    
    return (0);
}

/* Initializes the energy resource request strucutre */
static void
rsrc_en_list_init ()
{
    if (!preqlist) {
	ALLOC (preqlist, struct power_req_list_t);
	ds_hwlock_init (&preqlist->pl_lock);
	preqlist->pl_head = preqlist->pl_tail = NULL;
    }
}

/* Finds the entry for the specified process in the energy resource 
   request structure.  Returns a pointer to the entry if the process 
   is found.  Returns NULL if the process is not found */
static struct power_req_t*
rsrc_en_list_find_by_pid (pid_t pid)
{
    struct power_req_t* ppr;

    ds_hwlock_read (&preqlist->pl_lock);
    for (ppr = preqlist->pl_head; ppr != NULL; ppr = ppr->pr_pnext) {
	if (ppr->pr_pid == pid) {
	    break;
	}
    }
    ds_hwlock_release_read (&preqlist->pl_lock);
    return (ppr);
}

/* Finds the entry for the specified request id in the energy resource 
   request structure.  Returns a pointer to the entry if the request id
   is found.  Returns NULL if the process is not found */
static struct power_req_t*
rsrc_en_list_find_by_reqid (int reqid)
{
    struct power_req_t* ppr;

    ds_hwlock_read (&preqlist->pl_lock);
    for (ppr = preqlist->pl_head; ppr != NULL; ppr = ppr->pr_pnext) {
	if (ppr->pr_prequest && ppr->pr_prequest->reqid == reqid) {
	    break;
	}
    }
    ds_hwlock_release_read (&preqlist->pl_lock);
    return (ppr);
}

/* Adds a new entry to the energy resource request structure. */
int
rsrc_en_list_add_to_top (struct power_req_t* ppr)
{
    ds_hwlock_write (&preqlist->pl_lock);
    
    ppr->pr_pnext = preqlist->pl_head;
    ppr->pr_pprev = NULL;
    if (preqlist->pl_head) {
	preqlist->pl_head->pr_pprev = ppr;
    }
    preqlist->pl_head = ppr;
    if (!preqlist->pl_tail) {
	preqlist->pl_tail = ppr;
    }
    
    ds_hwlock_release_write (&preqlist->pl_lock);

    return (0);
}

/* Adds a new entry to the energy resource request structure. */
int
rsrc_en_list_add_to_end (struct power_req_t* ppr)
{
    ds_hwlock_write (&preqlist->pl_lock);
    
    ppr->pr_pprev = preqlist->pl_tail;
    ppr->pr_pnext = NULL;
    if (preqlist->pl_tail) {
	preqlist->pl_tail->pr_pnext = ppr;
    }
    preqlist->pl_tail = ppr;
    if (!preqlist->pl_head) {
	preqlist->pl_head = ppr;
    }
    
    ds_hwlock_release_write (&preqlist->pl_lock);

    return (0);
}

/* Prints out the status of the energy resource request structure */
void
rsrc_en_list_print ()
{
    struct power_req_t* ppr;
    
    ds_hwlock_read (&preqlist->pl_lock);
    
    printf ("Energy Aware Applications:\n");
    for (ppr = preqlist->pl_head; ppr != NULL; ppr = ppr->pr_pnext) {
	printf ("pid %d: range [%d,%d], current %d,", ppr->pr_pid,
		ppr->pr_min_fidelity, ppr->pr_max_fidelity,
		ppr->pr_cur_fidelity);
	if (ppr->pr_cur_adaptation == RSRC_EN_NONE) {
	    printf (" not adapating ");
	} else {
	    printf (" adapting to level %d ", ppr->pr_cur_adaptation);
	}
	printf (", request ");
	if (ppr->pr_prequest) {
	    printf ("yes");
	} else {
	    printf ("no");
	}
	printf ("\n");
    }
    
    printf ("Backwards\n");
    for (ppr = preqlist->pl_tail; ppr != NULL; ppr = ppr->pr_pprev) {
	printf ("pid %d\n", ppr->pr_pid);
    }
    
    ds_hwlock_release_read (&preqlist->pl_lock);
}

int 
rsrc_en_cancel (int reqid, ds_request_t** req)
{
    struct power_req_t* ppr;
    int rc = 0;

    /* See if the request exists */
    ppr = rsrc_en_list_find_by_reqid (reqid);
    if (!ppr) return (DS_RRLIST_NOSUCHREQ);

    ds_hwlock_write (&preqlist->pl_lock);    

    *req = ppr->pr_prequest;
    ppr->pr_prequest = NULL;

    ds_hwlock_release_write (&preqlist->pl_lock);
    
    return (rc);
}

int
rsrc_en_purge (pid_t pid)
{
    struct power_req_t* ppr;
    int rc = 0;

    /* See if the process exists */
    ppr = rsrc_en_list_find_by_pid (pid);
    if (!ppr) return (DS_RRLIST_NOSUCHPID);

    ds_hwlock_write (&preqlist->pl_lock);    

    /* Remove the process entry */
    if (preqlist->pl_tail == ppr) {
	preqlist->pl_tail = ppr->pr_pprev;
    }
    if (preqlist->pl_head == ppr) {
	preqlist->pl_head = ppr->pr_pnext;
    }
    if (ppr->pr_pnext) {
	ppr->pr_pnext->pr_pprev = ppr->pr_pprev;
    }
    if (ppr->pr_pprev) {
	ppr->pr_pprev->pr_pnext = ppr->pr_pnext;
    }

    if (ppr->pr_prequest) {

	/* Free the request */
	if (ppr->pr_prequest->argp) {
	    (ppr->pr_prequest->argp->free_fn)
		(ppr->pr_prequest->argp->argval);
	    FREE(ppr->pr_prequest->argp);
	}
	FREE(ppr->pr_prequest);

    } else {
	rc = DS_RRLIST_NOSUCHPID;
    }

    FREE (ppr);

    ds_hwlock_release_write (&preqlist->pl_lock);
    
    return (rc);
}

/* This function handles an application placing a power resource request
   when the minimal strategy is employes.  If the request is not placed
   at the mimimum fidelity, then EOUTOFWINDOW id returned with value 0. */
static int
rsrc_place_power_minimal(wp, op, low, high, pid, fn, argp, 
			 cancelled, value, reqid)
    struct warden *wp;
    struct viceroy_onode *op;
    int           low;
    int           high;
    pid_t         pid;
    REQFNP        fn;
    void          *argp;
    int           *cancelled;
    int           *value;
    int           *reqid;
{
    if (high != 0) {
	*value = 0;
	return (ODY_ENOTINWINDOW);
    }

    *reqid = ds_rrlist_get_reqid();

    return (0);
}

/* This function handles an application placing a power resource request. */
int
rsrc_place_power(wp, op, low, high, pid, fn, argp, cancelled, value, reqid)
    struct warden *wp;
    struct viceroy_onode *op;
    int           low;
    int           high;
    pid_t         pid;
    REQFNP        fn;
    void          *argp;
    int           *cancelled;
    int           *value;
    int           *reqid;
{
    struct power_req_t *ppr = NULL;
    ds_request_t       *reqp=NULL;
    ds_rrlist_return_t rc = 0;

    /* Peculiar format of power requests requires that low == high */
    if (low != high) return (EINVAL);

    if (rsrc_en_strategy == RSRC_EN_MINIMAL) {
	return (rsrc_place_power_minimal (wp, op, low, high, pid, fn, argp, 
					  cancelled, value, reqid));
    }
    ALLOC(reqp, ds_request_t);
    DS_REQUEST_ALLOCATE(reqp, low, high, pid, fn, argp);

    /* Allocate a new request id */
    *reqid = ds_rrlist_get_reqid();
    reqp->reqid = *reqid;

    /* Check if this is the first request for the process */
    ppr = rsrc_en_list_find_by_pid (pid);
    if (ppr) {
	
	/* The application appears to have adapted on its own.  */
	/* I don't think this should happen in the simple prototype */
	if (ppr->pr_prequest) {

	    /* Duplicate request */
	    *cancelled = ppr->pr_prequest->reqid;

	    /* Free the old request */
	    ds_thread_log(0, "\tReplaced power resource request %d", 
			  ppr->pr_prequest->pid);
	    if (ppr->pr_prequest->argp) {
		(ppr->pr_prequest->argp->free_fn)
		    (ppr->pr_prequest->argp->argval);
		FREE(ppr->pr_prequest->argp);
	    }
	    FREE(ppr->pr_prequest);
	    
	    /* Insert the new request */
	    ppr->pr_prequest = reqp;
	    ppr->pr_cur_adaptation = RSRC_EN_NONE;
	    ppr->pr_cur_fidelity = low;

	} else {

	    /* The application is probably responding to an adaption
	       request.  Alternatively, it could have cancelled and
	       re-requested.  The latter should not happen in the
	       simple prototype. */
	    if (ppr->pr_cur_adaptation != RSRC_EN_NONE &&
		ppr->pr_cur_adaptation != low) {
	      ds_thread_log (0, "App. %d did not move to requested fidelity", pid);
	    }
		
	    /* Add the request */
	    ppr->pr_prequest = reqp;
	    ppr->pr_cur_adaptation = RSRC_EN_NONE;
	    ppr->pr_cur_fidelity = low;
	}

    } else {
	
	ALLOC(ppr, struct power_req_t);
	ppr->pr_pid = pid;
	ppr->pr_min_fidelity = 0;
	ppr->pr_max_fidelity = high;
	ppr->pr_cur_fidelity = high;
	ppr->pr_cur_adaptation = RSRC_EN_NONE;
	ppr->pr_prequest = reqp;
	ppr->pr_pnext = NULL;
	ppr->pr_pprev = NULL;
	rsrc_en_list_add_to_top (ppr);

    }

    return rc;
}

/*   
   This function is caled when less power must be used.  It selects an
   application that is not at minimum fidelity and nudges it to adapt
   downwards.  If all applications are at minimum fidelity, the
   function returns NO_MORE_ADAPTATION, otherwise it returns zero.  In
   the simple prototype, application priority is indicated by list
   order, with the lowest-priority applications at the front of the
   list.  Therefore, the function selects the first application that
   is not at its minimum fidelity level and asks it to adapt downward. 
*/
static int 
rsrc_en_adapt_down ()
{
    struct power_req_t* ppr;
    ds_request_t*       preq;
    ds_list_t*          to_notify;

    /* Find victim application */
    ds_hwlock_write (&preqlist->pl_lock);
    for (ppr = preqlist->pl_head; ppr != NULL; ppr = ppr->pr_pnext) {
	preq = ppr->pr_prequest;
	if (preq) {
	    if (preq->low > ppr->pr_min_fidelity) {
		
		/* Found victim - ask it to adapt down one notch */
		ppr->pr_cur_adaptation = preq->low - 1;

		/* Remove the request */
		ppr->pr_prequest = NULL;

		/* Notify the application */
		to_notify = ds_list_create(NULL,FALSE,FALSE);
		ds_list_insert(to_notify,preq);
		do_notify (ppr->pr_cur_adaptation, to_notify, ODY_GR_POWER);

		ds_hwlock_release_write (&preqlist->pl_lock);
		return (0);
	    }
	} else if (ppr->pr_cur_adaptation != RSRC_EN_NONE) {
	    ds_thread_log (0, "Application %d did not adapt promptly", 
			   ppr->pr_pid);
	}
    }

    /* No more victims */
    ds_hwlock_release_write (&preqlist->pl_lock);
    return (NO_MORE_ADAPTATION);
}

/* 
   This function is caled when more power can be used.  It selects an
   application that is not at full  fidelity and nudges it to adapt
   upwards.  If all applications are at full fidelity, the function
   returns NO_MORE_ADAPTATION, otherwise it returns zero.  In the
   simple prototype, application priority is indicated by list order,
   with the lowest-priority applications at the front of the list.  
   Therefore, the function selects the last application that is not
   at its maximum fidelity level and asks it to adapt upward.
*/
static int 
rsrc_en_adapt_up ()
{
    struct power_req_t* ppr;
    ds_request_t*       preq;
    ds_list_t*          to_notify;

    /* Find victim application */
    ds_hwlock_write (&preqlist->pl_lock);
    for (ppr = preqlist->pl_tail; ppr != NULL; ppr = ppr->pr_pprev) {
	preq = ppr->pr_prequest;
	if (preq) {
	    if (preq->high < ppr->pr_max_fidelity) {
		
		/* Found victim - ask it to adapt down one notch */
		ppr->pr_cur_adaptation = preq->high + 1;

		/* Remove the request */
		ppr->pr_prequest = NULL;

		/* Notify the application */
		to_notify = ds_list_create(NULL,FALSE,FALSE);
		ds_list_insert(to_notify,preq);
		do_notify (ppr->pr_cur_adaptation, to_notify, ODY_GR_POWER);

		ds_hwlock_release_write (&preqlist->pl_lock);
		return (0);
	    }
	} else if (ppr->pr_cur_adaptation != RSRC_EN_NONE) {
	    ds_thread_log (0, "Application %d did not adapt promptly", 
			   ppr->pr_pid);
	}
    }

    /* No more victims */
    ds_hwlock_release_write (&preqlist->pl_lock);
    return (NO_MORE_ADAPTATION);
}

static int
rsrc_en_update (float power)
{
    float total_joules, charge, alpha, supply, demand;
    float demand_raw; 
    float half_life;
    struct timeval recharge_time;
    float frecharge_time;
    int rc;
    static int rsrc_en_cap = 0;

    if (rsrc_en_cap > 0) rsrc_en_cap--;

    total_joules = bmod_get_total_joules();
    charge = bmod_get_charge ();
    recharge_time = bmod_get_time_to_recharge();
    frecharge_time = (float) recharge_time.tv_sec + 
	((float) recharge_time.tv_usec / 1000000.0);

    /* To calculate alpha: 
       1) Calculate the total time it should take to adapt from minimum
       to maximum application fidelity.  This is a fixed percentage of
       the time to recharge. 
       2) Calcualte the time it should take for each step in the adptation. 
       3) Set alpha so this value is a "half-life". 
    */
    half_life = frecharge_time * apm_halflife;
    alpha = pow (0.5, 1.0 / half_life);

    if (wa_power == 0.0) {
	wa_power = power;
    } else {
	wa_power = wa_power * alpha + power * (1.0-alpha);
    }

    if (wa_power_raw == 0.0) {
      wa_power_raw = power;
    } else {
	wa_power_raw = wa_power_raw * alpha + power * (1.0-alpha);
    }

    /* Recalculate supply and demand */
    supply = bmod_get_remaining_joules ();

    demand = frecharge_time * wa_power;
    demand_raw = frecharge_time * wa_power_raw;

    ds_thread_log (0, "Last power reading: %.3f\tcharge: %.3f\talpha: %.3f",
		       power, charge, alpha);
    ds_thread_log (0, "Time to recharge: %d.%d", recharge_time.tv_sec,
		   recharge_time.tv_usec);
    ds_thread_log (0, "Avg. power: %.3f\tsupply: %.3f\tdemand: %.3f",
		   wa_power, supply, demand);
    ds_thread_log (0, "Avg. raw power: %.3f\traw demand: %.3f",
		   wa_power_raw, demand_raw);

    /* See if any applications need to adapt */
    if (supply < demand) {

	/* Adapt downwards */
	rc = rsrc_en_adapt_down ();
	if (rc == 0) {
	    /* After successful adaptation, set the average power to 
	       the middle of the allowable range.  This reflects the
	       (hoped for) benefit of the adaptation */
	    wa_power = (supply - (EN_FUDGEC * total_joules)) / 
	      ((1.0 + (EN_FUDGE / 2.0)) * frecharge_time);
	    ds_thread_log (0, "Adapted downward - reset avg. power to %.3f",
			   wa_power);
	} else if (rc == NO_MORE_ADAPTATION) {
	  ds_thread_log (0, "RSRC_EN: We're as low as we can go!!!");
	} else {
	    ASSERT (0);
	}
	
    } else if (rsrc_en_cap == 0 && ((supply - (EN_FUDGEC * total_joules)) / 
	       demand > (1.0 + EN_FUDGE))) {

	/* Adapt upwards */
	rc = rsrc_en_adapt_up ();
	if (rc == 0) {
	    /* After successful adaptation, set the average power to 
	       the middle of the allowable range.  This reflects the
	       (hoped for) benefit of the adaptation */
	    wa_power = (supply - (EN_FUDGEC * total_joules)) / 
	      ((1.0 + (EN_FUDGE / 2.0)) * frecharge_time);
	    ds_thread_log (0, "Adapted upward - reset avg. power to %.3f",
			   wa_power);
	    rsrc_en_cap = RSRC_EN_CAP;

	} else if (rc == NO_MORE_ADAPTATION) {
	  ds_thread_log (0, "RSRC_EN: We're as high as we can go!!!");
	} else {
	    ASSERT (0);
	}
    }

    return (0);
}

/* This function sets up a socket which will be used to communicate with
   the Energy Monitor.  It does not attempt to establish the connection
   so it should not block. */
static int
rsrc_en_socket_init ()
{
    int sock;
    struct sockaddr_in my_addr;

    /* Create a socket with which to talk to Energy Monitor */
    sock = socket (PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock < 0) {
	ds_thread_log (0, "Unable to create energy mon. thread, errno = %d",
		       errno);
	return (sock);
    }

    my_addr.sin_family = AF_INET;
    my_addr.sin_port = htons (RSRC_EN_PORT);
    my_addr.sin_addr.s_addr = INADDR_ANY;
    if (bind (sock, (struct sockaddr *) &my_addr, 
	      sizeof(struct sockaddr_in)) < 0) {
	ds_thread_log (0, "Unable to bind energy monitor thread, errno = %d",
		       errno);
	close (sock);
	return (-1);
    }

    /* Only expect one connection on this socket - the Energy Monitor */
    if (listen (sock, 1) < 0) {
	ds_thread_log (0, "Unable to listen on energy mon. thread, errno = %d",
		       errno);
	close (sock);
	return (-1);
    }

    return (sock);
}

void
rsrc_en_init ()
{
    if (!en_init) {

	/* Initialize the energy resource request structure */
	rsrc_en_list_init ();

	/* Initialize the connection to the energy monitor */
	rsrc_en_socket = rsrc_en_socket_init ();

	bmod_init ();

	ds_thread_log (50, "rsrc_en: inited, socket = %d\n", rsrc_en_socket);
    } 
}

/* This function establishes the connection with the energy monitor. It 
   should only be called once a connection has been detected on the 
   socket - otherwise it will block. */
static int 
rsrc_en_socket_connect (int sock) 
{
    int conn;
    struct sockaddr_in client_addr;
    int client_addrlen = sizeof(client_addr);

    /* Wait for the Energy Monitor to contact us */
    conn = accept (sock, (struct sockaddr *) &client_addr, &client_addrlen);
    if (conn < 0) {
	ds_thread_log (0, "Unable to connect energy mon. thread, errno = %d",
		       errno);
	return (conn);
    }

    /* Make sure the socket is non-blocking */
    fcntl (conn, F_SETFL, O_NONBLOCK);

    ds_thread_log (50, "rsrc_en: connected, connection = %d", conn);

    return (conn);
}

/* This function tries to read the EM init data */
static int
rsrc_en_get_em_data (int conn)
{
    struct timeval tv = {0, 0};
    fd_set reads;
    int rc;

    /* See if there has been any activity on the socket */
    FD_ZERO (&reads);
    FD_SET (conn, &reads);
    if (select (conn+1, &reads, NULL, NULL, &tv) > 0) {

	/* Read init info from Energy Monitor */
	ALLOC (rsrc_en_pemdata, struct em_data_t);
	rc = read (conn, rsrc_en_pemdata, sizeof(struct em_data_t));
	if (rc != sizeof (struct em_data_t)) {
	    ds_thread_log (0, "Can't read EM init data, rc = %d, errno = %d", 
			   rc, errno);
	    /* If this ever happens, then we'll need to put a small wait
	       in here to allow data to come over the link. */
	    ASSERT (0);
	} else {
	    rsrc_en_hz = ((int) (1.0 / rsrc_en_pemdata->em_intvl)) + 1;
	    ds_thread_log (50, "EM data, scale = %f, intvl = %f, %d\n",
			   rsrc_en_pemdata->em_scale, 
			   rsrc_en_pemdata->em_intvl,
			   rc);
	    return (0); /* Got the data OK */
	}
    }
    
    /* No data yet */
    return (-1);
}

/* This function reads outstanding power samples from the energy monitor.
   It accumulates samples until enough are received for the power
   estimates to be updated.  When this occurs, the update function is
   called, and a new set of power samples starts accumulating.  The
   function return zero if everything is working correctly, and a negative
   value otherwise. 
*/
static int
rsrc_en_read_power (int conn)
{
    static float total_power = 0.0; /* No readings yet */
    static int   num_readings = 0;
    int   data;
    int   rc;

    do {
	rc = read (conn, &data, sizeof(data));
	if (rc == sizeof (data)) {

	    /* That's a power reading */
	    float power, energy;

	    data = ntohl (data);

	    /* Calculate instantaneous power */
	    power = DEFAULT_VOLTAGE * (float) data * 
		(float) rsrc_en_pemdata->em_scale;

	    /* Subtract out interface overhead */
	    power -= RSRC_EN_OVERHEAD;

	    ds_thread_log (100, "data: %d, power %f", data, power);

	    /* Adjust battery modulation correctly */
	    energy = power * rsrc_en_pemdata->em_intvl;
	    bmod_drain (energy);

	    total_power += power;
	    num_readings++;
	    if (num_readings == rsrc_en_hz) {
		if (rsrc_en_strategy != RSRC_EN_MINIMAL &&
		    rsrc_en_strategy != RSRC_EN_MAXIMAL) {
		    /* No adaptation needed for minimal/maximal strategies */
		    rsrc_en_update (total_power / (float) num_readings);
		}
		total_power = 0.0;
		num_readings = 0;
	    }

	} else if (rc > 0) {
	    /* Returned an odd number of bytes - this shouldn't happen */
	    ds_thread_log (0, "Read an odd number of bytes from EM");
	    ASSERT (0);
	} else if (errno == EAGAIN) {
	    /* Expected - no more data to read */
	} else if (rc <= 0) {
	    ds_thread_log (0, "Unable to read energy mon. data, errno = %d",
			   errno);
	    return (-1);
	} else {
	    ASSERT (0);
	}
    } while (rc == sizeof(data));

    /* Return average power */
    return (0);
}

/* This is the heart of power monitoring.  It makes sure that the connection
   with the energy monitor is up and running.  Then reads the average 
   power consumption over the last period. */
void 
rsrc_en_monitor ()
{
    struct timeval tv = {0, 0};
    fd_set reads;

    /* Socket was never established - can't do anything */
    if (rsrc_en_socket < 0) {
	ds_thread_log (50, "rsrc_en_monitor: socket not established");
	return;
    }
    
    /* If no connection yet, see if we can establish it */
    if (rsrc_en_connection < 0) {
	/* Select on the socket */
	FD_ZERO (&reads);
	FD_SET (rsrc_en_socket, &reads);
	if (select (rsrc_en_socket+1, &reads, NULL, NULL, &tv) > 0) {
	    rsrc_en_connection = rsrc_en_socket_connect (rsrc_en_socket);
	    if (rsrc_en_connection < 0) {
		close (rsrc_en_socket);
		rsrc_en_socket = rsrc_en_connection;
		ds_thread_log (50, "rsrc_en_monitor: no connection");
		return;
	    }
	} else {
	    ds_thread_log (50, "rsrc_en_monitor: no connection yet\n");
	    return;
	}
    }

    /* If not already arrived, see if the Energy Monitor init data is there */
    if (!rsrc_en_pemdata) {
	if (rsrc_en_get_em_data (rsrc_en_connection) < 0) {
	    ds_thread_log (50, "rsrc_en_monitor: no init data yet");
	    return;
	}
    }

    /* All set - read as much data as has arrived */
    if (rsrc_en_read_power (rsrc_en_connection) < 0) {
	/* Fatal error - close the connection */
        ds_thread_log (0, "rsrc_en_monitor: closing connection");
	close (rsrc_en_connection);
	rsrc_en_connection = -1;
    }
}


/* 
 * wl_apm_init
 *
 * Perform one-time initialization on first call to a wl APM function.
 */
int 
wl_apm_init ()
{
    s = socket(AF_INET, SOCK_DGRAM, 0);
    if (s < 0) {
	ds_thread_log (0, "Unable to open APM socket, errno = %d\n", errno);
	return (-1);
    }

#ifdef __BSD44__
    ifr.ifr_addr.sa_family = AF_INET;
    strcpy (ifr.ifr_name, WL_NAME);

    if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0) {
	ds_thread_log (0, "Unable to get flags for interface %s, errno = %d\n",
		       WL_NAME);
	return (-1);
    }
#else
    strcpy (rq.ifr_ifrn.ifrn_name, WL_NAME);
#endif

    wl_inited = 1;

    return (0);
}

/* 
 * wl_apm_sleep
 *
 * Changes the APM status of the wavelan by putting it to sleep or by
 * waking it up.
 */
int 
wl_apm_sleep (int action) 
{
    int rc;
    if (!wl_inited) {
	rc = wl_apm_init();
	if (rc < 0) return (rc);
    }

    if (action == APM_WAKEUP) {
#ifdef __BSD44__
	ifr.ifr_flags |= IFF_UP;
#else
	rq.u.name[0] = APM_RESUME;
#endif
    } else if (action == APM_SLEEP) {
#ifdef __BSD44__
	ifr.ifr_flags |= IFF_UP; /* Down? */
#else
	rq.u.name[0] = APM_STANDBY;
#endif
    } else {
        close (s);
	return (EINVAL);
    }
  
#ifdef __BSD44__
    if (ioctl(s, SIOCAPM, (caddr_t)&ifr) < 0) {
#else
    if (ioctl(s, SIOCAPM, (caddr_t)&rq) < 0) {
#endif
	ds_thread_log (0, "Error updating state of interface %s, errno = %d\n",
		       WL_NAME, errno);
	return (-1);
    }

    return (0);
}

/* 
 * wd_apm_standby
 *
 * Changes the APM status of the hard drive by putting it into standby mode.
 */
int
wd_apm_standby ()
{
  int fd;
  int rc;
#ifdef __BSD44__
  int val = ATAPI_STANDBY_IMMEDIATE;
#else
  unsigned char val[4] = { ATAPI_STANDBY_IMMEDIATE, 0, 0, 0};
#endif

  fd = open (WD_DEV, O_RDWR, 0);
  if (fd < 0) {
      ds_thread_log (0, "Error opening %s, errno = %d\n", WD_DEV, errno);
      return (-1);
  }
 
  rc = ioctl (fd, WD_CMD, &val);
  if (rc < 0) {
      ds_thread_log (0, "Error updating APM state of %s, errno = %d\n", WD_DEV,
		     errno);
  }

  close (fd);
  return (rc);
}

/* 
 * wd_apm_sleep
 *
 * Changes the APM status of the hard drive by putting it into sleep mode.
 */
int
wd_apm_sleep ()
{
  int fd;
  int rc;
#ifdef __BSD44__
  int val = ATAPI_SLEEP;
#else
  unsigned char val[4] = { ATAPI_SLEEP, 0, 0, 0};
#endif

  fd = open (WD_DEV, O_RDWR, 0);
  if (fd < 0) {
      ds_thread_log (0, "Error opening %s, errno = %d\n", WD_DEV, errno);
      return (-1);
  }
 
  rc = ioctl (fd, WD_CMD, &val);
  if (rc < 0) {
      ds_thread_log (0, "Error updating APM state of %s, errno = %d\n", WD_DEV,
		     errno);
  }

  close (fd);
  return (rc);
}

static int apm_nw_cnt = 0;
extern int ViceroyAPMMode;

void APM_NW_INIT ()
{
    if (ViceroyAPMMode > 0) {
	apm_nw_cnt = 0;
	wl_apm_sleep (APM_SLEEP);
    }
}

void APM_NW_UP ()
{
    if (ViceroyAPMMode > 0) {
	if (apm_nw_cnt == 0) {
	    wl_apm_sleep (APM_WAKEUP);
	}
	apm_nw_cnt++;
    }
}

void APM_NW_DOWN ()
{
    if (ViceroyAPMMode > 0) {
	apm_nw_cnt--;
	ASSERT (apm_nw_cnt >= 0);
	if (apm_nw_cnt == 0) {
	    wl_apm_sleep (APM_SLEEP);
	}
    }
}

void APM_NW_CONN_UP (cmgr_id_t cid)
{
    if (!cmgr_islocal (cid)) {
	APM_NW_UP ();
    }
}

void APM_NW_CONN_DOWN (cmgr_id_t cid)
{
    if (!cmgr_islocal (cid)) {
	APM_NW_DOWN ();
    }
}







