#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include <sys/errno.h>

#include "kernel/vfs/includes/ody.h"
#include "include/viceroy_onode.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/onode-manager.h"

#include "include/odyssey.h"

#include "root_warden.h"
#include "vfscalls.h"

int
vfs_root(fidp)
    struct ofid *fidp;
{
    struct warden *wp;
    struct viceroy_onode  *op;
    int           rc;

    wp = warden_get(ROOT_TTYPE);
    ASSERT(wp);

    /* Do the call */
    do {
	rc = (wp->w_root)(wp, ROOT_TOME, &op);
	if (rc) {
	    ds_thread_log(0, "vfs_root: can't get root onode: (%d) %s", 
			  rc, strerror(rc));
	}
    } while (rc);

    *fidp = op->o_fid;
    /* Ensure that the onde returned is actually marked as a root */
#ifdef ODYDEBUG
    ASSERT(IS_AROOT(op));
#endif
    onode_release_read(op);
    onode_put(op);

    ds_thread_log(100, "vfs_root: returns (%x.%x.%x.%x)",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq);
    return 0;
}

int
vfs_lookup(dirp, name, fidp, vtp, cred)
    struct ofid *dirp;
    char        *name;
    struct ofid *fidp;
    enum ody_vtype *vtp;
    struct ody_cred *cred;
{
    char *remain = NULL; /* XXX - not currently useful */
    struct warden *wp;
    struct viceroy_onode *odp;
    struct viceroy_onode *op;
    char          *sname;
    int           rc;

    wp = warden_get(dirp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_lookup: cannot find warden for 0x%x!", 
		      dirp->ofid_ttype);
	return ENXIO;  /* Warden died/not there. */
    }

    odp = onode_get(dirp);
    if (!odp) {
	ds_thread_log(1,"vfs_lookup: cannot find onode for (%x.%x.%x.%x)!", 
		      dirp->ofid_ttype, dirp->ofid_tome,
		      dirp->ofid_vnum, dirp->ofid_uniq);
	return ENOENT;
    }
    onode_readlock(odp);

    /* 
     * Special case: odp is a root of a volume, and the first
     * component is '..'  In this case, we uncover the mount point
     * here in the viceroy, since it will be the common case.
     */
    if (IS_AROOT(odp)) {
	sname = name;
	while (name && *name == '/') name++;

	if (name && (!strcmp("..",name) 
		     || !strncmp("../", name, strlen("../"))))
	{
	    /* We are crossing upwards from a root. */
	    name += strlen("..");
	    while (name && *name == '/') name++;
	    /* Copy the fid name */
	    *dirp = odp->o_u.ou_mountedon;
	    onode_release_read(odp);
	    /* Do the new lookup */
	    return (vfs_lookup(dirp, name, fidp, vtp, cred));
	}
    }
       
    /* Make the call */
    /* 
     * XXX odp should (potentially) be writelocked here, since we'll
     * be changing it's last access time.  
     */
    rc = (wp->w_lookup)(wp, odp, cred, name, &remain, &op);
    /* odp's lock and reference are released by the warden */
    if (rc) {
	ds_thread_log(100, 
		      "vfs_lookup of %s in (%x.%x.%x.%x) failed: (%d) %s", 
		      name, dirp->ofid_ttype, dirp->ofid_tome,
		      dirp->ofid_vnum, dirp->ofid_uniq,
		      rc, strerror(rc));
	goto exit;
    }

    /* Check to see if this is a mountpoint */
    if (IS_MOUNTEDON(op)) {
	struct viceroy_onode  *rop;
	struct warden *rwp;
	struct ofid    rfid;
	
	rfid = op->o_u.ou_mountedon;
	rwp = warden_get(rfid.ofid_ttype);

	if (!rwp) { /* warden died/not there, don't try to cross mount point */
		
	    ds_thread_log(0,"%s (%x.%x.%x.%x) mounted on (%x.%x.%x.%x)",
			  "vfs_lookup could not cross mount point",
			  op->o_fid.ofid_ttype, op->o_fid.ofid_tome,
			  op->o_fid.ofid_vnum,  op->o_fid.ofid_uniq,
			  rfid.ofid_ttype, rfid.ofid_tome,
			  rfid.ofid_vnum, rfid.ofid_uniq);
	} else {
	    /* We need to grab the root */
	    rc = (rwp->w_root)(rwp, rfid.ofid_tome, &rop);

	    /* Drop the op reference */
	    onode_release_read(op);
	    onode_put(op);

	    if (rc) goto exit;
	    /* Ensure that the onde returned is actually marked as a root */
#ifdef ODYDEBUG
	    ASSERT(IS_AROOT(rop));
#endif
	    op = rop;
	}
    }

    *fidp = op->o_fid;
    *vtp = op->o_vtype;
    onode_release_read(op);
    onode_put(op);
    ds_thread_log(100, "vfs_lookup of %s in (%x.%x.%x.%x) succeeded",
		  name, dirp->ofid_ttype, dirp->ofid_tome,
		  dirp->ofid_vnum, dirp->ofid_uniq);
    ds_thread_log(100, "\t\t=>(%x.%x.%x.%x)",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq);		  
 exit:
    return (rc);
}

int
vfs_access(fidp, mode, cred)
    struct ofid  *fidp;
    int          mode;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;
    
    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_access: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_access: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    onode_readlock(op);
    rc = (wp->w_access)(wp, op, cred, mode);
    onode_release_read(op);
    onode_put(op);

    ds_thread_log(100,"vfs_access on (%x.%x.%x.%x), mode %x returns (%d) %s",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, mode, rc,
		  (rc) ? strerror(rc) : "");
		  
    return (rc);
}

int
vfs_open(fidp, flags, devp, inop, cred, uncacheable)
    struct ofid *fidp;
    int         flags;
    dev_t       *devp;
    ino_t       *inop;
    struct ody_cred *cred;
    int         *uncacheable;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int          rc;
    
    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_open: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_open: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }

    onode_writelock(op);
    rc = (wp->w_open)(wp, op, cred, flags);
    if (rc) {
	goto exit;
    }
    onode_open(op, flags);
    if (IS_UNCACHEABLE(op)) {
	ASSERT(wp->w_rdwr);   /* Better be able to cope */
	*uncacheable = 1;
	*devp = 0;
	*inop = 0;
    } else {
	*uncacheable = 0;
	*devp = op->o_device;
	*inop = op->o_inode;
    }

 exit:
    onode_release_write(op);
    onode_put(op);

    ds_thread_log(100,"vfs_open on (%x.%x.%x.%x), flags %x returns (%d) %s",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, flags, rc,
		  (rc) ? strerror(rc) : "");
    ds_thread_log(100,"\tUsing (%d, %d) as cache file",
		  *devp, *inop);
		  
    return rc;
}

int
vfs_close(fidp, flags, cred)
    struct ofid *fidp;
    int         flags;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int          rc;
    
    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"Cannot find warden for 0x%x!", fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"Cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }

    onode_writelock(op);
    rc = (wp->w_close)(wp, op, cred, flags);
    if (rc) {
	goto exit;
    }
    onode_close(op, flags);

 exit:
    onode_release_write(op);
    onode_put(op);

    ds_thread_log(100,"vfs_close on (%x.%x.%x.%x), flags %x returns (%d) %s",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, flags, rc,
		  (rc) ? strerror(rc) : "");
		  
    return rc;
}

int vfs_getattr(fidp, attrp, cred)
    struct ofid  *fidp;
    struct ody_vattr *attrp;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode  *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_getattr: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    if (!wp->w_getattr) {
	return EOPNOTSUPP;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_getattr: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    onode_readlock(op);
    rc = (wp->w_getattr)(wp, op, attrp, cred);
    onode_release_read(op);
    onode_put(op);

    ds_thread_log(100,"vfs_getattr on (%x.%x.%x.%x) returns (%d) %s",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq,
		  rc, (rc) ? strerror(rc) : "");
    
    return (rc);

}

int vfs_rdwr(fidp, rwflag, count, offset, ioflag, buf, cred)
    struct ofid   *fidp;
    int           rwflag;
    int           *count;
    int           offset;
    int           ioflag;
    void          **buf;
    struct ody_cred  *cred;
{
    struct warden *wp;
    struct viceroy_onode  *op;
    int           ocount = *count;
    void          *obuf = *buf;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_rdwr: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_rdwr: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    
    onode_writelock(op);
    rc = (wp->w_rdwr)(wp, op, cred, rwflag, count, offset, ioflag, buf);
    onode_release_write(op);
    onode_put(op);
    
    ds_thread_log(100, "vfs_rdwr on (%x.%x.%x.%x), at %d for %d ",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, offset, ocount);
    ds_thread_log(100, "returns %d (bytes %d)\n",
		  rc, *count);

    if (*buf != obuf) {
	ds_thread_log(100, "vfs_rdwr changed buf pointer from 0x%x to 0x%x\n",
		      obuf, *buf);
    }
   
    return rc;
}

int
vfs_rdwrdisp(fidp, count, offset, buf, cred)
    struct ofid  *fidp;
    int          count;
    int          offset;
    void         *buf;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode  *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_rdwrdisp: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_rdwrdisp: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    
    onode_writelock(op);
    rc = (wp->w_rdwrdisp)(wp, op, cred, count, offset, buf);
    onode_release_write(op);
    onode_put(op);
    
    ds_thread_log(100, "vfs_rdwrdisp on (%x.%x.%x.%x), disposes 0x%x\n",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, buf);
    
    return rc;
    
}

int vfs_readdir(fidp, count, offset, buf, cred)
    struct ofid   *fidp;
    int           *count;
    int           offset;
    struct ody_dirent **buf;
    struct ody_cred  *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           ocount = *count;
    struct ody_dirent *obuf = *buf;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_readdir: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    if (!wp->w_readdir) {
	return EOPNOTSUPP;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_readdir: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    
    onode_readlock(op);
    rc = (wp->w_readdir)(wp, op, cred, count, offset, buf);
    onode_release_read(op);
    onode_put(op);
    
    ds_thread_log(100, "vfs_readdir on (%x.%x.%x.%x), at %d for %d ",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, offset, ocount);
    ds_thread_log(100, "returns %d (bytes %d)\n",
		  rc, *count);

    if (*buf != obuf) {
	ds_thread_log(100, "vfs_readdir changed buf pointer from 0x%x to "
		      "0x%x\n",
		      obuf, *buf);
    }
   
    return rc;
}

int
vfs_readdirdisp(fidp, count, offset, buf, cred)
    struct ofid  *fidp;
    int          count;
    int          offset;
    void         *buf;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_readdirdisp: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_readdirdisp: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    
    onode_readlock(op);
    rc = (wp->w_readdirdisp)(wp, op, cred, count, offset, buf);
    onode_release_read(op);
    onode_put(op);
    
    ds_thread_log(100, "vfs_readdirdisp on (%x.%x.%x.%x), disposes 0x%x\n",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, buf);
    
    return rc;
    
}

int
vfs_tsop(fidp, gs, type, ts, opnum, argsz, argp, retsz, retpp, cred)
    struct ofid  *fidp;
    unsigned int gs;
    unsigned int type;
    unsigned int ts;
    unsigned int opnum;
    size_t       argsz;
    void         *argp;
    size_t       *retsz;
    void         **retpp;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;

    
    if ((gs != ODY_GENERIC_INTERFACE)
	|| (type != fidp->ofid_ttype))
    {
	ds_thread_log(0, "vfs_tsop: bad version (%d not %d) or (%d not %d)",
		      gs, ODY_GENERIC_INTERFACE, type, fidp->ofid_ttype);
	return ODY_EBADVERSION;
    }


    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_tsop: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_tsop: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }

    onode_writelock(op);
    rc = (wp->w_tsop)(wp, ts, op, cred, opnum, argsz, argp, retsz, retpp);
    onode_release_write(op);
    onode_put(op);

    ds_thread_log(100,"vfs_tsop on (%x.%x.%x.%x), op %d returns %d",
		  fidp->ofid_ttype, fidp->ofid_tome, fidp->ofid_vnum, 
		  fidp->ofid_uniq, opnum, rc);

    return rc;
}


int
vfs_dispose(fidp, opnum, argsz, argp, retsz, retp, cred)
    struct ofid *fidp;
    unsigned int opnum;
    size_t argsz;
    void *argp;
    size_t retsz;
    void *retp;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_dispose: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_dispose: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }

    onode_writelock(op);
    rc = (wp->w_tsopdisp)(wp, op, cred, opnum, argsz, argp, retsz, retp);
    onode_release_write(op);
    onode_put(op);

    ds_thread_log(100, "vfs_dispose on (%x.%x.%x.%x), op %d returns %d",
		  fidp->ofid_ttype, fidp->ofid_tome, fidp->ofid_vnum, 
		  fidp->ofid_uniq, opnum, rc);
    
    return rc;
}

#ifdef SETS
int
vfs_expand(dirp, name, fidp, vtp, curs, cred)
    struct ofid *dirp;
    char        *name;
    struct ofid *fidp;
    enum ody_vtype *vtp;
    int         *curs;
    struct ody_cred *cred;
{
    char *remain = NULL; /* XXX - not currently useful */
    struct warden *wp;
    struct viceroy_onode *odp;
    struct viceroy_onode *op;
    char          *sname;
    int           rc;

    wp = warden_get(dirp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_expand: cannot find warden for 0x%x!", 
		      dirp->ofid_ttype);
	return ENXIO;  /* Warden died/not there. */
    }

    odp = onode_get(dirp);
    if (!odp) {
	ds_thread_log(1,"vfs_expand: cannot find onode for (%x.%x.%x.%x)!", 
		      dirp->ofid_ttype, dirp->ofid_tome,
		      dirp->ofid_vnum, dirp->ofid_uniq);
	return ENOENT;
    }
    onode_readlock(odp);

    /* 
     * Special case: odp is a root of a volume, and the first
     * component is '..'  In this case, we uncover the mount point
     * here in the viceroy, since it will be the common case.
     */
    if (IS_AROOT(odp)) {
	sname = name;
	while (name && *name == '/') name++;

	if (name && (!strcmp("..",name) 
		     || !strncmp("../", name, strlen("../"))))
	{
	    /* We are crossing upwards from a root. */
	    name += strlen("..");
	    while (name && *name == '/') name++;
	    /* Copy the fid name */
	    *dirp = odp->o_u.ou_mountedon;
	    onode_release_read(odp);
	    /* Do the new lookup */
	    return (vfs_lookup(dirp, name, fidp, vtp, cred));
	}
    }
       
    /* this should exist actually -- pointing to lookup if cursors
     * aren't supported. */
    if (!wp->w_expand) {
	rc = EOPNOTSUPP;
	goto exit;
    }
    rc = (wp->w_expand)(wp, odp, cred, name, &remain, &op);
    /* odp's lock and reference are released by the warden */
    if (rc) {
	ds_thread_log(100, 
		      "vfs_expand of %s in (%x.%x.%x.%x) failed: (%d) %s", 
		      name, dirp->ofid_ttype, dirp->ofid_tome,
		      dirp->ofid_vnum, dirp->ofid_uniq,
		      rc, strerror(rc));
	goto exit;
    }

    /* Check to see if this is a mountpoint */
    if (IS_MOUNTEDON(op)) {
	struct viceroy_onode *rop;
	struct warden *rwp;
	struct ofid    rfid;
	
	rfid = op->o_u.ou_mountedon;
	rwp = warden_get(rfid.ofid_ttype);

	if (!rwp) { /* warden died/not there, don't try to cross mount point */
		
	    ds_thread_log(0,"%s (%x.%x.%x.%x) mounted on (%x.%x.%x.%x)",
			  "vfs_expand could not cross mount point",
			  op->o_fid.ofid_ttype, op->o_fid.ofid_tome,
			  op->o_fid.ofid_vnum,  op->o_fid.ofid_uniq,
			  rfid.ofid_ttype, rfid.ofid_tome,
			  rfid.ofid_vnum, rfid.ofid_uniq);
	} else {
	    /* We need to grab the root */
	    rc = (rwp->w_root)(rwp, rfid.ofid_tome, &rop);
	    /* Ensure that the onde returned is actually marked as a root */
#ifdef ODYDEBUG
	    ASSERT(IS_AROOT(rop));
#endif
	    /* Drop the op reference */
	    onode_release_read(op);
	    onode_put(op);
	    op = rop;
	    if (rc) goto exit;
	}
    }

    *fidp = op->o_fid;
    *vtp = op->o_vtype;
    *curs = IS_CURSOR(op);
    onode_release_read(op);
    onode_put(op);
    ds_thread_log(100, "vfs_expand of %s in (%x.%x.%x.%x) succeeded",
		  name, dirp->ofid_ttype, dirp->ofid_tome,
		  dirp->ofid_vnum, dirp->ofid_uniq);
    ds_thread_log(100, "\t\t=>(%x.%x.%x.%x)",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq);		  
 exit:
    return (rc);
}

int
vfs_cursexpand(fidp, size, retsz, buf, cred)
    struct ofid *fidp;
    size_t size;
    size_t *retsz;
    char *buf;
    struct ody_cred *cred;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_cursexpand: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;  /* Warden died/not there. */
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_curexpand: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    if (!IS_CURSOR(op) || !wp->w_cursexpand) {
	ds_thread_log(0,"vfs_cursexpand: non-cursor onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return EOPNOTSUPP;
    }
    onode_readlock(op);
    rc = (wp->w_cursexpand)(wp, op, cred, size, retsz, buf);
    onode_release_read(op);
    onode_put(op);
    if (rc) {
	ds_thread_log(100,"vfs_cursexpand: failed on (%x.%x.%x.%x) err %d", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq, rc);
    } else {
	ds_thread_log(100,"vfs_cursexpand: success on (%x.%x.%x.%x) retsz %d", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq, *retsz);
    }
    return rc;
}

/* could I combine this with vfs_open somehow, since they are identical
 * at this level? */
int
vfs_prefetch(fidp, mode, devp, inop, cred, uncacheable)
    struct ofid *fidp;
    int mode;
    dev_t *devp;
    ino_t *inop;
    struct ody_cred *cred;
    int *uncacheable;
{
    struct warden *wp;
    struct viceroy_onode *op;
    int           rc;

    wp = warden_get(fidp->ofid_ttype);
    if (!wp) {
	ds_thread_log(0,"vfs_prefetch: cannot find warden for 0x%x!", 
		      fidp->ofid_ttype);
	return ENXIO;  /* Warden died/not there. */
    }

    op = onode_get(fidp);
    if (!op) {
	ds_thread_log(1,"vfs_prefetch: cannot find onode for (%x.%x.%x.%x)!", 
		      fidp->ofid_ttype, fidp->ofid_tome,
		      fidp->ofid_vnum, fidp->ofid_uniq);
	return ENOENT;
    }
    if (!wp->w_prefetch) {
	ds_thread_log(1,"vfs_prefetch: warden has no prefetch operation!");
	return EOPNOTSUPP;
    }

    onode_writelock(op);
    rc = (wp->w_prefetch)(wp, op, cred, mode);
    if (rc) {
	goto exit;
    }

    onode_open(op, mode);
    if (IS_UNCACHEABLE(op)) {
	ASSERT(wp->w_rdwr);   /* Better be able to cope */
	*uncacheable = 1;
	*devp = 0;
	*inop = 0;
    } else {
	*uncacheable = 0;
	*devp = op->o_device;
	*inop = op->o_inode;
    }

exit:
    onode_release_write(op);
    onode_put(op);

    ds_thread_log(100,"vfs_prefetch on (%x.%x.%x.%x), flags %x returns (%d) %s",
		  fidp->ofid_ttype, fidp->ofid_tome,
		  fidp->ofid_vnum, fidp->ofid_uniq, mode, rc,
		  (rc) ? strerror(rc) : "");
    ds_thread_log(100,"\tUsing (%d, %d) as cache file",
		  *devp, *inop);
		  
    return rc;
}
#endif
