#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/* Standard includes */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>

/* System includes */
#include <sys/queue.h>
#include <sys/types.h>

#include <dirent.h>
#include <dlfcn.h>

/* Odyssey includes */
#include <kernel/vfs/includes/ody.h>
#include "include/viceroy_onode.h"
#include "include/odytypes.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <lwp/lwp.h>
#include <rpc2/rpc2.h>

#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/viceroy.h"

/* Warden definitions */
#include <base/simplewww/swww.h>
#include "odyssey/wardens/www/webwarden.h"
#include "odyssey/wardens/qt/qt_warden.h"
#include "odyssey/wardens/bit/bitwarden.h"
#include "odyssey/wardens/speech/speechwarden.h"
#include "odyssey/wardens/whiteboard/whiteboard.h"
#include "odyssey/wardens/map/mapwarden.h"

/* Local includes */
#include "root_warden.h"

/*
 * Warden table 
 */
struct wtab_ent {
    LIST_ENTRY(wtab_ent)  we_list;
    struct warden          *we_warden;
};

LIST_HEAD(wtab_list, wtab_ent);

static size_t             wtabsz = 13;
static struct wtab_list   *wtab;

#ifndef MAXWARDENS
#define MAXWARDENS 10
#endif

struct warden *wardens[MAXWARDENS] = {
    &root_warden,
};

static size_t  num_wdns = 1;

extern char *WardenDir;

static
void
warden_load(void)
{
    static int wardens_loaded = 0;
    DIR *dir;
    struct dirent *dent;

    if (wardens_loaded) return;

    /* Look for all the warden files */
    dir = opendir(WardenDir);

    if (!dir) {
	fprintf(stderr,"warden_load: could not open %s: %s\n", WardenDir,
		strerror(errno));
	exit(1);
    }

    while ((dent = readdir(dir)) != NULL) {
	char fname[256];
	char wname[32];
	int len = strlen(dent->d_name);
	struct warden *wdn;
	void *dl_handle;

	/* check if the file is lib<name>warden.so */
	if (strncmp(dent->d_name, "lib", 3)) continue;
	if (strncmp(dent->d_name+len-9,"warden.so", 9))
	    continue;
	sprintf(fname,"%s/%s", WardenDir, dent->d_name);
#ifdef __BSD44__
	strcpy(wname,"_");
	strncat(wname,dent->d_name + 3, len-12);
	len++;
#else
	strcpy(wname,"");
	strncat(wname,dent->d_name + 3, len-12);
#endif
	wname[len-12] = '\0';
	strcat(wname,"_warden");

	/* Fill in the next available warden table slot */
	dl_handle = dlopen(fname, RTLD_NOW);
	if (!dl_handle) {
	    fprintf(stderr, "warden_load: dlopen(%s): %s\n", fname, dlerror());
	    exit(1);
	}
	wdn = (struct warden *) dlsym(dl_handle, wname);
	if (!wdn) {
	    fprintf(stderr,"warden_load: dlsym(%s,%s): %s\n", fname, wname,
		    dlerror());
	    exit(1);
	}
	if (num_wdns < sizeof(wardens)/sizeof(wardens[0])) {
	    wardens[num_wdns++] = wdn;
	    fprintf(stderr,"warden_load: loaded %s (%s)\n", wname,
		    wdn->w_name);
	}
	else 
	    fprintf(stderr,"warden_load: could not load %s (%s): "
		    "warden table full\n", wname, wdn->w_name);

    }
    closedir(dir);
    wardens_loaded = 1;
}


int
warden_parseargs(cur, argc, argv)
    int cur;
    int argc;
    char *argv[];
{
    int result;
    int i;

    warden_load();

    for (i=0; i<num_wdns; i++) {
	if ((result = (wardens[i]->w_parseargs)(cur,argc,argv)) != 0)
	    return result;
    }
    return 0;
}

void
warden_usage(FILE *fp) {
    int i;
    
    for (i=0; i<num_wdns; i++) {
	fprintf(stderr,"\n\tArguments accepted by warden %s:\n", 
		wardens[i]->w_name);
	fprintf(stderr,"%s\n", wardens[i]->w_usage);

    }
}

void
warden_stats(FILE *fp)
{
    int i;

    for (i = 0; i < num_wdns; i++) {
	fprintf(fp, "Stats for warden: %s\n", wardens[i]->w_name);
	if (wardens[i]->w_pstats)
	    (wardens[i]->w_pstats)(wardens[i], fp);
    }
}

    
void
warden_init()
{
    int i;
    int ok;
    int bucket;
    struct wtab_ent *wep;

    warden_load();

    wtab = malloc(sizeof(struct wtab_list) * wtabsz);

    for (i=0; i<wtabsz; i++) {
	LIST_INIT(&wtab[i]);
    }
    
    for (i=0; i<num_wdns; i++) {
	ok = (wardens[i]->w_init)(wardens[i]);
	if (!ok) {
	    bucket = wardens[i]->w_typenum % wtabsz;
	    wep = malloc(sizeof(struct wtab_ent));
	    wep->we_warden = wardens[i];
	    LIST_INSERT_HEAD(&wtab[bucket], wep, we_list);
	} else {
	    ds_thread_log(0, "Warden %s initialization failed (%s)",
			  wardens[i]->w_name, strerror(errno));
	}
    }

}

struct warden *
warden_get(ttype)
    unsigned int ttype;
{
    int bucket = ttype % wtabsz;
    struct wtab_ent *wep;

    for (wep = wtab[bucket].lh_first; wep; wep = wep->we_list.le_next) {
	if (wep->we_warden->w_typenum == ttype)
	    return wep->we_warden;
    }
    return NULL;
}

/*
 * Default implementations:
 */

int
wdflt_parseargs(cur, argc, argv)
    int cur;
    int argc;
    char *argv[];
{
    return 0;
}

int
wdflt_init(wp)
    struct warden *wp;
{
    wp->w_data = NULL;
    return 0;
}
