#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * bitwarden_main.c
 *    (from webwarden_main.c, by et)
 *
 * Kip Walker, Jan 1997
 *
 * main routines for the bitwarden, including VFScalls
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>

#include <sys/uio.h>

#include <kernel/vfs/includes/ody.h>
#include "include/viceroy_onode.h"
#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"

#include "include/warden.h"
#include "include/viceroy.h"
#include "include/onode-manager.h"

#include "bitwarden.h"
#include "bitwarden_main.h"
#include "bitwarden_private.h"
#include "bitwarden_onode.h"
#include "bitwarden_rpc2.h"
#include "bitwarden_worker.h"

int bitw_loglevel = 0;

struct warden bit_warden = {
    BIT_TTYPE,
    "Bitstream",
    0,
    bitw_parse_args,
    "Usage: -bitw_server <tome #> <server name> [-bitw_debug]\n"
    "       [-bitw_workers <n>] [-bitw_loglevel <level>]\n",
    bitw_init,
    bitw_getcid,
    bitw_pstats,
    bitw_root,
    bitw_lookup,
    bitw_access,
    bitw_getattr,
    bitw_open,
    bitw_close,
    bitw_rdwr,
    bitw_rdwr_dispose,
    (w_readdir_fn)NULL,
    (w_readdir_dispose_fn) NULL,
    bitw_tsop,
    bitw_tsop_dispose,
#ifdef SETS
    bitw_expand,
    bitw_cursexpand,
    (w_prefetch_fn)NULL,
#endif
    0,
    0,
    (void *)NULL
};

static unsigned long servers[BITW_MAXTOME];
int bitw_nworkers = BITW_DEFAULT_WORKERS;
extern int ds_event_debug;
//extern char lwp_debug;
extern struct timeval cont_sw_threshold;

#define INET_ADDR_LEN  64
static char     inet_address[INET_ADDR_LEN];

static char *inatoa(unsigned long addr) 
{
    unsigned long haddr = ntohl(addr);

    sprintf(inet_address,"%lx.%lx.%lx.%lx",
	    (haddr & 0xff000000) >> 24,
	    (haddr & 0xff0000) >> 16,
	    (haddr & 0xff00) >> 8,
	    (haddr & 0xff));
    return inet_address;
}    

/*
 * try to consume viceroy command line switches.
 * Return number of args consumed, -1 if error.
 */
int bitw_parse_args(int cur, int argc, char *argv[])
{
    struct hostent *hp;
    int tome;

    if (strcmp(argv[cur], "-bitw_server") == 0) {
        cur++;
        if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
            return -1; /* no argument following flag */

        tome = atoi(argv[cur++]);
        if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
            return -1; /* no argument following tome # */

        if ((tome < 0) || (tome >= BITW_MAXTOME))
            return -1;

	if (!(hp = gethostbyname(argv[cur]))) {
	    ds_thread_log(0,"bitwarden: can't gethostbyname %s",
			  argv[cur]);
	    return 3;
	}
	if (hp->h_length != sizeof(unsigned long)) {
	    ds_thread_log(0,"bitwarden: host %s has address length %d!",
			  argv[cur], hp->h_length);
	    return 3;
	}
	bcopy(hp->h_addr, &servers[tome], sizeof(unsigned long));
	return 3;		/* ate two arguments */
    } else if (strcmp(argv[cur], "-bitw_debug") == 0) {
	ds_event_debug = 1;
	/* lwp_debug = 1; */    /* generates a LOT of printfs */
	return 1;		/* ate one argument */
    } else if (strcmp(argv[cur], "-bitw_workers") == 0) {
	if ((cur+1) >= argc)
	    return -1;
	if (*(argv[cur+1]) == '-')
	    return -1;
	sscanf(argv[cur+1], "%d", &bitw_nworkers);
	return 2;
    }
    else if (strcmp(argv[cur], "-bitw_loglevel") == 0) {
	if ((cur+1) >= argc)
	    return -1;
	if (*(argv[cur+1]) == '-')
	    return -1;
	sscanf(argv[cur+1], "%d", &bitw_loglevel);
	return 2;
    }
    return 0;
}

/*
 * setup warden data, check that we were given the name of a server.
 * Return 0 if successful.
 */
int bitw_init(struct warden *wp)
{
    struct bitw_data       *my_data;
    struct ofid            root_ofid;
    struct onode_create_in root_onode_params;
    struct viceroy_onode           *rop;
    int                    tome;
    int                    num_tomes = 0;
    int                    i;
#ifdef DO_PREFETCH
    struct ds_safeq_t      *workq = NULL;
#endif

    for (tome = 0; tome < BITW_MAXTOME; tome++) {
	if (servers[tome] != (unsigned long) NULL)
	    num_tomes++;
    }
    if (num_tomes == 0) {
	ds_thread_log(0, "bitw_init: no servers specified");
	return -1;
    }

    /* set up warden specific data. first we allocate the space
       in w_data, then we assign that same memory location to
       a pointer of the proper type. We then use the properly
       cast pointer in order to not have to cast the data every
       time we reference it. */
    ALLOC(wp->w_data, struct bitw_data);
    my_data = (struct bitw_data *)(wp->w_data);

    for (tome=0; tome<BITW_MAXTOME; tome++) {
	struct bitw_tome *bwt;

	my_data->bwd_tomes[tome] = NULL;
	if (servers[tome] == 0)	/* no servername given */
	    continue;

	ALLOC(bwt, struct bitw_tome);

	/* set up the root onode in init, since there is only one volume */
	root_ofid.ofid_ttype = BIT_TTYPE;
	root_ofid.ofid_tome  = tome;
	root_ofid.ofid_vnum  = 0;
	root_ofid.ofid_uniq  = 0;

	root_onode_params.oc_flags = 0;
	root_onode_params.oc_vtype = O_VDIR;
	root_onode_params.oc_ofree = NULL;
	root_onode_params.oc_data  = NULL;
	
	rop = onode_create(&root_ofid, &root_onode_params);
	if (rop == NULL) {
	    ds_thread_log(0, "bitw_init: could not create volume onode root!");
	    ASSERT(0);
	}

	onode_writelock(rop);
	rop->o_flags |= ON_ROOTOFVOL | ON_UNCACHEABLE;
	rop->o_cachefile = NULL;
	onode_release_write(rop);

	bwt->bwt_root_onode = rop;
	my_data->bwd_tomes[tome] = bwt;
    }

    /* set up remainder of bitwarden's w_data */
    my_data->bwd_current_uniq    = 1;/* set up current uniqifier */
    /* XXX need to set default quality parms */
    ALLOC(my_data->bwd_monitor, ds_hwlock_t);
    ds_hwlock_init(my_data->bwd_monitor);

    my_data->onode_list = NULL;

    gettimeofday(&my_data->last_pstats, NULL);


#ifdef DO_PREFETCH
	/* if we want to prefetch, start some worker threads to do it
	 * in the background and a safeq to give them messages. */
    workq = ds_safeq_create();
    ASSERT(workq);
    my_data->bwd_workq = workq;

    for (i=0; i<bitw_nworkers; i++) {
	struct bitw_worker_params *bwwp;

	ALLOC(bwwp, struct bitw_worker_params);
	bwwp->id = i;
	bwwp->workq = workq;

	ASSERT(ds_thread_create(bitw_worker_main, (char *)bwwp,
				DS_THREAD_DAEMON,
				"bit_prefetcher",
				bitw_loglevel));
    }
#endif

    ds_thread_log(0, "bitw_init: initialized successfully");
    return(0);			/* successful initialization */
}

/*
 * return the root_onde, readlocked
 * Return 0 if success, errno value otherwise
 */
int bitw_root(struct warden *wp,
	      tome_t        tome,
	      struct viceroy_onode  **root_onodep)
{
    struct viceroy_onode *root_onode;
    struct ofid  of;
    
    
    if (tome < 0 || tome >= BITW_MAXTOME) {
	ds_thread_log(100,"bitw_root: requested tome (%d) not supported",
		      tome);
	return EINVAL;
    }

    of.ofid_ttype = BIT_TTYPE;
    of.ofid_tome  = tome;
    of.ofid_vnum  = 0;		/* for bitwarden, {X,X,0,0} indicates */
    of.ofid_uniq  = 0;		/* an ofid for a volume. */

    root_onode = onode_get(&of);
    if (root_onode == NULL) {
	ds_thread_log(0, "bitw_root: onode_get() failed for warden tome "
		      "%d's root onode.", tome);
	return EINVAL;
    }

    onode_readlock(root_onode);

    *root_onodep = root_onode;

    /* Life is happy; we've gotten the onode of the root, and we've
       readlocked it, so return a happy return code. */
    return 0;
}

/*
 * create an onode for a pathname in our tome.  Since there isn't any
 * notion of a namespace for the bitwarden, accept anything given
 * to us.  Return a read_lock'ed onode.  Return code is 0 for success,
 * errno val for error.
 */
int bitw_lookup(struct warden *wp,
		struct viceroy_onode  *odp,
		struct ody_cred  *cred,
		char          *name,
		char          **remain,
		struct viceroy_onode  **opp)
{
    struct bitw_data       *my_data = (struct bitw_data *) wp->w_data;
    struct ofid            new_ofid;
    struct onode_create_in new_onode_params;
    struct viceroy_onode           *new_onode;
    int error = 0;

    /* deal with ".." */
    if (strncmp(name, "..", 2) == 0) {/* XXX this is broken! */
	*remain = name + 2;
	error = EMOUNTEDON;
        goto done;
    }

    *remain = NULL;		/* don't care what the name is */

    ds_hwlock_write(my_data->bwd_monitor);
    new_ofid.ofid_uniq = (my_data->bwd_current_uniq += 1);
    ds_hwlock_release_write(my_data->bwd_monitor);
    
    /* use the uniquifier for ofid.vnum, since it doesn't matter */
    new_ofid.ofid_vnum  = new_ofid.ofid_uniq;
    new_ofid.ofid_ttype = BIT_TTYPE;
    new_ofid.ofid_tome  = odp->o_fid.ofid_tome;

    /* all names correspond to uncacheable, regular files.  This way,
     * all reads and writes come to us for SFTP from/to the server */
    new_onode_params.oc_flags = ON_UNCACHEABLE;
    new_onode_params.oc_vtype = O_VREG;
    new_onode_params.oc_ofree = bitw_onode_data_free;
    ALLOC(new_onode_params.oc_data, struct bitw_onode_data);
    bzero(new_onode_params.oc_data, sizeof(struct bitw_onode_data));

    if (new_onode_params.oc_data == NULL) {
	ds_thread_log(0, "bitw_lookup: could not allocate bitw_onode_data");
	ASSERT(0);
    }

    new_onode = onode_create(&new_ofid, &new_onode_params);
    if (new_onode == NULL) {
	ds_thread_log(0, "bitw_lookup: could not create onode for %s",
		      name);
	ASSERT(0);
    }

    if (onode_get(&new_ofid) == new_onode) {
	*opp = new_onode;
    } else {
	ds_thread_log(0, "bitw_lookup: onode_get returned unexpected "
		      "onode.");
	ASSERT(0);
    }

    /* Must also be returned readlocked */
    onode_readlock(new_onode);

done:
    onode_release_read(odp);
    onode_put(odp);

    return error;
}

/*
 * accept any access permissions, since there is no ownership, and
 * writes (== put to server) are allowed.  
 * XXX should I disallow VEXEC?
 */
int bitw_access(struct warden *wp,
		struct viceroy_onode  *op,
		struct ody_cred  *cred,
		int           mode)
{
    return 0;
}

int bitw_getattr(struct warden *wp, struct viceroy_onode *op,
		 struct ody_vattr *vap, struct ody_cred *cred)
{
    time_t temp = 0;

    vap->va_type = O_VREG;
    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;
    vap->va_blocksize = 512;
    vap->va_atime = temp;
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    return 0;
}

/* Sample use of internal requests */

/* Request argument */
struct cbarg_t {
    struct warden *wp;
    struct viceroy_onode  *op;
    cmgr_id_t     cid;
};

#ifdef BITW_RR
/* Free a request argument.  We have to do a put on the onode */
static void
bitw_reqarg_free(argp)
    void *argp;
{
    struct cbarg_t *targp = (struct cbarg_t *)argp;
    
    onode_put(targp->op);
    FREE(targp);
}

/* Take a callback */
static void
bitw_callback(reqid, rsrc, val, argp)
    int reqid;
    unsigned int rsrc;
    int val;
    void *argp;
{
    struct cbarg_t *targp = (struct cbarg_t *)argp;
    int result;
    int value = val;
    int low;
    int high;
    int cancelled;
    int newreq;
    ds_reqarg_t *reqargp;
    struct bitw_onode_data *wod;

    ds_thread_log(0,"bitw_callback: cid %d, reqid %d, value %d",
		  targp->cid, reqid, value);

    /* Need a new reqargp */
    ALLOC(reqargp, ds_reqarg_t);
    reqargp->free_fn = bitw_reqarg_free;
    /* 
     * Rather than freeing the argp (which we are responsible for
     * doing) we reuse it in the next request 
     */
    reqargp->argval = argp;

    do {
	low = (value/10)*9;
	high = (value/10)*11;

	result = vr_rsrc_request(targp->wp, targp->op, ODY_GR_NETBW, 
				 ODY_GENERIC_INTERFACE, low, high,
				 bitw_callback, reqargp, &cancelled,
				 &value, &newreq);
	if (cancelled) {
	    ds_thread_log(0,"bitw_callback: old request %d CANCELLED",
			  cancelled);
	}
	if (result == ODY_ENOTINWINDOW) {
	    ds_thread_log(0,"bitw_callback: request OUTOFWINDOW (%d)",
			  value);
	    ASSERT(!newreq);
	} else {
	    ds_thread_log(0,"bitw_callback: request %d placed",
			  newreq);

	    /* Remember which request id is associated with this onode */
	    onode_writelock(targp->op);
	    wod = (struct bitw_onode_data *)targp->op->o_data;
	    wod->cur_reqid = newreq;
	    onode_release_write(targp->op);
	}
    } while (!newreq);
}
#endif

/*
 * given an onode, connect to the server and allocate the per-onode
 * data structure (includes prefetch buffer and such)
 */
int bitw_open(struct warden *wp,
	      struct viceroy_onode  *op,
	      struct ody_cred  *cred,
	      int           mode)
{
    cmgr_id_t         cid;
    struct bitw_data  *my_data = (struct bitw_data *) wp->w_data;
    struct bitw_onode_data *wod = (struct bitw_onode_data *)op->o_data;
    bitw_buffer_t     *buf = &wod->buf;
    struct in_addr    addr;
#ifdef DO_PREFETCH_ON_OPEN
    struct bitw_worker_msg *msg;
#endif
#ifdef BITW_RR
    ds_reqarg_t       *reqargp;
    struct cbarg_t    *argvalp;
    int               bwval;
    int               low,high;
    int               cancelled,reqid;
    int               result;
#endif
    
    if (op->o_fid.ofid_vnum == 0 || op->o_fid.ofid_uniq == 0) {
	/* this is the root.  Can't open it */
	return EISDIR;
    }

    if (my_data->bwd_tomes[op->o_fid.ofid_tome] == NULL) {
	/* this is a tome without a server! */
	return EINVAL;
    }

    /* someone used the name of a currently open bitstream. */
    if (wod->buf.buf != NULL) {
	return EBUSY;
    }
    
    /* Set up the RPC2 connection to ody web server. */
    addr.s_addr = servers[op->o_fid.ofid_tome];	
    cid = connect_to_bitserver(addr);
    if(cid < 0) {
	/* XXX want cm to turn errnum into string */
	ds_thread_log(0, "bitw_init: couldn't establish connection "
		      "to server %s (cm returned %d).",
		      inatoa(servers[op->o_fid.ofid_tome]), cid);
	return -1;
    }
    
    wod->cid = cid;
    wod->bytes_read = 0;
    wod->rpc_msec = 0.0;
    wod->chunk_count = 0;
    
#ifdef BITW_RR
    /* Set up a callback so that we can test internal-requests */

    /* Create the argument structure */
    ALLOC(reqargp, ds_reqarg_t);
    reqargp->free_fn = bitw_reqarg_free;

    /* Create the argument value buffer */
    ALLOC(argvalp, struct cbarg_t);
    argvalp->wp = wp;
    /* Need a new get for the onode reference stuffed in cbarg_t */
    argvalp->op = onode_get(&op->o_fid);
    argvalp->cid = cid;
    reqargp->argval = argvalp;

    /* Place the request, looping until we get one set up */
    bwval = vr_rsrc_value(op, ODY_GR_NETBW, ODY_GENERIC_INTERFACE);
    do {
	low = (bwval/10)*9;
	high = (bwval/10)*11;
	result = vr_rsrc_request(wp, op, ODY_GR_NETBW, ODY_GENERIC_INTERFACE,
				 low, high, bitw_callback, reqargp, 
				 &cancelled, &bwval, &reqid);
	if (cancelled) {
	    ds_thread_log(0,"bitw_open: old request %d CANCELLED", cancelled);
	}
	if (result == ODY_ENOTINWINDOW) {
	    ds_thread_log(0,"bitw_open: request OUTOFWINDOW (%d)", bwval);
	    ASSERT(!reqid);
	} else {

	    ds_thread_log(0,"bitw_open: request %d placed", reqid);
	    /* Remember which request id is associated with this onode */
	    onode_writelock(op);
	    wod = (struct bitw_onode_data *)op->o_data;
	    wod->cur_reqid = reqid;
	    onode_release_write(op);
	}
    } while (!reqid);
#endif

    /* allocate buffer structures */
    NALLOC(buf->buf, char, BITW_DEFAULT_DEPTH);
    assert(buf->buf);
    buf->consumer = buf->buf;
    buf->producer = buf->buf;
    buf->producer_left = BITW_DEFAULT_DEPTH;
    buf->consumer_left = BITW_DEFAULT_DEPTH;
    buf->prefetched = 0;
    buf->prefetching = 0;

    ds_hwlock_init(&wod->buf_lock);

    wod->params.prefetch_depth = BITW_DEFAULT_DEPTH;
    wod->params.chunk_size = BITW_DEFAULT_CHUNK;

    wod->prev = NULL;
    wod->next = my_data->onode_list;
    my_data->onode_list = wod;

    /* if prefetching, start a thread now to initially fill the 
     * buffer */
#ifdef DO_PREFETCH_ON_OPEN
    ALLOC(msg, struct bitw_worker_msg);
    msg->type = PREFETCH;
    msg->ofid_uniq = op->o_fid.ofid_uniq;
    msg->d.prefetch.wod = wod;
    ds_safeq_enq(my_data->bwd_workq, msg);
#endif

    /* no locally cached file */
    op->o_device = 0;
    op->o_inode  = 0;

    return 0;
}

/*
 * clean up an onode when the file is closed 
 */
int bitw_close(struct warden *wp,
	       struct viceroy_onode  *op,
	       struct ody_cred  *cred,
	       int           mode)
{
    struct bitw_data  *my_data = (struct bitw_data *) wp->w_data;
    struct bitw_onode_data *wod = (struct bitw_onode_data *)op->o_data;
    cmgr_id_t cid = wod->cid;
    int result;

#ifdef BITW_RR
    /* Cancel any oustanding request */
    onode_writelock(op);
    vr_rsrc_cancel(wod->cur_reqid);
    onode_release_write(op);
#endif

    /* allow prefetch to finish, so the buffer isn't deallocated while
     * writes/reads may be pending */
    while (wod->buf.prefetching)
	ds_thread_wait((char *)&wod->buf_lock);

    /* close connection and free prefetch buffer, keep the onode */
    if (cmgr_validate(cid) > 0) {
	result = cmgr_destroy(cid);
	if (result != CMGR_SUCCESS) {
	    ds_thread_log(0, "bitw_close: cm_destroy(%d) returned %d",
			  cid, result);
	    return ENOTCONN;
	}
    }

    if (wod->chunk_count != 0)
      ds_thread_log(1, "msec/chunk: %f", wod->rpc_msec/wod->chunk_count);

    /* update onode_list */
    if (!wod->prev && !wod->next)
	my_data->onode_list = NULL;
    else if (!wod->prev) {
	my_data->onode_list = wod->next;
	wod->next->prev = NULL;
    } else if (!wod->next) {
	wod->prev->next = NULL;
    } else {
	wod->prev->next = wod->next;
	wod->next->prev = wod->prev;
    }

    /* deallocate the buffer, set to NULL -- now the if the onode is
     * picked out of the namecache, bitw_open knows that it's not
     * actually being used */
    FREE(wod->buf.buf);
    wod->buf.buf = NULL;

    return 0;
}

/*
 *
 */
int bitw_rdwr (struct warden *wp,
	       struct viceroy_onode  *op,
	       struct ody_cred  *cred,
	       int           rwflag,
	       int           *count,
	       int           offset,
	       int           ioflag,
	       void          **data)
{
    long            len;
    struct bitw_onode_data *wod = (struct bitw_onode_data *)op->o_data;
#ifdef DO_PREFETCH
    struct bitw_data  *my_data = (struct bitw_data *) wp->w_data;
#else
    cmgr_id_t         cid = wod->cid;
    cmgr_connection_t *connp;
    float             el_msec;
    char              *buf;
    long              left;
#endif

#ifndef DO_PREFETCH
    if (cmgr_validate(cid) > 0) {
	connp = cmgr_get(cid);
	ASSERT(connp);
    } else {
        len = 0;
        goto exit;
    }
#endif

    if (rwflag == O_RW_WRITE) {
	len = 0;		/* nothing yet... */
    } else {
#ifdef DO_PREFETCH
	/* XXX fast and loose -- skip lock for this read (probably OK) */
	while (wod->buf.prefetched < *count) {
#ifndef DO_PREFETCH_ON_OPEN
	    if (!wod->buf.prefetching) {
		struct bitw_worker_msg *msg;
		ALLOC(msg, struct bitw_worker_msg);
		msg->type = PREFETCH;
		msg->ofid_uniq = op->o_fid.ofid_uniq;
		msg->d.prefetch.wod = wod;
		ds_safeq_enq(my_data->bwd_workq, msg);		
	    }
#endif
	    /* sleep this fella -- no data yet.  bitw_prefetch will signal */
	    ds_thread_log(100, "read sleeping");
	    ds_thread_wait((char *)&wod->buf_lock);
	}

	ds_hwlock_write(&wod->buf_lock);
	/* avoid overrunning the end of the buffer */
	len = (wod->buf.consumer_left < *count) ? 
	    wod->buf.consumer_left : *count;
	bcopy(wod->buf.consumer, *data, len);
	wod->bytes_read += len;

	/* update bookkeeping */
	if (len == wod->buf.consumer_left) {
	    /* need to wrap around */
	    wod->buf.consumer = wod->buf.buf;
	    wod->buf.consumer_left = wod->params.prefetch_depth;
	} else {
	    wod->buf.consumer += len;
	    wod->buf.consumer_left -= len;
	}
	wod->buf.prefetched -= len;
	ds_hwlock_release_write(&wod->buf_lock);

#ifdef DO_PREFETCH_REFILL
	/* initiate further prefetching if half of prefetch is empty.
	 * XXX better heuristic here? */
	if (!wod->buf.prefetching &&
	    wod->buf.prefetched << 1 <= wod->params.prefetch_depth) {
	    struct bitw_worker_msg *msg;
      
	    ds_thread_log(100, "starting a new prefetcher");
	    ALLOC(msg, struct bitw_worker_msg);
	    msg->type = PREFETCH;
	    msg->ofid_uniq = op->o_fid.ofid_uniq;
	    msg->d.prefetch.wod = wod;
	    ds_safeq_enq(my_data->bwd_workq, msg);
	}
	ds_thread_log(100, "finishing read");
#endif
#else
	/* just do the RPC if we aren't prefetching */
	left = *count;
	buf = *data;
	do {
	    len = left > wod->params.chunk_size ? wod->params.chunk_size :
		left;
	    len = bitw_read(buf, len, connp, &el_msec);
	    if (len > 0) {
		wod->rpc_msec += el_msec;
		wod->chunk_count++;
		left -= len;
		buf += len;
	    }
	} while (left > 0 && len > 0);
	cmgr_put(cid);
	if (len <0) {
	    cmgr_markdown(cid);
	    len = 0;
	} else
	    len = *count;
#endif
    }
  
#ifndef DO_PREFETCH
exit:
#endif
    *count = len;
    return 0;
}

int bitw_rdwr_dispose (struct warden *wp,
		       struct viceroy_onode  *op,
		       struct ody_cred  *cred,
		       int           count,
		       int           offset,
		       void          *data)
{
    /* nothing to dispose! */
    return 0;
}

/*
 * TSOPS
 *
 */

int bitw_tsop(struct warden *wp,
	      int           tsvnum,
	      struct viceroy_onode  *op,
	      struct ody_cred  *cred,
	      int           opnum,
	      size_t        argsz,
	      void          *argp,
	      size_t        *retsz,
	      void          **retp)
{
    long result;

    /* check the version number */
    if(tsvnum != BITWARDEN_VERSION) {
	return(ODY_EBADVERSION);
    }
  
    /* dispatch TSOPs, based on opnum */
    switch(opnum) {
    case BITW_SET_PARAMS:
	result = bitw_set_params(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    default:
	break;
    }
  
    return(EOPNOTSUPP);
}

int bitw_tsop_dispose(struct warden *wp,
		      struct viceroy_onode  *op,
		      struct ody_cred  *cred,
		      int           opnum,
		      size_t        argsz,
		      void          *argp,
		      size_t        retsz,
		      void          *retp)
{
    /* dispose of *retp, since we were responsible for allocating it */
    FREE(retp);			/* FREE() checks if retp is NULL */
    return(0);
}

cmgr_id_t bitw_getcid(struct warden *wp, struct viceroy_onode  *op)
{
    struct bitw_onode_data *wod = (struct bitw_onode_data *)op->o_data;

    return wod->cid;
}

void bitw_pstats(struct warden *wp, FILE *fp)
{
    struct bitw_data       *my_data = (struct bitw_data *) wp->w_data;
    struct bitw_onode_data *wod;
    struct timeval t;
    float msec_since;
    
    if (1 || !my_data)
	return;
    
    wod = my_data->onode_list;
    gettimeofday(&t, NULL);
    msec_since = ms(t) - ms(my_data->last_pstats);
    my_data->last_pstats = t;
    while (wod) {
	fprintf(fp, "%5ld %16.3f %10ld\n", wod->cid, msec_since, 
		wod->bytes_read);
	wod = wod->next;
    }
}

#ifdef SETS
int bitw_expand(struct warden *wp,
		struct viceroy_onode  *odp,
		struct ody_cred  *cred,
		char          *name,
		char          **remain,
		struct viceroy_onode  **opp)
{
    struct bitw_data       *my_data = (struct bitw_data *) wp->w_data;
    struct ofid            new_ofid;
    struct onode_create_in new_onode_params;
    struct viceroy_onode           *new_onode;
    int error = 0;

    /* deal with ".." */
    if (strncmp(name, "..", 2) == 0) {/* XXX this is broken! */
	*remain = name + 2;
	error = EMOUNTEDON;
        goto done;
    }

    if (strncmp(name, "cursor", 6) != 0) {
	*remain = name;
	error = ENOENT;
        goto done;
    }

    ds_hwlock_write(my_data->bwd_monitor);
    new_ofid.ofid_uniq = (my_data->bwd_current_uniq += 1);
    ds_hwlock_release_write(my_data->bwd_monitor);
    
    /* use the uniquifier for ofid.vnum, since it doesn't matter */
    new_ofid.ofid_vnum  = new_ofid.ofid_uniq;
    new_ofid.ofid_ttype = BIT_TTYPE;
    new_ofid.ofid_tome  = 1;

    /* this guy is a cursor!  and of course, not cacheable. */
    new_onode_params.oc_flags = ON_UNCACHEABLE | ON_CURSOR;
    new_onode_params.oc_vtype = O_VREG;
    new_onode_params.oc_ofree = bitw_onode_data_free;
    ALLOC(new_onode_params.oc_data, struct bitw_onode_data);
    bzero(new_onode_params.oc_data, sizeof(struct bitw_onode_data));

    if (new_onode_params.oc_data == NULL) {
	ds_thread_log(0, "bitw_lookup: could not allocate bitw_onode_data");
	ASSERT(0);
    }

    new_onode = onode_create(&new_ofid, &new_onode_params);
    if (new_onode == NULL) {
	ds_thread_log(0, "bitw_lookup: could not create onode for %s",
		      name);
	ASSERT(0);
    }

    if (onode_get(&new_ofid) == new_onode) {
	*opp = new_onode;
    } else {
	ds_thread_log(0, "bitw_lookup: onode_get returned unexpected "
		      "onode.");
	ASSERT(0);
    }

    /* Must also be returned readlocked */
    onode_readlock(new_onode);

done:
    onode_release_read(odp);
    onode_put(odp);

    return error;
}

int before = 0;

static int
bitw_cursexpand(struct warden *wp,
		struct viceroy_onode  *op,
		struct ody_cred  *cred,
		size_t        size,
		size_t        *retsz,
		char          *buf)
{
	struct {
		long next;
		char text[40];
	} jack[6] = {{44,"/ody/www/http:\\\\sports.yahoo.com\\nhl"},
		     {44,"/ody/www/http:\\\\www.phish.com"},
		     {44,"/ody/www/http:\\\\www.nhl.com"},
		     {44,"/ody/www/http:\\\\www.cs.cmu.edu"},
		     {44,"/ody/www/http:\\\\www.olestra.com"},
		     {0,""}};
	if (before == 0) {
		before = 1;
		bcopy(jack, buf, 220);
		*retsz = 220;
		return 0;
	} else {
		*retsz = 0;
		before = 0;	/* just for the demo --
				 * this assumes you don't try
				 * to cursexpand again.  *I won't* */
		return 0xe0f;
	}
}
#endif
