#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * bitwarden_rpc2.c -- stolen with permission from ...
 *   webwarden_rpc2.c -- RPC2 support code
 *
 * written by Eric Tilton <tilt+@cs.cmu.edu>
 *  adapted by Kip Walker <kwalker@cs.cmu.edu>
 *
 * This file provides support routines for RPC2.
 *
 */

/* system/C headers */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>

#include <kernel/vfs/includes/ody.h>
#include "include/viceroy_onode.h"

/* RPC2 headers */
#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

/* my headers */
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"

#include "ody_bits.h"
#include "bitwarden.h"
#include "bitwarden_onode.h"
#include "bitwarden_rpc2.h"
#include "bitwarden_worker.h"

cmgr_id_t connect_to_bitserver(srvraddr) 
    struct in_addr srvraddr;
{
    cmgr_initparams_t ci;
    cmgr_id_t cid;
    
    bzero(&ci, sizeof(cmgr_initparams_t));

    ci.ci_type    = CMGR_RPC2;

    ci.ci_restart.tv_sec = 60;	/* 1 minute between restarts */
    ci.ci_restart.tv_usec = 0;

    ci.ci_params.ci_rpc2.hid.Tag = RPC2_HOSTBYINETADDR;
    ci.ci_params.ci_rpc2.hid.Value.InetAddress = srvraddr;

    ci.ci_params.ci_rpc2.pid.Tag = RPC2_PORTBYINETNUMBER;
    ci.ci_params.ci_rpc2.pid.Value.InetPortNumber = htons(BITS_PORTAL);
    
    ci.ci_params.ci_rpc2.sid.Tag = RPC2_SUBSYSBYID;
    ci.ci_params.ci_rpc2.sid.Value.SubsysId = BITS_SUBSYSID;
    
    /* Bind Parameters */
    ci.ci_params.ci_rpc2.bp.SecurityLevel = RPC2_OPENKIMONO;
    ci.ci_params.ci_rpc2.bp.EncryptionType = 0;
    ci.ci_params.ci_rpc2.bp.SideEffectType = SMARTFTP;	/* allow use of SFTP */
    ci.ci_params.ci_rpc2.bp.ClientIdent = NULL;
    ci.ci_params.ci_rpc2.bp.SharedSecret = NULL;
    
    /* Create new binding */
    cid = cmgr_create(&ci);
    return cid;
}

/*
 * read a chunk from the server
 * assumes length bounded by chunk_size before called
 */
int bitw_read(char *buf, long length, cmgr_connection_t *connp, float *el_msec)
{
    SE_Descriptor sed;
    long result;
    struct timeval t1, t2;

    bzero(&sed, sizeof(sed));
    sed.Tag = SMARTFTP;
    sed.Value.SmartFTPD.Tag = FILEINVM;
    sed.Value.SmartFTPD.ByteQuota = -1;
    sed.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
    sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buf;
    sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.MaxSeqLen = length;
    gettimeofday(&t1, NULL);
    result = bits_get(CMGR_RPC2_HANDLE(connp), (RPC2_Integer *)&length, &sed);
    gettimeofday(&t2, NULL);
    ds_thread_log(2, "---->%f", ms(t2)-ms(t1));
    if (el_msec)
	*el_msec = ms(t2)-ms(t1);
    if (result == RPC2_SUCCESS &&
	sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen == length) {
	return length;
    } else
	return 0;
}

/*
 * write a chunk to the server
 * assumes length bounded by chunk_size before called
 */
int bitw_write(char *buf, long length, cmgr_connection_t *connp, 
	       float *el_msec)
{
    SE_Descriptor sed;
    long result;
    
    /* this should use the chunk size, looping until sent */
    bzero(&sed, sizeof(sed));
    sed.Tag = SMARTFTP;
    sed.Value.SmartFTPD.Tag = FILEINVM;
    sed.Value.SmartFTPD.ByteQuota = -1;
    sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;
    sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buf;
    sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = length;
    result = bits_put(CMGR_RPC2_HANDLE(connp), (RPC2_Integer)length, &sed);
    /* add timing bits */
    if (result == RPC2_SUCCESS)
	return sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen;
    else
	return 0;
}

/*
 * fill the prefetch buffer
 */
int bitw_prefetch(struct bitw_onode_data *wod)
{
    cmgr_connection_t *connp;
    int               result;
    int               to_prefetch, this_chunk;
    long              rc;
    float             el_msec;
    
    ds_thread_log(100, "called: bitw_prefetch");
    
    ds_hwlock_write(&wod->buf_lock);
    to_prefetch = wod->params.prefetch_depth - wod->buf.prefetched;
    if (to_prefetch < wod->params.chunk_size) {
	/* wait til it's worth it */
	ds_hwlock_release_write(&wod->buf_lock);
	return 0;
    }
    
    /* need to signal our presence */
    wod->buf.prefetching = 1;
    
    if (cmgr_validate(wod->cid) > 0) {
	connp = cmgr_get(wod->cid, CMGR_LOCK);
	ASSERT(connp);
    } else {
	rc = -1;
	goto exit;
    }
    
    rc = 0;			/* accumulator */
    
    /* allow consumer to access control variables if we block during
       this read.  it is safe to assume there will not be a conflict
       at wod->buf.producer.  same goes for release down below */
    ds_hwlock_release_write(&wod->buf_lock);
    
    do {
	/* don't overrun the buffer! Note: this is safe to be outside
	 * the lock because producer_left is changed by US only,
	 * and chunk_size is changed only when we aren't running */
	this_chunk = (wod->buf.producer_left < wod->params.chunk_size) ?
	    wod->buf.producer_left : wod->params.chunk_size;
	
	ds_thread_log(100, "pf before bitw_read");
	result = bitw_read(wod->buf.producer, this_chunk, connp, &el_msec);
	ds_thread_log(100, "pf after bitw_read");
	
	/* get the lock back to munge the state */
	ds_hwlock_write(&wod->buf_lock);
	
	if (result > 0) {
	    if (this_chunk == wod->buf.producer_left) {
		/* wrap around */
		wod->buf.producer = wod->buf.buf;
		wod->buf.producer_left = wod->params.prefetch_depth;
	    } else {
		wod->buf.producer += result;
		wod->buf.producer_left -= result;
	    }
	    wod->buf.prefetched += result;
	    rc += result;
	    wod->rpc_msec += el_msec;
	    wod->chunk_count++;
	} else {
	    ds_hwlock_release_write(&wod->buf_lock);
	    rc = -1;
	    goto exit;
	}
	
	/* give any blocked readers the chance to consume this
	   newly acquired chunk */
	ds_hwlock_release_write(&wod->buf_lock);
	
	if (wod->buf.prefetched << 1 >= wod->params.prefetch_depth)
	    /* once we've filled at least half the buffer, let
	     * any blocked readers continue
	     * XXX: is this necessary?  if so, better heuristic? */
	    ds_thread_wakeup((char *)&wod->buf_lock);
	
    } while (wod->buf.prefetched < wod->params.prefetch_depth);
    
    ds_thread_log(1, "exitting prefetcher (got %d)", rc);
    wod->buf.prefetching = 0;  /* we're outta here... */
    
    /* may awaken either a pending close, bitw_set_params or bitw_rdwr 
    ds_thread_wakeup((char *)&wod->buf_lock); */
    
    cmgr_put(wod->cid, CMGR_LOCK);
    
exit:
    if (rc < 0)
	cmgr_markdown(wod->cid);
    
    return rc;			/* XXX currently ignored by callers */
}
