#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/viceroy.h"
#include "include/onode-manager.h"

#include "mapw_private.h"
#include "ody_map.h"
#include "mapw_comm.h"

#include "odyssey/viceroy/rsrc_en.h" /* XXXAPM */

/* mapw_comm_connect creates a connection to the specified server */
cmgr_id_t mapw_comm_connect (char* server)
{
  cmgr_initparams_t cip;
  cmgr_id_t cid;

  ds_thread_log (10, "mapw: mapw_comm_connect entered");

  ds_thread_log (100, "mapw: connecting to hostname: %s, port: %d",
		 server, MAPW_SERVER_PORT);

  /* Initialize connection structure */
  bzero (&cip, sizeof(cmgr_initparams_t));

  cip.ci_type    = CMGR_RPC2;

  cip.ci_restart.tv_sec = 90;	/* 1.5 minutes between restarts */
  cip.ci_restart.tv_usec = 0;

  cip.ci_params.ci_rpc2.hid.Tag = RPC2_HOSTBYNAME;
  if (strlen (server) >= 64) {
    ds_thread_log(0, "mapw: Machine name %s too long!", server);
    return (CMGR_ERROR);
  } /* endif */

  strcpy(cip.ci_params.ci_rpc2.hid.Value.Name, server);
    
  cip.ci_params.ci_rpc2.pid.Tag = RPC2_PORTBYINETNUMBER;
  cip.ci_params.ci_rpc2.pid.Value.InetPortNumber = htons(MAPW_SERVER_PORT);
    
  cip.ci_params.ci_rpc2.sid.Tag = RPC2_SUBSYSBYID;
  cip.ci_params.ci_rpc2.sid.Value.SubsysId = MAP_SUBSYSID;

  /* Bind Parameters */
  cip.ci_params.ci_rpc2.bp.SecurityLevel = RPC2_OPENKIMONO;
  cip.ci_params.ci_rpc2.bp.EncryptionType = 0;
  cip.ci_params.ci_rpc2.bp.SideEffectType = SMARTFTP;	/* allow use of SFTP */
  cip.ci_params.ci_rpc2.bp.ClientIdent = NULL;
  cip.ci_params.ci_rpc2.bp.SharedSecret = NULL;
    
  /* Create new binding */
  cid = cmgr_create(&cip);
  if (cid < 0) {
    ds_thread_log (0, "mapw: connection create failed, hostname = %s, "
		   "port = %d", server, MAPW_SERVER_PORT);
  } /* endif */

  ds_thread_log (0, "mapw: connection created: %d", cid);

  ds_thread_log(10, "mapw: mapw_comm_connect completed");

  return (cid);
}

/* mapw_comm_getconn validates and gets a connection with the specified
   id.  If the function succeeds, it returns a pointer to the connection.
   If it fails, it returns 0 */
cmgr_connection_t* mapw_comm_getconn (struct warden* wp)
{
  MapClientData* wdata = wp->w_data;
  cmgr_connection_t* pconn = 0;
  cmgr_id_t connid;
  int validity;

  ds_thread_log(10, "mapw: mapw_comm_getconn entered");

  /* Establish connection if it does not yet exist */
  connid = wdata->mw_connid;
  if (connid == CMGR_CID_NONE) {
    connid = mapw_comm_connect (wdata->mw_server);
    wdata->mw_connid = connid;
  }

  /* Validate the connection */
  validity = cmgr_validate (connid);
  if (validity > 0) {

    /* Get the RPC2 handle from the specified connection */
    pconn = cmgr_get(connid, CMGR_LOCK);
    if (!pconn) {
      ds_thread_log (0, "mapw: unable to get connection %d", connid);
    } /* endif */
    
  } else {
    /* Connection is down */
    ds_thread_log (0, "mapw: connection is down");
  } /* endif */

  ds_thread_log(10, "mapw: mapw_comm_getconn completed");

  return (pconn);
}

/* mapw_comm_lookup checks if the specified file exists on the server */
int mapw_comm_lookup (struct warden* wp, char* name, int* vnum_out, 
		      int* vlen_out)
{
  MapClientData* wdata = wp->w_data;
  RPC2_CountedBS nameBS;
  RPC2_Integer vnum;
  RPC2_Integer vlen;
  cmgr_connection_t* pconn;
  int rc;

  ds_thread_log(10, "mapw: mapw_comm_lookup entered");

  /* Get the appropriate connection from the connection manager */
  pconn = mapw_comm_getconn (wp);
  if (!pconn) {
    return (CMGR_ERROR);
  } /* endif */

  /* Make the RPC */
  nameBS.SeqLen = (RPC2_Integer) strlen(name)+1;
  nameBS.SeqBody = (RPC2_ByteSeq) name;

  rc = map_lookup (CMGR_RPC2_HANDLE (pconn), &nameBS, &vnum, &vlen);
  if (rc < RPC2_FLIMIT) {

    ds_thread_log (0, "mapw: RPC failed, rc = %d", rc);

    /* Mark down the connection */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (wdata->mw_connid)) {
	ds_thread_log (0, "mapw: unable to mark down the connection");
      } /* endif */
    } /* endif */

  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (0, "mapw: lookup failed, name = %s, rc = %d", name, rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (wdata->mw_connid, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "mapw: unable to put connection %d", wdata->mw_connid);
  } /* endif */

  if (rc == RPC2_SUCCESS) {
    *vnum_out = vnum;
    *vlen_out = vlen;
  } /* endif */

  ds_thread_log(10, "mapw: mapw_comm_lookup completed");

  return (rc);
}

/* mapw_comm_open fetches the specified map from the server */
int mapw_comm_getmap (struct warden* wp, struct viceroy_onode* op)
{
  struct timeval rpc_tv1, rpc_tv2, rpc_tv3;
  MapClientData*      wdata = wp->w_data;
  SE_Descriptor       map_data_sftp;
  cmgr_connection_t*  pconn;
  struct map_onode_t* pmap;
  char*               buf;
  int rc;

  ds_thread_log(10, "mapw: mapw_comm_getmap entered");

  /* Get the appropriate connection from the connection manager */
  pconn = mapw_comm_getconn (wp);
  if (!pconn) {
    return (CMGR_ERROR);
  } /* endif */

  /* Need to allocate onode data field - include a "fudge factor" because
     distilled headers may be longer than the original file. */
  pmap = (struct map_onode_t *) op->o_data;
  if (pmap->mo_data) {
    FREE (pmap->mo_data);
  }
  pmap->mo_datalen = pmap->mo_len + 1024;
  NALLOC (pmap->mo_data, char, pmap->mo_datalen);

  /* Map will be SFTPd into the onode data */
  bzero(&map_data_sftp,sizeof(map_data_sftp));
  map_data_sftp.Tag = SMARTFTP;
  map_data_sftp.Value.SmartFTPD.Tag = FILEINVM;
  map_data_sftp.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
  map_data_sftp.Value.SmartFTPD.ByteQuota =
    (RPC2_Integer) pmap->mo_datalen;;
  map_data_sftp.Value.SmartFTPD.SeekOffset = 0;
  map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.MaxSeqLen =
    (RPC2_Integer) pmap->mo_datalen;
  map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = 
    (RPC2_ByteSeq) pmap->mo_data;

  /* Make the RPC */
  gettimeofday (&rpc_tv1, 0);

  rc = map_open (CMGR_RPC2_HANDLE (pconn), op->o_fid.ofid_vnum, 
		 &map_data_sftp);


  gettimeofday (&rpc_tv2, 0);
  timersub (&rpc_tv2, &rpc_tv1, &rpc_tv3);
  ds_thread_log (0, "fetched map in %d.%06d seconds",
		 rpc_tv3.tv_sec, rpc_tv3.tv_usec);

  if (rc < RPC2_FLIMIT) {

    ds_thread_log (0, "mapw: RPC failed, rc = %d", rc);

    /* Mark down the connection */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (wdata->mw_connid)) {
	ds_thread_log (0, "mapw: unable to mark down the connection");
      } /* endif */
    } /* endif */

  } /* endif */

  /* The buffer may have been way too large if the file was distilled.
     So, reallocate the buffer into a more reasonable space.
     NOTE: If RPC2 were to dynamically allocate the buffer, it would
     remove all this unnecessary work.  D'oh! */
  if (map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen <
      pmap->mo_datalen) {
    
    pmap->mo_datalen = 
      map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen;
    NALLOC (buf, char, pmap->mo_datalen);
    memcpy (buf, pmap->mo_data, pmap->mo_datalen);
    FREE (pmap->mo_data);
    pmap->mo_data = buf;

  }
      
      
  if (rc != RPC2_SUCCESS) {
    ds_thread_log (0, "mapw: open failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (wdata->mw_connid, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "mapw: unable to put connection");
  } /* endif */

  ds_thread_log(10, "mapw: mapw_comm_getmap completed");

  return (rc);
}

int mapw_comm_getdir (struct warden* wp, int offset, int* count, void* data)
{
  MapClientData*      wdata = wp->w_data;
  SE_Descriptor       map_data_sftp;
  cmgr_connection_t*  pconn;
  int rc;

  ds_thread_log(10, "mapw: mapw_comm_getdir entered");

  /* Get the appropriate connection from the connection manager */
  pconn = mapw_comm_getconn (wp);
  if (!pconn) {
    return (CMGR_ERROR);
  } /* endif */

  /* Dir will be SFTPd into the user buffer */
  bzero(&map_data_sftp,sizeof(map_data_sftp));
  map_data_sftp.Tag = SMARTFTP;
  map_data_sftp.Value.SmartFTPD.Tag = FILEINVM;
  map_data_sftp.Value.SmartFTPD.TransmissionDirection = SERVERTOCLIENT;
  map_data_sftp.Value.SmartFTPD.ByteQuota = (RPC2_Integer) -1;
  map_data_sftp.Value.SmartFTPD.SeekOffset = 0;
  map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.MaxSeqLen =
    (RPC2_Integer) *count;
  map_data_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = 
    (RPC2_ByteSeq) data;

  /* Make the RPC */
  rc = map_getdir (CMGR_RPC2_HANDLE (pconn), (RPC2_Integer *) count, 
		   &map_data_sftp);
  if (rc < RPC2_FLIMIT) {

    ds_thread_log (0, "mapw: getdir RPC failed, rc = %d", rc);

    /* Mark down the connection */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (wdata->mw_connid)) {
	ds_thread_log (0, "mapw: unable to mark down the connection");
      } /* endif */
    } /* endif */

  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (0, "mapw: open failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (wdata->mw_connid, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "mapw: unable to put connection");
  } /* endif */

  ds_thread_log(10, "mapw: mapw_comm_getdir completed");

  return (rc);
}

int mapw_comm_setfilter (struct warden* wp, void* filter, int size)
{
  MapClientData*      wdata = wp->w_data;
  RPC2_BoundedBS      filterBBS;
  cmgr_connection_t*  pconn;
  int rc;

  ds_thread_log(10, "mapw: mapw_comm_setfilter entered");

  /* Get the appropriate connection from the connection manager */
  pconn = mapw_comm_getconn (wp);
  if (!pconn) {
    return (CMGR_ERROR);
  } /* endif */

  filterBBS.SeqLen = size;
  filterBBS.MaxSeqLen = size;
  filterBBS.SeqBody = filter;

  /* Make the RPC */
  rc = map_setfilter (CMGR_RPC2_HANDLE (pconn), &filterBBS);
  if (rc < RPC2_FLIMIT) {

    ds_thread_log (0, "mapw: setfilter RPC failed, rc = %d", rc);

    /* Mark down the connection */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (wdata->mw_connid)) {
	ds_thread_log (0, "mapw: unable to mark down the connection");
      } /* endif */
    } /* endif */

  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (0, "mapw: set filter failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (wdata->mw_connid, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "mapw: unable to put connection");
  } /* endif */

  ds_thread_log(10, "mapw: mapw_comm_setfilter completed");

  return (rc);  

}
