#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>

#include <sys/queue.h>
#include <sys/uio.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_list.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/onode-manager.h"

#include "mapwarden.h"
#include "mapw_private.h"
#include "mapw_comm.h"
#include "libodymap_tsops.h"

/* Global data */
char* map_server_name = 0;

struct warden map_warden = {
  MAP_TTYPE,
  "Map",
  0,
  mapw_parse_arguments,
  "Usage: -mapw_server <server name> ",
  mapw_initialize,
  mapw_get_CID,
  NULL,
  mapw_get_root,
  mapw_lookup,
  mapw_access,
  mapw_get_attr,
  mapw_open,
  mapw_close,
  mapw_read_write,
  mapw_read_write_dispose,
  mapw_readdir,
  (w_readdir_dispose_fn) NULL,
  mapw_TSOP,
  mapw_TSOP_dispose,
#ifdef SETS
  (w_expand_fn) NULL,
  (w_cursexpand_fn) NULL,
  (w_prefetch_fn) NULL,
#endif
  0,
  0,
  (void *)NULL
};

/* mapw_parse_argments() takes in an argc/argv pair and an index to
   the current argument being parsed.  It processes as many arguments
   as it can starting at the specified index and returns the number of
   arguments parsed. */
int mapw_parse_arguments(int cur, int argc, char *argv[])
{
  int argsParsed = 0;
  int done = 0; /* Set when an argument that cannot be parsed is found */

  while (cur + argsParsed < argc && !done) {

    if (strcmp(argv[cur + argsParsed], "-mapw_server") == 0) {
      if (cur + argsParsed + 1 >= argc) {
	return -1;
      } /* endif */
      if (*(argv[cur + argsParsed + 1]) == '-') {
	return -1;
      } /* endif */
      map_server_name = strdup(argv[cur + argsParsed + 1]);
      argsParsed += 2;
    } else {
      done = 1; /* Found an argument we can't parse */
    } /* endif */

  } /* endwhile */

  return (argsParsed);
}

/* mapw_initialize() performs the initialization of the map warden.
   It allocates a private data area in the warden structure and copies
   data into the area.  It creates a root node for the single tome.  The
   function returns zero on succes, non-zero otherwise */
int mapw_initialize(struct warden *wp)
{
  MapClientData*         wdata;
  struct ofid            root_ofid;
  struct onode_create_in root_onode_params;
  struct viceroy_onode*  rop = 0;
  int                    rc = 0; /* Assume successful completion */
  
  ds_thread_log(10, "mapw: mapw_initialize entered");

  wp->w_data = NULL;

  /* A server name must be specified */
  if (map_server_name == 0) {
    ds_thread_log (0, "mapw: remote server name not specified");
    ds_thread_log (0, "mapw: map warden will not start");
    return (-1);
  } /* endif */

  /* Allocate client data for structure for map warden.  This is
     attached to the warden data structure */
  ALLOC (wp->w_data, MapClientData);
  wdata = wp->w_data;
  wdata->mw_server = map_server_name;
  ds_hwlock_init (&wdata->mw_lock); /* Initialize lock */
  wdata->mw_connid = CMGR_CID_NONE;

  /* set up the root onode in init, since there is only one volume */
  root_ofid.ofid_ttype = MAP_TTYPE;
  root_ofid.ofid_tome  = MAP_TOME;	/* we only support tome #1 for now */
  root_ofid.ofid_vnum  = 0;
  root_ofid.ofid_uniq  = 0;
  
  root_onode_params.oc_flags = ON_ROOTOFVOL | ON_UNCACHEABLE;
  root_onode_params.oc_vtype = O_VDIR; 
  root_onode_params.oc_ofree = NULL;
  root_onode_params.oc_data  = NULL;
  
  rop = onode_create(&root_ofid, &root_onode_params);
  if (!rop) {
    ds_thread_log(0, "mapw: Could not create volume onode root");
    return (-1);
  } /* endif */

  ds_thread_log(10, "mapw: mapw_initialize completed");
  
  return (rc);
}

/* mapw_get_root() returns the root o-node of the specified tome.
   Since there is only one tome, only that tome may be specified.  The
   node is found by looking for the o-node with vnum and uniq of zero.
   The o-node is readlocked before being returned. The function returns
   zero on success, non-zero otherwise. */
int mapw_get_root(struct warden *wp, tome_t tome, struct viceroy_onode **ppRoot)
{
  struct ofid of;

  ds_thread_log(10, "mapw: mapw_get_root entered");
  
  if (tome != MAP_TOME) {
    ds_thread_log(10, "mapw: requested tome (%d) not supported", tome);
    return(EINVAL);	
  } /* endif */
  
  of.ofid_ttype = MAP_TTYPE;
  of.ofid_tome  = tome;
  of.ofid_vnum  = 0;		/* for map warden, {X,X,0,0} indicates */
  of.ofid_uniq  = 0;		/* an ofid for a volume. */
  
  *ppRoot = onode_get(&of);
  if (*ppRoot == NULL) {
    ds_thread_log(0, "mapw: onode_get() failed for warden tome "
		  "%d's root onode.", tome);
    return(EINVAL);
  } /* endif */
  
  /* Indicate that the root is mounted on Odyssey root onode */
  (*ppRoot)->o_u.ou_mountedon.ofid_ttype = 0;
  (*ppRoot)->o_u.ou_mountedon.ofid_tome = 0;
  (*ppRoot)->o_u.ou_mountedon.ofid_vnum = 0;
  (*ppRoot)->o_u.ou_mountedon.ofid_uniq = 1;

  onode_readlock ((*ppRoot));

  ds_thread_log(10, "mapw: mapw_get_root completed");
  
  return (0);
}

/* mapw_lookup() resolves the specified name.  Basically, any name is
   considered valid, and a new o-node is created and returned.  Right now,
   there is no checking to see if an o-node with that particular name
   already exists - I might add this later.  The function returns zero
   on success. */
int mapw_lookup(struct warden* wp, struct viceroy_onode* odp, 
		struct ody_cred* cred, char* name, char** remain, 
		struct viceroy_onode** opp)
{
  int                    vnum, vlen;
  struct ofid            map_ofid;
  struct viceroy_onode*  map_onode;
  int                    rc;
  struct map_onode_t*    pmap = 0;

  ds_thread_log(10, "mapw: mapw_lookup entered");
  
  ds_thread_log (10, "mapw: lookup on name: %s", name);

  /* make sure we were passed the root map onode */

  if (((odp->o_flags | ON_ROOTOFVOL) == 0) || 
      (odp->o_fid.ofid_ttype != MAP_TTYPE)) {
    ds_thread_log (0, "mapw: mapw_lookup not passed root of map volume");
    onode_release_read(odp);
    onode_put(odp);
    return (EINVAL);
  }

  if (name[0] == '\0') {
    ds_thread_log (10, "mapw: mapw_lookup passed NULL name");
    return (0);
  }
    
  /* Do a lookup on the server to see if the file exists */
  rc = mapw_comm_lookup (wp, name, &vnum, &vlen);
  if (vlen <= 0) {
    ds_thread_log (0, "mapw: mapw_lookup returned 0 length file");
    onode_release_read(odp);
    onode_put (odp);
    return (EINVAL);
  }
  
  ds_thread_log (10, "mapw: lookup returned vnum: %d, rc = %d", vnum, rc);

  if (rc == 0) {
    map_ofid.ofid_vnum  = vnum;
    map_ofid.ofid_ttype = MAP_TTYPE;
    map_ofid.ofid_tome  = MAP_TOME;
    map_ofid.ofid_uniq  = 1;
	
    map_onode = onode_get(&map_ofid);
    if (map_onode == NULL) { /* haven't looked up this map before */

      struct onode_create_in map_onode_params;
      
      map_onode_params.oc_flags = ON_UNCACHEABLE;
      map_onode_params.oc_vtype = O_VREG;
      map_onode_params.oc_ofree = mapw_ofree;
      ALLOC (pmap, struct map_onode_t);
      pmap->mo_len = vlen;
      pmap->mo_datalen = 0;
      pmap->mo_data = 0;
      map_onode_params.oc_data = pmap;
      
      map_onode = onode_create(&map_ofid, &map_onode_params);
      ASSERT(map_onode);
      
      onode_writelock(map_onode);
      map_onode->o_cachefile = NULL; /* not cacheable */
      onode_release_write(map_onode);
      
      map_onode = onode_get(&map_ofid);
      ASSERT(map_onode);

    } else {
      pmap = (struct map_onode_t *) map_onode->o_data;

      /* Should not be any data in the onode??? */
      ASSERT (pmap->mo_data == 0);

      pmap->mo_len = vlen;
      pmap->mo_datalen = 0;
      pmap->mo_data = 0;
    }

    /* we have a referenced onode, readlock it */
    onode_readlock(map_onode);
    *opp = map_onode;

  } /* endif */

  /* release the lock and reference on odp that was handed in to us */
  onode_release_read(odp);
  onode_put(odp);

  ds_thread_log(10, "mapw: mapw_lookup completed");

  return (rc);
}

/* mapwAccess() checks the specified user's access to a specified o-node.
   Since there is no security implemented, this function always returns 
   zero (success). */
int mapw_access(struct warden* wp, struct viceroy_onode* op, 
	       struct ody_cred* cred, int mode)
{
  ds_thread_log(10, "mapw: mapw_access entered");
  ds_thread_log(10, "mapw: mapw_access completed");
  return (0);
}

int mapw_get_attr(struct warden *wp, struct viceroy_onode *op,
		struct ody_vattr *vap, struct ody_cred *cred)
{
    time_t temp = 0;

    ds_thread_log(10, "mapw: mapw_get_attr entered");

    if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0)
      vap->va_type = O_VDIR;	/* this is the root - a directory */
    else
      vap->va_type = O_VREG;
    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;
    vap->va_blocksize = 512;
    vap->va_atime = temp;
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    ds_thread_log(10, "mapw: mapw_get_attr completed");

    return (0);
}

/* mapw_open() opens a map o-node.  It first checks to make sure that
   it is not the root o-node and that the map has not already been 
   fetched from the server.  If the map has not been fetched, it is
   fetched in its entirity.  Only opens for reading are supported. */
int mapw_open (struct warden *wp, struct viceroy_onode *op, 
	       struct ody_cred *cred, int mode)
{
  struct map_onode_t* pmap;
  int rc = 0;

  ds_thread_log(10, "mapw: mapw_open entered");

  ds_thread_log (10, "mapw: open on onode: %d", op->o_fid.ofid_vnum);

  /* Check to see if o-node is the root */
  if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0) {
    /* Opening root is OK */
    return (0);
  } /* endif */

  pmap = (struct map_onode_t *) op->o_data;

  if (pmap->mo_data == 0) {
    /* Fetch the map from the server */
    rc = mapw_comm_getmap (wp, op);
  }

  ds_thread_log(10, "mapw: mapw_open completed");

  return (rc);
}

/* mapw_readdir() returns contents of server-side directory */
int mapw_readdir (struct warden *wp, struct viceroy_onode *op,
		  struct ody_cred *cred, int *count, int offset,
		  void *data)
{
  int rc;
  
  ds_thread_log(10, "mapw: mapw_readdir entered");

  if (op->o_fid.ofid_vnum != 0 || op->o_fid.ofid_uniq != 0) {
    ds_thread_log(0, "Readdir on a non-root directory.");
    return EIO;
  }
  
  ds_thread_log (10, "mapw: mapw_readdir - offset %d, count %d", offset,
		 *count);

  rc = mapw_comm_getdir(wp, offset, count, data);
  
  ds_thread_log (10, "mapw: returning count = %d, rc = %d", *count, rc);

  ds_thread_log(10, "mapw: mapw_readdir completed");

  return rc;
}

/* mapw_close () is called when the client process closes the onode.  It
   does not do anything right now. */
int mapw_close(struct warden *wp, struct viceroy_onode  *op, 
	       struct ody_cred  *cred, int mode)
{
  ds_thread_log(10, "mapw: mapw_close entered");

  if (op->o_fid.ofid_vnum != 0 || op->o_fid.ofid_uniq != 0) {
    ds_thread_log (10, "mapw: purging (hopefully) %d", op->o_fid.ofid_vnum);
    onode_put(op);
  }

  ds_thread_log(10, "mapw: mapw_close completed");

  return (0);
}

/* mapReadWrite () is not supported.  All meaningful interactions with
   an open reference take place through TSOPs. */
int mapw_read_write (struct warden* wp, struct viceroy_onode* op, 
		     struct ody_cred* cred, int rwflag, int* count, 
		     int offset, int ioflag, void **data)
{
  struct map_onode_t* pmap;

  ds_thread_log(10, "mapw: mapw_read_write entered");

  if (rwflag == O_RW_WRITE) {
    return (EACCES);
  }

  /* Check to see if o-node is the root */
  if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0) {
    return (EISDIR);
  } /* endif */

  ds_thread_log(10, "count: %d, offset: %d, data: %x", *count, offset, 
		*data);
  
  pmap = (struct map_onode_t*) op->o_data;
  if (offset > pmap->mo_datalen) {
    return (EFAULT);
  }
  *data = pmap->mo_data + offset;
  if (*count > (pmap->mo_datalen - offset)) {
    *count = pmap->mo_datalen - offset;
  }

  ds_thread_log(10, "mapw: mapw_read_write completed");

  return (0);
}

/* mapw_read_write_dispose() doesn't do much of anything. It probably
   should never be called. */
int mapw_read_write_dispose (struct warden* wp, struct viceroy_onode* op, 
			     struct ody_cred* cred, int count, int offset,
			     void* data)
{
  ds_thread_log(10, "mapw: mapw_read_write_dispose entered");
  ds_thread_log(10, "mapw: mapw_read_write_dispose completed");
  return (0);
}

/* mapw_TSOP() handles all type-specific operations for the map
   warden.  */
int mapw_TSOP(struct warden* wp, int tsvnum, struct viceroy_onode* op,
	      struct ody_cred* cred, int opnum, size_t argsz,
	      void* argp, size_t* retsz, void** retp)
{
  long result;

  ds_thread_log(10, "mapw: mapw_TSOP entered");

  /* check the version number */
  if(tsvnum != MAPWARDEN_VERSION) {
    return(ODY_EBADVERSION);
  } /* endif */

  /* dispatch TSOPs, based on opnum */
  switch(opnum) {

    case MAPW_SET_FILTER:
      
      /* Pass filter request to the server */
      /* Note: multiple clients are poorly supported by this */
      result = mapw_comm_setfilter (wp, argp, argsz);
      *retsz = 0;
      *retp = NULL;
      
      break;

    default:
      result = EOPNOTSUPP;
      break;
  } /* end switch */

  ds_thread_log(10, "mapw: mapw_TSOP completed");

  return (result);
}

/* mapw_TSOP_dispose() does nothing */
int mapw_TSOP_dispose(struct warden* wp, struct viceroy_onode* op, 
		      struct ody_cred* cred, int opnum, size_t argsz, 
		      void* argp, size_t retsz,void* retp)
{
  ds_thread_log(10, "mapw: mapw_TSOP_dispose entered");
  ds_thread_log(10, "mapw: mapw_TSOP_dispose completed");

  return(0);
}

/* mapw_get_CID() returns the connection used to communicate with the
   map server */
cmgr_id_t mapw_get_CID(struct warden *wp, struct viceroy_onode  *op)
{
  MapClientData* wdata = wp->w_data;
  
  ds_thread_log (10, "mapw: mapw_get_CID entered");
  ds_thread_log (10, "mapw: mapw_get_CID completed");

  return (wdata->mw_connid);
}

/* Deallocates a map onode */
void mapw_ofree(void* odata)
{
  struct map_onode_t* pmap;

  ds_thread_log(10, "mapw: mapw_ofree entered");
  
  pmap = (struct map_onode_t*) odata;
  if (pmap) {
    FREE (pmap->mo_data);
    pmap->mo_data = 0;
    FREE (pmap);
  }

  ds_thread_log(10, "mapw: mapw_ofree completed");
}
