#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * QuickTime warden
 *
 * This handles accesses on QuickTime movies under the Odyssey namespace
 *
 * 2 Sep 1996 -- First, dirty version
 *
 * 16 Oct 1996 -- Now handles all the TSOPs correctly
 *
 * 16 Jan 1997 -- Second, slightly cleaner version. Creates vnode numbers
 *                and maps them to file names, which the server understands.
 *                qt_openmovie now calls two RPC2 functions, QT_OpenMovie_RPC
 *                and QT_OpenTrack_RPC
 *
 * Specs for the functions provided here are in warden.h
 *
 * Dushyanth Narayanan <bumba+@cs.cmu.edu>
 *
 *
 */

#include <kernel/vfs/includes/ody.h>
#include <sys/queue.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <netinet/in.h>

#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_log.h"
#include "base/ds/ds_safeq.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <lwp/lock.h>

#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include "include/odytypes.h"

#include "include/cm_rpc2.h"
#include "include/cm.h"

#include "include/viceroy_onode.h"
#include "include/odyssey.h"
#include "include/warden.h"
#include "include/onode-manager.h"
#include "include/cache_mgr.h"

#include <db.h>
#include "qtw_cache.h"
#include "qt_server.h"
#include "ody_qt.h"
#include "qt.h"
#include "qt_warden.private.h"
#include "qtw_prefetch.h"

#include "odyssey/viceroy/rsrc_en.h"

#define MAX_HOSTNAME_LEN 64
static char *server_hostname[QT_MAXTOMES]; /* names of tome servers */

static int loglevel = 0;
static int num_prefetch_threads = 4;
static int prefetch_bufsize = 1000000; /* prefetch buffer size in bytes */
static int fetch_maxtime = 500;      /* at current bandwidth estimate,
					don't fetch at a time more than
					this many msec's worth */
static int fetch_mintime = 250;     /* at current bandwith estimate, a fetch
				       should take at least this long, unless
				       the buffer's more than half empty */
static int fetch_maxplaytime = 1000; /* don't fetch at a time more frames
					than would take this much time to
					play back. should be bigger than
					the fetch time, otherwise makes
					no sense */
static int drop_late_frames = 0;    /* whether to always return a frame
				       >= that requested by application */

int refetch_low_quality = 1;        /* when b/w improves, refetch frames that
				       have already been fetched at a lower
				       quality? */
int to_multiple = 2;                /* timeout RPC when fetch takes this
				       multiple of the expected time. */
				       

static int         qt_parseargs  __P((int cur,
				      int argc,
				      char *argv[]));

static int         qt_init       __P((struct warden *wp));

static int         qt_root       __P((struct warden *wp,
				      tome_t tome,
				      struct viceroy_onode **root_onodep));

static int         qt_lookup     __P((struct warden *wp,
				      struct viceroy_onode  *odp,
				      struct ody_cred  *cred,
				      char          *name,
				      char          **remain,
				      struct viceroy_onode  **opp));

static int         qt_access    __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         qt_getattr   __P((struct warden *wp, 
				     struct viceroy_onode  *op,
				     struct ody_vattr  *vap,
				     struct ody_cred  *cred));

static int         qt_open      __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         qt_close     __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int mode));

static int         qt_rdwr      __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int rwflag,
				     int *count,
				     int offset,
				     int ioflag,
				     void **data));

static int         qt_rdwr      __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int rwflag,
				     int *count,
				     int offset,
				     int ioflag,
				     void **data));

static int         qt_rdwr_dispose 
				__P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int count,
				     int offset,
				     void *data));

static int         qt_tsop      __P((struct warden *wp,
				     int tsvnum,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int opnum,
				     size_t argsz,
				     void *argp,
				     size_t *retsz,
				     void **retp));

static int         qt_tsop_dispose
                                __P((struct warden *wp,
				     struct viceroy_onode *op,
				     struct ody_cred *cred,
				     int opnum,
				     size_t argsz,
				     void *argp,
				     size_t retsz,
				     void *retp));

static cmgr_id_t     qt_getcid    __P((struct warden *wp,
				     struct viceroy_onode *op));

#ifdef INSTRUMENTED
static void          qt_pstats    __P((struct warden *wp,
				     FILE *fp));
#endif /* INSTRUMENTED */

struct warden qt_warden = {
    QT_TTYPE,
    "QuickTime",
    0,
    qt_parseargs,
    "Usage: -qtw_server <tome #> <server_hostname>\n"
    "      [-qtw_loglevel <level>]\n"
    "      [-qtw_to_multiple <multiple>]\n"
    "      [-qtw_droplate]\n"
    "      [-qtw_prefetch_threads <numthreads>]\n"
    "      [-qtw_prefetch_bufsize <bytes>]\n"
    "      [-qtw_fetch_maxtime <msec>]\n"
    "      [-qtw_fetch_mintime <msec>]\n"
    "      [-qtw_fetch_maxplaytime <msec>]\n"
    "      [-qtw_norefetch]",
    qt_init,
    qt_getcid,
#ifdef INSTRUMENTED
    qt_pstats,
#else /* INSTRUMENTED */
    (w_pstats_fn) NULL,
#endif /* INSTRUMENTED */
    qt_root,
    qt_lookup,
    qt_access,
    qt_getattr,
    qt_open,
    qt_close,
    qt_rdwr,
    qt_rdwr_dispose,
    (w_readdir_fn) NULL, /* qt_readdir not yet ready for prime time */
    (w_readdir_dispose_fn) NULL, /* since we never allocate our own bufs for
				    readdir */
    qt_tsop,
    qt_tsop_dispose,
#ifdef SETS
    (w_expand_fn) NULL,
    (w_cursexpand_fn) NULL,
    (w_prefetch_fn) NULL,
#endif

    0,
    0,

    (void *)NULL		/* qt_data */
};

/* Parse any viceroy command-line arguments that were meant for us */
/* Currently the only argument we take is of the form
   -qtw_server. This argument MUST be present, else it is an error */

static
int
qt_parseargs(cur, argc, argv)
    int cur;
    int argc;
    char *argv[];
{

    tome_t tome;

    if (argv[cur] && (strcmp(argv[cur], "-qtw_server") == 0)) {

	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	tome = atoi(argv[cur++]);
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following tome # */

	if ((tome < 0) || (tome >= QT_MAXTOMES))
	    return -1;
	
	server_hostname[tome] = strdup(argv[cur]);
	ASSERT(server_hostname[tome]);

	if (strlen(server_hostname[tome]) >= MAX_HOSTNAME_LEN)
	    return -1;

	return 3; /* we ate three arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_loglevel") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	loglevel = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_to_multiple") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	to_multiple = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_droplate") == 0)) {
	drop_late_frames = 1;
	return 1; /* we ate one argument */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_prefetch_threads") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	num_prefetch_threads = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_prefetch_bufsize") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	prefetch_bufsize = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_fetch_maxtime") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	fetch_maxtime = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_fetch_mintime") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	fetch_mintime = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_fetch_maxplaytime") == 0)) {
	cur++;
	if ((cur >= argc) || (argv[cur] == NULL) || (argv[cur][0] == '-')) 
	    return -1; /* no argument following flag */

	fetch_maxplaytime = atoi(argv[cur]);

	return 2; /* we ate two arguments */
    }

    if (argv[cur] && (strcmp(argv[cur], "-qtw_norefetch") == 0)) {
	refetch_low_quality = 0;
	return 1; /* we ate one argument */
    }

    return 0;
}

/* free warden-specific data associated with an onode */
static void qt_ofree(o_data)
    struct onode_data *o_data;
{
    if (o_data) {
	/* assumption: if the onode is unreferenced, so is the cache object */
	/* so we can safely get rid of the cache object */
	/* this will BREAK if multiple onodes reference the same cache obj */
	if (o_data->thumbnail) {
	    qtw_cache_release(o_data->thumbnail);
	    o_data->thumbnail = NULL;
	}
	ASSERT(o_data->thumbnail_fd == -1);
	FREE(o_data->all_hdrs);
	FREE(o_data->thumbnail_frame_list);
    }
    FREE(o_data);
}

/* function that prefetch threads execute; pick up a request from the
   prefetch request queue; call qtw_prefetch(); repeat */
static
int
Prefetch(cqtwdp)
    char *cqtwdp;
{
    struct qtw_data *qtwdp = (struct qtw_data *)cqtwdp;
    ds_safeq_t *req_q = qtwdp->qtw_prefetch_req;
    struct qtw_prefetch *qtwp;

    while (1) {
	ds_thread_log(100,"waiting for movie to prefetch");
	qtwdp->qtw_pf_avail++;
	qtwp = (struct qtw_prefetch *) ds_safeq_deq(req_q);
	qtwdp->qtw_pf_avail--;
	qtw_prefetch(qtwp);
    }
}
	

/* Create the root onodes for all the tomes that are being served
   and initialize the warden structure */

static
int
qt_init(wp)
    struct warden *wp;
{
    cmgr_initparams_t ci;
    int rc;
    struct qtw_data *qtwdp;
    struct ofid root_ofid;
    struct viceroy_onode *root_onode;
    struct onode_create_in root_onode_params;
    int result = 0;
    tome_t tome;
    int num_tomes = 0;
    int i;


    for (tome = 0; tome < QT_MAXTOMES; tome++) {
	if (server_hostname[tome] != NULL)
	    num_tomes++;
    }
    if (num_tomes == 0) {
	ds_thread_log(0, "qt_init: no servers specified");
	return -1;
    }

    ALLOC(qtwdp, struct qtw_data);
	
    /* try to connect to tome servers */

    /* XXX TMP : this is not the right way to do it; basically we want
       a flag that says, this connection is used very intermittently,
       don't give it "fair share" bandwidth */

    for (tome = 0; tome < QT_MAXTOMES; tome++) {
	struct qtt_data *qttdp;

	qtwdp->qtw_tome[tome] = NULL;
	if (server_hostname[tome] == NULL) /* not serving this volume */
	    continue;

	bzero(&ci, sizeof(cmgr_initparams_t));
// JNF - this isn't really true.
//	ci.ci_flags |= CMGR_CONNISLOCAL;
	ci.ci_type = CMGR_RPC2;
	ci.ci_restart.tv_sec = 2;/* don't restart oftener than once in 2 sec */
	ci.ci_restart.tv_usec = 0;
	
	ci.ci_params.ci_rpc2.hid.Tag = RPC2_HOSTBYNAME;
	strcpy(ci.ci_params.ci_rpc2.hid.Value.Name, server_hostname[tome]);
	ci.ci_params.ci_rpc2.pid.Tag = RPC2_PORTBYINETNUMBER;
	ci.ci_params.ci_rpc2.pid.Value.InetPortNumber = htons(QTPORTAL);
	ci.ci_params.ci_rpc2.sid.Tag = RPC2_SUBSYSBYID;
	ci.ci_params.ci_rpc2.sid.Value.SubsysId = QTSUBSYSID;
	
	ci.ci_params.ci_rpc2.bp.SecurityLevel = RPC2_OPENKIMONO;
	ci.ci_params.ci_rpc2.bp.EncryptionType = 0;
	ci.ci_params.ci_rpc2.bp.SideEffectType = SMARTFTP;
	ci.ci_params.ci_rpc2.bp.ClientIdent = NULL;
	ci.ci_params.ci_rpc2.bp.SharedSecret = NULL;
	
	rc = cmgr_create(&ci);
	
	if (rc < 0) {
	    result = ECONNREFUSED;
	    ds_thread_log(0,"qt_init: Cannot connect to server on %s",
			  server_hostname);
	    goto bad;
	}
	
	ds_thread_log(0,"qt_init: created conn %d to server %s for tome %d",
		      rc,server_hostname[tome],tome);
	
	/* we got the connection, squirrel away the connection id */

	ALLOC(qttdp, struct qtt_data);
	qttdp->qtd_cid = rc;
	
	/* make up an ofid for the root */
	root_ofid.ofid_ttype = QT_TTYPE;
	root_ofid.ofid_tome = tome;
	root_ofid.ofid_vnum = QT_ROOT_VNUM;
	root_ofid.ofid_uniq = QT_ROOT_UNIQ;

	/* Create the root onode */
	root_onode_params.oc_flags = ON_ROOTOFVOL | ON_UNCACHEABLE;
	root_onode_params.oc_vtype = O_VDIR;
	root_onode_params.oc_ofree = NULL;
	root_onode_params.oc_data = NULL;
	
	root_onode = onode_create(&root_ofid, &root_onode_params);
	ASSERT(root_onode);
	
	/* Grab a lock on the root onode so we can play with it */
	onode_writelock(root_onode);
	
	/* We have no cache file: All accesses go to the server */
	root_onode->o_cachefile = NULL;
	
	/* We're done fiddling with the onode, unlock it */
	onode_release_write(root_onode);
	
	/* Squirrel away root onode for qt_root() */
	qttdp->qtd_root_onode = root_onode;
	
	qtwdp->qtw_tome[tome] = qttdp;
    }

    /* create the prefetch threads and the request queue */
    qtwdp->qtw_pf_avail = 0;
    qtwdp->qtw_prefetch_req = ds_safeq_create();
    for (i = 0; i < num_prefetch_threads; i++)
	ASSERT(ds_thread_create(Prefetch, (char *)qtwdp, DS_THREAD_NORMAL,
				"prefetch", loglevel));
    
    /* get cache space from the viceroy */
    qtwdp->qtw_cache_files = cache_request(wp, CACHE_SPACE, CACHE_FILES);
    if (qtwdp->qtw_cache_files) {
	qtwdp->qtw_ncache_files = CACHE_FILES;
	ds_thread_log(100,"qt_init: Got %d cache files with %ld kilobytes\n",
		      CACHE_FILES, CACHE_SPACE);
    }
    else {
	qtwdp->qtw_ncache_files = 0;
	ds_thread_log(0,"qt_init: cache_request(%d,%ld) failed\n",CACHE_FILES,
		      CACHE_SPACE);
    }

    wp->w_data = qtwdp;

    ds_thread_log(100,"qt_init: created root (%x.%x.%x.%x)",
		  root_ofid.ofid_ttype,root_ofid.ofid_tome,
		  root_ofid.ofid_vnum,root_ofid.ofid_uniq);


    return 0;			/* Everything is fine. */
  
 bad:
    ASSERT(result);
    return result;
}


/* Return root onode for the specified volume. */

static
int
qt_root(wp, tome, root_onodep)
    struct warden *wp;
    tome_t tome;
    struct viceroy_onode **root_onodep;
{
    struct viceroy_onode *root_onode = NULL;
    struct ofid root_ofid;
    struct qtw_data *qtwdp;
    struct qtt_data *qttdp;
    int result = 0;

    if ((tome < 0) || (tome >= QT_MAXTOMES))
	return ENOENT;

    qtwdp = wp->w_data;
    qttdp = qtwdp->qtw_tome[tome];

    if (qttdp == NULL)
	return ENOENT;

    root_onode = qttdp->qtd_root_onode;

    /* Bump the refcount, because the caller is going to do a put */
    root_ofid = root_onode->o_fid;
    root_onode = onode_get(&root_ofid);

    if (root_onode == NULL) {
	result = errno;
	ds_thread_log(0,"qt_root: onode_get() failed for root onode: %s",
		      strerror(result));
	goto bad;
    }

    /* Return the onode to the viceroy read-locked */

    onode_readlock(root_onode);

    /* Everything's fine -- return locked, referenced root onode */
    *root_onodep = root_onode;

    ds_thread_log(100,"qt_root: success: ofid (%x.%x.%x.%x)",
		  root_ofid.ofid_ttype,root_ofid.ofid_tome,
		  root_ofid.ofid_vnum,root_ofid.ofid_uniq);

    return 0;
  
 bad:
    ASSERT(result);
    return result;
}

/* do a lookup on an onode and a relative path */
/* We operate in a flat namespace; so the onode passed in MUST be the
   root of the tome */

static
int
qt_lookup(wp, odp, cred, name, remain, opp)
    struct warden *wp;
    struct viceroy_onode *odp;
    struct ody_cred *cred;
    char *name;
    char **remain;
    struct viceroy_onode **opp;
{
    RPC2_CountedBS nameBS;
    RPC2_Integer vnum;
    struct ofid comp_ofid;
    struct viceroy_onode *comp_onode;
    struct qtw_data *qtwdp = wp->w_data;
    struct qtt_data *qttdp;
    tome_t tome = odp->o_fid.ofid_tome;
    cmgr_connection_t *connp;
    int result = 0;


    ASSERT((tome >= 0) && (tome < QT_MAXTOMES));
    qttdp = qtwdp->qtw_tome[tome];
    ASSERT(qttdp);

    /* make sure we were passed the root onode */

   ASSERT(odp);
    ASSERT((odp->o_flags) | ON_ROOTOFVOL);
    ASSERT(odp->o_fid.ofid_ttype == QT_TTYPE);

    nameBS.SeqLen = (RPC2_Integer) strlen(name)+1;
    nameBS.SeqBody = (RPC2_ByteSeq) name;

    *remain = name;
    *opp = NULL;

    if (cmgr_validate(qttdp->qtd_cid)) {
	connp = cmgr_get(qttdp->qtd_cid, CMGR_LOCK);
	ASSERT(connp);
	result = QT_Lookup_RPC(CMGR_RPC2_HANDLE(connp),&nameBS,&vnum);
	cmgr_put(qttdp->qtd_cid, CMGR_LOCK);
    }
    else
	result = ETIMEDOUT; /* we assume the RPC would've failed,
					   since the conn is marked down */

    if (result != RPC2_SUCCESS) {
	if (result < RPC2_FLIMIT) {
	    cmgr_markdown(qttdp->qtd_cid);
	    ds_thread_log(10,"QT_Lookup_RPC(%s): Connection %d down",
			  name, qttdp->qtd_cid);
	    result = ETIMEDOUT;
	}
	goto bad;
    }

    comp_ofid.ofid_ttype = QT_TTYPE; comp_ofid.ofid_tome = tome;
    comp_ofid.ofid_vnum = vnum; comp_ofid.ofid_uniq = QT_MOVIE_UNIQ;
	    
    comp_onode = onode_get(&comp_ofid);

    if (comp_onode == NULL) { /* haven't looked up this movie before */
	struct onode_create_in comp_onode_params;
	struct onode_data *o_data;

	ALLOC(o_data, struct onode_data);

	o_data->num_tracks = -1;
	o_data->all_hdrs = NULL;
	o_data->already_opened = 0;
	o_data->prefetch = NULL;
	o_data->od_cid = CMGR_CID_NONE;
	o_data->thumbnail = NULL;
	o_data->thumbnail_fd = -1;
	o_data->thumbnail_frame_list = NULL;
	o_data->thumbnail_track = 0;

	comp_onode_params.oc_flags = ON_UNCACHEABLE;
	comp_onode_params.oc_vtype = O_VREG;
	comp_onode_params.oc_ofree = qt_ofree;
	comp_onode_params.oc_data = o_data;

	comp_onode = onode_create(&comp_ofid, &comp_onode_params);
	ASSERT(comp_onode);

	onode_writelock(comp_onode);
	comp_onode->o_cachefile = NULL; /* not cacheable */
	onode_release_write(comp_onode);

	comp_onode = onode_get(&comp_ofid);

	ASSERT(comp_onode);
    }

    /* we have a referenced onode, readlock it */
    onode_readlock(comp_onode);
    *opp = comp_onode;

    /* release the lock and reference on odp that was handed in to us */
    onode_release_read(odp);
    onode_put(odp);

    return 0;

 bad:
    ASSERT(result);
    onode_release_read(odp);
    onode_put(odp);
    ds_thread_log(0,"qt_lookup: error %d\n", result);
    return result;
}


/* Does user 'cred' have 'mode' permissions on object 'op' */

static
int
qt_access(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    /* VREAD/VEXEC for all, VWRITE for none */

    switch (mode) {
    case O_A_READ:
    case O_A_EXEC:
	return 0;
    case O_A_WRITE:
	return EACCES;
    }
    return EINVAL;
}

/* Get attributes of an object */
/* NOTE: until readdir is implemented, only the top-level QuickTime mount
   point can be getattr'd */

static
int
qt_getattr(wp, op, vap, cred)
     struct warden *wp;
     struct viceroy_onode  *op;
     struct ody_vattr  *vap;
     struct ody_cred  *cred;
{
    struct qtw_data *qtwdp = wp->w_data;
    struct qtt_data *qttdp;
    tome_t tome = op->o_fid.ofid_tome;
    time_t temp;

    ASSERT((tome >= 0) && (tome < QT_MAXTOMES));
    qttdp = qtwdp->qtw_tome[tome];
    ASSERT(qttdp);

    
    ASSERT((op->o_flags) | ON_ROOTOFVOL);
    ASSERT(op->o_fid.ofid_ttype == QT_TTYPE);

    temp = 0;

    ds_thread_log(0,"qt_getattr on fid (%x.%x.%x.%x)",
		  op->o_fid.ofid_ttype,op->o_fid.ofid_tome,
		  op->o_fid.ofid_vnum,op->o_fid.ofid_uniq);

    if (op->o_flags | ON_ROOTOFVOL)
      vap->va_type = O_VDIR;
    else
      vap->va_type = O_VREG;

    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;		/* XXX should have better value! */
    vap->va_blocksize = 512;
    vap->va_atime = temp;	/* XXX again, better values exist */
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    return(0);
}

/* Open an onode in the specified mode */

static
int
qt_open(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    int result = 0;
    struct onode_data *o_data;

    /* Currently no writes are allowed */
    if (mode & O_O_WRITE) {
	result = EACCES;
	goto bad;
    }

    if (op->o_fid.ofid_uniq == QT_ROOT_UNIQ) /* root onode */
	return 0;

    o_data = op->o_data;
    ASSERT(o_data);

    /* currently we allow only one opener at at time */
    if (o_data->already_opened)
	return EPERM;

    o_data->already_opened = 1;

    if (!(op->o_flags & ON_UNCACHEABLE)) {
	ds_thread_log(0,"qt_open: cacheable onode (%x.%x.%x.%x), aborting",
		      op->o_fid.ofid_ttype, op->o_fid.ofid_tome,
		      op->o_fid.ofid_vnum, op->o_fid.ofid_uniq);
	ASSERT(0);
    }

    /* there's nothing that the warden needs to do for an open(), the real
       stuff happens in the tsops such as openmovie */
    return 0;

 bad:
    ASSERT(result);
    return result;
}


/* Close fd associated with movie. Clean up all the data structures,
   cache file, etc, associated with it.
   Later we may want to cache this info */

static
int
qt_close(wp, op, cred, mode)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int mode;
{
    struct qtw_data *qtwdp = wp->w_data;
    struct qtt_data *qttdp;
    struct onode_data *o_data;
    tome_t tome = op->o_fid.ofid_tome;
    RPC2_Integer vnum;
#ifdef CLOSE_ON_SERVER /* only used by currently disabled code */
    int result;
#endif /* CLOSE_ON_SERVER */
    
    ASSERT((tome >= 0) && (tome < QT_MAXTOMES));
    qttdp = qtwdp->qtw_tome[tome];
    ASSERT(qttdp);

    if (op->o_fid.ofid_uniq == QT_ROOT_UNIQ) /* root onode */
	return 0;

    o_data = op->o_data;
    ASSERT(o_data);
    ASSERT(o_data->already_opened);

    vnum = op->o_fid.ofid_vnum;
    
    if (o_data->od_cid != CMGR_CID_NONE)
	cmgr_destroy(o_data->od_cid);
    o_data->od_cid = CMGR_CID_NONE;

    if (o_data->prefetch) {
	struct qtw_prefetch *qtwp = o_data->prefetch;
	struct qtwp_frame *qtwfp;
	int i;
	struct qtwp_frame;

	ds_thread_log(100,"last frame fetched zztime %d\n",
		      qtwp->last_fetched_zztime);
	/* shut down prefetching */
	qtwp_stop(qtwp);

	/* if there are some tsop-returned buffers, clean them up */

	ds_hwlock_write(&qtwp->qtwp_lock);
	ASSERT(ds_list_valid(qtwp->retd_frames));

	for (i = ds_list_count(qtwp->retd_frames); i; i--) {
	    qtwfp = ds_list_get_first(qtwp->retd_frames);
	    ds_thread_log(0,"qt_close: had to dispose %x",qtwfp->retbuf);
	    ASSERT(qtwfp->retbuf);
	    FREE(qtwfp->retbuf);
	    QTWF_FREE(qtwp, qtwfp);
	}
	ds_list_destroy(qtwp->retd_frames);
	qtwp->retd_frames = NULL;
	ASSERT(qtwp->cur_bufsize == 0);
	ds_hwlock_release_write(&qtwp->qtwp_lock);
    }
	       
    o_data->prefetch = NULL;
    
    o_data->already_opened = 0;

    /* XXX Can't CloseMovie on server, because we cache meta-data and don't
       do an OpenMovie RPC next time someone opens the movie. Needs fixed */
#ifdef CLOSE_ON_SERVER
    if (cmgr_validate(qttdp->qtd_cid)) {
	cmgr_connection_t *connp = cmgr_get(qttdp->qtd_cid, CMGR_LOCK);
	ASSERT(connp);
	result = QT_CloseMovie_RPC(CMGR_RPC2_HANDLE(connp), vnum);
	cmgr_put(qttdp->qtd_cid, CMGR_LOCK);
    }
    else
	result = ETIMEDOUT;
    
    if (result != RPC2_SUCCESS) {
	if (result  < RPC2_FLIMIT) {
	    cmgr_markdown(qttdp->qtd_cid);
	    ds_thread_log(10,"QT_Close_RPC(%d): Connection %d down",
			  vnum, qttdp->qtd_cid);
	    result = ETIMEDOUT;
	}
	goto bad;
    }
#endif /* CLOSE_ON_SERVER */
	
    return 0;

    /* The commmented-out code above is the only thing that goto's here */
#ifdef CLOSE_ON_SERVER
 bad:
    ASSERT(result);
    return result;
#endif /* CLOSE_ON_SERVER */
}

/* function that does reads/writes. These are not meaningful for movies:
   only tsops are */

static
int
qt_rdwr(wp, op, cred, rwflag, count, offset, ioflag, data)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int rwflag;
    int *count;
    int offset;
    int ioflag;
    void **data;
{
    *count = 0;
    *data = NULL;
    return EOPNOTSUPP;
}

/* function to dispose buffers returned by qt_rdwr. Since qt_rdwr never
   returns any data, this should never be called
*/

static
int
qt_rdwr_dispose(wp, op, cred, count, offset, data)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int count;
    int offset;
    void *data;
{
    ds_thread_log(0,"qt_rdwr_dispose: called on bogus buffer %x, size %d",
		  data,count);
    ASSERT(0);

    /* NOTREACHED */
    return ELAST; /* bogus error code */
}

/* functions to do the type-specific operations */

/* Open a movie on all available tracks */
static int         qt_openmovie  __P((struct warden *wp,
				      struct viceroy_onode *op,
				      size_t argsz,
				      void *argp,
				      size_t *retsz,
				      void **retp));

/* Fetch a frame */
static int         qt_getframe  __P((struct warden *wp,
				     struct viceroy_onode *op,
				     size_t argsz,
				     void *argp,
				     size_t *retsz,
				     void **retp));

/* Do a type-specific operation */
static
int
qt_tsop(wp, tsvnum, op, cred, opnum, argsz, argp, retsz, retp)
    struct warden *wp;
    int tsvnum;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int opnum;
    size_t argsz;
    void *argp;
    size_t *retsz;
    void **retp;
{
  
    if (tsvnum != QT_VERSION)
	return EINVAL;

    switch(opnum) {

    case QT_OPENMOVIE: return qt_openmovie(wp,op,argsz,argp,retsz,retp);

    case QT_GETFRAME: return qt_getframe(wp,op,argsz,argp,retsz,retp);

    default: return EINVAL;

    }
}

/* dispose buffers allocated by tsops */
/* if necessary, this can be customised for each tsop */

static
int
qt_tsop_dispose(wp, op, cred, opnum, argsz, argp, retsz, retp)
    struct warden *wp;
    struct viceroy_onode *op;
    struct ody_cred *cred;
    int opnum;
    size_t argsz;
    void *argp;
    size_t retsz;
    void *retp;
{
    switch (opnum) {

    case QT_OPENMOVIE: /* buffers allocated by openmovie stay with onode */
	break;
	
    case QT_GETFRAME: {
	struct onode_data *o_data;
	struct qtw_prefetch *qtwp;
	struct qtwp_frame *qtwfp;
	int i;

	ASSERT(op);
	ASSERT(o_data = op->o_data);
	ASSERT(qtwp = o_data->prefetch);

	ds_hwlock_write(&qtwp->qtwp_lock);

	/* find the entry in qtwp->retd_frames that corresponds to this
	   dispose */

	ASSERT(ds_list_valid(qtwp->retd_frames));

	for (i = ds_list_count(qtwp->retd_frames), qtwfp=NULL; i; i--) {
	    qtwfp = ds_list_get_first(qtwp->retd_frames);
	    if (qtwfp->retbuf == retp) break; /* the one we want to dispose */
	    ds_list_insert(qtwp->retd_frames, qtwfp); /* else, put it back */
	    ds_thread_log(0,"qt_tsop_dispose(%x): %x not yet disposed",
			  retp, qtwfp->retbuf);
	}
	 	
	ds_hwlock_release_write(&qtwp->qtwp_lock);

	if (!qtwfp) {
	    ds_thread_log(0,"qt_tsop_dispose(%x): unknown buffer address");
	    return EINVAL;
	}


	/* TMP: free the buffer that we allocated specially for the tsop */
	/* this'll eventually become a uio that has pointers to the header and
	   into the prefetch buffer */

	ASSERT(qtwfp->retbuf == retp);
	FREE(qtwfp->retbuf);
	
	ASSERT(qtwfp->hdr.size + sizeof(struct qtw_frame_info) >= retsz);
	QTWF_FREE(qtwp, qtwfp);
    }
	break;

    default:
	ASSERT(0); /* we shouldn't have any other tsop numbers */
    }
    return 0;
}


/* return the ID of our connection to server. Used by viceroy for bandwidth
   estimation */

static
cmgr_id_t
qt_getcid(wp, op)
    struct warden *wp;
    struct viceroy_onode *op;
{
    struct onode_data *odp = op->o_data;
    if (!odp || !odp->already_opened)
	return CMGR_CID_NONE;
    return odp->od_cid;
}

#ifdef INSTRUMENTED
/* dump statistics that we've collected onto the give file pointer */
static
void
qt_pstats(wp, fp)
    struct warden *wp;
    FILE *fp;
{
}
    
#endif /* INSTRUMENTED */

/* the actual TSOPs */

#define MAX_TRACKS 20

/* Open a movie on all available tracks */
/* No input */
/* Output: *retp points to an array of struct qtw_track_hdr
 *retsz contains the size of this array in bytes
 */

static
int
qt_openmovie(wp, op, argsz, argp, retsz, retp)
    struct warden *wp;
    struct viceroy_onode *op;
    size_t argsz;
    void *argp;
    size_t *retsz;
    void **retp;
{
    struct onode_data *o_data = op->o_data;
    int retsize = *retsz;
    int result;
    QT_TrackHdr *all_hdrs = NULL;
    RPC2_Integer num_tracks, thumbnail_track;
    struct qtw_data *qtwdp = wp->w_data;
    tome_t tome = op->o_fid.ofid_tome;
    int i;
    cmgr_initparams_t ci;
    int rc;

    ASSERT((tome >= 0) && (tome < QT_MAXTOMES));
    *retp = NULL; *retsz = 0;

    /* open a new connection to the server, for frame fetches for this
       onode */
    bzero(&ci, sizeof(cmgr_initparams_t));
    ci.ci_type = CMGR_RPC2;
    ci.ci_restart.tv_sec =  0;   /* restart as soon as we try to validate */
    ci.ci_restart.tv_usec = 0;

    ci.ci_params.ci_rpc2.hid.Tag = RPC2_HOSTBYNAME;
    strcpy(ci.ci_params.ci_rpc2.hid.Value.Name, server_hostname[tome]);
    ci.ci_params.ci_rpc2.pid.Tag = RPC2_PORTBYINETNUMBER;
    ci.ci_params.ci_rpc2.pid.Value.InetPortNumber = htons(QTPORTAL);
    ci.ci_params.ci_rpc2.sid.Tag = RPC2_SUBSYSBYID;
    ci.ci_params.ci_rpc2.sid.Value.SubsysId = QTSUBSYSID;

    ci.ci_params.ci_rpc2.bp.SecurityLevel = RPC2_OPENKIMONO;
    ci.ci_params.ci_rpc2.bp.EncryptionType = 0;
    ci.ci_params.ci_rpc2.bp.SideEffectType = SMARTFTP;
    ci.ci_params.ci_rpc2.bp.ClientIdent = NULL;
    ci.ci_params.ci_rpc2.bp.SharedSecret = NULL;

    rc = cmgr_create(&ci);

    if (rc < 0) {
	result = ECONNREFUSED;
	ds_thread_log(0,"qt_openmovie: Cannot connect to server on %s",
		      server_hostname);
	goto bad;
    }

    /* we got the connection, squirrel away the connection id */
    o_data->od_cid = rc;
    ds_thread_log(0,"qt_openmovie: created connection %d to server",
		  o_data->od_cid);
    
    if (o_data->all_hdrs == NULL) { /* haven't yet opened this movie */
	RPC2_BoundedBS all_hdrs_BS;
	cmgr_connection_t *connp;
	RPC2_Integer vnum;

	vnum = op->o_fid.ofid_vnum;

	num_tracks = MAX_TRACKS;
	NALLOC(all_hdrs, QT_TrackHdr, num_tracks);
	
	all_hdrs_BS.MaxSeqLen = (RPC2_Integer) num_tracks*sizeof(QT_TrackHdr);
	all_hdrs_BS.SeqLen = (RPC2_Integer) 0;
	all_hdrs_BS.SeqBody = (RPC2_ByteSeq) all_hdrs;
	
	if (cmgr_validate(o_data->od_cid)) {
	    connp = cmgr_get(o_data->od_cid, CMGR_LOCK);
	    ASSERT(connp);
	    result = QT_OpenMovie_RPC(CMGR_RPC2_HANDLE(connp),
				      op->o_fid.ofid_vnum, &num_tracks,
				      &thumbnail_track, &all_hdrs_BS);
	    cmgr_put(o_data->od_cid, CMGR_LOCK);
	}
	else
	    result = ETIMEDOUT;

	if (result != RPC2_SUCCESS) {
	    if (result < RPC2_FLIMIT) {
		cmgr_markdown(o_data->od_cid);
		ds_thread_log(10,"QT_OpenMovie_RPC(%d): Connection %d down",
			      op->o_fid.ofid_vnum, o_data->od_cid);
		result = ETIMEDOUT;
	    }
	    FREE(all_hdrs);
	    goto bad;
	}

	o_data->thumbnail_track = thumbnail_track;
	o_data->num_tracks = num_tracks;
	NALLOC(o_data->all_hdrs, struct qtw_track_hdr, num_tracks);

	for (i = 0; i < num_tracks; i++) {
	    /* copy anim_hdr info */
	    memcpy(o_data->all_hdrs[i].codec_name, all_hdrs[i].codec_name,
	       sizeof(o_data->all_hdrs[i].codec_name));
	    o_data->all_hdrs[i].compression_type =
		ntohl(all_hdrs[i].compression_type);
	    o_data->all_hdrs[i].depth = ntohl(all_hdrs[i].depth);
	    o_data->all_hdrs[i].ave_fvid_size =
		ntohl(all_hdrs[i].ave_fvid_size);
	    o_data->all_hdrs[i].max_fvid_size =
		ntohl(all_hdrs[i].max_fvid_size);
	    o_data->all_hdrs[i].anim_type = ntohl(all_hdrs[i].anim_type);
	    o_data->all_hdrs[i].imagex = ntohl(all_hdrs[i].imagex);
	    o_data->all_hdrs[i].imagey = ntohl(all_hdrs[i].imagey);
	    o_data->all_hdrs[i].anim_flags = ntohl(all_hdrs[i].anim_flags);
	    o_data->all_hdrs[i].total_time = ntohl(all_hdrs[i].total_time);
	    o_data->all_hdrs[i].last_frame = ntohl(all_hdrs[i].last_frame);
	    o_data->all_hdrs[i].act_type = ntohl(all_hdrs[i].act_type);
	    o_data->all_hdrs[i].cmap_rev = ntohl(all_hdrs[i].cmap_rev);
	}

	FREE(all_hdrs);
    }

    if (retsize > o_data->num_tracks * sizeof(struct qtw_track_hdr))
	retsize = o_data->num_tracks * sizeof(struct qtw_track_hdr);
    
    *retsz = retsize;
    *retp = o_data->all_hdrs;

    /* Handle the thumbnail track */
    if (o_data->thumbnail_track) {
	    int tnail = 0; /* the thumbnail track must be the first one */
	/* fetch the frame list if we don't have it */
	    if (o_data->thumbnail_frame_list == NULL) {
		if (cmgr_validate(o_data->od_cid)) {
		    cmgr_connection_t *connp;
		    SE_Descriptor thumbnail_sftp;
		    int nframes = o_data->all_hdrs[tnail].last_frame+1;
		    
		    NALLOC(o_data->thumbnail_frame_list, QT_FrameInfo,
			   nframes);
		    
		    bzero(&thumbnail_sftp,sizeof(thumbnail_sftp));
		    thumbnail_sftp.Tag = SMARTFTP;
		    thumbnail_sftp.Value.SmartFTPD.Tag = FILEINVM;
		    thumbnail_sftp.Value.SmartFTPD.TransmissionDirection =
			SERVERTOCLIENT;
		    thumbnail_sftp.Value.SmartFTPD.ByteQuota =
			nframes * sizeof(QT_FrameInfo);
		    thumbnail_sftp.Value.SmartFTPD.SeekOffset = 0;
		    thumbnail_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.
			MaxSeqLen = nframes * sizeof(QT_FrameInfo);
		    thumbnail_sftp.Value.SmartFTPD.FileInfo.ByAddr.vmfile.
			SeqBody = (RPC2_ByteSeq) o_data->thumbnail_frame_list;

		    connp = cmgr_get(o_data->od_cid, CMGR_LOCK);
		    ASSERT(connp);
		    
		    /* fetch the thumbnail track */
		    result = QT_OpenTrack_RPC(CMGR_RPC2_HANDLE(connp),
					      op->o_fid.ofid_vnum, tnail,
					      &thumbnail_sftp);
		    cmgr_put(o_data->od_cid, CMGR_LOCK);
		}
		else
		    result = ETIMEDOUT;
		
		if (result != RPC2_SUCCESS) {
		    if (result < RPC2_FLIMIT) {
			cmgr_markdown(o_data->od_cid);
			ds_thread_log(0,"QT_FetchTrack_RPC(%d): "
				      "Connection %d down",
				      op->o_fid.ofid_vnum, o_data->od_cid);
			result = ETIMEDOUT;
		    }
		    ds_thread_log(0,"qt_open: Could not fetch "
				  "thumbnail frame list (%s)",
				  strerror(result));
		    if (o_data->thumbnail_frame_list) {
			FREE(o_data->thumbnail_frame_list);
			o_data->thumbnail_frame_list = NULL;
		    }
		}
		else {
		    for (i = 0; i <= o_data->all_hdrs[tnail].last_frame; i++) {
			QT_FrameInfo *frame_info =
			    &(o_data->thumbnail_frame_list[i]);
			frame_info->track_ID = ntohl(frame_info->track_ID);
			frame_info->frame_no = ntohl(frame_info->frame_no);
			frame_info->time_dur = ntohl(frame_info->time_dur);
			frame_info->zztime = ntohl(frame_info->zztime);
			frame_info->xapi_rev = ntohl(frame_info->xapi_rev);
			frame_info->flags = ntohl(frame_info->flags);
			frame_info->xpos = ntohl(frame_info->xpos);
			frame_info->ypos = ntohl(frame_info->ypos);
			frame_info->xsize = ntohl(frame_info->xsize);
			frame_info->ysize = ntohl(frame_info->ysize);
			frame_info->special = ntohl(frame_info->special);
			frame_info->bufpos = ntohl(frame_info->bufpos);
			frame_info->size = ntohl(frame_info->size);
		    }
		}
	    }
	    
	    if (o_data->thumbnail == NULL) {
		int fd = -1;
		
		o_data->thumbnail = qtw_cache_create(qtwdp->qtw_cache_files,
						     qtwdp->qtw_ncache_files,
						     op);
		if (!o_data->thumbnail) {
		    ds_thread_log(0, "qt_open: No cache entries left\n");
		    result = ENOSPC;
		}
		
		/* try to SFTP the thumbnail file */
		else if (cmgr_validate(o_data->od_cid)) {
		    SE_Descriptor thumbnail_sftp;
		    cmgr_connection_t *connp;
	    
		    fd = qtw_cache_open(o_data->thumbnail,
					O_WRONLY|O_CREAT|O_TRUNC);
		    ASSERT(fd >= 0);
	    
		    bzero(&thumbnail_sftp,sizeof(thumbnail_sftp));
		    thumbnail_sftp.Tag = SMARTFTP;
		    thumbnail_sftp.Value.SmartFTPD.Tag = FILEBYFD;
		    thumbnail_sftp.Value.SmartFTPD.TransmissionDirection =
			SERVERTOCLIENT;
		    thumbnail_sftp.Value.SmartFTPD.ByteQuota = -1;
		    thumbnail_sftp.Value.SmartFTPD.SeekOffset = 0;
		    thumbnail_sftp.Value.SmartFTPD.FileInfo.ByFD.fd = fd;
		    
		    connp = cmgr_get(o_data->od_cid, CMGR_LOCK);
		    ASSERT(connp);
		    
		    /* fetch track 0 -- assume that's the thumbnail track */
		    result = QT_FetchTrack_RPC(CMGR_RPC2_HANDLE(connp),
					       op->o_fid.ofid_vnum, tnail,
					       &thumbnail_sftp);
		    cmgr_put(o_data->od_cid, CMGR_LOCK);
	    
		}
		else
		    result = ETIMEDOUT;
		
		if (result != RPC2_SUCCESS) {
		    if (result < RPC2_FLIMIT) {
			cmgr_markdown(o_data->od_cid);
			ds_thread_log(0,"QT_FetchTrack_RPC(%d): Conn %d down",
				      op->o_fid.ofid_vnum, o_data->od_cid);
			result = ETIMEDOUT;
		    }
		    if (o_data->thumbnail) {
			/* we don't care about the close() retval, since we're
			   going to release it anyway */
			qtw_cache_close(o_data->thumbnail, fd, 1);
			qtw_cache_release(o_data->thumbnail);
			o_data->thumbnail = NULL;
		    }
		}
		
		/* close the cache file, make sure we have enough space */
		if (o_data->thumbnail) {
		    result = qtw_cache_close(o_data->thumbnail, fd, 1);
		    if (result) {
			ASSERT(result == ENOSPC);
			ds_thread_log(0,"qt_open: No space for thumbnail\n");
			qtw_cache_release(o_data->thumbnail);
			o_data->thumbnail = NULL;
		    }
		    o_data->thumbnail_fd = qtw_cache_open(o_data->thumbnail,
							  O_RDONLY);
		    ASSERT(o_data->thumbnail_fd >= 0);
		}
	    }
    }

    /* set up prefetching */
    
    ASSERT(qtwdp->qtw_pf_avail >= 0);

    if (qtwdp->qtw_pf_avail == 0) {
	ds_thread_log(0,"No prefetch threads available!!");
	result = ENFILE;
	goto bad;
    }

    qtwp_start(wp, op, prefetch_bufsize, fetch_maxtime, fetch_mintime,
	       fetch_maxplaytime);
    return 0;
    
 bad:
    ASSERT(result);
    return result;
}


/* fetch a frame from the cached thumbnail */
static
int
qtw_fetch_from_cache(QT_FrameInfo *frame_list,
		     int fd,
		     struct qtw_track_hdr *hdr,
		     int zztime,
		     struct qtwp_frame *qtwfp)
{

    int fstart, fsize, frame_no;
    int last_frame = hdr->last_frame;
    int total_time = hdr->total_time;
    struct qtwp_fbuf *qtwfbp = NULL;
    int result;

    /* compute the location of the required frame */
    frame_no = (long) last_frame * zztime / total_time;

    ASSERT(frame_no >= 0);
    ASSERT(frame_no <= hdr->last_frame);
    
    /* if it's behind, push it forward */
    while ((frame_no < last_frame)&&(frame_list[frame_no].zztime < zztime))
	frame_no++;
    
    /* if it's ahead, push it back */
    while ((frame_no > 0) && (frame_list[frame_no-1].zztime >= zztime))
	frame_no--;
    
    fstart = frame_list[frame_no].bufpos;
    fsize = frame_list[frame_no].size;

    QTWFB_ALLOC(qtwfbp, 1, fsize);
    
    result = lseek(fd, fstart, SEEK_SET);
    if (result< 0) {
	ds_thread_log(0, "qtw_fetch_from_cache: lseek(%d) returned %d\n",
		      fstart, errno);
	result = errno;
	goto bad;
    }

    result = read(fd, QTWFB_BUF(qtwfbp, 0), fsize);

    if (result < 0) {
	ds_thread_log(0, "qtw_fetch_from_cache: read(%d) returned %d\n",
		      fsize, errno);
	result = errno;
	goto bad;
    }

    /* Create a qtwfp structure */

    qtwfp->retbuf = NULL;
    qtwfp->buffer = qtwfbp;
    qtwfp->bufpos = 0;
    qtwfp->framesize = fsize;

    qtwfp->hdr.size = fsize;

    qtwfp->hdr.trackID = frame_list[frame_no].track_ID;
    qtwfp->hdr.frame_no = frame_list[frame_no].frame_no;
    qtwfp->hdr.time_dur = frame_list[frame_no].time_dur;
    qtwfp->hdr.zztime = frame_list[frame_no].zztime;
    qtwfp->hdr.xapi_rev = frame_list[frame_no].xapi_rev;
    qtwfp->hdr.flags = frame_list[frame_no].flags;
    qtwfp->hdr.xpos = frame_list[frame_no].xpos;
    qtwfp->hdr.ypos = frame_list[frame_no].ypos;
    qtwfp->hdr.xsize = frame_list[frame_no].xsize;
    qtwfp->hdr.ysize = frame_list[frame_no].ysize;
    qtwfp->hdr.special = frame_list[frame_no].special;

    return 0;

 bad:
    if (qtwfbp) QTWFB_FREE_ALL(qtwfbp);
    ASSERT(result);
    return result;
}

/* Fetch a frame from the prefetch queue or the cached thumbnail */

static
int 
qt_fetchframe(wp, op, track_ID, zztime, retsz, retp)
    struct warden *wp;
    struct viceroy_onode *op;
    int track_ID;
    int zztime;
    size_t *retsz;
    void **retp;
{
    struct onode_data *o_data = op->o_data;
    struct qtw_prefetch *qtwp = NULL;
    struct qtwp_frame *qtwfp = NULL;
    struct timeval tv_start;
    struct qtw_frame_info *retbuf = NULL;/* TMP this should become a struct uio */
    int result = -1;

    gettimeofday(&tv_start, NULL);

    ASSERT(o_data);

    *retp = NULL;

    qtwp = o_data->prefetch;

    /* We hold the lock until we exit this function, except when we block
       on an empty prefetch queue */

    ds_hwlock_write(&qtwp->qtwp_lock);

    /* update info about the app's time stream, so prefetch thread can
     synchronize */
    
    qtwp->last_req_zztime = zztime;
    qtwp->last_req_time = tv_start;

    /* in case the application has switched tracks, update that info */
    qtwp->track_ID = track_ID;

    /* If we have a local cached copy and there's no need to prefetch,
       indicate that */

    qtwp->is_cached = (o_data->thumbnail_track) && (o_data->thumbnail_fd >= 0)
	&& (o_data->thumbnail_frame_list != NULL) && (track_ID == 0);

    ASSERT(o_data->already_opened);


#ifndef BPROF
    /* JNF - This seems uncecessary since prefetch thread can only make
       progress after more space has been freed */
    /* wake prefetch thread up, in case it's waiting */
    ds_thread_wakeup((char *)qtwp);
#endif

    while(1) {

	/* If there's nothing in the prefetch buffer, and we're
	   playing the thumbnail, then we get it from the cached copy */
	if (qtwp->is_cached && (ds_safeq_count(qtwp->full_buffers) == 0)) {
	    ALLOC(qtwfp, struct qtwp_frame);
	    
	    result = qtw_fetch_from_cache(o_data->thumbnail_frame_list,
					  o_data->thumbnail_fd,
					  &o_data->all_hdrs[0],
					  zztime, qtwfp);
	    
	    ASSERT(result == 0); /* what could go wrong? */
	    
	    /* Do the bookkeeping */
	    ASSERT(qtwp->cur_bufsize == 0); /* queue was empty */
	    ASSERT(qtwp->max_bufsize >= qtwfp->framesize); 
	    qtwp->cur_bufsize += qtwfp->framesize;
	    
	    qtwp->last_fetched_track_ID = track_ID;
	    qtwp->last_fetched_zztime = qtwfp->hdr.zztime;
	    break; /* we're done */
	}
	    
	/* Otherwise we need to wait for something from the prefetch queue */
	ds_hwlock_release_write(&qtwp->qtwp_lock);
	qtwfp = ds_safeq_deq(qtwp->full_buffers); /* block */
	ds_hwlock_write(&qtwp->qtwp_lock);

	if (qtwfp->buffer == NULL) { /* error during prefetch */

	    /* return EINVAL if EOM token found */
  	    if (qtwfp->hdr.frame_no == -2) {
	      FREE (qtwfp);
	      ds_hwlock_release_write(&qtwp->qtwp_lock);
	      return EINVAL; 
	    }

	    ds_thread_log(10,"discarded badly prefetched frame");
	    FREE(qtwfp);
	    if (qtwp->is_cached)
		continue;       /* we'll just get it from the cache */
	    else { 
		ds_hwlock_release_write(&qtwp->qtwp_lock);
		return ETIMEDOUT;
	    }
	}
	    
	/* If we have a cached copy, then we definitely want to drop
	   late frames; the reason is that otherwise the late frame might
	   precede something we previously gave the app from thec cache,
	   thus violating time ordering in the movei.
	   */	   

	if (qtwp->is_cached && (qtwfp->hdr.zztime < zztime)) {
	    ds_thread_log(10,"qt_fetchframe: late prefetched frame, fetching"
			  " from cache");
	    QTWF_FREE(qtwp, qtwfp);
	    continue;
	}
	
	/* if it's a late frame, we might have to drop it */
	if (drop_late_frames && (qtwfp->hdr.zztime < zztime)) {
	    ds_thread_log(10, "qt_fetchframe: dropping late frame time "
			  "%d (req %d)", qtwfp->hdr.zztime, zztime);
	    QTWF_FREE(qtwp, qtwfp);
	    continue;
	}

	/* we got a frame for the app */
	break;
    }


    /* When we leave this loop, we STILL have the qtwp_lock */

    if (*retsz < qtwfp->hdr.size + sizeof(struct qtw_frame_info))
	ds_thread_log(0,
		      "qt_fetchframe: frame+header %d, app buffer only %d",
		      qtwfp->hdr.size + sizeof(struct qtw_frame_info),
		      *retsz);
    else
	*retsz = qtwfp->hdr.size + sizeof(struct qtw_frame_info);
    
    BALLOC(retbuf, *retsz);
    *retbuf = qtwfp->hdr;
    
    memcpy(&(retbuf[1]), /* skip over the header */
	   QTWFB_BUF(qtwfp->buffer, qtwfp->bufpos),
	   *retsz - sizeof(struct qtw_frame_info));
    
    qtwfp->retbuf = *retp = retbuf;
    
    /* store this frame away for tsop_dispose to dispose of */
    
    ASSERT(ds_list_valid(qtwp->retd_frames));
    ds_list_insert(qtwp->retd_frames, qtwfp);

    ds_hwlock_release_write(&qtwp->qtwp_lock);
    
    return 0;
}

/* Get a frame from a specific track given the time offset */

static
int
qt_getframe(wp, op, argsz, argp, retsz, retp)
    struct warden *wp;
    struct viceroy_onode *op;
    size_t argsz;
    void *argp;
    size_t *retsz;
    void **retp;
{
    struct qtw_frame_request *frame_spec = (struct qtw_frame_request *)argp;
    int track_ID, zztime;
    struct onode_data *o_data = op->o_data;
    int total_time;

    track_ID = frame_spec->track_ID;
    zztime = frame_spec->zztime;

    if (o_data->all_hdrs == NULL)
	return EINVAL;

    if ((track_ID < 0) || (track_ID >= o_data->num_tracks))
	return EINVAL;

    if (argsz != sizeof(struct qtw_frame_request)) 
	return EINVAL;

    total_time = o_data->all_hdrs[track_ID].total_time + 1;

    if (zztime > total_time) {
      ds_thread_log(100,"qt_getframebytime: zztime %d > last frame %d",
                     zztime, total_time);
	return EINVAL;
    }

    /* get the frame */
    return qt_fetchframe(wp, op, track_ID, zztime, retsz, retp);
}

