#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * qtw_prefetch.h
 *
 * prefetching/buffering in QuickTime warden
 */

/* Structure that holds multiple frames in a memory buffer, for efficiency.
   The number of frames is initialized on alloc, and decremented each time
   we free a frame. When it goes to zero, the buffer is freed */

struct qtwp_fbuf {
    int num_frames;
};

#define QTWFB_BUF(qtwfbp, bufpos) ((void *)(qtwfbp) + sizeof(qtwfbp->num_frames) + bufpos)

#define QTWFB_ALLOC(qtwfbp, nframes, size) do { \
    BALLOC((qtwfbp), size + sizeof(int)); \
    (qtwfbp)->num_frames = nframes; \
} while(0)

#define QTWFB_FREE(qtwfbp) do { \
    if ((qtwfbp)->num_frames-- == 1) FREE(qtwfbp); \
} while(0)

#define QTWFB_FREE_ALL(qtwfbp) FREE(qtwfbp)


/* structure that represents a single fetched frame. it is placed in the
   prefetch queue, and points to a location in a fetch buffer where the
   frame is */

struct qtwp_frame {
    struct qtw_frame_info hdr;
    void *retbuf;   /* the header+buffer that was returned to the app
		       TMP: this wil eventually become a uio */
    struct qtwp_fbuf *buffer;   /* buffer which contains this frame */
    int bufpos;     /* where this frame starts in that buffer */
    int framesize;    /* size of the frame */
};

#define MAX_FRAMES 32 /* maximum #frames to fetch on a single RPC */

/* macro to free a qtwp_frame structure, and any associated buffers */
#define QTWF_FREE(qtwp, qtwfp) do { \
    ds_hwlock_write(&qtwp->qtwp_lock); \
    ASSERT(qtwfp); \
    if (qtwfp->buffer) { \
        QTWFB_FREE(qtwfp->buffer); \
	ds_thread_wakeup((char *) qtwp); \
        qtwp->cur_bufsize -= qtwfp->framesize; \
    } \
    FREE(qtwfp); \
    ds_hwlock_release_write(&qtwp->qtwp_lock); \
} while (0)

/* the prefetch structure. each onode will have one of these when it is
   created. It will have empty queues of "full" and "empty" buffers.

   We will also have a worker thread that waits to pull a request out of the
   the "empty" list. The thread then locks the prefetch structure, reads
   the next frame to be prefetched, unlocks the prefetch structure, and
   fetches the frame. It then adds the buffer to the "full" list, bumps
   the "next frame to fetch", and loops.

   If however, the "is_cached" flag is set, it means the current track is
   available from the local cached copy, and the prefetch thread should
   just sleep until the warden signals it.

   When the movie is opened, we set up the prefetch structure, and add
   some number of buffers to the "empty" list. Then, when a frame is
   requested, the warden dequeues from the "full" list until it finds it.
   Anything that's dequeued from the "full" list is added to the "empty"
   list.

   When the movie is closed, the warden sets a flag that tells the prefetch
   thread to stop prefetching, but simply empty out the "empty" list and
   free the buffers. It then destroys the empty list, empties out the "full"
   list and destroys those buffers.
*/

struct qtw_prefetch {
    ds_hwlock_t qtwp_lock;
    struct warden *wp;
    struct viceroy_onode *op;
    int last_fetched_zztime;
    int last_fetched_frame_no;
    int last_fetched_track_ID;
    struct timeval last_req_time;
    int last_req_zztime;
    int track_ID;
    int is_cached;

    int prefetch_enable; /* cleared to tell the prefetch thread to close up */
    int max_bufsize;  /* total number of bytes that we can have ALLOC'ed
			 at any time. this should be at least twice the
			 chunksize, so we can be fetching one chunk while
			 another is waiting to be consumed */
    int cur_bufsize;  /* how many bytes we have ALLOC'ed right now */
    int bw;           /* latest bandwidth estimate */
    int maxchunktime; /* maximum time a transfer should take at current
			 bandwidth estimate, in ms */
    int maxchunksize; /* (maxchunktime/1000) * bw */
    int maxplaytime;  /* maximum "playing time" of a fetch */
    int bw_reqid;     /* request id for callback when bandwidth changes */
    ds_safeq_t *full_buffers;
    ds_list_t *retd_frames; /* list of qtwp_bufs that've been returned and
			       not yet disposed */
    struct qtwp_frame *last_returned; /* the last frame returned by a tsop
				       to the application. used by
				       tsop_dispose so it can free the
				       qtwp_frame and also (if the frame
				       was the last in its buffer) the
				       entire frame buffer */
    /* TMP: this should be a list of things, since we shouldn't assume
       there won't be a second tsop before the first one's dispose */
};

/* function to free a qtwp_frame structure, and any associated buffers */
void qtwf_free(struct qtwp_frame *qtwfp);

/* function that adds a prefetch structure to an onode, and
   starts the prefetch */

void qtwp_start(struct warden *wp, struct viceroy_onode *op, int bufsize,
		int fetch_maxtime, int fetch_mintime, int fetch_maxplaytime);

/* function that stops prefetch and deallocates buffers */
void qtwp_stop(struct qtw_prefetch *qtwp);

/* the function that does the entire prefetch for an open movie */
void qtw_prefetch(struct qtw_prefetch *qtwp);


