#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/viceroy.h"
#include "include/onode-manager.h"

#include "ody_speech.h"
#include "odyjanus.h"
#include "libodyspeech_tsops.h"
#include "speechw_private.h"
#include "speechw_comm.h"
#include "speechw_ping.h"

#include "odyssey/viceroy/rsrc_en.h" /* XXXAPM */

extern int noPing;
extern int speechw_rtimeout;

int speechwConnectToServers (ServerLocations* pLocations, 
			     SpeechConnections* pConnections)
{
  int rc = 0; /* Return code from this function - assume success */
  
  ds_thread_log (10, "speechw: speechwConnectToServers entered");

  /* First try to connect to the local */ 
  if (pLocations->localPort >= 0) {
    pConnections->localConnection = 
      speechwConnectToServer ("localhost", pLocations->localPort, ISLOCAL);
    if (pConnections->localConnection < 0) {
      ds_thread_log (1, "speechw: Unable to connect to local server, rc = %d",
		     pConnections->localConnection);
      return (pConnections->localConnection);
    } /* endif */
      
  } /* endif */

  if (pLocations->remotePort >= 0) {
    pConnections->remoteConnection 
      = speechwConnectToServer (pLocations->remoteHostname, 
				pLocations->remotePort, NOTLOCAL);
    if (pConnections->remoteConnection < 0) {
      ds_thread_log (1, "speechw: Unable to connect to remote server, rc = %d",
		     pConnections->remoteConnection);
      return (pConnections->remoteConnection);
    } /* endif */
  } /* endif */

  ds_thread_log (10, "speechw: speechwConnectToServers completed");

  return (rc);
}

/* speechwConnectToServer() determines whether the connection should be
   made to a local or remote Janus server and makes the connection.
   If the connection is successful, returns the connection id.  If the 
   connection is unsuccessful, returns a negative value */
cmgr_id_t speechwConnectToServer (char* hostname, int port, int islocal)
{
  cmgr_initparams_t cip;
  cmgr_id_t cid;

  ds_thread_log (10, "speechw: speechwConnectToServer entered");

  ds_thread_log (100, "speechw: connecting to hostname: %s, port: %d",
		 hostname, port);

  /* Initialize connection structure */
  bzero (&cip, sizeof(cmgr_initparams_t));

  cip.ci_type    = CMGR_RPC2;

  cip.ci_restart.tv_sec = 90;	/* 1.5 minutes between restarts */
  cip.ci_restart.tv_usec = 0;

  cip.ci_params.ci_rpc2.hid.Tag = RPC2_HOSTBYNAME;
  if (strlen (hostname) >= 64) {
    ds_thread_log(100, "speechw: Machine name %s too long!",
		      hostname);
    return (CMGR_ERROR);
  } /* endif */
  strcpy(cip.ci_params.ci_rpc2.hid.Value.Name, hostname);
    
  cip.ci_params.ci_rpc2.pid.Tag = RPC2_PORTBYINETNUMBER;
  cip.ci_params.ci_rpc2.pid.Value.InetPortNumber = htons(port);
    
  cip.ci_params.ci_rpc2.sid.Tag = RPC2_SUBSYSBYID;
  cip.ci_params.ci_rpc2.sid.Value.SubsysId = SPEECH_SUBSYSID;

  /* Flags */
  if (islocal == ISLOCAL) {
      cip.ci_flags |= CMGR_CONNISLOCAL;
  }
    
  /* Bind Parameters */
  cip.ci_params.ci_rpc2.bp.SecurityLevel = RPC2_OPENKIMONO;
  cip.ci_params.ci_rpc2.bp.EncryptionType = 0;
  cip.ci_params.ci_rpc2.bp.SideEffectType = SMARTFTP;	/* allow use of SFTP */
  cip.ci_params.ci_rpc2.bp.ClientIdent = NULL;
  cip.ci_params.ci_rpc2.bp.SharedSecret = NULL;

  /* Create new binding */
  cid = cmgr_create(&cip);
  if (cid < 0) {
    ds_thread_log (0, "speechw: connection create failed, hostname = %s, "
		   "port = %d", hostname, port);
  } else {
      ds_thread_log (0, "speechw: connection created: %d", cid);
      if (!islocal && speechw_rtimeout) {
	  struct timeval timo;

	  timo.tv_sec = speechw_rtimeout;
	  timo.tv_usec = 0;
	  if (cmgr_settimeout(cid, timo) == 0)
	      ds_thread_log(0,"speechw: timeout on conn %ld set to %d",
			    cid, speechw_rtimeout);
	  else
	      ds_thread_log(0, "speechw: cmgr_settimeout(%ld, %d sec) failed",
			    cid, speechw_rtimeout);
      }
  }

  ds_thread_log(10, "speechw: speechwConnectToServer completed");

  return (cid);
}

/* speechwGetConnection validates and gets a connection with the specified
   id.  If the function succeeds, it returns a pointer to the connection.
   If it fails, it returns 0 */
cmgr_connection_t* speechwGetConnection (cmgr_id_t connectionId)
{
  cmgr_connection_t* pConnection = 0;
  int validity;

  ds_thread_log(10, "speechw: speechwGetConnection entered");

  /* Validate the connection */
  validity = cmgr_validate (connectionId);
  if (validity > 0) {

    /* Get the RPC2 handle from the specified connection */
    pConnection = cmgr_get(connectionId, CMGR_LOCK);
    if (!pConnection) {
      ds_thread_log (0, "speechw: unable to get connection");
    } /* endif */
    
  } else {
    /* Connection is down */
    ds_thread_log (1, "speechw: connection is down");
  } /* endif */

  ds_thread_log(10, "speechw: speechwGetConnection completed");

  return (pConnection);
}

/* speechwInitModel makes a RPC to a speech server to initialize a
   speech model on that server.  If the model is initialized, the
   model parameters are returned.  If the RPC fails, it attempts to
   revalidate the connection and retry the RPC.  If the RPC succeeds,
   it returns zero and places the parameters in the specified
   structure, otherwise ir returns non-zero. */
int speechwInitModelRPC (cmgr_id_t connectionId, char* model, float* pFidelity,
			 float* pRecPerByteCost, float* pVQPerByteCost, 
			 float* pRecNoVQPerByteCost)
{
  ModelParms modelParms;
  cmgr_connection_t* pConnection;
  int validity;
  long rpc2_rc;
  int rc;

  ds_thread_log(10, "speechw: speechwInitModelRPC entered");

  /* Allocate the default amount of bytes for parameters */
  NALLOC (modelParms.fidelityBBS.SeqBody, char, 20);
  modelParms.fidelityBBS.MaxSeqLen = 20;
  modelParms.fidelityBBS.SeqLen = 0;
  NALLOC (modelParms.recBBS.SeqBody, char, 20);
  modelParms.recBBS.MaxSeqLen = 20;
  modelParms.recBBS.SeqLen = 0;
  NALLOC (modelParms.VQBBS.SeqBody, char, 20);
  modelParms.VQBBS.MaxSeqLen = 20;
  modelParms.VQBBS.SeqLen = 0;
  NALLOC (modelParms.recNoVQBBS.SeqBody, char, 20);
  modelParms.recNoVQBBS.MaxSeqLen = 20;
  modelParms.recNoVQBBS.SeqLen = 0;

  /* Get the appropriate connection from the connection manager */
  pConnection = speechwGetConnection (connectionId);
  if (!pConnection) {
    return (CMGR_ERROR);
  } /* endif */

  /* Make the RPC to a speech server */
  rc = speech_initModel (CMGR_RPC2_HANDLE (pConnection), model, &rpc2_rc, 
			 &modelParms);

  if (rc != RPC2_SUCCESS) {

    ds_thread_log (1, "speechw: RPC failed, rc = %d. Revalidating", rc);

    /* Mark down and revalidate and the conenction - this will correct some 
       transient errors like the server going down and coming back up */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (connectionId)) {
	ds_thread_log (0, "speechw: unable to mark down the connection");
      } /* endif */
    } /* endif */

    validity = cmgr_validate (connectionId);
    if (validity > 0) {
      /* OK - now retry the RPC */
      rc = speech_initModel (CMGR_RPC2_HANDLE (pConnection), model, &rpc2_rc, 
			     &modelParms);
    } else {
      /* No hope here */
      ds_thread_log (0, "speechw: connection does not validate");
    } /* endif */

  } /* endif */

  if (rc == RPC2_SUCCESS && rpc2_rc != RPC2_SUCCESS) {
    rc = rpc2_rc; 
  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (1, "speechw: initModel RPC failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (connectionId, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "speechw: unable to put connection");
  } /* endif */

  /* Copy and deallocate RPC data */
  if (rc == RPC2_SUCCESS) {
    if (modelParms.fidelityBBS.SeqBody) {
      *pFidelity = atof (modelParms.fidelityBBS.SeqBody);
      FREE (modelParms.fidelityBBS.SeqBody);
    } /* endif */
    if (modelParms.recBBS.SeqBody) {
      *pRecPerByteCost = atof (modelParms.recBBS.SeqBody);
      FREE (modelParms.recBBS.SeqBody);
    } /* endif */
    if (modelParms.VQBBS.SeqBody) {
      *pVQPerByteCost = atof (modelParms.VQBBS.SeqBody);
      FREE (modelParms.VQBBS.SeqBody);
    } /* endif */
    if (modelParms.recNoVQBBS.SeqBody) {
      *pRecNoVQPerByteCost = atof (modelParms.recNoVQBBS.SeqBody);
    FREE (modelParms.recNoVQBBS.SeqBody);
    } /* endif */
  } /* endif */

  ds_thread_log(10, "speechw: speechwInitModelRPC completed");

  return (rc);
}
/* speechwCloseModel makes a RPC to a speech server to close a
   speech model on that server.  If the RPC fails, it attempts to
   revalidate the connection and retry the RPC.  If the RPC succeeds,
   it returns zero. */
int speechwCloseModelRPC (cmgr_id_t connectionId, char* model)
{
  cmgr_connection_t* pConnection;
  int validity;
  long rpc2_rc;
  int rc;

  ds_thread_log(10, "speechw: speechwCloseModelRPC entered");

  /* Get the appropriate connection from the connection manager */
  pConnection = speechwGetConnection (connectionId);
  if (!pConnection) {
    return (CMGR_ERROR);
  } /* endif */

  /* Make the RPC to a speech server */
  rc = speech_closeModel (CMGR_RPC2_HANDLE (pConnection), model, &rpc2_rc);
  if (rc != RPC2_SUCCESS) {

    ds_thread_log (1, "speechw: RPC failed, rc = %d. Revalidating", rc);

    /* Mark down and revalidate and the conenction - this will correct some 
       transient errors like the server going down and coming back up */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (connectionId)) {
	ds_thread_log (0, "speechw: unable to mark down the connection");
      } /* endif */
    } /* endif */

    validity = cmgr_validate (connectionId);
    if (validity > 0) {
      /* OK - now retry the RPC */
      rc = speech_closeModel (CMGR_RPC2_HANDLE (pConnection), model, &rpc2_rc);
    } else {
      /* No hope here */
      ds_thread_log (0, "speechw: connection does not validate");
    } /* endif */

  } /* endif */

  if (rc == RPC2_SUCCESS && rpc2_rc != RPC2_SUCCESS) {
    rc = rpc2_rc; 
  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (1, "speechw: closeModel RPC failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (connectionId, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "speechw: unable to put connection");
  } /* endif */

  ds_thread_log(10, "speechw: speechwCloseModelRPC completed");

  return (rc);
}

/* speechwMakeRPC makes a RPC to a speech server.  If the RPC fails, it 
   attempts to revalidate the connection and retry the RPC.  If the RPC
   succeeds, it returns zero, otherwise non-zero. */
int speechwMakeRPC (cmgr_id_t connectionId, cmgr_connection_t* pConnection,
		    RecognitionParms* pRecognitionParms, int* pJanusRC, 
		    RPC2_BoundedBS* pTranslationBBS, SE_Descriptor sed,
		    struct timeval* ptvStart, struct timeval* ptvEnd) 
{
  int validity;
  int rc;
  long start_sec, start_usec, end_sec, end_usec;

  ds_thread_log(10, "speechw: speechwMakeRPC entered");

  /* Make the RPC to a speech server */
  rc = speech_getTranslation (CMGR_RPC2_HANDLE (pConnection), 
			      pRecognitionParms,
			      (RPC2_Integer *) pJanusRC, pTranslationBBS, 
			      &sed, (RPC2_Integer *) &start_sec, 
			      (RPC2_Integer *) &start_usec, 
			      (RPC2_Integer *) &end_sec, 
			      (RPC2_Integer *) &end_usec);
  if (rc != RPC2_SUCCESS) {

    ds_thread_log (1, "speechw: RPC failed, rc = %d. Revalidating", rc);

    /* Mark down and revalidate and the conenction - this will correct some 
       transient errors like the server going down and coming back up */
    if (rc < RPC2_FLIMIT) {
      if (cmgr_markdown (connectionId)) {
	ds_thread_log (0, "speechw: unable to mark down the connection");
      } /* endif */
    } /* endif */

    validity = cmgr_validate (connectionId);
    if (validity > 0) {
      /* OK - now retry the RPC */
      rc = speech_getTranslation (CMGR_RPC2_HANDLE (pConnection), 
				  pRecognitionParms, (RPC2_Integer *) pJanusRC,
				  pTranslationBBS, &sed,
				  (RPC2_Integer *) &start_sec,
				  (RPC2_Integer *) &start_usec, 
				  (RPC2_Integer *) &end_sec, 
				  (RPC2_Integer *) &end_usec);

    } else {
      /* No hope here */
      ds_thread_log (0, "speechw: connection does not validate");
    } /* endif */

  } /* endif */

  if (rc != RPC2_SUCCESS) {
    ds_thread_log (1, "speechw: getTranslation RPC failed, rc = %d", rc);
  } /* endif */

  /* RPC is done - put the connection back to the connection manager */
  if (cmgr_put (connectionId, CMGR_LOCK) != CMGR_SUCCESS) {
    ds_thread_log (0, "speechw: unable to put connection");
  } /* endif */

  /* Save timing data */
  ptvStart->tv_sec = start_sec;
  ptvStart->tv_usec = start_usec;
  ptvEnd->tv_sec = end_sec;
  ptvEnd->tv_usec = end_usec;

  ds_thread_log(10, "speechw: speechwMakeRPC completed");

  return (rc);
}

/* speechwRecognizeADC performs a full recognition of the data in the 
   input buffer by making a RPC to the Janus server on the specified 
   connection.  On success, allocates and returns a buffer containing
   the translation.  The function returns zero on success, non-zero 
   otherwise. */
int speechwRecognizeADC (cmgr_id_t connectionId, char* model, char* fidelity,
			 char* buffer, int len, char** pTranslation, 
			 /* int searchEffortPercent, */
			 struct timeval* ptvStart, struct timeval* ptvEnd)
{
  cmgr_connection_t* pConnection;
  RecognitionParms recognitionParms;
  SE_Descriptor sed;
  RPC2_BoundedBS translationBBS; /* translation buffer */
  int janusRC; /* Return code from janus */
  int rc = 0; /* Return code from called functions */

  ds_thread_log(10, "speechw: speechwRecognizeADC entered");

  pConnection = speechwGetConnection (connectionId);
  if (!pConnection) {
    return (CMGR_ERROR);
  } /* endif */

  /* Set up side-effect to transfer speech data to the server */
  bzero(&sed, sizeof(sed));
  sed.Tag = SMARTFTP;
  sed.Value.SmartFTPD.Tag = FILEINVM;
  sed.Value.SmartFTPD.ByteQuota = -1;
  sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buffer;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = len;

  /* Allocate the default amount of bytes for translation */
  NALLOC (translationBBS.SeqBody, char, DEF_TRANSLATION_SIZE);
  translationBBS.MaxSeqLen = DEF_TRANSLATION_SIZE;
  translationBBS.SeqLen = 0;

  /* Get the translation of the speech data */
  recognitionParms.inputType = DATATYPE_ADC;
  recognitionParms.outputType = DATATYPE_HYP;
  recognitionParms.model = model;
  recognitionParms.fidelity = fidelity;
#if 0
  recognitionParms.searchEffortPercent = searchEffortPercent;
#endif

  rc = speechwMakeRPC (connectionId, pConnection, &recognitionParms,
		       &janusRC, &translationBBS, sed, ptvStart, ptvEnd);

  /* Check the return code from Janus */
  if (rc == 0) {

    if (janusRC == 0) {
      
      /* Make a copy of the translation */
      NALLOC (*pTranslation, char, translationBBS.SeqLen);
      memcpy (*pTranslation, translationBBS.SeqBody, translationBBS.SeqLen);
      
      /* Save the timing data */

    } else {
      ds_thread_log (1, "speechw: translation failed, rc = %d", janusRC);
      return (janusRC);
    } /* endif */

  } /* endif */

  /* Deallocate the translation buffer */
  FREE (translationBBS.SeqBody);

  ds_thread_log (10, "speechw: speechwRecognizeADC completed");

  return (rc);
}

/* speechwCreateMEL converts the data in the specified buffer into MEL 
   format by making a RPC to the Janus server on the specified 
   connection.  On success, it allocates and returns a buffer containing
   the name of a fiel containing the MEL data.  The function returns zero 
   on success, non-zero otherwise. */
int speechwCreateMEL (cmgr_id_t connectionId, char* buffer, int len,
		      char** pFilename, struct timeval* ptvStart, 
		      struct timeval* ptvEnd)
{
  cmgr_connection_t* pConnection;
  RecognitionParms recognitionParms;
  SE_Descriptor sed;
  RPC2_BoundedBS filenameBBS; /* translation buffer */
  int janusRC; /* Return code from janus */
  int rc = 0; /* Return code from called functions */

  ds_thread_log(10, "speechw: speechwCreateMEL entered");

  pConnection = speechwGetConnection (connectionId);
  if (!pConnection) {
    return (CMGR_ERROR);
  } /* endif */

  /* Set up side-effect to transfer speech data to the server */
  bzero(&sed, sizeof(sed));
  sed.Tag = SMARTFTP;
  sed.Value.SmartFTPD.Tag = FILEINVM;
  sed.Value.SmartFTPD.ByteQuota = -1;
  sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buffer;
  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = len;

  /* Allocate the default amount of bytes for translation */
  NALLOC (filenameBBS.SeqBody, char, DEF_TRANSLATION_SIZE);
  filenameBBS.MaxSeqLen = DEF_TRANSLATION_SIZE;
  filenameBBS.SeqLen = 0;

  /* Get the translation of the speech data */
  recognitionParms.inputType = DATATYPE_ADC;
  recognitionParms.outputType = DATATYPE_MEL;
  recognitionParms.model = "N/A";
  recognitionParms.fidelity = "N/A";
  rc = speechwMakeRPC (connectionId, pConnection, &recognitionParms,
		       &janusRC, &filenameBBS, sed, ptvStart, ptvEnd);

  /* Check the return code from Janus */
  if (rc == 0) {

    if (janusRC == 0) {
      
      /* Make a copy of the filename */
      NALLOC (*pFilename, char, filenameBBS.SeqLen);
      memcpy (*pFilename, filenameBBS.SeqBody, filenameBBS.SeqLen);
      
    } else {
      ds_thread_log (1, "speechw: translation failed, rc = %d", janusRC);
      return (janusRC);
    } /* endif */

  } /* endif */

  /* Deallocate the buffer */
  FREE (filenameBBS.SeqBody);

  ds_thread_log (10, "speechw: speechwCreateMEL completed");

  return (rc);
}

/* speechwRecognizeMEL performs a recognition data stored in MEL
   format in the specified file by making a RPC to the Janus server on
   the specified connection.  On success, allocates and returns a
   buffer containing the translation.  The function returns zero on
   success, non-zero otherwise. */
int speechwRecognizeMEL (cmgr_id_t connectionId, char* model, char* fidelity,
			 char* filename,
			 char** pTranslation, /* int searchEffortPercent, */
			 struct timeval* ptvStart, struct timeval* ptvEnd)
{
  cmgr_connection_t* pConnection;
  RecognitionParms recognitionParms;
  SE_Descriptor sed;
  RPC2_BoundedBS translationBBS; /* translation buffer */
  int janusRC; /* Return code from janus */
  int rc = 0; /* Return code from called functions */

  ds_thread_log(10, "speechw: speechwRecognizeMEL entered");

  pConnection = speechwGetConnection (connectionId);
  if (!pConnection) {
    return (CMGR_ERROR);
  } /* endif */

  /* Set up side-effect to transfer speech data to the server */
  bzero(&sed, sizeof(sed));
  sed.Tag = SMARTFTP;
  sed.Value.SmartFTPD.Tag = FILEBYNAME;
  sed.Value.SmartFTPD.ByteQuota = -1;
  sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;
  if (strlen (filename) + 1 < 
      sizeof (sed.Value.SmartFTPD.FileInfo.ByName.LocalFileName)) {
    strcpy (sed.Value.SmartFTPD.FileInfo.ByName.LocalFileName, filename);
  } else {
    ds_thread_log (1, "speechw: MEL file name is too long: %s", filename);
    return (-1);
  } /* endif */

  /* Allocate the default amount of bytes for translation */
  NALLOC (translationBBS.SeqBody, char, DEF_TRANSLATION_SIZE);
  translationBBS.MaxSeqLen = DEF_TRANSLATION_SIZE;
  translationBBS.SeqLen = 0;

  /* Get the translation of the speech data */
  recognitionParms.inputType = DATATYPE_MEL;
  recognitionParms.outputType = DATATYPE_HYP;
  recognitionParms.model = model;
  recognitionParms.fidelity = fidelity;
#if 0
  recognitionParms.searchEffortPercent = searchEffortPercent;
#endif

  rc = speechwMakeRPC (connectionId, pConnection, &recognitionParms,
		       &janusRC, &translationBBS, sed, ptvStart, ptvEnd);

  /* Check the return code from Janus */
  if (rc == 0) {

    if (janusRC == 0) {
      
      /* Make a copy of the translation */
      NALLOC (*pTranslation, char, translationBBS.SeqLen);
      memcpy (*pTranslation, translationBBS.SeqBody, translationBBS.SeqLen);
      
    } else {
      ds_thread_log (1, "speechw: translation failed, rc = %d", janusRC);
      return (janusRC);
    } /* endif */

  } /* endif */

  /* Deallocate the translation buffer */
  FREE (translationBBS.SeqBody);

  ds_thread_log (10, "speechw: speechwRecognizeMEL completed");

  return (rc);
}

/* speechwGetTranslationLocal () performs speech recognition when the 
   configuration is set to local.  The function returns zero on success,
   non-zero otherwise. */
int speechwGetTranslationLocal (SpeechClientData* pData, 
				SpeechConnectionData* pConn, char* model,
				char* buffer, int len, char** pTranslation)
{
  struct timeval tvBegin, tvEnd;
  int rc = 0; /* Assume success */

  ds_thread_log(10, "speechw: speechwGetTranslationLocal entered");

  rc = speechwRecognizeADC (pData->connections.localConnection, model, 
			    pConn->fidelity, buffer, len, pTranslation, 
			    /* pConn->searchEffort*100, */
			    &tvBegin, &tvEnd);
#ifdef INSTRUMENTED
  pConn->currentMeasurements.tvBeginJanusRecognition = tvBegin;
  pConn->currentMeasurements.tvEndJanusRecognition = tvEnd;
#endif

  ds_thread_log(10, "speechw: speechwGetTranslationLocal completed");

  return (rc);
}

/* speechwGetTranslationIntermediate () performs speech recognition when the 
   configuration is set to weakly connected.  The function returns zero on 
   success, non-zero otherwise. */
int speechwGetTranslationIntermediate (SpeechClientData* pData, 
				       SpeechConnectionData* pConn, 
				       char* model, 
				       char* buffer, int len, 
				       char** pTranslation)
{
  char* pMELfile = 0;
  struct stat fs;
  struct timeval tvBegin, tvEnd;
  int rc = 0; /* Assume success */

  ds_thread_log(10, "speechw: speechwGetTranslationIntermediate entered");

  rc = speechwCreateMEL (pData->connections.localConnection, 
			 buffer, len, &pMELfile, &tvBegin, &tvEnd);

#ifdef INSTRUMENTED
    pConn->currentMeasurements.tvBeginMELConversion = tvBegin;
    pConn->currentMeasurements.tvEndMELConversion = tvEnd;
#endif

  if (rc == 0) {
    
    /* Make sure file exists */
    if (stat (pMELfile, &fs) == 0) {
      
      rc = speechwRecognizeMEL (pData->connections.remoteConnection,
				model, pConn->fidelity, 
				pMELfile, pTranslation, 
				/* pConn->searchEffort*100, */
				&tvBegin, &tvEnd);
      
#ifdef INSTRUMENTED
      pConn->currentMeasurements.MELfilesize = fs.st_size;
      pConn->currentMeasurements.tvBeginJanusRecognition = tvBegin;
      pConn->currentMeasurements.tvEndJanusRecognition = tvEnd;
#endif

    } else {
      ds_thread_log (0, "speechw: specified MEL file %s not found", pMELfile);
    } /* endif */

  } /* endif */
  
  FREE (pMELfile);

  ds_thread_log(10, "speechw: speechwGetTranslationIntermediate completed");

  return (rc);
}

/* speechwGetTranslationRemote () performs speech recognition when the 
   configuration is set to remote.  The function returns zero on success,
   non-zero otherwise. */
int speechwGetTranslationRemote (SpeechClientData* pData, 
				 SpeechConnectionData* pConn, char* model,
				 char* buffer, int len, char** pTranslation)
{
  struct timeval tvBegin, tvEnd;
  int rc = 0; /* Assume success */

  ds_thread_log(10, "speechw: speechwGetTranslationRemote entered");

  rc = speechwRecognizeADC (pData->connections.remoteConnection, model, 
			    pConn->fidelity, buffer, len, pTranslation, 
			    /* pConn->searchEffort*100, */
			    &tvBegin, &tvEnd);

#ifdef INSTRUMENTED
  pConn->currentMeasurements.tvBeginJanusRecognition = tvBegin;
  pConn->currentMeasurements.tvEndJanusRecognition = tvEnd;
#endif
  
  ds_thread_log(10, "speechw: speechwGetTranslationRemote completed");

  return (rc);
}

/* speechwGetTranslation() asks the server to recognize data in the
   specified input buffer by making a RPC.  The function determines,
   based upon the current strategy, which combination of local and
   remote janus servers to use and then invokes the appropriate
   functions to recognize the data.  The function returns zero on
   success, non-zero otherwise. */
int speechwGetTranslation (struct viceroy_onode* op, SpeechClientData* pData,
			   SpeechConnectionData* pConn, char* model,
			   char* buffer, int len, char** pTranslation)
{
  int size;
  int rc = 0; /* Assume success */

  ds_thread_log(10, "speechw: speechwGetTranslation entered");

  ds_thread_log (25, "speechw: strategy is %d", pConn->strategy);

#ifdef INSTRUMENTED
  pConn->currentMeasurements.strategy = pConn->strategy;
#endif

  if (pConn->strategy == SPEECHW_STRATEGY_LOCAL) {

    rc = speechwGetTranslationLocal (pData, pConn, model, buffer, len, 
				     pTranslation);

    /* This is a special case - because Odyssey uses passive bandwidth
       estimation, operating in local strategy will NEVER update the
       bandwidth estimation to the remote server, so there is no way
       to know when the bandwidth improves.  So, to remedy this, implement
       an asynchronous active ping of the server */
    size = len;
    if (size > PING_MAX_SIZE) {
      size = PING_MAX_SIZE;
    } /* endif */
    if (!noPing) {
      speechwPingRemoteServer (pData->connections.remoteConnection, size); 
    }

  } else if (pConn->strategy == SPEECHW_STRATEGY_HYBRID) {

    rc = speechwGetTranslationIntermediate (pData, pConn, model, buffer, len, 
					    pTranslation);

  } else if (pConn->strategy == SPEECHW_STRATEGY_REMOTE) { 

    rc = speechwGetTranslationRemote (pData, pConn, model, buffer, len, 
				      pTranslation);

  } else {
    ds_thread_log (0, "speechw: unsupported strategy");
    rc = EINVAL;
  } /* endif */

  ds_thread_log(10, "speechw: speechwGetTranslation completed");

  return (rc);
}


