#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/viceroy.h"
#include "include/onode-manager.h"

#include "ody_speech.h"
#include "libodyspeech_tsops.h"
#include "speechw_private.h"
#include "speechw_comm.h"
#include "speechw_ping.h"

#include "odyssey/viceroy/rsrc_en.h" /* XXXAPM */

static SpeechPingParameters pingParms;
static int threadCreated = 0;

/* speechwCreatePingThread creates a thread to actively ping the remote
   server.  If the thread has already been created, calling the function 
   again is a no-op.  The function returns zero on success, non-zero on
   error */
int speechwCreatePingThread ()
{
  ds_thread_log (10, "speechw: speechwCreatePingThread entered");

  if (!threadCreated) {

    /* Create the ping thread */
    if (ds_thread_create (speechwPingProcess, &pingParms, DS_THREAD_DAEMON,
			  PING_CLASS, 0)) {
      threadCreated = 1;
    } else {
      ds_thread_log (0, "speechw: error creating ping thread");
    } /* endif */
    
  } /* endif */
	 
  ds_thread_log (10, "speechw: speechwCreatePingThread completed");
	
  return (0);
}
  
/* speechwPingRemoteServer sends a ping of the specified size on the 
   connection specified by the given connection id.  It sets the values
   in the ping parameter structure, then signals the ping thread.  If the
   ping thread already has a ping in progress, the signal will be lost, but
   this is OK. This function returns zero on success, non-zero on failure. */
int speechwPingRemoteServer (int connectionId, int size)
{
  ds_thread_log (10, "speechw: speechwPingRemoteServer entered");

  pingParms.connectionId = connectionId;
  pingParms.size = size;
  
  ds_thread_wakeup (PING_EVENT);

  ds_thread_log (10, "speechw: speechwPingRemoteServer completed");

  return (0);
}

/* speechwPingProcess waits to be signalled, then initiates a ping of
   the server using the specified parametes in the parameter structure.
   The function never terminates since a thread can not be destroyed. */
int speechwPingProcess (void* pParameter)
{
  char* buffer;
  cmgr_connection_t* pConnection;
  SE_Descriptor sed;
  SpeechPingParameters* pParms = (SpeechPingParameters *) pParameter;
  int rc;

  ds_thread_log (10, "speechw: speechwPingProcess entered");

  while (1) {

    /* Wait until an active ping is needed */
    ds_thread_wait (PING_EVENT);

    NALLOC (buffer, char, pParms->size);
    
    /* Get the connection on which ping will be sent */
    pConnection = speechwGetConnection (pParms->connectionId);
    if (pConnection) {
    
      /* Initiate a ping RPC */
      bzero(&sed, sizeof(sed));
      sed.Tag = SMARTFTP;
      sed.Value.SmartFTPD.Tag = FILEINVM;
      sed.Value.SmartFTPD.ByteQuota = -1;
      sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqBody = buffer;
      sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen = pParms->size;
      sed.Value.SmartFTPD.TransmissionDirection = CLIENTTOSERVER;
      
      ds_thread_log (25, "speech2: pinging connection %d with buffer size %d", pParms->connectionId, pParms->size);

      rc = speech_pingServer (CMGR_RPC2_HANDLE (pConnection), pParms->size, &sed);

      if (rc != RPC2_SUCCESS) {
	
	ds_thread_log (1, "speechw: RPC ping failed, rc = %d", rc);
	
	/* Mark down the conenction */
	if (rc < RPC2_FLIMIT) {
	  if (cmgr_markdown (pParms->connectionId)) {
	    ds_thread_log (0, "speechw: unable to mark down the connection");
	  } /* endif */
	} /* endif */
	
      } /* endif */
      
      /* RPC is done - put the connection back to the connection manager */
      if (cmgr_put (pParms->connectionId, CMGR_LOCK) != CMGR_SUCCESS) {
	ds_thread_log (0, "speechw: unable to put connection");
      } /* endif */
    
    } /* endif */

    FREE (buffer);
    
  } /* endwhile */

  return (-1);
}

