#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

#include <stdio.h>
#include <string.h>

#include <sys/queue.h>

#include "include/odyssey.h"
#include "include/odytypes.h"
#include "base/ds/ds_thread.h"
#include "base/ds/ds_hash.h"
#include "base/ds/ds_hwlock.h"
#include "base/ds/ds_list.h"
#include "base/ds/ds_rrlist.h"

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include "include/cm_rpc2.h"
#include "include/cm.h"
#include "include/warden.h"
#include "include/onode-manager.h"

#include "ody_speech.h"
#include "odyjanus.h"
#include "libodyspeech_tsops.h"
#include "speechwarden.h"
#include "speechw_private.h"
#include "speechw_comm.h"
#include "speechw_ping.h"

/* Global data */
ServerLocations locations = { -1, 0, -1};
static int printStatistics = 0;
int noPing = 0;
int speechw_rtimeout = 0;

struct warden speech_warden = {
  SPEECH_TTYPE,
  "Speech",
  0,
  speechwParseArguments,
  "Usage: -speechw_server <server name> -speechw_lport <local port> "
  "-speechw_rport <remote port> [-speechw_printstats] [-speechw_rtimeout <sec>]",
  speechwInitialize,
  speechwGetCID,
  speechwPrintStatistics,
  speechwGetRoot,
  speechwLookup,
  speechwAccess,
  speechwGetAttr,
  speechwOpen,
  speechwClose,
  speechwReadWrite,
  speechwReadWriteDispose,
  (w_readdir_fn) NULL,
  (w_readdir_dispose_fn) NULL,
  speechwTSOP,
  speechwTSOPDispose,
#ifdef SETS
  (w_expand_fn) NULL,
  (w_cursexpand_fn) NULL,
  (w_prefetch_fn) NULL,
#endif
  0,
  0,
  (void *)NULL
};

/* speechwParseArguments() takes in an argc/argv pair and an index to
   the current argument being parsed.  It processes as many arguments
   as it can starting at the specified index and returns the number of
   arguments parsed. */
int speechwParseArguments(int cur, int argc, char *argv[])
{
  int argsParsed = 0;
  int done = 0; /* Set when an argument that cannot be parsed is found */

  while (cur + argsParsed < argc && !done) {

    if (strcmp(argv[cur + argsParsed], "-speechw_server") == 0) {
      if (cur + argsParsed + 1 >= argc) {
	return -1;
      } /* endif */
      if (*(argv[cur + argsParsed + 1]) == '-') {
	return -1;
      } /* endif */
      locations.remoteHostname = strdup(argv[cur + argsParsed + 1]);
      argsParsed += 2;
    } else if (strcmp (argv[cur + argsParsed], "-speechw_lport") == 0) {
      if (cur + argsParsed + 1 >= argc) {
	return -1;
      } /* endif */
      if (*(argv[cur + argsParsed + 1]) == '-') {
	return -1;
      } /* endif */
      locations.localPort = atoi (argv[cur + argsParsed + 1]);
      argsParsed += 2;
    } else if (strcmp (argv[cur + argsParsed], "-speechw_rport") == 0) {
      if (cur + argsParsed + 1 >= argc) {
	return -1;
      } /* endif */
      if (*(argv[cur + argsParsed + 1]) == '-') {
	return -1;
      } /* endif */
      locations.remotePort = atoi (argv[cur + argsParsed + 1]);
      argsParsed += 2;
    } else if (strcmp (argv[cur + argsParsed], "-speechw_noping") == 0) {
      noPing = 1;
      argsParsed++;
    } else if (strcmp (argv[cur + argsParsed], "-speechw_printstats") == 0) {
      printStatistics = 1;
      argsParsed++;
    } else if (strcmp (argv[cur + argsParsed], "-speechw_rtimeout") == 0) {
      if (cur + argsParsed + 1 >= argc) {
	return -1;
      } /* endif */
      if (*(argv[cur + argsParsed + 1]) == '-') {
	return -1;
      } /* endif */
      speechw_rtimeout = atoi (argv[cur + argsParsed + 1]);
      argsParsed += 2;
    } else {
      done = 1; /* Found an argument we can't parse */
    } /* endif */

  } /* endwhile */

  return (argsParsed);
}

/* speechwInitialize() performs the initialization of the speech warden.
   It allocates a private data area in the warden structure and copies
   data into the area.  It creates a root node for the single tome.  The
   function returns zero on succes, non-zero otherwise */
int speechwInitialize(struct warden *wp)
{
  SpeechClientData       *pData = 0;
  struct ofid            root_ofid;
  struct onode_create_in root_onode_params;
  struct viceroy_onode           *rop = 0;
  int                    rc = 0; /* Assume successful completion */
  
  ds_thread_log(10, "speechw: speechwInitialize entered");

  wp->w_data = NULL;

  /* To connect to remote Janus server, server and port must be specified */
  if (locations.remoteHostname == 0 || locations.remotePort == -1) {
    locations.remoteHostname = 0;
    locations.remotePort = -1;
    ds_thread_log (0, "speechw: remote server or port not specified");
    ds_thread_log (0, "speechw: speech warden will not connect remotely");
  } /* endif */

  if (locations.localPort == -1) {
    ds_thread_log (0, "speechw: local port not specified");
    ds_thread_log (0, "speechw: speech warden will not connect locally");
  } /* endif */

  if (locations.localPort == -1 && locations.remotePort == -1) {
    ds_thread_log (0, "speechw: speech warden can not start");
    return (-1);
  } /* endif */

  /* Allocate client data for structure for speech warden.  This is
     attached to the warden data structure */
  ALLOC (wp->w_data, SpeechClientData);
  pData = (SpeechClientData *) wp->w_data;
  pData->pLocations = &locations;
  pData->uniquifier = 0;
#ifdef INSTRUMENTED
  pData->lastMeasurements.totalBytes = 0; /* No measurements taken yet */
#endif /* INSTRUMENTED */
  ds_hwlock_init (&pData->lock); /* Initialize lock to guard uniquifier */

  /* set up the root onode in init, since there is only one volume */
  root_ofid.ofid_ttype = SPEECH_TTYPE;
  root_ofid.ofid_tome  = SPEECH_TOME;	/* we only support tome #1 for now */
  root_ofid.ofid_vnum  = 0;
  root_ofid.ofid_uniq  = pData->uniquifier++;
  
  root_onode_params.oc_flags = ON_ROOTOFVOL | ON_UNCACHEABLE;
  root_onode_params.oc_vtype = O_VDIR; 
  root_onode_params.oc_ofree = NULL;
  root_onode_params.oc_data  = NULL;
  
  rop = onode_create(&root_ofid, &root_onode_params);
  if (!rop) {
    ds_thread_log(0, "speechw: Could not create volume onode root");
    return (-1);
  } /* endif */

  /* Establish connections to servers */
  rc = speechwConnectToServers (pData->pLocations, &pData->connections);

  /* Start a thread to ping on the remote connection */
  if (locations.remotePort >= 0 && !noPing) {
    speechwCreatePingThread ();
  } /* endif */

  ds_thread_log(10, "speechw: speechwInitialize completed");
  
  return (rc);
}

/* speechwGetRoot() returns the root o-node of the specified tome.
   Since there is only one tome, only that tome may be specified.  The
   node is found by looking for the o-node with vnum and uniq of zero.
   The o-node is readlocked before being returned. The function returns
   zero on success, non-zero otherwise. */
int speechwGetRoot(struct warden *wp, tome_t tome, struct viceroy_onode **ppRoot)
{
  struct ofid of;

  ds_thread_log(10, "speechw: speechwGetRoot entered");
  
  if (tome != SPEECH_TOME) {
    ds_thread_log(10, "speechw: requested tome (%d) not supported", 
		  SPEECH_TOME);
    return(EINVAL);	
  } /* endif */
  
  of.ofid_ttype = SPEECH_TTYPE;
  of.ofid_tome  = tome;
  of.ofid_vnum  = 0;		/* for speech warden, {X,X,0,0} indicates */
  of.ofid_uniq  = 0;		/* an ofid for a volume. */
  
  *ppRoot = onode_get(&of);
  if (*ppRoot == NULL) {
    ds_thread_log(0, "speechw: onode_get() failed for warden tome "
		  "%d's root onode.", tome);
    return(EINVAL);
  } /* endif */
  
  onode_readlock ((*ppRoot));

  ds_thread_log(10, "speechw: speechwGetRoot completed");
  
  return (0);
}

/* speechwLookup() resolves the specified name.  Basically, any name is
   considered valid, and a new o-node is created and returned.  Right now,
   there is no checking to see if an o-node with that particular name
   already exists - I might add this later.  The function returns zero
   on success. */
int speechwLookup(struct warden* wp, struct viceroy_onode* odp, struct ody_cred* cred,
		  char* name, char** remain, struct viceroy_onode** opp)
{
  SpeechClientData*      pData = (SpeechClientData *) wp->w_data;
  SpeechConnectionData*  pConn;
  struct ofid            new_ofid;
  struct onode_create_in new_onode_params;
  struct viceroy_onode*          new_onode;

  ds_thread_log(10, "speechw: speechwLookup entered");
  
  /* Maybe should check to see if onode with name already exists */
  
  /* Release read lock and put directory onode */
  onode_release_read(odp);
  onode_put(odp);
  
  /* Don't bother looking at the name - just return a new o-node */
  *remain = NULL;

  /* Create the o-node */
  ds_hwlock_write (&pData->lock);
  new_ofid.ofid_uniq = pData->uniquifier++;
  ds_hwlock_release_write (&pData->lock);
  
  new_ofid.ofid_vnum  = new_ofid.ofid_uniq;
  new_ofid.ofid_ttype = SPEECH_TTYPE;
  new_ofid.ofid_tome  = SPEECH_TOME;
  
  new_onode_params.oc_flags = ON_UNCACHEABLE;
  new_onode_params.oc_vtype = O_VREG;
  new_onode_params.oc_ofree = speechwFreeONodeData;
  ALLOC (new_onode_params.oc_data, SpeechConnectionData);

  /* Initialize o-node to not busy */
  pConn = (SpeechConnectionData *) new_onode_params.oc_data;

  /* Default strategy is remote.  However, if no server is
     available remotely, use the local strategy. */
  if (pData->pLocations->remotePort >= 0) {
    pConn->strategy = SPEECHW_STRATEGY_REMOTE;
  } else {
    pConn->strategy = SPEECHW_STRATEGY_LOCAL;
  } /* endif */


  pConn->fidelity = "\"\""; /* Full fidelity */

#if 0
  pConn->searchEffort = 1.0; /* Use default for vocabulary */
#endif

  new_onode = onode_create (&new_ofid, &new_onode_params);
  if (!new_onode) {
    ds_thread_log(0, "speechw: could not create onode for %s", name);
    ASSERT(0);
  } /* endif */
  
  /* This is necessary becuase viceroy will put the o-node? */
  if (onode_get(&new_ofid) == new_onode) {
    *opp = new_onode;
  } else {
    ds_thread_log (0, "bitw_lookup: onode_get returned unexpected o-node");
    ASSERT(0);
  } /* endif */
  
  /* Must also be returned readlocked */
  onode_readlock(new_onode);
  
  ds_thread_log(10, "speechw: speechwLookup completed");
  
  return(0);
}

/* speechwAccess() checks the specified user's access to a specified o-node.
   Since there is no security implemented, this function always returns 
   zero (success). */
int speechwAccess(struct warden* wp, struct viceroy_onode* op, struct ody_cred* cred,
		  int mode)
{
  ds_thread_log(10, "speechw: speechwAccess entered");
  ds_thread_log(10, "speechw: speechwAccess completed");
  return (0);
}

int speechwGetAttr(struct warden *wp, struct viceroy_onode *op,
		   struct ody_vattr *vap, struct ody_cred *cred)
{
    time_t temp = 0;

    ds_thread_log(10, "speechw: speechwGetAttr entered");

    if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0)
      vap->va_type = O_VDIR;	/* this is the root - a directory */
    else
      vap->va_type = O_VREG;
    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;
    vap->va_blocksize = 512;
    vap->va_atime = temp;
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    ds_thread_log(10, "speechw: speechwGetAttr completed");

    return (0);
}

/* speechwOpen() opens an o-node.  It first checks to make sure that
   it is not the root o-node and that is not already busy (being used
   by another process. */
int speechwOpen (struct warden *wp, struct viceroy_onode *op, struct ody_cred *cred,
		 int mode)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;

  ds_thread_log(10, "speechw: speechwOpen entered");

  /* Check to see if o-node is the root */
  if (op->o_fid.ofid_vnum == 0 || op->o_fid.ofid_uniq == 0) {
    return (EISDIR);
  } /* endif */

  /* Set initial connection state */
  pConn->pModels = ds_list_create (NULL, FALSE, FALSE);

  ds_thread_log(10, "speechw: speechwOpen completed");

  return (0);
}

/* speechwClose () is called when the client process closes the o-node.  Any
   existing connection is cleaned up in this process.  The function returns
   zero on success. */
int speechwClose(struct warden *wp, struct viceroy_onode  *op, struct ody_cred  *cred,
		 int mode)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  SpeechModel* pModel;

  ds_thread_log(10, "speechw: speechwClose entered");

  /* Close any speech models which are still initialized */
  while ((pModel = ds_list_get_first (pConn->pModels)) != NULL) {
    speechwUnloadModel (wp, pModel);
    ds_list_remove (pConn->pModels, pModel);
    FREE (pModel);
  } /* endwhile */

  ds_list_destroy (pConn->pModels);

  ds_thread_log(10, "speechw: speechwClose completed");

  return (0);
}

/* speechReadWrite () is not supported.  All meaningful interactions with
   an open reference take place through TSOPs. */
int speechwReadWrite (struct warden* wp, struct viceroy_onode* op, struct ody_cred* cred,
		      int rwflag, int* count, int offset, int ioflag,
		      void **data)
{
  ds_thread_log(10, "speechw: speechwReadWrite entered");
  ds_thread_log(10, "speechw: speechwReadWrite completed");
  return (EOPNOTSUPP);
}

/* speechwReadWriteDispose() doesn't do much of anything. It probably
   should never be called. */
int speechwReadWriteDispose (struct warden* wp, struct viceroy_onode* op, 
			     struct ody_cred* cred, int count, int offset,
			     void* data)
{
  ds_thread_log(10, "speechw: speechwReadWriteDispose entered");
  ds_thread_log(10, "speechw: speechwReadWriteDispose completed");
  return (0);
}

/* speechwTSOP() handles all type-specific operations for the speech
   warden.  */
int speechwTSOP(struct warden* wp, int tsvnum, struct viceroy_onode* op,
		struct ody_cred* cred, int opnum, size_t argsz,
		void* argp, size_t* retsz, void** retp)
{
  long result;

  ds_thread_log(10, "speechw: speechwTSOP entered");

  /* check the version number */
  if(tsvnum != SPEECHWARDEN_VERSION) {
    return(ODY_EBADVERSION);
  } /* endif */

  /* dispatch TSOPs, based on opnum */
  switch(opnum) {
    case SPEECHW_SET_STRATEGY: 
      result = speechwSetStrategy (wp, op, argsz, argp, retsz, retp);
      break;
    case SPEECHW_SET_FIDELITY: 
      result = speechwSetFidelity (wp, op, argsz, argp, retsz, retp);
      break;
    case SPEECHW_INIT_MODELS:
      result = speechwInitModels (wp, op, argsz, argp, retsz, retp);
      break;
    case SPEECHW_RECOGNIZE:
      result = speechwRecognize (wp, op, argsz, argp, retsz, retp);
      break;
    case SPEECHW_CLOSE_MODELS:
      result = speechwCloseModels (wp, op, argsz, argp, retsz, retp);
      break;
#if 0
    case SPEECHW_SET_SEARCH_EFFORT:
      result = speechwSetSearchEffort (wp, op, argsz, argp, retsz, retp);
      break;
#endif
    case SPEECHW_DUMP_STATS:
      result = speechwDumpStats (wp, argsz, argp, retsz, retp);
      break;
    default:
      result = EOPNOTSUPP;
      break;
  } /* end switch */

  ds_thread_log(10, "speechw: speechwTSOP completed");

  return (result);
}

/* speechwTSOPDispose() deallocates retp which was allocated by speechwTSOP().
   It returns zero on success. */
int speechwTSOPDispose(struct warden* wp, struct viceroy_onode* op, struct ody_cred* cred,
		       int opnum, size_t argsz, void* argp, size_t retsz,
		       void* retp)
{
  ds_thread_log(10, "speechw: speechwTSOPDispose entered");

  /* dispose of *retp, since we were responsible for allocating it */
  FREE(retp);		
  
  ds_thread_log(10, "speechw: speechwTSOPDispose completed");

  return(0);
}

/* speechwFreeOnodeData() deallocates any connection data associated with
   an o-node */
void speechwFreeONodeData (void* ptr)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) ptr;

  ds_thread_log (10, "speechw: speechwFreeONodeData entered");

  FREE (pConn);

  ds_thread_log (10, "speechw: speechwFreeONodeData completed");
}

/* speechwSetStrategy() is a TSOP which allows a process to specify a
   strategy of local, hybrid, or remote.  The function returns zero if
   a valid strategy is specified. */
long speechwSetStrategy (struct warden* wp, struct viceroy_onode* op, size_t argsz, 
			 void* argp, size_t* retsz, void** retp)
{
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  enum strategyType strategy;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwSetStrategy entered");

  /* Argument should be an integer value (strategy) */
  if (argsz == sizeof (enum strategyType)) {

    strategy = *((enum strategyType *)argp);
    switch (strategy) {
    case SPEECHW_STRATEGY_LOCAL:
	if (pData->pLocations->localPort >= 0) {
	  pConn->strategy = strategy;
	  ds_thread_log (100, "speechw: Strategy set to: local");
	} else {
	  ds_thread_log (1, "speechw: local strategy not possible");
	  rc = ENODEV;
	} /* endif */
	break;
    case SPEECHW_STRATEGY_HYBRID:                
	if (pData->pLocations->localPort >= 0 && 
	    pData->pLocations->remotePort >= 0) {
	  pConn->strategy = strategy;
	  ds_thread_log (100, "speechw: Strategy set to: hybrid");
	} else {
	  ds_thread_log (1, "speechw: hybrid strategy no possible");
	  rc = ENODEV;
	} /* endif */
	break;
    case SPEECHW_STRATEGY_REMOTE:
	if (pData->pLocations->remotePort >= 0) {
	  pConn->strategy = strategy;
	  ds_thread_log (100, "speechw: Strategy set to: remote");
	} else {
	  ds_thread_log (1, "speechw: remote strategy no possible");
	  rc = ENODEV;
	} /* endif */
	break;
    default:
	ds_thread_log (1, "speechw: strategy not supported");
	rc = EINVAL;
    } /* end switch */
    
  } else {
    ds_thread_log (1, "speechw: strategy not specified");
    rc = EINVAL;
  } /* endif */

  /* Return result in retp */
  if (*retsz >= sizeof (rc)) {
    *retsz = sizeof(rc);
    ALLOC (*retp, int);
    memcpy (*retp, &rc, *retsz);
  } /* endif */
    
  ds_thread_log (10, "speechw: speechwSetStrategy completed");

  return (rc);
}

/* speechwSetFidelity() is a TSOP which allows a process to specify either
   full-fidelity or reduced fidelity data.  

   NOTE: If Janus is not loaded with both models - this will have no effect. */
long speechwSetFidelity (struct warden* wp, struct viceroy_onode* op, 
			 size_t argsz, void* argp, size_t* retsz, void** retp)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  enum fidelityType fidelity;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwSetFidelity entered");

  /* Argument should be an integer value (fidelity) */
  if (argsz == sizeof (enum fidelityType)) {

    fidelity = *((enum fidelityType *)argp);
    switch (fidelity) {
    case SPEECHW_FIDELITY_FULL: 
      pConn->fidelity = "\"\""; 
      ds_thread_log (50, "speechw: Fidelity set to full");
      break;
    case SPEECHW_FIDELITY_REDUCED: 
      pConn->fidelity = "Small";
      ds_thread_log (50, "speechw: Fidelity set to small");
      break;
    default: 
      ds_thread_log (0, "speechw: fidelity not supported");
      rc = EINVAL;
    }

  } else {
    ds_thread_log (1, "speechw: fidelity not specified");
    rc = EINVAL;
  } /* endif */
  
  /* Return result in retp */
  if (*retsz >= sizeof (rc)) {
    *retsz = sizeof(rc);
    ALLOC (*retp, int);
    memcpy (*retp, &rc, *retsz);
  } /* endif */
    
  ds_thread_log (10, "speechw: speechwSetFidelity completed");

  return (rc);
}

/* speechwFindModel() finds a specific model and returns it.  Otherwise, it
   returns NULL */
SpeechModel* speechwFindModel (SpeechConnectionData* pConn, char* model)
{
  SpeechModel* pModel;
  ds_list_iter_t* pIter;
  int done = 0;

  ds_thread_log (10, "speechw: speechwFindModel entered");

  pIter = ds_list_iter_create (pConn->pModels); 
  while (!done && (pModel = (SpeechModel*) ds_list_iter_next (pIter))) {
    if (!strcmp (pModel->model, model)) {
      done = 1;
    } /* endif */
  } /* endwhile */
  
  ds_list_iter_destroy (pIter);

  ds_thread_log (10, "speechw: speechwFindModel completed");

  return (pModel);
}

/* speechwInitModels() is a TSOP which initializes a specific speech
   model on all available servers, and returns to the application the
   various strategies available as well as the fidelity levels and CPU
   models associated with each strategy. */
long speechwInitModels (struct warden* wp, struct viceroy_onode* op, 
			size_t argsz, void* argp, size_t* retsz, 
			void** retp)
{
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  float fidelity, recPerByteCost, VQPerByteCost, recNoVQPerByteCost;
  StrategyModel strategyModel;
  SpeechModel* pModel;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwInitModels entered");

  /* Argument should be name of the model */
  if (argsz == 0) {
    ds_thread_log (1, "speechw: no model provided to speechwInitModels");
    return (EINVAL);
  } /* endif */

  /* Check for duplicates */
  if (speechwFindModel (pConn, argp)) {
    return (EEXIST);
  } /* endif */

  /* Try to initialize the model on all servers */
  if (pData->pLocations->localPort >= 0) {

    rc = speechwInitModelRPC (pData->connections.localConnection, argp,
			      &fidelity, &recPerByteCost, &VQPerByteCost, 
			      &recNoVQPerByteCost);
    if (rc == 0) {
      strategyModel.localEnabled = 1;
      strategyModel.localFidelity = fidelity;
      strategyModel.localRecPerByteCost = recPerByteCost;
      strategyModel.hybridVQPerByteCost = VQPerByteCost;
    } else {
      ds_thread_log (1, "speechw: local model init RPC failed, rc = %d", rc);
      strategyModel.localEnabled = 0;
      strategyModel.hybridEnabled = 0;
    } /* endif */
    
  } else {
    strategyModel.localEnabled = 0;
    strategyModel.hybridEnabled = 0;
  } /* endif */
  
  if (pData->pLocations->remotePort >= 0) {
    
    rc = speechwInitModelRPC (pData->connections.remoteConnection, argp, 
			      &fidelity, &recPerByteCost, &VQPerByteCost, 
			      &recNoVQPerByteCost);
    if (rc == 0) {
      strategyModel.hybridEnabled = strategyModel.localEnabled;
      strategyModel.remoteEnabled = 1;
      strategyModel.hybridFidelity = fidelity; 
      strategyModel.remoteFidelity = fidelity;
      strategyModel.hybridRecPerByteCost = recNoVQPerByteCost;
      strategyModel.remoteRecPerByteCost = recPerByteCost;
    } else {
      ds_thread_log (1, "speechw: remote model init RPC failed, rc = %d", rc);
      strategyModel.hybridEnabled = 0;
      strategyModel.remoteEnabled = 0;
    } /* endif */
    
  } else {
    strategyModel.hybridEnabled = 0;
    strategyModel.remoteEnabled = 0;
  } /* endif */
  
  /* Create a new model structure and add it to the list */
  ALLOC (pModel, SpeechModel);
  pModel->model = strdup (argp);
  pModel->initLocal = strategyModel.localEnabled;
  pModel->initRemote = strategyModel.remoteEnabled;
  ds_list_insert (pConn->pModels, pModel);
  
  /* Return result in retp */
  if (*retsz >= sizeof (strategyModel)) {
    *retsz = sizeof(strategyModel);
    ALLOC (*retp, StrategyModel);
    memcpy (*retp, &strategyModel, *retsz);
  } /* endif */
  
  ds_thread_log (10, "speechw: speechwInitModels completed");
  
  return (rc);
}
  
/* speechwRecognize() is a TSOP which recognizes an utterance using
   the specified speech model.  */
long speechwRecognize (struct warden* wp, struct viceroy_onode* op, 
		       size_t argsz, void* argp, size_t* retsz, 
		       void** retp)
{
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  SpeechModel* pModel;
  char* model;
  char* utterance;
  char* translation;
  int len;
  int rc;

  ds_thread_log (10, "speechw: speechwRecognize entered");

  /* Argument should be name of the model and an utterance */
  if (argsz == 0) {
    ds_thread_log (1, "speechw: no model provided to speechwRecognize");
    return (EINVAL);
  } /* endif */

  model = argp;
  
  if (argsz <= strlen(model)+1) {
    ds_thread_log (1, "speechw: no utterance provided to speechwRecognize");
    return (EINVAL);
  } /* endif */
    
  utterance = argp + strlen(model) + 1;
  len = argsz - strlen(model) - 1;

  /* Make sure the model is loaded */
  pModel = speechwFindModel (pConn, model);
  if (!pModel) {
    return (EINVAL);
  } /* endif */

  /* Do the recognition */
#ifdef INSTRUMENTED
  pConn->currentMeasurements.totalBytes = len;
  gettimeofday (&pConn->currentMeasurements.tvBeginTranslation, NULL);
#endif 
	
  /* Perform the speech recognition */
  rc = speechwGetTranslation (op, pData, pConn, model, utterance, len, 
			      &translation);
  if (rc) {
    ds_thread_log(1, "speechw: recognition failed, rc = %d", rc);
    return (rc);
  } /* endif */
    
#ifdef INSTRUMENTED
  gettimeofday (&pConn->currentMeasurements.tvEndTranslation, NULL);
  /* Copy measurements to warden data structure */
  ds_hwlock_write (&pData->lock);
  memcpy (&pData->lastMeasurements, &pConn->currentMeasurements,
	  sizeof (pConn->currentMeasurements));
  if (printStatistics) {
    speechwPrintStatistics (wp, stdout);
  } /* endif */
  ds_hwlock_release_write (&pData->lock);
#endif
    
  /* Return result in retp */
  if (*retsz >= strlen (translation)+1) {
    *retsz = strlen(translation)+1;
    NALLOC (*retp, char, strlen(translation)+1);
    strcpy (*retp, translation);
    FREE (translation);
  } else {
    ds_thread_log (1, "speechw: output buffer not big enough for translation");
    FREE (translation);
    return (ENOMEM);
  } /* endif */

  ds_thread_log (10, "speechw: speechwRecognize completed");
  
  return (rc);
}

/* speechwUnloadModel closes the model on all servers on which it
  is loaded. */
void speechwUnloadModel (struct warden* wp, SpeechModel* pModel)
{
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;
  int rc;

  ds_thread_log (10, "speechw: speechwUnloadModel entered");
  
  if (pModel->initLocal) {
    rc = speechwCloseModelRPC (pData->connections.localConnection, 
			       pModel->model);
    if (rc) {
      ds_thread_log (0, "speechw: unable to close local model");
    } /* endif */
  } /* endif */

  if (pModel->initRemote) {
    rc = speechwCloseModelRPC (pData->connections.remoteConnection, 
			       pModel->model);
    if (rc) {
      ds_thread_log (0, "speechw: unable to close remote model");
    } /* endif */
  } /* endif */

  ds_thread_log (10, "speechw: speechwUnloadModel entered");
}
  
/* speechwCloseModels() is a TSOP which closes a specific speech model
   on all servers on which it is loaded. */
long speechwCloseModels (struct warden* wp, struct viceroy_onode* op, 
			 size_t argsz, void* argp, size_t* retsz, 
			 void** retp)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  SpeechModel* pModel;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwCloseModels entered");

  /* Argument should be name of the model */
  if (argsz == 0) {
    ds_thread_log (1, "speechw: no model provided to speechwCloseModels");
    return (EINVAL);
  } /* endif */

  /* Find the model */
  pModel = speechwFindModel (pConn, argp);
  if (!pModel) {
    ds_thread_log (1, "speechw: unable to find specified model");
    return (EINVAL);
  } /* endif */

  /* Unload the model from any servers on which it is loaded */
  speechwUnloadModel (wp, pModel);

  /* Delete the model from the list */
  ds_list_remove (pConn->pModels, pModel);
  FREE (pModel);
  
  /* Return result in retp */
  if (*retsz >= sizeof (rc)) {
    *retsz = sizeof(rc);
    ALLOC (*retp, int);
    memcpy (*retp, &rc, *retsz);
  } /* endif */
    
  ds_thread_log (10, "speechw: speechwCloseModels completed");

  return (rc);
}

#if 0
/* speechwSetSearchEffort() is a TSOP which allows a process to specify
   the amount of effort to be expended in recognizing an utterance.  This
   value is used in setting Janus search parameters like beem width.  A
   search effort of one uses the default parameters.  Values smaller than
   one indicate that proportionally less effort should be expended (and
   that the recognition will be less accurate.)  The function returns zero 
   if a valid setting (between 0 and 10) is specified. */
long speechwSetSearchEffort (struct warden* wp, struct viceroy_onode* op, 
			     size_t argsz, void* argp, size_t* retsz, 
			     void** retp)
{
  SpeechConnectionData* pConn = (SpeechConnectionData *) op->o_data;
  float new_effort;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwSetSearchEffort entered");

  /* Argument should be a float value */
  if (argsz == sizeof (float)) {

    new_effort = *((float *)argp);
    if (new_effort > 0 && new_effort <= 10) {
      pConn->searchEffort = new_effort;
      ds_thread_log (100, "speechw: Search effort set to: %f", 
		     pConn->searchEffort);
    } else {
      ds_thread_log (1, "speechw: invalid search effort specified: %f",
		     new_effort);
      rc = 2;
    } /* endif */

  } else {
    ds_thread_log (1, "speechw: speechwSetSearchEffort parameter mismatch");
    rc = 1;
  } /* endif */

  /* Return result in retp */
  if (*retsz >= sizeof (rc)) {
    *retsz = sizeof(rc);
    ALLOC (*retp, int);
    memcpy (*retp, &rc, *retsz);
  } /* endif */
    
  ds_thread_log (10, "speechw: speechwSetSearchEffort completed");

  return (rc);
}
#endif

/* speechwDumpStats() is a TSOP which dumps the current 
   statistics to a specified log file.  This only is useful 
   when the warden is built with INSTRUMENTED defined.  The
   function returns zero if it is able to dump the statistics
   and non-zero otherwise */
long speechwDumpStats (struct warden* wp, size_t argsz, 
		       void* argp, size_t* retsz, void** retp)
{
  char* logfilename;
  FILE* logfile;
  struct timeval tv;
  long rc = 0; /* Assume success */

  ds_thread_log (10, "speechw: speechwDumpStats entered");

  /* Argument should be the name of a file to which statistics should
     be written. */
  if (argsz > 0) {

    logfilename = (char *) argp;
    logfile = fopen (logfilename, "a+");
    if (logfile) {

      gettimeofday (&tv, NULL);
      fprintf (logfile, "Speech statistics for: ");
      fprintf (logfile, "%ld.%ld\n", tv.tv_sec, tv.tv_usec);
      speechwPrintStatistics (wp, logfile);

      fclose (logfile);

    } else {
      ds_thread_log (1, "speechw: log file %s could not be opened", logfilename);
      rc = 2;
    } /* endif */

  } else {
    ds_thread_log (1, "speechw: log file not specified");
    rc = 1;
  } /* endif */

  /* Return result in retp */
  if (*retsz >= sizeof (rc)) {
    *retsz = sizeof(rc);
    ALLOC (*retp, int);
    memcpy (*retp, &rc, *retsz);
  } /* endif */
    
  ds_thread_log (10, "speechw: speechwDumpStats completed");

  return (rc);
}

/* speechwGetCID() returns the connection used to communicate with the
   remote Janus (if it exists).  Otherwise, returns CMGR_CID_NONE */
cmgr_id_t speechwGetCID(struct warden *wp, struct viceroy_onode  *op)
{
  cmgr_id_t connection = CMGR_CID_NONE; /* Default is no connection */
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;

  ds_thread_log (10, "speechw: speechwGetCID entered");

  if (pData->pLocations->remotePort >= 0) {
    connection = pData->connections.remoteConnection;
  } /* endif */

  ds_thread_log (10, "speechw: speechwGetCID completed");

  return (connection);
}

#ifdef INSTRUMENTED
/* speechwCalcTVDifference() returns a double that represnets the difference
   between two time value structures. */
double speechwCalcTVDifference (struct timeval* ptv1, struct timeval* ptv2)
{
  struct timeval tv3;
  double interval;

  timersub (ptv2, ptv1, &tv3);
  interval  = (double) tv3.tv_sec + (double) tv3.tv_usec/1000000.0;

  return (interval);
}
#endif

/* speechwPrintStatistics() prints out useful? statistics to the specified
   file.  Right now, it checks to see if the last translation has completed
   successfully, and prints out some information about it.  I might add
   some more data to the output if needed. */
void speechwPrintStatistics (struct warden* wp, FILE* file)
{
#ifdef INSTRUMENTED
  SpeechClientData* pData = (SpeechClientData *) wp->w_data;
  SpeechMeasurements* pMeasurements;
  float ttime; /* Translation time */
  float rtime; /* Recognition time */
  float mtime; /* MEL Conversion time */

  /* Check whehter warden has been initialized */
  if (!pData) {
    fprintf(file, "speech warden not initialized\n");
    return;
  } /* endif */

  pMeasurements = &pData->lastMeasurements;

  fprintf (file, "Last Translation:\n");
  if (pMeasurements->totalBytes == 0) {
    /* No measurements taken yet */
    fprintf (file, "None\n");
    return;
  } /* endif */

  fprintf (file, "Total bytes read: %ld\n", pMeasurements->totalBytes);
  fprintf (file, "Time to read data from client: %f\n", 
	   speechwCalcTVDifference (&pMeasurements->tvConnectToJanus,
				    &pMeasurements->tvBeginTranslation));
  ttime = speechwCalcTVDifference (&pMeasurements->tvBeginTranslation,
				   &pMeasurements->tvEndTranslation);
  fprintf (file, "Total time to recognize the data: %f\n", ttime);

  rtime = speechwCalcTVDifference (&pMeasurements->tvBeginJanusRecognition,
				   &pMeasurements->tvEndJanusRecognition);

  /* Strategy-specific data */
  if (pMeasurements->strategy == SPEECHW_STRATEGY_LOCAL) {
    fprintf (file, "Strategy: Local\n");
    fprintf (file, "Time to initiate RPC and transfer data locally: %f\n",
	     speechwCalcTVDifference (&pMeasurements->tvBeginTranslation,
				     &pMeasurements->tvBeginJanusRecognition));
    fprintf (file, "Time to perform Janus recognition: %f\n", rtime);
    fprintf (file, "Time to complete recognition: %f\n",
	     speechwCalcTVDifference (&pMeasurements->tvEndJanusRecognition,
				      &pMeasurements->tvEndTranslation));
  } else if (pMeasurements->strategy == SPEECHW_STRATEGY_HYBRID) {
    fprintf (file, "Strategy: Hybrid\n");
    fprintf (file, "MEL file size: %ld\n", pMeasurements->MELfilesize);
    fprintf (file, "Time to initiate MEL RPC and transfer data locally: %f\n",
	     speechwCalcTVDifference (&pMeasurements->tvBeginTranslation,
				     &pMeasurements->tvBeginMELConversion));
    fprintf (file, "Time to perform MEL Conversion: %f\n",
	     speechwCalcTVDifference (&pMeasurements->tvBeginMELConversion,
				      &pMeasurements->tvEndMELConversion));
    mtime = speechwCalcTVDifference (&pMeasurements->tvBeginTranslation,
				     &pMeasurements->tvEndMELConversion);
    fprintf (file, "Time to perform Janus recognition: %f\n", rtime);
    fprintf (file, "Time to perform RPC/SFTP: %f\n", ttime - rtime - mtime);
  } else if (pMeasurements->strategy == SPEECHW_STRATEGY_REMOTE) {
    fprintf (file, "Strategy: Remote\n");
    fprintf (file, "Time to perform Janus recognition: %f\n", rtime);
    fprintf (file, "Time to perform RPC/SFTP: %f\n", ttime - rtime);
  } else {
    fprintf (file, "Strategy: Unknown\n");
  } /* endif */

#endif
}
