#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * whiteboard_main.c -- exported interface which viceroy expects
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This code supplies the various hooks which the Viceroy expects in
 * order to be able to use the Whiteboard Warden.
 *
 */

#define WBOARDWARDEN

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>

#include <sys/queue.h>
#include <sys/stat.h>

#include <include/odyssey.h>
#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>
#include <base/ds/ds_rrlist.h>

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/onode-manager.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>

#include "whiteboard.h"
#include "whiteboard_private.h"
#include "whiteboard_tsops.h"
#include "whiteboard_rpc2.h"
#include "whiteboard_server.h"

#include "whiteboard_main.h"

char *peer_name = NULL; /* XXX this is the same hack as for server_name
			   in the web warden. the problem is we don't have
			   a place to squirrel away data during init. !!! */

/*
 * WARDEN METHODS
 *
 * the struct warden effectively provides a method lookup table,
 * which is exported by whiteboard.h
 *
 */

struct warden wboard_warden = {
    WBOARD_TTYPE,		/* type identifier */
    "Whiteboard",		/* name of warden */
    0,				/* flags */

    wboard_parse_arg,		/* parseargs function */
    "Usage: -wboard_peer <server name>",

    wboard_init,		/* various warden functionality */
    wboard_getcid,		/* get connection id */
#ifdef INSTRUMENTED
    wboard_pstats,		/* print statistics */
#else /* INSTRUMENTED */
    (w_pstats_fn) NULL,
#endif /* INSTRUMENTED */
    
    wboard_root,
    wboard_lookup,
    wboard_access,
    wboard_getattr,
    NULL,
    NULL,

    NULL,			/* read/write */
    NULL,			/* rdwr dispose */

    NULL,
    NULL,

    wboard_tsop,		/* tsops */
    wboard_tsop_dispose,

#ifdef SETS
    (w_expand_fn) NULL,
    (w_cursexpand_fn) NULL,
    (w_prefetch_fn) NULL,
#endif

    0,
    0,
    (void *)NULL		/* w_data, which will be set up by
				   wboard_init */
};

/*
 * INITIALIZATION FUNCTIONS
 */

int wboard_parse_arg(int cur, int argc, char *argv[])
{
    /* Takes argc and argv, and inspects arguments from 'cur' onward,
       dealing with those meant for the Web Warden. Returns number of
       arguments dealt with. */

    if(strcmp(argv[cur], "-wboard_peer") == 0) {
	if((cur + 1) >= argc)
	    return(-1);		/* flag, but no argument! */

	if(*(argv[cur + 1]) == '-')
	    return(-1);		/* where we expect a hostname, there
				   appears to be another flag */

	peer_name = strdup(argv[cur + 1]);
	return(2);
    }

    return(0);
}

int wboard_init(struct warden *wp)
{
    /* does necessary initialization, and returns 0 if successful.
       no arguments, since that sort of info presumably was passed in when
       wboard_parseargs was called. */
    struct wboard_data     *my_data;
    cmgr_id_t               cid = -1;
    struct ofid             root_ofid;
    ds_safeq_t             *workq = NULL;
    struct onode_create_in  root_onode_params;
    struct viceroy_onode   *rop;

    /* set up warden specific data. first we allocate the space
       in w_data, then we assign that same memory location to
       a pointer of the proper type. We then use the properly
       cast pointer in order to not have to cast the data every
       time we reference it. */
    ALLOC(wp->w_data, struct wboard_data);
    my_data = (struct wboard_data *)(wp->w_data);
    my_data->wbd_peer = peer_name;
    if (peer_name == NULL) {
	ds_thread_log(1, "wboard_init: no peer specified on command line.");
	goto failedinit;
    }

    peer_name = NULL;

    /* set up safeq for server thread to use */
    workq = ds_safeq_create();
    ASSERT(workq);
    my_data->wbd_workq = workq;

    /* set up server thread */
    ASSERT(ds_thread_create(wboard_server_thread,
			    (char *)workq,
			    DS_THREAD_NORMAL,
			    "wboard_server",
			    0));
    
    /* Set up the RPC2 connection to whiteboard peer. */
    cid = wboard_connect_to_machine(my_data->wbd_peer);
    if(cid < 0) {
	/* XXX want cm to turn errnum into string */
        ds_thread_log(0, "wboard_init: couldn't establish connection "
		         "to peer %s (cm returned %d).",
		      my_data->wbd_peer, cid);
	goto failedinit;	/* couldn't establish connection */
    } else {
	ds_thread_log(0, "wboard_init: established connection %d "
		         "to peer %s", cid, my_data->wbd_peer);
    }
    my_data->wbd_cid             = cid;

    /* set up remainder of webwarden's w_data */
    my_data->wbd_current_uniq    = 1;	/* set up current uniqifier */

    /* ALLOC(my_data->wwd_monitor, ds_hwlock_t);
       ds_hwlock_init(my_data->wwd_monitor); */

    /* set up the root onode in init, since there is only one volume */
    root_ofid.ofid_ttype = WBOARD_TTYPE;
    root_ofid.ofid_tome  = 1;	/* we only support tome #1 for now */
    root_ofid.ofid_vnum  = 0;
    root_ofid.ofid_uniq  = 0;

    root_onode_params.oc_flags = ON_ROOTOFVOL | ON_UNCACHEABLE;
    root_onode_params.oc_vtype = O_VDIR;
    root_onode_params.oc_ofree = NULL;
    root_onode_params.oc_data  = NULL;

    rop = onode_create(&root_ofid, &root_onode_params);
    if (rop == NULL) {
	ds_thread_log(0, "wboard_init: could not create volume onode root!");
	ASSERT(0);
    }

    ds_thread_log(10, "wboard_init: initialized successfully");
    return(0);			/* successful initialization */

failedinit:
    /* initialization failed, so free w_data and return -1. make sure
       to set wp->w_data to NULL so that we can tell later that this
       initialization failed. */
    if (cid >= 0) {
	/* be nice and close the connection */
	cmgr_destroy(cid);
    }
    FREE(wp->w_data);
    wp->w_data = NULL;
    return(-1);
}

/*
 * GENERAL INTERFACE
 */

int wboard_root(struct warden         *wp,
		tome_t                 tome,
		struct viceroy_onode **root_onodep)
{
    struct viceroy_onode *root_onode;
    struct ofid  of;

    if (tome != 1) {
	ds_thread_log(100,"wboard_root: requested tome (%d) not supported",
		      tome);
	return(EINVAL);		/* XXX what's the right error? */
    }

    of.ofid_ttype = WBOARD_TTYPE;
    of.ofid_tome  = tome;
    of.ofid_vnum  = 0;		/* for whiteboard warden, {X,X,0,0} */
    of.ofid_uniq  = 0;		/* indicates an ofid for a volume. */

    root_onode = onode_get(&of);
    if (root_onode == NULL) {
	ds_thread_log(0, "wboard_root: onode_get() failed for warden tome "
		         "%d's root onode.", tome);
	return(EINVAL);
    }

    onode_readlock(root_onode);

    *root_onodep = root_onode;

    /* Life is happy; we've gotten the onode of the root, and we've
       readlocked it, so return a happy return code. */
    return(0);
}

int wboard_lookup(struct warden         *wp,
		  struct viceroy_onode  *odp,
		  struct ody_cred       *cred,
		  char                  *name,
		  char                 **remain,
		  struct viceroy_onode **opp)
{
    /*
     * we don't use the namespace at all, currently.
     */

    /* we don't care about odp */
    if (odp) {			/* KIP: may be called by parsehtml, in
				   which case no parent dir is given */
        onode_release_read(odp);
        onode_put(odp);
    }

    *opp = NULL;
    return EINVAL;
}

int wboard_access(struct warden *wp,
		struct viceroy_onode  *op,
		struct ody_cred  *cred,
		int           mode)
{
    /* for web objects, everyone can read everything.
       this may be re-evaluated later. */
    switch(mode) {
    case O_A_READ:
    case O_A_EXEC:
    case O_A_WRITE:
	return 0;
	/* return EACCES; */
    }
    return EINVAL;
}

int wboard_getattr(struct warden *wp, struct viceroy_onode *op,
		   struct ody_vattr *vap, struct ody_cred *cred)
{
    time_t temp = 0;

    ds_thread_log(10, "wboard: wboard_getattr entered");

    if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0)
      vap->va_type = O_VDIR;	/* this is the root - a directory */
    else
      vap->va_type = O_VREG;
    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;
    vap->va_blocksize = 512;
    vap->va_atime = temp;
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    ds_thread_log(10, "wboard: wboard_getattr completed");

    return (0);
}

/*
 * TSOPS
 */

int wboard_tsop(struct warden *wp,
		int           tsvnum,
		struct viceroy_onode  *op,
		struct ody_cred  *cred,
		int           opnum,
		size_t        argsz,
		void          *argp,
		size_t        *retsz,
		void          **retp)
{
    int result;

    /* check the version number */
    if(tsvnum != WBOARD_VERSION) {
	return(ODY_EBADVERSION);
    }

    /* dispatch TSOPs, based on opnum */
    switch(opnum) {
    case WBOARD_SENDPOINTS:
	result = wboard_send_points(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WBOARD_GETPOINTS:
	result = wboard_get_points(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    }

    return(EOPNOTSUPP);
}

int wboard_tsop_dispose(struct warden *wp,
		      struct viceroy_onode  *op,
		      struct ody_cred  *cred,
		      int           opnum,
		      size_t        argsz,
		      void          *argp,
		      size_t        retsz,
		      void          *retp)
{
    /* dispose of *retp, since we were responsible for allocating it */
    FREE(retp);			/* FREE() checks if retp is NULL */
    return(0);
}

cmgr_id_t wboard_getcid(struct warden *wp, struct viceroy_onode  *op)
{
    struct wboard_data  *my_data = (struct wboard_data *) wp->w_data;
    return my_data->wbd_cid;
}

/*
 * OTHER
 */

#ifdef INSTRUMENTED

void wboard_pstats(struct warden *wp, FILE *statout)
{
}

#endif /* INSTRUMENTED */
