#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_cache.c  --  written by Kip Walker, Summer 1997
 *
 *  Some explanatory notes are in order:
 *   - viceroy_onodes are added to the cache on creation (lookup) even though
 *     there is no data associated with them.  prevent creation of multiple
 *     viceroy_onodes for the same object if VOP_OPEN is delayed
 *   - if no entries in the cache are reuseable, ENOSPC is returned --
 *     currently SETS will probably break when this happens.
 *   - the cache is a list of data objects -- struct page_data *'s.  There
 *     are zero or more viceroy_onodes pointing to each, a refcnt++ for each.
 *     A final refcnt is held to keep the page_data in the cache until it
 *     is kicked out.
 *   - the wdb_state indicates if the data can be referenced through a
 *     !ON_UNCACHEABLE viceroy_onode
 *   - this stuff will, of course, probably end up in a more general
 *     implementation to be shared with other wardens and deal with cache
 *     space as a resource (note that currently the reserved area is
 *     a number of files, not blocks!)
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/queue.h>
#include <sys/uio.h>

#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_rrlist.h>

#include <include/viceroy_onode.h>
#include <include/onode-manager.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>
#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_cache.h"

/* ought these to be in the w_data? at least they're static... */
static struct webw_cache_head webw_cache;
static ds_hwlock_t webw_cache_lock;
static struct webw_cache_head webw_cache_free;
static ds_hwlock_t webw_cache_free_lock;
static int webw_cache_size = 0;

/* webw_cache_init
 *   supplied with a list of cache_entries -- allocated by the viceroy.
 *   Put them on the cache list.
 */
void webw_cache_init(struct cache_entry *cache_list)
{
    TAILQ_INIT(&webw_cache);	/* init queue of webw onode ptrs */
    ds_hwlock_init(&webw_cache_lock);
    TAILQ_INIT(&webw_cache_free); /* init queue of webw onode ptrs */
    ds_hwlock_init(&webw_cache_free_lock);

    webw_cache_add_cfiles(cache_list);
}

/* webw_cache_add_cfiles
 *   add container files (centries) to our list
 */
void webw_cache_add_cfiles(struct cache_entry *cache_list)
{
    struct cache_entry *next_centry;

    ds_hwlock_write(&webw_cache_free_lock);
    while (cache_list) {
	next_centry = cache_list->ce_link.tqe_next;
	TAILQ_INSERT_HEAD(&webw_cache_free, cache_list, ce_link);
	webw_cache_size++;
	cache_list = next_centry;
    }
    ds_hwlock_release_write(&webw_cache_free_lock);
}

/* webw_cache_lookup
 *   search for given url in the onode cache.  Return 1 for cache hit 
 *   and set *c_entry, return 0 otherwise.
 *   XXX Any reason _not_ to reference the data here? XXX
 */
int webw_cache_lookup(char *url, struct cache_entry **c_entry)
{
    struct cache_entry *cached_op;
    struct webw_page_data   *wpdata;
    char                    *key;

    /* start with a read lock to allow more interleaving */
    ds_hwlock_read(&webw_cache_lock);
    cached_op = webw_cache.tqh_first;

    /* iterate through the list */
    while (cached_op != NULL) {
	int diff;

	key = (char *)cached_op->ce_key;
	diff = strcmp(url, key);
	if (!diff) {
	    ds_thread_log(10, "hit %s", url);

	    /* Grab a reference */
	    wpdata = (struct webw_page_data *)cached_op->ce_data;
	    ds_lock_write(&(wpdata->wpd_data_lock));
	    wpdata->wpd_refcnt++;
	    ds_lock_release(&(wpdata->wpd_data_lock));
	    
	    /* mark it MRU */
	    ds_hwlock_write(&webw_cache_lock);
	    TAILQ_REMOVE(&webw_cache, cached_op, ce_link);
	    TAILQ_INSERT_TAIL(&webw_cache, cached_op, ce_link);
	    ds_hwlock_release_write(&webw_cache_lock);
	    ds_hwlock_release_read(&webw_cache_lock);
	    *c_entry = cached_op;

	    return (1);
	}
	cached_op = cached_op->ce_link.tqe_next;
    }
    ds_hwlock_release_read(&webw_cache_lock);    
    
    return (0);
}

/* webw_cache_enter 
 *   Add "new_onode" to the onode cache if room.  If necessary,
 *   recycle an unreferenced onode return 1 for success, 0 for failure
 *   to find purgeable onode.
 *   Assumed to be called with a reference for the cache.
 */
struct cache_entry *webw_cache_enter(char *key,
				     struct viceroy_onode *op,
				     int flags,
				     struct webw_page_data *wpdata)
{
    struct cache_entry *cached_op=NULL;

    /* Currently adding cursor onodes and regulars to cache */
    /* allocate a new node, or take the LRU (available) onode */
    ds_hwlock_write(&webw_cache_lock);
    if (webw_cache_free.tqh_first != NULL) {
	ds_hwlock_write(&webw_cache_free_lock);
	cached_op = webw_cache_free.tqh_first;
	TAILQ_REMOVE(&webw_cache_free, cached_op, ce_link);
	ds_hwlock_release_write(&webw_cache_free_lock);
    } else {
	/* search for an onode to reuse */
	/* XXX might also want to make a request for more space and/or
           just create some new cfiles XXX */
	cached_op = webw_cache.tqh_first;

	/* look for a cache entry with only one reference */
	while (cached_op &&
	       ((struct webw_page_data *)cached_op->ce_data)->wpd_refcnt > 1)
	    cached_op = cached_op->ce_link.tqe_next;
	
	if (!cached_op) {
	    /* no purgeable onodes! */
	    ds_hwlock_release_write(&webw_cache_lock);
	    return 0;
	} else {
	    ds_thread_log(0, "cache: %s -> %s", (char *)cached_op->ce_key,
			  wpdata->wpd_url);
	    TAILQ_REMOVE(&webw_cache, cached_op, ce_link);
	    webw_release_page_data(cached_op->ce_data);
	    FREE(((struct webw_page_data *)cached_op->ce_data)->wpd_url);
	}
    }
    TAILQ_INSERT_TAIL(&webw_cache, cached_op, ce_link);

    cached_op->ce_key = (void *)key;
    cached_op->ce_flags = flags;
    cached_op->ce_data = (void *)wpdata;

    /* Can I release this lock here without consequences? */
    ds_hwlock_release_write(&webw_cache_lock);

    /* grab a reference for the cache */
    /* XXX don't do this with the cache_lock held!! probably safe,
       since the onode already has a ref, and is locked. XXX */
    ds_lock_write(&(wpdata->wpd_data_lock));
    wpdata->wpd_refcnt++;
    wpdata->wpd_centry = cached_op;
    ds_lock_release(&(wpdata->wpd_data_lock));

    /* should I let the caller do this? */
    ((struct webw_onode_data *)op->o_data)->wod_centry = cached_op;

    wpdata->wpd_state = VIRGIN;

    /* point at the container file */
    op->o_cachefile = cached_op->ce_fname;
    /* XXX consider filling in onode/device here for simplicity */

    ds_thread_log(5, "inserted %s", (char *)cached_op->ce_key);

    return (cached_op);
}

/*
 * webw_cache_readdir -- support readdir by filling dirents
 *                       Note: this is a hack, and may produce inaccurate
 *                       results.  In particular, if offset is > 0,
 *                       the cache contents/order may have changed and
 *                       conflict with previous readdir yielded dirents.
 */
int webw_cache_readdir(int offset, int count, void *data)
{
    struct cache_entry *cache_entry;
    struct ody_dirent *cur_dentry = (struct ody_dirent *)data;
    int size = sizeof(struct ody_dirent)-(ODY_MAXNAMLEN+1);
    int bytes_returned = 0;

    ds_hwlock_write(&webw_cache_lock);    
    cache_entry = webw_cache.tqh_first;
    while (cache_entry && offset) {
	cache_entry = cache_entry->ce_link.tqe_next;
	offset--;
    }
    while (cache_entry) {
	int namelen, reclen;
	/* NOTE: there will be cache entries for pages looked-up but not
	 * opened.  We can't help this, since the cache_entry is created 
	 * on lookup to avoid race conditions. */
	if (((struct webw_page_data *)cache_entry->ce_data)->wpd_data_avail == 0) {
	    /* skip objects with empty cache files */
	    cache_entry = cache_entry->ce_link.tqe_next;
	    continue;
	}
	namelen = strlen(cache_entry->ce_key);
	namelen = MIN(ODY_MAXNAMLEN, namelen);
	reclen = size + ((namelen+1 + 3) &~ 3);
	if (count >= reclen) {
	    char *c = cur_dentry->d_name;

	    cur_dentry->d_fileno = atoi(&cache_entry->ce_fname[1])+1;
	    cur_dentry->d_type = ODT_REG;
	    strncpy(cur_dentry->d_name, cache_entry->ce_key, ODY_MAXNAMLEN);
	    while (*c)
		if (*c == '/')
		    *c++ = '\\';
		else
		    c++;
	    cur_dentry->d_namlen = namelen;
	    cur_dentry->d_reclen = reclen;
		
	    count -= reclen;
	    bytes_returned += reclen;
	    (char *)cur_dentry += reclen;

	    cache_entry = cache_entry->ce_link.tqe_next;
	} else
	    break;
    }
    ds_hwlock_release_write(&webw_cache_lock);    

    return bytes_returned;
}

/*
 * webw_cache_flush -- flush all cached pages.  pages currently opened
 *                     are marked as FLUSHED, and the final webw_close
 *                     will complete the flush.
 */
int webw_cache_flush(void)
{
    struct cache_entry *cached_op, *next_cop;
    struct webw_page_data *wpd;
    int count = 0;

    ds_hwlock_write(&webw_cache_lock);
    
    cached_op = webw_cache.tqh_first;
    /* iterate through the list */
    while (cached_op != NULL) {
	wpd = (struct webw_page_data *)cached_op->ce_data;
	next_cop = cached_op->ce_link.tqe_next;

	if (wpd->wpd_state == VIRGIN) {
	    cached_op = next_cop;
	    continue;
	}
	if (wpd->wpd_openers == 0 && wpd->wpd_state >= FETCHED) {
	    webw_flush_page(wpd);
	} else {
	    /* mark it for destruction after last reference */
	    /* also defer when the data is still being fetched */
	    cached_op->ce_flags |= CE_FLUSHED;
	}
	cached_op = next_cop;
	count++;
    }

    ds_hwlock_release_write(&webw_cache_lock);

    ds_thread_log(0, "webw: flushed %d cache entries", count);

    return(0);
}

/* webw_dump_dbuf
 *   flush the dbuf list "chain" to the file in "fd".  Return -1 for
 *   error, 0 for success.
 */
int webw_dump_dbuf(struct webw_dbuf *chain, int fd)
{
    struct webw_dbuf *curr = chain;
    struct iovec *iova;
    size_t rc;
    int iovcnt = 0, i;

    if (!chain)
	return -1;
    
    while (curr) {
	iovcnt++;
	curr = curr->wdb_next;
    }
    
    curr = chain;
    iova = (struct iovec *)malloc(iovcnt * sizeof(struct iovec));
    for (i=0; i<iovcnt; i++) {
	iova[i].iov_base = curr->wdb_data;
	iova[i].iov_len  = curr->wdb_length;
	curr = curr->wdb_next;
    }
    
    rc = writev(fd, iova, iovcnt);
    
    free(iova);
    
    return (rc == -1) ? -1 : 0;
}

/* webw_print_centries
 *   for debugging, print url and container file names for all cached 
 *   webw onodes
 */
void webw_print_centries()
{
    struct cache_entry *cached_op;

    cached_op = webw_cache.tqh_first;
    while (cached_op) {
	ds_thread_log(0, "pg %8x cf %s url %s %d refs", cached_op->ce_data, 
		      cached_op->ce_fname, (char *)cached_op->ce_key,
		      ((struct webw_page_data *)cached_op->ce_data)->wpd_refcnt);
	cached_op = cached_op->ce_link.tqe_next;
    }
}

/* webw_print_centries_gdb
 *   for debugging, print url and container file names for all cached 
 *   webw onodes  (don't assume thread context)
 */
void webw_print_centries_gdb()
{
    struct cache_entry *cached_op;

    cached_op = webw_cache.tqh_first;
    while (cached_op) {
	fprintf(stderr, "pg %8lx cf %s url %s %d refs st=%d\n",
		(long)cached_op->ce_data, 
		cached_op->ce_fname, (char *)cached_op->ce_key,
		((struct webw_page_data *)cached_op->ce_data)->wpd_refcnt,
		((struct webw_page_data *)cached_op->ce_data)->wpd_state);
	cached_op = cached_op->ce_link.tqe_next;
    }
}
