#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_comm.c -- functions which talk to the server
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/queue.h> 
#include <sys/stat.h>

#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_log.h>
#include <base/ds/ds_rrlist.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>

#include <include/odyssey.h>
#include <kernel/vfs/includes/ody.h>
#include <include/viceroy_onode.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/onode-manager.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>

#include "distiller.h"
#include "webwarden.h"
#include "webwarden_onode.h"
#include "webwarden_worker.h"
#include "webwarden_comm.h"
#include "webwarden_private.h"

#ifdef INSTRUMENTED
#include <ds_histo.h>
#include "webwarden_instr.h"
#endif /* INSTRUMENTED */

#include "odyssey/viceroy/rsrc_en.h" /* XXXAPM */

extern int do_caching;

/*
 * webw_fetch_url -- opens a request for an HTTP object with the server
 *                   (by the object's URL), and requests a thread to
 *                   read the URL data. Returns a filename if there
 *                   is an error (where the file contains the error
 *                   message, and is suitable for giving to the minicache),
 *                   or NULL if the object was successfully opened
 *                   and is being streamed.  Return code signifies whether
 *                   a re-fetch occured.
 */

/* Possible states of the onode:
 *   new: no status, headers, dbufs
 *   expired: cleaned -- dates set
 *   cond getting: all set, data done
 */
int webw_fetch_url(cmgr_id_t               cid,     /* connection to
						       server */
		   struct webw_onode_data *wod,     /* webwarden-specific
						       onode data*/
		   webw_quality_t          img_quality,  /* the minimum
							    image quality
							    we're interested
							    in */
		   webw_quality_t          html_quality, /* ditto for HTML */
		   ds_safeq_t             *workq,   /* safe queue for
						       sending messages to
						       worker threads */
		   struct ofid            *ofp,     /* the ofid for this
						       onode */
		   char                  **cfile_name)
{
    long                        rc;
    SWWW_method_t               the_method;
    int                         status_code, header_size;
    int                         reqid;
    cmgr_connection_t           *connection;
    RPC2_BoundedBS              status, headers;
    struct webw_worker_request *req;
#ifdef INSTRUMENTED
    struct timeval              start;
#endif /* INSTRUMENTED */
    struct cache_entry         *cached_op = wod->wod_centry;
    struct webw_page_data      *wpdata = wod->wod_page_data;
    SWWW_headers_t             *reply_hdrs;
#if defined(SETS) && defined(BOUNCE_URLS)
    char                       *target_url = NULL; /* for 30[12] redirects */
#endif
#ifdef USE_LAST_MOD
    struct tm                   hdr_time;
    char                       *cur_header;
#endif
    time_t                      time_served = 0;

    /*
     * Setup IN parameters for the RPC, and then invoke it
     */
    
    ds_thread_log(100, "webw_fetch_url: trying to fetch %s", wpdata->wpd_url);
    
    /* set HTTP method to use */
    the_method = GET;		/* XXX eventually to be based on
				   flags */
    
    /* Set up IN OUT status parameter. Allow for 1K status line (which should
       be more than sufficient) */
    status.MaxSeqLen  = 1024;
    status.SeqLen     = 0;
    /* status.SeqBody    = (RPC2_ByteSeq) malloc(1024); */ /* XXX get rid of */
    NALLOC(status.SeqBody, RPC2_Byte, 1024);
    
    /* Set up IN OUT headers parameter. Allow for 4K worth of headers (again,
       this should be more than sufficient) */
    headers.MaxSeqLen = 4092;
    headers.SeqLen    = 0;
    /* headers.SeqBody   = (RPC2_ByteSeq) malloc(4192); */ /* XXX get rid of */
    NALLOC(headers.SeqBody, RPC2_Byte, 4192);

    ds_thread_log(1000, "webw_fetch_url: fetching with quality %d",
		  img_quality);
    
#ifdef INSTRUMENTED
    gettimeofday(&start, NULL);	/* get start time */
#endif /* INSTRUMENTED */

    ds_thread_log(3, "fetching %s", wpdata->wpd_url);

    if (!(cmgr_validate(cid) > 0)) {
	rc = RPC2_NOCONNECTION;
    } else {
	connection = cmgr_get(cid, CMGR_LOCK);
	ASSERT(connection);
	/* perform the rpc */
	ds_thread_log(1, "open_url on %s\n", wpdata->wpd_url);
	rc = open_url(                 CMGR_RPC2_HANDLE(connection),
				       (RPC2_Integer)     the_method,
				       (RPC2_String)      wpdata->wpd_url,
				       (RPC2_Integer)     img_quality,
				       (RPC2_Integer)     html_quality,
				       (RPC2_Integer)     wpdata->wpd_modified,
				       (RPC2_Integer *)   &reqid,
				       (RPC2_Integer *)   &status_code,
				       (RPC2_BoundedBS *) &status,
				       (RPC2_Integer *)   &header_size,
				       (RPC2_BoundedBS *) &headers);
	ds_thread_log(10, "open_url gives: %d\n", rc);
	cmgr_put(cid, CMGR_LOCK);
    }
    
    if(rc != RPC2_SUCCESS) {
	int   fd;
	char  tempfilename[ODY_MAXPATHLEN] = "/tmp/wwwtemp.XXXXX";
	char *error;
	
	/* XXX does this failure mean close_url is not necessary? */

	mkstemp(tempfilename);		/* get a filename */
	
	if (rc == ODYWEB_NODOC) {
	    ds_thread_log(0,
			  "webw_fetch_url: Remote fetch_url denied by server");
	    error = "The distillation server refused to serve that URL.";
	} else {
	    ds_thread_log(0, "webw_fetch_url: Remote fetch_url failed (%s)",
			  RPC2_ErrorMsg(rc));
	    
	    if (rc < RPC2_FLIMIT) {
		ds_thread_log(1, "webw_fetch_url: RPC2 failure was fatal.");
		error = "The connection is temporarily unavailable "
		    "(perhaps the network is unavailable).\n";
	    } else {
		/* connection was not fatal */
		error = "There was an internal error; Web Warden is "
		    "not able to fetch the requested document.\n";
	    }
	    
	    cmgr_markdown(cid);
	}

	FREE(headers.SeqBody);
	FREE(status.SeqBody);

	/* Might as well point to the cachefile already on disk. */
	if (wpdata->wpd_data_done) {
	    *cfile_name = cached_op->ce_fname;
	    return 0;
	}

	fd = open(tempfilename, O_WRONLY | O_CREAT | O_TRUNC,
		                S_IRWXU | S_IRGRP | S_IROTH);

	write(fd, error, strlen(error));
	close(fd);

	/* If this was a conditional re-get, may have to free
	 * these. */
	webw_release_page_data(wpdata);

	wpdata->wpd_status_code  = 500;
	wpdata->wpd_status_line  = strdup("HTTP/1.0 500 Internal Warden Error");
	wpdata->wpd_header_size  = 0;
	wpdata->wpd_header_bytes = 0;
	wpdata->wpd_headers      = NULL;
	wpdata->wpd_state        = CACHED;

	/* This might be a prefetch, in which case we need to tell him
	 * not to thread_wait for a worker to fetch the page. */
	wpdata->wpd_data_done    = 1;
	/* XXX This shouldn't be necessary -- sleeps are executed by
           the caller when "w_open" returns. XXX */
	ds_thread_wakeup((char *) &(wpdata->wpd_data_done));

	*cfile_name = strdup(tempfilename);
	return 0;
    }

    /* XXX might want to check content-length field to interact with
       the cache manager */

    reply_hdrs = SWWW_unpack_headers(-1, (char *)headers.SeqBody);

#if defined(SETS) && defined(BOUNCE_URLS)
    /* NOTE: Bouncing automatically at this point is bad in that
     * Netscape still thinks the URL for the page is that of the
     * "trigger" page, so that relative URLs in the target page will
     * be requested relative to the trigger page.  Similarly, we don't
     * want to make the trigger/target pages aliases, even when both
     * are fetched.  (XXX Any other reason to keep aliases?) */

    /* If this is a demand fetch, just do the normal thing.  If we
     * are attempting to prefetch, signal the worker that it should
     * prefetch the target url as well. */
    if ((status_code == 301 || status_code == 302) &&
	wod->wod_prefetching != NORMAL) {
	/* Find the Location header */
	target_url = SWWW_gethvalue(reply_hdrs, "location");
	/* check that target is given and doesn't pathologically point
           back to this page */
	if (target_url && strcmp(wpdata->wpd_url, target_url)) {
	    char *remain;

	    ds_thread_log(0, "%s bounces to %s", wpdata->wpd_url,
			  target_url);
	    rc = ((wpdata->wpd_wp)->w_lookup)(wpdata->wpd_wp, NULL,
					      NULL, target_url, &remain,
					      &wpdata->wpd_target);
	    
	    if (!rc) {
		onode_put(wpdata->wpd_target);
		onode_release_read(wpdata->wpd_target);
	    }
	}
    }
#endif

    if (status_code == 304) {
	/* Our conditional get came back "negative" -- can use the
	 * cached version. */

	/* a little sanity check */
	ASSERT(wpdata->wpd_data_done);

	ds_thread_log(3, "conditional get!");

	/* XXX should update any current headers with values in the
	   new ones.  For now, I just toss the new ones. (on exit) */

	if (cmgr_isvalid(cid) > 0) {
	    connection = cmgr_get(cid, CMGR_LOCK);
	    ASSERT(connection);
	    rc = close_url(CMGR_RPC2_HANDLE(connection), reqid);
	    cmgr_put(cid, CMGR_LOCK);
	} else {
	    ds_thread_log(1, "couldn't close url (no conn.)");
	    rc = RPC2_NOCONNECTION;
	}
	
	if(rc != RPC2_SUCCESS) {
	    ds_thread_log(10, "close_url failed (%s)", RPC2_ErrorMsg(rc));
	    
	    if (rc < RPC2_FLIMIT) {
		ds_thread_log(1, "RPC2 failure was fatal.");
	    }
	    
	    cmgr_markdown(cid);
	}

	goto keep_old_copy;
    } else {
	time_t old_wpd_modified = wpdata->wpd_modified;

	/* still have a lock on the onode */

	/* XXX could I use headerlists here? */

#ifdef USE_LAST_MOD
	/* Find an expiration date, last modified date */
	cur_header = SWWW_gethvalue(reply_hdrs, "last-modified");
	if (cur_header) {
	    SWWW_parse_date(cur_header, &hdr_time, 0, 1, 0);
	    hdr_time.tm_isdst = -1;	/* force check of DST */
	    wpdata->wpd_modified = mktime(&hdr_time);
	    ds_thread_log(10, "lm: %s", cur_header);
	}
	cur_header = SWWW_gethvalue(reply_hdrs, "expires");
	if (cur_header) {
	    SWWW_parse_date(cur_header, &hdr_time, 0, 0, 0);
	    hdr_time.tm_isdst = -1;	/* force check of DST */
	    wpdata->wpd_expires = mktime(&hdr_time);
	}
	cur_header = SWWW_gethvalue(reply_hdrs, "date");
	if (cur_header) {
		SWWW_parse_date(cur_header, &hdr_time, 0, 1, 0);
		hdr_time.tm_isdst = -1;	/* force check of DST */
		time_served = mktime(&hdr_time);
	}

#ifdef DONT_DATE_PAGES
	wpdata->wpd_expires = 0;
	wpdata->wpd_modified = 0;
#endif

	ds_thread_log(3, "old: %d new: %d", 
		      old_wpd_modified, wpdata->wpd_modified);
#endif

	/* Observed Netscape behavior, based on reply headers:
	   a) rep: Date, Last_Mod | req: If_Mod/Last_Mod, pragma no-cache 
	   b) rep: Date == Expires | req: <nothing> 
	   c) rep: Date | req: pragma no-cache
	   d) rep: Expires 0 | req: pragma no-cache
	   e) rep: Date < Expires | req: no-cache
	   */
	if (wpdata->wpd_expires && time_served == wpdata->wpd_expires) {
	    /* Case (b).  This must not be cached.  For simplicity, I
	     leave it in the cache, but mark it invalid so future
	     lookups will not hit. */
	    /* Do we really want to prevent disconnect hits from
               seeing this page? */

	    /* if this is a re-fetch, free the old headers */
	    FREE(wpdata->wpd_status_line);
	    FREE(wpdata->wpd_headers);

	    wod->wod_centry->ce_flags &= ~WCELEM_VALID;
	}
#define MOCK_COND_GET
#ifdef MOCK_COND_GET
	else if (old_wpd_modified && 
		 ((unsigned)wpdata->wpd_modified <= (unsigned) old_wpd_modified)) {
	    /* although the server didn't allow conditional get,
	     * this page isn't any newer.  Use the old data, and tell
	     * the server to quit the fetch. */
	    
	    ds_thread_log(3, "mock conditional get!");
	    
	    /* Do I use any of the header fields?  Anything to do? */
	    /* Need to do a data_done wakeup, since no worker will be
	     * doing it */

	    if (cmgr_isvalid(cid) > 0) {
		connection = cmgr_get(cid, CMGR_LOCK);
		ASSERT(connection);
		rc = close_url(CMGR_RPC2_HANDLE(connection), reqid);
		cmgr_put(cid, CMGR_LOCK);
	    } else {
		ds_thread_log(1, "couldn't close url (no conn.)");
		rc = RPC2_NOCONNECTION;
	    }
	    
	    if(rc != RPC2_SUCCESS) {
		ds_thread_log(10, "close_url failed (%s)", RPC2_ErrorMsg(rc));
		
		if (rc < RPC2_FLIMIT) {
		    ds_thread_log(1, "RPC2 failure was fatal.");
		}
		
		cmgr_markdown(cid);
	    }

	    goto keep_old_copy;
	}
#endif

	if (wpdata->wpd_data_done) {
	    /* Note: cursor onodes should not get to this point, since
	     * a re-expand of a cursor yields a new onode. XXX???XXX? */

	    /* If we get here, we must be replacing the data for an
	     * out-of-date page.  Trash the data so the worker isn't
	     * confused. */
	    webw_clean_page((struct webw_page_data *)cached_op->ce_data);
	    ((struct webw_page_data *)cached_op->ce_data)->wpd_refcnt++;
	    ds_thread_log(3, "%s outofdate", wpdata->wpd_url);
	}

	/*
	 * Deal with results of RPC, including storing object metadata
	 * We place the web object's metadata into the onode. This will
	 * all be freed whenever the onode is dealloc'd.
	 */
	
	wpdata->wpd_status_code  = status_code;
	wpdata->wpd_status_line  = status.SeqBody;
	wpdata->wpd_header_size  = header_size;
	wpdata->wpd_header_bytes = headers.SeqLen;
	wpdata->wpd_headers      = headers.SeqBody;
	
	ds_thread_log(1000, "webw_fetch_url: open succeeded with "
		      "status %d and %d headers (of length %d)",
		      status_code, header_size, headers.SeqLen);

	ALLOC(req, struct webw_worker_request);
	req->wwr_op = onode_get(ofp);
	if (req->wwr_op == NULL)
	    ASSERT(0);		/* this should never happen,
				   but just in case */
	req->wwr_cid        = cid;
	req->wwr_reqid      = reqid;
	req->wwr_reqtype    = WEBW_REQ_FETCH;
#ifdef INSTRUMENTED
	req->wwr_fetchstart = start;
#endif /* INSTRUMENTED */
	ds_safeq_enq(workq, req);
	
	/* only return the container file name if the file is finished
	 * being fetched. */
	SWWW_free_headerlist(reply_hdrs);
	if (do_caching && wpdata->wpd_data_done) {
	    ASSERT(cached_op);
	    *cfile_name = cached_op->ce_fname;
	} else {
	    *cfile_name = NULL;
	}
	return 1;
    }

keep_old_copy:
    FREE(status.SeqBody);
    FREE(headers.SeqBody);
    
    /* reset state (demoted to OPENED by webw_open) */
    wpdata->wpd_state = wpdata->wpd_old_state;
    
    /* no need to wakeup, since folks know to check data_done before
       blocking */
    SWWW_free_headerlist(reply_hdrs);
    *cfile_name = (wpdata->wpd_state >= ON_DISK ? wod->wod_centry->ce_fname : 0);
    return 0;
}

/*
 * webw_fetch_if_needed --
 *
 * this code used to be in webw_open; it's the core of the logic
 * which decides whether a refetch should be attempted.
 *
 *  wp         -- the warden data
 *  op         -- the onode in question
 *  force_cget -- if 1, force a conditional get regardless
 *   XXX do we ever get here if we DON'T want to force_cget? XXX
 */
int webw_fetch_if_needed(struct warden *wp,
			 struct viceroy_onode  *op,
			 int            force_cget,
			 char         **cfile_name)
{
    int                     result    = 0;

    struct webw_data       *my_data   = (struct webw_data *) wp->w_data;
    struct webw_onode_data *the_odata = (struct webw_onode_data *) op->o_data;
    struct webw_page_data  *wpdata    = the_odata->wod_page_data;

#ifdef INSTRUMENTED
    /* XXX should do something with HTML quality, but a) the experiments
       currently don't use it, and b) the cellophane currently keeps
       HTML and image quality the same (if HTML distillation is used) */
    UpdateHisto(&(my_data->wwd_histos->wh_fidelities),
		my_data->wwd_img_quality);
#endif /* INSTRUMENTED */
    
    *cfile_name = NULL;

    if (wpdata->wpd_state == OPENED) {
	/* someone is just ahead of us trying to get the headers on
           this page.  Just return (indicating that current copy is in
           memory, not disk). */
	return 0;
    }

    /* If we have headers for this thing, see if we can use the data
     * already here. */
    if (wpdata->wpd_state > OPENED && !force_cget) {
#ifdef USE_LAST_MOD
	time_t now;
	struct tm *ftm;
	
	/* Check conditions where re-fetch is PROHIBITED */
	if (wpdata->wpd_refcnt > 2 ||
	    !cmgr_isvalid(my_data->wwd_cid)) {
	    return 0;
	}
	
	time(&now);
	ftm = gmtime(&now);
	now = mktime(ftm);
	
	/* XXX perhaps if status_code is 500, should check if the
	 * connection is up now and throw away this old cached error
	 * message. */
	
	if ((wpdata->wpd_expires && now > wpdata->wpd_expires) ||
	    !(the_odata->wod_centry->ce_flags & WCELEM_VALID)) {
	    /* Toss the damn data */
	    webw_clean_page(wpdata);
	    wpdata->wpd_refcnt++;
	    
	    /* reset other fields */
	    wpdata->wpd_expires = 0;
	    wpdata->wpd_modified = 0;
#ifdef SETS
	    the_odata->wod_url_list->cur_dbuf = &wpdata->wpd_data;
#endif
	    
	    ds_thread_log(5, "%s expired", wpdata->wpd_url);
	} else if (/* cache copy OK || */ 0) {
	    *cfile_name = (wpdata->wpd_state >= ON_DISK ? the_odata->wod_centry->ce_fname : 0);
	    return 0;
	}
#else
	/* the data has already been fetched.  we aren't trying cond.
	   gets, so just return the thing we have. */
	*cfile_name = (wpdata->wpd_state >= ON_DISK ? the_odata->wod_centry->ce_fname : 0);
	return 0;
#endif
    }
    
    wpdata->wpd_old_state = wpdata->wpd_state;
    wpdata->wpd_state = OPENED;
#ifdef SETS
    /* send prefetch requests to lower priority workers */
    if (the_odata->wod_prefetching != NORMAL)
	result = webw_fetch_url(my_data->wwd_cid,
				(struct webw_onode_data *) op->o_data,
				my_data->wwd_img_quality,
				my_data->wwd_html_quality,
				my_data->wwd_prefq,
				&(op->o_fid),
				cfile_name);
    else
#endif
	result = webw_fetch_url(my_data->wwd_cid,
				(struct webw_onode_data *) op->o_data,
				my_data->wwd_img_quality,
				my_data->wwd_html_quality,
				my_data->wwd_workq,
				&(op->o_fid),
				cfile_name);
    
    return result;
}
