#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_main.c -- exported interface which viceroy expects
 *
 * Written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This code supplies the various hooks which the Viceroy expects in
 * order to be able to use the Web Warden.
 *
 */

#define WEBWARDEN

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>

#include <sys/queue.h>
#include <sys/stat.h>

#include <include/odyssey.h>
#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>
#include <base/ds/ds_rrlist.h>

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/onode-manager.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>

#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_comm.h"
#include "webwarden_rpc2.h"
#include "webwarden_tsops.h"
#include "webwarden_worker.h"
#ifdef SETS
#include "webwarden_sets.h"
#endif
#include "webwarden_cache.h"
#ifdef INSTRUMENTED
#include <ds_histo.h>
#include "webwarden_instr.h"
#endif /* INSTRUMENTED */

#include "webwarden_main.h"

/*
 * variables of universal interest to the webwarden
 */

char *server_name = NULL;	/* XXX this is a hack so that
				   webw_parse_arg() can do something
				   useful with the server name.
				   at some point we might have
				   multiple instances of web warden
				   code, at which point this will break. */

/* XXX these should probably be static and have webw_init set flags in
 * my_data->flags or something */

/* do_caching note -- the critical code to disable is:
      cache_lookup in webw_lookup/webw_expand
      cache_enter in webw_lookup/webw_expand
      webw_dump_dbuf in webw_worker_fetch
        (this prevents ON_DISK being set, which then avoids other code)
*/
int do_caching = 1;		/* caching on by default */
int prefetch   = 1;		/* prefetching on by default */

/*
 * WARDEN METHODS
 *
 * the struct warden effectively provides a method lookup table,
 * which is exported by webwarden.h
 *
 */

struct warden web_warden = {
    WEB_TTYPE,			/* type identifier */
    "Web",			/* name of warden */
    0,				/* flags */

    webw_parse_arg,		          /* parseargs function */
    "Usage: -webw_server <server name> [-webw_nocache] [-webw_nopf]",

    webw_init,			/* various warden functionality */
    webw_getcid,                /* get connection id */
#ifdef INSTRUMENTED
    webw_pstats,		/* print statistics */
#else /* INSTRUMENTED */
    (w_pstats_fn) NULL,
#endif /* INSTRUMENTED */
    
    webw_root,
    webw_lookup,
    webw_access,
    webw_getattr,
    webw_open,
    webw_close,

    webw_rdwr,			/* read/write */
    webw_rdwr_dispose,		/* rdwr dispose */

    webw_readdir,
    (w_readdir_dispose_fn) NULL, /* since we never allocate our own bufs for
				    readdir */

    webw_tsop,			/* tsops */
    webw_tsop_dispose,

#ifdef SETS
    (w_expand_fn) webw_expand,		/* sets expand ops */
    (w_cursexpand_fn) webw_cursexpand,
    (w_prefetch_fn) webw_prefetch,
#endif

    0,
    0,
    (void *)NULL		/* w_data, which will be set up by
				   webw_init */
};

/*
 * INITIALIZATION FUNCTIONS
 */

int webw_parse_arg(int cur, int argc, char *argv[])
{
    /* Takes argc and argv, and inspects arguments from 'cur' onward,
       dealing with those meant for the Web Warden. Returns number of
       arguments dealt with. */

    if(strcmp(argv[cur], "-webw_server") == 0) {
	if((cur + 1) >= argc)
	    return(-1);		/* flag, but no argument! */

	if(*(argv[cur + 1]) == '-')
	    return(-1);		/* where we expect a hostname, there
				   appears to be another flag */

	server_name = strdup(argv[cur + 1]);
	return(2);
    } else if (strcmp(argv[cur], "-webw_nocache") == 0) {
        do_caching = 0;
	return(1);
    } else if (strcmp(argv[cur], "-webw_nopf") == 0) {
        prefetch = 0;
	return(1);
    }

    return(0);
}

int webw_init(struct warden *wp)
{
    /* does necessary initialization, and returns 0 if successful.
       no arguments, since that sort of info presumably was passed in when
       webw_parseargs was called. */
    struct webw_data       *my_data;
    cmgr_id_t               cid = -1;
    struct ofid             root_ofid;
    struct onode_create_in  root_onode_params;
    struct viceroy_onode   *rop;
    int                     loop;
    ds_safeq_t             *workq = NULL;
    struct cache_entry *cache_list;
#ifdef SETS
    ds_safeq_t             *prefq = NULL;
#endif

    /* set up warden specific data. first we allocate the space
       in w_data, then we assign that same memory location to
       a pointer of the proper type. We then use the properly
       cast pointer in order to not have to cast the data every
       time we reference it. */
    ALLOC(wp->w_data, struct webw_data);
    my_data = (struct webw_data *)(wp->w_data);
    my_data->wwd_server = server_name;
    if (server_name == NULL) {
	ds_thread_log(1, "webw_init: no server specified on command line.");
	goto failedinit;
    }

    server_name = NULL;

    /* Set up the RPC2 connection to ody web server. */
    cid = webw_connect_to_machine(my_data->wwd_server);
    if(cid < 0) {
	/* XXX want cm to turn errnum into string */
        ds_thread_log(0, "webw_init: couldn't establish connection "
		         "to server %s (cm returned %d).",
		      my_data->wwd_server, cid);
	goto failedinit;	/* couldn't establish connection */
    } else {
	ds_thread_log(0, "webw_init: established connection %d "
		         "to server %s", cid, my_data->wwd_server);
    }
    my_data->wwd_cid             = cid;

    /* set up remainder of webwarden's w_data */
    my_data->wwd_current_uniq    = 1;	/* set up current uniqifier */
    my_data->wwd_img_quality = 100;     /* set default GIF compression
					   to none */
    my_data->wwd_html_quality = 100;    /* similar for HTML */
    ALLOC(my_data->wwd_monitor, ds_hwlock_t);
    ds_hwlock_init(my_data->wwd_monitor);

#ifdef INSTRUMENTED
    webw_init_histo(my_data);	/* initialize histograms */
#endif /* INSTRUMENTED */    
  
    /* set up safeq for workers to use */
    workq = ds_safeq_create();
    ASSERT(workq);
    my_data->wwd_workq = workq;

    /* set up workers */
    for(loop = 0; loop < WEBW_WORKERS; loop++) {
	struct webw_worker_params *pp;

	ALLOC(pp, struct webw_worker_params);
	pp->wwp_num   = loop;
	pp->wwp_workq = workq;
	pp->wwp_data  = my_data;

	ASSERT(ds_thread_create(webw_worker_shell,
				(char *)pp,
				DS_THREAD_NORMAL,
				"web_worker",
				0));
    }

#ifdef SETS
    /* set up safeq for prefetch workers to use */
    prefq = ds_safeq_create();
    ASSERT(prefq);
    my_data->wwd_prefq = prefq;

    /* set up prefetch workers */
    /* NOTE: need to have more of these threads than the number of
     * possible concurrent prefetches, or deadlock can occur! */
    for(loop = 0; loop < WEBW_PREFETCHERS; loop++) {
	struct webw_worker_params *pp;

	ALLOC(pp, struct webw_worker_params);
	pp->wwp_num   = loop;
	pp->wwp_workq = prefq;
	pp->wwp_data  = my_data;

	ASSERT(ds_thread_create(webw_worker_shell,
				(char *)pp,
				DS_THREAD_DAEMON,
				"web_prefetcher",
				0));
    }
#endif

    if(SWWW_version(SWWW_VERSION)) {
	/* if library version doesn't match header version, we're in
	   trouble */
        goto failedinit;
    }

    /* XXX For now, leave this here. I'll be moving it into
       SimpleWWW at some later point, once I figure out the
       proper way to do it. */

    signal(SIGPIPE, SIG_IGN);

    /* set up the root onode in init, since there is only one volume */
    root_ofid.ofid_ttype = WEB_TTYPE;
    root_ofid.ofid_tome  = 1;	/* we only support tome #1 for now */
    root_ofid.ofid_vnum  = 0;
    root_ofid.ofid_uniq  = 0;

    root_onode_params.oc_flags = ON_ROOTOFVOL | ON_UNCACHEABLE;
    root_onode_params.oc_vtype = O_VDIR;
    root_onode_params.oc_ofree = NULL;
    root_onode_params.oc_data  = NULL;

    rop = onode_create(&root_ofid, &root_onode_params);
    if (rop == NULL) {
	ds_thread_log(0, "webw_init: could not create volume onode root!");
	ASSERT(0);
    }

#ifdef SETS
    if (webw_sets_init())
	goto failedinit;
#endif

    if (do_caching) {
	/* Initialize the default cache space */
	cache_list = cache_request(wp, DEFAULT_CACHE_SIZE,
				   (DEFAULT_CACHE_SIZE/AVERAGE_PAGE_SIZE));
	if (cache_list == NULL) {
	    /* XXX For now, this is an error.  In the future, probably
	       want to proceed, and expect to hear from the viceroy later
	       that space became available XXX */
	    ds_thread_log(0, "webw_init: cache_request failed.");
	    goto failedinit;
	} else {
	    webw_cache_init(cache_list);
	}
    }

    ds_thread_log(10, "webw_init: initialized successfully");
    return(0);			/* successful initialization */

failedinit:
    /* initialization failed, so free w_data and return -1. make sure
       to set wp->w_data to NULL so that we can tell later that this
       initialization failed. */
    if (cid >= 0) {
	/* be nice and close the connection */
	cmgr_destroy(cid);
    }
    FREE(wp->w_data);
    wp->w_data = NULL;
    return(-1);
}


/*
 * GENERAL INTERFACE
 */


int webw_root(struct warden *wp,
	      tome_t        tome,
	      struct viceroy_onode  **root_onodep)
{
    struct viceroy_onode *root_onode;
    struct ofid  of;

    if (tome != 1) {
	ds_thread_log(100,"webw_root: requested tome (%d) not supported",
		      tome);
	return(EINVAL);		/* XXX what's the right error? */
    }

    of.ofid_ttype = WEB_TTYPE;
    of.ofid_tome  = tome;
    of.ofid_vnum  = 0;		/* for webwarden, {X,X,0,0} indicates */
    of.ofid_uniq  = 0;		/* an ofid for a volume. */

    root_onode = onode_get(&of);
    if (root_onode == NULL) {
	ds_thread_log(0, "webw_root: onode_get() failed for warden tome "
		         "%d's root onode.", tome);
	return(EINVAL);
    }

    onode_readlock(root_onode);

    *root_onodep = root_onode;

    /* Life is happy; we've gotten the onode of the root, and we've
       readlocked it, so return a happy return code. */
    return(0);
}

int webw_lookup(struct warden          *wp,
		struct viceroy_onode   *odp,
		struct ody_cred        *cred,
		char                   *name,
		char                  **remain,
		struct viceroy_onode  **opp)
{
    /* XXX for the moment, we make the following design decisions:
     *       - nothing can be mounted within this directory
     *       - there are no subdirectories of the root directory
     *       - files within the web root have the same name as the URL
     *         as they map onto.
     *
     * This makes our life here REALLY EASY.
     */
    char                    *url, *bookmark;
    struct cache_entry      *cache_entry;
    struct webw_page_data   *page_data;

    /* we don't care about odp */
    if (odp) {			/* KIP: may be called by parsehtml, in
				   which case no parent dir is given */
        onode_release_read(odp);
        onode_put(odp);
    }

    /* deal with ".." */
    if (strncmp(name, "..", 2) == 0) {/* XXX this is broken! */
	*remain = name + 2;
	return EMOUNTEDON;
    }

    /* if it's not "..", it must be a URL.
       get a new uniq number, and cons up an ofid/onode */
    url = strdup(name);
    *remain = NULL;

    if (strncmp("http", url, 4)) {
        *opp = NULL;
        return EINVAL;
    }

    /* XXX because URLs have slashes, we'll hack lookup to turn
           "\" into "/". this is because lookup still only takes
	   one component at a time. */
    bookmark = url;
    while(*bookmark != '\0') {
	if (*bookmark == '\\') {
	    *bookmark = '/';
	} else if (*bookmark == '#') {
	    /* eliminate anchors from looked-up URLs.  we want to
	     * fetch and cache the actual page.  The browser will deal with
	     * anchors.  lookup of an anchor likely came as a result of set
	     * membership. */
	    *bookmark = '\0';
	    break;
	}
	bookmark++;
    }

    ds_thread_log(1, "lookup %s", url);

    if (do_caching && webw_cache_lookup(url, &cache_entry)) {
	struct webw_page_data *wpdata = cache_entry->ce_data;

	/* the page data is in the cache, so create a new onode for it.
	   we're creating new onodes on each lookup so that we can
	   change the page data on a new open without affecting old
	   opens. */
	FREE(url);
	webw_build_onode(wp, wpdata, opp);

	((struct webw_onode_data *)(*opp)->o_data)->wod_centry = cache_entry;
	onode_readlock((*opp));
        return 0;
    }

    /* if it's not in the cache, we'll need to create a new onode */
    if (webw_new_onode(wp, url, opp)) {
	FREE(url);
	return(EINVAL);		/* XXX control currently can't reach here */
    }

    /* ...and enter this into the cache */
    page_data = ((struct webw_onode_data *) ((*opp)->o_data))->wod_page_data;

    if (do_caching) {
	if (!webw_cache_enter(url, *opp, 
			      WCELEM_VALID, page_data)) {
	    onode_put(*opp);
	    onode_put(*opp);	/* throw it away */
	    FREE(url);
	    return ENOSPC;		/* couldn't make room for this onode */
	}
    } else {
	/* force re-lookup of this onode (easier to do it this way
           than to flush the entry each time. */
	(*opp)->o_vtype |= ODY_NOCACHE;
    }

    /* Must also be returned readlocked */
    onode_readlock((*opp));
    return(0);
}

int webw_access(struct warden *wp,
		struct viceroy_onode  *op,
		struct ody_cred  *cred,
		int           mode)
{
    /* for web objects, everyone can read everything.
       this may be re-evaluated later. */
    switch(mode) {
    case O_A_READ:
    case O_A_EXEC:
    case O_A_WRITE:
	return 0;
	/* return EACCES; */
    }
    return EINVAL;
}

int webw_getattr(struct warden *wp, 
		 struct viceroy_onode  *op,
		 struct ody_vattr  *vap,
		 struct ody_cred  *cred)
{
    /* this is a simplistic implementation; this should become more
       sophisticated when I actually start doing some more interesting
       things in here. */
    struct webw_onode_data *the_odata = (struct webw_onode_data *)op->o_data;
    struct webw_page_data  *wpdata;
    time_t temp = 0;

    ds_thread_log(10,"webw_getattr on fid (%x.%x.%x.%x)",
		  op->o_fid.ofid_ttype,op->o_fid.ofid_tome,
		  op->o_fid.ofid_vnum,op->o_fid.ofid_uniq);

    if (the_odata && the_odata->wod_page_data && the_odata->wod_centry) {
	wpdata = the_odata->wod_page_data;
	vap->va_nlink = wpdata->wpd_refcnt;
	vap->va_fileid = atoi(&(the_odata->wod_centry->ce_fname[1])) + 1;
	vap->va_size = wpdata->wpd_data_avail;
    } else {
	vap->va_nlink = 1;
	vap->va_fileid = op->o_fid.ofid_vnum;
	vap->va_size = 0;
    }
    if (op->o_fid.ofid_vnum == 0 && op->o_fid.ofid_uniq == 0)
      vap->va_type = O_VDIR;
    else
      vap->va_type = O_VREG;
    vap->va_mode = O_S_READ;
    vap->va_nlink = 1;
    vap->va_uid = 0;
    vap->va_gid = 0;
    vap->va_fsid = op->o_fid.ofid_tome;
    vap->va_fileid = ody_f2i(&op->o_fid); /* for linux - break bsd44? */
    vap->va_size = 0;		/* XXX should have better value! */
    vap->va_blocksize = 512;
    vap->va_atime = temp;	/* XXX again, better values exist */
    vap->va_mtime = temp;
    vap->va_ctime = temp;
    vap->va_gen = op->o_fid.ofid_uniq;
    vap->va_flags = 0;
    vap->va_rdev = 0;
    vap->va_bytes = 0;
    vap->va_filerev = op->o_fid.ofid_uniq;

    return(0);
}

int webw_open(struct warden *wp,
	      struct viceroy_onode  *op,
	      struct ody_cred  *cred,
	      int            mode)
{
    char                   *result=NULL;
    int                     refetch = 0;
    struct webw_onode_data *the_odata = (struct webw_onode_data *) op->o_data;
    struct webw_page_data  *wpdata;
    struct stat             sbuf;
    int                     stat_result;
    int                     do_nothing = 0;

    if ((op->o_fid.ofid_vnum == 0) && (op->o_fid.ofid_uniq == 0)) {
	return 0;
    }

    /* XXX wpdata could have gone away if webw_lookup made this onode
       a long time ago.  Some mechanism should be added to catch this.
       (someday) XXX */
    wpdata = the_odata->wod_page_data;

    if (mode & O_O_WRITE) {
#ifdef SETS
	if (op->o_flags & ON_CURSOR) {
	    int rc;
	    /* This hack tells us that the user wants us to try and
               filter this page -- see if we know what it is. */
	    rc = webw_pick_filter(wpdata->wpd_url, the_odata->wod_url_list);
	} else {
#endif
	    return EACCES;
#ifdef SETS
	}
#endif
    }


	/*
	 * This is the meat of open. This is where we decide whether
	 * to fetch data or not, and if we have data, whether to flush
	 * it to disk or not:
	 *
	 * First, if don't have any data yet (this is the first demand
	 * fetch), or if we have data that has been marked
	 * questionable (because of a call to webw_invalidate_page())
	 * and there are no other openers, we do a conditional GET of
	 * the data.
	 *
	 * Else, if we do have the data and it's on disk, and we're
	 * already ON_CACHEABLE, we can leave things as is.
	 *
	 * If we don't meet those criteria, we then check if the data
	 * has been flushed to disk by the worker thread (which
	 * happens when it completes the demand fetch), and there are
	 * no other openers, but we haven't yet marked this onode as
	 * ON_CACHEABLE. If so, we can upgrade this to cacheable (We
	 * have to wait for there to be no other openers because then
	 * we can flip the onode from ON_UNCACHEABLE to ON_CACHEABLE.)
	 *
	 * Barring that, we finally resort to returning an uncacheable
	 * onode, representing that we are still demand fetching this
	 * data.
	 *
	 * kwalker note: For cursors, the wpdata behaves the same way.
	 * The only difference is that dropping the dbufs from memory
	 * is avoided, since parsehtml will need them pretty soon.
	 */

	ds_thread_log(1000, "in webw_open: openers is %d",
		      wpdata->wpd_openers);
	ds_thread_log(1000, "in webw_open: wpd_suspect is %d",
		      wpdata->wpd_suspect);

	/* bound to be a better structure for this logic?  (no
           offense, eric, I added some garbage!)  -kwalker */
	if (wpdata->wpd_state == VIRGIN ||
	    (wpdata->wpd_suspect == SUSPECT && wpdata->wpd_openers == 0)) {
	    refetch = webw_fetch_if_needed(wp, op, 1, &result);
	    wpdata->wpd_suspect = NOT_SUSPECT;
	} else if (!(op->o_flags & ON_UNCACHEABLE)) {
	    /* if we're already ON_CACHEABLE, do nothing */
	    do_nothing = 1;
	} else if (wpdata->wpd_state == CACHED) {
	    result = the_odata->wod_centry->ce_fname;
	    refetch = 0;
	} else {
	    result = NULL;
	    refetch = 0;
	}
	
	/* might want to check if state == FETCHED, and try to flush
           to disk (must have failed in the past because our cache was
           full) */
	if (wpdata->wpd_state == ON_DISK &&
	    wpdata->wpd_openers == 0) {
	    /* if we've got the data, but we're not already
	       ON_CACHEABLE, upgrade to ON_CACHEABLE (but only if
	       there are no other openers) */
	    if (!(op->o_flags & ON_CURSOR)) {
		struct cache_entry *centry = the_odata->wod_centry;
		wpdata->wpd_state = CACHED;
		ds_thread_log(0, "%s became CACHED", wpdata->wpd_url);
		webw_release_page_dbufs(wpdata);
		wpdata->wpd_data = NULL;
		result = centry->ce_fname; /* XXX think strdup is unnecc. */
	    } else {
		/* keep cursor accesses going to memory, keep the
                   dbufs around.  Technically, rdwr _could_ be
                   redirtected to disk, as long as the dbufs stay
                   around too.  for simplicity, I'll leave the onode
                   ON_UNCACHEABLE. */
		result = NULL;
	    }
	}

	if (!refetch && (op->o_flags & ON_CURSOR)) {
	    struct webw_worker_request *req;

	    ds_thread_log(5, "sending PARSE msg");
	    ALLOC(req, struct webw_worker_request);
	    req->wwr_op = onode_get(&(op->o_fid));
	    if (req->wwr_op == NULL)
		ASSERT(0);		/* this should never happen,
					   but just in case */
	    req->wwr_cid        = 0;    /* unused by PARSE handler */
	    req->wwr_reqid      = 0;    /* unused by PARSE handler */
	    req->wwr_reqtype    = WEBW_REQ_PARSE;
	    ds_safeq_enq(((struct webw_data *)(wp->w_data))->wwd_workq, req);
	}

	/*
	 * once we've gotten past the fetching decision, return
	 * either:
	 *   an ON_CACHEABLE onode which represents either:
	 *       the file which contains the error message
	 *       the disk cache file
	 *   an ON_UNCACHEABLE onode which represents:
	 *       data being streamed through the web warden
	 */

	if (!do_nothing) {
	    /* if do_nothing is set, leave the onode alone */
	    if (result && (do_caching || wpdata->wpd_status_code == 500)) {
		op->o_flags &= ~ON_UNCACHEABLE;	/* get rid of the
						   ON_UNCACHEABLE flag */
		op->o_cachefile = result;
		
		stat_result = stat(op->o_cachefile, &sbuf);
		if (stat_result < 0) {
		    return(stat_result); /* failed stat */
		}	
		op->o_device = sbuf.st_dev;
		op->o_inode  = sbuf.st_ino;
	    } else {
		op->o_device = 0;
		op->o_inode  = 0;
		op->o_flags |= ON_UNCACHEABLE;
	    }
	}

    ds_lock_write(&(wpdata->wpd_data_lock));
    wpdata->wpd_openers++;
    ds_lock_release(&(wpdata->wpd_data_lock));

    ds_thread_log(5, "open giving %s %d", wpdata->wpd_url,
		  (op->o_flags & ON_UNCACHEABLE));
    return 0;
}

int webw_close(struct warden *wp,
	       struct viceroy_onode  *op,
	       struct ody_cred  *cred,
	       int           mode)
{
    struct webw_page_data *wpdata;

    if ((op->o_fid.ofid_vnum == 0) && (op->o_fid.ofid_uniq == 0)) {
	return 0;
    }

    wpdata = (struct webw_page_data *)
	((struct webw_onode_data *)op->o_data)->wod_page_data;

    /* caller (vfs_close) does the necessary putting */

    ds_thread_log(1000, "in webw_close: openers is %d", wpdata->wpd_openers);

    ds_lock_write(&(wpdata->wpd_data_lock));
    wpdata->wpd_openers--;

    if (do_caching) {
	if ( (((struct webw_onode_data *)op->o_data)->wod_centry->ce_flags
	      & CE_FLUSHED)
	     && wpdata->wpd_openers == 0) {
	    webw_flush_page(wpdata);
	}
    }

    ds_lock_release(&(wpdata->wpd_data_lock));

    if (!do_caching) {
	/* extra onode_put, because there is no entry in the namecache --
	 * go ahead and destroy the onode. */
	onode_put(op);
    }

    return 0;
}

int webw_rdwr(struct warden *wp,
	      struct viceroy_onode *op,
	      struct ody_cred *cred,
	      int rwflag,
	      int *count,
	      int offset,
	      int ioflag,
	      void **data)
{
    struct webw_onode_data *odp = (struct webw_onode_data *) op->o_data;
    struct webw_page_data *wpdata = odp->wod_page_data;
    struct webw_dbuf       *bookmark;
    int                     remaining;
    int                     seen;
    char                   *bp;
    int                     index;

    if ((wpdata->wpd_data == NULL && wpdata->wpd_data_done)) {
	ds_thread_log(0, "BOGON -- got webw_rdwr after data flushed");
	*count = 0;
	*data = NULL;
	return(EINVAL);
    }

    if (rwflag != O_RW_READ) {
	/* only read is supported */
	*count = 0;
	*data = NULL;
	return(EOPNOTSUPP);
    }

    ds_thread_log(1000, "rdwr for %s", wpdata->wpd_url);

    ds_lock_read(&(wpdata->wpd_data_lock));

    if (wpdata->wpd_status_code == 500) {
      ds_thread_log(0, "get webw_rdwr for status 500 page");
      /*ASSERT(0);		shouldn't ever happen */
      /* XXX except that HTTP 0.9 replies seem to come here */
      /* Error occurred connecting to the webserver.  There will be 
	 no data to read, so just return now. */
      ds_lock_release(&(wpdata->wpd_data_lock));
      *count = 0;
      *data = NULL;
      return (0);
    } /* endif */

    /* check if IO_NDELAY is set */
    if (ioflag & O_IO_NDELAY) {
	/* XXX an alternate approach is to check only if the offset falls
	       within the available data */
	/* if so, then see if there's enough data available to fill
	   the request. if not, return EAGAIN. */
	if ( ((offset + *count - 1) > wpdata->wpd_data_avail ) &&
	     !(wpdata->wpd_data_done) ) {
	    ds_lock_release(&(wpdata->wpd_data_lock));
	    *count = 0;
	    *data = NULL;
	    return(EAGAIN);
	}
    }

    /*
     * find the appropriate swatch of data, and copy it into the
     * supplied buffer
     */

    /* find the start */
    seen     = 0;

    if (wpdata->wpd_data == NULL) {
	ds_lock_release(&(wpdata->wpd_data_lock));
	ds_thread_wait((char *) &(wpdata->wpd_data));
	ds_lock_read(&(wpdata->wpd_data_lock));
    }

    bookmark = wpdata->wpd_data;

    while ( !(seen <= offset && offset <= (seen + bookmark->wdb_length - 1))) {
	seen += bookmark->wdb_length;

	if (bookmark->wdb_next == NULL && bookmark->wdb_last == 0) {
	    ds_lock_release(&(wpdata->wpd_data_lock));
	    ds_thread_wait((char *) &(bookmark->wdb_next));
	    ds_lock_read(&(wpdata->wpd_data_lock));
	}

	bookmark = bookmark->wdb_next;
	if (bookmark == NULL) {
	    /* XXX needs fixed; this probably implies that the
	           offset is beyond the length of the file */
	    /* XXX my solution for now will be to return 0, implying EOF? */
	    ds_lock_release(&(wpdata->wpd_data_lock));
	    *count = 0;
	    *data = NULL;
	    return 0;
	}
    }

    index = offset - seen;

    /* do the copy */
    remaining = *count;
    bp = *data;

    while(remaining) {
	int amount;

	if (bookmark == NULL) {
	    /* XXX needs fixed for blocking read */
	    /* XXX for now, assume this is non-blocking,
	           and cut things short */
	    *count = *count - remaining;
	    break;
	}

	/* the amount to copy is the min of
	 * (bookmark->wdb_length - index) [the amount left in the dbuf], and
	 * remaining                      [the amount remaining to copy]
	 */
	amount = (remaining < (bookmark->wdb_length - index) ?
		  remaining :
		  (bookmark->wdb_length - index));

	/* copy out as much as is in this buffer */
	memcpy(bp, (bookmark->wdb_data) + index, amount);

	/* compute what remains */
	remaining -= amount;

	/* advance the bookmarks */
	index = 0;
	bp += amount;
	bookmark = bookmark->wdb_next;
    }

    ds_lock_release(&(wpdata->wpd_data_lock));

    return(0);
}

int webw_rdwr_dispose(struct warden *wp,
		      struct viceroy_onode *op,
		      struct ody_cred *cred,
		      int count,
		      int offset,
		      void *data)
{
    /* this shouldn't ever be called, since webw_rdwr uses the supplied
       buffer */
    ds_thread_log(0, "webw_rdwr_dispose: called on bogus buffer %x, size %d",
		  data, count);
    ASSERT(0);
    return(EOPNOTSUPP);
}

/* webw_readdir -- return contents of our top-level directory as the
 *                 pages we have cached.
 */
int webw_readdir(struct warden *wp,
		 struct viceroy_onode  *op,
		 struct ody_cred  *cred,
		 int           *count,
		 int            offset,
		 void          *data)
{
    int bytes_returned;

    if (op->o_fid.ofid_vnum != 0 || op->o_fid.ofid_uniq != 0) {
	ds_thread_log(0, "Readdir on a non-root directory.");
	return EIO;
    }

    bytes_returned = webw_cache_readdir(offset, *count, data);
    *count = bytes_returned;

    return 0;
}


/*
 * TSOPS
 */

int webw_tsop(struct warden *wp,
	      int           tsvnum,
	      struct viceroy_onode  *op,
	      struct ody_cred  *cred,
	      int           opnum,
	      size_t        argsz,
	      void          *argp,
	      size_t        *retsz,
	      void          **retp)
{
    int result;

    /* check the version number */
    if(tsvnum != WEBWARDEN_VERSION) {
	return(ODY_EBADVERSION);
    }

    /* dispatch TSOPs, based on opnum */
    switch(opnum) {
    case WEBW_OPEN_BY_URL:
	return(EOPNOTSUPP);	/* Not currently supported */
	/*
	result = webw_open_by_url(wp, op, argsz, argp, retsz, retp);
	return(result);
	*/
	break;
    case WEBW_STAT:
	result = webw_stat(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WEBW_SETMINQUALITY:
	result = webw_set_min_quality(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WEBW_GETMINQUALITY:
	result = webw_get_min_quality(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WEBW_SETMAXTIME:
	result = webw_set_max_time(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WEBW_INVALIDATEPAGE:
	result = webw_invalidate_page(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    case WEBW_FLUSHCACHE:
	result = webw_flush_cache(wp, op, argsz, argp, retsz, retp);
	return(result);
	break;
    }

    return(EOPNOTSUPP);
}

int webw_tsop_dispose(struct warden *wp,
		      struct viceroy_onode  *op,
		      struct ody_cred  *cred,
		      int           opnum,
		      size_t        argsz,
		      void          *argp,
		      size_t        retsz,
		      void          *retp)
{
    /* dispose of *retp, since we were responsible for allocating it */
    if (opnum == WEBW_GETMINQUALITY) /* don't free this */
        return(0);

    FREE(retp);			/* FREE() checks if retp is NULL */
    return(0);
}

cmgr_id_t webw_getcid(struct warden *wp, struct viceroy_onode  *op)
{
    struct webw_data  *my_data = (struct webw_data *) wp->w_data;
    return my_data->wwd_cid;
}

/*
 * OTHER
 */

#ifdef INSTRUMENTED

void webw_pstats(struct warden *wp, FILE *statout)
{
    struct webw_data   *my_data = (struct webw_data *) wp->w_data;
    struct webw_histos *histos;

    if (!my_data) {
      fprintf(statout, "webwarden not initialized\n");
      return;
    }

    histos = my_data->wwd_histos;

    /* object-centric */
    fprintf(statout, "\n##### OBJECT-CENTRIC\n");

    fprintf(statout, "\n## Object sizes (KB)\n");
    PrintHisto(statout, &(histos->wh_obj_sizes));

    fprintf(statout, "\n## Object fetch times (sec)\n");
    PrintHisto(statout, &(histos->wh_obj_times));

    fprintf(statout, "\n## Object fetch speeds (KB/sec)\n");
    PrintHisto(statout, &(histos->wh_obj_speeds));

    /* buffer-centric */
    fprintf(statout, "\n##### BUFFER-CENTRIC\n");

    fprintf(statout, "\n## Buffers/Object\n");
    PrintHisto(statout, &(histos->wh_buf_perobject));

    fprintf(statout, "\n## Buffer sizes (KB)\n");
    PrintHisto(statout, &(histos->wh_buf_sizes));

    fprintf(statout, "\n## Buffer fetch times (sec)\n");
    PrintHisto(statout, &(histos->wh_buf_times));

    fprintf(statout, "\n## Buffer fetch speeds (KB/sec)\n");
    PrintHisto(statout, &(histos->wh_buf_speeds));

    /* fidelity-centric */
    fprintf(statout, "\n##### FIDELITY-CENTRIC\n");

    fprintf(statout, "\n## Fidelities\n");
    PrintHisto(statout, &(histos->wh_fidelities));
}

#endif /* INSTRUMENTED */
