#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_onode.c -- handling of webwarden private onode data
 *
 * this code handles the data hung off of the onode by the warden.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include <sys/queue.h>

#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>
#include <base/ds/ds_rrlist.h>

#include <include/viceroy_onode.h>
#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/warden.h>
#include <include/onode-manager.h>
#include <include/cache_mgr.h>
#include <base/simplewww/swww.h>

#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_comm.h"
#include "webwarden_rpc2.h"

/*
 * webw_onode_data_free -- called when final ref to onode is removed.
 *                         release our hold on the page_data, free it
 *                         if necessary.
 */
void webw_onode_data_free(void *dp) {
    /* cast this right off into the right kind of thing */
    struct webw_onode_data *wodp = (struct webw_onode_data *) dp;
    struct webw_page_data  *wpdata;

    if (wodp == NULL)
	return;			/* ignore attempts to free NULL */
    
    wpdata = wodp->wod_page_data;

    ds_thread_log(5, "purging onode data (%s)", wpdata->wpd_url);

    ds_lock_write(&(wpdata->wpd_data_lock));

    /* this will only happen if the cache reference can be removed
       while an onode is referencing the page_data; otherwise, the
       cache will always remove the last ref and free things up. */
    if (--wpdata->wpd_refcnt == 0) {
	webw_release_page_data(wpdata);
	FREE(wpdata->wpd_url);	/* not freed in w_r_p_d for generality */
    } else
	ds_lock_release(&(wpdata->wpd_data_lock));

#ifdef SETS
    if (wodp->wod_url_list) {
	FREE(wodp->wod_url_list->urlbuf);
	FREE(wodp->wod_url_list->filter);
	FREE(wodp->wod_url_list);
    }
#endif

    FREE(wodp);
}

/*
 * webw_build_onode -- given the page_data, construct a new onode
 *                     Note: page_data assumed to have a reference for us.
 */

int webw_build_onode(struct warden         *wp,
                     struct webw_page_data *wpdata,
                     struct viceroy_onode           **opp)
{
    struct webw_data       *my_data = (struct webw_data *) wp->w_data;
    struct ofid             new_ofid;
    struct onode_create_in  new_onode_params;
    struct viceroy_onode           *new_onode;

    ds_hwlock_write(my_data->wwd_monitor);
    new_ofid.ofid_uniq = (my_data->wwd_current_uniq += 1);
    ds_hwlock_release_write(my_data->wwd_monitor);
    
    new_ofid.ofid_vnum  = 2;	/* want an even number -> not a directory */
    new_ofid.ofid_ttype = WEB_TTYPE;
    new_ofid.ofid_tome  = 1;

    new_onode_params.oc_flags = ON_UNCACHEABLE;
    new_onode_params.oc_vtype = O_VREG;
    new_onode_params.oc_ofree = webw_onode_data_free;
    ALLOC(new_onode_params.oc_data, struct webw_onode_data);

    if (new_onode_params.oc_data == NULL) {
	ds_thread_log(0, "webw_lookup: could not allocate webw_onode_data");
	ASSERT(0);
    } else {
	/* set up the webwarden specific onode data */
	struct webw_onode_data *odp;

	odp = (struct webw_onode_data *) new_onode_params.oc_data;
	odp->wod_centry      = NULL;
	odp->wod_page_data   = wpdata;
	/* reference the page_data for this onode */
	ds_lock_write(&(wpdata->wpd_data_lock));
	wpdata->wpd_refcnt++;
	ds_lock_release(&(wpdata->wpd_data_lock));

#ifdef SETS
	ALLOC(odp->wod_url_list, struct webw_urlbuf);
	odp->wod_url_list->urlbuf = NULL;
	odp->wod_url_list->filter = NULL;
	odp->wod_url_list->odata = odp;
	odp->wod_url_list->urlbuf_pos = 0;
	odp->wod_url_list->urlbuf_consumer_pos = 0;
	odp->wod_url_list->cur_dbuf = &(wpdata->wpd_data);
	odp->wod_url_list->cur_dbuf_pos = 0;
	odp->wod_url_list->partial_elmt = 0;
	odp->wod_url_list->base_url = NULL;
	odp->wod_url_list->yielded = 0;
	odp->wod_url_list->parse_done = 0;
	ds_lock_init(&(odp->wod_url_lock));
	odp->wod_prefetching = NORMAL;
#endif
    }

    new_onode = onode_create(&new_ofid, &new_onode_params);
    if (new_onode == NULL) {
	ds_thread_log(0, "webw_lookup: could not create onode for %s",
		      wpdata->wpd_url);
	ASSERT(0);
    }

    if (onode_get(&new_ofid) == new_onode) {
	*opp = new_onode;
    } else {
	ds_thread_log(0, "webw_lookup: onode_get returned unexpected "
		         "onode.");
	ASSERT(0);
    }

    return 0;
}

/*
 * webw_new_page_data -- construct new page_data
 */
struct webw_page_data *webw_new_page_data(struct warden *wp,
					  char          *url)
{
    struct webw_page_data  *wpdata;

    /* setup the page_data */
    ALLOC(wpdata, struct webw_page_data);
    
    ds_lock_init(&(wpdata->wpd_data_lock));
    wpdata->wpd_url         = url;
    wpdata->wpd_state       = VIRGIN;
    wpdata->wpd_suspect     = NOT_SUSPECT;
    wpdata->wpd_status_code = 0;
    wpdata->wpd_status_line = NULL;
    wpdata->wpd_headers     = NULL;
    wpdata->wpd_header_bytes = 0;
    wpdata->wpd_header_size  = 0;
    wpdata->wpd_expires     = 0;
    wpdata->wpd_modified    = 0;
    wpdata->wpd_target      = NULL;
    wpdata->wpd_wp          = wp;
    /* start with zero ref so onode->page_data binding can blindly ++
     * it, whether or not the page_data was just created.  Safe since
     * this function is always paired with a build_onode. */
    wpdata->wpd_refcnt      = 0;
    wpdata->wpd_openers     = 0;
    WEBW_INIT_DBUF_HDR(wpdata);

    return wpdata;
}

/*
 * webw_new_onode -- construct a new onode, onode_data, and page_data
 */
int webw_new_onode(struct warden *wp,
		   char          *url,
		   struct viceroy_onode  **opp)
{
    struct webw_page_data *wpdata;

    wpdata = webw_new_page_data(wp, url);

    /* now that we've create the page data, create the onode */
    return(webw_build_onode(wp, wpdata, opp));
}

/*
 * webw_clean_page -- release dbufs, reset info fields.  Assumed to be called
 *                    when safe (not opened by anyone) [no locking needed!]
 */
void webw_clean_page(struct webw_page_data *wpdata)
{
    webw_release_page_data(wpdata);
    WEBW_INIT_DBUF_HDR(wpdata);
    
    wpdata->wpd_status_code = 0;
    wpdata->wpd_status_line = NULL;
    wpdata->wpd_headers = NULL;
    wpdata->wpd_modified = 0;
    wpdata->wpd_expires = 0;

    /* Note: no way currently for cursor onodes using this page_data
     * to know that their urlbuf has outdated info -- but who
     * cares. */
}

/*
 * webw_release_page_dbufs -- release dbuf chain.  Lock assumed to be held!
 */
void webw_release_page_dbufs(struct webw_page_data *wpdata)
{
    struct webw_dbuf       *dbp, *temp;

    ds_thread_log(5, "tossing dbufs %s", wpdata->wpd_url);
    dbp = wpdata->wpd_data;
    while(dbp != NULL) {
	temp = dbp->wdb_next;
	FREE(dbp->wdb_data);
	FREE(dbp);
	dbp = temp;
    }
    wpdata->wpd_data = NULL;
    /* XXX reset data_avail, or does that apply to disk? */
}

/*
 * webw_collapse_dbufs -- replace a dbuf chain with a single contiguous buffer
 *                        It is assumed that the lock is held, and that 
 *                        wdp_data_done == 1.
 */
void webw_collapse_dbufs(struct webw_page_data *wpdata)
{
    struct webw_dbuf *dbp, *temp;
    char *new_buf, *b;

    ds_thread_log(10, "combining dbufs for %s", wpdata->wpd_url);
    dbp = wpdata->wpd_data;
    NALLOC(new_buf, char, wpdata->wpd_data_avail+1);
    b = new_buf;
    while (dbp != NULL) {
	temp = dbp->wdb_next;
	memcpy(b, dbp->wdb_data, dbp->wdb_length+1);
	b += dbp->wdb_length;
	FREE(dbp->wdb_data);
	FREE(dbp);
	dbp = temp;
    }
    ALLOC(dbp, struct webw_dbuf);
    dbp->wdb_data = new_buf;
    dbp->wdb_length = wpdata->wpd_data_avail;
    dbp->wdb_last = 1;
    dbp->wdb_next = NULL;
    wpdata->wpd_data = dbp;
}

/*
 * webw_release_page_data -- deference the page_data, if last
 *                           reference, free all memory.  Assumed to be
 *                           called with the lock held.
 */
void webw_release_page_data(struct webw_page_data *wpdata)
{
    ds_thread_log(5, "releasing page_data %s", wpdata->wpd_url);
    FREE(wpdata->wpd_status_line);
    FREE(wpdata->wpd_headers);
    webw_release_page_dbufs(wpdata);
}

/*
 * webw_flush_page -- flush all data and metadata for this page_data
 */
void webw_flush_page(struct webw_page_data *wpdata)
{
    /* toss dbufs, reset fields */
    webw_clean_page(wpdata);
    if (wpdata->wpd_state >= ON_DISK) {
	/* clean up cache file */
	cache_truncate(wpdata->wpd_centry);
    }
    /* downgrade to VIRGIN */
    wpdata->wpd_state = VIRGIN;
}
