#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_onode.h -- webwarden specific onode information and handlers
 *
 */

/*
 * data structure used to hold all or a portion of an HTTP object.
 * this is loosely patterned after BSD mbufs, but greatly simplified.
 *
 * N.B.: if wdb_last != 1, then a dbuf is not the last dbuf, even if its
 *       wdb_next pointer is NULL. If wdb_next is NULL, that just means
 *       that the next buffer-full hasn't arrived yet.
 */
    
struct webw_dbuf {
    int               wdb_length; /* length of data in this buffer */
    int               wdb_last;   /* set to 1 if this is known to be the
				     last dbuf */
    char             *wdb_data;   /* the buffer proper. if this is NULL,
				     the data for this dbuf hasn't arrived
				     yet. */
    struct webw_dbuf *wdb_next;   /* the buffer (if any) which follows
				     this one */
};

typedef enum {
    LYCOS_SEARCH = 1,
    ALTAVISTA_SEARCH,
    TEST_PAGE
} filter_type;

struct webw_filter {
    filter_type type;
    struct webw_dbuf *stop_dbuf;
    int stop_offset;
};

#define MAX_URL 1024		/* XXX this prolly exists elsewhere? */
struct webw_urlbuf {
    struct webw_onode_data *odata; /* backpointer */
    struct webw_filter     *filter; /* info for filtering this page */
    char      *urlbuf;		/* buffer for HREF urls */
    int        urlbuf_pos;	/* position in urlbuf */
    int        urlbuf_consumer_pos; /* consumer pos in urlbuf */
    struct webw_dbuf **cur_dbuf; /* current dbuf supplying html source */
    int        cur_dbuf_pos;	/* our position in cur_dbuf */
    char       *base_url;	/* base for relative urls */
    int        partial_elmt;	/* flag for elements that cross dbufs */
    int        yielded;		/* count of yielded urls */
    int        parse_done;	/* done parsing? */
};


/*
 * macro for initializing dbuf
 */
#define WEBW_INIT_DBUF_HDR(dbh) \
    do { \
	(dbh)->wpd_data       = NULL; \
	(dbh)->wpd_data_done  = 0; \
	(dbh)->wpd_data_avail = 0; \
    } while(0);

#define WEBW_INIT_DBUF(dbufp) \
    do { \
	(dbufp)->wdb_length = 0; \
	(dbufp)->wdb_last   = 0; \
	(dbufp)->wdb_data   = NULL; \
	(dbufp)->wdb_next   = NULL; \
    } while(0);

#ifdef SETS
typedef enum {
    NORMAL = 1,
    PREFETCH_PAGE,
    PREFETCH_CONTENTS
} fetch_type;
#endif

typedef enum {
    VIRGIN = 1,
    OPENED,
    FETCHED,
    ON_DISK,
    CACHED
} on_state;

/*
 * webwarden-specific per-onode data
 */
struct webw_onode_data {
    struct webw_page_data *wod_page_data; /* shared structure
					     with page info */
    struct cache_entry    *wod_centry;    /* cache info for this onode */

#ifdef SETS
    ds_lock_t           wod_url_lock;     /* protect the readers from the 
					     writer */
    struct webw_urlbuf *wod_url_list;     /* list of contained URL's.  This
					     buffer is a linked list */
    fetch_type          wod_prefetching;  /* webw_prefetch sets this so 
					     the worker can do the
					     right thing */
#endif
};

/*
 * definitions for values of cache_entry->ce_flags
 *   XXX now that centry is generic, revisit this.
 */
#define WCELEM_VALID      0x0100

enum webw_suspect {
    NOT_SUSPECT = 12,
    SUSPECT     = 39
};

/*
 * webw_page_data -- structure for information relating to cached
 *    web pages.  Instances of this structure map 1-1 with centries.
 *    (Contains many fields formerly in the webw_onode_data).
 */
struct webw_page_data {
    char             *wpd_url;	        /* URL which onode represents */
    on_state          wpd_state;        /* state of the onode */
    on_state          wpd_old_state;    /* state before cget */
    enum webw_suspect wpd_suspect;      /* is this data suspect?
					   (i.e. invalidated by
					   invalidePage TSOP) */
    int               wpd_status_code;  /* status code from status line
					   (e.g. 200) */
    char             *wpd_status_line;  /* status line (e.g.
					   "HTTP/1.0 200 Document follows") */
    int               wpd_header_size;  /* number of headers */
    int               wpd_header_bytes; /* number of bytes in the marshalled
					   headers */
    char             *wpd_headers;      /* marshalled list of headers */
    time_t            wpd_expires;      /* (opt.) expiration for page */
    time_t            wpd_modified; 	/* last modified date */
    struct viceroy_onode     *wpd_target;       /* target of 301/302 redirect */
    struct warden    *wpd_wp;           /* backpointer to warden -- used
					   in parsehtml (and to access cmid) */
    int               wpd_openers;      /* open onodes using this page_data */
    int               wpd_refcnt;       /* # of onodes using me */
    ds_lock_t         wpd_data_lock;    /* lock on page_data */
    int               wpd_data_done;    /* set to 1 if fetch has completed */
    int               wpd_data_avail;   /* number of bytes of data currently
					   available */
    struct webw_dbuf *wpd_data;         /* HTTP object data */
    struct cache_entry *wpd_centry;     /* cache_entry */
};

/* Prototypes */
void webw_onode_data_free(void *dp);

int webw_build_onode(struct warden          *wp,
                     struct webw_page_data  *wpdata,
                     struct viceroy_onode          **opp);

struct webw_page_data *webw_new_page_data(struct warden *wp,
					  char          *url);

int webw_new_onode(struct warden *wp,
	           char          *url,
		   struct viceroy_onode **opp);

void webw_clean_page(struct webw_page_data *wpdata);
void webw_release_page_dbufs(struct webw_page_data *wpdata);
void webw_collapse_dbufs(struct webw_page_data *wpdata);
void webw_release_page_data(struct webw_page_data *wpdata);
void webw_flush_page(struct webw_page_data *wpdata);
