#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_sets.c -- support for Dynamic Sets
 */

#ifdef SETS

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/queue.h>
#include <sys/stat.h>

#include <include/odyssey.h>
#include <include/odytypes.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>
#include <base/ds/ds_rrlist.h>

#include <include/viceroy_onode.h>
#include <kernel/vfs/includes/ody.h>
#include <kernel/sets/common/sets.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/onode-manager.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>

#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_worker.h"
#include "webwarden_comm.h"
#include "webwarden_sets.h"
#include "webwarden_cache.h"

extern int prefetch;

/* These define the tag names whose values we're interested in collecting,
   for each relevant element type */
struct html_elmt_attr {
    char *name;
    int   namelen;
};
static struct html_elmt_attr A_attr[] = {{"href", 4}, {NULL, 0}};
static struct html_elmt_attr AREA_attr[] = {{"href", 4}, {NULL, 0}};
static struct html_elmt_attr IMG_attr[] = {{"src", 3},{"lowsrc", 6},
					   {NULL, 0}};
static struct html_elmt_attr FRAME_attr[] = {{"src", 3}, {NULL,0}};
static struct html_elmt_attr BODY_attr[] = {{"background", 10}, {NULL,0}};
static struct html_elmt_attr INPUT_attr[] = {{"src", 3}, {NULL,0}};
#if 0
static struct html_elmt_attr OPTION_attr[] = {{"value", 5}, {NULL,0}};
#endif

extern int do_caching;

/* s_index: same as index, except take an "end" parameter to avoid depending
 *          on a NULL terminator */
inline char *s_index(char *str, char token, char *end)
{
  while (str < end && *str != token)
    str++;

  return (str == end) ? NULL : str;
}

/* Nothing to init at the moment */
int webw_sets_init(void)
{
    return 0;
}

/* Expand an onode -- ie open a cursor for the page.
   Currently, each request for a cursor creates a new onode.  The dbuf chain
   may be shared with a cached onode. */
int webw_expand(struct warden *wp,
		struct viceroy_onode  *odp,
		struct ody_cred  *cred,
		char          *name,
		char          **remain,
		struct viceroy_onode  **opp)
{
    char               *url, *bookmark;
    struct cache_entry *c_entry;

    /* This should be kept similar to webw_lookup in webwarden_main.c */

    onode_release_read(odp);
    onode_put(odp);
    
    /* deal with ".." */
    if (strncmp(name, "..", 2) == 0) {/* XXX this is broken! */
	*remain = name + 2;
	return EMOUNTEDON;
    }

    /* if it's not "..", it must be a URL.
       get a new uniq number, and cons up an ofid/onode */
    url = strdup(name);
    *remain = NULL;

    /* XXX because URLs have slashes, we'll hack lookup to turn
           "\" into "/". this is because lookup still only takes
	   one component at a time. */
    bookmark = url;
    while(*bookmark != 0) {
	if (*bookmark == '\\')
	    *bookmark = '/';
	bookmark++;
    }

    if (strncmp("http", url, 4)) {
	*opp = NULL;
	return EINVAL;
    }

    /* End of copy/paste from webw_lookup */
    
    if (do_caching && webw_cache_lookup(url, &c_entry)) {
	struct viceroy_onode *new_onode;
	struct webw_onode_data *new_odata;
	struct webw_page_data *wpdata;

	wpdata = (struct webw_page_data *)c_entry->ce_data;

	/* We have a readlock on this guy. */
	if (webw_build_onode(wp, wpdata, &new_onode)) {
	    FREE(url);
	    return(EINVAL);	/* XXX control currently can't reach here */
	}

	/* Let the data collection finish; this is much simpler, although
	   less efficient than, interleaving parsing with downloading. */
	if (!wpdata->wpd_data_done)
	    ds_thread_wait((char *)wpdata->wpd_data_done);
	
	if (wpdata->wpd_state == CACHED) {
	    int fd, rc;

	    /* this guy is on disk.  for ease of parsehtml'ing, read
               it into dbufs. again, this is for simplicity's sake -- the
	       parsing code only deals with in-memory dbufs. */
	    /* XXX figure out when it's safe to throw these away.
               Note that the wpdata is still CACHED, but we're taking
               up space both places. XXX */
	    ds_lock_write(&(wpdata->wpd_data_lock));
	    ALLOC(wpdata->wpd_data, struct webw_dbuf);
	    NALLOC(wpdata->wpd_data->wdb_data, char, wpdata->wpd_data_avail);
	    fd = cache_open(c_entry, O_RDONLY);
	    rc = read(fd, wpdata->wpd_data->wdb_data, wpdata->wpd_data_avail);
	    if (rc < 0) {
		/* bad news. make this more gentle XXX */
		perror("read");
		ASSERT(0);
	    }
	    rc = cache_close(c_entry, fd, 0);
	    if (rc) {
		perror("close");
		ASSERT(0);
	    }
	    wpdata->wpd_data->wdb_length = wpdata->wpd_data_avail;
	    wpdata->wpd_data->wdb_last = 1;
	    wpdata->wpd_data->wdb_next = NULL;
	    ds_lock_release(&(wpdata->wpd_data_lock));
	}

	new_odata = (struct webw_onode_data *)new_onode->o_data;

	/* setup urlbuf */
	NALLOC(new_odata->wod_url_list->urlbuf, char, URL_LIST_LEN);

	/* share the cache entry */
	new_odata->wod_centry = c_entry;
	new_odata->wod_url_list->cur_dbuf = &new_odata->wod_page_data->wpd_data;

	*opp = new_onode;
	/* delay parsehtml until webw_open -- need to know if we're
           refetching or not, filtering or not. */
    } else {
	if (webw_new_onode(wp, url, opp)) {
	    FREE(url);
	    return(EINVAL);	/* XXX control currently can't reach here */
	}
	
	if (do_caching && !webw_cache_enter(url,
					    *opp,
					    WCELEM_VALID,
					    ((struct webw_onode_data *)(*opp)->o_data)->wod_page_data)) {
	    onode_put(*opp);
	    onode_put(*opp);	/* throw it away */
	    FREE(url);
	    return ENOSPC;		/* couldn't make room for this onode */
	}
    }	

    (*opp)->o_flags |= ON_CURSOR;

    /* Must also be returned readlocked */
    onode_readlock((*opp));
    return(0);
}

int webw_cursexpand(struct warden *wp,
		    struct viceroy_onode  *op,
		    struct ody_cred  *cred,
		    size_t        size,
		    size_t        *retsz,
		    char          *buf)
{
    struct webw_onode_data *the_odata = (struct webw_onode_data *)(op)->o_data;
    struct webw_urlbuf     *ub = the_odata->wod_url_list;
    struct linktype         more = {0, "..."};
    int len;
    int err;

    ASSERT(size == URL_LIST_LEN);

    /* consume from the webw_urlbuf */
    ds_lock_read(&(the_odata->wod_url_lock));

    /* note: invariant that urlbuf_consumer_pos <= urlbuf_pos */
    len = ub->urlbuf_pos-ub->urlbuf_consumer_pos;
    if (the_odata->wod_url_list->parse_done) {
	if (ub->urlbuf_consumer_pos == ub->urlbuf_pos) {
	    ds_thread_log(100, "Yielded %d urls",
			  the_odata->wod_url_list->yielded);
	    *retsz = 0;
	    err = 0xe0f;
	} else {
	    bcopy(ub->urlbuf+ub->urlbuf_consumer_pos, buf, len);
	    *retsz = len;
	    ub->urlbuf_consumer_pos = ub->urlbuf_pos;
	    ds_thread_wakeup((char *)&(ub->urlbuf_consumer_pos));
	    err = 0;	
	}
    } else {
        /* assumes that the producer left space for the "more" entry  */
        if (len) bcopy(ub->urlbuf+ub->urlbuf_consumer_pos, buf, len);
	/* add "..." entry to end */
	bcopy(&more, buf+len, sizeof(more));
	*retsz = len+sizeof(more);
	ub->urlbuf_consumer_pos = ub->urlbuf_pos;
	ds_thread_wakeup((char *)&(ub->urlbuf_consumer_pos));
	err = 0;
    }
    ds_lock_release(&(the_odata->wod_url_lock));

    return err;
}

int webw_prefetch(struct warden *wp,
		  struct viceroy_onode  *op,
		  struct ody_cred  *cred,
		  int            mode)
{
    int rc = webw_do_prefetch(wp, op, cred, mode, PREFETCH_CONTENTS);

    ds_thread_log(0, "prefetched %s",
		  ((struct webw_onode_data *)op->o_data)->wod_page_data->wpd_url);
    return(rc);
}

/*
 * webw_parsehtml: parse an html document for either embedded links
 *                 or images.  parsing for links is used in SET membership
 *                 determination, while parsing for images is for 
 *                 prefetching the contents of the page
 *
 * May be called in the following circumstances:
 *     - each time a new dbuf of data has been received (cursor or pfetch pg)
 *     - when an entire document to filter has been received into dbufs
 *     - when a complete cursor is being re-parsed
 */
void webw_parsehtml(char *url, struct webw_urlbuf *list,
		    struct warden *wp, int action)
{
    struct webw_dbuf *curbuf = *(list->cur_dbuf);
    int               pos = list->cur_dbuf_pos;
    char             *ptr;
    char             *end;
    char             *next_elmt, *end_of_elmt, *link;
    int               link_len;
    struct html_elmt_attr *attrs;
    fetch_type        prefetch_mode = PREFETCH_PAGE;

    /* NOTE: since it was rarely used, and was likely flawed, cross-dbuf
     *       parsing was removed to fix the common case for memory bogons */
    while (curbuf) {
	/* use a BASE if one registered */
	if (list->base_url)
	    url = list->base_url;

	/* bound the valid region in curbuf */
	ptr = curbuf->wdb_data+pos;
	end  = curbuf->wdb_data + curbuf->wdb_length;

	while (pos < curbuf->wdb_length) {
	    /* find the beginning of the first element */
	    next_elmt = s_index(ptr, '<', end);

	    /* check if a filter is registered which wants us to ignore
	     * this part of the page */
	    if (list->filter && curbuf == list->filter->stop_dbuf) {
		char *next_pos = next_elmt ? next_elmt : end;
		if (next_pos - curbuf->wdb_data > list->filter->stop_offset) {
		    /* can't go any further */
		    return;
		}
	    }
	    
	    /* nothing else in this buffer */
	    if (!next_elmt) {
		/* force a move to the next buffer */
		pos = curbuf->wdb_length;
		ptr += pos;
		continue;
	    }

	    next_elmt++;	/* beginning of tag */
	    
	    if (!strncasecmp(next_elmt, "!--", 3)) {
		/* skip comments! */
		/* XXX can they be embedded? XXX
		   XXX the -- and > don't have to be next to each other XXX */
		end_of_elmt = strstr(next_elmt, "-->");
		if (!end_of_elmt)
		    ASSERT(0);	/* busted if crossing dbuf */
		end_of_elmt += 2;
		ds_thread_log(1000, "skipping comment (%d long)\n",
			      end_of_elmt-next_elmt);
		ptr = end_of_elmt;
		pos = ptr-curbuf->wdb_data;
		continue;
	    }

	    if (!(end_of_elmt = s_index(next_elmt, '>', end))) {
		ds_thread_log(0, "webw_parsehtml unhappy -- multi-dbuf element");
		ds_thread_log(0, "on url %s", url);
		pos = curbuf->wdb_length;
		ptr += pos;
		continue;
	    }

	    *end_of_elmt = '\0'; /* NULL-term for findAttrValue */

	    if (!strncasecmp(next_elmt, "base", 4)) {
		findAttrValue(next_elmt, "href", 4, &link, &link_len);
		if (link) {
		    char save = link[link_len];
		    FREE(list->base_url);
		    link[link_len] = '\0';
		    /* makeabsolute? */
		    list->base_url = strdup(link);
		    link[link_len] = save;
		    ds_thread_log(5, "new base: %s", list->base_url);
		    url = list->base_url;
		}
	    } else {
		switch (action) {
		case PARSE_LINKS:
		    if (next_elmt[0] == 'a' || next_elmt[0] == 'A') {
			if (isspace(next_elmt[1])) {
			    attrs = A_attr;
			} else if (!strncasecmp(next_elmt+1, "REA", 3)) {
			    attrs = AREA_attr;
			} else {
			    break;
			}
		    } else if (!strncasecmp(next_elmt, "frame", 5)) {
			attrs = FRAME_attr;
		    } else {
				/* not interested */
			break;
		    }
		    
		    /* Will return an absolute link if found. */
		    /* (this should loop on A_attr) */
		    while (attrs->name != NULL) {
			findAttrValue(next_elmt,
				      attrs->name,
				      attrs->namelen,
				      &link, &link_len);
			/* link is beginning of url[link_len] (not
			   \0-term) */
			if (link) {
			    link = makeAbsolute(url, link, link_len);
			    /* link is malloc'ed */
			    if (processAbsoluteLink(list, link))
				list->yielded++;
			    free(link);
			} else
			    /* No href means it's an internal
			       anchor...  <a name="name"> */
			    ;
			attrs++;
		    }
		    break;
		case PARSE_IMGS:
		    if (!strncasecmp(next_elmt, "img", 3)) {
			attrs = IMG_attr;
		    } else if (!strncasecmp(next_elmt, "frame", 5)) {
			attrs = FRAME_attr;
			prefetch_mode = PREFETCH_CONTENTS;
			/* XXX note this could "recurse" or go really
			 *     deep*/
		    } else if (!strncasecmp(next_elmt, "body", 4)) {
			attrs = BODY_attr;
		    } else if (!strncasecmp(next_elmt, "input", 5)) {
			findAttrValue(next_elmt, "type", 4, &link, &link_len);
			if (link && !strncasecmp(link, "image", 5))
			    attrs = INPUT_attr;
			else
			    break;
		    } else {
			break;
		    }
		    
		    while (attrs->name != NULL) {
			findAttrValue(next_elmt, attrs->name,
				      attrs->namelen, &link,
				      &link_len);
			if (link) {
			    struct viceroy_onode *op;
			    char *remain;
			    int err = 0;
			    
			    link = makeAbsolute(url, link, link_len);
			    ds_thread_log(5, "Found image: %s", link);
			    /* need to start prefetch of this guy, and
			       add it to the cache!  Need lookup to
			       create ofid/onode, add to cache, but
			       don't open -- may never get
			       requested. */
			    /* link is malloc'ed */
			    err = (wp->w_lookup)(wp, NULL, NULL,
						 link, &remain, &op);
			    if (!err) {
				onode_release_read(op);
				onode_writelock(op);
				ds_thread_log(0, "causing PF %s", link);
				err = webw_do_prefetch(wp, op,
						       NULL, O_O_READ,
						       prefetch_mode);
				/* this prolly aint right XXX */
				if (!err) {
				    /* XXX OK, this is getting
				     * bizarre.  I just wanted to
				     * cache these things and rely on
				     * LRU to keep them, but not
				     * actually reference them as
				     * opened -- if I left them
				     * referenced, they'd NEVER be
				     * purged from the cache.
				     * (Actually, this makes sense --
				     * this way, a prefetched page
				     * only "blocks" one page in the
				     * cache) */
				    onode_release_write(op);
				    /* block until this guy is
				     * fetched?  (ie, don't say this
				     * page is prefetched until all
				     * inlines are also prefetched */
				    /* XXX do this synchrony more
				       efficiently */
				    
				    if (!((struct webw_onode_data *)op->o_data)->wod_page_data->wpd_data_done)
					ds_thread_wait((char *) &((struct webw_onode_data *)op->o_data)->wod_page_data->wpd_data_done);
				    err = (wp->w_close)(wp, op,
							NULL, O_O_READ);
				    if (!err) {
					onode_put(op); /* leave only the effect on the cache */
				    }
				    /* XXX if one of these comes back
				     * with status_code 500, should we
				     * just bail?  we'll prolly just
				     * end up caching a bunch of error
				     * messages. */
				}
			    }
			    free(link);
			}
			attrs++;
		    }
		    break;
		}
		*end_of_elmt = '>';
	    }
	    ptr = end_of_elmt+1;
	    pos = ptr-curbuf->wdb_data;
	}
	/* NOTE: since the wpd_data chain isn't freed until close,
	 * it is safe to point into it here. */
	list->cur_dbuf = &(curbuf->wdb_next);
	list->cur_dbuf_pos = pos = 0;
	curbuf = curbuf->wdb_next;
    }
}

/* Note: ptr is NULL terminated */
void findAttrValue(char  *ptr,
		   char  *name,
		   int    namelen,
		   char **attr_value,
		   int   *aval_len)
{
    int i;
    char *c, *d;
    
    /* First, isoltate this anchor. Assume no html tags *within* link tag. */
    for (; *ptr; ptr++) {
	for (c = ptr; *c && tolower(*c) != name[0]; c++)
	    ;

	if (*c == 0) {
	    /* reached end of element without finding the attribute name */
	    break;
	}

	ptr = c;

	for (i = 1; ptr[i] && i < namelen; i++)
	    if (tolower(ptr[i]) != name[i]) 
		break;

	if (i == namelen) {		/* We found the field name. */
	    c+= namelen;		/* Skip over field name. */
	    SKIPWHITESPACE(c);
	    
	    if (*c == '=') {
		c++;
		SKIPWHITESPACE(c);
		
		if (*c == '"') {
		    /* We now have the link */
		    d = index(++c, '"');
		    if (!d || d == c) {
			/* no second '"', or empty string */
			break;
		    }
		} else {
		    /* no '"' around the link */
		    SKIPWHITESPACE(c);
		    d = c;
		    /* Find first whitespace after the start of the attrval */
		    while (*d && !isspace(*d))
			d++;
		}
		*attr_value = c;
		*aval_len = d-c;
		return;
	    }
	    else
		/* didn't find '=' */
		break;
	} else {
	    /* Don't print anything out, 'cause <A name =""> is legal as 
	       is <address> */
	}
    }
    
    *attr_value = NULL;
    return;
}

char *makeAbsolute(char *url, char *link, int len)
{
    char *newlink, *proto, *server, *path;
    char *c, *d;
    char *colon;
    char save;
    int n, port;
    
    /* null-terminate the discovered link */
    save = link[len];
    link[len] = '\0';

    colon = index(link, ':');
    if (colon) {
	/* If there is no '/' before the ':' then it is an absolute link */
	*colon = 0;
	c = index(link, '/');
	*colon = ':';
	if (c == NULL) {
	    char *abs_link;
	    d = index(colon+3, '/');
	    if (!d) {
		/* this is just a hostname -- put '/' on the end */
		abs_link = malloc(strlen(link)+2);
		strcpy(abs_link, link);
		strcat(abs_link, "/");
	    } else 
		abs_link = strdup(link);
	    link[len] = save;
	    return abs_link;
	}
    }

    /* Special case: null link is a self-pointer */
    if (link == "") {
	char *abs_link = strdup(url);
	link[len] = save;
	return abs_link;
    }

    /* Special case: links beginning with '#' are a self-pointer plus anchor */
    if (*link == '#') {
	newlink = (char *)malloc(strlen(url) + strlen(link) + 1);
	strcpy(newlink, url);
	strcat(newlink, link);
	link[len] = save;
	return newlink;
    }

    /* need space for NULL-term and the joining '/' */
    newlink = (char *)malloc(strlen(url) + strlen(link) + 2);

    /* Parse the reference URL */
    SWWW_parse_url(url, &proto, &server, &port, &path);

    /* The partial url in <link> will replace the suffix of the url.
     * Look at
     * http://info.cern.ch/hypertext/WWW/Addressing/URL/4_3_Partial.html
     * It says if partial begins with non-zero number of slashes,
     * replace suffix of url beginning with leftmost occurrence of
     * that number of slashes such that there isn't an occurrence of
     * more consecutive slashes in the suffix. Since that sounds
     * expensive, go for an easier approximation: If 2 slashes, you
     * have a new server + path, if 1 slash, you have a new path, if
     * no slashes, just replace the last component of url.  */
    n = 0;
    d = link;
    while (*d++ == '/')
	n++;

    if (n == 2) {	/* Replace host and rest of path with link */
	strcpy(newlink, proto);
	strcat(newlink, ":");
	strcat(newlink, link);
    } else if (n == 1) { /* Replace path with link */
	if (port != -1) {
	    sprintf(newlink, "%s://%s:%d%s", proto, server, port, link);
	} else { 
	    sprintf(newlink, "%s://%s%s", proto, server, link);
	}
    } else if (n == 0) { /* relative pathname */
	c = link;

	/* Skip back for as many ".." as we have, or at least 1 component. */
	for (d = rindex(path, '/'); d && *c; ) {
	    if (c[0] == '.' && c[1] == '.') {
		c += 2;
		if (*c == '/' || *c == 0) {
		    char *e = d;
		    if (*c) c++;
		    *e = 0;
		    d = rindex(path, '/'); /* skip back another component. */
		    *e = '/';
		    continue; 
		} else
		    /* Didn't match "../", don't skip any farther */
		    break;	
	    }
	    else if (c[0] == '.' && c[1] == '/') {
		c += 2;		/* "./" is the same as "" */
		continue;
	    } else
		break;	/* Didn't find a ".." or "." */
	}

	if (d) {
	    *d = 0;
	    if (port != -1) {
		sprintf(newlink, "%s://%s:%d%s/%s", proto, server, port, path, c);
	    } else {
		sprintf(newlink, "%s://%s%s/%s", proto, server, path, c);
	    }
	    *d = '/';
	} else
	    /* eprint(0, "Help! I'm hopelessly lost...-%s- link -%s-\n", path, link); */
	    /* report error! */
	    ;
    }
    
    free(proto);
    free(server);
    free(path);

    link[len] = save;

    return newlink;
    /* Get the protocol. */
}

int
processAbsoluteLink(struct webw_urlbuf *list, char *newlink)
{
    struct linktype *link;
    char            *c;
    int              len;

    if (list->filter && webw_filter_url(newlink, list))
	return 0;

    len = strlen(newlink);

    /* only "http" for now (no mailto, file, etc) */
    if (len > 5 && strncmp(newlink, "http:", 5))
        return 0;

    if (len > (URL_LIST_LEN - (2 + sizeof(char *)))) {
	ds_thread_log(0, "Got TOO long URL name");
	return 0;
    }

    /* pad to be sure there is room for a null, and round
     * to a word (sizeof(char *)) boundary. */
    len = (len + 2 + sizeof(char *)) & ~(sizeof(char *) - 1);

    /* use of sizeof(int) assumes no space in linktype packing */
    while (len + sizeof(int) > URL_LIST_LEN - list->urlbuf_pos) {
	/* need to wrap to the beginning --
	 * make sure there is enough room, and it's worthwhile to move */
	if (len+sizeof(struct linktype)-sizeof(char *) > list->urlbuf_consumer_pos ||
	    list->urlbuf_pos - list->urlbuf_consumer_pos > URL_LIST_LEN/2) {
	    /* gotta wait... must release locks (so that webw_cursexpand
	     * can get them, for one) */
	    ds_lock_release(&(list->odata->wod_page_data->wpd_data_lock));
	    ds_lock_release(&(list->odata->wod_url_lock));
	    ds_thread_log(100, "waiting for expand consumer!");
	    ds_thread_wait((char *)&(list->urlbuf_consumer_pos));
	    ds_lock_write(&(list->odata->wod_url_lock));
	    ds_lock_read(&(list->odata->wod_page_data->wpd_data_lock));
	}

	ds_thread_log(100, "expand producer back in action!");
	/* now safe to shift it back to the beginning
	 * note: bcopy claims to work with overlapping strings */
	bcopy(list->urlbuf+list->urlbuf_consumer_pos, list->urlbuf,
	      list->urlbuf_pos - list->urlbuf_consumer_pos);
	list->urlbuf_pos = list->urlbuf_pos - list->urlbuf_consumer_pos;
	list->urlbuf_consumer_pos = 0;
    }

    /* Fill in the new entry with the current link. */
    link = (struct linktype *)&list->urlbuf[list->urlbuf_pos];
    link->next = len + sizeof(int); /* see warning above about linktype
				     * packing */
    strcpy(&link->link[0], newlink);
    for (c = &link->link[0]; *c; c++)
      if (*c == '/')
	*c = '\\';		/* convert to Eric format */
    ds_thread_log(100, "added link %s", &link->link[0]);

    /* Update curlink to point to the next/null entry */
    list->urlbuf_pos += link->next;
    
    /* Make that last entry null-terminate the string. */
    ((struct linktype *)&list->urlbuf[list->urlbuf_pos])->next = 0; 
    
    /* WAKEUP(&msg->cond); If someone waiting */
    return 1;
}

int webw_pick_filter(char *url, struct webw_urlbuf *the_list)
{
    char *temp, *rest;
    struct webw_filter *new_filter;

    temp = strstr(url, "lycos");
    if (temp) {
	rest = temp;
	temp = strstr(rest, "query");
	if (temp) {
	    ds_thread_log(0, "filtering lycos search results");

	    /* Looks like we have a lycos query. */
	    ALLOC(new_filter, struct webw_filter);
	    the_list->filter = new_filter;
	    new_filter->type = LYCOS_SEARCH;
	    return 1;
	}
    }
    temp = strstr(url, "altavista");
    if (temp) {
	rest = temp;
	temp = strstr(rest, "query?");
	/* gross hack for DARPA demo XXX -kwalker */
	if (!temp) temp = strstr(rest, "_query_");
	if (temp) {
	    ds_thread_log(0, "filtering altavista search results");

	    /* Looks like we have an altavista query. */
	    ALLOC(new_filter, struct webw_filter);
	    the_list->filter = new_filter;
	    new_filter->type = ALTAVISTA_SEARCH;
	    return 1;
	}
    }
    temp = strstr(url, "test.html");
    if (temp) {
	ALLOC(new_filter, struct webw_filter);
	the_list->filter = new_filter;
	new_filter->type = TEST_PAGE;
	return 1;
    }

    return 0;
}

/* Go through a dbuf chain looking for the given string.  Assumes all
   data is there. */
int webw_find_string(char *text,
		     struct webw_dbuf *start_dbuf,
		     int start_offset,
		     struct webw_dbuf **found_dbuf,
		     int *found_offset)
{
    char carry[MAX_TEXT_LEN*2+1] = "";
    int carry_len = 0, carry_len2;
    struct webw_dbuf *curbuf = start_dbuf, *prevbuf = NULL;
    int pos = start_offset;
    int len = strlen(text);
    char *str_start;
    char *match;

    if (len > MAX_TEXT_LEN)
	return 0;

    while (curbuf) {
	if (carry_len) {
	    /* check carry-over */
	    carry_len2 = MIN(curbuf->wdb_length, (len-carry_len));
	    strncat(carry, curbuf->wdb_data, carry_len2);
	    match = strstr(carry, text);
	    if (match) {
		ds_thread_log(0, "found on boundary");
		if (match - carry < carry_len) {
		    *found_dbuf = prevbuf;
		    *found_offset = prevbuf->wdb_length - (match-carry);
		    return 1;
		}
	    }
	    pos = carry_len2;
	}
	str_start = curbuf->wdb_data+pos;
	match = strstr(str_start, text);
	if (match) {
	    *found_dbuf = curbuf;
	    *found_offset = match - curbuf->wdb_data;
	    return 1;
	} else {
	    if (!curbuf->wdb_next)
		return 0;
	    else {
		/* not found.  check the case of crossing the buffer */
		carry_len = MIN(curbuf->wdb_length, len);
		strcpy(carry, curbuf->wdb_data+curbuf->wdb_length-carry_len);
		prevbuf = curbuf;
		curbuf = curbuf->wdb_next;
	    }
	}
    }

    return 0;
}

/*
 * webw_do_prefetch -- utility routine to start prefetching a page.
 *                     called by wrappers which choose the appropriate
 *                     prefetch_mode
 *                     Note: this indirection allows 2 kinds of prefetching,
 *                     the upshot of which is that a prefetching page never
 *                     goes through w_open with fetch_type NORMAL, so
 *                     fetch_url/worker can tell it wasn't a demand load
 */

int webw_do_prefetch(struct warden *wp,
		     struct viceroy_onode  *op,
		     struct ody_cred  *cred,
		     int            mode,
		     fetch_type     prefetch_mode)
{
    struct webw_onode_data *the_odata;
    struct webw_page_data  *wpdata;
    int err;
    struct stat             sbuf;
    int                     stat_result;

    if (!prefetch)
	/* allow viceroy flag to disable prefetching */
        return EPERM;

    the_odata = (struct webw_onode_data *)op->o_data;
    /* use the "open" mechanism, but signal that we're prefetching */
    the_odata->wod_prefetching = prefetch_mode;
    err = (wp->w_open)(wp, op, cred, mode);
    if (err)
	return err;

    wpdata = the_odata->wod_page_data;

    /* wait until all data is prefetched (since the worker
     * synchronously fetches embedding imgs and frames, this blocks
     * until the page is truly prefetched */
    if (!wpdata->wpd_data_done)
	ds_thread_wait((char *) &(wpdata->wpd_data_done));

    ds_thread_log(0, "openers %d on PFed %s", wpdata->wpd_openers,
		  wpdata->wpd_url);

    /* If the only open reference to this file is us, set it up to be CACHED
     * so reads will redirect to UFS.  We don't want prefetched files being 
     * rdwr through the viceroy! */

    if (wpdata->wpd_openers == 1) {
	/* known to be ON_DISK */
	struct cache_entry *centry = the_odata->wod_centry;
	wpdata->wpd_state = CACHED;
	ds_thread_log(5, "%s became CACHED", wpdata->wpd_url);
	webw_release_page_dbufs(wpdata);
	wpdata->wpd_data = NULL;
	op->o_flags &= ~ON_UNCACHEABLE;
	op->o_cachefile = centry->ce_fname;
	stat_result = stat(op->o_cachefile, &sbuf);
	if (stat_result < 0) {
	    return(stat_result); /* failed stat */
	}	
	op->o_device = sbuf.st_dev;
	op->o_inode  = sbuf.st_ino;
	
    }
    
    return 0;
}

int webw_filter_url(char *url, struct webw_urlbuf *list)
{
    switch(list->filter->type) {
    case ALTAVISTA_SEARCH:
	if (strstr(url, "babelfish.altavista"))
	    return 1;
	break;
    default:
	break;
    }
    return 0;
}

#endif

