#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_tsops.c -- Web Warden TSOP handlers
 *
 * written by Eric Tilton <tilt+@cs.cmu.edu>
 *
 * This file provides handlers for the various Web Warden TSOPs.
 *
 * The currently supported TSOPs are:
 *
 *   WEBW_STAT
 *   WEBW_SETMINQUALITY
 *   WEBW_GETMINQUALITY
 *   WEBW_SETHEADERS
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/queue.h>

#include <include/odytypes.h>
#include <base/ds/ds_thread.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_hwlock.h>
#include <base/ds/ds_rrlist.h>
#include <base/ds/ds_safeq.h>

#include <include/viceroy_onode.h>
#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/warden.h>
#include <include/downcall.h>

#include <lwp/lwp.h>
#include <rpc2/rpc2.h>

#include <base/simplewww/swww.h>

#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_cache.h"
#include "webwarden_comm.h"
#include "webwarden_tsops.h"


extern int do_caching;

int webw_open_by_url(struct warden *wp,
		     struct viceroy_onode  *op,
		     size_t        argsz,
		     void          *argp,
		     size_t        *retsz,
		     void          **retp)
{
    /* XXX not clear this is actually useful. we can use regular
           open to do this more better (love that english).
	   this will probably end up as a library routine that
	   is linked in with the application that turns a URL
	   into a pathname. */

    *retsz = 0;
    *retp = NULL;
    return(EOPNOTSUPP);
}

int webw_stat(struct warden *wp,
	      struct viceroy_onode  *op,
	      size_t        argsz,
	      void          *argp,
	      size_t        *retsz,
	      void          **retp)
{
    /* our strategy here is to look at the metadata we stuffed into
       the onode earlier, which contains a header list. we'll
       take that headerlist and pack it into retp. */
    int                    size = 0;
    char                   *cbookmark;
    int                    *ibookmark;
    struct webw_onode_data *wodp = (struct webw_onode_data *) op->o_data;
    char                   *buffer;

    /* determine the size of buffer to be returned
       (see webwarden_onode.h for struct webw_onode_data, for
        further illumination) */
    ASSERT(sizeof(char) == 1);	/* XXX this is a dangerous assumption
				       I make. needs fixed. */

    size += sizeof(int);	               /* size of status code */
    size += sizeof(int);	               /* size of number of headers */
    size += strlen(wodp->wod_page_data->wpd_status_line) + 1;
				               /* size of status line */
    size += wodp->wod_page_data->wpd_header_bytes;/* size of headers */

    if (size >= *retsz) {
	*retsz = 0;
	*retp = NULL;
	return(E2BIG);		/* XXX too big. is this a reasonable error? */
    }

    /* allocate the return buffer */
    NALLOC(buffer, char, size);

    /* stuff it full of metadata */
    ibookmark = (int *) buffer;
    *ibookmark = wodp->wod_page_data->wpd_status_code;  ibookmark++;
    *ibookmark = wodp->wod_page_data->wpd_header_size;  ibookmark++;

    cbookmark = (char *) ibookmark;
    strcpy(cbookmark, wodp->wod_page_data->wpd_status_line);
    cbookmark += strlen(wodp->wod_page_data->wpd_status_line);
    cbookmark += 1;		/* don't forget the null terminator! */

    memcpy(cbookmark, wodp->wod_page_data->wpd_headers,
	   wodp->wod_page_data->wpd_header_bytes);

    /* finally, set retp to point to the buffer, and retsz to reflect
       the buffer's size */
    *retp  = buffer;
    *retsz = size;

    return(0);
}

int webw_set_min_quality(struct warden *wp,
			 struct viceroy_onode  *op,
			 size_t        argsz,
			 void          *argp,
			 size_t        *retsz,
			 void          **retp)
{
    char             *mime_type;
    webw_quality_t   *qp, *temp;
    struct webw_data *my_data;

    my_data = (struct webw_data *) wp->w_data;

    /* Minimal sanity check */
    if (argsz < sizeof(webw_quality_t) + 2) {
	*retsz = 0;
	*retp = NULL;
	return(EINVAL);
    }

    qp = (webw_quality_t *) argp;
    temp = qp + 1;

    mime_type = (char *) temp;

    ds_thread_log(1000, "webw_set_min_quality: mime type is %s", mime_type);

    /* make sure that the MIME type requested is one we handle */
    if (strcmp("*/*", mime_type) == 0) {
	my_data->wwd_img_quality  = *qp;
	my_data->wwd_html_quality = *qp;
	ds_thread_log(1000, "webw_set_min_quality: set quality for %s to %d",
		      mime_type, my_data->wwd_img_quality);
    } else if ( (strcmp("image/*", mime_type) == 0) ||
		(strcmp("image/gif", mime_type) == 0) ) {
	my_data->wwd_img_quality = *qp;
	ds_thread_log(1000, "webw_set_min_quality: set quality for %s to %d",
		      mime_type, my_data->wwd_img_quality);
    } else if ( (strcmp("text/html", mime_type) == 0) ) {
	my_data->wwd_html_quality = *qp;
	ds_thread_log(1000, "webw_set_min_quality: set quality for %s to %d",
		      mime_type, my_data->wwd_html_quality);
    } else {
	*retsz = 0;
	*retp = NULL;
	return(EOPNOTSUPP);
    }

    return(0);
}

int webw_get_min_quality(struct warden *wp,
			 struct viceroy_onode  *op,
			 size_t        argsz,
			 void          *argp,
			 size_t        *retsz,
			 void          **retp)
{
    char             *mime_type;
    struct webw_data *my_data;

    my_data = (struct webw_data *) wp->w_data;

    /* Minimal sanity check */
    if (*retsz != sizeof(webw_quality_t)) {
	*retsz = 0;
	*retp = NULL;
	return(EINVAL);
    }

    mime_type = (char *) argp;

    ds_thread_log(1000, "webw_get_min_quality: mime type is %s", mime_type);

    /* make sure that the MIME type requested is one we handle */
    if ( (strcmp("image/*", mime_type) == 0) ||
		(strcmp("image/gif", mime_type) == 0) ) {
	*retp = &(my_data->wwd_img_quality);
    } else if ( (strcmp("text/html", mime_type) == 0) ) {
	*retp = &(my_data->wwd_html_quality);
    } else {
	*retsz = 0;
	*retp = NULL;
	return(EOPNOTSUPP);
    }

    return(0);
}

int webw_set_max_time(struct warden *wp,
		      struct viceroy_onode  *op,
		      size_t        argsz,
		      void          *argp,
		      size_t        *retsz,
		      void          **retp)
{
    /* XXX not implemented */
    *retsz = 0;
    *retp = NULL;
    return(EOPNOTSUPP);
}

int webw_invalidate_page(struct warden *wp,
			 struct viceroy_onode  *op,
			 size_t        argsz,
			 void          *argp,
			 size_t        *retsz,
			 void          **retp)
{
    struct webw_onode_data *wodata = (struct webw_onode_data *) op->o_data;

    *retsz = 0;
    *retp = NULL;

    /* XXX locking could be needed here if we were using a truly pre-emptive
       threading package, since it's possible these pointers might
       change out from under us */

    /* mark the pagedata as suspect; all of the real work that results
       from this happens in webw_open() */

    wodata->wod_page_data->wpd_suspect = SUSPECT;

    return(0);
}

/*
 * webw_flush_cache -- flush all cached web objects
 *                     XXX for now, allow reuse of onodes/ofid.  This is
 *                     because we can't flush the namecache without deadlock.
 *                     Maybe it's OK anyway...
 */
int webw_flush_cache(struct warden *wp,
		     struct viceroy_onode  *op,
		     size_t        argsz,
		     void          *argp,
		     size_t        *retsz,
		     void          **retp)
{
    /* for now, the op is unused -- it is always '/ody/www/', since
       this tsop doesn't support flushing of arbitrary URLs */
    int result;
  
    *retsz = 0;
    *retp = NULL;

    if (!do_caching)
	return(0);

    /* flush_nc();   AVOID THIS -- leads to fsync deadlock! */
    result = webw_cache_flush();

    return(result);
}
