#ifndef _ODYBLURB_
#define _ODYBLURB_
/*
 *          Odyssey: an Experimental Platform for Mobile Computing
 *  
 *            Copyright (c) 1997-2000 Carnegie Mellon University
 *                   Additional copyrights listed below
 * 
 *  This  code  is  distributed "AS IS" without warranty of any kind under
 *  the terms of the GNU General Public Licence Version 2, as shown in the
 *  file  LICENSE.  The  technical and financial  contributors to Odyssey
 *  are listed in the file CREDITS.
 * 
 *                         Additional copyrights
 *                            none currently
 */
#endif /* _ODYBLURB_ */

/*
 * webwarden_worker.c -- support for web warden's worker threads
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/queue.h>

#include <base/ds/ds_thread.h>
#include <base/ds/ds_log.h>
#include <base/ds/ds_safeq.h>
#include <base/ds/ds_hash.h>
#include <base/ds/ds_histo.h>
#include <base/ds/ds_rrlist.h>

#include <lwp/lwp.h>
#include <lwp/timer.h>
#include <rpc2/rpc2.h>
#include <rpc2/se.h>
#include <rpc2/sftp.h>

#include <include/odyssey.h>
#include <include/odytypes.h>
#include <include/viceroy_onode.h>

#include <include/cm_rpc2.h>
#include <include/cm.h>
#include <include/onode-manager.h>
#include <include/warden.h>
#include <include/cache_mgr.h>

#include <base/simplewww/swww.h>

#include "distiller.h"
#include "webwarden.h"
#include "webwarden_private.h"
#include "webwarden_onode.h"
#include "webwarden_instr.h"
#include "webwarden_worker.h"
#include "webwarden_sets.h"
#include "webwarden_cache.h"

#include "odyssey/viceroy/rsrc_en.h" /* XXXAPM */

#define WEBW_BUFLENGTH 65536

extern int do_caching;

/* main loop for worker threads. the worker threads are responsible
   for grabbing buffers from the server */
int webw_worker_shell(void *pp)
{
    struct webw_worker_params  *params = (struct webw_worker_params *) pp;
    struct webw_worker_request *req;
    char                       *buffer;

    ds_thread_log(10, "Launched Web Warden worker #%d", params->wwp_num);

    NALLOC(buffer, char, WEBW_BUFLENGTH);

    while(1) {
	/*
	 * wait for a request to come in from webw_fetch_url. this
	 * request will contain an onode pointer (to a "gotten" onode
	 * which we must later put back) and a connection ID to use for
	 * fetching buffers.
	 */
	ds_thread_log(5, "waiting for message");
	req = (struct webw_worker_request *) ds_safeq_deq(params->wwp_workq);
	switch (req->wwr_reqtype) {
	case WEBW_REQ_FETCH:
	    webw_worker_fetch(params, req, buffer);
	    break;
	case WEBW_REQ_PARSE:
	    webw_worker_parse(req);
	    break;
	}
    }
}

void webw_worker_fetch(struct webw_worker_params  *params,
		       struct webw_worker_request *req,
		       char                       *buffer)
{
#ifdef INSTRUMENTED
    struct webw_histos         *histos = params->wwp_data->wwd_histos;
#endif /* INSTRUMENTED */
    struct ofid                 the_ofid;
    struct webw_onode_data     *the_odata;
    struct webw_page_data      *wpdata;
    RPC2_Integer                end_of_object;
    struct webw_dbuf           *dbp=NULL;
    struct webw_dbuf          **lastdbp;
    SE_Descriptor               sed;
    int                         reqid;
    int                         rc;
    cmgr_connection_t            *connection;
#ifdef INSTRUMENTED
    struct timeval              bufstart, bufend, buftime;
    struct timeval              objend, objtime;
    double                      time;
#endif /* INSTRUMENTED */
#ifdef SETS
    int                         cursor = 0;
    int                         prefetching = 0;
#endif
    int                         bufsused = 0;

#ifdef INSTRUMENTED
    bufsused = 0;
#endif /* INSTRUMENTED */

    reqid     = req->wwr_reqid;
    the_ofid  = req->wwr_op->o_fid;
    the_odata = (struct webw_onode_data *) req->wwr_op->o_data;
    wpdata    = (struct webw_page_data *) the_odata->wod_page_data;
    
#ifdef SETS
    cursor = req->wwr_op->o_flags & ON_CURSOR;
    prefetching = the_odata->wod_prefetching == PREFETCH_CONTENTS &&
	wpdata->wpd_status_code < 300;
    if (the_odata->wod_prefetching == PREFETCH_CONTENTS) {
	ds_thread_log(5, "PF %s", wpdata->wpd_url);
    }

    if (cursor) {
	/* we well be collecting a list of embedded urls. */
	NALLOC(the_odata->wod_url_list->urlbuf, char, URL_LIST_LEN);
    }
#endif
    
    lastdbp = &(wpdata->wpd_data);
    ASSERT(*lastdbp == NULL);
    
    ds_thread_log(100, "web worker %d: fetching %d.%d.%d.%d",
		  params->wwp_num,
		  the_ofid.ofid_ttype, the_ofid.ofid_tome,
		  the_ofid.ofid_vnum, the_ofid.ofid_uniq);
    
    end_of_object = 0;
    while(!end_of_object) {
	int rc;
	
	/* set up SFTP */
	setupVMSE(&sed, buffer, WEBW_BUFLENGTH);
	
	if (!(cmgr_isvalid(req->wwr_cid) > 0)) {
	    rc = RPC2_NOCONNECTION;
	} else {
	    connection = cmgr_get(req->wwr_cid, CMGR_LOCK);
	    ASSERT(connection);
	    
	    /* perform the rpc */
#ifdef INSTRUMENTED
	    gettimeofday(&bufstart, NULL);
#endif /* INSTRUMENTED */
	    rc = fetch_buffer(CMGR_RPC2_HANDLE(connection), reqid,
			      &end_of_object, &sed);
	    cmgr_put(req->wwr_cid, CMGR_LOCK);
#ifdef INSTRUMENTED
	    gettimeofday(&bufend, NULL);
#endif /* INSTRUMENTED */
	}
	
	if(rc != RPC2_SUCCESS) {
	    ds_thread_log(10, "web worker %d: fetch_buffer failed (%s)",
			  params->wwp_num,
			  RPC2_ErrorMsg(rc));
	    
	    if (rc < RPC2_FLIMIT) {
		ds_thread_log(11, "web worker %d: RPC2 failure was fatal.",
			      params->wwp_num);
	    }
	    
	    cmgr_markdown(req->wwr_cid);
	} else {
	    /* connection is OK */
	    ds_thread_log(1000, "web worker %d: "
			  "fetch_buffer returned %d bytes",
			  params->wwp_num,
			  (int)
			  sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen);
#ifdef INSTRUMENTED
	    UpdateHisto(&(histos->wh_buf_sizes),
			sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen
			/ 1024.0);
	    timersub(&bufend, &bufstart, &buftime);
	    time = buftime.tv_sec + (buftime.tv_usec / 1000000.0);
	    UpdateHisto(&(histos->wh_buf_times),
			time);
	    
	    if ( (time * 1024.0) != 0) {
		/* make sure this is a real number */
		UpdateHisto(&(histos->wh_buf_speeds),
			    sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen /
			    (time * 1024.0)); /* in KB/sec */
	    }
	    
#endif /* INSTRUMENTED */
	    bufsused++;
	}
	
	if (rc < 0) {
	    ds_thread_log(10, "web worker %d: "
			  "error %d in fetching %d.%d.%d.%d",
			  params->wwp_num,
			  rc,
			  the_ofid.ofid_ttype, the_ofid.ofid_tome,
			  the_ofid.ofid_vnum, the_ofid.ofid_uniq);
	    end_of_object = 1;
	}
	
	/* place the buffer into a dbuf */
	ALLOC(dbp, struct webw_dbuf);
	ASSERT(dbp);
	WEBW_INIT_DBUF(dbp);
	
	if (rc >= 0) {
	    dbp->wdb_length =
		sed.Value.SmartFTPD.FileInfo.ByAddr.vmfile.SeqLen;
	    NALLOC(dbp->wdb_data, char, dbp->wdb_length);
	    ASSERT(dbp->wdb_data);
	    memcpy(dbp->wdb_data, buffer, dbp->wdb_length);
	} else {
	    /* there was an error fetching the buffer */
	    dbp->wdb_length = strlen(WEBW_RPC2_ERROR_MESSAGE);
	    NALLOC(dbp->wdb_data, char, dbp->wdb_length);
	    memcpy(dbp->wdb_data, WEBW_RPC2_ERROR_MESSAGE, 
		   dbp->wdb_length);
	}
	
	/* once the dbuf is available, make it visible */
	ds_lock_write(&(wpdata->wpd_data_lock));
	*lastdbp = dbp;
	ds_thread_schedule((char *) lastdbp);
	lastdbp = &(dbp->wdb_next);
	wpdata->wpd_data_avail += dbp->wdb_length;
	ds_lock_release(&(wpdata->wpd_data_lock));
	
#if 0 /* SETS */
	/* XXX do I want to test rc here?  I think not, since we
	 *     may still have valid data in dbufs not yet parsed.
	 *     However, if I see deadlock in parsehtml, I need to
	 *     look here. (kwalker) */
	if (cursor) {
	    /* If filtering, wait til all data is collected.  It's
               probably a small price to pay for the simplicity. */
	    if (!the_odata->wod_url_list->filter) {
		/* this is a cursor, parse for "members" -- embedded
		 * urls */
		ds_lock_write(&(the_odata->wod_url_lock));
		
		ds_lock_read(&(wpdata->wpd_data_lock));
		/* do the parsing!  update the url_list! */
		webw_parsehtml(wpdata->wpd_url, the_odata->wod_url_list,
			       wpdata->wpd_wp, PARSE_LINKS);
		ds_lock_release(&(wpdata->wpd_data_lock));
		
		ds_thread_log(10, "Parsed them urls!");
		ds_lock_release(&(the_odata->wod_url_lock));
	    }
	} else if (prefetching) {
	    /* this is a page being prefetched -- look for
	     * inline images, frames, etc.  if this page is
	     * simply pointing us at another page, don't waste
	     * time here, since Netscape won't show us this
	     * page anyway. */
	    ds_lock_read(&(wpdata->wpd_data_lock));
	    webw_parsehtml(wpdata->wpd_url, the_odata->wod_url_list,
			   wpdata->wpd_wp, PARSE_IMGS);
	    ds_lock_release(&(wpdata->wpd_data_lock));
	}
#endif
#if 0
	/* if no one has this onode open anymore, stop reading it */
	/* XXX make sure this works! */
	/* XXX won't make sense once we have real caching! */
	/* XXXXXX WHAT IS/WAS THIS FOR??? XXXXXX */
	if (req->wwr_op->o_ocount == 0) {
	    ds_thread_log(5, "fetching but no openers!");
	    end_of_object = 1;
	}
#endif
    } /* end while(!end_of_object) */

#ifdef INSTRUMENTED
    UpdateHisto(&(histos->wh_buf_perobject), bufsused);
    
    UpdateHisto(&(histos->wh_obj_sizes),
		(wpdata->wpd_data_avail) / 1024.0);
    
    gettimeofday(&objend, NULL);
    timersub(&objend, &(req->wwr_fetchstart), &objtime);
    time = buftime.tv_sec + (buftime.tv_usec / 1000000.0);
    
    UpdateHisto(&(histos->wh_obj_times),
		time);
    
    if ( (time * 1024.0) != 0) {
	/* make sure this is a real number */
	UpdateHisto(&(histos->wh_obj_speeds),
		    wpdata->wpd_data_avail /
		    (time * 1024.0)); /* in KB/sec */
    }
#endif /* INSTRUMENTED */
    
    ds_thread_log(0, "fetched into %d bufs\n", bufsused);
#ifdef SETS
    if ((cursor || prefetching) && bufsused > 1) {
	ds_lock_write(&(wpdata->wpd_data_lock));
	ds_thread_log(0, "collapsing dbufs on %s", wpdata->wpd_url);
	webw_collapse_dbufs(wpdata);
	ds_lock_release(&(wpdata->wpd_data_lock));
    }

    if (cursor) {
	/* If filtering, wait til all data is collected.  It's
	   probably a small price to pay for the simplicity. */
	if (!the_odata->wod_url_list->filter) {
	    /* this is a cursor, parse for "members" -- embedded
	     * urls */
	    ds_lock_write(&(the_odata->wod_url_lock));
	    
	    ds_lock_read(&(wpdata->wpd_data_lock));
	    /* do the parsing!  update the url_list! */
	    webw_parsehtml(wpdata->wpd_url, the_odata->wod_url_list,
			   wpdata->wpd_wp, PARSE_LINKS);
	    ds_lock_release(&(wpdata->wpd_data_lock));
	    
	    ds_thread_log(10, "Parsed them urls!");
	    ds_lock_release(&(the_odata->wod_url_lock));
	}
    } else if (prefetching) {
	/* this is a page being prefetched -- look for
	 * inline images, frames, etc.  if this page is
	 * simply pointing us at another page, don't waste
	 * time here, since Netscape won't show us this
	 * page anyway. */
	ds_lock_read(&(wpdata->wpd_data_lock));
	webw_parsehtml(wpdata->wpd_url, the_odata->wod_url_list,
		       wpdata->wpd_wp, PARSE_IMGS);
	ds_lock_release(&(wpdata->wpd_data_lock));
    }
#endif

    if (cmgr_isvalid(req->wwr_cid) > 0) {
	connection = cmgr_get(req->wwr_cid, CMGR_LOCK);
	ASSERT(connection);
	rc = close_url(CMGR_RPC2_HANDLE(connection), reqid);
	cmgr_put(req->wwr_cid, CMGR_LOCK);
    } else {
	ds_thread_log(1, "web worker %d: couldn't close url (no conn.)",
		      params->wwp_num);
	rc = RPC2_NOCONNECTION;
    }
    
    if(rc != RPC2_SUCCESS) {
	ds_thread_log(10, "web worker %d: close_url failed (%s)",
		      params->wwp_num,
		      RPC2_ErrorMsg(rc));
	
	if (rc < RPC2_FLIMIT) {
	    ds_thread_log(1, "web worker %d: RPC2 failure was fatal.",
			  params->wwp_num);
	}
	
	cmgr_markdown(req->wwr_cid);
    }
    
    if (do_caching && the_odata->wod_centry) {
	/* dump to disk before marking data_done, make the onode point
	 * to the file */
	struct cache_entry *centry = the_odata->wod_centry;
	int fd;
	
	/* note: these flags are ok for now - only used within viceroy */
	fd = cache_open(centry, O_WRONLY | O_CREAT | O_TRUNC);
	webw_dump_dbuf(wpdata->wpd_data, fd);
	rc = cache_close(centry, fd, 1);
	
	if (rc == ENOSPC) {
	    /* didn't succeed in putting on disk, might want to
	       request more space.  Need to continue to access the
	       memory copy. */
	    wpdata->wpd_state = FETCHED;
	} else {
	    wpdata->wpd_state = ON_DISK;
	}
    } else {
	wpdata->wpd_state = FETCHED;
    }
    dbp->wdb_last = 1;		/* seal off dbuf */
    wpdata->wpd_data_done = 1;  /* mark fetch as done */
#ifdef SETS
    the_odata->wod_prefetching = NORMAL; /* reset this for future reloads */

    if (cursor && the_odata->wod_url_list->filter) {
	struct webw_worker_request *parse_req;

	/* tell a worker to do the filtered parsehtml */
	ALLOC(parse_req, struct webw_worker_request);
	parse_req->wwr_op = onode_get(&the_ofid);
	if (parse_req->wwr_op == NULL)
	  ASSERT(0);		/* this should never happen,
				   but just in case */
	parse_req->wwr_cid        = 0;    /* unused by PARSE handler */
	parse_req->wwr_reqid      = 0;    /* unused by PARSE handler */
	parse_req->wwr_reqtype    = WEBW_REQ_PARSE;
	ds_safeq_enq(params->wwp_workq, parse_req);
    } else {
	the_odata->wod_url_list->parse_done = 1;
    }

    /* this page is pointing us at a different url -- go ahead and
     * prefetch that one.  The VIRGIN test checks that we are the *
     * first to try to fetch it.  This way, if Netscape tries to * get
     * it before we have a chance, the demand fetch will do * the
     * job. */
    if (wpdata->wpd_target) {
	struct webw_page_data *bounce_wpdata = ((struct webw_onode_data *)wpdata->wpd_target->o_data)->wod_page_data;

	onode_writelock(wpdata->wpd_target);
	if (bounce_wpdata->wpd_state == VIRGIN) {
	    /* NOTE: if this happens within a frame, we could run into
	     * deadlock since we are consuming more prefetch workers
	     * */
	    
	    ds_thread_log(10, "got the chance to bounce");
	    
	    onode_get(&wpdata->wpd_target->o_fid);
	    rc = ((wpdata->wpd_wp)->w_prefetch)(wpdata->wpd_wp,
						wpdata->wpd_target, 
						NULL, O_O_READ);
	    onode_release_write(wpdata->wpd_target);
	    /* call close so the dump to disk occurs maybe pull the
	     * dumping code out of close and call directly */
	    if (!bounce_wpdata->wpd_data_done)
		ds_thread_wait((char *) &bounce_wpdata->wpd_data_done);
	    rc = ((wpdata->wpd_wp)->w_close)(wpdata->wpd_wp,
					     wpdata->wpd_target,
					     NULL,
					     O_O_READ);
	    onode_put(wpdata->wpd_target); /* leave only the effect on the cache */
	} else
	    onode_release_write(wpdata->wpd_target);
    }

    if (the_odata->wod_prefetching != NORMAL)
	ds_thread_log(5, "prefetched: %s", wpdata->wpd_url);
    else
#endif
	ds_thread_log(5, "fetched: %s", wpdata->wpd_url);
    
    ds_thread_schedule((char *) &(dbp->wdb_next));
    ds_thread_wakeup((char *) &(wpdata->wpd_data_done));
    
    onode_put(req->wwr_op);        /* put back the onode */
    FREE(req);                     /* free the request */
}

void webw_worker_parse(struct webw_worker_request *req)
{
#ifdef SETS
    struct webw_onode_data *the_odata = (struct webw_onode_data *)req->wwr_op->o_data;
    struct webw_page_data *wpdata = the_odata->wod_page_data;


    ds_thread_log(3, "parsehtml on formerly non-cursor onode");

    ds_lock_write(&(the_odata->wod_url_lock));
    ds_lock_read(&(wpdata->wpd_data_lock));
    if (the_odata->wod_url_list->filter) {
	/* do the filtered parsehtml */
	switch (the_odata->wod_url_list->filter->type) {
	case LYCOS_SEARCH:
	    ds_thread_log(0, "Trying Lycos filter\n");
	    webw_find_string("relevant results",
			     wpdata->wpd_data,
			     0,
			     the_odata->wod_url_list->cur_dbuf,
			     &the_odata->wod_url_list->cur_dbuf_pos);
	    webw_find_string("Previous Page",
			     wpdata->wpd_data,
			     0,
			     &the_odata->wod_url_list->filter->stop_dbuf,
			     &the_odata->wod_url_list->filter->stop_offset);
	    break;
	case ALTAVISTA_SEARCH:
	    ds_thread_log(0, "Trying AltaVista filter\n");
	    if (!webw_find_string("Real Name System",
				  wpdata->wpd_data,
				  0,
				  the_odata->wod_url_list->cur_dbuf,
				  &the_odata->wod_url_list->cur_dbuf_pos)) {
		ds_thread_log(0, "Not first results page...\n");
		if (!webw_find_string("matches were found",
				      wpdata->wpd_data,
				      0,
				      the_odata->wod_url_list->cur_dbuf,
				      &the_odata->wod_url_list->cur_dbuf_pos))
		    ds_thread_log(0, "Not a results page!\n");
	    }
	    webw_find_string("Pages: ",
			     wpdata->wpd_data,
			     0,
			     &the_odata->wod_url_list->filter->stop_dbuf,
			     &the_odata->wod_url_list->filter->stop_offset);
	    break;
	case TEST_PAGE:
	    webw_find_string("start here",
			     wpdata->wpd_data,
			     0,
			     the_odata->wod_url_list->cur_dbuf,
			     &the_odata->wod_url_list->cur_dbuf_pos);
	    webw_find_string("end here",
			     wpdata->wpd_data,
			     0,
			     &the_odata->wod_url_list->filter->stop_dbuf,
			     &the_odata->wod_url_list->filter->stop_offset);
	    break;
        }
    }
    /* do the parsing!  update the url_list! */
    webw_parsehtml(wpdata->wpd_url, the_odata->wod_url_list,
		   wpdata->wpd_wp, PARSE_LINKS);
    the_odata->wod_url_list->parse_done = 1;
    ds_lock_release(&(wpdata->wpd_data_lock));
    ds_lock_release(&(the_odata->wod_url_lock));
    onode_put(req->wwr_op);

    ds_thread_log(3, "parsehtml finished on formerly non-cursor onode");
#endif
}
