#!/usr/local/bin/perl

# BEGIN ODYSSEY COPYRIGHT

#          Odyssey: an Experimental Platform for Mobile Computing
#  
#            Copyright (c) 1997-2000 Carnegie Mellon University
#                   Additional copyrights listed below
# 
#  This  code  is  distributed "AS IS" without warranty of any kind under
#  the terms of the GNU General Public Licence Version 2, as shown in the
#  file  LICENSE.  The  technical and financial  contributors to Odyssey
#  are listed in the file CREDITS.
# 
#                         Additional copyrights
#                            none currently

# END ODYSSEY COPYRIGHT


# Usage:

# blurb.pl <options> [<filename>]* [<directories>]*
#
# where options include:
#     -r   recursively blurb files in all subdirectories of the
#          specified directories
#     -c   use this config file ("blurb.config" is default)
#     -b   use this blurb text ("blurb.text" is the default)
#     -u   update blurbs (i.e. don't modify unblurbed files)
#
# blurb.pl generates a list of candidate files for blurbing, and
# then adds the blurb text (typically a copyright statement) to those
# candidate files. The list of candidate files is generated in two
# ways:
#
#   1) all files specified directly on the command line
#   2) all files in directories that are specified on the command line,
#      given that they have the appropriate extension.
#
# A config file can be specified which contains a list of directories
# to be ignored when recursively generating a list of directories to
# check. (This is intended for source code directories which include
# code from other sources, if it has its own copyright notice.) The
# config also includes a list of extensions to check.

require "getopts.pl";

######################################################
# configuration
######################################################

# descriptions of how to treat various suffixes
#
# for a given suffix, there are three strings -- prefix, suffix, and comment.
# the prefix and suffix are placed before and after (respectively) the
# blurb text. The comment text is placed before each line of the blurb
# text.
#
# If there is no configuration information for a suffix, blurb will fail
# and complain.

# C and C++ files (.c, .cc and .h)

$comments{".c_prefix"} =
    "#ifndef _ODYBLURB_\n" .
    "#define _ODYBLURB_\n" .
    "/*\n";

$comments{".c_comment"} =
    " * ";

$comments{".c_suffix"} =
    " */\n" .
    "#endif /* _ODYBLURB_ */\n";

$comments{".cc_prefix"}  = $comments{".c_prefix"};
$comments{".cc_comment"} = $comments{".c_comment"};
$comments{".cc_suffix"}  = $comments{".c_suffix"};

$comments{".h_prefix"}  = $comments{".c_prefix"};
$comments{".h_comment"} = $comments{".c_comment"};
$comments{".h_suffix"}  = $comments{".c_suffix"};

# TCL files (.tcl)

$comments{".tcl_prefix"} =
    "# BEGIN ODYSSEY COPYRIGHT\n\n";

$comments{".tcl_comment"} =
    "# ";

$comments{".tcl_suffix"} =
    "\n# END ODYSSEY COPYRIGHT\n";

# Perl files (.pl)

$comments{".pl_prefix"} =
    "# BEGIN ODYSSEY COPYRIGHT\n\n";

$comments{".pl_comment"} =
    "# ";

$comments{".pl_suffix"} =
    "\n# END ODYSSEY COPYRIGHT\n";

# defaults for command line options

@extensions = (".c", ".cc", ".h", ".tcl", ".pl"); # default file exts
$update     = 0;
$recurse    = 0;
$configfile = "blurb.config";
$blurbfile  = "blurb.text";

######################################################
# subroutines
######################################################

# &getextension($file)
# returns the extension of the file (everything after the last dot)
sub getextension {
    local($file) = @_;
    local($pos);

    # find the last ., and return the string which follows
    $pos = rindex($file, ".");
    if ($pos == $[ - 1) {
	return "";
    }

    return substr($file, $pos);
}

# &check($file, @extensions);
# checks if a file has a valid extension
sub check {
    local($file, @extensions) = @_;
    local($ex, $myex);

    $myex = &getextension($file);

    for $ex (@extensions) {
	if ($ex eq $myex) { return 1; }
    }

    return 0;
}

# &multiline_eq($multiline_string, $index, @contents)
sub multiline_eq {
    local($multiline_string, $index, @contents) = @_;
    local(@multi);
    local($loop);

    @multi = split("\n", $multiline_string);

    for ($loop = 0; $loop < @multi; $loop++) {
	if ($multi[$loop] ne $contents[$index + $loop]) {
	    return 0;
	}
    }

    return 1;
}

# &blurb($file)
sub blurb {
    local($file) = @_;
    local($newfile, $fileext, $line);
    local($filedir, $filename, $pos);
    local(@contents);
    local($blurbstart, $blurbend, $loop);
    local($startlen, $endlen);
    local($spliceoffset, $splicelength);
    local($prefix, $suffix, $comment);
    local(@thisblurb);

    print "blurbing $file ... ";

    # read file into memory
    $pos = rindex($file, "/");
    if ($pos == $[ - 1) {
	$newfile = "blurbed-$file-.$$";
    } else {
	$filedir  = substr($file, 0, $pos);
	$filename = substr($file, $pos + 1);
	$newfile = "$filedir/blurbed-$filename-.$$";
    }
    $fileext = &getextension($file);

    $prefix  = $comments{$fileext . "_prefix"};
    $suffix  = $comments{$fileext . "_suffix"};
    $comment = $comments{$fileext . "_comment"};

    open(FILE, $file);

    $blurbstart = -1;
    $blurbend   = -1;

    while(<FILE>) {
	chop;
	push(@contents, $_);
    }

    close(FILE);

    # check file for blurb start and end

    $startlen = split("\n", $prefix);
    $endlen   = split("\n", $suffix);

    for ($loop = 0; $loop < @contents; $loop++) {
	# iterate through the loop
	if (&multiline_eq($prefix,
			  $loop,
			  @contents))
	{
	    if ($blurbstart != -1) {
		print "ERROR: found more than one blurb start.\n";
		exit -1;
	    }
	    $blurbstart = $loop;
	}

	if (&multiline_eq($suffix,
			  $loop,
			  @contents))
	{
	    if ($blurbend != -1) {
		print "ERROR: found more than one blurb end.\n";
		exit -1;
	    }
	    $blurbend = $loop;
	}

    }

    # sanity checking
    if (($blurbstart * $blurbend) < 0) {
	# this means a blurb start was found, but not a blurb end
	# (i.e., only one was -1)
	print " ERROR: found only one of blurb start or blurb end.\n";
	exit -1;
    }

    if ($blurbend < $blurbstart) {
	print " ERROR: blurb end precedes blurb start.\n";
	exit -1;
    }

    if ($blurbstart == -1 && $blurbend == -1) {
	if ($update) {
	    print " no blurb, not updating\n";
	    return;
	}
	# decide where to put the blurb. first, check for a magic
	# number. if there is a magic number, place the blurb one
	# blank line after the magic number. Otherwise, the blurb goes
	# at the start of the file.
	if ($contents[0] =~ /^\#\!/) {
	    # there is a magic number
	    splice(@contents, 1, 0, ""); # insert a blank line
	    $spliceoffset = 2;
	    $splicelength = 0;
	} else {
	    # there is no magic number
	    $spliceoffset = 0;
	    $splicelength = 0;
	}
    } else {
	# set up sliceoffest and splicelength based on blurbstart
	# and blurbend
	$spliceoffset = $blurbstart;
	$splicelength = $blurbend - $blurbstart + $endlen + 1;
    }
    
    # construct the blurb
    push(@thisblurb, $prefix);
    for ($loop = 0; $loop < @blurb; $loop++) {
	push(@thisblurb, $comment . $blurb[$loop]);
    }
    push(@thisblurb, $suffix);
    push(@thisblurb, "");

    # place the blurb via splice
    splice(@contents, $spliceoffset, $splicelength, @thisblurb);

    # output @contents to NEWFILE

    open(NEWFILE, ">$newfile");

    for ($loop = 0; $loop < @contents; $loop++) {
	$line = $contents[$loop];
	print NEWFILE $line;
	if (substr($line, -1, 1) ne "\n") {
	    print NEWFILE "\n";
	}
    }
    
    # clean up, and replace the old with the new
    close(NEWFILE);

    # when we believe this works
    rename($newfile, $file);
    print "done\n";
}

######################################################
# main program
######################################################

# parse command line options
&Getopts('c:b:ru');

if ($opt_c) { $configfile = $opt_c; }
if ($opt_b) { $blurbfile  = $opt_b; }
if ($opt_r) { $recurse    = 1; }
if ($opt_u) { $update     = 1; }

# parse config file

open(CONFIG, $configfile);

while(<CONFIG>) {
    chop;
    if ($_ =~ /^\#/) {next;}

    if ($_ eq "--EXTENSIONS--") { $state = 1; undef @extensions; next; }
    if ($_ eq "--IGNORE--")     { $state = 2; next; }

    if ($_) {
	if ($state == 1) {
	    # parse this line as an extension to add to the list
	    push (@extensions, $_);
	} elsif ($state == 2) {
	    # parse this line a directory to ignore
	    push (@ignore, $_);
	}
    }
}

close(CONFIG);

# read in blurb text

open(BLURB, $blurbfile);

while(<BLURB>) {
    push (@blurb, $_);
}

close(BLURB);

# generate list of candidate files

# start by sorting files on command line into directories and files

for $file (@ARGV) {
    if (-d $file) {
	unshift(@dirs, $file);
    } else {
	$result = &check($file, @extensions);
	if ($result == 0) {
	    print "$file doesn't match any of @extensions.\n";
	    exit(-1);
	}
	unshift(@files, $file);
    }
}

# now go through all directories and add appropriate files

SEARCH:
while($dir = shift(@dirs)) {
    for $skip (@ignore) {
	if ($dir =~ /$skip/) { next SEARCH; }
    }

    opendir(THISDIR, $dir);
    @dirfiles = grep(!/^\.\.?$/, readdir(THISDIR));
    closedir(THISDIR);

    for $file (@dirfiles) {
	$file = $dir . "/" . $file;
	if (-d $file) {
	    if ($recurse) {
		unshift(@dirs, $file);
	    }
	} else {
	    if (&check($file, @extensions)) {
		unshift(@files, $file);
	    }
	}
    }
}

# at this point, we have all files, so blurb 'em

for $file (@files) {
    &blurb($file);
}
